﻿unit SelfTest;

{ **************************************************************************** }
{ Rejbrand AlgoSim Automatic Tests                                             }
{ Copyright © 2019-2020 Andreas Rejbrand                                       }
{ https://english.rejbrand.se/                                                 }
{ **************************************************************************** }

{$DEFINE QuickTest}

interface

uses
  Windows, SysUtils, Types, Classes, Graphics, ASNum, ASKernelDefs, ASKernel,
  ASExpression, ASObjects, ASStructs, ASObjStore, ASPropMan, ASPropStores,
  ASFcnMgr, ASFcnExp, ASExecutionContext, ASFormatters, ASFunctions,
  ASTokenizer, ASParser, ASStrFcns, Generics.Defaults, Generics.Collections,
  ASColors, ASPixmap, ASSounds, UITypes;

type
  TSelfTestProgressEvent = procedure(ATotal, APerformed, ASucceeded,
    AFailed: Integer) of object;
  TSelfTestChapterEvent = procedure(AChapterNumber: Integer;
    const AChapterName: string) of object;

procedure Run(AKernel: TASKernel; AProgressEvent: TSelfTestProgressEvent = nil;
  AChapterEvent: TSelfTestChapterEvent = nil);

implementation

uses
  Math, StrUtils, DateUtils, Character, Clipbrd, MatLits, Registry, IOUtils,
  PngImage, ASMidi, SelfTestUI;

const
  TotalCount = 495194;

{$REGION 'Infrastructure'}

var
  ElapsedTime: Double;

type
  TSelfTester = class
  strict private
  type
    TTestItem = record
      Expr: string;
      Expected, Actual: string;
      constructor Create(const AExpr, AExpected, AActual: string);
      function ToString: string;
    end;
  const
    RealMatrixNames = AnsiString('ABCDEFGHIJKLMNOPQRSTUVWXYZÅÄÖabcdfghjklmnopqrstuvwxyzåä');
    ComplexMatrixNames = AnsiString('ABCDEFGHIJKLMNOPQRSTUVWXYZÅÄÖabcdfghjklmnopqrstuvwxyzåäöüë');
    CharSamples = 'GHNÖÅÉÏÔÂÑΓΒΘЧmbåñõéïγεφю75610½¼¾⅛ⅷ፹Ⅳ⑭〤⑽⒖㉑.,!?-;:()∫∅∂⌬⌨'#32#9#13#10#$A0#1#2#7#24#27#28#31;
  var
    FKernel: TASKernel;
    FTestCount,
    FSuccessCount,
    FFailCount: Integer;
    FChapterNumber: Integer;
    FFailList: TList<TTestItem>;
    FEps: TASR;
    FProgressEvent: TSelfTestProgressEvent;
    FChapterEvent: TSelfTestChapterEvent;
    FLastStatTime: UInt64;
    procedure Log(const AText: string; const AClass: string = '');
    procedure SendProgress(AForce: Boolean = False);
    procedure Chapter(const AChapterName: string);
    procedure Test(const AExpr: string; const AExpected: TAlgosimObject;
      ATestSingleLine: Boolean = False; const AExpSingleLine: string = ''); overload;
    procedure Test(const AExpr: string; const AExpected: ExceptClass;
      const AErrorMessage: string = ''); overload;
    procedure Test(const AExpr: string; const AExpected: TASI;
      TestSL: Boolean = False; const SL: string = ''); overload;
    procedure Test(const AExpr: string; const AExpected: TRationalNumber;
      TestSL: Boolean = False; const SL: string = ''); overload;
    procedure Test(const AExpr: string; const AExpected: TASR;
      TestSL: Boolean = False; const SL: string = ''); overload;
    procedure Test(const AExpr: string; const AExpected: TASC;
      TestSL: Boolean = False; const SL: string = ''); overload;
    procedure Test(const AExpr: string; const AExpected: string;
      TestSL: Boolean = False; const SL: string = ''); overload;
    procedure Test(const AExpr: string; const AExpected: array of TASR); overload;
    procedure Test(const AExpr: string; const AExpected: array of TASC); overload;
    procedure Test(const AExpr: string; AColCount: Integer; const AExpected: array of TASR); overload;
    procedure Test(const AExpr: string; AColCount: Integer; const AExpected: array of TASC); overload;
    procedure Test(const AExpr: string; const AExpected: TRGB;
      TestSL: Boolean = False; const SL: string = ''); overload;
    procedure Test(const AExpr: string; const AExpected: THSV;
      TestSL: Boolean = False; const SL: string = ''); overload;
    procedure Test(const AExpr: string; const AExpected: THSL;
      TestSL: Boolean = False; const SL: string = ''); overload;
        procedure Test(const AExpr: string; const AExpected: Boolean); overload;
    procedure Test(const AExpr: string; const AExpected: TASOSignal); overload;
    procedure Test(const AExpr: string; const AExpected: TASOSignal;
      const AErrorMessage: string); overload;
    procedure TestSL(const AExpr, AExpected: string);
    procedure TestML(const AExpr, AExpected: string);
    procedure TestDuration(const AMaxDur: Double); overload;
    procedure TestDuration(const AMinDur, AMaxDur: Double); overload;
    procedure TestExternal(const ATestName: string; AAssertion: Boolean);
    procedure TestNumFmt(const ARatValue: string; ANumberFormat: TNumberFormat;
      ANumDigits: Integer; const AExpOutput: array of string); overload;
    procedure TestNumFmt(const ARatValue: string; ANumberFormat: TNumberFormat;
      const AExpOutput: array of string); overload;
    procedure Eps(const AEpsilon: TASR = 0);
    function _ASOEq(A, B: TAlgosimObject): Boolean;
    procedure TestSeq;
    function Name(const AFirst, ALast: string): TAlgosimStructure;
    function Person(const AFirstName, ALastName, ASex: string;
      AAge: Integer): TAlgosimStructure;
    function Jane: TAlgosimStructure;
    function Mike: TAlgosimStructure;
    function John: TAlgosimStructure;
    function Sarah: TAlgosimStructure;
    function Mary: TAlgosimStructure;
    function SPerson(const AName: string; AAge: Integer;
      const ASex: string = ''): TAlgosimStructure;
    procedure TestRealMatrixQ(const AFcnName: string;
      const APositives: AnsiString);
    procedure TestRealMatrixQImplies(const ALeft, ARight: string);
    procedure TestComplexMatrixQ(const AFcnName: string;
      const APositives: AnsiString);
    procedure TestComplexMatrixQImplies(const ALeft, ARight: string);
    function LoremIpsum(AParagraphs: Integer): string;
    procedure TestCharType(const AFcnName: string; const APositives: string);
    procedure TestDefVars;
  public
    constructor Create(AKernel: TASKernel);
    procedure RunTests;
    destructor Destroy; override;
    property OnProgress: TSelfTestProgressEvent read FProgressEvent write FProgressEvent;
    property OnChapter: TSelfTestChapterEvent read FChapterEvent write FChapterEvent;
  end;

procedure Run(AKernel: TASKernel; AProgressEvent: TSelfTestProgressEvent;
  AChapterEvent: TSelfTestChapterEvent);
begin
  with TSelfTester.Create(AKernel) do
    try
      OnProgress := AProgressEvent;
      OnChapter := AChapterEvent;
      RunTests;
    finally
      Free;
    end;
end;

{ TSelfTester.TTestItem }

constructor TSelfTester.TTestItem.Create(const AExpr, AExpected,
  AActual: string);
begin
  Self.Expr := AExpr;
  Self.Expected := AExpected;
  Self.Actual := AActual;
end;

function TSelfTester.TTestItem.ToString: string;
begin
  Result :=
    '--------------------------------------------'#13#10 +
    'Test failed.'#13#10 +
    Self.Expr + #13#10 +
    '  Expected: ' + Self.Expected + #13#10 +
    '    Actual: ' + Self.Actual + #13#10 +
    '--------------------------------------------'#13#10;
end;

{ TSelfTester }

procedure TSelfTester.Chapter(const AChapterName: string);
begin
  Inc(FChapterNumber);
  if Assigned(FChapterEvent) then
    FChapterEvent(FChapterNumber, AChapterName);
end;

constructor TSelfTester.Create(AKernel: TASKernel);
begin
  FFailList := TList<TTestItem>.Create;
  FKernel := AKernel;
  FKernel.ClearAllVars;
  FKernel.LoadDefVars;
  FKernel.ClearAllBuffers;
  FKernel.ClearCommandHistory;
  FEps := -1;
end;

destructor TSelfTester.Destroy;
begin
  FFailList.Free;
  FKernel.ClearCommandHistory;
  FKernel.ClearAllBuffers;
  FKernel.ClearAllVars;
  FKernel.LoadDefVars;
  inherited;
end;

procedure TSelfTester.Eps(const AEpsilon: TASR);
begin
  FEps := AEpsilon;
end;

procedure TSelfTester.Log(const AText: string; const AClass: string);
begin
  FKernel.Perform(CLIENT_COMMAND_PRINT, 0, NativeInt(PChar(AText)),
    NativeInt(PChar(AClass)), 0);
end;

function TSelfTester.LoremIpsum(AParagraphs: Integer): string;
const
  Para =
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor ' +
    'incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis ' +
    'nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. ' +
    'Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu ' +
    'fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in ' +
    'culpa qui officia deserunt mollit anim id est laborum.';
var
  p: PChar;
  i: Integer;
const
  CRLF: string = #13#10;
begin

  if AParagraphs < 1 then
    Exit('');

  SetLength(Result, AParagraphs * Para.Length + (AParagraphs - 1) * CRLF.Length);

  p := PChar(Result);

  for i := 1 to AParagraphs do
  begin
    MoveChars(Para[1], p^, Para.Length);
    Inc(p, Para.Length);
    if i < AParagraphs then
    begin
      MoveChars(CRLF[1], p^, CRLF.Length);
      Inc(p, CRLF.Length);
    end;
  end;

end;

procedure TSelfTester.RunTests;
var
  OldSeed: Integer;
begin
  SendProgress;
  OldSeed := RandSeed;
  RandSeed := 9617237;
  try
    TestSeq;
  finally
    RandSeed := OldSeed;
  end;
  SendProgress(True);
end;

function TSelfTester.Name(const AFirst, ALast: string): TAlgosimStructure;
begin
  Result := TAlgosimStructure.CreateWithValue(
    ['first', 'last'],
    [ASO(AFirst), ASO(ALast)]
  );
end;

function TSelfTester.Person(const AFirstName, ALastName, ASex: string; AAge: Integer): TAlgosimStructure;
begin
  Result := TAlgosimStructure.CreateWithValue(
    ['name', 'sex', 'age'],
    [Name(AFirstName, ALastName), ASO(ASex), ASOInt(AAge)]
  );
end;

function TSelfTester.Jane: TAlgosimStructure;
begin
  Jane := Person('Jane', 'Smith', 'female', 40);
end;

function TSelfTester.John: TAlgosimStructure;
begin
  John := Person('John', 'Stone', 'male', 69);
end;

function TSelfTester.Mary: TAlgosimStructure;
begin
  Mary := TAlgosimStructure.CreateWithValue(
    ['mother', 'father'],
    [Jane, Mike]
  );
end;

function TSelfTester.Mike: TAlgosimStructure;
begin
  Mike := Person('Mike', 'Doe', 'male', 45);
end;

function TSelfTester.Sarah: TAlgosimStructure;
begin
  Sarah := Person('Sarah', 'Lester', 'female', 25);
end;

procedure TSelfTester.SendProgress(AForce: Boolean);
begin
  if not AForce and (GetTickCount64 - FLastStatTime < 100) then
    Exit;
  if Assigned(FProgressEvent) then
  begin
    FProgressEvent(TotalCount, FTestCount, FSuccessCount, FFailCount);
  end;
  FLastStatTime := GetTickCount64;
end;

function TSelfTester.SPerson(const AName: string;
  AAge: Integer; const ASex: string): TAlgosimStructure;
begin
  if ASex.IsEmpty then
    Result := TAlgosimStructure.CreateWithValue(
      ['name', 'age'],
      [ASO(AName), ASOInt(AAge)]
    )
  else
    Result := TAlgosimStructure.CreateWithValue(
      ['name', 'age', 'sex'],
      [ASO(AName), ASOInt(AAge), ASO(ASex)]
    )
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: array of TASR);
begin
  Test(AExpr, ASO(TRealVector.Create(AExpected)));
end;

procedure TSelfTester.Test(const AExpr, AExpected: string;
  TestSL: Boolean = False; const SL: string = '');
begin
  Test(AExpr, ASO(AExpected), TestSL, SL);
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: TASC;
  TestSL: Boolean; const SL: string);
begin
  Test(AExpr, ASO(AExpected), TestSL, SL);
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: TRationalNumber;
  TestSL: Boolean; const SL: string);
begin
  if TestSL then
    Test('SetNumberFormat((' + AExpr + '), "fraction")', ASORat(AExpected), TestSL, SL)
  else
    Test(AExpr, ASORat(AExpected), TestSL, SL);
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: TASI;
  TestSL: Boolean; const SL: string);
begin
  Test(AExpr, ASOInt(AExpected), TestSL, SL);
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: TAlgosimObject;
  ATestSingleLine: Boolean = False; const AExpSingleLine: string = '');
var
  res: TAlgosimObject;
begin
  try
    Inc(FTestCount);
    try
      res := FKernel.Evaluate(AExpr);
    except
      on E: Exception do
      begin
        Inc(FFailCount);
        FFailList.Add(TTestItem.Create(AExpr, AExpected.ClassName + '(' + AExpected.ToString + ')', E.ClassName + E.Message));
        Log(FFailList.Last.ToString, 'Error');
        Exit;
      end;
    end;
    if
      (res.ClassType = AExpected.ClassType) and
      _ASOEq(res, AExpected) and
      (not ATestSingleLine or res.GetAsSingleLineText(FKernel.FormatOptions).Equals(AExpSingleLine))
    then
      Inc(FSuccessCount)
    else
    begin
      Inc(FFailCount);
      if not (res.ClassType = AExpected.ClassType) or not _ASOEq(res, AExpected) then
      begin
        if res is TAlgosimFailure then
          FFailList.Add(TTestItem.Create(AExpr, AExpected.ClassName + '(' + AExpected.ToString + ')', res.ClassName + '(' + TAlgosimFailure(res).FailureReason + ')'))
        else
          FFailList.Add(TTestItem.Create(AExpr, AExpected.ClassName + '(' + AExpected.ToString + ')', res.ClassName + '(' + res.ToString + ')'))
      end
      else
        FFailList.Add(TTestItem.Create(AExpr, AExpSingleLine, res.GetAsSingleLineText(FKernel.FormatOptions)));
      Log(FFailList.Last.ToString, 'Error');
    end;
  finally
    FEps := -1; // noexcept
    AExpected.Free;
    SendProgress;
  end;
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: ExceptClass;
  const AErrorMessage: string);
var
  res: TAlgosimObject;
begin
  Inc(FTestCount);
  try
    res := FKernel.Evaluate(AExpr);
  except
    on E: Exception do
    begin
      if (E is AExpected) and (AErrorMessage.IsEmpty or E.Message.ToLower.Contains(AErrorMessage.ToLower)) then
        Inc(FSuccessCount)
      else
      begin
        Inc(FFailCount);
        if not (E is AExpected) then
          FFailList.Add(TTestItem.Create(AExpr, AExpected.ClassName, E.ClassName))
        else
          FFailList.Add(TTestItem.Create(AExpr, AExpected.ClassName + ': ' + AErrorMessage, E.ClassName + ': ' + E.Message));
        Log(FFailList.Last.ToString, 'Error');
      end;
      SendProgress;
      Exit;
    end;
  end;
  Inc(FFailCount);
  FFailList.Add(TTestItem.Create(AExpr, AExpected.ClassName, res.ClassName + '(' + res.ToString + ')'));
  Log(FFailList.Last.ToString, 'Error');
  SendProgress;
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: TASR;
  TestSL: Boolean; const SL: string);
begin
  Test(AExpr, ASO(AExpected), TestSL, SL);
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: array of TASC);
begin
  Test(AExpr, ASO(TComplexVector.Create(AExpected)));
end;

procedure TSelfTester.Test(const AExpr: string; AColCount: Integer;
  const AExpected: array of TASC);
begin
  Test(AExpr, ASO(TComplexMatrix.Create(AExpected, AColCount)));
end;

procedure TSelfTester.Test(const AExpr: string; AColCount: Integer;
  const AExpected: array of TASR);
begin
  Test(AExpr, ASO(TRealMatrix.Create(AExpected, AColCount)));
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: TRGB;
  TestSL: Boolean = False; const SL: string = '');
begin
  Test(AExpr, ASO(AExpected), TestSL, SL);
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: Boolean);
begin
  Test(AExpr, ASO(AExpected));
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: TASOSignal);
begin
  Test(AExpr, ASO(AExpected));
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: TASOSignal;
  const AErrorMessage: string);
var
  res: TAlgosimObject;
begin
  Assert(AExpected = failure);
  Inc(FTestCount);
  try
    res := FKernel.Evaluate(AExpr);
  except
    on E: Exception do
    begin
      Inc(FFailCount);
      FFailList.Add(TTestItem.Create(AExpr, 'failure' + ' (' + AErrorMessage + ')', E.ClassName));
      Log(FFailList.Last.ToString, 'Error');
      SendProgress;
      Exit;
    end;
  end;
  if
    (res is TAlgosimFailure)
      and
    TAlgosimFailure(res).FailureReason.ToLower.Contains(AErrorMessage.ToLower)
  then
    Inc(FSuccessCount)
  else
  begin
    Inc(FFailCount);
    if not (res is TAlgosimFailure) then
      FFailList.Add(TTestItem.Create(AExpr, 'failure', res.ClassName + '(' + res.ToString + ')'))
    else
      FFailList.Add(TTestItem.Create(AExpr, 'failure (' + AErrorMessage + ')', TAlgosimFailure(res).FailureReason));
    Log(FFailList.Last.ToString, 'Error');
  end;
  SendProgress;
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: THSL;
  TestSL: Boolean; const SL: string);
begin
  Test(AExpr, ASO(AExpected), TestSL, SL);
end;

procedure TSelfTester.Test(const AExpr: string; const AExpected: THSV;
  TestSL: Boolean; const SL: string);
begin
  Test(AExpr, ASO(AExpected), TestSL, SL);
end;

procedure TSelfTester.TestCharType(const AFcnName, APositives: string);
var
  c: Char;
begin

  for c in CharSamples do
    Test(Format('%s("%s")', [AFcnName, c]), Pos(c, APositives) <> 0);

end;

procedure TSelfTester.TestComplexMatrixQ(const AFcnName: string;
  const APositives: AnsiString);
var
  c: AnsiChar;
  x: set of AnsiChar;
begin
  x := [];
  for c in ComplexMatrixNames do
    Include(x, c);
  for c in APositives do
  begin
    Test(Format('%s(%s)', [AFcnName, c]), True);
    Exclude(x, c);
  end;
  for c in x do
    Test(Format('%s(%s)', [AFcnName, c]), False);
end;

procedure TSelfTester.TestComplexMatrixQImplies(const ALeft, ARight: string);
var
  c: AnsiChar;
begin
  for c in ComplexMatrixNames do
    Test(Format('%s(%s) ⇒ %s(%s)', [ALeft, c, ARight, c]), True);
end;

procedure TSelfTester.TestDefVars;
begin
  Eps; Test('e', 2.7182818284590452353602874);
  Eps; Test('π', 3.1415926535897932384626433);
       Test('ℎ', 6.62607015E-34); // No blasted [epsilons] [here]! No, sir. Not one, single, bloody [epsilon]! Not one! [...] No, sir! Not one blasted, miserable ...
  Eps; Test('ℏ', 6.62607015E-34 / (2*3.1415926535897932384626433));
  Test('i', ImaginaryUnit); Test('i^2', TASC(-1));
  Test('true', True);
  Test('false', False);
  Test('otherwise', True);
  Test('∅', TAlgosimSet.Create); Test('type(∅)', 'set'); Test('#∅', 0); Test('∅ = {}', True);
  Test('¶', sLineBreak);
end;

procedure TSelfTester.TestDuration(const AMinDur, AMaxDur: Double);
begin
  Test(
    AMinDur.ToString(TFormatSettings.Invariant) +
    '≤' +
    ElapsedTime.ToString(TFormatSettings.Invariant) +
    '∧' +
    ElapsedTime.ToString(TFormatSettings.Invariant) +
    '≤' +
    AMaxDur.ToString(TFormatSettings.Invariant),
    True
  );
end;

procedure TSelfTester.TestExternal(const ATestName: string;
  AAssertion: Boolean);
begin
  if AAssertion then
    Inc(FSuccessCount)
  else
  begin
    Inc(FFailCount);
    FFailList.Add(TTestItem.Create(ATestName, 'Success', 'Failure'));
    Log(FFailList.Last.ToString, 'Error');
  end;
end;

procedure TSelfTester.TestDuration(const AMaxDur: Double);
begin
  Test(
    ElapsedTime.ToString(TFormatSettings.Invariant) +
    '≤' +
    AMaxDur.ToString(TFormatSettings.Invariant),
    True
  );
end;

procedure TSelfTester.TestSL(const AExpr, AExpected: string);
begin
  Test('AsSingleLine(' + AExpr + ')', AExpected);
end;

procedure TSelfTester.TestML(const AExpr, AExpected: string);
begin
  Test('AsMultiLine(' + AExpr + ')', AExpected);
end;

procedure TSelfTester.TestNumFmt(const ARatValue: string;
  ANumberFormat: TNumberFormat; const AExpOutput: array of string);
begin
  TestNumFmt(ARatValue, ANumberFormat, -1, AExpOutput);
end;

procedure TSelfTester.TestNumFmt(const ARatValue: string;
  ANumberFormat: TNumberFormat; ANumDigits: Integer;
  const AExpOutput: array of string);

const
  NumFmtStrs: array[TNumberFormat] of string =
    ('default', 'fraction', 'fixed', 'scientific', 'power');

type
  TPrettyExpTestMode = (petmImplicit, petmOn, petmOff);

const
  PrettyExpTestModes = [Low(TPrettyExpTestMode)..High(TPrettyExpTestMode)];

  function ApplySettings(const S: string; AImplicitDef: Boolean;
    APrettyExp: TPrettyExpTestMode; ADigitGrouping: Integer): string;
  begin
    Result := S;
    if not AImplicitDef then
      Result := Format('SetNumberFormat(%s, "%s")', [Result, NumFmtStrs[ANumberFormat]]);
    if ANumDigits <> -1 then
      Result := Format('(%s) \ %d', [Result, ANumDigits]);
    if APrettyExp <> petmImplicit then
      Result := Format('SetPrettyExp(%s, %s)', [Result, IfThen(APrettyExp = petmOn, 'true', 'false')]);
    if ADigitGrouping <> -1 then
      Result := Format('SetDigitGrouping(%s, %d)', [Result, ADigitGrouping]);
  end;

  function ForceRat(const S: string): string;
  begin
    Result := Format('ToFraction(%s)', [S]);
  end;

  function ForceReal(const S: string): string;
  begin
    Result := Format('RealNumber(%s)', [S]);
  end;

  function ForceCplx(const S: string): string;
  begin
    Result := Format('ComplexNumber(%s)', [S]);
  end;

  function Unpretty(const S: string; APETM: TPrettyExpTestMode): string;
  begin
    if APETM = petmOff then
      Result := S
        .Replace('⋅10^−', 'E−')
        .Replace('⋅10^', 'E' + IfThen(ANumberFormat = nfExponent, '+'))
    else
      Result := S;
  end;

  function DoDigGr(const S: string; ADigGr: Integer): string;

    procedure DeE(var S: string);
    var
      i: Integer;
    begin
      i := S.Length;
      while (i > 0) and (S[i] <> 'E') and (S[i] <> 'e') and (S[i] <> DOT_OPERATOR) do
        Dec(i);
      if i > 0 then
        SetLength(S, i - 1);
    end;

  var
    p: Integer;
    Sgn, IntPart, IntPartGrpd, FracPart, FracPartGrpd, Suffix: string;
    i, j: Integer;
  begin

    if ADigGr < 1 then
      Exit(S);

    p := Pos('/', S);
    if p <> 0 then
      Exit(DoDigGr(Copy(S, 1, Pred(p)), ADigGr) + '/' + DoDigGr(Copy(S, Succ(p)), ADigGr));

    if IndexStr(Copy(S, 1, 1), [MINUS_SIGN, '-', '+']) <> -1 then
      Sgn := S[1];

    p := Pos('.', S);
    IntPart := Copy(S, 1 + Sgn.Length, IfThen(p = 0, MaxInt, p - 1 - Sgn.Length));
    if p <> 0 then
      FracPart := Copy(S, Succ(p));

    DeE(IntPart);
    DeE(FracPart);

    p := Pos('E', S);
    if p = 0 then
      p := Pos('e', S);
    if p = 0 then
      p := Pos(DOT_OPERATOR, S);
    if p <> 0 then
      Suffix := Copy(S, p);

    if not IntPart.IsEmpty then
    begin
      SetLength(IntPartGrpd, IntPart.Length + Ceil(IntPart.Length / ADigGr) - 1);
      j := 0;
      for i := 1 to IntPart.Length do
      begin
        if (i <> 1) and ((IntPart.Length - i + 1) mod ADigGr = 0) then
        begin
          Inc(j);
          IntPartGrpd[j] := FIGURE_SPACE;
        end;
        Inc(j);
        IntPartGrpd[j] := IntPart[i];
      end;
      Assert(j = IntPartGrpd.Length);
    end;

    if not FracPart.IsEmpty then
    begin
      SetLength(FracPartGrpd, FracPart.Length + Ceil(FracPart.Length / ADigGr) - 1);
      j := 0;
      for i := 1 to FracPart.Length do
      begin
        if (i <> 1) and (Pred(i) mod ADigGr = 0) then
        begin
          Inc(j);
          FracPartGrpd[j] := FIGURE_SPACE;
        end;
        Inc(j);
        FracPartGrpd[j] := FracPart[i];
      end;
      Assert(j = FracPartGrpd.Length);
    end;

    Result := Sgn + IntPartGrpd + IfThen(not FracPart.IsEmpty, '.') + FracPartGrpd + Suffix;

  end;

  function NZ(const S: string): Boolean;
  var
    i: Integer;
  begin
    for i := 1 to S.Length do
      if S[i].IsInArray(['E', 'e', DOT_OPERATOR]) then
        Break
      else if CharInSet(S[i], ['1'..'9']) then
        Exit(True);
    Result := False;
  end;

const
  Sgns: array[Boolean] of string = ('', MINUS_SIGN);

var
  dummy: Int64;
  i: Integer;
  LExpOutput: array[0..3] of string;
  ImplVals: array of Boolean;
  b: Boolean;
  petm: TPrettyExpTestMode;
  neg: Boolean;
  LRatValue: string;
  DigGr: Integer;

begin

  if not (Length(AExpOutput) in [1, 2, 3, 4]) then
    raise Exception.Create('TSelfTester.TestNumFmt: Incorrect array length.');

  for i := Low(LExpOutput) to High(LExpOutput) do
    LExpOutput[i] := AExpOutput[Min(i, High(AExpOutput))]
      .Replace('E', '⋅10^')
      .Replace(HYPHEN_MINUS, MINUS_SIGN)
      .Replace(#32, FIGURE_SPACE);

  if ANumberFormat = nfDefault then
    ImplVals := [False, True]
  else
    ImplVals := [False];

  for DigGr in TArray<Integer>.Create(-1, 0, 1, 2, 3, 4, 5) do
    for neg in [False, True] do
      for petm in PrettyExpTestModes do
        for b in ImplVals do
        begin

          if neg then
            LRatValue := '-' + ARatValue
          else
            LRatValue := ARatValue;

          if TryStrToInt64(ARatValue, dummy) then
          begin
            if LExpOutput[0] <> '*' then
              TestSL(ApplySettings(LRatValue, b, petm, DigGr), Sgns[neg and NZ(LExpOutput[0])] + DoDigGr(Unpretty(LExpOutput[0], petm), DigGr));
          end
          else
            if LExpOutput[0] <> '!' then
              raise Exception.CreateFmt('TSelfTester.TestNumFmt: Cannot convert "%s" to an integer.', [LRatValue]);

          if (LExpOutput[1] <> '*') and not b then
            TestSL(ApplySettings(ForceRat(LRatValue), b, petm, DigGr), Sgns[neg and NZ(LExpOutput[1])] + DoDigGr(Unpretty(LExpOutput[1], petm), DigGr));
          if LExpOutput[2] <> '*' then
            TestSL(ApplySettings(ForceReal(LRatValue), b, petm, DigGr), Sgns[neg and NZ(LExpOutput[2])] + DoDigGr(Unpretty(LExpOutput[2], petm), DigGr));
          if LExpOutput[3] <> '*' then
            TestSL(ApplySettings(ForceCplx(LRatValue), b, petm, DigGr), Sgns[neg and NZ(LExpOutput[3])] + DoDigGr(Unpretty(LExpOutput[3], petm), DigGr));

        end;

end;

function Rat(p, q: TASI): TRationalNumber;
begin
  Result := TRationalNumber.Create(p, q);
end;

function fs(const S: string): string;
var
  i: Integer;
begin
  Result := S;
  for i := 1 to S.Length do
    if Result[i] = #32 then
      Result[i] := FIGURE_SPACE;
end;

function mm(const S: string): string;
var
  i: Integer;
begin
  Result := S;
  for i := 1 to S.Length do
    if Result[i] = '-' then
      Result[i] := MINUS_SIGN
    else if Result[i] = '*' then
      Result[i] := DOT_OPERATOR;
end;


function TSelfTester._ASOEq(A, B: TAlgosimObject): Boolean;
begin
  if FEps >= 0 then
    Result := SameASO(A, B, FEps)
  else
    Result := A.Equals(B) or ((A is TAlgosimNullObject) and (B is TAlgosimNullObject));
end;

function intset(const AInts: array of TASI): TAlgosimSet;
var
  i: TASI;
begin
  Result := TAlgosimSet.Create;
  try
    for i in AInts do
      Result.AddElement(ASOInt(i));
  except
    Result.Free;
    raise;
  end;
end;

function asoset(const AElements: array of const): TAlgosimSet;
var
  i: Integer;
begin
  Result := TAlgosimSet.Create;
  try
    for i := Low(AElements) to High(AElements) do
      case AElements[i].VType of
        vtInteger:
          Result.AddElement(ASO(AElements[i].VInteger));
        vtInt64:
          Result.AddElement(ASO(AElements[i].VInt64^));
        vtBoolean:
          Result.AddElement(ASO(AElements[i].VBoolean));
        vtExtended:
          Result.AddElement(ASO(AElements[i].VExtended^));
        vtString:
          Result.AddElement(ASO(string(AElements[i].VString^)));
        vtUnicodeString:
          Result.AddElement(ASO(UnicodeString(AElements[i].VUnicodeString)));
        vtWideChar:
          Result.AddElement(ASO(AElements[i].VWideChar));
      else
        raise Exception.Create('asoset: Unsupported variant type.');
      end;
  except
    Result.Free;
    raise;
  end;
end;

function asosetex(const AElements: array of TAlgosimObject): TAlgosimSet;
var
  i: Integer;
  LastAdded: Integer;
begin
  LastAdded := -1;
  Result := TAlgosimSet.Create;
  try
    for i := Low(AElements) to High(AElements) do
    begin
      Result.AddElement(AElements[i]);
      LastAdded := i;
    end;
  except
    Result.Free;
    for i := LastAdded + 1 to High(AElements) do
      AElements[i].Free;
    raise;
  end;
end;

function asoarr(const AElements: array of const): TAlgosimArray;
var
  i: Integer;
begin
  Result := TAlgosimArray.Create;
  try
    for i := Low(AElements) to High(AElements) do
      case AElements[i].VType of
        vtInteger:
          Result.AddElement(ASO(AElements[i].VInteger));
        vtInt64:
          Result.AddElement(ASO(AElements[i].VInt64^));
        vtBoolean:
          Result.AddElement(ASO(AElements[i].VBoolean));
        vtExtended:
          Result.AddElement(ASO(AElements[i].VExtended^));
        vtString:
          Result.AddElement(ASO(string(AElements[i].VString^)));
        vtUnicodeString:
          Result.AddElement(ASO(UnicodeString(AElements[i].VUnicodeString)));
        vtWideChar:
          Result.AddElement(ASO(AElements[i].VWideChar));
      else
        raise Exception.Create('asoarr: Unsupported variant type.');
      end;
  except
    Result.Free;
    raise;
  end;
end;

function asoarrex(const AElements: array of TAlgosimObject): TAlgosimArray;
var
  i: Integer;
  LastAdded: Integer;
begin
  LastAdded := -1;
  Result := TAlgosimArray.Create;
  try
    for i := Low(AElements) to High(AElements) do
    begin
      Result.AddElement(AElements[i]);
      LastAdded := i;
    end;
  except
    Result.Free;
    for i := LastAdded + 1 to High(AElements) do
      AElements[i].Free;
    raise;
  end;
end;

function intarr(const AInts: array of TASI): TAlgosimArray;
var
  i: TASI;
begin
  Result := TAlgosimArray.Create;
  try
    for i in AInts do
      Result.AddElement(ASOInt(i));
  except
    Result.Free;
    raise;
  end;
end;

function intarrN(const AInts: array of TASI): TAlgosimArray;
var
  i: TASI;
begin
  Result := TAlgosimArray.Create;
  try
    for i in AInts do
      if i <> -1 then
        Result.AddElement(ASOInt(i))
      else
        Result.AddElement(ASO(null))
  except
    Result.Free;
    raise;
  end;
end;

function chrarr(const AChrs: array of Char): TAlgosimArray;
var
  c: Char;
begin
  Result := TAlgosimArray.Create;
  try
    for c in AChrs do
      Result.AddElement(ASO(c));
  except
    Result.Free;
    raise;
  end;
end;

function strarr(const AStrs: array of string): TAlgosimArray;
var
  s: string;
begin
  Result := TAlgosimArray.Create;
  try
    for s in AStrs do
      Result.AddElement(ASO(s));
  except
    Result.Free;
    raise;
  end;
end;

function idx2(A, B: Int64): TAlgosimArray;
begin
  Result := intarr([A, B]);
end;

function freq(AVal, AFreq: Int64): TAlgosimArray; overload;
begin
  Result := intarr([AVal, AFreq]);
end;

function freq(AVal: TASR; AFreq: Int64): TAlgosimArray; overload;
begin
  Result := asoarr([AVal, AFreq]);
end;

function freq(AVal: TASC; AFreq: Int64): TAlgosimArray; overload;
begin
  Result := asoarrex([ASO(AVal), ASOInt(AFreq)]);
end;

function freq(const AVal: string; AFreq: Int64): TAlgosimArray; overload;
begin
  Result := asoarrex([ASO(AVal), ASOInt(AFreq)]);
end;

function freq(AVal: Boolean; AFreq: Int64): TAlgosimArray; overload;
begin
  Result := asoarrex([ASO(AVal), ASOInt(AFreq)]);
end;

function freq(AVal: TAlgosimObject; AFreq: Int64): TAlgosimArray; overload;
begin
  Result := asoarrex([AVal, ASOInt(AFreq)]);
end;

var
  _hpc1, _hpc2, _hpcf: Int64;

procedure StartStopwatch;
begin
  ElapsedTime := 0.0;
  QueryPerformanceFrequency(_hpcf);
  QueryPerformanceCounter(_hpc1);
end;

procedure StopStopwatch;
begin
  QueryPerformanceCounter(_hpc2);
  ElapsedTime := (_hpc2 - _hpc1) / _hpcf;
end;


const
  SmallPrimes: array[0..999] of Integer =
    (
      2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73,
      79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163,
      167, 173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 251,
      257, 263, 269, 271, 277, 281, 283, 293, 307, 311, 313, 317, 331, 337, 347, 349,
      353, 359, 367, 373, 379, 383, 389, 397, 401, 409, 419, 421, 431, 433, 439, 443,
      449, 457, 461, 463, 467, 479, 487, 491, 499, 503, 509, 521, 523, 541, 547, 557,
      563, 569, 571, 577, 587, 593, 599, 601, 607, 613, 617, 619, 631, 641, 643, 647,
      653, 659, 661, 673, 677, 683, 691, 701, 709, 719, 727, 733, 739, 743, 751, 757,
      761, 769, 773, 787, 797, 809, 811, 821, 823, 827, 829, 839, 853, 857, 859, 863,
      877, 881, 883, 887, 907, 911, 919, 929, 937, 941, 947, 953, 967, 971, 977, 983,
      991, 997, 1009, 1013, 1019, 1021, 1031, 1033, 1039, 1049, 1051, 1061, 1063,
      1069, 1087, 1091, 1093, 1097, 1103, 1109, 1117, 1123, 1129, 1151, 1153, 1163,
      1171, 1181, 1187, 1193, 1201, 1213, 1217, 1223, 1229, 1231, 1237, 1249, 1259,
      1277, 1279, 1283, 1289, 1291, 1297, 1301, 1303, 1307, 1319, 1321, 1327, 1361,
      1367, 1373, 1381, 1399, 1409, 1423, 1427, 1429, 1433, 1439, 1447, 1451, 1453,
      1459, 1471, 1481, 1483, 1487, 1489, 1493, 1499, 1511, 1523, 1531, 1543, 1549,
      1553, 1559, 1567, 1571, 1579, 1583, 1597, 1601, 1607, 1609, 1613, 1619, 1621,
      1627, 1637, 1657, 1663, 1667, 1669, 1693, 1697, 1699, 1709, 1721, 1723, 1733,
      1741, 1747, 1753, 1759, 1777, 1783, 1787, 1789, 1801, 1811, 1823, 1831, 1847,
      1861, 1867, 1871, 1873, 1877, 1879, 1889, 1901, 1907, 1913, 1931, 1933, 1949,
      1951, 1973, 1979, 1987, 1993, 1997, 1999, 2003, 2011, 2017, 2027, 2029, 2039,
      2053, 2063, 2069, 2081, 2083, 2087, 2089, 2099, 2111, 2113, 2129, 2131, 2137,
      2141, 2143, 2153, 2161, 2179, 2203, 2207, 2213, 2221, 2237, 2239, 2243, 2251,
      2267, 2269, 2273, 2281, 2287, 2293, 2297, 2309, 2311, 2333, 2339, 2341, 2347,
      2351, 2357, 2371, 2377, 2381, 2383, 2389, 2393, 2399, 2411, 2417, 2423, 2437,
      2441, 2447, 2459, 2467, 2473, 2477, 2503, 2521, 2531, 2539, 2543, 2549, 2551,
      2557, 2579, 2591, 2593, 2609, 2617, 2621, 2633, 2647, 2657, 2659, 2663, 2671,
      2677, 2683, 2687, 2689, 2693, 2699, 2707, 2711, 2713, 2719, 2729, 2731, 2741,
      2749, 2753, 2767, 2777, 2789, 2791, 2797, 2801, 2803, 2819, 2833, 2837, 2843,
      2851, 2857, 2861, 2879, 2887, 2897, 2903, 2909, 2917, 2927, 2939, 2953, 2957,
      2963, 2969, 2971, 2999, 3001, 3011, 3019, 3023, 3037, 3041, 3049, 3061, 3067,
      3079, 3083, 3089, 3109, 3119, 3121, 3137, 3163, 3167, 3169, 3181, 3187, 3191,
      3203, 3209, 3217, 3221, 3229, 3251, 3253, 3257, 3259, 3271, 3299, 3301, 3307,
      3313, 3319, 3323, 3329, 3331, 3343, 3347, 3359, 3361, 3371, 3373, 3389, 3391,
      3407, 3413, 3433, 3449, 3457, 3461, 3463, 3467, 3469, 3491, 3499, 3511, 3517,
      3527, 3529, 3533, 3539, 3541, 3547, 3557, 3559, 3571, 3581, 3583, 3593, 3607,
      3613, 3617, 3623, 3631, 3637, 3643, 3659, 3671, 3673, 3677, 3691, 3697, 3701,
      3709, 3719, 3727, 3733, 3739, 3761, 3767, 3769, 3779, 3793, 3797, 3803, 3821,
      3823, 3833, 3847, 3851, 3853, 3863, 3877, 3881, 3889, 3907, 3911, 3917, 3919,
      3923, 3929, 3931, 3943, 3947, 3967, 3989, 4001, 4003, 4007, 4013, 4019, 4021,
      4027, 4049, 4051, 4057, 4073, 4079, 4091, 4093, 4099, 4111, 4127, 4129, 4133,
      4139, 4153, 4157, 4159, 4177, 4201, 4211, 4217, 4219, 4229, 4231, 4241, 4243,
      4253, 4259, 4261, 4271, 4273, 4283, 4289, 4297, 4327, 4337, 4339, 4349, 4357,
      4363, 4373, 4391, 4397, 4409, 4421, 4423, 4441, 4447, 4451, 4457, 4463, 4481,
      4483, 4493, 4507, 4513, 4517, 4519, 4523, 4547, 4549, 4561, 4567, 4583, 4591,
      4597, 4603, 4621, 4637, 4639, 4643, 4649, 4651, 4657, 4663, 4673, 4679, 4691,
      4703, 4721, 4723, 4729, 4733, 4751, 4759, 4783, 4787, 4789, 4793, 4799, 4801,
      4813, 4817, 4831, 4861, 4871, 4877, 4889, 4903, 4909, 4919, 4931, 4933, 4937,
      4943, 4951, 4957, 4967, 4969, 4973, 4987, 4993, 4999, 5003, 5009, 5011, 5021,
      5023, 5039, 5051, 5059, 5077, 5081, 5087, 5099, 5101, 5107, 5113, 5119, 5147,
      5153, 5167, 5171, 5179, 5189, 5197, 5209, 5227, 5231, 5233, 5237, 5261, 5273,
      5279, 5281, 5297, 5303, 5309, 5323, 5333, 5347, 5351, 5381, 5387, 5393, 5399,
      5407, 5413, 5417, 5419, 5431, 5437, 5441, 5443, 5449, 5471, 5477, 5479, 5483,
      5501, 5503, 5507, 5519, 5521, 5527, 5531, 5557, 5563, 5569, 5573, 5581, 5591,
      5623, 5639, 5641, 5647, 5651, 5653, 5657, 5659, 5669, 5683, 5689, 5693, 5701,
      5711, 5717, 5737, 5741, 5743, 5749, 5779, 5783, 5791, 5801, 5807, 5813, 5821,
      5827, 5839, 5843, 5849, 5851, 5857, 5861, 5867, 5869, 5879, 5881, 5897, 5903,
      5923, 5927, 5939, 5953, 5981, 5987, 6007, 6011, 6029, 6037, 6043, 6047, 6053,
      6067, 6073, 6079, 6089, 6091, 6101, 6113, 6121, 6131, 6133, 6143, 6151, 6163,
      6173, 6197, 6199, 6203, 6211, 6217, 6221, 6229, 6247, 6257, 6263, 6269, 6271,
      6277, 6287, 6299, 6301, 6311, 6317, 6323, 6329, 6337, 6343, 6353, 6359, 6361,
      6367, 6373, 6379, 6389, 6397, 6421, 6427, 6449, 6451, 6469, 6473, 6481, 6491,
      6521, 6529, 6547, 6551, 6553, 6563, 6569, 6571, 6577, 6581, 6599, 6607, 6619,
      6637, 6653, 6659, 6661, 6673, 6679, 6689, 6691, 6701, 6703, 6709, 6719, 6733,
      6737, 6761, 6763, 6779, 6781, 6791, 6793, 6803, 6823, 6827, 6829, 6833, 6841,
      6857, 6863, 6869, 6871, 6883, 6899, 6907, 6911, 6917, 6947, 6949, 6959, 6961,
      6967, 6971, 6977, 6983, 6991, 6997, 7001, 7013, 7019, 7027, 7039, 7043, 7057,
      7069, 7079, 7103, 7109, 7121, 7127, 7129, 7151, 7159, 7177, 7187, 7193, 7207,
      7211, 7213, 7219, 7229, 7237, 7243, 7247, 7253, 7283, 7297, 7307, 7309, 7321,
      7331, 7333, 7349, 7351, 7369, 7393, 7411, 7417, 7433, 7451, 7457, 7459, 7477,
      7481, 7487, 7489, 7499, 7507, 7517, 7523, 7529, 7537, 7541, 7547, 7549, 7559,
      7561, 7573, 7577, 7583, 7589, 7591, 7603, 7607, 7621, 7639, 7643, 7649, 7669,
      7673, 7681, 7687, 7691, 7699, 7703, 7717, 7723, 7727, 7741, 7753, 7757, 7759,
      7789, 7793, 7817, 7823, 7829, 7841, 7853, 7867, 7873, 7877, 7879, 7883, 7901,
      7907, 7919
    );

const
  JacobiSymbolValues: array[1..59] of array[1..30] of Integer =
    (
      (1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1, -1,  1,  0,  1, -1, -1,  1,  0,  1, -1, -1,  1,  0,  1, -1, -1,  1,  0,  1, -1, -1,  1,  0,  1, -1, -1,  1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1, -1,  1, -1, -1,  0,  1,  1, -1,  1, -1, -1,  0,  1,  1, -1,  1, -1, -1,  0,  1,  1, -1,  1, -1, -1,  0,  1,  1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  1,  1,  1, -1, -1, -1,  1, -1,  0,  1, -1,  1,  1,  1, -1, -1, -1,  1, -1,  0,  1, -1,  1,  1,  1, -1, -1, -1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  1,  1, -1, -1, -1, -1,  1,  1, -1,  1,  0,  1, -1,  1,  1, -1, -1, -1, -1,  1,  1, -1,  1,  0,  1, -1,  1,  1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1,  0,  1,  0,  0, -1,  1,  0,  0, -1,  0, -1, -1,  0,  1,  1,  0,  1,  0,  0, -1,  1,  0,  0, -1,  0, -1, -1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1, -1,  1, -1, -1, -1,  1,  1, -1, -1, -1,  1, -1,  1,  1,  0,  1,  1, -1,  1, -1, -1, -1,  1,  1, -1, -1, -1,  1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1, -1,  1,  1,  1,  1, -1,  1, -1,  1, -1, -1, -1, -1,  1,  1, -1,  0,  1, -1, -1,  1,  1,  1,  1, -1,  1, -1,  1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  0,  1,  1,  0,  0, -1,  0, -1, -1,  0, -1,  0,  0,  1,  1,  0, -1,  1,  0,  1, -1,  0,  1,  1,  0,  0, -1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1,  1,  1, -1,  1, -1,  1,  1, -1, -1,  1,  1, -1, -1,  1, -1,  1, -1, -1, -1, -1,  0,  1,  1,  1,  1, -1,  1, -1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1,  1,  1,  0,  1,  1,  1,  1,  0,  1,  1,  1,  1,  0,  1,  1,  1,  1,  0,  1,  1,  1,  1,  0,  1,  1,  1,  1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1, -1,  1,  1,  1,  1, -1,  1, -1, -1, -1,  1, -1, -1,  1, -1, -1, -1,  1, -1,  1,  1,  1,  1, -1, -1,  1,  0,  1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1, -1,  1,  1, -1,  1,  1,  1,  1, -1, -1, -1,  1, -1,  1, -1,  1,  1,  1, -1, -1, -1, -1,  1, -1, -1,  1, -1, -1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1,  0,  1, -1,  0, -1,  1,  0, -1,  0,  0, -1, -1,  0,  1,  1,  0, -1, -1,  0,  0, -1,  0,  1, -1,  0, -1,  1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  1,  1,  0, -1,  0, -1,  1,  0,  1,  1,  1,  0,  0,  1,  1, -1, -1,  0,  0, -1, -1, -1,  0, -1,  1,  0,  1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  1,  1, -1, -1,  1, -1,  1,  1,  1,  1, -1, -1, -1,  1, -1, -1, -1, -1,  1, -1, -1, -1,  1,  1,  1,  1, -1,  1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1,  0,  1,  1,  0, -1,  1,  0,  1,  1,  0,  0, -1,  0,  1, -1,  0, -1,  1,  0,  1, -1,  0,  1,  0,  0, -1, -1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1, -1,  1,  1, -1, -1,  1,  1,  1, -1, -1, -1, -1, -1,  1, -1,  1, -1,  1,  1, -1,  1, -1,  1, -1, -1, -1, -1, -1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1, -1,  1, -1,  1, -1, -1,  1,  1,  1, -1,  1,  1,  1,  1,  1, -1, -1, -1,  1, -1,  1,  1,  1, -1, -1, -1, -1, -1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  0,  1,  0,  0, -1, -1,  0,  0,  1,  0, -1,  1,  0,  1, -1,  0,  1,  0,  0, -1, -1,  0,  0,  1,  0, -1,  1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1,  1,  1, -1,  1,  1,  1,  1, -1, -1,  1, -1,  1, -1,  1,  1,  1, -1, -1,  1, -1, -1,  1,  1, -1,  1,  1, -1, -1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1,  1,  1,  1,  1,  0,  1,  1,  1,  1,  1,  1,  0,  1,  1,  1,  1,  1,  1,  0,  1,  1,  1,  1,  1,  1,  0,  1,  1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  0,  1,  1,  0, -1, -1,  0, -1,  1,  0,  1,  1,  0,  1,  0,  0,  1,  1,  0, -1,  1,  0,  1, -1,  0, -1,  1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1, -1,  1, -1,  1,  1, -1,  1,  1,  1, -1,  1, -1,  1,  1,  1, -1, -1, -1, -1, -1, -1,  1,  1, -1, -1,  1,  1, -1),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1, -1,  1,  0, -1,  1,  1,  1,  0,  0, -1,  1,  1,  0,  1,  1,  1, -1,  0, -1,  0, -1, -1,  0,  1, -1,  1, -1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1,  1,  0,  1, -1,  0,  1,  1,  0, -1, -1,  0, -1,  1,  0,  1, -1,  0,  0, -1,  0, -1, -1,  0,  1, -1,  0,  1,  1,  0),
      (0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0),
      (1, -1,  1,  1,  1, -1,  1, -1,  1, -1, -1,  1, -1, -1,  1,  1,  1, -1,  1,  1,  1,  1, -1, -1,  1,  1,  1,  1,  1, -1)
    );

  KroneckerSymbolValues: array[1..30] of array[1..30] of TASI =
  (
    (1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1),
    (1,  0, -1,  0, -1,  0,  1,  0,  1,  0, -1,  0, -1,  0,  1,  0,  1,  0, -1,  0, -1,  0,  1,  0,  1,  0, -1,  0, -1,  0),
    (1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0),
    (1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0),
    (1, -1, -1,  1,  0,  1, -1, -1,  1,  0,  1, -1, -1,  1,  0,  1, -1, -1,  1,  0,  1, -1, -1,  1,  0,  1, -1, -1,  1,  0),
    (1,  0,  0,  0,  1,  0,  1,  0,  0,  0,  1,  0, -1,  0,  0,  0, -1,  0, -1,  0,  0,  0, -1,  0,  1,  0,  0,  0,  1,  0),
    (1,  1, -1,  1, -1, -1,  0,  1,  1, -1,  1, -1, -1,  0,  1,  1, -1,  1, -1, -1,  0,  1,  1, -1,  1, -1, -1,  0,  1,  1),
    (1,  0, -1,  0, -1,  0,  1,  0,  1,  0, -1,  0, -1,  0,  1,  0,  1,  0, -1,  0, -1,  0,  1,  0,  1,  0, -1,  0, -1,  0),
    (1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0,  1,  1,  0),
    (1,  0,  1,  0,  0,  0, -1,  0,  1,  0, -1,  0,  1,  0,  0,  0, -1,  0, -1,  0, -1,  0, -1,  0,  0,  0,  1,  0, -1,  0),
    (1, -1,  1,  1,  1, -1, -1, -1,  1, -1,  0,  1, -1,  1,  1,  1, -1, -1, -1,  1, -1,  0,  1, -1,  1,  1,  1, -1, -1, -1),
    (1,  0,  0,  0, -1,  0,  1,  0,  0,  0, -1,  0,  1,  0,  0,  0, -1,  0,  1,  0,  0,  0, -1,  0,  1,  0,  0,  0, -1,  0),
    (1, -1,  1,  1, -1, -1, -1, -1,  1,  1, -1,  1,  0,  1, -1,  1,  1, -1, -1, -1, -1,  1,  1, -1,  1,  0,  1, -1,  1,  1),
    (1,  0,  1,  0,  1,  0,  0,  0,  1,  0, -1,  0,  1,  0,  1,  0, -1,  0,  1,  0,  0,  0,  1,  0,  1,  0,  1,  0, -1,  0),
    (1,  1,  0,  1,  0,  0, -1,  1,  0,  0, -1,  0, -1, -1,  0,  1,  1,  0,  1,  0,  0, -1,  1,  0,  0, -1,  0, -1, -1,  0),
    (1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0),
    (1,  1, -1,  1, -1, -1, -1,  1,  1, -1, -1, -1,  1, -1,  1,  1,  0,  1,  1, -1,  1, -1, -1, -1,  1,  1, -1, -1, -1,  1),
    (1,  0,  0,  0, -1,  0,  1,  0,  0,  0, -1,  0, -1,  0,  0,  0,  1,  0, -1,  0,  0,  0,  1,  0,  1,  0,  0,  0, -1,  0),
    (1, -1, -1,  1,  1,  1,  1, -1,  1, -1,  1, -1, -1, -1, -1,  1,  1, -1,  0,  1, -1, -1,  1,  1,  1,  1, -1,  1, -1,  1),
    (1,  0, -1,  0,  0,  0, -1,  0,  1,  0,  1,  0, -1,  0,  0,  0, -1,  0,  1,  0,  1,  0, -1,  0,  0,  0, -1,  0,  1,  0),
    (1, -1,  0,  1,  1,  0,  0, -1,  0, -1, -1,  0, -1,  0,  0,  1,  1,  0, -1,  1,  0,  1, -1,  0,  1,  1,  0,  0, -1,  0),
    (1,  0, -1,  0, -1,  0, -1,  0,  1,  0,  0,  0,  1,  0,  1,  0, -1,  0,  1,  0,  1,  0,  1,  0,  1,  0, -1,  0,  1,  0),
    (1,  1,  1,  1, -1,  1, -1,  1,  1, -1, -1,  1,  1, -1, -1,  1, -1,  1, -1, -1, -1, -1,  0,  1,  1,  1,  1, -1,  1, -1),
    (1,  0,  0,  0,  1,  0,  1,  0,  0,  0,  1,  0, -1,  0,  0,  0, -1,  0, -1,  0,  0,  0, -1,  0,  1,  0,  0,  0,  1,  0),
    (1,  1,  1,  1,  0,  1,  1,  1,  1,  0,  1,  1,  1,  1,  0,  1,  1,  1,  1,  0,  1,  1,  1,  1,  0,  1,  1,  1,  1,  0),
    (1,  0, -1,  0,  1,  0, -1,  0,  1,  0,  1,  0,  0,  0, -1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0, -1,  0, -1,  0),
    (1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0,  1, -1,  0),
    (1,  0, -1,  0, -1,  0,  0,  0,  1,  0,  1,  0, -1,  0,  1,  0, -1,  0, -1,  0,  0,  0,  1,  0,  1,  0, -1,  0,  1,  0),
    (1, -1, -1,  1,  1,  1,  1, -1,  1, -1, -1, -1,  1, -1, -1,  1, -1, -1, -1,  1, -1,  1,  1,  1,  1, -1, -1,  1,  0,  1),
    (1,  0,  0,  0,  0,  0, -1,  0,  0,  0,  1,  0,  1,  0,  0,  0,  1,  0, -1,  0,  0,  0,  1,  0,  0,  0,  0,  0,  1,  0)
  );

  Totients: array[1..500] of TASI =
    (
      1, 1, 2, 2, 4, 2, 6, 4, 6, 4, 10, 4, 12, 6, 8, 8, 16, 6, 18, 8, 12, 10, 22, 8, 20, 12, 18, 12, 28, 8, 30, 16, 20, 16,
      24, 12, 36, 18, 24, 16, 40, 12, 42, 20, 24, 22, 46, 16, 42, 20, 32, 24, 52, 18, 40, 24, 36, 28, 58, 16, 60, 30, 36, 32,
      48, 20, 66, 32, 44, 24, 70, 24, 72, 36, 40, 36, 60, 24, 78, 32, 54, 40, 82, 24, 64, 42, 56, 40, 88, 24, 72, 44, 60, 46,
      72, 32, 96, 42, 60, 40, 100, 32, 102, 48, 48, 52, 106, 36, 108, 40, 72, 48, 112, 36, 88, 56, 72, 58, 96, 32, 110, 60,
      80, 60, 100, 36, 126, 64, 84, 48, 130, 40, 108, 66, 72, 64, 136, 44, 138, 48, 92, 70, 120, 48, 112, 72, 84, 72, 148,
      40, 150, 72, 96, 60, 120, 48, 156, 78, 104, 64, 132, 54, 162, 80, 80, 82, 166, 48, 156, 64, 108, 84, 172, 56, 120, 80,
      116, 88, 178, 48, 180, 72, 120, 88, 144, 60, 160, 92, 108, 72, 190, 64, 192, 96, 96, 84, 196, 60, 198, 80, 132, 100,
      168, 64, 160, 102, 132, 96, 180, 48, 210, 104, 140, 106, 168, 72, 180, 108, 144, 80, 192, 72, 222, 96, 120, 112, 226,
      72, 228, 88, 120, 112, 232, 72, 184, 116, 156, 96, 238, 64, 240, 110, 162, 120, 168, 80, 216, 120, 164, 100, 250, 72,
      220, 126, 128, 128, 256, 84, 216, 96, 168, 130, 262, 80, 208, 108, 176, 132, 268, 72, 270, 128, 144, 136, 200, 88, 276,
      138, 180, 96, 280, 92, 282, 140, 144, 120, 240, 96, 272, 112, 192, 144, 292, 84, 232, 144, 180, 148, 264, 80, 252, 150,
      200, 144, 240, 96, 306, 120, 204, 120, 310, 96, 312, 156, 144, 156, 316, 104, 280, 128, 212, 132, 288, 108, 240, 162,
      216, 160, 276, 80, 330, 164, 216, 166, 264, 96, 336, 156, 224, 128, 300, 108, 294, 168, 176, 172, 346, 112, 348, 120,
      216, 160, 352, 116, 280, 176, 192, 178, 358, 96, 342, 180, 220, 144, 288, 120, 366, 176, 240, 144, 312, 120, 372, 160,
      200, 184, 336, 108, 378, 144, 252, 190, 382, 128, 240, 192, 252, 192, 388, 96, 352, 168, 260, 196, 312, 120, 396, 198,
      216, 160, 400, 132, 360, 200, 216, 168, 360, 128, 408, 160, 272, 204, 348, 132, 328, 192, 276, 180, 418, 96, 420, 210,
      276, 208, 320, 140, 360, 212, 240, 168, 430, 144, 432, 180, 224, 216, 396, 144, 438, 160, 252, 192, 442, 144, 352, 222,
      296, 192, 448, 120, 400, 224, 300, 226, 288, 144, 456, 228, 288, 176, 460, 120, 462, 224, 240, 232, 466, 144, 396, 184,
      312, 232, 420, 156, 360, 192, 312, 238, 478, 128, 432, 240, 264, 220, 384, 162, 486, 240, 324, 168, 490, 160, 448, 216,
      240, 240, 420, 164, 498, 200
    );

  Radicals: array[1..78] of TASI =
    (
      1, 2, 3, 2, 5, 6, 7, 2, 3, 10, 11, 6, 13, 14, 15, 2, 17, 6, 19, 10, 21, 22, 23, 6, 5, 26, 3, 14, 29, 30, 31, 2, 33, 34,
      35, 6, 37, 38, 39, 10, 41, 42, 43, 22, 15, 46, 47, 6, 7, 10, 51, 26, 53, 6, 55, 14, 57, 58, 59, 30, 61, 62, 21, 2, 65,
      66, 67, 34, 69, 70, 71, 6, 73, 74, 15, 38, 77, 78
    );

function IntInArray(const AValue: Integer; const AArray: array of Integer): Boolean;
var
  i: Integer;
begin
  for i := Low(AArray) to High(AArray) do
    if AArray[i] = AValue then
      Exit(True);
  Result := False;
end;

function IsSmallPrime(const AValue: Integer): Boolean;
begin
  if AValue > SmallPrimes[High(SmallPrimes)] then
    raise Exception.Create('IsSmallPrime: Prime candidate not small enough.');
  Result := IntInArray(AValue, SmallPrimes);
end;

procedure TSelfTester.TestRealMatrixQ(const AFcnName: string; const APositives: AnsiString);
var
  c: AnsiChar;
  x: set of AnsiChar;
begin
  x := [];
  for c in RealMatrixNames do
    Include(x, c);
  for c in APositives do
  begin
    Test(Format('%s(%s)', [AFcnName, c]), True);
    Exclude(x, c);
  end;
  for c in x do
    Test(Format('%s(%s)', [AFcnName, c]), False);
end;

procedure TSelfTester.TestRealMatrixQImplies(const ALeft, ARight: string);
var
  c: AnsiChar;
begin
  for c in RealMatrixNames do
    Test(Format('%s(%s) ⇒ %s(%s)', [ALeft, c, ARight, c]), True);
end;

procedure AvoidDateChange;
begin
  if SecondOfTheDay(Now) > 86397 then
    Sleep(5000);
end;


{$ENDREGION}

procedure TSelfTester.TestSeq;
var
  _i, _j: Integer;
  _x: TASR;
  fn, _s, _s2: string;
  _xarr, _yarr: TArray<TASR>;
  _cxarr, _cyarr: TArray<TASC>;
  _sxarr, _syarr: TArray<string>;
  _st: TStructType;
  _bin: TArray<Byte>;
  _bm: TBitmap;
  _png: TPngImage;
label
  Bookmark;
begin

  Test('SaveHistory(false)', null);

//  goto Bookmark;

  //
  //
  //  CHAPTER 1
  //  Literals and basic output formatting. Member access by value.
  //
  //

  Chapter('Literals, output formatting, member access by value');

  //
  // Integers
  //

  Test('0', 0, True, '0');
  Test('1', 1, True, '1');
  Test('2', 2, True, '2');
  Test('-1', -1, True, MINUS_SIGN + '1');
  Test(HYPHEN_MINUS + '1', -1, True, MINUS_SIGN + '1');         Assert(HYPHEN_MINUS = #$2D);
  Test(MINUS_SIGN + '1', -1, True, MINUS_SIGN + '1');           Assert(MINUS_SIGN = #$2212);
  Test('123', 123, True, '123');
  Test('-123', -123, True, MINUS_SIGN + '123');
  Test('123456789', 123456789, True, '123456789');
  Test('-123456789', -123456789, True, MINUS_SIGN + '123456789');
  Test('123456789123', 123456789123, True, '123456789123');
  Test('-123456789123', -123456789123, True, MINUS_SIGN + '123456789123');
  Test('123456789123456', 123456789123456, True, '123456789123456');
  Test('-123456789123456', -123456789123456, True, MINUS_SIGN + '123456789123456');
  Test('123456789123456789', 123456789123456789, True, '123456789123456789');
  Test('-123456789123456789', -123456789123456789, True, MINUS_SIGN + '123456789123456789');

  Test('10#0', 0, True, '0');
  Test('10#1', 1, True, '1');
  Test('-10#1', -1, True, MINUS_SIGN + '1');
  Test('10#123456', 123456, True, '123456');
  Test('-10#123456', -123456, True, MINUS_SIGN + '123456');

  Test('16#0', 0, True, '0');
  Test('16#1', 1, True, '1');
  Test('16#10', 16, True, '16');
  Test('16#FF', 255, True, '255');
  Test('-16#FF', -255, True, MINUS_SIGN + '255');
  Test('-16#ff', -255, True, MINUS_SIGN + '255');
  Test('-16#Ff', -255, True, MINUS_SIGN + '255');
  Test('-16#fF', -255, True, MINUS_SIGN + '255');
  Test('16#FFFFFFFF', 4294967295, True, '4294967295');
  Test('16#FFFFFFFFFFFFFFF', 1152921504606846975, True, '1152921504606846975');
  Test('16#fffffffffffffff', 1152921504606846975, True, '1152921504606846975');
  Test('16#20', 32, True, '32');
  Test('16#123ABCDEF', 4893429231, True, '4893429231');
  Test('16#A1B8EF81FA81FBA', 728332961084612538, True, '728332961084612538');
  Test('-16#A1B8EF81FA81FBA', -728332961084612538, True, MINUS_SIGN + '728332961084612538');
  Test('16#7FFFFFFFFFFFFFFF', 9223372036854775807, True, '9223372036854775807');
  Test('-16#7FFFFFFFFFFFFFFF', -9223372036854775807, True, MINUS_SIGN + '9223372036854775807');
  Test('16#8000000000000000', ESyntaxException, 'invalid numeric literal');

  Test('8#0', 0, True, '0');
  Test('8#1', 1, True, '1');
  Test('8#10', 8, True, '8');
  Test('8#11', 9, True, '9');
  Test('8#12', 10, True, '10');
  Test('8#15260', 6832, True, '6832');
  Test('8#4741716513576', 339557914494, True, '339557914494');
  Test('8#115651564324324455454', 1402087635523099436, True, '1402087635523099436');
  Test('-8#115651564324324455454', -1402087635523099436, True, MINUS_SIGN + '1402087635523099436');
  Test('8#777777777777777777777', 9223372036854775807, True, '9223372036854775807');
  Test('8#1000000000000000000000', ESyntaxException, 'invalid numeric literal');

  Test('2#0', 0, True, '0');
  Test('2#1', 1, True, '1');
  Test('2#10', 2, True, '2');
  Test('2#100', 4, True, '4');
  Test('2#1000', 8, True, '8');
  Test('2#10000', 16, True, '16');
  Test('2#100000', 32, True, '32');
  Test('2#1000000', 64, True, '64');
  Test('2#10000000', 128, True, '128');
  Test('2#100000000', 256, True, '256');
  Test('2#1000000000', 512, True, '512');
  Test('2#10000000000', 1024, True, '1024');
  Test('2#1000000000000000000000000', 16777216, True, '16777216');
  Test('2#100000000000000000000000000000000', 4294967296, True, '4294967296');
  Test('2#10000000000000000000000000000000000000000', 1099511627776, True, '1099511627776');
  Test('2#1011101010110110101100', 3059116, True, '3059116');
  Test('2#1001110101110100101010101010101', 1320834389, True, '1320834389');
  Test('-2#1001110101110100101010101010101', -1320834389, True, MINUS_SIGN + '1320834389');
  Test('2#101011110101010110101010101010101010101010110101010101010100110', 6317095989821483686, True, '6317095989821483686');
  Test('-2#101011110101010110101010101010101010101010110101010101010100110', -6317095989821483686, True, MINUS_SIGN + '6317095989821483686');
  Test('2#111111111111111111111111111111111111111111111111111111111111111', 9223372036854775807, True, '9223372036854775807');
  Test('2#1000000000000000000000000000000000000000000000000000000000000000', ESyntaxException, 'invalid numeric literal');

  Test('5#0', 0);
  Test('5#1', 1);
  Test('5#10', 5);
  Test('5#11', 6);
  Test('5#100', 25);
  Test('5#110', 30);
  Test('5#111', 31);
  Test('5#112', 32);
  Test('5#1000', 125);
  Test('5#123', 38);
  Test('5#1104332401304422434310311212', 9223372036854775807, True, '9223372036854775807');
  Test('5#1104332401304422434310311213', ESyntaxException, 'invalid numeric literal');

  Test('20#0', 0);
  Test('20#1', 1);
  Test('20#2', 2);
  Test('20#3', 3);
  Test('20#4', 4);
  Test('20#9', 9);
  Test('20#A', 10);
  Test('20#B', 11);
  Test('20#C', 12);
  Test('20#D', 13);
  Test('20#E', 14);
  Test('20#F', 15);
  Test('20#G', 16);
  Test('20#H', 17);
  Test('20#I', 18);
  Test('20#J', 19);
  Test('20#K', ESyntaxException, 'invalid base-20 digit "K" in numeric literal "20#K"');
  Test('20#10', 20);
  Test('20#11', 21);
  Test('20#12', 22);
  Test('20#123', 443);
  Test('20#1000', 8000);
  Test('20#a', 10);
  Test('20#b', 11);
  Test('20#c', 12);
  Test('20#5cbfjia3fh26ja7', 9223372036854775807, True, '9223372036854775807');
  Test('20#5cbfjia3fh26ja8', ESyntaxException, 'invalid numeric literal');

  Test('2#11010101010', 1706);
  Test('2#11010101020', ESyntaxException, 'invalid base-2 digit "2" in numeric literal "2#11010101020"');

  Test('10#123456', 123456);
  Test('10#12345A', ESyntaxException, 'invalid base-10 digit "A" in numeric literal "10#12345A"');

  Test('16#12345A', 1193050);
  Test('16#12345B', 1193051);
  Test('16#12345C', 1193052);
  Test('16#12345D', 1193053);
  Test('16#12345E', 1193054);
  Test('16#12345F', 1193055);
  Test('16#12345G', ESyntaxException, 'invalid base-16 digit "G" in numeric literal "16#12345G"');
  Test('16#123460', 1193056);
  Test('16#12345f', 1193055);

  Test('8#1234567', 342391);
  Test('8#1234568', ESyntaxException, 'invalid base-8 digit "8" in numeric literal "8#1234568"');
  Test('8#1234570', 342392);

  Test('36#AZ', 395);
  Test('36#ZZZZZ', 60466175);
  Test('36#az', 395);
  Test('36#zzzzz', 60466175);
  Test('36#ZzzzZ', 60466175);
  Test('36#1q7tyotd8hj0j', 8189181848448181891, True, '8189181848448181891');
  Test('-36#1q7tyotd8hj0j', -8189181848448181891, True, MINUS_SIGN + '8189181848448181891');
  Test('36#1y2p0ij32e8e7', 9223372036854775807, True, '9223372036854775807');
  Test('36#1y2p0ij32e8e8', ESyntaxException, 'invalid numeric literal');

  Test('1#156', ESyntaxException, 'invalid number base 1 in numeric literal "1#156"');
  Test('0#0', ESyntaxException, 'invalid number base 0 in numeric literal "0#0"');
  Test('37#123', ESyntaxException, 'invalid number base 37 in numeric literal "37#123"');
  Test('16#katt', ESyntaxException, 'invalid base-16 digit "k" in numeric literal "16#katt"');

  Test('36#0', 0, True, '0');
  Test('36#1', 1, True, '1');
  Test('36#2', 2, True, '2');
  Test('36#3', 3);
  Test('36#4', 4);
  Test('36#5', 5);
  Test('36#6', 6);
  Test('36#7', 7);
  Test('36#8', 8);
  Test('36#9', 9);
  Test('36#A', 10);
  Test('36#B', 11);
  Test('36#C', 12);
  Test('36#D', 13, True, '13');
  Test('36#E', 14);
  Test('36#F', 15);
  Test('36#G', 16);
  Test('36#H', 17);
  Test('36#I', 18);
  Test('36#J', 19);
  Test('36#K', 20);
  Test('36#L', 21);
  Test('36#M', 22);
  Test('36#N', 23);
  Test('36#O', 24, True, '24');
  Test('36#P', 25);
  Test('36#Q', 26);
  Test('36#R', 27);
  Test('36#S', 28);
  Test('36#T', 29);
  Test('36#U', 30);
  Test('36#V', 31);
  Test('36#W', 32);
  Test('36#X', 33);
  Test('36#Y', 34);
  Test('36#Z', 35, True, '35');
  Test('36#10', 36, True, '36');
  Test('36#11', 37, True, '37');

  Test('16#abc', 2748);
  Test('16#abc#2', ESyntaxException, 'invalid numeric literal "16#abc#2"');
  Test('1#2#3', ESyntaxException, 'invalid numeric literal "1#2#3"');

  // Format: number base

  Test('255', 255, True, '255');
  Test('SetNumberBase(255, 10)', 255, True, '255');
  Test('SetNumberBase(255, 16)', 255, True, 'FF');
  Test('SetNumberBase(255, 2)', 255, True, '11111111');
  Test('SetNumberBase(255, 3)', 255, True, '100110');
  Test('SetNumberBase(30, 30)', 30, True, '10');
  Test('SetNumberBase(36, 36)', 36, True, '10');
  Test('SetNumberBase(2#11111111, 2)', 255, True, '11111111');
  Test('SetNumberBase(2#11111111, 16)', 255, True, 'FF');

  Test('SetNumberBase(255, 1)', failure, 'Invalid number base 1.');
  Test('SetNumberBase(255, 37)', failure, 'Invalid number base 37.');
  Test('SetNumberBase(255, 999)', failure, 'Invalid number base 999.');
  Test('SetNumberBase(255, 0)', 255, True, '255');
  Test('SetNumberBase(255, -1)', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.');
  Test('SetNumberBase(255, 2.2)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');
  Test('SetNumberBase(255, "t")', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
  Test('SetNumberBase(255, 16, 5)', failure, 'Too many arguments.');
  Test('SetNumberBase(255)', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('-0', 0, True, '0');
  Test('-255', -255, True, MINUS_SIGN + '255');
  Test('SetNumberBase(-255, 0)', -255, True, MINUS_SIGN + '255');
  Test('SetNumberBase(-255, 10)', -255, True, MINUS_SIGN + '255');
  Test('SetNumberBase(-255, 16)', -255, True, MINUS_SIGN + 'FF');
  Test('SetNumberBase(-255, 2)', -255, True, MINUS_SIGN + '11111111');
  Test('SetNumberBase(-255, 3)', -255, True, MINUS_SIGN + '100110');
  Test('SetNumberBase(-30, 30)', -30, True, MINUS_SIGN + '10');
  Test('SetNumberBase(-36, 36)', -36, True, MINUS_SIGN + '10');
  Test('SetNumberBase(-2#11111111, 2)', -255, True, MINUS_SIGN + '11111111');
  Test('SetNumberBase(-2#11111111, 16)', -255, True, MINUS_SIGN + 'FF');

  Test('SetNumberBase(123456789123456, 0)', 123456789123456, True, '123456789123456');
  Test('SetNumberBase(123456789123456, 10)', 123456789123456, True, '123456789123456');
  Test('SetNumberBase(123456789123456, 16)', 123456789123456, True, '7048860F9180');
  Test('SetNumberBase(123456789123456, 8)', 123456789123456, True, '3404420603710600');
  Test('SetNumberBase(123456789123456, 2)', 123456789123456, True, '11100000100100010000110000011111001000110000000');
  Test('SetNumberBase(123456789123456, 3)', 123456789123456, True, '121012010100112220101001120010');
  Test('SetNumberBase(123456789123456, 32)', 123456789123456, True, '3G9230V4C0');
  Test('SetNumberBase(123456789123456, 36)', 123456789123456, True, '17RF9KOMTC');

  Test('SetNumberBase(-123456789123456, 0)', -123456789123456, True, MINUS_SIGN + '123456789123456');
  Test('SetNumberBase(-123456789123456, 10)', -123456789123456, True, MINUS_SIGN + '123456789123456');
  Test('SetNumberBase(-123456789123456, 16)', -123456789123456, True, MINUS_SIGN + '7048860F9180');
  Test('SetNumberBase(-123456789123456, 8)', -123456789123456, True, MINUS_SIGN + '3404420603710600');
  Test('SetNumberBase(-123456789123456, 2)', -123456789123456, True, MINUS_SIGN + '11100000100100010000110000011111001000110000000');
  Test('SetNumberBase(-123456789123456, 3)', -123456789123456, True, MINUS_SIGN + '121012010100112220101001120010');
  Test('SetNumberBase(-123456789123456, 32)', -123456789123456, True, MINUS_SIGN + '3G9230V4C0');
  Test('SetNumberBase(-123456789123456, 36)', -123456789123456, True, MINUS_SIGN + '17RF9KOMTC');

  Test('SetNumberBase(123456789123456789, 0)', 123456789123456789, True, '123456789123456789');
  Test('SetNumberBase(123456789123456789, 10)', 123456789123456789, True, '123456789123456789');
  Test('SetNumberBase(123456789123456789, 16)', 123456789123456789, True, '1B69B4BACD05F15');
  Test('SetNumberBase(123456789123456789, 8)', 123456789123456789, True, '6664664565464057425');
  Test('SetNumberBase(123456789123456789, 2)', 123456789123456789, True, '110110110100110110100101110101100110100000101111100010101');
  Test('SetNumberBase(123456789123456789, 3)', 123456789123456789, True, '211012121210012202221200111210202100');
  Test('SetNumberBase(8482137697523097365, 2)', 8482137697523097365, True, '111010110110110100110110100101110101100110100000101111100010101');

  Test('SetNumberBase(-123456789123456789, 0)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetNumberBase(-123456789123456789, 10)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetNumberBase(-123456789123456789, 16)', -123456789123456789, True, MINUS_SIGN + '1B69B4BACD05F15');
  Test('SetNumberBase(-123456789123456789, 8)', -123456789123456789, True, MINUS_SIGN + '6664664565464057425');
  Test('SetNumberBase(-123456789123456789, 2)', -123456789123456789, True, MINUS_SIGN + '110110110100110110100101110101100110100000101111100010101');
  Test('SetNumberBase(-123456789123456789, 3)', -123456789123456789, True, MINUS_SIGN + '211012121210012202221200111210202100');
  Test('SetNumberBase(-8482137697523097365, 2)', -8482137697523097365, True, MINUS_SIGN + '111010110110110100110110100101110101100110100000101111100010101');

  Test('SetNumberBase(0, 0)', 0, True, '0');
  Test('SetNumberBase(0, 2)', 0, True, '0');
  Test('SetNumberBase(0, 5)', 0, True, '0');
  Test('SetNumberBase(0, 8)', 0, True, '0');
  Test('SetNumberBase(0, 10)', 0, True, '0');
  Test('SetNumberBase(0, 16)', 0, True, '0');
  Test('SetNumberBase(0, 36)', 0, True, '0');

  Test('SetNumberBase(1, 0)', 1, True, '1');
  Test('SetNumberBase(1, 2)', 1, True, '1');
  Test('SetNumberBase(1, 5)', 1, True, '1');
  Test('SetNumberBase(1, 8)', 1, True, '1');
  Test('SetNumberBase(1, 10)', 1, True, '1');
  Test('SetNumberBase(1, 16)', 1, True, '1');
  Test('SetNumberBase(1, 36)', 1, True, '1');

  Test('SetNumberBase(2, 0)', 2, True, '2');
  Test('SetNumberBase(2, 2)', 2, True, '10');
  Test('SetNumberBase(2, 5)', 2, True, '2');
  Test('SetNumberBase(2, 8)', 2, True, '2');
  Test('SetNumberBase(2, 10)', 2, True, '2');
  Test('SetNumberBase(2, 16)', 2, True, '2');
  Test('SetNumberBase(2, 36)', 2, True, '2');

  Test('SetNumberBase(-2, 0)', -2, True, MINUS_SIGN + '2');
  Test('SetNumberBase(-2, 2)', -2, True, MINUS_SIGN + '10');
  Test('SetNumberBase(-2, 5)', -2, True, MINUS_SIGN + '2');
  Test('SetNumberBase(-2, 8)', -2, True, MINUS_SIGN + '2');
  Test('SetNumberBase(-2, 10)', -2, True, MINUS_SIGN + '2');
  Test('SetNumberBase(-2, 16)', -2, True, MINUS_SIGN + '2');
  Test('SetNumberBase(-2, 36)', -2, True, MINUS_SIGN + '2');

  Test('SetNumberBase(2, 2)', 2, True, '10');
  Test('SetNumberBase(3, 3)', 3, True, '10');
  Test('SetNumberBase(4, 4)', 4, True, '10');
  Test('SetNumberBase(5, 5)', 5, True, '10');
  Test('SetNumberBase(6, 6)', 6, True, '10');
  Test('SetNumberBase(7, 7)', 7, True, '10');
  Test('SetNumberBase(8, 8)', 8, True, '10');
  Test('SetNumberBase(9, 9)', 9, True, '10');
  Test('SetNumberBase(10, 10)', 10, True, '10');
  Test('SetNumberBase(11, 11)', 11, True, '10');

  Test('SetNumberBase(0, 36)', 0, True, '0');
  Test('SetNumberBase(1, 36)', 1, True, '1');
  Test('SetNumberBase(2, 36)', 2, True, '2');
  Test('SetNumberBase(3, 36)', 3, True, '3');
  Test('SetNumberBase(4, 36)', 4, True, '4');
  Test('SetNumberBase(5, 36)', 5, True, '5');
  Test('SetNumberBase(6, 36)', 6, True, '6');
  Test('SetNumberBase(7, 36)', 7, True, '7');
  Test('SetNumberBase(8, 36)', 8, True, '8');
  Test('SetNumberBase(9, 36)', 9, True, '9');
  Test('SetNumberBase(10, 36)', 10, True, 'A');
  Test('SetNumberBase(11, 36)', 11, True, 'B');
  Test('SetNumberBase(12, 36)', 12, True, 'C');
  Test('SetNumberBase(13, 36)', 13, True, 'D');
  Test('SetNumberBase(14, 36)', 14, True, 'E');
  Test('SetNumberBase(15, 36)', 15, True, 'F');
  Test('SetNumberBase(16, 36)', 16, True, 'G');
  Test('SetNumberBase(17, 36)', 17, True, 'H');
  Test('SetNumberBase(18, 36)', 18, True, 'I');
  Test('SetNumberBase(19, 36)', 19, True, 'J');
  Test('SetNumberBase(20, 36)', 20, True, 'K');
  Test('SetNumberBase(21, 36)', 21, True, 'L');
  Test('SetNumberBase(22, 36)', 22, True, 'M');
  Test('SetNumberBase(23, 36)', 23, True, 'N');
  Test('SetNumberBase(24, 36)', 24, True, 'O');
  Test('SetNumberBase(25, 36)', 25, True, 'P');
  Test('SetNumberBase(26, 36)', 26, True, 'Q');
  Test('SetNumberBase(27, 36)', 27, True, 'R');
  Test('SetNumberBase(28, 36)', 28, True, 'S');
  Test('SetNumberBase(29, 36)', 29, True, 'T');
  Test('SetNumberBase(30, 36)', 30, True, 'U');
  Test('SetNumberBase(31, 36)', 31, True, 'V');
  Test('SetNumberBase(32, 36)', 32, True, 'W');
  Test('SetNumberBase(33, 36)', 33, True, 'X');
  Test('SetNumberBase(34, 36)', 34, True, 'Y');
  Test('SetNumberBase(35, 36)', 35, True, 'Z');
  Test('SetNumberBase(36, 36)', 36, True, '10');
  Test('SetNumberBase(37, 36)', 37, True, '11');

  Test('SetNumberBase(856324759261834725, 2)', 856324759261834725, True, '101111100010010001101101010011100000001011000110110111100101');
  Test('SetNumberBase(856324759261834725, 3)', 856324759261834725, True, '12201001010002100111010102020021220020');
  Test('SetNumberBase(856324759261834725, 4)', 856324759261834725, True, '233202101231103200023012313211');
  Test('SetNumberBase(856324759261834725, 5)', 856324759261834725, True, '24140410144323432012202400');
  Test('SetNumberBase(856324759261834725, 6)', 856324759261834725, True, '103011414431331422425353');
  Test('SetNumberBase(856324759261834725, 7)', 856324759261834725, True, '1350602331133464501312');
  Test('SetNumberBase(856324759261834725, 8)', 856324759261834725, True, '57422155234013066745');
  Test('SetNumberBase(856324759261834725, 16)', 856324759261834725, True, 'BE246D4E02C6DE5');
  Test('SetNumberBase(856324759261834725, 32)', 856324759261834725, True, 'NOI6QJG2ORF5');
  Test('SetNumberBase(856324759261834725, 36)', 856324759261834725, True, '6I7PSJLAEQXX');

  Test('SetNumberBase(-856324759261834725, 2)', -856324759261834725, True, MINUS_SIGN + '101111100010010001101101010011100000001011000110110111100101');
  Test('SetNumberBase(-856324759261834725, 3)', -856324759261834725, True, MINUS_SIGN + '12201001010002100111010102020021220020');
  Test('SetNumberBase(-856324759261834725, 4)', -856324759261834725, True, MINUS_SIGN + '233202101231103200023012313211');
  Test('SetNumberBase(-856324759261834725, 5)', -856324759261834725, True, MINUS_SIGN + '24140410144323432012202400');
  Test('SetNumberBase(-856324759261834725, 6)', -856324759261834725, True, MINUS_SIGN + '103011414431331422425353');
  Test('SetNumberBase(-856324759261834725, 7)', -856324759261834725, True, MINUS_SIGN + '1350602331133464501312');
  Test('SetNumberBase(-856324759261834725, 8)', -856324759261834725, True, MINUS_SIGN + '57422155234013066745');
  Test('SetNumberBase(-856324759261834725, 16)', -856324759261834725, True, MINUS_SIGN + 'BE246D4E02C6DE5');
  Test('SetNumberBase(-856324759261834725, 32)', -856324759261834725, True, MINUS_SIGN + 'NOI6QJG2ORF5');
  Test('SetNumberBase(-856324759261834725, 36)', -856324759261834725, True, MINUS_SIGN + '6I7PSJLAEQXX');

  Test('SetNumberBase(9223372036854775807, 0)', 9223372036854775807, True, '9223372036854775807');
  Test('SetNumberBase(9223372036854775807, 10)', 9223372036854775807, True, '9223372036854775807');
  Test('SetNumberBase(9223372036854775807, 2)', 9223372036854775807, True, '111111111111111111111111111111111111111111111111111111111111111');
  Test('SetNumberBase(9223372036854775807, 8)', 9223372036854775807, True, '777777777777777777777');
  Test('SetNumberBase(9223372036854775807, 16)', 9223372036854775807, True, '7FFFFFFFFFFFFFFF');
  Test('SetNumberBase(9223372036854775807, 36)', 9223372036854775807, True, '1Y2P0IJ32E8E7');

  // Format: Minimum length

  Test('SetMinLength(123, 0)', 123, True, '123');
  Test('SetMinLength(123, 1)', 123, True, '123');
  Test('SetMinLength(123, 2)', 123, True, '123');
  Test('SetMinLength(123, 3)', 123, True, '123');
  Test('SetMinLength(123, 4)', 123, True, '0123');
  Test('SetMinLength(123, 5)', 123, True, '00123');
  Test('SetMinLength(123, 6)', 123, True, '000123');
  Test('SetMinLength(123, 7)', 123, True, '0000123');
  Test('SetMinLength(123, 8)', 123, True, '00000123');
  Test('SetMinLength(123, 9)', 123, True, '000000123');
  Test('SetMinLength(123, 10)', 123, True, '0000000123');
  Test('SetMinLength(123, 11)', 123, True, '00000000123');
  Test('SetMinLength(123, 12)', 123, True, '000000000123');
  Test('SetMinLength(123, 13)', 123, True, '0000000000123');
  Test('SetMinLength(123, 14)', 123, True, '00000000000123');
  Test('SetMinLength(123, 15)', 123, True, '000000000000123');
  Test('SetMinLength(123, 16)', 123, True, '0000000000000123');
  Test('SetMinLength(123, 17)', 123, True, '00000000000000123');
  Test('SetMinLength(123, 18)', 123, True, '000000000000000123');
  Test('SetMinLength(123, 19)', 123, True, '0000000000000000123');
  Test('SetMinLength(123, 20)', 123, True, '00000000000000000123');

  Test('SetMinLength(-123, 0)', -123, True, MINUS_SIGN + '123');
  Test('SetMinLength(-123, 1)', -123, True, MINUS_SIGN + '123');
  Test('SetMinLength(-123, 2)', -123, True, MINUS_SIGN + '123');
  Test('SetMinLength(-123, 3)', -123, True, MINUS_SIGN + '123');
  Test('SetMinLength(-123, 4)', -123, True, MINUS_SIGN + '0123');
  Test('SetMinLength(-123, 5)', -123, True, MINUS_SIGN + '00123');
  Test('SetMinLength(-123, 6)', -123, True, MINUS_SIGN + '000123');
  Test('SetMinLength(-123, 7)', -123, True, MINUS_SIGN + '0000123');
  Test('SetMinLength(-123, 8)', -123, True, MINUS_SIGN + '00000123');
  Test('SetMinLength(-123, 9)', -123, True, MINUS_SIGN + '000000123');
  Test('SetMinLength(-123, 10)', -123, True, MINUS_SIGN + '0000000123');
  Test('SetMinLength(-123, 11)', -123, True, MINUS_SIGN + '00000000123');
  Test('SetMinLength(-123, 12)', -123, True, MINUS_SIGN + '000000000123');
  Test('SetMinLength(-123, 13)', -123, True, MINUS_SIGN + '0000000000123');
  Test('SetMinLength(-123, 14)', -123, True, MINUS_SIGN + '00000000000123');
  Test('SetMinLength(-123, 15)', -123, True, MINUS_SIGN + '000000000000123');
  Test('SetMinLength(-123, 16)', -123, True, MINUS_SIGN + '0000000000000123');
  Test('SetMinLength(-123, 17)', -123, True, MINUS_SIGN + '00000000000000123');
  Test('SetMinLength(-123, 18)', -123, True, MINUS_SIGN + '000000000000000123');
  Test('SetMinLength(-123, 19)', -123, True, MINUS_SIGN + '0000000000000000123');
  Test('SetMinLength(-123, 20)', -123, True, MINUS_SIGN + '00000000000000000123');

  Test('SetMinLength(123, 100)', 123, True, '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000123');
  Test('SetMinLength(-123, 100)', -123, True, MINUS_SIGN + '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000123');

  Test('SetMinLength(0, 0)', 0, True, '0');
  Test('SetMinLength(0, 1)', 0, True, '0');
  Test('SetMinLength(0, 2)', 0, True, '00');
  Test('SetMinLength(0, 3)', 0, True, '000');
  Test('SetMinLength(0, 4)', 0, True, '0000');
  Test('SetMinLength(0, 5)', 0, True, '00000');
  Test('SetMinLength(0, 6)', 0, True, '000000');
  Test('SetMinLength(0, 20)', 0, True, '00000000000000000000');
  Test('SetMinLength(0, 100)', 0, True, '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000');

  Test('SetMinLength(-0, 0)', 0, True, '0');
  Test('SetMinLength(-0, 1)', 0, True, '0');
  Test('SetMinLength(-0, 2)', 0, True, '00');
  Test('SetMinLength(-0, 3)', 0, True, '000');
  Test('SetMinLength(-0, 4)', 0, True, '0000');
  Test('SetMinLength(-0, 5)', 0, True, '00000');
  Test('SetMinLength(-0, 6)', 0, True, '000000');
  Test('SetMinLength(-0, 20)', 0, True, '00000000000000000000');
  Test('SetMinLength(-0, 100)', 0, True, '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000');

  Test('SetMinLength(123456789123456789, 0)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 1)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 2)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 3)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 4)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 5)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 6)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 7)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 8)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 9)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 10)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 11)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 12)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 13)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 14)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 15)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 16)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 17)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 18)', 123456789123456789, True, '123456789123456789');
  Test('SetMinLength(123456789123456789, 19)', 123456789123456789, True, '0123456789123456789');
  Test('SetMinLength(123456789123456789, 20)', 123456789123456789, True, '00123456789123456789');
  Test('SetMinLength(123456789123456789, 21)', 123456789123456789, True, '000123456789123456789');
  Test('SetMinLength(123456789123456789, 22)', 123456789123456789, True, '0000123456789123456789');
  Test('SetMinLength(123456789123456789, 100)', 123456789123456789, True, '0000000000000000000000000000000000000000000000000000000000000000000000000000000000123456789123456789');

  Test('SetMinLength(-123456789123456789, 0)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 1)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 2)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 3)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 4)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 5)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 6)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 7)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 8)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 9)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 10)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 11)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 12)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 13)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 14)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 15)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 16)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 17)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 18)', -123456789123456789, True, MINUS_SIGN + '123456789123456789');
  Test('SetMinLength(-123456789123456789, 19)', -123456789123456789, True, MINUS_SIGN + '0123456789123456789');
  Test('SetMinLength(-123456789123456789, 20)', -123456789123456789, True, MINUS_SIGN + '00123456789123456789');
  Test('SetMinLength(-123456789123456789, 21)', -123456789123456789, True, MINUS_SIGN + '000123456789123456789');
  Test('SetMinLength(-123456789123456789, 22)', -123456789123456789, True, MINUS_SIGN + '0000123456789123456789');
  Test('SetMinLength(-123456789123456789, 100)', -123456789123456789, True, MINUS_SIGN + '0000000000000000000000000000000000000000000000000000000000000000000000000000000000123456789123456789');

  Test('SetMinLength(123, 255)', 123, True, StringOfChar('0', 252) + '123');
  Test('SetMinLength(123, 256)', failure, 'Invalid minimum length 256.');
  Test('SetMinLength(123, -1)', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.');
  Test('SetMinLength(123, 7.6)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');
  Test('SetMinLength(123, "h")', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
  Test('SetMinLength(123, 10, 20)', failure, 'Too many arguments.');
  Test('SetMinLength(123)', failure, 'Too few arguments. A required argument of type integer is missing.');

  // Format: number base + minimum length

  Test('SetMinLength(SetNumberBase(255, 10), 0)', 255, True, '255');
  Test('SetMinLength(SetNumberBase(255, 10), 1)', 255, True, '255');
  Test('SetMinLength(SetNumberBase(255, 10), 2)', 255, True, '255');
  Test('SetMinLength(SetNumberBase(255, 10), 3)', 255, True, '255');
  Test('SetMinLength(SetNumberBase(255, 10), 4)', 255, True, '0255');
  Test('SetMinLength(SetNumberBase(255, 10), 8)', 255, True, '00000255');
  Test('SetMinLength(SetNumberBase(255, 10), 16)', 255, True, '0000000000000255');
  Test('SetMinLength(SetNumberBase(255, 10), 100)', 255, True, '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000255');

  Test('SetMinLength(SetNumberBase(-255, 10), 0)', -255, True, MINUS_SIGN + '255');
  Test('SetMinLength(SetNumberBase(-255, 10), 1)', -255, True, MINUS_SIGN + '255');
  Test('SetMinLength(SetNumberBase(-255, 10), 2)', -255, True, MINUS_SIGN + '255');
  Test('SetMinLength(SetNumberBase(-255, 10), 3)', -255, True, MINUS_SIGN + '255');
  Test('SetMinLength(SetNumberBase(-255, 10), 4)', -255, True, MINUS_SIGN + '0255');
  Test('SetMinLength(SetNumberBase(-255, 10), 8)', -255, True, MINUS_SIGN + '00000255');
  Test('SetMinLength(SetNumberBase(-255, 10), 16)', -255, True, MINUS_SIGN + '0000000000000255');
  Test('SetMinLength(SetNumberBase(-255, 10), 100)', -255, True, MINUS_SIGN + '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000255');

  Test('SetMinLength(SetNumberBase(0, 10), 0)', 0, True, '0');
  Test('SetMinLength(SetNumberBase(0, 10), 1)', 0, True, '0');
  Test('SetMinLength(SetNumberBase(0, 10), 2)', 0, True, '00');
  Test('SetMinLength(SetNumberBase(0, 10), 8)', 0, True, '00000000');
  Test('SetMinLength(SetNumberBase(0, 10), 100)', 0, True, '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000');

  Test('SetMinLength(SetNumberBase(255, 16), 0)', 255, True, 'FF');
  Test('SetMinLength(SetNumberBase(255, 16), 1)', 255, True, 'FF');
  Test('SetMinLength(SetNumberBase(255, 16), 2)', 255, True, 'FF');
  Test('SetMinLength(SetNumberBase(255, 16), 3)', 255, True, '0FF');
  Test('SetMinLength(SetNumberBase(255, 16), 4)', 255, True, '00FF');
  Test('SetMinLength(SetNumberBase(255, 16), 8)', 255, True, '000000FF');
  Test('SetMinLength(SetNumberBase(255, 16), 16)', 255, True, '00000000000000FF');
  Test('SetMinLength(SetNumberBase(255, 16), 100)', 255, True, '00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000FF');

  Test('SetMinLength(SetNumberBase(-255, 16), 0)', -255, True, MINUS_SIGN + 'FF');
  Test('SetMinLength(SetNumberBase(-255, 16), 1)', -255, True, MINUS_SIGN + 'FF');
  Test('SetMinLength(SetNumberBase(-255, 16), 2)', -255, True, MINUS_SIGN + 'FF');
  Test('SetMinLength(SetNumberBase(-255, 16), 3)', -255, True, MINUS_SIGN + '0FF');
  Test('SetMinLength(SetNumberBase(-255, 16), 4)', -255, True, MINUS_SIGN + '00FF');
  Test('SetMinLength(SetNumberBase(-255, 16), 8)', -255, True, MINUS_SIGN + '000000FF');
  Test('SetMinLength(SetNumberBase(-255, 16), 16)', -255, True, MINUS_SIGN + '00000000000000FF');
  Test('SetMinLength(SetNumberBase(-255, 16), 100)', -255, True, MINUS_SIGN + '00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000FF');

  Test('SetMinLength(SetNumberBase(0, 16), 0)', 0, True, '0');
  Test('SetMinLength(SetNumberBase(0, 16), 1)', 0, True, '0');
  Test('SetMinLength(SetNumberBase(0, 16), 2)', 0, True, '00');
  Test('SetMinLength(SetNumberBase(0, 16), 8)', 0, True, '00000000');
  Test('SetMinLength(SetNumberBase(0, 16), 100)', 0, True, '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000');

  Test('SetMinLength(SetNumberBase(123456789123456789, 16), 0)', 123456789123456789, True, '1B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(123456789123456789, 16), 1)', 123456789123456789, True, '1B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(123456789123456789, 16), 8)', 123456789123456789, True, '1B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(123456789123456789, 16), 15)', 123456789123456789, True, '1B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(123456789123456789, 16), 16)', 123456789123456789, True, '01B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(123456789123456789, 16), 32)', 123456789123456789, True, '000000000000000001B69B4BACD05F15');

  Test('SetMinLength(SetNumberBase(-123456789123456789, 16), 0)', -123456789123456789, True, MINUS_SIGN + '1B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(-123456789123456789, 16), 1)', -123456789123456789, True, MINUS_SIGN + '1B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(-123456789123456789, 16), 8)', -123456789123456789, True, MINUS_SIGN + '1B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(-123456789123456789, 16), 15)', -123456789123456789, True, MINUS_SIGN + '1B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(-123456789123456789, 16), 16)', -123456789123456789, True, MINUS_SIGN + '01B69B4BACD05F15');
  Test('SetMinLength(SetNumberBase(-123456789123456789, 16), 32)', -123456789123456789, True, MINUS_SIGN + '000000000000000001B69B4BACD05F15');

  Test('SetMinLength(SetNumberBase(255, 2), 0)', 255, True, '11111111');
  Test('SetMinLength(SetNumberBase(255, 2), 1)', 255, True, '11111111');
  Test('SetMinLength(SetNumberBase(255, 2), 2)', 255, True, '11111111');
  Test('SetMinLength(SetNumberBase(255, 2), 3)', 255, True, '11111111');
  Test('SetMinLength(SetNumberBase(255, 2), 4)', 255, True, '11111111');
  Test('SetMinLength(SetNumberBase(255, 2), 5)', 255, True, '11111111');
  Test('SetMinLength(SetNumberBase(255, 2), 6)', 255, True, '11111111');
  Test('SetMinLength(SetNumberBase(255, 2), 7)', 255, True, '11111111');
  Test('SetMinLength(SetNumberBase(255, 2), 8)', 255, True, '11111111');
  Test('SetMinLength(SetNumberBase(255, 2), 9)', 255, True, '011111111');
  Test('SetMinLength(SetNumberBase(255, 2), 10)', 255, True, '0011111111');
  Test('SetMinLength(SetNumberBase(255, 2), 11)', 255, True, '00011111111');
  Test('SetMinLength(SetNumberBase(255, 2), 12)', 255, True, '000011111111');
  Test('SetMinLength(SetNumberBase(255, 2), 13)', 255, True, '0000011111111');
  Test('SetMinLength(SetNumberBase(255, 2), 14)', 255, True, '00000011111111');
  Test('SetMinLength(SetNumberBase(255, 2), 15)', 255, True, '000000011111111');
  Test('SetMinLength(SetNumberBase(255, 2), 16)', 255, True, '0000000011111111');
  Test('SetMinLength(SetNumberBase(255, 2), 32)', 255, True, '00000000000000000000000011111111');

  Test('SetMinLength(SetNumberBase(-255, 2), 0)', -255, True, MINUS_SIGN + '11111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 1)', -255, True, MINUS_SIGN + '11111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 2)', -255, True, MINUS_SIGN + '11111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 3)', -255, True, MINUS_SIGN + '11111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 4)', -255, True, MINUS_SIGN + '11111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 5)', -255, True, MINUS_SIGN + '11111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 6)', -255, True, MINUS_SIGN + '11111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 7)', -255, True, MINUS_SIGN + '11111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 8)', -255, True, MINUS_SIGN + '11111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 9)', -255, True, MINUS_SIGN + '011111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 10)', -255, True, MINUS_SIGN + '0011111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 11)', -255, True, MINUS_SIGN + '00011111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 12)', -255, True, MINUS_SIGN + '000011111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 13)', -255, True, MINUS_SIGN + '0000011111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 14)', -255, True, MINUS_SIGN + '00000011111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 15)', -255, True, MINUS_SIGN + '000000011111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 16)', -255, True, MINUS_SIGN + '0000000011111111');
  Test('SetMinLength(SetNumberBase(-255, 2), 32)', -255, True, MINUS_SIGN + '00000000000000000000000011111111');

  Test('SetMinLength(SetNumberBase(0, 2), 0)', 0, True, '0');
  Test('SetMinLength(SetNumberBase(0, 2), 1)', 0, True, '0');
  Test('SetMinLength(SetNumberBase(0, 2), 2)', 0, True, '00');
  Test('SetMinLength(SetNumberBase(0, 2), 8)', 0, True, '00000000');
  Test('SetMinLength(SetNumberBase(0, 2), 100)', 0, True, '0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000');

  Test('SetMinLength(SetNumberBase(123456789123456789, 2), 0)', 123456789123456789, True, '110110110100110110100101110101100110100000101111100010101');
  Test('SetMinLength(SetNumberBase(123456789123456789, 2), 1)', 123456789123456789, True, '110110110100110110100101110101100110100000101111100010101');
  Test('SetMinLength(SetNumberBase(123456789123456789, 2), 32)', 123456789123456789, True, '110110110100110110100101110101100110100000101111100010101');
  Test('SetMinLength(SetNumberBase(123456789123456789, 2), 57)', 123456789123456789, True, '110110110100110110100101110101100110100000101111100010101');
  Test('SetMinLength(SetNumberBase(123456789123456789, 2), 64)', 123456789123456789, True, '0000000110110110100110110100101110101100110100000101111100010101');

  Test('SetMinLength(SetNumberBase(-123456789123456789, 2), 0)', -123456789123456789, True, MINUS_SIGN + '110110110100110110100101110101100110100000101111100010101');
  Test('SetMinLength(SetNumberBase(-123456789123456789, 2), 1)', -123456789123456789, True, MINUS_SIGN + '110110110100110110100101110101100110100000101111100010101');
  Test('SetMinLength(SetNumberBase(-123456789123456789, 2), 32)', -123456789123456789, True, MINUS_SIGN + '110110110100110110100101110101100110100000101111100010101');
  Test('SetMinLength(SetNumberBase(-123456789123456789, 2), 57)', -123456789123456789, True, MINUS_SIGN + '110110110100110110100101110101100110100000101111100010101');
  Test('SetMinLength(SetNumberBase(-123456789123456789, 2), 64)', -123456789123456789, True, MINUS_SIGN + '0000000110110110100110110100101110101100110100000101111100010101');

  Test('SetMinLength(SetNumberBase(9223372036854775807, 36), 16)', 9223372036854775807, True, '0001Y2P0IJ32E8E7');

  Test('SetNumberBase(SetMinLength(-255, 8), 16)', -255, True, MINUS_SIGN + '000000FF');
  Test('SetNumberBase(SetMinLength(123456789123456789, 64), 2)', 123456789123456789, True, '0000000110110110100110110100101110101100110100000101111100010101');

  // Format: digit grouping

  Test('SetDigitGrouping(123456789, 3)', 123456789, True, fs('123 456 789'));

  Test('SetDigitGrouping(123456789, 0)', 123456789, True, fs('123456789'));
  Test('SetDigitGrouping(123456789, 1)', 123456789, True, fs('1 2 3 4 5 6 7 8 9'));
  Test('SetDigitGrouping(123456789, 2)', 123456789, True, fs('1 23 45 67 89'));
  Test('SetDigitGrouping(123456789, 3)', 123456789, True, fs('123 456 789'));
  Test('SetDigitGrouping(123456789, 4)', 123456789, True, fs('1 2345 6789'));
  Test('SetDigitGrouping(123456789, 5)', 123456789, True, fs('1234 56789'));
  Test('SetDigitGrouping(123456789, 6)', 123456789, True, fs('123 456789'));
  Test('SetDigitGrouping(123456789, 7)', 123456789, True, fs('12 3456789'));
  Test('SetDigitGrouping(123456789, 8)', 123456789, True, fs('1 23456789'));
  Test('SetDigitGrouping(123456789, 9)', 123456789, True, fs('123456789'));
  Test('SetDigitGrouping(123456789, 10)', 123456789, True, fs('123456789'));
  Test('SetDigitGrouping(123456789, 11)', 123456789, True, fs('123456789'));
  Test('SetDigitGrouping(123456789, 12)', 123456789, True, fs('123456789'));
  Test('SetDigitGrouping(123456789, 13)', 123456789, True, fs('123456789'));
  Test('SetDigitGrouping(123456789, 14)', 123456789, True, fs('123456789'));
  Test('SetDigitGrouping(123456789, 15)', 123456789, True, fs('123456789'));
  Test('SetDigitGrouping(123456789, 16)', 123456789, True, fs('123456789'));

  Test('SetDigitGrouping(-123456789, 0)', -123456789, True, MINUS_SIGN + fs('123456789'));
  Test('SetDigitGrouping(-123456789, 1)', -123456789, True, MINUS_SIGN + fs('1 2 3 4 5 6 7 8 9'));
  Test('SetDigitGrouping(-123456789, 2)', -123456789, True, MINUS_SIGN + fs('1 23 45 67 89'));
  Test('SetDigitGrouping(-123456789, 3)', -123456789, True, MINUS_SIGN + fs('123 456 789'));
  Test('SetDigitGrouping(-123456789, 4)', -123456789, True, MINUS_SIGN + fs('1 2345 6789'));
  Test('SetDigitGrouping(-123456789, 5)', -123456789, True, MINUS_SIGN + fs('1234 56789'));
  Test('SetDigitGrouping(-123456789, 6)', -123456789, True, MINUS_SIGN + fs('123 456789'));
  Test('SetDigitGrouping(-123456789, 7)', -123456789, True, MINUS_SIGN + fs('12 3456789'));
  Test('SetDigitGrouping(-123456789, 8)', -123456789, True, MINUS_SIGN + fs('1 23456789'));
  Test('SetDigitGrouping(-123456789, 9)', -123456789, True, MINUS_SIGN + fs('123456789'));
  Test('SetDigitGrouping(-123456789, 10)', -123456789, True, MINUS_SIGN + fs('123456789'));
  Test('SetDigitGrouping(-123456789, 11)', -123456789, True, MINUS_SIGN + fs('123456789'));
  Test('SetDigitGrouping(-123456789, 12)', -123456789, True, MINUS_SIGN + fs('123456789'));
  Test('SetDigitGrouping(-123456789, 13)', -123456789, True, MINUS_SIGN + fs('123456789'));
  Test('SetDigitGrouping(-123456789, 14)', -123456789, True, MINUS_SIGN + fs('123456789'));
  Test('SetDigitGrouping(-123456789, 15)', -123456789, True, MINUS_SIGN + fs('123456789'));
  Test('SetDigitGrouping(-123456789, 16)', -123456789, True, MINUS_SIGN + fs('123456789'));

  Test('SetDigitGrouping(0, 0)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 1)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 2)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 3)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 4)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 5)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 6)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 7)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 8)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 9)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 10)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 11)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 12)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 13)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 14)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 15)', 0, True, fs('0'));
  Test('SetDigitGrouping(0, 16)', 0, True, fs('0'));

  Test('SetDigitGrouping(123456789123456789, 0)', 123456789123456789, True, fs('123456789123456789'));
  Test('SetDigitGrouping(123456789123456789, 1)', 123456789123456789, True, fs('1 2 3 4 5 6 7 8 9 1 2 3 4 5 6 7 8 9'));
  Test('SetDigitGrouping(123456789123456789, 2)', 123456789123456789, True, fs('12 34 56 78 91 23 45 67 89'));
  Test('SetDigitGrouping(123456789123456789, 3)', 123456789123456789, True, fs('123 456 789 123 456 789'));
  Test('SetDigitGrouping(123456789123456789, 4)', 123456789123456789, True, fs('12 3456 7891 2345 6789'));
  Test('SetDigitGrouping(123456789123456789, 5)', 123456789123456789, True, fs('123 45678 91234 56789'));
  Test('SetDigitGrouping(123456789123456789, 6)', 123456789123456789, True, fs('123456 789123 456789'));
  Test('SetDigitGrouping(123456789123456789, 7)', 123456789123456789, True, fs('1234 5678912 3456789'));
  Test('SetDigitGrouping(123456789123456789, 8)', 123456789123456789, True, fs('12 34567891 23456789'));
  Test('SetDigitGrouping(123456789123456789, 9)', 123456789123456789, True, fs('123456789 123456789'));
  Test('SetDigitGrouping(123456789123456789, 10)', 123456789123456789, True, fs('12345678 9123456789'));
  Test('SetDigitGrouping(123456789123456789, 12)', 123456789123456789, True, fs('123456 789123456789'));
  Test('SetDigitGrouping(123456789123456789, 16)', 123456789123456789, True, fs('12 3456789123456789'));
  Test('SetDigitGrouping(123456789123456789, 18)', 123456789123456789, True, fs('123456789123456789'));
  Test('SetDigitGrouping(123456789123456789, 19)', 123456789123456789, True, fs('123456789123456789'));
  Test('SetDigitGrouping(123456789123456789, 20)', 123456789123456789, True, fs('123456789123456789'));

  Test('SetDigitGrouping(-123456789123456789, 0)', -123456789123456789, True, MINUS_SIGN + fs('123456789123456789'));
  Test('SetDigitGrouping(-123456789123456789, 1)', -123456789123456789, True, MINUS_SIGN + fs('1 2 3 4 5 6 7 8 9 1 2 3 4 5 6 7 8 9'));
  Test('SetDigitGrouping(-123456789123456789, 2)', -123456789123456789, True, MINUS_SIGN + fs('12 34 56 78 91 23 45 67 89'));
  Test('SetDigitGrouping(-123456789123456789, 3)', -123456789123456789, True, MINUS_SIGN + fs('123 456 789 123 456 789'));
  Test('SetDigitGrouping(-123456789123456789, 4)', -123456789123456789, True, MINUS_SIGN + fs('12 3456 7891 2345 6789'));
  Test('SetDigitGrouping(-123456789123456789, 5)', -123456789123456789, True, MINUS_SIGN + fs('123 45678 91234 56789'));
  Test('SetDigitGrouping(-123456789123456789, 6)', -123456789123456789, True, MINUS_SIGN + fs('123456 789123 456789'));
  Test('SetDigitGrouping(-123456789123456789, 7)', -123456789123456789, True, MINUS_SIGN + fs('1234 5678912 3456789'));
  Test('SetDigitGrouping(-123456789123456789, 8)', -123456789123456789, True, MINUS_SIGN + fs('12 34567891 23456789'));
  Test('SetDigitGrouping(-123456789123456789, 9)', -123456789123456789, True, MINUS_SIGN + fs('123456789 123456789'));
  Test('SetDigitGrouping(-123456789123456789, 10)', -123456789123456789, True, MINUS_SIGN + fs('12345678 9123456789'));
  Test('SetDigitGrouping(-123456789123456789, 12)', -123456789123456789, True, MINUS_SIGN + fs('123456 789123456789'));
  Test('SetDigitGrouping(-123456789123456789, 16)', -123456789123456789, True, MINUS_SIGN + fs('12 3456789123456789'));
  Test('SetDigitGrouping(-123456789123456789, 18)', -123456789123456789, True, MINUS_SIGN + fs('123456789123456789'));
  Test('SetDigitGrouping(-123456789123456789, 19)', -123456789123456789, True, MINUS_SIGN + fs('123456789123456789'));
  Test('SetDigitGrouping(-123456789123456789, 20)', -123456789123456789, True, MINUS_SIGN + fs('123456789123456789'));

  Test('SetDigitGrouping(7, 0)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 1)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 2)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 3)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 4)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 5)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 6)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 7)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 8)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 9)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 10)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 11)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 12)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 13)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 14)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 15)', 7, True, fs('7'));
  Test('SetDigitGrouping(7, 16)', 7, True, fs('7'));

  Test('SetDigitGrouping(-7, 0)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 1)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 2)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 3)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 4)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 5)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 6)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 7)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 8)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 9)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 10)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 11)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 12)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 13)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 14)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 15)', -7, True, MINUS_SIGN + fs('7'));
  Test('SetDigitGrouping(-7, 16)', -7, True, MINUS_SIGN + fs('7'));

  Test('SetDigitGrouping(123456, 127)', 123456, True, fs('123456'));
  Test('SetDigitGrouping(123456, 128)', failure, 'Invalid digit grouping size 128.');
  Test('SetDigitGrouping(123456, -1)', 123456, True, fs('123456'));
  Test('SetDigitGrouping(123456, 2.2)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');
  Test('SetDigitGrouping(123456, "k")', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
  Test('SetDigitGrouping(123456, 2, 2)', failure, 'Too many arguments.');
  Test('SetDigitGrouping(123456)', failure, 'Too few arguments. A required argument of type integer is missing.');

  // Format: minimum length + digit grouping

  Test('SetMinLength(SetDigitGrouping(123, 0), 0)', 123, True, fs('123'));
  Test('SetMinLength(SetDigitGrouping(123, 0), 6)', 123, True, fs('000123'));
  Test('SetMinLength(SetDigitGrouping(123, 1), 6)', 123, True, fs('0 0 0 1 2 3'));
  Test('SetMinLength(SetDigitGrouping(123, 2), 6)', 123, True, fs('00 01 23'));
  Test('SetMinLength(SetDigitGrouping(123, 3), 6)', 123, True, fs('000 123'));
  Test('SetMinLength(SetDigitGrouping(123, 4), 6)', 123, True, fs('00 0123'));
  Test('SetMinLength(SetDigitGrouping(123, 5), 6)', 123, True, fs('0 00123'));
  Test('SetMinLength(SetDigitGrouping(123, 6), 6)', 123, True, fs('000123'));
  Test('SetMinLength(SetDigitGrouping(123, 7), 6)', 123, True, fs('000123'));
  Test('SetMinLength(SetDigitGrouping(123, 4), 16)', 123, True, fs('0000 0000 0000 0123'));

  Test('SetMinLength(SetDigitGrouping(-123, 0), 0)', -123, True, MINUS_SIGN + fs('123'));
  Test('SetMinLength(SetDigitGrouping(-123, 0), 6)', -123, True, MINUS_SIGN + fs('000123'));
  Test('SetMinLength(SetDigitGrouping(-123, 1), 6)', -123, True, MINUS_SIGN + fs('0 0 0 1 2 3'));
  Test('SetMinLength(SetDigitGrouping(-123, 2), 6)', -123, True, MINUS_SIGN + fs('00 01 23'));
  Test('SetMinLength(SetDigitGrouping(-123, 3), 6)', -123, True, MINUS_SIGN + fs('000 123'));
  Test('SetMinLength(SetDigitGrouping(-123, 4), 6)', -123, True, MINUS_SIGN + fs('00 0123'));
  Test('SetMinLength(SetDigitGrouping(-123, 5), 6)', -123, True, MINUS_SIGN + fs('0 00123'));
  Test('SetMinLength(SetDigitGrouping(-123, 6), 6)', -123, True, MINUS_SIGN + fs('000123'));
  Test('SetMinLength(SetDigitGrouping(-123, 7), 6)', -123, True, MINUS_SIGN + fs('000123'));
  Test('SetMinLength(SetDigitGrouping(-123, 4), 16)', -123, True, MINUS_SIGN + fs('0000 0000 0000 0123'));

  Test('SetMinLength(SetDigitGrouping(0, 0), 0)', 0, True, fs('0'));
  Test('SetMinLength(SetDigitGrouping(0, 0), 8)', 0, True, fs('00000000'));
  Test('SetMinLength(SetDigitGrouping(0, 2), 8)', 0, True, fs('00 00 00 00'));
  Test('SetMinLength(SetDigitGrouping(0, 3), 8)', 0, True, fs('00 000 000'));
  Test('SetMinLength(SetDigitGrouping(0, 4), 8)', 0, True, fs('0000 0000'));
  Test('SetMinLength(SetDigitGrouping(0, 6), 8)', 0, True, fs('00 000000'));
  Test('SetMinLength(SetDigitGrouping(0, 7), 8)', 0, True, fs('0 0000000'));
  Test('SetMinLength(SetDigitGrouping(0, 8), 8)', 0, True, fs('00000000'));
  Test('SetMinLength(SetDigitGrouping(0, 9), 8)', 0, True, fs('00000000'));

  // Format: number base + digit grouping

  Test('SetDigitGrouping(SetNumberBase(123456, 10), 0)', 123456, True, fs('123456'));
  Test('SetDigitGrouping(SetNumberBase(123456, 10), 2)', 123456, True, fs('12 34 56'));
  Test('SetDigitGrouping(SetNumberBase(123456, 10), 3)', 123456, True, fs('123 456'));

  Test('SetDigitGrouping(SetNumberBase(-123456, 10), 0)', -123456, True, MINUS_SIGN + fs('123456'));
  Test('SetDigitGrouping(SetNumberBase(-123456, 10), 2)', -123456, True, MINUS_SIGN + fs('12 34 56'));
  Test('SetDigitGrouping(SetNumberBase(-123456, 10), 3)', -123456, True, MINUS_SIGN + fs('123 456'));

  Test('SetDigitGrouping(SetNumberBase(0, 10), 0)', 0, True, fs('0'));
  Test('SetDigitGrouping(SetNumberBase(0, 10), 2)', 0, True, fs('0'));
  Test('SetDigitGrouping(SetNumberBase(0, 10), 3)', 0, True, fs('0'));

  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 0)', 123456789123456789, True, fs('1B69B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 1)', 123456789123456789, True, fs('1 B 6 9 B 4 B A C D 0 5 F 1 5'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 2)', 123456789123456789, True, fs('1 B6 9B 4B AC D0 5F 15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 3)', 123456789123456789, True, fs('1B6 9B4 BAC D05 F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 4)', 123456789123456789, True, fs('1B6 9B4B ACD0 5F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 5)', 123456789123456789, True, fs('1B69B 4BACD 05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 6)', 123456789123456789, True, fs('1B6 9B4BAC D05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 7)', 123456789123456789, True, fs('1 B69B4BA CD05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 8)', 123456789123456789, True, fs('1B69B4B ACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 9)', 123456789123456789, True, fs('1B69B4 BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 10)', 123456789123456789, True, fs('1B69B 4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 12)', 123456789123456789, True, fs('1B6 9B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 14)', 123456789123456789, True, fs('1 B69B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 15)', 123456789123456789, True, fs('1B69B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 16)', 123456789123456789, True, fs('1B69B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 16), 32)', 123456789123456789, True, fs('1B69B4BACD05F15'));

  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 0)', -123456789123456789, True, MINUS_SIGN + fs('1B69B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 1)', -123456789123456789, True, MINUS_SIGN + fs('1 B 6 9 B 4 B A C D 0 5 F 1 5'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 2)', -123456789123456789, True, MINUS_SIGN + fs('1 B6 9B 4B AC D0 5F 15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 3)', -123456789123456789, True, MINUS_SIGN + fs('1B6 9B4 BAC D05 F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 4)', -123456789123456789, True, MINUS_SIGN + fs('1B6 9B4B ACD0 5F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 5)', -123456789123456789, True, MINUS_SIGN + fs('1B69B 4BACD 05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 6)', -123456789123456789, True, MINUS_SIGN + fs('1B6 9B4BAC D05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 7)', -123456789123456789, True, MINUS_SIGN + fs('1 B69B4BA CD05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 8)', -123456789123456789, True, MINUS_SIGN + fs('1B69B4B ACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 9)', -123456789123456789, True, MINUS_SIGN + fs('1B69B4 BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 10)', -123456789123456789, True, MINUS_SIGN + fs('1B69B 4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 12)', -123456789123456789, True, MINUS_SIGN + fs('1B6 9B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 14)', -123456789123456789, True, MINUS_SIGN + fs('1 B69B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 15)', -123456789123456789, True, MINUS_SIGN + fs('1B69B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 16)', -123456789123456789, True, MINUS_SIGN + fs('1B69B4BACD05F15'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 32)', -123456789123456789, True, MINUS_SIGN + fs('1B69B4BACD05F15'));

  Test('SetDigitGrouping(SetNumberBase(0, 16), 0)', 0, True, fs('0'));
  Test('SetDigitGrouping(SetNumberBase(0, 16), 1)', 0, True, fs('0'));
  Test('SetDigitGrouping(SetNumberBase(0, 16), 2)', 0, True, fs('0'));
  Test('SetDigitGrouping(SetNumberBase(0, 16), 3)', 0, True, fs('0'));

  Test('SetDigitGrouping(SetNumberBase(123, 2), 0)', 123, True, fs('1111011'));
  Test('SetDigitGrouping(SetNumberBase(123, 2), 1)', 123, True, fs('1 1 1 1 0 1 1'));
  Test('SetDigitGrouping(SetNumberBase(123, 2), 2)', 123, True, fs('1 11 10 11'));
  Test('SetDigitGrouping(SetNumberBase(123, 2), 3)', 123, True, fs('1 111 011'));
  Test('SetDigitGrouping(SetNumberBase(123, 2), 4)', 123, True, fs('111 1011'));
  Test('SetDigitGrouping(SetNumberBase(123, 2), 5)', 123, True, fs('11 11011'));
  Test('SetDigitGrouping(SetNumberBase(123, 2), 6)', 123, True, fs('1 111011'));
  Test('SetDigitGrouping(SetNumberBase(123, 2), 7)', 123, True, fs('1111011'));
  Test('SetDigitGrouping(SetNumberBase(123, 2), 8)', 123, True, fs('1111011'));
  Test('SetDigitGrouping(SetNumberBase(123, 2), 64)', 123, True, fs('1111011'));

  Test('SetDigitGrouping(SetNumberBase(-123, 2), 0)', -123, True, MINUS_SIGN + fs('1111011'));
  Test('SetDigitGrouping(SetNumberBase(-123, 2), 1)', -123, True, MINUS_SIGN + fs('1 1 1 1 0 1 1'));
  Test('SetDigitGrouping(SetNumberBase(-123, 2), 2)', -123, True, MINUS_SIGN + fs('1 11 10 11'));
  Test('SetDigitGrouping(SetNumberBase(-123, 2), 3)', -123, True, MINUS_SIGN + fs('1 111 011'));
  Test('SetDigitGrouping(SetNumberBase(-123, 2), 4)', -123, True, MINUS_SIGN + fs('111 1011'));
  Test('SetDigitGrouping(SetNumberBase(-123, 2), 5)', -123, True, MINUS_SIGN + fs('11 11011'));
  Test('SetDigitGrouping(SetNumberBase(-123, 2), 6)', -123, True, MINUS_SIGN + fs('1 111011'));
  Test('SetDigitGrouping(SetNumberBase(-123, 2), 7)', -123, True, MINUS_SIGN + fs('1111011'));
  Test('SetDigitGrouping(SetNumberBase(-123, 2), 8)', -123, True, MINUS_SIGN + fs('1111011'));
  Test('SetDigitGrouping(SetNumberBase(-123, 2), 64)', -123, True, MINUS_SIGN + fs('1111011'));

  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 2), 0)', 123456789123456789, True, fs('110110110100110110100101110101100110100000101111100010101'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 2), 1)', 123456789123456789, True, fs('1 1 0 1 1 0 1 1 0 1 0 0 1 1 0 1 1 0 1 0 0 1 0 1 1 1 0 1 0 1 1 0 0 1 1 0 1 0 0 0 0 0 1 0 1 1 1 1 1 0 0 0 1 0 1 0 1'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 2), 2)', 123456789123456789, True, fs('1 10 11 01 10 10 01 10 11 01 00 10 11 10 10 11 00 11 01 00 00 01 01 11 11 00 01 01 01'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 2), 8)', 123456789123456789, True, fs('1 10110110 10011011 01001011 10101100 11010000 01011111 00010101'));
  Test('SetDigitGrouping(SetNumberBase(123456789123456789, 2), 32)', 123456789123456789, True, fs('1101101101001101101001011 10101100110100000101111100010101'));

  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 2), 0)', -123456789123456789, True, MINUS_SIGN + fs('110110110100110110100101110101100110100000101111100010101'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 2), 1)', -123456789123456789, True, MINUS_SIGN + fs('1 1 0 1 1 0 1 1 0 1 0 0 1 1 0 1 1 0 1 0 0 1 0 1 1 1 0 1 0 1 1 0 0 1 1 0 1 0 0 0 0 0 1 0 1 1 1 1 1 0 0 0 1 0 1 0 1'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 2), 2)', -123456789123456789, True, MINUS_SIGN + fs('1 10 11 01 10 10 01 10 11 01 00 10 11 10 10 11 00 11 01 00 00 01 01 11 11 00 01 01 01'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 2), 8)', -123456789123456789, True, MINUS_SIGN + fs('1 10110110 10011011 01001011 10101100 11010000 01011111 00010101'));
  Test('SetDigitGrouping(SetNumberBase(-123456789123456789, 2), 32)', -123456789123456789, True, MINUS_SIGN + fs('1101101101001101101001011 10101100110100000101111100010101'));

  Test('SetDigitGrouping(SetNumberBase(0, 2), 0)', 0, True, fs('0'));
  Test('SetDigitGrouping(SetNumberBase(0, 2), 1)', 0, True, fs('0'));
  Test('SetDigitGrouping(SetNumberBase(0, 2), 2)', 0, True, fs('0'));
  Test('SetDigitGrouping(SetNumberBase(0, 2), 3)', 0, True, fs('0'));

  Test('SetDigitGrouping(SetNumberBase(9223372036854775807, 36), 0)', 9223372036854775807, True, fs('1Y2P0IJ32E8E7'));
  Test('SetDigitGrouping(SetNumberBase(9223372036854775807, 36), 1)', 9223372036854775807, True, fs('1 Y 2 P 0 I J 3 2 E 8 E 7'));
  Test('SetDigitGrouping(SetNumberBase(9223372036854775807, 36), 2)', 9223372036854775807, True, fs('1 Y2 P0 IJ 32 E8 E7'));
  Test('SetDigitGrouping(SetNumberBase(9223372036854775807, 36), 3)', 9223372036854775807, True, fs('1 Y2P 0IJ 32E 8E7'));
  Test('SetDigitGrouping(SetNumberBase(9223372036854775807, 36), 4)', 9223372036854775807, True, fs('1 Y2P0 IJ32 E8E7'));
  Test('SetDigitGrouping(SetNumberBase(9223372036854775807, 36), 8)', 9223372036854775807, True, fs('1Y2P0 IJ32E8E7'));
  Test('SetDigitGrouping(SetNumberBase(9223372036854775807, 36), 16)', 9223372036854775807, True, fs('1Y2P0IJ32E8E7'));

  Test('SetDigitGrouping(SetNumberBase(-9223372036854775807, 36), 0)', -9223372036854775807, True, MINUS_SIGN + fs('1Y2P0IJ32E8E7'));
  Test('SetDigitGrouping(SetNumberBase(-9223372036854775807, 36), 1)', -9223372036854775807, True, MINUS_SIGN + fs('1 Y 2 P 0 I J 3 2 E 8 E 7'));
  Test('SetDigitGrouping(SetNumberBase(-9223372036854775807, 36), 2)', -9223372036854775807, True, MINUS_SIGN + fs('1 Y2 P0 IJ 32 E8 E7'));
  Test('SetDigitGrouping(SetNumberBase(-9223372036854775807, 36), 3)', -9223372036854775807, True, MINUS_SIGN + fs('1 Y2P 0IJ 32E 8E7'));
  Test('SetDigitGrouping(SetNumberBase(-9223372036854775807, 36), 4)', -9223372036854775807, True, MINUS_SIGN + fs('1 Y2P0 IJ32 E8E7'));
  Test('SetDigitGrouping(SetNumberBase(-9223372036854775807, 36), 8)', -9223372036854775807, True, MINUS_SIGN + fs('1Y2P0 IJ32E8E7'));
  Test('SetDigitGrouping(SetNumberBase(-9223372036854775807, 36), 16)', -9223372036854775807, True, MINUS_SIGN + fs('1Y2P0IJ32E8E7'));

  Test('SetDigitGrouping(SetNumberBase(0, 36), 4)', 0, True, fs('0'));

  // Format: number base + digit grouping + minimum length

  Test('SetMinLength(SetDigitGrouping(SetNumberBase(123456, 10), 3), 10)', 123456, True, fs('0 000 123 456'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(123456, 16), 2), 8)', 123456, True, fs('00 01 E2 40'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(123456, 2), 8), 32)', 123456, True, fs('00000000 00000001 11100010 01000000'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(123456, 2), 4), 32)', 123456, True, fs('0000 0000 0000 0001 1110 0010 0100 0000'));

  Test('SetMinLength(SetDigitGrouping(SetNumberBase(123456789123456789, 10), 3), 20)', 123456789123456789, True, fs('00 123 456 789 123 456 789'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(123456789123456789, 16), 2), 16)', 123456789123456789, True, fs('01 B6 9B 4B AC D0 5F 15'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(123456789123456789, 2), 8), 64)', 123456789123456789, True, fs('00000001 10110110 10011011 01001011 10101100 11010000 01011111 00010101'));

  Test('SetMinLength(SetDigitGrouping(SetNumberBase(9223372036854775807, 10), 3), 20)', 9223372036854775807, True, fs('09 223 372 036 854 775 807'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(9223372036854775807, 16), 2), 16)', 9223372036854775807, True, fs('7F FF FF FF FF FF FF FF'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(9223372036854775807, 2), 8), 64)', 9223372036854775807, True, fs('01111111 11111111 11111111 11111111 11111111 11111111 11111111 11111111'));

  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-123456, 10), 3), 10)', -123456, True, MINUS_SIGN + fs('0 000 123 456'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-123456, 16), 2), 8)', -123456, True, MINUS_SIGN + fs('00 01 E2 40'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-123456, 2), 8), 32)', -123456, True, MINUS_SIGN + fs('00000000 00000001 11100010 01000000'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-123456, 2), 4), 32)', -123456, True, MINUS_SIGN + fs('0000 0000 0000 0001 1110 0010 0100 0000'));

  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-123456789123456789, 10), 3), 20)', -123456789123456789, True, MINUS_SIGN + fs('00 123 456 789 123 456 789'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-123456789123456789, 16), 2), 16)', -123456789123456789, True, MINUS_SIGN + fs('01 B6 9B 4B AC D0 5F 15'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-123456789123456789, 2), 8), 64)', -123456789123456789, True, MINUS_SIGN + fs('00000001 10110110 10011011 01001011 10101100 11010000 01011111 00010101'));

  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-9223372036854775807, 10), 3), 20)', -9223372036854775807, True, MINUS_SIGN + fs('09 223 372 036 854 775 807'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-9223372036854775807, 16), 2), 16)', -9223372036854775807, True, MINUS_SIGN + fs('7F FF FF FF FF FF FF FF'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-9223372036854775807, 2), 8), 64)', -9223372036854775807, True, MINUS_SIGN + fs('01111111 11111111 11111111 11111111 11111111 11111111 11111111 11111111'));

  Test('SetMinLength(SetDigitGrouping(SetNumberBase(853624750053, 2), 8), 64)', 853624750053, True, fs('00000000 00000000 00000000 11000110 10111111 11111111 11111111 11100101'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-853624750053, 2), 8), 64)', -853624750053, True, MINUS_SIGN + fs('00000000 00000000 00000000 11000110 10111111 11111111 11111111 11100101'));

  Test('SetMinLength(SetDigitGrouping(SetNumberBase(853624750053, 2), 8), 32)', 853624750053, True, fs('11000110 10111111 11111111 11111111 11100101'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(-853624750053, 2), 8), 32)', -853624750053, True, MINUS_SIGN + fs('11000110 10111111 11111111 11111111 11100101'));

  Test('SetNumberBase(SetMinLength(SetDigitGrouping(853624750053, 8), 64), 2)', 853624750053, True, fs('00000000 00000000 00000000 11000110 10111111 11111111 11111111 11100101'));
  Test('SetNumberBase(SetNumberBase(SetDigitGrouping(SetMinLength(853624750053, 64), 8), 36), 2)', 853624750053, True, fs('00000000 00000000 00000000 11000110 10111111 11111111 11111111 11100101'));
  Test('SetMinLength(SetNumberBase(SetNumberBase(SetDigitGrouping(SetMinLength(853624750053, 255), 8), 36), 2), 64)', 853624750053, True, fs('00000000 00000000 00000000 11000110 10111111 11111111 11111111 11100101'));

  Test('SetMinLength(SetDigitGrouping(SetNumberBase(0, 10), 3), 10)', 0, True, fs('0 000 000 000'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(0, 16), 2), 8)', 0, True, fs('00 00 00 00'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(0, 2), 8), 32)', 0, True, fs('00000000 00000000 00000000 00000000'));
  Test('SetMinLength(SetDigitGrouping(SetNumberBase(0, 2), 4), 32)', 0, True, fs('0000 0000 0000 0000 0000 0000 0000 0000'));


  //
  // Rational numbers (and the division operator on integers, and operators on rationals)
  //

  Test('1/2', Rat(1, 2), True, '1/2');
  Test('1/3', Rat(1, 3), True, '1/3');
  Test('1/1000000', Rat(1, 1000000), True, '1/1000000');
  Test('1237/17', Rat(1237, 17), True, '1237/17');
  Test('1237/1238', Rat(1237, 1238), True, '1237/1238');

  Test('-1/2', Rat(-1, 2), True, MINUS_SIGN + '1/2');
  Test('-1/3', Rat(-1, 3), True, MINUS_SIGN + '1/3');
  Test('-1/1000000', Rat(-1, 1000000), True, MINUS_SIGN + '1/1000000');
  Test('-1237/17', Rat(-1237, 17), True, MINUS_SIGN + '1237/17');
  Test('-1237/1238', Rat(-1237, 1238), True, MINUS_SIGN + '1237/1238');

  Test('1/1', 1, True, '1');
  Test('2/2', 1, True, '1');
  Test('987654321/987654321', 1, True, '1');
  Test('987654320/987654321', Rat(987654320, 987654321), True, '987654320/987654321');

  Test('-1/1', -1, True, MINUS_SIGN + '1');
  Test('-2/2', -1, True, MINUS_SIGN + '1');
  Test('-987654321/987654321', -1, True, MINUS_SIGN + '1');
  Test('-987654320/987654321', Rat(-987654320, 987654321), True, MINUS_SIGN + '987654320/987654321');

  Test('0/2', 0, True, '0');
  Test('0/3', 0, True, '0');
  Test('0/1000000', 0, True, '0');
  Test('0/17', 0, True, '0');
  Test('0/1238', 0, True, '0');

  Test('4/20', Rat(1, 5), True, '1/5');
  Test('40/200', Rat(1, 5), True, '1/5');
  Test('400000000/2000000000', Rat(1, 5), True, '1/5');
  Test('40000000000000/200000000000000', Rat(1, 5), True, '1/5');

  Test('-4/20', Rat(-1, 5), True, MINUS_SIGN + '1/5');
  Test('-40/200', Rat(-1, 5), True, MINUS_SIGN + '1/5');
  Test('-400000000/2000000000', Rat(-1, 5), True, MINUS_SIGN + '1/5');
  Test('-40000000000000/200000000000000', Rat(-1, 5), True, MINUS_SIGN + '1/5');

  Test('83250/47250', Rat(37, 21), True, '37/21');

  Test('1/3 + 1/4', Rat(7, 12), True, '7/12');
  Test('1/3 - 1/4', Rat(1, 12), True, '1/12');
  Test('(1/3) ⋅ (1/4)', Rat(1, 12), True, '1/12');
  Test('(7/3) ⋅ (3/4)', Rat(7, 4), True, '7/4');
  Test('(7/3) ⋅ (5/4)', Rat(35, 12), True, '35/12');
  Test('(7/3) / (3/4)', Rat(28, 9), True, '28/9');
  Test('(7/3) / (5/4)', Rat(28, 15), True, '28/15');

  Test('1/3 ⋅ 1/4', Rat(1, 12), True, '1/12');
  Test('7/3 ⋅ 3/4', Rat(7, 4), True, '7/4');
  Test('7/3 ⋅ 5/4', Rat(35, 12), True, '35/12');

  Test('(7/3) / (7/4)', Rat(4, 3), True, '4/3');
  Test('-(7/3) / (7/4)', Rat(-4, 3), True, MINUS_SIGN + '4/3');
  Test('(-7/3) / (7/4)', Rat(-4, 3), True, MINUS_SIGN + '4/3');
  Test('(7/-3) / (7/4)', Rat(-4, 3), True, MINUS_SIGN + '4/3');
  Test('(7/3) / -(7/4)', Rat(-4, 3), True, MINUS_SIGN + '4/3');
  Test('(7/3) / (-7/4)', Rat(-4, 3), True, MINUS_SIGN + '4/3');
  Test('(7/3) / (7/-4)', Rat(-4, 3), True, MINUS_SIGN + '4/3');
  Test('-(-7/-3) / -(-7/-4)', Rat(4, 3), True, '4/3');
  Test('--(-7/-3) / -(-7/-4)', Rat(-4, 3), True, MINUS_SIGN + '4/3');
  Test('-(-(-7/-3) / -(-7/-4))', Rat(-4, 3), True, MINUS_SIGN + '4/3');

  Test('0/3 + 1/4', Rat(1, 4), True, '1/4');
  Test('0 + 1/4', Rat(1, 4), True, '1/4');
  Test('(0/3) ⋅ (1/4)', Rat(0, 1), True, '0');
  Test('0 ⋅ (1/4)', Rat(0, 1), True, '0');

  Test('1/7 + 5', Rat(36, 7), True, '36/7');
  Test('1/7 - 5', Rat(-34, 7), True, MINUS_SIGN + '34/7');

  Test('5 + 1/7', Rat(36, 7), True, '36/7');
  Test('5 - 1/7', Rat(34, 7), True, '34/7');

  Test('5 ⋅ 1/7', Rat(5, 7), True, '5/7');
  Test('(5 ⋅ 1)/7', Rat(5, 7), True, '5/7');
  Test('5 ⋅ (1/7)', Rat(5, 7), True, '5/7');

  Test('123456789123456789/123456789123456', Rat(41152263041152263, 41152263041152), True, '41152263041152263/41152263041152');
  Test('-123456789123456789/123456789123456', Rat(-41152263041152263, 41152263041152), True, MINUS_SIGN + '41152263041152263/41152263041152');
  Test('123456789123456789/-123456789123456', Rat(-41152263041152263, 41152263041152), True, MINUS_SIGN + '41152263041152263/41152263041152');
  Test('-123456789123456789/--123456789123456', Rat(-41152263041152263, 41152263041152), True, MINUS_SIGN + '41152263041152263/41152263041152');
  Test('-(-(---123456789123456789/--123456789123456))', Rat(-41152263041152263, 41152263041152), True, MINUS_SIGN + '41152263041152263/41152263041152');
  Test('-(-(---123456789123456789/-(--123456789123456)))', Rat(41152263041152263, 41152263041152), True, '41152263041152263/41152263041152');

  Test('9223372036854775807/9223372036854775807', 1, True, '1');
  Test('9223372036854775806/9223372036854775807', Rat(9223372036854775806, 9223372036854775807), True, '9223372036854775806/9223372036854775807');

  Test('-9223372036854775807/9223372036854775807', -1, True, MINUS_SIGN + '1');
  Test('-9223372036854775806/9223372036854775807', Rat(-9223372036854775806, 9223372036854775807), True, MINUS_SIGN + '9223372036854775806/9223372036854775807');

  Test('1/357 + 512/879 - 12/67 + 22/17', Rat(3972039, 2336089), True, '3972039/2336089');
  Test('52/75 + 94/73 - 12/85 + 6523/51 + 152/7', Rat(32892473, 217175), True, '32892473/217175');
  Test('2/84 - 8563/1 + 5892/7 - 2/9853', Rat(-3195259013, 413826), True, MINUS_SIGN + '3195259013/413826');
  Test('532/751 + 9583/9571 - 9856/8522 + 3/7 - 89/2', Rat(-18659885062139, 428782273934), True, MINUS_SIGN + '18659885062139/428782273934');
  Test('15236/8562712 + 9/2', Rat(4818430, 1070339), True, '4818430/1070339');
  Test('21/5784632 - 8/3', Rat(-6610999, 2479128), True, MINUS_SIGN + '6610999/2479128');
  Test('2/7 ⋅ 5/1 ⋅ 5/9 ⋅ 3/7 ⋅ 7/2 ⋅ 2/7 ⋅ 1/3 ⋅ 6/4 ⋅ 7/2 ⋅ 4/6 ⋅ 9/8 ⋅ 2/7 ⋅ 7/1 ⋅ 4/5 ⋅ 6/8', Rat(15, 28), True, '15/28');
  Test('2/7 ⋅ 5/1 ⋅ 5/9 ⋅ 3/7 ⋅ 7/2 ⋅ 2/7 ⋅ 1/3 ⋅ 6/0 ⋅ 7/2 ⋅ 4/6 ⋅ 9/8 ⋅ 2/7 ⋅ 7/1 ⋅ 4/5 ⋅ 6/8', failure, 'division by zero');
  Test('598 ⋅ 752/411856', Rat(28106, 25741), True, '28106/25741');
  Test('752/411856 ⋅ 598', Rat(28106, 25741), True, '28106/25741');
  Test('0 ⋅ 752/411856', 0, True, '0');
  Test('0 ⋅ (752/411856)', Rat(0, 1), True, '0');

//  Test('9223372036854775806/9223372036854775807 + 1/9223372036854775807', Rat(1, 1), True, '1');

  Test('SetNumberBase(SetMinLength(SetDigitGrouping(255/256, 2), 8), 16)', Rat(255, 256), True, fs('00 00 00 FF/00 00 01 00'));


  //
  // Real numbers
  //

  Test('3.1415', 3.1415, True, '3.1415');
  Test('-3.1415', -3.1415, True, MINUS_SIGN + '3.1415');

  Test('3.141592653589793238', 3.141592653589793238, True, '3.14159265359');
  Test('-3.141592653589793238', -3.141592653589793238, True, MINUS_SIGN + '3.14159265359');

  Test('3', 3, True, '3');
  Test('3.', 3.0, True, '3');

  Test('-3', -3, True, MINUS_SIGN + '3');
  Test('-3.', -3.0, True, MINUS_SIGN + '3');

  Test('.1415', 0.1415, True, '0.1415');
  Test('-.1415', -0.1415, True, MINUS_SIGN + '0.1415');

  Test('.', 0.0, True, '0');
  Test('+.', 0.0, True, '0');
  Test('-.', 0.0, True, '0');
  Test('.+.', 0.0, True, '0');

  Test('123456.789', 123456.789, True, '123456.789');
  Test('-123456.789', -123456.789, True, MINUS_SIGN + '123456.789');

  Test('1/2', Rat(1, 2), True, '1/2');
  Test('1./2', 0.5, True, '0.5');
  Test('1/2.', 0.5, True, '0.5');

  Test('1/3', Rat(1, 3), True, '1/3');
  Test('1./3', 1/3, True, '0.333333333333');
  Test('1/3.', 1/3, True, '0.333333333333');

  Test('-1/2', Rat(-1, 2), True, MINUS_SIGN + '1/2');
  Test('-1./2', -0.5, True, MINUS_SIGN + '0.5');
  Test('-1/2.', -0.5, True, MINUS_SIGN + '0.5');

  Test('-1/3', Rat(-1, 3), True, MINUS_SIGN + '1/3');
  Test('-1./3', -1/3, True, MINUS_SIGN + '0.333333333333');
  Test('-1/3.', -1/3, True, MINUS_SIGN + '0.333333333333');

  // Format: default, number of digits

  Test('1./3', 1/3, True, '0.333333333333');
  Test('1./3 \ 2', 1/3, True, '0.33');
  Test('1./3 \ 3', 1/3, True, '0.333');
  Test('1./3 \ 4', 1/3, True, '0.3333');
  Test('1./3 \ 5', 1/3, True, '0.33333');
  Test('1./3 \ 12', 1/3, True, '0.333333333333');
  Test('1./3 \ 16', 1/3, True, '0.3333333333333333');

  Test('-1./3', -1/3, True, MINUS_SIGN + '0.333333333333');
  Test('-1./3 \ 2', -1/3, True, MINUS_SIGN + '0.33');
  Test('-1./3 \ 3', -1/3, True, MINUS_SIGN + '0.333');
  Test('-1./3 \ 4', -1/3, True, MINUS_SIGN + '0.3333');
  Test('-1./3 \ 5', -1/3, True, MINUS_SIGN + '0.33333');
  Test('-1./3 \ 12', -1/3, True, MINUS_SIGN + '0.333333333333');
  Test('-1./3 \ 16', -1/3, True, MINUS_SIGN + '0.3333333333333333');

  Test('1./3', 1/3, True, '0.333333333333'); // operator \ maps to SetMaxLen
  Test('SetMaxLen(1./3, 2)', 1/3, True, '0.33');
  Test('SetMaxLen(1./3, 3)', 1/3, True, '0.333');
  Test('SetMaxLen(1./3, 4)', 1/3, True, '0.3333');
  Test('SetMaxLen(1./3, 5)', 1/3, True, '0.33333');
  Test('SetMaxLen(1./3, 12)', 1/3, True, '0.333333333333');
  Test('SetMaxLen(1./3, 16)', 1/3, True, '0.3333333333333333');

  Test('-1./3', -1/3, True, MINUS_SIGN + '0.333333333333');
  Test('SetMaxLen(-1./3, 2)', -1/3, True, MINUS_SIGN + '0.33');
  Test('SetMaxLen(-1./3, 3)', -1/3, True, MINUS_SIGN + '0.333');
  Test('SetMaxLen(-1./3, 4)', -1/3, True, MINUS_SIGN + '0.3333');
  Test('SetMaxLen(-1./3, 5)', -1/3, True, MINUS_SIGN + '0.33333');
  Test('SetMaxLen(-1./3, 12)', -1/3, True, MINUS_SIGN + '0.333333333333');
  Test('SetMaxLen(-1./3, 16)', -1/3, True, MINUS_SIGN + '0.3333333333333333');

  Test('1./3', 1/3, True, '0.333333333333'); // SetMaxLen works like SetNumDigits when arg 1 is a number and arg 2 is >= 0
  Test('SetNumDigits(1./3, 2)', 1/3, True, '0.33');
  Test('SetNumDigits(1./3, 3)', 1/3, True, '0.333');
  Test('SetNumDigits(1./3, 4)', 1/3, True, '0.3333');
  Test('SetNumDigits(1./3, 5)', 1/3, True, '0.33333');
  Test('SetNumDigits(1./3, 12)', 1/3, True, '0.333333333333');
  Test('SetNumDigits(1./3, 16)', 1/3, True, '0.3333333333333333');

  Test('-1./3', -1/3, True, MINUS_SIGN + '0.333333333333');
  Test('SetNumDigits(-1./3, 2)', -1/3, True, MINUS_SIGN + '0.33');
  Test('SetNumDigits(-1./3, 3)', -1/3, True, MINUS_SIGN + '0.333');
  Test('SetNumDigits(-1./3, 4)', -1/3, True, MINUS_SIGN + '0.3333');
  Test('SetNumDigits(-1./3, 5)', -1/3, True, MINUS_SIGN + '0.33333');
  Test('SetNumDigits(-1./3, 12)', -1/3, True, MINUS_SIGN + '0.333333333333');
  Test('SetNumDigits(-1./3, 16)', -1/3, True, MINUS_SIGN + '0.3333333333333333');

  Test('SetNumDigits(1./3, -1)', 1/3, True, '0.333333333333'); // SetNumDigits: negative count means inherit
  Test('SetMaxLen(1./3, -1)', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.'); // SetMaxLength requires a non-negative second argument
  Test('1./3 \ -1', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.'); // SetMaxLength requires a non-negative second argument

  Test('SetNumDigits(1./3, "a")', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
  Test('SetMaxLen(1./3, "a")', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
  Test('1./3 \ "a"', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');

  Test('SetNumDigits(1./3, 6, 10)', failure, 'Too many arguments.');
  Test('SetNumDigits(1./3)', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('2./3', 2/3, True, '0.666666666667');
  Test('2./3 \ 2', 2/3, True, '0.67');
  Test('2./3 \ 3', 2/3, True, '0.667');
  Test('2./3 \ 4', 2/3, True, '0.6667');
  Test('2./3 \ 5', 2/3, True, '0.66667');
  Test('2./3 \ 12', 2/3, True, '0.666666666667');
  Test('2./3 \ 16', 2/3, True, '0.6666666666666667');

  Test('-2./3', -2/3, True, MINUS_SIGN + '0.666666666667');
  Test('-2./3 \ 2', -2/3, True, MINUS_SIGN + '0.67');
  Test('-2./3 \ 3', -2/3, True, MINUS_SIGN + '0.667');
  Test('-2./3 \ 4', -2/3, True, MINUS_SIGN + '0.6667');
  Test('-2./3 \ 5', -2/3, True, MINUS_SIGN + '0.66667');
  Test('-2./3 \ 12', -2/3, True, MINUS_SIGN + '0.666666666667');
  Test('-2./3 \ 16', -2/3, True, MINUS_SIGN + '0.6666666666666667');

  // Exp literals and default format's exp notation

  Test('7E20', 7E20, True, mm('7*10^20'));
  Test('1E20 + 7E20', 8E20, True, mm('8*10^20'));
  Test('7E-20', 7E-20, True, mm('7*10^-20'));
  Eps; Test('1E-20 + 7E-20', 8E-20, True, mm('8*10^-20'));
  Test('123.45E20', 123.45E20, True, mm('1.2345*10^22'));
  Test('123.45E-20', 123.45E-20, True, mm('1.2345*10^-18'));
  Test('-123.45E20', -123.45E20, True, mm('-1.2345*10^22'));
  Test('-123.45E-20', -123.45E-20, True, mm('-1.2345*10^-18'));
  Test('3E3 + 2E2 + 1E1 + 0E0 + 1E-1 + 2E-2 + 3E-3', 3210.123, True, mm('3210.123'));
  Test('-3E3 - 2E2 - 1E1 - 0E0 - 1E-1 - 2E-2 - 3E-3', -3210.123, True, mm('-3210.123'));
  Test('1E0', 1.0, True, '1');
  Test('7E0', 7.0, True, '7');
  Test('123456.123456E0', 123456.123456, True, '123456.123456');
  Test('-123456.123456E0', -123456.123456, True, MINUS_SIGN + '123456.123456');
  Test('12300000000000000000000000000000000000000000', 1.23E43, True, mm('1.23*10^43'));
  Test('-12300000000000000000000000000000000000000000', -1.23E43, True, mm('-1.23*10^43'));
  Test('0.0000000000000000000123', 1.23E-20, True, mm('1.23*10^-20'));
  Test('-0.0000000000000000000123', -1.23E-20, True, mm('-1.23*10^-20'));
  Test('1E20', 1E20, True, mm('1*10^20')); // a future version might skip the "1*" part
  Test('1E-20', 1E-20, True, mm('1*10^-20')); // -"-

  Test('123456.', 123456.0, True, '123456');
  Test('123456. \ 8', 123456.0, True, '123456');
  Test('123456. \ 6', 123456.0, True, '123456');
  Test('123456. \ 5', 123456.0, True, mm('1.2346*10^5'));
  Test('123456. \ 4', 123456.0, True, mm('1.235*10^5'));
  Test('123456. \ 3', 123456.0, True, mm('1.23*10^5'));
  Test('123456. \ 2', 123456.0, True, mm('1.2*10^5'));

  Test('-123456.', -123456.0, True, MINUS_SIGN + '123456');
  Test('-123456. \ 8', -123456.0, True, MINUS_SIGN + '123456');
  Test('-123456. \ 6', -123456.0, True, MINUS_SIGN + '123456');
  Test('-123456. \ 5', -123456.0, True, mm('-1.2346*10^5'));
  Test('-123456. \ 4', -123456.0, True, mm('-1.235*10^5'));
  Test('-123456. \ 3', -123456.0, True, mm('-1.23*10^5'));
  Test('-123456. \ 2', -123456.0, True, mm('-1.2*10^5'));

  // Format: default, pretty exp notation

  Test('7E20', 7E20, True, mm('7*10^20'));
  Test('SetPrettyExp(7E20, true)', 7E20, True, mm('7*10^20'));
  Test('SetPrettyExp(7E20, false)', 7E20, True, mm('7E20'));

  Test('7E-20', 7E-20, True, mm('7*10^-20'));
  Test('SetPrettyExp(7E-20, true)', 7E-20, True, mm('7*10^-20'));
  Test('SetPrettyExp(7E-20, false)', 7E-20, True, mm('7E-20'));

  Test('-7E20', -7E20, True, mm('-7*10^20'));
  Test('SetPrettyExp(-7E20, true)', -7E20, True, mm('-7*10^20'));
  Test('SetPrettyExp(-7E20, false)', -7E20, True, mm('-7E20'));

  Test('-7E-20', -7E-20, True, mm('-7*10^-20'));
  Test('SetPrettyExp(-7E-20, true)', -7E-20, True, mm('-7*10^-20'));
  Test('SetPrettyExp(-7E-20, false)', -7E-20, True, mm('-7E-20'));

  Test('-123456. \ 2', -123456.0, True, mm('-1.2*10^5'));
  Test('SetPrettyExp(-123456., true) \ 2', -123456.0, True, mm('-1.2*10^5'));
  Test('SetPrettyExp(-123456., false) \ 2', -123456.0, True, mm('-1.2E5'));
  Test('SetPrettyExp(-123456. \ 2, false)', -123456.0, True, mm('-1.2E5'));

  Test('SetPrettyExp(7E20, 7)', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('SetPrettyExp(7E20)', failure, 'Too few arguments. A required argument of type boolean is missing.');
  Test('SetPrettyExp(7E20, true, 2)', failure, 'Too many arguments.');

  // Format: fixed, number of digits

  Test('123.45678', 123.45678, True, '123.45678');
  Test('SetNumberFormat(123.45678, "inherited")', 123.45678, True, '123.45678');
  Test('SetNumberFormat(123.45678, "default")', 123.45678, True, '123.45678');
  Test('SetNumberFormat(123.45678, "fixed")', 123.45678, True, '123.456780000000');
  Test('SetNumberFormat(123.45678, "fixed") \ 10', 123.45678, True, '123.4567800000');
  Test('SetNumberFormat(123.45678, "fixed") \ 8', 123.45678, True, '123.45678000');
  Test('SetNumberFormat(123.45678, "fixed") \ 6', 123.45678, True, '123.456780');
  Test('SetNumberFormat(123.45678, "fixed") \ 3', 123.45678, True, '123.457');
  Test('SetNumberFormat(123.45678, "fixed") \ 2', 123.45678, True, '123.46');
  Test('SetNumberFormat(123.45678, "fixed") \ 1', 123.45678, True, '123.5');
  Test('SetNumberFormat(123.45678, "fixed") \ 0', 123.45678, True, '123');

  Test('SetNumberFormat(123.45678, "default") \ 10', 123.45678, True, '123.45678');
  Test('SetNumberFormat(123.45678, "default") \ 8', 123.45678, True, '123.45678');
  Test('SetNumberFormat(123.45678, "default") \ 6', 123.45678, True, '123.457');
  Test('SetNumberFormat(123.45678, "default") \ 3', 123.45678, True, '123');
  Test('SetNumberFormat(123.45678, "default") \ 2', 123.45678, True, mm('1.2*10^2'));

  Test('-123.45678', -123.45678, True, MINUS_SIGN + '123.45678');
  Test('SetNumberFormat(-123.45678, "inherited")', -123.45678, True, MINUS_SIGN + '123.45678');
  Test('SetNumberFormat(-123.45678, "default")', -123.45678, True, MINUS_SIGN + '123.45678');
  Test('SetNumberFormat(-123.45678, "fixed")', -123.45678, True, MINUS_SIGN + '123.456780000000');
  Test('SetNumberFormat(-123.45678, "fixed") \ 10', -123.45678, True, MINUS_SIGN + '123.4567800000');
  Test('SetNumberFormat(-123.45678, "fixed") \ 8', -123.45678, True, MINUS_SIGN + '123.45678000');
  Test('SetNumberFormat(-123.45678, "fixed") \ 6', -123.45678, True, MINUS_SIGN + '123.456780');
  Test('SetNumberFormat(-123.45678, "fixed") \ 3', -123.45678, True, MINUS_SIGN + '123.457');
  Test('SetNumberFormat(-123.45678, "fixed") \ 2', -123.45678, True, MINUS_SIGN + '123.46');
  Test('SetNumberFormat(-123.45678, "fixed") \ 1', -123.45678, True, MINUS_SIGN + '123.5');
  Test('SetNumberFormat(-123.45678, "fixed") \ 0', -123.45678, True, MINUS_SIGN + '123');

  Test('SetNumberFormat(-123.45678, "default") \ 10', -123.45678, True, MINUS_SIGN + '123.45678');
  Test('SetNumberFormat(-123.45678, "default") \ 8', -123.45678, True, MINUS_SIGN + '123.45678');
  Test('SetNumberFormat(-123.45678, "default") \ 6', -123.45678, True, MINUS_SIGN + '123.457');
  Test('SetNumberFormat(-123.45678, "default") \ 3', -123.45678, True, MINUS_SIGN + '123');
  Test('SetNumberFormat(-123.45678, "default") \ 2', -123.45678, True, mm('-1.2*10^2'));

  // Format: fixed, large and small numbers

  Test('SetNumberFormat(123.456E20, "fixed")', 123.456E20, True, mm('1.23456*10^22')); // reasonable to give up "fixed"
  Test('SetNumberFormat(-123.456E20, "fixed")', -123.456E20, True, mm('-1.23456*10^22'));

  Test('SetNumberFormat(123.456E-20, "fixed") \ 10', 123.456E-20, True, mm('0.0000000000')); // obvious
  Test('SetNumberFormat(-123.456E-20, "fixed") \ 10', -123.456E-20, True, mm('0.0000000000'));

  // Format: scientific

  Test('123.4567', 123.4567, True, '123.4567');
  Test('SetNumberFormat(123.4567, "default") \ 10', 123.4567, True, mm('123.4567'));
  Test('SetNumberFormat(123.4567, "fixed") \ 10', 123.4567, True, mm('123.4567000000'));
  Test('SetNumberFormat(123.4567, "scientific") \ 10', 123.4567, True, mm('1.234567000*10^2'));

  Test('SetNumberFormat(123.4567, "default") \ 8', 123.4567, True, mm('123.4567'));
  Test('SetNumberFormat(123.4567, "fixed") \ 8', 123.4567, True, mm('123.45670000'));
  Test('SetNumberFormat(123.4567, "scientific") \ 8', 123.4567, True, mm('1.2345670*10^2'));

  Test('SetNumberFormat(123.4567, "default") \ 6', 123.4567, True, mm('123.457'));
  Test('SetNumberFormat(123.4567, "fixed") \ 6', 123.4567, True, mm('123.456700'));
  Test('SetNumberFormat(123.4567, "scientific") \ 6', 123.4567, True, mm('1.23457*10^2'));

  Test('SetNumberFormat(123.4567, "default") \ 4', 123.4567, True, mm('123.5'));
  Test('SetNumberFormat(123.4567, "fixed") \ 4', 123.4567, True, mm('123.4567'));
  Test('SetNumberFormat(123.4567, "scientific") \ 4', 123.4567, True, mm('1.235*10^2'));

  Test('SetNumberFormat(123.4567, "default") \ 2', 123.4567, True, mm('1.2*10^2'));
  Test('SetNumberFormat(123.4567, "fixed") \ 2', 123.4567, True, mm('123.46'));
  Test('SetNumberFormat(123.4567, "scientific") \ 2', 123.4567, True, mm('1.2*10^2'));

  Test('-123.4567', -123.4567, True, MINUS_SIGN + '123.4567');
  Test('SetNumberFormat(-123.4567, "default") \ 10', -123.4567, True, mm('-123.4567'));
  Test('SetNumberFormat(-123.4567, "fixed") \ 10', -123.4567, True, mm('-123.4567000000'));
  Test('SetNumberFormat(-123.4567, "scientific") \ 10', -123.4567, True, mm('-1.234567000*10^2'));

  Test('SetNumberFormat(-123.4567, "default") \ 8', -123.4567, True, mm('-123.4567'));
  Test('SetNumberFormat(-123.4567, "fixed") \ 8', -123.4567, True, mm('-123.45670000'));
  Test('SetNumberFormat(-123.4567, "scientific") \ 8', -123.4567, True, mm('-1.2345670*10^2'));

  Test('SetNumberFormat(-123.4567, "default") \ 6', -123.4567, True, mm('-123.457'));
  Test('SetNumberFormat(-123.4567, "fixed") \ 6', -123.4567, True, mm('-123.456700'));
  Test('SetNumberFormat(-123.4567, "scientific") \ 6', -123.4567, True, mm('-1.23457*10^2'));

  Test('SetNumberFormat(-123.4567, "default") \ 4', -123.4567, True, mm('-123.5'));
  Test('SetNumberFormat(-123.4567, "fixed") \ 4', -123.4567, True, mm('-123.4567'));
  Test('SetNumberFormat(-123.4567, "scientific") \ 4', -123.4567, True, mm('-1.235*10^2'));

  Test('SetNumberFormat(-123.4567, "default") \ 2', -123.4567, True, mm('-1.2*10^2'));
  Test('SetNumberFormat(-123.4567, "fixed") \ 2', -123.4567, True, mm('-123.46'));
  Test('SetNumberFormat(-123.4567, "scientific") \ 2', -123.4567, True, mm('-1.2*10^2'));

  // Format: scientific, pretty exp off

  Test('SetPrettyExp(SetNumberFormat(-123.4567, "default"), false) \ 2', -123.4567, True, mm('-1.2E2'));
  Test('SetPrettyExp(SetNumberFormat(-123.4567, "fixed"), false) \ 2', -123.4567, True, mm('-123.46'));
  Test('SetPrettyExp(SetNumberFormat(-123.4567, "scientific"), false) \ 2', -123.4567, True, mm('-1.2E+2'));

  Test('SetPrettyExp(1E20, false)', 1E20, True, '1E20');
  Test('SetPrettyExp(SetNumberFormat(1E20, "default") \ 4, false)', 1E20, True, '1E20');
  Test('SetPrettyExp(SetNumberFormat(1E20, "scientific") \ 4, false)', 1E20, True, '1.000E+20');

  Test('SetPrettyExp(-1E20, false)', -1E20, True, mm('-1E20'));
  Test('SetPrettyExp(SetNumberFormat(-1E20, "default") \ 4, false)', -1E20, True, mm('-1E20'));
  Test('SetPrettyExp(SetNumberFormat(-1E20, "scientific") \ 4, false)', -1E20, True, mm('-1.000E+20'));

  Test('SetPrettyExp(1E-20, false)', 1E-20, True, mm('1E-20'));
  Test('SetPrettyExp(SetNumberFormat(1E-20, "default") \ 4, false)', 1E-20, True, mm('1E-20'));
  Test('SetPrettyExp(SetNumberFormat(1E-20, "scientific") \ 4, false)', 1E-20, True, mm('1.000E-20'));

  // Format: digit grouping (default style, no exps)

  Test('SetDigitGrouping(123456.123456, -1)', 123456.123456, True, '123456.123456');
  Test('SetDigitGrouping(123456.123456, 0)', 123456.123456, True, '123456.123456');
  Test('SetDigitGrouping(123456.123456, 1)', 123456.123456, True, fs('1 2 3 4 5 6.1 2 3 4 5 6'));
  Test('SetDigitGrouping(123456.123456, 2)', 123456.123456, True, fs('12 34 56.12 34 56'));
  Test('SetDigitGrouping(123456.123456, 3)', 123456.123456, True, fs('123 456.123 456'));
  Test('SetDigitGrouping(123456.123456, 4)', 123456.123456, True, fs('12 3456.1234 56'));
  Test('SetDigitGrouping(123456.123456, 5)', 123456.123456, True, fs('1 23456.12345 6'));
  Test('SetDigitGrouping(123456.123456, 6)', 123456.123456, True, fs('123456.123456'));
  Test('SetDigitGrouping(123456.123456, 7)', 123456.123456, True, fs('123456.123456'));
  Test('SetDigitGrouping(123456.123456, 8)', 123456.123456, True, fs('123456.123456'));

  Test('SetDigitGrouping(-123456.123456, -1)', -123456.123456, True, MINUS_SIGN + '123456.123456');
  Test('SetDigitGrouping(-123456.123456, 0)', -123456.123456, True, MINUS_SIGN + '123456.123456');
  Test('SetDigitGrouping(-123456.123456, 1)', -123456.123456, True, MINUS_SIGN + fs('1 2 3 4 5 6.1 2 3 4 5 6'));
  Test('SetDigitGrouping(-123456.123456, 2)', -123456.123456, True, MINUS_SIGN + fs('12 34 56.12 34 56'));
  Test('SetDigitGrouping(-123456.123456, 3)', -123456.123456, True, MINUS_SIGN + fs('123 456.123 456'));
  Test('SetDigitGrouping(-123456.123456, 4)', -123456.123456, True, MINUS_SIGN + fs('12 3456.1234 56'));
  Test('SetDigitGrouping(-123456.123456, 5)', -123456.123456, True, MINUS_SIGN + fs('1 23456.12345 6'));
  Test('SetDigitGrouping(-123456.123456, 6)', -123456.123456, True, MINUS_SIGN + fs('123456.123456'));
  Test('SetDigitGrouping(-123456.123456, 7)', -123456.123456, True, MINUS_SIGN + fs('123456.123456'));
  Test('SetDigitGrouping(-123456.123456, 8)', -123456.123456, True, MINUS_SIGN + fs('123456.123456'));

  // Format: digit grouping (default style, with exps)

  Test('SetDigitGrouping(1.23456123456E20, -1)', 1.23456123456E20, True, mm(fs('1.23456123456*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 0)', 1.23456123456E20, True, mm(fs('1.23456123456*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 1)', 1.23456123456E20, True, mm(fs('1.2 3 4 5 6 1 2 3 4 5 6*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 2)', 1.23456123456E20, True, mm(fs('1.23 45 61 23 45 6*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 3)', 1.23456123456E20, True, mm(fs('1.234 561 234 56*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 4)', 1.23456123456E20, True, mm(fs('1.2345 6123 456*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 5)', 1.23456123456E20, True, mm(fs('1.23456 12345 6*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 6)', 1.23456123456E20, True, mm(fs('1.234561 23456*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 7)', 1.23456123456E20, True, mm(fs('1.2345612 3456*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 8)', 1.23456123456E20, True, mm(fs('1.23456123 456*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 9)', 1.23456123456E20, True, mm(fs('1.234561234 56*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 10)', 1.23456123456E20, True, mm(fs('1.2345612345 6*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 11)', 1.23456123456E20, True, mm(fs('1.23456123456*10^20')));
  Test('SetDigitGrouping(1.23456123456E20, 12)', 1.23456123456E20, True, mm(fs('1.23456123456*10^20')));

  Test('SetDigitGrouping(1.23456123456E-20, -1)', 1.23456123456E-20, True, mm(fs('1.23456123456*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 0)', 1.23456123456E-20, True, mm(fs('1.23456123456*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 1)', 1.23456123456E-20, True, mm(fs('1.2 3 4 5 6 1 2 3 4 5 6*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 2)', 1.23456123456E-20, True, mm(fs('1.23 45 61 23 45 6*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 3)', 1.23456123456E-20, True, mm(fs('1.234 561 234 56*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 4)', 1.23456123456E-20, True, mm(fs('1.2345 6123 456*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 5)', 1.23456123456E-20, True, mm(fs('1.23456 12345 6*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 6)', 1.23456123456E-20, True, mm(fs('1.234561 23456*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 7)', 1.23456123456E-20, True, mm(fs('1.2345612 3456*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 8)', 1.23456123456E-20, True, mm(fs('1.23456123 456*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 9)', 1.23456123456E-20, True, mm(fs('1.234561234 56*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 10)', 1.23456123456E-20, True, mm(fs('1.2345612345 6*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 11)', 1.23456123456E-20, True, mm(fs('1.23456123456*10^-20')));
  Test('SetDigitGrouping(1.23456123456E-20, 12)', 1.23456123456E-20, True, mm(fs('1.23456123456*10^-20')));

  Test('SetDigitGrouping(-1.23456123456E20, -1)', -1.23456123456E20, True, mm(fs('-1.23456123456*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 0)', -1.23456123456E20, True, mm(fs('-1.23456123456*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 1)', -1.23456123456E20, True, mm(fs('-1.2 3 4 5 6 1 2 3 4 5 6*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 2)', -1.23456123456E20, True, mm(fs('-1.23 45 61 23 45 6*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 3)', -1.23456123456E20, True, mm(fs('-1.234 561 234 56*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 4)', -1.23456123456E20, True, mm(fs('-1.2345 6123 456*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 5)', -1.23456123456E20, True, mm(fs('-1.23456 12345 6*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 6)', -1.23456123456E20, True, mm(fs('-1.234561 23456*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 7)', -1.23456123456E20, True, mm(fs('-1.2345612 3456*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 8)', -1.23456123456E20, True, mm(fs('-1.23456123 456*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 9)', -1.23456123456E20, True, mm(fs('-1.234561234 56*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 10)', -1.23456123456E20, True, mm(fs('-1.2345612345 6*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 11)', -1.23456123456E20, True, mm(fs('-1.23456123456*10^20')));
  Test('SetDigitGrouping(-1.23456123456E20, 12)', -1.23456123456E20, True, mm(fs('-1.23456123456*10^20')));

  Test('SetDigitGrouping(-1.23456123456E-20, -1)', -1.23456123456E-20, True, mm(fs('-1.23456123456*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 0)', -1.23456123456E-20, True, mm(fs('-1.23456123456*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 1)', -1.23456123456E-20, True, mm(fs('-1.2 3 4 5 6 1 2 3 4 5 6*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 2)', -1.23456123456E-20, True, mm(fs('-1.23 45 61 23 45 6*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 3)', -1.23456123456E-20, True, mm(fs('-1.234 561 234 56*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 4)', -1.23456123456E-20, True, mm(fs('-1.2345 6123 456*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 5)', -1.23456123456E-20, True, mm(fs('-1.23456 12345 6*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 6)', -1.23456123456E-20, True, mm(fs('-1.234561 23456*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 7)', -1.23456123456E-20, True, mm(fs('-1.2345612 3456*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 8)', -1.23456123456E-20, True, mm(fs('-1.23456123 456*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 9)', -1.23456123456E-20, True, mm(fs('-1.234561234 56*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 10)', -1.23456123456E-20, True, mm(fs('-1.2345612345 6*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 11)', -1.23456123456E-20, True, mm(fs('-1.23456123456*10^-20')));
  Test('SetDigitGrouping(-1.23456123456E-20, 12)', -1.23456123456E-20, True, mm(fs('-1.23456123456*10^-20')));

  // Format: digit grouping (fixed)

  Test('SetNumberFormat(1234.56789, "fixed") \ 6', 1234.56789, True, '1234.567890');
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, -1)', 1234.56789, True, fs('1234.567890'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, 0)', 1234.56789, True, fs('1234.567890'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, 1)', 1234.56789, True, fs('1 2 3 4.5 6 7 8 9 0'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, 2)', 1234.56789, True, fs('12 34.56 78 90'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, 3)', 1234.56789, True, fs('1 234.567 890'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, 4)', 1234.56789, True, fs('1234.5678 90'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, 5)', 1234.56789, True, fs('1234.56789 0'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, 6)', 1234.56789, True, fs('1234.567890'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, 7)', 1234.56789, True, fs('1234.567890'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 6, 8)', 1234.56789, True, fs('1234.567890'));

  Test('SetNumberFormat(1234.56789, "fixed") \ 10', 1234.56789, True, '1234.5678900000');
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, -1)', 1234.56789, True, fs('1234.5678900000'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 0)', 1234.56789, True, fs('1234.5678900000'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 1)', 1234.56789, True, fs('1 2 3 4.5 6 7 8 9 0 0 0 0 0'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 2)', 1234.56789, True, fs('12 34.56 78 90 00 00'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 3)', 1234.56789, True, fs('1 234.567 890 000 0'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 4)', 1234.56789, True, fs('1234.5678 9000 00'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 5)', 1234.56789, True, fs('1234.56789 00000'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 6)', 1234.56789, True, fs('1234.567890 0000'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 7)', 1234.56789, True, fs('1234.5678900 000'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 8)', 1234.56789, True, fs('1234.56789000 00'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 9)', 1234.56789, True, fs('1234.567890000 0'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 10)', 1234.56789, True, fs('1234.5678900000'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 11)', 1234.56789, True, fs('1234.5678900000'));
  Test('SetDigitGrouping(SetNumberFormat(1234.56789, "fixed") \ 10, 12)', 1234.56789, True, fs('1234.5678900000'));

  Test('SetNumberFormat(-1234.56789, "fixed") \ 6', -1234.56789, True, mm('-1234.567890'));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, -1)', -1234.56789, True, fs(mm('-1234.567890')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, 0)', -1234.56789, True, fs(mm('-1234.567890')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, 1)', -1234.56789, True, fs(mm('-1 2 3 4.5 6 7 8 9 0')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, 2)', -1234.56789, True, fs(mm('-12 34.56 78 90')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, 3)', -1234.56789, True, fs(mm('-1 234.567 890')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, 4)', -1234.56789, True, fs(mm('-1234.5678 90')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, 5)', -1234.56789, True, fs(mm('-1234.56789 0')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, 6)', -1234.56789, True, fs(mm('-1234.567890')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, 7)', -1234.56789, True, fs(mm('-1234.567890')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 6, 8)', -1234.56789, True, fs(mm('-1234.567890')));

  Test('SetNumberFormat(-1234.56789, "fixed") \ 10', -1234.56789, True, mm('-1234.5678900000'));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, -1)', -1234.56789, True, fs(mm('-1234.5678900000')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 0)', -1234.56789, True, fs(mm('-1234.5678900000')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 1)', -1234.56789, True, fs(mm('-1 2 3 4.5 6 7 8 9 0 0 0 0 0')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 2)', -1234.56789, True, fs(mm('-12 34.56 78 90 00 00')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 3)', -1234.56789, True, fs(mm('-1 234.567 890 000 0')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 4)', -1234.56789, True, fs(mm('-1234.5678 9000 00')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 5)', -1234.56789, True, fs(mm('-1234.56789 00000')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 6)', -1234.56789, True, fs(mm('-1234.567890 0000')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 7)', -1234.56789, True, fs(mm('-1234.5678900 000')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 8)', -1234.56789, True, fs(mm('-1234.56789000 00')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 9)', -1234.56789, True, fs(mm('-1234.567890000 0')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 10)', -1234.56789, True, fs(mm('-1234.5678900000')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 11)', -1234.56789, True, fs(mm('-1234.5678900000')));
  Test('SetDigitGrouping(SetNumberFormat(-1234.56789, "fixed") \ 10, 12)', -1234.56789, True, fs(mm('-1234.5678900000')));


  //
  // Complex numbers
  //

  Test('i', ImaginaryUnit, True, 'i');
  Test('−i', -ImaginaryUnit, True, '−i');
  Test('77⋅i', 77*ImaginaryUnit, True, '77⋅i');
  Test('−77⋅i', -77*ImaginaryUnit, True, '−77⋅i');
  Test('−1⋅i', -ImaginaryUnit, True, '−i');
  Test('i^2', TASC(-1), True, mm('-1'));
  Test('1+i', 1 + ImaginaryUnit, True, mm('1 + i'));
  Test('7.5+i', 7.5 + ImaginaryUnit, True, mm('7.5 + i'));
  Test('7.5-i', 7.5 - ImaginaryUnit, True, mm('7.5 - i'));
  Test('2+3⋅i', 2 + 3*ImaginaryUnit, True, mm('2 + 3*i'));
  Test('-1+i', -1 + ImaginaryUnit, True, mm('-1 + i'));
  Test('-2+3⋅i', -2 + 3*ImaginaryUnit, True, mm('-2 + 3*i'));
  Test('1-i', 1 - ImaginaryUnit, True, mm('1 - i'));
  Test('2-3⋅i', 2 - 3*ImaginaryUnit, True, mm('2 - 3*i'));
  Test('-1-i', -1 - ImaginaryUnit, True, mm('-1 - i'));
  Test('-2-3⋅i', -2 - 3*ImaginaryUnit, True, mm('-2 - 3*i'));
  Test('sqrt(-1)', ImaginaryUnit, True, 'i');
  Eps; Test('i^i', TASC(exp(-pi/2)));
  Eps; Test('arcsin(-2)', -1.57079632679489662 + 1.31695789692481671*ImaginaryUnit);
  Test('1/3 + i/3', 1/3 + ImaginaryUnit/3, True, '0.333333333333 + 0.333333333333⋅i');
  Test('(1/3 + i/3) \ 5', 1/3 + ImaginaryUnit/3, True, '0.33333 + 0.33333⋅i');
  Test('(1/3 + i/3) \ 2', 1/3 + ImaginaryUnit/3, True, '0.33 + 0.33⋅i');


  //
  // Booleans
  //

  Test('true', true);
  Test('false', false);
  Test('true ∧ true ∧ true', true);
  Test('false ∧ true ∧ true', false);
  Test('true ∧ false ∧ true', false);
  Test('true ∧ true ∧ false', false);
  Test('false ∨ false ∨ false ∨ false', false);
  Test('false ∨ false ∨ true ∨ false', true);
  Test('¬false', true);
  Test('¬true', false);
  Test('¬¬¬¬¬¬¬¬¬¬true', true);
  Test('¬¬¬¬¬¬¬¬¬¬¬true', false);


  //
  // Strings
  //

  Test('"Lorem ipsum"', 'Lorem ipsum', True, 'Lorem ipsum');
  Test('"cat"', 'cat', True, 'cat');
  Test('""', '', True, '');
  Test('"Lorem" + " " + "ipsum"', 'Lorem ipsum', True, 'Lorem ipsum');
  Test('"Lorem" ~ " " ~ "ipsum"', 'Lorem ipsum', True, 'Lorem ipsum');
  Test('"cat" + "" + ""', 'cat', True, 'cat');
  Test('"cat" ~ "" ~ ""', 'cat', True, 'cat');
  Test('"" + "cat"', 'cat', True, 'cat');
  Test('"" ~ "cat"', 'cat', True, 'cat');
  Test('"" + ""', '', True, '');
  Test('"" ~ ""', '', True, '');

  _s2 := DupeString('cat ', 250) + '…';           Assert(_s2.Length - 1 = 1000);

  _s := 'cat' + DupeString(' cat', 100);
  Test('"' + _s + '"', _s, True, _s);

  _s := 'cat' + DupeString(' cat', 1000);
  Test('"' + _s + '"', _s, True, _s2);

  _s := 'cat' + DupeString(' cat', 10000);
  Test('"' + _s + '"', _s, True, _s2);

  _s := 'cat' + DupeString(' cat', 100000);
  Test('"' + _s + '"', _s, True, _s2);

  _s := 'cat' + DupeString(' cat', 1000000);
  Test('"' + _s + '"', _s, True, _s2);

  _s := 'cat' + DupeString(' cat', 10000000);
  Test('"' + _s + '"', _s, True, _s2);

  _s := 'Lorem ipsum dolor sit amet, consectetur adipiscing elit.';
  Test('"' + _s + '"', _s, True, _s);
  Test('"' + _s + '" \11', _s, True, 'Lorem ipsum…');
  Test('"' + _s + '" \5', _s, True, 'Lorem…');
  Test('"' + _s + '" \2', _s, True, 'Lo…');
  Test('"' + _s + '" \1', _s, True, 'L…');

  Test('SetMaxLen("' + _s + '", 11)', _s, True, 'Lorem ipsum…');

  Test('"' + _s + '" \0', _s, True, _s);
  Test('SetMaxLen("' + _s + '", 0)', _s, True, _s);

  Test('"' + _s + '" \ -1', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.');
  Test('SetMaxLen("' + _s + '", -1)', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.');

  Test('"alfa" + ¶ + "beta"', 'alfa'#13#10'beta', True, 'alfa↲beta');
  TestSL('"alfa" + ¶ + "beta"', 'alfa↲beta');
  TestML('"alfa" + ¶ + "beta"', 'alfa'#13#10'beta');

  Test('"A new book named ""Cats""."', 'A new book named "Cats".', True, 'A new book named "Cats".');
  Test('"The empty string is """"."', 'The empty string is "".', True, 'The empty string is "".');
  Test('""""""', '""', True, '""');

  Test('"alfa beta', ESyntaxException, 'Unterminated string literal "alfa beta" starting at column 1.');
  Test('"alfa" + "beta', ESyntaxException, 'Unterminated string literal "beta" starting at column 10.');
  Test('"', ESyntaxException, 'Unterminated string literal "" starting at column 1.');
  Test('"""""', ESyntaxException, 'Unterminated string literal """""" starting at column 1.');

  Test('"Dogs are cute."[1]', 'D');
  Test('"Dogs are cute."[2]', 'o');
  Test('"Dogs are cute."[3]', 'g');
  Test('"Dogs are cute."[4]', 's');
  Test('"Dogs are cute."[5]', ' ');
  Test('"Dogs are cute."[6]', 'a');
  Test('"Dogs are cute."[7]', 'r');
  Test('"Dogs are cute."[8]', 'e');
  Test('"Dogs are cute."[9]', ' ');
  Test('"Dogs are cute."[10]', 'c');
  Test('"Dogs are cute."[11]', 'u');
  Test('"Dogs are cute."[12]', 't');
  Test('"Dogs are cute."[13]', 'e');
  Test('"Dogs are cute."[14]', '.');

  Test('"Dogs are cute."[15]', failure, 'Index 15 out of bounds.');

  Test('"Dogs are cute."[-1]', '.');
  Test('"Dogs are cute."[-2]', 'e');
  Test('"Dogs are cute."[-3]', 't');
  Test('"Dogs are cute."[-4]', 'u');
  Test('"Dogs are cute."[-5]', 'c');
  Test('"Dogs are cute."[-6]', ' ');
  Test('"Dogs are cute."[-7]', 'e');
  Test('"Dogs are cute."[-8]', 'r');
  Test('"Dogs are cute."[-9]', 'a');
  Test('"Dogs are cute."[-10]', ' ');
  Test('"Dogs are cute."[-11]', 's');
  Test('"Dogs are cute."[-12]', 'g');
  Test('"Dogs are cute."[-13]', 'o');
  Test('"Dogs are cute."[-14]', 'D');

  Test('"Dogs are cute."[-15]', failure, 'Index -15 out of bounds.');

  Test('"Dogs are cute."[0]', failure, 'Index 0 out of bounds.');

  Test('"Dogs are cute."[5] ≔ "G"', failure, 'Left side cannot be assigned to.');
  Test('"Dogs are cute."[-2] ≔ "G"', failure, 'Left side cannot be assigned to.');
  Test('"Dogs are cute."[25] ≔ "G"', failure, 'Left side cannot be assigned to.');
  Test('"Dogs are cute."[-25] ≔ "G"', failure, 'Left side cannot be assigned to.');
  Test('"Dogs are cute."[0] ≔ "G"', failure, 'Left side cannot be assigned to.');

  Test('("A" + "B")[1]', 'A');
  Test('("A" + "B")[2]', 'B');
  Test('("A" + "B")[3]', failure, 'Index 3 out of bounds.');

  Test('("A" + "B")[-1]', 'B');
  Test('("A" + "B")[-2]', 'A');
  Test('("A" + "B")[-3]', failure, 'Index -3 out of bounds.');

  Test('("A" + "B")[0]', failure, 'Index 0 out of bounds.');

  Test('""[-3]', failure, 'Index -3 out of bounds.');
  Test('""[-2]', failure, 'Index -2 out of bounds.');
  Test('""[-1]', failure, 'Index -1 out of bounds.');
  Test('""[0]', failure, 'Index 0 out of bounds.');
  Test('""[1]', failure, 'Index 1 out of bounds.');
  Test('""[2]', failure, 'Index 2 out of bounds.');
  Test('""[3]', failure, 'Index 3 out of bounds.');

  Test('struct("a": 42, "b": "dog").b[1]', 'd');
  Test('struct("a": 42, "b": "dog").b[2]', 'o');
  Test('struct("a": 42, "b": "dog").b[3]', 'g');
  Test('struct("a": 42, "b": "dog").b[4]', failure, 'Index 4 out of bounds.');

  Test('struct("a": 42, "b": "dog").b[-1]', 'g');
  Test('struct("a": 42, "b": "dog").b[-2]', 'o');
  Test('struct("a": 42, "b": "dog").b[-3]', 'd');
  Test('struct("a": 42, "b": "dog").b[-4]', failure, 'Index -4 out of bounds.');

  Test('struct("a": 42, "b": "dog").b[0]', failure, 'Index 0 out of bounds.');

  Test('struct("a": 42, "b": "dog").b[2] ≔ "a"', failure, 'Left side cannot be assigned to.');


  //
  // Real vectors
  //

  Test('❨1, 0, 0❩', [1, 0, 0]);                                                  // ❨❩
  Test('❨1, 2, 3❩', [1, 2, 3]);
  Test('❨1, -1❩', [1, -1]);
  Test('❨1❩', [1]);
  Test('❨0❩', [0]);
  Test('❨-1❩', [-1]);

  Test('❨2, −3, 1❩[1]', 2.0);
  Test('❨2, −3, 1❩[2]', -3.0);
  Test('❨2, −3, 1❩[3]', 1.0);
  Test('❨2, −3, 1❩[4]', failure, 'Index 4 out of bounds.');

  Test('❨2, −3, 1❩[-1]', 1.0);
  Test('❨2, −3, 1❩[-2]', -3.0);
  Test('❨2, −3, 1❩[-3]', 2.0);
  Test('❨2, −3, 1❩[-4]', failure, 'Index -4 out of bounds.');

  Test('❨2, −3, 1❩[0]', failure, 'Index 0 out of bounds.');

  Test('❨2, −3, 1❩[2] ≔ 3', failure, 'Left side cannot be assigned to.');
  Test('❨2, −3, 1❩[20] ≔ 3', failure, 'Left side cannot be assigned to.');

  Test('vector(1, 2, 3)', [1, 2, 3]);

  Test('❨1, 2, 3❩ + ❨10, 100, 1000❩', [11, 102, 1003]);

  TestSL('❨1, 2, 3❩', '(1, 2, 3)');
  TestML('❨1, 2, 3❩', ' ⎛1⎞'#13#10'e⎜2⎟'#13#10' ⎝3⎠');

  TestML('❨1❩', '(1)');
  TestML('❨1, 2❩', ' ⎛1⎞'#13#10'e⎝2⎠');
  TestML('❨1, 2, 3❩', ' ⎛1⎞'#13#10'e⎜2⎟'#13#10' ⎝3⎠');

  Test('SequenceVector(100)', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100]);
  TestSL('SequenceVector(100)', '(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32,'+' 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70,'+' 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100)');
  TestML('SequenceVector(100)', '(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32,'+' 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70,'+' 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100)'); // (way) beyond default vertical until

  Test('SequenceVector(200)', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200]);
  TestSL('SequenceVector(200)', '(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, '+'33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, '+'67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, …)');
  TestML('SequenceVector(200)', '(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, '+'33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, '+'67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, …)'); // (way) beyond default vertical until

  Test('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩', '(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)');

  Test('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 10', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 10', '(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)');

  Test('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 9', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 9', '(1, 2, 3, 4, 5, 6, 7, 8, 9, …)');

  Test('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 5', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 5', '(1, 2, 3, 4, 5, …)');

  Test('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 2', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 2', '(1, 2, …)');

  Test('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 1', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 1', '(1, …)');

  Test('❨1, 2, 3, 4, 5, 6❩', [1, 2, 3, 4, 5, 6]);
  TestSL('❨1, 2, 3, 4, 5, 6❩', '(1, 2, 3, 4, 5, 6)');
  TestML('❨1, 2, 3, 4, 5, 6❩', ' ⎛1⎞'#13#10' ⎜2⎟'#13#10' ⎜3⎟'#13#10'e⎜4⎟'#13#10' ⎜5⎟'#13#10' ⎝6⎠');

  Test('❨1, 2, 3, 4, 5, 6❩ \ 3', [1, 2, 3, 4, 5, 6]);
  TestSL('❨1, 2, 3, 4, 5, 6❩ \ 3', '(1, 2, 3, …)');
  TestML('❨1, 2, 3, 4, 5, 6❩ \ 3', ' ⎛1⎞'#13#10' ⎜2⎟'#13#10'e⎜3⎟'#13#10' ⎝⋮⎠');

  Test('❨❩', ESyntaxException, 'Empty bracket at column 1.'); // would make sense to allow syntactically, to get vector's error message instead
  Test('vector()', failure, 'A vector must have dimension at least one.');

  Test('❨1, 1/2, 1/3❩', [1, 1/2, 1/3]);
  TestML('❨1, 1/2, 1/3❩', ' ⎛      1       ⎞'#13#10'e⎜     0.5      ⎟'#13#10' ⎝0.333333333333⎠');
  TestML('SetNumDigits(❨1, 1/2, 1/3❩, 4)', ' ⎛  1   ⎞'#13#10'e⎜ 0.5  ⎟'#13#10' ⎝0.3333⎠');
  TestML('SetNumDigits(❨1/2, 1/3, 1/4❩, 4) \ 2', ' ⎛ 0.5  ⎞'#13#10'e⎜0.3333⎟'#13#10' ⎝  ⋮   ⎠');

  TestSL('❨1, 1/2, 1/3❩', '(1, 0.5, 0.333333333333)');
  TestSL('SetNumDigits(❨1, 1/2, 1/3❩, 4)', '(1, 0.5, 0.3333)');
  TestSL('SetNumDigits(❨1/2, 1/3, 1/4❩, 4) \ 2', '(0.5, 0.3333, …)');

  TestML('SetDigitGrouping(❨12432435, 65426277, 763763588, 62565245246, 415❩, 3)', ' ⎛  12 432 435  ⎞'#13#10' ⎜  65 426 277  ⎟'#13#10'e⎜ 763 763 588  ⎟'#13#10' ⎜62 565 245 246⎟'#13#10' ⎝     415      ⎠');
  TestSL('SetDigitGrouping(❨12432435, 65426277, 763763588, 62565245246, 415❩, 3)', '(12 432 435, 65 426 277, 763 763 588, 62 565 245 246, 415)');


  //
  // Complex vectors
  //

  Test('❨i, 0, 0❩', [ImaginaryUnit, 0, 0]);                                      // ❨❩
  Test('❨i, 2, 3❩', [ImaginaryUnit, 2, 3]);
  Test('❨i, -1❩', [ImaginaryUnit, -1]);
  Test('❨i❩', [ImaginaryUnit]);
  Test('❨-i❩', [-ImaginaryUnit]);

  TestML('❨i❩', '(i)');
  TestML('❨i, 2❩', ' ⎛i⎞'#13#10'e⎝2⎠');
  TestML('❨i, 2, 3❩', ' ⎛i⎞'#13#10'e⎜2⎟'#13#10' ⎝3⎠');

  Test('❨2, −3, i❩[1]', TASC(2.0));
  Test('❨2, −3, i❩[2]', TASC(-3.0));
  Test('❨2, −3, i❩[3]', ImaginaryUnit);
  Test('❨2, −3, i❩[4]', failure, 'Index 4 out of bounds.');

  Test('❨2, −3, i❩[-1]', ImaginaryUnit);
  Test('❨2, −3, i❩[-2]', TASC(-3.0));
  Test('❨2, −3, i❩[-3]', TASC(2.0));
  Test('❨2, −3, i❩[-4]', failure, 'Index -4 out of bounds.');

  Test('❨2, −3, i❩[0]', failure, 'Index 0 out of bounds.');

  Test('❨2, −3, i❩[2] ≔ 3', failure, 'Left side cannot be assigned to.');
  Test('❨2, −3, i❩[20] ≔ 3', failure, 'Left side cannot be assigned to.');

  Test('vector(i, 2, 3)', [ImaginaryUnit, 2, 3]);

  Test('❨i, 2, 3❩ + ❨10, 100, 1000❩', [ImaginaryUnit + 10, 102, 1003]);

  TestSL('❨i, 2, 3❩', '(i, 2, 3)');
  TestML('❨i, 2, 3❩', ' ⎛i⎞'#13#10'e⎜2⎟'#13#10' ⎝3⎠');

  Test('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩', [ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩', '(i, 2, 3, 4, 5, 6, 7, 8, 9, 10)');

  Test('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 10', [ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 10', '(i, 2, 3, 4, 5, 6, 7, 8, 9, 10)');

  Test('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 9', [ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 9', '(i, 2, 3, 4, 5, 6, 7, 8, 9, …)');

  Test('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 5', [ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 5', '(i, 2, 3, 4, 5, …)');

  Test('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 2', [ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 2', '(i, 2, …)');

  Test('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 1', [ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  TestSL('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 1', '(i, …)');

  Test('❨i, 2, 3, 4, 5, 6❩', [ImaginaryUnit, 2, 3, 4, 5, 6]);
  TestSL('❨i, 2, 3, 4, 5, 6❩', '(i, 2, 3, 4, 5, 6)');
  TestML('❨i, 2, 3, 4, 5, 6❩', ' ⎛i⎞'#13#10' ⎜2⎟'#13#10' ⎜3⎟'#13#10'e⎜4⎟'#13#10' ⎜5⎟'#13#10' ⎝6⎠');

  Test('❨i, 2, 3, 4, 5, 6❩ \ 3', [ImaginaryUnit, 2, 3, 4, 5, 6]);
  TestSL('❨i, 2, 3, 4, 5, 6❩ \ 3', '(i, 2, 3, …)');
  TestML('❨i, 2, 3, 4, 5, 6❩ \ 3', ' ⎛i⎞'#13#10' ⎜2⎟'#13#10'e⎜3⎟'#13#10' ⎝⋮⎠');

  Test('❨i, 1/2, 1/3❩', [ImaginaryUnit, 1/2, 1/3]);
  TestML('❨i, 1/2, 1/3❩', ' ⎛      i       ⎞'#13#10'e⎜     0.5      ⎟'#13#10' ⎝0.333333333333⎠');
  TestML('SetNumDigits(❨i, 1/2, 1/3❩, 4)', ' ⎛  i   ⎞'#13#10'e⎜ 0.5  ⎟'#13#10' ⎝0.3333⎠');
  TestML('SetNumDigits(❨i/2, 1/3, 1/4❩, 4) \ 2', ' ⎛0.5⋅i ⎞'#13#10'e⎜0.3333⎟'#13#10' ⎝  ⋮   ⎠');

  TestSL('❨i, 1/2, 1/3❩', '(i, 0.5, 0.333333333333)');
  TestSL('SetNumDigits(❨i, 1/2, 1/3❩, 4)', '(i, 0.5, 0.3333)');
  TestSL('SetNumDigits(❨i/2, 1/3, 1/4❩, 4) \ 2', '(0.5⋅i, 0.3333, …)');

  TestML('SetDigitGrouping(❨12432435, 65426277, 763763588, 62565245246⋅i, 415❩, 3)', ' ⎛   12 432 435   ⎞'#13#10' ⎜   65 426 277   ⎟'#13#10'e⎜  763 763 588   ⎟'#13#10' ⎜62 565 245 246⋅i⎟'#13#10' ⎝      415       ⎠');
  TestSL('SetDigitGrouping(❨12432435, 65426277, 763763588, 62565245246⋅i, 415❩, 3)', '(12 432 435, 65 426 277, 763 763 588, 62 565 245 246⋅i, 415)');


  //
  // Real matrices
  //

  Test('❨❨1, 2❩, ❨3, 4❩❩', 2, [1, 2, 3, 4]);                                     // ❨❩
  Test('matrix(2, 1, 2, 3, 4)', 2, [1, 2, 3, 4]);
  Test('matrix(1, 1, 2, 3, 4)', 1, [1, 2, 3, 4]);
  Test('matrix(4, 1, 2, 3, 4)', 4, [1, 2, 3, 4]);
  Test('matrix(3, 1, 2, 3, 4)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('matrix(0, 1, 2, 3, 4)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('❨❨1, 2❩, ❨3, 4❩❩ + ❨❨1, 10❩, ❨100, 1000❩❩', 2, [2, 12, 103, 1004]);
  Test('❨❨1❩, ❨2❩, ❨3❩❩', 1, [1, 2, 3]);
  Test('❨❨1, 2, 3❩❩', 3, [1, 2, 3]);
  Test('❨❨1, 2❩, ❨3❩❩', failure, 'Attempt to create a non-rectangular matrix.');
  Test('ZeroMatrix(3)', 3, [0, 0, 0, 0, 0, 0, 0, 0, 0]);
  Test('ZeroMatrix(2, 3)', 3, [0, 0, 0, 0, 0, 0]);
  Test('IdentityMatrix(3)', 3, [1, 0, 0, 0, 1, 0, 0, 0, 1]);
  TestML('IdentityMatrix(5)', '⎛1  0  0  0  0⎞'#13#10'⎜0  1  0  0  0⎟'#13#10'⎜0  0  1  0  0⎟'#13#10'⎜0  0  0  1  0⎟'#13#10'⎝0  0  0  0  1⎠');
  TestML('IdentityMatrix(5) \ 5', '⎛1  0  0  0  0⎞'#13#10'⎜0  1  0  0  0⎟'#13#10'⎜0  0  1  0  0⎟'#13#10'⎜0  0  0  1  0⎟'#13#10'⎝0  0  0  0  1⎠');
  TestML('IdentityMatrix(5) \ 4', '⎛1  0  0  0  ⋯⎞'#13#10'⎜0  1  0  0  ⋯⎟'#13#10'⎜0  0  1  0  ⋯⎟'#13#10'⎜0  0  0  1  ⋯⎟'#13#10'⎝⋮  ⋮  ⋮  ⋮  ⋱⎠');
  TestML('IdentityMatrix(5) \ 3', '⎛1  0  0  ⋯⎞'#13#10'⎜0  1  0  ⋯⎟'#13#10'⎜0  0  1  ⋯⎟'#13#10'⎝⋮  ⋮  ⋮  ⋱⎠');
  TestML('IdentityMatrix(5) \ 2', '⎛1  0  ⋯⎞'#13#10'⎜0  1  ⋯⎟'#13#10'⎝⋮  ⋮  ⋱⎠');
  TestML('IdentityMatrix(5) \ 1', '⎛1  ⋯⎞'#13#10'⎝⋮  ⋱⎠');
  TestML('IdentityMatrix(5) \ 0', '⎛1  0  0  0  0⎞'#13#10'⎜0  1  0  0  0⎟'#13#10'⎜0  0  1  0  0⎟'#13#10'⎜0  0  0  1  0⎟'#13#10'⎝0  0  0  0  1⎠');
  Test('IdentityMatrix(5) \ -1', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.');
  TestSL('IdentityMatrix(5)', '((1, 0, 0, 0, 0), (0, 1, 0, 0, 0), (0, 0, 1, 0, 0), (0, 0, 0, 1, 0), (0, 0, 0, 0, 1))');
  TestSL('IdentityMatrix(5) \ 5', '((1, 0, 0, 0, 0), (0, 1, 0, 0, 0), (0, 0, 1, 0, 0), (0, 0, 0, 1, 0), (0, 0, 0, 0, 1))');
  TestSL('IdentityMatrix(5) \ 4', '((1, 0, 0, 0, …), (0, 1, 0, 0, …), (0, 0, 1, 0, …), (0, 0, 0, 1, …), …)');
  TestSL('IdentityMatrix(5) \ 3', '((1, 0, 0, …), (0, 1, 0, …), (0, 0, 1, …), …)');
  TestSL('IdentityMatrix(5) \ 2', '((1, 0, …), (0, 1, …), …)');
  TestSL('IdentityMatrix(5) \ 1', '((1, …), …)');
  TestSL('IdentityMatrix(5) \ 0', '((1, 0, 0, 0, 0), (0, 1, 0, 0, 0), (0, 0, 1, 0, 0), (0, 0, 0, 1, 0), (0, 0, 0, 0, 1))');
  TestML('ZeroMatrix(5, 8)', '⎛0  0  0  0  0  0  0  0⎞'#13#10'⎜0  0  0  0  0  0  0  0⎟'#13#10'⎜0  0  0  0  0  0  0  0⎟'#13#10'⎜0  0  0  0  0  0  0  0⎟'#13#10'⎝0  0  0  0  0  0  0  0⎠');
  TestML('ZeroMatrix(5, 8) \ 12', '⎛0  0  0  0  0  0  0  0⎞'#13#10'⎜0  0  0  0  0  0  0  0⎟'#13#10'⎜0  0  0  0  0  0  0  0⎟'#13#10'⎜0  0  0  0  0  0  0  0⎟'#13#10'⎝0  0  0  0  0  0  0  0⎠');
  TestML('ZeroMatrix(5, 8) \ 8', '⎛0  0  0  0  0  0  0  0⎞'#13#10'⎜0  0  0  0  0  0  0  0⎟'#13#10'⎜0  0  0  0  0  0  0  0⎟'#13#10'⎜0  0  0  0  0  0  0  0⎟'#13#10'⎝0  0  0  0  0  0  0  0⎠');
  TestML('ZeroMatrix(5, 8) \ 7', '⎛0  0  0  0  0  0  0  ⋯⎞'#13#10'⎜0  0  0  0  0  0  0  ⋯⎟'#13#10'⎜0  0  0  0  0  0  0  ⋯⎟'#13#10'⎜0  0  0  0  0  0  0  ⋯⎟'#13#10'⎝0  0  0  0  0  0  0  ⋯⎠');
  TestML('ZeroMatrix(5, 8) \ 4', '⎛0  0  0  0  ⋯⎞'#13#10'⎜0  0  0  0  ⋯⎟'#13#10'⎜0  0  0  0  ⋯⎟'#13#10'⎜0  0  0  0  ⋯⎟'#13#10'⎝⋮  ⋮  ⋮  ⋮  ⋱⎠');

  TestML('IdentityMatrix(1)', '(1)');
  TestML('IdentityMatrix(2)', '⎛1  0⎞'#13#10'⎝0  1⎠');
  TestML('IdentityMatrix(3)', '⎛1  0  0⎞'#13#10'⎜0  1  0⎟'#13#10'⎝0  0  1⎠');
  TestML('IdentityMatrix(4)', '⎛1  0  0  0⎞'#13#10'⎜0  1  0  0⎟'#13#10'⎜0  0  1  0⎟'#13#10'⎝0  0  0  1⎠');

  Test('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2)', 5, [1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, -2, 4, 5, 8, -9, 2]);
  Test('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 10', 5, [1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, -2, 4, 5, 8, -9, 2]);
  Test('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 8', 5, [1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, -2, 4, 5, 8, -9, 2]);
  Test('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 5', 5, [1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, -2, 4, 5, 8, -9, 2]);
  Test('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 2', 5, [1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, -2, 4, 5, 8, -9, 2]);
  Test('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 1', 5, [1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, -2, 4, 5, 8, -9, 2]);
  Test('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 0', 5, [1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, -2, 4, 5, 8, -9, 2]);

  TestML('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2)', '⎛ 1   5   3   4   7⎞'#13#10'⎜ 6   5   4   3   8⎟'#13#10'⎜ 7   6   5   3  −2⎟'#13#10'⎝ 4   5   8  −9   2⎠');
  TestML('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 10', '⎛ 1   5   3   4   7⎞'#13#10'⎜ 6   5   4   3   8⎟'#13#10'⎜ 7   6   5   3  −2⎟'#13#10'⎝ 4   5   8  −9   2⎠');
  TestML('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 6', '⎛ 1   5   3   4   7⎞'#13#10'⎜ 6   5   4   3   8⎟'#13#10'⎜ 7   6   5   3  −2⎟'#13#10'⎝ 4   5   8  −9   2⎠');
  TestML('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 5', '⎛ 1   5   3   4   7⎞'#13#10'⎜ 6   5   4   3   8⎟'#13#10'⎜ 7   6   5   3  −2⎟'#13#10'⎝ 4   5   8  −9   2⎠');
  TestML('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 0', '⎛ 1   5   3   4   7⎞'#13#10'⎜ 6   5   4   3   8⎟'#13#10'⎜ 7   6   5   3  −2⎟'#13#10'⎝ 4   5   8  −9   2⎠');
  TestML('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 4', '⎛ 1   5   3   4  ⋯ ⎞'#13#10'⎜ 6   5   4   3  ⋯ ⎟'#13#10'⎜ 7   6   5   3  ⋯ ⎟'#13#10'⎝ 4   5   8  −9  ⋯ ⎠');
  TestML('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 3', '⎛1  5  3  ⋯⎞'#13#10'⎜6  5  4  ⋯⎟'#13#10'⎜7  6  5  ⋯⎟'#13#10'⎝⋮  ⋮  ⋮  ⋱⎠');
  TestML('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 2', '⎛1  5  ⋯⎞'#13#10'⎜6  5  ⋯⎟'#13#10'⎝⋮  ⋮  ⋱⎠');
  TestML('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 1', '⎛1  ⋯⎞'#13#10'⎝⋮  ⋱⎠');

  TestSL('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2)', '((1, 5, 3, 4, 7), (6, 5, 4, 3, 8), (7, 6, 5, 3, −2), (4, 5, 8, −9, 2))');
  TestSL('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 10', '((1, 5, 3, 4, 7), (6, 5, 4, 3, 8), (7, 6, 5, 3, −2), (4, 5, 8, −9, 2))');
  TestSL('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 6', '((1, 5, 3, 4, 7), (6, 5, 4, 3, 8), (7, 6, 5, 3, −2), (4, 5, 8, −9, 2))');
  TestSL('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 5', '((1, 5, 3, 4, 7), (6, 5, 4, 3, 8), (7, 6, 5, 3, −2), (4, 5, 8, −9, 2))');
  TestSL('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 0', '((1, 5, 3, 4, 7), (6, 5, 4, 3, 8), (7, 6, 5, 3, −2), (4, 5, 8, −9, 2))');
  TestSL('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 4', '((1, 5, 3, 4, …), (6, 5, 4, 3, …), (7, 6, 5, 3, …), (4, 5, 8, −9, …))');
  TestSL('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 3', '((1, 5, 3, …), (6, 5, 4, …), (7, 6, 5, …), …)');
  TestSL('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 2', '((1, 5, …), (6, 5, …), …)');
  TestSL('matrix(5, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 1', '((1, …), …)');

  Test('matrix(2, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2)', 2, [1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, -2, 4, 5, 8, -9, 2]);
  TestML('matrix(2, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 10', '⎛ 1   5⎞'#13#10'⎜ 3   4⎟'#13#10'⎜ 7   6⎟'#13#10'⎜ 5   4⎟'#13#10'⎜ 3   8⎟'#13#10'⎜ 7   6⎟'#13#10'⎜ 5   3⎟'#13#10'⎜−2   4⎟'#13#10'⎜ 5   8⎟'#13#10'⎝−9   2⎠');
  TestSL('matrix(2, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 10', '((1, 5), (3, 4), (7, 6), (5, 4), (3, 8), (7, 6), (5, 3), (−2, 4), (5, 8), (−9, 2))');
  TestML('matrix(2, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 5', '⎛1  5⎞'#13#10'⎜3  4⎟'#13#10'⎜7  6⎟'#13#10'⎜5  4⎟'#13#10'⎜3  8⎟'#13#10'⎝⋮  ⋮⎠');
  TestSL('matrix(2, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 5', '((1, 5), (3, 4), (7, 6), (5, 4), (3, 8), …)');

  Test('matrix(4, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2)', 4, [1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, -2, 4, 5, 8, -9, 2]);
  TestML('matrix(4, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2)', '⎛ 1   5   3   4⎞'#13#10'⎜ 7   6   5   4⎟'#13#10'⎜ 3   8   7   6⎟'#13#10'⎜ 5   3  −2   4⎟'#13#10'⎝ 5   8  −9   2⎠');
  TestML('matrix(4, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 3', '⎛1  5  3  ⋯⎞'#13#10'⎜7  6  5  ⋯⎟'#13#10'⎜3  8  7  ⋯⎟'#13#10'⎝⋮  ⋮  ⋮  ⋱⎠');
  TestSL('matrix(4, 1, 5, 3, 4, 7, 6, 5, 4, 3, 8, 7, 6, 5, 3, −2, 4, 5, 8, −9, 2) \ 3', '((1, 5, 3, …), (7, 6, 5, …), (3, 8, 7, …), …)');

  TestML('IdentityMatrix(3)/3', '⎛0.333333333333               0               0⎞'#13#10'⎜             0  0.333333333333               0⎟'#13#10'⎝             0               0  0.333333333333⎠');
  TestML('SetNumDigits(IdentityMatrix(3)/3, 10)', '⎛0.3333333333             0             0⎞'#13#10'⎜           0  0.3333333333             0⎟'#13#10'⎝           0             0  0.3333333333⎠');
  TestML('SetNumDigits(IdentityMatrix(3)/3, 4)', '⎛0.3333       0       0⎞'#13#10'⎜     0  0.3333       0⎟'#13#10'⎝     0       0  0.3333⎠');
  TestML('SetNumDigits(IdentityMatrix(3)/3, 4) \ 2', '⎛0.3333       0    ⋯   ⎞'#13#10'⎜     0  0.3333    ⋯   ⎟'#13#10'⎝  ⋮       ⋮       ⋱   ⎠');

  TestSL('IdentityMatrix(3)/3', '((0.333333333333, 0, 0), (0, 0.333333333333, 0), (0, 0, 0.333333333333))');
  TestSL('SetNumDigits(IdentityMatrix(3)/3, 10)', '((0.3333333333, 0, 0), (0, 0.3333333333, 0), (0, 0, 0.3333333333))');
  TestSL('SetNumDigits(IdentityMatrix(3)/3, 4)', '((0.3333, 0, 0), (0, 0.3333, 0), (0, 0, 0.3333))');
  TestSL('SetNumDigits(IdentityMatrix(3)/3, 4) \ 2', '((0.3333, 0, …), (0, 0.3333, …), …)');

  TestML('SetDigitGrouping(SetNumDigits(IdentityMatrix(3)/3, 6), 3)', '⎛0.333 333          0          0⎞'#13#10'⎜        0  0.333 333          0⎟'#13#10'⎝        0          0  0.333 333⎠');
  TestSL('SetDigitGrouping(SetNumDigits(IdentityMatrix(3)/3, 6), 3)', '((0.333 333, 0, 0), (0, 0.333 333, 0), (0, 0, 0.333 333))');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1]', 10.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2]', 20.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[3]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[4]', 40.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[5]', 50.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[6]', 60.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[7]', failure, 'Index 7 out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1]', 60.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-2]', 50.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-3]', 40.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-4]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-5]', 20.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-6]', 10.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-7]', failure, 'Index -7 out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[0]', failure, 'Index 0 out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, 0]', failure, 'Index (1, 0) out of bounds.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, 1]', 10.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, 2]', 20.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, 3]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, 4]', failure, 'Index (1, 4) out of bounds.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2, 1]', 40.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2, 2]', 50.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2, 3]', 60.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2, 4]', failure, 'Index (2, 4) out of bounds.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[3, 1]', failure, 'Index (3, 1) out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, -1]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, -2]', 20.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, -3]', 10.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, -4]', failure, 'Index (1, -4) out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1, 1]', 40.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1, 2]', 50.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1, 3]', 60.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1, 4]', failure, 'Index (-1, 4) out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-2, 1]', 10.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-2, 2]', 20.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-2, 3]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-2, 4]', failure, 'Index (-2, 4) out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1, 0]', failure, 'Index (-1, 0) out of bounds.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1, -1]', 60.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1, -2]', 50.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-2, -1]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-2, -2]', 20.0);

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, 0)]', failure, 'Index (1, 0) out of bounds.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, 1)]', 10.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, 2)]', 20.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, 3)]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, 4)]', failure, 'Index (1, 4) out of bounds.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(2, 1)]', 40.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(2, 2)]', 50.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(2, 3)]', 60.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(2, 4)]', failure, 'Index (2, 4) out of bounds.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(3, 1)]', failure, 'Index (3, 1) out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, -1)]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, -2)]', 20.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, -3)]', 10.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, -4)]', failure, 'Index (1, -4) out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 1)]', 40.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 2)]', 50.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 3)]', 60.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 4)]', failure, 'Index (-1, 4) out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-2, 1)]', 10.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-2, 2)]', 20.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-2, 3)]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-2, 4)]', failure, 'Index (-2, 4) out of bounds.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 0)]', failure, 'Index (-1, 0) out of bounds.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-1, -1)]', 60.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-1, -2)]', 50.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-2, -1)]', 30.0);
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-2, -2)]', 20.0);

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[6] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[7] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[0] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-20] ≔ 100', failure, 'Left side cannot be assigned to.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, 1] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, 2] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, 5] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[1, -1] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2, 2] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2, 100] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2, -4] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[2, 0] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[-1, 1] ≔ 100', failure, 'Left side cannot be assigned to.');

  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, 1)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, 2)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, 5)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(1, -1)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(2, 2)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(2, 100)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(2, -4)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(2, 0)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨10, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 1)] ≔ 100', failure, 'Left side cannot be assigned to.');


  //
  // Complex matrices
  //

  Test('❨❨i, 2❩, ❨3, 4❩❩', 2, [ImaginaryUnit, 2, 3, 4]);
  Test('matrix(2, i, 2, 3, 4)', 2, [ImaginaryUnit, 2, 3, 4]);
  Test('❨❨i❩, ❨2❩, ❨3❩, ❨4❩❩', 1, [ImaginaryUnit, 2, 3, 4]);
  Test('❨❨i, 2, 3, 4❩❩', 4, [ImaginaryUnit, 2, 3, 4]);
  Test('❨❨i, 2❩, ❨3, 4, 5❩❩', failure, 'Attempt to create a non-rectangular matrix.');
  Test('❨❨i, 2❩, ❨3, 4❩❩ + ❨❨1, 0❩, ❨0, i❩❩', 2, [1+ImaginaryUnit, 2, 3, 4+ImaginaryUnit]);
  TestML('❨❨i, 2❩, ❨3, 4❩❩ + ❨❨1, 0❩, ❨0, i❩❩', '⎛1 + i      2⎞'#13#10'⎝    3  4 + i⎠');
  TestSL('❨❨i, 2❩, ❨3, 4❩❩ + ❨❨1, 0❩, ❨0, i❩❩', '((1 + i, 2), (3, 4 + i))');

  TestML('i⋅IdentityMatrix(1)', '(i)');
  TestML('i⋅IdentityMatrix(2)', '⎛i  0⎞'#13#10'⎝0  i⎠');
  TestML('i⋅IdentityMatrix(3)', '⎛i  0  0⎞'#13#10'⎜0  i  0⎟'#13#10'⎝0  0  i⎠');
  TestML('i⋅IdentityMatrix(4)', '⎛i  0  0  0⎞'#13#10'⎜0  i  0  0⎟'#13#10'⎜0  0  i  0⎟'#13#10'⎝0  0  0  i⎠');

  Test('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3)', 5, [3, ImaginaryUnit, 5, 7, ImaginaryUnit, 2, ImaginaryUnit, 3, 4, 5, 1, ImaginaryUnit, ImaginaryUnit, 2, 3]);
  TestML('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3)', '⎛3  i  5  7  i⎞'#13#10'⎜2  i  3  4  5⎟'#13#10'⎝1  i  i  2  3⎠');
  TestML('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 0', '⎛3  i  5  7  i⎞'#13#10'⎜2  i  3  4  5⎟'#13#10'⎝1  i  i  2  3⎠');
  TestML('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 10', '⎛3  i  5  7  i⎞'#13#10'⎜2  i  3  4  5⎟'#13#10'⎝1  i  i  2  3⎠');
  TestML('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 5', '⎛3  i  5  7  i⎞'#13#10'⎜2  i  3  4  5⎟'#13#10'⎝1  i  i  2  3⎠');
  TestML('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 4', '⎛3  i  5  7  ⋯⎞'#13#10'⎜2  i  3  4  ⋯⎟'#13#10'⎝1  i  i  2  ⋯⎠');
  TestML('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 3', '⎛3  i  5  ⋯⎞'#13#10'⎜2  i  3  ⋯⎟'#13#10'⎝1  i  i  ⋯⎠');
  TestML('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 2', '⎛3  i  ⋯⎞'#13#10'⎜2  i  ⋯⎟'#13#10'⎝⋮  ⋮  ⋱⎠');
  TestML('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 1', '⎛3  ⋯⎞'#13#10'⎝⋮  ⋱⎠');
  TestSL('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3)', '((3, i, 5, 7, i), (2, i, 3, 4, 5), (1, i, i, 2, 3))');
  TestSL('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 0', '((3, i, 5, 7, i), (2, i, 3, 4, 5), (1, i, i, 2, 3))');
  TestSL('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 10', '((3, i, 5, 7, i), (2, i, 3, 4, 5), (1, i, i, 2, 3))');
  TestSL('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 5', '((3, i, 5, 7, i), (2, i, 3, 4, 5), (1, i, i, 2, 3))');
  TestSL('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 4', '((3, i, 5, 7, …), (2, i, 3, 4, …), (1, i, i, 2, …))');
  TestSL('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 3', '((3, i, 5, …), (2, i, 3, …), (1, i, i, …))');
  TestSL('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 2', '((3, i, …), (2, i, …), …)');
  TestSL('matrix(5, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \ 1', '((3, …), …)');

  Test('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3)', 3, [3, ImaginaryUnit, 5, 7, ImaginaryUnit, 2, ImaginaryUnit, 3, 4, 5, 1, ImaginaryUnit, ImaginaryUnit, 2, 3]);
  TestML('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3)', '⎛3  i  5⎞'#13#10'⎜7  i  2⎟'#13#10'⎜i  3  4⎟'#13#10'⎜5  1  i⎟'#13#10'⎝i  2  3⎠');
  TestML('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \0', '⎛3  i  5⎞'#13#10'⎜7  i  2⎟'#13#10'⎜i  3  4⎟'#13#10'⎜5  1  i⎟'#13#10'⎝i  2  3⎠');
  TestML('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \10', '⎛3  i  5⎞'#13#10'⎜7  i  2⎟'#13#10'⎜i  3  4⎟'#13#10'⎜5  1  i⎟'#13#10'⎝i  2  3⎠');
  TestML('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \5', '⎛3  i  5⎞'#13#10'⎜7  i  2⎟'#13#10'⎜i  3  4⎟'#13#10'⎜5  1  i⎟'#13#10'⎝i  2  3⎠');
  TestML('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \4', '⎛3  i  5⎞'#13#10'⎜7  i  2⎟'#13#10'⎜i  3  4⎟'#13#10'⎜5  1  i⎟'#13#10'⎝⋮  ⋮  ⋮⎠');
  TestML('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \3', '⎛3  i  5⎞'#13#10'⎜7  i  2⎟'#13#10'⎜i  3  4⎟'#13#10'⎝⋮  ⋮  ⋮⎠');
  TestML('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \2', '⎛3  i  ⋯⎞'#13#10'⎜7  i  ⋯⎟'#13#10'⎝⋮  ⋮  ⋱⎠');
  TestML('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \1', '⎛3  ⋯⎞'#13#10'⎝⋮  ⋱⎠');
  TestSL('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3)', '((3, i, 5), (7, i, 2), (i, 3, 4), (5, 1, i), (i, 2, 3))');
  TestSL('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \0', '((3, i, 5), (7, i, 2), (i, 3, 4), (5, 1, i), (i, 2, 3))');
  TestSL('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \10', '((3, i, 5), (7, i, 2), (i, 3, 4), (5, 1, i), (i, 2, 3))');
  TestSL('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \5', '((3, i, 5), (7, i, 2), (i, 3, 4), (5, 1, i), (i, 2, 3))');
  TestSL('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \4', '((3, i, 5), (7, i, 2), (i, 3, 4), (5, 1, i), …)');
  TestSL('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \3', '((3, i, 5), (7, i, 2), (i, 3, 4), …)');
  TestSL('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \2', '((3, i, …), (7, i, …), …)');
  TestSL('matrix(3, 3, i, 5, 7, i, 2, i, 3, 4, 5, 1, i, i, 2, 3) \1', '((3, …), …)');

  TestML('❨❨1/3, i/3❩, ❨1/2, 1/4❩❩', '⎛  0.333333333333  0.333333333333⋅i⎞'#13#10'⎝             0.5              0.25⎠');
  TestML('SetNumDigits(❨❨1/3, i/3❩, ❨1/2, 1/4❩❩, 4)', '⎛  0.3333  0.3333⋅i⎞'#13#10'⎝     0.5      0.25⎠');
  TestML('SetNumDigits(❨❨1/3, i/3❩, ❨1/2, 1/4❩❩, 4) \ 1', '⎛0.3333    ⋯   ⎞'#13#10'⎝  ⋮       ⋱   ⎠');
  TestSL('❨❨1/3, i/3❩, ❨1/2, 1/4❩❩', '((0.333333333333, 0.333333333333⋅i), (0.5, 0.25))');
  TestSL('SetNumDigits(❨❨1/3, i/3❩, ❨1/2, 1/4❩❩, 4)', '((0.3333, 0.3333⋅i), (0.5, 0.25))');
  TestSL('SetNumDigits(❨❨1/3, i/3❩, ❨1/2, 1/4❩❩, 4) \ 1', '((0.3333, …), …)');

  Test('❨❨1, 2❩, ❨3, 4❩❩ + ❨❨i, 0❩, ❨0, i❩❩', 2, [1+ImaginaryUnit, 2, 3, 4+ImaginaryUnit]);
  Test('❨❨i, −i❩, ❨−i, i❩❩ + IdentityMatrix(2)', 2, [1+ImaginaryUnit, -ImaginaryUnit, -ImaginaryUnit, 1+ImaginaryUnit]);

  TestML('SetDigitGrouping(❨❨12345, 12345⋅i❩❩, 3)', '(  12 345  12 345⋅i)');
  TestSL('SetDigitGrouping(❨❨12345, 12345⋅i❩❩, 3)', '((12 345, 12 345⋅i))');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1]', ImaginaryUnit);
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2]', TASC(20.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[3]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[4]', TASC(40.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[5]', TASC(50.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[6]', TASC(60.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[7]', failure, 'Index 7 out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1]', TASC(60.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-2]', TASC(50.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-3]', TASC(40.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-4]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-5]', TASC(20.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-6]', ImaginaryUnit);
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-7]', failure, 'Index -7 out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[0]', failure, 'Index 0 out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, 0]', failure, 'Index (1, 0) out of bounds.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, 1]', ImaginaryUnit);
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, 2]', TASC(20.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, 3]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, 4]', failure, 'Index (1, 4) out of bounds.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2, 1]', TASC(40.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2, 2]', TASC(50.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2, 3]', TASC(60.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2, 4]', failure, 'Index (2, 4) out of bounds.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[3, 1]', failure, 'Index (3, 1) out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, -1]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, -2]', TASC(20.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, -3]', ImaginaryUnit);
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, -4]', failure, 'Index (1, -4) out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1, 1]', TASC(40.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1, 2]', TASC(50.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1, 3]', TASC(60.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1, 4]', failure, 'Index (-1, 4) out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-2, 1]', ImaginaryUnit);
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-2, 2]', TASC(20.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-2, 3]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-2, 4]', failure, 'Index (-2, 4) out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1, 0]', failure, 'Index (-1, 0) out of bounds.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1, -1]', TASC(60.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1, -2]', TASC(50.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-2, -1]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-2, -2]', TASC(20.0));

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, 0)]', failure, 'Index (1, 0) out of bounds.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, 1)]', ImaginaryUnit);
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, 2)]', TASC(20.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, 3)]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, 4)]', failure, 'Index (1, 4) out of bounds.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(2, 1)]', TASC(40.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(2, 2)]', TASC(50.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(2, 3)]', TASC(60.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(2, 4)]', failure, 'Index (2, 4) out of bounds.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(3, 1)]', failure, 'Index (3, 1) out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, -1)]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, -2)]', TASC(20.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, -3)]', ImaginaryUnit);
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, -4)]', failure, 'Index (1, -4) out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 1)]', TASC(40.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 2)]', TASC(50.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 3)]', TASC(60.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 4)]', failure, 'Index (-1, 4) out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-2, 1)]', ImaginaryUnit);
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-2, 2)]', TASC(20.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-2, 3)]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-2, 4)]', failure, 'Index (-2, 4) out of bounds.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 0)]', failure, 'Index (-1, 0) out of bounds.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-1, -1)]', TASC(60.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-1, -2)]', TASC(50.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-2, -1)]', TASC(30.0));
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-2, -2)]', TASC(20.0));

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[6] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[7] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[0] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-20] ≔ 100', failure, 'Left side cannot be assigned to.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, 1] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, 2] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, 5] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[1, -1] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2, 2] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2, 100] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2, -4] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[2, 0] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[-1, 1] ≔ 100', failure, 'Left side cannot be assigned to.');

  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, 1)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, 2)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, 5)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(1, -1)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(2, 2)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(2, 100)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(2, -4)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(2, 0)] ≔ 100', failure, 'Left side cannot be assigned to.');
  Test('❨❨i, 20, 30❩, ❨40, 50, 60❩❩[''(-1, 1)] ≔ 100', failure, 'Left side cannot be assigned to.');


  //
  // Colours
  //

  Test('color("red")', TRGB(clRed), True, 'rgba(1.000, 0.000, 0.000, 1.000)');
  Test('color("lime")', TRGB(clLime), True, 'rgba(0.000, 1.000, 0.000, 1.000)');
  Test('color("blue")', TRGB(clBlue), True, 'rgba(0.000, 0.000, 1.000, 1.000)');
  Test('color("white")', TRGB(clWhite), True, 'rgba(1.000, 1.000, 1.000, 1.000)');
  Test('color("black")', TRGB(clBlack), True, 'rgba(0.000, 0.000, 0.000, 1.000)');
  Test('color("yellow")', TRGB(clYellow), True, 'rgba(1.000, 1.000, 0.000, 1.000)');
  Test('color("aqua")', TRGB(clAqua), True, 'rgba(0.000, 1.000, 1.000, 1.000)');
  Test('color("fuchsia")', TRGB(clFuchsia), True, 'rgba(1.000, 0.000, 1.000, 1.000)');
  Test('rgb(0.2, 0.3, 0.4)', TRGB.Create(0.2, 0.3, 0.4), True, 'rgba(0.200, 0.300, 0.400, 1.000)');
  TestML('GetRGB(rgb(1/3, 1/7, 1/9)) @ (x ↦ SetNumDigits(x, 4))', 'red: 0.3333'#13#10'green: 0.1429'#13#10'blue: 0.1111');
  TestSL('GetRGB(rgb(1/3, 1/7, 1/9)) @ (x ↦ SetNumDigits(x, 4))', '(red: 0.3333, green: 0.1429, blue: 0.1111)');


  //
  // Lists
  //

  Test('''(1, 2, "cat", ❨1, 0❩)',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASO('cat'), ASO(ASR2(1, 0))]
    )
  );

  Test('''(1, 2, "cat", ❨1, 0❩)[1]', 1);
  Test('''(1, 2, "cat", ❨1, 0❩)[2]', 2);
  Test('''(1, 2, "cat", ❨1, 0❩)[3]', 'cat');
  Test('''(1, 2, "cat", ❨1, 0❩)[4]', ASO(ASR2(1, 0)));
  Test('''(1, 2, "cat", ❨1, 0❩)[5]', failure, 'Index 5 out of bounds.');

  Test('''(1, 2, "cat", ❨1, 0❩)[-1]', ASO(ASR2(1, 0)));
  Test('''(1, 2, "cat", ❨1, 0❩)[-2]', 'cat');
  Test('''(1, 2, "cat", ❨1, 0❩)[-3]', 2);
  Test('''(1, 2, "cat", ❨1, 0❩)[-4]', 1);
  Test('''(1, 2, "cat", ❨1, 0❩)[-5]', failure, 'Index -5 out of bounds.');

  Test('''(1, 2, "cat", ❨1, 0❩)[0]', failure, 'Index 0 out of bounds.');

  TestML('''(1, 2, 3)', '1'#13#10'2'#13#10'3');
  TestSL('''(1, 2, 3)', '(1, 2, 3)');

  TestML('''(123)', '123');
  TestSL('''(123)', '(123)');

  TestML('''(1/3, 1/7, 1/9)', '0.333333333333'#13#10'0.142857142857'#13#10'0.111111111111');
  TestML('''(1/3, 1/7, 1/9) @ RealNumber', '0.333333333333'#13#10'0.142857142857'#13#10'0.111111111111');
  TestML('''(1/3, 1/7, 1/9) @ RealNumber @ (x ↦ SetNumDigits(x, 4))', '0.3333'#13#10'0.1429'#13#10'0.1111');
  TestML('''(1/3, 1/7, 1/9) @ ToFraction', '1/3'#13#10'1/7'#13#10'1/9');

  TestSL('''(1/3, 1/7, 1/9)', '(0.333333333333, 0.142857142857, 0.111111111111)');
  TestSL('''(1/3, 1/7, 1/9) @ RealNumber', '(0.333333333333, 0.142857142857, 0.111111111111)');
  TestSL('''(1/3, 1/7, 1/9) @ RealNumber @ (x ↦ SetNumDigits(x, 4))', '(0.3333, 0.1429, 0.1111)');
  TestSL('''(1/3, 1/7, 1/9) @ ToFraction', '(1/3, 1/7, 1/9)');

  TestML('SequenceList(100)', '1'#13#10'2'#13#10'3'#13#10'4'#13#10'5'#13#10'6'#13#10'7'#13#10'8'#13#10'9'#13#10'10'#13#10'11'#13#10'12'#13#10'13'#13#10'14'#13#10'15'#13#10'16'
    +#13#10'17'#13#10'18'#13#10'19'#13#10'20'#13#10'21'#13#10'22'#13#10'23'#13#10'24'+#13#10'25'#13#10'26'#13#10'27'#13#10'28'#13#10'29'#13#10'30'#13#10'31'#13#10'32'#13#10'33'#13#10'34'#13#10'35'+#13#10'36'#13#10'37'#13#10'38'#13#10'39'#13#10'40'#13#10'41'#13#10'42'#13#10'43'#13#10'44'#13#10'45'#13#10'46'+#13#10'47'#13#10'48'#13#10'49'#13#10'50'#13#10'51'#13#10'52'#13#10'53'#13#10'54'#13#10'55'#13#10'56'
    +#13#10'57'#13#10'58'#13#10'59'#13#10'60'#13#10'61'#13#10'62'#13#10'63'#13#10'64'#13#10'65'#13#10'66'#13#10'67'#13#10'68'
    +#13#10'69'#13#10'70'#13#10'71'#13#10'72'#13#10'73'#13#10'74'#13#10'75'#13#10'76'#13#10'77'#13#10'78'#13#10'79'#13#10'80'#13#10'81'#13#10'82'#13#10'83'#13#10'84'#13#10'85'#13#10'86'#13#10'87'#13#10'88'#13#10'89'#13#10'90'#13#10'91'#13#10'92'#13#10'93'#13#10'94'#13#10'95'#13#10'96'#13#10'97'#13#10'98'#13#10'99'#13#10'100');
  TestSL('SequenceList(100)', '(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, '+'21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, '+'46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100)');

  TestML('SequenceList(200)', '1'#13#10'2'#13#10'3'#13#10'4'#13#10'5'#13#10'6'#13#10'7'#13#10'8'#13#10'9'#13#10'10'#13#10'11'#13#10'12'#13#10'13'#13#10'14'#13#10'15'#13#10'16'
    +#13#10'17'#13#10'18'#13#10'19'#13#10'20'#13#10'21'#13#10'22'#13#10'23'#13#10'24'
    +#13#10'25'#13#10'26'#13#10'27'#13#10'28'#13#10'29'#13#10'30'#13#10'31'#13#10'32'#13#10'33'#13#10'34'#13#10'35'
    +#13#10'36'#13#10'37'#13#10'38'#13#10'39'#13#10'40'#13#10'41'#13#10'42'#13#10'43'#13#10'44'#13#10'45'#13#10'46'
    +#13#10'47'#13#10'48'#13#10'49'#13#10'50'#13#10'51'#13#10'52'#13#10'53'#13#10'54'#13#10'55'#13#10'56'
    +#13#10'57'#13#10'58'#13#10'59'#13#10'60'#13#10'61'#13#10'62'#13#10'63'#13#10'64'#13#10'65'#13#10'66'#13#10'67'#13#10'68'
    +#13#10'69'#13#10'70'#13#10'71'#13#10'72'#13#10'73'#13#10'74'#13#10'75'#13#10'76'#13#10'77'#13#10'78'#13#10'79'#13#10'80'#13#10'81'#13#10'82'#13#10'83'#13#10'84'#13#10'85'#13#10'86'#13#10'87'#13#10'88'#13#10'89'#13#10'90'#13#10'91'#13#10'92'#13#10'93'#13#10'94'#13#10'95'#13#10'96'#13#10'97'#13#10'98'#13#10'99'#13#10'100'#13#10'⋮');
  TestSL('SequenceList(200)', '(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, '+'21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, '+'46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, …)');

  TestML('''(1, 2, 3, 4, 5) \ 0', '1'#13#10'2'#13#10'3'#13#10'4'#13#10'5');
  TestML('''(1, 2, 3, 4, 5) \ 10', '1'#13#10'2'#13#10'3'#13#10'4'#13#10'5');
  TestML('''(1, 2, 3, 4, 5) \ 5', '1'#13#10'2'#13#10'3'#13#10'4'#13#10'5');
  TestML('''(1, 2, 3, 4, 5) \ 4', '1'#13#10'2'#13#10'3'#13#10'4'#13#10'⋮');
  TestML('''(1, 2, 3, 4, 5) \ 3', '1'#13#10'2'#13#10'3'#13#10'⋮');
  TestML('''(1, 2, 3, 4, 5) \ 2', '1'#13#10'2'#13#10'⋮');
  TestML('''(1, 2, 3, 4, 5) \ 1', '1'#13#10'⋮');

  TestSL('''(1, 2, 3, 4, 5) \ 0', '(1, 2, 3, 4, 5)');
  TestSL('''(1, 2, 3, 4, 5) \ 10', '(1, 2, 3, 4, 5)');
  TestSL('''(1, 2, 3, 4, 5) \ 5', '(1, 2, 3, 4, 5)');
  TestSL('''(1, 2, 3, 4, 5) \ 4', '(1, 2, 3, 4, …)');
  TestSL('''(1, 2, 3, 4, 5) \ 3', '(1, 2, 3, …)');
  TestSL('''(1, 2, 3, 4, 5) \ 2', '(1, 2, …)');
  TestSL('''(1, 2, 3, 4, 5) \ 1', '(1, …)');

  TestML('''()', '');
  TestSL('''()', '()');

  TestSL('list(1, 2, 3)', '(1, 2, 3)');
  TestSL('list(1, 2, 3) \2', '(1, 2, …)');

  Test('''(1, 2, "cat", ❨1, 0❩)[3][1]', 'c');
  Test('''(1, 2, "cat", ❨1, 0❩)[3][2]', 'a');
  Test('''(1, 2, "cat", ❨1, 0❩)[3][3]', 't');
  Test('''(1, 2, "cat", ❨1, 0❩)[3][4]', failure, 'Index 4 out of bounds.');

  Test('''(1, 2, "cat", ❨1, 0❩)[3][-1]', 't');
  Test('''(1, 2, "cat", ❨1, 0❩)[3][-2]', 'a');
  Test('''(1, 2, "cat", ❨1, 0❩)[3][-3]', 'c');
  Test('''(1, 2, "cat", ❨1, 0❩)[3][-4]', failure, 'Index -4 out of bounds.');

  Test('''(1, 2, "cat", ❨1, 0❩)[3][0]', failure, 'Index 0 out of bounds.');

  Test('''(1, 2, "cat", ❨1, 0❩)[3][1] ≔ "C"', failure, 'Left side cannot be assigned to.');


  //
  // Structures
  //

  Test('struct("a": 42, "b": "dog")',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('dog')]
    )
  );

  TestML('struct("a": 1, "b": 2, "c": 3)', 'a: 1'#13#10'b: 2'#13#10'c: 3');
  TestSL('struct("a": 1, "b": 2, "c": 3)', '(a: 1, b: 2, c: 3)');

  TestML('struct("a": 1, "b": 2, "c": 3) @ square', 'a: 1'#13#10'b: 4'#13#10'c: 9');
  TestSL('struct("a": 1, "b": 2, "c": 3) @ square', '(a: 1, b: 4, c: 9)');

  TestML('struct("a": 1/3)', 'a: 0.333333333333');
  TestSL('struct("a": 1/3)', '(a: 0.333333333333)');

  TestML('struct("a": 1/3) @ ToFraction', 'a: 1/3');
  TestSL('struct("a": 1/3) @ ToFraction', '(a: 1/3)');

  TestML('struct("a": 1/3) @ (x ↦ SetNumDigits(x, 4))', 'a: 0.3333');
  TestSL('struct("a": 1/3) @ (x ↦ SetNumDigits(x, 4))', '(a: 0.3333)');

  TestML('s1 ≔ struct("a": 1, "b": 2); s2 ≔ struct("a": 10, "b": 20); struct("first": s1, "second": s2)',
    'first.a: 1'#13#10'first.b: 2'#13#10'second.a: 10'#13#10'second.b: 20');
  TestSL('s1 ≔ struct("a": 1, "b": 2); s2 ≔ struct("a": 10, "b": 20); struct("first": s1, "second": s2)',
    '(first: (a: 1, b: 2), second: (a: 10, b: 20))');

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  TestML('name("Andreas", "Rejbrand")', 'first: Andreas'#13#10'last: Rejbrand');
  TestSL('name("Andreas", "Rejbrand")', '(first: Andreas, last: Rejbrand)');
  TestML('person("Andreas", "Rejbrand", "male", 31)', 'name.first: Andreas'#13#10'name.last: Rejbrand'#13#10'sex: male'#13#10'age: 31');
  TestSL('person("Andreas", "Rejbrand", "male", 31)', '(name: (first: Andreas, last: Rejbrand), sex: male, age: 31)');
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
  TestML('struct("mother": Jane, "father": Mike)', 'mother.name.first: Jane'#13#10'mother.name.last: Smith'#13#10'mother.sex: female'#13#10'mother.age: 40'#13#10'father.name.first: Mike'#13#10'father.name.last: Doe'#13#10'father.sex: male'#13#10'father.age: 45');
  TestSL('struct("mother": Jane, "father": Mike)', '(mother: (name: (first: Jane, last: Smith), sex: female, age: 40), father: (name: (first: Mike, last: Doe), sex: male, age: 45))');

  TestML('''(Jane, Mike)', '(name: (first: Jane, last: Smith), sex: female, age: 40)'#13#10'(name: (first: Mike, last: Doe), sex: male, age: 45)');
  TestSL('''(Jane, Mike)', '((name: (first: Jane, last: Smith), sex: female, age: 40), (name: (first: Mike, last: Doe), sex: male, age: 45))');

  Test('delete(s1); delete(s2); delete(name); delete(person); delete(Jane); delete(Mike);', null);

  TestML('struct("a": struct("b": struct("c": struct("d": struct("e": struct("f": 42))))))', 'a.b.c.d.e.f: 42');
  TestSL('struct("a": struct("b": struct("c": struct("d": struct("e": struct("f": 42))))))', '(a: (b: (c: (d: (e: (f: 42))))))');

  TestML('struct("a": ''(1, 2, 3), "b": ''(10, 20, 30))', 'a: (1, 2, 3)'#13#10'b: (10, 20, 30)');
  TestSL('struct("a": ''(1, 2, 3), "b": ''(10, 20, 30))', '(a: (1, 2, 3), b: (10, 20, 30))');

  TestML('struct("a": ❨1, 2, 3❩, "b": ❨10, 20, 30❩)', 'a: (1, 2, 3)'#13#10'b: (10, 20, 30)');
  TestSL('struct("a": ❨1, 2, 3❩, "b": ❨10, 20, 30❩)', '(a: (1, 2, 3), b: (10, 20, 30))');

  TestML('struct("a": ❨1, 2, 3❩, "b": IdentityMatrix(2))', 'a: (1, 2, 3)'#13#10'b: ((1, 0), (0, 1))');
  TestSL('struct("a": ❨1, 2, 3❩, "b": IdentityMatrix(2))', '(a: (1, 2, 3), b: ((1, 0), (0, 1)))');

  Test('struct("a": 42, "b": "dog").a', 42);
  Test('struct("a": 42, "b": "dog").b', 'dog');
  Test('struct("a": 42, "b": "dog").c', failure, 'There is no member named "c".');

  Test('struct("a": 42, "b": "dog")["a"]', 42);
  Test('struct("a": 42, "b": "dog")["b"]', 'dog');
  Test('struct("a": 42, "b": "dog")["c"]', failure, 'There is no member named "c".');

  Test('struct("a": 42, "b": "dog")[1]', 42);
  Test('struct("a": 42, "b": "dog")[2]', 'dog');
  Test('struct("a": 42, "b": "dog")[3]', failure, 'Index 3 out of bounds.');

  Test('struct("a": 42, "b": "dog")[-1]', 'dog');
  Test('struct("a": 42, "b": "dog")[-2]', 42);
  Test('struct("a": 42, "b": "dog")[-3]', failure, 'Index -3 out of bounds.');

  Test('struct("a": 42, "b": "dog")[0]', failure, 'Index 0 out of bounds.');

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [
        ASOInt(42),
        ASO('dog'),
        TAlgosimStructure.CreateWithValue(
          ['dog', 'rat'],
          [ASO('Sally'), ASO('Fiona')]
        )
      ]
    )
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).a', 42);
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).b', 'dog');
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c',
    TAlgosimStructure.CreateWithValue(
      ['dog', 'rat'],
      [ASO('Sally'), ASO('Fiona')]
    )
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["a"]', 42);
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["b"]', 'dog');
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"]',
    TAlgosimStructure.CreateWithValue(
      ['dog', 'rat'],
      [ASO('Sally'), ASO('Fiona')]
    )
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[1]', 42);
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[2]', 'dog');
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3]',
    TAlgosimStructure.CreateWithValue(
      ['dog', 'rat'],
      [ASO('Sally'), ASO('Fiona')]
    )
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-3]', 42);
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-2]', 'dog');
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1]',
    TAlgosimStructure.CreateWithValue(
      ['dog', 'rat'],
      [ASO('Sally'), ASO('Fiona')]
    )
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c.rat',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c["rat"]',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c[2]',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c[-1]',
    'Fiona'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"].rat',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"]["rat"]',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"][2]',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"][-1]',
    'Fiona'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3].rat',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3]["rat"]',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3][2]',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3][-1]',
    'Fiona'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1].rat',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1]["rat"]',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1][2]',
    'Fiona'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1][-1]',
    'Fiona'
  );

  // ---

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c.dog',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c["dog"]',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c[1]',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c[-2]',
    'Sally'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"].dog',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"]["dog"]',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"][1]',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"][-2]',
    'Sally'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3].dog',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3]["dog"]',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3][1]',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3][-2]',
    'Sally'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1].dog',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1]["dog"]',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1][1]',
    'Sally'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1][-2]',
    'Sally'
  );

  // ---

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1].cat',
    failure, 'There is no member named "cat".'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1]["cat"]',
    failure, 'There is no member named "cat".'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1][3]',
    failure, 'Index 3 out of bounds.'
  );
  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[-1][-3]',
    failure, 'Index -3 out of bounds.'
  );


  // ---

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c.dog ≔ "Molly"',
    failure, 'Left side cannot be assigned to.');

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c ≔ "Molly"',
    failure, 'Left side cannot be assigned to.');

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"]["dog"] ≔ "Molly"',
    failure, 'Left side cannot be assigned to.');

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"] ≔ "Molly"',
    failure, 'Left side cannot be assigned to.');

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3]["dog"] ≔ "Molly"',
    failure, 'Left side cannot be assigned to.');

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3] ≔ "Molly"',
    failure, 'Left side cannot be assigned to.');

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c["dog"] ≔ "Molly"',
    failure, 'Left side cannot be assigned to.');

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"].dog ≔ "Molly"',
    failure, 'Left side cannot be assigned to.');

  // ---

  Test('struct("a": 42, "b": "dog").b[1]', 'd');
  Test('struct("a": 42, "b": "dog").b[2]', 'o');
  Test('struct("a": 42, "b": "dog").b[3]', 'g');
  Test('struct("a": 42, "b": "dog").b[4]', failure, 'Index 4 out of bounds.');

  Test('struct("a": 42, "b": "dog").b[-1]', 'g');
  Test('struct("a": 42, "b": "dog").b[-2]', 'o');
  Test('struct("a": 42, "b": "dog").b[-3]', 'd');
  Test('struct("a": 42, "b": "dog").b[-4]', failure, 'Index -4 out of bounds.');

  Test('struct("a": 42, "b": "dog").b[0]', failure, 'Index 0 out of bounds.');

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c.rat[1]',
    'F'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"].rat[2]',
    'i'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3].rat[3]',
    'o'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"]["rat"][4]',
    'n'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3][2][5]',
    'a'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c["rat"][6]',
    failure, 'Index 6 out of bounds.'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3]["rat"][-1]',
    'a'
  );

  // ---

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c.rat[1] ≔ "H"',
    failure, 'Left side cannot be assigned to.'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"].rat[2] ≔ "H"',
    failure, 'Left side cannot be assigned to.'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3].rat[3] ≔ "H"',
    failure, 'Left side cannot be assigned to.'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))["c"]["rat"][4] ≔ "H"',
    failure, 'Left side cannot be assigned to.'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3][2][5] ≔ "H"',
    failure, 'Left side cannot be assigned to.'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona")).c["rat"][6] ≔ "H"',
    failure, 'Left side cannot be assigned to.'
  );

  Test('struct("a": 42, "b": "dog", "c": struct("dog": "Sally", "rat": "Fiona"))[3]["rat"][-1] ≔ "H"',
    failure, 'Left side cannot be assigned to.'
  );


  //
  // Sets
  //

  TestML('{42}', '{42}');
  TestSL('{42}', '{42}');
  TestML('{}', '∅');
  TestSL('{}', '∅');
  TestML('∅', '∅');
  TestSL('∅', '∅');
  Test('#{}', 0);
  Test('#∅', 0);
  Test('#{1, 2, 3}', 3);
  Test('#{1, 2, 3, 2}', 3);
  Test('#{1, 2, 3, 1, 2, 3, 1, 2, 3, 3, 2, 1, 2, 3}', 3);
  Test('#{1, 2, 3, 1, 2, 3, 1, 4, 2, 3, 3, 2, 1, 2, 3}', 4);
  TestSL('{1/3} @ ToFraction', '{1/3}');
  TestSL('{1/3}', '{0.333333333333}');
  TestSL('{1/3} @ (x ↦ SetNumDigits(x, 4))', '{0.3333}');

  Test('{42}', intset([42]));
  Test('{42, 42, 42, 42}', intset([42]));
  Test('{}', intset([]));
  Test('{1, 2, 3}', intset([1, 2, 3]));
  Test('{1, 2, 3, 3, 3, 3}', intset([1, 2, 3]));
  Test('{1, 3, 2}', intset([1, 2, 3]));
  Test('{3, 2, 1}', intset([1, 2, 3]));
  Test('{3, 2, 2, 2, 3, 1}', intset([1, 2, 3]));

  Test('{1, 2, 3} = {3, 2, 1}', True);
  Test('{1, 2, 3} = {2, 3, 1}', True);
  Test('{1, 2, 3} = {2, 3, 1, 1, 1, 1}', True);
  Test('{1, 2, 3} = {1, 1, 2, 3}', True);
  Test('{1, 2, 3} = {1, 1, 2, 2, 3, 3}', True);
  Test('{1, 2, 3} = {3.0, 2, 1/2 + 1/2}', True);
  Test('{1, 2, 2, 3} = {1, 1, 2, 3}', True);
  Test('{1, 2, 2, 3} ≠ {1, 1, 2, 3, 4}', True);

  Test('{1} = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1}', true);

  Test('{10} @ type', asoset(['integer']));
  Test('{10, 10.0} @ type', asoset(['integer']));
  Test('{10.0} @ type', asoset(['integer']));
  Test('{10.2} @ type', asoset(['real number']));
  Test('{10.0, 10} @ type', asoset(['integer']));

  Test('{0} @ type', asoset(['integer']));
  Test('{0, 0.0} @ type', asoset(['integer']));
  Test('{0, 1/2 - 1/2} @ type', asoset(['integer']));
  Test('{0, i - i} @ type', asoset(['integer']));

  Test('{0.0} @ type', asoset(['integer']));
  Test('{0.0, 0} @ type', asoset(['integer']));
  Test('{0.0, 1/2 - 1/2} @ type', asoset(['integer']));
  Test('{0.0, i - i} @ type', asoset(['integer']));

  Test('{1/2 - 1/2} @ type', asoset(['rational number']));
  Test('{1/2 - 1/2, 0} @ type', asoset(['rational number']));
  Test('{1/2 - 1/2, 0.0} @ type', asoset(['rational number']));
  Test('{1/2 - 1/2, i - i} @ type', asoset(['rational number']));

  Test('{i - i} @ type', asoset(['integer']));
  Test('{i - i, 0} @ type', asoset(['integer']));
  Test('{i - i, 0.0} @ type', asoset(['integer']));
  Test('{i - i, 1/2 - 1/2} @ type', asoset(['integer']));

  Test('{10, 20, 30} @ type', asoset(['integer']));
  Test('{10, 20, 20, 30} @ type', asoset(['integer']));
  Test('{10, 20, 20.0, 30} @ type', asoset(['integer']));
  Test('{10, 20.0, 30} @ type', asoset(['integer']));
  Test('{10, 20.5, 30} @ type', asoset(['integer', 'real number']));
  Test('{10, 20.0, 20, 30} @ type', asoset(['integer']));
  Test('{10, 20.5, 20, 30} @ type', asoset(['integer', 'real number']));

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', asosetex(
    [
      ASO(ASR3(1, 0, 0)),
      ASO(ASR3(0, 1, 0)),
      ASO(ASR3(0, 0, 1))
    ]));

  Test('{❨1, 0, 0❩, ❨1, 0, 0❩, ❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', asosetex(
    [
      ASO(ASR3(1, 0, 0)),
      ASO(ASR3(0, 1, 0)),
      ASO(ASR3(0, 0, 1))
    ]));
  Test('{❨1, 0, 0❩, ❨1, 0, 0❩, ❨1, i-i, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', asosetex(
    [
      ASO(ASR3(1, 0, 0)),
      ASO(ASR3(0, 1, 0)),
      ASO(ASR3(0, 0, 1))
    ]));


  //
  // Binary data
  //

  TestML('BinaryData("test")', '74 00 65 00 73 00 74 00');
  TestSL('BinaryData("test")', '74 00 65 00 73 00 74 00');
  TestML('BinaryData(0)', '00 00 00 00 00 00 00 00');
  TestSL('BinaryData(0)', '00 00 00 00 00 00 00 00');
  TestML('BinaryData(255)', 'FF 00 00 00 00 00 00 00');
  TestSL('BinaryData(255)', 'FF 00 00 00 00 00 00 00');
  TestML('BinaryData(16#12ABCDEF123456)', '56 34 12 EF CD AB 12 00');
  TestSL('BinaryData(16#12ABCDEF123456)', '56 34 12 EF CD AB 12 00');
  TestML('BinaryData("∫")', '2B 22');
  TestSL('BinaryData("∫")', '2B 22');
  TestML('BinaryData("")', '');
  TestSL('BinaryData("")', '');





  //
  //
  //  CHAPTER 2
  //  Syntax and operators.
  //
  //

  Chapter('Syntax and operators');

  //
  // Syntax
  //

  Test('1+1', 2, True, '2');
  Test('1 + 1', 2, True, '2');
  Test(' 1 + 1 ', 2, True, '2');
  Test('     1   +   1  ', 2, True, '2');
  Test('1'#13#10'+1', 2, True, '2');
  Test('1'#13#10#9'+1', 2, True, '2');
  Test('   1   '#9#9#13#10#9'  +1'#13#10#13#10, 2, True, '2');
  Test('1++1', 2, True, '2');
  Test('+1+1', 2, True, '2');
  Test('+1+(+1)', 2, True, '2');
  Test('(1)+(1)', 2, True, '2');
  Test('(1+1)', 2, True, '2');
  Test('((1)+(1))', 2, True, '2');
  Test('((((1))))+((1))', 2, True, '2');
  Test('(((1))+((((1)))))', 2, True, '2');
  Test('+1+(+1)', 2, True, '2');
  Test('(+1)+(+1)', 2, True, '2');
  Test('((+1)+(+1))', 2, True, '2');
  Test('((+(1))+(+((1))))', 2, True, '2');
  Test('( ( +(1  ) ) + ( +( ( 1) )) )', 2, True, '2');
  Test('++( +( + (1  ) ) + ( +( + ( + +1) )) )', 2, True, '2');
  Test('+++++++++++++1', 1, True, '1');
  Test('(+(+(+(+(+(+(+(+(+(+(+(+(+(1))))))))))))))', 1, True, '1');
  Test('+'#13#10'1', 1, True, '1');
  Test('1'#13#10'+'#13#10'2', 3, True, '3');

  Test('−5', -5, True, '−5');
  Test('6−2', 4, True, '4');
  Test('3−8', -5, True, '−5');
  Test('-5', -5, True, '−5');
  Test('6-2', 4, True, '4');
  Test('3-8', -5, True, '−5');
  Test('−−2', 2, True, '2');
  Test('−−−2', -2, True, '−2');
  Test('−−−−2', 2, True, '2');
  Test('−(−(−(−2)))', 2, True, '2');
  Test('(−(−(−(−(2)))))', 2, True, '2');
  Test(' ( −( −(−(   −(  2)  ))  ) )', 2, True, '2');
  Test(' ( −( −(−(-   −(  2)  ))  ) )', -2, True, '−2');

  Test('1+6+3+8+2+9', 29, True, '29');
  Test('1 + 6 + 3 + 8 + 2 + 9', 29, True, '29');
  Test('1 + 6 + (3 + 8) + 2 + 9', 29, True, '29');
  Test('1 + ((6 + (3 + 8)) + (2 + 9))', 29, True, '29');
  Test('1 + ((6 + (3 + 8)) - (-2 - 9))', 29, True, '29');

  Test('2563 − 500 − 60 − 3', 2000, True, '2000');
  Test('2563 − (500 + 60 + 3)', 2000, True, '2000');
  Test('2563 − 500 + 60 + 3', 2126, True, '2126');

  Test('1+2⋅3', 7, True, '7');
  Test('(1+2)⋅3', 9, True, '9');
  Test('[1+2]⋅3', 9, True, '9');
  Test('-3 ⋅ -7', 21, True, '21');
  Test('(-3) ⋅ (-7)', 21, True, '21');
  Test('-(-3) ⋅ (-7)', -21, True, '−21');

  Test('5⋅4⋅3⋅2⋅1', 120, True, '120');
  Test('(5⋅4)⋅(3⋅[2⋅1])', 120, True, '120');

  Test('7 ⋅ (300 − [20 + 1])', 1953, True, '1953');
  Test('7 ⋅ (300 − 20 + 1)', 1967, True, '1967');
  Test('7 ⋅ 300 − 20 + 1', 2081, True, '2081');

  Test('-[7 ⋅ (300 − [20 + 1])] ⋅ (-1)', 1953, True, '1953');
  Test('-[7 ⋅ (300 − [20 + 1])] ⋅ -1', 1953, True, '1953');
  Test('---------[----7 ⋅ (--300 − --[--20 + ----1])] ⋅ -------1', 1953, True, '1953');
  Test('---------[----7 ⋅ (--300 − --[(-(-(20))) + -(--(-1))])] ⋅ -------1', 1953, True, '1953');

  Test('1.2-0.2', 1.0, True, '1');
  Test('(-1.2)-(-0.2)', -1.0, True, '−1');

  Test('2^10', 1024, True, '1024');
  Test('(2^3)^2', 64, True, '64');
  Test('2^(3^2)', 512, True, '512');
  Test('2^3^2', 512, True, '512');
  Test('3^2^3', 6561, True, '6561');
  Test('(3^2)^3', 729, True, '729');
  Test('−2^6', -64, True, '−64');
  Test('(−2)^6', 64, True, '64');
  Test('−(2^6)', -64, True, '−64');
  Test('2^-6', Rat(1, 64), True, '1/64');
  Test('-2^-6', Rat(-1, 64), True, '−1/64');

  Test('5!', 120, True, '120');
  Test('−5!', -120, True, '−120');
  Test('3!!', 720, True, '720');
  Test('1+3!!+2', 723, True, '723');

  Eps; Test('5% ⋅ 1100', 55.0, True, '55');

  Test('0.3 ⋅ 0.5', 0.15, True, '0.15');

  Test('100/5/10', 2, True, '2');
  Test('(100/5)/10', 2, True, '2');
  Test('100/(5/10)', Rat(200, 1), True, '200');

  Eps; Test('((7/[3+2])−5!+(4−5/3)^2)⋅(23−(5.369−2.15+3.1^3)⋅(3.11−0.5^3)^−2)', -2183.36671533110680270263782464652813044688670725050277755);

  Test('1/5^-1', Rat(5, 1), True, '5');
  Test('1/(1/5)', Rat(5, 1), True, '5');
  Test('1/-(1/5^-1)', Rat(-1, 5), True, '−1/5');

  Test('−√100', -10.0, True, '−10');
  Test('10−√100', 0.0, True, '0');
  Test('√100−10', 0.0, True, '0');
  Test('√100−19', -9.0, True, '−9');
  Test('√(100−19)', 9.0, True, '9');
  Test('√25^2', 25.0, True, '25');
  Test('√√16', 2.0, True, '2');
  Test('√(√16)', 2.0, True, '2');
  Test('√(√(16))', 2.0, True, '2');

  Test('√−(−20−5)', 5.0, True, '5');
  Test('√−−−(−20−5)', 5.0, True, '5');
  Test('√(−(−(−(−20−5))))', 5.0, True, '5');
  Test('−√−−−(−20−5)', -5.0, True, '−5');
  Test('−√−−√−−16', -2.0, True, '−2');

  Eps; Test('5!%', 1.2, True, '1.2');
  Eps; Test('3!!%', 7.2, True, '7.2');
  Eps; Test('-3!!%', -7.2, True, '−7.2');
  Eps; Test('---3!!%', -7.2, True, '−7.2');
  Eps; Test('----3!!%', 7.2, True, '7.2');
  Eps; Test('-(3!)!%', -7.2, True, '−7.2');
  Eps; Test('-(3!!)%', -7.2, True, '−7.2');
  Eps; Test('-((3!)!)%', -7.2, True, '−7.2');
  Eps; Test('---((-(((3!)!)%)))', 7.2, True, '7.2');
  Eps; Test('---((-(((3!)!))%))', 7.2, True, '7.2');
  Eps; Test('---((-(((3!)!)))%)', 7.2, True, '7.2');
  Eps; Test('---((-(((3!)!))))%', 7.2, True, '7.2');
  Eps; Test('1%%%%%', 1E-10);
  Eps; Test('(1%)%%%%', 1E-10);
  Eps; Test('(1%%)%%%', 1E-10);
  Eps; Test('((1%)%)%%%', 1E-10);
  Eps; Test('(((((1%)%)%)%)%)', 1E-10);
  Eps; Test('(((((-1%)%)%)%)%)', -1E-10);
  Eps; Test('-(-((((-1%)%)%)%)%)', -1E-10);
  Eps; Test('-(-((------((-1%)%)%)%)%)', -1E-10);
  Eps; Test('-(-((------((-0%)%)%)%)%)', 0.0);
  Test('360°', 2*Pi);

  Test('1!!!!!!!!!!!', 1, True, '1');
  Test('----------1!!!!!!!!!!!', 1, True, '1');
  Eps; Test('----------1!!!!!!!!!!!%', 0.01, True, '0.01');
  Eps; Test('----------1!!!!!!!!!!!%%%%%', 1E-10);
  Eps; Test('--−---−-----1!!!!!!!!!!!%%%%%', 1E-10);
  Eps; Test('--−---−-----1!!!!!!!!!!!%%%%%°', 1E-10 * pi/180);
  Eps; Test('--−---−-----1!!!!!!!!!!!%%%%%°°', 1E-10 * pi*pi/(180*180));
  Test('0%%°%%°%%°', 0.0);
  Test('-----0%%°%%°%%°', 0.0);

  Test('250 +', ESyntaxException, 'Operand missing after operator + at column 5.');
  Test('250 −', ESyntaxException, 'Operand missing after operator − at column 5.');
  Test('250 ⋅', ESyntaxException, 'Operand missing after operator ⋅ at column 5.');
  Test('250 /', ESyntaxException, 'Operand missing after operator / at column 5.');

  Test('+5', 5, True, '5');
  Test('−5', -5, True, '−5');
  Test('⋅5', ESyntaxException, 'Operand missing before operator ⋅ at column 1.');
  Test('/5', ESyntaxException, 'Operand missing before operator / at column 1.');

  Test('2 + (+ 4)', 6, True, '6');
  Test('2 + (− 4)', -2, True, '−2');
  Test('2 + (⋅ 4)', ESyntaxException, 'Operand missing before operator ⋅ at column 6.');
  Test('2 + (/ 4)', ESyntaxException, 'Operand missing before operator / at column 6.');

  Test('1 + 3 /', ESyntaxException, 'Operand missing after operator / at column 7.');
  Test('1 + (3 /)', ESyntaxException, 'Operand missing after operator / at column 8.');

  Test('!', ESyntaxException, 'Operand missing before operator ! at column 1.');
  Test('5+(!)', ESyntaxException, 'Operand missing before operator ! at column 4.');
  Test('5+(5!)', 125, True, '125');

  Test('10 + √36', 16.0, True, '16');
  Test('10 + √', ESyntaxException, 'Operand missing after operator √ at column 6.');
  Test('10 + √+36', 16.0, True, '16');
  Test('(10 + √)+36', ESyntaxException, 'Operand missing after operator √ at column 7.');

  Test('2^-1', Rat(1, 2), True, '1/2');
  Test('2.^-1.', 0.5, True, '0.5');

  Test('3⋅⋅6', ESyntaxException, 'Operand missing before operator ⋅ at column 3.');
  Test('3//6', ESyntaxException, 'Operand missing before operator / at column 3.');
  Test('3^^6', ESyntaxException, 'Operand missing before operator ^ at column 3.');
  Test('3/^6', ESyntaxException, 'Operand missing before operator ^ at column 3.');
  Test('3^⋅6', ESyntaxException, 'Operand missing before operator ⋅ at column 3.');
  Test('3⋅⋅⋅6', ESyntaxException, 'Operand missing before operator ⋅ at column 3.');
  Test('3/⋅^6', ESyntaxException, 'Operand missing before operator ⋅ at column 3.');
  Test('3⋅/^6', ESyntaxException, 'Operand missing before operator / at column 3.');
  Test('3⋅^/6', ESyntaxException, 'Operand missing before operator ^ at column 3.');

  Test('4**+8***+3**', 15, True, '15');
  Test('*4', ESyntaxException, 'Operand missing before operator * at column 1.');
  Test('**4', ESyntaxException, 'Operand missing before operator * at column 1.');
  Test('(4+i)*', 4-ImaginaryUnit, True, '4 − i');
  Test('(4+i)**', 4+ImaginaryUnit, True, '4 + i');
  Test('(4*+i**)**', 4+ImaginaryUnit, True, '4 + i');

  Test('100⋅[1+(2+7)⋅(5−2)^(6−9−2)]', Rat(2800, 27), True, '2800/27');
  Test('100⋅[1+(2+7)⋅(5−2^(6−9−2)]', ESyntaxException, 'Closing bracket ] at column 26 doesn''t match the most recent opening bracket ( at column 14.');
  Test('100⋅[1+(2+7)⋅(5−2)^(6−9−2]', ESyntaxException, 'Closing bracket ] at column 26 doesn''t match the most recent opening bracket ( at column 20.');
  Test('100⋅[1+(2+7)⋅5−2)^(6−9−2)]', ESyntaxException, 'Closing bracket ) at column 17 doesn''t match the most recent opening bracket [ at column 5.');
  Test('100⋅[1+(2+7)⋅(5−2)^(6−9−2)', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "[".');
  Test('100⋅[1+(2+7)⋅(5−2)^(6−9−2)]]', ESyntaxException, 'Negative bracket level starting at column 28.');
  Test('100)⋅[1+(2+7)⋅(5−2)^(6−9−2)]', ESyntaxException, 'Negative bracket level starting at column 4.');

  Test('(1+7)⋅(2+7)', 72, True, '72');
  Test('[1+7]⋅[2+7]', 72, True, '72');
  Test('[1+7]⋅(2+7)', 72, True, '72');
  Test('[1+7]⋅[2+7)', ESyntaxException, 'Closing bracket ) at column 11 doesn''t match the most recent opening bracket [ at column 7.');
  Test('(1+7)⋅(2+7', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "(".');
  Test('7+((1+7)⋅(2+7', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "((".');
  Test('[7+((1+7)⋅(2+7', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "[((".');
  Test('(4+7)⋅(1+3]', ESyntaxException, 'Closing bracket ] at column 11 doesn''t match the most recent opening bracket ( at column 7.');
  Test('(4+7)⋅(1+(3)]', ESyntaxException, 'Closing bracket ] at column 13 doesn''t match the most recent opening bracket ( at column 7.');
  Test('(4+7)⋅(1+3))', ESyntaxException, 'Negative bracket level starting at column 12.');
  Test('1 + (4+7)⋅(1+3)) + 3', ESyntaxException, 'Negative bracket level starting at column 16.');
  Test('1 + (4+7)⋅() + 3', ESyntaxException, 'Empty bracket at column 11.');
  Test('1 + (4+7)⋅[] + 3', ESyntaxException, 'Empty bracket at column 11.');
  Test('((((((((((', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "((((((((((".');
  Test('((((((((([', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "((((((((([".');
  Test('[(((((((([', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "[(((((((([".');
  Test('((([(((1)))])))', 1, True, '1');
  Test('((([((()))])))', ESyntaxException, 'Empty bracket at column 7.');
  Test('((([(((1)))]))', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "(".');
  Test('((([(((1)))]', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "(((".');
  Test('((([(((1)))', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "((([".');
  Test('((([(((1)', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "((([((".');
  Test('((([(((1', ESyntaxException, 'There is at least one open bracket at the end of the expression. The following brackets are open: "((([(((".');
  Test('((([(((1)))))))', ESyntaxException, 'Closing bracket ) at column 12 doesn''t match the most recent opening bracket [ at column 4.');
  Test('((([(((1)))])))]', ESyntaxException, 'Negative bracket level starting at column 16.');
  Test('((([(((1)))])))] + 2 + 3', ESyntaxException, 'Negative bracket level starting at column 16.');
  Test('((([(((1)))]))) + 2 + 3', 6, True, '6');

  Test('"alfa" + "beta', ESyntaxException, 'Unterminated string literal "beta" starting at column 10.');
  Test('10⋅("alfa" + "beta)', ESyntaxException, 'Unterminated string literal "beta)" starting at column 14.');
  Test('10⋅("alfa" + "beta")', 'alfabetaalfabetaalfabetaalfabetaalfabetaalfabetaalfabetaalfabetaalfabetaalfabeta');
  Test('"', ESyntaxException, 'Unterminated string literal "" starting at column 1.');
  Test('"test" + "', ESyntaxException, 'Unterminated string literal "" starting at column 10.');
  Test('"This is a ""good"" thing."', 'This is a "good" thing.');

  Test('123 456', ESyntaxException, 'Adjacent values at column 5; a comma or operator was expected.');
  Test('cat dog', ESyntaxException, 'Adjacent values at column 5; a comma or operator was expected.');
  Test('(1+2)3', ESyntaxException, 'Adjacent values at column 6; a comma or operator was expected.');
  Test('1(2+3)', failure, 'An object of type function was expected as argument 1, but an object of type integer was given.');

  Test('"alfa" + "beta + "gamma"', ESyntaxException, 'Adjacent values at column 19; a comma or operator was expected.');

  Test('5! 7!', ESyntaxException, 'Adjacent values at column 4; a comma or operator was expected.');
  Test('sin(5! 7!)', ESyntaxException, 'Adjacent values at column 8; a comma or operator was expected.');

  Test('5! (7!)', ESyntaxException, 'Adjacent values at column 4; a comma or operator was expected.');
  Test('sin(5! (7!))', ESyntaxException, 'Adjacent values at column 8; a comma or operator was expected.');

  Test('5!(5)', ESyntaxException, 'Adjacent values at column 3; a comma or operator was expected.');
  Test(' 5!(5)', ESyntaxException, 'Adjacent values at column 4; a comma or operator was expected.');
  Test(' 5! (5)', ESyntaxException, 'Adjacent values at column 5; a comma or operator was expected.');
  Test(' 5!   (5)', ESyntaxException, 'Adjacent values at column 7; a comma or operator was expected.');
  Test('( 5!   (5))', ESyntaxException, 'Adjacent values at column 8; a comma or operator was expected.');

  Test('1', 1, True, '1');
  Test('1; 2', 2, True, '2');
  Test('1; 2; 3', 3, True, '3');
  Test('1; 2; 3; 4', 4, True, '4');

  Test('1;', null);
  Test('1; 2;', null);
  Test('1; 2; 3;', null);
  Test('1; 2; 3; 4;', null);

  Eps; Test('41⋅(4325−[123+56^2−20%+(21−52+23+48−30)/2])^2', 46171963.04, True, '46171963.04');
  Test('41⋅(4325−[123+56^2−20%+(21−52+23+48−30)/2])^2; "cat"', 'cat', True, 'cat');
  Test('41⋅(4325−[123+56^2−20%+(21−52+23+48−30)/2])^2;', null);
  Test('x ≔ 5; x^2', 25, True, '25');
  Test('x ≔ 5; x ≔ x^2; x ≔ 10⋅x; x ≔ 2⋅x; x ≔ x/5; x', 100, True, '100');
  Test('f ≔ x ↦ 2⋅x^2; f(7)', 98, True, '98');
  Test('f ≔ x ↦ 2⋅x^2; y ≔ f(7); y + 2', 100, True, '100');
  Test('delete(f); delete(x); delete(y);', null);
  Test('g ≔ x ↦ √(2⋅x);', null);
  Test('g(50)', 10.0, True, '10');
  Test('delete(g)', success);

  Test('0;', null);
  Test('.;', null);
  Test('.;.;', null);
  Test('.;.', 0.0);

  Test('([1;])', null);
  Test('type([1;])', 'null');
  Test('type([1;]);', null);
  Test('type(type([1;]);)', 'null');
  Test('type(type([1;]););', null);
  Test('type(type(1;););', null);

  Test(';', ESyntaxException, 'Operand missing before operator ; at column 1.');
  Test('; 5^2', ESyntaxException, 'Operand missing before operator ; at column 1.');
  Test('7+(; 5^2)', ESyntaxException, 'Operand missing before operator ; at column 4.');
  Test('7+(10; 5^2)', 32);
  Test('7;;8', ESyntaxException, 'Operand missing before operator ; at column 3.');
  Test('7 ;; 8', ESyntaxException, 'Operand missing before operator ; at column 4.');
  Test('7 ; ; 8', ESyntaxException, 'Operand missing before operator ; at column 5.');
  Test('7 ;'#9'; 8', ESyntaxException, 'Operand missing before operator ; at column 5.');
  Test('7 ;'#13#10'; 8', ESyntaxException, 'Operand missing before operator ; at column 6.');


  //
  // Stack overflow protection
  //

  // If any of the following tests crashes the process, the self test is considered to have failed!

  Test(StringOfChar(MINUS_SIGN, 200) + '1', +1);
  Test(StringOfChar(MINUS_SIGN, 201) + '1', -1);
  Test('i' + StringOfChar('*', 200), +ImaginaryUnit);
  Test('i' + StringOfChar('*', 201), -ImaginaryUnit);

  Test(StringOfChar(MINUS_SIGN, 255) + '1', -1);
  Test('i' + StringOfChar('*', 255), -ImaginaryUnit);

  Test(StringOfChar(MINUS_SIGN, 256) + '1', EParseException, 'Expression is too deep.');
  Test('i' + StringOfChar('*', 256), EParseException, 'Expression is too deep.');

  // Collapsible operators

  Test(DupeString('123; ', 200) + '123456', 123456);
{$IFNDEF QuickTest}
  Test(DupeString('123; ', 1000) + '123456', 123456);
  Test(DupeString('123; ', 1023) + '123456', 123456);
  Test(DupeString('123; ', 1024) + '123456', EParseException, 'Expression is too deep.');

  Test('1' + DupeString('+1', 999), 1000);
  Test('1' + DupeString('+1', 1023), 1024);
  Test('1' + DupeString('+1', 1024), EParseException, 'Expression is too deep.');
{$ENDIF}

  // TODO: Parser should have better time complexity (in general).
  // TODO: Collapsible operators should not be parsed using recursion; there
  //       should not be a limit to how many operands a collapsible operator has.
  //       1M should not be an issue.



  //
  // Operators
  //

  // Unary plus (=identity)

  Test('+7', 7, True, '7');
  Test('+7.8', 7.8, True, '7.8');
  Test('+(-7)', -7, True, '−7');
  Test('+(-7.8)', -7.8, True, '−7.8');
  Test('+(7/3)', Rat(7, 3), True, '7/3');
  Test('+(-7/3)', Rat(-7, 3), True, '−7/3');
  Test('+i', ImaginaryUnit, True, 'i');
  Test('+48566935847526514', 48566935847526514, True, '48566935847526514');
  Test('+(2+3⋅i)', 2+3*ImaginaryUnit, True, '2 + 3⋅i');
  Test('+(−2−3⋅i)', -2-3*ImaginaryUnit, True, '−2 − 3⋅i');
  Test('+❨1, 2, 3❩', [1, 2, 3]);
  Test('+❨1, 2, 3⋅i❩', [1, 2, 3*ImaginaryUnit]);
  Test('+❨1, -2, 3⋅i❩', [1, -2, 3*ImaginaryUnit]);
  Test('+❨❨5, 6❩, ❨−5, −6❩❩', 2, [5, 6, -5, -6]);
  Test('+❨❨5, 6❩, ❨−5⋅i, −6❩❩', 2, [5, 6, -5*ImaginaryUnit, -6]);
  Test('+"cat"', 'cat', True, 'cat');
  Test('+true', true);
  Test('+false', false);
  Test('+color("red")', TRGB.Create(1, 0, 0));
  TestSL('+''(1,2)', '(1, 2)');
  Test('+(0;)', null);
  TestSL('+{}', '∅');
  TestSL('+{"cat"}', '{cat}');
  TestSL('+BinaryData("A")', '41 00');
  TestSL('+struct("a": 42)', '(a: 42)');
  Test('f ≔ x ↦ x^2; (+f)(4)', 16);
  Test('+delete(f)', success);

  // Unary minus

  Test('−0', 0, True, '0');
  Test('−0.0', 0.0, True, '0');
  Test('−0.0⋅i', TASC(0.0), True, '0');
  Test('−5', -5, True, '−5');
  Test('−(−5)', 5, True, '5');
  Test('−5.9', -5.9, True, '−5.9');
  Test('−(−5.9)', 5.9, True, '5.9');
  Test('−(2/3)', Rat(-2, 3), True, '−2/3');
  Test('−(−2/3)', Rat(2, 3), True, '2/3');
  Test('−((−2)/3)', Rat(2, 3), True, '2/3');
  Test('−(−(2/3))', Rat(2, 3), True, '2/3');
  Test('−π', -Pi, True, '−3.14159265359');
  Test('−(−π)', Pi, True, '3.14159265359');
  Test('−i', -ImaginaryUnit, True, '−i');
  Test('−3⋅i', -3*ImaginaryUnit, True, '−3⋅i');
  Test('−(−3⋅i)', 3*ImaginaryUnit, True, '3⋅i');
  Test('−(5+3⋅i)', -5-3*ImaginaryUnit, True, '−5 − 3⋅i');
  Test('−(5−3⋅i)', -5+3*ImaginaryUnit, True, '−5 + 3⋅i');
  Test('−(−5−3⋅i)', 5+3*ImaginaryUnit, True, '5 + 3⋅i');
  Test('−(−5.4−3.2⋅i)', 5.4+3.2*ImaginaryUnit, True, '5.4 + 3.2⋅i');
  Test('−48566935847526514', -48566935847526514, True, '−48566935847526514');

  Test('−❨8, 5.5, 3, 2❩', [-8, -5.5, -3, -2]);
  Test('−❨8, −5.5, 3, 2❩', [-8, 5.5, -3, -2]);
  Test('−❨8, −5.5, 3⋅i, 2❩', [-8, 5.5, -3*ImaginaryUnit, -2]);
  Test('−❨i, 2❩', [-ImaginaryUnit, -2]);
  Test('−❨1−i, 2❩', [-1+ImaginaryUnit, -2]);
  Test('−❨1❩', [-1]);
  Test('−❨i❩', [-ImaginaryUnit]);
  Test('−❨−1❩', [1]);
  Test('−❨−i❩', [ImaginaryUnit]);
  Test('−❨0❩', [0]);
  Test('−❨0, 0❩', [0, 0]);

  Test('−❨❨10, 20❩, ❨30, 40❩❩', 2, [-10, -20, -30, -40]);
  Test('−❨❨−10, 20❩, ❨30, 40❩❩', 2, [10, -20, -30, -40]);
  Test('−❨❨10, 20❩, ❨30, 40⋅i❩❩', 2, [-10, -20, -30, -40*ImaginaryUnit]);
  Test('−❨❨10, 20❩, ❨−30, −40⋅i❩❩', 2, [-10, -20, 30, 40*ImaginaryUnit]);

  Test('−"test"', failure, 'Cannot compute unary minus of an object of type "string".');
  Test('−true', failure, 'Cannot compute unary minus of an object of type "boolean".');
  Test('−false', failure, 'Cannot compute unary minus of an object of type "boolean".');
  Test('−(0;)', failure, 'Cannot compute unary minus of an object of type "null".');
  Test('a ≔ 4; −delete(a)', failure, 'Cannot compute unary minus of an object of type "success indication".');

  // Binary plus

  // - two numbers

  Test('1+1', 2, True, '2');
  Test('75283954 + 59852361', 135136315, True, '135136315');
  Test('500 + (-100)', 400, True, '400');
  Test('(-60) + 300', 240, True, '240');
  Test('(-60) + (-300)', -360, True, '−360');
  Test('48566935847526514 + 17584986538592057', 66151922386118571, True, '66151922386118571');
  Test('48566935847526514 + (-17584986538592057)', 30981949308934457, True, '30981949308934457');
  Test('(-17584986538592057) + 48566935847526514', 30981949308934457, True, '30981949308934457');
  Test('(-48566935847526514) + 17584986538592057', -30981949308934457, True, '−30981949308934457');
  Test('17584986538592057 + (-48566935847526514)', -30981949308934457, True, '−30981949308934457');
//  Test('9223372036854775807 + 0', 9223372036854775807, True, '9223372036854775807'); // TODO: see if it can be improved
  Test('0 + 9223372036854775807', 9223372036854775807, True, '9223372036854775807');
  Test('1000 + 200 + 30 + 4', 1234, True, '1234');

  Test('2/7 + 1/7', Rat(3, 7), True, '3/7');
  Test('3/7 + (-1/7)', Rat(2, 7), True, '2/7');
  Test('(-3/7) + (-1/7)', Rat(-4, 7), True, '−4/7');
  Test('(-3/7) + 1/7', Rat(-2, 7), True, '−2/7');
  Test('1/7 + 1/7 + 2/7 + 3/7', Rat(1, 1), True, '1');
  Test('1/7 + 1/7 + 2/7 + 4/7 + 0 + (-1/7)', Rat(1, 1), True, '1');
  Test('3/8 + 1/8', Rat(1, 2), True, '1/2');
  Test('11/10 + (-1/10)', Rat(1, 1), True, '1');
  Test('11/10 + 1/10', Rat(6, 5), True, '6/5');

  Test('5/12 + 3', Rat(41, 12), True, '41/12');
  Test('5/12 + (-1)', Rat(-7, 12), True, '−7/12');
  Test('1 + 2/3', Rat(5, 3), True, '5/3');
  Test('1 + (-2/3)', Rat(1, 3), True, '1/3');
  Test('0 + (-2/3)', Rat(-2, 3), True, '−2/3');
  Test('17584986538592057 + 1/2', Rat(35169973077184115, 2), True, '35169973077184115/2');
  Test('17584986538592057 + (-1/2)', Rat(35169973077184113, 2), True, '35169973077184113/2');

  Test('12.3 + 0.5', 12.8, True, '12.8');
  Test('12.3 + (−0.5)', 11.8, True, '11.8');
  Test('(−12.3) + (−0.5)', -12.8, True, '−12.8');
  Test('(−12.3) + 0.5', -11.8, True, '−11.8');

  Test('123456.789123 + 654321.000321', 777777.789444, True, '777777.789444');
  Test('123456.789123 + (-0.789123)', 123456.0, True, '123456');

  Test('123.456 + 5', 128.456, True, '128.456');
  Test('123.456 + (-5)', 118.456, True, '118.456');
  Test('(-123.456) + (-5)', -128.456, True, '−128.456');
  Test('(-123.456) + 5', -118.456, True, '−118.456');

  Test('5.25 + 1/4', 5.5, True, '5.5');
  Test('5.25 + (-1/4)', 5.0, True, '5');
  Test('(-5.25) + (-1/4)', -5.5, True, '−5.5');
  Test('(-5.25) + 1/4', -5.0, True, '−5');
  Test('1/4 + 5.25', 5.5, True, '5.5');
  Test('(-1/4) + 5.25', 5.0, True, '5');
  Test('(-1/4) + (-5.25)', -5.5, True, '−5.5');
  Test('1/4 + (-5.25)', -5.0, True, '−5');
  Test('17584986538592057 + 0.5', 17584986538592057.5);
  Test('17584986538592057 + (-0.5)', 17584986538592056.5);

  Test('(2 + 3⋅i) + (7 + 2⋅i)', 9 + 5*ImaginaryUnit, True, '9 + 5⋅i');
  Test('2 + 3⋅i + 7 + 2⋅i', 9 + 5*ImaginaryUnit, True, '9 + 5⋅i');
  Test('(2 + 3⋅i) + (-7 - 2⋅i)', -5 + ImaginaryUnit, True, '−5 + i');
  Test('i + i', 2*ImaginaryUnit, True, '2⋅i');
  Test('i + (-i)', TASC(0.0), True, '0');
  Test('123.456 + (1 + 2.2⋅i)', 124.456 + 2.2*ImaginaryUnit, True, '124.456 + 2.2⋅i');
  Test('1/4 + i/5', 1/4 + ImaginaryUnit/5, True, '0.25 + 0.2⋅i');
  Test('(1 + 2.2⋅i) + 123.456', 124.456 + 2.2*ImaginaryUnit, True, '124.456 + 2.2⋅i');
  Test('i/5 + 1/4', 1/4 + ImaginaryUnit/5, True, '0.25 + 0.2⋅i');
  Test('i + 1/i', TASC(0.0), True, '0');

  Test('8305843009213693952 + 8305843009213693952', 16611686018427387904.0);
  Test('52E20 + 8E20', 60E20);

  Test('(-8305843009213693952) + (-8305843009213693952)', -16611686018427387904.0);
  Test('(-52E20) + (-8E20)', -60E20);

  Test('2E500 + 3E500', 5E500);
  Eps; Test('(-2E500) + 3E500', 1E500);
  Eps; Test('2E500 + (-3E500)', -1E500);

  Test('1 + 2 + 3 + 4 + 5', 15);               // + is a collapsing operator, so these tests are not unnecessary
  Eps; Test('1.1 + 2.2 + 3.3', 6.6);
  Test('1 + 2 + 0.5', 3.5);
  Test('1 + 2 + 1/2', Rat(7, 2));
  Test('1 + 2 + 1/2 + 0.1', 3.6);
  Test('1 + 2 + 1/2 + 0.1 + i', 3.6 + ImaginaryUnit);

  // - two vectors

  Test('❨1, 2, 3❩ + ❨10, 20, 30❩', [11, 22, 33]);
  Test('❨1, 2, 3❩ + ❨1, -2, -3❩', [2, 0, 0]);
  Test('❨1, 2, 3❩ + ❨-1, -2, -3❩', [0, 0, 0]);
  Test('❨1, 0❩ + ❨0.2, -0.1❩', [1.2, -0.1]);
  Test('❨7❩ + ❨2❩', [9]);
  Test('❨-7❩ + ❨7❩', [0]);
  Test('❨7/2, 0❩ + ❨3/2, 1❩', [5, 1]);
  Test('❨1, 0, 0❩ + ❨1, 1❩', failure, 'Cannot add two vectors of unequal dimension.');

  Test('❨1, 2, 3❩ + ❨i, 20, 30❩', [1 + ImaginaryUnit, 22, 33]);
  Test('❨i, 2, 3❩ + ❨i, 20, 30❩', [2*ImaginaryUnit, 22, 33]);
  Test('❨i, 2, 3❩ + ❨10, 20, 30❩', [10 + ImaginaryUnit, 22, 33]);

  Test('❨1, 2, 3❩ + ❨i, 20❩', failure, 'Cannot add two vectors of unequal dimension.');
  Test('❨i, 2, 3❩ + ❨i, 20❩', failure, 'Cannot add two vectors of unequal dimension.');
  Test('❨i, 2, 3❩ + ❨10, 20❩', failure, 'Cannot add two vectors of unequal dimension.');

  Test('❨3, 5, 8, 43, 8, 5, 2, 3, 4, 5❩ + ❨54, 2, 8, 7, 5, 6, 9, 4, 2, 5❩', [57, 7, 16, 50, 13, 11, 11, 7, 6, 10]);
  Test('❨3, 5, 8, 43, 8, 5, 2, 3, 4, 5❩ + ❨54, 2, 8, 7, 5, 6, 9, 4, 2, 5 + i❩', [57, 7, 16, 50, 13, 11, 11, 7, 6, 10 + ImaginaryUnit]);

  Test('❨1, 2, 3❩ + ❨10, 20, 30❩ + ❨100, 200, 300❩', [111, 222, 333]);

  // - two matrices

  Test('❨❨1, 2❩, ❨3, 4❩❩ + ❨❨10, 20❩, ❨30, 40❩❩', 2, [11, 22, 33, 44]);
  Test('❨❨1❩, ❨10❩, ❨100❩❩ + ❨❨2❩, ❨20❩, ❨32❩❩', 1, [3, 30, 132]);
  Test('❨❨5, 7, 9, 11❩❩ + ❨❨10, 20, 30, 20❩❩', 4, [15, 27, 39, 31]);
  Test('❨❨1, 2, 3, 4❩, ❨10, 20, 30, 40❩❩ + ❨❨100, 100, 200, 200❩, ❨1000, 2000, 3000, 1000❩❩', 4, [101, 102, 203, 204, 1010, 2020, 3030, 1040]);
  Test('❨❨6❩❩ + ❨❨-7❩❩', 1, [-1]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ + ❨❨2❩, ❨20❩, ❨32❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨1, 2❩, ❨3, 4❩❩ + ❨❨2❩, ❨20❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨1, 2, 3, 4❩, ❨10, 20, 30, 40❩❩ + ❨❨10, 20, 30, 20❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨10, 20, 30, 40❩❩ + ❨❨20❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨2❩, ❨20❩, ❨32❩❩ + ❨❨20❩❩', failure, 'Cannot add two matrices of different sizes.');

  Test('❨❨1, 2❩, ❨3, 4❩❩ + ❨❨10, 20❩, ❨30, 40⋅i❩❩', 2, [11, 22, 33, 4 + 40*ImaginaryUnit]);
  Test('❨❨-i, 2❩, ❨3, 4❩❩ + ❨❨10, 20❩, ❨30, 40❩❩', 2, [10-ImaginaryUnit, 22, 33, 44]);
  Test('❨❨-i, 2❩, ❨3, 4❩❩ + ❨❨10, i❩, ❨30, 40❩❩', 2, [10-ImaginaryUnit, 2+ImaginaryUnit, 33, 44]);
  Test('❨❨-i, 2❩, ❨3, 4⋅i❩❩ + ❨❨i, i❩, ❨30, 40❩❩', 2, [0, 2+ImaginaryUnit, 33, 40 + 4*ImaginaryUnit]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ + ❨❨i❩, ❨20❩, ❨32❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨1, 2❩, ❨3, 4❩❩ + ❨❨i❩, ❨20❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨1, 2, 3, 4❩, ❨10, 20, 30, 40❩❩ + ❨❨10, i, 30, 20❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨10, 20, 30, 40❩❩ + ❨❨i❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨2❩, ❨20❩, ❨32❩❩ + ❨❨i❩❩', failure, 'Cannot add two matrices of different sizes.');

  Test('❨❨1, 2❩, ❨i, 4❩❩ + ❨❨2❩, ❨20❩, ❨32❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨1, i❩, ❨3, 4❩❩ + ❨❨2❩, ❨20❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨1, i, 3, 4❩, ❨10, 20, 30, 40❩❩ + ❨❨10, 20, 30, 20❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨10, i, 30, 40❩❩ + ❨❨20❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨i❩, ❨20❩, ❨32❩❩ + ❨❨20❩❩', failure, 'Cannot add two matrices of different sizes.');

  Test('❨❨1, 2❩, ❨3, i❩❩ + ❨❨i❩, ❨20❩, ❨32❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨1, 2❩, ❨i, 4❩❩ + ❨❨2❩, ❨i❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨1, 2, i, 4❩, ❨10, 20, i, 40❩❩ + ❨❨10, i, 30, 20❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨10, 20, i, 40❩❩ + ❨❨i❩❩', failure, 'Cannot add two matrices of different sizes.');
  Test('❨❨i❩, ❨20❩, ❨32❩❩ + ❨❨i❩❩', failure, 'Cannot add two matrices of different sizes.');

  Test('❨❨1, 2❩, ❨3❩❩ + ❨❨2❩, ❨20❩❩', failure, 'Attempt to create a non-rectangular matrix');
  Test('❨❨1❩, ❨3, 4❩❩ + ❨❨2❩, ❨20❩❩', failure, 'Attempt to create a non-rectangular matrix');

  Test('❨❨1, 2❩, ❨3, 4❩❩ + ❨❨10, 20❩, ❨30, 40❩❩ + ❨❨100, 200❩, ❨300, 400❩❩', 2, [111, 222, 333, 444]);

  // - two strings

  Test('"alfa" + "beta"', 'alfabeta', True, 'alfabeta');
  Test('"alfa" + ""', 'alfa', True, 'alfa');
  Test('"" + "beta"', 'beta', True, 'beta');
  Test('"" + ""', '', True, '');
  Test('"Name: " + """" + "Sir" + """"', 'Name: "Sir"', True, 'Name: "Sir"');

  // - string + any non-string object

  Test('"alfa " + 12', 'alfa 12', True, 'alfa 12');
  Test('"alfa " + 1/2', 'alfa 0.5', True, 'alfa 0.5');
  Test('"alfa " + 1.2', 'alfa 1.2', True, 'alfa 1.2');
  Test('"alfa " + true', 'alfa true', True, 'alfa true');
  Test('"alfa " + false', 'alfa false', True, 'alfa false');
  Test('"alfa " + (1 + 2⋅i)', 'alfa 1 + 2⋅i', True, 'alfa 1 + 2⋅i');
  Test('"alfa " + 1 + 2⋅i', 'alfa 12⋅i', True, 'alfa 12⋅i');
  Test('"alfa " + ❨1, 2, 3❩', 'alfa (1, 2, 3)', True, 'alfa (1, 2, 3)');
  Test('"alfa " + ❨1, 2, i❩', 'alfa (1, 2, i)', True, 'alfa (1, 2, i)');
  Test('"alfa " + ❨❨1, 2❩, ❨3, 4❩❩', 'alfa ((1, 2), (3, 4))', True, 'alfa ((1, 2), (3, 4))');
  Test('"alfa " + ❨❨i, 2❩, ❨3, 4❩❩', 'alfa ((i, 2), (3, 4))', True, 'alfa ((i, 2), (3, 4))');
  Test('"alfa " + ''(1, 2, 3)', 'alfa (1, 2, 3)', True, 'alfa (1, 2, 3)');
  Test('"alfa " + struct("a": 42)', 'alfa (a: 42)', True, 'alfa (a: 42)');
  Test('"alfa " + {5}', 'alfa {5}', True, 'alfa {5}');
  Test('"alfa " + {}', 'alfa ∅', True, 'alfa ∅');
  Test('"alfa " + color("red")', 'alfa rgba(1.000, 0.000, 0.000, 1.000)', True, 'alfa rgba(1.000, 0.000, 0.000, 1.000)');
  Test('"alfa " + BinaryData("a")', 'alfa 61 00', True, 'alfa 61 00');
  Test('"alfa " + (.;)', 'alfa ', True, 'alfa ');

  // - vector + number

  Test('❨1, 2, 3❩ + 1', [2, 3, 4]);
  Test('❨1, 2, 3❩ + (-1)', [0, 1, 2]);
  Test('❨1, 2, 3❩ + 10', [11, 12, 13]);
  Test('❨1, 2, 3❩ + 10.5', [11.5, 12.5, 13.5]);
  Test('❨1, 2, 3❩ + 1/2', [1.5, 2.5, 3.5]);
  Test('❨1, 2, 3❩ + i', [1 + ImaginaryUnit, 2 + ImaginaryUnit, 3 + ImaginaryUnit]);
  Test('❨1, 2, 3❩ + (2 + 3⋅i)', [3 + 3*ImaginaryUnit, 4 + 3*ImaginaryUnit, 5 + 3*ImaginaryUnit]);

  Test('❨1, 2.1, 3❩ + 1', [2, 3.1, 4]);
  Test('❨1, 1/2, 3❩ + 1', [2, 1.5, 4]);
  Test('❨1, 1/2, 3❩ + 1/2', [1.5, 1.0, 3.5]);

  Test('❨1.1, 2.1, 3❩ + 1', [2.1, 3.1, 4]);
  Test('❨1.1, 1/2, 3❩ + 1', [2.1, 1.5, 4]);
  Test('❨1.1, 1/2, 3❩ + 1/2', [1.6, 1.0, 3.5]);

  Test('❨i, 2, 3❩ + 1', [ImaginaryUnit + 1, 3, 4]);
  Test('❨i, 2, 3❩ + (-1)', [ImaginaryUnit - 1, 1, 2]);
  Test('❨i, 2, 3❩ + 10', [ImaginaryUnit + 10, 12, 13]);
  Test('❨i, 2, 3❩ + 10.5', [ImaginaryUnit + 10.5, 12.5, 13.5]);
  Test('❨i, 2, 3❩ + 1/2', [ImaginaryUnit + 0.5, 2.5, 3.5]);
  Test('❨i, 2, 3❩ + i', [2*ImaginaryUnit, 2 + ImaginaryUnit, 3 + ImaginaryUnit]);
  Test('❨i, 2, 3❩ + (2 + 3⋅i)', [2 + 4*ImaginaryUnit, 4 + 3*ImaginaryUnit, 5 + 3*ImaginaryUnit]);

  // - matrix + number

  Test('❨❨1, 2❩, ❨3, 4❩❩ + 1', 2, [2, 3, 4, 5]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ + (-1)', 2, [0, 1, 2, 3]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ + 10', 2, [11, 12, 13, 14]);
  Test('❨❨1❩, ❨10❩, ❨100❩❩ + 1/2', 1, [1.5, 10.5, 100.5]);
  Test('❨❨1, 2, 3, 4❩, ❨10, 20, 30, 40❩❩ + 0.5', 4, [1.5, 2.5, 3.5, 4.5, 10.5, 20.5, 30.5, 40.5]);
  Test('❨❨10, 20, 30, 40❩❩ + i', 4, [10 + ImaginaryUnit, 20 + ImaginaryUnit, 30 + ImaginaryUnit, 40 + ImaginaryUnit]);

  Test('❨❨1.1, 2❩, ❨3, 4❩❩ + 1', 2, [2.1, 3, 4, 5]);
  Eps; Test('❨❨1.1, 2❩, ❨3, 4❩❩ + (-1)', 2, [0.1, 1, 2, 3]);
  Test('❨❨1.1, 2❩, ❨3, 4❩❩ + 10', 2, [11.1, 12, 13, 14]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ + 1', 2, [ImaginaryUnit + 1, 3, 4, 5]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ + (-1)', 2, [ImaginaryUnit - 1, 1, 2, 3]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ + 10', 2, [ImaginaryUnit + 10, 12, 13, 14]);
  Test('❨❨i❩, ❨10❩, ❨100❩❩ + 1/2', 1, [ImaginaryUnit + 0.5, 10.5, 100.5]);
  Test('❨❨i, 2, 3, 4❩, ❨10, 20, 30, 40❩❩ + 0.5', 4, [ImaginaryUnit + 0.5, 2.5, 3.5, 4.5, 10.5, 20.5, 30.5, 40.5]);
  Test('❨❨i, 20, 30, 40❩❩ + i', 4, [2*ImaginaryUnit, 20 + ImaginaryUnit, 30 + ImaginaryUnit, 40 + ImaginaryUnit]);

  // - two sounds: postponed

  // - other combinations (testing the operator's collapsing nature)

  Test('❨1, 2, 3❩ + ❨10, 20, 30❩ + 100', [111, 122, 133]);

  // - invalid operands

  Test('5 + "test"', failure, 'Invalid function arguments.');
  Test('❨1, 0, 0❩ + "test"', failure, 'Invalid function arguments.');
  Test('❨❨i, 2❩, ❨3, 4❩❩ + "test"', failure, 'Invalid function arguments.');
  Test('❨1, 0, 0❩ + true', failure, 'Invalid function arguments.');
  Test('❨❨i, 2❩, ❨3, 4❩❩ + date()', failure, 'Invalid function arguments.');
  Test('1 + ❨1, 0, 0❩', failure, 'Invalid function arguments.');
  Test('1 + ❨❨i, 2❩, ❨3, 4❩❩', failure, 'Invalid function arguments.');
  Test('false + "test"', failure, 'Invalid function arguments.');
  Test('false + 42', failure, 'Invalid function arguments.');
  Test('❨❨i, 2❩, ❨3, 4❩❩ + (0;)', failure, 'Invalid function arguments.');
  Test('color("red") + ''(1, 2, 3, 4, 5)', failure, 'Invalid function arguments.');
  Test('color("red") + false', failure, 'Invalid function arguments.');
  Test('color("red") + time()', failure, 'Invalid function arguments.');
  Test('color("red") + ❨❨i, 2❩, ❨3, 4❩❩', failure, 'Invalid function arguments.');
  Test('color("red") + "nice"', failure, 'Invalid function arguments.');
  Test('BinaryData("a") + ❨❨i, 2❩, ❨3, 4❩❩', failure, 'Invalid function arguments.');
  Test('BinaryData("a") + now()', failure, 'Invalid function arguments.');
  Test('(0;) + (0;)', failure, 'Invalid function arguments.');

  // Binary minus

  // - two numbers

  Test('10 - 7', 3, True, '3');
  Test('10 - 21', -11, True, '−11');
  Test('10 - 1/2', Rat(19, 2), True, '19/2');
  Test('10 - 0.5', 9.5, True, '9.5');
  Test('10 − 7', 3, True, '3');
  Test('10 − 21', -11, True, '−11');
  Test('10 − 1/2', Rat(19, 2), True, '19/2');
  Test('10 − 0.5', 9.5, True, '9.5');

  Test('10 - (-7)', 17, True, '17');
  Test('10 - (-21)', 31, True, '31');
  Test('10 - (-1/2)', Rat(21, 2), True, '21/2');
  Test('10 − (−1/2)', Rat(21, 2), True, '21/2');
  Test('10 - (-0.5)', 10.5, True, '10.5');
  Test('10 − (−0.5)', 10.5, True, '10.5');

  Test('17584986538592057 - 57', 17584986538592000, True, '17584986538592000');
  Test('17584986538592057 - (-3)', 17584986538592060, True, '17584986538592060');
  Test('17584986538592057 - 07584986538592057', 10000000000000000, True, '10000000000000000');

  Test('10 - i', 10 - ImaginaryUnit, True, '10 − i');
  Test('10 - (5.4 + 2.3 ⋅ i)', 4.6 - 2.3*ImaginaryUnit, True, '4.6 − 2.3⋅i');

  Test('7/3 - 1', Rat(4, 3), True, '4/3');
  Test('7/3 - 1/3', Rat(2, 1), True, '2');
  Test('7/3 - 2/3', Rat(5, 3), True, '5/3');
  Test('7/3 - 1/4', Rat(25, 12), True, '25/12');
  Test('7/2 - 0.5', 3.0, True, '3');
  Test('1/2 - i', 0.5 - ImaginaryUnit, True, '0.5 − i');
  Test('1/2 − i', 0.5 - ImaginaryUnit, True, '0.5 − i');
  Test('1/2 - (5.4 + 2.3 ⋅ i)', -4.9 - 2.3*ImaginaryUnit, True, '−4.9 − 2.3⋅i');

  Test('10.61 - 1', 9.61, True, '9.61');
  Test('10.5 - 1/2', 10.0, True, '10');
  Test('10.6 - 1/2', 10.1, True, '10.1');
  Eps; Test('10.61 - 1.3', 9.31, True, '9.31');
  Test('0.5 - i', 0.5 - ImaginaryUnit, True, '0.5 − i');
  Test('0.5 - (5.4 + 2.3 ⋅ i)', -4.9 - 2.3*ImaginaryUnit, True, '−4.9 − 2.3⋅i');

  Test('(5.4 + 2.3 ⋅ i) - (3.2 - 1.1 ⋅ i)', 2.2 + 3.4*ImaginaryUnit, True, '2.2 + 3.4⋅i');
  Test('(5.4 + 2.3 ⋅ i) − (3.2 − 1.1 ⋅ i)', 2.2 + 3.4*ImaginaryUnit, True, '2.2 + 3.4⋅i');

  Test('0 − 0', 0, True, '0');
  Test('0.0 − 0', 0.0, True, '0');
  Test('0 − 0.0', 0.0, True, '0');
  Test('0.0 − 0.0', 0.0, True, '0');
  Test('3.2 − 0.0', 3.2, True, '3.2');
  Test('0 − 1', -1, True, '−1');
  Test('0 − 1.2', -1.2, True, '−1.2');
  Test('0 − 1/2', Rat(-1, 2), True, '−1/2');

  Test('(-8305843009213693952) - 8305843009213693952', -16611686018427387904.0);
  Test('50E100 - 27E100', 23E100);
  Test('50E100 - (-27E100)', 77E100);
  Test('(-50E100) - 27E100', -77E100);

  // - two vectors

  Test('❨1, 1, 0❩ − ❨0, 1, 2❩', [1, 0, -2]);
  Test('❨1, 1, 0❩ − ❨0, −1, 2❩', [1, 2, -2]);
  Test('❨1, 1, 0❩ − ❨0, −1❩', failure, 'Cannot subtract two vectors of unequal dimension.');

  Test('❨i, 1, 0❩ − ❨0, 1, 2❩', [ImaginaryUnit, 0, -2]);
  Test('❨i, 1, 0❩ − ❨0, −1, 2❩', [ImaginaryUnit, 2, -2]);
  Test('❨i, 1, 0❩ − ❨0, −1❩', failure, 'Cannot subtract two vectors of unequal dimension.');

  Test('❨1, 1, 0❩ − ❨i, 1, 2❩', [1 - ImaginaryUnit, 0, -2]);
  Test('❨1, 1, 0❩ − ❨i, −1, 2❩', [1 - ImaginaryUnit, 2, -2]);
  Test('❨1, 1, 0❩ − ❨i, −1❩', failure, 'Cannot subtract two vectors of unequal dimension.');

  Test('❨i, 1, i❩ − ❨i, 1, 2❩', [0, 0, ImaginaryUnit-2]);
  Test('❨i, 1, i❩ − ❨i, −1, 2❩', [0, 2, ImaginaryUnit-2]);
  Test('❨i, 1, i❩ − ❨i, −1❩', failure, 'Cannot subtract two vectors of unequal dimension.');

  Test('❨2, 5, 3.5, 1, −2, 4❩ − ❨−3.4, 5, 6, 7, 0, 0❩', [5.4, 0, -2.5, -6, -2, 4]);
  Test('❨2, 5, 3.5, 1, −2, 4❩ − ❨−3.4, 5, 6, 7, 0❩', failure, 'Cannot subtract two vectors of unequal dimension.');
  Test('❨2, 5, 3.5, 1, −2, 4❩ − ❨−3.4, 5, 6, 7, 0, 0, 0❩', failure, 'Cannot subtract two vectors of unequal dimension.');

  Test('❨1, 0❩ − ❨1, 1❩', [0, -1]);
  Test('❨1, 0❩ − ❨1/2, 0.5❩', [0.5, -0.5]);
  Test('❨1, 0❩ − ❨1, i❩', [0, -ImaginaryUnit]);

  Test('❨1❩ − ❨3❩', [-2]);

  // - two matrices

  Test('❨❨1, 2❩, ❨3, 4❩❩ − ❨❨3, 1❩, ❨2, 0❩❩', 2, [-2, 1, 1, 4]);
  Test('❨❨1, 2❩, ❨−3, 4❩❩ − ❨❨3, 1❩, ❨2, 0❩❩', 2, [-2, 1, -5, 4]);
  Test('❨❨1❩, ❨2❩, ❨3❩, ❨4❩❩ − ❨❨0❩, ❨1❩, ❨5❩, ❨−2❩❩', 1, [1, 1, -2, 6]);
  Test('❨❨5, 4, 3, 7❩❩ − ❨❨1, 2, 3, 6❩❩', 4, [4, 2, 0, 1]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ − ❨❨3, 1❩, ❨2, 0❩❩', 2, [ImaginaryUnit - 3, 1, 1, 4]);
  Test('❨❨i, 2❩, ❨−3, 4❩❩ − ❨❨3, 1❩, ❨2, 0❩❩', 2, [ImaginaryUnit - 3, 1, -5, 4]);
  Test('❨❨i❩, ❨2❩, ❨3❩, ❨4❩❩ − ❨❨0❩, ❨1❩, ❨5❩, ❨−2❩❩', 1, [ImaginaryUnit, 1, -2, 6]);
  Test('❨❨5⋅i, 4, 3, 7❩❩ − ❨❨1, 2, 3, 6❩❩', 4, [5*ImaginaryUnit - 1, 2, 0, 1]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ − ❨❨3⋅i, 1❩, ❨2, 0❩❩', 2, [1-3*ImaginaryUnit, 1, 1, 4]);
  Test('❨❨1, 2❩, ❨−3, 4❩❩ − ❨❨3⋅i, 1❩, ❨2, 0❩❩', 2, [1-3*ImaginaryUnit, 1, -5, 4]);
  Test('❨❨1❩, ❨2❩, ❨3❩, ❨4❩❩ − ❨❨i❩, ❨1❩, ❨5❩, ❨−2❩❩', 1, [1-ImaginaryUnit, 1, -2, 6]);
  Test('❨❨5, 4, 3, 7❩❩ − ❨❨i, 2, 3, 6❩❩', 4, [5-ImaginaryUnit, 2, 0, 1]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ − ❨❨3⋅i, 1❩, ❨2, 0❩❩', 2, [-2*ImaginaryUnit, 1, 1, 4]);
  Test('❨❨i, 2❩, ❨−3, 4❩❩ − ❨❨3⋅i, 1❩, ❨2, 0❩❩', 2, [-2*ImaginaryUnit, 1, -5, 4]);
  Test('❨❨i❩, ❨2❩, ❨3❩, ❨4❩❩ − ❨❨i❩, ❨1❩, ❨5❩, ❨−2❩❩', 1, [0*ImaginaryUnit, 1, -2, 6]);
  Test('❨❨5⋅i, 4, 3, 7❩❩ − ❨❨i, 2, 3, 6❩❩', 4, [4*ImaginaryUnit, 2, 0, 1]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ − ❨❨1❩, ❨2❩, ❨3❩, ❨4❩❩', failure, 'Cannot subtract two matrices of different sizes.');
  Test('❨❨i, 2❩, ❨3, 4❩❩ − ❨❨1❩, ❨2❩, ❨3❩, ❨4❩❩', failure, 'Cannot subtract two matrices of different sizes.');
  Test('❨❨1, 2❩, ❨3, 4❩❩ − ❨❨i❩, ❨2❩, ❨3❩, ❨4❩❩', failure, 'Cannot subtract two matrices of different sizes.');
  Test('❨❨i, 2❩, ❨3, 4❩❩ − ❨❨i❩, ❨2❩, ❨3❩, ❨4❩❩', failure, 'Cannot subtract two matrices of different sizes.');

  Test('❨❨5, 4, 3, 7❩❩ − ❨❨1, 2, 3❩❩', failure, 'Cannot subtract two matrices of different sizes.');
  Test('❨❨i, 4, 3, 7❩❩ − ❨❨1, 2, 3❩❩', failure, 'Cannot subtract two matrices of different sizes.');
  Test('❨❨5, 4, 3, 7❩❩ − ❨❨i, 2, 3❩❩', failure, 'Cannot subtract two matrices of different sizes.');
  Test('❨❨i, 4, 3, 7❩❩ − ❨❨i, 2, 3❩❩', failure, 'Cannot subtract two matrices of different sizes.');

  // - vector minus number

  Test('❨1, 2, 3❩ − 1', [0, 1, 2]);
  Test('❨1, 2, 3❩ − (−1)', [2, 3, 4]);
  Test('❨1, 2, 3❩ − 1/2', [0.5, 1.5, 2.5]);
  Test('❨1, 2, 3❩ − (−1/2)', [1.5, 2.5, 3.5]);
  Test('❨1, 2, 3❩ − 0.5', [0.5, 1.5, 2.5]);
  Test('❨1, 2, 3❩ − (−0.5)', [1.5, 2.5, 3.5]);
  Test('❨1, 2, 3❩ − i', [1-ImaginaryUnit, 2-ImaginaryUnit, 3-ImaginaryUnit]);
  Test('❨1, 2, 3❩ − (−i)', [1+ImaginaryUnit, 2+ImaginaryUnit, 3+ImaginaryUnit]);
  Test('❨1, 2, 3❩ − (2+i)', [-1-ImaginaryUnit, -ImaginaryUnit, 1-ImaginaryUnit]);

  Test('❨i, 2, 3❩ − 1', [ImaginaryUnit - 1, 1, 2]);
  Test('❨i, 2, 3❩ − (−1)', [ImaginaryUnit + 1, 3, 4]);
  Test('❨i, 2, 3❩ − 1/2', [ImaginaryUnit - 0.5, 1.5, 2.5]);
  Test('❨i, 2, 3❩ − (−1/2)', [ImaginaryUnit + 0.5, 2.5, 3.5]);
  Test('❨i, 2, 3❩ − 0.5', [ImaginaryUnit - 0.5, 1.5, 2.5]);
  Test('❨i, 2, 3❩ − (−0.5)', [ImaginaryUnit + 0.5, 2.5, 3.5]);
  Test('❨i, 2, 3❩ − i', [0, 2-ImaginaryUnit, 3-ImaginaryUnit]);
  Test('❨i, 2, 3❩ − (−i)', [2*ImaginaryUnit, 2+ImaginaryUnit, 3+ImaginaryUnit]);
  Test('❨i, 2, 3❩ − (2+i)', [-2, -ImaginaryUnit, 1-ImaginaryUnit]);

  Test('❨1, 2, 3, 4, 5❩ − 1', [0, 1, 2, 3, 4]);
  Test('❨1❩ − 10', [-9]);

  // - matrix minus number

  Test('❨❨1, 2❩, ❨3, 4❩❩ − 1', 2, [0, 1, 2, 3]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ − (−1)', 2, [2, 3, 4, 5]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ − 1/2', 2, [0.5, 1.5, 2.5, 3.5]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ − (−1/2)', 2, [1.5, 2.5, 3.5, 4.5]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ − 0.5', 2, [0.5, 1.5, 2.5, 3.5]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ − (−0.5)', 2, [1.5, 2.5, 3.5, 4.5]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ − i', 2, [1 - ImaginaryUnit, 2 - ImaginaryUnit, 3 - ImaginaryUnit, 4 - ImaginaryUnit]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ − (−i)', 2, [1 + ImaginaryUnit, 2 + ImaginaryUnit, 3 + ImaginaryUnit, 4 + ImaginaryUnit]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ − 1', 2, [ImaginaryUnit - 1, 1, 2, 3]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ − (−1)', 2, [ImaginaryUnit + 1, 3, 4, 5]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ − 1/2', 2, [ImaginaryUnit - 0.5, 1.5, 2.5, 3.5]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ − (−1/2)', 2, [ImaginaryUnit + 0.5, 2.5, 3.5, 4.5]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ − 0.5', 2, [ImaginaryUnit - 0.5, 1.5, 2.5, 3.5]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ − (−0.5)', 2, [ImaginaryUnit + 0.5, 2.5, 3.5, 4.5]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ − i', 2, [0, 2 - ImaginaryUnit, 3 - ImaginaryUnit, 4 - ImaginaryUnit]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ − (−i)', 2, [2*ImaginaryUnit, 2 + ImaginaryUnit, 3 + ImaginaryUnit, 4 + ImaginaryUnit]);

  Test('❨❨5, 4, 3, 7❩❩ − 2', 4, [3, 2, 1, 5]);
  Test('❨❨5⋅i, 4, 3, 7❩❩ − 2', 4, [-2 + 5*ImaginaryUnit, 2, 1, 5]);
  Test('❨❨5⋅i, 4, 3, 7❩❩ − 2⋅i', 4, [3*ImaginaryUnit, 4 - 2*ImaginaryUnit, 3 - 2*ImaginaryUnit, 7 - 2*ImaginaryUnit]);
  Test('❨❨5, 4, 3, 7❩❩ − 2⋅i', 4, [5 - 2*ImaginaryUnit, 4 - 2*ImaginaryUnit, 3 - 2*ImaginaryUnit, 7 - 2*ImaginaryUnit]);

  Test('❨❨123❩❩ - 100', 1, [23]);

  // - two sounds: postponed

  // - invalid operands

  Test('"alfa" - "beta"', failure, 'Invalid function arguments.');
  Test('"alfa" - 1', failure, 'Invalid function arguments.');
  Test('5 − "test"', failure, 'Invalid function arguments.');
  Test('❨1, 0, 0❩ − "test"', failure, 'Invalid function arguments.');
  Test('❨❨i, 2❩, ❨3, 4❩❩ − "test"', failure, 'Invalid function arguments.');
  Test('❨1, 0, 0❩ − true', failure, 'Invalid function arguments.');
  Test('❨❨i, 2❩, ❨3, 4❩❩ − date()', failure, 'Invalid function arguments.');
  Test('1 − ❨1, 0, 0❩', failure, 'Invalid function arguments.');
  Test('1 − ❨❨i, 2❩, ❨3, 4❩❩', failure, 'Invalid function arguments.');
  Test('false − "test"', failure, 'Invalid function arguments.');
  Test('false − 42', failure, 'Invalid function arguments.');
  Test('❨❨i, 2❩, ❨3, 4❩❩ − (0;)', failure, 'Invalid function arguments.');
  Test('color("red") − ''(1, 2, 3, 4, 5)', failure, 'Invalid function arguments.');
  Test('color("red") − false', failure, 'Invalid function arguments.');
  Test('color("red") − time()', failure, 'Invalid function arguments.');
  Test('color("red") − ❨❨i, 2❩, ❨3, 4❩❩', failure, 'Invalid function arguments.');
  Test('color("red") − "nice"', failure, 'Invalid function arguments.');
  Test('BinaryData("a") − ❨❨i, 2❩, ❨3, 4❩❩', failure, 'Invalid function arguments.');
  Test('BinaryData("a") − now()', failure, 'Invalid function arguments.');
  Test('(0;) − (0;)', failure, 'Invalid function arguments.');

  // Multiplication (cdot)

  // - two numbers

  Test('7⋅8', 56, True, '56');
  Test('(-7)⋅8', -56, True, '−56');
  Test('7⋅(-8)', -56, True, '−56');
  Test('(-7)⋅(-8)', 56, True, '56');

  Test('7.0⋅8', 56.0, True, '56');
  Test('(-7.0)⋅8', -56.0, True, '−56');
  Test('7.0⋅(-8)', -56.0, True, '−56');
  Test('(-7.0)⋅(-8)', 56.0, True, '56');

  Test('7⋅8.0', 56.0, True, '56');
  Test('(-7)⋅8.0', -56.0, True, '−56');
  Test('7⋅(-8.0)', -56.0, True, '−56');
  Test('(-7)⋅(-8.0)', 56.0, True, '56');

  Test('7⋅(1/2)', Rat(7, 2), True, '7/2');
  Test('7⋅(-1/2)', Rat(-7, 2), True, '−7/2');
  Test('(-7)⋅(1/2)', Rat(-7, 2), True, '−7/2');
  Test('(-7)⋅(-1/2)', Rat(7, 2), True, '7/2');

  Test('7⋅0.5', 3.5, True, '3.5');
  Test('7⋅(-0.5)', -3.5, True, '−3.5');
  Test('(-7)⋅0.5', -3.5, True, '−3.5');
  Test('(-7)⋅(-0.5)', 3.5, True, '3.5');

  Test('7⋅i', 7*ImaginaryUnit, True, '7⋅i');
  Test('7⋅(-i)', -7*ImaginaryUnit, True, '−7⋅i');
  Test('7 ⋅ (2 - 3⋅i)', 14 - 21*ImaginaryUnit, True, '14 − 21⋅i');

  Test('(1/2) ⋅ 5', Rat(5, 2), True, '5/2');
  Test('(1/2) ⋅ (-5)', Rat(-5, 2), True, '−5/2');
  Test('(-1/2) ⋅ 5', Rat(-5, 2), True, '−5/2');
  Test('(-1/2) ⋅ (-5)', Rat(5, 2), True, '5/2');

  Test('(3/5) ⋅ (2/7)', Rat(6, 35), True, '6/35');
  Test('(3/5) ⋅ (-2/7)', Rat(-6, 35), True, '−6/35');
  Test('(-3/5) ⋅ (2/7)', Rat(-6, 35), True, '−6/35');
  Test('(-3/5) ⋅ (-2/7)', Rat(6, 35), True, '6/35');

  Test('(3/7) ⋅ (7/3)', Rat(1, 1), True, '1');
  Test('(3/7) ⋅ (-7/3)', Rat(-1, 1), True, '−1');
  Test('(-3/7) ⋅ (7/3)', Rat(-1, 1), True, '−1');
  Test('(-3/7) ⋅ (-7/3)', Rat(1, 1), True, '1');

  Test('(7/12) ⋅ (4/3)', Rat(7, 9), True, '7/9');
  Test('(-7/12) ⋅ (4/3)', Rat(-7, 9), True, '−7/9');
  Test('(7/12) ⋅ (-4/3)', Rat(-7, 9), True, '−7/9');
  Test('(-7/12) ⋅ (-4/3)', Rat(7, 9), True, '7/9');

  Eps; Test('(3/5) ⋅ 0.1', 0.06, True, '0.06');
  Test('(3/5) ⋅ i', 0.6*ImaginaryUnit, True, '0.6⋅i');
  Test('(3/5) ⋅ (-i)', -0.6*ImaginaryUnit, True, '−0.6⋅i');
  Eps; Test('(3/5) ⋅ (3.2 + 3.6⋅i)', 1.92 + 2.16*ImaginaryUnit, True, '1.92 + 2.16⋅i');

  Test('0.6 ⋅ 5', 3.0, True, '3');
  Test('0.6 ⋅ 1/2', 0.3, True, '0.3');
  Test('0.6 ⋅ 0.5', 0.3, True, '0.3');
  Test('0.6 ⋅ i', 0.6*ImaginaryUnit, True, '0.6⋅i');

  Test('i ⋅ 5', 5*ImaginaryUnit, True, '5⋅i');
  Test('(-i) ⋅ 5', -5*ImaginaryUnit, True, '−5⋅i');
  Test('i ⋅ (-5)', -5*ImaginaryUnit, True, '−5⋅i');
  Test('(-i) ⋅ (-5)', 5*ImaginaryUnit, True, '5⋅i');

  Test('i ⋅ (1/2)', ImaginaryUnit/2, True, '0.5⋅i');
  Test('i ⋅ 0.5', ImaginaryUnit/2, True, '0.5⋅i');
  Test('i ⋅ i', TASC(-1.0), True, '−1');

  Eps; Test('(2.4 + 0.5⋅i)⋅(-2.4 + 9.2⋅i)', -10.36 + 20.88*ImaginaryUnit, True, '−10.36 + 20.88⋅i');

  Test('8305843009213693952 ⋅ 2', 16611686018427387904.0);
  Test('8305843009213693952 ⋅ 10000', 8.305843009213693952E22);
  Test('2⋅25E50', 50E50);
  Test('36E100 ⋅ 2E100', 72E200);

  Test('(-8305843009213693952) ⋅ 2', -16611686018427387904.0);
  Test('(-8305843009213693952) ⋅ 10000', -8.305843009213693952E22);
  Test('2⋅(-25E50)', -50E50);
  Test('(-36E100) ⋅ 2E100', -72E200);

  Test('1 ⋅ 2 ⋅ 3 ⋅ 4 ⋅ 5', 120);
  Test('1 ⋅ 2 ⋅ 3 ⋅ 4 ⋅ 5.0', 120.0);
  Test('1 ⋅ 2 ⋅ 3 ⋅ 4 ⋅ (5/2)', Rat(60, 1));

  // - vector inner (dot, scalar) product

  Test('❨1, 2, 3❩ ⋅ ❨5, 2, −1❩', 6.0, True, '6');
  Test('❨1, 0, 0❩ ⋅ ❨0, 1, 0❩', 0.0, True, '0');
  Test('❨1, 0, 0❩ ⋅ ❨1, 0, 0❩', 1.0, True, '1');

  Test('❨6, 4, 2❩ ⋅ ❨i, 3, 1❩', 14 - 6*ImaginaryUnit, True, '14 − 6⋅i');
  Test('❨1, 0, 0❩ ⋅ ❨2 + i, 4, 6❩', 2 - ImaginaryUnit, True, '2 − i');

  Test('❨2⋅i, 5, 3❩ ⋅ ❨1, 2, 1❩', 13 + 2*ImaginaryUnit, True, '13 + 2⋅i');
  Test('❨5+7⋅i, 2, 3❩ ⋅ ❨1, 0, 0❩', 5 + 7*ImaginaryUnit, True, '5 + 7⋅i');

  Test('❨3 + i, −2 − 5⋅i, 4 + 2⋅i❩ ⋅ ❨−5, −2 − 7⋅i, 3⋅i❩', 30 - 21*ImaginaryUnit, True, '30 − 21⋅i');
  Test('❨1, i, −1❩ ⋅ ❨1, i, −1❩', TASC(3.0), True, '3');

  Test('❨1, 0❩ ⋅ ❨0, 1❩', 0.0, True, '0');
  Test('❨1.0, 0❩ ⋅ ❨5.2, 1.3❩', 5.2, True, '5.2');

  Test('❨1, 2, 3❩ ⋅ ❨5, 2, −1❩ ⋅ ❨4, 5, 2❩', [24, 30, 12]);
  Test('❨1, 2, 3❩ ⋅ ❨5, 2, −1❩ ⋅ ❨4, 5, 2❩ ⋅ ❨2, 1, 7❩', 48.0 + 30 + 84);

  Test('❨1, 0❩ ⋅ ❨0, 2, 3, 4❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');
  Test('❨1, 0❩ ⋅ ❨0, 2, 3❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');
  Test('❨1, 0❩ ⋅ ❨0, 2❩', 0.0, True, '0');

  // - matrix multiplication

  Test('❨❨2, 0❩, ❨0, 3❩❩ ⋅ ❨❨1, 2❩, ❨3, 4❩❩', 2, [2, 4, 9, 12]);
  Test('❨❨4, 1, 2❩, ❨3, 5, 1❩, ❨−3, 5, 1❩❩ ⋅ ❨❨6, −4, 2❩, ❨−3, 7, 5❩, ❨0, 2, 1❩❩', 3, [21, -5, 15, 3, 25, 32, -33, 49, 20]);
  Test('IdentityMatrix(3) ⋅ ❨❨6, −4, 2❩, ❨−3, 7, 5❩, ❨0, 2, 1❩❩', 3, [6, -4, 2, -3, 7, 5, 0, 2, 1]);
  Test('❨❨4, 1, 2❩, ❨3, 5, 1❩, ❨−3, 5, 1❩❩ ⋅ IdentityMatrix(3)', 3, [4, 1, 2, 3, 5, 1, -3, 5, 1]);

  Test('❨❨1, 2❩, ❨5, −2❩, ❨5, 2❩❩ ⋅ ❨❨5, 6, 4, 2, 1❩, ❨4, −2, 3, 4, 6❩❩', 5, [13, 2, 10, 10, 13, 17, 34, 14, 2, -7, 33, 26, 26, 18, 17]);
  Test('❨❨1, 5, 4, 2❩, ❨−5, 2, 3, 4❩❩⋅❨❨1, 2❩, ❨−2, 1❩, ❨0, 3❩, ❨2, −3❩❩', 2, [-5, 13, -1, -11]);
  Test('❨❨4, 5, 1, 2, 3❩❩ ⋅ ❨❨2❩, ❨7❩, ❨−3❩, ❨4❩, ❨6❩❩', 1, [66]);
  Test('❨❨2❩, ❨7❩, ❨−3❩, ❨4❩, ❨6❩❩ ⋅ ❨❨4, 5, 1❩❩', 3, [8, 10, 2, 28, 35, 7, -12, -15, -3, 16, 20, 4, 24, 30, 6]);
  Test('❨❨7❩❩ ⋅ ❨❨9❩❩', 1, [63]);

  Test('❨❨3, i❩, ❨1, 2❩❩ ⋅ ❨❨4, 2❩, ❨1, 2❩❩', 2, [12 + ImaginaryUnit, 6 + 2*ImaginaryUnit, 6, 6]);
  Test('❨❨3, 2❩, ❨2, 1❩❩ ⋅ ❨❨i, i❩, ❨1, 2❩❩', 2, [2 + 3*ImaginaryUnit, 4 + 3*ImaginaryUnit, 1 + 2*ImaginaryUnit, 2 + 2*ImaginaryUnit]);

  Test('❨❨3, 2 + i, 2❩, ❨−2 − i, 1 + 2⋅i, 6❩, ❨0, i, 2❩❩ ⋅ ❨❨2, 3, 1❩, ❨0, 3, i❩, ❨1 + 2⋅i, 4, 6❩❩', 3, [8 + 4*ImaginaryUnit, 23 + 3*ImaginaryUnit, 14 + 2*ImaginaryUnit, 2 + 10*ImaginaryUnit, 21 + 3*ImaginaryUnit, 32, 2 + 4*ImaginaryUnit, 8 + 3*ImaginaryUnit, 11]);

  Test('❨❨3, 2❩, ❨3, −2❩❩ ⋅ ❨❨3, 2❩, ❨1, 2❩, ❨3, 2❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');
  Test('❨❨3, 2❩, ❨3, −2❩❩ ⋅ ❨❨3, 2❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');

  Test('❨❨3, i❩, ❨3, −2❩❩ ⋅ ❨❨3, 2❩, ❨1, 2❩, ❨3, 2❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');
  Test('❨❨3, i❩, ❨3, −2❩❩ ⋅ ❨❨3, 2❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');

  Test('❨❨3, 2❩, ❨3, −2❩❩ ⋅ ❨❨i, 2❩, ❨1, 2❩, ❨3, 2❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');
  Test('❨❨3, 2❩, ❨3, −2❩❩ ⋅ ❨❨i, 2❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');

  Test('❨❨3, i❩, ❨3, −2❩❩ ⋅ ❨❨3, i❩, ❨1, 2❩, ❨3, 2❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');
  Test('❨❨3, 2❩, ❨3, −i❩❩ ⋅ ❨❨3, i❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');

  Test('❨❨4, 5, 1, 2, 3❩❩ ⋅ ❨❨2❩, ❨7❩, ❨−3❩, ❨4❩, ❨6❩, ❨6❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');
  Test('❨❨4, 5, 1, 2, 3❩❩ ⋅ ❨❨2❩, ❨7❩, ❨−3❩, ❨4❩❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');

  Test('❨❨2, 0❩, ❨0, 3❩❩ ⋅ ❨❨1, 2❩, ❨3, 4❩❩ ⋅ IdentityMatrix(2)', 2, [2, 4, 9, 12]);

  // - matrix times vector (linear transformation)

  Test('IdentityMatrix(2) ⋅ ❨3, 2❩', [3, 2]);
  Test('IdentityMatrix(2) ⋅ ❨1.2, −2.3❩', [1.2, -2.3]);
  Test('❨❨2, 0❩, ❨0, 3❩❩ ⋅ ❨5, 7❩', [10, 21]);
  Test('❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 0❩❩ ⋅ ❨4, 5, 2❩', [4, 5, 0]);
  Test('❨❨cos(π/2), −sin(π/2)❩, ❨sin(π/2), cos(π/2)❩❩ ⋅ ❨1, 1❩', [-1, 1]);
  Test('❨❨1, 1❩, ❨0, 1❩❩ ⋅ ❨2, 3❩', [5, 3]);
  Test('❨❨1, 1❩, ❨0, 1❩, ❨0, -1❩❩ ⋅ ❨2, 3❩', [5, 3, -3]);

  Test('❨❨1, 1❩, ❨0, 1❩❩ ⋅ ❨2, 3, 4❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');
  Test('❨❨1, 1❩, ❨0, 1❩, ❨0, 1❩❩ ⋅ ❨2, 3, 4❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');
  Test('❨❨1, 1❩, ❨0, 1❩❩ ⋅ ❨2❩', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');

  // - vector times number or number times vector

  Test('❨1, 2, −3❩ ⋅ 2', [2, 4, -6]);
  Test('❨1, 2, −3❩ ⋅ (1/2)', [0.5, 1, -1.5]);
  Test('❨1, 2, −3❩ ⋅ 0.5', [0.5, 1, -1.5]);
  Test('❨1, 2, −3❩ ⋅ i', [ImaginaryUnit, 2*ImaginaryUnit, -3*ImaginaryUnit]);
  Test('❨i, 2, −3❩ ⋅ 2', [2*ImaginaryUnit, 4, -6]);
  Test('❨i, 2, −3❩ ⋅ (1/2)', [0.5*ImaginaryUnit, 1, -1.5]);
  Test('❨i, 2, −3❩ ⋅ 0.5', [0.5*ImaginaryUnit, 1, -1.5]);
  Test('❨i, 2, −3❩ ⋅ i', [-1, 2*ImaginaryUnit, -3*ImaginaryUnit]);

  Test('2 ⋅ ❨1, 2, −3❩', [2, 4, -6]);
  Test('(1/2) ⋅ ❨1, 2, −3❩', [0.5, 1, -1.5]);
  Test('0.5 ⋅ ❨1, 2, −3❩', [0.5, 1, -1.5]);
  Test('i ⋅ ❨1, 2, −3❩', [ImaginaryUnit, 2*ImaginaryUnit, -3*ImaginaryUnit]);
  Test('2 ⋅ ❨i, 2, −3❩', [2*ImaginaryUnit, 4, -6]);
  Test('(1/2) ⋅ ❨i, 2, −3❩', [0.5*ImaginaryUnit, 1, -1.5]);
  Test('0.5 ⋅ ❨i, 2, −3❩', [0.5*ImaginaryUnit, 1, -1.5]);
  Test('i ⋅ ❨i, 2, −3❩', [-1, 2*ImaginaryUnit, -3*ImaginaryUnit]);

  Test('❨1.1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ ⋅ 10', [11, 20, 30, 40, 50, 60, 70, 80, 90, 100]);
  Test('❨1.1, 2, 3, 4, 5, 6, 7, 8 + 7⋅i, 9, 10❩ ⋅ 10', [11, 20, 30, 40, 50, 60, 70, 80 + 70*ImaginaryUnit, 90, 100]);
  Test('10⋅❨1.1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩', [11, 20, 30, 40, 50, 60, 70, 80, 90, 100]);
  Test('10⋅❨1.1, 2, 3, 4, 5, 6, 7, 8 + 7⋅i, 9, 10❩', [11, 20, 30, 40, 50, 60, 70, 80 + 70*ImaginaryUnit, 90, 100]);

  Test('❨1, 2, −3❩ ⋅ 2 ⋅ 10', [20, 40, -60]);

  // - matrix times number or number times matrix

  Test('❨❨1, 2❩, ❨3, 4❩❩ ⋅ 10', 2, [10, 20, 30, 40]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ ⋅ (1/2)', 2, [0.5, 1, 1.5, 2]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ ⋅ 0.5', 2, [0.5, 1, 1.5, 2]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ ⋅ i', 2, [ImaginaryUnit, 2*ImaginaryUnit, 3*ImaginaryUnit, 4*ImaginaryUnit]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ ⋅ 10', 2, [10*ImaginaryUnit, 20, 30, 40]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ ⋅ (1/2)', 2, [0.5*ImaginaryUnit, 1, 1.5, 2]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ ⋅ 0.5', 2, [0.5*ImaginaryUnit, 1, 1.5, 2]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ ⋅ i', 2, [-1, 2*ImaginaryUnit, 3*ImaginaryUnit, 4*ImaginaryUnit]);

  Test('10 ⋅ ❨❨1, 2❩, ❨3, 4❩❩', 2, [10, 20, 30, 40]);
  Test('(1/2) ⋅ ❨❨1, 2❩, ❨3, 4❩❩', 2, [0.5, 1, 1.5, 2]);
  Test('0.5 ⋅ ❨❨1, 2❩, ❨3, 4❩❩', 2, [0.5, 1, 1.5, 2]);
  Test('i ⋅ ❨❨1, 2❩, ❨3, 4❩❩', 2, [ImaginaryUnit, 2*ImaginaryUnit, 3*ImaginaryUnit, 4*ImaginaryUnit]);
  Test('10 ⋅ ❨❨i, 2❩, ❨3, 4❩❩', 2, [10*ImaginaryUnit, 20, 30, 40]);
  Test('(1/2) ⋅ ❨❨i, 2❩, ❨3, 4❩❩', 2, [0.5*ImaginaryUnit, 1, 1.5, 2]);
  Test('0.5 ⋅ ❨❨i, 2❩, ❨3, 4❩❩', 2, [0.5*ImaginaryUnit, 1, 1.5, 2]);
  Test('i ⋅ ❨❨i, 2❩, ❨3, 4❩❩', 2, [-1, 2*ImaginaryUnit, 3*ImaginaryUnit, 4*ImaginaryUnit]);

  Test('10 ⋅ ❨❨1, 3, 4❩, ❨2, 1, 3❩, ❨4, 6, 7❩, ❨1, 2, 3❩, ❨5, 6, 7❩❩', 3, [10, 30, 40, 20, 10, 30, 40, 60, 70, 10, 20, 30, 50, 60, 70]);
  Test('10 ⋅ ❨❨i, 3, 4❩, ❨2, 1, 3❩, ❨4, 6, 7❩, ❨1, 2, 3❩, ❨5, 6, 7❩❩', 3, [10*ImaginaryUnit, 30, 40, 20, 10, 30, 40, 60, 70, 10, 20, 30, 50, 60, 70]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ ⋅ 10 ⋅ 10', 2, [100, 200, 300, 400]);
  Test('10 ⋅ ❨❨1, 2❩, ❨3, 4❩❩ ⋅ 10', 2, [100, 200, 300, 400]);
  Test('10 ⋅ 10 ⋅ ❨❨1, 2❩, ❨3, 4❩❩', 2, [100, 200, 300, 400]);

  // - string times non-negative 32-bit signed integer

  Test('0 ⋅ "test"', '', True, '');
  Test('1 ⋅ "test"', 'test', True, 'test');
  Test('2 ⋅ "test"', 'testtest', True, 'testtest');
  Test('3 ⋅ "test"', 'testtesttest', True, 'testtesttest');
  Test('4 ⋅ "test"', 'testtesttesttest', True, 'testtesttesttest');
  Test('5 ⋅ "test"', 'testtesttesttesttest', True, 'testtesttesttesttest');

  _s := DupeString('test', 1000);
  Test('1000 ⋅ "test"', _s);

  _s := DupeString('test', 10000);
  Test('10000 ⋅ "test"', _s);

  _s := DupeString('test', 100000);
  Test('100000 ⋅ "test"', _s);

  _s := DupeString('test', 1000000);
  Test('1000000 ⋅ "test"', _s);

  Test('"test" ⋅ 0', '', True, '');
  Test('"test" ⋅ 1', 'test', True, 'test');
  Test('"test" ⋅ 2', 'testtest', True, 'testtest');
  Test('"test" ⋅ 3', 'testtesttest', True, 'testtesttest');
  Test('"test" ⋅ 4', 'testtesttesttest', True, 'testtesttesttest');
  Test('"test" ⋅ 5', 'testtesttesttesttest', True, 'testtesttesttesttest');

  _s := DupeString('test', 1000);
  Test('"test" ⋅ 1000', _s);

  _s := DupeString('test', 10000);
  Test('"test" ⋅ 10000', _s);

  _s := DupeString('test', 100000);
  Test('"test" ⋅ 100000', _s);

  _s := DupeString('test', 1000000);
  Test('"test" ⋅ 1000000', _s);

  Test('"test" ⋅ -2', failure, 'Invalid function arguments.');
  Test('(-2) ⋅ "test"', failure, 'Invalid function arguments.');

  Test('2 ⋅ 2 ⋅ "test"', 'testtesttesttest', True, 'testtesttesttest');
  Test('2 ⋅ "test" ⋅ 2', 'testtesttesttest', True, 'testtesttesttest');
  Test('"test" ⋅ 2 ⋅ 2', 'testtesttesttest', True, 'testtesttesttest');

  Test('("test" ⋅ 2) ⋅ 2', 'testtesttesttest', True, 'testtesttesttest'); // same as without parentheses
  Test('"test" ⋅ (2 ⋅ 2)', 'testtesttesttest', True, 'testtesttesttest'); // same result, but faster (albeit unnoticeably; in theory, for huge strings)

  // - sound overloads: postponed

  // - invalid operands

  Test('"test" ⋅ "again"', failure, 'Invalid function arguments.');
  Test('"test" ⋅ ❨3, 2❩', failure, 'Invalid function arguments.');
  Test('❨3, 2❩ ⋅ "test"', failure, 'Invalid function arguments.');
  Test('"test" ⋅ ❨❨3, 2❩, ❨3, −2❩❩', failure, 'Invalid function arguments.');
  Test('❨❨3, 2❩, ❨3, −2❩❩ ⋅ "test"', failure, 'Invalid function arguments.');
  Test('date() ⋅ time()', failure, 'Invalid function arguments.');
  Test('5 ⋅ time()', failure, 'Invalid function arguments.');
  Test('date() ⋅ 5', failure, 'Invalid function arguments.');
  Test('true ⋅ 7', failure, 'Invalid function arguments.');
  Test('5 ⋅ false', failure, 'Invalid function arguments.');
  Test('5 ⋅ color("red")', failure, 'Invalid function arguments.');
  Test('5 ⋅ (.;)', failure, 'Invalid function arguments.');
  Test('(.;) ⋅ 10', failure, 'Invalid function arguments.');
  Test('(.;) ⋅ (.;)', failure, 'Invalid function arguments.');

  // Division

  // - two numbers

  Test('8642 / 1', 8642, True, '8642');
  Test('(-8642) / 1', -8642, True, '−8642');
  Test('8642 / (-1)', -8642, True, '−8642');
  Test('(-8642) / (-1)', 8642, True, '8642');

  Test('8642 / 2', 4321, True, '4321');
  Test('(-8642) / 2', -4321, True, '−4321');
  Test('8642 / (-2)', -4321, True, '−4321');
  Test('(-8642) / (-2)', 4321, True, '4321');

  Test('8642 / 29', 298, True, '298');
  Test('(-8642) / 29', -298, True, '−298');
  Test('8642 / (-29)', -298, True, '−298');
  Test('(-8642) / (-29)', 298, True, '298');

  Test('8642 / 149', 58, True, '58');
  Test('(-8642) / 149', -58, True, '−58');
  Test('8642 / (-149)', -58, True, '−58');
  Test('(-8642) / (-149)', 58, True, '58');

  Test('8642 / 4321', 2, True, '2');
  Test('(-8642) / 4321', -2, True, '−2');
  Test('8642 / (-4321)', -2, True, '−2');
  Test('(-8642) / (-4321)', 2, True, '2');

  Test('8642 / 8642', 1, True, '1');
  Test('(-8642) / 8642', -1, True, '−1');
  Test('8642 / (-8642)', -1, True, '−1');
  Test('(-8642) / (-8642)', 1, True, '1');

  Test('8642 / 3', Rat(8642, 3), True, '8642/3');
  Test('(-8642) / 3', Rat(-8642, 3), True, '−8642/3');
  Test('8642 / (-3)', Rat(-8642, 3), True, '−8642/3');
  Test('(-8642) / (-3)', Rat(8642, 3), True, '8642/3');

  Test('8642 / 123', Rat(8642, 123), True, '8642/123');
  Test('(-8642) / 123', Rat(-8642, 123), True, '−8642/123');
  Test('8642 / (-123)', Rat(-8642, 123), True, '−8642/123');
  Test('(-8642) / (-123)', Rat(8642, 123), True, '8642/123');

  Test('8642 / 124', Rat(4321, 62), True, '4321/62');
  Test('(-8642) / 124', Rat(-4321, 62), True, '−4321/62');
  Test('8642 / (-124)', Rat(-4321, 62), True, '−4321/62');
  Test('(-8642) / (-124)', Rat(4321, 62), True, '4321/62');

  Test('8642 / (1/2)', Rat(17284, 1), True, '17284');
  Test('8642 / (-1/2)', Rat(-17284, 1), True, '−17284');
  Test('(-8642) / (1/2)', Rat(-17284, 1), True, '−17284');
  Test('(-8642) / (-1/2)', Rat(17284, 1), True, '17284');

  Test('8642 / 0.5', 17284.0, True, '17284');
  Test('8642 / (-0.5)', -17284.0, True, '−17284');
  Test('(-8642) / 0.5', -17284.0, True, '−17284');
  Test('(-8642) / (-0.5)', 17284.0, True, '17284');

  Test('8642 / i', -8642*ImaginaryUnit, True, '−8642⋅i');
  Test('8642 / (-i)', 8642*ImaginaryUnit, True, '8642⋅i');
  Test('(-8642) / i', 8642*ImaginaryUnit, True, '8642⋅i');
  Test('(-8642) / (-i)', -8642*ImaginaryUnit, True, '−8642⋅i');

  Test('8642 / (2 + 6⋅i)', 432.1 - 1296.3*ImaginaryUnit, True, '432.1 − 1296.3⋅i');

  Test('4148888755878 / 8526387', 486594, True, '486594');

  Test('18364155781664328 / 738526387', 24865944, True, '24865944');
  Test('18364155781664328 / (-738526387)', -24865944, True, '−24865944');
  Test('(-18364155781664328) / 738526387', -24865944, True, '−24865944');
  Test('(-18364155781664328) / (-738526387)', 24865944, True, '24865944');

  Test('2189862929181664328 / 9738526387', 224865944, True, '224865944');
  Test('2189862929181664328 / (-9738526387)', -224865944, True, '−224865944');
  Test('(-2189862929181664328) / 9738526387', -224865944, True, '−224865944');
  Test('(-2189862929181664328) / (-9738526387)', 224865944, True, '224865944');

  Test('2189862929181664328 / 2', 1094931464590832164, True, '1094931464590832164');
  Test('2189862929181664328 / (-2)', -1094931464590832164, True, '−1094931464590832164');
  Test('(-2189862929181664328) / 2', -1094931464590832164, True, '−1094931464590832164');
  Test('(-2189862929181664328) / (-2)', 1094931464590832164, True, '1094931464590832164');

  Test('2189862929181664328 / 1094931464590832164', 2, True, '2');
  Test('2189862929181664328 / (-1094931464590832164)', -2, True, '−2');
  Test('(-2189862929181664328) / 1094931464590832164', -2, True, '−2');
  Test('(-2189862929181664328) / (-1094931464590832164)', 2, True, '2');

  Test('1094931464590832164 / 2189862929181664328', Rat(1, 2), True, '1/2');
  Test('1094931464590832164 / (-2189862929181664328)', Rat(-1, 2), True, '−1/2');
  Test('(-1094931464590832164) / 2189862929181664328', Rat(-1, 2), True, '−1/2');
  Test('(-1094931464590832164) / (-2189862929181664328)', Rat(1, 2), True, '1/2');

  Test('2189862929181664328 / 1', 2189862929181664328, True, '2189862929181664328');
  Test('2189862929181664328 / (-1)', -2189862929181664328, True, '−2189862929181664328');
  Test('(-2189862929181664328) / 1', -2189862929181664328, True, '−2189862929181664328');
  Test('(-2189862929181664328) / (-1)', 2189862929181664328, True, '2189862929181664328');

  Test('2189862929181664328 / i', -2189862929181664328 * ImaginaryUnit);
  Test('2189862929181664328 / (-i)', 2189862929181664328 * ImaginaryUnit);
  Test('(-2189862929181664328) / i', 2189862929181664328 * ImaginaryUnit);
  Test('(-2189862929181664328) / (-i)', -2189862929181664328 * ImaginaryUnit);

  Test('5/0', failure, 'Division by zero');
  Test('5/0.0', failure, 'Division by zero');
  Test('5/(1/3-1/3)', failure, 'Division by zero');
  Test('5/(i-i)', failure); // TODO: specify error message
  Test('2189862929181664328/0', failure, 'Division by zero');

  Test('(8/9) / 3', Rat(8, 27), True, '8/27');
  Test('(8/9) / -3', Rat(-8, 27), True, '−8/27');
  Test('(-8/9) / 3', Rat(-8, 27), True, '−8/27');
  Test('(-8/9) / -3', Rat(8, 27), True, '8/27');

  Test('(8/9) / (4/5)', Rat(10, 9), True, '10/9');
  Test('(8/9) / (-4/5)', Rat(-10, 9), True, '−10/9');
  Test('(-8/9) / (4/5)', Rat(-10, 9), True, '−10/9');
  Test('(-8/9) / (-4/5)', Rat(10, 9), True, '10/9');

  Test('(1/2) / 0.5', 1.0, True, '1');
  Test('(1/2) / -0.5', -1.0, True, '−1');
  Test('(-1/2) / 0.5', -1.0, True, '−1');
  Test('(-1/2) / -0.5', 1.0, True, '1');

  Test('(59/60) / i', -(59/60)*ImaginaryUnit);
  Test('(59/60) / -i', (59/60)*ImaginaryUnit);
  Test('(-59/60) / i', (59/60)*ImaginaryUnit);
  Test('(-59/60) / -i', -(59/60)*ImaginaryUnit);

  Test('(587312/895731) / (1452879/635217921)', Rat(41452345290928, 144598751061), True, '41452345290928/144598751061');
  Test('(587312/895731) / (-1452879/635217921)', Rat(-41452345290928, 144598751061), True, '−41452345290928/144598751061');
  Test('(-587312/895731) / (1452879/635217921)', Rat(-41452345290928, 144598751061), True, '−41452345290928/144598751061');
  Test('(-587312/895731) / (-1452879/635217921)', Rat(41452345290928, 144598751061), True, '41452345290928/144598751061');

  Test('(2/3)/0', failure, 'Division by zero');
  Test('(2/3)/0.0', failure, 'Division by zero');
  Test('(2/3)/(1/3-1/3)', failure, 'Division by zero');
  Test('(2/3)/(i-i)', failure); // TODO: specify error message

  Test('123456.789 / 2', 61728.3945, True, '61728.3945');
  Test('123456.789 / -2', -61728.3945, True, '−61728.3945');
  Test('(-123456.789) / 2', -61728.3945, True, '−61728.3945');
  Test('(-123456.789) / -2', 61728.3945, True, '61728.3945');

  Test('123456.789 / (1/2)', 246913.578, True, '246913.578');
  Test('123456.789 / (-1/2)', -246913.578, True, '−246913.578');
  Test('(-123456.789) / (1/2)', -246913.578, True, '−246913.578');
  Test('(-123456.789) / (-1/2)', 246913.578, True, '246913.578');

  Test('123456.789 / 0.5', 246913.578, True, '246913.578');
  Test('123456.789 / -0.5', -246913.578, True, '−246913.578');
  Test('(-123456.789) / 0.5', -246913.578, True, '−246913.578');
  Test('(-123456.789) / -0.5', 246913.578, True, '246913.578');

  Test('123456.789 / i', -123456.789*ImaginaryUnit, True, '−123456.789⋅i');
  Test('123456.789 / -i', 123456.789*ImaginaryUnit, True, '123456.789⋅i');
  Test('(-123456.789) / i', 123456.789*ImaginaryUnit, True, '123456.789⋅i');
  Test('(-123456.789) / -i', -123456.789*ImaginaryUnit, True, '−123456.789⋅i');

  Test('12.3 / (1 + 2⋅i)', 2.46 - 4.92*ImaginaryUnit, True, '2.46 − 4.92⋅i');
  Eps; Test('5268.5212 / (545.62 − 5.189⋅i)', 9.655153273823309642705 + 0.09182322923989068167588*ImaginaryUnit);

  Test('4.82E20 / 2', 2.41E20, True, '2.41⋅10^20');
  Test('4.82E20 / -2', -2.41E20, True, '−2.41⋅10^20');
  Test('(-4.82E20) / 2', -2.41E20, True, '−2.41⋅10^20');
  Test('(-4.82E20) / -2', 2.41E20, True, '2.41⋅10^20');

  Test('5.0/0', failure, 'Division by zero');
  Test('5.0/0.0', failure, 'Division by zero');
  Test('5.0/(1/3-1/3)', failure, 'Division by zero');
  Test('5.0/(i-i)', failure); // TODO: specify error message

  Test('i/2', ImaginaryUnit/2, True, '0.5⋅i');
  Test('i/(1/2)', 2*ImaginaryUnit, True, '2⋅i');
  Test('i/i', TASC(1), True, '1');

  Test('(7 - 3⋅i) / 2', 3.5 - 1.5*ImaginaryUnit, True, '3.5 − 1.5⋅i');
  Test('(7 - 3⋅i) / (1/2)', 14 - 6*ImaginaryUnit, True, '14 − 6⋅i');
  Test('(7 - 3⋅i) / 0.5', 14 - 6*ImaginaryUnit, True, '14 − 6⋅i');
  Test('(7 - 3⋅i) / i', -3 - 7*ImaginaryUnit, True, '−3 − 7⋅i');

  Eps; Test('(7 - 3⋅i) / (-2 + 20⋅i)', -0.18316831683168316831683168 - 0.3316831683168316831683*ImaginaryUnit);

  Test('i/0', failure); // TODO: specify error message
  Test('i/0.0', failure); // TODO: specify error message
  Test('i/(1/3-1/3)', failure); // TODO: specify error message
  Test('i/(i-i)', failure); // TODO: specify error message

  Test('8305843009213693952 / 0.1', 83058430092136939520.0);
  Test('5E20 / 2', 2.5E20);
  Test('5E200 / 2', 2.5E200);
  Test('5E20 / 2E20', 2.5);

  Test('(-8305843009213693952) / 0.1', -83058430092136939520.0);
  Test('(-5E20) / 2', -2.5E20);
  Test('(-5E200) / 2', -2.5E200);
  Test('(-5E20) / 2E20', -2.5);

  Test('8305843009213693952 / -0.1', -83058430092136939520.0);
  Test('5E20 / -2', -2.5E20);
  Test('5E200 / -2', -2.5E200);
  Test('5E20 / -2E20', -2.5);


  // - vector and number

  Test('❨5, 0, 1❩ / 1', [5, 0, 1]);
  Test('❨5, 0, 1❩ / 2', [2.5, 0, 0.5]);
  Test('❨5, 0, 1❩ / 3', [5/3, 0, 1/3]);
  Test('❨5, 0, 1❩ / 4', [1.25, 0, 0.25]);
  Test('❨5, 0, 1❩ / 5', [1, 0, 0.2]);
  Test('❨5, 0, 1❩ / 6', [5/6, 0, 1/6]);
  Test('❨5, 0, 1❩ / 7', [5/7, 0, 1/7]);
  Test('❨5, 0, 1❩ / 8', [5/8, 0, 1/8]);
  Test('❨5, 0, 1❩ / 9', [5/9, 0, 1/9]);
  Test('❨5, 0, 1❩ / 10', [0.5, 0, 0.1]);

  Test('❨5, 0, 1❩ / -1', [-5, 0, -1]);
  Test('❨5, 0, 1❩ / -2', [-2.5, 0, -0.5]);
  Test('❨5, 0, 1❩ / -3', [-5/3, 0, -1/3]);
  Test('❨5, 0, 1❩ / -4', [-1.25, 0, -0.25]);
  Test('❨5, 0, 1❩ / -5', [-1, 0, -0.2]);
  Test('❨5, 0, 1❩ / -6', [-5/6, 0, -1/6]);
  Test('❨5, 0, 1❩ / -7', [-5/7, 0, -1/7]);
  Test('❨5, 0, 1❩ / -8', [-5/8, 0, -1/8]);
  Test('❨5, 0, 1❩ / -9', [-5/9, 0, -1/9]);
  Test('❨5, 0, 1❩ / -10', [-0.5, 0, -0.1]);

  Test('❨5, 0, 1❩ / (1/2)', [10, 0, 2]);
  Test('❨5, 0, 1❩ / (-1/2)', [-10, 0, -2]);
  Test('❨5, 0, 1❩ / 0.5', [10, 0, 2]);
  Test('❨5, 0, 1❩ / -0.5', [-10, 0, -2]);

  Test('❨5, 0, 1❩ / i', [-5*ImaginaryUnit, 0, -ImaginaryUnit]);
  Test('❨5, 0, 1❩ / (2+i)', [2-ImaginaryUnit, 0, 0.4 - 0.2*ImaginaryUnit]);

  Test('❨5, 0, 1❩ / 0', failure, 'Division by zero');
  Test('❨5, 0, 1❩ / 0.0', failure, 'Division by zero');
  Test('❨5, 0, 1❩ / (1/3-1/3)', failure, 'Division by zero');
  Test('❨5, 0, 1❩ / (i-i)', failure); // TODO: specify error message

  Test('❨5, i, 1❩ / 1', [5, ImaginaryUnit, 1]);
  Test('❨5, i, 1❩ / 2', [2.5, 0.5*ImaginaryUnit, 0.5]);
  Test('❨5, i, 1❩ / 3', [5/3, ImaginaryUnit/3, 1/3]);
  Test('❨5, i, 1❩ / 4', [1.25, 0.25*ImaginaryUnit, 0.25]);
  Test('❨5, i, 1❩ / 5', [1, 0.2*ImaginaryUnit, 0.2]);
  Test('❨5, i, 1❩ / 6', [5/6, ImaginaryUnit/6, 1/6]);
  Test('❨5, i, 1❩ / 7', [5/7, ImaginaryUnit/7, 1/7]);
  Test('❨5, i, 1❩ / 8', [5/8, ImaginaryUnit/8, 1/8]);
  Test('❨5, i, 1❩ / 9', [5/9, ImaginaryUnit/9, 1/9]);
  Test('❨5, i, 1❩ / 10', [0.5, 0.1*ImaginaryUnit, 0.1]);

  Test('❨5, i, 1❩ / -1', [-5, -ImaginaryUnit, -1]);
  Test('❨5, i, 1❩ / -2', [-2.5, -0.5*ImaginaryUnit, -0.5]);
  Test('❨5, i, 1❩ / -3', [-5/3, -ImaginaryUnit/3, -1/3]);
  Test('❨5, i, 1❩ / -4', [-1.25, -0.25*ImaginaryUnit, -0.25]);
  Test('❨5, i, 1❩ / -5', [-1, -0.2*ImaginaryUnit, -0.2]);
  Test('❨5, i, 1❩ / -6', [-5/6, -ImaginaryUnit/6, -1/6]);
  Test('❨5, i, 1❩ / -7', [-5/7, -ImaginaryUnit/7, -1/7]);
  Test('❨5, i, 1❩ / -8', [-5/8, -ImaginaryUnit/8, -1/8]);
  Test('❨5, i, 1❩ / -9', [-5/9, -ImaginaryUnit/9, -1/9]);
  Test('❨5, i, 1❩ / -10', [-0.5, -0.1*ImaginaryUnit, -0.1]);

  Test('❨5, i, 1❩ / (1/2)', [10, 2*ImaginaryUnit, 2]);
  Test('❨5, i, 1❩ / (-1/2)', [-10, -2*ImaginaryUnit, -2]);
  Test('❨5, i, 1❩ / 0.5', [10, 2*ImaginaryUnit, 2]);
  Test('❨5, i, 1❩ / -0.5', [-10, -2*ImaginaryUnit, -2]);

  Test('❨5, i, 1❩ / i', [-5*ImaginaryUnit, 1, -ImaginaryUnit]);
  Test('❨5, i, 1❩ / (2+i)', [2-ImaginaryUnit, 0.2 + 0.4*ImaginaryUnit, 0.4 - 0.2*ImaginaryUnit]);

  Test('❨5, i, 1❩ / 0', failure); // TODO: specify error message
  Test('❨5, i, 1❩ / 0.0', failure); // TODO: specify error message
  Test('❨5, i, 1❩ / (1/3-1/3)', failure); // TODO: specify error message
  Test('❨5, i, 1❩ / (i-i)', failure); // TODO: specify error message

  Test('❨10, 20, 30, 40, 50, 60, 70, 80, 90, 100❩ / 10', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  Test('❨i, 20, 30, 40, 50, 60, 70, 80, 90, 100❩ / 10', [ImaginaryUnit/10, 2, 3, 4, 5, 6, 7, 8, 9, 10]);

  // - matrix and number

  Test('❨❨1, 2❩, ❨3, 4❩❩ / 1', 2, [1, 2, 3, 4]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 2', 2, [1/2, 2/2, 3/2, 4/2]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 3', 2, [1/3, 2/3, 3/3, 4/3]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 4', 2, [1/4, 2/4, 3/4, 4/4]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 5', 2, [1/5, 2/5, 3/5, 4/5]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 6', 2, [1/6, 2/6, 3/6, 4/6]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 7', 2, [1/7, 2/7, 3/7, 4/7]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 8', 2, [1/8, 2/8, 3/8, 4/8]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 9', 2, [1/9, 2/9, 3/9, 4/9]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 10', 2, [1/10, 2/10, 3/10, 4/10]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ / -1', 2, [-1, -2, -3, -4]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -2', 2, [-1/2, -2/2, -3/2, -4/2]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -3', 2, [-1/3, -2/3, -3/3, -4/3]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -4', 2, [-1/4, -2/4, -3/4, -4/4]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -5', 2, [-1/5, -2/5, -3/5, -4/5]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -6', 2, [-1/6, -2/6, -3/6, -4/6]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -7', 2, [-1/7, -2/7, -3/7, -4/7]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -8', 2, [-1/8, -2/8, -3/8, -4/8]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -9', 2, [-1/9, -2/9, -3/9, -4/9]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -10', 2, [-1/10, -2/10, -3/10, -4/10]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ / (1/2)', 2, [2, 4, 6, 8]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / (-1/2)', 2, [-2, -4, -6, -8]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 0.5', 2, [2, 4, 6, 8]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / -0.5', 2, [-2, -4, -6, -8]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ / i', 2, [-ImaginaryUnit, -2*ImaginaryUnit, -3*ImaginaryUnit, -4*ImaginaryUnit]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ / (1+i)', 2, [0.5-0.5*ImaginaryUnit, 1-ImaginaryUnit, 1.5-1.5*ImaginaryUnit, 2-2*ImaginaryUnit]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ / 0', failure, 'Division by zero');
  Test('❨❨1, 2❩, ❨3, 4❩❩ / 0.0', failure, 'Division by zero');
  Test('❨❨1, 2❩, ❨3, 4❩❩ / (1/3-1/3)', failure, 'Division by zero');
  Test('❨❨1, 2❩, ❨3, 4❩❩ / (i-i)', failure); // TODO: specify error message

  Test('❨❨i, 2❩, ❨3, 4❩❩ / 1', 2, [ImaginaryUnit, 2, 3, 4]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 2', 2, [ImaginaryUnit/2, 2/2, 3/2, 4/2]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 3', 2, [ImaginaryUnit/3, 2/3, 3/3, 4/3]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 4', 2, [ImaginaryUnit/4, 2/4, 3/4, 4/4]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 5', 2, [ImaginaryUnit/5, 2/5, 3/5, 4/5]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 6', 2, [ImaginaryUnit/6, 2/6, 3/6, 4/6]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 7', 2, [ImaginaryUnit/7, 2/7, 3/7, 4/7]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 8', 2, [ImaginaryUnit/8, 2/8, 3/8, 4/8]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 9', 2, [ImaginaryUnit/9, 2/9, 3/9, 4/9]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 10', 2, [ImaginaryUnit/10, 2/10, 3/10, 4/10]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ / -1', 2, [-ImaginaryUnit, -2, -3, -4]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -2', 2, [-ImaginaryUnit/2, -2/2, -3/2, -4/2]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -3', 2, [-ImaginaryUnit/3, -2/3, -3/3, -4/3]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -4', 2, [-ImaginaryUnit/4, -2/4, -3/4, -4/4]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -5', 2, [-ImaginaryUnit/5, -2/5, -3/5, -4/5]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -6', 2, [-ImaginaryUnit/6, -2/6, -3/6, -4/6]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -7', 2, [-ImaginaryUnit/7, -2/7, -3/7, -4/7]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -8', 2, [-ImaginaryUnit/8, -2/8, -3/8, -4/8]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -9', 2, [-ImaginaryUnit/9, -2/9, -3/9, -4/9]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -10', 2, [-ImaginaryUnit/10, -2/10, -3/10, -4/10]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ / (1/2)', 2, [2*ImaginaryUnit, 4, 6, 8]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / (-1/2)', 2, [-2*ImaginaryUnit, -4, -6, -8]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 0.5', 2, [2*ImaginaryUnit, 4, 6, 8]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / -0.5', 2, [-2*ImaginaryUnit, -4, -6, -8]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ / i', 2, [1, -2*ImaginaryUnit, -3*ImaginaryUnit, -4*ImaginaryUnit]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ / (1+i)', 2, [0.5+0.5*ImaginaryUnit, 1-ImaginaryUnit, 1.5-1.5*ImaginaryUnit, 2-2*ImaginaryUnit]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ / 0', failure); // TODO: specify error message
  Test('❨❨i, 2❩, ❨3, 4❩❩ / 0.0', failure); // TODO: specify error message
  Test('❨❨i, 2❩, ❨3, 4❩❩ / (1/3-1/3)', failure); // TODO: specify error message
  Test('❨❨i, 2❩, ❨3, 4❩❩ / (i-i)', failure); // TODO: specify error message

  Test('❨❨1, 4, 2, 3, 0, 2, 1❩, ❨4, 3, 2, 6, 0, −2, 3❩❩ / 2', 7, [0.5, 2, 1, 1.5, 0, 1, 0.5, 2, 1.5, 1, 3, 0, -1, 1.5]);
  Test('❨❨i, 4, 2, 3, 0, 2, 1❩, ❨4, 3, 2, 6, 0, −2, 3❩❩ / 2', 7, [0.5*ImaginaryUnit, 2, 1, 1.5, 0, 1, 0.5, 2, 1.5, 1, 3, 0, -1, 1.5]);
  Test('❨❨7❩❩ / 5', 1, [7/5]);
  Test('❨❨i❩❩ / 5', 1, [ImaginaryUnit/5]);

  // - invalid operands

  Test('"alfa" / 2', failure, 'Invalid function arguments.');
  Test('true / 2', failure, 'Invalid function arguments.');
  Test('now() / 2', failure, 'Invalid function arguments.');
  Test('color("red") / 2', failure, 'Invalid function arguments.');
  Test('(.;) / 2', failure, 'Invalid function arguments.');
  Test('"alfa" / "beta"', failure, 'Invalid function arguments.');
  Test('25 / "beta"', failure, 'Invalid function arguments.');
  Test('true / false', failure, 'Invalid function arguments.');
  Test('false / "gamma"', failure, 'Invalid function arguments.');
  Test('date() / "epsilon"', failure, 'Invalid function arguments.');
  Test('date() / time()', failure, 'Invalid function arguments.');
  Test('color("red") / false', failure, 'Invalid function arguments.');
  Test('(.;) / date()', failure, 'Invalid function arguments.');
  Test('9 / date()', failure, 'Invalid function arguments.');
  Test('"gamma" / date()', failure, 'Invalid function arguments.');
  Test('10 / (.;)', failure, 'Invalid function arguments.');
  Test('❨5, 0, 1❩ / ❨5, 2, 1❩', failure, 'Invalid function arguments.');
  Test('❨5, 0, 1❩ / "alfa"', failure, 'Invalid function arguments.');
  Test('❨5, 0, 1❩ / true', failure, 'Invalid function arguments.');
  Test('❨5, 0, 1❩ / date()', failure, 'Invalid function arguments.');
  Test('❨5, 0, 1❩ / color("red")', failure, 'Invalid function arguments.');
  Test('color("red") / ❨5, 0, 1❩', failure, 'Invalid function arguments.');
  Test('false / ❨5, 0, 1❩', failure, 'Invalid function arguments.');
  Test('"test" / ❨5, 0, 1❩', failure, 'Invalid function arguments.');
  Test('123 / ❨5, 0, 1❩', failure, 'Invalid function arguments.');

  // Exponentiation

  // - two numbers

  Test('2^0', 1);
  Test('2^1', 2);
  Test('2^2', 4);
  Test('2^3', 8);
  Test('2^4', 16);
  Test('2^5', 32);
  Test('2^6', 64);
  Test('2^7', 128);
  Test('2^8', 256);
  Test('2^9', 512);
  Test('2^10', 1024);
  Test('2^11', 2048);
  Test('2^12', 4096);
  Test('2^13', 8192);
  Test('2^14', 16384);
  Test('2^15', 32768);
  Test('2^16', 65536);
  Test('2^17', 131072);
  Test('2^18', 262144);
  Test('2^19', 524288);
  Test('2^20', 1048576);

  Test('2^50', 1125899906842624);

  Eps; Test('2^100', 1.267650600228229401496703205376E30);
  Eps; Test('2^200', 1.606938044258990275541962092341E60);
  Eps; Test('2^300', 2.037035976334486086268445688409E90);
  Eps; Test('2^400', 2.582249878086908589655919172003E120);
  Eps; Test('2^500', 3.273390607896141870013189696827E150);
  Eps; Test('2^600', 4.149515568880992958512407863691E180);
  Eps; Test('2^700', 5.260135901548373507240989882880E210);
  Eps; Test('2^800', 6.668014432879854274079851790721E240);
  Eps; Test('2^900', 8.452712498170643941637436558664E270);
  Eps; Test('2^1000', 1.071508607186267320948425049060E301);

  Eps; Test('2^2000', 1.148130695274254524232833201177E602);

  Test('2^-1', Rat(1, 2), True, '1/2');
  Test('2^-2', Rat(1, 4), True, '1/4');
  Test('2^-3', Rat(1, 8), True, '1/8');
  Test('2^-4', Rat(1, 16), True, '1/16');
  Test('2^-5', Rat(1, 32), True, '1/32');
  Test('2^-6', Rat(1, 64), True, '1/64');
  Test('2^-7', Rat(1, 128), True, '1/128');
  Test('2^-8', Rat(1, 256), True, '1/256');
  Test('2^-9', Rat(1, 512), True, '1/512');
  Test('2^-10', Rat(1, 1024), True, '1/1024');

  Test('2^-50', Rat(1, 1125899906842624), True, '1/1125899906842624');

  Eps; Test('2^-100/1E-31', 7.888609052210118054117285652827);
  Eps; Test('2^-200/1E-61', 6.223015277861141707144064053780);
  Eps; Test('2^-300/1E-91', 4.909093465297726553095771954986);
  Eps; Test('2^-400/1E-121', 3.872591914849318272818030633286);
  Eps; Test('2^-500/1E-151', 3.054936363499604682051979393213);
  Eps; Test('2^-600/1E-181', 2.409919865102884117740750034712);
  Eps; Test('2^-700/1E-211', 1.901091566295159823515072405835);
  Eps; Test('2^-800/1E-241', 1.499696813895630954817644437628);
  Eps; Test('2^-900/1E-271', 1.183052186166774710972751597518);
  Eps; Test('2^-1000/1E-302', 9.332636185032188789900895447238);

  Eps; Test('2^-2000/1E-603', 8.709809816217216675576195494778);

  Test('10^0', 1);
  Test('10^1', 10);
  Test('10^2', 100);
  Test('10^3', 1000);
  Test('10^4', 10000);
  Test('10^5', 100000);
  Test('10^6', 1000000);
  Test('10^7', 10000000);
  Test('10^8', 100000000);
  Test('10^9', 1000000000);
  Test('10^10', 10000000000);
  Test('10^11', 100000000000);
  Test('10^12', 1000000000000);
  Test('10^13', 10000000000000);
  Test('10^14', 100000000000000);
  Test('10^15', 1000000000000000);
  Test('10^16', 10000000000000000);
  Test('10^17', 100000000000000000);
  Test('10^18', 1000000000000000000);

  Eps; Test('10^50', 1E50);
  Eps; Test('10^100', 1E100);
  Eps; Test('10^200', 1E200);
  Eps; Test('10^300', 1E300);
  Eps; Test('10^400', 1E400);
  Eps; Test('10^500', 1E500);
  Eps; Test('10^600', 1E600);

  Test('10^-1', Rat(1, 10), True, '1/10');
  Test('10^-2', Rat(1, 100), True, '1/100');
  Test('10^-3', Rat(1, 1000), True, '1/1000');
  Test('10^-4', Rat(1, 10000), True, '1/10000');
  Test('10^-5', Rat(1, 100000), True, '1/100000');
  Test('10^-6', Rat(1, 1000000), True, '1/1000000');
  Test('10^-7', Rat(1, 10000000), True, '1/10000000');
  Test('10^-8', Rat(1, 100000000), True, '1/100000000');
  Test('10^-9', Rat(1, 1000000000), True, '1/1000000000');
  Test('10^-10', Rat(1, 10000000000), True, '1/10000000000');
  Test('10^-11', Rat(1, 100000000000), True, '1/100000000000');
  Test('10^-12', Rat(1, 1000000000000), True, '1/1000000000000');
  Test('10^-13', Rat(1, 10000000000000), True, '1/10000000000000');
  Test('10^-14', Rat(1, 100000000000000), True, '1/100000000000000');
  Test('10^-15', Rat(1, 1000000000000000), True, '1/1000000000000000');
  Test('10^-16', Rat(1, 10000000000000000), True, '1/10000000000000000');
  Test('10^-17', Rat(1, 100000000000000000), True, '1/100000000000000000');
  Test('10^-18', Rat(1, 1000000000000000000), True, '1/1000000000000000000');

  Eps; Test('10^−50/1E-50', 1.0);
  Eps; Test('10^−100/1E-100', 1.0);
  Eps; Test('10^−200/1E-200', 1.0);
  Eps; Test('10^−300/1E-300', 1.0);
  Eps; Test('10^−400/1E-400', 1.0);
  Eps; Test('10^−500/1E-500', 1.0);
  Eps; Test('10^−600/1E-600', 1.0);

  Eps; Test('253^214', 1.85264211176629736851898E514);
  Test('7^12', 13841287201, True, '13841287201');
  Test('7^-12', Rat(1, 13841287201), True, '1/13841287201');

  Test('0^1', 0);
  Test('0^2', 0);
  Test('0^3', 0);
  Test('0^4', 0);
  Test('0^5', 0);
  Test('0^6', 0);
  Test('0^7', 0);
  Test('0^8', 0);
  Test('0^9', 0);
  Test('0^10', 0);
  Test('0^100', 0);
  Test('0^1000', 0);

  Test('0^0', failure, 'Zero raised to the power of zero.');

  Test('0^−1', failure, 'Division by zero');
  Test('0^−2', failure, 'Division by zero');
  Test('0^−10', failure, 'Division by zero');
  Test('0^−100', failure, 'Division by zero');
  Test('0^−1000', failure, 'Division by zero');

  Test('-2^10', -1024, True, '−1024');
  Test('−2^10', -1024, True, '−1024');
  Test('-(2^10)', -1024, True, '−1024');
  Test('−(2^10)', -1024, True, '−1024');

  Test('(-2)^10', 1024, True, '1024');
  Test('(−2)^10', 1024, True, '1024');
  Test('(-2)^11', -2048, True, '−2048');
  Test('(−2)^11', -2048, True, '−2048');

  Test('(-10)^6', 1000000, True, '1000000');
  Test('(-10)^7', -10000000, True, '−10000000');

  Test('(-2)^3', -8);

  Test('(-2)^3', -8, True, '−8');
  Test('(-2)^2', 4, True, '4');
  Test('(-2)^1', -2, True, '−2');
  Test('(-2)^0', 1, True, '1');
  Test('(-2)^(-1)', Rat(-1, 2), True, '−1/2');
  Test('(-2)^(-2)', Rat(1, 4), True, '1/4');
  Test('(-2)^(-3)', Rat(-1, 8), True, '−1/8');
  Test('(-2)^(-4)', Rat(1, 16), True, '1/16');
  Test('(-2)^(-5)', Rat(-1, 32), True, '−1/32');
  Test('(-2)^(-6)', Rat(1, 64), True, '1/64');

  Test('(-2)^50', 1125899906842624);
  Test('(-2)^51', -2251799813685248);

  Test('(-2)^100',  1.267650600228229401496703E30);
  Test('(-2)^101', -2.535301200456458802993406E30);
  Test('(-2)^102',  5.070602400912917605986812E30);
  Test('(-2)^103', -1.014120480182583521197362E31);

  Eps; Test('(-2)^1000',  1.071508607186267320948425E301);
  Eps; Test('(-2)^1001', -2.143017214372534641896850E301);
  Eps; Test('(-2)^1002',  4.286034428745069283793700E301);
  Eps; Test('(-2)^1003', -8.572068857490138567587400E301);

  Test('(-2)^-50', Rat(1, 1125899906842624), True, '1/1125899906842624');
  Test('(-2)^-51', Rat(-1, 2251799813685248), True, '−1/2251799813685248');

  Eps; Test('(-2)^-100 / 1E-31',  7.88860905221011805411728565282);
  Eps; Test('(-2)^-101 / 1E-31', -3.94430452610505902705864282641);
  Eps; Test('(-2)^-102 / 1E-31',  1.97215226305252951352932141320);
  Eps; Test('(-2)^-103 / 1E-32', -9.86076131526264756764660706603);

  Eps; Test('(-2)^-1000 / 1E-302',  9.3326361850321887899008954472);
  Eps; Test('(-2)^-1001 / 1E-302', -4.6663180925160943949504477236);
  Eps; Test('(-2)^-1002 / 1E-302',  2.3331590462580471974752238618);
  Eps; Test('(-2)^-1003 / 1E-302', -1.1665795231290235987376119309);

  Test('1^(1/2)', 1.0, True, '1');
  Test('4^(1/2)', 2.0, True, '2');
  Test('9^(1/2)', 3.0, True, '3');
  Test('16^(1/2)', 4.0, True, '4');
  Eps; Test('25^(1/2)', 5.0, True, '5');
  Test('36^(1/2)', 6.0, True, '6');
  Test('49^(1/2)', 7.0, True, '7');
  Test('64^(1/2)', 8.0, True, '8');
  Test('81^(1/2)', 9.0, True, '9');
  Eps; Test('100^(1/2)', 10.0, True, '10');

  Test('1^(1/3)', 1.0, True, '1');
  Test('8^(1/3)', 2.0, True, '2');
  Eps; Test('27^(1/3)', 3.0, True, '3');
  Test('64^(1/3)', 4.0, True, '4');
  Eps; Test('125^(1/3)', 5.0, True, '5');

  Test('1^(2/3)', 1.0, True, '1');
  Test('8^(2/3)', 4.0, True, '4');
  Eps; Test('27^(2/3)', 9.0, True, '9');
  Test('64^(2/3)', 16.0, True, '16');
  Eps; Test('125^(2/3)', 25.0, True, '25');

  Test('1^(3/2)', 1.0, True, '1');
  Test('4^(3/2)', 8.0, True, '8');
  Eps; Test('9^(3/2)', 27.0, True, '27');
  Test('16^(3/2)', 64.0, True, '64');
  Eps; Test('25^(3/2)', 125.0, True, '125');

  Eps; Test('123^(1/2)', 11.09053650640941716205160010);
  Eps; Test('123^(5/7)', 31.10207507262337535852008947);
  Eps; Test('123^(9/7)', 486.4305665996165971556058156);

  Test('1^(−1/2)', 1.0, True, '1');
  Test('4^(−1/2)', 0.5, True, '0.5');
  Eps; Test('9^(−1/2)', 1/3);
  Test('16^(−1/2)', 0.25, True, '0.25');
  Eps; Test('25^(−1/2)', 0.2, True, '0.2');
  Eps; Test('36^(−1/2)', 1/6);
  Test('49^(−1/2)', 1/7);
  Test('64^(−1/2)', 1/8);
  Test('81^(−1/2)', 1/9);
  Eps; Test('100^(−1/2)', 0.1, True, '0.1');

  Test('1^(−1/3)', 1.0, True, '1');
  Test('8^(−1/3)', 0.5, True, '0.5');
  Eps; Test('27^(−1/3)', 1/3);
  Test('64^(−1/3)', 0.25, True, '0.25');
  Eps; Test('125^(−1/3)', 0.2, True, '0.2');

  Test('1^(−2/3)', 1.0, True, '1');
  Test('8^(−2/3)', 0.25, True, '0.25');
  Eps; Test('27^(−2/3)', 1/9);
  Test('64^(−2/3)', 1/16);
  Eps; Test('125^(−2/3)', 0.04, True, '0.04');

  Test('1^(−3/2)', 1.0, True, '1');
  Test('4^(−3/2)', 1/8);
  Eps; Test('9^(−3/2)', 1/27);
  Test('16^(−3/2)', 1/64);
  Eps; Test('25^(−3/2)', 1/125);

  Eps; Test('123^(−1/2)', 1/11.09053650640941716205160010);
  Eps; Test('123^(−5/7)', 1/31.10207507262337535852008947);
  Eps; Test('123^(−9/7)', 1/486.4305665996165971556058156);

  Test('1^0.5', 1.0, True, '1');
  Test('4^0.5', 2.0, True, '2');
  Test('9^0.5', 3.0, True, '3');
  Test('16^0.5', 4.0, True, '4');
  Eps; Test('25^0.5', 5.0, True, '5');
  Test('36^0.5', 6.0, True, '6');
  Test('49^0.5', 7.0, True, '7');
  Test('64^0.5', 8.0, True, '8');
  Test('81^0.5', 9.0, True, '9');
  Eps; Test('100^0.5', 10.0, True, '10');

  Test('1^(1./3)', 1.0, True, '1');
  Test('8^(1./3)', 2.0, True, '2');
  Eps; Test('27^(1./3)', 3.0, True, '3');
  Test('64^(1./3)', 4.0, True, '4');
  Eps; Test('125^(1./3)', 5.0, True, '5');

  Test('1^(2./3)', 1.0, True, '1');
  Test('8^(2./3)', 4.0, True, '4');
  Eps; Test('27^(2./3)', 9.0, True, '9');
  Test('64^(2./3)', 16.0, True, '16');
  Eps; Test('125^(2./3)', 25.0, True, '25');

  Test('1^(3./2)', 1.0, True, '1');
  Test('4^(3./2)', 8.0, True, '8');
  Eps; Test('9^(3./2)', 27.0, True, '27');
  Test('16^(3./2)', 64.0, True, '64');
  Eps; Test('25^(3./2)', 125.0, True, '125');

  Eps; Test('123^(1./2)', 11.09053650640941716205160010);
  Eps; Test('123^(5./7)', 31.10207507262337535852008947);
  Eps; Test('123^(9./7)', 486.4305665996165971556058156);

  Test('1^(−1./2)', 1.0, True, '1');
  Test('4^(−1./2)', 0.5, True, '0.5');
  Eps; Test('9^(−1./2)', 1/3);
  Test('16^(−1./2)', 0.25, True, '0.25');
  Eps; Test('25^(−1./2)', 0.2, True, '0.2');
  Eps; Test('36^(−1./2)', 1/6);
  Test('49^(−1./2)', 1/7);
  Test('64^(−1./2)', 1/8);
  Test('81^(−1./2)', 1/9);
  Eps; Test('100^(−1./2)', 0.1, True, '0.1');

  Test('1^(−1./3)', 1.0, True, '1');
  Test('8^(−1./3)', 0.5, True, '0.5');
  Eps; Test('27^(−1./3)', 1/3);
  Test('64^(−1./3)', 0.25, True, '0.25');
  Eps; Test('125^(−1./3)', 0.2, True, '0.2');

  Test('1^(−2./3)', 1.0, True, '1');
  Test('8^(−2./3)', 0.25, True, '0.25');
  Eps; Test('27^(−2./3)', 1/9);
  Test('64^(−2./3)', 1/16);
  Eps; Test('125^(−2./3)', 0.04, True, '0.04');

  Test('1^(−3./2)', 1.0, True, '1');
  Test('4^(−3./2)', 1/8);
  Eps; Test('9^(−3./2)', 1/27);
  Test('16^(−3./2)', 1/64);
  Eps; Test('25^(−3./2)', 1/125);

  Eps; Test('123^(−1./2)', 1/11.09053650640941716205160010);
  Eps; Test('123^(−5./7)', 1/31.10207507262337535852008947);
  Eps; Test('123^(−9./7)', 1/486.4305665996165971556058156);

  Test('1^(1/2 - 1/2)', 1, True, '1');
  Test('81^(1/3 - 1/3)', 1, True, '1');
  Test('0^(1/2 - 1/2)', failure, 'Zero raised to the power of zero.');

  Eps; Test('(-2)^(1/2)', Sqrt(2)*ImaginaryUnit);
  Eps; Test('(-2)^0.5', Sqrt(2)*ImaginaryUnit);
  Eps; Test('(-2)^-(1/2)', -Sqrt(1/2)*ImaginaryUnit);
  Eps; Test('(-2)^-0.5', -Sqrt(1/2)*ImaginaryUnit);

  Eps; Test('2^π', 8.8249778270762876238564296042);
  Eps; Test('2^e', 6.5808859910179209708515424038);
  Eps; Test('2^-π', 1/8.8249778270762876238564296042);
  Eps; Test('2^-e', 1/6.5808859910179209708515424038);

  Test('0^0.0', failure, 'Zero raised to the power of zero.');

  Test('(3/5)^0', Rat(1, 1), True, '1');
  Test('(3/5)^1', Rat(3, 5), True, '3/5');
  Test('(3/5)^2', Rat(9, 25), True, '9/25');
  Test('(3/5)^3', Rat(27, 125), True, '27/125');
  Test('(3/5)^4', Rat(81, 625), True, '81/625');
  Test('(3/5)^5', Rat(243, 3125), True, '243/3125');
  Test('(3/5)^6', Rat(729, 15625), True, '729/15625');
  Test('(3/5)^7', Rat(2187, 78125), True, '2187/78125');

  Test('(3/5)^20', Rat(3486784401, 95367431640625), True, '3486784401/95367431640625');
  Test('(3/2)^20', Rat(3486784401, 1048576), True, '3486784401/1048576');

  Eps; Test('(3/5)^100/1E-23', 6.53318623500070906096690);
  Eps; Test('(3/2)^100', 4.06561177535215237397279707E17);

  Test('(3/5)^−0', Rat(1, 1), True, '1');
  Test('(3/5)^−1', Rat(5, 3), True, '5/3');
  Test('(3/5)^−2', Rat(25, 9), True, '25/9');
  Test('(3/5)^−3', Rat(125, 27), True, '125/27');
  Test('(3/5)^−4', Rat(625, 81), True, '625/81');
  Test('(3/5)^−5', Rat(3125, 243), True, '3125/243');
  Test('(3/5)^−6', Rat(15625, 729), True, '15625/729');
  Test('(3/5)^−7', Rat(78125, 2187), True, '78125/2187');

  Test('(3/5)^−20', Rat(95367431640625, 3486784401), True, '95367431640625/3486784401');
  Test('(3/2)^−20', Rat(1048576, 3486784401), True, '1048576/3486784401');

  Eps; Test('(3/5)^−100', 1.53064670748650634144452844E22);
  Eps; Test('(3/2)^−100/1E-18', 2.45965442657982926924379399);

  Eps; Test('(1/2)^(1/2)', 1/sqrt(2));
  Eps; Test('(2/3)^(1/2)', sqrt(2)/sqrt(3));
  Eps; Test('(3/5)^(3/2)', pow(3/5, 3/2));

  Eps; Test('(1/2)^(-1/2)', sqrt(2));
  Eps; Test('(2/3)^(-1/2)', sqrt(3)/sqrt(2));
  Eps; Test('(3/5)^(-3/2)', pow(5/3, 3/2));

  Test('(3/5)^π', pow(3/5, Pi));
  Test('(3/5)^−π', pow(5/3, Pi));

  Test('(-3/5)^0', Rat(1, 1), True, '1');
  Test('(-3/5)^1', Rat(-3, 5), True, '−3/5');
  Test('(-3/5)^2', Rat(9, 25), True, '9/25');
  Test('(-3/5)^3', Rat(-27, 125), True, '−27/125');
  Test('(-3/5)^4', Rat(81, 625), True, '81/625');
  Test('(-3/5)^5', Rat(-243, 3125), True, '−243/3125');

  Test('(-3/5)^-1', Rat(-5, 3), True, '−5/3');
  Test('(-3/5)^-2', Rat(25, 9), True, '25/9');
  Test('(-3/5)^-3', Rat(-125, 27), True, '−125/27');
  Test('(-3/5)^-4', Rat(625, 81), True, '625/81');
  Test('(-3/5)^-5', Rat(-3125, 243), True, '−3125/243');

  Test('(-3/5)^20', Rat(3486784401, 95367431640625), True, '3486784401/95367431640625');
  Test('(-3/5)^21', Rat(-10460353203, 476837158203125), True, '−10460353203/476837158203125');

  Test('(-3/5)^-20', Rat(95367431640625, 3486784401), True, '95367431640625/3486784401');
  Test('(-3/5)^-21', Rat(-476837158203125, 10460353203), True, '−476837158203125/10460353203');

  Eps; Test('(-3/5)^100/1E-23', 6.53318623500070906096690267);
  Eps; Test('(-3/5)^101/1E-23', -3.9199117410004254365801416);

  Eps; Test('(-3/5)^-100',  1.530646707486506341444528441E22);
  Eps; Test('(-3/5)^-101', -2.551077845810843902407547401E22);

  Test('(1/2 - 1/2)^0', failure, 'Zero raised to the power of zero.');
  Test('(1/2 - 1/2)^(1/3 - 1/3)', failure, 'Zero raised to the power of zero.');
  Test('(1/2 - 1/2)^0.0', failure, 'Zero raised to the power of zero.');

  Eps; Test('(-4⋅(1/2))^(1/2)', Sqrt(2)*ImaginaryUnit);
  Eps; Test('(-4⋅(1/2))^0.5', Sqrt(2)*ImaginaryUnit);
  Eps; Test('(-4⋅(1/2))^-(1/2)', -Sqrt(1/2)*ImaginaryUnit);
  Eps; Test('(-4⋅(1/2))^-0.5', -Sqrt(1/2)*ImaginaryUnit);

  Test('π^0', 1.0);
  Eps; Test('π^1', Pi);
  Eps; Test('π^2', Pi*Pi);
  Eps; Test('π^3', Pi*Pi*Pi);
  Eps; Test('π^4', Pi*Pi*Pi*Pi);
  Eps; Test('π^-1', 1/Pi);
  Eps; Test('π^-2', 1/(Pi*Pi));
  Eps; Test('π^-3', 1/(Pi*Pi*Pi));
  Eps; Test('π^-4', 1/(Pi*Pi*Pi*Pi));

  Eps; Test('π^100', 5.187848314319613192086261524E49);
  Eps; Test('π^1000', 1.412123544515764812310439732E497);
  Eps; Test('π^-100/1E-50', 1.92758141605601299595132717748);
  Eps; Test('π^-1000/1E-498', 7.081533367839375300201218389281);

  Test('π^(1/2)', sqrt(Pi));
  Test('π^(3/2)', Pow(Pi, 3/2));
  Test('π^(-1/2)', 1/sqrt(Pi));
  Test('π^(-3/2)', Pow(Pi, -3/2));

  Test('π^π', Pow(Pi, Pi));
  Test('π^-π', Pow(Pi, -Pi));

  Eps; Test('(-π)^1', -Pi);
  Eps; Test('(-π)^2', Pi*Pi);
  Eps; Test('(-π)^3', -Pi*Pi*Pi);
  Eps; Test('(-π)^4', Pi*Pi*Pi*Pi);
  Eps; Test('(-π)^-1', -1/Pi);
  Eps; Test('(-π)^-2', 1/(Pi*Pi));
  Eps; Test('(-π)^-3', -1/(Pi*Pi*Pi));
  Eps; Test('(-π)^-4', 1/(Pi*Pi*Pi*Pi));

  Eps; Test('(-π)^100',  5.187848314319613192086261524E49);
  Eps; Test('(-π)^101', -1.629810615220468935601560707E50);
  Eps; Test('(-π)^1000',  1.412123544515764812310439732E497);
  Eps; Test('(-π)^1001', -4.436316953411906095459701104E497);
  Eps; Test('(-π)^-100/1E-50',  1.92758141605601299595132717748);
  Eps; Test('(-π)^-101/1E-51', -6.13568221154779549256440908236);
  Eps; Test('(-π)^-1000/1E-498',  7.081533367839375300201218389281);
  Eps; Test('(-π)^-1001/1E-498', -2.254122080323667391310572893983);

  Test('0.0^0', failure, 'Zero raised to the power of zero.');
  Test('0.0^(1/2-1/2)', failure, 'Zero raised to the power of zero.');
  Test('0.0^0.0', failure, 'Zero raised to the power of zero.');

  Eps; Test('(-2.0)^(1/2)', Sqrt(2)*ImaginaryUnit);
  Eps; Test('(-2.0)^0.5', Sqrt(2)*ImaginaryUnit);
  Eps; Test('(-2.0)^-(1/2)', -Sqrt(1/2)*ImaginaryUnit);
  Eps; Test('(-2.0)^-0.5', -Sqrt(1/2)*ImaginaryUnit);

  Eps; Test('π^e', Pow(Pi, Exp(1)));
  Eps; Test('e^π', Pow(Exp(1), Pi));
  Eps; Test('π^−e', Pow(Pi, -Exp(1)));
  Eps; Test('e^−π', Pow(Exp(1), -Pi));

  Test('i^2', TASC(-1));
  Test('i^-1', -ImaginaryUnit);
  Test('i^i', TASC(Exp(-Pi/2)));
  Test('i^-i', TASC(Exp(Pi/2)));
  Test('2^i', Cos(Ln(2)) + Sin(Ln(2)) * ImaginaryUnit);
  Test('2^(1+i)', 2*Cos(Ln(2)) + 2*Sin(Ln(2)) * ImaginaryUnit);
  Eps; Test('(3 − 2⋅i)^(5 + 1.2⋅i)', 208.462241802416852908684580 - 1216.2148793127138136850052642*ImaginaryUnit);
  Eps; Test('(−3 − 2⋅i)^(5 + 1.2⋅i)', 3018.78826570156423301465273+12697.8912161357663918391804618*ImaginaryUnit);
  Eps; Test('(π + e⋅i)^(−e + π⋅i)', -0.001820817983213375294947968925 + 0.00126281697465301499742259442*ImaginaryUnit);
  Eps; Test('e^(π⋅i)', TASC(-1));
  Eps; Test('e^(π⋅i/2)', ImaginaryUnit);
  Eps; Test('e^(-π⋅i/2)', -ImaginaryUnit);
  Eps; Test('e^(i⋅0)', TASC(1));
  Eps; Test('(−π)^π', -32.913857741893878178261463794-15.6897116534331706413921410*ImaginaryUnit);
  Eps; Test('(−π)^i', 0.01785997172427576861039491204+0.039350529116087611295006524*ImaginaryUnit);
  Eps; Test('(−π)^(π + e⋅i)', 0.006524875122424854072989223+0.0028742584114429763486324591*ImaginaryUnit);

  Test('(0)^(i-i)', failure, 'Zero raised to a complex number with non-positive real part.');
  Test('(0.0)^(i-i)', failure, 'Zero raised to a complex number with non-positive real part.');
  Test('(1/2-1/2)^(i-i)', failure, 'Zero raised to a complex number with non-positive real part.');

  Test('(i-i)^(i-i)', failure, 'Zero raised to a complex number with non-positive real part.');
  Test('(i-i)^(1/2-1/2)', failure, 'Zero raised to a complex number with non-positive real part.');
  Test('(i-i)^0.0', failure, 'Zero raised to a complex number with non-positive real part.');
  Test('(i-i)^0', failure, 'Zero raised to a complex number with non-positive real part.');

  Test('2^3^2', 512);
  Test('2^3^2 = 2^(3^2)', true);

//  Test('2^-3^2', Rat(1, 512)); TODO: Wishlist

  // - matrix raised to an integer

  Test('A ≔ ❨❨1, 2❩, ❨3, 4❩❩', 2, [1, 2, 3, 4]);
  Test('A^0', 2, [1, 0, 0, 1]);
  Test('A^1', 2, [1, 2, 3, 4]);
  Test('A^2', 2, [7, 10, 15, 22]);
  Test('A^3', 2, [37, 54, 81, 118]);
  Test('A^4', 2, [199, 290, 435, 634]);
  Test('A^5', 2, [1069, 1558, 2337, 3406]);
  Test('A^6', 2, [5743, 8370, 12555, 18298]);
  Test('A^7', 2, [30853, 44966, 67449, 98302]);
  Test('A^8', 2, [165751, 241570, 362355, 528106]);
  Test('A^9', 2, [890461, 1297782, 1946673, 2837134]);
  Test('A^10', 2, [4783807, 6972050, 10458075, 15241882]);
  Eps; Test('A^-1', 2, [-2, 1, 1.5, -0.5]);
  Eps; Test('A^-2', 2, [5.5, -2.5, -3.75, 1.75]);
  Eps; Test('A^-3', 2, [-14.75, 6.75, 10.125, -4.625]);
  Eps; Test('A^-4', 2, [39.625, -18.125, -27.1875, 12.4375]);
  Eps; Test('A^-5', 2, [-106.4375, 48.6875, 73.03125, -33.40625]);
  Eps; Test('A^-6', 2, [285.90625, -130.78125, -196.171875, 89.734375]);
  Eps; Test('A^-7', 2, [-767.984375, 351.296875, 526.9453125, -241.0390625]);
  Eps; Test('A^-8', 2, [2062.9140625, -943.6328125, -1415.44921875, 647.46484375]);
  Eps; Test('A^-9', 2, [-5541.27734375, 2534.73046875, 3802.095703125, -1739.181640625]);
  Eps; Test('A^-10', 2, [14884.650390625, -6808.642578125, -10212.9638671875, 4671.6865234375]);

  Eps; Test('A^100', 2, [2.4777692297556471353E72, 3.6111680422104224175E72, 5.4167520633156336263E72, 7.8945212930712807616E72]);
  Eps; Test('A^-100', 2, [6.2276792135387631750E42, -2.8487092906833027974E42, -4.2730639360249541961E42, 1.9546152775138089789E42]);

  Test('IsIdentity(A ⋅ A^−1)', True);
  Test('IsIdentity(A^−1 ⋅ A)', True);
  Test('IsIdentity(A^4 ⋅ A^−4, 1E−10)', True);
  Test('IsIdentity(A^−4 ⋅ A^4, 1E−10)', True);
  Test('IsIdentity(A^10 ⋅ A^−10, 1E−6)', True);
  Test('IsIdentity(A^−10 ⋅ A^10, 1E−6)', True);

  Test('IsIdentity(IdentityMatrix(10)^10)', True);
  Test('IsIdentity(IdentityMatrix(10)^100)', True);
  Test('IsIdentity(IdentityMatrix(10)^1000)', True);

  Test('IsIdentity(IdentityMatrix(10)^−10)', True);
  Test('IsIdentity(IdentityMatrix(10)^−100)', True);
  Test('IsIdentity(IdentityMatrix(10)^−1000)', True);

  Test('A ≔ ❨❨1, 0❩, ❨0, 0❩❩', 2, [1, 0, 0, 0]);

  Test('A^0', 2, [1, 0, 0, 1]);
  Test('A^1', 2, [1, 0, 0, 0]);
  Test('A^2', 2, [1, 0, 0, 0]);
  Test('A^3', 2, [1, 0, 0, 0]);
  Test('A^4', 2, [1, 0, 0, 0]);
  Test('A^5', 2, [1, 0, 0, 0]);
  Test('A^6', 2, [1, 0, 0, 0]);
  Test('A^7', 2, [1, 0, 0, 0]);
  Test('A^8', 2, [1, 0, 0, 0]);
  Test('A^9', 2, [1, 0, 0, 0]);
  Test('A^10', 2, [1, 0, 0, 0]);
  Test('A^100', 2, [1, 0, 0, 0]);
  Test('A^1000', 2, [1, 0, 0, 0]);

  Test('A^−1', failure, 'Matrix is singular.');
  Test('A^−2', failure, 'Matrix is singular.');
  Test('A^−10', failure, 'Matrix is singular.');
  Test('A^−100', failure, 'Matrix is singular.');
  Test('A^−1000', failure, 'Matrix is singular.');

  Test('A ≔ ❨❨1, 2, 3❩, ❨4, 5, 6❩❩', 3, [1, 2, 3, 4, 5, 6]);

  Test('A^0', failure, 'Cannot compute power of non-square matrix.');
  Test('A^1', failure, 'Cannot compute power of non-square matrix.');
  Test('A^2', failure, 'Cannot compute power of non-square matrix.');
  Test('A^3', failure, 'Cannot compute power of non-square matrix.');
  Test('A^4', failure, 'Cannot compute power of non-square matrix.');
  Test('A^5', failure, 'Cannot compute power of non-square matrix.');
  Test('A^6', failure, 'Cannot compute power of non-square matrix.');
  Test('A^7', failure, 'Cannot compute power of non-square matrix.');
  Test('A^8', failure, 'Cannot compute power of non-square matrix.');
  Test('A^9', failure, 'Cannot compute power of non-square matrix.');
  Test('A^10', failure, 'Cannot compute power of non-square matrix.');
  Test('A^100', failure, 'Cannot compute power of non-square matrix.');
  Test('A^1000', failure, 'Cannot compute power of non-square matrix.');
  Test('A^-1', failure, 'Cannot compute power of non-square matrix.');
  Test('A^-2', failure, 'Cannot compute power of non-square matrix.');
  Test('A^-10', failure, 'Cannot compute power of non-square matrix.');
  Test('A^-100', failure, 'Cannot compute power of non-square matrix.');
  Test('A^-1000', failure, 'Cannot compute power of non-square matrix.');

  Test('delete(A)', success);

  Test('❨❨2❩❩^10', 1, [1024]);
  Test('❨❨2❩❩^-10', 1, [1/1024]);

  // - invalid operands

  Test('"test"^2', failure, 'Invalid function arguments.');
  Test('true^2', failure, 'Invalid function arguments.');
  Test('date()^2', failure, 'Invalid function arguments.');
  Test('(.;)^2', failure, 'Invalid function arguments.');
  Test('color("red")^2', failure, 'Invalid function arguments.');
  Test('2^true', failure, 'Invalid function arguments.');
  Test('2^"test"', failure, 'Invalid function arguments.');
  Test('2^date()', failure, 'Invalid function arguments.');
  Test('2^(.;)', failure, 'Invalid function arguments.');
  Test('2^time()', failure, 'Invalid function arguments.');
  Test('2^color("red")', failure, 'Invalid function arguments.');
  Test('true^false', failure, 'Invalid function arguments.');
  Test('date()^time()', failure, 'Invalid function arguments.');
  Test('"cat"^"dog"', failure, 'Invalid function arguments.');

  // Factorial

  Test('0!', 1);
  Test('1!', 1);
  Test('2!', 2);
  Test('3!', 6);
  Test('4!', 24);
  Test('5!', 120);
  Test('6!', 720);
  Test('7!', 5040);
  Test('8!', 40320);
  Test('9!', 362880);
  Test('10!', 3628800);
  Test('11!', 39916800);
  Test('12!', 479001600);
  Test('13!', 6227020800);
  Test('14!', 87178291200);
  Test('15!', 1307674368000);
  Test('16!', 20922789888000);
  Test('17!', 355687428096000);
  Test('18!', 6402373705728000);
  Test('19!', 121645100408832000);
  Test('20!', 2432902008176640000);

  Eps; Test('21!', 5.109094217170944E19);
  Eps; Test('22!', 1.12400072777760768E21);
  Eps; Test('23!', 2.585201673888497664E22);
  Eps; Test('24!', 6.2044840173323943936E23);
  Eps; Test('25!', 1.5511210043330985984E25);

  Eps; Test('50!', 3.04140932017133780436126081660E64);
  Eps; Test('100!', 9.33262154439441526816992388562E157);
  Eps; Test('200!', 7.88657867364790503552363213932E374);
  Eps; Test('300!', 3.06057512216440636035370461297E614);

  Test('−5!', -120);

  Test('(−5)!', failure, 'A non-negative integer was expected as argument 1, but "-5" was given.');

  Test('"test"!', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');
  Test('true!', failure, 'An object of type integer was expected as argument 1, but an object of type boolean was given.');
  Test('5.2!', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('i!', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');
  Test('❨1, 0❩!', failure, 'An object of type integer was expected as argument 1, but an object of type real vector was given.');
  Test('❨i, 0❩!', failure, 'An object of type integer was expected as argument 1, but an object of type complex vector was given.');
  Test('(1/2)!', failure, 'An object of type integer was expected as argument 1, but an object of type rational number was given.');
  Test('❨❨1, 0❩, ❨0, 1❩❩!', failure, 'An object of type integer was expected as argument 1, but an object of type real matrix was given.');
  Test('❨❨i, 0❩, ❨0, 1❩❩!', failure, 'An object of type integer was expected as argument 1, but an object of type complex matrix was given.');
  Test('color("red")!', failure, 'An object of type integer was expected as argument 1, but an object of type color was given.');
  Test('date()!', failure, 'An object of type integer was expected as argument 1, but an object of type structure of type "Date" was given.');

  // Degrees

  Test('0°', 0.0);
  Test('90°', Pi/2);
  Test('180°', Pi);
  Test('270°', 3*Pi/2);
  Test('360°', 2*Pi);

  Eps; Test('1°', 1 * Pi / 180);
  Eps; Test('36°', 36 * Pi / 180);
  Eps; Test('1.7°', 1.7 * Pi / 180);
  Eps; Test('(-1.7)°', -1.7 * Pi / 180);
  Eps; Test('-1.7°', -1.7 * Pi / 180);
  Eps; Test('-(1.7°)', -1.7 * Pi / 180);

  Eps; Test('(123/19)°', (123/19) * Pi / 180);

  Eps; Test('(123/19 + 5.2⋅i)°', (123/19 + 5.2*ImaginaryUnit) * Pi / 180);

  Test('ChrName("°")', 'DEGREE SIGN');

  Test('"test"°', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');
  Test('false°', failure, 'An object of type number was expected as argument 1, but an object of type boolean was given.');
  Test('color("red")°', failure, 'An object of type number was expected as argument 1, but an object of type color was given.');

  // Percent

  Test('0%', 0.0);
  Test('0.0%', 0.0);
  Test('(1/2-1/2)%', 0.0);
  Test('(i-i)%', TASC(0.0));
  Test('1%', 0.01);
  Test('50%', 0.5);
  Test('100%', 1.0);
  Eps; Test('120%', 1.2);
  Test('1000 %', 10.0);
  Test('1.7%', 0.017);
  Test('-1.7%', -0.017);
  Test('-(1.7%)', -0.017);
  Test('(-1.7)%', -0.017);
  Test('(1/2)%', 0.005);
  Test('(1+2⋅i)%', 0.01 + 0.02*ImaginaryUnit);
  Test('5%%', 0.0005);
  Test('(5%)%', 0.0005);

  Test('"test"%', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');
  Test('false%', failure, 'An object of type number was expected as argument 1, but an object of type boolean was given.');
  Test('color("red")%', failure, 'An object of type number was expected as argument 1, but an object of type color was given.');

  // Permille

  Test('0‰', 0.0);
  Test('0.0‰', 0.0);
  Test('(1/2-1/2)‰', 0.0);
  Test('(i-i)‰', TASC(0.0));
  Eps; Test('1‰', 0.001);
  Eps; Test('50‰', 0.05);
  Eps; Test('100‰', 0.1);
  Eps; Test('500‰', 0.5);
  Eps; Test('1000‰', 1.0);
  Eps; Test('120‰', 0.12);
  Eps; Test('10000 ‰', 10.0);
  Eps; Test('1.7‰', 0.0017);
  Eps; Test('-1.7‰', -0.0017);
  Eps; Test('-(1.7‰)', -0.0017);
  Eps; Test('(-1.7)‰', -0.0017);
  Eps; Test('(1/2)‰', 0.0005);
  Eps; Test('(1+2⋅i)‰', 0.001 + 0.002*ImaginaryUnit);
  Eps; Test('5‰‰', 0.000005);
  Eps; Test('(5‰)‰', 0.000005);
  Eps; Test('(5 % ‰ % ‰ % %)/1E-14', 5.0);

  Test('"test"‰', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');
  Test('false‰', failure, 'An object of type number was expected as argument 1, but an object of type boolean was given.');
  Test('color("red")‰', failure, 'An object of type number was expected as argument 1, but an object of type color was given.');

  // Inner (dot, scalar) product (bar)

  Test('❨1, 2, 3❩ | ❨5, 2, −1❩', 6.0, True, '6');
  Test('❨1, 0, 0❩ | ❨0, 1, 0❩', 0.0, True, '0');
  Test('❨1, 0, 0❩ | ❨1, 0, 0❩', 1.0, True, '1');

  Test('❨6, 4, 2❩ | ❨i, 3, 1❩', 14 - 6*ImaginaryUnit, True, '14 − 6⋅i');
  Test('❨1, 0, 0❩ | ❨2 + i, 4, 6❩', 2 - ImaginaryUnit, True, '2 − i');

  Test('❨2⋅i, 5, 3❩ | ❨1, 2, 1❩', 13 + 2*ImaginaryUnit, True, '13 + 2⋅i');
  Test('❨5+7⋅i, 2, 3❩ | ❨1, 0, 0❩', 5 + 7*ImaginaryUnit, True, '5 + 7⋅i');

  Test('❨3 + i, −2 − 5⋅i, 4 + 2⋅i❩ | ❨−5, −2 − 7⋅i, 3⋅i❩', 30 - 21*ImaginaryUnit, True, '30 − 21⋅i');
  Test('❨1, i, −1❩ | ❨1, i, −1❩', TASC(3.0), True, '3');

  Test('❨1, 0❩ | ❨0, 1❩', 0.0, True, '0');
  Test('❨1.0, 0❩ | ❨5.2, 1.3❩', 5.2, True, '5.2');

  Test('❨1, 0❩ | 7', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');
  Test('❨1, 0❩ | 7.8', failure, 'An object of type vector was expected as argument 2, but an object of type real number was given.');
  Test('❨1, 0❩ | (7/5)', failure, 'An object of type vector was expected as argument 2, but an object of type rational number was given.');
  Test('❨1, 0❩ | i', failure, 'An object of type vector was expected as argument 2, but an object of type complex number was given.');
  Test('❨1, 0❩ | "test"', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');
  Test('❨1, 0❩ | color("blue")', failure, 'An object of type vector was expected as argument 2, but an object of type color was given.');
  Test('5 | ❨1, 0❩', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');
  Test('IdentityMatrix(4) | ❨1, 0❩', failure, 'An object of type vector was expected as argument 1, but an object of type real matrix was given.');
  Test('IdentityMatrix(4) | "cat"', failure, 'An object of type vector was expected as argument 1, but an object of type real matrix was given.');
  Test('false | "cat"', failure, 'An object of type vector was expected as argument 1, but an object of type boolean was given.');

  Test('u, v, x ≔ ❨1, 0, 0❩, ❨0, 1, 0❩, ❨5, 7, 3❩; (x|u)⋅u + (x|v)⋅v', [5, 7, 0]);
  Test('delete(u); delete(v); delete(x);', null);

  // Cross

  // - vector cross product

  Test('❨1, 0, 0❩ × ❨0, 1, 0❩', [0, 0, 1]);
  Test('❨0, 1, 0❩ × ❨0, 0, 1❩', [1, 0, 0]);
  Test('❨0, 0, 1❩ × ❨1, 0, 0❩', [0, 1, 0]);

  Test('❨0, 1, 0❩ × ❨1, 0, 0❩', [0, 0, -1]);
  Test('❨1, 0, 0❩ × ❨1, 0, 0❩', [0, 0, 0]);

  Test('❨1, 2, 3❩ × ❨−4, 3, 2❩', [-5, -14, 11]);

  Test('❨1, 2, 3❩ × ❨0, 0, 0❩', [0, 0, 0]);
  Test('❨0, 0, 0❩ × ❨−4, 3, 2❩', [0, 0, 0]);
  Test('❨0, 0, 0❩ × ❨0, 0, 0❩', [0, 0, 0]);

  Test('❨1, 2❩ × ❨−4, 3❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨1, 2, 4❩ × ❨−4, 3❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨1, 2, 4❩ × ❨−4, 3, 5, 6❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨1, 2, 4, 2❩ × ❨−4, 3, 5, 6❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨1, 2, 4, 2❩ × ❨−4, 3, 5❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨1, 2❩ × ❨−4, 3, 5❩', failure, 'Vector cross product only defined in ℝ³.');

  Test('❨1, 0, 0❩ × ❨0, 1, 0❩ × ❨1, 0, 0❩', [0, 1, 0]);
  Test('❨1, 0, 0❩ × ❨0, 1, 0❩ × ❨1, 0, 0❩ = [❨1, 0, 0❩ × ❨0, 1, 0❩] × ❨1, 0, 0❩', True);

  Test('❨1, 2, 3❩ × ❨−3, 2, 1❩ × ❨3, 2, −1❩', [-6, 20, 22]);
  Test('❨1, 2, 3❩ × ❨−3, 2, 1❩ × ❨3, 2, −1❩ = [❨1, 2, 3❩ × ❨−3, 2, 1❩] × ❨3, 2, −1❩', True);

  Test('❨2, 1, 4❩ × ❨−3, 2, 4❩ × ❨−4, 8, 5❩ × ❨1, 1, 3❩', [88, 356, -148]);
  Test('❨2, 1, 4❩ × ❨−3, 2, 4❩ × ❨−4, 8, 5❩ × ❨1, 1, 3❩ = [[❨2, 1, 4❩ × ❨−3, 2, 4❩] × ❨−4, 8, 5❩] × ❨1, 1, 3❩', true);

  // - Cartesian product

  Test('{"cat", "dog"} × {"female", "male"} = {''("cat", "female"), ''("cat", "male"), ''("dog", "female"), ''("dog", "male")}', True);
  Test('{"a", "b"} × {1, 2, 3} = {''("a", 1), ''("a", 2), ''("a", 3), ''("b", 1), ''("b", 2), ''("b", 3)}', True);
  Test('{1, 2, 3, 4, 5} × {true} = {''(1, true), ''(2, true), ''(3, true), ''(4, true), ''(5, true)}', True);
  Test('{10} × {10} = {''(10, 10)}', True);

  Test('{1, 2, 3, 4, 5} × {} = {}', True);
  Test('∅ × {1, 2, 3} = ∅', True);
  Test('∅ × ∅ = ∅', True);

  Test('{ 1, 2, 3 } × { 1, 2, 3 } = { ''(1, 1), ''(1, 2), ''(1, 3), ''(2, 1), ''(2, 2), ''(2, 3), ''(3, 1), ''(3, 2), ''(3, 3) }', True);
  Test('{ 1, 2, 3 } × { 1., 2., 3. } = { ''(1, 1.), ''(1, 2.), ''(1, 3.), ''(2, 1.), ''(2, 2.), ''(2, 3.), ''(3, 1.), ''(3, 2.), ''(3, 3.) }', True);
  Test('{ "a", "b" } × { 1/2, 1/3 } = { ''("a", 1/2), ''("a", 1/3), ''("b", 1/2), ''("b", 1/3) }', True);

  Test('{"a"} × {1, 2} × {"b"} = { ''("a", 1, "b"), ''("a", 2, "b") }', True);
  Test('{1, 2} × {10, 20} × {100, 200} = { ''(1, 10, 100), ''(1, 10, 200), ''(1, 20, 100), ''(1, 20, 200), ''(2, 10, 100), ''(2, 10, 200), ''(2, 20, 100), ''(2, 20, 200) }', True);

  Test('{"a"} × {"b"} × {"c"} × {"d"} = { ''("a", "b", "c", "d" ) }', True);
  Test('{"a"} × {1, 2} × {"c"} × {"d"} = { ''("a", 1, "c", "d" ), ''("a", 2, "c", "d" ) }', True);
  Test('{"a"} × {1, 2, 3} × {"c"} × {"d"} = { ''("a", 1, "c", "d" ), ''("a", 2, "c", "d" ), ''("a", 3, "c", "d" ) }', True);
  Test('{"a"} × {} × {"c"} × {"d"} = {}', True);
  Test('{} × {} × {"c"} × {"d"} = {}', True);
  Test('{"a"} × {1, 2, 3} × {"c"} × {} = {}', True);
  Test('{"a"} × {1, 2, 3} × {"c"} × {true} = { ''("a", 1, "c", true), ''("a", 2, "c", true), ''("a", 3, "c", true) }', True);

  // - invalid operands

  Test('❨1, 2, 3❩ × 5', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');
  Test('❨1, 2, 3❩ × 5.0', failure, 'An object of type vector was expected as argument 2, but an object of type real number was given.');
  Test('❨1, 2, 3❩ × i', failure, 'An object of type vector was expected as argument 2, but an object of type complex number was given.');
  Test('❨1, 2, 3❩ × true', failure, 'An object of type vector was expected as argument 2, but an object of type boolean was given.');
  Test('❨1, 2, 3❩ × "5"', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');
  Test('❨1, 2, 3❩ × "5" × true', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');
  Test('❨1, 2, 3❩ × (1/2) × true', failure, 'An object of type vector was expected as argument 2, but an object of type rational number was given.');

  Test('4 × 8', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');

  Test('{4} × 8', failure, 'An object of type set was expected as argument 2, but an object of type integer was given.');
  Test('{4} × {8} × 16', failure, 'An object of type set was expected as argument 3, but an object of type integer was given.');
  Test('{4} × {8} × {16} × 32', failure, 'An object of type set was expected as argument 4, but an object of type integer was given.');
  Test('{4} × {8} × {16} × 32 × 64', failure, 'An object of type set was expected as argument 4, but an object of type integer was given.');
  Test('{4} × {8} × {16} × "test" × 64', failure, 'An object of type set was expected as argument 4, but an object of type string was given.');

  // Square root

  // - numbers

  Test('√0', 0.0);
  Test('√1', 1.0);
  Test('√4', 2.0);
  Test('√9', 3.0);
  Test('√16', 4.0);
  Test('√25', 5.0);
  Test('√36', 6.0);
  Test('√49', 7.0);
  Test('√64', 8.0);
  Test('√81', 9.0);
  Test('√100', 10.0);

  Test('√0.0', 0.0);
  Test('√1.0', 1.0);
  Test('√4.0', 2.0);
  Test('√9.0', 3.0);
  Test('√16.0', 4.0);
  Test('√25.0', 5.0);
  Test('√36.0', 6.0);
  Test('√49.0', 7.0);
  Test('√64.0', 8.0);
  Test('√81.0', 9.0);
  Test('√100.0', 10.0);

  Test('√(31/2 + 1/2)', 4.0);
  Test('√(1/2 - 1/2)', 0.0);

  Eps; Test('√1E100', 1E50);
  Eps; Test('√16E100', 4E50);

  Test('√-1', ImaginaryUnit);
  Test('√-16', 4*ImaginaryUnit);

  Test('√-16.0', 4*ImaginaryUnit);
  Test('√-(31/2 + 1/2)', 4.0*ImaginaryUnit);

  Eps; Test('√(2 + 3⋅i)', 1.67414922803554004044803930 + 0.8959774761298381247157337*ImaginaryUnit);

  Eps; Test('√i', (1 + ImaginaryUnit) / Sqrt(2));
  Test('√(i-i)', TASC(0.0));

  // - matrices

  Eps; Test('√❨❨2, 3❩, ❨3, 5❩❩', 2,
    [
      1, 1,
      1, 2
    ]);


  Eps; Test('√❨❨1269, -658, -392❩, ❨-658, 1556, 644❩, ❨-392, 644, 281❩❩', 3,
    [
      34, -8, -7,
      -8, 36, 14,
      -7, 14, 6
    ]);

  Eps; Test('√❨❨890, 671, -28, 432, 614❩, ❨671, 1521, -1818, 526, -1354❩, ❨-28, -1818, 3265, -130, 3137❩, ❨432, 526, -130, 2048, -548❩, ❨614, -1354, 3137, -548, 4326❩❩', 5,
    [
      23, 15, 0, 6, 10,
      15, 26, -22, 6, -10,
      0, -22, 44, 2, 29,
      6, 6, 2, 44, -6,
      10, -10, 29, -6, 57
    ]);

  Eps; Test('√❨❨10, -3, 2❩, ❨-3, 2, 1❩, ❨2, 1, 5❩❩', 3,
    [
      3.0351922193590166925, -0.76708302760149100904, 0.44630910847552380925,
      -0.76708302760149100904, 1.1147905109963295083, 0.41088410209969516454,
      0.44630910847552380925, 0.41088410209969516454, 2.1522040875190044117
    ]);

  Eps; Test('√❨❨0.29, 0.18, -0.2❩, ❨0.18, 1.08, 0.78❩, ❨-0.2, 0.78, 1.05❩❩', 3,
    [
      0.44392349078718205112, 0.20871384021848979468, -0.22219466066621268935,
      0.20871384021848979468, 0.90820831798149495435, 0.45999585221007641237,
      -0.22219466066621268935, 0.45999585221007641237, 0.88827549145574884252
    ]);

  Eps; Test('√❨❨1, 0, 0, 0, 0, 0, 0, 0, 0, 0❩, ❨0, 5, 0, 0, 0, 3, 0, 0, 0, 0❩,'+' ❨0, 0, 1, 0, 0, 0, 0, 0, 0, 0❩, ❨0, 0, 0, 1, 0, 0, 0, 0, 0, 0❩, ❨0, 0, 0, 0, 1, 0, 0, 0, 0, 0❩,'+' ❨0, 3, 0, 0, 0, 2, 0, 0, 0, 0❩, ❨0, 0, 0, 0, 0, 0, 1, 0, 0, 0❩,'+' ❨0, 0, 0, 0, 0, 0, 0, 1, 0, 0❩, ❨0, 0, 0, 0, 0, 0, 0, 0, 1, 0❩,'+' ❨0, 0, 0, 0, 0, 0, 0, 0, 0, 1❩❩', 10,
    [
      1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 2, 0, 0, 0, 1, 0, 0, 0, 0,
      0, 0, 1, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 1, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
      0, 1, 0, 0, 0, 1, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 1, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 1, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 1, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 1
    ]);

  Test('√IdentityMatrix(1) = IdentityMatrix(1)', True);
  Test('√IdentityMatrix(2) = IdentityMatrix(2)', True);
  Test('√IdentityMatrix(3) = IdentityMatrix(3)', True);
  Test('√IdentityMatrix(4) = IdentityMatrix(4)', True);
  Test('√IdentityMatrix(5) = IdentityMatrix(5)', True);
  Test('√IdentityMatrix(6) = IdentityMatrix(6)', True);
  Test('√IdentityMatrix(7) = IdentityMatrix(7)', True);
  Test('√IdentityMatrix(8) = IdentityMatrix(8)', True);
  Test('√IdentityMatrix(9) = IdentityMatrix(9)', True);
  Test('√IdentityMatrix(10) = IdentityMatrix(10)', True);

  Test('√ZeroMatrix(1) = ZeroMatrix(1)', True);
  Test('√ZeroMatrix(2) = ZeroMatrix(2)', True);
  Test('√ZeroMatrix(3) = ZeroMatrix(3)', True);
  Test('√ZeroMatrix(4) = ZeroMatrix(4)', True);
  Test('√ZeroMatrix(5) = ZeroMatrix(5)', True);
  Test('√ZeroMatrix(6) = ZeroMatrix(6)', True);
  Test('√ZeroMatrix(7) = ZeroMatrix(7)', True);
  Test('√ZeroMatrix(8) = ZeroMatrix(8)', True);
  Test('√ZeroMatrix(9) = ZeroMatrix(9)', True);
  Test('√ZeroMatrix(10) = ZeroMatrix(10)', True);

  Eps; Test('√❨❨50, 3+28⋅i, 18-27⋅i❩, ❨3-28⋅i, 18, -14-19⋅i❩, ❨18+27⋅i, -14+19⋅i, 46❩❩', 3,
    [
      6, 3*ImaginaryUnit, 1-2*ImaginaryUnit,
      -3*ImaginaryUnit, 2, -1-2*ImaginaryUnit,
      1+2*ImaginaryUnit, -1+2*ImaginaryUnit, 6
    ]);

  Test('√❨❨1, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩', failure, 'Matrix square root only defined for positive semidefinite symmetric matrices.');
  Test('√❨❨1, 2, 3❩, ❨4, 5, 6❩❩', failure, 'Matrix square root only defined for positive semidefinite symmetric matrices.');

  Test('√❨❨i, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩', failure, 'Matrix square root only defined for positive semidefinite Hermitian matrices.');
  Test('√❨❨i, 2, 3❩, ❨4, 5, 6❩❩', failure, 'Matrix square root only defined for positive semidefinite Hermitian matrices.');

  // - invalid operands

  Test('√"test"', failure, 'An object of type numeric entity was expected as argument 1, but an object of type string was given.');
  Test('√false', failure, 'An object of type numeric entity was expected as argument 1, but an object of type boolean was given.');
  Test('√color("red")', failure, 'An object of type numeric entity was expected as argument 1, but an object of type color was given.');
  Test('√❨1, 2, 3❩', failure, 'Invalid function arguments.');

  // Asterisk (complex conjugate and matrix conjugate transpose)

  // - numbers

  Test('0*', 0);
  Test('(1/2 - 1/2)*', Rat(0, 1));
  Test('0.*', 0.0);
  Test('(i-i)*', TASC(0));

  Test('5*', 5);
  Test('(-3)*', -3);
  Test('(1/3)*', Rat(1, 3));
  Test('(-7/8)*', Rat(-7, 8));
  Test('3.5*', 3.5);
  Test('(-3.5)*', -3.5);
  Test('1E100*', 1E100);
  Test('(-1E100)*', -1E100);

  Test('0**', 0);
  Test('(1/2 - 1/2)**', Rat(0, 1));
  Test('0.**', 0.0);
  Test('(i-i)**', TASC(0));

  Test('5**', 5);
  Test('(-3)**', -3);
  Test('(1/3)**', Rat(1, 3));
  Test('(-7/8)**', Rat(-7, 8));
  Test('3.5**', 3.5);
  Test('(-3.5)**', -3.5);
  Test('1E100**', 1E100);
  Test('(-1E100)**', -1E100);

  Test('i*', -ImaginaryUnit);
  Test('(-i)*', ImaginaryUnit);

  Test('(2+4⋅i)*', 2-4*ImaginaryUnit);
  Test('(2-4⋅i)*', 2+4*ImaginaryUnit);
  Test('(-2+4⋅i)*', -2-4*ImaginaryUnit);
  Test('(-2-4⋅i)*', -2+4*ImaginaryUnit);

  Test('(1.235 + 63.22⋅i)*', 1.235 - 63.22*ImaginaryUnit);
  Test('(1.235 - 63.22⋅i)*', 1.235 + 63.22*ImaginaryUnit);
  Test('(-1.235 + 63.22⋅i)*', -1.235 - 63.22*ImaginaryUnit);
  Test('(-1.235 - 63.22⋅i)*', -1.235 + 63.22*ImaginaryUnit);

  Test('(e+π⋅i)*', Exp(1) - Pi*ImaginaryUnit);
  Test('(e−π⋅i)*', Exp(1) + Pi*ImaginaryUnit);

  Test('i**', ImaginaryUnit);
  Test('(-i)**', -ImaginaryUnit);

  Test('(2+4⋅i)**', 2+4*ImaginaryUnit);
  Test('(2-4⋅i)**', 2-4*ImaginaryUnit);
  Test('(-2+4⋅i)**', -2+4*ImaginaryUnit);
  Test('(-2-4⋅i)**', -2-4*ImaginaryUnit);

  Test('(1.235 + 63.22⋅i)**', 1.235 + 63.22*ImaginaryUnit);
  Test('(1.235 - 63.22⋅i)**', 1.235 - 63.22*ImaginaryUnit);
  Test('(-1.235 + 63.22⋅i)**', -1.235 + 63.22*ImaginaryUnit);
  Test('(-1.235 - 63.22⋅i)**', -1.235 - 63.22*ImaginaryUnit);

  Test('(e+π⋅i)**', Exp(1) + Pi*ImaginaryUnit);
  Test('(e−π⋅i)**', Exp(1) - Pi*ImaginaryUnit);

  Test('i', +ImaginaryUnit);
  Test('i*', -ImaginaryUnit);
  Test('i**', +ImaginaryUnit);
  Test('i***', -ImaginaryUnit);
  Test('i****', +ImaginaryUnit);
  Test('i*****', -ImaginaryUnit);
  Test('i******', +ImaginaryUnit);
  Test('i*******', -ImaginaryUnit);
  Test('i********', +ImaginaryUnit);
  Test('i*********', -ImaginaryUnit);
  Test('i**********', +ImaginaryUnit);

  Test('i' + StringOfChar('*', 100), +ImaginaryUnit);
  Test('i' + StringOfChar('*', 101), -ImaginaryUnit);
  Test('i' + StringOfChar('*', 102), +ImaginaryUnit);
  Test('i' + StringOfChar('*', 103), -ImaginaryUnit);

  // - vectors (~ column matrices)

  Test('❨1, 2, 3❩*', 3, [1, 2, 3]);
  Test('❨−2, 0, i❩*', 3, [-2, 0, -ImaginaryUnit]);
  Test('❨2, 1 + 3⋅i, 5 - 2⋅i, -i❩*', 4, [2, 1 - 3*ImaginaryUnit, 5 + 2*ImaginaryUnit, ImaginaryUnit]);

  // - matrices

  Test('❨❨1, 2❩, ❨3, 4❩❩*', 2, [1, 3, 2, 4]);
  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩*', 3, [1, 4, 7, 2, 5, 8, 3, 6, 9]);
  Test('❨❨1, 2, 3, 4, 5❩, ❨6, 7, 8, 9, 10❩❩*', 2, [1, 6, 2, 7, 3, 8, 4, 9, 5, 10]);
  Test('❨❨1, 2❩, ❨−3, −4❩, ❨5, −6❩❩*', 3, [1, -3, 5, 2, -4, -6]);
  Test('❨❨7❩❩*', 1, [7]);
  Test('❨❨1, 2, 3, 4, −5❩❩*', 1, [1, 2, 3, 4, -5]);

  Test('A ≔ ❨❨1, 6, 4❩, ❨3, 2, −5❩, ❨12, 9, −9.5❩❩; B ≔ ❨❨1, 2, 3, −3, 0❩, ❨−6, 4, 5.6, 3, 1❩❩; C ≔ ❨❨6, 7❩, ❨2, 1❩, ❨−4, 5.6❩, ❨π, e^2❩❩; D ≔ ❨❨5, 4, 3, 2❩, ❨−5, 4, 7, 5.5❩, ❨2, 3, 0, 1❩, ❨−5, 6, 3, 2❩❩;', null);
  Test('A* = transpose(A) ∧ B* = transpose(B) ∧ C* = transpose(C) ∧ D* = transpose(D)', True);
  Test('A** = A ∧ B** = B ∧ C** = C ∧ D** = D', True);
  Test('delete(A); delete(B); delete(C); delete(D);', null);

  Test('IdentityMatrix(10)* = IdentityMatrix(10)', True);
  Test('IdentityMatrix(100)* = IdentityMatrix(100)', True);

  Test('ZeroMatrix(10)* = ZeroMatrix(10)', True);
  Test('ZeroMatrix(100)* = ZeroMatrix(100)', True);

  Test('❨❨2, i, 2+i❩, ❨−4, −2 − 3⋅i, 4❩, ❨0, 1, 1−2⋅i❩❩*', 3, [2, -4, 0, -ImaginaryUnit, -2 + 3*ImaginaryUnit, 1, 2 - ImaginaryUnit, 4, 1 + 2*ImaginaryUnit]);
  Test('❨❨1, i, 1+i, 3−2⋅i❩, ❨−4, −2−3⋅i, 0, −i❩❩*', 2, [1, -4, -ImaginaryUnit, -2 + 3*ImaginaryUnit, 1 - ImaginaryUnit, 0, 3 + 2*ImaginaryUnit, ImaginaryUnit]);
  Test('❨❨5 + i, 1 − i, 2 + 3⋅i❩, ❨−2, 0, i❩, ❨1 − 5⋅i, 0, 3❩❩*', 3, [5-ImaginaryUnit, -2, 1+5*ImaginaryUnit, 1+ImaginaryUnit, 0, 0, 2-3*ImaginaryUnit, -ImaginaryUnit, 3]);
  Test('❨❨i❩❩*', 1, [-ImaginaryUnit]);
  Test('❨❨-i❩❩*', 1, [ImaginaryUnit]);
  Test('❨❨1, 2, i, 1 + 2⋅i❩❩*', 1, [1, 2, -ImaginaryUnit, 1 - 2*ImaginaryUnit]);

  Test('[IdentityMatrix(10) + i⋅ZeroMatrix(10)]* = [IdentityMatrix(10) + i⋅ZeroMatrix(10)]', True);
  Test('[IdentityMatrix(100) + i⋅ZeroMatrix(100)]* = [IdentityMatrix(100) + i⋅ZeroMatrix(100)]', True);

  Test('A ≔ ❨❨2 + i, −3, 1 + 5⋅i❩, ❨0, −i, 4❩, ❨1 + 3⋅i, −5 + 3⋅i, 9❩❩; B ≔ ❨❨1, 2, i, 5, 2+7⋅i❩, ❨−3, 5, −2, −i, 1−i❩❩; C ≔ ❨❨2+i❩, ❨2−4⋅i❩, ❨π⋅i❩, ❨e+π⋅i❩❩; D ≔ ❨❨i, 2❩, ❨−2, i❩❩;', null);
  Test('A* = ConjugateTranspose(A) ∧ B* = ConjugateTranspose(B) ∧ C* = ConjugateTranspose(C) ∧ D* = ConjugateTranspose(D)', True);
  Test('A** = A ∧ B** = B ∧ C** = C ∧ D** = D', True);
  Test('A*** = ConjugateTranspose(A) ∧ B*** = ConjugateTranspose(B) ∧ C*** = ConjugateTranspose(C) ∧ D*** = ConjugateTranspose(D)', True);
  Test('A**** = A ∧ B**** = B ∧ C**** = C ∧ D**** = D', True);
  Test('delete(A); delete(B); delete(C); delete(D);', null);

  // - invalid operands

  Test('"test"*', failure, 'An object of type numeric entity was expected as argument 1, but an object of type string was given.');
  Test('false*', failure, 'An object of type numeric entity was expected as argument 1, but an object of type boolean was given.');
  Test('color("red")*', failure, 'An object of type numeric entity was expected as argument 1, but an object of type color was given.');

  // Colon (create structure member)

  Test('("a": 42) = struct("name": "a", "value": 42)', true);
  Test('type("a": 24)', 'structure of type "StructMember"');

  Test('("a": 42.0) = struct("name": "a", "value": 42.0)', True);
  Test('("a": 42/7) = struct("name": "a", "value": 42/7)', True);
  Test('("a": "cat") = struct("name": "a", "value": "cat")', True);
  Test('("a": true) = struct("name": "a", "value": true)', True);
  Test('("a": color("red")) = struct("name": "a", "value": color("red"))', True);

  Test('5: 4', failure, 'An object of type string was expected as argument 1, but an object of type integer was given.');
  Test('5.0: 4', failure, 'An object of type string was expected as argument 1, but an object of type real number was given.');
  Test('false: 4', failure, 'An object of type string was expected as argument 1, but an object of type boolean was given.');

  Test('struct("name": "Smith", "age": 50).name', 'Smith');
  Test('struct("name": "Smith", "age": 50).age', 50);

  // Semicolon (also tested in the section on syntax in this chapter)

  Test('42', 42);
  Test('42;', null);
  Test('1; 2', 2);
  Test('1; 2;', null);
  Test('1; 2; 3', 3);
  Test('1; 2; 3;', null);
  Test('1; 2; 3; 4', 4);
  Test('1; 2; 3; 4;', null);

  Test('"alfa"; "beta"', 'beta');
  Test('"alfa"; "beta";', null);

  Test('5; true; 6/7; color("red"); "test"', 'test');
  Test('5.4; "alfa"; color("blue"); false; 11/12', Rat(11, 12));
  Test('color("red"); 22/23; "fish"; true', True);
  Test('"rat"; false; 1/2; 22.3', 22.3);
  Test('22/23; false; color("red"); 10', 10);
  Test('true; 1; 2.3; "dog"; color("red")', TRGB.Create(1, 0, 0));

  Test('5; true; 6/7; color("red"); "test";', null);
  Test('5.4; "alfa"; color("blue"); false; 11/12;', null);
  Test('color("red"); 22/23; "fish"; true;', null);
  Test('"rat"; false; 1/2; 22.3;', null);
  Test('22/23; false; color("red"); 10;', null);
  Test('true; 1; 2.3; "dog"; color("red");', null);

  Test('a ≔ 4; b ≔ 10; a^2 + b^2', 116);
  Test('delete(a); delete(b)', success);
  Test('a ≔ 4; b ≔ 10; a^2 + b^2', 116);
  Test('delete(a); delete(b);', null);
  Test('f ≔ x ↦ x^2 + 1; f(10)', 101);
  Test('delete(f);', null);

  // Less than

  // - numbers

  Test('1 < 2', True);
  Test('1 < 3', True);
  Test('2 < 2', False);
  Test('2 < 3', True);
  Test('3 < 3', False);
  Test('4 < 3', False);
  Test('7 < 1561161515665', True);
  Test('1551118 < 156', False);
  Test('-18918919877171156 < 75', True);
  Test('-58952 < -25', True);
  Test('-1451891 < -89849456561', False);
  Test('-819448811198 < 89144545665', True);
  Test('167474488913 < -48974545415', False);
  Test('876127175165 < 876127175166', True);
  Test('876127175166 < 876127175166', False);
  Test('876127175167 < 876127175166', False);

  Test('3/7 < 4/7', True);
  Test('4/7 < 4/7', False);
  Test('5/7 < 4/7', False);
  Test('-5/7 < 4/7', True);

  Test('15615.0 < 54894894.756', True);
  Test('7563165.156 < 7563165.157', True);
  Test('7563165.157 < 7563165.157', False);
  Test('7563165.158 < 7563165.157', False);

  Test('5 < 5.8', True);
  Test('5 < 5.0', False);
  Test('5 < 5.1', True);
  Test('4.9 < 5', True);
  Test('5.0 < 5', False);
  Test('5.1 < 5', False);

  Test('0.4999 < 1/2', True);
  Test('0.5000 < 1/2', False);
  Test('0.5001 < 1/2', False);

  Test('1/2 + 1/2 < 1', False);
  Test('1/2 + 1/2 < 1.000000001', True);
  Test('1/2 + 1/2 < 2', True);

  Test('-992/3 < 7/8', True);

  Test('i < 5', failure, 'Complex numbers have no order.');
  Test('i < 5.5', failure, 'Complex numbers have no order.');
  Test('i < 5/2', failure, 'Complex numbers have no order.');
  Test('0 < i', failure, 'Complex numbers have no order.');
  Test('0.22 < i', failure, 'Complex numbers have no order.');
  Test('1/3 < i', failure, 'Complex numbers have no order.');

  Test('i < i', failure, 'Complex numbers have no order.');

  Test('2 + 3⋅i < 10 + 20⋅i', failure, 'Complex numbers have no order.');

  // - real vectors

  Test('❨2, 1, 4❩ < ❨5, 3, 5❩', True);
  Test('❨3, 1, 4❩ < ❨5, 3, 5❩', True);
  Test('❨4, 1, 4❩ < ❨5, 3, 5❩', True);
  Test('❨5, 1, 4❩ < ❨5, 3, 5❩', False);
  Test('❨6, 1, 4❩ < ❨5, 3, 5❩', False);
  Test('❨2, 2, 4❩ < ❨5, 3, 5❩', True);
  Test('❨2, 3, 4❩ < ❨5, 3, 5❩', False);
  Test('❨2, 4, 4❩ < ❨5, 3, 5❩', False);
  Test('❨2, 1, 5❩ < ❨5, 3, 5❩', False);
  Test('❨2, 1, 6❩ < ❨5, 3, 5❩', False);
  Test('❨2, 1, -7❩ < ❨5, 3, 5❩', True);
  Test('❨2, 1, -7❩ < ❨5, -3, 5❩', False);

  Test('❨4, 1❩ < ❨5, 3❩', True);
  Test('❨5, 1❩ < ❨5, 3❩', False);
  Test('❨-5, 1❩ < ❨5, 3❩', True);
  Test('❨4, 2❩ < ❨5, 3❩', True);
  Test('❨0, 0❩ < ❨5, 3❩', True);
  Test('❨-4, -2❩ < ❨5, 3❩', True);
  Test('-❨4, 2❩ < ❨5, 3❩', True);

  Test('❨2, 2, 4❩ < ❨5, 3❩', failure, 'Cannot subtract two vectors of unequal dimension.'); // OK, but leaks implementation

  // - real matrices

  Test('❨❨1, 2❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨2, 2❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨3, 2❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨4, 2❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨8, 2❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨9, 2❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨10, 2❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨11, 2❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨-11, 2❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);

  Test('❨❨1, 3❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 4❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 5❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨1, 6❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', False);

  Test('❨❨1, 2❩, ❨4, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨5, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨8, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨9, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨10, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨1, 2❩, ❨11, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨1, 2❩, ❨-11, 4❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);

  Test('❨❨1, 2❩, ❨3, 5❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨1, 2❩, ❨3, 6❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨1, 2❩, ❨3, -7❩❩ < ❨❨10, 5❩, ❨10, 5❩❩', True);

  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨2, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨3, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨4, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);
  Test('❨❨5, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);

  Test('❨❨1, 6, 4, 2❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);
  Test('❨❨1, 5, 5, 2❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);

  Test('❨❨1, 5, 4, 3❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 4❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);
  Test('❨❨1, 5, 4, 5❩, ❨0, −2, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);

  Test('❨❨1, 5, 4, 2❩, ❨0, −1, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 2❩, ❨0, 0, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);
  Test('❨❨1, 5, 4, 2❩, ❨0, 1, 4, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);

  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 6, 5❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);

  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 6❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 7❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 8❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);
  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 9❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);
  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, -9❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);

  Test('❨❨1, 2❩, ❨3, 4❩❩ < ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', failure, 'Cannot subtract two matrices of different sizes.');

  // - datetime objects: postponed

  // - invalid operands

  Test('"test" < 7', failure, 'Invalid function arguments.');
  Test('7 < false', failure, 'Invalid function arguments.');
  Test('"LiU rocks" < true', failure, 'Invalid function arguments.'); // could make it return True
  Test('❨2, 1, 4❩ < ❨3, 2, i❩', failure, 'Invalid function arguments.');
  Test('❨❨1, 3❩, ❨3, 4❩❩ < ❨❨10, 5❩, ❨10, i❩❩', failure, 'Invalid function arguments.');
  Test('❨2, 1, 4❩ < ❨❨1, 5, 4, 2❩, ❨0, −1, 4, 5❩❩', failure, 'Invalid function arguments.');
  Test('❨2, 1, 4❩ < 7', failure, 'Invalid function arguments.');
  Test('❨❨1, 2❩, ❨9, 4❩❩ < 7', failure, 'Invalid function arguments.');
  Test('"test" < "testing"', failure, 'Invalid function arguments.');

  // - n-ary extension

  Test('1 < 2 < 3', True);
  Test('1 < 2 < 3 < 4', True);
  Test('1 < 2 < 3 < 4 < 5', True);
  Test('1 < 2 < 3 < 4 < 5 < 6 < 7 < 8 < 9 < 10 < 100 < 1000 < 1E10 < 2E10 < 1E100', True);
  Test('-5 < 0 < 5', True);
  Test('1 < 2 < 2.01 < 4', True);

  Test('1 < 2 < 2 < 4', False);
  Test('1 < 2 < 3 < 2', False);
  Test('1 < -2 < 3 < 4', False);
  Test('1 < 2 < 3 < -4', False);
  Test('1 < 2 < 3 < 4 < 5 < 6 < 7 < 8 < 9 < 10 < 100 < 1000 < 1E10 < 2E10 < 1E-100', False);
  Test('1 < 2 < 3 < 4 < 5 < 6 < 7 < 8 < 9 < 10 < 100 < 1000 < 1E10 < -2E10 < 1E100', False);
  Test('1 < 2 < 3 < 4 < 5 < 6 < 7 < 8 < 9 < 10 < 100 < 1000 < 1E10 < 1E10 < 1E100', False);
  Test('1 < 2 < 1.99 < 4', False);
  Test('-1 < 2 < 1.9999 < 4', False);


  // Less than or equal to

  // - numbers

  Test('1 ≤ 2', True);
  Test('1 ≤ 3', True);
  Test('2 ≤ 2', True);
  Test('2 ≤ 3', True);
  Test('3 ≤ 3', True);
  Test('4 ≤ 3', False);
  Test('7 ≤ 1561161515665', True);
  Test('1551118 ≤ 156', False);
  Test('-18918919877171156 ≤ 75', True);
  Test('-58952 ≤ -25', True);
  Test('-1451891 ≤ -89849456561', False);
  Test('-819448811198 ≤ 89144545665', True);
  Test('167474488913 ≤ -48974545415', False);
  Test('876127175165 ≤ 876127175166', True);
  Test('876127175166 ≤ 876127175166', True);
  Test('876127175167 ≤ 876127175166', False);

  Test('3/7 ≤ 4/7', True);
  Test('4/7 ≤ 4/7', True);
  Test('5/7 ≤ 4/7', False);
  Test('-5/7 ≤ 4/7', True);

  Test('15615.0 ≤ 54894894.756', True);
  Test('7563165.156 ≤ 7563165.157', True);
  Test('7563165.157 ≤ 7563165.157', True);
  Test('7563165.158 ≤ 7563165.157', False);

  Test('5 ≤ 5.8', True);
  Test('5 ≤ 5.0', True);
  Test('5 ≤ 5.1', True);
  Test('4.9 ≤ 5', True);
  Test('5.0 ≤ 5', True);
  Test('5.1 ≤ 5', False);

  Test('0.4999 ≤ 1/2', True);
  Test('0.5000 ≤ 1/2', True);
  Test('0.5001 ≤ 1/2', False);

  Test('1/2 + 1/2 ≤ 0.999999999', False);
  Test('1/2 + 1/2 ≤ 1', True);
  Test('1/2 + 1/2 ≤ 1.000000001', True);
  Test('1/2 + 1/2 ≤ 2', True);

  Test('-992/3 ≤ 7/8', True);

  Test('i ≤ 5', failure, 'Complex numbers have no order.');
  Test('i ≤ 5.5', failure, 'Complex numbers have no order.');
  Test('i ≤ 5/2', failure, 'Complex numbers have no order.');
  Test('0 ≤ i', failure, 'Complex numbers have no order.');
  Test('0.22 ≤ i', failure, 'Complex numbers have no order.');
  Test('1/3 ≤ i', failure, 'Complex numbers have no order.');

  Test('i ≤ i', failure, 'Complex numbers have no order.');

  Test('2 + 3⋅i ≤ 10 + 20⋅i', failure, 'Complex numbers have no order.');

  // - real vectors

  Test('❨2, 1, 4❩ ≤ ❨5, 3, 5❩', True);
  Test('❨3, 1, 4❩ ≤ ❨5, 3, 5❩', True);
  Test('❨4, 1, 4❩ ≤ ❨5, 3, 5❩', True);
  Test('❨5, 1, 4❩ ≤ ❨5, 3, 5❩', True);
  Test('❨6, 1, 4❩ ≤ ❨5, 3, 5❩', False);
  Test('❨2, 2, 4❩ ≤ ❨5, 3, 5❩', True);
  Test('❨2, 3, 4❩ ≤ ❨5, 3, 5❩', True);
  Test('❨2, 4, 4❩ ≤ ❨5, 3, 5❩', False);
  Test('❨2, 1, 5❩ ≤ ❨5, 3, 5❩', True);
  Test('❨2, 1, 6❩ ≤ ❨5, 3, 5❩', False);
  Test('❨2, 1, -7❩ ≤ ❨5, 3, 5❩', True);
  Test('❨2, 1, -7❩ ≤ ❨5, -3, 5❩', False);

  Test('❨4, 1❩ ≤ ❨5, 3❩', True);
  Test('❨5, 1❩ ≤ ❨5, 3❩', True);
  Test('❨6, 1❩ ≤ ❨5, 3❩', False);
  Test('❨-6, 1❩ ≤ ❨5, 3❩', True);
  Test('❨4, 2❩ ≤ ❨5, 3❩', True);
  Test('❨0, 0❩ ≤ ❨5, 3❩', True);
  Test('❨-4, -2❩ ≤ ❨5, 3❩', True);
  Test('-❨4, 2❩ ≤ ❨5, 3❩', True);

  Test('❨2, 2, 4❩ ≤ ❨5, 3❩', failure, 'Cannot subtract two vectors of unequal dimension.'); // OK, but leaks implementation

  // - real matrices

  Test('❨❨1, 2❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨2, 2❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨3, 2❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨4, 2❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨8, 2❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨9, 2❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨10, 2❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨11, 2❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨-11, 2❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);

  Test('❨❨1, 3❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 4❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 5❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 6❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', False);

  Test('❨❨1, 2❩, ❨4, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨5, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨8, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨9, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨10, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨11, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨1, 2❩, ❨-11, 4❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);

  Test('❨❨1, 2❩, ❨3, 5❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);
  Test('❨❨1, 2❩, ❨3, 6❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', False);
  Test('❨❨1, 2❩, ❨3, -7❩❩ ≤ ❨❨10, 5❩, ❨10, 5❩❩', True);

  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨2, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨3, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨4, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨5, 5, 4, 2❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);

  Test('❨❨1, 6, 4, 2❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 5, 2❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);

  Test('❨❨1, 5, 4, 3❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 4❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 5❩, ❨0, −2, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);

  Test('❨❨1, 5, 4, 2❩, ❨0, −1, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 2❩, ❨0, 0, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 2❩, ❨0, 1, 4, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);

  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 6, 5❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);

  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 6❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 7❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 8❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);
  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, 9❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', False);
  Test('❨❨1, 5, 4, 2❩, ❨0, −2, 4, -9❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', True);

  Test('❨❨1, 2❩, ❨3, 4❩❩ ≤ ❨❨4, 6, 5, 4❩, ❨2, 0, 5, 8❩❩', failure, 'Cannot subtract two matrices of different sizes.');

  // - datetime objects: postponed

  // - invalid operands

  Test('"test" ≤ 7', failure, 'Invalid function arguments.');
  Test('7 ≤ false', failure, 'Invalid function arguments.');
  Test('"LiU rocks" ≤ true', failure, 'Invalid function arguments.'); // could make it return True
  Test('❨2, 1, 4❩ ≤ ❨3, 2, i❩', failure, 'Invalid function arguments.');
  Test('❨❨1, 3❩, ❨3, 4❩❩ ≤ ❨❨10, 5❩, ❨10, i❩❩', failure, 'Invalid function arguments.');
  Test('❨2, 1, 4❩ ≤ ❨❨1, 5, 4, 2❩, ❨0, −1, 4, 5❩❩', failure, 'Invalid function arguments.');
  Test('❨2, 1, 4❩ ≤ 7', failure, 'Invalid function arguments.');
  Test('❨❨1, 2❩, ❨9, 4❩❩ ≤ 7', failure, 'Invalid function arguments.');
  Test('"test" ≤ "testing"', failure, 'Invalid function arguments.');

  // - n-ary extension

  Test('1 ≤ 2 ≤ 3', True);
  Test('1 ≤ 1 ≤ 3', True);
  Test('1 ≤ 1 ≤ 1', True);
  Test('1 ≤ 2 ≤ 3 ≤ 4', True);
  Test('1 ≤ 2 ≤ 3 ≤ 3', True);
  Test('1 ≤ 2 ≤ 2 ≤ 4', True);
  Test('2 ≤ 2 ≤ 2 ≤ 2', True);
  Test('1 ≤ 2 ≤ 3 ≤ 4 ≤ 5 ≤ 6 ≤ 7 ≤ 8 ≤ 9 ≤ 10', True);
  Test('1 ≤ 2 ≤ 3 ≤ 4 ≤ 5 ≤ 5 ≤ 7 ≤ 8 ≤ 9 ≤ 10', True);
  Test('1 ≤ 2 ≤ 3 ≤ 4 ≤ 5 ≤ 6 ≤ 7 ≤ 8 ≤ 9 ≤ 10 ≤ 1E10 ≤ 1E100', True);
  Test('1 ≤ 2 ≤ 3 ≤ 4 ≤ 5 ≤ 6 ≤ 7 ≤ 8 ≤ 9 ≤ 10 ≤ 1E10 ≤ 1E100 ≤ 1E100', True);
  Test('-1 ≤ 2 ≤ 3 ≤ 4', True);

  Test('-1 ≤ -2 ≤ 3 ≤ 4', False);

  Test('1 ≤ 2 ≤ 1.9', False);
  Test('1 ≤ 1 ≤ -3', False);
  Test('1 ≤ 0.9 ≤ 1', False);
  Test('1 ≤ -2 ≤ 3 ≤ 4', False);
  Test('1 ≤ 2 ≤ -3 ≤ 3', False);
  Test('1 ≤ 2 ≤ 2 ≤ -4', False);
  Test('2 ≤ 2 ≤ 2 ≤ 1.5', False);
  Test('1 ≤ 2 ≤ 3 ≤ 4 ≤ 5 ≤ -6 ≤ 7 ≤ 8 ≤ 9 ≤ 10', False);
  Test('1 ≤ 2 ≤ 3 ≤ 4 ≤ 5 ≤ 5 ≤ 7 ≤ 8 ≤ 9 ≤ 8', False);
  Test('1 ≤ 2 ≤ 3 ≤ 4 ≤ 5 ≤ 6 ≤ 7 ≤ 8 ≤ 7 ≤ 10 ≤ 1E10 ≤ 1E100', False);
  Test('1 ≤ 2 ≤ 3 ≤ 4 ≤ 5 ≤ 6 ≤ 7 ≤ 8 ≤ 9 ≤ 10 ≤ 1E10 ≤ 1E-100 ≤ 1E100', False);
  Test('-1 ≤ 2 ≤ 3 ≤ -4', False);

  // Greater than

  // - numbers

  Test('5 > 7', False);
  Test('6 > 7', False);
  Test('7 > 7', False);
  Test('8 > 7', True);
  Test('9 > 7', True);
  Test('-5 > 7', False);
  Test('0 > 7', False);

  Test('-5 > -3', False);
  Test('-4 > -3', False);
  Test('-3 > -3', False);
  Test('-2 > -3', True);
  Test('-1 > -3', True);
  Test('0 > -3', True);
  Test('1 > -3', True);

  Test('0 > 0', False);

  Test('71176761671531055 > 71176761671531057', False);
  Test('71176761671531056 > 71176761671531057', False);
  Test('71176761671531057 > 71176761671531057', False);
  Test('71176761671531058 > 71176761671531057', True);
  Test('71176761671531059 > 71176761671531057', True);

  Test('71176761671531055 > -71176761671531057', True);
  Test('-71176761671531055 > -71176761671531057', True);
  Test('-71176761671531056 > -71176761671531057', True);
  Test('-71176761671531057 > -71176761671531057', False);
  Test('-71176761671531058 > -71176761671531057', False);
  Test('-71176761671531059 > -71176761671531057', False);

  Test('7/11 > 8/11', False);
  Test('8/11 > 8/11', False);
  Test('9/11 > 8/11', True);

  Test('16.95 > 16.97', False);
  Test('16.96 > 16.97', False);
  Test('16.97 > 16.97', False);
  Test('16.98 > 16.97', True);
  Test('16.99 > 16.97', True);
  Test('17.00 > 16.97', True);

  Test('5 > 14/3', True);
  Test('5 > 14/3 + 1/3', False);
  Test('5 > 16/3', False);

  Test('17/3 > 6', False);
  Test('17/3 + 1/3 > 6', False);
  Test('19/3 > 6', True);

  Test('-5/9 > -1', True);
  Test('-5/9 - 4/9 > -1', False);
  Test('-15/9 > -1', False);

  Test('5 > 4.9', True);
  Test('5 > 5.0', False);
  Test('5 > 5.1', False);

  Test('4.8 > 5', False);
  Test('4.9 > 5', False);
  Test('5.0 > 5', False);
  Test('5.1 > 5', True);
  Test('5.2 > 5', True);
  Test('4.8 > -5', True);

  Test('1/2 > 0.49', True);
  Test('1/2 > 0.50', False);
  Test('1/2 > 0.51', False);
  Test('1/2 > -0.51', True);

  Test('2.0 > 13/7', True);
  Test('2.0 > 13/7 + 1/7', False);
  Test('2.0 > 15/7', False);
  Test('2.0 > -15/7', True);
  Test('-2.0 > -15/7', True);

  Test('18234/9117319 > 0.002', False);
  Test('18235/9117319 > 0.002', True);

  Test('18234/9117319 > 2/1000', False);
  Test('18235/9117319 > 2/1000', True);

  Test('18234/9117319 > 2‰', False);
  Test('18235/9117319 > 2‰', True);

  Test('18234/9117319 > -2‰', True);
  Test('18235/9117319 > -2‰', True);

  Test('i > 8', failure, 'Complex numbers have no order.');
  Test('5 > i', failure, 'Complex numbers have no order.');
  Test('3 + 5⋅i > 4 + 3⋅i', failure, 'Complex numbers have no order.');
  Test('i > 8/9', failure, 'Complex numbers have no order.');
  Test('i > 8.9', failure, 'Complex numbers have no order.');
  Test('9999E99 > i', failure, 'Complex numbers have no order.');
  Test('i > 0', failure, 'Complex numbers have no order.');
  Test('i > i', failure, 'Complex numbers have no order.');

  // - real vectors

  Test('❨5, 7, 3, 2❩ > ❨2, 0, 2, −2❩', True);
  Test('❨5, 7, 3, 2❩ > ❨3, 0, 2, −2❩', True);
  Test('❨5, 7, 3, 2❩ > ❨4, 0, 2, −2❩', True);
  Test('❨5, 7, 3, 2❩ > ❨5, 0, 2, −2❩', False);
  Test('❨5, 7, 3, 2❩ > ❨6, 0, 2, −2❩', False);
  Test('❨5, 7, 3, 2❩ > ❨7, 0, 2, −2❩', False);

  Test('❨5, 7, 3, -2❩ > ❨2, 0, 2, −2❩', False);

  Test('❨5, 7, 3, -2❩ > ❨2, 0, 2❩', failure, 'Cannot subtract two vectors of unequal dimension.');

  // - vector is positive

  Test('❨1, 2❩ > 0', True);
  Test('❨5, 7, 3, 2❩ > 0', True);
  Test('❨5, 7, 3, 2, 2, 7❩ > 0', True);
  Test('❨5, 7, 3, 2, 2, 7, 5, 7, 22.1, 1E10❩ > 0', True);

  Test('❨1, 2, 0❩ > 0', False);
  Test('❨5, 7, -3, 2❩ > 0', False);
  Test('❨5, -7, 3, 2, 2, 7❩ > 0', False);
  Test('❨5, 7, 3, 2, 2, 0, 7, 5, 7, 22.1, 1E10❩ > 0', False);

  Test('❨-1, 2❩ > 0', False);
  Test('❨-5, -7, 3, 2❩ > 0', False);
  Test('❨0, 5, 7, 3, 2, 2, 7❩ > 0', False);
  Test('❨5, 7, 3, 2, 2, 7, 5, 7, 22.1, -1E10❩ > 0', False);

  Test('❨1, 0❩ > 0', False);
  Test('❨5, 0, -3, 2❩ > 0', False);
  Test('❨5, 7, 3, -2, 2, 7❩ > 0', False);
  Test('❨5, 7, 3, 2, 0, 7, 5, 7, 22.1, 1E10❩ > 0', False);

  // - real matrices

  Test('❨❨5, 7❩, ❨3, −5❩❩ > ❨❨2, 4❩, ❨0, −10❩❩', True);
  Test('❨❨5, 7❩, ❨3, −5❩❩ > ❨❨3, 4❩, ❨0, −10❩❩', True);
  Test('❨❨5, 7❩, ❨3, −5❩❩ > ❨❨4, 4❩, ❨0, −10❩❩', True);
  Test('❨❨5, 7❩, ❨3, −5❩❩ > ❨❨5, 4❩, ❨0, −10❩❩', False);
  Test('❨❨5, 7❩, ❨3, −5❩❩ > ❨❨6, 4❩, ❨0, −10❩❩', False);
  Test('❨❨5, 7❩, ❨3, −5❩❩ > ❨❨7, 4❩, ❨0, −10❩❩', False);

  Test('❨❨5, 7, 7❩, ❨3, −5, 2❩❩ > ❨❨7, 4❩, ❨0, −10❩❩', failure, 'Cannot subtract two matrices of different sizes.');

  // - matrix is positive

  Test('❨❨5, 7❩, ❨3, 5❩❩ > 0', True);

  Test('❨❨5, 0❩, ❨3, 5❩❩ > 0', False);
  Test('❨❨5, 7❩, ❨3, 0❩❩ > 0', False);
  Test('❨❨5, -7❩, ❨3, 5❩❩ > 0', False);
  Test('❨❨5, 7❩, ❨3, -5❩❩ > 0', False);
  Test('❨❨5, -7❩, ❨0, 5❩❩ > 0', False);

  Test('IdentityMatrix(100) > 0', False);

  // - datetime objects: postponed

  // - invalid operands

  Test('"test" > 7', failure, 'Invalid function arguments.');
  Test('7 > false', failure, 'Invalid function arguments.');
  Test('"LiU rocks" > true', failure, 'Invalid function arguments.'); // could make it return False
  Test('❨2, 1, 4❩ > ❨3, 2, i❩', failure, 'Invalid function arguments.');
  Test('❨❨1, 3❩, ❨3, 4❩❩ > ❨❨10, 5❩, ❨10, i❩❩', failure, 'Invalid function arguments.');
  Test('❨2, 1, 4❩ > ❨❨1, 5, 4, 2❩, ❨0, −1, 4, 5❩❩', failure, 'Invalid function arguments.');
  Test('❨2, 1, 4❩ > 7', failure, 'Invalid function arguments.');
  Test('❨❨1, 2❩, ❨9, 4❩❩ > 7', failure, 'Invalid function arguments.');
  Test('"test" > "testing"', failure, 'Invalid function arguments.');

  // - n-ary extension

  Test('3 > 2 > 1', True);
  Test('4 > 3 > 2 > 1', True);
  Test('4 > 3.2 > 2 > 1', True);
  Test('5 > 4 > 3 > 2 > 1', True);
  Test('5 > 4 > 3 > 2 > 1 > 0 > -1 > -2', True);
  Test('10 > 9 > 8 > 7 > 6 > 5 > 4 > 3 > 2 > 1', True);
  Test('1E100 > 1E10 > 10 > 9 > 8 > 7 > 6 > 5 > 4 > 3 > 2 > 1', True);

  Test('3 > 2 > 2', False);
  Test('4 > 3 > 1 > 1', False);
  Test('3.2 > 3.2 > 3.2 > 1', False);
  Test('3 > -2 > 1', False);
  Test('4 > 3 > 2 > 3', False);
  Test('4 > 4.1 > 2 > 1', False);
  Test('5 > 4 > 3 > -2 > 1', False);
  Test('5 > 4 > 3 > 2 > -1 > 0 > -1 > -2', False);
  Test('10 > 9 > 8 > 7 > 6 > 5 > -4.5 > 3 > 2 > 1', False);
  Test('1E100 > 1E-10 > 10 > 9 > 8 > 7 > 6 > 5 > 4 > 3 > 2 > 1', False);

  // Greater than or equal to

  // - numbers

  Test('5 ≥ 7', False);
  Test('6 ≥ 7', False);
  Test('7 ≥ 7', True);
  Test('8 ≥ 7', True);
  Test('9 ≥ 7', True);
  Test('-5 ≥ 7', False);
  Test('0 ≥ 7', False);

  Test('-5 ≥ -3', False);
  Test('-4 ≥ -3', False);
  Test('-3 ≥ -3', True);
  Test('-2 ≥ -3', True);
  Test('-1 ≥ -3', True);
  Test('0 ≥ -3', True);
  Test('1 ≥ -3', True);

  Test('0 ≥ 0', True);

  Test('71176761671531055 ≥ 71176761671531057', False);
  Test('71176761671531056 ≥ 71176761671531057', False);
  Test('71176761671531057 ≥ 71176761671531057', True);
  Test('71176761671531058 ≥ 71176761671531057', True);
  Test('71176761671531059 ≥ 71176761671531057', True);

  Test('71176761671531055 ≥ -71176761671531057', True);
  Test('-71176761671531055 ≥ -71176761671531057', True);
  Test('-71176761671531056 ≥ -71176761671531057', True);
  Test('-71176761671531057 ≥ -71176761671531057', True);
  Test('-71176761671531058 ≥ -71176761671531057', False);
  Test('-71176761671531059 ≥ -71176761671531057', False);

  Test('7/11 ≥ 8/11', False);
  Test('8/11 ≥ 8/11', True);
  Test('9/11 ≥ 8/11', True);

  Test('16.95 ≥ 16.97', False);
  Test('16.96 ≥ 16.97', False);
  Test('16.97 ≥ 16.97', True);
  Test('16.98 ≥ 16.97', True);
  Test('16.99 ≥ 16.97', True);
  Test('17.00 ≥ 16.97', True);

  Test('5 ≥ 14/3', True);
  Test('5 ≥ 14/3 + 1/3', True);
  Test('5 ≥ 16/3', False);

  Test('17/3 ≥ 6', False);
  Test('17/3 + 1/3 ≥ 6', True);
  Test('19/3 ≥ 6', True);

  Test('-5/9 ≥ -1', True);
  Test('-5/9 - 4/9 ≥ -1', True);
  Test('-15/9 ≥ -1', False);

  Test('5 ≥ 4.9', True);
  Test('5 ≥ 5.0', True);
  Test('5 ≥ 5.1', False);

  Test('4.8 ≥ 5', False);
  Test('4.9 ≥ 5', False);
  Test('5.0 ≥ 5', True);
  Test('5.1 ≥ 5', True);
  Test('5.2 ≥ 5', True);
  Test('4.8 ≥ -5', True);

  Test('1/2 ≥ 0.49', True);
  Test('1/2 ≥ 0.50', True);
  Test('1/2 ≥ 0.51', False);
  Test('1/2 ≥ -0.51', True);

  Test('2.0 ≥ 13/7', True);
  Test('2.0 ≥ 13/7 + 1/7', True);
  Test('2.0 ≥ 15/7', False);
  Test('2.0 ≥ -15/7', True);
  Test('-2.0 ≥ -15/7', True);

  Test('18234/9117319 ≥ 0.002', False);
  Test('18235/9117319 ≥ 0.002', True);

  Test('18234/9117319 ≥ 2/1000', False);
  Test('18235/9117319 ≥ 2/1000', True);

  Test('18234/9117319 ≥ 2‰', False);
  Test('18235/9117319 ≥ 2‰', True);

  Test('18234/9117319 ≥ -2‰', True);
  Test('18235/9117319 ≥ -2‰', True);

  Test('i ≥ 8', failure, 'Complex numbers have no order.');
  Test('5 ≥ i', failure, 'Complex numbers have no order.');
  Test('3 + 5⋅i ≥ 4 + 3⋅i', failure, 'Complex numbers have no order.');
  Test('i ≥ 8/9', failure, 'Complex numbers have no order.');
  Test('i ≥ 8.9', failure, 'Complex numbers have no order.');
  Test('9999E99 ≥ i', failure, 'Complex numbers have no order.');
  Test('i ≥ 0', failure, 'Complex numbers have no order.');
  Test('i ≥ i', failure, 'Complex numbers have no order.');

  // - real vectors

  Test('❨5, 7, 3, 2❩ ≥ ❨2, 0, 2, −2❩', True);
  Test('❨5, 7, 3, 2❩ ≥ ❨3, 0, 2, −2❩', True);
  Test('❨5, 7, 3, 2❩ ≥ ❨4, 0, 2, −2❩', True);
  Test('❨5, 7, 3, 2❩ ≥ ❨5, 0, 2, −2❩', True);
  Test('❨5, 7, 3, 2❩ ≥ ❨6, 0, 2, −2❩', False);
  Test('❨5, 7, 3, 2❩ ≥ ❨7, 0, 2, −2❩', False);

  Test('❨5, 7, 3, -2❩ ≥ ❨2, 0, 2, −2❩', True);

  Test('❨5, 7, 3, -2❩ ≥ ❨2, 0, 2❩', failure, 'Cannot subtract two vectors of unequal dimension.');

  // - vector is non-negative

  Test('❨1, 2❩ ≥ 0', True);
  Test('❨5, 7, 3, 2❩ ≥ 0', True);
  Test('❨5, 7, 3, 2, 2, 7❩ ≥ 0', True);
  Test('❨5, 7, 3, 2, 2, 7, 5, 7, 22.1, 1E10❩ ≥ 0', True);

  Test('❨1, 2, 0❩ ≥ 0', True);
  Test('❨5, 7, -3, 2❩ ≥ 0', False);
  Test('❨5, -7, 3, 2, 2, 7❩ ≥ 0', False);
  Test('❨5, 7, 3, 2, 2, 0, 7, 5, 7, 22.1, 1E10❩ ≥ 0', True);

  Test('❨-1, 2❩ ≥ 0', False);
  Test('❨-5, -7, 3, 2❩ ≥ 0', False);
  Test('❨0, 5, 7, 3, 2, 2, 7❩ ≥ 0', True);
  Test('❨5, 7, 3, 2, 2, 7, 5, 7, 22.1, -1E10❩ ≥ 0', False);

  Test('❨1, 0❩ ≥ 0', True);
  Test('❨5, 0, -3, 2❩ ≥ 0', False);
  Test('❨5, 7, 3, -2, 2, 7❩ ≥ 0', False);
  Test('❨5, 7, 3, 2, 0, 7, 5, 7, 22.1, 1E10❩ ≥ 0', True);

  // - real matrices

  Test('❨❨5, 7❩, ❨3, −5❩❩ ≥ ❨❨2, 4❩, ❨0, −10❩❩', True);
  Test('❨❨5, 7❩, ❨3, −5❩❩ ≥ ❨❨3, 4❩, ❨0, −10❩❩', True);
  Test('❨❨5, 7❩, ❨3, −5❩❩ ≥ ❨❨4, 4❩, ❨0, −10❩❩', True);
  Test('❨❨5, 7❩, ❨3, −5❩❩ ≥ ❨❨5, 4❩, ❨0, −10❩❩', True);
  Test('❨❨5, 7❩, ❨3, −5❩❩ ≥ ❨❨6, 4❩, ❨0, −10❩❩', False);
  Test('❨❨5, 7❩, ❨3, −5❩❩ ≥ ❨❨7, 4❩, ❨0, −10❩❩', False);

  Test('❨❨5, 7, 7❩, ❨3, −5, 2❩❩ ≥ ❨❨7, 4❩, ❨0, −10❩❩', failure, 'Cannot subtract two matrices of different sizes.');

  // - matrix is non-negative

  Test('❨❨5, 7❩, ❨3, 5❩❩ ≥ 0', True);

  Test('❨❨5, 0❩, ❨3, 5❩❩ ≥ 0', True);
  Test('❨❨5, 7❩, ❨3, 0❩❩ ≥ 0', True);
  Test('❨❨5, -7❩, ❨3, 5❩❩ ≥ 0', False);
  Test('❨❨5, 7❩, ❨3, -5❩❩ ≥ 0', False);
  Test('❨❨5, -7❩, ❨0, 5❩❩ ≥ 0', False);

  Test('IdentityMatrix(100) ≥ 0', True);

  // - datetime objects: postponed

  // - invalid operands

  Test('"test" ≥ 7', failure, 'Invalid function arguments.');
  Test('7 ≥ false', failure, 'Invalid function arguments.');
  Test('"LiU rocks" ≥ true', failure, 'Invalid function arguments.'); // could make it return False
  Test('❨2, 1, 4❩ ≥ ❨3, 2, i❩', failure, 'Invalid function arguments.');
  Test('❨❨1, 3❩, ❨3, 4❩❩ ≥ ❨❨10, 5❩, ❨10, i❩❩', failure, 'Invalid function arguments.');
  Test('❨2, 1, 4❩ ≥ ❨❨1, 5, 4, 2❩, ❨0, −1, 4, 5❩❩', failure, 'Invalid function arguments.');
  Test('❨2, 1, 4❩ ≥ 7', failure, 'Invalid function arguments.');
  Test('❨❨1, 2❩, ❨9, 4❩❩ ≥ 7', failure, 'Invalid function arguments.');
  Test('"test" ≥ "testing"', failure, 'Invalid function arguments.');

  // - n-ary extension

  Test('5 ≥ 4 ≥ 3', True);
  Test('5 ≥ 4 ≥ 3 ≥ 2 ≥ 1', True);
  Test('5 ≥ 4 ≥ 3 ≥ 2 ≥ 1 ≥ 0 ≥ -1', True);
  Test('5 ≥ 5 ≥ 3', True);
  Test('5 ≥ 4 ≥ 4 ≥ 4 ≥ 1', True);
  Test('5 ≥ 4 ≥ 3 ≥ 2 ≥ 1 ≥ 1 ≥ -1', True);
  Test('5 ≥ 5 ≥ 5', True);
  Test('5 ≥ 4 ≥ 3 ≥ 2 ≥ 1', True);
  Test('4 ≥ 4 ≥ 4 ≥ 2 ≥ 1 ≥ 0 ≥ 0', True);
  Test('5 ≥ 4.3 ≥ 3', True);
  Test('5 ≥ 4.3 ≥ 4.3 ≥ 2 ≥ 1', True);
  Test('5 ≥ 4 ≥ 3 ≥ 2 ≥ 1.999 ≥ 0 ≥ -1', True);
  Test('1E10 ≥ 5 ≥ 4 ≥ 3 ≥ 2 ≥ 1.999 ≥ 0 ≥ -1', True);

  Test('5 ≥ 4 ≥ 5', False);
  Test('5 ≥ 4 ≥ 4.1 ≥ 2 ≥ 1', False);
  Test('5 ≥ 8 ≥ 3 ≥ 2 ≥ 1 ≥ 0 ≥ -1', False);
  Test('5 ≥ 5 ≥ 6', False);
  Test('5 ≥ 4 ≥ -4 ≥ 4 ≥ 1', False);
  Test('-5 ≥ 4 ≥ 3 ≥ 2 ≥ 1 ≥ 1 ≥ -1', False);
  Test('5 ≥ 5 ≥ 5.000000001', False);
  Test('5 ≥ 4 ≥ -3 ≥ 2 ≥ 1', False);
  Test('4 ≥ 4 ≥ 4.001 ≥ 2 ≥ 1 ≥ 0 ≥ 0', False);
  Test('-5 ≥ 4.3 ≥ 3', False);
  Test('5 ≥ 4.3 ≥ 4.31 ≥ 2 ≥ 1', False);
  Test('5 ≥ 4 ≥ 3 ≥ -2 ≥ 1.999 ≥ 0 ≥ -1', False);
  Test('1E-10 ≥ 5 ≥ 4 ≥ 3 ≥ 2 ≥ 1.999 ≥ 0 ≥ -1', False);

  // Equals

  Test('1 + 1 = 2', True);
  Test('1 + 1/2 + 0.5 + i - i = 2', True);
  Test('1 + 1/2 + 0.5 + i - i = 2.0', True);
  Test('1 + 1/2 + 0.5 + i - i = 1/2 + 3/2', True);
  Test('1 + 1/2 + 0.5 + i - i = 2 + i - i', True);

  Test('1 = 0', False);
  Test('1 = 1', True);
  Test('1 = 2', False);

  Test('123456789123455 = 123456789123456', False);
  Test('123456789123456 = 123456789123456', True);
  Test('123456789123457 = 123456789123456', False);

  Test('123456789123456 = -123456789123456', False);

  Test('0 = 0', True);
  Test('0 = -0', True);

  Test('52697415 = 52697415', True);
  Test('-52697415 = -52697415', True);
  Test('-52697415 = 52697415', False);
  Test('52697416 = 52697415', False);

  Test('84012636 = 8490516', False);
  Test('84012636 = -8490516', False);

  Test('123456789123456789 = 123456789123456789', True);
  Test('123456789123456789 = 123456789123456788', False);
  Test('123456789123456789 = -123456789123456789', False);

  Test('42 = 40 + 2', True);
  Test('42 = 40 + 3', False);
  Test('42 = 40 - 2', False);

  Test('100 + 23 = 100 + 20 + 3', True);

  Test('75520864896 = 75520864895', False);
  Test('75520864896 = 75520864896', True);
  Test('75520864896 = 75520864897', False);
  Test('75520864896 = 0', False);

  Test('17/2 = 17/2', True);
  Test('17/2 = 19/2', False);
  Test('17/2 = 17/3', False);
  Test('17/2 = 15/2 + 1/2 + 1/2', True);
  Test('17/2 = 15/2 + 1/2 + 3/2', False);

  Test('-17/2 = 17/2', False);
  Test('17/2 = -17/2', False);
  Test('-17/2 = -17/2', True);
  Test('15/2 = 17/2', False);

  Test('189561/9849133 = 189561/9849133', True);
  Test('189561/9849133 = 189562/9849133', False);
  Test('189561/9849133 = 189561/9849132', False);
  Test('189561/9849133 = -189561/9849133', False);
  Test('189562/9849133 = 189561/9849134', False);
  Test('-189561/9849133 = -189561/9849133', True);

  Test('189561/9849133 = 189560/9849133 + 1/9849133', True);

  Test('189561/9849133 = 1/2', False);
  Test('189561/9849133 = -1/2', False);
  Test('189561/9849133 = 0', False);
  Test('189561/9849133 = 18916518/1898105687', False);
  Test('189561/9849133 = 9849133/189561', False);

  Test('1/2 - 1/2 = 1/3 - 1/3', True);
  Test('1/99 - 1/99 = 7/5 - 7/5', True);
  Test('1/2 - 1/2 = 1/3 - 1/3', True);

  Test('1/21 − 38/51 + 83/119 = 523/517 − 325/17 + 92/3 − 331186/26367', True);

  Test('1/21 − 38/51 + 83/119 = 523/517 − 325/17 + 92/3 − 331187/26367', False);
  Test('1/21 − 38/51 + 83/119 = 523/516 − 325/17 + 92/3 − 331186/26367', False);
  Test('1/21 − 38/51 + 83/119 = 523/517 − 325/17 + 92/3 + 331186/26367', False);
  Test('2/21 − 38/51 + 83/119 = 523/517 − 325/17 + 92/3 − 331186/26367', False);
  Test('1/21 − 38/51 + 83/120 = 523/517 − 325/17 + 92/3 − 331186/26367', False);
  Test('0 = 523/517 − 325/17 + 92/3 − 331186/26367', True);
  Test('1/21 − 38/51 + 83/119 = 0', True);

  Test('123.456 = 123.456', True);
  Test('123.456 = 123.457', False);
  Test('123.456 = 123.4560000001', False);
  Test('123.456000000001 = 123.456', False);
  Test('123.456 = -123.456', False);
  Test('123.456 = 123.0', False);

  Test('-123.456 = -123.456', True);
  Test('-123.456 = -123.556', False);
  Test('-123.456 = -123.4559999999', False);

  Test('123E100 = 123E100', True);

  Test('123E100 = 123E101', False);
  Test('123E100 = 123.000000001E100', False);
  Test('122.99999999999E100 = 123E100', False);
  Test('123.0E100 = 123E100', True);
  Test('-123E100 = -123E100', True);
  Test('123E100 = -123E100', False);
  Test('123E100 = 123E200', False);
  Test('123E100 = 0', False);

  Test('189.5613E200 = 189.5613E200', True);
  Test('189.5613E200 = 189.5614E200', False);
  Test('189.5613E200 = 189.5613000001E200', False);
  Test('189.5612999999E200 = 189.5613E200', False);
  Test('-189.5613E200 = -189.5613E200', True);
  Test('189.5613E200 = 189.5613E-200', False);
  Test('189.5613E-200 = 189.5613E-200', True);
  Test('-189.5613E-200 = -189.5613E-200', True);

  Test('189.5613E-200 = 189.5613E-200', True);
  Test('189.5614E-200 = 189.5613E-200', False);
  Test('189.561300001E-200 = 189.5613E-200', False);
  Test('189.5613E-200 = 189.5612999999E-200', False);

  Test('π = 3.1415926535797932384626', False);

  Test('e = exp(1)', True);
  Test('e = exp(1.01)', False);

  Test('e = π', False);
  Test('0⋅e = 0⋅π', True);

  Test('3 + 5⋅i = 3 + 5⋅i', True);
  Test('3 + 5⋅i = 3 + 6⋅i', False);
  Test('3 + 5⋅i = 3 + 5.0000000000001⋅i', False);
  Test('3 + 5⋅i = 2.999999999999 + 5⋅i', False);
  Test('3 + 5⋅i = 3', False);
  Test('3 + 5⋅i = 5⋅i', False);
  Test('3 + 5⋅i = 3 − 5⋅i', False);
  Test('3 − 5⋅i = 3 − 5⋅i', True);
  Test('3 + 5⋅i = 3.00000000001 + 5⋅i', False);
  Test('3.0000000001 + 5⋅i = 3 + 5⋅i', False);

  Test('i = i', True);
  Test('i = -i', False);
  Test('1/i = -i', True);

  Test('1/7 + 6/7 = 1', True);
  Test('2/7 + 6/7 = 1', False);

  Test('2 = 12/9 + 6/9', True);
  Test('2 = 12/9 + 7/9', False);

  Test('1/2 + 1/2 = 1', True);
  Test('1/2 + 1/2 = 3/2 - 1/2', True);
  Test('1/2 + 1/2 = 1.0', True);
  Test('1/2 + 1/2 = 1.0 + i - i', True);

  Test('1/2 = 0.5', True);
  Test('1/2 = 0.5 + i - i', True);
  Test('0.75 = 3/4', True);
  Test('-0.75 = -3/4', True);
  Test('0.75 + i - i = 3/4', True);
  Test('0.75 = 3/4 + i - i', True);

  Test('0 = 0.0', True);
  Test('0 = 1/2 - 1/2', True);
  Test('0 = i - i', True);
  Test('1/2 - 1/2 = 0.0', True);
  Test('1/2 - 1/2 = i - i', True);
  Test('0.0 = i - i', True);

  Test('5 + 0 = 5 + 0.0', True);
  Test('5 + 0 = 5 + 1/2 - 1/2', True);
  Test('5 + 0 = 5 + i - i', True);
  Test('5 + 1/2 - 1/2 = 5 + 0.0', True);
  Test('5 + 1/2 - 1/2 = 5 + i - i', True);
  Test('5 + 0.0 = 5 + i - i', True);

  Test('6 + 0 = 5 + 0.0', False);
  Test('6 + 0 = 5 + 1/2 - 1/2', False);
  Test('6 + 0 = 5 + i - i', False);
  Test('6 + 1/2 - 1/2 = 5 + 0.0', False);
  Test('6 + 1/2 - 1/2 = 5 + i - i', False);
  Test('6 + 0.0 = 5 + i - i', False);

  Test('0 = (.;)', False);
  Test('0.0 = (.;)', False);
  Test('1/2-1/2 = (.;)', False);
  Test('i-i = (.;)', False);

  Test('1/7 + 58/2 − 63/7 + 25/7 = 166/7', True);
  Test('1/7 + 58/2 − 63/7 + 25/7 − 5/7 = 23', True);
  Test('1/7 + 58/2 − 63/7 + 25/7 − 5/7 = 23.0', True);
  Test('1/7 + 58/2 − 63/7 + 25/7 − 5/7 = 23.0 + i − i', True);
  Test('i + 1/7 + 58/2 − 63/7 + 25/7 − 5/7 = 23.0 + i', True);

  Test('59/10 = 5.9', True);
  Test('59/100 = 0.59', True);
  Test('59/1000 = 0.059', True);

  Test('5⋅i − 3⋅i − 2⋅i + 7/2 = 7/2', True);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 = 3.5', True);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 = 3.5 + 0⋅i', True);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 = 3.5 + 2⋅i − 2⋅i', True);

  Test('5⋅i − 3⋅i − 2⋅i + 7/2 = 7/3', False);
  Test('5⋅i − 3⋅i − 3⋅i + 7/2 = 3.5', False);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 = 3.6 + 0⋅i', False);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 = 3.5 + 2⋅i − 2.2⋅i', False);

  Test('1/4 = 0.25', True);
  Test('4⋅0.25 = 1', True);
  Test('4⋅0.25 = 1.0', True);
  Test('4⋅0.25 = 3/4 + 1/4', True);
  Test('4⋅0.25 = 3/4 + 1/4 + i - i', True);
  Test('i - i = 1/7 + 2/7 - 3/7', True);
  Test('i - i = 1/7 + 2/7 - 3.00000001/7', False);

  Test('1/4 = 0.24', False);
  Test('4⋅0.25000000001 = 1', False);
  Test('4⋅0.25 = 0.999999999', False);
  Test('4⋅0.25 = 3/4 + 2/4', False);
  Test('4⋅0.25 = 3/4 + 1/4 - i', False);

  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i = 0', True);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i = 0.0', True);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i = 1/2 - 1/2', True);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i = i - i', True);

  Test('6 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i = 0', False);
  Test('5 + 3⋅i + 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i = 0.0', False);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 3⋅i + 1 − 10⋅i = 1/2 - 1/2', False);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i = 7 - i', False);

  Test('5 = ❨5❩', False);
  Test('5 = ❨❨5❩❩', False);
  Test('❨5❩ = ❨❨5❩❩', False);

  Test('0 = ❨0❩', False);
  Test('0 = ❨❨0❩❩', False);
  Test('❨0❩ = ❨❨0❩❩', False);

  Test('0.0 = ❨0❩', False);
  Test('0.0 = ❨❨0❩❩', False);

  Test('1/2-1/2 = ❨0❩', False);
  Test('1/2-1/2 = ❨❨0❩❩', False);
  Test('❨0❩ = ❨❨0❩❩', False);

  Test('i-i = ❨0❩', False);
  Test('i-i = ❨❨0❩❩', False);
  Test('❨0❩ = ❨❨0❩❩', False);

  Test('0 = ❨i-i❩', False);
  Test('0 = ❨❨i-i❩❩', False);
  Test('❨0❩ = ❨❨i-i❩❩', False);

  Test('i-i = ❨i-i❩', False);
  Test('i-i = ❨❨i-i❩❩', False);
  Test('❨i-i❩ = ❨❨0❩❩', False);

  Test('❨i-i❩ = ❨❨i-i❩❩', False);

  Test('❨1, 2, 3, 4, 5❩ = ❨1, 2, 3, 4, 5❩', True);

  Test('❨2, 2, 3, 4, 5❩ = ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, -2, 3, 4, 5❩ = ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3.000000000001, 4, 5❩ = ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3, 3.9999999999, 5❩ = ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3, 4, 500❩ = ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3, 4❩ = ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3, 4, 5❩ = ❨1, 2, 3❩', False);

  Test('❨π, e, sin(1)❩ = ❨π, e, sin(1)❩', True);

  Test('❨5 + i, 2, −i❩ = ❨5 + i, 2, −i❩', True);
  Test('❨5 + i, 2, −i❩ = ❨5 + i, 2, i❩', False);
  Test('❨5 + i, 0, −i❩ = ❨5 + i, 2, −i❩', False);

  Test('❨1, 2, 3, 4, 5❩ = ❨1, 2, 3, 4, 5 + i - i❩', True);
  Test('❨1 + i - i, 2, 3, 4, 5❩ = ❨1, 2, 3, 4, 5❩', True);

  Test('❨1, 2, 3❩ = ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', True);

  Test('❨2, 2, 3❩ = ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, -2, 3❩ = ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, 2, 3.0000001❩ = ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, 2, 3❩ = ❨1, 1, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, 2, 3❩ = ❨1, i, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, 2, 3❩ = ❨1, 1 + i, 2❩ + ❨0, 1, 1❩', False);

  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);

  Test('❨❨5, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, 2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1.0000000000001❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨3.9999999999, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0.1, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2.5❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4.000000001, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, 2, -1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, -1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, -4, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9.2❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨40, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2❩, ❨4, 0❩, ❨−1, 0❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨4, −2, 1❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = 4', False);
  Test('❨❨4❩❩ = ❨4❩', False);
  Test('❨❨4❩❩ = 4', False);
  Test('❨4❩ = 4', False);
  Test('4 = ❨4❩', False);

  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, i-i, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, i-i, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4 + i-i, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, i-i, 9❩, ❨−1, 0, 2❩❩ = ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, i-i, 2❩❩', True);

  Test('❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩ = ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', True);
  Test('❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩ = ❨❨2+i, 2−3⋅i❩, ❨i, i-i❩❩', True);

  Test('❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩ = ❨❨2+i, 2−3⋅i❩, ❨i, 1❩❩', False);
  Test('❨❨2+i, 2−3⋅i❩, ❨i, 2❩❩ = ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨2+i, 2−4⋅i❩, ❨i, 0❩❩ = ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨2+i, 2+3⋅i❩, ❨i, 0❩❩ = ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨2-i, 2−3⋅i❩, ❨i, 0❩❩ = ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨6+i, 2−3⋅i❩, ❨i, 0❩❩ = ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨2+i, 2−3⋅i❩, ❨i, 6❩❩ = ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);

  Test('❨❨2, 6❩, ❨−5, 3❩❩ = ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', True);

  Test('❨❨3, 6❩, ❨−5, 3❩❩ = ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 7❩, ❨−5, 3❩❩ = ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨5, 3❩❩ = ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨−5, 3.0000000001❩❩ = ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ = ❨❨1, 3 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ = ❨❨1, 2 + 3⋅i❩, ❨5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ = ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4❩, ❨3, 3❩❩', False);

  Test('"alpha" = "alpha"', True);
  Test('"alpha" = "beta"', False);
  Test('"alpha" = ""', False);
  Test('"" = ""', True);

  Test('"7" = 7', False);
  Test('"7" = 7.0', False);
  Test('"7" = 7+i-i', False);

  Test('true = true', True);
  Test('true = false', False);
  Test('false = true', False);
  Test('false = false', True);

  Test('false = 0', False);
  Test('false = 0.0', False);
  Test('false = 1/2 - 1/2', False);
  Test('false = i - i', False);
  Test('false = (.;)', False);

  Test('true = 1', False);
  Test('true = 1.0', False);

  Test('true = "true"', False);
  Test('"false" = false', False);
  Test('"false" = ""', False);
  Test('"false" = "false"', True);

  Test('"LiU rocks" = true', False);

  Test('❨❨2, 6❩, ❨−5, 3❩❩ = 5', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ = "alpha"', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ = true', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ = 5.0', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ = 0', False);

  Test('❨❨0, 0❩, ❨0, 0❩❩ = 0', False);
  Test('❨0, 0❩ = 0', False);

  Test('struct("a": 42, "b": 100) = struct("a": 42, "b": 100)', True);

  Test('struct("a": 42.0000001, "b": 100) = struct("a": 42, "b": 100)', False);
  Test('struct("A": 42, "b": 100) = struct("a": 42, "b": 100)', False);
  Test('struct("a": 43, "b": 100) = struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "c": 100) = struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 101) = struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100) = struct("d": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100) = struct("a": 420, "b": 100)', False);
  Test('struct("a": 42, "b": 100) = struct("a": 42, "e": 100)', False);
  Test('struct("a": 42, "b": 100) = struct("a": 42, "b": 100.01)', False);
  Test('struct("a": 42, "b": 100) = struct("a": 42)', False);
  Test('struct("a": 42) = struct("a": 42, "b": 100)', False);
  Test('struct("b": 100) = struct("a": 42, "b": 100)', False);
  Test('struct("b": 100, "a": 42) = struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100) = struct("a": 100, "b": 42)', False);
  Test('struct("a": 42, "b": 100) = struct("a": 42, "b": 100, "c": 1000)', False);
  Test('struct("a": 42, "b": 100, "c": 1000) = struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100, "c": 1000) = struct("a": 42, "b": 100, "c": 1000)', True);
  Test('struct("a": 42, "b": 100, "c": "1000") = struct("a": 42, "b": 100, "c": 1000)', False);
  Test('struct("a": 42, "b": 100, "c": "1000") = struct("a": 42, "b": 100, "c": "1000")', True);
  Test('struct("a": 42, "b": 100, "c": "1000") = struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100, "c": "1000") = struct("a": 42)', False);
  Test('struct("a": 42, "b": 100) = struct("a": 42, "b": 100, "c": "1000")', False);
  Test('struct("a": 42) = struct("a": 42, "b": 100, "c": "1000")', False);

  Test('struct("a": 42, "b": 100) = 42', False);
  Test('struct("a": 42, "b": 100) = "a"', False);
  Test('struct("a": 42, "b": 100) = ("a": 42)', False);
  Test('struct("a": 42, "b": 100) = true', False);
  Test('struct("a": 42, "b": 100) = ❨❨2, 6❩, ❨−5, 3❩❩', False);
  Test('struct("a": 42, "b": 100) = 0', False);
  Test('struct("a": 42, "b": 100) = 0.0', False);
  Test('struct("a": 42, "b": 100) = i-i', False);
  Test('struct("a": 42, "b": 100) = (.;)', False);

  Test('color("red") = color("red")', True);
  Test('color("red") = rgb(1, 0, 0)', True);

  Test('color("red") = rgb(1, 0.000001, 0)', False);
  Test('color("red") = rgb(0.7, 0, 0)', False);
  Test('color("red") = rgb(1, 1, 0)', False);
  Test('color("red") = rgb(1, 1, 1)', False);
  Test('color("red") = rgb(0, 0, 0)', False);
  Test('color("green") = rgb(1, 0, 0)', False);

  Test('color("red") = "red"', False);
  Test('color("red") = 16#FF0000', False);
  Test('color("red") = true', False);
  Test('color("red") = color("green")', False);

  Test('color("black") = rgb(0, 0, 0)', True);

  Test('color("black") = rgb(0.0001, 0, 0)', False);
  Test('color("black") = rgb(0, 0.0001, 0)', False);
  Test('color("black") = rgb(0, 0, 0.001)', False);
  Test('color("black") = rgb(0, 1, 0)', False);
  Test('color("black") = rgb(1, 1, 0)', False);
  Test('color("black") = rgb(1, 1, 1)', False);
  Test('color("black") = rgb(0, 1, 1)', False);

  Test('color("black") = 0', False);
  Test('color("black") = 0.0', False);

  Test('color("white") = rgb(1, 1, 1)', True);

  Test('color("white") = rgb(0.9999, 1, 1)', False);
  Test('color("white") = rgb(0, 1, 1)', False);
  Test('color("white") = rgb(1, 0.999, 1)', False);
  Test('color("white") = rgb(1, 1, 0)', False);
  Test('color("white") = rgb(1, 1, 0)', False);
  Test('color("white") = rgb(0.3, 0.5, 0.7)', False);
  Test('color("white") = rgb(0.1, 0.1, 0.1)', False);

  Test('''(1, 2, 3) = ''(1, 2, 3)', True);
  Test('''(1, 2.0, 3) = ''(1, 2, 3)', True);
  Test('''(1, 2, 3) = ''(1, 2, 3.0)', True);
  Test('''(1, 2, 3) = ''(1/2 + 1/2, 2, 3)', True);
  Test('''(1., 2., 3.) = ''(1/2 + 1/2, 2, 3 + i - i)', True);
  Test('''(1, 2 + i - i, 3) = ''(1, 2, 3)', True);
  Test('''(1, 2 + i - i, 3.) = ''(1/2 + 1/2, 2, 3)', True);
  Test('''(1, 2., 3 + i) = ''(1, 2, 3 + i)', True);

  Test('''(2, 2, 3) = ''(1, 2, 3)', False);
  Test('''(1, 1, 3) = ''(1, 2, 3)', False);
  Test('''(1, 2, -3) = ''(1, 2, 3)', False);
  Test('''(1, 2, 3.01) = ''(1, 2, 3)', False);
  Test('''(1, 2, 3) = ''(10, 2, 3)', False);
  Test('''(1, 2, 3) = ''(1, 0.2, 3)', False);
  Test('''(1, 2, 3) = ''(1, 2, 3.000000001)', False);
  Test('''(1, 2, 3) = ''(1, 2, 2.999999999)', False);
  Test('''(2, 2, 3) = ''(1, 2, 3/2)', False);
  Test('''(2, 2, 3) = ''(1, 2, 3 + i)', False);

  Test('''(1, 2, "test") = ''(1, 2, 3)', False);
  Test('''(1, 2, "test") = ''(1, 2, "test")', True);
  Test('''(1, 2, "test") = ''(1/2 + 1/2, 2.0, "te"+"st")', True);

  Test('''(1, 2, 3) = ''(1, 2, 3, 4)', False);
  Test('''(1, 2, 3) = ''(1, 2)', False);
  Test('''(1, 2, 3) = ''(3, 2, 1)', False);

  Test('''(1, 2, 3) = ''(1, 2, 3)', True);

  Test('''(1, 2, ❨1, 2❩) = ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', True);

  Test('''(2, 2, ❨1, 1❩) = ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', False);
  Test('''(1, 2.000000001, ❨1, 2❩) = ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', False);
  Test('''(1, 2, ❨1, 2❩) = ''(1/2 + 1/2, 1.1 + 1.0, ❨1 + i − i, 2❩)', False);
  Test('''(1, 2, ❨1, 2❩) = ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i, 2❩)', False);
  Test('''(1, 2, ❨1, 2❩) = ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2, 3❩)', False);
  Test('''(1, 2, ❨1, 2❩) = ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩, 5)', False);
  Test('''(1, 2, ❨1, 2❩) = ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩, "test")', False);
  Test('''(1, 2, ❨1, 2❩, false) = ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', False);
  Test('''(1, 2, ❨1, 2❩) = ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i❩)', False);
  Test('''(1, 2, ❨1, 2❩) = ''(1/2 + 1/2, 1.0 + 1.0, 1 + i − i, 2)', False);

  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);

  Test('''(3, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π^2, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨2, 1❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨i, 0❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, true, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "dog", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("green")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1/2 + 1/2, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨i, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(0.9, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), color("green")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), true) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), 0) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("gold")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);

  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), 0) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0), 0)', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), 0) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0), (.;))', False);

  Test('''(1.9 + 0.1 + i - i, π, ❨1+i-i, 2❩, ❨❨0, i❩, ❨i, 0❩❩, not(true), "cat", color("red")) = ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);

  if SecondsBetween(Now, Tomorrow) < 2 then
    Sleep(3000);

  Test('date() = date()', True);
  Test('date() = tomorrow()', False);
  Test('date() = tomorrow()', False);

  Test('''() = ''()', True);
  Test('''() = {}', False);
  Test('''() = (.;)', False);

  Test('{1, 2, 3} = {1, 2, 3}', True);
  Test('{1, 2, 3} = {1, 2, 3, 3}', True);
  Test('{1, 2, 3} = {1, 2, 3, 3, 3}', True);
  Test('{1, 2, 3} = {3, 2, 1}', True);
  Test('{1, 2, 3} = {1, 2, 2, 3}', True);
  Test('{1, 2, 3} = {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', True);
  Test('{2, 1, 2, 3} = {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', True);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} = {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', True);

  Test('{1, 2, 3} = {1.0, 2.0, 3.0}', True);
  Test('{1, 2, 3} = {1/2 + 1/2, 2, 3}', True);
  Test('{1.0, 2, 3} = {3, 2, 1/2 + 1/2, 3}', True);
  Test('{1.0, 2, 3, 1, 1 + i - i} = {3, 2, 1/2 + 1/2, 3}', True);

  Test('{1, 1.0, sqrt(4), 1.0, 2 + i - i, 3.0, 3, 3, 3, 1/2 + 1/2} = {3.0, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1/2 + 1/2, 2, 3.0}', True);
  Test('{1, 1.0, sqrt(4), 1.0, 2 + i - i, 3.0, 3, 3, 3, 1/2 + 1/2} = {3.0, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2.5, 1, 1/2 + 1/2, 2, 3.0}', False);
  Test('{2 + 1/2, 1, 1.0, sqrt(4), 1.0, 2 + i - i, 3.0, 3, 3, 3, 1/2 + 1/2} = {3.0, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2.5, 1, 1/2 + 1/2, 2, 3.0}', True);

  Test('{2, 2, 2, 3, 3, 3, 3, 2} = {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', False);
  Test('{2, 1, 2, 3, 3, 4, 3, 1} = {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', False);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} = {3, 3, 2, 0, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', False);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} = {3, 3, 2, 1, 2, 5, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', False);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} = {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, -3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', False);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} = {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 10, 2, 2, 3, 2, 1, 1, 2, 3}', False);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} = {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, i, 2, 3, 2, 1, 1, 2, 3}', False);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} = {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2/9, 3, 2, 1, 1, 2, 3}', False);

  Test('{1, true, "dog"} = {"dog", 1.0, true, 1/2 + 1/2, "dog", "dog"}', True);
  Test('{1, true, "dog"} = {"dog", 1.0, true, 1/2 + 1/2, "dog", "dog", "dog"}', True);
  Test('{1, true, "dog"} = {"dog", 1.0, true, 1/2 + 1/2, "dog", "dog", "dog", "cat"}', False);
  Test('{1, true, "dog", "cat"} = {"dog", 1.0, true, 1/2 + 1/2, "dog", "dog", "dog", "cat"}', True);
  Test('{color("blue"), 1, true, "dog", "cat"} = {"dog", 1.0, true, 1/2 + 1/2, rgb(0, 0, 1), "dog", "dog", "dog", "cat"}', True);
  Test('{color("blue"), 1, true, "dog", "cat"} = {"dog", 1.0, true, 1/2 + 1/2, rgb(0, 1, 1), "dog", "dog", "dog", "cat"}', False);

  Test('{} = {}', True);
  Test('{} = {0}', False);
  Test('{0,0,0} = {0}', True);
  Test('{} = ∅', True);

  Test('{"dog", "cat"} = {"dog"} ∪ {"cat"}', True);
  Test('{"dog", "cat"} = {"dog"} ∪ {"cat"} ∪ ∅', True);
  Test('{"dog", "cat"} = {"dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', True);
  Test('{"dog", "cat", "rat"} = {"dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', False);
  Test('{"dog", "cat", "rat"} = {"dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"rat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', True);
  Test('{"dog", "cat", "rat"} = {"dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"rat", "dog"}', True);
  Test('{"dog", "cat", "rat"} = {"rat", "dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', True);
  Test('{"dog", "cat", "rat"} ∪ ∅ ∪ {"dog"} = {"rat", "dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', True);
  Test('{"cat", "rat"} ∪ ∅ ∪ {"dog"} = {"rat", "dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', True);
  Test('{"cat", "rat"} ∪ ∅ ∪ {"dog"} = {"rat", "dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"Dog!"}', False);

  Test('''("dog", "cat") = {"dog", "cat"}', False);
  Test('''() = {}', False);

  Test('4 = {4}', False);
  Test('4 = ''(4)', False);
  Test('4 = true', False);
  Test('true = {4}', False);
  Test('false = {}', False);

  Test('0 = ZeroVector(100)', False); {arguably, could make this return true, at least if LHS and RHS are swapped}
  Test('0 = ZeroMatrix(100)', False); {and this}

  Test('BinaryData("asterism") = BinaryData("asterism")', True);
  Test('BinaryData("asterism") = BinaryData("asterisk")', False);

  Test('BinaryData(0) = BinaryData(0)', True);
  Test('BinaryData(1000) = BinaryData(1000)', True);
  Test('BinaryData(1000) = BinaryData(10000)', False);
  Test('BinaryData(1000) = BinaryData(1001)', False);
  Test('BinaryData(1000) = BinaryData(1000.001)', False);

  Test('BinaryData("rat") = BinaryData(1000)', False);

  Test('BinaryData(1000) = 10', False);
  Test('BinaryData(1000) = "cat"', False);
  Test('BinaryData(1000) = color("red")', False);
  Test('BinaryData(1000) = {1, 2, 3}', False);
  Test('BinaryData(1000) = true', False);
  Test('BinaryData(1000) = now()', False);

  Test('BinaryData(0) = BinaryData(0.0)', {$IFDEF CPU64BITS}True{$ELSE}False{$ENDIF});

  Test('5 = "cat"', False);
  Test('5.5 = ''(1, 2)', False);
  Test('1 + i = {"dog", "cat"}', False);
  Test('{"dog"} = "dog"', False);
  Test('5 + i = "rat"', False);

  // - pixmaps, sounds: postponed

  // - n-ary extension

  Test('1 = 0! = i^0', True);
  Test('1 = 0! = i^0 = exp(0)', True);
  Test('1 = 0! = i^0 = exp(0) = 5/5', True);

  Test('1 = 0! = i^i', False);
  Test('1 = 5! = i^0 = exp(0)', False);
  Test('1 = 0! = i^0 = exp(0) = 5/4', False);

  Test('"AAA" = 3⋅"A" = "AA" + "A"', True);
  Test('"AAA" = 3⋅"A" = "AA" + "A" = "A" + "AA"', True);
  Test('"AAA" = 3⋅"A" = "AA" + "A" = "A" + "AA" = "A" + "A" + "A"', True);

  Test('"ABA" = 3⋅"A" = "AA" + "A"', False);
  Test('"AAA" = 3⋅"B" = "AA" + "A" = "A" + "AA"', False);
  Test('"AAA" = 3⋅"A" = "AB" + "A" = "A" + "AA" = "A" + "A" + "A"', False);

  Test('"A" = 0! = i^0 = exp(0) = 5/5', False);
  Test('1 = "A" = i^0 = exp(0) = 5/5', False);
  Test('1 = 0! = "A" = exp(0) = 5/5', False);
  Test('1 = 0! = i^0 = "A" = 5/5', False);
  Test('1 = 0! = i^0 = exp(0) = "A"', False);

  // Not equal

  Test('1 + 1 ≠ 2', False);
  Test('1 + 1/2 + 0.5 + i - i ≠ 2', False);
  Test('1 + 1/2 + 0.5 + i - i ≠ 2.0', False);
  Test('1 + 1/2 + 0.5 + i - i ≠ 1/2 + 3/2', False);
  Test('1 + 1/2 + 0.5 + i - i ≠ 2 + i - i', False);

  Test('1 ≠ 0', True);
  Test('1 ≠ 1', False);
  Test('1 ≠ 2', True);

  Test('123456789123455 ≠ 123456789123456', True);
  Test('123456789123456 ≠ 123456789123456', False);
  Test('123456789123457 ≠ 123456789123456', True);

  Test('123456789123456 ≠ -123456789123456', True);

  Test('0 ≠ 0', False);
  Test('0 ≠ -0', False);

  Test('52697415 ≠ 52697415', False);
  Test('-52697415 ≠ -52697415', False);
  Test('-52697415 ≠ 52697415', True);
  Test('52697416 ≠ 52697415', True);

  Test('84012636 ≠ 8490516', True);
  Test('84012636 ≠ -8490516', True);

  Test('123456789123456789 ≠ 123456789123456789', False);
  Test('123456789123456789 ≠ 123456789123456788', True);
  Test('123456789123456789 ≠ -123456789123456789', True);

  Test('42 ≠ 40 + 2', False);
  Test('42 ≠ 40 + 3', True);
  Test('42 ≠ 40 - 2', True);

  Test('100 + 23 ≠ 100 + 20 + 3', False);

  Test('75520864896 ≠ 75520864895', True);
  Test('75520864896 ≠ 75520864896', False);
  Test('75520864896 ≠ 75520864897', True);
  Test('75520864896 ≠ 0', True);

  Test('17/2 ≠ 17/2', False);
  Test('17/2 ≠ 19/2', True);
  Test('17/2 ≠ 17/3', True);
  Test('17/2 ≠ 15/2 + 1/2 + 1/2', False);
  Test('17/2 ≠ 15/2 + 1/2 + 3/2', True);

  Test('-17/2 ≠ 17/2', True);
  Test('17/2 ≠ -17/2', True);
  Test('-17/2 ≠ -17/2', False);
  Test('15/2 ≠ 17/2', True);

  Test('189561/9849133 ≠ 189561/9849133', False);
  Test('189561/9849133 ≠ 189562/9849133', True);
  Test('189561/9849133 ≠ 189561/9849132', True);
  Test('189561/9849133 ≠ -189561/9849133', True);
  Test('189562/9849133 ≠ 189561/9849134', True);
  Test('-189561/9849133 ≠ -189561/9849133', False);

  Test('189561/9849133 ≠ 189560/9849133 + 1/9849133', False);

  Test('189561/9849133 ≠ 1/2', True);
  Test('189561/9849133 ≠ -1/2', True);
  Test('189561/9849133 ≠ 0', True);
  Test('189561/9849133 ≠ 18916518/1898105687', True);
  Test('189561/9849133 ≠ 9849133/189561', True);

  Test('1/2 - 1/2 ≠ 1/3 - 1/3', False);
  Test('1/99 - 1/99 ≠ 7/5 - 7/5', False);
  Test('1/2 - 1/2 ≠ 1/3 - 1/3', False);

  Test('1/21 − 38/51 + 83/119 ≠ 523/517 − 325/17 + 92/3 − 331186/26367', False);

  Test('1/21 − 38/51 + 83/119 ≠ 523/517 − 325/17 + 92/3 − 331187/26367', True);
  Test('1/21 − 38/51 + 83/119 ≠ 523/516 − 325/17 + 92/3 − 331186/26367', True);
  Test('1/21 − 38/51 + 83/119 ≠ 523/517 − 325/17 + 92/3 + 331186/26367', True);
  Test('2/21 − 38/51 + 83/119 ≠ 523/517 − 325/17 + 92/3 − 331186/26367', True);
  Test('1/21 − 38/51 + 83/120 ≠ 523/517 − 325/17 + 92/3 − 331186/26367', True);
  Test('0 ≠ 523/517 − 325/17 + 92/3 − 331186/26367', False);
  Test('1/21 − 38/51 + 83/119 ≠ 0', False);

  Test('123.456 ≠ 123.456', False);
  Test('123.456 ≠ 123.457', True);
  Test('123.456 ≠ 123.4560000001', True);
  Test('123.456000000001 ≠ 123.456', True);
  Test('123.456 ≠ -123.456', True);
  Test('123.456 ≠ 123.0', True);

  Test('-123.456 ≠ -123.456', False);
  Test('-123.456 ≠ -123.556', True);
  Test('-123.456 ≠ -123.4559999999', True);

  Test('123E100 ≠ 123E100', False);

  Test('123E100 ≠ 123E101', True);
  Test('123E100 ≠ 123.000000001E100', True);
  Test('122.99999999999E100 ≠ 123E100', True);
  Test('123.0E100 ≠ 123E100', False);
  Test('-123E100 ≠ -123E100', False);
  Test('123E100 ≠ -123E100', True);
  Test('123E100 ≠ 123E200', True);
  Test('123E100 ≠ 0', True);

  Test('189.5613E200 ≠ 189.5613E200', False);
  Test('189.5613E200 ≠ 189.5614E200', True);
  Test('189.5613E200 ≠ 189.5613000001E200', True);
  Test('189.5612999999E200 ≠ 189.5613E200', True);
  Test('-189.5613E200 ≠ -189.5613E200', False);
  Test('189.5613E200 ≠ 189.5613E-200', True);
  Test('189.5613E-200 ≠ 189.5613E-200', False);
  Test('-189.5613E-200 ≠ -189.5613E-200', False);

  Test('189.5613E-200 ≠ 189.5613E-200', False);
  Test('189.5614E-200 ≠ 189.5613E-200', True);
  Test('189.561300001E-200 ≠ 189.5613E-200', True);
  Test('189.5613E-200 ≠ 189.5612999999E-200', True);

  Test('π ≠ 3.1415926535797932384626', True);

  Test('e ≠ exp(1)', False);
  Test('e ≠ exp(1.01)', True);

  Test('e ≠ π', True);
  Test('0⋅e ≠ 0⋅π', False);

  Test('3 + 5⋅i ≠ 3 + 5⋅i', False);
  Test('3 + 5⋅i ≠ 3 + 6⋅i', True);
  Test('3 + 5⋅i ≠ 3 + 5.0000000000001⋅i', True);
  Test('3 + 5⋅i ≠ 2.999999999999 + 5⋅i', True);
  Test('3 + 5⋅i ≠ 3', True);
  Test('3 + 5⋅i ≠ 5⋅i', True);
  Test('3 + 5⋅i ≠ 3 − 5⋅i', True);
  Test('3 − 5⋅i ≠ 3 − 5⋅i', False);
  Test('3 + 5⋅i ≠ 3.00000000001 + 5⋅i', True);
  Test('3.0000000001 + 5⋅i ≠ 3 + 5⋅i', True);

  Test('i ≠ i', False);
  Test('i ≠ -i', True);
  Test('1/i ≠ -i', False);

  Test('1/7 + 6/7 ≠ 1', False);
  Test('2/7 + 6/7 ≠ 1', True);

  Test('2 ≠ 12/9 + 6/9', False);
  Test('2 ≠ 12/9 + 7/9', True);

  Test('1/2 + 1/2 ≠ 1', False);
  Test('1/2 + 1/2 ≠ 3/2 - 1/2', False);
  Test('1/2 + 1/2 ≠ 1.0', False);
  Test('1/2 + 1/2 ≠ 1.0 + i - i', False);

  Test('1/2 ≠ 0.5', False);
  Test('1/2 ≠ 0.5 + i - i', False);
  Test('0.75 ≠ 3/4', False);
  Test('-0.75 ≠ -3/4', False);
  Test('0.75 + i - i ≠ 3/4', False);
  Test('0.75 ≠ 3/4 + i - i', False);

  Test('0 ≠ 0.0', False);
  Test('0 ≠ 1/2 - 1/2', False);
  Test('0 ≠ i - i', False);
  Test('1/2 - 1/2 ≠ 0.0', False);
  Test('1/2 - 1/2 ≠ i - i', False);
  Test('0.0 ≠ i - i', False);

  Test('5 + 0 ≠ 5 + 0.0', False);
  Test('5 + 0 ≠ 5 + 1/2 - 1/2', False);
  Test('5 + 0 ≠ 5 + i - i', False);
  Test('5 + 1/2 - 1/2 ≠ 5 + 0.0', False);
  Test('5 + 1/2 - 1/2 ≠ 5 + i - i', False);
  Test('5 + 0.0 ≠ 5 + i - i', False);

  Test('6 + 0 ≠ 5 + 0.0', True);
  Test('6 + 0 ≠ 5 + 1/2 - 1/2', True);
  Test('6 + 0 ≠ 5 + i - i', True);
  Test('6 + 1/2 - 1/2 ≠ 5 + 0.0', True);
  Test('6 + 1/2 - 1/2 ≠ 5 + i - i', True);
  Test('6 + 0.0 ≠ 5 + i - i', True);

  Test('0 ≠ (.;)', True);
  Test('0.0 ≠ (.;)', True);
  Test('1/2-1/2 ≠ (.;)', True);
  Test('i-i ≠ (.;)', True);

  Test('1/7 + 58/2 − 63/7 + 25/7 ≠ 166/7', False);
  Test('1/7 + 58/2 − 63/7 + 25/7 − 5/7 ≠ 23', False);
  Test('1/7 + 58/2 − 63/7 + 25/7 − 5/7 ≠ 23.0', False);
  Test('1/7 + 58/2 − 63/7 + 25/7 − 5/7 ≠ 23.0 + i − i', False);
  Test('i + 1/7 + 58/2 − 63/7 + 25/7 − 5/7 ≠ 23.0 + i', False);

  Test('59/10 ≠ 5.9', False);
  Test('59/100 ≠ 0.59', False);
  Test('59/1000 ≠ 0.059', False);

  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≠ 7/2', False);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≠ 3.5', False);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≠ 3.5 + 0⋅i', False);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≠ 3.5 + 2⋅i − 2⋅i', False);

  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≠ 7/3', True);
  Test('5⋅i − 3⋅i − 3⋅i + 7/2 ≠ 3.5', True);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≠ 3.6 + 0⋅i', True);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≠ 3.5 + 2⋅i − 2.2⋅i', True);

  Test('1/4 ≠ 0.25', False);
  Test('4⋅0.25 ≠ 1', False);
  Test('4⋅0.25 ≠ 1.0', False);
  Test('4⋅0.25 ≠ 3/4 + 1/4', False);
  Test('4⋅0.25 ≠ 3/4 + 1/4 + i - i', False);
  Test('i - i ≠ 1/7 + 2/7 - 3/7', False);
  Test('i - i ≠ 1/7 + 2/7 - 3.00000001/7', True);

  Test('1/4 ≠ 0.24', True);
  Test('4⋅0.25000000001 ≠ 1', True);
  Test('4⋅0.25 ≠ 0.999999999', True);
  Test('4⋅0.25 ≠ 3/4 + 2/4', True);
  Test('4⋅0.25 ≠ 3/4 + 1/4 - i', True);

  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≠ 0', False);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≠ 0.0', False);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≠ 1/2 - 1/2', False);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≠ i - i', False);

  Test('6 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≠ 0', True);
  Test('5 + 3⋅i + 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≠ 0.0', True);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 3⋅i + 1 − 10⋅i ≠ 1/2 - 1/2', True);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≠ 7 - i', True);

  Test('5 ≠ ❨5❩', True);
  Test('5 ≠ ❨❨5❩❩', True);
  Test('❨5❩ ≠ ❨❨5❩❩', True);

  Test('0 ≠ ❨0❩', True);
  Test('0 ≠ ❨❨0❩❩', True);
  Test('❨0❩ ≠ ❨❨0❩❩', True);

  Test('0.0 ≠ ❨0❩', True);
  Test('0.0 ≠ ❨❨0❩❩', True);

  Test('1/2-1/2 ≠ ❨0❩', True);
  Test('1/2-1/2 ≠ ❨❨0❩❩', True);
  Test('❨0❩ ≠ ❨❨0❩❩', True);

  Test('i-i ≠ ❨0❩', True);
  Test('i-i ≠ ❨❨0❩❩', True);
  Test('❨0❩ ≠ ❨❨0❩❩', True);

  Test('0 ≠ ❨i-i❩', True);
  Test('0 ≠ ❨❨i-i❩❩', True);
  Test('❨0❩ ≠ ❨❨i-i❩❩', True);

  Test('i-i ≠ ❨i-i❩', True);
  Test('i-i ≠ ❨❨i-i❩❩', True);
  Test('❨i-i❩ ≠ ❨❨0❩❩', True);

  Test('❨i-i❩ ≠ ❨❨i-i❩❩', True);

  Test('❨1, 2, 3, 4, 5❩ ≠ ❨1, 2, 3, 4, 5❩', False);

  Test('❨2, 2, 3, 4, 5❩ ≠ ❨1, 2, 3, 4, 5❩', True);
  Test('❨1, -2, 3, 4, 5❩ ≠ ❨1, 2, 3, 4, 5❩', True);
  Test('❨1, 2, 3.000000000001, 4, 5❩ ≠ ❨1, 2, 3, 4, 5❩', True);
  Test('❨1, 2, 3, 3.9999999999, 5❩ ≠ ❨1, 2, 3, 4, 5❩', True);
  Test('❨1, 2, 3, 4, 500❩ ≠ ❨1, 2, 3, 4, 5❩', True);
  Test('❨1, 2, 3, 4❩ ≠ ❨1, 2, 3, 4, 5❩', True);
  Test('❨1, 2, 3, 4, 5❩ ≠ ❨1, 2, 3❩', True);

  Test('❨π, e, sin(1)❩ ≠ ❨π, e, sin(1)❩', False);

  Test('❨5 + i, 2, −i❩ ≠ ❨5 + i, 2, −i❩', False);
  Test('❨5 + i, 2, −i❩ ≠ ❨5 + i, 2, i❩', True);
  Test('❨5 + i, 0, −i❩ ≠ ❨5 + i, 2, −i❩', True);

  Test('❨1, 2, 3, 4, 5❩ ≠ ❨1, 2, 3, 4, 5 + i - i❩', False);
  Test('❨1 + i - i, 2, 3, 4, 5❩ ≠ ❨1, 2, 3, 4, 5❩', False);

  Test('❨1, 2, 3❩ ≠ ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', False);

  Test('❨2, 2, 3❩ ≠ ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', True);
  Test('❨1, -2, 3❩ ≠ ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', True);
  Test('❨1, 2, 3.0000001❩ ≠ ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', True);
  Test('❨1, 2, 3❩ ≠ ❨1, 1, 2❩ + ❨0, 1 − i, 1❩', True);
  Test('❨1, 2, 3❩ ≠ ❨1, i, 2❩ + ❨0, 1 − i, 1❩', True);
  Test('❨1, 2, 3❩ ≠ ❨1, 1 + i, 2❩ + ❨0, 1, 1❩', True);

  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);

  Test('❨❨5, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, 2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1.0000000000001❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨3.9999999999, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0.1, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2.5❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4.000000001, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, 2, -1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, -1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, -4, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9.2❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨40, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2❩, ❨4, 0❩, ❨−1, 0❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨4, −2, 1❩', True);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ 4', True);
  Test('❨❨4❩❩ ≠ ❨4❩', True);
  Test('❨❨4❩❩ ≠ 4', True);
  Test('❨4❩ ≠ 4', True);
  Test('4 ≠ ❨4❩', True);

  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, i-i, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, i-i, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4 + i-i, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, i-i, 9❩, ❨−1, 0, 2❩❩ ≠ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, i-i, 2❩❩', False);

  Test('❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩ ≠ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩ ≠ ❨❨2+i, 2−3⋅i❩, ❨i, i-i❩❩', False);

  Test('❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩ ≠ ❨❨2+i, 2−3⋅i❩, ❨i, 1❩❩', True);
  Test('❨❨2+i, 2−3⋅i❩, ❨i, 2❩❩ ≠ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', True);
  Test('❨❨2+i, 2−4⋅i❩, ❨i, 0❩❩ ≠ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', True);
  Test('❨❨2+i, 2+3⋅i❩, ❨i, 0❩❩ ≠ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', True);
  Test('❨❨2-i, 2−3⋅i❩, ❨i, 0❩❩ ≠ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', True);
  Test('❨❨6+i, 2−3⋅i❩, ❨i, 0❩❩ ≠ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', True);
  Test('❨❨2+i, 2−3⋅i❩, ❨i, 6❩❩ ≠ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', True);

  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≠ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);

  Test('❨❨3, 6❩, ❨−5, 3❩❩ ≠ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', True);
  Test('❨❨2, 7❩, ❨−5, 3❩❩ ≠ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', True);
  Test('❨❨2, 6❩, ❨5, 3❩❩ ≠ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', True);
  Test('❨❨2, 6❩, ❨−5, 3.0000000001❩❩ ≠ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', True);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≠ ❨❨1, 3 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', True);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≠ ❨❨1, 2 + 3⋅i❩, ❨5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', True);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≠ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4❩, ❨3, 3❩❩', True);

  Test('"alpha" ≠ "alpha"', False);
  Test('"alpha" ≠ "beta"', True);
  Test('"alpha" ≠ ""', True);
  Test('"" ≠ ""', False);

  Test('"7" ≠ 7', True);
  Test('"7" ≠ 7.0', True);
  Test('"7" ≠ 7+i-i', True);

  Test('true ≠ true', False);
  Test('true ≠ false', True);
  Test('false ≠ true', True);
  Test('false ≠ false', False);

  Test('false ≠ 0', True);
  Test('false ≠ 0.0', True);
  Test('false ≠ 1/2 - 1/2', True);
  Test('false ≠ i - i', True);
  Test('false ≠ (.;)', True);

  Test('true ≠ 1', True);
  Test('true ≠ 1.0', True);

  Test('true ≠ "true"', True);
  Test('"false" ≠ false', True);
  Test('"false" ≠ ""', True);
  Test('"false" ≠ "false"', False);

  Test('"LiU rocks" ≠ true', True);

  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≠ 5', True);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≠ "alpha"', True);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≠ true', True);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≠ 5.0', True);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≠ 0', True);

  Test('❨❨0, 0❩, ❨0, 0❩❩ ≠ 0', True);
  Test('❨0, 0❩ ≠ 0', True);

  Test('struct("a": 42, "b": 100) ≠ struct("a": 42, "b": 100)', False);

  Test('struct("a": 42.0000001, "b": 100) ≠ struct("a": 42, "b": 100)', True);
  Test('struct("A": 42, "b": 100) ≠ struct("a": 42, "b": 100)', True);
  Test('struct("a": 43, "b": 100) ≠ struct("a": 42, "b": 100)', True);
  Test('struct("a": 42, "c": 100) ≠ struct("a": 42, "b": 100)', True);
  Test('struct("a": 42, "b": 101) ≠ struct("a": 42, "b": 100)', True);
  Test('struct("a": 42, "b": 100) ≠ struct("d": 42, "b": 100)', True);
  Test('struct("a": 42, "b": 100) ≠ struct("a": 420, "b": 100)', True);
  Test('struct("a": 42, "b": 100) ≠ struct("a": 42, "e": 100)', True);
  Test('struct("a": 42, "b": 100) ≠ struct("a": 42, "b": 100.01)', True);
  Test('struct("a": 42, "b": 100) ≠ struct("a": 42)', True);
  Test('struct("a": 42) ≠ struct("a": 42, "b": 100)', True);
  Test('struct("b": 100) ≠ struct("a": 42, "b": 100)', True);
  Test('struct("b": 100, "a": 42) ≠ struct("a": 42, "b": 100)', True);
  Test('struct("a": 42, "b": 100) ≠ struct("a": 100, "b": 42)', True);
  Test('struct("a": 42, "b": 100) ≠ struct("a": 42, "b": 100, "c": 1000)', True);
  Test('struct("a": 42, "b": 100, "c": 1000) ≠ struct("a": 42, "b": 100)', True);
  Test('struct("a": 42, "b": 100, "c": 1000) ≠ struct("a": 42, "b": 100, "c": 1000)', False);
  Test('struct("a": 42, "b": 100, "c": "1000") ≠ struct("a": 42, "b": 100, "c": 1000)', True);
  Test('struct("a": 42, "b": 100, "c": "1000") ≠ struct("a": 42, "b": 100, "c": "1000")', False);
  Test('struct("a": 42, "b": 100, "c": "1000") ≠ struct("a": 42, "b": 100)', True);
  Test('struct("a": 42, "b": 100, "c": "1000") ≠ struct("a": 42)', True);
  Test('struct("a": 42, "b": 100) ≠ struct("a": 42, "b": 100, "c": "1000")', True);
  Test('struct("a": 42) ≠ struct("a": 42, "b": 100, "c": "1000")', True);

  Test('struct("a": 42, "b": 100) ≠ 42', True);
  Test('struct("a": 42, "b": 100) ≠ "a"', True);
  Test('struct("a": 42, "b": 100) ≠ ("a": 42)', True);
  Test('struct("a": 42, "b": 100) ≠ true', True);
  Test('struct("a": 42, "b": 100) ≠ ❨❨2, 6❩, ❨−5, 3❩❩', True);
  Test('struct("a": 42, "b": 100) ≠ 0', True);
  Test('struct("a": 42, "b": 100) ≠ 0.0', True);
  Test('struct("a": 42, "b": 100) ≠ i-i', True);
  Test('struct("a": 42, "b": 100) ≠ (.;)', True);

  Test('color("red") ≠ color("red")', False);
  Test('color("red") ≠ rgb(1, 0, 0)', False);

  Test('color("red") ≠ rgb(1, 0.000001, 0)', True);
  Test('color("red") ≠ rgb(0.7, 0, 0)', True);
  Test('color("red") ≠ rgb(1, 1, 0)', True);
  Test('color("red") ≠ rgb(1, 1, 1)', True);
  Test('color("red") ≠ rgb(0, 0, 0)', True);
  Test('color("green") ≠ rgb(1, 0, 0)', True);

  Test('color("red") ≠ "red"', True);
  Test('color("red") ≠ 16#FF0000', True);
  Test('color("red") ≠ true', True);
  Test('color("red") ≠ color("green")', True);

  Test('color("black") ≠ rgb(0, 0, 0)', False);

  Test('color("black") ≠ rgb(0.0001, 0, 0)', True);
  Test('color("black") ≠ rgb(0, 0.0001, 0)', True);
  Test('color("black") ≠ rgb(0, 0, 0.001)', True);
  Test('color("black") ≠ rgb(0, 1, 0)', True);
  Test('color("black") ≠ rgb(1, 1, 0)', True);
  Test('color("black") ≠ rgb(1, 1, 1)', True);
  Test('color("black") ≠ rgb(0, 1, 1)', True);

  Test('color("black") ≠ 0', True);
  Test('color("black") ≠ 0.0', True);

  Test('color("white") ≠ rgb(1, 1, 1)', False);

  Test('color("white") ≠ rgb(0.9999, 1, 1)', True);
  Test('color("white") ≠ rgb(0, 1, 1)', True);
  Test('color("white") ≠ rgb(1, 0.999, 1)', True);
  Test('color("white") ≠ rgb(1, 1, 0)', True);
  Test('color("white") ≠ rgb(1, 1, 0)', True);
  Test('color("white") ≠ rgb(0.3, 0.5, 0.7)', True);
  Test('color("white") ≠ rgb(0.1, 0.1, 0.1)', True);

  Test('''(1, 2, 3) ≠ ''(1, 2, 3)', False);
  Test('''(1, 2.0, 3) ≠ ''(1, 2, 3)', False);
  Test('''(1, 2, 3) ≠ ''(1, 2, 3.0)', False);
  Test('''(1, 2, 3) ≠ ''(1/2 + 1/2, 2, 3)', False);
  Test('''(1., 2., 3.) ≠ ''(1/2 + 1/2, 2, 3 + i - i)', False);
  Test('''(1, 2 + i - i, 3) ≠ ''(1, 2, 3)', False);
  Test('''(1, 2 + i - i, 3.) ≠ ''(1/2 + 1/2, 2, 3)', False);
  Test('''(1, 2., 3 + i) ≠ ''(1, 2, 3 + i)', False);

  Test('''(2, 2, 3) ≠ ''(1, 2, 3)', True);
  Test('''(1, 1, 3) ≠ ''(1, 2, 3)', True);
  Test('''(1, 2, -3) ≠ ''(1, 2, 3)', True);
  Test('''(1, 2, 3.01) ≠ ''(1, 2, 3)', True);
  Test('''(1, 2, 3) ≠ ''(10, 2, 3)', True);
  Test('''(1, 2, 3) ≠ ''(1, 0.2, 3)', True);
  Test('''(1, 2, 3) ≠ ''(1, 2, 3.000000001)', True);
  Test('''(1, 2, 3) ≠ ''(1, 2, 2.999999999)', True);
  Test('''(2, 2, 3) ≠ ''(1, 2, 3/2)', True);
  Test('''(2, 2, 3) ≠ ''(1, 2, 3 + i)', True);

  Test('''(1, 2, "test") ≠ ''(1, 2, 3)', True);
  Test('''(1, 2, "test") ≠ ''(1, 2, "test")', False);
  Test('''(1, 2, "test") ≠ ''(1/2 + 1/2, 2.0, "te"+"st")', False);

  Test('''(1, 2, 3) ≠ ''(1, 2, 3, 4)', True);
  Test('''(1, 2, 3) ≠ ''(1, 2)', True);
  Test('''(1, 2, 3) ≠ ''(3, 2, 1)', True);

  Test('''(1, 2, 3) ≠ ''(1, 2, 3)', False);

  Test('''(1, 2, ❨1, 2❩) ≠ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', False);

  Test('''(2, 2, ❨1, 1❩) ≠ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', True);
  Test('''(1, 2.000000001, ❨1, 2❩) ≠ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', True);
  Test('''(1, 2, ❨1, 2❩) ≠ ''(1/2 + 1/2, 1.1 + 1.0, ❨1 + i − i, 2❩)', True);
  Test('''(1, 2, ❨1, 2❩) ≠ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i, 2❩)', True);
  Test('''(1, 2, ❨1, 2❩) ≠ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2, 3❩)', True);
  Test('''(1, 2, ❨1, 2❩) ≠ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩, 5)', True);
  Test('''(1, 2, ❨1, 2❩) ≠ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩, "test")', True);
  Test('''(1, 2, ❨1, 2❩, false) ≠ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', True);
  Test('''(1, 2, ❨1, 2❩) ≠ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i❩)', True);
  Test('''(1, 2, ❨1, 2❩) ≠ ''(1/2 + 1/2, 1.0 + 1.0, 1 + i − i, 2)', True);

  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);

  Test('''(3, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π^2, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨2, 1❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨i, 0❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, true, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "dog", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("green")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1/2 + 1/2, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨i, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(0.9, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), color("green")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), true) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), 0) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("gold")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);

  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), 0) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0), 0)', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), 0) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0), (.;))', True);

  Test('''(1.9 + 0.1 + i - i, π, ❨1+i-i, 2❩, ❨❨0, i❩, ❨i, 0❩❩, not(true), "cat", color("red")) ≠ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);

  if SecondsBetween(Now, Tomorrow) < 2 then
    Sleep(3000);

  Test('date() ≠ date()', False);
  Test('date() ≠ tomorrow()', True);
  Test('date() ≠ tomorrow()', True);

  Test('''() ≠ ''()', False);
  Test('''() ≠ {}', True);
  Test('''() ≠ (.;)', True);

  Test('{1, 2, 3} ≠ {1, 2, 3}', False);
  Test('{1, 2, 3} ≠ {1, 2, 3, 3}', False);
  Test('{1, 2, 3} ≠ {1, 2, 3, 3, 3}', False);
  Test('{1, 2, 3} ≠ {3, 2, 1}', False);
  Test('{1, 2, 3} ≠ {1, 2, 2, 3}', False);
  Test('{1, 2, 3} ≠ {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', False);
  Test('{2, 1, 2, 3} ≠ {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', False);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} ≠ {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', False);

  Test('{1, 2, 3} ≠ {1.0, 2.0, 3.0}', False);
  Test('{1, 2, 3} ≠ {1/2 + 1/2, 2, 3}', False);
  Test('{1.0, 2, 3} ≠ {3, 2, 1/2 + 1/2, 3}', False);
  Test('{1.0, 2, 3, 1, 1 + i - i} ≠ {3, 2, 1/2 + 1/2, 3}', False);

  Test('{1, 1.0, sqrt(4), 1.0, 2 + i - i, 3.0, 3, 3, 3, 1/2 + 1/2} ≠ {3.0, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1/2 + 1/2, 2, 3.0}', False);
  Test('{1, 1.0, sqrt(4), 1.0, 2 + i - i, 3.0, 3, 3, 3, 1/2 + 1/2} ≠ {3.0, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2.5, 1, 1/2 + 1/2, 2, 3.0}', True);
  Test('{2 + 1/2, 1, 1.0, sqrt(4), 1.0, 2 + i - i, 3.0, 3, 3, 3, 1/2 + 1/2} ≠ {3.0, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2.5, 1, 1/2 + 1/2, 2, 3.0}', False);

  Test('{2, 2, 2, 3, 3, 3, 3, 2} ≠ {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', True);
  Test('{2, 1, 2, 3, 3, 4, 3, 1} ≠ {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', True);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} ≠ {3, 3, 2, 0, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', True);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} ≠ {3, 3, 2, 1, 2, 5, 2, 2, 1, 2, 3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', True);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} ≠ {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, -3, 1, 2, 2, 3, 2, 1, 1, 2, 3}', True);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} ≠ {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 10, 2, 2, 3, 2, 1, 1, 2, 3}', True);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} ≠ {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, i, 2, 3, 2, 1, 1, 2, 3}', True);
  Test('{2, 1, 2, 3, 3, 3, 3, 1} ≠ {3, 3, 2, 1, 2, 3, 2, 2, 1, 2, 3, 1, 2, 2/9, 3, 2, 1, 1, 2, 3}', True);

  Test('{1, true, "dog"} ≠ {"dog", 1.0, true, 1/2 + 1/2, "dog", "dog"}', False);
  Test('{1, true, "dog"} ≠ {"dog", 1.0, true, 1/2 + 1/2, "dog", "dog", "dog"}', False);
  Test('{1, true, "dog"} ≠ {"dog", 1.0, true, 1/2 + 1/2, "dog", "dog", "dog", "cat"}', True);
  Test('{1, true, "dog", "cat"} ≠ {"dog", 1.0, true, 1/2 + 1/2, "dog", "dog", "dog", "cat"}', False);
  Test('{color("blue"), 1, true, "dog", "cat"} ≠ {"dog", 1.0, true, 1/2 + 1/2, rgb(0, 0, 1), "dog", "dog", "dog", "cat"}', False);
  Test('{color("blue"), 1, true, "dog", "cat"} ≠ {"dog", 1.0, true, 1/2 + 1/2, rgb(0, 1, 1), "dog", "dog", "dog", "cat"}', True);

  Test('{} ≠ {}', False);
  Test('{} ≠ {0}', True);
  Test('{0,0,0} ≠ {0}', False);
  Test('{} ≠ ∅', False);

  Test('{"dog", "cat"} ≠ {"dog"} ∪ {"cat"}', False);
  Test('{"dog", "cat"} ≠ {"dog"} ∪ {"cat"} ∪ ∅', False);
  Test('{"dog", "cat"} ≠ {"dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', False);
  Test('{"dog", "cat", "rat"} ≠ {"dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', True);
  Test('{"dog", "cat", "rat"} ≠ {"dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"rat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', False);
  Test('{"dog", "cat", "rat"} ≠ {"dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"rat", "dog"}', False);
  Test('{"dog", "cat", "rat"} ≠ {"rat", "dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', False);
  Test('{"dog", "cat", "rat"} ∪ ∅ ∪ {"dog"} ≠ {"rat", "dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', False);
  Test('{"cat", "rat"} ∪ ∅ ∪ {"dog"} ≠ {"rat", "dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"dog"}', False);
  Test('{"cat", "rat"} ∪ ∅ ∪ {"dog"} ≠ {"rat", "dog"} ∪ {"cat"} ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ ∅ ∪ {"Dog!"}', True);

  Test('''("dog", "cat") ≠ {"dog", "cat"}', True);
  Test('''() ≠ {}', True);

  Test('4 ≠ {4}', True);
  Test('4 ≠ ''(4)', True);
  Test('4 ≠ true', True);
  Test('true ≠ {4}', True);
  Test('false ≠ {}', True);

  Test('0 ≠ ZeroVector(100)', True); {arguably, could make this return true, at least if LHS and RHS are swapped}
  Test('0 ≠ ZeroMatrix(100)', True); {and this}

  Test('BinaryData("asterism") ≠ BinaryData("asterism")', False);
  Test('BinaryData("asterism") ≠ BinaryData("asterisk")', True);

  Test('BinaryData(0) ≠ BinaryData(0)', False);
  Test('BinaryData(1000) ≠ BinaryData(1000)', False);
  Test('BinaryData(1000) ≠ BinaryData(10000)', True);
  Test('BinaryData(1000) ≠ BinaryData(1001)', True);
  Test('BinaryData(1000) ≠ BinaryData(1000.001)', True);

  Test('BinaryData("rat") ≠ BinaryData(1000)', True);

  Test('BinaryData(1000) ≠ 10', True);
  Test('BinaryData(1000) ≠ "cat"', True);
  Test('BinaryData(1000) ≠ color("red")', True);
  Test('BinaryData(1000) ≠ {1, 2, 3}', True);
  Test('BinaryData(1000) ≠ true', True);
  Test('BinaryData(1000) ≠ now()', True);

  Test('BinaryData(0) ≠ BinaryData(0.0)', {$IFDEF CPU64BITS}False{$ELSE}True{$ENDIF});

  Test('5 ≠ "cat"', True);
  Test('5.5 ≠ ''(1, 2)', True);
  Test('1 + i ≠ {"dog", "cat"}', True);
  Test('{"dog"} ≠ "dog"', True);
  Test('5 + i ≠ "rat"', True);

  // - pixmaps, sounds: postponed

  // Same value/approximation

  Test('1 + 1 ≈ 2', True);
  Test('1 + 1/2 + 0.5 + i - i ≈ 2', True);
  Test('1 + 1/2 + 0.5 + i - i ≈ 2.0', True);
  Test('1 + 1/2 + 0.5 + i - i ≈ 1/2 + 3/2', True);
  Test('1 + 1/2 + 0.5 + i - i ≈ 2 + i - i', True);

  Test('1 ≈ 0', False);
  Test('1 ≈ 1', True);
  Test('1 ≈ 2', False);

  Test('0 ≈ 0', True);
  Test('0 ≈ -0', True);

  Test('52697415 ≈ 52697415', True);
  Test('-52697415 ≈ -52697415', True);
  Test('-52697415 ≈ 52697415', False);
  Test('52697416 ≈ 52697415', False);

  Test('84012636 ≈ 8490516', False);
  Test('84012636 ≈ -8490516', False);

  Test('123456789123456789 ≈ 123456789123456789', True);
  Test('123456789123456789 ≈ 123456789123456788', True); // all numbers are compared as floating-point values with epsilon
  Test('123456789123456789 ≈ -123456789123456789', False);

  Test('42 ≈ 40 + 2', True);
  Test('42 ≈ 40 + 3', False);
  Test('42 ≈ 40 - 2', False);

  Test('100 + 23 ≈ 100 + 20 + 3', True);

  Test('75520864896 ≈ 75520864895', False);
  Test('75520864896 ≈ 75520864896', True);
  Test('75520864896 ≈ 75520864897', False);
  Test('75520864896 ≈ 0', False);

  Test('17/2 ≈ 17/2', True);
  Test('17/2 ≈ 19/2', False);
  Test('17/2 ≈ 17/3', False);
  Test('17/2 ≈ 15/2 + 1/2 + 1/2', True);
  Test('17/2 ≈ 15/2 + 1/2 + 3/2', False);

  Test('-17/2 ≈ 17/2', False);
  Test('17/2 ≈ -17/2', False);
  Test('-17/2 ≈ -17/2', True);
  Test('15/2 ≈ 17/2', False);

  Test('189561/9849133 ≈ 189561/9849133', True);
  Test('189561/9849133 ≈ 189562/9849133', False);
  Test('189561/9849133 ≈ 189561/9849132', False);
  Test('189561/9849133 ≈ -189561/9849133', False);
  Test('189562/9849133 ≈ 189561/9849134', False);
  Test('-189561/9849133 ≈ -189561/9849133', True);

  Test('189561/9849133 ≈ 189560/9849133 + 1/9849133', True);

  Test('189561/9849133 ≈ 1/2', False);
  Test('189561/9849133 ≈ -1/2', False);
  Test('189561/9849133 ≈ 0', False);
  Test('189561/9849133 ≈ 18916518/1898105687', False);
  Test('189561/9849133 ≈ 9849133/189561', False);

  Test('1/2 - 1/2 ≈ 1/3 - 1/3', True);
  Test('1/99 - 1/99 ≈ 7/5 - 7/5', True);
  Test('1/2 - 1/2 ≈ 1/3 - 1/3', True);

  Test('1/21 − 38/51 + 83/119 ≈ 523/517 − 325/17 + 92/3 − 331186/26367', True);

  Test('1/21 − 38/51 + 83/119 ≈ 523/517 − 325/17 + 92/3 − 331187/26367', False);
  Test('1/21 − 38/51 + 83/119 ≈ 523/516 − 325/17 + 92/3 − 331186/26367', False);
  Test('1/21 − 38/51 + 83/119 ≈ 523/517 − 325/17 + 92/3 + 331186/26367', False);
  Test('2/21 − 38/51 + 83/119 ≈ 523/517 − 325/17 + 92/3 − 331186/26367', False);
  Test('1/21 − 38/51 + 83/120 ≈ 523/517 − 325/17 + 92/3 − 331186/26367', False);
  Test('0 ≈ 523/517 − 325/17 + 92/3 − 331186/26367', True);
  Test('1/21 − 38/51 + 83/119 ≈ 0', True);

  Test('123.456 ≈ 123.456', True);
  Test('123.456 ≈ 123.457', False);
  Test('123.456 ≈ 123.4560000001', False);
  Test('123.456000000001 ≈ 123.456', False);
  Test('123.456 ≈ -123.456', False);
  Test('123.456 ≈ 123.0', False);

  Test('-123.456 ≈ -123.456', True);
  Test('-123.456 ≈ -123.556', False);
  Test('-123.456 ≈ -123.4559999999', False);

  Test('123E100 ≈ 123E100', True);

  Test('123E100 ≈ 123E101', False);
  Test('123E100 ≈ 123.000000001E100', False);
  Test('122.99999999999E100 ≈ 123E100', False);
  Test('123.0E100 ≈ 123E100', True);
  Test('-123E100 ≈ -123E100', True);
  Test('123E100 ≈ -123E100', False);
  Test('123E100 ≈ 123E200', False);
  Test('123E100 ≈ 0', False);

  Test('189.5613E200 ≈ 189.5613E200', True);
  Test('189.5613E200 ≈ 189.5614E200', False);
  Test('189.5613E200 ≈ 189.5613000001E200', False);
  Test('189.5612999999E200 ≈ 189.5613E200', False);
  Test('-189.5613E200 ≈ -189.5613E200', True);
  Test('189.5613E200 ≈ 189.5613E-200', False);
  Test('189.5613E-200 ≈ 189.5613E-200', True);
  Test('-189.5613E-200 ≈ -189.5613E-200', True);

  Test('189.5613E-200 ≈ 189.5613E-200', True);
  Test('189.5614E-200 ≈ 189.5613E-200', True); // both are zero
  Test('189.5614E-200 ≈ 9238.4217E-200', True); // both are zero
  Test('189.5614E-200 ≈ 9238.4217E-210', True); // both are zero
  Test('189.5614E-200 ≈ -9238.4217E-210', True); // both are zero
  Test('189.5614E-200 ≈ 0', True);
  Test('189.561300001E-200 ≈ 189.5613E-200', True);
  Test('189.5613E-200 ≈ 189.5612999999E-200', True);

  Test('π ≈ 3.1415926535897932384626', True); // <-------------------------------------------------
  Test('π ≈ 3.1415926635897932384626', False);

  Test('e ≈ exp(1)', True);
  Test('e ≈ exp(1.01)', False);

  Test('e ≈ π', False);
  Test('0⋅e ≈ 0⋅π', True);

  Test('3 + 5⋅i ≈ 3 + 5⋅i', True);
  Test('3 + 5⋅i ≈ 3 + 6⋅i', False);
  Test('3 + 5⋅i ≈ 3 + 5.001⋅i', False);
  Test('3 + 5⋅i ≈ 2.9999 + 5⋅i', False);
  Test('3 + 5⋅i ≈ 3', False);
  Test('3 + 5⋅i ≈ 5⋅i', False);
  Test('3 + 5⋅i ≈ 3 − 5⋅i', False);
  Test('3 − 5⋅i ≈ 3 − 5⋅i', True);
  Test('3 + 5⋅i ≈ 3.000001 + 5⋅i', False);
  Test('3.0001 + 5⋅i ≈ 3 + 5⋅i', False);

  Test('i ≈ i', True);
  Test('i ≈ -i', False);
  Test('1/i ≈ -i', True);

  Test('1/7 + 6/7 ≈ 1', True);
  Test('2/7 + 6/7 ≈ 1', False);

  Test('2 ≈ 12/9 + 6/9', True);
  Test('2 ≈ 12/9 + 7/9', False);

  Test('1/2 + 1/2 ≈ 1', True);
  Test('1/2 + 1/2 ≈ 3/2 - 1/2', True);
  Test('1/2 + 1/2 ≈ 1.0', True);
  Test('1/2 + 1/2 ≈ 1.0 + i - i', True);

  Test('1/2 ≈ 0.5', True);
  Test('1/2 ≈ 0.5 + i - i', True);
  Test('0.75 ≈ 3/4', True);
  Test('-0.75 ≈ -3/4', True);
  Test('0.75 + i - i ≈ 3/4', True);
  Test('0.75 ≈ 3/4 + i - i', True);

  Test('0 ≈ 0.0', True);
  Test('0 ≈ 1/2 - 1/2', True);
  Test('0 ≈ i - i', True);
  Test('1/2 - 1/2 ≈ 0.0', True);
  Test('1/2 - 1/2 ≈ i - i', True);
  Test('0.0 ≈ i - i', True);

  Test('5 + 0 ≈ 5 + 0.0', True);
  Test('5 + 0 ≈ 5 + 1/2 - 1/2', True);
  Test('5 + 0 ≈ 5 + i - i', True);
  Test('5 + 1/2 - 1/2 ≈ 5 + 0.0', True);
  Test('5 + 1/2 - 1/2 ≈ 5 + i - i', True);
  Test('5 + 0.0 ≈ 5 + i - i', True);

  Test('6 + 0 ≈ 5 + 0.0', False);
  Test('6 + 0 ≈ 5 + 1/2 - 1/2', False);
  Test('6 + 0 ≈ 5 + i - i', False);
  Test('6 + 1/2 - 1/2 ≈ 5 + 0.0', False);
  Test('6 + 1/2 - 1/2 ≈ 5 + i - i', False);
  Test('6 + 0.0 ≈ 5 + i - i', False);

  Test('0 ≈ (.;)', False);
  Test('0.0 ≈ (.;)', False);
  Test('1/2-1/2 ≈ (.;)', False);
  Test('i-i ≈ (.;)', False);

  Test('1/7 + 58/2 − 63/7 + 25/7 ≈ 166/7', True);
  Test('1/7 + 58/2 − 63/7 + 25/7 − 5/7 ≈ 23', True);
  Test('1/7 + 58/2 − 63/7 + 25/7 − 5/7 ≈ 23.0', True);
  Test('1/7 + 58/2 − 63/7 + 25/7 − 5/7 ≈ 23.0 + i − i', True);
  Test('i + 1/7 + 58/2 − 63/7 + 25/7 − 5/7 ≈ 23.0 + i', True);

  Test('59/10 ≈ 5.9', True);
  Test('59/100 ≈ 0.59', True);
  Test('59/1000 ≈ 0.059', True);

  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≈ 7/2', True);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≈ 3.5', True);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≈ 3.5 + 0⋅i', True);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≈ 3.5 + 2⋅i − 2⋅i', True);

  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≈ 7/3', False);
  Test('5⋅i − 3⋅i − 3⋅i + 7/2 ≈ 3.5', False);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≈ 3.6 + 0⋅i', False);
  Test('5⋅i − 3⋅i − 2⋅i + 7/2 ≈ 3.5 + 2⋅i − 2.2⋅i', False);

  Test('1/4 ≈ 0.25', True);
  Test('4⋅0.25 ≈ 1', True);
  Test('4⋅0.25 ≈ 1.0', True);
  Test('4⋅0.25 ≈ 3/4 + 1/4', True);
  Test('4⋅0.25 ≈ 3/4 + 1/4 + i - i', True);
  Test('i - i ≈ 1/7 + 2/7 - 3/7', True);
  Test('i - i ≈ 1/7 + 2/7 - 3.00000001/7', False);

  Test('1/4 ≈ 0.24', False);
  Test('4⋅0.2500001 ≈ 1', False);
  Test('4⋅0.25 ≈ 0.9999999', False);
  Test('4⋅0.25 ≈ 3/4 + 2/4', False);
  Test('4⋅0.25 ≈ 3/4 + 1/4 - i', False);

  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≈ 0', True);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≈ 0.0', True);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≈ 1/2 - 1/2', True);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≈ i - i', True);

  Test('6 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≈ 0', False);
  Test('5 + 3⋅i + 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≈ 0.0', False);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 3⋅i + 1 − 10⋅i ≈ 1/2 - 1/2', False);
  Test('5 + 3⋅i − 2 + 5⋅i − 4 + 2⋅i + 1 − 10⋅i ≈ 7 - i', False);

  Test('5 ≈ ❨5❩', False);
  Test('5 ≈ ❨❨5❩❩', False);
  Test('❨5❩ ≈ ❨❨5❩❩', False);

  Test('0 ≈ ❨0❩', False);
  Test('0 ≈ ❨❨0❩❩', False);
  Test('❨0❩ ≈ ❨❨0❩❩', False);

  Test('0.0 ≈ ❨0❩', False);
  Test('0.0 ≈ ❨❨0❩❩', False);

  Test('1/2-1/2 ≈ ❨0❩', False);
  Test('1/2-1/2 ≈ ❨❨0❩❩', False);
  Test('❨0❩ ≈ ❨❨0❩❩', False);

  Test('i-i ≈ ❨0❩', False);
  Test('i-i ≈ ❨❨0❩❩', False);
  Test('❨0❩ ≈ ❨❨0❩❩', False);

  Test('0 ≈ ❨i-i❩', False);
  Test('0 ≈ ❨❨i-i❩❩', False);
  Test('❨0❩ ≈ ❨❨i-i❩❩', False);

  Test('i-i ≈ ❨i-i❩', False);
  Test('i-i ≈ ❨❨i-i❩❩', False);
  Test('❨i-i❩ ≈ ❨❨0❩❩', False);

  Test('❨i-i❩ ≈ ❨❨i-i❩❩', False);

  Test('❨1, 2, 3, 4, 5❩ ≈ ❨1, 2, 3, 4, 5❩', True);

  Test('❨2, 2, 3, 4, 5❩ ≈ ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, -2, 3, 4, 5❩ ≈ ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3.00000001, 4, 5❩ ≈ ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3, 3.9999999, 5❩ ≈ ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3, 4, 500❩ ≈ ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3, 4❩ ≈ ❨1, 2, 3, 4, 5❩', False);
  Test('❨1, 2, 3, 4, 5❩ ≈ ❨1, 2, 3❩', False);

  Test('❨π, e, sin(1)❩ ≈ ❨π, e, sin(1)❩', True);

  Test('❨5 + i, 2, −i❩ ≈ ❨5 + i, 2, −i❩', True);
  Test('❨5 + i, 2, −i❩ ≈ ❨5 + i, 2, i❩', False);
  Test('❨5 + i, 0, −i❩ ≈ ❨5 + i, 2, −i❩', False);

  Test('❨1, 2, 3, 4, 5❩ ≈ ❨1, 2, 3, 4, 5 + i - i❩', True);
  Test('❨1 + i - i, 2, 3, 4, 5❩ ≈ ❨1, 2, 3, 4, 5❩', True);

  Test('❨1, 2, 3❩ ≈ ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', True);

  Test('❨2, 2, 3❩ ≈ ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, -2, 3❩ ≈ ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, 2, 3.000001❩ ≈ ❨1, 1 + i, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, 2, 3❩ ≈ ❨1, 1, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, 2, 3❩ ≈ ❨1, i, 2❩ + ❨0, 1 − i, 1❩', False);
  Test('❨1, 2, 3❩ ≈ ❨1, 1 + i, 2❩ + ❨0, 1, 1❩', False);

  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);

  Test('❨❨5, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, 2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1.000001❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨3.99999, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0.1, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2.5❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4.000001, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, 2, -1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, -1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, -4, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9.2❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨40, 0, 9❩, ❨−1, 0, 2❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2❩, ❨4, 0❩, ❨−1, 0❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨4, −2, 1❩', False);
  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ 4', False);
  Test('❨❨4❩❩ ≈ ❨4❩', False);
  Test('❨❨4❩❩ ≈ 4', False);
  Test('❨4❩ ≈ 4', False);
  Test('4 ≈ ❨4❩', False);

  Test('❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, i-i, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, i-i, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4 + i-i, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, 0, 2❩❩', True);
  Test('❨❨4, −2, 1❩, ❨4, i-i, 9❩, ❨−1, 0, 2❩❩ ≈ ❨❨4, −2, 1❩, ❨4, 0, 9❩, ❨−1, i-i, 2❩❩', True);

  Test('❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩ ≈ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', True);
  Test('❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩ ≈ ❨❨2+i, 2−3⋅i❩, ❨i, i-i❩❩', True);

  Test('❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩ ≈ ❨❨2+i, 2−3⋅i❩, ❨i, 1❩❩', False);
  Test('❨❨2+i, 2−3⋅i❩, ❨i, 2❩❩ ≈ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨2+i, 2−4⋅i❩, ❨i, 0❩❩ ≈ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨2+i, 2+3⋅i❩, ❨i, 0❩❩ ≈ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨2-i, 2−3⋅i❩, ❨i, 0❩❩ ≈ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨6+i, 2−3⋅i❩, ❨i, 0❩❩ ≈ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);
  Test('❨❨2+i, 2−3⋅i❩, ❨i, 6❩❩ ≈ ❨❨2+i, 2−3⋅i❩, ❨i, 0❩❩', False);

  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≈ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', True);

  Test('❨❨3, 6❩, ❨−5, 3❩❩ ≈ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 7❩, ❨−5, 3❩❩ ≈ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨5, 3❩❩ ≈ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨−5, 3.0000001❩❩ ≈ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≈ ❨❨1, 3 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≈ ❨❨1, 2 + 3⋅i❩, ❨5, 0❩❩ + ❨❨1, 4 − 3⋅i❩, ❨0, 3❩❩', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≈ ❨❨1, 2 + 3⋅i❩, ❨−5, 0❩❩ + ❨❨1, 4❩, ❨3, 3❩❩', False);

  Test('"alpha" ≈ "alpha"', True);
  Test('"alpha" ≈ "beta"', False);
  Test('"alpha" ≈ ""', False);
  Test('"" ≈ ""', True);

  Test('"7" ≈ 7', False);
  Test('"7" ≈ 7.0', False);
  Test('"7" ≈ 7+i-i', False);

  Test('true ≈ true', True);
  Test('true ≈ false', False);
  Test('false ≈ true', False);
  Test('false ≈ false', True);

  Test('false ≈ 0', False);
  Test('false ≈ 0.0', False);
  Test('false ≈ 1/2 - 1/2', False);
  Test('false ≈ i - i', False);
  Test('false ≈ (.;)', False);

  Test('true ≈ 1', False);
  Test('true ≈ 1.0', False);

  Test('true ≈ "true"', False);
  Test('"false" ≈ false', False);
  Test('"false" ≈ ""', False);
  Test('"false" ≈ "false"', True);

  Test('"LiU rocks" ≈ true', False);

  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≈ 5', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≈ "alpha"', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≈ true', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≈ 5.0', False);
  Test('❨❨2, 6❩, ❨−5, 3❩❩ ≈ 0', False);

  Test('❨❨0, 0❩, ❨0, 0❩❩ ≈ 0', False);
  Test('❨0, 0❩ ≈ 0', False);

  Test('struct("a": 42, "b": 100) ≈ struct("a": 42, "b": 100)', True);

  Test('struct("a": 42.000001, "b": 100) ≈ struct("a": 42, "b": 100)', False);
  Test('struct("A": 42, "b": 100) ≈ struct("a": 42, "b": 100)', False);
  Test('struct("a": 43, "b": 100) ≈ struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "c": 100) ≈ struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 101) ≈ struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100) ≈ struct("d": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100) ≈ struct("a": 420, "b": 100)', False);
  Test('struct("a": 42, "b": 100) ≈ struct("a": 42, "e": 100)', False);
  Test('struct("a": 42, "b": 100) ≈ struct("a": 42, "b": 100.01)', False);
  Test('struct("a": 42, "b": 100) ≈ struct("a": 42)', False);
  Test('struct("a": 42) ≈ struct("a": 42, "b": 100)', False);
  Test('struct("b": 100) ≈ struct("a": 42, "b": 100)', False);
  Test('struct("b": 100, "a": 42) ≈ struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100) ≈ struct("a": 100, "b": 42)', False);
  Test('struct("a": 42, "b": 100) ≈ struct("a": 42, "b": 100, "c": 1000)', False);
  Test('struct("a": 42, "b": 100, "c": 1000) ≈ struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100, "c": 1000) ≈ struct("a": 42, "b": 100, "c": 1000)', True);
  Test('struct("a": 42, "b": 100, "c": "1000") ≈ struct("a": 42, "b": 100, "c": 1000)', False);
  Test('struct("a": 42, "b": 100, "c": "1000") ≈ struct("a": 42, "b": 100, "c": "1000")', True);
  Test('struct("a": 42, "b": 100, "c": "1000") ≈ struct("a": 42, "b": 100)', False);
  Test('struct("a": 42, "b": 100, "c": "1000") ≈ struct("a": 42)', False);
  Test('struct("a": 42, "b": 100) ≈ struct("a": 42, "b": 100, "c": "1000")', False);
  Test('struct("a": 42) ≈ struct("a": 42, "b": 100, "c": "1000")', False);

  Test('struct("a": 42, "b": 100) ≈ 42', False);
  Test('struct("a": 42, "b": 100) ≈ "a"', False);
  Test('struct("a": 42, "b": 100) ≈ ("a": 42)', False);
  Test('struct("a": 42, "b": 100) ≈ true', False);
  Test('struct("a": 42, "b": 100) ≈ ❨❨2, 6❩, ❨−5, 3❩❩', False);
  Test('struct("a": 42, "b": 100) ≈ 0', False);
  Test('struct("a": 42, "b": 100) ≈ 0.0', False);
  Test('struct("a": 42, "b": 100) ≈ i-i', False);
  Test('struct("a": 42, "b": 100) ≈ (.;)', False);

  Test('color("red") ≈ color("red")', True);
  Test('color("red") ≈ rgb(1, 0, 0)', True);

  Test('color("red") ≈ rgb(1, 0.000001, 0)', True); // two colours are considered almost equal if their RGBA values are the same as bytes
  Test('color("red") ≈ rgb(0.7, 0, 0)', False);
  Test('color("red") ≈ rgb(1, 1, 0)', False);
  Test('color("red") ≈ rgb(1, 1, 1)', False);
  Test('color("red") ≈ rgb(0, 0, 0)', False);
  Test('color("green") ≈ rgb(1, 0, 0)', False);

  Test('color("red") ≈ "red"', False);
  Test('color("red") ≈ 16#FF0000', False);
  Test('color("red") ≈ true', False);
  Test('color("red") ≈ color("green")', False);

  Test('color("black") ≈ rgb(0, 0, 0)', True);

  Test('color("black") ≈ rgb(0.0001, 0, 0)', True);
  Test('color("black") ≈ rgb(0, 0.0001, 0)', True);
  Test('color("black") ≈ rgb(0, 0, 0.001)', True);
  Test('color("black") ≈ rgb(0, 1, 0)', False);
  Test('color("black") ≈ rgb(1, 1, 0)', False);
  Test('color("black") ≈ rgb(1, 1, 1)', False);
  Test('color("black") ≈ rgb(0, 1, 1)', False);

  Test('color("black") ≈ 0', False);
  Test('color("black") ≈ 0.0', False);

  Test('color("white") ≈ rgb(1, 1, 1)', True);

  Test('color("white") ≈ rgb(0.9999, 1, 1)', True);
  Test('color("white") ≈ rgb(0, 1, 1)', False);
  Test('color("white") ≈ rgb(1, 0.999, 1)', True);
  Test('color("white") ≈ rgb(1, 1, 0)', False);
  Test('color("white") ≈ rgb(1, 1, 0)', False);
  Test('color("white") ≈ rgb(0.3, 0.5, 0.7)', False);
  Test('color("white") ≈ rgb(0.1, 0.1, 0.1)', False);

  Test('''(1, 2, 3) ≈ ''(1, 2, 3)', True);
  Test('''(1, 2.0, 3) ≈ ''(1, 2, 3)', True);
  Test('''(1, 2, 3) ≈ ''(1, 2, 3.0)', True);
  Test('''(1, 2, 3) ≈ ''(1/2 + 1/2, 2, 3)', True);
  Test('''(1., 2., 3.) ≈ ''(1/2 + 1/2, 2, 3 + i - i)', True);
  Test('''(1, 2 + i - i, 3) ≈ ''(1, 2, 3)', True);
  Test('''(1, 2 + i - i, 3.) ≈ ''(1/2 + 1/2, 2, 3)', True);
  Test('''(1, 2., 3 + i) ≈ ''(1, 2, 3 + i)', True);

  Test('''(2, 2, 3) ≈ ''(1, 2, 3)', False);
  Test('''(1, 1, 3) ≈ ''(1, 2, 3)', False);
  Test('''(1, 2, -3) ≈ ''(1, 2, 3)', False);
  Test('''(1, 2, 3.01) ≈ ''(1, 2, 3)', False);
  Test('''(1, 2, 3) ≈ ''(10, 2, 3)', False);
  Test('''(1, 2, 3) ≈ ''(1, 0.2, 3)', False);
  Test('''(1, 2, 3) ≈ ''(1, 2, 3.000001)', False);
  Test('''(1, 2, 3) ≈ ''(1, 2, 2.999999)', False);
  Test('''(2, 2, 3) ≈ ''(1, 2, 3/2)', False);
  Test('''(2, 2, 3) ≈ ''(1, 2, 3 + i)', False);

  Test('''(1, 2, "test") ≈ ''(1, 2, 3)', False);
  Test('''(1, 2, "test") ≈ ''(1, 2, "test")', True);
  Test('''(1, 2, "test") ≈ ''(1/2 + 1/2, 2.0, "te"+"st")', True);

  Test('''(1, 2, 3) ≈ ''(1, 2, 3, 4)', False);
  Test('''(1, 2, 3) ≈ ''(1, 2)', False);
  Test('''(1, 2, 3) ≈ ''(3, 2, 1)', False);

  Test('''(1, 2, 3) ≈ ''(1, 2, 3)', True);

  Test('''(1, 2, ❨1, 2❩) ≈ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', True);

  Test('''(2, 2, ❨1, 1❩) ≈ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', False);
  Test('''(1, 2.000001, ❨1, 2❩) ≈ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', False);
  Test('''(1, 2, ❨1, 2❩) ≈ ''(1/2 + 1/2, 1.1 + 1.0, ❨1 + i − i, 2❩)', False);
  Test('''(1, 2, ❨1, 2❩) ≈ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i, 2❩)', False);
  Test('''(1, 2, ❨1, 2❩) ≈ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2, 3❩)', False);
  Test('''(1, 2, ❨1, 2❩) ≈ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩, 5)', False);
  Test('''(1, 2, ❨1, 2❩) ≈ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩, "test")', False);
  Test('''(1, 2, ❨1, 2❩, false) ≈ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i, 2❩)', False);
  Test('''(1, 2, ❨1, 2❩) ≈ ''(1/2 + 1/2, 1.0 + 1.0, ❨1 + i − i❩)', False);
  Test('''(1, 2, ❨1, 2❩) ≈ ''(1/2 + 1/2, 1.0 + 1.0, 1 + i − i, 2)', False);

  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);

  Test('''(3, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π^2, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨2, 1❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨i, 0❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, true, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "dog", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("green")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1/2 + 1/2, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨i, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(0.9, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), color("green")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), true) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), 0) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("gold")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', False);

  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), 0) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0), 0)', True);
  Test('''(2, π, ❨1, 2❩, ❨❨0, i❩, ❨i, 0❩❩, false, "cat", color("red"), 0) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0), (.;))', False);

  Test('''(1.9 + 0.1 + i - i, π, ❨1+i-i, 2❩, ❨❨0, i❩, ❨i, 0❩❩, not(true), "cat", color("red")) ≈ ''(1.0 + 1.0, π + i − i, ❨1, 2 + i − i❩, ❨❨0, i❩, ❨i, 0❩❩, true ∧ false, "c"+"at", rgb(1, 0, 0))', True);

  if SecondsBetween(Now, Tomorrow) < 2 then
    Sleep(3000);

  Test('date() ≈ date()', True);
  Test('date() ≈ tomorrow()', False);
  Test('date() ≈ tomorrow()', False);

  Test('''() ≈ ''()', True);
  Test('''() ≈ (.;)', False);
  Test('''("dog", "cat") ≈ {"dog", "cat"}', False);
  Test('''() ≈ {}', False);

  Test('4 ≈ {4}', False);
  Test('4 ≈ ''(4)', False);
  Test('4 ≈ true', False);
  Test('true ≈ {4}', False);
  Test('false ≈ {}', False);

  Test('0 ≈ ZeroVector(100)', False); {arguably, could make this return true, at least if LHS and RHS are swapped}
  Test('0 ≈ ZeroMatrix(100)', False); {and this}

  Test('BinaryData("asterism") ≈ BinaryData("asterism")', True);
  Test('BinaryData("asterism") ≈ BinaryData("asterisk")', False);

  Test('BinaryData(0) ≈ BinaryData(0)', True);
  Test('BinaryData(1000) ≈ BinaryData(1000)', True);
  Test('BinaryData(1000) ≈ BinaryData(10000)', False);
  Test('BinaryData(1000) ≈ BinaryData(1001)', False);
  Test('BinaryData(1000) ≈ BinaryData(1000.001)', False);

  Test('BinaryData("rat") ≈ BinaryData(1000)', False);

  Test('BinaryData(1000) ≈ 10', False);
  Test('BinaryData(1000) ≈ "cat"', False);
  Test('BinaryData(1000) ≈ color("red")', False);
  Test('BinaryData(1000) ≈ {1, 2, 3}', False);
  Test('BinaryData(1000) ≈ true', False);
  Test('BinaryData(1000) ≈ now()', False);

  Test('BinaryData(0) ≈ BinaryData(0.0)', {$IFDEF CPU64BITS}True{$ELSE}False{$ENDIF});

  Test('5 ≈ "cat"', False);
  Test('5.5 ≈ ''(1, 2)', False);
  Test('1 + i ≈ {"dog", "cat"}', False);
  Test('{"dog"} ≈ "dog"', False);
  Test('5 + i ≈ "rat"', False);

  // - pixmaps, sounds: postponed

  // Logical and (logical conjunction)

  Test('false ∧ false', false);
  Test('false ∧ true', false);
  Test('true ∧ false', false);
  Test('true ∧ true', true);

  Test('false ∧ false ∧ false', false);
  Test('false ∧ false ∧ true', false);
  Test('false ∧ true ∧ false', false);
  Test('false ∧ true ∧ true', false);
  Test('true ∧ false ∧ false', false);
  Test('true ∧ false ∧ true', false);
  Test('true ∧ true ∧ false', false);
  Test('true ∧ true ∧ true', true);

  Test('true ∧ true ∧ true ∧ true ∧ true ∧ true ∧ true ∧ true', true);
  Test('true ∧ true ∧ true ∧ true ∧ true ∧ true ∧ true ∧ false', false);
  Test('true ∧ false ∧ true ∧ true ∧ true ∧ true ∧ true ∧ true', false);
  Test('true ∧ true ∧ false ∧ true ∧ true ∧ true ∧ true ∧ true', false);
  Test('true ∧ true ∧ true ∧ false ∧ false ∧ true ∧ true ∧ true', false);
  Test('true ∧ false ∧ true ∧ true ∧ true ∧ true ∧ true ∧ true', false);
  Test('false ∧ true ∧ true ∧ true ∧ true ∧ true ∧ true ∧ true', false);
  Test('true ∧ true ∧ true ∧ true ∧ true ∧ false ∧ false ∧ true', false);
  Test('false ∧ true ∧ false ∧ false ∧ false ∧ true ∧ true ∧ true', false);
  Test('false ∧ false ∧ false ∧ false ∧ false ∧ true ∧ false ∧ false', false);
  Test('false ∧ false ∧ false ∧ false ∧ false ∧ false ∧ false ∧ false', false);

  Test('false ∧ error()', false);
  Test('true ∧ error()', failure);

  Test('false ∧ true ∧ true ∧ true ∧ true ∧ true ∧ true ∧ error()', false);
  Test('true ∧ false ∧ true ∧ true ∧ true ∧ true ∧ true ∧ error()', false);
  Test('true ∧ true ∧ true ∧ true ∧ true ∧ true ∧ false ∧ error()', false);
  Test('true ∧ true ∧ true ∧ true ∧ false ∧ true ∧ true ∧ error()', false);
  Test('true ∧ true ∧ true ∧ true ∧ true ∧ true ∧ true ∧ error()', failure);

  Test('a ≔ 123; false ∧ delete(a); a − 1', 122);
  Test('delete(a)', success);

  Test('n ≔ 3; IsNumber(n) ∧ ¬IsZero(n) ∧ (10/n ≥ 5)', false);
  Test('n ≔ 2; IsNumber(n) ∧ ¬IsZero(n) ∧ (10/n ≥ 5)', true);
  Test('n ≔ 1; IsNumber(n) ∧ ¬IsZero(n) ∧ (10/n ≥ 5)', true);
  Test('n ≔ 0; IsNumber(n) ∧ ¬IsZero(n) ∧ (10/n ≥ 5)', false);
  Test('n ≔ "cat"; IsNumber(n) ∧ ¬IsZero(n) ∧ (10/n ≥ 5)', false);

  Test('n ≔ 0; (10/n ≥ 5) ∧ IsNumber(n) ∧ ¬IsZero(n)', failure);
  Test('n ≔ "dog"; (10/n ≥ 5) ∧ IsNumber(n) ∧ ¬IsZero(n)', failure);

  Test('delete(n)', success);

  Test('"cat" ∧ "dog"', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('true ∧ 101', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('false ∧ 101', false);    
  Test('true ∧ true ∧ 13', failure, 'An object of type boolean was expected as argument 3, but an object of type integer was given.');
  Test('false ∧ true ∧ 13', false);
  Test('true ∧ false ∧ 13', false);
  Test('false ∧ false ∧ 13', false);

  // Logical or (logical disjunction)

  Test('false ∨ false', false);
  Test('false ∨ true', true);
  Test('true ∨ false', true);
  Test('true ∨ true', true);

  Test('false ∨ false ∨ false', false);
  Test('false ∨ false ∨ true', true);
  Test('false ∨ true ∨ false', true);
  Test('false ∨ true ∨ true', true);
  Test('true ∨ false ∨ false', true);
  Test('true ∨ false ∨ true', true);
  Test('true ∨ true ∨ false', true);
  Test('true ∨ true ∨ true', true);

  Test('false ∨ false ∨ false ∨ false ∨ false ∨ false ∨ false', false);
  Test('false ∨ false ∨ false ∨ false ∨ false ∨ false ∨ true', true);
  Test('true ∨ false ∨ false ∨ false ∨ false ∨ false ∨ false', true);
  Test('false ∨ true ∨ false ∨ false ∨ false ∨ false ∨ false', true);
  Test('false ∨ false ∨ true ∨ false ∨ false ∨ false ∨ false', true);
  Test('false ∨ false ∨ false ∨ true ∨ false ∨ false ∨ false', true);
  Test('false ∨ false ∨ false ∨ false ∨ true ∨ false ∨ false', true);
  Test('false ∨ false ∨ false ∨ false ∨ false ∨ true ∨ false', true);
  Test('true ∨ false ∨ false ∨ false ∨ true ∨ false ∨ false', true);
  Test('true ∨ true ∨ true ∨ false ∨ false ∨ false ∨ false', true);
  Test('false ∨ false ∨ false ∨ true ∨ true ∨ true ∨ true', true);
  Test('true ∨ true ∨ true ∨ true ∨ true ∨ true ∨ true', true);  

  Test('true ∨ error()', true);
  Test('false ∨ error()', failure);

  Test('false ∨ false ∨ false ∨ false ∨ false ∨ false ∨ error()', failure);
  Test('false ∨ false ∨ false ∨ true ∨ false ∨ false ∨ error()', true);
  Test('false ∨ false ∨ false ∨ false ∨ false ∨ true ∨ error()', true);
  Test('true ∨ false ∨ false ∨ false ∨ false ∨ false ∨ error()', true);
  Test('true ∨ true ∨ true ∨ true ∨ true ∨ true ∨ error()', true);

  Test('n ≔ "rat"; ¬IsNumber(n) ∨ IsZero(n) ∨ (10/n ≥ 5)', true);
  Test('n ≔ 0; ¬IsNumber(n) ∨ IsZero(n) ∨ (10/n ≥ 5)', true);
  Test('n ≔ 1; ¬IsNumber(n) ∨ IsZero(n) ∨ (10/n ≥ 5)', true);
  Test('n ≔ 2; ¬IsNumber(n) ∨ IsZero(n) ∨ (10/n ≥ 5)', true);
  Test('n ≔ 3; ¬IsNumber(n) ∨ IsZero(n) ∨ (10/n ≥ 5)', false);

  Test('n ≔ "rat"; (10/n ≥ 5) ∨ ¬IsNumber(n) ∨ IsZero(n)', failure);
  Test('n ≔ 0; (10/n ≥ 5) ∨ ¬IsNumber(n) ∨ IsZero(n)', failure);

  Test('"cat" ∨ "dog"', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('false ∨ 101', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('true ∨ 101', true);
  Test('false ∨ false ∨ 13', failure, 'An object of type boolean was expected as argument 3, but an object of type integer was given.');
  Test('false ∨ true ∨ 13', true);
  Test('true ∨ false ∨ 13', true);
  Test('true ∨ true ∨ 13', true);

  // Logical not (logical negation)

  Test('¬false', true);
  Test('¬true', false);

  Test('¬"test"', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('¬{1, 2, 3}', failure, 'An object of type boolean was expected as argument 1, but an object of type set was given.');

  Test('¬¬true', true);
  Test('¬¬false', false);

  Test(StringOfChar(Char('¬'), 100) + 'true', true);
  Test(StringOfChar(Char('¬'), 101) + 'true', false);
  Test(StringOfChar(Char('¬'), 102) + 'true', true);
  Test(StringOfChar(Char('¬'), 103) + 'true', false);

  // Logical exclusive or (xor)

  Test('false ⊻ false', false);
  Test('false ⊻ true', true);
  Test('true ⊻ false', true);
  Test('true ⊻ true', false);

  Test('true ⊻ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('false ⊻ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('"cat" ⊻ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"dog" ⊻ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"rabbit" ⊻ 5', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');

  // Logical nand

  Test('false ⊼ false', true);
  Test('false ⊼ true', true);
  Test('true ⊼ false', true);
  Test('true ⊼ true', false);

  Test('true ⊼ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('false ⊼ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('"cat" ⊼ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"dog" ⊼ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"rabbit" ⊼ 5', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');  

  // Logical nor

  Test('false ⊽ false', true);
  Test('false ⊽ true', false);
  Test('true ⊽ false', false);
  Test('true ⊽ true', false);

  Test('true ⊽ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('false ⊽ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('"cat" ⊽ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"dog" ⊽ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"rabbit" ⊽ 5', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');    

  // Right double arrow (implication, implies right)

  Test('false ⇒ false', true);
  Test('false ⇒ true', true);
  Test('true ⇒ false', false);
  Test('true ⇒ true', true);

  Test('true ⇒ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('false ⇒ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('"cat" ⇒ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"dog" ⇒ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"rabbit" ⇒ 5', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');

  // Left double arrow (implication, implies left)

  Test('false ⇐ false', true);
  Test('false ⇐ true', false);
  Test('true ⇐ false', true);
  Test('true ⇐ true', true);

  Test('true ⇐ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('false ⇐ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('"cat" ⇐ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"dog" ⇐ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"rabbit" ⇐ 5', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');

  // Left and right double arrow (equivalence, equivalent)

  Test('false ⇔ false', true);
  Test('false ⇔ true', false);
  Test('true ⇔ false', false);
  Test('true ⇔ true', true);

  Test('true ⇔ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('false ⇔ 12', failure, 'An object of type boolean was expected as argument 2, but an object of type integer was given.');
  Test('"cat" ⇔ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"dog" ⇔ true', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');
  Test('"rabbit" ⇔ 5', failure, 'An object of type boolean was expected as argument 1, but an object of type string was given.');

  // Union

  Test('{1, 2, 3} ∪ {2, 3, 4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{1, 2, 3} ∪ {3, 4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{1, 2, 3} ∪ {2, 4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{1, 2, 3} ∪ {4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{1, 2} ∪ {2, 3, 4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{1} ∪ {2, 3, 4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{5, 1} ∪ {5, 1, 2, 3, 4}', intset([1, 2, 3, 4, 5]));
  Test('{5, 1, 2, 2, 2, 1} ∪ {5, 1, 2, 3, 4}', intset([1, 2, 3, 4, 5]));
  Test('{5, 1, 2, 2, 2, 1, 4} ∪ {5, 1, 1, 2, 3, 4}', intset([1, 2, 3, 4, 5]));

  Test('{1, 2, 3, 4, 5} ∪ {}', intset([1, 2, 3, 4, 5]));
  Test('{1, 2, 3, 4, 5, 5, 5} ∪ {}', intset([1, 2, 3, 4, 5]));
  Test('{3, 4, 1, 2, 2, 4, 5, 1, 2, 3} ∪ {}', intset([1, 2, 3, 4, 5]));

  Test('{} ∪ {1, 2, 3, 4, 5}', intset([1, 2, 3, 4, 5]));

  Test('{1, 2} ∪ {3} ∪ {4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{1, 2, 3} ∪ {3} ∪ {4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{1} ∪ {2, 3} ∪ {4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{1, 4, 5, 4, 2} ∪ {3} ∪ {4, 5}', intset([1, 2, 3, 4, 5]));
  Test('{1, 2} ∪ {1, 2, 3} ∪ {1, 2, 4, 5}', intset([1, 2, 3, 4, 5]));

  Test('{1, 2} ∪ {1, 2, 3} ∪ {1, 2, 4, 5, 6}', intset([1, 2, 3, 4, 5, 6]));
  Test('{6, 1, 2} ∪ {1, 2, 3} ∪ {1, 2, 4, 5}', intset([1, 2, 3, 4, 5, 6]));
  Test('{1, 2} ∪ {1, 2, 6, 3} ∪ {1, 2, 4, 5}', intset([1, 2, 3, 4, 5, 6]));
  Test('{1, 2} ∪ {1, 2, 3} ∪ {1, 2, 4, 5} ∪ {6}', intset([1, 2, 3, 4, 5, 6]));

  Test('{1} ∪ {2} ∪ {3} ∪ {4} ∪ {5}', intset([1, 2, 3, 4, 5]));
  Test('{1} ∪ {} ∪ {2} ∪ {3} ∪ {4} ∪ {5}', intset([1, 2, 3, 4, 5]));
  Test('{4} ∪ {1} ∪ {2} ∪ {3} ∪ {5, 4} ∪ {2}', intset([1, 2, 3, 4, 5]));

  Test('{4} ∪ {1} ∪ {2} ∪ {3} ∪ {5, 4} ∪ {2, 10}', intset([1, 2, 3, 4, 5, 10]));
  Test('{10, 4} ∪ {1} ∪ {2} ∪ {3} ∪ {5, 4} ∪ {2}', intset([1, 2, 3, 4, 5, 10]));
  Test('{4} ∪ {1} ∪ {2} ∪ {3, 10} ∪ {5, 4} ∪ {2}', intset([1, 2, 3, 4, 5, 10]));
  Test('{4} ∪ {1} ∪ {10} ∪ {2, 3} ∪ {5, 4} ∪ {}', intset([1, 2, 3, 4, 5, 10]));

  Test('{42, "cat"} ∪ {10, "dog"} ∪ {"rat"} ∪ {100, 101, 102}', asoset([10, 42, 100, 101, 102, 'cat', 'dog', 'rat']));
  Test('{42, "cat"} ∪ {42, 10, "dog"} ∪ {101, "rat"} ∪ {101, 101, 100, 102}', asoset([10, 42, 100, 101, 102, 'cat', 'dog', 'rat']));

  Test('{10} ∪ {10.0}', intset([10]));
  Test('{10.0} ∪ {10}', intset([10]));
  Test('#({10} ∪ {10.0})', 1);
  Test('#({10.0} ∪ {10})', 1);
  Test('({10} ∪ {10.0}) @ type', asoset(['integer']));
  Test('({10.0} ∪ {10}) @ type', asoset(['integer']));

  Test('{} ∪ {}', intset([]));
  Test('{} ∪ {} ∪ {}', intset([]));
  Test('{} ∪ {} ∪ {7}', intset([7]));

  Test('{color("red"), color("lime"), color("blue")} ∪ {color("yellow"), color("black")}',
    asosetex([
      ASO(TRGB.Create(1, 0, 0)),
      ASO(TRGB.Create(0, 1, 0)),
      ASO(TRGB.Create(0, 0, 1)),
      ASO(TRGB.Create(1, 1, 0)),
      ASO(TRGB.Create(0, 0, 0))
    ]));
  Test('{color("red"), color("lime"), color("blue")} ∪ {color("red"), color("yellow"), color("black")}',
    asosetex([
      ASO(TRGB.Create(1, 0, 0)),
      ASO(TRGB.Create(0, 1, 0)),
      ASO(TRGB.Create(0, 0, 1)),
      ASO(TRGB.Create(1, 1, 0)),
      ASO(TRGB.Create(0, 0, 0))
    ]));
  Test('{color("red"), color("lime"), color("blue")} ∪ {color("red"), color("yellow"), color("black"), color("red"), color("lime")}',
    asosetex([
      ASO(TRGB.Create(1, 0, 0)),
      ASO(TRGB.Create(0, 1, 0)),
      ASO(TRGB.Create(0, 0, 1)),
      ASO(TRGB.Create(1, 1, 0)),
      ASO(TRGB.Create(0, 0, 0))
    ]));

  Test('{❨1, 0, 0❩} ∪ {❨0, 1, 0❩} ∪ {❨0, 0, 1❩}',
    asosetex([
      ASO(ASR3(1, 0, 0)),
      ASO(ASR3(0, 1, 0)),
      ASO(ASR3(0, 0, 1))
    ]));
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩} ∪ {❨0, 1, 0❩, ❨0, 0, 1❩, ❨0, i−i, 1❩}',
    asosetex([
      ASO(ASR3(1, 0, 0)),
      ASO(ASR3(0, 1, 0)),
      ASO(ASR3(0, 0, 1))
    ]));

  Test('{IdentityMatrix(1), IdentityMatrix(2), IdentityMatrix(3), IdentityMatrix(4)} ∪ {"rat", IdentityMatrix(4), IdentityMatrix(5)}',
    asosetex([
      ASO(IdentityMatrix(1)),
      ASO(IdentityMatrix(2)),
      ASO(IdentityMatrix(3)),
      ASO(IdentityMatrix(4)),
      ASO(IdentityMatrix(5)),
      ASO('rat')
    ]));

  Test('{10, 20} ∪ "test"', failure, 'An object of type set was expected as argument 2, but an object of type string was given.');
  Test('123 ∪ {"test"}', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');

  // Intersection

  Test('{10, 20, 30, 40} ∩ {30, 40, 50, 60}', intset([30, 40]));
  Test('{10, 20, 30, 40} ∩ {30, 40, 50, 60, 70, 80}', intset([30, 40]));
  Test('{10, 20, 30, 40} ∩ {30, 40, 50, 60, 70, 80, 20}', intset([20, 30, 40]));
  Test('{10, 20, 30} ∩ {30, 40, 50}', intset([30]));
  Test('{10, 20, 30} ∩ {40, 50, 60}', intset([]));
  Test('{10, 20, 30} ∩ {}', intset([]));
  Test('{} ∩ {}', intset([]));
  Test('{} ∩ {} ∩ {}', intset([]));
  Test('{} ∩ {} ∩ {} ∩ {"rat"}', intset([]));
  Test('{10, "rat"} ∩ {"rat", false} ∩ {color("red"), "rat"} ∩ {"rat", IdentityMatrix(100)}', asoset(['rat']));

  Test('{10, 20, 30, 40, 50} ∩ {10, 30, 50, 70} ∩ {30, 31, 32}', intset([30]));
  Test('{10, 20, 30, 40, 50} ∩ {10, 30, 50, 70} ∩ {10, 30, 31, 32}', intset([10, 30]));
  Test('{10, 20, 30, 40, 50} ∩ {10, 30, 50, 70} ∩ {10, 30, 31, 32} ∩ {10}', intset([10]));
  Test('{10, 20, 30, 40, 50} ∩ {10, 30, 50, 70} ∩ {10, 30, 31, 32} ∩ {20}', intset([]));
  Test('{10, 20, 30, 40, 50} ∩ {10, 30, 50, 70} ∩ {10, 30, 31, 32} ∩ {10} ∩ {11}', intset([]));

  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90, 100} ∩ {10, 20, 25.0}', intset([10, 20]));
  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90, 100} ∩ {10.0, 20, 25.0}', intset([10, 20]));
  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90, 100} ∩ {10.0, 20.0, 25.0}', intset([10, 20]));

  Test('{10, true, i, "cat", "dog", 0, "rat", "table", 123} ∩ {"cat", "dog", "rat", "rabbit", "hourse"}', asoset(['cat', 'dog', 'rat']));
  Test('{0, 123} ∩ {10, true, i, "cat", "dog", 0, "rat", "table", 123} ∩ {0, 10, 20, 30}', asoset([0]));
  Test('{0, 123} ∩ {10, true, i, "cat", "dog", i-i, "rat", "table", 123} ∩ {0.0, 10, 20, 30}', asoset([0]));

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨1, 1, 0❩, ❨0, 1, 1❩, ❨0, 0, 1❩} ∩ {❨0, 0, 0❩, ❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}',
    asosetex([
      ASO(ASR3(1, 0, 0)),
      ASO(ASR3(0, 1, 0)),
      ASO(ASR3(0, 0, 1))
    ]));
  Test('{"dog", 5, ❨1, 0, 0❩, ❨1, 0, 0❩, ❨0, 1, 0❩, ❨1, 1, 0❩, ❨0, 1, 1❩, ❨0, 0, 1❩} ∩ {❨0, 0, 0❩, ❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "rat"}',
    asosetex([
      ASO(ASR3(1, 0, 0)),
      ASO(ASR3(0, 1, 0)),
      ASO(ASR3(0, 0, 1))
    ]));

  Test('{ZeroMatrix(2), ZeroMatrix(3), ZeroMatrix(4)} ∩ {ZeroMatrix(3), 1024}',
    asosetex([
      ASO(ZeroMatrix(3))
    ]));

  Test('{ZeroMatrix(2), ZeroMatrix(3), ZeroMatrix(3), ZeroMatrix(4)} ∩ {ZeroMatrix(3), 1024} ∩ {ZeroMatrix(3)}',
    asosetex([
      ASO(ZeroMatrix(3))
    ]));

  Test('{ZeroMatrix(2), ZeroMatrix(3), ZeroMatrix(4)} ∩ {ZeroMatrix(3), 1024} ∩ {ZeroMatrix(2)}',
    asosetex([

    ]));

  Test('{10, 20, 30} ∩ true', failure, 'An object of type set was expected as argument 2, but an object of type boolean was given.');
  Test('7 ∩ {"rat", "dog"}', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('color("red") ∩ color("yellow")', failure, 'An object of type set was expected as argument 1, but an object of type color was given.');

  // Set difference

  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90} ∖ {50, 60, 70}', intset([10, 20, 30, 40, 80, 90]));
  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90} ∖ {50, 60, 70, 80, 90}', intset([10, 20, 30, 40]));
  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90} ∖ {10, 50, 60, 70, 80, 90}', intset([20, 30, 40]));
  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90} ∖ {10, 20, 25, 50, 60, 70, 80, 90}', intset([30, 40]));
  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90} ∖ {10, 20, 25, 50, 55, 60, 65, 70, 80, 90}', intset([30, 40]));
  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90} ∖ {10, 20, 25, 50, 55, 60, 65, 70, 80, 90, 30}', intset([40]));
  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90} ∖ {10, 20, 25, 50, 55, 60, 65, 70, 80, 90, 30, 40.0}', intset([]));
  Test('{10, 20, 30, 40, 50, 60, 70, 80, 90} ∖ {10, 20, 25, 50, 55, 60, 65, 70, 80, 90, 30, 40.0, 42}', intset([]));

  Test('{1.1, 1.2, 1.3, 1.4, 1.5} ∖ {0.5, 1.5, 2.5}', asoset([1.1, 1.2, 1.3, 1.4]));

  Test('{1, 2, 3, 4, 5} ∖ {1, 3, 5, 7, 9, 11}', intset([2, 4]));
  Test('{1, 2, 3, 4, 5} ∖ {10, 20, 30, "rat"}', intset([1, 2, 3, 4, 5]));
  Test('{1, 2, 3, 4, 5} ∖ {}', intset([1, 2, 3, 4, 5]));

  Test('{5, 2, 3, 4, 1, 5, 1} ∖ {2, 4, 4, 4, 2, 4, 2, 4}', intset([1, 3, 5]));

  Test('{1, 2, 3, 4, 5} ∖ {2, 4, 6, 8} ∖ {5} ∖ {50, 100}', intset([1, 3]));

  Test('{1, 2, 3, 4, 5} ∖ {2, 4, 6, 8} ∖ {5} ∖ {50, 100} ∖ {-5, -3, -1, 1, 3, 5}', intset([]));
  Test('{1, 2, 3, 4, 5} ∖ {2, 4, 6, 8} ∖ {5} ∖ {50, 100} ∖ {-5, -3, -1, 1, 3, 5} ∖ {42}', intset([]));

  Test('{1, 2, 3, "dog", 4, 5, "rat", 6, 7, "cat", 8, 9} ∖ {1, 2, 3, 4, 5, 6, 7, 8, 9}', asoset(['dog', 'cat', 'rat']));
  Test('{1, 2, 3, "dog", 4, 5, "rat", 6, 7, "cat", 8, 9} ∖ {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12}', asoset(['dog', 'cat', 'rat']));

  Test('{10, i, 20, -i} ∖ {10, 20, 30, 40, 50}', asosetex([ASO(ImaginaryUnit), ASO(-ImaginaryUnit)]));

  Test('{} ∖ {10, 20, 30}', intset([]));
  Test('{} ∖ {}', intset([]));

  Test('{1, 2, 3, 4, 5} ∖ {"cat", "dog"}', intset([1, 2, 3, 4, 5]));

  Test('{IdentityMatrix(2), IdentityMatrix(3), IdentityMatrix(4)} ∖ {IdentityMatrix(2)}',
    asosetex([
      ASO(IdentityMatrix(3)),
      ASO(IdentityMatrix(4))
    ]));

  Test('{1, 2, 3} ∖ 2', failure, 'An object of type set was expected as argument 2, but an object of type integer was given.');
  Test('"dog" ∖ 2', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  // Symmetric difference

  Test('{1, 2, 3, 4, 5, 6, 7, 8, 9} ∆ {6, 7, 8, 9, 10, 11, 12, 13}', intset([1, 2, 3, 4, 5, 10, 11, 12, 13]));

  Test('{1, 2, 3, 4, 5} ∆ {1, 10, 100}', intset([2, 3, 4, 5, 10, 100]));
  Test('{2, 3, 5, 7, 11, 13, 17, 19} ∆ {1, 3, 5, 7, 9, 11, 13, 15, 17, 19}', intset([1, 2, 9, 15]));
  Test('{2, 3, 5, 7, 11, 13, 17, 19} ∆ {}', intset([2, 3, 5, 7, 11, 13, 17, 19]));
  Test('{} ∆ {0, 1, 2, 3, 4, 5}', intset([0, 1, 2, 3, 4, 5]));
  Test('{} ∆ {}', intset([]));
  Test('{0, 1, 2, 3, 4, 5} ∆ {0, 1, 2, 3, 4, 5}', intset([]));
  Test('{0, 1, 2, 3, 4, 5, 7} ∆ {0, 1, 2, 3, 4, 5}', intset([7]));
  Test('{0, 1, 2, 3, 4, 5} ∆ {0, 1, 2, 3, 4, 5, 7}', intset([7]));
  Test('{0, 1, 2, 3, 4, 5, 5} ∆ {0, 1, 2, 3, 4, 5}', intset([]));
  Test('{0, 1, 2, 3, 4, 5, 5, 5, 5} ∆ {5, 5, 5, 0, 1, 2, 3, 4, 5}', intset([]));
  Test('{0, 1, 2, 3, 4, 5, 5, 5, 5, 6} ∆ {5, 5, 5, -1, 0, 1, 2, 3, 4, 5}', intset([-1, 6]));

  Test('{"human", "dog", "cat", "wolf", "tiger", "lion", "rat"} ∆ {"rabbit", "rat", "human", "dog", "guinea pig", "horse", "cow", "sheep"}',
    asoset(['cat', 'wolf', 'tiger', 'lion', 'rabbit', 'guinea pig', 'horse', 'cow', 'sheep']));

  Test('{1, 2, 3} ∆ 3', failure, 'An object of type set was expected as argument 2, but an object of type integer was given.');
  Test('true ∆ 3', failure, 'An object of type set was expected as argument 1, but an object of type boolean was given.');

  // Complement

  Test('try(delete(universe))', null);

  Test('∁{1, 2, 3}', failure, 'Unknown identifier "universe".');
  Test('universe ≔ {1, 2, 3, 4, 5}; ∁{1, 2, 3}', intset([4, 5]));
  Test('∁{1, 3, 5}', intset([2, 4]));
  Test('∁{2, 4}', intset([1, 3, 5]));
  Test('∁{1, 2, 3, 4}', intset([5]));
  Test('∁{1, 2, 3, 5}', intset([4]));
  Test('∁{1, 2, 3, 4, 5}', intset([]));
  Test('∁{1, 2, 3, 4, 5, 6, 7, 8, 9, 10}', intset([]));
  Test('∁{"rabbit", "cat", "horse", 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, "dog"}', intset([]));

  Test('universe ≔ {1, 2, 3, 4, 5, "rabbit", "guinea pig"}; ∁{"rabbit", "cat", "horse", 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, "dog"}',
    asoset(['guinea pig']));

  Test('∁ 5', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('∁5', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('∁(5)', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');

  Test('∁ "rat"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');
  Test('∁"dog"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');
  Test('∁("cat")', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  Test('delete(universe); ∁{1, 2, 3}', failure, 'Unknown identifier "universe".');

  // Element of (set membership, the "in" operator)

  Test('0 ∈ {1, 2, 3}', False);
  Test('1 ∈ {1, 2, 3}', True);
  Test('2 ∈ {1, 2, 3}', True);
  Test('3 ∈ {1, 2, 3}', True);
  Test('4 ∈ {1, 2, 3}', False);
  Test('5 ∈ {1, 2, 3}', False);

  Test('0 ∈ {1, 2, 3, 3, 2, 1, 2, 3}', False);
  Test('1 ∈ {1, 1, 2, 3, 2, 1, 2, 3}', True);
  Test('2 ∈ {3, 2, 1, 2, 3, 1, 2, 3}', True);
  Test('3 ∈ {2, 3, 1, 2, 3}', True);
  Test('4 ∈ {1, 1, 1, 1, 1, 2, 3}', False);
  Test('5 ∈ {3, 3, 3, 3, 3, 3, 3, 2, 1, 1, 2, 3}', False);

  Test('"cat" ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('"dog" ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('"rat" ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('"rabbit" ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('"guinea pig" ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('"horse" ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('"cow" ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('"tarantula" ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('i ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('true ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('color("red") ∈ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);

  Test('6 ∈ {}', False);
  Test('"dog" ∈ {}', False);
  Test('i ∈ {}', False);

  Test('42 ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('i ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('color("red") ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('"rat" ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('IdentityMatrix(10) ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);

  Test('123 ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('1 ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('color("green") ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('"dog" ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('IdentityMatrix(4) ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('false ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('BinaryData(123) ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('ZeroVector(3) ∈ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);

  Test('❨1, 0, 0❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);
  Test('❨0, 1, 0❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);
  Test('❨0, 0, 1❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);

  Test('❨1, 1, 0❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);
  Test('❨0, 1, 1❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);
  Test('❨1, 1, 1❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);
  Test('❨1, 2, 3❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);

  Test('10.5 ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);
  Test('"guinea pig" ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);

  Test('"guinea pig" ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', True);
  Test('❨1, 0, 0❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', True);
  Test('❨0, 1, 0❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', True);
  Test('❨0, 0, 1❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', True);
  Test('❨1, 1, 0❩ ∈ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', False);

  Test('10 ∈ {10, 20, 30}', True);
  Test('10 ∈ {10.0, 20.0, 30.0}', True);
  Test('10.0 ∈ {10, 20, 30}', True);
  Test('10 ∈ {10+i-i, 20, 30}', True);
  Test('10+i-i ∈ {10, 20, 30}', True);

  Test('100 ∈ {10, 20, 30}', False);
  Test('100 ∈ {10.0, 20.0, 30.0}', False);
  Test('100.0 ∈ {10, 20, 30}', False);
  Test('100 ∈ {10+i-i, 20, 30}', False);
  Test('100+i-i ∈ {10, 20, 30}', False);

  Test('0 ∈ {0}', True);
  Test('0.0 ∈ {0}', True);
  Test('1/2 - 1/2 ∈ {0}', True);
  Test('i - i ∈ {0}', True);

  Test('0 ∈ {0.0}', True);
  Test('0.0 ∈ {0.0}', True);
  Test('1/2 - 1/2 ∈ {0.0}', True);
  Test('i - i ∈ {0.0}', True);

  Test('0 ∈ {1/2 - 1/2}', True);
  Test('0.0 ∈ {1/2 - 1/2}', True);
  Test('1/2 - 1/2 ∈ {1/2 - 1/2}', True);
  Test('i - i ∈ {1/2 - 1/2}', True);

  Test('0 ∈ {i-i}', True);
  Test('0.0 ∈ {i-i}', True);
  Test('1/2 - 1/2 ∈ {i-i}', True);
  Test('i - i ∈ {i-i}', True);

  Test('1/2 + 1/2 ∈ {1, 5}', True);

  Test('5 ∈ 125.5', failure, 'An object of type set was expected as argument 2, but an object of type real number was given.');
  Test('"rat" ∈ "rational"', failure, 'An object of type set was expected as argument 2, but an object of type string was given.');

  // Not an element of (set membership, the "not in" operator)

  Test('0 ∉ {1, 2, 3}', True);
  Test('1 ∉ {1, 2, 3}', False);
  Test('2 ∉ {1, 2, 3}', False);
  Test('3 ∉ {1, 2, 3}', False);
  Test('4 ∉ {1, 2, 3}', True);
  Test('5 ∉ {1, 2, 3}', True);

  Test('0 ∉ {1, 2, 3, 3, 2, 1, 2, 3}', True);
  Test('1 ∉ {1, 1, 2, 3, 2, 1, 2, 3}', False);
  Test('2 ∉ {3, 2, 1, 2, 3, 1, 2, 3}', False);
  Test('3 ∉ {2, 3, 1, 2, 3}', False);
  Test('4 ∉ {1, 1, 1, 1, 1, 2, 3}', True);
  Test('5 ∉ {3, 3, 3, 3, 3, 3, 3, 2, 1, 1, 2, 3}', True);

  Test('"cat" ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('"dog" ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('"rat" ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('"rabbit" ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('"guinea pig" ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', False);
  Test('"horse" ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('"cow" ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('"tarantula" ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('i ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('true ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);
  Test('color("red") ∉ {"dog", "cat", "rat", "rabbit", "guinea pig"}', True);

  Test('6 ∉ {}', True);
  Test('"dog" ∉ {}', True);
  Test('i ∉ {}', True);

  Test('42 ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('i ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('color("red") ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('"rat" ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);
  Test('IdentityMatrix(10) ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', False);

  Test('123 ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('1 ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('color("green") ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('"dog" ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('IdentityMatrix(4) ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('false ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('BinaryData(123) ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);
  Test('ZeroVector(3) ∉ {42, i, color("red"), "rat", IdentityMatrix(10)}', True);

  Test('❨1, 0, 0❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);
  Test('❨0, 1, 0❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);
  Test('❨0, 0, 1❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);

  Test('❨1, 1, 0❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);
  Test('❨0, 1, 1❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);
  Test('❨1, 1, 1❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);
  Test('❨1, 2, 3❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);

  Test('10.5 ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);
  Test('"guinea pig" ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);

  Test('"guinea pig" ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', False);
  Test('❨1, 0, 0❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', False);
  Test('❨0, 1, 0❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', False);
  Test('❨0, 0, 1❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', False);
  Test('❨1, 1, 0❩ ∉ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"}', True);

  Test('10 ∉ {10, 20, 30}', False);
  Test('10 ∉ {10.0, 20.0, 30.0}', False);
  Test('10.0 ∉ {10, 20, 30}', False);
  Test('10 ∉ {10+i-i, 20, 30}', False);
  Test('10+i-i ∉ {10, 20, 30}', False);

  Test('100 ∉ {10, 20, 30}', True);
  Test('100 ∉ {10.0, 20.0, 30.0}', True);
  Test('100.0 ∉ {10, 20, 30}', True);
  Test('100 ∉ {10+i-i, 20, 30}', True);
  Test('100+i-i ∉ {10, 20, 30}', True);

  Test('0 ∉ {0}', False);
  Test('0.0 ∉ {0}', False);
  Test('1/2 - 1/2 ∉ {0}', False);
  Test('i - i ∉ {0}', False);

  Test('0 ∉ {0.0}', False);
  Test('0.0 ∉ {0.0}', False);
  Test('1/2 - 1/2 ∉ {0.0}', False);
  Test('i - i ∉ {0.0}', False);

  Test('0 ∉ {1/2 - 1/2}', False);
  Test('0.0 ∉ {1/2 - 1/2}', False);
  Test('1/2 - 1/2 ∉ {1/2 - 1/2}', False);
  Test('i - i ∉ {1/2 - 1/2}', False);

  Test('0 ∉ {i-i}', False);
  Test('0.0 ∉ {i-i}', False);
  Test('1/2 - 1/2 ∉ {i-i}', False);
  Test('i - i ∉ {i-i}', False);

  Test('1/2 + 1/2 ∉ {1, 5}', False);

  Test('5 ∉ 125.5', failure, 'An object of type set was expected as argument 2, but an object of type real number was given.');
  Test('"rat" ∉ "rational"', failure, 'An object of type set was expected as argument 2, but an object of type string was given.');

  // Contains as member (set membership)

  Test('{1, 2, 3} ∋ 0', False);
  Test('{1, 2, 3} ∋ 1', True);
  Test('{1, 2, 3} ∋ 2', True);
  Test('{1, 2, 3} ∋ 3', True);
  Test('{1, 2, 3} ∋ 4', False);
  Test('{1, 2, 3} ∋ 5', False);

  Test('{1, 2, 3, 3, 2, 1, 2, 3} ∋ 0', False);
  Test('{1, 1, 2, 3, 2, 1, 2, 3} ∋ 1', True);
  Test('{3, 2, 1, 2, 3, 1, 2, 3} ∋ 2', True);
  Test('{2, 3, 1, 2, 3} ∋ 3', True);
  Test('{1, 1, 1, 1, 1, 2, 3} ∋ 4', False);
  Test('{3, 3, 3, 3, 3, 3, 3, 2, 1, 1, 2, 3} ∋ 5', False);

  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ "cat"', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ "dog"', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ "rat"', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ "rabbit"', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ "guinea pig"', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ "horse"', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ "cow"', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ "tarantula"', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ i', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ true', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∋ color("red")', False);

  Test('{} ∋ 6', False);
  Test('{} ∋ "dog"', False);
  Test('{} ∋ i', False);

  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ 42', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ i', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ color("red")', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ "rat"', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ IdentityMatrix(10)', True);

  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ 123', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ 1', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ color("green")', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ "dog"', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ IdentityMatrix(4)', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ false', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ BinaryData(123)', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∋ ZeroVector(3)', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∋ ❨1, 0, 0❩', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∋ ❨0, 1, 0❩', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∋ ❨0, 0, 1❩', True);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∋ ❨1, 1, 0❩', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∋ ❨0, 1, 1❩', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∋ ❨1, 1, 1❩', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∋ ❨1, 2, 3❩', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∋ 10.5', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∋ "guinea pig"', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∋ "guinea pig"', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∋ ❨1, 0, 0❩', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∋ ❨0, 1, 0❩', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∋ ❨0, 0, 1❩', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∋ ❨1, 1, 0❩', False);

  Test('{1, 5} ∋ 1/2 + 1/2', True);

  Test('125.5 ∋ 5', failure, 'An object of type set was expected as argument 1, but an object of type real number was given.');
  Test('"rational" ∋ "rat"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  // Does not contain as member (set membership)

  Test('{1, 2, 3} ∌ 0', True);
  Test('{1, 2, 3} ∌ 1', False);
  Test('{1, 2, 3} ∌ 2', False);
  Test('{1, 2, 3} ∌ 3', False);
  Test('{1, 2, 3} ∌ 4', True);
  Test('{1, 2, 3} ∌ 5', True);

  Test('{1, 2, 3, 3, 2, 1, 2, 3} ∌ 0', True);
  Test('{1, 1, 2, 3, 2, 1, 2, 3} ∌ 1', False);
  Test('{3, 2, 1, 2, 3, 1, 2, 3} ∌ 2', False);
  Test('{2, 3, 1, 2, 3} ∌ 3', False);
  Test('{1, 1, 1, 1, 1, 2, 3} ∌ 4', True);
  Test('{3, 3, 3, 3, 3, 3, 3, 2, 1, 1, 2, 3} ∌ 5', True);

  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ "cat"', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ "dog"', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ "rat"', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ "rabbit"', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ "guinea pig"', False);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ "horse"', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ "cow"', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ "tarantula"', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ i', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ true', True);
  Test('{"dog", "cat", "rat", "rabbit", "guinea pig"} ∌ color("red")', True);

  Test('{} ∌ 6', True);
  Test('{} ∌ "dog"', True);
  Test('{} ∌ i', True);

  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ 42', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ i', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ color("red")', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ "rat"', False);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ IdentityMatrix(10)', False);

  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ 123', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ 1', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ color("green")', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ "dog"', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ IdentityMatrix(4)', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ false', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ BinaryData(123)', True);
  Test('{42, i, color("red"), "rat", IdentityMatrix(10)} ∌ ZeroVector(3)', True);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∌ ❨1, 0, 0❩', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∌ ❨0, 1, 0❩', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∌ ❨0, 0, 1❩', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∌ ❨1, 1, 0❩', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∌ ❨0, 1, 1❩', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∌ ❨1, 1, 1❩', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∌ ❨1, 2, 3❩', True);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∌ 10.5', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ∌ "guinea pig"', True);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∌ "guinea pig"', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∌ ❨1, 0, 0❩', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∌ ❨0, 1, 0❩', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∌ ❨0, 0, 1❩', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, "guinea pig"} ∌ ❨1, 1, 0❩', True);

  Test('{1, 5} ∌ 1/2 + 1/2', False);

  Test('125.5 ∌ 5', failure, 'An object of type set was expected as argument 1, but an object of type real number was given.');
  Test('"rational" ∌ "rat"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  // Subset
  // ⊂: U+2282: SUBSET OF

  Test('{1, 2, 3} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{1, 2, 3, 4, 5} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{1, 2, 3, 4, 5, 6, 7, 8} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{1, 2, 3, 4, 5, 6, 7, 8, 9} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, 3, 7, 8, 9} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);

  Test('{-1, 0, 1, 2, 3} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, 3, 4, 5, 6, 7, 8, 9, 10} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, 30} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, "cat"} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, "cat"} ⊂ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, "cat"}', True);

  Test('{"dog", "cat"} ⊂ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat"} ⊂ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat", "rabbit"} ⊂ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "cat", "horse"} ⊂ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);
  Test('{"dog", "cat", "horse", "guinea pig"} ⊂ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊂ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig", "cow", "horse", "sheep"} ⊂ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩} ⊂ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);
  Test('{❨1, 1, 0❩, ❨0, 1, 0❩} ⊂ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊂ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);

  Test('{"dog", "cat", "dog"} ⊂ {"rat", "dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "cat", "dog", "tarantula"} ⊂ {"rat", "dog", "rat", "cat", "rabbit", "guinea pig"}', False);

  Test('{} ⊂ {1, 2, 3}', True);
  Test('{} ⊂ {}', True);
  Test('{0} ⊂ {}', False);
  Test('{0} ⊂ {0}', True);

  Test('{4, "rat", true, IdentityMatrix(10)} ⊂ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', True);
  Test('{4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")} ⊂ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', True);
  Test('{4, "rat", true, IdentityMatrix(10), 25.5, color("red"), 123} ⊂ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', False);
  Test('{"rat", true, IdentityMatrix(10), 25.5, color("blue")} ⊂ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', False);
  Test('{25.5, false, true, 4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")} ⊂ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', True);

  Test('2 ⊂ {1, 2, 3}', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('2 ⊂ 123', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('{2} ⊂ 123', failure, 'An object of type set was expected as argument 2, but an object of type integer was given.');
  Test('"rat" ⊂ "rational"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  // Subset (alternative operator symbol)
  // ⊆: U+2286: SUBSET OF OR EQUAL TO

  Test('{1, 2, 3} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{1, 2, 3, 4, 5} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{1, 2, 3, 4, 5, 6, 7, 8} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{1, 2, 3, 4, 5, 6, 7, 8, 9} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, 3, 7, 8, 9} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);

  Test('{-1, 0, 1, 2, 3} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, 3, 4, 5, 6, 7, 8, 9, 10} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, 30} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, "cat"} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, "cat"} ⊆ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, "cat"}', True);

  Test('{"dog", "cat"} ⊆ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat"} ⊆ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat", "rabbit"} ⊆ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "cat", "horse"} ⊆ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);
  Test('{"dog", "cat", "horse", "guinea pig"} ⊆ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊆ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig", "cow", "horse", "sheep"} ⊆ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩} ⊆ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);
  Test('{❨1, 1, 0❩, ❨0, 1, 0❩} ⊆ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊆ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);

  Test('{"dog", "cat", "dog"} ⊆ {"rat", "dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "cat", "dog", "tarantula"} ⊆ {"rat", "dog", "rat", "cat", "rabbit", "guinea pig"}', False);

  Test('{} ⊆ {1, 2, 3}', True);
  Test('{} ⊆ {}', True);
  Test('{0} ⊆ {}', False);
  Test('{0} ⊆ {0}', True);

  Test('{4, "rat", true, IdentityMatrix(10)} ⊆ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', True);
  Test('{4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")} ⊆ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', True);
  Test('{4, "rat", true, IdentityMatrix(10), 25.5, color("red"), 123} ⊆ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', False);
  Test('{"rat", true, IdentityMatrix(10), 25.5, color("blue")} ⊆ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', False);
  Test('{25.5, false, true, 4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")} ⊆ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', True);

  Test('2 ⊆ {1, 2, 3}', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('2 ⊆ 123', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('{2} ⊆ 123', failure, 'An object of type set was expected as argument 2, but an object of type integer was given.');
  Test('"rat" ⊆ "rational"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  // Proper subset
  // ⊊: U+228A: SUBSET OF WITH NOT EQUAL TO

  Test('{1, 2, 3} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{1, 2, 3, 4, 5} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{1, 2, 3, 4, 5, 6, 7, 8} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{1, 2, 3, 4, 5, 6, 7, 8, 9} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, 3, 7, 8, 9} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);

  Test('{-1, 0, 1, 2, 3} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, 3, 4, 5, 6, 7, 8, 9, 10} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, 30} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, "cat"} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{1, 2, "cat"} ⊊ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, "cat"}', True);

  Test('{"dog", "cat"} ⊊ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat"} ⊊ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat", "rabbit"} ⊊ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "cat", "horse"} ⊊ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);
  Test('{"dog", "cat", "horse", "guinea pig"} ⊊ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊊ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig", "cow", "horse", "sheep"} ⊊ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩} ⊊ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);
  Test('{❨1, 1, 0❩, ❨0, 1, 0❩} ⊊ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊊ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);

  Test('{"dog", "cat", "dog"} ⊊ {"rat", "dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "cat", "dog", "tarantula"} ⊊ {"rat", "dog", "rat", "cat", "rabbit", "guinea pig"}', False);

  Test('{} ⊊ {1, 2, 3}', True);
  Test('{} ⊊ {}', False);
  Test('{0} ⊊ {}', False);
  Test('{0} ⊊ {0}', False);

  Test('{4, "rat", true, IdentityMatrix(10)} ⊊ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', True);
  Test('{4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")} ⊊ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', True);
  Test('{4, "rat", true, IdentityMatrix(10), 25.5, color("red"), 123} ⊊ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', False);
  Test('{"rat", true, IdentityMatrix(10), 25.5, color("blue")} ⊊ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', False);
  Test('{25.5, false, true, 4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")} ⊊ {1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")}', True);

  Test('2 ⊊ {1, 2, 3}', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('2 ⊊ 123', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('{2} ⊊ 123', failure, 'An object of type set was expected as argument 2, but an object of type integer was given.');
  Test('"rat" ⊊ "rational"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  // Superset
  // ⊃: U+2283: SUPERSET OF

  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {1, 2, 3}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {0, 1, 2, 3}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {1, 2, 3, 4, 5}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {1, 2, 3, 4, 5, 6, 7, 8}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {1, 2, 3, 7, 8, 9}', True);

  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {-1, 0, 1, 2, 3}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {1, 2, 30}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊃ {1, 2, "cat"}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, "cat"} ⊃ {1, 2, "cat"}', True);

  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊃ {"dog", "cat"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊃ {"dog", "rat"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊃ {"dog", "rat", "rabbit"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊃ {"dog", "cat", "horse"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊃ {"dog", "cat", "horse", "guinea pig"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊃ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊃ {"dog", "rat", "cat", "rabbit", "guinea pig", "cow", "horse", "sheep"}', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊃ {❨1, 0, 0❩, ❨0, 1, 0❩}', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊃ {❨1, 1, 0❩, ❨0, 1, 0❩}', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊃ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);

  Test('{"rat", "dog", "rat", "cat", "rabbit", "guinea pig"} ⊃ {"dog", "cat", "dog"}', True);
  Test('{"rat", "dog", "rat", "cat", "rabbit", "guinea pig"} ⊃ {"dog", "cat", "dog", "tarantula"}', False);

  Test('{1, 2, 3} ⊃ {}', True);
  Test('{} ⊃ {}', True);
  Test('{} ⊃ {0}', False);
  Test('{0} ⊃ {0}', True);

  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊃ {4, "rat", true, IdentityMatrix(10)}', True);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊃ {4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")}', True);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊃ {4, "rat", true, IdentityMatrix(10), 25.5, color("red"), 123}', False);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊃ {"rat", true, IdentityMatrix(10), 25.5, color("blue")}', False);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊃ {25.5, false, true, 4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")}', True);

  Test('{1, 2, 3} ⊃ 2', failure, 'An object of type set was expected as argument 2, but an object of type integer was given.');
  Test('123 ⊃ 2', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('123 ⊃ {2}', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('"rational" ⊃ "rat"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  // Superset (alternative operator symbol)
  // ⊇: U+2287: SUPERSET OF OR EQUAL TO

  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {1, 2, 3}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {0, 1, 2, 3}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {1, 2, 3, 4, 5}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {1, 2, 3, 4, 5, 6, 7, 8}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {1, 2, 3, 7, 8, 9}', True);

  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {-1, 0, 1, 2, 3}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {1, 2, 30}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊇ {1, 2, "cat"}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, "cat"} ⊇ {1, 2, "cat"}', True);

  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊇ {"dog", "cat"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊇ {"dog", "rat"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊇ {"dog", "rat", "rabbit"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊇ {"dog", "cat", "horse"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊇ {"dog", "cat", "horse", "guinea pig"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊇ {"dog", "rat", "cat", "rabbit", "guinea pig"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊇ {"dog", "rat", "cat", "rabbit", "guinea pig", "cow", "horse", "sheep"}', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊇ {❨1, 0, 0❩, ❨0, 1, 0❩}', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊇ {❨1, 1, 0❩, ❨0, 1, 0❩}', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊇ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', True);

  Test('{"rat", "dog", "rat", "cat", "rabbit", "guinea pig"} ⊇ {"dog", "cat", "dog"}', True);
  Test('{"rat", "dog", "rat", "cat", "rabbit", "guinea pig"} ⊇ {"dog", "cat", "dog", "tarantula"}', False);

  Test('{1, 2, 3} ⊇ {}', True);
  Test('{} ⊇ {}', True);
  Test('{} ⊇ {0}', False);
  Test('{0} ⊇ {0}', True);

  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊇ {4, "rat", true, IdentityMatrix(10)}', True);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊇ {4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")}', True);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊇ {4, "rat", true, IdentityMatrix(10), 25.5, color("red"), 123}', False);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊇ {"rat", true, IdentityMatrix(10), 25.5, color("blue")}', False);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊇ {25.5, false, true, 4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")}', True);

  Test('{1, 2, 3} ⊇ 2', failure, 'An object of type set was expected as argument 2, but an object of type integer was given.');
  Test('123 ⊇ 2', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('123 ⊇ {2}', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('"rational" ⊇ "rat"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  // Proper superset
  // ⊋: U+228B: SUPERSET OF WITH NOT EQUAL TO

  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {1, 2, 3}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {0, 1, 2, 3}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {1, 2, 3, 4, 5}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {1, 2, 3, 4, 5, 6, 7, 8}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {1, 2, 3, 4, 5, 6, 7, 8, 9}', True);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {1, 2, 3, 7, 8, 9}', True);

  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {-1, 0, 1, 2, 3}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {1, 2, 30}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9} ⊋ {1, 2, "cat"}', False);
  Test('{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, "cat"} ⊋ {1, 2, "cat"}', True);

  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊋ {"dog", "cat"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊋ {"dog", "rat"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊋ {"dog", "rat", "rabbit"}', True);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊋ {"dog", "cat", "horse"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊋ {"dog", "cat", "horse", "guinea pig"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊋ {"dog", "rat", "cat", "rabbit", "guinea pig"}', False);
  Test('{"dog", "rat", "cat", "rabbit", "guinea pig"} ⊋ {"dog", "rat", "cat", "rabbit", "guinea pig", "cow", "horse", "sheep"}', False);

  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊋ {❨1, 0, 0❩, ❨0, 1, 0❩}', True);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊋ {❨1, 1, 0❩, ❨0, 1, 0❩}', False);
  Test('{❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩} ⊋ {❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩}', False);

  Test('{"rat", "dog", "rat", "cat", "rabbit", "guinea pig"} ⊋ {"dog", "cat", "dog"}', True);
  Test('{"rat", "dog", "rat", "cat", "rabbit", "guinea pig"} ⊋ {"dog", "cat", "dog", "tarantula"}', False);

  Test('{1, 2, 3} ⊋ {}', True);
  Test('{} ⊋ {}', False);
  Test('{} ⊋ {0}', False);
  Test('{0} ⊋ {0}', False);

  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊋ {4, "rat", true, IdentityMatrix(10)}', True);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊋ {4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")}', True);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊋ {4, "rat", true, IdentityMatrix(10), 25.5, color("red"), 123}', False);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊋ {"rat", true, IdentityMatrix(10), 25.5, color("blue")}', False);
  Test('{1, 4, 25.5, "rat", "cow", true, false, IdentityMatrix(10), IdentityMatrix(20), color("red")} ⊋ {25.5, false, true, 4.0, "rat", true, IdentityMatrix(10), 25.5, color("red")}', True);

  Test('{1, 2, 3} ⊋ 2', failure, 'An object of type set was expected as argument 2, but an object of type integer was given.');
  Test('123 ⊋ 2', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('123 ⊋ {2}', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.');
  Test('"rational" ⊋ "rat"', failure, 'An object of type set was expected as argument 1, but an object of type string was given.');

  // Miscellaneous set expressions (and other tests)

  Test(
    'numbers ≔ ToSet(SequenceList(100));' +
    'EvenNumbers ≔ filter(numbers, even);' +
    'OddNumbers ≔ filter(numbers, odd);' +
    'primes ≔ filter(numbers, IsPrime);' +
    'CompositeNumbers ≔ filter(numbers, n ↦ #PrimeFactors(n) > 1);' +
    'squares ≔ filter(numbers, n ↦ IsInteger(√n));' +
    'MultiplesOf5 ≔ filter(numbers, n ↦ 5 ∣ n);' +
    'MultiplesOf10 ≔ filter(numbers, n ↦ 10 ∣ n);' +
    'universe ≔ numbers;',
    null);

  Test('numbers',
    intset([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100]));

  Test('join(sort(ToList(numbers)))',
    '1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, '+'51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100');

  Test('EvenNumbers',
    intset([2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32, 34, 36, 38, 40, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 62, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 84, 86, 88, 90, 92, 94, 96, 98, 100]));

  Test('join(sort(ToList(EvenNumbers)))',
    '2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32, 34, 36, 38, 40, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 62, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 84, 86, 88, 90, 92, 94, 96, 98, 100');

  Test('OddNumbers',
    intset([1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31, 33, 35, 37, 39, 41, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 63, 65, 67, 69, 71, 73, 75, 77, 79, 81, 83, 85, 87, 89, 91, 93, 95, 97, 99]));

  Test('join(sort(ToList(OddNumbers)))',
    '1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31, 33, 35, 37, 39, 41, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 63, 65, 67, 69, 71, 73, 75, 77, 79, 81, 83, 85, 87, 89, 91, 93, 95, 97, 99');

  Test('primes',
    intset([2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]));

  Test('join(sort(ToList(primes)))',
    '2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97');

  Test('CompositeNumbers',
    intset([4, 6, 8, 9, 10, 12, 14, 15, 16, 18, 20, 21, 22, 24, 25, 26, 27, 28, 30, 32, 33, 34, 35, 36, 38, 39, 40, 42, 44, 45, 46, 48, 49, 50, 51, 52, 54, 55, 56, 57, 58, 60, 62, 63, 64, 65, 66, 68, 69, 70, 72, 74, 75, 76, 77, 78, 80, 81, 82, 84, 85, 86, 87, 88, 90, 91, 92, 93, 94, 95, 96, 98, 99, 100]));

  Test('join(sort(ToList(CompositeNumbers)))',
    '4, 6, 8, 9, 10, 12, 14, 15, 16, 18, 20, 21, 22, 24, 25, 26, 27, 28, 30, 32, 33, 34, 35, 36, 38, 39, 40, 42, 44, 45, 46, 48, 49, '+'50, 51, 52, 54, 55, 56, 57, 58, 60, 62, 63, 64, 65, 66, 68, 69, 70, 72, 74, 75, 76, 77, 78, 80, 81, 82, 84, 85, 86, 87, 88, 90, 91, 92, 93, 94, 95, 96, 98, 99, 100');

  Test('squares',
    intset([1, 4, 9, 16, 25, 36, 49, 64, 81, 100]));

  Test('join(sort(ToList(squares)))',
    '1, 4, 9, 16, 25, 36, 49, 64, 81, 100');

  Test('MultiplesOf5',
    intset([5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100]));

  Test('join(sort(ToList(MultiplesOf5)))',
    '5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100');

  Test('MultiplesOf10',
    intset([10, 20, 30, 40, 50, 60, 70, 80, 90, 100]));

  Test('join(sort(ToList(MultiplesOf10)))',
    '10, 20, 30, 40, 50, 60, 70, 80, 90, 100');

  Test('#numbers', 100);
  Test('#EvenNumbers', 50);
  Test('#OddNumbers', 50);
  Test('#primes = PrimePi(100)', True);

  Test('EvenNumbers ⊂ numbers', True);
  Test('OddNumbers ⊂ numbers', True);
  Test('primes ⊂ numbers', True);
  Test('CompositeNumbers ⊂ numbers', True);
  Test('MultiplesOf5 ⊂ numbers', True);
  Test('MultiplesOf10 ⊂ numbers', True);

  Test('EvenNumbers ⊊ numbers', True);
  Test('OddNumbers ⊊ numbers', True);
  Test('primes ⊊ numbers', True);
  Test('CompositeNumbers ⊊ numbers', True);
  Test('MultiplesOf5 ⊊ numbers', True);
  Test('MultiplesOf10 ⊊ numbers', True);

  Test('∀(numbers, IsNumber)', True);
  Test('∀(numbers, even)', False);
  Test('∀(numbers, odd)', False);
  Test('∀(numbers, IsPrime)', False);

  Test('∃(numbers, IsNumber)', True);
  Test('∃(numbers, even)', True);
  Test('∃(numbers, odd)', True);
  Test('∃(numbers, IsPrime)', True);

  Test('∀(EvenNumbers, even)', True);
  Test('∀(OddNumbers, odd)', True);
  Test('∀(primes, IsPrime)', True);

  Test('count(EvenNumbers, even) = #EvenNumbers', True);
  Test('count(OddNumbers, odd) = #OddNumbers', True);
  Test('count(primes, IsPrime) = #primes', True);

  Test('count(numbers, even) = #EvenNumbers', True);
  Test('count(numbers, odd) = #OddNumbers', True);
  Test('count(numbers, IsPrime) = #primes', True);

  Test('∁numbers', intset([]));
  Test('∁numbers = ∅', True);
  Test('#∁numbers', 0);
  Test('∁EvenNumbers = OddNumbers', True);
  Test('∁OddNumbers = EvenNumbers', True);

  Test('∁EvenNumbers = numbers ∖ EvenNumbers', True);
  Test('∁OddNumbers = numbers ∖ OddNumbers', True);
  Test('∁squares = numbers ∖ squares', True);

  Test('∁∁numbers = numbers', True);
  Test('∁∁EvenNumbers = EvenNumbers', True);
  Test('∁∁OddNumbers = OddNumbers', True);

  Test('numbers ∖ EvenNumbers = OddNumbers', True);
  Test('numbers ∖ OddNumbers = EvenNumbers', True);

  Test('#EvenNumbers + #OddNumbers = #numbers', True);
  Test('EvenNumbers ∪ OddNumbers = numbers', True);
  Test('#(EvenNumbers ∪ OddNumbers) = #numbers', True);

  Test('EvenNumbers ∩ OddNumbers', intset([]));
  Test('#(EvenNumbers ∩ OddNumbers)', 0);
  Test('∁(EvenNumbers ∩ OddNumbers) = numbers', True);

  Test('primes ∩ EvenNumbers', intset([2]));
  Test('#(primes ∩ EvenNumbers)', 1);
  Test('1 ∈ primes ∩ EvenNumbers', False);
  Test('2 ∈ primes ∩ EvenNumbers', True);
  Test('3 ∈ primes ∩ EvenNumbers', False);
  Test('4 ∈ primes ∩ EvenNumbers', False);
  Test('∃(numbers, n ↦ IsPrime(n) ∧ even(n))', True);
  Test('count(numbers, n ↦ IsPrime(n) ∧ even(n))', 1);

  Test('∀(primes, odd)', False);
  Test('∀(primes ∖ {2}, odd)', True);
  Test('∀(primes ∖ EvenNumbers, odd)', True);

  Test('primes ∖ EvenNumbers = primes ∖ {2}', True);

  Test('1 ∈ primes ∩ OddNumbers', False);
  Test('2 ∈ primes ∩ OddNumbers', False);
  Test('3 ∈ primes ∩ OddNumbers', True);
  Test('4 ∈ primes ∩ OddNumbers', False);
  Test('5 ∈ primes ∩ OddNumbers', True);

  Test('8 ∉ primes', True);
  Test('9 ∉ primes', True);
  Test('10 ∉ primes', True);
  Test('11 ∉ primes', False);
  Test('12 ∉ primes', True);

  Test('squares ⊊ numbers', True);
  Test('25 ∈ squares', True);
  Test('{25, 36, 49} ⊂ squares', True);
  Test('35 ∈ squares', False);
  Test('{35, 36, 49} ⊂ squares', False);
  Test('squares ∩ primes', intset([]));
  Test('squares ∩ primes = ∅', True);
  Test('#(squares ∩ primes)', 0);

  Test('#(squares ∪ MultiplesOf5) = #squares + #MultiplesOf5 − #(squares ∩ MultiplesOf5)', True);
  Test('∁(squares ∪ MultiplesOf5) = ∁squares ∩ ∁MultiplesOf5', True);

  Test('MultiplesOf10 ⊂ MultiplesOf5', True);
  Test('#MultiplesOf10 ≤ #MultiplesOf5', True);
  Test('MultiplesOf10 ∪ MultiplesOf5 = MultiplesOf5', True);

  Test('∁MultiplesOf5 ⊂ ∁MultiplesOf10', True);

  Test('∁primes = CompositeNumbers', False);
  Test('∁primes ≠ CompositeNumbers', True);
  Test('∁primes = CompositeNumbers ∪ {1}', True);
  Test('∁primes ∖ {1} = CompositeNumbers', True);
  Test('∁primes ∖ CompositeNumbers', intset([1]));
  Test('∁primes ∩ ∁CompositeNumbers', intset([1]));
  Test('∁(primes ∪ CompositeNumbers)', intset([1]));
  Test('∁primes ∩ ∁CompositeNumbers = ∁(primes ∪ CompositeNumbers)', True);
  Test('primes ∪ CompositeNumbers = numbers ∖ {1}', True);
  Test('primes ∪ CompositeNumbers ∪ {1} = numbers', True);
  Test('primes ∪ CompositeNumbers ⊊ numbers', True);

  Test('squares ∪ ∁squares = numbers', True);
  Test('#squares + #∁squares = #numbers', True);

  Test('delete(numbers); delete(universe); delete(EvenNumbers); delete(OddNumbers); ' +
       'delete(primes); delete(CompositeNumbers); delete(MultiplesOf5); delete(MultiplesOf10);', null);

  // Divides

  Test('1 ∣ 25', True);
  Test('2 ∣ 25', False);
  Test('3 ∣ 25', False);
  Test('4 ∣ 25', False);
  Test('5 ∣ 25', True);
  Test('6 ∣ 25', False);
  Test('7 ∣ 25', False);
  Test('24 ∣ 25', False);
  Test('25 ∣ 25', True);
  Test('26 ∣ 25', False);
  Test('50 ∣ 25', False);

  Test('1 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('2 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('4 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('3 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('5 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('10 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('6 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('12 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('7 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('23 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('35 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('20 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('60 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('14 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('21 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('42 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('84 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('46 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('92 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('805 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('2415 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('4830 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('9660 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);

  Test('9 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('11 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('53 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('99 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('200 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('123 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('1234 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('12345 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('100 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);

  Test('1 ∣ −25', True);
  Test('−2 ∣ 25', False);
  Test('−3 ∣ 25', False);
  Test('−4 ∣ −25', False);
  Test('−5 ∣ −25', True);
  Test('6 ∣ 25', False);
  Test('−7 ∣ −25', False);
  Test('24 ∣ 25', False);
  Test('−25 ∣ −25', True);
  Test('26 ∣ 25', False);
  Test('50 ∣ −25', False);

  Test('1 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−2 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−4 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('3 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('5 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−10 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−6 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('12 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('7 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−23 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−35 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−20 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−60 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−14 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('21 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−42 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−84 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('46 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−92 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−805 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('2415 ∣ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−4830 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−9660 ∣ 2⋅2⋅3⋅5⋅7⋅23', True);

  Test('−9 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−11 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('53 ∣ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−99 ∣ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−200 ∣ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('123 ∣ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−1234 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−12345 ∣ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−100 ∣ −2⋅2⋅3⋅5⋅7⋅23', False);

  Test('1786524853 ∣ 1786524853⋅239574163', True);
  Test('239574163 ∣ 1786524853⋅239574163', True);
  Test('−1786524853 ∣ 1786524853⋅239574163', True);
  Test('239574163 ∣ −1786524853⋅239574163', True);
  Test('−239574163 ∣ −1786524853⋅239574163', True);

  Test('1786524852 ∣ 1786524853⋅239574163', False);
  Test('1786524854 ∣ 1786524853⋅239574163', False);
  Test('2 ∣ 1786524853⋅239574163', False);
  Test('55 ∣ 1786524853⋅239574163', False);
  Test('17 ∣ 1786524853⋅239574163', False);
  Test('1 ∣ 1786524853⋅239574163', True);
  Test('−1 ∣ 1786524853⋅239574163', True);
  Test('1786524853⋅239574163 ∣ 1786524853⋅239574163', True);
  Test('−1786524853⋅239574163 ∣ 1786524853⋅239574163', True);
  Test('−1786524853⋅239574163 ∣ −1786524853⋅239574163', True);
  Test('1786524853⋅239574163 ∣ −1786524853⋅239574163', True);

  Test('1 ∣ 1', True);
  Test('1 ∣ 123', True);
  Test('1 ∣ 431', True);
  Test('1 ∣ 24143122', True);
  Test('1 ∣ 4324565465', True);
  Test('1 ∣ 87581790', True);
  Test('1 ∣ 165653516', True);

  Test('−1 ∣ 1', True);
  Test('−1 ∣ 123', True);
  Test('−1 ∣ 431', True);
  Test('−1 ∣ 24143122', True);
  Test('−1 ∣ 4324565465', True);
  Test('−1 ∣ 87581790', True);
  Test('−1 ∣ 165653516', True);

  Test('1 ∣ 1', True);
  Test('123 ∣ 123', True);
  Test('431 ∣ 431', True);
  Test('24143122 ∣ 24143122', True);
  Test('4324565465 ∣ 4324565465', True);
  Test('87581790 ∣ 87581790', True);
  Test('165653516 ∣ 165653516', True);

  Test('−1 ∣ 1', True);
  Test('−123 ∣ 123', True);
  Test('−431 ∣ 431', True);
  Test('−24143122 ∣ 24143122', True);
  Test('−4324565465 ∣ 4324565465', True);
  Test('−87581790 ∣ 87581790', True);
  Test('−165653516 ∣ 165653516', True);

  Test('1 ∣ 0', True);
  Test('2 ∣ 0', True);
  Test('3 ∣ 0', True);
  Test('4 ∣ 0', True);
  Test('5 ∣ 0', True);
  Test('10 ∣ 0', True);
  Test('100 ∣ 0', True);
  Test('-1 ∣ 0', True);
  Test('-2 ∣ 0', True);
  Test('-3 ∣ 0', True);
  Test('-4 ∣ 0', True);
  Test('-5 ∣ 0', True);
  Test('-10 ∣ 0', True);
  Test('-100 ∣ 0', True);

  Test('0 ∣ 21', failure, 'Division by zero');
  Test('0 ∣ 0', failure, 'Division by zero');

  Test('5.5 ∣ 11', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('1/2 ∣ 11', failure, 'An object of type integer was expected as argument 1, but an object of type rational number was given.');
  Test('"Brexit" ∣ "UK"', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');
  Test('"impeachment" ∣ "US"', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');
  Test('10 ∣ "one hundred"', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');

  // Does not divide

  Test('1 ∤ 25', False);
  Test('2 ∤ 25', True);
  Test('3 ∤ 25', True);
  Test('4 ∤ 25', True);
  Test('5 ∤ 25', False);
  Test('6 ∤ 25', True);
  Test('7 ∤ 25', True);
  Test('24 ∤ 25', True);
  Test('25 ∤ 25', False);
  Test('26 ∤ 25', True);
  Test('50 ∤ 25', True);

  Test('1 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('2 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('4 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('3 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('5 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('10 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('6 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('12 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('7 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('23 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('35 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('20 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('60 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('14 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('21 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('42 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('84 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('46 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('92 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('805 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('2415 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('4830 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('9660 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);

  Test('9 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('11 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('53 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('99 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('200 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('123 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('1234 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('12345 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('100 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);

  Test('1 ∤ −25', False);
  Test('−2 ∤ 25', True);
  Test('−3 ∤ 25', True);
  Test('−4 ∤ −25', True);
  Test('−5 ∤ −25', False);
  Test('6 ∤ 25', True);
  Test('−7 ∤ −25', True);
  Test('24 ∤ 25', True);
  Test('−25 ∤ −25', False);
  Test('26 ∤ 25', True);
  Test('50 ∤ −25', True);

  Test('1 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−2 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−4 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('3 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('5 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−10 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−6 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('12 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('7 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−23 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−35 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−20 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−60 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−14 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('21 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−42 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−84 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('46 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−92 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−805 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('2415 ∤ −2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−4830 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);
  Test('−9660 ∤ 2⋅2⋅3⋅5⋅7⋅23', False);

  Test('−9 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−11 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('53 ∤ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−99 ∤ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−200 ∤ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('123 ∤ −2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−1234 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−12345 ∤ 2⋅2⋅3⋅5⋅7⋅23', True);
  Test('−100 ∤ −2⋅2⋅3⋅5⋅7⋅23', True);

  Test('1786524853 ∤ 1786524853⋅239574163', False);
  Test('239574163 ∤ 1786524853⋅239574163', False);
  Test('−1786524853 ∤ 1786524853⋅239574163', False);
  Test('239574163 ∤ −1786524853⋅239574163', False);
  Test('−239574163 ∤ −1786524853⋅239574163', False);

  Test('1786524852 ∤ 1786524853⋅239574163', True);
  Test('1786524854 ∤ 1786524853⋅239574163', True);
  Test('2 ∤ 1786524853⋅239574163', True);
  Test('55 ∤ 1786524853⋅239574163', True);
  Test('17 ∤ 1786524853⋅239574163', True);
  Test('1 ∤ 1786524853⋅239574163', False);
  Test('−1 ∤ 1786524853⋅239574163', False);
  Test('1786524853⋅239574163 ∤ 1786524853⋅239574163', False);
  Test('−1786524853⋅239574163 ∤ 1786524853⋅239574163', False);
  Test('−1786524853⋅239574163 ∤ −1786524853⋅239574163', False);
  Test('1786524853⋅239574163 ∤ −1786524853⋅239574163', False);

  Test('1 ∤ 1', False);
  Test('1 ∤ 123', False);
  Test('1 ∤ 431', False);
  Test('1 ∤ 24143122', False);
  Test('1 ∤ 4324565465', False);
  Test('1 ∤ 87581790', False);
  Test('1 ∤ 165653516', False);

  Test('−1 ∤ 1', False);
  Test('−1 ∤ 123', False);
  Test('−1 ∤ 431', False);
  Test('−1 ∤ 24143122', False);
  Test('−1 ∤ 4324565465', False);
  Test('−1 ∤ 87581790', False);
  Test('−1 ∤ 165653516', False);

  Test('1 ∤ 1', False);
  Test('123 ∤ 123', False);
  Test('431 ∤ 431', False);
  Test('24143122 ∤ 24143122', False);
  Test('4324565465 ∤ 4324565465', False);
  Test('87581790 ∤ 87581790', False);
  Test('165653516 ∤ 165653516', False);

  Test('−1 ∤ 1', False);
  Test('−123 ∤ 123', False);
  Test('−431 ∤ 431', False);
  Test('−24143122 ∤ 24143122', False);
  Test('−4324565465 ∤ 4324565465', False);
  Test('−87581790 ∤ 87581790', False);
  Test('−165653516 ∤ 165653516', False);

  Test('1 ∤ 0', False);
  Test('2 ∤ 0', False);
  Test('3 ∤ 0', False);
  Test('4 ∤ 0', False);
  Test('5 ∤ 0', False);
  Test('10 ∤ 0', False);
  Test('100 ∤ 0', False);
  Test('-1 ∤ 0', False);
  Test('-2 ∤ 0', False);
  Test('-3 ∤ 0', False);
  Test('-4 ∤ 0', False);
  Test('-5 ∤ 0', False);
  Test('-10 ∤ 0', False);
  Test('-100 ∤ 0', False);

  Test('0 ∤ 21', failure, 'Division by zero');
  Test('0 ∤ 0', failure, 'Division by zero');

  Test('5.5 ∤ 11', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('1/2 ∤ 11', failure, 'An object of type integer was expected as argument 1, but an object of type rational number was given.');
  Test('"Brexit" ∤ "UK"', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');
  Test('"impeachment" ∤ "US"', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');
  Test('10 ∤ "one hundred"', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');

  // Parallel to

  Test('❨4, 8, 1❩ ∥ ❨4, 8, 1❩', True);
  Test('❨8, 16, 2❩ ∥ ❨4, 8, 1❩', True);
  Test('❨12, 24, 3❩ ∥ ❨4, 8, 1❩', True);
  Test('❨40, 80, 10❩ ∥ ❨4, 8, 1❩', True);
  Test('❨-4, -8, -1❩ ∥ ❨4, 8, 1❩', True);
  Test('❨-400, -800, -100❩ ∥ ❨4, 8, 1❩', True);
  Test('❨0, 0, 0❩ ∥ ❨4, 8, 1❩', True);
  Test('❨4⋅i, 8⋅i, i❩ ∥ ❨4, 8, 1❩', True);
  Test('❨12+4⋅i, 24+8⋅i, 3+i❩ ∥ ❨4, 8, 1❩', True);

  Test('❨4, 8, 0❩ ∥ ❨4, 8, 1❩', False);
  Test('❨4, 7, 1❩ ∥ ❨4, 8, 1❩', False);
  Test('❨8, 8, 2❩ ∥ ❨4, 8, 1❩', False);
  Test('❨4, -8, 1❩ ∥ ❨4, 8, 1❩', False);
  Test('❨12, 24.0000001, 3❩ ∥ ❨4, 8, 1❩', False);
  Test('❨7, 2, 6❩ ∥ ❨4, 8, 1❩', False);
  Test('❨1, 0, 0❩ ∥ ❨4, 8, 1❩', False);
  Test('❨6, 9, 3❩ ∥ ❨4, 8, 1❩', False);
  Test('❨12+4⋅i, 21+8⋅i, 3+i❩ ∥ ❨4, 8, 1❩', False);

  Test('❨1, −1, 0❩ ∥ ❨1, −1, 0❩', True);
  Test('❨−1, 1, 0❩ ∥ ❨1, −1, 0❩', True);
  Test('❨2, −2, 0❩ ∥ ❨1, −1, 0❩', True);
  Test('❨10, −10, 0❩ ∥ ❨1, −1, 0❩', True);
  Test('❨−7, 7, 0❩ ∥ ❨1, −1, 0❩', True);
  Test('❨0.5, −0.5, 0❩ ∥ ❨1, −1, 0❩', True);
  Test('❨1/77, −1/77, 0❩ ∥ ❨1, −1, 0❩', True);
  Test('❨i, −i, 0❩ ∥ ❨1, −1, 0❩', True);
  Test('❨−i, i, 0❩ ∥ ❨1, −1, 0❩', True);
  Test('❨0, 0, 0❩ ∥ ❨1, −1, 0❩', True);

  Test('❨1, 1, 0❩ ∥ ❨1, −1, 0❩', False);
  Test('❨1, −1, 1❩ ∥ ❨1, −1, 0❩', False);
  Test('❨2, −1, 0❩ ∥ ❨1, −1, 0❩', False);
  Test('❨10, −9, 0❩ ∥ ❨1, −1, 0❩', False);
  Test('❨1, −1, 5❩ ∥ ❨1, −1, 0❩', False);
  Test('❨0, 0, 1❩ ∥ ❨1, −1, 0❩', False);
  Test('❨1, 1, 1❩ ∥ ❨1, −1, 0❩', False);
  Test('❨5, 2, 0❩ ∥ ❨1, −1, 0❩', False);
  Test('❨i, −1, 0❩ ∥ ❨1, −1, 0❩', False);

  Test('❨0, 0, 0❩ ∥ ❨1, 2, 3❩', True);
  Test('❨0, 0, 0❩ ∥ ❨1, 0, 0❩', True);
  Test('❨0, 0, 0❩ ∥ ❨7, −2, −3❩', True);
  Test('❨0, 0, 0❩ ∥ ❨1, i, 3❩', True);
  Test('❨0, 0, 0❩ ∥ ❨1, 2, 3 + i❩', True);

  Test('❨0, 0, 0❩ ∥ ❨0, 0, 0❩', True);
  Test('❨1, 0, 0❩ ∥ ❨0, 0, 0❩', True);
  Test('❨1, 2, 3❩ ∥ ❨0, 0, 0❩', True);
  Test('❨1000, i, 5❩ ∥ ❨0, 0, 0❩', True);

  Test('❨0, 0, i-i❩ ∥ ❨1, 2, 3❩', True);
  Test('❨0, i-i, 0❩ ∥ ❨1, 0, 0❩', True);
  Test('❨0, i-i, 0❩ ∥ ❨1, i-i, 0❩', True);
  Test('❨i-i, 0, 0❩ ∥ ❨7, −2, −3❩', True);
  Test('❨0, i-i, 0❩ ∥ ❨1, i, 3❩', True);
  Test('❨0, 0, i-i❩ ∥ ❨1, 2, 3 + i❩', True);

  Test('❨0, 0, i-i❩ ∥ ❨0, 0, 0❩', True);
  Test('❨0, 0, i-i❩ ∥ ❨0, 0, i-i❩', True);
  Test('❨1, 0, 0❩ ∥ ❨i-i, 0, 0❩', True);
  Test('❨1, i-i, 0❩ ∥ ❨i-i, 0, 0❩', True);
  Test('❨1, 2, 3❩ ∥ ❨0, 0, 0❩', True);
  Test('❨1000, i, 5❩ ∥ ❨0, i-i, 0❩', True);

  Test('❨7, 0, 0❩ ∥ ❨1, 0, 0❩', True);
  Test('❨1, 0, 0❩ ∥ ❨1, 0, 0❩', True);
  Test('❨−10, 0, 0❩ ∥ ❨1, 0, 0❩', True);
  Test('❨−i, 0, 0❩ ∥ ❨1, 0, 0❩', True);
  Test('❨7 − 9⋅i, 0, 0❩ ∥ ❨1, 0, 0❩', True);

  Test('❨7, 1, 0❩ ∥ ❨1, 0, 0❩', False);
  Test('❨7, i, 0❩ ∥ ❨1, 0, 0❩', False);
  Test('❨7, 0, 1❩ ∥ ❨1, 0, 0❩', False);
  Test('❨7, 0.00000001, 0❩ ∥ ❨1, 0, 0❩', False);
  Test('❨7, 0, -0.00000001❩ ∥ ❨1, 0, 0❩', False);

  Test('❨1, 2❩ ∥ ❨1, 2❩', True);
  Test('❨2, 4❩ ∥ ❨1, 2❩', True);
  Test('❨4, 8❩ ∥ ❨1, 2❩', True);
  Test('❨1.01, 2.02❩ ∥ ❨1, 2❩', True);
  Test('❨-1, -2❩ ∥ ❨1, 2❩', True);
  Test('❨i, 2⋅i❩ ∥ ❨1, 2❩', True);
  Test('❨7+i, 14+2⋅i❩ ∥ ❨1, 2❩', True);
  Test('❨0, 0❩ ∥ ❨1, 2❩', True);
  Test('❨1, 2❩ ∥ ❨0, 0❩', True);
  Test('❨0, 0❩ ∥ ❨0, 0❩', True);

  Test('❨-1, 2❩ ∥ ❨1, 2❩', False);
  Test('❨1.0001, 2❩ ∥ ❨1, 2❩', False);
  Test('❨5, 2❩ ∥ ❨1, 2❩', False);
  Test('❨7+i, 14+2.000001⋅i❩ ∥ ❨1, 2❩', False);

  Test('❨1, 0, 0❩ ∥ ❨0, 1, 0❩', False);
  Test('❨1, 0, 0❩ ∥ ❨0, 1, i-i❩', False);

  Test('❨2, 3⋅i, 0❩ ∥ ❨2, 3⋅i, 0❩', True);
  Test('❨20, 30⋅i, 0❩ ∥ ❨2, 3⋅i, 0❩', True);
  Test('❨2, 4⋅i, 0❩ ∥ ❨2, 3⋅i, 0❩', False);

  Test('❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩ ∥ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', True);
  Test('❨2, 4, 0, 2, 0, −2, 4, 6, 2, 2⋅π❩ ∥ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', True);
  Test('❨10, 20, 0, 10, 0, −10, 20, 30, 10, 10⋅π❩ ∥ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', True);
  Test('❨−1, −2, 0, −1, 0, 1, −2, −3, −1, −π❩ ∥ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', True);

  Test('❨1, 2, 1, 1, 0, −1, 2, 3, 1, π❩ ∥ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', False);

  Test('❨1.00000000000000000001, 2, 0❩ ∥ ❨1, 2, 0❩', True);
  Test('❨1.0000000000000001, 2, 0❩ ∥ ❨1, 2, 0❩', True);
  Test('❨1.00000000001, 2, 0❩ ∥ ❨1, 2, 0❩', False);

  Test('❨1, 0, 0❩ ∥ ❨0, 2❩', failure, 'Vectors are of different dimension.');
  Test('❨1, 0, 0❩ ∥ ❨i-i, 2❩', failure, 'Vectors are of different dimension.');

  Test('"line" ∥ ❨0, 2❩', failure, 'An object of type vector was expected as argument 1, but an object of type string was given.');
  Test('❨1, 0, 0❩ ∥ 6', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');

  // Not parallel to

  Test('❨4, 8, 1❩ ∦ ❨4, 8, 1❩', False);
  Test('❨8, 16, 2❩ ∦ ❨4, 8, 1❩', False);
  Test('❨12, 24, 3❩ ∦ ❨4, 8, 1❩', False);
  Test('❨40, 80, 10❩ ∦ ❨4, 8, 1❩', False);
  Test('❨-4, -8, -1❩ ∦ ❨4, 8, 1❩', False);
  Test('❨-400, -800, -100❩ ∦ ❨4, 8, 1❩', False);
  Test('❨0, 0, 0❩ ∦ ❨4, 8, 1❩', False);
  Test('❨4⋅i, 8⋅i, i❩ ∦ ❨4, 8, 1❩', False);
  Test('❨12+4⋅i, 24+8⋅i, 3+i❩ ∦ ❨4, 8, 1❩', False);

  Test('❨4, 8, 0❩ ∦ ❨4, 8, 1❩', True);
  Test('❨4, 7, 1❩ ∦ ❨4, 8, 1❩', True);
  Test('❨8, 8, 2❩ ∦ ❨4, 8, 1❩', True);
  Test('❨4, -8, 1❩ ∦ ❨4, 8, 1❩', True);
  Test('❨12, 24.0000001, 3❩ ∦ ❨4, 8, 1❩', True);
  Test('❨7, 2, 6❩ ∦ ❨4, 8, 1❩', True);
  Test('❨1, 0, 0❩ ∦ ❨4, 8, 1❩', True);
  Test('❨6, 9, 3❩ ∦ ❨4, 8, 1❩', True);
  Test('❨12+4⋅i, 21+8⋅i, 3+i❩ ∦ ❨4, 8, 1❩', True);

  Test('❨1, −1, 0❩ ∦ ❨1, −1, 0❩', False);
  Test('❨−1, 1, 0❩ ∦ ❨1, −1, 0❩', False);
  Test('❨2, −2, 0❩ ∦ ❨1, −1, 0❩', False);
  Test('❨10, −10, 0❩ ∦ ❨1, −1, 0❩', False);
  Test('❨−7, 7, 0❩ ∦ ❨1, −1, 0❩', False);
  Test('❨0.5, −0.5, 0❩ ∦ ❨1, −1, 0❩', False);
  Test('❨1/77, −1/77, 0❩ ∦ ❨1, −1, 0❩', False);
  Test('❨i, −i, 0❩ ∦ ❨1, −1, 0❩', False);
  Test('❨−i, i, 0❩ ∦ ❨1, −1, 0❩', False);
  Test('❨0, 0, 0❩ ∦ ❨1, −1, 0❩', False);

  Test('❨1, 1, 0❩ ∦ ❨1, −1, 0❩', True);
  Test('❨1, −1, 1❩ ∦ ❨1, −1, 0❩', True);
  Test('❨2, −1, 0❩ ∦ ❨1, −1, 0❩', True);
  Test('❨10, −9, 0❩ ∦ ❨1, −1, 0❩', True);
  Test('❨1, −1, 5❩ ∦ ❨1, −1, 0❩', True);
  Test('❨0, 0, 1❩ ∦ ❨1, −1, 0❩', True);
  Test('❨1, 1, 1❩ ∦ ❨1, −1, 0❩', True);
  Test('❨5, 2, 0❩ ∦ ❨1, −1, 0❩', True);
  Test('❨i, −1, 0❩ ∦ ❨1, −1, 0❩', True);

  Test('❨0, 0, 0❩ ∦ ❨1, 2, 3❩', False);
  Test('❨0, 0, 0❩ ∦ ❨1, 0, 0❩', False);
  Test('❨0, 0, 0❩ ∦ ❨7, −2, −3❩', False);
  Test('❨0, 0, 0❩ ∦ ❨1, i, 3❩', False);
  Test('❨0, 0, 0❩ ∦ ❨1, 2, 3 + i❩', False);

  Test('❨0, 0, 0❩ ∦ ❨0, 0, 0❩', False);
  Test('❨1, 0, 0❩ ∦ ❨0, 0, 0❩', False);
  Test('❨1, 2, 3❩ ∦ ❨0, 0, 0❩', False);
  Test('❨1000, i, 5❩ ∦ ❨0, 0, 0❩', False);

  Test('❨0, 0, i-i❩ ∦ ❨1, 2, 3❩', False);
  Test('❨0, i-i, 0❩ ∦ ❨1, 0, 0❩', False);
  Test('❨0, i-i, 0❩ ∦ ❨1, i-i, 0❩', False);
  Test('❨i-i, 0, 0❩ ∦ ❨7, −2, −3❩', False);
  Test('❨0, i-i, 0❩ ∦ ❨1, i, 3❩', False);
  Test('❨0, 0, i-i❩ ∦ ❨1, 2, 3 + i❩', False);

  Test('❨0, 0, i-i❩ ∦ ❨0, 0, 0❩', False);
  Test('❨0, 0, i-i❩ ∦ ❨0, 0, i-i❩', False);
  Test('❨1, 0, 0❩ ∦ ❨i-i, 0, 0❩', False);
  Test('❨1, i-i, 0❩ ∦ ❨i-i, 0, 0❩', False);
  Test('❨1, 2, 3❩ ∦ ❨0, 0, 0❩', False);
  Test('❨1000, i, 5❩ ∦ ❨0, i-i, 0❩', False);

  Test('❨7, 0, 0❩ ∦ ❨1, 0, 0❩', False);
  Test('❨1, 0, 0❩ ∦ ❨1, 0, 0❩', False);
  Test('❨−10, 0, 0❩ ∦ ❨1, 0, 0❩', False);
  Test('❨−i, 0, 0❩ ∦ ❨1, 0, 0❩', False);
  Test('❨7 − 9⋅i, 0, 0❩ ∦ ❨1, 0, 0❩', False);

  Test('❨7, 1, 0❩ ∦ ❨1, 0, 0❩', True);
  Test('❨7, i, 0❩ ∦ ❨1, 0, 0❩', True);
  Test('❨7, 0, 1❩ ∦ ❨1, 0, 0❩', True);
  Test('❨7, 0.00000001, 0❩ ∦ ❨1, 0, 0❩', True);
  Test('❨7, 0, -0.00000001❩ ∦ ❨1, 0, 0❩', True);

  Test('❨1, 2❩ ∦ ❨1, 2❩', False);
  Test('❨2, 4❩ ∦ ❨1, 2❩', False);
  Test('❨4, 8❩ ∦ ❨1, 2❩', False);
  Test('❨1.01, 2.02❩ ∦ ❨1, 2❩', False);
  Test('❨-1, -2❩ ∦ ❨1, 2❩', False);
  Test('❨i, 2⋅i❩ ∦ ❨1, 2❩', False);
  Test('❨7+i, 14+2⋅i❩ ∦ ❨1, 2❩', False);
  Test('❨0, 0❩ ∦ ❨1, 2❩', False);
  Test('❨1, 2❩ ∦ ❨0, 0❩', False);
  Test('❨0, 0❩ ∦ ❨0, 0❩', False);

  Test('❨-1, 2❩ ∦ ❨1, 2❩', True);
  Test('❨1.0001, 2❩ ∦ ❨1, 2❩', True);
  Test('❨5, 2❩ ∦ ❨1, 2❩', True);
  Test('❨7+i, 14+2.000001⋅i❩ ∦ ❨1, 2❩', True);

  Test('❨1, 0, 0❩ ∦ ❨0, 1, 0❩', True);
  Test('❨1, 0, 0❩ ∦ ❨0, 1, i-i❩', True);

  Test('❨2, 3⋅i, 0❩ ∦ ❨2, 3⋅i, 0❩', False);
  Test('❨20, 30⋅i, 0❩ ∦ ❨2, 3⋅i, 0❩', False);
  Test('❨2, 4⋅i, 0❩ ∦ ❨2, 3⋅i, 0❩', True);

  Test('❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩ ∦ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', False);
  Test('❨2, 4, 0, 2, 0, −2, 4, 6, 2, 2⋅π❩ ∦ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', False);
  Test('❨10, 20, 0, 10, 0, −10, 20, 30, 10, 10⋅π❩ ∦ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', False);
  Test('❨−1, −2, 0, −1, 0, 1, −2, −3, −1, −π❩ ∦ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', False);

  Test('❨1, 2, 1, 1, 0, −1, 2, 3, 1, π❩ ∦ ❨1, 2, 0, 1, 0, −1, 2, 3, 1, π❩', True);

  Test('❨1.00000000000000000001, 2, 0❩ ∦ ❨1, 2, 0❩', False);
  Test('❨1.0000000000000001, 2, 0❩ ∦ ❨1, 2, 0❩', False);
  Test('❨1.00000000001, 2, 0❩ ∦ ❨1, 2, 0❩', True);

  Test('❨1, 0, 0❩ ∦ ❨0, 2❩', failure, 'Vectors are of different dimension.');
  Test('❨1, 0, 0❩ ∦ ❨i-i, 2❩', failure, 'Vectors are of different dimension.');

  Test('"line" ∦ ❨0, 2❩', failure, 'An object of type vector was expected as argument 1, but an object of type string was given.');
  Test('❨1, 0, 0❩ ∦ 6', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');

  // Perpendicular, coprime

  // - numbers: coprime

  Test('14 ⟂ 15', True);
  Test('100 ⟂ 49', True);
  Test('25 ⟂ 36', True);
  Test('81 ⟂ 1024', True);
  Test('895965 ⟂ 14278', True);

  Test('15 ⟂ 9', False);
  Test('25 ⟂ 10', False);
  Test('81 ⟂ 33', False);
  Test('100 ⟂ 72', False);
  Test('49 ⟂ 28', False);
  Test('895965 ⟂ 21417', False);
  Test('895965 ⟂ 12826', False);

  Test('−14 ⟂ 15', True);
  Test('100 ⟂ −49', True);
  Test('−25 ⟂ −36', True);
  Test('−81 ⟂ 1024', True);
  Test('−895965 ⟂ 14278', True);

  Test('−15 ⟂ 9', False);
  Test('25 ⟂ −10', False);
  Test('−81 ⟂ −33', False);
  Test('−100 ⟂ 72', False);
  Test('49 ⟂ −28', False);
  Test('−895965 ⟂ −21417', False);
  Test('−895965 ⟂ 12826', False);

  Test('1 ⟂ 1', True);
  Test('1 ⟂ 2', True);
  Test('1 ⟂ 5', True);
  Test('1 ⟂ 15', True);
  Test('1 ⟂ 1024', True);
  Test('1 ⟂ 10000000', True);
  Test('1 ⟂ 895965', True);

  Test('−1 ⟂ 1', True);
  Test('1 ⟂ −2', True);
  Test('−1 ⟂ −5', True);
  Test('−1 ⟂ 15', True);
  Test('1 ⟂ −1024', True);
  Test('−1 ⟂ −10000000', True);
  Test('1 ⟂ −895965', True);

  Test('1 ⟂ 0', True);
  Test('0 ⟂ 1', True);
  Test('(−1) ⟂ 0', True);
  Test('0 ⟂ −1', True);

  Test('2 ⟂ 0', False);
  Test('9 ⟂ 0', False);
  Test('25 ⟂ 0', False);
  Test('36 ⟂ 0', False);
  Test('53 ⟂ 0', False);
  Test('106 ⟂ 0', False);
  Test('1000 ⟂ 0', False);
  Test('1024 ⟂ 0', False);

  Test('−2 ⟂ 0', False);
  Test('−9 ⟂ 0', False);
  Test('−25 ⟂ 0', False);
  Test('−36 ⟂ 0', False);
  Test('−53 ⟂ 0', False);
  Test('−106 ⟂ 0', False);
  Test('−1000 ⟂ 0', False);
  Test('−1024 ⟂ 0', False);

  Test('0 ⟂ 2', False);
  Test('0 ⟂ 9', False);
  Test('0 ⟂ 25', False);
  Test('0 ⟂ 36', False);
  Test('0 ⟂ 53', False);
  Test('0 ⟂ 106', False);
  Test('0 ⟂ 1000', False);
  Test('0 ⟂ 1024', False);

  Test('0 ⟂ −2', False);
  Test('0 ⟂ −9', False);
  Test('0 ⟂ −25', False);
  Test('0 ⟂ −36', False);
  Test('0 ⟂ −53', False);
  Test('0 ⟂ −106', False);
  Test('0 ⟂ −1000', False);
  Test('0 ⟂ −1024', False);

  Test('1152921504606846976 ⟂ 576460752303423488', False);
  Test('1152921504606846976 ⟂ 16677181699666569', True);

  Test('14181656094375 ⟂ 5489965305721', True);
  Test('14181656094375 ⟂ 16469895917163', False);
  Test('14181656094375 ⟂ 38429757140047', False);
  Test('14181656094375 ⟂ 27449826528605', False);
  Test('14181656094375 ⟂ 10979930611442', True);

  Test('−1152921504606846976 ⟂ 576460752303423488', False);
  Test('1152921504606846976 ⟂ −16677181699666569', True);

  Test('14181656094375 ⟂ −5489965305721', True);
  Test('−14181656094375 ⟂ −16469895917163', False);
  Test('−14181656094375 ⟂ 38429757140047', False);
  Test('−14181656094375 ⟂ −27449826528605', False);
  Test('14181656094375 ⟂ −10979930611442', True);

  Test('14181656094375 ⟂ 14181656094375', False);
  Test('14181656094375 ⟂ −14181656094375', False);
  Test('−14181656094375 ⟂ 14181656094375', False);
  Test('−14181656094375 ⟂ −14181656094375', False);

  Test('12 ⟂ 13.5', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');
  Test('1152921504606846976 ⟂ 16677181699666569.4', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');

  Test('12.5 ⟂ 13', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('1152921504606846976.4 ⟂ 16677181699666569', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('5 ⟂ 1/3', failure, 'An object of type integer was expected as argument 2, but an object of type rational number was given.');
  Test('5 ⟂ i', failure, 'An object of type integer was expected as argument 2, but an object of type complex number was given.');

  Test('1/3 ⟂ 3', failure, 'An object of type integer was expected as argument 1, but an object of type rational number was given.');
  Test('i ⟂ 7', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('−895965 ⟂ 14278.02', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');

  // - vectors: orthogonal

  Test('❨1, 0, 0❩ ⟂ ❨0, 1, 0❩', True);
  Test('❨1, 0, 0❩ ⟂ ❨0, 0, 1❩', True);
  Test('❨0, 1, 0❩ ⟂ ❨0, 0, 1❩', True);
  Test('❨0, 0, 1❩ ⟂ ❨1, 0, 0❩', True);
  Test('❨1, 0, 0❩ ⟂ ❨0, 2, 0❩', True);
  Test('❨1, 0, 0❩ ⟂ ❨0, 1, 1❩', True);
  Test('❨1, 0, 1❩ ⟂ ❨0, 1, 0❩', True);
  Test('❨1, 1, 1❩ ⟂ ❨1, 1, -2❩', True);
  Test('❨1, 2, 3❩ ⟂ ❨3, 2, -7/3❩', True);

  Test('❨1, 0, 0❩ ⟂ ❨1, 0, 0❩', False);
  Test('❨1, 2, 3❩ ⟂ ❨1, 3, 0❩', False);
  Test('❨2, 0, 1❩ ⟂ ❨1, 1, 1❩', False);
  Test('❨5, -4, 2❩ ⟂ ❨3, -5, 2❩', False);
  Test('❨4, -2.5, 1❩ ⟂ ❨-4, 2.5, 7.1❩', False);
  Test('❨4, 2, -5.4❩ ⟂ ❨2.45, 1.2, 0❩', False);

  Test('❨1, 0❩ ⟂ ❨0, 1❩', True);
  Test('❨1, 1❩ ⟂ ❨-1, 1❩', True);
  Test('❨1, 1❩ ⟂ ❨0, 1❩', False);
  Test('❨1, 2❩ ⟂ ❨3, -4.1❩', False);

  Test('❨0, 0, 0❩ ⟂ ❨1, 2, 3❩', True);
  Test('❨0, 0, 0❩ ⟂ ❨2, 0, 0❩', True);
  Test('❨0, 0, 0❩ ⟂ ❨1, 0, 0❩', True);
  Test('❨0, 0, 0❩ ⟂ ❨-2.1, 2.5, -6.3❩', True);
  Test('❨0, 0, 0❩ ⟂ ❨10, 0, 0❩', True);
  Test('❨0, 0, 0❩ ⟂ ❨0, 0, 1❩', True);
  Test('❨0, 0, 0❩ ⟂ ❨0, 0, 0❩', True);

  Test('❨1, 0, 0❩ ⟂ ❨0, 0, 0❩', True);
  Test('❨1, 2, 3❩ ⟂ ❨0, 0, 0❩', True);
  Test('❨7.6, 4.6, -2.5❩ ⟂ ❨0, 0, 0❩', True);

  Test('❨1, 0❩ ⟂ ❨0, 0❩', True);

  Test('❨1, 0, 0, 0, 0❩ ⟂ ❨0, 1, 0, 0, 0❩', True);
  Test('❨1, 0, 0, 0, 0❩ ⟂ ❨0, 1, 2, 1, -3❩', True);
  Test('❨1, 0, 0, 0, 0❩ ⟂ ❨1, 1, -4, 2, 1❩', False);
  Test('❨0, 0, 0, 0, 0❩ ⟂ ❨0, 1, 0, 0, 0❩', True);
  Test('❨0, 0, 0, 0, 0❩ ⟂ ❨0, 0, 0, 0, 0❩', True);

  Test('❨1, 2, 1, 4, 4❩ ⟂ ❨2, 1, 4, 2, -4❩', True);
  Test('❨1, 2, 1, 4, 4❩ ⟂ ❨2, 1, 4, 2, 5❩', False);
  Test('❨1, 2, 1, 4, 4❩ ⟂ ❨2, 1, 4, 2, 4❩', False);
  Test('❨1, -2, 1, 4, 4❩ ⟂ ❨2, 1, 4, 2, -4❩', False);
  Test('❨1, 2.01, 1, 4, 4❩ ⟂ ❨2, 1, 4, 2, -4❩', False);

  Test('❨i, 0, 0❩ ⟂ ❨0, 2, 3❩', True);
  Test('❨i, 1, 0❩ ⟂ ❨0, 2, 3❩', False);
  Test('❨3, 2, 0❩ ⟂ ❨0, 0, 1 + 2⋅i❩', True);
  Test('❨3, 2, 1❩ ⟂ ❨0, 0, 1 + 2⋅i❩', False);
  Test('❨0, 0, 0, 0❩ ⟂ ❨1, i, 2+i, −π❩', True);
  Test('❨2 + i, 3, −1, −i, 1❩ ⟂ ❨i, 2 − i, 3, 3⋅i, −1 + i❩', True);
  Test('❨2 + i, 3, −1, −i, 1❩ ⟂ ❨i, 2 − i, 3, 4⋅i, −1 + i❩', False);
  Test('❨2 + i, 3, −1, −i, 1❩ ⟂ ❨i, 2 − i, -3, 3⋅i, −1 + i❩', False);
  Test('❨2 + i, 3, −1, −i, 1❩ ⟂ ❨i, 2 − i, 3, 3⋅i, 1 + i❩', False);
  Test('❨1 + i, 3.9, 1, i, 1❩ ⟂ ❨i, 1 − i, 3, 3⋅i, −1 + i❩', False);
  Test('❨1 + i, 3.9, 1, i, 1❩ ⟂ ❨0, 0, 0, 0, 0❩', True);
  Test('❨2 + i, 3, −1, −i, 1❩ ⟂ ❨i, 2 − i, 3.000001, 3⋅i, −1 + i❩', False);

  Test('❨1, 0, 0❩ ⟂ ❨1, 0❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');
  Test('❨1, i, 0❩ ⟂ ❨1, 0❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');
  Test('❨1, 0, 0❩ ⟂ ❨1, i❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');
  Test('❨i, 0, 0❩ ⟂ ❨i, 0❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');

  // - invalid operands

  Test('❨1, 0, 0❩ ⟂ "cat"', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');
  Test('"x" ⟂ "y"', failure, 'An object of type vector was expected as argument 1, but an object of type string was given.');
  Test('true ⟂ false', failure, 'An object of type vector was expected as argument 1, but an object of type boolean was given.');

  // Direct sum

  Test('❨❨1, 2❩, ❨3, 4❩❩ ⊕ ❨❨10, 10, 20❩, ❨10, 10, 20❩, ❨10, 10, 20❩❩ ⊕ ❨❨1, 0❩, ❨0, 1❩❩',
    7, [1, 2, 0, 0, 0, 0, 0, 3, 4, 0, 0, 0, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1]);

  Test('❨❨1, 5❩, ❨3, 2❩❩ ⊕ ❨❨100❩❩ ⊕ ❨❨200❩❩ ⊕ ❨❨7, 6❩, ❨4, −6❩❩',
    6, [1, 5, 0, 0, 0, 0, 3, 2, 0, 0, 0, 0, 0, 0, 100, 0, 0, 0, 0, 0, 0, 200, 0, 0, 0, 0, 0, 0, 7, 6, 0, 0, 0, 0, 4, -6]);

  Test('❨❨1❩❩ ⊕ ❨❨2❩❩ ⊕ ❨❨3❩❩ ⊕ ❨❨4❩❩ ⊕ ❨❨5❩❩',
    5, [1, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 5]);

  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊕ ❨❨1, 1❩, ❨2, 2❩❩',
    5, [1, 2, 3, 0, 0, 4, 5, 6, 0, 0, 7, 8, 9, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 2, 2]);

  Test('2⋅IdentityMatrix(2) ⊕ 3⋅IdentityMatrix(2)',
    4, [2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 3, 0, 0, 0, 0, 3]);

  Test('3⋅IdentityMatrix(3) ⊕ ❨❨10❩❩ ⊕ 5⋅IdentityMatrix(2)',
    6, [3, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 5]);

  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩❩ ⊕ IdentityMatrix(2)',
    5, [1, 2, 3, 0, 0, 4, 5, 6, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1]);

  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩❩ ⊕ ❨❨1, 10❩, ❨2, 20❩, ❨3, 30❩❩',
    5, [1, 2, 3, 0, 0, 4, 5, 6, 0, 0, 0, 0, 0, 1, 10, 0, 0, 0, 2, 20, 0, 0, 0, 3, 30]);

  Test('❨❨1, 2, 3❩, ❨10, 20, 30❩❩ ⊕ ❨❨7❩, ❨9❩❩ ⊕ ❨❨−1, −2❩❩',
    6, [1, 2, 3, 0, 0, 0, 10, 20, 30, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 0, 9, 0, 0, 0, 0, 0, 0, -1, -2]);

  Test('❨❨1, 2, 3❩❩ ⊕ ❨❨10❩, ❨20❩, ❨30❩❩',
    4, [1, 2, 3, 0, 0, 0, 0, 10, 0, 0, 0, 20, 0, 0, 0, 30]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ ⊕ ❨❨10❩❩',
    3, [1, 2, 0, 3, 4, 0, 0, 0, 10]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ ⊕ ❨❨i❩❩',
    3, [1, 2, 0, 3, 4, 0, 0, 0, ImaginaryUnit]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ ⊕ ❨❨10❩❩',
    3, [ImaginaryUnit, 2, 0, 3, 4, 0, 0, 0, 10]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ ⊕ ❨❨i❩❩',
    3, [ImaginaryUnit, 2, 0, 3, 4, 0, 0, 0, ImaginaryUnit]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ ⊕ ❨❨10, 10, 20❩, ❨10, 10, 20❩, ❨10, 10, 20❩❩ ⊕ ❨❨1, 0❩, ❨0, i❩❩',
    7, [1, 2, 0, 0, 0, 0, 0, 3, 4, 0, 0, 0, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, ImaginaryUnit]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ ⊕ ❨❨10, 10, 20❩, ❨10, i, 20❩, ❨10, 10, 20❩❩ ⊕ ❨❨1, 0❩, ❨0, 1❩❩',
    7, [1, 2, 0, 0, 0, 0, 0, 3, 4, 0, 0, 0, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 10, ImaginaryUnit, 20, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ ⊕ ❨❨10, 10, 20❩, ❨10, 10, 20❩, ❨10, 10, 20❩❩ ⊕ ❨❨1, 0❩, ❨0, 1❩❩',
    7, [ImaginaryUnit, 2, 0, 0, 0, 0, 0, 3, 4, 0, 0, 0, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ ⊕ ❨❨10, 10, 20❩, ❨10, 10, 20❩, ❨10, 10, 20❩❩ ⊕ ❨❨1, 0❩, ❨0, i❩❩',
    7, [ImaginaryUnit, 2, 0, 0, 0, 0, 0, 3, 4, 0, 0, 0, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 10, 10, 20, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, ImaginaryUnit]);

  Test('❨❨i❩❩ ⊕ ❨❨2⋅i❩❩ ⊕ ❨❨3⋅i❩❩ ⊕ ❨❨4⋅i❩❩ ⊕ ❨❨5⋅i❩❩',
    5, [ImaginaryUnit, 0, 0, 0, 0, 0, 2*ImaginaryUnit, 0, 0, 0, 0, 0, 3*ImaginaryUnit, 0, 0, 0, 0, 0, 4*ImaginaryUnit, 0, 0, 0, 0, 0, 5*ImaginaryUnit]);

  Test('(2⋅IdentityMatrix(2) ⊕ 3⋅IdentityMatrix(3) + IdentityMatrix(5)) ⊕ ❨❨10❩❩ + 1',
    6, [4, 1, 1, 1, 1, 1, 1, 4, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 1, 1, 1, 11]);

  Test('IdentityMatrix(7) ⊕ IdentityMatrix(13) = IdentityMatrix(20)', True);

  Test('tr(5⋅IdentityMatrix(100) ⊕ 7⋅IdentityMatrix(200)) = 5⋅100 + 7⋅200', True);

  Test('tr(5⋅IdentityMatrix(1000) ⊕ 7⋅IdentityMatrix(2000)) = 5⋅1000 + 7⋅2000', True);

  Test('tr(i⋅IdentityMatrix(1000) ⊕ IdentityMatrix(2000))', 2000 + 1000*ImaginaryUnit);

  Test('A ≔ IdentityMatrix(15);', null);
  StartStopwatch;
  Test('tr(A' + DupeString(' ⊕ A', 200) + ')', 3015.0);
  StopStopwatch;
  TestDuration(0.150);

{$IFNDEF QuickTest}
  Test('A ≔ IdentityMatrix(50);', null);
  StartStopwatch;
  Test('tr(A' + DupeString(' ⊕ A', 200) + ')', 10050.0);
  StopStopwatch;
  TestDuration(1.5);
{$ENDIF}

  Test('delete(A)', success);

  Test('true ⊕ IdentityMatrix(3)', failure, 'An object of type matrix was expected as argument 1, but an object of type boolean was given.');
  Test('IdentityMatrix(3) ⊕ "A"', failure, 'An object of type matrix was expected as argument 2, but an object of type string was given.');
  Test('IdentityMatrix(3) ⊕ IdentityMatrix(5) ⊕ 43', failure, 'An object of type matrix was expected as argument 3, but an object of type integer was given.');

  // Outer product (of vectors)

  Test('❨1, 2, 3❩ ⊗ ❨1, 10, 100❩', 3, [1, 10, 100, 2, 20, 200, 3, 30, 300]);
  Test('❨1, 2, 3, 4❩ ⊗ ❨1, 10, 100❩', 3, [1, 10, 100, 2, 20, 200, 3, 30, 300, 4, 40, 400]);
  Test('❨1, 2, 3, 4, 5❩ ⊗ ❨1, 10, 100❩', 3, [1, 10, 100, 2, 20, 200, 3, 30, 300, 4, 40, 400, 5, 50, 500]);
  Test('❨1, 2, 3, 4, 5❩ ⊗ ❨1, 10❩', 2, [1, 10, 2, 20, 3, 30, 4, 40, 5, 50]);
  Test('❨1, 2❩ ⊗ ❨1, 10, 100❩', 3, [1, 10, 100, 2, 20, 200]);

  Test('❨1, i, 2❩ ⊗ ❨1, 10, 100❩', 3, [1, 10, 100, ImaginaryUnit, 10*ImaginaryUnit, 100*ImaginaryUnit, 2, 20, 200]);
  Test('❨1, 2, 3, 4❩ ⊗ ❨1, i, 2❩', 3, [1, -ImaginaryUnit, 2, 2, -2*ImaginaryUnit, 4, 3, -3*ImaginaryUnit, 6, 4, -4*ImaginaryUnit, 8]);
  Test('❨1, i, 2❩ ⊗ ❨1, i, 2❩', 3, [1, -ImaginaryUnit, 2, ImaginaryUnit, 1, 2*ImaginaryUnit, 2, -2*ImaginaryUnit, 4]);

  Test('❨1, 2, 3❩ ⊗ ❨2+3⋅i, 1−2⋅i❩', 2, [2 - 3*ImaginaryUnit, 1 + 2*ImaginaryUnit, 4 - 6*ImaginaryUnit, 2 + 4*ImaginaryUnit, 6 - 9*ImaginaryUnit, 3 + 6*ImaginaryUnit]);

  Test('❨1, 2, 3❩ ⊗ "cat"', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');
  Test('123 ⊗ ❨2+3⋅i, 1−2⋅i❩', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');
  Test('❨1, 2, 3, 4❩ ⊗ ❨1, 10❩ ⊗ ❨1, 10, 100❩', failure, 'An object of type vector was expected as argument 1, but an object of type real matrix was given.');
  Test('❨1, 2, 3, 4❩ ⊗ ❨i, 10❩ ⊗ ❨1, 10, 100❩', failure, 'An object of type vector was expected as argument 1, but an object of type complex matrix was given.');
  Test('❨1, 2, 3, 4❩ ⊗ ❨i, 10❩ ⊗ ❨i, 10, 100❩', failure, 'An object of type vector was expected as argument 1, but an object of type complex matrix was given.');

  // Hadamard product (entrywise matrix multiplication)

  Test('❨❨1, 2❩, ❨3, 4❩❩ ⊙ ❨❨10, 1❩, ❨1, −10❩❩', 2, [10, 2, 3, -40]);
  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊙ IdentityMatrix(3)', 3, [1, 0, 0, 0, 5, 0, 0, 0, 9]);

  Test('❨❨1, 2❩, ❨3, 4❩❩ ⊙ ❨❨10, i❩, ❨1, −10❩❩', 2, [10, 2*ImaginaryUnit, 3, -40]);
  Test('❨❨i, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊙ IdentityMatrix(3)', 3, [ImaginaryUnit, 0, 0, 0, 5, 0, 0, 0, 9]);

  Test('❨❨i, 2❩, ❨3, 4❩❩ ⊙ ❨❨10, i❩, ❨1, −10❩❩', 2, [10*ImaginaryUnit, 2*ImaginaryUnit, 3, -40]);

  Test('IdentityMatrix(10) ⊙ IdentityMatrix(10) = IdentityMatrix(10)', True);
  Test('IdentityMatrix(10) ⊙ ZeroMatrix(10) = ZeroMatrix(10)', True);

  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩❩ ⊙ ❨❨1, 0, 1❩, ❨−1, 0, −1❩❩', 3, [1, 0, 3, -4, 0, -6]);
  Test('❨❨i, 2, 3❩, ❨4, 5, 6❩❩ ⊙ ❨❨1, 0, 1❩, ❨−1, 0, −1❩❩', 3, [ImaginaryUnit, 0, 3, -4, 0, -6]);
  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩❩ ⊙ ❨❨i, 0, 1❩, ❨−1, 0, −1❩❩', 3, [ImaginaryUnit, 0, 3, -4, 0, -6]);
  Test('❨❨i, 2, 3❩, ❨4, 5, 6❩❩ ⊙ ❨❨i, 0, 1❩, ❨−1, 0, −1❩❩', 3, [-1, 0*ImaginaryUnit, 3, -4, 0, -6]);

  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊙ IdentityMatrix(3) ⊙ ❨❨10, 10, 10❩, ❨100, 100, 100❩, ❨1000, 1000, 1000❩❩',
    3, [10, 0, 0, 0, 500, 0, 0, 0, 9000]);

  Test('❨❨i, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊙ IdentityMatrix(3) ⊙ ❨❨10, 10, 10❩, ❨100, 100, 100❩, ❨1000, 1000, 1000❩❩',
    3, [10*ImaginaryUnit, 0, 0, 0, 500, 0, 0, 0, 9000]);

  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊙ IdentityMatrix(3) ⊙ ❨❨i, 10, 10❩, ❨100, 100, 100❩, ❨1000, 1000, 1000❩❩',
    3, [ImaginaryUnit, 0, 0, 0, 500, 0, 0, 0, 9000]);

  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊙ IdentityMatrix(2)', failure, 'Hadamard product only defined for matrices of the same size.');
  Test('❨❨i, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊙ IdentityMatrix(2)', failure, 'Hadamard product only defined for matrices of the same size.');

  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊙ IdentityMatrix(3) ⊙ ❨❨i, 10, 10❩, ❨100, 100, 100❩❩',
    failure, 'Hadamard product only defined for matrices of the same size.');

  Test('❨❨1, 2, 3❩, ❨4, 5, 6❩, ❨7, 8, 9❩❩ ⊙ "test"', failure, 'An object of type matrix was expected as argument 2, but an object of type string was given.');

  Test('10 ⊙ IdentityMatrix(3) ⊙ ❨❨10, 10, 10❩, ❨100, 100, 100❩, ❨1000, 1000, 1000❩❩', failure, 'An object of type matrix was expected as argument 1, but an object of type integer was given.');

  // Length, dimension, cardinality, value count

  Test('#"Dogs are marvellous animals."', 28);
  Test('#""', 0);
  Test('#(123⋅"A")', 123);
  Test('#❨1, 0, 0❩', 3);
  Test('#❨1, 2+i, −3, −i❩', 4);
  Test('#SequenceVector(100)', 100);
  Test('#IdentityMatrix(3)', 9);
  Test('#❨❨1, 2, i❩, ❨−3, 1+i, 0❩❩', 6);
  Test('#''(1, true, 42, IdentityMatrix(10), ❨1, 0, 0❩, color("red"), "Fancy rat")', 7);
  Test('#''()', 0);
  Test('#SequenceList(100)', 100);
  Test('#filter(SequenceList(100), IsPositive)', 100);
  Test('#filter(SequenceList(100), IsPrime)', 25);
  Test('#filter(SequenceList(100), odd)', 50);
  Test('#filter(SequenceList(100), even)', 50);
  Test('#filter(SequenceList(100), IsNegative)', 0);
  Test('#{1, 2, 3, 4, 5, 6, 7, 8, 9, 10}', 10);
  Test('#{}', 0);
  Test('#∅', 0);
  Test('#date()', 3);
  Test('#time()', 4);
  Test('#datetime()', 2);

  Test('#struct("a": 42, "b": "dog", "c": "horse").c', 5);
  Test('#struct("a": 42, "b": "dog", "c": "horse").b', 3);
  Test('#struct("a": 42, "b": "dog", "c": "horse").a', failure, 'Object type "integer" is not a suitable container.');
  Test('#struct("a": 42, "b": "dog", "c": "horse")', 3);

  Test('#10', failure, 'Object type "integer" is not a suitable container.');
  Test('#10.5', failure, 'Object type "real number" is not a suitable container.');
  Test('#i', failure, 'Object type "complex number" is not a suitable container.');
  Test('#true', failure, 'Object type "boolean" is not a suitable container.');
  Test('#(.;)', failure, 'Object type "null" is not a suitable container.');

  // FCN_Length is not a simple function and first attempts to access any object
  // through an L-value reference; this path must be tested separately.

  Test('x ≔ "Dogs are marvellous animals."; #x', 28);
  Test('x ≔ ""; #x', 0);
  Test('x ≔ (123⋅"A"); #x', 123);
  Test('x ≔ ❨1, 0, 0❩; #x', 3);
  Test('x ≔ ❨1, 2+i, −3, −i❩; #x', 4);
  Test('x ≔ SequenceVector(100); #x', 100);
  Test('x ≔ IdentityMatrix(3); #x', 9);
  Test('x ≔ ❨❨1, 2, i❩, ❨−3, 1+i, 0❩❩; #x', 6);
  Test('x ≔ ''(1, true, 42, IdentityMatrix(10), ❨1, 0, 0❩, color("red"), "Fancy rat"); #x', 7);
  Test('x ≔ ''(); #x', 0);
  Test('x ≔ SequenceList(100); #x', 100);
  Test('x ≔ filter(SequenceList(100), IsPositive); #x', 100);
  Test('x ≔ filter(SequenceList(100), IsPrime); #x', 25);
  Test('x ≔ filter(SequenceList(100), odd); #x', 50);
  Test('x ≔ filter(SequenceList(100), even); #x', 50);
  Test('x ≔ filter(SequenceList(100), IsNegative); #x', 0);
  Test('x ≔ {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}; #x', 10);
  Test('x ≔ {}; #x', 0);
  Test('x ≔ ∅; #x', 0);
  Test('x ≔ date(); #x', 3);
  Test('x ≔ time(); #x', 4);
  Test('x ≔ datetime(); #x', 2);

  Test('x ≔ struct("a": 42, "b": "dog", "c": "horse"); #x.c', 5);
  Test('x ≔ struct("a": 42, "b": "dog", "c": "horse"); #x.b', 3);
  Test('x ≔ struct("a": 42, "b": "dog", "c": "horse"); #x.a', failure, 'Object type "integer" is not a suitable container.');
  Test('x ≔ struct("a": 42, "b": "dog", "c": "horse"); #x', 3);

  Test('x ≔ 10; #x', failure, 'Object type "integer" is not a suitable container.');
  Test('x ≔ 10.5; #x', failure, 'Object type "real number" is not a suitable container.');
  Test('x ≔ i; #x', failure, 'Object type "complex number" is not a suitable container.');
  Test('x ≔ true; #x', failure, 'Object type "boolean" is not a suitable container.');
  Test('x ≔ (.;); #x', failure, 'Object type "null" is not a suitable container.');

  Test('delete(x)', success);

  // Primorial

  Test('0#', 1);
  Test('1#', 1);
  Test('2#', 2);
  Test('3#', 6);
  Test('4#', 6);
  Test('5#', 30);
  Test('6#', 30);
  Test('7#', 210);
  Test('8#', 210);
  Test('9#', 210);
  Test('10#', 210);
  Test('11#', 2310);
  Test('12#', 2310);
  Test('13#', 30030);

  Test('20#', 9699690);
  Test('30#', 6469693230);
  Test('40#', 7420738134810);
  Test('43#', 13082761331670030);
  Test('47#', 614889782588491410);

  Test('53#', 614889782588491410.0 * 53.0);
  Test('59#', 614889782588491410.0 * 53.0 * 59.0);

  Eps; Test('100#', 2.305567963945518424753102147331756070E36);

  Test('−(10#)', -210);
  Test('−10#', -210);

  Test('(−1)#', failure, 'A non-negative integer was expected as argument 1, but "-1" was given.');
  Test('(−10)#', failure, 'A non-negative integer was expected as argument 1, but "-10" was given.');

  Test('1.5#', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('"rabbit"#', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');
  Test('false#', failure, 'An object of type integer was expected as argument 1, but an object of type boolean was given.');

  Test('#IdentityMatrix(5) − (#ComplexZeroVector(5))# + 3##', 25);
  Test('#IdentityMatrix(6) − (#ComplexZeroVector(4))# + 4##', 60);
  Test('#IdentityMatrix(7) − (#ComplexZeroVector(6))# + 5##', 19 + 6469693230);

  // Set max length

  // Was extensively tested in Chapter 1, inlcuding output formatting.

  Test('π \ 2', Pi, True, '3.1');
  Test('π \ 3', Pi, True, '3.14');
  Test('π \ 4', Pi, True, '3.142');
  Test('π \ 5', Pi, True, '3.1416');
  Test('π \ 6', Pi, True, '3.14159');

  Test('π \ 10', Pi, True, '3.141592654');
  Test('π \ 12', Pi, True, '3.14159265359');

  Test('π \ 15', Pi, True, '3.14159265358979');

  Test('(π − 1) \ 2', Pi - 1, True, '2.1');
  Test('(π − 1) \ 3', Pi - 1, True, '2.14');
  Test('(π − 1) \ 4', Pi - 1, True, '2.142');
  Test('(π − 1) \ 5', Pi - 1, True, '2.1416');
  Test('(π − 1) \ 6', Pi - 1, True, '2.14159');

  Test('(π − 1) \ 10', Pi - 1, True, '2.141592654');
  Test('(π − 1) \ 12', Pi - 1, True, '2.14159265359');

  Test('(π − 1) \ 15', Pi - 1, True, '2.14159265358979');

  Test('(1/3 + π⋅i) \ 4', 1/3 + Pi*ImaginaryUnit, True, '0.3333 + 3.142⋅i');

  Test('π − 1 \ 2', Pi - 1, True, '2.1');
  Test('π − 1 \ 3', Pi - 1, True, '2.14');
  Test('π − 1 \ 4', Pi - 1, True, '2.142');
  Test('π − 1 \ 5', Pi - 1, True, '2.1416');
  Test('π − 1 \ 6', Pi - 1, True, '2.14159');

  Test('π − 1 \ 10', Pi - 1, True, '2.141592654');
  Test('π − 1 \ 12', Pi - 1, True, '2.14159265359');

  Test('π − 1 \ 15', Pi - 1, True, '2.14159265358979');

  Test('1/3 + π⋅i \ 4', 1/3 + Pi*ImaginaryUnit, True, '0.3333 + 3.142⋅i');

  Test('"Dogs are marvellous animals." \ 10', 'Dogs are marvellous animals.', True, 'Dogs are m…');
  Test('"Dogs are marvellous animals." \ 4', 'Dogs are marvellous animals.', True, 'Dogs…');
  Test('"Dogs are marvellous animals." \ 1', 'Dogs are marvellous animals.', True, 'D…');
  Test('"Dogs are marvellous animals." \ 0', 'Dogs are marvellous animals.', True, 'Dogs are marvellous animals.');

  TestSL('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 4', '(1, 2, 3, 4, …)');
  TestSL('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 4', '(i, 2, 3, 4, …)');

  TestSL('❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 3', '(1, 2, 3, …)');
  TestSL('❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩ \ 3', '(i, 2, 3, …)');

  TestML('IdentityMatrix(10) \ 3', '⎛1  0  0  ⋯⎞'#13#10'⎜0  1  0  ⋯⎟'#13#10'⎜0  0  1  ⋯⎟'#13#10'⎝⋮  ⋮  ⋮  ⋱⎠');
  TestML('i⋅IdentityMatrix(10) \ 3', '⎛i  0  0  ⋯⎞'#13#10'⎜0  i  0  ⋯⎟'#13#10'⎜0  0  i  ⋯⎟'#13#10'⎝⋮  ⋮  ⋮  ⋱⎠');

  TestSL('SequenceList(100) \ 5', '(1, 2, 3, 4, 5, …)');
  TestML('SequenceList(100) \ 5', '1'#13#10'2'#13#10'3'#13#10'4'#13#10'5'#13#10'⋮');

  Test('π \ "cat"', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
  Test('true \ 3', failure, 'An object of type "boolean" has no maximum display length setting.');
  Test('date() \ 2', failure, 'An object of type "typed structure" has no maximum display length setting.');
  Test('{1, 2, 3, 4, 5, 6, "dog"} \ 3', failure, 'An object of type "set" has no maximum display length setting.'); // but should it?!

  Test('\4', ESyntaxException, 'Operand missing before operator \ at column 1.');
  Test('π\', ESyntaxException, 'Operand missing after operator \ at column 2.');

  // Catenate

  Test('"Alpha" ~ "Beta"', 'AlphaBeta');
  Test('"alpha" ~ " beta" ~ " gamma" ~ " delta" ~ " epsilon" ~ " phi"', 'alpha beta gamma delta epsilon phi');

  Test('❨1, 2, 3❩ ~ ❨10, 20, 30❩', [1, 2, 3, 10, 20, 30]);
  Test('❨1, 2, 3❩ ~ ❨4, 5, 6❩ ~ ❨7, 8❩ ~ ❨9❩ ~ ❨10❩', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  Test('❨1, 2, i❩ ~ ❨3, 4, i❩', [1, 2, ImaginaryUnit, 3, 4, ImaginaryUnit]);
  Test('❨1, 2, i❩ ~ ❨3, 4, i❩ ~ ❨5, i❩ ~ ❨i❩', [1, 2, ImaginaryUnit, 3, 4, ImaginaryUnit, 5, ImaginaryUnit, ImaginaryUnit]);
  Test('❨1, 2, 3❩ ~ ❨10, i❩', [1, 2, 3, 10, ImaginaryUnit]);
  Test('❨1, 2, 3❩ ~ ❨10, i❩ ~ ❨4, 5❩', [1, 2, 3, 10, ImaginaryUnit, 4, 5]);
  Test('❨i, 2, 3❩ ~ ❨4, 5❩', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('SequenceVector(1, 10) ~ SequenceVector(11, 20) = SequenceVector(20)', True);
  Test('SequenceVector(1, 10000) ~ SequenceVector(10001, 20000) = SequenceVector(20000)', True);

  Test('#(ZeroVector(100)' + DupeString(' ~ ZeroVector(100)', 99) + ')', 10000);

{$IFNDEF QuickTest}
  StartStopwatch;
  Test('#(ZeroVector(1000)' + DupeString(' ~ ZeroVector(1000)', 999) + ')', 1000000);
  StopStopwatch;
  TestDuration(1);
{$ENDIF}

  Test('''(1, 2) ~ ''("dog", "cat") ~ ''(true, false) ~ ''(❨1, 0❩, ❨0, 1❩)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(1),
        ASO(2),
        ASO('dog'),
        ASO('cat'),
        ASO(true),
        ASO(false),
        ASO(ASR2(1, 0)),
        ASO(ASR2(0, 1))
      ]
    ));

  Test('''(1, 2) ~ ''("dog") ~ ''(true, 42.0, false) ~ ''(❨1, 0❩)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(1),
        ASO(2),
        ASO('dog'),
        ASO(true),
        ASO(42.0),
        ASO(false),
        ASO(ASR2(1, 0))
      ]
    ));

  Test('SequenceList(1, 10) ~ SequenceList(11, 20) = SequenceList(20)', True);
  Test('SequenceList(1, 10000) ~ SequenceList(10001, 20000) = SequenceList(20000)', True);

  Test('#(SequenceList(100)' + DupeString(' ~ SequenceList(100)', 99) + ')', 10000);

{$IFNDEF QuickTest}
  StartStopwatch;
  Test('#(SequenceList(1000)' + DupeString(' ~ SequenceList(1000)', 999) + ')', 1000000);
  StopStopwatch;
  TestDuration(1);
{$ENDIF}

  Test('"alpha" ~ ❨1, 0❩', failure, 'Invalid function arguments.');
  Test('"alpha" ~ 123', failure, 'Invalid function arguments.');
  Test('"alpha" ~ " beta" ~ " gamma" ~ " delta" ~ 10 ~ " phi"', failure, 'Invalid function arguments.');
  Test('❨1, 2, 3❩ ~ ❨4, 5, 6❩ ~ ❨7, 8❩ ~ "cat" ~ ❨10❩', failure, 'Invalid function arguments.');
  Test('❨1, 2, 3❩ ~ ❨4, 5, 6❩ ~ ❨7, 8❩ ~ 9 ~ ❨10❩', failure, 'Invalid function arguments.');
  Test('❨1, 2, 3❩ ~ ❨4, 5, 6❩ ~ ❨7, 8❩ ~ "9" ~ ❨10❩', failure, 'Invalid function arguments.');
  Test('7 ~ 7.0001', failure, 'Invalid function arguments.');
  Test('true ~ false', failure, 'Invalid function arguments.');
  Test('2 ~ 3', failure, 'Invalid function arguments.');

  // Assignment

  // Only for simple L-values. More extensive testing is done in a following
  // chapter.

  Test('x ≔ 42', 42);
  Test('x + 1', 43);
  Test('x ≔ x ⋅ 2', 84);
  Test('x − 4', 80);
  Test('x ≔ x − 3', 81);
  Test('√x', 9.0);
  Test('y ≔ x', 81);
  Test('x + y', 162);
  Test('x ≔ 50', 50);
  Test('x + y', 131);
  Test('delete(x); delete(y)', success);

  Test('a ≔ b ≔ 123', 123);
  Test('(a − 23) ⋅ (b − 100)', 2300);
  Test('b ≔ 100', 100);
  Test('(a − 23) ⋅ (b − 100)', 0);
  Test('delete(a); delete(b)', success);

  Test('a, b, c, d ≔ 5', 5);
  Test('a', 5);
  Test('b', 5);
  Test('c', 5);
  Test('d', 5);
  Test('d ≔ 6', 6);
  Test('a + b + c + d', 21);
  Test('delete(a); delete(b); delete(c); delete(d)', success);
  Test('a', failure, 'Unknown identifier "a".');

  Test('a, b, c ≔ 100, 50, 7', 7);
  Test('a + b + c', 157);
  Test('delete(a); delete(b); delete(c)', success);

  Test('d ≔ a, b, c ≔ 55', 55);
  Test('a', 55);
  Test('b', 55);
  Test('c', 55);
  Test('d', 55);
  Test('delete(a); delete(b); delete(c); delete(d)', success);

  Test('d ≔ a, b, c ≔ 100, 50, 7', 7);
  Test('a + b + c + d', 164);
  Test('delete(a); delete(b); delete(c); delete(d)', success);
  Test('a + b + c + d', failure, 'Unknown identifier "a".');

  Test('x ≔ 10.2; y ≔ 5.6; x^2 + y^2', 135.4);
  Test('x, y ≔ 7.1, 2.9; x^2 + y^2', 58.82);
  Test('delete(x); delete(y)', success);

  Test('a, b, c ≔ 1, 2', failure, 'The number of symbols (3) doesn''t match the number of values (2) in multiple assignment expression.');
  Test('a, b, c ≔ 1, 2, 3, 4', failure, 'The number of symbols (3) doesn''t match the number of values (4) in multiple assignment expression.');
  Test('a, b, c ≔ 1, 2, 3; 2⋅a + 3⋅b − c', 5);
  Test('delete(a); delete(b); delete(c);', null);

  Test('A ≔ ❨❨2, 1, 3❩, ❨3, −2, 4❩, ❨0, 1, 2❩❩; b ≔ ❨1, 3, 6❩; c ≔ 10; A⋅b + c', [33, 31, 25]);
  Eps; Test('det(A) + tr(A) + #A + #b', 1.0);
  Test('delete(A); delete(b); delete(c)', success);
  Test('det(A) + tr(A) + #A + #b', failure, 'Unknown identifier "A".');

  Test('5 ≔ 6', failure, 'Left side cannot be assigned to.');
  Test('5, 6 ≔ 2, 5', failure, 'Left side cannot be assigned to.');
  Test('5, 6 ≔ 2', failure, 'Left side cannot be assigned to.');
  Test('5, 6 ≔ 2, 5, 7', failure, 'The number of symbols (2) doesn''t match the number of values (3) in multiple assignment expression.');
  Test('1 + 1 ≔ 3', failure, 'Left side cannot be assigned to.');
  Test('sin(e) ≔ "test"', failure, 'Left side cannot be assigned to.');
  Test('a ≔ ≔ 5', ESyntaxException, 'Operand missing before operator ≔ at column 5.');
  Test('a ≔', ESyntaxException, 'Operand missing after operator ≔ at column 3.');
  Test('≔ 5', ESyntaxException, 'Operand missing before operator ≔ at column 1.');
  Test('≔', ESyntaxException, 'Operand missing before operator ≔ at column 1.');
  Test('"h" ≔ 4', failure, 'Left side cannot be assigned to.');

  Test('π ≔ 3', failure, 'Cannot modify or delete protected variable "π".');
  Test('delete(π)', failure, 'Cannot modify or delete protected variable "π".');

  // Maps to (create function)

  Test('f ≔ x ↦ x^2 + 1; f(5)', 26);
  Test('delete(f)', success);
  Test('f(5)', failure, 'Unknown identifier "f".');

  Test('(x ↦ 2⋅x − 1)(10)', 19);

  Test('f ≔ x ↦ x^2; g ≔ x ↦ x + 1; h ≔ x ↦ x/2;', null);
  Test('g(f(h(g(h(g(h(h(g(99)))))))))', 50);

  Test('f ≔ x ↦ 4⋅x^3 − x^2 + 5⋅x + 1;', null);
  Test('f(10)', 3951);

  Test('f ≔ (x, y) ↦ 3⋅x^2 + y^2;', null);
  Test('f(3, 5)', 52);

  Test('f ≔ (a, b, c, d) ↦ 2⋅a − b + 3⋅c + d;', null);
  Test('f(1, 2, 3, 4)', 13);

  Test('f ≔ A ↦ sqr(det(A));', null);
  Test('A ≔ ❨❨2, 1, 3❩, ❨3, −2, 4❩, ❨0, 1, 2❩❩;', null);
  Eps; Test('f(A)', 169.0);

  Test('delete(f); delete(g); delete(h); delete(A)', success);

  Test('f, g ≔ (x ↦ 3⋅x + 1), (x ↦ x^2/2);', null);
  Test('f(2)', 7);
  Test('g(5)', Rat(25, 2));
  Test('delete(f); delete(g)', success);

  Test('f ≔ x ↦ x^2.1;', null);
  Eps; Test('f(7)', 59.5258881579142800305386216888);
  Test('f(3, 5)', failure, '1 argument(s) expected, but 2 given.');
  Test('f≔(x,y)↦x+y/2;', null);
  Test('f(7, 8)', 11);
  Test('f(7)', failure, '2 argument(s) expected, but 1 given.');
  Test('delete(f)', success);

  Test('6 ↦ x^2', failure, 'A symbol was expected as argument 1.');
  Test('f ≔ 6 ↦ x^2', failure, 'A symbol was expected as argument 1.');
  Test('delete(f)', failure, 'Unknown identifier "f".');

  // Apply

  Test('"Hello, World!" @ UpperCase', 'HELLO, WORLD!'); // inefficient, but should work (better: UpperCase("Hello, World!"))
  Test('❨1, 2, 3, 0, 4, 5❩ @ square', [1, 4, 9, 0, 16, 25]);
  Test('❨i, 2, 3, 0, 4, 5❩ @ square', [TASC(-1), 4, 9, 0, 16, 25]);
  Test('❨❨1, 2❩, ❨3, 4❩❩ @ square', 2, [1, 4, 9, 16]);
  Test('❨❨i, 2❩, ❨3, 4❩❩ @ square', 2, [TASC(-1), 4, 9, 16]);
  Test('''(1, 2, 0, 3, 4, 5) @ square',
    TAlgosimArray.CreateWithValue(
      [
        ASO(1),
        ASO(4),
        ASO(0),
        ASO(9),
        ASO(16),
        ASO(25)
      ]
    ));
  Test('{0, 1, 3, 5, 7} @ square', intset([0, 1, 9, 25, 49]));

  Test('❨1, 2, 3, 1, 2, 3❩ @ square', [1, 4, 9, 1, 4, 9]);
  Test('❨i, 2, 3, i, 2, 3❩ @ square', [TASC(-1), 4, 9, TASC(-1), 4, 9]);
  Test('❨❨1, 2❩, ❨1, 2❩❩ @ square', 2, [1, 4, 1, 4]);
  Test('❨❨i, 2❩, ❨i, 2❩❩ @ square', 2, [TASC(-1), 4, TASC(-1), 4]);
  Test('''(1, 2, 3, 1, 2, 3) @ square',
    TAlgosimArray.CreateWithValue(
      [
        ASO(1),
        ASO(4),
        ASO(9),
        ASO(1),
        ASO(4),
        ASO(9)
      ]
    ));
  Test('{1, 2, 3, 1, 2, 3} @ square', intset([1, 4, 9]));

  Test('struct("a": 6, "b": 7) @ square', TAlgosimStructure.CreateWithValue(['a', 'b'], [ASOInt(36), ASOInt(49)]));

  Test('"" @ UpperCase', '');
  Test('''() @ square', TAlgosimArray.Create);
  Test('∅ @ square', intset([]));
  Test('struct() @ square', TAlgosimStructure.CreateWithValue([]));

  Test('SequenceList(10) @ square @ succ @ IsPrime',
    TAlgosimArray.CreateWithValue(
      [
        ASO(true), ASO(true), ASO(false), ASO(true), ASO(false), ASO(true), ASO(false), ASO(false), ASO(false), ASO(true)
      ]
    )
  );

  Test('❨4, 3, 1, 2, −5❩ @ square @ (x ↦ x/2) @ trunc', [8, 4, 0, 2, 12]);

  Test('{1, 10, 100, 1000} @ square @ (x ↦ x/100)',
    asosetex(
      [
        ASOInt(1),
        ASOInt(100),
        ASOInt(10000),
        ASO(Rat(1, 100))
      ]
    )
  );

  Test('❨−2, 2, −3, 3, −4, 4❩ @ square', [4, 4, 9, 9, 16, 16]);
  Test('{−2, 2, −3, 3, −4, 4} @ square', intset([4, 9, 16]));

  Test('f ≔ x ↦ x^2; ❨4, 3, 1, 2, −5❩ @ f @ (x ↦ x/2) @ trunc', [8, 4, 0, 2, 12]);
  Test('delete(f)', success);

  Test('102 @ square', failure, 'Cannot apply function to elements in an object of type "integer".');
  Test('true @ not', failure, 'Cannot apply function to elements in an object of type "boolean".');
  Test('color("red") @ square', failure, 'Cannot apply function to elements in an object of type "color".');

  Test('struct("a": 6, "b": 7) @ 3', failure, 'An object of type function was expected as argument 2, but an object of type integer was given.');

  // Access member

  Test('struct("a": 6, "b": 7).a', 6);
  Test('struct("a": 6, "b": 7).b', 7);
  Test('struct("a": 6, "b": 7).c', failure, 'There is no member named "c".');

  Test('S ≔ struct("a": 6, "b": 7);', null);
  Test('S.a', 6);
  Test('S.b', 7);
  Test('S.c', failure, 'There is no member named "c".');
  Test('S = struct("a": 6, "b": 7)', True);
  Test('delete(S)', success);
  Test('S.a', failure, 'Unknown identifier "S".');

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null);

  Test('Mary.mother = Jane', True);

  Test('Mary.mother.name.first', 'Jane');
  Test('Mary.mother.name.last', 'Smith');
  Test('Mary.mother.sex', 'female');
  Test('Mary.mother.age', 40);

  Test('Mary.father = Mike', True);

  Test('Mary.father.name.first', 'Mike');
  Test('Mary.father.name.last', 'Doe');
  Test('Mary.father.sex', 'male');
  Test('Mary.father.age', 45);

  Test('Mary.father.name.middle', failure, 'There is no member named "middle".');
  Test('Mary.father.dog', failure, 'There is no member named "dog".');
  Test('Mary.father.dog.name', failure, 'There is no member named "dog".');
  Test('Mary.mentor', failure, 'There is no member named "mentor".');
  Test('Daniel.father.name.first', failure, 'Unknown identifier "Daniel".');

  Test('delete(Mary); delete(Jane); delete(Mike); delete(name); delete(person)', success);

  Test('"Rats are intelligent animals.".length', failure, 'An object of type structure was expected as argument 1, but an object of type string was given.');
  Test('"Rats are intelligent animals.".length.prime', failure, 'An object of type structure was expected as argument 1, but an object of type string was given.');
  Test('(45).square', failure, 'An object of type structure was expected as argument 1, but an object of type integer was given.');
  Test('(45).square.square', failure, 'An object of type structure was expected as argument 1, but an object of type integer was given.');
  Test('(45).square.square.inverse', failure, 'An object of type structure was expected as argument 1, but an object of type integer was given.');
  Test('IdentityMatrix(100).det', failure, 'An object of type structure was expected as argument 1, but an object of type real matrix was given.');
  Test('IdentityMatrix(100).det.square', failure, 'An object of type structure was expected as argument 1, but an object of type real matrix was given.');

  Test('i.square', failure, 'Unsupported subscript operation.');
  Test('i.square.square', failure, 'Unsupported subscript operation.');
  Test('e.square', failure, 'Unsupported subscript operation.');
  Test('e.square.inverse', failure, 'Unsupported subscript operation.');
  Test('e.square.inverse.square', failure, 'Unsupported subscript operation.');
  Test('true.negation', failure, 'Unsupported subscript operation.');

  // Output

  Test('ClearHistory(); SaveHistory(true);', null);
  Test('1+1', 2);
  Test('5!', 120);
  Test('IdentityMatrix(3)', 3, [1, 0, 0, 0, 1, 0, 0, 0, 1]);
  Test('"Hello, World!"', 'Hello, World!');

  Test('§1', null);
  Test('§2', 2);
  Test('§3', 120);
  Test('§4', 3, [1, 0, 0, 0, 1, 0, 0, 0, 1]);
  Test('§5', 'Hello, World!');
  Test('§6', null);
  Test('§7', 2);
  Test('§8', 120);
  Test('§14', failure, 'Index 14 out of bounds.');

  Test('2^10', 1024);
  Test('§-1', 1024);

  Test('2^5', 32);
  Test('2^6', 64);
  Test('§-2', 32);

  Test('§-20', failure, 'Index -20 out of bounds.');
  Test('§-20', null);
  Test('§-20', 2);
  Test('§-20', 120);

  Test('§0', failure, 'Index 0 out of bounds.');

  Test('ClearHistory();', null);
  Test('1+1', 2);
  Test('5!', 120);
  Test('IdentityMatrix(3)', 3, [1, 0, 0, 0, 1, 0, 0, 0, 1]);
  Test('"Hello, World!"', 'Hello, World!');
  Test('HistoryLength()', 5);

  Test('ClearHistory(); SaveHistory(false);', null);

  Test('1+1', 2);
  Test('5!', 120);
  Test('IdentityMatrix(3)', 3, [1, 0, 0, 0, 1, 0, 0, 0, 1]);
  Test('"Hello, World!"', 'Hello, World!');

  Test('§1', null);
  Test('§2', null);
  Test('§3', null);
  Test('§4', null);
  Test('§5', null);

  Test('§"cat"', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');





  //
  //
  //  CHAPTER 3
  //  Variables, L-values, indices, and assignment.
  //
  //

  Chapter('Variables, L-values, indices, assignment');

  //
  // Simple assignments
  //

  Test('x ≔ 120', 120);
  Test('√(x + 1)', 11.0);
  Test('x ≔ x − 20', 100);
  Test('√x', 10.0);
  Test('x ≔ 55', 55);
  Test('(x − 1) / 2', 27);
  Test('y ≔ x', 55);
  Test('x + y', 110);
  Test('y ≔ 12', 12);
  Test('x + y', 67);
  Test('delete(y)', success);
  Test('x + y', failure, 'Unknown identifier "y".');
  Test('x^2', 3025);
  Test('delete(x)', success);
  Test('x^2', failure, 'Unknown identifier "x".');
  Test('x ≔ "Hello, "; y ≔ "World"', 'World');
  Test('x ~ y ~ "!"', 'Hello, World!');
  Test('delete(x); delete(y)', success);

  Test('x ≔ y ≔ 21', 21);
  Test('x', 21);
  Test('y', 21);
  Test('2⋅x + y^2', 483);
  Test('delete(x); delete(y)', success);
  Test('delete(x); delete(y)', failure, 'Unknown identifier "x".');

  Test('a, b ≔ 15', 15);
  Test('a', 15);
  Test('b', 15);
  Test('2⋅(a+b)', 60);
  Test('delete(a); delete(b)', success);
  Test('delete(a); delete(b)', failure, 'Unknown identifier "a".');

  Test('a, b ≔ 11, 23', 23);
  Test('a', 11);
  Test('b', 23);
  Test('2⋅(a+b)', 68);

  Test('a, b, c, d ≔ 10, 20, 30, "dogs! "', 'dogs! ');
  Test('a', 10);
  Test('b', 20);
  Test('c', 30);
  Test('d', 'dogs! ');
  Test('(a^2 + b − 3⋅c) ⋅ d', DupeString('dogs! ', 30));

  Test('delete(a); delete(b); delete(c); delete(d)', success);
  Test('delete(a); delete(b); delete(c); delete(d)', failure, 'Unknown identifier "a".');

  Test('a ≔ b, c ≔ 7', 7);
  Test('a', 7);
  Test('b', 7);
  Test('c', 7);

  Test('a ≔ b, c ≔ 7, 11', 11);
  Test('a', 11);
  Test('b', 7);
  Test('c', 11);
  Test('a + b + c', 29);

  Test('delete(a); delete(b); delete(c)', success);

  Test('a, b, c, d ≔ 1, 2, 3', failure, 'The number of symbols (4) doesn''t match the number of values (3) in multiple assignment expression.');
  Test('a, b, c, d ≔ 1, 2, 3, 4, 5', failure, 'The number of symbols (4) doesn''t match the number of values (5) in multiple assignment expression.');

  Test('a, b, c, d ≔ 1, 2, 3, true', true);
  Test('''(a, b, c, d) = ''(1, 2, 3, true)', true);
  Test('delete(a); delete(b); delete(c); delete(d)', success);
  Test('''(a, b, c, d) = ''(1, 2, 3, true)', failure, 'Unknown identifier "a".');

  Test('A ≔ ❨❨2, 1, 0❩, ❨0, 1, 0❩, ❨0, 2, −1❩❩; b ≔ ❨1, 2, 3❩; A⋅b', [4, 2, 1]);
  Test('''(det(A), tr(A))', TAlgosimArray.CreateWithValue([ASO(-2.0), ASO(2.0)]));
  Test('delete(A); delete(b)', success);

  Test('A, b ≔ ❨❨0, 1, 2❩, ❨0, 1, 1❩, ❨1, 0, 3❩❩, ❨5, 1, −2❩; A⋅b', [-3, -1, -1]);
  Test('''(det(A), tr(A))', TAlgosimArray.CreateWithValue([ASO(-1.0), ASO(4.0)]));
  Test('delete(A); delete(b)', success);

  Test('''(det(A), tr(A))', failure, 'Unknown identifier "A".');
  Test('delete(A); delete(b)', failure, 'Unknown identifier "A".');

  Test('a, b ≔ SequenceList(1, 10), SequenceList(11, 20); a~b = SequenceList(1, 20)', true);
  Test('delete(a); delete(b)', success);

  Eps; Test('a ≔ 2; b ≔ 3; c ≔ −1; x ≔ 1.5; y ≔ 1.2; z ≔ 1.3; a⋅x^2 + b⋅y^2 + c⋅y^2', 7.38);
  Test('delete(a); delete(b); delete(c); delete(x); delete(y); delete(z)', success);

  Test('5 ≔ 3', failure, 'Left side cannot be assigned to.');
  Test('1 + 1 ≔ 3', failure, 'Left side cannot be assigned to.');
  Test('IdentityMatrix(3) ≔ 0', failure, 'Left side cannot be assigned to.');
  Test('"cats" ≔ "dogs"', failure, 'Left side cannot be assigned to.');
  Test('1, 2 ≔ 2, 1', failure, 'Left side cannot be assigned to.');
  Test('"cats", "dogs" ≔ "mammals"', failure, 'Left side cannot be assigned to.');

  Test('π ≔ 3.1', failure, 'Cannot modify or delete protected variable "π".');
  Test('delete(π)', failure, 'Cannot modify or delete protected variable "π".');
  Test('e ≔ 2', failure, 'Cannot modify or delete protected variable "e".');
  Test('delete(e)', failure, 'Cannot modify or delete protected variable "e".');
  Test('true ≔ false', failure, 'Cannot modify or delete protected variable "true".');
  Test('delete(true)', failure, 'Cannot modify or delete protected variable "true".');

  Test('a ≔ 42', 42);
  Test('protect(a)', success);
  Test('a ≔ 100', failure, 'Cannot modify or delete protected variable "a".');
  Test('delete(a)', failure, 'Cannot modify or delete protected variable "a".');
  Test('a^2', 42*42);
  Test('a ≔ 100', failure, 'Cannot modify or delete protected variable "a".');
  Test('delete(a)', failure, 'Cannot modify or delete protected variable "a".');
  Test('unprotect(a)', success);
  Test('a ≔ 100', 100);
  Test('a^2', 100*100);
  Test('delete(a)', success);
  Test('a^2', failure, 'Unknown identifier "a".');

  Test('a ≔ 42', 42);
  Test('a#', 304250263527210);
  Test('a ≔ 32', 32);
  Test('a#', 200560490130);
  Test('protect(a)', success);
  Test('a ≔ 22', failure, 'Cannot modify or delete protected variable "a".');
  Test('a#', 200560490130);
  Test('delete(a)', failure, 'Cannot modify or delete protected variable "a".');
  Test('a#', 200560490130);
  Test('protect(a)', success); // idempotency
  Test('a ≔ 22', failure, 'Cannot modify or delete protected variable "a".');
  Test('a#', 200560490130);
  Test('delete(a)', failure, 'Cannot modify or delete protected variable "a".');
  Test('a#', 200560490130);
  Test('unprotect(a)', success);
  Test('a ≔ 22', 22);
  Test('a#', 9699690);
  Test('unprotect(a)', success); // idempotency
  Test('delete(a)', success);
  Test('a#', failure, 'Unknown identifier "a".');


  //
  // Structure members
  //

  Test('S ≔ struct("a": 42, "b": "dog", "c": "cat")',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(42), ASO('dog'), ASO('cat')]
    )
  );

  Test('S.a', 42);
  Test('S.b', 'dog');
  Test('S.c', 'cat');
  Test('S.d', failure, 'There is no member named "d".');

  Test('#S', 3);
  Test('#S.b', 'dog'.Length);

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(42), ASO('dog'), ASO('cat')]
    )
  );

  Test('S.a ≔ 123', 123);

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(123), ASO('dog'), ASO('cat')]
    )
  );

  Test('S.a', 123);
  Test('S.b', 'dog');
  Test('S.c', 'cat');
  Test('S.d', failure, 'There is no member named "d".');

  Test('S.b ≔ "fancy rat"', 'fancy rat');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(123), ASO('fancy rat'), ASO('cat')]
    )
  );

  Test('S.a', 123);
  Test('S.b', 'fancy rat');
  Test('S.c', 'cat');
  Test('S.d', failure, 'There is no member named "d".');

  Test('#S.b', 'fancy rat'.Length);
  Test('#S.c', 'cat'.Length);

  Test('S.c ≔ "rabbit"', 'rabbit');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(123), ASO('fancy rat'), ASO('rabbit')]
    )
  );

  Test('S.a', 123);
  Test('S.b', 'fancy rat');
  Test('S.c', 'rabbit');
  Test('S.d', failure, 'There is no member named "d".');

  Test('#S.b', 'fancy rat'.Length);
  Test('#S.c', 'rabbit'.Length);

  Test('S.d ≔ "horse"', failure); // TODO: specify an error message

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(123), ASO('fancy rat'), ASO('rabbit')]
    )
  );

  Test('S.a ≔ S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(123), ASO('fancy rat'), ASO('rabbit')]
    )
  );

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [
        TAlgosimStructure.CreateWithValue(
          ['a', 'b', 'c'],
          [ASOInt(123), ASO('fancy rat'), ASO('rabbit')]
        ),
        ASO('fancy rat'),
        ASO('rabbit')]
    )
  );

  Test('S.a.a', 123);
  Test('S.a.b', 'fancy rat');
  Test('S.a.c', 'rabbit');
  Test('S.a.d', failure, 'There is no member named "d".');
  Test('S.b', 'fancy rat');
  Test('S.c', 'rabbit');
  Test('S.d', failure, 'There is no member named "d".');

  Test('S.a',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(123), ASO('fancy rat'), ASO('rabbit')]
    )
  );

  Test('S.a.d ≔ "horse"', failure); // TODO: specify an error message

  Test('S.a.a ≔ 1024', 1024);

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [
        TAlgosimStructure.CreateWithValue(
          ['a', 'b', 'c'],
          [ASOInt(1024), ASO('fancy rat'), ASO('rabbit')]
        ),
        ASO('fancy rat'),
        ASO('rabbit')]
    )
  );

  Test('S.a.a', 1024);
  Test('S.a.b', 'fancy rat');
  Test('S.a.c', 'rabbit');
  Test('S.a.d', failure, 'There is no member named "d".');
  Test('S.b', 'fancy rat');
  Test('S.c', 'rabbit');
  Test('S.d', failure, 'There is no member named "d".');

  Test('S.a.b ≔ "guinea pig"', 'guinea pig');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [
        TAlgosimStructure.CreateWithValue(
          ['a', 'b', 'c'],
          [ASOInt(1024), ASO('guinea pig'), ASO('rabbit')]
        ),
        ASO('fancy rat'),
        ASO('rabbit')]
    )
  );

  Test('S.a.a', 1024);
  Test('S.a.b', 'guinea pig');
  Test('S.a.c', 'rabbit');
  Test('S.a.d', failure, 'There is no member named "d".');
  Test('S.b', 'fancy rat');
  Test('S.c', 'rabbit');
  Test('S.d', failure, 'There is no member named "d".');

  Test('#S', 3);
  Test('#S.a', 3);
  Test('#S.a.b', 'guinea pig'.Length);
  Test('#S.a.c', 'rabbit'.Length);
  Test('#S.b', 'fancy rat'.Length);
  Test('#S.c', 'rabbit'.Length);

  Test('S.a ≔ "horse"', 'horse');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASO('horse'), ASO('fancy rat'), ASO('rabbit')]
    )
  );

  Test('S.a', 'horse');
  Test('S.b', 'fancy rat');
  Test('S.c', 'rabbit');
  Test('S.d', failure, 'There is no member named "d".');

  Test('#S', 3);
  Test('#S.a', 'horse'.Length);
  Test('#S.b', 'fancy rat'.Length);
  Test('#S.c', 'rabbit'.Length);

  Test('delete(S)', success);

  Test('S', failure, 'Unknown identifier "S".');

  Test('S.a', failure, 'Unknown identifier "S".');
  Test('S.b', failure, 'Unknown identifier "S".');
  Test('S.c', failure, 'Unknown identifier "S".');
  Test('S.d', failure, 'Unknown identifier "S".');

  Test('#S', failure, 'Unknown identifier "S".');
  Test('#S.a', failure, 'Unknown identifier "S".');
  Test('#S.b', failure, 'Unknown identifier "S".');
  Test('#S.c', failure, 'Unknown identifier "S".');

  Test('S ≔ struct("a": 100, "b": 200)',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(100), ASOInt(200)]
    )
  );

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(100), ASOInt(200)]
    )
  );

  Test('S.a', 100);
  Test('S.b', 200);

  Test('S.a ≔ 150', 150);

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(150), ASOInt(200)]
    )
  );

  Test('S.a', 150);
  Test('S.b', 200);

  Test('protect(S)', success);

  Test('S.a ≔ 175', failure, 'Cannot modify or delete protected variable "S".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(150), ASOInt(200)]
    )
  );

  Test('S.a', 150);
  Test('S.b', 200);

  Test('delete(S)', failure, 'Cannot modify or delete protected variable "S".');
  Test('unprotect(S)', success);

  Test('S.a ≔ 175', 175);

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(175), ASOInt(200)]
    )
  );

  Test('S.a', 175);
  Test('S.b', 200);

  Test('delete(S)', success);

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45); John ≔ person("John", "Stone", "male", 69); Sarah ≔ person("Sarah", "Lester", "female", 25)', Sarah);
  Test('Mary ≔ struct("mother": Jane, "father": Mike)', Mary);

  Test('Jane', Jane);
  Test('Mike', Mike);
  Test('John', John);
  Test('Sarah', Sarah);
  Test('Mary', Mary);

  Test('Jane.name', Name('Jane', 'Smith'));
  Test('Jane.name.first', 'Jane');
  Test('Jane.name.last', 'Smith');
  Test('Jane.name.middle', failure, 'There is no member named "middle".');
  Test('Jane.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Jane.sex', 'female');
  Test('Jane.age', 40);
  Test('Jane.address', failure, 'There is no member named "address".');

  Test('Mike.name', Name('Mike', 'Doe'));
  Test('Mike.name.first', 'Mike');
  Test('Mike.name.last', 'Doe');
  Test('Mike.name.middle', failure, 'There is no member named "middle".');
  Test('Mike.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Mike.sex', 'male');
  Test('Mike.age', 45);
  Test('Mike.address', failure, 'There is no member named "address".');

  Test('John.name', Name('John', 'Stone'));
  Test('John.name.first', 'John');
  Test('John.name.last', 'Stone');
  Test('John.name.middle', failure, 'There is no member named "middle".');
  Test('John.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('John.sex', 'male');
  Test('John.age', 69);
  Test('John.address', failure, 'There is no member named "address".');

  Test('Sarah.name', Name('Sarah', 'Lester'));
  Test('Sarah.name.first', 'Sarah');
  Test('Sarah.name.last', 'Lester');
  Test('Sarah.name.middle', failure, 'There is no member named "middle".');
  Test('Sarah.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Sarah.sex', 'female');
  Test('Sarah.age', 25);
  Test('Sarah.address', failure, 'There is no member named "address".');

  Test('Emmy.name', failure, 'Unknown identifier "Emmy".');
  Test('Emmy.name.first', failure, 'Unknown identifier "Emmy".');
  Test('Emmy.name.last', failure, 'Unknown identifier "Emmy".');
  Test('Emmy.name.middle', failure, 'Unknown identifier "Emmy".');
  Test('Emmy.name.first.etymology', failure, 'Unknown identifier "Emmy".');
  Test('Emmy.sex', failure, 'Unknown identifier "Emmy".');
  Test('Emmy.age', failure, 'Unknown identifier "Emmy".');
  Test('Emmy.address', failure, 'Unknown identifier "Emmy".');

  Test('Mary.mother', Jane);
  Test('Mary.mother.name', Name('Jane', 'Smith'));
  Test('Mary.mother.name.first', 'Jane');
  Test('Mary.mother.name.last', 'Smith');
  Test('Mary.mother.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Mary.mother.name.middle', failure, 'There is no member named "middle".');
  Test('Mary.mother.sex', 'female');
  Test('Mary.mother.age', 40);
  Test('Mary.mother.address', failure, 'There is no member named "address".');

  Test('Mary.father', Mike);
  Test('Mary.father.name', Name('Mike', 'Doe'));
  Test('Mary.father.name.first', 'Mike');
  Test('Mary.father.name.last', 'Doe');
  Test('Mary.father.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Mary.father.name.middle', failure, 'There is no member named "middle".');
  Test('Mary.father.sex', 'male');
  Test('Mary.father.age', 45);
  Test('Mary.father.address', failure, 'There is no member named "address".');

  Test('Mary.address', failure, 'There is no member named "address".');

  Test('Harry ≔ struct("mother": Sarah, "father": John)',
    TAlgosimStructure.CreateWithValue(
      ['mother', 'father'],
      [Sarah, John]
    )
  );

  Test('Harry.mother', Sarah);
  Test('Harry.mother.name', Name('Sarah', 'Lester'));
  Test('Harry.mother.name.first', 'Sarah');
  Test('Harry.mother.name.last', 'Lester');
  Test('Harry.mother.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.mother.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.mother.sex', 'female');
  Test('Harry.mother.age', 25);
  Test('Harry.mother.address', failure, 'There is no member named "address".');

  Test('Harry.father', John);
  Test('Harry.father.name', Name('John', 'Stone'));
  Test('Harry.father.name.first', 'John');
  Test('Harry.father.name.last', 'Stone');
  Test('Harry.father.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.father.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.father.sex', 'male');
  Test('Harry.father.age', 69);
  Test('Harry.father.address', failure, 'There is no member named "address".');

  Test('Harry.address', failure, 'There is no member named "address".');

  Test('Harry.father.name.first ≔ "Jonathan"', 'Jonathan');

  Test('Harry.mother', Sarah);
  Test('Harry.mother.name', Name('Sarah', 'Lester'));
  Test('Harry.mother.name.first', 'Sarah');
  Test('Harry.mother.name.last', 'Lester');
  Test('Harry.mother.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.mother.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.mother.sex', 'female');
  Test('Harry.mother.age', 25);
  Test('Harry.mother.address', failure, 'There is no member named "address".');

  Test('Harry.father', Person('Jonathan', 'Stone', 'male', 69));
  Test('Harry.father.name', Name('Jonathan', 'Stone'));
  Test('Harry.father.name.first', 'Jonathan');
  Test('Harry.father.name.last', 'Stone');
  Test('Harry.father.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.father.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.father.sex', 'male');
  Test('Harry.father.age', 69);
  Test('Harry.father.address', failure, 'There is no member named "address".');

  Test('Harry.address', failure, 'There is no member named "address".');

  Test('Harry.father.name ≔ name("Bill", "Gates")', Name('Bill', 'Gates'));

  Test('Harry.mother', Sarah);
  Test('Harry.mother.name', Name('Sarah', 'Lester'));
  Test('Harry.mother.name.first', 'Sarah');
  Test('Harry.mother.name.last', 'Lester');
  Test('Harry.mother.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.mother.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.mother.sex', 'female');
  Test('Harry.mother.age', 25);
  Test('Harry.mother.address', failure, 'There is no member named "address".');

  Test('Harry.father', Person('Bill', 'Gates', 'male', 69));
  Test('Harry.father.name', Name('Bill', 'Gates'));
  Test('Harry.father.name.first', 'Bill');
  Test('Harry.father.name.last', 'Gates');
  Test('Harry.father.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.father.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.father.sex', 'male');
  Test('Harry.father.age', 69);
  Test('Harry.father.address', failure, 'There is no member named "address".');

  Test('Harry.address', failure, 'There is no member named "address".');

  Test('Harry.father ≔ Mike', Mike);

  Test('Harry.mother', Sarah);
  Test('Harry.mother.name', Name('Sarah', 'Lester'));
  Test('Harry.mother.name.first', 'Sarah');
  Test('Harry.mother.name.last', 'Lester');
  Test('Harry.mother.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.mother.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.mother.sex', 'female');
  Test('Harry.mother.age', 25);
  Test('Harry.mother.address', failure, 'There is no member named "address".');

  Test('Harry.father', Mike);
  Test('Harry.father.name', Name('Mike', 'Doe'));
  Test('Harry.father.name.first', 'Mike');
  Test('Harry.father.name.last', 'Doe');
  Test('Harry.father.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.father.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.father.sex', 'male');
  Test('Harry.father.age', 45);
  Test('Harry.father.address', failure, 'There is no member named "address".');

  Test('Harry.address', failure, 'There is no member named "address".');

  Test('Harry.mother ≔ Jane', Jane);

  Test('Harry.mother', Jane);
  Test('Harry.mother.name', Name('Jane', 'Smith'));
  Test('Harry.mother.name.first', 'Jane');
  Test('Harry.mother.name.last', 'Smith');
  Test('Harry.mother.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.mother.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.mother.sex', 'female');
  Test('Harry.mother.age', 40);
  Test('Harry.mother.address', failure, 'There is no member named "address".');

  Test('Harry.father', Mike);
  Test('Harry.father.name', Name('Mike', 'Doe'));
  Test('Harry.father.name.first', 'Mike');
  Test('Harry.father.name.last', 'Doe');
  Test('Harry.father.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.father.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.father.sex', 'male');
  Test('Harry.father.age', 45);
  Test('Harry.father.address', failure, 'There is no member named "address".');

  Test('Harry.address', failure, 'There is no member named "address".');

  Test('Harry.father.name.first.etymology ≔ "From ..."', failure, 'Object "Harry.father.name.first" is not a structure.');
  Test('Harry.father.name.middle.etymology ≔ "From ..."', failure); // TODO: could specify error message
  Test('Harry.mentor.name.first.etymology ≔ "From ..."', failure); // TODO: could specify error message

  Test('Harry.mother', Jane);
  Test('Harry.mother.name', Name('Jane', 'Smith'));
  Test('Harry.mother.name.first', 'Jane');
  Test('Harry.mother.name.last', 'Smith');
  Test('Harry.mother.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.mother.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.mother.sex', 'female');
  Test('Harry.mother.age', 40);
  Test('Harry.mother.address', failure, 'There is no member named "address".');

  Test('Harry.father', Mike);
  Test('Harry.father.name', Name('Mike', 'Doe'));
  Test('Harry.father.name.first', 'Mike');
  Test('Harry.father.name.last', 'Doe');
  Test('Harry.father.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.father.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.father.sex', 'male');
  Test('Harry.father.age', 45);
  Test('Harry.father.address', failure, 'There is no member named "address".');

  Test('Harry.address', failure, 'There is no member named "address".');

  Test('Jane', Jane);
  Test('Mike', Mike);
  Test('John', John);
  Test('Sarah', Sarah);
  Test('Mary', Mary);

  Test('Harry = Mary', True); // parent data

  Test('Harry.mother.age ≔ 70', 70);
  Test('Harry.father.age ≔ 75.5', 75.5);

  Test('Harry.mother', Person('Jane', 'Smith', 'female', 70));
  Test('Harry.mother.name', Name('Jane', 'Smith'));
  Test('Harry.mother.name.first', 'Jane');
  Test('Harry.mother.name.last', 'Smith');
  Test('Harry.mother.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.mother.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.mother.sex', 'female');
  Test('Harry.mother.age', 70);
  Test('Harry.mother.address', failure, 'There is no member named "address".');

  Test('Harry.father.name', Name('Mike', 'Doe'));
  Test('Harry.father.name.first', 'Mike');
  Test('Harry.father.name.last', 'Doe');
  Test('Harry.father.name.first.etymology', failure, 'Unsupported subscript operation.');
  Test('Harry.father.name.middle', failure, 'There is no member named "middle".');
  Test('Harry.father.sex', 'male');
  Test('Harry.father.age', 75.5);
  Test('Harry.father.address', failure, 'There is no member named "address".');

  Test('Harry.address', failure, 'There is no member named "address".');

  Test('Harry.mother ≔ "Mrs Jane Smith"; Harry.father ≔ "Mr Mike Doe"', 'Mr Mike Doe');

  Test('Harry',
    TAlgosimStructure.CreateWithValue(
      ['mother', 'father'],
      [ASO('Mrs Jane Smith'), ASO('Mr Mike Doe')]
    )
  );

  Test('Harry.mother', 'Mrs Jane Smith');
  Test('Harry.father', 'Mr Mike Doe');
  Test('Harry.address', failure, 'There is no member named "address".');

  Test('Harry.mother, Harry.father ≔ "Ann", "Richard"', 'Richard');

  Test('Harry',
    TAlgosimStructure.CreateWithValue(
      ['mother', 'father'],
      [ASO('Ann'), ASO('Richard')]
    )
  );

  Test('Harry.mother', 'Ann');
  Test('Harry.father', 'Richard');
  Test('Harry.address', failure, 'There is no member named "address".');

  Test('winner ≔ Jane', Jane);
  Test('winner.name.first', 'Jane');
  Test('winner ≔ Mike', Mike);
  Test('winner.name.first', 'Mike');
  Test('protect(winner)', success);
  Test('winner ≔ Sarah', failure, 'Cannot modify or delete protected variable "winner".');
  Test('winner.name.first', 'Mike');
  Test('delete(winner)', failure, 'Cannot modify or delete protected variable "winner".');
  Test('unprotect(winner)', success);
  Test('winner ≔ Sarah', Sarah);
  Test('winner.name.first', 'Sarah');
  Test('winner.name.first ≔ "Sara"', 'Sara');
  Test('winner', Person('Sara', 'Lester', 'female', 25));
  Test('delete(winner)', success);

  Test('winner ≔ Sarah', Sarah);
  Test('winner', Sarah);
  Test('winner.name.first', 'Sarah');
  Test('protect(winner)', success);
  Test('winner.name.first ≔ "Sara"', failure, 'Cannot modify or delete protected variable "winner".');
  Test('winner.name ≔ "Sara Lester"', failure, 'Cannot modify or delete protected variable "winner".');
  Test('winner', Sarah);
  Test('winner.name.first', 'Sarah');
  Test('unprotect(winner)', success);
  Test('winner.name.first ≔ "Sara"', 'Sara');
  Test('winner', person('Sara', 'Lester', 'female', 25));
  Test('winner.name.first', 'Sara');
  Test('delete(winner)', success);

  Test('Jane', Jane);
  Test('Mike', Mike);
  Test('John', John);
  Test('Sarah', Sarah);
  Test('Mary', Mary);

  Test('Mary = Harry', False);
  Test('Mary ≠ Harry', True);
  Test('#Mary = #Harry', True);

  Test('#Mary', 2);
  Test('#Mary.mother', 3);
  Test('#Mary.mother.name', 2);
  Test('#Mary.mother.name.last', 'Smith'.Length);
  Test('#Mary.mother.name.last.etymology', failure, 'Unsupported subscript operation.');

  Test('delete(Jane); delete(Mike); delete(John); delete(Sarah); delete(Harry)', success);
  Test('Mary', Mary);

  Test('protect(Mary.mother)', failure); // cannot protect only a part of a variable

  Test('delete(Mary)', success);
  Test('Mary', failure, 'Unknown identifier "Mary".');

  // Subscripts

  Test('S ≔ struct("a": 100, "b": 200, "c": 300)',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(100), ASOInt(200), ASOInt(300)]
    )
  );

  Test('S["a"]', 100);
  Test('S["b"]', 200);
  Test('S["c"]', 300);
  Test('S["d"]', failure, 'There is no member named "d".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(100), ASOInt(200), ASOInt(300)]
    )
  );

  Test('S["a"] ≔ 150', 150);

  Test('S["a"]', 150);
  Test('S["b"]', 200);
  Test('S["c"]', 300);
  Test('S["d"]', failure, 'There is no member named "d".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c'],
      [ASOInt(150), ASOInt(200), ASOInt(300)]
    )
  );

  Test('S["d"] ≔ 400', 400);

  Test('S["a"]', 150);
  Test('S["b"]', 200);
  Test('S["c"]', 300);
  Test('S["d"]', 400);
  Test('S["e"]', failure, 'There is no member named "e".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c', 'd'],
      [ASOInt(150), ASOInt(200), ASOInt(300), ASOInt(400)]
    )
  );

  Test('S["b"] ≔ 250', 250);

  Test('S["a"]', 150);
  Test('S["b"]', 250);
  Test('S["c"]', 300);
  Test('S["d"]', 400);
  Test('S["e"]', failure, 'There is no member named "e".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c', 'd'],
      [ASOInt(150), ASOInt(250), ASOInt(300), ASOInt(400)]
    )
  );

  Test('S["d"] ≔ 1024', 1024);

  Test('S["a"]', 150);
  Test('S["b"]', 250);
  Test('S["c"]', 300);
  Test('S["d"]', 1024);
  Test('S["e"]', failure, 'There is no member named "e".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c', 'd'],
      [ASOInt(150), ASOInt(250), ASOInt(300), ASOInt(1024)]
    )
  );

  Test('S["dog"] ≔ "Sally"', 'Sally');

  Test('S["a"]', 150);
  Test('S["b"]', 250);
  Test('S["c"]', 300);
  Test('S["d"]', 1024);
  Test('S["dog"]', 'Sally');
  Test('S["e"]', failure, 'There is no member named "e".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c', 'd', 'dog'],
      [ASOInt(150), ASOInt(250), ASOInt(300), ASOInt(1024), ASO('Sally')]
    )
  );

  Test('protect(S)', success);

  Test('S["c"] ≔ 350', failure, 'Cannot modify or delete protected variable "S".');

  Test('S["a"]', 150);
  Test('S["b"]', 250);
  Test('S["c"]', 300);
  Test('S["d"]', 1024);
  Test('S["dog"]', 'Sally');
  Test('S["e"]', failure, 'There is no member named "e".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c', 'd', 'dog'],
      [ASOInt(150), ASOInt(250), ASOInt(300), ASOInt(1024), ASO('Sally')]
    )
  );

  Test('S["rat"] ≔ "Fiona"', failure, 'Cannot modify or delete protected variable "S".');

  Test('S["a"]', 150);
  Test('S["b"]', 250);
  Test('S["c"]', 300);
  Test('S["d"]', 1024);
  Test('S["dog"]', 'Sally');
  Test('S["e"]', failure, 'There is no member named "e".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c', 'd', 'dog'],
      [ASOInt(150), ASOInt(250), ASOInt(300), ASOInt(1024), ASO('Sally')]
    )
  );

  Test('unprotect(S)', success);

  Test('S["c"], S["rat"] ≔ 350, "Fiona"', 'Fiona');

  Test('S["a"]', 150);
  Test('S["b"]', 250);
  Test('S["c"]', 350);
  Test('S["d"]', 1024);
  Test('S["dog"]', 'Sally');
  Test('S["rat"]', 'Fiona');
  Test('S["e"]', failure, 'There is no member named "e".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b', 'c', 'd', 'dog', 'rat'],
      [ASOInt(150), ASOInt(250), ASOInt(350), ASOInt(1024), ASO('Sally'), ASO('Fiona')]
    )
  );

  Test('S["a"]["part1"]', failure, 'Unsupported subscript operation.');
  Test('S["a"]["part1"] ≔ 10.5', failure, 'Unsupported subscript operation.');
  Test('S["a"].part1', failure, 'Unsupported subscript operation.'); // could be more specific
  Test('S["a"].part1 ≔ "cat"', failure, 'Object "S[a]" is not a structure.');

  Test('delete(S)', success);

  Test('S["a"]', failure, 'Unknown identifier "S".');
  Test('S["b"]', failure, 'Unknown identifier "S".');
  Test('S["c"]', failure, 'Unknown identifier "S".');
  Test('S["d"]', failure, 'Unknown identifier "S".');
  Test('S["dog"]', failure, 'Unknown identifier "S".');
  Test('S["rat"]', failure, 'Unknown identifier "S".');
  Test('S["e"]', failure, 'Unknown identifier "S".');

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45); John ≔ person("John", "Stone", "male", 69); Sarah ≔ person("Sarah", "Lester", "female", 25)', Sarah);
  Test('Mary ≔ struct("mother": Jane, "father": Mike)', Mary);

  Test('Jane', Jane);
  Test('Mike', Mike);
  Test('John', John);
  Test('Sarah', Sarah);
  Test('Mary', Mary);

  Test('Jane["name"]', Name('Jane', 'Smith'));
  Test('Jane["name"]["first"]', 'Jane');
  Test('Jane["name"]["last"]', 'Smith');
  Test('Jane["name"]["middle"]', failure, 'There is no member named "middle".');
  Test('Jane["name"]["first"]["etymology"]', failure, 'Unsupported subscript operation.');
  Test('Jane["sex"]', 'female');
  Test('Jane["age"]', 40);
  Test('Jane["address"]', failure, 'There is no member named "address".');

  Test('Jane["name"].first', 'Jane');
  Test('Jane["name"].last', 'Smith');
  Test('Jane["name"].middle', failure, 'There is no member named "middle".');
  Test('Jane["name"].first["etymology"]', failure, 'Unsupported subscript operation.');
  Test('Jane["name"]["first"].etymology', failure, 'Unsupported subscript operation.');
  Test('Jane["name"].first.etymology', failure, 'Unsupported subscript operation.');

  Test('Jane.name["first"]', 'Jane');
  Test('Jane.name["last"]', 'Smith');
  Test('Jane.name["middle"]', failure, 'There is no member named "middle".');
  Test('Jane.name["first"]["etymology"]', failure, 'Unsupported subscript operation.');
  Test('Jane.name.first["etymology"]', failure, 'Unsupported subscript operation.');
  Test('Jane.name["first"].etymology', failure, 'Unsupported subscript operation.');
  Test('Jane.name["first"]["etymology"]', failure, 'Unsupported subscript operation.');

  Test('Jane["address"] ≔ "Storgatan 5"', 'Storgatan 5');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [Name('Jane', 'Smith'), ASO('female'), ASOInt(40), ASO('Storgatan 5')]
    )
  );

  Test('Jane["address"] ≔ "Storgatan 6"', 'Storgatan 6');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [Name('Jane', 'Smith'), ASO('female'), ASOInt(40), ASO('Storgatan 6')]
    )
  );

  Test('Jane["name"]["first"] ≔ "Jenny"', 'Jenny');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [Name('Jenny', 'Smith'), ASO('female'), ASOInt(40), ASO('Storgatan 6')]
    )
  );

  Test('Jane["name"]["middle"] ≔ "Sara"', 'Sara');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Jenny'), ASO('Smith'), ASO('Sara')]
        ),
        ASO('female'),
        ASOInt(40),
        ASO('Storgatan 6')]
    )
  );

  Test('Jane["name"].middle ≔ "Sarah"', 'Sarah');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Jenny'), ASO('Smith'), ASO('Sarah')]
        ),
        ASO('female'),
        ASOInt(40),
        ASO('Storgatan 6')]
    )
  );

  Test('Jane.name["middle"] ≔ "Sophie"', 'Sophie');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Jenny'), ASO('Smith'), ASO('Sophie')]
        ),
        ASO('female'),
        ASOInt(40),
        ASO('Storgatan 6')]
    )
  );

  Test('protect("Jane")', success);

  Test('Jane.name["middle"] ≔ "Josephine"', failure, 'Cannot modify or delete protected variable "Jane".');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Jenny'), ASO('Smith'), ASO('Sophie')]
        ),
        ASO('female'),
        ASOInt(40),
        ASO('Storgatan 6')]
    )
  );

  Test('Jane["name"]["middle"] ≔ "Josephine"', failure, 'Cannot modify or delete protected variable "Jane".');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Jenny'), ASO('Smith'), ASO('Sophie')]
        ),
        ASO('female'),
        ASOInt(40),
        ASO('Storgatan 6')]
    )
  );

  Test('Jane["name"].middle ≔ "Josephine"', failure, 'Cannot modify or delete protected variable "Jane".');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Jenny'), ASO('Smith'), ASO('Sophie')]
        ),
        ASO('female'),
        ASOInt(40),
        ASO('Storgatan 6')]
    )
  );

  Test('unprotect(Jane)', success);

  Test('Jane["name"].middle ≔ "Josephine"', 'Josephine');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Jenny'), ASO('Smith'), ASO('Josephine')]
        ),
        ASO('female'),
        ASOInt(40),
        ASO('Storgatan 6')]
    )
  );

  Test('Jane["name"]["middle"] ≔ "Emma"', 'Emma');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Jenny'), ASO('Smith'), ASO('Emma')]
        ),
        ASO('female'),
        ASOInt(40),
        ASO('Storgatan 6')]
    )
  );

  Test('Jane["name"]["first"]["etymology"] ≔ "From ..."', failure, 'Unsupported subscript operation.');
  Test('Jane["name"]["first"].etymology ≔ "From ..."', failure, 'Object "Jane[name][first]" is not a structure.');

  Test('Jane',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age', 'address'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Jenny'), ASO('Smith'), ASO('Emma')]
        ),
        ASO('female'),
        ASOInt(40),
        ASO('Storgatan 6')]
    )
  );

  Test('delete(Jane); delete(Mike); delete(John); delete(Sarah); delete(Mary)', success);

  // Numerical subscripts

  Test('S ≔ struct("alpha": 100, "beta": 200, "gamma": "dogs!")',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASOInt(100), ASOInt(200), ASO('dogs!')]
    )
  );

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASOInt(100), ASOInt(200), ASO('dogs!')]
    )
  );

  Test('S[1]', 100);
  Test('S[2]', 200);
  Test('S[3]', 'dogs!');
  Test('S[4]', failure, 'Index 4 out of bounds.');

  Test('S[0]', failure, 'Index 0 out of bounds.');

  Test('S[−1]', 'dogs!');
  Test('S[−2]', 200);
  Test('S[−3]', 100);
  Test('S[−4]', failure, 'Index -4 out of bounds.');

  Test('S[1] ≔ 150', 150);

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASOInt(150), ASOInt(200), ASO('dogs!')]
    )
  );

  Test('S[2] ≔ 250.5', 250.5);

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASOInt(150), ASO(250.5), ASO('dogs!')]
    )
  );

  Test('S[3] ≔ ❨1, 0, 0❩', ASO(ASR3(1, 0, 0)));

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASOInt(150), ASO(250.5), ASO(ASR3(1, 0, 0))]
    )
  );

  Test('S[−1] ≔ "cats!"', 'cats!');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASOInt(150), ASO(250.5), ASO('cats!')]
    )
  );

  Test('S[−2] ≔ 1.5', 1.5);

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASOInt(150), ASO(1.5), ASO('cats!')]
    )
  );

  Test('S[−3] ≔ true', true);

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASO(True), ASO(1.5), ASO('cats!')]
    )
  );

  Test('S[4] ≔ "rats!"', failure, 'Index 4 out of bounds.');
  Test('S[-4] ≔ "rats!"', failure, 'Index -4 out of bounds.');
  Test('S[0] ≔ "rats!"', failure, 'Index 0 out of bounds.');

  Test('protect(S)', success);

  Test('S[1] ≔ "horses!"', failure, 'Cannot modify or delete protected variable "S".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASO(True), ASO(1.5), ASO('cats!')]
    )
  );

  Test('unprotect(S)', success);

  Test('S[1] ≔ "horses!"', 'horses!');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASO('horses!'), ASO(1.5), ASO('cats!')]
    )
  );

  Test('protect(S)', success);

  Test('S[-1] ≔ "dogs!"', failure, 'Cannot modify or delete protected variable "S".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASO('horses!'), ASO(1.5), ASO('cats!')]
    )
  );

  Test('unprotect(S)', success);

  Test('S[-1] ≔ "dogs!"', 'dogs!');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASO('horses!'), ASO(1.5), ASO('dogs!')]
    )
  );

  Test('protect(S)', success);

  Test('S[4] ≔ "horses!"', failure, 'Cannot modify or delete protected variable "S".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASO('horses!'), ASO(1.5), ASO('dogs!')]
    )
  );

  Test('delete(S)', failure, 'Cannot modify or delete protected variable "S".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['alpha', 'beta', 'gamma'],
      [ASO('horses!'), ASO(1.5), ASO('dogs!')]
    )
  );

  Test('unprotect(S); delete(S)', success);

  Test('S[1]', failure, 'Unknown identifier "S".');

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45); John ≔ person("John", "Stone", "male", 69); Sarah ≔ person("Sarah", "Lester", "female", 25)', Sarah);
  Test('Mary ≔ struct("mother": Jane, "father": Mike)', Mary);

  Test('Jane', Jane);
  Test('Mike', Mike);
  Test('John', John);
  Test('Sarah', Sarah);
  Test('Mary', Mary);

  Test('Mike[1]', Name('Mike', 'Doe'));
  Test('Mike[2]', 'male');
  Test('Mike[3]', 45);
  Test('Mike[4]', failure, 'Index 4 out of bounds.');

  Test('Mike[0]', failure, 'Index 0 out of bounds.');

  Test('Mike[-1]', 45);
  Test('Mike[-2]', 'male');
  Test('Mike[-3]', Name('Mike', 'Doe'));
  Test('Mike[-4]', failure, 'Index -4 out of bounds.');

  Test('Mike.name.first', 'Mike');
  Test('Mike.name.last', 'Doe');
  Test('Mike.name["first"]', 'Mike');
  Test('Mike.name["last"]', 'Doe');
  Test('Mike.name[1]', 'Mike');
  Test('Mike.name[2]', 'Doe');
  Test('Mike.name[-2]', 'Mike');
  Test('Mike.name[-1]', 'Doe');

  Test('Mike["name"].first', 'Mike');
  Test('Mike["name"].last', 'Doe');
  Test('Mike["name"]["first"]', 'Mike');
  Test('Mike["name"]["last"]', 'Doe');
  Test('Mike["name"][1]', 'Mike');
  Test('Mike["name"][2]', 'Doe');
  Test('Mike["name"][-2]', 'Mike');
  Test('Mike["name"][-1]', 'Doe');

  Test('Mike[1].first', 'Mike');
  Test('Mike[1].last', 'Doe');
  Test('Mike[1]["first"]', 'Mike');
  Test('Mike[1]["last"]', 'Doe');
  Test('Mike[1][1]', 'Mike');
  Test('Mike[1][2]', 'Doe');
  Test('Mike[1][-2]', 'Mike');
  Test('Mike[1][-1]', 'Doe');

  Test('Mike[1][3]', failure, 'Index 3 out of bounds.');
  Test('Mike[1][0]', failure, 'Index 0 out of bounds.');
  Test('Mike[1][-3]', failure, 'Index -3 out of bounds.');

  Test('Mike[−1] ≔ 46', 46);

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Mike', 'Doe'), ASO('male'), ASOInt(46)]
    )
  );

  Test('Mike[2] ≔ "man"', 'man');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Mike', 'Doe'), ASO('man'), ASOInt(46)]
    )
  );

  Test('Mike[1] ≔ name("Michael", "Daniels")',
    Name('Michael', 'Daniels'));


  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Michael', 'Daniels'), ASO('man'), ASOInt(46)]
    )
  );

  Test('Mike[1][2] ≔ "Gambon"', 'Gambon');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Michael', 'Gambon'), ASO('man'), ASOInt(46)]
    )
  );

  Test('Mike[1].last ≔ "Smith"', 'Smith');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Michael', 'Smith'), ASO('man'), ASOInt(46)]
    )
  );

  Test('Mike[1]["last"] ≔ "Bell"', 'Bell');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Michael', 'Bell'), ASO('man'), ASOInt(46)]
    )
  );

  Test('Mike.name["last"] ≔ "Potter"', 'Potter');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Michael', 'Potter'), ASO('man'), ASOInt(46)]
    )
  );

  Test('Mike.name.last ≔ "Hall"', 'Hall');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Michael', 'Hall'), ASO('man'), ASOInt(46)]
    )
  );

  Test('Mike.name[2] ≔ "Rodgers"', 'Rodgers');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Michael', 'Rodgers'), ASO('man'), ASOInt(46)]
    )
  );

  Test('Mike.name[3] ≔ "John"', failure, 'Index 3 out of bounds.');

  Test('Mike[1]["middle"] ≔ "John"', 'John');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Michael'), ASO('Rodgers'), ASO('John')]
        ),
        ASO('man'),
        ASOInt(46)
      ]
    )
  );

  Test('Mike.name[3] ≔ "Jonathan"', 'Jonathan');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Michael'), ASO('Rodgers'), ASO('Jonathan')]
        ),
        ASO('man'),
        ASOInt(46)
      ]
    )
  );

  Test('protect(Mike)', success);

  Test('Mike.name[3] ≔ "Gilbert"', failure, 'Cannot modify or delete protected variable "Mike".');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Michael'), ASO('Rodgers'), ASO('Jonathan')]
        ),
        ASO('man'),
        ASOInt(46)
      ]
    )
  );

  Test('unprotect(Mike)', success);

  Test('Mike.name[3] ≔ "Gilbert"', 'Gilbert');

  Test('Mike',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [
        TAlgosimStructure.CreateWithValue(
          ['first', 'last', 'middle'],
          [ASO('Michael'), ASO('Rodgers'), ASO('Gilbert')]
        ),
        ASO('man'),
        ASOInt(46)
      ]
    )
  );

  Test('Mary[1]', Jane);
  Test('Mary[1][1][1]', 'Jane');
  Test('Mary[2]', Mike);
  Test('Mary[2][1][2]', 'Doe');

  Test('Mary[1] ≔ Sarah', Sarah);

  Test('Mary[1]', Sarah);
  Test('Mary[1][1][1]', 'Sarah');
  Test('Mary[2]', Mike);
  Test('Mary[2][1][2]', 'Doe');

  Test('Mary[2] ≔ John', John);

  Test('Mary[1]', Sarah);
  Test('Mary[1][1][1]', 'Sarah');
  Test('Mary[2]', John);
  Test('Mary[2][1][2]', 'Stone');

  Test('Mary.father[1][2]', 'Stone');
  Test('Mary[2].name[2]', 'Stone');
  Test('Mary[2][1].last', 'Stone');
  Test('Mary[2].name.last', 'Stone');
  Test('Mary.father.name[2]', 'Stone');

  Test('Mary[2][1][2] ≔ "Frog"', 'Frog');
  Test('Mary.father.name', Name('John', 'Frog'));

  Test('Mary.father[1][2] ≔ "Armstrong"', 'Armstrong');
  Test('Mary.father.name', Name('John', 'Armstrong'));

  Test('Mary.father.name[2] ≔ "Smith"', 'Smith');
  Test('Mary.father.name', Name('John', 'Smith'));

  Test('Mary[2].name.last ≔ "Gates"', 'Gates');
  Test('Mary.father.name', Name('John', 'Gates'));

  Test('Mary[-1].name.last ≔ "Roberts"', 'Roberts');
  Test('Mary.father.name', Name('John', 'Roberts'));

  Test('protect(Mary)', success);

  Test('Mary[2][1][2] ≔ "Frog"', failure, 'Cannot modify or delete protected variable "Mary".');
  Test('Mary.father[1][2] ≔ "Frog"', failure, 'Cannot modify or delete protected variable "Mary".');
  Test('Mary.father.name[2] ≔ "Frog"', failure, 'Cannot modify or delete protected variable "Mary".');
  Test('Mary[2].name.last ≔ "Frog"', failure, 'Cannot modify or delete protected variable "Mary".');

  Test('Mary.father.name', Name('John', 'Roberts'));

  Test('unprotect(Mary)', success);

  Test('Mary[2][1][2] ≔ "Frog"', 'Frog');
  Test('Mary.father.name', Name('John', 'Frog'));

  Test('protect(Mary)', success);
  Test('Mary[2][1]["middle"] ≔ "Big"', failure, 'Cannot modify or delete protected variable "Mary".');
  Test('Mary.father.name', Name('John', 'Frog'));

  Test('unprotect(Mary)', success);
  Test('Mary[2][1]["middle"] ≔ "Big"', 'Big');
  Test('Mary.father.name',
    TAlgosimStructure.CreateWithValue(
      ['first', 'last', 'middle'],
      [ASO('John'), ASO('Frog'), ASO('Big')]
    )
  );

  Test('protect(Mary[1])', failure);

  Test('delete(Jane); delete(Mike); delete(John); delete(Sarah); delete(Mary)', success);


  //
  // Lists
  //

  Test('L ≔ ''(10, 20, "dog", ❨1, 0, 0❩)',
    TAlgosimArray.CreateWithValue(
      [ASOInt(10), ASOInt(20), ASO('dog'), ASO(ASR3(1, 0, 0))]
    )
  );

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASOInt(10), ASOInt(20), ASO('dog'), ASO(ASR3(1, 0, 0))]
    )
  );

  Test('L[1]', 10);
  Test('L[2]', 20);
  Test('L[3]', 'dog');
  Test('L[4]', ASO(ASR3(1, 0, 0)));
  Test('L[5]', failure, 'Index 5 out of bounds.');

  Test('L[-5]', failure, 'Index -5 out of bounds.');
  Test('L[-4]', 10);
  Test('L[-3]', 20);
  Test('L[-2]', 'dog');
  Test('L[-1]', ASO(ASR3(1, 0, 0)));

  Test('L[0]', failure, 'Index 0 out of bounds.');

  Test('L["a"]', failure, 'Unsupported subscript operation.');
  Test('L.a', failure, 'Unsupported subscript operation.');

  Test('L[1] ≔ 10.5', 10.5);

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASO(10.5), ASOInt(20), ASO('dog'), ASO(ASR3(1, 0, 0))]
    )
  );

  Test('L[1]', 10.5);
  Test('L[2]', 20);
  Test('L[3]', 'dog');
  Test('L[4]', ASO(ASR3(1, 0, 0)));
  Test('L[5]', failure, 'Index 5 out of bounds.');

  Test('L[-5]', failure, 'Index -5 out of bounds.');
  Test('L[-4]', 10.5);
  Test('L[-3]', 20);
  Test('L[-2]', 'dog');
  Test('L[-1]', ASO(ASR3(1, 0, 0)));

  Test('L[0]', failure, 'Index 0 out of bounds.');

  Test('L["a"]', failure, 'Unsupported subscript operation.');
  Test('L.a', failure, 'Unsupported subscript operation.');

  Test('L[2] ≔ "rat"', 'rat');

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASO(10.5), ASO('rat'), ASO('dog'), ASO(ASR3(1, 0, 0))]
    )
  );

  Test('L[1]', 10.5);
  Test('L[2]', 'rat');
  Test('L[3]', 'dog');
  Test('L[4]', ASO(ASR3(1, 0, 0)));
  Test('L[5]', failure, 'Index 5 out of bounds.');

  Test('L[-5]', failure, 'Index -5 out of bounds.');
  Test('L[-4]', 10.5);
  Test('L[-3]', 'rat');
  Test('L[-2]', 'dog');
  Test('L[-1]', ASO(ASR3(1, 0, 0)));

  Test('L[0]', failure, 'Index 0 out of bounds.');

  Test('L["a"]', failure, 'Unsupported subscript operation.');
  Test('L.a', failure, 'Unsupported subscript operation.');

  Test('L[−1] ≔ ❨0, 1, 0❩', ASO(ASR3(0, 1, 0)));

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASO(10.5), ASO('rat'), ASO('dog'), ASO(ASR3(0, 1, 0))]
    )
  );

  Test('L[1]', 10.5);
  Test('L[2]', 'rat');
  Test('L[3]', 'dog');
  Test('L[4]', ASO(ASR3(0, 1, 0)));
  Test('L[5]', failure, 'Index 5 out of bounds.');

  Test('L[-5]', failure, 'Index -5 out of bounds.');
  Test('L[-4]', 10.5);
  Test('L[-3]', 'rat');
  Test('L[-2]', 'dog');
  Test('L[-1]', ASO(ASR3(0, 1, 0)));

  Test('L[0]', failure, 'Index 0 out of bounds.');

  Test('L["a"]', failure, 'Unsupported subscript operation.');
  Test('L.a', failure, 'Unsupported subscript operation.');

  Test('L[5] ≔ "cat"', failure, 'Index 5 out of bounds.');
  Test('L[-5] ≔ "cat"', failure, 'Index -5 out of bounds.');
  Test('L[0] ≔ "cat"', failure, 'Index 0 out of bounds.');

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASO(10.5), ASO('rat'), ASO('dog'), ASO(ASR3(0, 1, 0))]
    )
  );

  Test('protect(L)', success);
  Test('L[1] ≔ 123', failure, 'Cannot modify or delete protected variable "L".');
  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASO(10.5), ASO('rat'), ASO('dog'), ASO(ASR3(0, 1, 0))]
    )
  );

  Test('unprotect(L)', success);
  Test('L[1] ≔ 123', 123);
  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASOInt(123), ASO('rat'), ASO('dog'), ASO(ASR3(0, 1, 0))]
    )
  );

  Test('delete(L)', success);
  Test('L[1]', failure, 'Unknown identifier "L".');

  Test('animals ≔ ''("dog", "cat", "rat", "rabbit")',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('colors ≔ ''("red", "green", "blue")',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('ints ≔ ''(1, 2, 3, 4, 5)',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(4), ASOInt(5)]
    )
  );

  Test('animals',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('colors',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('ints',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(4), ASOInt(5)]
    )
  );

  Test('arr ≔ ''(animals, colors, ints)',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(4), ASOInt(5)]
        )
      ]
    )
  );

  Test('arr',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(4), ASOInt(5)]
        )
      ]
    )
  );

  Test('arr[1]',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('arr[2]',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('arr[3]',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(4), ASOInt(5)]
    )
  );

  Test('arr[-3]',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('arr[-2]',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('arr[-1]',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(4), ASOInt(5)]
    )
  );

  Test('arr[1][2]', 'cat');
  Test('arr[2][3]', 'blue');
  Test('arr[3][-2]', 4);
  Test('arr[-1][2]', 2);
  Test('arr[4][2]', failure, 'Index 4 out of bounds.');
  Test('arr[-4][2]', failure, 'Index -4 out of bounds.');
  Test('arr[0][2]', failure, 'Index 0 out of bounds.');
  Test('arr[3][5]', 5);
  Test('arr[3][6]', failure, 'Index 6 out of bounds.');
  Test('arr[3][-6]', failure, 'Index -6 out of bounds.');
  Test('arr[3][0]', failure, 'Index 0 out of bounds.');

  Test('arr[1][−1] ≔ "guinea pig"', 'guinea pig');

  Test('arr',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(4), ASOInt(5)]
        )
      ]
    )
  );

  Test('arr[1]',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
    )
  );

  Test('arr[2]',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('arr[3]',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(4), ASOInt(5)]
    )
  );

  Test('arr[3][2] ≔ 20.5', 20.5);

  Test('arr',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(1), ASO(20.5), ASOInt(3), ASOInt(4), ASOInt(5)]
        )
      ]
    )
  );

  Test('arr[1]',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
    )
  );

  Test('arr[2]',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('arr[3]',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASO(20.5), ASOInt(3), ASOInt(4), ASOInt(5)]
    )
  );

  Test('arr[2] ≔ ''("white", "black")',
    TAlgosimArray.CreateWithValue(
      [ASO('white'), ASO('black')]
    )
  );

  Test('arr',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('white'), ASO('black')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(1), ASO(20.5), ASOInt(3), ASOInt(4), ASOInt(5)]
        )
      ]
    )
  );

  Test('arr[1]',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
    )
  );

  Test('arr[2]',
    TAlgosimArray.CreateWithValue(
      [ASO('white'), ASO('black')]
    )
  );

  Test('arr[3]',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASO(20.5), ASOInt(3), ASOInt(4), ASOInt(5)]
    )
  );

  Test('arr[4] ≔ ''(true, false)', failure, 'Index 4 out of bounds.');
  Test('arr[2][3] ≔ "pink"', failure, 'Index 3 out of bounds.');

  Test('arr',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('white'), ASO('black')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(1), ASO(20.5), ASOInt(3), ASOInt(4), ASOInt(5)]
        )
      ]
    )
  );

  Test('arr[1]',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
    )
  );

  Test('arr[2]',
    TAlgosimArray.CreateWithValue(
      [ASO('white'), ASO('black')]
    )
  );

  Test('arr[3]',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASO(20.5), ASOInt(3), ASOInt(4), ASOInt(5)]
    )
  );

  Test('animals',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('colors',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('ints',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(4), ASOInt(5)]
    )
  );

  Test('protect(arr)', success);

  Test('arr[1][2] ≔ "horse"', failure, 'Cannot modify or delete protected variable "arr".');

  Test('arr',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('white'), ASO('black')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(1), ASO(20.5), ASOInt(3), ASOInt(4), ASOInt(5)]
        )
      ]
    )
  );

  Test('unprotect(arr)', success);

  Test('arr[1][2] ≔ "horse"', 'horse');

  Test('arr',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('horse'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('white'), ASO('black')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(1), ASO(20.5), ASOInt(3), ASOInt(4), ASOInt(5)]
        )
      ]
    )
  );

  Test('protect(arr)', success);

  Test('arr[3] ≔ ''(7, 8)', failure, 'Cannot modify or delete protected variable "arr".');

  Test('arr',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('horse'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('white'), ASO('black')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(1), ASO(20.5), ASOInt(3), ASOInt(4), ASOInt(5)]
        )
      ]
    )
  );

  Test('unprotect(arr)', success);

  Test('arr[3] ≔ ''(7, 8)',
    TAlgosimArray.CreateWithValue(
      [ASOInt(7), ASOInt(8)]
    )
  );

  Test('arr',
    TAlgosimArray.CreateWithValue(
      [
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('horse'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('white'), ASO('black')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASOInt(7), ASOInt(8)]
        )
      ]
    )
  );

  Test('delete(animals); delete(colors); delete(ints); delete(arr)', success);
  Test('arr[1][−1] ≔ "guinea pig"', failure, 'Unknown identifier "arr".');


  //
  // Lists of structures
  //

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45); John ≔ person("John", "Stone", "male", 69); Sarah ≔ person("Sarah", "Lester", "female", 25)', Sarah);

  Test('Jane', Jane);
  Test('Mike', Mike);
  Test('John', John);
  Test('Sarah', Sarah);

  Test('Members ≔ ''(Jane, Mike, John, Sarah)',
    TAlgosimArray.CreateWithValue(
      [Jane, Mike, John, Sarah]
    )
  );

  Test('Members[1]', Jane);
  Test('Members[2]', Mike);
  Test('Members[3]', John);
  Test('Members[4]', Sarah);

  Test('Members[1].name', Name('Jane', 'Smith'));
  Test('Members[1].name.first', 'Jane');
  Test('Members[1].name.last', 'Smith');

  Test('Members[1]["name"]', Name('Jane', 'Smith'));
  Test('Members[1]["name"]["first"]', 'Jane');
  Test('Members[1]["name"]["last"]', 'Smith');

  Test('Members[1][1]', Name('Jane', 'Smith'));
  Test('Members[1][1][1]', 'Jane');
  Test('Members[1][1][2]', 'Smith');

  Test('Members[1].name["last"]', 'Smith');
  Test('Members[1]["name"].last', 'Smith');

  Test('Members[1].name[2]', 'Smith');
  Test('Members[1][1].last', 'Smith');

  Test('Members[3].name.first ≔ "Jonathan"', 'Jonathan');
  Test('Members[3]', Person('Jonathan', 'Stone', 'male', 69));

  Test('Members[3]["name"][2] ≔ "Rodgers"', 'Rodgers');
  Test('Members[3]', Person('Jonathan', 'Rodgers', 'male', 69));

  Test('Members[3].name ≔ name("Bill", "Gates")', Name('Bill', 'Gates'));
  Test('Members[3]', Person('Bill', 'Gates', 'male', 69));

  Test('Members[3] ≔ person("Michael", "Gambon", "male", 79)', Person('Michael', 'Gambon', 'male', 79));
  Test('Members[3]', Person('Michael', 'Gambon', 'male', 79));

  Test('Members[1]', Jane);
  Test('Members[2]', Mike);
  Test('Members[4]', Sarah);

  Test('protect(Members)', success);
  Test('Members[3].name.first ≔ "Jonathan"', failure, 'Cannot modify or delete protected variable "Members".');
  Test('Members[3].name.last ≔ "Rodgers"', failure, 'Cannot modify or delete protected variable "Members".');
  Test('Members[3].name ≔ name("Bill", "Gates")', failure, 'Cannot modify or delete protected variable "Members".');
  Test('Members[3] ≔ person("Andrew", "Smith", "male", 85)', failure, 'Cannot modify or delete protected variable "Members".');
  Test('Members[3]', Person('Michael', 'Gambon', 'male', 79));
  Test('delete(Members)', failure, 'Cannot modify or delete protected variable "Members".');

  Test('unprotect(Members)', success);
  Test('Members[3].name.first ≔ "Jonathan"', 'Jonathan');
  Test('Members[3]', Person('Jonathan', 'Gambon', 'male', 79));

  Test('delete(Jane); delete(Mike); delete(John); delete(Sarah); delete(Members)', success);
  Test('Members[3].name.last ≔ "Rodgers"', failure, 'Unknown identifier "Members".');
  Test('Members[3]["name"][-1] ≔ "Rodgers"', failure, 'Unknown identifier "Members".');


  //
  // Structures with lists as members
  //

  Test('ints ≔ ''(10, 20, 30, 40)',
    TAlgosimArray.CreateWithValue(
      [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
    )
  );

  Test('animals ≔ ''("dog", "cat", "rat", "rabbit")',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('colors ≔ ''("red", "green", "blue")',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('vectors ≔ ''(❨1, 0❩, ❨0, 1❩)',
    TAlgosimArray.CreateWithValue(
      [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
    )
  );

  Test('things ≔ struct("ints": ints, "animals": animals, "colors": colors, "vectors": vectors)',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
        )
      ]
    )
  );

  Test('ints',
    TAlgosimArray.CreateWithValue(
      [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
    )
  );

  Test('animals',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('colors',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('vectors',
    TAlgosimArray.CreateWithValue(
      [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
    )
  );

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
        )
      ]
    )
  );

  Test('things.ints',
    TAlgosimArray.CreateWithValue(
      [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
    )
  );

  Test('things.animals',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('things.colors',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('things.vectors',
    TAlgosimArray.CreateWithValue(
      [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
    )
  );

  Test('things["ints"]',
    TAlgosimArray.CreateWithValue(
      [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
    )
  );

  Test('things["animals"]',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('things["colors"]',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('things["vectors"]',
    TAlgosimArray.CreateWithValue(
      [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
    )
  );

  Test('things[1]',
    TAlgosimArray.CreateWithValue(
      [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
    )
  );

  Test('things[2]',
    TAlgosimArray.CreateWithValue(
      [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
    )
  );

  Test('things[3]',
    TAlgosimArray.CreateWithValue(
      [ASO('red'), ASO('green'), ASO('blue')]
    )
  );

  Test('things[4]',
    TAlgosimArray.CreateWithValue(
      [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
    )
  );

  Test('things.matrices', failure, 'There is no member named "matrices".');
  Test('things["matrices"]', failure, 'There is no member named "matrices".');
  Test('things[5]', failure, 'Index 5 out of bounds.');

  Test('things.ints[3]', 30);
  Test('things.ints[−1]', 40);

  Test('things.animals[1]', 'dog');
  Test('things.animals[2]', 'cat');
  Test('things.animals[3]', 'rat');
  Test('things.animals[4]', 'rabbit');
  Test('things.animals[5]', failure, 'Index 5 out of bounds.');

  Test('things.colors[-1]', 'blue');
  Test('things.colors[-2]', 'green');
  Test('things.colors[-3]', 'red');
  Test('things.colors[-4]', failure, 'Index -4 out of bounds.');

  Test('things.vectors[1]', ASO(ASR2(1, 0)));
  Test('things.vectors[2]', ASO(ASR2(0, 1)));
  Test('things.vectors[3]', failure, 'Index 3 out of bounds.');

  Test('things.matrices[1]', failure, 'There is no member named "matrices".');

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('rabbit')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
        )
      ]
    )
  );

  Test('things.animals[4] ≔ "guinea pig"', 'guinea pig');

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
        )
      ]
    )
  );

  Test('protect(things)', success);

  Test('things.animals[4] ≔ "horse"', failure, 'Cannot modify or delete protected variable "things".');

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('guinea pig')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
        )
      ]
    )
  );

  Test('unprotect(things)', success);

  Test('things.animals[4] ≔ "horse"', 'horse');

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('horse')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
        )
      ]
    )
  );

  Test('protect(things)', success);

  Test('things.vectors ≔ ''(❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩)', failure, 'Cannot modify or delete protected variable "things".');

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('horse')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
        )
      ]
    )
  );

  Test('unprotect(things)', success);

  Test('things.vectors ≔ ''(❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩)',
    TAlgosimArray.CreateWithValue(
      [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 1, 0)), ASO(ASR3(0, 0, 1))]
    )
  );

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('horse')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 1, 0)), ASO(ASR3(0, 0, 1))]
        )
      ]
    )
  );

  Test('things["vectors"] ≔ ''(❨1, 0, 0❩, ❨0, 2, 0❩, ❨0, 0, 3❩)',
    TAlgosimArray.CreateWithValue(
      [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 2, 0)), ASO(ASR3(0, 0, 3))]
    )
  );

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('horse')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 2, 0)), ASO(ASR3(0, 0, 3))]
        )
      ]
    )
  );

  Test('things[4] ≔ ''(❨1, 0, 0❩, ❨0, 2, 0❩, ❨0, 0, 30❩)',
    TAlgosimArray.CreateWithValue(
      [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 2, 0)), ASO(ASR3(0, 0, 30))]
    )
  );

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('horse')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 2, 0)), ASO(ASR3(0, 0, 30))]
        )
      ]
    )
  );

  Test('things.bools ≔ ''(true, false)', failure); // could be more specific

  Test('things["bools"] ≔ ''(true, false)',
    TAlgosimArray.CreateWithValue(
      [ASO(True), ASO(False)]
    )
  );

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors', 'bools'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(20), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('horse')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 2, 0)), ASO(ASR3(0, 0, 30))]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(True), ASO(False)]
        )
      ]
    )
  );

  Test('things["ints"][2] ≔ 200', 200);

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors', 'bools'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(200), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('horse')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('green'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 2, 0)), ASO(ASR3(0, 0, 30))]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(True), ASO(False)]
        )
      ]
    )
  );

  Test('things[3][2] ≔ "white"', 'white');

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors', 'bools'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(200), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('horse')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('white'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 2, 0)), ASO(ASR3(0, 0, 30))]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(True), ASO(False)]
        )
      ]
    )
  );

  Test('things[30][2] ≔ "white"', failure, 'Index 30 out of bounds.');

  Test('things',
    TAlgosimStructure.CreateWithValue(
      ['ints', 'animals', 'colors', 'vectors', 'bools'],
      [
        TAlgosimArray.CreateWithValue(
          [ASOInt(10), ASOInt(200), ASOInt(30), ASOInt(40)]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('dog'), ASO('cat'), ASO('rat'), ASO('horse')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO('red'), ASO('white'), ASO('blue')]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(ASR3(1, 0, 0)), ASO(ASR3(0, 2, 0)), ASO(ASR3(0, 0, 30))]
        ),
        TAlgosimArray.CreateWithValue(
          [ASO(True), ASO(False)]
        )
      ]
    )
  );

  Test('delete(ints); delete(animals); delete(colors); delete(vectors); delete(things)', success);


  //
  // Non-ASO members
  //

  // Strings

  Test('S ≔ "Dogs are cute."', 'Dogs are cute.');

  Test('S', 'Dogs are cute.');

  Test('S[1]', 'D');
  Test('S[2]', 'o');
  Test('S[3]', 'g');
  Test('S[4]', 's');
  Test('S[5]', ' ');
  Test('S[6]', 'a');
  Test('S[7]', 'r');
  Test('S[8]', 'e');
  Test('S[9]', ' ');
  Test('S[10]', 'c');
  Test('S[11]', 'u');
  Test('S[12]', 't');
  Test('S[13]', 'e');
  Test('S[14]', '.');

  Test('S[15]', failure, 'Index 15 out of bounds.');

  Test('S[-1]', '.');
  Test('S[-2]', 'e');
  Test('S[-3]', 't');
  Test('S[-4]', 'u');
  Test('S[-5]', 'c');
  Test('S[-6]', ' ');
  Test('S[-7]', 'e');
  Test('S[-8]', 'r');
  Test('S[-9]', 'a');
  Test('S[-10]', ' ');
  Test('S[-11]', 's');
  Test('S[-12]', 'g');
  Test('S[-13]', 'o');
  Test('S[-14]', 'D');

  Test('S[-15]', failure, 'Index -15 out of bounds.');

  Test('S[0]', failure, 'Index 0 out of bounds.');

  Test('S[14] ≔ "!"', '!');

  Test('S', 'Dogs are cute!');

  Test('S[1]', 'D');
  Test('S[2]', 'o');
  Test('S[3]', 'g');
  Test('S[4]', 's');
  Test('S[5]', ' ');
  Test('S[6]', 'a');
  Test('S[7]', 'r');
  Test('S[8]', 'e');
  Test('S[9]', ' ');
  Test('S[10]', 'c');
  Test('S[11]', 'u');
  Test('S[12]', 't');
  Test('S[13]', 'e');
  Test('S[14]', '!');

  Test('S[15]', failure, 'Index 15 out of bounds.');

  Test('S[-1]', '!');
  Test('S[-2]', 'e');
  Test('S[-3]', 't');
  Test('S[-4]', 'u');
  Test('S[-5]', 'c');
  Test('S[-6]', ' ');
  Test('S[-7]', 'e');
  Test('S[-8]', 'r');
  Test('S[-9]', 'a');
  Test('S[-10]', ' ');
  Test('S[-11]', 's');
  Test('S[-12]', 'g');
  Test('S[-13]', 'o');
  Test('S[-14]', 'D');

  Test('S[-15]', failure, 'Index -15 out of bounds.');

  Test('S[0]', failure, 'Index 0 out of bounds.');

  Test('S[-1] ≔ "‼"', '‼');

  Test('S', 'Dogs are cute‼');

  Test('S[1]', 'D');
  Test('S[2]', 'o');
  Test('S[3]', 'g');
  Test('S[4]', 's');
  Test('S[5]', ' ');
  Test('S[6]', 'a');
  Test('S[7]', 'r');
  Test('S[8]', 'e');
  Test('S[9]', ' ');
  Test('S[10]', 'c');
  Test('S[11]', 'u');
  Test('S[12]', 't');
  Test('S[13]', 'e');
  Test('S[14]', '‼');

  Test('S[15]', failure, 'Index 15 out of bounds.');

  Test('S[-1]', '‼');
  Test('S[-2]', 'e');
  Test('S[-3]', 't');
  Test('S[-4]', 'u');
  Test('S[-5]', 'c');
  Test('S[-6]', ' ');
  Test('S[-7]', 'e');
  Test('S[-8]', 'r');
  Test('S[-9]', 'a');
  Test('S[-10]', ' ');
  Test('S[-11]', 's');
  Test('S[-12]', 'g');
  Test('S[-13]', 'o');
  Test('S[-14]', 'D');

  Test('S[-15]', failure, 'Index -15 out of bounds.');

  Test('S[0]', failure, 'Index 0 out of bounds.');

  Test('S[2] ≔ "ö"', 'ö');

  Test('S', 'Dögs are cute‼');

  Test('S[1]', 'D');
  Test('S[2]', 'ö');
  Test('S[3]', 'g');
  Test('S[4]', 's');
  Test('S[5]', ' ');
  Test('S[6]', 'a');
  Test('S[7]', 'r');
  Test('S[8]', 'e');
  Test('S[9]', ' ');
  Test('S[10]', 'c');
  Test('S[11]', 'u');
  Test('S[12]', 't');
  Test('S[13]', 'e');
  Test('S[14]', '‼');

  Test('S[15]', failure, 'Index 15 out of bounds.');

  Test('S[-1]', '‼');
  Test('S[-2]', 'e');
  Test('S[-3]', 't');
  Test('S[-4]', 'u');
  Test('S[-5]', 'c');
  Test('S[-6]', ' ');
  Test('S[-7]', 'e');
  Test('S[-8]', 'r');
  Test('S[-9]', 'a');
  Test('S[-10]', ' ');
  Test('S[-11]', 's');
  Test('S[-12]', 'g');
  Test('S[-13]', 'ö');
  Test('S[-14]', 'D');

  Test('S[-15]', failure, 'Index -15 out of bounds.');

  Test('S[0]', failure, 'Index 0 out of bounds.');

  Test('S[15] ≔ "!"', failure, 'Index 15 out of bounds.');
  Test('S[25] ≔ "!"', failure, 'Index 25 out of bounds.');
  Test('S[-15] ≔ "!"', failure, 'Index -15 out of bounds.');
  Test('S[-25] ≔ "!"', failure, 'Index -25 out of bounds.');
  Test('S[0] ≔ "!"', failure, 'Index 0 out of bounds.');

  Test('S ≔ "0000000000"; for(i, 0, 9, (S[i + 1] ≔ chr(ord("0") + i))); S', '0123456789');

  Test('S ≔ "cats"', 'cats');
  Test('S', 'cats');
  Test('protect(S)', success);
  Test('S[1] ≔ "C"', failure, 'Cannot modify or delete protected variable "S".');
  Test('S', 'cats');
  Test('unprotect(S)', success);
  Test('S[1] ≔ "C"', 'C');
  Test('S', 'Cats');
  Test('protect(S)', success);
  Test('S[-1] ≔ "!"', failure, 'Cannot modify or delete protected variable "S".');
  Test('S', 'Cats');
  Test('unprotect(S)', success);
  Test('S[-1] ≔ "!"', '!');
  Test('S', 'Cat!');
  Test('delete(S)', success);

  Test('S ≔ "Dogs are marvellous animals."', 'Dogs are marvellous animals.');
  Test('S', 'Dogs are marvellous animals.');
  Test('S[1] ≔ 5', failure, 'A character was expected, but an object of type "integer" given, when setting a character in a string.');
  Test('S[2] ≔ 3.0', failure, 'A character was expected, but an object of type "real number" given, when setting a character in a string.');
  Test('S[3] ≔ i', failure, 'A character was expected, but an object of type "complex number" given, when setting a character in a string.');
  Test('S[4] ≔ true', failure, 'A character was expected, but an object of type "boolean" given, when setting a character in a string.');
  Test('S[5] ≔ ZeroVector(4)', failure, 'A character was expected, but an object of type "real vector" given, when setting a character in a string.');
  Test('S[6] ≔ IdentityMatrix(10)', failure, 'A character was expected, but an object of type "real matrix" given, when setting a character in a string.');
  Test('S[7] ≔ "dog"', failure, 'A character was expected, but an object of type "string" given, when setting a character in a string.');
  Test('S[8] ≔ "E"', 'E');
  Test('S', 'Dogs arE marvellous animals.');
  Test('delete(S)', success);

  // Strings in structs

  Test('S ≔ struct("a": 42, "b": "dog")',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('dog')]
    )
  );

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('dog')]
    )
  );

  Test('S.b[1]', 'd');
  Test('S.b[2]', 'o');
  Test('S.b[3]', 'g');
  Test('S.b[4]', failure, 'Index 4 out of bounds.');

  Test('S.b[-1]', 'g');
  Test('S.b[-2]', 'o');
  Test('S.b[-3]', 'd');
  Test('S.b[-4]', failure, 'Index -4 out of bounds.');

  Test('S.b[0]', failure, 'Index 0 out of bounds.');

  Test('S.b[1] ≔ "D"', 'D');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('Dog')]
    )
  );

  Test('S.b[2] ≔ "a"', 'a');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('Dag')]
    )
  );

  Test('S.b[3] ≔ "l"', 'l');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('Dal')]
    )
  );

  Test('S.b[4] ≔ "!"', failure, 'Index 4 out of bounds.');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('Dal')]
    )
  );

  Test('S.b[-1] ≔ "m"', 'm');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('Dam')]
    )
  );

  Test('S.b[-2] ≔ "u"', 'u');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('Dum')]
    )
  );

  Test('S.b[-3] ≔ "r"', 'r');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('rum')]
    )
  );

  Test('S.b[-4] ≔ "b"', failure, 'Index -4 out of bounds.');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('rum')]
    )
  );

  Test('S.b[0] ≔ "b"', failure, 'Index 0 out of bounds.');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('rum')]
    )
  );

  Test('protect(S)', success);

  Test('S.b[1] ≔ "R"', failure, 'Cannot modify or delete protected variable "S".');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('rum')]
    )
  );

  Test('unprotect(S)', success);

  Test('S.b[1] ≔ "R"', 'R');

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASOInt(42), ASO('Rum')]
    )
  );

  Test('delete(S)', success);

  // Strings in lists

  Test('L ≔ ''(1, 2, "cat", ❨1, 0❩)',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASO('cat'), ASO(ASR2(1, 0))]
    )
  );

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASO('cat'), ASO(ASR2(1, 0))]
    )
  );

  Test('L[3][1] ≔ "C"', 'C');

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASO('Cat'), ASO(ASR2(1, 0))]
    )
  );

  Test('L[3][-1] ≔ "r"', 'r');

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASO('Car'), ASO(ASR2(1, 0))]
    )
  );

  Test('protect(L)', success);

  Test('L[3][2] ≔ "o"', failure, 'Cannot modify or delete protected variable "L".');

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASO('Car'), ASO(ASR2(1, 0))]
    )
  );

  Test('unprotect(L)', success);

  Test('L[3][2] ≔ "o"', 'o');

  Test('L',
    TAlgosimArray.CreateWithValue(
      [ASOInt(1), ASOInt(2), ASO('Cor'), ASO(ASR2(1, 0))]
    )
  );

  Test('delete(L)', success);

  // Real vectors

  Test('v ≔ ❨4, 1, −2❩', ASO(ASR3(4, 1, -2)));
  Test('v', ASO(ASR3(4, 1, -2)));

  Test('v[1]', 4.0);
  Test('v[2]', 1.0);
  Test('v[3]', -2.0);
  Test('v[4]', failure, 'Index 4 out of bounds.');

  Test('v[-1]', -2.0);
  Test('v[-2]', 1.0);
  Test('v[-3]', 4.0);
  Test('v[-4]', failure, 'Index -4 out of bounds.');

  Test('v[0]', failure, 'Index 0 out of bounds.');

  Test('v', ASO(ASR3(4, 1, -2)));

  Test('v[1] ≔ 10', 10);
  Test('v', ASO(ASR3(10, 1, -2)));

  Test('v[2] ≔ 20', 20);
  Test('v', ASO(ASR3(10, 20, -2)));

  Test('v[3] ≔ 30', 30);
  Test('v', ASO(ASR3(10, 20, 30)));

  Test('v[4] ≔ 40', failure, 'Index 4 out of bounds.');
  Test('v', ASO(ASR3(10, 20, 30)));

  Test('v[-1] ≔ 100', 100);
  Test('v', ASO(ASR3(10, 20, 100)));

  Test('v[3] ≔ 30', 30);
  Test('v', ASO(ASR3(10, 20, 30)));

  Test('protect(v)', success);
  Test('v[3] ≔ 300', failure, 'Cannot modify or delete protected variable "v".');
  Test('v', ASO(ASR3(10, 20, 30)));

  Test('unprotect(v)', success);
  Test('v[3] ≔ 300', 300);
  Test('v', ASO(ASR3(10, 20, 300)));

  Test('v[1] ≔ "horse"', failure, 'A real number was expected, but an object of type "string" given, when setting a component in a real vector.');
  Test('v[2] ≔ true', failure, 'A real number was expected, but an object of type "boolean" given, when setting a component in a real vector.');
  Test('v[3] ≔ ZeroVector(2)', failure, 'A real number was expected, but an object of type "real vector" given, when setting a component in a real vector.');
  Test('v', ASO(ASR3(10, 20, 300)));

  Test('v[1] ≔ i', failure, 'A real number was expected, but an object of type "complex number" given, when setting a component in a real vector.');
  Test('v', ASO(ASR3(10, 20, 300)));

  Test('v[3] ≔ 30.5', 30.5);
  Test('v', ASO(ASR3(10, 20, 30.5)));

  Test('v[3] ≔ 3/5', Rat(3, 5));
  Test('v', ASO(ASR3(10, 20, 0.6)));

  // Complex vectors

  Test('v ≔ ❨4, i, −2❩', ASO(ASC3(4, ImaginaryUnit, -2)));
  Test('v', ASO(ASC3(4, ImaginaryUnit, -2)));

  Test('v[1]', TASC(4.0));
  Test('v[2]', ImaginaryUnit);
  Test('v[3]', TASC(-2.0));
  Test('v[4]', failure, 'Index 4 out of bounds.');

  Test('v[-1]', TASC(-2.0));
  Test('v[-2]', ImaginaryUnit);
  Test('v[-3]', TASC(4.0));
  Test('v[-4]', failure, 'Index -4 out of bounds.');

  Test('v[0]', failure, 'Index 0 out of bounds.');

  Test('v', ASO(ASC3(4, ImaginaryUnit, -2)));

  Test('v[1] ≔ 10', 10);
  Test('v', ASO(ASC3(10, ImaginaryUnit, -2)));

  Test('v[2] ≔ 20', 20);
  Test('v', ASO(ASC3(10, 20, -2)));

  Test('v[3] ≔ 30', 30);
  Test('v', ASO(ASC3(10, 20, 30)));

  Test('v[4] ≔ 40', failure, 'Index 4 out of bounds.');
  Test('v', ASO(ASC3(10, 20, 30)));

  Test('v[-1] ≔ 100', 100);
  Test('v', ASO(ASC3(10, 20, 100)));

  Test('v[3] ≔ 30', 30);
  Test('v', ASO(ASC3(10, 20, 30)));

  Test('protect(v)', success);
  Test('v[3] ≔ 300', failure, 'Cannot modify or delete protected variable "v".');
  Test('v', ASO(ASC3(10, 20, 30)));

  Test('unprotect(v)', success);
  Test('v[3] ≔ 300', 300);
  Test('v', ASO(ASC3(10, 20, 300)));

  Test('v[1] ≔ "horse"', failure, 'A real or complex number was expected, but an object of type "string" given, when setting a component in a complex vector.');
  Test('v[2] ≔ true', failure, 'A real or complex number was expected, but an object of type "boolean" given, when setting a component in a complex vector.');
  Test('v[3] ≔ ZeroVector(2)', failure, 'A real or complex number was expected, but an object of type "real vector" given, when setting a component in a complex vector.');
  Test('v', ASO(ASC3(10, 20, 300)));
  Test('v[1] ≔ i', ImaginaryUnit);
  Test('v', ASO(ASC3(ImaginaryUnit, 20, 300)));

  Test('v[3] ≔ 30.5', 30.5);
  Test('v', ASO(ASC3(ImaginaryUnit, 20, 30.5)));

  Test('v[3] ≔ 3/5', Rat(3, 5));
  Test('v', ASO(ASC3(ImaginaryUnit, 20, 0.6)));

  Test('v[3] ≔ 2+3⋅i', 2 + 3*ImaginaryUnit);
  Test('v', ASO(ASC3(ImaginaryUnit, 20, 2 + 3*ImaginaryUnit)));

  // Real matrices

  Test('A ≔ ❨❨10, 20, 30❩, ❨40, 50, 60❩❩', 3, [10, 20, 30, 40, 50, 60]);
  Test('A', 3, [10, 20, 30, 40, 50, 60]);

  Test('A[1]', 10.0);
  Test('A[2]', 20.0);
  Test('A[3]', 30.0);
  Test('A[4]', 40.0);
  Test('A[5]', 50.0);
  Test('A[6]', 60.0);
  Test('A[7]', failure, 'Index 7 out of bounds.');

  Test('A[-1]', 60.0);
  Test('A[-2]', 50.0);
  Test('A[-3]', 40.0);
  Test('A[-4]', 30.0);
  Test('A[-5]', 20.0);
  Test('A[-6]', 10.0);
  Test('A[-7]', failure, 'Index -7 out of bounds.');

  Test('A[0]', failure, 'Index 0 out of bounds.');

  Test('A[1, 1]', 10.0);
  Test('A[1, 2]', 20.0);
  Test('A[1, 3]', 30.0);
  Test('A[2, 1]', 40.0);
  Test('A[2, 2]', 50.0);
  Test('A[2, 3]', 60.0);

  Test('A[1, -3]', 10.0);
  Test('A[-2, -2]', 20.0);
  Test('A[1, -1]', 30.0);
  Test('A[-1, 1]', 40.0);
  Test('A[-1, 2]', 50.0);
  Test('A[-1, -1]', 60.0);

  Test('A[1, 4]', failure, 'Index (1, 4) out of bounds.');
  Test('A[1, 40]', failure, 'Index (1, 40) out of bounds.');
  Test('A[1, 0]', failure, 'Index (1, 0) out of bounds.');
  Test('A[1, -4]', failure, 'Index (1, -4) out of bounds.');
  Test('A[1, -40]', failure, 'Index (1, -40) out of bounds.');
  Test('A[-1, 4]', failure, 'Index (-1, 4) out of bounds.');
  Test('A[0, 4]', failure, 'Index (0, 4) out of bounds.');
  Test('A[0, 2]', failure, 'Index (0, 2) out of bounds.');
  Test('A[-8, 1]', failure, 'Index (-8, 1) out of bounds.');
  Test('A[-2, 0]', failure, 'Index (-2, 0) out of bounds.');
  Test('A[0, 0]', failure, 'Index (0, 0) out of bounds.');

  Test('A[''(1, 1)]', 10.0);
  Test('A[''(1, 2)]', 20.0);
  Test('A[''(1, 3)]', 30.0);
  Test('A[''(2, 1)]', 40.0);
  Test('A[''(2, 2)]', 50.0);
  Test('A[''(2, 3)]', 60.0);

  Test('A[''(1, -3)]', 10.0);
  Test('A[''(-2, -2)]', 20.0);
  Test('A[''(1, -1)]', 30.0);
  Test('A[''(-1, 1)]', 40.0);
  Test('A[''(-1, 2)]', 50.0);
  Test('A[''(-1, -1)]', 60.0);

  Test('A[''(1, 4)]', failure, 'Index (1, 4) out of bounds.');
  Test('A[''(1, 40)]', failure, 'Index (1, 40) out of bounds.');
  Test('A[''(1, 0)]', failure, 'Index (1, 0) out of bounds.');
  Test('A[''(1, -4)]', failure, 'Index (1, -4) out of bounds.');
  Test('A[''(1, -40)]', failure, 'Index (1, -40) out of bounds.');
  Test('A[''(-1, 4)]', failure, 'Index (-1, 4) out of bounds.');
  Test('A[''(0, 4)]', failure, 'Index (0, 4) out of bounds.');
  Test('A[''(0, 2)]', failure, 'Index (0, 2) out of bounds.');
  Test('A[''(-8, 1)]', failure, 'Index (-8, 1) out of bounds.');
  Test('A[''(-2, 0)]', failure, 'Index (-2, 0) out of bounds.');
  Test('A[''(0, 0)]', failure, 'Index (0, 0) out of bounds.');

  Test('A[1] ≔ 100', 100);
  Test('A', 3, [100, 20, 30, 40, 50, 60]);
  Test('A[1, 3] ≔ 300', 300);
  Test('A', 3, [100, 20, 300, 40, 50, 60]);
  Test('A[−1, −1] ≔ 600', 600);
  Test('A', 3, [100, 20, 300, 40, 50, 600]);
  Test('A[−2] ≔ 500', 500);
  Test('A', 3, [100, 20, 300, 40, 500, 600]);
  Test('A[−1, 1] ≔ 400', 400);
  Test('A', 3, [100, 20, 300, 400, 500, 600]);
  Test('A[1, −2] ≔ 200', 200);
  Test('A', 3, [100, 200, 300, 400, 500, 600]);

  Test('A[''(-1, 1)] ≔ 7', 7);
  Test('A', 3, [100, 200, 300, 7, 500, 600]);
  Test('A[''(-1, 2)] ≔ 8', 8);
  Test('A', 3, [100, 200, 300, 7, 8, 600]);
  Test('A[''(-1, 3)] ≔ 9', 9);
  Test('A', 3, [100, 200, 300, 7, 8, 9]);
  Test('A[''(1, 1)] ≔ 1', 1);
  Test('A', 3, [1, 200, 300, 7, 8, 9]);
  Test('A[''(1, 2)] ≔ 2', 2);
  Test('A', 3, [1, 2, 300, 7, 8, 9]);
  Test('A[''(1, -1)] ≔ 3', 3);
  Test('A', 3, [1, 2, 3, 7, 8, 9]);

  Test('A[8] ≔ 14', failure, 'Index 8 out of bounds.');
  Test('A[-8] ≔ -4', failure, 'Index -8 out of bounds.');
  Test('A[1, 4] ≔ 4', failure, 'Index (1, 4) out of bounds.');
  Test('A[1, -4] ≔ 4', failure, 'Index (1, -4) out of bounds.');
  Test('A[0, 1] ≔ 4', failure, 'Index (0, 1) out of bounds.');
  Test('A[''(3, 1)] ≔ 4', failure, 'Index (3, 1) out of bounds.');
  Test('A[''(-4, 2)] ≔ 4', failure, 'Index (-4, 2) out of bounds.');
  Test('A[0, 0] ≔ 4', failure, 'Index (0, 0) out of bounds.');

  Test('A[1] ≔ i', failure, 'A real number was expected, but an object of type "complex number" given, when setting an entry in a real matrix.');
  Test('A[2] ≔ "cat"', failure, 'A real number was expected, but an object of type "string" given, when setting an entry in a real matrix.');
  Test('A[1, 3] ≔ true', failure, 'A real number was expected, but an object of type "boolean" given, when setting an entry in a real matrix.');
  Test('A[-1] ≔ color("red")', failure, 'A real number was expected, but an object of type "color" given, when setting an entry in a real matrix.');
  Test('A[-1, -1] ≔ i', failure, 'A real number was expected, but an object of type "complex number" given, when setting an entry in a real matrix.');
  Test('A[''(1, -1)] ≔ i', failure, 'A real number was expected, but an object of type "complex number" given, when setting an entry in a real matrix.');

  Test('A', 3, [1, 2, 3, 7, 8, 9]);
  Test('protect(A)', success);
  Test('A[1, 1] ≔ 10', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [1, 2, 3, 7, 8, 9]);
  Test('unprotect(A)', success);
  Test('A[1, 1] ≔ 10', 10);
  Test('A', 3, [10, 2, 3, 7, 8, 9]);

  Test('protect(A)', success);
  Test('A[-1, 1] ≔ 70', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [10, 2, 3, 7, 8, 9]);
  Test('unprotect(A)', success);
  Test('A[-1, 1] ≔ 70', 70);
  Test('A', 3, [10, 2, 3, 70, 8, 9]);

  Test('protect(A)', success);
  Test('A[2] ≔ 20', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [10, 2, 3, 70, 8, 9]);
  Test('unprotect(A)', success);
  Test('A[2] ≔ 20', 20);
  Test('A', 3, [10, 20, 3, 70, 8, 9]);

  Test('protect(A)', success);
  Test('A[-1] ≔ 90', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [10, 20, 3, 70, 8, 9]);
  Test('unprotect(A)', success);
  Test('A[-1] ≔ 90', 90);
  Test('A', 3, [10, 20, 3, 70, 8, 90]);

  Test('delete(A)', success);
  Test('A[1, -40]', failure, 'Unknown identifier "A".');

  // Complex matrices

  Test('A ≔ ❨❨i, 20, 30❩, ❨40, 50, 60❩❩', 3, [ImaginaryUnit, 20, 30, 40, 50, 60]);
  Test('A', 3, [ImaginaryUnit, 20, 30, 40, 50, 60]);

  Test('A[1]', ImaginaryUnit);
  Test('A[2]', TASC(20.0));
  Test('A[3]', TASC(30.0));
  Test('A[4]', TASC(40.0));
  Test('A[5]', TASC(50.0));
  Test('A[6]', TASC(60.0));
  Test('A[7]', failure, 'Index 7 out of bounds.');

  Test('A[-1]', TASC(60.0));
  Test('A[-2]', TASC(50.0));
  Test('A[-3]', TASC(40.0));
  Test('A[-4]', TASC(30.0));
  Test('A[-5]', TASC(20.0));
  Test('A[-6]', ImaginaryUnit);
  Test('A[-7]', failure, 'Index -7 out of bounds.');

  Test('A[0]', failure, 'Index 0 out of bounds.');

  Test('A[1, 1]', ImaginaryUnit);
  Test('A[1, 2]', TASC(20.0));
  Test('A[1, 3]', TASC(30.0));
  Test('A[2, 1]', TASC(40.0));
  Test('A[2, 2]', TASC(50.0));
  Test('A[2, 3]', TASC(60.0));

  Test('A[1, -3]', ImaginaryUnit);
  Test('A[-2, -2]', TASC(20.0));
  Test('A[1, -1]', TASC(30.0));
  Test('A[-1, 1]', TASC(40.0));
  Test('A[-1, 2]', TASC(50.0));
  Test('A[-1, -1]', TASC(60.0));

  Test('A[1, 4]', failure, 'Index (1, 4) out of bounds.');
  Test('A[1, 40]', failure, 'Index (1, 40) out of bounds.');
  Test('A[1, 0]', failure, 'Index (1, 0) out of bounds.');
  Test('A[1, -4]', failure, 'Index (1, -4) out of bounds.');
  Test('A[1, -40]', failure, 'Index (1, -40) out of bounds.');
  Test('A[-1, 4]', failure, 'Index (-1, 4) out of bounds.');
  Test('A[0, 4]', failure, 'Index (0, 4) out of bounds.');
  Test('A[0, 2]', failure, 'Index (0, 2) out of bounds.');
  Test('A[-8, 1]', failure, 'Index (-8, 1) out of bounds.');
  Test('A[-2, 0]', failure, 'Index (-2, 0) out of bounds.');
  Test('A[0, 0]', failure, 'Index (0, 0) out of bounds.');

  Test('A[''(1, 1)]', ImaginaryUnit);
  Test('A[''(1, 2)]', TASC(20.0));
  Test('A[''(1, 3)]', TASC(30.0));
  Test('A[''(2, 1)]', TASC(40.0));
  Test('A[''(2, 2)]', TASC(50.0));
  Test('A[''(2, 3)]', TASC(60.0));

  Test('A[''(1, -3)]', ImaginaryUnit);
  Test('A[''(-2, -2)]', TASC(20.0));
  Test('A[''(1, -1)]', TASC(30.0));
  Test('A[''(-1, 1)]', TASC(40.0));
  Test('A[''(-1, 2)]', TASC(50.0));
  Test('A[''(-1, -1)]', TASC(60.0));

  Test('A[''(1, 4)]', failure, 'Index (1, 4) out of bounds.');
  Test('A[''(1, 40)]', failure, 'Index (1, 40) out of bounds.');
  Test('A[''(1, 0)]', failure, 'Index (1, 0) out of bounds.');
  Test('A[''(1, -4)]', failure, 'Index (1, -4) out of bounds.');
  Test('A[''(1, -40)]', failure, 'Index (1, -40) out of bounds.');
  Test('A[''(-1, 4)]', failure, 'Index (-1, 4) out of bounds.');
  Test('A[''(0, 4)]', failure, 'Index (0, 4) out of bounds.');
  Test('A[''(0, 2)]', failure, 'Index (0, 2) out of bounds.');
  Test('A[''(-8, 1)]', failure, 'Index (-8, 1) out of bounds.');
  Test('A[''(-2, 0)]', failure, 'Index (-2, 0) out of bounds.');
  Test('A[''(0, 0)]', failure, 'Index (0, 0) out of bounds.');

  Test('A[1] ≔ 100', 100);
  Test('A', 3, [100+0*ImaginaryUnit, 20, 30, 40, 50, 60]);
  Test('A[1, 3] ≔ 300', 300);
  Test('A', 3, [100+0*ImaginaryUnit, 20, 300, 40, 50, 60]);
  Test('A[−1, −1] ≔ 600', 600);
  Test('A', 3, [100+0*ImaginaryUnit, 20, 300, 40, 50, 600]);
  Test('A[−2] ≔ 500', 500);
  Test('A', 3, [100+0*ImaginaryUnit, 20, 300, 40, 500, 600]);
  Test('A[−1, 1] ≔ 400', 400);
  Test('A', 3, [100+0*ImaginaryUnit, 20, 300, 400, 500, 600]);
  Test('A[1, −2] ≔ 200', 200);
  Test('A', 3, [100+0*ImaginaryUnit, 200, 300, 400, 500, 600]);

  Test('A[''(-1, 1)] ≔ 7', 7);
  Test('A', 3, [100+0*ImaginaryUnit, 200, 300, 7, 500, 600]);
  Test('A[''(-1, 2)] ≔ 8', 8);
  Test('A', 3, [100+0*ImaginaryUnit, 200, 300, 7, 8, 600]);
  Test('A[''(-1, 3)] ≔ 9', 9);
  Test('A', 3, [100+0*ImaginaryUnit, 200, 300, 7, 8, 9]);
  Test('A[''(1, 1)] ≔ 1', 1);
  Test('A', 3, [1+0*ImaginaryUnit, 200, 300, 7, 8, 9]);
  Test('A[''(1, 2)] ≔ 2', 2);
  Test('A', 3, [1+0*ImaginaryUnit, 2, 300, 7, 8, 9]);
  Test('A[''(1, -1)] ≔ 3', 3);
  Test('A', 3, [1+0*ImaginaryUnit, 2, 3, 7, 8, 9]);

  Test('A[8] ≔ 14', failure, 'Index 8 out of bounds.');
  Test('A[-8] ≔ -4', failure, 'Index -8 out of bounds.');
  Test('A[1, 4] ≔ 4', failure, 'Index (1, 4) out of bounds.');
  Test('A[1, -4] ≔ 4', failure, 'Index (1, -4) out of bounds.');
  Test('A[0, 1] ≔ 4', failure, 'Index (0, 1) out of bounds.');
  Test('A[''(3, 1)] ≔ 4', failure, 'Index (3, 1) out of bounds.');
  Test('A[''(-4, 2)] ≔ 4', failure, 'Index (-4, 2) out of bounds.');
  Test('A[0, 0] ≔ 4', failure, 'Index (0, 0) out of bounds.');

  Test('A[1] ≔ i', ImaginaryUnit);
  Test('A', 3, [ImaginaryUnit, 2, 3, 7, 8, 9]);
  Test('A[2] ≔ "cat"', failure, 'A real or complex number was expected, but an object of type "string" given, when setting an entry in a complex matrix.');
  Test('A[1, 3] ≔ true', failure, 'A real or complex number was expected, but an object of type "boolean" given, when setting an entry in a complex matrix.');
  Test('A[-1] ≔ color("red")', failure, 'A real or complex number was expected, but an object of type "color" given, when setting an entry in a complex matrix.');
  Test('A[-1, -1] ≔ "cat"', failure, 'A real or complex number was expected, but an object of type "string" given, when setting an entry in a complex matrix.');
  Test('A[''(1, -1)] ≔ "cat"', failure, 'A real or complex number was expected, but an object of type "string" given, when setting an entry in a complex matrix.');

  Test('A', 3, [ImaginaryUnit, 2, 3, 7, 8, 9]);
  Test('protect(A)', success);
  Test('A[1, 1] ≔ 10', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [ImaginaryUnit, 2, 3, 7, 8, 9]);
  Test('unprotect(A)', success);
  Test('A[1, 1] ≔ 10', 10);
  Test('A', 3, [10+0*ImaginaryUnit, 2, 3, 7, 8, 9]);

  Test('protect(A)', success);
  Test('A[-1, 1] ≔ 70', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [10+0*ImaginaryUnit, 2, 3, 7, 8, 9]);
  Test('unprotect(A)', success);
  Test('A[-1, 1] ≔ 70', 70);
  Test('A', 3, [10+0*ImaginaryUnit, 2, 3, 70, 8, 9]);

  Test('protect(A)', success);
  Test('A[2] ≔ 20', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [10+0*ImaginaryUnit, 2, 3, 70, 8, 9]);
  Test('unprotect(A)', success);
  Test('A[2] ≔ 20', 20);
  Test('A', 3, [10+0*ImaginaryUnit, 20, 3, 70, 8, 9]);

  Test('protect(A)', success);
  Test('A[-1] ≔ 90⋅i', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [10+0*ImaginaryUnit, 20, 3, 70, 8, 9]);
  Test('unprotect(A)', success);
  Test('A[-1] ≔ 90⋅i', 90*ImaginaryUnit);
  Test('A', 3, [10+0*ImaginaryUnit, 20, 3, 70, 8, 90*ImaginaryUnit]);

  Test('delete(A)', success);
  Test('A[1, -40]', failure, 'Unknown identifier "A".');

  // Vectors and matrices in structures and lists

  Test('S ≔ struct("a": ❨1, 2, 3❩, "b": ❨10, 20, 30❩)',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR3(1, 2, 3)), ASO(ASR3(10, 20, 30))]
    )
  );

  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR3(1, 2, 3)), ASO(ASR3(10, 20, 30))]
    )
  );

  Test('S.a', [1, 2, 3]);
  Test('S.b', [10, 20, 30]);

  Test('S.a[1]', 1.0);
  Test('S.a[2]', 2.0);
  Test('S.a[3]', 3.0);
  Test('S.a[4]', failure, 'Index 4 out of bounds.');

  Test('S.b[1]', 10.0);
  Test('S.b[2]', 20.0);
  Test('S.b[3]', 30.0);

  Test('S.b[-1]', 30.0);
  Test('S.b[-2]', 20.0);
  Test('S.b[-3]', 10.0);
  Test('S.b[-4]', failure, 'Index -4 out of bounds.');

  Test('S.a[1] ≔ 100', 100);
  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR3(100, 2, 3)), ASO(ASR3(10, 20, 30))]
    )
  );

  Test('S.a[−1] ≔ 300', 300);
  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR3(100, 2, 300)), ASO(ASR3(10, 20, 30))]
    )
  );

  Test('protect(S)', success);
  Test('S.a[2] ≔ 200', failure, 'Cannot modify or delete protected variable "S".');
  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR3(100, 2, 300)), ASO(ASR3(10, 20, 30))]
    )
  );

  Test('protect(S)', success); // idempotency
  Test('S.a[2] ≔ 200', failure, 'Cannot modify or delete protected variable "S".');
  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR3(100, 2, 300)), ASO(ASR3(10, 20, 30))]
    )
  );

  Test('unprotect(S)', success);
  Test('S.a[2] ≔ 200', 200);
  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR3(100, 200, 300)), ASO(ASR3(10, 20, 30))]
    )
  );

  Test('S.a ≔ ❨1, 0❩', ASO(ASR2(1, 0)));
  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR2(1, 0)), ASO(ASR3(10, 20, 30))]
    )
  );

  Test('protect(S)', success);
  Test('S.b ≔ ❨0, 1❩', failure, 'Cannot modify or delete protected variable "S".');
  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR2(1, 0)), ASO(ASR3(10, 20, 30))]
    )
  );

  Test('unprotect(S)', success);
  Test('S.b ≔ ❨0, 1❩', ASO(ASR2(0, 1)));
  Test('S',
    TAlgosimStructure.CreateWithValue(
      ['a', 'b'],
      [ASO(ASR2(1, 0)), ASO(ASR2(0, 1))]
    )
  );

  Test('delete(S)', success);
  Test('S.b ≔ ❨1, 1❩', failure, 'Unknown identifier "S".');

  Test('L ≔ ''(IdentityMatrix(2), 2⋅IdentityMatrix(2))',
    TAlgosimArray.CreateWithValue(
      [
        ASO(IdentityMatrix(2)),
        ASO(2*IdentityMatrix(2))
      ]
    )
  );

  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(IdentityMatrix(2)),
        ASO(2*IdentityMatrix(2))
      ]
    )
  );

  Test('L[1]', 2, [1, 0, 0, 1]);
  Test('L[2]', 2, [2, 0, 0, 2]);

  Test('L[1][1]', 1.0);
  Test('L[1][2]', 0.0);
  Test('L[1][3]', 0.0);
  Test('L[1][4]', 1.0);
  Test('L[1][5]', failure, 'Index 5 out of bounds.');

  Test('L[2][-1]', 2.0);
  Test('L[2][-2]', 0.0);
  Test('L[2][-3]', 0.0);
  Test('L[2][-4]', 2.0);
  Test('L[2][-5]', failure, 'Index -5 out of bounds.');

  Test('L[2][1, 1]', 2.0);
  Test('L[2][1, 2]', 0.0);
  Test('L[2][2, 1]', 0.0);
  Test('L[2][2, 2]', 2.0);
  Test('L[2][2, 3]', failure, 'Index (2, 3) out of bounds.');

  Test('L[1][-1, -1]', 1.0);

  Test('L[1][1] ≔ 10', 10);
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 1], 2)),
        ASO(TRealMatrix.Create([2, 0, 0, 2], 2))
      ]
    )
  );

  Test('L[1][-1] ≔ 100', 100);
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 0, 0, 2], 2))
      ]
    )
  );

  Test('L[2][1, 2] ≔ 3', 3);
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 3, 0, 2], 2))
      ]
    )
  );

  Test('L[2][-1, 2] ≔ 4', 4);
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 3, 0, 4], 2))
      ]
    )
  );

  Test('L[2][-1, -2] ≔ 5', 5);
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 3, 5, 4], 2))
      ]
    )
  );

  Test('L[2][1, 2] ≔ 50', 50);
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 50, 5, 4], 2))
      ]
    )
  );

  Test('L[2][1, 3] ≔ 50', failure, 'Index (1, 3) out of bounds.');
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 50, 5, 4], 2))
      ]
    )
  );

  Test('L[2][-1, -3] ≔ 50', failure, 'Index (-1, -3) out of bounds.');
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 50, 5, 4], 2))
      ]
    )
  );

  Test('protect(L)', success);
  Test('L[2][1, 2] ≔ 3', failure, 'Cannot modify or delete protected variable "L".');
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 50, 5, 4], 2))
      ]
    )
  );

  Test('unprotect(L)', success);
  Test('L[2][1, 2] ≔ 3', 3);
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 3, 5, 4], 2))
      ]
    )
  );

  Test('unprotect(L)', success); // idempotency
  Test('L[2][''(1, 2)] ≔ 3.1', 3.1);
  Test('L',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TRealMatrix.Create([10, 0, 0, 100], 2)),
        ASO(TRealMatrix.Create([2, 3.1, 5, 4], 2))
      ]
    )
  );

  Test('delete(L)', success);
  Test('L[2][1, 2] ≔ 50', failure, 'Unknown identifier "L".');





  //
  //
  //  CHAPTER 4
  //  Basic mathematical functions.
  //
  //

  Chapter('Basic mathematical functions');

  //
  // Elementary functions of a real variable
  //

  // Trigonometric functions

  Eps; Test('sin(0)', 0.0);
  Eps; Test('sin(π/2)', 1.0);
  Eps; Test('sin(π)', 0.0);
  Eps; Test('sin(-π/2)', -1.0);
  Eps; Test('sin(π/4)', 0.707106781186547524400844362104849039284835937688474036588);
  Eps; Test('sin(π/3)', 0.866025403784438646763723170752936183471402626905190314027);
  Eps; Test('sin(π/6)', 1/2);

  Eps; Test('sin(1)', 0.841470984807896506652502321630298999622563060798371065672);
  Eps; Test('sin(123)', -0.459903490689591251292435715293231810808580607381042);
  Eps; Test('sin(0.001)', 0.0009999998333333416666664682539710097001513147348086);
  Eps; Test('sin(-1)', -0.841470984807896506652502321630298999622563060798371065672);
  Eps; Test('sin(-123)', 0.459903490689591251292435715293231810808580607381042);
  Eps; Test('sin(-0.001)', -0.0009999998333333416666664682539710097001513147348086);

  Eps(1E-12); Test('sin(123456)', -0.74028345388665753672466063491709017804265556263134076537);
  Eps(1E-12); Test('sin(-123456)', 0.74028345388665753672466063491709017804265556263134076537);

  Test('sin(1, 2)', failure, 'Too many arguments.');

  Eps; Test('cos(0)', 1.0);
  Eps; Test('cos(π/2)', 0.0);
  Eps; Test('cos(π)', -1.0);
  Eps; Test('cos(-π/2)', 0.0);
  Eps; Test('cos(π/4)', 0.707106781186547524400844362104849039284835937688474036588);
  Eps; Test('cos(π/3)', 1/2);
  Eps; Test('cos(π/6)', 0.866025403784438646763723170752936183471402626905190314027);

  Eps; Test('cos(1)', 0.540302305868139717400936607442976603732310420617922227670);
  Eps; Test('cos(123)', -0.887968906691855428978322569442621150811865560981614);
  Eps; Test('cos(0.001)', 0.9999995000000416666652777778025793648037918892128961);
  Eps; Test('cos(-1)', 0.540302305868139717400936607442976603732310420617922227670);
  Eps; Test('cos(-123)', -0.887968906691855428978322569442621150811865560981614);
  Eps; Test('cos(-0.001)', 0.9999995000000416666652777778025793648037918892128961);

  Eps(1E-12); Test('cos(123456)', -0.67229488165658451960808772768357726667866595466767514385);
  Eps(1E-12); Test('cos(-123456)', -0.67229488165658451960808772768357726667866595466767514385);

  Test('cos(1, 2)', failure, 'Too many arguments.');

  Eps; Test('tan(0)', 0.0);
  Eps; Test('tan(π)', 0.0);
  Eps; Test('tan(π/4)', 1.0);
  Eps; Test('tan(π/3)', 1.732050807568877293527446341505872366942805253810380628055);
  Eps; Test('tan(π/6)', 1/1.732050807568877293527446341505872366942805253810380628055);

  Eps; Test('tan(1)', 1.557407724654902230506974807458360173087250772381520038383);
  Eps; Test('tan(123)', 0.5179274715856551831319240756392856882012784222828477);
  Eps; Test('tan(0.001)', 0.0010000003333334666667206349425044180343149597741933);
  Eps; Test('tan(-1)', -1.557407724654902230506974807458360173087250772381520038383);
  Eps; Test('tan(-123)', -0.5179274715856551831319240756392856882012784222828477);
  Eps; Test('tan(-0.001)', -0.0010000003333334666667206349425044180343149597741933);

  Eps(1E-12); Test('tan(123456)', 1.1011290939216198190353988040045654938483143691480705);
  Eps(1E-12); Test('tan(-123456)', -1.1011290939216198190353988040045654938483143691480705);

  Test('tan(1, 2)', failure, 'Too many arguments.');

  Eps; Test('cot(π/2)', 0.0);
  Eps; Test('cot(-π/2)', 0.0);
  Eps; Test('cot(π/4)', 1.0);
  Eps; Test('cot(π/3)', 1/1.732050807568877293527446341505872366942805253810380628055);
  Eps; Test('cot(π/6)', 1.732050807568877293527446341505872366942805253810380628055);

  Eps; Test('cot(1)', 1/1.557407724654902230506974807458360173087250772381520038383);
  Eps; Test('cot(123)', 1/0.5179274715856551831319240756392856882012784222828477);
  Eps; Test('cot(0.001)', 1/0.0010000003333334666667206349425044180343149597741933);
  Eps; Test('cot(-1)', -1/1.557407724654902230506974807458360173087250772381520038383);
  Eps; Test('cot(-123)', -1/0.5179274715856551831319240756392856882012784222828477);
  Eps; Test('cot(-0.001)', -1/0.0010000003333334666667206349425044180343149597741933);

  Eps(1E-12); Test('cot(123456)', 1/1.1011290939216198190353988040045654938483143691480705);
  Eps(1E-12); Test('cot(-123456)', -1/1.1011290939216198190353988040045654938483143691480705);

  Test('cot(1, 2)', failure, 'Too many arguments.');

  Eps; Test('sec(0)', 1.0);
  Eps; Test('sec(π)', -1.0);
  Eps; Test('sec(π/4)', 1/0.707106781186547524400844362104849039284835937688474036588);
  Eps; Test('sec(π/3)', 2.0);
  Eps; Test('sec(π/6)', 1/0.866025403784438646763723170752936183471402626905190314027);

  Eps; Test('sec(1)', 1/0.540302305868139717400936607442976603732310420617922227670);
  Eps; Test('sec(123)', -1/0.887968906691855428978322569442621150811865560981614);
  Eps; Test('sec(0.001)', 1/0.9999995000000416666652777778025793648037918892128961);
  Eps; Test('sec(-1)', 1/0.540302305868139717400936607442976603732310420617922227670);
  Eps; Test('sec(-123)', -1/0.887968906691855428978322569442621150811865560981614);
  Eps; Test('sec(-0.001)', 1/0.9999995000000416666652777778025793648037918892128961);

  Eps(1E-12); Test('sec(123456)', -1/0.67229488165658451960808772768357726667866595466767514385);
  Eps(1E-12); Test('sec(-123456)', -1/0.67229488165658451960808772768357726667866595466767514385);

  Test('sec(1, 2)', failure, 'Too many arguments.');

  Eps; Test('csc(π/2)', 1.0);
  Eps; Test('csc(-π/2)', -1.0);
  Eps; Test('csc(π/4)', 1/0.707106781186547524400844362104849039284835937688474036588);
  Eps; Test('csc(π/3)', 1/0.866025403784438646763723170752936183471402626905190314027);
  Eps; Test('csc(π/6)', 2.0);

  Eps; Test('csc(1)', 1/0.841470984807896506652502321630298999622563060798371065672);
  Eps; Test('csc(123)', -1/0.459903490689591251292435715293231810808580607381042);
  Eps; Test('csc(0.001)', 1/0.0009999998333333416666664682539710097001513147348086);
  Eps; Test('csc(-1)', -1/0.841470984807896506652502321630298999622563060798371065672);
  Eps; Test('csc(-123)', 1/0.459903490689591251292435715293231810808580607381042);
  Eps; Test('csc(-0.001)', -1/0.0009999998333333416666664682539710097001513147348086);

  Eps(1E-12); Test('csc(123456)', -1/0.74028345388665753672466063491709017804265556263134076537);
  Eps(1E-12); Test('csc(-123456)', 1/0.74028345388665753672466063491709017804265556263134076537);

  Test('csc(1, 2)', failure, 'Too many arguments.');

  Eps; Test('sin(cos(tan(1.23456789)))', -0.819740325959989183836632561433919612688949870316605);
  Eps; Test('sin(e)', 0.410781290502908695476009492018360591888306970393415345304);
  Eps; Test('sin(e) + sin(-e)', 0.0);

  Eps; Test('sin(cos(tan(1) + 2) + 3)', 0.870580073950121068697344985465662031887377093858778495887);
  Eps; Test('sin(cos(-tan(-1) + 2) -- 3)', 0.870580073950121068697344985465662031887377093858778495887);
  Eps; Test('sin(cos(-tan(1) - 2) + 3)', 0.870580073950121068697344985465662031887377093858778495887);
  Eps; Test('sin(cos(tan(-1) - 2) + 3)', 0.870580073950121068697344985465662031887377093858778495887);
  Eps; Test('sin(-cos(tan(1) + 2) - 3)', -0.870580073950121068697344985465662031887377093858778495887);
  Eps; Test('sin(-cos(-tan(1) - 2) - 3)', -0.870580073950121068697344985465662031887377093858778495887);
  Eps; Test('sin(-cos(tan(-1) - 2) - 3)', -0.870580073950121068697344985465662031887377093858778495887);

  // Inverse trigonometric functions

  Eps; Test('arcsin(−1)', -Pi/2);
  Eps; Test('arcsin(−√3/2)', -Pi/3);
  Eps; Test('arcsin(−1/√2)', -Pi/4);
  Eps; Test('arcsin(−1/2)', -Pi/6);
  Eps; Test('arcsin(0)', 0.0);
  Eps; Test('arcsin(1/2)', Pi/6);
  Eps; Test('arcsin(1/√2)', Pi/4);
  Eps; Test('arcsin(√3/2)', Pi/3);
  Eps; Test('arcsin(1)', Pi/2);

  Test('arcsin(1, 2)', failure, 'Too many arguments.');

  Eps; Test('arccos(−1)', Pi);
  Eps; Test('arccos(−√3/2)', 5*Pi/6);
  Eps; Test('arccos(−1/√2)', 3*Pi/4);
  Eps; Test('arccos(−1/2)', 2*Pi/3);
  Eps; Test('arccos(0)', Pi/2);
  Eps; Test('arccos(1/2)', Pi/3);
  Eps; Test('arccos(1/√2)', Pi/4);
  Eps; Test('arccos(√3/2)', Pi/6);
  Eps; Test('arccos(1)', 0.0);

  Test('arccos(1, 2)', failure, 'Too many arguments.');

  Eps; Test('arctan(−1E30)', -Pi/2);
  Eps; Test('arctan(−√3)', -Pi/3);
  Eps; Test('arctan(−1)', -Pi/4);
  Eps; Test('arctan(−1/√3)', -Pi/6);
  Eps; Test('arctan(0)', 0.0);
  Eps; Test('arctan(1/√3)', Pi/6);
  Eps; Test('arctan(1)', Pi/4);
  Eps; Test('arctan(√3)', Pi/3);
  Eps; Test('arctan(1E30)', Pi/2);

  Test('arctan(1, 2)', failure, 'Too many arguments.');

  Eps; Test('arccot(−1E30)', 0.0);
  Eps; Test('arccot(−√3)', -Pi/6);
  Eps; Test('arccot(−1)', -Pi/4);
  Eps; Test('arccot(−1/√3)', -Pi/3);
  Eps(1E-10); Test('arccot(-1E-12)', -Pi/2);
  Eps(1E-10); Test('arccot(1E-12)', Pi/2);
  Eps; Test('arccot(1/√3)', Pi/3);
  Eps; Test('arccot(1)', Pi/4);
  Eps; Test('arccot(√3)', Pi/6);
  Eps; Test('arccot(1E30)', 0.0);

  Test('arccot(1, 2)', failure, 'Too many arguments.');

  Eps; Test('arccsc(−1E30)', 0.0);
  Eps; Test('arccsc(−1)', -Pi/2);
  Eps; Test('arccsc(−2/√3)', -Pi/3);
  Eps; Test('arccsc(−√2)', -Pi/4);
  Eps; Test('arccsc(−2)', -Pi/6);
  Eps; Test('arccsc(2)', Pi/6);
  Eps; Test('arccsc(√2)', Pi/4);
  Eps; Test('arccsc(2/√3)', Pi/3);
  Eps; Test('arccsc(1)', Pi/2);
  Eps; Test('arccsc(1E30)', 0.0);

  Test('arccsc(1, 2)', failure, 'Too many arguments.');

  Eps; Test('arcsec(-1E30)', Pi/2);
  Eps; Test('arcsec(−1)', Pi);
  Eps; Test('arcsec(−2/√3)', 5*Pi/6);
  Eps; Test('arcsec(−√2)', 3*Pi/4);
  Eps; Test('arcsec(−2)', 2*Pi/3);
  Eps; Test('arcsec(2)', Pi/3);
  Eps; Test('arcsec(√2)', Pi/4);
  Eps; Test('arcsec(2/√3)', Pi/6);
  Eps; Test('arcsec(1)', 0.0);
  Eps; Test('arcsec(1E30)', Pi/2);

  Test('arcsec(1, 2)', failure, 'Too many arguments.');

  // Exponential function and logarithms

  Test('exp(0)', 1.0);

  Eps; Test('exp(1)', 2.718281828459045235360287471352662497757247093699959574966);
  Eps; Test('exp(2)', 7.389056098930650227230427460575007813180315570551847324087);
  Eps; Test('exp(5)', 148.4131591025766034211155800405522796234876675938789890467);
  Eps; Test('exp(10)', 22026.46579480671651695790064528424436635351261855678107423);
  Eps; Test('exp(20)', 4.85165195409790277969106830541540558684638988944847254E8);
  Eps; Test('exp(30)', 1.0686474581524462146990468650741401650024495005473054E13);
  Eps; Test('exp(40)', 2.3538526683701998540789991074903480450887161725455546E17);
  Eps; Test('exp(50)', 5.1847055285870724640874533229334853848274691005838464E21);
  Eps; Test('exp(60)', 1.1420073898156842836629571831447656301980459595563958E26);
  Eps; Test('exp(70)', 2.5154386709191670062657811742521129614074129733831405E30);
  Eps; Test('exp(80)', 5.5406223843935100525711733958316612924856728832685322E34);
  Eps; Test('exp(90)', 1.2204032943178408020027100351363697539707464210997675E39);
  Eps; Test('exp(100)', 2.6881171418161354484126255515800135873611118773741922E43);
  Eps; Test('exp(150)', 1.3937095806663796973183419371414574774736900614021843E65);
  Eps; Test('exp(200)', 7.2259737681257492581774770421893056973568744285273192E86);
  Eps; Test('exp(250)', 3.7464546145026732603499548122029201458088307542813159E108);

  Eps; Test('exp(-1)', 1/2.718281828459045235360287471352662497757247093699959574966);
  Eps; Test('exp(-2)', 1/7.389056098930650227230427460575007813180315570551847324087);
  Eps; Test('exp(-5)', 1/148.4131591025766034211155800405522796234876675938789890467);
  Eps; Test('exp(-10)', 1/22026.46579480671651695790064528424436635351261855678107423);
  Eps; Test('exp(-20)', 1/4.85165195409790277969106830541540558684638988944847254E8);
  Eps; Test('exp(-30)', 1/1.0686474581524462146990468650741401650024495005473054E13);
  Eps; Test('exp(-40)', 1/2.3538526683701998540789991074903480450887161725455546E17);
  Eps; Test('exp(-50)', 1/5.1847055285870724640874533229334853848274691005838464E21); // actually, the wrong kind of epsilon comparison, so this test is not useful
  Eps; Test('exp(-60)', 1/1.1420073898156842836629571831447656301980459595563958E26); // -"-
  Eps; Test('exp(-70)', 1/2.5154386709191670062657811742521129614074129733831405E30);
  Eps; Test('exp(-80)', 1/5.5406223843935100525711733958316612924856728832685322E34);
  Eps; Test('exp(-90)', 1/1.2204032943178408020027100351363697539707464210997675E39);
  Eps; Test('exp(-100)', 1/2.6881171418161354484126255515800135873611118773741922E43);
  Eps; Test('exp(-150)', 1/1.3937095806663796973183419371414574774736900614021843E65);
  Eps; Test('exp(-200)', 1/7.2259737681257492581774770421893056973568744285273192E86);
  Eps; Test('exp(-250)', 1/3.7464546145026732603499548122029201458088307542813159E108);

  Eps; Test('exp(1/3)', 1.395612425086089528628125319602586837597906515199406982617);
  Eps; Test('exp(π)', 23.14069263277926900572908636794854738026610624260021199344);
  Eps; Test('exp(−π)', 1/23.14069263277926900572908636794854738026610624260021199344);

  Test('exp(1, 2)', failure, 'Too many arguments.');

  Test('e^0', 1.0);

  Eps; Test('e^1', 2.718281828459045235360287471352662497757247093699959574966);
  Eps; Test('e^2', 7.389056098930650227230427460575007813180315570551847324087);
  Eps; Test('e^5', 148.4131591025766034211155800405522796234876675938789890467);
  Eps; Test('e^10', 22026.46579480671651695790064528424436635351261855678107423);
  Eps; Test('e^20', 4.85165195409790277969106830541540558684638988944847254E8);
  Eps; Test('e^30', 1.0686474581524462146990468650741401650024495005473054E13);
  Eps; Test('e^40', 2.3538526683701998540789991074903480450887161725455546E17);
  Eps; Test('e^50', 5.1847055285870724640874533229334853848274691005838464E21);
  Eps; Test('e^60', 1.1420073898156842836629571831447656301980459595563958E26);
  Eps; Test('e^70', 2.5154386709191670062657811742521129614074129733831405E30);
  Eps; Test('e^80', 5.5406223843935100525711733958316612924856728832685322E34);
  Eps; Test('e^90', 1.2204032943178408020027100351363697539707464210997675E39);
  Eps; Test('e^100', 2.6881171418161354484126255515800135873611118773741922E43);
  Eps; Test('e^150', 1.3937095806663796973183419371414574774736900614021843E65);
  Eps; Test('e^200', 7.2259737681257492581774770421893056973568744285273192E86);
  Eps; Test('e^250', 3.7464546145026732603499548122029201458088307542813159E108);

  Eps; Test('e^-1', 1/2.718281828459045235360287471352662497757247093699959574966);
  Eps; Test('e^-2', 1/7.389056098930650227230427460575007813180315570551847324087);
  Eps; Test('e^-5', 1/148.4131591025766034211155800405522796234876675938789890467);
  Eps; Test('e^-10', 1/22026.46579480671651695790064528424436635351261855678107423);
  Eps; Test('e^-20', 1/4.85165195409790277969106830541540558684638988944847254E8);
  Eps; Test('e^-30', 1/1.0686474581524462146990468650741401650024495005473054E13);
  Eps; Test('e^-40', 1/2.3538526683701998540789991074903480450887161725455546E17);
  Eps; Test('e^-50', 1/5.1847055285870724640874533229334853848274691005838464E21);
  Eps; Test('e^-60', 1/1.1420073898156842836629571831447656301980459595563958E26);
  Eps; Test('e^-70', 1/2.5154386709191670062657811742521129614074129733831405E30);
  Eps; Test('e^-80', 1/5.5406223843935100525711733958316612924856728832685322E34);
  Eps; Test('e^-90', 1/1.2204032943178408020027100351363697539707464210997675E39);
  Eps; Test('e^-100', 1/2.6881171418161354484126255515800135873611118773741922E43);
  Eps; Test('e^-150', 1/1.3937095806663796973183419371414574774736900614021843E65);
  Eps; Test('e^-200', 1/7.2259737681257492581774770421893056973568744285273192E86);
  Eps; Test('e^-250', 1/3.7464546145026732603499548122029201458088307542813159E108);

  Eps; Test('e^(1/3)', 1.395612425086089528628125319602586837597906515199406982617);
  Eps; Test('e^π', 23.14069263277926900572908636794854738026610624260021199344);
  Eps; Test('e^−π', 1/23.14069263277926900572908636794854738026610624260021199344);

  Eps; Test('ln(1)', 0.0);
  Eps; Test('ln(e)', 1.0);
  Eps; Test('ln(e^2)', 2.0);
  Eps; Test('ln(e^3)', 3.0);
  Eps; Test('ln(e^4)', 4.0);
  Eps; Test('ln(e^5)', 5.0);
  Eps; Test('ln(e^6)', 6.0);
  Eps; Test('ln(e^7)', 7.0);
  Eps; Test('ln(e^8)', 8.0);
  Eps; Test('ln(e^9)', 9.0);
  Eps; Test('ln(e^10)', 10.0);
  Eps; Test('ln(e^20)', 20.0);
  Eps; Test('ln(e^50)', 50.0);
  Eps; Test('ln(e^100)', 100.0);
  Eps; Test('ln(e^150)', 150.0);
  Eps; Test('ln(e^200)', 200.0);
  Eps; Test('ln(e^250)', 250.0);

  Eps; Test('ln(e^−1)', -1.0);
  Eps; Test('ln(e^−2)', -2.0);
  Eps; Test('ln(e^−3)', -3.0);
  Eps; Test('ln(e^−4)', -4.0);
  Eps; Test('ln(e^−5)', -5.0);
  Eps; Test('ln(e^−6)', -6.0);
  Eps; Test('ln(e^−7)', -7.0);
  Eps; Test('ln(e^−8)', -8.0);
  Eps; Test('ln(e^−9)', -9.0);
  Eps; Test('ln(e^−10)', -10.0);
  Eps; Test('ln(e^−20)', -20.0);
  Eps; Test('ln(e^−50)', -50.0);
  Eps; Test('ln(e^−100)', -100.0);
  Eps; Test('ln(e^−150)', -150.0);
  Eps; Test('ln(e^−200)', -200.0);
  Eps; Test('ln(e^−250)', -250.0);

  Eps; Test('ln(123)', 4.812184355372417495262008609959933293023901027222051085353);
  Eps; Test('ln(1/3)', -1.09861228866810969139524523692252570464749055782274945173);
  Eps; Test('ln(π)', 1.144729885849400174143427351353058711647294812915311571513);
  Eps; Test('ln(e^π)', Pi);

  Test('ln(1, 2)', failure, 'Too many arguments.');

  Eps; Test('log(1)', 0.0);
  Eps; Test('log(10)', 1.0);
  Eps; Test('log(10^2)', 2.0);
  Eps; Test('log(10^3)', 3.0);
  Eps; Test('log(10^4)', 4.0);
  Eps; Test('log(10^5)', 5.0);
  Eps; Test('log(10^6)', 6.0);
  Eps; Test('log(10^7)', 7.0);
  Eps; Test('log(10^8)', 8.0);
  Eps; Test('log(10^9)', 9.0);
  Eps; Test('log(10^10)', 10.0);
  Eps; Test('log(10^20)', 20.0);
  Eps; Test('log(10^50)', 50.0);
  Eps; Test('log(10^100)', 100.0);
  Eps; Test('log(10^150)', 150.0);
  Eps; Test('log(10^200)', 200.0);
  Eps; Test('log(10^250)', 250.0);

  Eps; Test('log(10^−1)', -1.0);
  Eps; Test('log(10^−2)', -2.0);
  Eps; Test('log(10^−3)', -3.0);
  Eps; Test('log(10^−4)', -4.0);
  Eps; Test('log(10^−5)', -5.0);
  Eps; Test('log(10^−6)', -6.0);
  Eps; Test('log(10^−7)', -7.0);
  Eps; Test('log(10^−8)', -8.0);
  Eps; Test('log(10^−9)', -9.0);
  Eps; Test('log(10^−10)', -10.0);
  Eps; Test('log(10^−20)', -20.0);
  Eps; Test('log(10^−50)', -50.0);
  Eps; Test('log(10^−100)', -100.0);
  Eps; Test('log(10^−150)', -150.0);
  Eps; Test('log(10^−200)', -200.0);
  Eps; Test('log(10^−250)', -250.0);

  Eps; Test('log(123)', 2.089905111439397931804439753223296108730642498024382953837);
  Eps; Test('log(1/3)', -0.47712125471966243729502790325511530920012886419069586482);
  Eps; Test('log(π)', 0.497149872694133854351268288290898873651678324380442446134);
  Eps; Test('log(10^π)', Pi);

  Test('log(1, 2)', failure, 'Too many arguments.');

  // Powers and roots

  Test('sqr(0)', 0);
  Test('sqr(1)', 1);
  Test('sqr(2)', 4);
  Test('sqr(3)', 9);
  Test('sqr(4)', 16);
  Test('sqr(5)', 25);
  Test('sqr(6)', 36);
  Test('sqr(7)', 49);
  Test('sqr(8)', 64);
  Test('sqr(9)', 81);
  Test('sqr(10)', 100);
  Test('sqr(11)', 121);
  Test('sqr(12)', 144);

  Test('sqr(20)', 400);
  Test('sqr(50)', 2500);
  Test('sqr(100)', 10000);
  Test('sqr(150)', 22500);
  Test('sqr(200)', 40000);
  Test('sqr(1000)', 1000000);
  Test('sqr(10000)', 100000000);
  Test('sqr(100000)', 10000000000);
  Test('sqr(1000000)', 1000000000000);
  Test('sqr(10000000)', 100000000000000);
  Test('sqr(100000000)', 10000000000000000);
  Test('sqr(1000000000)', 1000000000000000000);
  Eps; Test('sqr(1E50)', 1E100);

  Eps; Test('sqr(0.5)', 0.25);
  Eps; Test('sqr(0.3)', 0.09);
  Eps; Test('sqr(0.2)', 0.04);
  Eps; Test('sqr(0.1)', 0.01);
  Eps; Test('sqr(0.01)', 0.0001);
  Eps; Test('sqr(0.001)', 0.000001);
  Eps; Test('sqr(0.0001)', 0.00000001);
  Eps; Test('sqr(0.00001)', 0.0000000001);
  Eps; Test('sqr(0.000001)', 0.000000000001);
  Eps; Test('sqr(0.0000001)', 0.00000000000001);
  Eps; Test('sqr(0.00000001)', 0.0000000000000001);
  Eps; Test('sqr(0.000000001)', 0.000000000000000001);
  Eps; Test('sqr(1E-50)', 1E-100);

  Test('sqr(7/3)', Rat(49, 9));
  Test('sqr(4/5)', Rat(16, 25));
  Test('sqr(9/100)', Rat(81, 10000));

  Test('sqr(1/2)', Rat(1, 4));
  Test('sqr(3/10)', Rat(9, 100));
  Test('sqr(1/5)', Rat(1, 25));
  Test('sqr(1/10)', Rat(1, 100));
  Test('sqr(1/100)', Rat(1, 10000));
  Test('sqr(1/1000)', Rat(1, 1000000));
  Test('sqr(1/10000)', Rat(1, 100000000));
  Test('sqr(1/100000)', Rat(1, 10000000000));
  Test('sqr(1/1000000)', Rat(1, 1000000000000));
  Test('sqr(1/10000000)', Rat(1, 100000000000000));
  Test('sqr(1/100000000)', Rat(1, 10000000000000000));

  Test('sqr(−0)', 0);
  Test('sqr(−1)', 1);
  Test('sqr(−2)', 4);
  Test('sqr(−3)', 9);
  Test('sqr(−4)', 16);
  Test('sqr(−5)', 25);
  Test('sqr(−6)', 36);
  Test('sqr(−7)', 49);
  Test('sqr(−8)', 64);
  Test('sqr(−9)', 81);
  Test('sqr(−10)', 100);
  Test('sqr(−11)', 121);
  Test('sqr(−12)', 144);

  Test('sqr(−20)', 400);
  Test('sqr(−50)', 2500);
  Test('sqr(−100)', 10000);
  Test('sqr(−150)', 22500);
  Test('sqr(−200)', 40000);
  Test('sqr(−1000)', 1000000);
  Test('sqr(−10000)', 100000000);
  Test('sqr(−100000)', 10000000000);
  Test('sqr(−1000000)', 1000000000000);
  Test('sqr(−10000000)', 100000000000000);
  Test('sqr(−100000000)', 10000000000000000);
  Test('sqr(−1000000000)', 1000000000000000000);
  Eps; Test('sqr(−1E50)', 1E100);

  Eps; Test('sqr(−0.5)', 0.25);
  Eps; Test('sqr(−0.3)', 0.09);
  Eps; Test('sqr(−0.2)', 0.04);
  Eps; Test('sqr(−0.1)', 0.01);
  Eps; Test('sqr(−0.01)', 0.0001);
  Eps; Test('sqr(−0.001)', 0.000001);
  Eps; Test('sqr(−0.0001)', 0.00000001);
  Eps; Test('sqr(−0.00001)', 0.0000000001);
  Eps; Test('sqr(−0.000001)', 0.000000000001);
  Eps; Test('sqr(−0.0000001)', 0.00000000000001);
  Eps; Test('sqr(−0.00000001)', 0.0000000000000001);
  Eps; Test('sqr(−0.000000001)', 0.000000000000000001);
  Eps; Test('sqr(−1E-50)', 1E-100);

  Test('sqr(−7/3)', Rat(49, 9));
  Test('sqr(−4/5)', Rat(16, 25));
  Test('sqr(−9/100)', Rat(81, 10000));

  Test('sqr(−1/2)', Rat(1, 4));
  Test('sqr(−3/10)', Rat(9, 100));
  Test('sqr(−1/5)', Rat(1, 25));
  Test('sqr(−1/10)', Rat(1, 100));
  Test('sqr(−1/100)', Rat(1, 10000));
  Test('sqr(−1/1000)', Rat(1, 1000000));
  Test('sqr(−1/10000)', Rat(1, 100000000));
  Test('sqr(−1/100000)', Rat(1, 10000000000));
  Test('sqr(−1/1000000)', Rat(1, 1000000000000));
  Test('sqr(−1/10000000)', Rat(1, 100000000000000));
  Test('sqr(−1/100000000)', Rat(1, 10000000000000000));

  Test('sqr(1, 2)', failure, 'Too many arguments.');

  Test('sqrt(0)', 0.0);
  Test('sqrt(1)', 1.0);
  Test('sqrt(4)', 2.0);
  Test('sqrt(9)', 3.0);
  Test('sqrt(16)', 4.0);
  Test('sqrt(25)', 5.0);
  Test('sqrt(36)', 6.0);
  Test('sqrt(49)', 7.0);
  Test('sqrt(64)', 8.0);
  Test('sqrt(81)', 9.0);
  Test('sqrt(100)', 10.0);
  Test('sqrt(121)', 11.0);
  Test('sqrt(144)', 12.0);

  Test('sqrt(0.0)', 0.0);
  Test('sqrt(1.0)', 1.0);
  Test('sqrt(4.0)', 2.0);
  Test('sqrt(9.0)', 3.0);
  Test('sqrt(16.0)', 4.0);
  Test('sqrt(25.0)', 5.0);
  Test('sqrt(36.0)', 6.0);
  Test('sqrt(49.0)', 7.0);
  Test('sqrt(64.0)', 8.0);
  Test('sqrt(81.0)', 9.0);
  Test('sqrt(100.0)', 10.0);
  Test('sqrt(121.0)', 11.0);
  Test('sqrt(144.0)', 12.0);

  Test('sqrt(100)', 10.0);
  Test('sqrt(10000)', 100.0);
  Test('sqrt(1000000)', 1000.0);
  Test('sqrt(100000000)', 10000.0);
  Test('sqrt(10000000000)', 100000.0);
  Test('sqrt(1000000000000)', 1000000.0);
  Test('sqrt(100000000000000)', 10000000.0);
  Test('sqrt(10000000000000000)', 100000000.0);
  Test('sqrt(1000000000000000000)', 1000000000.0);
  Test('sqrt(100000000000000000000)', 10000000000.0);
  Test('sqrt(10000000000000000000000)', 100000000000.0);
  Test('sqrt(1000000000000000000000000)', 1000000000000.0);

  Test('sqrt(100.0)', 10.0);
  Test('sqrt(10000.0)', 100.0);
  Test('sqrt(1000000.0)', 1000.0);
  Test('sqrt(100000000.0)', 10000.0);
  Test('sqrt(10000000000.0)', 100000.0);
  Test('sqrt(1000000000000.0)', 1000000.0);
  Test('sqrt(100000000000000.0)', 10000000.0);
  Test('sqrt(10000000000000000.0)', 100000000.0);
  Test('sqrt(1000000000000000000.0)', 1000000000.0);
  Test('sqrt(100000000000000000000.0)', 10000000000.0);
  Test('sqrt(10000000000000000000000.0)', 100000000000.0);
  Test('sqrt(1000000000000000000000000.0)', 1000000000000.0);
  Eps; Test('sqrt(1E200)', 1E100);

  Eps; Test('sqrt(12345)', 111.1080555135405112450044387430752414899113774596977299764);
  Eps; Test('sqrt(12345.6789)', 111.11111060555555440541666143353469245878409860134351);

  Eps; Test('sqrt(0.25)', 0.5);
  Eps; Test('sqrt(0.16)', 0.4);
  Eps; Test('sqrt(0.09)', 0.3);
  Eps; Test('sqrt(0.04)', 0.2);
  Eps; Test('sqrt(0.01)', 0.1);
  Eps; Test('sqrt(0.0001)', 0.01);
  Eps; Test('sqrt(0.000001)', 0.001);
  Eps; Test('sqrt(0.00000001)', 0.0001);
  Eps; Test('sqrt(0.0000000001)', 0.00001);
  Eps; Test('sqrt(0.000000000001)', 0.000001);
  Eps; Test('sqrt(0.00000000000001)', 0.0000001);
  Eps; Test('sqrt(0.0000000000000001)', 0.00000001);
  Eps; Test('sqrt(0.000000000000000001)', 0.000000001);
  Eps; Test('sqrt(1E-200)', 1E-100);

  Eps; Test('sqrt(π)', 1.772453850905516027298167483341145182797549456122387128213);
  Eps; Test('sqrt(81/25)', 9/5);
  Eps; Test('sqrt(81/25 - 81/25)', 0.0);

  Test('sqrt(1, 2)', failure, 'Too many arguments.');

  // Hyperbolic functions

  Test('sinh(0)', 0.0);

  Eps; Test('sinh(1/3)', 0.339540557256150139101260611338603585072397227688963912738);
  Eps; Test('sinh(0.5)', 0.5210953054937473616224256264114915591059289826114805);
  Eps; Test('sinh(1)', 1.175201193643801456882381850595600815155717981334095870229);
  Eps; Test('sinh(2)', 3.626860407847018767668213982801261704886342012321135721309);
  Eps; Test('sinh(5)', 74.20321057778875897700947199606456559961940900442581698066);
  Eps; Test('sinh(10)', 11013.23287470339337723652455484636440290145119031934610383);
  Eps; Test('sinh(20)', 2.42582597704895137953976604051491365359349304394513138E8);
  Eps; Test('sinh(50)', 2.5923527642935362320437266614667426924137344538544308E21);

  Eps; Test('sinh(-1/3)', -0.339540557256150139101260611338603585072397227688963912738);
  Eps; Test('sinh(-0.5)', -0.5210953054937473616224256264114915591059289826114805);
  Eps; Test('sinh(-1)', -1.175201193643801456882381850595600815155717981334095870229);
  Eps; Test('sinh(-2)', -3.626860407847018767668213982801261704886342012321135721309);
  Eps; Test('sinh(-5)', -74.20321057778875897700947199606456559961940900442581698066);
  Eps; Test('sinh(-10)', -11013.23287470339337723652455484636440290145119031934610383);
  Eps; Test('sinh(-20)', -2.42582597704895137953976604051491365359349304394513138E8);
  Eps; Test('sinh(-50)', -2.5923527642935362320437266614667426924137344538544308E21);

  Test('sinh(1, 2)', failure, 'Too many arguments.');

  Test('cosh(0)', 1.0);

  Eps; Test('cosh(1/3)', 1.056071867829939389526864708263983252525509287510443069879);
  Eps; Test('cosh(0.5)', 1.1276259652063807852262251614026720125478471180986674);
  Eps; Test('cosh(1)', 1.543080634815243778477905620757061682601529112365863704737);
  Eps; Test('cosh(2)', 3.762195691083631459562213477773746108293973558230711602777);
  Eps; Test('cosh(5)', 74.20994852478784444410610804448771402386825858945317206609);
  Eps; Test('cosh(10)', 11013.23292010332313972137609043787996345206142823743497040);
  Eps; Test('cosh(20)', 2.42582597704895140015130226490049193325289684550334115E8);
  Eps; Test('cosh(50)', 2.5923527642935362320437266614667426924137346467294155E21);

  Eps; Test('cosh(-1/3)', 1.056071867829939389526864708263983252525509287510443069879);
  Eps; Test('cosh(-0.5)', 1.1276259652063807852262251614026720125478471180986674);
  Eps; Test('cosh(-1)', 1.543080634815243778477905620757061682601529112365863704737);
  Eps; Test('cosh(-2)', 3.762195691083631459562213477773746108293973558230711602777);
  Eps; Test('cosh(-5)', 74.20994852478784444410610804448771402386825858945317206609);
  Eps; Test('cosh(-10)', 11013.23292010332313972137609043787996345206142823743497040);
  Eps; Test('cosh(-20)', 2.42582597704895140015130226490049193325289684550334115E8);
  Eps; Test('cosh(-50)', 2.5923527642935362320437266614667426924137346467294155E21);

  Test('cosh(1, 2)', failure, 'Too many arguments.');

  Test('tanh(0)', 0.0);

  Eps; Test('tanh(1/3)', 0.321512737531634344719406222425206466005292002502083511679);
  Eps; Test('tanh(0.5)', 0.4621171572600097585023184836436725487302892803301130);
  Eps; Test('tanh(1)', 0.761594155955764888119458282604793590412768597257936551596);
  Eps; Test('tanh(2)', 0.964027580075816883946413724100923150255029976240934776048);
  Eps; Test('tanh(5)', 0.999909204262595131210990447534473021089812615990547862736);
  Eps; Test('tanh(10)', 0.999999995877692763619592837138275741050814618495019962261);
  Eps; Test('tanh(20)', 0.999999999999999991503291489416822045438558191190987257130);
  Eps; Test('tanh(50)', 0.999999999999999999999999999999999999999999925598480479583);
  Eps; Test('tanh(100)', 1.0);
  Eps; Test('tanh(1000)', 1.0);
  Eps; Test('tanh(1E10)', 1.0);
  Eps; Test('tanh(1E50)', 1.0);
  Eps; Test('tanh(1E100)', 1.0);

  Eps; Test('tanh(-1/3)', -0.321512737531634344719406222425206466005292002502083511679);
  Eps; Test('tanh(-0.5)', -0.4621171572600097585023184836436725487302892803301130);
  Eps; Test('tanh(-1)', -0.761594155955764888119458282604793590412768597257936551596);
  Eps; Test('tanh(-2)', -0.964027580075816883946413724100923150255029976240934776048);
  Eps; Test('tanh(-5)', -0.999909204262595131210990447534473021089812615990547862736);
  Eps; Test('tanh(-10)', -0.999999995877692763619592837138275741050814618495019962261);
  Eps; Test('tanh(-20)', -0.999999999999999991503291489416822045438558191190987257130);
  Eps; Test('tanh(-50)', -0.999999999999999999999999999999999999999999925598480479583);
  Eps; Test('tanh(-100)', -1.0);
  Eps; Test('tanh(-1000)', -1.0);
  Eps; Test('tanh(-1E10)', -1.0);
  Eps; Test('tanh(-1E50)', -1.0);
  Eps; Test('tanh(-1E100)', -1.0);

  Test('tanh(1, 2)', failure, 'Too many arguments.');

  Eps; Test('coth(1E-10)', 1.0000000000000000000033333333333333333333311111111111E10);
  Eps; Test('coth(1/10000)', 10000.00003333333331111111113227513225396825398963176738790);
  Eps; Test('coth(1/1000)', 1000.000333333311111113227513015873037250812864188744823643);
  Eps; Test('coth(1/100)', 100.0033333111113227492063705839319236906316350002943191638);
  Eps; Test('coth(1/10)', 10.033311132253989610145270492851499126734254107108310);
  Eps; Test('coth(1/3)', 3.110296679619443701378375231417624794566543581787011086057);
  Eps; Test('coth(0.5)', 2.1639534137386528487700040102180231170937386021507922);
  Eps; Test('coth(1)', 1.313035285499331303636161246930847832912013941240452655543);
  Eps; Test('coth(2)', 1.037314720727548095877809764767820711662391269249194603569);
  Eps; Test('coth(5)', 1.000090803982019375536657920521687603830547437521890754159);
  Eps; Test('coth(10)', 1.000000004122307253373824184028080312460180026756219308447);
  Eps; Test('coth(20)', 1.000000000000000008496708510583178026755497322625619635355);
  Eps; Test('coth(50)', 1.000000000000000000000000000000000000000000074401519520417);
  Eps; Test('coth(100)', 1.0);
  Eps; Test('coth(1000)', 1.0);
  Eps; Test('coth(1E10)', 1.0);
  Eps; Test('coth(1E50)', 1.0);
  Eps; Test('coth(1E100)', 1.0);

  Eps; Test('coth(-1E-10)', -1.0000000000000000000033333333333333333333311111111111E10);
  Eps; Test('coth(-1/10000)', -10000.00003333333331111111113227513225396825398963176738790);
  Eps; Test('coth(-1/1000)', -1000.000333333311111113227513015873037250812864188744823643);
  Eps; Test('coth(-1/100)', -100.0033333111113227492063705839319236906316350002943191638);
  Eps; Test('coth(-1/10)', -10.033311132253989610145270492851499126734254107108310);
  Eps; Test('coth(-1/3)', -3.110296679619443701378375231417624794566543581787011086057);
  Eps; Test('coth(-0.5)', -2.1639534137386528487700040102180231170937386021507922);
  Eps; Test('coth(-1)', -1.313035285499331303636161246930847832912013941240452655543);
  Eps; Test('coth(-2)', -1.037314720727548095877809764767820711662391269249194603569);
  Eps; Test('coth(-5)', -1.000090803982019375536657920521687603830547437521890754159);
  Eps; Test('coth(-10)', -1.000000004122307253373824184028080312460180026756219308447);
  Eps; Test('coth(-20)', -1.000000000000000008496708510583178026755497322625619635355);
  Eps; Test('coth(-50)', -1.000000000000000000000000000000000000000000074401519520417);
  Eps; Test('coth(-100)', -1.0);
  Eps; Test('coth(-1000)', -1.0);
  Eps; Test('coth(-1E10)', -1.0);
  Eps; Test('coth(-1E50)', -1.0);
  Eps; Test('coth(-1E100)', -1.0);

  Test('coth(1, 2)', failure, 'Too many arguments.');

  Eps(1); Test('csch(1E-10)', 9.999999999999999999983333333333333333333352777777778E9);
  Eps; Test('csch(1/10000)', 9999.999983333333352777777757275132296130952359616335332413);
  Eps; Test('csch(1/1000)', 999.9998333333527777757275134374999786639565217460263892157);
  Eps; Test('csch(1/100)', 99.99833335277757275342259771178973856982340751490977001337);
  Eps; Test('csch(1/10)', 9.983352757296109637946884127643214807091454313327291581209);
  Eps; Test('csch(1/3)', 2.945156266694814347645586003495165676460850560888490147110);
  Eps; Test('csch(0.5)', 1.919034751334943719492202878727006159587179871532971946414);
  Eps; Test('csch(1)', 0.850918128239321545133842763287175284181724660910339616990);
  Eps; Test('csch(2)', 0.275720564771783207758351482163027121249622671991258051973);
  Eps; Test('csch(5)', 0.013476505830589086655381881284337964618035455336483814697);
  Eps; Test('csch(10)', 0.000090799859712122162833736493607291370367410765671445711);
  Eps; Test('csch(20)', 4.1223072448771156734449022857046827041305996730932590E-9);
  Eps; Test('csch(50)', 3.8574996959278355660346856330540251495056654459624452E-22);
  Eps; Test('csch(100)', 7.4401519520416719259193916077262366747177845847535639E-44);
  Eps; Test('csch(1000)', 0.0);
  Eps; Test('csch(1E4)', 0.0);
  Eps; Test('csch(1E5)', 0.0);
  Eps; Test('csch(1E6)', 0.0);
  Eps; Test('csch(1E7)', 0.0);
  Eps; Test('csch(1E8)', 0.0);
  Eps; Test('csch(1E9)', 0.0);
  Eps; Test('csch(1E10)', 0.0);
  Eps; Test('csch(1E20)', 0.0);
  Eps; Test('csch(1E30)', 0.0);
  Eps; Test('csch(1E40)', 0.0);
  Eps; Test('csch(1E50)', 0.0);
  Eps; Test('csch(1E60)', 0.0);
  Eps; Test('csch(1E70)', 0.0);
  Eps; Test('csch(1E80)', 0.0);
  Eps; Test('csch(1E90)', 0.0);
  Eps; Test('csch(1E100)', 0.0);
  Eps; Test('csch(1E150)', 0.0);

  Eps(1); Test('csch(-1E-10)', -9.999999999999999999983333333333333333333352777777778E9);
  Eps; Test('csch(-1/10000)', -9999.999983333333352777777757275132296130952359616335332413);
  Eps; Test('csch(-1/1000)', -999.9998333333527777757275134374999786639565217460263892157);
  Eps; Test('csch(-1/100)', -99.99833335277757275342259771178973856982340751490977001337);
  Eps; Test('csch(-1/10)', -9.983352757296109637946884127643214807091454313327291581209);
  Eps; Test('csch(-1/3)', -2.945156266694814347645586003495165676460850560888490147110);
  Eps; Test('csch(-0.5)', -1.919034751334943719492202878727006159587179871532971946414);
  Eps; Test('csch(-1)', -0.850918128239321545133842763287175284181724660910339616990);
  Eps; Test('csch(-2)', -0.275720564771783207758351482163027121249622671991258051973);
  Eps; Test('csch(-5)', -0.013476505830589086655381881284337964618035455336483814697);
  Eps; Test('csch(-10)', -0.000090799859712122162833736493607291370367410765671445711);
  Eps; Test('csch(-20)', -4.1223072448771156734449022857046827041305996730932590E-9);
  Eps; Test('csch(-50)', -3.8574996959278355660346856330540251495056654459624452E-22);
  Eps; Test('csch(-100)', -7.4401519520416719259193916077262366747177845847535639E-44);
  Eps; Test('csch(-1000)', 0.0);
  Eps; Test('csch(-1E4)', 0.0);
  Eps; Test('csch(-1E5)', 0.0);
  Eps; Test('csch(-1E6)', 0.0);
  Eps; Test('csch(-1E7)', 0.0);
  Eps; Test('csch(-1E8)', 0.0);
  Eps; Test('csch(-1E9)', 0.0);
  Eps; Test('csch(-1E10)', 0.0);
  Eps; Test('csch(-1E20)', 0.0);
  Eps; Test('csch(-1E30)', 0.0);
  Eps; Test('csch(-1E40)', 0.0);
  Eps; Test('csch(-1E50)', 0.0);
  Eps; Test('csch(-1E60)', 0.0);
  Eps; Test('csch(-1E70)', 0.0);
  Eps; Test('csch(-1E80)', 0.0);
  Eps; Test('csch(-1E90)', 0.0);
  Eps; Test('csch(-1E100)', 0.0);
  Eps; Test('csch(-1E150)', 0.0);

  Test('csch(1, 2)', failure, 'Too many arguments.');

  Test('sech(0)', 1.0);

  Eps; Test('sech(1/10)', 0.995020748953226491634506512599732588375439771795889118064);
  Eps; Test('sech(1/3)', 0.946905253763497926558952280139811647196865554127211019972);
  Eps; Test('sech(0.5)', 0.886818883970073908658897797783408562534089088712613924836);
  Eps; Test('sech(1)', 0.648054273663885399574977353226150323108489312071942023037);
  Eps; Test('sech(2)', 0.265802228834079692120862739819888971530782654432268069714);
  Eps; Test('sech(5)', 0.013475282221304557305519138244882155290837353941700686833);
  Eps; Test('sech(10)', 0.000090799859337817244080129507820478187068803766001825381);
  Eps; Test('sech(20)', 4.1223072448771156384188592349186013501756684701365942E-9);
  Eps; Test('sech(50)', 3.8574996959278355660346856330540251495056651589586063E-22);
  Eps; Test('sech(100)', 7.4401519520416719259193916077262366747177845847535639E-44);
  Eps; Test('sech(1000)', 0.0);
  Eps; Test('sech(1E4)', 0.0);
  Eps; Test('sech(1E5)', 0.0);
  Eps; Test('sech(1E6)', 0.0);
  Eps; Test('sech(1E7)', 0.0);
  Eps; Test('sech(1E8)', 0.0);
  Eps; Test('sech(1E9)', 0.0);
  Eps; Test('sech(1E10)', 0.0);
  Eps; Test('sech(1E20)', 0.0);
  Eps; Test('sech(1E30)', 0.0);
  Eps; Test('sech(1E40)', 0.0);
  Eps; Test('sech(1E50)', 0.0);
  Eps; Test('sech(1E60)', 0.0);
  Eps; Test('sech(1E70)', 0.0);
  Eps; Test('sech(1E80)', 0.0);
  Eps; Test('sech(1E90)', 0.0);
  Eps; Test('sech(1E100)', 0.0);
  Eps; Test('sech(1E150)', 0.0);

  Eps; Test('sech(-1/10)', 0.995020748953226491634506512599732588375439771795889118064);
  Eps; Test('sech(-1/3)', 0.946905253763497926558952280139811647196865554127211019972);
  Eps; Test('sech(-0.5)', 0.886818883970073908658897797783408562534089088712613924836);
  Eps; Test('sech(-1)', 0.648054273663885399574977353226150323108489312071942023037);
  Eps; Test('sech(-2)', 0.265802228834079692120862739819888971530782654432268069714);
  Eps; Test('sech(-5)', 0.013475282221304557305519138244882155290837353941700686833);
  Eps; Test('sech(-10)', 0.000090799859337817244080129507820478187068803766001825381);
  Eps; Test('sech(-20)', 4.1223072448771156384188592349186013501756684701365942E-9);
  Eps; Test('sech(-50)', 3.8574996959278355660346856330540251495056651589586063E-22);
  Eps; Test('sech(-100)', 7.4401519520416719259193916077262366747177845847535639E-44);
  Eps; Test('sech(-1000)', 0.0);
  Eps; Test('sech(-1E4)', 0.0);
  Eps; Test('sech(-1E5)', 0.0);
  Eps; Test('sech(-1E6)', 0.0);
  Eps; Test('sech(-1E7)', 0.0);
  Eps; Test('sech(-1E8)', 0.0);
  Eps; Test('sech(-1E9)', 0.0);
  Eps; Test('sech(-1E10)', 0.0);
  Eps; Test('sech(-1E20)', 0.0);
  Eps; Test('sech(-1E30)', 0.0);
  Eps; Test('sech(-1E40)', 0.0);
  Eps; Test('sech(-1E50)', 0.0);
  Eps; Test('sech(-1E60)', 0.0);
  Eps; Test('sech(-1E70)', 0.0);
  Eps; Test('sech(-1E80)', 0.0);
  Eps; Test('sech(-1E90)', 0.0);
  Eps; Test('sech(-1E100)', 0.0);
  Eps; Test('sech(-1E150)', 0.0);

  Test('sech(1, 2)', failure, 'Too many arguments.');

  // Inverse hyperbolic functions

  Test('arcsinh(0)', 0.0);

  Eps; Test('arcsinh(1/3)', 0.327450150237258443322535259988258127700524528990767451275);
  Eps; Test('arcsinh(0.5)', 0.481211825059603447497758913424368423135184334385660519661);
  Eps; Test('arcsinh(1)', 0.881373587019543025232609324979792309028160328261635410753);
  Eps; Test('arcsinh(2)', 1.443635475178810342493276740273105269405553003156981558983);
  Eps; Test('arcsinh(5)', 2.312438341272752620253562341364414383658245072646559237167);
  Eps; Test('arcsinh(10)', 2.998222950297969738846595537596453476607058054877303655734);
  Eps; Test('arcsinh(20)', 3.689503868988905640821653570961093329678760832868086366520);
  Eps; Test('arcsinh(50)', 4.605270170991423826621239267208306413008902245692188338724);
  Eps; Test('arcsinh(100)', 5.298342365610588757368825689112906302142383535156218238322);
  Eps; Test('arcsinh(1000)', 7.600902709541988611523289784664939633568339120417227831061);
  Eps; Test('arcsinh(1E4)', 9.903487555036128036114197992278966390016366883146161692759);
  Eps; Test('arcsinh(1E5)', 12.20607264555517372950625189488004968941433749286870371371);
  Eps; Test('arcsinh(1E6)', 14.50865773852446941352518075581436181373419239946619226396);
  Eps; Test('arcsinh(1E7)', 16.81124283151826759754317230423935102128321060684499941968);
  Eps; Test('arcsinh(1E8)', 19.11382792451231080656116375893308929138431204339049114572);
  Eps; Test('arcsinh(1E9)', 21.41641301750635646582915521361745443639166353201921203847);
  Eps; Test('arcsinh(1E10)', 23.71899811050040214959964666830181864408650564564798501445);
  Eps; Test('arcsinh(1E11)', 26.02158320349444783361516312298618285168761650833925799048);
  Eps; Test('arcsinh(1E12)', 28.32416829648849351763312982767054705928871799790543721652);
  Eps; Test('arcsinh(1E50)', 115.8224018302622295103168048556763869481305745657989040557);
  Eps; Test('arcsinh(1E100)', 230.9516564799645137112163775898945973281856489972375528574);
  Eps; Test('arcsinh(1E150)', 346.0809111296667979121159503241128077082407234286762016591);

  Eps; Test('arcsinh(-1/3)', -0.327450150237258443322535259988258127700524528990767451275);
  Eps; Test('arcsinh(-0.5)', -0.481211825059603447497758913424368423135184334385660519661);
  Eps; Test('arcsinh(-1)', -0.881373587019543025232609324979792309028160328261635410753);
  Eps; Test('arcsinh(-2)', -1.443635475178810342493276740273105269405553003156981558983);
  Eps; Test('arcsinh(-5)', -2.312438341272752620253562341364414383658245072646559237167);
  Eps; Test('arcsinh(-10)', -2.998222950297969738846595537596453476607058054877303655734);
  Eps; Test('arcsinh(-20)', -3.689503868988905640821653570961093329678760832868086366520);
  Eps; Test('arcsinh(-50)', -4.605270170991423826621239267208306413008902245692188338724);
  Eps; Test('arcsinh(-100)', -5.298342365610588757368825689112906302142383535156218238322);
  Eps; Test('arcsinh(-1000)', -7.600902709541988611523289784664939633568339120417227831061);
  Eps; Test('arcsinh(-1E4)', -9.903487555036128036114197992278966390016366883146161692759);
  Eps; Test('arcsinh(-1E5)', -12.20607264555517372950625189488004968941433749286870371371);
  Eps; Test('arcsinh(-1E6)', -14.50865773852446941352518075581436181373419239946619226396);
  Eps; Test('arcsinh(-1E7)', -16.81124283151826759754317230423935102128321060684499941968);
  Eps; Test('arcsinh(-1E8)', -19.11382792451231080656116375893308929138431204339049114572);
  Eps; Test('arcsinh(-1E9)', -21.41641301750635646582915521361745443639166353201921203847);
  Eps; Test('arcsinh(-1E10)', -23.71899811050040214959964666830181864408650564564798501445);
  Eps; Test('arcsinh(-1E11)', -26.02158320349444783361516312298618285168761650833925799048);
  Eps; Test('arcsinh(-1E12)', -28.32416829648849351763312982767054705928871799790543721652);
  Eps; Test('arcsinh(-1E50)', -115.8224018302622295103168048556763869481305745657989040557);
  Eps; Test('arcsinh(-1E100)', -230.9516564799645137112163775898945973281856489972375528574);
  Eps; Test('arcsinh(-1E150)', -346.0809111296667979121159503241128077082407234286762016591);

  Test('arcsinh(1, 2)', failure, 'Too many arguments.');

  Eps; Test('arccosh(1)', 0.0);
  Eps; Test('arccosh(1.5)', 0.9624236501192068949955178268487368462703686687713210);
  Eps; Test('arccosh(2)', 1.316957896924816708625046347307968444026981971467516479768);
  Eps; Test('arccosh(5/2)', 1.5667992369724110786640568625804834938620823510926588);
  Eps; Test('arccosh(5)', 2.292431669561177687800787311348015431621868240015710247605);
  Eps; Test('arccosh(10)', 2.993222846126380897912667713774182913083660451180980642685);
  Eps; Test('arccosh(20)', 3.688253867361296667618167572032351887833155697655874258829);
  Eps; Test('arccosh(50)', 4.605070170984757159450572551513062915547943483476541577357);
  Eps; Test('arccosh(100)', 5.298292365610484590701666834943247168937251883140100722832);
  Eps; Test('arccosh(1000)', 7.600902209541988611419123117998223748151672423825895985802);
  Eps; Test('arccosh(1E4)', 9.903487550036128036114197888112299723349695294604495026092);
  Eps; Test('arccosh(1E5)', 12.20607264550517372950625189487994552274767082620203655486);
  Eps; Test('arccosh(1E6)', 14.50865773852396941352518075581436181363002573279952559729);
  Eps; Test('arccosh(1E7)', 16.81124283151826259754317230423935102128321050267833275302);
  Eps; Test('arccosh(1E8)', 19.11382792451231075656116375893308929138431204339038697905);
  Eps; Test('arccosh(1E9)', 21.41641301750635646532915521361745443639166353201921203836);
  Eps; Test('arccosh(1E10)', 23.71899811050040214959464666830181864408650564564798501445);
  Eps; Test('arccosh(1E11)', 26.02158320349444783361511312298618285168761650833925799048);
  Eps; Test('arccosh(1E12)', 28.32416829648849351763312932767054705928871799790543721652);
  Eps; Test('arccosh(1E50)', 115.8224018302622295103168048556763869481305745657989040557);
  Eps; Test('arccosh(1E100)', 230.9516564799645137112163775898945973281856489972375528574);
  Eps; Test('arccosh(1E150)', 346.0809111296667979121159503241128077082407234286762016591);

  Test('arccosh(1, 2)', failure, 'Too many arguments.');

  Test('arctanh(0)', 0.0);

  Eps; Test('arctanh(0.01)', 0.0100003333533347620158821075514042238870973088073443);
  Eps; Test('arctanh(0.1)', 0.1003353477310755806357265520600389452633628691459591);
  Eps; Test('arctanh(1/3)', 0.346573590279972654708616060729088284037750067180127627060);
  Eps; Test('arctanh(1/2)', 0.549306144334054845697622618461262852323745278911374725867);
  Eps; Test('arctanh(0.8)', 1.098612288668109691395245236922525704647490557822749451734);
  Eps; Test('arctanh(0.9)', 1.472219489583220230004513715943926768618689630649564409268);
  Eps; Test('arctanh(0.99)', 2.646652412362246197705060645934268600945552640284736249453);
  Eps; Test('arctanh(0.999)', 3.800201167250200031775967267809004462263675695230499890921);
  Eps; Test('arctanh(0.9999)', 4.951718775643043188629984385462344226560343044514956583248);
  Eps(1E-10); Test('arctanh(0.99999)', 6.103033822758836843920183238794164167618256740528365250187);
  Eps(1E-10); Test('arctanh(0.999999)', 7.254328619262047206741757083636344448506419115841743930067);
  Eps(1E-9); Test('arctanh(0.9999999)', 8.405621390759131923771565318790248427277021942745416320706);
  Eps(1E-8); Test('arctanh(0.99999999)', 9.556913959756155384530581858633211702983822375861884895776);
  Eps(1E-8); Test('arctanh(0.999999999)', 10.70820650850317823272707760678789388490156093267314768587);
  Eps(1E-8); Test('arctanh(0.9999999999)', 11.85949905522520107479794833415088848870992339574065914264);
  Eps(1E-8); Test('arctanh(0.99999999999)', 13.01079160174472391680756281149309140501047492122692066190);
  Eps(1E-6); Test('arctanh(0.999999999999)', 14.16208414824399675881656472633527352962352566561942433742);

  Eps; Test('arctanh(-0.01)', -0.0100003333533347620158821075514042238870973088073443);
  Eps; Test('arctanh(-0.1)', -0.1003353477310755806357265520600389452633628691459591);
  Eps; Test('arctanh(-1/3)', -0.346573590279972654708616060729088284037750067180127627060);
  Eps; Test('arctanh(-1/2)', -0.549306144334054845697622618461262852323745278911374725867);
  Eps; Test('arctanh(-0.8)', -1.098612288668109691395245236922525704647490557822749451734);
  Eps; Test('arctanh(-0.9)', -1.472219489583220230004513715943926768618689630649564409268);
  Eps; Test('arctanh(-0.99)', -2.646652412362246197705060645934268600945552640284736249453);
  Eps; Test('arctanh(-0.999)', -3.800201167250200031775967267809004462263675695230499890921);
  Eps; Test('arctanh(-0.9999)', -4.951718775643043188629984385462344226560343044514956583248);
  Eps(1E-10); Test('arctanh(-0.99999)', -6.103033822758836843920183238794164167618256740528365250187);
  Eps(1E-10); Test('arctanh(-0.999999)', -7.254328619262047206741757083636344448506419115841743930067);
  Eps(1E-9); Test('arctanh(-0.9999999)', -8.405621390759131923771565318790248427277021942745416320706);
  Eps(1E-8); Test('arctanh(-0.99999999)', -9.556913959756155384530581858633211702983822375861884895776);
  Eps(1E-8); Test('arctanh(-0.999999999)', -10.70820650850317823272707760678789388490156093267314768587);
  Eps(1E-8); Test('arctanh(-0.9999999999)', -11.85949905522520107479794833415088848870992339574065914264);
  Eps(1E-8); Test('arctanh(-0.99999999999)', -13.01079160174472391680756281149309140501047492122692066190);
  Eps(1E-6); Test('arctanh(-0.999999999999)', -14.16208414824399675881656472633527352962352566561942433742);

  Test('arctanh(1, 2)', failure, 'Too many arguments.');

  Eps(1E-8); Test('arccoth(1.000000001)', 10.708206509003178232727077606829560551568227599346064);
  Eps(1E-8); Test('arccoth(1.00000001)', 9.5569139647561553845305819002998783696504896675285515);
  Eps(1E-9); Test('arccoth(1.0000001)', 8.4056214407591319237716069854569150940061886094120830);
  Eps(1E-10); Test('arccoth(1.000001)', 7.2543291192620472067834237503030173651730857836244820);
  Eps(1E-10); Test('arccoth(1.00001)', 6.1030388227588368855868499060858308342960841214809632);
  Eps; Test('arccoth(1.0001)', 4.9517687756430848552967135521291225003700838679280988);
  Eps; Test('arccoth(1.001)', 3.8007011672918667046926350505473167142921918932929646);
  Eps; Test('arccoth(1.01)', 2.6516524540295378755326586166431242276072117305223223);
  Eps; Test('arccoth(1.1)', 1.5222612188617114982502989901828527171422876437023053);
  Eps; Test('arccoth(1.5)', 0.8047189562170501873003796666130938197628006771342588);
  Eps; Test('arccoth(2)', 0.549306144334054845697622618461262852323745278911374725867);
  Eps; Test('arccoth(5/2)', 0.423648930193601806855053753260327012494797085879555868362);
  Eps; Test('arccoth(5)', 0.202732554054082190989006557732174568285995211731247098807);
  Eps; Test('arccoth(10)', 0.100335347731075580635726552060038945263362869145959135874);
  Eps; Test('arccoth(100)', 0.010000333353334762015882107551404223887097308807344335621);
  Eps; Test('arccoth(1000)', 0.001000000333333533333476190587301678210755133821800480624);
  Eps; Test('arccoth(1E4)', 0.000100000000333333335333333347619047730158731067821075513);
  Eps; Test('arccoth(1E5)', 0.000010000000000333333333353333333334761904762015873015882);
  Eps; Test('arccoth(1E6)', 1.0000000000003333333333335333333333334761904761905873E-6);
  Eps; Test('arccoth(1E7)', 1.0000000000000033333333333333533333333333334761904761E-7);
  Eps; Test('arccoth(1E8)', 1.0000000000000000333333333333333353333333333333334762E-8);
  Eps; Test('arccoth(1E9)', 1.0000000000000000003333333333333333335333333333333333E-9);
  Eps; Test('arccoth(1E10)', 1.0000000000000000000033333333333333333333533333333333E-10);
  Eps; Test('arccoth(1E20)', 1.0000000000000000000000000000000000000000333333333333E-20);
  Eps; Test('arccoth(1E50)', 1.000000000000000000000000000000000000000000000000000E-50);
  Eps; Test('arccoth(1E100)', 1.000000000000000000000000000000000000000000000000000E-100);
  Eps; Test('arccoth(1E200)', 1.000000000000000000000000000000000000000000000000000E-200);

  Eps(1E-8); Test('arccoth(-1.000000001)', -10.708206509003178232727077606829560551568227599346064);
  Eps(1E-8); Test('arccoth(-1.00000001)', -9.5569139647561553845305819002998783696504896675285515);
  Eps(1E-9); Test('arccoth(-1.0000001)', -8.4056214407591319237716069854569150940061886094120830);
  Eps(1E-10); Test('arccoth(-1.000001)', -7.2543291192620472067834237503030173651730857836244820);
  Eps(1E-10); Test('arccoth(-1.00001)', -6.1030388227588368855868499060858308342960841214809632);
  Eps; Test('arccoth(-1.0001)', -4.9517687756430848552967135521291225003700838679280988);
  Eps; Test('arccoth(-1.001)', -3.8007011672918667046926350505473167142921918932929646);
  Eps; Test('arccoth(-1.01)', -2.6516524540295378755326586166431242276072117305223223);
  Eps; Test('arccoth(-1.1)', -1.5222612188617114982502989901828527171422876437023053);
  Eps; Test('arccoth(-1.5)', -0.8047189562170501873003796666130938197628006771342588);
  Eps; Test('arccoth(-2)', -0.549306144334054845697622618461262852323745278911374725867);
  Eps; Test('arccoth(-5/2)', -0.423648930193601806855053753260327012494797085879555868362);
  Eps; Test('arccoth(-5)', -0.202732554054082190989006557732174568285995211731247098807);
  Eps; Test('arccoth(-10)', -0.100335347731075580635726552060038945263362869145959135874);
  Eps; Test('arccoth(-100)', -0.010000333353334762015882107551404223887097308807344335621);
  Eps; Test('arccoth(-1000)', -0.001000000333333533333476190587301678210755133821800480624);
  Eps; Test('arccoth(-1E4)', -0.000100000000333333335333333347619047730158731067821075513);
  Eps; Test('arccoth(-1E5)', -0.000010000000000333333333353333333334761904762015873015882);
  Eps; Test('arccoth(-1E6)', -1.0000000000003333333333335333333333334761904761905873E-6);
  Eps; Test('arccoth(-1E7)', -1.0000000000000033333333333333533333333333334761904761E-7);
  Eps; Test('arccoth(-1E8)', -1.0000000000000000333333333333333353333333333333334762E-8);
  Eps; Test('arccoth(-1E9)', -1.0000000000000000003333333333333333335333333333333333E-9);
  Eps; Test('arccoth(-1E10)', -1.0000000000000000000033333333333333333333533333333333E-10);
  Eps; Test('arccoth(-1E20)', -1.0000000000000000000000000000000000000000333333333333E-20);
  Eps; Test('arccoth(-1E50)', -1.000000000000000000000000000000000000000000000000000E-50);
  Eps; Test('arccoth(-1E100)', -1.000000000000000000000000000000000000000000000000000E-100);
  Eps; Test('arccoth(-1E200)', -1.000000000000000000000000000000000000000000000000000E-200);

  Test('arccoth(1, 2)', failure, 'Too many arguments.');

  Eps; Test('arcsech(1E-10)', 23.71899811050040214959464666830181864408650564564798501445);
  Eps; Test('arcsech(1E-9)', 21.41641301750635646532915521361745443639166353201921203836);
  Eps; Test('arcsech(1E-8)', 19.11382792451231075656116375893308929138431204339038697905);
  Eps; Test('arcsech(1E-7)', 16.81124283151826259754317230423935102128321050267833275302);
  Eps; Test('arcsech(1E-6)', 14.50865773852396941352518075581436181363002573279952559729);
  Eps; Test('arcsech(1E-5)', 12.20607264550517372950625189487994552274767082620203655486);
  Eps; Test('arcsech(1E-4)', 9.903487550036128036114197888112299723349695294604495026092);
  Eps; Test('arcsech(1E-3)', 7.600902209541988611419123117998223748151672423825895985802);
  Eps; Test('arcsech(1E-2)', 5.298292365610484590701666834943247168937251883140100722832);
  Eps; Test('arcsech(0.1)', 2.993222846126380897912667713774182913083660451180980642685);
  Eps; Test('arcsech(0.2)', 2.292431669561177687800787311348015431621868240015710247605);
  Eps; Test('arcsech(1/2)', 1.316957896924816708625046347307968444026981971467516479768);
  Eps; Test('arcsech(0.99)', 0.1420144407460770687457967942996676582857380154568418);
  Eps; Test('arcsech(0.9999)', 0.0141427249173917226895200406600229693658867021124513);
  Eps; Test('arcsech(1)', 0.0);

  Test('arcsech(1, 2)', failure, 'Too many arguments.');

  Eps; Test('arccsch(1E-10)', 23.71899811050040214959964666830181864408650564564798501445);
  Eps; Test('arccsch(1E-9)', 21.41641301750635646582915521361745443639166353201921203847);
  Eps; Test('arccsch(1E-8)', 19.11382792451231080656116375893308929138431204339049114572);
  Eps; Test('arccsch(1E-7)', 16.81124283151826759754317230423935102128321060684499941968);
  Eps; Test('arccsch(1E-6)', 14.50865773852446941352518075581436181373419239946619226396);
  Eps; Test('arccsch(1E-5)', 12.20607264555517372950625189488004968941433749286870371371);
  Eps; Test('arccsch(1E-4)', 9.903487555036128036114197992278966390016366883146161692759);
  Eps; Test('arccsch(1E-3)', 7.600902709541988611523289784664939633568339120417227831061);
  Eps; Test('arccsch(1E-2)', 5.298342365610588757368825689112906302142383535156218238322);
  Eps; Test('arccsch(0.1)', 2.998222950297969738846595537596453476607058054877303655734);
  Eps; Test('arccsch(1/2)', 1.443635475178810342493276740273105269405553003156981558983);
  Eps; Test('arccsch(1)', 0.881373587019543025232609324979792309028160328261635410753);
  Eps; Test('arccsch(2)', 0.481211825059603447497758913424368423135184334385660519661);
  Eps; Test('arccsch(3)', 0.327450150237258443322535259988258127700524528990767451275);
  Eps; Test('arccsch(4)', 0.247466461547263452944781549788359289253766903098567696469);
  Eps; Test('arccsch(5)', 0.198690110349241406474636915950206968221308794224453773021);
  Eps; Test('arccsch(10)', 0.099834078899207563327303124704769443267712911708825010742);
  Eps; Test('arccsch(50)', 0.019998666906609539360498967978559796611784199862956818634);
  Eps; Test('arccsch(100)', 0.009999833340832886935141612163953795982217152792954153612);
  Eps; Test('arccsch(1E3)', 0.000999999833333408333288690506572398262778896762002256268);
  Eps; Test('arccsch(1E4)', 0.000099999999833333334083333328869047649429563268341902889);
  Eps; Test('arccsch(1E5)', 9.999999999833333333340833333332886904761935143849204E-6);
  Eps; Test('arccsch(1E6)', 9.999999999998333333333334083333333332886904761905066E-7);
  Eps; Test('arccsch(1E7)', 9.999999999999983333333333333408333333333332886904762E-8);
  Eps; Test('arccsch(1E8)', 9.999999999999999833333333333333340833333333333332887E-9);
  Eps; Test('arccsch(1E9)', 9.999999999999999998333333333333333334083333333333333E-10);
  Eps; Test('arccsch(1E10)', 9.999999999999999999983333333333333333333408333333333E-11);
  Eps; Test('arccsch(1E50)', 1E-50);
  Eps; Test('arccsch(1E100)', 1E-100);
  Eps; Test('arccsch(1E200)', 1E-200);

  Eps; Test('arccsch(-1E-10)', -23.71899811050040214959964666830181864408650564564798501445);
  Eps; Test('arccsch(-1E-9)', -21.41641301750635646582915521361745443639166353201921203847);
  Eps; Test('arccsch(-1E-8)', -19.11382792451231080656116375893308929138431204339049114572);
  Eps; Test('arccsch(-1E-7)', -16.81124283151826759754317230423935102128321060684499941968);
  Eps; Test('arccsch(-1E-6)', -14.50865773852446941352518075581436181373419239946619226396);
  Eps; Test('arccsch(-1E-5)', -12.20607264555517372950625189488004968941433749286870371371);
  Eps; Test('arccsch(-1E-4)', -9.903487555036128036114197992278966390016366883146161692759);
  Eps; Test('arccsch(-1E-3)', -7.600902709541988611523289784664939633568339120417227831061);
  Eps; Test('arccsch(-1E-2)', -5.298342365610588757368825689112906302142383535156218238322);
  Eps; Test('arccsch(-0.1)', -2.998222950297969738846595537596453476607058054877303655734);
  Eps; Test('arccsch(-1/2)', -1.443635475178810342493276740273105269405553003156981558983);
  Eps; Test('arccsch(-1)', -0.881373587019543025232609324979792309028160328261635410753);
  Eps; Test('arccsch(-2)', -0.481211825059603447497758913424368423135184334385660519661);
  Eps; Test('arccsch(-3)', -0.327450150237258443322535259988258127700524528990767451275);
  Eps; Test('arccsch(-4)', -0.247466461547263452944781549788359289253766903098567696469);
  Eps; Test('arccsch(-5)', -0.198690110349241406474636915950206968221308794224453773021);
  Eps; Test('arccsch(-10)', -0.099834078899207563327303124704769443267712911708825010742);
  Eps; Test('arccsch(-50)', -0.019998666906609539360498967978559796611784199862956818634);
  Eps; Test('arccsch(-100)', -0.009999833340832886935141612163953795982217152792954153612);
  Eps; Test('arccsch(-1E3)', -0.000999999833333408333288690506572398262778896762002256268);
  Eps; Test('arccsch(-1E4)', -0.000099999999833333334083333328869047649429563268341902889);
  Eps; Test('arccsch(-1E5)', -9.999999999833333333340833333332886904761935143849204E-6);
  Eps; Test('arccsch(-1E6)', -9.999999999998333333333334083333333332886904761905066E-7);
  Eps; Test('arccsch(-1E7)', -9.999999999999983333333333333408333333333332886904762E-8);
  Eps; Test('arccsch(-1E8)', -9.999999999999999833333333333333340833333333333332887E-9);
  Eps; Test('arccsch(-1E9)', -9.999999999999999998333333333333333334083333333333333E-10);
  Eps; Test('arccsch(-1E10)', -9.999999999999999999983333333333333333333408333333333E-11);
  Eps; Test('arccsch(-1E50)', -1E-50);
  Eps; Test('arccsch(-1E100)', -1E-100);
  Eps; Test('arccsch(-1E200)', -1E-200);

  Test('arccsch(1, 2)', failure, 'Too many arguments.');

  // Absolute value

  Test('abs(0)', 0);

  Test('abs(1)', 1);
  Test('abs(2)', 2);
  Test('abs(3)', 3);
  Test('abs(4)', 4);
  Test('abs(5)', 5);
  Test('abs(6)', 6);
  Test('abs(7)', 7);
  Test('abs(8)', 8);
  Test('abs(9)', 9);
  Test('abs(10)', 10);
  Test('abs(50)', 50);
  Test('abs(100)', 100);
  Test('abs(1000)', 1000);
  Test('abs(10000)', 10000);
  Test('abs(100000)', 100000);
  Test('abs(1000000)', 1000000);
  Test('abs(10000000)', 10000000);
  Test('abs(100000000)', 100000000);
  Test('abs(1000000000)', 1000000000);
  Test('abs(10000000000)', 10000000000);
  Test('abs(100000000000)', 100000000000);
  Test('abs(1000000000000)', 1000000000000);
  Test('abs(10000000000000)', 10000000000000);
  Test('abs(100000000000000)', 100000000000000);
  Test('abs(1000000000000000)', 1000000000000000);
  Test('abs(10000000000000000)', 10000000000000000);
  Test('abs(100000000000000000)', 100000000000000000);
  Test('abs(1000000000000000000)', 1000000000000000000);
  Test('abs(8000000000000000000)', 8000000000000000000);

  Test('abs(-1)', 1);
  Test('abs(-2)', 2);
  Test('abs(-3)', 3);
  Test('abs(-4)', 4);
  Test('abs(-5)', 5);
  Test('abs(-6)', 6);
  Test('abs(-7)', 7);
  Test('abs(-8)', 8);
  Test('abs(-9)', 9);
  Test('abs(-10)', 10);
  Test('abs(-50)', 50);
  Test('abs(-100)', 100);
  Test('abs(-1000)', 1000);
  Test('abs(-10000)', 10000);
  Test('abs(-100000)', 100000);
  Test('abs(-1000000)', 1000000);
  Test('abs(-10000000)', 10000000);
  Test('abs(-100000000)', 100000000);
  Test('abs(-1000000000)', 1000000000);
  Test('abs(-10000000000)', 10000000000);
  Test('abs(-100000000000)', 100000000000);
  Test('abs(-1000000000000)', 1000000000000);
  Test('abs(-10000000000000)', 10000000000000);
  Test('abs(-100000000000000)', 100000000000000);
  Test('abs(-1000000000000000)', 1000000000000000);
  Test('abs(-10000000000000000)', 10000000000000000);
  Test('abs(-100000000000000000)', 100000000000000000);
  Test('abs(-1000000000000000000)', 1000000000000000000);
  Test('abs(-8000000000000000000)', 8000000000000000000);

  Test('abs(1/2 - 1/2)', Rat(0, 1));

  Test('abs(1/2)', Rat(1, 2));
  Test('abs(7/29)', Rat(7, 29));
  Test('abs(811/13)', Rat(811, 13));
  Test('abs(-1/2)', Rat(1, 2));
  Test('abs(-7/29)', Rat(7, 29));
  Test('abs(-811/13)', Rat(811, 13));

  Test('abs(0.0)', 0.0);

  Test('abs(0.123456789)', 0.123456789);
  Test('abs(1.0)', 1.0);
  Test('abs(1.1)', 1.1);
  Test('abs(1.5)', 1.5);
  Test('abs(2.0)', 2.0);
  Test('abs(3.0)', 3.0);
  Test('abs(4.0)', 4.0);
  Test('abs(5.0)', 5.0);
  Test('abs(6.0)', 6.0);
  Test('abs(7.0)', 7.0);
  Test('abs(7.123456)', 7.123456);
  Test('abs(8.0)', 8.0);
  Test('abs(9.0)', 9.0);
  Test('abs(10.0)', 10.0);
  Test('abs(50.0)', 50.0);
  Test('abs(100.0)', 100.0);
  Test('abs(1000.0)', 1000.0);
  Test('abs(10000.0)', 10000.0);
  Test('abs(100000.0)', 100000.0);
  Test('abs(1000000.0)', 1000000.0);
  Test('abs(10000000.0)', 10000000.0);
  Test('abs(100000000.0)', 100000000.0);
  Test('abs(1000000000.0)', 1000000000.0);
  Test('abs(10000000000.0)', 10000000000.0);
  Test('abs(100000000000.0)', 100000000000.0);
  Test('abs(1000000000000.0)', 1000000000000.0);
  Test('abs(10000000000000.0)', 10000000000000.0);
  Test('abs(100000000000000.0)', 100000000000000.0);
  Test('abs(1000000000000000.0)', 1000000000000000.0);
  Test('abs(10000000000000000.0)', 10000000000000000.0);
  Test('abs(100000000000000000.0)', 100000000000000000.0);
  Test('abs(1000000000000000000.0)', 1000000000000000000.0);
  Test('abs(8000000000000000000.0)', 8000000000000000000.0);
  Test('abs(1E50)', 1E50);
  Test('abs(1E100)', 1E100);
  Test('abs(1E200)', 1E200);

  Test('abs(1E-50)', 1E-50);
  Test('abs(1E-100)', 1E-100);
  Test('abs(1E-200)', 1E-200);

  Test('abs(-0.123456789)', 0.123456789);
  Test('abs(-1.0)', 1.0);
  Test('abs(-1.1)', 1.1);
  Test('abs(-1.5)', 1.5);
  Test('abs(-2.0)', 2.0);
  Test('abs(-3.0)', 3.0);
  Test('abs(-4.0)', 4.0);
  Test('abs(-5.0)', 5.0);
  Test('abs(-6.0)', 6.0);
  Test('abs(-7.0)', 7.0);
  Test('abs(-7.123456)', 7.123456);
  Test('abs(-8.0)', 8.0);
  Test('abs(-9.0)', 9.0);
  Test('abs(-10.0)', 10.0);
  Test('abs(-50.0)', 50.0);
  Test('abs(-100.0)', 100.0);
  Test('abs(-1000.0)', 1000.0);
  Test('abs(-10000.0)', 10000.0);
  Test('abs(-100000.0)', 100000.0);
  Test('abs(-1000000.0)', 1000000.0);
  Test('abs(-10000000.0)', 10000000.0);
  Test('abs(-100000000.0)', 100000000.0);
  Test('abs(-1000000000.0)', 1000000000.0);
  Test('abs(-10000000000.0)', 10000000000.0);
  Test('abs(-100000000000.0)', 100000000000.0);
  Test('abs(-1000000000000.0)', 1000000000000.0);
  Test('abs(-10000000000000.0)', 10000000000000.0);
  Test('abs(-100000000000000.0)', 100000000000000.0);
  Test('abs(-1000000000000000.0)', 1000000000000000.0);
  Test('abs(-10000000000000000.0)', 10000000000000000.0);
  Test('abs(-100000000000000000.0)', 100000000000000000.0);
  Test('abs(-1000000000000000000.0)', 1000000000000000000.0);
  Test('abs(-8000000000000000000.0)', 8000000000000000000.0);
  Test('abs(-1E50)', 1E50);
  Test('abs(-1E100)', 1E100);
  Test('abs(-1E200)', 1E200);

  Test('abs(-1E-50)', 1E-50);
  Test('abs(-1E-100)', 1E-100);
  Test('abs(-1E-200)', 1E-200);

  Test('abs(2, 3)', failure, 'Too many arguments.');
  Test('abs("cat")', failure, 'An object of type numeric entity was expected as argument 1, but an object of type string was given.');

  // Misc elementary functions

  Test('sinc(0)', 1.0);
  Test('sinc(0.0)', 1.0);
  Test('sinc(1/2-1/2)', 1.0);

  for _i := -100 to 100 do
  begin
    if _i = 0 then
      Continue;
    _x := _i / 10;
    Eps; Test(Format('sinc(%g)', [_x], DefaultFormatSettings), Sin(TASR(_x)) / _x);
  end;

  // Misc simple functions

  Test('Heaviside(-1E50)', 0.0);
  Test('Heaviside(-5.7)', 0.0);
  Test('Heaviside(-5/8)', 0.0);
  Test('Heaviside(-5)', 0.0);
  Test('Heaviside(-3)', 0.0);
  Test('Heaviside(0)', 0.5);
  Test('Heaviside(1/2 - 1/2)', 0.5);
  Test('Heaviside(0.0)', 0.5);
  Test('Heaviside(3)', 1.0);
  Test('Heaviside(5)', 1.0);
  Test('Heaviside(5.3)', 1.0);
  Test('Heaviside(5/3)', 1.0);
  Test('Heaviside(1E50)', 1.0);

  Test('Heaviside(1 + i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');

  Test('ramp(-1E50)', 0.0);
  Test('ramp(-123.456)', 0.0);
  Test('ramp(-4)', 0.0);
  Test('ramp(-4.5)', 0.0);
  Test('ramp(-4/9)', 0.0);
  Test('ramp(0)', 0.0);
  Test('ramp(1/2-1/2)', 0.0);
  Test('ramp(0.0)', 0.0);

  Test('ramp(0.123)', 0.123);
  Test('ramp(1/2)', 1/2);
  Test('ramp(1)', 1.0);
  Test('ramp(21)', 21.0);
  Test('ramp(123)', 123.0);
  Test('ramp(123.456)', 123.456);
  Test('ramp(21E50)', 21E50);

  Test('ramp(1 + i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');

  Test('rect(-123E50)', 0.0);
  Test('rect(-123)', 0.0);
  Test('rect(-123/7)', 0.0);
  Test('rect(-123.456)', 0.0);
  Test('rect(-0.7)', 0.0);
  Test('rect(-0.6)', 0.0);

  Test('rect(-0.5)', 0.5);
  Test('rect(-1/2)', 0.5);

  Test('rect(-0.4)', 1.0);
  Test('rect(-0.3)', 1.0);
  Test('rect(-0.2)', 1.0);
  Test('rect(-1/10)', 1.0);
  Test('rect(0)', 1.0);
  Test('rect(1/2-1/2)', 1.0);
  Test('rect(0.0)', 1.0);
  Test('rect(0.1)', 1.0);
  Test('rect(1/10)', 1.0);
  Test('rect(0.2)', 1.0);

  Test('rect(1/2)', 0.5);
  Test('rect(0.5)', 0.5);

  Test('rect(0.6)', 0.0);
  Test('rect(0.7)', 0.0);
  Test('rect(4/5)', 0.0);
  Test('rect(1)', 0.0);
  Test('rect(2)', 0.0);
  Test('rect(9)', 0.0);
  Test('rect(9E50)', 0.0);

  Test('rect(1 + i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');

  Test('tri(-5E50)', 0.0);
  Test('tri(-5)', 0.0);
  Test('tri(-4.5)', 0.0);
  Test('tri(-11/3)', 0.0);
  Test('tri(-1.2)', 0.0);
  Test('tri(-1.0)', 0.0);
  Test('tri(-1)', 0.0);
  Eps; Test('tri(-0.9)', 0.1);
  Eps; Test('tri(-0.8)', 0.2);
  Eps; Test('tri(-0.7)', 0.3);
  Eps; Test('tri(-1/2)', 0.5);
  Eps; Test('tri(-0.1)', 0.9);
  Eps; Test('tri(0)', 1.0);
  Eps; Test('tri(0.0)', 1.0);
  Eps; Test('tri(1/2-1/2)', 1.0);
  Eps; Test('tri(0.1)', 0.9);
  Eps; Test('tri(0.2)', 0.8);
  Eps; Test('tri(1/3)', 2/3);
  Eps; Test('tri(1/2)', 0.5);
  Eps; Test('tri(0.5)', 0.5);
  Eps; Test('tri(0.7)', 0.3);
  Eps; Test('tri(0.8)', 0.2);
  Eps; Test('tri(0.99)', 0.01);
  Test('tri(1)', 0.0);
  Test('tri(3/4 + 1/4)', 0.0);
  Test('tri(1.0)', 0.0);
  Test('tri(1.1)', 0.0);
  Test('tri(1.4)', 0.0);
  Test('tri(3/2)', 0.0);
  Test('tri(2)', 0.0);
  Test('tri(123)', 0.0);
  Test('tri(123E50)', 0.0);
  Test('tri(1 + i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');

  for _i := -100 to 100 do
  begin

    Test(Format('SquareWave(0.1 + 2⋅π⋅%d)', [_i]), 1.0);
    Test(Format('SquareWave(1.2 + 2⋅π⋅%d)', [_i]), 1.0);
    Test(Format('SquareWave(2.5 + 2⋅π⋅%d)', [_i]), 1.0);
    Test(Format('SquareWave(3.1415 + 2⋅π⋅%d)', [_i]), 1.0);
    Test(Format('SquareWave(3.1416 + 2⋅π⋅%d)', [_i]), -1.0);
    Test(Format('SquareWave(4 + 2⋅π⋅%d)', [_i]), -1.0);
    Test(Format('SquareWave(5.5 + 2⋅π⋅%d)', [_i]), -1.0);
    Test(Format('SquareWave(6.28 + 2⋅π⋅%d)', [_i]), -1.0);
    Test(Format('SquareWave(7 + 2⋅π⋅%d)', [_i]), 1.0);

    Eps; Test(Format('TriangleWave(0.1 + 2⋅π⋅%d)', [_i]), 0.0636619772367581343);
    Eps; Test(Format('TriangleWave(1.2 + 2⋅π⋅%d)', [_i]), 0.763943726841097612);
    Eps; Test(Format('TriangleWave(2.5 + 2⋅π⋅%d)', [_i]), 0.408450569081046642);
    Eps; Test(Format('TriangleWave(3.1415 + 2⋅π⋅%d)', [_i]), 5.89851072432107364E-5);
    Eps; Test(Format('TriangleWave(3.1416 + 2⋅π⋅%d)', [_i]), -4.67686999354737128E-6);
    Eps; Test(Format('TriangleWave(4 + 2⋅π⋅%d)', [_i]), -0.546479089470325372);
    Eps; Test(Format('TriangleWave(5.5 + 2⋅π⋅%d)', [_i]), -0.498591251978302613);
    Eps; Test(Format('TriangleWave(6.28 + 2⋅π⋅%d)', [_i]), -0.00202782953158916535);
    Eps; Test(Format('TriangleWave(7 + 2⋅π⋅%d)', [_i]), 0.456338406573069402);

    Eps; Test(Format('SawtoothWave(0.1 + 2⋅π⋅%d)', [_i]), 0.0318309886183790672);
    Eps; Test(Format('SawtoothWave(1.2 + 2⋅π⋅%d)', [_i]), 0.381971863420548806);
    Eps; Test(Format('SawtoothWave(2.5 + 2⋅π⋅%d)', [_i]), 0.795774715459476679);
    Eps; Test(Format('SawtoothWave(3.1415 + 2⋅π⋅%d)', [_i]), 0.999970507446378395);
    Eps; Test(Format('SawtoothWave(3.1416 + 2⋅π⋅%d)', [_i]), -0.999997661565003226);
    Eps; Test(Format('SawtoothWave(4 + 2⋅π⋅%d)', [_i]), -0.726760455264837314);
    Eps; Test(Format('SawtoothWave(5.5 + 2⋅π⋅%d)', [_i]), -0.249295625989151307);
    Eps; Test(Format('SawtoothWave(6.28 + 2⋅π⋅%d)', [_i]), -0.00101391476579458268);
    Eps; Test(Format('SawtoothWave(7 + 2⋅π⋅%d)', [_i]), 0.228169203286534701);

  end;

  Test('percent(0)', 0.0);
  Test('percent(0.0)', 0.0);
  Test('percent(1/2-1/2)', 0.0);
  Test('percent(i-i)', TASC(0.0));
  Eps; Test('percent(1)', 0.01);
  Eps; Test('percent(7.5)', 0.075);
  Eps; Test('percent(99)', 0.99);
  Eps; Test('percent(99.9)', 0.999);
  Eps; Test('percent(120)', 1.2);
  Eps; Test('percent(150)', 1.5);
  Eps; Test('percent(200)', 2.0);
  Eps; Test('percent(1/2)', 0.005);
  Eps; Test('percent(-23)', -0.23);
  Eps; Test('percent(3 + i)', 0.03 + 0.01*ImaginaryUnit);
  Test('percent("cat")', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');

  Test('permille(0)', 0.0);
  Test('permille(0.0)', 0.0);
  Test('permille(1/2-1/2)', 0.0);
  Test('permille(i-i)', TASC(0.0));
  Eps; Test('permille(1)', 0.001);
  Eps; Test('permille(500)', 0.5);
  Eps; Test('permille(999)', 0.999);
  Eps; Test('permille(999.9)', 0.9999);
  Eps; Test('permille(1/2)', 0.0005);
  Eps; Test('permille(1200)', 1.2);
  Eps; Test('permille(-1500)', -1.5);
  Eps; Test('permille(3 - 10⋅i)', 0.003 - 0.01*ImaginaryUnit);
  Test('permille("cat")', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');

  Test('degrees(0)', 0.0);
  Test('degrees(0.0)', 0.0);
  Test('degrees(1/2-1/2)', 0.0);
  Test('degrees(i-i)', TASC(0.0));
  Eps; Test('degrees(30.0)', Pi/6);
  Eps; Test('degrees(45)', Pi/4);
  Eps; Test('degrees(60)', Pi/3);
  Eps; Test('degrees(90)', Pi/2);
  Eps; Test('degrees(180)', Pi);
  Eps; Test('degrees(-45)', -Pi/4);
  Eps; Test('degrees(360)', 2*Pi);
  Eps; Test('degrees(720)', 4*Pi);
  Eps; Test('degrees(-1/2)', -Pi/360);
  Eps; Test('degrees(45 + 180⋅i)', Pi/4 + Pi*ImaginaryUnit);
  Test('degrees("cat")', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');

  // Piecewise

  Test('f ≔ x ↦ piecewise(−10, x < −5, 0, x < −2, 1, x ≤ 0, 5, otherwise);', null);
  Test('f(-10)', -10);
  Test('f(-8)', -10);
  Test('f(-6)', -10);
  Test('f(-5)', 0);
  Test('f(-4)', 0);
  Test('f(-3)', 0);
  Test('f(-2)', 1);
  Test('f(-1)', 1);
  Test('f(0)', 1);
  Test('f(1)', 5);
  Test('f(2)', 5);
  Test('f(5)', 5);
  Test('f(10)', 5);

  Test('f ≔ x ↦ piecewise(−x, x < 0, x, otherwise);', null);
  Test('f(5)', 5);
  Test('f(3.52)', 3.52);
  Test('f(-1.19)', 1.19);
  Test('f(3.1415)', 3.1415);
  Test('f(-3.1415)', 3.1415);
  Test('f(-10E20)', 10E20);
  Test('f(5E-3)', 5E-3);
  Test('f(0)', 0);

  Test('f ≔ x ↦ piecewise(sin(x)/x, x ≠ 0, 1, otherwise);', null);
  Test('f(7.5)', sinc(7.5));
  Test('f(-3.2)', sinc(-3.2));
  Test('f(10.201)', sinc(10.201));
  Test('f(0)', 1);
  Test('f(1.22)', sinc(1.22));

  Test('f ≔ x ↦ piecewise(x, even(#x), "odd", otherwise);', null);
  Test('f("test")', 'test');
  Test('f("nargle")', 'nargle');
  Test('f("hippogriff")', 'hippogriff');
  Test('f("wrackspurt")', 'wrackspurt');
  Test('f("thestral")', 'thestral');
  Test('f("unicorn")', 'odd');
  Test('f("owl")', 'odd');
  Test('f("boggart")', 'odd');
  Test('f("hinkypunk")', 'odd');

  Test('delete(f)', success);

  // Additional tests

  Eps; Test('arcsin(arccsch(coth(arccos(√(sin(e)^2 + cos(e/2)^2)))))', 0.819175391468022550591124065781692738450290782438187087868);
  Eps; Test('arcsin(arccsch(10⋅coth(arccos(√(sin(e)^2 + cos(e/2)^2))))/π)', 0.025355746351156947652467226401424265852418977020860741394);


  //
  // Elementary functions of a complex variable
  //

  // Complex exponential function

  Eps; Test('exp(i)', Cos(1) + ImaginaryUnit * Sin(1));
  Eps; Test('exp(5⋅i)', Cos(5) + ImaginaryUnit * Sin(5));
  Eps; Test('exp(2 + 5⋅i)', Exp(2) * (Cos(5) + ImaginaryUnit * Sin(5)));
  Eps; Test('exp(2 + 0⋅i)', TASC(Exp(2)));
  Eps; Test('exp(-3 - 7⋅i)', Exp(-3) * (Cos(-7) + ImaginaryUnit * Sin(-7)));
  Eps; Test('exp(-10 + 2⋅i)', Exp(-10) * (Cos(2) + ImaginaryUnit * Sin(2)));

  Eps; Test('e^i', Cos(1) + ImaginaryUnit * Sin(1));
  Eps; Test('e^(5⋅i)', Cos(5) + ImaginaryUnit * Sin(5));
  Eps; Test('e^(2 + 5⋅i)', Exp(2) * (Cos(5) + ImaginaryUnit * Sin(5)));
  Eps; Test('e^(2 + 0⋅i)', TASC(Exp(2)));
  Eps; Test('e^(-3 - 7⋅i)', Exp(-3) * (Cos(-7) + ImaginaryUnit * Sin(-7)));
  Eps; Test('e^(-10 + 2⋅i)', Exp(-10) * (Cos(2) + ImaginaryUnit * Sin(2)));

  // Squares and square roots

  Eps; Test('sqr(i)', TASC(-1.0));
  Eps; Test('sqr(1 + i)', 2*ImaginaryUnit);
  Eps; Test('sqr(3 + 0⋅i)', TASC(9));

  Eps; Test('i^2', TASC(-1.0));
  Eps; Test('(1 + i)^2', 2*ImaginaryUnit);
  Eps; Test('(3 + 0⋅i)^2', TASC(9));

  Eps; Test('sqrt(i)', 1/Sqrt(2) + ImaginaryUnit/Sqrt(2));
  Eps; Test('sqrt(1 + i)', 1.0986841134678099660398011952406783785443931209271577437 + 0.45508986056222734130435775782246856962019037848315009258*ImaginaryUnit);
  Eps; Test('sqrt(25 + 0⋅i)', TASC(5));
  Eps; Test('sqrt(-25)', 5*ImaginaryUnit);

  Eps; Test('√i', 1/Sqrt(2) + ImaginaryUnit/Sqrt(2));
  Eps; Test('√(1 + i)', 1.0986841134678099660398011952406783785443931209271577437 + 0.45508986056222734130435775782246856962019037848315009258*ImaginaryUnit);
  Eps; Test('√(25 + 0⋅i)', TASC(5));
  Eps; Test('√-25', 5*ImaginaryUnit);

  Eps; Test('i^(1/2)', 1/Sqrt(2) + ImaginaryUnit/Sqrt(2));
  Eps; Test('(1 + i)^(1/2)', 1.0986841134678099660398011952406783785443931209271577437 + 0.45508986056222734130435775782246856962019037848315009258*ImaginaryUnit);
  Eps; Test('(25 + 0⋅i)^(1/2)', TASC(5));
  Eps; Test('(-25)^(1/2)', 5*ImaginaryUnit);

  // Complex logarithm (principal branch)

  Eps; Test('ln(i)', Pi/2 * ImaginaryUnit);
  Eps; Test('ln(-1)', Pi * ImaginaryUnit);
  Eps; Test('ln(-i)', -Pi/2 * ImaginaryUnit);
  Eps; Test('ln(1 + i)', Ln(Sqrt(2)) + Pi/4 * ImaginaryUnit);
  Eps; Test('ln(-3 + 5⋅i)', Ln(Sqrt(9 + 25)) + (Pi/2 + ArcTan(TASR(3/5))) * ImaginaryUnit);
  Eps; Test('ln(7 + 0⋅i)', TASC(Ln(7)));
  Test('ln(0 + 0⋅i)', failure);

  Eps; Test('log(i)', (Pi/2 * ImaginaryUnit) / Ln(10));
  Eps; Test('log(-1)', Pi * ImaginaryUnit / Ln(10));
  Eps; Test('log(1 + i)', (Ln(Sqrt(2)) + Pi/4 * ImaginaryUnit) / Ln(10));

  // Complex trigonometric functions

  Eps; Test('sin(i)', (Exp(-1) - Exp(1)) / (2*ImaginaryUnit));
  Eps; Test('sin(-i)', (Exp(1) - Exp(-1)) / (2*ImaginaryUnit));
  Eps; Test('sin(3 + 5⋅i)', 10.4725085339403922766733225368535032711264199503880135285 - 73.4606216956736763667911925050817504072139228144746370129*ImaginaryUnit);
  Eps; Test('sin(-7 + 2⋅i)', -2.4717121507995191847930660445841074964134826715188594020 + 2.7342982376643447360299256844111515022153999388350714301*ImaginaryUnit);
  Eps; Test('sin(12 - 17⋅i)', -6.480446741574280814816562751675775997334405457597648E6 - 1.019162625205039564741173947560666935396866815626899E7*ImaginaryUnit);
  Eps; Test('sin(1.2 + 0⋅i)', TASC(Sin(TASR(1.2))));

  Eps; Test('cos(i)', TASC(Exp(-1) + Exp(1)) / 2);
  Eps; Test('cos(-i)', TASC(Exp(1) + Exp(-1)) / 2);
  Eps; Test('cos(3 + i)', -1.527638250116543444482360747972750291068812349268136176 - 0.1658444019189788416185872450717075757108286108502988833*ImaginaryUnit);
  Eps; Test('cos(4 + 11⋅i)', -19568.175398869946707199055477095770480641155637545146163 + 22656.449920921169676201134323269640536589365445079209622*ImaginaryUnit);
  Eps; Test('cos(4 - 3⋅i)', -6.5806630405511564325607441265388036167112673455158977732 - 7.5815527427465443537163452865384260093875275909488528129*ImaginaryUnit);
  Eps; Test('cos(-1 + 10⋅i)', 5950.47514179473124548048736467827488569449324482969186870 + 9267.31591299536549970566883311875637513437382939959314367*ImaginaryUnit);
  Eps; Test('cos(-4 - 3⋅i)', -6.5806630405511564325607441265388036167112673455158977732 + 7.5815527427465443537163452865384260093875275909488528129*ImaginaryUnit);
  Eps; Test('cos(-2 + 0⋅i)', TASC(Cos(TASR(-2))));

  Eps; Test('tan(i)', (Exp(1) - Exp(-1)) / (Exp(1) + Exp(-1)) * ImaginaryUnit);
  Eps; Test('tan(2 + 3⋅i)', -0.0037640256415042482927512211303226908396306202016580864 + 1.0032386273536098014463585978219272598077897241071003*ImaginaryUnit);
  Eps; Test('tan(-3 + 8⋅i)', 6.288813022767552217559561635063193878124352096306372E-8 + 0.9999997838941594365304882856089471795699421781331468*ImaginaryUnit);
  Eps; Test('tan(1 - i)', 0.27175258531951171652884372249858892070946411146177945309 - 1.0839233273386945434757520612119717213449675274753898563*ImaginaryUnit);
  Eps; Test('tan(-5 - 6⋅i)', 6.685231390277019376683702646896512582735741129324424E-6 - 1.000010310898119777267799874225580594599226881899293*ImaginaryUnit);
  Eps; Test('tan(2 + 0⋅i)', TASC(Tan(TASR(2))));

  Eps; Test('cot(i)', - (Exp(1) + Exp(-1)) / (Exp(1) - Exp(-1)) * ImaginaryUnit);
  Eps; Test('cot(1 + 4⋅i)', 0.00060990025382222757285631662318375227678357820322110016 - 0.99972064953393074903673898504649165658496308506765458*ImaginaryUnit);
  Eps; Test('cot(1 - i)', 0.21762156185440268136513424360523807352075436916785404091 + 0.86801414289592494863584920891627388827343874994609327121*ImaginaryUnit);
  Eps; Test('cot(-3 - 5⋅i)', 0.00002537310004454597738376334678946965675405003735598581 + 1.000087186805896774328531688104521857713161283189077*ImaginaryUnit);
  Eps; Test('cot(-3 + 0⋅i)', TASC(Cot(TASR(-3))));

  Eps; Test('sec(i)', 2 / TASC(Exp(-1) + Exp(1)));
  Eps; Test('sec(-i)', 2 / TASC(Exp(1) + Exp(-1)));
  Eps; Test('sec(1 + i)', 0.49833703055518678521380589177216953443287793247109398814 + 0.59108384172104504805039169297433507150365868212588646424*ImaginaryUnit);
  Eps; Test('sec(1 - i)', 0.49833703055518678521380589177216953443287793247109398814 - 0.59108384172104504805039169297433507150365868212588646424*ImaginaryUnit);
  Eps; Test('sec(10 + 3⋅i)', -0.0835873060411529967625937748938324021264546729173126659 - 0.0539267285969059792466480103788546294283694271488789939*ImaginaryUnit);
  Eps; Test('sec(-10 + 0⋅i)', TASC(Sec(TASR(-10))));

  Eps; Test('csc(i)', (2*ImaginaryUnit) / (Exp(-1) - Exp(1)));
  Eps; Test('csc(-i)', (2*ImaginaryUnit) / (Exp(1) - Exp(-1)));
  Eps; Test('csc(-3 + 6⋅i)', -0.0006996156075332090800431212818069060533041247773525188 + 0.004907919864796477925059893863335768820975451569217730*ImaginaryUnit);
  Eps; Test('csc(-3 - 2⋅i)', -0.0403005788568915218751324795428698677808424754242681263 - 0.272548661462940199512498477932708924053539864495446014*ImaginaryUnit);
  Eps; Test('csc(5 - i)', -0.6431695477297066634515600499073498501588588193164601810 + 0.1448992736120180695021103750974513144270986152547713184*ImaginaryUnit);
  Eps; Test('csc(5 + i/3)', -0.978614127392489059539368178720765073632608455227481827 - 0.0930736608190462454007343145164138993466562018858647873*ImaginaryUnit);
  Eps; Test('csc(5 + 0⋅i)', TASC(Csc(TASR(5))));

  Eps; Test('sin(cos(tan(i)))', TASC(0.964700726543061195952074901399353982125433421940661228439));
  Eps; Test('sin(cos(tan(i - cot(2 - sec(3 - csc(4 + i))))))', 0.99947576215872888090100797214952990182876247189561699 - 0.00052935905324478176387128818016362908140290415574362707*ImaginaryUnit);
  Eps; Test('sin(1 + cos(1 + tan(1 + i)))', 1.8829620368282747450196685869873148055712749579883178254 - 0.13694664035900467885407097870352858896326713080974450332*ImaginaryUnit);

  // Inverse complex trigonometric functions

  Eps; Test('arcsin(i)', 0.881373587019543025232609324979792309028160328261635410753*ImaginaryUnit);
  Eps; Test('arcsin(2 + i)', 1.06344002357775205618949199708955100285184240862790481816 + 1.46935174436818527325584431736164761678780335733478817287*ImaginaryUnit);
  Eps; Test('arcsin(5 - i/2)', 1.46910811624953398156705597913659617349950317338046023488 - 2.29771708098104931807962984598024793598189646836182468243*ImaginaryUnit);
  Eps; Test('arcsin(2)', 1.57079632679489661923132169163975144209858469968755291048 - 1.31695789692481670862504634730796844402698197146751647976*ImaginaryUnit);
  Eps; Test('arcsin(-3)', -1.5707963267948966192313216916397514420985846996875529104 + 1.7627471740390860504652186499595846180563206565232708215*ImaginaryUnit);
  Eps; Test('arcsin(-2 + 3⋅i)', -0.5706527843210994007102838796856696501828032450960401365 + 1.983387029916535432347076902894039565014248302909345356*ImaginaryUnit);
  Eps; Test('arcsin(-0.5 + 0⋅i)', TASC(-Pi/6));

  Eps; Test('arccos(i)', 1.5707963267948966192313216916397514420985846996875529104 - 0.88137358701954302523260932497979230902816032826163541075*ImaginaryUnit);
  Eps; Test('arccos(-i)', 1.5707963267948966192313216916397514420985846996875529104 + 0.88137358701954302523260932497979230902816032826163541075*ImaginaryUnit);
  Eps; Test('arccos(1 - i)', 0.90455689430238136412731679566195872143109456096160506765 + 1.0612750619050356520330189162135734858067854989386336963*ImaginaryUnit);
  Eps; Test('arccos(2 - 3⋅i)', 1.00014354247379721852103781195408179191578145459151277395 + 1.98338702991653543234707690289403956501424830290934535612*ImaginaryUnit);
  Eps; Test('arccos(1 + i/5)', 0.43944584185260459623793162127732986501155823842528417214 - 0.45431323077957737705529998828019449802364511931170001629*ImaginaryUnit);
  Eps; Test('arccos(5)', 2.292431669561177687800787311348015431621868240015710247605*ImaginaryUnit);
  Eps; Test('arccos(-10)', 3.14159265358979323846264338327950288419716939937510582097 - 2.99322284612638089791266771377418291308366045118098064268*ImaginaryUnit);
  Eps; Test('arccos(1/2 + 0⋅i)', TASC(Pi/3));

  Eps; Test('arctan(2⋅i)', 1.5707963267948966192313216916397514420985846996875529104 + 0.54930614433405484569762261846126285232374527891137472586*ImaginaryUnit);
  Eps; Test('arctan(1 + 3⋅i/2)', 1.2074751564540338108745829199556994743122637338158870059 + 0.43946447938809341314562817478393099149222788991355658925*ImaginaryUnit);
  Eps; Test('arctan(2 - i/2)', 1.126556440834822348743875836675452217620757354939304685 - 0.09641562020299616723799601214481218586574992398782266413*ImaginaryUnit);
  Eps; Test('arctan(-3 + 5⋅i)', -1.480869576898657483937045193011697324092097931272956850 + 0.1469466662255297520474327851547159424423449403442452953*ImaginaryUnit);
  Eps; Test('arctan(10 + 0⋅i)', TASC(ArcTan(TASR(10))));

  Eps; Test('arccot(i/2)', -1.570796326794896619231321691639751442098584699687552910 - 0.5493061443340548456976226184612628523237452789113747258*ImaginaryUnit);
  Eps; Test('arccot(2 + 5⋅i/2)', 0.20407455187754464032852904168448859579436685461336611733 - 0.23887786125685909036318202708478274131991664762292234862*ImaginaryUnit);
  Eps; Test('arccot(2 - 3⋅i)', 0.16087527719832109670070230717933065951037764777882809571 + 0.22907268296853876629588180294200276786252530497706561694*ImaginaryUnit);
  Eps; Test('arccot(-7 + 5⋅i)', -0.0947400789405736593074421665595193002371703766608980189 - 0.0670659966486698360490117940769519826957337669563844177*ImaginaryUnit);
  Eps; Test('arccot(-7 + 0⋅i)', TASC(ArcCot(TASR(-7))));

  Eps; Test('arcsec(i)', 1.5707963267948966192313216916397514420985846996875529104 + 0.88137358701954302523260932497979230902816032826163541075*ImaginaryUnit);
  Eps; Test('arcsec(1 + i)', 1.1185178796437059371676632938087720813830374192067503766 + 0.53063753095251782601650945810678674290339274946931684819*ImaginaryUnit);
  Eps; Test('arcsec(1/2 + 3⋅i)', 1.5193626859652739339438753648952886006484791830697021677 + 0.31930046431291188976224194180507391249323888240415051292*ImaginaryUnit);
  Eps; Test('arcsec(-5 - i)', 1.764159192442527195069605058846098455942470086412192650 - 0.03918190971789783355608094243759799949168435597631752405*ImaginaryUnit);
  Eps; Test('arcsec(3 - i/3)', 1.235529175718658836097423919553647710160672671956774308 - 0.03873276540858785871745750235620968971542824863383183700*ImaginaryUnit);
  Eps; Test('arcsec(1/2)', 1.316957896924816708625046347307968444026981971467516479768*ImaginaryUnit);
  Eps; Test('arcsec(-1/3)', 3.14159265358979323846264338327950288419716939937510582097 - 1.76274717403908605046521864995958461805632065652327082150*ImaginaryUnit);
  Eps; Test('arcsec(3 - 0⋅i)', TASC(ArcSec(TASR(3))));

  Eps; Test('arccsc(i)', -0.88137358701954302523260932497979230902816032826163541075*ImaginaryUnit);
  Eps; Test('arccsc(-i/2)', 1.443635475178810342493276740273105269405553003156981558983*ImaginaryUnit);
  Eps; Test('arccsc(3 + 7⋅i)', 0.05137311698382470436831234274226759204172025421216207297 - 0.1205568521335811427147324807681089362882002013727558169*ImaginaryUnit);
  Eps; Test('arccsc(2 - 4⋅i)', 0.09819751113349835325555360252611032475049647810490203295 + 0.1996393853895331198054284409849021595828912596512481146*ImaginaryUnit);
  Eps; Test('arccsc(-2 - i/5)', -0.516970434895694519606668267057593866015099072138245782 + 0.0569159834311293664547894794781437598484977806186101116*ImaginaryUnit);
  Eps; Test('arccsc(-2 + i)', -0.4015863916678060682834364455101079062076992541428898361 - 0.2156124185558296449669175649377748138325837860327226893*ImaginaryUnit);
  Eps; Test('arccsc(1/5)', 1.57079632679489661923132169163975144209858469968755291048 - 2.29243166956117768780078731134801543162186824001571024760*ImaginaryUnit);
  Eps; Test('arccsc(-2/3)', -1.570796326794896619231321691639751442098584699687552910 + 0.9624236501192068949955178268487368462703686687713210393*ImaginaryUnit);
  Eps; Test('arccsc(-2 + 0⋅i)', TASC(-Pi/6));

  // Complex hyperbolic functions

  Eps; Test('sinh(i)', 0.841470984807896506652502321630298999622563060798371065672*ImaginaryUnit);
  Eps; Test('sinh(2 + i)', 1.95960104142160589707035204998935827843632016018455965880 + 3.16577851321616814674073461719190553837911076789146893228*ImaginaryUnit);
  Eps; Test('sinh(3 - i)', 5.41268092317819278427941063379302764217781251017215863262 - 8.47164545430014942489758363997688120727985059927333986461*ImaginaryUnit);
  Eps; Test('sinh(-1 - 3⋅i)', 1.1634403637032505573309761202206068610104166170444687227 - 0.21775955162215223162459475156621247326955061443448722968*ImaginaryUnit);
  Eps; Test('sinh(-1 + 5⋅i/2)', 0.94150493327086727602538902375051252126864401328674467157 + 0.92349077604317324082303630121758402071294281076115355911*ImaginaryUnit);
  Eps; Test('sinh(-1 + 0⋅i)', TASC(Sinh(TASR(-1))));

  Eps; Test('cosh(i)', TASC(0.540302305868139717400936607442976603732310420617922227670));
  Eps; Test('cosh(-i/2)', TASC(0.877582561890372716116281582603829651991645197109744052997));
  Eps; Test('cosh(5 + 2⋅i)', -30.882235318916737411494024025148077933069785517444009293 + 67.472788440587524023225603430961109383148842533217249527*ImaginaryUnit);
  Eps; Test('cosh(3 - i)', 5.43958099101976439206063642199155745864401966972358255280 - 8.42975108084994488020976990589329511266286413185248634648*ImaginaryUnit);
  Eps; Test('cosh(-2 + 7⋅i)', 2.83632781278861669021310122787681746277278223612433925845 - 2.38279868337925305816574006377397602961262743125163959136*ImaginaryUnit);
  Eps; Test('cosh(-2 - 3⋅i/2)', 0.26612719531354575762276492476273912918619384684594107300 + 3.6177750739401373755640400182597583467227886309770489798*ImaginaryUnit);
  Eps; Test('cosh(-2 - 0⋅i)', TASC(Cosh(TASR(-2))));

  Eps; Test('tanh(i)', 1.557407724654902230506974807458360173087250772381520038383*ImaginaryUnit);
  Eps; Test('tanh(2 + i)', 1.014793616146633568117054175417967614163217471123533242 + 0.03381282607989669028437055972528730164018766933075587436*ImaginaryUnit);
  Eps; Test('tanh(2 - 3⋅i)', 0.965385879022133124278480269394560685879729650005757773 + 0.00988437503832249372031403430350121097961813353467039031*ImaginaryUnit);
  Eps; Test('tanh(-5 + 3⋅i)', -0.999912820151353808282092630139729541405660204620864 - 0.0000253686762076760324178061367074262881955607026024778*ImaginaryUnit);
  Eps; Test('tanh(-5 - 3⋅i/2)', -1.00008989513790955090797973923373112862250987297077 - 0.0000128148288216646398570254799433458544569865051830556*ImaginaryUnit);
  Eps; Test('tanh(-5 + 0⋅i)', TASC(Tanh(TASR(-5))));

  Eps; Test('coth(i)', -0.64209261593433070300641998659426562023027811391817137910*ImaginaryUnit);
  Eps; Test('coth(2 + 3⋅i)', 1.035746637764995396112758656897908320248306959923214042 + 0.01060478347033710175031689620777929239726759093914072464*ImaginaryUnit);
  Eps; Test('coth(3 - 5⋅i)', 0.995845318575854129760420015871648417110265572041024444 - 0.00268579840575852564465377110128147493789774393611081331*ImaginaryUnit);
  Eps; Test('coth(-1 + i)', -0.8680141428959249486358492089162738882734387499460932712 - 0.2176215618544026813651342436052380735207543691678540409*ImaginaryUnit);
  Eps; Test('coth(-2 - i/5)', -1.03421152371004768222472143351272382681025692126150923 + 0.0147578658539154012710311300865370447509718498417579195*ImaginaryUnit);
  Eps; Test('coth(-2 + 0⋅i)', TASC(Coth(TASR(-2))));

  Eps; Test('sech(i)', TASC(1.850815717680925617911753241398650193470396655094009298835));
  Eps; Test('sech(1 + i)', 0.49833703055518678521380589177216953443287793247109398814 - 0.59108384172104504805039169297433507150365868212588646424*ImaginaryUnit);
  Eps; Test('sech(1 - i/2)', 0.62948578135956955432815912338170185757161889630431625195 + 0.26190434748843763385912602613483627554864495961004915373*ImaginaryUnit);
  Eps; Test('sech(-5 + 3⋅i/2)', 0.00095337600687830383757118206204968467427180346898398407 + 0.013442734786700773269694515686799513138414286685751969*ImaginaryUnit);
  Eps; Test('sech(-5 - 4⋅i)', -0.0088089484097708918699776567438125690868890185381799680 + 0.010198261901161904258119499040075400553200436744025406*ImaginaryUnit);
  Eps; Test('sech(-5 + 0⋅i)', TASC(Sech(TASR(-5))));

  Eps; Test('csch(i)', -1.18839510577812121626159945237455100352782983409796262526*ImaginaryUnit);
  Eps; Test('csch(1 + 5⋅i)', 0.14489927361201806950211037509745131442709861525477131849 + 0.64316954772970666345156004990734985015885881931646018104*ImaginaryUnit);
  Eps; Test('csch(1 - 5⋅i/2)', -0.5413229061900023322691170635360287966693582652982964675 + 0.5309655776211742076846759350844320866020737940438825701*ImaginaryUnit);
  Eps; Test('csch(-2 + 3⋅i)', 0.2725486614629401995124984779327089240535398644954460147 - 0.04030057885689152187513247954286986778084247542426812633*ImaginaryUnit);
  Eps; Test('csch(-2 - 7⋅i)', -0.2012622548368106688480845612177196901938387522524683595 + 0.1819341993953043695306632153698328079117400980687779202*ImaginaryUnit);
  Eps; Test('csch(5 + 0⋅i)', TASC(Csch(TASR(5))));

  // Inverse complex hyperbolic functions

  Eps; Test('arcsinh(i)', Pi/2 * ImaginaryUnit);
  Eps; Test('arcsinh(-i)', -Pi/2 * ImaginaryUnit);
  Eps; Test('arcsinh(i/2)', Pi/6 * ImaginaryUnit);
  Eps; Test('arcsinh(i/4)', 0.252680255142078653485657436993710972252193733096838193633*ImaginaryUnit);
  Eps; Test('arcsinh(1 + i)', 1.0612750619050356520330189162135734858067854989386336963 + 0.66623943249251525510400489597779272066749013872594784283*ImaginaryUnit);
  Eps; Test('arcsinh(2 + 3⋅i)', 1.9686379257930962917886650952454981895207310126820105738 + 0.96465850440760279204541105949953235551977737250733165271*ImaginaryUnit);
  Eps; Test('arcsinh(2 - 3⋅i/2)', 1.6224941488715938425880479489804754465917756048177896209 - 0.60651151819975475395980252363910403305626429517951463818*ImaginaryUnit);
  Eps; Test('arcsinh(-1 + 2⋅i)', -1.4693517443681852732558443173616476167878033573347881728 + 1.0634400235777520561894919970895510028518424086279048181*ImaginaryUnit);
  Eps; Test('arcsinh(-1 + 0⋅i)', TASC(ArcSinh(TASR(-1))));

  Eps; Test('arccosh(i)', 0.88137358701954302523260932497979230902816032826163541075 + 1.5707963267948966192313216916397514420985846996875529104*ImaginaryUnit);
  Eps; Test('arccosh(1 + i)', 1.0612750619050356520330189162135734858067854989386336963 + 0.90455689430238136412731679566195872143109456096160506765*ImaginaryUnit);
  Eps; Test('arccosh(2 + 4⋅i)', 2.19857302792093578266601733129592145699144272657407248531 + 1.11692611683177409796864848134297958857023880515686366472*ImaginaryUnit);
  Eps; Test('arccosh(2 - i/2)', 1.3618009008578457882050378978645319263266616591142375121 - 0.27775425655771396017752906244863676219669029912827933295*ImaginaryUnit);
  Eps; Test('arccosh(-1 + 3⋅i/2)', 1.31695789692481670862504634730796844402698197146751647976 + 2.09439510239319549230842892218633525613144626625007054731*ImaginaryUnit);
  Eps; Test('arccosh(-1 - 3⋅i/2)', 1.31695789692481670862504634730796844402698197146751647976 - 2.09439510239319549230842892218633525613144626625007054731*ImaginaryUnit);
  Eps; Test('arccosh(-1)', Pi * ImaginaryUnit);
  Eps; Test('arccosh(0)', Pi/2 * ImaginaryUnit);
  Eps; Test('arccosh(2 + 0⋅i)', TASC(ArcCosh(TASR(2))));

  Eps; Test('arctanh(i)', Pi/4 * ImaginaryUnit);
  Eps; Test('arctanh(-i)', -Pi/4 * ImaginaryUnit);
  Eps; Test('arctanh(2 + i)', 0.40235947810852509365018983330654690988140033856712943047 + 1.3389725222944935611241935759091442410843161725444927785*ImaginaryUnit);
  Eps; Test('arctanh(2 - 3⋅i)', 0.14694666622552975204743278515471594244234494034424529538 - 1.3389725222944935611241935759091442410843161725444927785*ImaginaryUnit);
  Eps; Test('arctanh(-1 + 5⋅i)', -0.0371050012795683194954383414773840816108275513464271587 + 1.38054313823871417607952773323453488984988184285850249*ImaginaryUnit);
  Eps; Test('arctanh(-3 - 2⋅i)', -0.2290726829685387662958818029420027678625253049770656169 - 1.409921049596575522530619384460420782588207051908724814*ImaginaryUnit);
  Eps; Test('arctanh(2)', 0.54930614433405484569762261846126285232374527891137472586 - 1.5707963267948966192313216916397514420985846996875529104*ImaginaryUnit);
  Eps; Test('arctanh(-4)', -0.2554128118829953416027570481518309674390553982228841350 + 1.570796326794896619231321691639751442098584699687552910*ImaginaryUnit);
  Eps; Test('arctanh(0.52 + 0⋅i)', TASC(ArcTanh(TASR(0.52))));

  Eps; Test('arccoth(i)', -Pi/4 * ImaginaryUnit);
  Eps; Test('arccoth(-i)', Pi/4 * ImaginaryUnit);
  Eps; Test('arccoth(1 + i)', 0.40235947810852509365018983330654690988140033856712943047 - 0.55357435889704525150853273008926852003502382270071632333*ImaginaryUnit);
  Eps; Test('arccoth(2 - 3⋅i)', 0.14694666622552975204743278515471594244234494034424529538 + 0.23182380450040305810712811573060720101426852714306013190*ImaginaryUnit);
  Eps; Test('arccoth(-3 - 2⋅i)', -0.2290726829685387662958818029420027678625253049770656169 + 0.1608752771983210967007023071793306595103776477788280957*ImaginaryUnit);
  Eps; Test('arccoth(-3 + 5⋅i)', -0.0865690591794584441708728351688739957204743888691393886 - 0.147117282555593888376466528399690255288300270591305792*ImaginaryUnit);
  Eps; Test('arccoth(0.5)', 0.54930614433405484569762261846126285232374527891137472586 - 1.5707963267948966192313216916397514420985846996875529104*ImaginaryUnit);
  Eps; Test('arccoth(-1/3)', -0.3465735902799726547086160607290882840377500671801276270 + 1.570796326794896619231321691639751442098584699687552910*ImaginaryUnit);
  Eps; Test('arccoth(2 + 0⋅i)', TASC(ArcCoth(TASR(2))));

  Eps; Test('arcsech(i)', 0.88137358701954302523260932497979230902816032826163541075 - 1.5707963267948966192313216916397514420985846996875529104*ImaginaryUnit);
  Eps; Test('arcsech(2 + i)', 0.21561241855582964496691756493777481383258378603272268933 - 1.1692099351270905509478852461296435358908854455446630743*ImaginaryUnit);
  Eps; Test('arcsech(3 - 5⋅i)', 0.14709131539545412329207998996527146452483240278423548659 + 1.4833957949709502506043286970054860391604491466255433102*ImaginaryUnit);
  Eps; Test('arcsech(-1 + 2⋅i)', 0.39656823011232897891653433445907000552240234314617462124 - 1.7571143809027121450534918838480825107611837575945558325*ImaginaryUnit);
  Eps; Test('arcsech(-1 - 4⋅i)', 0.23355169780604010274248336571769587732347980013880148526 + 1.6280825489345943146036806147142453102717643620820134808*ImaginaryUnit);
  Eps; Test('arcsech(-1)', Pi * ImaginaryUnit);
  Eps; Test('arcsech(2)', Pi/3 * ImaginaryUnit);
  Eps; Test('arcsech(0.456 + 0⋅i)', TASC(ArcSech(TASR(0.456))));

  Eps; Test('arccsch(i)', -Pi/2 * ImaginaryUnit);
  Eps; Test('arccsch(-i)', Pi/2 * ImaginaryUnit);
  Eps; Test('arccsch(1 + 7⋅i)', 0.02019747349130953633698187409053590539862163336505798341 - 0.1404325799868232167518917095779004230428547022063631278*ImaginaryUnit);
  Eps; Test('arccsch(1 - 5⋅i)', 0.03918190971789783355608094243759799949168435597631752405 + 0.1933628656476305758382833672063470138438853867246397398*ImaginaryUnit);
  Eps; Test('arccsch(-2 + 3⋅i)', -0.1573554988449854287823288407037480934887571166858323920 - 0.2299629023772078545139672970137605536403087370229113715*ImaginaryUnit);
  Eps; Test('arccsch(-2 - i)', -0.3965682301123289789165343344590700055224023431461746212 + 0.1863180541078155258221701922083310686625990579070029220*ImaginaryUnit);
  Eps; Test('arccsch(-2 - i)', -0.3965682301123289789165343344590700055224023431461746212 + 0.1863180541078155258221701922083310686625990579070029220*ImaginaryUnit);
  Eps; Test('arccsch(-2 + 0⋅i)', TASC(ArcCsch(TASR(-2))));

  // Misc elementary functions

  Test('sinc(i - i)', TASC(1.0));

  for _i := -10 to 10 do
    for _j := -10 to 10 do
    begin
      if (_i = 0) and (_j = 0) then
        Continue;
      Eps; Test(Format('sinc(%d + %d⋅i)', [_i, _j]),
             csin(_i + _j*ImaginaryUnit) / (_i + _j*ImaginaryUnit));
    end;

  // Conjugate

  Test('ConjugateTranspose(0)', 0);
  Test('ConjugateTranspose(1/2-1/2)', Rat(0, 1));
  Test('ConjugateTranspose(0.0)', 0.0);
  Test('ConjugateTranspose(0 + 0⋅i)', TASC(0));
  Test('ConjugateTranspose(3 + 4⋅i)', 3 - 4*ImaginaryUnit);
  Test('ConjugateTranspose(7.2 + 9.25⋅i)', 7.2 - 9.25*ImaginaryUnit);
  Test('ConjugateTranspose(1 - 10.1⋅i)', 1 + 10.1*ImaginaryUnit);
  Test('ConjugateTranspose(-2 + 95⋅i)', -2 - 95*ImaginaryUnit);
  Test('ConjugateTranspose(-3 - 3.14⋅i)', -3 + 3.14*ImaginaryUnit);
  Test('ConjugateTranspose(3 + 0⋅i)', TASC(3));
  Test('ConjugateTranspose(3.0)', 3.0);
  Test('ConjugateTranspose(8/3 + 1/3)', Rat(3, 1));
  Test('ConjugateTranspose(3)', 3);
  Test('ConjugateTranspose(4⋅i)', -4*ImaginaryUnit);
  Test('ConjugateTranspose(-4.5⋅i)', 4.5*ImaginaryUnit);

  // Absolute value

  Eps; Test('abs(i)', 1.0);
  Eps; Test('abs(-i)', 1.0);
  Eps; Test('abs(1 + i)', Sqrt(2));
  Eps; Test('abs(1 - i)', Sqrt(2));
  Eps; Test('abs(-1 + i)', Sqrt(2));
  Eps; Test('abs(-1 - i)', Sqrt(2));
  Eps; Test('abs(3 + 4⋅i)', 5.0);
  Eps; Test('abs(-3 + 0⋅i)', 3.0);

  // Real and imaginary parts. Argument. Sign.

  Test('Re(3 + 7⋅i)', 3.0);
  Test('Im(3 + 7⋅i)', 7.0);
  Test('Re(2 − 10⋅i)', 2.0);
  Test('Im(2 − 10⋅i)', -10.0);
  Test('Re(−5 + i)', -5.0);
  Test('Im(−5 + i)', 1.0);

  Test('Re(3)', 3.0);
  Test('Re(3/2)', 3/2);
  Test('Re(3.)', 3.0);
  Test('Re(3 + 0⋅i)', 3.0);
  Test('Re(-3)', -3.0);
  Test('Re(-3/2)', -3/2);
  Test('Re(-3.)', -3.0);
  Test('Re(-3 + 0⋅i)', -3.0);

  Test('Im(3)', 0.0);
  Test('Im(3/2)', 0.0);
  Test('Im(3.)', 0.0);
  Test('Im(3 + 0⋅i)', 0.0);
  Test('Im(-3)', 0.0);
  Test('Im(-3/2)', 0.0);
  Test('Im(-3.)', 0.0);
  Test('Im(-3 + 0⋅i)', 0.0);

  Test('Re(21⋅i)', 0.0);
  Test('Im(21⋅i)', 21.0);

  Test('arg(-5)', Pi);
  Test('arg(-5/7)', Pi);
  Test('arg(-5.0)', Pi);
  Test('arg(-5 + 0⋅i)', Pi);

  Test('arg(-√3/2 - i/2)', -5*Pi/6);
  Test('arg(-1 - i)', -3*Pi/4);
  Test('arg(-1/2 - √3⋅i/2)', -2*Pi/3);
  Test('arg(-i)', -Pi/2);
  Test('arg(1/2 - √3⋅i/2)', -Pi/3);
  Test('arg(1 - i)', -Pi/4);
  Test('arg(√3/2 - i/2)', -Pi/6);

  Test('arg(5)', 0.0);
  Test('arg(5/8)', 0.0);
  Test('arg(5.0)', 0.0);
  Test('arg(5 + 0⋅i)', 0.0);

  Test('arg(√3/2 + i/2)', Pi/6);
  Test('arg(1 + i)', Pi/4);
  Test('arg(1/2 + √3⋅i/2)', Pi/3);
  Test('arg(i)', Pi/2);
  Test('arg(-1/2 + √3⋅i/2)', 2*Pi/3);
  Test('arg(-1 + i)', 3*Pi/4);
  Test('arg(-√3/2 + i/2)', 5*Pi/6);
  Test('arg(-5)', Pi);

  Test('sgn(1 + i − i)', TASC(1));
  Test('sgn(i)', ImaginaryUnit);
  Test('sgn(-1 + i − i)', TASC(-1));
  Test('sgn(-i)', -ImaginaryUnit);
  Eps; Test('sgn(1 + i)', (1 + ImaginaryUnit) / Sqrt(TASR(2)));
  Eps; Test('sgn(1 - i)', (1 - ImaginaryUnit) / Sqrt(TASR(2)));
  Eps; Test('sgn(-1 + i)', (-1 + ImaginaryUnit) / Sqrt(TASR(2)));
  Eps; Test('sgn(-1 - i)', (-1 - ImaginaryUnit) / Sqrt(TASR(2)));
  Eps; Test('sgn(4 + 2⋅i)', (4 + 2*ImaginaryUnit) / Sqrt(TASR(20)));
  Eps; Test('sgn(4 - 2⋅i)', (4 - 2*ImaginaryUnit) / Sqrt(TASR(20)));
  Eps; Test('sgn(-4 + 2⋅i)', (-4 + 2*ImaginaryUnit) / Sqrt(TASR(20)));
  Eps; Test('sgn(-4 - 2⋅i)', (-4 - 2*ImaginaryUnit) / Sqrt(TASR(20)));


  //
  // Integer functions. Rounding. Number theory.
  //

  // Parity

  for _i := -100 to 100 do
  begin
    Test(Format('odd(%d)', [_i]), Odd(_i));
    Test(Format('even(%d)', [_i]), not Odd(_i));
  end;

  Test('odd(30410216477784515)', True);
  Test('odd(30410216477784516)', False);
  Test('odd(30410216477784517)', True);
  Test('odd(30410216477784518)', False);
  Test('odd(30410216477784519)', True);
  Test('odd(30410216477784520)', False);

  Test('even(30410216477784515)', False);
  Test('even(30410216477784516)', True);
  Test('even(30410216477784517)', False);
  Test('even(30410216477784518)', True);
  Test('even(30410216477784519)', False);
  Test('even(30410216477784520)', True);

  Test('odd(-30410216477784515)', True);
  Test('odd(-30410216477784516)', False);
  Test('odd(-30410216477784517)', True);
  Test('odd(-30410216477784518)', False);
  Test('odd(-30410216477784519)', True);
  Test('odd(-30410216477784520)', False);

  Test('even(-30410216477784515)', False);
  Test('even(-30410216477784516)', True);
  Test('even(-30410216477784517)', False);
  Test('even(-30410216477784518)', True);
  Test('even(-30410216477784519)', False);
  Test('even(-30410216477784520)', True);

  Test('odd(2.4)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('odd(2.4 + i)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');
  Test('odd(2, 4)', failure, 'Too many arguments.');
  Test('odd()', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('even(2.4)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('even(2.4 + i)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');
  Test('even(2, 4)', failure, 'Too many arguments.');
  Test('even()', failure, 'Too few arguments. A required argument of type integer is missing.');

  // Round

  Test('round(-5.0)', -5);
  Test('round(-4.9)', -5);
  Test('round(-4.8)', -5);
  Test('round(-4.7)', -5);
  Test('round(-4.6)', -5);
  Test('round(-4.5)', -4);
  Test('round(-4.4)', -4);
  Test('round(-4.3)', -4);
  Test('round(-4.2)', -4);
  Test('round(-4.1)', -4);
  Test('round(-4.0)', -4);
  Test('round(-3.9)', -4);
  Test('round(-3.8)', -4);
  Test('round(-3.7)', -4);
  Test('round(-3.6)', -4);
  Test('round(-3.5)', -4);
  Test('round(-3.4)', -3);
  Test('round(-3.3)', -3);
  Test('round(-3.2)', -3);
  Test('round(-3.1)', -3);
  Test('round(-3.0)', -3);
  Test('round(-2.9)', -3);
  Test('round(-2.8)', -3);
  Test('round(-2.7)', -3);
  Test('round(-2.6)', -3);
  Test('round(-2.5)', -2);
  Test('round(-2.4)', -2);
  Test('round(-2.3)', -2);
  Test('round(-2.2)', -2);
  Test('round(-2.1)', -2);
  Test('round(-2.0)', -2);
  Test('round(-1.9)', -2);
  Test('round(-1.8)', -2);
  Test('round(-1.7)', -2);
  Test('round(-1.6)', -2);
  Test('round(-1.5)', -2);
  Test('round(-1.4)', -1);
  Test('round(-1.3)', -1);
  Test('round(-1.2)', -1);
  Test('round(-1.1)', -1);
  Test('round(-1.0)', -1);
  Test('round(-0.9)', -1);
  Test('round(-0.8)', -1);
  Test('round(-0.7)', -1);
  Test('round(-0.6)', -1);
  Test('round(-0.5)', -0);
  Test('round(-0.4)', -0);
  Test('round(-0.3)', -0);
  Test('round(-0.2)', -0);
  Test('round(-0.1)', -0);
  Test('round(0)', 0);
  Test('round(0.0)', 0);
  Test('round(0.1)', 0);
  Test('round(0.2)', 0);
  Test('round(0.3)', 0);
  Test('round(0.4)', 0);
  Test('round(0.5)', 0);
  Test('round(0.6)', 1);
  Test('round(0.7)', 1);
  Test('round(0.8)', 1);
  Test('round(0.9)', 1);
  Test('round(1.0)', 1);
  Test('round(1.1)', 1);
  Test('round(1.2)', 1);
  Test('round(1.3)', 1);
  Test('round(1.4)', 1);
  Test('round(1.5)', 2);
  Test('round(1.6)', 2);
  Test('round(1.7)', 2);
  Test('round(1.8)', 2);
  Test('round(1.9)', 2);
  Test('round(2.0)', 2);
  Test('round(2.1)', 2);
  Test('round(2.2)', 2);
  Test('round(2.3)', 2);
  Test('round(2.4)', 2);
  Test('round(2.5)', 2);
  Test('round(2.6)', 3);
  Test('round(2.7)', 3);
  Test('round(2.8)', 3);
  Test('round(2.9)', 3);
  Test('round(3.0)', 3);
  Test('round(3.1)', 3);
  Test('round(3.2)', 3);
  Test('round(3.3)', 3);
  Test('round(3.4)', 3);
  Test('round(3.5)', 4);
  Test('round(3.6)', 4);
  Test('round(3.7)', 4);
  Test('round(3.8)', 4);
  Test('round(3.9)', 4);
  Test('round(4.0)', 4);
  Test('round(4.1)', 4);
  Test('round(4.2)', 4);
  Test('round(4.3)', 4);
  Test('round(4.4)', 4);
  Test('round(4.5)', 4);
  Test('round(4.6)', 5);
  Test('round(4.7)', 5);
  Test('round(4.8)', 5);
  Test('round(4.9)', 5);
  Test('round(5.0)', 5);

  Test('round(5)', 5);

  Test('round(5/2)', 2);
  Test('round(5/3)', 2);
  Test('round(5/4)', 1);

  Test('round(123456789.123456)', 123456789);
  Test('round(123456789.987564231)', 123456790);
  Test('round(-123456789.123456)', -123456789);
  Test('round(-123456789.987564231)', -123456790);

  Test('round(1234567891234567)', 1234567891234567);
  Test('round(1234567891234567.2)', 1234567891234567);
  Test('round(1234567891234567.5)', 1234567891234568);
  Test('round(1234567891234567.8)', 1234567891234568);
  Test('round(1234567891234568)', 1234567891234568);

  Test('round(-1234567891234567)', -1234567891234567);
  Test('round(-1234567891234567.2)', -1234567891234567);
  Test('round(-1234567891234567.5)', -1234567891234568);
  Test('round(-1234567891234567.8)', -1234567891234568);
  Test('round(-1234567891234568)', -1234567891234568);

  Test('round(1E50)', failure, 'Invalid floating point operation');

  Test('round(π)', 3);

  Test('round(1, 2)', failure, 'Too many arguments.');
  Test('round("rectangle")', failure, 'An object of type real number was expected as argument 1, but an object of type string was given.');

  // Floor

  Test('floor(-5.0)', -5);
  Test('floor(-4.9)', -5);
  Test('floor(-4.8)', -5);
  Test('floor(-4.7)', -5);
  Test('floor(-4.6)', -5);
  Test('floor(-4.5)', -5);
  Test('floor(-4.4)', -5);
  Test('floor(-4.3)', -5);
  Test('floor(-4.2)', -5);
  Test('floor(-4.1)', -5);
  Test('floor(-4.0)', -4);
  Test('floor(-3.9)', -4);
  Test('floor(-3.8)', -4);
  Test('floor(-3.7)', -4);
  Test('floor(-3.6)', -4);
  Test('floor(-3.5)', -4);
  Test('floor(-3.4)', -4);
  Test('floor(-3.3)', -4);
  Test('floor(-3.2)', -4);
  Test('floor(-3.1)', -4);
  Test('floor(-3.0)', -3);
  Test('floor(-2.9)', -3);
  Test('floor(-2.8)', -3);
  Test('floor(-2.7)', -3);
  Test('floor(-2.6)', -3);
  Test('floor(-2.5)', -3);
  Test('floor(-2.4)', -3);
  Test('floor(-2.3)', -3);
  Test('floor(-2.2)', -3);
  Test('floor(-2.1)', -3);
  Test('floor(-2.0)', -2);
  Test('floor(-1.9)', -2);
  Test('floor(-1.8)', -2);
  Test('floor(-1.7)', -2);
  Test('floor(-1.6)', -2);
  Test('floor(-1.5)', -2);
  Test('floor(-1.4)', -2);
  Test('floor(-1.3)', -2);
  Test('floor(-1.2)', -2);
  Test('floor(-1.1)', -2);
  Test('floor(-1.0)', -1);
  Test('floor(-0.9)', -1);
  Test('floor(-0.8)', -1);
  Test('floor(-0.7)', -1);
  Test('floor(-0.6)', -1);
  Test('floor(-0.5)', -1);
  Test('floor(-0.4)', -1);
  Test('floor(-0.3)', -1);
  Test('floor(-0.2)', -1);
  Test('floor(-0.1)', -1);
  Test('floor(0)', 0);
  Test('floor(0.0)', 0);
  Test('floor(0.1)', 0);
  Test('floor(0.2)', 0);
  Test('floor(0.3)', 0);
  Test('floor(0.4)', 0);
  Test('floor(0.5)', 0);
  Test('floor(0.6)', 0);
  Test('floor(0.7)', 0);
  Test('floor(0.8)', 0);
  Test('floor(0.9)', 0);
  Test('floor(1.0)', 1);
  Test('floor(1.1)', 1);
  Test('floor(1.2)', 1);
  Test('floor(1.3)', 1);
  Test('floor(1.4)', 1);
  Test('floor(1.5)', 1);
  Test('floor(1.6)', 1);
  Test('floor(1.7)', 1);
  Test('floor(1.8)', 1);
  Test('floor(1.9)', 1);
  Test('floor(2.0)', 2);
  Test('floor(2.1)', 2);
  Test('floor(2.2)', 2);
  Test('floor(2.3)', 2);
  Test('floor(2.4)', 2);
  Test('floor(2.5)', 2);
  Test('floor(2.6)', 2);
  Test('floor(2.7)', 2);
  Test('floor(2.8)', 2);
  Test('floor(2.9)', 2);
  Test('floor(3.0)', 3);
  Test('floor(3.1)', 3);
  Test('floor(3.2)', 3);
  Test('floor(3.3)', 3);
  Test('floor(3.4)', 3);
  Test('floor(3.5)', 3);
  Test('floor(3.6)', 3);
  Test('floor(3.7)', 3);
  Test('floor(3.8)', 3);
  Test('floor(3.9)', 3);
  Test('floor(4.0)', 4);
  Test('floor(4.1)', 4);
  Test('floor(4.2)', 4);
  Test('floor(4.3)', 4);
  Test('floor(4.4)', 4);
  Test('floor(4.5)', 4);
  Test('floor(4.6)', 4);
  Test('floor(4.7)', 4);
  Test('floor(4.8)', 4);
  Test('floor(4.9)', 4);
  Test('floor(5.0)', 5);

  Test('floor(5)', 5);

  Test('floor(5/2)', 2);
  Test('floor(5/3)', 1);
  Test('floor(5/4)', 1);

  Test('floor(123456789.123456)', 123456789);
  Test('floor(123456789.987564231)', 123456789);
  Test('floor(-123456789.123456)', -123456790);
  Test('floor(-123456789.987564231)', -123456790);

  Test('floor(1234567891234567)', 1234567891234567);
  Test('floor(1234567891234567.2)', 1234567891234567);
  Test('floor(1234567891234567.5)', 1234567891234567);
  Test('floor(1234567891234567.8)', 1234567891234567);
  Test('floor(1234567891234568)', 1234567891234568);

  Test('floor(-1234567891234567)', -1234567891234567);
  Test('floor(-1234567891234567.2)', -1234567891234568);
  Test('floor(-1234567891234567.5)', -1234567891234568);
  Test('floor(-1234567891234567.8)', -1234567891234568);
  Test('floor(-1234567891234568)', -1234567891234568);

  Test('floor(1E50)', failure, 'Invalid floating point operation');

  Test('floor(π)', 3);

  Test('floor(1, 2)', failure, 'Too many arguments.');
  Test('floor("rectangle")', failure, 'An object of type real number was expected as argument 1, but an object of type string was given.');

  // Ceiling

  Test('ceil(-5.0)', -5);
  Test('ceil(-4.9)', -4);
  Test('ceil(-4.8)', -4);
  Test('ceil(-4.7)', -4);
  Test('ceil(-4.6)', -4);
  Test('ceil(-4.5)', -4);
  Test('ceil(-4.4)', -4);
  Test('ceil(-4.3)', -4);
  Test('ceil(-4.2)', -4);
  Test('ceil(-4.1)', -4);
  Test('ceil(-4.0)', -4);
  Test('ceil(-3.9)', -3);
  Test('ceil(-3.8)', -3);
  Test('ceil(-3.7)', -3);
  Test('ceil(-3.6)', -3);
  Test('ceil(-3.5)', -3);
  Test('ceil(-3.4)', -3);
  Test('ceil(-3.3)', -3);
  Test('ceil(-3.2)', -3);
  Test('ceil(-3.1)', -3);
  Test('ceil(-3.0)', -3);
  Test('ceil(-2.9)', -2);
  Test('ceil(-2.8)', -2);
  Test('ceil(-2.7)', -2);
  Test('ceil(-2.6)', -2);
  Test('ceil(-2.5)', -2);
  Test('ceil(-2.4)', -2);
  Test('ceil(-2.3)', -2);
  Test('ceil(-2.2)', -2);
  Test('ceil(-2.1)', -2);
  Test('ceil(-2.0)', -2);
  Test('ceil(-1.9)', -1);
  Test('ceil(-1.8)', -1);
  Test('ceil(-1.7)', -1);
  Test('ceil(-1.6)', -1);
  Test('ceil(-1.5)', -1);
  Test('ceil(-1.4)', -1);
  Test('ceil(-1.3)', -1);
  Test('ceil(-1.2)', -1);
  Test('ceil(-1.1)', -1);
  Test('ceil(-1.0)', -1);
  Test('ceil(-0.9)', -0);
  Test('ceil(-0.8)', -0);
  Test('ceil(-0.7)', -0);
  Test('ceil(-0.6)', -0);
  Test('ceil(-0.5)', -0);
  Test('ceil(-0.4)', -0);
  Test('ceil(-0.3)', -0);
  Test('ceil(-0.2)', -0);
  Test('ceil(-0.1)', -0);
  Test('ceil(0)', 0);
  Test('ceil(0.0)', 0);
  Test('ceil(0.1)', 1);
  Test('ceil(0.2)', 1);
  Test('ceil(0.3)', 1);
  Test('ceil(0.4)', 1);
  Test('ceil(0.5)', 1);
  Test('ceil(0.6)', 1);
  Test('ceil(0.7)', 1);
  Test('ceil(0.8)', 1);
  Test('ceil(0.9)', 1);
  Test('ceil(1.0)', 1);
  Test('ceil(1.1)', 2);
  Test('ceil(1.2)', 2);
  Test('ceil(1.3)', 2);
  Test('ceil(1.4)', 2);
  Test('ceil(1.5)', 2);
  Test('ceil(1.6)', 2);
  Test('ceil(1.7)', 2);
  Test('ceil(1.8)', 2);
  Test('ceil(1.9)', 2);
  Test('ceil(2.0)', 2);
  Test('ceil(2.1)', 3);
  Test('ceil(2.2)', 3);
  Test('ceil(2.3)', 3);
  Test('ceil(2.4)', 3);
  Test('ceil(2.5)', 3);
  Test('ceil(2.6)', 3);
  Test('ceil(2.7)', 3);
  Test('ceil(2.8)', 3);
  Test('ceil(2.9)', 3);
  Test('ceil(3.0)', 3);
  Test('ceil(3.1)', 4);
  Test('ceil(3.2)', 4);
  Test('ceil(3.3)', 4);
  Test('ceil(3.4)', 4);
  Test('ceil(3.5)', 4);
  Test('ceil(3.6)', 4);
  Test('ceil(3.7)', 4);
  Test('ceil(3.8)', 4);
  Test('ceil(3.9)', 4);
  Test('ceil(4.0)', 4);
  Test('ceil(4.1)', 5);
  Test('ceil(4.2)', 5);
  Test('ceil(4.3)', 5);
  Test('ceil(4.4)', 5);
  Test('ceil(4.5)', 5);
  Test('ceil(4.6)', 5);
  Test('ceil(4.7)', 5);
  Test('ceil(4.8)', 5);
  Test('ceil(4.9)', 5);
  Test('ceil(5.0)', 5);

  Test('ceil(5)', 5);

  Test('ceil(5/2)', 3);
  Test('ceil(5/3)', 2);
  Test('ceil(5/4)', 2);

  Test('ceil(123456789.123456)', 123456790);
  Test('ceil(123456789.987564231)', 123456790);
  Test('ceil(-123456789.123456)', -123456789);
  Test('ceil(-123456789.987564231)', -123456789);

  Test('ceil(1234567891234567)', 1234567891234567);
  Test('ceil(1234567891234567.2)', 1234567891234568);
  Test('ceil(1234567891234567.5)', 1234567891234568);
  Test('ceil(1234567891234567.8)', 1234567891234568);
  Test('ceil(1234567891234568)', 1234567891234568);

  Test('ceil(-1234567891234567)', -1234567891234567);
  Test('ceil(-1234567891234567.2)', -1234567891234567);
  Test('ceil(-1234567891234567.5)', -1234567891234567);
  Test('ceil(-1234567891234567.8)', -1234567891234567);
  Test('ceil(-1234567891234568)', -1234567891234568);

  Test('ceil(1E50)', failure, 'Invalid floating point operation');

  Test('ceil(π)', 4);

  Test('ceil(1, 2)', failure, 'Too many arguments.');
  Test('ceil("rectangle")', failure, 'An object of type real number was expected as argument 1, but an object of type string was given.');

  // Truncation (trunc)

  Test('trunc(-5.0)', -5);
  Test('trunc(-4.9)', -4);
  Test('trunc(-4.8)', -4);
  Test('trunc(-4.7)', -4);
  Test('trunc(-4.6)', -4);
  Test('trunc(-4.5)', -4);
  Test('trunc(-4.4)', -4);
  Test('trunc(-4.3)', -4);
  Test('trunc(-4.2)', -4);
  Test('trunc(-4.1)', -4);
  Test('trunc(-4.0)', -4);
  Test('trunc(-3.9)', -3);
  Test('trunc(-3.8)', -3);
  Test('trunc(-3.7)', -3);
  Test('trunc(-3.6)', -3);
  Test('trunc(-3.5)', -3);
  Test('trunc(-3.4)', -3);
  Test('trunc(-3.3)', -3);
  Test('trunc(-3.2)', -3);
  Test('trunc(-3.1)', -3);
  Test('trunc(-3.0)', -3);
  Test('trunc(-2.9)', -2);
  Test('trunc(-2.8)', -2);
  Test('trunc(-2.7)', -2);
  Test('trunc(-2.6)', -2);
  Test('trunc(-2.5)', -2);
  Test('trunc(-2.4)', -2);
  Test('trunc(-2.3)', -2);
  Test('trunc(-2.2)', -2);
  Test('trunc(-2.1)', -2);
  Test('trunc(-2.0)', -2);
  Test('trunc(-1.9)', -1);
  Test('trunc(-1.8)', -1);
  Test('trunc(-1.7)', -1);
  Test('trunc(-1.6)', -1);
  Test('trunc(-1.5)', -1);
  Test('trunc(-1.4)', -1);
  Test('trunc(-1.3)', -1);
  Test('trunc(-1.2)', -1);
  Test('trunc(-1.1)', -1);
  Test('trunc(-1.0)', -1);
  Test('trunc(-0.9)', -0);
  Test('trunc(-0.8)', -0);
  Test('trunc(-0.7)', -0);
  Test('trunc(-0.6)', -0);
  Test('trunc(-0.5)', -0);
  Test('trunc(-0.4)', -0);
  Test('trunc(-0.3)', -0);
  Test('trunc(-0.2)', -0);
  Test('trunc(-0.1)', -0);
  Test('trunc(0)', 0);
  Test('trunc(0.0)', 0);
  Test('trunc(0.1)', 0);
  Test('trunc(0.2)', 0);
  Test('trunc(0.3)', 0);
  Test('trunc(0.4)', 0);
  Test('trunc(0.5)', 0);
  Test('trunc(0.6)', 0);
  Test('trunc(0.7)', 0);
  Test('trunc(0.8)', 0);
  Test('trunc(0.9)', 0);
  Test('trunc(1.0)', 1);
  Test('trunc(1.1)', 1);
  Test('trunc(1.2)', 1);
  Test('trunc(1.3)', 1);
  Test('trunc(1.4)', 1);
  Test('trunc(1.5)', 1);
  Test('trunc(1.6)', 1);
  Test('trunc(1.7)', 1);
  Test('trunc(1.8)', 1);
  Test('trunc(1.9)', 1);
  Test('trunc(2.0)', 2);
  Test('trunc(2.1)', 2);
  Test('trunc(2.2)', 2);
  Test('trunc(2.3)', 2);
  Test('trunc(2.4)', 2);
  Test('trunc(2.5)', 2);
  Test('trunc(2.6)', 2);
  Test('trunc(2.7)', 2);
  Test('trunc(2.8)', 2);
  Test('trunc(2.9)', 2);
  Test('trunc(3.0)', 3);
  Test('trunc(3.1)', 3);
  Test('trunc(3.2)', 3);
  Test('trunc(3.3)', 3);
  Test('trunc(3.4)', 3);
  Test('trunc(3.5)', 3);
  Test('trunc(3.6)', 3);
  Test('trunc(3.7)', 3);
  Test('trunc(3.8)', 3);
  Test('trunc(3.9)', 3);
  Test('trunc(4.0)', 4);
  Test('trunc(4.1)', 4);
  Test('trunc(4.2)', 4);
  Test('trunc(4.3)', 4);
  Test('trunc(4.4)', 4);
  Test('trunc(4.5)', 4);
  Test('trunc(4.6)', 4);
  Test('trunc(4.7)', 4);
  Test('trunc(4.8)', 4);
  Test('trunc(4.9)', 4);
  Test('trunc(5.0)', 5);

  Test('trunc(5)', 5);

  Test('trunc(5/2)', 2);
  Test('trunc(5/3)', 1);
  Test('trunc(5/4)', 1);

  Test('trunc(123456789.123456)', 123456789);
  Test('trunc(123456789.987564231)', 123456789);
  Test('trunc(-123456789.123456)', -123456789);
  Test('trunc(-123456789.987564231)', -123456789);

  Test('trunc(1234567891234567)', 1234567891234567);
  Test('trunc(1234567891234567.2)', 1234567891234567);
  Test('trunc(1234567891234567.5)', 1234567891234567);
  Test('trunc(1234567891234567.8)', 1234567891234567);
  Test('trunc(1234567891234568)', 1234567891234568);

  Test('trunc(-1234567891234567)', -1234567891234567);
  Test('trunc(-1234567891234567.2)', -1234567891234567);
  Test('trunc(-1234567891234567.5)', -1234567891234567);
  Test('trunc(-1234567891234567.8)', -1234567891234567);
  Test('trunc(-1234567891234568)', -1234567891234568);

  Test('trunc(1E50)', failure, 'Invalid floating point operation');

  Test('trunc(π)', 3);

  Test('trunc(1, 2)', failure, 'Too many arguments.');
  Test('trunc("rectangle")', failure, 'An object of type real number was expected as argument 1, but an object of type string was given.');

  // Comparison of rounding routines

  Test('floor(1.4)', 1);
  Test('floor(1.6)', 1);
  Test('floor(-1.4)', -2);
  Test('floor(-1.6)', -2);

  Test('ceil(1.4)', 2);
  Test('ceil(1.6)', 2);
  Test('ceil(-1.4)', -1);
  Test('ceil(-1.6)', -1);

  Test('round(1.4)', 1);
  Test('round(1.6)', 2);
  Test('round(-1.4)', -1);
  Test('round(-1.6)', -2);

  Test('trunc(1.4)', 1);
  Test('trunc(1.6)', 1);
  Test('trunc(-1.4)', -1);
  Test('trunc(-1.6)', -1);

  Test('⌊1.4⌋', 1);
  Test('⌊1.6⌋', 1);
  Test('⌊−1.4⌋', -2);
  Test('⌊−1.6⌋', -2);

  Test('⌈1.4⌉', 2);
  Test('⌈1.6⌉', 2);
  Test('⌈−1.4⌉', -1);
  Test('⌈−1.6⌉', -1);

  // Signum function (sign, sgn)

  Test('sgn(0)', 0);
  Test('sgn(0.1)', 1);
  Test('sgn(-0.1)', -1);

  Test('sgn(45134)', 1);
  Test('sgn(45134.54325)', 1);
  Test('sgn(45134E50)', 1);

  Test('sgn(-45134)', -1);
  Test('sgn(-45134.54325)', -1);
  Test('sgn(-45134E50)', -1);

  Test('sgn(1E-100)', 1);
  Test('sgn(-1E-100)', -1);

  Test('sgn(-3/7)', -1);
  Test('sgn(5111511/9532151)', 1);

  Test('sgn(1234567891234567)', 1);
  Test('sgn(1234567891234567.5)', 1);
  Test('sgn(-1234567891234567)', -1);
  Test('sgn(-1234567891234567.5)', -1);

  Test('sgn(π)', 1);

  Test('sgn(2, 3)', failure, 'Too many arguments.');
  Test('sgn("rectangle")', failure, 'An object of type real number was expected as argument 1, but an object of type string was given.');

  // Modulo

  Test('mod(-15, 5)', 0);
  Test('mod(-14, 5)', 1);
  Test('mod(-13, 5)', 2);
  Test('mod(-12, 5)', 3);
  Test('mod(-11, 5)', 4);
  Test('mod(-10, 5)', 0);
  Test('mod(-9, 5)', 1);
  Test('mod(-8, 5)', 2);
  Test('mod(-7, 5)', 3);
  Test('mod(-6, 5)', 4);
  Test('mod(-5, 5)', 0);
  Test('mod(-4, 5)', 1);
  Test('mod(-3, 5)', 2);
  Test('mod(-2, 5)', 3);
  Test('mod(-1, 5)', 4);
  Test('mod(0, 5)', 0);
  Test('mod(1, 5)', 1);
  Test('mod(2, 5)', 2);
  Test('mod(3, 5)', 3);
  Test('mod(4, 5)', 4);
  Test('mod(5, 5)', 0);
  Test('mod(6, 5)', 1);
  Test('mod(7, 5)', 2);
  Test('mod(8, 5)', 3);
  Test('mod(9, 5)', 4);
  Test('mod(10, 5)', 0);
  Test('mod(11, 5)', 1);
  Test('mod(12, 5)', 2);
  Test('mod(13, 5)', 3);
  Test('mod(14, 5)', 4);
  Test('mod(15, 5)', 0);
  Test('mod(16, 5)', 1);
  Test('mod(17, 5)', 2);
  Test('mod(18, 5)', 3);
  Test('mod(19, 5)', 4);
  Test('mod(20, 5)', 0);
  Test('mod(21, 5)', 1);
  Test('mod(22, 5)', 2);
  Test('mod(23, 5)', 3);
  Test('mod(24, 5)', 4);
  Test('mod(25, 5)', 0);
  Test('mod(26, 5)', 1);
  Test('mod(27, 5)', 2);
  Test('mod(28, 5)', 3);
  Test('mod(29, 5)', 4);
  Test('mod(30, 5)', 0);

  Test('mod(0, 1)', 0);

  Test('mod(29, 1)', 0);
  Test('mod(25213459, 1)', 0);
  Test('mod(26164647157579, 1)', 0);

  Test('mod(-29, 1)', 0);
  Test('mod(-25213459, 1)', 0);
  Test('mod(-26164647157579, 1)', 0);

  Test('mod(1234567891234567, 156397337)', 0);
  Test('mod(1234567891234568, 156397337)', 1);
  Test('mod(1234567891234578, 156397337)', 11);
  Test('mod(1234567891234678, 156397337)', 111);
  Test('mod(1234567891235678, 156397337)', 1111);
  Test('mod(1234567891245678, 156397337)', 11111);

  Test('mod(-15.0, 5)', 0.0);
  Test('mod(-14.0, 5)', 1.0);
  Test('mod(-13.0, 5)', 2.0);
  Test('mod(-12.0, 5)', 3.0);
  Test('mod(-11.0, 5)', 4.0);
  Test('mod(-10.0, 5)', 0.0);
  Test('mod(-9.0, 5)', 1.0);
  Test('mod(-8.0, 5)', 2.0);
  Test('mod(-7.0, 5)', 3.0);
  Test('mod(-6.0, 5)', 4.0);
  Test('mod(-5.0, 5)', 0.0);
  Test('mod(-4.0, 5)', 1.0);
  Test('mod(-3.0, 5)', 2.0);
  Test('mod(-2.0, 5)', 3.0);
  Test('mod(-1.0, 5)', 4.0);
  Test('mod(0.0, 5)', 0.0);
  Test('mod(1.0, 5)', 1.0);
  Test('mod(2.0, 5)', 2.0);
  Test('mod(3.0, 5)', 3.0);
  Test('mod(4.0, 5)', 4.0);
  Test('mod(5.0, 5)', 0.0);
  Test('mod(6.0, 5)', 1.0);
  Test('mod(7.0, 5)', 2.0);
  Test('mod(8.0, 5)', 3.0);
  Test('mod(9.0, 5)', 4.0);
  Test('mod(10, 5.0)', 0.0);
  Test('mod(11, 5.0)', 1.0);
  Test('mod(12, 5.0)', 2.0);
  Test('mod(13, 5.0)', 3.0);
  Test('mod(14, 5.0)', 4.0);
  Test('mod(15, 5.0)', 0.0);
  Test('mod(16, 5.0)', 1.0);
  Test('mod(17, 5.0)', 2.0);
  Test('mod(18, 5.0)', 3.0);
  Test('mod(19, 5.0)', 4.0);
  Test('mod(20, 5.0)', 0.0);
  Test('mod(21, 5.0)', 1.0);
  Test('mod(22, 5.0)', 2.0);
  Test('mod(23.0, 5.0)', 3.0);
  Test('mod(24.0, 5.0)', 4.0);
  Test('mod(25.0, 5.0)', 0.0);
  Test('mod(26.0, 5.0)', 1.0);
  Test('mod(27.0, 5.0)', 2.0);
  Test('mod(28.0, 5.0)', 3.0);
  Test('mod(29.0, 5.0)', 4.0);
  Test('mod(30.0, 5.0)', 0.0);

  Test('mod(0.0, 1)', 0.0);
  Test('mod(0, 1.0)', 0.0);

  Test('mod(29.0, 1)', 0.0);
  Test('mod(25213459.0, 1)', 0.0);
  Test('mod(26164647157579.0, 1.0)', 0.0);
  Test('mod(26164647157579, 1.0)', 0.0);

  Test('mod(-29.0, 1)', 0.0);
  Test('mod(-25213459, 1.0)', 0.0);
  Test('mod(-26164647157579.0, 1.0)', 0.0);

  Test('mod(1234567891234567.0, 156397337)', 0.0);
  Test('mod(1234567891234568.0, 156397337)', 1.0);
  Test('mod(1234567891234578, 156397337.0)', 11.0);
  Test('mod(1234567891234678, 156397337.0)', 111.0);
  Test('mod(1234567891235678.0, 156397337.0)', 1111.0);
  Test('mod(1234567891245678.0, 156397337.0)', 11111.0);

  Test('mod(-10.5, 5)', 4.5);
  Test('mod(-9.5, 5)', 0.5);
  Test('mod(-8.5, 5)', 1.5);
  Test('mod(-7.5, 5)', 2.5);
  Test('mod(-6.5, 5)', 3.5);
  Test('mod(-5.5, 5)', 4.5);
  Test('mod(-4.5, 5)', 0.5);
  Test('mod(-3.5, 5)', 1.5);
  Test('mod(-2.5, 5)', 2.5);
  Test('mod(-1.5, 5)', 3.5);
  Test('mod(-0.5, 5)', 4.5);
  Test('mod(0.5, 5)', 0.5);
  Test('mod(1.5, 5)', 1.5);
  Test('mod(2.5, 5)', 2.5);
  Test('mod(3.5, 5)', 3.5);
  Test('mod(4.5, 5)', 4.5);
  Test('mod(5.5, 5)', 0.5);
  Test('mod(6.5, 5)', 1.5);
  Test('mod(7.5, 5)', 2.5);
  Test('mod(8.5, 5)', 3.5);
  Test('mod(9.5, 5)', 4.5);
  Test('mod(10.5, 5)', 0.5);
  Test('mod(11.5, 5)', 1.5);
  Test('mod(12.5, 5)', 2.5);
  Test('mod(13.5, 5)', 3.5);
  Test('mod(14.5, 5)', 4.5);
  Test('mod(15.5, 5)', 0.5);
  Test('mod(16.5, 5)', 1.5);
  Test('mod(17.5, 5)', 2.5);
  Test('mod(18.5, 5)', 3.5);
  Test('mod(19.5, 5)', 4.5);
  Test('mod(20.5, 5)', 0.5);

  Test('mod(1, 1.5)', 1.0);
  Test('mod(2, 1.5)', 0.5);
  Test('mod(3, 1.5)', 0.0);
  Test('mod(4, 1.5)', 1.0);
  Test('mod(5, 1.5)', 0.5);
  Test('mod(6, 1.5)', 0.0);
  Test('mod(7, 1.5)', 1.0);

  Eps; Test('mod(π, e)', 0.423310825130748003102355911926840386439922305675146246007);
  Eps; Test('mod(e, π)', 2.718281828459045235360287471352662497757247093699959574966);

  Eps; Test('mod(−π, e)', 2.294971003328297232257931559425822111317324788024813328958);
  Eps; Test('mod(−e, π)', 0.423310825130748003102355911926840386439922305675146246007);

  Eps(1E-6); Test('mod(123456789123, π)', 1.125871901464596442733771596141163579944705707241703723047);

  Test('mod(5, 0)', failure, 'Division by zero');
  Test('mod(5.0, 0)', failure, 'Division by zero');

  Test('mod(5)', failure, 'A required argument of type real number is missing.');
  Test('mod(5, 2, 1)', failure, 'Too many arguments.');
  Test('mod("rectangle", 5)', failure, 'An object of type real number was expected as argument 1, but an object of type string was given.');

  // Successor function

  Test('succ(-6)', -5);
  Test('succ(-5)', -4);
  Test('succ(-4)', -3);
  Test('succ(-3)', -2);
  Test('succ(-2)', -1);
  Test('succ(-1)', 0);
  Test('succ(0)', 1);
  Test('succ(1)', 2);
  Test('succ(2)', 3);
  Test('succ(3)', 4);
  Test('succ(4)', 5);
  Test('succ(5)', 6);
  Test('succ(6)', 7);
  Test('succ(7)', 8);

  Test('succ(-123456789123456)', -123456789123455);
  Test('succ(123456789123456)', 123456789123457);

  Test('succ(0.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('succ(1/2)', failure, 'An object of type integer was expected as argument 1, but an object of type rational number was given.');

  Test('succ(-6, 5)', 0);
  Test('succ(-5, 5)', 1);
  Test('succ(-4, 5)', 2);
  Test('succ(-3, 5)', 3);
  Test('succ(-2, 5)', 4);
  Test('succ(-1, 5)', 0);
  Test('succ(0, 5)', 1);
  Test('succ(1, 5)', 2);
  Test('succ(2, 5)', 3);
  Test('succ(3, 5)', 4);
  Test('succ(4, 5)', 0);
  Test('succ(5, 5)', 1);
  Test('succ(6, 5)', 2);
  Test('succ(7, 5)', 3);
  Test('succ(8, 5)', 4);
  Test('succ(9, 5)', 0);
  Test('succ(10, 5)', 1);
  Test('succ(11, 5)', 2);
  Test('succ(12, 5)', 3);

  Test('succ(-123456789123456, 5)', 0);
  Test('succ(123456789123456, 5)', 2);

  Test('succ(-6, 5, 10)', 10);
  Test('succ(-5, 5, 10)', 11);
  Test('succ(-4, 5, 10)', 12);
  Test('succ(-3, 5, 10)', 13);
  Test('succ(-2, 5, 10)', 14);
  Test('succ(-1, 5, 10)', 10);
  Test('succ(0, 5, 10)', 11);
  Test('succ(1, 5, 10)', 12);
  Test('succ(2, 5, 10)', 13);
  Test('succ(3, 5, 10)', 14);
  Test('succ(4, 5, 10)', 10);
  Test('succ(5, 5, 10)', 11);
  Test('succ(6, 5, 10)', 12);
  Test('succ(7, 5, 10)', 13);
  Test('succ(8, 5, 10)', 14);
  Test('succ(9, 5, 10)', 10);
  Test('succ(10, 5, 10)', 11);
  Test('succ(11, 5, 10)', 12);
  Test('succ(12, 5, 10)', 13);

  Test('succ(-123456789123456, 5, 10)', 10);
  Test('succ(123456789123456, 5, 10)', 12);

  Test('succ(-123456789123456, 5, 100000000000)', 100000000000);
  Test('succ(123456789123456, 5, 100000000000)', 100000000002);

  Test('succ(5, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('succ()', failure, 'There is no version of this function that can take 0 argument(s).');
  Test('succ(1, 2, 3, 4)', failure, 'There is no version of this function that can take 4 argument(s).');
  Test('succ("rectangle")', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');

  // Predecessor function

  Test('pred(-6)', -7);
  Test('pred(-5)', -6);
  Test('pred(-4)', -5);
  Test('pred(-3)', -4);
  Test('pred(-2)', -3);
  Test('pred(-1)', -2);
  Test('pred(0)', -1);
  Test('pred(1)', 0);
  Test('pred(2)', 1);
  Test('pred(3)', 2);
  Test('pred(4)', 3);
  Test('pred(5)', 4);
  Test('pred(6)', 5);
  Test('pred(7)', 6);

  Test('pred(-123456789123456)', -123456789123457);
  Test('pred(123456789123456)', 123456789123455);

  Test('pred(0.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('pred(1/2)', failure, 'An object of type integer was expected as argument 1, but an object of type rational number was given.');

  Test('pred(-6, 5)', 3);
  Test('pred(-5, 5)', 4);
  Test('pred(-4, 5)', 0);
  Test('pred(-3, 5)', 1);
  Test('pred(-2, 5)', 2);
  Test('pred(-1, 5)', 3);
  Test('pred(0, 5)', 4);
  Test('pred(1, 5)', 0);
  Test('pred(2, 5)', 1);
  Test('pred(3, 5)', 2);
  Test('pred(4, 5)', 3);
  Test('pred(5, 5)', 4);
  Test('pred(6, 5)', 0);
  Test('pred(7, 5)', 1);
  Test('pred(8, 5)', 2);
  Test('pred(9, 5)', 3);
  Test('pred(10, 5)', 4);
  Test('pred(11, 5)', 0);
  Test('pred(12, 5)', 1);

  Test('pred(-123456789123456, 5)', 3);
  Test('pred(123456789123456, 5)', 0);

  Test('pred(-6, 5, 10)', 13);
  Test('pred(-5, 5, 10)', 14);
  Test('pred(-4, 5, 10)', 10);
  Test('pred(-3, 5, 10)', 11);
  Test('pred(-2, 5, 10)', 12);
  Test('pred(-1, 5, 10)', 13);
  Test('pred(0, 5, 10)', 14);
  Test('pred(1, 5, 10)', 10);
  Test('pred(2, 5, 10)', 11);
  Test('pred(3, 5, 10)', 12);
  Test('pred(4, 5, 10)', 13);
  Test('pred(5, 5, 10)', 14);
  Test('pred(6, 5, 10)', 10);
  Test('pred(7, 5, 10)', 11);
  Test('pred(8, 5, 10)', 12);
  Test('pred(9, 5, 10)', 13);
  Test('pred(10, 5, 10)', 14);
  Test('pred(11, 5, 10)', 10);
  Test('pred(12, 5, 10)', 11);

  Test('pred(-123456789123456, 5, 10)', 13);
  Test('pred(123456789123456, 5, 10)', 10);

  Test('pred(-123456789123456, 5, 100000000000)', 100000000003);
  Test('pred(123456789123456, 5, 100000000000)', 100000000000);

  Test('pred(5, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('pred()', failure, 'There is no version of this function that can take 0 argument(s).');
  Test('pred(1, 2, 3, 4)', failure, 'There is no version of this function that can take 4 argument(s).');
  Test('pred("rectangle")', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');

  // Factorial

  Test('factorial(0)', 1);
  Test('factorial(1)', 1);
  Test('factorial(2)', 2);
  Test('factorial(3)', 6);
  Test('factorial(4)', 24);
  Test('factorial(5)', 120);
  Test('factorial(6)', 720);
  Test('factorial(7)', 5040);
  Test('factorial(8)', 40320);
  Test('factorial(9)', 362880);
  Test('factorial(10)', 3628800);
  Test('factorial(11)', 39916800);
  Test('factorial(12)', 479001600);
  Test('factorial(13)', 6227020800);
  Test('factorial(14)', 87178291200);
  Test('factorial(15)', 1307674368000);
  Test('factorial(16)', 20922789888000);
  Test('factorial(17)', 355687428096000);
  Test('factorial(18)', 6402373705728000);
  Test('factorial(19)', 121645100408832000);
  Test('factorial(20)', 2432902008176640000);

  Test('factorial(21)', 5.109094217170944E19);
  Test('factorial(22)', 1.12400072777760768E21);
  Test('factorial(23)', 2.58520167388849766E22);

  Eps; Test('factorial(50)', 3.0414093201713378043612608166064768844377641568960512E64);
  Eps; Test('factorial(100)', 9.3326215443944152681699238856266700490715968264381621E157);
  Eps; Test('factorial(1000)', 4.023872600770937735437024339230039857193748642107146E2567);

  Test('factorial(-1)', failure, 'A non-negative integer was expected as argument 1, but "-1" was given.');
  Test('factorial(1.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('factorial("cat")', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');
  Test('factorial(1, 2)', failure, 'Too many arguments.');
  Test('factorial()', failure, 'A required argument of type integer is missing.');

  // Primorial

  Test('primorial(0)', 1);
  Test('primorial(1)', 1);
  Test('primorial(2)', 2);
  Test('primorial(3)', 6);
  Test('primorial(4)', 6);
  Test('primorial(5)', 30);
  Test('primorial(6)', 30);
  Test('primorial(7)', 210);
  Test('primorial(8)', 210);
  Test('primorial(9)', 210);
  Test('primorial(10)', 210);
  Test('primorial(11)', 2310);
  Test('primorial(12)', 2310);
  Test('primorial(13)', 30030);
  Test('primorial(14)', 30030);
  Test('primorial(15)', 30030);
  Test('primorial(16)', 30030);
  Test('primorial(17)', 510510);
  Test('primorial(18)', 510510);
  Test('primorial(19)', 9699690);
  Test('primorial(20)', 9699690);
  Test('primorial(21)', 9699690);
  Test('primorial(22)', 9699690);
  Test('primorial(23)', 223092870);
  Test('primorial(24)', 223092870);
  Test('primorial(25)', 223092870);
  Test('primorial(26)', 223092870);
  Test('primorial(27)', 223092870);
  Test('primorial(28)', 223092870);
  Test('primorial(29)', 6469693230);
  Test('primorial(30)', 6469693230);
  Test('primorial(31)', 200560490130);
  Test('primorial(32)', 200560490130);
  Test('primorial(33)', 200560490130);
  Test('primorial(34)', 200560490130);
  Test('primorial(35)', 200560490130);
  Test('primorial(36)', 200560490130);
  Test('primorial(37)', 7420738134810);
  Test('primorial(38)', 7420738134810);
  Test('primorial(39)', 7420738134810);
  Test('primorial(40)', 7420738134810);
  Test('primorial(41)', 304250263527210);
  Test('primorial(42)', 304250263527210);
  Test('primorial(43)', 13082761331670030);
  Test('primorial(44)', 13082761331670030);
  Test('primorial(45)', 13082761331670030);
  Test('primorial(46)', 13082761331670030);
  Test('primorial(47)', 614889782588491410);
  Test('primorial(48)', 614889782588491410);
  Test('primorial(49)', 614889782588491410);
  Test('primorial(50)', 614889782588491410);
  Test('primorial(51)', 614889782588491410);
  Test('primorial(52)', 614889782588491410);
  Eps; Test('primorial(53)', 3.25891584771900447E19);
  Eps; Test('primorial(54)', 3.25891584771900447E19);
  Eps; Test('primorial(55)', 3.25891584771900447E19);
  Eps; Test('primorial(56)', 3.25891584771900447E19);
  Eps; Test('primorial(57)', 3.25891584771900447E19);
  Eps; Test('primorial(58)', 3.25891584771900447E19);
  Eps; Test('primorial(59)', 1.92276035015421264E21);
  Eps; Test('primorial(60)', 1.92276035015421264E21);
  Eps; Test('primorial(61)', 1.17288381359406971E23);
  Eps; Test('primorial(62)', 1.17288381359406971E23);
  Eps; Test('primorial(63)', 1.17288381359406971E23);
  Eps; Test('primorial(64)', 1.17288381359406971E23);
  Eps; Test('primorial(65)', 1.17288381359406971E23);
  Eps; Test('primorial(66)', 1.17288381359406971E23);
  Eps; Test('primorial(67)', 7.85832155108026706E24);
  Eps; Test('primorial(68)', 7.85832155108026706E24);
  Eps; Test('primorial(69)', 7.85832155108026706E24);
  Eps; Test('primorial(70)', 7.85832155108026706E24);
  Eps; Test('primorial(71)', 5.57940830126698961E26);
  Eps; Test('primorial(72)', 5.57940830126698961E26);
  Eps; Test('primorial(73)', 4.07296805992490242E28);
  Eps; Test('primorial(74)', 4.07296805992490242E28);
  Eps; Test('primorial(75)', 4.07296805992490242E28);
  Eps; Test('primorial(76)', 4.07296805992490242E28);
  Eps; Test('primorial(77)', 4.07296805992490242E28);
  Eps; Test('primorial(78)', 4.07296805992490242E28);
  Eps; Test('primorial(79)', 3.21764476734067291E30);
  Eps; Test('primorial(80)', 3.21764476734067291E30);
  Eps; Test('primorial(81)', 3.21764476734067291E30);
  Eps; Test('primorial(82)', 3.21764476734067291E30);
  Eps; Test('primorial(83)', 2.67064515689275851E32);
  Eps; Test('primorial(84)', 2.67064515689275851E32);
  Eps; Test('primorial(85)', 2.67064515689275851E32);
  Eps; Test('primorial(86)', 2.67064515689275851E32);
  Eps; Test('primorial(87)', 2.67064515689275851E32);
  Eps; Test('primorial(88)', 2.67064515689275851E32);
  Eps; Test('primorial(89)', 2.37687418963455508E34);
  Eps; Test('primorial(90)', 2.37687418963455508E34);
  Eps; Test('primorial(91)', 2.37687418963455508E34);
  Eps; Test('primorial(92)', 2.37687418963455508E34);
  Eps; Test('primorial(93)', 2.37687418963455508E34);
  Eps; Test('primorial(94)', 2.37687418963455508E34);
  Eps; Test('primorial(95)', 2.37687418963455508E34);
  Eps; Test('primorial(96)', 2.37687418963455508E34);
  Eps; Test('primorial(97)', 2.30556796394551842E36);
  Eps; Test('primorial(98)', 2.30556796394551842E36);
  Eps; Test('primorial(99)', 2.30556796394551842E36);
  Eps; Test('primorial(100)', 2.30556796394551842E36);

  Eps; Test('primorial(173)', 1.66589903787325219E68);

  Eps; Test('primorial(541)', 4.7119307999061849531624878347602604220205747734096755E219);
  Eps; Test('primorial(1583)', 1.2546824917742270487305744119826516778755973335815816E670);
  Eps; Test('primorial(3571)', 5.575898468972228945658434397398021879422236041037048E1519);

  Test('primorial(-1)', failure, 'A non-negative integer was expected as argument 1, but "-1" was given.');
  Test('primorial(5, 6)', failure, 'Too many arguments.');
  Test('primorial()', failure, 'A required argument of type integer is missing.');
  Test('primorial(true)', failure, 'An object of type integer was expected as argument 1, but an object of type boolean was given.');

  // Greatest common divisor (GCD)

  Test('gcd(9, 15)', 3);
  Test('gcd(8, 12)', 4);
  Test('gcd(36, 18)', 18);
  Test('gcd(100, 48)', 4);
  Test('gcd(100, 45)', 5);
  Test('gcd(1, 123)', 1);
  Test('gcd(1, 1024)', 1);
  Test('gcd(1024, 4096)', 1024);
  Test('gcd(16, 4096)', 16);
  Test('gcd(1, 1)', 1);
  Test('gcd(204110088, 1147278130)', 2);
  Test('gcd(204110088, 1147278131)', 1);
  Test('gcd(204110088, 1147278132)', 1121484);
  Test('gcd(204110088, 1147278133)', 1);
  Test('gcd(204110088, 1147278134)', 2);
  Test('gcd(204110088, 204110088)', 204110088);
  Test('gcd(38237143473243, 135261142842969)', 73391830083);
  Test('gcd(14139833799843612, 6808068125850628)', 523697548142356);
  Test('gcd(80864667758636586, 34792835773219524)', 191169427325382);
  Test('gcd(182, 521)', 1);
  Test('gcd(80864667758636586, 34792835773219541)', 1);
  Test('gcd(80864667758636586, 34792835773219542)', 18);

  Test('gcd(1024)', 1024);
  Test('gcd(16)', 16);
  Test('gcd(1)', 1);
  Test('gcd(204110088)', 204110088);

  Test('gcd(-9, 15)', 3);
  Test('gcd(-8, 12)', 4);
  Test('gcd(-36, 18)', 18);
  Test('gcd(-100, 48)', 4);
  Test('gcd(100, -45)', 5);
  Test('gcd(1, -123)', 1);
  Test('gcd(1, -1024)', 1);
  Test('gcd(1024, -4096)', 1024);
  Test('gcd(16, -4096)', 16);
  Test('gcd(-1, -1)', 1);
  Test('gcd(-204110088, -1147278130)', 2);
  Test('gcd(-204110088, -1147278131)', 1);
  Test('gcd(-204110088, -1147278132)', 1121484);
  Test('gcd(-204110088, -1147278133)', 1);
  Test('gcd(204110088, -1147278134)', 2);
  Test('gcd(204110088, -204110088)', 204110088);
  Test('gcd(38237143473243, -135261142842969)', 73391830083);
  Test('gcd(-14139833799843612, 6808068125850628)', 523697548142356);
  Test('gcd(-80864667758636586, -34792835773219524)', 191169427325382);
  Test('gcd(-182, -521)', 1);
  Test('gcd(-80864667758636586, -34792835773219541)', 1);
  Test('gcd(80864667758636586, -34792835773219542)', 18);

  Test('gcd(1234, 0)', 1234);
  Test('gcd(1, 0)', 1);
  Test('gcd(891156, 0)', 891156);
  Test('gcd(-84129156165137369, 0)', 84129156165137369);
  Test('gcd(-34315618861, 0)', 34315618861);
  Test('gcd(0, 151651561)', 151651561);
  Test('gcd(0, -151818)', 151818);

  Test('gcd(0, 0)', 0); // by definition (with good reason)

  Test('gcd(1.2, 5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('gcd("fifty-seven", 5)', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');

  Test('gcd(0, 9, 15)', 3);
  Test('gcd(8, 0, 12)', 4);
  Test('gcd(36, 18, 0)', 18);
  Test('gcd(100, 48, 0)', 4);
  Test('gcd(100, 0, 45)', 5);
  Test('gcd(1, 123, 0)', 1);
  Test('gcd(1, 1024, 0)', 1);
  Test('gcd(1024, 4096, 0)', 1024);
  Test('gcd(0, 16, 0, 4096)', 16);
  Test('gcd(1, 1, 0, 0, 0)', 1);
  Test('gcd(204110088, 0, 1147278130)', 2);
  Test('gcd(204110088, 0, 1147278131, 0)', 1);
  Test('gcd(0, 204110088, 1147278132, 0, 0)', 1121484);
  Test('gcd(0, 204110088, 0, 0, 1147278133)', 1);
  Test('gcd(204110088, 1147278134, 0, 0)', 2);
  Test('gcd(204110088, 204110088, 0)', 204110088);
  Test('gcd(38237143473243, 0, 135261142842969)', 73391830083);
  Test('gcd(14139833799843612, 0, 6808068125850628)', 523697548142356);
  Test('gcd(80864667758636586, 0, 0, 34792835773219524)', 191169427325382);
  Test('gcd(182, 521, 0)', 1);
  Test('gcd(80864667758636586, 0, 34792835773219541)', 1);
  Test('gcd(80864667758636586, 0, 34792835773219542)', 18);

  Test('gcd(-9, 0, 15)', 3);
  Test('gcd(-8, 12, 0)', 4);
  Test('gcd(-36, 18, 0)', 18);
  Test('gcd(0, -100, 0, 48, 0, 0)', 4);
  Test('gcd(100, -45, 0)', 5);
  Test('gcd(1, -123, 0)', 1);
  Test('gcd(1, -1024, 0)', 1);
  Test('gcd(1024, -4096, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 1024);
  Test('gcd(16, -4096, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 16);
  Test('gcd(-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1)', 1);
  Test('gcd(-204110088, -1147278130, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 2);
  Test('gcd(-204110088, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1147278131)', 1);
  Test('gcd(-204110088, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1147278132)', 1121484);
  Test('gcd(-204110088, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1147278133, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 1);
  Test('gcd(204110088, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1147278134)', 2);
  Test('gcd(204110088, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -204110088)', 204110088);
  Test('gcd(38237143473243, -135261142842969, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 73391830083);
  Test('gcd(-14139833799843612, 6808068125850628, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 523697548142356);
  Test('gcd(-80864667758636586, -34792835773219524, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 191169427325382);
  Test('gcd(-182, -521, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 1);
  Test('gcd(-80864667758636586, -34792835773219541, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 1);
  Test('gcd(80864667758636586, -34792835773219542, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 18);

  Test('gcd(24, 112, 48)', 8);
  Test('gcd(555895039266567, 13254159421503, 34391511701547)', 86628492951);
  Test('gcd(555895039266568, 13254159421504, 34391511701548)', 4);
  Test('gcd(-24, 112, 48)', 8);
  Test('gcd(555895039266567, -13254159421503, 34391511701547)', 86628492951);
  Test('gcd(555895039266568, 13254159421504, -34391511701548)', 4);
  Test('gcd(-24, -112, 48)', 8);
  Test('gcd(555895039266567, -13254159421503, -34391511701547)', 86628492951);
  Test('gcd(-555895039266568, -13254159421504, -34391511701548)', 4);
  Test('gcd(-24, 112, 48)', 8);
  Test('gcd(555895039266567, -13254159421503, 34391511701547)', 86628492951);
  Test('gcd(555895039266568, 13254159421504, -34391511701548)', 4);
  Test('gcd(0, -24, 0, -112, 48, 0, 48, 0, 48, 0, 0, 0, 0, 48)', 8);
  Test('gcd(555895039266567, -13254159421503, -34391511701547, 0, 0, 0, 0)', 86628492951);
  Test('gcd(-555895039266568, -13254159421504, 0, -34391511701548)', 4);
  Test('gcd(-555895039266568, -13254159421504, 0, -34391511701548, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)', 4);

  Test('gcd(17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, 1703472120174825, 3017579184309690, 2306987957036763)', 3244708800333);
  Test('gcd(7000413963543, 13947287476122, 5313889758177, 10293151697829, 2489630969826, 13063870035216, 3868297582149)', 13385112741);

  Test('gcd(17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, 1703472120174825, 3017579184309690, 2306987957036763, 17083391833753245, '+'1658046196970163, 3124654574720679, 2728800101080053, 1703472120174825, 3017579184309690, 2306987957036763, 17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, 1703472120174825, '+'3017579184309690, 2306987957036763, 17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, 1703472120174825, 3017579184309690, 2306987957036763, 17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, '+'1703472120174825, 3017579184309690, 2306987957036763, 17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, 1703472120174825, 3017579184309690, 2306987957036763)', 3244708800333);

  Test('gcd(-17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, -1703472120174825, 3017579184309690, 2306987957036763, -17083391833753245, '+'1658046196970163, 3124654574720679, -2728800101080053, 1703472120174825, 3017579184309690, -2306987957036763, -17083391833753245, -1658046196970163, -3124654574720679, 2728800101080053, -1703472120174825, '+'3017579184309690, 2306987957036763, 0, 17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, 1703472120174825, 3017579184309690, 2306987957036763, 17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, '+'1703472120174825, 3017579184309690, 2306987957036763, 17083391833753245, 1658046196970163, 3124654574720679, 2728800101080053, 1703472120174825, 3017579184309690, 0, 0, -2306987957036763)', 3244708800333);

  Test('gcd(0, 0, 0)', 0);
  Test('gcd(0, 1, 0)', 1);
  Test('gcd(0, 1, 1024)', 1);
  Test('gcd(0, -1, 0)', 1);
  Test('gcd(0, -1, -1024)', 1);

  Test('gcd()', failure, 'Cannot compute the GCD of an empty list.');

  // Least common multiple (LCM)

  Test('lcm(3, 7)', 21);
  Test('lcm(10, 4)', 20);
  Test('lcm(4, 12)', 12);
  Test('lcm(24486, 5538)', 22600578);
  Test('lcm(217932, 301158)', 76494132);
  Test('lcm(4249920670453458, 4157864627050134)', 1151728501692887118);

  Test('lcm(4)', 4);
  Test('lcm(24486)', 24486);
  Test('lcm(217932)', 217932);

  Test('lcm(-3, 7)', 21);
  Test('lcm(-10, 4)', 20);
  Test('lcm(4, -12)', 12);
  Test('lcm(24486, -5538)', 22600578);
  Test('lcm(-217932, -301158)', 76494132);
  Test('lcm(-4249920670453458, -4157864627050134)', 1151728501692887118);

  Test('lcm(1234, 1)', 1234);
  Test('lcm(1, 1)', 1);
  Test('lcm(123456789123456, 1)', 123456789123456);

  Test('lcm(-1234, 1)', 1234);
  Test('lcm(-1, 1)', 1);
  Test('lcm(-123456789123456, 1)', 123456789123456);

  Test('lcm(-3, 7, 1)', 21);
  Test('lcm(-10, 1, 4, 1)', 20);
  Test('lcm(4, -12, 1, 1, -1)', 12);
  Test('lcm(24486, -5538, 1)', 22600578);
  Test('lcm(-217932, 1, -301158)', 76494132);
  Test('lcm(1, 1, -4249920670453458, -1, -4157864627050134, 1)', 1151728501692887118);

  Test('lcm(151327422246, 231607221846, 311887021446)', 67844471888126934);
  Test('lcm(-151327422246, 231607221846, 311887021446)', 67844471888126934);
  Test('lcm(-151327422246, -231607221846, 311887021446)', 67844471888126934);
  Test('lcm(-151327422246, -231607221846, -311887021446)', 67844471888126934);
  Test('lcm(-151327422246, -231607221846, 1, 1, -311887021446, -1, 1, 1)', 67844471888126934);

  Test('lcm(3, 0)', 0);
  Test('lcm(0, 4)', 0);
  Test('lcm(0, 0)', 0);
  Test('lcm(0)', 0);
  Test('lcm(217932, 301158, 0)', 0);
  Test('lcm(0, 4249920670453458, 4157864627050134)', 0);
  Test('lcm(1, 1, -4249920670453458, -1, 0, -4157864627050134, 1)', 0);
  Test('lcm(151327422246, 0, 0, 0, 231607221846, 311887021446)', 0);

  Test('lcm(627619846691838, 2060592594005946, 2838174704951586)', failure, 'LCM too large for the integer type.');
  Test('lcm(627619846691838, 2060592594005946, 2838174704951586, 3949006292016786)', failure, 'LCM too large for the integer type.');

  Test('lcm(1.2, 5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('lcm("fifty-seven", 5)', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');

  Test('lcm()', failure, 'Cannot compute the LCM of an empty list.');

  // Binomial coefficients (combinations)

  Test('binomial(0, 0)', 1);

  Test('binomial(1, 0)', 1);
  Test('binomial(1, 1)', 1);

  Test('binomial(2, 0)', 1);
  Test('binomial(2, 1)', 2);
  Test('binomial(2, 2)', 1);

  Test('binomial(5, 0)', 1);
  Test('binomial(5, 1)', 5);
  Test('binomial(5, 2)', 10);
  Test('binomial(5, 3)', 10);
  Test('binomial(5, 4)', 5);
  Test('binomial(5, 5)', 1);

  Test('binomial(10, 0)', 1);
  Test('binomial(10, 1)', 10);
  Test('binomial(10, 2)', 45);
  Test('binomial(10, 3)', 120);
  Test('binomial(10, 4)', 210);
  Test('binomial(10, 5)', 252);
  Test('binomial(10, 6)', 210);
  Test('binomial(10, 7)', 120);
  Test('binomial(10, 8)', 45);
  Test('binomial(10, 9)', 10);
  Test('binomial(10, 10)', 1);

  Test('binomial(20, 0)', 1);
  Test('binomial(20, 1)', 20);
  Test('binomial(20, 2)', 190);
  Test('binomial(20, 3)', 1140);
  Test('binomial(20, 4)', 4845);
  Test('binomial(20, 5)', 15504);
  Test('binomial(20, 6)', 38760);
  Test('binomial(20, 7)', 77520);
  Test('binomial(20, 8)', 125970);
  Test('binomial(20, 9)', 167960);
  Test('binomial(20, 10)', 184756);
  Test('binomial(20, 11)', 167960);
  Test('binomial(20, 12)', 125970);
  Test('binomial(20, 13)', 77520);
  Test('binomial(20, 14)', 38760);
  Test('binomial(20, 15)', 15504);
  Test('binomial(20, 16)', 4845);
  Test('binomial(20, 17)', 1140);
  Test('binomial(20, 18)', 190);
  Test('binomial(20, 19)', 20);
  Test('binomial(20, 20)', 1);

  Test('binomial(50, 0)', 1);
  Test('binomial(50, 1)', 50);
  Test('binomial(50, 2)', 1225);
  Test('binomial(50, 3)', 19600);
  Test('binomial(50, 4)', 230300);
  Test('binomial(50, 5)', 2118760);
  Test('binomial(50, 6)', 15890700);
  Test('binomial(50, 7)', 99884400);
  Test('binomial(50, 8)', 536878650);
  Test('binomial(50, 9)', 2505433700);
  Test('binomial(50, 10)', 10272278170);
  Test('binomial(50, 11)', 37353738800);
  Test('binomial(50, 12)', 121399651100);
  Test('binomial(50, 13)', 354860518600);
  Test('binomial(50, 14)', 937845656300);
  Test('binomial(50, 15)', 2250829575120);
  Test('binomial(50, 16)', 4923689695575);
  Test('binomial(50, 17)', 9847379391150);
  Test('binomial(50, 18)', 18053528883775);
  Test('binomial(50, 19)', 30405943383200);
  Test('binomial(50, 20)', 47129212243960);
  Test('binomial(50, 21)', 67327446062800);
  Test('binomial(50, 22)', 88749815264600);
  Test('binomial(50, 23)', 108043253365600);
  Test('binomial(50, 24)', 121548660036300);
  Test('binomial(50, 25)', 126410606437752);
  Test('binomial(50, 26)', 121548660036300);
  Test('binomial(50, 27)', 108043253365600);
  Test('binomial(50, 28)', 88749815264600);
  Test('binomial(50, 29)', 67327446062800);
  Test('binomial(50, 30)', 47129212243960);
  Test('binomial(50, 31)', 30405943383200);
  Test('binomial(50, 32)', 18053528883775);
  Test('binomial(50, 33)', 9847379391150);
  Test('binomial(50, 34)', 4923689695575);
  Test('binomial(50, 35)', 2250829575120);
  Test('binomial(50, 36)', 937845656300);
  Test('binomial(50, 37)', 354860518600);
  Test('binomial(50, 38)', 121399651100);
  Test('binomial(50, 39)', 37353738800);
  Test('binomial(50, 40)', 10272278170);
  Test('binomial(50, 41)', 2505433700);
  Test('binomial(50, 42)', 536878650);
  Test('binomial(50, 43)', 99884400);
  Test('binomial(50, 44)', 15890700);
  Test('binomial(50, 45)', 2118760);
  Test('binomial(50, 46)', 230300);
  Test('binomial(50, 47)', 19600);
  Test('binomial(50, 48)', 1225);
  Test('binomial(50, 49)', 50);
  Test('binomial(50, 50)', 1);

  Test('binomial(100, 0   )', 1);
  Test('binomial(100, 1   )', 100);
  Test('binomial(100, 2   )', 4950);
  Test('binomial(100, 3   )', 161700);
  Test('binomial(100, 4   )', 3921225);
  Test('binomial(100, 5   )', 75287520);
  Test('binomial(100, 6   )', 1192052400);
  Test('binomial(100, 7   )', 16007560800);
  Test('binomial(100, 8   )', 186087894300);
  Test('binomial(100, 9   )', 1902231808400);
  Test('binomial(100, 10  )', 17310309456440);
  Test('binomial(100, 11  )', 141629804643600);
  Test('binomial(100, 12  )', 1050421051106700);
  Test('binomial(100, 13  )', 7110542499799200);
  Test('binomial(100, 14  )', 44186942677323600);
  Test('binomial(100, 15  )', 253338471349988640);
  Test('binomial(100, 16  )', 1345860629046814650);
  Test('binomial(100, 17  )', 6650134872937201800);
  Eps; Test('binomial(100, 18  )', 30664510802988208300.0);
  Eps; Test('binomial(100, 19  )', 132341572939212267400.0);
  Eps; Test('binomial(100, 20  )', 535983370403809682970.0);
  Eps; Test('binomial(100, 21  )', 2041841411062132125600.0);
  Eps; Test('binomial(100, 22  )', 7332066885177656269200.0);
  Eps; Test('binomial(100, 23  )', 24865270306254660391200.0);
  Eps; Test('binomial(100, 24  )', 79776075565900368755100.0);
  Eps; Test('binomial(100, 25  )', 242519269720337121015504.0);
  Eps; Test('binomial(100, 26  )', 699574816500972464467800.0);
  Eps; Test('binomial(100, 27  )', 1917353200780443050763600.0);
  Eps; Test('binomial(100, 28  )', 4998813702034726525205100.0);
  Eps; Test('binomial(100, 29  )', 12410847811948286545336800.0);
  Eps; Test('binomial(100, 30  )', 29372339821610944823963760.0);
  Eps; Test('binomial(100, 31  )', 66324638306863423796047200.0);
  Eps; Test('binomial(100, 32  )', 143012501349174257560226775.0);
  Eps; Test('binomial(100, 33  )', 294692427022540894366527900.0);
  Eps; Test('binomial(100, 34  )', 580717429720889409486981450.0);
  Eps; Test('binomial(100, 35  )', 1095067153187962886461165020.0);
  Eps; Test('binomial(100, 36  )', 1977204582144932989443770175.0);
  Eps; Test('binomial(100, 37  )', 3420029547493938143902737600.0);
  Eps; Test('binomial(100, 38  )', 5670048986634686922786117600.0);
  Eps; Test('binomial(100, 39  )', 9013924030034630492634340800.0);
  Eps; Test('binomial(100, 40  )', 13746234145802811501267369720.0);
  Eps; Test('binomial(100, 41  )', 20116440213369968050635175200.0);
  Eps; Test('binomial(100, 42  )', 28258808871162574166368460400.0);
  Eps; Test('binomial(100, 43  )', 38116532895986727945334202400.0);
  Eps; Test('binomial(100, 44  )', 49378235797073715747364762200.0);
  Eps; Test('binomial(100, 45  )', 61448471214136179596720592960.0);
  Eps; Test('binomial(100, 46  )', 73470998190814997343905056800.0);
  Eps; Test('binomial(100, 47  )', 84413487283064039501507937600.0);
  Eps; Test('binomial(100, 48  )', 93206558875049876949581681100.0);
  Eps; Test('binomial(100, 49  )', 98913082887808032681188722800.0);
  Eps; Test('binomial(100, 50  )', 100891344545564193334812497256.0);
  Eps; Test('binomial(100, 51  )', 98913082887808032681188722800.0);
  Eps; Test('binomial(100, 52  )', 93206558875049876949581681100.0);
  Eps; Test('binomial(100, 53  )', 84413487283064039501507937600.0);
  Eps; Test('binomial(100, 54  )', 73470998190814997343905056800.0);
  Eps; Test('binomial(100, 55  )', 61448471214136179596720592960.0);
  Eps; Test('binomial(100, 56  )', 49378235797073715747364762200.0);
  Eps; Test('binomial(100, 57  )', 38116532895986727945334202400.0);
  Eps; Test('binomial(100, 58  )', 28258808871162574166368460400.0);
  Eps; Test('binomial(100, 59  )', 20116440213369968050635175200.0);
  Eps; Test('binomial(100, 60  )', 13746234145802811501267369720.0);
  Eps; Test('binomial(100, 61  )', 9013924030034630492634340800.0);
  Eps; Test('binomial(100, 62  )', 5670048986634686922786117600.0);
  Eps; Test('binomial(100, 63  )', 3420029547493938143902737600.0);
  Eps; Test('binomial(100, 64  )', 1977204582144932989443770175.0);
  Eps; Test('binomial(100, 65  )', 1095067153187962886461165020.0);
  Eps; Test('binomial(100, 66  )', 580717429720889409486981450.0);
  Eps; Test('binomial(100, 67  )', 294692427022540894366527900.0);
  Eps; Test('binomial(100, 68  )', 143012501349174257560226775.0);
  Eps; Test('binomial(100, 69  )', 66324638306863423796047200.0);
  Eps; Test('binomial(100, 70  )', 29372339821610944823963760.0);
  Eps; Test('binomial(100, 71  )', 12410847811948286545336800.0);
  Eps; Test('binomial(100, 72  )', 4998813702034726525205100.0);
  Eps; Test('binomial(100, 73  )', 1917353200780443050763600.0);
  Eps; Test('binomial(100, 74  )', 699574816500972464467800.0);
  Eps; Test('binomial(100, 75  )', 242519269720337121015504.0);
  Eps; Test('binomial(100, 76  )', 79776075565900368755100.0);
  Eps; Test('binomial(100, 77  )', 24865270306254660391200.0);
  Eps; Test('binomial(100, 78  )', 7332066885177656269200.0);
  Eps; Test('binomial(100, 79  )', 2041841411062132125600.0);
  Eps; Test('binomial(100, 80  )', 535983370403809682970.0);
  Eps; Test('binomial(100, 81  )', 132341572939212267400.0);
  Eps; Test('binomial(100, 82  )', 30664510802988208300.0);
  Test('binomial(100, 83  )', 6650134872937201800);
  Test('binomial(100, 84  )', 1345860629046814650);
  Test('binomial(100, 85  )', 253338471349988640);
  Test('binomial(100, 86  )', 44186942677323600);
  Test('binomial(100, 87  )', 7110542499799200);
  Test('binomial(100, 88  )', 1050421051106700);
  Test('binomial(100, 89  )', 141629804643600);
  Test('binomial(100, 90  )', 17310309456440);
  Test('binomial(100, 91  )', 1902231808400);
  Test('binomial(100, 92  )', 186087894300);
  Test('binomial(100, 93  )', 16007560800);
  Test('binomial(100, 94  )', 1192052400);
  Test('binomial(100, 95  )', 75287520);
  Test('binomial(100, 96  )', 3921225);
  Test('binomial(100, 97  )', 161700);
  Test('binomial(100, 98  )', 4950);
  Test('binomial(100, 99  )', 100);
  Test('binomial(100, 100 )', 1);

  Test('binomial(200, 0   )', 1);
  Test('binomial(200, 1   )', 200);
  Test('binomial(200, 2   )', 19900);
  Test('binomial(200, 3   )', 1313400);
  Test('binomial(200, 4   )', 64684950);
  Test('binomial(200, 5   )', 2535650040);
  Test('binomial(200, 6   )', 82408626300);
  Test('binomial(200, 7   )', 2283896214600);
  Test('binomial(200, 8   )', 55098996177225);
  Test('binomial(200, 9   )', 1175445251780800);
  Test('binomial(200, 10  )', 22451004309013280);
  Test('binomial(200, 11  )', 387790074428411200);
  Test('binomial(200, 12  )', 6107693672247476400);
  Eps; Test('binomial(200, 13  )', 88326646952501966400.0);
  Eps; Test('binomial(200, 14  )', 1179791641436990551200.0);
  Eps; Test('binomial(200, 15  )', 14629416353818682834880.0);
  Eps; Test('binomial(200, 16  )', 169152626591028520278300.0);
  Eps; Test('binomial(200, 17  )', 1830828428985249866541600.0);
  Eps; Test('binomial(200, 18  )', 18613422361350040309839600.0);
  Eps; Test('binomial(200, 19  )', 178296993145563544020568800.0);
  Eps; Test('binomial(200, 20  )', 1613587787967350073386147640.0);
  Eps; Test('binomial(200, 21  )', 13830752468291572057595551200.0);
  Eps; Test('binomial(200, 22  )', 112532031446554154468618348400.0);
  Eps; Test('binomial(200, 23  )', 870900069455940847626698522400.0);
  Eps; Test('binomial(200, 24  )', 6422888012237563751246901602700.0);
  Eps; Test('binomial(200, 25  )', 45217131606152448808778187283008.0);
  Eps; Test('binomial(200, 26  )', 304346078118333790059083952866400.0);
  Eps; Test('binomial(200, 27  )', 1961341392318151091491874362916800.0);
  Eps; Test('binomial(200, 28  )', 12118287888251433529574795170878800.0);
  Eps; Test('binomial(200, 29  )', 71873983337215398865064302392798400.0);
  Eps; Test('binomial(200, 30  )', 409681705022127773530866523638950880.0);
  Eps; Test('binomial(200, 31  )', 2246641608185861983878945452213601600.0);
  Eps; Test('binomial(200, 32  )', 11865075993231583602360680669503083450.0);
  Eps; Test('binomial(200, 33  )', 60404023238269880157472556135652061200.0);
  Eps; Test('binomial(200, 34  )', 296690349435031470185232849254526300600.0);
  Eps; Test('binomial(200, 35  )', 1407159943034720687164247227892896168560.0);
  Eps; Test('binomial(200, 36  )', 6449483072242469816169466461175774105900.0);
  Eps; Test('binomial(200, 37  )', 28586897941831487833832229719806133874800.0);
  Eps; Test('binomial(200, 38  )', 122622746434698224129332985377063153199800.0);
  Eps; Test('binomial(200, 39  )', 509356023651823392537229323873954636368400.0);
  Eps; Test('binomial(200, 40  )', 2050157995198589154962348028592667411382810.0);
  Eps; Test('binomial(200, 41  )', 8000616566628640604731114257922604532225600.0);
  Eps; Test('binomial(200, 42  )', 30288048430808425146482075404992717157711200.0);
  Eps; Test('binomial(200, 43  )', 111290968652737934259166695674159286300427200.0);
  Eps; Test('binomial(200, 44  )', 397106410874542174515662982291886544299251600.0);
  Eps; Test('binomial(200, 45  )', 1376635557698412871654298338611873353570738880.0);
  Eps; Test('binomial(200, 46  )', 4638663292244652067530787880105225430510098400.0);
  Eps; Test('binomial(200, 47  )', 15199024404376094008505134756089462048905428800.0);
  Eps; Test('binomial(200, 48  )', 48446890288948799652110117035035160280886054300.0);
  Eps; Test('binomial(200, 49  )', 150284231100412603002464036516843762503973066400.0);
  Eps; Test('binomial(200, 50  )', 453858377923246061067441390280868162761998660528.0);
  Eps; Test('binomial(200, 51  )', 1334877582127194297257180559649612243417643119200.0);
  Eps; Test('binomial(200, 52  )', 3824937687249075967140767372842158159023631245400.0);
  Eps; Test('binomial(200, 53  )', 10680958070054023455411954173219611462933913666400.0);
  Eps; Test('binomial(200, 54  )', 29075941412924841628621430804875608982431209425200.0);
  Eps; Test('binomial(200, 55  )', 77183408114309579595976889045669798389726483201440.0);
  Eps; Test('binomial(200, 56  )', 199849896010265875739583016278966442259113215432300.0);
  Eps; Test('binomial(200, 57  )', 504883947815408528184209725336336275180917596881600.0);
  Eps; Test('binomial(200, 58  )', 1244800078234541716040379150398208402601227868173600.0);
  Eps; Test('binomial(200, 59  )', 2995959510327202096232776938246535477447023004756800.0);
  Eps; Test('binomial(200, 60  )', 7040504849268924926147025804879358372000504061178480.0);
  Eps; Test('binomial(200, 61  )', 16158535719633598191157108404641150361968369976475200.0);
  Eps; Test('binomial(200, 62  )', 36226394597243066912432872068469675811509732689194400.0);
  Eps; Test('binomial(200, 63  )', 79353054832056241808186291197600242253783223985854400.0);
  Eps; Test('binomial(200, 64  )', 169865132999870392620648779594863018574504713844719575.0);
  Eps; Test('binomial(200, 65  )', 355410124430498052252434369613867238863579093582797880.0);
  Eps; Test('binomial(200, 66  )', 726975254516927834152706665119273897675502691419359300.0);
  Eps; Test('binomial(200, 67  )', 1453950509033855668305413330238547795351005382838718600.0);
  Eps; Test('binomial(200, 68  )', 2843756142669158880656176072378336129142407587022787850.0);
  Eps; Test('binomial(200, 69  )', 5440229142497521336907467268897686507924605818652289800.0);
  Eps; Test('binomial(200, 70  )', 10181000252388218501926831603222813321973190889192142340.0);
  Eps; Test('binomial(200, 71  )', 18641268067753076130288564907309376505021335430915190200.0);
  Eps; Test('binomial(200, 72  )', 33398938621390928066767012125595966238163225980389715775.0);
  Eps; Test('binomial(200, 73  )', 58562522514219709486933939069538132581984834595751830400.0);
  Eps; Test('binomial(200, 74  )', 100505950801431123038386625159883011323136135049465979200.0);
  Eps; Test('binomial(200, 75  )', 168849997346404286704489530268603459022868706883102845056.0);
  Eps; Test('binomial(200, 76  )', 277713811425007050500805148468097794445507741584050732000.0);
  Eps; Test('binomial(200, 77  )', 447227436580530834572725174156417227418739739693795984000.0);
  Eps; Test('binomial(200, 78  )', 705243265376990931441605082323581012468012666440216744000.0);
  Eps; Test('binomial(200, 79  )', 1089109852860669539694630633461732702798703105135524592000.0);
  Eps; Test('binomial(200, 80  )', 1647278652451762678788128833110870712983038446517480945400.0);
  Eps; Test('binomial(200, 81  )', 2440412818447055820426857530534623278493390291137008808000.0);
  Eps; Test('binomial(200, 82  )', 3541574699941459056473122513824636221228212739576878636000.0);
  Eps; Test('binomial(200, 83  )', 5035009814374604441732872971461531013312398834579176856000.0);
  Eps; Test('binomial(200, 84  )', 7013049384307484758127930210249989625685126948163853478000.0);
  Eps; Test('binomial(200, 85  )', 9570749747996096846386351816341162312699702658670670628800.0);
  Eps; Test('binomial(200, 86  )', 12798095593250594620167796033479461232098439601710780492000.0);
  Eps; Test('binomial(200, 87  )', 16769918363569744674702629285248949200680713960862402024000.0);
  Eps; Test('binomial(200, 88  )', 21534099716856603957288603514012855223601371336107402599000.0);
  Eps; Test('binomial(200, 89  )', 27099091778516175766475545995162244775768017861168866192000.0);
  Eps; Test('binomial(200, 90  )', 33422213193503283445319840060700101890113888695441601636800.0);
  Eps; Test('binomial(200, 91  )', 40400477486652320648188817655791331955082722598885452528000.0);
  Eps; Test('binomial(200, 92  )', 47865783109185901637528055700883208512000182209549068756000.0);
  Eps; Test('binomial(200, 93  )', 55586070707441692224226129201025661497806663211089241136000.0);
  Eps; Test('binomial(200, 94  )', 63273506018045330510555274728827082768779925144537753208000.0);
  Eps; Test('binomial(200, 95  )', 70599911978029526674935359171112323931480758582326335158400.0);
  Eps; Test('binomial(200, 96  )', 77218653725969794800710549093404104300057079699419429079500.0);
  Eps; Test('binomial(200, 97  )', 82791133891761429477050485625917802548514807100408460044000.0);
  Eps; Test('binomial(200, 98  )', 87015171335218645266695918566015649617316582972878279434000.0);
  Eps; Test('binomial(200, 99  )', 89651994709013149668717007007410063242083752153874590932000.0);
  Eps; Test('binomial(200, 100 )', 90548514656103281165404177077484163874504589675413336841320.0);
  Eps; Test('binomial(200, 101 )', 89651994709013149668717007007410063242083752153874590932000.0);
  Eps; Test('binomial(200, 102 )', 87015171335218645266695918566015649617316582972878279434000.0);
  Eps; Test('binomial(200, 103 )', 82791133891761429477050485625917802548514807100408460044000.0);
  Eps; Test('binomial(200, 104 )', 77218653725969794800710549093404104300057079699419429079500.0);
  Eps; Test('binomial(200, 105 )', 70599911978029526674935359171112323931480758582326335158400.0);
  Eps; Test('binomial(200, 106 )', 63273506018045330510555274728827082768779925144537753208000.0);
  Eps; Test('binomial(200, 107 )', 55586070707441692224226129201025661497806663211089241136000.0);
  Eps; Test('binomial(200, 108 )', 47865783109185901637528055700883208512000182209549068756000.0);
  Eps; Test('binomial(200, 109 )', 40400477486652320648188817655791331955082722598885452528000.0);
  Eps; Test('binomial(200, 110 )', 33422213193503283445319840060700101890113888695441601636800.0);
  Eps; Test('binomial(200, 111 )', 27099091778516175766475545995162244775768017861168866192000.0);
  Eps; Test('binomial(200, 112 )', 21534099716856603957288603514012855223601371336107402599000.0);
  Eps; Test('binomial(200, 113 )', 16769918363569744674702629285248949200680713960862402024000.0);
  Eps; Test('binomial(200, 114 )', 12798095593250594620167796033479461232098439601710780492000.0);
  Eps; Test('binomial(200, 115 )', 9570749747996096846386351816341162312699702658670670628800.0);
  Eps; Test('binomial(200, 116 )', 7013049384307484758127930210249989625685126948163853478000.0);
  Eps; Test('binomial(200, 117 )', 5035009814374604441732872971461531013312398834579176856000.0);
  Eps; Test('binomial(200, 118 )', 3541574699941459056473122513824636221228212739576878636000.0);
  Eps; Test('binomial(200, 119 )', 2440412818447055820426857530534623278493390291137008808000.0);
  Eps; Test('binomial(200, 120 )', 1647278652451762678788128833110870712983038446517480945400.0);
  Eps; Test('binomial(200, 121 )', 1089109852860669539694630633461732702798703105135524592000.0);
  Eps; Test('binomial(200, 122 )', 705243265376990931441605082323581012468012666440216744000.0);
  Eps; Test('binomial(200, 123 )', 447227436580530834572725174156417227418739739693795984000.0);
  Eps; Test('binomial(200, 124 )', 277713811425007050500805148468097794445507741584050732000.0);
  Eps; Test('binomial(200, 125 )', 168849997346404286704489530268603459022868706883102845056.0);
  Eps; Test('binomial(200, 126 )', 100505950801431123038386625159883011323136135049465979200.0);
  Eps; Test('binomial(200, 127 )', 58562522514219709486933939069538132581984834595751830400.0);
  Eps; Test('binomial(200, 128 )', 33398938621390928066767012125595966238163225980389715775.0);
  Eps; Test('binomial(200, 129 )', 18641268067753076130288564907309376505021335430915190200.0);
  Eps; Test('binomial(200, 130 )', 10181000252388218501926831603222813321973190889192142340.0);
  Eps; Test('binomial(200, 131 )', 5440229142497521336907467268897686507924605818652289800.0);
  Eps; Test('binomial(200, 132 )', 2843756142669158880656176072378336129142407587022787850.0);
  Eps; Test('binomial(200, 133 )', 1453950509033855668305413330238547795351005382838718600.0);
  Eps; Test('binomial(200, 134 )', 726975254516927834152706665119273897675502691419359300.0);
  Eps; Test('binomial(200, 135 )', 355410124430498052252434369613867238863579093582797880.0);
  Eps; Test('binomial(200, 136 )', 169865132999870392620648779594863018574504713844719575.0);
  Eps; Test('binomial(200, 137 )', 79353054832056241808186291197600242253783223985854400.0);
  Eps; Test('binomial(200, 138 )', 36226394597243066912432872068469675811509732689194400.0);
  Eps; Test('binomial(200, 139 )', 16158535719633598191157108404641150361968369976475200.0);
  Eps; Test('binomial(200, 140 )', 7040504849268924926147025804879358372000504061178480.0);
  Eps; Test('binomial(200, 141 )', 2995959510327202096232776938246535477447023004756800.0);
  Eps; Test('binomial(200, 142 )', 1244800078234541716040379150398208402601227868173600.0);
  Eps; Test('binomial(200, 143 )', 504883947815408528184209725336336275180917596881600.0);
  Eps; Test('binomial(200, 144 )', 199849896010265875739583016278966442259113215432300.0);
  Eps; Test('binomial(200, 145 )', 77183408114309579595976889045669798389726483201440.0);
  Eps; Test('binomial(200, 146 )', 29075941412924841628621430804875608982431209425200.0);
  Eps; Test('binomial(200, 147 )', 10680958070054023455411954173219611462933913666400.0);
  Eps; Test('binomial(200, 148 )', 3824937687249075967140767372842158159023631245400.0);
  Eps; Test('binomial(200, 149 )', 1334877582127194297257180559649612243417643119200.0);
  Eps; Test('binomial(200, 150 )', 453858377923246061067441390280868162761998660528.0);
  Eps; Test('binomial(200, 151 )', 150284231100412603002464036516843762503973066400.0);
  Eps; Test('binomial(200, 152 )', 48446890288948799652110117035035160280886054300.0);
  Eps; Test('binomial(200, 153 )', 15199024404376094008505134756089462048905428800.0);
  Eps; Test('binomial(200, 154 )', 4638663292244652067530787880105225430510098400.0);
  Eps; Test('binomial(200, 155 )', 1376635557698412871654298338611873353570738880.0);
  Eps; Test('binomial(200, 156 )', 397106410874542174515662982291886544299251600.0);
  Eps; Test('binomial(200, 157 )', 111290968652737934259166695674159286300427200.0);
  Eps; Test('binomial(200, 158 )', 30288048430808425146482075404992717157711200.0);
  Eps; Test('binomial(200, 159 )', 8000616566628640604731114257922604532225600.0);
  Eps; Test('binomial(200, 160 )', 2050157995198589154962348028592667411382810.0);
  Eps; Test('binomial(200, 161 )', 509356023651823392537229323873954636368400.0);
  Eps; Test('binomial(200, 162 )', 122622746434698224129332985377063153199800.0);
  Eps; Test('binomial(200, 163 )', 28586897941831487833832229719806133874800.0);
  Eps; Test('binomial(200, 164 )', 6449483072242469816169466461175774105900.0);
  Eps; Test('binomial(200, 165 )', 1407159943034720687164247227892896168560.0);
  Eps; Test('binomial(200, 166 )', 296690349435031470185232849254526300600.0);
  Eps; Test('binomial(200, 167 )', 60404023238269880157472556135652061200.0);
  Eps; Test('binomial(200, 168 )', 11865075993231583602360680669503083450.0);
  Eps; Test('binomial(200, 169 )', 2246641608185861983878945452213601600.0);
  Eps; Test('binomial(200, 170 )', 409681705022127773530866523638950880.0);
  Eps; Test('binomial(200, 171 )', 71873983337215398865064302392798400.0);
  Eps; Test('binomial(200, 172 )', 12118287888251433529574795170878800.0);
  Eps; Test('binomial(200, 173 )', 1961341392318151091491874362916800.0);
  Eps; Test('binomial(200, 174 )', 304346078118333790059083952866400.0);
  Eps; Test('binomial(200, 175 )', 45217131606152448808778187283008.0);
  Eps; Test('binomial(200, 176 )', 6422888012237563751246901602700.0);
  Eps; Test('binomial(200, 177 )', 870900069455940847626698522400.0);
  Eps; Test('binomial(200, 178 )', 112532031446554154468618348400.0);
  Eps; Test('binomial(200, 179 )', 13830752468291572057595551200.0);
  Eps; Test('binomial(200, 180 )', 1613587787967350073386147640.0);
  Eps; Test('binomial(200, 181 )', 178296993145563544020568800.0);
  Eps; Test('binomial(200, 182 )', 18613422361350040309839600.0);
  Eps; Test('binomial(200, 183 )', 1830828428985249866541600.0);
  Eps; Test('binomial(200, 184 )', 169152626591028520278300.0);
  Eps; Test('binomial(200, 185 )', 14629416353818682834880.0);
  Eps; Test('binomial(200, 186 )', 1179791641436990551200.0);
  Eps; Test('binomial(200, 187 )', 88326646952501966400.0);
  Test('binomial(200, 188 )', 6107693672247476400);
  Test('binomial(200, 189 )', 387790074428411200);
  Test('binomial(200, 190 )', 22451004309013280);
  Test('binomial(200, 191 )', 1175445251780800);
  Test('binomial(200, 192 )', 55098996177225);
  Test('binomial(200, 193 )', 2283896214600);
  Test('binomial(200, 194 )', 82408626300);
  Test('binomial(200, 195 )', 2535650040);
  Test('binomial(200, 196 )', 64684950);
  Test('binomial(200, 197 )', 1313400);
  Test('binomial(200, 198 )', 19900);
  Test('binomial(200, 199 )', 200);
  Test('binomial(200, 200 )', 1);

  Test('binomial(400, 0   )', 1);
  Test('binomial(400, 1   )', 400);
  Test('binomial(400, 2   )', 79800);
  Test('binomial(400, 3   )', 10586800);
  Test('binomial(400, 4   )', 1050739900);
  Test('binomial(400, 5   )', 83218600080);
  Test('binomial(400, 6   )', 5478557838600);
  Test('binomial(400, 7   )', 308364541201200);
  Test('binomial(400, 8   )', 15148408086508950);
  Test('binomial(400, 9   )', 659797329990167600);
  Eps; Test('binomial(400, 10  )', 25798075602615553160.0);
  Eps; Test('binomial(400, 11  )', 914659044092733248400.0);
  Eps; Test('binomial(400, 12  )', 29650197346006102802300.0);
  Eps; Test('binomial(400, 13  )', 884944351557720606714800.0);
  Eps; Test('binomial(400, 14  )', 24462390289488419628473400.0);
  Eps; Test('binomial(400, 15  )', 629498843449501998439382160.0);
  Eps; Test('binomial(400, 16  )', 15147315920503641837447633225.0);
  Eps; Test('binomial(400, 17  )', 342151136086670497975287715200.0);
  Eps; Test('binomial(400, 18  )', 7280215840066377818029733051200.0);
  Eps; Test('binomial(400, 19  )', 146370655310808227709860948713600.0);
  Eps; Test('binomial(400, 20  )', 2788360983670896737872851072994080.0);
  Eps; Test('binomial(400, 21  )', 50456055894997179066270638463702400.0);
  Eps; Test('binomial(400, 22  )', 869220235645633221187116908079236800.0);
  Eps; Test('binomial(400, 23  )', 14285445611915189461249138750171804800.0);
  Eps; Test('binomial(400, 24  )', 224400541487167767787121887867282100400.0);
  Eps; Test('binomial(400, 25  )', 3374984143967003227518313193523922790016.0);
  Eps; Test('binomial(400, 26  )', 48677655922601008089206440291210424856000.0);
  Eps; Test('binomial(400, 27  )', 674275678335288037976415135885655514672000.0);
  Eps; Test('binomial(400, 28  )', 8982315286395087077328673060191053820452000.0);
  Eps; Test('binomial(400, 29  )', 115221423673757668716078150979002483489936000.0);
  Eps; Test('binomial(400, 30  )', 1424904939432136503122166467106997379158875200.0);
  Eps; Test('binomial(400, 31  )', 17006929922254532456619406220309323557702704000.0);
  Eps; Test('binomial(400, 32  )', 196111160665997577390392527977941887274759305500.0);
  Eps; Test('binomial(400, 33  )', 2186936579548094196353468190784321652033679528000.0);
  Eps; Test('binomial(400, 34  )', 23606050726298546178285965471113119008716481964000.0);
  Eps; Test('binomial(400, 35  )', 246851844737864797178647524640782901634006639966400.0);
  Eps; Test('binomial(400, 36  )', 2502803425814462526950176291496826641567011766326000.0);
  Eps; Test('binomial(400, 37  )', 24622174243147685400266599192022835068388980620072000.0);
  Eps; Test('binomial(400, 38  )', 235206559217437100007809881755376029732242104344372000.0);
  Eps; Test('binomial(400, 39  )', 2183199344531082825713517363985798019565939532632376000.0);
  Eps; Test('binomial(400, 40  )', 19703374084393022502064494209971827126582604282007193400.0);
  Eps; Test('binomial(400, 41  )', 173005235862963124408371168672923360135847257110307064000.0);
  Eps; Test('binomial(400, 42  )', 1478782849400089563395363084609035387827837269109529428000.0);
  Eps; Test('binomial(400, 43  )', 12311726978726327062686976378838015554473621914911896168000.0);
  Eps; Test('binomial(400, 44  )', 99892875713756790031346603801026626203342795991444248454000.0);
  Eps; Test('binomial(400, 45  )', 790263638979942605581319798959232865075334119398981165547200.0);
  Eps; Test('binomial(400, 46  )', 6098773735606078803942794100663644936994426356231267690636000.0);
  Eps; Test('binomial(400, 47  )', 45935444732011742480760619396487878887149509151188697074152000.0);
  Eps; Test('binomial(400, 48  )', 337816916466669689493927055145004609315912015216033543066159500.0);
  Eps; Test('binomial(400, 49  )', 2426766420331994504119639253286563724065327129715179737944656000.0);
  Eps; Test('binomial(400, 50  )', 17035900270730601418919867558071677342938596450600561760371485120.0);
  Eps; Test('binomial(400, 51  )', 116913041073641382286704973437746805294676642308043070904510192000.0);
  Eps; Test('binomial(400, 52  )', 784666371821170046501154533264877597073887464721289072032193404000.0);
  Eps; Test('binomial(400, 53  )', 5152149007429569361932109010871271769466279957037898057871760464000.0);
  Eps; Test('binomial(400, 54  )', 33107327881075195714637811606895024148237021205410196779287053352000.0);
  Eps; Test('binomial(400, 55  )', 208275189942763958495721505745194151914363806128580510647878553814400.0);
  Eps; Test('binomial(400, 56  )', 1283123938040242244303998562180213971615277019899290645955680376178000.0);
  Eps; Test('binomial(400, 57  )', 7743765520804268983167991322631466776064127979743087407171123673776000.0);
  Eps; Test('binomial(400, 58  )', 45795027131652832090114155580389536279137860293997913459649921036296000.0);
  Eps; Test('binomial(400, 59  )', 265455919983479128386763410313444430635002512212665871240682593125648000.0);
  Eps; Test('binomial(400, 60  )', 1508674478572773046331438715281409180775597611075317701551212737597432800.0);
  Eps; Test('binomial(400, 61  )', 8409005290405620258240805954027526581372183405993574074219874275133232000.0);
  Eps; Test('binomial(400, 62  )', 45978270862056536573284406748634379211051131848900348567105441601131704000.0);
  Eps; Test('binomial(400, 63  )', 246677072244049354948732213984736828148179088332195520883835543828293904000.0);
  Eps; Test('binomial(400, 64  )', 1298908958535072384651918064263379860717755511999217039653946535470860088250.0);
  Eps; Test('binomial(400, 65  )', 6714360154888989557585299532192240510787166954334414235749631321818599840800.0);
  Eps; Test('binomial(400, 66  )', 34080464422542598511985990049763645016874256510636799529941310497109559798000.0);
  Eps; Test('binomial(400, 67  )', 169893658464615341835870457860015782621432860814219269298513398597531238396000.0);
  Eps; Test('binomial(400, 68  )', 831979239245836894578600918638018464896134450751985539358896496073204446851000.0);
  Eps; Test('binomial(400, 69  )', 4003146484487215202899934854895972903558212139850133319813820821685563425428000.0);
  Eps; Test('binomial(400, 70  )', 18929164090932403316569691956722386158253831689862773269405352742541735625952400.0);
  Eps; Test('binomial(400, 71  )', 87980621831094268936168990784766020172165696586686129280334738099137644458652000.0);
  Eps; Test('binomial(400, 72  )', 402022563644861312222216638447055842175590474680829674072640678258559514262451500.0);
  Eps; Test('binomial(400, 73  )', 1806347957198828909710781608364853647035529804045371686244193732449418091480604000.0);
  Eps; Test('binomial(400, 74  )', 7982105162216446668586832242369015440278624944903196505430423655553509674515642000.0);
  Eps; Test('binomial(400, 75  )', 34695550438434154852790764146830653780411089760512560810270908156139255385227990560.0);
  Eps; Test('binomial(400, 76  )', 148369130164356583252065767733157401034652686475876082412342699351911289476303907000.0);
  Eps; Test('binomial(400, 77  )', 624306469782487441216484529162896077080876239197192866254532916753496854419772284000.0);
  Eps; Test('binomial(400, 78  )', 2585269099227480045037493627174556832014397759752478151284796565530506204840851894000.0);
  Eps; Test('binomial(400, 79  )', 10537425948749981955722442379116548100109317451143012211565879672162316429857649492000.0);
  Eps; Test('binomial(400, 80  )', 42281421619359302597336300046205149251688636272711336498908092184551294674803818586650.0);
  Eps; Test('binomial(400, 81  )', 167037715039444158409229827343032688401732884040341082464822092580943386369595332688000.0);
  Eps; Test('binomial(400, 82  )', 649817452409544957713955060029602775611619390352058601296076189430743173803669647896000.0);
  Eps; Test('binomial(400, 83  )', 2489662046581148151241418181800164851138493567854875123037978653481642521320083711216000.0);
  Eps; Test('binomial(400, 84  )', 9395510342455047189803923376555384021558362631071373976226657537543817610219839719708000.0);
  Eps; Test('binomial(400, 85  )', 34929191390774057787976938670488251186028736369630049135148515080751133703876110016796800.0);
  Eps; Test('binomial(400, 86  )', 127938317303416607014101577688416268879058743679458900901997468028332640892104356456872000.0);
  Eps; Test('binomial(400, 87  )', 461754386589342696579630981542100096873844201325863159577324194952832749886445608361584000.0);
  Eps; Test('binomial(400, 88  )', 1642376397755275727607096559348606026380832216079490556223891738866325576300653129740634000.0);
  Eps; Test('binomial(400, 89  )', 5757544225838719404645102545132191912705838779964056781369148567711163818042739061562672000.0);
  Eps; Test('binomial(400, 90  )', 19895513935953797053829187683734574276127954006320240655620057828424132749014353868288788800.0);
  Eps; Test('binomial(400, 91  )', 67775926595007440513044485516018879402194129032519501134529867327598693980158787902961808000.0);
  Eps; Test('binomial(400, 92  )', 227638709976709773027508108961411236253021585554875280984453576132913004781185494152339116000.0);
  Eps; Test('binomial(400, 93  )', 753900243793834517123360188818437212536888691945178349927007542461690381425861636547531696000.0);
  Eps; Test('binomial(400, 94  )', 2462206115369225497413527425183619406902391791778401632208418250380201564869569387447789688000.0);
  Eps; Test('binomial(400, 95  )', 7930895487399821075879362022170395142232967245096746310060799838066754514211455079568669942400.0);
  Eps; Test('binomial(400, 96  )', 25197115871426514876491723091270526233135989684942787755922332818857917987859310409046295129500.0);
  Eps; Test('binomial(400, 97  )', 78968280669212995076840039378827216235807637775490798740210197700338217199064230560310038344000.0);
  Eps; Test('binomial(400, 98  )', 244157031048689158247780938079435168565813410673201143043711121461249793993025121018101445084000.0);
  Eps; Test('binomial(400, 99  )', 744802256330344704957877205050398190978541919427340860597987461427246846322157439873400367832000.0);
  Eps; Test('binomial(400, 100 )', 2241854791554337561923210387201698554845411177476295990399942258896013007429693894018935107174320.0);
  Eps; Test('binomial(400, 101 )', 6658974628379220480960030853074352143105181715276126704158244333354494081474338299066143882696000.0);
  Eps; Test('binomial(400, 102 )', 19519935430248891409873031618325796968514209145760410632777598585029350297655168151184088440452000.0);
  Eps; Test('binomial(400, 103 )', 56475152992370579030506440992826092200167323547928178335608974546978120278652816592746197623832000.0);
  Eps; Test('binomial(400, 104 )', 161280004218596749731350124758359128687016298978217970823806398465889439641921985846592506675751000.0);
  Eps; Test('binomial(400, 105 )', 454656011892425123052187018366421924679588804738595422512825656627650229847703883910394114057355200.0);
  Eps; Test('binomial(400, 106 )', 1265316259511937842456558211491457243212063182998921223030977063256196394387477790127983619310564000.0);
  Eps; Test('binomial(400, 107 )', 3476663367257100240020823496995218967330341829922269528701936977545062990186153928015207327825288000.0);
  Eps; Test('binomial(400, 108 )', 9432058950058614540056493376107399605812871816363194184348773466858365334486510193596812472711198000.0);
  Eps; Test('binomial(400, 109 )', 25267534068046930694463266658929914540342739177780300016787539929565529152936339234222653596620824000.0);
  Eps; Test('binomial(400, 110 )', 66844112852742334837171005434078228465815791824855157317137946540941536213677042883261747241969634400.0);
  Eps; Test('binomial(400, 111 )', 174637772317975469394410734917862038334113329992864825423153193665522932450147229154467627929470216000.0);
  Eps; Test('binomial(400, 112 )', 450627823213347416562363414207697581058560288999445844172243508654786852482969189514653075639436539500.0);
  Eps; Test('binomial(400, 113 )', 1148502770667646513008501445060326578273144807361419496651381685775031978009691385665664475965997552000.0);
  Eps; Test('binomial(400, 114 )', 2891406098084338151170525567827313403196425962392345574903039858047668225340187962158295654405625416000.0);
  Eps; Test('binomial(400, 115 )', 7190801252627136619432785325205318550558068045601833342802342603492461847367771801541500497043555382400.0);
  Eps; Test('binomial(400, 116 )', 17667054801713223590847791531754446438871115456866573299126445189615100228446680719304548634977700724000.0);
  Eps; Test('binomial(400, 117 )', 42884133022962012818810023889044981099482023844018006982494960973082807392126985677628135148151000048000.0);
  Eps; Test('binomial(400, 118 )', 102849234283883471421383362377963810602995023286924542169881982672732495694677431752277646160396042488000.0);
  Eps; Test('binomial(400, 119 )', 243726756874412932275883262105762979748273920730358999091653101795887090637807023144052909388501546064000.0);
  Eps; Test('binomial(400, 120 )', 570726822347583616412693305430994977577208097710257322872954346705368937243531445862323896151407787033200.0);
  Eps; Test('binomial(400, 121 )', 1320690167415895971864083681989079286955522870734479755408489397334738036596601692904551164647885788176000.0);
  Eps; Test('binomial(400, 122 )', 3020266858270778493033437272745517385742548204384588948844004441445835345987310428855489958498033892632000.0);
  Eps; Test('binomial(400, 123 )', 6826294199994117244417037087993933603548198380641591282753115729446684765727417066844115515954905871152000.0);
  Eps; Test('binomial(400, 124 )', 15249060430632019973415478011083222646635894769659038591311395621425255484729794576740483854189588115396000.0);
  Eps; Test('binomial(400, 125 )', 33669925430835500101301375448471755603772055651407157209615561532106964110283386425442988350050610558794368.0);
  Eps; Test('binomial(400, 126 )', 73485948360950496252840303558172482468550121461404509782891106518487421669269295769816046002094586537051200.0);
  Eps; Test('binomial(400, 127 )', 158544487014964062781718450196372127530572702995471147090646954220988610530549504259288162240739501662614400.0);
  Eps; Test('binomial(400, 128 )', 338145663711603040151633882059449928248799593107528305904270457049452270897187614553013033529077218389794775.0);
  Eps; Test('binomial(400, 129 )', 712989306430666875358483844342406050261034800970912396945438483081015640961511869445112752867511654279257200.0);
  Eps; Test('binomial(400, 130 )', 1486308477251620947862685552436861843236464854331671227478567914730424913081305512458658123285351217766759240.0);
  Eps; Test('binomial(400, 131 )', 3063383884411737831472710680595058760869049699767566652055063641047440660549255636365173231198815487000190800.0);
  Eps; Test('binomial(400, 132 )', 6242805037172405126258781614242960656619502797253601737900091813952738921876892168047209084791525500023116100.0);
  Eps; Test('binomial(400, 133 )', 12579486841821087021333484756519650044917494358375678689903944407062661887691782714561293494166382210572895600.0);
  Eps; Test('binomial(400, 134 )', 25065096916165897273851047985005571358156499952882882165704128034968139731445567050655711663749433210619127800.0);
  Eps; Test('binomial(400, 135 )', 49387524294075027221069472326010977639034288796050715970942948572603890137514969151662365204128512844627318480.0);
  Eps; Test('binomial(400, 136 )', 96233043661249133923407427694065507899588871551128233325734421851029638870893138420518579258044528704604701450.0);
  Eps; Test('binomial(400, 137 )', 185441777566202710626128181833819664857601913062028128452510126778626457386246631700853320614042011518362344400.0);
  Eps; Test('binomial(400, 138 )', 353414402173270383294722549436917187373545674893575346253696835817237378931759885053075531315167021951661569400.0);
  Eps; Test('binomial(400, 139 )', 666148009851775830382858330593325921524237171382134825312723532259828728633964675423782656148012660081549145200.0);
  Eps; Test('binomial(400, 140 )', 1241890218366524940928043030606129039413042155219551352904434585141537844096177002040051951818795030580602334980.0);
  Eps; Test('binomial(400, 141 )', 2290010331739691380434689276294989718066602555723995402518815547069502407553234188158961045906997928730188702800.0);
  Eps; Test('binomial(400, 142 )', 4176849830426620193891440299721143218163732830510667670791360751345078334903434188261766978097975095359992070600.0);
  Eps; Test('binomial(400, 143 )', 7535854938811664405762179002294090561442259232669596217231965551377833639196405738262488673771171850369775903600.0);
  Eps; Test('binomial(400, 144 )', 13449407772740262168617222247149869960351809880528376582143160741000716981065807463426802702494383094062725050175.0);
  Eps; Test('binomial(400, 145 )', 23745161309113842173558682036347356619655609168381133827783787239284024463123080763015596495438359117793500778240.0);
  Eps; Test('binomial(400, 146 )', 41472713245370066809982629584031342041179317383131432370444285931626207110249216401157377440662887500255771907200.0);
  Eps; Test('binomial(400, 147 )', 71660334451183652855344135471727625023534330716431182463216657324034398680294564393836556938288254592278680710400.0);
  Eps; Test('binomial(400, 148 )', 122500436595604487651365312664507358992933686967953305156714961506626370716990032375950330441803570350314231214400.0);
  Eps; Test('binomial(400, 149 )', 207181946456995509316403079137287613867243551113585455701289733554831177320009987642546867592848991464960981651200.0);
  Eps; Test('binomial(400, 150 )', 346684457071372485589447819089727940537854208863399662540158154148417503382150045988528425105367312384701375963008.0);
  Eps; Test('binomial(400, 151 )', 573980889191014048989152018360476722744791736528807388311520122762280634738659016537298717061866411232949297952000.0);
  Eps; Test('binomial(400, 152 )', 940271325056332224988808240603675683970086463129427892694529674788209724012671678406495924660557476296081415724000.0);
  Eps; Test('binomial(400, 153 )', 1524099925581505828740029043592886075977656489255543250903551368284156938268905727090267904024955909290380333984000.0);
  Eps; Test('binomial(400, 154 )', 2444497932588519088953163466022356238743384109390384305020631090689524439950777367475949170741325386978726899312000.0);
  Eps; Test('binomial(400, 155 )', 3879654783334036747628891694461287966005628973613126058290808053610471046631556338058603200015264807721076240198400.0);
  Eps; Test('binomial(400, 156 )', 6093047576389993610058195289378304818406276272661640283854153673939521836055969889899729384639358191613228710568000.0);
  Eps; Test('binomial(400, 157 )', 9469449736555149304803819430626155259179180958786243498473971314912377885335392695130789616891741393335208951456000.0);
  Eps; Test('binomial(400, 158 )', 14563773961917096715616000769887061569497094765728209937526424237491821684408230537447986562687931383420606172176000.0);
  Eps; Test('binomial(400, 159 )', 22166247162163128334459573498821816979989288888718407577870406701088181431615042704795048730631945879168469771488000.0);
  Eps; Test('binomial(400, 160 )', 33387909788008212053779732582600361826108866388632101414167300093514073281370158074097542150514368480497507593303800.0);
  Eps; Test('binomial(400, 161 )', 49770797199515347160913887079652713281156074119700026952795975294679363897694645576294472770953095871549079642192000.0);
  Eps; Test('binomial(400, 162 )', 73427287226445481305298882790351842433310504411162385442705173428570172663882841313175178964554258724075494040024000.0);
  Eps; Test('binomial(400, 163 )', 107212848833705672089945608000636432509987116870286182425544977153372399349718504494084003641496402308772807248624000.0);
  Eps; Test('binomial(400, 164 )', 154935641302367343203153104244822161615042357916206251432037558447251577109044424177426273555089313092555825109292000.0);
  Eps; Test('binomial(400, 165 )', 221604917256719351490570500616836546309999978595301062654308265415462861804451418823470306418188350847534392277532800.0);
  Eps; Test('binomial(400, 166 )', 313717804550174985543879925572027640860542138372865962191340014292974533277386044719973024146230496681750495091688000.0);
  Eps; Test('binomial(400, 167 )', 439580636315814051600406602298529748271657846582339132651338702662012220280888230326189746408490636069039615876976000.0);
  Eps; Test('binomial(400, 168 )', 609656477747527821564849632949746615162477846748125106593820938811005043603850938488108398292728084548132324400806000.0);
  Eps; Test('binomial(400, 169 )', 836924868860511565698491803812669909572158937547722039821103300616290947432505430350539339668123761036489344739568000.0);
  Eps; Test('binomial(400, 170 )', 1137233204157518656919715333416039700653933615138375242345146249660960052099463261241026985078450522349582580204942400.0);
  Eps; Test('binomial(400, 171 )', 1529611912024732696441722378278883807897103692876177226546103142818835157794599708101966120280956842926339142965712000.0);
  Eps; Test('binomial(400, 172 )', 2036518185195719694681130375731769720979283405050259214413125695962286343807926355554361869443832075756579440343884000.0);
  Eps; Test('binomial(400, 173 )', 2683966163148116129406345235068459516666338822840804051365275483695961192995417393447367088053142851286127817331824000.0);
  Eps; Test('binomial(400, 174 )', 3501496086405875640087588323911151208524476510257830572758146751718294200057239932830760511425651880700867899622552000.0);
  Eps; Test('binomial(400, 175 )', 4521932088729873683770256921165258132151609664675826911104806662219054224073921284684296431898270428790835116083981440.0);
  Eps; Test('binomial(400, 176 )', 5780879090705804425274476177626040225762000991773074176128303971586859093276319824170265324869947991351919892720999000.0);
  Eps; Test('binomial(400, 177 )', 7315914781458193170968828608973067856331571876594172968659548529013878174541783280305872501530329661371921220166688000.0);
  Eps; Test('binomial(400, 178 )', 9165443799242567849022745953938169280685059148766857146129659112191544005184368941057357122703727609471564225265008000.0);
  Eps; Test('binomial(400, 179 )', 11367198454926536661916478222202645700067503525286269756652426384952641168440949189467783694079483403925627139714144000.0);
  Eps; Test('binomial(400, 180 )', 13956393658548692234908564928371026109527323772712586756778812394858520545696943171513223313286476845930908877093476800.0);
  Eps; Test('binomial(400, 181 )', 16963572402655869014805990520671965436994537182302591638073694623584942099742140871452536623884115503341436204202016000.0);
  Eps; Test('binomial(400, 182 )', 20412210748250743484848966615533848520339580455627843784275489684423639120019389290374206157311106017757002905056272000.0);
  Eps; Test('binomial(400, 183 )', 24316185481522743604902047662220650149912724258616775655585009569422695782318179591811895859529077114049325865039712000.0);
  Eps; Test('binomial(400, 184 )', 28677240486361061751433393166858049361581854152825219115554060198721331438929592235995551095205487683416868003878356000.0);
  Eps; Test('binomial(400, 185 )', 33482615919210753180051961751574803578928002686541877453836091907696257247615091475540751548996677511448883723447161600.0);
  Eps; Test('binomial(400, 186 )', 38703023777582322224253611702089154674567314933368299207391181506208039291598089608823987005560675618072634411511504000.0);
  Eps; Test('binomial(400, 187 )', 44291160900548753775349052963888123531323023506635379841613437659510804323005300407959001172138954985387934567184288000.0);
  Eps; Test('binomial(400, 188 )', 50180942935196194436964618517596650596658531951666680352466288412105326174468771206889719413114879850466117355373688000.0);
  Eps; Test('binomial(400, 189 )', 56287618530484620215007931882171904372971474993403895421814037795589043116335341247939791087726743535972576081159904000.0);
  Eps; Test('binomial(400, 190 )', 62508881631222394028245650669148799066826216966359062810540852499312042618667142122712083786896541505737966069077577600.0);
  Eps; Test('binomial(400, 191 )', 68727042631186925371369563562938470178185892999661796807400937302908528533613088197746270132189914744528653793226656000.0);
  Eps; Test('binomial(400, 192 )', 74812249530823267721959577003406980558546102275673518399722895293270221164193413715255054466810896779200461681168599500.0);
  Eps; Test('binomial(400, 193 )', 80626673069488288529365761744604414280712897789326900658768716170985523327213627216440680461640759223179772174523672000.0);
  Eps; Test('binomial(400, 194 )', 86029491367959153224632539593469658536636957950467363074047032203061872828521756875274334307008438964939241443950516000.0);
  Eps; Test('binomial(400, 195 )', 90882437034869669560381041826947434146395965834852701503865069917080747706028112391315450601249940650140942243352852800.0);
  Eps; Test('binomial(400, 196 )', 95055610164021848264684252931246040816383535694616345960675200678579353468039607348059527414572642006524965101465994000.0);
  Eps; Test('binomial(400, 197 )', 98433220677464248964444607096315697089046910059399667898364167200153239124264364969564180672958471925538542541619608000.0);
  Eps; Test('binomial(400, 198 )', 100918908068309305756476036568444881358972337081101679713979423947631856273867000448593579174800857580223859272468588000.0);
  Eps; Test('binomial(400, 199 )', 102440298642203415893508338627265658464886492916495172372984138881515753604628814525708055242762679553795073231350024000.0);
  Eps; Test('binomial(400, 200 )', 102952500135414432972975880320401986757210925381077648234849059575923332372651958598336595518976492951564048597506774120.0);
  Eps; Test('binomial(400, 201 )', 102440298642203415893508338627265658464886492916495172372984138881515753604628814525708055242762679553795073231350024000.0);
  Eps; Test('binomial(400, 202 )', 100918908068309305756476036568444881358972337081101679713979423947631856273867000448593579174800857580223859272468588000.0);
  Eps; Test('binomial(400, 203 )', 98433220677464248964444607096315697089046910059399667898364167200153239124264364969564180672958471925538542541619608000.0);
  Eps; Test('binomial(400, 204 )', 95055610164021848264684252931246040816383535694616345960675200678579353468039607348059527414572642006524965101465994000.0);
  Eps; Test('binomial(400, 205 )', 90882437034869669560381041826947434146395965834852701503865069917080747706028112391315450601249940650140942243352852800.0);
  Eps; Test('binomial(400, 206 )', 86029491367959153224632539593469658536636957950467363074047032203061872828521756875274334307008438964939241443950516000.0);
  Eps; Test('binomial(400, 207 )', 80626673069488288529365761744604414280712897789326900658768716170985523327213627216440680461640759223179772174523672000.0);
  Eps; Test('binomial(400, 208 )', 74812249530823267721959577003406980558546102275673518399722895293270221164193413715255054466810896779200461681168599500.0);
  Eps; Test('binomial(400, 209 )', 68727042631186925371369563562938470178185892999661796807400937302908528533613088197746270132189914744528653793226656000.0);
  Eps; Test('binomial(400, 210 )', 62508881631222394028245650669148799066826216966359062810540852499312042618667142122712083786896541505737966069077577600.0);
  Eps; Test('binomial(400, 211 )', 56287618530484620215007931882171904372971474993403895421814037795589043116335341247939791087726743535972576081159904000.0);
  Eps; Test('binomial(400, 212 )', 50180942935196194436964618517596650596658531951666680352466288412105326174468771206889719413114879850466117355373688000.0);
  Eps; Test('binomial(400, 213 )', 44291160900548753775349052963888123531323023506635379841613437659510804323005300407959001172138954985387934567184288000.0);
  Eps; Test('binomial(400, 214 )', 38703023777582322224253611702089154674567314933368299207391181506208039291598089608823987005560675618072634411511504000.0);
  Eps; Test('binomial(400, 215 )', 33482615919210753180051961751574803578928002686541877453836091907696257247615091475540751548996677511448883723447161600.0);
  Eps; Test('binomial(400, 216 )', 28677240486361061751433393166858049361581854152825219115554060198721331438929592235995551095205487683416868003878356000.0);
  Eps; Test('binomial(400, 217 )', 24316185481522743604902047662220650149912724258616775655585009569422695782318179591811895859529077114049325865039712000.0);
  Eps; Test('binomial(400, 218 )', 20412210748250743484848966615533848520339580455627843784275489684423639120019389290374206157311106017757002905056272000.0);
  Eps; Test('binomial(400, 219 )', 16963572402655869014805990520671965436994537182302591638073694623584942099742140871452536623884115503341436204202016000.0);
  Eps; Test('binomial(400, 220 )', 13956393658548692234908564928371026109527323772712586756778812394858520545696943171513223313286476845930908877093476800.0);
  Eps; Test('binomial(400, 221 )', 11367198454926536661916478222202645700067503525286269756652426384952641168440949189467783694079483403925627139714144000.0);
  Eps; Test('binomial(400, 222 )', 9165443799242567849022745953938169280685059148766857146129659112191544005184368941057357122703727609471564225265008000.0);
  Eps; Test('binomial(400, 223 )', 7315914781458193170968828608973067856331571876594172968659548529013878174541783280305872501530329661371921220166688000.0);
  Eps; Test('binomial(400, 224 )', 5780879090705804425274476177626040225762000991773074176128303971586859093276319824170265324869947991351919892720999000.0);
  Eps; Test('binomial(400, 225 )', 4521932088729873683770256921165258132151609664675826911104806662219054224073921284684296431898270428790835116083981440.0);
  Eps; Test('binomial(400, 226 )', 3501496086405875640087588323911151208524476510257830572758146751718294200057239932830760511425651880700867899622552000.0);
  Eps; Test('binomial(400, 227 )', 2683966163148116129406345235068459516666338822840804051365275483695961192995417393447367088053142851286127817331824000.0);
  Eps; Test('binomial(400, 228 )', 2036518185195719694681130375731769720979283405050259214413125695962286343807926355554361869443832075756579440343884000.0);
  Eps; Test('binomial(400, 229 )', 1529611912024732696441722378278883807897103692876177226546103142818835157794599708101966120280956842926339142965712000.0);
  Eps; Test('binomial(400, 230 )', 1137233204157518656919715333416039700653933615138375242345146249660960052099463261241026985078450522349582580204942400.0);
  Eps; Test('binomial(400, 231 )', 836924868860511565698491803812669909572158937547722039821103300616290947432505430350539339668123761036489344739568000.0);
  Eps; Test('binomial(400, 232 )', 609656477747527821564849632949746615162477846748125106593820938811005043603850938488108398292728084548132324400806000.0);
  Eps; Test('binomial(400, 233 )', 439580636315814051600406602298529748271657846582339132651338702662012220280888230326189746408490636069039615876976000.0);
  Eps; Test('binomial(400, 234 )', 313717804550174985543879925572027640860542138372865962191340014292974533277386044719973024146230496681750495091688000.0);
  Eps; Test('binomial(400, 235 )', 221604917256719351490570500616836546309999978595301062654308265415462861804451418823470306418188350847534392277532800.0);
  Eps; Test('binomial(400, 236 )', 154935641302367343203153104244822161615042357916206251432037558447251577109044424177426273555089313092555825109292000.0);
  Eps; Test('binomial(400, 237 )', 107212848833705672089945608000636432509987116870286182425544977153372399349718504494084003641496402308772807248624000.0);
  Eps; Test('binomial(400, 238 )', 73427287226445481305298882790351842433310504411162385442705173428570172663882841313175178964554258724075494040024000.0);
  Eps; Test('binomial(400, 239 )', 49770797199515347160913887079652713281156074119700026952795975294679363897694645576294472770953095871549079642192000.0);
  Eps; Test('binomial(400, 240 )', 33387909788008212053779732582600361826108866388632101414167300093514073281370158074097542150514368480497507593303800.0);
  Eps; Test('binomial(400, 241 )', 22166247162163128334459573498821816979989288888718407577870406701088181431615042704795048730631945879168469771488000.0);
  Eps; Test('binomial(400, 242 )', 14563773961917096715616000769887061569497094765728209937526424237491821684408230537447986562687931383420606172176000.0);
  Eps; Test('binomial(400, 243 )', 9469449736555149304803819430626155259179180958786243498473971314912377885335392695130789616891741393335208951456000.0);
  Eps; Test('binomial(400, 244 )', 6093047576389993610058195289378304818406276272661640283854153673939521836055969889899729384639358191613228710568000.0);
  Eps; Test('binomial(400, 245 )', 3879654783334036747628891694461287966005628973613126058290808053610471046631556338058603200015264807721076240198400.0);
  Eps; Test('binomial(400, 246 )', 2444497932588519088953163466022356238743384109390384305020631090689524439950777367475949170741325386978726899312000.0);
  Eps; Test('binomial(400, 247 )', 1524099925581505828740029043592886075977656489255543250903551368284156938268905727090267904024955909290380333984000.0);
  Eps; Test('binomial(400, 248 )', 940271325056332224988808240603675683970086463129427892694529674788209724012671678406495924660557476296081415724000.0);
  Eps; Test('binomial(400, 249 )', 573980889191014048989152018360476722744791736528807388311520122762280634738659016537298717061866411232949297952000.0);
  Eps; Test('binomial(400, 250 )', 346684457071372485589447819089727940537854208863399662540158154148417503382150045988528425105367312384701375963008.0);
  Eps; Test('binomial(400, 251 )', 207181946456995509316403079137287613867243551113585455701289733554831177320009987642546867592848991464960981651200.0);
  Eps; Test('binomial(400, 252 )', 122500436595604487651365312664507358992933686967953305156714961506626370716990032375950330441803570350314231214400.0);
  Eps; Test('binomial(400, 253 )', 71660334451183652855344135471727625023534330716431182463216657324034398680294564393836556938288254592278680710400.0);
  Eps; Test('binomial(400, 254 )', 41472713245370066809982629584031342041179317383131432370444285931626207110249216401157377440662887500255771907200.0);
  Eps; Test('binomial(400, 255 )', 23745161309113842173558682036347356619655609168381133827783787239284024463123080763015596495438359117793500778240.0);
  Eps; Test('binomial(400, 256 )', 13449407772740262168617222247149869960351809880528376582143160741000716981065807463426802702494383094062725050175.0);
  Eps; Test('binomial(400, 257 )', 7535854938811664405762179002294090561442259232669596217231965551377833639196405738262488673771171850369775903600.0);
  Eps; Test('binomial(400, 258 )', 4176849830426620193891440299721143218163732830510667670791360751345078334903434188261766978097975095359992070600.0);
  Eps; Test('binomial(400, 259 )', 2290010331739691380434689276294989718066602555723995402518815547069502407553234188158961045906997928730188702800.0);
  Eps; Test('binomial(400, 260 )', 1241890218366524940928043030606129039413042155219551352904434585141537844096177002040051951818795030580602334980.0);
  Eps; Test('binomial(400, 261 )', 666148009851775830382858330593325921524237171382134825312723532259828728633964675423782656148012660081549145200.0);
  Eps; Test('binomial(400, 262 )', 353414402173270383294722549436917187373545674893575346253696835817237378931759885053075531315167021951661569400.0);
  Eps; Test('binomial(400, 263 )', 185441777566202710626128181833819664857601913062028128452510126778626457386246631700853320614042011518362344400.0);
  Eps; Test('binomial(400, 264 )', 96233043661249133923407427694065507899588871551128233325734421851029638870893138420518579258044528704604701450.0);
  Eps; Test('binomial(400, 265 )', 49387524294075027221069472326010977639034288796050715970942948572603890137514969151662365204128512844627318480.0);
  Eps; Test('binomial(400, 266 )', 25065096916165897273851047985005571358156499952882882165704128034968139731445567050655711663749433210619127800.0);
  Eps; Test('binomial(400, 267 )', 12579486841821087021333484756519650044917494358375678689903944407062661887691782714561293494166382210572895600.0);
  Eps; Test('binomial(400, 268 )', 6242805037172405126258781614242960656619502797253601737900091813952738921876892168047209084791525500023116100.0);
  Eps; Test('binomial(400, 269 )', 3063383884411737831472710680595058760869049699767566652055063641047440660549255636365173231198815487000190800.0);
  Eps; Test('binomial(400, 270 )', 1486308477251620947862685552436861843236464854331671227478567914730424913081305512458658123285351217766759240.0);
  Eps; Test('binomial(400, 271 )', 712989306430666875358483844342406050261034800970912396945438483081015640961511869445112752867511654279257200.0);
  Eps; Test('binomial(400, 272 )', 338145663711603040151633882059449928248799593107528305904270457049452270897187614553013033529077218389794775.0);
  Eps; Test('binomial(400, 273 )', 158544487014964062781718450196372127530572702995471147090646954220988610530549504259288162240739501662614400.0);
  Eps; Test('binomial(400, 274 )', 73485948360950496252840303558172482468550121461404509782891106518487421669269295769816046002094586537051200.0);
  Eps; Test('binomial(400, 275 )', 33669925430835500101301375448471755603772055651407157209615561532106964110283386425442988350050610558794368.0);
  Eps; Test('binomial(400, 276 )', 15249060430632019973415478011083222646635894769659038591311395621425255484729794576740483854189588115396000.0);
  Eps; Test('binomial(400, 277 )', 6826294199994117244417037087993933603548198380641591282753115729446684765727417066844115515954905871152000.0);
  Eps; Test('binomial(400, 278 )', 3020266858270778493033437272745517385742548204384588948844004441445835345987310428855489958498033892632000.0);
  Eps; Test('binomial(400, 279 )', 1320690167415895971864083681989079286955522870734479755408489397334738036596601692904551164647885788176000.0);
  Eps; Test('binomial(400, 280 )', 570726822347583616412693305430994977577208097710257322872954346705368937243531445862323896151407787033200.0);
  Eps; Test('binomial(400, 281 )', 243726756874412932275883262105762979748273920730358999091653101795887090637807023144052909388501546064000.0);
  Eps; Test('binomial(400, 282 )', 102849234283883471421383362377963810602995023286924542169881982672732495694677431752277646160396042488000.0);
  Eps; Test('binomial(400, 283 )', 42884133022962012818810023889044981099482023844018006982494960973082807392126985677628135148151000048000.0);
  Eps; Test('binomial(400, 284 )', 17667054801713223590847791531754446438871115456866573299126445189615100228446680719304548634977700724000.0);
  Eps; Test('binomial(400, 285 )', 7190801252627136619432785325205318550558068045601833342802342603492461847367771801541500497043555382400.0);
  Eps; Test('binomial(400, 286 )', 2891406098084338151170525567827313403196425962392345574903039858047668225340187962158295654405625416000.0);
  Eps; Test('binomial(400, 287 )', 1148502770667646513008501445060326578273144807361419496651381685775031978009691385665664475965997552000.0);
  Eps; Test('binomial(400, 288 )', 450627823213347416562363414207697581058560288999445844172243508654786852482969189514653075639436539500.0);
  Eps; Test('binomial(400, 289 )', 174637772317975469394410734917862038334113329992864825423153193665522932450147229154467627929470216000.0);
  Eps; Test('binomial(400, 290 )', 66844112852742334837171005434078228465815791824855157317137946540941536213677042883261747241969634400.0);
  Eps; Test('binomial(400, 291 )', 25267534068046930694463266658929914540342739177780300016787539929565529152936339234222653596620824000.0);
  Eps; Test('binomial(400, 292 )', 9432058950058614540056493376107399605812871816363194184348773466858365334486510193596812472711198000.0);
  Eps; Test('binomial(400, 293 )', 3476663367257100240020823496995218967330341829922269528701936977545062990186153928015207327825288000.0);
  Eps; Test('binomial(400, 294 )', 1265316259511937842456558211491457243212063182998921223030977063256196394387477790127983619310564000.0);
  Eps; Test('binomial(400, 295 )', 454656011892425123052187018366421924679588804738595422512825656627650229847703883910394114057355200.0);
  Eps; Test('binomial(400, 296 )', 161280004218596749731350124758359128687016298978217970823806398465889439641921985846592506675751000.0);
  Eps; Test('binomial(400, 297 )', 56475152992370579030506440992826092200167323547928178335608974546978120278652816592746197623832000.0);
  Eps; Test('binomial(400, 298 )', 19519935430248891409873031618325796968514209145760410632777598585029350297655168151184088440452000.0);
  Eps; Test('binomial(400, 299 )', 6658974628379220480960030853074352143105181715276126704158244333354494081474338299066143882696000.0);
  Eps; Test('binomial(400, 300 )', 2241854791554337561923210387201698554845411177476295990399942258896013007429693894018935107174320.0);
  Eps; Test('binomial(400, 301 )', 744802256330344704957877205050398190978541919427340860597987461427246846322157439873400367832000.0);
  Eps; Test('binomial(400, 302 )', 244157031048689158247780938079435168565813410673201143043711121461249793993025121018101445084000.0);
  Eps; Test('binomial(400, 303 )', 78968280669212995076840039378827216235807637775490798740210197700338217199064230560310038344000.0);
  Eps; Test('binomial(400, 304 )', 25197115871426514876491723091270526233135989684942787755922332818857917987859310409046295129500.0);
  Eps; Test('binomial(400, 305 )', 7930895487399821075879362022170395142232967245096746310060799838066754514211455079568669942400.0);
  Eps; Test('binomial(400, 306 )', 2462206115369225497413527425183619406902391791778401632208418250380201564869569387447789688000.0);
  Eps; Test('binomial(400, 307 )', 753900243793834517123360188818437212536888691945178349927007542461690381425861636547531696000.0);
  Eps; Test('binomial(400, 308 )', 227638709976709773027508108961411236253021585554875280984453576132913004781185494152339116000.0);
  Eps; Test('binomial(400, 309 )', 67775926595007440513044485516018879402194129032519501134529867327598693980158787902961808000.0);
  Eps; Test('binomial(400, 310 )', 19895513935953797053829187683734574276127954006320240655620057828424132749014353868288788800.0);
  Eps; Test('binomial(400, 311 )', 5757544225838719404645102545132191912705838779964056781369148567711163818042739061562672000.0);
  Eps; Test('binomial(400, 312 )', 1642376397755275727607096559348606026380832216079490556223891738866325576300653129740634000.0);
  Eps; Test('binomial(400, 313 )', 461754386589342696579630981542100096873844201325863159577324194952832749886445608361584000.0);
  Eps; Test('binomial(400, 314 )', 127938317303416607014101577688416268879058743679458900901997468028332640892104356456872000.0);
  Eps; Test('binomial(400, 315 )', 34929191390774057787976938670488251186028736369630049135148515080751133703876110016796800.0);
  Eps; Test('binomial(400, 316 )', 9395510342455047189803923376555384021558362631071373976226657537543817610219839719708000.0);
  Eps; Test('binomial(400, 317 )', 2489662046581148151241418181800164851138493567854875123037978653481642521320083711216000.0);
  Eps; Test('binomial(400, 318 )', 649817452409544957713955060029602775611619390352058601296076189430743173803669647896000.0);
  Eps; Test('binomial(400, 319 )', 167037715039444158409229827343032688401732884040341082464822092580943386369595332688000.0);
  Eps; Test('binomial(400, 320 )', 42281421619359302597336300046205149251688636272711336498908092184551294674803818586650.0);
  Eps; Test('binomial(400, 321 )', 10537425948749981955722442379116548100109317451143012211565879672162316429857649492000.0);
  Eps; Test('binomial(400, 322 )', 2585269099227480045037493627174556832014397759752478151284796565530506204840851894000.0);
  Eps; Test('binomial(400, 323 )', 624306469782487441216484529162896077080876239197192866254532916753496854419772284000.0);
  Eps; Test('binomial(400, 324 )', 148369130164356583252065767733157401034652686475876082412342699351911289476303907000.0);
  Eps; Test('binomial(400, 325 )', 34695550438434154852790764146830653780411089760512560810270908156139255385227990560.0);
  Eps; Test('binomial(400, 326 )', 7982105162216446668586832242369015440278624944903196505430423655553509674515642000.0);
  Eps; Test('binomial(400, 327 )', 1806347957198828909710781608364853647035529804045371686244193732449418091480604000.0);
  Eps; Test('binomial(400, 328 )', 402022563644861312222216638447055842175590474680829674072640678258559514262451500.0);
  Eps; Test('binomial(400, 329 )', 87980621831094268936168990784766020172165696586686129280334738099137644458652000.0);
  Eps; Test('binomial(400, 330 )', 18929164090932403316569691956722386158253831689862773269405352742541735625952400.0);
  Eps; Test('binomial(400, 331 )', 4003146484487215202899934854895972903558212139850133319813820821685563425428000.0);
  Eps; Test('binomial(400, 332 )', 831979239245836894578600918638018464896134450751985539358896496073204446851000.0);
  Eps; Test('binomial(400, 333 )', 169893658464615341835870457860015782621432860814219269298513398597531238396000.0);
  Eps; Test('binomial(400, 334 )', 34080464422542598511985990049763645016874256510636799529941310497109559798000.0);
  Eps; Test('binomial(400, 335 )', 6714360154888989557585299532192240510787166954334414235749631321818599840800.0);
  Eps; Test('binomial(400, 336 )', 1298908958535072384651918064263379860717755511999217039653946535470860088250.0);
  Eps; Test('binomial(400, 337 )', 246677072244049354948732213984736828148179088332195520883835543828293904000.0);
  Eps; Test('binomial(400, 338 )', 45978270862056536573284406748634379211051131848900348567105441601131704000.0);
  Eps; Test('binomial(400, 339 )', 8409005290405620258240805954027526581372183405993574074219874275133232000.0);
  Eps; Test('binomial(400, 340 )', 1508674478572773046331438715281409180775597611075317701551212737597432800.0);
  Eps; Test('binomial(400, 341 )', 265455919983479128386763410313444430635002512212665871240682593125648000.0);
  Eps; Test('binomial(400, 342 )', 45795027131652832090114155580389536279137860293997913459649921036296000.0);
  Eps; Test('binomial(400, 343 )', 7743765520804268983167991322631466776064127979743087407171123673776000.0);
  Eps; Test('binomial(400, 344 )', 1283123938040242244303998562180213971615277019899290645955680376178000.0);
  Eps; Test('binomial(400, 345 )', 208275189942763958495721505745194151914363806128580510647878553814400.0);
  Eps; Test('binomial(400, 346 )', 33107327881075195714637811606895024148237021205410196779287053352000.0);
  Eps; Test('binomial(400, 347 )', 5152149007429569361932109010871271769466279957037898057871760464000.0);
  Eps; Test('binomial(400, 348 )', 784666371821170046501154533264877597073887464721289072032193404000.0);
  Eps; Test('binomial(400, 349 )', 116913041073641382286704973437746805294676642308043070904510192000.0);
  Eps; Test('binomial(400, 350 )', 17035900270730601418919867558071677342938596450600561760371485120.0);
  Eps; Test('binomial(400, 351 )', 2426766420331994504119639253286563724065327129715179737944656000.0);
  Eps; Test('binomial(400, 352 )', 337816916466669689493927055145004609315912015216033543066159500.0);
  Eps; Test('binomial(400, 353 )', 45935444732011742480760619396487878887149509151188697074152000.0);
  Eps; Test('binomial(400, 354 )', 6098773735606078803942794100663644936994426356231267690636000.0);
  Eps; Test('binomial(400, 355 )', 790263638979942605581319798959232865075334119398981165547200.0);
  Eps; Test('binomial(400, 356 )', 99892875713756790031346603801026626203342795991444248454000.0);
  Eps; Test('binomial(400, 357 )', 12311726978726327062686976378838015554473621914911896168000.0);
  Eps; Test('binomial(400, 358 )', 1478782849400089563395363084609035387827837269109529428000.0);
  Eps; Test('binomial(400, 359 )', 173005235862963124408371168672923360135847257110307064000.0);
  Eps; Test('binomial(400, 360 )', 19703374084393022502064494209971827126582604282007193400.0);
  Eps; Test('binomial(400, 361 )', 2183199344531082825713517363985798019565939532632376000.0);
  Eps; Test('binomial(400, 362 )', 235206559217437100007809881755376029732242104344372000.0);
  Eps; Test('binomial(400, 363 )', 24622174243147685400266599192022835068388980620072000.0);
  Eps; Test('binomial(400, 364 )', 2502803425814462526950176291496826641567011766326000.0);
  Eps; Test('binomial(400, 365 )', 246851844737864797178647524640782901634006639966400.0);
  Eps; Test('binomial(400, 366 )', 23606050726298546178285965471113119008716481964000.0);
  Eps; Test('binomial(400, 367 )', 2186936579548094196353468190784321652033679528000.0);
  Eps; Test('binomial(400, 368 )', 196111160665997577390392527977941887274759305500.0);
  Eps; Test('binomial(400, 369 )', 17006929922254532456619406220309323557702704000.0);
  Eps; Test('binomial(400, 370 )', 1424904939432136503122166467106997379158875200.0);
  Eps; Test('binomial(400, 371 )', 115221423673757668716078150979002483489936000.0);
  Eps; Test('binomial(400, 372 )', 8982315286395087077328673060191053820452000.0);
  Eps; Test('binomial(400, 373 )', 674275678335288037976415135885655514672000.0);
  Eps; Test('binomial(400, 374 )', 48677655922601008089206440291210424856000.0);
  Eps; Test('binomial(400, 375 )', 3374984143967003227518313193523922790016.0);
  Eps; Test('binomial(400, 376 )', 224400541487167767787121887867282100400.0);
  Eps; Test('binomial(400, 377 )', 14285445611915189461249138750171804800.0);
  Eps; Test('binomial(400, 378 )', 869220235645633221187116908079236800.0);
  Eps; Test('binomial(400, 379 )', 50456055894997179066270638463702400.0);
  Eps; Test('binomial(400, 380 )', 2788360983670896737872851072994080.0);
  Eps; Test('binomial(400, 381 )', 146370655310808227709860948713600.0);
  Eps; Test('binomial(400, 382 )', 7280215840066377818029733051200.0);
  Eps; Test('binomial(400, 383 )', 342151136086670497975287715200.0);
  Eps; Test('binomial(400, 384 )', 15147315920503641837447633225.0);
  Eps; Test('binomial(400, 385 )', 629498843449501998439382160.0);
  Eps; Test('binomial(400, 386 )', 24462390289488419628473400.0);
  Eps; Test('binomial(400, 387 )', 884944351557720606714800.0);
  Eps; Test('binomial(400, 388 )', 29650197346006102802300.0);
  Eps; Test('binomial(400, 389 )', 914659044092733248400.0);
  Eps; Test('binomial(400, 390 )', 25798075602615553160.0);
  Test('binomial(400, 391 )', 659797329990167600);
  Test('binomial(400, 392 )', 15148408086508950);
  Test('binomial(400, 393 )', 308364541201200);
  Test('binomial(400, 394 )', 5478557838600);
  Test('binomial(400, 395 )', 83218600080);
  Test('binomial(400, 396 )', 1050739900);
  Test('binomial(400, 397 )', 10586800);
  Test('binomial(400, 398 )', 79800);
  Test('binomial(400, 399 )', 400);
  Test('binomial(400, 400 )', 1);

  Eps; Test('binomial(500, 250)', 1.1674431578827768292093473476217661965923008118031144E149);
  Eps; Test('binomial(800, 400)', 1.8804244186835312700958607615195351332156581822914058E239);
  Eps; Test('binomial(1000, 500)', 2.7028824094543656951561469362597527549615200844654828E299);
  Eps; Test('binomial(1200, 600)', 3.9650964622610201403671601486260572953160819570679165E359);
  Eps; Test('binomial(1500, 750)', 7.2246289244893021702811607322848529594437634304054652E449);
  Eps; Test('binomial(2000, 1000)', 2.0481516269894897143351625029808250443964248879813970E600);
  Eps; Test('binomial(3000, 1500)', 1.7919679375475600507326903666685292846560897853633339E901);
  Eps; Test('binomial(4000, 2000)', 1.662897875033835069539536826460381558016225597388640E1202);
  Eps; Test('binomial(5000, 2500)', 1.593718685349438375691025792935919087257129434168729E1503);
  Eps; Test('binomial(8000, 4000)', 1.550053928509880709855382381537398942593569455195730E2406);
  Eps; Test('binomial(10000, 5000)', 1.591790263532438948337597273641521188653005837457614E3008);

  Test('binomial(0, 1)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(1, 2)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(4, 5)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(10, 15)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(10, 11)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(-10, 5)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(10, -1)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(10, -3)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(10, -11)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(-10, -5)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(-10, -50)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(-1, 0)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('binomial(10, 5, 2)', failure, 'Too many arguments.');
  Test('binomial(10)', failure, 'A required argument of type integer is missing.');
  Test('binomial(10, "cat")', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
  Test('binomial(10, 5.5)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');

  Test('combinations(0, 0)', 1);

  Test('combinations(1, 0)', 1);
  Test('combinations(1, 1)', 1);

  Test('combinations(2, 0)', 1);
  Test('combinations(2, 1)', 2);
  Test('combinations(2, 2)', 1);

  Test('combinations(5, 0)', 1);
  Test('combinations(5, 1)', 5);
  Test('combinations(5, 2)', 10);
  Test('combinations(5, 3)', 10);
  Test('combinations(5, 4)', 5);
  Test('combinations(5, 5)', 1);

  Test('combinations(10, 0)', 1);
  Test('combinations(10, 1)', 10);
  Test('combinations(10, 2)', 45);
  Test('combinations(10, 3)', 120);
  Test('combinations(10, 4)', 210);
  Test('combinations(10, 5)', 252);
  Test('combinations(10, 6)', 210);
  Test('combinations(10, 7)', 120);
  Test('combinations(10, 8)', 45);
  Test('combinations(10, 9)', 10);
  Test('combinations(10, 10)', 1);

  Test('combinations(20, 0)', 1);
  Test('combinations(20, 1)', 20);
  Test('combinations(20, 2)', 190);
  Test('combinations(20, 3)', 1140);
  Test('combinations(20, 4)', 4845);
  Test('combinations(20, 5)', 15504);
  Test('combinations(20, 6)', 38760);
  Test('combinations(20, 7)', 77520);
  Test('combinations(20, 8)', 125970);
  Test('combinations(20, 9)', 167960);
  Test('combinations(20, 10)', 184756);
  Test('combinations(20, 11)', 167960);
  Test('combinations(20, 12)', 125970);
  Test('combinations(20, 13)', 77520);
  Test('combinations(20, 14)', 38760);
  Test('combinations(20, 15)', 15504);
  Test('combinations(20, 16)', 4845);
  Test('combinations(20, 17)', 1140);
  Test('combinations(20, 18)', 190);
  Test('combinations(20, 19)', 20);
  Test('combinations(20, 20)', 1);

  Test('combinations(50, 0)', 1);
  Test('combinations(50, 1)', 50);
  Test('combinations(50, 2)', 1225);
  Test('combinations(50, 3)', 19600);
  Test('combinations(50, 4)', 230300);
  Test('combinations(50, 5)', 2118760);
  Test('combinations(50, 6)', 15890700);
  Test('combinations(50, 7)', 99884400);
  Test('combinations(50, 8)', 536878650);
  Test('combinations(50, 9)', 2505433700);
  Test('combinations(50, 10)', 10272278170);
  Test('combinations(50, 11)', 37353738800);
  Test('combinations(50, 12)', 121399651100);
  Test('combinations(50, 13)', 354860518600);
  Test('combinations(50, 14)', 937845656300);
  Test('combinations(50, 15)', 2250829575120);
  Test('combinations(50, 16)', 4923689695575);
  Test('combinations(50, 17)', 9847379391150);
  Test('combinations(50, 18)', 18053528883775);
  Test('combinations(50, 19)', 30405943383200);
  Test('combinations(50, 20)', 47129212243960);
  Test('combinations(50, 21)', 67327446062800);
  Test('combinations(50, 22)', 88749815264600);
  Test('combinations(50, 23)', 108043253365600);
  Test('combinations(50, 24)', 121548660036300);
  Test('combinations(50, 25)', 126410606437752);
  Test('combinations(50, 26)', 121548660036300);
  Test('combinations(50, 27)', 108043253365600);
  Test('combinations(50, 28)', 88749815264600);
  Test('combinations(50, 29)', 67327446062800);
  Test('combinations(50, 30)', 47129212243960);
  Test('combinations(50, 31)', 30405943383200);
  Test('combinations(50, 32)', 18053528883775);
  Test('combinations(50, 33)', 9847379391150);
  Test('combinations(50, 34)', 4923689695575);
  Test('combinations(50, 35)', 2250829575120);
  Test('combinations(50, 36)', 937845656300);
  Test('combinations(50, 37)', 354860518600);
  Test('combinations(50, 38)', 121399651100);
  Test('combinations(50, 39)', 37353738800);
  Test('combinations(50, 40)', 10272278170);
  Test('combinations(50, 41)', 2505433700);
  Test('combinations(50, 42)', 536878650);
  Test('combinations(50, 43)', 99884400);
  Test('combinations(50, 44)', 15890700);
  Test('combinations(50, 45)', 2118760);
  Test('combinations(50, 46)', 230300);
  Test('combinations(50, 47)', 19600);
  Test('combinations(50, 48)', 1225);
  Test('combinations(50, 49)', 50);
  Test('combinations(50, 50)', 1);

  Test('combinations(100, 0   )', 1);
  Test('combinations(100, 1   )', 100);
  Test('combinations(100, 2   )', 4950);
  Test('combinations(100, 3   )', 161700);
  Test('combinations(100, 4   )', 3921225);
  Test('combinations(100, 5   )', 75287520);
  Test('combinations(100, 6   )', 1192052400);
  Test('combinations(100, 7   )', 16007560800);
  Test('combinations(100, 8   )', 186087894300);
  Test('combinations(100, 9   )', 1902231808400);
  Test('combinations(100, 10  )', 17310309456440);
  Test('combinations(100, 11  )', 141629804643600);
  Test('combinations(100, 12  )', 1050421051106700);
  Test('combinations(100, 13  )', 7110542499799200);
  Test('combinations(100, 14  )', 44186942677323600);
  Test('combinations(100, 15  )', 253338471349988640);
  Test('combinations(100, 16  )', 1345860629046814650);
  Test('combinations(100, 17  )', 6650134872937201800);
  Eps; Test('combinations(100, 18  )', 30664510802988208300.0);
  Eps; Test('combinations(100, 19  )', 132341572939212267400.0);
  Eps; Test('combinations(100, 20  )', 535983370403809682970.0);
  Eps; Test('combinations(100, 21  )', 2041841411062132125600.0);
  Eps; Test('combinations(100, 22  )', 7332066885177656269200.0);
  Eps; Test('combinations(100, 23  )', 24865270306254660391200.0);
  Eps; Test('combinations(100, 24  )', 79776075565900368755100.0);
  Eps; Test('combinations(100, 25  )', 242519269720337121015504.0);
  Eps; Test('combinations(100, 26  )', 699574816500972464467800.0);
  Eps; Test('combinations(100, 27  )', 1917353200780443050763600.0);
  Eps; Test('combinations(100, 28  )', 4998813702034726525205100.0);
  Eps; Test('combinations(100, 29  )', 12410847811948286545336800.0);
  Eps; Test('combinations(100, 30  )', 29372339821610944823963760.0);
  Eps; Test('combinations(100, 31  )', 66324638306863423796047200.0);
  Eps; Test('combinations(100, 32  )', 143012501349174257560226775.0);
  Eps; Test('combinations(100, 33  )', 294692427022540894366527900.0);
  Eps; Test('combinations(100, 34  )', 580717429720889409486981450.0);
  Eps; Test('combinations(100, 35  )', 1095067153187962886461165020.0);
  Eps; Test('combinations(100, 36  )', 1977204582144932989443770175.0);
  Eps; Test('combinations(100, 37  )', 3420029547493938143902737600.0);
  Eps; Test('combinations(100, 38  )', 5670048986634686922786117600.0);
  Eps; Test('combinations(100, 39  )', 9013924030034630492634340800.0);
  Eps; Test('combinations(100, 40  )', 13746234145802811501267369720.0);
  Eps; Test('combinations(100, 41  )', 20116440213369968050635175200.0);
  Eps; Test('combinations(100, 42  )', 28258808871162574166368460400.0);
  Eps; Test('combinations(100, 43  )', 38116532895986727945334202400.0);
  Eps; Test('combinations(100, 44  )', 49378235797073715747364762200.0);
  Eps; Test('combinations(100, 45  )', 61448471214136179596720592960.0);
  Eps; Test('combinations(100, 46  )', 73470998190814997343905056800.0);
  Eps; Test('combinations(100, 47  )', 84413487283064039501507937600.0);
  Eps; Test('combinations(100, 48  )', 93206558875049876949581681100.0);
  Eps; Test('combinations(100, 49  )', 98913082887808032681188722800.0);
  Eps; Test('combinations(100, 50  )', 100891344545564193334812497256.0);
  Eps; Test('combinations(100, 51  )', 98913082887808032681188722800.0);
  Eps; Test('combinations(100, 52  )', 93206558875049876949581681100.0);
  Eps; Test('combinations(100, 53  )', 84413487283064039501507937600.0);
  Eps; Test('combinations(100, 54  )', 73470998190814997343905056800.0);
  Eps; Test('combinations(100, 55  )', 61448471214136179596720592960.0);
  Eps; Test('combinations(100, 56  )', 49378235797073715747364762200.0);
  Eps; Test('combinations(100, 57  )', 38116532895986727945334202400.0);
  Eps; Test('combinations(100, 58  )', 28258808871162574166368460400.0);
  Eps; Test('combinations(100, 59  )', 20116440213369968050635175200.0);
  Eps; Test('combinations(100, 60  )', 13746234145802811501267369720.0);
  Eps; Test('combinations(100, 61  )', 9013924030034630492634340800.0);
  Eps; Test('combinations(100, 62  )', 5670048986634686922786117600.0);
  Eps; Test('combinations(100, 63  )', 3420029547493938143902737600.0);
  Eps; Test('combinations(100, 64  )', 1977204582144932989443770175.0);
  Eps; Test('combinations(100, 65  )', 1095067153187962886461165020.0);
  Eps; Test('combinations(100, 66  )', 580717429720889409486981450.0);
  Eps; Test('combinations(100, 67  )', 294692427022540894366527900.0);
  Eps; Test('combinations(100, 68  )', 143012501349174257560226775.0);
  Eps; Test('combinations(100, 69  )', 66324638306863423796047200.0);
  Eps; Test('combinations(100, 70  )', 29372339821610944823963760.0);
  Eps; Test('combinations(100, 71  )', 12410847811948286545336800.0);
  Eps; Test('combinations(100, 72  )', 4998813702034726525205100.0);
  Eps; Test('combinations(100, 73  )', 1917353200780443050763600.0);
  Eps; Test('combinations(100, 74  )', 699574816500972464467800.0);
  Eps; Test('combinations(100, 75  )', 242519269720337121015504.0);
  Eps; Test('combinations(100, 76  )', 79776075565900368755100.0);
  Eps; Test('combinations(100, 77  )', 24865270306254660391200.0);
  Eps; Test('combinations(100, 78  )', 7332066885177656269200.0);
  Eps; Test('combinations(100, 79  )', 2041841411062132125600.0);
  Eps; Test('combinations(100, 80  )', 535983370403809682970.0);
  Eps; Test('combinations(100, 81  )', 132341572939212267400.0);
  Eps; Test('combinations(100, 82  )', 30664510802988208300.0);
  Test('combinations(100, 83  )', 6650134872937201800);
  Test('combinations(100, 84  )', 1345860629046814650);
  Test('combinations(100, 85  )', 253338471349988640);
  Test('combinations(100, 86  )', 44186942677323600);
  Test('combinations(100, 87  )', 7110542499799200);
  Test('combinations(100, 88  )', 1050421051106700);
  Test('combinations(100, 89  )', 141629804643600);
  Test('combinations(100, 90  )', 17310309456440);
  Test('combinations(100, 91  )', 1902231808400);
  Test('combinations(100, 92  )', 186087894300);
  Test('combinations(100, 93  )', 16007560800);
  Test('combinations(100, 94  )', 1192052400);
  Test('combinations(100, 95  )', 75287520);
  Test('combinations(100, 96  )', 3921225);
  Test('combinations(100, 97  )', 161700);
  Test('combinations(100, 98  )', 4950);
  Test('combinations(100, 99  )', 100);
  Test('combinations(100, 100 )', 1);

  Test('combinations(200, 0   )', 1);
  Test('combinations(200, 1   )', 200);
  Test('combinations(200, 2   )', 19900);
  Test('combinations(200, 3   )', 1313400);
  Test('combinations(200, 4   )', 64684950);
  Test('combinations(200, 5   )', 2535650040);
  Test('combinations(200, 6   )', 82408626300);
  Test('combinations(200, 7   )', 2283896214600);
  Test('combinations(200, 8   )', 55098996177225);
  Test('combinations(200, 9   )', 1175445251780800);
  Test('combinations(200, 10  )', 22451004309013280);
  Test('combinations(200, 11  )', 387790074428411200);
  Test('combinations(200, 12  )', 6107693672247476400);
  Eps; Test('combinations(200, 13  )', 88326646952501966400.0);
  Eps; Test('combinations(200, 14  )', 1179791641436990551200.0);
  Eps; Test('combinations(200, 15  )', 14629416353818682834880.0);
  Eps; Test('combinations(200, 16  )', 169152626591028520278300.0);
  Eps; Test('combinations(200, 17  )', 1830828428985249866541600.0);
  Eps; Test('combinations(200, 18  )', 18613422361350040309839600.0);
  Eps; Test('combinations(200, 19  )', 178296993145563544020568800.0);
  Eps; Test('combinations(200, 20  )', 1613587787967350073386147640.0);
  Eps; Test('combinations(200, 21  )', 13830752468291572057595551200.0);
  Eps; Test('combinations(200, 22  )', 112532031446554154468618348400.0);
  Eps; Test('combinations(200, 23  )', 870900069455940847626698522400.0);
  Eps; Test('combinations(200, 24  )', 6422888012237563751246901602700.0);
  Eps; Test('combinations(200, 25  )', 45217131606152448808778187283008.0);
  Eps; Test('combinations(200, 26  )', 304346078118333790059083952866400.0);
  Eps; Test('combinations(200, 27  )', 1961341392318151091491874362916800.0);
  Eps; Test('combinations(200, 28  )', 12118287888251433529574795170878800.0);
  Eps; Test('combinations(200, 29  )', 71873983337215398865064302392798400.0);
  Eps; Test('combinations(200, 30  )', 409681705022127773530866523638950880.0);
  Eps; Test('combinations(200, 31  )', 2246641608185861983878945452213601600.0);
  Eps; Test('combinations(200, 32  )', 11865075993231583602360680669503083450.0);
  Eps; Test('combinations(200, 33  )', 60404023238269880157472556135652061200.0);
  Eps; Test('combinations(200, 34  )', 296690349435031470185232849254526300600.0);
  Eps; Test('combinations(200, 35  )', 1407159943034720687164247227892896168560.0);
  Eps; Test('combinations(200, 36  )', 6449483072242469816169466461175774105900.0);
  Eps; Test('combinations(200, 37  )', 28586897941831487833832229719806133874800.0);
  Eps; Test('combinations(200, 38  )', 122622746434698224129332985377063153199800.0);
  Eps; Test('combinations(200, 39  )', 509356023651823392537229323873954636368400.0);
  Eps; Test('combinations(200, 40  )', 2050157995198589154962348028592667411382810.0);
  Eps; Test('combinations(200, 41  )', 8000616566628640604731114257922604532225600.0);
  Eps; Test('combinations(200, 42  )', 30288048430808425146482075404992717157711200.0);
  Eps; Test('combinations(200, 43  )', 111290968652737934259166695674159286300427200.0);
  Eps; Test('combinations(200, 44  )', 397106410874542174515662982291886544299251600.0);
  Eps; Test('combinations(200, 45  )', 1376635557698412871654298338611873353570738880.0);
  Eps; Test('combinations(200, 46  )', 4638663292244652067530787880105225430510098400.0);
  Eps; Test('combinations(200, 47  )', 15199024404376094008505134756089462048905428800.0);
  Eps; Test('combinations(200, 48  )', 48446890288948799652110117035035160280886054300.0);
  Eps; Test('combinations(200, 49  )', 150284231100412603002464036516843762503973066400.0);
  Eps; Test('combinations(200, 50  )', 453858377923246061067441390280868162761998660528.0);
  Eps; Test('combinations(200, 51  )', 1334877582127194297257180559649612243417643119200.0);
  Eps; Test('combinations(200, 52  )', 3824937687249075967140767372842158159023631245400.0);
  Eps; Test('combinations(200, 53  )', 10680958070054023455411954173219611462933913666400.0);
  Eps; Test('combinations(200, 54  )', 29075941412924841628621430804875608982431209425200.0);
  Eps; Test('combinations(200, 55  )', 77183408114309579595976889045669798389726483201440.0);
  Eps; Test('combinations(200, 56  )', 199849896010265875739583016278966442259113215432300.0);
  Eps; Test('combinations(200, 57  )', 504883947815408528184209725336336275180917596881600.0);
  Eps; Test('combinations(200, 58  )', 1244800078234541716040379150398208402601227868173600.0);
  Eps; Test('combinations(200, 59  )', 2995959510327202096232776938246535477447023004756800.0);
  Eps; Test('combinations(200, 60  )', 7040504849268924926147025804879358372000504061178480.0);
  Eps; Test('combinations(200, 61  )', 16158535719633598191157108404641150361968369976475200.0);
  Eps; Test('combinations(200, 62  )', 36226394597243066912432872068469675811509732689194400.0);
  Eps; Test('combinations(200, 63  )', 79353054832056241808186291197600242253783223985854400.0);
  Eps; Test('combinations(200, 64  )', 169865132999870392620648779594863018574504713844719575.0);
  Eps; Test('combinations(200, 65  )', 355410124430498052252434369613867238863579093582797880.0);
  Eps; Test('combinations(200, 66  )', 726975254516927834152706665119273897675502691419359300.0);
  Eps; Test('combinations(200, 67  )', 1453950509033855668305413330238547795351005382838718600.0);
  Eps; Test('combinations(200, 68  )', 2843756142669158880656176072378336129142407587022787850.0);
  Eps; Test('combinations(200, 69  )', 5440229142497521336907467268897686507924605818652289800.0);
  Eps; Test('combinations(200, 70  )', 10181000252388218501926831603222813321973190889192142340.0);
  Eps; Test('combinations(200, 71  )', 18641268067753076130288564907309376505021335430915190200.0);
  Eps; Test('combinations(200, 72  )', 33398938621390928066767012125595966238163225980389715775.0);
  Eps; Test('combinations(200, 73  )', 58562522514219709486933939069538132581984834595751830400.0);
  Eps; Test('combinations(200, 74  )', 100505950801431123038386625159883011323136135049465979200.0);
  Eps; Test('combinations(200, 75  )', 168849997346404286704489530268603459022868706883102845056.0);
  Eps; Test('combinations(200, 76  )', 277713811425007050500805148468097794445507741584050732000.0);
  Eps; Test('combinations(200, 77  )', 447227436580530834572725174156417227418739739693795984000.0);
  Eps; Test('combinations(200, 78  )', 705243265376990931441605082323581012468012666440216744000.0);
  Eps; Test('combinations(200, 79  )', 1089109852860669539694630633461732702798703105135524592000.0);
  Eps; Test('combinations(200, 80  )', 1647278652451762678788128833110870712983038446517480945400.0);
  Eps; Test('combinations(200, 81  )', 2440412818447055820426857530534623278493390291137008808000.0);
  Eps; Test('combinations(200, 82  )', 3541574699941459056473122513824636221228212739576878636000.0);
  Eps; Test('combinations(200, 83  )', 5035009814374604441732872971461531013312398834579176856000.0);
  Eps; Test('combinations(200, 84  )', 7013049384307484758127930210249989625685126948163853478000.0);
  Eps; Test('combinations(200, 85  )', 9570749747996096846386351816341162312699702658670670628800.0);
  Eps; Test('combinations(200, 86  )', 12798095593250594620167796033479461232098439601710780492000.0);
  Eps; Test('combinations(200, 87  )', 16769918363569744674702629285248949200680713960862402024000.0);
  Eps; Test('combinations(200, 88  )', 21534099716856603957288603514012855223601371336107402599000.0);
  Eps; Test('combinations(200, 89  )', 27099091778516175766475545995162244775768017861168866192000.0);
  Eps; Test('combinations(200, 90  )', 33422213193503283445319840060700101890113888695441601636800.0);
  Eps; Test('combinations(200, 91  )', 40400477486652320648188817655791331955082722598885452528000.0);
  Eps; Test('combinations(200, 92  )', 47865783109185901637528055700883208512000182209549068756000.0);
  Eps; Test('combinations(200, 93  )', 55586070707441692224226129201025661497806663211089241136000.0);
  Eps; Test('combinations(200, 94  )', 63273506018045330510555274728827082768779925144537753208000.0);
  Eps; Test('combinations(200, 95  )', 70599911978029526674935359171112323931480758582326335158400.0);
  Eps; Test('combinations(200, 96  )', 77218653725969794800710549093404104300057079699419429079500.0);
  Eps; Test('combinations(200, 97  )', 82791133891761429477050485625917802548514807100408460044000.0);
  Eps; Test('combinations(200, 98  )', 87015171335218645266695918566015649617316582972878279434000.0);
  Eps; Test('combinations(200, 99  )', 89651994709013149668717007007410063242083752153874590932000.0);
  Eps; Test('combinations(200, 100 )', 90548514656103281165404177077484163874504589675413336841320.0);
  Eps; Test('combinations(200, 101 )', 89651994709013149668717007007410063242083752153874590932000.0);
  Eps; Test('combinations(200, 102 )', 87015171335218645266695918566015649617316582972878279434000.0);
  Eps; Test('combinations(200, 103 )', 82791133891761429477050485625917802548514807100408460044000.0);
  Eps; Test('combinations(200, 104 )', 77218653725969794800710549093404104300057079699419429079500.0);
  Eps; Test('combinations(200, 105 )', 70599911978029526674935359171112323931480758582326335158400.0);
  Eps; Test('combinations(200, 106 )', 63273506018045330510555274728827082768779925144537753208000.0);
  Eps; Test('combinations(200, 107 )', 55586070707441692224226129201025661497806663211089241136000.0);
  Eps; Test('combinations(200, 108 )', 47865783109185901637528055700883208512000182209549068756000.0);
  Eps; Test('combinations(200, 109 )', 40400477486652320648188817655791331955082722598885452528000.0);
  Eps; Test('combinations(200, 110 )', 33422213193503283445319840060700101890113888695441601636800.0);
  Eps; Test('combinations(200, 111 )', 27099091778516175766475545995162244775768017861168866192000.0);
  Eps; Test('combinations(200, 112 )', 21534099716856603957288603514012855223601371336107402599000.0);
  Eps; Test('combinations(200, 113 )', 16769918363569744674702629285248949200680713960862402024000.0);
  Eps; Test('combinations(200, 114 )', 12798095593250594620167796033479461232098439601710780492000.0);
  Eps; Test('combinations(200, 115 )', 9570749747996096846386351816341162312699702658670670628800.0);
  Eps; Test('combinations(200, 116 )', 7013049384307484758127930210249989625685126948163853478000.0);
  Eps; Test('combinations(200, 117 )', 5035009814374604441732872971461531013312398834579176856000.0);
  Eps; Test('combinations(200, 118 )', 3541574699941459056473122513824636221228212739576878636000.0);
  Eps; Test('combinations(200, 119 )', 2440412818447055820426857530534623278493390291137008808000.0);
  Eps; Test('combinations(200, 120 )', 1647278652451762678788128833110870712983038446517480945400.0);
  Eps; Test('combinations(200, 121 )', 1089109852860669539694630633461732702798703105135524592000.0);
  Eps; Test('combinations(200, 122 )', 705243265376990931441605082323581012468012666440216744000.0);
  Eps; Test('combinations(200, 123 )', 447227436580530834572725174156417227418739739693795984000.0);
  Eps; Test('combinations(200, 124 )', 277713811425007050500805148468097794445507741584050732000.0);
  Eps; Test('combinations(200, 125 )', 168849997346404286704489530268603459022868706883102845056.0);
  Eps; Test('combinations(200, 126 )', 100505950801431123038386625159883011323136135049465979200.0);
  Eps; Test('combinations(200, 127 )', 58562522514219709486933939069538132581984834595751830400.0);
  Eps; Test('combinations(200, 128 )', 33398938621390928066767012125595966238163225980389715775.0);
  Eps; Test('combinations(200, 129 )', 18641268067753076130288564907309376505021335430915190200.0);
  Eps; Test('combinations(200, 130 )', 10181000252388218501926831603222813321973190889192142340.0);
  Eps; Test('combinations(200, 131 )', 5440229142497521336907467268897686507924605818652289800.0);
  Eps; Test('combinations(200, 132 )', 2843756142669158880656176072378336129142407587022787850.0);
  Eps; Test('combinations(200, 133 )', 1453950509033855668305413330238547795351005382838718600.0);
  Eps; Test('combinations(200, 134 )', 726975254516927834152706665119273897675502691419359300.0);
  Eps; Test('combinations(200, 135 )', 355410124430498052252434369613867238863579093582797880.0);
  Eps; Test('combinations(200, 136 )', 169865132999870392620648779594863018574504713844719575.0);
  Eps; Test('combinations(200, 137 )', 79353054832056241808186291197600242253783223985854400.0);
  Eps; Test('combinations(200, 138 )', 36226394597243066912432872068469675811509732689194400.0);
  Eps; Test('combinations(200, 139 )', 16158535719633598191157108404641150361968369976475200.0);
  Eps; Test('combinations(200, 140 )', 7040504849268924926147025804879358372000504061178480.0);
  Eps; Test('combinations(200, 141 )', 2995959510327202096232776938246535477447023004756800.0);
  Eps; Test('combinations(200, 142 )', 1244800078234541716040379150398208402601227868173600.0);
  Eps; Test('combinations(200, 143 )', 504883947815408528184209725336336275180917596881600.0);
  Eps; Test('combinations(200, 144 )', 199849896010265875739583016278966442259113215432300.0);
  Eps; Test('combinations(200, 145 )', 77183408114309579595976889045669798389726483201440.0);
  Eps; Test('combinations(200, 146 )', 29075941412924841628621430804875608982431209425200.0);
  Eps; Test('combinations(200, 147 )', 10680958070054023455411954173219611462933913666400.0);
  Eps; Test('combinations(200, 148 )', 3824937687249075967140767372842158159023631245400.0);
  Eps; Test('combinations(200, 149 )', 1334877582127194297257180559649612243417643119200.0);
  Eps; Test('combinations(200, 150 )', 453858377923246061067441390280868162761998660528.0);
  Eps; Test('combinations(200, 151 )', 150284231100412603002464036516843762503973066400.0);
  Eps; Test('combinations(200, 152 )', 48446890288948799652110117035035160280886054300.0);
  Eps; Test('combinations(200, 153 )', 15199024404376094008505134756089462048905428800.0);
  Eps; Test('combinations(200, 154 )', 4638663292244652067530787880105225430510098400.0);
  Eps; Test('combinations(200, 155 )', 1376635557698412871654298338611873353570738880.0);
  Eps; Test('combinations(200, 156 )', 397106410874542174515662982291886544299251600.0);
  Eps; Test('combinations(200, 157 )', 111290968652737934259166695674159286300427200.0);
  Eps; Test('combinations(200, 158 )', 30288048430808425146482075404992717157711200.0);
  Eps; Test('combinations(200, 159 )', 8000616566628640604731114257922604532225600.0);
  Eps; Test('combinations(200, 160 )', 2050157995198589154962348028592667411382810.0);
  Eps; Test('combinations(200, 161 )', 509356023651823392537229323873954636368400.0);
  Eps; Test('combinations(200, 162 )', 122622746434698224129332985377063153199800.0);
  Eps; Test('combinations(200, 163 )', 28586897941831487833832229719806133874800.0);
  Eps; Test('combinations(200, 164 )', 6449483072242469816169466461175774105900.0);
  Eps; Test('combinations(200, 165 )', 1407159943034720687164247227892896168560.0);
  Eps; Test('combinations(200, 166 )', 296690349435031470185232849254526300600.0);
  Eps; Test('combinations(200, 167 )', 60404023238269880157472556135652061200.0);
  Eps; Test('combinations(200, 168 )', 11865075993231583602360680669503083450.0);
  Eps; Test('combinations(200, 169 )', 2246641608185861983878945452213601600.0);
  Eps; Test('combinations(200, 170 )', 409681705022127773530866523638950880.0);
  Eps; Test('combinations(200, 171 )', 71873983337215398865064302392798400.0);
  Eps; Test('combinations(200, 172 )', 12118287888251433529574795170878800.0);
  Eps; Test('combinations(200, 173 )', 1961341392318151091491874362916800.0);
  Eps; Test('combinations(200, 174 )', 304346078118333790059083952866400.0);
  Eps; Test('combinations(200, 175 )', 45217131606152448808778187283008.0);
  Eps; Test('combinations(200, 176 )', 6422888012237563751246901602700.0);
  Eps; Test('combinations(200, 177 )', 870900069455940847626698522400.0);
  Eps; Test('combinations(200, 178 )', 112532031446554154468618348400.0);
  Eps; Test('combinations(200, 179 )', 13830752468291572057595551200.0);
  Eps; Test('combinations(200, 180 )', 1613587787967350073386147640.0);
  Eps; Test('combinations(200, 181 )', 178296993145563544020568800.0);
  Eps; Test('combinations(200, 182 )', 18613422361350040309839600.0);
  Eps; Test('combinations(200, 183 )', 1830828428985249866541600.0);
  Eps; Test('combinations(200, 184 )', 169152626591028520278300.0);
  Eps; Test('combinations(200, 185 )', 14629416353818682834880.0);
  Eps; Test('combinations(200, 186 )', 1179791641436990551200.0);
  Eps; Test('combinations(200, 187 )', 88326646952501966400.0);
  Test('combinations(200, 188 )', 6107693672247476400);
  Test('combinations(200, 189 )', 387790074428411200);
  Test('combinations(200, 190 )', 22451004309013280);
  Test('combinations(200, 191 )', 1175445251780800);
  Test('combinations(200, 192 )', 55098996177225);
  Test('combinations(200, 193 )', 2283896214600);
  Test('combinations(200, 194 )', 82408626300);
  Test('combinations(200, 195 )', 2535650040);
  Test('combinations(200, 196 )', 64684950);
  Test('combinations(200, 197 )', 1313400);
  Test('combinations(200, 198 )', 19900);
  Test('combinations(200, 199 )', 200);
  Test('combinations(200, 200 )', 1);

  Test('combinations(400, 0   )', 1);
  Test('combinations(400, 1   )', 400);
  Test('combinations(400, 2   )', 79800);
  Test('combinations(400, 3   )', 10586800);
  Test('combinations(400, 4   )', 1050739900);
  Test('combinations(400, 5   )', 83218600080);
  Test('combinations(400, 6   )', 5478557838600);
  Test('combinations(400, 7   )', 308364541201200);
  Test('combinations(400, 8   )', 15148408086508950);
  Test('combinations(400, 9   )', 659797329990167600);
  Eps; Test('combinations(400, 10  )', 25798075602615553160.0);
  Eps; Test('combinations(400, 11  )', 914659044092733248400.0);
  Eps; Test('combinations(400, 12  )', 29650197346006102802300.0);
  Eps; Test('combinations(400, 13  )', 884944351557720606714800.0);
  Eps; Test('combinations(400, 14  )', 24462390289488419628473400.0);
  Eps; Test('combinations(400, 15  )', 629498843449501998439382160.0);
  Eps; Test('combinations(400, 16  )', 15147315920503641837447633225.0);
  Eps; Test('combinations(400, 17  )', 342151136086670497975287715200.0);
  Eps; Test('combinations(400, 18  )', 7280215840066377818029733051200.0);
  Eps; Test('combinations(400, 19  )', 146370655310808227709860948713600.0);
  Eps; Test('combinations(400, 20  )', 2788360983670896737872851072994080.0);
  Eps; Test('combinations(400, 21  )', 50456055894997179066270638463702400.0);
  Eps; Test('combinations(400, 22  )', 869220235645633221187116908079236800.0);
  Eps; Test('combinations(400, 23  )', 14285445611915189461249138750171804800.0);
  Eps; Test('combinations(400, 24  )', 224400541487167767787121887867282100400.0);
  Eps; Test('combinations(400, 25  )', 3374984143967003227518313193523922790016.0);
  Eps; Test('combinations(400, 26  )', 48677655922601008089206440291210424856000.0);
  Eps; Test('combinations(400, 27  )', 674275678335288037976415135885655514672000.0);
  Eps; Test('combinations(400, 28  )', 8982315286395087077328673060191053820452000.0);
  Eps; Test('combinations(400, 29  )', 115221423673757668716078150979002483489936000.0);
  Eps; Test('combinations(400, 30  )', 1424904939432136503122166467106997379158875200.0);
  Eps; Test('combinations(400, 31  )', 17006929922254532456619406220309323557702704000.0);
  Eps; Test('combinations(400, 32  )', 196111160665997577390392527977941887274759305500.0);
  Eps; Test('combinations(400, 33  )', 2186936579548094196353468190784321652033679528000.0);
  Eps; Test('combinations(400, 34  )', 23606050726298546178285965471113119008716481964000.0);
  Eps; Test('combinations(400, 35  )', 246851844737864797178647524640782901634006639966400.0);
  Eps; Test('combinations(400, 36  )', 2502803425814462526950176291496826641567011766326000.0);
  Eps; Test('combinations(400, 37  )', 24622174243147685400266599192022835068388980620072000.0);
  Eps; Test('combinations(400, 38  )', 235206559217437100007809881755376029732242104344372000.0);
  Eps; Test('combinations(400, 39  )', 2183199344531082825713517363985798019565939532632376000.0);
  Eps; Test('combinations(400, 40  )', 19703374084393022502064494209971827126582604282007193400.0);
  Eps; Test('combinations(400, 41  )', 173005235862963124408371168672923360135847257110307064000.0);
  Eps; Test('combinations(400, 42  )', 1478782849400089563395363084609035387827837269109529428000.0);
  Eps; Test('combinations(400, 43  )', 12311726978726327062686976378838015554473621914911896168000.0);
  Eps; Test('combinations(400, 44  )', 99892875713756790031346603801026626203342795991444248454000.0);
  Eps; Test('combinations(400, 45  )', 790263638979942605581319798959232865075334119398981165547200.0);
  Eps; Test('combinations(400, 46  )', 6098773735606078803942794100663644936994426356231267690636000.0);
  Eps; Test('combinations(400, 47  )', 45935444732011742480760619396487878887149509151188697074152000.0);
  Eps; Test('combinations(400, 48  )', 337816916466669689493927055145004609315912015216033543066159500.0);
  Eps; Test('combinations(400, 49  )', 2426766420331994504119639253286563724065327129715179737944656000.0);
  Eps; Test('combinations(400, 50  )', 17035900270730601418919867558071677342938596450600561760371485120.0);
  Eps; Test('combinations(400, 51  )', 116913041073641382286704973437746805294676642308043070904510192000.0);
  Eps; Test('combinations(400, 52  )', 784666371821170046501154533264877597073887464721289072032193404000.0);
  Eps; Test('combinations(400, 53  )', 5152149007429569361932109010871271769466279957037898057871760464000.0);
  Eps; Test('combinations(400, 54  )', 33107327881075195714637811606895024148237021205410196779287053352000.0);
  Eps; Test('combinations(400, 55  )', 208275189942763958495721505745194151914363806128580510647878553814400.0);
  Eps; Test('combinations(400, 56  )', 1283123938040242244303998562180213971615277019899290645955680376178000.0);
  Eps; Test('combinations(400, 57  )', 7743765520804268983167991322631466776064127979743087407171123673776000.0);
  Eps; Test('combinations(400, 58  )', 45795027131652832090114155580389536279137860293997913459649921036296000.0);
  Eps; Test('combinations(400, 59  )', 265455919983479128386763410313444430635002512212665871240682593125648000.0);
  Eps; Test('combinations(400, 60  )', 1508674478572773046331438715281409180775597611075317701551212737597432800.0);
  Eps; Test('combinations(400, 61  )', 8409005290405620258240805954027526581372183405993574074219874275133232000.0);
  Eps; Test('combinations(400, 62  )', 45978270862056536573284406748634379211051131848900348567105441601131704000.0);
  Eps; Test('combinations(400, 63  )', 246677072244049354948732213984736828148179088332195520883835543828293904000.0);
  Eps; Test('combinations(400, 64  )', 1298908958535072384651918064263379860717755511999217039653946535470860088250.0);
  Eps; Test('combinations(400, 65  )', 6714360154888989557585299532192240510787166954334414235749631321818599840800.0);
  Eps; Test('combinations(400, 66  )', 34080464422542598511985990049763645016874256510636799529941310497109559798000.0);
  Eps; Test('combinations(400, 67  )', 169893658464615341835870457860015782621432860814219269298513398597531238396000.0);
  Eps; Test('combinations(400, 68  )', 831979239245836894578600918638018464896134450751985539358896496073204446851000.0);
  Eps; Test('combinations(400, 69  )', 4003146484487215202899934854895972903558212139850133319813820821685563425428000.0);
  Eps; Test('combinations(400, 70  )', 18929164090932403316569691956722386158253831689862773269405352742541735625952400.0);
  Eps; Test('combinations(400, 71  )', 87980621831094268936168990784766020172165696586686129280334738099137644458652000.0);
  Eps; Test('combinations(400, 72  )', 402022563644861312222216638447055842175590474680829674072640678258559514262451500.0);
  Eps; Test('combinations(400, 73  )', 1806347957198828909710781608364853647035529804045371686244193732449418091480604000.0);
  Eps; Test('combinations(400, 74  )', 7982105162216446668586832242369015440278624944903196505430423655553509674515642000.0);
  Eps; Test('combinations(400, 75  )', 34695550438434154852790764146830653780411089760512560810270908156139255385227990560.0);
  Eps; Test('combinations(400, 76  )', 148369130164356583252065767733157401034652686475876082412342699351911289476303907000.0);
  Eps; Test('combinations(400, 77  )', 624306469782487441216484529162896077080876239197192866254532916753496854419772284000.0);
  Eps; Test('combinations(400, 78  )', 2585269099227480045037493627174556832014397759752478151284796565530506204840851894000.0);
  Eps; Test('combinations(400, 79  )', 10537425948749981955722442379116548100109317451143012211565879672162316429857649492000.0);
  Eps; Test('combinations(400, 80  )', 42281421619359302597336300046205149251688636272711336498908092184551294674803818586650.0);
  Eps; Test('combinations(400, 81  )', 167037715039444158409229827343032688401732884040341082464822092580943386369595332688000.0);
  Eps; Test('combinations(400, 82  )', 649817452409544957713955060029602775611619390352058601296076189430743173803669647896000.0);
  Eps; Test('combinations(400, 83  )', 2489662046581148151241418181800164851138493567854875123037978653481642521320083711216000.0);
  Eps; Test('combinations(400, 84  )', 9395510342455047189803923376555384021558362631071373976226657537543817610219839719708000.0);
  Eps; Test('combinations(400, 85  )', 34929191390774057787976938670488251186028736369630049135148515080751133703876110016796800.0);
  Eps; Test('combinations(400, 86  )', 127938317303416607014101577688416268879058743679458900901997468028332640892104356456872000.0);
  Eps; Test('combinations(400, 87  )', 461754386589342696579630981542100096873844201325863159577324194952832749886445608361584000.0);
  Eps; Test('combinations(400, 88  )', 1642376397755275727607096559348606026380832216079490556223891738866325576300653129740634000.0);
  Eps; Test('combinations(400, 89  )', 5757544225838719404645102545132191912705838779964056781369148567711163818042739061562672000.0);
  Eps; Test('combinations(400, 90  )', 19895513935953797053829187683734574276127954006320240655620057828424132749014353868288788800.0);
  Eps; Test('combinations(400, 91  )', 67775926595007440513044485516018879402194129032519501134529867327598693980158787902961808000.0);
  Eps; Test('combinations(400, 92  )', 227638709976709773027508108961411236253021585554875280984453576132913004781185494152339116000.0);
  Eps; Test('combinations(400, 93  )', 753900243793834517123360188818437212536888691945178349927007542461690381425861636547531696000.0);
  Eps; Test('combinations(400, 94  )', 2462206115369225497413527425183619406902391791778401632208418250380201564869569387447789688000.0);
  Eps; Test('combinations(400, 95  )', 7930895487399821075879362022170395142232967245096746310060799838066754514211455079568669942400.0);
  Eps; Test('combinations(400, 96  )', 25197115871426514876491723091270526233135989684942787755922332818857917987859310409046295129500.0);
  Eps; Test('combinations(400, 97  )', 78968280669212995076840039378827216235807637775490798740210197700338217199064230560310038344000.0);
  Eps; Test('combinations(400, 98  )', 244157031048689158247780938079435168565813410673201143043711121461249793993025121018101445084000.0);
  Eps; Test('combinations(400, 99  )', 744802256330344704957877205050398190978541919427340860597987461427246846322157439873400367832000.0);
  Eps; Test('combinations(400, 100 )', 2241854791554337561923210387201698554845411177476295990399942258896013007429693894018935107174320.0);
  Eps; Test('combinations(400, 101 )', 6658974628379220480960030853074352143105181715276126704158244333354494081474338299066143882696000.0);
  Eps; Test('combinations(400, 102 )', 19519935430248891409873031618325796968514209145760410632777598585029350297655168151184088440452000.0);
  Eps; Test('combinations(400, 103 )', 56475152992370579030506440992826092200167323547928178335608974546978120278652816592746197623832000.0);
  Eps; Test('combinations(400, 104 )', 161280004218596749731350124758359128687016298978217970823806398465889439641921985846592506675751000.0);
  Eps; Test('combinations(400, 105 )', 454656011892425123052187018366421924679588804738595422512825656627650229847703883910394114057355200.0);
  Eps; Test('combinations(400, 106 )', 1265316259511937842456558211491457243212063182998921223030977063256196394387477790127983619310564000.0);
  Eps; Test('combinations(400, 107 )', 3476663367257100240020823496995218967330341829922269528701936977545062990186153928015207327825288000.0);
  Eps; Test('combinations(400, 108 )', 9432058950058614540056493376107399605812871816363194184348773466858365334486510193596812472711198000.0);
  Eps; Test('combinations(400, 109 )', 25267534068046930694463266658929914540342739177780300016787539929565529152936339234222653596620824000.0);
  Eps; Test('combinations(400, 110 )', 66844112852742334837171005434078228465815791824855157317137946540941536213677042883261747241969634400.0);
  Eps; Test('combinations(400, 111 )', 174637772317975469394410734917862038334113329992864825423153193665522932450147229154467627929470216000.0);
  Eps; Test('combinations(400, 112 )', 450627823213347416562363414207697581058560288999445844172243508654786852482969189514653075639436539500.0);
  Eps; Test('combinations(400, 113 )', 1148502770667646513008501445060326578273144807361419496651381685775031978009691385665664475965997552000.0);
  Eps; Test('combinations(400, 114 )', 2891406098084338151170525567827313403196425962392345574903039858047668225340187962158295654405625416000.0);
  Eps; Test('combinations(400, 115 )', 7190801252627136619432785325205318550558068045601833342802342603492461847367771801541500497043555382400.0);
  Eps; Test('combinations(400, 116 )', 17667054801713223590847791531754446438871115456866573299126445189615100228446680719304548634977700724000.0);
  Eps; Test('combinations(400, 117 )', 42884133022962012818810023889044981099482023844018006982494960973082807392126985677628135148151000048000.0);
  Eps; Test('combinations(400, 118 )', 102849234283883471421383362377963810602995023286924542169881982672732495694677431752277646160396042488000.0);
  Eps; Test('combinations(400, 119 )', 243726756874412932275883262105762979748273920730358999091653101795887090637807023144052909388501546064000.0);
  Eps; Test('combinations(400, 120 )', 570726822347583616412693305430994977577208097710257322872954346705368937243531445862323896151407787033200.0);
  Eps; Test('combinations(400, 121 )', 1320690167415895971864083681989079286955522870734479755408489397334738036596601692904551164647885788176000.0);
  Eps; Test('combinations(400, 122 )', 3020266858270778493033437272745517385742548204384588948844004441445835345987310428855489958498033892632000.0);
  Eps; Test('combinations(400, 123 )', 6826294199994117244417037087993933603548198380641591282753115729446684765727417066844115515954905871152000.0);
  Eps; Test('combinations(400, 124 )', 15249060430632019973415478011083222646635894769659038591311395621425255484729794576740483854189588115396000.0);
  Eps; Test('combinations(400, 125 )', 33669925430835500101301375448471755603772055651407157209615561532106964110283386425442988350050610558794368.0);
  Eps; Test('combinations(400, 126 )', 73485948360950496252840303558172482468550121461404509782891106518487421669269295769816046002094586537051200.0);
  Eps; Test('combinations(400, 127 )', 158544487014964062781718450196372127530572702995471147090646954220988610530549504259288162240739501662614400.0);
  Eps; Test('combinations(400, 128 )', 338145663711603040151633882059449928248799593107528305904270457049452270897187614553013033529077218389794775.0);
  Eps; Test('combinations(400, 129 )', 712989306430666875358483844342406050261034800970912396945438483081015640961511869445112752867511654279257200.0);
  Eps; Test('combinations(400, 130 )', 1486308477251620947862685552436861843236464854331671227478567914730424913081305512458658123285351217766759240.0);
  Eps; Test('combinations(400, 131 )', 3063383884411737831472710680595058760869049699767566652055063641047440660549255636365173231198815487000190800.0);
  Eps; Test('combinations(400, 132 )', 6242805037172405126258781614242960656619502797253601737900091813952738921876892168047209084791525500023116100.0);
  Eps; Test('combinations(400, 133 )', 12579486841821087021333484756519650044917494358375678689903944407062661887691782714561293494166382210572895600.0);
  Eps; Test('combinations(400, 134 )', 25065096916165897273851047985005571358156499952882882165704128034968139731445567050655711663749433210619127800.0);
  Eps; Test('combinations(400, 135 )', 49387524294075027221069472326010977639034288796050715970942948572603890137514969151662365204128512844627318480.0);
  Eps; Test('combinations(400, 136 )', 96233043661249133923407427694065507899588871551128233325734421851029638870893138420518579258044528704604701450.0);
  Eps; Test('combinations(400, 137 )', 185441777566202710626128181833819664857601913062028128452510126778626457386246631700853320614042011518362344400.0);
  Eps; Test('combinations(400, 138 )', 353414402173270383294722549436917187373545674893575346253696835817237378931759885053075531315167021951661569400.0);
  Eps; Test('combinations(400, 139 )', 666148009851775830382858330593325921524237171382134825312723532259828728633964675423782656148012660081549145200.0);
  Eps; Test('combinations(400, 140 )', 1241890218366524940928043030606129039413042155219551352904434585141537844096177002040051951818795030580602334980.0);
  Eps; Test('combinations(400, 141 )', 2290010331739691380434689276294989718066602555723995402518815547069502407553234188158961045906997928730188702800.0);
  Eps; Test('combinations(400, 142 )', 4176849830426620193891440299721143218163732830510667670791360751345078334903434188261766978097975095359992070600.0);
  Eps; Test('combinations(400, 143 )', 7535854938811664405762179002294090561442259232669596217231965551377833639196405738262488673771171850369775903600.0);
  Eps; Test('combinations(400, 144 )', 13449407772740262168617222247149869960351809880528376582143160741000716981065807463426802702494383094062725050175.0);
  Eps; Test('combinations(400, 145 )', 23745161309113842173558682036347356619655609168381133827783787239284024463123080763015596495438359117793500778240.0);
  Eps; Test('combinations(400, 146 )', 41472713245370066809982629584031342041179317383131432370444285931626207110249216401157377440662887500255771907200.0);
  Eps; Test('combinations(400, 147 )', 71660334451183652855344135471727625023534330716431182463216657324034398680294564393836556938288254592278680710400.0);
  Eps; Test('combinations(400, 148 )', 122500436595604487651365312664507358992933686967953305156714961506626370716990032375950330441803570350314231214400.0);
  Eps; Test('combinations(400, 149 )', 207181946456995509316403079137287613867243551113585455701289733554831177320009987642546867592848991464960981651200.0);
  Eps; Test('combinations(400, 150 )', 346684457071372485589447819089727940537854208863399662540158154148417503382150045988528425105367312384701375963008.0);
  Eps; Test('combinations(400, 151 )', 573980889191014048989152018360476722744791736528807388311520122762280634738659016537298717061866411232949297952000.0);
  Eps; Test('combinations(400, 152 )', 940271325056332224988808240603675683970086463129427892694529674788209724012671678406495924660557476296081415724000.0);
  Eps; Test('combinations(400, 153 )', 1524099925581505828740029043592886075977656489255543250903551368284156938268905727090267904024955909290380333984000.0);
  Eps; Test('combinations(400, 154 )', 2444497932588519088953163466022356238743384109390384305020631090689524439950777367475949170741325386978726899312000.0);
  Eps; Test('combinations(400, 155 )', 3879654783334036747628891694461287966005628973613126058290808053610471046631556338058603200015264807721076240198400.0);
  Eps; Test('combinations(400, 156 )', 6093047576389993610058195289378304818406276272661640283854153673939521836055969889899729384639358191613228710568000.0);
  Eps; Test('combinations(400, 157 )', 9469449736555149304803819430626155259179180958786243498473971314912377885335392695130789616891741393335208951456000.0);
  Eps; Test('combinations(400, 158 )', 14563773961917096715616000769887061569497094765728209937526424237491821684408230537447986562687931383420606172176000.0);
  Eps; Test('combinations(400, 159 )', 22166247162163128334459573498821816979989288888718407577870406701088181431615042704795048730631945879168469771488000.0);
  Eps; Test('combinations(400, 160 )', 33387909788008212053779732582600361826108866388632101414167300093514073281370158074097542150514368480497507593303800.0);
  Eps; Test('combinations(400, 161 )', 49770797199515347160913887079652713281156074119700026952795975294679363897694645576294472770953095871549079642192000.0);
  Eps; Test('combinations(400, 162 )', 73427287226445481305298882790351842433310504411162385442705173428570172663882841313175178964554258724075494040024000.0);
  Eps; Test('combinations(400, 163 )', 107212848833705672089945608000636432509987116870286182425544977153372399349718504494084003641496402308772807248624000.0);
  Eps; Test('combinations(400, 164 )', 154935641302367343203153104244822161615042357916206251432037558447251577109044424177426273555089313092555825109292000.0);
  Eps; Test('combinations(400, 165 )', 221604917256719351490570500616836546309999978595301062654308265415462861804451418823470306418188350847534392277532800.0);
  Eps; Test('combinations(400, 166 )', 313717804550174985543879925572027640860542138372865962191340014292974533277386044719973024146230496681750495091688000.0);
  Eps; Test('combinations(400, 167 )', 439580636315814051600406602298529748271657846582339132651338702662012220280888230326189746408490636069039615876976000.0);
  Eps; Test('combinations(400, 168 )', 609656477747527821564849632949746615162477846748125106593820938811005043603850938488108398292728084548132324400806000.0);
  Eps; Test('combinations(400, 169 )', 836924868860511565698491803812669909572158937547722039821103300616290947432505430350539339668123761036489344739568000.0);
  Eps; Test('combinations(400, 170 )', 1137233204157518656919715333416039700653933615138375242345146249660960052099463261241026985078450522349582580204942400.0);
  Eps; Test('combinations(400, 171 )', 1529611912024732696441722378278883807897103692876177226546103142818835157794599708101966120280956842926339142965712000.0);
  Eps; Test('combinations(400, 172 )', 2036518185195719694681130375731769720979283405050259214413125695962286343807926355554361869443832075756579440343884000.0);
  Eps; Test('combinations(400, 173 )', 2683966163148116129406345235068459516666338822840804051365275483695961192995417393447367088053142851286127817331824000.0);
  Eps; Test('combinations(400, 174 )', 3501496086405875640087588323911151208524476510257830572758146751718294200057239932830760511425651880700867899622552000.0);
  Eps; Test('combinations(400, 175 )', 4521932088729873683770256921165258132151609664675826911104806662219054224073921284684296431898270428790835116083981440.0);
  Eps; Test('combinations(400, 176 )', 5780879090705804425274476177626040225762000991773074176128303971586859093276319824170265324869947991351919892720999000.0);
  Eps; Test('combinations(400, 177 )', 7315914781458193170968828608973067856331571876594172968659548529013878174541783280305872501530329661371921220166688000.0);
  Eps; Test('combinations(400, 178 )', 9165443799242567849022745953938169280685059148766857146129659112191544005184368941057357122703727609471564225265008000.0);
  Eps; Test('combinations(400, 179 )', 11367198454926536661916478222202645700067503525286269756652426384952641168440949189467783694079483403925627139714144000.0);
  Eps; Test('combinations(400, 180 )', 13956393658548692234908564928371026109527323772712586756778812394858520545696943171513223313286476845930908877093476800.0);
  Eps; Test('combinations(400, 181 )', 16963572402655869014805990520671965436994537182302591638073694623584942099742140871452536623884115503341436204202016000.0);
  Eps; Test('combinations(400, 182 )', 20412210748250743484848966615533848520339580455627843784275489684423639120019389290374206157311106017757002905056272000.0);
  Eps; Test('combinations(400, 183 )', 24316185481522743604902047662220650149912724258616775655585009569422695782318179591811895859529077114049325865039712000.0);
  Eps; Test('combinations(400, 184 )', 28677240486361061751433393166858049361581854152825219115554060198721331438929592235995551095205487683416868003878356000.0);
  Eps; Test('combinations(400, 185 )', 33482615919210753180051961751574803578928002686541877453836091907696257247615091475540751548996677511448883723447161600.0);
  Eps; Test('combinations(400, 186 )', 38703023777582322224253611702089154674567314933368299207391181506208039291598089608823987005560675618072634411511504000.0);
  Eps; Test('combinations(400, 187 )', 44291160900548753775349052963888123531323023506635379841613437659510804323005300407959001172138954985387934567184288000.0);
  Eps; Test('combinations(400, 188 )', 50180942935196194436964618517596650596658531951666680352466288412105326174468771206889719413114879850466117355373688000.0);
  Eps; Test('combinations(400, 189 )', 56287618530484620215007931882171904372971474993403895421814037795589043116335341247939791087726743535972576081159904000.0);
  Eps; Test('combinations(400, 190 )', 62508881631222394028245650669148799066826216966359062810540852499312042618667142122712083786896541505737966069077577600.0);
  Eps; Test('combinations(400, 191 )', 68727042631186925371369563562938470178185892999661796807400937302908528533613088197746270132189914744528653793226656000.0);
  Eps; Test('combinations(400, 192 )', 74812249530823267721959577003406980558546102275673518399722895293270221164193413715255054466810896779200461681168599500.0);
  Eps; Test('combinations(400, 193 )', 80626673069488288529365761744604414280712897789326900658768716170985523327213627216440680461640759223179772174523672000.0);
  Eps; Test('combinations(400, 194 )', 86029491367959153224632539593469658536636957950467363074047032203061872828521756875274334307008438964939241443950516000.0);
  Eps; Test('combinations(400, 195 )', 90882437034869669560381041826947434146395965834852701503865069917080747706028112391315450601249940650140942243352852800.0);
  Eps; Test('combinations(400, 196 )', 95055610164021848264684252931246040816383535694616345960675200678579353468039607348059527414572642006524965101465994000.0);
  Eps; Test('combinations(400, 197 )', 98433220677464248964444607096315697089046910059399667898364167200153239124264364969564180672958471925538542541619608000.0);
  Eps; Test('combinations(400, 198 )', 100918908068309305756476036568444881358972337081101679713979423947631856273867000448593579174800857580223859272468588000.0);
  Eps; Test('combinations(400, 199 )', 102440298642203415893508338627265658464886492916495172372984138881515753604628814525708055242762679553795073231350024000.0);
  Eps; Test('combinations(400, 200 )', 102952500135414432972975880320401986757210925381077648234849059575923332372651958598336595518976492951564048597506774120.0);
  Eps; Test('combinations(400, 201 )', 102440298642203415893508338627265658464886492916495172372984138881515753604628814525708055242762679553795073231350024000.0);
  Eps; Test('combinations(400, 202 )', 100918908068309305756476036568444881358972337081101679713979423947631856273867000448593579174800857580223859272468588000.0);
  Eps; Test('combinations(400, 203 )', 98433220677464248964444607096315697089046910059399667898364167200153239124264364969564180672958471925538542541619608000.0);
  Eps; Test('combinations(400, 204 )', 95055610164021848264684252931246040816383535694616345960675200678579353468039607348059527414572642006524965101465994000.0);
  Eps; Test('combinations(400, 205 )', 90882437034869669560381041826947434146395965834852701503865069917080747706028112391315450601249940650140942243352852800.0);
  Eps; Test('combinations(400, 206 )', 86029491367959153224632539593469658536636957950467363074047032203061872828521756875274334307008438964939241443950516000.0);
  Eps; Test('combinations(400, 207 )', 80626673069488288529365761744604414280712897789326900658768716170985523327213627216440680461640759223179772174523672000.0);
  Eps; Test('combinations(400, 208 )', 74812249530823267721959577003406980558546102275673518399722895293270221164193413715255054466810896779200461681168599500.0);
  Eps; Test('combinations(400, 209 )', 68727042631186925371369563562938470178185892999661796807400937302908528533613088197746270132189914744528653793226656000.0);
  Eps; Test('combinations(400, 210 )', 62508881631222394028245650669148799066826216966359062810540852499312042618667142122712083786896541505737966069077577600.0);
  Eps; Test('combinations(400, 211 )', 56287618530484620215007931882171904372971474993403895421814037795589043116335341247939791087726743535972576081159904000.0);
  Eps; Test('combinations(400, 212 )', 50180942935196194436964618517596650596658531951666680352466288412105326174468771206889719413114879850466117355373688000.0);
  Eps; Test('combinations(400, 213 )', 44291160900548753775349052963888123531323023506635379841613437659510804323005300407959001172138954985387934567184288000.0);
  Eps; Test('combinations(400, 214 )', 38703023777582322224253611702089154674567314933368299207391181506208039291598089608823987005560675618072634411511504000.0);
  Eps; Test('combinations(400, 215 )', 33482615919210753180051961751574803578928002686541877453836091907696257247615091475540751548996677511448883723447161600.0);
  Eps; Test('combinations(400, 216 )', 28677240486361061751433393166858049361581854152825219115554060198721331438929592235995551095205487683416868003878356000.0);
  Eps; Test('combinations(400, 217 )', 24316185481522743604902047662220650149912724258616775655585009569422695782318179591811895859529077114049325865039712000.0);
  Eps; Test('combinations(400, 218 )', 20412210748250743484848966615533848520339580455627843784275489684423639120019389290374206157311106017757002905056272000.0);
  Eps; Test('combinations(400, 219 )', 16963572402655869014805990520671965436994537182302591638073694623584942099742140871452536623884115503341436204202016000.0);
  Eps; Test('combinations(400, 220 )', 13956393658548692234908564928371026109527323772712586756778812394858520545696943171513223313286476845930908877093476800.0);
  Eps; Test('combinations(400, 221 )', 11367198454926536661916478222202645700067503525286269756652426384952641168440949189467783694079483403925627139714144000.0);
  Eps; Test('combinations(400, 222 )', 9165443799242567849022745953938169280685059148766857146129659112191544005184368941057357122703727609471564225265008000.0);
  Eps; Test('combinations(400, 223 )', 7315914781458193170968828608973067856331571876594172968659548529013878174541783280305872501530329661371921220166688000.0);
  Eps; Test('combinations(400, 224 )', 5780879090705804425274476177626040225762000991773074176128303971586859093276319824170265324869947991351919892720999000.0);
  Eps; Test('combinations(400, 225 )', 4521932088729873683770256921165258132151609664675826911104806662219054224073921284684296431898270428790835116083981440.0);
  Eps; Test('combinations(400, 226 )', 3501496086405875640087588323911151208524476510257830572758146751718294200057239932830760511425651880700867899622552000.0);
  Eps; Test('combinations(400, 227 )', 2683966163148116129406345235068459516666338822840804051365275483695961192995417393447367088053142851286127817331824000.0);
  Eps; Test('combinations(400, 228 )', 2036518185195719694681130375731769720979283405050259214413125695962286343807926355554361869443832075756579440343884000.0);
  Eps; Test('combinations(400, 229 )', 1529611912024732696441722378278883807897103692876177226546103142818835157794599708101966120280956842926339142965712000.0);
  Eps; Test('combinations(400, 230 )', 1137233204157518656919715333416039700653933615138375242345146249660960052099463261241026985078450522349582580204942400.0);
  Eps; Test('combinations(400, 231 )', 836924868860511565698491803812669909572158937547722039821103300616290947432505430350539339668123761036489344739568000.0);
  Eps; Test('combinations(400, 232 )', 609656477747527821564849632949746615162477846748125106593820938811005043603850938488108398292728084548132324400806000.0);
  Eps; Test('combinations(400, 233 )', 439580636315814051600406602298529748271657846582339132651338702662012220280888230326189746408490636069039615876976000.0);
  Eps; Test('combinations(400, 234 )', 313717804550174985543879925572027640860542138372865962191340014292974533277386044719973024146230496681750495091688000.0);
  Eps; Test('combinations(400, 235 )', 221604917256719351490570500616836546309999978595301062654308265415462861804451418823470306418188350847534392277532800.0);
  Eps; Test('combinations(400, 236 )', 154935641302367343203153104244822161615042357916206251432037558447251577109044424177426273555089313092555825109292000.0);
  Eps; Test('combinations(400, 237 )', 107212848833705672089945608000636432509987116870286182425544977153372399349718504494084003641496402308772807248624000.0);
  Eps; Test('combinations(400, 238 )', 73427287226445481305298882790351842433310504411162385442705173428570172663882841313175178964554258724075494040024000.0);
  Eps; Test('combinations(400, 239 )', 49770797199515347160913887079652713281156074119700026952795975294679363897694645576294472770953095871549079642192000.0);
  Eps; Test('combinations(400, 240 )', 33387909788008212053779732582600361826108866388632101414167300093514073281370158074097542150514368480497507593303800.0);
  Eps; Test('combinations(400, 241 )', 22166247162163128334459573498821816979989288888718407577870406701088181431615042704795048730631945879168469771488000.0);
  Eps; Test('combinations(400, 242 )', 14563773961917096715616000769887061569497094765728209937526424237491821684408230537447986562687931383420606172176000.0);
  Eps; Test('combinations(400, 243 )', 9469449736555149304803819430626155259179180958786243498473971314912377885335392695130789616891741393335208951456000.0);
  Eps; Test('combinations(400, 244 )', 6093047576389993610058195289378304818406276272661640283854153673939521836055969889899729384639358191613228710568000.0);
  Eps; Test('combinations(400, 245 )', 3879654783334036747628891694461287966005628973613126058290808053610471046631556338058603200015264807721076240198400.0);
  Eps; Test('combinations(400, 246 )', 2444497932588519088953163466022356238743384109390384305020631090689524439950777367475949170741325386978726899312000.0);
  Eps; Test('combinations(400, 247 )', 1524099925581505828740029043592886075977656489255543250903551368284156938268905727090267904024955909290380333984000.0);
  Eps; Test('combinations(400, 248 )', 940271325056332224988808240603675683970086463129427892694529674788209724012671678406495924660557476296081415724000.0);
  Eps; Test('combinations(400, 249 )', 573980889191014048989152018360476722744791736528807388311520122762280634738659016537298717061866411232949297952000.0);
  Eps; Test('combinations(400, 250 )', 346684457071372485589447819089727940537854208863399662540158154148417503382150045988528425105367312384701375963008.0);
  Eps; Test('combinations(400, 251 )', 207181946456995509316403079137287613867243551113585455701289733554831177320009987642546867592848991464960981651200.0);
  Eps; Test('combinations(400, 252 )', 122500436595604487651365312664507358992933686967953305156714961506626370716990032375950330441803570350314231214400.0);
  Eps; Test('combinations(400, 253 )', 71660334451183652855344135471727625023534330716431182463216657324034398680294564393836556938288254592278680710400.0);
  Eps; Test('combinations(400, 254 )', 41472713245370066809982629584031342041179317383131432370444285931626207110249216401157377440662887500255771907200.0);
  Eps; Test('combinations(400, 255 )', 23745161309113842173558682036347356619655609168381133827783787239284024463123080763015596495438359117793500778240.0);
  Eps; Test('combinations(400, 256 )', 13449407772740262168617222247149869960351809880528376582143160741000716981065807463426802702494383094062725050175.0);
  Eps; Test('combinations(400, 257 )', 7535854938811664405762179002294090561442259232669596217231965551377833639196405738262488673771171850369775903600.0);
  Eps; Test('combinations(400, 258 )', 4176849830426620193891440299721143218163732830510667670791360751345078334903434188261766978097975095359992070600.0);
  Eps; Test('combinations(400, 259 )', 2290010331739691380434689276294989718066602555723995402518815547069502407553234188158961045906997928730188702800.0);
  Eps; Test('combinations(400, 260 )', 1241890218366524940928043030606129039413042155219551352904434585141537844096177002040051951818795030580602334980.0);
  Eps; Test('combinations(400, 261 )', 666148009851775830382858330593325921524237171382134825312723532259828728633964675423782656148012660081549145200.0);
  Eps; Test('combinations(400, 262 )', 353414402173270383294722549436917187373545674893575346253696835817237378931759885053075531315167021951661569400.0);
  Eps; Test('combinations(400, 263 )', 185441777566202710626128181833819664857601913062028128452510126778626457386246631700853320614042011518362344400.0);
  Eps; Test('combinations(400, 264 )', 96233043661249133923407427694065507899588871551128233325734421851029638870893138420518579258044528704604701450.0);
  Eps; Test('combinations(400, 265 )', 49387524294075027221069472326010977639034288796050715970942948572603890137514969151662365204128512844627318480.0);
  Eps; Test('combinations(400, 266 )', 25065096916165897273851047985005571358156499952882882165704128034968139731445567050655711663749433210619127800.0);
  Eps; Test('combinations(400, 267 )', 12579486841821087021333484756519650044917494358375678689903944407062661887691782714561293494166382210572895600.0);
  Eps; Test('combinations(400, 268 )', 6242805037172405126258781614242960656619502797253601737900091813952738921876892168047209084791525500023116100.0);
  Eps; Test('combinations(400, 269 )', 3063383884411737831472710680595058760869049699767566652055063641047440660549255636365173231198815487000190800.0);
  Eps; Test('combinations(400, 270 )', 1486308477251620947862685552436861843236464854331671227478567914730424913081305512458658123285351217766759240.0);
  Eps; Test('combinations(400, 271 )', 712989306430666875358483844342406050261034800970912396945438483081015640961511869445112752867511654279257200.0);
  Eps; Test('combinations(400, 272 )', 338145663711603040151633882059449928248799593107528305904270457049452270897187614553013033529077218389794775.0);
  Eps; Test('combinations(400, 273 )', 158544487014964062781718450196372127530572702995471147090646954220988610530549504259288162240739501662614400.0);
  Eps; Test('combinations(400, 274 )', 73485948360950496252840303558172482468550121461404509782891106518487421669269295769816046002094586537051200.0);
  Eps; Test('combinations(400, 275 )', 33669925430835500101301375448471755603772055651407157209615561532106964110283386425442988350050610558794368.0);
  Eps; Test('combinations(400, 276 )', 15249060430632019973415478011083222646635894769659038591311395621425255484729794576740483854189588115396000.0);
  Eps; Test('combinations(400, 277 )', 6826294199994117244417037087993933603548198380641591282753115729446684765727417066844115515954905871152000.0);
  Eps; Test('combinations(400, 278 )', 3020266858270778493033437272745517385742548204384588948844004441445835345987310428855489958498033892632000.0);
  Eps; Test('combinations(400, 279 )', 1320690167415895971864083681989079286955522870734479755408489397334738036596601692904551164647885788176000.0);
  Eps; Test('combinations(400, 280 )', 570726822347583616412693305430994977577208097710257322872954346705368937243531445862323896151407787033200.0);
  Eps; Test('combinations(400, 281 )', 243726756874412932275883262105762979748273920730358999091653101795887090637807023144052909388501546064000.0);
  Eps; Test('combinations(400, 282 )', 102849234283883471421383362377963810602995023286924542169881982672732495694677431752277646160396042488000.0);
  Eps; Test('combinations(400, 283 )', 42884133022962012818810023889044981099482023844018006982494960973082807392126985677628135148151000048000.0);
  Eps; Test('combinations(400, 284 )', 17667054801713223590847791531754446438871115456866573299126445189615100228446680719304548634977700724000.0);
  Eps; Test('combinations(400, 285 )', 7190801252627136619432785325205318550558068045601833342802342603492461847367771801541500497043555382400.0);
  Eps; Test('combinations(400, 286 )', 2891406098084338151170525567827313403196425962392345574903039858047668225340187962158295654405625416000.0);
  Eps; Test('combinations(400, 287 )', 1148502770667646513008501445060326578273144807361419496651381685775031978009691385665664475965997552000.0);
  Eps; Test('combinations(400, 288 )', 450627823213347416562363414207697581058560288999445844172243508654786852482969189514653075639436539500.0);
  Eps; Test('combinations(400, 289 )', 174637772317975469394410734917862038334113329992864825423153193665522932450147229154467627929470216000.0);
  Eps; Test('combinations(400, 290 )', 66844112852742334837171005434078228465815791824855157317137946540941536213677042883261747241969634400.0);
  Eps; Test('combinations(400, 291 )', 25267534068046930694463266658929914540342739177780300016787539929565529152936339234222653596620824000.0);
  Eps; Test('combinations(400, 292 )', 9432058950058614540056493376107399605812871816363194184348773466858365334486510193596812472711198000.0);
  Eps; Test('combinations(400, 293 )', 3476663367257100240020823496995218967330341829922269528701936977545062990186153928015207327825288000.0);
  Eps; Test('combinations(400, 294 )', 1265316259511937842456558211491457243212063182998921223030977063256196394387477790127983619310564000.0);
  Eps; Test('combinations(400, 295 )', 454656011892425123052187018366421924679588804738595422512825656627650229847703883910394114057355200.0);
  Eps; Test('combinations(400, 296 )', 161280004218596749731350124758359128687016298978217970823806398465889439641921985846592506675751000.0);
  Eps; Test('combinations(400, 297 )', 56475152992370579030506440992826092200167323547928178335608974546978120278652816592746197623832000.0);
  Eps; Test('combinations(400, 298 )', 19519935430248891409873031618325796968514209145760410632777598585029350297655168151184088440452000.0);
  Eps; Test('combinations(400, 299 )', 6658974628379220480960030853074352143105181715276126704158244333354494081474338299066143882696000.0);
  Eps; Test('combinations(400, 300 )', 2241854791554337561923210387201698554845411177476295990399942258896013007429693894018935107174320.0);
  Eps; Test('combinations(400, 301 )', 744802256330344704957877205050398190978541919427340860597987461427246846322157439873400367832000.0);
  Eps; Test('combinations(400, 302 )', 244157031048689158247780938079435168565813410673201143043711121461249793993025121018101445084000.0);
  Eps; Test('combinations(400, 303 )', 78968280669212995076840039378827216235807637775490798740210197700338217199064230560310038344000.0);
  Eps; Test('combinations(400, 304 )', 25197115871426514876491723091270526233135989684942787755922332818857917987859310409046295129500.0);
  Eps; Test('combinations(400, 305 )', 7930895487399821075879362022170395142232967245096746310060799838066754514211455079568669942400.0);
  Eps; Test('combinations(400, 306 )', 2462206115369225497413527425183619406902391791778401632208418250380201564869569387447789688000.0);
  Eps; Test('combinations(400, 307 )', 753900243793834517123360188818437212536888691945178349927007542461690381425861636547531696000.0);
  Eps; Test('combinations(400, 308 )', 227638709976709773027508108961411236253021585554875280984453576132913004781185494152339116000.0);
  Eps; Test('combinations(400, 309 )', 67775926595007440513044485516018879402194129032519501134529867327598693980158787902961808000.0);
  Eps; Test('combinations(400, 310 )', 19895513935953797053829187683734574276127954006320240655620057828424132749014353868288788800.0);
  Eps; Test('combinations(400, 311 )', 5757544225838719404645102545132191912705838779964056781369148567711163818042739061562672000.0);
  Eps; Test('combinations(400, 312 )', 1642376397755275727607096559348606026380832216079490556223891738866325576300653129740634000.0);
  Eps; Test('combinations(400, 313 )', 461754386589342696579630981542100096873844201325863159577324194952832749886445608361584000.0);
  Eps; Test('combinations(400, 314 )', 127938317303416607014101577688416268879058743679458900901997468028332640892104356456872000.0);
  Eps; Test('combinations(400, 315 )', 34929191390774057787976938670488251186028736369630049135148515080751133703876110016796800.0);
  Eps; Test('combinations(400, 316 )', 9395510342455047189803923376555384021558362631071373976226657537543817610219839719708000.0);
  Eps; Test('combinations(400, 317 )', 2489662046581148151241418181800164851138493567854875123037978653481642521320083711216000.0);
  Eps; Test('combinations(400, 318 )', 649817452409544957713955060029602775611619390352058601296076189430743173803669647896000.0);
  Eps; Test('combinations(400, 319 )', 167037715039444158409229827343032688401732884040341082464822092580943386369595332688000.0);
  Eps; Test('combinations(400, 320 )', 42281421619359302597336300046205149251688636272711336498908092184551294674803818586650.0);
  Eps; Test('combinations(400, 321 )', 10537425948749981955722442379116548100109317451143012211565879672162316429857649492000.0);
  Eps; Test('combinations(400, 322 )', 2585269099227480045037493627174556832014397759752478151284796565530506204840851894000.0);
  Eps; Test('combinations(400, 323 )', 624306469782487441216484529162896077080876239197192866254532916753496854419772284000.0);
  Eps; Test('combinations(400, 324 )', 148369130164356583252065767733157401034652686475876082412342699351911289476303907000.0);
  Eps; Test('combinations(400, 325 )', 34695550438434154852790764146830653780411089760512560810270908156139255385227990560.0);
  Eps; Test('combinations(400, 326 )', 7982105162216446668586832242369015440278624944903196505430423655553509674515642000.0);
  Eps; Test('combinations(400, 327 )', 1806347957198828909710781608364853647035529804045371686244193732449418091480604000.0);
  Eps; Test('combinations(400, 328 )', 402022563644861312222216638447055842175590474680829674072640678258559514262451500.0);
  Eps; Test('combinations(400, 329 )', 87980621831094268936168990784766020172165696586686129280334738099137644458652000.0);
  Eps; Test('combinations(400, 330 )', 18929164090932403316569691956722386158253831689862773269405352742541735625952400.0);
  Eps; Test('combinations(400, 331 )', 4003146484487215202899934854895972903558212139850133319813820821685563425428000.0);
  Eps; Test('combinations(400, 332 )', 831979239245836894578600918638018464896134450751985539358896496073204446851000.0);
  Eps; Test('combinations(400, 333 )', 169893658464615341835870457860015782621432860814219269298513398597531238396000.0);
  Eps; Test('combinations(400, 334 )', 34080464422542598511985990049763645016874256510636799529941310497109559798000.0);
  Eps; Test('combinations(400, 335 )', 6714360154888989557585299532192240510787166954334414235749631321818599840800.0);
  Eps; Test('combinations(400, 336 )', 1298908958535072384651918064263379860717755511999217039653946535470860088250.0);
  Eps; Test('combinations(400, 337 )', 246677072244049354948732213984736828148179088332195520883835543828293904000.0);
  Eps; Test('combinations(400, 338 )', 45978270862056536573284406748634379211051131848900348567105441601131704000.0);
  Eps; Test('combinations(400, 339 )', 8409005290405620258240805954027526581372183405993574074219874275133232000.0);
  Eps; Test('combinations(400, 340 )', 1508674478572773046331438715281409180775597611075317701551212737597432800.0);
  Eps; Test('combinations(400, 341 )', 265455919983479128386763410313444430635002512212665871240682593125648000.0);
  Eps; Test('combinations(400, 342 )', 45795027131652832090114155580389536279137860293997913459649921036296000.0);
  Eps; Test('combinations(400, 343 )', 7743765520804268983167991322631466776064127979743087407171123673776000.0);
  Eps; Test('combinations(400, 344 )', 1283123938040242244303998562180213971615277019899290645955680376178000.0);
  Eps; Test('combinations(400, 345 )', 208275189942763958495721505745194151914363806128580510647878553814400.0);
  Eps; Test('combinations(400, 346 )', 33107327881075195714637811606895024148237021205410196779287053352000.0);
  Eps; Test('combinations(400, 347 )', 5152149007429569361932109010871271769466279957037898057871760464000.0);
  Eps; Test('combinations(400, 348 )', 784666371821170046501154533264877597073887464721289072032193404000.0);
  Eps; Test('combinations(400, 349 )', 116913041073641382286704973437746805294676642308043070904510192000.0);
  Eps; Test('combinations(400, 350 )', 17035900270730601418919867558071677342938596450600561760371485120.0);
  Eps; Test('combinations(400, 351 )', 2426766420331994504119639253286563724065327129715179737944656000.0);
  Eps; Test('combinations(400, 352 )', 337816916466669689493927055145004609315912015216033543066159500.0);
  Eps; Test('combinations(400, 353 )', 45935444732011742480760619396487878887149509151188697074152000.0);
  Eps; Test('combinations(400, 354 )', 6098773735606078803942794100663644936994426356231267690636000.0);
  Eps; Test('combinations(400, 355 )', 790263638979942605581319798959232865075334119398981165547200.0);
  Eps; Test('combinations(400, 356 )', 99892875713756790031346603801026626203342795991444248454000.0);
  Eps; Test('combinations(400, 357 )', 12311726978726327062686976378838015554473621914911896168000.0);
  Eps; Test('combinations(400, 358 )', 1478782849400089563395363084609035387827837269109529428000.0);
  Eps; Test('combinations(400, 359 )', 173005235862963124408371168672923360135847257110307064000.0);
  Eps; Test('combinations(400, 360 )', 19703374084393022502064494209971827126582604282007193400.0);
  Eps; Test('combinations(400, 361 )', 2183199344531082825713517363985798019565939532632376000.0);
  Eps; Test('combinations(400, 362 )', 235206559217437100007809881755376029732242104344372000.0);
  Eps; Test('combinations(400, 363 )', 24622174243147685400266599192022835068388980620072000.0);
  Eps; Test('combinations(400, 364 )', 2502803425814462526950176291496826641567011766326000.0);
  Eps; Test('combinations(400, 365 )', 246851844737864797178647524640782901634006639966400.0);
  Eps; Test('combinations(400, 366 )', 23606050726298546178285965471113119008716481964000.0);
  Eps; Test('combinations(400, 367 )', 2186936579548094196353468190784321652033679528000.0);
  Eps; Test('combinations(400, 368 )', 196111160665997577390392527977941887274759305500.0);
  Eps; Test('combinations(400, 369 )', 17006929922254532456619406220309323557702704000.0);
  Eps; Test('combinations(400, 370 )', 1424904939432136503122166467106997379158875200.0);
  Eps; Test('combinations(400, 371 )', 115221423673757668716078150979002483489936000.0);
  Eps; Test('combinations(400, 372 )', 8982315286395087077328673060191053820452000.0);
  Eps; Test('combinations(400, 373 )', 674275678335288037976415135885655514672000.0);
  Eps; Test('combinations(400, 374 )', 48677655922601008089206440291210424856000.0);
  Eps; Test('combinations(400, 375 )', 3374984143967003227518313193523922790016.0);
  Eps; Test('combinations(400, 376 )', 224400541487167767787121887867282100400.0);
  Eps; Test('combinations(400, 377 )', 14285445611915189461249138750171804800.0);
  Eps; Test('combinations(400, 378 )', 869220235645633221187116908079236800.0);
  Eps; Test('combinations(400, 379 )', 50456055894997179066270638463702400.0);
  Eps; Test('combinations(400, 380 )', 2788360983670896737872851072994080.0);
  Eps; Test('combinations(400, 381 )', 146370655310808227709860948713600.0);
  Eps; Test('combinations(400, 382 )', 7280215840066377818029733051200.0);
  Eps; Test('combinations(400, 383 )', 342151136086670497975287715200.0);
  Eps; Test('combinations(400, 384 )', 15147315920503641837447633225.0);
  Eps; Test('combinations(400, 385 )', 629498843449501998439382160.0);
  Eps; Test('combinations(400, 386 )', 24462390289488419628473400.0);
  Eps; Test('combinations(400, 387 )', 884944351557720606714800.0);
  Eps; Test('combinations(400, 388 )', 29650197346006102802300.0);
  Eps; Test('combinations(400, 389 )', 914659044092733248400.0);
  Eps; Test('combinations(400, 390 )', 25798075602615553160.0);
  Test('combinations(400, 391 )', 659797329990167600);
  Test('combinations(400, 392 )', 15148408086508950);
  Test('combinations(400, 393 )', 308364541201200);
  Test('combinations(400, 394 )', 5478557838600);
  Test('combinations(400, 395 )', 83218600080);
  Test('combinations(400, 396 )', 1050739900);
  Test('combinations(400, 397 )', 10586800);
  Test('combinations(400, 398 )', 79800);
  Test('combinations(400, 399 )', 400);
  Test('combinations(400, 400 )', 1);

  Eps; Test('combinations(500, 250)', 1.1674431578827768292093473476217661965923008118031144E149);
  Eps; Test('combinations(800, 400)', 1.8804244186835312700958607615195351332156581822914058E239);
  Eps; Test('combinations(1000, 500)', 2.7028824094543656951561469362597527549615200844654828E299);
  Eps; Test('combinations(1200, 600)', 3.9650964622610201403671601486260572953160819570679165E359);
  Eps; Test('combinations(1500, 750)', 7.2246289244893021702811607322848529594437634304054652E449);
  Eps; Test('combinations(2000, 1000)', 2.0481516269894897143351625029808250443964248879813970E600);
  Eps; Test('combinations(3000, 1500)', 1.7919679375475600507326903666685292846560897853633339E901);
  Eps; Test('combinations(4000, 2000)', 1.662897875033835069539536826460381558016225597388640E1202);
  Eps; Test('combinations(5000, 2500)', 1.593718685349438375691025792935919087257129434168729E1503);
  Eps; Test('combinations(8000, 4000)', 1.550053928509880709855382381537398942593569455195730E2406);
  Eps; Test('combinations(10000, 5000)', 1.591790263532438948337597273641521188653005837457614E3008);

  Test('combinations(0, 1)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(1, 2)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(4, 5)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(10, 15)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(10, 11)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(-10, 5)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(10, -1)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(10, -3)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(10, -11)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(-10, -5)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(-10, -50)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(-1, 0)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('combinations(10, 5, 2)', failure, 'Too many arguments.');
  Test('combinations(10)', failure, 'A required argument of type integer is missing.');
  Test('combinations(10, "cat")', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
  Test('combinations(10, 5.5)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');

  // Permutations

  Test('permutations(0, 0)', 1);

  Test('permutations(1, 0)', 1);
  Test('permutations(1, 1)', 1);

  Test('permutations(2, 0)', 1);
  Test('permutations(2, 1)', 2);
  Test('permutations(2, 2)', 2);

  Test('permutations(5, 0)', 1);
  Test('permutations(5, 1)', 5);
  Test('permutations(5, 2)', 20);
  Test('permutations(5, 3)', 60);
  Test('permutations(5, 4)', 120);
  Test('permutations(5, 5)', 120);

  Test('permutations(10, 0   )', 1);
  Test('permutations(10, 1   )', 10);
  Test('permutations(10, 2   )', 90);
  Test('permutations(10, 3   )', 720);
  Test('permutations(10, 4   )', 5040);
  Test('permutations(10, 5   )', 30240);
  Test('permutations(10, 6   )', 151200);
  Test('permutations(10, 7   )', 604800);
  Test('permutations(10, 8   )', 1814400);
  Test('permutations(10, 9   )', 3628800);
  Test('permutations(10, 10  )', 3628800);

  Test('permutations(20, 0   )', 1);
  Test('permutations(20, 1   )', 20);
  Test('permutations(20, 2   )', 380);
  Test('permutations(20, 3   )', 6840);
  Test('permutations(20, 4   )', 116280);
  Test('permutations(20, 5   )', 1860480);
  Test('permutations(20, 6   )', 27907200);
  Test('permutations(20, 7   )', 390700800);
  Test('permutations(20, 8   )', 5079110400);
  Test('permutations(20, 9   )', 60949324800);
  Test('permutations(20, 10  )', 670442572800);
  Test('permutations(20, 11  )', 6704425728000);
  Test('permutations(20, 12  )', 60339831552000);
  Test('permutations(20, 13  )', 482718652416000);
  Test('permutations(20, 14  )', 3379030566912000);
  Test('permutations(20, 15  )', 20274183401472000);
  Test('permutations(20, 16  )', 101370917007360000);
  Test('permutations(20, 17  )', 405483668029440000);
  Test('permutations(20, 18  )', 1216451004088320000);
  Test('permutations(20, 19  )', 2432902008176640000);
  Test('permutations(20, 20  )', 2432902008176640000);

  Test('permutations(50, 0   )', 1);
  Test('permutations(50, 1   )', 50);
  Test('permutations(50, 2   )', 2450);
  Test('permutations(50, 3   )', 117600);
  Test('permutations(50, 4   )', 5527200);
  Test('permutations(50, 5   )', 254251200);
  Test('permutations(50, 6   )', 11441304000);
  Test('permutations(50, 7   )', 503417376000);
  Test('permutations(50, 8   )', 21646947168000);
  Test('permutations(50, 9   )', 909171781056000);
  Test('permutations(50, 10  )', 37276043023296000);
  Test('permutations(50, 11  )', 1491041720931840000);
  Eps; Test('permutations(50, 12  )', 58150627116341760000.0);
  Eps; Test('permutations(50, 13  )', 2209723830420986880000.0);
  Eps; Test('permutations(50, 14  )', 81759781725576514560000.0);
  Eps; Test('permutations(50, 15  )', 2943352142120754524160000.0);
  Eps; Test('permutations(50, 16  )', 103017324974226408345600000.0);
  Eps; Test('permutations(50, 17  )', 3502589049123697883750400000.0);
  Eps; Test('permutations(50, 18  )', 115585438621082030163763200000.0);
  Eps; Test('permutations(50, 19  )', 3698734035874624965240422400000.0);
  Eps; Test('permutations(50, 20  )', 114660755112113373922453094400000.0);
  Eps; Test('permutations(50, 21  )', 3439822653363401217673592832000000.0);
  Eps; Test('permutations(50, 22  )', 99754856947538635312534192128000000.0);
  Eps; Test('permutations(50, 23  )', 2793135994531081788750957379584000000.0);
  Eps; Test('permutations(50, 24  )', 75414671852339208296275849248768000000.0);
  Eps; Test('permutations(50, 25  )', 1960781468160819415703172080467968000000.0);
  Eps; Test('permutations(50, 26  )', 49019536704020485392579302011699200000000.0);
  Eps; Test('permutations(50, 27  )', 1176468880896491649421903248280780800000000.0);
  Eps; Test('permutations(50, 28  )', 27058784260619307936703774710457958400000000.0);
  Eps; Test('permutations(50, 29  )', 595293253733624774607483043630075084800000000.0);
  Eps; Test('permutations(50, 30  )', 12501158328406120266757143916231576780800000000.0);
  Eps; Test('permutations(50, 31  )', 250023166568122405335142878324631535616000000000.0);
  Eps; Test('permutations(50, 32  )', 4750440164794325701367714688167999176704000000000.0);
  Eps; Test('permutations(50, 33  )', 85507922966297862624618864387023985180672000000000.0);
  Eps; Test('permutations(50, 34  )', 1453634690427063664618520694579407748071424000000000.0);
  Eps; Test('permutations(50, 35  )', 23258155046833018633896331113270523969142784000000000.0);
  Eps; Test('permutations(50, 36  )', 348872325702495279508444966699057859537141760000000000.0);
  Eps; Test('permutations(50, 37  )', 4884212559834933913118229533786810033519984640000000000.0);
  Eps; Test('permutations(50, 38  )', 63494763277854140870536983939228530435759800320000000000.0);
  Eps; Test('permutations(50, 39  )', 761937159334249690446443807270742365229117603840000000000.0);
  Eps; Test('permutations(50, 40  )', 8381308752676746594910881879978166017520293642240000000000.0);
  Eps; Test('permutations(50, 41  )', 83813087526767465949108818799781660175202936422400000000000.0);
  Eps; Test('permutations(50, 42  )', 754317787740907193541979369198034941576826427801600000000000.0);
  Eps; Test('permutations(50, 43  )', 6034542301927257548335834953584279532614611422412800000000000.0);
  Eps; Test('permutations(50, 44  )', 42241796113490802838350844675089956728302279956889600000000000.0);
  Eps; Test('permutations(50, 45  )', 253450776680944817030105068050539740369813679741337600000000000.0);
  Eps; Test('permutations(50, 46  )', 1267253883404724085150525340252698701849068398706688000000000000.0);
  Eps; Test('permutations(50, 47  )', 5069015533618896340602101361010794807396273594826752000000000000.0);
  Eps; Test('permutations(50, 48  )', 15207046600856689021806304083032384422188820784480256000000000000.0);
  Eps; Test('permutations(50, 49  )', 30414093201713378043612608166064768844377641568960512000000000000.0);
  Eps; Test('permutations(50, 50  )', 30414093201713378043612608166064768844377641568960512000000000000.0);

  Test('permutations(100, 0   )', 1);
  Test('permutations(100, 1   )', 100);
  Test('permutations(100, 2   )', 9900);
  Test('permutations(100, 3   )', 970200);
  Test('permutations(100, 4   )', 94109400);
  Test('permutations(100, 5   )', 9034502400);
  Test('permutations(100, 6   )', 858277728000);
  Test('permutations(100, 7   )', 80678106432000);
  Test('permutations(100, 8   )', 7503063898176000);
  Test('permutations(100, 9   )', 690281878632192000);
  Eps; Test('permutations(100, 10  )', 62815650955529472000.0);
  Eps; Test('permutations(100, 11  )', 5653408585997652480000.0);
  Eps; Test('permutations(100, 12  )', 503153364153791070720000.0);
  Eps; Test('permutations(100, 13  )', 44277496045533614223360000.0);
  Eps; Test('permutations(100, 14  )', 3852142155961424437432320000.0);
  Eps; Test('permutations(100, 15  )', 331284225412682501619179520000.0);
  Eps; Test('permutations(100, 16  )', 28159159160078012637630259200000.0);
  Eps; Test('permutations(100, 17  )', 2365369369446553061560941772800000.0);
  Eps; Test('permutations(100, 18  )', 196325657664063904109558167142400000.0);
  Eps; Test('permutations(100, 19  )', 16098703928453240136983769705676800000.0);
  Eps; Test('permutations(100, 20  )', 1303995018204712451095685346159820800000.0);
  Eps; Test('permutations(100, 21  )', 104319601456376996087654827692785664000000.0);
  Eps; Test('permutations(100, 22  )', 8241248515053782690924731387730067456000000.0);
  Eps; Test('permutations(100, 23  )', 642817384174195049892129048242945261568000000.0);
  Eps; Test('permutations(100, 24  )', 49496938581413018841693936714706785140736000000.0);
  Eps; Test('permutations(100, 25  )', 3761767332187389431968739190317715670695936000000.0);
  Eps; Test('permutations(100, 26  )', 282132549914054207397655439273828675302195200000000.0);
  Eps; Test('permutations(100, 27  )', 20877808693640011347426502506263321972362444800000000.0);
  Eps; Test('permutations(100, 28  )', 1524080034635720828362134682957222503982458470400000000.0);
  Eps; Test('permutations(100, 29  )', 109733762493771899642073697172920020286737009868800000000.0);
  Eps; Test('permutations(100, 30  )', 7791097137057804874587232499277321440358327700684800000000.0);
  Eps; Test('permutations(100, 31  )', 545376799594046341221106274949412500825082939047936000000000.0);
  Eps; Test('permutations(100, 32  )', 37630999171989197544256332971509462556930722794307584000000000.0);
  Eps; Test('permutations(100, 33  )', 2558907943695265433009430642062643453871289150012915712000000000.0);
  Eps; Test('permutations(100, 34  )', 171446832227582784011631853018197111409376373050865352704000000000.0);
  Eps; Test('permutations(100, 35  )', 11315490927020463744767702299201009353018840621357113278464000000000.0);
  Eps; Test('permutations(100, 36  )', 735506910256330143409900649448065607946224640388212363100160000000000.0);
  Eps; Test('permutations(100, 37  )', 47072442256405129178233641564676198908558376984845591238410240000000000.0);
  Eps; Test('permutations(100, 38  )', 2965563862153523138228719418574600531239177750045272248019845120000000000.0);
  Eps; Test('permutations(100, 39  )', 183864959453518434570180603951625232936829020502806879377230397440000000000.0);
  Eps; Test('permutations(100, 40  )', 11215762526664624508781016841049139209146570250671219642011054243840000000000.0);
  Eps; Test('permutations(100, 41  )', 672945751599877470526861010462948352548794215040273178520663254630400000000000.0);
  Eps; Test('permutations(100, 42  )', 39703799344392770761084799617313952800378858687376117532719132023193600000000000.0);
  Eps; Test('permutations(100, 43  )', 2302820361974780704142918377804209262421973803867814816897709657345228800000000000.0);
  Eps; Test('permutations(100, 44  )', 131260760632562500136146347534839927958052506820465444563169450468678041600000000000.0);
  Eps; Test('permutations(100, 45  )', 7350602595423500007624195461951035965650940381946064895537489226245970329600000000000.0);
  Eps; Test('permutations(100, 46  )', 404283142748292500419330750407306978110801721007033569254561907443528368128000000000000.0);
  Eps; Test('permutations(100, 47  )', 21831289708407795022643860521994576817983292934379812739746343001950531878912000000000000.0);
  Eps; Test('permutations(100, 48  )', 1157058354545613136200124607665712571353114525522130075206556179103378189582336000000000000.0);
  Eps; Test('permutations(100, 49  )', 60167034436371883082406479598617053710361955327150763910740921313375665858281472000000000000.0);
  Eps; Test('permutations(100, 50  )', 3068518756254966037202730459529469739228459721684688959447786986982158958772355072000000000000.0);
  Eps; Test('permutations(100, 51  )', 153425937812748301860136522976473486961422986084234447972389349349107947938617753600000000000000.0);
  Eps; Test('permutations(100, 52  )', 7517870952824666791146689625847200861109726318127487950647078118106289448992269926400000000000000.0);
  Eps; Test('permutations(100, 53  )', 360857805735584005975041102040665641333266863270119421631059749669101893551628956467200000000000000.0);
  Eps; Test('permutations(100, 54  )', 16960316869572448280826931795911285142663542573695612816659808234447788996926560953958400000000000000.0);
  Eps; Test('permutations(100, 55  )', 780174576000332620918038862611919116562522958389998189566351178784598293858621803882086400000000000000.0);
  Eps; Test('permutations(100, 56  )', 35107855920014967941311748817536360245313533127549918530485803045306923223637981174693888000000000000000.0);
  Eps; Test('permutations(100, 57  )', 1544745660480658589417716947971599850793795457612196415341375333993504621840071171686531072000000000000000.0);
  Eps; Test('permutations(100, 58  )', 66424063400668319344961828762778793584133204677324445859679139361720698739123060382520836096000000000000000.0);
  Eps; Test('permutations(100, 59  )', 2789810662828069412488396808036709330533594596447626726106523853192269347043168536065875116032000000000000000.0);
  Eps; Test('permutations(100, 60  )', 114382237175950845912024269129505082551877378454352695770367477980883043228769909978700879757312000000000000000.0);
  Eps; Test('permutations(100, 61  )', 4575289487038033836480970765180203302075095138174107830814699119235321729150796399148035190292480000000000000000.0);
  Eps; Test('permutations(100, 62  )', 178436289994483319622757859842027928780928710388790205401773265650177547436881059566773372421406720000000000000000.0);
  Eps; Test('permutations(100, 63  )', 6780579019790366145664798673997061293675290994774027805267384094706746802601480263537388152013455360000000000000000.0);
  Eps; Test('permutations(100, 64  )', 250881423732243547389597550937891267865985766806639028794893211504149631696254769750883361624497848320000000000000000.0);
  Eps; Test('permutations(100, 65  )', 9031731254360767706025511833764085643175487605039005036616155614149386741065171711031801018481922539520000000000000000.0);
  Eps; Test('permutations(100, 66  )', 316110593902626869710892914181742997511142066176365176281565446495228535937281009886113035646867288883200000000000000000.0);
  Eps; Test('permutations(100, 67  )', 10747760192689313570170359082179261915378830249996415993573225180837770221867554336127843211993487822028800000000000000000.0);
  Eps; Test('permutations(100, 68  )', 354676086358747347815621849711915643207501398249881727787916430967646417321629293092218825995785098126950400000000000000000.0);
  Eps; Test('permutations(100, 69  )', 11349634763479915130099899190781300582640044743996215289213325790964685354292137378951002431865123140062412800000000000000000.0);
  Eps; Test('permutations(100, 70  )', 351838677667877369033096874914220318061841387063882673965613099519905245983056258747481075387818817341934796800000000000000000.0);
  Eps; Test('permutations(100, 71  )', 10555160330036321070992906247426609541855241611916480218968392985597157379491687762424432261634564520258043904000000000000000000.0);
  Eps; Test('permutations(100, 72  )', 306099649571053311058794281175371676713802006745577926350083396582317564005258945110308535587402371087483273216000000000000000000.0);
  Eps; Test('permutations(100, 73  )', 8570790187989492709646239872910406947986456188876181937802335104304891792147250463088638996447266390449531650048000000000000000000.0);
  Eps; Test('permutations(100, 74  )', 231411335075716303160448476568580987595634317099656912320663047816232078387975762503393252904076192542137354551296000000000000000000.0);
  Eps; Test('permutations(100, 75  )', 6016694711968623882171660390783105677486492244591079720337239243222034038087369825088224575505981006095571218333696000000000000000000.0);
  Eps; Test('permutations(100, 76  )', 150417367799215597054291509769577641937162306114776993008430981080550850952184245627205614387649525152389280458342400000000000000000000.0);
  Eps; Test('permutations(100, 77  )', 3610016827181174329302996234469863406491895346754647832202343545933220422852421895052934745303588603657342731000217600000000000000000000.0);
  Eps; Test('permutations(100, 78  )', 83030387025167009573968913392806858349313592975356900140653901556464069725605703586217499141982537884118882813005004800000000000000000000.0);
  Eps; Test('permutations(100, 79  )', 1826668514553674210627316094641750883684899045457851803094385834242209533963325478896784981123615833450615421886110105600000000000000000000.0);
  Eps; Test('permutations(100, 80  )', 38360038805627158423173637987476768557382879954614887864982102519086400213229835056832484603595932502462923859608312217600000000000000000000.0);
  Eps; Test('permutations(100, 81  )', 767200776112543168463472759749535371147657599092297757299642050381728004264596701136649692071918650049258477192166244352000000000000000000000.0);
  Eps; Test('permutations(100, 82  )', 14576814746138320200805982435241172051805494382753657388693198957252832081027337321596344149366454350935911066651158642688000000000000000000000.0);
  Eps; Test('permutations(100, 83  )', 262382665430489763614507683834341096932498898889565832996477581230550977458492071788734194688596178316846399199720855568384000000000000000000000.0);
  Eps; Test('permutations(100, 84  )', 4460505312318325981446630625183798647852481281122619160940118880919366616794365220408481309706135031386388786395254544662528000000000000000000000.0);
  Eps; Test('permutations(100, 85  )', 71368084997093215703146090002940778365639700497961906575041902094709865868709843526535700955298160502182220582324072714600448000000000000000000000.0);
  Eps; Test('permutations(100, 86  )', 1070521274956398235547191350044111675484595507469428598625628531420647988030647652898035514329472407532733308734861090719006720000000000000000000000.0);
  Eps; Test('permutations(100, 87  )', 14987297849389575297660678900617563456784337104572000380758799439889071832429067140572497200612613705458266322288055270066094080000000000000000000000.0);
  Eps; Test('permutations(100, 88  )', 194834872042064478869588825708028324938196382359436004949864392718557933821577872827442463607963978170957462189744718510859223040000000000000000000000.0);
  Eps; Test('permutations(100, 89  )', 2338018464504773746435065908496339899258356588313232059398372712622695205858934473929309563295567738051489546276936622130310676480000000000000000000000.0);
  Eps; Test('permutations(100, 90  )', 25718203109552511210785724993459738891841922471445552653382099838849647264448279213222405196251245118566385009046302843433417441280000000000000000000000.0);
  Eps; Test('permutations(100, 91  )', 257182031095525112107857249934597388918419224714455526533820998388496472644482792132224051962512451185663850090463028434334174412800000000000000000000000.0);
  Eps; Test('permutations(100, 92  )', 2314638279859726008970715249411376500265773022430099738804388985496468253800345129190016467662612060670974650814167255909007569715200000000000000000000000.0);
  Eps; Test('permutations(100, 93  )', 18517106238877808071765721995291012002126184179440797910435111883971746030402761033520131741300896485367797206513338047272060557721600000000000000000000000.0);
  Eps; Test('permutations(100, 94  )', 129619743672144656502360053967037084014883289256085585373045783187802222212819327234640922189106275397574580445593366330904423904051200000000000000000000000.0);
  Eps; Test('permutations(100, 95  )', 777718462032867939014160323802222504089299735536513512238274699126813333276915963407845533134637652385447482673560197985426543424307200000000000000000000000.0);
  Eps; Test('permutations(100, 96  )', 3888592310164339695070801619011112520446498677682567561191373495634066666384579817039227665673188261927237413367800989927132717121536000000000000000000000000.0);
  Eps; Test('permutations(100, 97  )', 15554369240657358780283206476044450081785994710730270244765493982536266665538319268156910662692753047708949653471203959708530868486144000000000000000000000000.0);
  Eps; Test('permutations(100, 98  )', 46663107721972076340849619428133350245357984132190810734296481947608799996614957804470731988078259143126848960413611879125592605458432000000000000000000000000.0);
  Eps; Test('permutations(100, 99  )', 93326215443944152681699238856266700490715968264381621468592963895217599993229915608941463976156518286253697920827223758251185210916864000000000000000000000000.0);
  Eps; Test('permutations(100, 100 )', 93326215443944152681699238856266700490715968264381621468592963895217599993229915608941463976156518286253697920827223758251185210916864000000000000000000000000.0);

  Eps; Test('permutations(200, 200)', 7.8865786736479050355236321393218506229513597768717326E374);
  Eps; Test('permutations(500, 500)', 1.220136825991110068701238785423046926253574342803192E1134);

  for _i := 0 to 1000 do
    Test(Format('permutations(%d, %d) ≈ %d!', [_i, _i, _i]), True);

  // Prime numbers

  for _i := -100 to SmallPrimes[High(SmallPrimes)] do
    Test('IsPrime(' + _i.ToString + ')', IsSmallPrime(_i));

  Test('IsPrime(99871)', True);
  Test('IsPrime(99872)', False);
  Test('IsPrime(99873)', False);
  Test('IsPrime(99874)', False);
  Test('IsPrime(99875)', False);
  Test('IsPrime(99876)', False);
  Test('IsPrime(99877)', True);
  Test('IsPrime(99878)', False);
  Test('IsPrime(99879)', False);
  Test('IsPrime(99880)', False);
  Test('IsPrime(99881)', True);
  Test('IsPrime(99882)', False);
  Test('IsPrime(99883)', False);
  Test('IsPrime(99884)', False);
  Test('IsPrime(99885)', False);
  Test('IsPrime(99886)', False);
  Test('IsPrime(99887)', False);
  Test('IsPrime(99888)', False);
  Test('IsPrime(99889)', False);
  Test('IsPrime(99890)', False);
  Test('IsPrime(99891)', False);
  Test('IsPrime(99892)', False);
  Test('IsPrime(99893)', False);
  Test('IsPrime(99894)', False);
  Test('IsPrime(99895)', False);
  Test('IsPrime(99896)', False);
  Test('IsPrime(99897)', False);
  Test('IsPrime(99898)', False);
  Test('IsPrime(99899)', False);
  Test('IsPrime(99900)', False);
  Test('IsPrime(99901)', True);
  Test('IsPrime(99902)', False);
  Test('IsPrime(99903)', False);
  Test('IsPrime(99904)', False);
  Test('IsPrime(99905)', False);
  Test('IsPrime(99906)', False);
  Test('IsPrime(99907)', True);

  Test('IsPrime(1000000000000000000)', False); // because the first prime to be tested, 2, will show it isn't prime

  Test('IsPrime(574298344)', False);
  Test('IsPrime(574298345)', False);
  Test('IsPrime(574298346)', False);
  Test('IsPrime(574298347)', False);
  Test('IsPrime(574298348)', False);
  Test('IsPrime(574298349)', False);
  Test('IsPrime(574298350)', False);
  Test('IsPrime(574298351)', False);
  Test('IsPrime(574298352)', False);
  Test('IsPrime(574298353)', True);
  Test('IsPrime(574298354)', False);
  Test('IsPrime(574298355)', False);
  Test('IsPrime(574298356)', False);
  Test('IsPrime(574298357)', False);
  Test('IsPrime(574298358)', False);
  Test('IsPrime(574298359)', True);
  Test('IsPrime(574298360)', False);
  Test('IsPrime(574298361)', False);
  Test('IsPrime(574298362)', False);
  Test('IsPrime(574298363)', False);
  Test('IsPrime(574298364)', False);
  Test('IsPrime(574298365)', False);
  Test('IsPrime(574298366)', False);
  Test('IsPrime(574298367)', False);
  Test('IsPrime(574298368)', False);
  Test('IsPrime(574298369)', False);
  Test('IsPrime(574298370)', False);
  Test('IsPrime(574298371)', True);
  Test('IsPrime(574298372)', False);
  Test('IsPrime(574298373)', False);
  Test('IsPrime(574298374)', False);
  Test('IsPrime(574298375)', False);
  Test('IsPrime(574298376)', False);

  Test('IsPrime(253241523120330)', False);
  Test('IsPrime(253241523120331)', False);
  Test('IsPrime(253241523120332)', False);
  Test('IsPrime(253241523120333)', False);
  Test('IsPrime(253241523120334)', False);
  Test('IsPrime(253241523120335)', False);
  Test('IsPrime(253241523120336)', False);
  Test('IsPrime(253241523120337)', True);
  Test('IsPrime(253241523120338)', False);
  Test('IsPrime(253241523120339)', False);
  Test('IsPrime(253241523120340)', False);
  Test('IsPrime(253241523120341)', False);
  Test('IsPrime(253241523120342)', False);
  Test('IsPrime(253241523120343)', False);

  Test('IsPrime(11253241543723530)', False);
  Test('IsPrime(11253241543723531)', False);
  Test('IsPrime(11253241543723532)', False);
  Test('IsPrime(11253241543723533)', False);
  Test('IsPrime(11253241543723534)', False);
  Test('IsPrime(11253241543723535)', False);
  Test('IsPrime(11253241543723536)', False);
  Test('IsPrime(11253241543723537)', True);
  Test('IsPrime(11253241543723538)', False);
  Test('IsPrime(11253241543723539)', False);
  Test('IsPrime(11253241543723540)', False);
  Test('IsPrime(11253241543723541)', False);
  Test('IsPrime(11253241543723542)', False);
  Test('IsPrime(11253241543723543)', False);
  Test('IsPrime(11253241543723544)', False);

  Test('IsPrime(51213576)', False);
  Test('IsPrime(51213577)', False);
  Test('IsPrime(51213578)', False);
  Test('IsPrime(51213579)', False);
  Test('IsPrime(51213580)', False);
  Test('IsPrime(51213581)', False);
  Test('IsPrime(51213582)', False);
  Test('IsPrime(51213583)', True);
  Test('IsPrime(51213584)', False);
  Test('IsPrime(51213585)', False);
  Test('IsPrime(51213586)', False);
  Test('IsPrime(51213587)', False);
  Test('IsPrime(51213588)', False);
  Test('IsPrime(51213589)', False);
  Test('IsPrime(51213590)', False);

  Test('IsPrime(-11253241543723537)', False);
  Test('IsPrime(-51213583)', False);

{$IFNDEF QuickTest}
  Test('IsPrime(611253241543723537)', False);
  Test('IsPrime(611253241543723538)', False);
  Test('IsPrime(611253241543723539)', False);
  Test('IsPrime(611253241543723540)', False);
  Test('IsPrime(611253241543723541)', False);
  Test('IsPrime(611253241543723542)', False);
  Test('IsPrime(611253241543723543)', False);
  Test('IsPrime(611253241543723544)', False);
  Test('IsPrime(611253241543723545)', False);
  Test('IsPrime(611253241543723546)', False);
  Test('IsPrime(611253241543723547)', False);
  Test('IsPrime(611253241543723548)', False);
  Test('IsPrime(611253241543723549)', False);
  Test('IsPrime(611253241543723550)', False);
  Test('IsPrime(611253241543723551)', False);
  Test('IsPrime(611253241543723552)', False);
  Test('IsPrime(611253241543723553)', False);
  Test('IsPrime(611253241543723554)', False);
  Test('IsPrime(611253241543723555)', False);
  Test('IsPrime(611253241543723556)', False);
  Test('IsPrime(611253241543723557)', False);
  Test('IsPrime(611253241543723558)', False);
  Test('IsPrime(611253241543723559)', False);
  Test('IsPrime(611253241543723560)', False);
  Test('IsPrime(611253241543723561)', False);
  Test('IsPrime(611253241543723562)', False);
  Test('IsPrime(611253241543723563)', False);
  Test('IsPrime(611253241543723564)', False);
  Test('IsPrime(611253241543723565)', False);
  Test('IsPrime(611253241543723566)', False);
  Test('IsPrime(611253241543723567)', True);
  Test('IsPrime(611253241543723568)', False);
  Test('IsPrime(611253241543723569)', False);
  Test('IsPrime(611253241543723570)', False);
  Test('IsPrime(611253241543723571)', False);
  Test('IsPrime(611253241543723572)', False);
  Test('IsPrime(611253241543723573)', False);
  Test('IsPrime(611253241543723574)', False);
  Test('IsPrime(611253241543723575)', False);
  Test('IsPrime(611253241543723576)', False);
  Test('IsPrime(611253241543723577)', True);
  Test('IsPrime(611253241543723578)', False);
  Test('IsPrime(611253241543723579)', False);
  Test('IsPrime(611253241543723580)', False);
{$ENDIF}

  Test('IsPrime(1000000000000000001)', failure, 'Cannot test integers this large for primality.');
  Test('IsPrime(1611253241543723539)', failure, 'Cannot test integers this large for primality.');
  Test('IsPrime(18611253241543723539)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('IsPrime(-3)', False);
  Test('IsPrime(-3000000)', False);
  Test('IsPrime(3.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('IsPrime(2, 3)', failure, 'Too many arguments.');
  Test('IsPrime()', failure, 'A required argument of type integer is missing.');

  for _i := 1 to Length(SmallPrimes) do
    Test('prime(' + _i.ToString + ')', SmallPrimes[_i - 1]);

  Test('prime(10000)', 104729);

  Test('prime(0)', failure, 'To obtain the Nth prime number, N must be a positive integer.');
  Test('prime(-1)', failure, 'To obtain the Nth prime number, N must be a positive integer.');
  Test('prime(-10000000000)', failure, 'To obtain the Nth prime number, N must be a positive integer.');

  Test('prime(3.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('prime(2, 3)', failure, 'Too many arguments.');
  Test('prime()', failure, 'A required argument of type integer is missing.');

{$IFNDEF QuickTest}
  Test('prime(20000)', 224737);
  Test('prime(50000)', 611953);
  Test('prime(100000)', 1299709);
  Test('prime(200000)', 2750159);
  Test('prime(500000)', 7368787);
  Test('prime(1000000)', 15485863);
  Test('prime(2000000)', 32452843);
  Test('prime(5000000)', 86028121);
  Test('prime(10000000)', 179424673);
  Test('prime(20000000)', 373587883);
  Test('prime(50000000)', 982451653);
  Test('prime(100000000)', failure, 'Cannot test integers this large for primality.');
{$ENDIF}

  Test('NextPrime(-5)', 2);
  Test('NextPrime(0)', 2);
  Test('NextPrime(2)', 3);
  Test('NextPrime(3)', 5);
  Test('NextPrime(4)', 5);
  Test('NextPrime(5)', 7);
  Test('NextPrime(90)', 97);
  Test('NextPrime(97)', 101);
  Test('NextPrime(50000)', 50021);
  Test('NextPrime(50010)', 50021);
  Test('NextPrime(50021)', 50023);

  Test('NextPrime(536281473)', 536281511);

{$IFNDEF QuickTest}
  Test('NextPrime(11253241543723530)', 11253241543723537);
  Test('NextPrime(11253241543723531)', 11253241543723537);
  Test('NextPrime(11253241543723532)', 11253241543723537);
  Test('NextPrime(11253241543723533)', 11253241543723537);
  Test('NextPrime(11253241543723534)', 11253241543723537);
  Test('NextPrime(11253241543723535)', 11253241543723537);
  Test('NextPrime(11253241543723536)', 11253241543723537);
  Test('NextPrime(11253241543723537)', 11253241543723549);
  Test('NextPrime(1500000000000000000)', failure, 'Cannot test integers this large for primality.');
{$ENDIF}

  Test('PrevPrime(-45)', failure, 'There are no prime numbers smaller than 2.');
  Test('PrevPrime(-1)', failure, 'There are no prime numbers smaller than 2.');
  Test('PrevPrime(0)', failure, 'There are no prime numbers smaller than 2.');
  Test('PrevPrime(1)', failure, 'There are no prime numbers smaller than 2.');
  Test('PrevPrime(2)', failure, 'There are no prime numbers smaller than 2.');
  Test('PrevPrime(3)', 2);
  Test('PrevPrime(4)', 3);
  Test('PrevPrime(5)', 3);
  Test('PrevPrime(6)', 5);
  Test('PrevPrime(7)', 5);
  Test('PrevPrime(8)', 7);
  Test('PrevPrime(9)', 7);
  Test('PrevPrime(10)', 7);
  Test('PrevPrime(11)', 7);
  Test('PrevPrime(12)', 11);
  Test('PrevPrime(100)', 97);
  Test('PrevPrime(60000)', 59999);
  Test('PrevPrime(59999)', 59981);
  Test('PrevPrime(59981)', 59971);
  Test('PrevPrime(59980)', 59971);

{$IFNDEF QuickTest}
  Test('PrevPrime(8441167879117811)', 8441167879117751);
  Test('PrevPrime(8441167879117811121)', failure, 'Cannot test integers this large for primality.');
{$ENDIF}

  Test('PrevPrime(84411678791)', 84411678787);
  Test('PrevPrime(84411678787)', 84411678737);

  Test('NextPrime(3.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('NextPrime(2, 3)', failure, 'Too many arguments.');
  Test('NextPrime()', failure, 'A required argument of type integer is missing.');

  Test('PrevPrime(3.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('PrevPrime(2, 3)', failure, 'Too many arguments.');
  Test('PrevPrime()', failure, 'A required argument of type integer is missing.');

{$IFNDEF QuickTest}
  Test('NextPrime(574298344)', 574298353);
  Test('NextPrime(574298353)', 574298359);
  Test('PrevPrime(574298357)', 574298353);
  Test('PrevPrime(574298375)', 574298371);

  Test('IsPrime(0999999935)', False);
  Test('IsPrime(0999999936)', False);
  Test('IsPrime(0999999937)', True);
  Test('IsPrime(0999999938)', False);
  Test('IsPrime(0999999939)', False);
  Test('IsPrime(0999999995)', False);
  Test('IsPrime(0999999996)', False);
  Test('IsPrime(0999999997)', False);
  Test('IsPrime(0999999998)', False);
  Test('IsPrime(0999999999)', False);
  Test('IsPrime(1000000000)', False);

  Test('IsPrime(25324152301)', False);
  Test('IsPrime(25324152302)', False);
  Test('IsPrime(25324152303)', False);
  Test('IsPrime(25324152304)', False);
  Test('IsPrime(25324152305)', False);
  Test('IsPrime(25324152306)', False);
  Test('IsPrime(25324152307)', True);
  Test('IsPrime(25324152308)', False);
  Test('IsPrime(25324152309)', False);
  Test('IsPrime(25324152310)', False);
  Test('IsPrime(25324152311)', False);
  Test('IsPrime(25324152312)', False);

  Test('IsPrime(6523847052383)', False);
  Test('IsPrime(6523847052384)', False);
  Test('IsPrime(6523847052385)', False);
  Test('IsPrime(6523847052386)', False);
  Test('IsPrime(6523847052387)', False);
  Test('IsPrime(6523847052388)', False);
  Test('IsPrime(6523847052389)', True);
  Test('IsPrime(6523847052390)', False);
  Test('IsPrime(6523847052391)', False);
  Test('IsPrime(6523847052392)', False);
  Test('IsPrime(6523847052393)', False);
  Test('IsPrime(6523847052394)', False);

  Test('IsPrime(695235874512176)', False);
  Test('IsPrime(695235874512177)', False);
  Test('IsPrime(695235874512178)', False);
  Test('IsPrime(695235874512179)', False);
  Test('IsPrime(695235874512180)', False);
  Test('IsPrime(695235874512181)', True);
  Test('IsPrime(695235874512182)', False);
  Test('IsPrime(695235874512183)', False);
  Test('IsPrime(695235874512184)', False);
  Test('IsPrime(695235874512185)', False);
  Test('IsPrime(695235874512186)', False);

  Test('IsPrime(12032578956417524)', False);
  Test('IsPrime(12032578956417525)', False);
  Test('IsPrime(12032578956417526)', False);
  Test('IsPrime(12032578956417527)', False);
  Test('IsPrime(12032578956417528)', False);
  Test('IsPrime(12032578956417529)', True);
  Test('IsPrime(12032578956417530)', False);
  Test('IsPrime(12032578956417531)', False);
  Test('IsPrime(12032578956417532)', False);
  Test('IsPrime(12032578956417533)', False);
  Test('IsPrime(12032578956417534)', False);

  Test('prime(50847532)', 999999893);
  Test('prime(50847533)', 999999929);
  Test('prime(50847534)', 999999937);

  Test('NextPrime(0999999935)', 0999999937);

  Test('PrevPrime(0999999932)', 0999999929);
  Test('PrevPrime(0999999937)', 0999999929);
  Test('PrevPrime(0999999938)', 0999999937);
  Test('PrevPrime(0999999996)', 0999999937);
  Test('PrevPrime(1000000000)', 0999999937);
  Test('PrevPrime(1000000001)', 0999999937);

{$ENDIF}

  Test('PrimePi(-4000000000000000    )', 0);
  Test('PrimePi(-3    )', 0);
  Test('PrimePi(-2    )', 0);
  Test('PrimePi(-1    )', 0);
  Test('PrimePi( 0    )', 0);
  Test('PrimePi( 1    )', 0);
  Test('PrimePi( 2    )', 1);
  Test('PrimePi( 3    )', 2);
  Test('PrimePi( 4    )', 2);
  Test('PrimePi( 5    )', 3);
  Test('PrimePi( 6    )', 3);
  Test('PrimePi( 7    )', 4);
  Test('PrimePi( 8    )', 4);
  Test('PrimePi( 9    )', 4);
  Test('PrimePi( 10   )', 4);
  Test('PrimePi( 11   )', 5);
  Test('PrimePi( 12   )', 5);
  Test('PrimePi( 13   )', 6);
  Test('PrimePi( 14   )', 6);
  Test('PrimePi( 15   )', 6);
  Test('PrimePi( 16   )', 6);
  Test('PrimePi( 17   )', 7);
  Test('PrimePi( 18   )', 7);
  Test('PrimePi( 19   )', 8);
  Test('PrimePi( 20   )', 8);
  Test('PrimePi( 21   )', 8);
  Test('PrimePi( 22   )', 8);
  Test('PrimePi( 23   )', 9);
  Test('PrimePi( 24   )', 9);
  Test('PrimePi( 25   )', 9);
  Test('PrimePi( 26   )', 9);
  Test('PrimePi( 27   )', 9);
  Test('PrimePi( 28   )', 9);
  Test('PrimePi( 29   )', 10);
  Test('PrimePi( 30   )', 10);
  Test('PrimePi( 31   )', 11);
  Test('PrimePi( 32   )', 11);
  Test('PrimePi( 33   )', 11);
  Test('PrimePi( 34   )', 11);
  Test('PrimePi( 35   )', 11);
  Test('PrimePi( 36   )', 11);
  Test('PrimePi( 37   )', 12);
  Test('PrimePi( 38   )', 12);
  Test('PrimePi( 39   )', 12);
  Test('PrimePi( 40   )', 12);
  Test('PrimePi( 41   )', 13);
  Test('PrimePi( 42   )', 13);
  Test('PrimePi( 43   )', 14);
  Test('PrimePi( 44   )', 14);
  Test('PrimePi( 45   )', 14);
  Test('PrimePi( 46   )', 14);
  Test('PrimePi( 47   )', 15);
  Test('PrimePi( 48   )', 15);
  Test('PrimePi( 49   )', 15);
  Test('PrimePi( 50   )', 15);
  Test('PrimePi( 51   )', 15);
  Test('PrimePi( 52   )', 15);
  Test('PrimePi( 53   )', 16);
  Test('PrimePi( 54   )', 16);
  Test('PrimePi( 55   )', 16);
  Test('PrimePi( 56   )', 16);
  Test('PrimePi( 57   )', 16);
  Test('PrimePi( 58   )', 16);
  Test('PrimePi( 59   )', 17);
  Test('PrimePi( 60   )', 17);
  Test('PrimePi( 61   )', 18);
  Test('PrimePi( 62   )', 18);
  Test('PrimePi( 63   )', 18);
  Test('PrimePi( 64   )', 18);
  Test('PrimePi( 65   )', 18);
  Test('PrimePi( 66   )', 18);
  Test('PrimePi( 67   )', 19);
  Test('PrimePi( 68   )', 19);
  Test('PrimePi( 69   )', 19);
  Test('PrimePi( 70   )', 19);
  Test('PrimePi( 71   )', 20);
  Test('PrimePi( 72   )', 20);
  Test('PrimePi( 73   )', 21);
  Test('PrimePi( 74   )', 21);
  Test('PrimePi( 75   )', 21);
  Test('PrimePi( 76   )', 21);
  Test('PrimePi( 77   )', 21);
  Test('PrimePi( 78   )', 21);
  Test('PrimePi( 79   )', 22);
  Test('PrimePi( 80   )', 22);
  Test('PrimePi( 81   )', 22);
  Test('PrimePi( 82   )', 22);
  Test('PrimePi( 83   )', 23);
  Test('PrimePi( 84   )', 23);
  Test('PrimePi( 85   )', 23);
  Test('PrimePi( 86   )', 23);
  Test('PrimePi( 87   )', 23);
  Test('PrimePi( 88   )', 23);
  Test('PrimePi( 89   )', 24);
  Test('PrimePi( 90   )', 24);
  Test('PrimePi( 91   )', 24);
  Test('PrimePi( 92   )', 24);
  Test('PrimePi( 93   )', 24);
  Test('PrimePi( 94   )', 24);
  Test('PrimePi( 95   )', 24);
  Test('PrimePi( 96   )', 24);
  Test('PrimePi( 97   )', 25);
  Test('PrimePi( 98   )', 25);
  Test('PrimePi( 99   )', 25);
  Test('PrimePi( 100  )', 25);
  Test('PrimePi( 101  )', 26);
  Test('PrimePi( 102  )', 26);
  Test('PrimePi( 103  )', 27);
  Test('PrimePi( 104  )', 27);
  Test('PrimePi( 105  )', 27);
  Test('PrimePi( 106  )', 27);
  Test('PrimePi( 107  )', 28);
  Test('PrimePi( 108  )', 28);
  Test('PrimePi( 109  )', 29);
  Test('PrimePi( 110  )', 29);
  Test('PrimePi( 111  )', 29);
  Test('PrimePi( 112  )', 29);
  Test('PrimePi( 113  )', 30);
  Test('PrimePi( 114  )', 30);
  Test('PrimePi( 115  )', 30);
  Test('PrimePi( 116  )', 30);
  Test('PrimePi( 117  )', 30);
  Test('PrimePi( 118  )', 30);
  Test('PrimePi( 119  )', 30);
  Test('PrimePi( 120  )', 30);
  Test('PrimePi( 121  )', 30);
  Test('PrimePi( 122  )', 30);
  Test('PrimePi( 123  )', 30);
  Test('PrimePi( 124  )', 30);
  Test('PrimePi( 125  )', 30);
  Test('PrimePi( 126  )', 30);
  Test('PrimePi( 127  )', 31);
  Test('PrimePi( 128  )', 31);
  Test('PrimePi( 129  )', 31);
  Test('PrimePi( 130  )', 31);
  Test('PrimePi( 131  )', 32);
  Test('PrimePi( 132  )', 32);
  Test('PrimePi( 133  )', 32);
  Test('PrimePi( 134  )', 32);
  Test('PrimePi( 135  )', 32);
  Test('PrimePi( 136  )', 32);
  Test('PrimePi( 137  )', 33);
  Test('PrimePi( 138  )', 33);
  Test('PrimePi( 139  )', 34);
  Test('PrimePi( 140  )', 34);
  Test('PrimePi( 141  )', 34);
  Test('PrimePi( 142  )', 34);
  Test('PrimePi( 143  )', 34);
  Test('PrimePi( 144  )', 34);
  Test('PrimePi( 145  )', 34);
  Test('PrimePi( 146  )', 34);
  Test('PrimePi( 147  )', 34);
  Test('PrimePi( 148  )', 34);
  Test('PrimePi( 149  )', 35);
  Test('PrimePi( 150  )', 35);
  Test('PrimePi( 151  )', 36);
  Test('PrimePi( 152  )', 36);
  Test('PrimePi( 153  )', 36);
  Test('PrimePi( 154  )', 36);
  Test('PrimePi( 155  )', 36);
  Test('PrimePi( 156  )', 36);
  Test('PrimePi( 157  )', 37);
  Test('PrimePi( 158  )', 37);
  Test('PrimePi( 159  )', 37);
  Test('PrimePi( 160  )', 37);
  Test('PrimePi( 161  )', 37);
  Test('PrimePi( 162  )', 37);
  Test('PrimePi( 163  )', 38);
  Test('PrimePi( 164  )', 38);
  Test('PrimePi( 165  )', 38);
  Test('PrimePi( 166  )', 38);
  Test('PrimePi( 167  )', 39);
  Test('PrimePi( 168  )', 39);
  Test('PrimePi( 169  )', 39);
  Test('PrimePi( 170  )', 39);
  Test('PrimePi( 171  )', 39);
  Test('PrimePi( 172  )', 39);
  Test('PrimePi( 173  )', 40);
  Test('PrimePi( 174  )', 40);
  Test('PrimePi( 175  )', 40);
  Test('PrimePi( 176  )', 40);
  Test('PrimePi( 177  )', 40);
  Test('PrimePi( 178  )', 40);
  Test('PrimePi( 179  )', 41);
  Test('PrimePi( 180  )', 41);
  Test('PrimePi( 181  )', 42);
  Test('PrimePi( 182  )', 42);
  Test('PrimePi( 183  )', 42);
  Test('PrimePi( 184  )', 42);
  Test('PrimePi( 185  )', 42);
  Test('PrimePi( 186  )', 42);
  Test('PrimePi( 187  )', 42);
  Test('PrimePi( 188  )', 42);
  Test('PrimePi( 189  )', 42);
  Test('PrimePi( 190  )', 42);
  Test('PrimePi( 191  )', 43);
  Test('PrimePi( 192  )', 43);
  Test('PrimePi( 193  )', 44);
  Test('PrimePi( 194  )', 44);
  Test('PrimePi( 195  )', 44);
  Test('PrimePi( 196  )', 44);
  Test('PrimePi( 197  )', 45);
  Test('PrimePi( 198  )', 45);
  Test('PrimePi( 199  )', 46);
  Test('PrimePi( 200  )', 46);
  Test('PrimePi( 201  )', 46);
  Test('PrimePi( 202  )', 46);
  Test('PrimePi( 203  )', 46);
  Test('PrimePi( 204  )', 46);
  Test('PrimePi( 205  )', 46);
  Test('PrimePi( 206  )', 46);
  Test('PrimePi( 207  )', 46);
  Test('PrimePi( 208  )', 46);
  Test('PrimePi( 209  )', 46);
  Test('PrimePi( 210  )', 46);
  Test('PrimePi( 211  )', 47);
  Test('PrimePi( 212  )', 47);
  Test('PrimePi( 213  )', 47);
  Test('PrimePi( 214  )', 47);
  Test('PrimePi( 215  )', 47);
  Test('PrimePi( 216  )', 47);
  Test('PrimePi( 217  )', 47);
  Test('PrimePi( 218  )', 47);
  Test('PrimePi( 219  )', 47);
  Test('PrimePi( 220  )', 47);
  Test('PrimePi( 221  )', 47);
  Test('PrimePi( 222  )', 47);
  Test('PrimePi( 223  )', 48);
  Test('PrimePi( 224  )', 48);
  Test('PrimePi( 225  )', 48);
  Test('PrimePi( 226  )', 48);
  Test('PrimePi( 227  )', 49);
  Test('PrimePi( 228  )', 49);
  Test('PrimePi( 229  )', 50);
  Test('PrimePi( 230  )', 50);
  Test('PrimePi( 231  )', 50);
  Test('PrimePi( 232  )', 50);
  Test('PrimePi( 233  )', 51);
  Test('PrimePi( 234  )', 51);
  Test('PrimePi( 235  )', 51);
  Test('PrimePi( 236  )', 51);
  Test('PrimePi( 237  )', 51);
  Test('PrimePi( 238  )', 51);
  Test('PrimePi( 239  )', 52);
  Test('PrimePi( 240  )', 52);
  Test('PrimePi( 241  )', 53);
  Test('PrimePi( 242  )', 53);
  Test('PrimePi( 243  )', 53);
  Test('PrimePi( 244  )', 53);
  Test('PrimePi( 245  )', 53);
  Test('PrimePi( 246  )', 53);
  Test('PrimePi( 247  )', 53);
  Test('PrimePi( 248  )', 53);
  Test('PrimePi( 249  )', 53);
  Test('PrimePi( 250  )', 53);
  Test('PrimePi( 251  )', 54);
  Test('PrimePi( 252  )', 54);
  Test('PrimePi( 253  )', 54);
  Test('PrimePi( 254  )', 54);
  Test('PrimePi( 255  )', 54);
  Test('PrimePi( 256  )', 54);
  Test('PrimePi( 257  )', 55);
  Test('PrimePi( 258  )', 55);
  Test('PrimePi( 259  )', 55);
  Test('PrimePi( 260  )', 55);
  Test('PrimePi( 261  )', 55);
  Test('PrimePi( 262  )', 55);
  Test('PrimePi( 263  )', 56);
  Test('PrimePi( 264  )', 56);
  Test('PrimePi( 265  )', 56);
  Test('PrimePi( 266  )', 56);
  Test('PrimePi( 267  )', 56);
  Test('PrimePi( 268  )', 56);
  Test('PrimePi( 269  )', 57);
  Test('PrimePi( 270  )', 57);
  Test('PrimePi( 271  )', 58);
  Test('PrimePi( 272  )', 58);
  Test('PrimePi( 273  )', 58);
  Test('PrimePi( 274  )', 58);
  Test('PrimePi( 275  )', 58);
  Test('PrimePi( 276  )', 58);
  Test('PrimePi( 277  )', 59);
  Test('PrimePi( 278  )', 59);
  Test('PrimePi( 279  )', 59);
  Test('PrimePi( 280  )', 59);
  Test('PrimePi( 281  )', 60);
  Test('PrimePi( 282  )', 60);
  Test('PrimePi( 283  )', 61);
  Test('PrimePi( 284  )', 61);
  Test('PrimePi( 285  )', 61);
  Test('PrimePi( 286  )', 61);
  Test('PrimePi( 287  )', 61);
  Test('PrimePi( 288  )', 61);
  Test('PrimePi( 289  )', 61);
  Test('PrimePi( 290  )', 61);
  Test('PrimePi( 291  )', 61);
  Test('PrimePi( 292  )', 61);
  Test('PrimePi( 293  )', 62);
  Test('PrimePi( 294  )', 62);
  Test('PrimePi( 295  )', 62);
  Test('PrimePi( 296  )', 62);
  Test('PrimePi( 297  )', 62);
  Test('PrimePi( 298  )', 62);
  Test('PrimePi( 299  )', 62);
  Test('PrimePi( 300  )', 62);
  Test('PrimePi( 301  )', 62);
  Test('PrimePi( 302  )', 62);
  Test('PrimePi( 303  )', 62);
  Test('PrimePi( 304  )', 62);
  Test('PrimePi( 305  )', 62);
  Test('PrimePi( 306  )', 62);
  Test('PrimePi( 307  )', 63);
  Test('PrimePi( 308  )', 63);
  Test('PrimePi( 309  )', 63);
  Test('PrimePi( 310  )', 63);
  Test('PrimePi( 311  )', 64);
  Test('PrimePi( 312  )', 64);
  Test('PrimePi( 313  )', 65);
  Test('PrimePi( 314  )', 65);
  Test('PrimePi( 315  )', 65);
  Test('PrimePi( 316  )', 65);
  Test('PrimePi( 317  )', 66);
  Test('PrimePi( 318  )', 66);
  Test('PrimePi( 319  )', 66);
  Test('PrimePi( 320  )', 66);
  Test('PrimePi( 321  )', 66);
  Test('PrimePi( 322  )', 66);
  Test('PrimePi( 323  )', 66);
  Test('PrimePi( 324  )', 66);
  Test('PrimePi( 325  )', 66);
  Test('PrimePi( 326  )', 66);
  Test('PrimePi( 327  )', 66);
  Test('PrimePi( 328  )', 66);
  Test('PrimePi( 329  )', 66);
  Test('PrimePi( 330  )', 66);
  Test('PrimePi( 331  )', 67);
  Test('PrimePi( 332  )', 67);
  Test('PrimePi( 333  )', 67);
  Test('PrimePi( 334  )', 67);
  Test('PrimePi( 335  )', 67);
  Test('PrimePi( 336  )', 67);
  Test('PrimePi( 337  )', 68);
  Test('PrimePi( 338  )', 68);
  Test('PrimePi( 339  )', 68);
  Test('PrimePi( 340  )', 68);
  Test('PrimePi( 341  )', 68);
  Test('PrimePi( 342  )', 68);
  Test('PrimePi( 343  )', 68);
  Test('PrimePi( 344  )', 68);
  Test('PrimePi( 345  )', 68);
  Test('PrimePi( 346  )', 68);
  Test('PrimePi( 347  )', 69);
  Test('PrimePi( 348  )', 69);
  Test('PrimePi( 349  )', 70);
  Test('PrimePi( 350  )', 70);
  Test('PrimePi( 351  )', 70);
  Test('PrimePi( 352  )', 70);
  Test('PrimePi( 353  )', 71);
  Test('PrimePi( 354  )', 71);
  Test('PrimePi( 355  )', 71);
  Test('PrimePi( 356  )', 71);
  Test('PrimePi( 357  )', 71);
  Test('PrimePi( 358  )', 71);
  Test('PrimePi( 359  )', 72);
  Test('PrimePi( 360  )', 72);
  Test('PrimePi( 361  )', 72);
  Test('PrimePi( 362  )', 72);
  Test('PrimePi( 363  )', 72);
  Test('PrimePi( 364  )', 72);
  Test('PrimePi( 365  )', 72);
  Test('PrimePi( 366  )', 72);
  Test('PrimePi( 367  )', 73);
  Test('PrimePi( 368  )', 73);
  Test('PrimePi( 369  )', 73);
  Test('PrimePi( 370  )', 73);
  Test('PrimePi( 371  )', 73);
  Test('PrimePi( 372  )', 73);
  Test('PrimePi( 373  )', 74);
  Test('PrimePi( 374  )', 74);
  Test('PrimePi( 375  )', 74);
  Test('PrimePi( 376  )', 74);
  Test('PrimePi( 377  )', 74);
  Test('PrimePi( 378  )', 74);
  Test('PrimePi( 379  )', 75);
  Test('PrimePi( 380  )', 75);
  Test('PrimePi( 381  )', 75);
  Test('PrimePi( 382  )', 75);
  Test('PrimePi( 383  )', 76);
  Test('PrimePi( 384  )', 76);
  Test('PrimePi( 385  )', 76);
  Test('PrimePi( 386  )', 76);
  Test('PrimePi( 387  )', 76);
  Test('PrimePi( 388  )', 76);
  Test('PrimePi( 389  )', 77);
  Test('PrimePi( 390  )', 77);
  Test('PrimePi( 391  )', 77);
  Test('PrimePi( 392  )', 77);
  Test('PrimePi( 393  )', 77);
  Test('PrimePi( 394  )', 77);
  Test('PrimePi( 395  )', 77);
  Test('PrimePi( 396  )', 77);
  Test('PrimePi( 397  )', 78);
  Test('PrimePi( 398  )', 78);
  Test('PrimePi( 399  )', 78);
  Test('PrimePi( 400  )', 78);
  Test('PrimePi( 401  )', 79);
  Test('PrimePi( 402  )', 79);
  Test('PrimePi( 403  )', 79);
  Test('PrimePi( 404  )', 79);
  Test('PrimePi( 405  )', 79);
  Test('PrimePi( 406  )', 79);
  Test('PrimePi( 407  )', 79);
  Test('PrimePi( 408  )', 79);
  Test('PrimePi( 409  )', 80);
  Test('PrimePi( 410  )', 80);
  Test('PrimePi( 411  )', 80);
  Test('PrimePi( 412  )', 80);
  Test('PrimePi( 413  )', 80);
  Test('PrimePi( 414  )', 80);
  Test('PrimePi( 415  )', 80);
  Test('PrimePi( 416  )', 80);
  Test('PrimePi( 417  )', 80);
  Test('PrimePi( 418  )', 80);
  Test('PrimePi( 419  )', 81);
  Test('PrimePi( 420  )', 81);
  Test('PrimePi( 421  )', 82);
  Test('PrimePi( 422  )', 82);
  Test('PrimePi( 423  )', 82);
  Test('PrimePi( 424  )', 82);
  Test('PrimePi( 425  )', 82);
  Test('PrimePi( 426  )', 82);
  Test('PrimePi( 427  )', 82);
  Test('PrimePi( 428  )', 82);
  Test('PrimePi( 429  )', 82);
  Test('PrimePi( 430  )', 82);
  Test('PrimePi( 431  )', 83);
  Test('PrimePi( 432  )', 83);
  Test('PrimePi( 433  )', 84);
  Test('PrimePi( 434  )', 84);
  Test('PrimePi( 435  )', 84);
  Test('PrimePi( 436  )', 84);
  Test('PrimePi( 437  )', 84);
  Test('PrimePi( 438  )', 84);
  Test('PrimePi( 439  )', 85);
  Test('PrimePi( 440  )', 85);
  Test('PrimePi( 441  )', 85);
  Test('PrimePi( 442  )', 85);
  Test('PrimePi( 443  )', 86);
  Test('PrimePi( 444  )', 86);
  Test('PrimePi( 445  )', 86);
  Test('PrimePi( 446  )', 86);
  Test('PrimePi( 447  )', 86);
  Test('PrimePi( 448  )', 86);
  Test('PrimePi( 449  )', 87);
  Test('PrimePi( 450  )', 87);
  Test('PrimePi( 451  )', 87);
  Test('PrimePi( 452  )', 87);
  Test('PrimePi( 453  )', 87);
  Test('PrimePi( 454  )', 87);
  Test('PrimePi( 455  )', 87);
  Test('PrimePi( 456  )', 87);
  Test('PrimePi( 457  )', 88);
  Test('PrimePi( 458  )', 88);
  Test('PrimePi( 459  )', 88);
  Test('PrimePi( 460  )', 88);
  Test('PrimePi( 461  )', 89);
  Test('PrimePi( 462  )', 89);
  Test('PrimePi( 463  )', 90);
  Test('PrimePi( 464  )', 90);
  Test('PrimePi( 465  )', 90);
  Test('PrimePi( 466  )', 90);
  Test('PrimePi( 467  )', 91);
  Test('PrimePi( 468  )', 91);
  Test('PrimePi( 469  )', 91);
  Test('PrimePi( 470  )', 91);
  Test('PrimePi( 471  )', 91);
  Test('PrimePi( 472  )', 91);
  Test('PrimePi( 473  )', 91);
  Test('PrimePi( 474  )', 91);
  Test('PrimePi( 475  )', 91);
  Test('PrimePi( 476  )', 91);
  Test('PrimePi( 477  )', 91);
  Test('PrimePi( 478  )', 91);
  Test('PrimePi( 479  )', 92);
  Test('PrimePi( 480  )', 92);
  Test('PrimePi( 481  )', 92);
  Test('PrimePi( 482  )', 92);
  Test('PrimePi( 483  )', 92);
  Test('PrimePi( 484  )', 92);
  Test('PrimePi( 485  )', 92);
  Test('PrimePi( 486  )', 92);
  Test('PrimePi( 487  )', 93);
  Test('PrimePi( 488  )', 93);
  Test('PrimePi( 489  )', 93);
  Test('PrimePi( 490  )', 93);
  Test('PrimePi( 491  )', 94);
  Test('PrimePi( 492  )', 94);
  Test('PrimePi( 493  )', 94);
  Test('PrimePi( 494  )', 94);
  Test('PrimePi( 495  )', 94);
  Test('PrimePi( 496  )', 94);
  Test('PrimePi( 497  )', 94);
  Test('PrimePi( 498  )', 94);
  Test('PrimePi( 499  )', 95);
  Test('PrimePi( 500  )', 95);

  Test('PrimePi( 1000 )', 168);
  Test('PrimePi( 10000 )', 1229);
  Test('PrimePi( 100000 )', 9592);
  Test('PrimePi( 1000000 )', 78498);
  Test('PrimePi( 10000000 )', 664579);
{$IFNDEF QuickTest}
  Test('PrimePi( 100000000 )', 5761455);
  Test('PrimePi( 1000000000 )', 50847534);
  Test('PrimePi( 10000000000 )', failure, 'Cannot test integers this large for primality.');
  Test('PrimePi( 100000000000 )', failure, 'Cannot test integers this large for primality.');
  Test('PrimePi( 1000000000000 )', failure, 'Cannot test integers this large for primality.');
  Test('PrimePi( 10000000000000 )', failure, 'Cannot test integers this large for primality.');
  Test('PrimePi( 100000000000000 )', failure, 'Cannot test integers this large for primality.');
{$ENDIF}

  Test('PrimePi(3.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('PrimePi(2, 3)', failure, 'Too many arguments.');
  Test('PrimePi()', failure, 'A required argument of type integer is missing.');

  Test('IsPrime(6897311)', False);
  Test('IsPrime(5896419)', False);
  Test('IsPrime(4528917)', False);
  Test('IsPrime(1852641)', False);
  Test('IsPrime(458233)', False);
  Test('IsPrime(917637)', False);
  Test('IsPrime(527911)', False);
  Test('IsPrime(5978327)', False);
  Test('IsPrime(4289711)', False);
  Test('IsPrime(689713)', True);
  Test('IsPrime(485279)', False);
  Test('NextPrime(485279)', 485311);
  Test('PrevPrime(485279)', 485263);
  Test('prime(123456)', 1632899);

  // Fibonacci numbers

  Test('Fibonacci(0)', 0);
  Test('Fibonacci(1)', 1);
  Test('Fibonacci(2)', 1);
  Test('Fibonacci(3)', 2);
  Test('Fibonacci(4)', 3);
  Test('Fibonacci(5)', 5);
  Test('Fibonacci(6)', 8);
  Test('Fibonacci(7)', 13);
  Test('Fibonacci(8)', 21);
  Test('Fibonacci(9)', 34);
  Test('Fibonacci(10)', 55);
  Test('Fibonacci(11)', 89);
  Test('Fibonacci(12)', 144);

  Test('Fibonacci(90)', 2880067194370816120);
  Test('Fibonacci(91)', 4660046610375530309);
  Test('Fibonacci(92)', 7540113804746346429);
  Eps; Test('Fibonacci(93)', 12200160415121876738.0);
  Eps; Test('Fibonacci(94)', 19740274219868223167.0);

  Eps; Test('Fibonacci(100)', 354224848179261915075.0);
  Eps; Test('Fibonacci(500)', 1.3942322456169788013972438287040728395007025658769730E104);
  Eps; Test('Fibonacci(1000)', 4.3466557686937456435688527675040625802564660517371780E208);
  Eps; Test('Fibonacci(2000)', 4.2246963333923048787067256023414827825798528402506810E417);
  Eps; Test('Fibonacci(5000)', 3.878968454388325633701916308325905312082127714646245E1044);

  for _i := 2 to 1000 do
    Test('Fibonacci(' + _i.ToString + ') = Fibonacci(' + _i.ToString + ' - 1) + Fibonacci(' + _i.ToString + ' - 2)', True);

  Test('Fibonacci(-1)', failure, 'To obtain the Nth Fibonacci number, N must be zero or a positive integer.');
  Test('Fibonacci(-100)', failure, 'To obtain the Nth Fibonacci number, N must be zero or a positive integer.');

  Test('Fibonacci()', failure, 'A required argument of type integer is missing.');
  Test('Fibonacci(1, 2)', failure, 'Too many arguments.');
  Test('Fibonacci(π)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  // Lucas numbers

  Test('Lucas(0)', 2);
  Test('Lucas(1)', 1);
  Test('Lucas(2)', 3);
  Test('Lucas(3)', 4);
  Test('Lucas(4)', 7);
  Test('Lucas(5)', 11);
  Test('Lucas(6)', 18);
  Test('Lucas(7)', 29);
  Test('Lucas(8)', 47);
  Test('Lucas(9)', 76);
  Test('Lucas(10)', 123);
  Test('Lucas(11)', 199);
  Test('Lucas(12)', 322);

  Test('Lucas(89)', 3980154972736918051);
  Test('Lucas(90)', 6440026026380244498);
  Eps; Test('Lucas(91)', 10420180999117162549.0);
  Eps; Test('Lucas(92)', 16860207025497407047.0);

  Eps; Test('Lucas(100)', 7.92070839848372253127E20);
  Eps; Test('Lucas(500)', 3.1175980776217478160530100720173686014195239323981907E104);
  Eps; Test('Lucas(1000)', 9.7194177735908175207981982079326473737797879155345685E208);
  Eps; Test('Lucas(2000)', 9.4467081857593084153840674959996774315309632184803680E417);
  Eps; Test('Lucas(5000)', 8.673637146589588538368589908373462798874929690826669E1044);

  for _i := 2 to 1000 do
    Test('Lucas(' + _i.ToString + ') = Lucas(' + _i.ToString + ' - 1) + Lucas(' + _i.ToString + ' - 2)', True);

  Test('Lucas(-1)', failure, 'To obtain the Nth Lucas number, N must be zero or a positive integer.');
  Test('Lucas(-100)', failure, 'To obtain the Nth Lucas number, N must be zero or a positive integer.');

  Test('Lucas()', failure, 'A required argument of type integer is missing.');
  Test('Lucas(1, 2)', failure, 'Too many arguments.');
  Test('Lucas(π)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  // Möbius (μ) function

  Test('MöbiusMu( 1    )', 1);
  Test('MöbiusMu( 2    )', -1);
  Test('MöbiusMu( 3    )', -1);
  Test('MöbiusMu( 4    )', 0);
  Test('MöbiusMu( 5    )', -1);
  Test('MöbiusMu( 6    )', 1);
  Test('MöbiusMu( 7    )', -1);
  Test('MöbiusMu( 8    )', 0);
  Test('MöbiusMu( 9    )', 0);
  Test('MöbiusMu( 10   )', 1);
  Test('MöbiusMu( 11   )', -1);
  Test('MöbiusMu( 12   )', 0);
  Test('MöbiusMu( 13   )', -1);
  Test('MöbiusMu( 14   )', 1);
  Test('MöbiusMu( 15   )', 1);
  Test('MöbiusMu( 16   )', 0);
  Test('MöbiusMu( 17   )', -1);
  Test('MöbiusMu( 18   )', 0);
  Test('MöbiusMu( 19   )', -1);
  Test('MöbiusMu( 20   )', 0);
  Test('MöbiusMu( 21   )', 1);
  Test('MöbiusMu( 22   )', 1);
  Test('MöbiusMu( 23   )', -1);
  Test('MöbiusMu( 24   )', 0);
  Test('MöbiusMu( 25   )', 0);
  Test('MöbiusMu( 26   )', 1);
  Test('MöbiusMu( 27   )', 0);
  Test('MöbiusMu( 28   )', 0);
  Test('MöbiusMu( 29   )', -1);
  Test('MöbiusMu( 30   )', -1);
  Test('MöbiusMu( 31   )', -1);
  Test('MöbiusMu( 32   )', 0);
  Test('MöbiusMu( 33   )', 1);
  Test('MöbiusMu( 34   )', 1);
  Test('MöbiusMu( 35   )', 1);
  Test('MöbiusMu( 36   )', 0);
  Test('MöbiusMu( 37   )', -1);
  Test('MöbiusMu( 38   )', 1);
  Test('MöbiusMu( 39   )', 1);
  Test('MöbiusMu( 40   )', 0);
  Test('MöbiusMu( 41   )', -1);
  Test('MöbiusMu( 42   )', -1);
  Test('MöbiusMu( 43   )', -1);
  Test('MöbiusMu( 44   )', 0);
  Test('MöbiusMu( 45   )', 0);
  Test('MöbiusMu( 46   )', 1);
  Test('MöbiusMu( 47   )', -1);
  Test('MöbiusMu( 48   )', 0);
  Test('MöbiusMu( 49   )', 0);
  Test('MöbiusMu( 50   )', 0);
  Test('MöbiusMu( 51   )', 1);
  Test('MöbiusMu( 52   )', 0);
  Test('MöbiusMu( 53   )', -1);
  Test('MöbiusMu( 54   )', 0);
  Test('MöbiusMu( 55   )', 1);
  Test('MöbiusMu( 56   )', 0);
  Test('MöbiusMu( 57   )', 1);
  Test('MöbiusMu( 58   )', 1);
  Test('MöbiusMu( 59   )', -1);
  Test('MöbiusMu( 60   )', 0);
  Test('MöbiusMu( 61   )', -1);
  Test('MöbiusMu( 62   )', 1);
  Test('MöbiusMu( 63   )', 0);
  Test('MöbiusMu( 64   )', 0);
  Test('MöbiusMu( 65   )', 1);
  Test('MöbiusMu( 66   )', -1);
  Test('MöbiusMu( 67   )', -1);
  Test('MöbiusMu( 68   )', 0);
  Test('MöbiusMu( 69   )', 1);
  Test('MöbiusMu( 70   )', -1);
  Test('MöbiusMu( 71   )', -1);
  Test('MöbiusMu( 72   )', 0);
  Test('MöbiusMu( 73   )', -1);
  Test('MöbiusMu( 74   )', 1);
  Test('MöbiusMu( 75   )', 0);
  Test('MöbiusMu( 76   )', 0);
  Test('MöbiusMu( 77   )', 1);
  Test('MöbiusMu( 78   )', -1);
  Test('MöbiusMu( 79   )', -1);
  Test('MöbiusMu( 80   )', 0);
  Test('MöbiusMu( 81   )', 0);
  Test('MöbiusMu( 82   )', 1);
  Test('MöbiusMu( 83   )', -1);
  Test('MöbiusMu( 84   )', 0);
  Test('MöbiusMu( 85   )', 1);
  Test('MöbiusMu( 86   )', 1);
  Test('MöbiusMu( 87   )', 1);
  Test('MöbiusMu( 88   )', 0);
  Test('MöbiusMu( 89   )', -1);
  Test('MöbiusMu( 90   )', 0);
  Test('MöbiusMu( 91   )', 1);
  Test('MöbiusMu( 92   )', 0);
  Test('MöbiusMu( 93   )', 1);
  Test('MöbiusMu( 94   )', 1);
  Test('MöbiusMu( 95   )', 1);
  Test('MöbiusMu( 96   )', 0);
  Test('MöbiusMu( 97   )', -1);
  Test('MöbiusMu( 98   )', 0);
  Test('MöbiusMu( 99   )', 0);
  Test('MöbiusMu( 100  )', 0);
  Test('MöbiusMu( 101  )', -1);
  Test('MöbiusMu( 102  )', -1);
  Test('MöbiusMu( 103  )', -1);
  Test('MöbiusMu( 104  )', 0);
  Test('MöbiusMu( 105  )', -1);
  Test('MöbiusMu( 106  )', 1);
  Test('MöbiusMu( 107  )', -1);
  Test('MöbiusMu( 108  )', 0);
  Test('MöbiusMu( 109  )', -1);
  Test('MöbiusMu( 110  )', -1);
  Test('MöbiusMu( 111  )', 1);
  Test('MöbiusMu( 112  )', 0);
  Test('MöbiusMu( 113  )', -1);
  Test('MöbiusMu( 114  )', -1);
  Test('MöbiusMu( 115  )', 1);
  Test('MöbiusMu( 116  )', 0);
  Test('MöbiusMu( 117  )', 0);
  Test('MöbiusMu( 118  )', 1);
  Test('MöbiusMu( 119  )', 1);
  Test('MöbiusMu( 120  )', 0);
  Test('MöbiusMu( 121  )', 0);
  Test('MöbiusMu( 122  )', 1);
  Test('MöbiusMu( 123  )', 1);
  Test('MöbiusMu( 124  )', 0);
  Test('MöbiusMu( 125  )', 0);
  Test('MöbiusMu( 126  )', 0);
  Test('MöbiusMu( 127  )', -1);
  Test('MöbiusMu( 128  )', 0);
  Test('MöbiusMu( 129  )', 1);
  Test('MöbiusMu( 130  )', -1);
  Test('MöbiusMu( 131  )', -1);
  Test('MöbiusMu( 132  )', 0);
  Test('MöbiusMu( 133  )', 1);
  Test('MöbiusMu( 134  )', 1);
  Test('MöbiusMu( 135  )', 0);
  Test('MöbiusMu( 136  )', 0);
  Test('MöbiusMu( 137  )', -1);
  Test('MöbiusMu( 138  )', -1);
  Test('MöbiusMu( 139  )', -1);
  Test('MöbiusMu( 140  )', 0);
  Test('MöbiusMu( 141  )', 1);
  Test('MöbiusMu( 142  )', 1);
  Test('MöbiusMu( 143  )', 1);
  Test('MöbiusMu( 144  )', 0);
  Test('MöbiusMu( 145  )', 1);
  Test('MöbiusMu( 146  )', 1);
  Test('MöbiusMu( 147  )', 0);
  Test('MöbiusMu( 148  )', 0);
  Test('MöbiusMu( 149  )', -1);
  Test('MöbiusMu( 150  )', 0);
  Test('MöbiusMu( 151  )', -1);
  Test('MöbiusMu( 152  )', 0);
  Test('MöbiusMu( 153  )', 0);
  Test('MöbiusMu( 154  )', -1);
  Test('MöbiusMu( 155  )', 1);
  Test('MöbiusMu( 156  )', 0);
  Test('MöbiusMu( 157  )', -1);
  Test('MöbiusMu( 158  )', 1);
  Test('MöbiusMu( 159  )', 1);
  Test('MöbiusMu( 160  )', 0);
  Test('MöbiusMu( 161  )', 1);
  Test('MöbiusMu( 162  )', 0);
  Test('MöbiusMu( 163  )', -1);
  Test('MöbiusMu( 164  )', 0);
  Test('MöbiusMu( 165  )', -1);
  Test('MöbiusMu( 166  )', 1);
  Test('MöbiusMu( 167  )', -1);
  Test('MöbiusMu( 168  )', 0);
  Test('MöbiusMu( 169  )', 0);
  Test('MöbiusMu( 170  )', -1);
  Test('MöbiusMu( 171  )', 0);
  Test('MöbiusMu( 172  )', 0);
  Test('MöbiusMu( 173  )', -1);
  Test('MöbiusMu( 174  )', -1);
  Test('MöbiusMu( 175  )', 0);
  Test('MöbiusMu( 176  )', 0);
  Test('MöbiusMu( 177  )', 1);
  Test('MöbiusMu( 178  )', 1);
  Test('MöbiusMu( 179  )', -1);
  Test('MöbiusMu( 180  )', 0);
  Test('MöbiusMu( 181  )', -1);
  Test('MöbiusMu( 182  )', -1);
  Test('MöbiusMu( 183  )', 1);
  Test('MöbiusMu( 184  )', 0);
  Test('MöbiusMu( 185  )', 1);
  Test('MöbiusMu( 186  )', -1);
  Test('MöbiusMu( 187  )', 1);
  Test('MöbiusMu( 188  )', 0);
  Test('MöbiusMu( 189  )', 0);
  Test('MöbiusMu( 190  )', -1);
  Test('MöbiusMu( 191  )', -1);
  Test('MöbiusMu( 192  )', 0);
  Test('MöbiusMu( 193  )', -1);
  Test('MöbiusMu( 194  )', 1);
  Test('MöbiusMu( 195  )', -1);
  Test('MöbiusMu( 196  )', 0);
  Test('MöbiusMu( 197  )', -1);
  Test('MöbiusMu( 198  )', 0);
  Test('MöbiusMu( 199  )', -1);
  Test('MöbiusMu( 200  )', 0);

  Test('MöbiusMu( 1000 )', 0);
  Test('MöbiusMu( 1001 )', -1);
  Test('MöbiusMu( 1002 )', -1);
  Test('MöbiusMu( 1003 )', 1);
  Test('MöbiusMu( 1004 )', 0);

  Test('MöbiusMu( 10000)', 0);
  Test('MöbiusMu( 10001)', 1);
  Test('MöbiusMu( 10002)', -1);
  Test('MöbiusMu( 10003)', 1);
  Test('MöbiusMu( 10004)', 0);
  Test('MöbiusMu( 10005)', 1);
  Test('MöbiusMu( 10006)', 1);
  Test('MöbiusMu( 10007)', -1);
  Test('MöbiusMu( 10008)', 0);

  Test('MöbiusMu(100000)', 0);
  Test('MöbiusMu(100001)', 1);
  Test('MöbiusMu(100002)', 1);
  Test('MöbiusMu(100003)', -1);
  Test('MöbiusMu(100004)', 0);
  Test('MöbiusMu(100005)', 1);

  Test('MöbiusMu(123456)', 0);
  Test('MöbiusMu(123457)', -1);

  Test('MöbiusMu(4⋅91233)', 0);
  Test('MöbiusMu(4⋅84905)', 0);
  Test('MöbiusMu(4⋅47179)', 0);
  Test('MöbiusMu(4⋅81074)', 0);
  Test('MöbiusMu(4⋅89406)', 0);
  Test('MöbiusMu(4⋅84617)', 0);
  Test('MöbiusMu(4⋅87171)', 0);

  Test('MöbiusMu(1152921504606846976)', 0);
  Test('MöbiusMu(3743906242624487424)', 0);
  Test('MöbiusMu(1350851717672992089)', 0);
  Test('MöbiusMu(1490116119384765625)', 0);
  Test('MöbiusMu(3909821048582988049)', 0);
  Test('MöbiusMu(505447028499293771)', 0);
  Test('MöbiusMu(174887470365513049)', 0);

{$IFNDEF QuickTest}
  Test('MöbiusMu(123456789)', 0);
  Test('MöbiusMu(123476789)', 1);
  Test('MöbiusMu(321654987)', 0);
  Test('MöbiusMu(121654987)', 1);
  Test('MöbiusMu(582346917)', 0);
  Test('MöbiusMu(485230577)', -1);
  Test('MöbiusMu(690427315)', 1);
{$ENDIF}

  Test('MöbiusMu( 0    )', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('MöbiusMu( -1   )', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('MöbiusMu( -2   )', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('MöbiusMu( -3   )', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('MöbiusMu( -1350851717672992089   )', failure, 'Cannot find the prime factors of a non-positive integer.');

  Test('MöbiusMu( 1, 2 )', failure, 'Too many arguments.');
  Test('MöbiusMu()', failure, 'A required argument of type integer is missing.');
  Test('MöbiusMu(3.14)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  // Mertens function

  Test('Mertens( 1    )', 1);
  Test('Mertens( 2    )', 0);
  Test('Mertens( 3    )', -1);
  Test('Mertens( 4    )', -1);
  Test('Mertens( 5    )', -2);
  Test('Mertens( 6    )', -1);
  Test('Mertens( 7    )', -2);
  Test('Mertens( 8    )', -2);
  Test('Mertens( 9    )', -2);
  Test('Mertens( 10   )', -1);
  Test('Mertens( 11   )', -2);
  Test('Mertens( 12   )', -2);
  Test('Mertens( 13   )', -3);
  Test('Mertens( 14   )', -2);
  Test('Mertens( 15   )', -1);
  Test('Mertens( 16   )', -1);
  Test('Mertens( 17   )', -2);
  Test('Mertens( 18   )', -2);
  Test('Mertens( 19   )', -3);
  Test('Mertens( 20   )', -3);
  Test('Mertens( 21   )', -2);
  Test('Mertens( 22   )', -1);
  Test('Mertens( 23   )', -2);
  Test('Mertens( 24   )', -2);
  Test('Mertens( 25   )', -2);
  Test('Mertens( 26   )', -1);
  Test('Mertens( 27   )', -1);
  Test('Mertens( 28   )', -1);
  Test('Mertens( 29   )', -2);
  Test('Mertens( 30   )', -3);
  Test('Mertens( 31   )', -4);
  Test('Mertens( 32   )', -4);
  Test('Mertens( 33   )', -3);
  Test('Mertens( 34   )', -2);
  Test('Mertens( 35   )', -1);
  Test('Mertens( 36   )', -1);
  Test('Mertens( 37   )', -2);
  Test('Mertens( 38   )', -1);
  Test('Mertens( 39   )', 0);
  Test('Mertens( 40   )', 0);
  Test('Mertens( 41   )', -1);
  Test('Mertens( 42   )', -2);
  Test('Mertens( 43   )', -3);
  Test('Mertens( 44   )', -3);
  Test('Mertens( 45   )', -3);
  Test('Mertens( 46   )', -2);
  Test('Mertens( 47   )', -3);
  Test('Mertens( 48   )', -3);
  Test('Mertens( 49   )', -3);
  Test('Mertens( 50   )', -3);
  Test('Mertens( 51   )', -2);
  Test('Mertens( 52   )', -2);
  Test('Mertens( 53   )', -3);
  Test('Mertens( 54   )', -3);
  Test('Mertens( 55   )', -2);
  Test('Mertens( 56   )', -2);
  Test('Mertens( 57   )', -1);
  Test('Mertens( 58   )', 0);
  Test('Mertens( 59   )', -1);
  Test('Mertens( 60   )', -1);
  Test('Mertens( 61   )', -2);
  Test('Mertens( 62   )', -1);
  Test('Mertens( 63   )', -1);
  Test('Mertens( 64   )', -1);
  Test('Mertens( 65   )', 0);
  Test('Mertens( 66   )', -1);
  Test('Mertens( 67   )', -2);
  Test('Mertens( 68   )', -2);
  Test('Mertens( 69   )', -1);
  Test('Mertens( 70   )', -2);
  Test('Mertens( 71   )', -3);
  Test('Mertens( 72   )', -3);
  Test('Mertens( 73   )', -4);
  Test('Mertens( 74   )', -3);
  Test('Mertens( 75   )', -3);
  Test('Mertens( 76   )', -3);
  Test('Mertens( 77   )', -2);
  Test('Mertens( 78   )', -3);
  Test('Mertens( 79   )', -4);
  Test('Mertens( 80   )', -4);
  Test('Mertens( 81   )', -4);
  Test('Mertens( 82   )', -3);
  Test('Mertens( 83   )', -4);
  Test('Mertens( 84   )', -4);
  Test('Mertens( 85   )', -3);
  Test('Mertens( 86   )', -2);
  Test('Mertens( 87   )', -1);
  Test('Mertens( 88   )', -1);
  Test('Mertens( 89   )', -2);
  Test('Mertens( 90   )', -2);
  Test('Mertens( 91   )', -1);
  Test('Mertens( 92   )', -1);
  Test('Mertens( 93   )', 0);
  Test('Mertens( 94   )', 1);
  Test('Mertens( 95   )', 2);
  Test('Mertens( 96   )', 2);
  Test('Mertens( 97   )', 1);
  Test('Mertens( 98   )', 1);
  Test('Mertens( 99   )', 1);
  Test('Mertens( 100  )', 1);
  Test('Mertens( 101  )', 0);
  Test('Mertens( 102  )', -1);
  Test('Mertens( 103  )', -2);
  Test('Mertens( 104  )', -2);
  Test('Mertens( 105  )', -3);
  Test('Mertens( 106  )', -2);
  Test('Mertens( 107  )', -3);
  Test('Mertens( 108  )', -3);
  Test('Mertens( 109  )', -4);
  Test('Mertens( 110  )', -5);
  Test('Mertens( 111  )', -4);
  Test('Mertens( 112  )', -4);
  Test('Mertens( 113  )', -5);
  Test('Mertens( 114  )', -6);
  Test('Mertens( 115  )', -5);
  Test('Mertens( 116  )', -5);
  Test('Mertens( 117  )', -5);
  Test('Mertens( 118  )', -4);
  Test('Mertens( 119  )', -3);
  Test('Mertens( 120  )', -3);
  Test('Mertens( 121  )', -3);
  Test('Mertens( 122  )', -2);
  Test('Mertens( 123  )', -1);
  Test('Mertens( 124  )', -1);
  Test('Mertens( 125  )', -1);
  Test('Mertens( 126  )', -1);
  Test('Mertens( 127  )', -2);
  Test('Mertens( 128  )', -2);
  Test('Mertens( 129  )', -1);
  Test('Mertens( 130  )', -2);
  Test('Mertens( 131  )', -3);
  Test('Mertens( 132  )', -3);
  Test('Mertens( 133  )', -2);
  Test('Mertens( 134  )', -1);
  Test('Mertens( 135  )', -1);
  Test('Mertens( 136  )', -1);
  Test('Mertens( 137  )', -2);
  Test('Mertens( 138  )', -3);
  Test('Mertens( 139  )', -4);
  Test('Mertens( 140  )', -4);
  Test('Mertens( 141  )', -3);
  Test('Mertens( 142  )', -2);
  Test('Mertens( 143  )', -1);

  Test('Mertens( 1000 )', 2);
  Test('Mertens( 10000 )', -23);
  Test('Mertens( 100000 )', -48);
  Test('Mertens( 1000000 )', 212);

  Test('Mertens(0)', failure, 'Cannot compute Mertens number for a non-positive index.');
  Test('Mertens(-1)', failure, 'Cannot compute Mertens number for a non-positive index.');

  Test('Mertens( 1, 2 )', failure, 'Too many arguments.');
  Test('Mertens()', failure, 'A required argument of type integer is missing.');
  Test('Mertens(3.14)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  // Coprime test

  Test('coprime(2, 3)', True);
  Test('coprime(2, 9)', True);
  Test('coprime(5, 3)', True);
  Test('coprime(5, 11)', True);
  Test('coprime(2⋅7⋅11, 5⋅53⋅53)', True);
  Test('coprime(13⋅17, 2⋅2⋅2⋅2⋅101)', True);
  Test('coprime(53, 3)', True);
  Test('coprime(123456, 1)', True);
  Test('coprime(10000000, 3)', True);
  Test('coprime(1000000, 1)', True);
  Test('coprime(17, 19)', True);
  Test('coprime(1, 1)', True);
  Test('coprime(1, 2048)', True);

  Test('coprime(2, 2)', False);
  Test('coprime(2, 2048)', False);
  Test('coprime(1024, 32)', False);
  Test('coprime(3, 27)', False);
  Test('coprime(2⋅7⋅11, 2⋅53⋅53)', False);
  Test('coprime(2⋅7⋅11, 5⋅7⋅53⋅53)', False);
  Test('coprime(5⋅7⋅11, 5⋅7⋅53⋅53)', False);
  Test('coprime(5⋅7⋅11, 2⋅3⋅11⋅53)', False);
  Test('coprime(2⋅3⋅5⋅7⋅11, 2⋅3⋅11⋅53)', False);
  Test('coprime(119135437, 119135437)', False);

  Test('coprime(2, 0)', False);
  Test('coprime(2, 0)', False);
  Test('coprime(5, 0)', False);
  Test('coprime(5, 0)', False);
  Test('coprime(2⋅7⋅11, 0)', False);
  Test('coprime(13⋅17, 0)', False);
  Test('coprime(53, 0)', False);
  Test('coprime(123456, 0)', False);
  Test('coprime(10000000, 0)', False);
  Test('coprime(0, 10000000)', False);
  Test('coprime(1000000, 0)', False);
  Test('coprime(17, 0)', False);
  Test('coprime(1, 0)', True);
  Test('coprime(0, 1)', True);
  Test('coprime(0, 0)', False); // gcd(0, 0) = 0 <> 1

  Test('coprime(89408484632772, 1861618814930)', False);

  Test('coprime(-2, 3)', True);
  Test('coprime(-2, 9)', True);
  Test('coprime(-5, 3)', True);
  Test('coprime(-5, -11)', True);
  Test('coprime(2⋅7⋅11, -5⋅53⋅53)', True);
  Test('coprime(13⋅17, -2⋅2⋅2⋅2⋅101)', True);
  Test('coprime(53, -3)', True);
  Test('coprime(123456, -1)', True);
  Test('coprime(-10000000, -3)', True);
  Test('coprime(-1000000, 1)', True);
  Test('coprime(-17, -19)', True);
  Test('coprime(-1, -1)', True);
  Test('coprime(1, -2048)', True);

  Test('coprime(2, -2)', False);
  Test('coprime(-2, 2048)', False);
  Test('coprime(-1024, 32)', False);
  Test('coprime(-3, 27)', False);
  Test('coprime(-2⋅7⋅11, -2⋅53⋅53)', False);
  Test('coprime(-2⋅7⋅11, -5⋅7⋅53⋅53)', False);
  Test('coprime(-5⋅7⋅11, 5⋅7⋅53⋅53)', False);
  Test('coprime(-5⋅7⋅11, 2⋅3⋅11⋅53)', False);
  Test('coprime(2⋅3⋅5⋅7⋅11, -2⋅3⋅11⋅53)', False);
  Test('coprime(119135437, -119135437)', False);

  Test('coprime(-2, 0)', False);
  Test('coprime(-2, 0)', False);
  Test('coprime(-5, 0)', False);
  Test('coprime(-5, 0)', False);
  Test('coprime(-2⋅7⋅11, 0)', False);
  Test('coprime(-13⋅17, 0)', False);
  Test('coprime(-53, 0)', False);
  Test('coprime(-123456, 0)', False);
  Test('coprime(-10000000, 0)', False);
  Test('coprime(0, -10000000)', False);
  Test('coprime(-1000000, 0)', False);
  Test('coprime(-17, 0)', False);
  Test('coprime(-1, 0)', True);
  Test('coprime(0, -1)', True);

  Test('coprime(89408484632772, -1861618814930)', False);

  Test('coprime(41353513, 75642727)', True);
  Test('coprime(846060, 4894083)', False);
  Test('coprime(-41353513, -75642727)', True);
  Test('coprime(846060, -4894083)', False);

  Test('coprime(48240917733, 81981650157)', False);
  Test('coprime(48940917733, 81981650166)', True);
  Test('coprime(48240917733, -81981650157)', False);
  Test('coprime(-48940917733, -81981650166)', True);

  Test('coprime(-48940917733)', failure, 'A required argument of type integer is missing.');
  Test('coprime(-48940917733, 123, 5608)', failure, 'Too many arguments.');
  Test('coprime(-48940917733, 123.08)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');

  // The Iverson bracket

  Test('Iverson(false)', 0);
  Test('Iverson(true)', 1);

  Test('Iverson(1)', failure, 'An object of type boolean was expected as argument 1, but an object of type integer was given.');
  Test('Iverson()', failure, 'A required argument of type boolean is missing.');
  Test('Iverson(false, true)', failure, 'Too many arguments.');

  // Kronecker delta

  for _i := -10 to 10 do
    for _j := -10 to 10 do
      Test(Format('Kronecker(%d, %d)', [_i, _j]), IfThen(_i = _j, 1, 0));

  for _i := -10 to 10 do
    for _j := -10 to 10 do
      Test(Format('Kronecker(%d, %d) = Iverson(%d = %d)', [_i, _j, _i, _j]), True);

  Test('Kronecker(5)', failure, 'A required argument of type integer is missing.');
  Test('Kronecker(5, 6, 7)', failure, 'Too many arguments.');
  Test('Kronecker(5, 6.7)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');

  // Jaboci symbol

  Test('JacobiSymbol(15, 7)', 1);
  Test('JacobiSymbol(27, 29)', -1);
  Test('JacobiSymbol(4, 35)', 1);
  Test('JacobiSymbol(10, 35)', 0);
  Test('JacobiSymbol(18, 35)', -1);
  Test('JacobiSymbol(30, 35)', 0);
  Test('JacobiSymbol(1, 53)', 1);
  Test('JacobiSymbol(10, 53)', 1);
  Test('JacobiSymbol(12, 53)', -1);
  Test('JacobiSymbol(17, 53)', 1);
  Test('JacobiSymbol(25, 53)', 1);
  Test('JacobiSymbol(26, 53)', -1);

  Test('JacobiSymbol(19, 45)', 1);
  Test('JacobiSymbol(8, 21)', -1);
  Test('JacobiSymbol(5, 21)', 1);
  Test('JacobiSymbol(988, 9875)', -1);
  Test('JacobiSymbol(1001, 9907)', -1);
  Test('JacobiSymbol(1121, 10253)', -1);
  Test('JacobiSymbol(1234, 10253)', -1);
  Test('JacobiSymbol(2345, 10253)', -1);
  Test('JacobiSymbol(2345, 13511)', -1);
  Test('JacobiSymbol(2345, 13555)', 0);
  Test('JacobiSymbol(2347, 13557)', -1);
  Test('JacobiSymbol(2347, 22321)', -1);
  Test('JacobiSymbol(23471, 22321)', -1);
  Test('JacobiSymbol(13471, 12321)', 1);

  Test('JacobiSymbol(-230, 597)', -1);
  Test('JacobiSymbol(-5221, 899)', 1);

  for _i := Low(JacobiSymbolValues) to High(JacobiSymbolValues) do
    if Odd(_i) then
      for _j := Low(JacobiSymbolValues[_i]) to High(JacobiSymbolValues[_i]) do
        Test(Format('JacobiSymbol(%d, %d)', [_j, _i]), JacobiSymbolValues[_i, _j]);

  Test('JacobiSymbol(15, -2)', failure, 'Invalid modulus (not an odd positive integer) for Jacobi symbol: -2');
  Test('JacobiSymbol(15, -1)', failure, 'Invalid modulus (not an odd positive integer) for Jacobi symbol: -1');
  Test('JacobiSymbol(15, 0)', failure, 'Invalid modulus (not an odd positive integer) for Jacobi symbol: 0');
  Test('JacobiSymbol(15, 2)', failure, 'Invalid modulus (not an odd positive integer) for Jacobi symbol: 2');
  Test('JacobiSymbol(15, 4)', failure, 'Invalid modulus (not an odd positive integer) for Jacobi symbol: 4');
  Test('JacobiSymbol(15, 6)', failure, 'Invalid modulus (not an odd positive integer) for Jacobi symbol: 6');

  for _i := Low(JacobiSymbolValues) to High(JacobiSymbolValues) do
    if Odd(_i) then
      for _j := Low(JacobiSymbolValues[_i]) to High(JacobiSymbolValues[_i]) do
        if gcd(_i, _j) = 1 then
          Test(Format('abs(JacobiSymbol(%d, %d))', [_j, _i]), 1)
        else
          Test(Format('JacobiSymbol(%d, %d)', [_j, _i]), 0);

  for _i := 1 to 99 do
    if Odd(_i) then
      Test(Format('JacobiSymbol(-1, %d)', [_i]), AltSgn((_i - 1) div 2));

  for _i := 1 to 99 do
    if Odd(_i) then
      Test(Format('JacobiSymbol(2, %d)', [_i]), AltSgn((Sqr(_i) - 1) div 8));

  Test('JacobiSymbol(5)', failure, 'A required argument of type integer is missing.');
  Test('JacobiSymbol(5, 3, 1)', failure, 'Too many arguments.');
  Test('JacobiSymbol(19, 45.2)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');

  // Legendre symbol

  for _i := Low(JacobiSymbolValues) to High(JacobiSymbolValues) do
    if Odd(_i) and IsPrime(_i) then
      for _j := Low(JacobiSymbolValues[_i]) to High(JacobiSymbolValues[_i]) do
        Test(Format('LegendreSymbol(%d, %d)', [_j, _i]), JacobiSymbolValues[_i, _j]);

  for _i := Low(JacobiSymbolValues) to High(JacobiSymbolValues) do
    if Odd(_i) and IsPrime(_i) then
      for _j := Low(JacobiSymbolValues[_i]) to High(JacobiSymbolValues[_i]) do
        Test(Format('LegendreSymbol(%d + %d, %d)', [_j, _i, _i]), JacobiSymbolValues[_i, _j]);

  for _i := 3 to 100 do
    Test(Format('LegendreSymbol(3, %d)', [NthPrime(_i)]), AltSgn(Math.Floor((NthPrime(_i) + 1) / 6)));

  for _i := 4 to 100 do
    Test(Format('LegendreSymbol(5, %d)', [NthPrime(_i)]), AltSgn(Math.Floor((2*NthPrime(_i) + 2) / 5)));

  for _i := 2 to 20 do
    for _j := 2 to 20 do
      if _i <> _j then
        Test(
          Format('LegendreSymbol(%d, %d) ⋅ LegendreSymbol(%d, %d)',
            [NthPrime(_j), NthPrime(_i), NthPrime(_i), NthPrime(_j)]),
          AltSgn((NthPrime(_i) - 1)*(NthPrime(_j) - 1) div 4)
        );

  Test('LegendreSymbol(15, 7)', 1);
  Test('LegendreSymbol(27, 29)', -1);
  Test('LegendreSymbol(4, 35)', failure, 'Non-prime modulus for Legendre symbol: 35');
  Test('LegendreSymbol(4, 51)', failure, 'Non-prime modulus for Legendre symbol: 51');
  Test('LegendreSymbol(1, 53)', 1);
  Test('LegendreSymbol(10, 53)', 1);
  Test('LegendreSymbol(12, 53)', -1);
  Test('LegendreSymbol(17, 53)', 1);
  Test('LegendreSymbol(25, 53)', 1);
  Test('LegendreSymbol(26, 53)', -1);

  Test('LegendreSymbol(12345, 331)', -1);

  Test('LegendreSymbol(19, 45)', failure, 'Non-prime modulus for Legendre symbol: 45');
  Test('LegendreSymbol(8, 21)', failure, 'Non-prime modulus for Legendre symbol: 21');
  Test('LegendreSymbol(5, 21)', failure, 'Non-prime modulus for Legendre symbol: 21');
  Test('LegendreSymbol(1001, 9907)', -1);

  Test('LegendreSymbol(19, 1)', failure, 'Non-prime modulus for Legendre symbol: 1');
  Test('LegendreSymbol(19, 0)', failure, 'Non-prime modulus for Legendre symbol: 0');
  Test('LegendreSymbol(19, -1)', failure, 'Non-prime modulus for Legendre symbol: -1');
  Test('LegendreSymbol(19, -2)', failure, 'Non-prime modulus for Legendre symbol: -2');

  Test('LegendreSymbol(19, 2)', failure, 'Invalid modulus (not an odd positive integer) for Jacobi symbol: 2');

  Test('LegendreSymbol(5)', failure, 'A required argument of type integer is missing.');
  Test('LegendreSymbol(5, 3, 1)', failure, 'Too many arguments.');
  Test('LegendreSymbol(19, 45.2)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');

  // Kronecker symbol

  Test('KroneckerSymbol(4, 7)', 1);
  Test('KroneckerSymbol(4, 10)', 0);
  Test('KroneckerSymbol(4, 13)', 1);
  Test('KroneckerSymbol(4, 22)', 0);

  Test('KroneckerSymbol(7, 13)', -1);
  Test('KroneckerSymbol(7, 23)', -1);
  Test('KroneckerSymbol(7, 24)', 1);
  Test('KroneckerSymbol(7, 25)', 1);
  Test('KroneckerSymbol(7, 26)', -1);

  Test('KroneckerSymbol(19, 3)', 1);
  Test('KroneckerSymbol(19, 4)', 1);
  Test('KroneckerSymbol(19, 5)', 1);
  Test('KroneckerSymbol(19, 6)', -1);

  Test('KroneckerSymbol(20, 10)', 0);
  Test('KroneckerSymbol(20, 11)', 1);
  Test('KroneckerSymbol(20, 12)', 0);
  Test('KroneckerSymbol(20, 13)', -1);

  Test('KroneckerSymbol(2837, 2458)', -1);
  Test('KroneckerSymbol(2837, 2459)', 1);
  Test('KroneckerSymbol(2837, 2460)', 1);

  Test('KroneckerSymbol(10234, 8642)', 0);
  Test('KroneckerSymbol(10235, 8642)', 1);
  Test('KroneckerSymbol(10235, 8643)', 1);
  Test('KroneckerSymbol(10235, 8644)', -1);
  Test('KroneckerSymbol(10235, 8645)', 0);
  Test('KroneckerSymbol(10235, 8646)', -1);
  Test('KroneckerSymbol(10235, 8647)', 1);
  Test('KroneckerSymbol(10235, 8648)', 0);

  Test('KroneckerSymbol(9, 2)', 1);
  Test('KroneckerSymbol(10, 2)', 0);
  Test('KroneckerSymbol(11, 2)', -1);
  Test('KroneckerSymbol(12, 2)', 0);
  Test('KroneckerSymbol(13, 2)', -1);
  Test('KroneckerSymbol(14, 2)', 0);
  Test('KroneckerSymbol(15, 2)', 1);
  Test('KroneckerSymbol(12345, 2)', 1);
  Test('KroneckerSymbol(77777, 2)', 1);
  Test('KroneckerSymbol(77779, 2)', -1);

  Test('KroneckerSymbol(-9, 2)', 1);
  Test('KroneckerSymbol(-10, 2)', 0);
  Test('KroneckerSymbol(-11, 2)', -1);
  Test('KroneckerSymbol(-12, 2)', 0);
  Test('KroneckerSymbol(-13, 2)', -1);
  Test('KroneckerSymbol(-14, 2)', 0);
  Test('KroneckerSymbol(-15, 2)', 1);
  Test('KroneckerSymbol(-12345, 2)', 1);
  Test('KroneckerSymbol(-77777, 2)', 1);
  Test('KroneckerSymbol(-77779, 2)', -1);

  Test('KroneckerSymbol(-2317, 17)', -1);
  Test('KroneckerSymbol(-2317, 18)', -1);
  Test('KroneckerSymbol(-2317, 19)', 1);
  Test('KroneckerSymbol(-2317, 20)', -1);
  Test('KroneckerSymbol(-2317, 21)', 0);
  Test('KroneckerSymbol(-2317, 22)', -1);
  Test('KroneckerSymbol(-2317, 23)', 1);

  Test('KroneckerSymbol(-2318, 49)', 1);
  Test('KroneckerSymbol(-2318, 50)', 0);
  Test('KroneckerSymbol(-2318, 51)', -1);
  Test('KroneckerSymbol(-2318, 52)', 0);
  Test('KroneckerSymbol(-2318, 53)', -1);
  Test('KroneckerSymbol(-2318, 54)', 0);
  Test('KroneckerSymbol(-2318, 55)', -1);
  Test('KroneckerSymbol(-2318, 56)', 0);
  Test('KroneckerSymbol(-2318, 57)', 0);
  Test('KroneckerSymbol(-2318, 58)', 0);
  Test('KroneckerSymbol(-2318, 59)', -1);

  Test('KroneckerSymbol(2531, -17)', 1);
  Test('KroneckerSymbol(2531, -18)', -1);
  Test('KroneckerSymbol(2531, -19)', 1);
  Test('KroneckerSymbol(2531, -20)', 1);
  Test('KroneckerSymbol(2531, -21)', -1);
  Test('KroneckerSymbol(2531, -22)', -1);
  Test('KroneckerSymbol(2531, -23)', 1);
  Test('KroneckerSymbol(2531, -24)', 1);
  Test('KroneckerSymbol(2531, -25)', 1);
  Test('KroneckerSymbol(2531, -26)', -1);
  Test('KroneckerSymbol(2531, -27)', -1);
  Test('KroneckerSymbol(2531, -28)', 1);
  Test('KroneckerSymbol(2531, -29)', -1);

  Test('KroneckerSymbol(-5127, -3123)', 0);
  Test('KroneckerSymbol(-5127, -3124)', -1);
  Test('KroneckerSymbol(-5127, -3125)', 1);
  Test('KroneckerSymbol(-5127, -3126)', 0);
  Test('KroneckerSymbol(-5127, -3127)', -1);
  Test('KroneckerSymbol(-5127, -3128)', 1);
  Test('KroneckerSymbol(-5127, -3129)', 0);
  Test('KroneckerSymbol(-5127, -3130)', 1);
  Test('KroneckerSymbol(-5127, -3131)', -1);
  Test('KroneckerSymbol(-5127, -3132)', 0);
  Test('KroneckerSymbol(-5127, -3133)', -1);
  Test('KroneckerSymbol(-5127, -3134)', 1);

  Test('KroneckerSymbol(-28531, -207)', -1);
  Test('KroneckerSymbol(-28531, -208)', -1);
  Test('KroneckerSymbol(-28531, -209)', -1);
  Test('KroneckerSymbol(-28531, -210)', -1);
  Test('KroneckerSymbol(-28531, -211)', 1);
  Test('KroneckerSymbol(-28531, -212)', -1);
  Test('KroneckerSymbol(-28531, -213)', -1);
  Test('KroneckerSymbol(-28531, -214)', -1);
  Test('KroneckerSymbol(-28531, -215)', -1);
  Test('KroneckerSymbol(-28531, -216)', -1);
  Test('KroneckerSymbol(-28531, -217)', -1);
  Test('KroneckerSymbol(-28531, -218)', 1);
  Test('KroneckerSymbol(-28531, -219)', 1);
  Test('KroneckerSymbol(-28531, -220)', -1);
  Test('KroneckerSymbol(-28531, -221)', 1);

  Test('KroneckerSymbol(-291, -123456)', 0);
  Test('KroneckerSymbol(-291, -123457)', -1);
  Test('KroneckerSymbol(-291, -123458)', -1);
  Test('KroneckerSymbol(-291, -123459)', 0);
  Test('KroneckerSymbol(-291, -123460)', 1);
  Test('KroneckerSymbol(-291, -123461)', -1);
  Test('KroneckerSymbol(-291, -123462)', 0);

  for _i := -100 to 100 do
    Test('KroneckerSymbol(' + _i.ToString + ', 0)', IfThen(Abs(_i) = 1, 1, 0));

  for _i := -100 to 100 do
    Test('KroneckerSymbol(0, ' + _i.ToString + ')', IfThen(Abs(_i) = 1, 1, 0));

  for _i := Low(KroneckerSymbolValues) to High(KroneckerSymbolValues) do
    for _j := Low(KroneckerSymbolValues[_i]) to High(KroneckerSymbolValues[_i]) do
      Test(Format('KroneckerSymbol(%d, %d)', [_j, _i]), KroneckerSymbolValues[_i, _j]);

  for _i := Low(KroneckerSymbolValues) to High(KroneckerSymbolValues) do
    for _j := Low(KroneckerSymbolValues[_i]) to High(KroneckerSymbolValues[_i]) do
      if gcd(_i, _j) = 1 then
        Test(Format('abs(KroneckerSymbol(%d, %d))', [_j, _i]), 1)
      else
        Test(Format('KroneckerSymbol(%d, %d)', [_j, _i]), 0);

  // Euler's totient (phi) function

  for _i := Low(Totients) to High(Totients) do
    Test('totient(' + _i.ToString + ')', Totients[_i]);

  Test('totient(1000)', 400);
  Test('totient(5000)', 2000);
  Test('totient(10000)', 4000);
  Test('totient(50000)', 20000);
  Test('totient(100000)', 40000);
  Test('totient(500000)', 200000);
  Test('totient(1000000)', 400000);
  Test('totient(5000000)', 2000000);
  Test('totient(10000000)', 4000000);

  Test('totient(-123)', failure, 'Totient only defined for positive integers.');
  Test('totient(-1)', failure, 'Totient only defined for positive integers.');
  Test('totient(0)', failure, 'Totient only defined for positive integers.');
  Test('totient(1.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('totient()', failure, 'A required argument of type integer is missing.');
  Test('totient(10, 20)', failure, 'Too many arguments.');

  // Cototient

  for _i := Low(Totients) to High(Totients) do
    Test('cototient(' + _i.ToString + ')', _i - Totients[_i]);

  Test('cototient(1000)', 600);
  Test('cototient(5000)', 3000);
  Test('cototient(10000)', 6000);
  Test('cototient(50000)', 30000);
  Test('cototient(100000)', 60000);
  Test('cototient(500000)', 300000);
  Test('cototient(1000000)', 600000);
  Test('cototient(5000000)', 3000000);
  Test('cototient(10000000)', 6000000);

  Test('cototient(-123)', failure, 'Totient only defined for positive integers.');
  Test('cototient(-1)', failure, 'Totient only defined for positive integers.');
  Test('cototient(0)', failure, 'Totient only defined for positive integers.');
  Test('cototient(1.5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('cototient()', failure, 'A required argument of type integer is missing.');
  Test('cototient(10, 20)', failure, 'Too many arguments.');

  // Prime factorization

  Test('PrimeFactors(1)', intarr([]));
  Test('PrimeFactors(12)', intarr([2, 2, 3]));
  Test('PrimeFactors(123)', intarr([3, 41]));
  Test('PrimeFactors(1234)', intarr([2, 617]));
  Test('PrimeFactors(12345)', intarr([3, 5, 823]));
  Test('PrimeFactors(123456)', intarr([2, 2, 2, 2, 2, 2, 3, 643]));
  Test('PrimeFactors(1234567)', intarr([127, 9721]));
  Test('PrimeFactors(12345678)', intarr([2, 3, 3, 47, 14593]));
  Test('PrimeFactors(123456789)', intarr([3, 3, 3607, 3803]));

  Test('PrimeFactors(10)', intarr([2, 5]));
  Test('PrimeFactors(100)', intarr([2, 2, 5, 5]));
  Test('PrimeFactors(1000)', intarr([2, 2, 2, 5, 5, 5]));
  Test('PrimeFactors(10000)', intarr([2, 2, 2, 2, 5, 5, 5, 5]));
  Test('PrimeFactors(100000)', intarr([2, 2, 2, 2, 2, 5, 5, 5, 5, 5]));
  Test('PrimeFactors(1000000)', intarr([2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5]));
  Test('PrimeFactors(10000000)', intarr([2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 5]));
  Test('PrimeFactors(100000000)', intarr([2, 2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 5, 5]));
  Test('PrimeFactors(1000000000)', intarr([2, 2, 2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 5, 5, 5]));

  Test('PrimeFactors(84910)', intarr([2, 5, 7, 1213]));
  Test('PrimeFactors(93874)', intarr([2, 11, 17, 251]));
  Test('PrimeFactors(925714)', intarr([2, 257, 1801]));
  Test('PrimeFactors(911775)', intarr([3, 5, 5, 12157]));
  Test('PrimeFactors(3529810)', intarr([2, 5, 23, 103, 149]));
  Test('PrimeFactors(3529811)', intarr([277, 12743]));
  Test('PrimeFactors(7482571)', intarr([541, 13831]));
  Test('PrimeFactors(8267182)', intarr([2, 7, 7, 11, 7669]));
  Test('PrimeFactors(68217455)', intarr([5, 37, 368743]));
  Test('PrimeFactors(716084055)', intarr([3, 3, 5, 15912979]));
  Test('PrimeFactors(1817854564)', intarr([2, 2, 19, 23919139]));
  Test('PrimeFactors(84042420832)', intarr([2, 2, 2, 2, 2, 211, 12447041]));
  Test('PrimeFactors(84914005284)', intarr([2, 2, 3, 3, 211, 1283, 8713]));
  Test('PrimeFactors(415041487431)', intarr([3, 17, 22271, 365411]));
  Test('PrimeFactors(480818410450)', intarr([2, 5, 5, 7, 89, 15435583]));
  Test('PrimeFactors(876900466545)', intarr([3, 5, 17, 47, 6113, 11969]));
  Test('PrimeFactors(4080854248054)', intarr([2, 347, 5880193441]));
  Test('PrimeFactors(78148891090419)', intarr([3, 26049630363473]));
  Test('PrimeFactors(71814860460404)', intarr([2, 2, 19, 944932374479]));
  Test('PrimeFactors(21289442464403)', intarr([857, 24841823179]));
  Test('PrimeFactors(654145381543015)', intarr([5, 389, 30859, 10898653]));
  Test('PrimeFactors(6895186126573215)', intarr([3, 5, 2767, 166129047743]));
  Test('PrimeFactors(1125899906842624)', intarr([2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]));
  Test('PrimeFactors(1752489248702471)', intarr([109, 2531, 20563, 308923]));
  Test('PrimeFactors(461529845206245646)', intarr([2, 23, 47, 83, 499, 563, 1327, 6899]));

{$IFNDEF QuickTest}
  Test('PrimeFactors(461529845206245640)', intarr([2, 2, 2, 5, 11, 1048931466377831]));
  Test('PrimeFactors(461529845206245641)', intarr([7, 29, 179, 12701374499993]));
  Test('PrimeFactors(461529845206245642)', intarr([2, 3, 1091, 236087, 298642571]));
  Test('PrimeFactors(461529845206245643)', intarr([2798921, 164895631283]));
  Test('PrimeFactors(461529845206245644)', intarr([2, 2, 3533, 218287, 149612641]));
  Test('PrimeFactors(461529845206245645)', intarr([3, 5, 1373, 841979, 26615629]));
  Test('PrimeFactors(461529845206245646)', intarr([2, 23, 47, 83, 499, 563, 1327, 6899]));
  Test('PrimeFactors(461529845206245647)', intarr([13, 557, 2387941, 26691787]));
  Test('PrimeFactors(461529845206245648)', intarr([2, 2, 2, 2, 3, 3, 7, 43, 1361, 9973, 784489]));
  Test('PrimeFactors(461529845206245649)', intarr([103, 107, 3116699, 13436431]));
  Test('PrimeFactors(461529845206245650)', intarr([2, 5, 5, 1229, 7510656553397]));
  Test('PrimeFactors(461529845206245651)', intarr([3, 11, 11, 17, 74790122379881]));
  Test('PrimeFactors(461529845206245652)', intarr([2, 2, 27419911, 4207980883]));
  Test('PrimeFactors(461529845206245653)', intarr([373, 1237345429507361]));
{$ENDIF}

  Test('PrimeFactors(2)', intarr([2]));
  Test('PrimeFactors(3)', intarr([3]));
  Test('PrimeFactors(4)', intarr([2, 2]));
  Test('PrimeFactors(5)', intarr([5]));
  Test('PrimeFactors(6)', intarr([2, 3]));
  Test('PrimeFactors(7)', intarr([7]));
  Test('PrimeFactors(8)', intarr([2, 2, 2]));
  Test('PrimeFactors(9)', intarr([3, 3]));
  Test('PrimeFactors(10)', intarr([2, 5]));
  Test('PrimeFactors(11)', intarr([11]));

  Test('PrimeFactors(574298353)', intarr([574298353]));

{$IFNDEF QuickTest}
  Test('PrimeFactors(253241523120337)', intarr([253241523120337]));
  Test('PrimeFactors(11253241543723537)', intarr([11253241543723537]));
{$ENDIF}

  Test('PrimeFactors(0)', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('PrimeFactors(-1)', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('PrimeFactors(-2)', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('PrimeFactors(-100000000000000)', failure, 'Cannot find the prime factors of a non-positive integer.');

  Test('PrimeFactors(5.6)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('PrimeFactors()', failure, 'A required argument of type integer is missing.');
  Test('PrimeFactors(5, 6)', failure, 'Too many arguments.');

  // Radical of an integer

  for _i := Low(Radicals) to High(Radicals) do
    Test('rad(' + _i.ToString + ')', Radicals[_i]);

  for _i := 1 to 1000 do
    Test(Format('rad(%d) = product(unique(PrimeFactors(%d)))', [_i, _i]), True);

  Test('rad(0)', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('rad(-1)', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('rad(-2)', failure, 'Cannot find the prime factors of a non-positive integer.');
  Test('rad(-100000000000000)', failure, 'Cannot find the prime factors of a non-positive integer.');

  Test('rad(5.6)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('rad()', failure, 'A required argument of type integer is missing.');
  Test('rad(5, 6)', failure, 'Too many arguments.');

  // Square free integers

  Test('IsSquareFree(1)', True);
  Test('IsSquareFree(12)', False);
  Test('IsSquareFree(123)', True);
  Test('IsSquareFree(1234)', True);
  Test('IsSquareFree(12345)', True);
  Test('IsSquareFree(123456)', False);
  Test('IsSquareFree(1234567)', True);
  Test('IsSquareFree(12345678)', False);
  Test('IsSquareFree(123456789)', False);

  Test('IsSquareFree(10)', True);
  Test('IsSquareFree(100)', False);
  Test('IsSquareFree(1000)', False);
  Test('IsSquareFree(10000)', False);
  Test('IsSquareFree(100000)', False);
  Test('IsSquareFree(1000000)', False);
  Test('IsSquareFree(10000000)', False);
  Test('IsSquareFree(100000000)', False);
  Test('IsSquareFree(1000000000)', False);

  Test('IsSquareFree(84910)', True);
  Test('IsSquareFree(93874)', True);
  Test('IsSquareFree(925714)', True);
  Test('IsSquareFree(911775)', False);
  Test('IsSquareFree(3529810)', True);
  Test('IsSquareFree(3529811)', True);
  Test('IsSquareFree(7482571)', True);
  Test('IsSquareFree(8267182)', False);
  Test('IsSquareFree(68217455)', True);
  Test('IsSquareFree(716084055)', False);
  Test('IsSquareFree(1817854564)', False);
  Test('IsSquareFree(84042420832)', False);
  Test('IsSquareFree(84914005284)', False);
  Test('IsSquareFree(415041487431)', True);
  Test('IsSquareFree(480818410450)', False);
  Test('IsSquareFree(876900466545)', True);
  Test('IsSquareFree(4080854248054)', True);
  Test('IsSquareFree(78148891090419)', True);
  Test('IsSquareFree(71814860460404)', False);
  Test('IsSquareFree(21289442464403)', True);
  Test('IsSquareFree(232152425429401)', False);
  Test('IsSquareFree(654145381543015)', True);
  Test('IsSquareFree(6895186126573215)', True);
  Test('IsSquareFree(1125899906842624)', False);
  Test('IsSquareFree(1752489248702471)', True);
  Test('IsSquareFree(461529845206245646)', True);

{$IFNDEF QuickTest}
  Test('IsSquareFree(461529845206245640)', False);
  Test('IsSquareFree(461529845206245641)', True);
  Test('IsSquareFree(461529845206245642)', True);
  Test('IsSquareFree(461529845206245643)', True);
  Test('IsSquareFree(461529845206245644)', False);
  Test('IsSquareFree(461529845206245645)', True);
  Test('IsSquareFree(461529845206245646)', True);
  Test('IsSquareFree(461529845206245647)', True);
  Test('IsSquareFree(461529845206245648)', False);
  Test('IsSquareFree(461529845206245649)', True);
  Test('IsSquareFree(461529845206245650)', False);
  Test('IsSquareFree(461529845206245651)', False);
  Test('IsSquareFree(461529845206245652)', False);
  Test('IsSquareFree(461529845206245653)', True);
{$ENDIF}

  Test('IsSquareFree(2)', True);
  Test('IsSquareFree(3)', True);
  Test('IsSquareFree(4)', False);
  Test('IsSquareFree(5)', True);
  Test('IsSquareFree(6)', True);
  Test('IsSquareFree(7)', True);
  Test('IsSquareFree(8)', False);
  Test('IsSquareFree(9)', False);
  Test('IsSquareFree(10)', True);
  Test('IsSquareFree(11)', True);

  Test('IsSquareFree(574298353)', True);

{$IFNDEF QuickTest}
  Test('IsSquareFree(253241523120337)', True);
  Test('IsSquareFree(11253241543723537)', True);
{$ENDIF}



  Test('IsSquareFree(-1)', True);
  Test('IsSquareFree(-12)', False);
  Test('IsSquareFree(-123)', True);
  Test('IsSquareFree(-1234)', True);
  Test('IsSquareFree(-12345)', True);
  Test('IsSquareFree(-123456)', False);
  Test('IsSquareFree(-1234567)', True);
  Test('IsSquareFree(-12345678)', False);
  Test('IsSquareFree(-123456789)', False);

  Test('IsSquareFree(-10)', True);
  Test('IsSquareFree(-100)', False);
  Test('IsSquareFree(-1000)', False);
  Test('IsSquareFree(-10000)', False);
  Test('IsSquareFree(-100000)', False);
  Test('IsSquareFree(-1000000)', False);
  Test('IsSquareFree(-10000000)', False);
  Test('IsSquareFree(-100000000)', False);
  Test('IsSquareFree(-1000000000)', False);

  Test('IsSquareFree(-84910)', True);
  Test('IsSquareFree(-93874)', True);
  Test('IsSquareFree(-925714)', True);
  Test('IsSquareFree(-911775)', False);
  Test('IsSquareFree(-3529810)', True);
  Test('IsSquareFree(-3529811)', True);
  Test('IsSquareFree(-7482571)', True);
  Test('IsSquareFree(-8267182)', False);
  Test('IsSquareFree(-68217455)', True);
  Test('IsSquareFree(-716084055)', False);
  Test('IsSquareFree(-1817854564)', False);
  Test('IsSquareFree(-84042420832)', False);
  Test('IsSquareFree(-84914005284)', False);
  Test('IsSquareFree(-415041487431)', True);
  Test('IsSquareFree(-480818410450)', False);
  Test('IsSquareFree(-876900466545)', True);
  Test('IsSquareFree(-4080854248054)', True);
  Test('IsSquareFree(-78148891090419)', True);
  Test('IsSquareFree(-71814860460404)', False);
  Test('IsSquareFree(-21289442464403)', True);
  Test('IsSquareFree(-232152425429401)', False);
  Test('IsSquareFree(-654145381543015)', True);
  Test('IsSquareFree(-6895186126573215)', True);
  Test('IsSquareFree(-1125899906842624)', False);
  Test('IsSquareFree(-1752489248702471)', True);
  Test('IsSquareFree(-461529845206245646)', True);

{$IFNDEF QuickTest}
  Test('IsSquareFree(-461529845206245640)', False);
  Test('IsSquareFree(-461529845206245641)', True);
  Test('IsSquareFree(-461529845206245642)', True);
  Test('IsSquareFree(-461529845206245643)', True);
  Test('IsSquareFree(-461529845206245644)', False);
  Test('IsSquareFree(-461529845206245645)', True);
  Test('IsSquareFree(-461529845206245646)', True);
  Test('IsSquareFree(-461529845206245647)', True);
  Test('IsSquareFree(-461529845206245648)', False);
  Test('IsSquareFree(-461529845206245649)', True);
  Test('IsSquareFree(-461529845206245650)', False);
  Test('IsSquareFree(-461529845206245651)', False);
  Test('IsSquareFree(-461529845206245652)', False);
  Test('IsSquareFree(-461529845206245653)', True);
{$ENDIF}

  Test('IsSquareFree(-2)', True);
  Test('IsSquareFree(-3)', True);
  Test('IsSquareFree(-4)', False);
  Test('IsSquareFree(-5)', True);
  Test('IsSquareFree(-6)', True);
  Test('IsSquareFree(-7)', True);
  Test('IsSquareFree(-8)', False);
  Test('IsSquareFree(-9)', False);
  Test('IsSquareFree(-10)', True);
  Test('IsSquareFree(-11)', True);

  Test('IsSquareFree(-574298353)', True);

{$IFNDEF QuickTest}
  Test('IsSquareFree(-253241523120337)', True);
  Test('IsSquareFree(-11253241543723537)', True);
{$ENDIF}

  Test('IsSquareFree(5.6)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('IsSquareFree()', failure, 'A required argument of type integer is missing.');
  Test('IsSquareFree(5, 6)', failure, 'Too many arguments.');

  for _i := 1 to 1000 do
    Test(Format('IsSquareFree(%d) = (%d = rad(%d))', [_i, _i, _i]), True);

  for _i := 1 to 1000 do
    Test(Format('IsSquareFree(%d) = (PrimeFactors(%d) = unique(PrimeFactors(%d)))', [_i, _i, _i]), True);

  // Integer factorisation

  Test('factors(1)', intarr([1])); // differs from PrimeFactors
  Test('factors(12)', intarr([2, 2, 3]));
  Test('factors(123)', intarr([3, 41]));
  Test('factors(1234)', intarr([2, 617]));
  Test('factors(12345)', intarr([3, 5, 823]));
  Test('factors(123456)', intarr([2, 2, 2, 2, 2, 2, 3, 643]));
  Test('factors(1234567)', intarr([127, 9721]));
  Test('factors(12345678)', intarr([2, 3, 3, 47, 14593]));
  Test('factors(123456789)', intarr([3, 3, 3607, 3803]));

  Test('factors(10)', intarr([2, 5]));
  Test('factors(100)', intarr([2, 2, 5, 5]));
  Test('factors(1000)', intarr([2, 2, 2, 5, 5, 5]));
  Test('factors(10000)', intarr([2, 2, 2, 2, 5, 5, 5, 5]));
  Test('factors(100000)', intarr([2, 2, 2, 2, 2, 5, 5, 5, 5, 5]));
  Test('factors(1000000)', intarr([2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5]));
  Test('factors(10000000)', intarr([2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 5]));
  Test('factors(100000000)', intarr([2, 2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 5, 5]));
  Test('factors(1000000000)', intarr([2, 2, 2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 5, 5, 5]));

  Test('factors(84910)', intarr([2, 5, 7, 1213]));
  Test('factors(93874)', intarr([2, 11, 17, 251]));
  Test('factors(925714)', intarr([2, 257, 1801]));
  Test('factors(911775)', intarr([3, 5, 5, 12157]));
  Test('factors(3529810)', intarr([2, 5, 23, 103, 149]));
  Test('factors(3529811)', intarr([277, 12743]));
  Test('factors(7482571)', intarr([541, 13831]));
  Test('factors(8267182)', intarr([2, 7, 7, 11, 7669]));
  Test('factors(68217455)', intarr([5, 37, 368743]));
  Test('factors(716084055)', intarr([3, 3, 5, 15912979]));
  Test('factors(1817854564)', intarr([2, 2, 19, 23919139]));
  Test('factors(84042420832)', intarr([2, 2, 2, 2, 2, 211, 12447041]));
  Test('factors(84914005284)', intarr([2, 2, 3, 3, 211, 1283, 8713]));
  Test('factors(415041487431)', intarr([3, 17, 22271, 365411]));
  Test('factors(480818410450)', intarr([2, 5, 5, 7, 89, 15435583]));
  Test('factors(876900466545)', intarr([3, 5, 17, 47, 6113, 11969]));
  Test('factors(4080854248054)', intarr([2, 347, 5880193441]));
  Test('factors(78148891090419)', intarr([3, 26049630363473]));
  Test('factors(71814860460404)', intarr([2, 2, 19, 944932374479]));
  Test('factors(21289442464403)', intarr([857, 24841823179]));
  Test('factors(654145381543015)', intarr([5, 389, 30859, 10898653]));
  Test('factors(6895186126573215)', intarr([3, 5, 2767, 166129047743]));
  Test('factors(1125899906842624)', intarr([2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]));
  Test('factors(1752489248702471)', intarr([109, 2531, 20563, 308923]));
  Test('factors(461529845206245646)', intarr([2, 23, 47, 83, 499, 563, 1327, 6899]));

{$IFNDEF QuickTest}
  Test('factors(461529845206245640)', intarr([2, 2, 2, 5, 11, 1048931466377831]));
  Test('factors(461529845206245641)', intarr([7, 29, 179, 12701374499993]));
  Test('factors(461529845206245642)', intarr([2, 3, 1091, 236087, 298642571]));
  Test('factors(461529845206245643)', intarr([2798921, 164895631283]));
  Test('factors(461529845206245644)', intarr([2, 2, 3533, 218287, 149612641]));
  Test('factors(461529845206245645)', intarr([3, 5, 1373, 841979, 26615629]));
  Test('factors(461529845206245646)', intarr([2, 23, 47, 83, 499, 563, 1327, 6899]));
  Test('factors(461529845206245647)', intarr([13, 557, 2387941, 26691787]));
  Test('factors(461529845206245648)', intarr([2, 2, 2, 2, 3, 3, 7, 43, 1361, 9973, 784489]));
  Test('factors(461529845206245649)', intarr([103, 107, 3116699, 13436431]));
  Test('factors(461529845206245650)', intarr([2, 5, 5, 1229, 7510656553397]));
  Test('factors(461529845206245651)', intarr([3, 11, 11, 17, 74790122379881]));
  Test('factors(461529845206245652)', intarr([2, 2, 27419911, 4207980883]));
  Test('factors(461529845206245653)', intarr([373, 1237345429507361]));
{$ENDIF}

  Test('factors(2)', intarr([2]));
  Test('factors(3)', intarr([3]));
  Test('factors(4)', intarr([2, 2]));
  Test('factors(5)', intarr([5]));
  Test('factors(6)', intarr([2, 3]));
  Test('factors(7)', intarr([7]));
  Test('factors(8)', intarr([2, 2, 2]));
  Test('factors(9)', intarr([3, 3]));
  Test('factors(10)', intarr([2, 5]));
  Test('factors(11)', intarr([11]));

  Test('factors(574298353)', intarr([574298353]));

{$IFNDEF QuickTest}
  Test('factors(253241523120337)', intarr([253241523120337]));
  Test('factors(11253241543723537)', intarr([11253241543723537]));
{$ENDIF}

  Test('factors(0)', intarr([0])); // differs from PrimeFactors

  Test('factors(-1)', intarr([-1])); // all negative ones differ from PrimeFactors
  Test('factors(-12)', intarr([-2, 2, 3]));
  Test('factors(-123)', intarr([-3, 41]));
  Test('factors(-1234)', intarr([-2, 617]));
  Test('factors(-12345)', intarr([-3, 5, 823]));
  Test('factors(-123456)', intarr([-2, 2, 2, 2, 2, 2, 3, 643]));
  Test('factors(-1234567)', intarr([-127, 9721]));
  Test('factors(-12345678)', intarr([-2, 3, 3, 47, 14593]));
  Test('factors(-123456789)', intarr([-3, 3, 3607, 3803]));

  Test('factors(-10)', intarr([-2, 5]));
  Test('factors(-100)', intarr([-2, 2, 5, 5]));
  Test('factors(-1000)', intarr([-2, 2, 2, 5, 5, 5]));
  Test('factors(-10000)', intarr([-2, 2, 2, 2, 5, 5, 5, 5]));
  Test('factors(-100000)', intarr([-2, 2, 2, 2, 2, 5, 5, 5, 5, 5]));
  Test('factors(-1000000)', intarr([-2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5]));
  Test('factors(-10000000)', intarr([-2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 5]));
  Test('factors(-100000000)', intarr([-2, 2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 5, 5]));
  Test('factors(-1000000000)', intarr([-2, 2, 2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 5, 5, 5]));

  Test('factors(-84910)', intarr([-2, 5, 7, 1213]));
  Test('factors(-93874)', intarr([-2, 11, 17, 251]));
  Test('factors(-925714)', intarr([-2, 257, 1801]));
  Test('factors(-911775)', intarr([-3, 5, 5, 12157]));
  Test('factors(-3529810)', intarr([-2, 5, 23, 103, 149]));
  Test('factors(-3529811)', intarr([-277, 12743]));
  Test('factors(-7482571)', intarr([-541, 13831]));
  Test('factors(-8267182)', intarr([-2, 7, 7, 11, 7669]));
  Test('factors(-68217455)', intarr([-5, 37, 368743]));
  Test('factors(-716084055)', intarr([-3, 3, 5, 15912979]));
  Test('factors(-1817854564)', intarr([-2, 2, 19, 23919139]));
  Test('factors(-84042420832)', intarr([-2, 2, 2, 2, 2, 211, 12447041]));
  Test('factors(-84914005284)', intarr([-2, 2, 3, 3, 211, 1283, 8713]));
  Test('factors(-415041487431)', intarr([-3, 17, 22271, 365411]));
  Test('factors(-480818410450)', intarr([-2, 5, 5, 7, 89, 15435583]));
  Test('factors(-876900466545)', intarr([-3, 5, 17, 47, 6113, 11969]));
  Test('factors(-4080854248054)', intarr([-2, 347, 5880193441]));
  Test('factors(-78148891090419)', intarr([-3, 26049630363473]));
  Test('factors(-71814860460404)', intarr([-2, 2, 19, 944932374479]));
  Test('factors(-21289442464403)', intarr([-857, 24841823179]));
  Test('factors(-654145381543015)', intarr([-5, 389, 30859, 10898653]));
  Test('factors(-6895186126573215)', intarr([-3, 5, 2767, 166129047743]));
  Test('factors(-1125899906842624)', intarr([-2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]));
  Test('factors(-1752489248702471)', intarr([-109, 2531, 20563, 308923]));
  Test('factors(-461529845206245646)', intarr([-2, 23, 47, 83, 499, 563, 1327, 6899]));

{$IFNDEF QuickTest}
  Test('factors(-461529845206245640)', intarr([-2, 2, 2, 5, 11, 1048931466377831]));
  Test('factors(-461529845206245641)', intarr([-7, 29, 179, 12701374499993]));
  Test('factors(-461529845206245642)', intarr([-2, 3, 1091, 236087, 298642571]));
  Test('factors(-461529845206245643)', intarr([-2798921, 164895631283]));
  Test('factors(-461529845206245644)', intarr([-2, 2, 3533, 218287, 149612641]));
  Test('factors(-461529845206245645)', intarr([-3, 5, 1373, 841979, 26615629]));
  Test('factors(-461529845206245646)', intarr([-2, 23, 47, 83, 499, 563, 1327, 6899]));
  Test('factors(-461529845206245647)', intarr([-13, 557, 2387941, 26691787]));
  Test('factors(-461529845206245648)', intarr([-2, 2, 2, 2, 3, 3, 7, 43, 1361, 9973, 784489]));
  Test('factors(-461529845206245649)', intarr([-103, 107, 3116699, 13436431]));
  Test('factors(-461529845206245650)', intarr([-2, 5, 5, 1229, 7510656553397]));
  Test('factors(-461529845206245651)', intarr([-3, 11, 11, 17, 74790122379881]));
  Test('factors(-461529845206245652)', intarr([-2, 2, 27419911, 4207980883]));
  Test('factors(-461529845206245653)', intarr([-373, 1237345429507361]));
{$ENDIF}

  Test('factors(-2)', intarr([-2]));
  Test('factors(-3)', intarr([-3]));
  Test('factors(-4)', intarr([-2, 2]));
  Test('factors(-5)', intarr([-5]));
  Test('factors(-6)', intarr([-2, 3]));
  Test('factors(-7)', intarr([-7]));
  Test('factors(-8)', intarr([-2, 2, 2]));
  Test('factors(-9)', intarr([-3, 3]));
  Test('factors(-10)', intarr([-2, 5]));
  Test('factors(-11)', intarr([-11]));

  Test('factors(-574298353)', intarr([-574298353]));

{$IFNDEF QuickTest}
  Test('factors(-253241523120337)', intarr([-253241523120337]));
  Test('factors(-11253241543723537)', intarr([-11253241543723537]));
{$ENDIF}

  Test('factors(5.6)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('factors()', failure, 'A required argument of type integer is missing.');
  Test('factors(5, 6)', failure, 'Too many arguments.');

  // Factorised expression as string

  Test('FactorizedExpression(1)', '1');
  Test('FactorizedExpression(12)', '2⋅2⋅3');
  Test('FactorizedExpression(123)', '3⋅41');
  Test('FactorizedExpression(1234)', '2⋅617');
  Test('FactorizedExpression(12345)', '3⋅5⋅823');
  Test('FactorizedExpression(123456)', '2⋅2⋅2⋅2⋅2⋅2⋅3⋅643');
  Test('FactorizedExpression(1234567)', '127⋅9721');
  Test('FactorizedExpression(12345678)', '2⋅3⋅3⋅47⋅14593');
  Test('FactorizedExpression(123456789)', '3⋅3⋅3607⋅3803');

  Test('FactorizedExpression(10)', '2⋅5');
  Test('FactorizedExpression(100)', '2⋅2⋅5⋅5');
  Test('FactorizedExpression(1000)', '2⋅2⋅2⋅5⋅5⋅5');
  Test('FactorizedExpression(10000)', '2⋅2⋅2⋅2⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(100000)', '2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(1000000)', '2⋅2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(10000000)', '2⋅2⋅2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(100000000)', '2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(1000000000)', '2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5⋅5⋅5⋅5⋅5');

  Test('FactorizedExpression(84910)', '2⋅5⋅7⋅1213');
  Test('FactorizedExpression(93874)', '2⋅11⋅17⋅251');
  Test('FactorizedExpression(925714)', '2⋅257⋅1801');
  Test('FactorizedExpression(911775)', '3⋅5⋅5⋅12157');
  Test('FactorizedExpression(3529810)', '2⋅5⋅23⋅103⋅149');
  Test('FactorizedExpression(3529811)', '277⋅12743');
  Test('FactorizedExpression(7482571)', '541⋅13831');
  Test('FactorizedExpression(8267182)', '2⋅7⋅7⋅11⋅7669');
  Test('FactorizedExpression(68217455)', '5⋅37⋅368743');
  Test('FactorizedExpression(716084055)', '3⋅3⋅5⋅15912979');
  Test('FactorizedExpression(1817854564)', '2⋅2⋅19⋅23919139');
  Test('FactorizedExpression(84042420832)', '2⋅2⋅2⋅2⋅2⋅211⋅12447041');
  Test('FactorizedExpression(84914005284)', '2⋅2⋅3⋅3⋅211⋅1283⋅8713');
  Test('FactorizedExpression(415041487431)', '3⋅17⋅22271⋅365411');
  Test('FactorizedExpression(480818410450)', '2⋅5⋅5⋅7⋅89⋅15435583');
  Test('FactorizedExpression(876900466545)', '3⋅5⋅17⋅47⋅6113⋅11969');
  Test('FactorizedExpression(4080854248054)', '2⋅347⋅5880193441');
  Test('FactorizedExpression(78148891090419)', '3⋅26049630363473');
  Test('FactorizedExpression(71814860460404)', '2⋅2⋅19⋅944932374479');
  Test('FactorizedExpression(21289442464403)', '857⋅24841823179');
  Test('FactorizedExpression(654145381543015)', '5⋅389⋅30859⋅10898653');
  Test('FactorizedExpression(6895186126573215)', '3⋅5⋅2767⋅166129047743');
  Test('FactorizedExpression(1125899906842624)', '2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2');
  Test('FactorizedExpression(1752489248702471)', '109⋅2531⋅20563⋅308923');
  Test('FactorizedExpression(461529845206245646)', '2⋅23⋅47⋅83⋅499⋅563⋅1327⋅6899');

  Test('FactorizedExpression(0)', '0');

  Test('FactorizedExpression(-1)', '−1');
  Test('FactorizedExpression(-12)', '−2⋅2⋅3');
  Test('FactorizedExpression(-123)', '−3⋅41');
  Test('FactorizedExpression(-1234)', '−2⋅617');
  Test('FactorizedExpression(-12345)', '−3⋅5⋅823');
  Test('FactorizedExpression(-123456)', '−2⋅2⋅2⋅2⋅2⋅2⋅3⋅643');
  Test('FactorizedExpression(-1234567)', '−127⋅9721');
  Test('FactorizedExpression(-12345678)', '−2⋅3⋅3⋅47⋅14593');
  Test('FactorizedExpression(-123456789)', '−3⋅3⋅3607⋅3803');

  Test('FactorizedExpression(-10)', '−2⋅5');
  Test('FactorizedExpression(-100)', '−2⋅2⋅5⋅5');
  Test('FactorizedExpression(-1000)', '−2⋅2⋅2⋅5⋅5⋅5');
  Test('FactorizedExpression(-10000)', '−2⋅2⋅2⋅2⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(-100000)', '−2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(-1000000)', '−2⋅2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(-10000000)', '−2⋅2⋅2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(-100000000)', '−2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5⋅5⋅5⋅5');
  Test('FactorizedExpression(-1000000000)', '−2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅5⋅5⋅5⋅5⋅5⋅5⋅5⋅5⋅5');

  Test('FactorizedExpression(-84910)', '−2⋅5⋅7⋅1213');
  Test('FactorizedExpression(-93874)', '−2⋅11⋅17⋅251');
  Test('FactorizedExpression(-925714)', '−2⋅257⋅1801');
  Test('FactorizedExpression(-911775)', '−3⋅5⋅5⋅12157');
  Test('FactorizedExpression(-3529810)', '−2⋅5⋅23⋅103⋅149');
  Test('FactorizedExpression(-3529811)', '−277⋅12743');
  Test('FactorizedExpression(-7482571)', '−541⋅13831');
  Test('FactorizedExpression(-8267182)', '−2⋅7⋅7⋅11⋅7669');
  Test('FactorizedExpression(-68217455)', '−5⋅37⋅368743');
  Test('FactorizedExpression(-716084055)', '−3⋅3⋅5⋅15912979');
  Test('FactorizedExpression(-1817854564)', '−2⋅2⋅19⋅23919139');
  Test('FactorizedExpression(-84042420832)', '−2⋅2⋅2⋅2⋅2⋅211⋅12447041');
  Test('FactorizedExpression(-84914005284)', '−2⋅2⋅3⋅3⋅211⋅1283⋅8713');
  Test('FactorizedExpression(-415041487431)', '−3⋅17⋅22271⋅365411');
  Test('FactorizedExpression(-480818410450)', '−2⋅5⋅5⋅7⋅89⋅15435583');
  Test('FactorizedExpression(-876900466545)', '−3⋅5⋅17⋅47⋅6113⋅11969');
  Test('FactorizedExpression(-4080854248054)', '−2⋅347⋅5880193441');
  Test('FactorizedExpression(-78148891090419)', '−3⋅26049630363473');
  Test('FactorizedExpression(-71814860460404)', '−2⋅2⋅19⋅944932374479');
  Test('FactorizedExpression(-21289442464403)', '−857⋅24841823179');
  Test('FactorizedExpression(-654145381543015)', '−5⋅389⋅30859⋅10898653');
  Test('FactorizedExpression(-6895186126573215)', '−3⋅5⋅2767⋅166129047743');
  Test('FactorizedExpression(-1125899906842624)', '−2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2⋅2');
  Test('FactorizedExpression(-1752489248702471)', '−109⋅2531⋅20563⋅308923');
  Test('FactorizedExpression(-461529845206245646)', '−2⋅23⋅47⋅83⋅499⋅563⋅1327⋅6899');

  Test('FactorizedExpression(5.6)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('FactorizedExpression()', failure, 'A required argument of type integer is missing.');
  Test('FactorizedExpression(5, 6)', failure, 'Too many arguments.');

  for _i := 123000 to 124000 do
    Test(Format('FactorizedExpression(%d) = join(factors(%d), "⋅")', [_i, _i]), True);

  // Divisors

  Test('divisors(1)', intarr([1]));
  Test('divisors(2)', intarr([1, 2]));
  Test('divisors(3)', intarr([1, 3]));
  Test('divisors(4)', intarr([1, 2, 4]));
  Test('divisors(5)', intarr([1, 5]));
  Test('divisors(6)', intarr([1, 2, 3, 6]));
  Test('divisors(7)', intarr([1, 7]));
  Test('divisors(8)', intarr([1, 2, 4, 8]));
  Test('divisors(9)', intarr([1, 3, 9]));
  Test('divisors(10)', intarr([1, 2, 5, 10]));
  Test('divisors(11)', intarr([1, 11]));
  Test('divisors(12)', intarr([1, 2, 3, 4, 6, 12]));
  Test('divisors(13)', intarr([1, 13]));
  Test('divisors(14)', intarr([1, 2, 7, 14]));
  Test('divisors(15)', intarr([1, 3, 5, 15]));
  Test('divisors(16)', intarr([1, 2, 4, 8, 16]));
  Test('divisors(17)', intarr([1, 17]));
  Test('divisors(18)', intarr([1, 2, 3, 6, 9, 18]));
  Test('divisors(19)', intarr([1, 19]));
  Test('divisors(20)', intarr([1, 2, 4, 5, 10, 20]));
  Test('divisors(21)', intarr([1, 3, 7, 21]));

  Test('divisors(123)', intarr([1, 3, 41, 123]));
  Test('divisors(1234)', intarr([1, 2, 617, 1234]));
  Test('divisors(12345)', intarr([1, 3, 5, 15, 823, 2469, 4115, 12345]));
  Test('divisors(123456)', intarr([1, 2, 3, 4, 6, 8, 12, 16, 24, 32, 48, 64, 96, 192, 643, 1286, 1929, 2572, 3858, 5144, 7716, 10288, 15432, 20576, 30864, 41152, 61728, 123456]));
  Test('divisors(1234567)', intarr([1, 127, 9721, 1234567]));
  Test('divisors(12345678)', intarr([1, 2, 3, 6, 9, 18, 47, 94, 141, 282, 423, 846, 14593, 29186, 43779, 87558, 131337, 262674, 685871, 1371742, 2057613, 4115226, 6172839, 12345678]));
  Test('divisors(123456789)', intarr([1, 3, 9, 3607, 3803, 10821, 11409, 32463, 34227, 13717421, 41152263, 123456789]));

  Test('divisors(1073741824)', intarr([1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288, 1048576, 2097152, 4194304, 8388608, 16777216, 33554432, 67108864, 134217728, 268435456, 536870912, 1073741824]));
  Test('divisors(1125899906842624)', intarr([1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288, 1048576, 2097152, 4194304, 8388608, 16777216, 33554432, 67108864, 134217728, 268435456, 536870912, 1073741824, 2147483648, 4294967296, 8589934592, 17179869184, 34359738368, 68719476736, 137438953472, 274877906944, 549755813888, 1099511627776, 2199023255552, 4398046511104, 8796093022208, 17592186044416, 35184372088832, 70368744177664, 140737488355328, 281474976710656, 562949953421312, 1125899906842624]));
  Test('divisors(1125899906842625)', intarr([1, 5, 25, 41, 101, 125, 205, 505, 1025, 2525, 4141, 5125, 8101, 12625, 20705, 40505, 103525, 202525, 268501, 332141, 517625, 818201, 1012625, 1342505, 1660705, 4091005, 6712525, 8303525, 11008541, 20455025, 27118601, 33546241, 33562625, 41517625, 55042705, 102275125, 135593005, 167731205, 275213525, 677965025, 838656025, 1111862641, 1376067625, 2175126601, 3389825125, 4193280125, 5559313205, 10875633005, 27796566025, 54378165025, 89180190641, 138982830125, 219687786701, 271890825125, 445900953205, 1098438933505, 2229504766025, 5492194667525, 9007199254741, 11147523830125, 27460973337625, 45035996273705, 225179981368525, 1125899906842625]));
  Test('divisors(1125899906842625123)', intarr([1, 11, 41, 373, 451, 4103, 13003, 15293, 143033, 168223, 533123, 4850119, 5864353, 53351309, 198854879, 514719767, 2187403669, 5661917437, 21103510447, 191990473091, 232138614917, 2111895204001, 6692901130301, 7871609396731, 73621912433311, 86587703364041, 274408946342341, 2496452121602273, 3018498409765751, 27460973337625003, 102354536985693193, 1125899906842625123]));

  Test('divisors(-14)', intarr([1, 2, 7, 14]));
  Test('divisors(-12345)', intarr([1, 3, 5, 15, 823, 2469, 4115, 12345]));
  Test('divisors(-1125899906842625123)', intarr([1, 11, 41, 373, 451, 4103, 13003, 15293, 143033, 168223, 533123, 4850119, 5864353, 53351309, 198854879, 514719767, 2187403669, 5661917437, 21103510447, 191990473091, 232138614917, 2111895204001, 6692901130301, 7871609396731, 73621912433311, 86587703364041, 274408946342341, 2496452121602273, 3018498409765751, 27460973337625003, 102354536985693193, 1125899906842625123]));

  Test('divisors(0)', failure, 'Cannot compute divisors of zero (all non-zero integers are divisors of zero).');

  Test('divisors(5.6)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('divisors()', failure, 'A required argument of type integer is missing.');
  Test('divisors(5, 6)', failure, 'Too many arguments.');

  // Divisibility (divides function)

  Test('divides(0, 100)', failure, 'Division by zero');

  Test('divides(1, 100)', True);
  Test('divides(2, 100)', True);
  Test('divides(3, 100)', False);
  Test('divides(4, 100)', True);
  Test('divides(5, 100)', True);
  Test('divides(6, 100)', False);
  Test('divides(7, 100)', False);
  Test('divides(8, 100)', False);
  Test('divides(9, 100)', False);
  Test('divides(10, 100)', True);
  Test('divides(50, 100)', True);
  Test('divides(51, 100)', False);
  Test('divides(75, 100)', False);
  Test('divides(99, 100)', False);
  Test('divides(100, 100)', True);
  Test('divides(101, 100)', False);
  Test('divides(200, 100)', False);

  Test('divides(-1, 100)', True);
  Test('divides(-2, 100)', True);
  Test('divides(-3, 100)', False);
  Test('divides(-4, 100)', True);
  Test('divides(-5, 100)', True);
  Test('divides(-6, 100)', False);
  Test('divides(-7, 100)', False);
  Test('divides(-8, 100)', False);
  Test('divides(-9, 100)', False);
  Test('divides(-10, 100)', True);
  Test('divides(-50, 100)', True);
  Test('divides(-51, 100)', False);
  Test('divides(-75, 100)', False);
  Test('divides(-99, 100)', False);
  Test('divides(-100, 100)', True);
  Test('divides(-101, 100)', False);
  Test('divides(-200, 100)', False);

  Test('divides(1, 0)', True);
  Test('divides(2, 0)', True);
  Test('divides(3, 0)', True);
  Test('divides(4, 0)', True);
  Test('divides(5, 0)', True);
  Test('divides(6, 0)', True);
  Test('divides(7, 0)', True);
  Test('divides(8, 0)', True);
  Test('divides(9, 0)', True);
  Test('divides(10, 0)', True);
  Test('divides(50, 0)', True);
  Test('divides(51, 0)', True);
  Test('divides(75, 0)', True);
  Test('divides(99, 0)', True);
  Test('divides(100, 0)', True);
  Test('divides(101, 0)', True);
  Test('divides(200, 0)', True);

  Test('divides(-1, 0)', True);
  Test('divides(-2, 0)', True);
  Test('divides(-3, 0)', True);
  Test('divides(-4, 0)', True);
  Test('divides(-5, 0)', True);
  Test('divides(-6, 0)', True);
  Test('divides(-7, 0)', True);
  Test('divides(-8, 0)', True);
  Test('divides(-9, 0)', True);
  Test('divides(-10, 0)', True);
  Test('divides(-50, 0)', True);
  Test('divides(-51, 0)', True);
  Test('divides(-75, 0)', True);
  Test('divides(-99, 0)', True);
  Test('divides(-100, 0)', True);
  Test('divides(-101, 0)', True);
  Test('divides(-200, 0)', True);

  Test('divides(1, 30410216477784)', True);
  Test('divides(2, 30410216477784)', True);
  Test('divides(3, 30410216477784)', True);
  Test('divides(4, 30410216477784)', True);
  Test('divides(5, 30410216477784)', False);

  Test('divides(231893, 30410216477784)', False);
  Test('divides(231894, 30410216477784)', True);
  Test('divides(231895, 30410216477784)', False);

  Test('divides(194937285110, 30410216477784)', False);
  Test('divides(194937285111, 30410216477784)', False);
  Test('divides(194937285112, 30410216477784)', False);
  Test('divides(194937285113, 30410216477784)', False);
  Test('divides(194937285114, 30410216477784)', True);
  Test('divides(194937285115, 30410216477784)', False);
  Test('divides(194937285116, 30410216477784)', False);
  Test('divides(194937285117, 30410216477784)', False);
  Test('divides(194937285118, 30410216477784)', False);
  Test('divides(194937285119, 30410216477784)', False);




  Test('NotDivides(0, 100)', failure, 'Division by zero');

  Test('NotDivides(1, 100)', False);
  Test('NotDivides(2, 100)', False);
  Test('NotDivides(3, 100)', True);
  Test('NotDivides(4, 100)', False);
  Test('NotDivides(5, 100)', False);
  Test('NotDivides(6, 100)', True);
  Test('NotDivides(7, 100)', True);
  Test('NotDivides(8, 100)', True);
  Test('NotDivides(9, 100)', True);
  Test('NotDivides(10, 100)', False);
  Test('NotDivides(50, 100)', False);
  Test('NotDivides(51, 100)', True);
  Test('NotDivides(75, 100)', True);
  Test('NotDivides(99, 100)', True);
  Test('NotDivides(100, 100)', False);
  Test('NotDivides(101, 100)', True);
  Test('NotDivides(200, 100)', True);

  Test('NotDivides(-1, 100)', False);
  Test('NotDivides(-2, 100)', False);
  Test('NotDivides(-3, 100)', True);
  Test('NotDivides(-4, 100)', False);
  Test('NotDivides(-5, 100)', False);
  Test('NotDivides(-6, 100)', True);
  Test('NotDivides(-7, 100)', True);
  Test('NotDivides(-8, 100)', True);
  Test('NotDivides(-9, 100)', True);
  Test('NotDivides(-10, 100)', False);
  Test('NotDivides(-50, 100)', False);
  Test('NotDivides(-51, 100)', True);
  Test('NotDivides(-75, 100)', True);
  Test('NotDivides(-99, 100)', True);
  Test('NotDivides(-100, 100)', False);
  Test('NotDivides(-101, 100)', True);
  Test('NotDivides(-200, 100)', True);

  Test('NotDivides(1, 0)', False);
  Test('NotDivides(2, 0)', False);
  Test('NotDivides(3, 0)', False);
  Test('NotDivides(4, 0)', False);
  Test('NotDivides(5, 0)', False);
  Test('NotDivides(6, 0)', False);
  Test('NotDivides(7, 0)', False);
  Test('NotDivides(8, 0)', False);
  Test('NotDivides(9, 0)', False);
  Test('NotDivides(10, 0)', False);
  Test('NotDivides(50, 0)', False);
  Test('NotDivides(51, 0)', False);
  Test('NotDivides(75, 0)', False);
  Test('NotDivides(99, 0)', False);
  Test('NotDivides(100, 0)', False);
  Test('NotDivides(101, 0)', False);
  Test('NotDivides(200, 0)', False);

  Test('NotDivides(-1, 0)', False);
  Test('NotDivides(-2, 0)', False);
  Test('NotDivides(-3, 0)', False);
  Test('NotDivides(-4, 0)', False);
  Test('NotDivides(-5, 0)', False);
  Test('NotDivides(-6, 0)', False);
  Test('NotDivides(-7, 0)', False);
  Test('NotDivides(-8, 0)', False);
  Test('NotDivides(-9, 0)', False);
  Test('NotDivides(-10, 0)', False);
  Test('NotDivides(-50, 0)', False);
  Test('NotDivides(-51, 0)', False);
  Test('NotDivides(-75, 0)', False);
  Test('NotDivides(-99, 0)', False);
  Test('NotDivides(-100, 0)', False);
  Test('NotDivides(-101, 0)', False);
  Test('NotDivides(-200, 0)', False);

  Test('NotDivides(1, 30410216477784)', False);
  Test('NotDivides(2, 30410216477784)', False);
  Test('NotDivides(3, 30410216477784)', False);
  Test('NotDivides(4, 30410216477784)', False);
  Test('NotDivides(5, 30410216477784)', True);

  Test('NotDivides(231893, 30410216477784)', True);
  Test('NotDivides(231894, 30410216477784)', False);
  Test('NotDivides(231895, 30410216477784)', True);

  Test('NotDivides(194937285110, 30410216477784)', True);
  Test('NotDivides(194937285111, 30410216477784)', True);
  Test('NotDivides(194937285112, 30410216477784)', True);
  Test('NotDivides(194937285113, 30410216477784)', True);
  Test('NotDivides(194937285114, 30410216477784)', False);
  Test('NotDivides(194937285115, 30410216477784)', True);
  Test('NotDivides(194937285116, 30410216477784)', True);
  Test('NotDivides(194937285117, 30410216477784)', True);
  Test('NotDivides(194937285118, 30410216477784)', True);
  Test('NotDivides(194937285119, 30410216477784)', True);

  // Continued fractions

  Test('ContinuedFraction(-2)', intarr([-2]));
  Test('ContinuedFraction(-1)', intarr([-1]));
  Test('ContinuedFraction(0)', intarr([0]));
  Test('ContinuedFraction(1)', intarr([1]));
  Test('ContinuedFraction(2)', intarr([2]));
  Test('ContinuedFraction(3)', intarr([3]));
  Test('ContinuedFraction(4)', intarr([4]));
  Test('ContinuedFraction(5)', intarr([5]));
  Test('ContinuedFraction(6)', intarr([6]));

  Test('ContinuedFraction(0.1357)', intarr([0, 7, 2, 1, 2, 2, 3, 6, 1, 2]));

  Test('ContinuedFraction(6.5)', intarr([6, 2]));
  Test('ContinuedFraction(6.25)', intarr([6, 4]));
  Test('ContinuedFraction(6.257)', intarr([6, 3, 1, 8, 5, 1, 1, 2]));

  Test('ContinuedFraction(10.7)', intarr([10, 1, 2, 3]));
  Test('ContinuedFraction(10.71)', intarr([10, 1, 2, 2, 4, 3]));
  Test('ContinuedFraction(10.711)', intarr([10, 1, 2, 2, 5, 1, 3, 1, 1, 2]));
  Test('ContinuedFraction(10.7115)', intarr([10, 1, 2, 2, 6, 1, 8, 1, 3]));

  Test('ContinuedFraction(123.456)', intarr([123, 2, 5, 5, 2]));
  Test('ContinuedFraction(123.456789)', intarr([123, 2, 5, 3, 1, 1, 135, 10, 9]));

  Test('ContinuedFraction(10.17536)', intarr([10, 5, 1, 2, 2, 1, 3, 4, 1, 2]));

  Test('ContinuedFraction(10.101010)', intarr([10, 9, 1, 9, 1010]));
  Test('ContinuedFraction(1000.121416)', intarr([1000, 8, 4, 4, 3, 1, 4, 1, 1, 1, 3, 1, 2]));

  Test('ContinuedFraction(1357/10000)', intarr([0, 7, 2, 1, 2, 2, 3, 6, 1, 2]));

  Test('ContinuedFraction(13/2)', intarr([6, 2]));
  Test('ContinuedFraction(25/4)', intarr([6, 4]));
  Test('ContinuedFraction(6257/1000)', intarr([6, 3, 1, 8, 5, 1, 1, 2]));

  Test('ContinuedFraction(107/10)', intarr([10, 1, 2, 3]));
  Test('ContinuedFraction(1071/100)', intarr([10, 1, 2, 2, 4, 3]));
  Test('ContinuedFraction(10711/1000)', intarr([10, 1, 2, 2, 5, 1, 3, 1, 1, 2]));
  Test('ContinuedFraction(107115/10000)', intarr([10, 1, 2, 2, 6, 1, 8, 1, 3]));

  Test('ContinuedFraction(123456/1000)', intarr([123, 2, 5, 5, 2]));
  Test('ContinuedFraction(123456789/1000000)', intarr([123, 2, 5, 3, 1, 1, 135, 10, 9]));

  Test('ContinuedFraction(1017536/100000)', intarr([10, 5, 1, 2, 2, 1, 3, 4, 1, 2]));

  Test('ContinuedFraction(10101010/1000000)', intarr([10, 9, 1, 9, 1010]));
  Test('ContinuedFraction(1000121416/1000000)', intarr([1000, 8, 4, 4, 3, 1, 4, 1, 1, 1, 3, 1, 2]));

  Test('ContinuedFraction(0.0)', intarr([0]));
  Test('ContinuedFraction(-1.1415)', intarr([-2, 1, 6, 14, 1, 8, 2]));
  Test('ContinuedFraction(0.1357902468)', intarr([0, 7, 2, 1, 2, 1, 11, 1, 2, 1, 6, 1, 15, 2, 34, 1, 5, 1]));

  Test('ContinuedFraction(17875057260063841/6305034377045117, 1000)',
    intarr([2, 1, 5, 16, 16, 1, 1, 1, 1, 4, 1, 19, 1, 4, 12, 1, 1, 1, 1, 1, 2, 2, 1, 3, 1, 1, 1, 1, 1, 1, 2, 1, 4, 2, 1, 1, 9, 4, 7]));

  Test('ContinuedFraction(17875057260063841/6305034377045117, 5)',
    intarr([2, 1, 5, 16, 16]));

  Test('ContinuedFraction(2 + i)', failure, 'Object isn''t a real number.');

  // To rational number

  // - continued fraction

  Test('ToFraction(ContinuedFraction(1/2))', Rat(1, 2));
  Test('ToFraction(ContinuedFraction(1/7))', Rat(1, 7));
  Test('ToFraction(ContinuedFraction(23/53))', Rat(23, 53));
  Test('ToFraction(ContinuedFraction(144/125))', Rat(144, 125));
  Test('ToFraction(ContinuedFraction(913/280))', Rat(913, 280));

  Test('ToFraction(ContinuedFraction(1/5))', Rat(1, 5));
  Test('ToFraction(ContinuedFraction(0.2))', Rat(1, 5));
  Test('ToFraction(ContinuedFraction(0.123))', Rat(123, 1000));

  Test('ToFraction(''(10, 1, 2, 3))', Rat(107, 10));
  Test('ToFraction(''(6, 4))', Rat(25, 4));
  Test('ToFraction(''(123, 2, 5, 5, 2))', Rat(15432, 125));
  Test('ToFraction(''(0))', Rat(0, 1));
  Test('ToFraction(''(7))', Rat(7, 1));
  Test('ToFraction(''(-10))', Rat(-10, 1));
  Test('ToFraction(''(-2, 1, 6, 14, 1, 8, 2))', Rat(-2283, 2000));

  Test('ToFraction(10, 1, 2, 3)', Rat(107, 10));
  Test('ToFraction(6, 4)', Rat(25, 4));
  Test('ToFraction(123, 2, 5, 5, 2)', Rat(15432, 125));
  Test('ToFraction(0)', Rat(0, 1));
  Test('ToFraction(7)', Rat(7, 1));
  Test('ToFraction(-10)', Rat(-10, 1));
  Test('ToFraction(-2, 1, 6, 14, 1, 8, 2)', Rat(-2283, 2000));

  Test('ToFraction(''(2, 4))', Rat(9, 4));
  Test('ToFraction(''(2, 4, "cat"))', failure, 'The array must contain only objects of type integer, but it contains an object of type string.');

  Test('ToFraction(2, 4)', Rat(9, 4));
  Test('ToFraction(2, 4, "rabbit")', failure, 'An object of type integer was expected as argument 3, but an object of type string was given.');
                                                                 ///////
  // - real number

  Test('ToFraction(12)', Rat(12, 1));
  Test('ToFraction(12.0)', Rat(12, 1));
  Test('ToFraction(-123)', Rat(-123, 1));
  Test('ToFraction(0)', Rat(0, 1));
  Test('ToFraction(0.0)', Rat(0, 1));
  Test('ToFraction(0.5)', Rat(1, 2));
  Test('ToFraction(0.25)', Rat(1, 4));
  Test('ToFraction(0.1)', Rat(1, 10));
  Test('ToFraction(0.333333333333333333333333333333333333333333333333333333333333333333333)', Rat(1, 3));
  Test('ToFraction(0.142857142857142857142857142857142857142857142857142857142)', Rat(1, 7));
  Test('ToFraction(17636.57142857142857142857142857142857142857142857142857142)', Rat(123456, 7));
  Test('ToFraction(-173.3333333333333333333333333333333333333333333333333333333)', Rat(-520, 3));
  Test('ToFraction(0.001)', Rat(1, 1000));
  Test('ToFraction(0.228070175438596491228070175438596491228070175438596491228)', Rat(13, 57));
  Test('ToFraction(2831.3125)', Rat(45301, 16));
  Test('ToFraction(290.96875)', Rat(9311, 32));
  Test('ToFraction(0.012409513960703205791106514994829369183040330920372285418)', Rat(12, 967));
  Test('ToFraction(0.029001074113856068743286788399570354457572502685284640171)', Rat(27, 931));
  Test('ToFraction(0.057363848505099860863087086560688892053858036526178555386)', Rat(5236, 91277));
  Test('ToFraction(-0.001279732925302545555710112672137988593684796216441786062)', Rat(-46, 35945));
  Test('ToFraction(-11841.42857142857142857142857142857142857142857142857142857)', Rat(-82890, 7));
  Test('ToFraction(-0.517792)', Rat(-16181, 31250));
  Test('ToFraction(0.999940688018979833926453143534994068801897983392645314353)', Rat(16859, 16860));
  Test('ToFraction(0.036210104479748103621010447974810362101044797481036210104)', Rat(253, 6987));
  Test('ToFraction(0.00390625)', Rat(1, 256));
  Test('ToFraction(-0.0009765625)', Rat(-1, 1024));
  Test('ToFraction(0.464285714285714285714285714285714285714285714285714285714)', Rat(13, 28));
  Test('ToFraction(853/9111)', Rat(853, 9111));
  Test('ToFraction(0.123456)', Rat(1929, 15625));
  Test('ToFraction(-0.123456)', Rat(-1929, 15625));

  Test('type(ToFraction(π))', 'rational number');
  Test('ToFraction(π) ≈ π', True);
  Test('RealNumber(ToFraction(π)) ≈ π', True);

  Test('ToFraction(2 + i)', failure, 'Object isn''t a rational number.');

  Test('ToFraction("dog")', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');
                                                        //////
  Test('ToFraction(5, "dog")', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
                                                           ///////
  Test('ToFraction()', failure, 'Too few arguments. A required argument of type number is missing.');
                                                                                //////
  // To symbolic form

  Test('ToSymbolicForm(−1)', '−1');
  Test('ToSymbolicForm(0)', '0');
  Test('ToSymbolicForm(1)', '1');
  Test('ToSymbolicForm(123456789)', '123456789');
  Test('ToSymbolicForm(123456789123456)', '123456789123456');
  Test('ToSymbolicForm(−123456789123456)', '−123456789123456');

  Test('ToSymbolicForm(1/2)', '1/2');
  Test('ToSymbolicForm(0.5)', '1/2');
  Test('ToSymbolicForm(1/4)', '1/4');
  Test('ToSymbolicForm(0.25)', '1/4');
  Test('ToSymbolicForm(1/3)', '1/3');
  Test('ToSymbolicForm(0.33333333333333333333333333333333)', '1/3');
  Test('ToSymbolicForm(−1/2)', '−1/2');
  Test('ToSymbolicForm(−1/7)', '−1/7');

  Test('ToSymbolicForm(297/311)', '297/311');
  Test('ToSymbolicForm(−1953/1954)', '−1953/1954');

  Test('ToSymbolicForm(0.123456)', '1929/15625');
  Test('ToSymbolicForm(7.145539906103286384976525821596244131455399061032863849765)', '1522/213');
  Test('ToSymbolicForm(0.252747252747252747252747252747252747252747252747252747252)', '23/91');
  Test('ToSymbolicForm(0.525)', '21/40');
  Test('ToSymbolicForm(−0.472222222222222222222222222222222222222222222222222222222)', '−17/36');
  Test('ToSymbolicForm(0.002089864158829676071055381400208986415882967607105538140)', '2/957');
  Test('ToSymbolicForm(0.001582278481012658227848101265822784810126582278481012658)', '1/632');
  Test('ToSymbolicForm(0.00390625)', '1/256');
  Test('ToSymbolicForm(0.000244140625)', '1/4096');
  Test('ToSymbolicForm(3.694026700604197249295943247040441523912728464053728823169)', '586327/158723');
  Test('ToSymbolicForm(−1241.173913043478260869565217391304347826086956521739130434)', '−28547/23');
  Test('ToSymbolicForm(1.3188610316130989277659812985505717262572042783851865E-6)', '1/758230');
  Test('ToSymbolicForm(−0.000031309684085287579448323366417232850120542283728357180)', '−1/31939');
  Test('ToSymbolicForm(1.034502587694077055779183438757906843013225991949396204715)', '1799/1739');
  Test('ToSymbolicForm(14.05174512689387457771302575033856292870378066011215355356)', '2874158/204541');
  Test('ToSymbolicForm(0.000511858044702269237331513393618836376045043507933799692)', '3/5861');
  Test('ToSymbolicForm(0.296296296296296296296296296296296296296296296296296296296)', '8/27');
  Test('ToSymbolicForm(0.034883720930232558139534883720930232558139534883720930232)', '3/86');
  Test('ToSymbolicForm(−0.005446333687566418703506907545164718384697130712008501594)', '−41/7528');
  Test('ToSymbolicForm(0.0001)', '1/10000');
  Test('ToSymbolicForm(3.1415)', '6283/2000');

  Test('ToSymbolicForm(sqrt(2))', '√2');
  Test('ToSymbolicForm(−sqrt(2))', '−√2');
  Test('ToSymbolicForm(5⋅sqrt(2))', '5⋅√2');
  Test('ToSymbolicForm(−3⋅sqrt(2))', '−3⋅√2');
  Test('ToSymbolicForm(sqrt(2)/3)', '√2/3');
  Test('ToSymbolicForm(−sqrt(2)/3)', '−√2/3');
  Test('ToSymbolicForm(1/sqrt(2))', '√2/2'); {would prefer '1/√2'}
  Test('ToSymbolicForm(−1/sqrt(2))', '−√2/2'); {would prefer '−1/√2'}
  Test('ToSymbolicForm(176⋅sqrt(2)/3437)', '(176/3437)⋅√2');
  Test('ToSymbolicForm(−17597⋅sqrt(2)/695)', '(−17597/695)⋅√2');
  Test('ToSymbolicForm(-35.8070734634235303219616064459252619979719654546879931563)', '(−17597/695)⋅√2');

  Test('ToSymbolicForm(sqrt(5))', '√5');
  Test('ToSymbolicForm(sqrt(5)/2)', '√5/2');
  Test('ToSymbolicForm(−sqrt(5)/4)', '−√5/4');
  Test('ToSymbolicForm(2⋅sqrt(5))', '2⋅√5');
  Test('ToSymbolicForm(2⋅sqrt(5)/7)', '(2/7)⋅√5');
  Test('ToSymbolicForm(−2⋅sqrt(5)/795)', '(−2/795)⋅√5');
  Test('ToSymbolicForm(-0.00562532824528248980228722935529880813947325373487176282)', '(−2/795)⋅√5');

  Test('ToSymbolicForm(5⋅sqrt(37)/21)', '(5/21)⋅√37');
  Test('ToSymbolicForm(−10⋅sqrt(53)/99)', '(−10/99)⋅√53');
  Test('ToSymbolicForm(−10/sqrt(53))', '(−10/53)⋅√53'); {would prefer '−10/√53'}
  Test('ToSymbolicForm(-1.37360563948689023982967971538245901769389994010876932799)', '(−10/53)⋅√53'); {would prefer '−10/√53'}

  Test('ToSymbolicForm(π)', 'π');
  Test('ToSymbolicForm(2⋅π)', '2⋅π');
  Test('ToSymbolicForm(7⋅π)', '7⋅π');
  Test('ToSymbolicForm(10⋅π)', '10⋅π');
  Test('ToSymbolicForm(2000⋅π)', '2000⋅π');
  Test('ToSymbolicForm(1000000⋅π)', '1000000⋅π');
  Test('ToSymbolicForm(π/2)', 'π/2');
  Test('ToSymbolicForm(π/3)', 'π/3');
  Test('ToSymbolicForm(π/256)', 'π/256');
  Test('ToSymbolicForm(7⋅π/8)', '(7/8)⋅π');
  Test('ToSymbolicForm(91⋅π/53)', '(91/53)⋅π');
  Test('ToSymbolicForm(1000⋅π/999)', '(1000/999)⋅π');
  Test('ToSymbolicForm(−π)', '−π');
  Test('ToSymbolicForm(−4⋅π)', '−4⋅π');
  Test('ToSymbolicForm(−π/6)', '−π/6');
  Test('ToSymbolicForm(−2⋅π/3)', '(−2/3)⋅π');
  Test('ToSymbolicForm(−8523⋅π/9197)', '(−8523/9197)⋅π');
  Test('ToSymbolicForm(-2.91136176867954852358563765963805622290012773631336597935)', '(−8523/9197)⋅π');

  Test('ToSymbolicForm(π^2)', 'π²');
  Test('ToSymbolicForm(−π^2)', '−π²');
  Test('ToSymbolicForm(3⋅π^2)', '3⋅π²');
  Test('ToSymbolicForm(10000⋅π^2)', '10000⋅π²');
  Test('ToSymbolicForm(−10000⋅π^2)', '−10000⋅π²');
  Test('ToSymbolicForm(π^2/2)', 'π²/2');
  Test('ToSymbolicForm(π^2/1024)', 'π²/1024');
  Test('ToSymbolicForm(3⋅π^2/2)', '(3/2)⋅π²');
  Test('ToSymbolicForm(581⋅π^2/53)', '(581/53)⋅π²');
  Test('ToSymbolicForm(−581⋅π^2/53)', '(−581/53)⋅π²');
  Test('ToSymbolicForm(-108.193210510055044481940363602415920936174704822771685923)', '(−581/53)⋅π²');

  Test('ToSymbolicForm(2/π)', '2⋅π⁻¹');
  Test('ToSymbolicForm(1/(4⋅π))', 'π⁻¹/4');
  Test('ToSymbolicForm(3/(4⋅π))', '(3/4)⋅π⁻¹');
  Test('ToSymbolicForm(−3/(4⋅π))', '(−3/4)⋅π⁻¹');
  Test('ToSymbolicForm(-0.23873241463784300365332564505877154305168946861068467312)', '(−3/4)⋅π⁻¹');

  Test('ToSymbolicForm(e)', 'e');
  Test('ToSymbolicForm(2⋅e)', '2⋅e');
  Test('ToSymbolicForm(−e)', '−e');
  Test('ToSymbolicForm(−10⋅e)', '−10⋅e');
  Test('ToSymbolicForm(1000000⋅e)', '1000000⋅e');
  Test('ToSymbolicForm(e/5)', 'e/5');
  Test('ToSymbolicForm(−e/7)', '−e/7');
  Test('ToSymbolicForm(2⋅e/3)', '(2/3)⋅e');
  Test('ToSymbolicForm(−2⋅e/31735)', '(−2/31735)⋅e');
  Test('ToSymbolicForm(−2⋅e^2/31735)', '(−2/31735)⋅e²');
  Test('ToSymbolicForm(−e^2/7)', '−e²/7');
  Test('ToSymbolicForm(−5⋅e^2/271)', '(−5/271)⋅e²');
  Test('ToSymbolicForm(-0.13632944831975369422934368008440973825055932786996028273)', '(−5/271)⋅e²');

  Test('ToSymbolicForm(1/e)', 'e⁻¹');
  Test('ToSymbolicForm(7/e)', '7⋅e⁻¹');
  Test('ToSymbolicForm(7/(256⋅e))', '(7/256)⋅e⁻¹');
  Test('ToSymbolicForm(0.010059203469531625981127603090352445594221398114149901724)', '(7/256)⋅e⁻¹');

  Test('ToSymbolicForm(10/e + sqrt(2)/71 ⋅ i)',
    TAlgosimStructure.CreateWithValue(
      ['Re', 'Im'],
      [ASO('10⋅e⁻¹'), ASO('√2/71')]
    )
  );

  Test('ToSymbolicForm()', failure, 'Too few arguments. A required argument of type number is missing.');
  Test('ToSymbolicForm("cat")', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');
  Test('ToSymbolicForm(5, 6)', failure, 'Too many arguments.');

  // Carol numbers

  Test('for(n, 1, 30, assert(IsCarolNumber(4^n − 2^(n+1) − 1)))', null);
  Test('filter(SequenceList(-100, 1000000), IsCarolNumber)', intarr([-1, 7, 47, 223, 959, 3967, 16127, 65023, 261119]));
  Test('IsCarolNumber(1125899839733759)', True);
  Test('IsCarolNumber(281474943156223)', True);
  Test('IsCarolNumber(70368727400447)', True);
  Test('IsCarolNumber(17592177655807)', True);
  Test('IsCarolNumber(4398042316799)', True);
  Test('IsCarolNumber(1099509530623)', True);
  Test('IsCarolNumber(274876858367)', True);
  Test('IsCarolNumber(68718952447)', True);
  Test('IsCarolNumber(17179607039)', True);
  Test('IsCarolNumber(4294836223)', True);
  Test('IsCarolNumber(1073676287)', True);
  Test('IsCarolNumber(268402687)', True);
  Test('IsCarolNumber(67092479)', True);
  Test('IsCarolNumber(16769023)', True);
  Test('IsCarolNumber(4190207)', True);
  Test('IsCarolNumber(1046527)', True);

  Test('IsCarolNumber(1115899839733759)', False);
  Test('IsCarolNumber(381474943156223)', False);
  Test('IsCarolNumber(72368727400447)', False);
  Test('IsCarolNumber(17542177655807)', False);
  Test('IsCarolNumber(4398072316799)', False);
  Test('IsCarolNumber(1099529530623)', False);
  Test('IsCarolNumber(274876658367)', False);
  Test('IsCarolNumber(68718952446)', False);
  Test('IsCarolNumber(17179607038)', False);
  Test('IsCarolNumber(4294836224)', False);
  Test('IsCarolNumber(1073676288)', False);
  Test('IsCarolNumber(268402617)', False);
  Test('IsCarolNumber(67092470)', False);
  Test('IsCarolNumber(16769025)', False);
  Test('IsCarolNumber(4190227)', False);
  Test('IsCarolNumber(1046507)', False);
  Test('IsCarolNumber(0)', False);
  Test('IsCarolNumber(1)', False);


  //
  // Special functions
  //

  // Error function

  Test('erf(0)', 0.0);

  Eps(1E-12); Test('erf(0.1)', 0.1124629160182848922032750717439683832216962991597025);
  Eps(1E-12); Test('erf(0.2)', 0.2227025892104784541401390068001438163882690384302276);
  Eps(1E-12); Test('erf(0.5)', 0.5204998778130465376827466538919645287364515757579637);
  Eps(1E-12); Test('erf(0.8)', 0.7421009647076604861671105865029458773176895799147087);
  Eps(1E-12); Test('erf(0.9)', 0.7969082124228321285187247851418859375486580415858037);
  Eps(1E-12); Test('erf(1.0)', 0.842700792949714869341220635082609259296066997966302908459);
  Eps(1E-12); Test('erf(1.5)', 0.9661051464753107270669762616459478586814104792576367);
  Eps(1E-12); Test('erf(2.0)', 0.995322265018952734162069256367252928610891797040060076738);
  Eps(1E-12); Test('erf(5.0)', 0.99999999999846254020557196514981165651461662110988194968);
  Eps(1E-12); Test('erf(10)', 0.999999999999999999999999999999999999999999997911512416237);
  Eps(1E-12); Test('erf(100)', 1.0);
  Eps(1E-12); Test('erf(1000)', 1.0);
  Eps(1E-12); Test('erf(1E6)', 1.0);
  Eps(1E-12); Test('erf(1E50)', 1.0);

  Eps(1E-12); Test('erf(-0.1)', -0.1124629160182848922032750717439683832216962991597025);
  Eps(1E-12); Test('erf(-0.2)', -0.2227025892104784541401390068001438163882690384302276);
  Eps(1E-12); Test('erf(-0.5)', -0.5204998778130465376827466538919645287364515757579637);
  Eps(1E-12); Test('erf(-0.8)', -0.7421009647076604861671105865029458773176895799147087);
  Eps(1E-12); Test('erf(-0.9)', -0.7969082124228321285187247851418859375486580415858037);
  Eps(1E-12); Test('erf(-1.0)', -0.842700792949714869341220635082609259296066997966302908459);
  Eps(1E-12); Test('erf(-1.5)', -0.9661051464753107270669762616459478586814104792576367);
  Eps(1E-12); Test('erf(-2.0)', -0.995322265018952734162069256367252928610891797040060076738);
  Eps(1E-12); Test('erf(-5.0)', -0.99999999999846254020557196514981165651461662110988194968);
  Eps(1E-12); Test('erf(-10)', -0.999999999999999999999999999999999999999999997911512416237);
  Eps(1E-12); Test('erf(-100)', -1.0);
  Eps(1E-12); Test('erf(-1000)', -1.0);
  Eps(1E-12); Test('erf(-1E6)', -1.0);
  Eps(1E-12); Test('erf(-1E50)', -1.0);

  Test('erf()', failure, 'Too few arguments. A required argument of type real number is missing.');
  Test('erf(i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');
  Test('erf(5, 6)', failure, 'Too many arguments.');

  // Complementary error function

  for _i := -100 to 100 do
    Test(Format('erfc(%d/10) = 1 - erf(%d/10)', [_i, _i]), True);

  Test('erfc()', failure, 'Too few arguments. A required argument of type real number is missing.');
  Test('erfc(i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');
  Test('erfc(5, 6)', failure, 'Too many arguments.');

  // Fresnel cosine integral

  Test('FresnelC(0)', 0.0);

  Eps(1E-12); Test('FresnelC(0.1)', 0.0999990000046296189458835350248242780395810402727072);
  Eps(1E-12); Test('FresnelC(0.2)', 0.1999680023702828509150530424733365293665368925309025);
  Eps(1E-12); Test('FresnelC(0.5)', 0.4968840292147947147473274109631733512829027022886211);
  Eps(1E-12); Test('FresnelC(1)', 0.9045242379002720814747883668325571458079918205955017);
  Eps(1E-12); Test('FresnelC(2)', 0.4614614624332163728664741555277730352693828342010849);
  Eps(1E-12); Test('FresnelC(3)', 0.7028635577302687301744098681780622019533373198375706);
  Eps(1E-12); Test('FresnelC(4)', 0.5944603274978229817857606619295196128665370118407959);
  Eps(1E-12); Test('FresnelC(5)', 0.6114667663964626117940160504610222269957333983754062);
  Eps(1E-10); Test('FresnelC(6)', 0.5442040253871845829563310031105696899884509003539315);
  Eps(1E-10); Test('FresnelC(7)', 0.5583343318609714183625036376682713907282436800602031);
  Eps(1E-10); Test('FresnelC(8)', 0.6839570275081139328869052603547318682673629909632490);
  Eps(1E-10); Test('FresnelC(9)', 0.5914010847421237956093145859746907294150313098800105);
  Eps(1E-10); Test('FresnelC(10)', 0.6011251848134443481311911611406102655902173340569311);
{$IFNDEF QuickTest}
  Eps(1E-4); Test('FresnelC(50)', 0.62015814775327796);
  Eps(1E-4); Test('FresnelC(100)', 0.625129463563781268792952450934527);
{$ENDIF}

  Eps(1E-12); Test('FresnelC(-0.1)', -0.0999990000046296189458835350248242780395810402727072);
  Eps(1E-12); Test('FresnelC(-0.2)', -0.1999680023702828509150530424733365293665368925309025);
  Eps(1E-12); Test('FresnelC(-0.5)', -0.4968840292147947147473274109631733512829027022886211);
  Eps(1E-12); Test('FresnelC(-1)', -0.9045242379002720814747883668325571458079918205955017);
  Eps(1E-12); Test('FresnelC(-2)', -0.4614614624332163728664741555277730352693828342010849);
  Eps(1E-12); Test('FresnelC(-3)', -0.7028635577302687301744098681780622019533373198375706);
  Eps(1E-12); Test('FresnelC(-4)', -0.5944603274978229817857606619295196128665370118407959);
  Eps(1E-12); Test('FresnelC(-5)', -0.6114667663964626117940160504610222269957333983754062);
  Eps(1E-10); Test('FresnelC(-6)', -0.5442040253871845829563310031105696899884509003539315);
  Eps(1E-10); Test('FresnelC(-7)', -0.5583343318609714183625036376682713907282436800602031);
  Eps(1E-10); Test('FresnelC(-8)', -0.6839570275081139328869052603547318682673629909632490);
  Eps(1E-10); Test('FresnelC(-9)', -0.5914010847421237956093145859746907294150313098800105);
  Eps(1E-10); Test('FresnelC(-10)', -0.6011251848134443481311911611406102655902173340569311);
{$IFNDEF QuickTest}
  Eps(1E-4); Test('FresnelC(-50)', -0.62015814775327796);
  Eps(1E-4); Test('FresnelC(-100)', -0.625129463563781268792952450934527);
{$ENDIF}

  Test('FresnelC()', failure, 'Too few arguments. A required argument of type real number is missing.');
  Test('FresnelC(i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');
  Test('FresnelC(5, 6)', failure, 'Too many arguments.');

  // Fresnel sine integral

  Test('FresnelS(0)', 0.0);

  Eps(1E-12); Test('FresnelS(0.1)', 0.0003333309523885281253006398044666315606800327779084);
  Eps(1E-12); Test('FresnelS(0.2)', 0.0026663619202766228455069422074347799143665632386424);
  Eps(1E-12); Test('FresnelS(0.5)', 0.0414810242685474815985281654834999914474996368906335);
  Eps(1E-12); Test('FresnelS(1)', 0.3102683017233811018081524231653965075745093888324467);
  Eps(1E-12); Test('FresnelS(2)', 0.8047764893437561102962756017926325942235849454136770);
  Eps(1E-12); Test('FresnelS(3)', 0.7735625268937690171497721678263028632270460209291312);
  Eps(1E-12); Test('FresnelS(4)', 0.7471338446481146561998171240629835783238065841321865);
  Eps(1E-12); Test('FresnelS(5)', 0.5279172811653224138446156871849314538314043858781680);
  Eps(1E-10); Test('FresnelS(6)', 0.6384591893150103778454080609408122940299921139678892);
  Eps(1E-10); Test('FresnelS(7)', 0.6058869316278271054377626517795394767807921931762645);
  Eps(1E-10); Test('FresnelS(8)', 0.6017216449352555816074986779173591595323815044572577);
  Eps(1E-10); Test('FresnelS(9)', 0.5837286611515135900361274584163298717364608203093222);
  Eps(1E-10); Test('FresnelS(10)', 0.5836708999296233421575724092855749812634352727740758);
{$IFNDEF QuickTest}
  Eps(1E-4); Test('FresnelS(50)', 0.61906464542382491);
  Eps(1E-4); Test('FresnelS(100)', 0.63141720902969061149002048488957);
{$ENDIF}

  Eps(1E-12); Test('FresnelS(-0.1)', -0.0003333309523885281253006398044666315606800327779084);
  Eps(1E-12); Test('FresnelS(-0.2)', -0.0026663619202766228455069422074347799143665632386424);
  Eps(1E-12); Test('FresnelS(-0.5)', -0.0414810242685474815985281654834999914474996368906335);
  Eps(1E-12); Test('FresnelS(-1)', -0.3102683017233811018081524231653965075745093888324467);
  Eps(1E-12); Test('FresnelS(-2)', -0.8047764893437561102962756017926325942235849454136770);
  Eps(1E-12); Test('FresnelS(-3)', -0.7735625268937690171497721678263028632270460209291312);
  Eps(1E-12); Test('FresnelS(-4)', -0.7471338446481146561998171240629835783238065841321865);
  Eps(1E-12); Test('FresnelS(-5)', -0.5279172811653224138446156871849314538314043858781680);
  Eps(1E-10); Test('FresnelS(-6)', -0.6384591893150103778454080609408122940299921139678892);
  Eps(1E-10); Test('FresnelS(-7)', -0.6058869316278271054377626517795394767807921931762645);
  Eps(1E-10); Test('FresnelS(-8)', -0.6017216449352555816074986779173591595323815044572577);
  Eps(1E-10); Test('FresnelS(-9)', -0.5837286611515135900361274584163298717364608203093222);
  Eps(1E-10); Test('FresnelS(-10)', -0.5836708999296233421575724092855749812634352727740758);
{$IFNDEF QuickTest}
  Eps(1E-4); Test('FresnelS(-50)', -0.61906464542382491);
  Eps(1E-4); Test('FresnelS(-100)', -0.63141720902969061149002048488957);
{$ENDIF}

  Test('FresnelS()', failure, 'Too few arguments. A required argument of type real number is missing.');
  Test('FresnelS(i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');
  Test('FresnelS(5, 6)', failure, 'Too many arguments.');

  // Sine integral

  Test('Si(0)', 0.0);

  Eps(1E-12); Test('Si(0.1)', 0.099944461108276950160592118554190930431785170581484030555);
  Eps(1E-12); Test('Si(0.2)', 0.199556088526233821400456944764165950986603878805513219862);
  Eps(1E-12); Test('Si(0.5)', 0.493107418043066689161626707572764653641337138428721131660);
  Eps(1E-12); Test('Si(1)', 0.946083070367183014941353313823179657812337954738111790471);
  Eps(1E-12); Test('Si(2)', 1.605412976802694848576720148198588940848583422328499660289);
  Eps(1E-12); Test('Si(3)', 1.848652527999468256397730251111973245164512730309465694979);
  Eps(1E-12); Test('Si(4)', 1.758203138949053058105559303358501617209579460962609792592);
  Eps(1E-12); Test('Si(5)', 1.549931244944674137274408400730639012183184893966372210477);
  Eps(1E-10); Test('Si(6)', 1.424687551280506535769031027917142018492285449190238213174);
  Eps(1E-10); Test('Si(7)', 1.454596614248093590614768493836160420066907406317066317971);
  Eps(1E-10); Test('Si(8)', 1.574186821706942052082971451206658507176789221924863859521);
  Eps(1E-10); Test('Si(9)', 1.665040075829602495106653427897108463186698581348304888159);
  Eps(1E-10); Test('Si(10)', 1.658347594218874049330971879389672480630254348309579842195);
{$IFNDEF QuickTest}
  Eps(1E-4); Test('Si(50)', 1.551617072485935894727985594859377543420004615915493673419);
  Eps(1E-4); Test('Si(100)', 1.562225466889056293352345138804502677227824980541083456384);
{$ENDIF}

  Eps(1E-12); Test('Si(-0.1)', -0.099944461108276950160592118554190930431785170581484030555);
  Eps(1E-12); Test('Si(-0.2)', -0.199556088526233821400456944764165950986603878805513219862);
  Eps(1E-12); Test('Si(-0.5)', -0.493107418043066689161626707572764653641337138428721131660);
  Eps(1E-12); Test('Si(-1)', -0.946083070367183014941353313823179657812337954738111790471);
  Eps(1E-12); Test('Si(-2)', -1.605412976802694848576720148198588940848583422328499660289);
  Eps(1E-12); Test('Si(-3)', -1.848652527999468256397730251111973245164512730309465694979);
  Eps(1E-12); Test('Si(-4)', -1.758203138949053058105559303358501617209579460962609792592);
  Eps(1E-12); Test('Si(-5)', -1.549931244944674137274408400730639012183184893966372210477);
  Eps(1E-10); Test('Si(-6)', -1.424687551280506535769031027917142018492285449190238213174);
  Eps(1E-10); Test('Si(-7)', -1.454596614248093590614768493836160420066907406317066317971);
  Eps(1E-10); Test('Si(-8)', -1.574186821706942052082971451206658507176789221924863859521);
  Eps(1E-10); Test('Si(-9)', -1.665040075829602495106653427897108463186698581348304888159);
  Eps(1E-10); Test('Si(-10)', -1.658347594218874049330971879389672480630254348309579842195);
{$IFNDEF QuickTest}
  Eps(1E-4); Test('Si(-50)', -1.551617072485935894727985594859377543420004615915493673419);
  Eps(1E-4); Test('Si(-100)', -1.562225466889056293352345138804502677227824980541083456384);
{$ENDIF}

  Test('Si()', failure, 'Too few arguments. A required argument of type real number is missing.');
  Test('Si(i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');
  Test('Si(5, 6)', failure, 'Too many arguments.');

  // Cosine integral

  Eps(1E-12); Test('Ci(0.1)', -1.72786838665729663899772515290657363282848171769703311388);
  Eps(1E-12); Test('Ci(0.2)', -1.04220559567278197536291690667588952345312516810593890777);
  Eps(1E-12); Test('Ci(0.5)', -0.17778407880661290133581027107056907809051947481262196866);
  Eps(1E-12); Test('Ci(1)', 0.337403922900968134662646203889150769997578032585731894801);
  Eps(1E-12); Test('Ci(2)', 0.422980828774864995698565153198255894135737756306187688110);
  Eps(1E-12); Test('Ci(3)', 0.119629786008000327626472281176677850546836524987071897103);
  Eps(1E-12); Test('Ci(4)', -0.14098169788693041163914489869403592671296835546169386158);
  Eps(1E-12); Test('Ci(5)', -0.19002974965664387861845890011630080649673915610185662891);
  Eps(1E-10); Test('Ci(6)', -0.06805724389324712620416830484061742841596919013564899125);
  Eps(1E-10); Test('Ci(7)', 0.076695278482184518382915763031436846848903015516736613052);
  Eps(1E-10); Test('Ci(8)', 0.122433882532009557292295958268735027035373787175898220629);
  Eps(1E-10); Test('Ci(9)', 0.055347531333133607085641648449793252002663554318911601082);
  Eps(1E-10); Test('Ci(10)', -0.04545643300445537263453282995262785288764695795731688693);
{$IFNDEF QuickTest}
  Eps(1E-4); Test('Ci(50)', -0.00562838632411630544018589549846409957473375839534749948);
  Eps(1E-4); Test('Ci(100)', -0.00514882514261049214444355390534449785032633791932324580);
{$ENDIF}


  Test('Ci(0)', failure, 'Cosine integral only defined for positive real numbers.');
  Test('Ci(-1)', failure, 'Cosine integral only defined for positive real numbers.');
  Test('Ci(-1000)', failure, 'Cosine integral only defined for positive real numbers.');

  Test('Ci()', failure, 'Too few arguments. A required argument of type real number is missing.');
  Test('Ci(i)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');
  Test('Ci(5, 6)', failure, 'Too many arguments.');

  // Bessel functions

  Test('Bessel(0, 0)', 1.0);

  Eps(1E-12); Test('Bessel(0, 0.1)', 0.9975015620660400322812868984747920848320082326843663);
  Eps(1E-12); Test('Bessel(0, 0.2)', 0.9900249722395763908175076200832671191563303563557559);
  Eps(1E-12); Test('Bessel(0, 0.5)', 0.9384698072408129042284046735997126255689267970968215);
{$IFNDEF QuickTest} // Bessel functions don't use any integration cache; should implement that
  Eps(1E-12); Test('Bessel(0, 1)', 0.765197686557966551449717526102663220909274289755325241861);
  Eps(1E-12); Test('Bessel(0, 2)', 0.223890779141235668051827454649948625825154482218607603128);
  Eps(1E-12); Test('Bessel(0, 3)', -0.26005195490193343762415469597733143681960865351129327705);
  Eps(1E-12); Test('Bessel(0, 4)', -0.39714980986384737228659076845169804197561868528938884014);
  Eps(1E-12); Test('Bessel(0, 5)', -0.17759677131433830434739701307475871107113035600850912899);
  Eps(1E-12); Test('Bessel(0, 10)', -0.24593576445134833519776086248532875382960007282656656969);
  Eps(1E-12); Test('Bessel(0, 15)', -0.01422447282678077323386427061184007334439812577187924319);
  Eps(1E-12); Test('Bessel(0, 20)', 0.167024664340583154727320544701384038875333378408533084201);

  Eps(1E-12); Test('Bessel(0, -0.1)', 0.9975015620660400322812868984747920848320082326843663);
  Eps(1E-12); Test('Bessel(0, -0.2)', 0.9900249722395763908175076200832671191563303563557559);
  Eps(1E-12); Test('Bessel(0, -0.5)', 0.9384698072408129042284046735997126255689267970968215);
  Eps(1E-12); Test('Bessel(0, -1)', 0.765197686557966551449717526102663220909274289755325241861);
  Eps(1E-12); Test('Bessel(0, -2)', 0.223890779141235668051827454649948625825154482218607603128);
  Eps(1E-12); Test('Bessel(0, -3)', -0.26005195490193343762415469597733143681960865351129327705);
  Eps(1E-12); Test('Bessel(0, -4)', -0.39714980986384737228659076845169804197561868528938884014);
  Eps(1E-12); Test('Bessel(0, -5)', -0.17759677131433830434739701307475871107113035600850912899);
  Eps(1E-12); Test('Bessel(0, -10)', -0.24593576445134833519776086248532875382960007282656656969);
  Eps(1E-12); Test('Bessel(0, -15)', -0.01422447282678077323386427061184007334439812577187924319);
  Eps(1E-12); Test('Bessel(0, -20)', 0.167024664340583154727320544701384038875333378408533084201);
{$ENDIF}

  Test('Bessel(1, 0)', 0.0);

  Eps(1E-12); Test('Bessel(1, 0.1)', 0.0499375260362419975563365524378064840585613541415665);
  Eps(1E-12); Test('Bessel(1, 0.2)', 0.0995008326392359953979227693278841598213067853770768);
  Eps(1E-12); Test('Bessel(1, 0.5)', 0.2422684576748738863839545761415316408006286544379597);
{$IFNDEF QuickTest}
  Eps(1E-12); Test('Bessel(1, 1)', 0.440050585744933515959682203718914913127372301992765251136);
  Eps(1E-12); Test('Bessel(1, 2)', 0.576724807756873387202448242269137086920302689719675440121);
  Eps(1E-12); Test('Bessel(1, 3)', 0.339058958525936458925514597206478896973080418198008175481);
  Eps(1E-12); Test('Bessel(1, 4)', -0.06604332802354913614318542080327502872742341953170756044);
  Eps(1E-12); Test('Bessel(1, 5)', -0.32757913759146522203773432191016913276084990462405401868);
  Eps(1E-12); Test('Bessel(1, 10)', 0.043472746168861436669748768025859288306272867118594208135);
  Eps(1E-12); Test('Bessel(1, 15)', 0.205104038613522761147137412076923635363529311493888391999);
  Eps(1E-12); Test('Bessel(1, 20)', 0.066833124175850045578992974193646719982994444621062459423);

  Eps(1E-12); Test('Bessel(1, -0.1)', -0.0499375260362419975563365524378064840585613541415665);
  Eps(1E-12); Test('Bessel(1, -0.2)', -0.0995008326392359953979227693278841598213067853770768);
  Eps(1E-12); Test('Bessel(1, -0.5)', -0.2422684576748738863839545761415316408006286544379597);
  Eps(1E-12); Test('Bessel(1, -1)', -0.440050585744933515959682203718914913127372301992765251136);
  Eps(1E-12); Test('Bessel(1, -2)', -0.576724807756873387202448242269137086920302689719675440121);
  Eps(1E-12); Test('Bessel(1, -3)', -0.339058958525936458925514597206478896973080418198008175481);
  Eps(1E-12); Test('Bessel(1, -4)', 0.06604332802354913614318542080327502872742341953170756044);
  Eps(1E-12); Test('Bessel(1, -5)', 0.32757913759146522203773432191016913276084990462405401868);
  Eps(1E-12); Test('Bessel(1, -10)', -0.043472746168861436669748768025859288306272867118594208135);
  Eps(1E-12); Test('Bessel(1, -15)', -0.205104038613522761147137412076923635363529311493888391999);
  Eps(1E-12); Test('Bessel(1, -20)', -0.066833124175850045578992974193646719982994444621062459423);
{$ENDIF}

  Eps(1E-12); Test('Bessel(1, i)', 0.565159103992485027207696027609863307328899621621092009480*ImaginaryUnit);
  Eps(1E-12); Test('Bessel(2, i)', TASC(-0.13574766976703828118285256999499092294987106811277818784));
  Eps(1E-12); Test('Bessel(2, 1 + i)', 0.04157988694396212208283336257837005994406413253165264166 + 0.2473976415133063105105919644784222846179418163179320472*ImaginaryUnit);
  Eps(1E-12); Test('Bessel(3, 2 - i/3)', 0.1243481479093985857655785476068830597554279374515206589 - 0.05370855043413167121050052711946281700829906437960445723*ImaginaryUnit);

  // Laguerre polynomials

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Laguerre(0, %d)', [_i]), 1.0);
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Laguerre(1, %d)', [_i]), 1.0 - _i);
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Laguerre(2, %d)', [_i]), (1/2) * (_i*_i - 4*_i + 2));
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Laguerre(3, %d)', [_i]), (1/6) * (-_i*_i*_i + 9*_i*_i - 18*_i + 6));
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Laguerre(4, %d)', [_i]), (1/24) * (_i*_i*_i*_i - 16*_i*_i*_i + 72*_i*_i - 96*_i + 24));
  end;

  Test('Laguerre(-1, 1)', failure, 'Laguerre polynomial not defined for negative n.');

  Eps; Test('Laguerre(0, 1 + i)', TASC(1.0));
  Eps; Test('Laguerre(1, 2 + 3⋅i)', -1 - 3*ImaginaryUnit);
  Eps; Test('Laguerre(2, 5 - 2⋅i)', 3/2 - 6*ImaginaryUnit);
  Eps; Test('Laguerre(3, 7 - 10⋅i)', 589/3 - 305/3 * ImaginaryUnit);
  Eps; Test('Laguerre(4, -1 + 2⋅i)', -93/8 - 53/3 * ImaginaryUnit);
  Eps; Test('Laguerre(5, -2 - i)', 3307/120 + 1947/40 * ImaginaryUnit);
  Eps; Test('Laguerre(6, 5 - i/2)', -39007/15360 - 7237/11520 * ImaginaryUnit);
  Eps; Test('Laguerre(7, -i/3)', -78037/524880 + 23345657/11022480 * ImaginaryUnit);

  // (Physicists') Hermite polynomials

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Hermite(0, %d)', [_i]), 1.0);
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Hermite(1, %d)', [_i]), 2.0*_i);
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Hermite(2, %d)', [_i]), 4.0*_i*_i - 2);
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Hermite(3, %d)', [_i]), 8.0*_i*_i*_i - 12*_i);
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Hermite(4, %d)', [_i], DefaultFormatSettings), 16.0*_i*_i*_i*_i - 48*_i*_i + 12);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Hermite(5, %g)', [_x], DefaultFormatSettings), 32*_x*_x*_x*_x*_x - 160*_x*_x*_x + 120*_x);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Hermite(6, %g)', [_x], DefaultFormatSettings), 64*_x*_x*_x*_x*_x*_x - 480*_x*_x*_x*_x + 720*_x*_x - 120);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Hermite(7, %g)', [_x], DefaultFormatSettings), 128*_x*_x*_x*_x*_x*_x*_x - 1344*_x*_x*_x*_x*_x + 3360*_x*_x*_x - 1680*_x);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Hermite(8, %g)', [_x], DefaultFormatSettings), 256*_x*_x*_x*_x*_x*_x*_x*_x - 3584*_x*_x*_x*_x*_x*_x + 13440*_x*_x*_x*_x - 13440*_x*_x + 1680);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Hermite(9, %g)', [_x], DefaultFormatSettings), 512*_x*_x*_x*_x*_x*_x*_x*_x*_x - 9216*_x*_x*_x*_x*_x*_x*_x + 48384*_x*_x*_x*_x*_x - 80640*_x*_x*_x + 30240*_x);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Hermite(10, %g)', [_x], DefaultFormatSettings), 1024*_x*_x*_x*_x*_x*_x*_x*_x*_x*_x - 23040*_x*_x*_x*_x*_x*_x*_x*_x + 161280*_x*_x*_x*_x*_x*_x - 403200*_x*_x*_x*_x + 302400*_x*_x - 30240);
  end;

  Test('Hermite(-1, 1)', failure, 'Hermite polynomial only implemented for non-negative n ≤ 10.');
  Test('Hermite(11, 1)', failure, 'Hermite polynomial only implemented for non-negative n ≤ 10.');

  Eps; Test('Hermite(0, 1 + i)', TASC(1.0));
  Eps; Test('Hermite(1, 2 + i/5)', 4 + 2/5 * ImaginaryUnit);
  Eps; Test('Hermite(2, 2 - 3⋅i)', -22 - 48 * ImaginaryUnit);
  Eps; Test('Hermite(3, -1 + 5⋅i)', 604 - 940 * ImaginaryUnit);
  Eps; Test('Hermite(4, -2 + 7⋅i)', 22028 + 41664 * ImaginaryUnit);
  Eps; Test('Hermite(5, -2 - i)', 1296 + 328 * ImaginaryUnit);
  Eps; Test('Hermite(6, -4 - i/10)', 2322835749/15625 + 86048048/3125 * ImaginaryUnit);
  Eps; Test('Hermite(7, -4 - 3⋅i)', -1969472 + 9078096 * ImaginaryUnit);
  Eps; Test('Hermite(8, -1 - i)', -47984 + 1792 * ImaginaryUnit);
  Eps; Test('Hermite(9, 1/2 - i/3)', 10872793/729 + 120515662/19683 * ImaginaryUnit);
  Eps; Test('Hermite(10, -i)', TASC(-921184));

  // Legendre polynomials

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Legendre(0, %d)', [_i]), 1.0);
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Legendre(1, %d)', [_i]), TASR(_i));
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Legendre(2, %d)', [_i]), (1/2) * (3*_i*_i - 1));
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Legendre(3, %d)', [_i]), (1/2) * (5*_i*_i*_i - 3*_i));
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Legendre(4, %g)', [_x], DefaultFormatSettings), (1/8) * (35*_x*_x*_x*_x - 30*_x*_x + 3));
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Legendre(5, %g)', [_x], DefaultFormatSettings), (1/8) * (63*_x*_x*_x*_x*_x - 70*_x*_x*_x + 15*_x));
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Legendre(6, %g)', [_x], DefaultFormatSettings), (1/16) * (231*_x*_x*_x*_x*_x*_x - 315*_x*_x*_x*_x + 105*_x*_x - 5));
  end;

  Test('Legendre(-1, 1)', failure, 'Legendre polynomial not defined for negative n.');

  Eps; Test('Legendre(0, 1 + i)', TASC(1.0));
  Eps; Test('Legendre(1, 2 + i)', 2 + ImaginaryUnit);
  Eps; Test('Legendre(3, 5 + 3⋅i)', -65/2 + 981/2 * ImaginaryUnit);
  Eps; Test('Legendre(4, 7 - 5⋅i)', -152057/8 - 28875/2 * ImaginaryUnit);
  Eps; Test('Legendre(5, -2 + 3⋅i)', -1367 - 9549/2 * ImaginaryUnit);
  Eps; Test('Legendre(6, -8 - 2⋅i)', 6512935/16 + 4441122 * ImaginaryUnit);
  Eps; Test('Legendre(7, -2⋅i/3)', 84941/5832 * ImaginaryUnit);
  Eps; Test('Legendre(8, 2 - i/5)', 3136050367/625000 - 100528071/15625 * ImaginaryUnit);
  Eps; Test('Legendre(9, 1 + i/3)', 618947/93312 - 99724117/2519424 * ImaginaryUnit);
  Eps; Test('Legendre(10, -1 - i)', -1630263/256 + 382129/128 * ImaginaryUnit);

  // Chebyshev polynomials of the first kind

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Chebyshev(0, %d)', [_i]), 1.0);
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Chebyshev(1, %d)', [_i]), TASR(_i));
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Chebyshev(2, %d)', [_i]), 2.0*_i*_i - 1);
  end;

  for _i := -100 to 100 do
  begin
    Eps; Test(Format('Chebyshev(3, %d)', [_i]), 4.0*_i*_i*_i - 3*_i);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Chebyshev(4, %g)', [_x], DefaultFormatSettings), 8*_x*_x*_x*_x - 8*_x*_x + 1);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Chebyshev(5, %g)', [_x], DefaultFormatSettings), 16*_x*_x*_x*_x*_x - 20*_x*_x*_x + 5*_x);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Chebyshev(6, %g)', [_x], DefaultFormatSettings), 32*_x*_x*_x*_x*_x*_x - 48*_x*_x*_x*_x + 18*_x*_x - 1);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Chebyshev(7, %g)', [_x], DefaultFormatSettings), 64*_x*_x*_x*_x*_x*_x*_x - 112*_x*_x*_x*_x*_x + 56*_x*_x*_x - 7*_x);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Chebyshev(8, %g)', [_x], DefaultFormatSettings), 128*_x*_x*_x*_x*_x*_x*_x*_x - 256*_x*_x*_x*_x*_x*_x + 160*_x*_x*_x*_x - 32*_x*_x + 1);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Chebyshev(9, %g)', [_x], DefaultFormatSettings), 256*_x*_x*_x*_x*_x*_x*_x*_x*_x - 576*_x*_x*_x*_x*_x*_x*_x + 432*_x*_x*_x*_x*_x - 120*_x*_x*_x + 9*_x);
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Chebyshev(10, %g)', [_x], DefaultFormatSettings), 512*_x*_x*_x*_x*_x*_x*_x*_x*_x*_x - 1280*_x*_x*_x*_x*_x*_x*_x*_x + 1120*_x*_x*_x*_x*_x*_x - 400*_x*_x*_x*_x + 50*_x*_x - 1);
  end;

  Eps; Test('Chebyshev(0, 1 + i)', TASC(1.0));
  Eps; Test('Chebyshev(1, 2 + 3⋅i)', 2 + 3*ImaginaryUnit);
  Eps; Test('Chebyshev(2, 5 + 7⋅i)', -49 + 140*ImaginaryUnit);
  Eps; Test('Chebyshev(3, 5 - 2⋅i)', 245 - 562*ImaginaryUnit);
  Eps; Test('Chebyshev(4, -1 + 6⋅i)', 8929 + 6816*ImaginaryUnit);
  Eps; Test('Chebyshev(5, -3 - i)', 537 - 4541*ImaginaryUnit);
  Eps; Test('Chebyshev(6, -2 - i/3)', 471469/729 + 112280/81 * ImaginaryUnit);
  Eps; Test('Chebyshev(7, -1 - i)', -841 - 41 * ImaginaryUnit);
  Eps; Test('Chebyshev(8, 1/2 + i)', -1553/2 - 280 * ImaginaryUnit);
  Eps; Test('Chebyshev(9, -i/2)', -38 * ImaginaryUnit);
  Eps; Test('Chebyshev(10, 1/2 + 5⋅i/2)', 2221687 + 8046225 * ImaginaryUnit);
  Eps; Test('Chebyshev(10, 1/10 + i/5)', -40668013/19531250 + 29343942/9765625 * ImaginaryUnit);

  // Bernstein polynomials

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Bernstein(0, 3, %g)', [_x], DefaultFormatSettings), pow(1 - _x, 3));
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Bernstein(1, 3, %g)', [_x], DefaultFormatSettings), 3*_x*pow(1 - _x, 2));
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Bernstein(2, 3, %g)', [_x], DefaultFormatSettings), 3*_x*_x*(1-_x));
  end;

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Bernstein(3, 3, %g)', [_x], DefaultFormatSettings), _x*_x*_x);
  end;

  Test('Bernstein(4, 3, 10)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');
  Test('Bernstein(-1, 3, 10)', failure, 'Binomial coefficient C(n, k) only defined for integers n and k satisfying 0 ≤ k ≤ n.');

  Test('Bernstein(2, 3)', failure, 'Too few arguments. A required argument of type number is missing.');

  for _i := -100 to 100 do
  begin
    _x := _i / 10;
    Eps; Test(Format('Bernstein(6, 10, %g)', [_x], DefaultFormatSettings), 210*pow(_x, 6)*pow(1 - _x, 4));
  end;

  // Gamma function

  for _i := 1 to 100 do
    Test(Format('gamma(%d)', [_i]), factorial(_i - 1));

  Eps(1E-12); Test('gamma(0.5)', Sqrt(Pi));
  Eps(1E-12); Test('gamma(1.5)', Sqrt(Pi) / 2);
  Eps(1E-12); Test('gamma(2.5)', 3 * Sqrt(Pi) / 4);
  Eps(1E-11); Test('gamma(3.5)', 15 * Sqrt(Pi) / 8);
  Eps(1E-10); Test('gamma(4.5)', 105 * Sqrt(Pi) / 16);
  Eps(1E-9); Test('gamma(5.5)', 945 * Sqrt(Pi) / 32);
  Eps(1E-7); Test('gamma(6.5)', 10395 * Sqrt(Pi) / 64);
  Eps(1E-6); Test('gamma(7.5)', 135135 * Sqrt(Pi) / 128);
  Eps(1E-5); Test('gamma(8.5)', 2027025 * Sqrt(Pi) / 256);
  Eps(1E-4); Test('gamma(9.5)', 34459425 * Sqrt(Pi) / 512);
  Eps(1E-4); Test('gamma(10.5)', 654729075 * Sqrt(Pi) / 1024);

  Eps(1E-12); Test('gamma(-0.5)', -2*Sqrt(Pi));
  Eps(1E-12); Test('gamma(-1.5)', 4*Sqrt(Pi)/3);
  Eps(1E-12); Test('gamma(-2.5)', -8*Sqrt(Pi)/15);
  Eps(1E-12); Test('gamma(-3.5)', 16*Sqrt(Pi)/105);
  Eps(1E-12); Test('gamma(-4.5)', -32*Sqrt(Pi)/945);
  Eps(1E-12); Test('gamma(-5.5)', 64*Sqrt(Pi)/10395);
  Eps(1E-12); Test('gamma(-6.5)', -128*Sqrt(Pi)/135135);
  Eps(1E-12); Test('gamma(-7.5)', 256*Sqrt(Pi)/2027025);
  Eps(1E-12); Test('gamma(-8.5)', -512*Sqrt(Pi)/34459425);
  Eps(1E-12); Test('gamma(-9.5)', 1024*Sqrt(Pi)/654729075);

  for _i := 0 downto -10 do
    Test(Format('gamma(%d)', [_i]), failure, 'Gamma function not defined for non-positive integers.');

  Eps(1E-12); Test('gamma(i)', -0.1549498283018106851249551304838866051958796520793249302 - 0.4980156681183560427136911174621980919529629675876500928*ImaginaryUnit);
  Eps(1E-12); Test('gamma(1 + i)', 0.49801566811835604271369111746219809195296296758765009289 - 0.15494982830181068512495513048388660519587965207932493026*ImaginaryUnit);
  Eps(1E-12); Test('gamma(2 + 3⋅i)', -0.0823952726656118836738703143646259774892907379038426752 + 0.0917742874352593145956674172937769177383779146310396588*ImaginaryUnit);
  Eps(1E-12); Test('gamma(5 - 3⋅i/2)', -12.319673342713806765304246118251878830714082981365460957 - 14.189577272950092932314465054480288923528977805372384971*ImaginaryUnit);
  Eps(1E-12); Test('gamma(-3 + 7⋅i)', 3.888686566766278461568905334904512455202847233180156E-8 + 1.145975420127756784498523571901256179427475536073713E-8*ImaginaryUnit);
  Eps(1E-12); Test('gamma(-2 - i/4)', 0.41931135601757055406034043510793363519171658454211308328 + 1.7820281003262372417054411133763631011586288081694451113*ImaginaryUnit);
  Eps(1E-12); Test('gamma(-i/4)', -0.5241053229606349827871712420106509076850207235767397346 + 3.767162961395262571349547483497653734888630754234884715*ImaginaryUnit);

  // Harmonic numbers

  Test('HarmonicNumber(0)', 0.0);

  Test('HarmonicNumber(1)', 1.0);

  Test('HarmonicNumber(2)', 1 + 1/2);
  Test('HarmonicNumber(3)', 1 + 1/2 + 1/3);
  Test('HarmonicNumber(4)', 1 + 1/2 + 1/3 + 1/4);
  Test('HarmonicNumber(5)', 1 + 1/2 + 1/3 + 1/4 + 1/5);
  Test('HarmonicNumber(6)', 1 + 1/2 + 1/3 + 1/4 + 1/5 + 1/6);
  Test('HarmonicNumber(7)', 1 + 1/2 + 1/3 + 1/4 + 1/5 + 1/6 + 1/7);
  Test('HarmonicNumber(8)', 1 + 1/2 + 1/3 + 1/4 + 1/5 + 1/6 + 1/7 + 1/8);
  Test('HarmonicNumber(9)', 1 + 1/2 + 1/3 + 1/4 + 1/5 + 1/6 + 1/7 + 1/8 + 1/9);
  Test('HarmonicNumber(10)', 1 + 1/2 + 1/3 + 1/4 + 1/5 + 1/6 + 1/7 + 1/8 + 1/9 + 1/10);

  for _i := 1 to 1000 do
    Test(Format('HarmonicNumber(%d) = HarmonicNumber(%d - 1) + 1/%d', [_i, _i, _i]), True);

  Eps; Test('HarmonicNumber(100)', 5.187377517639620260805117675658253157908972126708451653176);
  Eps; Test('HarmonicNumber(1000)', 7.485470860550344912656518204333900176521679169708803665773);
  Eps; Test('HarmonicNumber(10000)', 9.787606036044382264178477904851605334859262945577691718389);
  Eps; Test('HarmonicNumber(100000)', 12.09014612986342794736321936350421950079369894178220110162);
  Eps; Test('HarmonicNumber(1000000)', 14.39272672286572363138112749318858767664480001374431165341);
  Eps(1E-14); Test('HarmonicNumber(10000000)', 16.69531136585985181539911893954045188424986975237308046278);
  Eps(1E-14); Test('HarmonicNumber(100000000)', 18.99789641385389832441711039422398284185097124497010343881);

{$IFNDEF QuickTest}
  Eps(1E-14); Test('HarmonicNumber(1000000000)', 21.30048150234794401668510184890834696612707273359888038310);
  Eps(1E-14); Test('HarmonicNumber(1000000001)', 21.30048150334794401568510184990834696512707273459888038210);
  Eps(1E-14); Test('HarmonicNumber(1000000002)', 21.30048150434794401368510185390834695712707275059888035010);
  Eps(1E-14); Test('HarmonicNumber(1000000003)', 21.30048150534794401068510186290834693012707283159888010710);
{$ENDIF}

  Eps(1E-8); Test('HarmonicNumber(2000000000)', 21.99362868265788932616483397036652353419476036795913563722);

  Eps(1E-10); Test('HarmonicNumber(1E10)', 23.60306659489198970078559330359271117371984172222765335913);
  Eps(1E-10); Test('HarmonicNumber(2E10)', 24.29621377542693501020345042505088774179534107533790861325);
  Eps(1E-10); Test('HarmonicNumber(5E10)', 25.21250450728609007538715263681889881324544224449617108105);
  Eps(1E-10); Test('HarmonicNumber(1E11)', 25.90565168784103538480440975827707538132094237760642633517);
  Eps(1E-12); Test('HarmonicNumber(1E12)', 28.20823678083058106882240946296143958892204386615187431120);
  Eps(1E-12); Test('HarmonicNumber(1E13)', 30.51082187382417675284040100014580379652314535478063895473);
  Eps(1E-12); Test('HarmonicNumber(1E14)', 32.81340696681817743685839245565516800412424684340941192993);
  Eps(1E-12); Test('HarmonicNumber(1E15)', 35.11599205981221862087638391034778221172534833203818490597);
  Eps(1E-14); Test('HarmonicNumber(1E16)', 37.41857715280626385489437536503222891932644982066695788200);
  Eps(1E-14); Test('HarmonicNumber(1E17)', 39.72116224580030949391236681971659395192755130929573085803);
  Eps(1E-14); Test('HarmonicNumber(1E18)', 42.02374733879435517343035827440095816777865279792450383407);
  Eps(1E-14); Test('HarmonicNumber(1E19)', 44.32633243178840085699834972908532237546225428655327681010);
  Eps(1E-14); Test('HarmonicNumber(1E20)', 46.62891752478244654097134118376968658306418077518204978613);

  Eps(1E-14); Test('HarmonicNumber(1E50)', 115.7064703146038170615060848243006128110972337673785774004);
  Eps(1E-14); Test('HarmonicNumber(1E100)', 230.8357249643061012624056575585188231911523081988172212021);
  Eps(1E-14); Test('HarmonicNumber(1E200)', 461.0942342637106696642048030269552439512624570616945188054);
  Eps(1E-14); Test('HarmonicNumber(1E500)', 1151.869762161924374869602239432264506231592903650326411615);

  Test('HarmonicNumber(-1)', failure, 'To compute the Nth harmonic number, N must be a non-negative integer.');
  Test('HarmonicNumber(-2)', failure, 'To compute the Nth harmonic number, N must be a non-negative integer.');
  Test('HarmonicNumber(-1E50)', failure);
  Test('HarmonicNumber(1.5)', failure);
  Test('HarmonicNumber(-1.5)', failure);

  Test('HarmonicNumber()', failure, 'Too few arguments. A required argument of type number is missing.');
  Test('HarmonicNumber("cat")', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');
  Test('HarmonicNumber(1, 2)', failure, 'Too many arguments.');
  Test('HarmonicNumber(1 + i)', failure, 'Object isn''t a real number.');


  //
  // Parameterisable container aggregate functions.
  // Sums and products. Max and min. Averages (means).
  // (Tests using parameterisation also test non-standard expression evaluation.)
  //

  // Parameterised values: real numbers (and integers and rational numbers)

  // - sum

  Test('∑(n, n, 1, 5)', 15);
  Test('∑(n, n, 1, 1000)', 500500);
  Test('∑(n, n, -1000, 1000)', 0); // cancellation
  Test('∑(n, n, -1000, 1002)', 2003);
  Test('∑(n, n, -1001, 1000)', -1001);
  Test('∑(n, n, 1, 2)', 3);
  Test('∑(n, n, 1, 1)', 1);
  Test('∑(n, n, 123456, 123456)', 123456);

  Test('∑(n, n, 123457, 123456)', 0); // empty sum
  Test('∑(n, n, 1, 0)', 0);
  Test('∑(n, n, 10, 5)', 0);
  Test('∑(n, n, 1, -5)', 0);
  Test('∑(n, n, 1000000, -1000000)', 0);
  Test('∑(n, n, 200000000, -100000000)', 0);
  Test('∑(n, n, 1000000000000000000, -1000000000000000000)', 0); // returns 0 immediately (fast)

  Test('∑(n, n, -1000000000000000000, 1000000000000000000)', failure, 'Too many values for the index variable.');

  Test('∑(n, n, 1000000000000000000, 1000000000000000000)', 1000000000000000000); // large indices work
  Test('∑(n, n, 1000000000000000000, 1000000000000000001)', 2000000000000000001);

  Test('∑(n^2, n, 1, 1000)', 333833500);
  Test('∑(n!, n, 1, 10)', 4037913);
  Test('∑(n!, n, -10, 10)', failure, 'A non-negative integer was expected as argument 1, but "-10" was given.');
  Test('∑(prime(n), n, 1, 100)', 24133);
  Test('∑(prime(n), n, -100, 100)', failure, 'To obtain the Nth prime number, N must be a positive integer.');

  Eps(    ); Test('∑(2^−k, k, 1, 1000)', 1.0); // geometric series, q = 1/2
  Eps(1E-4); Test('∑(1/n^2, n, 1, 10000)', Pi*Pi/6);
  Eps(    ); Test('∑(1/n!, n, 0, 100)', Exp(1)); // Taylor: exp(1)
  Eps(1E-4); Test('∑((−1)^k/(2⋅k + 1), k, 0, 10000)', Pi/4); // Taylor: arctan(1)
  Eps(    ); Test('∑((−1)^k ⋅ (π/6)^(2⋅k+1)/(2⋅k+1)!, k, 0, 100)', 0.5); // Taylor: sin(pi/6)
  Eps(1E-4); Test('∑((−1)^(k+1)/k, k, 1, 10000)', Ln(2)); // Taylor: ln(2)
  Eps(    ); Test('∑(1/(2^k⋅k), k, 1, 100)', Ln(2));

  Test('n ≔ 6; ∑(binomial(n, k), k, 0, n)', 64);
  Test('n ≔ 10; ∑(binomial(n, k), k, 0, n)', 1024);

  Test('∑(Iverson(IsPrime(n)), n, 1, 100) = PrimePi(100)', True);
  Test('∑(Iverson(IsPrime(n)), n, 1, 500) = PrimePi(500)', True);
  Test('∑(Iverson(IsPrime(n)), n, 1, 1000) = PrimePi(1000)', True);
  Test('∑(Iverson(IsPrime(n)), n, 1, 10000) = PrimePi(10000)', True);

  Test('∑(1/n, n, 1, 50)', HarmonicNumber(50));
  Test('∑(1/n, n, 1, 100)', HarmonicNumber(100));
  Test('∑(1/n, n, 1, 1000)', HarmonicNumber(1000));

  Test('∑(MöbiusMu(k), k, 1, 100) = Mertens(100)', True);
  Test('∑(MöbiusMu(k), k, 1, 101) = Mertens(101)', True);
  Test('∑(MöbiusMu(k), k, 1, 102) = Mertens(102)', True);
  Test('∑(MöbiusMu(k), k, 1, 103) = Mertens(103)', True);
  Test('∑(MöbiusMu(k), k, 1, 104) = Mertens(104)', True);
  Test('∑(MöbiusMu(k), k, 1, 105) = Mertens(105)', True);
  Test('∑(MöbiusMu(k), k, 1, 106) = Mertens(106)', True);
  Test('∑(MöbiusMu(k), k, 1, 107) = Mertens(107)', True);
  Test('∑(MöbiusMu(k), k, 1, 108) = Mertens(108)', True);
  Test('∑(MöbiusMu(k), k, 1, 109) = Mertens(109)', True);

  Test('n ≔ 50; ∑(Fibonacci(k), k, 1, n) = Fibonacci(n + 2) − 1', True);
  Test('n ≔ 100; ∑(Fibonacci(k), k, 1, n) = Fibonacci(n + 2) − 1', True);
  Test('n ≔ 150; ∑(Fibonacci(k), k, 1, n) = Fibonacci(n + 2) − 1', True);
  Test('n ≔ 200; ∑(Fibonacci(k), k, 1, n) = Fibonacci(n + 2) − 1', True);

  Test('∑(∑(k, k, 1, n), n, 1, 100)', 171700); // sum of sum (variable in upper bound)
  Test('∑(∑(a⋅k, k, 1, 10), a, 1, 20)', 11550); // sum of sum (variable in summand)
  Test('∑(∑(a⋅k, k, 1, a), a, 1, 10)', 1705); // sum of sum (variable in summand and upper bound)
  Test('∑(∑(a⋅k, k, 1, a), a, 1, 100)', 12920425);

  Test('∑(∑(∑(a⋅b⋅k, k, 1, b), a, 1, b), b, 1, 10)', 68629); // sum of sum of sum (mixed positions)
  Test('∑(∑(∑(a⋅b⋅k, k, 1, b), a, 1, b), b, 9, 10)', 48475);
  Test('∑(∑(∑(a⋅b⋅k, k, 1, b), a, 1, b), b, 3, 13)', 297187);
  Test('∑(∑(∑(a⋅b⋅k, k, 1, b), a, 1, b), b, -10, 10)', 68629); // Empty sums don't alter the value
  Test('∑(∑(∑(a⋅b⋅k, k, 1, b), a, 1, b), b, -10, -3)', 0); // Empty sums

  Test('∑(∑(∑(a⋅b⋅k, k, 1, b), a, 1, b), b, 1, 10)', 68629);
  Test('∑(∑(∑(a⋅b⋅k, k, 1, b), a, 1, b), b, 1, 20)', 3455683);
  Test('∑(∑(∑(a⋅b⋅k, k, 1, b), a, 1, b), b, 1, 30)', 36187912);

  Test('f ≔ n ↦ ∑(∑(∑(a⋅b⋅k, k, 1, b), a, 1, b), b, 1, n);', null);
  Test('f(10)', 68629);
  Test('f(10) + 1', 68630);
  Test('f(20/2) + 1', 68630);
  Test('f(20)', 3455683);
  Test('f(30)', 36187912);
  Test('f(0)', 0);
  Test('f(-100)', 0);
  Test('f(2, 5)', failure, '1 argument(s) expected, but 2 given.');

  Test('g ≔ (a, b, k) ↦ a⋅b⋅k; f ≔ n ↦ ∑(∑(∑(g(a, b, k), k, 1, b), a, 1, b), b, 1, n);', null);
  Test('f(10)', 68629);
  Test('f(10) + 1', 68630);
  Test('f(20/2) + 1', 68630);
  Test('f(20)', 3455683);
  Test('f(30)', 36187912);
  Test('f(0)', 0);
  Test('f(-100)', 0);
  Test('f(2, 5)', failure, '1 argument(s) expected, but 2 given.');

  Eps(1E-9); Test('N ≔ 100000; π/N ⋅ ∑(sin(n⋅π/N), n, 1, N)', 2.0); // Riemann sum for the integral of sine from 0 to pi

  Test('delete(f); delete(g); delete(n); delete(N)', success);

  // - product

  Test('∏(k, k, 1, 5)', 120);
  Test('∏(k, k, 1, 10)', 3628800);
  Eps; Test('∏(k, k, 1, 50)', 3.0414093201713378043612608166064768844377641568960512E64);

  Test('∏(k, k, 10, 10)', 10);
  Test('∏(k, k, 10, 11)', 110);
  Test('∏(k, k, 99, 100)', 9900);

  Test('∏(k, k, -10, -10)', -10);
  Test('∏(k, k, -10, -9)', 90);
  Test('∏(k, k, -10, -8)', -720);
  Test('∏(k, k, -10, -7)', 5040);

  Test('∏(k, k, -10, 10)', 0);

  Test('∏(k, k, 10, 9)', 1);
  Test('∏(k, k, 10, -1000000000)', 1);
  Test('∏(k, k, 1, -1)', 1);
  Test('∏(k, k, 0, -1)', 1);

  for _i := 0 to 10 do
    Test(Format('∏(k, k, 1, %d) = %d!', [_i, _i]), True);

  Eps(1E-4); Test('∏((2⋅k)^2/((2⋅k−1)⋅(2⋅k+1)), k, 1, 10000)', Pi/2);

  Eps; Test('a ≔ n ↦ (n + 1)/n; ∏(a(n), n, 1, 100)', 101.0);
  Test('delete(a)', success);

  Test('∏(∑(a⋅k, k, 1, a), a, 1, 5)', 324000);

  Test('∑(∏(∑(a⋅k, k, 1, a), a, 1, n), n, 1, 1)', 1);
  Test('∑(∏(∑(a⋅k, k, 1, a), a, 1, n), n, 1, 2)', 7);
  Test('∑(∏(∑(a⋅k, k, 1, a), a, 1, n), n, 1, 3)', 115);
  Test('∑(∏(∑(a⋅k, k, 1, a), a, 1, n), n, 1, 4)', 4435);
  Test('∑(∏(∑(a⋅k, k, 1, a), a, 1, n), n, 1, 5)', 328435);
  Test('∑(∏(∑(a⋅k, k, 1, a), a, 1, n), n, 1, 6)', 41152435);
  Test('∑(∏(∑(a⋅k, k, 1, a), a, 1, n), n, 1, 7)', 8042656435);

  Test('a ≔ N ↦ ∑(∏(∑(a⋅k, k, 1, a), a, 1, n), n, 1, N);', null);

  Test('a(1)', 1);
  Test('a(2)', 7);
  Test('a(3)', 115);
  Test('a(4)', 4435);
  Test('a(5)', 328435);
  Test('a(6)', 41152435);
  Test('a(7)', 8042656435);

  Test('∑(a(n), n, 1, 7)', 8084141863);

  Test('delete(a)', success);

  Test('∑(∑(∏(∑(a⋅k, k, 1, a), a, 1, n), n, 1, N), N, 1, 7)', 8084141863);

  // - min

  Test('min(k, k, 5, 37)', 5);
  Test('min(k, k, -101, 37)', -101);

  Test('min(k/2, k, 5, 37)', 2.5);
  Test('min(k/2, k, -102, 37)', -51.0); // switches from Z to R at the second value
  Test('min(k/2, k, -102, -102)', -51);

  Test('min(n^2 − 6⋅n + 14, n, −100, 100)', 5);
  Test('min(n^2 − 6⋅n + 14.0, n, −100, 100)', 5.0);

  Test('min(Mertens(n), n, 1, 10000)', -43);

  Test('min(Mertens(n), n, 0, 10000)', failure, 'Cannot compute Mertens number for a non-positive index.');
  Test('min(Mertens(n), n, 20, 19)', failure, 'Cannot find the smallest value in an empty list.');

  // - max

  Test('max(k, k, 5, 37)', 37);
  Test('max(k, k, -101, 39)', 39);

  Test('max(k/2, k, 5, 37)', 18.5);
  Test('max(k/2, k, -102, 39)', 19.5); // switches from Z to R at the second value

  Test('max(3 − n^2 − 4⋅n, n, −100, 100)', 7);
  Test('max(3. − n^2 − 4⋅n, n, −100, 100)', 7.0);

  Test('max(Mertens(n), n, 1, 10000)', 35);

  Test('max(Mertens(n), n, 0, 10000)', failure, 'Cannot compute Mertens number for a non-positive index.');
  Test('max(Mertens(n), n, 20, 19)', failure, 'Cannot find the largest value in an empty list.');

  // - averages

  Test('mean(n^2, n, 1, 10)', 38.5);
  Test('average(n^2, n, 1, 10)', 38.5);
  Test('ArithmeticMean(n^2, n, 1, 10)', 38.5);

  Eps; Test('GeometricMean(n^2, n, 1, 10)', 20.50938353057179320068888417748352765252839967057751932878);

  Eps; Test('HarmonicMean(n^2, n, 1, 10)', 6.452579827864142630627298586770809148267388226256890997389);
  Test('ToFraction(HarmonicMean(n^2, n, 1, 10))', Rat(12700800, 1968329)); // exact

  // Parameterised values: complex numbers

  // - sum

  Test('∑(i^n, n, 0, 5)', 1 + ImaginaryUnit);

  Eps; Test('N ≔ 100; ∑(e^(n⋅2⋅π⋅i/N), n, 0, N−1)', TASC(0.0));
  Test('delete(N)', success);

  Eps; Test('∑(arcsin(k/10), k, 0, 20)', // first 11 terms are real, then non-real
    22.2860565507812216694373875292991693492858041083232776054 - 9.58906539767535588955651797835686581388739377534929228494*ImaginaryUnit);

  // - product

  Eps; Test('∏((i + k + 1)/(i + k), k, 1, 10)', 6 - 5*ImaginaryUnit);

  Eps; Test('∏(arcsin(k/10), k, 1, 20)', // first 10 factors are real, then non-real
    0.3168575833230835118697219099861197069587 + 0.3968699397031834013477504509421124402082*ImaginaryUnit);

  // - min and max (not applicable)

  Test('min(arcsin(k/10), k, 1, 20)', failure, 'Object isn''t a real number.');
  Test('max(arcsin(k/10), k, 1, 20)', failure, 'Object isn''t a real number.');

  // - averages

  Eps; Test('N ≔ 100; average(e^(n⋅2⋅π⋅i/N), n, 0, N−1)', TASC(0.0));
  Eps(1E-4); Test('N ≔ 100000; average(e^(n⋅π⋅i/N), n, 0, N)', 0.6366*ImaginaryUnit);
  Test('delete(N)', success);

  Test('mean(n^2 + i, n, 1, 10)', 38.5 + ImaginaryUnit);
  Test('average(n^2 + i, n, 1, 10)', 38.5 + ImaginaryUnit);
  Test('ArithmeticMean(n^2 + i, n, 1, 10)', 38.5 + ImaginaryUnit);

  Eps; Test('GeometricMean(n^2 + i, n, 1, 10)', 21.1296794732483723780501106219994500192481607582975984365 + 2.82602359727550791780719604001273696999214552115880735774*ImaginaryUnit);

  Eps; Test('HarmonicMean(n^2 + i, n, 1, 10)', 7.36987810444769310699592007330432761598639272697814191183 + 4.12361984435842182947838233538305872903090270124293967516*ImaginaryUnit);

  // Parameterised values: vectors

  // - sum

  Test('∑(❨1, t, t^2❩, t, 0, 100)', ASO(ASR3(101, 5050, 338350)));
  Test('∑(❨1, t, t^2❩, t, -100, 100)', ASO(ASR3(201, 0, 676700)));
  Test('∑(❨i, t, t^2❩, t, -100, 100)', ASO(ASC3(201*ImaginaryUnit, 0, 676700)));

  // - product (not applicable)

  Test('∏(❨1, t, t^2❩, t, 0, 100)', failure, 'Cannot compute product of this container.');
  Test('∏(❨i, t, t^2❩, t, 0, 100)', failure, 'Cannot compute product of this container.');

  // - min (not applicable)

  Test('min(❨1, t, t^2❩, t, 0, 100)', failure, 'Cannot find the smallest value in this list.');
  Test('min(❨i, t, t^2❩, t, 0, 100)', failure, 'Cannot find the smallest value in this list.');

  // - max (not applicable)

  Test('max(❨1, t, t^2❩, t, 0, 100)', failure, 'Cannot find the largest value in this list.');
  Test('max(❨i, t, t^2❩, t, 0, 100)', failure, 'Cannot find the largest value in this list.');

  // - averages

  Eps; Test('mean(❨1, t, t^2❩, t, -100, 100)', ASO(ASR3(1, 0, 676700/201)));
  Eps; Test('average(❨1, t, t^2❩, t, -100, 100)', ASO(ASR3(1, 0, 676700/201)));
  Eps; Test('ArithmeticMean(❨1, t, t^2❩, t, -100, 100)', ASO(ASR3(1, 0, 676700/201)));

  Eps; Test('mean(❨i, t, t^2❩, t, -100, 100)', ASO(ASC3(ImaginaryUnit, 0, 676700/201)));
  Eps; Test('average(❨i, t, t^2❩, t, -100, 100)', ASO(ASC3(ImaginaryUnit, 0, 676700/201)));
  Eps; Test('ArithmeticMean(❨i, t, t^2❩, t, -100, 100)', ASO(ASC3(ImaginaryUnit, 0, 676700/201)));

  Test('GeometricMean(❨1, t, t^2❩, t, -100, 100)', failure, 'Cannot compute mean of this object.');
  Test('HarmonicMean(❨1, t, t^2❩, t, -100, 100)', failure, 'Cannot compute mean of this object.');
  Test('GeometricMean(❨i, t, t^2❩, t, -100, 100)', failure, 'Cannot compute mean of this object.');
  Test('HarmonicMean(❨i, t, t^2❩, t, -100, 100)', failure, 'Cannot compute mean of this object.');

  // Parameterised values: matrices

  // - sum

  Test('∑(❨❨0, 1❩, ❨t, t^2❩❩, t, 0, 100)', 2, [0, 101, 5050, 338350]);
  Test('∑(❨❨0, 1❩, ❨t, t^2❩❩, t, -100, 100)', 2, [0, 201, 0, 676700]);

  // - product

  Eps; Test('∏(❨❨1, 0❩, ❨1/t!, 1❩❩, t, 0, 100)', 2, [1, 0, Exp(1), 1]);
  Eps; Test('∏(❨❨i, 0❩, ❨1/t!, i❩❩, t, 0, 100)', 2, [ImaginaryUnit, 0, Exp(1), ImaginaryUnit]);
  Eps; Test('∏(❨❨1, 0❩, ❨1/t!, i❩❩, t, 0, 100)', 2, [1, 0, cexp(ImaginaryUnit), ImaginaryUnit]);

  Test('rot ≔ φ ↦ ❨❨cos(φ), −sin(φ)❩, ❨sin(φ), cos(φ)❩❩;', null);
  Eps; Test('∏(rot(n/(100⋅2⋅π)), n, 1, 10)', 2, [Cos(TASR(11/(40*Pi))), -Sin(TASR(11/(40*Pi))), Sin(TASR(11/(40*Pi))), Cos(TASR(11/(40*Pi)))]);
  Test('delete(rot)', success);

  // - min (not applicable)

  Test('min(❨❨1, 0❩, ❨1/t!, 1❩❩, t, 0, 100)', failure, 'Cannot find the smallest value in this list.');
  Test('min(❨❨i, 0❩, ❨1/t!, 1❩❩, t, 0, 100)', failure, 'Cannot find the smallest value in this list.');

  // - max (not applicable)

  Test('max(❨❨1, 0❩, ❨1/t!, 1❩❩, t, 0, 100)', failure, 'Cannot find the largest value in this list.');
  Test('max(❨❨i, 0❩, ❨1/t!, 1❩❩, t, 0, 100)', failure, 'Cannot find the largest value in this list.');

  // - averages

  Eps; Test('mean(❨❨0, 1❩, ❨t, t^2❩❩, t, -100, 100)', 2, [0, 1, 0, 676700/201]);
  Eps; Test('average(❨❨0, 1❩, ❨t, t^2❩❩, t, -100, 100)', 2, [0, 1, 0, 676700/201]);
  Eps; Test('ArithmeticMean(❨❨0, 1❩, ❨t, t^2❩❩, t, -100, 100)', 2, [0, 1, 0, 676700/201]);

  Eps; Test('mean(❨❨0, i❩, ❨t, t^2❩❩, t, -100, 100)', 2, [0, ImaginaryUnit, 0, 676700/201]);
  Eps; Test('average(❨❨0, i❩, ❨t, t^2❩❩, t, -100, 100)', 2, [0, ImaginaryUnit, 0, 676700/201]);
  Eps; Test('ArithmeticMean(❨❨0, i❩, ❨t, t^2❩❩, t, -100, 100)', 2, [0, ImaginaryUnit, 0, 676700/201]);

  Eps; Test('GeometricMean(❨❨0, 1❩, ❨t, t^2❩❩, t, -100, 100)', failure, 'Cannot compute mean of this object.');
  Eps; Test('HarmonicMean(❨❨0, 1❩, ❨t, t^2❩❩, t, -100, 100)', failure, 'Cannot compute mean of this object.');
  Eps; Test('GeometricMean(❨❨0, i❩, ❨t, t^2❩❩, t, -100, 100)', failure, 'Cannot compute mean of this object.');
  Eps; Test('HarmonicMean(❨❨0, i❩, ❨t, t^2❩❩, t, -100, 100)', failure, 'Cannot compute mean of this object.');

  // Given container

  // - sum

  Test('∑(''(1, 5, 3, 10))', 19);
  Test('∑(''(1, 5, 3, 10, 10))', 29);
  Test('∑(''(1, 5, 3, 10, 1000))', 1019);
  Test('∑(''(21))', 21);
  Test('∑(''())', 0);

  Test('∑(''(1, 5, 3, 10.0))', 19.0);
  Test('∑(''(1, 5, 3, 10, i))', 19 + ImaginaryUnit);

  Eps; Test('∑(''(1, 5, 3, 10.5 + i/2 + i/3))', 19.5 + 5*ImaginaryUnit/6);

  Test('∑(❨1, 5, 3, 10❩)', 19.0);
  Test('∑(❨1, 5, 3, 10, i❩)', 19.0 + ImaginaryUnit);

  Test('∑(❨❨1, 2❩, ❨3, 4❩❩)', 10.0);
  Test('∑(❨❨i, 2❩, ❨3, 4❩❩)', 9.0 + ImaginaryUnit);

  Test('∑({1, 2, 5})', 8);
  Test('∑({1, 2, 5.0})', 8);
  Test('∑({1, 2, 5 + i})', 8.0 + ImaginaryUnit);

  Test('∑({1, 2, 5, 5})', 8);

  Test('∑({})', 0);
  Test('∑(∅)', 0);

  Test('∑(''(❨5, 0, 1❩, ❨1, 4, 9❩, ❨−2, 4, 6❩))', [4, 8, 16]);
  Test('∑(''(❨5, 0, 1❩, ❨1, 4, 9❩, ❨−2, 4, 6, 8❩))', failure, 'Cannot add two vectors of unequal dimension.');

  Test('∑(''(❨❨4, 2❩, ❨1, −3❩❩, ❨❨7, 5❩, ❨9, 3❩❩, ❨❨−1, 0❩, ❨4, 4❩❩))', 2, [10, 7, 14, 4]);
  Test('∑(''(❨❨4, 2❩, ❨1, −3❩❩, ❨❨7, 5❩, ❨9, 3❩❩, ❨❨−1, 0, 5❩, ❨4, 1, 4❩❩))', failure, 'Cannot add two matrices of different sizes.');

  Test('∑(5)', failure, 'Cannot compute sum of this object.');
  Test('∑("rat")', failure, 'Cannot compute sum of this container.');
  Test('∑("5")', failure, 'Cannot compute sum of this container.');

  // - product

  Test('∏(''(5, 2, 7, 1))', 70);
  Test('∏(''(5, 2, 7, 1, 2))', 140);
  Test('∏(''(5, 2, 7, 1.0))', 70.0);
  Test('∏(''(5, 2, 7, i))', 70 * ImaginaryUnit);
  Test('∏(''())', 1);

  Test('∏(❨5, 8, 2❩)', 80.0);
  Test('∏(❨5, 8, i❩)', 40.0 * ImaginaryUnit);

  Test('∏(❨❨1, 2❩, ❨3, 4❩❩)', 24.0);
  Test('∏(❨❨i, 2❩, ❨3, 4❩❩)', 24.0 * ImaginaryUnit);

  Test('∏({1, 2, 5})', 10);
  Test('∏({1, 2, 5.0})', 10);
  Test('∏({1, 2, 5 + i})', 10.0 + 2*ImaginaryUnit);

  Test('∏({1, 2, 5, 5})', 10);

  Test('∏({})', 1);
  Test('∏(∅)', 1);

  Test('∏(''(❨5, 0, 1❩, ❨1, 4, 9❩, ❨−2, 4, 6❩))', failure, 'Cannot compute product of this container.');
  Test('∏(''(❨❨4, 2❩, ❨1, −3❩❩, ❨❨7, 5❩, ❨9, 3❩❩, ❨❨−1, 0❩, ❨4, 4❩❩))', 2, [58, 104, 4, -16]);
  Test('∏(''(❨❨4, 2❩, ❨1, −3❩❩, ❨❨7, 5❩, ❨9, 3❩❩, ❨❨−1, 0❩❩))', failure, 'When multiplying two matrices, the number of columns in the first matrix has to equal the number of rows in the second matrix.');

  Test('∏(5)', failure, 'Cannot compute product of this object.');
  Test('∏("rat")', failure, 'Cannot compute product of this container.');
  Test('∏("5")', failure, 'Cannot compute product of this container.');

  // - max

  Test('max(''(1, 7, 2, 0, −3, 5))', 7);
  Test('max(''(1, 7, 2, 0, −3, 5.0))', 7.0);
  Test('max(''())', failure, 'Cannot find the largest value in an empty list.');

  Test('max(❨1, 7, 2, 0, −3, 5❩)', 7.0);
  Test('max(❨❨1, 7, 2❩, ❨0, −3, 5❩❩)', 7.0);

  Test('max({1, 7, 2, 0, −3, 5})', 7);
  Test('max({1, 7, 2, 0, −3, 5.0})', 7);

  Test('max(''(1, 7, 2, 0, −3, 5, i))', failure, 'Object isn''t a real number.');
  Test('max(❨1, 7, 2, 0, −3, 5, i❩)', failure, 'Object isn''t a real number.');
  Test('max(❨❨1, 7, 2❩, ❨0, −3, i❩❩)', failure, 'Object isn''t a real number.');
  Test('max({1, 7, 2, 0, −3, 5, i})', failure, 'Object isn''t a real number.');

  Test('max(5)', failure, 'Cannot find the largest number in an object of type "integer".');
  Test('max(5.0)', failure, 'Cannot find the largest number in an object of type "real number".');
  Test('max(i)', failure, 'Cannot find the largest number in an object of type "complex number".');
  Test('max("test")', failure, 'Cannot find the largest value in this list.');

  Test('max(2, 7)', 7);
  Test('max(20, 7)', 20);
  Test('max(5/2, 7)', 7.0);
  Test('max(2, 7/5)', 2.0);
  Test('max(2.0, 7)', 7.0);
  Test('max(2, 7.0)', 7.0);
  Test('max(5.3, -2.1)', 5.3);
  Test('max(-52, -52.001)', -52.0);

  Test('max(5, i)', failure, 'An object of type real number was expected as argument 2, but an object of type complex number was given.');
  Test('max(i, 2)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');

  Test('max(5, 6, 1)', failure, 'There is no version of this function that can take 3 argument(s).');

  // - min

  Test('min(''(1, 7, 2, 0, −3, 5))', -3);
  Test('min(''(1, 7, 2, 0, −3, 5.0))', -3.0);
  Test('min(''())', failure, 'Cannot find the smallest value in an empty list.');

  Test('min(❨1, 7, 2, 0, −3, 5❩)', -3.0);
  Test('min(❨❨1, 7, 2❩, ❨0, −3, 5❩❩)', -3.0);

  Test('min({1, 7, 2, 0, −3, 5})', -3);
  Test('min({1, 7, 2, 0, −3, 5.0})', -3);

  Test('min(''(1, 7, 2, 0, −3, 5, i))', failure, 'Object isn''t a real number.');
  Test('min(❨1, 7, 2, 0, −3, 5, i❩)', failure, 'Object isn''t a real number.');
  Test('min(❨❨1, 7, 2❩, ❨0, −3, i❩❩)', failure, 'Object isn''t a real number.');
  Test('min({1, 7, 2, 0, −3, 5, i})', failure, 'Object isn''t a real number.');

  Test('min(5)', failure, 'Cannot find the smallest number in an object of type "integer".');
  Test('min(5.0)', failure, 'Cannot find the smallest number in an object of type "real number".');
  Test('min(i)', failure, 'Cannot find the smallest number in an object of type "complex number".');
  Test('min("test")', failure, 'Cannot find the smallest value in this list.');

  Test('min(2, 7)', 2);
  Test('min(20, 7)', 7);
  Test('min(5/2, 7)', 2.5);
  Test('min(2, 7/5)', 1.4);
  Test('min(2.0, 7)', 2.0);
  Test('min(2, 7.0)', 2.0);
  Test('min(5.3, -2.1)', -2.1);
  Test('min(-52, -52.001)', -52.001);

  Test('min(5, i)', failure, 'An object of type real number was expected as argument 2, but an object of type complex number was given.');
  Test('min(i, 2)', failure, 'An object of type real number was expected as argument 1, but an object of type complex number was given.');

  Test('min(5, 6, 1)', failure, 'There is no version of this function that can take 3 argument(s).');

  // - averages

  Test('mean(''(1, 7, 2, 0, −3, 5))', 2.0);
  Test('mean(''(1, 7, 2, 0, −3, 5, i))', 12/7 + ImaginaryUnit/7);
  Test('mean(''())', failure, 'Cannot compute the mean of an empty list.');

  Test('average(''(1, 7, 2, 0, −3, 5))', 2.0);
  Test('average(''(1, 7, 2, 0, −3, 5, i))', 12/7 + ImaginaryUnit/7);
  Test('average(''())', failure, 'Cannot compute the mean of an empty list.');

  Test('ArithmeticMean(''(1, 7, 2, 0, −3, 5))', 2.0);
  Test('ArithmeticMean(''(1, 7, 2, 0, −3, 5, i))', 12/7 + ImaginaryUnit/7);
  Test('ArithmeticMean(''())', failure, 'Cannot compute the mean of an empty list.');

  Test('mean(❨1, 7, 2, 0, −3, 5❩)', 2.0);
  Test('mean(❨❨1, 7, 2❩, ❨0, −3, 5❩❩)', 2.0);
  Test('mean(❨i, 7, 2, 0, −3, 5❩)', 11/6 + ImaginaryUnit/6);
  Test('mean(❨❨i, 7, 2❩, ❨0, −3, 5❩❩)', 11/6 + ImaginaryUnit/6);

  Test('mean({1, 7, 2, 0, −3, 5})', 2.0);
  Test('mean({1, 7, 2, 0, −3, 5.0})', 2.0);
  Test('mean({1, 7, 2, 0, −3, 5, i})', 12/7 + ImaginaryUnit/7);
  Test('mean({})', failure, 'Cannot compute the mean of an empty list.');
  Test('mean(∅)', failure, 'Cannot compute the mean of an empty list.');

  Test('average(❨1, 7, 2, 0, −3, 5❩)', 2.0);
  Test('average(❨❨1, 7, 2❩, ❨0, −3, 5❩❩)', 2.0);
  Test('average(❨i, 7, 2, 0, −3, 5❩)', 11/6 + ImaginaryUnit/6);
  Test('average(❨❨i, 7, 2❩, ❨0, −3, 5❩❩)', 11/6 + ImaginaryUnit/6);

  Test('average({1, 7, 2, 0, −3, 5})', 2.0);
  Test('average({1, 7, 2, 0, −3, 5.0})', 2.0);
  Test('average({1, 7, 2, 0, −3, 5, i})', 12/7 + ImaginaryUnit/7);
  Test('average({})', failure, 'Cannot compute the mean of an empty list.');
  Test('average(∅)', failure, 'Cannot compute the mean of an empty list.');

  Test('ArithmeticMean(❨1, 7, 2, 0, −3, 5❩)', 2.0);
  Test('ArithmeticMean(❨❨1, 7, 2❩, ❨0, −3, 5❩❩)', 2.0);
  Test('ArithmeticMean(❨i, 7, 2, 0, −3, 5❩)', 11/6 + ImaginaryUnit/6);
  Test('ArithmeticMean(❨❨i, 7, 2❩, ❨0, −3, 5❩❩)', 11/6 + ImaginaryUnit/6);

  Test('ArithmeticMean({1, 7, 2, 0, −3, 5})', 2.0);
  Test('ArithmeticMean({1, 7, 2, 0, −3, 5.0})', 2.0);
  Test('ArithmeticMean({1, 7, 2, 0, −3, 5, i})', 12/7 + ImaginaryUnit/7);
  Test('ArithmeticMean({})', failure, 'Cannot compute the mean of an empty list.');
  Test('ArithmeticMean(∅)', failure, 'Cannot compute the mean of an empty list.');

  Eps; Test('GeometricMean(''(2, 5, 7, 3, 1, 4))', 3.07170766197453541);
  Eps; Test('GeometricMean(❨2, 5, 7, 3, 1, 4❩)', 3.07170766197453541);
  Eps; Test('GeometricMean(❨❨2, 5, 7❩, ❨3, 1, 4❩❩)', 3.07170766197453541);
  Eps; Test('GeometricMean({2, 5, 7, 3, 1, 4})', 3.07170766197453541);

  Eps; Test('GeometricMean(''(2, 5, 7, 3, 1, i))', 2.3549426067282413 + 0.631004969694475251*ImaginaryUnit);
  Eps; Test('GeometricMean(❨2, 5, 7, 3, 1, i❩)', 2.3549426067282413 + 0.631004969694475251*ImaginaryUnit);
  Eps; Test('GeometricMean(❨❨2, 5, 7❩, ❨3, 1, i❩❩)', 2.3549426067282413 + 0.631004969694475251*ImaginaryUnit);
  Eps; Test('GeometricMean({2, 5, 7, 3, 1, i})', 2.3549426067282413 + 0.631004969694475251*ImaginaryUnit);

  Eps; Test('GeometricMean(''())', failure, 'Cannot compute the mean of an empty list.');
  Eps; Test('GeometricMean({})', failure, 'Cannot compute the mean of an empty list.');
  Eps; Test('GeometricMean(∅)', failure, 'Cannot compute the mean of an empty list.');

  Eps; Test('HarmonicMean(''(2, 5, 7, 3, 1, 4))', 2.47301275760549558);
  Eps; Test('HarmonicMean(❨2, 5, 7, 3, 1, 4❩)', 2.47301275760549558);
  Eps; Test('HarmonicMean(❨❨2, 5, 7❩, ❨3, 1, 4❩❩)', 2.47301275760549558);
  Eps; Test('HarmonicMean({2, 5, 7, 3, 1, 4})', 2.47301275760549558);

  Eps; Test('HarmonicMean(''(2, 5, 7, 3, 1, i))', 2.27642726399392763 + 1.04606066835607178*ImaginaryUnit);
  Eps; Test('HarmonicMean(❨2, 5, 7, 3, 1, i❩)', 2.27642726399392763 + 1.04606066835607178*ImaginaryUnit);
  Eps; Test('HarmonicMean(❨❨2, 5, 7❩, ❨3, 1, i❩❩)', 2.27642726399392763 + 1.04606066835607178*ImaginaryUnit);
  Eps; Test('HarmonicMean({2, 5, 7, 3, 1, i})', 2.27642726399392763 + 1.04606066835607178*ImaginaryUnit);

  Eps; Test('HarmonicMean(''())', failure, 'Cannot compute the mean of an empty list.');
  Eps; Test('HarmonicMean({})', failure, 'Cannot compute the mean of an empty list.');
  Eps; Test('HarmonicMean(∅)', failure, 'Cannot compute the mean of an empty list.');


  //
  // Integrals
  //

  Eps(1E-8); Test('∫(sin(x), x, 0, π)', 2.0);
  Eps(1E-12); Test('∫(sin(x)^2, x, 0, π)', Pi/2);
  Eps(1E-10); Test('∫(x^2, x, 0, 1)', 1/3);

  Test('err ≔ Δx ↦ ∫(x^2, x, 0, 10, Δx) − 1000./3;', null);
  Test('err(1E-1) <  0.02', True);
  Test('err(1E-2) <  0.0002', True);
  Test('err(1E-3) <  0.000002', True);
  Test('err(1E-4) <  0.00000002', True);
  Test('err(1E-5) <  0.0000000002', True);
  Test('delete(err)', success);

  Eps(1E-4); Test('∫(∫(x^2 + y, y, 0, 1, 0.01), x, 0, 1, 0.01)', 5/6);

  Eps(1E-4); Test('f ≔ (x, y) ↦ x^2 + y; ∫(∫(f(x, y), y, 0, 1, 0.01), x, 0, 1, 0.01)', 5/6);
  Eps(1E-4); Test('f ≔ (x, y) ↦ x^2 + y; ∫(∫(f(x, y), y, 0, x, 0.005), x, 0, 1, 0.005)', 5/12);
  Eps(1E-4); Test('f ≔ (x, y) ↦ x^2 + y; ∫(∫(f(x, y), y, 0, x^2, 0.005), x, 0, 1, 0.005)', 3/10);

  Test('delete(f)', success);

  Test('∫(prime(x)!, x, 0, 1)', failure, 'To obtain the Nth prime number, N must be a positive integer.');
  Test('∫(∫(prime(x), y, 0, 1, 0.01), x, 0, 1, 0.01)', failure, 'To obtain the Nth prime number, N must be a positive integer.');

  Test('f ≔ n ↦ ∫(x^2, x, 0, n);', null);
  Eps(1E-8); Test('f(0)', 0.0);
  Eps(1E-8); Test('f(1)', 1/3);
  Eps(1E-8); Test('f(2)', 8/3);
  Eps(1E-8); Test('f(3)', 9.0);
  Eps(1E-8); Test('f(4)', 64/3);
  Eps(1E-8); Test('∑(f(n), n, 1, 4)', 100/3);
  Eps(1E-8); Test('∑(∫(x^2, x, 0, n), n, 1, 4)', 100/3);
  Test('delete(f)', success);

  Test('∫(ln(x), x, −2, −1)', failure, 'Integrand must be a real-valued function.');


  //
  // Differentiation (derivative at a point)
  //

  Test('diff(7, x, 1)', 0.0);
  Eps(1E-12); Test('diff(7⋅x, x, 5)', 7.0);
  Eps(1E-12); Test('diff(x^2, x, 2)', 4.0);
  Eps(1E-12); Test('diff(sin(x), x, π/6)', Sqrt(3)/2);
  Eps(1E-12); Test('diff(exp(x), x, 1)', Exp(1));
  Eps(1E-12); Test('diff(ln(x), x, 1)', 1.0);

  Test('diff(x!, x, 5)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('diff(ln(x), x, −2)', failure, 'Function to differentiate must be real-valued.');


  //
  // Intervals
  //

  Test('[5, 4]', asoarr([]));
  Test('[5, 5]', asoarr([5.0]));
  Eps; Test('[5, 6, 0.1]', asoarr([5.0, 5.1, 5.2, 5.3, 5.4, 5.5, 5.6, 5.7, 5.8, 5.9, 6.0]));
  Test('#[5, 6, 0.1]', 10 + 1);
  Test('#[5, 6, 0.01]', 100 + 1);
  Test('#[5, 6, 0.001]', 1000 + 1);
  Test('#[5, 6, 0.0001]', 10000 + 1);
  Test('#[5, 6, 0.00001]', 100000 + 1);
  Test('#[5, 6, 0.000001]', 1000000 + 1);
  Test('min([5, 6, 0.000001])', 5.0);
  Test('max([5, 6, 0.000001])', 6.0);
  Eps; Test('[5, 6, 0.2]', asoarr([5.0, 5.2, 5.4, 5.6, 5.8, 6.0]));
  Eps; Test('[5, 6, 0.3]', asoarr([5.0, 5.3, 5.6, 5.9]));
  Eps; Test('[5, 6, 0.4]', asoarr([5.0, 5.4, 5.8]));
  Eps; Test('[5, 6, 0.5]', asoarr([5.0, 5.5, 6.0]));
  Eps; Test('[5, 6, 0.6]', asoarr([5.0, 5.6]));
  Eps; Test('[5, 6, 0.7]', asoarr([5.0, 5.7]));
  Eps; Test('[5, 6, 0.8]', asoarr([5.0, 5.8]));
  Eps; Test('[5, 6, 0.9]', asoarr([5.0, 5.9]));
  Test('[5, 6, 1.0]', asoarr([5.0, 6.0]));
  Test('[5, 6, 1.1]', asoarr([5.0]));
  Test('[5, 6, 10]', asoarr([5.0]));
  Test('[5, 6, 100]', asoarr([5.0]));
  Test('[5, 6, 0]', failure, 'A positive real number was expected as argument 3, but "0" was given.');
  Test('[5, 6, -1]', failure, 'A positive real number was expected as argument 3, but "-1" was given.');
  Test('[5, 6, 0.001, 5]', failure, 'Too many arguments.');
  Test('[5, 6, true]', failure, 'An object of type real number was expected as argument 3, but an object of type boolean was given.');
  Eps; Test('[-5, 5, 1]', asoarr([-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0]));

  Test('interval(5, 4)', asoarr([]));
  Test('interval(5, 5)', asoarr([5.0]));
  Eps; Test('interval(5, 6, 0.1)', asoarr([5.0, 5.1, 5.2, 5.3, 5.4, 5.5, 5.6, 5.7, 5.8, 5.9, 6.0]));
  Test('#interval(5, 6, 0.1)', 10 + 1);
  Test('#interval(5, 6, 0.01)', 100 + 1);
  Test('#interval(5, 6, 0.001)', 1000 + 1);
  Test('#interval(5, 6, 0.0001)', 10000 + 1);
  Test('#interval(5, 6, 0.00001)', 100000 + 1);
  Test('#interval(5, 6, 0.000001)', 1000000 + 1);
  Test('min(interval(5, 6, 0.000001))', 5.0);
  Test('max(interval(5, 6, 0.000001))', 6.0);
  Eps; Test('interval(5, 6, 0.2)', asoarr([5.0, 5.2, 5.4, 5.6, 5.8, 6.0]));
  Eps; Test('interval(5, 6, 0.3)', asoarr([5.0, 5.3, 5.6, 5.9]));
  Eps; Test('interval(5, 6, 0.4)', asoarr([5.0, 5.4, 5.8]));
  Eps; Test('interval(5, 6, 0.5)', asoarr([5.0, 5.5, 6.0]));
  Eps; Test('interval(5, 6, 0.6)', asoarr([5.0, 5.6]));
  Eps; Test('interval(5, 6, 0.7)', asoarr([5.0, 5.7]));
  Eps; Test('interval(5, 6, 0.8)', asoarr([5.0, 5.8]));
  Eps; Test('interval(5, 6, 0.9)', asoarr([5.0, 5.9]));
  Test('interval(5, 6, 1.0)', asoarr([5.0, 6.0]));
  Test('interval(5, 6, 1.1)', asoarr([5.0]));
  Test('interval(5, 6, 10)', asoarr([5.0]));
  Test('interval(5, 6, 100)', asoarr([5.0]));
  Test('interval(5, 6, 0)', failure, 'A positive real number was expected as argument 3, but "0" was given.');
  Test('interval(5, 6, -1)', failure, 'A positive real number was expected as argument 3, but "-1" was given.');
  Test('interval(5, 6, 0.001, 5)', failure, 'Too many arguments.');
  Test('interval(5, 6, true)', failure, 'An object of type real number was expected as argument 3, but an object of type boolean was given.');
  Eps; Test('interval(-5, 5, 1)', asoarr([-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0]));

  Test('ClosedInterval(5, 4)', asoarr([]));
  Test('ClosedInterval(5, 5)', asoarr([5.0]));
  Eps; Test('ClosedInterval(5, 6, 0.1)', asoarr([5.0, 5.1, 5.2, 5.3, 5.4, 5.5, 5.6, 5.7, 5.8, 5.9, 6.0]));
  Test('#ClosedInterval(5, 6, 0.1)', 10 + 1);
  Test('#ClosedInterval(5, 6, 0.01)', 100 + 1);
  Test('#ClosedInterval(5, 6, 0.001)', 1000 + 1);
  Test('#ClosedInterval(5, 6, 0.0001)', 10000 + 1);
  Test('#ClosedInterval(5, 6, 0.00001)', 100000 + 1);
  Test('#ClosedInterval(5, 6, 0.000001)', 1000000 + 1);
  Test('min(ClosedInterval(5, 6, 0.000001))', 5.0);
  Test('max(ClosedInterval(5, 6, 0.000001))', 6.0);
  Eps; Test('ClosedInterval(5, 6, 0.2)', asoarr([5.0, 5.2, 5.4, 5.6, 5.8, 6.0]));
  Eps; Test('ClosedInterval(5, 6, 0.3)', asoarr([5.0, 5.3, 5.6, 5.9]));
  Eps; Test('ClosedInterval(5, 6, 0.4)', asoarr([5.0, 5.4, 5.8]));
  Eps; Test('ClosedInterval(5, 6, 0.5)', asoarr([5.0, 5.5, 6.0]));
  Eps; Test('ClosedInterval(5, 6, 0.6)', asoarr([5.0, 5.6]));
  Eps; Test('ClosedInterval(5, 6, 0.7)', asoarr([5.0, 5.7]));
  Eps; Test('ClosedInterval(5, 6, 0.8)', asoarr([5.0, 5.8]));
  Eps; Test('ClosedInterval(5, 6, 0.9)', asoarr([5.0, 5.9]));
  Test('ClosedInterval(5, 6, 1.0)', asoarr([5.0, 6.0]));
  Test('ClosedInterval(5, 6, 1.1)', asoarr([5.0]));
  Test('ClosedInterval(5, 6, 10)', asoarr([5.0]));
  Test('ClosedInterval(5, 6, 100)', asoarr([5.0]));
  Test('ClosedInterval(5, 6, 0)', failure, 'A positive real number was expected as argument 3, but "0" was given.');
  Test('ClosedInterval(5, 6, -1)', failure, 'A positive real number was expected as argument 3, but "-1" was given.');
  Test('ClosedInterval(5, 6, 0.001, 5)', failure, 'Too many arguments.');
  Test('ClosedInterval(5, 6, true)', failure, 'An object of type real number was expected as argument 3, but an object of type boolean was given.');
  Eps; Test('ClosedInterval(-5, 5, 1)', asoarr([-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0]));

  Test('OpenInterval(5, 4)', asoarr([]));
  Test('OpenInterval(5, 5)', asoarr([]));
  Eps; Test('OpenInterval(5, 6, 0.1)', asoarr([5.1, 5.2, 5.3, 5.4, 5.5, 5.6, 5.7, 5.8, 5.9]));
  Test('#OpenInterval(5, 6, 0.1)', 10 - 1);
  Test('#OpenInterval(5, 6, 0.01)', 100 - 1);
  Test('#OpenInterval(5, 6, 0.001)', 1000 - 1);
  Test('#OpenInterval(5, 6, 0.0001)', 10000 - 1);
  Test('#OpenInterval(5, 6, 0.00001)', 100000 - 1);
  Test('#OpenInterval(5, 6, 0.000001)', 1000000 - 1);
  Test('min(OpenInterval(5, 6, 0.000001))', 5.000001);
  Test('max(OpenInterval(5, 6, 0.000001))', 5.999999);
  Eps; Test('OpenInterval(5, 6, 0.2)', asoarr([5.2, 5.4, 5.6, 5.8]));
  Eps; Test('OpenInterval(5, 6, 0.3)', asoarr([5.3, 5.6]));
  Eps; Test('OpenInterval(5, 6, 0.4)', asoarr([5.4]));
  Eps; Test('OpenInterval(5, 6, 0.5)', asoarr([5.5]));
  Eps; Test('OpenInterval(5, 6, 0.6)', asoarr([]));
  Eps; Test('OpenInterval(5, 6, 0.7)', asoarr([]));
  Eps; Test('OpenInterval(5, 6, 0.8)', asoarr([]));
  Eps; Test('OpenInterval(5, 6, 0.9)', asoarr([]));
  Test('OpenInterval(5, 6, 1.0)', asoarr([]));
  Test('OpenInterval(5, 6, 1.1)', asoarr([]));
  Test('OpenInterval(5, 6, 10)', asoarr([]));
  Test('OpenInterval(5, 6, 100)', asoarr([]));
  Test('OpenInterval(5, 6, 0)', failure, 'A positive real number was expected as argument 3, but "0" was given.');
  Test('OpenInterval(5, 6, -1)', failure, 'A positive real number was expected as argument 3, but "-1" was given.');
  Test('OpenInterval(5, 6, 0.001, 5)', failure, 'Too many arguments.');
  Test('OpenInterval(5, 6, true)', failure, 'An object of type real number was expected as argument 3, but an object of type boolean was given.');
  Eps; Test('OpenInterval(-5, 5, 1)', asoarr([-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0]));


  //
  // Random number functions
  //

  Test('freqs ≔ ''(0, 0, 0, 0, 0, 0, 0);', null);

  for _i := 1 to 10000 do
    Test('n ≔ RandomInt(7); (type(n) = "integer") ∧ (n ≥ 0) ∧ (n < 7) ∧ succeeded(inc(freqs[n+1]))', True);

  for _i := 1 to 7 do
    Test(Format('(freqs[%d] ≥ 1150) ∧ (freqs[%d] ≤ 1700)', [_i, _i]), True);

  Test('freqs ≔ ''(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);', null);

  for _i := 1 to 10000 do
    Test('n ≔ RandomInt(20, 30); (type(n) = "integer") ∧ (n ≥ 20) ∧ (n < 30) ∧ succeeded(inc(freqs[n-19]))', True);

  for _i := 1 to 10 do
    Test(Format('(freqs[%d] ≥ 750) ∧ (freqs[%d] ≤ 1250)', [_i, _i]), True);

  Test('delete(freqs); delete(n)', success);

  for _i := 0 to 10 do
    Test('RandomInt(1)', 0);

  Test('RandomInt(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('RandomInt(-1)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');

  for _i := 0 to 10 do
    Test('RandomInt(4, 5)', 4);

  Test('RandomInt(4, 4)', failure, 'The set of possible values is empty.');
  Test('RandomInt(4, 3)', failure, 'The set of possible values is empty.');
  Test('RandomInt(4, -10)', failure, 'The set of possible values is empty.');

  Test('s ≔ 0.', 0.0);
  for _i := 1 to 10000 do
    Test('x ≔ RandomReal(); s ≔ s + x; (x ≥ 0) ∧ (x < 1)', True);
  Test('avg ≔ s / 10000; (avg ≥ 0.4) ∧ (avg ≤ 0.6)', True);

  Test('s ≔ 0.', 0.0);
  for _i := 1 to 10000 do
    Test('x ≔ RandomReal(5); s ≔ s + x; (x ≥ 0) ∧ (x < 5)', True);
  Test('avg ≔ s / 10000; (avg ≥ 2.2) ∧ (avg ≤ 2.8)', True);

  Test('s ≔ 0.', 0.0);
  for _i := 1 to 10000 do
    Test('x ≔ RandomReal(2, 5); s ≔ s + x; (x ≥ 2) ∧ (x < 5)', True);
  Test('avg ≔ s / 10000; (avg ≥ 3.2) ∧ (avg ≤ 3.8)', True);

  Test('delete(x); delete(s); delete(avg)', success);


  //
  // Bitwise logical operators
  //

  // Bitwise and (bitwise conjunction)

  Test('157 ∧ 197', 133);
  Test('219 ∧ 85', 81);
  Test('219 ∧ 131', 131);
  Test('91 ∧ 64', 64);
  Test('87 ∧ 32', 0);
  Test('22355 ∧ 42621', 1617);
  Test('55964 ∧ 20506', 20504);
  Test('54743 ∧ 5188', 5188);
  Test('45731 ∧ 38101 ∧ 20981', 4225);
  Test('219 ∧ 159 ∧ 187', 155);
  Test('219 ∧ 187 ∧ 106 ∧ 106 ∧ 127', 10);
  Test('219 ∧ 187 ∧ 106 ∧ 106 ∧ 127 ∧ 1', 0);
  Test('219 ∧ 187 ∧ 106 ∧ 106 ∧ 127 ∧ 2', 2);
  Test('175 ∧ 117 ∧ 92 ∧ 115 ∧ 89', 0);
  Test('519846545615848421 ∧ 251635489456321632', 231052631100129376);
  Test('5156113200231656 ∧ 120651616515132123', 4423942344);
  Test('4565165560505615611 ∧ 156151135151510131 ∧ 86515121545115612', 562955393450064);
  Test('4565165560505615611 ∧ 156151135151510131 ∧ 86515121545115612 ∧ 0', 0);

  Test('157 ∧ true', failure, 'An object of type integer was expected as argument 2, but an object of type boolean was given.');
  Test('219 ∧ 157 ∧ "cat"', failure, 'An object of type integer was expected as argument 3, but an object of type string was given.');

  // Bitwise or (bitwise disjunction)

  Test('157 ∨ 197', 221);
  Test('219 ∨ 85', 223);
  Test('219 ∨ 131', 219);
  Test('91 ∨ 64', 91);
  Test('87 ∨ 32', 119);
  Test('22355 ∨ 42621', 63359);
  Test('55964 ∨ 20506', 55966);
  Test('54743 ∨ 5188', 54743);
  Test('45731 ∨ 38101 ∨ 20981', 63479);
  Test('219 ∨ 159 ∨ 187', 255);
  Test('219 ∨ 187 ∨ 106 ∨ 106 ∨ 127', 255);
  Test('219 ∨ 187 ∨ 106 ∨ 106 ∨ 127 ∨ 1', 255);
  Test('219 ∨ 187 ∨ 106 ∨ 106 ∨ 127 ∨ 2', 255);
  Test('175 ∨ 117 ∨ 92 ∨ 115 ∨ 89', 255);
  Test('519846545615848421 ∨ 251635489456321632', 540429403972040677);
  Test('5156113200231656 ∨ 120651616515132123', 125807725291421435);
  Test('4565165560505615611 ∨ 156151135151510131 ∨ 86515121545115612', 4574531235602233343);
  Test('4565165560505615611 ∨ 156151135151510131 ∨ 86515121545115612 ∨ 0', 4574531235602233343);

  Test('157 ∨ true', failure, 'An object of type integer was expected as argument 2, but an object of type boolean was given.');
  Test('219 ∨ 157 ∨ "cat"', failure, 'An object of type integer was expected as argument 3, but an object of type string was given.');

  // Bitwise xor

  Test('47547 ⊻ 23421', 58054);
  Test('56053 ⊻ 29622', 43331);
  Test('57093 ⊻ 30173', 43736);
  Test('222 ⊻ 182', 104);
  Test('221 ⊻ 0', 221);
  Test('846515616156056061 ⊻ 185151561564412123', 661649996338864422);
  Test('510531025361320112 ⊻ 450645161561', 510530924772937449);

  Test('157 ⊻ true', failure, 'An object of type integer was expected as argument 2, but an object of type boolean was given.');
  Test('157 ⊻ "nargle"', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');

  // Bitwise not (bitwise negation)

  Test('¬0', -1);
  Test('¬-1', 0);
  Test('¬871065116153', -871065116154);
  Test('¬61651216505160651', -61651216505160652);
  Test('¬156156123110567121', -156156123110567122);

  // Misc

  Test('56123 ∧ ¬2048', 54075);
  Test('123 ∧ 255', 123);
  Test('123 ∧ 0', 0);
  Test('123 ∨ 255', 255);
  Test('123 ∨ 0', 123);





  //
  //
  //  CHAPTER 5
  //  Vectors and matrices.
  //
  //

  Chapter('Vectors and matrices');

  // Sizes

  Test('a ≔ ❨21❩; u ≔ ❨1, 0❩; v ≔ ❨3, 0, −1❩; w ≔ ❨3 + 5⋅i, −7, 2⋅i, 1 − i❩; A ≔ ❨❨1, 2❩, ❨3, 4❩❩; B ≔ ❨❨5, 3, 1❩, ❨1, 6, 8❩❩; C ≔ ❨❨2 + i, 7❩, ❨i, 1 − 2⋅i❩❩; D ≔ ❨❨1, 6, 8, 2, 5❩❩; E ≔ ❨❨4❩, ❨6❩, ❨i❩❩; F ≔ ❨❨6❩❩;', null);

  Test('dim(a)', 1);
  Test('dim(u)', 2);
  Test('dim(v)', 3);
  Test('dim(w)', 4);

  Test('dim(A)', failure, 'An object of type vector was expected as argument 1, but an object of type real matrix was given.');
  Test('dim(5)', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');

  for _i := 1 to 100 do
    Test(Format('dim(ZeroVector(%d))', [_i]), _i);

  Test('size(π)', ASOSize(1, 1)); Test('height(π)', 1); Test('width(π)', 1);
  Test('size(a)', ASOSize(1, 1)); Test('height(a)', 1); Test('width(a)', 1);
  Test('size(u)', ASOSize(2, 1)); Test('height(u)', 2); Test('width(u)', 1);
  Test('size(v)', ASOSize(3, 1)); Test('height(v)', 3); Test('width(v)', 1);
  Test('size(w)', ASOSize(4, 1)); Test('height(w)', 4); Test('width(w)', 1);
  Test('size(A)', ASOSize(2, 2)); Test('height(A)', 2); Test('width(A)', 2);
  Test('size(B)', ASOSize(2, 3)); Test('height(B)', 2); Test('width(B)', 3);
  Test('size(C)', ASOSize(2, 2)); Test('height(C)', 2); Test('width(C)', 2);
  Test('size(D)', ASOSize(1, 5)); Test('height(D)', 1); Test('width(D)', 5);
  Test('size(E)', ASOSize(3, 1)); Test('height(E)', 3); Test('width(E)', 1);
  Test('size(F)', ASOSize(1, 1)); Test('height(F)', 1); Test('width(F)', 1);

  for _i := 1 to 10 do
    for _j := 1 to 10 do
      Test(Format('size(ZeroMatrix(%d, %d))', [_i, _j]), ASOSize(_i, _j));

  // Real and imaginary parts

  Test('Re(a)', [21]);
  Test('Im(a)', [0]);

  Test('Re(u)', [1, 0]);
  Test('Im(u)', [0, 0]);

  Test('Re(v)', [3, 0, -1]);
  Test('Im(v)', [0, 0, 0]);

  Test('Re(w)', [3, -7, 0, 1]);
  Test('Im(w)', [5, 0, 2, -1]);

  Test('Re(A)', 2, [1, 2, 3, 4]);
  Test('Im(A)', 2, [0, 0, 0, 0]);

  Test('Re(B)', 3, [5, 3, 1, 1, 6, 8]);
  Test('Im(B)', 3, [0, 0, 0, 0, 0, 0]);

  Test('Re(C)', 2, [2, 7, 0, 1]);
  Test('Im(C)', 2, [1, 0, 1, -2]);

  Test('Re(D)', 5, [1, 6, 8, 2, 5]);
  Test('Im(D)', 5, [0, 0, 0, 0, 0]);

  Test('Re(E)', 1, [4, 6, 0]);
  Test('Im(E)', 1, [0, 0, 1]);

  Test('Re(F)', 1, [6]);
  Test('Im(F)', 1, [0]);

  // Vector constructors

  Test('❨1❩', [1]);
  Test('❨1, 0❩', [1, 0]);
  Test('❨1, 0, 0❩', [1, 0, 0]);

  Test('❨i❩', [ImaginaryUnit]);
  Test('❨1, i❩', [1, ImaginaryUnit]);
  Test('❨1, 0, -i❩', [1, 0, -ImaginaryUnit]);

  Test('❨❩', ESyntaxException, 'Empty bracket at column 1.');

  Test('ZeroVector(1)', [0]);
  Test('ZeroVector(2)', [0, 0]);
  Test('ZeroVector(3)', [0, 0, 0]);
  Test('ZeroVector(4)', [0, 0, 0, 0]);
  Test('ZeroVector(5)', [0, 0, 0, 0, 0]);
  Test('ZeroVector(6)', [0, 0, 0, 0, 0, 0]);

  Test('ZeroVector(100)', ASO(ZeroVector(100)));
  Test('v ≔ ZeroVector(1000)', ASO(ZeroVector(1000)));

  Test('type(v)', 'real vector');
  Test('dim(v)', 1000);
  Test('size(v)', ASOSize(1000, 1));
  Test('sum(v)', 0.0);
  Test('product(v)', 0.0);
  Test('average(v)', 0.0);
  Test('min(v)', 0.0);
  Test('max(v)', 0.0);
  Test('delete(v)', success);

  Test('ZeroVector(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('ZeroVector(-1)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');
  Test('ZeroVector(5, 6)', failure, 'Too many arguments.');

  Test('ComplexZeroVector(1)', [TASC(0)]);
  Test('ComplexZeroVector(2)', [TASC(0), 0]);
  Test('ComplexZeroVector(3)', [TASC(0), 0, 0]);
  Test('ComplexZeroVector(4)', [TASC(0), 0, 0, 0]);
  Test('ComplexZeroVector(5)', [TASC(0), 0, 0, 0, 0]);
  Test('ComplexZeroVector(6)', [TASC(0), 0, 0, 0, 0, 0]);

  Test('ComplexZeroVector(100)', ASO(ComplexZeroVector(100)));
  Test('v ≔ ComplexZeroVector(1000)', ASO(ComplexZeroVector(1000)));

  Test('type(v)', 'complex vector');
  Test('dim(v)', 1000);
  Test('size(v)', ASOSize(1000, 1));
  Test('sum(v)', TASC(0.0));
  Test('product(v)', TASC(0.0));
  Test('average(v)', TASC(0.0));
  Test('min(v)', 0.0);
  Test('max(v)', 0.0);
  Test('delete(v)', success);

  Test('ComplexZeroVector(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('ComplexZeroVector(-1)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');
  Test('ComplexZeroVector(5, 6)', failure, 'Too many arguments.');

  Test('BasisVector(1, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('BasisVector(1, 1)', [1]);
  Test('BasisVector(1, 2)', failure, 'Invalid unit vector index.');

  Test('BasisVector(2, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('BasisVector(2, 1)', [1, 0]);
  Test('BasisVector(2, 2)', [0, 1]);
  Test('BasisVector(2, 3)', failure, 'Invalid unit vector index.');

  Test('BasisVector(3, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('BasisVector(3, 1)', [1, 0, 0]);
  Test('BasisVector(3, 2)', [0, 1, 0]);
  Test('BasisVector(3, 3)', [0, 0, 1]);
  Test('BasisVector(3, 4)', failure, 'Invalid unit vector index.');

  Test('BasisVector(4, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('BasisVector(4, 1)', [1, 0, 0, 0]);
  Test('BasisVector(4, 2)', [0, 1, 0, 0]);
  Test('BasisVector(4, 3)', [0, 0, 1, 0]);
  Test('BasisVector(4, 4)', [0, 0, 0, 1]);
  Test('BasisVector(4, 5)', failure, 'Invalid unit vector index.');

  Test('BasisVector(0, 0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('BasisVector(-1, 0)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');
  Test('BasisVector(5, 2, 3)', failure, 'Too many arguments.');
  Test('BasisVector(5)', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('v ≔ BasisVector(1000, 359)', ASO(UnitVector(1000, Pred(359))));
  Test('type(v)', 'real vector');
  Test('dim(v)', 1000);
  Test('size(v)', ASOSize(1000, 1));
  Test('sum(v)', 1.0);
  Test('product(v)', 0.0);
  Test('average(v)', 1/1000);
  Test('min(v)', 0.0);
  Test('max(v)', 1.0);
  Test('delete(v)', success);

  for _i := 1 to 100 do
  begin
    Test(Format('v ≔ RandomVector(%d);', [_i]), null);
    Test('dim(v)', _i);
    Test('min(v) ≥ 0', True);
    Test('max(v) < 1', True);
  end;

  Test('avg ≔ mean(RandomVector(10000)); (avg > 0.48) ∧ (avg < 0.52)', True);
  Test('delete(v); delete(avg)', success);

  Test('RandomVector(5, 3)', failure, 'Too many arguments.');
  Test('RandomVector()', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('RandomVector(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');

  for _i := 1 to 100 do
  begin
    Test(Format('v ≔ RandomSignedVector(%d);', [_i]), null);
    Test('dim(v)', _i);
    Test('min(v) ≥ -1', True);
    Test('max(v) < 1', True);
  end;

  Test('avg ≔ mean(RandomSignedVector(10000)); abs(avg) < 0.02', True);
  Test('delete(v); delete(avg)', success);

  Test('RandomSignedVector(5, 3)', failure, 'Too many arguments.');
  Test('RandomSignedVector()', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('RandomSignedVector(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');

  for _i := 1 to 100 do
  begin
    Test(Format('v ≔ RandomIntVector(%d, 17, 21);', [_i]), null);
    Test('dim(v)', _i);
    Test('∀(v, x↦IsInteger(x))', True);
    Test('min(v) ≥ 17', True);
    Test('max(v) ≤ 20', True);
  end;

  Test('avg ≔ mean(RandomIntVector(10000, 17, 21)); (avg > 18.4) ∧ (avg < 18.6)', True);
  Test('delete(v); delete(avg)', success);

  Test('RandomIntVector()', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('RandomIntVector(10)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('RandomIntVector(10, 5)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('RandomIntVector(10, 5, 8, 9)', failure, 'Too many arguments.');
  Test('RandomIntVector(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');

  for _i := 1 to 100 do
  begin
    Test(Format('v ≔ RandomIntVector(%d, 17, 18);', [_i]), null);
    Test('dim(v)', _i);
    Test('∀(v, x↦x=17)', True);
    Test('min(v) = 17', True);
    Test('max(v) = 17', True);
  end;

  Test('delete(v)', success);

  Test('RandomIntVector(100, 4, 4)', failure, 'The set of possible values is empty.');
  Test('RandomIntVector(100, 4, 3)', failure, 'The set of possible values is empty.');
  Test('RandomIntVector(100, 4, -4)', failure, 'The set of possible values is empty.');

  Test('SequenceVector()', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('SequenceVector(5)', [1, 2, 3, 4, 5]);
  Test('SequenceVector(10)', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  Test('SequenceVector(1)', [1]);
  Test('SequenceVector(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');

  Test('SequenceVector(1, 5)', [1, 2, 3, 4, 5]);
  Test('SequenceVector(0, 5)', [0, 1, 2, 3, 4, 5]);
  Test('SequenceVector(-3, 5)', [-3, -2, -1, 0, 1, 2, 3, 4, 5]);
  Test('SequenceVector(-10, -5)', [-10, -9, -8, -7, -6, -5]);
  Test('SequenceVector(7, 10)', [7, 8, 9, 10]);
  Test('SequenceVector(21, 21)', [21]);

  Test('SequenceVector(21, 20)', [21, 20]);

  Test('SequenceVector(0, 100, 10)', [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]);
  Test('SequenceVector(0, 100, 20)', [0, 20, 40, 60, 80, 100]);
  Test('SequenceVector(0, 100, 25)', [0, 25, 50, 75, 100]);
  Test('SequenceVector(0, 100, 50)', [0, 50, 100]);
  Test('SequenceVector(0, 100, 100)', [0, 100]);
  Test('SequenceVector(0, 100, 200)', [0]);

  Test('SequenceVector(-10, 10, 5)', [-10, -5, 0, 5, 10]);

  Test('SequenceVector(0, 100, 10, 5)', failure, 'Too many arguments.');
  Test('SequenceVector(0.5, 100, 25)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('SequenceVector(0, 100.5, 25)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');
  Test('SequenceVector(0, 100, 25.6)', failure, 'An object of type integer was expected as argument 3, but an object of type real number was given.');

  // Matrix constructors

  Test('❨❨1, 2❩, ❨3, 4❩❩', 2, [1, 2, 3, 4]);
  Test('❨❨1, 2, 3❩, ❨10, 20, 30❩❩', 3, [1, 2, 3, 10, 20, 30]);
  Test('❨❨5, 4, 3, 2, 1❩❩', 5, [5, 4, 3, 2, 1]);
  Test('❨❨1❩, ❨2❩, ❨10❩❩', 1, [1, 2, 10]);
  Test('❨❨1, 3❩, ❨5, 7❩, ❨9, 0❩❩', 2, [1, 3, 5, 7, 9, 0]);

  Test('❨❨1, 3❩, ❨5, 7❩, ❨9, 0, 6❩❩', failure, 'Attempt to create a non-rectangular matrix.');

  Test('❨❨1, 2❩, ❨i, 4❩❩', 2, [1, 2, ImaginaryUnit, 4]);
  Test('❨❨i, 2, 3❩, ❨10, 20, 30❩❩', 3, [ImaginaryUnit, 2, 3, 10, 20, 30]);
  Test('❨❨5, 4, 3, 2, 1 + i❩❩', 5, [5, 4, 3, 2, 1 + ImaginaryUnit]);
  Test('❨❨1❩, ❨2❩, ❨i❩❩', 1, [1, 2, ImaginaryUnit]);
  Test('❨❨1, 3❩, ❨5, i/7❩, ❨9, 0❩❩', 2, [1, 3, 5, ImaginaryUnit/7, 9, 0]);

  Test('❨❨❩❩', ESyntaxException, 'Empty bracket at column 2.');

  Test('❨❨1, 3❩, ❨5, 7❩, ❨9, i, 6❩❩', failure, 'Attempt to create a non-rectangular matrix.');

  Test('matrix(2, 1, 2, 3, -4)', 2, [1, 2, 3, -4]);
  Test('matrix(3, 1, 2, 3, 10, 20, 30)', 3, [1, 2, 3, 10, 20, 30]);
  Test('matrix(3, 10, 20, 30, 40, 50, 60, 70, 80, 90)', 3, [10, 20, 30, 40, 50, 60, 70, 80, 90]);
  Test('matrix(2, 10, 20, 100, 200, 1000, 2000)', 2, [10, 20, 100, 200, 1000, 2000]);
  Test('matrix(1, -1, 1, 0)', 1, [-1, 1, 0]);

  Test('A ≔ matrix(1, 1, 2, 3, 4, 5, 6)', 1, [1, 2, 3, 4, 5, 6]);
  Test('size(A)', ASOSize(6, 1));
  Test('A ≔ matrix(2, 1, 2, 3, 4, 5, 6)', 2, [1, 2, 3, 4, 5, 6]);
  Test('size(A)', ASOSize(3, 2));
  Test('A ≔ matrix(3, 1, 2, 3, 4, 5, 6)', 3, [1, 2, 3, 4, 5, 6]);
  Test('size(A)', ASOSize(2, 3));
  Test('A ≔ matrix(6, 1, 2, 3, 4, 5, 6)', 6, [1, 2, 3, 4, 5, 6]);
  Test('size(A)', ASOSize(1, 6));
  Test('type(A)', 'real matrix');
  Test('delete(A)', success);

  Test('matrix(4, 1, 2, 3, 4, 5, 6)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('matrix(5, 1, 2, 3, 4, 5, 6)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('matrix(7, 1, 2, 3, 4, 5, 6)', failure, 'Attempt to create a non-rectangular matrix.');

  Test('matrix(0, 1, 2, 3, 4, 5, 6)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('matrix(-1, 1, 2, 3, 4, 5, 6)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');

  Test('matrix(1, 21)', 1, [21]);

  Test('matrix(2, i, 2, 3, -4)', 2, [ImaginaryUnit, 2, 3, -4]);
  Test('matrix(3, i, 2, 3, 10, 20, 30)', 3, [ImaginaryUnit, 2, 3, 10, 20, 30]);
  Test('matrix(3, i, 20, 30, 40, 50, 60, 70, 80, 90)', 3, [ImaginaryUnit, 20, 30, 40, 50, 60, 70, 80, 90]);
  Test('matrix(2, i, 20, 100, 200, 1000, 2000)', 2, [ImaginaryUnit, 20, 100, 200, 1000, 2000]);
  Test('matrix(1, -1, 1, i/2)', 1, [-1, 1, ImaginaryUnit/2]);

  Test('A ≔ matrix(1, i, 2, 3, 4, 5, 6)', 1, [ImaginaryUnit, 2, 3, 4, 5, 6]);
  Test('size(A)', ASOSize(6, 1));
  Test('A ≔ matrix(2, i, 2, 3, 4, 5, 6)', 2, [ImaginaryUnit, 2, 3, 4, 5, 6]);
  Test('size(A)', ASOSize(3, 2));
  Test('A ≔ matrix(3, i, 2, 3, 4, 5, 6)', 3, [ImaginaryUnit, 2, 3, 4, 5, 6]);
  Test('size(A)', ASOSize(2, 3));
  Test('A ≔ matrix(6, i, 2, 3, 4, 5, 6)', 6, [ImaginaryUnit, 2, 3, 4, 5, 6]);
  Test('size(A)', ASOSize(1, 6));
  Test('type(A)', 'complex matrix');
  Test('delete(A)', success);

  Test('matrix(4, i, 2, 3, 4, 5, 6)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('matrix(5, i, 2, 3, 4, 5, 6)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('matrix(7, i, 2, 3, 4, 5, 6)', failure, 'Attempt to create a non-rectangular matrix.');

  Test('matrix(0, i, 2, 3, 4, 5, 6)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('matrix(-1, i, 2, 3, 4, 5, 6)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');

  Test('matrix(1, i/21)', 1, [ImaginaryUnit/21]);

  Test('matrix(2, ''(1, 2, 3, 4))', 2, [1, 2, 3, 4]);
  Test('matrix(2, ToList(❨1, 2, 3, 4, 5, 6❩))', 2, [1, 2, 3, 4, 5, 6]);
  Test('matrix(2, ToList(❨1, 2, 3, 4, 5, 6, 7❩))', failure, 'Attempt to create a non-rectangular matrix.');

  Test('ZeroMatrix(1, 1)', 1, [0]);
  Test('ZeroMatrix(2, 2)', 2, [0, 0, 0, 0]);
  Test('ZeroMatrix(2, 3)', 3, [0, 0, 0, 0, 0, 0]);
  Test('ZeroMatrix(4, 1)', 1, [0, 0, 0, 0]);
  Test('ZeroMatrix(1, 3)', 3, [0, 0, 0]);

  Test('A ≔ ZeroMatrix(3, 2)', 2, [0, 0, 0, 0, 0, 0]);
  Test('type(A)', 'real matrix');
  Test('size(A)', ASOSize(3, 2));
  Test('sum(A)', 0.0);
  Test('product(A)', 0.0);
  Test('average(A)', 0.0);
  Test('min(A)', 0.0);
  Test('max(A)', 0.0);
  Test('delete(A)', success);

  Test('ZeroMatrix(1)', 1, [0]);
  Test('ZeroMatrix(2)', 2, [0, 0, 0, 0]);
  Test('ZeroMatrix(3)', 3, [0, 0, 0, 0, 0, 0, 0, 0, 0]);
  Test('ZeroMatrix(4)', 4, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]);

  Test('ZeroMatrix()', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('ZeroMatrix(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('ZeroMatrix(-1)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');
  Test('ZeroMatrix(0, 4)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('ZeroMatrix(4, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('ZeroMatrix(4, 4, 6)', failure, 'Too many arguments.');

  Test('ComplexZeroMatrix(1, 1)', 1, [TASC(0)]);
  Test('ComplexZeroMatrix(2, 2)', 2, [TASC(0), 0, 0, 0]);
  Test('ComplexZeroMatrix(2, 3)', 3, [TASC(0), 0, 0, 0, 0, 0]);
  Test('ComplexZeroMatrix(4, 1)', 1, [TASC(0), 0, 0, 0]);
  Test('ComplexZeroMatrix(1, 3)', 3, [TASC(0), 0, 0]);

  Test('A ≔ ComplexZeroMatrix(3, 2)', 2, [TASC(0), 0, 0, 0, 0, 0]);
  Test('type(A)', 'complex matrix');
  Test('size(A)', ASOSize(3, 2));
  Test('sum(A)', TASC(0));
  Test('product(A)', TASC(0));
  Test('average(A)', TASC(0));
  Test('min(A)', 0.0);
  Test('max(A)', 0.0);
  Test('delete(A)', success);

  Test('ComplexZeroMatrix(1)', 1, [TASC(0)]);
  Test('ComplexZeroMatrix(2)', 2, [TASC(0), 0, 0, 0]);
  Test('ComplexZeroMatrix(3)', 3, [TASC(0), 0, 0, 0, 0, 0, 0, 0, 0]);
  Test('ComplexZeroMatrix(4)', 4, [TASC(0), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]);

  Test('ComplexZeroMatrix()', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('ComplexZeroMatrix(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('ComplexZeroMatrix(-1)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');
  Test('ComplexZeroMatrix(0, 4)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('ComplexZeroMatrix(4, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('ComplexZeroMatrix(4, 4, 6)', failure, 'Too many arguments.');

  Test('IdentityMatrix(1)', 1, [1]);
  Test('IdentityMatrix(2)', 2, [1, 0, 0, 1]);
  Test('IdentityMatrix(3)', 3, [1, 0, 0, 0, 1, 0, 0, 0, 1]);
  Test('IdentityMatrix(4)', 4, [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]);

  Test('A ≔ IdentityMatrix(100)', ASO(IdentityMatrix(100)));
  Test('type(A)', 'real matrix');
  Test('size(A)', ASOSize(100, 100));
  Test('sum(A)', 100.0);
  Test('product(A)', 0.0);
  Test('average(A)', 1/100);
  Test('min(A)', 0.0);
  Test('max(A)', 1.0);
  Test('delete(A)', success);

  Test('IdentityMatrix(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('IdentityMatrix()', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('IdentityMatrix(3, 4)', failure, 'Too many arguments.');

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ RandomMatrix(%d);', [_i]), null);
    Test('size(A)', ASOSize(_i, _i));
    Test('min(A) ≥ 0', True);
    Test('max(A) < 1', True);
  end;

  Test('A ≔ RandomMatrix(100); avg ≔ mean(A); (avg > 0.49) ∧ (avg < 0.51)', True);
  Test('delete(A)', success);

  for _i := 1 to 10 do
    for _j := 1 to 10 do
    begin
      Test(Format('A ≔ RandomMatrix(%d, %d);', [_i, _j]), null);
      Test('size(A)', ASOSize(_i, _j));
      Test('min(A) ≥ 0', True);
      Test('max(A) < 1', True);
    end;

  Test('A ≔ RandomMatrix(70, 140); avg ≔ mean(A); (avg > 0.49) ∧ (avg < 0.51)', True);
  Test('delete(A)', success);

  Test('RandomMatrix()', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('RandomMatrix(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('RandomMatrix(-1)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');
  Test('RandomMatrix(0, 5)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('RandomMatrix(5, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('RandomMatrix(5, 6, 7)', failure, 'Too many arguments.');

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ RandomIntMatrix(%d, 7, 13);', [_i]), null);
    Test('size(A)', ASOSize(_i, _i));
    Test('∀(A, x↦IsInteger(x))', True);
    Test('min(A) ≥ 7', True);
    Test('max(A) ≤ 12', True);
  end;

  Test('A ≔ RandomIntMatrix(100, 7, 13); avg ≔ mean(A); (avg > 9.4) ∧ (avg < 9.6)', True);
  Test('delete(A)', success);

  for _i := 1 to 10 do
    for _j := 1 to 10 do
    begin
      Test(Format('A ≔ RandomIntMatrix(%d, %d, 7, 13);', [_i, _j]), null);
      Test('size(A)', ASOSize(_i, _j));
      Test('∀(A, x↦IsInteger(x))', True);
      Test('min(A) ≥ 7', True);
      Test('max(A) ≤ 12', True);
    end;

  Test('A ≔ RandomIntMatrix(70, 140, 7, 13); avg ≔ mean(A); (avg > 9.4) ∧ (avg < 9.6)', True);
  Test('delete(A)', success);

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ RandomIntMatrix(%d, 7, 8);', [_i]), null);
    Test('size(A)', ASOSize(_i, _i));
    Test('∀(A, x↦x=7)', True);
    Test('min(A) = 7', True);
    Test('max(A) = 7', True);
  end;

  for _i := 1 to 10 do
    for _j := 1 to 10 do
    begin
      Test(Format('A ≔ RandomIntMatrix(%d, %d, 7, 8);', [_i, _j]), null);
      Test('size(A)', ASOSize(_i, _j));
      Test('∀(A, x↦x=7)', True);
      Test('min(A) = 7', True);
      Test('max(A) = 7', True);
    end;

  Test('RandomIntMatrix(100, 7, 7)', failure, 'The set of possible values is empty.');
  Test('RandomIntMatrix(70, 120, 7, 7)', failure, 'The set of possible values is empty.');

  Test('RandomIntMatrix(100, 7, 2)', failure, 'The set of possible values is empty.');
  Test('RandomIntMatrix(70, 120, 7, 2)', failure, 'The set of possible values is empty.');

  Test('RandomIntMatrix(0, 1, 10)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('RandomIntMatrix(0, 10, 1, 10)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('RandomIntMatrix(10, 0, 1, 10)', failure, 'A positive integer was expected as argument 2, but "0" was given.');

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ RandomIntMatrix(%d, -10, 11);', [_i]), null);
    Test('size(A)', ASOSize(_i, _i));
    Test('∀(A, x↦IsInteger(x))', True);
    Test('min(A) ≥ -10', True);
    Test('max(A) ≤ 10', True);
  end;

  Test('A ≔ RandomIntMatrix(100, -10, 11); avg ≔ mean(A); abs(avg) < 0.25', True);
  Test('delete(A)', success);

  for _i := 1 to 10 do
    for _j := 1 to 10 do
    begin
      Test(Format('A ≔ RandomIntMatrix(%d, %d, -10, 11);', [_i, _j]), null);
      Test('size(A)', ASOSize(_i, _j));
      Test('∀(A, x↦IsInteger(x))', True);
      Test('min(A) ≥ -10', True);
      Test('max(A) ≤ 10', True);
    end;

  Test('A ≔ RandomIntMatrix(70, 140, -10, 11); avg ≔ mean(A); abs(avg) < 0.25', True);
  Test('delete(A)', success);

  Test('RandomIntMatrix(10, 0, 1, 10)', failure, 'A positive integer was expected as argument 2, but "0" was given.');

  Test('RandomIntMatrix()', failure, 'There is no version of this function that can take 0 argument(s).');
  Test('RandomIntMatrix(10)', failure, 'There is no version of this function that can take 1 argument(s).');
  Test('RandomIntMatrix(10, 20)', failure, 'There is no version of this function that can take 2 argument(s).');
  Test('RandomIntMatrix(10, 10, 2, 5, 1)', failure, 'There is no version of this function that can take 5 argument(s).');

  Test('MatFromCols(❨1, 1, 0❩, ❨0, 2, 0❩, ❨0, 0, 3❩)', 3, [1, 0, 0, 1, 2, 0, 0, 0, 3]);
  Test('MatFromCols(❨1, 0❩, ❨2, 0❩, ❨3, 0❩, ❨4, 0❩, ❨5, 10❩)', 5, [1, 2, 3, 4, 5, 0, 0, 0, 0, 10]);
  Test('MatFromCols(❨1, 2, 3❩, ❨10, 20, 30❩)', 2, [1, 10, 2, 20, 3, 30]);
  Test('MatFromCols(❨1, 0❩)', 1, [1, 0]);

  Test('MatFromCols(❨1, 0, 0❩, ❨0, 2, 0❩, ❨0, 3❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromCols(❨1, 0❩, ❨2, 0❩, ❨3, 0❩, ❨4, 0❩, ❨5, 10, 25❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromCols(❨1, 2, 3, 4❩, ❨10, 20, 30❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromCols(❨1, 0❩, ❨1❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromCols(❨1, 0, 0❩, ❨0, 2, 0, 1❩, ❨0, 3❩)', failure, 'Attempt to create a non-rectangular matrix.');

  Test('MatFromCols()', failure, 'A matrix must have size at least 1×1.');

  Test('MatFromCols(1)', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');

  Test('MatFromCols(❨i, 0, 0❩, ❨0, 2, 0❩, ❨0, 0, 3❩)', 3, [ImaginaryUnit, 0, 0, 0, 2, 0, 0, 0, 3]);
  Test('MatFromCols(❨1, 0, 0❩, ❨0, 2, i❩, ❨0, 0, 3❩)', 3, [1, 0, 0, 0, 2, 0, 0, ImaginaryUnit, 3]);
  Test('MatFromCols(❨1, 0, 0❩, ❨0, 2, 0❩, ❨i, 0, 3❩)', 3, [1, 0, ImaginaryUnit, 0, 2, 0, 0, 0, 3]);

  Test('MatFromCols(❨i, 0❩, ❨2, 0❩, ❨3, 0❩, ❨4, 0❩, ❨5, 10❩)', 5, [ImaginaryUnit, 2, 3, 4, 5, 0, 0, 0, 0, 10]);
  Test('MatFromCols(❨i, 2, 3❩, ❨10, 20, 30❩)', 2, [ImaginaryUnit, 10, 2, 20, 3, 30]);
  Test('MatFromCols(❨i, 0❩)', 1, [ImaginaryUnit, 0]);

  Test('MatFromCols(❨i, 0, 0❩, ❨0, 2, 0❩, ❨0, 3❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromCols(❨i, 0❩, ❨2, 0❩, ❨3, 0❩, ❨4, 0❩, ❨5, 10, 25❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromCols(❨i, 2, 3, 4❩, ❨10, 20, 30❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromCols(❨i, 0❩, ❨1❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromCols(❨i, 0, 0❩, ❨0, 2, 0, 1❩, ❨0, 3❩)', failure, 'Attempt to create a non-rectangular matrix.');

  Test('MatFromCols(i)', failure, 'An object of type vector was expected as argument 1, but an object of type complex number was given.');

  Test('MatFromRows(❨1, 2, 3❩, ❨10, 20, 30❩, ❨0, 0, 1❩)', 3, [1, 2, 3, 10, 20, 30, 0, 0, 1]);
  Test('MatFromRows(❨1, 2❩, ❨3, 4❩, ❨5, 6❩, ❨7, 8❩)', 2, [1, 2, 3, 4, 5, 6, 7, 8]);
  Test('MatFromRows(❨1, 2, 3❩, ❨10, 20, 30❩)', 3, [1, 2, 3, 10, 20, 30]);
  Test('MatFromRows(❨10, 11, 12, 13❩)', 4, [10, 11, 12, 13]);

  Test('MatFromRows(❨1, 2, 3❩, ❨10, 20, 30, 40❩, ❨0, 0, 1❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromRows(❨1, 2, 3❩, ❨10, 20, 30❩, ❨0, 0❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromRows(❨1, 2❩, ❨3, 4❩, ❨5, 6, 7❩, ❨7❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromRows(❨1, 2, 3❩, ❨10, 20, 30, 40❩)', failure, 'Attempt to create a non-rectangular matrix.');

  Test('MatFromRows()', failure, 'A matrix must have size at least 1×1.');

  Test('MatFromRows(1)', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');

  Test('MatFromRows(❨i, 2, 3❩, ❨10, 20, 30❩, ❨0, 0, 1❩)', 3, [ImaginaryUnit, 2, 3, 10, 20, 30, 0, 0, 1]);
  Test('MatFromRows(❨1, 2, 3❩, ❨10, 20, 30❩, ❨0, 0, i❩)', 3, [1, 2, 3, 10, 20, 30, 0, 0, ImaginaryUnit]);
  Test('MatFromRows(❨i, 2❩, ❨3, 4❩, ❨5, 6❩, ❨7, 8❩)', 2, [ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8]);
  Test('MatFromRows(❨i, 2, 3❩, ❨10, 20, 30❩)', 3, [ImaginaryUnit, 2, 3, 10, 20, 30]);
  Test('MatFromRows(❨10, 11, i, 13❩)', 4, [10, 11, ImaginaryUnit, 13]);

  Test('MatFromRows(❨1, 2, 3❩, ❨10, 20, 30, 40❩, ❨0, 0, i❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromRows(❨i, 2, 3❩, ❨10, 20, 30❩, ❨0, 0❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromRows(❨i, 2❩, ❨3, 4❩, ❨5, 6, 7❩, ❨7❩)', failure, 'Attempt to create a non-rectangular matrix.');
  Test('MatFromRows(❨i, 2, 3❩, ❨10, 20, 30, 40❩)', failure, 'Attempt to create a non-rectangular matrix.');

  Test('MatFromRows()', failure, 'A matrix must have size at least 1×1.');

  Test('MatFromRows(i)', failure, 'An object of type vector was expected as argument 1, but an object of type complex number was given.');

  Test('MatFromBlocks(2, ❨❨1, 2❩, ❨3, 4❩❩, ❨❨10, 20❩, ❨30, 40❩❩, ❨❨12, 13❩, ❨14, 15❩❩, ❨❨100, 200❩, ❨300, 400❩❩)',
    4, [1, 2, 10, 20, 3, 4, 30, 40, 12, 13, 100, 200, 14, 15, 300, 400]);

  Test('MatFromBlocks(2, ❨❨1, 2❩, ❨3, 4❩❩, ❨❨10, 20, 30❩, ❨40, 50, 60❩❩, ❨❨100, 200❩, ❨300, 400❩❩, ❨❨1, 2, 3❩, ❨4, 5, 6❩❩)',
    5, [1, 2, 10, 20, 30, 3, 4, 40, 50, 60, 100, 200, 1, 2, 3, 300, 400, 4, 5, 6]);

  Test('MatFromBlocks(2, ❨❨1, 2❩, ❨3, 4❩❩, ❨❨10, 20, 30❩, ❨40, 50, 60❩❩, ❨❨1, 2, 3❩, ❨4, 5, 6❩❩,  ❨❨100, 200❩, ❨300, 400❩❩)',
    failure, 'Invalid block at position (2, 1), index 3.');

  Test('MatFromBlocks(2, ❨❨1, 2❩, ❨3, 4❩❩, ❨❨10, 20, 30❩, ❨40, 50, 60❩❩, ❨❨100, 200❩, ❨300, 400❩❩, ❨❨1, 2, 3❩❩)',
    failure, 'Invalid block at position (2, 2), index 4.');

  Test('MatFromBlocks(2, ❨❨1, 2❩, ❨3, 4❩❩, ❨❨10, 20, 30❩, ❨40, 50, 60❩, ❨70, 80, 90❩❩, ❨❨100, 200❩, ❨300, 400❩❩, ❨❨1, 2, 3❩, ❨4, 5, 6❩❩)',
    failure, 'Invalid block at position (1, 2), index 2.');

  Test('MatFromBlocks(3, ❨❨1, 2❩, ❨3, 4❩❩, IdentityMatrix(2), IdentityMatrix(2), ZeroMatrix(2), ❨❨10, 20❩, ❨30, 40❩❩, IdentityMatrix(2))',
    6, [1, 2, 1, 0, 1, 0, 3, 4, 0, 1, 0, 1, 0, 0, 10, 20, 1, 0, 0, 0, 30, 40, 0, 1]);

  Test('I ≔ IdentityMatrix(2); MatFromBlocks(4, I, I, I, I, I, I, I, I)',
    8, [1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1]);

  Test('delete(I)', success);

  Test('MatFromBlocks(3, ❨❨1❩, ❨2❩❩, ❨❨10❩, ❨20❩❩, ❨❨100❩, ❨200❩❩)',
    3, [1, 10, 100, 2, 20, 200]);

  Test('MatFromBlocks(1, ❨❨1, 2❩, ❨3, 4❩❩, ❨❨10, 100❩, ❨100, 10❩❩, ❨❨5, 6❩, ❨7, 8❩❩)',
    2, [1, 2, 3, 4, 10, 100, 100, 10, 5, 6, 7, 8]);

  Test('MatFromBlocks(2, ❨❨1, 2❩, ❨3, 4❩❩, ❨❨100, 100❩, ❨200, 200❩❩, ❨❨10, 12❩, ❨14, 16❩❩)',
    4, [1, 2, 100, 100, 3, 4, 200, 200, 10, 12, 0, 0, 14, 16, 0, 0]);

  Test('MatFromBlocks(3, IdentityMatrix(2), IdentityMatrix(2), IdentityMatrix(2), IdentityMatrix(2))',
    6, [1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0]);

  Test('MatFromBlocks(3, IdentityMatrix(2), IdentityMatrix(2), IdentityMatrix(2), IdentityMatrix(3))',
    failure, 'Invalid block at position (2, 1), index 4.');

  Test('MatFromBlocks(3, IdentityMatrix(2), IdentityMatrix(2), IdentityMatrix(3), IdentityMatrix(3))',
    failure, 'Invalid block at position (1, 3), index 3.');

  Test('MatFromBlocks(3, IdentityMatrix(3))',
    failure, 'Too few blocks to fill the first row.');

  Test('MatFromBlocks(3, IdentityMatrix(3), IdentityMatrix(3))',
    failure, 'Too few blocks to fill the first row.');

  Test('MatFromBlocks(3, IdentityMatrix(3), IdentityMatrix(3), IdentityMatrix(3))',
    9, [1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1]);

  Test('MatFromBlocks(3)', failure, 'The number of blocks has to be at least 1.');

  Test('MatFromBlocks(0, IdentityMatrix(3), IdentityMatrix(3))',
    failure, 'A positive integer was expected as argument 1, but "0" was given.');

  Test('MatFromBlocks()', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('MatFromBlocks(2, ❨❨1, 2❩, ❨3, 4❩❩, ❨❨10, 20❩, ❨30, 40❩❩, ❨❨12, 13❩, ❨14, 15❩❩, ❨❨100, 200❩, ❨300, 400⋅i❩❩)',
    4, [1, 2, 10, 20, 3, 4, 30, 40, 12, 13, 100, 200, 14, 15, 300, 400*ImaginaryUnit]);

  Test('MatFromBlocks(2, ❨❨1, 2❩, ❨3, 4❩❩, ❨❨10, 20, 30❩, ❨40, 50, 60❩❩, ❨❨100, 200❩, ❨300, 400❩❩, ❨❨i, 2, 3❩, ❨4, 5, 6❩❩)',
    5, [1, 2, 10, 20, 30, 3, 4, 40, 50, 60, 100, 200, ImaginaryUnit, 2, 3, 300, 400, 4, 5, 6]);

  Test('MatFromBlocks(2, ❨❨i, 2❩, ❨3, 4❩❩, ❨❨10, 20, 30❩, ❨40, 50, 60❩❩, ❨❨1, 2, 3❩, ❨4, 5, 6❩❩,  ❨❨100, 200❩, ❨300, 400❩❩)',
    failure, 'Invalid block at position (2, 1), index 3.');

  Test('MatFromBlocks(2, ❨❨i, 2❩, ❨3, 4❩❩, ❨❨10, 20, 30❩, ❨40, 50, 60❩❩, ❨❨100, 200❩, ❨300, 400❩❩, ❨❨1, 2, 3❩❩)',
    failure, 'Invalid block at position (2, 2), index 4.');

  Test('MatFromBlocks(2, ❨❨i, 2❩, ❨3, 4❩❩, ❨❨10, 20, 30❩, ❨40, 50, 60❩, ❨70, 80, 90❩❩, ❨❨100, 200❩, ❨300, 400❩❩, ❨❨1, 2, 3❩, ❨4, 5, 6❩❩)',
    failure, 'Invalid block at position (1, 2), index 2.');

  Test('MatFromBlocks(3, ❨❨i, 2❩, ❨3, 4❩❩, IdentityMatrix(2), IdentityMatrix(2), ZeroMatrix(2), ❨❨10, 20❩, ❨30, 40❩❩, IdentityMatrix(2))',
    6, [ImaginaryUnit, 2, 1, 0, 1, 0, 3, 4, 0, 1, 0, 1, 0, 0, 10, 20, 1, 0, 0, 0, 30, 40, 0, 1]);

  Test('I ≔ IdentityMatrix(2); MatFromBlocks(4, I, I, I, I, I, ComplexMatrix(I), I, I)',
    8, [TASC(1), 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1]);

  Test('delete(I)', success);

  Test('MatFromBlocks(3, ❨❨i❩, ❨2❩❩, ❨❨10❩, ❨20❩❩, ❨❨100❩, ❨200❩❩)',
    3, [ImaginaryUnit, 10, 100, 2, 20, 200]);

  Test('MatFromBlocks(1, ❨❨i, 2❩, ❨3, 4❩❩, ❨❨10, 100❩, ❨100, 10❩❩, ❨❨5, 6❩, ❨7, 8❩❩)',
    2, [ImaginaryUnit, 2, 3, 4, 10, 100, 100, 10, 5, 6, 7, 8]);

  Test('MatFromBlocks(2, ❨❨i, 2❩, ❨3, 4❩❩, ❨❨100, 100❩, ❨200, 200❩❩, ❨❨10, 12❩, ❨14, 16❩❩)',
    4, [ImaginaryUnit, 2, 100, 100, 3, 4, 200, 200, 10, 12, 0, 0, 14, 16, 0, 0]);

  Test('MatFromBlocks(3, IdentityMatrix(2), IdentityMatrix(2), IdentityMatrix(2), ComplexMatrix(IdentityMatrix(2)))',
    6, [1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, TASC(0)]);

  Test('MatFromBlocks(3, IdentityMatrix(2), IdentityMatrix(2), IdentityMatrix(2), ComplexMatrix(IdentityMatrix(3)))',
    failure, 'Invalid block at position (2, 1), index 4.');

  Test('MatFromBlocks(3, IdentityMatrix(2), IdentityMatrix(2), IdentityMatrix(3), ComplexMatrix(IdentityMatrix(3)))',
    failure, 'Invalid block at position (1, 3), index 3.');

  Test('MatFromBlocks(3, ComplexMatrix(IdentityMatrix(3)))',
    failure, 'Too few blocks to fill the first row.');

  Test('MatFromBlocks(3, IdentityMatrix(3), ComplexMatrix(IdentityMatrix(3)))',
    failure, 'Too few blocks to fill the first row.');

  Test('MatFromBlocks(3, IdentityMatrix(3), IdentityMatrix(3), ComplexMatrix(IdentityMatrix(3)))',
    9, [1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, TASC(1)]);

  Test('MatFromBlocks(0, IdentityMatrix(3), ComplexMatrix(IdentityMatrix(3)))',
    failure, 'A positive integer was expected as argument 1, but "0" was given.');

  Test('diag()', failure, 'A matrix must have size at least 1×1.');

  Test('diag(1)', 1, [1]);

  Test('diag(1, 2)', 2, [1, 0, 0, 2]);

  Test('diag(1, 2, 3)', 3, [1, 0, 0, 0, 2, 0, 0, 0, 3]);
  Test('diag(i, 2, 3)', 3, [ImaginaryUnit, 0, 0, 0, 2, 0, 0, 0, 3]);

  Test('diag(1, 2, 3, 4 + i)', 4, [1, 0, 0, 0, 0, 2, 0, 0, 0, 0, 3, 0, 0, 0, 0, 4 + ImaginaryUnit]);

  Test('diag(1, 1, 1, 1, 1, 1, 1) = IdentityMatrix(7)', True);
  Test('diag(0, 0, 0, 0, 0, 0, 0, 0, 0) = ZeroMatrix(9)', True);

  // Matrix parts

  Test('A ≔ ❨❨1, −4, 2❩, ❨0, 1, 5❩, ❨0, −2, 7❩❩; B ≔ ❨❨4, 6, 2, 0❩, ❨−2, 5, 8, 0❩, ❨2, 5, 0, 1❩❩; C ≔ ❨❨1, 0❩, ❨0, 4❩, ❨2, 0❩, ❨1, 1❩❩; Z ≔ ❨❨i, 2, 3❩, ❨0, 0, 1❩❩; W ≔ ❨❨i, 1❩, ❨2, −i❩❩; Q ≔ ❨❨i, 2, −i❩, ❨0, 3, −1❩, ❨2⋅i, 1, 0❩❩;', null);

  Test('cols(A)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ASR3(1, 0, 0)),
        ASO(ASR3(-4, 1, -2)),
        ASO(ASR3(2, 5, 7))
      ]
    )
  );

  Test('rows(A)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ASR3(1, -4, 2)),
        ASO(ASR3(0, 1, 5)),
        ASO(ASR3(0, -2, 7))
      ]
    )
  );

  Test('cols(B)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ASR3(4, -2, 2)),
        ASO(ASR3(6, 5, 5)),
        ASO(ASR3(2, 8, 0)),
        ASO(ASR3(0, 0, 1))
      ]
    )
  );

  Test('rows(B)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ASR4(4, 6, 2, 0)),
        ASO(ASR4(-2, 5, 8, 0)),
        ASO(ASR4(2, 5, 0, 1))
      ]
    )
  );

  Test('cols(C)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ASR4(1, 0, 2, 1)),
        ASO(ASR4(0, 4, 0, 1))
      ]
    )
  );

  Test('rows(C)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ASR2(1, 0)),
        ASO(ASR2(0, 4)),
        ASO(ASR2(2, 0)),
        ASO(ASR2(1, 1))
      ]
    )
  );

  Test('cols(Z)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ASC2(ImaginaryUnit, 0)),
        ASO(ASC2(2, 0)),
        ASO(ASC2(3, 1))
      ]
    )
  );

  Test('rows(Z)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ASC3(ImaginaryUnit, 2, 3)),
        ASO(ASC3(0, 0, 1))
      ]
    )
  );

  // row, col, MainDiagonal, antidiagonal, superdiagonal, and subdiagonal
  // can be part of lvalues.

  // - lvalue, read

  Test('col(A, 1)', [1, 0, 0]);
  Test('col(A, 2)', [-4, 1, -2]);
  Test('col(A, 3)', [2, 5, 7]);

  Test('col(A, 4)', failure, 'Index 4 out of bounds.');
  Test('col(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('col(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('row(A, 1)', [1, -4, 2]);
  Test('row(A, 2)', [0, 1, 5]);
  Test('row(A, 3)', [0, -2, 7]);

  Test('row(A, 4)', failure, 'Index 4 out of bounds.');
  Test('row(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('row(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('col(B, 1)', [4, -2, 2]);
  Test('col(B, 2)', [6, 5, 5]);
  Test('col(B, 3)', [2, 8, 0]);
  Test('col(B, 4)', [0, 0, 1]);
  Test('col(B, 5)', failure, 'Index 5 out of bounds.');

  Test('row(B, 1)', [4, 6, 2, 0]);
  Test('row(B, 2)', [-2, 5, 8, 0]);
  Test('row(B, 3)', [2, 5, 0, 1]);
  Test('row(B, 4)', failure, 'Index 4 out of bounds.');

  Test('col(C, 1)', [1, 0, 2, 1]);
  Test('col(C, 2)', [0, 4, 0, 1]);
  Test('col(C, 3)', failure, 'Index 3 out of bounds.');

  Test('row(C, 1)', [1, 0]);
  Test('row(C, 2)', [0, 4]);
  Test('row(C, 3)', [2, 0]);
  Test('row(C, 4)', [1, 1]);
  Test('row(C, 5)', failure, 'Index 5 out of bounds.');

  Test('col(Z, 1)', [ImaginaryUnit, 0]);
  Test('col(Z, 2)', [TASC(2), 0]);
  Test('col(Z, 3)', [TASC(3), 1]);
  Test('col(Z, 4)', failure, 'Index 4 out of bounds.');

  Test('row(Z, 1)', [ImaginaryUnit, 2, 3]);
  Test('row(Z, 2)', [TASC(0), 0, 1]);
  Test('row(Z, 3)', failure, 'Index 3 out of bounds.');

  Test('col(A, 2, 3)', failure, '2 argument(s) expected, but 3 given.');
  Test('col(A)', failure, '2 argument(s) expected, but 1 given.');
  Test('row(A, 2, 3)', failure, '2 argument(s) expected, but 3 given.');
  Test('row(A)', failure, '2 argument(s) expected, but 1 given.');

  Test('col(Z, 2, 3)', failure, '2 argument(s) expected, but 3 given.');
  Test('col(Z)', failure, '2 argument(s) expected, but 1 given.');
  Test('row(Z, 2, 3)', failure, '2 argument(s) expected, but 3 given.');
  Test('row(Z)', failure, '2 argument(s) expected, but 1 given.');

  Test('MainDiagonal(A)', [1, 1, 7]);
  Test('MainDiagonal(B)', [4, 5, 0]);
  Test('MainDiagonal(C)', [1, 4]);
  Test('MainDiagonal(Z)', [ImaginaryUnit, 0]);
  Test('MainDiagonal(Q)', [ImaginaryUnit, 3, 0]);

  Test('antidiagonal(A)', [2, 1, 0]);
  Test('antidiagonal(B)', failure, 'A non-square matrix has no antidiagonal.');
  Test('antidiagonal(C)', failure, 'A non-square matrix has no antidiagonal.');
  Test('antidiagonal(Z)', failure, 'A non-square matrix has no antidiagonal.');
  Test('antidiagonal(W)', [TASC(1), 2]);
  Test('antidiagonal(Q)', [-ImaginaryUnit, 3, 2*ImaginaryUnit]);

  Test('superdiagonal(A)', [-4, 5]);
  Test('superdiagonal(B)', [6, 8, 1]);
  Test('superdiagonal(C)', [0]);
  Test('superdiagonal(Z)', [TASC(2), 1]);
  Test('superdiagonal(W)', [TASC(1)]);
  Test('superdiagonal(Q)', [TASC(2), -1]);

  Test('subdiagonal(A)', [0, -2]);
  Test('subdiagonal(B)', [-2, 5]);
  Test('subdiagonal(C)', [0, 0]);
  Test('subdiagonal(Z)', [TASC(0)]);
  Test('subdiagonal(W)', [TASC(2)]);
  Test('subdiagonal(Q)', [TASC(0), 1]);

  // - rvalue, read

  Test('row(❨❨1, 2, 3❩, ❨10, 20, 30❩❩, 1)', [1, 2, 3]);
  Test('row(❨❨1, 2, 3❩, ❨10, 20, 30❩❩, 2)', [10, 20, 30]);
  Test('row(❨❨1, 2, 3❩, ❨10, 20, 30❩❩, 3)', failure, 'Index 3 out of bounds.');

  Test('col(❨❨1, 2, 3❩, ❨10, 20, 30❩❩, 1)', [1, 10]);
  Test('col(❨❨1, 2, 3❩, ❨10, 20, 30❩❩, 2)', [2, 20]);
  Test('col(❨❨1, 2, 3❩, ❨10, 20, 30❩❩, 3)', [3, 30]);
  Test('col(❨❨1, 2, 3❩, ❨10, 20, 30❩❩, 4)', failure, 'Index 4 out of bounds.');

  Test('MainDiagonal(❨❨1, 2, 3❩, ❨10, 20, 30❩❩)', [1, 20]);
  Test('antidiagonal(❨❨1, 2, 3❩, ❨10, 20, 30❩❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('superdiagonal(❨❨1, 2, 3❩, ❨10, 20, 30❩❩)', [2, 30]);
  Test('subdiagonal(❨❨1, 2, 3❩, ❨10, 20, 30❩❩)', [10]);

  Test('MainDiagonal(❨❨1, 2, 3❩, ❨10, 20, 30❩, ❨100, 200, 300❩❩)', [1, 20, 300]);
  Test('antidiagonal(❨❨1, 2, 3❩, ❨10, 20, 30❩, ❨100, 200, 300❩❩)', [3, 20, 100]);
  Test('superdiagonal(❨❨1, 2, 3❩, ❨10, 20, 30❩, ❨100, 200, 300❩❩)', [2, 30]);
  Test('subdiagonal(❨❨1, 2, 3❩, ❨10, 20, 30❩, ❨100, 200, 300❩❩)', [10, 200]);

  Test('row(❨❨i, 2, 3❩, ❨10, 20, 30❩❩, 1)', [ImaginaryUnit, 2, 3]);
  Test('row(❨❨i, 2, 3❩, ❨10, 20, 30❩❩, 2)', [TASC(10), 20, 30]);
  Test('row(❨❨i, 2, 3❩, ❨10, 20, 30❩❩, 3)', failure, 'Index 3 out of bounds.');

  Test('col(❨❨i, 2, 3❩, ❨10, 20, 30❩❩, 1)', [ImaginaryUnit, 10]);
  Test('col(❨❨i, 2, 3❩, ❨10, 20, 30❩❩, 2)', [TASC(2), 20]);
  Test('col(❨❨i, 2, 3❩, ❨10, 20, 30❩❩, 3)', [TASC(3), 30]);
  Test('col(❨❨i, 2, 3❩, ❨10, 20, 30❩❩, 4)', failure, 'Index 4 out of bounds.');

  Test('MainDiagonal(❨❨i, 2, 3❩, ❨10, 20, 30❩❩)', [ImaginaryUnit, 20]);
  Test('antidiagonal(❨❨i, 2, 3❩, ❨10, 20, 30❩❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('superdiagonal(❨❨i, 2, 3❩, ❨10, 20, 30❩❩)', [TASC(2), 30]);
  Test('subdiagonal(❨❨i, 2, 3❩, ❨10, 20, 30❩❩)', [TASC(10)]);

  Test('MainDiagonal(❨❨i, 2, 3❩, ❨10, 20, i/3❩, ❨100, 200, 300❩❩)', [ImaginaryUnit, 20, 300]);
  Test('antidiagonal(❨❨i, 2, 3❩, ❨10, 20, i/3❩, ❨100, 200, 300❩❩)', [TASC(3), 20, 100]);
  Test('superdiagonal(❨❨i, 2, 3❩, ❨10, 20, i/3❩, ❨100, 200, 300❩❩)', [2, ImaginaryUnit/3]);
  Test('subdiagonal(❨❨i, 2, 3❩, ❨10, 20, i/3❩, ❨100, 200, 300❩❩)', [TASC(10), 200]);

  // - lvalue, assigned to

  Test('A', 3, [1, -4, 2, 0, 1, 5, 0, -2, 7]);

  Test('col(A, 1) ≔ ❨7, 8, 9❩', [7, 8, 9]);

  Test('A', 3, [7, -4, 2, 8, 1, 5, 9, -2, 7]);

  Test('col(A, 2) ≔ ❨1, 2, 3❩', [1, 2, 3]);

  Test('A', 3, [7, 1, 2, 8, 2, 5, 9, 3, 7]);

  Test('col(A, 3) ≔ ❨10, 20, 30❩', [10, 20, 30]);

  Test('A', 3, [7, 1, 10, 8, 2, 20, 9, 3, 30]);

  Test('col(A, 4) ≔ ❨5, 6, 7❩', failure, 'Index 4 out of bounds.');
  Test('col(A, 0) ≔ ❨5, 6, 7❩', failure, 'Index 0 out of bounds.');
  Test('col(A, -1) ≔ ❨5, 6, 7❩', failure, 'Index -1 out of bounds.');

  Test('row(A, 1) ≔ ❨1, 2, 3❩', [1, 2, 3]);

  Test('A', 3, [1, 2, 3, 8, 2, 20, 9, 3, 30]);

  Test('row(A, 2) ≔ ❨10, 20, 30❩', [10, 20, 30]);

  Test('A', 3, [1, 2, 3, 10, 20, 30, 9, 3, 30]);

  Test('row(A, 3) ≔ ❨0, 0, 1❩', [0, 0, 1]);

  Test('A', 3, [1, 2, 3, 10, 20, 30, 0, 0, 1]);

  Test('row(A, 4) ≔ ❨5, 6, 7❩', failure, 'Index 4 out of bounds.');
  Test('row(A, 0) ≔ ❨5, 6, 7❩', failure, 'Index 0 out of bounds.');
  Test('row(A, -1) ≔ ❨5, 6, 7❩', failure, 'Index -1 out of bounds.');

  Test('col(A, 1) ≔ ❨1, 2, 3, 4❩', failure, 'Incorrect length of array.');
  Test('row(A, 3) ≔ ❨1, 2, 3, 4❩', failure, 'Incorrect length of array.');

  Test('col(A, 1, 2) ≔ ❨5, 66, 777❩', failure, 'Left side cannot be assigned to.');
  Test('row(A) ≔ ❨5, 66, 777❩', failure, 'Left side cannot be assigned to.');

  Test('A', 3, [1, 2, 3, 10, 20, 30, 0, 0, 1]);

  Test('col(A, 1+1) ≔ ❨9, 9, 9❩', [9, 9, 9]);

  Test('A', 3, [1, 9, 3, 10, 9, 30, 0, 9, 1]);

  Test('B', 4, [4, 6, 2, 0, -2, 5, 8, 0, 2, 5, 0, 1]);

  Test('col(B, 1) ≔ ❨1, 2, 3❩', [1, 2, 3]);

  Test('B', 4, [1, 6, 2, 0, 2, 5, 8, 0, 3, 5, 0, 1]);

  Test('col(B, 2) ≔ ❨10, 20, 30❩', [10, 20, 30]);

  Test('B', 4, [1, 10, 2, 0, 2, 20, 8, 0, 3, 30, 0, 1]);

  Test('col(B, 3) ≔ ❨1, 0, 1❩', [1, 0, 1]);

  Test('B', 4, [1, 10, 1, 0, 2, 20, 0, 0, 3, 30, 1, 1]);

  Test('col(B, 4) ≔ ❨5, 6, 7❩', [5, 6, 7]);

  Test('B', 4, [1, 10, 1, 5, 2, 20, 0, 6, 3, 30, 1, 7]);

  Test('col(B, 5) ≔ ❨1, 1, 0❩', failure, 'Index 5 out of bounds.');
  Test('col(B, 0) ≔ ❨1, 1, 0❩', failure, 'Index 0 out of bounds.');
  Test('col(B, -1) ≔ ❨1, 1, 0❩', failure, 'Index -1 out of bounds.');

  Test('row(B, 1) ≔ ❨1, 2, 3, 4❩', [1, 2, 3, 4]);

  Test('B', 4, [1, 2, 3, 4, 2, 20, 0, 6, 3, 30, 1, 7]);

  Test('row(B, 2) ≔ ❨5, 6, 7, 8❩', [5, 6, 7, 8]);

  Test('B', 4, [1, 2, 3, 4, 5, 6, 7, 8, 3, 30, 1, 7]);

  Test('row(B, 3) ≔ ❨9, 10, 11, 12❩', [9, 10, 11, 12]);

  Test('B', 4, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]);

  Test('row(B, 4) ≔ ❨13, 14, 15, 16❩', failure, 'Index 4 out of bounds.');
  Test('row(B, 0) ≔ ❨13, 14, 15, 16❩', failure, 'Index 0 out of bounds.');
  Test('row(B, -1) ≔ ❨13, 14, 15, 16❩', failure, 'Index -1 out of bounds.');

  Test('row(B, 2) ≔ ❨5, 6, 7, 8, 9❩', failure, 'Incorrect length of array.');

  Test('Z', 3, [ImaginaryUnit, 2, 3, 0, 0, 1]);

  Test('col(Z, 1) ≔ ❨i, 2⋅i❩', [ImaginaryUnit, 2*ImaginaryUnit]);

  Test('Z', 3, [ImaginaryUnit, 2, 3, 2*ImaginaryUnit, 0, 1]);

  Test('col(Z, 2) ≔ ❨10, 20❩', [10, 20]);

  Test('Z', 3, [ImaginaryUnit, 10, 3, 2*ImaginaryUnit, 20, 1]);

  Test('col(Z, 3) ≔ ❨0, i❩', [0, ImaginaryUnit]);

  Test('Z', 3, [ImaginaryUnit, 10, 0, 2*ImaginaryUnit, 20, ImaginaryUnit]);

  Test('col(Z, 4) ≔ ❨i, i❩', failure, 'Index 4 out of bounds.');
  Test('col(Z, 0) ≔ ❨i, i❩', failure, 'Index 0 out of bounds.');
  Test('col(Z, -1) ≔ ❨i, i❩', failure, 'Index -1 out of bounds.');

  Test('Z', 3, [ImaginaryUnit, 10, 0, 2*ImaginaryUnit, 20, ImaginaryUnit]);

  Test('row(Z, 1) ≔ ❨i, 2⋅i, 3⋅i❩', [ImaginaryUnit, 2*ImaginaryUnit, 3*ImaginaryUnit]);

  Test('Z', 3, [ImaginaryUnit, 2*ImaginaryUnit, 3*ImaginaryUnit, 2*ImaginaryUnit, 20, ImaginaryUnit]);

  Test('row(Z, 2) ≔ ❨1, 2, 3❩', [1, 2, 3]);

  Test('Z', 3, [ImaginaryUnit, 2*ImaginaryUnit, 3*ImaginaryUnit, 1, 2, 3]);

  Test('row(Z, 3) ≔ ❨10, 20, 30❩', failure, 'Index 3 out of bounds.');
  Test('row(Z, 0) ≔ ❨10, 20, 30❩', failure, 'Index 0 out of bounds.');
  Test('row(Z, -1) ≔ ❨10, 20, 30❩', failure, 'Index -1 out of bounds.');

  Test('col(Z, 1) ≔ ❨i, 2, 3❩', failure, 'Incorrect length of array.');
  Test('row(Z, 2) ≔ ❨1, 2, 3, 4❩', failure, 'Incorrect length of array.');

  Test('A', 3, [1, 9, 3, 10, 9, 30, 0, 9, 1]);

  Test('MainDiagonal(A) ≔ ❨100, 200, 300❩', [100, 200, 300]);

  Test('A', 3, [100, 9, 3, 10, 200, 30, 0, 9, 300]);

  Test('MainDiagonal(A) ≔ ❨10, 11❩', failure, 'Incorrect number of elements in diagonal.');
  Test('MainDiagonal(A) ≔ ❨10, 11, 12, 13❩', failure, 'Incorrect number of elements in diagonal.');

  Test('A', 3, [100, 9, 3, 10, 200, 30, 0, 9, 300]);

  Test('superdiagonal(A) ≔ ❨101, 201❩', [101, 201]);

  Test('A', 3, [100, 101, 3, 10, 200, 201, 0, 9, 300]);

  Test('superdiagonal(A) ≔ ❨1❩', failure, 'Incorrect number of elements in superdiagonal.');
  Test('superdiagonal(A) ≔ ❨1, 2, 3❩', failure, 'Incorrect number of elements in superdiagonal.');

  Test('A', 3, [100, 101, 3, 10, 200, 201, 0, 9, 300]);

  Test('subdiagonal(A) ≔ ❨199, 299❩', [199, 299]);

  Test('A', 3, [100, 101, 3, 199, 200, 201, 0, 299, 300]);

  Test('subdiagonal(A) ≔ ❨7❩', failure, 'Incorrect number of elements in subdiagonal.');
  Test('subdiagonal(A) ≔ ❨7, 8, 7❩', failure, 'Incorrect number of elements in subdiagonal.');

  Test('A', 3, [100, 101, 3, 199, 200, 201, 0, 299, 300]);

  Test('antidiagonal(A) ≔ ❨102, 200, 298❩', [102, 200, 298]);

  Test('A', 3, [100, 101, 102, 199, 200, 201, 298, 299, 300]);

  Test('antidiagonal(A) ≔ ❨55, 66❩', failure, 'Incorrect number of elements in antidiagonal.');
  Test('antidiagonal(A) ≔ ❨55, 66, 77, 88❩', failure, 'Incorrect number of elements in antidiagonal.');

  Test('A', 3, [100, 101, 102, 199, 200, 201, 298, 299, 300]);

  Test('B', 4, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]);

  Test('superdiagonal(B) ≔ ❨20, 70, 120❩', [20, 70, 120]);
  Test('B', 4, [1, 20, 3, 4, 5, 6, 70, 8, 9, 10, 11, 120]);
  Test('subdiagonal(B) ≔ ❨50, 100❩', [50, 100]);
  Test('B', 4, [1, 20, 3, 4, 50, 6, 70, 8, 9, 100, 11, 120]);

  Test('B ≔ ❨❨1, 2, 3, 4❩, ❨5, 6, 7, 8❩, ❨9, 10, 11, 12❩❩;', null);

  Test('superdiagonal(B) ≔ ❨20, 70, 120, 150❩', failure, 'Incorrect number of elements in superdiagonal.');
  Test('superdiagonal(B) ≔ ❨20, 70❩', failure, 'Incorrect number of elements in superdiagonal.');
  Test('subdiagonal(B) ≔ ❨50, 100, 150❩', failure, 'Incorrect number of elements in subdiagonal.');
  Test('subdiagonal(B) ≔ ❨50❩', failure, 'Incorrect number of elements in subdiagonal.');

  Test('B', 4, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]);

  Test('C', 2, [1, 0, 0, 4, 2, 0, 1, 1]);

  Test('superdiagonal(C) ≔ ❨21❩', [21]);
  Test('C', 2, [1, 21, 0, 4, 2, 0, 1, 1]);
  Test('subdiagonal(C) ≔ ❨7, 8❩', [7, 8]);
  Test('C', 2, [1, 21, 7, 4, 2, 8, 1, 1]);

  Test('C ≔ ❨❨1, 0❩, ❨0, 4❩, ❨2, 0❩, ❨1, 1❩❩;', null);

  Test('superdiagonal(C) ≔ ❨21, 23❩', failure, 'Incorrect number of elements in superdiagonal.');
  Test('subdiagonal(C) ≔ ❨7, 8, 9❩', failure, 'Incorrect number of elements in subdiagonal.');
  Test('subdiagonal(C) ≔ ❨7❩', failure, 'Incorrect number of elements in subdiagonal.');

  Test('C', 2, [1, 0, 0, 4, 2, 0, 1, 1]);

  Test('antidiagonal(B) ≔ ❨1, 2, 3❩', failure, 'A non-square matrix has no antidiagonal.');

  Test('Q', 3, [ImaginaryUnit, 2, -ImaginaryUnit, 0, 3, -1, 2*ImaginaryUnit, 1, 0]);

  Test('MainDiagonal(Q) ≔ ❨10, 20, 30⋅i❩', [10, 20, 30*ImaginaryUnit]);

  Test('Q', 3, [10, 2, -ImaginaryUnit, 0, 20, -1, 2*ImaginaryUnit, 1, 30*ImaginaryUnit]);

  Test('MainDiagonal(Q) ≔ ❨1, 2❩', failure, 'Incorrect number of elements in diagonal.');
  Test('MainDiagonal(Q) ≔ ❨1, 2, 3, 4❩', failure, 'Incorrect number of elements in diagonal.');

  Test('Q', 3, [10, 2, -ImaginaryUnit, 0, 20, -1, 2*ImaginaryUnit, 1, 30*ImaginaryUnit]);

  Test('superdiagonal(Q) ≔ ❨1, 2❩', [1, 2]);

  Test('Q', 3, [10, 1, -ImaginaryUnit, 0, 20, 2, 2*ImaginaryUnit, 1, 30*ImaginaryUnit]);

  Test('superdiagonal(Q) ≔ ❨6❩', failure, 'Incorrect number of elements in superdiagonal.');
  Test('superdiagonal(Q) ≔ ❨6, 7, 8❩', failure, 'Incorrect number of elements in superdiagonal.');

  Test('Q', 3, [10, 1, -ImaginaryUnit, 0, 20, 2, 2*ImaginaryUnit, 1, 30*ImaginaryUnit]);

  Test('subdiagonal(Q) ≔ ❨i, 2⋅i❩', [ImaginaryUnit, 2*ImaginaryUnit]);

  Test('Q', 3, [10, 1, -ImaginaryUnit, ImaginaryUnit, 20, 2, 2*ImaginaryUnit, 2*ImaginaryUnit, 30*ImaginaryUnit]);

  Test('subdiagonal(Q) ≔ ❨21❩', failure, 'Incorrect number of elements in subdiagonal.');
  Test('subdiagonal(Q) ≔ ❨21, 22, 23❩', failure, 'Incorrect number of elements in subdiagonal.');

  Test('Q', 3, [10, 1, -ImaginaryUnit, ImaginaryUnit, 20, 2, 2*ImaginaryUnit, 2*ImaginaryUnit, 30*ImaginaryUnit]);

  Test('antidiagonal(Q) ≔ ❨10, 20, 30❩', [10, 20, 30]);

  Test('Q', 3, [10, 1, 10, ImaginaryUnit, 20, 2, 30, 2*ImaginaryUnit, 30*ImaginaryUnit]);

  Test('antidiagonal(Q) ≔ ❨−1, −2❩', failure, 'Incorrect number of elements in antidiagonal.');
  Test('antidiagonal(Q) ≔ ❨−1, −2, 1, 2❩', failure, 'Incorrect number of elements in antidiagonal.');

  Test('Q', 3, [10, 1, 10, ImaginaryUnit, 20, 2, 30, 2*ImaginaryUnit, 30*ImaginaryUnit]);

  Test('antidiagonal(Z) ≔ ❨500, 600❩', failure, 'A non-square matrix has no antidiagonal.');

  Test('MainDiagonal(A, 2) ≔ ❨5❩', failure, 'Left side cannot be assigned to.');
  Test('superdiagonal(A, 2) ≔ ❨5❩', failure, 'Left side cannot be assigned to.');
  Test('subdiagonal(A, 2) ≔ ❨5❩', failure, 'Left side cannot be assigned to.');
  Test('antidiagonal(A, 2) ≔ ❨5❩', failure, 'Left side cannot be assigned to.');

  Test('col(A, 1) ≔ "Test"', failure, 'Cannot convert string "Test" to a real number.');
  Test('row(A, 2) ≔ true', failure, 'An object of type "boolean" cannot be converted to a real vector.');

  Test('col(Q, 1) ≔ "Test"', failure, 'Cannot convert string "Test" to a complex number.');
  Test('row(Q, 2) ≔ true', failure, 'An object of type "boolean" cannot be converted to a complex vector.');

  Test('A', 3, [100, 101, 102, 199, 200, 201, 298, 299, 300]);
  Test('Q', 3, [10, 1, 10, ImaginaryUnit, 20, 2, 30, 2*ImaginaryUnit, 30*ImaginaryUnit]);

  Test('row(A, 1) ≔ ''(1, 2, 3)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(1.0),
        ASO(2.0),
        ASO(3.0)
      ]
    )
  );

  Test('A', 3, [1, 2, 3, 199, 200, 201, 298, 299, 300]);

  Test('col(A, 2) ≔ ''(1.2, 1.3, 1.4)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(1.2),
        ASO(1.3),
        ASO(1.4)
      ]
    )
  );

  Test('A', 3, [1, 1.2, 3, 199, 1.3, 201, 298, 1.4, 300]);

  Test('MainDiagonal(A) ≔ ''(10, 20, 30)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(10.0),
        ASO(20.0),
        ASO(30.0)
      ]
    )
  );

  Test('A', 3, [10, 1.2, 3, 199, 20, 201, 298, 1.4, 30]);

  Test('superdiagonal(A) ≔ ''(0.1, 0.2)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(0.1),
        ASO(0.2)
      ]
    )
  );

  Test('A', 3, [10, 0.1, 3, 199, 20, 0.2, 298, 1.4, 30]);

  Test('subdiagonal(A) ≔ ''(-1, -2)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(-1.0),
        ASO(-2.0)
      ]
    )
  );

  Test('A', 3, [10, 0.1, 3, -1, 20, 0.2, 298, -2, 30]);

  Test('antidiagonal(A) ≔ ''(7, 8, 9)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(7.0),
        ASO(8.0),
        ASO(9.0)
      ]
    )
  );

  Test('A', 3, [10, 0.1, 7, -1, 8, 0.2, 9, -2, 30]);

  Test('MainDiagonal(A) ≔ ''(7, 8, 9, 10)', failure, 'Incorrect number of elements in diagonal.');
  Test('antidiagonal(A) ≔ ''(7, 8, 9, 10)', failure, 'Incorrect number of elements in antidiagonal.');
  Test('superdiagonal(A) ≔ ''(7, 8, 9, 10)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('subdiagonal(A) ≔ ''(7, 8, 9, 10)', failure, 'Incorrect number of elements in subdiagonal.');

  Test('A', 3, [10, 0.1, 7, -1, 8, 0.2, 9, -2, 30]);

  Test('row(Q, 3) ≔ ''(i, 2, 3)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ImaginaryUnit),
        ASO(2),
        ASO(3)
      ]
    )
  );

  Test('Q', 3, [10, 1, 10, ImaginaryUnit, 20, 2, ImaginaryUnit, 2, 3]);

  Test('col(Q, 1) ≔ ''(5, 7, 9)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(5),
        ASO(7),
        ASO(9)
      ]
    )
  );

  Test('Q', 3, [TASC(5), 1, 10, 7, 20, 2, 9, 2, 3]);

  Test('MainDiagonal(Q) ≔ ''(i, 2⋅i, 3⋅i)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ImaginaryUnit),
        ASO(2*ImaginaryUnit),
        ASO(3*ImaginaryUnit)
      ]
    )
  );

  Test('Q', 3, [ImaginaryUnit, 1, 10, 7, 2*ImaginaryUnit, 2, 9, 2, 3*ImaginaryUnit]);

  Test('superdiagonal(Q) ≔ ''(0, 1)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(0),
        ASO(1)
      ]
    )
  );

  Test('Q', 3, [ImaginaryUnit, 0, 10, 7, 2*ImaginaryUnit, 1, 9, 2, 3*ImaginaryUnit]);

  Test('subdiagonal(Q) ≔ ''(i, -i)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(ImaginaryUnit),
        ASO(-ImaginaryUnit)
      ]
    )
  );

  Test('Q', 3, [ImaginaryUnit, 0, 10, ImaginaryUnit, 2*ImaginaryUnit, 1, 9, -ImaginaryUnit, 3*ImaginaryUnit]);

  Test('antidiagonal(Q) ≔ ''(-1, -2, -3)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(-1),
        ASO(-2),
        ASO(-3)
      ]
    )
  );

  Test('Q', 3, [ImaginaryUnit, 0, -1, ImaginaryUnit, -2, 1, -3, -ImaginaryUnit, 3*ImaginaryUnit]);

  Test('MainDiagonal(Q) ≔ ''(7, 8, 9, 10)', failure, 'Incorrect number of elements in diagonal.');
  Test('antidiagonal(Q) ≔ ''(7, 8, 9, 10)', failure, 'Incorrect number of elements in antidiagonal.');
  Test('superdiagonal(Q) ≔ ''(7, 8, 9, 10)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('subdiagonal(Q) ≔ ''(7, 8, 9, 10)', failure, 'Incorrect number of elements in subdiagonal.');

  Test('A', 3, [10, 0.1, 7, -1, 8, 0.2, 9, -2, 30]);

  Test('row(A, 1) ≔ ❨1, i, 3❩', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component i.');
  Test('col(A, 3) ≔ ❨2 + 7⋅i, -2, i❩', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component 2 + 7⋅i.');
  Test('MainDiagonal(A) ≔ ❨-4⋅i, 2, 1❩', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component −4⋅i.');
  Test('antidiagonal(A) ≔ ❨2⋅i, 2, 1❩', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component 2⋅i.');
  Test('superdiagonal(A) ≔ ❨7, i❩', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component i.');
  Test('subdiagonal(A) ≔ ❨7, 3 + i❩', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component 3 + i.');

  Test('row(A, 1) ≔ ''(1, i, 3)', failure, 'Complex number i cannot be converted to a real number.');
  Test('col(A, 3) ≔ ''(1, 5 + 2⋅i, 3)', failure, 'Complex number 5 + 2⋅i cannot be converted to a real number.');
  Test('MainDiagonal(A) ≔ ''(3 - 5⋅i, 3)', failure, 'Complex number 3 − 5⋅i cannot be converted to a real number.');
  Test('antidiagonal(A) ≔ ''(5, -2⋅i, 3)', failure, 'Complex number −2⋅i cannot be converted to a real number.');
  Test('superdiagonal(A) ≔ ''(5, 10⋅i)', failure, 'Complex number 10⋅i cannot be converted to a real number.');
  Test('superdiagonal(A) ≔ ''(5, -i)', failure, 'Complex number −i cannot be converted to a real number.');
  Test('subdiagonal(A) ≔ ''(5, -10 - 9⋅i)', failure, 'Complex number −10 − 9⋅i cannot be converted to a real number.');

  Test('A', 3, [10, 0.1, 7, -1, 8, 0.2, 9, -2, 30]);

  // - lvalue, assigned to, protected

  Test('A', 3, [10, 0.1, 7, -1, 8, 0.2, 9, -2, 30]);

  Test('protect(A)', success);
  Test('row(A, 1) ≔ ❨1, 2, 3❩', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [10, 0.1, 7, -1, 8, 0.2, 9, -2, 30]);
  Test('unprotect(A)', success);
  Test('row(A, 1) ≔ ❨1, 2, 3❩', [1, 2, 3]);
  Test('A', 3, [1, 2, 3, -1, 8, 0.2, 9, -2, 30]);

  Test('protect(A)', success);
  Test('col(A, 3) ≔ ❨30, 40, 50❩', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [1, 2, 3, -1, 8, 0.2, 9, -2, 30]);
  Test('unprotect(A)', success);
  Test('col(A, 3) ≔ ❨30, 40, 50❩', [30, 40, 50]);
  Test('A', 3, [1, 2, 30, -1, 8, 40, 9, -2, 50]);

  Test('protect(A)', success);
  Test('MainDiagonal(A) ≔ ❨−1, −2, −3❩', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [1, 2, 30, -1, 8, 40, 9, -2, 50]);
  Test('unprotect(A)', success);
  Test('MainDiagonal(A) ≔ ❨−1, −2, −3❩', [-1, -2, -3]);
  Test('A', 3, [-1, 2, 30, -1, -2, 40, 9, -2, -3]);

  Test('protect(A)', success);
  Test('superdiagonal(A) ≔ ❨10, 20❩', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [-1, 2, 30, -1, -2, 40, 9, -2, -3]);
  Test('unprotect(A)', success);
  Test('superdiagonal(A) ≔ ❨10, 20❩', [10, 20]);
  Test('A', 3, [-1, 10, 30, -1, -2, 20, 9, -2, -3]);

  Test('protect(A)', success);
  Test('subdiagonal(A) ≔ ❨100, 200❩', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [-1, 10, 30, -1, -2, 20, 9, -2, -3]);
  Test('unprotect(A)', success);
  Test('subdiagonal(A) ≔ ❨100, 200❩', [100, 200]);
  Test('A', 3, [-1, 10, 30, 100, -2, 20, 9, 200, -3]);

  Test('protect(A)', success);
  Test('antidiagonal(A) ≔ ❨0, 1, 0❩', failure, 'Cannot modify or delete protected variable "A".');
  Test('A', 3, [-1, 10, 30, 100, -2, 20, 9, 200, -3]);
  Test('unprotect(A)', success);
  Test('antidiagonal(A) ≔ ❨0, 1, 0❩', [0, 1, 0]);
  Test('A', 3, [-1, 10, 0, 100, 1, 20, 0, 200, -3]);

  Test('Q', 3, [ImaginaryUnit, 0, -1, ImaginaryUnit, -2, 1, -3, -ImaginaryUnit, 3*ImaginaryUnit]);
  Test('protect(Q)', success);
  Test('row(Q, 2) ≔ ❨1, 2, i❩', failure, 'Cannot modify or delete protected variable "Q".');
  Test('Q', 3, [ImaginaryUnit, 0, -1, ImaginaryUnit, -2, 1, -3, -ImaginaryUnit, 3*ImaginaryUnit]);
  Test('unprotect(Q)', success);
  Test('row(Q, 2) ≔ ❨1, 2, i❩', [1, 2, ImaginaryUnit]);
  Test('Q', 3, [ImaginaryUnit, 0, -1, 1, 2, ImaginaryUnit, -3, -ImaginaryUnit, 3*ImaginaryUnit]);

  // - lvalue, assigned to, complex converted to real

  Test('A ≔ matrix(3, 10, 0.1, 7, −1, 8, 0.2, 9, −2, 30)', 3, [10, 0.1, 7, -1, 8, 0.2, 9, -2, 30]);
  Test('type(A)', 'real matrix');
  Test('col(A, 1) ≔ ❨1, i − i + 3, 5❩', [1, TASC(3), 5]);
  Test('A', 3, [1, 0.1, 7, 3, 8, 0.2, 5, -2, 30]);
  Test('type(A)', 'real matrix');
  Test('col(A, 2) ≔ ''(i − i + 2, 4, 6)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(TASC(2)),
        ASO(4),
        ASO(6)
      ]
    )
  );
  Test('A', 3, [1, 2, 7, 3, 4, 0.2, 5, 6, 30]);
  Test('type(A)', 'real matrix');

  // - additional tests (missing from the above sections)

  Test('A ≔ ❨❨4, 5, 7, 8❩❩', 4, [4, 5, 7, 8]);

  Test('superdiagonal(A)', [5]);
  Test('subdiagonal(A)', ASO(TRealVector.Create([]))); // returns empty vector!

  Test('superdiagonal(A) ≔ ❨50❩', [50]);
  Test('A', 4, [4, 50, 7, 8]);

  Test('subdiagonal(A) ≔ ❨50❩', failure, 'Incorrect number of elements in subdiagonal.');
  Test('A', 4, [4, 50, 7, 8]);

  Test('subdiagonal(A) ≔ DebugObject("empty vector")', ASO(TRealVector.Create([])));
  Test('A', 4, [4, 50, 7, 8]);

  Test('B ≔ ❨❨4❩, ❨5❩, ❨6❩, ❨7❩❩', 1, [4, 5, 6, 7]);

  Test('superdiagonal(B)', ASO(TRealVector.Create([]))); // returns empty vector!
  Test('subdiagonal(B)', [5]);

  Test('subdiagonal(B) ≔ ❨50❩', [50]);
  Test('B', 1, [4, 50, 6, 7]);

  Test('superdiagonal(B) ≔ ❨100❩', failure, 'Incorrect number of elements in superdiagonal.');
  Test('B', 1, [4, 50, 6, 7]);

  Test('superdiagonal(B) ≔ DebugObject("empty vector")', ASO(TRealVector.Create([])));
  Test('B', 1, [4, 50, 6, 7]);

  Test('C ≔ ❨❨21❩❩', 1, [21]);
  Test('C', 1, [21]);
  Test('MainDiagonal(C)', [21]);
  Test('superdiagonal(C)', ASO(TRealVector.Create([])));
  Test('subdiagonal(C)', ASO(TRealVector.Create([])));

  Test('MainDiagonal(C) ≔ ❨23❩', [23]);
  Test('C', 1, [23]);

  Test('superdiagonal(C) ≔ ❨23❩', failure, 'Incorrect number of elements in superdiagonal.');
  Test('subdiagonal(C) ≔ ❨23❩', failure, 'Incorrect number of elements in subdiagonal.');
  Test('C', 1, [23]);

  Test('superdiagonal(C) ≔ DebugObject("empty vector")', ASO(TRealVector.Create([])));
  Test('subdiagonal(C) ≔ DebugObject("empty vector")', ASO(TRealVector.Create([])));
  Test('C', 1, [23]);

  Test('A ≔ ❨❨i, 5, 7, 8❩❩', 4, [ImaginaryUnit, 5, 7, 8]);

  Test('superdiagonal(A)', [TASC(5)]);
  Test('subdiagonal(A)', ASO(TComplexVector.Create([]))); // returns empty vector!

  Test('superdiagonal(A) ≔ ❨i/2❩', [ImaginaryUnit/2]);
  Test('A', 4, [ImaginaryUnit, ImaginaryUnit/2, 7, 8]);

  Test('subdiagonal(A) ≔ ❨i/4❩', failure, 'Incorrect number of elements in subdiagonal.');
  Test('A', 4, [ImaginaryUnit, ImaginaryUnit/2, 7, 8]);

  Test('subdiagonal(A) ≔ ComplexVector(DebugObject("empty vector"))', ASO(TComplexVector.Create([])));
  Test('A', 4, [ImaginaryUnit, ImaginaryUnit/2, 7, 8]);

  Test('B ≔ ❨❨i❩, ❨5❩, ❨6❩, ❨7❩❩', 1, [ImaginaryUnit, 5, 6, 7]);

  Test('superdiagonal(B)', ASO(TComplexVector.Create([]))); // returns empty vector!
  Test('subdiagonal(B)', [TASC(5)]);

  Test('subdiagonal(B) ≔ ❨i/2❩', [ImaginaryUnit/2]);
  Test('B', 1, [ImaginaryUnit, ImaginaryUnit/2, 6, 7]);

  Test('superdiagonal(B) ≔ ❨i/4❩', failure, 'Incorrect number of elements in superdiagonal.');
  Test('B', 1, [ImaginaryUnit, ImaginaryUnit/2, 6, 7]);

  Test('superdiagonal(B) ≔ ComplexVector(DebugObject("empty vector"))', ASO(TComplexVector.Create([])));
  Test('B', 1, [ImaginaryUnit, ImaginaryUnit/2, 6, 7]);

  Test('C ≔ ❨❨i❩❩', 1, [ImaginaryUnit]);
  Test('C', 1, [ImaginaryUnit]);
  Test('MainDiagonal(C)', [ImaginaryUnit]);
  Test('superdiagonal(C)', ASO(TComplexVector.Create([])));
  Test('subdiagonal(C)', ASO(TComplexVector.Create([])));

  Test('MainDiagonal(C) ≔ ❨i/2❩', [ImaginaryUnit/2]);
  Test('C', 1, [ImaginaryUnit/2]);

  Test('superdiagonal(C) ≔ ❨i/4❩', failure, 'Incorrect number of elements in superdiagonal.');
  Test('subdiagonal(C) ≔ ❨i/4❩', failure, 'Incorrect number of elements in subdiagonal.');
  Test('C', 1, [ImaginaryUnit/2]);

  Test('superdiagonal(C) ≔ ComplexVector(DebugObject("empty vector"))', ASO(TComplexVector.Create([])));
  Test('subdiagonal(C) ≔ ComplexVector(DebugObject("empty vector"))', ASO(TComplexVector.Create([])));
  Test('C', 1, [ImaginaryUnit/2]);

  Test('delete(A); delete(B); delete(C); delete(Z); delete(W); delete(Q)', success);





  //
  //
  //  CHAPTER 6
  //  Linear algebra.
  //
  //

  Chapter('Linear algebra');

  //
  //  Basic operations
  //

  // Operations on vectors

  // - inner (scalar, dot) product

  Test('❨2, 6❩ ⋅ ❨−2, 3❩', 14.0);
  Test('❨1, 4, 2❩ ⋅ ❨6, −1, 2❩', 6.0);
  Test('❨7, −1, 2❩ ⋅ ❨5, −2, 0❩', 37.0);
  Test('❨4, 1, 3.5❩ ⋅ ❨−2.4, 6, 4❩', 10.4);
  Test('❨1, 5, 2, 8, 6❩ ⋅ ❨0, 1, 0, 0, 0❩', 5.0);

  Test('❨1, 5, 2, 8, 6❩ ⋅ ❨0, 1, 0, 0❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');

  Test('InnerProduct(❨2, 5❩, ❨−4, 6❩)', 22.0);
  Eps; Test('InnerProduct(❨−1.2, 4.5, 6.2❩, ❨9.4, −0.2, 0.1❩)', -11.56);
  Test('InnerProduct(❨1, 4, 2❩, ❨1, 0❩)', failure, 'Cannot compute dot product of two vectors of unequal dimension.');

  Test('InnerProduct()', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('InnerProduct(❨4, 5❩)', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('InnerProduct(❨4, 5❩, ❨1, 5❩, ❨3, 6❩)', failure, 'Too many arguments.');
  Test('InnerProduct(❨2, 5❩, 7)', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');
  Test('InnerProduct("rat", ❨2, 5❩)', failure, 'An object of type vector was expected as argument 1, but an object of type string was given.');

  Test('❨i❩ ⋅ ❨1❩', ImaginaryUnit);
  Test('❨1❩ ⋅ ❨i❩', -ImaginaryUnit);

  Test('❨2, i, −1❩ ⋅ ❨1, 2, 3❩', -1 + 2*ImaginaryUnit);
  Test('❨2, 4, −2❩ ⋅ ❨3, i, −i❩', 6 - 6*ImaginaryUnit);
  Test('❨2 + i, 3, 5 − i❩ ⋅ ❨1, 2 + 3⋅i, −i❩', 9 - 3*ImaginaryUnit);
  Test('❨2, 3❩ ⋅ ❨2 + i, 5❩', 19 - 2*ImaginaryUnit);
  Test('❨2, 3❩ ⋅ ❨2 + i, 5, 3❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');

  Test('InnerProduct(❨2, 5, −1❩, ❨3, 2 + i, 7❩)', 9 - 5*ImaginaryUnit);
  Test('InnerProduct(❨2+i, 3❩, ❨2, −3❩)', -5 + 2*ImaginaryUnit);
  Test('InnerProduct(❨3+i, 2−5⋅i, 3❩, ❨1−i, 2−4⋅i, 5⋅i❩)', 26 - 13*ImaginaryUnit);
  Test('InnerProduct(❨3+i, 2−5⋅i, 3❩, ❨1−i, 2−4⋅i, 5⋅i, 2❩)', failure, 'Cannot compute dot product of two vectors of unequal dimension.');
  Test('InnerProduct(❨3+i, 2−5⋅i, 3, 7❩, ❨5⋅i, 2❩)', failure, 'Cannot compute dot product of two vectors of unequal dimension.');

  Test('InnerProduct()', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('InnerProduct(❨5⋅i, 2❩)', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('InnerProduct(❨5⋅i, 2❩, ❨i, -1❩, ❨10, 4❩)', failure, 'Too many arguments.');
  Test('InnerProduct(7, ❨2, −3❩)', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');
  Test('InnerProduct(❨2, −3❩, "horse")', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');

  // - vector (cross) product

  Test('❨1, 0, 0❩ × ❨0, 1, 0❩', [0, 0, 1]);
  Test('❨6, −2, 3❩ × ❨5, 1, 2❩', [-7, 3, 16]);
  Test('❨1, 4❩ × ❨6, 2❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨1, 4, 7❩ × ❨6, 2❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨1, 4❩ × ❨6, 2, 2❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨1, 7, 4❩ × ❨6, 2, 2, -2❩', failure, 'Vector cross product only defined in ℝ³.');

  Test('❨0, 1, 2❩ × ❨−1, 0, 1❩ × ❨1, 2, 3❩', [-8, -2, 4]);
  Test('❨0, 1, 2❩ × ❨−1, 0, 1❩ × ❨1, 2, 3❩ × ❨0, 3, 1❩', [-14, 8, -24]);

  Test('❨0, 1, 2❩ × 7', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');
  Test('❨0, 1, 2❩ × ❨−1, 0, 1❩ × "cat"', failure, 'An object of type vector was expected as argument 3, but an object of type string was given.');
  Test('❨0, 1, 2❩ × ❨−1, 0, 1❩ × ❨1, 2, 3❩ × true', failure, 'An object of type vector was expected as argument 4, but an object of type boolean was given.');
  Test('❨0, 1, 2❩ × ❨−1, 0, 1❩ × ❨1, 2, 3❩ × {1, 2, 3}', failure, 'An object of type vector was expected as argument 4, but an object of type set was given.');

  Test('5 × ❨0, 1, 2❩', failure, 'An object of type set was expected as argument 1, but an object of type integer was given.'); // non-elegant asymmetry

  Test('❨0, 1, 2❩ × ❨−1, 0, 1❩ × ❨1, 2, 3, 4❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨0, 1, 2❩ × ❨−1, 0, 1❩ × ❨1, 2❩ × ❨0, 3, 1❩', failure, 'Vector cross product only defined in ℝ³.');
  Test('❨0, 1, 2❩ × ❨−1, 0, 1❩ × ❨1, 2, 3❩ × ❨0, 3, 1, -1❩', failure, 'Vector cross product only defined in ℝ³.');

  Test('CrossProduct(❨5, −2, 3❩, ❨−1, 2, 10❩)', [-26, -53, 8]);
  Test('CrossProduct(❨2, 7, 1❩, ❨−3, 4, 5❩)', [31, -13, 29]);
  Test('CrossProduct(❨0, −1, 2❩, ❨1, 2, 4❩)', [-8, 2, 1]);
  Test('CrossProduct(❨0, −1, 2❩, ❨1, 2, 4, 6❩)', failure, 'Vector cross product only defined in ℝ³.');
  Test('CrossProduct(❨0, −1, 2❩, ❨1, 6❩)', failure, 'Vector cross product only defined in ℝ³.');
  Test('CrossProduct(❨0, 2❩, ❨1, 2, 6❩)', failure, 'Vector cross product only defined in ℝ³.');

  Test('CrossProduct(❨0, 1, 2❩, ❨−1, 0, 1❩, ❨1, 2, 3❩)', [-8, -2, 4]);
  Test('CrossProduct(❨0, 1, 2❩, ❨−1, 0, 1❩, ❨1, 2, 3❩, ❨0, 3, 1❩)', [-14, 8, -24]);

  Test('CrossProduct()', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('CrossProduct(❨0, 1, 2❩)', [0, 1, 2]);
  Test('CrossProduct(5, ❨−1, 0, 1❩, ❨1, 2, 3❩, ❨0, 3, 1❩)', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');
  Test('CrossProduct(❨0, 1, 2❩, "dog", ❨1, 2, 3❩, ❨0, 3, 1❩)', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');
  Test('CrossProduct(❨0, 1, 2❩, ❨−1, 0, 1❩, true, ❨0, 3, 1❩)', failure, 'An object of type vector was expected as argument 3, but an object of type boolean was given.');
  Test('CrossProduct(❨0, 1, 2❩, ❨−1, 0, 1❩, ❨1, 2, 3❩, {0, 3, 1})', failure, 'An object of type vector was expected as argument 4, but an object of type set was given.');
  Test('CrossProduct(❨0, 1, 2❩, ❨−1, 0, 1❩, ❨1, 2, 3❩, ❨0, 3, 1❩, IdentityMatrix(3))', failure, 'An object of type vector was expected as argument 5, but an object of type real matrix was given.');

  Test('CrossProduct(❨0, 1, 2❩, ❨−1, 0, 1❩, ❨1, 2❩)', failure, 'Vector cross product only defined in ℝ³.');
  Test('CrossProduct(❨0, 1, 2❩, ❨−1, 0, 1, 1❩, ❨1, 2, 3❩, ❨0, 3, 1❩)', failure, 'Vector cross product only defined in ℝ³.');
  Test('CrossProduct(❨0, 1, 2❩, ❨−1, 0, 1❩, ❨1, 2❩, ❨0, 3, 1❩)', failure, 'Vector cross product only defined in ℝ³.');
  Test('CrossProduct(❨0, 1, 2❩, ❨−1, 0, 1❩, ❨1, 2, 3❩, ❨0, 3, 1, 7, 5, 1❩)', failure, 'Vector cross product only defined in ℝ³.');

  Test('❨3, 2, 1❩ × ❨2, i, 5❩', [10 - ImaginaryUnit, -13, -4 + 3*ImaginaryUnit]);
  Test('❨2, i, −i❩ × ❨1, 4, −2❩', [2*ImaginaryUnit, 4 - ImaginaryUnit, 8 - ImaginaryUnit]);
  Test('❨2 + 3⋅i, 5, 0❩ × ❨0, 1 − 2⋅i, 3❩', [15, -6-9*ImaginaryUnit, 8-ImaginaryUnit]);
  Test('❨1, 0, i❩ × ❨2, −i, 0❩', [-1, 2*ImaginaryUnit, -ImaginaryUnit]);
  Test('❨1, 0, i❩ × ❨2, −i, 0❩ × ❨2, 0, i❩', [-2, -ImaginaryUnit, -4*ImaginaryUnit]);

  Test('❨3, 2❩ × ❨2, i❩', failure, 'Vector cross product only defined in ℂ³.');
  Test('❨3, 2, 5❩ × ❨2, i❩', failure, 'Vector cross product only defined in ℂ³.');
  Test('❨3, 2❩ × ❨2, i, 1❩', failure, 'Vector cross product only defined in ℂ³.');
  Test('❨3, i, 5❩ × ❨2, 1, 1, 2❩', failure, 'Vector cross product only defined in ℂ³.');
  Test('❨3, 3, 5❩ × ❨2, 1, i, 2❩', failure, 'Vector cross product only defined in ℂ³.');
  Test('❨3, 4, 5❩ × ❨2, 1, 2❩ × ❨2, i❩', failure, 'Vector cross product only defined in ℂ³.');
  Test('❨3, 4, 5❩ × ❨2, i, 2❩ × ❨2, 8❩', failure, 'Vector cross product only defined in ℂ³.');

  Test('❨3, 4, 5❩ × 4 × ❨2, i❩', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');
  Test('❨3, 4, 5❩ × ❨2, 1, 2❩ × "guinea pig"', failure, 'An object of type vector was expected as argument 3, but an object of type string was given.');
  Test('❨3, 4, 5❩ × ❨2, 1, 2❩ × {1, 2, 3}', failure, 'An object of type vector was expected as argument 3, but an object of type set was given.');

  Test('CrossProduct(❨2, 1, 0❩, ❨3, i, 2❩)', [2, -4, -3 + 2*ImaginaryUnit]);
  Test('CrossProduct(❨1, i, −i❩, ❨0, i, 2❩)', [-1 + 2*ImaginaryUnit, -2, ImaginaryUnit]);
  Test('CrossProduct(❨2 + i, 3, 1 − i❩, ❨2, 1, 1 + 2⋅i❩)', [2 + 7*ImaginaryUnit, 2 - 7*ImaginaryUnit, -4 + ImaginaryUnit]);
  Test('CrossProduct(❨2, 3, i❩)', [2, 3, ImaginaryUnit]);
  Test('CrossProduct(❨1, 0, 1❩, ❨2, i, 0❩)', [-ImaginaryUnit, 2, ImaginaryUnit]);
  Test('CrossProduct(❨1, 0, 1❩, ❨2, i, 0❩, ❨2, 0, 1❩)', [2, 3*ImaginaryUnit, -4]);

  Test('CrossProduct(❨1, 0, 1❩, ❨2, i❩)', failure, 'Vector cross product only defined in ℂ³.');
  Test('CrossProduct(❨1, 0, 1❩, ❨2, i, 0❩, ❨1, 2, 0, 1❩)', failure, 'Vector cross product only defined in ℂ³.');
  Test('CrossProduct(❨1, 0, 1❩, ❨2, 3, 0❩, ❨i, 2, 0, 1❩)', failure, 'Vector cross product only defined in ℂ³.');
  Test('CrossProduct(❨1, 0, i❩, ❨2, 0❩, ❨2, 0, 1❩)', failure, 'Vector cross product only defined in ℂ³.');

  Test('CrossProduct(5, ❨2, i, 0❩, ❨2, 0, 1❩)', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');
  Test('CrossProduct(❨1, 0, 1❩, "cat", ❨2, 0, 1❩)', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');
  Test('CrossProduct(❨1, 0, 1❩, ❨2, i, 0❩, ZeroMatrix(2))', failure, 'An object of type vector was expected as argument 3, but an object of type real matrix was given.');

  // - angle between vectors

  Test('∠(❨1, 0❩, ❨2, 0❩)', 0.0);
  Eps; Test('∠(❨1, 0❩, ❨0, 1❩)', Pi/2);
  Eps; Test('∠(❨1, 0❩, ❨0, -4❩)', Pi/2);
  Eps; Test('∠(❨1, 0❩, ❨1, 1❩)', Pi/4);
  Eps; Test('∠(❨2, 0❩, ❨3, 3❩)', Pi/4);
  Eps; Test('∠(❨−1, −1❩, ❨−1, 0❩)', Pi/4);
  Eps; Test('∠(❨1, 0❩, ❨−1, 1❩)', 3*Pi/4);
  Eps; Test('∠(❨1, 0❩, ❨0, -1E20❩)', Pi/2);
  Eps; Test('∠(❨5, 0❩, ❨−1, 0❩)', Pi);
  Eps; Test('∠(❨1, 1❩, ❨1, -1❩)', Pi/2);
  Eps; Test('∠(❨1, 0❩, ❨cos(1.23), sin(1.23)❩)', 1.23);

  Eps; Test('∠(❨1, 4, 0❩, ❨0, 0, 1❩)', Pi/2);
  Eps; Test('∠(❨1, 4, 0❩, ❨−1, 2, 1❩)', 0.805011035735014204);

  Test('∠(❨1, 4, 0❩, ❨−1, 2❩)', failure, 'Cannot compute angle between vectors of different dimension.');

  Test('∠(❨1, 1❩, ❨0, 0❩)', failure);

  Test('∠()', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('∠(❨1, 1❩)', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('∠(❨1, 1❩, ❨1, 0❩, ❨1, -1❩)', failure, 'Too many arguments.');
  Test('∠(7, ❨1, -1❩)', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');
  Test('∠(❨1, -1❩, "cat")', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');

  Eps; Test('∠(❨1, 2 + 3⋅i, −2❩, ❨1 − 2⋅i, i, −i❩)', 1.20643960142319382);
  Eps; Test('∠(❨1, i, 0❩, ❨0, 0, 2 + 7⋅i/3❩)', Pi/2);
  Eps; Test('∠(❨1, -5, 0❩, ❨0, 0, 2 + 7⋅i/3❩)', Pi/2);

  Test('angle(❨1, 0❩, ❨2, 0❩)', 0.0);
  Eps; Test('angle(❨1, 0❩, ❨0, 1❩)', Pi/2);
  Eps; Test('angle(❨1, 0❩, ❨0, -4❩)', Pi/2);
  Eps; Test('angle(❨1, 0❩, ❨1, 1❩)', Pi/4);
  Eps; Test('angle(❨2, 0❩, ❨3, 3❩)', Pi/4);
  Eps; Test('angle(❨−1, −1❩, ❨−1, 0❩)', Pi/4);
  Eps; Test('angle(❨1, 0❩, ❨−1, 1❩)', 3*Pi/4);
  Eps; Test('angle(❨1, 0❩, ❨0, -1E20❩)', Pi/2);
  Eps; Test('angle(❨5, 0❩, ❨−1, 0❩)', Pi);
  Eps; Test('angle(❨1, 1❩, ❨1, -1❩)', Pi/2);
  Eps; Test('angle(❨1, 0❩, ❨cos(1.23), sin(1.23)❩)', 1.23);

  Eps; Test('angle(❨1, 4, 0❩, ❨0, 0, 1❩)', Pi/2);
  Eps; Test('angle(❨1, 4, 0❩, ❨−1, 2, 1❩)', 0.805011035735014204);

  Test('angle(❨1, 4, 0❩, ❨−1, 2❩)', failure, 'Cannot compute angle between vectors of different dimension.');

  Test('angle(❨1, 1❩, ❨0, 0❩)', failure);

  Test('angle()', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('angle(❨1, 1❩)', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('angle(❨1, 1❩, ❨1, 0❩, ❨1, -1❩)', failure, 'Too many arguments.');
  Test('angle(7, ❨1, -1❩)', failure, 'An object of type vector was expected as argument 1, but an object of type integer was given.');
  Test('angle(❨1, -1❩, "cat")', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');

  Eps; Test('angle(❨1, 2 + 3⋅i, −2❩, ❨1 − 2⋅i, i, −i❩)', 1.20643960142319382);
  Eps; Test('angle(❨1, i, 0❩, ❨0, 0, 2 + 7⋅i/3❩)', Pi/2);
  Eps; Test('angle(❨1, -5, 0❩, ❨0, 0, 2 + 7⋅i/3❩)', Pi/2);

  // - parallel

  Test('❨1, 3, 2❩ ∥ ❨10, 30, 20❩', True);
  Test('❨1, 3, 2❩ ∥ ❨-10, -30, -20❩', True);
  Test('❨10, 30, 20❩ ∥ ❨-0.2, -0.6, -0.4❩', True);
  Test('❨1, 3, 2❩ ∥ ❨10, 30, 21❩', False);
  Test('❨1, 3, 2❩ ∥ ❨2, 0, -3❩', False);

  Test('❨1, 3, 2, 7❩ ∥ ❨2, 6, 4, 14❩', True);
  Test('❨1, 3, 2, 7❩ ∥ ❨-0.2, -0.6, -0.4, -1.4❩', True);
  Test('❨1, 3, 2, 7❩ ∥ ❨2, 6, 4, 15❩', False);

  Test('❨1, 3, 2❩ ∥ ❨0, 0, 0❩', True);
  Test('❨e, √π, 2❩ ∥ ❨0, 0, 0❩', True);
  Test('❨sin(1), cos(2), arcsinh(2)❩ ∥ ❨0, 0, 0❩', True);
  Test('❨0, 0, 0❩ ∥ ❨-3, 3, 6❩', True);
  Test('❨0, 0, 0❩ ∥ ❨0, 0, 0❩', True);
  Test('❨0, 0, 0, 0❩ ∥ ❨0, 0, 0, 0❩', True);

  Test('❨arcsin(0.2), 0, 0❩ ∥ ❨2, 0, 0❩', True);
  Test('❨sinh(e), 0, 0❩ ∥ ❨-e, 0, 0❩', True);
  Test('❨sinh(e), 0, 0❩ ∥ ❨-e, 0, 5❩', False);
  Test('❨sinh(e), 1, 0❩ ∥ ❨-e, 0, 0❩', False);

  Test('❨sin(1), sin(2), cos(1)❩ ∥ √7⋅❨sin(1), sin(2), cos(1)❩', True);
  Test('❨sin(1), sin(2), cos(1)❩ ∥ −π⋅❨sin(1), sin(2), cos(1)❩/√2', True);
  Test('❨sin(1), sin(2), cos(1)❩ ∥ √7⋅❨sin(1), sin(2), tan(1)❩', False);

  Test('❨1, 3, 2❩ ∥ ❨2, 0, -3, 5❩', failure, 'Vectors are of different dimension.');

  Test('AreParallel(❨1, 3, 2❩, ❨10, 30, 20❩)', True);
  Test('AreParallel(❨1, 3, 2❩, ❨-10, -30, -20❩)', True);
  Test('AreParallel(❨10, 30, 20❩, ❨-0.2, -0.6, -0.4❩)', True);
  Test('AreParallel(❨1, 3, 2❩, ❨10, 30, 21❩)', False);
  Test('AreParallel(❨1, 3, 2❩, ❨2, 0, -3❩)', False);

  Test('AreParallel(❨1, 3, 2, 7❩, ❨2, 6, 4, 14❩)', True);
  Test('AreParallel(❨1, 3, 2, 7❩, ❨-0.2, -0.6, -0.4, -1.4❩)', True);
  Test('AreParallel(❨1, 3, 2, 7❩, ❨2, 6, 4, 14.0001❩)', False);
  Test('AreParallel(❨1, 3, 2, 7❩, ❨2, 6, 4, 15❩)', False);

  Test('AreParallel(❨1, 3, 2❩, ❨0, 0, 0❩)', True);
  Test('AreParallel(❨e, √π, 2❩, ❨0, 0, 0❩)', True);
  Test('AreParallel(❨sin(1), cos(2), arcsinh(2)❩, ❨0, 0, 0❩)', True);
  Test('AreParallel(❨0, 0, 0❩, ❨-3, 3, 6❩)', True);
  Test('AreParallel(❨0, 0, 0❩, ❨0, 0, 0❩)', True);
  Test('AreParallel(❨0, 0, 0, 0❩, ❨0, 0, 0, 0❩)', True);

  Test('AreParallel(❨arcsin(0.2), 0, 0❩, ❨2, 0, 0❩)', True);
  Test('AreParallel(❨sinh(e), 0, 0❩, ❨-e, 0, 0❩)', True);
  Test('AreParallel(❨sinh(e), 0, 0❩, ❨-e, 0, 5❩)', False);
  Test('AreParallel(❨sinh(e), 1, 0❩, ❨-e, 0, 0❩)', False);

  Test('AreParallel(❨sin(1), sin(2), cos(1)❩, √7⋅❨sin(1), sin(2), cos(1)❩)', True);
  Test('AreParallel(❨sin(1), sin(2), cos(1)❩, −π⋅❨sin(1), sin(2), cos(1)❩/√2)', True);
  Test('AreParallel(❨sin(1), sin(2), cos(1)❩, √7⋅❨sin(1), sin(2), tan(1)❩)', False);

  Test('AreParallel(❨1, 3, 2❩, ❨2, 0, -3, 5❩)', failure, 'Vectors are of different dimension.');

  Test('AreParallel()', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('AreParallel(❨1, 3, 2❩)', failure, 'Too few arguments. A required argument of type vector is missing.');

  Test('AreParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩)', False);
  Test('AreParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩, 0.01)', True);
  Test('AreParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩, 0.001)', True);
  Test('AreParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩, 0.0001)', False);
  Test('AreParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩, 0.00001)', False);

  Test('❨1, 2, i❩ ∥ ❨2, 4, 2⋅i❩', True);
  Test('❨1, 2, i❩ ∥ ❨−0.1, −0.2, −i/10❩', True);
  Test('❨1, 2, i❩ ∥ ❨i, 2⋅i, −1❩', True);
  Test('❨1, 2, i❩ ∥ ❨2 + i, 4 + 2⋅i, −1 + 2⋅i❩', True);
  Test('❨1, 2, i❩ ∥ ❨2, 4, i❩', False);
  Test('❨1, 2, i❩ ∥ ❨−0.1, −0.3, −i/10❩', False);
  Test('❨1, 2, i❩ ∥ ❨i, 2⋅i, 1❩', False);
  Test('❨1, 2, i❩ ∥ ❨2 + i, 4 + 2⋅i, 1 + 2⋅i❩', False);

  Test('❨1, 2, i❩ ∥ √3⋅❨2, 4, 2⋅i❩/9999', True);
  Test('❨1, 2, i❩ ∥ -√π⋅❨2, 4, 2⋅i❩/e^5', True);
  Test('❨1, 2, i❩ ∥ -√π⋅❨2, 4, 2⋅i❩/(2 - i)', True);
  Test('❨1, 2, i❩ ∥ -√π⋅e^2⋅❨2, 4, 2⋅i❩/sin(2 - i)', True);
  Test('❨1, 2, i❩ ∥ -√π⋅e^2⋅❨2, 4, 2.01⋅i❩/sin(2 - i)', False);

  Test('❨0, 1, 2❩ ∥ ❨0, 7 + 3⋅i/2, 14 + 3⋅i❩', True);
  Test('❨0, 1, 2❩ ∥ ❨0, 7 + 3⋅i/2, 15 + 3⋅i❩', False);
  Test('❨0, 1, 2❩ ∥ ❨1, 7 + 3⋅i/2, 14 + 3⋅i❩', False);

  Test('❨0, 0, 0❩ ∥ ❨1, 7 + 3⋅i/2, 14 + 3⋅i❩', True);
  Test('❨1, 2, i❩ ∥ ❨0, 0, 0❩', True);

  Test('❨1, 2, i❩ ∥ 5', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');
  Test('❨1, 2, i❩ ∥ true', failure, 'An object of type vector was expected as argument 2, but an object of type boolean was given.');
  Test('IdentityMatrix(4) ∥ 5', failure, 'An object of type vector was expected as argument 1, but an object of type real matrix was given.');

  Test('AreParallel(❨1, 2, i❩, ❨2, 4, 2⋅i❩)', True);
  Test('AreParallel(❨1, 2, i❩, ❨−0.1, −0.2, −i/10❩)', True);
  Test('AreParallel(❨1, 2, i❩, ❨i, 2⋅i, −1❩)', True);
  Test('AreParallel(❨1, 2, i❩, ❨2 + i, 4 + 2⋅i, −1 + 2⋅i❩)', True);
  Test('AreParallel(❨1, 2, i❩, ❨2, 4, i❩)', False);
  Test('AreParallel(❨1, 2, i❩, ❨−0.1, −0.3, −i/10❩)', False);
  Test('AreParallel(❨1, 2, i❩, ❨i, 2⋅i, 1❩)', False);
  Test('AreParallel(❨1, 2, i❩, ❨2 + i, 4 + 2⋅i, 1 + 2⋅i❩)', False);

  Test('AreParallel(❨1, 2, i❩, √3⋅❨2, 4, 2⋅i❩/9999)', True);
  Test('AreParallel(❨1, 2, i❩, -√π⋅❨2, 4, 2⋅i❩/e^5)', True);
  Test('AreParallel(❨1, 2, i❩, -√π⋅❨2, 4, 2⋅i❩/(2 - i))', True);
  Test('AreParallel(❨1, 2, i❩, -√π⋅e^2⋅❨2, 4, 2⋅i❩/sin(2 - i))', True);
  Test('AreParallel(❨1, 2, i❩, -√π⋅e^2⋅❨2, 4, 2.01⋅i❩/sin(2 - i))', False);

  Test('AreParallel(❨0, 1, 2❩, ❨0, 7 + 3⋅i/2, 14 + 3⋅i❩)', True);
  Test('AreParallel(❨0, 1, 2❩, ❨0, 7 + 3⋅i/2, 15 + 3⋅i❩)', False);
  Test('AreParallel(❨0, 1, 2❩, ❨1, 7 + 3⋅i/2, 14 + 3⋅i❩)', False);

  Test('AreParallel(❨0, 0, 0❩, ❨1, 7 + 3⋅i/2, 14 + 3⋅i❩)', True);
  Test('AreParallel(❨1, 2, i❩, ❨0, 0, 0❩)', True);

  Test('AreParallel(❨5.31 + 2.57⋅i, 4.56 − 9.63⋅i, −2.74 + 3.78⋅i❩, ❨4.84 − 7.27⋅i, −13.21 − 8.62⋅i, 5 + 4.78⋅i❩)', False);
  Test('AreParallel(❨5.31 + 2.57⋅i, 4.56 − 9.63⋅i, −2.74 + 3.78⋅i❩, ❨4.84 − 7.27⋅i, −13.21 − 8.62⋅i, 5 + 4.78⋅i❩, 0.01)', True);
  Test('AreParallel(❨5.31 + 2.57⋅i, 4.56 − 9.63⋅i, −2.74 + 3.78⋅i❩, ❨4.84 − 7.27⋅i, −13.21 − 8.62⋅i, 5 + 4.78⋅i❩, 0.001)', True);
  Test('AreParallel(❨5.31 + 2.57⋅i, 4.56 − 9.63⋅i, −2.74 + 3.78⋅i❩, ❨4.84 − 7.27⋅i, −13.21 − 8.62⋅i, 5 + 4.78⋅i❩, 0.0001)', False);

  Test('❨1, 3, 2❩ ∦ ❨10, 30, 20❩', False);
  Test('❨1, 3, 2❩ ∦ ❨-10, -30, -20❩', False);
  Test('❨10, 30, 20❩ ∦ ❨-0.2, -0.6, -0.4❩', False);
  Test('❨1, 3, 2❩ ∦ ❨10, 30, 21❩', True);
  Test('❨1, 3, 2❩ ∦ ❨2, 0, -3❩', True);

  Test('❨1, 3, 2, 7❩ ∦ ❨2, 6, 4, 14❩', False);
  Test('❨1, 3, 2, 7❩ ∦ ❨-0.2, -0.6, -0.4, -1.4❩', False);
  Test('❨1, 3, 2, 7❩ ∦ ❨2, 6, 4, 15❩', True);

  Test('❨1, 3, 2, 7❩ ∦ ❨2, 6, 4❩', failure, 'Vectors are of different dimension.');

  Test('❨1, 3, 2❩ ∦ ❨0, 0, 0❩', False);
  Test('❨e, √π, 2❩ ∦ ❨0, 0, 0❩', False);
  Test('❨sin(1), cos(2), arcsinh(2)❩ ∦ ❨0, 0, 0❩', False);
  Test('❨0, 0, 0❩ ∦ ❨-3, 3, 6❩', False);
  Test('❨0, 0, 0❩ ∦ ❨0, 0, 0❩', False);
  Test('❨0, 0, 0, 0❩ ∦ ❨0, 0, 0, 0❩', False);

  Test('❨arcsin(0.2), 0, 0❩ ∦ ❨2, 0, 0❩', False);
  Test('❨sinh(e), 0, 0❩ ∦ ❨-e, 0, 0❩', False);
  Test('❨sinh(e), 0, 0❩ ∦ ❨-e, 0, 5❩', True);
  Test('❨sinh(e), 1, 0❩ ∦ ❨-e, 0, 0❩', True);

  Test('❨sin(1), sin(2), cos(1)❩ ∦ √7⋅❨sin(1), sin(2), cos(1)❩', False);
  Test('❨sin(1), sin(2), cos(1)❩ ∦ −π⋅❨sin(1), sin(2), cos(1)❩/√2', False);
  Test('❨sin(1), sin(2), cos(1)❩ ∦ √7⋅❨sin(1), sin(2), tan(1)❩', True);

  Test('❨1, 3, 2❩ ∦ ❨2, 0, -3, 5❩', failure, 'Vectors are of different dimension.');

  Test('AreNotParallel(❨1, 3, 2❩, ❨10, 30, 20❩)', False);
  Test('AreNotParallel(❨1, 3, 2❩, ❨-10, -30, -20❩)', False);
  Test('AreNotParallel(❨10, 30, 20❩, ❨-0.2, -0.6, -0.4❩)', False);
  Test('AreNotParallel(❨1, 3, 2❩, ❨10, 30, 21❩)', True);
  Test('AreNotParallel(❨1, 3, 2❩, ❨2, 0, -3❩)', True);

  Test('AreNotParallel(❨1, 3, 2, 7❩, ❨2, 6, 4, 14❩)', False);
  Test('AreNotParallel(❨1, 3, 2, 7❩, ❨-0.2, -0.6, -0.4, -1.4❩)', False);
  Test('AreNotParallel(❨1, 3, 2, 7❩, ❨2, 6, 4, 15❩)', True);

  Test('AreNotParallel(❨1, 3, 2❩, ❨0, 0, 0❩)', False);
  Test('AreNotParallel(❨e, √π, 2❩, ❨0, 0, 0❩)', False);
  Test('AreNotParallel(❨sin(1), cos(2), arcsinh(2)❩, ❨0, 0, 0❩)', False);
  Test('AreNotParallel(❨0, 0, 0❩, ❨-3, 3, 6❩)', False);
  Test('AreNotParallel(❨0, 0, 0❩, ❨0, 0, 0❩)', False);
  Test('AreNotParallel(❨0, 0, 0, 0❩, ❨0, 0, 0, 0❩)', False);

  Test('AreNotParallel(❨arcsin(0.2), 0, 0❩, ❨2, 0, 0❩)', False);
  Test('AreNotParallel(❨sinh(e), 0, 0❩, ❨-e, 0, 0❩)', False);
  Test('AreNotParallel(❨sinh(e), 0, 0❩, ❨-e, 0, 5❩)', True);
  Test('AreNotParallel(❨sinh(e), 1, 0❩, ❨-e, 0, 0❩)', True);

  Test('AreNotParallel(❨sin(1), sin(2), cos(1)❩, √7⋅❨sin(1), sin(2), cos(1)❩)', False);
  Test('AreNotParallel(❨sin(1), sin(2), cos(1)❩, −π⋅❨sin(1), sin(2), cos(1)❩/√2)', False);
  Test('AreNotParallel(❨sin(1), sin(2), cos(1)❩, √7⋅❨sin(1), sin(2), tan(1)❩)', True);

  Test('AreNotParallel(❨1, 3, 2❩, ❨2, 0, -3, 5❩)', failure, 'Vectors are of different dimension.');

  Test('AreNotParallel()', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('AreNotParallel(❨1, 3, 2❩)', failure, 'Too few arguments. A required argument of type vector is missing.');

  Test('AreNotParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩)', True);
  Test('AreNotParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩, 0.01)', False);
  Test('AreNotParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩, 0.001)', False);
  Test('AreNotParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩, 0.0001)', True);
  Test('AreNotParallel(❨1.23, 5.19, −3.66❩, ❨2.13, 8.99, −6.34❩, 0.00001)', True);

  Test('❨1, 2, i❩ ∦ ❨2, 4, 2⋅i❩', False);
  Test('❨1, 2, i❩ ∦ ❨−0.1, −0.2, −i/10❩', False);
  Test('❨1, 2, i❩ ∦ ❨i, 2⋅i, −1❩', False);
  Test('❨1, 2, i❩ ∦ ❨2 + i, 4 + 2⋅i, −1 + 2⋅i❩', False);
  Test('❨1, 2, i❩ ∦ ❨2, 4, i❩', True);
  Test('❨1, 2, i❩ ∦ ❨−0.1, −0.3, −i/10❩', True);
  Test('❨1, 2, i❩ ∦ ❨i, 2⋅i, 1❩', True);
  Test('❨1, 2, i❩ ∦ ❨2 + i, 4 + 2⋅i, 1 + 2⋅i❩', True);

  Test('❨1, 2, i❩ ∦ ❨2, 4, 2⋅i, i❩', failure, 'Vectors are of different dimension.');

  Test('❨1, 2, i❩ ∦ √3⋅❨2, 4, 2⋅i❩/9999', False);
  Test('❨1, 2, i❩ ∦ -√π⋅❨2, 4, 2⋅i❩/e^5', False);
  Test('❨1, 2, i❩ ∦ -√π⋅❨2, 4, 2⋅i❩/(2 - i)', False);
  Test('❨1, 2, i❩ ∦ -√π⋅e^2⋅❨2, 4, 2⋅i❩/sin(2 - i)', False);
  Test('❨1, 2, i❩ ∦ -√π⋅e^2⋅❨2, 4, 2.01⋅i❩/sin(2 - i)', True);

  Test('❨0, 1, 2❩ ∦ ❨0, 7 + 3⋅i/2, 14 + 3⋅i❩', False);
  Test('❨0, 1, 2❩ ∦ ❨0, 7 + 3⋅i/2, 15 + 3⋅i❩', True);
  Test('❨0, 1, 2❩ ∦ ❨1, 7 + 3⋅i/2, 14 + 3⋅i❩', True);

  Test('❨0, 0, 0❩ ∦ ❨1, 7 + 3⋅i/2, 14 + 3⋅i❩', False);
  Test('❨1, 2, i❩ ∦ ❨0, 0, 0❩', False);

  Test('❨1, 2, i❩ ∦ 5', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');
  Test('❨1, 2, i❩ ∦ true', failure, 'An object of type vector was expected as argument 2, but an object of type boolean was given.');
  Test('IdentityMatrix(4) ∦ 5', failure, 'An object of type vector was expected as argument 1, but an object of type real matrix was given.');

  Test('AreNotParallel(❨1, 2, i❩, ❨2, 4, 2⋅i❩)', False);
  Test('AreNotParallel(❨1, 2, i❩, ❨−0.1, −0.2, −i/10❩)', False);
  Test('AreNotParallel(❨1, 2, i❩, ❨i, 2⋅i, −1❩)', False);
  Test('AreNotParallel(❨1, 2, i❩, ❨2 + i, 4 + 2⋅i, −1 + 2⋅i❩)', False);
  Test('AreNotParallel(❨1, 2, i❩, ❨2, 4, i❩)', True);
  Test('AreNotParallel(❨1, 2, i❩, ❨−0.1, −0.3, −i/10❩)', True);
  Test('AreNotParallel(❨1, 2, i❩, ❨i, 2⋅i, 1❩)', True);
  Test('AreNotParallel(❨1, 2, i❩, ❨2 + i, 4 + 2⋅i, 1 + 2⋅i❩)', True);

  Test('AreNotParallel(❨1, 2, i❩, √3⋅❨2, 4, 2⋅i❩/9999)', False);
  Test('AreNotParallel(❨1, 2, i❩, -√π⋅❨2, 4, 2⋅i❩/e^5)', False);
  Test('AreNotParallel(❨1, 2, i❩, -√π⋅❨2, 4, 2⋅i❩/(2 - i))', False);
  Test('AreNotParallel(❨1, 2, i❩, -√π⋅e^2⋅❨2, 4, 2⋅i❩/sin(2 - i))', False);
  Test('AreNotParallel(❨1, 2, i❩, -√π⋅e^2⋅❨2, 4, 2.01⋅i❩/sin(2 - i))', True);

  Test('AreNotParallel(❨0, 1, 2❩, ❨0, 7 + 3⋅i/2, 14 + 3⋅i❩)', False);
  Test('AreNotParallel(❨0, 1, 2❩, ❨0, 7 + 3⋅i/2, 15 + 3⋅i❩)', True);
  Test('AreNotParallel(❨0, 1, 2❩, ❨1, 7 + 3⋅i/2, 14 + 3⋅i❩)', True);

  Test('AreNotParallel(❨0, 0, 0❩, ❨1, 7 + 3⋅i/2, 14 + 3⋅i❩)', False);
  Test('AreNotParallel(❨1, 2, i❩, ❨0, 0, 0❩)', False);

  Test('AreNotParallel(❨5.31 + 2.57⋅i, 4.56 − 9.63⋅i, −2.74 + 3.78⋅i❩, ❨4.84 − 7.27⋅i, −13.21 − 8.62⋅i, 5 + 4.78⋅i❩)', True);
  Test('AreNotParallel(❨5.31 + 2.57⋅i, 4.56 − 9.63⋅i, −2.74 + 3.78⋅i❩, ❨4.84 − 7.27⋅i, −13.21 − 8.62⋅i, 5 + 4.78⋅i❩, 0.01)', False);
  Test('AreNotParallel(❨5.31 + 2.57⋅i, 4.56 − 9.63⋅i, −2.74 + 3.78⋅i❩, ❨4.84 − 7.27⋅i, −13.21 − 8.62⋅i, 5 + 4.78⋅i❩, 0.001)', False);
  Test('AreNotParallel(❨5.31 + 2.57⋅i, 4.56 − 9.63⋅i, −2.74 + 3.78⋅i❩, ❨4.84 − 7.27⋅i, −13.21 − 8.62⋅i, 5 + 4.78⋅i❩, 0.0001)', True);

  // - orthogonality

  Test('❨1, 0❩ ⟂ ❨0, 1❩', True);
  Test('❨-2, 0❩ ⟂ ❨0, 8❩', True);
  Test('❨1, 0❩ ⟂ ❨2, 1❩', False);

  Test('❨1, 2, 0❩ ⟂ ❨0, 0, −1❩', True);
  Test('❨1, 2, 0❩ ⟂ ❨1, 0, −1❩', False);
  Test('❨1, 2, 0❩ ⟂ ❨1, 3, −1❩', False);
  Test('❨1, 2, 2❩ ⟂ ❨1, 0, −1❩', False);

  Test('❨1, 2, −1❩ ⟂ ❨2, 0, 2❩', True);
  Test('❨1, 2, −1❩ ⟂ ❨2, 0, 1❩', False);
  Test('❨1, 2, 1❩ ⟂ ❨2, 0, 2❩', False);

  Test('❨1, 2, 2❩ ⟂ ❨0, 0, 0❩', True);
  Test('❨0, 0, 0❩ ⟂ ❨2, 0, 2❩', True);
  Test('❨0, 0, 0❩ ⟂ ❨0, 0, 0❩', True);

  Test('❨1, 0, 2, −1, 3, 1❩ ⟂ ❨0, 1, −1, 2, 2, −2❩', True);
  Test('❨1, 0, 2, −1, 3, 1❩ ⟂ ❨0, 1, −1, 2, 2, −1❩', False);
  Test('❨1, 0, 2, −1, 3, 1❩ ⟂ ❨1, 1, −1, 2, 2, −2❩', False);
  Test('❨1, 0, -2, −1, 3, 1❩ ⟂ ❨0, 1, −1, 2, 2, −2❩', False);

  Test('❨1, 2, 1❩ ⟂ ❨2, 0❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');

  Test('❨1, 2, 6❩ ⟂ 6', failure, 'An object of type vector was expected as argument 2, but an object of type integer was given.');

  Test('ArePerpendicular(❨2, 1❩, ❨−1, 2❩)', True);
  Test('ArePerpendicular(❨2, 1❩, ❨−1, 1❩)', False);
  Test('ArePerpendicular(❨2, 1❩, ❨1, 2❩)', False);

  Test('ArePerpendicular(❨0, 1, 0❩, ❨0, 0, 1❩)', True);
  Test('ArePerpendicular(❨0, 1, 0❩, ❨0, 1, 1❩)', False);
  Test('ArePerpendicular(❨0, 0, 0❩, ❨0, 1, 1❩)', True);

  Test('ArePerpendicular(❨2, −1, 4, 2, 1, 0❩, ❨1, 2, −3, 4, 4, 2❩)', True);
  Test('ArePerpendicular(❨2, −1, 4, 2, 1, 0❩, ❨1, 2, −3, 4, 4, 5❩)', True);
  Test('ArePerpendicular(❨2, −1, 4, 2, 1, 0❩, ❨1, 2, −3, 3, 4, 2❩)', False);
  Test('ArePerpendicular(❨2, 1, 4, 2, 1, 0❩, ❨1, 2, −3, 4, 4, 2❩)', False);
  Test('ArePerpendicular(❨1, −1, 4, 2, 1, 0❩, ❨1, 2, −3, 4, 4, 2❩)', False);

  Test('ArePerpendicular(❨0.80, 1.55, −2.00, 1.41❩, ❨2.45, 1.03, 2.23, 0.64❩)', False);
  Test('ArePerpendicular(❨0.80, 1.55, −2.00, 1.41❩, ❨2.45, 1.03, 2.23, 0.64❩, 0.1)', True);
  Test('ArePerpendicular(❨0.80, 1.55, −2.00, 1.41❩, ❨2.45, 1.03, 2.23, 0.64❩, 0.01)', True);
  Test('ArePerpendicular(❨0.80, 1.55, −2.00, 1.41❩, ❨2.45, 1.03, 2.23, 0.64❩, 0.001)', False);

  Test('❨2, i, 1❩ ⟂ ❨1, i, −3❩', True);
  Test('❨2, i, 1❩ ⟂ ❨1, i, 3❩', False);
  Test('❨1, 2, 0, 1❩ ⟂ ❨3, 1, i, −5❩', True);
  Test('❨1, 2, 0, 1❩ ⟂ ❨3, 1, i, −4❩', False);
  Test('❨1, 2, 2, 1❩ ⟂ ❨3, 1, i, −5❩', False);

  Test('❨1, 2, 2, 1❩ ⟂ ❨3, 1, i, −5, 4❩', failure, 'Cannot compute dot product of two vectors of unequal dimension.');
  Test('❨1, 2, i❩ ⟂ "cat"', failure, 'An object of type vector was expected as argument 2, but an object of type string was given.');

  Test('ArePerpendicular(❨1 + i, 2, −1❩, ❨i, −i, 1 − i❩)', True);
  Test('ArePerpendicular(❨1 + i, 2, −1❩, ❨i, −i, 1 + i❩)', False);

  Test('ArePerpendicular(❨0.11 + 0.45⋅i, 0.17 − 0.60⋅i, −0.18 + 0.60⋅i❩, ❨2.46 − 0.93⋅i, 2.41 − 0.70⋅i, 0.48 − 0.98⋅i❩)', False);
  Test('ArePerpendicular(❨0.11 + 0.45⋅i, 0.17 − 0.60⋅i, −0.18 + 0.60⋅i❩, ❨2.46 − 0.93⋅i, 2.41 − 0.70⋅i, 0.48 − 0.98⋅i❩, 0.1)', True);
  Test('ArePerpendicular(❨0.11 + 0.45⋅i, 0.17 − 0.60⋅i, −0.18 + 0.60⋅i❩, ❨2.46 − 0.93⋅i, 2.41 − 0.70⋅i, 0.48 − 0.98⋅i❩, 0.01)', True);
  Test('ArePerpendicular(❨0.11 + 0.45⋅i, 0.17 − 0.60⋅i, −0.18 + 0.60⋅i❩, ❨2.46 − 0.93⋅i, 2.41 − 0.70⋅i, 0.48 − 0.98⋅i❩, 0.001)', False);

  Test('ArePerpendicular()', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('ArePerpendicular(❨1, 2, −1❩)', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('ArePerpendicular(❨1 + i, 2, −1❩)', failure, 'Too few arguments. A required argument of type vector is missing.');

  // - norms, real spaces

  Test('u ≔ ❨1, 2❩; v ≔ ❨1, 2, −4❩; w ≔ ❨2, 0, 1, −1❩; x ≔ ❨7, 4, −6, 5, 2❩;', null);

  Eps; Test('norm(u)', Sqrt(5));
  Eps; Test('norm(v)', Sqrt(21));
  Eps; Test('norm(w)', Sqrt(6));
  Eps; Test('norm(x)', Sqrt(130));

  Eps; Test('norm(u, "Euclidean")', Sqrt(5));
  Eps; Test('norm(v, "Euclidean")', Sqrt(21));
  Eps; Test('norm(w, "Euclidean")', Sqrt(6));
  Eps; Test('norm(x, "Euclidean")', Sqrt(130));

  Eps; Test('norm(u, "Frobenius")', Sqrt(5));
  Eps; Test('norm(v, "Frobenius")', Sqrt(21));
  Eps; Test('norm(w, "Frobenius")', Sqrt(6));
  Eps; Test('norm(x, "Frobenius")', Sqrt(130));

  Eps; Test('norm(u, "p", 2)', Sqrt(5));
  Eps; Test('norm(v, "p", 2)', Sqrt(21));
  Eps; Test('norm(w, "p", 2)', Sqrt(6));
  Eps; Test('norm(x, "p", 2)', Sqrt(130));

  Eps; Test('norm(u, "p", 3)', cbrt(9));
  Eps; Test('norm(v, "p", 3)', cbrt(73));
  Eps; Test('norm(w, "p", 3)', cbrt(10));
  Eps; Test('norm(x, "p", 3)', cbrt(756));

  Eps; Test('norm(u, "p", 4)', frt(17));
  Eps; Test('norm(v, "p", 4)', frt(273));
  Eps; Test('norm(w, "p", 4)', frt(18));
  Eps; Test('norm(x, "p", 4)', frt(4594));

  Eps; Test('norm(u, "p", 1)', 3.0);
  Eps; Test('norm(v, "p", 1)', 7.0);
  Eps; Test('norm(w, "p", 1)', 4.0);
  Eps; Test('norm(x, "p", 1)', 24.0);

  Eps; Test('norm(u, "sum")', 3.0);
  Eps; Test('norm(v, "sum")', 7.0);
  Eps; Test('norm(w, "sum")', 4.0);
  Eps; Test('norm(x, "sum")', 24.0);

  Eps; Test('norm(u, "max column sum")', 3.0);
  Eps; Test('norm(v, "max column sum")', 7.0);
  Eps; Test('norm(w, "max column sum")', 4.0);
  Eps; Test('norm(x, "max column sum")', 24.0);

  Eps; Test('norm(u, "max")', 2.0);
  Eps; Test('norm(v, "max")', 4.0);
  Eps; Test('norm(w, "max")', 2.0);
  Eps; Test('norm(x, "max")', 7.0);

  Eps; Test('norm(u, "max row sum")', 2.0);
  Eps; Test('norm(v, "max row sum")', 4.0);
  Eps; Test('norm(w, "max row sum")', 2.0);
  Eps; Test('norm(x, "max row sum")', 7.0);

  Eps; Test('norm(u, "k", 1)', 2.0);
  Eps; Test('norm(u, "k", 2)', 3.0);
  Eps; Test('norm(u, "k", 3)', 3.0);
  Eps; Test('norm(v, "k", 1)', 4.0);
  Eps; Test('norm(v, "k", 2)', 6.0);
  Eps; Test('norm(v, "k", 3)', 7.0);
  Eps; Test('norm(v, "k", 4)', 7.0);
  Eps; Test('norm(w, "k", 1)', 2.0);
  Eps; Test('norm(w, "k", 2)', 3.0);
  Eps; Test('norm(w, "k", 3)', 4.0);
  Eps; Test('norm(w, "k", 4)', 4.0);
  Eps; Test('norm(w, "k", 5)', 4.0);
  Eps; Test('norm(x, "k", 1)', 7.0);
  Eps; Test('norm(x, "k", 2)', 13.0);
  Eps; Test('norm(x, "k", 3)', 18.0);
  Eps; Test('norm(x, "k", 4)', 22.0);
  Eps; Test('norm(x, "k", 5)', 24.0);
  Eps; Test('norm(x, "k", 6)', 24.0);

  Test('norm(u, "spectral")', failure, 'Cannot compute spectral norm of an object of type "real vector".');
  Test('norm(x, "nargle")', failure, 'Unknown norm type "nargle".');

  Eps; Test('NormSquared(u)', 5.0);
  Eps; Test('NormSquared(v)', 21.0);
  Eps; Test('NormSquared(w)', 6.0);
  Eps; Test('NormSquared(x)', 130.0);

  Eps; Test('normalized(u)', ASO(ASR2(1, 2) / Sqrt(5)));
  Eps; Test('normalized(v)', ASO(ASR3(1, 2, -4) / Sqrt(21)));
  Eps; Test('normalized(w)', ASO(ASR4(2, 0, 1, -1) / Sqrt(6)));
  Eps; Test('normalized(x)', ASO(ASR5(7, 4, -6, 5, 2) / Sqrt(130)));

  // - norms, complex spaces

  Test('u ≔ ❨1, 2⋅i❩; v ≔ ❨3, 1 + i, −4❩; w ≔ ❨2, 4 + 5⋅i, 3, −i❩; x ≔ ❨5, 2 + i, 1, 1 − 2⋅i, −4❩;', null);

  Eps; Test('norm(u)', Sqrt(5));
  Eps; Test('norm(v)', Sqrt(27));
  Eps; Test('norm(w)', Sqrt(55));
  Eps; Test('norm(x)', Sqrt(52));

  Eps; Test('norm(u, "Euclidean")', Sqrt(5));
  Eps; Test('norm(v, "Euclidean")', Sqrt(27));
  Eps; Test('norm(w, "Euclidean")', Sqrt(55));
  Eps; Test('norm(x, "Euclidean")', Sqrt(52));

  Eps; Test('norm(u, "Frobenius")', Sqrt(5));
  Eps; Test('norm(v, "Frobenius")', Sqrt(27));
  Eps; Test('norm(w, "Frobenius")', Sqrt(55));
  Eps; Test('norm(x, "Frobenius")', Sqrt(52));

  Eps; Test('norm(u, "p", 2)', Sqrt(5));
  Eps; Test('norm(v, "p", 2)', Sqrt(27));
  Eps; Test('norm(w, "p", 2)', Sqrt(55));
  Eps; Test('norm(x, "p", 2)', Sqrt(52));

  Eps; Test('norm(u, "p", 3)', cbrt(9));
  Eps; Test('norm(v, "p", 3)', cbrt(93.8284271247461901));
  Eps; Test('norm(w, "p", 3)', cbrt(298.528093734746796));
  Eps; Test('norm(x, "p", 3)', cbrt(212.360679774997897));

  Eps; Test('norm(u, "p", 4)', frt(17));
  Eps; Test('norm(v, "p", 4)', frt(341));
  Eps; Test('norm(w, "p", 4)', frt(1779));
  Eps; Test('norm(x, "p", 4)', frt(932));

  Eps; Test('norm(u, "p", 1)', 3.0);
  Eps; Test('norm(v, "p", 1)', 7 + Sqrt(2));
  Eps; Test('norm(w, "p", 1)', 6 + Sqrt(16 + 25));
  Eps; Test('norm(x, "p", 1)', 10 + Sqrt(5) + Sqrt(5));

  Eps; Test('norm(u, "sum")', 3.0);
  Eps; Test('norm(v, "sum")', 7 + Sqrt(2));
  Eps; Test('norm(w, "sum")', 6 + Sqrt(16 + 25));
  Eps; Test('norm(x, "sum")', 10 + Sqrt(5) + Sqrt(5));

  Eps; Test('norm(u, "max column sum")', 3.0);
  Eps; Test('norm(v, "max column sum")', 7 + Sqrt(2));
  Eps; Test('norm(w, "max column sum")', 6 + Sqrt(16 + 25));
  Eps; Test('norm(x, "max column sum")', 10 + Sqrt(5) + Sqrt(5));

  Eps; Test('norm(u, "max")', 2.0);
  Eps; Test('norm(v, "max")', 4.0);
  Eps; Test('norm(w, "max")', Sqrt(16 + 25));
  Eps; Test('norm(x, "max")', 5.0);

  Eps; Test('norm(u, "max row sum")', 2.0);
  Eps; Test('norm(v, "max row sum")', 4.0);
  Eps; Test('norm(w, "max row sum")', Sqrt(16 + 25));
  Eps; Test('norm(x, "max row sum")', 5.0);

  Eps; Test('norm(u, "k", 1)', 2.0);
  Eps; Test('norm(u, "k", 2)', 3.0);
  Eps; Test('norm(u, "k", 3)', 3.0);
  Eps; Test('norm(v, "k", 1)', 4.0);
  Eps; Test('norm(v, "k", 2)', 7.0);
  Eps; Test('norm(v, "k", 3)', 7.0 + Sqrt(2));
  Eps; Test('norm(v, "k", 4)', 7.0 + Sqrt(2));
  Eps; Test('norm(w, "k", 1)', Sqrt(16 + 25));
  Eps; Test('norm(w, "k", 2)', Sqrt(16 + 25) + 3);
  Eps; Test('norm(w, "k", 3)', Sqrt(16 + 25) + 3 + 2);
  Eps; Test('norm(w, "k", 4)', Sqrt(16 + 25) + 3 + 2 + 1);
  Eps; Test('norm(w, "k", 5)', Sqrt(16 + 25) + 3 + 2 + 1);
  Eps; Test('norm(x, "k", 1)', 5.0);
  Eps; Test('norm(x, "k", 2)', 5.0 + 4.0);
  Eps; Test('norm(x, "k", 3)', 5.0 + 4.0 + Sqrt(5));
  Eps; Test('norm(x, "k", 4)', 5.0 + 4.0 + Sqrt(5) + Sqrt(5));
  Eps; Test('norm(x, "k", 5)', 5.0 + 4.0 + Sqrt(5) + Sqrt(5) + 1);
  Eps; Test('norm(x, "k", 6)', 5.0 + 4.0 + Sqrt(5) + Sqrt(5) + 1);

  Test('norm(u, "spectral")', failure, 'Cannot compute spectral norm of an object of type "complex vector".');
  Test('norm(x, "nargle")', failure, 'Unknown norm type "nargle".');

  Test('NormSquared(u)', 5.0);
  Test('NormSquared(v)', 27.0);
  Test('NormSquared(w)', 55.0);
  Test('NormSquared(x)', 52.0);

  Test('normalized(u)', ASO(ASC2(1, 2*ImaginaryUnit) / Sqrt(5)));
  Test('normalized(v)', ASO(ASC3(3, 1 + ImaginaryUnit, -4) / Sqrt(27)));
  Test('normalized(w)', ASO(ASC4(2, 4 + 5*ImaginaryUnit, 3, -ImaginaryUnit) / Sqrt(55)));
  Test('normalized(x)', ASO(ASC5(5, 2 + ImaginaryUnit, 1, 1 - 2*ImaginaryUnit, -4) / Sqrt(52)));

  Test('delete(u); delete(v); delete(w); delete(x)', success);

  // - transpose and conjugate transpose

  Test('❨1, 2, 3❩*', 3, [1, 2, 3]);
  Test('❨2 + 3⋅i, 5 − i, 7 + 5⋅i❩*', 3, [2 - 3*ImaginaryUnit, 5 + ImaginaryUnit, 7 - 5*ImaginaryUnit]);
  Test('❨2, i, 5, 1 − i, 6⋅i❩*', 5, [2, -ImaginaryUnit, 5, 1 + ImaginaryUnit, -6*ImaginaryUnit]);

  Test('transpose(❨6, 2, 8❩)', 3, [6, 2, 8]);
  Test('transpose(❨2 + 3⋅i, 5 − i, 7 + 5⋅i❩)', 3, [2 + 3*ImaginaryUnit, 5 - ImaginaryUnit, 7 + 5*ImaginaryUnit]);
  Test('transpose(❨2, i, 5, 1 − i, 6⋅i❩)', 5, [2, ImaginaryUnit, 5, 1 - ImaginaryUnit, 6*ImaginaryUnit]);

  Test('ConjugateTranspose(❨6, 2, 8❩)', 3, [6, 2, 8]);
  Test('ConjugateTranspose(❨2 + 3⋅i, 5 − i, 7 + 5⋅i❩)', 3, [2 - 3*ImaginaryUnit, 5 + ImaginaryUnit, 7 - 5*ImaginaryUnit]);
  Test('ConjugateTranspose(❨2, i, 5, 1 − i, 6⋅i❩)', 5, [2, -ImaginaryUnit, 5, 1 + ImaginaryUnit, -6*ImaginaryUnit]);

  // Operations on matrices

  // - norms, real spaces

  Test('A ≔ ❨❨1, 5, −2❩, ❨2, −1, −8❩, ❨0, 1, 3❩❩; B ≔ ❨❨2, 1, −3❩, ❨5, 0, 1❩❩; C ≔ ❨❨1, 5, 2, 3❩, ❨−4, 1, 2, 0❩, ❨3, 2, 7, −1❩, ❨5, 2, 0, 3❩❩;', null);

  Eps; Test('norm(A)', Sqrt(109));
  Eps; Test('norm(B)', Sqrt(40));
  Eps; Test('norm(C)', Sqrt(161));

  Eps; Test('norm(A, "Euclidean")', Sqrt(109));
  Eps; Test('norm(B, "Euclidean")', Sqrt(40));
  Eps; Test('norm(C, "Euclidean")', Sqrt(161));

  Eps; Test('norm(A, "Frobenius")', Sqrt(109));
  Eps; Test('norm(B, "Frobenius")', Sqrt(40));
  Eps; Test('norm(C, "Frobenius")', Sqrt(161));

  Eps; Test('norm(A, "p", 2)', Sqrt(109));
  Eps; Test('norm(B, "p", 2)', Sqrt(40));
  Eps; Test('norm(C, "p", 2)', Sqrt(161));

  Eps; Test('norm(A, "p", 3)', cbrt(683));
  Eps; Test('norm(B, "p", 3)', cbrt(162));
  Eps; Test('norm(C, "p", 3)', cbrt(773));

  Eps; Test('norm(A, "p", 4)', frt(4837));
  Eps; Test('norm(B, "p", 4)', frt(724));
  Eps; Test('norm(C, "p", 4)', frt(4217));

  Eps; Test('norm(A, "p", 1)', 23.0);
  Eps; Test('norm(B, "p", 1)', 12.0);
  Eps; Test('norm(C, "p", 1)', 41.0);

  Eps; Test('norm(A, "sum")', 23.0);
  Eps; Test('norm(B, "sum")', 12.0);
  Eps; Test('norm(C, "sum")', 41.0);

  Eps; Test('norm(A, "max")', 8.0);
  Eps; Test('norm(B, "max")', 5.0);
  Eps; Test('norm(C, "max")', 7.0);

  Eps; Test('norm(A, "k", 1)', 8.0);
  Eps; Test('norm(A, "k", 2)', 13.0);
  Eps; Test('norm(A, "k", 3)', 16.0);
  Eps; Test('norm(A, "k", 4)', 18.0);
  Eps; Test('norm(A, "k", 5)', 20.0);
  Eps; Test('norm(A, "k", 6)', 21.0);
  Eps; Test('norm(A, "k", 7)', 22.0);
  Eps; Test('norm(A, "k", 8)', 23.0);
  Eps; Test('norm(A, "k", 9)', 23.0);
  Eps; Test('norm(A, "k", 10)', 23.0);

  Eps; Test('norm(B, "k", 1)', 5.0);
  Eps; Test('norm(B, "k", 2)', 8.0);
  Eps; Test('norm(B, "k", 3)', 10.0);
  Eps; Test('norm(B, "k", 4)', 11.0);
  Eps; Test('norm(B, "k", 5)', 12.0);
  Eps; Test('norm(B, "k", 6)', 12.0);
  Eps; Test('norm(B, "k", 7)', 12.0);

  Eps; Test('norm(C, "k", 1)', 7.0);
  Eps; Test('norm(C, "k", 2)', 12.0);
  Eps; Test('norm(C, "k", 3)', 17.0);
  Eps; Test('norm(C, "k", 4)', 21.0);
  Eps; Test('norm(C, "k", 5)', 24.0);
  Eps; Test('norm(C, "k", 6)', 27.0);
  Eps; Test('norm(C, "k", 7)', 30.0);
  Eps; Test('norm(C, "k", 8)', 32.0);
  Eps; Test('norm(C, "k", 9)', 34.0);
  Eps; Test('norm(C, "k", 10)', 36.0);
  Eps; Test('norm(C, "k", 11)', 38.0);
  Eps; Test('norm(C, "k", 12)', 39.0);
  Eps; Test('norm(C, "k", 13)', 40.0);
  Eps; Test('norm(C, "k", 14)', 41.0);
  Eps; Test('norm(C, "k", 15)', 41.0);
  Eps; Test('norm(C, "k", 16)', 41.0);
  Eps; Test('norm(C, "k", 17)', 41.0);

  Eps; Test('norm(A, "max column sum")', 13.0);
  Eps; Test('norm(B, "max column sum")', 7.0);
  Eps; Test('norm(C, "max column sum")', 13.0);

  Eps; Test('norm(A, "max row sum")', 11.0);
  Eps; Test('norm(B, "max row sum")', 6.0);
  Eps; Test('norm(C, "max row sum")', 13.0);

  Eps; Test('norm(A, "spectral")', 9.01391777220033201);
  Eps; Test('norm(B, "spectral")', 5.40551056397939965);
  Eps; Test('norm(C, "spectral")', 9.56001072060560776);

  // - norms, complex spaces

  Test('A ≔ ❨❨2, 1, 1 + i❩, ❨3 − 2⋅i, 2, 4❩, ❨1, 0, i❩❩; B ≔ ❨❨1, 0, i❩, ❨−i, 2, 1❩❩; C ≔ ❨❨2, 1, 3, 2 − 3⋅i❩, ❨3, i, 2, 0❩, ❨1 + 2⋅i, 3, 1 − i, 3❩, ❨0, 2, i, −2⋅i❩❩;', null);

  Eps; Test('norm(A)', Sqrt(42));
  Eps; Test('norm(B)', Sqrt(8));
  Eps; Test('norm(C)', Sqrt(75));

  Eps; Test('norm(A, "Euclidean")', Sqrt(42));
  Eps; Test('norm(B, "Euclidean")', Sqrt(8));
  Eps; Test('norm(C, "Euclidean")', Sqrt(75));

  Eps; Test('norm(A, "Frobenius")', Sqrt(42));
  Eps; Test('norm(B, "Frobenius")', Sqrt(8));
  Eps; Test('norm(C, "Frobenius")', Sqrt(75));

  Eps; Test('norm(A, "p", 2)', Sqrt(42));
  Eps; Test('norm(B, "p", 2)', Sqrt(8));
  Eps; Test('norm(C, "p", 2)', Sqrt(75));

  Eps; Test('norm(A, "p", 3)', cbrt(132.700593705778051));
  Eps; Test('norm(B, "p", 3)', cbrt(12));
  Eps; Test('norm(C, "p", 3)', cbrt(203.880933593276999));

  Eps; Test('norm(A, "p", 4)', frt(464));
  Eps; Test('norm(B, "p", 4)', frt(20));
  Eps; Test('norm(C, "p", 4)', frt(589));

  Eps; Test('norm(A, "p", 1)', 11.0 + Sqrt(2) + Sqrt(13));
  Eps; Test('norm(B, "p", 1)', 6.0);
  Eps; Test('norm(C, "p", 1)', 23.0 + Sqrt(13) + Sqrt(5) + Sqrt(2));

  Eps; Test('norm(A, "sum")', 11.0 + Sqrt(2) + Sqrt(13));
  Eps; Test('norm(B, "sum")', 6.0);
  Eps; Test('norm(C, "sum")', 23.0 + Sqrt(13) + Sqrt(5) + Sqrt(2));

  Eps; Test('norm(A, "max")', 4.0);
  Eps; Test('norm(B, "max")', 2.0);
  Eps; Test('norm(C, "max")', Sqrt(13));

  Eps; Test('norm(A, "k", 1)', 4.0);
  Eps; Test('norm(A, "k", 2)', 4.0 + Sqrt(13));
  Eps; Test('norm(A, "k", 3)', 4.0 + Sqrt(13) + 2);
  Eps; Test('norm(A, "k", 4)', 4.0 + Sqrt(13) + 4);
  Eps; Test('norm(A, "k", 5)', 4.0 + Sqrt(13) + 4 + Sqrt(2));
  Eps; Test('norm(A, "k", 6)', 4.0 + Sqrt(13) + 4 + Sqrt(2) + 1);
  Eps; Test('norm(A, "k", 7)', 4.0 + Sqrt(13) + 4 + Sqrt(2) + 2);
  Eps; Test('norm(A, "k", 8)', 4.0 + Sqrt(13) + 4 + Sqrt(2) + 3);
  Eps; Test('norm(A, "k", 9)', 4.0 + Sqrt(13) + 4 + Sqrt(2) + 3);
  Eps; Test('norm(A, "k", 10)', 4.0 + Sqrt(13) + 4 + Sqrt(2) + 3);

  Eps; Test('norm(B, "k", 1)', 2.0);
  Eps; Test('norm(B, "k", 2)', 3.0);
  Eps; Test('norm(B, "k", 3)', 4.0);
  Eps; Test('norm(B, "k", 4)', 5.0);
  Eps; Test('norm(B, "k", 5)', 6.0);
  Eps; Test('norm(B, "k", 6)', 6.0);
  Eps; Test('norm(B, "k", 7)', 6.0);

  Eps; Test('norm(C, "k", 1)', Sqrt(13));
  Eps; Test('norm(C, "k", 2)', Sqrt(13) + 3);
  Eps; Test('norm(C, "k", 3)', Sqrt(13) + 6);
  Eps; Test('norm(C, "k", 4)', Sqrt(13) + 9);
  Eps; Test('norm(C, "k", 5)', Sqrt(13) + 12);
  Eps; Test('norm(C, "k", 6)', Sqrt(13) + 12 + Sqrt(5));
  Eps; Test('norm(C, "k", 7)', Sqrt(13) + 12 + Sqrt(5) + 2);
  Eps; Test('norm(C, "k", 8)', Sqrt(13) + 12 + Sqrt(5) + 4);
  Eps; Test('norm(C, "k", 9)', Sqrt(13) + 12 + Sqrt(5) + 6);
  Eps; Test('norm(C, "k", 10)', Sqrt(13) + 12 + Sqrt(5) + 8);
  Eps; Test('norm(C, "k", 11)', Sqrt(13) + 12 + Sqrt(5) + 8 + Sqrt(2));
  Eps; Test('norm(C, "k", 12)', Sqrt(13) + 12 + Sqrt(5) + 8 + Sqrt(2) + 1);
  Eps; Test('norm(C, "k", 13)', Sqrt(13) + 12 + Sqrt(5) + 8 + Sqrt(2) + 2);
  Eps; Test('norm(C, "k", 14)', Sqrt(13) + 12 + Sqrt(5) + 8 + Sqrt(2) + 3);
  Eps; Test('norm(C, "k", 15)', Sqrt(13) + 12 + Sqrt(5) + 8 + Sqrt(2) + 3);
  Eps; Test('norm(C, "k", 16)', Sqrt(13) + 12 + Sqrt(5) + 8 + Sqrt(2) + 3);
  Eps; Test('norm(C, "k", 17)', Sqrt(13) + 12 + Sqrt(5) + 8 + Sqrt(2) + 3);

  Eps; Test('norm(A, "max column sum")', 3 + Sqrt(13));
  Eps; Test('norm(B, "max column sum")', 2.0);
  Eps; Test('norm(C, "max column sum")', Sqrt(13) + 5);

  Eps; Test('norm(A, "max row sum")', 6 + Sqrt(13));
  Eps; Test('norm(B, "max row sum")', 4.0);
  Eps; Test('norm(C, "max row sum")', 6 + Sqrt(5) + Sqrt(2));

  Eps; Test('norm(A, "spectral")', 6.38943232171595646);
  Eps; Test('norm(B, "spectral")', 2.61312592975275306);
  Eps; Test('norm(C, "spectral")', 7.08203992175305035);

  // - transpose and conjugate transpose

  Test('A ≔ ❨❨1, 5, −2❩, ❨2, −1, −8❩, ❨0, 1, 3❩❩; B ≔ ❨❨2, 1, −3❩, ❨5, 0, 1❩❩; C ≔ ❨❨1, 5, 2, 3❩, ❨−4, 1, 2, 0❩, ❨3, 2, 7, −1❩, ❨5, 2, 0, 3❩❩;', null);

  Test('A*', 3, [1, 2, 0, 5, -1, 1, -2, -8, 3]);
  Test('B*', 2, [2, 5, 1, 0, -3, 1]);
  Test('C*', 4, [1, -4, 3, 5, 5, 1, 2, 2, 2, 2, 7, 0, 3, 0, -1, 3]);

  Test('ConjugateTranspose(A)', 3, [1, 2, 0, 5, -1, 1, -2, -8, 3]);
  Test('ConjugateTranspose(B)', 2, [2, 5, 1, 0, -3, 1]);
  Test('ConjugateTranspose(C)', 4, [1, -4, 3, 5, 5, 1, 2, 2, 2, 2, 7, 0, 3, 0, -1, 3]);

  Test('transpose(A)', 3, [1, 2, 0, 5, -1, 1, -2, -8, 3]);
  Test('transpose(B)', 2, [2, 5, 1, 0, -3, 1]);
  Test('transpose(C)', 4, [1, -4, 3, 5, 5, 1, 2, 2, 2, 2, 7, 0, 3, 0, -1, 3]);

  Test('A ≔ ❨❨2, 1, 1 + i❩, ❨3 − 2⋅i, 2, 4❩, ❨1, 0, i❩❩; B ≔ ❨❨1, 0, i❩, ❨−i, 2, 1❩❩; C ≔ ❨❨2, 1, 3, 2 − 3⋅i❩, ❨3, i, 2, 0❩, ❨1 + 2⋅i, 3, 1 − i, 3❩, ❨0, 2, i, −2⋅i❩❩;', null);

  Test('A*', 3, [2, 3 + 2*ImaginaryUnit, 1, 1, 2, 0, 1 - ImaginaryUnit, 4, -ImaginaryUnit]);
  Test('B*', 2, [1, ImaginaryUnit, 0, 2, -ImaginaryUnit, 1]);
  Test('C*', 4, [2, 3, 1 - 2*ImaginaryUnit, 0, 1, -ImaginaryUnit, 3, 2, 3, 2, 1 + ImaginaryUnit, -ImaginaryUnit, 2 + 3*ImaginaryUnit, 0, 3, 2*ImaginaryUnit]);

  Test('ConjugateTranspose(A)', 3, [2, 3 + 2*ImaginaryUnit, 1, 1, 2, 0, 1 - ImaginaryUnit, 4, -ImaginaryUnit]);
  Test('ConjugateTranspose(B)', 2, [1, ImaginaryUnit, 0, 2, -ImaginaryUnit, 1]);
  Test('ConjugateTranspose(C)', 4, [2, 3, 1 - 2*ImaginaryUnit, 0, 1, -ImaginaryUnit, 3, 2, 3, 2, 1 + ImaginaryUnit, -ImaginaryUnit, 2 + 3*ImaginaryUnit, 0, 3, 2*ImaginaryUnit]);

  Test('transpose(A)', 3, [2, 3 - 2*ImaginaryUnit, 1, 1, 2, 0, 1 + ImaginaryUnit, 4, ImaginaryUnit]);
  Test('transpose(B)', 2, [1, -ImaginaryUnit, 0, 2, ImaginaryUnit, 1]);
  Test('transpose(C)', 4, [2, 3, 1 + 2*ImaginaryUnit, 0, 1, ImaginaryUnit, 3, 2, 3, 2, 1 - ImaginaryUnit, ImaginaryUnit, 2 - 3*ImaginaryUnit, 0, 3, -2*ImaginaryUnit]);

  Test('delete(A); delete(B); delete(C)', success);


  //
  // Special real matrices
  //

  Test('A ≔ ❨❨5, 3, 1❩, ❨0, −2, 4❩, ❨−7, 1, 6❩❩; B ≔ ❨❨4, 2, 0❩, ❨−1, 2, 5❩❩; C ≔ ❨❨2, 3, 0, −1❩, ❨2, 4, −5, 5❩, ❨0, 7, 2, 6❩, ❨2, 8, 7, 1❩❩; D ≔ ❨❨1❩, ❨0❩, ❨−1❩, ❨2❩❩;', null);
  Test('E ≔ ❨❨4, 8, 1❩❩; F ≔ ❨❨4, 0, 0❩, ❨0, 6, 0❩, ❨0, 0, −2❩❩; g ≔ ❨❨2, 3, 0❩, ❨1, 2, 6❩, ❨3, 2, 0❩❩; h ≔ ❨❨1, 2, 1❩, ❨2, 3, 8❩, ❨1, 6, 2❩❩;', null);
  Test('G ≔ ❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩, ❨0, 0, 0❩❩; H ≔ ❨❨3, 2, 0, 3❩, ❨0, 1, 1, 0❩, ❨0, 0, 5, 2❩, ❨0, 0, 0, −2❩❩; I ≔ ❨❨2, 0, 0, 0❩, ❨0, 3, 0, 0❩, ❨0, 2, 5, 0❩, ❨1, 3, 3, 0❩❩;', null);
  Test('J ≔ ❨❨2, 2, 0, 0❩, ❨0, 3, 3, 0❩, ❨0, 2, 5, 1❩, ❨1, 3, 3, 0❩❩; K ≔ ❨❨−2, 0, −3, −5❩, ❨0, 0, 0, −1❩❩; L ≔ ❨❨−1, −2, −4, −5❩, ❨−2, −5, −3, −4❩, ❨−8, −6, −9, −7❩❩;', null);
  Test('M ≔ ❨❨1, 2, 0, 0❩, ❨0, 2, 3, 0❩, ❨0, 0, 8, 4❩, ❨0, 0, 0, −1❩❩; N ≔ ❨❨0, 1❩, ❨0, 1❩, ❨1, 0❩, ❨0, 1❩❩; O ≔ ❨❨1, 0, 0❩, ❨0, 0, 1❩, ❨0, 1, 0❩❩; P ≔ ❨❨1, 2, 3, 4❩, ❨4, 1, 2, 3❩, ❨3, 4, 1, 2❩, ❨2, 3, 4, 1❩❩;', null);
  Test('Q ≔ ❨❨1, 4, 16, 64❩, ❨1, 2, 4, 8❩, ❨1, 3, 9, 27❩, ❨1, 5, 25, 125❩❩; R ≔ ❨❨1, 1, 1, 1❩, ❨1, 2, 3, 4❩, ❨1, 4, 9, 16❩, ❨1, 8, 27, 64❩❩; S ≔ ❨❨4, −2, 7❩, ❨−2, 1, 9❩, ❨7, 9, 3❩❩;', null);
  Test('T ≔ ❨❨1, 0, 0❩, ❨0, 1/√2, −1/√2❩, ❨0, 1/√2, 1/√2❩❩; U ≔ ❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 0❩❩; V ≔ ❨❨−1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩; W ≔ ❨❨2, 0, 0❩, ❨2, 2, 0❩, ❨0, 0, 2❩❩;', null);
  Test('X ≔ ❨❨5, 2, 7, −6❩, ❨0, 5, 2, 7❩, ❨0, 0, 5, 2❩, ❨0, 0, 0, 5❩❩; Y ≔ ❨❨0, 0, 0, 2❩, ❨0, 0, 3, 0❩, ❨0, −2, 0, 0❩, ❨0, 0, 0, 0❩❩; Z ≔ ❨❨2, 0, 0, 0❩, ❨1, 3, 0, 0❩, ❨0, 0, −1, 0❩, ❨0, 0, 5, 2❩❩;'+' j ≔ ❨❨0, 5, −2, 3❩, ❨−5, 0, −4, −2❩, ❨2, 4, 0, −1❩, ❨−3, 2, 1, 0❩❩;', null);
  Test('k ≔ ❨❨1, 1, 1❩, ❨1, 2, 4❩, ❨1, 3, 9❩, ❨1, 4, 16❩❩; l ≔ ❨❨2, −2, 3, 5, 7❩, ❨6, 2, −2, 3, 5❩, ❨−10, 6, 2, −2, 3❩, ❨−1, −10, 6, 2, −2❩❩; m ≔ ❨❨6, 3, −5, 10, 1❩, ❨1, 6, 3, −5, 10❩, ❨10, 1, 6, 3, −5❩, ❨−5, 10, 1, 6, 3❩❩;', null);
  Test('n ≔ ❨❨7, 2, −3❩, ❨1, 5, 1❩, ❨−3, 2, 7❩❩; o ≔ ❨❨2, −5, 7, 1❩, ❨5, 3, −2, 2❩, ❨2, −2, 3, 5❩, ❨1, 7, −5, 8❩❩; p ≔ ❨❨0, 0, 0, 1❩, ❨0, 0, 1, 0❩, ❨0, 1, 0, 0❩, ❨1, 0, 0, 0❩❩; q ≔ ❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩;', null);
  Test('r ≔ ❨❨1, 1, 1, 1❩, ❨1, 1, −1, −1❩, ❨1, −1, 1, −1❩, ❨1, −1, −1, 1❩❩; s ≔ ❨❨0, 1, 3, 5❩, ❨0, 0, 2, 1❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩; a ≔ ❨❨1, 2, 0, 0❩, ❨5, 2, 3, 0❩, ❨0, 3, 8, 4❩, ❨0, 0, 2, −1❩❩;', null);
  Test('b ≔ ❨❨1, 0, 0, 0❩, ❨5, 2, 0, 0❩, ❨0, 3, 8, 0❩, ❨0, 0, 2, −1❩❩; c ≔ ❨❨5, 2, 1, 7❩, ❨2, 1, 7, 9❩, ❨1, 7, 9, 5❩, ❨7, 9, 5, 2❩❩; d ≔ ❨❨−5, 0, 0❩, ❨0, −9, 0❩, ❨0, 0, −1❩❩; f ≔ ❨❨−5, 0, 0❩, ❨0, 0, 0❩, ❨0, 0, −1❩❩;', null);
  Test('t ≔ ❨❨0, 0❩, ❨0, 0❩❩; u ≔ ❨❨0, 0, 0, 0❩, ❨0, 0, 0, 0❩❩; v ≔ ❨❨1, 0❩, ❨0, 1❩❩; w ≔ ❨❨1❩, ❨0❩❩; x ≔ ❨❨0, 0, 0, 0❩, ❨0, 2, 0, 1❩, ❨0, 0, 3, 1❩, ❨0, 0, 0, 2❩❩; y ≔ ❨❨1, 0, 3, 0, 0❩, ❨0, 1, 2, 0, 0❩, ❨0, 0, 0, 1, 0❩, ❨0, 0, 0, 0, 1❩❩; '+'Ö ≔ ❨❨1, 4, 0❩, ❨0, 0, 1❩, ❨0, 0, 0❩, ❨0, 0, 0❩❩;', null);
  Test('z ≔ ❨❨1, 0, 3, 0, 0, 2❩, ❨0, 1, 2, 0, 0, 1❩, ❨0, 0, 0, 1, 0, 1❩, ❨0, 0, 0, 0, 1, 2❩, ❨0, 0, 0, 0, 0, 0❩❩;', null);
  Test('Å ≔ ❨❨1, 0, 3, 0, 0, 2❩, ❨0, 1, 2, 1, 0, 1❩, ❨0, 0, 0, 1, 0, 1❩, ❨0, 0, 0, 0, 1, 2❩, ❨0, 0, 0, 0, 0, 0❩❩; å ≔ ❨❨7, 0❩, ❨0, 7❩❩;', null);
  Test('Ä ≔ ❨❨1, 0, 3, 0, 0, 2❩, ❨0, 1, 2, 0, 0, 1❩, ❨0, 0, 0, −1, 0, 1❩, ❨0, 0, 0, 0, 1, 2❩, ❨0, 0, 0, 0, 0, 0❩❩; ä ≔ ❨❨0, 0❩, ❨0, 0❩, ❨0, 0❩, ❨0, 0❩❩;', null);

  TestRealMatrixQ('IsMatrix', RealMatrixNames);
  TestRealMatrixQ('IsRow', 'E');
  TestRealMatrixQ('IsCol', 'Dw');
  TestRealMatrixQ('IsSquare', 'ACFghHIJMOPQRSTUVWXYZjnopqrsabcdftvxå');
  TestRealMatrixQ('IsIdentity', 'qv');
  TestRealMatrixQ('IsZeroMatrix', 'tuä');
  TestRealMatrixQ('IsDiagonal', 'FGUVqdftuvwåä');
  TestRealMatrixQ('IsAntidiagonal', 'Ypt');
  TestRealMatrixQ('IsReversal', 'p');
  TestRealMatrixQ('IsUpperTriangular', 'EFGHKMUVXqsdftuvwxyÖzÅÄåä');
  TestRealMatrixQ('IsLowerTriangular', 'DFGIUVWZqbdftuvwåä');
  TestRealMatrixQ('IsTriangular', 'EFGHKMUVXqsdftuvwDFGIUVWZqbdftuvwxyÖzÅÄåä');
  TestRealMatrixQ('IsRowEchelonForm', 'EFGHKMUVXqsdtuvwyÖzÅÄåä');
  TestRealMatrixQ('IsReducedRowEchelonForm', 'GUqtuvwyÖzä');
  TestRealMatrixQ('IsScalar', 'qtvå');
  TestRealMatrixQ('IsSymmetric', 'FOSUVpqrcdftvå');
  TestRealMatrixQ('IsSkewSymmetric', 'jt');
  TestRealMatrixQ('IsHermitian', 'FOSUVpqrcdftvå');
  TestRealMatrixQ('IsSkewHermitian', 'jt');
  TestRealMatrixQ('IsOrthogonal', 'OTVpqv');
  TestRealMatrixQ('IsUnitary', 'OTVpqv');
  TestRealMatrixQ('IsNormal', 'FOPSTUVpqrcdfjtvå');
  TestRealMatrixQ('IsBinary', 'GNOUpqtuvwä');
  TestRealMatrixQ('IsPermutation', 'Opqv');
  TestRealMatrixQ('IsCirculant', 'EPämqtuåv');
  TestRealMatrixQ('IsToeplitz', 'DEGPäXlmqtuåvw');
  TestRealMatrixQ('IsHankel', 'DEOäpcftuåvw');
  TestRealMatrixQ('IsUpperHessenberg', 'BEFGKHMOTUVWXYZÅÄÖqsabdftuwyzåävx');
  TestRealMatrixQ('IsLowerHessenberg', 'BDGFgIJMNOTUVWZÖqabdftuwåäv');
  TestRealMatrixQ('IsTridiagonal', 'FMOTUVWZqabdftåv');
  TestRealMatrixQ('IsUpperBidiagonal', 'FMUVqdftåv');
  TestRealMatrixQ('IsLowerBidiagonal', 'FUVWZqbdftåv');
  TestRealMatrixQ('IsBidiagonal', 'FMUVqdftåvWZb');
  TestRealMatrixQ('IsCentrosymmetric', 'npqtvå');
  TestRealMatrixQ('IsVandermonde', 'Qk');
  TestRealMatrixQ('IsIdempotent', 'Uqtv');
  TestRealMatrixQ('IsInvolution', 'OVpqv');
  TestRealMatrixQ('IsPositiveDefinite', 'qvå');
  TestRealMatrixQ('IsPositiveSemidefinite', 'Uqtvå');
  TestRealMatrixQ('IsNegativeDefinite', 'd');
  TestRealMatrixQ('IsNegativeSemidefinite', 'dft');
  TestRealMatrixQ('IsIndefinite', 'FOSVprc');
  TestRealMatrixQ('IsNilpotent', 'st');
  TestRealMatrixQ('IsPositive', 'EhPQRkc');
  TestRealMatrixQ('IsNonNegative', 'EghGIJNOPäQRUWkpqsctuåvwxyÖzÅ');
  TestRealMatrixQ('IsNegative', 'L');
  TestRealMatrixQ('IsNonPositive', 'KLädftu');
  TestRealMatrixQ('IsZero', 'tuä');
  TestRealMatrixQ('IsNonZero', 'ABCDEFGHIJKLMNOPQRSTUVWXYZÅÄÖabcdfghjklmnopqrsvwxyzå');
  TestRealMatrixQ('IsDiagonallyDominant', 'FTUVWnqdftåvx');
  TestRealMatrixQ('IsStrictlyDiagonallyDominant', 'FVnqdåv');

  TestRealMatrixQImplies('IsUpperTriangular', 'IsTriangular');
  TestRealMatrixQImplies('IsLowerTriangular', 'IsTriangular');

  TestRealMatrixQImplies('IsUpperTriangular', 'IsUpperHessenberg');
  TestRealMatrixQImplies('IsLowerTriangular', 'IsLowerHessenberg');

  TestRealMatrixQImplies('IsUpperBidiagonal', 'IsBidiagonal');
  TestRealMatrixQImplies('IsLowerBidiagonal', 'IsBidiagonal');

  TestRealMatrixQImplies('IsBidiagonal', 'IsTridiagonal');

  TestRealMatrixQImplies('IsRowEchelonForm', 'IsUpperTriangular');
  TestRealMatrixQImplies('IsReducedRowEchelonForm', 'IsRowEchelonForm');

  TestRealMatrixQImplies('IsIdentity', 'IsScalar');
  TestRealMatrixQImplies('IsScalar', 'IsDiagonal');

  TestRealMatrixQImplies('IsReversal', 'IsAntidiagonal');

  TestRealMatrixQImplies('IsCirculant', 'IsToeplitz');

  TestRealMatrixQImplies('IsIdentity', 'IsPermutation');
  TestRealMatrixQImplies('IsReversal', 'IsPermutation');
  TestRealMatrixQImplies('IsPermutation', 'IsBinary');

  TestRealMatrixQImplies('IsPositiveDefinite', 'IsPositiveSemidefinite');
  TestRealMatrixQImplies('IsNegativeDefinite', 'IsNegativeSemidefinite');

  TestRealMatrixQImplies('IsPositiveDefinite', 'IsSymmetric');
  TestRealMatrixQImplies('IsNegativeDefinite', 'IsSymmetric');
  TestRealMatrixQImplies('IsPositiveSemidefinite', 'IsSymmetric');
  TestRealMatrixQImplies('IsNegativeSemidefinite', 'IsSymmetric');
  TestRealMatrixQImplies('IsIndefinite', 'IsSymmetric');

  TestRealMatrixQImplies('IsSymmetric', 'IsNormal');
  TestRealMatrixQImplies('IsSkewSymmetric', 'IsNormal');
  TestRealMatrixQImplies('IsOrthogonal', 'IsNormal');

  TestRealMatrixQImplies('IsSymmetric', 'IsSquare');
  TestRealMatrixQImplies('IsSkewSymmetric', 'IsSquare');
  TestRealMatrixQImplies('IsOrthogonal', 'IsSquare');
  TestRealMatrixQImplies('IsNormal', 'IsSquare');
  TestRealMatrixQImplies('IsIdentity', 'IsSquare');
  TestRealMatrixQImplies('IsReversal', 'IsSquare');
  TestRealMatrixQImplies('IsScalar', 'IsSquare');
  TestRealMatrixQImplies('IsNilpotent', 'IsSquare');

  TestRealMatrixQImplies('IsPositive', 'IsNonNegative');
  TestRealMatrixQImplies('IsNegative', 'IsNonPositive');

  TestRealMatrixQImplies('IsStrictlyDiagonallyDominant', 'IsDiagonallyDominant');

  TestRealMatrixQImplies('IsRow', 'IsToeplitz');
  TestRealMatrixQImplies('IsRow', 'IsCirculant');
  TestRealMatrixQImplies('IsRow', 'IsHankel');
  TestRealMatrixQImplies('IsRow', 'IsUpperTriangular');

  TestRealMatrixQImplies('IsCol', 'IsToeplitz');
  TestRealMatrixQImplies('IsCol', 'IsHankel');
  TestRealMatrixQImplies('IsCol', 'IsLowerTriangular');

  Test('delete(A); delete(B); delete(C); delete(D); delete(E); delete(F); delete(g); delete(h); delete(G); delete(H); delete(I); delete(J); delete(K); delete(L); delete(M)', success);
  Test('delete(N), delete(O); delete(P); delete(Q); delete(R); delete(S); delete(T); delete(U); delete(V); delete(W); delete(X); delete(Y); delete(Z); delete(j); delete(k)', success);
  Test('delete(l); delete(m); delete(n); delete(o); delete(p); delete(q); delete(r); delete(s); delete(a); delete(b); delete(c); delete(f); delete(t); delete(u); delete(v)', success);
  Test('delete(w); delete(x); delete(y); delete(z); delete(Å); delete(Ä); delete(Ö); delete(å); delete(ä)', success);

  Test('IsIdentity(❨❨1, 0.00002❩, ❨0, 1❩❩)', False);
  Test('IsIdentity(❨❨1, 0.00002❩, ❨0, 1❩❩, 1E-6)', False);
  Test('IsIdentity(❨❨1, 0.00002❩, ❨0, 1❩❩, 1E-5)', False);
  Test('IsIdentity(❨❨1, 0.00002❩, ❨0, 1❩❩, 1E-4)', True);
  Test('IsIdentity(❨❨1, 0.00002❩, ❨0, 1❩❩, 1E-3)', True);

  Test('IsIdentity(❨❨1.002, 0.00002❩, ❨0, 1❩❩)', False);
  Test('IsIdentity(❨❨1.002, 0.00002❩, ❨0, 1❩❩, 1E-6)', False);
  Test('IsIdentity(❨❨1.002, 0.00002❩, ❨0, 1❩❩, 1E-5)', False);
  Test('IsIdentity(❨❨1.002, 0.00002❩, ❨0, 1❩❩, 1E-4)', False);
  Test('IsIdentity(❨❨1.002, 0.00002❩, ❨0, 1❩❩, 1E-3)', False);
  Test('IsIdentity(❨❨1.002, 0.00002❩, ❨0, 1❩❩, 1E-2)', True);
  Test('IsIdentity(❨❨1.002, 0.00002❩, ❨0, 1❩❩, 1E-1)', True);

  Test('IsZeroMatrix(❨❨0, 0, 0❩, ❨0, 0, 0.02❩❩)', False);
  Test('IsZeroMatrix(❨❨0, 0, 0❩, ❨0, 0, 0.02❩❩, 1E-3)', False);
  Test('IsZeroMatrix(❨❨0, 0, 0❩, ❨0, 0, 0.02❩❩, 1E-2)', False);
  Test('IsZeroMatrix(❨❨0, 0, 0❩, ❨0, 0, 0.02❩❩, 1E-1)', True);

  Test('IsDiagonal(❨❨6, 0, 0❩, ❨0, 7, 0❩, ❨0, 0.0002, 2❩❩)', False);
  Test('IsDiagonal(❨❨6, 0, 0❩, ❨0, 7, 0❩, ❨0, 0.0002, 2❩❩, 1E-5)', False);
  Test('IsDiagonal(❨❨6, 0, 0❩, ❨0, 7, 0❩, ❨0, 0.0002, 2❩❩, 1E-4)', False);
  Test('IsDiagonal(❨❨6, 0, 0❩, ❨0, 7, 0❩, ❨0, 0.0002, 2❩❩, 1E-3)', True);
  Test('IsDiagonal(❨❨6, 0, 0❩, ❨0, 7, 0❩, ❨0, 0.0002, 2❩❩, 1E-2)', True);

  Test('IsAntidiagonal(❨❨0, 0, 4❩, ❨0, 3, 0.000002❩, ❨1, 0, 0❩❩)', False);
  Test('IsAntidiagonal(❨❨0, 0, 4❩, ❨0, 3, 0.000002❩, ❨1, 0, 0❩❩, 1E-7)', False);
  Test('IsAntidiagonal(❨❨0, 0, 4❩, ❨0, 3, 0.000002❩, ❨1, 0, 0❩❩, 1E-6)', False);
  Test('IsAntidiagonal(❨❨0, 0, 4❩, ❨0, 3, 0.000002❩, ❨1, 0, 0❩❩, 1E-5)', True);
  Test('IsAntidiagonal(❨❨0, 0, 4❩, ❨0, 3, 0.000002❩, ❨1, 0, 0❩❩, 1E-4)', True);

  Test('IsReversal(❨❨0, 0, 1❩, ❨0, 1, 0❩, ❨1.0002, 0, 0❩❩)', False);
  Test('IsReversal(❨❨0, 0, 1❩, ❨0, 1, 0❩, ❨1.0002, 0, 0❩❩, 1E-5)', False);
  Test('IsReversal(❨❨0, 0, 1❩, ❨0, 1, 0❩, ❨1.0002, 0, 0❩❩, 1E-4)', False);
  Test('IsReversal(❨❨0, 0, 1❩, ❨0, 1, 0❩, ❨1.0002, 0, 0❩❩, 1E-3)', True);
  Test('IsReversal(❨❨0, 0, 1❩, ❨0, 1, 0❩, ❨1.0002, 0, 0❩❩, 1E-2)', True);

  Test('IsUpperTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩)', False);
  Test('IsUpperTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩, 1E-5)', False);
  Test('IsUpperTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩, 1E-4)', False);
  Test('IsUpperTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩, 1E-3)', True);
  Test('IsUpperTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩, 1E-2)', True);

  Test('IsLowerTriangular(❨❨1, 0, 0❩, ❨1, 6, 0.02❩, ❨2, 6, 2❩❩)', False);
  Test('IsLowerTriangular(❨❨1, 0, 0❩, ❨1, 6, 0.02❩, ❨2, 6, 2❩❩, 1E-3)', False);
  Test('IsLowerTriangular(❨❨1, 0, 0❩, ❨1, 6, 0.02❩, ❨2, 6, 2❩❩, 1E-2)', False);
  Test('IsLowerTriangular(❨❨1, 0, 0❩, ❨1, 6, 0.02❩, ❨2, 6, 2❩❩, 1E-1)', True);

  Test('IsTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩)', False);
  Test('IsTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩, 1E-5)', False);
  Test('IsTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩, 1E-4)', False);
  Test('IsTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩, 1E-3)', True);
  Test('IsTriangular(❨❨3, 5, 1, 2❩, ❨0, 4, 1, 6❩, ❨0, 0, 1, 1❩, ❨0, 0.0002, 0, 7❩❩, 1E-2)', True);

  Test('IsTriangular(❨❨1, 0, 0❩, ❨1, 6, 0.02❩, ❨2, 6, 2❩❩)', False);
  Test('IsTriangular(❨❨1, 0, 0❩, ❨1, 6, 0.02❩, ❨2, 6, 2❩❩, 1E-3)', False);
  Test('IsTriangular(❨❨1, 0, 0❩, ❨1, 6, 0.02❩, ❨2, 6, 2❩❩, 1E-2)', False);
  Test('IsTriangular(❨❨1, 0, 0❩, ❨1, 6, 0.02❩, ❨2, 6, 2❩❩, 1E-1)', True);

  Test('IsRowEchelonForm(❨❨5, 2, 3, 1❩, ❨0.002, 0, 1, 7❩, ❨0, 0, 0, 2❩❩)', False);
  Test('IsRowEchelonForm(❨❨5, 2, 3, 1❩, ❨0.002, 0, 1, 7❩, ❨0, 0, 0, 2❩❩, 1E-4)', False);
  Test('IsRowEchelonForm(❨❨5, 2, 3, 1❩, ❨0.002, 0, 1, 7❩, ❨0, 0, 0, 2❩❩, 1E-3)', False);
  Test('IsRowEchelonForm(❨❨5, 2, 3, 1❩, ❨0.002, 0, 1, 7❩, ❨0, 0, 0, 2❩❩, 1E-2)', True);
  Test('IsRowEchelonForm(❨❨5, 2, 3, 1❩, ❨0.002, 0, 1, 7❩, ❨0, 0, 0, 2❩❩, 1E-1)', True);

  Test('IsReducedRowEchelonForm(❨❨1, 3, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩)', False);
  Test('IsReducedRowEchelonForm(❨❨1, 3, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩, 1E-4)', False);
  Test('IsReducedRowEchelonForm(❨❨1, 3, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩, 1E-3)', False);
  Test('IsReducedRowEchelonForm(❨❨1, 3, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩, 1E-2)', True);
  Test('IsReducedRowEchelonForm(❨❨1, 3, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩, 1E-1)', True);

  Test('IsReducedRowEchelonForm(❨❨1.02, 0, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩)', False);
  Test('IsReducedRowEchelonForm(❨❨1.02, 0, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩, 1E-4)', False);
  Test('IsReducedRowEchelonForm(❨❨1.02, 0, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩, 1E-3)', False);
  Test('IsReducedRowEchelonForm(❨❨1.02, 0, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩, 1E-2)', False);
  Test('IsReducedRowEchelonForm(❨❨1.02, 0, 0.002, 0❩, ❨0, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩, 1E-1)', True);

  Test('IsScalar(❨❨2.55, 0, 0❩, ❨0, 2.56, 0❩, ❨0, 0, 2.55❩❩)', False);
  Test('IsScalar(❨❨2.55, 0, 0❩, ❨0, 2.56, 0❩, ❨0, 0, 2.55❩❩, 1E-3)', False);
  Test('IsScalar(❨❨2.55, 0, 0❩, ❨0, 2.56, 0❩, ❨0, 0, 2.55❩❩, 1E-2)', True);
  Test('IsScalar(❨❨2.55, 0, 0❩, ❨0, 2.56, 0❩, ❨0, 0, 2.55❩❩, 1E-1)', True);

  Test('IsScalar(❨❨2.55, 0, 0❩, ❨0, 2.55, 0.02❩, ❨0, 0, 2.55❩❩)', False);
  Test('IsScalar(❨❨2.55, 0, 0❩, ❨0, 2.55, 0.02❩, ❨0, 0, 2.55❩❩, 1E-3)', False);
  Test('IsScalar(❨❨2.55, 0, 0❩, ❨0, 2.55, 0.02❩, ❨0, 0, 2.55❩❩, 1E-2)', False);
  Test('IsScalar(❨❨2.55, 0, 0❩, ❨0, 2.55, 0.02❩, ❨0, 0, 2.55❩❩, 1E-1)', True);

  Test('IsSymmetric(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩)', False);
  Test('IsSymmetric(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩, 1E-3)', False);
  Test('IsSymmetric(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩, 1E-2)', True);
  Test('IsSymmetric(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩, 1E-1)', True);

  Test('IsSkewSymmetric(❨❨0, 2.65, −4.32, 0.47❩, ❨−2.64, 0, −1.16, −7.02❩, ❨4.31, 1.16, 0, 0.43❩, ❨−0.48, 7.02, −0.43, 0❩❩)', False);
  Test('IsSkewSymmetric(❨❨0, 2.65, −4.32, 0.47❩, ❨−2.64, 0, −1.16, −7.02❩, ❨4.31, 1.16, 0, 0.43❩, ❨−0.48, 7.02, −0.43, 0❩❩, 1E-3)', False);
  Test('IsSkewSymmetric(❨❨0, 2.65, −4.32, 0.47❩, ❨−2.64, 0, −1.16, −7.02❩, ❨4.31, 1.16, 0, 0.43❩, ❨−0.48, 7.02, −0.43, 0❩❩, 1E-2)', True);
  Test('IsSkewSymmetric(❨❨0, 2.65, −4.32, 0.47❩, ❨−2.64, 0, −1.16, −7.02❩, ❨4.31, 1.16, 0, 0.43❩, ❨−0.48, 7.02, −0.43, 0❩❩, 1E-1)', True);

  Test('IsHermitian(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩)', False);
  Test('IsHermitian(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩, 1E-3)', False);
  Test('IsHermitian(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩, 1E-2)', True);
  Test('IsHermitian(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩, 1E-1)', True);

  Test('IsSkewHermitian(❨❨0, 2.65, −4.32, 0.47❩, ❨−2.64, 0, −1.16, −7.02❩, ❨4.31, 1.16, 0, 0.43❩, ❨−0.48, 7.02, −0.43, 0❩❩)', False);
  Test('IsSkewHermitian(❨❨0, 2.65, −4.32, 0.47❩, ❨−2.64, 0, −1.16, −7.02❩, ❨4.31, 1.16, 0, 0.43❩, ❨−0.48, 7.02, −0.43, 0❩❩, 1E-3)', False);
  Test('IsSkewHermitian(❨❨0, 2.65, −4.32, 0.47❩, ❨−2.64, 0, −1.16, −7.02❩, ❨4.31, 1.16, 0, 0.43❩, ❨−0.48, 7.02, −0.43, 0❩❩, 1E-2)', True);
  Test('IsSkewHermitian(❨❨0, 2.65, −4.32, 0.47❩, ❨−2.64, 0, −1.16, −7.02❩, ❨4.31, 1.16, 0, 0.43❩, ❨−0.48, 7.02, −0.43, 0❩❩, 1E-1)', True);

  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩)', False);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩, 1E-5)', False);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩, 1E-4)', False);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩, 1E-3)', True);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩, 1E-2)', True);

  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩)', False);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩, 1E-5)', False);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩, 1E-4)', False);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩, 1E-3)', True);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.8775, −0.4794❩, ❨0, 0.4794, 0.8775❩❩, 1E-2)', True);

  Test('IsNormal(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩)', False);
  Test('IsNormal(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩, 1E-3)', False);
  Test('IsNormal(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩, 1E-2)', True);
  Test('IsNormal(❨❨2.44, 7.70, 9.13, 12.98❩, ❨7.71, 10.24, 10.62, 9.48❩, ❨9.14, 10.62, 2.24, 7.98❩, ❨12.98, 9.48, 7.99, 10.42❩❩, 1E-1)', True);

  Test('IsBinary(❨❨0.0001, 1.0019, 0.000001, 0.999999❩, ❨1.0002, 0.99998, −0.0003, 0.0001❩, ❨−0.00045, 0.9995, 0.0004, 1.0005❩❩)', False);
  Test('IsBinary(❨❨0.0001, 1.0019, 0.000001, 0.999999❩, ❨1.0002, 0.99998, −0.0003, 0.0001❩, ❨−0.00045, 0.9995, 0.0004, 1.0005❩❩, 1E-5)', False);
  Test('IsBinary(❨❨0.0001, 1.0019, 0.000001, 0.999999❩, ❨1.0002, 0.99998, −0.0003, 0.0001❩, ❨−0.00045, 0.9995, 0.0004, 1.0005❩❩, 1E-4)', False);
  Test('IsBinary(❨❨0.0001, 1.0019, 0.000001, 0.999999❩, ❨1.0002, 0.99998, −0.0003, 0.0001❩, ❨−0.00045, 0.9995, 0.0004, 1.0005❩❩, 1E-3)', False);
  Test('IsBinary(❨❨0.0001, 1.0019, 0.000001, 0.999999❩, ❨1.0002, 0.99998, −0.0003, 0.0001❩, ❨−0.00045, 0.9995, 0.0004, 1.0005❩❩, 1E-2)', True);
  Test('IsBinary(❨❨0.0001, 1.0019, 0.000001, 0.999999❩, ❨1.0002, 0.99998, −0.0003, 0.0001❩, ❨−0.00045, 0.9995, 0.0004, 1.0005❩❩, 1E-1)', True);

  Test('IsPermutation(❨❨0, 0, 1.001, 0❩, ❨0, 1, 0, 0❩, ❨0, 0, 0, 1❩, ❨1, 0.002, 0, 0❩❩)', False);
  Test('IsPermutation(❨❨0, 0, 1.001, 0❩, ❨0, 1, 0, 0❩, ❨0, 0, 0, 1❩, ❨1, 0.002, 0, 0❩❩, 1E-3)', False);
  Test('IsPermutation(❨❨0, 0, 1.001, 0❩, ❨0, 1, 0, 0❩, ❨0, 0, 0, 1❩, ❨1, 0.002, 0, 0❩❩, 1E-2)', True);
  Test('IsPermutation(❨❨0, 0, 1.001, 0❩, ❨0, 1, 0, 0❩, ❨0, 0, 0, 1❩, ❨1, 0.002, 0, 0❩❩, 1E-1)', True);

  Test('IsCirculant(❨❨6, 2, 3, 7❩, ❨7, 6, 2, 3❩, ❨3, 7, 6, 2❩, ❨2, 3.002, 7, 6❩❩)', False);
  Test('IsCirculant(❨❨6, 2, 3, 7❩, ❨7, 6, 2, 3❩, ❨3, 7, 6, 2❩, ❨2, 3.002, 7, 6❩❩, 1E-4)', False);
  Test('IsCirculant(❨❨6, 2, 3, 7❩, ❨7, 6, 2, 3❩, ❨3, 7, 6, 2❩, ❨2, 3.002, 7, 6❩❩, 1E-3)', False);
  Test('IsCirculant(❨❨6, 2, 3, 7❩, ❨7, 6, 2, 3❩, ❨3, 7, 6, 2❩, ❨2, 3.002, 7, 6❩❩, 1E-2)', True);
  Test('IsCirculant(❨❨6, 2, 3, 7❩, ❨7, 6, 2, 3❩, ❨3, 7, 6, 2❩, ❨2, 3.002, 7, 6❩❩, 1E-1)', True);

  Test('IsToeplitz(❨❨4, 1, 2, 5❩, ❨8, 4, 1.002, 2❩, ❨6, 8, 4, 1❩, ❨9, 6, 8, 4❩❩)', False);
  Test('IsToeplitz(❨❨4, 1, 2, 5❩, ❨8, 4, 1.002, 2❩, ❨6, 8, 4, 1❩, ❨9, 6, 8, 4❩❩, 1E-4)', False);
  Test('IsToeplitz(❨❨4, 1, 2, 5❩, ❨8, 4, 1.002, 2❩, ❨6, 8, 4, 1❩, ❨9, 6, 8, 4❩❩, 1E-3)', False);
  Test('IsToeplitz(❨❨4, 1, 2, 5❩, ❨8, 4, 1.002, 2❩, ❨6, 8, 4, 1❩, ❨9, 6, 8, 4❩❩, 1E-2)', True);
  Test('IsToeplitz(❨❨4, 1, 2, 5❩, ❨8, 4, 1.002, 2❩, ❨6, 8, 4, 1❩, ❨9, 6, 8, 4❩❩, 1E-1)', True);

  Test('IsHankel(❨❨4, 2, 1.002, 3❩, ❨2.003, 1, 3, 7❩, ❨1, 3, 6.999, 5❩, ❨3, 7, 5, 0❩❩)', False);
  Test('IsHankel(❨❨4, 2, 1.002, 3❩, ❨2.003, 1, 3, 7❩, ❨1, 3, 6.999, 5❩, ❨3, 7, 5, 0❩❩, 1E-4)', False);
  Test('IsHankel(❨❨4, 2, 1.002, 3❩, ❨2.003, 1, 3, 7❩, ❨1, 3, 6.999, 5❩, ❨3, 7, 5, 0❩❩, 1E-3)', False);
  Test('IsHankel(❨❨4, 2, 1.002, 3❩, ❨2.003, 1, 3, 7❩, ❨1, 3, 6.999, 5❩, ❨3, 7, 5, 0❩❩, 1E-2)', True);
  Test('IsHankel(❨❨4, 2, 1.002, 3❩, ❨2.003, 1, 3, 7❩, ❨1, 3, 6.999, 5❩, ❨3, 7, 5, 0❩❩, 1E-1)', True);

  Test('IsUpperHessenberg(❨❨1, 5, 2, 3, 6❩, ❨2, 5, 7, 3, 6❩, ❨−0.001, 1, 4, 5, 6❩, ❨0.003, 0.002, 2, 1, 2❩, ❨0.001, 0.002, −0.001, 6, 2❩❩)', False);
  Test('IsUpperHessenberg(❨❨1, 5, 2, 3, 6❩, ❨2, 5, 7, 3, 6❩, ❨−0.001, 1, 4, 5, 6❩, ❨0.003, 0.002, 2, 1, 2❩, ❨0.001, 0.002, −0.001, 6, 2❩❩, 1E-4)', False);
  Test('IsUpperHessenberg(❨❨1, 5, 2, 3, 6❩, ❨2, 5, 7, 3, 6❩, ❨−0.001, 1, 4, 5, 6❩, ❨0.003, 0.002, 2, 1, 2❩, ❨0.001, 0.002, −0.001, 6, 2❩❩, 1E-3)', False);
  Test('IsUpperHessenberg(❨❨1, 5, 2, 3, 6❩, ❨2, 5, 7, 3, 6❩, ❨−0.001, 1, 4, 5, 6❩, ❨0.003, 0.002, 2, 1, 2❩, ❨0.001, 0.002, −0.001, 6, 2❩❩, 1E-2)', True);
  Test('IsUpperHessenberg(❨❨1, 5, 2, 3, 6❩, ❨2, 5, 7, 3, 6❩, ❨−0.001, 1, 4, 5, 6❩, ❨0.003, 0.002, 2, 1, 2❩, ❨0.001, 0.002, −0.001, 6, 2❩❩, 1E-1)', True);

  Test('IsLowerHessenberg(❨❨1, 4, 0.003, -0.001❩, ❨5, 2, 3, 0.002❩, ❨4, 6, 2, 2❩, ❨6, 3, 1, 5❩❩)', False);
  Test('IsLowerHessenberg(❨❨1, 4, 0.003, -0.001❩, ❨5, 2, 3, 0.002❩, ❨4, 6, 2, 2❩, ❨6, 3, 1, 5❩❩, 1E-4)', False);
  Test('IsLowerHessenberg(❨❨1, 4, 0.003, -0.001❩, ❨5, 2, 3, 0.002❩, ❨4, 6, 2, 2❩, ❨6, 3, 1, 5❩❩, 1E-3)', False);
  Test('IsLowerHessenberg(❨❨1, 4, 0.003, -0.001❩, ❨5, 2, 3, 0.002❩, ❨4, 6, 2, 2❩, ❨6, 3, 1, 5❩❩, 1E-2)', True);
  Test('IsLowerHessenberg(❨❨1, 4, 0.003, -0.001❩, ❨5, 2, 3, 0.002❩, ❨4, 6, 2, 2❩, ❨6, 3, 1, 5❩❩, 1E-1)', True);

  Test('IsTridiagonal(❨❨1, 3, 0.02, 0.01❩, ❨5, 2, 1, 0.01❩, ❨0.03, 1, 3, 2❩, ❨−0.01, 0.02, 1, 2❩❩)', False);
  Test('IsTridiagonal(❨❨1, 3, 0.02, 0.01❩, ❨5, 2, 1, 0.01❩, ❨0.03, 1, 3, 2❩, ❨−0.01, 0.02, 1, 2❩❩, 1E-3)', False);
  Test('IsTridiagonal(❨❨1, 3, 0.02, 0.01❩, ❨5, 2, 1, 0.01❩, ❨0.03, 1, 3, 2❩, ❨−0.01, 0.02, 1, 2❩❩, 1E-2)', False);
  Test('IsTridiagonal(❨❨1, 3, 0.02, 0.01❩, ❨5, 2, 1, 0.01❩, ❨0.03, 1, 3, 2❩, ❨−0.01, 0.02, 1, 2❩❩, 1E-1)', True);

  Test('IsUpperBidiagonal(❨❨1, 3, 0.02, 0.01❩, ❨−0.01, 2, 1, 0.01❩, ❨0.03, 0.01, 3, 2❩, ❨−0.01, 0.02, 0.01, 2❩❩)', False);
  Test('IsUpperBidiagonal(❨❨1, 3, 0.02, 0.01❩, ❨−0.01, 2, 1, 0.01❩, ❨0.03, 0.01, 3, 2❩, ❨−0.01, 0.02, 0.01, 2❩❩, 1E-3)', False);
  Test('IsUpperBidiagonal(❨❨1, 3, 0.02, 0.01❩, ❨−0.01, 2, 1, 0.01❩, ❨0.03, 0.01, 3, 2❩, ❨−0.01, 0.02, 0.01, 2❩❩, 1E-2)', False);
  Test('IsUpperBidiagonal(❨❨1, 3, 0.02, 0.01❩, ❨−0.01, 2, 1, 0.01❩, ❨0.03, 0.01, 3, 2❩, ❨−0.01, 0.02, 0.01, 2❩❩, 1E-1)', True);

  Test('IsLowerBidiagonal(❨❨1, 0.01, 0.02, 0.01❩, ❨5, 2, 0.02, 0.01❩, ❨0.03, 1, 3, −0.03❩, ❨−0.01, 0.02, 1, 2❩❩)', False);
  Test('IsLowerBidiagonal(❨❨1, 0.01, 0.02, 0.01❩, ❨5, 2, 0.02, 0.01❩, ❨0.03, 1, 3, −0.03❩, ❨−0.01, 0.02, 1, 2❩❩, 1E-3)', False);
  Test('IsLowerBidiagonal(❨❨1, 0.01, 0.02, 0.01❩, ❨5, 2, 0.02, 0.01❩, ❨0.03, 1, 3, −0.03❩, ❨−0.01, 0.02, 1, 2❩❩, 1E-2)', False);
  Test('IsLowerBidiagonal(❨❨1, 0.01, 0.02, 0.01❩, ❨5, 2, 0.02, 0.01❩, ❨0.03, 1, 3, −0.03❩, ❨−0.01, 0.02, 1, 2❩❩, 1E-1)', True);

  Test('IsBidiagonal(❨❨1, 0.01, 0.02, 0.01❩, ❨5, 2, 0.02, 0.01❩, ❨0.03, 1, 3, −0.03❩, ❨−0.01, 0.02, 1, 2❩❩)', False);
  Test('IsBidiagonal(❨❨1, 0.01, 0.02, 0.01❩, ❨5, 2, 0.02, 0.01❩, ❨0.03, 1, 3, −0.03❩, ❨−0.01, 0.02, 1, 2❩❩, 1E-3)', False);
  Test('IsBidiagonal(❨❨1, 0.01, 0.02, 0.01❩, ❨5, 2, 0.02, 0.01❩, ❨0.03, 1, 3, −0.03❩, ❨−0.01, 0.02, 1, 2❩❩, 1E-2)', False);
  Test('IsBidiagonal(❨❨1, 0.01, 0.02, 0.01❩, ❨5, 2, 0.02, 0.01❩, ❨0.03, 1, 3, −0.03❩, ❨−0.01, 0.02, 1, 2❩❩, 1E-1)', True);

  Test('IsBidiagonal(❨❨1, 3, 0.02, 0.01❩, ❨−0.01, 2, 1, 0.01❩, ❨0.03, 0.01, 3, 2❩, ❨−0.01, 0.02, 0.01, 2❩❩)', False);
  Test('IsBidiagonal(❨❨1, 3, 0.02, 0.01❩, ❨−0.01, 2, 1, 0.01❩, ❨0.03, 0.01, 3, 2❩, ❨−0.01, 0.02, 0.01, 2❩❩, 1E-3)', False);
  Test('IsBidiagonal(❨❨1, 3, 0.02, 0.01❩, ❨−0.01, 2, 1, 0.01❩, ❨0.03, 0.01, 3, 2❩, ❨−0.01, 0.02, 0.01, 2❩❩, 1E-2)', False);
  Test('IsBidiagonal(❨❨1, 3, 0.02, 0.01❩, ❨−0.01, 2, 1, 0.01❩, ❨0.03, 0.01, 3, 2❩, ❨−0.01, 0.02, 0.01, 2❩❩, 1E-1)', True);

  Test('IsCentrosymmetric(❨❨2, 4, 7❩, ❨5, 1, 5.002❩, ❨7, 4, 2❩❩)', False);
  Test('IsCentrosymmetric(❨❨2, 4, 7❩, ❨5, 1, 5.002❩, ❨7, 4, 2❩❩, 1E-4)', False);
  Test('IsCentrosymmetric(❨❨2, 4, 7❩, ❨5, 1, 5.002❩, ❨7, 4, 2❩❩, 1E-3)', False);
  Test('IsCentrosymmetric(❨❨2, 4, 7❩, ❨5, 1, 5.002❩, ❨7, 4, 2❩❩, 1E-2)', True);
  Test('IsCentrosymmetric(❨❨2, 4, 7❩, ❨5, 1, 5.002❩, ❨7, 4, 2❩❩, 1E-1)', True);

  Test('IsVandermonde(❨❨1, 2, 4, 8❩, ❨1.001, −1, 1, −1❩, ❨1, 3.002, 9, 27❩, ❨1, 5, 24.998, 125❩, ❨1, 1, 1, 1.002❩❩)', False);
  Test('IsVandermonde(❨❨1, 2, 4, 8❩, ❨1.001, −1, 1, −1❩, ❨1, 3.002, 9, 27❩, ❨1, 5, 24.998, 125❩, ❨1, 1, 1, 1.002❩❩, 1E-3)', False);
  Test('IsVandermonde(❨❨1, 2, 4, 8❩, ❨1.001, −1, 1, −1❩, ❨1, 3.002, 9, 27❩, ❨1, 5, 24.998, 125❩, ❨1, 1, 1, 1.002❩❩, 1E-2)', False);
  Test('IsVandermonde(❨❨1, 2, 4, 8❩, ❨1.001, −1, 1, −1❩, ❨1, 3.002, 9, 27❩, ❨1, 5, 24.998, 125❩, ❨1, 1, 1, 1.002❩❩, 1E-1)', True);

  Test('IsIdempotent(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, 0❩❩)', False);
  Test('IsIdempotent(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, 0❩❩, 1E-4)', False);
  Test('IsIdempotent(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, 0❩❩, 1E-3)', False);
  Test('IsIdempotent(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, 0❩❩, 1E-2)', True);
  Test('IsIdempotent(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, 0❩❩, 1E-1)', True);

  Test('IsInvolution(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, −1.001❩❩)', False);
  Test('IsInvolution(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, −1.001❩❩, 1E-4)', False);
  Test('IsInvolution(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, −1.001❩❩, 1E-3)', False);
  Test('IsInvolution(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, −1.001❩❩, 1E-2)', True);
  Test('IsInvolution(❨❨1, 0, 0.002, 0❩, ❨0, 1, 0, 0❩, ❨−0.002, 0, 1, 0.003❩, ❨−0.002, 0, 0.003, −1.001❩❩, 1E-1)', True);

  Test('IsPositive(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, 1❩❩)', True);
  Test('IsPositive(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, 1❩❩, 1E-6)', True);
  Test('IsPositive(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, 1❩❩, 1E-5)', True);
  Test('IsPositive(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, 1❩❩, 1E-4)', True);
  Test('IsPositive(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, 1❩❩, 1E-3)', False);
  Test('IsPositive(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, 1❩❩, 1E-2)', False);
  Test('IsPositive(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, 1❩❩, 1E-1)', False);

  Test('IsNonNegative(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, −0.0002❩❩)', False);
  Test('IsNonNegative(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, −0.0002❩❩, 1E-6)', False);
  Test('IsNonNegative(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, −0.0002❩❩, 1E-5)', False);
  Test('IsNonNegative(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, −0.0002❩❩, 1E-4)', False);
  Test('IsNonNegative(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, −0.0002❩❩, 1E-3)', True);
  Test('IsNonNegative(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, −0.0002❩❩, 1E-2)', True);
  Test('IsNonNegative(❨❨3, 1, 5, 2❩, ❨0.0002, 2, 3, −0.0002❩❩, 1E-1)', True);

  Test('IsZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩)', False);
  Test('IsZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-6)', False);
  Test('IsZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-5)', False);
  Test('IsZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-4)', False);
  Test('IsZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-3)', True);
  Test('IsZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-2)', True);
  Test('IsZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-1)', True);

  Test('IsNonZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩)', True);
  Test('IsNonZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-6)', True);
  Test('IsNonZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-5)', True);
  Test('IsNonZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-4)', True);
  Test('IsNonZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-3)', False);
  Test('IsNonZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-2)', False);
  Test('IsNonZero(❨❨0.0002, 0.0001, −0.0003, 0.0000052❩, ❨−0.00013, 0.00042, 0.00021, 0.000051❩❩, 1E-1)', False);

  Test('IsNegative(❨❨−4, −5, −2, −1❩, ❨−0.003, −4, −2, −1❩❩)', True);
  Test('IsNegative(❨❨−4, −5, −2, −1❩, ❨−0.003, −4, −2, −1❩❩, 1E-5)', True);
  Test('IsNegative(❨❨−4, −5, −2, −1❩, ❨−0.003, −4, −2, −1❩❩, 1E-4)', True);
  Test('IsNegative(❨❨−4, −5, −2, −1❩, ❨−0.003, −4, −2, −1❩❩, 1E-3)', True);
  Test('IsNegative(❨❨−4, −5, −2, −1❩, ❨−0.003, −4, −2, −1❩❩, 1E-2)', False);
  Test('IsNegative(❨❨−4, −5, −2, −1❩, ❨−0.003, −4, −2, −1❩❩, 1E-1)', False);

  Test('IsNonPositive(❨❨−4, −5, −2, −1❩, ❨0.003, −4, −2, −1❩❩)', False);
  Test('IsNonPositive(❨❨−4, −5, −2, −1❩, ❨0.003, −4, −2, −1❩❩, 1E-5)', False);
  Test('IsNonPositive(❨❨−4, −5, −2, −1❩, ❨0.003, −4, −2, −1❩❩, 1E-4)', False);
  Test('IsNonPositive(❨❨−4, −5, −2, −1❩, ❨0.003, −4, −2, −1❩❩, 1E-3)', False);
  Test('IsNonPositive(❨❨−4, −5, −2, −1❩, ❨0.003, −4, −2, −1❩❩, 1E-2)', True);
  Test('IsNonPositive(❨❨−4, −5, −2, −1❩, ❨0.003, −4, −2, −1❩❩, 1E-1)', True);


  //
  // Special complex matrices
  //

  Test('A ≔ ❨❨2, i, 1❩, ❨3, −1, −i❩, ❨0, 1, 2❩❩; B ≔ ❨❨1, 0, i❩, ❨0, 1, −2❩❩; C ≔ ❨❨1, 2❩, ❨i, −i❩, ❨0, 1❩❩; D ≔ ❨❨1❩, ❨0❩, ❨i❩, ❨1❩❩; E ≔ ❨❨2, 0, i❩❩;', null);
  Test('F ≔ ComplexMatrix(❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩); G ≔ ❨❨2, 5, 2❩, ❨0, i, 2❩, ❨0, 0, 5❩❩; H ≔ ❨❨i, 0, 0❩, ❨3, i, 0❩, ❨7, 4, −1❩❩; I ≔ ❨❨1, 0, 0❩, ❨0, 0, −i❩, ❨0, i, 0❩❩; J ≔ ComplexMatrix(❨❨1, 0, 0❩, ❨0, 0, 1❩, ❨0, 1, 0❩❩);', null);
  Test('K ≔ ❨❨2, 1, 0, 0❩, ❨0, 3, i, 0❩, ❨0, 0, 1, 3❩, ❨0, 0, 0, i❩❩; L ≔ ❨❨i, 0, 0, 0❩, ❨1, −i, 0, 0❩, ❨0, 1, 1, 0❩, ❨0, 0, 3, −1❩❩; M ≔ ❨❨1, 0, 0, 0❩, ❨1, 2, 3, 0❩, ❨0, 0, 6, 1❩, ❨0, 0, 5, i❩❩; N ≔ ❨❨2, 2, 0, 0❩, ❨3, 6, 1, 0❩, ❨0, 5, 4, 1❩, ❨1, 2, i, 2❩❩;', null);
  Test('O ≔ ❨❨2, 2, 3, 2❩, ❨4, 1, 1, 1❩, ❨0, 6, i, 2❩, ❨0, 0, 0, 1❩❩; P ≔ ❨❨1, 0, 2, i❩, ❨0, 1, 2, 3❩, ❨2, 2, 5, i❩, ❨i, 3, i, 3❩❩; Q ≔ ❨❨2, i, 1, 3❩, ❨−i, 5, 0, i❩, ❨1, 0, 9, 1❩, ❨3, −i, 1, 2❩❩; '+'R ≔ ComplexMatrix(❨❨0, 0, 0, 1❩, ❨0, 0, 1, 0❩, ❨0, 1, 0, 0❩, ❨1, 0, 0, 0❩❩);', null);
  Test('S ≔ ❨❨2, 1, 0, 0❩, ❨0, 6, 5, i❩, ❨0, 0, 3, 3❩, ❨0, 0, 0, 2❩❩; T ≔ ❨❨3, 2, 0, 1❩, ❨0, 0, i, 5❩, ❨0, 0, 0, 5❩, ❨0, 0, 0, 0❩❩; U ≔ ❨❨3, 2, 0, 1❩, ❨0, 0, 0, 0❩, ❨0, 0, i, 5❩, ❨0, 0, 0, 5❩❩; V ≔ ❨❨5, 2, 1, 2❩, ❨0, 6, i, −i❩, ❨0, 0, 0, 1❩❩;', null);
  Test('W ≔ ❨❨1, 0, i, 0❩, ❨0, 1, 4, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩; X ≔ ❨❨1, 1, i, 0❩, ❨0, 1, 4, 0❩, ❨0, 0, 0, 1❩, ❨0, 0, 0, 0❩❩; Y ≔ ❨❨1, 0, i, 0❩, ❨0, 1, 4, 0❩, ❨0, 0, 0, i❩, ❨0, 0, 0, 0❩❩; Z ≔ ❨❨1, 0, 5, 0❩, ❨0, 1, i, 0❩, ❨0, 0, 0, 1❩❩;', null);
  Test('a ≔ ❨❨1, 0, 5, 0❩, ❨0, 1, i, i❩, ❨0, 0, 0, 1❩❩; b ≔ ❨❨1, 0, 5, 0❩, ❨0, 1, i, 0❩, ❨0, 0, 0, i❩❩; c ≔ ComplexMatrix(❨❨1, 0, 0❩, ❨0, 1/√2, −1/√2❩, ❨0, 1/√2, 1/√2❩❩); d ≔ ❨❨1, i, 5, 2❩, ❨i, 5, 2, 7❩, ❨5, 2, 7, 1❩, ❨2, 7, 1, 0❩❩;', null);
  Test('f ≔ ❨❨2, 1, i, 3❩, ❨1, i, 3, 2❩, ❨i, 3, 2, 1❩, ❨3, 2, 1, i❩❩; g ≔ ❨❨5, 1, 2, i❩, ❨8, 5, 1, 2❩, ❨3, 8, 5, 1❩, ❨7, 3, 8, 5❩❩; h ≔ ❨❨i, 2, 5, 9❩, ❨9, i, 2, 5❩, ❨5, 9, i, 2❩, ❨2, 5, 9, i❩❩; j ≔ ❨❨0, 2, 5, i❩, ❨−2, 0, 1, −1❩, ❨−5, −1, 0, 7❩, ❨−i, 1, −7, 0❩❩;', null);
  Test('k ≔ ❨❨0, 3, 1, i❩, ❨−3, 0, −i, 5❩, ❨−1, −i, 0, −2❩, ❨i, −5, 2, 0❩❩; l ≔ ❨❨1, 0, 0❩, ❨0, i/√2, −1/√2❩, ❨0, i/√2, 1/√2❩❩; m ≔ ❨❨1, 0, 0❩, ❨0, i, √2❩, ❨0, −√2, i❩❩; n ≔ ❨❨i, 0, 0❩, ❨0, i, 0❩, ❨0, 0, i❩❩;', null);
  Test('p ≔ ComplexMatrix(❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 0❩❩); q ≔ ComplexMatrix(❨❨−1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩); r ≔ ❨❨1, 2, 3❩, ❨4, i, 4❩, ❨3, 2, 1❩❩; s ≔ ComplexMatrix(❨❨0, 1, 0❩, ❨0, 0, 2❩, ❨0, 0, 0❩❩); t ≔ ComplexMatrix(❨❨3, 0, 5❩, ❨0, 1, 2❩, ❨5, 2, 7❩❩);', null);
  Test('u ≔ ❨❨−3, i, 1❩, ❨−i, −1, 1❩, ❨1, 1, −3❩❩; v ≔ ❨❨−1, 0, i❩, ❨0, −1, −1❩, ❨−i, −1, 1❩❩; w ≔ ❨❨1, 0, i❩, ❨0, 1, −1❩, ❨−i, −1, 2❩❩; x ≔ ❨❨−2, i, 1❩, ❨−i, −1, 1❩, ❨1, 1, −3❩❩; ë ≔ ❨❨0, 0❩❩;', null);
  Test('y ≔ ❨❨1, 3, i❩, ❨3, −2, 1❩, ❨−i, 1, 3❩❩; z ≔ ❨❨2, 0, i❩, ❨0, 1, −1❩, ❨−i, −1, 2❩❩; o ≔ ❨❨1, 2, 4, 3❩, ❨0, 5, i, 7❩, ❨7, i, 5, 0❩, ❨3, 4, 2, 1❩❩; Å ≔ ❨❨1, 2, 4, 8❩, ❨1, i, −1, −i❩, ❨1, 5, 25, 125❩, ❨1, −1, 1, −1❩❩; ü ≔ ComplexMatrix(❨❨0❩, ❨0❩, ❨0❩, ❨0❩❩);', null);
  Test('Ä ≔ ComplexMatrix(❨❨2, 5, 7❩, ❨6, 2, 5❩, ❨8, 4, 3❩❩); Ö ≔ ComplexMatrix(❨❨−2, −3, −1❩, ❨−4, −5, −1❩, ❨−1, −1, −5❩❩); å ≔ ComplexMatrix(❨❨5, 1, 2❩, ❨1, −6, 1❩, ❨1, 2, 3❩❩); ä ≔ ComplexMatrix(❨❨7, 3, 1❩, ❨1, 5, 0❩, ❨1, 1, −3❩❩); '+'ö ≔ ComplexMatrix(❨❨0, 0, 0❩, ❨0, 0, 0❩, ❨0, 0, 0❩❩);', null);

  TestComplexMatrixQ('IsMatrix', ComplexMatrixNames);
  TestComplexMatrixQ('IsRow', 'Eë');
  TestComplexMatrixQ('IsCol', 'Dü');
  TestComplexMatrixQ('IsSquare', 'AFGHIJKLMNOPQRSTUWXYcdfghjklmnpqrstuvwxyzoÅÄÖåäö');
  TestComplexMatrixQ('IsIdentity', 'F');
  TestComplexMatrixQ('IsZeroMatrix', 'ëüö');
  TestComplexMatrixQ('IsDiagonal', 'Fnpqëüö');
  TestComplexMatrixQ('IsAntidiagonal', 'Rö');
  TestComplexMatrixQ('IsReversal', 'R');
  TestComplexMatrixQ('IsUpperTriangular', 'BEFGKSTUVWXYZabnpqsëüö');
  TestComplexMatrixQ('IsLowerTriangular', 'DFHLnpqëüö');
  TestComplexMatrixQ('IsTriangular', 'BEFGKSTUVWXYZabnpqsëüöDFHLnpqëüö');
  TestComplexMatrixQ('IsRowEchelonForm', 'BEFGKSTVWXYZabnpqsëüö');
  TestComplexMatrixQ('IsReducedRowEchelonForm', 'BFWZpëüö');
  TestComplexMatrixQ('IsScalar', 'Fnö');
  TestComplexMatrixQ('IsSymmetric', 'FJPRdfnpqtö');
  TestComplexMatrixQ('IsSkewSymmetric', 'jö');
  TestComplexMatrixQ('IsHermitian', 'FIJQRpqtuvwxyzö');
  TestComplexMatrixQ('IsSkewHermitian', 'knö');
  TestComplexMatrixQ('IsOrthogonal', 'FJRcmq');
  TestComplexMatrixQ('IsUnitary', 'FIJRclnq');
  TestComplexMatrixQ('IsNormal', 'FIJQRchklmnpqtuvwxyzö');
  TestComplexMatrixQ('IsBinary', 'FJRpëüö');
  TestComplexMatrixQ('IsPermutation', 'FJR');
  TestComplexMatrixQ('IsCirculant', 'EFhnëüö');
  TestComplexMatrixQ('IsToeplitz', 'DEFghnëüö');
  TestComplexMatrixQ('IsHankel', 'DEJRdfëüö');
  TestComplexMatrixQ('IsUpperHessenberg', 'ABCEFGIJKLMOSTUVWXYZabclmnpqsëüö');
  TestComplexMatrixQ('IsLowerHessenberg', 'CDFHIJKLMNclmnpqsëüö');
  TestComplexMatrixQ('IsTridiagonal', 'FIJKLMclmnpqsö');
  TestComplexMatrixQ('IsUpperBidiagonal', 'FKnpqsö');
  TestComplexMatrixQ('IsLowerBidiagonal', 'FLnpqö');
  TestComplexMatrixQ('IsBidiagonal', 'FKLnpqsö');
  TestComplexMatrixQ('IsCentrosymmetric', 'FRnroö');
  TestComplexMatrixQ('IsVandermonde', 'Å');
  TestComplexMatrixQ('IsIdempotent', 'Fpö');
  TestComplexMatrixQ('IsInvolution', 'FIJRq');
//  TestComplexMatrixQ('IsPositiveDefinite', 'Fz');                              // TODO (5279121). Cannot compute spectrum of v and w. Investigate. Wishlist.
//  TestComplexMatrixQ('IsPositiveSemidefinite', 'Fpwzö');
//  TestComplexMatrixQ('IsNegativeDefinite', 'u');
//  TestComplexMatrixQ('IsNegativeSemidefinite', 'uxö');
//  TestComplexMatrixQ('IsIndefinite', 'QRIJqtvy');
//  TestComplexMatrixQ('IsNilpotent', 'sö');
//  TestComplexMatrixQ('IsPositive', '');                                        // N/A to complex matrices
//  TestComplexMatrixQ('IsNonNegative', '');
//  TestComplexMatrixQ('IsNegative', '');
//  TestComplexMatrixQ('IsNonPositive', '');
  TestComplexMatrixQ('IsZero', 'ëüö');
  TestComplexMatrixQ('IsNonZero', 'ABCDEFGHIJKLMNOPQRSTUVWXYZÅÄÖabcdfghjklmnopqrstuvwxyzåä');
  TestComplexMatrixQ('IsDiagonallyDominant', 'FSclnpqwzåäö');
  TestComplexMatrixQ('IsStrictlyDiagonallyDominant', 'Fnqä');

  TestComplexMatrixQImplies('IsUpperTriangular', 'IsTriangular');
  TestComplexMatrixQImplies('IsLowerTriangular', 'IsTriangular');

  TestComplexMatrixQImplies('IsUpperTriangular', 'IsUpperHessenberg');
  TestComplexMatrixQImplies('IsLowerTriangular', 'IsLowerHessenberg');

  TestComplexMatrixQImplies('IsUpperBidiagonal', 'IsBidiagonal');
  TestComplexMatrixQImplies('IsLowerBidiagonal', 'IsBidiagonal');

  TestComplexMatrixQImplies('IsBidiagonal', 'IsTridiagonal');

  TestComplexMatrixQImplies('IsRowEchelonForm', 'IsUpperTriangular');
  TestComplexMatrixQImplies('IsReducedRowEchelonForm', 'IsRowEchelonForm');

  TestComplexMatrixQImplies('IsIdentity', 'IsScalar');
  TestComplexMatrixQImplies('IsScalar', 'IsDiagonal');

  TestComplexMatrixQImplies('IsReversal', 'IsAntidiagonal');

  TestComplexMatrixQImplies('IsCirculant', 'IsToeplitz');

  TestComplexMatrixQImplies('IsIdentity', 'IsPermutation');
  TestComplexMatrixQImplies('IsReversal', 'IsPermutation');
  TestComplexMatrixQImplies('IsPermutation', 'IsBinary');

//  TestComplexMatrixQImplies('IsPositiveDefinite', 'IsPositiveSemidefinite');   // TODO (see bm 5279121)
//  TestComplexMatrixQImplies('IsNegativeDefinite', 'IsNegativeSemidefinite');

//  TestComplexMatrixQImplies('IsPositiveDefinite', 'IsSymmetric');
//  TestComplexMatrixQImplies('IsNegativeDefinite', 'IsSymmetric');
//  TestComplexMatrixQImplies('IsPositiveSemidefinite', 'IsSymmetric');
//  TestComplexMatrixQImplies('IsNegativeSemidefinite', 'IsSymmetric');
//  TestComplexMatrixQImplies('IsIndefinite', 'IsSymmetric');

  TestComplexMatrixQImplies('IsHermitian', 'IsNormal');
  TestComplexMatrixQImplies('IsSkewHermitian', 'IsNormal');
  TestComplexMatrixQImplies('IsUnitary', 'IsNormal');

  TestComplexMatrixQImplies('IsSymmetric', 'IsSquare');
  TestComplexMatrixQImplies('IsSkewSymmetric', 'IsSquare');
  TestComplexMatrixQImplies('IsHermitian', 'IsSquare');
  TestComplexMatrixQImplies('IsSkewHermitian', 'IsSquare');
  TestComplexMatrixQImplies('IsOrthogonal', 'IsSquare');
  TestComplexMatrixQImplies('IsUnitary', 'IsSquare');
  TestComplexMatrixQImplies('IsNormal', 'IsSquare');
  TestComplexMatrixQImplies('IsIdentity', 'IsSquare');
  TestComplexMatrixQImplies('IsReversal', 'IsSquare');
  TestComplexMatrixQImplies('IsScalar', 'IsSquare');
//  TestComplexMatrixQImplies('IsNilpotent', 'IsSquare');                        // TODO (see bm 5279121)

  TestComplexMatrixQImplies('IsStrictlyDiagonallyDominant', 'IsDiagonallyDominant');

  TestComplexMatrixQImplies('IsRow', 'IsToeplitz');
  TestComplexMatrixQImplies('IsRow', 'IsCirculant');
  TestComplexMatrixQImplies('IsRow', 'IsHankel');
  TestComplexMatrixQImplies('IsRow', 'IsUpperTriangular');

  TestComplexMatrixQImplies('IsCol', 'IsToeplitz');
  TestComplexMatrixQImplies('IsCol', 'IsHankel');
  TestComplexMatrixQImplies('IsCol', 'IsLowerTriangular');

  Test('delete(A); delete(B); delete(C); delete(D); delete(E); delete(F); delete(g); delete(h); delete(G); delete(H); delete(I); delete(J); delete(K); delete(L); delete(M)', success);
  Test('delete(N), delete(O); delete(P); delete(Q); delete(R); delete(S); delete(T); delete(U); delete(V); delete(W); delete(X); delete(Y); delete(Z); delete(j); delete(k)', success);
  Test('delete(l); delete(m); delete(n); delete(o); delete(p); delete(q); delete(r); delete(s); delete(a); delete(b); delete(c); delete(f); delete(t); delete(u); delete(v)', success);
  Test('delete(w); delete(x); delete(y); delete(z); delete(Å); delete(Ä); delete(Ö); delete(å); delete(ä); delete(ö); delete(ü); delete(ë)', success);

  Test('IsIdentity(❨❨1.0002, 0, 0.00001⋅i❩, ❨0, 0.9992, 0❩, ❨0, 0, 1.0003❩❩)', False);
  Test('IsIdentity(❨❨1.0002, 0, 0.00001⋅i❩, ❨0, 0.9992, 0❩, ❨0, 0, 1.0003❩❩, 1E-6)', False);
  Test('IsIdentity(❨❨1.0002, 0, 0.00001⋅i❩, ❨0, 0.9992, 0❩, ❨0, 0, 1.0003❩❩, 1E-5)', False);
  Test('IsIdentity(❨❨1.0002, 0, 0.00001⋅i❩, ❨0, 0.9992, 0❩, ❨0, 0, 1.0003❩❩, 1E-4)', False);
  Test('IsIdentity(❨❨1.0002, 0, 0.00001⋅i❩, ❨0, 0.9992, 0❩, ❨0, 0, 1.0003❩❩, 1E-3)', True);
  Test('IsIdentity(❨❨1.0002, 0, 0.00001⋅i❩, ❨0, 0.9992, 0❩, ❨0, 0, 1.0003❩❩, 1E-2)', True);
  Test('IsIdentity(❨❨1.0002, 0, 0.00001⋅i❩, ❨0, 0.9992, 0❩, ❨0, 0, 1.0003❩❩, 1E-1)', True);

  Test('IsZeroMatrix(❨❨0, 0.002, 0.003⋅i, −0.001❩, ❨0.0002, 0.004, 0.001⋅i, 0❩❩)', False);
  Test('IsZeroMatrix(❨❨0, 0.002, 0.003⋅i, −0.001❩, ❨0.0002, 0.004, 0.001⋅i, 0❩❩, 1E-6)', False);
  Test('IsZeroMatrix(❨❨0, 0.002, 0.003⋅i, −0.001❩, ❨0.0002, 0.004, 0.001⋅i, 0❩❩, 1E-5)', False);
  Test('IsZeroMatrix(❨❨0, 0.002, 0.003⋅i, −0.001❩, ❨0.0002, 0.004, 0.001⋅i, 0❩❩, 1E-4)', False);
  Test('IsZeroMatrix(❨❨0, 0.002, 0.003⋅i, −0.001❩, ❨0.0002, 0.004, 0.001⋅i, 0❩❩, 1E-3)', False);
  Test('IsZeroMatrix(❨❨0, 0.002, 0.003⋅i, −0.001❩, ❨0.0002, 0.004, 0.001⋅i, 0❩❩, 1E-2)', True);
  Test('IsZeroMatrix(❨❨0, 0.002, 0.003⋅i, −0.001❩, ❨0.0002, 0.004, 0.001⋅i, 0❩❩, 1E-1)', True);

  Test('IsDiagonal(❨❨1, 0.0003, 0.0002⋅i❩, ❨−0.0005, 2, 0.0002⋅i❩, ❨0.0003⋅i, 0.0001 + 0.0002⋅i, 3❩❩)', False);
  Test('IsDiagonal(❨❨1, 0.0003, 0.0002⋅i❩, ❨−0.0005, 2, 0.0002⋅i❩, ❨0.0003⋅i, 0.0001 + 0.0002⋅i, 3❩❩, 1E-6)', False);
  Test('IsDiagonal(❨❨1, 0.0003, 0.0002⋅i❩, ❨−0.0005, 2, 0.0002⋅i❩, ❨0.0003⋅i, 0.0001 + 0.0002⋅i, 3❩❩, 1E-5)', False);
  Test('IsDiagonal(❨❨1, 0.0003, 0.0002⋅i❩, ❨−0.0005, 2, 0.0002⋅i❩, ❨0.0003⋅i, 0.0001 + 0.0002⋅i, 3❩❩, 1E-4)', False);
  Test('IsDiagonal(❨❨1, 0.0003, 0.0002⋅i❩, ❨−0.0005, 2, 0.0002⋅i❩, ❨0.0003⋅i, 0.0001 + 0.0002⋅i, 3❩❩, 1E-3)', True);
  Test('IsDiagonal(❨❨1, 0.0003, 0.0002⋅i❩, ❨−0.0005, 2, 0.0002⋅i❩, ❨0.0003⋅i, 0.0001 + 0.0002⋅i, 3❩❩, 1E-2)', True);
  Test('IsDiagonal(❨❨1, 0.0003, 0.0002⋅i❩, ❨−0.0005, 2, 0.0002⋅i❩, ❨0.0003⋅i, 0.0001 + 0.0002⋅i, 3❩❩, 1E-1)', True);

  Test('IsAntidiagonal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 2, 0.0004❩, ❨3⋅i, −0.00031, 0.0004❩❩)', False);
  Test('IsAntidiagonal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 2, 0.0004❩, ❨3⋅i, −0.00031, 0.0004❩❩, 1E-6)', False);
  Test('IsAntidiagonal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 2, 0.0004❩, ❨3⋅i, −0.00031, 0.0004❩❩, 1E-5)', False);
  Test('IsAntidiagonal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 2, 0.0004❩, ❨3⋅i, −0.00031, 0.0004❩❩, 1E-4)', False);
  Test('IsAntidiagonal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 2, 0.0004❩, ❨3⋅i, −0.00031, 0.0004❩❩, 1E-3)', True);
  Test('IsAntidiagonal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 2, 0.0004❩, ❨3⋅i, −0.00031, 0.0004❩❩, 1E-2)', True);
  Test('IsAntidiagonal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 2, 0.0004❩, ❨3⋅i, −0.00031, 0.0004❩❩, 1E-1)', True);

  Test('IsReversal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 1.00001, 0.0004❩, ❨1 + 0.0001⋅i, −0.00031, 0.0004❩❩)', False);
  Test('IsReversal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 1.00001, 0.0004❩, ❨1 + 0.0001⋅i, −0.00031, 0.0004❩❩, 1E-6)', False);
  Test('IsReversal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 1.00001, 0.0004❩, ❨1 + 0.0001⋅i, −0.00031, 0.0004❩❩, 1E-5)', False);
  Test('IsReversal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 1.00001, 0.0004❩, ❨1 + 0.0001⋅i, −0.00031, 0.0004❩❩, 1E-4)', False);
  Test('IsReversal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 1.00001, 0.0004❩, ❨1 + 0.0001⋅i, −0.00031, 0.0004❩❩, 1E-3)', True);
  Test('IsReversal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 1.00001, 0.0004❩, ❨1 + 0.0001⋅i, −0.00031, 0.0004❩❩, 1E-2)', True);
  Test('IsReversal(❨❨0.0002, 0.0003⋅i, 1❩, ❨0.0001 + 0.0002⋅i, 1.00001, 0.0004❩, ❨1 + 0.0001⋅i, −0.00031, 0.0004❩❩, 1E-1)', True);

  Test('IsUpperTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩)', False);
  Test('IsUpperTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-6)', False);
  Test('IsUpperTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-5)', False);
  Test('IsUpperTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-4)', False);
  Test('IsUpperTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-3)', True);
  Test('IsUpperTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-2)', True);
  Test('IsUpperTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-1)', True);

  Test('IsLowerTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩)', False);
  Test('IsLowerTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-6)', False);
  Test('IsLowerTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-5)', False);
  Test('IsLowerTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-4)', False);
  Test('IsLowerTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-3)', False);
  Test('IsLowerTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-2)', True);
  Test('IsLowerTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-1)', True);

  Test('IsTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩)', False);
  Test('IsTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-6)', False);
  Test('IsTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-5)', False);
  Test('IsTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-4)', False);
  Test('IsTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-3)', True);
  Test('IsTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-2)', True);
  Test('IsTriangular(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-1)', True);

  Test('IsTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩)', False);
  Test('IsTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-6)', False);
  Test('IsTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-5)', False);
  Test('IsTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-4)', False);
  Test('IsTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-3)', False);
  Test('IsTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-2)', True);
  Test('IsTriangular(❨❨4, 0.001, 0.002❩, ❨5, i, 0.001 + 0.002⋅i❩, ❨4, 1, i❩❩, 1E-1)', True);

  Test('IsRowEchelonForm(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩)', False);
  Test('IsRowEchelonForm(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-6)', False);
  Test('IsRowEchelonForm(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-5)', False);
  Test('IsRowEchelonForm(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-4)', False);
  Test('IsRowEchelonForm(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-3)', True);
  Test('IsRowEchelonForm(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-2)', True);
  Test('IsRowEchelonForm(❨❨3, 2⋅i, 5, 6❩, ❨0.0001, 5, 2, 1❩, ❨−0.0002, 0.0005, i, 3❩, ❨0.0002⋅i, 0.0001, 0.0003, 5❩❩, 1E-1)', True);

  Test('IsReducedRowEchelonForm(❨❨1, 0.00001⋅i, 0.0001, 0❩, ❨0.0001, 1.00001, 0.0002, 0.0003❩, ❨−0.0002, 0.0005, 0.999999, 0.0001⋅i❩, ❨0.0002⋅i, 0.0001, 0.0003, 1❩❩)', False);
  Test('IsReducedRowEchelonForm(❨❨1, 0.00001⋅i, 0.0001, 0❩, ❨0.0001, 1.00001, 0.0002, 0.0003❩, ❨−0.0002, 0.0005, 0.999999, 0.0001⋅i❩, ❨0.0002⋅i, 0.0001, 0.0003, 1❩❩, 1E-6)', False);
  Test('IsReducedRowEchelonForm(❨❨1, 0.00001⋅i, 0.0001, 0❩, ❨0.0001, 1.00001, 0.0002, 0.0003❩, ❨−0.0002, 0.0005, 0.999999, 0.0001⋅i❩, ❨0.0002⋅i, 0.0001, 0.0003, 1❩❩, 1E-5)', False);
  Test('IsReducedRowEchelonForm(❨❨1, 0.00001⋅i, 0.0001, 0❩, ❨0.0001, 1.00001, 0.0002, 0.0003❩, ❨−0.0002, 0.0005, 0.999999, 0.0001⋅i❩, ❨0.0002⋅i, 0.0001, 0.0003, 1❩❩, 1E-4)', False);
  Test('IsReducedRowEchelonForm(❨❨1, 0.00001⋅i, 0.0001, 0❩, ❨0.0001, 1.00001, 0.0002, 0.0003❩, ❨−0.0002, 0.0005, 0.999999, 0.0001⋅i❩, ❨0.0002⋅i, 0.0001, 0.0003, 1❩❩, 1E-3)', True);
  Test('IsReducedRowEchelonForm(❨❨1, 0.00001⋅i, 0.0001, 0❩, ❨0.0001, 1.00001, 0.0002, 0.0003❩, ❨−0.0002, 0.0005, 0.999999, 0.0001⋅i❩, ❨0.0002⋅i, 0.0001, 0.0003, 1❩❩, 1E-2)', True);
  Test('IsReducedRowEchelonForm(❨❨1, 0.00001⋅i, 0.0001, 0❩, ❨0.0001, 1.00001, 0.0002, 0.0003❩, ❨−0.0002, 0.0005, 0.999999, 0.0001⋅i❩, ❨0.0002⋅i, 0.0001, 0.0003, 1❩❩, 1E-1)', True);

  Test('IsScalar(❨❨1.0002⋅i, 0, 0.0001❩, ❨0.0001⋅i, 1.0001⋅i, 0.0003❩, ❨0.0004, 0.0001, 1.0004⋅i❩❩)', False);
  Test('IsScalar(❨❨1.0002⋅i, 0, 0.0001❩, ❨0.0001⋅i, 1.0001⋅i, 0.0003❩, ❨0.0004, 0.0001, 1.0004⋅i❩❩, 1E-6)', False);
  Test('IsScalar(❨❨1.0002⋅i, 0, 0.0001❩, ❨0.0001⋅i, 1.0001⋅i, 0.0003❩, ❨0.0004, 0.0001, 1.0004⋅i❩❩, 1E-5)', False);
  Test('IsScalar(❨❨1.0002⋅i, 0, 0.0001❩, ❨0.0001⋅i, 1.0001⋅i, 0.0003❩, ❨0.0004, 0.0001, 1.0004⋅i❩❩, 1E-4)', False);
  Test('IsScalar(❨❨1.0002⋅i, 0, 0.0001❩, ❨0.0001⋅i, 1.0001⋅i, 0.0003❩, ❨0.0004, 0.0001, 1.0004⋅i❩❩, 1E-3)', True);
  Test('IsScalar(❨❨1.0002⋅i, 0, 0.0001❩, ❨0.0001⋅i, 1.0001⋅i, 0.0003❩, ❨0.0004, 0.0001, 1.0004⋅i❩❩, 1E-2)', True);
  Test('IsScalar(❨❨1.0002⋅i, 0, 0.0001❩, ❨0.0001⋅i, 1.0001⋅i, 0.0003❩, ❨0.0004, 0.0001, 1.0004⋅i❩❩, 1E-1)', True);

  Test('IsSymmetric(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 + 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 + 0.42⋅i, 7.62, 4.66❩❩)', False);
  Test('IsSymmetric(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 + 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 + 0.42⋅i, 7.62, 4.66❩❩, 1E-4)', False);
  Test('IsSymmetric(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 + 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 + 0.42⋅i, 7.62, 4.66❩❩, 1E-3)', False);
  Test('IsSymmetric(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 + 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 + 0.42⋅i, 7.62, 4.66❩❩, 1E-2)', True);
  Test('IsSymmetric(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 + 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 + 0.42⋅i, 7.62, 4.66❩❩, 1E-1)', True);

  Test('IsSkewSymmetric(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 − 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 − 0.42⋅i, 2.8, 0❩❩)', False);
  Test('IsSkewSymmetric(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 − 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 − 0.42⋅i, 2.8, 0❩❩, 1E-4)', False);
  Test('IsSkewSymmetric(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 − 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 − 0.42⋅i, 2.8, 0❩❩, 1E-3)', False);
  Test('IsSkewSymmetric(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 − 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 − 0.42⋅i, 2.8, 0❩❩, 1.1E-2)', True);
  Test('IsSkewSymmetric(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 − 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 − 0.42⋅i, 2.8, 0❩❩, 1E-1)', True);

  Test('IsHermitian(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩)', False);
  Test('IsHermitian(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩, 1E-4)', False);
  Test('IsHermitian(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩, 1E-3)', False);
  Test('IsHermitian(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩, 1E-2)', True);
  Test('IsHermitian(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩, 1E-1)', True);

  Test('IsSkewHermitian(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 + 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 + 0.42⋅i, 2.8, 0❩❩)', False);
  Test('IsSkewHermitian(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 + 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 + 0.42⋅i, 2.8, 0❩❩, 1E-4)', False);
  Test('IsSkewHermitian(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 + 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 + 0.42⋅i, 2.8, 0❩❩, 1E-3)', False);
  Test('IsSkewHermitian(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 + 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 + 0.42⋅i, 2.8, 0❩❩, 1.1E-2)', True);
  Test('IsSkewHermitian(❨❨0, 1.78, −4.22 + 3.1⋅i, 0.98❩, ❨−1.79, 0, 2.11, −2.13 + 0.43⋅i❩, ❨4.21 + 3.1⋅i, −2.11, 0, −2.8❩, ❨−0.97, 2.13 + 0.42⋅i, 2.8, 0❩❩, 1E-1)', True);

  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, i, 1.414❩, ❨0, −1.414, i❩❩)', False);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, i, 1.414❩, ❨0, −1.414, i❩❩, 1E-6)', False);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, i, 1.414❩, ❨0, −1.414, i❩❩, 1E-5)', False);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, i, 1.414❩, ❨0, −1.414, i❩❩, 1E-4)', False);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, i, 1.414❩, ❨0, −1.414, i❩❩, 1E-3)', True);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, i, 1.414❩, ❨0, −1.414, i❩❩, 1E-2)', True);
  Test('IsOrthogonal(❨❨1, 0, 0❩, ❨0, i, 1.414❩, ❨0, −1.414, i❩❩, 1E-1)', True);

  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.707⋅i, −0.707❩, ❨0, 0.707⋅i, 0.707❩❩)', False);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.707⋅i, −0.707❩, ❨0, 0.707⋅i, 0.707❩❩, 1E-6)', False);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.707⋅i, −0.707❩, ❨0, 0.707⋅i, 0.707❩❩, 1E-5)', False);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.707⋅i, −0.707❩, ❨0, 0.707⋅i, 0.707❩❩, 1E-4)', False);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.707⋅i, −0.707❩, ❨0, 0.707⋅i, 0.707❩❩, 1E-3)', True);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.707⋅i, −0.707❩, ❨0, 0.707⋅i, 0.707❩❩, 1E-2)', True);
  Test('IsUnitary(❨❨1, 0, 0❩, ❨0, 0.707⋅i, −0.707❩, ❨0, 0.707⋅i, 0.707❩❩, 1E-1)', True);

  Test('IsNormal(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩)', False);
  Test('IsNormal(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩, 1E-4)', False);
  Test('IsNormal(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩, 1E-3)', False);
  Test('IsNormal(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩, 1E-2)', False);
  Test('IsNormal(❨❨2.46, 3.24, 4.22 + 3.1⋅i, 1.88❩, ❨3.25, 2.84, 3.23, 2.13 + 0.41⋅i❩, ❨4.21 − 3.1⋅i, 3.23, 4.92, 7.62❩, ❨1.87, 2.13 − 0.42⋅i, 7.62, 4.66❩❩, 1E-1)', True);

  Test('IsBinary(❨❨1.0001, 0.0002⋅i, 1.0001 − 0.0003⋅i❩, ❨1.00031, 0.00051⋅i, −0.0001⋅i❩❩)', False);
  Test('IsBinary(❨❨1.0001, 0.0002⋅i, 1.0001 − 0.0003⋅i❩, ❨1.00031, 0.00051⋅i, −0.0001⋅i❩❩, 1E-6)', False);
  Test('IsBinary(❨❨1.0001, 0.0002⋅i, 1.0001 − 0.0003⋅i❩, ❨1.00031, 0.00051⋅i, −0.0001⋅i❩❩, 1E-5)', False);
  Test('IsBinary(❨❨1.0001, 0.0002⋅i, 1.0001 − 0.0003⋅i❩, ❨1.00031, 0.00051⋅i, −0.0001⋅i❩❩, 1E-4)', False);
  Test('IsBinary(❨❨1.0001, 0.0002⋅i, 1.0001 − 0.0003⋅i❩, ❨1.00031, 0.00051⋅i, −0.0001⋅i❩❩, 1E-3)', True);
  Test('IsBinary(❨❨1.0001, 0.0002⋅i, 1.0001 − 0.0003⋅i❩, ❨1.00031, 0.00051⋅i, −0.0001⋅i❩❩, 1E-2)', True);
  Test('IsBinary(❨❨1.0001, 0.0002⋅i, 1.0001 − 0.0003⋅i❩, ❨1.00031, 0.00051⋅i, −0.0001⋅i❩❩, 1E-1)', True);

  Test('IsPermutation(❨❨1.0001, 0.0001, 0.0001⋅i❩, ❨−0.0002, 0.00003⋅i, 0.999999❩, ❨0.0002, 1.000003, 0.000012❩❩)', False);
  Test('IsPermutation(❨❨1.0001, 0.0001, 0.0001⋅i❩, ❨−0.0002, 0.00003⋅i, 0.999999❩, ❨0.0002, 1.000003, 0.000012❩❩, 1E-6)', False);
  Test('IsPermutation(❨❨1.0001, 0.0001, 0.0001⋅i❩, ❨−0.0002, 0.00003⋅i, 0.999999❩, ❨0.0002, 1.000003, 0.000012❩❩, 1E-5)', False);
  Test('IsPermutation(❨❨1.0001, 0.0001, 0.0001⋅i❩, ❨−0.0002, 0.00003⋅i, 0.999999❩, ❨0.0002, 1.000003, 0.000012❩❩, 1E-4)', False);
  Test('IsPermutation(❨❨1.0001, 0.0001, 0.0001⋅i❩, ❨−0.0002, 0.00003⋅i, 0.999999❩, ❨0.0002, 1.000003, 0.000012❩❩, 1E-3)', True);
  Test('IsPermutation(❨❨1.0001, 0.0001, 0.0001⋅i❩, ❨−0.0002, 0.00003⋅i, 0.999999❩, ❨0.0002, 1.000003, 0.000012❩❩, 1E-2)', True);
  Test('IsPermutation(❨❨1.0001, 0.0001, 0.0001⋅i❩, ❨−0.0002, 0.00003⋅i, 0.999999❩, ❨0.0002, 1.000003, 0.000012❩❩, 1E-1)', True);

  Test('IsCirculant(❨❨3, i, 2, 5❩, ❨5, 3, i, 2❩, ❨2, 5, 2.995, i❩, ❨i, 2, 5, 3❩❩)', False);
  Test('IsCirculant(❨❨3, i, 2, 5❩, ❨5, 3, i, 2❩, ❨2, 5, 2.995, i❩, ❨i, 2, 5, 3❩❩, 1E-6)', False);
  Test('IsCirculant(❨❨3, i, 2, 5❩, ❨5, 3, i, 2❩, ❨2, 5, 2.995, i❩, ❨i, 2, 5, 3❩❩, 1E-5)', False);
  Test('IsCirculant(❨❨3, i, 2, 5❩, ❨5, 3, i, 2❩, ❨2, 5, 2.995, i❩, ❨i, 2, 5, 3❩❩, 1E-4)', False);
  Test('IsCirculant(❨❨3, i, 2, 5❩, ❨5, 3, i, 2❩, ❨2, 5, 2.995, i❩, ❨i, 2, 5, 3❩❩, 1E-3)', False);
  Test('IsCirculant(❨❨3, i, 2, 5❩, ❨5, 3, i, 2❩, ❨2, 5, 2.995, i❩, ❨i, 2, 5, 3❩❩, 1E-2)', True);
  Test('IsCirculant(❨❨3, i, 2, 5❩, ❨5, 3, i, 2❩, ❨2, 5, 2.995, i❩, ❨i, 2, 5, 3❩❩, 1E-1)', True);

  Test('IsToeplitz(❨❨1, 3, 2, i❩, ❨5, 1.0002, 3, 2❩, ❨6, 5, 1, 3❩, ❨8, 6, 5, 1❩❩)', False);
  Test('IsToeplitz(❨❨1, 3, 2, i❩, ❨5, 1.0002, 3, 2❩, ❨6, 5, 1, 3❩, ❨8, 6, 5, 1❩❩, 1E-6)', False);
  Test('IsToeplitz(❨❨1, 3, 2, i❩, ❨5, 1.0002, 3, 2❩, ❨6, 5, 1, 3❩, ❨8, 6, 5, 1❩❩, 1E-5)', False);
  Test('IsToeplitz(❨❨1, 3, 2, i❩, ❨5, 1.0002, 3, 2❩, ❨6, 5, 1, 3❩, ❨8, 6, 5, 1❩❩, 1E-4)', False);
  Test('IsToeplitz(❨❨1, 3, 2, i❩, ❨5, 1.0002, 3, 2❩, ❨6, 5, 1, 3❩, ❨8, 6, 5, 1❩❩, 1E-3)', True);
  Test('IsToeplitz(❨❨1, 3, 2, i❩, ❨5, 1.0002, 3, 2❩, ❨6, 5, 1, 3❩, ❨8, 6, 5, 1❩❩, 1E-2)', True);
  Test('IsToeplitz(❨❨1, 3, 2, i❩, ❨5, 1.0002, 3, 2❩, ❨6, 5, 1, 3❩, ❨8, 6, 5, 1❩❩, 1E-1)', True);

  Test('IsHankel(❨❨7, 4, i, 2❩, ❨4, i, 2, 8❩, ❨i, 2, 8, 0❩, ❨2, 8, 0.00002, 1❩❩)', False);
  Test('IsHankel(❨❨7, 4, i, 2❩, ❨4, i, 2, 8❩, ❨i, 2, 8, 0❩, ❨2, 8, 0.00002, 1❩❩, 1E-6)', False);
  Test('IsHankel(❨❨7, 4, i, 2❩, ❨4, i, 2, 8❩, ❨i, 2, 8, 0❩, ❨2, 8, 0.00002, 1❩❩, 1E-5)', False);
  Test('IsHankel(❨❨7, 4, i, 2❩, ❨4, i, 2, 8❩, ❨i, 2, 8, 0❩, ❨2, 8, 0.00002, 1❩❩, 1E-4)', True);
  Test('IsHankel(❨❨7, 4, i, 2❩, ❨4, i, 2, 8❩, ❨i, 2, 8, 0❩, ❨2, 8, 0.00002, 1❩❩, 1E-3)', True);
  Test('IsHankel(❨❨7, 4, i, 2❩, ❨4, i, 2, 8❩, ❨i, 2, 8, 0❩, ❨2, 8, 0.00002, 1❩❩, 1E-2)', True);
  Test('IsHankel(❨❨7, 4, i, 2❩, ❨4, i, 2, 8❩, ❨i, 2, 8, 0❩, ❨2, 8, 0.00002, 1❩❩, 1E-1)', True);

  Test('IsUpperHessenberg(❨❨6, 2, 3, 1❩, ❨7, 3, i, i❩, ❨0.003, 4, 1, 2❩, ❨0.001, 0.002, 5, 8❩❩)', False);
  Test('IsUpperHessenberg(❨❨6, 2, 3, 1❩, ❨7, 3, i, i❩, ❨0.003, 4, 1, 2❩, ❨0.001, 0.002, 5, 8❩❩, 1E-6)', False);
  Test('IsUpperHessenberg(❨❨6, 2, 3, 1❩, ❨7, 3, i, i❩, ❨0.003, 4, 1, 2❩, ❨0.001, 0.002, 5, 8❩❩, 1E-5)', False);
  Test('IsUpperHessenberg(❨❨6, 2, 3, 1❩, ❨7, 3, i, i❩, ❨0.003, 4, 1, 2❩, ❨0.001, 0.002, 5, 8❩❩, 1E-4)', False);
  Test('IsUpperHessenberg(❨❨6, 2, 3, 1❩, ❨7, 3, i, i❩, ❨0.003, 4, 1, 2❩, ❨0.001, 0.002, 5, 8❩❩, 1E-3)', False);
  Test('IsUpperHessenberg(❨❨6, 2, 3, 1❩, ❨7, 3, i, i❩, ❨0.003, 4, 1, 2❩, ❨0.001, 0.002, 5, 8❩❩, 1E-2)', True);
  Test('IsUpperHessenberg(❨❨6, 2, 3, 1❩, ❨7, 3, i, i❩, ❨0.003, 4, 1, 2❩, ❨0.001, 0.002, 5, 8❩❩, 1E-1)', True);

  Test('IsLowerHessenberg(❨❨4, 5, 0.0001, 0.0002❩, ❨6, 2, 4, 0.0001❩, ❨i, 2, 4, 6❩, ❨6, 2, 2, 1❩❩)', False);
  Test('IsLowerHessenberg(❨❨4, 5, 0.0001, 0.0002❩, ❨6, 2, 4, 0.0001❩, ❨i, 2, 4, 6❩, ❨6, 2, 2, 1❩❩, 1E-6)', False);
  Test('IsLowerHessenberg(❨❨4, 5, 0.0001, 0.0002❩, ❨6, 2, 4, 0.0001❩, ❨i, 2, 4, 6❩, ❨6, 2, 2, 1❩❩, 1E-5)', False);
  Test('IsLowerHessenberg(❨❨4, 5, 0.0001, 0.0002❩, ❨6, 2, 4, 0.0001❩, ❨i, 2, 4, 6❩, ❨6, 2, 2, 1❩❩, 1E-4)', False);
  Test('IsLowerHessenberg(❨❨4, 5, 0.0001, 0.0002❩, ❨6, 2, 4, 0.0001❩, ❨i, 2, 4, 6❩, ❨6, 2, 2, 1❩❩, 1E-3)', True);
  Test('IsLowerHessenberg(❨❨4, 5, 0.0001, 0.0002❩, ❨6, 2, 4, 0.0001❩, ❨i, 2, 4, 6❩, ❨6, 2, 2, 1❩❩, 1E-2)', True);
  Test('IsLowerHessenberg(❨❨4, 5, 0.0001, 0.0002❩, ❨6, 2, 4, 0.0001❩, ❨i, 2, 4, 6❩, ❨6, 2, 2, 1❩❩, 1E-1)', True);

  Test('IsTridiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨5, 1, 2, 0.0003❩, ❨0.0001, 2, i, 4❩, ❨0.0002, 0.0003, 6, 2❩❩)', False);
  Test('IsTridiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨5, 1, 2, 0.0003❩, ❨0.0001, 2, i, 4❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-6)', False);
  Test('IsTridiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨5, 1, 2, 0.0003❩, ❨0.0001, 2, i, 4❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-5)', False);
  Test('IsTridiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨5, 1, 2, 0.0003❩, ❨0.0001, 2, i, 4❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-4)', False);
  Test('IsTridiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨5, 1, 2, 0.0003❩, ❨0.0001, 2, i, 4❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-3)', True);
  Test('IsTridiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨5, 1, 2, 0.0003❩, ❨0.0001, 2, i, 4❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-2)', True);
  Test('IsTridiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨5, 1, 2, 0.0003❩, ❨0.0001, 2, i, 4❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-1)', True);

  Test('IsUpperBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩)', False);
  Test('IsUpperBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-6)', False);
  Test('IsUpperBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-5)', False);
  Test('IsUpperBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-4)', False);
  Test('IsUpperBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-3)', True);
  Test('IsUpperBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-2)', True);
  Test('IsUpperBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-1)', True);

  Test('IsLowerBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩)', False);
  Test('IsLowerBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-6)', False);
  Test('IsLowerBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-5)', False);
  Test('IsLowerBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-4)', False);
  Test('IsLowerBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-3)', True);
  Test('IsLowerBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-2)', True);
  Test('IsLowerBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-1)', True);

  Test('IsBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩)', False);
  Test('IsBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-6)', False);
  Test('IsBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-5)', False);
  Test('IsBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-4)', False);
  Test('IsBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-3)', True);
  Test('IsBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-2)', True);
  Test('IsBidiagonal(❨❨6, 2, 0.0001, 0.0002❩, ❨0, 1, 2, 0.0003❩, ❨0.0001, 0.0001, i, 4❩, ❨0.0002, 0.0003, 0, 2❩❩, 1E-1)', True);

  Test('IsBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩)', False);
  Test('IsBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-6)', False);
  Test('IsBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-5)', False);
  Test('IsBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-4)', False);
  Test('IsBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-3)', True);
  Test('IsBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-2)', True);
  Test('IsBidiagonal(❨❨6, 0.0001, 0.0001, 0.0002❩, ❨5, 1, 0.0001, 0.0003❩, ❨0.0001, 2, i, 0.0001❩, ❨0.0002, 0.0003, 6, 2❩❩, 1E-1)', True);

  Test('IsCentrosymmetric(❨❨1, 3.0002, i, 6❩, ❨4.0003, 2, 7, 5❩, ❨5.0001, 7, 2.0003, 4❩, ❨6, i, 3, 0.99999❩❩)', False);
  Test('IsCentrosymmetric(❨❨1, 3.0002, i, 6❩, ❨4.0003, 2, 7, 5❩, ❨5.0001, 7, 2.0003, 4❩, ❨6, i, 3, 0.99999❩❩, 1E-6)', False);
  Test('IsCentrosymmetric(❨❨1, 3.0002, i, 6❩, ❨4.0003, 2, 7, 5❩, ❨5.0001, 7, 2.0003, 4❩, ❨6, i, 3, 0.99999❩❩, 1E-5)', False);
  Test('IsCentrosymmetric(❨❨1, 3.0002, i, 6❩, ❨4.0003, 2, 7, 5❩, ❨5.0001, 7, 2.0003, 4❩, ❨6, i, 3, 0.99999❩❩, 1E-4)', False);
  Test('IsCentrosymmetric(❨❨1, 3.0002, i, 6❩, ❨4.0003, 2, 7, 5❩, ❨5.0001, 7, 2.0003, 4❩, ❨6, i, 3, 0.99999❩❩, 1E-3)', True);
  Test('IsCentrosymmetric(❨❨1, 3.0002, i, 6❩, ❨4.0003, 2, 7, 5❩, ❨5.0001, 7, 2.0003, 4❩, ❨6, i, 3, 0.99999❩❩, 1E-2)', True);
  Test('IsCentrosymmetric(❨❨1, 3.0002, i, 6❩, ❨4.0003, 2, 7, 5❩, ❨5.0001, 7, 2.0003, 4❩, ❨6, i, 3, 0.99999❩❩, 1E-1)', True);

  Test('IsVandermonde(❨❨1, 3.000002, 9, 27❩, ❨1, 2, 4.000001, 8❩, ❨0.99999999, 5, 25, 125❩, ❨1, i, −1, −i❩, ❨1, −1.0000003, 1, −1❩, ❨1, 10, 100, 1000❩❩)', False);
  Test('IsVandermonde(❨❨1, 3.000002, 9, 27❩, ❨1, 2, 4.000001, 8❩, ❨0.99999999, 5, 25, 125❩, ❨1, i, −1, −i❩, ❨1, −1.0000003, 1, −1❩, ❨1, 10, 100, 1000❩❩, 1E-6)', False);
  Test('IsVandermonde(❨❨1, 3.000002, 9, 27❩, ❨1, 2, 4.000001, 8❩, ❨0.99999999, 5, 25, 125❩, ❨1, i, −1, −i❩, ❨1, −1.0000003, 1, −1❩, ❨1, 10, 100, 1000❩❩, 1E-5)', False);
  Test('IsVandermonde(❨❨1, 3.000002, 9, 27❩, ❨1, 2, 4.000001, 8❩, ❨0.99999999, 5, 25, 125❩, ❨1, i, −1, −i❩, ❨1, −1.0000003, 1, −1❩, ❨1, 10, 100, 1000❩❩, 1E-4)', True);
  Test('IsVandermonde(❨❨1, 3.000002, 9, 27❩, ❨1, 2, 4.000001, 8❩, ❨0.99999999, 5, 25, 125❩, ❨1, i, −1, −i❩, ❨1, −1.0000003, 1, −1❩, ❨1, 10, 100, 1000❩❩, 1E-3)', True);
  Test('IsVandermonde(❨❨1, 3.000002, 9, 27❩, ❨1, 2, 4.000001, 8❩, ❨0.99999999, 5, 25, 125❩, ❨1, i, −1, −i❩, ❨1, −1.0000003, 1, −1❩, ❨1, 10, 100, 1000❩❩, 1E-2)', True);
  Test('IsVandermonde(❨❨1, 3.000002, 9, 27❩, ❨1, 2, 4.000001, 8❩, ❨0.99999999, 5, 25, 125❩, ❨1, i, −1, −i❩, ❨1, −1.0000003, 1, −1❩, ❨1, 10, 100, 1000❩❩, 1E-1)', True);

  Test('IsIdempotent(❨❨0, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩)', False);
  Test('IsIdempotent(❨❨0, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-6)', False);
  Test('IsIdempotent(❨❨0, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-5)', False);
  Test('IsIdempotent(❨❨0, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-4)', False);
  Test('IsIdempotent(❨❨0, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-3)', True);
  Test('IsIdempotent(❨❨0, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-2)', True);
  Test('IsIdempotent(❨❨0, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-1)', True);

  Test('IsInvolution(❨❨−0.9999999, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩)', False);
  Test('IsInvolution(❨❨−0.9999999, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-6)', False);
  Test('IsInvolution(❨❨−0.9999999, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-5)', False);
  Test('IsInvolution(❨❨−0.9999999, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-4)', False);
  Test('IsInvolution(❨❨−0.9999999, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-3)', False);
  Test('IsInvolution(❨❨−0.9999999, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-2)', True);
  Test('IsInvolution(❨❨−0.9999999, 0.0001, 0.0002⋅i❩, ❨0.00032, 1.00001, 0.0001❩, ❨−0.000036, 0.0005⋅i, 0.999999❩❩, 1E-1)', True);

  Test('IsZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩)', False);
  Test('IsZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-6)', False);
  Test('IsZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-5)', False);
  Test('IsZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-4)', False);
  Test('IsZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-3)', True);
  Test('IsZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-2)', True);
  Test('IsZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-1)', True);

  Test('IsNonZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩)', True);
  Test('IsNonZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-6)', True);
  Test('IsNonZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-5)', True);
  Test('IsNonZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-4)', True);
  Test('IsNonZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-3)', False);
  Test('IsNonZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-2)', False);
  Test('IsNonZero(❨❨0.0003, 0.0001 + 0.0001⋅i, −0.0002❩❩, 1E-1)', False);


  //
  // Basic functions on matrices
  //

  // Determinant

  Eps; Test('det(❨❨2, 1❩, ❨5, 3❩❩)', 1.0);
  Eps; Test('det(❨❨2, 8❩, ❨4, 5❩❩)', -22.0);
  Eps; Test('det(❨❨7, −3❩, ❨−2, 8❩❩)', 50.0);

  Eps; Test('det(❨❨5, 2, 1❩, ❨−2, 5, 1❩, ❨5, 2, 8❩❩)', 203.0);
  Eps; Test('det(❨❨−2, 4, 6❩, ❨2.7, −5.3, 5❩, ❨7, 2, 1❩❩)', 414.8);
  Eps; Test('det(❨❨1.5, 2.4, 7.4❩, ❨−2.3, 0.4, 7.5❩, ❨0.1, 4.6, 0❩❩)', -128.538);

  Eps; Test('det(❨❨5, 2, 4, 8❩, ❨2, 4, 1, 6❩, ❨0, 5, 2, 2❩, ❨5, 7, 8, 2❩❩)', -300.0);
  Eps; Test('det(❨❨7.5, 3.4, 2.2, −6.7❩, ❨4.1, 2.6, −2.4, 7.8❩, ❨2.6, 3.3, 7.2, 3.4❩, ❨2.5, 7.3, 2.2, 1.6❩❩)', -3788.5386);
  Eps; Test('det(❨❨6, 2, 4, 5❩, ❨5, 1, 0, 5❩, ❨1, 3, 0, 0❩, ❨0, 1, 0, 0❩❩)', 20.0);   // fun textbook example!

  Eps; Test('det(❨❨6, 2, 4, 1, 2❩, ❨6, 2, 1, 4, −2❩, ❨2, 4, 6, 8, 9❩, ❨−1, 5, 2, 2, 1❩, ❨4, 6, 2, 7, 2❩❩)', -2646.0);
  Eps; Test('det(❨❨6, 2, 3, −4, 6❩, ❨1, 5, 7, 3, 2❩, ❨5, 1042, 2, 5, 7❩, ❨−5, 6, 2, 2, 1❩, ❨4, 2, 5, 7, 8❩❩)', 2279440.0);
  Eps; Test('det(❨❨1534, 5432, 72, 527, 245❩, ❨63, 245, 752, 8375, 123❩, ❨35, 542, 876, 23, 524❩, ❨2314, 6453, 62, 7, 345❩, ❨35, 643, 87, 563, 345❩❩)', 5927413833983547.0);

  Eps; Test('det(❨❨5, 6, 2, 3, 4, 7, 8, 4, 6, 7❩, ❨1, 5, 7, 2, 6, 8, 9, 4, 5, 7❩, ❨4, 3, 7, 6, 4, 8, 9, 0, 2, 4❩, ❨4, 7, 8, 4, 3, 7, 8, 5, 6, 9❩, ❨1, 2, 5, 7, 8, 4, 3, 7, 6, 4❩, ❨4, 7, 8, 7, 5, 2, 5, 7, 1, 3❩, ❨2, 4, 6, 7, 5, 4, 2, 4, 8, 6❩, '+'❨4, 6, 1, 5, 8, 3, 2, 4, 6, 8❩, ❨4, 6, 2, 6, 8, 2, 3, 4, 4, 3❩, ❨1, 4, 7, 3, 2, 6, 8, 5, 2, 3❩❩)', 76224.0);
  Eps; Test('det(❨❨7, 6, 5, 6, 7, 5, 2, 3, 4, 5❩, ❨7, 6, 4, 8, 4, 3, 2, 1, 6, 7❩, ❨5, 6, 2, 3, 4, 5, 7, 8, 5, 3❩, ❨4, 6, 8, 3, 1, 3, 3, 6, 7, 8❩, ❨2, 8, 5, 4, 3, 3, 2, 4, 7, 8❩, ❨5, 2, 1, 3, 4, 7, 5, 4, 3, 2❩, ❨6, 3, 2, 6, 8, 5, 4, 3, 2, 6❩, '+'❨6, 7, 3, 3, 2, 1, 7, 5, 4, 2❩, ❨5, 7, 8, 3, 3, 2, 6, 7, 3, 7❩, ❨5, 2, 2, 2, 3, 1, 6, 2, 7, 4❩❩)', -1262250.0);
  Eps; Test('det(❨❨5, 6, 5, 3, 8, 6, 8, 6, 8, 9❩, ❨4, 5, 7, 2, −6, 7, 3, 5, 8, 1❩, ❨0, 5, 2, 7, 8, −3, 6, 7, 7, 6❩, ❨2, 4, 5, 6, 7, 2, 123, 5, 6, 7❩, ❨1, 2, 4, 0, 5, 6, 3, 5.6, 7, 8❩, ❨4, 3, 2, 3, 7, 2, −3, 5, 2, 1❩, ❨5, 3, 6, 7, 5, 2, 1, 3, 4, −5❩, '+'❨4, 6, 6, 2, 1, 3, 7, 4, 2, 4.5❩, ❨5, 2, 4, 6, 7, 3, 4, 1, 2, 5❩, ❨6, 3, 3, 2, 21, 22, 5, 6, 0, 3❩❩)', 5192807543.4);

  Eps; Test('det(' + MatLit_50x50_Rand_A + ')', 1460982.95465541113);
  Eps; Test('det(' + MatLit_50x50_Rand_B + ')', 28063.2591503821555);

  Test('A ≔ ' + MatLit_100x100_Rand_A + ';', null);
  Test('B ≔ ' + MatLit_100x100_Rand_B + ';', null);
  Test('C ≔ ' + MatLit_100x100_Rand_C + ';', null);

  Eps; Test('det(A)', 5.12406946272844878E24);
  Eps; Test('det(B)', 2.83947627757255657E25);
  Eps; Test('det(C)', -4.14560652548419427E23);

  Eps(1E-13 * 1E50); Test('det(A⋅B)', 5.12406946272844878E24 * 2.83947627757255657E25);
  Eps(1E-13 * 2E48); Test('det(A⋅C)', 5.12406946272844878E24 * -4.14560652548419427E23);
  Eps(1E-13 * 1E49); Test('det(B⋅C)', 2.83947627757255657E25 * -4.14560652548419427E23);

{$IFDEF QuickTest}
  Test('A ≔ RandomMatrix(1000);', null);
  StartStopwatch;
  Test('det(A);', null);
  StopStopwatch;
  TestDuration(2.5);
{$ELSE}
  Test('A, B, C ≔ RandomMatrix(1000), RandomMatrix(1000), RandomMatrix(1000);', null);
  StartStopwatch;
  Test('''(det(A), det(B), det(C));', null);
  StopStopwatch;
  TestDuration(8);
{$ENDIF}

  Test('A ≔ ' + MatLit_100x100_Rand_A + ';', null);
  Test('B ≔ ' + MatLit_100x100_Rand_B + ';', null);
  Test('C ≔ ' + MatLit_100x100_Rand_C + ';', null);

  Test('det(ZeroMatrix(10))', 0.0);
  Test('det(ZeroMatrix(50))', 0.0);
  Test('det(ZeroMatrix(100))', 0.0);
  Test('det(ZeroMatrix(500))', 0.0);

  Test('det(IdentityMatrix(10))', 1.0);
  Test('det(IdentityMatrix(50))', 1.0);
  Test('det(IdentityMatrix(100))', 1.0);
  Test('det(IdentityMatrix(500))', 1.0);

  Eps; Test('det(❨❨2 + 3⋅i, 1 − 2⋅i❩, ❨4 + 5⋅i, −2 + i❩❩)', -21 - ImaginaryUnit);
  Eps; Test('det(❨❨3 + 2⋅i, 1 − 2⋅i, 2 + 5⋅i❩, ❨−2 + 2⋅i, 3 + 4⋅i, 1 − i❩, ❨5 − 3⋅i, 2 + i, 1 + 4⋅i❩❩)', -95 - 190*ImaginaryUnit);
  Eps; Test('det(❨❨5, 2, 7, i❩, ❨1, −2, 4, 7❩, ❨4, −i, 2, 1+i❩, ❨2, 5, 7, 2❩❩)', 528 - 134*ImaginaryUnit);
  Eps; Test('det(❨❨6, 2, 4, 5, 1❩, ❨4, 6, i, 2, 1❩, ❨5, 2, 1, 8, 6❩, ❨5, −7, 2, 5, 8❩, ❨i, 2, 5, 6, 8❩❩)', 10199 - 831*ImaginaryUnit);
  Eps; Test('det(❨❨6, 7, 2, 4, 5, i❩, ❨2, 3, 5, 1, 2, 1+i❩, ❨5, 2, −i, 3, 5+2⋅i, 4❩, ❨5, 6, 2, 1, 5.6, 7❩, ❨i, 2, 3, 9, 6, 7❩, ❨4, 5, 1 − i, 5, 7, 2❩❩)', 14796 - 1437.8*ImaginaryUnit);

  Eps; Test('det(A + B⋅i)', -4.52570721632352749E40 + 4.52896057650712066E40*ImaginaryUnit);
  Eps; Test('det(A + C⋅i)', 8.91088961396279729E40 + 3.56936742573770336E40*ImaginaryUnit);
  Eps; Test('det(B + C⋅i)', 5.73508107803082928E40 + 2.07593237007640487E41*ImaginaryUnit);

  Test('delete(A); delete(B); delete(C)', success);

  // Trace

  Test('A ≔ ' + MatLit_100x100_Rand_A + ';', null);
  Test('B ≔ ' + MatLit_100x100_Rand_B + ';', null);
  Test('C ≔ ' + MatLit_100x100_Rand_C + ';', null);

  Test('tr(❨❨5, 1❩, ❨2, 6❩❩)', 11.0);
  Test('tr(❨❨5, 1, 6❩, ❨2, 6, 5❩, ❨4, 1, 2❩❩)', 13.0);
  Test('tr(❨❨5, 1, 6, 4❩, ❨2, 6, 5, 1❩, ❨4, 1, 2, 7❩, ❨5, 2, 7, 3❩❩)', 16.0);

  Eps; Test('tr(A)', 49.531483373139054);
  Eps; Test('tr(B)', 45.333559668157250);
  Eps; Test('tr(C)', 45.182895293924958);

  Test('tr(ZeroMatrix(10))', 0.0);
  Test('tr(ZeroMatrix(50))', 0.0);
  Test('tr(ZeroMatrix(100))', 0.0);
  Test('tr(ZeroMatrix(500))', 0.0);

  Test('tr(IdentityMatrix(10))', 10.0);
  Test('tr(IdentityMatrix(50))', 50.0);
  Test('tr(IdentityMatrix(100))', 100.0);
  Test('tr(IdentityMatrix(500))', 500.0);

  Test('tr(❨❨5 + 2⋅i, 2 + 3⋅i❩, ❨−5 + 2⋅i, 4 − 5⋅i❩❩)', 9 - 3*ImaginaryUnit);
  Test('tr(❨❨3, 5, 1❩, ❨2, i, 7❩, ❨4, 2, i❩❩)', 3 + 2*ImaginaryUnit);
  Test('tr(❨❨5, 6, 2, i❩, ❨4, −i, 2, 6❩, ❨5, 2, 1, 7❩, ❨0, 2, 1, 2⋅i❩❩)', 6 + ImaginaryUnit);

  Eps; Test('tr(A + B⋅i)', 49.5314833731390536 + 45.3335596681572497*ImaginaryUnit);
  Eps; Test('tr(A + C⋅i)', 49.5314833731390536 + 45.1828952939249575*ImaginaryUnit);
  Eps; Test('tr(B + C⋅i)', 45.3335596681572497 + 45.1828952939249575*ImaginaryUnit);

  Test('delete(A); delete(B); delete(C)', success);

  // Inverse

  Test('A ≔ ❨❨3, 1❩, ❨-2, 4❩❩', 2, [3, 1, -2, 4]);
  Eps; Test('B ≔ inv(A)', 2, [0.285714285714285714, -0.0714285714285714286, 0.142857142857142857, 0.214285714285714286]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨4, 2, 1❩, ❨4, −1, 2❩, ❨1, 5, 2❩❩', 3, [4, 2, 1, 4, -1, 2, 1, 5, 2]);
  Eps; Test('B ≔ inv(A)', 3, [0.307692307692307692, -0.025641025641025641, -0.128205128205128205, 0.153846153846153846, -0.179487179487179487, 0.102564102564102564, -0.538461538461538462, 0.461538461538461538, 0.307692307692307692]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨5, 1, 2, 3❩, ❨6, 2, 1, −4❩, ❨9, 2, 4, 7❩, ❨4, 1, 6, 8❩❩', 4, [5, 1, 2, 3, 6, 2, 1, -4, 9, 2, 4, 7, 4, 1, 6, 8]);
  Eps; Test('B ≔ inv(A)', 4, [1.38636363636363636, -0.0909090909090909091, -0.568181818181818182, -0.0681818181818181818, -5.81818181818181818, 0.545454545454545455, 2.90909090909090909, -0.090909090909090909, 0.863636363636363636, 0.0909090909090909091, -0.681818181818181818, 0.318181818181818182, -0.613636363636363636, -0.0909090909090909091, 0.431818181818181818, -0.0681818181818181818]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨5, 2, −6, 1, 2❩, ❨0, 2, 6, −7, 2❩, ❨1, 2, 7, 2, 3❩, ❨9, 8, −7, 1, 6❩, ❨5, 1, 2, 7, 6❩❩', 5, [5, 2, -6, 1, 2, 0, 2, 6, -7, 2, 1, 2, 7, 2, 3, 9, 8, -7, 1, 6, 5, 1, 2, 7, 6]);
  Eps; Test('B ≔ inv(A)', 5, [0.963923337091319053, -0.00338218714768883875, 0.553175497933107854, -0.347237880496054115, -0.249530251785043217, -0.226606538895152198, -0.114994363021420519, 0.141300263059000376, 0.19391206313416009, -0.150695227358136039, 0.285231116121758737, -0.00450958286358511838, 0.293122886133032694, -0.129650507328072153, -0.1104847801578354, -0.0496054114994363021, -0.129650507328072153, 0.0939496429913566329, 0.0225479143179255919, -0.00977076287110108982, -0.802705749718151071, 0.174746335963923337, -0.691845170988350244, 0.273957158962795941, 0.447951897782788425]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ' + MatLit_100x100_Rand_A + ';', null);
  Test('B ≔ ' + MatLit_100x100_Rand_B + ';', null);
  Test('C ≔ ' + MatLit_100x100_Rand_C + ';', null);

  StartStopwatch;
  Test('Ai ≔ inv(A);', null);
  Test('Bi ≔ inv(B);', null);
  Test('Ci ≔ inv(C);', null);
  StopStopwatch;
  TestDuration(0.1);

  Test('IsIdentity(A⋅Ai)', True);
  Test('IsIdentity(Ai⋅A)', True);
  Test('IsIdentity(B⋅Bi)', True);
  Test('IsIdentity(Bi⋅B)', True);
  Test('IsIdentity(C⋅Ci)', True);
  Test('IsIdentity(Ci⋅C)', True);

  Test('inv(❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 0❩❩)', failure, 'Matrix is singular.');
  Test('inv(❨❨1, 0, 0❩, ❨0, 1, 0❩❩)', failure, 'Cannot compute inverse of non-square matrix.');

  Test('A ≔ ❨❨2 + 3⋅i, 1 − 2⋅i❩, ❨−4⋅i, 6 + 5⋅i❩❩', 2, [2 + 3*ImaginaryUnit, 1 - 2*ImaginaryUnit, -4*ImaginaryUnit, 6 + 5*ImaginaryUnit]);
  Eps; Test('B ≔ inv(A)', 2, [0.181124880838894185 - 0.159199237368922784*ImaginaryUnit, 0.0562440419447092469 + 0.0400381315538608198*ImaginaryUnit, 0.122020972354623451 + 0.0190657769304099142*ImaginaryUnit, 0.101048617731172545 - 0.0467111534795042898*ImaginaryUnit]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨2, 1, i❩, ❨5, 1, 2❩, ❨−2, 5, −i❩❩', 3, [2, 1, ImaginaryUnit, 5, 1, 2, -2, 5, -ImaginaryUnit]);
  Eps; Test('B ≔ inv(A)', 3, [0.142276422764227642 + 0.21951219512195122*ImaginaryUnit, 0.121951219512195122 - 0.0975609756097560976*ImaginaryUnit, -0.0528455284552845528 - 0.0243902439024390244*ImaginaryUnit, 0.166666666666666667 + 5.01945450224770571E-21*ImaginaryUnit, -5.01945450224770571E-21, 0.166666666666666667, -0.439024390243902439 - 0.548780487804878049*ImaginaryUnit, 0.195121951219512195 + 0.243902439024390244*ImaginaryUnit, 0.0487804878048780488 + 0.060975609756097561*ImaginaryUnit]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨5, 2, i, 1❩, ❨6, 2, 1, 2❩, ❨−4, 1, 2, i❩, ❨−i, 1, 5, 8❩❩', 4, [5, 2, ImaginaryUnit, 1, 6, 2, 1, 2, -4, 1, 2, ImaginaryUnit, -ImaginaryUnit, 1, 5, 8]);
  Eps; Test('B ≔ inv(A)', 4, [-0.173183213920163767 - 0.142169907881269191*ImaginaryUnit, 0.242579324462640737 + 0.129989764585465711*ImaginaryUnit, -0.103070624360286592 + 0.026202661207778915*ImaginaryUnit, -0.035721596724667349 - 0.00184237461617195497*ImaginaryUnit, 0.50962128966223132 + 0.424564994882292733*ImaginaryUnit, -0.180143295803480041 - 0.396110542476970317*ImaginaryUnit, 0.366837256908904811 - 0.0570112589559877175*ImaginaryUnit, -0.0257932446264073695 + 0.000102354145342886399*ImaginaryUnit, -0.446878198567041965 - 0.61330603889457523*ImaginaryUnit, 0.430910951893551689 + 0.589559877175025589*ImaginaryUnit, 0.0679631525076765609 + 0.126714431934493347*ImaginaryUnit, -0.0360286591606960082 - 0.0792221084953940635*ImaginaryUnit, 0.233367451381780962 + 0.308597748208802456*ImaginaryUnit, -0.263050153531218014 - 0.288638689866939611*ImaginaryUnit, -0.0916069600818833163 - 0.0849539406345957011*ImaginaryUnit, 0.150972364380757421 + 0.0450358239508700102*ImaginaryUnit]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ' + MatLit_100x100_Rand_A + ';', null);
  Test('B ≔ ' + MatLit_100x100_Rand_B + ';', null);
  Test('C ≔ ' + MatLit_100x100_Rand_C + ';', null);

  Test('X ≔ A + B⋅i;', null);
  Test('Y ≔ A + C⋅i;', null);
  Test('Z ≔ B + C⋅i;', null);

  StartStopwatch;
  Test('Xi ≔ inv(X);', null);
  Test('Yi ≔ inv(Y);', null);
  Test('Zi ≔ inv(Z);', null);
  StopStopwatch;
  TestDuration(0.2);

  Test('IsIdentity(X⋅Xi)', True);
  Test('IsIdentity(Xi⋅X)', True);
  Test('IsIdentity(Y⋅Yi)', True);
  Test('IsIdentity(Yi⋅Y)', True);
  Test('IsIdentity(Z⋅Zi)', True);
  Test('IsIdentity(Zi⋅Z)', True);

  Test('inv(❨❨i, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 0❩❩)', failure, 'Matrix is singular.');
  Test('inv(❨❨i, 0, 0❩, ❨0, 1, 0❩❩)', failure, 'Cannot compute inverse of non-square matrix.');

  Test('A ≔ ❨❨3, 1❩, ❨-2, 4❩❩', 2, [3, 1, -2, 4]);
  Eps; Test('B ≔ A^−1', 2, [0.285714285714285714, -0.0714285714285714286, 0.142857142857142857, 0.214285714285714286]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨4, 2, 1❩, ❨4, −1, 2❩, ❨1, 5, 2❩❩', 3, [4, 2, 1, 4, -1, 2, 1, 5, 2]);
  Eps; Test('B ≔ A^−1', 3, [0.307692307692307692, -0.025641025641025641, -0.128205128205128205, 0.153846153846153846, -0.179487179487179487, 0.102564102564102564, -0.538461538461538462, 0.461538461538461538, 0.307692307692307692]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨5, 1, 2, 3❩, ❨6, 2, 1, −4❩, ❨9, 2, 4, 7❩, ❨4, 1, 6, 8❩❩', 4, [5, 1, 2, 3, 6, 2, 1, -4, 9, 2, 4, 7, 4, 1, 6, 8]);
  Eps; Test('B ≔ A^−1', 4, [1.38636363636363636, -0.0909090909090909091, -0.568181818181818182, -0.0681818181818181818, -5.81818181818181818, 0.545454545454545455, 2.90909090909090909, -0.090909090909090909, 0.863636363636363636, 0.0909090909090909091, -0.681818181818181818, 0.318181818181818182, -0.613636363636363636, -0.0909090909090909091, 0.431818181818181818, -0.0681818181818181818]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨5, 2, −6, 1, 2❩, ❨0, 2, 6, −7, 2❩, ❨1, 2, 7, 2, 3❩, ❨9, 8, −7, 1, 6❩, ❨5, 1, 2, 7, 6❩❩', 5, [5, 2, -6, 1, 2, 0, 2, 6, -7, 2, 1, 2, 7, 2, 3, 9, 8, -7, 1, 6, 5, 1, 2, 7, 6]);
  Eps; Test('B ≔ A^−1', 5, [0.963923337091319053, -0.00338218714768883875, 0.553175497933107854, -0.347237880496054115, -0.249530251785043217, -0.226606538895152198, -0.114994363021420519, 0.141300263059000376, 0.19391206313416009, -0.150695227358136039, 0.285231116121758737, -0.00450958286358511838, 0.293122886133032694, -0.129650507328072153, -0.1104847801578354, -0.0496054114994363021, -0.129650507328072153, 0.0939496429913566329, 0.0225479143179255919, -0.00977076287110108982, -0.802705749718151071, 0.174746335963923337, -0.691845170988350244, 0.273957158962795941, 0.447951897782788425]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨2 + 3⋅i, 1 − 2⋅i❩, ❨−4⋅i, 6 + 5⋅i❩❩', 2, [2 + 3*ImaginaryUnit, 1 - 2*ImaginaryUnit, -4*ImaginaryUnit, 6 + 5*ImaginaryUnit]);
  Eps; Test('B ≔ A^−1', 2, [0.181124880838894185 - 0.159199237368922784*ImaginaryUnit, 0.0562440419447092469 + 0.0400381315538608198*ImaginaryUnit, 0.122020972354623451 + 0.0190657769304099142*ImaginaryUnit, 0.101048617731172545 - 0.0467111534795042898*ImaginaryUnit]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨2, 1, i❩, ❨5, 1, 2❩, ❨−2, 5, −i❩❩', 3, [2, 1, ImaginaryUnit, 5, 1, 2, -2, 5, -ImaginaryUnit]);
  Eps; Test('B ≔ A^−1', 3, [0.142276422764227642 + 0.21951219512195122*ImaginaryUnit, 0.121951219512195122 - 0.0975609756097560976*ImaginaryUnit, -0.0528455284552845528 - 0.0243902439024390244*ImaginaryUnit, 0.166666666666666667 + 5.01945450224770571E-21*ImaginaryUnit, -5.01945450224770571E-21, 0.166666666666666667, -0.439024390243902439 - 0.548780487804878049*ImaginaryUnit, 0.195121951219512195 + 0.243902439024390244*ImaginaryUnit, 0.0487804878048780488 + 0.060975609756097561*ImaginaryUnit]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('A ≔ ❨❨5, 2, i, 1❩, ❨6, 2, 1, 2❩, ❨−4, 1, 2, i❩, ❨−i, 1, 5, 8❩❩', 4, [5, 2, ImaginaryUnit, 1, 6, 2, 1, 2, -4, 1, 2, ImaginaryUnit, -ImaginaryUnit, 1, 5, 8]);
  Eps; Test('B ≔ A^−1', 4, [-0.173183213920163767 - 0.142169907881269191*ImaginaryUnit, 0.242579324462640737 + 0.129989764585465711*ImaginaryUnit, -0.103070624360286592 + 0.026202661207778915*ImaginaryUnit, -0.035721596724667349 - 0.00184237461617195497*ImaginaryUnit, 0.50962128966223132 + 0.424564994882292733*ImaginaryUnit, -0.180143295803480041 - 0.396110542476970317*ImaginaryUnit, 0.366837256908904811 - 0.0570112589559877175*ImaginaryUnit, -0.0257932446264073695 + 0.000102354145342886399*ImaginaryUnit, -0.446878198567041965 - 0.61330603889457523*ImaginaryUnit, 0.430910951893551689 + 0.589559877175025589*ImaginaryUnit, 0.0679631525076765609 + 0.126714431934493347*ImaginaryUnit, -0.0360286591606960082 - 0.0792221084953940635*ImaginaryUnit, 0.233367451381780962 + 0.308597748208802456*ImaginaryUnit, -0.263050153531218014 - 0.288638689866939611*ImaginaryUnit, -0.0916069600818833163 - 0.0849539406345957011*ImaginaryUnit, 0.150972364380757421 + 0.0450358239508700102*ImaginaryUnit]);
  Test('IsIdentity(A⋅B)', True);
  Test('IsIdentity(B⋅A)', True);

  Test('delete(A); delete(B); delete(C); delete(Ai); delete(Bi); delete(Ci); delete(X); delete(Y); delete(Z); delete(Xi); delete(Yi); delete(Zi)', success);

  // Rank, nullity, singularity

  Test('A ≔ ❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩; B ≔ ❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 0❩❩; C ≔ ❨❨1, 0, 0❩, ❨0, 0, 0❩, ❨0, 0, 0❩❩; D ≔ ❨❨0, 0, 0❩, ❨0, 0, 0❩, ❨0, 0, 0❩❩;', null);
  Test('E ≔ ❨❨5, 6, 5❩, ❨2, 3, 2❩, ❨1, −2, 5❩❩; F ≔ ❨❨5, 6, 2❩, ❨2, 3, 2❩, ❨1, −2, −6❩❩; G ≔ ❨❨2, −6, 4❩, ❨−1, 3, −2❩, ❨3, −9, 6❩❩;', null);
  Test('H ≔ ❨❨2, 5❩, ❨1, 4❩, ❨7, 4❩❩; I ≔ ❨❨2, 4❩, ❨1, 2❩, ❨7, 14❩❩; J ≔ ❨❨3, −2, 7❩, ❨1, 4, 6❩❩; K ≔ ❨❨3, −2, −1❩, ❨1, 4, 5❩❩; L ≔ ❨❨3, 6, 9❩, ❨1, 2, 3❩❩;', null);
  Test('M ≔ ❨❨5, 2, 4, 7❩, ❨8, 5, 3, −5❩, ❨3, 6, 8, 2❩, ❨1, −1, 7, 4❩❩; N ≔ ❨❨5, 2, 14, 7❩, ❨8, 5, 8, −5❩, ❨3, 6, 11, 2❩, ❨1, −1, 4, 4❩❩; O ≔ ❨❨5, 2, 3, −2❩, ❨8, 5, 3, 4❩, ❨3, 6, −3, 18❩, ❨1, −1, 2, −6❩❩;'+' P ≔ ❨❨5, 10, 15, −5❩, ❨8, 16, 24, −8❩, ❨3, 6, 9, −3❩, ❨1, 2, 3, −1❩❩;', null);
  Test('Q ≔ ❨❨2, 5, 4, 1, 5, 8❩, ❨2, −5, 6, 2, 7, 0❩, ❨2, 6, 3, 7, 5, 8❩, ❨5, 6, 9, −1, 5, 7❩, ❨9, 3, −9, 2, 0, 1❩, ❨4, 8, 6, 8, 3, 5❩❩; '+'R ≔ ❨❨2, 5, 4, 1, 5, 8❩, ❨3, −5, 6, 2, 7, 0❩, ❨2, 6, 3, 0, 5, 8❩, ❨5, 6, 9, 9, 5, 7❩, ❨9, 3, −9, 0, 0, 1❩, ❨4, 8, 6, 7, 3, 5❩❩; '+'S ≔ ❨❨2, 4, 4, 1, 5, 8❩, ❨3, −5, 6, 2, 7, 0❩, ❨2, 3, 3, 0, 5, 8❩, ❨5, 11, 9, 9, 5, 7❩, ❨9, 1, −9, 0, 0, 1❩, ❨4, 9, 6, 7, 3, 5❩❩;', null);

  Test('rank(A)', 3);
  Test('nullity(A)', 0);
  Test('IsSingular(A)', False);

  Test('rank(B)', 2);
  Test('nullity(B)', 1);
  Test('IsSingular(B)', True);

  Test('rank(C)', 1);
  Test('nullity(C)', 2);
  Test('IsSingular(C)', True);

  Test('rank(D)', 0);
  Test('nullity(D)', 3);
  Test('IsSingular(D)', True);

  Test('rank(E)', 3);
  Test('nullity(E)', 0);
  Test('IsSingular(E)', False);

  Test('rank(F)', 2);
  Test('nullity(F)', 1);
  Test('IsSingular(F)', True);

  Test('rank(G)', 1);
  Test('nullity(G)', 2);
  Test('IsSingular(G)', True);

  Test('rank(H)', 2);
  Test('nullity(H)', 0);
  Test('IsSingular(H)', failure, 'Matrix is not square.');

  Test('rank(I)', 1);
  Test('nullity(I)', 1);
  Test('IsSingular(I)', failure, 'Matrix is not square.');

  Test('rank(J)', 2);
  Test('nullity(J)', 1);
  Test('IsSingular(J)', failure, 'Matrix is not square.');

  Test('rank(K)', 2);
  Test('nullity(K)', 1);
  Test('IsSingular(K)', failure, 'Matrix is not square.');

  Test('rank(L)', 1);
  Test('nullity(L)', 2);
  Test('IsSingular(L)', failure, 'Matrix is not square.');

  Test('rank(M)', 4);
  Test('nullity(M)', 0);
  Test('IsSingular(M)', False);

  Test('rank(N)', 3);
  Test('nullity(N)', 1);
  Test('IsSingular(N)', True);

  Test('rank(O)', 2);
  Test('nullity(O)', 2);
  Test('IsSingular(O)', True);

  Test('rank(P)', 1);
  Test('nullity(P)', 3);
  Test('IsSingular(P)', True);

  Test('rank(Q)', 6);
  Test('nullity(Q)', 0);
  Test('IsSingular(Q)', False);

  Test('rank(R)', 5);
  Test('nullity(R)', 1);
  Test('IsSingular(R)', True);

  Test('rank(S)', 4);
  Test('nullity(S)', 2);
  Test('IsSingular(S)', True);

  Test('delete(A); delete(B); delete(C); delete(D); delete(E); delete(F); delete(G); delete(H); delete(I); delete(J); delete(K); delete(L); delete(M); delete(N); delete(O); delete(P); delete(Q); delete(R); delete(S)', success);

  Test('A ≔ ' + MatLit_100x100_Rand_A + ';', null);
  Test('B ≔ ' + MatLit_100x100_Rand_B + ';', null);
  Test('C ≔ ' + MatLit_100x100_Rand_C + ';', null);

  StartStopwatch;

  Test('rank(A)', 100);
  Test('nullity(A)', 0);
  Test('IsSingular(A)', False);

  Test('rank(B)', 100);
  Test('nullity(B)', 0);
  Test('IsSingular(B)', False);

  Test('rank(C)', 100);
  Test('nullity(C)', 0);
  Test('IsSingular(C)', False);

  StopStopwatch;
  TestDuration(0.15);

  Test('col(A, 3) ≔ col(A, 1) + col(A, 2);', null);
  Test('rank(A)', 99);
  Test('nullity(A)', 1);
  Test('IsSingular(A)', True);

  Test('A ≔ ❨❨i, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩; B ≔ ❨❨i, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 0❩❩; C ≔ ❨❨i, 0, 0❩, ❨0, 0, 0❩, ❨0, 0, 0❩❩; D ≔ ComplexMatrix(❨❨0, 0, 0❩, ❨0, 0, 0❩, ❨0, 0, 0❩❩);', null);
  Test('E ≔ ❨❨5, 6, 5❩, ❨2, 3, 2❩, ❨i, −2, 5❩❩; F ≔ ❨❨5, 6, 1❩, ❨2, 3, 1❩, ❨i, −2, −2−i❩❩; G ≔ ❨❨2, −6, 4❩, ❨−1, 3, −2❩, ❨i, −3⋅i, 2⋅i❩❩;', null);
  Test('H ≔ ❨❨2, 5❩, ❨i, 4❩, ❨7, 4❩❩; I ≔ ❨❨2, 4❩, ❨i, 2⋅i❩, ❨7, 14❩❩; J ≔ ❨❨3, −2, 7❩, ❨i, 4, 6❩❩; K ≔ ❨❨3, −2, 1❩, ❨i, 4, 4+i❩❩; L ≔ ❨❨3, 6, 9❩, ❨i, 2⋅i, 3⋅i❩❩;', null);
  Test('M ≔ ❨❨5, 2, 4, 7❩, ❨8, 5, 3, −5❩, ❨3, 6, 8, 2❩, ❨i, −1, 7, 4❩❩; N ≔ ❨❨5, 2, 14, 7❩, ❨8, 5, 8, −5❩, ❨3, 6, 11, 2❩, ❨i, −1, 3+i, 4❩❩; O ≔ ❨❨5, 2, 7, 3❩, ❨8, 5, 13, 3❩, ❨3, 6, 9, −3❩, ❨i, −1, −1+i, 1+i❩❩;'+' P ≔ ❨❨5, 10, 15, −5❩, ❨8, 16, 24, −8❩, ❨3, 6, 9, −3❩, ❨i, 2⋅i, 3⋅i, −i❩❩;', null);
  Test('Q ≔ ❨❨2, 5, 4, i, 5, 8❩, ❨3, −5, 6, 2, 7, 0❩, ❨2, 6, 3, 7, 5, 8❩, ❨5, 6, 9, −1, 5, 7❩, ❨9, 3, −9, 2, 0, 1❩, ❨4, 8, 6, 8, 3, 5❩❩; '+'R ≔ ❨❨2, 5, 4, i, 4+i, 8❩, ❨3, -5, 6, 2, 8, 0❩, ❨2, 6, 3, 7, 10, 8❩, ❨5, 6, 9, -1, 8, 7❩, ❨9, 3, −9, 2, -7, 1❩, ❨4, 8, 6, 8, 14, 5❩❩; '+'S ≔ ❨❨2, 5, 4, i, 4+i, 7❩, ❨3, −5, 6, 2, 8, -2❩, ❨2, 6, 3, 7, 10, 8❩, ❨5, 6, 9, -1, 8, 11❩, ❨9, 3, −9, 2, -7, 12❩, ❨4, 8, 6, 8, 14, 12❩❩;', null);

  Test('rank(A)', 3);
  Test('nullity(A)', 0);
  Test('IsSingular(A)', False);

  Test('rank(B)', 2);
  Test('nullity(B)', 1);
  Test('IsSingular(B)', True);

  Test('rank(C)', 1);
  Test('nullity(C)', 2);
  Test('IsSingular(C)', True);

  Test('rank(D)', 0);
  Test('nullity(D)', 3);
  Test('IsSingular(D)', True);

  Test('rank(E)', 3);
  Test('nullity(E)', 0);
  Test('IsSingular(E)', False);

  Test('rank(F)', 2);
  Test('nullity(F)', 1);
  Test('IsSingular(F)', True);

  Test('rank(G)', 1);
  Test('nullity(G)', 2);
  Test('IsSingular(G)', True);

  Test('rank(H)', 2);
  Test('nullity(H)', 0);
  Test('IsSingular(H)', failure, 'Matrix is not square.');

  Test('rank(I)', 1);
  Test('nullity(I)', 1);
  Test('IsSingular(I)', failure, 'Matrix is not square.');

  Test('rank(J)', 2);
  Test('nullity(J)', 1);
  Test('IsSingular(J)', failure, 'Matrix is not square.');

  Test('rank(K)', 2);
  Test('nullity(K)', 1);
  Test('IsSingular(K)', failure, 'Matrix is not square.');

  Test('rank(L)', 1);
  Test('nullity(L)', 2);
  Test('IsSingular(L)', failure, 'Matrix is not square.');

  Test('rank(M)', 4);
  Test('nullity(M)', 0);
  Test('IsSingular(M)', False);

  Test('rank(N)', 3);
  Test('nullity(N)', 1);
  Test('IsSingular(N)', True);

  Test('rank(O)', 2);
  Test('nullity(O)', 2);
  Test('IsSingular(O)', True);

  Test('rank(P)', 1);
  Test('nullity(P)', 3);
  Test('IsSingular(P)', True);

  Test('rank(Q)', 6);
  Test('nullity(Q)', 0);
  Test('IsSingular(Q)', False);

  Test('rank(R)', 5);
  Test('nullity(R)', 1);
  Test('IsSingular(R)', True);

  Test('rank(S)', 4);
  Test('nullity(S)', 2);
  Test('IsSingular(S)', True);

  Test('A ≔ ' + MatLit_100x100_Rand_A + ';', null);
  Test('B ≔ ' + MatLit_100x100_Rand_B + ';', null);
  Test('C ≔ ' + MatLit_100x100_Rand_C + ';', null);

  Test('X ≔ A + B⋅i;', null);
  Test('Y ≔ A + C⋅i;', null);
  Test('Z ≔ B + C⋅i;', null);

  Test('rank(X)', 100);
  Test('nullity(X)', 0);
  Test('IsSingular(X)', False);

  Test('rank(Y)', 100);
  Test('nullity(Y)', 0);
  Test('IsSingular(Y)', False);

  Test('rank(Z)', 100);
  Test('nullity(Z)', 0);
  Test('IsSingular(Z)', False);

  Test('col(X, 85) ≔ col(X, 31) + col(X, 57);', null);
  Test('rank(X)', 99);
  Test('nullity(X)', 1);
  Test('IsSingular(X)', True);

  Test('delete(A); delete(B); delete(C); delete(D); delete(E); delete(F); delete(G); delete(H); delete(I); delete(J); delete(K); delete(L); delete(M); delete(N); delete(O); delete(P); delete(Q); delete(R); delete(S); delete(X); delete(Y); delete(Z)', success);

  // Matrix square root

  Eps; Test('√❨❨97, 52, 8❩, ❨52, 36, 18❩, ❨8, 18, 29❩❩', 3, [9, 4, 0, 4, 4, 2, 0, 2, 5]);
  Eps; Test('√❨❨254, 223, 159❩, ❨223, 326, 309❩, ❨159, 309, 414❩❩', 3, [14, 7, 3, 7, 14, 9, 3, 9, 18]);
  Eps(1E-15); Test('√❨❨78, 29, -110❩, ❨29, 174, 10❩, ❨-110, 10, 171❩❩', 3, [5, 2, -7, 2, 13, 1, -7, 1, 11]);
  Eps; Test('√❨❨110, 43, 81❩, ❨43, 38, 43❩, ❨81, 43, 70❩❩', 3, [9, 2, 5, 2, 5, 3, 5, 3, 6]);

  Eps; Test('√❨❨3851, -3325, 2089, -1042, -228, -1852❩, ❨-3325, 8385, -239, -2309, 2257, 2174❩, ❨2089, -239, 9449, -477, 1633, -432❩, ❨-1042, -2309, -477, 3079, -170, -1546❩, ❨-228, 2257, 1633, -170, 2070, -1689❩, ❨-1852, 2174, -432, -1546, -1689, 5123❩❩',
    6, [46, -27, 15, -21, -4, -18, -27, 81, -1, -21, 22, 13, 15, -1, 95, -1, 14, 1, -21, -21, -1, 44, -2, -16, -4, 22, 14, -2, 29, -23, -18, 13, 1, -16, -23, 62]);

  Eps; Test('√❨❨2.7188, 1.673, -2.6948, -1.7114, -2.4302, -3.0205❩, ❨1.673, 3.0471, -1.7734, -0.1466, -1.7598, -1.1892❩, ❨-2.6948, -1.7734, 2.9028, 1.5929, 2.7186, 3.4337❩, ❨-1.7114, -0.1466, 1.5929, 1.7656, 1.295, 1.5489❩, '+'❨-2.4302, -1.7598, 2.7186, 1.295, 5.3695, 6.4108❩, ❨-3.0205, -1.1892, 3.4337, 1.5489, 6.4108, 9.1433❩❩',
    6, [1.03, 0.5, -0.77, -0.63, -0.34, -0.55, 0.5, 1.52, -0.49, 0.21, -0.45, 0, -0.77, -0.49, 1.12, 0.47, 0.43, 0.64, -0.63, 0.21, 0.47, 1, 0.29, 0.14, -0.34, -0.45, 0.43, 0.29, 1.68, 1.4, -0.55, 0, 0.64, 0.14, 1.4, 2.54]);

  Test('√❨❨4, 1, 2❩, ❨6, 2, 2❩, ❨6, 7, 2❩❩', failure, 'Matrix square root only defined for positive semidefinite symmetric matrices.');

  Eps; Test('√❨❨359, 165 - 231⋅i, -115 + 184⋅i❩, ❨165 + 231⋅i, 435, -81 + 5⋅i❩, ❨-115 - 184⋅i, -81 - 5⋅i, 170❩❩',
    3, [14, 5 - 7*ImaginaryUnit, -5 + 8*ImaginaryUnit, 5 + 7*ImaginaryUnit, 19, 0, -5 - 8*ImaginaryUnit, 0, 9]);

  Eps; Test('√❨❨26.4021, 0.2278 + 10.1106⋅i, 13.3148 + 23.222⋅i, 5.0529 - 20.124⋅i, -7.901 - 14.2179⋅i, -13.1404 + 0.042⋅i❩, ❨0.2278 - 10.1106⋅i, 9.4228, 8.4996 - 2.7067⋅i, -13.4986 - 6.9818⋅i, -6.7041 + 1.6279⋅i, 8.5688 + 5.9575⋅i❩, '+'❨13.3148 - 23.222⋅i, 8.4996 + 2.7067⋅i, 32.7353, -21.7024 - 7.3565⋅i, -22.3923 + 0.4343⋅i, -2.345 + 4.8311⋅i❩, ❨5.0529 + 20.124⋅i, -13.4986 + 6.9818⋅i, -21.7024 + 7.3565⋅i, 40.997, 19.3339 - 0.0229999999999999992⋅i, -22.8175 - 2.0227⋅i❩, '+'❨-7.901 + 14.2179⋅i, -6.7041 - 1.6279⋅i, -22.3923 - 0.4343⋅i, 19.3339 + 0.0229999999999999992⋅i, 20.7363, -2.6089 - 0.8045⋅i❩, ❨-13.1404 - 0.042⋅i, 8.5688 - 5.9575⋅i, -2.345 - 4.8311⋅i, -22.8175 + 2.0227⋅i, -2.6089 + 0.8045⋅i, 33.2686❩❩',
    6, [3.27, 0.08 + 0.88*ImaginaryUnit, 1.52 + 2.23*ImaginaryUnit, 0.7 - 1.8*ImaginaryUnit, -0.98 - 1.03*ImaginaryUnit, -1.27 - 0.53*ImaginaryUnit, 0.08 - 0.88*ImaginaryUnit, 2.06, 0.43 - 0.25*ImaginaryUnit, -1.23 - 1.02*ImaginaryUnit, -0.62 + 0.54*ImaginaryUnit, 0.89 + 0.36*ImaginaryUnit, 1.52 - 2.23*ImaginaryUnit, 0.43 + 0.25*ImaginaryUnit, 4.14, -1.78 - 0.43*ImaginaryUnit, -2.11 + 0.24*ImaginaryUnit, -0.42 + 0.16*ImaginaryUnit, 0.7 + 1.8*ImaginaryUnit, -1.23 + 1.02*ImaginaryUnit, -1.78 + 0.43*ImaginaryUnit, 4.77, 1.77 + 0.67*ImaginaryUnit, -2.24 + 0.09*ImaginaryUnit, -0.98 + 1.03*ImaginaryUnit, -0.62 - 0.54*ImaginaryUnit, -2.11 - 0.24*ImaginaryUnit, 1.77 - 0.67*ImaginaryUnit, 3.15, -0.13 - 0.09*ImaginaryUnit, -1.27 + 0.53*ImaginaryUnit, 0.89 - 0.36*ImaginaryUnit, -0.42 - 0.16*ImaginaryUnit, -2.24 - 0.09*ImaginaryUnit, -0.13 + 0.09*ImaginaryUnit, 5.02]);

  // Hermitian square and modulus

  Test('HermitianSquare(❨❨3, 5, 1❩, ❨2, 1, 5❩❩)', 3, [13, 17, 13, 17, 26, 10, 13, 10, 26]);
  Test('HermitianSquare(❨❨6, 1, 3, 5❩❩)', 4, [36, 6, 18, 30, 6, 1, 3, 5, 18, 3, 9, 15, 30, 5, 15, 25]);
  Test('HermitianSquare(❨❨5, 1, 2, 3❩, ❨6, 2, 1, 1❩, ❨0, 6, 7, −3❩, ❨5, 1, 3, 8❩❩)', 4, [86, 22, 31, 61, 22, 42, 49, -5, 31, 49, 63, 10, 61, -5, 10, 83]);

  Test('HermitianSquare(❨❨3, i❩, ❨1, −i❩, ❨4, 7❩❩)', 2, [26, 28 + 2*ImaginaryUnit, 28 - 2*ImaginaryUnit, 51]);
  Test('HermitianSquare(❨❨6, 1 + i, 2 − 3⋅i❩, ❨6, 1, i❩❩)', 3, [72, 12+6*ImaginaryUnit, 12-12*ImaginaryUnit, 12-6*ImaginaryUnit, 3, -1-4*ImaginaryUnit, 12+12*ImaginaryUnit, -1+4*ImaginaryUnit, 14]);
  Test('HermitianSquare(❨❨1, 4, i❩, ❨4, −2, 1❩, ❨0, 1, 5❩❩)', 3, [17, -4, 4+ImaginaryUnit, -4, 21, 3+4*ImaginaryUnit, 4-ImaginaryUnit, 3-4*ImaginaryUnit, 27]);

  Test('A ≔ ❨❨4, 3, 5❩, ❨7, −3, 1❩, ❨1, 4, 8❩❩', 3, [4, 3, 5, 7, -3, 1, 1, 4, 8]);
  Eps; Test('M ≔ modulus(A)', 3, [7.66161489598277892, -1.15637609417388669, 2.44181316125495937, -1.15637609417388669, 4.38980424381820705, 3.6595645956559922, 2.44181316125495937, 3.6595645956559922, 8.4050660589755781]);
  Test('IsPositiveSemidefinite(M)', True);
  Test('M^2 ≈ A*⋅A', True);
  Test('norm(A) ≈ norm(M)', True);
  Test('abs(det(A)) ≈ det(M)', True);

  Test('A ≔ ❨❨5, 2, 3, 1❩, ❨−4, 7, 5, 9❩, ❨2, 0, −3, 4❩, ❨4, 4, 7, 5❩❩', 4, [5, 2, 3, 1, -4, 7, 5, 9, 2, 0, -3, 4, 4, 4, 7, 5]);
  Eps; Test('M ≔ modulus(A)', 4, [7.69100520985240193, -0.46899091993611459, 1.251703432597207, -0.248444955402715359, -0.46899091993611459, 5.13206129403951794, 4.19592279964112954, 4.98359571501246989, 1.251703432597207, 4.19592279964112954, 8.03390062336959419, 2.87817844306042996, -0.248444955402715359, 4.98359571501246989, 2.87817844306042996, 9.47724314889865696]);
  Test('IsPositiveSemidefinite(M)', True);
  Test('M^2 ≈ A*⋅A', True);
  Test('norm(A) ≈ norm(M)', True);
  Test('abs(det(A)) ≈ det(M)', True);

  Test('A ≔ ❨❨5, 1, 2❩, ❨6, 2, 8❩❩', 3, [5, 1, 2, 6, 2, 8]);
  Eps; Test('M ≔ modulus(A)', 3, [6.43800469034064263, 1.51325427264365605, 4.15477521816494974, 1.51325427264365605, 0.442691589144319934, 1.58558685136658348, 4.15477521816494974, 1.58558685136658348, 6.94433274140113465]);
  Test('IsPositiveSemidefinite(M)', True);
  Test('M^2 ≈ A*⋅A', True);
  Test('norm(A) ≈ norm(M)', True);
  Test('det(M)', 0.0);

  Test('A ≔ ❨❨5, 3 + i, 2❩, ❨1, 7, 5❩, ❨−i, 3, 4❩❩;', null);
  Eps; Test('M ≔ modulus(A)', 3, [4.83787163654284113, 1.59310281607896703 + 0.620663163540410443*ImaginaryUnit, 0.805031261583889188 + 0.154023865138550882*ImaginaryUnit, 1.59310281607896703 - 0.620663163540410443*ImaginaryUnit, 6.77228650762633288, 4.38075209080517553 - 0.148145983251944481*ImaginaryUnit, 0.805031261583889188 - 0.154023865138550882*ImaginaryUnit, 4.38075209080517553 + 0.148145983251944481*ImaginaryUnit, 5.01151326481261855]);
  Test('IsPositiveSemidefinite(M)', True);
  Test('M^2 ≈ A*⋅A', True);
  Test('norm(A) ≈ norm(M)', True);
  Test('abs(det(A)) ≈ det(M)', True);


  //
  // Eigenvalues and eigenvectors
  //

  // Eigenvalues

  // - eigenvalues of real symmetric matrices

  Test('A ≔ ❨❨21❩❩;', null);
  Test('Λ ≔ eigenvalues(A)', [21]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨3, 5❩, ❨5, 2❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [7.524937810560445, -2.524937810560445]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨4, 1, 8❩, ❨1, 2, 5❩, ❨8, 5, 3❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [13.136085649145485, -5.802319968129476, 1.666234318983990]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨2, 0, 1❩, ❨0, 2, 0❩, ❨1, 0, 2❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [3, 2, 1]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨0, 0, 2❩, ❨0, 2, 0❩, ❨2, 0, 0❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [-2, 2, 2]);
  Test('∏(Λ) ≈ det(A)', True);               {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨6, 1, 3, 2❩, ❨1, 8, 5, 1❩, ❨3, 5, 2, 0❩, ❨2, 1, 0, 7❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [12.317576522344481, 7.443877742227379, 4.830339943837674, -1.591794208409535]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨0, 1, 0, 2❩, ❨1, -2, 2, 0❩, ❨0, 2, 0, -1❩, ❨2, 0, -1, 2❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [-3.449489742783178, 3.449489742783178, -1.449489742783178, 1.449489742783178]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨0, 1, 0, 2❩, ❨1, -2, 2, 0❩, ❨0, 2, 0, -1❩, ❨2, 0, -1, -2❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [-3.449489742783178, -3.449489742783178, 1.449489742783178, 1.449489742783178]);
  Test('∏(Λ) ≈ det(A)', True);            {2}                                    {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨0, 0, 1, 2❩, ❨0, -2, 0, 0❩, ❨1, 0, 0, 1❩, ❨2, 0, 1, 0❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [2.732050807568878, -2, -2, -0.732050807568878]);
  Test('∏(Λ) ≈ det(A)', True);                               {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨3.09090210854135331, -0.0989480787770248203, 0.206696174292443883, -0.173565216686099517❩, '+'❨-0.0989480787770248203, 3.10770621771892455, -0.224991363401589891, 0.188927902874714558❩, ❨0.206696174292443883, -0.224991363401589891, 3.4699924914029535, -0.394658241210492851❩, '+'❨-0.173565216686099517, 0.188927902874714558, -0.394658241210492851, 3.33139918233676864❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [4, 3, 3, 3]);
  Test('∏(Λ) ≈ det(A)', True);              {3}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨3, 2, 1, 0, 1❩, ❨2, 5, 7, 3, 8❩, ❨1, 7, 6, 0, 2❩, ❨0, 3, 0, 5, 3❩, ❨1, 8, 2, 3, 9❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [19.613776982166058, 6.482010475356414, -3.831681427313180, 3.142965451970841, 2.592928517819864]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨2, 0, 1, 0, 0❩, ❨0, 0, 0, 2, 0❩, ❨1, 0, 0, 1, 1❩, ❨0, 2, 1, 0, 0❩, ❨0, 0, 1, 0, 2❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [2.895106515927531, -2.292401585224621, 2, 2, -0.602704930702910]);
  Test('∏(Λ) ≈ det(A)', True);                                                  {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨0, 1, 1, 0, 0❩, ❨1, 0, 0, 0, 1❩, ❨1, 0, 0, 1, 0❩, ❨0, 0, 1, 0, 1❩, ❨0, 1, 0, 1, 0❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [2, -1.618033988749895, -1.618033988749895, 0.618033988749895, 0.618033988749895]);
  Test('∏(Λ) ≈ det(A)', True);               {2}                                    {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨4.88518185435250995, -0.0503410125523369264, 2.41500433970266516, -0.3526553784998275, 0.377477078463321959❩, '+'❨-0.0503410125523369266, 6.99880168536003428, 0.0574866278829369505, -0.00839458885495496514, 0.00898544320902995499❩, ❨2.41500433970266516, 0.0574866278829369505, 4.24219977363724693, 0.402712768116436254, -0.431057764708227601❩, '+'❨-0.3526553784998275, -0.00839458885495496536, 0.402712768116436254, 6.94119313935299163, 0.0629459900627696699❩, ❨0.377477078463321959, 0.00898544320902995499, -0.431057764708227601, 0.0629459900627696697, 6.93262354729721721❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [7, 7, 7, 7, 2]);
  Test('∏(Λ) ≈ det(A)', True);           {4}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨2, −1, 0, 3, 8, 2❩, ❨−1, 7, 5, 4, 8, 1❩, ❨0, 5, 2, 9, 3, 0❩, ❨3, 4, 9, −7, 2, 5❩, ❨8, 8, 3, 2, 1, 3❩, ❨2, 1, 0, 5, 3, 8❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [20.369654342915162, -14.157794981021667, -8.732880602208134, 8.619138715827635, 5.073173708519922, 1.828708815967091]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨10, 16, 9, 8, 0, 7, 8❩, ❨16, 12, 5, 10, 4, 11, 14❩, ❨9, 5, 2, 12, 12, 5, 14❩, ❨8, 10, 12, 16, 12, 9, 3❩, ❨0, 4, 12, 12, 16, 17, 7❩, ❨7, 11, 5, 9, 17, 18, 7❩, ❨8, 14, 14, 3, 7, 7, 14❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [67.387186371351689, 22.239241106426356, -13.887324208091441, 10.960115323234580, 9.767062366458459, -5.676556998202810, -2.789723961176876]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨10, 5, 14, 10, 9, 13, 9, 15❩, ❨5, 4, 8, 7, 7, 4, 9, 13❩, ❨14, 8, 4, 6, 7, 5, 1, 14❩, ❨10, 7, 6, 18, 13, 12, 5, 11❩, ❨9, 7, 7, 13, 14, 12, 3, 14❩, ❨13, 4, 5, 12, 12, 18, 10, 15❩, ❨9, 9, 1, 5, 3, 10, 14, 14❩, ❨15, 13, 14, 11, 14, 15, 14, 6❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [79.906455303672033, 15.007966465439917, -14.248879078663860, 10.321739350164814, -10.196871321023851, 6.862620180127339, 3.257195862090613, -2.910226761807029]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨10, 7, 11, 13, 15, 11, 9, 17, 7❩, ❨7, 8, 12, 3, 6, 10, 14, 7, 8❩, ❨11, 12, 18, 4, 11, 7, 9, 11, 2❩, ❨13, 3, 4, 0, 4, 9, 16, 11, 13❩, ❨15, 6, 11, 4, 4, 6, 11, 4, 11❩, ❨11, 10, 7, 9, 6, 8, 8, 10, 14❩, ❨9, 14, 9, 16, 11, 8, 10, 11, 13❩, '+'❨17, 7, 11, 11, 4, 10, 11, 14, 9❩, ❨7, 8, 2, 13, 11, 14, 13, 9, 6❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [86.264296094535837, -19.275500362885520, 16.623302682936547, -12.030952595962669, 10.184197789201297, -6.930309037497170, 4.449115111645037, -3.598599758479221, 2.314450076505832]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨6, 6, 8, 14, 12, 14, 5, 9, 12, 10❩, ❨6, 0, 4, 4, 10, 8, 15, 14, 13, 11❩, ❨8, 4, 10, 6, 3, 11, 11, 9, 9, 13❩, ❨14, 4, 6, 10, 9, 17, 16, 9, 11, 8❩, ❨12, 10, 3, 9, 10, 8, 12, 10, 9, 8❩, ❨14, 8, 11, 17, 8, 2, 4, 12, 12, 14❩, '+'❨5, 15, 11, 16, 12, 4, 0, 9, 14, 11❩, ❨9, 14, 9, 9, 10, 12, 9, 18, 9, 8❩, ❨12, 13, 9, 11, 9, 12, 14, 9, 2, 7❩, ❨10, 11, 13, 8, 8, 14, 11, 8, 7, 6❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [96.530449307303996, -23.793421334022021, -16.961843873973226, -11.581585212083253, 11.447090905627016, 9.824786945678492, 5.683157788780861, -4.401418812239814, -3.708291806892070, 0.961076091819928]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨4.45528983757093495, -0.0878217640871465801, 0.587220808461187196, 0.381484594913282418, 0.806374962066634831, 0.415539821124804545, -1.07626065503415, -0.174476853005194865, 0.6117764065082528, 0.479713560928682166❩, '+'❨-0.0878217640871465801, 4.39333906061644079, -1.16146861964970503, 0.189604671732296926, 0.549866251807311702, 0.959584846946758171, -0.559867163033034075, 0.601342029002866998, 0.908395415505047237, -0.0481674127191538071❩, '+
    '❨0.587220808461187195, -1.16146861964970503, 4.731699905531466, -0.297368715494811859, 0.174137444046764151, 0.158862374235097095, 0.0708587702680495379, 0.527044617773810588, 0.717806236876411773, 0.100941787359586736❩, '+'❨0.381484594913282417, 0.189604671732296926, -0.297368715494811859, 5.20093245260149577, -0.422111106264731792, -0.609505049911372394, 0.430772405093700426, -0.075450459824149398, -0.0231177557010034982, 1.02271233462064678❩, '+
    '❨0.806374962066634831, 0.549866251807311702, 0.174137444046764151, -0.422111106264731792, 4.65483119820909985, 0.289795090828546352, 0.795841033040542465, -0.226199611183290607, -1.41597888676298831, 0.547038811317080901❩, '+'❨0.415539821124804545, 0.959584846946758171, 0.158862374235097095, -0.609505049911372394, 0.289795090828546352, 3.97252873681661176, 0.586614812369869111, -0.203645718668121608, 0.548222506108191402, -0.0260783651922211985❩, '+
    '❨-1.07626065503415, -0.559867163033034075, 0.0708587702680495379, 0.430772405093700426, 0.795841033040542465, 0.586614812369869111, 5.07303988296985066, 0.0852567827171805209, 0.721798919103536904, 0.0913845737686953809❩, '+'❨-0.174476853005194865, 0.601342029002866998, 0.527044617773810588, -0.0754504598241493979, -0.226199611183290607, -0.203645718668121608, 0.0852567827171805208, 5.72552879232507247, -0.381852810701084683, 0.248741195041378638❩, '+'❨0.6117764065082528, 0.908395415505047237, 0.717806236876411773, -0.0231177557010034982, -1.41597888676298831, 0.548222506108191402, 0.721798919103536904, -0.381852810701084683, 4.15028770968589867, 0.0862862168111860338❩, '+'❨0.479713560928682166, -0.0481674127191538071, 0.100941787359586736, 1.02271233462064678, 0.547038811317080901, -0.0260783651922211984, 0.0913845737686953809, 0.248741195041378638, 0.0862862168111860338, 3.64252242367312907❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [6, 6, 6, 6, 6, 6, 3, 3, 3, 1]);
  Test('∏(Λ) ≈ det(A)', True);           {6}               {3}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ReversalMatrix(10);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [-1, -1, -1, -1, -1, 1, 1, 1, 1, 1]);
  Test('∏(Λ) ≈ det(A)', True);            {5}                {5}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨0.947217218464429805, -0.0797032009555874838, 0.455681246175844459, -0.890506530889571131, 0.0810219664601976321, -0.4241414032969027, -0.325611817072624359, 0.442131323976784809, 0.696001040652687542, -0.363371579820000821❩, '+'❨-0.0797032009555874838, 0.664262086761071853, 0.151183874385912782, 0.366175763890158299, -0.362499363441370523, 0.345609605325798349, -0.36153060182114391, -0.370383371480650959, -0.351151983839811129, -0.192995673028588803❩, '+
    '❨0.455681246175844459, 0.151183874385912782, 0.965112845520727124, -0.379442604818156945, -0.0868548299522641688, -0.703208127497219295, -0.29273075622571167, 0.392533036681992209, 0.433785161673175889, -0.0198556703071645884❩, '+'❨-0.890506530889571131, 0.366175763890158299, -0.379442604818156945, 1.60196556896465841, 0.0343051071296422081, 0.587093406299823787, -0.485735414772669726, -0.766652742695679709, -0.63672778172637717, -0.0824745994586258031❩, '+
    '❨0.0810219664601976321, -0.362499363441370523, -0.0868548299522641688, 0.0343051071296422081, 0.327534361557092576, -0.222948382709437083, -0.0500390113869091228, 0.161756725513035523, 0.21826384416565025, -0.135644537928391234❩, '+'❨-0.4241414032969027, 0.345609605325798349, -0.703208127497219295, 0.587093406299823787, -0.222948382709437083, 0.989037781112786131, -0.126211410841058615, -0.699854415829559306, -0.413031006405435901, 0.131956448780506888❩, '+
    '❨-0.325611817072624359, -0.36153060182114391, -0.29273075622571167, -0.485735414772669726, -0.0500390113869091228, -0.126211410841058615, 1.00217393220066957, 0.281471732316985816, -0.335032854995363073, 0.398331000774434439❩, '+'❨0.442131323976784809, -0.370383371480650959, 0.392533036681992209, -0.766652742695679709, 0.161756725513035523, -0.699854415829559306, 0.281471732316985816, 0.61286300125401041, 0.371247429623465179, -0.0611706531297214894❩, '+
    '❨0.696001040652687542, -0.351151983839811129, 0.433785161673175889, -0.63672778172637717, 0.21826384416565025, -0.413031006405435901, -0.335032854995363073, 0.371247429623465179, 0.960656361843560564, 0.149752326353616183❩, '+'❨-0.363371579820000821, -0.192995673028588803, -0.0198556703071645884, -0.0824745994586258031, -0.135644537928391234, 0.131956448780506888, 0.398331000774434439, -0.0611706531297214894, 0.149752326353616183, 0.929176842320993553❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [4, 2, 1, 1, 1, 0, 0, 0, 0, 0]);
  Test('∏(Λ) ≈ det(A)', True);                 {3}      {5}
  Test('∑(Λ) ≈ tr(A)', True);

  // - eigenvalues of real non-symmetric matrices

  Test('A ≔ ❨❨1, 2❩, ❨3, 4❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [5.372281323269014, -0.372281323269014]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨5, 7, 4❩, ❨0, 2, 6❩, ❨1, 5, 3❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [9.334366656868244, 3.386344569047903, -2.720711225916147]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 0, 2❩, ❨1, 0, -1❩, ❨1, 1, 2❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [3, ImaginaryUnit, -ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 1, 0❩, ❨1, 0, -1❩, ❨1, 2, 1❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [1 + ImaginaryUnit, 1 - ImaginaryUnit, 0]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 0, 1❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [1, 1, 0]);
  Test('∏(Λ) ≈ det(A)', True);           {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 0, 2❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [2, 1, 0]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 1, 2❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [1, 1, 1]);
  Test('∏(Λ) ≈ det(A)', True);           {3}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨5, 4, 7, 2❩, ❨6, 3, 5, 7❩, ❨0, 5, 7, 8❩, ❨3, 4, 1, 9❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [18.722583143817452, 3.494347146728337 + 3.992775601519464*ImaginaryUnit, 3.494347146728337 - 3.992775601519464*ImaginaryUnit, -1.711277437274135]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨0, 0, 1, 0❩, ❨0, 0, 0, 0❩, ❨1, 0, 0, 0❩, ❨0, 0, 0, 2❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [2, -1, 1, 0]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨5, 8, 4, 0, 2❩, ❨6, 3, 0, 4, 8❩, ❨6, 5, 4, 9, 2❩, ❨5, 6, 3, 8, 7❩, ❨1, 0, 2, 9, 7❩❩;', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [22.814428493209221, 4.634582172214291, -3.664877452474182, 1.607933393525328 + 2.605736529318042*ImaginaryUnit, 1.607933393525328 - 2.605736529318042*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨5, 4, 7, 0, 2, 5❩, ❨6, 5, 8, 9, 5, 1❩, ❨0, 5, 8, 2, 3, 6❩, ❨4, 5, 8, 4, 6, 2❩, ❨7, 5, 9, 6, 1, 0❩, ❨5, 4, 7, 5, 2, 6❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [27.475808456817589, -5.000337888753362, 4.427726124516064, 0.935932122112036 + 3.027434842214514*ImaginaryUnit, 0.935932122112036 - 3.027434842214514*ImaginaryUnit, 0.224939063195599]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨4, 2, 5, 8, 9, 6, 3❩, ❨5, 8, 4, 5, 8, 6, 3❩, ❨0, 5, 4, 7, 8, 9, 6❩, ❨4, 1, 2, 5, 0, 8, 6❩, ❨6, 6, 5, 0, 7, 4, 8❩, ❨4, 7, 5, 1, 9, 8, 6❩, ❨5, 3, 6, 8, 4, 1, 2❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [35.314786301350608, -4.702400410143717 + 3.178692995712497*ImaginaryUnit, -4.702400410143717 - 3.178692995712497*ImaginaryUnit, 3.918347191792202 + 2.136398063157340*ImaginaryUnit, 3.918347191792202 - 2.136398063157340*ImaginaryUnit, 2.126660067676175 + 2.808465290149118*ImaginaryUnit, 2.126660067676175 - 2.808465290149118*ImaginaryUnit]);
  Test('SameValue(∏(Λ), det(A), 1E-12)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-12)', True);

  Test('A ≔ ❨❨5, 8, 6, 5, 4, 7, 1, 2❩, ❨6, 3, 2, 5, 6, 9, 8, 5❩, ❨4, 5, 6, 9, 2, 0, 4, 7❩, ❨3, 6, 0, 8, 2, 4, 5, 9❩, ❨6, 3, 5, 4, 7, 8, 5, 2❩, ❨5, 3, 2, 1, 9, 8, 5, 6❩, ❨5, 4, 6, 8, 2, 1, 0, 7❩, ❨6, 3, 9, 5, 8, 3, 4, 7❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [39.293880621805755, 8.438596968391407, 1.305365440546036 + 5.337213095091579*ImaginaryUnit, 1.305365440546036 - 5.337213095091579*ImaginaryUnit, -3.112298271457913 + 2.269218103413521*ImaginaryUnit, -3.112298271457913 - 2.269218103413521*ImaginaryUnit, -0.854171490023212, 0.735559561649804]);
  Test('SameValue(∏(Λ), det(A), 1E-12)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-12)', True);

  Test('A ≔ ❨❨5, 7, 6, 9, 5, 6, 3, 2, 5❩, ❨5, 4, 0, 2, 5, 8, 9, 6, 5❩, ❨7, 4, 5, 2, 1, 5, 6, 9, 8❩, ❨3, 6, 5, 4, 7, 8, 5, 2, 0❩, ❨2, 0, 8, 4, 6, 0, 8, 7, 3❩, ❨0, 1, 6, 8, 5, 3, 7, 9, 4❩, ❨5, 6, 3, 8, 7, 4, 1, 0, 5❩, '+'❨6, 5, 3, 8, 9, 4, 5, 2, 1❩, ❨4, 6, 3, 5, 5, 8, 9, 3, 2❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [42.560831728333504, 3.031197245527941 + 5.541734598766467*ImaginaryUnit, 3.031197245527941 - 5.541734598766467*ImaginaryUnit, -4.182206830119297 + 4.492519814382946*ImaginaryUnit, -4.182206830119297 - 4.492519814382946*ImaginaryUnit, -5.948944747014972, -2.836934201726346, 1.910903175127180, -1.383836785536604]);
  Test('SameValue(∏(Λ), det(A), 1E-11)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-11)', True);

  Test('A ≔ ❨❨5, 4, 1, 5, 2, 9, 8, 3, 0, 5❩, ❨5, 6, 3, 2, 0, 8, 5, 6, 9, 7❩, ❨4, 5, 6, 8, 2, 3, 6, 5, 7, 0❩, ❨4, 5, 9, 6, 3, 6, 5, 6, 8, 9❩, ❨4, 5, 5, 2, 6, 8, 7, 9, 2, 5❩, ❨5, 6, 3, 2, 0, 5, 4, 7, 8, 5❩, ❨4, 5, 4, 5, 6, 0, 2, 3, 7, 9❩, '+'❨5, 6, 3, 2, 5, 4, 7, 8, 4, 1❩, ❨5, 6, 3, 2, 0, 4, 5, 1, 2, 7❩, ❨6, 5, 6, 4, 1, 2, 7, 8, 4, 0❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [45.848201925978358, 7.971524351830347, -4.126771995866862 + 4.957293600551936*ImaginaryUnit, -4.126771995866862 - 4.957293600551936*ImaginaryUnit, -4.767288464064574, 4.390252727621196 + 0.962177492102693*ImaginaryUnit, 4.390252727621196 - 0.962177492102693*ImaginaryUnit, -2.883665412954859, -0.347866932148989 + 0.497719878697392*ImaginaryUnit, -0.347866932148989 - 0.497719878697392*ImaginaryUnit]);
  Test('SameValue(∏(Λ), det(A), 1E-11)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-11)', True);

  // - eigenvalues of real symmetric matrices (complex data type)

  Test('A ≔ ComplexMatrix(❨❨21❩❩);', null);
  Test('Λ ≔ eigenvalues(A)', [21]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨3, 5❩, ❨5, 2❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [7.524937810560445, -2.524937810560445]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨4, 1, 8❩, ❨1, 2, 5❩, ❨8, 5, 3❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [13.136085649145485, -5.802319968129476, 1.666234318983990]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨0, 0, 2❩, ❨0, 2, 0❩, ❨2, 0, 0❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [-2, 2, 2]);
  Test('∏(Λ) ≈ det(A)', True);               {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨6, 1, 3, 2❩, ❨1, 8, 5, 1❩, ❨3, 5, 2, 0❩, ❨2, 1, 0, 7❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [12.317576522344481, 7.443877742227379, 4.830339943837674, -1.591794208409535]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨0, 1, 0, 2❩, ❨1, -2, 2, 0❩, ❨0, 2, 0, -1❩, ❨2, 0, -1, 2❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [-3.449489742783178, 3.449489742783178, -1.449489742783178, 1.449489742783178]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨0, 1, 0, 2❩, ❨1, -2, 2, 0❩, ❨0, 2, 0, -1❩, ❨2, 0, -1, -2❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [-3.449489742783178, -3.449489742783178, 1.449489742783178, 1.449489742783178]);
  Test('∏(Λ) ≈ det(A)', True);            {2}                                    {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨3, 2, 1, 0, 1❩, ❨2, 5, 7, 3, 8❩, ❨1, 7, 6, 0, 2❩, ❨0, 3, 0, 5, 3❩, ❨1, 8, 2, 3, 9❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [19.613776982166058, 6.482010475356414, -3.831681427313180, 3.142965451970841, 2.592928517819864]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨2, −1, 0, 3, 8, 2❩, ❨−1, 7, 5, 4, 8, 1❩, ❨0, 5, 2, 9, 3, 0❩, ❨3, 4, 9, −7, 2, 5❩, ❨8, 8, 3, 2, 1, 3❩, ❨2, 1, 0, 5, 3, 8❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [20.369654342915162, -14.157794981021667, -8.732880602208134, 8.619138715827635, 5.073173708519922, 1.828708815967091]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨10, 16, 9, 8, 0, 7, 8❩, ❨16, 12, 5, 10, 4, 11, 14❩, ❨9, 5, 2, 12, 12, 5, 14❩, ❨8, 10, 12, 16, 12, 9, 3❩, ❨0, 4, 12, 12, 16, 17, 7❩, ❨7, 11, 5, 9, 17, 18, 7❩, ❨8, 14, 14, 3, 7, 7, 14❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [67.387186371351689, 22.239241106426356, -13.887324208091441, 10.960115323234580, 9.767062366458459, -5.676556998202810, -2.789723961176876]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨10, 5, 14, 10, 9, 13, 9, 15❩, ❨5, 4, 8, 7, 7, 4, 9, 13❩, ❨14, 8, 4, 6, 7, 5, 1, 14❩, ❨10, 7, 6, 18, 13, 12, 5, 11❩, '+'❨9, 7, 7, 13, 14, 12, 3, 14❩, ❨13, 4, 5, 12, 12, 18, 10, 15❩, ❨9, 9, 1, 5, 3, 10, 14, 14❩, ❨15, 13, 14, 11, 14, 15, 14, 6❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [79.906455303672033, 15.007966465439917, -14.248879078663860, 10.321739350164814, -10.196871321023851, 6.862620180127339, 3.257195862090613, -2.910226761807029]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨10, 7, 11, 13, 15, 11, 9, 17, 7❩, ❨7, 8, 12, 3, 6, 10, 14, 7, 8❩, ❨11, 12, 18, 4, 11, 7, 9, 11, 2❩, ❨13, 3, 4, 0, 4, 9, 16, 11, 13❩, ❨15, 6, 11, 4, 4, 6, 11, 4, 11❩, ❨11, 10, 7, 9, 6, 8, 8, 10, 14❩, ❨9, 14, 9, 16, 11, 8, 10, 11, 13❩, '+'❨17, 7, 11, 11, 4, 10, 11, 14, 9❩, ❨7, 8, 2, 13, 11, 14, 13, 9, 6❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [86.264296094535837, -19.275500362885520, 16.623302682936547, -12.030952595962669, 10.184197789201297, -6.930309037497170, 4.449115111645037, -3.598599758479221, 2.314450076505832]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨6, 6, 8, 14, 12, 14, 5, 9, 12, 10❩, ❨6, 0, 4, 4, 10, 8, 15, 14, 13, 11❩, ❨8, 4, 10, 6, 3, 11, 11, 9, 9, 13❩, ❨14, 4, 6, 10, 9, 17, 16, 9, 11, 8❩, ❨12, 10, 3, 9, 10, 8, 12, 10, 9, 8❩, ❨14, 8, 11, 17, 8, 2, 4, 12, 12, 14❩, '+'❨5, 15, 11, 16, 12, 4, 0, 9, 14, 11❩, ❨9, 14, 9, 9, 10, 12, 9, 18, 9, 8❩, ❨12, 13, 9, 11, 9, 12, 14, 9, 2, 7❩, ❨10, 11, 13, 8, 8, 14, 11, 8, 7, 6❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [96.530449307303996, -23.793421334022021, -16.961843873973226, -11.581585212083253, 11.447090905627016, 9.824786945678492, 5.683157788780861, -4.401418812239814, -3.708291806892070, 0.961076091819928]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  // - eigenvalues of real non-symmetric matrices (complex data type)

  Test('A ≔ ComplexMatrix(❨❨1, 2❩, ❨3, 4❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [5.372281323269014, -0.372281323269014]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨5, 7, 4❩, ❨0, 2, 6❩, ❨1, 5, 3❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [9.334366656868244, 3.386344569047903, -2.720711225916147]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨1, 0, 2❩, ❨1, 0, -1❩, ❨1, 1, 2❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [3, ImaginaryUnit, -ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨1, 1, 0❩, ❨1, 0, -1❩, ❨1, 2, 1❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [1 + ImaginaryUnit, 1 - ImaginaryUnit, 0]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 0, 1❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [1, 1, 0]);
  Test('∏(Λ) ≈ det(A)', True);           {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 0, 2❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [2, 1, 0]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 1, 2❩❩);', null);          // slightly lower precision
  Eps(1E-9); Test('Λ ≔ eigenvalues(A)', [ASC(1), 1, 1]);                         // too great imaginary part to be automatically dropped
  Test('∏(Λ) ≈ det(A)', True);           {3}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ComplexMatrix(❨❨5, 4, 7, 2❩, ❨6, 3, 5, 7❩, ❨0, 5, 7, 8❩, ❨3, 4, 1, 9❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [18.722583143817452, 3.494347146728337 + 3.992775601519464*ImaginaryUnit, 3.494347146728337 - 3.992775601519464*ImaginaryUnit, -1.711277437274135]);
  Test('SameValue(∏(Λ), det(A), 1E-12)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-12)', True);

  Test('A ≔ ComplexMatrix(❨❨0, 0, 1, 0❩, ❨0, 0, 0, 0❩, ❨1, 0, 0, 0❩, ❨0, 0, 0, 2❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [2, -1, 1, 0]);
  Test('SameValue(∏(Λ), det(A), 1E-12)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-12)', True);

  Test('A ≔ ComplexMatrix(❨❨5, 8, 4, 0, 2❩, ❨6, 3, 0, 4, 8❩, ❨6, 5, 4, 9, 2❩, ❨5, 6, 3, 8, 7❩, ❨1, 0, 2, 9, 7❩❩);', null);
  Eps(1E-14); Test('Λ ≔ eigenvalues(A)', [22.814428493209221, 4.634582172214291, -3.664877452474182, 1.607933393525328 + 2.605736529318042*ImaginaryUnit, 1.607933393525328 - 2.605736529318042*ImaginaryUnit]);
  Test('SameValue(∏(Λ), det(A), 1E-12)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-12)', True);

  Test('A ≔ ComplexMatrix(❨❨5, 4, 7, 0, 2, 5❩, ❨6, 5, 8, 9, 5, 1❩, ❨0, 5, 8, 2, 3, 6❩, ❨4, 5, 8, 4, 6, 2❩, ❨7, 5, 9, 6, 1, 0❩, ❨5, 4, 7, 5, 2, 6❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [27.475808456817589, -5.000337888753362, 4.427726124516064, 0.935932122112036 + 3.027434842214514*ImaginaryUnit, 0.935932122112036 - 3.027434842214514*ImaginaryUnit, 0.224939063195599]);
  Test('SameValue(∏(Λ), det(A), 1E-12)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-12)', True);

  Test('A ≔ ComplexMatrix(❨❨4, 2, 5, 8, 9, 6, 3❩, ❨5, 8, 4, 5, 8, 6, 3❩, ❨0, 5, 4, 7, 8, 9, 6❩, ❨4, 1, 2, 5, 0, 8, 6❩, ❨6, 6, 5, 0, 7, 4, 8❩, ❨4, 7, 5, 1, 9, 8, 6❩, ❨5, 3, 6, 8, 4, 1, 2❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [35.314786301350608, -4.702400410143717 + 3.178692995712497*ImaginaryUnit, -4.702400410143717 - 3.178692995712497*ImaginaryUnit, 3.918347191792202 + 2.136398063157340*ImaginaryUnit, 3.918347191792202 - 2.136398063157340*ImaginaryUnit, 2.126660067676175 + 2.808465290149118*ImaginaryUnit, 2.126660067676175 - 2.808465290149118*ImaginaryUnit]);
  Test('SameValue(∏(Λ), det(A), 1E-12)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-12)', True);

  Test('A ≔ ComplexMatrix(❨❨5, 8, 6, 5, 4, 7, 1, 2❩, ❨6, 3, 2, 5, 6, 9, 8, 5❩, ❨4, 5, 6, 9, 2, 0, 4, 7❩, ❨3, 6, 0, 8, 2, 4, 5, 9❩, ❨6, 3, 5, 4, 7, 8, 5, 2❩, ❨5, 3, 2, 1, 9, 8, 5, 6❩, ❨5, 4, 6, 8, 2, 1, 0, 7❩, ❨6, 3, 9, 5, 8, 3, 4, 7❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [39.293880621805755, 8.438596968391407, 1.305365440546036 + 5.337213095091579*ImaginaryUnit, 1.305365440546036 - 5.337213095091579*ImaginaryUnit, -3.112298271457913 + 2.269218103413521*ImaginaryUnit, -3.112298271457913 - 2.269218103413521*ImaginaryUnit, -0.854171490023212, 0.735559561649804]);
  Test('SameValue(∏(Λ), det(A), 1E-12)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-12)', True);

  Test('A ≔ ComplexMatrix(❨❨5, 7, 6, 9, 5, 6, 3, 2, 5❩, ❨5, 4, 0, 2, 5, 8, 9, 6, 5❩, ❨7, 4, 5, 2, 1, 5, 6, 9, 8❩, ❨3, 6, 5, 4, 7, 8, 5, 2, 0❩, ❨2, 0, 8, 4, 6, 0, 8, 7, 3❩, ❨0, 1, 6, 8, 5, 3, 7, 9, 4❩, ❨5, 6, 3, 8, 7, 4, 1, 0, 5❩, '+'❨6, 5, 3, 8, 9, 4, 5, 2, 1❩, ❨4, 6, 3, 5, 5, 8, 9, 3, 2❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [42.560831728333504, 3.031197245527941 + 5.541734598766467*ImaginaryUnit, 3.031197245527941 - 5.541734598766467*ImaginaryUnit, -4.182206830119297 + 4.492519814382946*ImaginaryUnit, -4.182206830119297 - 4.492519814382946*ImaginaryUnit, -5.948944747014972, -2.836934201726346, 1.910903175127180, -1.383836785536604]);
  Test('SameValue(∏(Λ), det(A), 1E-10)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-11)', True);

  Test('A ≔ ComplexMatrix(❨❨5, 4, 1, 5, 2, 9, 8, 3, 0, 5❩, ❨5, 6, 3, 2, 0, 8, 5, 6, 9, 7❩, ❨4, 5, 6, 8, 2, 3, 6, 5, 7, 0❩, ❨4, 5, 9, 6, 3, 6, 5, 6, 8, 9❩, ❨4, 5, 5, 2, 6, 8, 7, 9, 2, 5❩, ❨5, 6, 3, 2, 0, 5, 4, 7, 8, 5❩, ❨4, 5, 4, 5, 6, 0, 2, 3, 7, 9❩, '+'❨5, 6, 3, 2, 5, 4, 7, 8, 4, 1❩, ❨5, 6, 3, 2, 0, 4, 5, 1, 2, 7❩, ❨6, 5, 6, 4, 1, 2, 7, 8, 4, 0❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [45.848201925978358, 7.971524351830347, -4.126771995866862 + 4.957293600551936*ImaginaryUnit, -4.126771995866862 - 4.957293600551936*ImaginaryUnit, -4.767288464064574, 4.390252727621196 + 0.962177492102693*ImaginaryUnit, 4.390252727621196 - 0.962177492102693*ImaginaryUnit, -2.883665412954859, -0.347866932148989 + 0.497719878697392*ImaginaryUnit, -0.347866932148989 - 0.497719878697392*ImaginaryUnit]);
  Test('SameValue(∏(Λ), det(A), 1E-10)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-11)', True);

  // - eigenvalues of complex Hermitian matrices

  Test('A ≔ ComplexMatrix(❨❨3❩❩);', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [3]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, i❩, ❨-i, 2❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [2.618033988749895, 0.381966011250105]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨2, i, 3❩, ❨-i, 5, 1❩, ❨3, 1, 0❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [5.754405707945917, 3.547781109928436, -2.302186817874350]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

//  Test('A ≔ ❨❨−1, 0, i❩, ❨0, −1, −1❩, ❨−i, −1, 1❩❩;', null);                                   // TODO (see bm 5279121)
//  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [-1.732050807568877, 1.732050807568878, -1]);
//  Test('∏(Λ) ≈ det(A)', True);
//  Test('∑(Λ) ≈ tr(A)', True);

//  Test('A ≔ ❨❨1, 0, i❩, ❨0, 1, −1❩, ❨−i, −1, 2❩❩;', null);
//  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [3, 1, 0]);
//  Test('∏(Λ) ≈ det(A)', True);
//  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨3, 2, i, 5❩, ❨2, 7, 0, 1❩, ❨−i, 0, 2, −i❩, ❨5, 1, i, 3❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [9.854644953717658, 5.564431100382422, -2.058222288315847, 1.639146234215764]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨5, i, 1, 2, 3❩, ❨−i, 4, 8, 3, 5❩, ❨1, 8, 1, 2, −i❩, ❨2, 3, 2, 5, 9❩, ❨3, 5, i, 9, 3❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [18.188313753340950, -7.992020539830704, 7.386862064294815, 3.719849391723610, -3.303004669528673]);
  Test('SameValue(∏(Λ), det(A), 1E-13)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-13)', True);

  Test('A ≔ ❨❨3, 8, 5, 1, 2, 0❩, ❨8, 8, 0, i, 7, 2❩, ❨5, 0, 5, 0, 1, 8❩, ❨1, −i, 0, 3, i, 2❩, ❨2, 7, 1, −i, 7, 9❩, ❨0, 2, 8, 2, 9, 2❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [22.508205131012872, 9.731573786579915, -9.030969704262825, 6.534082020312578, -4.802456736301678, 3.059565502659126]);
  Test('SameValue(∏(Λ), det(A), 1E-13)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-13)', True);

  Test('A ≔ ❨❨2, 8, i, 0, 9, 3, 1❩, ❨8, 5, 2, 7, −i, 3, 0❩, ❨−i, 2, 0, 4, 3, 9, 1❩, ❨0, 7, 4, 6, 2, 1, i❩, ❨9, i, 3, 2, 8, 0, 5❩, ❨3, 3, 9, 1, 0, 3, 2❩, ❨1, 0, 1, −i, 5, 2, 4❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [22.174399966828524, 11.679014591078959, -10.611248368420329, -7.740114065328601, 7.675160638625790, 3.968782044056573, 0.854005193159086]);
  Test('SameValue(∏(Λ), det(A), 1E-11)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-11)', True);

  Test('A ≔ ❨❨2, 8, 5, 3, i, 2, 7, 1❩, ❨8, 2, 7, 8, 0, 3, 4, 6❩, ❨5, 7, 1, 2, i, 5, 7, 1❩, ❨3, 8, 2, 6, 3, 2, −i, i❩, ❨−i, 0, −i, 3, 5, 7, 0, 8❩, ❨2, 3, 5, 2, 7, 3, 2, i❩, ❨7, 4, 7, i, 0, 2, 5, 7❩, ❨1, 6, 1, −i, 8, −i, 7, 8❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [29.502669186726628, 12.778533826477519, -11.255008945938719, 9.448283569095759, -6.801121864010132, 5.153648915796556, -3.757699283727493, -3.069305404420104]);
  Test('SameValue(∏(Λ), det(A), 1E-10)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-10)', True);

  Test('A ≔ ❨❨3, 2, 8, 5, 0, 7, 9, 6, 5❩, ❨2, 1, 0, i, 3, 8, 0, 2, 2❩, ❨8, 0, 5, 7, 9, 3, 0, 8, 6❩, ❨5, −i, 7, 1, 2, 4, 9, 3, 5❩, ❨0, 3, 9, 2, 6, −i, i, 2, 8❩, ❨7, 8, 3, 4, i, 5, 4, 9, 8❩, ❨9, 0, 0, 9, −i, 4, 3, 6, 7❩, '+'❨6, 2, 8, 3, 2, 9, 6, 2, 8❩, ❨5, 2, 6, 5, 8, 8, 7, 8, 5❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [42.725142660244707, -13.296262193399452, 12.514481402540193, -9.322363301852313, 8.085873429233777, -6.520876606050006, -4.072755888595437, 1.921419591541651, -1.034659093663105]);
  Test('SameValue(∏(Λ), det(A), 1E-10)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-10)', True);

  Test('A ≔ ❨❨1, 2, 3, 4, 5, 6, 7, 8, 9, i❩, ❨2, 3, 4, 5, 6, 7, 8, 9, i, 1❩, ❨3, 4, 5, 6, 7, 8, 9, i, 1, 2❩, ❨4, 5, 6, 7, 8, 9, i, 1, 2, 3❩, ❨5, 6, 7, 8, 9, i, 1, 2, 3, 4❩, ❨6, 7, 8, 9, −i, 1, 2, 3, 4, 5❩, '+'❨7, 8, 9, −i, 1, 2, 3, 4, 5, 6❩, ❨8, 9, −i, 1, 2, 3, 4, 5, 6, 7❩, ❨9, −i, 1, 2, 3, 4, 5, 6, 7, 8❩, ❨−i, 1, 2, 3, 4, 5, 6, 7, 8, 9❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [45.027811551304595, -16.240738290136697, 16.217270719887843, -8.583046327320011, 8.582382067187316, -6.236363349096755, 6.234384102220631, -5.275735156559826, 5.274352985531571, 4.999681696981297]);
  Test('SameValue(∏(Λ), det(A), 1E-7)', True);
  Test('SameValue(∑(Λ), tr(A), 1E-7)', True);

  // - eigenvalues of complex non-Hermitian matrices

  Test('A ≔ ❨❨i❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨2, i❩, ❨4, 1❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [2.959075886558056 + 1.370730623694961*ImaginaryUnit, 0.040924113441945 - 1.370730623694960*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨9, 3, i❩, ❨5, 0, −i❩, ❨3, 4, 1❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [10.459935146351672 + 0.323571498463665*ImaginaryUnit, -1.957811822820754 + 1.128535366894418*ImaginaryUnit, 1.497876676469088 - 1.452106865358084*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 1, i❩, ❨0, i, 0❩, ❨1, 0, i❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [1 + ImaginaryUnit, ImaginaryUnit, 0]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨0, 1, i❩, ❨0, i, 1❩, ❨1, i, 0❩❩;', null);
//  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [-1 - ImaginaryUnit, 1 + ImaginaryUnit, ImaginaryUnit]);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [1 + ImaginaryUnit, -1 - ImaginaryUnit, ImaginaryUnit]); // for numerical reasons, the absolute values of the first two eigenvalues differ slightly
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 0, 0❩, ❨i, 1, i❩, ❨1, i, 1❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [1 + ImaginaryUnit, 1 - ImaginaryUnit, 1]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨-1, 0, -1❩, ❨i, 1, 0❩, ❨0, 0, 1❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [-1, 1, 1]);
  Test('∏(Λ) ≈ det(A)', True);               {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 0, 1❩, ❨i, 1, 0❩, ❨0, 0, 1❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [1, 1, 1]);
  Test('∏(Λ) ≈ det(A)', True);           {3}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 0, 0, 0❩, ❨0, i, i, 0❩, ❨1, 0, 0, 1❩, ❨0, 0, 0, i❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [ImaginaryUnit, ImaginaryUnit, 1, 0]);
  Test('∏(Λ) ≈ det(A)', True);           {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 0, 0, 0❩, ❨0, 1, i, 0❩, ❨1, 0, 0, 1❩, ❨0, 0, 0, 1❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [1, 1, 1, 0]);
  Test('∏(Λ) ≈ det(A)', True);           {3}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 0, 0, 0❩, ❨0, 1, i, 0❩, ❨1, 1, 0, 1❩, ❨0, 0, 0, 1❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [1.300242590220120 + 0.624810533843827*ImaginaryUnit, 1, 1, -0.300242590220120 - 0.624810533843827*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);                                                                {2}
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨6, 2, 1, 4❩, ❨i, 0, 3, 1❩, ❨2, 8, 7, i❩, ❨3, 7, 5, i❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [11.610435836773123 + 0.971168150072798*ImaginaryUnit, 5.288062915106729 - 0.432329623857917*ImaginaryUnit, -3.546816515929571 + 0.571906149712652*ImaginaryUnit, -0.351682235950283 - 0.110744675927530*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 5, 2, 8, 6❩, ❨4, 3, 2, 0, i❩, ❨2, 5, 3, 9, 4❩, ❨0, 7, 5, 2, i❩, ❨6, 5, 0, 8, 7❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [17.007411042679909 + 0.894883766897211*ImaginaryUnit, 4.053715298377216 - 0.463689356139747*ImaginaryUnit, -1.557595666415852 - 2.588121240940120*ImaginaryUnit, -2.024809742194188 + 2.164016139445834*ImaginaryUnit, -1.478720932447073 - 0.007089309263174*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨5, i, 3, 0, 5, 1❩, ❨7, 6, 3, 2, 6, 9❩, ❨4, 2, i, 4, 5, 0❩, ❨2, 3, 8, 5, 6, 9❩, ❨4, 2, 8, i, 6, 3❩, ❨4, 8, 3, 2, 9, 7❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [23.728019358667826 + 0.814189552450136*ImaginaryUnit, 6.679408809815122 - 1.673047202539371*ImaginaryUnit, -6.327308146805672 + 0.677513478539412*ImaginaryUnit, 5.914415432219343 + 1.336712335249757*ImaginaryUnit, -0.762401540029106 - 2.264439635456942*ImaginaryUnit, -0.232133913867463 + 2.109071471757015*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨5, 2, 3, 6, 9, 7, 5❩, ❨0, 9, 6, 3, 5, 7, 6❩, ❨5, 4, i, 4, 2, 9, 6❩, ❨7, 6, 3, 0, 8, 4, i❩, ❨9, 6, 3, 2, 7, 1, 4❩, ❨2, 5, 3, 6, 0, 1, 1❩, ❨8, 6, 3, i, 4, 5, 6❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [30.739187794060115 + 0.331697323990786*ImaginaryUnit, 8.059962976283682 - 0.006684792990958*ImaginaryUnit, -6.737671725109822 + 0.024618872830778*ImaginaryUnit, -6.368560021382779 + 0.537842454048868*ImaginaryUnit, 2.826230004301034 - 2.925394611754156*ImaginaryUnit, -3.175369396823525 + 1.089346221020855*ImaginaryUnit, 2.656220368671315 + 1.948574532853833*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨5, i, 3, 6, 2, i, 4, 8❩, ❨3, 6, 2, 5, 8, 0, 4, 7❩, ❨4, 5, 3, 8, i, 0, 2, 4❩, ❨7, 5, 6, 3, 0, 2, 5, 2❩, ❨5, 8, 9, 0, 2, 1, 5, 3❩, ❨6, 3, 2, 5, 7, 4, 0, 8❩, ❨9, 3, 5, 8, 0, 2, 5, 4❩, ❨6, i, 3, 0, 2, 5, 6, 7❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [30.772312425651990 + 0.781844424960348*ImaginaryUnit, -8.161318213160694 - 0.407042381222211*ImaginaryUnit, 7.548847048048358 - 0.954391117238968*ImaginaryUnit, 6.117233002689495 + 0.248167551071669*ImaginaryUnit, -0.742832773919414 - 3.105136633221925*ImaginaryUnit, 0.258068235369258 + 2.689896604426479*ImaginaryUnit, -0.207188727333182 + 1.961737071914902*ImaginaryUnit, -0.585120997345780 - 1.215075520690294*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨3, 2, 5, 6, 8, 0, 2, 5, 6❩, ❨8, 5, 0, 3, 6, 5, 1, 4, 7❩, ❨4, 3, 6, 5, 2, 8, 9, 6, 0❩, ❨4, 5, 6, i, 8, 9, 6, 6, 3❩, ❨0, 2, 5, 1, 3, 5, 4, 9, 7❩, ❨6, 3, 2, 5, 6, 4, 1, 7, 0❩, ❨8, 2, 5, 0, 1, 9, 6, 4, 7❩, '+'❨6, i, 3, 6, 4, 7, 8, 0, 2❩, ❨3, 5, 0, 4, 9, 2, 1, i, 8❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [37.908238069312972 + 0.345722875150198*ImaginaryUnit, 9.987869809597976 - 0.223815022736451*ImaginaryUnit, -1.372978560723225 + 7.134561075565384*ImaginaryUnit, -1.083409905970040 - 7.158553458460958*ImaginaryUnit, -6.153342951102821 + 3.302132200426763*ImaginaryUnit, -6.469416592038341 - 2.437820528872786*ImaginaryUnit, 2.449249834737968 - 0.134944071188362*ImaginaryUnit, 0.419736230114992 + 1.959426976919182*ImaginaryUnit, -0.685945933929473 - 1.786710046802955*ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  Test('A ≔ ❨❨1, 2, 3, 4, 5, 6, 7, 8, 9, i❩, ❨2, 3, 4, 5, 6, 7, 8, 9, i, 1❩, ❨3, 4, 5, 6, 7, 8, 9, i, 1, 2❩, ❨4, 5, 6, 7, 8, 9, i, 1, 2, 3❩, ❨5, 6, 7, 8, 9, i, 1, 2, 3, 4❩, ❨6, 7, 8, 9, i, 1, 2, 3, 4, 5❩, '+'❨7, 8, 9, i, 1, 2, 3, 4, 5, 6❩, ❨8, 9, i, 1, 2, 3, 4, 5, 6, 7❩, ❨9, i, 1, 2, 3, 4, 5, 6, 7, 8❩, ❨i, 1, 2, 3, 4, 5, 6, 7, 8, 9❩❩;', null);
  Eps(1E-13); Test('Λ ≔ eigenvalues(A)', [45 + ImaginaryUnit, -16.152375098265388 + 0.309551998983540*ImaginaryUnit, 16.152375098265420 - 0.309551998983539*ImaginaryUnit, -8.468134950725567 + 0.590448785841750*ImaginaryUnit, 8.468134950725586 - 0.590448785841751*ImaginaryUnit, -6.152802719222792 + 0.812637789340920*ImaginaryUnit, 6.152802719222786 - 0.812637789340917*ImaginaryUnit, -5.248510696295683 + 0.952651197515692*ImaginaryUnit, 5.248510696295685 - 0.952651197515693*ImaginaryUnit, 5 - ImaginaryUnit]);
  Test('∏(Λ) ≈ det(A)', True);
  Test('∑(Λ) ≈ tr(A)', True);

  // Eigenvectors

  // - eigenvectors of real symmetric matrices

  Test('A ≔ ❨❨21❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Test('Λ', [21]);

  // -

  Test('A ≔ ❨❨7, 4❩, ❨4, 1❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [9, -1]);

  // -

  Test('A ≔ ❨❨5, 1, 7❩, ❨1, 2, 4❩, ❨7, 4, 3❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [12.227056515812240, -4.115127714597712, 1.888071198785468]);
  Test('col(E, 1) ∥ ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩', True);
  Test('col(E, 2) ∥ ❨0.529539308281996, 0.402013635885016, -0.746976008682163❩', True);
  Test('col(E, 3) ∥ ❨0.505747762545161, -0.856574274410522, -0.102468107714796❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);

  Test('IsEigenvector(A, ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩)', True);
  Test('IsEigenvector(A, ❨0.681034009183674, 0.323520554185660, 0.656917245624993❩)', False);
  Test('IsEigenvector(A, ❨0.681034009183674, 0.323620554185660, 0.656907245624993❩)', False);
  Test('IsEigenvector(A, ❨0.681044009183674, 0.323520554185660, 0.656907245624993❩)', False);

  Test('IsEigenvector(A, ❨0.529539308281996, 0.402013635885016, -0.746976008682163❩)', True);
  Test('IsEigenvector(A, ❨0.529539308281996, 0.402013635885016, -0.746986008682163❩)', False);
  Test('IsEigenvector(A, ❨0.529539308281996, 0.402014635885016, -0.746976008682163❩)', False);
  Test('IsEigenvector(A, ❨0.529549308281996, 0.402013635885016, -0.746976008682163❩)', False);

  Test('IsEigenvector(A, ❨0.505747762545161, -0.856574274410522, -0.102468107714796❩)', True);
  Test('IsEigenvector(A, ❨0.505747762545161, -0.856574274410522, -0.102469107714796❩)', False);
  Test('IsEigenvector(A, ❨0.505747762545161, 0.856574274410522, -0.102468107714796❩)', False);
  Test('IsEigenvector(A, ❨0.505757762545161, -0.856574274410522, -0.102468107714796❩)', False);

  Test('IsEigenvector(A, ❨1, 2❩)', failure, 'Vector is of wrong dimension.');

  Test('EigenvalueOf(A, ❨1, 2, 3❩)', failure, 'A vector which isn''t an eigenvector has no associated eigenvalue.');
  Test('EigenvalueOf(A, ❨1, 2❩)', failure, 'Vector is of wrong dimension.');

  Test('EigenvectorOf(A, 5)', failure, 'Couldn''t compute associated eigenvector.');

  Test('IsEigenpair(A, ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩, 12.227056515812240)', True);
  Test('IsEigenpair(A, ❨0.681014009183674, 0.323520554185660, 0.656907245624993❩, 12.227056515812240)', False);
  Test('IsEigenpair(A, ❨0.681034009183674, 0.323510554185660, 0.656907245624993❩, 12.227056515812240)', False);
  Test('IsEigenpair(A, ❨0.681034009183674, 0.323520554185660, 0.656917245624993❩, 12.227056515812240)', False);
  Test('IsEigenpair(A, ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩, 12.227256515812240)', False);

  Test('IsEigenpair(A, ❨0.681034009183674, 0.323520554185660❩, 12.227056515812240)', failure, 'Vector is of wrong dimension.');

  Test('IsEigenvector(❨❨1, 2, 3❩, ❨5, 6, 7❩❩, ❨3, 1❩)', failure, 'Non-square matrix cannot have eigenvectors.');
  Test('IsEigenvector(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1❩)', failure, 'Vector is of wrong dimension.');
  Test('IsEigenvector(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1, 6❩)', False);

  Test('EigenvalueOf(❨❨1, 2, 3❩, ❨5, 6, 7❩❩, ❨3, 1❩)', failure, 'Non-square matrix cannot have eigenvectors.');
  Test('EigenvalueOf(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1❩)', failure, 'Vector is of wrong dimension.');
  Test('EigenvalueOf(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1, 6❩)', failure, 'A vector which isn''t an eigenvector has no associated eigenvalue.');

  Test('IsEigenpair(❨❨1, 2, 3❩, ❨5, 6, 7❩❩, ❨3, 1, 9❩, 6)', failure, 'Non-square matrix cannot have eigenvectors.');
  Test('IsEigenpair(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1❩, 6)', failure, 'Vector is of wrong dimension.');
  Test('IsEigenpair(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1, 7❩, 6)', False);

  Test('IsEigenvector(A, ❨0, 0, 0❩)', False);
  Test('IsEigenvector(A, ❨1, 0, 0❩)', False);

  // --

  Test('A ≔ ❨❨0, 0, 2❩, ❨0, 2, 0❩, ❨2, 0, 0❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [-2, 2, 2]);
                            {2}

  // --

  Test('A ≔ ❨❨2, 8, 0, 5❩, ❨8, 3, 9, 1❩, ❨0, 9, 7, 0❩, ❨5, 1, 0, 1❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [16.697147626100790, -8.901304306038915, 6.690502206743226, -1.486345526805109]);

  Test('col(E, 1) ∥ ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩', True);
  Test('col(E, 2) ∥ ❨-0.597306724171091, 0.667546370768639, -0.377825442573040, 0.234210279616640❩', True);
  Test('col(E, 3) ∥ ❨-0.605102871175434, -0.020256078111149, 0.589033934885276, -0.535237545533168❩', True);
  Test('col(E, 4) ∥ ❨-0.323277235566436, -0.354560706938906, 0.376021262906496, 0.792708359929243❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);

  Test('IsEigenvector(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩)', True);
  Test('IsEigenvector(A, ❨0.415421869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩)', False);
  Test('IsEigenvector(A, ❨0.415411869738660, 0.654454424709722, 0.607367243387605, 0.174010835501171❩)', False);
  Test('IsEigenvector(A, ❨0.415411869738660, 0.654414424709722, 0.607368243387605, 0.174010835501171❩)', False);
  Test('IsEigenvector(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174011835501171❩)', False);

  Test('IsEigenvector(A, ❨-0.597306724171091, 0.667546370768639, -0.377825442573040, 0.234210279616640❩)', True);
  Test('IsEigenvector(A, ❨0.597306724171091, 0.667546370768639, -0.377825442573040, 0.234210279616640❩)', False);
  Test('IsEigenvector(A, ❨-0.597316724171091, 0.667546370768639, -0.377825442573040, 0.234210279616640❩)', False);
  Test('IsEigenvector(A, ❨-0.597306724171091, 0.667556370768639, -0.377825442573040, 0.234210279616640❩)', False);
  Test('IsEigenvector(A, ❨-0.597306724171091, 0.667546370768639, -0.377815442573040, 0.234220279616640❩)', False);

  Test('IsEigenvector(A, ❨-0.605102871175434, -0.020256078111149, 0.589033934885276, -0.535237545533168❩)', True);
  Test('IsEigenvector(A, ❨-0.605107871175434, -0.020256078111149, 0.589033934885276, -0.535237545533168❩)', False);
  Test('IsEigenvector(A, ❨-0.605102871175434, -0.020258078111149, 0.589033934885276, -0.535237545533168❩)', False);
  Test('IsEigenvector(A, ❨-0.605102871175434, -0.020256078111149, 0.589034934885276, -0.535237545533168❩)', False);
  Test('IsEigenvector(A, ❨-0.605102871175434, -0.020256078111149, 0.589033934885276, 0.535237545533168❩)', False);

  Test('IsEigenvector(A, ❨-0.323277235566436, -0.354560706938906, 0.376021262906496, 0.792708359929243❩)', True);
  Test('IsEigenvector(A, ❨-0.323217235566436, -0.354560706938906, 0.376021262906496, 0.792708359929243❩)', False);
  Test('IsEigenvector(A, ❨-0.323277235566436, -0.354564706938906, 0.376021262906496, 0.792708359929243❩)', False);
  Test('IsEigenvector(A, ❨-0.323277235566436, -0.354560706938906, 0.376023262906496, 0.792708359929243❩)', False);
  Test('IsEigenvector(A, ❨-0.323277235566436, -0.354560706938906, 0.376021262906496, 0.792718359929243❩)', False);

  Test('IsEigenvector(A, ❨-0.35, -0.75, 0.05, 0.43❩)', False);

  Test('IsEigenvector(A, ❨0, 0, 0, 0❩)', False);
  Test('IsEigenvector(A, ❨1, 0, 0, 0❩)', False);

  Test('IsEigenvector(A, ❨0, 0, 0❩)', failure, 'Vector is of wrong dimension.');

  Test('IsEigenpair(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩, 16.697147626100790)', True);
  Test('IsEigenpair(A, ❨0.415421869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩, 16.697147626100790)', False);
  Test('IsEigenpair(A, ❨0.415411869738660, 0.654416424709722, 0.607367243387605, 0.174010835501171❩, 16.697147626100790)', False);
  Test('IsEigenpair(A, ❨0.415411869738660, 0.654414424709722, 0.607377243387605, 0.174010835501171❩, 16.697147626100790)', False);
  Test('IsEigenpair(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174014835501171❩, 16.697147626100790)', False);
  Test('IsEigenpair(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩, 16.697137626100790)', False);

  // --

  Test('A ≔ ❨❨0, 1, 0, 2❩, ❨1, -2, 2, 0❩, ❨0, 2, 0, -1❩, ❨2, 0, -1, -2❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [-3.449489742783178, -3.449489742783178, 1.449489742783178, 1.449489742783178]);
                         {2}                                    {2}

  // --

  Test('A ≔ ❨❨0, 0, 1, 2❩, ❨0, -2, 0, 0❩, ❨1, 0, 0, 1❩, ❨2, 0, 1, 0❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [2.732050807568878, -2, -2, -0.732050807568878]);
                                            {2}

  // --

  Test('A ≔ ❨❨3.09090210854135331, -0.0989480787770248203, 0.206696174292443883, -0.173565216686099517❩, '+'❨-0.0989480787770248203, 3.10770621771892455, -0.224991363401589891, 0.188927902874714558❩, ❨0.206696174292443883, -0.224991363401589891, 3.4699924914029535, -0.394658241210492851❩, '+'❨-0.173565216686099517, 0.188927902874714558, -0.394658241210492851, 3.33139918233676864❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [4, 3, 3, 3]);
                           {3}

  // --

  Test('A ≔ ❨❨3, 9, 2, 5, 1❩, ❨9, 2, 8, 2, 7❩, ❨2, 8, 3, 2, 5❩, ❨5, 2, 2, 6, 2❩, ❨1, 7, 5, 2, 4❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [21.437075526003142, -9.608848480894173, 6.159627148842105, 1.764018208004057, -1.751872401955133]);

  Test('col(E, 1) ∥ ❨0.438718328790090, 0.571040527424050, 0.444736969767807, 0.328057963353101, 0.419554459427209❩', True);
  Test('col(E, 2) ∥ ❨-0.517317497184398, 0.742773501365916, -0.312648829278767, 0.142651379319237, -0.250142811818416❩', True);
  Test('col(E, 3) ∥ ❨0.384970537144515, -0.154132214465530, -0.335181955611742, 0.737631277204982, -0.414239183945520❩', True);
  Test('col(E, 4) ∥ ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩', True);
  Test('col(E, 5) ∥ ❨-0.212274890341188, -0.050664616873610, 0.761222130728086, 0.108633310275057, -0.600926067928628❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('SameValue(EigenvalueOf(A, col(E, 5)), Λ[5], 1E-15)', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);

  Test('IsEigenvector(A, ❨0.438718328790090, 0.571040527424050, 0.444736969767807, 0.328057963353101, 0.419554459427209❩)', True);
  Test('IsEigenvector(A, ❨0.438728328790090, 0.571040527424050, 0.444736969767807, 0.328057963353101, 0.419554459427209❩)', False);
  Test('IsEigenvector(A, ❨0.438718328790090, 0.571041527424050, 0.444736969767807, 0.328057963353101, 0.419554459427209❩)', False);
  Test('IsEigenvector(A, ❨0.438718328790090, 0.571040527424050, 0.444746969767807, 0.328057963353101, 0.419554459427209❩)', False);
  Test('IsEigenvector(A, ❨0.438718328790090, 0.571040527424050, 0.444736969767807, 0.328027963353101, 0.419554459427209❩)', False);
  Test('IsEigenvector(A, ❨0.438718328790090, 0.571040527424050, 0.444736969767807, 0.328057963353101, 0.419555459427209❩)', False);

  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩, 1.764018208004057)', True);
  Test('IsEigenpair(A, ❨0.588761415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309632465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112468495387063, -0.562252131562491, -0.478222112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562257131562491, -0.478222112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478223112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩, 1.764014208004057)', False);

  // --

  Test('A ≔ ❨❨2, 0, 1, 0, 0❩, ❨0, 0, 0, 2, 0❩, ❨1, 0, 0, 1, 1❩, ❨0, 2, 1, 0, 0❩, ❨0, 0, 1, 0, 2❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [2.895106515927531, -2.292401585224621, 2, 2, -0.602704930702910]);
                                                               {2}
  // --

  Test('A ≔ ❨❨0, 1, 1, 0, 0❩, ❨1, 0, 0, 0, 1❩, ❨1, 0, 0, 1, 0❩, ❨0, 0, 1, 0, 1❩, ❨0, 1, 0, 1, 0❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [2, -1.618033988749895, -1.618033988749895, 0.618033988749895, 0.618033988749895]);
                            {2}                                    {2}

  // --

  Test('A ≔ ❨❨4.88518185435250995, -0.0503410125523369264, 2.41500433970266516, -0.3526553784998275, 0.377477078463321959❩, '+'❨-0.0503410125523369266, 6.99880168536003428, 0.0574866278829369505, -0.00839458885495496514, 0.00898544320902995499❩, ❨2.41500433970266516, 0.0574866278829369505, 4.24219977363724693, 0.402712768116436254, -0.431057764708227601❩, '+'❨-0.3526553784998275, -0.00839458885495496536, 0.402712768116436254, 6.94119313935299163, 0.0629459900627696699❩, ❨0.377477078463321959, 0.00898544320902995499, -0.431057764708227601, 0.0629459900627696697, 6.93262354729721721❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [7, 7, 7, 7, 2]);
                        {4}

  // --

  Test('A ≔ ❨❨4, 2, 8, 1, 7, 5❩, ❨2, 3, 9, 4, 1, 2❩, ❨8, 9, 7, 5, 2, 1❩, ❨1, 4, 5, 8, 0, 9❩, ❨7, 1, 2, 0, 3, 2❩, ❨5, 2, 1, 9, 2, 6❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [25.699931912397357, 10.666684085724908, -7.835696371596078, 7.409797588520018, -2.807084819210859, -2.133632395835343]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);

  // --

  Test('A ≔ ❨❨9, 2, 1, 3, 5, 4, 0❩, ❨2, 7, 2, 8, 3, 5, 2❩, ❨1, 2, 6, 9, 0, 3, 1❩, ❨3, 8, 9, 5, 4, 9, 2❩, ❨5, 3, 0, 4, 3, 7, 4❩, ❨4, 5, 3, 9, 7, 8, 4❩, ❨0, 2, 1, 2, 4, 4, 3❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [30.641044817744506384, 9.8517787330299055632, -6.5550797165196624249, 6.2412143228197048481, 3.6651048576610945760, -3.1057270151052322463, 0.26166400036968330015]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('SameValue(EigenvalueOf(A, col(E, 3)), Λ[3], 1E-14)', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);

  // --

  Test('A ≔ ❨❨3, 7, 1, 5, 2, 0, 6, 8❩, ❨7, 3, 9, 4, 7, 8, 2, 1❩, ❨1, 9, 5, 6, 0, 1, 9, 2❩, ❨5, 4, 6, 6, 3, 9, 1, 1❩, ❨2, 7, 0, 3, 4, 2, 6, 2❩, ❨0, 8, 1, 9, 2, 9, 0, 5❩, ❨6, 2, 9, 1, 6, 0, 4, 3❩, ❨8, 1, 2, 1, 2, 5, 3, 1❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-13); Test('Λ', [32.776937694011345, -15.258689129816567, 12.644822532655207, 6.717980397487340, -5.958148232873256, 4.141024331011820, -0.325483553395741, 0.261555960919853]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);
  Test('IsEigenvector(A, col(E, 8))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);
  Test('EigenvalueOf(A, col(E, 8)) ≈ Λ[8]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);
  Test('EigenvectorOf(A, Λ[8]) ∥ col(E, 8)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);
  Test('IsEigenpair(A, col(E, 8), Λ[8])', True);

  // --

  Test('A ≔ ❨❨3, 2, 1, 5, 9, 8, 7, 0, 2❩, ❨2, 5, 4, 8, 3, 0, 7, 1, 8❩, ❨1, 4, 2, 9, 3, 4, 5, 9, 2❩, ❨5, 8, 9, 8, 2, 6, 4, 8, 5❩, ❨9, 3, 3, 2, 0, 2, 9, 7, 1❩, ❨8, 0, 4, 6, 2, 5, 4, 8, 0❩, '+'❨7, 7, 5, 4, 9, 4, 3, 2, 5❩, ❨0, 1, 9, 8, 7, 8, 2, 8, 1❩, ❨2, 8, 2, 5, 1, 0, 5, 1, 9❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-13); Test('Λ', [41.547567826293928, 15.294422185576229, -12.701166678543432, 11.116110823642684, -8.128109328768399, -4.470204023755420, -3.750820326108707, 2.912735795595040, 1.179463726068098]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);
  Test('IsEigenvector(A, col(E, 8))', True);
  Test('IsEigenvector(A, col(E, 9))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);
  Test('EigenvalueOf(A, col(E, 8)) ≈ Λ[8]', True);
  Test('EigenvalueOf(A, col(E, 9)) ≈ Λ[9]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);
  Test('EigenvectorOf(A, Λ[8]) ∥ col(E, 8)', True);
  Test('EigenvectorOf(A, Λ[9]) ∥ col(E, 9)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);
  Test('IsEigenpair(A, col(E, 8), Λ[8])', True);
  Test('IsEigenpair(A, col(E, 9), Λ[9])', True);

  // --

  Test('A ≔ ❨❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩, ❨2, 3, 4, 5, 6, 7, 8, 9, 10, 1❩, ❨3, 4, 5, 6, 7, 8, 9, 10, 1, 2❩, ❨4, 5, 6, 7, 8, 9, 10, 1, 2, 3❩, ❨5, 6, 7, 8, 9, 10, 1, 2, 3, 4❩, '+'❨6, 7, 8, 9, 10, 1, 2, 3, 4, 5❩, ❨7, 8, 9, 10, 1, 2, 3, 4, 5, 6❩, ❨8, 9, 10, 1, 2, 3, 4, 5, 6, 7❩, ❨9, 10, 1, 2, 3, 4, 5, 6, 7, 8❩, ❨10, 1, 2, 3, 4, 5, 6, 7, 8, 9❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-13); Test('Λ', [55, -16.180339887498950, 16.180339887498950, -8.506508083520398, 8.506508083520398, -6.180339887498945, 6.180339887498937, -5.257311121191338, 5.257311121191338, -5]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);
  Test('IsEigenvector(A, col(E, 8))', True);
  Test('IsEigenvector(A, col(E, 9))', True);
  Test('IsEigenvector(A, col(E, 10))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);
  Test('EigenvalueOf(A, col(E, 8)) ≈ Λ[8]', True);
  Test('EigenvalueOf(A, col(E, 9)) ≈ Λ[9]', True);
  Test('EigenvalueOf(A, col(E, 10)) ≈ Λ[10]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);
  Test('EigenvectorOf(A, Λ[8]) ∥ col(E, 8)', True);
  Test('EigenvectorOf(A, Λ[9]) ∥ col(E, 9)', True);
  Test('EigenvectorOf(A, Λ[10]) ∥ col(E, 10)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);
  Test('IsEigenpair(A, col(E, 8), Λ[8])', True);
  Test('IsEigenpair(A, col(E, 9), Λ[9])', True);
  Test('IsEigenpair(A, col(E, 10), Λ[10])', True);

  // --

  Test('A ≔ ❨❨4.45528983757093495, -0.0878217640871465801, 0.587220808461187196, 0.381484594913282418, 0.806374962066634831, 0.415539821124804545, -1.07626065503415, -0.174476853005194865, 0.6117764065082528, 0.479713560928682166❩, '+'❨-0.0878217640871465801, 4.39333906061644079, -1.16146861964970503, 0.189604671732296926, 0.549866251807311702, 0.959584846946758171, -0.559867163033034075, 0.601342029002866998, 0.908395415505047237, -0.0481674127191538071❩, '+
    '❨0.587220808461187195, -1.16146861964970503, 4.731699905531466, -0.297368715494811859, 0.174137444046764151, 0.158862374235097095, 0.0708587702680495379, 0.527044617773810588, 0.717806236876411773, 0.100941787359586736❩, '+'❨0.381484594913282417, 0.189604671732296926, -0.297368715494811859, 5.20093245260149577, -0.422111106264731792, -0.609505049911372394, 0.430772405093700426, -0.075450459824149398, -0.0231177557010034982, 1.02271233462064678❩, '+
    '❨0.806374962066634831, 0.549866251807311702, 0.174137444046764151, -0.422111106264731792, 4.65483119820909985, 0.289795090828546352, 0.795841033040542465, -0.226199611183290607, -1.41597888676298831, 0.547038811317080901❩, '+'❨0.415539821124804545, 0.959584846946758171, 0.158862374235097095, -0.609505049911372394, 0.289795090828546352, 3.97252873681661176, 0.586614812369869111, -0.203645718668121608, 0.548222506108191402, -0.0260783651922211985❩, '+
    '❨-1.07626065503415, -0.559867163033034075, 0.0708587702680495379, 0.430772405093700426, 0.795841033040542465, 0.586614812369869111, 5.07303988296985066, 0.0852567827171805209, 0.721798919103536904, 0.0913845737686953809❩, '+'❨-0.174476853005194865, 0.601342029002866998, 0.527044617773810588, -0.0754504598241493979, -0.226199611183290607, -0.203645718668121608, 0.0852567827171805208, 5.72552879232507247, -0.381852810701084683, 0.248741195041378638❩, '+'❨0.6117764065082528, 0.908395415505047237, 0.717806236876411773, -0.0231177557010034982, -1.41597888676298831, 0.548222506108191402, 0.721798919103536904, -0.381852810701084683, 4.15028770968589867, 0.0862862168111860338❩, '+'❨0.479713560928682166, -0.0481674127191538071, 0.100941787359586736, 1.02271233462064678, 0.547038811317080901, -0.0260783651922211984, 0.0913845737686953809, 0.248741195041378638, 0.0862862168111860338, 3.64252242367312907❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [6, 6, 6, 6, 6, 6, 3, 3, 3, 1]);
                        {6}               {3}

  // --

  Test('A ≔ ❨❨0.947217218464429805, -0.0797032009555874838, 0.455681246175844459, -0.890506530889571131, 0.0810219664601976321, -0.4241414032969027, -0.325611817072624359, 0.442131323976784809, 0.696001040652687542, -0.363371579820000821❩, '+'❨-0.0797032009555874838, 0.664262086761071853, 0.151183874385912782, 0.366175763890158299, -0.362499363441370523, 0.345609605325798349, -0.36153060182114391, -0.370383371480650959, -0.351151983839811129, -0.192995673028588803❩, '+
    '❨0.455681246175844459, 0.151183874385912782, 0.965112845520727124, -0.379442604818156945, -0.0868548299522641688, -0.703208127497219295, -0.29273075622571167, 0.392533036681992209, 0.433785161673175889, -0.0198556703071645884❩, '+'❨-0.890506530889571131, 0.366175763890158299, -0.379442604818156945, 1.60196556896465841, 0.0343051071296422081, 0.587093406299823787, -0.485735414772669726, -0.766652742695679709, -0.63672778172637717, -0.0824745994586258031❩, '+
    '❨0.0810219664601976321, -0.362499363441370523, -0.0868548299522641688, 0.0343051071296422081, 0.327534361557092576, -0.222948382709437083, -0.0500390113869091228, 0.161756725513035523, 0.21826384416565025, -0.135644537928391234❩, '+'❨-0.4241414032969027, 0.345609605325798349, -0.703208127497219295, 0.587093406299823787, -0.222948382709437083, 0.989037781112786131, -0.126211410841058615, -0.699854415829559306, -0.413031006405435901, 0.131956448780506888❩, '+
    '❨-0.325611817072624359, -0.36153060182114391, -0.29273075622571167, -0.485735414772669726, -0.0500390113869091228, -0.126211410841058615, 1.00217393220066957, 0.281471732316985816, -0.335032854995363073, 0.398331000774434439❩, '+'❨0.442131323976784809, -0.370383371480650959, 0.392533036681992209, -0.766652742695679709, 0.161756725513035523, -0.699854415829559306, 0.281471732316985816, 0.61286300125401041, 0.371247429623465179, -0.0611706531297214894❩, '+
    '❨0.696001040652687542, -0.351151983839811129, 0.433785161673175889, -0.63672778172637717, 0.21826384416565025, -0.413031006405435901, -0.335032854995363073, 0.371247429623465179, 0.960656361843560564, 0.149752326353616183❩, '+'❨-0.363371579820000821, -0.192995673028588803, -0.0198556703071645884, -0.0824745994586258031, -0.135644537928391234, 0.131956448780506888, 0.398331000774434439, -0.0611706531297214894, 0.149752326353616183, 0.929176842320993553❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [4, 2, 1, 1, 1, 0, 0, 0, 0, 0]);
                              {3}      {5}

  // --

  Test('A ≔ ReversalMatrix(10);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [-1, -1, -1, -1, -1, 1, 1, 1, 1, 1]);
                         {5}                {5}
  // --

  Test('A ≔ diag(12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Test('IsIdentity(E)', True);
  Test('A = diag(Λ)', True);

  // --

  Test('A ≔ IdentityMatrix(10);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Test('IsIdentity(E)', True);
  Test('IsIdentity(diag(Λ))', True);

  // --

  Test('A ≔ ZeroMatrix(10);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Test('IsIdentity(E)', True);
  Test('IsZero(Λ)', True);

  // - eigenvectors of real non-symmetric matrices

  Test('A ≔ ❨❨5, 2, 1❩, ❨0, 2, 4❩, ❨1, 2, 1❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [6, 3.236067977499790, -1.236067977499789]);

  Test('col(E, 1) ∥ ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩', True);
  Test('col(E, 2) ∥ ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩', True);
  Test('col(E, 3) ∥ ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);

  Test('IsEigenvector(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩)', True);
  Test('IsEigenvector(A, ❨0.904554033733291, 0.301511344577764, 0.301511344577764❩)', False);
  Test('IsEigenvector(A, ❨0.904534033733291, 0.301571344577764, 0.301511344577764❩)', False);
  Test('IsEigenvector(A, ❨0.904534033733291, 0.301511344577764, 0.301510344577764❩)', False);

  Test('IsEigenvector(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩)', True);
  Test('IsEigenvector(A, ❨0.781038693640056, -0.596654357427190, -0.184376336212866❩)', False);
  Test('IsEigenvector(A, ❨0.781030693640056, -0.596650357427190, -0.184376336212866❩)', False);
  Test('IsEigenvector(A, ❨0.781030693640056, -0.596654357427190, -0.184374336212866❩)', False);

  Test('IsEigenvector(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩)', True);
  Test('IsEigenvector(A, ❨0.146862297110534, -0.769007151342417, 0.622139854231882❩)', False);
  Test('IsEigenvector(A, ❨0.146867297110534, -0.769004151342417, 0.622139854231882❩)', False);
  Test('IsEigenvector(A, ❨0.146867297110534, -0.769007151342417, 0.622199854231882❩)', False);

  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩, 6)', True);
  Test('IsEigenpair(A, ❨0.904535033733291, 0.301511344577764, 0.301511344577764❩, 6)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301514344577764, 0.301511344577764❩, 6)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301510344577764❩, 6)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩, 7)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩, 6.0001)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩, -6)', False);

  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, 3.236067977499790)', True);
  Test('IsEigenpair(A, ❨0.781034693640056, -0.596654357427190, -0.184376336212866❩, 3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596644357427190, -0.184376336212866❩, 3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184370336212866❩, 3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, 3.236068977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, -3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, 0)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, 6)', False);

  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, -1.236067977499789)', True);
  Test('IsEigenpair(A, ❨0.146865297110534, -0.769007151342417, 0.622139854231882❩, -1.236067977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769047151342417, 0.622139854231882❩, -1.236067977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622130854231882❩, -1.236067977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, -1.236064977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 1.236067977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 6)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 0)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 1E100)', False);

  // -

  Test('A ≔ ❨❨2, 3, 0, 1❩, ❨5, 6, 8, 1❩, ❨4, 2, 5, 7❩, ❨3, 0, 2, 6❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [13.109772228646445, 3.890227771353548, 3, -1]);

  Test('col(E, 1) ∥ ❨0.233782511971403, 0.784104790434032, 0.520117226972313, 0.244956087740999❩', True);
  Test('col(E, 2) ∥ ❨-0.596105141205652, -0.513975993546188, 0.456218061914745, 0.415153488085013❩', True);
  Test('col(E, 3) ∥ ❨-0.696310623822791, -0.348155311911396, 0.522232967867093, 0.348155311911397❩', True);
  Test('col(E, 4) ∥ ❨-0.713563032023560, 0.600041640565266, -0.121630062276743, 0.340564174374881❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);

  // -

  Test('A ≔ ❨❨6, 4, 2, 1, 3❩, ❨5, 0, 4, 7, 3❩, ❨9, 5, 3, 0, 2❩, ❨4, 4, 2, 8, 5❩, ❨3, 6, 0, 4, 7❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [19.317943284512630, 6.195170660848524, -3.837738100539873, 2.388435089272784, -0.063810934094071]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);

  // -

  Test('A ≔ ❨❨3, 6, 2, 5, 4, 7❩, ❨2, 0, 1, 4, 2, 0❩, ❨3, 6, 9, 8, 5, 2❩, ❨2, 7, 6, 4, 3, 1❩, ❨5, 6, 9, 8, 2, 4❩, ❨6, 5, 3, 9, 5, 7❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [25.608230542132779785, 6.8775642608200556778, -4.2970302684870303417, -3.5268871389883464722, 0.86082458257068375765, -0.52270197804814240633]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);

  // -

  Test('A ≔ ❨❨0, 2, 0❩, ❨1, −1, 1❩, ❨2, 0, 1❩❩;', null);
  Eps(1E-14); Test('eigenvalues(A)', [2, -1, -1]); // -1: geometric multiplicity = 1 < 2 = algebraic multiplicity
  Test('eigenvectors(A)', failure, 'Couldn''t compute eigenvectors.');

  // -

  Test('A ≔ ❨❨0, 1, 0❩, ❨−1, −2, 0❩, ❨1, 0, −1❩❩;', null);
  Eps(1E-14); Test('eigenvalues(A)', [-1, -1, -1]); // -1: geometric multiplicity = 1 < 3 = algebraic multiplicity
  Test('eigenvectors(A)', failure, 'Couldn''t compute eigenvectors.');

  // - eigenvectors of real symmetric matrices (complex data type)

  Test('A ≔ ComplexMatrix(❨❨21❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Test('Λ', [TASC(21)]);

  // -

  Test('A ≔ ComplexMatrix(❨❨7, 4❩, ❨4, 1❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(9), -1]);

  // -

  Test('A ≔ ComplexMatrix(❨❨5, 1, 7❩, ❨1, 2, 4❩, ❨7, 4, 3❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(12.227056515812240), -4.115127714597712, 1.888071198785468]);
  Test('col(E, 1) ∥ ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩', True);
  Test('col(E, 2) ∥ ❨0.529539308281996, 0.402013635885016, -0.746976008682163❩', True);
  Test('col(E, 3) ∥ ❨0.505747762545161, -0.856574274410522, -0.102468107714796❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);

  Test('IsEigenvector(A, ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩)', True);
  Test('IsEigenvector(A, ❨0.681034009183674, 0.323520554185660, 0.656917245624993❩)', False);
  Test('IsEigenvector(A, ❨0.681034009183674, 0.323620554185660, 0.656907245624993❩)', False);
  Test('IsEigenvector(A, ❨0.681044009183674, 0.323520554185660, 0.656907245624993❩)', False);

  Test('IsEigenvector(A, ❨0.529539308281996, 0.402013635885016, -0.746976008682163❩)', True);
  Test('IsEigenvector(A, ❨0.529539308281996, 0.402013635885016, -0.746986008682163❩)', False);
  Test('IsEigenvector(A, ❨0.529539308281996, 0.402014635885016, -0.746976008682163❩)', False);
  Test('IsEigenvector(A, ❨0.529549308281996, 0.402013635885016, -0.746976008682163❩)', False);

  Test('IsEigenvector(A, ❨0.505747762545161, -0.856574274410522, -0.102468107714796❩)', True);
  Test('IsEigenvector(A, ❨0.505747762545161, -0.856574274410522, -0.102469107714796❩)', False);
  Test('IsEigenvector(A, ❨0.505747762545161, 0.856574274410522, -0.102468107714796❩)', False);
  Test('IsEigenvector(A, ❨0.505757762545161, -0.856574274410522, -0.102468107714796❩)', False);

  Test('IsEigenvector(A, ❨1, 2❩)', failure, 'Vector is of wrong dimension.');

  Test('EigenvalueOf(A, ❨1, 2, 3❩)', failure, 'A vector which isn''t an eigenvector has no associated eigenvalue.');
  Test('EigenvalueOf(A, ❨1, 2❩)', failure, 'Vector is of wrong dimension.');

  Test('EigenvectorOf(A, 5)', failure, 'Couldn''t compute associated eigenvector.');

  Test('IsEigenpair(A, ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩, 12.227056515812240)', True);
  Test('IsEigenpair(A, ❨0.681014009183674, 0.323520554185660, 0.656907245624993❩, 12.227056515812240)', False);
  Test('IsEigenpair(A, ❨0.681034009183674, 0.323510554185660, 0.656907245624993❩, 12.227056515812240)', False);
  Test('IsEigenpair(A, ❨0.681034009183674, 0.323520554185660, 0.656917245624993❩, 12.227056515812240)', False);
  Test('IsEigenpair(A, ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩, 12.227256515812240)', False);

  Test('IsEigenpair(A, ❨0.681034009183674, 0.323520554185660❩, 12.227056515812240)', failure, 'Vector is of wrong dimension.');

  Test('IsEigenvector(❨❨1, 2, 3❩, ❨5, 6, 7❩❩, ❨3, 1❩)', failure, 'Non-square matrix cannot have eigenvectors.');
  Test('IsEigenvector(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1❩)', failure, 'Vector is of wrong dimension.');
  Test('IsEigenvector(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1, 6❩)', False);

  Test('EigenvalueOf(❨❨1, 2, 3❩, ❨5, 6, 7❩❩, ❨3, 1❩)', failure, 'Non-square matrix cannot have eigenvectors.');
  Test('EigenvalueOf(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1❩)', failure, 'Vector is of wrong dimension.');
  Test('EigenvalueOf(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1, 6❩)', failure, 'A vector which isn''t an eigenvector has no associated eigenvalue.');

  Test('IsEigenpair(❨❨1, 2, 3❩, ❨5, 6, 7❩❩, ❨3, 1, 9❩, 6)', failure, 'Non-square matrix cannot have eigenvectors.');
  Test('IsEigenpair(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1❩, 6)', failure, 'Vector is of wrong dimension.');
  Test('IsEigenpair(❨❨1, 2, 3❩, ❨5, 6, 7❩, ❨6, 7, 8❩❩, ❨3, 1, 7❩, 6)', False);

  Test('IsEigenvector(A, ❨0, 0, 0❩)', False);
  Test('IsEigenvector(A, ❨1, 0, 0❩)', False);

  // --

  Test('A ≔ ComplexMatrix(❨❨0, 0, 2❩, ❨0, 2, 0❩, ❨2, 0, 0❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(-2), 2, 2]);
                                  {2}

  // --

  Test('A ≔ ComplexMatrix(❨❨2, 8, 0, 5❩, ❨8, 3, 9, 1❩, ❨0, 9, 7, 0❩, ❨5, 1, 0, 1❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(16.697147626100790), -8.901304306038915, 6.690502206743226, -1.486345526805109]);

  Test('col(E, 1) ∥ ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩', True);
  Test('col(E, 2) ∥ ❨-0.597306724171091, 0.667546370768639, -0.377825442573040, 0.234210279616640❩', True);
  Test('col(E, 3) ∥ ❨-0.605102871175434, -0.020256078111149, 0.589033934885276, -0.535237545533168❩', True);
  Test('col(E, 4) ∥ ❨-0.323277235566436, -0.354560706938906, 0.376021262906496, 0.792708359929243❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);

  Test('IsEigenvector(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩)', True);
  Test('IsEigenvector(A, ❨0.415421869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩)', False);
  Test('IsEigenvector(A, ❨0.415411869738660, 0.654454424709722, 0.607367243387605, 0.174010835501171❩)', False);
  Test('IsEigenvector(A, ❨0.415411869738660, 0.654414424709722, 0.607368243387605, 0.174010835501171❩)', False);
  Test('IsEigenvector(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174011835501171❩)', False);

  Test('IsEigenvector(A, ❨-0.597306724171091, 0.667546370768639, -0.377825442573040, 0.234210279616640❩)', True);
  Test('IsEigenvector(A, ❨0.597306724171091, 0.667546370768639, -0.377825442573040, 0.234210279616640❩)', False);
  Test('IsEigenvector(A, ❨-0.597316724171091, 0.667546370768639, -0.377825442573040, 0.234210279616640❩)', False);
  Test('IsEigenvector(A, ❨-0.597306724171091, 0.667556370768639, -0.377825442573040, 0.234210279616640❩)', False);
  Test('IsEigenvector(A, ❨-0.597306724171091, 0.667546370768639, -0.377815442573040, 0.234220279616640❩)', False);

  Test('IsEigenvector(A, ❨-0.605102871175434, -0.020256078111149, 0.589033934885276, -0.535237545533168❩)', True);
  Test('IsEigenvector(A, ❨-0.605107871175434, -0.020256078111149, 0.589033934885276, -0.535237545533168❩)', False);
  Test('IsEigenvector(A, ❨-0.605102871175434, -0.020258078111149, 0.589033934885276, -0.535237545533168❩)', False);
  Test('IsEigenvector(A, ❨-0.605102871175434, -0.020256078111149, 0.589034934885276, -0.535237545533168❩)', False);
  Test('IsEigenvector(A, ❨-0.605102871175434, -0.020256078111149, 0.589033934885276, 0.535237545533168❩)', False);

  Test('IsEigenvector(A, ❨-0.323277235566436, -0.354560706938906, 0.376021262906496, 0.792708359929243❩)', True);
  Test('IsEigenvector(A, ❨-0.323217235566436, -0.354560706938906, 0.376021262906496, 0.792708359929243❩)', False);
  Test('IsEigenvector(A, ❨-0.323277235566436, -0.354564706938906, 0.376021262906496, 0.792708359929243❩)', False);
  Test('IsEigenvector(A, ❨-0.323277235566436, -0.354560706938906, 0.376023262906496, 0.792708359929243❩)', False);
  Test('IsEigenvector(A, ❨-0.323277235566436, -0.354560706938906, 0.376021262906496, 0.792718359929243❩)', False);

  Test('IsEigenvector(A, ❨-0.35, -0.75, 0.05, 0.43❩)', False);

  Test('IsEigenvector(A, ❨0, 0, 0, 0❩)', False);
  Test('IsEigenvector(A, ❨1, 0, 0, 0❩)', False);

  Test('IsEigenvector(A, ❨0, 0, 0❩)', failure, 'Vector is of wrong dimension.');

  Test('IsEigenpair(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩, 16.697147626100790)', True);
  Test('IsEigenpair(A, ❨0.415421869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩, 16.697147626100790)', False);
  Test('IsEigenpair(A, ❨0.415411869738660, 0.654416424709722, 0.607367243387605, 0.174010835501171❩, 16.697147626100790)', False);
  Test('IsEigenpair(A, ❨0.415411869738660, 0.654414424709722, 0.607377243387605, 0.174010835501171❩, 16.697147626100790)', False);
  Test('IsEigenpair(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174014835501171❩, 16.697147626100790)', False);
  Test('IsEigenpair(A, ❨0.415411869738660, 0.654414424709722, 0.607367243387605, 0.174010835501171❩, 16.697137626100790)', False);

  // --

  Test('A ≔ ComplexMatrix(❨❨0, 0, 1, 2❩, ❨0, -2, 0, 0❩, ❨1, 0, 0, 1❩, ❨2, 0, 1, 0❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(2.732050807568878), -2, -2, -0.732050807568878]);
                                                  {2}

  // --

  Test('A ≔ ComplexMatrix(❨❨3.09090210854135331, -0.0989480787770248203, 0.206696174292443883, -0.173565216686099517❩, '+'❨-0.0989480787770248203, 3.10770621771892455, -0.224991363401589891, 0.188927902874714558❩, ❨0.206696174292443883, -0.224991363401589891, 3.4699924914029535, -0.394658241210492851❩, '+'❨-0.173565216686099517, 0.188927902874714558, -0.394658241210492851, 3.33139918233676864❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(4), 3, 3, 3]);
                                 {3}

  // --

  Test('A ≔ ComplexMatrix(❨❨3, 9, 2, 5, 1❩, ❨9, 2, 8, 2, 7❩, ❨2, 8, 3, 2, 5❩, ❨5, 2, 2, 6, 2❩, ❨1, 7, 5, 2, 4❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(21.437075526003142), -9.608848480894173, 6.159627148842105, 1.764018208004057, -1.751872401955133]);

  Test('col(E, 1) ∥ ❨0.438718328790090, 0.571040527424050, 0.444736969767807, 0.328057963353101, 0.419554459427209❩', True);
  Test('col(E, 2) ∥ ❨-0.517317497184398, 0.742773501365916, -0.312648829278767, 0.142651379319237, -0.250142811818416❩', True);
  Test('col(E, 3) ∥ ❨0.384970537144515, -0.154132214465530, -0.335181955611742, 0.737631277204982, -0.414239183945520❩', True);
  Test('col(E, 4) ∥ ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩', True);
  Test('col(E, 5) ∥ ❨-0.212274890341188, -0.050664616873610, 0.761222130728086, 0.108633310275057, -0.600926067928628❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('SameValue(EigenvalueOf(A, col(E, 5)), Λ[5], 1E-15)', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);

  Test('IsEigenvector(A, ❨0.438718328790090, 0.571040527424050, 0.444736969767807, 0.328057963353101, 0.419554459427209❩)', True);
  Test('IsEigenvector(A, ❨0.438728328790090, 0.571040527424050, 0.444736969767807, 0.328057963353101, 0.419554459427209❩)', False);
  Test('IsEigenvector(A, ❨0.438718328790090, 0.571041527424050, 0.444736969767807, 0.328057963353101, 0.419554459427209❩)', False);
  Test('IsEigenvector(A, ❨0.438718328790090, 0.571040527424050, 0.444746969767807, 0.328057963353101, 0.419554459427209❩)', False);
  Test('IsEigenvector(A, ❨0.438718328790090, 0.571040527424050, 0.444736969767807, 0.328027963353101, 0.419554459427209❩)', False);
  Test('IsEigenvector(A, ❨0.438718328790090, 0.571040527424050, 0.444736969767807, 0.328057963353101, 0.419555459427209❩)', False);

  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩, 1.764018208004057)', True);
  Test('IsEigenpair(A, ❨0.588761415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309632465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112468495387063, -0.562252131562491, -0.478222112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562257131562491, -0.478222112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478223112243157❩, 1.764018208004057)', False);
  Test('IsEigenpair(A, ❨0.588766415101779, 0.309639465780580, -0.112488495387063, -0.562252131562491, -0.478222112243157❩, 1.764014208004057)', False);

  // --

  Test('A ≔ ComplexMatrix(❨❨2, 0, 1, 0, 0❩, ❨0, 0, 0, 2, 0❩, ❨1, 0, 0, 1, 1❩, ❨0, 2, 1, 0, 0❩, ❨0, 0, 1, 0, 2❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(2.895106515927531), -2.292401585224621, 2, 2, -0.602704930702910]);
                                                                     {2}

  // --

  Test('A ≔ ComplexMatrix(❨❨0, 1, 1, 0, 0❩, ❨1, 0, 0, 0, 1❩, ❨1, 0, 0, 1, 0❩, ❨0, 0, 1, 0, 1❩, ❨0, 1, 0, 1, 0❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(2), -1.618033988749895, -1.618033988749895, 0.618033988749895, 0.618033988749895]);
                                  {2}                                    {2}

  // --

  Test('A ≔ ComplexMatrix(❨❨4.88518185435250995, -0.0503410125523369264, 2.41500433970266516, -0.3526553784998275, 0.377477078463321959❩, '+'❨-0.0503410125523369266, 6.99880168536003428, 0.0574866278829369505, -0.00839458885495496514, 0.00898544320902995499❩, ❨2.41500433970266516, 0.0574866278829369505, 4.24219977363724693, 0.402712768116436254, -0.431057764708227601❩, '+'❨-0.3526553784998275, -0.00839458885495496536, 0.402712768116436254, 6.94119313935299163, 0.0629459900627696699❩, ❨0.377477078463321959, 0.00898544320902995499, -0.431057764708227601, 0.0629459900627696697, 6.93262354729721721❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(7), 7, 7, 7, 2]);
                             {4}

  // --

  Test('A ≔ ComplexMatrix(❨❨4, 2, 8, 1, 7, 5❩, ❨2, 3, 9, 4, 1, 2❩, ❨8, 9, 7, 5, 2, 1❩, ❨1, 4, 5, 8, 0, 9❩, ❨7, 1, 2, 0, 3, 2❩, ❨5, 2, 1, 9, 2, 6❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(25.699931912397357), 10.666684085724908, -7.835696371596078, 7.409797588520018, -2.807084819210859, -2.133632395835343]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);

  // --

  Test('A ≔ ComplexMatrix(❨❨9, 2, 1, 3, 5, 4, 0❩, ❨2, 7, 2, 8, 3, 5, 2❩, ❨1, 2, 6, 9, 0, 3, 1❩, ❨3, 8, 9, 5, 4, 9, 2❩, ❨5, 3, 0, 4, 3, 7, 4❩, ❨4, 5, 3, 9, 7, 8, 4❩, ❨0, 2, 1, 2, 4, 4, 3❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(30.641044817744506384), 9.8517787330299055632, -6.5550797165196624249, 6.2412143228197048481, 3.6651048576610945760, -3.1057270151052322463, 0.26166400036968330015]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('SameValue(EigenvalueOf(A, col(E, 3)), Λ[3], 1E-14)', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);

  // --

  Test('A ≔ ComplexMatrix(❨❨3, 7, 1, 5, 2, 0, 6, 8❩, ❨7, 3, 9, 4, 7, 8, 2, 1❩, ❨1, 9, 5, 6, 0, 1, 9, 2❩, ❨5, 4, 6, 6, 3, 9, 1, 1❩, ❨2, 7, 0, 3, 4, 2, 6, 2❩, ❨0, 8, 1, 9, 2, 9, 0, 5❩, ❨6, 2, 9, 1, 6, 0, 4, 3❩, ❨8, 1, 2, 1, 2, 5, 3, 1❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-13); Test('Λ', [TASC(32.776937694011345), -15.258689129816567, 12.644822532655207, 6.717980397487340, -5.958148232873256, 4.141024331011820, -0.325483553395741, 0.261555960919853]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);
  Test('IsEigenvector(A, col(E, 8))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);
  Test('EigenvalueOf(A, col(E, 8)) ≈ Λ[8]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);
  Test('EigenvectorOf(A, Λ[8]) ∥ col(E, 8)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);
  Test('IsEigenpair(A, col(E, 8), Λ[8])', True);

  // --

  Test('A ≔ ComplexMatrix(❨❨3, 2, 1, 5, 9, 8, 7, 0, 2❩, ❨2, 5, 4, 8, 3, 0, 7, 1, 8❩, ❨1, 4, 2, 9, 3, 4, 5, 9, 2❩, ❨5, 8, 9, 8, 2, 6, 4, 8, 5❩, ❨9, 3, 3, 2, 0, 2, 9, 7, 1❩, ❨8, 0, 4, 6, 2, 5, 4, 8, 0❩, '+'❨7, 7, 5, 4, 9, 4, 3, 2, 5❩, ❨0, 1, 9, 8, 7, 8, 2, 8, 1❩, ❨2, 8, 2, 5, 1, 0, 5, 1, 9❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-13); Test('Λ', [TASC(41.547567826293928), 15.294422185576229, -12.701166678543432, 11.116110823642684, -8.128109328768399, -4.470204023755420, -3.750820326108707, 2.912735795595040, 1.179463726068098]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);
  Test('IsEigenvector(A, col(E, 8))', True);
  Test('IsEigenvector(A, col(E, 9))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);
  Test('EigenvalueOf(A, col(E, 8)) ≈ Λ[8]', True);
  Test('EigenvalueOf(A, col(E, 9)) ≈ Λ[9]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);
  Test('EigenvectorOf(A, Λ[8]) ∥ col(E, 8)', True);
  Test('EigenvectorOf(A, Λ[9]) ∥ col(E, 9)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);
  Test('IsEigenpair(A, col(E, 8), Λ[8])', True);
  Test('IsEigenpair(A, col(E, 9), Λ[9])', True);

  // --

  Test('A ≔ ComplexMatrix(❨❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩, ❨2, 3, 4, 5, 6, 7, 8, 9, 10, 1❩, ❨3, 4, 5, 6, 7, 8, 9, 10, 1, 2❩, ❨4, 5, 6, 7, 8, 9, 10, 1, 2, 3❩, ❨5, 6, 7, 8, 9, 10, 1, 2, 3, 4❩, '+'❨6, 7, 8, 9, 10, 1, 2, 3, 4, 5❩, ❨7, 8, 9, 10, 1, 2, 3, 4, 5, 6❩, ❨8, 9, 10, 1, 2, 3, 4, 5, 6, 7❩, ❨9, 10, 1, 2, 3, 4, 5, 6, 7, 8❩, ❨10, 1, 2, 3, 4, 5, 6, 7, 8, 9❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-13); Test('Λ', [TASC(55), -16.180339887498950, 16.180339887498950, -8.506508083520398, 8.506508083520398, -6.180339887498945, 6.180339887498937, -5.257311121191338, 5.257311121191338, -5]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);
  Test('IsEigenvector(A, col(E, 8))', True);
  Test('IsEigenvector(A, col(E, 9))', True);
  Test('IsEigenvector(A, col(E, 10))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);
  Test('EigenvalueOf(A, col(E, 8)) ≈ Λ[8]', True);
  Test('EigenvalueOf(A, col(E, 9)) ≈ Λ[9]', True);
  Test('EigenvalueOf(A, col(E, 10)) ≈ Λ[10]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);
  Test('EigenvectorOf(A, Λ[8]) ∥ col(E, 8)', True);
  Test('EigenvectorOf(A, Λ[9]) ∥ col(E, 9)', True);
  Test('EigenvectorOf(A, Λ[10]) ∥ col(E, 10)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);
  Test('IsEigenpair(A, col(E, 8), Λ[8])', True);
  Test('IsEigenpair(A, col(E, 9), Λ[9])', True);
  Test('IsEigenpair(A, col(E, 10), Λ[10])', True);

  // --

  Test('A ≔ ComplexMatrix(❨❨4.45528983757093495, -0.0878217640871465801, 0.587220808461187196, 0.381484594913282418, 0.806374962066634831, 0.415539821124804545, -1.07626065503415, -0.174476853005194865, 0.6117764065082528, 0.479713560928682166❩, '+'❨-0.0878217640871465801, 4.39333906061644079, -1.16146861964970503, 0.189604671732296926, 0.549866251807311702, 0.959584846946758171, -0.559867163033034075, 0.601342029002866998, 0.908395415505047237, -0.0481674127191538071❩, '+
    '❨0.587220808461187195, -1.16146861964970503, 4.731699905531466, -0.297368715494811859, 0.174137444046764151, 0.158862374235097095, 0.0708587702680495379, 0.527044617773810588, 0.717806236876411773, 0.100941787359586736❩, '+'❨0.381484594913282417, 0.189604671732296926, -0.297368715494811859, 5.20093245260149577, -0.422111106264731792, -0.609505049911372394, 0.430772405093700426, -0.075450459824149398, -0.0231177557010034982, 1.02271233462064678❩, '+
    '❨0.806374962066634831, 0.549866251807311702, 0.174137444046764151, -0.422111106264731792, 4.65483119820909985, 0.289795090828546352, 0.795841033040542465, -0.226199611183290607, -1.41597888676298831, 0.547038811317080901❩, '+'❨0.415539821124804545, 0.959584846946758171, 0.158862374235097095, -0.609505049911372394, 0.289795090828546352, 3.97252873681661176, 0.586614812369869111, -0.203645718668121608, 0.548222506108191402, -0.0260783651922211985❩, '+
    '❨-1.07626065503415, -0.559867163033034075, 0.0708587702680495379, 0.430772405093700426, 0.795841033040542465, 0.586614812369869111, 5.07303988296985066, 0.0852567827171805209, 0.721798919103536904, 0.0913845737686953809❩, '+'❨-0.174476853005194865, 0.601342029002866998, 0.527044617773810588, -0.0754504598241493979, -0.226199611183290607, -0.203645718668121608, 0.0852567827171805208, 5.72552879232507247, -0.381852810701084683, 0.248741195041378638❩, '+'❨0.6117764065082528, 0.908395415505047237, 0.717806236876411773, -0.0231177557010034982, -1.41597888676298831, 0.548222506108191402, 0.721798919103536904, -0.381852810701084683, 4.15028770968589867, 0.0862862168111860338❩, '+'❨0.479713560928682166, -0.0481674127191538071, 0.100941787359586736, 1.02271233462064678, 0.547038811317080901, -0.0260783651922211984, 0.0913845737686953809, 0.248741195041378638, 0.0862862168111860338, 3.64252242367312907❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(6), 6, 6, 6, 6, 6, 3, 3, 3, 1]);
                             {6}                {3}

  // --

  Test('A ≔ ComplexMatrix(❨❨0.947217218464429805, -0.0797032009555874838, 0.455681246175844459, -0.890506530889571131, 0.0810219664601976321, -0.4241414032969027, -0.325611817072624359, 0.442131323976784809, 0.696001040652687542, -0.363371579820000821❩, '+'❨-0.0797032009555874838, 0.664262086761071853, 0.151183874385912782, 0.366175763890158299, -0.362499363441370523, 0.345609605325798349, -0.36153060182114391, -0.370383371480650959, -0.351151983839811129, -0.192995673028588803❩, '+
    '❨0.455681246175844459, 0.151183874385912782, 0.965112845520727124, -0.379442604818156945, -0.0868548299522641688, -0.703208127497219295, -0.29273075622571167, 0.392533036681992209, 0.433785161673175889, -0.0198556703071645884❩, '+'❨-0.890506530889571131, 0.366175763890158299, -0.379442604818156945, 1.60196556896465841, 0.0343051071296422081, 0.587093406299823787, -0.485735414772669726, -0.766652742695679709, -0.63672778172637717, -0.0824745994586258031❩, '+
    '❨0.0810219664601976321, -0.362499363441370523, -0.0868548299522641688, 0.0343051071296422081, 0.327534361557092576, -0.222948382709437083, -0.0500390113869091228, 0.161756725513035523, 0.21826384416565025, -0.135644537928391234❩, '+'❨-0.4241414032969027, 0.345609605325798349, -0.703208127497219295, 0.587093406299823787, -0.222948382709437083, 0.989037781112786131, -0.126211410841058615, -0.699854415829559306, -0.413031006405435901, 0.131956448780506888❩, '+
    '❨-0.325611817072624359, -0.36153060182114391, -0.29273075622571167, -0.485735414772669726, -0.0500390113869091228, -0.126211410841058615, 1.00217393220066957, 0.281471732316985816, -0.335032854995363073, 0.398331000774434439❩, '+'❨0.442131323976784809, -0.370383371480650959, 0.392533036681992209, -0.766652742695679709, 0.161756725513035523, -0.699854415829559306, 0.281471732316985816, 0.61286300125401041, 0.371247429623465179, -0.0611706531297214894❩, '+
    '❨0.696001040652687542, -0.351151983839811129, 0.433785161673175889, -0.63672778172637717, 0.21826384416565025, -0.413031006405435901, -0.335032854995363073, 0.371247429623465179, 0.960656361843560564, 0.149752326353616183❩, '+'❨-0.363371579820000821, -0.192995673028588803, -0.0198556703071645884, -0.0824745994586258031, -0.135644537928391234, 0.131956448780506888, 0.398331000774434439, -0.0611706531297214894, 0.149752326353616183, 0.929176842320993553❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(4), 2, 1, 1, 1, 0, 0, 0, 0, 0]);
                                    {3}      {5}

  // --

  Test('A ≔ ComplexMatrix(ReversalMatrix(10));', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(-1), -1, -1, -1, -1, 1, 1, 1, 1, 1]);
                              {5}                 {5}

  // --

  Test('A ≔ ComplexMatrix(diag(12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0));', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Test('IsIdentity(E)', True);
  Test('A = diag(Λ)', True);

  // --

  Test('A ≔ ComplexMatrix(IdentityMatrix(10));', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Test('IsIdentity(E)', True);
  Test('IsIdentity(diag(Λ))', True);

  // --

  Test('A ≔ ComplexMatrix(ZeroMatrix(10));', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsOrthogonal(E)', True);
  Test('IsUnitary(E)', True);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Test('IsIdentity(E)', True);
  Test('IsZero(Λ)', True);

  // - eigenvectors of real non-symmetric matrices (complex data type)

  Test('A ≔ ComplexMatrix(❨❨5, 2, 1❩, ❨0, 2, 4❩, ❨1, 2, 1❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [TASC(6), 3.236067977499790, -1.236067977499789]);

  Test('col(E, 1) ∥ ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩', True);
  Test('col(E, 2) ∥ ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩', True);
  Test('col(E, 3) ∥ ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);

  Test('IsEigenvector(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩)', True);
  Test('IsEigenvector(A, ❨0.904554033733291, 0.301511344577764, 0.301511344577764❩)', False);
  Test('IsEigenvector(A, ❨0.904534033733291, 0.301571344577764, 0.301511344577764❩)', False);
  Test('IsEigenvector(A, ❨0.904534033733291, 0.301511344577764, 0.301510344577764❩)', False);

  Test('IsEigenvector(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩)', True);
  Test('IsEigenvector(A, ❨0.781038693640056, -0.596654357427190, -0.184376336212866❩)', False);
  Test('IsEigenvector(A, ❨0.781030693640056, -0.596650357427190, -0.184376336212866❩)', False);
  Test('IsEigenvector(A, ❨0.781030693640056, -0.596654357427190, -0.184374336212866❩)', False);

  Test('IsEigenvector(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩)', True);
  Test('IsEigenvector(A, ❨0.146862297110534, -0.769007151342417, 0.622139854231882❩)', False);
  Test('IsEigenvector(A, ❨0.146867297110534, -0.769004151342417, 0.622139854231882❩)', False);
  Test('IsEigenvector(A, ❨0.146867297110534, -0.769007151342417, 0.622199854231882❩)', False);

  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩, 6)', True);
  Test('IsEigenpair(A, ❨0.904535033733291, 0.301511344577764, 0.301511344577764❩, 6)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301514344577764, 0.301511344577764❩, 6)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301510344577764❩, 6)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩, 7)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩, 6.0001)', False);
  Test('IsEigenpair(A, ❨0.904534033733291, 0.301511344577764, 0.301511344577764❩, -6)', False);

  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, 3.236067977499790)', True);
  Test('IsEigenpair(A, ❨0.781034693640056, -0.596654357427190, -0.184376336212866❩, 3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596644357427190, -0.184376336212866❩, 3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184370336212866❩, 3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, 3.236068977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, -3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, 0)', False);
  Test('IsEigenpair(A, ❨0.781030693640056, -0.596654357427190, -0.184376336212866❩, 6)', False);

  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, -1.236067977499789)', True);
  Test('IsEigenpair(A, ❨0.146865297110534, -0.769007151342417, 0.622139854231882❩, -1.236067977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769047151342417, 0.622139854231882❩, -1.236067977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622130854231882❩, -1.236067977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, -1.236064977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 1.236067977499789)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 6)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 3.236067977499790)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 0)', False);
  Test('IsEigenpair(A, ❨0.146867297110534, -0.769007151342417, 0.622139854231882❩, 1E100)', False);

  // -

  Test('A ≔ ComplexMatrix(❨❨2, 3, 0, 1❩, ❨5, 6, 8, 1❩, ❨4, 2, 5, 7❩, ❨3, 0, 2, 6❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [TASC(13.109772228646445), 3.890227771353548, 3, -1]);

  Test('col(E, 1) ∥ ❨0.233782511971403, 0.784104790434032, 0.520117226972313, 0.244956087740999❩', True);
  Test('col(E, 2) ∥ ❨-0.596105141205652, -0.513975993546188, 0.456218061914745, 0.415153488085013❩', True);
  Test('col(E, 3) ∥ ❨-0.696310623822791, -0.348155311911396, 0.522232967867093, 0.348155311911397❩', True);
  Test('col(E, 4) ∥ ❨-0.713563032023560, 0.600041640565266, -0.121630062276743, 0.340564174374881❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);

  // -

  Test('A ≔ ComplexMatrix(❨❨6, 4, 2, 1, 3❩, ❨5, 0, 4, 7, 3❩, ❨9, 5, 3, 0, 2❩, ❨4, 4, 2, 8, 5❩, ❨3, 6, 0, 4, 7❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [TASC(19.317943284512630), 6.195170660848524, -3.837738100539873, 2.388435089272784, -0.063810934094071]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);

  // -

  Test('A ≔ ComplexMatrix(❨❨3, 6, 2, 5, 4, 7❩, ❨2, 0, 1, 4, 2, 0❩, ❨3, 6, 9, 8, 5, 2❩, ❨2, 7, 6, 4, 3, 1❩, ❨5, 6, 9, 8, 2, 4❩, ❨6, 5, 3, 9, 5, 7❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [TASC(25.608230542132779785), 6.8775642608200556778, -4.2970302684870303417, -3.5268871389883464722, 0.86082458257068375765, -0.52270197804814240633]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);

  // -

  Test('A ≔ ComplexMatrix(❨❨0, 2, 0❩, ❨1, −1, 1❩, ❨2, 0, 1❩❩);', null);
  Eps(1E-14); Test('eigenvalues(A)', [2, -1, -1]); // -1: geometric multiplicity = 1 < 2 = algebraic multiplicity
  Test('eigenvectors(A)', failure, 'Couldn''t compute eigenvectors.');

  // - eigenvectors of non-real Hermitian matrices

  Test('A ≔ ❨❨2, i❩, ❨-i, 7❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(7.192582403567252), 1.807417596432748]);

  // -

  Test('A ≔ ❨❨2, 5, i❩, ❨5, 7, −i❩, ❨−i, i, 4❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(10.107946815755056), 4.331111110781892, -1.439057926536950]);

  Test('col(E, 1) ∥ ❨0.519222484861137⋅i,  0.852891389063453⋅i, -0.054628652518984❩', True);
  Test('col(E, 2) ∥ ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.967804067246890❩', True);
  Test('col(E, 3) ∥ ❨0.820912311491764⋅i, -0.515491977723137⋅i, -0.245705103211833❩', True);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);

  Test('IsEigenvector(A, ❨0.519222484861137⋅i,  0.852891389063453⋅i, -0.054628652518984❩)', True);
  Test('IsEigenvector(A, ❨0.519232484861137⋅i,  0.852891389063453⋅i, -0.054628652518984❩)', False);
  Test('IsEigenvector(A, ❨0.519222484861137⋅i,  0.852898389063453⋅i, -0.054628652518984❩)', False);
  Test('IsEigenvector(A, ❨0.519222484861137⋅i,  0.852891389063453⋅i, -0.054620652518984❩)', False);
  Test('IsEigenvector(A, ❨0.519222484861137⋅i,  0.00003 + 0.852891389063453⋅i, -0.054628652518984❩)', False);
  Test('IsEigenvector(A, ❨0.00004 + 0.519222484861137⋅i,  0.852891389063453⋅i, -0.054628652518984❩)', False);
  Test('IsEigenvector(A, ❨0.519222484861137,  0.852891389063453⋅i, -0.054628652518984❩)', False);
  Test('IsEigenvector(A, ❨0.519222484861137,  0.852891389063453, -0.054628652518984❩)', False);

  Test('IsEigenvector(A, ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.967804067246890❩)', True);
  Test('IsEigenvector(A, ❨0.237725398905681⋅i, -0.082730280819671⋅i, 0.967804067246890❩)', False);
  Test('IsEigenvector(A, ❨0.237720398905681⋅i, -0.082737280819671⋅i, 0.967804067246890❩)', False);
  Test('IsEigenvector(A, ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.967864067246890❩)', False);
  Test('IsEigenvector(A, ❨0.237720398905681⋅i, -0.082730280819671, 0.967804067246890❩)', False);
  Test('IsEigenvector(A, ❨0.237720398905681, -0.082730280819671⋅i, 0.967804067246890❩)', False);
  Test('IsEigenvector(A, ❨0.237720398905681, -0.082730280819671, 0.967804067246890❩)', False);
  Test('IsEigenvector(A, ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.967804067246890⋅i❩)', False);
  Test('IsEigenvector(A, ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.6 + 0.967804067246890❩)', False);
  Test('IsEigenvector(A, ❨0.237720398905681⋅i, 0.082730280819671⋅i, 0.967804067246890❩)', False);

  Test('IsEigenvector(A, ❨0.820912311491764⋅i, -0.515491977723137⋅i, -0.245705103211833❩)', True);
  Test('IsEigenvector(A, ❨0.820911311491764⋅i, -0.515491977723137⋅i, -0.245705103211833❩)', False);
  Test('IsEigenvector(A, ❨0.820912311491764⋅i, -0.515495977723137⋅i, -0.245705103211833❩)', False);
  Test('IsEigenvector(A, ❨0.820912311491764⋅i, -0.515491977723137⋅i, -0.245707103211833❩)', False);
  Test('IsEigenvector(A, ❨0.820912311491764⋅i, -0.515491977723137⋅i, 0.245705103211833❩)', False);
  Test('IsEigenvector(A, ❨0.820912311491764⋅i, -0.515491977723137⋅i + 0.0006, -0.245705103211833❩)', False);
  Test('IsEigenvector(A, ❨0.820912311491764, -0.515491977723137⋅i, -0.245705103211833❩)', False);
  Test('IsEigenvector(A, ❨0.00006 - 0.820912311491764⋅i, -0.515491977723137⋅i, -0.245705103211833❩)', False);

  Test('IsEigenvector(A, ❨1, 2❩)', failure, 'Vector is of wrong dimension.');

  Test('EigenvalueOf(A, ❨1, 2, 3❩)', failure, 'A vector which isn''t an eigenvector has no associated eigenvalue.');
  Test('EigenvalueOf(A, ❨1, 2❩)', failure, 'Vector is of wrong dimension.');

  Test('EigenvectorOf(A, 5)', failure, 'Couldn''t compute associated eigenvector.');

  Test('IsEigenpair(A, ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.967804067246890❩, 4.331111110781892)', True);
  Test('IsEigenpair(A, ❨0.237730398905681⋅i, -0.082730280819671⋅i, 0.967804067246890❩, 4.331111110781892)', False);
  Test('IsEigenpair(A, ❨0.237720398905681⋅i, -0.082740280819671⋅i, 0.967804067246890❩, 4.331111110781892)', False);
  Test('IsEigenpair(A, ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.967864067246890❩, 4.331111110781892)', False);
  Test('IsEigenpair(A, ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.967804067246890❩, 4.331171110781892)', False);
  Test('IsEigenpair(A, ❨0.237720398905681⋅i, 0.00001-0.082730280819671⋅i, 0.967804067246890❩, 4.331111110781892)', False);
  Test('IsEigenpair(A, ❨0.237720398905681, -0.082730280819671⋅i, 0.967804067246890❩, 4.331111110781892)', False);
  Test('IsEigenpair(A, ❨0.237720398905681, -0.082730280819671, 0.967804067246890❩, 4.331111110781892)', False);
  Test('IsEigenpair(A, ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.967804067246890 + 0.0005⋅i❩, 4.331111110781892)', False);
  Test('IsEigenpair(A, ❨0.237720398905681⋅i, -0.082730280819671⋅i, 0.967804067246890❩, 10.107946815755056)', False);

  // -

  Test('A ≔ ❨❨1, i, 0❩, ❨−i, 3, 0❩, ❨0, 0, 0❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(3.414213562373095), 0.585786437626905, 0]);
                                                                    {0}
  // -

  Test('A ≔ ❨❨3, i, 8, 2❩, ❨−i, 5, 3, i❩, ❨8, 3, 2, 7❩, ❨2, −i, 7, 1❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(14.219175497193367), -8.046249768016203, 5.279072848214093, -0.451998577391262]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('SameValue(EigenvalueOf(A, col(E, 3)), Λ[3], 1E-15)', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);

  // -

  Test('A ≔ ❨❨0, 0, 0, i❩, ❨0, 0, 1, 0❩, ❨0, 1, 0, 0❩, ❨−i, 0, 0, 0❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(-1), -1, 1, 1]);
                              {2}     {2}

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);

  // -

  Test('A ≔ ❨❨5, 2, −i, 7, 4❩, ❨2, 3, 9, 1, i❩, ❨i, 9, 5, 9, 8❩, ❨7, 1, 9, 2, 3❩, ❨4, −i, 8, 3, 6❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(22.691445962489553), -11.503186518668644, 7.060493996470923, 3.839332450777222, -1.088085891069047]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);

  // -

  Test('A ≔ ❨❨0, 0, −i, −1, 0❩, ❨0, 0, 0, 0, 1❩, ❨i, 0, 0, −i, 0❩, ❨−1, 0, i, 0, 0❩, ❨0, 1, 0, 0, 0❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E, 1E-15)', True);
  Test('SameValue(A, E ⋅ diag(Λ) ⋅ E*, 1E-14)', True);

  Eps(1E-14); Test('Λ', [TASC(2), -1, -1, -1, 1]);
                                  {3}

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('SameValue(EigenvalueOf(A, col(E, 2)), Λ[2], 1E-14)', True);
  Test('SameValue(EigenvalueOf(A, col(E, 3)), Λ[3], 1E-14)', True);
  Test('SameValue(EigenvalueOf(A, col(E, 4)), Λ[4], 1E-13)', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);

  // -

//  Test('A ≔ ❨❨0, 0, 0, 1, 1−i❩, ❨0, 0, 0, i, 0❩, ❨0, 0, 0, 0, 1❩, ❨1, −i, 0, 0, 0❩, ❨1+i, 0, 1, 0, 0❩❩;', null);  // TODO. Wishlist. (Currently obtains the complete spectrum and all eigenspaces except the 0 one (the nullspace).)
//  Test('(Λ, E) ≔ eigenvectors(A);', null);
//  Test('IsReal(Λ)', True);
//  Test('IsUnitary(E)', True);
//  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);
//
//  Eps(1E-14); Test('Λ', [TASC(-2), 2, -1, 1, 0]);

  // -

  Test('A ≔ ❨❨9, 2, i, 2, i, 5❩, ❨2, 8, 0, 2, 3, 1❩, ❨−i, 0, 5, −i, 3, 4❩, ❨2, 2, i, 6, 2, 1❩, ❨−i, 3, 3, 2, 3, i❩, ❨5, 1, 4, 1, −i, 0❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E)', True);
  Test('SameValue(A, E ⋅ diag(Λ) ⋅ E*, 1E-15)', True);

  Eps(1E-14); Test('Λ', [TASC(14.589433207624165), 8.499935664587964, 7.224883679054878, 4.563236528531559, -4.322019577002851, 0.444530497204293]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('SameValue(EigenvalueOf(A, col(E, 4)), Λ[4], 1E-15)', True);
  Test('SameValue(EigenvalueOf(A, col(E, 5)), Λ[5], 1E-15)', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);

  // -

  Test('A ≔ ❨❨2, 0, 8, 3, i, 4, 5❩, ❨0, 6, 3, 7, 1, 2, 0❩, ❨8, 3, 8, 9, 3, i, 6❩, ❨3, 7, 9, 5, 2, 9, −i❩, ❨−i, 1, 3, 2, 7, 3, 1❩, ❨4, 2, −i, 9, 3, 5, i❩, ❨5, 0, 6, i, 1, −i, 2❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-14); Test('Λ', [TASC(26.784922520579256993), 10.967618684890724839, -9.3773999257987389970, 6.6504097990874679751, 4.7855475613601607466, -4.2277575354832616959, -0.58334110463560986115]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('SameValue(EigenvalueOf(A, col(E, 5)), Λ[5], 1E-14)', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);

  // -

  Test('A ≔ ❨❨3, 2, 8, 0, i, 5, 7, 2❩, ❨2, 3, 9, 2, 7, 1, 2, 5❩, ❨8, 9, 9, 0, 3, 1, 5, 7❩, ❨0, 2, 0, 6, i, 3, 0, 8❩, ❨−i, 7, 3, −i, 7, 6, 1, −i❩, ❨5, 1, 1, 3, 6, 5, 2, 1❩, ❨7, 2, 5, 0, 1, 2, 3, 9❩, ❨2, 5, 7, 8, i, 1, 9, 6❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('IsReal(Λ)', True);
  Test('IsUnitary(E)', True);
  Test('A ≈ E ⋅ diag(Λ) ⋅ E*', True);

  Eps(1E-13); Test('Λ', [TASC(31.541450879650441), 12.451388642998566, 9.945326338353777, -9.132216213673990, 6.957054234677591, -5.929873813622360, -4.821944371529606, 0.988814303145571]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);
  Test('IsEigenvector(A, col(E, 6))', True);
  Test('IsEigenvector(A, col(E, 7))', True);
  Test('IsEigenvector(A, col(E, 8))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);
  Test('EigenvalueOf(A, col(E, 6)) ≈ Λ[6]', True);
  Test('EigenvalueOf(A, col(E, 7)) ≈ Λ[7]', True);
  Test('EigenvalueOf(A, col(E, 8)) ≈ Λ[8]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);
  Test('EigenvectorOf(A, Λ[6]) ∥ col(E, 6)', True);
  Test('EigenvectorOf(A, Λ[7]) ∥ col(E, 7)', True);
  Test('EigenvectorOf(A, Λ[8]) ∥ col(E, 8)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);
  Test('IsEigenpair(A, col(E, 6), Λ[6])', True);
  Test('IsEigenpair(A, col(E, 7), Λ[7])', True);
  Test('IsEigenpair(A, col(E, 8), Λ[8])', True);

  // - eigenvectors of non-real non-Hermitian matrices

  Test('A ≔ ❨❨3, 5, i❩, ❨1, i, −2❩, ❨4, 0, 1❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [3.796762780724884 + 2.608464621094820*ImaginaryUnit, 2.828938723402460 - 1.856597124782495*ImaginaryUnit, -2.625701504127345 + 0.248132503687670*ImaginaryUnit]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);

  // -

//  Test('A ≔ ❨❨2, 0, i❩, ❨1, 0, 0❩, ❨0, 0, i❩❩;', null);                        // TODO. Wishlist.
//  Test('(Λ, E) ≔ eigenvectors(A);', null);
//  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);
//
//  Eps(1E-14); Test('Λ', [2, ImaginaryUnit, 0]);

  // -

  Test('A ≔ ❨❨3, 5, i, 4❩, ❨0, 5, 3, 8❩, ❨i, 4, 2, 6❩, ❨1, 3, 2, i❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [11.107149435941459 + 0.533069641032313*ImaginaryUnit, -3.378066601503166 + 0.872708325271380*ImaginaryUnit, 2.593743367634961 - 1.064199910826719*ImaginaryUnit, -0.322826202073264 + 0.658421944523025*ImaginaryUnit]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);

  // -

  Test('A ≔ ❨❨1.57519401408085795 + 0.432308033321589517⋅i, -0.315432944876384471 + 0.121542553437335565⋅i, -1.53883608840310379 + 0.248576747227350505⋅i, -0.632999052211440081 + 0.410919464237228956⋅i❩, '+'❨-0.315432944876384471 + 0.121542553437335565⋅i, 1.94355816819509714 + 0.0341714498862397758⋅i, 0.0315808797266594304 + 0.069886863658359134⋅i, -0.262122163108155187 + 0.115529199299754214⋅i❩, '+'❨-1.53883608840310379 + 0.248576747227350505⋅i, 0.0315808797266594304 + 0.069886863658359134⋅i, 4.18134443375707839 + 0.142931415794822536⋅i, -0.919820668174180992 + 0.23627833840531786⋅i❩, '+'❨-0.632999052211440081 + 0.410919464237228956⋅i, -0.262122163108155187 + 0.115529199299754214⋅i, -0.919820668174180992 + 0.23627833840531786⋅i, 1.29990338396696651 + 0.390589100997348171⋅i❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('SameValue(A, E ⋅ diag(Λ) ⋅ inv(E), 1E-15)', True);

  Eps(1E-14); Test('Λ', [5, 2, 2, ImaginaryUnit]);
                           {2}
  // -

  Test('A ≔ ❨❨2, 4, i, 3, 6❩, ❨i, 2, 6, 3, 0❩, ❨9, 3, 0, 2, 5❩, ❨4, 2, 6, i, 4❩, ❨7, 3, 0, 6, 8❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [17.606360994938018071 + 0.565120228698768441*ImaginaryUnit, -2.8607289727538659348 - 4.4336014674947297475*ImaginaryUnit, -1.9673163287759711292 + 4.6514948618516669997*ImaginaryUnit, -3.4143709271513932924 + 0.1742140430021373782*ImaginaryUnit, 2.6360552337432122859 + 0.0427723339421569283*ImaginaryUnit]);

  Test('IsEigenvector(A, col(E, 1))', True);
  Test('IsEigenvector(A, col(E, 2))', True);
  Test('IsEigenvector(A, col(E, 3))', True);
  Test('IsEigenvector(A, col(E, 4))', True);
  Test('IsEigenvector(A, col(E, 5))', True);

  Test('EigenvalueOf(A, col(E, 1)) ≈ Λ[1]', True);
  Test('EigenvalueOf(A, col(E, 2)) ≈ Λ[2]', True);
  Test('EigenvalueOf(A, col(E, 3)) ≈ Λ[3]', True);
  Test('EigenvalueOf(A, col(E, 4)) ≈ Λ[4]', True);
  Test('EigenvalueOf(A, col(E, 5)) ≈ Λ[5]', True);

  Test('EigenvectorOf(A, Λ[1]) ∥ col(E, 1)', True);
  Test('EigenvectorOf(A, Λ[2]) ∥ col(E, 2)', True);
  Test('EigenvectorOf(A, Λ[3]) ∥ col(E, 3)', True);
  Test('EigenvectorOf(A, Λ[4]) ∥ col(E, 4)', True);
  Test('EigenvectorOf(A, Λ[5]) ∥ col(E, 5)', True);

  Test('IsEigenpair(A, col(E, 1), Λ[1])', True);
  Test('IsEigenpair(A, col(E, 2), Λ[2])', True);
  Test('IsEigenpair(A, col(E, 3), Λ[3])', True);
  Test('IsEigenpair(A, col(E, 4), Λ[4])', True);
  Test('IsEigenpair(A, col(E, 5), Λ[5])', True);

  // -

  Test('A ≔ ❨❨3, 6, 2, 0, 1, 4❩, ❨i, 5, 2, 6, i, 8❩, ❨4, 1, 5, 2, 0, i❩, ❨3, 6, 2, 5, 4, 7❩, ❨i, 5, 2, 3, 0, 9❩, ❨9, 5, 4, 2, i, 4❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [20.338027259929576518 + 0.850874501029312238*ImaginaryUnit, -3.3305909558888180389 - 4.0193288703545500750*ImaginaryUnit, -1.7862303576469919606 + 3.6512377259081217118*ImaginaryUnit, 3.1004341929871214890 + 0.2640193696616336506*ImaginaryUnit, 1.5283992101013815889 - 2.6017126813340652695*ImaginaryUnit, 2.1499606505177304036 + 1.8549099550895477446*ImaginaryUnit]);

  // -

  Test('A ≔ ❨❨2, 5, 1, i, 5, 3, 6❩, ❨4, 5, 2, 8, 5, 6, i❩, ❨i, i, 2, 4, 0, 1, 8❩, ❨5, 0, 4, 7, 8, 6, 5❩, ❨2, 5, i, 3, 6, 9, 4❩, ❨7, 4, 5, 2, 0, 3, 6❩, ❨i, 4, 7, 1, 2, 4, 5❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  // -

  Test('A ≔ ❨❨4, 2, 5, 3, 0, i, 4, 5❩, ❨5, 6, 3, 2, 1, 4, 5, 7❩, ❨5, 4, 3, 6, 2, 5, 7, 8❩, ❨5, 6, 3, 0, 2, 8, 5, 4❩, ❨7, i, 5, 6, 4, 1, 2, 2❩, ❨6, 9, 8, 7, i, 4, 5, 6❩, ❨4, 3, 0, 8, 5, 7, 9, i❩, ❨5, 6, 3, 6, 1, 7, 4, 5❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  // -

  Test('A ≔ ❨❨5, 4, 1, 2, 5, 3, 6, 9, 5❩, ❨4, 1, 7, i, 5, 6, 3, 2, 5❩, ❨8, 5, 2, 5, 3, 6, 2, 1, 0❩, ❨4, 5, 9, 6, 5, 8, 7, 4, 0❩, ❨6, 3, 5, 2, 6, 5, 8, 9, 4❩, ❨4, 5, 2, 3, 1, 7, 5, 6, 8❩, ❨5, 4, 2, 1, 3, 6, 5, 8, 9❩, '+'❨4, 1, 1, 5, 6, 8, 5, 9, 7❩, ❨5, i, 1, 2, 3, 6, 5, 7, 8❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  // -

  Test('A ≔ ❨❨−i, 1, i, 1, i, 1, i, 1, i, −1❩, ❨i, 2, −i, 2, i, 2, i, −2, i, 2❩, ❨i, 3, i, 3, −i, −3, i, 3, i, 3❩, ❨i, 4, i, −4, i, 4, −i, 4, i, 4❩, ❨i, −5, i, 5, i, 5, i, 5, −i, 5❩, ❨−i, 6, i, 6, i, 6, i, 6, i, −6❩, '+'❨i, 7, −i, 7, i, 7, i, −7, i, 7❩, ❨i, 8, i, 8, −i, −8, i, 8, i, 8❩, ❨i, 9, i, −9, i, 9, −i, 9, i, 9❩, ❨i, 0, i, 0, i, 0, i, 0, −i, 0❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  // -

  Test('A ≔ ❨❨1, 0, i❩, ❨1, 1, 1❩, ❨0, 0, i❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', failure, 'Couldn''t compute eigenvectors.'); // eigenvalue 1: geometric multiplicity = 1 < 2 = algebraic multiplicity

  // - real matrices with non-real eigenvalues

  Test('A ≔ ❨❨2, 3, 4❩, ❨4, 5, −7❩, ❨1, 2, −2❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [6.475176326140790,  -0.737588163070393 + 1.734355939710355*ImaginaryUnit, -0.737588163070393 - 1.734355939710355*ImaginaryUnit]);

  Test('A ≔ ComplexMatrix(❨❨2, 3, 4❩, ❨4, 5, −7❩, ❨1, 2, −2❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [6.475176326140790,  -0.737588163070393 + 1.734355939710355*ImaginaryUnit, -0.737588163070393 - 1.734355939710355*ImaginaryUnit]);

  // -

  Test('A ≔ ❨❨5, 2, 0, 1❩, ❨3, 0, 5, 8❩, ❨4, 3, 1, 9❩, ❨3, 6, 0, 7❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [14.036206157473359, 4, -2.518103078736678 + 1.829859219363527*ImaginaryUnit, -2.518103078736678 - 1.829859219363527*ImaginaryUnit]);

  Test('A ≔ ComplexMatrix(❨❨5, 2, 0, 1❩, ❨3, 0, 5, 8❩, ❨4, 3, 1, 9❩, ❨3, 6, 0, 7❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [14.036206157473359, 4, -2.518103078736678 + 1.829859219363527*ImaginaryUnit, -2.518103078736678 - 1.829859219363527*ImaginaryUnit]);

  // -

  Test('A ≔ ❨❨−6, 2, 3, 5, −8❩, ❨4, 0, 2, 6, 5❩, ❨0, 0, 1, 6, 7❩, ❨5, 4, 8, 0, 2❩, ❨4, 3, 6, 6, 2❩❩;', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [14.738054917903042, -9.047156047492546, -3.898428540473577 + 1.621125566153341*ImaginaryUnit, -3.898428540473577 - 1.621125566153341*ImaginaryUnit, -0.894041789463348]);

  Test('A ≔ ComplexMatrix(❨❨−6, 2, 3, 5, −8❩, ❨4, 0, 2, 6, 5❩, ❨0, 0, 1, 6, 7❩, ❨5, 4, 8, 0, 2❩, ❨4, 3, 6, 6, 2❩❩);', null);
  Test('(Λ, E) ≔ eigenvectors(A);', null);
  Test('A ≈ E ⋅ diag(Λ) ⋅ inv(E)', True);

  Eps(1E-14); Test('Λ', [14.738054917903042, -9.047156047492546, -3.898428540473577 + 1.621125566153341*ImaginaryUnit, -3.898428540473577 - 1.621125566153341*ImaginaryUnit, -0.894041789463348]);

  // - epsilons

  Test('A ≔ ❨❨5, 1, 7❩, ❨1, 2, 4❩, ❨7, 4, 3❩❩;', null);

  Test('IsEigenvector(A, ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩)', True);

  Test('IsEigenvector(A, ❨0.681, 0.323, 0.656❩)', False);

  Test('IsEigenvector(A, ❨0.681, 0.323, 0.656❩, 1E-6)', False);
  Test('IsEigenvector(A, ❨0.681, 0.323, 0.656❩, 1E-5)', False);
  Test('IsEigenvector(A, ❨0.681, 0.323, 0.656❩, 1E-4)', False);

  Test('IsEigenvector(A, ❨0.681, 0.323, 0.656❩, 1E-2)', True);
  Test('IsEigenvector(A, ❨0.681, 0.323, 0.656❩, 1E-1)', True);

  Test('IsEigenpair(A, ❨0.681034009183674, 0.323520554185660, 0.656907245624993❩, 12.227056515812240)', True);

  Test('IsEigenpair(A, ❨0.68103, 0.32352, 0.65690❩, 12.22705)', False);

  Test('IsEigenpair(A, ❨0.68103, 0.32352, 0.65690❩, 12.22705, 1E-8)', False);
  Test('IsEigenpair(A, ❨0.68103, 0.32352, 0.65690❩, 12.22705, 1E-7)', False);
  Test('IsEigenpair(A, ❨0.68103, 0.32352, 0.65690❩, 12.22705, 1E-6)', False);

  Test('IsEigenpair(A, ❨0.68103, 0.32352, 0.65690❩, 12.22705, 1E-4)', True);
  Test('IsEigenpair(A, ❨0.68103, 0.32352, 0.65690❩, 12.22705, 1E-3)', True);
  Test('IsEigenpair(A, ❨0.68103, 0.32352, 0.65690❩, 12.22705, 1E-2)', True);
  Test('IsEigenpair(A, ❨0.68103, 0.32352, 0.65690❩, 12.22705, 1E-1)', True);

  // - spectral radius

  Eps(1E-13); Test('A ≔ ❨❨21❩❩; SpectralRadius(A)', 21.0);
  Eps(1E-13); Test('A ≔ ❨❨3, 5❩, ❨5, 2❩❩; SpectralRadius(A)', 7.524937810560445);
  Eps(1E-13); Test('A ≔ ❨❨4, 1, 8❩, ❨1, 2, 5❩, ❨8, 5, 3❩❩; SpectralRadius(A)', 13.136085649145485);
  Eps(1E-13); Test('A ≔ ❨❨2, 0, 1❩, ❨0, 2, 0❩, ❨1, 0, 2❩❩; SpectralRadius(A)', 3.0);
  Eps(1E-13); Test('A ≔ ❨❨0, 0, 2❩, ❨0, 2, 0❩, ❨2, 0, 0❩❩; SpectralRadius(A)', 2.0);
  Eps(1E-13); Test('A ≔ ❨❨6, 1, 3, 2❩, ❨1, 8, 5, 1❩, ❨3, 5, 2, 0❩, ❨2, 1, 0, 7❩❩; SpectralRadius(A)', 12.317576522344481);
  Eps(1E-13); Test('A ≔ ❨❨0, 1, 0, 2❩, ❨1, -2, 2, 0❩, ❨0, 2, 0, -1❩, ❨2, 0, -1, 2❩❩; SpectralRadius(A)', 3.449489742783178);
  Eps(1E-13); Test('A ≔ ❨❨0, 1, 0, 2❩, ❨1, -2, 2, 0❩, ❨0, 2, 0, -1❩, ❨2, 0, -1, -2❩❩; SpectralRadius(A)', 3.449489742783178);
  Eps(1E-13); Test('A ≔ ❨❨0, 0, 1, 2❩, ❨0, -2, 0, 0❩, ❨1, 0, 0, 1❩, ❨2, 0, 1, 0❩❩; SpectralRadius(A)', 2.732050807568878);
  Eps(1E-13); Test('A ≔ ❨❨3.09090210854135331, -0.0989480787770248203, 0.206696174292443883, -0.173565216686099517❩, '+'❨-0.0989480787770248203, 3.10770621771892455, -0.224991363401589891, 0.188927902874714558❩, ❨0.206696174292443883, -0.224991363401589891, 3.4699924914029535, -0.394658241210492851❩, '+'❨-0.173565216686099517, 0.188927902874714558, -0.394658241210492851, 3.33139918233676864❩❩; SpectralRadius(A)', 4.0);
  Eps(1E-13); Test('A ≔ ❨❨3, 2, 1, 0, 1❩, ❨2, 5, 7, 3, 8❩, ❨1, 7, 6, 0, 2❩, ❨0, 3, 0, 5, 3❩, ❨1, 8, 2, 3, 9❩❩; SpectralRadius(A)', 19.613776982166058);
  Eps(1E-13); Test('A ≔ ❨❨2, 0, 1, 0, 0❩, ❨0, 0, 0, 2, 0❩, ❨1, 0, 0, 1, 1❩, ❨0, 2, 1, 0, 0❩, ❨0, 0, 1, 0, 2❩❩; SpectralRadius(A)', 2.895106515927531);
  Eps(1E-13); Test('A ≔ ❨❨0, 1, 1, 0, 0❩, ❨1, 0, 0, 0, 1❩, ❨1, 0, 0, 1, 0❩, ❨0, 0, 1, 0, 1❩, ❨0, 1, 0, 1, 0❩❩; SpectralRadius(A)', 2.0);
  Eps(1E-13); Test('A ≔ ❨❨4.88518185435250995, -0.0503410125523369264, 2.41500433970266516, -0.3526553784998275, 0.377477078463321959❩, '+'❨-0.0503410125523369266, 6.99880168536003428, 0.0574866278829369505, -0.00839458885495496514, 0.00898544320902995499❩, ❨2.41500433970266516, 0.0574866278829369505, 4.24219977363724693, 0.402712768116436254, -0.431057764708227601❩, '+'❨-0.3526553784998275, -0.00839458885495496536, 0.402712768116436254, 6.94119313935299163, 0.0629459900627696699❩, ❨0.377477078463321959, 0.00898544320902995499, -0.431057764708227601, 0.0629459900627696697, 6.93262354729721721❩❩; SpectralRadius(A)', 7.0);
  Eps(1E-13); Test('A ≔ ❨❨2, −1, 0, 3, 8, 2❩, ❨−1, 7, 5, 4, 8, 1❩, ❨0, 5, 2, 9, 3, 0❩, ❨3, 4, 9, −7, 2, 5❩, ❨8, 8, 3, 2, 1, 3❩, ❨2, 1, 0, 5, 3, 8❩❩; SpectralRadius(A)', 20.369654342915162);
  Eps(1E-13); Test('A ≔ ❨❨10, 16, 9, 8, 0, 7, 8❩, ❨16, 12, 5, 10, 4, 11, 14❩, ❨9, 5, 2, 12, 12, 5, 14❩, ❨8, 10, 12, 16, 12, 9, 3❩, ❨0, 4, 12, 12, 16, 17, 7❩, ❨7, 11, 5, 9, 17, 18, 7❩, ❨8, 14, 14, 3, 7, 7, 14❩❩; SpectralRadius(A)', 67.387186371351689);
  Eps(1E-13); Test('A ≔ ❨❨10, 5, 14, 10, 9, 13, 9, 15❩, ❨5, 4, 8, 7, 7, 4, 9, 13❩, ❨14, 8, 4, 6, 7, 5, 1, 14❩, ❨10, 7, 6, 18, 13, 12, 5, 11❩, ❨9, 7, 7, 13, 14, 12, 3, 14❩, ❨13, 4, 5, 12, 12, 18, 10, 15❩, ❨9, 9, 1, 5, 3, 10, 14, 14❩, '+'❨15, 13, 14, 11, 14, 15, 14, 6❩❩; SpectralRadius(A)', 79.906455303672033);
  Eps(1E-13); Test('A ≔ ❨❨10, 7, 11, 13, 15, 11, 9, 17, 7❩, ❨7, 8, 12, 3, 6, 10, 14, 7, 8❩, ❨11, 12, 18, 4, 11, 7, 9, 11, 2❩, ❨13, 3, 4, 0, 4, 9, 16, 11, 13❩, ❨15, 6, 11, 4, 4, 6, 11, 4, 11❩, ❨11, 10, 7, 9, 6, 8, 8, 10, 14❩, ❨9, 14, 9, 16, 11, 8, 10, 11, 13❩, '+'❨17, 7, 11, 11, 4, 10, 11, 14, 9❩, ❨7, 8, 2, 13, 11, 14, 13, 9, 6❩❩; SpectralRadius(A)', 86.264296094535837);
  Eps(1E-13); Test('A ≔ ❨❨6, 6, 8, 14, 12, 14, 5, 9, 12, 10❩, ❨6, 0, 4, 4, 10, 8, 15, 14, 13, 11❩, ❨8, 4, 10, 6, 3, 11, 11, 9, 9, 13❩, ❨14, 4, 6, 10, 9, 17, 16, 9, 11, 8❩, ❨12, 10, 3, 9, 10, 8, 12, 10, 9, 8❩, ❨14, 8, 11, 17, 8, 2, 4, 12, 12, 14❩, '+'❨5, 15, 11, 16, 12, 4, 0, 9, 14, 11❩, ❨9, 14, 9, 9, 10, 12, 9, 18, 9, 8❩, ❨12, 13, 9, 11, 9, 12, 14, 9, 2, 7❩, ❨10, 11, 13, 8, 8, 14, 11, 8, 7, 6❩❩; SpectralRadius(A)', 96.530449307303996);
  Eps(1E-13); Test('A ≔ ❨❨4.45528983757093495, -0.0878217640871465801, 0.587220808461187196, 0.381484594913282418, 0.806374962066634831, 0.415539821124804545, -1.07626065503415, -0.174476853005194865, 0.6117764065082528, 0.479713560928682166❩, '+'❨-0.0878217640871465801, 4.39333906061644079, -1.16146861964970503, 0.189604671732296926, 0.549866251807311702, 0.959584846946758171, -0.559867163033034075, 0.601342029002866998, 0.908395415505047237, -0.0481674127191538071❩, '+
    '❨0.587220808461187195, -1.16146861964970503, 4.731699905531466, -0.297368715494811859, 0.174137444046764151, 0.158862374235097095, 0.0708587702680495379, 0.527044617773810588, 0.717806236876411773, 0.100941787359586736❩, '+'❨0.381484594913282417, 0.189604671732296926, -0.297368715494811859, 5.20093245260149577, -0.422111106264731792, -0.609505049911372394, 0.430772405093700426, -0.075450459824149398, -0.0231177557010034982, 1.02271233462064678❩, '+
    '❨0.806374962066634831, 0.549866251807311702, 0.174137444046764151, -0.422111106264731792, 4.65483119820909985, 0.289795090828546352, 0.795841033040542465, -0.226199611183290607, -1.41597888676298831, 0.547038811317080901❩, '+'❨0.415539821124804545, 0.959584846946758171, 0.158862374235097095, -0.609505049911372394, 0.289795090828546352, 3.97252873681661176, 0.586614812369869111, -0.203645718668121608, 0.548222506108191402, -0.0260783651922211985❩, '+
    '❨-1.07626065503415, -0.559867163033034075, 0.0708587702680495379, 0.430772405093700426, 0.795841033040542465, 0.586614812369869111, 5.07303988296985066, 0.0852567827171805209, 0.721798919103536904, 0.0913845737686953809❩, '+'❨-0.174476853005194865, 0.601342029002866998, 0.527044617773810588, -0.0754504598241493979, -0.226199611183290607, -0.203645718668121608, 0.0852567827171805208, 5.72552879232507247, -0.381852810701084683, 0.248741195041378638❩, '+'❨0.6117764065082528, 0.908395415505047237, 0.717806236876411773, -0.0231177557010034982, -1.41597888676298831, 0.548222506108191402, 0.721798919103536904, -0.381852810701084683, 4.15028770968589867, 0.0862862168111860338❩, '+'❨0.479713560928682166, -0.0481674127191538071, 0.100941787359586736, 1.02271233462064678, 0.547038811317080901, -0.0260783651922211984, 0.0913845737686953809, 0.248741195041378638, 0.0862862168111860338, 3.64252242367312907❩❩; SpectralRadius(A)',
    6.0);
  Eps(1E-13); Test('A ≔ ReversalMatrix(10); SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨0.947217218464429805, -0.0797032009555874838, 0.455681246175844459, -0.890506530889571131, 0.0810219664601976321, -0.4241414032969027, -0.325611817072624359, 0.442131323976784809, 0.696001040652687542, -0.363371579820000821❩, '+'❨-0.0797032009555874838, 0.664262086761071853, 0.151183874385912782, 0.366175763890158299, -0.362499363441370523, 0.345609605325798349, -0.36153060182114391, -0.370383371480650959, -0.351151983839811129, -0.192995673028588803❩, '+
    '❨0.455681246175844459, 0.151183874385912782, 0.965112845520727124, -0.379442604818156945, -0.0868548299522641688, -0.703208127497219295, -0.29273075622571167, 0.392533036681992209, 0.433785161673175889, -0.0198556703071645884❩, '+'❨-0.890506530889571131, 0.366175763890158299, -0.379442604818156945, 1.60196556896465841, 0.0343051071296422081, 0.587093406299823787, -0.485735414772669726, -0.766652742695679709, -0.63672778172637717, -0.0824745994586258031❩, '+
    '❨0.0810219664601976321, -0.362499363441370523, -0.0868548299522641688, 0.0343051071296422081, 0.327534361557092576, -0.222948382709437083, -0.0500390113869091228, 0.161756725513035523, 0.21826384416565025, -0.135644537928391234❩, '+'❨-0.4241414032969027, 0.345609605325798349, -0.703208127497219295, 0.587093406299823787, -0.222948382709437083, 0.989037781112786131, -0.126211410841058615, -0.699854415829559306, -0.413031006405435901, 0.131956448780506888❩, '+
    '❨-0.325611817072624359, -0.36153060182114391, -0.29273075622571167, -0.485735414772669726, -0.0500390113869091228, -0.126211410841058615, 1.00217393220066957, 0.281471732316985816, -0.335032854995363073, 0.398331000774434439❩, '+'❨0.442131323976784809, -0.370383371480650959, 0.392533036681992209, -0.766652742695679709, 0.161756725513035523, -0.699854415829559306, 0.281471732316985816, 0.61286300125401041, 0.371247429623465179, -0.0611706531297214894❩, '+
    '❨0.696001040652687542, -0.351151983839811129, 0.433785161673175889, -0.63672778172637717, 0.21826384416565025, -0.413031006405435901, -0.335032854995363073, 0.371247429623465179, 0.960656361843560564, 0.149752326353616183❩, '+'❨-0.363371579820000821, -0.192995673028588803, -0.0198556703071645884, -0.0824745994586258031, -0.135644537928391234, 0.131956448780506888, 0.398331000774434439, -0.0611706531297214894, 0.149752326353616183, 0.929176842320993553❩❩; SpectralRadius(A)',
    4.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 2❩, ❨3, 4❩❩; SpectralRadius(A)', 5.372281323269014);
  Eps(1E-13); Test('A ≔ ❨❨5, 7, 4❩, ❨0, 2, 6❩, ❨1, 5, 3❩❩; SpectralRadius(A)', 9.334366656868244);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 2❩, ❨1, 0, -1❩, ❨1, 1, 2❩❩; SpectralRadius(A)', 3.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 1, 0❩, ❨1, 0, -1❩, ❨1, 2, 1❩❩; SpectralRadius(A)', Sqrt(2));
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 0, 1❩❩; SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 0, 2❩❩; SpectralRadius(A)', 2.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 1, 2❩❩; SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨5, 4, 7, 2❩, ❨6, 3, 5, 7❩, ❨0, 5, 7, 8❩, ❨3, 4, 1, 9❩❩; SpectralRadius(A)', 18.722583143817452);
  Eps(1E-13); Test('A ≔ ❨❨0, 0, 1, 0❩, ❨0, 0, 0, 0❩, ❨1, 0, 0, 0❩, ❨0, 0, 0, 2❩❩; SpectralRadius(A)', 2.0);
  Eps(1E-13); Test('A ≔ ❨❨5, 8, 4, 0, 2❩, ❨6, 3, 0, 4, 8❩, ❨6, 5, 4, 9, 2❩, ❨5, 6, 3, 8, 7❩, ❨1, 0, 2, 9, 7❩❩; SpectralRadius(A)', 22.814428493209221);
  Eps(1E-13); Test('A ≔ ❨❨5, 4, 7, 0, 2, 5❩, ❨6, 5, 8, 9, 5, 1❩, ❨0, 5, 8, 2, 3, 6❩, ❨4, 5, 8, 4, 6, 2❩, ❨7, 5, 9, 6, 1, 0❩, ❨5, 4, 7, 5, 2, 6❩❩; SpectralRadius(A)', 27.475808456817589);
  Eps(1E-13); Test('A ≔ ❨❨4, 2, 5, 8, 9, 6, 3❩, ❨5, 8, 4, 5, 8, 6, 3❩, ❨0, 5, 4, 7, 8, 9, 6❩, ❨4, 1, 2, 5, 0, 8, 6❩, ❨6, 6, 5, 0, 7, 4, 8❩, ❨4, 7, 5, 1, 9, 8, 6❩, ❨5, 3, 6, 8, 4, 1, 2❩❩; SpectralRadius(A)', 35.314786301350608);
  Eps(1E-13); Test('A ≔ ❨❨5, 8, 6, 5, 4, 7, 1, 2❩, ❨6, 3, 2, 5, 6, 9, 8, 5❩, ❨4, 5, 6, 9, 2, 0, 4, 7❩, ❨3, 6, 0, 8, 2, 4, 5, 9❩, ❨6, 3, 5, 4, 7, 8, 5, 2❩, ❨5, 3, 2, 1, 9, 8, 5, 6❩, ❨5, 4, 6, 8, 2, 1, 0, 7❩, ❨6, 3, 9, 5, 8, 3, 4, 7❩❩; SpectralRadius(A)', 39.293880621805755);
  Eps(1E-13); Test('A ≔ ❨❨5, 7, 6, 9, 5, 6, 3, 2, 5❩, ❨5, 4, 0, 2, 5, 8, 9, 6, 5❩, ❨7, 4, 5, 2, 1, 5, 6, 9, 8❩, ❨3, 6, 5, 4, 7, 8, 5, 2, 0❩, ❨2, 0, 8, 4, 6, 0, 8, 7, 3❩, ❨0, 1, 6, 8, 5, 3, 7, 9, 4❩, ❨5, 6, 3, 8, 7, 4, 1, 0, 5❩, '+'❨6, 5, 3, 8, 9, 4, 5, 2, 1❩, ❨4, 6, 3, 5, 5, 8, 9, 3, 2❩❩; SpectralRadius(A)', 42.560831728333504);
  Eps(1E-13); Test('A ≔ ❨❨5, 4, 1, 5, 2, 9, 8, 3, 0, 5❩, ❨5, 6, 3, 2, 0, 8, 5, 6, 9, 7❩, ❨4, 5, 6, 8, 2, 3, 6, 5, 7, 0❩, ❨4, 5, 9, 6, 3, 6, 5, 6, 8, 9❩, ❨4, 5, 5, 2, 6, 8, 7, 9, 2, 5❩, ❨5, 6, 3, 2, 0, 5, 4, 7, 8, 5❩, ❨4, 5, 4, 5, 6, 0, 2, 3, 7, 9❩, '+'❨5, 6, 3, 2, 5, 4, 7, 8, 4, 1❩, ❨5, 6, 3, 2, 0, 4, 5, 1, 2, 7❩, ❨6, 5, 6, 4, 1, 2, 7, 8, 4, 0❩❩; SpectralRadius(A)', 45.848201925978358);

  Eps(1E-13); Test('A ≔ ❨❨21❩❩; SpectralRadius(ComplexMatrix(A))', 21.0);
  Eps(1E-13); Test('A ≔ ❨❨3, 5❩, ❨5, 2❩❩; SpectralRadius(ComplexMatrix(A))', 7.524937810560445);
  Eps(1E-13); Test('A ≔ ❨❨4, 1, 8❩, ❨1, 2, 5❩, ❨8, 5, 3❩❩; SpectralRadius(ComplexMatrix(A))', 13.136085649145485);
  Eps(1E-13); Test('A ≔ ❨❨2, 0, 1❩, ❨0, 2, 0❩, ❨1, 0, 2❩❩; SpectralRadius(ComplexMatrix(A))', 3.0);
  Eps(1E-13); Test('A ≔ ❨❨0, 0, 2❩, ❨0, 2, 0❩, ❨2, 0, 0❩❩; SpectralRadius(ComplexMatrix(A))', 2.0);
  Eps(1E-13); Test('A ≔ ❨❨6, 1, 3, 2❩, ❨1, 8, 5, 1❩, ❨3, 5, 2, 0❩, ❨2, 1, 0, 7❩❩; SpectralRadius(ComplexMatrix(A))', 12.317576522344481);
  Eps(1E-13); Test('A ≔ ❨❨0, 1, 0, 2❩, ❨1, -2, 2, 0❩, ❨0, 2, 0, -1❩, ❨2, 0, -1, 2❩❩; SpectralRadius(ComplexMatrix(A))', 3.449489742783178);
  Eps(1E-13); Test('A ≔ ❨❨0, 1, 0, 2❩, ❨1, -2, 2, 0❩, ❨0, 2, 0, -1❩, ❨2, 0, -1, -2❩❩; SpectralRadius(ComplexMatrix(A))', 3.449489742783178);
  Eps(1E-13); Test('A ≔ ❨❨0, 0, 1, 2❩, ❨0, -2, 0, 0❩, ❨1, 0, 0, 1❩, ❨2, 0, 1, 0❩❩; SpectralRadius(ComplexMatrix(A))', 2.732050807568878);
  Eps(1E-13); Test('A ≔ ❨❨3.09090210854135331, -0.0989480787770248203, 0.206696174292443883, -0.173565216686099517❩, '+'❨-0.0989480787770248203, 3.10770621771892455, -0.224991363401589891, 0.188927902874714558❩, ❨0.206696174292443883, -0.224991363401589891, 3.4699924914029535, -0.394658241210492851❩, '+'❨-0.173565216686099517, 0.188927902874714558, -0.394658241210492851, 3.33139918233676864❩❩; SpectralRadius(ComplexMatrix(A))', 4.0);
  Eps(1E-13); Test('A ≔ ❨❨3, 2, 1, 0, 1❩, ❨2, 5, 7, 3, 8❩, ❨1, 7, 6, 0, 2❩, ❨0, 3, 0, 5, 3❩, ❨1, 8, 2, 3, 9❩❩; SpectralRadius(ComplexMatrix(A))', 19.613776982166058);
  Eps(1E-13); Test('A ≔ ❨❨2, 0, 1, 0, 0❩, ❨0, 0, 0, 2, 0❩, ❨1, 0, 0, 1, 1❩, ❨0, 2, 1, 0, 0❩, ❨0, 0, 1, 0, 2❩❩; SpectralRadius(ComplexMatrix(A))', 2.895106515927531);
  Eps(1E-13); Test('A ≔ ❨❨0, 1, 1, 0, 0❩, ❨1, 0, 0, 0, 1❩, ❨1, 0, 0, 1, 0❩, ❨0, 0, 1, 0, 1❩, ❨0, 1, 0, 1, 0❩❩; SpectralRadius(ComplexMatrix(A))', 2.0);
  Eps(1E-13); Test('A ≔ ❨❨4.88518185435250995, -0.0503410125523369264, 2.41500433970266516, -0.3526553784998275, 0.377477078463321959❩, '+'❨-0.0503410125523369266, 6.99880168536003428, 0.0574866278829369505, -0.00839458885495496514, 0.00898544320902995499❩, '+'❨2.41500433970266516, 0.0574866278829369505, 4.24219977363724693, 0.402712768116436254, -0.431057764708227601❩, '+'❨-0.3526553784998275, -0.00839458885495496536, 0.402712768116436254, 6.94119313935299163, 0.0629459900627696699❩, '+'❨0.377477078463321959, 0.00898544320902995499, -0.431057764708227601, 0.0629459900627696697, 6.93262354729721721❩❩; SpectralRadius(ComplexMatrix(A))', 7.0);
  Eps(1E-13); Test('A ≔ ❨❨2, −1, 0, 3, 8, 2❩, ❨−1, 7, 5, 4, 8, 1❩, ❨0, 5, 2, 9, 3, 0❩, ❨3, 4, 9, −7, 2, 5❩, ❨8, 8, 3, 2, 1, 3❩, ❨2, 1, 0, 5, 3, 8❩❩; SpectralRadius(ComplexMatrix(A))', 20.369654342915162);
  Eps(1E-13); Test('A ≔ ❨❨10, 16, 9, 8, 0, 7, 8❩, ❨16, 12, 5, 10, 4, 11, 14❩, ❨9, 5, 2, 12, 12, 5, 14❩, ❨8, 10, 12, 16, 12, 9, 3❩, ❨0, 4, 12, 12, 16, 17, 7❩, ❨7, 11, 5, 9, 17, 18, 7❩, ❨8, 14, 14, 3, 7, 7, 14❩❩; SpectralRadius(ComplexMatrix(A))', 67.387186371351689);
  Eps(1E-13); Test('A ≔ ❨❨10, 5, 14, 10, 9, 13, 9, 15❩, ❨5, 4, 8, 7, 7, 4, 9, 13❩, ❨14, 8, 4, 6, 7, 5, 1, 14❩, ❨10, 7, 6, 18, 13, 12, 5, 11❩, ❨9, 7, 7, 13, 14, 12, 3, 14❩, ❨13, 4, 5, 12, 12, 18, 10, 15❩, ❨9, 9, 1, 5, 3, 10, 14, 14❩, '+'❨15, 13, 14, 11, 14, 15, 14, 6❩❩; SpectralRadius(ComplexMatrix(A))', 79.906455303672033);
  Eps(1E-13); Test('A ≔ ❨❨10, 7, 11, 13, 15, 11, 9, 17, 7❩, ❨7, 8, 12, 3, 6, 10, 14, 7, 8❩, ❨11, 12, 18, 4, 11, 7, 9, 11, 2❩, ❨13, 3, 4, 0, 4, 9, 16, 11, 13❩, ❨15, 6, 11, 4, 4, 6, 11, 4, 11❩, ❨11, 10, 7, 9, 6, 8, 8, 10, 14❩, ❨9, 14, 9, 16, 11, 8, 10, 11, 13❩, '+'❨17, 7, 11, 11, 4, 10, 11, 14, 9❩, ❨7, 8, 2, 13, 11, 14, 13, 9, 6❩❩; SpectralRadius(ComplexMatrix(A))', 86.264296094535837);
  Eps(1E-13); Test('A ≔ ❨❨6, 6, 8, 14, 12, 14, 5, 9, 12, 10❩, ❨6, 0, 4, 4, 10, 8, 15, 14, 13, 11❩, ❨8, 4, 10, 6, 3, 11, 11, 9, 9, 13❩, ❨14, 4, 6, 10, 9, 17, 16, 9, 11, 8❩, ❨12, 10, 3, 9, 10, 8, 12, 10, 9, 8❩, ❨14, 8, 11, 17, 8, 2, 4, 12, 12, 14❩, '+'❨5, 15, 11, 16, 12, 4, 0, 9, 14, 11❩, ❨9, 14, 9, 9, 10, 12, 9, 18, 9, 8❩, ❨12, 13, 9, 11, 9, 12, 14, 9, 2, 7❩, ❨10, 11, 13, 8, 8, 14, 11, 8, 7, 6❩❩; SpectralRadius(ComplexMatrix(A))', 96.530449307303996);
  Eps(1E-13); Test('A ≔ ❨❨4.45528983757093495, -0.0878217640871465801, 0.587220808461187196, 0.381484594913282418, 0.806374962066634831, 0.415539821124804545, -1.07626065503415, -0.174476853005194865, 0.6117764065082528, 0.479713560928682166❩, '+'❨-0.0878217640871465801, 4.39333906061644079, -1.16146861964970503, 0.189604671732296926, 0.549866251807311702, 0.959584846946758171, -0.559867163033034075, 0.601342029002866998, 0.908395415505047237, -0.0481674127191538071❩, '+
    '❨0.587220808461187195, -1.16146861964970503, 4.731699905531466, -0.297368715494811859, 0.174137444046764151, 0.158862374235097095, 0.0708587702680495379, 0.527044617773810588, 0.717806236876411773, 0.100941787359586736❩, '+'❨0.381484594913282417, 0.189604671732296926, -0.297368715494811859, 5.20093245260149577, -0.422111106264731792, -0.609505049911372394, 0.430772405093700426, -0.075450459824149398, -0.0231177557010034982, 1.02271233462064678❩, '+
    '❨0.806374962066634831, 0.549866251807311702, 0.174137444046764151, -0.422111106264731792, 4.65483119820909985, 0.289795090828546352, 0.795841033040542465, -0.226199611183290607, -1.41597888676298831, 0.547038811317080901❩, '+'❨0.415539821124804545, 0.959584846946758171, 0.158862374235097095, -0.609505049911372394, 0.289795090828546352, 3.97252873681661176, 0.586614812369869111, -0.203645718668121608, 0.548222506108191402, -0.0260783651922211985❩, '+
    '❨-1.07626065503415, -0.559867163033034075, 0.0708587702680495379, 0.430772405093700426, 0.795841033040542465, 0.586614812369869111, 5.07303988296985066, 0.0852567827171805209, 0.721798919103536904, 0.0913845737686953809❩, '+'❨-0.174476853005194865, 0.601342029002866998, 0.527044617773810588, -0.0754504598241493979, -0.226199611183290607, -0.203645718668121608, 0.0852567827171805208, 5.72552879232507247, -0.381852810701084683, 0.248741195041378638❩, '+'❨0.6117764065082528, 0.908395415505047237, 0.717806236876411773, -0.0231177557010034982, -1.41597888676298831, 0.548222506108191402, 0.721798919103536904, -0.381852810701084683, 4.15028770968589867, 0.0862862168111860338❩, '+'❨0.479713560928682166, -0.0481674127191538071, 0.100941787359586736, 1.02271233462064678, 0.547038811317080901, -0.0260783651922211984, 0.0913845737686953809, 0.248741195041378638, 0.0862862168111860338, 3.64252242367312907❩❩; '+'SpectralRadius(ComplexMatrix(A))',
    6.0);
  Eps(1E-13); Test('A ≔ ReversalMatrix(10); SpectralRadius(ComplexMatrix(A))', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨0.947217218464429805, -0.0797032009555874838, 0.455681246175844459, -0.890506530889571131, 0.0810219664601976321, -0.4241414032969027, -0.325611817072624359, 0.442131323976784809, 0.696001040652687542, -0.363371579820000821❩, '+'❨-0.0797032009555874838, 0.664262086761071853, 0.151183874385912782, 0.366175763890158299, -0.362499363441370523, 0.345609605325798349, -0.36153060182114391, -0.370383371480650959, -0.351151983839811129, -0.192995673028588803❩, '+
    '❨0.455681246175844459, 0.151183874385912782, 0.965112845520727124, -0.379442604818156945, -0.0868548299522641688, -0.703208127497219295, -0.29273075622571167, 0.392533036681992209, 0.433785161673175889, -0.0198556703071645884❩, '+'❨-0.890506530889571131, 0.366175763890158299, -0.379442604818156945, 1.60196556896465841, 0.0343051071296422081, 0.587093406299823787, -0.485735414772669726, -0.766652742695679709, -0.63672778172637717, -0.0824745994586258031❩, '+
    '❨0.0810219664601976321, -0.362499363441370523, -0.0868548299522641688, 0.0343051071296422081, 0.327534361557092576, -0.222948382709437083, -0.0500390113869091228, 0.161756725513035523, 0.21826384416565025, -0.135644537928391234❩, '+'❨-0.4241414032969027, 0.345609605325798349, -0.703208127497219295, 0.587093406299823787, -0.222948382709437083, 0.989037781112786131, -0.126211410841058615, -0.699854415829559306, -0.413031006405435901, 0.131956448780506888❩, '+
    '❨-0.325611817072624359, -0.36153060182114391, -0.29273075622571167, -0.485735414772669726, -0.0500390113869091228, -0.126211410841058615, 1.00217393220066957, 0.281471732316985816, -0.335032854995363073, 0.398331000774434439❩, '+'❨0.442131323976784809, -0.370383371480650959, 0.392533036681992209, -0.766652742695679709, 0.161756725513035523, -0.699854415829559306, 0.281471732316985816, 0.61286300125401041, 0.371247429623465179, -0.0611706531297214894❩, '+
    '❨0.696001040652687542, -0.351151983839811129, 0.433785161673175889, -0.63672778172637717, 0.21826384416565025, -0.413031006405435901, -0.335032854995363073, 0.371247429623465179, 0.960656361843560564, 0.149752326353616183❩, '+'❨-0.363371579820000821, -0.192995673028588803, -0.0198556703071645884, -0.0824745994586258031, -0.135644537928391234, 0.131956448780506888, 0.398331000774434439, -0.0611706531297214894, 0.149752326353616183, 0.929176842320993553❩❩; '+'SpectralRadius(ComplexMatrix(A))',
    4.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 2❩, ❨3, 4❩❩; SpectralRadius(ComplexMatrix(A))', 5.372281323269014);
  Eps(1E-13); Test('A ≔ ❨❨5, 7, 4❩, ❨0, 2, 6❩, ❨1, 5, 3❩❩; SpectralRadius(ComplexMatrix(A))', 9.334366656868244);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 2❩, ❨1, 0, -1❩, ❨1, 1, 2❩❩; SpectralRadius(ComplexMatrix(A))', 3.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 1, 0❩, ❨1, 0, -1❩, ❨1, 2, 1❩❩; SpectralRadius(ComplexMatrix(A))', Sqrt(2));
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 0, 1❩❩; SpectralRadius(ComplexMatrix(A))', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 0, 2❩❩; SpectralRadius(ComplexMatrix(A))', 2.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0❩, ❨1, 0, -1❩, ❨1, 1, 2❩❩; SpectralRadius(ComplexMatrix(A))', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨5, 4, 7, 2❩, ❨6, 3, 5, 7❩, ❨0, 5, 7, 8❩, ❨3, 4, 1, 9❩❩; SpectralRadius(ComplexMatrix(A))', 18.722583143817452);
  Eps(1E-13); Test('A ≔ ❨❨0, 0, 1, 0❩, ❨0, 0, 0, 0❩, ❨1, 0, 0, 0❩, ❨0, 0, 0, 2❩❩; SpectralRadius(ComplexMatrix(A))', 2.0);
  Eps(1E-13); Test('A ≔ ❨❨5, 8, 4, 0, 2❩, ❨6, 3, 0, 4, 8❩, ❨6, 5, 4, 9, 2❩, ❨5, 6, 3, 8, 7❩, ❨1, 0, 2, 9, 7❩❩; SpectralRadius(ComplexMatrix(A))', 22.814428493209221);
  Eps(1E-13); Test('A ≔ ❨❨5, 4, 7, 0, 2, 5❩, ❨6, 5, 8, 9, 5, 1❩, ❨0, 5, 8, 2, 3, 6❩, ❨4, 5, 8, 4, 6, 2❩, ❨7, 5, 9, 6, 1, 0❩, ❨5, 4, 7, 5, 2, 6❩❩; SpectralRadius(ComplexMatrix(A))', 27.475808456817589);
  Eps(1E-13); Test('A ≔ ❨❨4, 2, 5, 8, 9, 6, 3❩, ❨5, 8, 4, 5, 8, 6, 3❩, ❨0, 5, 4, 7, 8, 9, 6❩, ❨4, 1, 2, 5, 0, 8, 6❩, ❨6, 6, 5, 0, 7, 4, 8❩, ❨4, 7, 5, 1, 9, 8, 6❩, ❨5, 3, 6, 8, 4, 1, 2❩❩; SpectralRadius(ComplexMatrix(A))', 35.314786301350608);
  Eps(1E-13); Test('A ≔ ❨❨5, 8, 6, 5, 4, 7, 1, 2❩, ❨6, 3, 2, 5, 6, 9, 8, 5❩, ❨4, 5, 6, 9, 2, 0, 4, 7❩, ❨3, 6, 0, 8, 2, 4, 5, 9❩, ❨6, 3, 5, 4, 7, 8, 5, 2❩, ❨5, 3, 2, 1, 9, 8, 5, 6❩, ❨5, 4, 6, 8, 2, 1, 0, 7❩, ❨6, 3, 9, 5, 8, 3, 4, 7❩❩; SpectralRadius(ComplexMatrix(A))', 39.293880621805755);
  Eps(1E-13); Test('A ≔ ❨❨5, 7, 6, 9, 5, 6, 3, 2, 5❩, ❨5, 4, 0, 2, 5, 8, 9, 6, 5❩, ❨7, 4, 5, 2, 1, 5, 6, 9, 8❩, ❨3, 6, 5, 4, 7, 8, 5, 2, 0❩, ❨2, 0, 8, 4, 6, 0, 8, 7, 3❩, ❨0, 1, 6, 8, 5, 3, 7, 9, 4❩, ❨5, 6, 3, 8, 7, 4, 1, 0, 5❩, '+'❨6, 5, 3, 8, 9, 4, 5, 2, 1❩, ❨4, 6, 3, 5, 5, 8, 9, 3, 2❩❩; SpectralRadius(ComplexMatrix(A))', 42.560831728333504);
  Eps(1E-13); Test('A ≔ ❨❨5, 4, 1, 5, 2, 9, 8, 3, 0, 5❩, ❨5, 6, 3, 2, 0, 8, 5, 6, 9, 7❩, ❨4, 5, 6, 8, 2, 3, 6, 5, 7, 0❩, ❨4, 5, 9, 6, 3, 6, 5, 6, 8, 9❩, ❨4, 5, 5, 2, 6, 8, 7, 9, 2, 5❩, ❨5, 6, 3, 2, 0, 5, 4, 7, 8, 5❩, ❨4, 5, 4, 5, 6, 0, 2, 3, 7, 9❩, '+'❨5, 6, 3, 2, 5, 4, 7, 8, 4, 1❩, ❨5, 6, 3, 2, 0, 4, 5, 1, 2, 7❩, ❨6, 5, 6, 4, 1, 2, 7, 8, 4, 0❩❩; SpectralRadius(ComplexMatrix(A))', 45.848201925978358);


  Eps(1E-13); Test('A ≔ ❨❨1, i❩, ❨-i, 2❩❩; SpectralRadius(A)', 2.618033988749895);
  Eps(1E-13); Test('A ≔ ❨❨2, i, 3❩, ❨-i, 5, 1❩, ❨3, 1, 0❩❩; SpectralRadius(A)', 5.754405707945917);
  Eps(1E-13); Test('A ≔ ❨❨3, 2, i, 5❩, ❨2, 7, 0, 1❩, ❨−i, 0, 2, −i❩, ❨5, 1, i, 3❩❩; SpectralRadius(A)', 9.854644953717658);
  Eps(1E-13); Test('A ≔ ❨❨5, i, 1, 2, 3❩, ❨−i, 4, 8, 3, 5❩, ❨1, 8, 1, 2, −i❩, ❨2, 3, 2, 5, 9❩, ❨3, 5, i, 9, 3❩❩; SpectralRadius(A)', 18.188313753340950);
  Eps(1E-13); Test('A ≔ ❨❨3, 8, 5, 1, 2, 0❩, ❨8, 8, 0, i, 7, 2❩, ❨5, 0, 5, 0, 1, 8❩, ❨1, −i, 0, 3, i, 2❩, ❨2, 7, 1, −i, 7, 9❩, ❨0, 2, 8, 2, 9, 2❩❩; SpectralRadius(A)', 22.508205131012872);
  Eps(1E-13); Test('A ≔ ❨❨2, 8, i, 0, 9, 3, 1❩, ❨8, 5, 2, 7, −i, 3, 0❩, ❨−i, 2, 0, 4, 3, 9, 1❩, ❨0, 7, 4, 6, 2, 1, i❩, ❨9, i, 3, 2, 8, 0, 5❩, ❨3, 3, 9, 1, 0, 3, 2❩, ❨1, 0, 1, −i, 5, 2, 4❩❩; SpectralRadius(A)', 22.174399966828524);
  Eps(1E-13); Test('A ≔ ❨❨2, 8, 5, 3, i, 2, 7, 1❩, ❨8, 2, 7, 8, 0, 3, 4, 6❩, ❨5, 7, 1, 2, i, 5, 7, 1❩, ❨3, 8, 2, 6, 3, 2, −i, i❩, ❨−i, 0, −i, 3, 5, 7, 0, 8❩, ❨2, 3, 5, 2, 7, 3, 2, i❩, ❨7, 4, 7, i, 0, 2, 5, 7❩, ❨1, 6, 1, −i, 8, −i, 7, 8❩❩; SpectralRadius(A)', 29.502669186726628);
  Eps(1E-13); Test('A ≔ ❨❨3, 2, 8, 5, 0, 7, 9, 6, 5❩, ❨2, 1, 0, i, 3, 8, 0, 2, 2❩, ❨8, 0, 5, 7, 9, 3, 0, 8, 6❩, ❨5, −i, 7, 1, 2, 4, 9, 3, 5❩, ❨0, 3, 9, 2, 6, −i, i, 2, 8❩, ❨7, 8, 3, 4, i, 5, 4, 9, 8❩, ❨9, 0, 0, 9, −i, 4, 3, 6, 7❩, '+'❨6, 2, 8, 3, 2, 9, 6, 2, 8❩, ❨5, 2, 6, 5, 8, 8, 7, 8, 5❩❩; SpectralRadius(A)', 42.725142660244707);
  Eps(1E-13); Test('A ≔ ❨❨1, 2, 3, 4, 5, 6, 7, 8, 9, i❩, ❨2, 3, 4, 5, 6, 7, 8, 9, i, 1❩, ❨3, 4, 5, 6, 7, 8, 9, i, 1, 2❩, ❨4, 5, 6, 7, 8, 9, i, 1, 2, 3❩, ❨5, 6, 7, 8, 9, i, 1, 2, 3, 4❩, ❨6, 7, 8, 9, −i, 1, 2, 3, 4, 5❩, '+'❨7, 8, 9, −i, 1, 2, 3, 4, 5, 6❩, ❨8, 9, −i, 1, 2, 3, 4, 5, 6, 7❩, ❨9, −i, 1, 2, 3, 4, 5, 6, 7, 8❩, ❨−i, 1, 2, 3, 4, 5, 6, 7, 8, 9❩❩; SpectralRadius(A)', 45.027811551304595);
  Eps(1E-13); Test('A ≔ ❨❨i❩❩; SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨2, i❩, ❨4, 1❩❩; SpectralRadius(A)', 3.26113976166991682);
  Eps(1E-13); Test('A ≔ ❨❨9, 3, i❩, ❨5, 0, −i❩, ❨3, 4, 1❩❩; SpectralRadius(A)', 10.4649386897631176);
  Eps(1E-13); Test('A ≔ ❨❨1, 1, i❩, ❨0, i, 0❩, ❨1, 0, i❩❩; SpectralRadius(A)', Sqrt(2));
  Eps(1E-13); Test('A ≔ ❨❨0, 1, i❩, ❨0, i, 1❩, ❨1, i, 0❩❩; SpectralRadius(A)', Sqrt(2));
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0❩, ❨i, 1, i❩, ❨1, i, 1❩❩; SpectralRadius(A)', Sqrt(2));
  Eps(1E-13); Test('A ≔ ❨❨-1, 0, -1❩, ❨i, 1, 0❩, ❨0, 0, 1❩❩; SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 1❩, ❨i, 1, 0❩, ❨0, 0, 1❩❩; SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0, 0❩, ❨0, i, i, 0❩, ❨1, 0, 0, 1❩, ❨0, 0, 0, i❩❩; SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0, 0❩, ❨0, 1, i, 0❩, ❨1, 0, 0, 1❩, ❨0, 0, 0, 1❩❩; SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨1, 0, 0, 0❩, ❨0, 1, i, 0❩, ❨1, 1, 0, 1❩, ❨0, 0, 0, 1❩❩; SpectralRadius(A)', 1.44257374044605948);
  Eps(1E-13); Test('A ≔ ❨❨6, 2, 1, 4❩, ❨i, 0, 3, 1❩, ❨2, 8, 7, i❩, ❨3, 7, 5, i❩❩; SpectralRadius(A)', 11.6509822717031644);
  Eps(1E-13); Test('A ≔ ❨❨1, 5, 2, 8, 6❩, ❨4, 3, 2, 0, i❩, ❨2, 5, 3, 9, 4❩, ❨0, 7, 5, 2, i❩, ❨6, 5, 0, 8, 7❩❩; SpectralRadius(A)', 17.0309379463060299);
  Eps(1E-13); Test('A ≔ ❨❨5, i, 3, 0, 5, 1❩, ❨7, 6, 3, 2, 6, 9❩, ❨4, 2, i, 4, 5, 0❩, ❨2, 3, 8, 5, 6, 9❩, ❨4, 2, 8, i, 6, 3❩, ❨4, 8, 3, 2, 9, 7❩❩; SpectralRadius(A)', 23.7419840643665259);
  Eps(1E-13); Test('A ≔ ❨❨5, 2, 3, 6, 9, 7, 5❩, ❨0, 9, 6, 3, 5, 7, 6❩, ❨5, 4, i, 4, 2, 9, 6❩, ❨7, 6, 3, 0, 8, 4, i❩, ❨9, 6, 3, 2, 7, 1, 4❩, ❨2, 5, 3, 6, 0, 1, 1❩, ❨8, 6, 3, i, 4, 5, 6❩❩; SpectralRadius(A)', 30.7409773649641304);
  Eps(1E-13); Test('A ≔ ❨❨5, i, 3, 6, 2, i, 4, 8❩, ❨3, 6, 2, 5, 8, 0, 4, 7❩, ❨4, 5, 3, 8, i, 0, 2, 4❩, ❨7, 5, 6, 3, 0, 2, 5, 2❩, ❨5, 8, 9, 0, 2, 1, 5, 3❩, ❨6, 3, 2, 5, 7, 4, 0, 8❩, ❨9, 3, 5, 8, 0, 2, 5, 4❩, ❨6, i, 3, 0, 2, 5, 6, 7❩❩; SpectralRadius(A)', 30.7822431399464038);
  Eps(1E-13); Test('A ≔ ❨❨3, 2, 5, 6, 8, 0, 2, 5, 6❩, ❨8, 5, 0, 3, 6, 5, 1, 4, 7❩, ❨4, 3, 6, 5, 2, 8, 9, 6, 0❩, ❨4, 5, 6, i, 8, 9, 6, 6, 3❩, ❨0, 2, 5, 1, 3, 5, 4, 9, 7❩, ❨6, 3, 2, 5, 6, 4, 1, 7, 0❩, ❨8, 2, 5, 0, 1, 9, 6, 4, 7❩, '+'❨6, i, 3, 6, 4, 7, 8, 0, 2❩, ❨3, 5, 0, 4, 9, 2, 1, i, 8❩❩; SpectralRadius(A)', 37.9098145316765089);
  Eps(1E-13); Test('A ≔ ❨❨1, 2, 3, 4, 5, 6, 7, 8, 9, i❩, ❨2, 3, 4, 5, 6, 7, 8, 9, i, 1❩, ❨3, 4, 5, 6, 7, 8, 9, i, 1, 2❩, ❨4, 5, 6, 7, 8, 9, i, 1, 2, 3❩, ❨5, 6, 7, 8, 9, i, 1, 2, 3, 4❩, ❨6, 7, 8, 9, i, 1, 2, 3, 4, 5❩, '+'❨7, 8, 9, i, 1, 2, 3, 4, 5, 6❩, ❨8, 9, i, 1, 2, 3, 4, 5, 6, 7❩, ❨9, i, 1, 2, 3, 4, 5, 6, 7, 8❩, ❨i, 1, 2, 3, 4, 5, 6, 7, 8, 9❩❩; SpectralRadius(A)', 45.0111097397075959);
  Eps(1E-13); Test('A ≔ ❨❨2, 8, 0, 5❩, ❨8, 3, 9, 1❩, ❨0, 9, 7, 0❩, ❨5, 1, 0, 1❩❩; SpectralRadius(A)', 16.697147626100790);
  Eps(1E-13); Test('A ≔ ❨❨4.88518185435250995, -0.0503410125523369264, 2.41500433970266516, -0.3526553784998275, 0.377477078463321959❩, '+'❨-0.0503410125523369266, 6.99880168536003428, 0.0574866278829369505, -0.00839458885495496514, 0.00898544320902995499❩, ❨2.41500433970266516, 0.0574866278829369505, 4.24219977363724693, 0.402712768116436254, -0.431057764708227601❩, '+'❨-0.3526553784998275, -0.00839458885495496536, 0.402712768116436254, 6.94119313935299163, 0.0629459900627696699❩, ❨0.377477078463321959, 0.00898544320902995499, -0.431057764708227601, 0.0629459900627696697, 6.93262354729721721❩❩; SpectralRadius(A)', 7.0);
  Eps(1E-13); Test('A ≔ ❨❨4, 2, 8, 1, 7, 5❩, ❨2, 3, 9, 4, 1, 2❩, ❨8, 9, 7, 5, 2, 1❩, ❨1, 4, 5, 8, 0, 9❩, ❨7, 1, 2, 0, 3, 2❩, ❨5, 2, 1, 9, 2, 6❩❩; SpectralRadius(A)', 25.699931912397357);
  Eps(1E-13); Test('A ≔ ❨❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩, ❨2, 3, 4, 5, 6, 7, 8, 9, 10, 1❩, ❨3, 4, 5, 6, 7, 8, 9, 10, 1, 2❩, ❨4, 5, 6, 7, 8, 9, 10, 1, 2, 3❩, ❨5, 6, 7, 8, 9, 10, 1, 2, 3, 4❩, '+'❨6, 7, 8, 9, 10, 1, 2, 3, 4, 5❩, ❨7, 8, 9, 10, 1, 2, 3, 4, 5, 6❩, ❨8, 9, 10, 1, 2, 3, 4, 5, 6, 7❩, ❨9, 10, 1, 2, 3, 4, 5, 6, 7, 8❩, ❨10, 1, 2, 3, 4, 5, 6, 7, 8, 9❩❩; SpectralRadius(A)', 55.0);
  Eps(1E-13); Test('A ≔ ❨❨4.45528983757093495, -0.0878217640871465801, 0.587220808461187196, 0.381484594913282418, 0.806374962066634831, 0.415539821124804545, -1.07626065503415, -0.174476853005194865, 0.6117764065082528, 0.479713560928682166❩, '+'❨-0.0878217640871465801, 4.39333906061644079, -1.16146861964970503, 0.189604671732296926, 0.549866251807311702, 0.959584846946758171, -0.559867163033034075, 0.601342029002866998, 0.908395415505047237, -0.0481674127191538071❩, '+
    '❨0.587220808461187195, -1.16146861964970503, 4.731699905531466, -0.297368715494811859, 0.174137444046764151, 0.158862374235097095, 0.0708587702680495379, 0.527044617773810588, 0.717806236876411773, 0.100941787359586736❩, '+'❨0.381484594913282417, 0.189604671732296926, -0.297368715494811859, 5.20093245260149577, -0.422111106264731792, -0.609505049911372394, 0.430772405093700426, -0.075450459824149398, -0.0231177557010034982, 1.02271233462064678❩, '+
    '❨0.806374962066634831, 0.549866251807311702, 0.174137444046764151, -0.422111106264731792, 4.65483119820909985, 0.289795090828546352, 0.795841033040542465, -0.226199611183290607, -1.41597888676298831, 0.547038811317080901❩, '+'❨0.415539821124804545, 0.959584846946758171, 0.158862374235097095, -0.609505049911372394, 0.289795090828546352, 3.97252873681661176, 0.586614812369869111, -0.203645718668121608, 0.548222506108191402, -0.0260783651922211985❩, '+
    '❨-1.07626065503415, -0.559867163033034075, 0.0708587702680495379, 0.430772405093700426, 0.795841033040542465, 0.586614812369869111, 5.07303988296985066, 0.0852567827171805209, 0.721798919103536904, 0.0913845737686953809❩, '+'❨-0.174476853005194865, 0.601342029002866998, 0.527044617773810588, -0.0754504598241493979, -0.226199611183290607, -0.203645718668121608, 0.0852567827171805208, 5.72552879232507247, -0.381852810701084683, 0.248741195041378638❩, '+'❨0.6117764065082528, 0.908395415505047237, 0.717806236876411773, -0.0231177557010034982, -1.41597888676298831, 0.548222506108191402, 0.721798919103536904, -0.381852810701084683, 4.15028770968589867, 0.0862862168111860338❩, '+'❨0.479713560928682166, -0.0481674127191538071, 0.100941787359586736, 1.02271233462064678, 0.547038811317080901, -0.0260783651922211984, 0.0913845737686953809, 0.248741195041378638, 0.0862862168111860338, 3.64252242367312907❩❩; SpectralRadius(A)', 6.0);
  Eps(1E-13); Test('A ≔ ❨❨0.947217218464429805, -0.0797032009555874838, 0.455681246175844459, -0.890506530889571131, 0.0810219664601976321, -0.4241414032969027, -0.325611817072624359, 0.442131323976784809, 0.696001040652687542, -0.363371579820000821❩, '+'❨-0.0797032009555874838, 0.664262086761071853, 0.151183874385912782, 0.366175763890158299, -0.362499363441370523, 0.345609605325798349, -0.36153060182114391, -0.370383371480650959, -0.351151983839811129, -0.192995673028588803❩, '+
    '❨0.455681246175844459, 0.151183874385912782, 0.965112845520727124, -0.379442604818156945, -0.0868548299522641688, -0.703208127497219295, -0.29273075622571167, 0.392533036681992209, 0.433785161673175889, -0.0198556703071645884❩, '+'❨-0.890506530889571131, 0.366175763890158299, -0.379442604818156945, 1.60196556896465841, 0.0343051071296422081, 0.587093406299823787, -0.485735414772669726, -0.766652742695679709, -0.63672778172637717, -0.0824745994586258031❩, '+
    '❨0.0810219664601976321, -0.362499363441370523, -0.0868548299522641688, 0.0343051071296422081, 0.327534361557092576, -0.222948382709437083, -0.0500390113869091228, 0.161756725513035523, 0.21826384416565025, -0.135644537928391234❩, '+'❨-0.4241414032969027, 0.345609605325798349, -0.703208127497219295, 0.587093406299823787, -0.222948382709437083, 0.989037781112786131, -0.126211410841058615, -0.699854415829559306, -0.413031006405435901, 0.131956448780506888❩, '+
    '❨-0.325611817072624359, -0.36153060182114391, -0.29273075622571167, -0.485735414772669726, -0.0500390113869091228, -0.126211410841058615, 1.00217393220066957, 0.281471732316985816, -0.335032854995363073, 0.398331000774434439❩, '+'❨0.442131323976784809, -0.370383371480650959, 0.392533036681992209, -0.766652742695679709, 0.161756725513035523, -0.699854415829559306, 0.281471732316985816, 0.61286300125401041, 0.371247429623465179, -0.0611706531297214894❩, '+
    '❨0.696001040652687542, -0.351151983839811129, 0.433785161673175889, -0.63672778172637717, 0.21826384416565025, -0.413031006405435901, -0.335032854995363073, 0.371247429623465179, 0.960656361843560564, 0.149752326353616183❩, '+'❨-0.363371579820000821, -0.192995673028588803, -0.0198556703071645884, -0.0824745994586258031, -0.135644537928391234, 0.131956448780506888, 0.398331000774434439, -0.0611706531297214894, 0.149752326353616183, 0.929176842320993553❩❩; SpectralRadius(A)', 4.0);
  Eps(1E-13); Test('A ≔ IdentityMatrix(10); SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ZeroMatrix(10); SpectralRadius(A)', 0.0);
  Eps(1E-13); Test('A ≔ ❨❨2, 3, 0, 1❩, ❨5, 6, 8, 1❩, ❨4, 2, 5, 7❩, ❨3, 0, 2, 6❩❩; SpectralRadius(A)', 13.109772228646445);

  Eps(1E-13); Test('A ≔ ❨❨0, 0, 0, i❩, ❨0, 0, 1, 0❩, ❨0, 1, 0, 0❩, ❨−i, 0, 0, 0❩❩; SpectralRadius(A)', 1.0);
  Eps(1E-13); Test('A ≔ ❨❨5, 2, −i, 7, 4❩, ❨2, 3, 9, 1, i❩, ❨i, 9, 5, 9, 8❩, ❨7, 1, 9, 2, 3❩, ❨4, −i, 8, 3, 6❩❩; SpectralRadius(A)', 22.691445962489553);
  Eps(1E-13); Test('A ≔ ❨❨0, 0, −i, −1, 0❩, ❨0, 0, 0, 0, 1❩, ❨i, 0, 0, −i, 0❩, ❨−1, 0, i, 0, 0❩, ❨0, 1, 0, 0, 0❩❩; SpectralRadius(A)', 2.0);
  Eps(1E-13); Test('A ≔ ❨❨3, 2, 8, 0, i, 5, 7, 2❩, ❨2, 3, 9, 2, 7, 1, 2, 5❩, ❨8, 9, 9, 0, 3, 1, 5, 7❩, ❨0, 2, 0, 6, i, 3, 0, 8❩, ❨−i, 7, 3, −i, 7, 6, 1, −i❩, ❨5, 1, 1, 3, 6, 5, 2, 1❩, ❨7, 2, 5, 0, 1, 2, 3, 9❩, ❨2, 5, 7, 8, i, 1, 9, 6❩❩; SpectralRadius(A)', 31.541450879650441);
  Eps(1E-13); Test('A ≔ ❨❨−6, 2, 3, 5, −8❩, ❨4, 0, 2, 6, 5❩, ❨0, 0, 1, 6, 7❩, ❨5, 4, 8, 0, 2❩, ❨4, 3, 6, 6, 2❩❩; SpectralRadius(A)', 14.738054917903042);

  Test('delete(A)', success);

  //
  // Other matrix decompositions
  //

  // QR decomposition

  Test('A ≔ ❨❨21❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨6, 1❩, ❨5, 4❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨5, 2, 1❩, ❨4, 6, 8❩, ❨4, 3, 2❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨6, 3, 2, 5❩, ❨7, 1, 4, 5❩, ❨5, 6, 0, 4❩, ❨−5, 2, 8, 7❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨2, 1, 5, 3, 0❩, ❨5, 2, 0, 4, 7❩, ❨6, 3, 4, 2, 1❩, ❨−5, 3, 9, 5, 0❩, ❨5, 2, 1, −7, 9❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨3, 2, 5, 0, 1, 4❩, ❨7, 5, 6, 3, 2, 6❩, ❨4, 7, 8, 9, 5, 3❩, ❨4, 2, 0, 3, −4, −7❩, ❨6, 3, 5, 0, 1, 7❩, ❨5, 0, 2, 1, 0, 9❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨6, 3, 0, 2, 4, 1, 7❩, ❨5, 2, 0, −3, 1, 4, 7❩, ❨2, 5, 9, −6, −20, 4, 1❩, ❨7, 5, 2, 3, 6, −7, 4❩, ❨5, 6, 3, 0, 2, 8, 9❩, ❨0, 0, 0, 0, 0, 0, 0❩, ❨9, 5, 3, 4, 1, 5, −7❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨4, 7, 5, 6, 2, 1, 0, 5❩, ❨5, 4, 7, 6, 3, 2, 1, 4❩, ❨5, 2, 0, 8, 4, 6, 3, 2❩, ❨4, 7, 8, 5, 1, 6, 9, 0❩, ❨6, 3, 2, 7, 9, 1, 3, 4❩, ❨5, 6, 2, 5, 8, 0, 1, 4❩, ❨6, 3, 9, 5, 7, 4, 1, 8❩, ❨7, 4, 5, 6, 3, 2, 2, 8❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨4, 5, 2, 1, 5, 8, 9, 6, 3❩, ❨4, 1, 2, 5, 6, 4, 7, 8, 2❩, ❨0, 0, 0, 0, 0, 0, 0, 0, 0❩, ❨4, 6, 3, 2, 5, 8, 9, 60, 1❩, ❨4, 5, 6, 3, 5, 2, 8, 0, 1❩, ❨4, 3, 6, 3, 2, 2, 5, 9, 7❩, '+'❨4, 1, 2, 5, 6, 3, 8, 7, 5❩, ❨4, 5, 6, 3, 2, 5, 8, 9, 0❩, ❨5, 0, 2, 3, 6, 1, 4, 8, −7❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ IdentityMatrix(10);', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ReversalMatrix(10);', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ZeroMatrix(10);', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ CirculantMatrix(❨1, 2, 3, 4, 5, 6, 7, 8, 9, 0❩);', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ForwardShiftMatrix(10);', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ HilbertMatrix(10);', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨i/21❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨i, 3❩, ❨5, -i❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨3, 0, i❩, ❨4, i, i❩, ❨1, 2, 8❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨3, 5, 2, i❩, ❨1, 0, 4, 7❩, ❨5, i, i, 6❩, ❨1, 0, 7, 4❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨3, 6, 9, 8, i❩, ❨5, 4, i, 2, 3❩, ❨4, 0, 7, 1, i❩, ❨5, i, 3, 6, 9❩, ❨8, 5, 4, 4, 0❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨i, 4, 5, 6, 0, 2❩, ❨3, 0, 4, 7, 0, 4❩, ❨3, 6, i, 4, 1, i❩, ❨7, 0, 4, 1, 5, 9❩, ❨−5, 20, 100, 6, 7, i/1000❩, ❨1 + i, 5, 8, 4, −i, 6❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨i, 4, 0, i, 36, 4❩, ❨7, 0, −4, 5, 6, 9❩, ❨7, 3, 0, 4, i, 4❩, ❨0, 0, 0, 0, 0, 0❩, ❨0, 0, 0, 0, 0, 0❩, ❨4, i, 3, 6, 9, 8❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨4, 8, 3, i, 4, 1, 5❩, ❨4, 2, 0, 1, 4, 5, 9❩, ❨6, 3, 0, 5, 7, i, 4❩, ❨−4, 5, 8, 1, 1000, 6, 3❩, ❨4, 7, 5, 4, i, 6, 3❩, ❨1, 0, 2, 5, 9, 3, 1❩, ❨4, 6, 8, 5, 0, i, i❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨2, 5, 3, 0, 2, 1, 4, 8❩, ❨5, 6, 3, 0, 4, 2, 8, 9❩, ❨5, i, 4, 1, 2, 0, 7, 9❩, ❨5, 3, 6, 0, 5, 8, i, 7❩, ❨4, 2, 3, 6, 0, 11, 4, 5❩, ❨−5, −6, −7, −2, −4, −5, −9, −8❩, ❨4, 6, 0, i, 5, 7, 9, 6❩, ❨5, 3, 1, 4, 5, 0, 2, 8❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  StartStopwatch;
  _i := RandSeed;
  try
    RandSeed := 4318675;
    Test('A ≔ RandomMatrix(100);', null);
    Test('(Q, R) ≔ QR(A);', null);
    Test('IsUnitary(Q)', True);
    Test('IsUpperTriangular(R)', True);
    Test('A ≈ Q⋅R', True);
  finally
    RandSeed := _i;
  end;
  StopStopwatch;
  TestDuration(0.7);

{$IFNDEF QuickTest}
  StartStopwatch;
  _i := RandSeed;
  try
    RandSeed := 4318675;
    for _j := 1 to 100 do
    begin
      Test(Format('A ≔ RandomMatrix(%d);', [_j]), null);
      Test('(Q, R) ≔ QR(A);', null);
      Test('IsUnitary(Q)', True);
      Test('IsUpperTriangular(R)', True);
      Test('A ≈ Q⋅R', True);
    end;
  finally
    RandSeed := _i;
  end;
  StopStopwatch;
  TestDuration(15);
{$ENDIF}

  Test('A ≔ ❨❨5, 3, 6❩, ❨5, 0, 1❩, ❨8, 1, 7❩, ❨3, 0, 2❩, ❨4, 5, 8❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsOrthogonal(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('IsRealType(Q) ∧ IsRealType(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨6, i, 1, 5❩, ❨7, i, −2, 0❩, ❨4, 5, 3, 6❩, ❨4, i, 9, 0❩, ❨4, 1, 2, i❩, ❨5, 6, 3, 2❩, ❨4, 7, 8, 5❩, ❨1, 5, i, 6❩❩;', null);
  Test('(Q, R) ≔ QR(A);', null);
  Test('IsUnitary(Q)', True);
  Test('IsUpperTriangular(R)', True);
  Test('A ≈ Q⋅R', True);

  Test('A ≔ ❨❨6, 7, 1, 5❩, ❨7, 5, −2, 0❩❩;', null);
  Test('(Q, R) ≔ QR(A);', failure, 'QR decomposition only implemented for square and tall matrices.');

  Test('A ≔ ❨❨6, 7, 1, 5❩, ❨7, 5, −2, i❩❩;', null);
  Test('(Q, R) ≔ QR(A);', failure, 'QR decomposition only implemented for square and tall matrices.');

  Test('delete(A); delete(Q); delete(R)', success);

  // LU decomposition

  Test('A ≔ ❨❨21❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨6, 8❩, ❨5, 3❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨5, 1, 3❩, ❨6, 1, 0❩, ❨6, 8, 4❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨4, 0, 2, 3❩, ❨4, 8, 9, 6❩, ❨2, 0, 1, 4❩, ❨5, 6, 3, 2❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨4, 5, 6, 0, 1❩, ❨2, 3, 1, 0, 5❩, ❨8, 6, 6, 7, 9❩, ❨4, 0, 2, −1, 2❩, ❨1, 5, 6, 3, 2❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨3, 6, 2, 5, 4, 7❩, ❨1, 0, 2, 6, 9, 8❩, ❨4, 5, 6, 3, 2, 9❩, ❨4, 0, 1, 2, 8, 7❩, ❨5, 6, 3, 2, 9, 8❩, ❨5, 2, 1, 4, 7, 8❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨5, 7, 4, 9, 0, 3, 2❩, ❨5, 2, 1, 4, 6, 9, 0❩, ❨5, 3, 2, 8, 5, 9, 6❩, ❨4, 5, 2, 1, 6, 9, 0❩, ❨8, 5, 4, 7, 7, −2, 6❩, ❨5, 6, 3, 2, 5, 8, 9❩, ❨4, 0, 5, 7, 8, 0, 6❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨5, 4, 7, 5, 0, 3, 2, 6❩, ❨4, 5, 2, 1, 3, 6, 5, 8❩, ❨7, 3, 1, 0, 5, 2, 6, 9❩, ❨4, 2, 5, 0, 4, 7, 9, 6❩, ❨4, 3, 0, 2, 1, 5, 6, 7❩, ❨6, 5, 4, 8, 2, 3, 3, 6❩, ❨5, 2, 0, 1, 3, 5, 7, 8❩, ❨6, 5, 2, 0, 1, 4, 1, 2❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨5, 4, 7, 5, 0, 3, 2, 6❩, ❨4, 5, 2, 1, 3, 6, 5, 8❩, ❨7, 3, 1, 0, 5, 2, 6, 9❩, ❨10, 8, 14, 10, 0, 6, 4, 12❩, ❨4, 3, 0, 2, 1, 5, 6, 7❩, ❨6, 5, 4, 8, 2, 3, 3, 6❩, ❨5, 2, 0, 1, 3, 5, 7, 8❩, ❨6, 5, 2, 0, 1, 4, 1, 2❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ CirculantMatrix(❨1, 2, 3, 4, 5, 6, 7, 8, 9❩);', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  StartStopwatch;
  _i := RandSeed;
  try
    RandSeed := 4318675;
    for _j := 1 to 1000 do
    begin
      Test('A ≔ RandomMatrix(10);', null);
      Test('(P, L, U) ≔ LU(A);', null);
      Test('IsPermutation(P)', True);
      Test('IsLowerTriangular(L)', True);
      Test('IsUpperTriangular(U)', True);
      Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
      Test('A ≈ P*⋅L⋅U', True);
    end;
  finally
    RandSeed := _i;
  end;
  StopStopwatch;
  TestDuration(0.5);

{$IFNDEF QuickTest}
  StartStopwatch;
  _i := RandSeed;
  try
    RandSeed := 4318675;
    for _j := 1 to 100 do
    begin
      Test(Format('A ≔ RandomMatrix(%d);', [_j]), null);
      Test('(P, L, U) ≔ LU(A);', null);
      Test('IsPermutation(P)', True);
      Test('IsLowerTriangular(L)', True);
      Test('IsUpperTriangular(U)', True);
      Test('IsRealType(P) ∧ IsRealType(L) ∧ IsRealType(U)', True);
      Test('A ≈ P*⋅L⋅U', True);
    end;
  finally
    RandSeed := _i;
  end;
  StopStopwatch;
  TestDuration(20);
{$ENDIF}

  Test('A ≔ ❨❨4, 0, 2, 3❩, ❨4, 8, 9, 6❩, ❨2, 0, 1, 4❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', failure, 'LU decomposition only implemented for square matrices.');

  Test('A ≔ ❨❨4, 0❩, ❨4, 8❩, ❨2, 0❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', failure, 'LU decomposition only implemented for square matrices.');

  Test('A ≔ ❨❨5, 3, i❩, ❨1, −i, 7❩, ❨4, 9, 2❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨5, 3, i, 5❩, ❨1, 0, 8, i❩, ❨9, 2, 6, 5❩, ❨i, i, 2, 3❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨5, 3, i, 5, 9❩, ❨1, 2, 9, 2, 3❩, ❨5, 7, i, 6, 3❩, ❨4, 1, 2, i, 6❩, ❨5, 7, 0, 1, 4❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨5, i, 3, 0, 2, 4❩, ❨9, 0, 1, 2, 4, 6❩, ❨9, 0, 2, 3, 6, i❩, ❨4, 1, 7, 0, 9, 5❩, ❨6, 3, 2, 5, 0, i❩, ❨1, 20, 3, −5, 6, −9❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨5, 3, 0, 2, i, 4, 6❩, ❨2, 8, 0, 4, i, 6, 3❩, ❨7, 0, 21, i, 6, 3, 5❩, ❨0, 0, 0, 0, 0, 0, 0❩, ❨4, 8, i, 6, i, 1, 2❩, ❨5, 6, 9, 0, 2, 6, 7❩, ❨9, i, 4, 6, 3, 2, 4❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  Test('A ≔ ❨❨i, i, i, 1, 5, 2, 3, 0❩, ❨4, 6, 0, 3, 5, 9, 4, i❩, ❨1 + i, 2, 3, 6, 5, 8, 4 − 3⋅i, 5❩, ❨4, 6, 3, 0, 5, 8, 9, 7❩, ❨1, i, 6, 3, 2, 5, 0, 4❩, ❨5, 6, −7, −2, 1, 2, 33, 1000❩, ❨5, 6, 9, 8, 0, 1, 2, 0❩, ❨0, 0, 0, 0, 0, 0, 0, 0❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', null);
  Test('IsPermutation(P)', True);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ P*⋅L⋅U', True);

  StartStopwatch;
  _i := RandSeed;
  try
    RandSeed := 4318675;
    for _j := 1 to 1000 do
    begin
      Test('A ≔ RandomMatrix(10) + RandomMatrix(10)⋅i;', null);
      Test('(P, L, U) ≔ LU(A);', null);
      Test('IsPermutation(P)', True);
      Test('IsLowerTriangular(L)', True);
      Test('IsUpperTriangular(U)', True);
      Test('A ≈ P*⋅L⋅U', True);
    end;
  finally
    RandSeed := _i;
  end;
  StopStopwatch;
  TestDuration(0.8);

  Test('A ≔ ❨❨4, i, 2, 3❩, ❨4, 8, 9, 6❩, ❨2, 0, 1, 4❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', failure, 'LU decomposition only implemented for square matrices.');

  Test('A ≔ ❨❨4, i❩, ❨4, 8❩, ❨2, 0❩❩;', null);
  Test('(P, L, U) ≔ LU(A);', failure, 'LU decomposition only implemented for square matrices.');

  Test('A ≔ ❨❨6, 2, 3❩, ❨4, 2, 8❩, ❨1, 0, 7❩❩;', null);
  Test('(L, U) ≔ LU(A);', null);
  Test('IsLowerTriangular(L)', True);
  Test('IsUpperTriangular(U)', True);
  Test('IsRealType(L) ∧ IsRealType(U)', True);
  Test('A ≈ L⋅U', True);

  Test('delete(A); delete(P); delete(L); delete(U)', success);

  // Cholesky decomposition

  Test('A ≔ ❨❨6, 3, 2❩, ❨4, 0, 7❩, ❨4, 5, 8❩❩; U ≔ Cholesky(A)', failure, 'Couldn''t compute Cholesky factor.');
  Test('A ≔ ❨❨1, 3, 6❩, ❨5, 7, 0❩, ❨4, 5, 2❩❩; U ≔ Cholesky(A)', failure, 'Couldn''t compute Cholesky factor.');
  Test('A ≔ ❨❨4, 5, 8❩, ❨0, 1, 2❩, ❨6, 3, 7❩❩; U ≔ Cholesky(A)', failure, 'Couldn''t compute Cholesky factor.');
  Test('A ≔ ❨❨5, 6, 2, 4❩, ❨0, 1, −2, 7❩, ❨6, 3, 9, 8❩, ❨4, 3, 2, 5❩❩; U ≔ Cholesky(A)', failure, 'Couldn''t compute Cholesky factor.');

  Test('A ≔ ❨❨4.93597888649073641, 0.120575953845758162, -1.26079836241812257❩, ❨0.120575953845758162, 5.13324261553148423, -0.672122365157388638❩, ❨-1.26079836241812257, -0.672122365157388638, 5.93077849797777936❩❩;', null);
  Eps; Test('eigenvalues(A)', [7, 5, 4]);
  Test('IsPositiveDefinite(A)', True);
  Test('U ≔ Cholesky(A);', null);
  Test('IsRealType(U)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ U*⋅U', True);

  Test('A ≔ ❨❨3.71446288866910665, 1.97136689961010517, 1.36957261324154093, 1.08432078976074785❩, ❨1.97136689961010517, 2.64085379111499924, 1.03939132918314869, 0.307712302509145893❩, '+'❨1.36957261324154093, 1.03939132918314869, 3.7026136585939855, 1.34869495616383325❩, ❨1.08432078976074785, 0.307712302509145893, 1.34869495616383325, 2.94206966162190861❩❩;', null);
  Eps; Test('eigenvalues(A)', [7, 3, 2, 1]);
  Test('IsPositiveDefinite(A)', True);
  Test('U ≔ Cholesky(A);', null);
  Test('IsRealType(U)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ U*⋅U', True);

  Test('A ≔ ❨❨4.69244961335086436, -6.99003683346844873, -5.70144991196682375, 14.5654009984260046, -1.79267787409596933❩, ❨-6.99003683346844873, 14.102246678978645, 6.84207579510503739, -29.4503294077709206, 5.333709056019228❩, '+'❨-5.70144991196682375, 6.84207579510503739, 9.34933653581810971, -16.8252095806600843, 2.26175847549628046❩, ❨14.5654009984260046, -29.4503294077709206, -16.8252095806600843, 79.0669934855477114, -14.656691988681412❩, '+'❨-1.79267787409596933, 5.333709056019228, 2.26175847549628046, -14.656691988681412, 3.79897368630466946❩❩;', null);
  Eps; Test('eigenvalues(A)', [100, 7, 3, 1, 0.01]);
  Test('IsPositiveDefinite(A)', True);
  Test('U ≔ Cholesky(A);', null);
  Test('IsRealType(U)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ U*⋅U', True);

  Test('A ≔ ❨❨3.25569141373115836, 0.354843150024904775, -0.985274339831795524, 0.055669454145733476, -0.239379057083828038, 0.313951103595871587❩, '+'❨0.354843150024904775, 3.92438030278724386, -0.139064360631726472, -0.486972902917455806, 0.335571278864453504, 1.57495113789696396❩, '+'❨-0.985274339831795524, -0.139064360631726472, 4.1059099176304882, 0.457914417220631855, 0.0478988777066919346, -1.57111201529170735❩, '+'❨0.055669454145733476, -0.486972902917455806, 0.457914417220631855, 5.11247207849193247, -0.421896385014884285, -1.28677508019808249❩, '+'❨-0.239379057083828038, 0.335571278864453504, 0.0478988777066919348, -0.421896385014884284, 4.11143853521532303, -0.599959767468462238❩, '+'❨0.313951103595871587, 1.57495113789696396, -1.57111201529170735, -1.28677508019808249, -0.599959767468462238, 5.49010775214385409❩❩;', null);
  Eps; Test('eigenvalues(A)', [8, 5, 4, 4, 3, 2]);
  Test('IsPositiveDefinite(A)', True);
  Test('U ≔ Cholesky(A);', null);
  Test('IsRealType(U)', True);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ U*⋅U', True);

  StartStopwatch;
  _i := RandSeed;
  try
    RandSeed := 4318675;
    for _j := 1 to 1000 do
    begin
      Test('E ≔ GramSchmidt(RandomMatrix(10)); D ≔ diag(RandomIntVector(10, 1, 100)); A ≔ E⋅D⋅E*;', null);
      Test('IsPositiveDefinite(A)', True);
      Test('U ≔ Cholesky(A);', null);
      Test('IsRealType(U)', True);
      Test('IsUpperTriangular(U)', True);
      Test('A ≈ U*⋅U', True);
    end;
  finally
    RandSeed := _i;
  end;
  StopStopwatch;
  TestDuration(1.75);

  Test('Cholesky(❨❨3, 5, 1❩, ❨1, 2, 3❩❩)', failure, 'Cannot compute Cholesky decomposition of non-square matrix.');

  Test('A ≔ ❨❨6, 3, 2❩, ❨4, i, 7❩, ❨4, 5, 8❩❩; U ≔ Cholesky(A)', failure, 'Couldn''t compute Cholesky factor.');
  Test('A ≔ ❨❨1, 3, 6❩, ❨5, 7, 0❩, ❨4, 5, i❩❩; U ≔ Cholesky(A)', failure, 'Couldn''t compute Cholesky factor.');
  Test('A ≔ ❨❨i, 5, 8❩, ❨i, 1, 2❩, ❨6, 3, 7❩❩; U ≔ Cholesky(A)', failure, 'Couldn''t compute Cholesky factor.');
  Test('A ≔ ❨❨5, 6, 2, 4❩, ❨i, 1, −2, 7❩, ❨6, 3, i, 8❩, ❨4, i, 2, 5❩❩; U ≔ Cholesky(A)', failure, 'Couldn''t compute Cholesky factor.');

  Test('A ≔ ❨❨4.50691891236022514, 0.805836912347700675 - 0.781252166154897276⋅i, -1.60956385463046235 + 1.10788519536159756⋅i❩, '+'❨0.805836912347700675 + 0.781252166154897276⋅i, 5.86871548240192601 - 2.16840434497100887E-19⋅i, 0.0519096184238235582 - 0.318264548670979053⋅i❩, '+'❨-1.60956385463046235 - 1.10788519536159756⋅i, 0.0519096184238235582 + 0.318264548670979053⋅i, 6.62436560523784885 - 4.06575814682064163E-20⋅i❩❩;', null);
  Eps; Test('eigenvalues(A)', [8, 6, 3]);
  Test('IsPositiveDefinite(A)', True);
  Test('U ≔ Cholesky(A);', null);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ U*⋅U', True);

  Test('A ≔ ❨❨2.42567155884810701 + 1.69406589450860068E-21⋅i, -0.541458499166476172 + 0.56837332764644428⋅i, -0.393760848951031678 - 0.200974439686655447⋅i, 0.178116003053974962 - 0.306653013032412579⋅i❩, '+'❨-0.541458499166476172 - 0.56837332764644428⋅i, 2.9538256920912331, -0.536922443662477092 - 0.253491327740968221⋅i, -0.689023915988372674 + 0.192306297849908853⋅i❩, '+'❨-0.393760848951031678 + 0.200974439686655447⋅i, -0.536922443662477092 + 0.253491327740968221⋅i, 2.31141835941757323, -0.22890956682993686 + 0.710034350116213099⋅i❩, '+'❨0.178116003053974962 + 0.306653013032412579⋅i, -0.689023915988372674 - 0.192306297849908853⋅i, -0.22890956682993686 - 0.710034350116213099⋅i, 2.30908438964308665❩❩;', null);
  Eps; Test('eigenvalues(A)', [4, 3, 2, 1]);
  Test('IsPositiveDefinite(A)', True);
  Test('U ≔ Cholesky(A);', null);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ U*⋅U', True);

  Test('A ≔ ❨❨36.3817607324922451 + 8.67361737988403547E-19⋅i, -36.6631490448397316 + 11.9438621354627077⋅i, 16.1462702944291521 - 6.45264231481579467⋅i, 8.32436985104583655 + 0.472261109207546581⋅i, -9.19237247530613032 + 12.7318875244064349⋅i❩, '+
    '❨-36.6631490448397316 - 11.9438621354627077⋅i, 67.8912853886941801 - 1.73472347597680709E-18⋅i, -3.70294041953983936 - 6.47584262976698282⋅i, -10.0301011261191165 + 4.80142282853594383⋅i, -15.6514488868485211 - 1.84532777081835844⋅i❩, '+
    '❨16.1462702944291521 + 6.45264231481579467⋅i, -3.70294041953983936 + 6.47584262976698282⋅i, 20.2238960752888459, 1.0433984716634792 + 6.91050096705167366⋅i, -27.3678999125304748 - 1.06525611362937123⋅i❩, '+
    '❨8.32436985104583655 - 0.472261109207546581⋅i, -10.0301011261191165 - 4.80142282853594383⋅i, 1.0433984716634792 - 6.91050096705167366⋅i, 8.00119570117276258, 3.66804497600672684 + 15.1619742113008918⋅i❩, '+
    '❨-9.19237247530613032 - 12.7318875244064349⋅i, -15.6514488868485211 + 1.84532777081835844⋅i, -27.3678999125304748 + 1.06525611362937123⋅i, 3.66804497600672684 - 15.1619742113008918⋅i, 53.5118621023519662❩❩;', null);
  Eps; Test('eigenvalues(A)', [100, 80, 4, 2, 0.01]);
  Test('IsPositiveDefinite(A)', True);
  Test('U ≔ Cholesky(A);', null);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ U*⋅U', True);

  Test('A ≔ ❨❨4.34340365873964928 + 2.71050543121376108E-20⋅i, -0.270944788535373213 + 0.579123555674773119⋅i, 0.19735481426685348 + 0.0124292686296485832⋅i, '+'3.97275958037236907 - 3.12323842085080762⋅i, 0.561937145744118512 - 0.72958734512521⋅i, -0.6262804092054267 + 1.7607648350037341⋅i❩, '+
    '❨-0.270944788535373213 - 0.579123555674773119⋅i, 5.79497317381224502 + 3.81164826264435153E-21⋅i, -0.0730171048477857757 - 0.0286433657347114163⋅i, '+'-3.11349167800771644 + 0.666002984820350029⋅i, -0.364589053672550824 + 1.59369429955185477⋅i, 3.30418029981095185 + 1.09829976554579104⋅i❩, '+
    '❨0.19735481426685348 - 0.0124292686296485833⋅i, -0.0730171048477857757 + 0.0286433657347114163⋅i, 5.13651914269719606 - 1.33407689192552303E-20⋅i, 0.867664101100140022 - 1.21101352008563466⋅i, '+'1.73306691447269943 - 1.40209341706461094⋅i, 0.726979190979306421 + 0.98143185800994314⋅i❩, '+
    '❨3.97275958037236907 + 3.12323842085080762⋅i, -3.11349167800771644 - 0.666002984820350029⋅i, 0.867664101100140022 + 1.21101352008563466⋅i, 10.4528342533782368 + 1.74700545371199445E-21⋅i, '+'4.1144040114747787 - 0.33680898741005859⋅i, -3.97640945258550905 - 2.41659196047885716⋅i❩, '+
    '❨0.561937145744118512 + 0.72958734512521⋅i, -0.364589053672550824 - 1.59369429955185477⋅i, 1.73306691447269943 + 1.40209341706461094⋅i, 4.1144040114747787 + 0.33680898741005859⋅i, '+'5.20913085698579575 - 1.07573184301296143E-19⋅i, -2.57546656541027782 - 2.58593478583507784⋅i❩, '+
    '❨-0.6262804092054267 - 1.7607648350037341⋅i, 3.30418029981095185 - 1.09829976554579104⋅i, 0.726979190979306421 - 0.98143185800994314⋅i, -3.97640945258550905 + 2.41659196047885716⋅i, '+'-2.57546656541027782 + 2.58593478583507784⋅i, 8.36313891438687709 - 1.21972744404619249E-19⋅i❩❩;', null);
  Eps; Test('eigenvalues(A)', [20, 8, 6, 4, 1, 0.3]);
  Test('IsPositiveDefinite(A)', True);
  Test('U ≔ Cholesky(A);', null);
  Test('IsUpperTriangular(U)', True);
  Test('A ≈ U*⋅U', True);

  StartStopwatch;
  _i := RandSeed;
  try
    RandSeed := 4318675;
    for _j := 1 to 100 do
    begin
      Test('E ≔ GramSchmidt(RandomMatrix(10) + RandomMatrix(10)⋅i); D ≔ diag(RandomIntVector(10, 1, 100)); A ≔ E⋅D⋅E*;', null);
      Test('IsPositiveDefinite(A)', True);
      Test('U ≔ Cholesky(A);', null);
      Test('IsUpperTriangular(U)', True);
      Test('A ≈ U*⋅U', True);
    end;
  finally
    RandSeed := _i;
  end;
  StopStopwatch;
  TestDuration(0.75);

  Test('Cholesky(❨❨3, 5, i❩, ❨1, 2, 3❩❩)', failure, 'Cannot compute Cholesky decomposition of non-square matrix.');

  Test('delete(A); delete(E); delete(D); delete(U)', success);

  // Singular values

  Eps(1E-14); Test('A ≔ ❨❨6, 3, 2❩, ❨1, 0, 4❩, ❨7, 5, 9❩❩; SingularValues(A)',
    [14.299417303552637, 3.955854687507967, 0.936951690419907]);

  Eps(1E-14); Test('A ≔ ❨❨6, 3, 2, 8❩, ❨−1, 4, 5, 3❩, ❨5, 6, 9, 0❩, ❨1, 3, 4, 2❩❩; SingularValues(A)',
    [15.972205617385749, 7.872260828866542, 4.348624447378992, 0.074983824272194]);

  Eps(1E-14); Test('A ≔ ❨❨0, 2, 8, 5, 1❩, ❨9, 3, 5, 4, 6❩, ❨1, 0, 2, 4, 7❩, ❨5, 2, 6, 3, 7❩, ❨4, 0, 1, 2, 4❩❩; SingularValues(A)',
    [20.060555101346626, 7.551329780423160, 5.251105360879447, 1.872338213422416, 0.686869475687951]);

  Eps(1E-14); Test('A ≔ ❨❨2, 3, 6, 5, 8, 1❩, ❨4, 5, 6, 3, 2, 6❩, ❨9, 0, 1, 4, 5, 2❩, ❨4, 1, 4, 2, 8, 0❩, ❨6, 9, 0, 1, 4, 5❩, ❨6, 3, 2, 5, 8, 7❩❩; SingularValues(A)',
    [25.493993327503308, 9.561016689465456, 7.373500539267359, 5.077075421109402, 3.761916202225462, 1.829219925051135]);

  Eps(1E-14); Test('A ≔ ❨❨5, 6, 3, 2, 1❩, ❨4, 5, 2, 8, 0❩, ❨4, 9, 6, 3, 5❩❩; SingularValues(A)',
    [17.527362377345845, 6.092045778009372, 2.584288360814148]);

  Eps(1E-14); Test('A ≔ ❨❨5, 8❩, ❨1, 3❩, ❨5, 9❩, ❨5, −4❩, ❨1, 8❩, ❨3, −4❩❩; SingularValues(A)',
    [16.493005456603921, 7.998798097741513]);

  Eps(1E-14); Test('A ≔ ❨❨5, i, 3❩, ❨1, 0, i❩, ❨4, 5, 7❩❩; SingularValues(A)',
    [10.628784294206016, 3.642097016433973, 0.874113122039879]);

  Eps(1E-14); Test('A ≔ ❨❨5, 2, i, 4❩, ❨1, 0, 3, 6❩, ❨6, 3, i, 1❩, ❨i, 7, 4, 9❩❩; SingularValues(A)',
    [14.604621042142906, 7.538203029364286, 3.837684363640320, 1.073647068896616]);

  Eps(1E-14); Test('A ≔ ❨❨3, 6, 2, 4, i❩, ❨5, 7, 4, 1, 5❩, ❨1, 0, i, 6, 9❩, ❨6, 3, 2, 5, 8❩, ❨4, 5, i, 8, 9❩❩; SingularValues(A)',
    [22.668597020219032, 9.181152640007719, 4.621352049130089, 2.227629797014103, 1.233659621310555]);

  Eps(1E-14); Test('A ≔ ❨❨3, 2, 5, 1, i, 4❩, ❨4, 0, 3, 2, 1, 6❩, ❨8, 7, i, 4, 5, 6❩, ❨4, 5, 6, 1, 0, 8❩, ❨6, 5, 2, 7, 8, i❩, ❨7, 5, 0, 2, 9, 6❩❩; SingularValues(A)',
    [25.412555473507545, 11.281608112707431, 5.787205131709382, 3.683491102796632, 2.882401896112685, 0.747831575701784]);

  Eps(1E-14); Test('A ≔ ❨❨2, 5, 6, 8, 7❩, ❨i, 4, 5, 1, 2❩❩; SingularValues(A)',
    [14.486417927389791, 3.891490155840011]);

  Eps(1E-14); Test('A ≔ ❨❨5, 1, 2❩, ❨6, i, 2❩, ❨9, 7, 4❩, ❨5, 6, i❩, ❨5, 3, −3❩❩; SingularValues(A)',
    [16.624178394796832, 5.433725873857354, 4.013890360453734]);

  Eps(1E-14); Test('A ≔ ❨❨3, 5, 0, 2, −8, 7, 3, 2❩❩; SingularValues(A)',
    [12.806248474865699]);

  Eps(1E-14); Test('A ≔ ❨❨5❩, ❨2❩, ❨i❩, ❨−3❩❩; SingularValues(A)',
    [6.244997998398398]);

  Test('delete(A)', success);

  // Similar Hessenberg matrix

  Test('A ≔ ❨❨6, 1, 3❩, ❨7, 3, 1❩, ❨2, 4, 6❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', null);
  Test('IsRealType(H) ∧ IsRealType(U)', True);
  Test('IsOrthogonal(U) ∧ IsUpperHessenberg(H)', True);
  Test('A ≈ U ⋅ H ⋅ U*', True);

  Test('A ≔ ❨❨3, 0, 2, 5❩, ❨7, 1, 4, 5❩, ❨−3, 6, 8, 9❩, ❨4, 5, 2, 6❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', null);
  Test('IsRealType(H) ∧ IsRealType(U)', True);
  Test('IsOrthogonal(U) ∧ IsUpperHessenberg(H)', True);
  Test('A ≈ U ⋅ H ⋅ U*', True);

  Test('A ≔ ❨❨6, 3, 0, 2, 5❩, ❨7, 1, 0, 6, 5❩, ❨8, 1, 4, 5, 6❩, ❨8, 5, 2, 3, 6❩, ❨4, 5, 2, 0, 1❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', null);
  Test('IsRealType(H) ∧ IsRealType(U)', True);
  Test('IsOrthogonal(U) ∧ IsUpperHessenberg(H)', True);
  Test('A ≈ U ⋅ H ⋅ U*', True);

  Test('A ≔ ❨❨6, 5, 2, 0, 1, 4❩, ❨8, 5, 0, 4, 1, 9❩, ❨−5, 2, 3, 6, −4, 7❩, ❨8, 0, 1, 4, 5, 3❩, ❨6, 5, 2, 3, 0, 4❩, ❨8, 5, 6, 9, −7, 5❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', null);
  Test('IsRealType(H) ∧ IsRealType(U)', True);
  Test('IsOrthogonal(U) ∧ IsUpperHessenberg(H)', True);
  Test('A ≈ U ⋅ H ⋅ U*', True);

  Test('A ≔ ❨❨8, 0, 1, 4, 5, 3❩, ❨6, 5, 2, 3, 0, 4❩, ❨8, 5, 6, 9, −7, 5❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', failure, 'Cannot find similar Hessenberg matrix of non-square matrix.');

  Test('A ≔ ❨❨5, i, 6❩, ❨1, 3, 4❩, ❨6, 0, 4❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', null);
  Test('IsUnitary(U) ∧ IsUpperHessenberg(H)', True);
  Test('A ≈ U ⋅ H ⋅ U*', True);

  Test('A ≔ ❨❨3, 5, i, 9❩, ❨5, 2, 1, 4❩, ❨1, 2, 5, i❩, ❨7, 0, 9, 6❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', null);
  Test('IsUnitary(U) ∧ IsUpperHessenberg(H)', True);
  Test('A ≈ U ⋅ H ⋅ U*', True);

  Test('A ≔ ❨❨6, 2, 5, 4, i❩, ❨5, 0, 1, 4, 9❩, ❨8, 0, 3, 2, 6❩, ❨5, 2, 1, i, 4❩, ❨6, 5, 2, 3, i❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', null);
  Test('IsUnitary(U) ∧ IsUpperHessenberg(H)', True);
  Test('A ≈ U ⋅ H ⋅ U*', True);

  Test('A ≔ ❨❨6, 5, 2, 4, i, 1❩, ❨4, 7, 2, 5, 3, 6❩, ❨5, 4, 7, 8, 9, i❩, ❨6, 0, 2, 4, 3, 1❩, ❨5, 1, 2, 4, 7, 3❩, ❨6, 5, 4, 0, 3, 7❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', null);
  Test('IsUnitary(U) ∧ IsUpperHessenberg(H)', True);
  Test('A ≈ U ⋅ H ⋅ U*', True);

  Test('A ≔ ❨❨8, i❩, ❨6, i❩, ❨8, 5❩❩;', null);
  Test('(H, U) ≔ Hessenberg(A);', failure, 'Cannot find similar Hessenberg matrix of non-square matrix.');

  Test('delete(A); delete(H); delete(U)', success);


  //
  // Misc. matrix functions
  //

  // Minors

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −1, 4, 5❩, ❨9, 6, 3, 0, 5❩, ❨7, 3, 0, 6, 4❩❩;', null);
  Eps; Test('minor(A, 3, 2)', -479.0);
  Eps; Test('minor(A, 1, 5)', 45.0);
  Eps; Test('minor(A, 4, 3)', -74.0);

  Test('minor(A, 6, 3)', failure, 'Invalid row or column index');
  Test('minor(A, 2, 7)', failure, 'Invalid row or column index');

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −1, 4, 5❩❩;', null);
  Test('minor(A, 3, 2)', failure, 'Cannot compute determinant of non-square matrix.');

  Test('A ≔ ❨❨5, i, 3, 6, 2, −1❩, ❨4, 5, 8, i, 9, −5❩, ❨4, 2, 5, 3, −1, i❩, ❨5, 7, 0, 2, 3, 6❩, ❨−i, 4, 5, 0, 2, 9❩, ❨7, 5, 9, 0, −4, i❩❩;', null);
  Eps; Test('minor(A, 5, 2)', 7275 + 1755*ImaginaryUnit);
  Eps; Test('minor(A, 3, 6)', 19222 + 2216*ImaginaryUnit);
  Eps; Test('minor(A, 4, 4)', 2393 + 213*ImaginaryUnit);
  Eps; Test('minor(A, 1, 3)', 63 + 1347*ImaginaryUnit);

  Test('minor(A, 7, 3)', failure, 'Invalid row or column index');
  Test('minor(A, 2, 120)', failure, 'Invalid row or column index');

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −i, 4, 5❩❩;', null);
  Test('minor(A, 3, 2)', failure, 'Cannot compute determinant of non-square matrix.');

  Test('A ≔ ❨❨5, 3❩, ❨4, 1❩❩;', null);
  Eps; Test('minor(A, 1, 1)', 1.0);
  Eps; Test('minor(A, 1, 2)', 4.0);
  Eps; Test('minor(A, 2, 1)', 3.0);
  Eps; Test('minor(A, 2, 2)', 5.0);

  Test('minor(A, 2, 2, 1)', failure, 'Too many arguments.');
  Test('minor(A, 2)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('minor(A, 2, -1)', failure, 'A positive integer was expected as argument 3, but "-1" was given.');

  Test('A ≔ ❨❨21❩❩;', null);
  Test('minor(A, 1, 1)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨21, 22, 23, 24, 25❩❩;', null);
  Test('minor(A, 1, 4)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨1❩, ❨2❩, ❨3❩, ❨4❩❩;', null);
  Test('minor(A, 3, 1)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨21⋅i❩❩;', null);
  Test('minor(A, 1, 1)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨21, 22⋅i, 23, 24, 25❩❩;', null);
  Test('minor(A, 1, 4)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨1❩, ❨2⋅i❩, ❨3❩, ❨4❩❩;', null);
  Test('minor(A, 3, 1)', failure, 'A matrix must have size at least 1×1.');

  // Cofactors

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −1, 4, 5❩, ❨9, 6, 3, 0, 5❩, ❨7, 3, 0, 6, 4❩❩;', null);
  Eps; Test('cofactor(A, 3, 2)', 479.0);
  Eps; Test('cofactor(A, 1, 5)', 45.0);
  Eps; Test('cofactor(A, 4, 3)', 74.0);

  Test('cofactor(A, 6, 3)', failure, 'Invalid row or column index');
  Test('cofactor(A, 2, 7)', failure, 'Invalid row or column index');

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −1, 4, 5❩❩;', null);
  Test('cofactor(A, 3, 2)', failure, 'Cannot compute determinant of non-square matrix.');

  Test('A ≔ ❨❨5, i, 3, 6, 2, −1❩, ❨4, 5, 8, i, 9, −5❩, ❨4, 2, 5, 3, −1, i❩, ❨5, 7, 0, 2, 3, 6❩, ❨−i, 4, 5, 0, 2, 9❩, ❨7, 5, 9, 0, −4, i❩❩;', null);
  Eps; Test('cofactor(A, 5, 2)', -7275 - 1755*ImaginaryUnit);
  Eps; Test('cofactor(A, 3, 6)', -19222 - 2216*ImaginaryUnit);
  Eps; Test('cofactor(A, 4, 4)', 2393 + 213*ImaginaryUnit);
  Eps; Test('cofactor(A, 1, 3)', 63 + 1347*ImaginaryUnit);

  Test('cofactor(A, 7, 3)', failure, 'Invalid row or column index');
  Test('cofactor(A, 2, 120)', failure, 'Invalid row or column index');

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −i, 4, 5❩❩;', null);
  Test('cofactor(A, 3, 2)', failure, 'Cannot compute determinant of non-square matrix.');

  Test('A ≔ ❨❨5, 3❩, ❨4, 1❩❩;', null);
  Eps; Test('cofactor(A, 1, 1)', 1.0);
  Eps; Test('cofactor(A, 1, 2)', -4.0);
  Eps; Test('cofactor(A, 2, 1)', -3.0);
  Eps; Test('cofactor(A, 2, 2)', 5.0);

  Test('cofactor(A, 2, 2, 1)', failure, 'Too many arguments.');
  Test('cofactor(A, 2)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('cofactor(A, 2, -1)', failure, 'A positive integer was expected as argument 3, but "-1" was given.');

  Test('A ≔ ❨❨21❩❩;', null);
  Test('cofactor(A, 1, 1)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨21, 22, 23, 24, 25❩❩;', null);
  Test('cofactor(A, 1, 4)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨1❩, ❨2❩, ❨3❩, ❨4❩❩;', null);
  Test('cofactor(A, 3, 1)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨21⋅i❩❩;', null);
  Test('cofactor(A, 1, 1)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨21, 22⋅i, 23, 24, 25❩❩;', null);
  Test('cofactor(A, 1, 4)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨1❩, ❨2⋅i❩, ❨3❩, ❨4❩❩;', null);
  Test('cofactor(A, 3, 1)', failure, 'A matrix must have size at least 1×1.');

  // Cofactor matrices

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −1, 4, 5❩, ❨9, 6, 3, 0, 5❩, ❨7, 3, 0, 6, 4❩❩;', null);
  Eps; Test('CofactorMatrix(A)', 5, [-360, 580, -155, 100, 45, 270, -452, 129, -75, -21, -285, 479, -173, 65, 42, 145, -227, 74, -45, -16, 265, -426, 137, -50, -48]);
  Test('A ≔ ❨❨5, i, 3, 6, 2, −1❩, ❨4, 5, 8, i, 9, −5❩, ❨4, 2, 5, 3, −1, i❩, ❨5, 7, 0, 2, 3, 6❩, ❨−i, 4, 5, 0, 2, 9❩, ❨7, 5, 9, 0, −4, i❩❩;', null);
  Eps; Test('CofactorMatrix(A)', 6, [24305 - 1220*ImaginaryUnit, -26508 - 954*ImaginaryUnit, 63 + 1347*ImaginaryUnit, -11186 - 2548*ImaginaryUnit, 9067 + 2170*ImaginaryUnit,
    9867 + 1894*ImaginaryUnit, -2217 - 1021*ImaginaryUnit, 2724 + 1074*ImaginaryUnit, 1851 - 1011*ImaginaryUnit, -564 + 2152*ImaginaryUnit, 3539 - 3447*ImaginaryUnit,
    -2912 + 604*ImaginaryUnit, -49001 + 5799*ImaginaryUnit, 47382 + 2256*ImaginaryUnit, 3117 - 5727*ImaginaryUnit, 42767 - 8661*ImaginaryUnit, -18957 - 4723*ImaginaryUnit,
    -19222 - 2216*ImaginaryUnit, 76 - 3930*ImaginaryUnit, 8988 - 1884*ImaginaryUnit, -5370 + 3624*ImaginaryUnit, 2393 + 213*ImaginaryUnit, -489 - 1195*ImaginaryUnit,
    -466 - 902*ImaginaryUnit, 1535 + 5113*ImaginaryUnit, -7275 - 1755*ImaginaryUnit, 4077 - 3333*ImaginaryUnit, -3572 - 2032*ImaginaryUnit, 3268 + 946*ImaginaryUnit,
    6765 - 2009*ImaginaryUnit, 20239 - 4748*ImaginaryUnit, -15867 - 915*ImaginaryUnit, 1428 + 882*ImaginaryUnit, -17545 + 4877*ImaginaryUnit, 2548 + 4439*ImaginaryUnit,
    6220 + 1179*ImaginaryUnit]);
  Test('A ≔ ❨❨5, 3❩, ❨4, 1❩❩;', null);
  Eps; Test('CofactorMatrix(A)', 2, [1, -4, -3, 5]);

  Test('CofactorMatrix(A, 2)', failure, 'Too many arguments.');

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −1, 4, 5❩❩;', null);
  Test('CofactorMatrix(A)', failure, 'Cannot compute cofactor matrix of non-square matrix.');

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −i, 4, 5❩❩;', null);
  Test('CofactorMatrix(A)', failure, 'Cannot compute cofactor matrix of non-square matrix.');

  Test('A ≔ ❨❨21❩❩;', null);
  Test('CofactorMatrix(A)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨21, 22⋅i, 23, 24, 25❩❩;', null);
  Test('CofactorMatrix(A)', failure, 'Cannot compute cofactor matrix of non-square matrix.');

  Test('A ≔ ❨❨1❩, ❨2❩, ❨3❩, ❨4❩❩;', null);
  Test('CofactorMatrix(A)', failure, 'Cannot compute cofactor matrix of non-square matrix.');

  // Adjugate matrices

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −1, 4, 5❩, ❨9, 6, 3, 0, 5❩, ❨7, 3, 0, 6, 4❩❩;', null);
  Eps; Test('AdjugateMatrix(A)', 5, [-360, 270, -285, 145, 265, 580, -452, 479, -227, -426, -155, 129, -173, 74, 137, 100, -75, 65, -45, -50, 45, -21, 42, -16, -48]);
  Test('A ≔ ❨❨5, i, 3, 6, 2, −1❩, ❨4, 5, 8, i, 9, −5❩, ❨4, 2, 5, 3, −1, i❩, ❨5, 7, 0, 2, 3, 6❩, ❨−i, 4, 5, 0, 2, 9❩, ❨7, 5, 9, 0, −4, i❩❩;', null);
  Eps; Test('AdjugateMatrix(A)', 6, [24305 - 1220*ImaginaryUnit, -2217 - 1021*ImaginaryUnit, -49001 + 5799*ImaginaryUnit, 76 - 3930*ImaginaryUnit,
    1535 + 5113*ImaginaryUnit, 20239 - 4748*ImaginaryUnit, -26508 - 954*ImaginaryUnit, 2724 + 1074*ImaginaryUnit, 47382 + 2256*ImaginaryUnit,
    8988 - 1884*ImaginaryUnit, -7275 - 1755*ImaginaryUnit, -15867 - 915*ImaginaryUnit, 63 + 1347*ImaginaryUnit, 1851 - 1011*ImaginaryUnit,
    3117 - 5727*ImaginaryUnit, -5370 + 3624*ImaginaryUnit, 4077 - 3333*ImaginaryUnit, 1428 + 882*ImaginaryUnit, -11186 - 2548*ImaginaryUnit,
    -564 + 2152*ImaginaryUnit, 42767 - 8661*ImaginaryUnit, 2393 + 213*ImaginaryUnit, -3572 - 2032*ImaginaryUnit, -17545 + 4877*ImaginaryUnit,
    9067 + 2170*ImaginaryUnit, 3539 - 3447*ImaginaryUnit, -18957 - 4723*ImaginaryUnit, -489 - 1195*ImaginaryUnit, 3268 + 946*ImaginaryUnit,
    2548 + 4439*ImaginaryUnit, 9867 + 1894*ImaginaryUnit, -2912 + 604*ImaginaryUnit, -19222 - 2216*ImaginaryUnit, -466 - 902*ImaginaryUnit,
    6765 - 2009*ImaginaryUnit, 6220 + 1179*ImaginaryUnit]);
  Test('A ≔ ❨❨5, 3❩, ❨4, 1❩❩;', null);
  Eps; Test('AdjugateMatrix(A)', 2, [1, -3, -4, 5]);

  Test('AdjugateMatrix(A, 2)', failure, 'Too many arguments.');

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −1, 4, 5❩❩;', null);
  Test('AdjugateMatrix(A)', failure, 'Cannot compute cofactor matrix of non-square matrix.');

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −i, 4, 5❩❩;', null);
  Test('AdjugateMatrix(A)', failure, 'Cannot compute cofactor matrix of non-square matrix.');

  Test('A ≔ ❨❨21❩❩;', null);
  Test('AdjugateMatrix(A)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨21, 22⋅i, 23, 24, 25❩❩;', null);
  Test('AdjugateMatrix(A)', failure, 'Cannot compute cofactor matrix of non-square matrix.');

  Test('A ≔ ❨❨1❩, ❨2❩, ❨3❩, ❨4❩❩;', null);
  Test('AdjugateMatrix(A)', failure, 'Cannot compute cofactor matrix of non-square matrix.');

  Test('A ≔ ❨❨6, 3, 2, 5, 7❩, ❨4, 1, 0, 5, 8❩, ❨7, 3, −1, 4, 5❩, ❨9, 6, 3, 0, 5❩, ❨7, 3, 0, 6, 4❩❩;', null);
  Test('C ≔ CofactorMatrix(A); Adj ≔ AdjugateMatrix(A);', null);
  Test('C = transpose(Adj)', True);
  Test('SameValue(Adj ⋅ A, det(A) ⋅ IdentityMatrix(5), 1E−15)', True);
  Test('SameValue(A ⋅ Adj, det(A) ⋅ IdentityMatrix(5), 1E−15)', True);
  Test('A^−1 ≈ det(A)^−1 ⋅ Adj', True);

  Test('A ≔ ❨❨5, i, 3, 6, 2, −1❩, ❨4, 5, 8, i, 9, −5❩, ❨4, 2, 5, 3, −1, i❩, ❨5, 7, 0, 2, 3, 6❩, ❨−i, 4, 5, 0, 2, 9❩, ❨7, 5, 9, 0, −4, i❩❩;', null);
  Test('C ≔ CofactorMatrix(A); Adj ≔ AdjugateMatrix(A);', null);
  Test('C = transpose(Adj)', True);
  Test('SameValue(Adj ⋅ A, det(A) ⋅ IdentityMatrix(6), 1E−13)', True);
  Test('SameValue(A ⋅ Adj, det(A) ⋅ IdentityMatrix(6), 1E−13)', True);
  Test('A^−1 ≈ det(A)^−1 ⋅ Adj', True);

  Test('x ≔ ❨1, 5, i, 3, 2, −3❩; b ≔ A⋅x;', null);
  Test('x ≈ Adj ⋅ b / det(A)', True);
  Test('x[1] ≈ det(ReplaceCol(A, 1, b)) / det(A)', True);
  Test('x[2] ≈ det(ReplaceCol(A, 2, b)) / det(A)', True);
  Test('x[3] ≈ det(ReplaceCol(A, 3, b)) / det(A)', True);
  Test('x[4] ≈ det(ReplaceCol(A, 4, b)) / det(A)', True);
  Test('x[5] ≈ det(ReplaceCol(A, 5, b)) / det(A)', True);
  Test('x[6] ≈ det(ReplaceCol(A, 6, b)) / det(A)', True);

  Test('delete(A); delete(C); delete(Adj); delete(x); delete(b)', success);

  // Nilpotency index

  Test('A ≔ ❨❨0, 5, 3, 1, 4❩, ❨0, 0, 2, 6, 3❩, ❨0, 0, 0, 5, 3❩, ❨0, 0, 0, 0, 7❩, ❨0, 0, 0, 0, 0❩❩;', null);
  Test('NilpotencyIndex(A)', 5);

  Test('A ≔ ❨❨0, 6, -96, 90, -30, -54, 44, -73, -91, -53❩, ❨0, 0, 84, 63, -10, 10, -19, 54, 54, -17❩, '+'❨0, 0, 0, -98, -38, 87, -86, -30, -29, 10❩, ❨0, 0, 0, 0, 15, -35, -18, 95, 14, 35❩, '+'❨0, 0, 0, 0, 0, 58, 14, -61, -32, -15❩, ❨0, 0, 0, 0, 0, 0, -88, -86, -69, 35❩, '+'❨0, 0, 0, 0, 0, 0, 0, -87, -33, -42❩, ❨0, 0, 0, 0, 0, 0, 0, 0, 90, -85❩, '+'❨0, 0, 0, 0, 0, 0, 0, 0, 0, -52❩, ❨0, 0, 0, 0, 0, 0, 0, 0, 0, 0❩❩;', null);
  Test('NilpotencyIndex(A)', 10);

  Test('A ≔ ReplaceDiagonal(ToUpperTriangular(RandomIntMatrix(50, 1, 100)), ZeroVector(50));', null);
  Test('NilpotencyIndex(A)', 50);

  Test('A ≔ ❨❨5, −3, 2❩, ❨15, −9, 6❩, ❨10, −6, 4❩❩;', null);
  Test('NilpotencyIndex(A)', 2);

  Test('A ≔ ❨❨−4, 2, 5❩, ❨8, −4, −6❩, ❨−8, 4, 8❩❩;', null);
  Test('NilpotencyIndex(A)', 3);

  Test('A ≔ ❨❨−2, 1, 3, 2❩, ❨−2, 0, 4, −2❩, ❨−4, 2, 4, 4❩, ❨2, −1, −3, −2❩❩;', null);
  Test('NilpotencyIndex(A)', 4);

  Test('A ≔ ❨❨2, 2, 2, 2, −4❩, ❨7, 1, 1, 1, −5❩, ❨1, 7, 1, 1, −5❩, ❨1, 1, 7, 1, −5❩, ❨1, 1, 1, 7, −5❩❩;', null);
  Test('NilpotencyIndex(A)', 5);

  Test('A ≔ ❨❨2, -3, -2, -1, 2❩, ❨2, 0, -2, 2, 2❩, ❨-3, -2, -1, -2, 1❩, ❨-1, -2, 1, -3, -1❩, ❨-3, -2, -1, -3, 2❩❩;', null);
  Test('NilpotencyIndex(A)', 5);

  Test('A ≔ ❨❨0, -4, -1, 1❩, ❨2, -4, 7, -2❩, ❨5, 5, -3, 9❩, ❨5, -2, 5, -5❩❩;', null);
  Test('NilpotencyIndex(A)', -1);

  Test('A ≔ ❨❨1, 2, -8, -5, 8❩, ❨2, -1, 1, 2, 4❩, ❨8, 7, -6, -9, -5❩, ❨5, -7, -10, 1, 6❩, ❨1, -5, -1, 4, 4❩❩;', null);
  Test('NilpotencyIndex(A)', -1);

  Test('A ≔ ❨❨5, 4, 7, 8, 2❩, ❨4, 6, 5, 5, 2❩❩;', null);
  Test('NilpotencyIndex(A)', failure, 'Cannot find nilpotency index of non-square matrix.');

  for _i := 1 to 50 do
    Test(Format('NilpotencyIndex(BackwardShiftMatrix(%d))', [_i]), _i);

  _j := RandSeed;
  RandSeed := 6411481; // Never needed in practice, of course.
  try
    StartStopwatch;
    for _i := 1 to 100 do
      Test(Format('NilpotencyIndex(RandomMatrix(%d))', [_i]), -1);
    StopStopwatch;
    TestDuration(0.01);
  finally
    RandSeed := _j;
  end;

  Test('A ≔ ❨❨0, 5, 3, i, 4❩, ❨0, 0, 2, 6, 3❩, ❨0, 0, 0, 5, 3❩, ❨0, 0, 0, 0, 7❩, ❨0, 0, 0, 0, 0❩❩;', null);
  Test('NilpotencyIndex(A)', 5);

  Test('A ≔ ❨❨0, 6, -96, i, -30, -54, 44, -73, -91, -53❩, ❨0, 0, 84, 63, -10, i, -19, 54, 54, -17❩, '+'❨0, 0, 0, -98, -38, i, -86, -30, -29, 10❩, ❨0, 0, 0, 0, 15, -35, -18, i, 14, 35❩, '+'❨0, 0, 0, 0, 0, 58, 14, -61, -32, -15❩, ❨0, 0, 0, 0, 0, 0, -88, -86, -69, 35❩, '+'❨0, 0, 0, 0, 0, 0, 0, -87, -i, -42❩, ❨0, 0, 0, 0, 0, 0, 0, 0, 90, -85❩, '+'❨0, 0, 0, 0, 0, 0, 0, 0, 0, -52❩, ❨0, 0, 0, 0, 0, 0, 0, 0, 0, 0❩❩;', null);
  Test('NilpotencyIndex(A)', 10);

  Test('A ≔ ReplaceDiagonal(ToUpperTriangular(RandomIntMatrix(50, 1, 100) + RandomIntMatrix(50, 1, 100)⋅i), ZeroVector(50));', null);
  Test('NilpotencyIndex(A)', 50);

  Test('A ≔ ❨❨1+i, −3−2⋅i, −1−3⋅i❩, ❨1+i, 2⋅i, −3+i❩, ❨−3+i, −1+2⋅i, −1−3⋅i❩❩;', null);
  Test('NilpotencyIndex(A)', 3);

  for _i := 1 to 50 do
    Test(Format('NilpotencyIndex(ComplexMatrix(BackwardShiftMatrix(%d)))', [_i]), _i);

  _j := RandSeed;
  RandSeed := 6411481; // Never needed in practice, of course.
  try
    StartStopwatch;
    for _i := 1 to 100 do
      Test(Format('NilpotencyIndex(RandomMatrix(%d) + RandomMatrix(%d)⋅i)', [_i, _i]), -1);
    StopStopwatch;
    TestDuration(0.05);
  finally
    RandSeed := _j;
  end;

  Test('D ≔ diag(0.1, 0.1, 0.1); NilpotencyIndex(D)', -1);
  Test('D ≔ diag(−0.1, −0.1, 0.2); NilpotencyIndex(D)', -1);

  Test('A ≔ ❨❨0, 3, 6❩, ❨0.0001, 0, 1❩, ❨0, 0, 0❩❩; NilpotencyIndex(A)', -1);
  Test('A ≔ ❨❨0, 3, 6❩, ❨0.0001, 0, 1❩, ❨0, 0, 0❩❩; NilpotencyIndex(A, 1E-4)', -1);
  Test('A ≔ ❨❨0, 3, 6❩, ❨0.0001, 0, 1❩, ❨0, 0, 0❩❩; NilpotencyIndex(A, 1E-3)', -1);
  Test('A ≔ ❨❨0, 3, 6❩, ❨0.0001, 0, 1❩, ❨0, 0, 0❩❩; NilpotencyIndex(A, 1E-2)', 3);

  Test('D ≔ diag(i/10, i/10, i/10); NilpotencyIndex(D)', -1);
  Test('D ≔ diag(−i/10, −i/10, i/5); NilpotencyIndex(D)', -1);

  Test('A ≔ ❨❨0, 3, i❩, ❨0.0001, 0, 1❩, ❨0, 0, 0❩❩; NilpotencyIndex(A)', -1);
  Test('A ≔ ❨❨0, 3, i❩, ❨0.0001, 0, 1❩, ❨0, 0, 0❩❩; NilpotencyIndex(A, 1E-4)', -1);
  Test('A ≔ ❨❨0, 3, i❩, ❨0.0001, 0, 1❩, ❨0, 0, 0❩❩; NilpotencyIndex(A, 1E-3)', 3);
  Test('A ≔ ❨❨0, 3, i❩, ❨0.0001, 0, 1❩, ❨0, 0, 0❩❩; NilpotencyIndex(A, 1E-2)', 3);

  Test('delete(A); delete(D)', success);

  // Pivot pos

  Test('A ≔ ❨❨6, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('PivotPos(A, 1)', 1);
  Test('PivotPos(A, 2)', 2);
  Test('PivotPos(A, 3)', 4);
  Test('PivotPos(A, 4)', 0);
  Test('PivotPos(A, 5)', 3);
  Test('PivotPos(A, 6)', 1);
  Test('PivotPos(A, 7)', 5);
  Test('PivotPos(A, 8)', failure, 'The specified row does not exist.');
  Test('PivotPos(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('PivotPos(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('A ≔ ❨❨6, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨0.00001, 0.0001, 0.001, 0.01, 0.1❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('PivotPos(A, 6)', 1);
  Test('PivotPos(A, 6, 2E-6)', 1);
  Test('PivotPos(A, 6, 2E-5)', 2);
  Test('PivotPos(A, 6, 2E-4)', 3);
  Test('PivotPos(A, 6, 2E-3)', 4);
  Test('PivotPos(A, 6, 2E-2)', 5);
  Test('PivotPos(A, 6, 2E-1)', 0);

  Test('A ≔ ❨❨i, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('PivotPos(A, 1)', 1);
  Test('PivotPos(A, 2)', 2);
  Test('PivotPos(A, 3)', 4);
  Test('PivotPos(A, 4)', 0);
  Test('PivotPos(A, 5)', 3);
  Test('PivotPos(A, 6)', 1);
  Test('PivotPos(A, 7)', 5);
  Test('PivotPos(A, 8)', failure, 'The specified row does not exist.');
  Test('PivotPos(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('PivotPos(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('A ≔ ❨❨i, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨0.00001, 0.0001, 0.001, 0.01, 0.1❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('PivotPos(A, 6)', 1);
  Test('PivotPos(A, 6, 2E-6)', 1);
  Test('PivotPos(A, 6, 2E-5)', 2);
  Test('PivotPos(A, 6, 2E-4)', 3);
  Test('PivotPos(A, 6, 2E-3)', 4);
  Test('PivotPos(A, 6, 2E-2)', 5);
  Test('PivotPos(A, 6, 2E-1)', 0);

  // Zero row

  Test('A ≔ ❨❨6, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('IsZeroRow(A, 1)', False);
  Test('IsZeroRow(A, 2)', False);
  Test('IsZeroRow(A, 3)', False);
  Test('IsZeroRow(A, 4)', True);
  Test('IsZeroRow(A, 5)', False);
  Test('IsZeroRow(A, 6)', False);
  Test('IsZeroRow(A, 7)', False);
  Test('IsZeroRow(A, 8)', failure, 'The specified row does not exist.');
  Test('IsZeroRow(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('IsZeroRow(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('A ≔ ❨❨6, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0.0001, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('IsZeroRow(A, 4)', False);
  Test('IsZeroRow(A, 4, 2E-6)', False);
  Test('IsZeroRow(A, 4, 2E-5)', False);
  Test('IsZeroRow(A, 4, 2E-4)', True);
  Test('IsZeroRow(A, 4, 2E-3)', True);
  Test('IsZeroRow(A, 4, 2E-2)', True);

  Test('A ≔ ❨❨i, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('IsZeroRow(A, 1)', False);
  Test('IsZeroRow(A, 2)', False);
  Test('IsZeroRow(A, 3)', False);
  Test('IsZeroRow(A, 4)', True);
  Test('IsZeroRow(A, 5)', False);
  Test('IsZeroRow(A, 6)', False);
  Test('IsZeroRow(A, 7)', False);
  Test('IsZeroRow(A, 8)', failure, 'The specified row does not exist.');
  Test('IsZeroRow(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('IsZeroRow(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('A ≔ ❨❨i, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0.0001, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('IsZeroRow(A, 4)', False);
  Test('IsZeroRow(A, 4, 2E-6)', False);
  Test('IsZeroRow(A, 4, 2E-5)', False);
  Test('IsZeroRow(A, 4, 2E-4)', True);
  Test('IsZeroRow(A, 4, 2E-3)', True);
  Test('IsZeroRow(A, 4, 2E-2)', True);

  Test('A ≔ ❨❨6, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('IsEssentiallyZeroRow(A, 1)', False);
  Test('IsEssentiallyZeroRow(A, 2)', False);
  Test('IsEssentiallyZeroRow(A, 3)', False);
  Test('IsEssentiallyZeroRow(A, 4)', True);
  Test('IsEssentiallyZeroRow(A, 5)', False);
  Test('IsEssentiallyZeroRow(A, 6)', False);
  Test('IsEssentiallyZeroRow(A, 7)', True);
  Test('IsEssentiallyZeroRow(A, 8)', failure, 'The specified row does not exist.');
  Test('IsEssentiallyZeroRow(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('IsEssentiallyZeroRow(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('A ≔ ❨❨6, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0.0001, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, 5❩❩;', null);

  Test('IsEssentiallyZeroRow(A, 4)', False);
  Test('IsEssentiallyZeroRow(A, 4, 2E-6)', False);
  Test('IsEssentiallyZeroRow(A, 4, 2E-5)', False);
  Test('IsEssentiallyZeroRow(A, 4, 2E-4)', True);
  Test('IsEssentiallyZeroRow(A, 4, 2E-3)', True);
  Test('IsEssentiallyZeroRow(A, 4, 2E-2)', True);

  Test('A ≔ ❨❨i, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, i❩❩;', null);

  Test('IsEssentiallyZeroRow(A, 1)', False);
  Test('IsEssentiallyZeroRow(A, 2)', False);
  Test('IsEssentiallyZeroRow(A, 3)', False);
  Test('IsEssentiallyZeroRow(A, 4)', True);
  Test('IsEssentiallyZeroRow(A, 5)', False);
  Test('IsEssentiallyZeroRow(A, 6)', False);
  Test('IsEssentiallyZeroRow(A, 7)', True);
  Test('IsEssentiallyZeroRow(A, 8)', failure, 'The specified row does not exist.');
  Test('IsEssentiallyZeroRow(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('IsEssentiallyZeroRow(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('A ≔ ❨❨i, 5, 3, 2, 5❩, ❨0, 4, 5, 3, 1❩, ❨0, 0, 0, 5, 4❩, ❨0, 0, 0.0001, 0, 0❩, ❨0, 0, 5, 4, 8❩, ❨6, 5, 2, 3, 6❩, ❨0, 0, 0, 0, i❩❩;', null);

  Test('IsEssentiallyZeroRow(A, 4)', False);
  Test('IsEssentiallyZeroRow(A, 4, 2E-6)', False);
  Test('IsEssentiallyZeroRow(A, 4, 2E-5)', False);
  Test('IsEssentiallyZeroRow(A, 4, 2E-4)', True);
  Test('IsEssentiallyZeroRow(A, 4, 2E-3)', True);
  Test('IsEssentiallyZeroRow(A, 4, 2E-2)', True);

  Test('delete(A)', success);

  // Commuting matrices

  Test('A ≔ ❨❨6, 3, 2❩, ❨4, 2, 0❩, ❨9, 0, 1❩❩; B ≔ ❨❨3, 0, 5❩, ❨4, 5, 2❩, ❨4, 5, 1❩❩; C ≔ diag(4, 4, 4); D ≔ diag(5, 2, 3); E ≔ diag(2, 7, 4);', null);

  Test('commute(A, B)', False);
  Test('commute(B, A)', False);

  Test('commute(A, C)', True);
  Test('commute(C, A)', True);

  Test('commute(A, D)', False);
  Test('commute(D, A)', False);

  Test('commute(B, C)', True);
  Test('commute(C, B)', True);

  Test('commute(B, D)', False);
  Test('commute(D, B)', False);

  Test('commute(C, D)', True);
  Test('commute(D, C)', True);

  Test('commute(D, E)', True);
  Test('commute(E, D)', True);

  Test('A ≔ ❨❨6, 3, 2❩, ❨4, 2, i❩, ❨9, 0, 1❩❩; B ≔ ❨❨3, 0, 5❩, ❨4, 5, 2❩, ❨4, 5, 1❩❩; C ≔ diag(4, 4, 4); D ≔ diag(5, 2, 3); E ≔ diag(2, 7, 4);', null);

  Test('commute(A, B)', False);
  Test('commute(B, A)', False);

  Test('commute(A, C)', True);
  Test('commute(C, A)', True);

  Test('commute(A, D)', False);
  Test('commute(D, A)', False);

  Test('commute(B, C)', True);
  Test('commute(C, B)', True);

  Test('commute(B, D)', False);
  Test('commute(D, B)', False);

  Test('commute(C, D)', True);
  Test('commute(D, C)', True);

  Test('commute(D, E)', True);
  Test('commute(E, D)', True);

  Test('A ≔ ❨❨6, 3, 2❩, ❨4, 2, 0❩, ❨9, 0, 1❩❩; B ≔ ❨❨3, 0, 5❩, ❨4, 5, 2❩, ❨4, 5, 1❩❩; C ≔ diag(i, i, i); D ≔ diag(5, 2, 3); E ≔ diag(2, 7, 4);', null);

  Test('commute(A, B)', False);
  Test('commute(B, A)', False);

  Test('commute(A, C)', True);
  Test('commute(C, A)', True);

  Test('commute(A, D)', False);
  Test('commute(D, A)', False);

  Test('commute(B, C)', True);
  Test('commute(C, B)', True);

  Test('commute(B, D)', False);
  Test('commute(D, B)', False);

  Test('commute(C, D)', True);
  Test('commute(D, C)', True);

  Test('commute(D, E)', True);
  Test('commute(E, D)', True);

  Test('A ≔ ❨❨6, 3, 2❩, ❨4, 2, i❩, ❨9, 0, 1❩❩; B ≔ ❨❨3, 0, i❩, ❨4, 5, 2❩, ❨4, 5, 1❩❩; C ≔ diag(1+i, 1+i, 1+i); D ≔ diag(5, i/2, 3); E ≔ diag(2, 7, 3 + i);', null);

  Test('commute(A, B)', False);
  Test('commute(B, A)', False);

  Test('commute(A, C)', True);
  Test('commute(C, A)', True);

  Test('commute(A, D)', False);
  Test('commute(D, A)', False);

  Test('commute(B, C)', True);
  Test('commute(C, B)', True);

  Test('commute(B, D)', False);
  Test('commute(D, B)', False);

  Test('commute(C, D)', True);
  Test('commute(D, C)', True);

  Test('commute(D, E)', True);
  Test('commute(E, D)', True);

  Test('G ≔ ❨❨4, 1, 3❩, ❨6, 7, 7❩❩; K ≔ ❨❨4, 5, 2, 6, 7, 3, 3, 5, 6❩❩*;', null);
  Test('commute(G, K)', failure, 'When multiplying two matrices, ');

  Test('G ≔ ❨❨4, i, 3❩, ❨6, 7, 7❩❩; K ≔ ❨❨4, 5, 2, 6, 7, 3, 3, 5, 6❩❩*;', null);
  Test('commute(G, K)', failure, 'When multiplying two matrices, ');

  Test('G ≔ ❨❨4, 1, 3❩, ❨6, 7, 7❩❩; K ≔ ❨❨4, 5, i, 6, 7, 3, 3, 5, 6❩❩*;', null);
  Test('commute(G, K)', failure, 'When multiplying two matrices, ');

  Test('G ≔ ❨❨4, i, 3❩, ❨6, 7, 7❩❩; K ≔ ❨❨4, 5, i, 6, 7, 3, 3, 5, 6❩❩*;', null);
  Test('commute(G, K)', failure, 'When multiplying two matrices, ');

  Test('delete(A); delete(B); delete(C); delete(D); delete(E); delete(G); delete(K)', success);

  // Matrix zeroing functions

  Test('A ≔ ❨❨6, 3, 2, 5, 4❩, ❨7, 1, 4, 2, 6❩, ❨3, 9, 5, 6, 8❩, ❨4, 1, 2, 5, 8❩, ❨6, 5, 2, 3, 1❩❩;', null);
  Test('B ≔ ❨❨6, 3, 2, 5, 6, 8, 7, 9, 5, 6❩, ❨4, 5, 2, 3, 1, 4, 5, 6, 8, 9❩, ❨2, 3, 1, 4, 5, 2, 3, 6, 7, 8❩, ❨4, 5, 2, 3, 6, 1, 2, 3, 5, 9❩❩;', null);
  Test('C ≔ B*;', null);

  Test('ToLowerTriangular(A)', 5, [6, 0, 0, 0, 0, 7, 1, 0, 0, 0, 3, 9, 5, 0, 0, 4, 1, 2, 5, 0, 6, 5, 2, 3, 1]);
  Test('IsLowerTriangular(ToLowerTriangular(A))', True);

  Test('ToLowerTriangular(B)', 10, [6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 5, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 1, 0, 0, 0, 0, 0, 0, 0, 4, 5, 2, 3, 0, 0, 0, 0, 0, 0]);
  Test('IsLowerTriangular(ToLowerTriangular(B))', True);

  Test('ToLowerTriangular(C)', 4, [6, 0, 0, 0, 3, 5, 0, 0, 2, 2, 1, 0, 5, 3, 4, 3, 6, 1, 5, 6, 8, 4, 2, 1, 7, 5, 3, 2, 9, 6, 6, 3, 5, 8, 7, 5, 6, 9, 8, 9]);
  Test('IsLowerTriangular(ToLowerTriangular(C))', True);

  Test('ToUpperTriangular(A)', 5, [6, 3, 2, 5, 4, 0, 1, 4, 2, 6, 0, 0, 5, 6, 8, 0, 0, 0, 5, 8, 0, 0, 0, 0, 1]);
  Test('IsUpperTriangular(ToUpperTriangular(A))', True);

  Test('ToUpperTriangular(B)', 10, [6, 3, 2, 5, 6, 8, 7, 9, 5, 6, 0, 5, 2, 3, 1, 4, 5, 6, 8, 9, 0, 0, 1, 4, 5, 2, 3, 6, 7, 8, 0, 0, 0, 3, 6, 1, 2, 3, 5, 9]);
  Test('IsUpperTriangular(ToUpperTriangular(B))', True);

  Test('ToUpperTriangular(C)', 4, [6, 4, 2, 4, 0, 5, 3, 5, 0, 0, 1, 2, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]);
  Test('IsUpperTriangular(ToUpperTriangular(C))', True);

  Test('ToUpperHessenberg(A)', 5, [6, 3, 2, 5, 4, 7, 1, 4, 2, 6, 0, 9, 5, 6, 8, 0, 0, 2, 5, 8, 0, 0, 0, 3, 1]);
  Test('IsUpperHessenberg(ToUpperHessenberg(A))', True);

  Test('ToUpperHessenberg(B)', 10, [6, 3, 2, 5, 6, 8, 7, 9, 5, 6, 4, 5, 2, 3, 1, 4, 5, 6, 8, 9, 0, 3, 1, 4, 5, 2, 3, 6, 7, 8, 0, 0, 2, 3, 6, 1, 2, 3, 5, 9]);
  Test('IsUpperHessenberg(ToUpperHessenberg(B))', True);

  Test('ToUpperHessenberg(C)', 4, [6, 4, 2, 4, 3, 5, 3, 5, 0, 2, 1, 2, 0, 0, 4, 3, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]);
  Test('IsUpperHessenberg(ToUpperHessenberg(C))', True);

  Test('A ≔ ❨❨i, 3, 2, 5, 4❩, ❨7, 1, 4, 2, 6❩, ❨3, 9, 5, 6, 8❩, ❨4, 1, 2, 5, 8❩, ❨6, 5, 2, 3, 1❩❩;', null);
  Test('B ≔ ❨❨i, 3, 2, 5, 6, 8, 7, 9, 5, 6❩, ❨4, 5, 2, 3, 1, 4, 5, 6, 8, 9❩, ❨2, 3, 1, 4, 5, 2, 3, 6, 7, 8❩, ❨4, 5, 2, 3, 6, 1, 2, 3, 5, 9❩❩;', null);
  Test('C ≔ B*;', null);

  Test('ToLowerTriangular(A)', 5, [ImaginaryUnit, 0, 0, 0, 0, 7, 1, 0, 0, 0, 3, 9, 5, 0, 0, 4, 1, 2, 5, 0, 6, 5, 2, 3, 1]);
  Test('IsLowerTriangular(ToLowerTriangular(A))', True);

  Test('ToLowerTriangular(B)', 10, [ImaginaryUnit, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 5, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 1, 0, 0, 0, 0, 0, 0, 0, 4, 5, 2, 3, 0, 0, 0, 0, 0, 0]);
  Test('IsLowerTriangular(ToLowerTriangular(B))', True);

  Test('ToLowerTriangular(C)', 4, [-ImaginaryUnit, 0, 0, 0, 3, 5, 0, 0, 2, 2, 1, 0, 5, 3, 4, 3, 6, 1, 5, 6, 8, 4, 2, 1, 7, 5, 3, 2, 9, 6, 6, 3, 5, 8, 7, 5, 6, 9, 8, 9]);
  Test('IsLowerTriangular(ToLowerTriangular(C))', True);

  Test('ToUpperTriangular(A)', 5, [ImaginaryUnit, 3, 2, 5, 4, 0, 1, 4, 2, 6, 0, 0, 5, 6, 8, 0, 0, 0, 5, 8, 0, 0, 0, 0, 1]);
  Test('IsUpperTriangular(ToUpperTriangular(A))', True);

  Test('ToUpperTriangular(B)', 10, [ImaginaryUnit, 3, 2, 5, 6, 8, 7, 9, 5, 6, 0, 5, 2, 3, 1, 4, 5, 6, 8, 9, 0, 0, 1, 4, 5, 2, 3, 6, 7, 8, 0, 0, 0, 3, 6, 1, 2, 3, 5, 9]);
  Test('IsUpperTriangular(ToUpperTriangular(B))', True);

  Test('ToUpperTriangular(C)', 4, [-ImaginaryUnit, 4, 2, 4, 0, 5, 3, 5, 0, 0, 1, 2, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]);
  Test('IsUpperTriangular(ToUpperTriangular(C))', True);

  Test('delete(A); delete(B); delete(C)', success);

  // Deleted absolute row sum

  Test('A ≔ ❨❨-6, -3, 2, 5, 4❩, ❨-7, 1, 4, 2, 6❩, ❨3, 9, 5, 6, 8❩, ❨4, 1, -2, -5, -8❩, ❨6, -5, 2, -3, -1❩❩;', null);
  Test('B ≔ ❨❨6, 3, -2, 5, 6, 8, -7, 9, 5, 6❩, ❨4, -5, 2, 3, 1, -4, 5, 6, 8, 9❩, ❨-2, 3, 1, -4, 5, 2, 3, 6, 7, -8❩, ❨4, -5, -2, -3, -6, 1, 2, 3, 5, 9❩❩;', null);
  Test('C ≔ B*;', null);

  Test('DeletedAbsoluteRowSum(A, 1)', 14.0);
  Test('DeletedAbsoluteRowSum(A, 2)', 19.0);
  Test('DeletedAbsoluteRowSum(A, 3)', 26.0);
  Test('DeletedAbsoluteRowSum(A, 4)', 15.0);
  Test('DeletedAbsoluteRowSum(A, 5)', 16.0);
  Test('DeletedAbsoluteRowSum(A, 6)', failure, 'Index 6 out of bounds.');
  Test('DeletedAbsoluteRowSum(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('DeletedAbsoluteRowSum(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('DeletedAbsoluteRowSum(B, 1)', 51.0);
  Test('DeletedAbsoluteRowSum(B, 2)', 42.0);
  Test('DeletedAbsoluteRowSum(B, 3)', 40.0);
  Test('DeletedAbsoluteRowSum(B, 4)', 37.0);
  Test('DeletedAbsoluteRowSum(B, 5)', failure, 'Index 5 out of bounds.');
  Test('DeletedAbsoluteRowSum(B, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('DeletedAbsoluteRowSum(B, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('DeletedAbsoluteRowSum(C, 1)', 10.0);
  Test('DeletedAbsoluteRowSum(C, 2)', 11.0);
  Test('DeletedAbsoluteRowSum(C, 3)', 6.0);
  Test('DeletedAbsoluteRowSum(C, 4)', 12.0);
  Test('DeletedAbsoluteRowSum(C, 5)', 18.0);
  Test('DeletedAbsoluteRowSum(C, 6)', 15.0);
  Test('DeletedAbsoluteRowSum(C, 7)', 17.0);
  Test('DeletedAbsoluteRowSum(C, 8)', 24.0);
  Test('DeletedAbsoluteRowSum(C, 9)', 25.0);
  Test('DeletedAbsoluteRowSum(C, 10)', 32.0);
  Test('DeletedAbsoluteRowSum(C, 11)', failure, 'Index 11 out of bounds.');
  Test('DeletedAbsoluteRowSum(C, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('DeletedAbsoluteRowSum(C, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('A ≔ ❨❨-i, -3, 2, 5, 4❩, ❨-7, i, 4, 2, 6❩, ❨3, 9, 5, 6, 8❩, ❨4, -i, -2, 3 + 4⋅i, -8❩, ❨6, -5, 2, -3, -1❩❩;', null);
  Test('B ≔ ❨❨i, 3, -2, 5, 6, 8, -7, 9, 5, 6❩, ❨4, -3 + 4⋅i, 2, 3, i, -4, 3 - 4⋅i, 6, 8, 9❩, ❨-2, 3, i, -4, 5, 2, 3, 6, 7, -8❩, ❨4, -5, -2, -3, -6, i, 2, 3, 3 + 4⋅i, 9❩❩;', null);
  Test('C ≔ B*;', null);

  Test('DeletedAbsoluteRowSum(A, 1)', 14.0);
  Test('DeletedAbsoluteRowSum(A, 2)', 19.0);
  Test('DeletedAbsoluteRowSum(A, 3)', 26.0);
  Test('DeletedAbsoluteRowSum(A, 4)', 15.0);
  Test('DeletedAbsoluteRowSum(A, 5)', 16.0);
  Test('DeletedAbsoluteRowSum(A, 6)', failure, 'Index 6 out of bounds.');
  Test('DeletedAbsoluteRowSum(A, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('DeletedAbsoluteRowSum(A, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('DeletedAbsoluteRowSum(B, 1)', 51.0);
  Test('DeletedAbsoluteRowSum(B, 2)', 42.0);
  Test('DeletedAbsoluteRowSum(B, 3)', 40.0);
  Test('DeletedAbsoluteRowSum(B, 4)', 37.0);
  Test('DeletedAbsoluteRowSum(B, 5)', failure, 'Index 5 out of bounds.');
  Test('DeletedAbsoluteRowSum(B, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('DeletedAbsoluteRowSum(B, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('DeletedAbsoluteRowSum(C, 1)', 10.0);
  Test('DeletedAbsoluteRowSum(C, 2)', 11.0);
  Test('DeletedAbsoluteRowSum(C, 3)', 6.0);
  Test('DeletedAbsoluteRowSum(C, 4)', 12.0);
  Test('DeletedAbsoluteRowSum(C, 5)', 18.0);
  Test('DeletedAbsoluteRowSum(C, 6)', 15.0);
  Test('DeletedAbsoluteRowSum(C, 7)', 17.0);
  Test('DeletedAbsoluteRowSum(C, 8)', 24.0);
  Test('DeletedAbsoluteRowSum(C, 9)', 25.0);
  Test('DeletedAbsoluteRowSum(C, 10)', 32.0);
  Test('DeletedAbsoluteRowSum(C, 11)', failure, 'Index 11 out of bounds.');
  Test('DeletedAbsoluteRowSum(C, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('DeletedAbsoluteRowSum(C, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('delete(A); delete(B); delete(C)', success);

  // Zero row counts

  Test('A ≔ ❨❨2, 3, 6, 5❩, ❨5, 1, 2, 5❩, ❨0, 0, 0, 0❩, ❨6, 3, 2, 5❩, ❨7, 4, 1, 0❩, '+
    '❨2, 0, 1, 0❩, ❨0, 0, 0, 1❩, ❨2, 0, 0, 0❩, ❨3, 6, 3, 0❩, ❨0, 0, 0, 0❩, ❨5, 2, 1, 0❩, '+
    '❨0, 0, 0, 0.0001❩, ❨0, 4, 2, 0❩, ❨0, 0, 0, 0.01❩, ❨0, 0.000001, 0.00000001, 0❩, '+
    '❨0.001, 0.0001, 0, 0❩, ❨4, 0, 1, 7❩, ❨6, 3, 2, 1❩, ❨0, 0, 0, 0.01❩, ❨0, 0.00001, 0.0001, 0❩, '+
    '❨0, 0, 0, 0❩, ❨0, 0, 0, 0❩, ❨0, 0, 0, 0❩❩;', null);

  Test('ZeroRowCount(A)', 5);
  Test('ZeroRowCount(A, 2E-9)', 5);
  Test('ZeroRowCount(A, 2E-8)', 5);
  Test('ZeroRowCount(A, 2E-7)', 5);
  Test('ZeroRowCount(A, 2E-6)', 6);
  Test('ZeroRowCount(A, 2E-5)', 6);
  Test('ZeroRowCount(A, 2E-4)', 8);
  Test('ZeroRowCount(A, 2E-3)', 9);
  Test('ZeroRowCount(A, 2E-2)', 11);
  Test('ZeroRowCount(A, 2E-1)', 11);

  Test('TrailingZeroRowCount(A)', 3);
  Test('TrailingZeroRowCount(A, 2E-9)', 3);
  Test('TrailingZeroRowCount(A, 2E-8)', 3);
  Test('TrailingZeroRowCount(A, 2E-7)', 3);
  Test('TrailingZeroRowCount(A, 2E-6)', 3);
  Test('TrailingZeroRowCount(A, 2E-5)', 3);
  Test('TrailingZeroRowCount(A, 2E-4)', 4);
  Test('TrailingZeroRowCount(A, 2E-3)', 4);
  Test('TrailingZeroRowCount(A, 2E-2)', 5);
  Test('TrailingZeroRowCount(A, 2E-1)', 5);

  Test('A ≔ ❨❨2, 3, 6, 5❩, ❨5, i, 2, 5❩, ❨0, 0, 0, 0❩, ❨6, 3, 2, 5❩, ❨7, 4, 1, 0❩, '+
    '❨2, 0, 1 + i, 0❩, ❨0, 0, 0, i❩, ❨2, 0, 0, 0❩, ❨3, 6, 3, 0❩, ❨0, 0, 0, 0❩, ❨i/10000, i/2, i, 0❩, '+
    '❨0, 0, 0, 0.0001❩, ❨0, 4, 2, 0❩, ❨0, 0, 0, 0.01❩, ❨0, 0.000001⋅i, 0.00000001⋅i, 0❩, '+
    '❨0.001, 0.0001⋅i, 0, 0❩, ❨4, 0, 1, 7❩, ❨6, 3, 2, 1❩, ❨0, 0, 0, 0.01⋅i❩, ❨0, 0.00001, 0.0001⋅i, 0❩, '+
    '❨0, 0, 0, 0❩, ❨0, 0, 0, 0❩, ❨0, 0, 0, 0❩❩;', null);

  Test('ZeroRowCount(A)', 5);
  Test('ZeroRowCount(A, 2E-9)', 5);
  Test('ZeroRowCount(A, 2E-8)', 5);
  Test('ZeroRowCount(A, 2E-7)', 5);
  Test('ZeroRowCount(A, 2E-6)', 6);
  Test('ZeroRowCount(A, 2E-5)', 6);
  Test('ZeroRowCount(A, 2E-4)', 8);
  Test('ZeroRowCount(A, 2E-3)', 9);
  Test('ZeroRowCount(A, 2E-2)', 11);
  Test('ZeroRowCount(A, 2E-1)', 11);

  Test('TrailingZeroRowCount(A)', 3);
  Test('TrailingZeroRowCount(A, 2E-9)', 3);
  Test('TrailingZeroRowCount(A, 2E-8)', 3);
  Test('TrailingZeroRowCount(A, 2E-7)', 3);
  Test('TrailingZeroRowCount(A, 2E-6)', 3);
  Test('TrailingZeroRowCount(A, 2E-5)', 3);
  Test('TrailingZeroRowCount(A, 2E-4)', 4);
  Test('TrailingZeroRowCount(A, 2E-3)', 4);
  Test('TrailingZeroRowCount(A, 2E-2)', 5);
  Test('TrailingZeroRowCount(A, 2E-1)', 5);

  Test('delete(A)', success);

  for _i := 1 to 100 do
    Test(Format('ZeroRowCount(ZeroMatrix(%d))', [_i]), _i);

  for _i := 1 to 100 do
    Test(Format('TrailingZeroRowCount(ZeroMatrix(%d))', [_i]), _i);

  for _i := 1 to 100 do
    Test(Format('ZeroRowCount(ComplexZeroMatrix(%d))', [_i]), _i);

  for _i := 1 to 100 do
    Test(Format('TrailingZeroRowCount(ComplexZeroMatrix(%d))', [_i]), _i);

  // The Gram-Schmidt orthogonalization procedure

  Test('A ≔ ❨❨5, 3, 2❩, ❨1, 4, −1❩, ❨7, 2, 0❩❩;', null);
  Eps; Test('B ≔ GramSchmidt(A)', 3, [0.577350269189625765, 0.210235332491054675, 0.788966267744725799, 0.115470053837925153, 0.935547229585193303, -0.333793420968922454, 0.80829037686547607, -0.283817698862923811, -0.515862559679243792]);
  Test('IsOrthogonal(B)', True);

  Test('A ≔ ❨❨5, 3, 2, 9❩, ❨1, 4, −1, 5❩, ❨7, 2, 0, 6❩❩;', null);
  Test('rank(A)', 3);
  Eps; Test('B ≔ GramSchmidt(A)', 4, [0.577350269189625765, 0.210235332491054675, 0.788966267744725799, 0, 0.115470053837925153, 0.935547229585193303, -0.333793420968922454, 0, 0.80829037686547607, -0.283817698862923811, -0.515862559679243792, 0]);
  Test('rank(B)', 3);

  Test('A ≔ ❨❨6, 3, 9❩, ❨5, 2, 9❩, ❨6, 1, 15❩❩;', null);
  Test('rank(A)', 2);
  Eps; Test('B ≔ GramSchmidt(A)', 3, [0.609207699080171426, 0.621523709890218689, 0, 0.507673082566809522, 0.171454816521439638, 0, 0.609207699080171426, -0.764402723658085054, 0]);
  Test('rank(B)', 2);

  StartStopwatch;
  _j := RandSeed;
  RandSeed := 6230517;
  try
    for _i := 1 to 100 do
      Test('IsOrthogonal(GramSchmidt(RandomMatrix(6)))', True);
  finally
    RandSeed := _j;
  end;
  StopStopwatch;
  TestDuration(0.007);

  Test('A ≔ ❨❨2, 3, i❩, ❨i, 1, 5❩, ❨4, 7, 2❩❩;', null);
  Eps; Test('B ≔ GramSchmidt(A)', 3, [0.436435780471984763, -0.120490673177969969 + 0.0481962692711879875*ImaginaryUnit, -0.121196321612797194 + 0.882039896182024025*ImaginaryUnit, 0.218217890235992381*ImaginaryUnit, 0.481962692711879875 - 0.819336577610195788*ImaginaryUnit, 0.175061353440707058 + 0.13466257956977466*ImaginaryUnit, 0.872871560943969525, 0.265079480991533932 + 0.096392538542375975*ImaginaryUnit, 0.0269325159139549317 - 0.397254609730835248*ImaginaryUnit]);
  Test('IsUnitary(B)', True);

  Test('A ≔ ❨❨2, 3, i, 5❩, ❨i, 1, 5, -7❩, ❨4, 7, 2, 1+i❩❩;', null);
  Test('rank(A)', 3);
  Eps; Test('B ≔ GramSchmidt(A)', 4, [0.436435780471984763, -0.120490673177969969 + 0.0481962692711879875*ImaginaryUnit, -0.121196321612797194 + 0.882039896182024025*ImaginaryUnit, 0, 0.218217890235992381*ImaginaryUnit, 0.481962692711879875 - 0.819336577610195788*ImaginaryUnit, 0.175061353440707058 + 0.13466257956977466*ImaginaryUnit, 0, 0.872871560943969525, 0.265079480991533932 + 0.096392538542375975*ImaginaryUnit, 0.0269325159139549317 - 0.397254609730835248*ImaginaryUnit, 0]);
  Test('rank(B)', 3);

  StartStopwatch;
  _j := RandSeed;
  RandSeed := 6230517;
  try
    for _i := 1 to 100 do
      Test('IsUnitary(GramSchmidt(RandomMatrix(6) + RandomMatrix(6)⋅i))', True);
  finally
    RandSeed := _j;
  end;
  StopStopwatch;
  TestDuration(0.01);

  Test('GramSchmidt(❨❨5❩❩)', 1, [1]);
  Test('GramSchmidt(❨❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, -21❩❩)', 12, [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]);

  Test('delete(A); delete(B)', success);

  // Column space basis

  Test('A ≔ ❨❨6, 2, -5, 7❩, ❨3, 0, 3, 1❩, ❨2, 1, 0, -4❩, ❨5, 3, 2, 5❩❩;', null);
  Test('ColumnSpaceBasis(A)', 4, [6, 2, -5, 7, 3, 0, 3, 1, 2, 1, 0, -4, 5, 3, 2, 5]);
  Test('rank(A)', 4);

  Test('A ≔ ❨❨6, 2, -5, 7, 5, -3❩, ❨3, 0, 3, 1, 3, 5❩, ❨2, 1, 0, -4, 0, -4❩, ❨5, 3, 2, 5, 8, 8❩❩;', null);
  Test('ColumnSpaceBasis(A)', 4, [6, 2, -5, 7, 3, 0, 3, 1, 2, 1, 0, -4, 5, 3, 2, 5]);
  Test('rank(A)', 4);

  Test('A ≔ ❨❨6, 2, 6, 20, -14, 7❩, ❨3, 0, 6, 15, -9, 1❩, ❨2, 1, 1, 5, -4, -4❩, ❨5, 3, 1, 10, -9, 5❩❩;', null);
  Test('ColumnSpaceBasis(A)', 3, [6, 2, 7, 3, 0, 1, 2, 1, -4, 5, 3, 5]);
  Test('rank(A)', 3);

  Test('A ≔ ❨❨6, -12, 2, 4, 7, 5, -11, -6❩, ❨3, -6, 0, 3, 1, 3, -2, 6❩, ❨2, -4, 1, 1, -4, 0, 11, 12❩, ❨5, -10, 3, 2, 5, 8, -10, 9❩❩;', null);
  Test('ColumnSpaceBasis(A)', 4, [6, 2, 7, 5, 3, 0, 1, 3, 2, 1, -4, 0, 5, 3, 5, 8]);
  Test('rank(A)', 4);

  Test('A ≔ ❨❨7, 14, -7, 21, 5, 1, 13, 10❩, ❨1, 2, -1, 3, 3, 7, -5, -10❩, ❨-4, -8, 4, -12, 0, 8, -16, -20❩, ❨5, 10, -5, 15, 8, 14, -4, -15❩❩;', null);
  Test('ColumnSpaceBasis(A)', 2, [7, 5, 1, 3, -4, 0, 5, 8]);
  Test('rank(A)', 2);

  Test('A ≔ ❨❨7, 21, -14, 28, 35, -7, -21, 0, 7❩, ❨1, 3, -2, 4, 5, -1, -3, 0, 1❩, ❨-4, -12, 8, -16, -20, 4, 12, 0, -4❩, ❨5, 15, -10, 20, 25, -5, -15, 0, 5❩❩;', null);
  Test('ColumnSpaceBasis(A)', 1, [7, 1, -4, 5]);
  Test('rank(A)', 1);

  Test('A ≔ ❨❨21❩❩;', null);
  Test('ColumnSpaceBasis(A)', 1, [21]);
  Test('rank(A)', 1);

  Test('A ≔ ❨❨1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 12❩❩;', null);
  Test('ColumnSpaceBasis(A)', 1, [1]);
  Test('rank(A)', 1);

  Test('A ≔ IdentityMatrix(100);', null);
  Test('ColumnSpaceBasis(A) = IdentityMatrix(100)', True);
  Test('rank(A)', 100);

  Test('A ≔ ❨❨i, 2, -5, 7❩, ❨3, 0, 3, 1❩, ❨2, 1, 0, -4❩, ❨5, 3, 2, 5❩❩;', null);
  Test('ColumnSpaceBasis(A)', 4, [ImaginaryUnit, 2, -5, 7, 3, 0, 3, 1, 2, 1, 0, -4, 5, 3, 2, 5]);
  Test('rank(A)', 4);

  Test('A ≔ ❨❨i, 2, 2 + i, -2 + i, 7, 5 + i, -5 + i❩, ❨3, 0, 3, 3, 1, 4, 2❩, ❨2, 1, 3, 1, -4, -3, 7❩, ❨5, 3, 8, 2, 5, 7, 3❩❩;', null);
  Test('ColumnSpaceBasis(A)', 3, [ImaginaryUnit, 2, 7, 3, 0, 1, 2, 1, -4, 5, 3, 5]);
  Test('rank(A)', 3);

  Test('A ≔ ❨❨i/2❩❩;', null);
  Test('ColumnSpaceBasis(A)', 1, [ImaginaryUnit/2]);
  Test('rank(A)', 1);

  Test('A ≔ ❨❨i, 2, 3, 4, 5, 6, 7, 8, 10, 11, 12❩❩;', null);
  Test('ColumnSpaceBasis(A)', 1, [ImaginaryUnit]);
  Test('rank(A)', 1);

  Test('A ≔ ComplexMatrix(IdentityMatrix(100));', null);
  Test('ColumnSpaceBasis(A) = IdentityMatrix(100)', True);
  Test('rank(A)', 100);

  Test('delete(A)', success);

  // Column space projection

  Test('A ≔ ❨❨3, 2, 1❩, ❨2, 0, 7❩, ❨−3, 2, 5❩❩;', null);

  Eps; Test('ColumnSpaceProjection(A, ❨12, 30, 5❩)', [12, 30, 5]);
  Eps; Test('DistanceFromColumnSpace(A, ❨12, 30, 5❩)', 0.0);
  Eps; Test('ColumnSpaceProjection(A, ❨−3, 20, 1❩)', [-3, 20, 1]);
  Eps; Test('DistanceFromColumnSpace(A, ❨−3, 20, 1❩)', 0.0);

  Test('A ≔ ❨❨6, 3, 2❩, ❨0, −2, 4❩, ❨7, 0, 9❩, ❨6, 3, 0❩❩; v ≔ ❨5, 3, 0, 1❩;', null);
  Eps; Test('ColumnSpaceProjection(A, v)', [2.95361527967257844, -0.581173260572987722, 2.04638472032742156, 0.658935879945429741]);
  Eps; Test('DistanceFromColumnSpace(A, v)', 4.61698034470698621);

  Test('A ≔ ❨❨4, 7, 9, 0❩, ❨1, 3, -6, 6❩, ❨6, 5, 0, 1❩, ❨2, 6, 1, 0❩, ❨1, 1, 4, 2❩❩; v ≔ ❨2, 0, 1, −1, 0❩;', null);
  Eps; Test('ColumnSpaceProjection(A, v)', [1.28748228789665694, -0.330209821415635332, 1.13177746933524194, -0.267562437880864553, 0.924740729579285025]);
  Eps; Test('DistanceFromColumnSpace(A, v)', 1.42326930585556422);

  Test('A ≔ ❨❨4, 7, 9, 6❩, ❨1, 3, -6, -8❩, ❨6, 5, 0, 1❩, ❨2, 6, 1, -3❩, ❨1, 1, 4, 4❩❩; v ≔ ❨3, 5, 1, 4, 1❩;', null);
  Test('rank(A)', 3);
  Eps; Test('ColumnSpaceProjection(A, v)', [2.79078582702608652, 3.83413494401153325, 1.30851447249046145, 4.81697177584728384, -0.482308638753215764]);
  Eps; Test('DistanceFromColumnSpace(A, v)', 2.08874959287553019);

  Test('A ≔ ❨❨4, 7, -2, 1❩, ❨1, 3, -3, -1❩, ❨6, 5, 8, 7❩, ❨2, 6, -6, -2❩, ❨1, 1, 1, 1❩❩; v ≔ ❨3, 5, 1, 4, 1❩;', null);
  Test('rank(A)', 2);
  Eps; Test('ColumnSpaceProjection(A, v)', [4.07277628032345013, 2.24797843665768194, 0.698113207547169812, 4.49595687331536388, 0.28032345013477089]);
  Eps; Test('DistanceFromColumnSpace(A, v)', 3.09507913061973969);

  Test('A ≔ ❨❨4, -8, 12, -4❩, ❨1, -2, 3, -1❩, ❨6, -12, 18, -6❩, ❨2, -4, 6, -2❩, ❨1, -2, 3, -1❩❩; v ≔ ❨3, 5, 1, 4, 1❩;', null);
  Test('rank(A)', 1);
  Eps; Test('ColumnSpaceProjection(A, v)', [2.20689655172413793, 0.551724137931034483, 3.3103448275862069, 1.10344827586206897, 0.551724137931034483]);
  Eps; Test('DistanceFromColumnSpace(A, v)', 5.8604460228046548);

  Test('A ≔ ❨❨4, 7, 1, -2, 9, -4❩, ❨1, 3, -1, -3, -6, 6❩, ❨6, 5, 7, 8, 0, 13❩, ❨2, 6, -2, -6, 1, -1❩, ❨1, 1, 1, 1, 4, -2❩❩; v ≔ ❨1, 0, −2, 1, 2❩;', null);
  Test('rank(A)', 3);
  Eps; Test('ColumnSpaceProjection(A, v)', [1.38405229732920721, -0.515292743248449598, -1.89602674521208506, 0.995600462324298125, 0.364043100555534842]);
  Eps; Test('DistanceFromColumnSpace(A, v)', 1.76074062508117727);

  Test('ColumnSpaceProjection(ComplexMatrix(A), v)', failure, 'Not implemented for complex matrices.');
  Test('ColumnSpaceProjection(A, ComplexVector(v))', failure, 'Not implemented for complex matrices.');
  Test('DistanceFromColumnSpace(ComplexMatrix(A), v)', failure, 'Not implemented for complex matrices.');
  Test('DistanceFromColumnSpace(A, ComplexVector(v))', failure, 'Not implemented for complex matrices.');

  Test('A ≔ ❨❨6, 3, 2❩, ❨0, −2, 4❩, ❨7, 0, 9❩, ❨6, 3, 0❩❩; v ≔ ❨5, 3, 0❩;', null);
  Eps; Test('ColumnSpaceProjection(A, v)', failure, 'When multiplying two matrices, ');
  Eps; Test('DistanceFromColumnSpace(A, v)', failure, 'When multiplying two matrices, ');

  Test('delete(A); delete(v)', success);

  // Additional column space tests (with explicit matrix constructions)

  Test('(a, b, c, d) ≔ ❨3, 0, 1, −2, 1❩, ❨0, 2, 4, −3, 1❩, ❨6, 2, 3, −1, 2❩, ❨0, 0, 2, 3, −1❩;', null);

  Test('A ≔ MatFromCols(a, b, c, d);', null);
  Test('rank(A)', 4);
  Test('ColumnSpaceBasis(A)', 4, [3, 0, 6, 0, 0, 2, 2, 0, 1, 4, 3, 2, -2, -3, -1, 3, 1, 1, 2, -1]); // = A

  Test('A ≔ MatFromCols(a, b, c, 2⋅a − b − c);', null);
  Test('rank(A)', 3);
  Test('ColumnSpaceBasis(A)', 3, [3, 0, 6, 0, 2, 2, 1, 4, 3, -2, -3, -1, 1, 1, 2]);

  Test('A ≔ MatFromCols(a, b, a − 2⋅b, c);', null);
  Test('rank(A)', 3);
  Test('ColumnSpaceBasis(A)', 3, [3, 0, 6, 0, 2, 2, 1, 4, 3, -2, -3, -1, 1, 1, 2]);

  Test('A ≔ MatFromCols(a, b, a − 2⋅b, 3⋅a − 2⋅b);', null);
  Test('rank(A)', 2);
  Test('ColumnSpaceBasis(A)', 2, [3, 0, 0, 2, 1, 4, -2, -3, 1, 1]);

  Test('A ≔ MatFromCols(a, b, 2⋅b − a, c, a − 2⋅b + c, 2⋅c − b, d);', null);
  Test('rank(A)', 4);
  Test('ColumnSpaceBasis(A)', 4, [3, 0, 6, 0, 0, 2, 2, 0, 1, 4, 3, 2, -2, -3, -1, 3, 1, 1, 2, -1]);

  Test('A ≔ MatFromCols(a, b, 2⋅b − a, c, a − 2⋅b + c, 2⋅c − b, 2⋅b − a − c);', null);
  Test('rank(A)', 3);
  Test('ColumnSpaceBasis(A)', 3, [3, 0, 6, 0, 2, 2, 1, 4, 3, -2, -3, -1, 1, 1, 2]);

  Test('A ≔ MatFromCols(a, b, 5⋅a − 4⋅b, 3⋅b − 2⋅a, 2⋅a − b, b − a, 5⋅b − 3⋅a);', null);
  Test('rank(A)', 2);
  Test('ColumnSpaceBasis(A)', 2, [3, 0, 0, 2, 1, 4, -2, -3, 1, 1]);

  Test('A ≔ MatFromCols(a, b, 5⋅a − 4⋅b, 3⋅b − 2⋅a, 2⋅a − b, c, 5⋅b − 3⋅a);', null);
  Test('rank(A)', 3);
  Test('ColumnSpaceBasis(A)', 3, [3, 0, 6, 0, 2, 2, 1, 4, 3, -2, -3, -1, 1, 1, 2]);

  Test('A ≔ MatFromCols(a, b);', null);
  Test('rank(A)', 2);
  Test('ColumnSpaceBasis(A)', 2, [3, 0, 0, 2, 1, 4, -2, -3, 1, 1]);

  Test('A ≔ MatFromCols(a);', null);
  Test('rank(A)', 1);
  Test('ColumnSpaceBasis(A)', 1, [3, 0, 1, -2, 1]);

  Test('delete(A)', success);

  // Row operations

  Test('A ≔ ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩;', null);

  Test('RowSwap(A, 3, 3) = A', True);
  Test('RowSwap(A, 2, 4) = ❨❨3, 2, 1, 4❩, ❨8, 3, 0, 1❩, ❨7, 2, 1, 6❩, ❨6, 0, 2, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowSwap(A, 1, 2) = ❨❨6, 0, 2, 1❩, ❨3, 2, 1, 4❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowSwap(A, 1, 6) = ❨❨0, 4, 8, 2❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨3, 2, 1, 4❩❩', True);
  Test('RowSwap(A, 1, 1) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowSwap(A, 6, 6) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowSwap(RowSwap(A, 1, 3), 3, 1) = A', True);
  Test('RowSwap(RowSwap(A, 1, 3), 1, 3) = A', True);
  Test('RowSwap(RowSwap(A, 3, 1), 3, 1) = A', True);
  Test('RowSwap(RowSwap(A, 3, 1), 1, 3) = A', True);
  Test('RowSwap(A, 1, 7)', failure, 'Index 7 out of bounds.');
  Test('RowSwap(A, 1, 0)', failure, 'A positive integer was expected as argument 3, but "0" was given.');
  Test('RowSwap(A, 1, -2)', failure, 'A positive integer was expected as argument 3, but "-2" was given.');
  Test('RowSwap(A, 7, 2)', failure, 'Index 7 out of bounds.');
  Test('RowSwap(A, 0, 2)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('RowSwap(A, -2, 2)', failure, 'A positive integer was expected as argument 2, but "-2" was given.');
  Test('RowSwap(4, 1, 2)', failure, 'An object of type matrix was expected as argument 1, but an object of type integer was given.');
  Test('RowSwap(❨1, 2, 3, 4❩, 1, 2)', failure, 'An object of type matrix was expected as argument 1, but an object of type real vector was given.');

  Test('A ≔ ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩;', null);

  Test('RowSwap(A, 3, 3) = A', True);
  Test('RowSwap(A, 2, 4) = ❨❨3, 2, i, 4❩, ❨8, 3, 0, 1❩, ❨7, 2, 1, 6❩, ❨6, 0, 2, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowSwap(A, 1, 2) = ❨❨6, 0, 2, 1❩, ❨3, 2, i, 4❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowSwap(A, 1, 6) = ❨❨0, 4, 8, 2❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨3, 2, i, 4❩❩', True);
  Test('RowSwap(A, 1, 1) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowSwap(A, 6, 6) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowSwap(RowSwap(A, 1, 3), 3, 1) = A', True);
  Test('RowSwap(RowSwap(A, 1, 3), 1, 3) = A', True);
  Test('RowSwap(RowSwap(A, 3, 1), 3, 1) = A', True);
  Test('RowSwap(RowSwap(A, 3, 1), 1, 3) = A', True);
  Test('RowSwap(A, 1, 7)', failure, 'Index 7 out of bounds.');
  Test('RowSwap(A, 1, 0)', failure, 'A positive integer was expected as argument 3, but "0" was given.');
  Test('RowSwap(A, 1, -2)', failure, 'A positive integer was expected as argument 3, but "-2" was given.');
  Test('RowSwap(A, 7, 2)', failure, 'Index 7 out of bounds.');
  Test('RowSwap(A, 0, 2)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('RowSwap(A, -2, 2)', failure, 'A positive integer was expected as argument 2, but "-2" was given.');
  Test('RowSwap(i, 1, 2)', failure, 'An object of type matrix was expected as argument 1, but an object of type complex number was given.');
  Test('RowSwap(❨i, 2, 3, 4❩, 1, 2)', failure, 'An object of type matrix was expected as argument 1, but an object of type complex vector was given.');

  Test('A ≔ ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩;', null);

  Test('RowScale(A, 3, 10) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨70, 20, 10, 60❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 5, 2) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨10, 10, 6, 2❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 1, -1) = ❨❨-3, -2, -1, -4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 4, -3) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨-24, -9, 0, -3❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 6, 1/2) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 2, 4, 1❩❩', True);
  Test('RowScale(A, 1, 1/3) = ❨❨1, 2/3, 1/3, 4/3❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 1, 1) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 1, 0) = ❨❨0, 0, 0, 0❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 7, 1)', failure, 'Index 7 out of bounds.');
  Test('RowScale(A, 0, 1)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('RowScale(A, -5, 1)', failure, 'A positive integer was expected as argument 2, but "-5" was given.');
  Test('RowScale(❨4, 1, 2❩, 2, 3)', failure, 'An object of type matrix was expected as argument 1, but an object of type real vector was given.');

  Test('A ≔ ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩;', null);

  Test('RowScale(A, 3, 10) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨70, 20, 10, 60❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 5, 2) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨10, 10, 6, 2❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 1, -1) = ❨❨-3, -2, -i, -4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 4, -3) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨-24, -9, 0, -3❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 6, 1/2) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 2, 4, 1❩❩', True);
  Test('RowScale(A, 1, 1/3) = ❨❨1, 2/3, i/3, 4/3❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 1, 1) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 1, 0) = ❨❨0, 0, 0, 0❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowScale(A, 7, 1)', failure, 'Index 7 out of bounds.');
  Test('RowScale(A, 0, 1)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('RowScale(A, -5, 1)', failure, 'A positive integer was expected as argument 2, but "-5" was given.');
  Test('RowScale(❨4, i, 2❩, 2, 3)', failure, 'An object of type matrix was expected as argument 1, but an object of type complex vector was given.');

  Test('A ≔ ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩;', null);

  Test('RowAddMul(A, 3, 2, 5) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨37, 2, 11, 11❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 6, 3, -1) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨-7, 2, 7, -4❩❩', True);
  Test('RowAddMul(A, 5, 6, 2) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 13, 19, 5❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 5, 6, 0) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 3, 1, 10) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨37, 22, 11, 46❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 4, 4, 2) = ❨❨3, 2, 1, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨24, 9, 0, 3❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 7, 6, 2)', failure, 'Index 7 out of bounds.');
  Test('RowAddMul(A, 5, 7, 2)', failure, 'Index 7 out of bounds.');
  Test('RowAddMul(A, 0, 6, 2)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('RowAddMul(A, 5, -2, 2)', failure, 'A positive integer was expected as argument 3, but "-2" was given.');

  Test('A ≔ ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩;', null);

  Test('RowAddMul(A, 3, 2, 5) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨37, 2, 11, 11❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 6, 3, -1) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨-7, 2, 7, -4❩❩', True);
  Test('RowAddMul(A, 5, 6, 2) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 13, 19, 5❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 5, 6, 0) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 3, 1, 10) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨37, 22, 1 + 10⋅i, 46❩, ❨8, 3, 0, 1❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 4, 4, 2) = ❨❨3, 2, i, 4❩, ❨6, 0, 2, 1❩, ❨7, 2, 1, 6❩, ❨24, 9, 0, 3❩, ❨5, 5, 3, 1❩, ❨0, 4, 8, 2❩❩', True);
  Test('RowAddMul(A, 7, 6, 2)', failure, 'Index 7 out of bounds.');
  Test('RowAddMul(A, 5, 7, 2)', failure, 'Index 7 out of bounds.');
  Test('RowAddMul(A, 0, 6, 2)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('RowAddMul(A, 5, -2, 2)', failure, 'A positive integer was expected as argument 3, but "-2" was given.');

  // Row echelon form and reduced row echelon form

  StartStopwatch;

  _j := RandSeed;
  RandSeed := 312537;
  try
    for _i := 1 to 100 do
    begin
      Test('A ≔ RandomMatrix(10);', null);
      Test('B ≔ RowEchelonForm(A);', null);
      Test('IsRowEchelonForm(B)', True);
      Test('abs(det(A)) ≈ abs(det(B))', True);
      Test('C ≔ ReducedRowEchelonForm(A);', null);
      Test('IsReducedRowEchelonForm(C)', True);
    end;

    for _i := 1 to 100 do
    begin

      Test('A ≔ RandomMatrix(10, 11);', null);
      Test('u ≔ SysSolve(A);', null);

      Test('B ≔ RowEchelonForm(A);', null);
      Test('IsRowEchelonForm(B)', True);
      Test('v ≔ SysSolve(B);', null);
      Test('u ≈ v', True);

      Test('C ≔ ReducedRowEchelonForm(A);', null);
      Test('IsReducedRowEchelonForm(C)', True);
      Test('w ≔ SysSolve(C);', null);
      Test('u ≈ w', True);

    end;

  finally
    RandSeed := _j;
  end;

  StopStopwatch;
  TestDuration(0.1);

  StartStopwatch;

  _j := RandSeed;
  RandSeed := 312537;
  try
    for _i := 1 to 100 do
    begin
      Test('A ≔ RandomMatrix(10) + RandomMatrix(10)⋅i;', null);
      Test('B ≔ RowEchelonForm(A);', null);
      Test('IsRowEchelonForm(B)', True);
      Test('abs(det(A)) ≈ abs(det(B))', True);
      Test('C ≔ ReducedRowEchelonForm(A);', null);
      Test('IsReducedRowEchelonForm(C)', True);
    end;

    for _i := 1 to 100 do
    begin

      Test('A ≔ RandomMatrix(10, 11) + RandomMatrix(10, 11)⋅i;', null);
      Test('u ≔ SysSolve(A);', null);

      Test('B ≔ RowEchelonForm(A);', null);
      Test('IsRowEchelonForm(B)', True);
      Test('v ≔ SysSolve(B);', null);
      Test('u ≈ v', True);

      Test('C ≔ ReducedRowEchelonForm(A);', null);
      Test('IsReducedRowEchelonForm(C)', True);
      Test('w ≔ SysSolve(C);', null);
      Test('u ≈ w', True);

    end;

  finally
    RandSeed := _j;
  end;

  StopStopwatch;
  TestDuration(0.2);

  Test('delete(A); delete(B); delete(C); delete(u); delete(v); delete(w)', success);

  Test('RowEchelonForm(❨❨21❩❩)', 1, [21]);
  Test('ReducedRowEchelonForm(❨❨21❩❩)', 1, [1]);

  Test('RowEchelonForm(❨❨21⋅i❩❩)', 1, [21*ImaginaryUnit]);
  Test('ReducedRowEchelonForm(❨❨21⋅i❩❩)', 1, [TASC(1)]);

  Test('RowEchelonForm(❨❨21, 22, 23❩❩)', 3, [21, 22, 23]);
  Test('ReducedRowEchelonForm(❨❨21, 22, 23❩❩)', 3, [1, 22/21, 23/21]);

  Test('RowEchelonForm(ComplexMatrix(❨❨21, 22, 23❩❩))', 3, [TASC(21), 22, 23]);
  Test('ReducedRowEchelonForm(ComplexMatrix(❨❨21, 22, 23❩❩))', 3, [TASC(1), 22/21, 23/21]);

  Test('RowEchelonForm(❨❨21, 22, 23❩❩*)', 1, [23, 0, 0]); // the greatest pivot element is selected
  Test('ReducedRowEchelonForm(❨❨21, 22, 23❩❩*)', 1, [1, 0, 0]);

  Test('RowEchelonForm(ComplexMatrix(❨❨21, 22, 23❩❩*))', 1, [TASC(23), 0, 0]); // the greatest pivot element is selected
  Test('ReducedRowEchelonForm(ComplexMatrix(❨❨21, 22, 23❩❩*))', 1, [TASC(1), 0, 0]);


  //
  // Special matrix constructors
  //

  // Reversal matrices

  Test('ReversalMatrix(1)', 1, [1]);
  Test('ReversalMatrix(2)', 2, [0, 1, 1, 0]);
  Test('ReversalMatrix(3)', 3, [0, 0, 1, 0, 1, 0, 1, 0, 0]);
  Test('ReversalMatrix(4)', 4, [0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0]);
  Test('ReversalMatrix(5)', 5, [0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0]);

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ ReversalMatrix(%d);', [_i]), null);
    Test('IsReversal(A)', True);
    Test('IsPermutation(A)', True);
    Test('IsBinary(A)', True);
    Test('IsOrthogonal(A)', True);
    Test('IsSymmetric(A)', True);
    Test('IsNormal(A)', True);
  end;

  Test('ReversalMatrix(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('ReversalMatrix(-1)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');

  // Circulant matrices

  Test('CirculantMatrix(❨21❩)', 1, [21]);
  Test('CirculantMatrix(❨1, 2❩)', 2, [1, 2, 2, 1]);
  Test('CirculantMatrix(❨1, 2, 3❩)', 3, [1, 2, 3, 3, 1, 2, 2, 3, 1]);
  Test('CirculantMatrix(❨1, 2, 3, 4❩)', 4, [1, 2, 3, 4, 4, 1, 2, 3, 3, 4, 1, 2, 2, 3, 4, 1]);
  Test('CirculantMatrix(❨1, 2, 3, 4, 5❩)', 5, [1, 2, 3, 4, 5, 5, 1, 2, 3, 4, 4, 5, 1, 2, 3, 3, 4, 5, 1, 2, 2, 3, 4, 5, 1]);

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ CirculantMatrix(SequenceVector(%d));', [_i]), null);
    Test('IsCirculant(A)', True);
    Test('IsToeplitz(A)', True);
  end;

  Test('CirculantMatrix(❨1, i, 2, −i❩)', 4, [1, ImaginaryUnit, 2, -ImaginaryUnit, -ImaginaryUnit, 1, ImaginaryUnit, 2, 2, -ImaginaryUnit, 1, ImaginaryUnit, ImaginaryUnit, 2, -ImaginaryUnit, 1]);
  Test('IsCirculant(CirculantMatrix(❨1, i, 2, −i❩))', True);
  Test('IsToeplitz(CirculantMatrix(❨1, i, 2, −i❩))', True);

  // Toeplitz matrices

  Test('ToeplitzMatrix(❨1, 2, 3, 4, 5❩)', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('ToeplitzMatrix(❨1, 2, 3, 4, 5❩, ❨1, 20, 30, 40, 50❩)', 5, [1, 2, 3, 4, 5, 20, 1, 2, 3, 4, 30, 20, 1, 2, 3, 40, 30, 20, 1, 2, 50, 40, 30, 20, 1]);
  Test('ToeplitzMatrix(❨1, 2, 3, 4, 5❩, ❨10, 20, 30, 40, 50❩)', failure, 'The first element of the first row must equal the first element of the first column.');

  Test('ToeplitzMatrix(❨1, 2, 3, 4❩, ❨1, 20, 30, 40, 50❩)', 4, [1, 2, 3, 4, 20, 1, 2, 3, 30, 20, 1, 2, 40, 30, 20, 1, 50, 40, 30, 20]);
  Test('ToeplitzMatrix(❨1, 2, 3, 4, 5❩, ❨1, 20, 30, 40❩)', 5, [1, 2, 3, 4, 5, 20, 1, 2, 3, 4, 30, 20, 1, 2, 3, 40, 30, 20, 1, 2]);

  Test('ToeplitzMatrix(❨i, 2, 3, 4, 5❩)', failure, 'Too few arguments. A required argument of type vector is missing.');
  Test('ToeplitzMatrix(❨i, 2, 3, 4, 5❩, ❨i, 20, 30, 40, 50❩)', 5, [ImaginaryUnit, 2, 3, 4, 5, 20, ImaginaryUnit, 2, 3, 4, 30, 20, ImaginaryUnit, 2, 3, 40, 30, 20, ImaginaryUnit, 2, 50, 40, 30, 20, ImaginaryUnit]);
  Test('ToeplitzMatrix(❨i, 2, 3, 4, 5❩, ❨10, 20, 30, 40, 50❩)', failure, 'The first element of the first row must equal the first element of the first column.');

  Test('ToeplitzMatrix(❨i, 2, 3, 4❩, ❨i, 20, 30, 40, 50❩)', 4, [ImaginaryUnit, 2, 3, 4, 20, ImaginaryUnit, 2, 3, 30, 20, ImaginaryUnit, 2, 40, 30, 20, ImaginaryUnit, 50, 40, 30, 20]);
  Test('ToeplitzMatrix(❨i, 2, 3, 4, 5❩, ❨i, 20, 30, 40❩)', 5, [ImaginaryUnit, 2, 3, 4, 5, 20, ImaginaryUnit, 2, 3, 4, 30, 20, ImaginaryUnit, 2, 3, 40, 30, 20, ImaginaryUnit, 2]);

  Test('ToeplitzMatrix(❨1, 2, 3, 4, i, 6❩, ❨1, 20, 30❩)', 6, [1, 2, 3, 4, ImaginaryUnit, 6, 20, 1, 2, 3, 4, ImaginaryUnit, 30, 20, 1, 2, 3, 4]);
  Test('ToeplitzMatrix(❨1, 2, 3❩, ❨1, 20, 30, 40, i, 50❩)', 3, [1, 2, 3, 20, 1, 2, 30, 20, 1, 40, 30, 20, ImaginaryUnit, 40, 30, 50, ImaginaryUnit, 40]);

  Test('IsToeplitz(ToeplitzMatrix(❨1, 2, 3, 4❩, ❨1, 20, 30, 40, 50❩))', True);
  Test('IsToeplitz(ToeplitzMatrix(❨1, 2, 3, 4, 5❩, ❨1, 20, 30, 40❩))', True);

  Test('IsToeplitz(ToeplitzMatrix(❨i, 2, 3, 4, 5❩, ❨i, 20, 30, 40, 50❩))', True);

  Test('IsToeplitz(ToeplitzMatrix(❨i, 2, 3, 4❩, ❨i, 20, 30, 40, 50❩))', True);
  Test('IsToeplitz(ToeplitzMatrix(❨i, 2, 3, 4, 5❩, ❨i, 20, 30, 40❩))', True);

  Test('IsToeplitz(ToeplitzMatrix(❨1, 2, 3, 4, i, 6❩, ❨1, 20, 30❩))', True);
  Test('IsToeplitz(ToeplitzMatrix(❨1, 2, 3❩, ❨1, 20, 30, 40, i, 50❩))', True);

  Test('ToeplitzMatrix(❨1, 2, 3, 4, 5❩, ❨1❩)', 5, [1, 2, 3, 4, 5]);
  Test('ToeplitzMatrix(❨1❩, ❨1, 2, 3❩)', 1, [1, 2, 3]);
  Test('ToeplitzMatrix(❨21❩, ❨21❩)', 1, [21]);
  Test('ToeplitzMatrix(❨21❩, ❨22❩)', failure, 'The first element of the first row must equal the first element of the first column.');
  Test('ToeplitzMatrix(❨21❩, ❨22❩, ❨23❩)', failure, 'Too many arguments.');

  // Hankel matrices

  Test('HankelMatrix(❨1, 2, 3, 4, 5❩, ❨5, 40, 30, 20, 10❩)', 5, [1, 2, 3, 4, 5, 2, 3, 4, 5, 40, 3, 4, 5, 40, 30, 4, 5, 40, 30, 20, 5, 40, 30, 20, 10]);
  Test('HankelMatrix(❨1, 2, 3, 4, 5❩, ❨50, 40, 30, 20, 10❩)', failure, 'The last element of the first row must equal the first element of the last column.');

  Test('HankelMatrix(❨1, 2, 3, 4, 5❩, ❨5, 40, 30❩)', 5, [1, 2, 3, 4, 5, 2, 3, 4, 5, 40, 3, 4, 5, 40, 30]);
  Test('HankelMatrix(❨1, 2, 3❩, ❨3, 40, 50, 60, 70❩)', 3, [1, 2, 3, 2, 3, 40, 3, 40, 50, 40, 50, 60, 50, 60, 70]);

  Test('HankelMatrix(❨i, 2, 3, 4, 5❩, ❨5, 40, 30, 20, 10❩)', 5, [ImaginaryUnit, 2, 3, 4, 5, 2, 3, 4, 5, 40, 3, 4, 5, 40, 30, 4, 5, 40, 30, 20, 5, 40, 30, 20, 10]);
  Test('HankelMatrix(❨i, 2, 3, 4, 5❩, ❨50, 40, 30, 20, 10❩)', failure, 'The last element of the first row must equal the first element of the last column.');

  Test('HankelMatrix(❨i, 2, 3, 4, 5❩, ❨5, 40, 30❩)', 5, [ImaginaryUnit, 2, 3, 4, 5, 2, 3, 4, 5, 40, 3, 4, 5, 40, 30]);
  Test('HankelMatrix(❨i, 2, 3❩, ❨3, 40, 50, 60, 70❩)', 3, [ImaginaryUnit, 2, 3, 2, 3, 40, 3, 40, 50, 40, 50, 60, 50, 60, 70]);

  Test('HankelMatrix(❨1, 2, 3, 4, 5❩, ❨5, 40, 30, 20, i❩)', 5, [1, 2, 3, 4, 5, 2, 3, 4, 5, 40, 3, 4, 5, 40, 30, 4, 5, 40, 30, 20, 5, 40, 30, 20, ImaginaryUnit]);
  Test('HankelMatrix(❨1, 2, 3, 4, 5❩, ❨50, 40, 30, 20, i❩)', failure, 'The last element of the first row must equal the first element of the last column.');

  Test('HankelMatrix(❨1, 2, 3, 4, 5❩, ❨5, 40, i❩)', 5, [1, 2, 3, 4, 5, 2, 3, 4, 5, 40, 3, 4, 5, 40, ImaginaryUnit]);
  Test('HankelMatrix(❨1, 2, 3❩, ❨3, 40, 50, i, 70❩)', 3, [1, 2, 3, 2, 3, 40, 3, 40, 50, 40, 50, ImaginaryUnit, 50, ImaginaryUnit, 70]);

  Test('HankelMatrix(❨i, 2, 3, 4, 5❩, ❨i, 40, i❩)', failure, 'The last element of the first row must equal the first element of the last column.');
  Test('HankelMatrix(❨i, 2, 3, 4, 5❩, ❨5, 40, i❩)', 5, [ImaginaryUnit, 2, 3, 4, 5, 2, 3, 4, 5, 40, 3, 4, 5, 40, ImaginaryUnit]);
  Test('HankelMatrix(❨i, 2, 3❩, ❨3, 40, 50, i, 70❩)', 3, [ImaginaryUnit, 2, 3, 2, 3, 40, 3, 40, 50, 40, 50, ImaginaryUnit, 50, ImaginaryUnit, 70]);

  Test('IsHankel(HankelMatrix(❨1, 2, 3, 4, 5❩, ❨5, 40, 30, 20, 10❩))', True);
  Test('IsHankel(HankelMatrix(❨1, 2, 3, 4, 5❩, ❨5, 40, 30❩))', True);
  Test('IsHankel(HankelMatrix(❨i, 2, 3, 4, 5❩, ❨5, 40, 30❩))', True);
  Test('IsHankel(HankelMatrix(❨i, 2, 3❩, ❨3, 40, 50, 60, 70❩))', True);
  Test('IsHankel(HankelMatrix(❨1, 2, 3❩, ❨3, 40, 50, 60, 70❩))', True);
  Test('IsHankel(HankelMatrix(❨1, 2, 3, 4, 5❩, ❨5, 40, i❩))', True);
  Test('IsHankel(HankelMatrix(❨1, 2, 3❩, ❨3, 40, 50, i, 70❩))', True);
  Test('IsHankel(HankelMatrix(❨i, 2, 3, 4, 5❩, ❨5, 40, i❩))', True);
  Test('IsHankel(HankelMatrix(❨i, 2, 3❩, ❨3, 40, 50, i, 70❩))', True);

  Test('IsSymmetric(HankelMatrix(❨1, 2, 3, 4, 5❩, ❨5, 40, 30, 20, 10❩))', True);

  // Backward shift matrices

  Test('BackwardShiftMatrix(1)', 1, [0]);
  Test('BackwardShiftMatrix(2)', 2, [0, 1, 0, 0]);
  Test('BackwardShiftMatrix(3)', 3, [0, 1, 0, 0, 0, 1, 0, 0, 0]);
  Test('BackwardShiftMatrix(4)', 4, [0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0]);
  Test('BackwardShiftMatrix(5)', 5, [0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0]);

  Test('BackwardShiftMatrix(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('BackwardShiftMatrix(-1)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ BackwardShiftMatrix(%d);', [_i]), null);
    Test('IsBinary(A)', True);
    Test('rank(A)', _i - 1);
    if _i mod 25 = 0 then
      Test('NilpotencyIndex(A)', _i);
    if _i >= 2 then
      Test(Format('IsZero(ReplaceSuperdiagonal(A, ZeroVector(%d)))', [_i - 1]), True);
  end;

  Test('BackwardShiftMatrix(1, 2)', failure, 'Too many arguments.');
  Test('BackwardShiftMatrix()', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('BackwardShiftMatrix(π)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('BackwardShiftMatrix(i)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  // Forward shift matrices

  Test('ForwardShiftMatrix(1)', 1, [0]);
  Test('ForwardShiftMatrix(2)', 2, [0, 0, 1, 0]);
  Test('ForwardShiftMatrix(3)', 3, [0, 0, 0, 1, 0, 0, 0, 1, 0]);
  Test('ForwardShiftMatrix(4)', 4, [0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0]);
  Test('ForwardShiftMatrix(5)', 5, [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0]);

  Test('ForwardShiftMatrix(0)', failure, 'A positive integer was expected as argument 1, but "0" was given.');
  Test('ForwardShiftMatrix(-1)', failure, 'A positive integer was expected as argument 1, but "-1" was given.');

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ ForwardShiftMatrix(%d);', [_i]), null);
    Test('IsBinary(A)', True);
    Test('rank(A)', _i - 1);
    if _i mod 25 = 0 then
      Test('NilpotencyIndex(A)', _i);
    if _i >= 2 then
      Test(Format('IsZero(ReplaceSubdiagonal(A, ZeroVector(%d)))', [_i - 1]), True);
  end;

  Test('ForwardShiftMatrix(1, 2)', failure, 'Too many arguments.');
  Test('ForwardShiftMatrix()', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('ForwardShiftMatrix(π)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('ForwardShiftMatrix(i)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  // Vandermonde matrices

  Test('VandermondeMatrix(❨1, 2, 3, 4❩)', 4, [1, 1, 1, 1, 1, 2, 4, 8, 1, 3, 9, 27, 1, 4, 16, 64]);

  Test('VandermondeMatrix(❨1, 2, 3, 4❩, 5)', 5, [1, 1, 1, 1, 1, 1, 2, 4, 8, 16, 1, 3, 9, 27, 81, 1, 4, 16, 64, 256]);
  Test('VandermondeMatrix(❨1, 2, 3, 4❩, 4)', 4, [1, 1, 1, 1, 1, 2, 4, 8, 1, 3, 9, 27, 1, 4, 16, 64]);
  Test('VandermondeMatrix(❨1, 2, 3, 4❩, 3)', 3, [1, 1, 1, 1, 2, 4, 1, 3, 9, 1, 4, 16]);
  Test('VandermondeMatrix(❨1, 2, 3, 4❩, 2)', 2, [1, 1, 1, 2, 1, 3, 1, 4]);
  Test('VandermondeMatrix(❨1, 2, 3, 4❩, 1)', 1, [1, 1, 1, 1]);
  Test('VandermondeMatrix(❨1, 2, 3, 4❩, 0)', 4, [1, 1, 1, 1, 1, 2, 4, 8, 1, 3, 9, 27, 1, 4, 16, 64]);

  Test('VandermondeMatrix(❨1, 2, 3, 4❩, -1)', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.');

  Test('VandermondeMatrix(❨2❩, 11)', 11, [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]);

  Test('VandermondeMatrix(❨2, −1, i❩, 11)', 11, [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 1, ImaginaryUnit, -1, -ImaginaryUnit, 1, ImaginaryUnit, -1, -ImaginaryUnit, 1, ImaginaryUnit, -1]);
  Test('VandermondeMatrix(❨2, −1, i❩)', 3, [1, 2, 4, 1, -1, 1, 1, ImaginaryUnit, -1]);
  Test('VandermondeMatrix(❨2, −1, i❩, 3)', 3, [1, 2, 4, 1, -1, 1, 1, ImaginaryUnit, -1]);
  Test('VandermondeMatrix(❨2, −1, i❩, 0)', 3, [1, 2, 4, 1, -1, 1, 1, ImaginaryUnit, -1]);
  Test('VandermondeMatrix(❨i❩, 5)', 5, [1, ImaginaryUnit, -1, -ImaginaryUnit, 1]);

  // Hilbert matrices

  Eps; Test('HilbertMatrix(1)', 1, [1]);
  Eps; Test('HilbertMatrix(2)', 2, [1, 1/2, 1/2, 1/3]);
  Eps; Test('HilbertMatrix(3)', 3, [1, 1/2, 1/3, 1/2, 1/3, 1/4, 1/3, 1/4, 1/5]);
  Eps; Test('HilbertMatrix(4)', 4, [1, 1/2, 1/3, 1/4, 1/2, 1/3, 1/4, 1/5, 1/3, 1/4, 1/5, 1/6, 1/4, 1/5, 1/6, 1/7]);
  Eps; Test('HilbertMatrix(5)', 5, [1, 1/2, 1/3, 1/4, 1/5, 1/2, 1/3, 1/4, 1/5, 1/6, 1/3, 1/4, 1/5, 1/6, 1/7, 1/4, 1/5, 1/6, 1/7, 1/8, 1/5, 1/6, 1/7, 1/8, 1/9]);

  Eps; Test('HilbertMatrix(3, 5)', 5, [1, 1/2, 1/3, 1/4, 1/5, 1/2, 1/3, 1/4, 1/5, 1/6, 1/3, 1/4, 1/5, 1/6, 1/7]);
  Eps; Test('HilbertMatrix(6, 3)', 3, [1, 1/2, 1/3, 1/2, 1/3, 1/4, 1/3, 1/4, 1/5, 1/4, 1/5, 1/6, 1/5, 1/6, 1/7, 1/6, 1/7, 1/8]);

  Eps; Test('HilbertMatrix(1, 10)', 10, [1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]);
  Eps; Test('HilbertMatrix(10, 1)', 1, [1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]);

  for _i := 1 to 10 do
    for _j := 1 to 10 do
    begin
      Test(Format('A ≔ HilbertMatrix(%d, %d);', [_i, _j]), null);
      Test('IsHankel(A)', True);
      if _i = _j then
      begin
        Test('IsSymmetric(A)', True);
        Test('IsPositiveDefinite(A)', True);
      end;
    end;

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ HilbertMatrix(%d);', [_i]), null);
    Test('IsHankel(A)', True);
    Test('IsSymmetric(A)', True);
  end;

  Test('delete(A)', success);

  // Rotation matrices

  Eps; Test('RotationMatrix(1.234)', 2, [0.330465108071729857, -0.943818209374633705, 0.943818209374633705, 0.330465108071729857]);
  Eps; Test('RotationMatrix(1.234, 2)', 2, [0.330465108071729857, -0.943818209374633705, 0.943818209374633705, 0.330465108071729857]);

  Eps; Test('RotationMatrix(1.234, 3)', 3, [0.330465108071729857, -0.943818209374633705, 0, 0.943818209374633705, 0.330465108071729857, 0, 0, 0, 1]);
  Eps; Test('RotationMatrix(1.234, 3, 1, 2)', 3, [0.330465108071729857, -0.943818209374633705, 0, 0.943818209374633705, 0.330465108071729857, 0, 0, 0, 1]);
  Eps; Test('RotationMatrix(1.234, 3, 2, 1)', 3, [0.330465108071729857, -0.943818209374633705, 0, 0.943818209374633705, 0.330465108071729857, 0, 0, 0, 1]);
  Eps; Test('RotationMatrix(1.234, 3, 2, 3)', 3, [1, 0, 0, 0, 0.330465108071729857, -0.943818209374633705, 0, 0.943818209374633705, 0.330465108071729857]);
  Eps; Test('RotationMatrix(1.234, 3, 3, 2)', 3, [1, 0, 0, 0, 0.330465108071729857, -0.943818209374633705, 0, 0.943818209374633705, 0.330465108071729857]);

  Eps; Test('RotationMatrix(1.234, 10, 3, 7)', 10, [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.330465108071729857, 0, 0, 0, -0.943818209374633705, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0.943818209374633705, 0, 0, 0, 0.330465108071729857, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]);
  Eps; Test('RotationMatrix(1.234, 10, 7, 3)', 10, [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.330465108071729857, 0, 0, 0, -0.943818209374633705, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0.943818209374633705, 0, 0, 0, 0.330465108071729857, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]);

  Test('RotationMatrix(1.234, 10, 3, 11)', failure, 'Invalid coordinate indices.');
  Test('RotationMatrix(1.234, 10, 11, 3)', failure, 'Invalid coordinate indices.');

  Test('RotationMatrix(1.234, 10, 0, 7)', failure, 'A positive integer was expected as argument 3, but "0" was given.');
  Test('RotationMatrix(1.234, 10, -3, 7)', failure, 'A positive integer was expected as argument 3, but "-3" was given.');

  Test('RotationMatrix(1.234, 10, 3, 0)', failure, 'A positive integer was expected as argument 4, but "0" was given.');
  Test('RotationMatrix(1.234, 10, 3, -3)', failure, 'A positive integer was expected as argument 4, but "-3" was given.');

  Test('RotationMatrix(1.234, 10, 3, 3)', failure, 'Invalid coordinate indices.');

  Test('RotationMatrix(1.234, 1)', failure, 'Invalid coordinate indices.');
  Test('RotationMatrix(1.234, 1, 1, 1)', failure, 'Invalid coordinate indices.');
  Test('RotationMatrix(1.234, 1, 1, 2)', failure, 'Invalid coordinate indices.');

  Test('RotationMatrix(1.234, 10, 7, 3, 5)', failure, 'Too many arguments.');
  Test('RotationMatrix()', failure, 'Too few arguments. A required argument of type real number is missing.');

  Eps; Test('RotationMatrix(90°, ❨1, 0, 0❩) ⋅ ❨1, 0, 0❩', [1, 0, 0]);
  Eps; Test('RotationMatrix(90°, ❨1, 0, 0❩) ⋅ ❨0, 1, 0❩', [0, 0, 1]);
  Eps; Test('RotationMatrix(90°, ❨1, 0, 0❩) ⋅ ❨0, 0, 1❩', [0, -1, 0]);

  Eps; Test('RotationMatrix(90°, ❨0, 1, 0❩) ⋅ ❨1, 0, 0❩', [0, 0, -1]);
  Eps; Test('RotationMatrix(90°, ❨0, 1, 0❩) ⋅ ❨0, 1, 0❩', [0, 1, 0]);
  Eps; Test('RotationMatrix(90°, ❨0, 1, 0❩) ⋅ ❨0, 0, 1❩', [1, 0, 0]);

  Eps; Test('RotationMatrix(90°, ❨0, 0, 1❩) ⋅ ❨1, 0, 0❩', [0, 1, 0]);
  Eps; Test('RotationMatrix(90°, ❨0, 0, 1❩) ⋅ ❨0, 1, 0❩', [-1, 0, 0]);
  Eps; Test('RotationMatrix(90°, ❨0, 0, 1❩) ⋅ ❨0, 0, 1❩', [0, 0, 1]);

  Eps; Test('RotationMatrix(90°, ❨0, 0, 1❩) ⋅ ❨1, 1, 1❩', [-1, 1, 1]); // (really nothing but linearity...)

  Eps; Test('RotationMatrix(90°, ❨1, 1, 0❩) ⋅ ❨1, 0, 0❩', [0.5, 0.5, -1/Sqrt(TASR(2))]);
  Eps; Test('RotationMatrix(90°, ❨1, 1, 0❩) ⋅ ❨0, 1, 0❩', [0.5, 0.5, +1/Sqrt(TASR(2))]);
  Eps; Test('RotationMatrix(90°, ❨1, 1, 0❩) ⋅ ❨0, 0, 1❩', [1/Sqrt(TASR(2)), -1/Sqrt(TASR(2)), 0]);

  Eps; Test('RotationMatrix(90°, ❨1, 0, 1❩) ⋅ ❨1, 0, 0❩', [1/2, 1/Sqrt(TASR(2)), 1/2]);
  Eps; Test('RotationMatrix(90°, ❨1, 0, 1❩) ⋅ ❨0, 1, 0❩', [-1/Sqrt(TASR(2)), 0, 1/Sqrt(TASR(2))]);
  Eps; Test('RotationMatrix(90°, ❨1, 0, 1❩) ⋅ ❨0, 0, 1❩', [1/2, -1/Sqrt(TASR(2)), 1/2]);

  Eps; Test('RotationMatrix(90°, ❨0, 1, 1❩) ⋅ ❨1, 0, 0❩', [0, 1/Sqrt(TASR(2)), -1/Sqrt(TASR(2))]);
  Eps; Test('RotationMatrix(90°, ❨0, 1, 1❩) ⋅ ❨0, 1, 0❩', [-1/Sqrt(TASR(2)), 1/2, 1/2]);
  Eps; Test('RotationMatrix(90°, ❨0, 1, 1❩) ⋅ ❨0, 0, 1❩', [1/Sqrt(TASR(2)), 1/2, 1/2]);

  Eps; Test('RotationMatrix(90°, ❨1, 1, 1❩) ⋅ ❨1, 0, 0❩', [1/3, 0.910683602522959098, -0.244016935856292431]);
  Eps; Test('RotationMatrix(90°, ❨1, 1, 1❩) ⋅ ❨0, 1, 0❩', [-0.244016935856292431, 1/3, 0.910683602522959098]);
  Eps; Test('RotationMatrix(90°, ❨1, 1, 1❩) ⋅ ❨0, 0, 1❩', [0.910683602522959098, -0.244016935856292431, 1/3]);

  Eps; Test('RotationMatrix(38°, ❨1, 2, -1❩) ⋅ ❨1, 2, 3❩', [2.86941579418709284, 0.711976692250990176, 2.29336917868907319]);

  Eps; Test('RotationMatrix(360°, ❨1, 2, -1❩) ⋅ ❨1, 2, 3❩', [1, 2, 3]);
  Eps; Test('RotationMatrix(0°, ❨1, 2, -1❩) ⋅ ❨1, 2, 3❩', [1, 2, 3]);
  Eps; Test('RotationMatrix(180°, ❨1, 2, -1❩) ⋅ RotationMatrix(180°, ❨1, 2, -1❩) ⋅ ❨1, 2, 3❩', [1, 2, 3]);
  Eps; Test('RotationMatrix(17°, ❨1, 2, -1❩) ⋅ RotationMatrix(-17°, ❨1, 2, -1❩) ⋅ ❨1, 2, 3❩', [1, 2, 3]);
  Eps; Test('RotationMatrix(377°, ❨1, 2, -1❩) ⋅ RotationMatrix(-17°, ❨1, 2, -1❩) ⋅ ❨1, 2, 3❩', [1, 2, 3]);

  Test('RotationMatrix(38°, ❨0, 0, 0❩) ⋅ ❨1, 2, 3❩', failure, 'Cannot rotate about the zero vector.');

  _j := RandSeed;
  RandSeed := 632014737;
  try

    for _i := 1 to 100 do
    begin
      Test(Format('A ≔ RotationMatrix(%d°, ❨%d, %d, %d❩);', [Random(1000)-500, 1+Random(10), Random(20)-10, Random(20)-10]), null);
      Test('IsOrthogonal(A)', True);
      Eps; Test('det(A)', 1.0);
    end;

    for _i := 1 to 100 do
    begin
      Test(Format('v ≔ %d°; A ≔ RotationMatrix(v, ❨1, 0, −1❩); B ≔ RotationMatrix(−v, ❨1, 0, −1❩);', [Random(1000)-500]), null);
      Test('SameValue(inv(A), B)', True);
    end;

  finally
    RandSeed := _j
  end;

  Test('delete(A); delete(B); delete(v)', success);

  // Reflection matrices

  Eps; Test('ReflectionMatrix(❨21❩)', 1, [-1]);
  Eps; Test('ReflectionMatrix(❨-21❩)', 1, [-1]);
  Eps; Test('ReflectionMatrix(❨1❩)', 1, [-1]);

  Eps; Test('ReflectionMatrix(❨1, 0❩)', 2, [-1, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(❨0, 1❩)', 2, [1, 0, 0, -1]);

  Eps; Test('ReflectionMatrix(❨1, 0, 0❩)', 3, [-1, 0, 0, 0, 1, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(❨0, 1, 0❩)', 3, [1, 0, 0, 0, -1, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(❨0, 0, 1❩)', 3, [1, 0, 0, 0, 1, 0, 0, 0, -1]);

  Eps; Test('ReflectionMatrix(❨1, 0, 0, 0❩)', 4, [-1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(❨0, 1, 0, 0❩)', 4, [1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(❨0, 0, 1, 0❩)', 4, [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(❨0, 0, 0, 1❩)', 4, [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1]);

  Eps; Test('ReflectionMatrix(❨1, -1❩)', 2, [0, 1, 1, 0]);

  Eps; Test('ReflectionMatrix(❨1, 1, 0❩)', 3, [0, -1, 0, -1, 0, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(❨1, 1, 1❩)', 3, [1/3, -2/3, -2/3, -2/3, 1/3, -2/3, -2/3, -2/3, 1/3]);

  Test('ReflectionMatrix(❨0❩)', failure, 'Vector cannot be zero.');
  Test('ReflectionMatrix(❨0, 0❩)', failure, 'Vector cannot be zero.');
  Test('ReflectionMatrix(❨0, 0, 0❩)', failure, 'Vector cannot be zero.');
  Test('ReflectionMatrix(❨0, 0, 0, 0❩)', failure, 'Vector cannot be zero.');

  Eps; Test('ReflectionMatrix(ComplexVector(❨21❩))', 1, [TASC(-1)]);
  Eps; Test('ReflectionMatrix(ComplexVector(❨-21❩))', 1, [TASC(-1)]);
  Eps; Test('ReflectionMatrix(ComplexVector(❨1❩))', 1, [TASC(-1)]);

  Eps; Test('ReflectionMatrix(ComplexVector(❨1, 0❩))', 2, [TASC(-1), 0, 0, 1]);
  Eps; Test('ReflectionMatrix(ComplexVector(❨0, 1❩))', 2, [TASC(1), 0, 0, -1]);

  Eps; Test('ReflectionMatrix(ComplexVector(❨1, 0, 0❩))', 3, [TASC(-1), 0, 0, 0, 1, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(ComplexVector(❨0, 1, 0❩))', 3, [1, 0, 0, 0, TASC(-1), 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(ComplexVector(❨0, 0, 1❩))', 3, [1, 0, 0, 0, 1, 0, 0, 0, TASC(-1)]);

  Eps; Test('ReflectionMatrix(ComplexVector(❨1, 0, 0, 0❩))', 4, [TASC(-1), 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(ComplexVector(❨0, 1, 0, 0❩))', 4, [1, 0, 0, 0, 0, TASC(-1), 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(ComplexVector(❨0, 0, 1, 0❩))', 4, [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, TASC(-1), 0, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(ComplexVector(❨0, 0, 0, 1❩))', 4, [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, TASC(-1)]);

  Eps; Test('ReflectionMatrix(ComplexVector(❨1, -1❩))', 2, [0, TASC(1), 1, 0]);

  Eps; Test('ReflectionMatrix(ComplexVector(❨1, 1, 0❩))', 3, [0, TASC(-1), 0, TASC(-1), 0, 0, 0, 0, 1]);
  Eps; Test('ReflectionMatrix(ComplexVector(❨1, 1, 1❩))', 3, [TASC(1/3), -2/3, -2/3, -2/3, 1/3, -2/3, -2/3, -2/3, 1/3]);

  Test('ReflectionMatrix(ComplexVector(❨0❩))', failure, 'Vector cannot be zero.');
  Test('ReflectionMatrix(ComplexVector(❨0, 0❩))', failure, 'Vector cannot be zero.');
  Test('ReflectionMatrix(ComplexVector(❨0, 0, 0❩))', failure, 'Vector cannot be zero.');
  Test('ReflectionMatrix(ComplexVector(❨0, 0, 0, 0❩))', failure, 'Vector cannot be zero.');

  Test('A ≔ ReflectionMatrix(❨5, 3, 0, 1, 4❩);', null);
  Test('IsSymmetric(A)', True);
  Test('IsOrthogonal(A)', True);
  Test('IsNormal(A)', True);
  Test('IsInvolution(A)', True);
  Eps; Test('det(A)', -1.0);

  Test('A ≔ ReflectionMatrix(❨5, 3, 0, i, 4❩);', null);
  Test('IsHermitian(A)', True);
  Test('IsUnitary(A)', True);
  Test('IsNormal(A)', True);
  Test('IsInvolution(A)', True);
  Eps; Test('det(A)', TASC(-1.0));

  Test('delete(A)', success);


  //
  // Systems of linear equations
  //

  // General system solver (real)

  Test('A ≔ ❨❨3, 5, 0❩, ❨1, 4, −2❩, ❨6, 2, 9❩❩;', null);

  // - one by one

  Test('SysSolve(A, ❨13, 3, 37❩)', [1, 2, 3]);
  Test('SysSolve(A, ❨30, 15, 45❩)', [5, 3, 1]);
  Test('SysSolve(A, ❨14, −1, 56❩)', [3, 1, 4]);

  // - one by one, augmented matrix

  Test('SysSolve(aug(A, ❨13, 3, 37❩))', [1, 2, 3]);
  Test('SysSolve(aug(A, ❨30, 15, 45❩))', [5, 3, 1]);
  Test('SysSolve(aug(A, ❨14, −1, 56❩))', [3, 1, 4]);

  // - all at once

  Test('SysSolve(A, ❨❨13, 30, 14❩, ❨3, 15, -1❩, ❨37, 45, 56❩❩)', 3, [1, 5, 3, 2, 3, 1, 3, 1, 4]);

  // -

  Test('A ≔ ❨❨3, 0, 2, 1❩, ❨8, 4, 2, 9❩, ❨5, 3, −4, 2❩, ❨7, 0, 2, 5❩❩;', null);

  Test('SysSolve(A, ❨26, 52, 15, 42❩)', [6, 3, 5, -2]);
  Test('SysSolve(A, ❨1, 6, 9, 5❩)', [1, 0, -1, 0]);
  Test('SysSolve(A, ❨0, 0, 0, 0❩)', [0, 0, 0, 0]);

  Test('SysSolve(aug(A, ❨26, 52, 15, 42❩))', [6, 3, 5, -2]);
  Test('SysSolve(aug(A, ❨1, 6, 9, 5❩))', [1, 0, -1, 0]);
  Test('SysSolve(aug(A, ❨0, 0, 0, 0❩))', [0, 0, 0, 0]);

  Test('SysSolve(A, ❨❨26, 1, 0❩, ❨52, 6, 0❩, ❨15, 9, 0❩, ❨42, 5, 0❩❩)', 3, [6, 1, 0, 3, 0, 0, 5, -1, 0, -2, 0, 0]);

  // -

  Test('A ≔ ❨❨3, 6, 5, 2, 0, 1, 4, 7, 0, 2❩, ❨5, −3, 2, 1, 4, 5, 0, 7, 8, 6❩, ❨6, 2, 3, 5, 0, 7, 4, 9, 6, 5❩, ❨−5, 0, 3, 7, 4, 1, 5, 6, 3, 3❩, ❨5, 2, 0, 4, 1, −6, 7, 8, 5, 0❩, ❨4, 2, 5, 3, 3, 1, 0, 4, 9, 5❩, ❨7, 4, 5, 2, 6, 0, 7, 4, 9, 6❩, '+'❨−5, 6, 3, 2, −8, 4, 6, 2, 1, 7❩, ❨5, 6, 6, 3, 9, 5, 0, 4, 1, 2❩, ❨0, 5, 8, 6, 2, 1, 4, 2, 3, 6❩❩;', null);

  Eps; Test('SysSolve(A, ❨148, 247, 285, 198, 152, 219, 290, 145, 183, 205❩)', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  Eps; Test('SysSolve(A, ❨12, 19, 19, 10, 18, 21, 34, -3, 21, 17❩)', [1, 0, 1, 0, 1, 0, 1, 0, 1, 0]);
  Eps; Test('SysSolve(A, ❨2, 6, 5, 3, 0, 5, 6, 7, 2, 6❩)', [0, 0, 0, 0, 0, 0, 0, 0, 0, 1]);
  Eps; Test('SysSolve(A, ❨3, 5, 6, -5, 5, 4, 7, -5, 5, 0❩)', [1, 0, 0, 0, 0, 0, 0, 0, 0, 0]);
  Eps; Test('SysSolve(A, ❨129, 137, 196, 79, 91, 86, 153, 58, 160, 90❩)', [6, 3, 0, 1, 4, 7, 5, 8, -2, 4]);
  Eps(1E-12); Test('SysSolve(A, ❨16, 400030, 47, 400021, 100008, 300048, 600051, -799970, 900039, 200049❩)', [1, 2, 1, 3, 100000, 2, 0, -2, 3, 1]);
  Eps; Test('SysSolve(A, ❨0, 0, 0, 0, 0, 0, 0, 0, 0, 0❩)', [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]);

  // -

  Test('A ≔ ❨❨3, 6, 5, 2, 0, 1, 4, 2, 0, 2❩, ❨5, -3, 2, 1, 4, 5, 0, -3, 8, 6❩, ❨6, 2, 3, 5, 0, 7, 4, 5, 6, 5❩, ❨-5, 0, 3, 7, 4, 1, 5, 3, 3, 3❩, ❨5, 2, 0, 4, 1, -6, 7, 3, 5, 0❩, ❨4, 2, 5, 3, 3, 1, 0, 0, 9, 5❩, ❨7, 4, 5, 2, 6, 0, 7, -4, 9, 6❩, '+'❨-5, 6, 3, 2, -8, 4, 6, 10, 1, 7❩, ❨5, 6, 6, 3, 9, 5, 0, -6, 1, 2❩, ❨0, 5, 8, 6, 2, 1, 4, 4, 3, 6❩❩;', null);
  Eps; Test('SysSolve(A, ❨12, 19, 19, 10, 18, 21, 34, -3, 21, 17❩)', failure, 'Matrix is singular.');

  Test('SysSolve(ZeroMatrix(5), ❨5, 6, 3, 2, 1❩)', failure, 'Matrix is singular.');
  Test('SysSolve(aug(ZeroMatrix(5), ❨5, 6, 3, 2, 1❩))', failure, 'Matrix is singular.');
  Test('SysSolve(ZeroMatrix(5), ❨❨5❩, ❨6❩, ❨3❩, ❨2❩, ❨1❩❩)', failure, 'Matrix is singular.');

  // -

  Test('SysSolve(❨❨6, 3, 2, 1❩, ❨5, 4, 0, 1❩, ❨−5, 2, 1, 7❩, ❨0, 1, 2, 9❩❩, ❨6❩)', failure, 'RHS is of wrong dimension.');
  Test('SysSolve(❨❨6, 3, 2, 1❩, ❨5, 4, 0, 1❩, ❨−5, 2, 1, 7❩, ❨0, 1, 2, 9❩❩, ❨6, 3❩)', failure, 'RHS is of wrong dimension.');
  Test('SysSolve(❨❨6, 3, 2, 1❩, ❨5, 4, 0, 1❩, ❨−5, 2, 1, 7❩, ❨0, 1, 2, 9❩❩, ❨6, 3, 0❩)', failure, 'RHS is of wrong dimension.');
  Test('SysSolve(❨❨6, 3, 2, 1❩, ❨5, 4, 0, 1❩, ❨−5, 2, 1, 7❩, ❨0, 1, 2, 9❩❩, ❨6, 3, 0, 7, 8❩)', failure, 'RHS is of wrong dimension.');

  Test('SysSolve(❨❨6, 3, 2, 1❩, ❨5, 4, 0, 1❩, ❨−5, 2, 1, 7❩, ❨0, 1, 2, 9❩❩, ❨❨6❩, ❨3❩, ❨0❩❩)', failure, 'RHS is of wrong dimension.');

  Test('SysSolve(RandomMatrix(100), ❨6, 3, 0❩)', failure, 'RHS is of wrong dimension.');

  // -

  Test('SysSolve(❨❨5, 4, 1, 2❩, ❨6, 3, 0, 2❩, ❨1, 5, 7, 2❩❩, ❨4, 5, 7, 6❩)', failure, 'Coefficient matrix isn''t square.');
  Test('SysSolve(❨❨5, 4, 1, 2❩, ❨6, 3, 0, 2❩, ❨1, 5, 7, 2❩, ❨7, 5, 6, 1❩, ❨0, −2, 4, 8❩❩, ❨4, 5, 7, 6❩)', failure, 'Coefficient matrix isn''t square.');
  Test('SysSolve(❨❨3, 2, 5, 6❩, ❨7, 4, 2, 1❩❩)', failure);

  Test('SysSolve(❨❨5, 4, 1, 2❩, ❨6, 3, 0, 2❩, ❨1, 5, 7, 2❩❩, ❨❨4❩, ❨5❩, ❨7❩, ❨6❩❩)', failure, 'Coefficient matrix isn''t square.');

  // General system solver (complex)

  Test('A ≔ ❨❨2, i, 6❩, ❨−7, 0, i❩, ❨3, 2, 4❩❩;', null);

  // - one by one

  Eps; Test('SysSolve(A, ❨6 + 6⋅i, −22, 9 + 4⋅i❩)', [3, 0, ImaginaryUnit]);
  Eps; Test('SysSolve(A, ❨16 − 2⋅i, −35 + i, 15❩)', [TASC(5), -2, 1]);
  Eps; Test('SysSolve(A, ❨1 − 6⋅i, −6, 3 − 2⋅i❩)', [1, ImaginaryUnit, -ImaginaryUnit]);
  Eps; Test('SysSolve(A, ❨7, 4, −2❩)', [-0.622467771639042357 + 0.173112338858195212*ImaginaryUnit, -2.48987108655616943 - 0.97421731123388582*ImaginaryUnit, 1.21178637200736648 + 0.357274401473296501*ImaginaryUnit]);

  // - one by one, augmented matrix

  Eps; Test('SysSolve(aug(A, ❨6 + 6⋅i, −22, 9 + 4⋅i❩))', [3, 0, ImaginaryUnit]);
  Eps; Test('SysSolve(aug(A, ❨16 − 2⋅i, −35 + i, 15❩))', [TASC(5), -2, 1]);
  Eps; Test('SysSolve(aug(A, ❨1 − 6⋅i, −6, 3 − 2⋅i❩))', [1, ImaginaryUnit, -ImaginaryUnit]);
  Eps; Test('SysSolve(aug(A, ❨7, 4, −2❩))', [-0.622467771639042357 + 0.173112338858195212*ImaginaryUnit, -2.48987108655616943 - 0.97421731123388582*ImaginaryUnit, 1.21178637200736648 + 0.357274401473296501*ImaginaryUnit]);

  // - all at once

  Eps; Test('SysSolve(A, ❨❨6 + 6⋅i, 16 - 2⋅i, 1 - 6⋅i, 7❩, ❨-22, -35 + i, -6, 4❩, ❨9 + 4⋅i, 15, 3 - 2⋅i, -2❩❩)', 4, [3, 5, 1, -0.622467771639042357 + 0.173112338858195212*ImaginaryUnit, 0, -2, ImaginaryUnit, -2.48987108655616943 - 0.97421731123388582*ImaginaryUnit, ImaginaryUnit, 1, -ImaginaryUnit, 1.21178637200736648 + 0.357274401473296501*ImaginaryUnit]);

  // -

  Test('A ≔ ❨❨3, 6, 5, 1❩, ❨7, 4, 5, 2❩, ❨0, 8, 6, 3❩, ❨5, 1, 4, 2❩❩;', null);

  Eps; Test('SysSolve(A, ❨38 + 5⋅i, 52 + 2⋅i, 30 + 5⋅i, 32 - i❩)', [5, 3 + ImaginaryUnit, 1, -ImaginaryUnit]);
  Eps; Test('SysSolve(aug(A, ❨38 + 5⋅i, 52 + 2⋅i, 30 + 5⋅i, 32 - i❩))', [5, 3 + ImaginaryUnit, 1, -ImaginaryUnit]);

  // -

  Test('A ≔ ❨❨6, 3, 0, i, 1, 2, 4, 5, 3, 0❩, ❨1, 4, 2, 3, 0, 5, 9, i, 6, 3❩, ❨5, 7, i, 1, 2, 4, 4, 7, 1, 0❩, ❨3, 2, 0, 6, 2, 7, i, 7, 4, 1❩, ❨2, 1, 4, 5, 3, 0, 2, 8, 7, 6❩, ❨5, 2, 3, 1, 0, 4, 7, 8, 6, 0❩, ❨7, 4, 8, 6, i, 1, 2, 3, 6, 9❩, '+'❨4, 1, 5, 2, 3, 0, 8, 2, 3, 6❩, ❨1, 0, 4, i, i, 6, 8, 7, 9, 6❩, ❨9, 6, 3, 0, 1, i, 4, 4, 7, 1❩❩;', null);

  Eps; Test('SysSolve(A, ❨67 - 4⋅i, 103 + 2⋅i, 84 - 3⋅i, 92 + 8⋅i, 114 + 15⋅i, 93 - 6⋅i, 159 + 21⋅i, 84 + 6⋅i, 142 + 3⋅i, 114 + 2⋅i❩)',
    [3, 5, 0, ImaginaryUnit, -1, 4, -ImaginaryUnit, 2, 6, 8 + 2*ImaginaryUnit]);

  Eps; Test('SysSolve(A, ❨0, 3 + 3⋅i, 0, 1 + i, 6 + 6⋅i, 0, 9 + 9⋅i, 6 + 6⋅i, 6 + 6⋅i, 1 + i❩)',
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 1 + ImaginaryUnit]);

  Eps; Test('SysSolve(A, ❨6 - 6⋅i, 1 - i, 5 - 5⋅i, 3 - 3⋅i, 2 - 2⋅i, 5 - 5⋅i, 7 - 7⋅i, 4 - 4⋅i, 1 - i, 9 - 9⋅i❩)',
    [1 - ImaginaryUnit, 0, 0, 0, 0, 0, 0, 0, 0, 0]);

  // -

  Test('A ≔ ❨❨6, 3, 0, i, 1, 2, 4, -9 + i, 3, 0❩, ❨1, 4, 2, 3, 0, 5, 9, -20, 6, 3❩, ❨5, 7, i, 1, 2, 4, 4, -4, 1, 0❩, ❨3, 2, 0, 6, 2, 7, i, 8 - 3⋅i, 4, 1❩, ❨2, 1, 4, 5, 3, 0, 2, 0, 7, 6❩, ❨5, 2, 3, 1, 0, 4, 7, -18, 6, 0❩, ❨7, 4, 8, 6, i, 1, 2, 4, 6, 9❩, '+'❨4, 1, 5, 2, 3, 0, 8, -21, 3, 6❩, ❨1, 0, 4, i, i, 6, 8, -24 + i, 9, 6❩, ❨9, 6, 3, 0, 1, i, 4, -6, 7, 1❩❩;', null);

  Test('SysSolve(A, ❨67 - 4⋅i, 103 + 2⋅i, 84 - 3⋅i, 92 + 8⋅i, 114 + 15⋅i, 93 - 6⋅i, 159 + 21⋅i, 84 + 6⋅i, 142 + 3⋅i, 114 + 2⋅i❩)', failure, 'Matrix is singular.');
  Test('SysSolve(aug(A, ❨67 - 4⋅i, 103 + 2⋅i, 84 - 3⋅i, 92 + 8⋅i, 114 + 15⋅i, 93 - 6⋅i, 159 + 21⋅i, 84 + 6⋅i, 142 + 3⋅i, 114 + 2⋅i❩))', failure, 'Matrix is singular.');

  Test('SysSolve(ComplexZeroMatrix(5), IdentityMatrix(5))', failure, 'Matrix is singular.');

  // -

  Test('SysSolve(❨❨6, 3, i, 1❩, ❨1, i, 5, 2❩, ❨0, 4, 9, i❩, ❨6, 3, 2, 1❩❩, ❨4❩)', failure, 'RHS is of wrong dimension.');
  Test('SysSolve(❨❨6, 3, i, 1❩, ❨1, i, 5, 2❩, ❨0, 4, 9, i❩, ❨6, 3, 2, 1❩❩, ❨4, 7❩)', failure, 'RHS is of wrong dimension.');
  Test('SysSolve(❨❨6, 3, i, 1❩, ❨1, i, 5, 2❩, ❨0, 4, 9, i❩, ❨6, 3, 2, 1❩❩, ❨4, 7, i❩)', failure, 'RHS is of wrong dimension.');
  Test('SysSolve(❨❨6, 3, i, 1❩, ❨1, i, 5, 2❩, ❨0, 4, 9, i❩, ❨6, 3, 2, 1❩❩, ❨4, 7, i, 7, 8❩)', failure, 'RHS is of wrong dimension.');

  Test('SysSolve(❨❨6, 3, i, 1❩, ❨1, i, 5, 2❩, ❨0, 4, 9, i❩, ❨6, 3, 2, 1❩❩, IdentityMatrix(3))', failure, 'RHS is of wrong dimension.');

  // -

  Test('SysSolve(❨❨6, 3, i, 1❩, ❨1, i, 5, 2❩, ❨0, 4, 9, i❩❩, ❨4, 7, i, 5❩)', failure, 'Coefficient matrix isn''t square.');
  Test('SysSolve(❨❨6, 3, i, 1❩, ❨1, i, 5, 2❩, ❨0, 4, 9, i❩, ❨5, i, -1, i❩, ❨4, 1, 1, -1❩❩, ❨4, 7, i, 5❩)', failure, 'Coefficient matrix isn''t square.');

  Test('SysSolve(❨❨6, 3, i, 1❩, ❨1, i, 5, 2❩, ❨0, 4, 9, i❩❩, IdentityMatrix(3))', failure, 'Coefficient matrix isn''t square.');
  Test('SysSolve(❨❨6, 3, i, 1❩, ❨1, i, 5, 2❩, ❨0, 4, 9, i❩❩, IdentityMatrix(4))', failure, 'Coefficient matrix isn''t square.');

  // Back substitution

  Test('A ≔ ❨❨4, 5, 3, 0❩, ❨0, 5, 1, 2❩, ❨0, 0, 2, 1❩, ❨0, 0, 0, 7❩❩;', null);
  Test('BackSubstitution(A, ❨23, 21, 10, 28❩)', [1, 2, 3, 4]);

  Test('BackSubstitution(A, ❨23, 21, 10❩)', failure, 'RHS is of wrong dimension.');
  Test('BackSubstitution(A, ❨23, 21❩)', failure, 'RHS is of wrong dimension.');
  Test('BackSubstitution(A, ❨23, 21, 10, 56, 98, 12❩)', failure, 'RHS is of wrong dimension.');

  Test('A ≔ ❨❨4, 5, 3, 0❩, ❨0, 5, 1, 2❩, ❨0, 0, 2, 1❩, ❨0, 0, 0, 0❩❩;', null);
  Test('BackSubstitution(A, ❨23, 21, 10, 28❩)', failure, 'Matrix is singular.');

  Test('A ≔ ❨❨4, 5, 3, 0❩, ❨0, 5, 1, 2❩, ❨0, 0, 2, 1❩, ❨0, 0, 0, 7❩, ❨0, 3, 2, 7❩❩;', null);
  Test('BackSubstitution(A, ❨23, 21, 10, 28❩)', failure, 'Coefficient matrix isn''t square.');
  Test('A ≔ ❨❨4, 5, 3, 0❩, ❨0, 5, 1, 2❩, ❨0, 0, 2, 1❩❩;', null);
  Test('BackSubstitution(A, ❨23, 21, 10, 28❩)', failure, 'Coefficient matrix isn''t square.');

  Test('A ≔ ❨❨3, 5, i, 4❩, ❨0, i, 2, 1❩, ❨0, 0, 4, 8❩, ❨0, 0, 0, 3❩❩;', null);
  Test('BackSubstitution(A, ❨29 + 23⋅i, 10 + 7⋅i, 44 + 40⋅i, 12 + 15⋅i❩)', [1, 2, 3, 4 + 5*ImaginaryUnit]);

  Test('BackSubstitution(A, ❨29 + 23⋅i, 10 + 7⋅i, 44 + 40⋅i, 12 + 15⋅i, 82❩)', failure, 'RHS is of wrong dimension.');
  Test('BackSubstitution(A, ❨29 + 23⋅i, 10 + 7⋅i, 44 + 40⋅i❩)', failure, 'RHS is of wrong dimension.');
  Test('BackSubstitution(A, ❨29 + 23⋅i, 10 + 7⋅i❩)', failure, 'RHS is of wrong dimension.');
  Test('BackSubstitution(A, ❨29 + 23⋅i❩)', failure, 'RHS is of wrong dimension.');

  Test('A ≔ ❨❨3, 5, i, 4❩, ❨0, i, 2, 1❩, ❨0, 0, 4, 8❩, ❨0, 0, 0, 0❩❩;', null);
  Test('BackSubstitution(A, ❨29 + 23⋅i, 10 + 7⋅i, 44 + 40⋅i, 12 + 15⋅i❩)', failure, 'Matrix is singular.');

  Test('BackSubstitution(❨❨3, 5, i, 4❩, ❨0, i, 2, 1❩, ❨0, 0, 4, 8❩, ❨0, 0, 0, 3❩, ❨0, 0, 0, 6❩❩, ❨29 + 23⋅i, 10 + 7⋅i, 44 + 40⋅i, 12 + 15⋅i❩)', failure, 'Coefficient matrix isn''t square.');
  Test('BackSubstitution(❨❨3, 5, i, 4❩, ❨0, i, 2, 1❩, ❨0, 0, 4, 8❩❩, ❨29 + 23⋅i, 10 + 7⋅i, 44 + 40⋅i, 12 + 15⋅i❩)', failure, 'Coefficient matrix isn''t square.');

  // - performance comparison

  Test('N ≔ 10;', null);
  Test('M, Y, X, X2 ≔ SequenceList(10000);', null);
  Test('for(i, 1, length(M), (M[i] ≔ ToUpperTriangular(RandomMatrix(N))));', null);
  Test('for(i, 1, length(Y), (Y[i] ≔ RandomVector(N)));', null);

  StartStopwatch;
  Test('for(i, 1, length(M), (X[i] ≔ SysSolve(M[i], Y[i])));', null);
  Test('t1 ≔ history(-1).EvalTime;', null);
  StopStopwatch;
  TestDuration(0.17);

  StartStopwatch;
  Test('for(i, 1, length(M), (X2[i] ≔ BackSubstitution(M[i], Y[i])));', null);
  Test('t2 ≔ history(-1).EvalTime;', null);
  StopStopwatch;
  TestDuration(0.12);

  Test('t2 < t1', True);

  Test('for(i, 1, length(M), (X[i] = X2[i]));', null);

  Test('delete(A); delete(N); delete(M); delete(Y); delete(X); delete(X2); delete(t1); delete(t2)', success);

  // Forward substitution

  Test('A ≔ ❨❨6, 0, 0, 0❩, ❨1, 2, 0, 0❩, ❨5, 4, 9, 0❩, ❨7, 3, 2, 1❩❩;', null);
  Test('ForwardSubstitution(A, ❨48, 10, 116, 80❩)', [8, 1, 8, 5]);
  Test('ForwardSubstitution(A, ❨48, 10❩)', failure, 'RHS is of wrong dimension.');
  Test('ForwardSubstitution(A, ❨48, 10, 116❩)', failure, 'RHS is of wrong dimension.');
  Test('ForwardSubstitution(A, ❨48, 10, 116, 80, 21❩)', failure, 'RHS is of wrong dimension.');

  Test('A ≔ ❨❨0, 0, 0, 0❩, ❨1, 2, 0, 0❩, ❨5, 4, 9, 0❩, ❨7, 3, 2, 1❩❩;', null);
  Test('ForwardSubstitution(A, ❨48, 10, 116, 80❩)', failure, 'Matrix is singular.');

  Test('A ≔ ❨❨0, 0, 0, 0❩, ❨1, 2, 0, 0❩, ❨5, 4, 9, 0❩, ❨7, 3, 2, 1❩, ❨1, 4, 0, 1❩❩;', null);
  Test('ForwardSubstitution(A, ❨48, 10, 116, 80❩)', failure, 'Coefficient matrix isn''t square.');

  Test('A ≔ ❨❨0, 0, 0, 0❩, ❨1, 2, 0, 0❩, ❨5, 4, 9, 0❩❩;', null);
  Test('ForwardSubstitution(A, ❨48, 10, 116, 80❩)', failure, 'Coefficient matrix isn''t square.');

  Test('A ≔ ❨❨3, 0, 0, 0, 0❩, ❨2, i, 0, 0, 0❩, ❨1, 3, 4, 0, 0❩, ❨5, 3, i, 7, 0❩, ❨4, i, 2, 0, 1❩❩;', null);
  Test('ForwardSubstitution(A, ❨12 + 3⋅i, 8 + 4⋅i, 10 + i, 26 - 2⋅i, 23 + 6⋅i❩)', [4 + ImaginaryUnit, 2, 0, -ImaginaryUnit, 7]);

  Test('A ≔ ❨❨0, 0, 0, 0, 0❩, ❨2, i, 0, 0, 0❩, ❨1, 3, 4, 0, 0❩, ❨5, 3, i, 7, 0❩, ❨4, i, 2, 0, 1❩❩;', null);
  Test('ForwardSubstitution(A, ❨12 + 3⋅i, 8 + 4⋅i, 10 + i, 26 - 2⋅i, 23 + 6⋅i❩)', failure, 'Matrix is singular.');

  Test('A ≔ ❨❨3, 0, 0, 0, 0❩, ❨2, i, 0, 0, 0❩, ❨1, 3, 4, 0, 0❩, ❨5, 3, i, 7, 0❩, ❨4, i, 2, 0, 1❩❩;', null);
  Test('ForwardSubstitution(A, ❨12 + 3⋅i, 8 + 4⋅i, 10 + i, 26 - 2⋅i, 23 + 6⋅i, 4❩)', failure, 'RHS is of wrong dimension.');
  Test('ForwardSubstitution(A, ❨12 + 3⋅i, 8 + 4⋅i, 10 + i, 26 - 2⋅i❩)', failure, 'RHS is of wrong dimension.');

  Test('A ≔ ❨❨3, 0, 0, 0, 0❩, ❨2, i, 0, 0, 0❩, ❨1, 3, 4, 0, 0❩, ❨5, 3, i, 7, 0❩, ❨4, i, 2, 0, 1❩, ❨-4, 7, 4, -3, 1❩❩;', null);
  Test('ForwardSubstitution(A, ❨12 + 3⋅i, 8 + 4⋅i, 10 + i, 26 - 2⋅i, 23 + 6⋅i❩)', failure, 'Coefficient matrix isn''t square.');

  Test('A ≔ ❨❨3, 0, 0, 0, 0❩, ❨2, i, 0, 0, 0❩, ❨1, 3, 4, 0, 0❩, ❨5, 3, i, 7, 0❩❩;', null);
  Test('ForwardSubstitution(A, ❨12 + 3⋅i, 8 + 4⋅i, 10 + i, 26 - 2⋅i, 23 + 6⋅i❩)', failure, 'Coefficient matrix isn''t square.');

  Test('delete(A)', success);


  //
  // Least-squares polynomial curve fitting
  //

  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨2, 4, 6, 8, 10❩, 1)', [0, 2]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨3, 6, 9, 12, 15❩, 1)', [0, 3]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨10, 20, 30, 40, 50❩, 1)', [0, 10]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨15, 25, 35, 45, 55❩, 1)', [5, 10]);

  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨2, 4, 6, 8, 10❩, 2)', [0, 2, 0]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨3, 6, 9, 12, 15❩, 2)', [0, 3, 0]);
  Eps(1E-15); Test('polyfit(❨1, 2, 3, 4, 5❩, ❨10, 20, 30, 40, 50❩, 2)', [0, 10, 0]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨15, 25, 35, 45, 55❩, 2)', [5, 10, 0]);

  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨1, 4, 9, 16, 25❩, 2)', [0, 0, 1]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨3, 8, 15, 24, 35❩, 2)', [0, 2, 1]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5❩, ❨6, 11, 18, 27, 38❩, 2)', [3, 2, 1]);

  Eps(1E-15); Test('polyfit(❨1, 2, 3, 4, 5❩, ❨6, 11, 18, 27, 38❩, 3)', [3, 2, 1, 0]);

  Test('polyfit(❨1, 2, 3, 4, 5❩, ❨6, 11, 18, 27❩, 2)', failure, 'X and Y vectors must have the same dimension.');

  Eps; Test('polyfit(❨1, 2❩, ❨2, 4❩, 1)', [0, 2]);
  Eps; Test('polyfit(❨1, 2❩, ❨1, 4❩, 1)', [-2, 3]);
  Eps; Test('polyfit(❨1, 2❩, ❨2, 4❩, 0)', [3]);
  Eps; Test('polyfit(❨1, 2❩, ❨1, 4❩, 0)', [2.5]);
  Eps; Test('polyfit(❨1, 2❩, ❨1, 2❩, 0)', [1.5]);

  Eps; Test('polyfit(❨1❩, ❨1❩, 0)', [1]);

  Test('polyfit(❨1, 2❩, ❨1, 4❩, -1)', failure, 'A non-negative integer was expected as argument 3, but "-1" was given.');

  SetLength(_xarr, 1000);
  SetLength(_yarr, 1000);
  SetLength(_sxarr, 1000);
  SetLength(_syarr, 1000);
  for _i := 0 to High(_xarr) do
  begin
    _xarr[_i] := _i;
    _yarr[_i] := 2*_i*_i*_i + 4*_i*_i - _i + 1;
    _sxarr[_i] := _xarr[_i].ToString(DefaultFormatSettings);
    _syarr[_i] := _yarr[_i].ToString(DefaultFormatSettings);
  end;

  Test(Format('X ≔ ❨%s❩;', [string.Join(', ', _sxarr)]), null);
  Test(Format('Y ≔ ❨%s❩;', [string.Join(', ', _syarr)]), null);

  Eps(1E-6); Test('polyfit(X, Y, 3)', [1, -1, 4, 2]);

  for _i := 0 to High(_xarr) do
  begin
    _xarr[_i] := _i;
    _yarr[_i] := 2*_i*_i*_i + 4*_i*_i - _i + 1 + 0.1*Random;
    _sxarr[_i] := _xarr[_i].ToString(DefaultFormatSettings);
    _syarr[_i] := _yarr[_i].ToString(DefaultFormatSettings);
  end;

  Test(Format('X ≔ ❨%s❩;', [string.Join(', ', _sxarr)]), null);
  Test(Format('Y ≔ ❨%s❩;', [string.Join(', ', _syarr)]), null);

  Eps(1E-1); Test('polyfit(X, Y, 3)', [1, -1, 4, 2]);

  Eps; Test('polyfit(❨1, 2, 3, 4, 5, i❩, ❨2, 4, 6, 8, 10, 2⋅i❩, 1)', [TASC(0), 2]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5, i❩, ❨3, 6, 9, 12, 15, 3⋅i❩, 1)', [TASC(0), 3]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5, i❩, ❨4, 7, 10, 13, 16, 3⋅i + 1❩, 1)', [TASC(1), 3]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5, i❩, ❨2⋅i, 4⋅i, 6⋅i, 8⋅i, 10⋅i, -2❩, 1)', [0, 2*ImaginaryUnit]);
  Eps; Test('polyfit(❨1, 2, 3, 4, 5, i❩, ❨3 + 2⋅i, 3 + 4⋅i, 3 + 6⋅i, 3 + 8⋅i, 3 + 10⋅i, 1❩, 1)', [3, 2*ImaginaryUnit]);

  SetLength(_cxarr, 1000);
  SetLength(_cyarr, 1000);
  SetLength(_sxarr, 1000);
  SetLength(_syarr, 1000);
  for _i := 0 to High(_cxarr) do
  begin
    _cxarr[_i] := _i;
    _cyarr[_i] := (3 + 7*ImaginaryUnit)*_i*_i*_i - 4*_i*_i - ImaginaryUnit*_i + 4 + 3*ImaginaryUnit;
    _sxarr[_i] := ComplexToStr(_cxarr[_i], False, InputFormOptions);
    _syarr[_i] := ComplexToStr(_cyarr[_i], False, InputFormOptions);
  end;

  Test(Format('X ≔ ❨%s❩;', [string.Join(', ', _sxarr)]), null);
  Test(Format('Y ≔ ❨%s❩;', [string.Join(', ', _syarr)]), null);

  Eps(1E-6); Test('polyfit(X, Y, 3)', [4 + 3*ImaginaryUnit, -ImaginaryUnit, -4, 3 + 7*ImaginaryUnit]);

  _xarr := nil;
  _yarr := nil;
  _cxarr := nil;
  _cyarr := nil;
  _sxarr := nil;
  _syarr := nil;

  Test('delete(X); delete(Y)', success);


  //
  // Supermatrices, submatrices, vector conversions, replacing parts
  //

  // Replacing parts

  Test('A ≔ ❨❨4, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨4, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨1, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨5, 6, 3❩❩; E ≔ ❨❨7❩, ❨1❩❩; F ≔ ❨❨21❩❩;', null);

  Test('ReplaceRow(A, 1, ❨1, 2, 3, 4❩)', 4, [1, 2, 3, 4, 0, 1, 8, 7, 9, 3, 5, 4, 1, 7, 5, 2]);
  Test('ReplaceRow(A, 2, ❨10, 20, 30, 40❩)', 4, [4, 2, 3, 6, 10, 20, 30, 40, 9, 3, 5, 4, 1, 7, 5, 2]);
  Test('ReplaceRow(A, 3, ❨9, 8, 7, 6❩)', 4, [4, 2, 3, 6, 0, 1, 8, 7, 9, 8, 7, 6, 1, 7, 5, 2]);
  Test('ReplaceRow(A, 4, ❨0, 1, 0, 1❩)', 4, [4, 2, 3, 6, 0, 1, 8, 7, 9, 3, 5, 4, 0, 1, 0, 1]);

  Test('ReplaceRow(A, 5, ❨0, 1, 0, 1❩)', failure, 'Index 5 out of bounds.');
  Test('ReplaceRow(A, 0, ❨0, 1, 0, 1❩)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('ReplaceRow(A, -1, ❨0, 1, 0, 1❩)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('ReplaceRow(A, 1, ❨1, 2, 3❩)', failure, 'Incorrect length of array.');
  Test('ReplaceRow(A, 1, ❨1, 2, 3, 4, 5❩)', failure, 'Incorrect length of array.');

  Test('ReplaceRow(B, 1, ❨5, 4, 3, 2, 1❩)', 5, [5, 4, 3, 2, 1, 7, 4, 5, 1, 2]);
  Test('ReplaceRow(B, 2, ❨0, 0, 0, 0, 0❩)', 5, [4, 6, 5, 3, 2, 0, 0, 0, 0, 0]);
  Test('ReplaceRow(B, 3, ❨0, 0, 0, 0, 0❩)', failure, 'Index 3 out of bounds.');

  Test('ReplaceRow(C, 4, ❨10, 20, 30❩)', 3, [1, 4, 0, 7, 5, 6, 7, 8, 5, 10, 20, 30, 4, 7, 8]);

  Test('ReplaceRow(D, 1, ❨4, 5, 6❩)', 3, [4, 5, 6]);
  Test('ReplaceRow(E, 2, ❨21❩)', 1, [7, 21]);

  Test('ReplaceRow(F, 1, ❨6❩)', 1, [6]);

  Test('ReplaceCol(A, 1, ❨1, 2, 3, 4❩)', 4, [1, 2, 3, 6, 2, 1, 8, 7, 3, 3, 5, 4, 4, 7, 5, 2]);
  Test('ReplaceCol(A, 2, ❨9, 8, 7, 6❩)', 4, [4, 9, 3, 6, 0, 8, 8, 7, 9, 7, 5, 4, 1, 6, 5, 2]);
  Test('ReplaceCol(A, 3, ❨0, 0, 0, 0❩)', 4, [4, 2, 0, 6, 0, 1, 0, 7, 9, 3, 0, 4, 1, 7, 0, 2]);
  Test('ReplaceCol(A, 4, ❨5, 0, 4, 0❩)', 4, [4, 2, 3, 5, 0, 1, 8, 0, 9, 3, 5, 4, 1, 7, 5, 0]);

  Test('ReplaceCol(A, 5, ❨5, 0, 4, 0❩)', failure, 'Index 5 out of bounds.');
  Test('ReplaceCol(A, 0, ❨5, 0, 4, 0❩)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('ReplaceCol(A, -1, ❨5, 0, 4, 0❩)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('ReplaceCol(A, 3, ❨1, 2, 3❩)', failure, 'Incorrect length of array.');
  Test('ReplaceCol(A, 3, ❨1, 2, 3, 4, 5❩)', failure, 'Incorrect length of array.');

  Test('ReplaceCol(B, 4, ❨10, 20❩)', 5, [4, 6, 5, 10, 2, 7, 4, 5, 20, 2]);
  Test('ReplaceCol(C, 3, ❨10, 20, 30, 40, 50❩)', 3, [1, 4, 10, 7, 5, 20, 7, 8, 30, 2, 0, 40, 4, 7, 50]);
  Test('ReplaceCol(D, 2, ❨1❩)', 3, [5, 1, 3]);
  Test('ReplaceCol(E, 1, ❨9, 8❩)', 1, [9, 8]);
  Test('ReplaceCol(F, 1, ❨4❩)', 1, [4]);

  Test('ReplaceDiagonal(A, ❨1, 2, 3, 4❩)', 4, [1, 2, 3, 6, 0, 2, 8, 7, 9, 3, 3, 4, 1, 7, 5, 4]);
  Test('ReplaceDiagonal(A, ❨1, 2, 3❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(A, ❨1, 2, 3, 4, 5❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(B, ❨1, 2❩)', 5, [1, 6, 5, 3, 2, 7, 2, 5, 1, 2]);
  Test('ReplaceDiagonal(B, ❨1, 2, 3❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(C, ❨1, 2, 3❩)', 3, [1, 4, 0, 7, 2, 6, 7, 8, 3, 2, 0, 3, 4, 7, 8]);
  Test('ReplaceDiagonal(C, ❨1, 2, 3, 4❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(D, ❨21❩)', 3, [21, 6, 3]);
  Test('ReplaceDiagonal(D, ❨21, 22❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(E, ❨21❩)', 1, [21, 1]);
  Test('ReplaceDiagonal(E, ❨21, 11❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(F, ❨23❩)', 1, [23]);
  Test('ReplaceDiagonal(F, ❨23, 24❩)', failure, 'Incorrect number of elements in diagonal.');

  Test('ReplaceSuperdiagonal(A, ❨10, 20, 30❩)', 4, [4, 10, 3, 6, 0, 1, 20, 7, 9, 3, 5, 30, 1, 7, 5, 2]);
  Test('ReplaceSuperdiagonal(A, ❨10, 20❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(A, ❨10, 20, 30, 40❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(B, ❨60, 50❩)', 5, [4, 60, 5, 3, 2, 7, 4, 50, 1, 2]);
  Test('ReplaceSuperdiagonal(B, ❨60❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(B, ❨60, 50, 40❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(C, ❨40, 60❩)', 3, [1, 40, 0, 7, 5, 60, 7, 8, 5, 2, 0, 3, 4, 7, 8]);
  Test('ReplaceSuperdiagonal(C, ❨40❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(C, ❨40, 60, 80❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(D, ❨60❩)', 3, [5, 60, 3]);
  Test('ReplaceSuperdiagonal(D, ❨60, 70❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(E, DebugObject("empty vector"))', 1, [7, 1]);
  Test('ReplaceSuperdiagonal(E, ❨80❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(F, DebugObject("empty vector"))', 1, [21]);
  Test('ReplaceSuperdiagonal(F, ❨23❩)', failure, 'Incorrect number of elements in superdiagonal.');

  Test('ReplaceSubdiagonal(A, ❨10, 20, 30❩)', 4, [4, 2, 3, 6, 10, 1, 8, 7, 9, 20, 5, 4, 1, 7, 30, 2]);
  Test('ReplaceSubdiagonal(A, ❨10, 20❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(A, ❨10, 20, 30, 40❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(B, ❨70❩)', 5, [4, 6, 5, 3, 2, 70, 4, 5, 1, 2]);
  Test('ReplaceSubdiagonal(B, ❨70, 80❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(C, ❨70, 80, 90❩)', 3, [1, 4, 0, 70, 5, 6, 7, 80, 5, 2, 0, 90, 4, 7, 8]);
  Test('ReplaceSubdiagonal(C, ❨70, 80❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(C, ❨70, 80, 90, 100❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(D, ❨100❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(D, DebugObject("empty vector"))', 3, [5, 6, 3]);
  Test('ReplaceSubdiagonal(E, ❨10❩)', 1, [7, 10]);
  Test('ReplaceSubdiagonal(E, ❨10, 20❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(F, ❨1000❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(F, DebugObject("empty vector"))', 1, [21]);

  Test('ReplaceAntidiagonal(A, ❨10, 20, 30, 40❩)', 4, [4, 2, 3, 10, 0, 1, 20, 7, 9, 30, 5, 4, 40, 7, 5, 2]);
  Test('ReplaceAntidiagonal(A, ❨10, 20, 30❩)', failure, 'Incorrect number of elements in antidiagonal.');
  Test('ReplaceAntidiagonal(A, ❨10, 20, 30, 40, 50❩)', failure, 'Incorrect number of elements in antidiagonal.');
  Test('ReplaceAntidiagonal(B, ❨10, 20❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('ReplaceAntidiagonal(C, ❨10, 20, 30❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('ReplaceAntidiagonal(D, ❨10❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('ReplaceAntidiagonal(E, ❨10❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('ReplaceAntidiagonal(F, ❨10❩)', 1, [10]);
  Test('ReplaceAntidiagonal(F, ❨10, 20❩)', failure, 'Incorrect number of elements in antidiagonal.');

  Test('A ≔ ❨❨i, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨i, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨i, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨i, 6, 3❩❩; E ≔ ❨❨i❩, ❨1❩❩; F ≔ ❨❨i❩❩;', null);

  Test('ReplaceRow(A, 1, ❨1, 2, 3, 4❩)', 4, [TASC(1), 2, 3, 4, 0, 1, 8, 7, 9, 3, 5, 4, 1, 7, 5, 2]);
  Test('ReplaceRow(A, 1, ❨1, 2, 3, i/4❩)', 4, [TASC(1), 2, 3, ImaginaryUnit/4, 0, 1, 8, 7, 9, 3, 5, 4, 1, 7, 5, 2]);
  Test('ReplaceRow(A, 2, ❨10, 20, 30, 40❩)', 4, [ImaginaryUnit, 2, 3, 6, 10, 20, 30, 40, 9, 3, 5, 4, 1, 7, 5, 2]);
  Test('ReplaceRow(A, 3, ❨9, 8, 7, 6❩)', 4, [ImaginaryUnit, 2, 3, 6, 0, 1, 8, 7, 9, 8, 7, 6, 1, 7, 5, 2]);
  Test('ReplaceRow(A, 4, ❨0, 1, 0, 1❩)', 4, [ImaginaryUnit, 2, 3, 6, 0, 1, 8, 7, 9, 3, 5, 4, 0, 1, 0, 1]);

  Test('ReplaceRow(A, 5, ❨0, 1, 0, 1❩)', failure, 'Index 5 out of bounds.');
  Test('ReplaceRow(A, 0, ❨0, 1, 0, 1❩)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('ReplaceRow(A, -1, ❨0, 1, 0, 1❩)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('ReplaceRow(A, 1, ❨1, 2, 3❩)', failure, 'Incorrect length of array.');
  Test('ReplaceRow(A, 1, ❨1, 2, 3, 4, 5❩)', failure, 'Incorrect length of array.');

  Test('ReplaceRow(B, 1, ❨5, 4, 3, 2, 1❩)', 5, [TASC(5), 4, 3, 2, 1, 7, 4, 5, 1, 2]);
  Test('ReplaceRow(B, 2, ❨0, 0, 0, 0, 0❩)', 5, [ImaginaryUnit, 6, 5, 3, 2, 0, 0, 0, 0, 0]);
  Test('ReplaceRow(B, 3, ❨0, 0, 0, 0, 0❩)', failure, 'Index 3 out of bounds.');

  Test('ReplaceRow(C, 4, ❨10, 20, 30❩)', 3, [ImaginaryUnit, 4, 0, 7, 5, 6, 7, 8, 5, 10, 20, 30, 4, 7, 8]);

  Test('ReplaceRow(D, 1, ❨4, 5, 6❩)', 3, [TASC(4), 5, 6]);
  Test('ReplaceRow(E, 2, ❨21❩)', 1, [ImaginaryUnit, 21]);

  Test('ReplaceRow(F, 1, ❨i/2❩)', 1, [ImaginaryUnit/2]);

  Test('ReplaceCol(A, 1, ❨1, 2, 3, 4❩)', 4, [TASC(1), 2, 3, 6, 2, 1, 8, 7, 3, 3, 5, 4, 4, 7, 5, 2]);
  Test('ReplaceCol(A, 1, ❨1, 2, 3, i/4❩)', 4, [TASC(1), 2, 3, 6, 2, 1, 8, 7, 3, 3, 5, 4, ImaginaryUnit/4, 7, 5, 2]);
  Test('ReplaceCol(A, 2, ❨9, 8, 7, 6❩)', 4, [ImaginaryUnit, 9, 3, 6, 0, 8, 8, 7, 9, 7, 5, 4, 1, 6, 5, 2]);
  Test('ReplaceCol(A, 3, ❨0, 0, 0, 0❩)', 4, [ImaginaryUnit, 2, 0, 6, 0, 1, 0, 7, 9, 3, 0, 4, 1, 7, 0, 2]);
  Test('ReplaceCol(A, 4, ❨5, 0, 4, 0❩)', 4, [ImaginaryUnit, 2, 3, 5, 0, 1, 8, 0, 9, 3, 5, 4, 1, 7, 5, 0]);

  Test('ReplaceCol(A, 5, ❨5, 0, 4, 0❩)', failure, 'Index 5 out of bounds.');
  Test('ReplaceCol(A, 0, ❨5, 0, 4, 0❩)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('ReplaceCol(A, -1, ❨5, i, 4, 0❩)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('ReplaceCol(A, 3, ❨i, 2, 3❩)', failure, 'Incorrect length of array.');
  Test('ReplaceCol(A, 3, ❨1, 2, 3, 4, 5❩)', failure, 'Incorrect length of array.');

  Test('ReplaceCol(B, 4, ❨10, 20❩)', 5, [ImaginaryUnit, 6, 5, 10, 2, 7, 4, 5, 20, 2]);
  Test('ReplaceCol(C, 3, ❨10, 20, 30, 40, 50❩)', 3, [ImaginaryUnit, 4, 10, 7, 5, 20, 7, 8, 30, 2, 0, 40, 4, 7, 50]);
  Test('ReplaceCol(D, 2, ❨1❩)', 3, [ImaginaryUnit, 1, 3]);
  Test('ReplaceCol(E, 1, ❨9, 8❩)', 1, [TASC(9), 8]);
  Test('ReplaceCol(F, 1, ❨4❩)', 1, [TASC(4)]);

  Test('ReplaceDiagonal(A, ❨1, 2, 3, 4❩)', 4, [TASC(1), 2, 3, 6, 0, 2, 8, 7, 9, 3, 3, 4, 1, 7, 5, 4]);
  Test('ReplaceDiagonal(A, ❨1, 2, 3, i/4❩)', 4, [TASC(1), 2, 3, 6, 0, 2, 8, 7, 9, 3, 3, 4, 1, 7, 5, ImaginaryUnit/4]);
  Test('ReplaceDiagonal(A, ❨1, 2, 3❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(A, ❨1, i, 3, 4, 5❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(B, ❨1, 2❩)', 5, [TASC(1), 6, 5, 3, 2, 7, 2, 5, 1, 2]);
  Test('ReplaceDiagonal(B, ❨1, 2, 3❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(C, ❨1, 2, 3❩)', 3, [TASC(1), 4, 0, 7, 2, 6, 7, 8, 3, 2, 0, 3, 4, 7, 8]);
  Test('ReplaceDiagonal(C, ❨1, 2, 3, 4❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(D, ❨21❩)', 3, [TASC(21), 6, 3]);
  Test('ReplaceDiagonal(D, ❨i, 22❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(E, ❨i/4❩)', 1, [ImaginaryUnit/4, 1]);
  Test('ReplaceDiagonal(E, ❨21, 11❩)', failure, 'Incorrect number of elements in diagonal.');
  Test('ReplaceDiagonal(F, ❨23❩)', 1, [TASC(23)]);
  Test('ReplaceDiagonal(F, ❨23, 24❩)', failure, 'Incorrect number of elements in diagonal.');

  Test('ReplaceSuperdiagonal(A, ❨10, 20, 30❩)', 4, [ImaginaryUnit, 10, 3, 6, 0, 1, 20, 7, 9, 3, 5, 30, 1, 7, 5, 2]);
  Test('ReplaceSuperdiagonal(A, ❨10, 20❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(A, ❨10, 20, 30, 40❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(B, ❨60, 50❩)', 5, [ImaginaryUnit, 60, 5, 3, 2, 7, 4, 50, 1, 2]);
  Test('ReplaceSuperdiagonal(B, ❨60❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(B, ❨60, 50, 40❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(C, ❨40, 60⋅i❩)', 3, [ImaginaryUnit, 40, 0, 7, 5, 60*ImaginaryUnit, 7, 8, 5, 2, 0, 3, 4, 7, 8]);
  Test('ReplaceSuperdiagonal(C, ❨40❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(C, ❨40, 60, 80❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(D, ❨60❩)', 3, [ImaginaryUnit, 60, 3]);
  Test('ReplaceSuperdiagonal(D, ❨60, 70❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(E, DebugObject("empty vector"))', 1, [ImaginaryUnit, 1]);
  Test('ReplaceSuperdiagonal(E, ❨80❩)', failure, 'Incorrect number of elements in superdiagonal.');
  Test('ReplaceSuperdiagonal(F, DebugObject("empty vector"))', 1, [ImaginaryUnit]);
  Test('ReplaceSuperdiagonal(F, ❨23❩)', failure, 'Incorrect number of elements in superdiagonal.');

  Test('ReplaceSubdiagonal(A, ❨10, 20, 30❩)', 4, [ImaginaryUnit, 2, 3, 6, 10, 1, 8, 7, 9, 20, 5, 4, 1, 7, 30, 2]);
  Test('ReplaceSubdiagonal(A, ❨10, 20❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(A, ❨10, 20, 30, 40❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(B, ❨70⋅i❩)', 5, [ImaginaryUnit, 6, 5, 3, 2, 70*ImaginaryUnit, 4, 5, 1, 2]);
  Test('ReplaceSubdiagonal(B, ❨70, 80❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(C, ❨70, 80, 90❩)', 3, [ImaginaryUnit, 4, 0, 70, 5, 6, 7, 80, 5, 2, 0, 90, 4, 7, 8]);
  Test('ReplaceSubdiagonal(C, ❨70, 80❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(C, ❨70, 80, 90, 100❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(D, ❨100❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(D, DebugObject("empty vector"))', 3, [ImaginaryUnit, 6, 3]);
  Test('ReplaceSubdiagonal(E, ❨10❩)', 1, [ImaginaryUnit, 10]);
  Test('ReplaceSubdiagonal(E, ❨10, 20❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(F, ❨1000❩)', failure, 'Incorrect number of elements in subdiagonal.');
  Test('ReplaceSubdiagonal(F, DebugObject("empty vector"))', 1, [ImaginaryUnit]);

  Test('ReplaceAntidiagonal(A, ❨10, 20, 30, 40❩)', 4, [ImaginaryUnit, 2, 3, 10, 0, 1, 20, 7, 9, 30, 5, 4, 40, 7, 5, 2]);
  Test('ReplaceAntidiagonal(A, ❨10, 20, 30❩)', failure, 'Incorrect number of elements in antidiagonal.');
  Test('ReplaceAntidiagonal(A, ❨10, 20, 30, 40, 50❩)', failure, 'Incorrect number of elements in antidiagonal.');
  Test('ReplaceAntidiagonal(B, ❨10, 20❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('ReplaceAntidiagonal(C, ❨10, 20, 30❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('ReplaceAntidiagonal(D, ❨10❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('ReplaceAntidiagonal(E, ❨10❩)', failure, 'A non-square matrix has no antidiagonal.');
  Test('ReplaceAntidiagonal(F, ❨10❩)', 1, [TASC(10)]);
  Test('ReplaceAntidiagonal(F, ❨10, 20❩)', failure, 'Incorrect number of elements in antidiagonal.');

  // Vector conversions

  Test('A ≔ ❨❨4, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨4, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨1, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨5, 6, 3❩❩; E ≔ ❨❨7❩, ❨1❩❩; F ≔ ❨❨21❩❩;', null);

  Test('vector(A)', [4, 2, 3, 6, 0, 1, 8, 7, 9, 3, 5, 4, 1, 7, 5, 2]);
  Test('vector(B)', [4, 6, 5, 3, 2, 7, 4, 5, 1, 2]);
  Test('vector(C)', [1, 4, 0, 7, 5, 6, 7, 8, 5, 2, 0, 3, 4, 7, 8]);
  Test('vector(D)', [5, 6, 3]);
  Test('vector(E)', [7, 1]);
  Test('vector(F)', [21]);

  Test('vec(A)', [4, 0, 9, 1, 2, 1, 3, 7, 3, 8, 5, 5, 6, 7, 4, 2]);
  Test('vec(B)', [4, 7, 6, 4, 5, 5, 3, 1, 2, 2]);
  Test('vec(C)', [1, 7, 7, 2, 4, 4, 5, 8, 0, 7, 0, 6, 5, 3, 8]);
  Test('vec(D)', [5, 6, 3]);
  Test('vec(E)', [7, 1]);
  Test('vec(F)', [21]);

  Test('vectorization(A)', [4, 0, 9, 1, 2, 1, 3, 7, 3, 8, 5, 5, 6, 7, 4, 2]);
  Test('vectorization(B)', [4, 7, 6, 4, 5, 5, 3, 1, 2, 2]);
  Test('vectorization(C)', [1, 7, 7, 2, 4, 4, 5, 8, 0, 7, 0, 6, 5, 3, 8]);
  Test('vectorization(D)', [5, 6, 3]);
  Test('vectorization(E)', [7, 1]);
  Test('vectorization(F)', [21]);

  Test('A ≔ ❨❨i, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨i, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨i, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨i, 6, 3❩❩; E ≔ ❨❨i❩, ❨1❩❩; F ≔ ❨❨i❩❩;', null);

  Test('vector(A)', [ImaginaryUnit, 2, 3, 6, 0, 1, 8, 7, 9, 3, 5, 4, 1, 7, 5, 2]);
  Test('vector(B)', [ImaginaryUnit, 6, 5, 3, 2, 7, 4, 5, 1, 2]);
  Test('vector(C)', [ImaginaryUnit, 4, 0, 7, 5, 6, 7, 8, 5, 2, 0, 3, 4, 7, 8]);
  Test('vector(D)', [ImaginaryUnit, 6, 3]);
  Test('vector(E)', [ImaginaryUnit, 1]);
  Test('vector(F)', [ImaginaryUnit]);

  Test('vec(A)', [ImaginaryUnit, 0, 9, 1, 2, 1, 3, 7, 3, 8, 5, 5, 6, 7, 4, 2]);
  Test('vec(B)', [ImaginaryUnit, 7, 6, 4, 5, 5, 3, 1, 2, 2]);
  Test('vec(C)', [ImaginaryUnit, 7, 7, 2, 4, 4, 5, 8, 0, 7, 0, 6, 5, 3, 8]);
  Test('vec(D)', [ImaginaryUnit, 6, 3]);
  Test('vec(E)', [ImaginaryUnit, 1]);
  Test('vec(F)', [ImaginaryUnit]);

  Test('vectorization(A)', [ImaginaryUnit, 0, 9, 1, 2, 1, 3, 7, 3, 8, 5, 5, 6, 7, 4, 2]);
  Test('vectorization(B)', [ImaginaryUnit, 7, 6, 4, 5, 5, 3, 1, 2, 2]);
  Test('vectorization(C)', [ImaginaryUnit, 7, 7, 2, 4, 4, 5, 8, 0, 7, 0, 6, 5, 3, 8]);
  Test('vectorization(D)', [ImaginaryUnit, 6, 3]);
  Test('vectorization(E)', [ImaginaryUnit, 1]);
  Test('vectorization(F)', [ImaginaryUnit]);

  // Matrix augmentation

  Test('A ≔ ❨❨4, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨4, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨1, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨5, 6, 3❩❩; E ≔ ❨❨7❩, ❨1❩❩; F ≔ ❨❨21❩❩;', null);

  Test('aug(A, ❨10, 20, 30, 40❩)', 5, [4, 2, 3, 6, 10, 0, 1, 8, 7, 20, 9, 3, 5, 4, 30, 1, 7, 5, 2, 40]);
  Test('aug(A, ❨10, 20, 30❩)', failure, 'Cannot augment matrix with different number of rows.');
  Test('aug(A, ❨10, 20, 30, 40, 50❩)', failure, 'Cannot augment matrix with different number of rows.');

  Test('aug(A, ❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩, ❨7, 8❩❩)', 6, [4, 2, 3, 6, 1, 2, 0, 1, 8, 7, 3, 4, 9, 3, 5, 4, 5, 6, 1, 7, 5, 2, 7, 8]);
  Test('aug(A, ❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩)', failure, 'Cannot augment matrix with different number of rows.');
  Test('aug(A, ❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩, ❨7, 8❩, ❨9, 10❩❩)', failure, 'Cannot augment matrix with different number of rows.');

  Test('aug(B, ❨0, 1❩)', 6, [4, 6, 5, 3, 2, 0, 7, 4, 5, 1, 2, 1]);
  Test('aug(D, ❨8❩)', 4, [5, 6, 3, 8]);
  Test('aug(E, ❨4, 2❩)', 2, [7, 4, 1, 2]);
  Test('aug(F, ❨23❩)', 2, [21, 23]);

  Test('aug(❨1, 2, 3❩, ❨4, 5, 6❩)', 2, [1, 4, 2, 5, 3, 6]);

  Test('aug(❨1, 2, 3❩, IdentityMatrix(3))', 4, [1, 1, 0, 0, 2, 0, 1, 0, 3, 0, 0, 1]);

  Test('aug(IdentityMatrix(3), IdentityMatrix(3))', 6, [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1]);
  Test('aug(IdentityMatrix(3), IdentityMatrix(2))', failure, 'Cannot augment matrix with different number of rows.');

  Test('A ≔ ❨❨i, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨i, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨i, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨i, 6, 3❩❩; E ≔ ❨❨i❩, ❨1❩❩; F ≔ ❨❨i❩❩;', null);

  Test('aug(A, ❨10, 20, 30, 40❩)', 5, [ImaginaryUnit, 2, 3, 6, 10, 0, 1, 8, 7, 20, 9, 3, 5, 4, 30, 1, 7, 5, 2, 40]);
  Test('aug(A, ❨10, 20, 30❩)', failure, 'Cannot augment matrix with different number of rows.');
  Test('aug(A, ❨10, 20, 30, 40, 50❩)', failure, 'Cannot augment matrix with different number of rows.');

  Test('aug(A, ❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩, ❨7, 8❩❩)', 6, [ImaginaryUnit, 2, 3, 6, 1, 2, 0, 1, 8, 7, 3, 4, 9, 3, 5, 4, 5, 6, 1, 7, 5, 2, 7, 8]);
  Test('aug(A, ❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩)', failure, 'Cannot augment matrix with different number of rows.');
  Test('aug(A, ❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩, ❨7, 8❩, ❨9, 10❩❩)', failure, 'Cannot augment matrix with different number of rows.');

  Test('aug(B, ❨0, 1❩)', 6, [ImaginaryUnit, 6, 5, 3, 2, 0, 7, 4, 5, 1, 2, 1]);
  Test('aug(D, ❨8❩)', 4, [ImaginaryUnit, 6, 3, 8]);
  Test('aug(E, ❨4, 2❩)', 2, [ImaginaryUnit, 4, 1, 2]);
  Test('aug(F, ❨23❩)', 2, [ImaginaryUnit, 23]);

  Test('aug(❨i, 2, 3❩, ❨4, 5, 6❩)', 2, [ImaginaryUnit, 4, 2, 5, 3, 6]);
  Test('aug(❨1, 2, 3❩, ❨i/4, 5, 6❩)', 2, [1, ImaginaryUnit/4, 2, 5, 3, 6]);
  Test('aug(❨i, 2, 3❩, ❨i/4, 5, 6❩)', 2, [ImaginaryUnit, ImaginaryUnit/4, 2, 5, 3, 6]);

  Test('aug(❨i, 2, 3❩, IdentityMatrix(3))', 4, [ImaginaryUnit, 1, 0, 0, 2, 0, 1, 0, 3, 0, 0, 1]);

  Test('aug(IdentityMatrix(3), ComplexMatrix(IdentityMatrix(3)))', 6, [TASC(1), 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1]);
  Test('aug(IdentityMatrix(3), ComplexMatrix(IdentityMatrix(2)))', failure, 'Cannot augment matrix with different number of rows.');

  Test('A ≔ ❨❨4, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨4, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨1, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨5, 6, 3❩❩; E ≔ ❨❨7❩, ❨1❩❩; F ≔ ❨❨21❩❩;', null);

  Test('lessen(A)', 3, [4, 2, 3, 0, 1, 8, 9, 3, 5, 1, 7, 5]);
  Test('lessen(B)', 4, [4, 6, 5, 3, 7, 4, 5, 1]);
  Test('lessen(C)', 2, [1, 4, 7, 5, 7, 8, 2, 0, 4, 7]);
  Test('lessen(D)', 2, [5, 6]);
  Test('lessen(E)', failure, 'Cannot lessen a single-column matrix.');
  Test('lessen(F)', failure, 'Cannot lessen a single-column matrix.');

  Test('A ≔ ❨❨i, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨i, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨i, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨i, 6, 3❩❩; E ≔ ❨❨i❩, ❨1❩❩; F ≔ ❨❨i❩❩;', null);

  Test('lessen(A)', 3, [ImaginaryUnit, 2, 3, 0, 1, 8, 9, 3, 5, 1, 7, 5]);
  Test('lessen(B)', 4, [ImaginaryUnit, 6, 5, 3, 7, 4, 5, 1]);
  Test('lessen(C)', 2, [ImaginaryUnit, 4, 7, 5, 7, 8, 2, 0, 4, 7]);
  Test('lessen(D)', 2, [ImaginaryUnit, 6]);
  Test('lessen(E)', failure, 'Cannot lessen a single-column matrix.');
  Test('lessen(F)', failure, 'Cannot lessen a single-column matrix.');

  for _i := 1 to 100 do
  begin
    Test(Format('A ≔ RandomMatrix(%d); v ≔ RandomVector(%d);', [_i, _i]), null); // random seed irrelevant
    Test('lessen(aug(A, v)) = A', True);
  end;

  // Submatrices

  Test('A ≔ ❨❨4, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨4, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨1, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨5, 6, 3❩❩; E ≔ ❨❨7❩, ❨1❩❩; F ≔ ❨❨21❩❩;', null);

  Test('SubmatrixByRemoval(A, 2, 3)', 3, [4, 2, 6, 9, 3, 4, 1, 7, 2]);
  Test('SubmatrixByRemoval(A, 2, 5)', failure, 'Invalid row or column index');
  Test('SubmatrixByRemoval(A, 5, 3)', failure, 'Invalid row or column index');

  Test('SubmatrixByRemoval(B, 1, 4)', 4, [7, 4, 5, 2]);
  Test('SubmatrixByRemoval(C, 4, 1)', 2, [4, 0, 5, 6, 8, 5, 7, 8]);
  Test('SubmatrixByRemoval(D, 1, 2)', failure, 'A matrix must have size at least 1×1.');
  Test('SubmatrixByRemoval(E, 1, 1)', failure, 'A matrix must have size at least 1×1.');
  Test('SubmatrixByRemoval(F, 1, 1)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨i, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨i, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨i, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨i, 6, 3❩❩; E ≔ ❨❨i❩, ❨1❩❩; F ≔ ❨❨i❩❩;', null);

  Test('SubmatrixByRemoval(A, 2, 3)', 3, [ImaginaryUnit, 2, 6, 9, 3, 4, 1, 7, 2]);
  Test('SubmatrixByRemoval(A, 2, 5)', failure, 'Invalid row or column index');
  Test('SubmatrixByRemoval(A, 5, 3)', failure, 'Invalid row or column index');

  Test('SubmatrixByRemoval(B, 1, 4)', 4, [TASC(7), 4, 5, 2]);
  Test('SubmatrixByRemoval(C, 4, 1)', 2, [TASC(4), 0, 5, 6, 8, 5, 7, 8]);
  Test('SubmatrixByRemoval(D, 1, 2)', failure, 'A matrix must have size at least 1×1.');
  Test('SubmatrixByRemoval(E, 1, 1)', failure, 'A matrix must have size at least 1×1.');
  Test('SubmatrixByRemoval(F, 1, 1)', failure, 'A matrix must have size at least 1×1.');

  Test('A ≔ ❨❨4, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨4, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨1, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨5, 6, 3❩❩; E ≔ ❨❨7❩, ❨1❩❩; F ≔ ❨❨21❩❩;', null);

  Test('LeadingPrincipalSubmatrix(A)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('LeadingPrincipalSubmatrix(A, 1)', 1, [4]);
  Test('LeadingPrincipalSubmatrix(A, 2)', 2, [4, 2, 0, 1]);
  Test('LeadingPrincipalSubmatrix(A, 3)', 3, [4, 2, 3, 0, 1, 8, 9, 3, 5]);
  Test('LeadingPrincipalSubmatrix(A, 4)', 4, [4, 2, 3, 6, 0, 1, 8, 7, 9, 3, 5, 4, 1, 7, 5, 2]);
  Test('LeadingPrincipalSubmatrix(A, 5)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(B, 1)', 1, [4]);
  Test('LeadingPrincipalSubmatrix(B, 2)', 2, [4, 6, 7, 4]);
  Test('LeadingPrincipalSubmatrix(B, 3)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(C, 1)', 1, [1]);
  Test('LeadingPrincipalSubmatrix(C, 2)', 2, [1, 4, 7, 5]);
  Test('LeadingPrincipalSubmatrix(C, 3)', 3, [1, 4, 0, 7, 5, 6, 7, 8, 5]);
  Test('LeadingPrincipalSubmatrix(C, 4)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(D, 1)', 1, [5]);
  Test('LeadingPrincipalSubmatrix(D, 2)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(E, 1)', 1, [7]);
  Test('LeadingPrincipalSubmatrix(E, 2)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(F, 1)', 1, [21]);
  Test('LeadingPrincipalSubmatrix(F, 2)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('A ≔ ❨❨i, 2, 3, 6❩, ❨0, 1, 8, 7❩, ❨9, 3, 5, 4❩, ❨1, 7, 5, 2❩❩; B ≔ ❨❨i, 6, 5, 3, 2❩, ❨7, 4, 5, 1, 2❩❩; C ≔ ❨❨i, 4, 0❩, ❨7, 5, 6❩, ❨7, 8, 5❩, ❨2, 0, 3❩, ❨4, 7, 8❩❩; D ≔ ❨❨i, 6, 3❩❩; E ≔ ❨❨i❩, ❨1❩❩; F ≔ ❨❨i❩❩;', null);

  Test('LeadingPrincipalSubmatrix(A)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('LeadingPrincipalSubmatrix(A, 1)', 1, [ImaginaryUnit]);
  Test('LeadingPrincipalSubmatrix(A, 2)', 2, [ImaginaryUnit, 2, 0, 1]);
  Test('LeadingPrincipalSubmatrix(A, 3)', 3, [ImaginaryUnit, 2, 3, 0, 1, 8, 9, 3, 5]);
  Test('LeadingPrincipalSubmatrix(A, 4)', 4, [ImaginaryUnit, 2, 3, 6, 0, 1, 8, 7, 9, 3, 5, 4, 1, 7, 5, 2]);
  Test('LeadingPrincipalSubmatrix(A, 5)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(B, 1)', 1, [ImaginaryUnit]);
  Test('LeadingPrincipalSubmatrix(B, 2)', 2, [ImaginaryUnit, 6, 7, 4]);
  Test('LeadingPrincipalSubmatrix(B, 3)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(C, 1)', 1, [ImaginaryUnit]);
  Test('LeadingPrincipalSubmatrix(C, 2)', 2, [ImaginaryUnit, 4, 7, 5]);
  Test('LeadingPrincipalSubmatrix(C, 3)', 3, [ImaginaryUnit, 4, 0, 7, 5, 6, 7, 8, 5]);
  Test('LeadingPrincipalSubmatrix(C, 4)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(D, 1)', 1, [ImaginaryUnit]);
  Test('LeadingPrincipalSubmatrix(D, 2)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(E, 1)', 1, [ImaginaryUnit]);
  Test('LeadingPrincipalSubmatrix(E, 2)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('LeadingPrincipalSubmatrix(F, 1)', 1, [ImaginaryUnit]);
  Test('LeadingPrincipalSubmatrix(F, 2)', failure, 'Invalid index arrays passed to function Submatrix.');

  Test('delete(A); delete(B); delete(C); delete(D); delete(E); delete(F)', success);





  //
  //
  //  CHAPTER 7
  //  String functions.
  //
  //

  Chapter('String functions');

  //
  //  Basic string functions
  //

  // Length

  Test('length("Hello, World!")', 13);
  Test('length("It does not do to dwell on dreams, and forget to live.")', 54);
  Test('length("")', 0);
  Test('length("∫⌬dx")', 4);

  _s := LoremIpsum(100);
  Test('length("' + _s + '")', _s.Length);

  Test('s ≔ "Hello, World!"; length(s)', 13);
  Test('s ≔ "It does not do to dwell on dreams, and forget to live."; length(s)', 54);
  Test('s ≔ ""; length(s)', 0);
  Test('s ≔ "∫⌬dx"; length(s)', 4);
  Test('s ≔ "' + _s + '"; length(s)', _s.Length);

  Test('#"Hello, World!"', 13);
  Test('#"It does not do to dwell on dreams, and forget to live."', 54);
  Test('#""', 0);
  Test('#"∫⌬dx"', 4);

  _s := LoremIpsum(100);
  Test('#"' + _s + '"', _s.Length);

  Test('s ≔ "Hello, World!"; #s', 13);
  Test('s ≔ "It does not do to dwell on dreams, and forget to live."; #s', 54);
  Test('s ≔ ""; #s', 0);
  Test('s ≔ "∫⌬dx"; #s', 4);
  Test('s ≔ "' + _s + '"; #s', _s.Length);

  _s := '';
  Test('LoremStr ≔ s; delete(s)', success);

  // Reversal

  Test('reverse("Hello, World!")', '!dlroW ,olleH');
  Test('reverse("PARIS")', 'SIRAP');
  Test('reverse("")', '');
  Test('reverse("A")', 'A');
  Test('reverse("∫⌬dx")', 'xd⌬∫');
  Test('reverse("erised")', 'desire');

  Test('length(reverse(LoremStr)) = length(LoremStr)', True);

  // Shuffling

  Test('s ≔ "Hello, World!"; s2 ≔ shuffle(s); #s = #s2 ∧ ToSet(s) = ToSet(s2)', True);
  Test('s ≔ "It does not do to dwell on dreams, and forget to live."; s2 ≔ shuffle(s); #s = #s2 ∧ ToSet(s) = ToSet(s2)', True);
  Test('s ≔ ""; s2 ≔ shuffle(s); #s = #s2 ∧ ToSet(s) = ToSet(s2)', True);
  Test('s ≔ "∫⌬dx"; s2 ≔ shuffle(s); #s = #s2 ∧ ToSet(s) = ToSet(s2)', True);
  Test('s ≔ LoremStr; s2 ≔ shuffle(s); #s = #s2 ∧ ToSet(s) = ToSet(s2)', True);
  Test('delete(s)', success);

  // Substring search

  Test('StringPos("test", "t")', 1);
  Test('StringPos("test", "te")', 1);
  Test('StringPos("test", "tes")', 1);
  Test('StringPos("test", "test")', 1);
  Test('StringPos("test", "testa")', 0);
  Test('StringPos("test", "katt")', 0);
  Test('StringPos("test", "e")', 2);
  Test('StringPos("test", "st")', 3);
  Test('StringPos("test", "st", "w")', 0);
  Test('StringPos("test", "ST")', 0);
  Test('StringPos("test", "ST", "i")', 3);
  Test('StringPos("test", "ST", "iw")', 0);
  Test('StringPos("test", "ST", "wi")', 0);
  Test('StringPos("test", "TEST", "wi")', 1);
  Test('StringPos("test", "TEST", "w")', 0);
  Test('StringPos("test", "t", "")', 1);
  Test('StringPos("test", "t", "", 1)', 1);
  Test('StringPos("test", "t", "", 2)', 4);
  Test('StringPos("test", "t", "", 5)', 0);
  Test('StringPos("test", "")', 0);
  Test('StringPos("", "t")', 0);
  Test('StringPos("", "")', 0);
  Test('StringPos("test")', failure, 'Too few arguments. A required argument of type string is missing.');
  Test('StringPos("test", "t", "", 1, 2)', failure, 'Too many arguments.');
  Test('StringPos("test", "t", "", "cat")', failure, 'An object of type integer was expected as argument 4, but an object of type string was given.');
  Test('StringPos("test", "t", 1)', failure, 'An object of type string was expected as argument 3, but an object of type integer was given.');
  Test('StringPos("test", 2)', failure, 'An object of type string was expected as argument 2, but an object of type integer was given.');
  Test('StringPos(5, "t")', failure, 'An object of type string was expected as argument 1, but an object of type integer was given.');

  Test('LoremStr ≔ "' + LoremIpsum(1) + '";', null);

  // 'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor ' +
  // 'incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis ' +
  // 'nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. ' +
  // 'Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu ' +
  // 'fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in ' +
  // 'culpa qui officia deserunt mollit anim id est laborum.';

  Test('StringPos(LoremStr, "con")', 29);
  Test('StringPos(LoremStr, "laboris")', 184);
  Test('StringPos(LoremStr, "Lorem")', 1);
  Test('StringPos(LoremStr, "laborum.")', 438);
  Test('StringPos(LoremStr, ".")', 123);
  Test('StringPos(LoremStr, " ")', 6);
  Test('StringPos(LoremStr, "..")', 0);
  Test('StringPos(LoremStr, "acetylkolinesteras")', 0);

  Test('StringPos(LoremStr, "LOREM")', 0);
  Test('StringPos(LoremStr, "lorem")', 0);
  Test('StringPos(LoremStr, "DOLOR")', 0);

  Test('StringPos(LoremStr, "LOREM", "i")', 1);
  Test('StringPos(LoremStr, "lorem", "i")', 1);
  Test('StringPos(LoremStr, "DOLOR", "i")', 13);
  Test('StringPos(LoremStr, "ELIT", "i")', 52);
  Test('StringPos(LoremStr, "ELIT")', 0);

  Test('StringPos(LoremStr, "con")', 29);
  Test('StringPos(LoremStr, "con", "i")', 29);
  Test('StringPos(LoremStr, "con", "w")', 0);
  Test('StringPos(LoremStr, "con", "iw")', 0);
  Test('StringPos(LoremStr, "con", "wi")', 0);

  Test('StringPos(LoremStr, "do")', 13);
  Test('StringPos(LoremStr, "do", "w")', 62);

  Test('StringPos(LoremStr, "Pariat", "iw")', 0);
  Test('StringPos(LoremStr, "Pariat", "wi")', 0);
  Test('StringPos(LoremStr, "Pariat", "i")', 326);
  Test('StringPos(LoremStr, "Pariat", "w")', 0);
  Test('StringPos(LoremStr, "Pariat", "")', 0);

  Test('StringPos(LoremStr, "IN", "i")', 48);
  Test('StringPos(LoremStr, "IN", "iw")', 255);
  Test('StringPos(LoremStr, "IN", "wi")', 255);
  Test('StringPos(LoremStr, "IN", "")', 0);
  Test('StringPos(LoremStr, "IN", "w")', 0);

  Test('StringPos(LoremStr, "in")', 48);
  Test('StringPos(LoremStr, "in", "")', 48);
  Test('StringPos(LoremStr, "in", "", 47)', 48);
  Test('StringPos(LoremStr, "in", "", 48)', 48);
  Test('StringPos(LoremStr, "in", "", 49)', 80);
  Test('StringPos(LoremStr, "in", "", 79)', 80);
  Test('StringPos(LoremStr, "in", "", 80)', 80);
  Test('StringPos(LoremStr, "in", "", 81)', 137);
  Test('StringPos(LoremStr, "in", "", 136)', 137);
  Test('StringPos(LoremStr, "in", "", 137)', 137);
  Test('StringPos(LoremStr, "in", "", 138)', 255);
  Test('StringPos(LoremStr, "in", "", 254)', 255);
  Test('StringPos(LoremStr, "in", "", 255)', 255);
  Test('StringPos(LoremStr, "in", "", 256)', 272);
  Test('StringPos(LoremStr, "in", "", 271)', 272);
  Test('StringPos(LoremStr, "in", "", 272)', 272);
  Test('StringPos(LoremStr, "in", "", 273)', 347);
  Test('StringPos(LoremStr, "in", "", 346)', 347);
  Test('StringPos(LoremStr, "in", "", 347)', 347);
  Test('StringPos(LoremStr, "in", "", 348)', 389);
  Test('StringPos(LoremStr, "in", "", 388)', 389);
  Test('StringPos(LoremStr, "in", "", 389)', 389);
  Test('StringPos(LoremStr, "in", "", 390)', 0);
  Test('StringPos(LoremStr, "in", "", 500)', 0);
  Test('StringPos(LoremStr, "in", "", 5000)', 0);
  Test('StringPos(LoremStr, "in", "", 50000)', 0);

  Test('StringPos(LoremStr, "IN")', 0);
  Test('StringPos(LoremStr, "IN", "")', 0);
  Test('StringPos(LoremStr, "IN", "", 47)', 0);
  Test('StringPos(LoremStr, "IN", "", 48)', 0);
  Test('StringPos(LoremStr, "IN", "", 49)', 0);
  Test('StringPos(LoremStr, "IN", "", 79)', 0);
  Test('StringPos(LoremStr, "IN", "", 80)', 0);
  Test('StringPos(LoremStr, "IN", "", 81)', 0);
  Test('StringPos(LoremStr, "IN", "", 136)', 0);
  Test('StringPos(LoremStr, "IN", "", 137)', 0);
  Test('StringPos(LoremStr, "IN", "", 138)', 0);
  Test('StringPos(LoremStr, "IN", "", 254)', 0);
  Test('StringPos(LoremStr, "IN", "", 255)', 0);
  Test('StringPos(LoremStr, "IN", "", 256)', 0);
  Test('StringPos(LoremStr, "IN", "", 271)', 0);
  Test('StringPos(LoremStr, "IN", "", 272)', 0);
  Test('StringPos(LoremStr, "IN", "", 273)', 0);
  Test('StringPos(LoremStr, "IN", "", 346)', 0);
  Test('StringPos(LoremStr, "IN", "", 347)', 0);
  Test('StringPos(LoremStr, "IN", "", 348)', 0);
  Test('StringPos(LoremStr, "IN", "", 388)', 0);
  Test('StringPos(LoremStr, "IN", "", 389)', 0);
  Test('StringPos(LoremStr, "IN", "", 390)', 0);
  Test('StringPos(LoremStr, "IN", "", 500)', 0);
  Test('StringPos(LoremStr, "IN", "", 5000)', 0);
  Test('StringPos(LoremStr, "IN", "", 50000)', 0);

  Test('StringPos(LoremStr, "IN")', 0);
  Test('StringPos(LoremStr, "IN", "i")', 48);
  Test('StringPos(LoremStr, "IN", "i", 47)', 48);
  Test('StringPos(LoremStr, "IN", "i", 48)', 48);
  Test('StringPos(LoremStr, "IN", "i", 49)', 80);
  Test('StringPos(LoremStr, "IN", "i", 79)', 80);
  Test('StringPos(LoremStr, "IN", "i", 80)', 80);
  Test('StringPos(LoremStr, "IN", "i", 81)', 137);
  Test('StringPos(LoremStr, "IN", "i", 136)', 137);
  Test('StringPos(LoremStr, "IN", "i", 137)', 137);
  Test('StringPos(LoremStr, "IN", "i", 138)', 255);
  Test('StringPos(LoremStr, "IN", "i", 254)', 255);
  Test('StringPos(LoremStr, "IN", "i", 255)', 255);
  Test('StringPos(LoremStr, "IN", "i", 256)', 272);
  Test('StringPos(LoremStr, "IN", "i", 271)', 272);
  Test('StringPos(LoremStr, "IN", "i", 272)', 272);
  Test('StringPos(LoremStr, "IN", "i", 273)', 347);
  Test('StringPos(LoremStr, "IN", "i", 346)', 347);
  Test('StringPos(LoremStr, "IN", "i", 347)', 347);
  Test('StringPos(LoremStr, "IN", "i", 348)', 389);
  Test('StringPos(LoremStr, "IN", "i", 388)', 389);
  Test('StringPos(LoremStr, "IN", "i", 389)', 389);
  Test('StringPos(LoremStr, "IN", "i", 390)', 0);
  Test('StringPos(LoremStr, "IN", "i", 500)', 0);
  Test('StringPos(LoremStr, "IN", "i", 5000)', 0);
  Test('StringPos(LoremStr, "IN", "i", 50000)', 0);

  Test('StringPos(LoremStr, "IN")', 0);
  Test('StringPos(LoremStr, "IN", "i")', 48);
  Test('StringPos(LoremStr, "IN", "iw")', 255);
  Test('StringPos(LoremStr, "IN", "iw", 254)', 255);
  Test('StringPos(LoremStr, "IN", "wi", 255)', 255);
  Test('StringPos(LoremStr, "IN", "iw", 256)', 272);
  Test('StringPos(LoremStr, "IN", "iw", 271)', 272);
  Test('StringPos(LoremStr, "IN", "iw", 272)', 272);
  Test('StringPos(LoremStr, "IN", "wi", 273)', 389);
  Test('StringPos(LoremStr, "IN", "iw", 288)', 389);
  Test('StringPos(LoremStr, "IN", "iw", 389)', 389);
  Test('StringPos(LoremStr, "IN", "wi", 390)', 0);

  Test('StringPos(LoremStr, "IN")', 0);
  Test('StringPos(LoremStr, "IN", "")', 0);
  Test('StringPos(LoremStr, "IN", "w")', 0);
  Test('StringPos(LoremStr, "IN", "w", 254)', 0);
  Test('StringPos(LoremStr, "IN", "w", 255)', 0);
  Test('StringPos(LoremStr, "IN", "w", 256)', 0);
  Test('StringPos(LoremStr, "IN", "w", 271)', 0);
  Test('StringPos(LoremStr, "IN", "w", 272)', 0);
  Test('StringPos(LoremStr, "IN", "w", 273)', 0);
  Test('StringPos(LoremStr, "IN", "w", 288)', 0);
  Test('StringPos(LoremStr, "IN", "w", 389)', 0);
  Test('StringPos(LoremStr, "IN", "w", 390)', 0);

  Test('StringPos(LoremStr, "in")', 48);
  Test('StringPos(LoremStr, "in", "w")', 255);
  Test('StringPos(LoremStr, "in", "w", 254)', 255);
  Test('StringPos(LoremStr, "in", "w", 255)', 255);
  Test('StringPos(LoremStr, "in", "w", 256)', 272);
  Test('StringPos(LoremStr, "in", "w", 271)', 272);
  Test('StringPos(LoremStr, "in", "w", 272)', 272);
  Test('StringPos(LoremStr, "in", "w", 273)', 389);
  Test('StringPos(LoremStr, "in", "w", 288)', 389);
  Test('StringPos(LoremStr, "in", "w", 389)', 389);
  Test('StringPos(LoremStr, "in", "w", 390)', 0);

  Test('StringPos(LoremStr, "")', 0);
  Test('StringPos(LoremStr, "", "i")', 0);
  Test('StringPos(LoremStr, "", "w")', 0);
  Test('StringPos(LoremStr, "", "iw")', 0);
  Test('StringPos(LoremStr, "", "wi")', 0);
  Test('StringPos(LoremStr, "", "", 20)', 0);
  Test('StringPos(LoremStr, "", "i", 20)', 0);
  Test('StringPos(LoremStr, "", "w", 20)', 0);
  Test('StringPos(LoremStr, "", "iw", 20)', 0);
  Test('StringPos(LoremStr, "", "wi", 20)', 0);
  Test('StringPos(LoremStr, "", "", 5000)', 0);
  Test('StringPos(LoremStr, "", "i", 5000)', 0);
  Test('StringPos(LoremStr, "", "w", 5000)', 0);
  Test('StringPos(LoremStr, "", "iw", 5000)', 0);
  Test('StringPos(LoremStr, "", "wi", 5000)', 0);

  Test('StringPos("", "in")', 0);
  Test('StringPos("", "in", "i")', 0);
  Test('StringPos("", "in", "w")', 0);
  Test('StringPos("", "in", "iw")', 0);
  Test('StringPos("", "in", "wi")', 0);
  Test('StringPos("", "in", "", 20)', 0);
  Test('StringPos("", "in", "i", 20)', 0);
  Test('StringPos("", "in", "w", 20)', 0);
  Test('StringPos("", "in", "iw", 20)', 0);
  Test('StringPos("", "in", "wi", 20)', 0);
  Test('StringPos("", "in", "", 5000)', 0);
  Test('StringPos("", "in", "i", 5000)', 0);
  Test('StringPos("", "in", "w", 5000)', 0);
  Test('StringPos("", "in", "iw", 5000)', 0);
  Test('StringPos("", "in", "wi", 5000)', 0);

  Test('StringPos("", "")', 0);
  Test('StringPos("", "", "i")', 0);
  Test('StringPos("", "", "w")', 0);
  Test('StringPos("", "", "iw")', 0);
  Test('StringPos("", "", "wi")', 0);
  Test('StringPos("", "", "", 20)', 0);
  Test('StringPos("", "", "i", 20)', 0);
  Test('StringPos("", "", "w", 20)', 0);
  Test('StringPos("", "", "iw", 20)', 0);
  Test('StringPos("", "", "wi", 20)', 0);
  Test('StringPos("", "", "", 5000)', 0);
  Test('StringPos("", "", "i", 5000)', 0);
  Test('StringPos("", "", "w", 5000)', 0);
  Test('StringPos("", "", "iw", 5000)', 0);
  Test('StringPos("", "", "wi", 5000)', 0);

  Test('StringPos(LoremStr, "Ex")', 336);
  Test('StringPos(LoremStr, "Ex", "i")', 163);
  Test('StringPos(LoremStr, "Ex", "iw")', 208);
  Test('StringPos(LoremStr, "Ex", "wi")', 208);
  Test('StringPos(LoremStr, "Ex", "w")', 0);

  Test('StringPos(LoremStr, "lorem", "j")', failure, 'Unsupported character "j".');
  Test('StringPos(LoremStr, "lorem", "c")', failure, 'Unsupported character "c".');
  Test('StringPos(LoremStr, "lorem", "iwb")', failure, 'Unsupported character "b".');
  Test('StringPos(LoremStr, "lorem", "ibw")', failure, 'Unsupported character "b".');
  Test('StringPos(LoremStr, "lorem", "biw")', failure, 'Unsupported character "b".');

  Test('StringPos(LoremStr, "lorem", "i", 0)', failure, 'A positive integer was expected as argument 4, but "0" was given.');
  Test('StringPos(LoremStr, "lorem", "i", -1)', failure, 'A positive integer was expected as argument 4, but "-1" was given.');

  Test('StringPos("t", "test")', 0);
  Test('StringPos("t", "test", "iw")', 0);
  Test('StringPos("t", "test", "iw", 5)', 0);

  _s := LoremIpsum(1000);
  Test('LoremStr ≔ "' + _s + ' nargle";', null);

  StartStopwatch;
  Test('StringPos(LoremStr, "nargle")', _s.Length + 2);
  Test('StringPos(LoremStr, "nargl")', _s.Length + 2);
  StopStopwatch;
  TestDuration(0.02);

  StartStopwatch;
  Test('StringPos(LoremStr, "nargle", "w")', _s.Length + 2);
  Test('StringPos(LoremStr, "nargl", "w")', 0);
  StopStopwatch;
  TestDuration(0.04);

  StartStopwatch;
  Test('StringPos(LoremStr, "Nargle", "wi")', _s.Length + 2);
  Test('StringPos(LoremStr, "Nargle", "w")', 0);
  StopStopwatch;
  TestDuration(0.04);

  Test('StringPos(LoremStr, "laborum." + chr(13) + chr(10) + "Lorem")', 438);

  Test('delete(LoremStr)', success);
  _s := '';

  Test('s ≔ "A cat is a good/excellent" + chr(13) + chr(10) + "mister. –Indeed!"', 'A cat is a good/excellent'#13#10'mister. –Indeed!');

  Test('StringPos(s, "a")', 4);
  Test('StringPos(s, "a", "i")', 1);
  Test('StringPos(s, "a", "", 3)', 4);
  Test('StringPos(s, "a", "w", 3)', 10);
  Test('StringPos(s, "go")', 12);
  Test('StringPos(s, "go", "w")', 0);
  Test('StringPos(s, "good", "")', 12);
  Test('StringPos(s, "good", "w")', 12);
  Test('StringPos(s, "excel")', 17);
  Test('StringPos(s, "excel", "w")', 0);
  Test('StringPos(s, "excellent")', 17);
  Test('StringPos(s, "excellent", "w")', 17);
  Test('StringPos(s, "mist")', 28);
  Test('StringPos(s, "mist", "w")', 0);
  Test('StringPos(s, "mister")', 28);
  Test('StringPos(s, "mister", "w")', 28);
  Test('StringPos(s, "deed")', 39);
  Test('StringPos(s, "deed", "w")', 0);
  Test('StringPos(s, "indeed")', 0);
  Test('StringPos(s, "indeed", "i")', 37);
  Test('StringPos(s, "indeed", "iw")', 37);

  Test('s ≔ "Överskatta inte köpmannens förkärlek för sångfågeln."', 'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringPos(s, "skatt")', 5);
  Test('StringPos(s, "skatt", "w")', 0);
  Test('StringPos(s, "överskatta", "")', 0);
  Test('StringPos(s, "överskatta", "i")', 1);
  Test('StringPos(s, "överskatta", "iw")', 1);
  Test('StringPos(s, "köpman")', 17);
  Test('StringPos(s, "köpman", "w")', 0);
  Test('StringPos(s, "köpmannens")', 17);
  Test('StringPos(s, "köpmannens", "w")', 17);
  Test('StringPos(s, "kärlek")', 31);
  Test('StringPos(s, "kärlek", "w")', 0);
  Test('StringPos(s, "förkärlek")', 28);
  Test('StringPos(s, "förkärlek", "w")', 28);
  Test('StringPos(s, "sång")', 42);
  Test('StringPos(s, "sång", "w")', 0);
  Test('StringPos(s, "sångfågeln")', 42);
  Test('StringPos(s, "sångfågeln", "w")', 42);
  Test('StringPos(s, "fågel")', 46);
  Test('StringPos(s, "fågel", "w")', 0);
  Test('StringPos(s, "fågeln")', 46);
  Test('StringPos(s, "fågeln", "w")', 0);

  Test('s ≔ "Ansökan om måsten?"', 'Ansökan om måsten?');

  Test('StringPos(s, "ansökan")', 0);
  Test('StringPos(s, "ansökan", "i")', 1);
  Test('StringPos(s, "ansökan", "iw")', 1);
  Test('StringPos(s, "kan")', 5);
  Test('StringPos(s, "kan", "w")', 0);
  Test('StringPos(s, "måste")', 12);
  Test('StringPos(s, "måste", "w")', 0);
  Test('StringPos(s, "måsten")', 12);
  Test('StringPos(s, "måsten", "w")', 12);
  Test('StringPos(s, "sten")', 14);
  Test('StringPos(s, "sten", "w")', 0);

  Test('s ≔ "aaaaaaAAAAAAA"', 'aaaaaaAAAAAAA');

  Test('StringPos(s, "a")', 1);
  Test('StringPos(s, "a", "i")', 1);
  Test('StringPos(s, "a", "iw")', 0);
  Test('StringPos(s, "A")', 7);
  Test('StringPos(s, "A", "i")', 1);
  Test('StringPos(s, "A", "iw")', 0);

  Test('StringPos(s, "aa")', 1);
  Test('StringPos(s, "aaa")', 1);
  Test('StringPos(s, "aaaa")', 1);
  Test('StringPos(s, "aaaaa")', 1);
  Test('StringPos(s, "aaaaaa")', 1);
  Test('StringPos(s, "aaaaaaa")', 0);
  Test('StringPos(s, "aaaaaaaa")', 0);
  Test('StringPos(s, "aaaaaaaaa")', 0);
  Test('StringPos(s, "aaaaaaaaaa")', 0);
  Test('StringPos(s, "aaaaaaaaaaa")', 0);
  Test('StringPos(s, "aaaaaaaaaaaa")', 0);
  Test('StringPos(s, "aaaaaaaaaaaaa")', 0);
  Test('StringPos(s, "aaaaaaaaaaaaaa")', 0);

  Test('StringPos(s, "aa", "i")', 1);
  Test('StringPos(s, "aaa", "i")', 1);
  Test('StringPos(s, "aaaa", "i")', 1);
  Test('StringPos(s, "aaaaa", "i")', 1);
  Test('StringPos(s, "aaaaaa", "i")', 1);
  Test('StringPos(s, "aaaaaaa", "i")', 1);
  Test('StringPos(s, "aaaaaaaa", "i")', 1);
  Test('StringPos(s, "aaaaaaaaa", "i")', 1);
  Test('StringPos(s, "aaaaaaaaaa", "i")', 1);
  Test('StringPos(s, "aaaaaaaaaaa", "i")', 1);
  Test('StringPos(s, "aaaaaaaaaaaa", "i")', 1);
  Test('StringPos(s, "aaaaaaaaaaaaa", "i")', 1);
  Test('StringPos(s, "aaaaaaaaaaaaaa", "i")', 0);

  Test('StringPos(s, "a", "", 1)', 1);
  Test('StringPos(s, "a", "", 2)', 2);
  Test('StringPos(s, "a", "", 3)', 3);
  Test('StringPos(s, "a", "", 4)', 4);
  Test('StringPos(s, "a", "", 5)', 5);
  Test('StringPos(s, "a", "", 6)', 6);
  Test('StringPos(s, "a", "", 7)', 0);

  Test('StringPos(s, "aa", "", 1)', 1);
  Test('StringPos(s, "aa", "", 2)', 2);
  Test('StringPos(s, "aa", "", 3)', 3);
  Test('StringPos(s, "aa", "", 4)', 4);
  Test('StringPos(s, "aa", "", 5)', 5);
  Test('StringPos(s, "aa", "", 6)', 0);
  Test('StringPos(s, "aa", "", 7)', 0);

  Test('StringPos(s, "aaa", "", 1)', 1);
  Test('StringPos(s, "aaa", "", 2)', 2);
  Test('StringPos(s, "aaa", "", 3)', 3);
  Test('StringPos(s, "aaa", "", 4)', 4);
  Test('StringPos(s, "aaa", "", 5)', 0);
  Test('StringPos(s, "aaa", "", 6)', 0);
  Test('StringPos(s, "aaa", "", 7)', 0);

  Test('StringPos(s, "aaaa", "", 1)', 1);
  Test('StringPos(s, "aaaa", "", 2)', 2);
  Test('StringPos(s, "aaaa", "", 3)', 3);
  Test('StringPos(s, "aaaa", "", 4)', 0);
  Test('StringPos(s, "aaaa", "", 5)', 0);
  Test('StringPos(s, "aaaa", "", 6)', 0);
  Test('StringPos(s, "aaaa", "", 7)', 0);


  // -

  Test('LoremStr ≔ "' + LoremIpsum(1) + '";', null);

  Test('StringContains(LoremStr, "Lorem")', True);
  Test('StringContains(LoremStr, "ipsum")', True);
  Test('StringContains(LoremStr, "tempor")', True);
  Test('StringContains(LoremStr, "nostrud")', True);
  Test('StringContains(LoremStr, "pariatur")', True);
  Test('StringContains(LoremStr, "Duis aute irure dolor in reprehenderit in voluptate")', True);
  Test('StringContains(LoremStr, "horcrux")', False);
  Test('StringContains(LoremStr, "diadem")', False);
  Test('StringContains(LoremStr, "diary")', False);
  Test('StringContains(LoremStr, "Defence Against the Dark Arts")', False);

  Test('StringContains(LoremStr, "lorem")', False);
  Test('StringContains(LoremStr, "lorem", "")', False);
  Test('StringContains(LoremStr, "lorem", "i")', True);

  Test('StringContains(LoremStr, "LABORIS")', False);
  Test('StringContains(LoremStr, "LABORIS", "")', False);
  Test('StringContains(LoremStr, "LABORIS", "i")', True);

  Test('StringContains(LoremStr, "mini")', True);
  Test('StringContains(LoremStr, "mini", "")', True);
  Test('StringContains(LoremStr, "mini", "w")', False);
  Test('StringContains(LoremStr, "minim", "w")', True);

  Test('StringContains(LoremStr, "MOLL")', False);
  Test('StringContains(LoremStr, "MOLL", "")', False);
  Test('StringContains(LoremStr, "MOLL", "i")', True);
  Test('StringContains(LoremStr, "MOLL", "iw")', False);
  Test('StringContains(LoremStr, "MOLL", "wi")', False);
  Test('StringContains(LoremStr, "MOLL", "w")', False);

  Test('StringContains(LoremStr, "in")', True);
  Test('StringContains(LoremStr, "in", "", 388)', True);
  Test('StringContains(LoremStr, "in", "", 389)', True);
  Test('StringContains(LoremStr, "in", "", 390)', False);
  Test('StringContains(LoremStr, "in", "", 1000000)', False);

  Test('StringContains(LoremStr, "ut")', True);
  Test('StringContains(LoremStr, "ut", "", 238)', True);
  Test('StringContains(LoremStr, "ut", "", 239)', True);
  Test('StringContains(LoremStr, "ut", "", 240)', False);
  Test('StringContains(LoremStr, "ut", "", 240000)', False);

  Test('StringContains(LoremStr, "ut", "w")', True);
  Test('StringContains(LoremStr, "ut", "w", 196)', True);
  Test('StringContains(LoremStr, "ut", "w", 197)', True);
  Test('StringContains(LoremStr, "ut", "w", 198)', False);
  Test('StringContains(LoremStr, "ut", "w", 238)', False);
  Test('StringContains(LoremStr, "ut", "w", 239)', False);
  Test('StringContains(LoremStr, "ut", "w", 240)', False);
  Test('StringContains(LoremStr, "ut", "w", 24000)', False);

  Test('StringContains(LoremStr, "Ex")', True);
  Test('StringContains(LoremStr, "Ex", "")', True);
  Test('StringContains(LoremStr, "Ex", "w")', False);
  Test('StringContains(LoremStr, "Ex", "wi")', True);
  Test('StringContains(LoremStr, "Ex", "iw")', True);
  Test('StringContains(LoremStr, "Ex", "i")', True);
  Test('StringContains(LoremStr, "Ex", "", 233)', True);
  Test('StringContains(LoremStr, "Ex", "w", 233)', False);
  Test('StringContains(LoremStr, "Ex", "w", 200)', False);
  Test('StringContains(LoremStr, "Ex", "wi", 200)', True);

  Test('StringContains(LoremStr, "")', False);
  Test('StringContains("", "ipsum")', False);
  Test('StringContains("", "")', False);

  Test('StringContains(LoremStr, "", "i")', False);
  Test('StringContains("", "ipsum", "i")', False);
  Test('StringContains("", "", "i")', False);

  Test('StringContains(LoremStr, "", "w")', False);
  Test('StringContains("", "ipsum", "w")', False);
  Test('StringContains("", "", "w")', False);

  Test('StringContains(LoremStr, "", "wi")', False);
  Test('StringContains("", "ipsum", "wi")', False);
  Test('StringContains("", "", "wi")', False);

  Test('StringContains(LoremStr, "in", "c")', failure, 'Unsupported character "c".');
  Test('StringContains(LoremStr, "in", "iwc")', failure, 'Unsupported character "c".');
  Test('StringContains(LoremStr, "in", "iw", "cat")', failure, 'An object of type integer was expected as argument 4, but an object of type string was given.');
  Test('StringContains(LoremStr, "in", 20, 50)', failure, 'An object of type string was expected as argument 3, but an object of type integer was given.');

  _s := LoremIpsum(1000);
  Test('LoremStr ≔ "' + _s + ' nargle";', null);

  Test('StringContains(LoremStr, "nargle")', True);
  Test('StringContains(LoremStr, "laborum." + chr(13) + chr(10) + "Lorem")', True);

  Test('delete(LoremStr)', success);
  _s := '';

  Test('s ≔ "A cat is a good/excellent" + chr(13) + chr(10) + "mister. –Indeed!"', 'A cat is a good/excellent'#13#10'mister. –Indeed!');

  Test('StringContains(s, "a")', True);
  Test('StringContains(s, "a", "i")', True);
  Test('StringContains(s, "a", "", 3)', True);
  Test('StringContains(s, "a", "w", 3)', True);
  Test('StringContains(s, "go")', True);
  Test('StringContains(s, "go", "w")', False);
  Test('StringContains(s, "good", "")', True);
  Test('StringContains(s, "good", "w")', True);
  Test('StringContains(s, "excel")', True);
  Test('StringContains(s, "excel", "w")', False);
  Test('StringContains(s, "excellent")', True);
  Test('StringContains(s, "excellent", "w")', True);
  Test('StringContains(s, "mist")', True);
  Test('StringContains(s, "mist", "w")', False);
  Test('StringContains(s, "mister")', True);
  Test('StringContains(s, "mister", "w")', True);
  Test('StringContains(s, "deed")', True);
  Test('StringContains(s, "deed", "w")', False);
  Test('StringContains(s, "indeed")', False);
  Test('StringContains(s, "indeed", "i")', True);
  Test('StringContains(s, "indeed", "iw")', True);

  Test('s ≔ "Överskatta inte köpmannens förkärlek för sångfågeln."', 'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringContains(s, "skatt")', True);
  Test('StringContains(s, "skatt", "w")', False);
  Test('StringContains(s, "överskatta", "")', False);
  Test('StringContains(s, "överskatta", "i")', True);
  Test('StringContains(s, "överskatta", "iw")', True);
  Test('StringContains(s, "köpman")', True);
  Test('StringContains(s, "köpman", "w")', False);
  Test('StringContains(s, "köpmannens")', True);
  Test('StringContains(s, "köpmannens", "w")', True);
  Test('StringContains(s, "kärlek")', True);
  Test('StringContains(s, "kärlek", "w")', False);
  Test('StringContains(s, "förkärlek")', True);
  Test('StringContains(s, "förkärlek", "w")', True);
  Test('StringContains(s, "sång")', True);
  Test('StringContains(s, "sång", "w")', False);
  Test('StringContains(s, "sångfågeln")', True);
  Test('StringContains(s, "sångfågeln", "w")', True);
  Test('StringContains(s, "fågel")', True);
  Test('StringContains(s, "fågel", "w")', False);
  Test('StringContains(s, "fågeln")', True);
  Test('StringContains(s, "fågeln", "w")', False);

  Test('s ≔ "Ansökan om måsten?"', 'Ansökan om måsten?');

  Test('StringContains(s, "ansökan")', False);
  Test('StringContains(s, "ansökan", "i")', True);
  Test('StringContains(s, "ansökan", "iw")', True);
  Test('StringContains(s, "kan")', True);
  Test('StringContains(s, "kan", "w")', False);
  Test('StringContains(s, "måste")', True);
  Test('StringContains(s, "måste", "w")', False);
  Test('StringContains(s, "måsten")', True);
  Test('StringContains(s, "måsten", "w")', True);
  Test('StringContains(s, "sten")', True);
  Test('StringContains(s, "sten", "w")', False);

  Test('s ≔ "aaaaaaAAAAAAA"', 'aaaaaaAAAAAAA');

  Test('StringContains(s, "a")', True);
  Test('StringContains(s, "a", "i")', True);
  Test('StringContains(s, "a", "iw")', False);
  Test('StringContains(s, "A")', True);
  Test('StringContains(s, "A", "i")', True);
  Test('StringContains(s, "A", "iw")', False);

  Test('StringContains(s, "aa")', True);
  Test('StringContains(s, "aaa")', True);
  Test('StringContains(s, "aaaa")', True);
  Test('StringContains(s, "aaaaa")', True);
  Test('StringContains(s, "aaaaaa")', True);
  Test('StringContains(s, "aaaaaaa")', False);
  Test('StringContains(s, "aaaaaaaa")', False);
  Test('StringContains(s, "aaaaaaaaa")', False);
  Test('StringContains(s, "aaaaaaaaaa")', False);
  Test('StringContains(s, "aaaaaaaaaaa")', False);
  Test('StringContains(s, "aaaaaaaaaaaa")', False);
  Test('StringContains(s, "aaaaaaaaaaaaa")', False);
  Test('StringContains(s, "aaaaaaaaaaaaaa")', False);

  Test('StringContains(s, "aa", "i")', True);
  Test('StringContains(s, "aaa", "i")', True);
  Test('StringContains(s, "aaaa", "i")', True);
  Test('StringContains(s, "aaaaa", "i")', True);
  Test('StringContains(s, "aaaaaa", "i")', True);
  Test('StringContains(s, "aaaaaaa", "i")', True);
  Test('StringContains(s, "aaaaaaaa", "i")', True);
  Test('StringContains(s, "aaaaaaaaa", "i")', True);
  Test('StringContains(s, "aaaaaaaaaa", "i")', True);
  Test('StringContains(s, "aaaaaaaaaaa", "i")', True);
  Test('StringContains(s, "aaaaaaaaaaaa", "i")', True);
  Test('StringContains(s, "aaaaaaaaaaaaa", "i")', True);
  Test('StringContains(s, "aaaaaaaaaaaaaa", "i")', False);

  Test('StringContains(s, "a", "", 1)', True);
  Test('StringContains(s, "a", "", 2)', True);
  Test('StringContains(s, "a", "", 3)', True);
  Test('StringContains(s, "a", "", 4)', True);
  Test('StringContains(s, "a", "", 5)', True);
  Test('StringContains(s, "a", "", 6)', True);
  Test('StringContains(s, "a", "", 7)', False);

  Test('StringContains(s, "aa", "", 1)', True);
  Test('StringContains(s, "aa", "", 2)', True);
  Test('StringContains(s, "aa", "", 3)', True);
  Test('StringContains(s, "aa", "", 4)', True);
  Test('StringContains(s, "aa", "", 5)', True);
  Test('StringContains(s, "aa", "", 6)', False);
  Test('StringContains(s, "aa", "", 7)', False);

  Test('StringContains(s, "aaa", "", 1)', True);
  Test('StringContains(s, "aaa", "", 2)', True);
  Test('StringContains(s, "aaa", "", 3)', True);
  Test('StringContains(s, "aaa", "", 4)', True);
  Test('StringContains(s, "aaa", "", 5)', False);
  Test('StringContains(s, "aaa", "", 6)', False);
  Test('StringContains(s, "aaa", "", 7)', False);

  Test('StringContains(s, "aaaa", "", 1)', True);
  Test('StringContains(s, "aaaa", "", 2)', True);
  Test('StringContains(s, "aaaa", "", 3)', True);
  Test('StringContains(s, "aaaa", "", 4)', False);
  Test('StringContains(s, "aaaa", "", 5)', False);
  Test('StringContains(s, "aaaa", "", 6)', False);
  Test('StringContains(s, "aaaa", "", 7)', False);

  // -

  Test('LoremStr ≔ "' + LoremIpsum(1) + '";', null);

  Test('StringCount(LoremStr, "dolor")', 4);
  Test('StringCount(LoremStr, "et")', 3);
  Test('StringCount(LoremStr, "a")', 29);
  Test('StringCount(LoremStr, " ")', 68);

  Test('StringCount(LoremStr, "in")', 7);
  Test('StringCount(LoremStr, "in", "w")', 3);

  Test('StringCount(LoremStr, "ex")', 2);
  Test('StringCount(LoremStr, "ex", "i")', 3);
  Test('StringCount(LoremStr, "ex", "wi")', 1);
  Test('StringCount(LoremStr, "ex", "iw")', 1);
  Test('StringCount(LoremStr, "ex", "w")', 1);

  Test('StringCount(LoremStr, "Ex")', 1);
  Test('StringCount(LoremStr, "Ex", "i")', 3);
  Test('StringCount(LoremStr, "Ex", "wi")', 1);
  Test('StringCount(LoremStr, "Ex", "iw")', 1);
  Test('StringCount(LoremStr, "Ex", "w")', 0);

  Test('StringCount(LoremStr, "do")', 6);
  Test('StringCount(LoremStr, "do", "w")', 1);

  Test('StringCount(LoremStr, "")', 0);
  Test('StringCount("", "in")', 0);
  Test('StringCount("", "")', 0);

  Test('StringCount(LoremStr, "", "w")', 0);
  Test('StringCount("", "in", "w")', 0);
  Test('StringCount("", "", "w")', 0);

  Test('StringCount(LoremStr, "", "i")', 0);
  Test('StringCount("", "in", "i")', 0);
  Test('StringCount("", "", "i")', 0);

  Test('StringCount(LoremStr, "", "wi")', 0);
  Test('StringCount("", "in", "wi")', 0);
  Test('StringCount("", "", "wi")', 0);

  _s := LoremIpsum(1000);
  Test('LoremStr ≔ "' + _s + ' nargle";', null);

  Test('StringCount(LoremStr, "ipsum")', 1000);
  Test('StringCount(LoremStr, "et")', 3000);
  Test('StringCount(LoremStr, "et", "w")', 1000);

  Test('StringCount(LoremStr, "nargle")', 1);

  Test('StringCount(LoremStr, chr(13) + chr(10))', 999);
  Test('StringCount(LoremStr, chr(13) + chr(10), "i")', 999);
  Test('StringCount(LoremStr, "laborum." + chr(13) + chr(10) + "Lorem")', 999);
  Test('StringCount(LoremStr, "laborum." + chr(13) + chr(10) + "lorem")', 0);
  Test('StringCount(LoremStr, "laborum." + chr(13) + chr(10) + "lorem", "i")', 999);

  Test('s ≔ "A cat is a good/excellent" + chr(13) + chr(10) + "mister. –Indeed!"', 'A cat is a good/excellent'#13#10'mister. –Indeed!');

  Test('StringCount(s, "a")', 2);
  Test('StringCount(s, "a", "i")', 3);
  Test('StringCount(s, "a", "", 3)', 2);
  Test('StringCount(s, "a", "w", 3)', 1);
  Test('StringCount(s, "go")', 1);
  Test('StringCount(s, "go", "w")', 0);
  Test('StringCount(s, "good", "")', 1);
  Test('StringCount(s, "good", "w")', 1);
  Test('StringCount(s, "excel")', 1);
  Test('StringCount(s, "excel", "w")', 0);
  Test('StringCount(s, "excellent")', 1);
  Test('StringCount(s, "excellent", "w")', 1);
  Test('StringCount(s, "mist")', 1);
  Test('StringCount(s, "mist", "w")', 0);
  Test('StringCount(s, "mister")', 1);
  Test('StringCount(s, "mister", "w")', 1);
  Test('StringCount(s, "deed")', 1);
  Test('StringCount(s, "deed", "w")', 0);
  Test('StringCount(s, "indeed")', 0);
  Test('StringCount(s, "indeed", "i")', 1);
  Test('StringCount(s, "indeed", "iw")', 1);

  Test('s ≔ "Överskatta inte köpmannens förkärlek för sångfågeln."', 'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringCount(s, "skatt")', 1);
  Test('StringCount(s, "skatt", "w")', 0);
  Test('StringCount(s, "överskatta", "")', 0);
  Test('StringCount(s, "överskatta", "i")', 1);
  Test('StringCount(s, "överskatta", "iw")', 1);
  Test('StringCount(s, "köpman")', 1);
  Test('StringCount(s, "köpman", "w")', 0);
  Test('StringCount(s, "köpmannens")', 1);
  Test('StringCount(s, "köpmannens", "w")', 1);
  Test('StringCount(s, "kärlek")', 1);
  Test('StringCount(s, "kärlek", "w")', 0);
  Test('StringCount(s, "förkärlek")', 1);
  Test('StringCount(s, "förkärlek", "w")', 1);
  Test('StringCount(s, "sång")', 1);
  Test('StringCount(s, "sång", "w")', 0);
  Test('StringCount(s, "sångfågeln")', 1);
  Test('StringCount(s, "sångfågeln", "w")', 1);
  Test('StringCount(s, "fågel")', 1);
  Test('StringCount(s, "fågel", "w")', 0);
  Test('StringCount(s, "fågeln")', 1);
  Test('StringCount(s, "fågeln", "w")', 0);

  Test('s ≔ "Ansökan om måsten?"', 'Ansökan om måsten?');

  Test('StringCount(s, "ansökan")', 0);
  Test('StringCount(s, "ansökan", "i")', 1);
  Test('StringCount(s, "ansökan", "iw")', 1);
  Test('StringCount(s, "kan")', 1);
  Test('StringCount(s, "kan", "w")', 0);
  Test('StringCount(s, "måste")', 1);
  Test('StringCount(s, "måste", "w")', 0);
  Test('StringCount(s, "måsten")', 1);
  Test('StringCount(s, "måsten", "w")', 1);
  Test('StringCount(s, "sten")', 1);
  Test('StringCount(s, "sten", "w")', 0);

  Test('s ≔ "aaaaaaAAAAAAA"', 'aaaaaaAAAAAAA');

  Test('StringCount(s, "a")', 6);
  Test('StringCount(s, "a", "i")', 13);
  Test('StringCount(s, "a", "iw")', 0);
  Test('StringCount(s, "A")', 7);
  Test('StringCount(s, "A", "i")', 13);
  Test('StringCount(s, "A", "iw")', 0);

  Test('StringCount(s, "aa")', 3);
  Test('StringCount(s, "aaa")', 2);
  Test('StringCount(s, "aaaa")', 1);
  Test('StringCount(s, "aaaaa")', 1);
  Test('StringCount(s, "aaaaaa")', 1);
  Test('StringCount(s, "aaaaaaa")', 0);
  Test('StringCount(s, "aaaaaaaa")', 0);
  Test('StringCount(s, "aaaaaaaaa")', 0);
  Test('StringCount(s, "aaaaaaaaaa")', 0);
  Test('StringCount(s, "aaaaaaaaaaa")', 0);
  Test('StringCount(s, "aaaaaaaaaaaa")', 0);
  Test('StringCount(s, "aaaaaaaaaaaaa")', 0);
  Test('StringCount(s, "aaaaaaaaaaaaaa")', 0);

  Test('StringCount(s, "aa", "i")', 6);
  Test('StringCount(s, "aaa", "i")', 4);
  Test('StringCount(s, "aaaa", "i")', 3);
  Test('StringCount(s, "aaaaa", "i")', 2);
  Test('StringCount(s, "aaaaaa", "i")', 2);
  Test('StringCount(s, "aaaaaaa", "i")', 1);
  Test('StringCount(s, "aaaaaaaa", "i")', 1);
  Test('StringCount(s, "aaaaaaaaa", "i")', 1);
  Test('StringCount(s, "aaaaaaaaaa", "i")', 1);
  Test('StringCount(s, "aaaaaaaaaaa", "i")', 1);
  Test('StringCount(s, "aaaaaaaaaaaa", "i")', 1);
  Test('StringCount(s, "aaaaaaaaaaaaa", "i")', 1);
  Test('StringCount(s, "aaaaaaaaaaaaaa", "i")', 0);

  Test('StringCount(s, "a", "", 1)', 6);
  Test('StringCount(s, "a", "", 2)', 5);
  Test('StringCount(s, "a", "", 3)', 4);
  Test('StringCount(s, "a", "", 4)', 3);
  Test('StringCount(s, "a", "", 5)', 2);
  Test('StringCount(s, "a", "", 6)', 1);
  Test('StringCount(s, "a", "", 7)', 0);

  Test('StringCount(s, "aa", "", 1)', 3);
  Test('StringCount(s, "aa", "", 2)', 2);
  Test('StringCount(s, "aa", "", 3)', 2);
  Test('StringCount(s, "aa", "", 4)', 1);
  Test('StringCount(s, "aa", "", 5)', 1);
  Test('StringCount(s, "aa", "", 6)', 0);
  Test('StringCount(s, "aa", "", 7)', 0);

  Test('StringCount(s, "aaa", "", 1)', 2);
  Test('StringCount(s, "aaa", "", 2)', 1);
  Test('StringCount(s, "aaa", "", 3)', 1);
  Test('StringCount(s, "aaa", "", 4)', 1);
  Test('StringCount(s, "aaa", "", 5)', 0);
  Test('StringCount(s, "aaa", "", 6)', 0);
  Test('StringCount(s, "aaa", "", 7)', 0);

  Test('StringCount(s, "aaaa", "", 1)', 1);
  Test('StringCount(s, "aaaa", "", 2)', 1);
  Test('StringCount(s, "aaaa", "", 3)', 1);
  Test('StringCount(s, "aaaa", "", 4)', 0);
  Test('StringCount(s, "aaaa", "", 5)', 0);
  Test('StringCount(s, "aaaa", "", 6)', 0);
  Test('StringCount(s, "aaaa", "", 7)', 0);

  // -

  Test('LoremStr ≔ "' + LoremIpsum(1) + '";', null);

  Test('StringIndices(LoremStr, "dolor")', intarr([13, 104, 249, 303]));
  Test('StringIndices(LoremStr, "et")', intarr([25, 36, 101]));
  Test('StringIndices(LoremStr, "a")', intarr([23, 41, 95, 112, 115, 117, 122, 133, 146, 170, 179, 185, 200, 212, 229, 238, 281, 317, 324, 327, 330, 354, 357, 365, 367, 396, 408, 426, 439]));
  Test('StringIndices(LoremStr, " ")', intarr([6, 12, 18, 22, 28, 40, 51, 57, 61, 64, 72, 79, 90, 93, 100, 103, 110, 116, 124, 127, 132, 135, 141, 149, 154, 162, 175, 183, 191, 196, 199, 207, 210, 213, 221, 232, 237, 242, 248, 254, 257, 271, 274, 284, 290, 295, 302, 309, 312, 319, 325, 335, 345, 350, 359, 369, 373, 383, 388, 391, 397, 401, 409, 418, 425, 430, 433, 437]));

  Test('StringIndices(LoremStr, "in")', intarr([48, 80, 137, 255, 272, 347, 389]));
  Test('StringIndices(LoremStr, "in", "w")', intarr([255, 272, 389]));

  Test('StringIndices(LoremStr, "ex")', intarr([163, 208]));
  Test('StringIndices(LoremStr, "ex", "i")', intarr([163, 208, 336]));
  Test('StringIndices(LoremStr, "ex", "wi")', intarr([208]));
  Test('StringIndices(LoremStr, "ex", "iw")', intarr([208]));
  Test('StringIndices(LoremStr, "ex", "w")', intarr([208]));

  Test('StringIndices(LoremStr, "Ex")', intarr([336]));
  Test('StringIndices(LoremStr, "Ex", "i")', intarr([163, 208, 336]));
  Test('StringIndices(LoremStr, "Ex", "wi")', intarr([208]));
  Test('StringIndices(LoremStr, "Ex", "iw")', intarr([208]));
  Test('StringIndices(LoremStr, "Ex", "w")', intarr([]));

  Test('StringIndices(LoremStr, "do")', intarr([13, 62, 104, 219, 249, 303]));
  Test('StringIndices(LoremStr, "do", "w")', intarr([62]));

  Test('StringIndices(LoremStr, "")', intarr([]));
  Test('StringIndices("", "in")', intarr([]));
  Test('StringIndices("", "")', intarr([]));

  Test('StringIndices(LoremStr, "", "w")', intarr([]));
  Test('StringIndices("", "in", "w")', intarr([]));
  Test('StringIndices("", "", "w")', intarr([]));

  Test('StringIndices(LoremStr, "", "i")', intarr([]));
  Test('StringIndices("", "in", "i")', intarr([]));
  Test('StringIndices("", "", "i")', intarr([]));

  Test('StringIndices(LoremStr, "", "wi")', intarr([]));
  Test('StringIndices("", "in", "wi")', intarr([]));
  Test('StringIndices("", "", "wi")', intarr([]));

  _s := LoremIpsum(1000);
  Test('LoremStr ≔ "' + _s + ' nargle";', null);

  Test('StringIndices(LoremStr, "nargle")', intarr([_s.Length + 2]));

  Test('s ≔ "A cat is a good/excellent" + chr(13) + chr(10) + "mister. –Indeed!"', 'A cat is a good/excellent'#13#10'mister. –Indeed!');

  Test('StringIndices(s, "a")', intarr([4, 10]));
  Test('StringIndices(s, "a", "i")', intarr([1, 4, 10]));
  Test('StringIndices(s, "a", "", 3)', intarr([4, 10]));
  Test('StringIndices(s, "a", "w", 3)', intarr([10]));
  Test('StringIndices(s, "go")', intarr([12]));
  Test('StringIndices(s, "go", "w")', intarr([]));
  Test('StringIndices(s, "good", "")', intarr([12]));
  Test('StringIndices(s, "good", "w")', intarr([12]));
  Test('StringIndices(s, "excel")', intarr([17]));
  Test('StringIndices(s, "excel", "w")', intarr([]));
  Test('StringIndices(s, "excellent")', intarr([17]));
  Test('StringIndices(s, "excellent", "w")', intarr([17]));
  Test('StringIndices(s, "mist")', intarr([28]));
  Test('StringIndices(s, "mist", "w")', intarr([]));
  Test('StringIndices(s, "mister")', intarr([28]));
  Test('StringIndices(s, "mister", "w")', intarr([28]));
  Test('StringIndices(s, "deed")', intarr([39]));
  Test('StringIndices(s, "deed", "w")', intarr([]));
  Test('StringIndices(s, "indeed")', intarr([]));
  Test('StringIndices(s, "indeed", "i")', intarr([37]));
  Test('StringIndices(s, "indeed", "iw")', intarr([37]));

  Test('s ≔ "Överskatta inte köpmannens förkärlek för sångfågeln."', 'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringIndices(s, "skatt")', intarr([5]));
  Test('StringIndices(s, "skatt", "w")', intarr([]));
  Test('StringIndices(s, "överskatta", "")', intarr([]));
  Test('StringIndices(s, "överskatta", "i")', intarr([1]));
  Test('StringIndices(s, "överskatta", "iw")', intarr([1]));
  Test('StringIndices(s, "köpman")', intarr([17]));
  Test('StringIndices(s, "köpman", "w")', intarr([]));
  Test('StringIndices(s, "köpmannens")', intarr([17]));
  Test('StringIndices(s, "köpmannens", "w")', intarr([17]));
  Test('StringIndices(s, "kärlek")', intarr([31]));
  Test('StringIndices(s, "kärlek", "w")', intarr([]));
  Test('StringIndices(s, "förkärlek")', intarr([28]));
  Test('StringIndices(s, "förkärlek", "w")', intarr([28]));
  Test('StringIndices(s, "sång")', intarr([42]));
  Test('StringIndices(s, "sång", "w")', intarr([]));
  Test('StringIndices(s, "sångfågeln")', intarr([42]));
  Test('StringIndices(s, "sångfågeln", "w")', intarr([42]));
  Test('StringIndices(s, "fågel")', intarr([46]));
  Test('StringIndices(s, "fågel", "w")', intarr([]));
  Test('StringIndices(s, "fågeln")', intarr([46]));
  Test('StringIndices(s, "fågeln", "w")', intarr([]));

  Test('s ≔ "Ansökan om måsten?"', 'Ansökan om måsten?');

  Test('StringIndices(s, "ansökan")', intarr([]));
  Test('StringIndices(s, "ansökan", "i")', intarr([1]));
  Test('StringIndices(s, "ansökan", "iw")', intarr([1]));
  Test('StringIndices(s, "kan")', intarr([5]));
  Test('StringIndices(s, "kan", "w")', intarr([]));
  Test('StringIndices(s, "måste")', intarr([12]));
  Test('StringIndices(s, "måste", "w")', intarr([]));
  Test('StringIndices(s, "måsten")', intarr([12]));
  Test('StringIndices(s, "måsten", "w")', intarr([12]));
  Test('StringIndices(s, "sten")', intarr([14]));
  Test('StringIndices(s, "sten", "w")', intarr([]));

  Test('s ≔ "aaaaaaAAAAAAA"', 'aaaaaaAAAAAAA');

  Test('StringIndices(s, "a")', intarr([1, 2, 3, 4, 5, 6]));
  Test('StringIndices(s, "a", "i")', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]));
  Test('StringIndices(s, "a", "iw")', intarr([]));
  Test('StringIndices(s, "A")', intarr([7, 8, 9, 10, 11, 12, 13]));
  Test('StringIndices(s, "A", "i")', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]));
  Test('StringIndices(s, "A", "iw")', intarr([]));

  Test('StringIndices(s, "aa")', intarr([1, 3, 5]));
  Test('StringIndices(s, "aaa")', intarr([1, 4]));
  Test('StringIndices(s, "aaaa")', intarr([1]));
  Test('StringIndices(s, "aaaaa")', intarr([1]));
  Test('StringIndices(s, "aaaaaa")', intarr([1]));
  Test('StringIndices(s, "aaaaaaa")', intarr([]));
  Test('StringIndices(s, "aaaaaaaa")', intarr([]));
  Test('StringIndices(s, "aaaaaaaaa")', intarr([]));
  Test('StringIndices(s, "aaaaaaaaaa")', intarr([]));
  Test('StringIndices(s, "aaaaaaaaaaa")', intarr([]));
  Test('StringIndices(s, "aaaaaaaaaaaa")', intarr([]));
  Test('StringIndices(s, "aaaaaaaaaaaaa")', intarr([]));
  Test('StringIndices(s, "aaaaaaaaaaaaaa")', intarr([]));

  Test('StringIndices(s, "aa", "i")', intarr([1, 3, 5, 7, 9, 11]));
  Test('StringIndices(s, "aaa", "i")', intarr([1, 4, 7, 10]));
  Test('StringIndices(s, "aaaa", "i")', intarr([1, 5, 9]));
  Test('StringIndices(s, "aaaaa", "i")', intarr([1, 6]));
  Test('StringIndices(s, "aaaaaa", "i")', intarr([1, 7]));
  Test('StringIndices(s, "aaaaaaa", "i")', intarr([1]));
  Test('StringIndices(s, "aaaaaaaa", "i")', intarr([1]));
  Test('StringIndices(s, "aaaaaaaaa", "i")', intarr([1]));
  Test('StringIndices(s, "aaaaaaaaaa", "i")', intarr([1]));
  Test('StringIndices(s, "aaaaaaaaaaa", "i")', intarr([1]));
  Test('StringIndices(s, "aaaaaaaaaaaa", "i")', intarr([1]));
  Test('StringIndices(s, "aaaaaaaaaaaaa", "i")', intarr([1]));
  Test('StringIndices(s, "aaaaaaaaaaaaaa", "i")', intarr([]));

  Test('StringIndices(s, "a", "", 1)', intarr([1, 2, 3, 4, 5, 6]));
  Test('StringIndices(s, "a", "", 2)', intarr([2, 3, 4, 5, 6]));
  Test('StringIndices(s, "a", "", 3)', intarr([3, 4, 5, 6]));
  Test('StringIndices(s, "a", "", 4)', intarr([4, 5, 6]));
  Test('StringIndices(s, "a", "", 5)', intarr([5, 6]));
  Test('StringIndices(s, "a", "", 6)', intarr([6]));
  Test('StringIndices(s, "a", "", 7)', intarr([]));

  Test('StringIndices(s, "aa", "", 1)', intarr([1, 3, 5]));
  Test('StringIndices(s, "aa", "", 2)', intarr([2, 4]));
  Test('StringIndices(s, "aa", "", 3)', intarr([3, 5]));
  Test('StringIndices(s, "aa", "", 4)', intarr([4]));
  Test('StringIndices(s, "aa", "", 5)', intarr([5]));
  Test('StringIndices(s, "aa", "", 6)', intarr([]));
  Test('StringIndices(s, "aa", "", 7)', intarr([]));

  Test('StringIndices(s, "aaa", "", 1)', intarr([1, 4]));
  Test('StringIndices(s, "aaa", "", 2)', intarr([2]));
  Test('StringIndices(s, "aaa", "", 3)', intarr([3]));
  Test('StringIndices(s, "aaa", "", 4)', intarr([4]));
  Test('StringIndices(s, "aaa", "", 5)', intarr([]));
  Test('StringIndices(s, "aaa", "", 6)', intarr([]));
  Test('StringIndices(s, "aaa", "", 7)', intarr([]));

  Test('StringIndices(s, "aaaa", "", 1)', intarr([1]));
  Test('StringIndices(s, "aaaa", "", 2)', intarr([2]));
  Test('StringIndices(s, "aaaa", "", 3)', intarr([3]));
  Test('StringIndices(s, "aaaa", "", 4)', intarr([]));
  Test('StringIndices(s, "aaaa", "", 5)', intarr([]));
  Test('StringIndices(s, "aaaa", "", 6)', intarr([]));
  Test('StringIndices(s, "aaaa", "", 7)', intarr([]));

  // Replace all

  Test('s ≔ "A car named Carry was careful not to care about roundabouts being too round to care too much about cars with carry.";', null); // No, it's not supposed to make sense.

  Test('StringReplace(s, "car", "bike")',
    'A bike named Carry was bikeeful not to bikee about roundabouts being too round to bikee too much about bikes with bikery.');

  Test('StringReplace(s, "car", "bike", "i")',
    'A bike named bikery was bikeeful not to bikee about roundabouts being too round to bikee too much about bikes with bikery.');

  Test('StringReplace(s, "car", "bike", "w")',
    'A bike named Carry was careful not to care about roundabouts being too round to care too much about cars with carry.');

  Test('StringReplace(s, "car", "bike", "iw")',
    'A bike named Carry was careful not to care about roundabouts being too round to care too much about cars with carry.');

  Test('StringReplace(s, "care", "forget")',
    'A car named Carry was forgetful not to forget about roundabouts being too round to forget too much about cars with carry.');

  Test('StringReplace(s, "care", "forget", "w")',
    'A car named Carry was careful not to forget about roundabouts being too round to forget too much about cars with carry.');

  Test('StringReplace(s, "carry", "bikery")',
    'A car named Carry was careful not to care about roundabouts being too round to care too much about cars with bikery.');

  Test('StringReplace(s, "carry", "bikery", "i")',
    'A car named bikery was careful not to care about roundabouts being too round to care too much about cars with bikery.');

  Test('StringReplace(s, "about", "for")',
    'A car named Carry was careful not to care for roundfors being too round to care too much for cars with carry.');

  Test('StringReplace(s, "about", "for", "w")',
    'A car named Carry was careful not to care for roundabouts being too round to care too much for cars with carry.');

  Test('StringReplace(s, "a", "u")',
    'A cur numed Curry wus cureful not to cure ubout roundubouts being too round to cure too much ubout curs with curry.');

  Test('StringReplace(s, "a", "u", "i")',
    'u cur numed Curry wus cureful not to cure ubout roundubouts being too round to cure too much ubout curs with curry.');

  Test('StringReplace(s, "a", "u", "w")',
    'A car named Carry was careful not to care about roundabouts being too round to care too much about cars with carry.');

  Test('StringReplace(s, "a", "u", "iw")',
    'u car named Carry was careful not to care about roundabouts being too round to care too much about cars with carry.');

  Test('StringReplace(s, "e", "ee")',
    'A car nameed Carry was careeful not to caree about roundabouts beeing too round to caree too much about cars with carry.');

  Test('StringReplace(s, "much", "little", "w")',
    'A car named Carry was careful not to care about roundabouts being too round to care too little about cars with carry.');

  Test('StringReplace(s, "roundabout", "hippogriff")',
    'A car named Carry was careful not to care about hippogriffs being too round to care too much about cars with carry.');

  Test('StringReplace(s, "roundabout", "hippogriff", "w")',
    'A car named Carry was careful not to care about roundabouts being too round to care too much about cars with carry.');

  Test('StringReplace(s, "nargle", "wrackspurt")',
    'A car named Carry was careful not to care about roundabouts being too round to care too much about cars with carry.');

  Test('StringReplace(s, "being too round", "not being round enough")',
    'A car named Carry was careful not to care about roundabouts not being round enough to care too much about cars with carry.');

  Test('StringReplace(s, "being too round", "not being round enough", "iw")',
    'A car named Carry was careful not to care about roundabouts not being round enough to care too much about cars with carry.');

  Test('StringReplace(s, "not", "")',
    'A car named Carry was careful  to care about roundabouts being too round to care too much about cars with carry.');

  Test('StringReplace(s, "not ", "")',
    'A car named Carry was careful to care about roundabouts being too round to care too much about cars with carry.');

  Test('StringReplace(s, " ", "")',
    'AcarnamedCarrywascarefulnottocareaboutroundaboutsbeingtooroundtocaretoomuchaboutcarswithcarry.');

  Test('StringReplace(s, " ", "-")',
    'A-car-named-Carry-was-careful-not-to-care-about-roundabouts-being-too-round-to-care-too-much-about-cars-with-carry.');

  Test('StringReplace(s, "c", "b")',
    'A bar named Carry was bareful not to bare about roundabouts being too round to bare too mubh about bars with barry.');

  Test('StringReplace(s, "c", "b", "i")',
    'A bar named barry was bareful not to bare about roundabouts being too round to bare too mubh about bars with barry.');

  Test('s ≔ "A cat is a good/excellent" + chr(13) + chr(10) + "mister. –Indeed!"', 'A cat is a good/excellent'#13#10'mister. –Indeed!');

  Test('StringReplace(s, "go", "run")',
    'A cat is a runod/excellent'#13#10'mister. –Indeed!');

  Test('StringReplace(s, "go", "run", "w")',
    'A cat is a good/excellent'#13#10'mister. –Indeed!');

  Test('StringReplace(s, "good", "terrific")',
    'A cat is a terrific/excellent'#13#10'mister. –Indeed!');

  Test('StringReplace(s, "good", "terrific", "w")',
    'A cat is a terrific/excellent'#13#10'mister. –Indeed!');

  Test('StringReplace(s, "excellent", "splendid", "w")',
    'A cat is a good/splendid'#13#10'mister. –Indeed!');

  Test('StringReplace(s, "mister", "friend", "w")',
    'A cat is a good/excellent'#13#10'friend. –Indeed!');

  Test('StringReplace(s, "indeed", "yes")',
    'A cat is a good/excellent'#13#10'mister. –Indeed!');

  Test('StringReplace(s, "indeed", "yes", "i")',
    'A cat is a good/excellent'#13#10'mister. –yes!');

  Test('StringReplace(s, "indeed", "yes", "iw")',
    'A cat is a good/excellent'#13#10'mister. –yes!');

  Test('s ≔ "Överskatta inte köpmannens förkärlek för sångfågeln."', 'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringReplace(s, "överskatta", "underskatta")',
    'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringReplace(s, "överskatta", "underskatta", "i")',
    'underskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringReplace(s, "skatta", "deklarera")',
    'Överdeklarera inte köpmannens förkärlek för sångfågeln.');

  Test('StringReplace(s, "skatta", "deklarera", "w")',
    'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringReplace(s, "man", "kvinna")',
    'Överskatta inte köpkvinnanens förkärlek för sångfågeln.');

  Test('StringReplace(s, "man", "kvinna", "w")',
    'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringReplace(s, "kärlek", "hopp")',
    'Överskatta inte köpmannens förhopp för sångfågeln.');

  Test('StringReplace(s, "kärlek", "hopp", "w")',
    'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringReplace(s, "sång", "musik")',
    'Överskatta inte köpmannens förkärlek för musikfågeln.');

  Test('StringReplace(s, "sång", "musik", "w")',
    'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('StringReplace(s, "fågel", "uggla")',
    'Överskatta inte köpmannens förkärlek för sångugglan.');

  Test('StringReplace(s, "fågel", "uggla", "w")',
    'Överskatta inte köpmannens förkärlek för sångfågeln.');

  Test('s ≔ "Ansökan om måsten?"', 'Ansökan om måsten?');

  Test('StringReplace(s, "kan", "bör")',
    'Ansöbör om måsten?');

  Test('StringReplace(s, "kan", "bör", "w")',
    'Ansökan om måsten?');

  Test('StringReplace(s, "sten", "rot")',
    'Ansökan om mårot?');

  Test('StringReplace(s, "sten", "rot", "w")',
    'Ansökan om måsten?');

  Test('StringReplace(s, "om", "utan", "w")',
    'Ansökan utan måsten?');

  Test('s ≔ "aaaaaaAAAAAAA"', 'aaaaaaAAAAAAA');

  Test('StringReplace(s, "a", "b")',
    'bbbbbbAAAAAAA');

  Test('StringReplace(s, "a", "b", "i")',
    'bbbbbbbbbbbbb');

  Test('StringReplace(s, "a", "(a)")',
    '(a)(a)(a)(a)(a)(a)AAAAAAA');

  Test('StringReplace(s, "aa", "cd")',
    'cdcdcdAAAAAAA');

  Test('StringReplace(s, "aa", "cat")',
    'catcatcatAAAAAAA');

  Test('StringReplace(s, "aaa", "cat")',
    'catcatAAAAAAA');

  Test('StringReplace(s, "aaa", "!")',
    '!!AAAAAAA');

  Test('StringReplace(s, "aaaa", "cat")',
    'cataaAAAAAAA');

  Test('StringReplace(s, "a", "")',
    'AAAAAAA');

  Test('StringReplace(s, "aa", "")',
    'AAAAAAA');

  Test('StringReplace(s, "aaa", "")',
    'AAAAAAA');

  Test('StringReplace(s, "aaaa", "")',
    'aaAAAAAAA');

  Test('StringReplace(s, "aaaaa", "")',
    'aAAAAAAA');

  Test('StringReplace(s, "aaaaa", "", "i")',
    'AAA');

  Test('StringReplace(s, "aA", "a" + chr(13) + chr(10) + "A")',
    'aaaaaa'#13#10'AAAAAAA');

  _s := LoremIpsum(10000);
  Test('LoremStr ≔ "' + _s + '";', null);

  _s2 := SysUtils.StringReplace(_s, 'ipsum', 'hipster', [rfReplaceAll]);

  StartStopwatch;
  Test('StringReplace(LoremStr, "ipsum", "hipster")', _s2);
  StopStopwatch;
  TestDuration(0.1);

  StartStopwatch;
  Test('StringReplace(LoremStr, "ipsum", "hipster", "iw")', _s2);
  StopStopwatch;
  TestDuration(0.1);

  _s2 := SysUtils.StringReplace(_s, ' ', #13#10, [rfReplaceAll]);

  StartStopwatch;
  Test('StringReplace(LoremStr, " ", chr(13) + chr(10))', _s2);
  StopStopwatch;
  TestDuration(0.12);

  Test('delete(s); delete(LoremStr)', success);
  _s := '';
  _s2 := '';

  // Splitting

  Test('split("alfa beta gamma", " ")', strarr(['alfa', 'beta', 'gamma']));
  Test('split("alfa,beta,gamma", " ")', strarr(['alfa,beta,gamma']));
  Test('split("alfa,beta,gamma", ",")', strarr(['alfa', 'beta', 'gamma']));

  Test('split(" alfa, beta,  gamma ", ",")', strarr([' alfa', ' beta', '  gamma ']));
  Test('split(" alfa, beta,  gamma ", ",") @ trim', strarr(['alfa', 'beta', 'gamma']));

  Test('split("alfa beta,gamma", " ")', strarr(['alfa', 'beta,gamma']));
  Test('split("alfa beta,gamma", ",")', strarr(['alfa beta', 'gamma']));
  Test('split("alfa beta,gamma", ''(" ", ","))', strarr(['alfa', 'beta', 'gamma']));

  Test('split("alfa beta  gamma", " ")', strarr(['alfa', 'beta', '', 'gamma']));
  Test('split("alfa beta gamma  ", " ")', strarr(['alfa', 'beta', 'gamma', '', '']));
  Test('split(" alfa beta gamma", " ")', strarr(['', 'alfa', 'beta', 'gamma']));

  Test('split("", " ")', strarr([]));
  Test('split(" ", " ")', strarr(['', '']));

  Test('split("", ",")', strarr([]));
  Test('split("alfa", ",")', strarr(['alfa']));
  Test('split("alfa,beta", ",")', strarr(['alfa', 'beta']));
  Test('split("alfa,beta,gamma", ",")', strarr(['alfa', 'beta', 'gamma']));
  Test('split("alfa,beta,gamma,delta", ",")', strarr(['alfa', 'beta', 'gamma', 'delta']));

  Test('split("1 2 3;4 5 6;7 8 9", ''(" ", ";"))', strarr(['1', '2', '3', '4', '5', '6', '7', '8', '9']));

  Test('s ≔ "' + LoremIpsum(10000) + '";', null);
  StartStopwatch;
  Test('#split(s, ''(" ", chr(13) + chr(10)))', 69 * 10000);
  StopStopwatch;
  TestDuration(0.3);

  Test('delete(s)', success);

  Test('split("alfa beta gamma", "")', strarr(['alfa beta gamma']));
  Test('split("alfa beta gamma", ''())', strarr(['alfa beta gamma']));

  Test('split("alfa beta gamma", 5)', failure, 'An object of type array was expected as argument 2, but an object of type integer was given.');
  Test('split("alfa beta gamma", ''(1, 3))', failure, 'The array must contain only objects of type string, but it contains an object of type integer.');

  // Joining

  Test('letters ≔ ''("alpha", "beta", "gamma", "delta", "epsilon")',
    strarr(['alpha', 'beta', 'gamma', 'delta', 'epsilon']));

  Test('join(letters)', 'alpha, beta, gamma, delta, epsilon');
  Test('join(letters, " ")', 'alpha beta gamma delta epsilon');
  Test('join(letters, ",")', 'alpha,beta,gamma,delta,epsilon');
  Test('join(letters, ";")', 'alpha;beta;gamma;delta;epsilon');

  Test('join(letters, UpperCase)', 'ALPHA, BETA, GAMMA, DELTA, EPSILON');
  Test('join(letters, UpperCase, " ")', 'ALPHA BETA GAMMA DELTA EPSILON');
  Test('join(letters, UpperCase, ",")', 'ALPHA,BETA,GAMMA,DELTA,EPSILON');
  Test('join(letters, UpperCase, ";")', 'ALPHA;BETA;GAMMA;DELTA;EPSILON');

  Test('join(letters, ", ", """", """")', '"alpha", "beta", "gamma", "delta", "epsilon"');
  Test('join(letters, ", ", "(", ")")', '(alpha), (beta), (gamma), (delta), (epsilon)');
  Test('join(letters, ", ", "(")', '(alpha, (beta, (gamma, (delta, (epsilon');

  Test('join(letters, UpperCase, ", ", """", """")', '"ALPHA", "BETA", "GAMMA", "DELTA", "EPSILON"');
  Test('join(letters, UpperCase, ", ", "(", ")")', '(ALPHA), (BETA), (GAMMA), (DELTA), (EPSILON)');
  Test('join(letters, UpperCase, ", ", "(")', '(ALPHA, (BETA, (GAMMA, (DELTA, (EPSILON');

  Test('join(letters, ", ", """", """", "[", "]")', '["alpha", "beta", "gamma", "delta", "epsilon"]');
  Test('join(letters, ", ", "(", ")", "[", "]")', '[(alpha), (beta), (gamma), (delta), (epsilon)]');
  Test('join(letters, ", ", "(", ")", "[")', '[(alpha), (beta), (gamma), (delta), (epsilon)');

  Test('join(letters, UpperCase, ", ", """", """", "[", "]")', '["ALPHA", "BETA", "GAMMA", "DELTA", "EPSILON"]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[", "]")', '[(ALPHA), (BETA), (GAMMA), (DELTA), (EPSILON)]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[")', '[(ALPHA), (BETA), (GAMMA), (DELTA), (EPSILON)');

  Test('letters ≔ ''("alpha")', strarr(['alpha']));

  Test('join(letters)', 'alpha');
  Test('join(letters, " ")', 'alpha');
  Test('join(letters, ",")', 'alpha');
  Test('join(letters, ";")', 'alpha');

  Test('join(letters, UpperCase)', 'ALPHA');
  Test('join(letters, UpperCase, " ")', 'ALPHA');
  Test('join(letters, UpperCase, ",")', 'ALPHA');
  Test('join(letters, UpperCase, ";")', 'ALPHA');

  Test('join(letters, ", ", """", """")', '"alpha"');
  Test('join(letters, ", ", "(", ")")', '(alpha)');
  Test('join(letters, ", ", "(")', '(alpha');

  Test('join(letters, UpperCase, ", ", """", """")', '"ALPHA"');
  Test('join(letters, UpperCase, ", ", "(", ")")', '(ALPHA)');
  Test('join(letters, UpperCase, ", ", "(")', '(ALPHA');

  Test('join(letters, ", ", """", """", "[", "]")', '["alpha"]');
  Test('join(letters, ", ", "(", ")", "[", "]")', '[(alpha)]');
  Test('join(letters, ", ", "(", ")", "[")', '[(alpha)');

  Test('join(letters, UpperCase, ", ", """", """", "[", "]")', '["ALPHA"]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[", "]")', '[(ALPHA)]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[")', '[(ALPHA)');

  Test('letters ≔ ''()', strarr([]));

  Test('join(letters)', '');
  Test('join(letters, " ")', '');
  Test('join(letters, ",")', '');
  Test('join(letters, ";")', '');

  Test('join(letters, UpperCase)', '');
  Test('join(letters, UpperCase, " ")', '');
  Test('join(letters, UpperCase, ",")', '');
  Test('join(letters, UpperCase, ";")', '');

  Test('join(letters, ", ", """", """")', '');
  Test('join(letters, ", ", "(", ")")', '');
  Test('join(letters, ", ", "(")', '');

  Test('join(letters, UpperCase, ", ", """", """")', '');
  Test('join(letters, UpperCase, ", ", "(", ")")', '');
  Test('join(letters, UpperCase, ", ", "(")', '');

  Test('join(letters, ", ", """", """", "[", "]")', '[]');
  Test('join(letters, ", ", "(", ")", "[", "]")', '[]');
  Test('join(letters, ", ", "(", ")", "[")', '[');

  Test('join(letters, UpperCase, ", ", """", """", "[", "]")', '[]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[", "]")', '[]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[")', '[');

  Test('letters ≔ ''("alpha", "beta", "gamma", "", "epsilon")',
    strarr(['alpha', 'beta', 'gamma', '', 'epsilon']));

  Test('join(letters)', 'alpha, beta, gamma, , epsilon');
  Test('join(letters, " ")', 'alpha beta gamma  epsilon');
  Test('join(letters, ",")', 'alpha,beta,gamma,,epsilon');
  Test('join(letters, ";")', 'alpha;beta;gamma;;epsilon');

  Test('join(letters, UpperCase)', 'ALPHA, BETA, GAMMA, , EPSILON');
  Test('join(letters, UpperCase, " ")', 'ALPHA BETA GAMMA  EPSILON');
  Test('join(letters, UpperCase, ",")', 'ALPHA,BETA,GAMMA,,EPSILON');
  Test('join(letters, UpperCase, ";")', 'ALPHA;BETA;GAMMA;;EPSILON');

  Test('join(letters, ", ", """", """")', '"alpha", "beta", "gamma", "", "epsilon"');
  Test('join(letters, ", ", "(", ")")', '(alpha), (beta), (gamma), (), (epsilon)');
  Test('join(letters, ", ", "(")', '(alpha, (beta, (gamma, (, (epsilon');

  Test('join(letters, UpperCase, ", ", """", """")', '"ALPHA", "BETA", "GAMMA", "", "EPSILON"');
  Test('join(letters, UpperCase, ", ", "(", ")")', '(ALPHA), (BETA), (GAMMA), (), (EPSILON)');
  Test('join(letters, UpperCase, ", ", "(")', '(ALPHA, (BETA, (GAMMA, (, (EPSILON');

  Test('join(letters, ", ", """", """", "[", "]")', '["alpha", "beta", "gamma", "", "epsilon"]');
  Test('join(letters, ", ", "(", ")", "[", "]")', '[(alpha), (beta), (gamma), (), (epsilon)]');
  Test('join(letters, ", ", "(", ")", "[")', '[(alpha), (beta), (gamma), (), (epsilon)');

  Test('join(letters, UpperCase, ", ", """", """", "[", "]")', '["ALPHA", "BETA", "GAMMA", "", "EPSILON"]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[", "]")', '[(ALPHA), (BETA), (GAMMA), (), (EPSILON)]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[")', '[(ALPHA), (BETA), (GAMMA), (), (EPSILON)');

  Test('letters ≔ ''("")', strarr(['']));

  Test('join(letters)', '');
  Test('join(letters, " ")', '');
  Test('join(letters, ",")', '');
  Test('join(letters, ";")', '');

  Test('join(letters, UpperCase)', '');
  Test('join(letters, UpperCase, " ")', '');
  Test('join(letters, UpperCase, ",")', '');
  Test('join(letters, UpperCase, ";")', '');

  Test('join(letters, ", ", """", """")', '""');
  Test('join(letters, ", ", "(", ")")', '()');
  Test('join(letters, ", ", "(")', '(');

  Test('join(letters, UpperCase, ", ", """", """")', '""');
  Test('join(letters, UpperCase, ", ", "(", ")")', '()');
  Test('join(letters, UpperCase, ", ", "(")', '(');

  Test('join(letters, ", ", """", """", "[", "]")', '[""]');
  Test('join(letters, ", ", "(", ")", "[", "]")', '[()]');
  Test('join(letters, ", ", "(", ")", "[")', '[()');

  Test('join(letters, UpperCase, ", ", """", """", "[", "]")', '[""]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[", "]")', '[()]');
  Test('join(letters, UpperCase, ", ", "(", ")", "[")', '[()');

  Test('letters ≔ ''("alpha", "beta", "gamma", "delta", "epsilon")',
    strarr(['alpha', 'beta', 'gamma', 'delta', 'epsilon']));

  Test('join(letters, (s ↦ UpperCase(reverse(s))))', 'AHPLA, ATEB, AMMAG, ATLED, NOLISPE');
  Test('join(letters, (s ↦ UpperCase(reverse(s))), " ")', 'AHPLA ATEB AMMAG ATLED NOLISPE');
  Test('join(letters, (s ↦ UpperCase(reverse(s))), ",")', 'AHPLA,ATEB,AMMAG,ATLED,NOLISPE');
  Test('join(letters, (s ↦ UpperCase(reverse(s))), ";")', 'AHPLA;ATEB;AMMAG;ATLED;NOLISPE');

  Test('delete(letters)', success);

  // Trimming

  Test('trim("   Hello, World!  ")', 'Hello, World!');
  Test('TrimLeft("   Hello, World!  ")', 'Hello, World!  ');
  Test('TrimRight("   Hello, World!  ")', '   Hello, World!');

  Test('trim("   ")', '');
  Test('TrimLeft("   ")', '');
  Test('TrimRight("   ")', '');

  Test('trim("")', '');
  Test('TrimLeft("")', '');
  Test('TrimRight("")', '');

  Test('trim(chr(13) + chr(10) + "  Hello, World!   " + chr(13) + chr(10))', 'Hello, World!');
  Test('TrimLeft(chr(13) + chr(10) + "  Hello, World!   " + chr(13) + chr(10))', 'Hello, World!   '#13#10);
  Test('TrimRight(chr(13) + chr(10) + "  Hello, World!   " + chr(13) + chr(10))', #13#10'  Hello, World!');

  Test('trim("  ' + LoremIpsum(1000) + '        ")', LoremIpsum(1000));

  Test('trim(5)', failure, 'An object of type string was expected as argument 1, but an object of type integer was given.');
  Test('TrimLeft(5)', failure, 'An object of type string was expected as argument 1, but an object of type integer was given.');
  Test('TrimRight(5)', failure, 'An object of type string was expected as argument 1, but an object of type integer was given.');

  Test('trim()', failure, 'Too few arguments. A required argument of type string is missing.');
  Test('trim("  test  ", 5)', failure, 'Too many arguments.');

  // Padding

  Test('StringPad("test")', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('StringPad("test", 0)', 'test');
  Test('StringPad("test", 1)', 'test');
  Test('StringPad("test", 2)', 'test');
  Test('StringPad("test", 3)', 'test');
  Test('StringPad("test", 4)', 'test');
  Test('StringPad("test", 5)', 'test ');
  Test('StringPad("test", 6)', 'test  ');
  Test('StringPad("test", 7)', 'test   ');
  Test('StringPad("test", 8)', 'test    ');
  Test('StringPad("test", 9)', 'test     ');
  Test('StringPad("test", 10)', 'test      ');
  Test('StringPad("test", -1)', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.');

  Test('StringPad("test", 0, "l")', 'test');
  Test('StringPad("test", 1, "l")', 'test');
  Test('StringPad("test", 2, "l")', 'test');
  Test('StringPad("test", 3, "l")', 'test');
  Test('StringPad("test", 4, "l")', 'test');
  Test('StringPad("test", 5, "l")', 'test ');
  Test('StringPad("test", 6, "l")', 'test  ');
  Test('StringPad("test", 7, "l")', 'test   ');
  Test('StringPad("test", 8, "l")', 'test    ');
  Test('StringPad("test", 9, "l")', 'test     ');
  Test('StringPad("test", 10, "l")', 'test      ');

  Test('StringPad("test", 0, "r")', 'test');
  Test('StringPad("test", 1, "r")', 'test');
  Test('StringPad("test", 2, "r")', 'test');
  Test('StringPad("test", 3, "r")', 'test');
  Test('StringPad("test", 4, "r")', 'test');
  Test('StringPad("test", 5, "r")', ' test');
  Test('StringPad("test", 6, "r")', '  test');
  Test('StringPad("test", 7, "r")', '   test');
  Test('StringPad("test", 8, "r")', '    test');
  Test('StringPad("test", 9, "r")', '     test');
  Test('StringPad("test", 10, "r")', '      test');

  Test('StringPad("test", 0, "c")', 'test');
  Test('StringPad("test", 1, "c")', 'test');
  Test('StringPad("test", 2, "c")', 'test');
  Test('StringPad("test", 3, "c")', 'test');
  Test('StringPad("test", 4, "c")', 'test');
  Test('StringPad("test", 5, "c")', 'test ');
  Test('StringPad("test", 6, "c")', ' test ');
  Test('StringPad("test", 7, "c")', ' test  ');
  Test('StringPad("test", 8, "c")', '  test  ');
  Test('StringPad("test", 9, "c")', '  test   ');
  Test('StringPad("test", 10, "c")', '   test   ');

  Test('StringPad("test", 6, "a")', failure, 'Character "a" not allowed as argument. Expected one of "lrc".');

  Test('StringPad("test", 0, "l", ".")', 'test');
  Test('StringPad("test", 1, "l", ".")', 'test');
  Test('StringPad("test", 2, "l", ".")', 'test');
  Test('StringPad("test", 3, "l", ".")', 'test');
  Test('StringPad("test", 4, "l", ".")', 'test');
  Test('StringPad("test", 5, "l", ".")', 'test.');
  Test('StringPad("test", 6, "l", ".")', 'test..');
  Test('StringPad("test", 7, "l", ".")', 'test...');
  Test('StringPad("test", 8, "l", ".")', 'test....');
  Test('StringPad("test", 9, "l", ".")', 'test.....');
  Test('StringPad("test", 10, "l", ".")', 'test......');

  Test('StringPad("test", 0, "r", ".")', 'test');
  Test('StringPad("test", 1, "r", ".")', 'test');
  Test('StringPad("test", 2, "r", ".")', 'test');
  Test('StringPad("test", 3, "r", ".")', 'test');
  Test('StringPad("test", 4, "r", ".")', 'test');
  Test('StringPad("test", 5, "r", ".")', '.test');
  Test('StringPad("test", 6, "r", ".")', '..test');
  Test('StringPad("test", 7, "r", ".")', '...test');
  Test('StringPad("test", 8, "r", ".")', '....test');
  Test('StringPad("test", 9, "r", ".")', '.....test');
  Test('StringPad("test", 10, "r", ".")', '......test');

  Test('StringPad("test", 0, "c", ".")', 'test');
  Test('StringPad("test", 1, "c", ".")', 'test');
  Test('StringPad("test", 2, "c", ".")', 'test');
  Test('StringPad("test", 3, "c", ".")', 'test');
  Test('StringPad("test", 4, "c", ".")', 'test');
  Test('StringPad("test", 5, "c", ".")', 'test.');
  Test('StringPad("test", 6, "c", ".")', '.test.');
  Test('StringPad("test", 7, "c", ".")', '.test..');
  Test('StringPad("test", 8, "c", ".")', '..test..');
  Test('StringPad("test", 9, "c", ".")', '..test...');
  Test('StringPad("test", 10, "c", ".")', '...test...');

  Test('StringPad("test", 10, "c", "..")', failure, 'A character was expected as argument 4, but ".." was given.');

  Test('StringPad("", 10, "l", ".")', '..........');
  Test('StringPad("", 11, "l", ".")', '...........');
  Test('StringPad("", 10, "r", ".")', '..........');
  Test('StringPad("", 11, "r", ".")', '...........');
  Test('StringPad("", 10, "c", ".")', '..........');
  Test('StringPad("", 11, "c", ".")', '...........');


  //
  // Transformations
  //

  // Case conversion

  Test('s ≔ "Luna Lovegood loves to care for magical creatures. And she does it well."',
    'Luna Lovegood loves to care for magical creatures. And she does it well.');

  Test('UpperCase(s)', 'LUNA LOVEGOOD LOVES TO CARE FOR MAGICAL CREATURES. AND SHE DOES IT WELL.');

  Test('LowerCase(s)', 'luna lovegood loves to care for magical creatures. and she does it well.');

  Test('InvertCase(s)', 'lUNA lOVEGOOD LOVES TO CARE FOR MAGICAL CREATURES. aND SHE DOES IT WELL.');

  Test('TitleCase(s)', 'Luna Lovegood Loves To Care For Magical Creatures. And She Does It Well.');

  Test('SentenceCase(s)', 'Luna Lovegood loves to care for magical creatures. And she does it well.');
  Test('SentenceCase(LowerCase(s))', 'Luna lovegood loves to care for magical creatures. And she does it well.');

  Test('UpperCase("")', '');
  Test('LowerCase("")', '');
  Test('InvertCase("")', '');
  Test('TitleCase("")', '');
  Test('SentenceCase("")', '');

  Test('SentenceCase("this is merely an example. it isn''t important! or is it? maybe, I wouldn''t know. many would care; however, I don''t care.")',
    'This is merely an example. It isn''t important! Or is it? Maybe, I wouldn''t know. Many would care; however, I don''t care.');

  Test('SentenceCase("(most) people like dogs. (almost) everyone was there.")',
    '(Most) people like dogs. (Almost) everyone was there.');

  Test('SentenceCase("wait... almost there!")', 'Wait... Almost there!');

  Test('SentenceCase("no?! really?")', 'No?! Really?');
  Test('SentenceCase("no‽ really?")', 'No‽ Really?');

  // Obfuscation

  Test('rot13("Past the point of no return.")', 'Cnfg gur cbvag bs ab erghea.');
  Test('rot13("Gur svany guerfubyq.")', 'The final threshold.');
  Test('rot13(rot13("What warm, unspoken secrets will we learn?"))', 'What warm, unspoken secrets will we learn?');

  Test('rot13("Sjung om igen med mig, vår kärlekssång.")', 'Fwhat bz vtra zrq zvt, iåe xäeyrxffåat.');

  Test('rot13("")', '');

  Test('rot13("∫sin(x)dx")', '∫fva(k)qk');

  // -

  Test('Caesar("What raging fire shall flood the soul?", 10)',
    'Grkd bkqsxq psbo crkvv pvyyn dro cyev?');

  Test('Caesar("What rich desire unlocks its door?", 15)',
    'Lwpi gxrw sthxgt jcadrzh xih sddg?');

  Test('Caesar("What sweet seduction lies before us?", 2)',
    'Yjcv uyggv ugfwevkqp nkgu dghqtg wu?');

  Test('Caesar("What raging fire shall flood the soul?", 10 + 26)',
    'Grkd bkqsxq psbo crkvv pvyyn dro cyev?');

  Test('Caesar("What rich desire unlocks its door?", 15 + 26 + 26)',
    'Lwpi gxrw sthxgt jcadrzh xih sddg?');

  Test('Caesar("What sweet seduction lies before us?", 2 + 26 + 26 + 26)',
    'Yjcv uyggv ugfwevkqp nkgu dghqtg wu?');

  Test('Caesar(Caesar("What raging fire shall flood the soul?", 10), -10)',
    'What raging fire shall flood the soul?');

  Test('Caesar(Caesar("What rich desire unlocks its door?", 15), -15)',
    'What rich desire unlocks its door?');

  Test('Caesar(Caesar("What sweet seduction lies before us?", 2), -2)',
    'What sweet seduction lies before us?');

  Test('Caesar("Och du må fly min blick, och frukta mig.", 5)',
    'Thm iz rå kqd rns gqnhp, thm kwzpyf rnl.');

  Test('Caesar("Och du må fly min blick, och frukta mig.", 6)',
    'Uin ja så lre sot hroiq, uin lxaqzg som.');

  Test('Caesar(Caesar("Och du må fly min blick, och frukta mig.", 6), -6)',
    'Och du må fly min blick, och frukta mig.');

  Test('Caesar("", 10)', '');

  // -

  Test('VigenèreEncode("And do I dream again?", "opera")',
    'Och dc M dfted ovezn?');

  Test('VigenèreDecode(VigenèreEncode("And do I dream again?", "opera"), "opera")',
    'And do I dream again?');

  Test('VigenèreEncode("Nu är sista gränsen nådd, och överstigen.", "opera")',
    'Bj är hmjto kiäbhie båhu, dgy ökiishxkvn.');

  Test('VigenèreDecode(VigenèreEncode("Nu är sista gränsen nådd, och överstigen.", "opera"), "opera")',
    'Nu är sista gränsen nådd, och överstigen.');

  Test('VigenèreEncode("And do I dream again?", "opera!")', failure, 'Invalid character in Vigenère key.');
  Test('VigenèreEncode("And do I dream again?", "opera ghost")', failure, 'Invalid character in Vigenère key.');
  Test('VigenèreEncode("And do I dream again?", "ändlös")', failure, 'Invalid character in Vigenère key.');

  Test('VigenèreEncode("", "opera")', '');
  Test('VigenèreDecode("", "opera")', '');

  Test('VigenèreEncode("And do I dream again?", "")', failure, 'Vigenère key is empty.');
  Test('VigenèreDecode("And do I dream again?", "")', failure, 'Vigenère key is empty.');


  //
  // The Format function
  //

  Test('format("Welcome %1. You have %2 new message(s).", ''("Mr Weasley", 517))',
    'Welcome Mr Weasley. You have 517 new message(s).');

  Test('format("Welcome %1. You have %2 new message(s), %1.", ''("Mr Weasley", 517))',
    'Welcome Mr Weasley. You have 517 new message(s), Mr Weasley.');

  Test('format("There are %1 messages, %2%% of which are internal.", ''(517, 85))',
    'There are 517 messages, 85% of which are internal.');

  Test('format("(%1, %2, %3)", ''(25, 36, 411))',
    '(25, 36, 411)');

  Test('format("(%3, %2, %1)", ''(25, 36, 411))',
    '(411, 36, 25)');

  Test('format("(%1, %2, %3) ~ (%3, %1, %2)", ''(25, 36, 411))',
    '(25, 36, 411) ~ (411, 25, 36)');

  Test('format("B%1m%2", ''("oooooooo", "!!!!!!"))',
    'Boooooooom!!!!!!');

  Test('format("B%1m%2", ''("oooooooo", "!!!!!!", "????"))',
    'Boooooooom!!!!!!');

  Test('format("Welcome %1. You have %2 new message(s) and %3 new alert(s).", ''("Mr Weasley", 517))',
    failure, 'Format string references non-existing object.');

  Test('format("Welcome %1. You have % new message(s).", ''("Mr Weasley", 517))',
    failure, 'Invalid format string.');

  Test('format("Welcome!", ''())', 'Welcome!');

  Test('format("Welcome!", ''(1, 2, 3))', 'Welcome!');

  Test('format("%1", ''("Welcome!"))', 'Welcome!');

  Test('format("%1%2%1", ''("Welcome!", "  "))', 'Welcome!  Welcome!');

  Test('format("", ''())', '');
  Test('format("", ''(1, 2, 3))', '');

  Test('format("Is 1+1=2? Answer: %1", ''(1+1=2))',
    'Is 1+1=2? Answer: true');


  //
  // Character tests
  //

  TestCharType('ChrIsLetter', 'GHNÖÅÉÏÔÂÑΓΒΘЧmbåñõéïγεφю');
  TestCharType('ChrIsDigit', '75610');
  TestCharType('ChrIsLetterOrDigit', 'GHNÖÅÉÏÔÂÑΓΒΘЧmbåñõéïγεφю75610');
  TestCharType('ChrIsNumber', '75610½¼¾⅛ⅷ፹Ⅳ⑭〤⑽⒖㉑');
  TestCharType('ChrIsPunctuation', '.,!?-;:()');
  TestCharType('ChrIsSeparator', #32#$A0);
  TestCharType('ChrIsWhitespace', #32#9#13#10#$A0);
  TestCharType('ChrIsUpperCase', 'GHNÖÅÉÏÔÂÑΓΒΘЧ');
  TestCharType('ChrIsLowerCase', 'mbåñõéïγεφю');
  TestCharType('ChrIsSymbol', '∫∅∂⌬⌨');
  TestCharType('ChrIsASCII', 'GHNmb75610.,!?-;:()'#32#9#13#10#1#2#7#24#27#28#31);
  TestCharType('ChrIsControl', #9#13#10#1#2#7#24#27#28#31);

  Test('ChrNumVal("0")', 0);
  Test('ChrNumVal("1")', 1);
  Test('ChrNumVal("2")', 2);
  Test('ChrNumVal("3")', 3);
  Test('ChrNumVal("4")', 4);
  Test('ChrNumVal("5")', 5);
  Test('ChrNumVal("6")', 6);
  Test('ChrNumVal("7")', 7);
  Test('ChrNumVal("8")', 8);
  Test('ChrNumVal("9")', 9);

  Test('ChrNumVal("½")', 1/2);
  Test('ChrNumVal("¼")', 1/4);
  Test('ChrNumVal("¾")', 3/4);
  Test('ChrNumVal("⅛")', 1/8);

  Test('ChrNumVal("Ⅳ")', 4);
  Test('ChrNumVal("ⅷ")', 8);

  Test('ChrNumVal("፹")', 80);

  Test('ChrNumVal("〤")', 4);

  Test('ChrNumVal("⑭")', 14);
  Test('ChrNumVal("⑽")', 10);
  Test('ChrNumVal("⒖")', 15);
  Test('ChrNumVal("㉑")', 21);


  //
  // UCD functions
  //

  Test('ChrName("A")', 'LATIN CAPITAL LETTER A');
  Test('ChrName("B")', 'LATIN CAPITAL LETTER B');
  Test('ChrName("C")', 'LATIN CAPITAL LETTER C');
  Test('ChrName("D")', 'LATIN CAPITAL LETTER D');
  Test('ChrName("E")', 'LATIN CAPITAL LETTER E');
  Test('ChrName("F")', 'LATIN CAPITAL LETTER F');
  Test('ChrName("G")', 'LATIN CAPITAL LETTER G');
  Test('ChrName("H")', 'LATIN CAPITAL LETTER H');
  Test('ChrName("I")', 'LATIN CAPITAL LETTER I');
  Test('ChrName("J")', 'LATIN CAPITAL LETTER J');
  Test('ChrName("K")', 'LATIN CAPITAL LETTER K');
  Test('ChrName("L")', 'LATIN CAPITAL LETTER L');
  Test('ChrName("M")', 'LATIN CAPITAL LETTER M');
  Test('ChrName("N")', 'LATIN CAPITAL LETTER N');
  Test('ChrName("O")', 'LATIN CAPITAL LETTER O');
  Test('ChrName("P")', 'LATIN CAPITAL LETTER P');
  Test('ChrName("Q")', 'LATIN CAPITAL LETTER Q');
  Test('ChrName("R")', 'LATIN CAPITAL LETTER R');
  Test('ChrName("S")', 'LATIN CAPITAL LETTER S');
  Test('ChrName("T")', 'LATIN CAPITAL LETTER T');
  Test('ChrName("U")', 'LATIN CAPITAL LETTER U');
  Test('ChrName("V")', 'LATIN CAPITAL LETTER V');
  Test('ChrName("W")', 'LATIN CAPITAL LETTER W');
  Test('ChrName("X")', 'LATIN CAPITAL LETTER X');
  Test('ChrName("Y")', 'LATIN CAPITAL LETTER Y');
  Test('ChrName("Z")', 'LATIN CAPITAL LETTER Z');

  Test('ChrName("a")', 'LATIN SMALL LETTER A');
  Test('ChrName("b")', 'LATIN SMALL LETTER B');
  Test('ChrName("c")', 'LATIN SMALL LETTER C');
  Test('ChrName("d")', 'LATIN SMALL LETTER D');
  Test('ChrName("e")', 'LATIN SMALL LETTER E');
  Test('ChrName("f")', 'LATIN SMALL LETTER F');
  Test('ChrName("g")', 'LATIN SMALL LETTER G');
  Test('ChrName("h")', 'LATIN SMALL LETTER H');
  Test('ChrName("i")', 'LATIN SMALL LETTER I');
  Test('ChrName("j")', 'LATIN SMALL LETTER J');
  Test('ChrName("k")', 'LATIN SMALL LETTER K');
  Test('ChrName("l")', 'LATIN SMALL LETTER L');
  Test('ChrName("m")', 'LATIN SMALL LETTER M');
  Test('ChrName("n")', 'LATIN SMALL LETTER N');
  Test('ChrName("o")', 'LATIN SMALL LETTER O');
  Test('ChrName("p")', 'LATIN SMALL LETTER P');
  Test('ChrName("q")', 'LATIN SMALL LETTER Q');
  Test('ChrName("r")', 'LATIN SMALL LETTER R');
  Test('ChrName("s")', 'LATIN SMALL LETTER S');
  Test('ChrName("t")', 'LATIN SMALL LETTER T');
  Test('ChrName("u")', 'LATIN SMALL LETTER U');
  Test('ChrName("v")', 'LATIN SMALL LETTER V');
  Test('ChrName("w")', 'LATIN SMALL LETTER W');
  Test('ChrName("x")', 'LATIN SMALL LETTER X');
  Test('ChrName("y")', 'LATIN SMALL LETTER Y');
  Test('ChrName("z")', 'LATIN SMALL LETTER Z');

  Test('ChrName("Å")', 'LATIN CAPITAL LETTER A WITH RING ABOVE');
  Test('ChrName("Ä")', 'LATIN CAPITAL LETTER A WITH DIAERESIS');
  Test('ChrName("Ö")', 'LATIN CAPITAL LETTER O WITH DIAERESIS');
  Test('ChrName("å")', 'LATIN SMALL LETTER A WITH RING ABOVE');
  Test('ChrName("ä")', 'LATIN SMALL LETTER A WITH DIAERESIS');
  Test('ChrName("ö")', 'LATIN SMALL LETTER O WITH DIAERESIS');
  Test('ChrName("ï")', 'LATIN SMALL LETTER I WITH DIAERESIS');
  Test('ChrName("ñ")', 'LATIN SMALL LETTER N WITH TILDE');
  Test('ChrName("ô")', 'LATIN SMALL LETTER O WITH CIRCUMFLEX');
  Test('ChrName("é")', 'LATIN SMALL LETTER E WITH ACUTE');

  Test('ChrName("Γ")', 'GREEK CAPITAL LETTER GAMMA');
  Test('ChrName("Λ")', 'GREEK CAPITAL LETTER LAMDA');
  Test('ChrName("Δ")', 'GREEK CAPITAL LETTER DELTA');
  Test('ChrName("Π")', 'GREEK CAPITAL LETTER PI');
  Test('ChrName("Ω")', 'GREEK CAPITAL LETTER OMEGA');
  Test('ChrName("Ξ")', 'GREEK CAPITAL LETTER XI');
  Test('ChrName("α")', 'GREEK SMALL LETTER ALPHA');
  Test('ChrName("β")', 'GREEK SMALL LETTER BETA');
  Test('ChrName("φ")', 'GREEK SMALL LETTER PHI');
  Test('ChrName("π")', 'GREEK SMALL LETTER PI');
  Test('ChrName("ξ")', 'GREEK SMALL LETTER XI');
  Test('ChrName("μ")', 'GREEK SMALL LETTER MU');

  Test('ChrName("Ѩ")', 'CYRILLIC CAPITAL LETTER IOTIFIED LITTLE YUS');
  Test('ChrName("Խ")', 'ARMENIAN CAPITAL LETTER XEH');
  Test('ChrName("ל")', 'HEBREW LETTER LAMED');
  Test('ChrName("س")', 'ARABIC LETTER SEEN');
  Test('ChrName("ޘ")', 'THAANA LETTER TTAA');
  Test('ChrName("߷")', 'NKO SYMBOL GBAKURUNEN');
  Test('ChrName("औ")', 'DEVANAGARI LETTER AU');
  Test('ChrName("ঊ")', 'BENGALI LETTER UU');
  Test('ChrName("ஊ")', 'TAMIL LETTER UU');
  Test('ChrName("ಊ")', 'KANNADA LETTER UU');
  Test('ChrName("ഊ")', 'MALAYALAM LETTER UU');
  Test('ChrName("ඖ")', 'SINHALA LETTER AUYANNA');
  Test('ChrName("ช")', 'THAI CHARACTER CHO CHANG');
  Test('ChrName("༒")', 'TIBETAN MARK RGYA GRAM SHAD');
  Test('ChrName("ማ")', 'ETHIOPIC SYLLABLE MAA');
  Test('ChrName("Ꮚ")', 'CHEROKEE LETTER QUU');
  Test('ChrName("ᚠ")', 'RUNIC LETTER FEHU FEOH FE F');
  Test('ChrName("៚")', 'KHMER SIGN KOOMUUT');
  Test('ChrName("ᠿ")', 'MONGOLIAN LETTER ZRA');
  Test('ChrName("⡨")', 'BRAILLE PATTERN DOTS-467');
  Test('ChrName("ꕥ")', 'VAI SYLLABLE ZHA');

  Test('ChrName("0")', 'DIGIT ZERO');
  Test('ChrName("1")', 'DIGIT ONE');
  Test('ChrName("2")', 'DIGIT TWO');
  Test('ChrName("3")', 'DIGIT THREE');
  Test('ChrName("4")', 'DIGIT FOUR');
  Test('ChrName("5")', 'DIGIT FIVE');
  Test('ChrName("6")', 'DIGIT SIX');
  Test('ChrName("7")', 'DIGIT SEVEN');
  Test('ChrName("8")', 'DIGIT EIGHT');
  Test('ChrName("9")', 'DIGIT NINE');

  Test('ChrName("Ⅳ")', 'ROMAN NUMERAL FOUR');
  Test('ChrName("ⅷ")', 'SMALL ROMAN NUMERAL EIGHT');

  Test('ChrName("㉑")', 'CIRCLED NUMBER TWENTY ONE');

  Test('ChrName(" ")', 'SPACE');
  Test('ChrName(".")', 'FULL STOP');
  Test('ChrName(",")', 'COMMA');
  Test('ChrName(":")', 'COLON');
  Test('ChrName(";")', 'SEMICOLON');
  Test('ChrName("!")', 'EXCLAMATION MARK');
  Test('ChrName("?")', 'QUESTION MARK');
  Test('ChrName("-")', 'HYPHEN-MINUS');
  Test('ChrName("–")', 'EN DASH');
  Test('ChrName("_")', 'LOW LINE');
  Test('ChrName("@")', 'COMMERCIAL AT');

  Test('ChrName("µ")', 'MICRO SIGN');

  Test('ChrName("""")', 'QUOTATION MARK');
  Test('ChrName("''")', 'APOSTROPHE');
  Test('ChrName("+")', 'PLUS SIGN');
  Test('ChrName("−")', 'MINUS SIGN');
  Test('ChrName("*")', 'ASTERISK');
  Test('ChrName("/")', 'SOLIDUS');
  Test('ChrName("\")', 'REVERSE SOLIDUS');

  Test('ChrName("&")', 'AMPERSAND');
  Test('ChrName("%")', 'PERCENT SIGN');
  Test('ChrName("½")', 'VULGAR FRACTION ONE HALF');
  Test('ChrName("¾")', 'VULGAR FRACTION THREE QUARTERS');
  Test('ChrName("⅛")', 'VULGAR FRACTION ONE EIGHTH');
  Test('ChrName("°")', 'DEGREE SIGN');
  Test('ChrName("©")', 'COPYRIGHT SIGN');

  Test('ChrName("≈")', 'ALMOST EQUAL TO');
  Test('ChrName("×")', 'MULTIPLICATION SIGN');
  Test('ChrName("√")', 'SQUARE ROOT');
  Test('ChrName("∫")', 'INTEGRAL');
  Test('ChrName("∬")', 'DOUBLE INTEGRAL');
  Test('ChrName("∭")', 'TRIPLE INTEGRAL');
  Test('ChrName("∑")', 'N-ARY SUMMATION');
  Test('ChrName("∏")', 'N-ARY PRODUCT');
  Test('ChrName("∅")', 'EMPTY SET');
  Test('ChrName("∂")', 'PARTIAL DIFFERENTIAL');
  Test('ChrName("∇")', 'NABLA');
  Test('ChrName("∈")', 'ELEMENT OF');
  Test('ChrName("⊂")', 'SUBSET OF');
  Test('ChrName("∪")', 'UNION');
  Test('ChrName("∩")', 'INTERSECTION');
  Test('ChrName("∖")', 'SET MINUS');
  Test('ChrName("∁")', 'COMPLEMENT');
  Test('ChrName("∧")', 'LOGICAL AND');
  Test('ChrName("∨")', 'LOGICAL OR');
  Test('ChrName("¬")', 'NOT SIGN');
  Test('ChrName("∴")', 'THEREFORE');
  Test('ChrName("∎")', 'END OF PROOF');
  Test('ChrName("≤")', 'LESS-THAN OR EQUAL TO');
  Test('ChrName("≥")', 'GREATER-THAN OR EQUAL TO');
  Test('ChrName("∃")', 'THERE EXISTS');
  Test('ChrName("∀")', 'FOR ALL');
  Test('ChrName("∞")', 'INFINITY');

  Test('ChrName("ℕ")', 'DOUBLE-STRUCK CAPITAL N');
  Test('ChrName("ℤ")', 'DOUBLE-STRUCK CAPITAL Z');
  Test('ChrName("ℚ")', 'DOUBLE-STRUCK CAPITAL Q');
  Test('ChrName("ℝ")', 'DOUBLE-STRUCK CAPITAL R');
  Test('ChrName("ℂ")', 'DOUBLE-STRUCK CAPITAL C');
  Test('ChrName("ℍ")', 'DOUBLE-STRUCK CAPITAL H');

  Test('ChrName("⚘")', 'FLOWER');
  Test('ChrName("⌬")', 'BENZENE RING');
  Test('ChrName("⌨")', 'KEYBOARD');
  Test('ChrName("⌛")', 'HOURGLASS');
  Test('ChrName("☘")', 'SHAMROCK');

  Test('ChrName("cat")', failure, 'A character was expected as argument 1, but "cat" was given.');
  Test('ChrName("")', failure, 'A character was expected as argument 1, but "" was given.');

  Test('ChrBlock("A")', 'Basic Latin');
  Test('ChrBlock("b")', 'Basic Latin');
  Test('ChrBlock("3")', 'Basic Latin');
  Test('ChrBlock("!")', 'Basic Latin');
  Test('ChrBlock("ñ")', 'Latin-1 Supplement');
  Test('ChrBlock("–")', 'General Punctuation');
  Test('ChrBlock("Λ")', 'Greek and Coptic');
  Test('ChrBlock("Ѩ")', 'Cyrillic');
  Test('ChrBlock("Խ")', 'Armenian');
  Test('ChrBlock("ל")', 'Hebrew');
  Test('ChrBlock("س")', 'Arabic');
  Test('ChrBlock("ޘ")', 'Thaana');
  Test('ChrBlock("߷")', 'NKo');
  Test('ChrBlock("औ")', 'Devanagari');
  Test('ChrBlock("ঊ")', 'Bengali');
  Test('ChrBlock("ஊ")', 'Tamil');
  Test('ChrBlock("ಊ")', 'Kannada');
  Test('ChrBlock("ഊ")', 'Malayalam');
  Test('ChrBlock("ඖ")', 'Sinhala');
  Test('ChrBlock("ช")', 'Thai');
  Test('ChrBlock("༒")', 'Tibetan');
  Test('ChrBlock("ማ")', 'Ethiopic');
  Test('ChrBlock("Ꮚ")', 'Cherokee');
  Test('ChrBlock("ᚠ")', 'Runic');
  Test('ChrBlock("៚")', 'Khmer');
  Test('ChrBlock("ᠿ")', 'Mongolian');
  Test('ChrBlock("⡨")', 'Braille Patterns');
  Test('ChrBlock("ꕥ")', 'Vai');
  Test('ChrBlock("Ⅳ")', 'Number Forms');
  Test('ChrBlock("−")', 'Mathematical Operators');
  Test('ChrBlock("∫")', 'Mathematical Operators');
  Test('ChrBlock("∑")', 'Mathematical Operators');
  Test('ChrBlock("∂")', 'Mathematical Operators');
  Test('ChrBlock("∧")', 'Mathematical Operators');
  Test('ChrBlock("≤")', 'Mathematical Operators');
  Test('ChrBlock("∀")', 'Mathematical Operators');
  Test('ChrBlock("∞")', 'Mathematical Operators');
  Test('ChrBlock("ℕ")', 'Letterlike Symbols');
  Test('ChrBlock("ℝ")', 'Letterlike Symbols');
  Test('ChrBlock("ℂ")', 'Letterlike Symbols');
  Test('ChrBlock("⚘")', 'Miscellaneous Symbols');
  Test('ChrBlock("☘")', 'Miscellaneous Symbols');
  Test('ChrBlock("⌨")', 'Miscellaneous Technical');
  Test('ChrBlock("⌬")', 'Miscellaneous Technical');
  Test('ChrBlock("▉")', 'Block Elements');
  Test('ChrBlock("▲")', 'Geometric Shapes');

  Test('#ChrBlocks() ≥ 262', True);
  Test('contains(ChrBlocks(), "Basic Latin")', True);
  Test('contains(ChrBlocks(), "Latin-1 Supplement")', True);
  Test('contains(ChrBlocks(), "General Punctuation")', True);
  Test('contains(ChrBlocks(), "Greek and Coptic")', True);
  Test('contains(ChrBlocks(), "Cyrillic")', True);
  Test('contains(ChrBlocks(), "Armenian")', True);
  Test('contains(ChrBlocks(), "Hebrew")', True);
  Test('contains(ChrBlocks(), "Arabic")', True);
  Test('contains(ChrBlocks(), "Tamil")', True);
  Test('contains(ChrBlocks(), "Thai")', True);
  Test('contains(ChrBlocks(), "Runic")', True);
  Test('contains(ChrBlocks(), "Mongolian")', True);
  Test('contains(ChrBlocks(), "Braille Patterns")', True);
  Test('contains(ChrBlocks(), "Number Forms")', True);
  Test('contains(ChrBlocks(), "Mathematical Operators")', True);
  Test('contains(ChrBlocks(), "Letterlike Symbols")', True);
  Test('contains(ChrBlocks(), "Miscellaneous Symbols")', True);
  Test('contains(ChrBlocks(), "Miscellaneous Technical")', True);
  Test('contains(ChrBlocks(), "Block Elements")', True);
  Test('contains(ChrBlocks(), "Geometric Shapes")', True);
  Test('¬contains(ChrBlocks(), "Supplemental Death Eater Lullaby Symbols 2")', True);

  Test('ChrBlockRange("Basic Latin")',
    TAlgosimStructure.CreateWithValue(
      ['start', 'end'],
      [ASOInt(0), ASOInt(127)]
    )
  );

  Test('ChrBlockRange("Mathematical Operators")',
    TAlgosimStructure.CreateWithValue(
      ['start', 'end'],
      [ASOInt(8704), ASOInt(8959)]
    )
  );

  Test('ord("'#0'")', 0);
  Test('chr(0)', #0);

  Test('ChrName(chr(0))', '<control> (NULL)');

  Test('ord("A")', 65);
  Test('chr(65)', 'A');

  Test('ord("∫")', $222B);
  Test('chr(16#222B)', '∫');


  //
  // Word wrapping
  //

  Test('s ≔ "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat."',
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.');

  Test('WordWrap(s, 10000)',
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.');

  Test('WordWrap(s, 231)',
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.');

  Test('WordWrap(s, 230)',
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo'#13#10'consequat.');

  Test('WordWrap(s)',
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor'#13#10 +
    'incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis'#13#10 +
    'nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.');

  Test('WordWrap(s, 80)',
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor'#13#10 +
    'incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis'#13#10 +
    'nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.');

  Test('WordWrap(s, 70)',
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do'#13#10 +
    'eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad'#13#10 +
    'minim veniam, quis nostrud exercitation ullamco laboris nisi ut'#13#10 +
    'aliquip ex ea commodo consequat.');

  Test('WordWrap(s, 60)',
    'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed'#13#10 +
    'do eiusmod tempor incididunt ut labore et dolore magna'#13#10 +
    'aliqua. Ut enim ad minim veniam, quis nostrud exercitation'#13#10 +
    'ullamco laboris nisi ut aliquip ex ea commodo consequat.');

  Test('WordWrap(s, 50)',
    'Lorem ipsum dolor sit amet, consectetur adipiscing'#13#10 +
    'elit, sed do eiusmod tempor incididunt ut labore'#13#10 +
    'et dolore magna aliqua. Ut enim ad minim veniam,'#13#10 +
    'quis nostrud exercitation ullamco laboris nisi ut'#13#10 +
    'aliquip ex ea commodo consequat.');

  Test('WordWrap(s, 40)',
    'Lorem ipsum dolor sit amet, consectetur'#13#10 +
    'adipiscing elit, sed do eiusmod tempor'#13#10 +
    'incididunt ut labore et dolore magna'#13#10 +
    'aliqua. Ut enim ad minim veniam, quis'#13#10 +
    'nostrud exercitation ullamco laboris'#13#10 +
    'nisi ut aliquip ex ea commodo consequat.');

  Test('WordWrap(s, 30)',
    'Lorem ipsum dolor sit amet,'#13#10 +
    'consectetur adipiscing elit,'#13#10 +
    'sed do eiusmod tempor'#13#10 +
    'incididunt ut labore et dolore'#13#10 +
    'magna aliqua. Ut enim ad minim'#13#10 +
    'veniam, quis nostrud'#13#10 +
    'exercitation ullamco laboris'#13#10 +
    'nisi ut aliquip ex ea commodo'#13#10 +
    'consequat.');

  Test('WordWrap(s, 20)',
    'Lorem ipsum dolor'#13#10 +
    'sit amet,'#13#10 +
    'consectetur'#13#10 +
    'adipiscing elit, sed'#13#10 +
    'do eiusmod tempor'#13#10 +
    'incididunt ut labore'#13#10 +
    'et dolore magna'#13#10 +
    'aliqua. Ut enim ad'#13#10 +
    'minim veniam, quis'#13#10 +
    'nostrud exercitation'#13#10 +
    'ullamco laboris nisi'#13#10 +
    'ut aliquip ex ea'#13#10 +
    'commodo consequat.');

  Test('WordWrap(s, 10)',
    'Lorem'#13#10 +
    'ipsum'#13#10 +
    'dolor sit'#13#10 +
    'amet,'#13#10 +
    'consectetu'#13#10 +
    'r'#13#10 +
    'adipiscing'#13#10 +
    'elit, sed'#13#10 +
    'do eiusmod'#13#10 +
    'tempor'#13#10 +
    'incididunt'#13#10 +
    'ut labore'#13#10 +
    'et dolore'#13#10 +
    'magna'#13#10 +
    'aliqua. Ut'#13#10 +
    'enim ad'#13#10 +
    'minim'#13#10 +
    'veniam,'#13#10 +
    'quis'#13#10 +
    'nostrud'#13#10 +
    'exercitati'#13#10 +
    'on ullamco'#13#10 +
    'laboris'#13#10 +
    'nisi ut'#13#10 +
    'aliquip ex'#13#10 +
    'ea commodo'#13#10 +
    'consequat.');

  Test('WordWrap(s, 1)',
    'L'#13#10+'o'#13#10+'r'#13#10+'e'#13#10+'m'#13#10+'i'#13#10+'p'#13#10+'s'#13#10+'u'#13#10+'m'#13#10+'d'#13#10+'o'#13#10+'l'#13#10+'o'#13#10+'r'#13#10+'s'#13#10+'i'#13#10+'t'#13#10+'a'#13#10+'m'#13#10+'e'#13#10+'t'#13#10+','#13#10+'c'#13#10+'o'#13#10+'n'#13#10+'s'#13#10+'e'#13#10+'c'#13#10+'t'#13#10+'e'#13#10+'t'#13#10+'u'#13#10+'r'#13#10+'a'#13#10+
    'd'#13#10+'i'#13#10+'p'#13#10+'i'#13#10+'s'#13#10+'c'#13#10+'i'#13#10+'n'#13#10+'g'#13#10+'e'#13#10+'l'#13#10+'i'#13#10+'t'#13#10+','#13#10+'s'#13#10+'e'#13#10+'d'#13#10+'d'#13#10+'o'#13#10+'e'#13#10+'i'#13#10+'u'#13#10+'s'#13#10+'m'#13#10+'o'#13#10+'d'#13#10+'t'#13#10+'e'#13#10+'m'#13#10+'p'#13#10+'o'#13#10+'r'#13#10+'i'#13#10+'n'#13#10+'c'#13#10+
    'i'#13#10+'d'#13#10+'i'#13#10+'d'#13#10+'u'#13#10+'n'#13#10+'t'#13#10+'u'#13#10+'t'#13#10+'l'#13#10+'a'#13#10+'b'#13#10+'o'#13#10+'r'#13#10+'e'#13#10+'e'#13#10+'t'#13#10+'d'#13#10+'o'#13#10+'l'#13#10+'o'#13#10+'r'#13#10+'e'#13#10+'m'#13#10+'a'#13#10+'g'#13#10+'n'#13#10+'a'#13#10+'a'#13#10+'l'#13#10+'i'#13#10+'q'#13#10+'u'#13#10+'a'#13#10+'.'#13#10+
    'U'#13#10+'t'#13#10+'e'#13#10+'n'#13#10+'i'#13#10+'m'#13#10+'a'#13#10+'d'#13#10+'m'#13#10+'i'#13#10+'n'#13#10+'i'#13#10+'m'#13#10+'v'#13#10+'e'#13#10+'n'#13#10+'i'#13#10+'a'#13#10+'m'#13#10+','#13#10+'q'#13#10+'u'#13#10+'i'#13#10+'s'#13#10+'n'#13#10+'o'#13#10+'s'#13#10+'t'#13#10+'r'#13#10+'u'#13#10+'d'#13#10+'e'#13#10+'x'#13#10+'e'#13#10+'r'#13#10+
    'c'#13#10+'i'#13#10+'t'#13#10+'a'#13#10+'t'#13#10+'i'#13#10+'o'#13#10+'n'#13#10+'u'#13#10+'l'#13#10+'l'#13#10+'a'#13#10+'m'#13#10+'c'#13#10+'o'#13#10+'l'#13#10+'a'#13#10+'b'#13#10+'o'#13#10+'r'#13#10+'i'#13#10+'s'#13#10+'n'#13#10+'i'#13#10+'s'#13#10+'i'#13#10+'u'#13#10+'t'#13#10+'a'#13#10+'l'#13#10+'i'#13#10+'q'#13#10+'u'#13#10+'i'#13#10+'p'#13#10+
    'e'#13#10+'x'#13#10+'e'#13#10+'a'#13#10+'c'#13#10+'o'#13#10+'m'#13#10+'m'#13#10+'o'#13#10+'d'#13#10+'o'#13#10+'c'#13#10+'o'#13#10+'n'#13#10+'s'#13#10+'e'#13#10+'q'#13#10+'u'#13#10+'a'#13#10+'t'#13#10+'.');

  Test('WordWrap(s, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');

  Test('WordWrap("Word wrapping is great fun.", 27)', 'Word wrapping is great fun.');
  Test('WordWrap("Word wrapping is great fun.", 26)', 'Word wrapping is great'#13#10'fun.');

  Test('WordWrap("Word wrapping is great fun, isn''t it?", 37)', 'Word wrapping is great fun, isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 36)', 'Word wrapping is great fun, isn''t'#13#10'it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 35)', 'Word wrapping is great fun, isn''t'#13#10'it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 34)', 'Word wrapping is great fun, isn''t'#13#10'it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 33)', 'Word wrapping is great fun, isn''t'#13#10'it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 32)', 'Word wrapping is great fun,'#13#10'isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 31)', 'Word wrapping is great fun,'#13#10'isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 30)', 'Word wrapping is great fun,'#13#10'isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 29)', 'Word wrapping is great fun,'#13#10'isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 28)', 'Word wrapping is great fun,'#13#10'isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 27)', 'Word wrapping is great fun,'#13#10'isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 26)', 'Word wrapping is great'#13#10'fun, isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 25)', 'Word wrapping is great'#13#10'fun, isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 24)', 'Word wrapping is great'#13#10'fun, isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 23)', 'Word wrapping is great'#13#10'fun, isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 22)', 'Word wrapping is great'#13#10'fun, isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 21)', 'Word wrapping is'#13#10'great fun, isn''t it?');
  Test('WordWrap("Word wrapping is great fun, isn''t it?", 20)', 'Word wrapping is'#13#10'great fun, isn''t it?');

  Test('WordWrap("Word wrapping is great fun, isn’t it?", 37)', 'Word wrapping is great fun, isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 36)', 'Word wrapping is great fun, isn’t'#13#10'it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 35)', 'Word wrapping is great fun, isn’t'#13#10'it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 34)', 'Word wrapping is great fun, isn’t'#13#10'it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 33)', 'Word wrapping is great fun, isn’t'#13#10'it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 32)', 'Word wrapping is great fun,'#13#10'isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 31)', 'Word wrapping is great fun,'#13#10'isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 30)', 'Word wrapping is great fun,'#13#10'isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 29)', 'Word wrapping is great fun,'#13#10'isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 28)', 'Word wrapping is great fun,'#13#10'isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 27)', 'Word wrapping is great fun,'#13#10'isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 26)', 'Word wrapping is great'#13#10'fun, isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 25)', 'Word wrapping is great'#13#10'fun, isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 24)', 'Word wrapping is great'#13#10'fun, isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 23)', 'Word wrapping is great'#13#10'fun, isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 22)', 'Word wrapping is great'#13#10'fun, isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 21)', 'Word wrapping is'#13#10'great fun, isn’t it?');
  Test('WordWrap("Word wrapping is great fun, isn’t it?", 20)', 'Word wrapping is'#13#10'great fun, isn’t it?');

  Test('WordWrap("He’s a crime-fighting canine.")', 'He’s a crime-fighting canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 29)', 'He’s a crime-fighting canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 28)', 'He’s a crime-fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 27)', 'He’s a crime-fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 26)', 'He’s a crime-fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 25)', 'He’s a crime-fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 24)', 'He’s a crime-fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 23)', 'He’s a crime-fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 22)', 'He’s a crime-fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 21)', 'He’s a crime-fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 20)', 'He’s a crime-'#13#10'fighting canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 19)', 'He’s a crime-'#13#10'fighting canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 18)', 'He’s a crime-'#13#10'fighting canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 17)', 'He’s a crime-'#13#10'fighting canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 16)', 'He’s a crime-'#13#10'fighting canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 15)', 'He’s a crime-'#13#10'fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 14)', 'He’s a crime-'#13#10'fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 13)', 'He’s a crime-'#13#10'fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 12)', 'He’s a'#13#10'crime-'#13#10'fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 11)', 'He’s a'#13#10'crime-'#13#10'fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime-fighting canine.", 10)', 'He’s a'#13#10'crime-'#13#10'fighting'#13#10'canine.');

                            {nbhyp}
  Test('WordWrap("He’s a crime‑fighting canine.")', 'He’s a crime‑fighting canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 29)', 'He’s a crime‑fighting canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 28)', 'He’s a crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 27)', 'He’s a crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 26)', 'He’s a crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 25)', 'He’s a crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 24)', 'He’s a crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 23)', 'He’s a crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 22)', 'He’s a crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 21)', 'He’s a crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 20)', 'He’s a'#13#10'crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 19)', 'He’s a'#13#10'crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 18)', 'He’s a'#13#10'crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 17)', 'He’s a'#13#10'crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 16)', 'He’s a'#13#10'crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 15)', 'He’s a'#13#10'crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 14)', 'He’s a'#13#10'crime‑fighting'#13#10'canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 13)', 'He’s a'#13#10'crime‑fightin'#13#10'g canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 12)', 'He’s a'#13#10'crime‑fighti'#13#10'ng canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 11)', 'He’s a'#13#10'crime‑fight'#13#10'ing canine.');
  Test('WordWrap("He’s a crime‑fighting canine.", 10)', 'He’s a'#13#10'crime‑figh'#13#10'ting'#13#10'canine.');

  Test('WordWrap("Jag är 100 % säker.")', 'Jag är 100 % säker.');
  Test('WordWrap("Jag är 100 % säker.", 19)', 'Jag är 100 % säker.');
  Test('WordWrap("Jag är 100 % säker.", 18)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 17)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 16)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 15)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 14)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 13)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 12)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 11)', 'Jag är 100'#13#10'% säker.');
  Test('WordWrap("Jag är 100 % säker.", 10)', 'Jag är 100'#13#10'% säker.');
  Test('WordWrap("Jag är 100 % säker.", 9)', 'Jag är'#13#10'100 %'#13#10'säker.');

                          {nbsp}
  Test('WordWrap("Jag är 100 % säker.")', 'Jag är 100 % säker.');
  Test('WordWrap("Jag är 100 % säker.", 19)', 'Jag är 100 % säker.');
  Test('WordWrap("Jag är 100 % säker.", 18)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 17)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 16)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 15)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 14)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 13)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 12)', 'Jag är 100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 11)', 'Jag är'#13#10'100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 10)', 'Jag är'#13#10'100 %'#13#10'säker.');
  Test('WordWrap("Jag är 100 % säker.", 9)', 'Jag är'#13#10'100 %'#13#10'säker.');

  Test('nl ≔ chr(13) + chr(10); s ≔ "Lorem ipsum dolor sit amet," + nl + "consectetur adipiscing elit," + nl + "sed do eiusmod tempor" + nl + "incididunt ut labore et dolore" + nl + "magna aliqua."',
    'Lorem ipsum dolor sit amet,'#13#10 +
    'consectetur adipiscing elit,'#13#10 +
    'sed do eiusmod tempor'#13#10 +
    'incididunt ut labore et dolore'#13#10 +
    'magna aliqua.'
    );

  Test('WordWrap(s)',
    'Lorem ipsum dolor sit amet,'#13#10 +
    'consectetur adipiscing elit,'#13#10 +
    'sed do eiusmod tempor'#13#10 +
    'incididunt ut labore et dolore'#13#10 +
    'magna aliqua.'
    );

  Test('WordWrap(s, 25)',
    'Lorem ipsum dolor sit'#13#10 +
    'amet,'#13#10 +
    'consectetur adipiscing'#13#10 +
    'elit,'#13#10 +
    'sed do eiusmod tempor'#13#10 +
    'incididunt ut labore et'#13#10 +
    'dolore'#13#10 +
    'magna aliqua.'
    );

  Test('s ≔ "Lorem ipsum dolor sit amet," + 2⋅nl + "consectetur adipiscing elit," + 2⋅nl + "sed do eiusmod tempor incididunt" + 2⋅nl + "ut labore et dolore magna aliqua."',
    'Lorem ipsum dolor sit amet,'#13#10#13#10 +
    'consectetur adipiscing elit,'#13#10#13#10 +
    'sed do eiusmod tempor incididunt'#13#10#13#10 +
    'ut labore et dolore magna aliqua.'
    );

  Test('WordWrap(s)',
    'Lorem ipsum dolor sit amet,'#13#10#13#10 +
    'consectetur adipiscing elit,'#13#10#13#10 +
    'sed do eiusmod tempor incididunt'#13#10#13#10 +
    'ut labore et dolore magna aliqua.'
    );

  Test('WordWrap(s, 25)',
    'Lorem ipsum dolor sit'#13#10'amet,'#13#10#13#10 +
    'consectetur adipiscing'#13#10'elit,'#13#10#13#10 +
    'sed do eiusmod tempor'#13#10'incididunt'#13#10#13#10 +
    'ut labore et dolore magna'#13#10'aliqua.'
    );

  Test('WordWrap("system.properties.registry.keys.count")',
    'system.properties.registry.keys.count');

  Test('WordWrap("system.properties.registry.keys.count", 34)',
    'system.properties.registry.keys.'#13#10'count');

  Test('WordWrap("system.properties.registry.keys.count", 25)',
    'system.properties.'#13#10'registry.keys.count');

  Test('WordWrap("system.properties.registry.keys.count", 18)',
    'system.properties.'#13#10'registry.keys.'#13#10'count');

  Test('WordWrap("system.properties.registry.keys.count", 17)',
    'system.'#13#10'properties.'#13#10'registry.keys.'#13#10'count');

  Test('WordWrap("")', '');
  Test('WordWrap("       ")', '       ');
  Test('WordWrap("       " + nl + "   ")', #13#10'   ');
  Test('WordWrap(nl + nl + nl)', #13#10#13#10#13#10);

  Test('trim(WordWrap("     " + nl + 1000⋅"         " + 1000⋅nl + nl + "   cat   " + nl + nl + 1000⋅"     " + nl))', 'cat');

  Test('delete(s); delete(nl)', success);


  //
  // Word and character extraction
  //

  // Word extraction

  Test('s ≔ "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."',
       'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.');

  Test('words(s)',
    strarr(
      ['Lorem', 'ipsum', 'dolor', 'sit', 'amet', 'consectetur', 'adipiscing', 'elit', 'sed', 'do', 'eiusmod', 'tempor', 'incididunt', 'ut', 'labore', 'et', 'dolore', 'magna', 'aliqua']
    ));

  Test('s ≔ "He''s a super-cute crime-fighting canine.";', null);
  Test('words(s)',
    strarr(
      ['He''s', 'a', 'super-cute', 'crime-fighting', 'canine']
    ));

  Test('s ≔ "It’s not hard to see he hasn’t done it before.";', null);
  Test('words(s)',
    strarr(
      ['It’s', 'not', 'hard', 'to', 'see', 'he', 'hasn’t', 'done', 'it', 'before']
    ));

  Test('s ≔ "He’s done it before, hasn’t he?";', null);
  Test('words(s)',
    strarr(
      ['He’s', 'done', 'it', 'before', 'hasn’t', 'he']
    ));

  Test('s ≔ "What?! Is that what you think?";', null);
  Test('words(s)',
    strarr(
      ['What', 'Is', 'that', 'what', 'you', 'think']
    ));

  Test('s ≔ "What‽ Is that what you think?";', null);
  Test('words(s)',
    strarr(
      ['What', 'Is', 'that', 'what', 'you', 'think']
    ));

  Test('s ≔ "This is because of a hormone (aldosterone) produced by the adrenal glands.";', null);
  Test('words(s)',
    strarr(
      ['This', 'is', 'because', 'of', 'a', 'hormone', 'aldosterone', 'produced', 'by', 'the', 'adrenal', 'glands']
    ));

  Test('s ≔ "Vitamin D3 is produced from 7-dehydrocholesterol in the skin.";', null);
  Test('words(s)',
    strarr(
      ['Vitamin', 'D3', 'is', 'produced', 'from', '7-dehydrocholesterol', 'in', 'the', 'skin']
    ));

  Test('s ≔ "We saw 531 paintings; they were all awesome.";', null);
  Test('words(s)',
    strarr(
      ['We', 'saw', 'paintings', 'they', 'were', 'all', 'awesome']
    ));

  Test('s ≔ "We found that d ≈ 0.591.";', null);
  Test('words(s)',
    strarr(
      ['We', 'found', 'that', 'd']
    ));

  Test('s ≔ "d ≈ 5.7; ε ≤ 0.4";', null);
  Test('words(s)',
    strarr(
      ['d', 'ε']
    ));

  Test('s ≔ "What do you have to say about this ""horrible"" suggestion?";', null);
  Test('words(s)',
    strarr(
      ['What', 'do', 'you', 'have', 'to', 'say', 'about', 'this', 'horrible', 'suggestion']
    ));

  Test('s ≔ "What do you have to say about this ''horrible'' suggestion?";', null);
  Test('words(s)',
    strarr(
      ['What', 'do', 'you', 'have', 'to', 'say', 'about', 'this', 'horrible', 'suggestion']
    ));

  Test('s ≔ "What do you have to say about this “horrible” suggestion?";', null);
  Test('words(s)',
    strarr(
      ['What', 'do', 'you', 'have', 'to', 'say', 'about', 'this', 'horrible', 'suggestion']
    ));

  Test('s ≔ "What do you have to say about this ‘horrible’ suggestion?";', null);
  Test('words(s)',
    strarr(
      ['What', 'do', 'you', 'have', 'to', 'say', 'about', 'this', 'horrible', 'suggestion']
    ));

  Test('s ≔ "He was a good player – and a dear friend – before the accident.";', null);
  Test('words(s)',
    strarr(
      ['He', 'was', 'a', 'good', 'player', 'and', 'a', 'dear', 'friend', 'before', 'the', 'accident']
    ));

  Test('s ≔ "He was a good player—and a dear friend—before the accident.";', null);
  Test('words(s)',
    strarr(
      ['He', 'was', 'a', 'good', 'player', 'and', 'a', 'dear', 'friend', 'before', 'the', 'accident']
    ));

  Test('s ≔ "Compiling...";', null);
  Test('words(s)',
    strarr(
      ['Compiling']
    ));

  Test('s ≔ "Please submit your question/suggestion before 6 pm this Thursday.";', null);
  Test('words(s)',
    strarr(
      ['Please', 'submit', 'your', 'question', 'suggestion', 'before', 'pm', 'this', 'Thursday']
    ));

  Test('s ≔ "Jag köpte en ny tv-apparat som jag kan koppla dvd-spelaren till.";', null);
  Test('words(s)',
    strarr(
      ['Jag', 'köpte', 'en', 'ny', 'tv-apparat', 'som', 'jag', 'kan', 'koppla', 'dvd-spelaren', 'till']
    ));

  Test('s ≔ "En 22-årig grönsaksodlare hittade en 500-kronorssedel under 100-årsjubileet.";', null);
  Test('words(s)',
    strarr(
      ['En', '22-årig', 'grönsaksodlare', 'hittade', 'en', '500-kronorssedel', 'under', '100-årsjubileet']
    ));

  Test('s ≔ "en 5-åring med β-talassemi";', null);
  Test('words(s)',
    strarr(
      ['en', '5-åring', 'med', 'β-talassemi']
    ));

  Test('s ≔ "en C++-programmerare med ledvärk";', null);
  Test('words(s)',
    strarr(
      ['en', 'C++-programmerare', 'med', 'ledvärk']
    ));

  Test('s ≔ "SVT:s ledning kände till sms:ets innehåll.";', null);
  Test('words(s)',
    strarr(
      ['SVT:s', 'ledning', 'kände', 'till', 'sms:ets', 'innehåll']
    ));

  Test('s ≔ "Kan man ta igen en 7–3-förlust?";', null);
  Test('words(s)',
    strarr(
      ['Kan', 'man', 'ta', 'igen', 'en', '7–3-förlust']
    ));

  Test('s ≔ "I have one dog, two rats (!), and three birds (all parakeets).";', null);
  Test('words(s)',
    strarr(
      ['I', 'have', 'one', 'dog', 'two', 'rats', 'and', 'three', 'birds', 'all', 'parakeets']
    ));

  Test('s ≔ "I have three cars: one Volvo, one Peugeot, and one BMW.";', null);
  Test('words(s)',
    strarr(
      ['I', 'have', 'three', 'cars', 'one', 'Volvo', 'one', 'Peugeot', 'and', 'one', 'BMW']
    ));

  Test('s ≔ "It can record more than 560 actions/second.";', null);
  Test('words(s)',
    strarr(
      ['It', 'can', 'record', 'more', 'than', 'actions', 'second']
    ));

  Test('s ≔ "Please read pages 20–34.";', null);
  Test('words(s)',
    strarr(
      ['Please', 'read', 'pages']
    ));

  Test('s ≔ "These toys are designed for kids aged 7–10.";', null);
  Test('words(s)',
    strarr(
      ['These', 'toys', 'are', 'designed', 'for', 'kids', 'aged']
    ));

  Test('s ≔ "What''s your favo(u)rite colo(u)r?";', null);
  Test('words(s)',
    strarr(
      ['What''s', 'your', 'favo(u)rite', 'colo(u)r']
    ));

  Test('s ≔ "Hello, World!" + chr(13) + chr(10) + "How are you?";', null);
  Test('words(s)',
    strarr(
      ['Hello', 'World', 'How', 'are', 'you']
    ));

  Test('s ≔ "acetylkolinesteras";', null);
  Test('words(s)',
    strarr(
      ['acetylkolinesteras']
    ));

  Test('s ≔ "";', null);
  Test('words(s)',
    strarr(
      []
    ));

  Test('s ≔ "         ";', null);
  Test('words(s)',
    strarr(
      []
    ));

  Test('s ≔ "     7!((.    ";', null);
  Test('words(s)',
    strarr(
      []
    ));

  Test('s ≔ ".........";', null);
  Test('words(s)',
    strarr(
      []
    ));

  Test('s ≔ "A million people in lockdown because of Covid-19.";', null);
  Test('words(s)',
    strarr(
      ['A', 'million', 'people', 'in', 'lockdown', 'because', 'of', 'Covid-19']
    ));

  _s :=
    'program randword;'#13#10 +
    ''#13#10 +
    '{$APPTYPE CONSOLE}'#13#10 +
    ''#13#10 +
    '{$R *.res}'#13#10 +
    ''#13#10 +
    'uses'#13#10 +
    '  System.SysUtils;'#13#10 +
    ''#13#10 +
    'function RandomWordLength: Integer;'#13#10 +
    'begin'#13#10 +
    '  Result := 1 + Random(12);'#13#10 +
    'end;'#13#10 +
    ''#13#10 +
    'function RandomWord(ALength: Integer): string;'#13#10 +
    'var'#13#10 +
    '  i: Integer;'#13#10 +
    'begin'#13#10 +
    '  SetLength(Result, ALength);'#13#10 +
    '  for i := 1 to Result.Length do'#13#10 +
    '    Result[i] := Chr(Ord(''a'') + Random(Ord(''z'') - Ord(''a'') + 1));'#13#10 +
    'end;'#13#10 +
    ''#13#10 +
    'begin'#13#10 +
    '  try'#13#10 +
    '    Randomize;'#13#10 +
    '    try'#13#10 +
    '      while True do'#13#10 +
    '      begin'#13#10 +
    '        Writeln(RandomWord(RandomWordLength));'#13#10 +
    '        Sleep(1000);'#13#10 +
    '      end;'#13#10 +
    '    except'#13#10 +
    '      on E: Exception do'#13#10 +
    '        Writeln(E.ClassName, '': '', E.Message);'#13#10 +
    '    end;'#13#10 +
    '  finally'#13#10 +
    '    Writeln(''Done.'');'#13#10 +
    '    Readln;'#13#10 +
    '  end;'#13#10 +
    'end.'#13#10;

  Test('s ≔ "' + _s + '"', _s);

  Test('words(s)', // "wrong" settings
    strarr(
      [
        'program',
        'randword',
        '$APPTYPE',
        'CONSOLE',
        '$R',
        'res',
        'uses',
        'System.SysUtils',
        'function',
        'RandomWordLength',
        'Integer',
        'begin',
        'Result',
        'Random(12',
        'end',
        'function',
        'RandomWord(ALength',
        'Integer',
        'string',
        'var',
        'i',
        'Integer',
        'begin',
        'SetLength(Result',
        'ALength',
        'for',
        'i',
        'to',
        'Result.Length',
        'do',
        'Result[i',
        'Chr(Ord(''a',
        'Random(Ord(''z',
        'Ord(''a',
        'end',
        'begin',
        'try',
        'Randomize',
        'try',
        'while',
        'True',
        'do',
        'begin',
        'Writeln(RandomWord(RandomWordLength',
        'Sleep(1000',
        'end',
        'except',
        'on',
        'E',
        'Exception',
        'do',
        'Writeln(E.ClassName',
        'E.Message',
        'end',
        'finally',
        'Writeln(''Done',
        'Readln',
        'end',
        'end'
      ]
    ));


  Test('words(s, "source code")',
    strarr(
      [
        'program',
        'randword',
        'APPTYPE',
        'CONSOLE',
        'R',
        'res',
        'uses',
        'System',
        'SysUtils',
        'function',
        'RandomWordLength',
        'Integer',
        'begin',
        'Result',
        'Random',
        'end',
        'function',
        'RandomWord',
        'ALength',
        'Integer',
        'string',
        'var',
        'i',
        'Integer',
        'begin',
        'SetLength',
        'Result',
        'ALength',
        'for',
        'i',
        'to',
        'Result',
        'Length',
        'do',
        'Result',
        'i',
        'Chr',
        'Ord',
        'a',
        'Random',
        'Ord',
        'z',
        'Ord',
        'a',
        'end',
        'begin',
        'try',
        'Randomize',
        'try',
        'while',
        'True',
        'do',
        'begin',
        'Writeln',
        'RandomWord',
        'RandomWordLength',
        'Sleep',
        'end',
        'except',
        'on',
        'E',
        'Exception',
        'do',
        'Writeln',
        'E',
        'ClassName',
        'E',
        'Message',
        'end',
        'finally',
        'Writeln',
        'Done',
        'Readln',
        'end',
        'end'
      ]
    ));

  Test('s ≔ "x := fcn(@val) + ptr^; { great, isn''t it? }";', null);
  Test('words(s, "source code")',
    strarr(
      ['x', 'fcn', 'val', 'ptr', 'great', 'isn''t', 'it']
    ));

  Test('s ≔ "∫cos(x)dx = sin(x) + C";', null);
  Test('words(s, "math")',
    strarr(
      ['cos', 'x', 'dx', 'sin', 'x', 'C']
    ));

  // Character extraction

  Test('characters("Hello, World!")', chrarr(['H', 'e', 'l', 'l', 'o', ',', ' ', 'W', 'o', 'r', 'l', 'd', '!']));
  Test('characters("∫⌬dx")', chrarr(['∫', '⌬', 'd', 'x']));
  Test('characters("ABC")', chrarr(['A', 'B', 'C']));
  Test('characters("A")', chrarr(['A']));
  Test('characters("")', chrarr([]));
  Test('characters(10000⋅"A")', chrarr(StringOfChar('A', 10000).ToCharArray));

  Test('characters(1024)', failure, 'An object of type string was expected as argument 1, but an object of type integer was given.');
  Test('characters()', failure, 'Too few arguments. A required argument of type string is missing.');
  Test('characters("cat", "dog")', failure, 'Too many arguments.');





  //
  //
  //  CHAPTER 8
  //  Date and time functions.
  //
  //

  Chapter('Date and time functions');

  FixedTime := EncodeDateTime(2020, 05, 15, 14, 30, 20, 123);
  try


    //
    // Datetime constructors
    //

    // date

    Test('date()', ASODate(2020, 05, 15)); // current date

    Test('date(43966)', ASODate(2020, 05, 15)); // "cast": get date struct from timestamp (double)
    Test('date(32137)', ASODate(1987, 12, 26));
    Test('date(18464)', ASODate(1950, 07, 20));
    Test('date(54989)', ASODate(2050, 07, 20));
    Test('date(43890)', ASODate(2020, 02, 29));
    Test('date(0)', ASODate(1899, 12, 30));
    Test('date(-10)', ASODate(1899, 12, 20));

    Test('date(date(1963, 08, 02))', ASODate(1963, 08, 02)); // "cast": get date struct from date struct (testing outer fcn)
    Test('date(date(1999, 12, 24))', ASODate(1999, 12, 24));
    Test('date(date(2043, 06, 07))', ASODate(2043, 06, 07));

    Test('date(datetime(2004, 06, 05, 18, 12, 32, 123))', ASODate(2004, 06, 05)); // extract date part of datetime struct
    Test('date(datetime(2119, 04, 28, 14, 42, 00))', ASODate(2119, 04, 28));

    Test('date(1963, 08, 02)', ASODate(1963, 08, 02)); // encode a specific date
    Test('date(1999, 12, 24)', ASODate(1999, 12, 24));
    Test('date(2043, 06, 07)', ASODate(2043, 06, 07));

    Test('date(2043, 06, 07, 08)', failure, 'There is no version of this function that can take 4 argument(s).');
    Test('date(2043, 06, 07, 08, 30, 25)', failure, 'There is no version of this function that can take 6 argument(s).');

    // time

    Test('time()', ASOTime(14, 30, 20, 123)); // current time

    Test('time(0.604399571759259259)', ASOTime(14, 30, 20, 123)); // "cast": get time struct from timestamp (double)
    Test('time(0.524268391203703704)', ASOTime(12, 34, 56, 789));
    Test('time(0.5)', ASOTime(12, 00, 00, 000));
    Test('time(0.0)', ASOTime(00, 00, 00, 000));
    Test('time(0.947916666666666667)', ASOTime(22, 45, 00, 000));

    Test('time(time(16, 30, 10, 521))', ASOTime(16, 30, 10, 521)); // "cast": get time struct from time struct (testing outer fcn)
    Test('time(time(18, 07, 03))', ASOTime(18, 07, 03, 000));
    Test('time(time(00, 00, 00))', ASOTime(00, 00, 00, 000));
    Test('time(time(23, 59, 59, 999))', ASOTime(23, 59, 59, 999));

    Test('time(datetime(2004, 06, 05, 18, 12, 32, 123))', ASOTime(18, 12, 32, 123)); // extract time part of datetime struct
    Test('time(datetime(2119, 04, 28, 14, 42, 00))', ASOTime(14, 42, 00, 000));

    Test('time(16, 30, 10, 521)', ASOTime(16, 30, 10, 521)); // encode a specific time
    Test('time(18, 07, 03)', ASOTime(18, 07, 03, 000));
    Test('time(00, 00, 00)', ASOTime(00, 00, 00, 000));
    Test('time(23, 59, 59, 999)', ASOTime(23, 59, 59, 999));

    // now

    Test('now()', ASODateTime(2020, 05, 15, 14, 30, 20, 123));

    // datetime

    Test('datetime()', ASODateTime(2020, 05, 15, 14, 30, 20, 123)); // current time

    Test('datetime(43966.604399571759259259)', ASODateTime(2020, 05, 15, 14, 30, 20, 123)); // "cast": get datetime struct from timestamp (double)
    Test('datetime(32137.524268391203703704)', ASODateTime(1987, 12, 26, 12, 34, 56, 789));
    Test('datetime(18464.5)', ASODateTime(1950, 07, 20, 12, 00, 00, 000));
    Test('datetime(54989.0)', ASODateTime(2050, 07, 20, 00, 00, 00, 000));
    Test('datetime(43890.947916666666666667)', ASODateTime(2020, 02, 29, 22, 45, 00, 000));
    Test('datetime(0.25)', ASODateTime(1899, 12, 30, 06, 00, 00, 000));

    Test('datetime(datetime(1942, 07, 08, 15, 03, 22, 987))', ASODateTime(1942, 07, 08, 15, 03, 22, 987)); // "cast": get datetime struct from datetime struct
    Test('datetime(datetime(2032, 04, 30, 19, 28, 40))', ASODateTime(2032, 04, 30, 19, 28, 40, 000));
    Test('datetime(date(1995, 08, 20))', ASODateTime(1995, 08, 20, 00, 00, 00, 000));
    Test('datetime(time(23, 36, 10))', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');

    Test('datetime(1942, 07, 08, 15, 03, 22, 987)', ASODateTime(1942, 07, 08, 15, 03, 22, 987)); // encode a specific datetime
    Test('datetime(2032, 04, 30, 19, 28, 40)', ASODateTime(2032, 04, 30, 19, 28, 40, 000));
    Test('datetime(2008, 05, 02, 09, 41)', ASODateTime(2008, 05, 02, 09, 41, 00, 000));
    Test('datetime(2008, 05, 02, 09)', ASODateTime(2008, 05, 02, 09, 00, 00, 000));
    Test('datetime(2008, 05, 02)', ASODateTime(2008, 05, 02, 00, 00, 00, 000));
    Test('datetime(2008, 05)', failure, 'An object of type Date was expected as argument 1, but an object of type integer was given.');

    Test('datetime(date(1994, 06, 20), time(18, 50, 55))', ASODateTime(1994, 06, 20, 18, 50, 55, 000)); // get datetime strucutre from date and time structures
    Test('datetime(date(2000, 05, 13), time(09, 50, 11, 456))', ASODateTime(2000, 05, 13, 09, 50, 11, 456));

    // named days relative today

    Test('today()', ASODate(2020, 05, 15));
    Test('yesterday()', ASODate(2020, 05, 14));
    Test('tomorrow()', ASODate(2020, 05, 16));


    //
    // Addition
    //

    // milliseconds

    Test('AddMilliseconds(datetime(2020, 04, 10, 18, 33, 20, 123), 0)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 123));

    Test('AddMilliseconds(datetime(2020, 04, 10, 18, 33, 20, 123), 5)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 128));

    Test('AddMilliseconds(datetime(2020, 04, 10, 18, 33, 20, 123), 1000)',
      ASODateTime(2020, 04, 10, 18, 33, 21, 123));

    Test('AddMilliseconds(datetime(2020, 04, 10, 18, 33, 20, 123), -5)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 118));

    // seconds

    Test('AddSeconds(datetime(2020, 04, 10, 18, 33, 20, 123), 0)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 123));

    Test('AddSeconds(datetime(2020, 04, 10, 18, 33, 20, 123), 5)',
      ASODateTime(2020, 04, 10, 18, 33, 25, 123));

    Test('AddSeconds(datetime(2020, 04, 10, 18, 33, 20, 123), 60)',
      ASODateTime(2020, 04, 10, 18, 34, 20, 123));

    Test('AddSeconds(datetime(2020, 04, 10, 18, 33, 20, 123), -5)',
      ASODateTime(2020, 04, 10, 18, 33, 15, 123));

    Test('AddSeconds(datetime(2020, 04, 10, 18, 33, 20, 123), 123456)',
      ASODateTime(2020, 04, 12, 04, 50, 56, 123));

    // minutes

    Test('AddMinutes(datetime(2020, 04, 10, 18, 33, 20, 123), 0)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 123));

    Test('AddMinutes(datetime(2020, 04, 10, 18, 33, 20, 123), 5)',
      ASODateTime(2020, 04, 10, 18, 38, 20, 123));

    Test('AddMinutes(datetime(2020, 04, 10, 18, 33, 20, 123), 60)',
      ASODateTime(2020, 04, 10, 19, 33, 20, 123));

    Test('AddMinutes(datetime(2020, 04, 10, 18, 33, 20, 123), -5)',
      ASODateTime(2020, 04, 10, 18, 28, 20, 123));

    // hours

    Test('AddHours(datetime(2020, 04, 10, 18, 33, 20, 123), 0)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 123));

    Test('AddHours(datetime(2020, 04, 10, 18, 33, 20, 123), 5)',
      ASODateTime(2020, 04, 10, 23, 33, 20, 123));

    Test('AddHours(datetime(2020, 04, 10, 18, 33, 20, 123), 24)',
      ASODateTime(2020, 04, 11, 18, 33, 20, 123));

    Test('AddHours(datetime(2020, 04, 10, 18, 33, 20, 123), -5)',
      ASODateTime(2020, 04, 10, 13, 33, 20, 123));

    // days

    Test('AddDays(datetime(2020, 04, 10, 18, 33, 20, 123), 0)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 123));

    Test('AddDays(datetime(2020, 04, 10, 18, 33, 20, 123), 5)',
      ASODateTime(2020, 04, 15, 18, 33, 20, 123));

    Test('AddDays(datetime(2020, 04, 10, 18, 33, 20, 123), 30)',
      ASODateTime(2020, 05, 10, 18, 33, 20, 123));

    Test('AddDays(datetime(2020, 04, 10, 18, 33, 20, 123), -5)',
      ASODateTime(2020, 04, 5, 18, 33, 20, 123));

    // weeks

    Test('AddWeeks(datetime(2020, 04, 10, 18, 33, 20, 123), 0)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 123));

    Test('AddWeeks(datetime(2020, 04, 10, 18, 33, 20, 123), 5)',
      ASODateTime(2020, 05, 15, 18, 33, 20, 123));

    Test('AddWeeks(datetime(2020, 04, 10, 18, 33, 20, 123), -1)',
      ASODateTime(2020, 04, 3, 18, 33, 20, 123));

    // months (special)

    Test('AddMonths(datetime(2020, 04, 10, 18, 33, 20, 123), 0)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 123));

    Test('AddMonths(datetime(2020, 04, 10, 18, 33, 20, 123), 5)',
      ASODateTime(2020, 09, 10, 18, 33, 20, 123));

    Test('AddMonths(datetime(2020, 04, 10, 18, 33, 20, 123), 12)',
      ASODateTime(2021, 04, 10, 18, 33, 20, 123));

    Test('AddMonths(datetime(2020, 04, 10, 18, 33, 20, 123), 1)',
      ASODateTime(2020, 05, 10, 18, 33, 20, 123));

    Test('AddMonths(datetime(2020, 04, 10, 18, 33, 20, 123), -2)',
      ASODateTime(2020, 02, 10, 18, 33, 20, 123));

    Test('AddMonths(datetime(2020, 04, 10, 18, 33, 20, 123), -5)',
      ASODateTime(2019, 11, 10, 18, 33, 20, 123));

    Test('AddMonths(datetime(2020, 01, 29, 18, 33, 20, 123), 1)',
      ASODateTime(2020, 02, 29, 18, 33, 20, 123));

    Test('AddMonths(datetime(2020, 01, 30, 18, 33, 20, 123), 1)',
      ASODateTime(2020, 02, 29, 18, 33, 20, 123));

    Test('AddMonths(datetime(2020, 03, 30, 18, 33, 20, 123), -1)',
      ASODateTime(2020, 02, 29, 18, 33, 20, 123));

    Test('AddMonths(datetime(2021, 01, 30, 18, 33, 20, 123), 1)',
      ASODateTime(2021, 02, 28, 18, 33, 20, 123));

    Test('AddMonths(datetime(2021, 03, 30, 18, 33, 20, 123), -1)',
      ASODateTime(2021, 02, 28, 18, 33, 20, 123));

    Test('AddMonths(datetime(2021, 01, 31, 18, 33, 20, 123), 1)',
      ASODateTime(2021, 02, 28, 18, 33, 20, 123));

    Test('AddMonths(datetime(2021, 03, 31, 18, 33, 20, 123), -1)',
      ASODateTime(2021, 02, 28, 18, 33, 20, 123));

    Test('AddMonths(datetime(2021, 03, 31, 18, 33, 20, 123), 1)',
      ASODateTime(2021, 04, 30, 18, 33, 20, 123));

    Test('AddMonths(datetime(2020, 02, 29, 18, 33, 20, 123), 12)',
      ASODateTime(2021, 02, 28, 18, 33, 20, 123));

    // years

    Test('AddYears(datetime(2020, 04, 10, 18, 33, 20, 123), 0)',
      ASODateTime(2020, 04, 10, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 04, 10, 18, 33, 20, 123), 1)',
      ASODateTime(2021, 04, 10, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 04, 10, 18, 33, 20, 123), 10)',
      ASODateTime(2030, 04, 10, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 04, 10, 18, 33, 20, 123), 20)',
      ASODateTime(2040, 04, 10, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 04, 10, 18, 33, 20, 123), 100)',
      ASODateTime(2120, 04, 10, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 04, 10, 18, 33, 20, 123), 1000)',
      ASODateTime(3020, 04, 10, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 04, 10, 18, 33, 20, 123), -1)',
      ASODateTime(2019, 04, 10, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 02, 29, 18, 33, 20, 123), 1)',
      ASODateTime(2021, 02, 28, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 02, 29, 18, 33, 20, 123), -1)',
      ASODateTime(2019, 02, 28, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 02, 29, 18, 33, 20, 123), 4)',
      ASODateTime(2024, 02, 29, 18, 33, 20, 123));

    Test('AddYears(datetime(2020, 02, 29, 18, 33, 20, 123), 5)',
      ASODateTime(2025, 02, 28, 18, 33, 20, 123));


    //
    // Validation
    //

    Test('IsValidDate(date(2020, 04, 10))', True);
    Test('IsValidDate(date(1987, 12, 26))', True);
    Test('IsValidDate(date(1940, 06, 18))', True);
    Test('IsValidDate(date(2150, 08, 01))', True);
    Test('IsValidDate(date(2000, 02, 29))', True);
    Test('IsValidDate(date(2020, 02, 29))', True);
    Test('IsValidDate(date(2400, 02, 29))', True);

    Test('IsValidDate(date(2020, 13, 10))', False);
    Test('IsValidDate(date(2020, 0, 10))', False);
    Test('IsValidDate(date(2020, -2, 10))', False);
    Test('IsValidDate(date(2020, 04, 0))', False);
    Test('IsValidDate(date(2020, 04, 32))', False);
    Test('IsValidDate(date(2020, 04, 100))', False);
    Test('IsValidDate(date(2020, 02, 30))', False);
    Test('IsValidDate(date(2020, 04, 31))', False);
    Test('IsValidDate(date(2020, 11, 31))', False);
    Test('IsValidDate(date(2020, 04, -5))', False);
    Test('IsValidDate(date(2300, 02, 29))', False);
    Test('IsValidDate(date(1900, 02, 29))', False);

    Test('IsValidTime(time(20, 40, 30))', True);
    Test('IsValidTime(time(20, 40, 30, 000))', True);
    Test('IsValidTime(time(20, 40, 30, 123))', True);
    Test('IsValidTime(time(20, 40, 30, 999))', True);
    Test('IsValidTime(time(20, 40, 00))', True);
    Test('IsValidTime(time(23, 59, 59))', True);
    Test('IsValidTime(time(23, 59, 59, 999))', True);
    Test('IsValidTime(time(00, 00, 00))', True);

    Test('IsValidTime(time(24, 40, 30))', False);
    Test('IsValidTime(time(25, 40, 30))', False);
    Test('IsValidTime(time(200, 40, 30))', False);
    Test('IsValidTime(time(-1, 40, 30))', False);
    Test('IsValidTime(time(-2, 40, 30))', False);
    Test('IsValidTime(time(20, 60, 30))', False);
    Test('IsValidTime(time(20, 120, 30))', False);
    Test('IsValidTime(time(20, -2, 30))', False);
    Test('IsValidTime(time(20, 40, 60))', False);
    Test('IsValidTime(time(20, 40, 70))', False);
    Test('IsValidTime(time(20, 40, -1))', False);
    Test('IsValidTime(time(20, 40, 30, 1000))', False);
    Test('IsValidTime(time(20, 40, 30, 1234))', False);
    Test('IsValidTime(time(20, 40, 30, -2))', False);

    Test('IsValidDatetime(datetime(2020, 04, 10, 20, 45, 30, 123))', True);
    Test('IsValidDatetime(datetime(2020, 02, 29, 20, 45, 30, 123))', True);
    Test('IsValidDatetime(datetime(2000, 02, 29, 20, 45, 30, 123))', True);
    Test('IsValidDatetime(datetime(2400, 02, 29, 20, 45, 30, 123))', True);
    Test('IsValidDatetime(datetime(1987, 12, 26, 12, 30, 30, 000))', True);
    Test('IsValidDatetime(datetime(2610, 10, 02, 09, 10, 20, 000))', True);
    Test('IsValidDatetime(datetime(1920, 05, 07, 13, 08, 02, 000))', True);
    Test('IsValidDatetime(datetime(1999, 12, 31, 12, 59, 59, 999))', True);
    Test('IsValidDatetime(datetime(2000, 01, 01, 00, 00, 00, 000))', True);
    Test('IsValidDatetime(datetime(1900, 02, 28, 09, 52, 12, 766))', True);
    Test('IsValidDatetime(datetime(2062, 07, 19, 07, 41, 20, 547))', True);
    Test('IsValidDatetime(datetime(2020, 04, 05, 22, 17, 52, 711))', True);

    Test('IsValidDatetime(datetime(2020, 00, 10, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 13, 10, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, -1, 10, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 00, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 31, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 05, 32, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 50, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, -5, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 24, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 28, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, -1, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 20, 60, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 20, 80, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 20, -2, 30, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 20, 45, 60, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 20, 45, 90, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 20, 45, -4, 123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 20, 45, 30, 1123))', False);
    Test('IsValidDatetime(datetime(2020, 04, 10, 20, 45, 30, -50))', False);
    Test('IsValidDatetime(datetime(2021, 02, 29, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(2100, 02, 29, 20, 45, 30, 123))', False);
    Test('IsValidDatetime(datetime(1900, 02, 29, 01, 50, 00, 000))', False);


    //
    // Distance
    //

    // milliseconds

    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 123))', 123.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 123), datetime(2020, 04, 10, 21, 00, 00, 000))', 123.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 30, 000))', 30000.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 30, 00, 000))', 30*60*1000.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 30, 00, 123))', 30*60*1000.0+123.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 23, 00, 00, 000))', 2*60*60*1000.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 15, 21, 00, 00, 000))', 5*24*60*60*1000.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 05, 10, 21, 00, 00, 000))', 30*24*60*60*1000.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2021, 04, 10, 21, 00, 00, 000))', 365*24*60*60*1000.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2021, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 365*24*60*60*1000.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2022, 04, 10, 21, 00, 00, 000))', 2*365*24*60*60*1000.0);
    Eps(1E-3); Test('MillisecondsBetween(datetime(2019, 04, 10, 21, 00, 00, 000), datetime(2021, 04, 10, 21, 00, 00, 000))', (366+365)*24*60*60*1000.0);

    Eps(1E-3); Test('MillisecondsBetween(date(2020, 04, 10), date(2020, 05, 10))', 30*24*60*60*1000.0);
    Test('MillisecondsBetween(time(14, 20, 30), time(14, 50, 00))', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');

    // seconds

    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.0);
    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 123))', 0.123);
    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 123), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.123);
    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 20, 000))', 20.0);
    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 10, 00, 000))', 600.0);
    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 10, 23, 000))', 623.0);
    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 10, 23, 500))', 623.5);
    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 22, 00, 00, 000))', 3600.0);
    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 15, 21, 00, 00, 000))', 5*24*60*60.0);
    Eps(1E-6); Test('SecondsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 05, 10, 21, 00, 00, 000))', 30*24*60*60.0);
    Eps(1E-6); Test('SecondsBetween(datetime(2021, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 365*24*60*60.0);
    Eps(1E-6); Test('SecondsBetween(datetime(2022, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 2*365*24*60*60.0);
    Eps(1E-6); Test('SecondsBetween(datetime(2022, 04, 10, 21, 00, 00, 000), datetime(2019, 04, 10, 21, 00, 00, 000))', (2*365+366)*24*60*60.0);

    Eps(1E-6); Test('SecondsBetween(date(2020, 04, 10), date(2020, 04, 15))', 5*24*60*60.0);
    Test('SecondsBetween(time(14, 20, 30), time(14, 50, 00))', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');

    // minutes

    Eps(1E-8); Test('MinutesBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.0);
    Eps(1E-8); Test('MinutesBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 123))', 0.123/60);
    Eps(1E-8); Test('MinutesBetween(datetime(2020, 04, 10, 21, 00, 00, 123), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.123/60);
    Eps(1E-8); Test('MinutesBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 57, 000))', 57/60);
    Eps(1E-8); Test('MinutesBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 45, 00, 000))', 45.0);
    Eps(1E-8); Test('MinutesBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 23, 00, 00, 000))', 120.0);
    Eps(1E-8); Test('MinutesBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 23, 15, 00, 000))', 135.0);
    Eps(1E-8); Test('MinutesBetween(datetime(2020, 04, 12, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 2*24*60.0);
    Eps(1E-8); Test('MinutesBetween(datetime(2020, 06, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', (30+31)*24*60.0);
    Eps(1E-8); Test('MinutesBetween(datetime(2021, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 365*24*60.0);
    Eps(1E-8); Test('MinutesBetween(datetime(2022, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 2*365*24*60.0);
    Eps(1E-8); Test('MinutesBetween(datetime(2023, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 3*365*24*60.0);
    Eps(1E-8); Test('MinutesBetween(datetime(2024, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', (3*365 + 366)*24*60.0);

    Eps(1E-8); Test('MinutesBetween(date(2020, 04, 12), date(2020, 04, 10))', 2*24*60.0);
    Test('MinutesBetween(time(14, 20, 30), time(14, 50, 00))', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');

    // hours

    Eps(1E-9); Test('HoursBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.0);
    Eps(1E-9); Test('HoursBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 123))', 0.123/3600);
    Eps(1E-9); Test('HoursBetween(datetime(2020, 04, 10, 21, 00, 00, 123), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.123/3600);
    Eps(1E-9); Test('HoursBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 40, 000))', 40/3600);
    Eps(1E-9); Test('HoursBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 20, 00, 000))', 20/60);
    Eps(1E-9); Test('HoursBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 20, 40, 000))', 20/60 + 40/3600);
    Eps(1E-9); Test('HoursBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 11, 05, 00, 00, 000))', 8.0);
    Eps(1E-9); Test('HoursBetween(datetime(2020, 04, 11, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 24.0);
    Eps(1E-9); Test('HoursBetween(datetime(2020, 05, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 30*24.0);
    Eps(1E-9); Test('HoursBetween(datetime(2021, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 365*24.0);
    Eps(1E-9); Test('HoursBetween(datetime(2022, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 2*365*24.0);
    Eps(1E-9); Test('HoursBetween(datetime(2023, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 3*365*24.0);
    Eps(1E-9); Test('HoursBetween(datetime(2024, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', (3*365+366)*24.0);

    Eps(1E-9); Test('HoursBetween(date(2020, 05, 10), date(2020, 04, 10))', 30*24.0);
    Test('HoursBetween(time(14, 20, 30), time(14, 50, 00))', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');

    // days

    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.0);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 123))', 0.123/86400);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 123), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.123/86400);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 10, 000))', 10/86400);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 50, 00, 000))', 50/1440);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 50, 30, 000))', 50/1440 + 30/86400);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 11, 00, 00, 00, 000))', 3/24);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 17, 21, 00, 00, 000))', 7.0);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 05, 10, 21, 00, 00, 000))', 30.0);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2021, 04, 10, 21, 00, 00, 000))', 365.0);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2022, 04, 10, 21, 00, 00, 000))', 2*365.0);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2023, 04, 10, 21, 00, 00, 000))', 3*365.0);
    Eps(1E-11); Test('DaysBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2024, 04, 10, 21, 00, 00, 000))', 3*365.0 + 366);

    Eps(1E-11); Test('DaysBetween(date(2020, 04, 10), date(2023, 04, 10))', 3*365.0);
    Test('DaysBetween(time(14, 20, 30), time(14, 50, 00))', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');

    // weeks

    Eps(1E-11); Test('WeeksBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.0);
    Eps(1E-11); Test('WeeksBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 123))', 0.123/604800);
    Eps(1E-11); Test('WeeksBetween(datetime(2020, 04, 10, 21, 00, 00, 123), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.123/604800);
    Eps(1E-11); Test('WeeksBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 30, 000))', 30/604800);
    Eps(1E-11); Test('WeeksBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 20, 00, 000))', 20/10080);
    Eps(1E-11); Test('WeeksBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 22, 00, 00, 000))', 1/168);
    Eps(1E-11); Test('WeeksBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 12, 21, 00, 00, 000))', 2/7);
    Eps(1E-11); Test('WeeksBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 30, 21, 00, 00, 000))', 20/7);
    Eps(1E-11); Test('WeeksBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 06, 10, 21, 00, 00, 000))', (30+31)/7);
    Eps(1E-11); Test('WeeksBetween(datetime(2021, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 365/7);
    Eps(1E-11); Test('WeeksBetween(datetime(2022, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 2*365/7);
    Eps(1E-11); Test('WeeksBetween(datetime(2023, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 3*365/7);
    Eps(1E-11); Test('WeeksBetween(datetime(2024, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', (3*365 + 366)/7);

    Eps(1E-11); Test('WeeksBetween(date(2020, 04, 10), date(2020, 06, 10))', (30+31)/7);
    Test('WeeksBetween(time(14, 20, 30), time(14, 50, 00))', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');

    // months

    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.0);
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 123))', 0.123/(60*60*24*30.4375));
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 123), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.123/(60*60*24*30.4375));
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 15, 000))', 15/(60*60*24*30.4375));
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 45, 00, 000))', 45/(60*24*30.4375));
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 22, 00, 00, 000))', 1/(24*30.4375));
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 17, 21, 00, 00, 000))', 7/30.4375);
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 05, 10, 21, 00, 00, 000))', 30/30.4375);
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2021, 04, 10, 21, 00, 00, 000))', 365/30.4375);
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2022, 04, 10, 21, 00, 00, 000))', 2*365/30.4375);
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2023, 04, 10, 21, 00, 00, 000))', 3*365/30.4375);
    Eps(1E-12); Test('MonthsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2024, 04, 10, 21, 00, 00, 000))', 4*12.0);

    Eps(1E-12); Test('MonthsBetween(date(2020, 04, 10), date(2021, 04, 10))', 365/30.4375);
    Test('MonthsBetween(time(14, 20, 30), time(14, 50, 00))', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');

    // years

    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.0);
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 00, 123))', 0.123/(60*60*24*365.25));
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 123), datetime(2020, 04, 10, 21, 00, 00, 000))', 0.123/(60*60*24*365.25));
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 00, 17, 000))', 17/(60*60*24*365.25));
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 21, 20, 00, 000))', 20/(60*24*365.25));
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 10, 22, 00, 00, 000))', 1/(24*365.25));
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 04, 19, 21, 00, 00, 000))', 9/365.25);
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2020, 05, 10, 21, 00, 00, 000))', 30/365.25);
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2021, 04, 10, 21, 00, 00, 000))', 365/365.25);
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2022, 04, 10, 21, 00, 00, 000))', 2*365/365.25);
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2023, 04, 10, 21, 00, 00, 000))', 3*365/365.25);
    Eps(1E-13); Test('YearsBetween(datetime(2020, 04, 10, 21, 00, 00, 000), datetime(2024, 04, 10, 21, 00, 00, 000))', 4.0);

    Eps(1E-13); Test('YearsBetween(date(2020, 04, 10), date(2024, 04, 10))', 4.0);
    Eps(1E-13); Test('YearsBetween(date(2020, 04, 10), date(2020, 06, 10))', (30+31) / 365.25);
    Test('YearsBetween(time(14, 20, 30), time(14, 50, 00))', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');


    //
    // Extraction
    //

    Test('DayOfTheWeek()', 5);
    Test('DayOfTheWeek(now())', 5);
    Test('DayOfTheWeek(date())', 5);
    Test('DayOfTheWeek(today())', 5);
    Test('DayOfTheWeek(yesterday())', 4);
    Test('DayOfTheWeek(tomorrow())', 6);
    Test('DayOfTheWeek(time())', failure, 'An object of type DateTime was expected as argument 1, but an object of type structure of type "Time" was given.');

    Test('DayOfTheWeek(date(2020, 04, 11))', 6);
    Test('DayOfTheWeek(date(2020, 04, 12))', 7); // Sunday
    Test('DayOfTheWeek(date(2020, 04, 13))', 1); // Monday

    Test('DayOfTheWeek(date(1944, 06, 14))', 3);
    Test('DayOfTheWeek(date(2076, 08, 20))', 4);

    // -

    Test('DayOfTheMonth()', 15);
    Test('DayOfTheMonth(date(1944, 06, 14))', 14);
    Test('DayOfTheMonth(date(2076, 08, 20))', 20);

    // -

    Test('DayOfTheYear()', 136);
    Test('DayOfTheYear(date(1944, 06, 14))', 166);
    Test('DayOfTheYear(date(2077, 08, 20))', 232);

    // -

    Test('WeekOfTheYear()', 20);
    Test('WeekOfTheYear(date(1944, 06, 14))', 24);
    Test('WeekOfTheYear(date(2076, 08, 20))', 34);

    // -

    Test('SecondOfTheDay()', 52220);
    Test('SecondOfTheDay(date(1944, 06, 14))', 0);
    Test('SecondOfTheDay(date(2076, 08, 20))', 0);

    Test('SecondOfTheDay(datetime(1944, 06, 14, 08, 45, 30))', 31530);
    Test('SecondOfTheDay(datetime(2076, 08, 20, 20, 14, 07, 610))', 72847{.61});

    // -

    Test('SecondOfTheWeek()', 4*86400 + 52220);
    Test('SecondOfTheWeek(date(1944, 06, 14))', 2*24*60*60);
    Test('SecondOfTheWeek(date(2076, 08, 20))', 3*24*60*60);

    Test('SecondOfTheWeek(datetime(1944, 06, 14, 08, 45, 30))', 2*24*60*60 + 31530);
    Test('SecondOfTheWeek(datetime(2076, 08, 20, 20, 14, 07, 610))', 3*24*60*60 + 72847{.61});

    // -

    Test('SecondOfTheMonth()', 14*86400 + 52220);
    Test('SecondOfTheMonth(date(1944, 06, 14))', 13*24*60*60);
    Test('SecondOfTheMonth(date(2076, 08, 20))', 19*24*60*60);

    Test('SecondOfTheMonth(datetime(1944, 06, 14, 08, 45, 30))', 13*24*60*60 + 31530);
    Test('SecondOfTheMonth(datetime(2076, 08, 20, 20, 14, 07, 610))', 19*24*60*60 + 72847{.61});

    // -

    Test('SecondOfTheYear()', 11716220);
    Test('SecondOfTheYear(date(1944, 06, 14))', 165*24*60*60);
    Test('SecondOfTheYear(date(2077, 08, 20))', 231*24*60*60);

    Test('SecondOfTheYear(datetime(1944, 06, 14, 08, 45, 30))', 165*24*60*60 + 31530);
    Test('SecondOfTheYear(datetime(2077, 08, 20, 20, 14, 07, 610))', 231*24*60*60 + 72847{.61});

    // -

    Test('MillisecondOfTheDay()', 52220123);
    Test('MillisecondOfTheDay(date(1944, 06, 14))', 1000*(0));
    Test('MillisecondOfTheDay(date(2076, 08, 20))', 1000*(0));

    Test('MillisecondOfTheDay(datetime(1944, 06, 14, 08, 45, 30))', 1000*(31530));
    Test('MillisecondOfTheDay(datetime(2076, 08, 20, 20, 14, 07, 610))', 1000*(72847{.61}) + 610);

    // -

    Test('MillisecondOfTheWeek()', 397820123);
    Test('MillisecondOfTheWeek(date(1944, 06, 14))', 1000*(2*24*60*60));
    Test('MillisecondOfTheWeek(date(2076, 08, 20))', 1000*(3*24*60*60));

    Test('MillisecondOfTheWeek(datetime(1944, 06, 14, 08, 45, 30))', 1000*(2*24*60*60 + 31530));
    Test('MillisecondOfTheWeek(datetime(2076, 08, 20, 20, 14, 07, 610))', 1000*(3*24*60*60 + 72847{.61}) + 610);

    // -

    Test('MillisecondOfTheMonth()', 1261820123);
    Test('MillisecondOfTheMonth(date(1944, 06, 14))', 1000*(13*24*60*60));
    Test('MillisecondOfTheMonth(date(2076, 08, 20))', 1000*(19*24*60*60));

    Test('MillisecondOfTheMonth(datetime(1944, 06, 14, 08, 45, 30))', 1000*(13*24*60*60 + 31530));
    Test('MillisecondOfTheMonth(datetime(2076, 08, 20, 20, 14, 07, 610))', 1000*(19*24*60*60 + 72847{.61}) + 610);

    // -

    Test('MillisecondOfTheYear()', 11716220123);
    Test('MillisecondOfTheYear(date(1944, 06, 14))', Int64(1000)*(165*24*60*60));
    Test('MillisecondOfTheYear(date(2077, 08, 20))', Int64(1000)*(231*24*60*60));

    Test('MillisecondOfTheYear(datetime(1944, 06, 14, 08, 45, 30))', Int64(1000)*(165*24*60*60 + 31530));
    Test('MillisecondOfTheYear(datetime(2077, 08, 20, 20, 14, 07, 610))', Int64(1000)*(231*24*60*60 + 72847{.61}) + 610);


    //
    // Miscellaneous functions
    //

    Test('IsLeapYear(1996)', True);
    Test('IsLeapYear(1997)', False);
    Test('IsLeapYear(1998)', False);
    Test('IsLeapYear(1999)', False);
    Test('IsLeapYear(2000)', True);
    Test('IsLeapYear(2001)', False);
    Test('IsLeapYear(2002)', False);
    Test('IsLeapYear(2003)', False);
    Test('IsLeapYear(2004)', True);
    Test('IsLeapYear(2005)', False);
    Test('IsLeapYear(2006)', False);
    Test('IsLeapYear(2007)', False);
    Test('IsLeapYear(2008)', True);

    Test('IsLeapYear(2100)', False);
    Test('IsLeapYear(2200)', False);
    Test('IsLeapYear(2300)', False);
    Test('IsLeapYear(2400)', True);

    // -

    Test('DaysInYear(1996)', 366);
    Test('DaysInYear(1997)', 365);
    Test('DaysInYear(1998)', 365);
    Test('DaysInYear(1999)', 365);
    Test('DaysInYear(2000)', 366);
    Test('DaysInYear(2001)', 365);
    Test('DaysInYear(2002)', 365);
    Test('DaysInYear(2003)', 365);
    Test('DaysInYear(2004)', 366);
    Test('DaysInYear(2005)', 365);
    Test('DaysInYear(2006)', 365);
    Test('DaysInYear(2007)', 365);
    Test('DaysInYear(2008)', 366);

    Test('DaysInYear(2100)', 365);
    Test('DaysInYear(2200)', 365);
    Test('DaysInYear(2300)', 365);
    Test('DaysInYear(2400)', 366);

    // -

    Test('DaysInMonth(1996, 2)', 29);
    Test('DaysInMonth(1997, 2)', 28);
    Test('DaysInMonth(1998, 2)', 28);
    Test('DaysInMonth(1999, 2)', 28);
    Test('DaysInMonth(2000, 2)', 29);
    Test('DaysInMonth(2001, 2)', 28);
    Test('DaysInMonth(2002, 2)', 28);
    Test('DaysInMonth(2003, 2)', 28);
    Test('DaysInMonth(2004, 2)', 29);
    Test('DaysInMonth(2005, 2)', 28);
    Test('DaysInMonth(2006, 2)', 28);
    Test('DaysInMonth(2007, 2)', 28);
    Test('DaysInMonth(2008, 2)', 29);

    Test('DaysInMonth(2100, 2)', 28);
    Test('DaysInMonth(2200, 2)', 28);
    Test('DaysInMonth(2300, 2)', 28);
    Test('DaysInMonth(2400, 2)', 29);

    Test('DaysInMonth(1996, 3)', 31);
    Test('DaysInMonth(1997, 3)', 31);
    Test('DaysInMonth(1998, 3)', 31);
    Test('DaysInMonth(1999, 3)', 31);
    Test('DaysInMonth(2000, 3)', 31);
    Test('DaysInMonth(2001, 3)', 31);
    Test('DaysInMonth(2002, 3)', 31);
    Test('DaysInMonth(2003, 3)', 31);
    Test('DaysInMonth(2004, 3)', 31);
    Test('DaysInMonth(2005, 3)', 31);
    Test('DaysInMonth(2006, 3)', 31);
    Test('DaysInMonth(2007, 3)', 31);
    Test('DaysInMonth(2008, 3)', 31);

    Test('DaysInMonth(2100, 3)', 31);
    Test('DaysInMonth(2200, 3)', 31);
    Test('DaysInMonth(2300, 3)', 31);
    Test('DaysInMonth(2400, 3)', 31);

    // -

    Eps; Test('timestamp()', 43966.604399571759259291);
    Eps; Test('timestamp(now())', 43966.604399571759259291);
    Eps; Test('timestamp(datetime())', 43966.604399571759259291);
    Eps; Test('timestamp(date())', 43966.0);
    Eps(1E-10); Test('timestamp(time())', 0.604399571759259291);
    Eps; Test('timestamp(today())', 43966.0);
    Eps; Test('timestamp(yesterday())', 43965.0);
    Eps; Test('timestamp(tomorrow())', 43967.0);

    Eps; Test('timestamp(date(2050, 07, 05))', 54974.0);
    Eps; Test('timestamp(datetime(2050, 07, 05, 09, 20, 00, 000))', 54974.388888888888888889);
    Eps; Test('timestamp(time(09, 20, 00, 000))', 0.388888888888888889);

    Eps; Test('timestamp(54974.0)', 54974.0);
    Eps; Test('timestamp(54974.388888888888888889)', 54974.388888888888888889);
    Eps; Test('timestamp(0.388888888888888889)', 0.388888888888888889);


    //
    // Truncation
    //

    Test('TruncToMillisecond(now())', ASODateTime(2020, 05, 15, 14, 30, 20, 123));
    Test('TruncToSecond(now())', ASODateTime(2020, 05, 15, 14, 30, 20, 000));
    Test('TruncToMinute(now())', ASODateTime(2020, 05, 15, 14, 30, 00, 000));
    Test('TruncToHour(now())', ASODateTime(2020, 05, 15, 14, 00, 00, 000));
    Test('TruncToDay(now())', ASODateTime(2020, 05, 15, 00, 00, 00, 000));
    Test('TruncToMonth(now())', ASODateTime(2020, 05, 01, 00, 00, 00, 000));
    Test('TruncToYear(now())', ASODateTime(2020, 01, 01, 00, 00, 00, 000));

    Test('TruncToMillisecond(timestamp(now()))', ASOToDateTimeFree(ASODateTime(2020, 05, 15, 14, 30, 20, 123)));
    Test('TruncToSecond(timestamp(now()))', ASOToDateTimeFree(ASODateTime(2020, 05, 15, 14, 30, 20, 000)));
    Test('TruncToMinute(timestamp(now()))', ASOToDateTimeFree(ASODateTime(2020, 05, 15, 14, 30, 00, 000)));
    Test('TruncToHour(timestamp(now()))', ASOToDateTimeFree(ASODateTime(2020, 05, 15, 14, 00, 00, 000)));
    Test('TruncToDay(timestamp(now()))', ASOToDateTimeFree(ASODateTime(2020, 05, 15, 00, 00, 00, 000)));
    Test('TruncToMonth(timestamp(now()))', ASOToDateTimeFree(ASODateTime(2020, 05, 01, 00, 00, 00, 000)));
    Test('TruncToYear(timestamp(now()))', ASOToDateTimeFree(ASODateTime(2020, 01, 01, 00, 00, 00, 000)));

    Eps(1E-11); Test('TruncToMillisecond(43966.60439957703703703703703703703703703703703703703703703)', 43966.60439957175925925925925925925925925925925925925925925);


    //
    // Formatting
    //

    Test('FormatDateTime(now())', '2020-05-15 14:30:20');

    Test('FormatDateTime(now(), "yyyy""-""mm""-""dd"" ""hh"":""nn"":""ss")', '2020-05-15 14:30:20');
    Test('FormatDateTime(now(), "yyyy""-""mm""-""dd"" ""hh"":""mm"":""ss")', '2020-05-15 14:30:20');
    Test('FormatDateTime(now(), "yyyy-mm-dd hh:nn:ss")', '2020-05-15 14:30:20');

    Test('FormatDateTime(now(), "yyyy-mm-dd hh:nn:ss"".""zzz")', '2020-05-15 14:30:20.123');
    Test('FormatDateTime(now(), "yyyy-mm-dd""T""hh:nn:ss.zzz""Z""")', '2020-05-15T14:30:20.123Z');

    Test('FormatDateTime(now(), "yyyy-mm-dd")', '2020-05-15');
    Test('FormatDateTime(now(), "dd""/""mm ""-""yy")', '15/05 -20');
    Test('FormatDateTime(now(), "dd""/""m ""-""yy")', '15/5 -20');
    Test('FormatDateTime(now(), "(yyyy, mm, dd; hh, nn, ss)")', '(2020, 05, 15; 14, 30, 20)');

    Test('FormatDateTime(now(), """klockan"" hh:nn")', 'klockan 14:30');

    Test('FormatDateTime(now(), "mmmm dd")', 'May 15');
    Test('FormatDateTime(now(), "dd mmmm yyyy")', '15 May 2020');

    Test('DateTimeString(now(), "sv-SE")', '2020-05-15 14:30:20');
    Test('DateTimeString(now(), "da-DK")', '15-05-2020 14:30:20');
    Test('DateTimeString(now(), "nb-NO")', '15.05.2020 14:30:20');
    Test('DateTimeString(now(), "en-GB")', '15/05/2020 14:30:20');
    Test('DateTimeString(now(), "en-US")', '5/15/2020 2:30:20 PM');
    Test('DateTimeString(now(), "en-CA")', '15/05/2020 2:30:20 PM');
    Test('DateTimeString(now(), "en-ZA")', '2020/05/15 02:30:20 PM');
    Test('DateTimeString(now(), "de-DE")', '15.05.2020 14:30:20');
    Test('DateTimeString(now(), "el")', '15/5/2020 2:30:20 μμ');
    Test('DateTimeString(now(), "ko")', '2020-05-15 오후 2:30:20');

    Test('DateString(now(), "sv-SE")', '2020-05-15');
    Test('DateString(now(), "da-DK")', '15-05-2020');
    Test('DateString(now(), "nb-NO")', '15.05.2020');
    Test('DateString(now(), "en-GB")', '15/05/2020');
    Test('DateString(now(), "en-US")', '5/15/2020');
    Test('DateString(now(), "en-CA")', '15/05/2020');
    Test('DateString(now(), "en-ZA")', '2020/05/15');
    Test('DateString(now(), "de-DE")', '15.05.2020');
    Test('DateString(now(), "el")', '15/5/2020');
    Test('DateString(now(), "ko")', '2020-05-15');

    Test('TimeString(now(), "sv-SE")', '14:30:20');
    Test('TimeString(now(), "da-DK")', '14:30:20');
    Test('TimeString(now(), "nb-NO")', '14:30:20');
    Test('TimeString(now(), "en-GB")', '14:30:20');
    Test('TimeString(now(), "en-US")', '2:30:20 PM');
    Test('TimeString(now(), "en-CA")', '2:30:20 PM');
    Test('TimeString(now(), "en-ZA")', '02:30:20 PM');
    Test('TimeString(now(), "de-DE")', '14:30:20');
    Test('TimeString(now(), "el")', '2:30:20 μμ');
    Test('TimeString(now(), "ko")', '오후 2:30:20');

  finally
    FixedTime := 0.0;
  end;





  //
  //
  //  CHAPTER 9
  //  Container functions.
  //
  //

  Chapter('Container functions');

  //
  // first, rest, and last
  //

  Test('first("Expecto Patronum")', 'E');
  Test('rest("Expecto Patronum")', 'xpecto Patronum');
  Test('last("Expecto Patronum")', 'm');

  Test('first("")', failure, 'Index 1 out of bounds.');
  Test('rest("")', '');
  Test('last("")', failure, 'Index -1 out of bounds.');

  Test('s ≔ "Expecto Patronum"; EmptyStr ≔ "";', null);

  Test('first(s)', 'E');
  Test('rest(s)', 'xpecto Patronum');
  Test('last(s)', 'm');

  Test('first(EmptyStr)', failure, 'Index 1 out of bounds.');
  Test('rest(EmptyStr)', '');
  Test('last(EmptyStr)', failure, 'Index -1 out of bounds.');

  Test('first(s) ≔ "I"', 'I');
  Test('s', 'Ixpecto Patronum');

  Test('last(s) ≔ "s"', 's');
  Test('s', 'Ixpecto Patronus');

  Test('rest(s) ≔ "test"', failure, 'Left side cannot be assigned to.');
  Test('s', 'Ixpecto Patronus');

  Test('first(EmptyStr) ≔ "A"', failure, 'Index 1 out of bounds.');
  Test('last(EmptyStr) ≔ "A"', failure, 'Index -1 out of bounds.');

  Test('protect(s)', success);
  Test('first(s) ≔ "E"', failure, 'Cannot modify or delete protected variable "s".');
  Test('s', 'Ixpecto Patronus');
  Test('unprotect(s)', success);
  Test('first(s) ≔ "E"', 'E');
  Test('s', 'Expecto Patronus');

  Test('protect(s)', success);
  Test('last(s) ≔ "m"', failure, 'Cannot modify or delete protected variable "s".');
  Test('s', 'Expecto Patronus');
  Test('unprotect(s)', success);
  Test('last(s) ≔ "m"', 'm');
  Test('s', 'Expecto Patronum');

  Test('first("test") ≔ "T"', failure, 'Left side cannot be assigned to.');
  Test('last("test") ≔ "T"', failure, 'Left side cannot be assigned to.');

  Test('first(❨7, 3, 4❩)', 7.0);
  Test('rest(❨7, 3, 4❩)', [3, 4]);
  Test('last(❨7, 3, 4❩)', 4.0);

  Test('first(DebugObject("empty vector"))', failure, 'Index 1 out of bounds.');
  Test('rest(DebugObject("empty vector"))', TAlgosimRealVector.Create);
  Test('last(DebugObject("empty vector"))', failure, 'Index -1 out of bounds.');

  Test('v ≔ ❨5, 0, 2, 8❩', [5, 0, 2, 8]);
  Test('first(v)', 5.0);
  Test('rest(v)', [0, 2, 8]);
  Test('last(v)', 8.0);

  Test('first(v) ≔ 50', 50);
  Test('v', [50, 0, 2, 8]);
  Test('last(v) ≔ 80', 80);
  Test('v', [50, 0, 2, 80]);

  Test('protect(v)', success);
  Test('first(v) ≔ 500', failure, 'Cannot modify or delete protected variable "v".');
  Test('v', [50, 0, 2, 80]);
  Test('unprotect(v)', success);
  Test('first(v) ≔ 500', 500);
  Test('v', [500, 0, 2, 80]);

  Test('protect(v)', success);
  Test('last(v) ≔ 800', failure, 'Cannot modify or delete protected variable "v".');
  Test('v', [500, 0, 2, 80]);
  Test('unprotect(v)', success);
  Test('last(v) ≔ 800', 800);
  Test('v', [500, 0, 2, 800]);

  Test('first(❨7⋅i, 3, 4❩)', 7.0 * ImaginaryUnit);
  Test('rest(❨7⋅i, 3, 4❩)', [TASC(3), 4]);
  Test('last(❨7⋅i, 3, 4❩)', TASC(4.0));

  Test('first(ComplexVector(DebugObject("empty vector")))', failure, 'Index 1 out of bounds.');
  Test('rest(ComplexVector(DebugObject("empty vector")))', TAlgosimComplexVector.Create);
  Test('last(ComplexVector(DebugObject("empty vector")))', failure, 'Index -1 out of bounds.');

  Test('v ≔ ❨5, 0, 2, 8⋅i❩', [5, 0, 2, 8*ImaginaryUnit]);
  Test('first(v)', TASC(5.0));
  Test('rest(v)', [0, 2, 8*ImaginaryUnit]);
  Test('last(v)', 8.0*ImaginaryUnit);

  Test('first(v) ≔ 50', 50);
  Test('v', [50, 0, 2, 8*ImaginaryUnit]);
  Test('last(v) ≔ 80⋅i', 80*ImaginaryUnit);
  Test('v', [50, 0, 2, 80*ImaginaryUnit]);

  Test('protect(v)', success);
  Test('first(v) ≔ 500', failure, 'Cannot modify or delete protected variable "v".');
  Test('v', [50, 0, 2, 80*ImaginaryUnit]);
  Test('unprotect(v)', success);
  Test('first(v) ≔ 500', 500);
  Test('v', [500, 0, 2, 80*ImaginaryUnit]);

  Test('protect(v)', success);
  Test('last(v) ≔ 800⋅i', failure, 'Cannot modify or delete protected variable "v".');
  Test('v', [500, 0, 2, 80*ImaginaryUnit]);
  Test('unprotect(v)', success);
  Test('last(v) ≔ 800⋅i', 800*ImaginaryUnit);
  Test('v', [500, 0, 2, 800*ImaginaryUnit]);

  Test('first(❨1, 2, 3❩) ≔ 10', failure, 'Left side cannot be assigned to.');
  Test('first(❨i, 2, 3❩) ≔ 10', failure, 'Left side cannot be assigned to.');
  Test('last(❨1, 2, 3❩) ≔ 10', failure, 'Left side cannot be assigned to.');
  Test('last(❨i, 2, 3❩) ≔ 10', failure, 'Left side cannot be assigned to.');

  Test('first(❨❨1, 2❩, ❨3, 4❩❩)', 1.0);
  Test('rest(❨❨1, 2❩, ❨3, 4❩❩)', failure, 'Object type "real matrix" is not a suitable container.');
  Test('last(❨❨1, 2❩, ❨3, 4❩❩)', 4.0);
  Test('first(❨❨3, 5, 2❩, ❨7, 1, 0❩❩)', 3.0);
  Test('rest(❨❨3, 5, 2❩, ❨7, 1, 0❩❩)', failure, 'Object type "real matrix" is not a suitable container.');
  Test('last(❨❨3, 5, 2❩, ❨7, 1, 0❩❩)', 0.0);

  Test('first(DebugObject("empty matrix"))', failure, 'Index 1 out of bounds.');
  Test('last(DebugObject("empty matrix"))', failure, 'Index -1 out of bounds.');

  Test('M ≔ ❨❨4, 2, 1❩, ❨0, 5, 9❩❩', 3, [4, 2, 1, 0, 5, 9]);
  Test('first(M)', 4.0);
  Test('last(M)', 9.0);

  Test('first(M) ≔ 40', 40);
  Test('M', 3, [40, 2, 1, 0, 5, 9]);
  Test('last(M) ≔ 90', 90);
  Test('M', 3, [40, 2, 1, 0, 5, 90]);

  Test('protect(M)', success);
  Test('first(M) ≔ 400', failure, 'Cannot modify or delete protected variable "M".');
  Test('M', 3, [40, 2, 1, 0, 5, 90]);
  Test('unprotect(M)', success);
  Test('first(M) ≔ 400', 400);
  Test('M', 3, [400, 2, 1, 0, 5, 90]);

  Test('protect(M)', success);
  Test('last(M) ≔ 900', failure, 'Cannot modify or delete protected variable "M".');
  Test('M', 3, [400, 2, 1, 0, 5, 90]);
  Test('unprotect(M)', success);
  Test('last(M) ≔ 900', 900);
  Test('M', 3, [400, 2, 1, 0, 5, 900]);

  Test('first(❨❨5, 3❩, ❨7, 8❩, ❨1, 2❩❩) ≔ π', failure, 'Left side cannot be assigned to.');
  Test('last(❨❨5, 3❩, ❨7, 8❩, ❨1, 2❩❩) ≔ π^2', failure, 'Left side cannot be assigned to.');

  Test('first(❨❨i, 2❩, ❨3, 4❩❩)', ImaginaryUnit);
  Test('rest(❨❨i, 2❩, ❨3, 4❩❩)', failure, 'Object type "complex matrix" is not a suitable container.');
  Test('last(❨❨i, 2❩, ❨3, 4❩❩)', TASC(4.0));
  Test('first(❨❨3, 5, 2❩, ❨7, i, 0❩❩)', TASC(3.0));
  Test('rest(❨❨3, 5, 2❩, ❨7, i, 0❩❩)', failure, 'Object type "complex matrix" is not a suitable container.');
  Test('last(❨❨3, 5, 2❩, ❨7, i, 0❩❩)', TASC(0.0));

  Test('first(ComplexMatrix(DebugObject("empty matrix")))', failure, 'Index 1 out of bounds.');
  Test('last(ComplexMatrix(DebugObject("empty matrix")))', failure, 'Index -1 out of bounds.');

  Test('M ≔ ❨❨4, 2, i❩, ❨0, 5, 9❩❩', 3, [4, 2, ImaginaryUnit, 0, 5, 9]);
  Test('first(M)', TASC(4.0));
  Test('last(M)', TASC(9.0));

  Test('first(M) ≔ 40', 40);
  Test('M', 3, [40, 2, ImaginaryUnit, 0, 5, 9]);
  Test('last(M) ≔ 90', 90);
  Test('M', 3, [40, 2, ImaginaryUnit, 0, 5, 90]);

  Test('protect(M)', success);
  Test('first(M) ≔ 400', failure, 'Cannot modify or delete protected variable "M".');
  Test('M', 3, [40, 2, ImaginaryUnit, 0, 5, 90]);
  Test('unprotect(M)', success);
  Test('first(M) ≔ 400', 400);
  Test('M', 3, [400, 2, ImaginaryUnit, 0, 5, 90]);

  Test('protect(M)', success);
  Test('last(M) ≔ 900', failure, 'Cannot modify or delete protected variable "M".');
  Test('M', 3, [400, 2, ImaginaryUnit, 0, 5, 90]);
  Test('unprotect(M)', success);
  Test('last(M) ≔ 900', 900);
  Test('M', 3, [400, 2, ImaginaryUnit, 0, 5, 900]);

  Test('first(❨❨5, 3❩, ❨7, 8❩, ❨i, 2❩❩) ≔ π', failure, 'Left side cannot be assigned to.');
  Test('last(❨❨5, 3❩, ❨7, 8❩, ❨i, 2❩❩) ≔ π^2', failure, 'Left side cannot be assigned to.');

  Test('first(''(5, true, "cat", "dog"))', 5);
  Test('rest(''(5, true, "cat", "dog"))', asoarr([True, 'cat', 'dog']));
  Test('last(''(5, true, "cat", "dog"))', 'dog');

  Test('first(''())', failure, 'Index 1 out of bounds.');
  Test('last(''())', failure, 'Index -1 out of bounds.');

  Test('L ≔ ''(true, 3.1415, "rat", 2.5)', asoarr([True, 3.1415, 'rat', 2.5]));

  Test('first(L)', True);
  Test('last(L)', 2.5);

  Test('first(L) ≔ "thestral"', 'thestral');
  Test('L', asoarr(['thestral', 3.1415, 'rat', 2.5]));
  Test('last(L) ≔ 99', 99);
  Test('L', asoarr(['thestral', 3.1415, 'rat', 99]));

  Test('protect(L)', success);
  Test('first(L) ≔ "hippogriff"', failure, 'Cannot modify or delete protected variable "L".');
  Test('L', asoarr(['thestral', 3.1415, 'rat', 99]));
  Test('unprotect(L)', success);
  Test('first(L) ≔ "hippogriff"', 'hippogriff');
  Test('L', asoarr(['hippogriff', 3.1415, 'rat', 99]));

  Test('protect(L)', success);
  Test('last(L) ≔ 99.9', failure, 'Cannot modify or delete protected variable "L".');
  Test('L', asoarr(['hippogriff', 3.1415, 'rat', 99]));
  Test('unprotect(L)', success);
  Test('last(L) ≔ 99.9', 99.9);
  Test('L', asoarr(['hippogriff', 3.1415, 'rat', 99.9]));

  Test('S ≔ ToSet(L);', null);
  Test('first(S) ∈ S', True);
  Test('rest(S)', failure, 'Object type "set" is not a suitable container.');
  Test('last(S) ∈ S', True);

  Test('first(''(true, 3.1415, "rat", 2.5)) ≔ false', failure, 'Left side cannot be assigned to.');
  Test('last(''(true, 3.1415, "rat", 2.5)) ≔ false', failure, 'Left side cannot be assigned to.');

  Test('L ≔ ''();', null);
  Test('first(L) ≔ 5', failure, 'Index 1 out of bounds.');
  Test('last(L) ≔ 5', failure, 'Index -1 out of bounds.');

  Test('first("hippogriff", 0)', '');
  Test('first("hippogriff", 1)', 'h');
  Test('first("hippogriff", 2)', 'hi');
  Test('first("hippogriff", 3)', 'hip');
  Test('first("hippogriff", 4)', 'hipp');
  Test('first("hippogriff", 5)', 'hippo');
  Test('first("hippogriff", 6)', 'hippog');
  Test('first("hippogriff", 7)', 'hippogr');
  Test('first("hippogriff", 8)', 'hippogri');
  Test('first("hippogriff", 9)', 'hippogrif');
  Test('first("hippogriff", 10)', 'hippogriff');
  Test('first("hippogriff", 11)', 'hippogriff');
  Test('first("hippogriff", 12)', 'hippogriff');
  Test('first("hippogriff", 13)', 'hippogriff');
  Test('first("hippogriff", 14)', 'hippogriff');
  Test('first("hippogriff", 15)', 'hippogriff');
  Test('first("hippogriff", 1000000)', 'hippogriff');

  Test('first(❨10, 20, 30❩, 0)', TAlgosimRealVector.Create);
  Test('first(❨10, 20, 30❩, 1)', [10]);
  Test('first(❨10, 20, 30❩, 2)', [10, 20]);
  Test('first(❨10, 20, 30❩, 3)', [10, 20, 30]);
  Test('first(❨10, 20, 30❩, 4)', [10, 20, 30]);
  Test('first(❨10, 20, 30❩, 5)', [10, 20, 30]);
  Test('first(❨10, 20, 30❩, 1000000)', [10, 20, 30]);

  Test('first(❨i, 20, 30❩, 0)', TAlgosimComplexVector.Create);
  Test('first(❨i, 20, 30❩, 1)', [ImaginaryUnit]);
  Test('first(❨i, 20, 30❩, 2)', [ImaginaryUnit, 20]);
  Test('first(❨i, 20, 30❩, 3)', [ImaginaryUnit, 20, 30]);
  Test('first(❨i, 20, 30❩, 4)', [ImaginaryUnit, 20, 30]);
  Test('first(❨i, 20, 30❩, 5)', [ImaginaryUnit, 20, 30]);
  Test('first(❨i, 20, 30❩, 1000000)', [ImaginaryUnit, 20, 30]);

  Test('first(''(10, "dog", 20, true, 30), 0)', asoarr([]));
  Test('first(''(10, "dog", 20, true, 30), 1)', asoarr([10]));
  Test('first(''(10, "dog", 20, true, 30), 2)', asoarr([10, 'dog']));
  Test('first(''(10, "dog", 20, true, 30), 3)', asoarr([10, 'dog', 20]));
  Test('first(''(10, "dog", 20, true, 30), 4)', asoarr([10, 'dog', 20, True]));
  Test('first(''(10, "dog", 20, true, 30), 5)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(''(10, "dog", 20, true, 30), 6)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(''(10, "dog", 20, true, 30), 7)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(''(10, "dog", 20, true, 30), 8)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(''(10, "dog", 20, true, 30), 9)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(''(10, "dog", 20, true, 30), 10000)', asoarr([10, 'dog', 20, True, 30]));

  Test('x ≔ "hippogriff"', 'hippogriff');
  Test('first(x, 0)', '');
  Test('first(x, 1)', 'h');
  Test('first(x, 2)', 'hi');
  Test('first(x, 3)', 'hip');
  Test('first(x, 4)', 'hipp');
  Test('first(x, 5)', 'hippo');
  Test('first(x, 6)', 'hippog');
  Test('first(x, 7)', 'hippogr');
  Test('first(x, 8)', 'hippogri');
  Test('first(x, 9)', 'hippogrif');
  Test('first(x, 10)', 'hippogriff');
  Test('first(x, 11)', 'hippogriff');
  Test('first(x, 12)', 'hippogriff');
  Test('first(x, 13)', 'hippogriff');
  Test('first(x, 14)', 'hippogriff');
  Test('first(x, 15)', 'hippogriff');
  Test('first(x, 1000000)', 'hippogriff');

  Test('x ≔ ❨10, 20, 30❩', [10, 20, 30]);
  Test('first(x, 0)', TAlgosimRealVector.Create);
  Test('first(x, 1)', [10]);
  Test('first(x, 2)', [10, 20]);
  Test('first(x, 3)', [10, 20, 30]);
  Test('first(x, 4)', [10, 20, 30]);
  Test('first(x, 5)', [10, 20, 30]);
  Test('first(x, 1000000)', [10, 20, 30]);

  Test('x ≔ ❨i, 20, 30❩', [ImaginaryUnit, 20, 30]);
  Test('first(x, 0)', TAlgosimComplexVector.Create);
  Test('first(x, 1)', [ImaginaryUnit]);
  Test('first(x, 2)', [ImaginaryUnit, 20]);
  Test('first(x, 3)', [ImaginaryUnit, 20, 30]);
  Test('first(x, 4)', [ImaginaryUnit, 20, 30]);
  Test('first(x, 5)', [ImaginaryUnit, 20, 30]);
  Test('first(x, 1000000)', [ImaginaryUnit, 20, 30]);

  Test('x ≔ ''(10, "dog", 20, true, 30)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(x, 0)', asoarr([]));
  Test('first(x, 1)', asoarr([10]));
  Test('first(x, 2)', asoarr([10, 'dog']));
  Test('first(x, 3)', asoarr([10, 'dog', 20]));
  Test('first(x, 4)', asoarr([10, 'dog', 20, True]));
  Test('first(x, 5)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(x, 6)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(x, 7)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(x, 8)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(x, 9)', asoarr([10, 'dog', 20, True, 30]));
  Test('first(x, 10000)', asoarr([10, 'dog', 20, True, 30]));

  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 0)', asoarr([]));
  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 1)', asoarr([1.0]));
  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 2)', asoarr([1.0, 2.0]));
  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 3)', asoarr([1.0, 2.0, 3.0]));
  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 4)', asoarr([1.0, 2.0, 3.0, 4.0]));
  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 5)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0]));
  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 6)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]));
  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 7)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]));
  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 8)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]));
  Test('first(❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩, 100000)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]));

  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 0)', asoarrex([]));
  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 1)', asoarrex([ASO(ImaginaryUnit)]));
  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 2)', asoarrex([ASO(ImaginaryUnit), ASO(2.0)]));
  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 3)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0)]));
  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 4)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0)]));
  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 5)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0)]));
  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 6)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0), ASO(6.0)]));
  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 7)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0), ASO(6.0)]));
  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 8)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0), ASO(6.0)]));
  Test('first(❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩, 100000)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0), ASO(6.0)]));

  Test('x ≔ ❨❨1, 2❩, ❨3, 4❩, ❨5, 6❩❩', 2, [1, 2, 3, 4, 5, 6]);
  Test('first(x, 0)', asoarr([]));
  Test('first(x, 1)', asoarr([1.0]));
  Test('first(x, 2)', asoarr([1.0, 2.0]));
  Test('first(x, 3)', asoarr([1.0, 2.0, 3.0]));
  Test('first(x, 4)', asoarr([1.0, 2.0, 3.0, 4.0]));
  Test('first(x, 5)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0]));
  Test('first(x, 6)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]));
  Test('first(x, 7)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]));
  Test('first(x, 8)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]));
  Test('first(x, 100000)', asoarr([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]));

  Test('x ≔ ❨❨i, 2❩, ❨3, 4❩, ❨5, 6❩❩', 2, [ImaginaryUnit, 2, 3, 4, 5, 6]);
  Test('first(x, 0)', asoarrex([]));
  Test('first(x, 1)', asoarrex([ASO(ImaginaryUnit)]));
  Test('first(x, 2)', asoarrex([ASO(ImaginaryUnit), ASO(2.0)]));
  Test('first(x, 3)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0)]));
  Test('first(x, 4)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0)]));
  Test('first(x, 5)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0)]));
  Test('first(x, 6)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0), ASO(6.0)]));
  Test('first(x, 7)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0), ASO(6.0)]));
  Test('first(x, 8)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0), ASO(6.0)]));
  Test('first(x, 100000)', asoarrex([ASO(ImaginaryUnit), ASO(2.0), ASO(3.0), ASO(4.0), ASO(5.0), ASO(6.0)]));

  Test('first(ColorGradient(1, 2, "white", "black"))', ASOColor(clWhite));
  Test('first(ColorGradient(1, 2, "white", "black"), 0)', asoarrex([]));
  Test('first(ColorGradient(1, 2, "white", "black"), 1)', asoarrex([ASOColor(clWhite)]));
  Test('first(ColorGradient(1, 2, "white", "black"), 2)', asoarrex([ASOColor(clWhite), ASOColor(clBlack)]));
  Test('first(ColorGradient(1, 2, "white", "black"), 10000)', asoarrex([ASOColor(clWhite), ASOColor(clBlack)]));

  Test('pm ≔ ColorGradient(1, 2, "white", "black");', null);
  Test('first(pm)', ASOColor(clWhite));
  Test('first(pm, 0)', asoarrex([]));
  Test('first(pm, 1)', asoarrex([ASOColor(clWhite)]));
  Test('first(pm, 2)', asoarrex([ASOColor(clWhite), ASOColor(clBlack)]));
  Test('first(pm, 10000)', asoarrex([ASOColor(clWhite), ASOColor(clBlack)]));

  Test('last("hippogriff")', 'f');
  Test('last("hippogriff", 0)', '');
  Test('last("hippogriff", 1)', 'f');
  Test('last("hippogriff", 2)', 'ff');
  Test('last("hippogriff", 3)', 'iff');
  Test('last("hippogriff", 4)', 'riff');
  Test('last("hippogriff", 5)', 'griff');
  Test('last("hippogriff", 6)', 'ogriff');
  Test('last("hippogriff", 7)', 'pogriff');
  Test('last("hippogriff", 8)', 'ppogriff');
  Test('last("hippogriff", 9)', 'ippogriff');
  Test('last("hippogriff", 10)', 'hippogriff');
  Test('last("hippogriff", 11)', 'hippogriff');
  Test('last("hippogriff", 12)', 'hippogriff');
  Test('last("hippogriff", 10000000)', 'hippogriff');

  Test('x ≔ "hippogriff"', 'hippogriff');
  Test('last(x)', 'f');
  Test('last(x, 0)', '');
  Test('last(x, 1)', 'f');
  Test('last(x, 2)', 'ff');
  Test('last(x, 3)', 'iff');
  Test('last(x, 4)', 'riff');
  Test('last(x, 5)', 'griff');
  Test('last(x, 6)', 'ogriff');
  Test('last(x, 7)', 'pogriff');
  Test('last(x, 8)', 'ppogriff');
  Test('last(x, 9)', 'ippogriff');
  Test('last(x, 10)', 'hippogriff');
  Test('last(x, 11)', 'hippogriff');
  Test('last(x, 12)', 'hippogriff');
  Test('last(x, 10000000)', 'hippogriff');

  Test('last(❨5, 2, 1❩)', 1.0);
  Test('last(❨5, 2, 1❩, 0)', TAlgosimRealVector.Create);
  Test('last(❨5, 2, 1❩, 1)', [1]);
  Test('last(❨5, 2, 1❩, 2)', [2, 1]);
  Test('last(❨5, 2, 1❩, 3)', [5, 2, 1]);
  Test('last(❨5, 2, 1❩, 4)', [5, 2, 1]);
  Test('last(❨5, 2, 1❩, 5)', [5, 2, 1]);
  Test('last(❨5, 2, 1❩, 100000000)', [5, 2, 1]);

  Test('x ≔ ❨5, 2, 1❩', [5, 2, 1]);
  Test('last(x)', 1.0);
  Test('last(x, 0)', TAlgosimRealVector.Create);
  Test('last(x, 1)', [1]);
  Test('last(x, 2)', [2, 1]);
  Test('last(x, 3)', [5, 2, 1]);
  Test('last(x, 4)', [5, 2, 1]);
  Test('last(x, 5)', [5, 2, 1]);
  Test('last(x, 100000000)', [5, 2, 1]);

  Test('last(❨5, 2, i❩)', ImaginaryUnit);
  Test('last(❨5, 2, i❩, 0)', TAlgosimComplexVector.Create);
  Test('last(❨5, 2, i❩, 1)', [ImaginaryUnit]);
  Test('last(❨5, 2, i❩, 2)', [2, ImaginaryUnit]);
  Test('last(❨5, 2, i❩, 3)', [5, 2, ImaginaryUnit]);
  Test('last(❨5, 2, i❩, 4)', [5, 2, ImaginaryUnit]);
  Test('last(❨5, 2, i❩, 5)', [5, 2, ImaginaryUnit]);
  Test('last(❨5, 2, i❩, 100000000)', [5, 2, ImaginaryUnit]);

  Test('x ≔ ❨5, 2, i❩', [5, 2, ImaginaryUnit]);
  Test('last(x)', ImaginaryUnit);
  Test('last(x, 0)', TAlgosimComplexVector.Create);
  Test('last(x, 1)', [ImaginaryUnit]);
  Test('last(x, 2)', [2, ImaginaryUnit]);
  Test('last(x, 3)', [5, 2, ImaginaryUnit]);
  Test('last(x, 4)', [5, 2, ImaginaryUnit]);
  Test('last(x, 5)', [5, 2, ImaginaryUnit]);
  Test('last(x, 100000000)', [5, 2, ImaginaryUnit]);

  Test('last(''("dog", 4, true))', True);
  Test('last(''("dog", 4, true), 0)', asoarr([]));
  Test('last(''("dog", 4, true), 1)', asoarr([True]));
  Test('last(''("dog", 4, true), 2)', asoarr([4.0, True]));
  Test('last(''("dog", 4, true), 3)', asoarr(['dog', 4.0, True]));
  Test('last(''("dog", 4, true), 4)', asoarr(['dog', 4.0, True]));
  Test('last(''("dog", 4, true), 5)', asoarr(['dog', 4.0, True]));
  Test('last(''("dog", 4, true), 10000000)', asoarr(['dog', 4.0, True]));

  Test('x ≔ ''("dog", 4, true)', asoarr(['dog', 4.0, True]));
  Test('last(x)', True);
  Test('last(x, 0)', asoarr([]));
  Test('last(x, 1)', asoarr([True]));
  Test('last(x, 2)', asoarr([4.0, True]));
  Test('last(x, 3)', asoarr(['dog', 4.0, True]));
  Test('last(x, 4)', asoarr(['dog', 4.0, True]));
  Test('last(x, 5)', asoarr(['dog', 4.0, True]));
  Test('last(x, 10000000)', asoarr(['dog', 4.0, True]));

  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩)', 3.0);
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 0)', asoarr([]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 1)', asoarr([3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 2)', asoarr([10, 3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 3)', asoarr([2, 10, 3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 4)', asoarr([5, 2, 10, 3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 5)', asoarr([1, 5, 2, 10, 3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 6)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 7)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 8)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 9)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 10)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩, 10000000)', asoarr([2, 1, 5, 2, 10, 3]));

  Test('x ≔ ❨❨2, 1❩, ❨5, 2❩, ❨10, 3❩❩', 2, [2, 1, 5, 2, 10, 3]);
  Test('last(x)', 3.0);
  Test('last(x, 0)', asoarr([]));
  Test('last(x, 1)', asoarr([3]));
  Test('last(x, 2)', asoarr([10, 3]));
  Test('last(x, 3)', asoarr([2, 10, 3]));
  Test('last(x, 4)', asoarr([5, 2, 10, 3]));
  Test('last(x, 5)', asoarr([1, 5, 2, 10, 3]));
  Test('last(x, 6)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(x, 7)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(x, 8)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(x, 9)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(x, 10)', asoarr([2, 1, 5, 2, 10, 3]));
  Test('last(x, 10000000)', asoarr([2, 1, 5, 2, 10, 3]));

  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩)', ImaginaryUnit);
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 0)', asoarrex([]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 1)', asoarrex([ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 2)', asoarrex([ASO(10), ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 3)', asoarrex([ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 4)', asoarrex([ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 5)', asoarrex([ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 6)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 7)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 8)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 9)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 10)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩, 10000000)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));

  Test('x ≔ ❨❨2, 1❩, ❨5, 2❩, ❨10, i❩❩', 2, [2, 1, 5, 2, 10, ImaginaryUnit]);
  Test('last(x)', ImaginaryUnit);
  Test('last(x, 0)', asoarrex([]));
  Test('last(x, 1)', asoarrex([ASO(ImaginaryUnit)]));
  Test('last(x, 2)', asoarrex([ASO(10), ASO(ImaginaryUnit)]));
  Test('last(x, 3)', asoarrex([ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(x, 4)', asoarrex([ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(x, 5)', asoarrex([ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(x, 6)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(x, 7)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(x, 8)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(x, 9)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(x, 10)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));
  Test('last(x, 10000000)', asoarrex([ASO(2), ASO(1), ASO(5), ASO(2), ASO(10), ASO(ImaginaryUnit)]));

  Test('last(ColorGradient(1, 2, "white", "black"))', ASOColor(clBlack));
  Test('last(ColorGradient(1, 2, "white", "black"), 0)', asoarrex([]));
  Test('last(ColorGradient(1, 2, "white", "black"), 1)', asoarrex([ASOColor(clBlack)]));
  Test('last(ColorGradient(1, 2, "white", "black"), 2)', asoarrex([ASOColor(clWhite), ASOColor(clBlack)]));
  Test('last(ColorGradient(1, 2, "white", "black"), 10000)', asoarrex([ASOColor(clWhite), ASOColor(clBlack)]));

  Test('pm ≔ ColorGradient(1, 2, "white", "black");', null);
  Test('last(pm)', ASOColor(clBlack));
  Test('last(pm, 0)', asoarrex([]));
  Test('last(pm, 1)', asoarrex([ASOColor(clBlack)]));
  Test('last(pm, 2)', asoarrex([ASOColor(clWhite), ASOColor(clBlack)]));
  Test('last(pm, 10000)', asoarrex([ASOColor(clWhite), ASOColor(clBlack)]));


  //
  // part and range
  //

  Test('part("Expecto Patronum", range(1, 7))', 'Expecto');
  Test('part("Expecto Patronum", range(9, 16))', 'Patronum');
  Test('part("Expecto Patronum", range(-8, -1))', 'Patronum');

  Test('part("Expecto Patronum", range(0, 7))', 'Expecto');  // The `part` function is forgiving.
  Test('part("Expecto Patronum", range(9, 20))', 'Patronum'); // The `part` function is forgiving.
  Test('part("Expecto Patronum", range(9, 200000000))', 'Patronum'); // The `part` function is forgiving.

  Test('part("Expecto Patronum", range(1))', 'E');
  Test('part("Expecto Patronum", range(2))', 'x');
  Test('part("Expecto Patronum", range(5))', 'c');
  Test('part("Expecto Patronum", range(16))', 'm');
  Test('part("Expecto Patronum", range(17))', '');

  Test('part("Expecto Patronum", 1)', 'E');
  Test('part("Expecto Patronum", 2)', 'x');
  Test('part("Expecto Patronum", 5)', 'c');
  Test('part("Expecto Patronum", 16)', 'm');
  Test('part("Expecto Patronum", 17)', '');

  Test('part("Expecto Patronum", range(7, 1))', 'otcepxE'); // picking members in the reverse order
  Test('part("Expecto Patronum", range(-1, -8))', 'munortaP'); // picking members in the reverse order
  Test('part("Expecto Patronum", range(-10, -16))', 'otcepxE'); // picking members in the reverse order
  Test('part("Expecto Patronum", range(-10, -10000))', 'otcepxE'); // still forgiving

  Test('part("Expecto Patronum", range(1, 100, 2))', 'EpcoPtou'); // take every other
  Test('part("Expecto Patronum", range(9, 100, 2))', 'Ptou');
  Test('part("Expecto Patronum", range(9, 11, 2))', 'Pt');
  Test('part("Expecto Patronum", range(9, 10, 2))', 'P');
  Test('part("Expecto Patronum", range(9, 9, 2))', 'P');
  Test('part("Expecto Patronum", range(-8, -1, 2))', 'Ptou');
  Test('part("Expecto Patronum", range(7, 1, 2))', 'ocpE');
  Test('part("Expecto Patronum", range(7, -100, 2))', 'ocpE');
  Test('part("Expecto Patronum", range(-1, -8, 2))', 'mnra');

  Test('part("Expecto Patronum", ''(1, 5, 6, 8))', 'Ect ');
  Test('part("Expecto Patronum", ''(1, 5, 6, 8, -2))', 'Ect u');
  Test('part("Expecto Patronum", ''(1, 5, 6, 8, -2, -3))', 'Ect un');
  Test('part("Expecto Patronum", ''(1, 5, 6, 8, -2, -3, 1))', 'Ect unE');
  Test('part("Expecto Patronum", ''(1, 5, 6, 8, 50000, -2, -3, -123, 1))', 'Ect unE');
  Test('part("Expecto Patronum", ''(4, 0))', 'e');
  Test('part("Expecto Patronum", ''(4))', 'e');
  Test('part("Expecto Patronum", ''())', '');
  Test('part("Expecto Patronum", ''(1000))', '');

  Test('s ≔ "Expecto Patronum"', 'Expecto Patronum');

  Test('part(s, range(1, 7))', 'Expecto');
  Test('part(s, range(9, 16))', 'Patronum');
  Test('part(s, range(-8, -1))', 'Patronum');

  Test('part(s, range(0, 7))', 'Expecto');  // The `part` function is forgiving.
  Test('part(s, range(9, 20))', 'Patronum'); // The `part` function is forgiving.
  Test('part(s, range(9, 200000000))', 'Patronum'); // The `part` function is forgiving.

  Test('part(s, range(1))', 'E');
  Test('part(s, range(2))', 'x');
  Test('part(s, range(5))', 'c');
  Test('part(s, range(16))', 'm');
  Test('part(s, range(17))', '');

  Test('part(s, 1)', 'E');
  Test('part(s, 2)', 'x');
  Test('part(s, 5)', 'c');
  Test('part(s, 16)', 'm');
  Test('part(s, 17)', '');

  Test('part(s, range(7, 1))', 'otcepxE'); // picking members in the reverse order
  Test('part(s, range(-1, -8))', 'munortaP'); // picking members in the reverse order
  Test('part(s, range(-10, -16))', 'otcepxE'); // picking members in the reverse order
  Test('part(s, range(-10, -10000))', 'otcepxE'); // still forgiving

  Test('part(s, range(1, 100, 2))', 'EpcoPtou'); // take every other
  Test('part(s, range(9, 100, 2))', 'Ptou');
  Test('part(s, range(9, 11, 2))', 'Pt');
  Test('part(s, range(9, 10, 2))', 'P');
  Test('part(s, range(9, 9, 2))', 'P');
  Test('part(s, range(-8, -1, 2))', 'Ptou');
  Test('part(s, range(7, 1, 2))', 'ocpE');
  Test('part(s, range(7, -100, 2))', 'ocpE');
  Test('part(s, range(-1, -8, 2))', 'mnra');

  Test('part(s, ''(1, 5, 6, 8))', 'Ect ');
  Test('part(s, ''(1, 5, 6, 8, -2))', 'Ect u');
  Test('part(s, ''(1, 5, 6, 8, -2, -3))', 'Ect un');
  Test('part(s, ''(1, 5, 6, 8, -2, -3, 1))', 'Ect unE');
  Test('part(s, ''(1, 5, 6, 8, 50000, -2, -3, -123, 1))', 'Ect unE');
  Test('part(s, ''(4, 0))', 'e');
  Test('part(s, ''(4))', 'e');
  Test('part(s, ''())', '');
  Test('part(s, ''(1000))', '');

  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(2, 5))', [1, 4, 1, 5]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(2, 100))', [1, 4, 1, 5, 9, 2]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(1, 3))', [3, 1, 4]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(-2, -1))', [9, 2]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(100, 3))', [2, 9, 5, 1, 4]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(-1, -100))', [2, 9, 5, 1, 4, 1, 3]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(1, 100, 2))', [3, 4, 5, 2]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(1, 100, 3))', [3, 1, 2]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(1, 100, 4))', [3, 5]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(1, 100, 5))', [3, 9]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(1, 100, 6))', [3, 2]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(1, 100, 7))', [3]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, range(-1, -100, 2))', [2, 5, 4, 3]);

  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, ''(2, 5, 6))', [1, 5, 9]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, ''(2, 5, 6, -2))', [1, 5, 9, 9]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, ''(2, 5, 6, -2, -1))', [1, 5, 9, 9, 2]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, ''(2, -5, 6, -2, -1))', [1, 4, 9, 9, 2]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, ''(2, -5, 6, 5000, -2, -1))', [1, 4, 9, 9, 2]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, ''(2, -5, 6, -5000, -2, -1))', [1, 4, 9, 9, 2]);
  Test('part(❨3, 1, 4, 1, 5, 9, 2❩, ''(2))', [1]);
  Test('#part(❨3, 1, 4, 1, 5, 9, 2❩, ''(500))', 0);
  Test('#part(❨3, 1, 4, 1, 5, 9, 2❩, ''())', 0);

  Test('v ≔ ❨3, 1, 4, 1, 5, 9, 2❩', [3, 1, 4, 1, 5, 9, 2]);

  Test('part(v, range(2, 5))', [1, 4, 1, 5]);
  Test('part(v, range(2, 100))', [1, 4, 1, 5, 9, 2]);
  Test('part(v, range(1, 3))', [3, 1, 4]);
  Test('part(v, range(-2, -1))', [9, 2]);
  Test('part(v, range(100, 3))', [2, 9, 5, 1, 4]);
  Test('part(v, range(-1, -100))', [2, 9, 5, 1, 4, 1, 3]);
  Test('part(v, range(1, 100, 2))', [3, 4, 5, 2]);
  Test('part(v, range(1, 100, 3))', [3, 1, 2]);
  Test('part(v, range(1, 100, 4))', [3, 5]);
  Test('part(v, range(1, 100, 5))', [3, 9]);
  Test('part(v, range(1, 100, 6))', [3, 2]);
  Test('part(v, range(1, 100, 7))', [3]);
  Test('part(v, range(-1, -100, 2))', [2, 5, 4, 3]);

  Test('part(v, ''(2, 5, 6))', [1, 5, 9]);
  Test('part(v, ''(2, 5, 6, -2))', [1, 5, 9, 9]);
  Test('part(v, ''(2, 5, 6, -2, -1))', [1, 5, 9, 9, 2]);
  Test('part(v, ''(2, -5, 6, -2, -1))', [1, 4, 9, 9, 2]);
  Test('part(v, ''(2, -5, 6, 5000, -2, -1))', [1, 4, 9, 9, 2]);
  Test('part(v, ''(2, -5, 6, -5000, -2, -1))', [1, 4, 9, 9, 2]);
  Test('part(v, ''(2))', [1]);
  Test('#part(v, ''(500))', 0);
  Test('#part(v, ''())', 0);

  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(2, 5))', [ImaginaryUnit, 4, 1, 5]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(2, 100))', [ImaginaryUnit, 4, 1, 5, 9, 2]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(1, 3))', [3, ImaginaryUnit, 4]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(-2, -1))', [TASC(9), 2]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(100, 3))', [TASC(2), 9, 5, 1, 4]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(-1, -100))', [2, 9, 5, 1, 4, ImaginaryUnit, 3]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(1, 100, 2))', [TASC(3), 4, 5, 2]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(1, 100, 3))', [TASC(3), 1, 2]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(1, 100, 4))', [TASC(3), 5]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(1, 100, 5))', [TASC(3), 9]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(1, 100, 6))', [TASC(3), 2]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(1, 100, 7))', [TASC(3)]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, range(-1, -100, 2))', [2, 5, 4, TASC(3)]);

  Test('part(❨3, i, 4, 1, 5, 9, 2❩, ''(2, 5, 6))', [ImaginaryUnit, 5, 9]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, ''(2, 5, 6, -2))', [ImaginaryUnit, 5, 9, 9]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, ''(2, 5, 6, -2, -1))', [ImaginaryUnit, 5, 9, 9, 2]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, ''(2, -5, 6, -2, -1))', [ImaginaryUnit, 4, 9, 9, 2]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, ''(2, -5, 6, 5000, -2, -1))', [ImaginaryUnit, 4, 9, 9, 2]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, ''(2, -5, 6, -5000, -2, -1))', [ImaginaryUnit, 4, 9, 9, 2]);
  Test('part(❨3, i, 4, 1, 5, 9, 2❩, ''(2))', [ImaginaryUnit]);
  Test('#part(❨3, i, 4, 1, 5, 9, 2❩, ''(500))', 0);
  Test('#part(❨3, i, 4, 1, 5, 9, 2❩, ''())', 0);

  Test('v ≔ ❨3, i, 4, 1, 5, 9, 2❩', [3, ImaginaryUnit, 4, 1, 5, 9, 2]);

  Test('part(v, range(2, 5))', [ImaginaryUnit, 4, 1, 5]);
  Test('part(v, range(2, 100))', [ImaginaryUnit, 4, 1, 5, 9, 2]);
  Test('part(v, range(1, 3))', [3, ImaginaryUnit, 4]);
  Test('part(v, range(-2, -1))', [TASC(9), 2]);
  Test('part(v, range(100, 3))', [TASC(2), 9, 5, 1, 4]);
  Test('part(v, range(-1, -100))', [2, 9, 5, 1, 4, ImaginaryUnit, 3]);
  Test('part(v, range(1, 100, 2))', [TASC(3), 4, 5, 2]);
  Test('part(v, range(1, 100, 3))', [TASC(3), 1, 2]);
  Test('part(v, range(1, 100, 4))', [TASC(3), 5]);
  Test('part(v, range(1, 100, 5))', [TASC(3), 9]);
  Test('part(v, range(1, 100, 6))', [TASC(3), 2]);
  Test('part(v, range(1, 100, 7))', [TASC(3)]);
  Test('part(v, range(-1, -100, 2))', [2, 5, 4, TASC(3)]);

  Test('part(v, ''(2, 5, 6))', [ImaginaryUnit, 5, 9]);
  Test('part(v, ''(2, 5, 6, -2))', [ImaginaryUnit, 5, 9, 9]);
  Test('part(v, ''(2, 5, 6, -2, -1))', [ImaginaryUnit, 5, 9, 9, 2]);
  Test('part(v, ''(2, -5, 6, -2, -1))', [ImaginaryUnit, 4, 9, 9, 2]);
  Test('part(v, ''(2, -5, 6, 5000, -2, -1))', [ImaginaryUnit, 4, 9, 9, 2]);
  Test('part(v, ''(2, -5, 6, -5000, -2, -1))', [ImaginaryUnit, 4, 9, 9, 2]);
  Test('part(v, ''(2))', [ImaginaryUnit]);
  Test('#part(v, ''(500))', 0);
  Test('#part(v, ''())', 0);

  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 5))', asoarr([1.0, 4.0, 1.0, 5.0]));
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(-4, -1))', asoarr([2.0, 6.0, 5.0, 3.0]));
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(5, 1))', asoarr([5.0, 1.0, 4.0, 1.0, 3.0]));
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(5, 6))', asoarr([5.0, 9.0]));
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 100))', asoarr([1.0, 4.0, 1.0, 5.0, 9.0, 2.0, 6.0, 5.0, 3.0]));
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 100, 2))', asoarr([1.0, 1.0, 9.0, 6.0, 3.0]));

  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(8, 2, -1))', asoarr([6, 1, 3]));
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(8, 2, 8000, -1))', asoarr([6, 1, 3]));
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(8000))', asoarr([]));
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''())', asoarr([]));

  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 4), range(1, 2))', 3, [1, 4, 1, 2, 6, 5]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 5), range(1, 2))', 4, [1, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 50000), range(1, 2))', 4, [1, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 50000), range(1, 1))', 4, [1, 4, 1, 5]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 50000), 1)', 4, [1, 4, 1, 5]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 50000), range(1, 20000))', 4, [1, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(1, 2), range(1, 2))', 2, [3, 1, 9, 2]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(1, 1), range(1, 2))', 1, [3, 9]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, 1, range(1, 2))', 1, [3, 9]);

  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(3, 5), range(1, 2))', 3, [4, 1, 5, 6, 5, 3]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(-3, -1), range(1, 2))', 3, [4, 1, 5, 6, 5, 3]);

  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(5, 3), range(1, 2))', 3, [5, 1, 4, 3, 5, 6]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(5, 3), range(2, 1))', 3, [3, 5, 6, 5, 1, 4]);

  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(3, 5), range(-1, -2))', 3, [6, 5, 3, 4, 1, 5]);

  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(3, 5, 2), range(1, 2))', 2, [4, 5, 6, 3]);

  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(1, 100, 2), range(1, 2))', 3, [3, 4, 5, 9, 6, 3]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(-1, -100, 2), range(1, 2))', 3, [5, 4, 3, 3, 6, 9]);

  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(1, 2, -2), ''(1))', 3, [3, 1, 1]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(1, 2, -2), ''(2, 1))', 3, [9, 2, 5, 3, 1, 1]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(500, 1, 2, -2), ''(2, -500, -1))', 3, [9, 2, 5, 9, 2, 5]);
  Test('part(❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(1, 2, -2), ''())', failure, 'A matrix must have size at least 1×1.');

  Test('M ≔ ❨❨3, 1, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩', 5, [3, 1, 4, 1, 5, 9, 2, 6, 5, 3]);

  Test('part(M, range(2, 5))', asoarr([1.0, 4.0, 1.0, 5.0]));
  Test('part(M, range(-4, -1))', asoarr([2.0, 6.0, 5.0, 3.0]));
  Test('part(M, range(5, 1))', asoarr([5.0, 1.0, 4.0, 1.0, 3.0]));
  Test('part(M, range(5, 6))', asoarr([5.0, 9.0]));
  Test('part(M, range(2, 100))', asoarr([1.0, 4.0, 1.0, 5.0, 9.0, 2.0, 6.0, 5.0, 3.0]));
  Test('part(M, range(2, 100, 2))', asoarr([1.0, 1.0, 9.0, 6.0, 3.0]));

  Test('part(M, ''(8, 2, -1))', asoarr([6, 1, 3]));
  Test('part(M, ''(8, 2, 8000, -1))', asoarr([6, 1, 3]));
  Test('part(M, ''(8000))', asoarr([]));
  Test('part(M, ''())', asoarr([]));

  Test('part(M, range(2, 4), range(1, 2))', 3, [1, 4, 1, 2, 6, 5]);
  Test('part(M, range(2, 5), range(1, 2))', 4, [1, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(M, range(2, 50000), range(1, 2))', 4, [1, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(M, range(2, 50000), range(1, 1))', 4, [1, 4, 1, 5]);
  Test('part(M, range(2, 50000), 1)', 4, [1, 4, 1, 5]);
  Test('part(M, range(2, 50000), range(1, 20000))', 4, [1, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(M, range(1, 2), range(1, 2))', 2, [3, 1, 9, 2]);
  Test('part(M, range(1, 1), range(1, 2))', 1, [3, 9]);
  Test('part(M, 1, range(1, 2))', 1, [3, 9]);

  Test('part(M, range(3, 5), range(1, 2))', 3, [4, 1, 5, 6, 5, 3]);
  Test('part(M, range(-3, -1), range(1, 2))', 3, [4, 1, 5, 6, 5, 3]);

  Test('part(M, range(5, 3), range(1, 2))', 3, [5, 1, 4, 3, 5, 6]);
  Test('part(M, range(5, 3), range(2, 1))', 3, [3, 5, 6, 5, 1, 4]);

  Test('part(M, range(3, 5), range(-1, -2))', 3, [6, 5, 3, 4, 1, 5]);

  Test('part(M, range(3, 5, 2), range(1, 2))', 2, [4, 5, 6, 3]);

  Test('part(M, range(1, 100, 2), range(1, 2))', 3, [3, 4, 5, 9, 6, 3]);
  Test('part(M, range(-1, -100, 2), range(1, 2))', 3, [5, 4, 3, 3, 6, 9]);

  Test('part(M, ''(1, 2, -2), ''(1))', 3, [3, 1, 1]);
  Test('part(M, ''(1, 2, -2), ''(2, 1))', 3, [9, 2, 5, 3, 1, 1]);
  Test('part(M, ''(500, 1, 2, -2), ''(2, -500, -1))', 3, [9, 2, 5, 9, 2, 5]);
  Test('part(M, ''(1, 2, -2), ''())', failure, 'A matrix must have size at least 1×1.');

  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 5))', asoarrex([ASO(ImaginaryUnit), ASO(TASC(4.0)), ASO(TASC(1.0)), ASO(TASC(5.0))]));
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(-4, -1))', asoarrex([ASO(TASC(2.0)), ASO(TASC(6.0)), ASO(TASC(5.0)), ASO(TASC(3.0))]));
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(5, 1))', asoarrex([ASO(TASC(5.0)), ASO(TASC(1.0)), ASO(TASC(4.0)), ASO(ImaginaryUnit), ASO(TASC(3.0))]));
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(5, 6))', asoarrex([ASO(TASC(5.0)), ASO(TASC(9.0))]));
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 100))', asoarrex([ASO(ImaginaryUnit), ASO(TASC(4.0)), ASO(TASC(1.0)), ASO(TASC(5.0)), ASO(TASC(9.0)), ASO(TASC(2.0)), ASO(TASC(6.0)), ASO(TASC(5.0)), ASO(TASC(3.0))]));
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 100, 2))', asoarrex([ASO(ImaginaryUnit), ASO(TASC(1.0)), ASO(TASC(9.0)), ASO(TASC(6.0)), ASO(TASC(3.0))]));

  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(8, 2, -1))', asoarrex([ASO(TASC(6)), ASO(ImaginaryUnit), ASO(TASC(3))]));
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(8, 2, 8000, -1))', asoarrex([ASO(TASC(6)), ASO(ImaginaryUnit), ASO(TASC(3))]));
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(8000))', asoarrex([]));
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''())', asoarrex([]));

  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 4), range(1, 2))', 3, [ImaginaryUnit, 4, 1, 2, 6, 5]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 5), range(1, 2))', 4, [ImaginaryUnit, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 50000), range(1, 2))', 4, [ImaginaryUnit, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 50000), range(1, 1))', 4, [ImaginaryUnit, 4, 1, 5]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 50000), 1)', 4, [ImaginaryUnit, 4, 1, 5]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(2, 50000), range(1, 20000))', 4, [ImaginaryUnit, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(1, 2), range(1, 2))', 2, [3, ImaginaryUnit, 9, 2]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(1, 1), range(1, 2))', 1, [TASC(3), 9]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, 1, range(1, 2))', 1, [TASC(3), 9]);

  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(3, 5), range(1, 2))', 3, [TASC(4), 1, 5, 6, 5, 3]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(-3, -1), range(1, 2))', 3, [TASC(4), 1, 5, 6, 5, 3]);

  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(5, 3), range(1, 2))', 3, [TASC(5), 1, 4, 3, 5, 6]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(5, 3), range(2, 1))', 3, [TASC(3), 5, 6, 5, 1, 4]);

  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(3, 5), range(-1, -2))', 3, [TASC(6), 5, 3, 4, 1, 5]);

  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(3, 5, 2), range(1, 2))', 2, [TASC(4), 5, 6, 3]);

  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(1, 100, 2), range(1, 2))', 3, [TASC(3), 4, 5, 9, 6, 3]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, range(-1, -100, 2), range(1, 2))', 3, [TASC(5), 4, 3, 3, 6, 9]);

  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(1, 2, -2), ''(1))', 3, [3, ImaginaryUnit, 1]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(1, 2, -2), ''(2, 1))', 3, [9, 2, 5, 3, ImaginaryUnit, 1]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(500, 1, 2, -2), ''(2, -500, -1))', 3, [TASC(9), 2, 5, 9, 2, 5]);
  Test('part(❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩, ''(1, 2, -2), ''())', failure, 'A matrix must have size at least 1×1.');

  Test('M ≔ ❨❨3, i, 4, 1, 5❩, ❨9, 2, 6, 5, 3❩❩', 5, [3, ImaginaryUnit, 4, 1, 5, 9, 2, 6, 5, 3]);

  Test('part(M, range(2, 5))', asoarrex([ASO(ImaginaryUnit), ASO(TASC(4.0)), ASO(TASC(1.0)), ASO(TASC(5.0))]));
  Test('part(M, range(-4, -1))', asoarrex([ASO(TASC(2.0)), ASO(TASC(6.0)), ASO(TASC(5.0)), ASO(TASC(3.0))]));
  Test('part(M, range(5, 1))', asoarrex([ASO(TASC(5.0)), ASO(TASC(1.0)), ASO(TASC(4.0)), ASO(ImaginaryUnit), ASO(TASC(3.0))]));
  Test('part(M, range(5, 6))', asoarrex([ASO(TASC(5.0)), ASO(TASC(9.0))]));
  Test('part(M, range(2, 100))', asoarrex([ASO(ImaginaryUnit), ASO(TASC(4.0)), ASO(TASC(1.0)), ASO(TASC(5.0)), ASO(TASC(9.0)), ASO(TASC(2.0)), ASO(TASC(6.0)), ASO(TASC(5.0)), ASO(TASC(3.0))]));
  Test('part(M, range(2, 100, 2))', asoarrex([ASO(ImaginaryUnit), ASO(TASC(1.0)), ASO(TASC(9.0)), ASO(TASC(6.0)), ASO(TASC(3.0))]));

  Test('part(M, ''(8, 2, -1))', asoarrex([ASO(TASC(6)), ASO(ImaginaryUnit), ASO(TASC(3))]));
  Test('part(M, ''(8, 2, 8000, -1))', asoarrex([ASO(TASC(6)), ASO(ImaginaryUnit), ASO(TASC(3))]));
  Test('part(M, ''(8000))', asoarrex([]));
  Test('part(M, ''())', asoarrex([]));

  Test('part(M, range(2, 4), range(1, 2))', 3, [ImaginaryUnit, 4, 1, 2, 6, 5]);
  Test('part(M, range(2, 5), range(1, 2))', 4, [ImaginaryUnit, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(M, range(2, 50000), range(1, 2))', 4, [ImaginaryUnit, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(M, range(2, 50000), range(1, 1))', 4, [ImaginaryUnit, 4, 1, 5]);
  Test('part(M, range(2, 50000), 1)', 4, [ImaginaryUnit, 4, 1, 5]);
  Test('part(M, range(2, 50000), range(1, 20000))', 4, [ImaginaryUnit, 4, 1, 5, 2, 6, 5, 3]);
  Test('part(M, range(1, 2), range(1, 2))', 2, [3, ImaginaryUnit, 9, 2]);
  Test('part(M, range(1, 1), range(1, 2))', 1, [TASC(3), 9]);
  Test('part(M, 1, range(1, 2))', 1, [TASC(3), 9]);

  Test('part(M, range(3, 5), range(1, 2))', 3, [TASC(4), 1, 5, 6, 5, 3]);
  Test('part(M, range(-3, -1), range(1, 2))', 3, [TASC(4), 1, 5, 6, 5, 3]);

  Test('part(M, range(5, 3), range(1, 2))', 3, [TASC(5), 1, 4, 3, 5, 6]);
  Test('part(M, range(5, 3), range(2, 1))', 3, [TASC(3), 5, 6, 5, 1, 4]);

  Test('part(M, range(3, 5), range(-1, -2))', 3, [TASC(6), 5, 3, 4, 1, 5]);

  Test('part(M, range(3, 5, 2), range(1, 2))', 2, [TASC(4), 5, 6, 3]);

  Test('part(M, range(1, 100, 2), range(1, 2))', 3, [TASC(3), 4, 5, 9, 6, 3]);
  Test('part(M, range(-1, -100, 2), range(1, 2))', 3, [TASC(5), 4, 3, 3, 6, 9]);

  Test('part(M, ''(1, 2, -2), ''(1))', 3, [3, ImaginaryUnit, 1]);
  Test('part(M, ''(1, 2, -2), ''(2, 1))', 3, [9, 2, 5, 3, ImaginaryUnit, 1]);
  Test('part(M, ''(500, 1, 2, -2), ''(2, -500, -1))', 3, [TASC(9), 2, 5, 9, 2, 5]);
  Test('part(M, ''(1, 2, -2), ''())', failure, 'A matrix must have size at least 1×1.');

  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), range(2, 4))', asoarr([True, 4.1, 4.2]));
  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), range(2, 6))', asoarr([True, 4.1, 4.2, 4.3, 'astronomy']));
  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), range(2, 1000000))', asoarr([True, 4.1, 4.2, 4.3, 'astronomy']));

  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), range(-3, -2))', asoarr([4.2, 4.3]));
  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), range(-100, -2))', asoarr([6, True, 4.1, 4.2, 4.3]));

  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), range(4, 2))', asoarr([4.2, 4.1, True]));

  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), range(2, 100, 2))', asoarr([True, 4.2, 'astronomy']));

  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), ''(5, -1, 2, 2))', asoarr([4.3, 'astronomy', True, True]));
  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), ''(5, 5000, -1, -123, -123, 2, 2))', asoarr([4.3, 'astronomy', True, True]));
  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), ''(1000))', asoarr([]));
  Test('part(''(6, true, 4.1, 4.2, 4.3, "astronomy"), ''())', asoarr([]));

  Test('L ≔ ''(6, true, 4.1, 4.2, 4.3, "astronomy");', null);

  Test('part(L, range(2, 4))', asoarr([True, 4.1, 4.2]));
  Test('part(L, range(2, 6))', asoarr([True, 4.1, 4.2, 4.3, 'astronomy']));
  Test('part(L, range(2, 1000000))', asoarr([True, 4.1, 4.2, 4.3, 'astronomy']));

  Test('part(L, range(-3, -2))', asoarr([4.2, 4.3]));
  Test('part(L, range(-100, -2))', asoarr([6, True, 4.1, 4.2, 4.3]));

  Test('part(L, range(4, 2))', asoarr([4.2, 4.1, True]));

  Test('part(L, range(2, 100, 2))', asoarr([True, 4.2, 'astronomy']));

  Test('part(L, ''(5, -1, 2, 2))', asoarr([4.3, 'astronomy', True, True]));
  Test('part(L, ''(5, 5000, -1, -123, -123, 2, 2))', asoarr([4.3, 'astronomy', True, True]));
  Test('part(L, ''(1000))', asoarr([]));
  Test('part(L, ''())', asoarr([]));

  Test('S ≔ ToSet(L);', null);
  Test('#part(S, range(2, 4))', 3);

  Test('delete(s); delete(v); delete(M); delete(L); delete(S)', success);


  //
  // Sorting
  //

  // sort

  Test('v ≔ ❨6, 2, 0, −10, 1, −2, 4, 8, −3, 9, 2❩', [6, 2, 0, -10, 1, -2, 4, 8, -3, 9, 2]);

  Test('sort(v)', [-10, -3, -2, 0, 1, 2, 2, 4, 6, 8, 9]);
  Test('sort(v, "standard order")', [-10, -3, -2, 0, 1, 2, 2, 4, 6, 8, 9]);
  Test('sort(v, "standard order descending")', [9, 8, 6, 4, 2, 2, 1, 0, -2, -3, -10]);
  Test('sort(v, "absolute value")', [0, 1, -2, 2, 2, -3, 4, 6, 8, 9, -10]);
  Test('sort(v, "absolute value descending")', [-10, 9, 8, 6, 4, -3, 2, 2, -2, 1, 0]);
  Test('sort(v, "real and imaginary parts")', [-10, -3, -2, 0, 1, 2, 2, 4, 6, 8, 9]);
  Test('sort(v, "real and imaginary parts descending")', [9, 8, 6, 4, 2, 2, 1, 0, -2, -3, -10]);
  Test('sort(v, "modulus")', [0, 1, -2, 2, 2, -3, 4, 6, 8, 9, -10]);
  Test('sort(v, "modulus descending")', [-10, 9, 8, 6, 4, -3, 2, 2, -2, 1, 0]);
  Test('sort(v, "argument")', [0, 1, 2, 2, 4, 6, 8, 9, -2, -3, -10]);
  Test('sort(v, "argument descending")', [-10, -3, -2, 9, 8, 6, 4, 2, 2, 1, 0]);
  Test('sort(v, "modulus argument")', [0, 1, 2, 2, -2, -3, 4, 6, 8, 9, -10]);
  Test('sort(v, "modulus argument descending")', [-10, 9, 8, 6, 4, -3, -2, 2, 2, 1, 0]);

  Test('v ≔ ComplexVector(❨6, 2, 0, −10, 1, −2, 4, 8, −3, 9, 2❩)', [TASC(6), 2, 0, -10, 1, -2, 4, 8, -3, 9, 2]);

  Test('sort(v)', failure, 'No comparison method specified.');
  Test('sort(v, "standard order")', [TASC(-10), -3, -2, 0, 1, 2, 2, 4, 6, 8, 9]);
  Test('sort(v, "standard order descending")', [TASC(9), 8, 6, 4, 2, 2, 1, 0, -2, -3, -10]);
  Test('sort(v, "absolute value")', [TASC(0), 1, -2, 2, 2, -3, 4, 6, 8, 9, -10]);
  Test('sort(v, "absolute value descending")', [TASC(-10), 9, 8, 6, 4, -3, 2, 2, -2, 1, 0]);
  Test('sort(v, "real and imaginary parts")', [TASC(-10), -3, -2, 0, 1, 2, 2, 4, 6, 8, 9]);
  Test('sort(v, "real and imaginary parts descending")', [TASC(9), 8, 6, 4, 2, 2, 1, 0, -2, -3, -10]);
  Test('sort(v, "modulus")', [TASC(0), 1, -2, 2, 2, -3, 4, 6, 8, 9, -10]);
  Test('sort(v, "modulus descending")', [TASC(-10), 9, 8, 6, 4, -3, 2, 2, -2, 1, 0]);
  Test('sort(v, "argument")', [TASC(0), 1, 2, 2, 4, 6, 8, 9, -2, -3, -10]);
  Test('sort(v, "argument descending")', [TASC(-10), -3, -2, 9, 8, 6, 4, 2, 2, 1, 0]);
  Test('sort(v, "modulus argument")', [TASC(0), 1, 2, 2, -2, -3, 4, 6, 8, 9, -10]);
  Test('sort(v, "modulus argument descending")', [TASC(-10), 9, 8, 6, 4, -3, -2, 2, 2, 1, 0]);

  Test('v ≔ ❨3, −5, 5⋅i, −1, 1 + i, −4⋅i, −10, −5⋅i, 12 + 10⋅i, 12 − 10⋅i, 0, 5❩',
    [3, -5, 5*ImaginaryUnit, -1, 1 + ImaginaryUnit, -4*ImaginaryUnit, -10, -5*ImaginaryUnit, 12 + 10*ImaginaryUnit, 12 - 10*ImaginaryUnit, 0, 5]);

  Test('sort(v)', failure, 'No comparison method specified.');
  Test('sort(v, "standard order")', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component 5⋅i.');
  Test('sort(v, "standard order descending")', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component 5⋅i.');
  Test('sort(v, "absolute value")', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component 5⋅i.');
  Test('sort(v, "absolute value descending")', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component 5⋅i.');
  Test('sort(v, "real and imaginary parts")', [-10, -5, -1, -5*ImaginaryUnit, -4*ImaginaryUnit, 0, 5*ImaginaryUnit, 1+ImaginaryUnit, 3, 5, 12-10*ImaginaryUnit, 12+10*ImaginaryUnit]);
  Test('sort(v, "real and imaginary parts descending")', [12+10*ImaginaryUnit, 12-10*ImaginaryUnit, 5, 3, 1+ImaginaryUnit, 5*ImaginaryUnit, 0, -4*ImaginaryUnit, -5*ImaginaryUnit, -1, -5, -10]);
  Test('sort(v, "modulus")', [0, -1, 1 + ImaginaryUnit, 3, -4*ImaginaryUnit, -5, -5*ImaginaryUnit, 5*ImaginaryUnit, 5, -10, 12 - 10*ImaginaryUnit, 12 + 10*ImaginaryUnit]);
  Test('sort(v, "modulus descending")', [12 + 10*ImaginaryUnit, 12 - 10*ImaginaryUnit, -10, 5, 5*ImaginaryUnit, -5*ImaginaryUnit, -5, -4*ImaginaryUnit, 3, 1 + ImaginaryUnit, -1, 0]);
  Test('sort(v, "argument")', [-4*ImaginaryUnit, -5*ImaginaryUnit, 12 - 10*ImaginaryUnit, 0, 3, 5, 12 + 10*ImaginaryUnit, 1 + ImaginaryUnit, 5*ImaginaryUnit, -1, -5, -10]);
  Test('sort(v, "argument descending")', [-10, -5, -1, 5*ImaginaryUnit, 1 + ImaginaryUnit, 12 + 10*ImaginaryUnit, 5, 3, 0, 12 - 10*ImaginaryUnit, -5*ImaginaryUnit, -4*ImaginaryUnit]);
  Test('sort(v, "modulus argument")', [0, -1, 1 + ImaginaryUnit, 3, -4*ImaginaryUnit, -5*ImaginaryUnit, 5, 5*ImaginaryUnit, -5, -10, 12 - 10*ImaginaryUnit, 12 + 10*ImaginaryUnit]);
  Test('sort(v, "modulus argument descending")', [12 + 10*ImaginaryUnit, 12 - 10*ImaginaryUnit, -10, -5, 5*ImaginaryUnit, 5, -5*ImaginaryUnit, -4*ImaginaryUnit, 3, 1 + ImaginaryUnit, -1, 0]);

  Test('A ≔ ❨❨3, 7, 0, −4❩, ❨−10, 4, 1, 4❩, ❨5, 8, 0, 1❩❩', 4, [3, 7, 0, -4, -10, 4, 1, 4, 5, 8, 0, 1]);

  Test('sort(A)', 4, [-10, -4, 0, 0, 1, 1, 3, 4, 4, 5, 7, 8]);
  Test('sort(A, "standard order")', 4, [-10, -4, 0, 0, 1, 1, 3, 4, 4, 5, 7, 8]);
  Test('sort(A, "standard order descending")', 4, [8, 7, 5, 4, 4, 3, 1, 1, 0, 0, -4, -10]);
  Test('sort(A, "absolute value")', 4, [0, 0, 1, 1, 3, -4, 4, 4, 5, 7, 8, -10]);
  Test('sort(A, "absolute value descending")', 4, [-10, 8, 7, 5, 4, 4, -4, 3, 1, 1, 0, 0]);
  Test('sort(A, "real and imaginary parts")', 4, [-10, -4, 0, 0, 1, 1, 3, 4, 4, 5, 7, 8]);
  Test('sort(A, "real and imaginary parts descending")', 4, [8, 7, 5, 4, 4, 3, 1, 1, 0, 0, -4, -10]);
  Test('sort(A, "modulus")', 4, [0, 0, 1, 1, 3, -4, 4, 4, 5, 7, 8, -10]);
  Test('sort(A, "modulus descending")', 4, [-10, 8, 7, 5, 4, 4, -4, 3, 1, 1, 0, 0]);
  Test('sort(A, "argument")', 4, [0, 0, 1, 1, 3, 4, 4, 5, 7, 8, -4, -10]);
  Test('sort(A, "argument descending")', 4, [-10, -4, 8, 7, 5, 4, 4, 3, 1, 1, 0, 0]);
  Test('sort(A, "modulus argument")', 4, [0, 0, 1, 1, 3, 4, 4, -4, 5, 7, 8, -10]);
  Test('sort(A, "modulus argument descending")', 4, [-10, 8, 7, 5, -4, 4, 4, 3, 1, 1, 0, 0]);

  Test('A ≔ ComplexMatrix(❨❨3, 7, 0, −4❩, ❨−10, 4, 1, 4❩, ❨5, 8, 0, 1❩❩)', 4, [TASC(3), 7, 0, -4, -10, 4, 1, 4, 5, 8, 0, 1]);

  Test('sort(A)', failure, 'No comparison method specified.');
  Test('sort(A, "standard order")', 4, [TASC(-10), -4, 0, 0, 1, 1, 3, 4, 4, 5, 7, 8]);
  Test('sort(A, "standard order descending")', 4, [TASC(8), 7, 5, 4, 4, 3, 1, 1, 0, 0, -4, -10]);
  Test('sort(A, "absolute value")', 4, [TASC(0), 0, 1, 1, 3, -4, 4, 4, 5, 7, 8, -10]);
  Test('sort(A, "absolute value descending")', 4, [TASC(-10), 8, 7, 5, 4, 4, -4, 3, 1, 1, 0, 0]);
  Test('sort(A, "real and imaginary parts")', 4, [TASC(-10), -4, 0, 0, 1, 1, 3, 4, 4, 5, 7, 8]);
  Test('sort(A, "real and imaginary parts descending")', 4, [TASC(8), 7, 5, 4, 4, 3, 1, 1, 0, 0, -4, -10]);
  Test('sort(A, "modulus")', 4, [TASC(0), 0, 1, 1, 3, -4, 4, 4, 5, 7, 8, -10]);
  Test('sort(A, "modulus descending")', 4, [TASC(-10), 8, 7, 5, 4, 4, -4, 3, 1, 1, 0, 0]);
  Test('sort(A, "argument")', 4, [TASC(0), 0, 1, 1, 3, 4, 4, 5, 7, 8, -4, -10]);
  Test('sort(A, "argument descending")', 4, [TASC(-10), -4, 8, 7, 5, 4, 4, 3, 1, 1, 0, 0]);
  Test('sort(A, "modulus argument")', 4, [TASC(0), 0, 1, 1, 3, 4, 4, -4, 5, 7, 8, -10]);
  Test('sort(A, "modulus argument descending")', 4, [TASC(-10), 8, 7, 5, -4, 4, 4, 3, 1, 1, 0, 0]);

  Test('A ≔ ❨❨4, −2, 5, 0❩, ❨−5, 1, 5⋅i, 10❩, ❨−5⋅i, −2⋅i, −7, i❩❩',
    4, [4, -2, 5, 0, -5, 1, 5*ImaginaryUnit, 10, -5*ImaginaryUnit, -2*ImaginaryUnit, -7, ImaginaryUnit]);

  Test('sort(A)', failure, 'No comparison method specified.');
  Test('sort(A, "standard order")', failure, 'Complex matrix cannot be converted to real matrix because it contains a non-real component 5⋅i.');
  Test('sort(A, "standard order descending")', failure, 'Complex matrix cannot be converted to real matrix because it contains a non-real component 5⋅i.');
  Test('sort(A, "absolute value")', failure, 'Complex matrix cannot be converted to real matrix because it contains a non-real component 5⋅i.');
  Test('sort(A, "absolute value descending")', failure, 'Complex matrix cannot be converted to real matrix because it contains a non-real component 5⋅i.');
  Test('sort(A, "real and imaginary parts")', 4, [-7, -5, -2, -5*ImaginaryUnit, -2*ImaginaryUnit, 0, ImaginaryUnit, 5*ImaginaryUnit, 1, 4, 5, 10]);
  Test('sort(A, "real and imaginary parts descending")', 4, [10, 5, 4, 1, 5*ImaginaryUnit, ImaginaryUnit, 0, -2*ImaginaryUnit, -5*ImaginaryUnit, -2, -5, -7]);
  Test('sort(A, "modulus")', 4, [0, ImaginaryUnit, 1, -2, -2*ImaginaryUnit, 4, -5, -5*ImaginaryUnit, 5*ImaginaryUnit, 5, -7, 10]);
  Test('sort(A, "modulus descending")', 4, [10, -7, 5, 5*ImaginaryUnit, -5*ImaginaryUnit, -5, 4, -2*ImaginaryUnit, -2, 1, ImaginaryUnit, 0]);
  Test('sort(A, "argument")', 4, [-2*ImaginaryUnit, -5*ImaginaryUnit, 0, 1, 4, 5, 10, ImaginaryUnit, 5*ImaginaryUnit, -2, -5, -7]);
  Test('sort(A, "argument descending")', 4, [-7, -5, -2, 5*ImaginaryUnit, ImaginaryUnit, 10, 5, 4, 1, 0, -5*ImaginaryUnit, -2*ImaginaryUnit]);
  Test('sort(A, "modulus argument")', 4, [0, 1, ImaginaryUnit, -2*ImaginaryUnit, -2, 4, -5*ImaginaryUnit, 5, 5*ImaginaryUnit, -5, -7, 10]);
  Test('sort(A, "modulus argument descending")', 4, [10, -7, -5, 5*ImaginaryUnit, 5, -5*ImaginaryUnit, 4, -2, -2*ImaginaryUnit, ImaginaryUnit, 1, 0]);

  Test('sort(v, "I will have order!")', failure, 'Unknown comparison method "i will have order!".');

  Test('#sort(DebugObject("empty vector"))', 0);
  Test('#sort(ComplexVector(DebugObject("empty vector")), "modulus")', 0);
  Test('#sort(DebugObject("empty matrix"))', 0);
  Test('#sort(ComplexMatrix(DebugObject("empty matrix")), "modulus")', 0);

  Test('L ≔ ''(5, 2, −3, 0, −2, 2, 1, 7, −10, 4, 2, 6)', intarr([5, 2, -3, 0, -2, 2, 1, 7, -10, 4, 2, 6]));

  Test('sort(L)', intarr([-10, -3, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "standard order")', intarr([-10, -3, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "standard order descending")', intarr([7, 6, 5, 4, 2, 2, 2, 1, 0, -2, -3, -10]));
  Test('sort(L, "absolute value")', intarr([0, 1, -2, 2, 2, 2, -3, 4, 5, 6, 7, -10]));
  Test('sort(L, "absolute value descending")', intarr([-10, 7, 6, 5, 4, -3, 2, 2, 2, -2, 1, 0]));
  Test('sort(L, "real and imaginary parts")', intarr([-10, -3, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "real and imaginary parts descending")', intarr([7, 6, 5, 4, 2, 2, 2, 1, 0, -2, -3, -10]));
  Test('sort(L, "modulus")', intarr([0, 1, -2, 2, 2, 2, -3, 4, 5, 6, 7, -10]));
  Test('sort(L, "modulus descending")', intarr([-10, 7, 6, 5, 4, -3, 2, 2, 2, -2, 1, 0]));
  Test('sort(L, "argument")', intarr([0, 1, 2, 2, 2, 4, 5, 6, 7, -2, -3, -10]));
  Test('sort(L, "argument descending")', intarr([-10, -3, -2, 7, 6, 5, 4, 2, 2, 2, 1, 0]));
  Test('sort(L, "modulus argument")', intarr([0, 1, 2, 2, 2, -2, -3, 4, 5, 6, 7, -10]));
  Test('sort(L, "modulus argument descending")', intarr([-10, 7, 6, 5, 4, -3, -2, 2, 2, 2, 1, 0]));

  Test('L ≔ ''(5, 2, −3.0, 0, −2.0, 2, 1, 7, −10, 4, 2, 6.0)', asoarr([5, 2, -3.0, 0, -2.0, 2, 1, 7, -10, 4, 2, 6.0]));

  Test('sort(L)', asoarr([-10, -3.0, -2.0, 0, 1, 2, 2, 2, 4, 5, 6.0, 7]));
  Test('sort(L, "standard order")', asoarr([-10, -3.0, -2.0, 0, 1, 2, 2, 2, 4, 5, 6.0, 7]));
  Test('sort(L, "standard order descending")', asoarr([7, 6.0, 5, 4, 2, 2, 2, 1, 0, -2.0, -3.0, -10]));
  Test('sort(L, "absolute value")', asoarr([0, 1, -2.0, 2, 2, 2, -3.0, 4, 5, 6.0, 7, -10]));
  Test('sort(L, "absolute value descending")', asoarr([-10, 7, 6.0, 5, 4, -3.0, 2, 2, 2, -2.0, 1, 0]));
  Test('sort(L, "real and imaginary parts")', asoarr([-10, -3.0, -2.0, 0, 1, 2, 2, 2, 4, 5, 6.0, 7]));
  Test('sort(L, "real and imaginary parts descending")', asoarr([7, 6.0, 5, 4, 2, 2, 2, 1, 0, -2.0, -3.0, -10]));
  Test('sort(L, "modulus")', asoarr([0, 1, -2.0, 2, 2, 2, -3.0, 4, 5, 6.0, 7, -10]));
  Test('sort(L, "modulus descending")', asoarr([-10, 7, 6.0, 5, 4, -3.0, 2, 2, 2, -2.0, 1, 0]));
  Test('sort(L, "argument")', asoarr([0, 1, 2, 2, 2, 4, 5, 6.0, 7, -2.0, -3.0, -10]));
  Test('sort(L, "argument descending")', asoarr([-10, -3.0, -2.0, 7, 6.0, 5, 4, 2, 2, 2, 1, 0]));
  Test('sort(L, "modulus argument")', asoarr([0, 1, 2, 2, 2, -2.0, -3.0, 4, 5, 6.0, 7, -10]));
  Test('sort(L, "modulus argument descending")', asoarr([-10, 7, 6.0, 5, 4, -3.0, -2.0, 2, 2, 2, 1, 0]));

  Test('L ≔ ''(5, 2, −3.5, 0, −2, 2, 1, 7, −10, 4, 2, 6)', asoarr([5, 2, -3.5, 0, -2, 2, 1, 7, -10, 4, 2, 6]));

  Test('sort(L)', asoarr([-10, -3.5, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "standard order")', asoarr([-10, -3.5, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "standard order descending")', asoarr([7, 6, 5, 4, 2, 2, 2, 1, 0, -2, -3.5, -10]));
  Test('sort(L, "absolute value")', asoarr([0, 1, -2, 2, 2, 2, -3.5, 4, 5, 6, 7, -10]));
  Test('sort(L, "absolute value descending")', asoarr([-10, 7, 6, 5, 4, -3.5, 2, 2, 2, -2, 1, 0]));
  Test('sort(L, "real and imaginary parts")', asoarr([-10, -3.5, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "real and imaginary parts descending")', asoarr([7, 6, 5, 4, 2, 2, 2, 1, 0, -2, -3.5, -10]));
  Test('sort(L, "modulus")', asoarr([0, 1, -2, 2, 2, 2, -3.5, 4, 5, 6, 7, -10]));
  Test('sort(L, "modulus descending")', asoarr([-10, 7, 6, 5, 4, -3.5, 2, 2, 2, -2, 1, 0]));
  Test('sort(L, "argument")', asoarr([0, 1, 2, 2, 2, 4, 5, 6, 7, -2, -3.5, -10]));
  Test('sort(L, "argument descending")', asoarr([-10, -3.5, -2, 7, 6, 5, 4, 2, 2, 2, 1, 0]));
  Test('sort(L, "modulus argument")', asoarr([0, 1, 2, 2, 2, -2, -3.5, 4, 5, 6, 7, -10]));
  Test('sort(L, "modulus argument descending")', asoarr([-10, 7, 6, 5, 4, -3.5, -2, 2, 2, 2, 1, 0]));

  Test('L ≔ ''(5, 2, −3.5, 0, −2, 2.0, 1, 7, −10, 4, 2, 6)', asoarr([5, 2, -3.5, 0, -2, 2.0, 1, 7, -10, 4, 2, 6]));

  Test('sort(L)', asoarr([-10, -3.5, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "standard order")', asoarr([-10, -3.5, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "standard order descending")', asoarr([7, 6, 5, 4, 2, 2, 2, 1, 0, -2, -3.5, -10]));
  Test('sort(L, "absolute value")', asoarr([0, 1, -2, 2, 2, 2, -3.5, 4, 5, 6, 7, -10]));
  Test('sort(L, "absolute value descending")', asoarr([-10, 7, 6, 5, 4, -3.5, 2, 2, 2, -2, 1, 0]));
  Test('sort(L, "real and imaginary parts")', asoarr([-10, -3.5, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "real and imaginary parts descending")', asoarr([7, 6, 5, 4, 2, 2, 2, 1, 0, -2, -3.5, -10]));
  Test('sort(L, "modulus")', asoarr([0, 1, -2, 2, 2, 2, -3.5, 4, 5, 6, 7, -10]));
  Test('sort(L, "modulus descending")', asoarr([-10, 7, 6, 5, 4, -3.5, 2, 2, 2, -2, 1, 0]));
  Test('sort(L, "argument")', asoarr([0, 1, 2, 2, 2, 4, 5, 6, 7, -2, -3.5, -10]));
  Test('sort(L, "argument descending")', asoarr([-10, -3.5, -2, 7, 6, 5, 4, 2, 2, 2, 1, 0]));
  Test('sort(L, "modulus argument")', asoarr([0, 1, 2, 2, 2, -2, -3.5, 4, 5, 6, 7, -10]));
  Test('sort(L, "modulus argument descending")', asoarr([-10, 7, 6, 5, 4, -3.5, -2, 2, 2, 2, 1, 0]));

  Test('L ≔ ''(5, 2, −3.5, 0, −2, 2.0, 1, 7, −10, 4, 2, ComplexNumber(6));', null);

  Test('sort(L)', asoarr([-10, -3.5, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "standard order")', asoarr([-10, -3.5, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "standard order descending")', asoarr([7, 6, 5, 4, 2, 2, 2, 1, 0, -2, -3.5, -10]));
  Test('sort(L, "absolute value")', asoarr([0, 1, -2, 2, 2, 2, -3.5, 4, 5, 6, 7, -10]));
  Test('sort(L, "absolute value descending")', asoarr([-10, 7, 6, 5, 4, -3.5, 2, 2, 2, -2, 1, 0]));
  Test('sort(L, "real and imaginary parts")', asoarr([-10, -3.5, -2, 0, 1, 2, 2, 2, 4, 5, 6, 7]));
  Test('sort(L, "real and imaginary parts descending")', asoarr([7, 6, 5, 4, 2, 2, 2, 1, 0, -2, -3.5, -10]));
  Test('sort(L, "modulus")', asoarr([0, 1, -2, 2, 2, 2, -3.5, 4, 5, 6, 7, -10]));
  Test('sort(L, "modulus descending")', asoarr([-10, 7, 6, 5, 4, -3.5, 2, 2, 2, -2, 1, 0]));
  Test('sort(L, "argument")', asoarr([0, 1, 2, 2, 2, 4, 5, 6, 7, -2, -3.5, -10]));
  Test('sort(L, "argument descending")', asoarr([-10, -3.5, -2, 7, 6, 5, 4, 2, 2, 2, 1, 0]));
  Test('sort(L, "modulus argument")', asoarr([0, 1, 2, 2, 2, -2, -3.5, 4, 5, 6, 7, -10]));
  Test('sort(L, "modulus argument descending")', asoarr([-10, 7, 6, 5, 4, -3.5, -2, 2, 2, 2, 1, 0]));

  Test('L ≔ ''("b" : 23, 5 + i, "rabbit", 5, −3, "dog", 1, "cat", 5.8, 3⋅i, true, −10, "a" : 21, 0, −i, "rat", false)',
    TAlgosimArray.CreateWithValue(
      [
        ASOMember(sm('b', ASOInt(23))),
        ASO(5 + ImaginaryUnit),
        ASO('rabbit'),
        ASOInt(5),
        ASOInt(-3),
        ASO('dog'),
        ASOInt(1),
        ASO('cat'),
        ASO(5.8),
        ASO(3*ImaginaryUnit),
        ASO(True),
        ASOInt(-10),
        ASOMember(sm('a', ASOInt(21))),
        ASOInt(0),
        ASO(-ImaginaryUnit),
        ASO('rat'),
        ASO(False)
      ]
    )
  );

  Test('sort(L)',
    TAlgosimArray.CreateWithValue(
      [
        ASO(False),
        ASO(True),
        ASOInt(-10),
        ASOInt(-3),
        ASO(-ImaginaryUnit),
        ASOInt(0),
        ASO(3*ImaginaryUnit),
        ASOInt(1),
        ASOInt(5),
        ASO(5 + ImaginaryUnit),
        ASO(5.8),
        ASO('cat'),
        ASO('dog'),
        ASO('rabbit'),
        ASO('rat'),
        ASOMember(sm('a', ASOInt(21))),
        ASOMember(sm('b', ASOInt(23)))
      ]
    )
  );

  Test('sort(L, "standard order")', failure, 'Object isn''t a real number.');
  Test('sort(L, "standard order descending")', failure, 'Object isn''t a real number.');
  Test('sort(L, "absolute value")', failure, 'Object isn''t a real number.');
  Test('sort(L, "absolute value descending")', failure, 'Object isn''t a real number.');
  Test('sort(L, "real and imaginary parts")', failure, 'Object isn''t a complex number.');
  Test('sort(L, "real and imaginary parts descending")', failure, 'Object isn''t a complex number.');
  Test('sort(L, "modulus")', failure, 'Object isn''t a complex number.');
  Test('sort(L, "modulus descending")', failure, 'Object isn''t a complex number.');
  Test('sort(L, "argument")', failure, 'Object isn''t a complex number.');
  Test('sort(L, "argument descending")', failure, 'Object isn''t a complex number.');
  Test('sort(L, "modulus argument")', failure, 'Object isn''t a complex number.');
  Test('sort(L, "modulus argument descending")', failure, 'Object isn''t a complex number.');

  Test('sort(''())', asoarr([]));
  Test('sort(''(), "standard order")', asoarr([]));
  Test('sort(''(), "modulus")', asoarr([]));

  Test('L ≔ ''(3, 5.5, 0, −2/3, i, 3⋅i, 7, −10, −3, 5, 1+i, −3⋅i, 6, 5/2, −3.8, 5.0)',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(3),
        ASO(5.5),
        ASOInt(0),
        ASO(rat(-2, 3)),
        ASO(ImaginaryUnit),
        ASO(3*ImaginaryUnit),
        ASOInt(7),
        ASOInt(-10),
        ASOInt(-3),
        ASOInt(5),
        ASO(1+ImaginaryUnit),
        ASO(-3*ImaginaryUnit),
        ASOInt(6),
        ASO(rat(5, 2)),
        ASO(-3.8),
        ASO(5.0)
      ]
    )
  );

  Test('sort(L)',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(-10),
        ASO(-3.8),
        ASOInt(-3),
        ASO(rat(-2, 3)),
        ASO(-3*ImaginaryUnit),
        ASOInt(0),
        ASO(ImaginaryUnit),
        ASO(3*ImaginaryUnit),
        ASO(1+ImaginaryUnit),
        ASO(rat(5, 2)),
        ASOInt(3),
        ASOInt(5),
        ASO(5.0),
        ASO(5.5),
        ASOInt(6),
        ASOInt(7)
      ]
    )
  );

  Test('sort(L, "real and imaginary parts")',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(-10),
        ASO(-3.8),
        ASOInt(-3),
        ASO(rat(-2, 3)),
        ASO(-3*ImaginaryUnit),
        ASOInt(0),
        ASO(ImaginaryUnit),
        ASO(3*ImaginaryUnit),
        ASO(1+ImaginaryUnit),
        ASO(rat(5, 2)),
        ASOInt(3),
        ASOInt(5),
        ASO(5.0),
        ASO(5.5),
        ASOInt(6),
        ASOInt(7)
      ]
    )
  );

  Test('sort(L, "real and imaginary parts descending")',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(7),
        ASOInt(6),
        ASO(5.5),
        ASO(5.0),
        ASOInt(5),
        ASOInt(3),
        ASO(rat(5, 2)),
        ASO(1+ImaginaryUnit),
        ASO(3*ImaginaryUnit),
        ASO(ImaginaryUnit),
        ASOInt(0),
        ASO(-3*ImaginaryUnit),
        ASO(rat(-2, 3)),
        ASOInt(-3),
        ASO(-3.8),
        ASOInt(-10)
      ]
    )
  );

  Test('sort(L, "standard order")', failure, 'Object isn''t a real number.');
  Test('sort(L, "standard order descending")', failure, 'Object isn''t a real number.');
  Test('sort(L, "absolute value")', failure, 'Object isn''t a real number.');
  Test('sort(L, "absolute value descending")', failure, 'Object isn''t a real number.');

  Test('sort(L, "modulus")',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(0),
        ASO(rat(-2, 3)),
        ASO(ImaginaryUnit),
        ASO(1+ImaginaryUnit),
        ASO(rat(5, 2)),
        ASOInt(-3),
        ASO(-3*ImaginaryUnit),
        ASO(3*ImaginaryUnit),
        ASOInt(3),
        ASO(-3.8),
        ASOInt(5),
        ASO(5.0),
        ASO(5.5),
        ASOInt(6),
        ASOInt(7),
        ASOInt(-10)
      ]
    )
  );

  Test('sort(L, "modulus descending")',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(-10),
        ASOInt(7),
        ASOInt(6),
        ASO(5.5),
        ASO(5.0),
        ASOInt(5),
        ASO(-3.8),
        ASOInt(3),
        ASO(3*ImaginaryUnit),
        ASO(-3*ImaginaryUnit),
        ASOInt(-3),
        ASO(rat(5, 2)),
        ASO(1+ImaginaryUnit),
        ASO(ImaginaryUnit),
        ASO(rat(-2, 3)),
        ASOInt(0)
      ]
    )
  );

  Test('sort(L, "argument")',
    TAlgosimArray.CreateWithValue(
      [
        ASO(-3*ImaginaryUnit),
        ASOInt(0),
        ASO(rat(5, 2)),
        ASOInt(3),
        ASO(5.0),
        ASOInt(5),
        ASO(5.5),
        ASOInt(6),
        ASOInt(7),
        ASO(1+ImaginaryUnit),
        ASO(ImaginaryUnit),
        ASO(3*ImaginaryUnit),
        ASO(rat(-2, 3)),
        ASOInt(-3),
        ASO(-3.8),
        ASOInt(-10)
      ]
    )
  );

  Test('sort(L, "argument descending")',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(-10),
        ASO(-3.8),
        ASOInt(-3),
        ASO(rat(-2, 3)),
        ASO(3*ImaginaryUnit),
        ASO(ImaginaryUnit),
        ASO(1+ImaginaryUnit),
        ASOInt(7),
        ASOInt(6),
        ASO(5.5),
        ASOInt(5),
        ASO(5.0),
        ASOInt(3),
        ASO(rat(5, 2)),
        ASOInt(0),
        ASO(-3*ImaginaryUnit)
      ]
    )
  );

  Test('sort(L, "modulus argument")',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(0),
        ASO(rat(-2, 3)),
        ASO(ImaginaryUnit),
        ASO(1+ImaginaryUnit),
        ASO(rat(5, 2)),
        ASO(-3*ImaginaryUnit),
        ASOInt(3),
        ASO(3*ImaginaryUnit),
        ASOInt(-3),
        ASO(-3.8),
        ASOInt(5),
        ASO(5.0),
        ASO(5.5),
        ASOInt(6),
        ASOInt(7),
        ASOInt(-10)
      ]
    )
  );

  Test('sort(L, "modulus argument descending")',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(-10),
        ASOInt(7),
        ASOInt(6),
        ASO(5.5),
        ASO(5.0),
        ASOInt(5),
        ASO(-3.8),
        ASOInt(-3),
        ASO(3*ImaginaryUnit),
        ASOInt(3),
        ASO(-3*ImaginaryUnit),
        ASO(rat(5, 2)),
        ASO(1+ImaginaryUnit),
        ASO(ImaginaryUnit),
        ASO(rat(-2, 3)),
        ASOInt(0)
      ]
    )
  );

  Test('L ≔ ''("cat", "RAT", "rabbit", "dog", "HORSE", "bird", "guinea pig", "hamster", "PIG", "thestral", "elephant", "ACROMANTULA", "owl")',
    TAlgosimArray.CreateWithValue(
      [
        ASO('cat'),
        ASO('RAT'),
        ASO('rabbit'),
        ASO('dog'),
        ASO('HORSE'),
        ASO('bird'),
        ASO('guinea pig'),
        ASO('hamster'),
        ASO('PIG'),
        ASO('thestral'),
        ASO('elephant'),
        ASO('ACROMANTULA'),
        ASO('owl')
      ]
    )
  );

  Test('sort(L)',
    TAlgosimArray.CreateWithValue(
      [
        ASO('ACROMANTULA'),
        ASO('bird'),
        ASO('cat'),
        ASO('dog'),
        ASO('elephant'),
        ASO('guinea pig'),
        ASO('hamster'),
        ASO('HORSE'),
        ASO('owl'),
        ASO('PIG'),
        ASO('rabbit'),
        ASO('RAT'),
        ASO('thestral')
      ]
    )
  );

  Test('L ≔ ''("cat", "CAT", "Cat", "cAT", "cAt", "CaT", "caT", "CAt")',
    strarr(
      [
        'cat',
        'CAT',
        'Cat',
        'cAT',
        'cAt',
        'CaT',
        'caT',
        'CAt'
      ]
    )
  );

  Test('sort(L)',
    strarr(
      [
        'CAT',
        'CAt',
        'CaT',
        'Cat',
        'cAT',
        'cAt',
        'caT',
        'cat'
      ]
    )
  );

  Test('S ≔ ToSet(L);', null);
  Test('sort(S)', failure, 'An object of type "set" cannot be sorted');

  // CustomSort

  Test('v ≔ ❨6, 3, −8, 2, 9, 5, −1, 0, 7, −4❩', [6, 3, -8, 2, 9, 5, -1, 0, 7, -4]);

  Test('CustomSort(v, CompareValue(left, right))',
    [-8, -4, -1, 0, 2, 3, 5, 6, 7, 9]);

  Test('CustomSort(v, left − right)',
    [-8, -4, -1, 0, 2, 3, 5, 6, 7, 9]);

  Test('CustomSort(v, CompareValue(abs(left), abs(right)))',
    [0, -1, 2, 3, -4, 5, 6, 7, -8, 9]);

  Test('w ≔ CustomSort(v, CompareValue(Iverson(odd(left)), Iverson(odd(right))));', null);
  Test('sort(part(w, range(1, 5)))', [-8, -4, 0, 2, 6]);
  Test('sort(part(w, range(6, 10)))', [-1, 3, 5, 7, 9]);

  Test('v ≔ ❨6.3, 4.2, 8.7, 5.1, 7.8, 1.9, 2.6, 3.5❩', [6.3, 4.2, 8.7, 5.1, 7.8, 1.9, 2.6, 3.5]);

  Test('CustomSort(v, CompareValue(left, right))',
    [1.9, 2.6, 3.5, 4.2, 5.1, 6.3, 7.8, 8.7]);

  Test('CustomSort(v, CompareValue(frac(left), frac(right)))',
    [5.1, 4.2, 6.3, 3.5, 2.6, 8.7, 7.8, 1.9]);

  Test('CustomSort(v, frac(left) - frac(right))',
    [5.1, 4.2, 6.3, 3.5, 2.6, 8.7, 7.8, 1.9]);

  Test('CustomSort(v, wrong - right)', failure, 'Unknown identifier "wrong".');

  Test('CustomSort(v, left − right / 0)', failure, 'Floating point division by zero');

  Test('v ≔ ❨−3⋅i, 5, 7, −1, 6, 8⋅i, 2⋅i, 9, −4❩', [-3*ImaginaryUnit, 5, 7, -1, 6, 8*ImaginaryUnit, 2*ImaginaryUnit, 9, -4]);

  Test('w ≔ CustomSort(v, CompareValue(Re(left), Re(right)));', null);
  Test('part(w, range(1, 2))', [TASC(-4), -1]);
  Test('sort(part(w, range(3, 5)), "real and imaginary parts")', [-3*ImaginaryUnit, 2*ImaginaryUnit, 8*ImaginaryUnit]);
  Test('part(w, range(6, 9))', [TASC(5), 6, 7, 9]);

  Test('w ≔ CustomSort(v, CompareValue(Im(left), Im(right)));', null);
  Test('part(w, range(1, 1))', [-3*ImaginaryUnit]);
  Test('sort(part(w, range(2, 7)), "real and imaginary parts")', [TASC(-4), -1, 5, 6, 7, 9]);
  Test('part(w, range(8, 9))', [2*ImaginaryUnit, 8*ImaginaryUnit]);

  Test('CustomSort(v, CompareValue(abs(left), abs(right)))',
    [-1, 2*ImaginaryUnit, -3*ImaginaryUnit, -4, 5, 6, 7, 8*ImaginaryUnit, 9]);

  Test('CustomSort(v, sin(left + right))', failure, 'The comparer was expected to return a real number, but an object of type complex number was returned.');

  Test('A ≔ ❨❨5.3, 2.5, −7.2, 1.6❩, ❨9.7, 3.4, −8.8, 4.1❩❩', 4,
    [5.3, 2.5, -7.2, 1.6, 9.7, 3.4, -8.8, 4.1]);

  Test('CustomSort(A, CompareValue(left, right))', 4,
    [-8.8, -7.2, 1.6, 2.5, 3.4, 4.1, 5.3, 9.7]);

  Test('CustomSort(A, CompareValue(abs(left), abs(right)))', 4,
    [1.6, 2.5, 3.4, 4.1, 5.3, -7.2, -8.8, 9.7]);

  Test('CustomSort(A, CompareValue(abs(frac(left)), abs(frac(right))))', 4,
    [4.1, -7.2, 5.3, 3.4, 2.5, 1.6, 9.7, -8.8]);

  Test('CustomSort(A, abs(frac(left)) − abs(frac(right)))', 4,
    [4.1, -7.2, 5.3, 3.4, 2.5, 1.6, 9.7, -8.8]);

  Test('A ≔ ❨❨2, 3⋅i, 5, 7, −3❩, ❨−3⋅i, −8, 3, 0, 1 + i❩❩', 5,
    [2, 3*ImaginaryUnit, 5, 7, -3, -3*ImaginaryUnit, -8, 3, 0, 1+ImaginaryUnit]);

  Test('B ≔ CustomSort(A, Re(left) − Re(right));', null);

  Test('part(vector(B), range(1, 2))', [TASC(-8), -3]);
  Test('sort(part(vector(B), range(3, 5)), "real and imaginary parts")', [-3*ImaginaryUnit, 0, 3*ImaginaryUnit]);
  Test('part(vector(B), range(6, 10))', [1 + ImaginaryUnit, 2, 3, 5, 7]);

  Test('B ≔ CustomSort(A, Im(left) − Im(right));', null);
  Test('B[1, 1]', -3*ImaginaryUnit);
  Test('sort(part(vector(B), range(2, 8)), "real and imaginary parts")', [TASC(-8), -3, 0, 2, 3, 5, 7]);
  Test('B[2, 4]', 1 + ImaginaryUnit);
  Test('B[2, 5]', 3*ImaginaryUnit);

  Test('B ≔ CustomSort(A, abs(left) − abs(right));', null);
  Test('part(vector(B), range(1, 3))', [0, 1 + ImaginaryUnit, 2]);
  Test('sort(part(vector(B), range(4, 7)), "real and imaginary parts")', [-3, -3*ImaginaryUnit, 3*ImaginaryUnit, 3]);
  Test('part(vector(B), range(8, 10))', [TASC(5), 7, -8]);

  Test('L ≔ ''(5, 5/2, 3.6, −4.2, 3⋅i, 5.0, 7/2, −6, 1−i)',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(5),
        ASO(rat(5, 2)),
        ASO(3.6),
        ASO(-4.2),
        ASO(3*ImaginaryUnit),
        ASO(5.0),
        ASO(rat(7, 2)),
        ASOInt(-6),
        ASO(1-ImaginaryUnit)
      ]
    )
  );

  Test('CustomSort(L, Re(left) - Re(right))',
    TAlgosimArray.CreateWithValue(
      [
        ASOInt(-6),
        ASO(-4.2),
        ASO(3*ImaginaryUnit),
        ASO(1-ImaginaryUnit),
        ASO(rat(5, 2)),
        ASO(rat(7, 2)),
        ASO(3.6),
        ASOInt(5),
        ASO(5.0)
      ]
    )
  );

  Test('CustomSort(L, abs(left) - abs(right))',
    TAlgosimArray.CreateWithValue(
      [
        ASO(1-ImaginaryUnit),
        ASO(rat(5, 2)),
        ASO(3*ImaginaryUnit),
        ASO(rat(7, 2)),
        ASO(3.6),
        ASO(-4.2),
        ASOInt(5),
        ASO(5.0),
        ASOInt(-6)
      ]
    )
  );

  Test('L ≔ ''("RAT", "rabbit", "cat", "DOG", "horse", "guinea pig", "THESTRAL")',
    strarr(
      [
        'RAT',
        'rabbit',
        'cat',
        'DOG',
        'horse',
        'guinea pig',
        'THESTRAL'
      ]
    )
  );

  Test('CustomSort(L, CompareValue(left, right))',
    strarr(
      [
        'cat',
        'DOG',
        'guinea pig',
        'horse',
        'rabbit',
        'RAT',
        'THESTRAL'
      ]
    )
  );

  Test('CustomSort(L, CompareString(left, right))',
    strarr(
      [
        'DOG',
        'RAT',
        'THESTRAL',
        'cat',
        'guinea pig',
        'horse',
        'rabbit'
      ]
    )
  );

  Test('CustomSort(L, #left − #right + CompareValue(left, right) / 1000)',
    strarr(
      [
        'cat',
        'DOG',
        'RAT',
        'horse',
        'rabbit',
        'THESTRAL',
        'guinea pig'
      ]
    )
  );

  Test('CustomSort(L, left + right)', failure, 'The comparer was expected to return a real number, but an object of type string was returned.');

  Test('p ≔ (name, age) ↦ struct("name": name, "age": age);', null);
  Test('L ≔ ''(p("Andrew", 32), p("Sarah", 28), p("John", 53), p("David", 45), p("Elizabeth", 80), p("Eric", 16), p("Sophie", 22), p("Bill", 75))',
    TAlgosimArray.CreateWithValue(
      [
        SPerson('Andrew', 32),
        SPerson('Sarah', 28),
        SPerson('John', 53),
        SPerson('David', 45),
        SPerson('Elizabeth', 80),
        SPerson('Eric', 16),
        SPerson('Sophie', 22),
        SPerson('Bill', 75)
      ]
    )
  );

  Test('sort(L)',
    TAlgosimArray.CreateWithValue(
      [
        SPerson('Andrew', 32),
        SPerson('Bill', 75),
        SPerson('David', 45),
        SPerson('Elizabeth', 80),
        SPerson('Eric', 16),
        SPerson('John', 53),
        SPerson('Sarah', 28),
        SPerson('Sophie', 22)
      ]
    )
  );

  Test('CustomSort(L, CompareValue(left.name, right.name))',
    TAlgosimArray.CreateWithValue(
      [
        SPerson('Andrew', 32),
        SPerson('Bill', 75),
        SPerson('David', 45),
        SPerson('Elizabeth', 80),
        SPerson('Eric', 16),
        SPerson('John', 53),
        SPerson('Sarah', 28),
        SPerson('Sophie', 22)
      ]
    )
  );

  Test('CustomSort(L, CompareValue(left.age, right.age))',
    TAlgosimArray.CreateWithValue(
      [
        SPerson('Eric', 16),
        SPerson('Sophie', 22),
        SPerson('Sarah', 28),
        SPerson('Andrew', 32),
        SPerson('David', 45),
        SPerson('John', 53),
        SPerson('Bill', 75),
        SPerson('Elizabeth', 80)
      ]
    )
  );

  Test('CustomSort(L, CompareValue(left.height, right.height))', failure, 'There is no member named "height".');

  Test('S ≔ ToSet(L);', null);
  Test('CustomSort(S, CompareValue(left.age, right.age))', failure, 'An object of type "set" cannot be sorted');

  // SortBy

  Test('v ≔ ❨3, −5, 0, 7, 4, −9, 1, 2, −6, 8❩',
    [3, -5, 0, 7, 4, -9, 1, 2, -6, 8]);

  Test('SortBy(v, identity)',
    [-9, -6, -5, 0, 1, 2, 3, 4, 7, 8]);

  Test('SortBy(v, abs)',
    [0, 1, 2, 3, 4, -5, -6, 7, 8, -9]);

  Test('w ≔ SortBy(v, odd);', null);
  Test('sort(part(w, range(1, 5)))', [-6, 0, 2, 4, 8]);
  Test('sort(part(w, range(6, 10)))', [-9, -5, 1, 3, 7]);

  Test('SortBy(v, sin)',
    [4, -9, 0, 3, -6, 7, 1, 2, -5, 8]);

  Test('v ≔ ❨6.2, 7.1, 8.8, 5.3, 1.9, 4.7, 9.5, 3.6, 2.4❩',
    [6.2, 7.1, 8.8, 5.3, 1.9, 4.7, 9.5, 3.6, 2.4]);

  Test('SortBy(v, identity)',
    [1.9, 2.4, 3.6, 4.7, 5.3, 6.2, 7.1, 8.8, 9.5]);

  Test('SortBy(v, trunc)',
    [1.9, 2.4, 3.6, 4.7, 5.3, 6.2, 7.1, 8.8, 9.5]);

  Test('SortBy(v, frac)',
    [7.1, 6.2, 5.3, 2.4, 9.5, 3.6, 4.7, 8.8, 1.9]);

  Test('v ≔ ❨2, 3 + i, −1 − i, 4 + 15⋅i, 7 − 2⋅i, −5 + 5⋅i, −6 − 3⋅i❩',
    [2, 3 + ImaginaryUnit, -1 - ImaginaryUnit, 4 + 15*ImaginaryUnit, 7 - 2*ImaginaryUnit, -5 + 5*ImaginaryUnit, -6 - 3*ImaginaryUnit]);

  Test('SortBy(v, Re)',
    [-6 - 3*ImaginaryUnit, -5 + 5*ImaginaryUnit, -1 - ImaginaryUnit, 2, 3 + ImaginaryUnit, 4 + 15*ImaginaryUnit, 7 - 2*ImaginaryUnit]);

  Test('SortBy(v, Im)',
    [-6 - 3*ImaginaryUnit, 7 - 2*ImaginaryUnit, -1 - ImaginaryUnit, 2, 3 + ImaginaryUnit, -5 + 5*ImaginaryUnit, 4 + 15*ImaginaryUnit]);

  Test('SortBy(v, abs)',
    [-1 - ImaginaryUnit, 2, 3 + ImaginaryUnit, -6 - 3*ImaginaryUnit, -5 + 5*ImaginaryUnit, 7 - 2*ImaginaryUnit, 4 + 15*ImaginaryUnit]);

  Test('SortBy(v, arg)',
    [-6 - 3*ImaginaryUnit, -1 - ImaginaryUnit, 7 - 2*ImaginaryUnit, 2, 3 + ImaginaryUnit, 4 + 15*ImaginaryUnit, -5 + 5*ImaginaryUnit]);

  Test('SortBy(v, z ↦ Re(z) + Im(z))',
    [-6 - 3*ImaginaryUnit, -1 - ImaginaryUnit, -5 + 5*ImaginaryUnit, 2, 3 + ImaginaryUnit, 7 - 2*ImaginaryUnit, 4 + 15*ImaginaryUnit]);

  Test('SortBy(v, z ↦ max(abs(Re(z)), abs(Im(z))))',
    [-1 - ImaginaryUnit, 2, 3 + ImaginaryUnit, -5 + 5*ImaginaryUnit, -6 - 3*ImaginaryUnit, 7 - 2*ImaginaryUnit, 4 + 15*ImaginaryUnit]);

  Test('A ≔ ❨❨6.3, 5.2, 3.7, 9.9❩, ❨4.5, 1.1, 2.0, 8.8❩❩', 4,
    [6.3, 5.2, 3.7, 9.9, 4.5, 1.1, 2.0, 8.8]);

  Test('SortBy(A, trunc)', 4,
    [1.1, 2.0, 3.7, 4.5, 5.2, 6.3, 8.8, 9.9]);

  Test('SortBy(A, frac)', 4,
    [2.0, 1.1, 5.2, 6.3, 4.5, 3.7, 8.8, 9.9]);

  Test('A ≔ ❨❨1 + i/100, −2, 1 + i, 3⋅i❩, ❨−5⋅i, 2 − 2⋅i, 0, −3 − 3⋅i❩❩', 4,
    [1 + ImaginaryUnit/100, -2, 1 + ImaginaryUnit, 3*ImaginaryUnit, -5*ImaginaryUnit, 2 - 2*ImaginaryUnit, 0, -3-3*ImaginaryUnit]);

  Test('SortBy(A, abs)', 4,
    [0, 1 + ImaginaryUnit/100, 1 + ImaginaryUnit, -2, 2 - 2*ImaginaryUnit, 3*ImaginaryUnit, -3 - 3*ImaginaryUnit, -5*ImaginaryUnit]);

  Test('SortBy(A, arg)', 4,
    [-3 - 3*ImaginaryUnit, -5*ImaginaryUnit, 2 - 2*ImaginaryUnit, 0, 1 + ImaginaryUnit/100, 1 + ImaginaryUnit, 3*ImaginaryUnit, -2]);

  Test('L ≔ ''("owl", "hippogriff", "rabbit", "thestral", "snake", "bird", "acromantula", "short-tailed shrew tenrec")',
    strarr(
      [
        'owl',
        'hippogriff',
        'rabbit',
        'thestral',
        'snake',
        'bird',
        'acromantula',
        'short-tailed shrew tenrec'
      ]
    )
  );

  Test('SortBy(L, length)',
    strarr(
      [
        'owl',
        'bird',
        'snake',
        'rabbit',
        'thestral',
        'hippogriff',
        'acromantula',
        'short-tailed shrew tenrec'
      ]
    )
  );

  Test('SortBy(L, (s ↦ count(s, "a") + count(s, "e") + count(s, "i") + count(s, "r") + count(s, "b") + count(s, "d") + count(s, "o") + count(s, "g") + count(s, "p") + count(s, "n") + count(s, "k") + count(s, "i") + count(s, "m")))',
    strarr(
      [
        'owl',
        'thestral',
        'snake',
        'bird',
        'rabbit',
        'acromantula',
        'hippogriff',
        'short-tailed shrew tenrec'
      ]
    )
  );

  Test('p ≔ (name, age) ↦ struct("name": name, "age": age);', null);
  Test('L ≔ ''(p("Andrew", 32), p("Sarah", 28), p("John", 53), p("David", 45), p("Elizabeth", 80), p("Eric", 16), p("Sophie", 22), p("Bill", 75))',
    TAlgosimArray.CreateWithValue(
      [
        SPerson('Andrew', 32),
        SPerson('Sarah', 28),
        SPerson('John', 53),
        SPerson('David', 45),
        SPerson('Elizabeth', 80),
        SPerson('Eric', 16),
        SPerson('Sophie', 22),
        SPerson('Bill', 75)
      ]
    )
  );

  Test('SortBy(L, member("name"))',
    TAlgosimArray.CreateWithValue(
      [
        SPerson('Andrew', 32),
        SPerson('Bill', 75),
        SPerson('David', 45),
        SPerson('Elizabeth', 80),
        SPerson('Eric', 16),
        SPerson('John', 53),
        SPerson('Sarah', 28),
        SPerson('Sophie', 22)
      ]
    )
  );

  Test('SortBy(L, member("age"))',
    TAlgosimArray.CreateWithValue(
      [
        SPerson('Eric', 16),
        SPerson('Sophie', 22),
        SPerson('Sarah', 28),
        SPerson('Andrew', 32),
        SPerson('David', 45),
        SPerson('John', 53),
        SPerson('Bill', 75),
        SPerson('Elizabeth', 80)
      ]
    )
  );

  Test('SortBy(L, member("sex"))', failure, 'There is no member named "sex".');

  Test('S ≔ ToSet(L);', null);
  Test('SortBy(S, member("age"))', failure, 'Object type "set" is not a suitable container.');

  Test('delete(v); delete(A); delete(L); delete(S); delete(w); delete(p)', success);


  //
  // Grouping
  //

  // GroupBy

  Test('L ≔ ''(6, 0, −2, −7, 4, −1, 3, 8, −6, 2, −3, −4, −8, 1, −9, 7)',
    intarr([6, 0, -2, -7, 4, -1, 3, 8, -6, 2, -3, -4, -8, 1, -9, 7]));

  Test('GroupBy(L, IsNegative) @ sort',
    asoarrex(
      [
        intarr([0, 1, 2, 3, 4, 6, 7, 8]),
        intarr([-9, -8, -7, -6, -4, -3, -2, -1])
      ]
    )
  );

  Test('GroupBy(L, odd) @ sort',
    asoarrex(
      [
        intarr([-8, -6, -4, -2, 0, 2, 4, 6, 8]),
        intarr([-9, -7, -3, -1, 1, 3, 7])
      ]
    )
  );

  Test('GroupBy(L, sgn) @ sort',
    asoarrex(
      [
        intarr([-9, -8, -7, -6, -4, -3, -2, -1]),
        intarr([0]),
        intarr([1, 2, 3, 4, 6, 7, 8])
      ]
    )
  );

  Test('GroupBy(L, IsNegative, "neg") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['neg', 'members'],
          [ASO(False), intarr([0, 1, 2, 3, 4, 6, 7, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['neg', 'members'],
          [ASO(True), intarr([-9, -8, -7, -6, -4, -3, -2, -1])]
          )
      ]
    )
  );

  Test('GroupBy(L, odd, "odd") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['odd', 'members'],
          [ASO(False), intarr([-8, -6, -4, -2, 0, 2, 4, 6, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['odd', 'members'],
          [ASO(True), intarr([-9, -7, -3, -1, 1, 3, 7])]
          )
      ]
    )
  );

  Test('GroupBy(L, IsNegative, "neg", "ints") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['neg', 'ints'],
          [ASO(False), intarr([0, 1, 2, 3, 4, 6, 7, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['neg', 'ints'],
          [ASO(True), intarr([-9, -8, -7, -6, -4, -3, -2, -1])]
          )
      ]
    )
  );

  Test('GroupBy(L, odd, "odd", "values") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['odd', 'values'],
          [ASO(False), intarr([-8, -6, -4, -2, 0, 2, 4, 6, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['odd', 'values'],
          [ASO(True), intarr([-9, -7, -3, -1, 1, 3, 7])]
          )
      ]
    )
  );

  Test('S ≔ ToSet(L)',
    intset([6, 0, -2, -7, 4, -1, 3, 8, -6, 2, -3, -4, -8, 1, -9, 7]));

  Test('GroupBy(S, IsNegative) @ sort',
    asoarrex(
      [
        intarr([0, 1, 2, 3, 4, 6, 7, 8]),
        intarr([-9, -8, -7, -6, -4, -3, -2, -1])
      ]
    )
  );

  Test('GroupBy(S, odd) @ sort',
    asoarrex(
      [
        intarr([-8, -6, -4, -2, 0, 2, 4, 6, 8]),
        intarr([-9, -7, -3, -1, 1, 3, 7])
      ]
    )
  );

  Test('GroupBy(S, sgn) @ sort',
    asoarrex(
      [
        intarr([-9, -8, -7, -6, -4, -3, -2, -1]),
        intarr([0]),
        intarr([1, 2, 3, 4, 6, 7, 8])
      ]
    )
  );

  Test('GroupBy(S, IsNegative, "neg") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['neg', 'members'],
          [ASO(False), intarr([0, 1, 2, 3, 4, 6, 7, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['neg', 'members'],
          [ASO(True), intarr([-9, -8, -7, -6, -4, -3, -2, -1])]
          )
      ]
    )
  );

  Test('GroupBy(S, odd, "odd") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['odd', 'members'],
          [ASO(False), intarr([-8, -6, -4, -2, 0, 2, 4, 6, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['odd', 'members'],
          [ASO(True), intarr([-9, -7, -3, -1, 1, 3, 7])]
          )
      ]
    )
  );

  Test('GroupBy(S, IsNegative, "neg", "ints") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['neg', 'ints'],
          [ASO(False), intarr([0, 1, 2, 3, 4, 6, 7, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['neg', 'ints'],
          [ASO(True), intarr([-9, -8, -7, -6, -4, -3, -2, -1])]
          )
      ]
    )
  );

  Test('GroupBy(S, odd, "odd", "values") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['odd', 'values'],
          [ASO(False), intarr([-8, -6, -4, -2, 0, 2, 4, 6, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['odd', 'values'],
          [ASO(True), intarr([-9, -7, -3, -1, 1, 3, 7])]
          )
      ]
    )
  );

  Test('v ≔ ❨6, 0, −2, −7, 4, −1, 3, 8, −6, 2, −3, −4, −8, 1, −9, 7❩',
    [6, 0, -2, -7, 4, -1, 3, 8, -6, 2, -3, -4, -8, 1, -9, 7]);

  Test('GroupBy(v, IsNegative) @ sort',
    asoarrex(
      [
        intarr([0, 1, 2, 3, 4, 6, 7, 8]),
        intarr([-9, -8, -7, -6, -4, -3, -2, -1])
      ]
    )
  );

  Test('GroupBy(v, odd) @ sort',
    asoarrex(
      [
        intarr([-8, -6, -4, -2, 0, 2, 4, 6, 8]),
        intarr([-9, -7, -3, -1, 1, 3, 7])
      ]
    )
  );

  Test('GroupBy(v, sgn) @ sort',
    asoarrex(
      [
        intarr([-9, -8, -7, -6, -4, -3, -2, -1]),
        intarr([0]),
        intarr([1, 2, 3, 4, 6, 7, 8])
      ]
    )
  );

  Test('GroupBy(v, IsNegative, "neg") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['neg', 'members'],
          [ASO(False), intarr([0, 1, 2, 3, 4, 6, 7, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['neg', 'members'],
          [ASO(True), intarr([-9, -8, -7, -6, -4, -3, -2, -1])]
          )
      ]
    )
  );

  Test('GroupBy(v, odd, "odd") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['odd', 'members'],
          [ASO(False), intarr([-8, -6, -4, -2, 0, 2, 4, 6, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['odd', 'members'],
          [ASO(True), intarr([-9, -7, -3, -1, 1, 3, 7])]
          )
      ]
    )
  );

  Test('GroupBy(v, IsNegative, "neg", "ints") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['neg', 'ints'],
          [ASO(False), intarr([0, 1, 2, 3, 4, 6, 7, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['neg', 'ints'],
          [ASO(True), intarr([-9, -8, -7, -6, -4, -3, -2, -1])]
          )
      ]
    )
  );

  Test('GroupBy(v, odd, "odd", "values") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['odd', 'values'],
          [ASO(False), intarr([-8, -6, -4, -2, 0, 2, 4, 6, 8])]
          ),
        TAlgosimStructure.CreateWithValue(
          ['odd', 'values'],
          [ASO(True), intarr([-9, -7, -3, -1, 1, 3, 7])]
          )
      ]
    )
  );

  Test('p ≔ (name, age, sex) ↦ struct("name": name, "age": age, "sex": sex);', null);
  Test('L ≔ ''(p("Ian", 89, "male"), p("Andrew", 32, "male"), p("Sarah", 28, "female"), p("John", 53, "male"), p("David", 45, "male"), p("Elizabeth", 80, "female"), p("Eric", 16, "male"), p("Sophie", 22, "female"), '+'p("Bill", 75, "male"), p("Michael", 79, "male"), p("Richard", 85, "male"))',
    asoarrex(
      [
        SPerson('Ian', 89, 'male'),
        SPerson('Andrew', 32, 'male'),
        SPerson('Sarah', 28, 'female'),
        SPerson('John', 53, 'male'),
        SPerson('David', 45, 'male'),
        SPerson('Elizabeth', 80, 'female'),
        SPerson('Eric', 16, 'male'),
        SPerson('Sophie', 22, 'female'),
        SPerson('Bill', 75, 'male'),
        SPerson('Michael', 79, 'male'),
        SPerson('Richard', 85, 'male')
      ]
    )
  );

  Test('GroupBy(L, member("sex")) @ sort',
    asoarrex(
      [
        asoarrex(
          [
            SPerson('Elizabeth', 80, 'female'),
            SPerson('Sarah', 28, 'female'),
            SPerson('Sophie', 22, 'female')
          ]
        ),
        asoarrex(
          [
            SPerson('Andrew', 32, 'male'),
            SPerson('Bill', 75, 'male'),
            SPerson('David', 45, 'male'),
            SPerson('Eric', 16, 'male'),
            SPerson('Ian', 89, 'male'),
            SPerson('John', 53, 'male'),
            SPerson('Michael', 79, 'male'),
            SPerson('Richard', 85, 'male')
          ]
        )
      ]
    )
  );

  Test('GroupBy(L, member("sex"), "gender") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['gender', 'members'],
          [
            ASO('female'),
            asoarrex(
              [
                SPerson('Elizabeth', 80, 'female'),
                SPerson('Sarah', 28, 'female'),
                SPerson('Sophie', 22, 'female')
              ]
            )
          ]
        ),
        TAlgosimStructure.CreateWithValue(
          ['gender', 'members'],
          [
            ASO('male'),
            asoarrex(
              [
                SPerson('Andrew', 32, 'male'),
                SPerson('Bill', 75, 'male'),
                SPerson('David', 45, 'male'),
                SPerson('Eric', 16, 'male'),
                SPerson('Ian', 89, 'male'),
                SPerson('John', 53, 'male'),
                SPerson('Michael', 79, 'male'),
                SPerson('Richard', 85, 'male')
              ]
            )
          ]
        )
      ]
    )
  );

  Test('GroupBy(L, member("sex"), "gender", "students") @ (x↦ApplyIf(x, (y↦type(y)="array"), sort))',
    asoarrex(
      [
        TAlgosimStructure.CreateWithValue(
          ['gender', 'students'],
          [
            ASO('female'),
            asoarrex(
              [
                SPerson('Elizabeth', 80, 'female'),
                SPerson('Sarah', 28, 'female'),
                SPerson('Sophie', 22, 'female')
              ]
            )
          ]
        ),
        TAlgosimStructure.CreateWithValue(
          ['gender', 'students'],
          [
            ASO('male'),
            asoarrex(
              [
                SPerson('Andrew', 32, 'male'),
                SPerson('Bill', 75, 'male'),
                SPerson('David', 45, 'male'),
                SPerson('Eric', 16, 'male'),
                SPerson('Ian', 89, 'male'),
                SPerson('John', 53, 'male'),
                SPerson('Michael', 79, 'male'),
                SPerson('Richard', 85, 'male')
              ]
            )
          ]
        )
      ]
    )
  );

  Test('GroupBy(L, (s ↦ ⌊s.age/10⌋)) @ sort',
    asoarrex(
      [
        asoarrex(
          [
            SPerson('Eric', 16, 'male')
          ]
        ),
        asoarrex(
          [
            SPerson('Sarah', 28, 'female'),
            SPerson('Sophie', 22, 'female')
          ]
        ),
        asoarrex(
          [
            SPerson('Andrew', 32, 'male')
          ]
        ),
        asoarrex(
          [
            SPerson('David', 45, 'male')
          ]
        ),
        asoarrex(
          [
            SPerson('John', 53, 'male')
          ]
        ),
        asoarrex(
          [
            SPerson('Bill', 75, 'male'),
            SPerson('Michael', 79, 'male')
          ]
        ),
        asoarrex(
          [
            SPerson('Elizabeth', 80, 'female'),
            SPerson('Ian', 89, 'male'),
            SPerson('Richard', 85, 'male')
          ]
        )
      ]
    )
  );

  Test('delete(v); delete(L); delete(S); delete(p)', success);


  //
  // entrywise
  //

  Test('L ≔ ''(❨❨6, 1❩, ❨2, 4❩❩, ❨❨5, 7❩, ❨6, 3❩❩, ❨❨4, 1❩, ❨5, 6❩❩);', null);

  Eps; Test('entrywise(L, sum)', 2, [15, 9, 13, 13]);   // = ∑(L)
  Eps; Test('entrywise(L, product)', 2, [120, 7, 60, 72]);  // ≠ ∏(L)
  Eps; Test('entrywise(L, mean)', 2, [5, 3, 13/3, 13/3]); // = mean(L)
  Eps; Test('entrywise(L, average)', 2, [5, 3, 13/3, 13/3]); // = average(L)
  Eps; Test('entrywise(L, ArithmeticMean)', 2, [5, 3, 13/3, 13/3]); // = ArithmeticMean(L)
  Eps; Test('entrywise(L, GeometricMean)', 2, [4.9324241486609402, 1.9129311827723891, 3.9148676411688636, 4.16016764610380823]); // GeometricMean(L) not defined
  Eps; Test('entrywise(L, HarmonicMean)', 2, [4.86486486486486486, 1.4, 3.46153846153846154, 4]); // HarmonicMean(L) not defined

  Test('entrywise(''(❨❨6, 1❩, ❨2, 4❩❩, ❨❨5, 7❩, ❨6, 3❩❩, ❨❨4, 1, 8❩, ❨5, 6, 4❩❩), sum)',
    failure, 'Cannot compute entrywise images in a list of vectors or matrices of different sizes.');

  Test('L ≔ ''(❨6, 2, 1, 4❩, ❨7, 4, 2, 3❩, ❨1, 2, 1, 3❩);', null);

  Eps; Test('entrywise(L, sum)', [14, 8, 4, 10]);   // = ∑(L)
  Eps; Test('entrywise(L, product)', [42, 16, 2, 36]);  // ∏(L) not defined
  Eps; Test('entrywise(L, mean)', [14/3, 8/3, 4/3, 10/3]); // = mean(L)
  Eps; Test('entrywise(L, average)', [14/3, 8/3, 4/3, 10/3]); // = average(L)
  Eps; Test('entrywise(L, ArithmeticMean)', [14/3, 8/3, 4/3, 10/3]); // = ArithmeticMean(L)
  Eps; Test('entrywise(L, GeometricMean)', [3.47602664488644979, 2.51984209978974633, 1.25992104989487316, 3.30192724889462668]); // GeometricMean(L) not defined
  Eps; Test('entrywise(L, HarmonicMean)', [2.29090909090909091, 2.4, 1.2, 3.27272727272727273]); // HarmonicMean(L) not defined

  Test('entrywise(''(❨6, 2, 1, 4❩, ❨7, 4, 2, 3❩, ❨1, 2, 1, 3, 5❩), sum)',
    failure, 'Cannot compute entrywise images in a list of vectors or matrices of different sizes.');

  Test('entrywise(''(❨1, 2, 3❩, ❨❨5, 2, 7❩❩*), sum)', 1, [6, 4, 10]);
  Test('entrywise(''(❨1, 2, 3❩, ❨❨5, 2, 7❩❩), sum)',
    failure, 'Cannot compute entrywise images in a list of vectors or matrices of different sizes.');
  Test('entrywise(''(❨1, 2, 3❩, ❨❨5, 2, 7, 4❩❩*), sum)',
    failure, 'Cannot compute entrywise images in a list of vectors or matrices of different sizes.');

  Test('entrywise(''(❨1, 2, 3❩, ❨❨5, 2, 7❩❩*, 10), sum)',
    failure, 'The array must contain only objects of type numeric array, but it contains an object of type integer.');

  Test('entrywise(''(❨1, 2, 3❩), sum)', [1, 2, 3]);
  Test('entrywise(''(), sum)',
    failure, 'Cannot compute entrywise images in an empty list of vectors or matrices.'); // indeed a zero matrix or vector, but of what size/dimension?

  Test('entrywise(''(❨1, 2, 3❩), sum, product)', failure, 'Too many arguments.');

  Test('delete(L)', success);


  //
  // Accumulation
  //

  // accumulate

  Test('v ≔ ❨5, 1, 3, −2, 7, 5❩', [5, 1, 3, -2, 7, 5]);

  Test('accumulate(v, 0, add)', 19.0);
  Test('accumulate(v, 1, multiply)', -1050.0);
  Test('accumulate(v, 0, subtract)', -19.0);
  Test('accumulate(v, 1, lcm)', 210);
  Test('accumulate(v, -1000, max)', 7.0);
  Test('accumulate(v, +1000, min)', -2.0);

  Test('accumulate(v, "test", add)', 'test513−275');
  Test('accumulate(v, true, add)', failure, 'Invalid function arguments.');

  Test('accumulate(DebugObject("empty vector"), 0, add)', 0);
  Test('accumulate(DebugObject("empty vector"), 7, add)', 7);
  Test('accumulate(DebugObject("empty vector"), 7.0, add)', 7.0);

  Test('v ≔ ❨3, i, 5, 1, 2, 2⋅i❩', [3, ImaginaryUnit, 5, 1, 2, 2*ImaginaryUnit]);

  Test('accumulate(v, 0, add)', 11 + 3*ImaginaryUnit);
  Test('accumulate(v, 1, multiply)', TASC(-60));
  Test('accumulate(v, 0, subtract)', -11 - 3*ImaginaryUnit);
  Test('accumulate(v, 1, lcm)', failure, 'An object of type integer was expected as argument 2, but an object of type complex number was given.');

  Test('accumulate(ComplexVector(DebugObject("empty vector")), 0, add)', 0);
  Test('accumulate(ComplexVector(DebugObject("empty vector")), 51, add)', 51);

  Test('A ≔ ❨❨3, 1, 5, −2, 7❩, ❨3, 2, 2, 5, 1❩❩', 5, [3, 1, 5, -2, 7, 3, 2, 2, 5, 1]);

  Test('accumulate(A, 0, add)', 27.0);
  Test('accumulate(A, 1, multiply)', -12600.0);
  Test('accumulate(A, 0, subtract)', -27.0);
  Test('accumulate(A, 1, lcm)', 210);
  Test('accumulate(A, -1000, max)', 7.0);
  Test('accumulate(A, +1000, min)', -2.0);

  Test('accumulate(DebugObject("empty matrix"), 0, add)', 0);
  Test('accumulate(DebugObject("empty matrix"), 1.2, add)', 1.2);

  Test('A ≔ ❨❨4, 3, 2, i, 1❩, ❨−2, 1, i, 5, 3❩❩', 5, [4, 3, 2, ImaginaryUnit, 1, -2, 1, ImaginaryUnit, 5, 3]);

                    {sic!}
  Test('accumulate(A, 3, add)', 20 + 2*ImaginaryUnit);
  Test('accumulate(A, 1, multiply)', TASC(720));
  Test('accumulate(A, 0, subtract)', -17 - 2*ImaginaryUnit);
  Test('accumulate(A, 1, lcm)', failure, 'An object of type integer was expected as argument 2, but an object of type complex number was given.');

  Test('accumulate(ComplexMatrix(DebugObject("empty matrix")), 0, add)', 0);
  Test('accumulate(ComplexMatrix(DebugObject("empty matrix")), i, add)', ImaginaryUnit);

  Test('L ≔ ''(5, 3, 2, −11, 7, 5, 6)', intarr([5, 3, 2, -11, 7, 5, 6]));

  Test('accumulate(L, 0, add)', 17);
  Test('accumulate(L, 1, multiply)', -69300);
  Test('accumulate(L, 0, subtract)', -17);
  Test('accumulate(L, 1, lcm)', 2310);
  Test('accumulate(L, -1000, max)', 7);
  Test('accumulate(L, +1000, min)', -11);

  Test('accumulate(L, 0.0, add)', 17.0);
  Test('accumulate(L, 3.0, add)', 20.0);
  Test('accumulate(L, 3, add)', 20);

  Test('accumulate(L, 0, (a, b) ↦ (2⋅a + 3⋅b)/5 )', rat(15972, 3125));

  Test('L[3] ≔ 2.0', 2.0);
  Test('accumulate(L, 0, add)', 17.0);
  Test('accumulate(L, 1, multiply)', -69300.0);
  Test('accumulate(L, 0, subtract)', -17.0);
  Test('accumulate(L, 1, lcm)', 2310);
  Test('accumulate(L, -1000, max)', 7.0);
  Test('accumulate(L, +1000, min)', -11.0);

  Test('L ≔ ''("alpha", "beta", "gamma", "delta")', strarr(['alpha', 'beta', 'gamma', 'delta']));

  Test('accumulate(L, "", add)', 'alphabetagammadelta');
  Test('accumulate(L, "start", (s, t) ↦ s + ", " + t)', 'start, alpha, beta, gamma, delta');

  Test('accumulate(''(), 0, add)', 0);
  Test('accumulate(''(), 10, add)', 10);
  Test('accumulate(''(), "cat", add)', 'cat');
  Test('accumulate(''(), true, add)', True);

  Test('accumulate("alpha", "x", (s, t) ↦ s + ", " + t)', 'x, a, l, p, h, a');

  Test('accumulate({3, 20, 4, 5, −1, −5, 2}, 0, add)', 28);
  Test('accumulate({3, 20, 4, 5, −1, −5, 2}, -100, max)', 20);
  Test('accumulate({3, 20, 4, 5, −1, −5, 2}, +100, min)', -5);
  Test('accumulate({}, 0, add)', 0);
  Test('accumulate({6}, "cat", add)', 'cat6');

  // AccumulateSteps and AccumulateList

  Test('v ≔ ❨5, 1, 3, −2, 7, 5❩', [5, 1, 3, -2, 7, 5]);
  Test('AccumulateSteps(v, 0, add)', [5, 6, 9, 7, 14, 19]);
  Test('AccumulateList(v, 0, add)', asoarr([5.0, 6.0, 9.0, 7.0, 14.0, 19.0]));
  Test('AccumulateSteps(v, 1, multiply)', [5, 5, 15, -30, -210, -1050]);
  Test('AccumulateList(v, 1, multiply)', asoarr([5.0, 5.0, 15.0, -30.0, -210.0, -1050.0]));
  Test('AccumulateSteps(v, -100, max)', [5, 5, 5, 5, 7, 7]);
  Test('AccumulateList(v, -100, max)', asoarr([5.0, 5.0, 5.0, 5.0, 7.0, 7.0]));
  Test('AccumulateSteps(v, 100, min)', [5, 1, 1, -2, -2, -2]);
  Test('AccumulateList(v, 100, min)', asoarr([5.0, 1.0, 1.0, -2.0, -2.0, -2.0]));

  Test('v ≔ ❨8, i, 4, i, 3, −5, 2❩', [8, ImaginaryUnit, 4, ImaginaryUnit, 3, -5, 2]);
  Test('AccumulateSteps(v, 0, add)', [8, 8 + ImaginaryUnit, 12 + ImaginaryUnit, 12 + 2*ImaginaryUnit, 15 + 2*ImaginaryUnit, 10 + 2*ImaginaryUnit, 12 + 2*ImaginaryUnit]);
  Test('AccumulateList(v, 0, add)', asoarrex([ASO(8.0), ASO(8 + ImaginaryUnit), ASO(12 + ImaginaryUnit), ASO(12 + 2*ImaginaryUnit), ASO(15 + 2*ImaginaryUnit), ASO(10 + 2*ImaginaryUnit), ASO(12 + 2*ImaginaryUnit)]));
  Test('AccumulateSteps(v, 1, multiply)', [8, 8*ImaginaryUnit, 32*ImaginaryUnit, -32, -96, 480, 960]);
  Test('AccumulateList(v, 1, multiply)', asoarrex([ASO(8.0), ASO(8*ImaginaryUnit), ASO(32*ImaginaryUnit), ASO(-32.0), ASO(-96.0), ASO(480.0), ASO(960.0)]));

  Test('A ≔ ❨❨6, −2, 3, 4❩, ❨1, 10, 2, 1❩❩', 4, [6, -2, 3, 4, 1, 10, 2, 1]);
  Test('AccumulateSteps(A, 0, add)', 4, [6, 4, 7, 11, 12, 22, 24, 25]);
  Test('AccumulateList(A, 0, add)', asoarr([6.0, 4.0, 7.0, 11.0, 12.0, 22.0, 24.0, 25.0]));
  Test('AccumulateSteps(A, 1, multiply)', 4, [6, -12, -36, -144, -144, -1440, -2880, -2880]);
  Test('AccumulateList(A, 1, multiply)', asoarr([6.0, -12.0, -36.0, -144.0, -144.0, -1440.0, -2880.0, -2880.0]));
  Test('AccumulateSteps(A, -100, max)', 4, [6, 6, 6, 6, 6, 10, 10, 10]);
  Test('AccumulateList(A, -100, max)', asoarr([6.0, 6.0, 6.0, 6.0, 6.0, 10.0, 10.0, 10.0]));
  Test('AccumulateSteps(A, 100, min)', 4, [6, -2, -2, -2, -2, -2, -2, -2]);
  Test('AccumulateList(A, 100, min)', asoarr([6.0, -2.0, -2.0, -2.0, -2.0, -2.0, -2.0, -2.0]));

  Test('A ≔ ❨❨3, i, 4, 2❩, ❨−1, i, i, 2❩❩', 4, [3, ImaginaryUnit, 4, 2, -1, ImaginaryUnit, ImaginaryUnit, 2]);
  Test('AccumulateSteps(A, 0, add)', 4, [3, 3 + ImaginaryUnit, 7 + ImaginaryUnit, 9 + ImaginaryUnit, 8 + ImaginaryUnit, 8 + 2*ImaginaryUnit, 8 + 3*ImaginaryUnit, 10 + 3*ImaginaryUnit]);
  Test('AccumulateList(A, 0, add)', asoarrex([ASO(3.0), ASO(3 + ImaginaryUnit), ASO(7 + ImaginaryUnit), ASO(9 + ImaginaryUnit), ASO(8 + ImaginaryUnit), ASO(8 + 2*ImaginaryUnit), ASO(8 + 3*ImaginaryUnit), ASO(10 + 3*ImaginaryUnit)]));
  Test('AccumulateSteps(A, 1, multiply)', 4, [3, 3*ImaginaryUnit, 12*ImaginaryUnit, 24*ImaginaryUnit, -24*ImaginaryUnit, 24, 24*ImaginaryUnit, 48*ImaginaryUnit]);
  Test('AccumulateList(A, 1, multiply)', asoarrex([ASO(3.0), ASO(3*ImaginaryUnit), ASO(12*ImaginaryUnit), ASO(24*ImaginaryUnit), ASO(-24*ImaginaryUnit), ASO(24.0), ASO(24*ImaginaryUnit), ASO(48*ImaginaryUnit)]));

  Test('L ≔ ''(5, 1/2, 0.1, i, 0)', asoarrex([ASOInt(5), ASO(rat(1, 2)), ASO(0.1), ASO(ImaginaryUnit), ASOInt(0)]));
  Test('AccumulateSteps(L, 0, add)', asoarrex([ASOInt(5), ASO(rat(11, 2)), ASO(5.6), ASO(5.6 + ImaginaryUnit), ASO(5.6 + ImaginaryUnit)]));
  Test('AccumulateList(L, 0, add)', asoarrex([ASOInt(5), ASO(rat(11, 2)), ASO(5.6), ASO(5.6 + ImaginaryUnit), ASO(5.6 + ImaginaryUnit)]));
  Test('AccumulateSteps(L, 1, multiply)', asoarrex([ASOInt(5), ASO(rat(5, 2)), ASO(0.25), ASO(0.25*ImaginaryUnit), ASO(0.0*ImaginaryUnit)]));
  Test('AccumulateList(L, 1, multiply)', asoarrex([ASOInt(5), ASO(rat(5, 2)), ASO(0.25), ASO(0.25*ImaginaryUnit), ASO(0.0*ImaginaryUnit)]));

  Test('AccumulateSteps(DebugObject("empty vector"), 0, add)', TAlgosimRealVector.Create);
  Test('AccumulateSteps(ComplexVector(DebugObject("empty vector")), 0, add)', TAlgosimComplexVector.Create);
  Test('AccumulateSteps(DebugObject("empty matrix"), 0, add)', failure, 'A matrix must have size at least 1×1.');
  Test('AccumulateSteps(ComplexMatrix(DebugObject("empty matrix")), 0, add)', failure, 'A matrix must have size at least 1×1.');
  Test('AccumulateSteps(''(), 0, add)', TAlgosimArray.Create);

  Test('AccumulateList(DebugObject("empty vector"), 0, add)', TAlgosimArray.Create);
  Test('AccumulateList(ComplexVector(DebugObject("empty vector")), 0, add)', TAlgosimArray.Create);
  Test('AccumulateList(DebugObject("empty matrix"), 0, add)', TAlgosimArray.Create);
  Test('AccumulateList(ComplexMatrix(DebugObject("empty matrix")), 0, add)', TAlgosimArray.Create);
  Test('AccumulateList(''(), 0, add)', TAlgosimArray.Create);

  Test('AccumulateSteps(''("alpha", "beta"), 0, add)', failure, 'Invalid function arguments.');
  Test('AccumulateList(''("alpha", "beta"), 0, add)', failure, 'Invalid function arguments.');

  Test('AccumulateSteps({3, 20, 4, 5, −1, −5, 2}, 0, add)', failure, 'Object type "set" is not a suitable container.');
  Test('#AccumulateList({3, 20, 4, 5, −1, −5, 2}, 0, add)', 7); // few invariants to choose from...

  Test('delete(v); delete(A); delete(L)', success);


  //
  // Inserting, appending, moving, and removing
  //

  // Appending a single element

  Test('append(❨5, 3, 2, 6❩, 10)', [5, 3, 2, 6, 10]);
  Test('append(❨5, 3, 2, 6❩, ❨7, 8, 9❩)', failure, 'Object isn''t a real number.');
  Test('append(❨5, 3, 2, 6❩, 10 + i)', failure, 'Object isn''t a real number.');
  Test('append(ComplexVector(❨5, 3, 2, 6❩), 10 + i)', [5, 3, 2, 6, 10 + ImaginaryUnit]);
  Test('append(❨5, 3, 2, 6❩, "nargle")', failure, 'Object isn''t a real number.');
  Test('append(DebugObject("empty vector"), 3.1415)', [3.1415]);

  Test('append(❨3, i, 1 + i, 5❩, 10)', [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 10]);
  Test('append(❨3, i, 1 + i, 5❩, 10 + i)', [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 10 + ImaginaryUnit]);
  Test('append(❨3, i, 1 + i, 5❩, "nargle")', failure, 'Object isn''t a complex number.');
  Test('append(ComplexVector(DebugObject("empty vector")), 3.1415)', [TASC(3.1415)]);

  Test('append(❨❨4, 1❩, ❨0, 5❩❩, 5)', failure, 'Object type "real matrix" is not a suitable container.');
  Test('append(❨❨4, i❩, ❨0, 5❩❩, 5)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('append(''(3, 1, 4, "one", "five"), "Ⅸ")', asoarr([3, 1, 4, 'one', 'five', 'Ⅸ']));
  Test('append(''(), "a start")', asoarr(['a start']));

  Test('append({1, 2, 3}, 10)', failure, 'Object type "set" is not a suitable container.');

  Test('append("test", 2)', failure, 'An object of type "integer" cannot be converted to a character.');
  Test('append("test", "2")', 'test2');
  Test('append("test", "ing")', failure, 'A string of length 3 cannot be converted to a character.');

  Test('append(''(3, 1, 4, "one", "five"), ''(9, 2))', asoarrex([ASOInt(3), ASOInt(1), ASOInt(4), ASO('one'), ASO('five'), intarr([9, 2])]));

  // Extension

  Test('ExtendWith(❨5, 3, 2, 6❩, 10)', [5, 3, 2, 6, 10]);
  Test('ExtendWith(❨5, 3, 2, 6❩, ❨7, 8, 9❩)', [5, 3, 2, 6, 7, 8, 9]);
  Test('ExtendWith(❨5, 3, 2, 6❩, 10 + i)', failure, 'Object isn''t a real number.');
  Test('ExtendWith(ComplexVector(❨5, 3, 2, 6❩), 10 + i)', [5, 3, 2, 6, 10 + ImaginaryUnit]);
  Test('ExtendWith(❨5, 3, 2, 6❩, "nargle")', failure, 'Cannot convert string "nargle" to a real number.');
  Test('ExtendWith(DebugObject("empty vector"), 3.1415)', [3.1415]);

  Test('ExtendWith(❨3, i, 1 + i, 5❩, 10)', [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 10]);
  Test('ExtendWith(❨3, i, 1 + i, 5❩, 10 + i)', [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 10 + ImaginaryUnit]);
  Test('ExtendWith(❨3, i, 1 + i, 5❩, "nargle")', failure, 'Cannot convert string "nargle" to a complex number.');
  Test('ExtendWith(ComplexVector(DebugObject("empty vector")), 3.1415)', [TASC(3.1415)]);

  Test('ExtendWith(❨❨4, 1❩, ❨0, 5❩❩, 5)', failure, 'Object type "real matrix" is not a suitable container.');
  Test('ExtendWith(❨❨4, i❩, ❨0, 5❩❩, 5)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('ExtendWith(''(3, 1, 4, "one", "five"), ''("Ⅸ"))', asoarr([3, 1, 4, 'one', 'five', 'Ⅸ']));
  Test('ExtendWith(''(3, 1, 4, "one", "five"), "Ⅸ")', failure, 'Object type "string" is not a suitable container.');
  Test('ExtendWith(''(), ''("a start"))', asoarr(['a start']));

  Test('ExtendWith({1, 2, 3}, 10)', failure, 'Object type "set" is not a suitable container.');

  Test('ExtendWith("test", 2)', 'test2');
  Test('ExtendWith("test", "2")', 'test2');
  Test('ExtendWith("test", "ing")', 'testing');

  Test('ExtendWith(''(3, 1, 4, "one", "five"), ''(9, 2))', asoarrex([ASOInt(3), ASOInt(1), ASOInt(4), ASO('one'), ASO('five'), ASOInt(9), ASOInt(2)]));

  // Insertion

  Test('insert(❨5, 3, 0, 1, 2❩, 1, 100)', [100, 5, 3, 0, 1, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, 2, 100)', [5, 100, 3, 0, 1, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, 3, 100)', [5, 3, 100, 0, 1, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, 4, 100)', [5, 3, 0, 100, 1, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, 5, 100)', [5, 3, 0, 1, 100, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, 6, 100)', [5, 3, 0, 1, 2, 100]);
  Test('insert(❨5, 3, 0, 1, 2❩, 7, 100)', failure, 'Index 7 out of bounds.');

  Test('insert(❨5, 3, 0, 1, 2❩, 0, 100)', failure, 'Index 0 out of bounds.');

  Test('insert(❨5, 3, 0, 1, 2❩, -7, 100)', failure, 'Index -7 out of bounds.');
  Test('insert(❨5, 3, 0, 1, 2❩, -6, 100)', [100, 5, 3, 0, 1, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, -5, 100)', [5, 100, 3, 0, 1, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, -4, 100)', [5, 3, 100, 0, 1, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, -3, 100)', [5, 3, 0, 100, 1, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, -2, 100)', [5, 3, 0, 1, 100, 2]);
  Test('insert(❨5, 3, 0, 1, 2❩, -1, 100)', [5, 3, 0, 1, 2, 100]);

  Test('insert(❨5, 3, 0, 1, 2❩, 3, 1 + i)', failure, 'Object isn''t a real number.');
  Test('insert(❨5, 3, 0, 1, 2❩, 3, "nargle")', failure, 'Object isn''t a real number.');

  Test('insert(DebugObject("empty vector"), 1, 123)', [123]);
  Test('insert(DebugObject("empty vector"), -1, 123)', [123]);
  Test('insert(DebugObject("empty vector"), 0, 123)', failure, 'Index 0 out of bounds.');
  Test('insert(DebugObject("empty vector"), 2, 123)', failure, 'Index 2 out of bounds.');
  Test('insert(DebugObject("empty vector"), -2, 123)', failure, 'Index -2 out of bounds.');

  Test('insert(❨5, 3, 0, i, 2❩, 1, 100)', [100, 5, 3, 0, ImaginaryUnit, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, 2, 100)', [5, 100, 3, 0, ImaginaryUnit, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, 3, 100)', [5, 3, 100, 0, ImaginaryUnit, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, 4, 100)', [5, 3, 0, 100, ImaginaryUnit, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, 5, 100)', [5, 3, 0, ImaginaryUnit, 100, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, 6, 100)', [5, 3, 0, ImaginaryUnit, 2, 100]);
  Test('insert(❨5, 3, 0, i, 2❩, 7, 100)', failure, 'Index 7 out of bounds.');

  Test('insert(❨5, 3, 0, i, 2❩, 0, 100)', failure, 'Index 0 out of bounds.');

  Test('insert(❨5, 3, 0, i, 2❩, -7, 100)', failure, 'Index -7 out of bounds.');
  Test('insert(❨5, 3, 0, i, 2❩, -6, 100)', [100, 5, 3, 0, ImaginaryUnit, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, -5, 100)', [5, 100, 3, 0, ImaginaryUnit, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, -4, 100)', [5, 3, 100, 0, ImaginaryUnit, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, -3, 100)', [5, 3, 0, 100, ImaginaryUnit, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, -2, 100)', [5, 3, 0, ImaginaryUnit, 100, 2]);
  Test('insert(❨5, 3, 0, i, 2❩, -1, 100)', [5, 3, 0, ImaginaryUnit, 2, 100]);

  Test('insert(❨5, 3, 0, i, 2❩, 3, "nargle")', failure, 'Object isn''t a complex number.');

  Test('insert(ComplexVector(DebugObject("empty vector")), 1, 123)', [TASC(123)]);
  Test('insert(ComplexVector(DebugObject("empty vector")), -1, 123)', [TASC(123)]);
  Test('insert(ComplexVector(DebugObject("empty vector")), 0, 123)', failure, 'Index 0 out of bounds.');
  Test('insert(ComplexVector(DebugObject("empty vector")), 2, 123)', failure, 'Index 2 out of bounds.');
  Test('insert(ComplexVector(DebugObject("empty vector")), -2, 123)', failure, 'Index -2 out of bounds.');

  Test('insert(❨❨5, 3❩, ❨4, −1.0000002E10❩❩, 2, 12.34)', failure, 'Object type "real matrix" is not a suitable container.');
  Test('insert(❨❨5, 3/i❩, ❨4, −1.0000002E10❩❩, 2, 12.34)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('s ≔ "Hello World"', 'Hello World');
  Test('insert(s, 6, ",")', 'Hello, World');
  Test('insert(s, -1, "!")', 'Hello World!');
  Test('insert(insert(s, 6, ","), -1, "!")', 'Hello, World!');
  Test('insert(insert(s, 6, ","), #s + 2, "!")', 'Hello, World!');

  Test('insert(s, 50, ",")', failure, 'Index 50 out of bounds.');

  Test('insert(s, 6, "comma")', 'Hellocomma World');

  Test('L ≔ insert(''(3, 1, 4, "one", "nine", "two"), 5, "five")', asoarr([3, 1, 4, 'one', 'five', 'nine', 'two']));
  Test('L ≔ insert(L, -1, 6)', asoarr([3, 1, 4, 'one', 'five', 'nine', 'two', 6]));
  Test('insert(L, -1, 5)', asoarr([3, 1, 4, 'one', 'five', 'nine', 'two', 6, 5]));
  Test('insert(''(), 1, "a")', asoarr(['a']));
  Test('insert(''("a", "b"), 2, ''(1, 2))', asoarrex([ASO('a'), intarr([1, 2]), ASO('b')]));

  // Deletion

  Test('remove(❨3, 1, 0, 7, 4, 8❩, 1)', [1, 0, 7, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, 2)', [3, 0, 7, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, 3)', [3, 1, 7, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, 4)', [3, 1, 0, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, 5)', [3, 1, 0, 7, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, 6)', [3, 1, 0, 7, 4]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, 7)', [3, 1, 0, 7, 4, 8]);

  Test('remove(❨3, 1, 0, 7, 4, 8❩, 00)', [3, 1, 0, 7, 4, 8]);

  Test('remove(❨3, 1, 0, 7, 4, 8❩, -1)', [3, 1, 0, 7, 4]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, -2)', [3, 1, 0, 7, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, -3)', [3, 1, 0, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, -4)', [3, 1, 7, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, -5)', [3, 0, 7, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, -6)', [1, 0, 7, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, -7)', [3, 1, 0, 7, 4, 8]);

  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(1, 4))', [1, 0, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(2, 3))', [3, 7, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(4, 6))', [3, 1, 0, 4]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(5, 6))', [3, 1, 0, 7]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(5, -1))', [3, 1, 0, 7]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(-2, -1))', [3, 1, 0, 7]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(-2, -1, 7))', [3, 1, 0, 7]);

  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(1, 2, 3, 4, 6))', [4]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(1, 2, 3, 4, 6, 5))', TAlgosimRealVector.Create);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(1, 2, 2, 5, 1, 10, -1, -2, -999, 3, 4, 6, 5))', TAlgosimRealVector.Create);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(1, 2, 2, 5, 1, 10, -1, -2, -999, 4, 6, 5))', [0]);

  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(1, 4))', [4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(2, 100))', [3]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(-100, 100))', TAlgosimRealVector.Create);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(5, 100))', [3, 1, 0, 7]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(-3, -1))', [3, 1, 0]);

  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(range(1, 2), range(-2, -1)))', [0, 7]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, ''(range(1, 2), -2))', [0, 7, 8]);

  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(4, 1))', [4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(100, 2))', [3]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(100, -100))', TAlgosimRealVector.Create);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(100, 5))', [3, 1, 0, 7]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(-1, -3))', [3, 1, 0]);

  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(1, -1))', TAlgosimRealVector.Create);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(1, -1, 2))', [1, 7, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(1, -1, 3))', [1, 0, 4, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(1, -1, 4))', [1, 0, 7, 8]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(1, -1, 5))', [1, 0, 7, 4]);

  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(-1, 1))', TAlgosimRealVector.Create);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(-1, 1, 2))', [3, 0, 4]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(-1, 1, 3))', [3, 1, 7, 4]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(-1, 1, 4))', [3, 0, 7, 4]);
  Test('remove(❨3, 1, 0, 7, 4, 8❩, range(-1, 1, 5))', [1, 0, 7, 4]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, 1)', [1, ImaginaryUnit, 7, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, 2)', [3, ImaginaryUnit, 7, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, 3)', [3, TASC(1), 7, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, 4)', [3, 1, ImaginaryUnit, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, 5)', [3, 1, ImaginaryUnit, 7, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, 6)', [3, 1, ImaginaryUnit, 7, 4]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, 7)', [3, 1, ImaginaryUnit, 7, 4, 8]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, 00)', [3, 1, ImaginaryUnit, 7, 4, 8]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, -1)', [3, 1, ImaginaryUnit, 7, 4]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, -2)', [3, 1, ImaginaryUnit, 7, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, -3)', [3, 1, ImaginaryUnit, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, -4)', [3, TASC(1), 7, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, -5)', [3, ImaginaryUnit, 7, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, -6)', [1, ImaginaryUnit, 7, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, -7)', [3, 1, ImaginaryUnit, 7, 4, 8]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(1, 4))', [1, ImaginaryUnit, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(2, 3))', [3, TASC(7), 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(4, 6))', [3, 1, ImaginaryUnit, 4]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(5, 6))', [3, 1, ImaginaryUnit, 7]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(5, -1))', [3, 1, ImaginaryUnit, 7]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(-2, -1))', [3, 1, ImaginaryUnit, 7]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(-2, -1, 7))', [3, 1, ImaginaryUnit, 7]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(1, 2, 3, 4, 6))', [TASC(4)]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(1, 2, 3, 4, 6, 5))', TAlgosimComplexVector.Create);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(1, 2, 2, 5, 1, 10, -1, -2, -999, 3, 4, 6, 5))', TAlgosimComplexVector.Create);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(1, 2, 2, 5, 1, 10, -1, -2, -999, 4, 6, 5))', [ImaginaryUnit]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, range(1, 4))', [TASC(4), 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(2, 100))', [TASC(3)]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(-100, 100))', TAlgosimComplexVector.Create);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(5, 100))', [3, 1, ImaginaryUnit, 7]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(-3, -1))', [3, 1, ImaginaryUnit]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(range(1, 2), range(-2, -1)))', [ImaginaryUnit, 7]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, ''(range(1, 2), -2))', [ImaginaryUnit, 7, 8]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, range(4, 1))', [TASC(4), 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(100, 2))', [TASC(3)]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(100, -100))', TAlgosimComplexVector.Create);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(100, 5))', [3, 1, ImaginaryUnit, 7]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(-1, -3))', [3, 1, ImaginaryUnit]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, range(1, -1))', TAlgosimComplexVector.Create);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(1, -1, 2))', [TASC(1), 7, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(1, -1, 3))', [1, ImaginaryUnit, 4, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(1, -1, 4))', [1, ImaginaryUnit, 7, 8]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(1, -1, 5))', [1, ImaginaryUnit, 7, 4]);

  Test('remove(❨3, 1, i, 7, 4, 8❩, range(-1, 1))', TAlgosimComplexVector.Create);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(-1, 1, 2))', [3, ImaginaryUnit, 4]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(-1, 1, 3))', [TASC(3), 1, 7, 4]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(-1, 1, 4))', [3, ImaginaryUnit, 7, 4]);
  Test('remove(❨3, 1, i, 7, 4, 8❩, range(-1, 1, 5))', [1, ImaginaryUnit, 7, 4]);

  Test('remove(❨❨6, 3, 2, 5, 4❩, ❨7, 0, −2, 3, 5❩❩, 2)', failure, 'Object type "real matrix" is not a suitable container.');
  Test('remove(❨❨6, 3, 2, 5, 4❩, ❨7, i, −2, 3, 5❩❩, 2)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('remove("Technically, it’s a ferret.", 5)', 'Techically, it’s a ferret.');
  Test('remove("Technically, it’s a ferret.", 1)', 'echnically, it’s a ferret.');
  Test('remove("Technically, it’s a ferret.", -1)', 'Technically, it’s a ferret');
  Test('remove("Technically, it’s a ferret.", ''(-1, -200))', 'Technically, it’s a ferret');
  Test('remove("Technically, it’s a ferret.", ''(1, 10, -1))', 'echnicaly, it’s a ferret');
  Test('remove("Technically, it’s a ferret.", ''(1, 10, -1, -1, 500))', 'echnicaly, it’s a ferret');
  Test('remove("Technically, it’s a ferret.", range(1, 13))', 'it’s a ferret.');
  Test('remove("Technically, it’s a ferret.", ''(range(1, 13), −1))', 'it’s a ferret');
  Test('remove("Technically, it’s a ferret.", ''(range(1, 13), −1, 7))', 'it’s a ferret');
  Test('remove("Technically, it’s a ferret.", ''(range(1, 13), −1, 7, 700))', 'it’s a ferret');
  Test('remove("Technically, it’s a ferret.", ''(range(6, 1300), range(−3, −700)))', '');
  Test('remove("Technically, it’s a ferret.", range(1000, 1000000000))', 'Technically, it’s a ferret.');

  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 1)', asoarr([1, 4, 'one', 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 2)', asoarr([3, 4, 'one', 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 3)', asoarr([3, 1, 'one', 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 4)', asoarr([3, 1, 4, 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 5)', asoarr([3, 1, 4, 'one', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 6)', asoarr([3, 1, 4, 'one', 'five', 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 7)', asoarr([3, 1, 4, 'one', 'five', 9]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 8)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 9)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));

  Test('remove(''(3, 1, 4, "one", "five", 9, 2), -1)', asoarr([3, 1, 4, 'one', 'five', 9]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), -2)', asoarr([3, 1, 4, 'one', 'five', 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), -3)', asoarr([3, 1, 4, 'one', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), -4)', asoarr([3, 1, 4, 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), -5)', asoarr([3, 1, 'one', 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), -6)', asoarr([3, 4, 'one', 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), -7)', asoarr([1, 4, 'one', 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), -8)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), -9)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));

  Test('remove(''(3, 1, 4, "one", "five", 9, 2), 0)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));

  Test('remove(''(3, 1, 4, "one", "five", 9, 2), range(1, 3))', asoarr(['one', 'five', 9, 2]));
  Test('remove(''(3, 1, 4, "one", "five", 9, 2), ''(range(1, 3), range(-1, -2)))', asoarr(['one', 'five']));

  Test('remove(''(3, 1, 4, "one", "five", 9, 2), ''(range(1, 5), range(-1, -4)))', asoarr([]));

  Test('remove({1, 2, 3}, 2)', failure, 'Object type "set" is not a suitable container.');

  // Truncation

  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 0)', TAlgosimRealVector.Create);
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 1)', [0]);
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 2)', [0, 1]);
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 5)', [0, 1, 2, -1, 3]);
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 8)', [0, 1, 2, -1, 3, -2, 0, 1]);
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 10)', [0, 1, 2, -1, 3, -2, 0, 1, 2, 4]);
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 11)', [0, 1, 2, -1, 3, -2, 0, 1, 2, 4, -1]);
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 12)', [0, 1, 2, -1, 3, -2, 0, 1, 2, 4, -1]);
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 20)', [0, 1, 2, -1, 3, -2, 0, 1, 2, 4, -1]);
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 100000000)', [0, 1, 2, -1, 3, -2, 0, 1, 2, 4, -1]);

  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, -1)', failure, 'New length must be non-negative.');
  Test('truncate(❨0, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, -10)', failure, 'New length must be non-negative.');

  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 0)', TAlgosimComplexVector.Create);
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 1)', [ImaginaryUnit]);
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 2)', [ImaginaryUnit, 1]);
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 5)', [ImaginaryUnit, 1, 2, -1, 3]);
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 8)', [ImaginaryUnit, 1, 2, -1, 3, -2, 0, 1]);
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 10)', [ImaginaryUnit, 1, 2, -1, 3, -2, 0, 1, 2, 4]);
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 11)', [ImaginaryUnit, 1, 2, -1, 3, -2, 0, 1, 2, 4, -1]);
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 12)', [ImaginaryUnit, 1, 2, -1, 3, -2, 0, 1, 2, 4, -1]);
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 20)', [ImaginaryUnit, 1, 2, -1, 3, -2, 0, 1, 2, 4, -1]);
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, 100000000)', [ImaginaryUnit, 1, 2, -1, 3, -2, 0, 1, 2, 4, -1]);

  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, -1)', failure, 'New length must be non-negative.');
  Test('truncate(❨i, 1, 2, −1, 3, −2, 0, 1, 2, 4, −1❩, -10)', failure, 'New length must be non-negative.');

  Test('truncate(❨❨2, 3, 1❩, ❨0, 2, 1❩❩, 2)', failure, 'Object type "real matrix" is not a suitable container.');
  Test('truncate(❨❨2, 3, i❩, ❨0, 2, 1❩❩, 2)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('truncate("Hello, World!", 5)', 'Hello');
  Test('truncate("Hello, World!", 12)', 'Hello, World');
  Test('truncate("Hello, World!", 13)', 'Hello, World!');
  Test('truncate("Hello, World!", 14)', 'Hello, World!');
  Test('truncate("Hello, World!", 20)', 'Hello, World!');
  Test('truncate("Hello, World!", 100000000)', 'Hello, World!');
  Test('truncate("Hello, World!", 4)', 'Hell');
  Test('truncate("Hello, World!", 3)', 'Hel');
  Test('truncate("Hello, World!", 2)', 'He');
  Test('truncate("Hello, World!", 1)', 'H');
  Test('truncate("Hello, World!", 0)', '');

  Test('truncate("Hello, World!", -1)', failure, 'New length must be non-negative.');
  Test('truncate("Hello, World!", -10)', failure, 'New length must be non-negative.');

  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 3)', asoarr([3, 1, 4]));
  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 5)', asoarr([3, 1, 4, 'one', 'five']));
  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 7)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));
  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 8)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));
  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 10)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));
  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 100000000)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));

  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 3)', asoarr([3, 1, 4]));
  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 2)', asoarr([3, 1]));
  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 1)', asoarr([3]));
  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), 0)', asoarr([]));

  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), -1)', failure, 'New length must be non-negative.');
  Test('truncate(''(3, 1, 4, "one", "five", 9, 2), -123)', failure, 'New length must be non-negative.');

  Test('truncate({1, 2, 3}, 2)', failure, 'Object type "set" is not a suitable container.');

  Test('truncate(1024, 2)', failure, 'Object type "integer" is not a suitable container.');

  Test('truncate("Hello, World!")', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('truncate("Hello, World!", 2, 8)', failure, 'Too many arguments.');

  Test('truncate(struct("a": 687, "b": 713, "c": 394, "d": 426), 10000)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('a', ASOInt(687)),
        sm('b', ASOInt(713)),
        sm('c', ASOInt(394)),
        sm('d', ASOInt(426))
      ]
    )
  );

  Test('truncate(struct("a": 687, "b": 713, "c": 394, "d": 426), 5)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('a', ASOInt(687)),
        sm('b', ASOInt(713)),
        sm('c', ASOInt(394)),
        sm('d', ASOInt(426))
      ]
    )
  );

  Test('truncate(struct("a": 687, "b": 713, "c": 394, "d": 426), 4)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('a', ASOInt(687)),
        sm('b', ASOInt(713)),
        sm('c', ASOInt(394)),
        sm('d', ASOInt(426))
      ]
    )
  );

  Test('truncate(struct("a": 687, "b": 713, "c": 394, "d": 426), 3)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('a', ASOInt(687)),
        sm('b', ASOInt(713)),
        sm('c', ASOInt(394))
      ]
    )
  );

  Test('truncate(struct("a": 687, "b": 713, "c": 394, "d": 426), 2)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('a', ASOInt(687)),
        sm('b', ASOInt(713))
      ]
    )
  );

  Test('truncate(struct("a": 687, "b": 713, "c": 394, "d": 426), 1)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('a', ASOInt(687))
      ]
    )
  );

  Test('truncate(struct("a": 687, "b": 713, "c": 394, "d": 426), 0)',
    TAlgosimStructure.CreateWithValue(
      []
    )
  );

  Test('truncate(BinaryData("test"), 1000)', TAlgosimBinaryData.CreateWithValue([$74, 0, $65, 0, $73, 0, $74, 0]));
  Test('truncate(BinaryData("test"), 10)', TAlgosimBinaryData.CreateWithValue([$74, 0, $65, 0, $73, 0, $74, 0]));
  Test('truncate(BinaryData("test"), 9)', TAlgosimBinaryData.CreateWithValue([$74, 0, $65, 0, $73, 0, $74, 0]));
  Test('truncate(BinaryData("test"), 8)', TAlgosimBinaryData.CreateWithValue([$74, 0, $65, 0, $73, 0, $74, 0]));
  Test('truncate(BinaryData("test"), 7)', TAlgosimBinaryData.CreateWithValue([$74, 0, $65, 0, $73, 0, $74]));
  Test('truncate(BinaryData("test"), 6)', TAlgosimBinaryData.CreateWithValue([$74, 0, $65, 0, $73, 0]));
  Test('truncate(BinaryData("test"), 5)', TAlgosimBinaryData.CreateWithValue([$74, 0, $65, 0, $73]));
  Test('truncate(BinaryData("test"), 4)', TAlgosimBinaryData.CreateWithValue([$74, 0, $65, 0]));
  Test('truncate(BinaryData("test"), 3)', TAlgosimBinaryData.CreateWithValue([$74, 0, $65]));
  Test('truncate(BinaryData("test"), 2)', TAlgosimBinaryData.CreateWithValue([$74, 0]));
  Test('truncate(BinaryData("test"), 1)', TAlgosimBinaryData.CreateWithValue([$74]));
  Test('truncate(BinaryData("test"), 0)', TAlgosimBinaryData.CreateWithValue([]));

  // Swapping

  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 2, 4)', [10, 3, -5, 0, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 1, 4)', [3, 0, -5, 10, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 3, 7)', [10, 0, 4, 3, 8, -7, -5]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 1, 2)', [0, 10, -5, 3, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 1, 7)', [4, 0, -5, 3, 8, -7, 10]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 7, 1)', [4, 0, -5, 3, 8, -7, 10]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 2, 2)', [10, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 1, 1)', [10, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 7, 7)', [10, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 6, 3)', [10, 0, -7, 3, 8, -5, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 4, -1)', [10, 0, -5, 4, 8, -7, 3]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 2, -2)', [10, -7, -5, 3, 8, 0, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -2, -1)', [10, 0, -5, 3, 8, 4, -7]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -1, -2)', [10, 0, -5, 3, 8, 4, -7]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -1, -1)', [10, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -1, 4)', [10, 0, -5, 4, 8, -7, 3]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 4, -4)', [10, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 4, -5)', [10, 0, 3, -5, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -5, 4)', [10, 0, 3, -5, 8, -7, 4]);
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -7, 7)', [4, 0, -5, 3, 8, -7, 10]);

  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 4, 0)', failure, 'Index 0 out of bounds.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 4, 8)', failure, 'Index 8 out of bounds.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 4, 10)', failure, 'Index 10 out of bounds.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 4, 100)', failure, 'Index 100 out of bounds.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 4, -8)', failure, 'Index -8 out of bounds.');

  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 0, 5)', failure, 'Index 0 out of bounds.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 8, 2)', failure, 'Index 8 out of bounds.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 10, -4)', failure, 'Index 10 out of bounds.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, 100, -7)', failure, 'Index 100 out of bounds.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -8, 1)', failure, 'Index -8 out of bounds.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -8, 100)', failure, 'Index -8 out of bounds.');

  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -3, 1, 5)', failure, 'Too many arguments');

  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -3)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩)', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -3.5, 2)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');
  Test('swap(❨10, 0, −5, 3, 8, −7, 4❩, -3, i)', failure, 'An object of type integer was expected as argument 3, but an object of type complex number was given.');

  Test('swap(DebugObject("empty vector"), 1, 1)', failure, 'Index 1 out of bounds.');
  Test('swap(DebugObject("empty vector"), -1, 1)', failure, 'Index -1 out of bounds.');
  Test('swap(DebugObject("empty vector"), -1, -1)', failure, 'Index -1 out of bounds.');

  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 2, 4)', [ImaginaryUnit, 3, -5, 0, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 1, 4)', [3, 0, -5, ImaginaryUnit, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 3, 7)', [ImaginaryUnit, 0, 4, 3, 8, -7, -5]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 1, 2)', [0, ImaginaryUnit, -5, 3, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 1, 7)', [4, 0, -5, 3, 8, -7, ImaginaryUnit]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 7, 1)', [4, 0, -5, 3, 8, -7, ImaginaryUnit]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 2, 2)', [ImaginaryUnit, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 1, 1)', [ImaginaryUnit, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 7, 7)', [ImaginaryUnit, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 6, 3)', [ImaginaryUnit, 0, -7, 3, 8, -5, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 4, -1)', [ImaginaryUnit, 0, -5, 4, 8, -7, 3]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 2, -2)', [ImaginaryUnit, -7, -5, 3, 8, 0, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -2, -1)', [ImaginaryUnit, 0, -5, 3, 8, 4, -7]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -1, -2)', [ImaginaryUnit, 0, -5, 3, 8, 4, -7]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -1, -1)', [ImaginaryUnit, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -1, 4)', [ImaginaryUnit, 0, -5, 4, 8, -7, 3]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 4, -4)', [ImaginaryUnit, 0, -5, 3, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 4, -5)', [ImaginaryUnit, 0, 3, -5, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -5, 4)', [ImaginaryUnit, 0, 3, -5, 8, -7, 4]);
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -7, 7)', [4, 0, -5, 3, 8, -7, ImaginaryUnit]);

  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 4, 0)', failure, 'Index 0 out of bounds.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 4, 8)', failure, 'Index 8 out of bounds.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 4, 10)', failure, 'Index 10 out of bounds.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 4, 100)', failure, 'Index 100 out of bounds.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 4, -8)', failure, 'Index -8 out of bounds.');

  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 0, 5)', failure, 'Index 0 out of bounds.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 8, 2)', failure, 'Index 8 out of bounds.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 10, -4)', failure, 'Index 10 out of bounds.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, 100, -7)', failure, 'Index 100 out of bounds.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -8, 1)', failure, 'Index -8 out of bounds.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -8, 100)', failure, 'Index -8 out of bounds.');

  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -3, 1, 5)', failure, 'Too many arguments');

  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -3)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩)', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -3.5, 2)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');
  Test('swap(❨i, 0, −5, 3, 8, −7, 4❩, -3, i)', failure, 'An object of type integer was expected as argument 3, but an object of type complex number was given.');

  Test('swap(ComplexVector(DebugObject("empty vector")), 1, 1)', failure, 'Index 1 out of bounds.');
  Test('swap(ComplexVector(DebugObject("empty vector")), -1, 1)', failure, 'Index -1 out of bounds.');
  Test('swap(ComplexVector(DebugObject("empty vector")), -1, -1)', failure, 'Index -1 out of bounds.');

  Test('swap(❨❨5, −2, 10, 21, −10❩, ❨0, 1, 20, 5, −2❩❩, 3, 4)', failure, 'Object type "real matrix" is not a suitable container.'); // But it would be possible. Would be even neater if the indices were 2D.
  Test('swap(❨❨5, −2 + 5⋅i, 10, 21, −10❩, ❨0, 1, 20, 5, −2❩❩, 3, 4)', failure, 'Object type "complex matrix" is not a suitable container.'); // But it would be possible. Would be even neater if the indices were 2D.

  Test('swap("Hello, World!", 1, 8)', 'Wello, Horld!');
  Test('swap("Hello, World!", 1, -6)', 'Wello, Horld!');
  Test('swap("Hello, World!", 8, 1)', 'Wello, Horld!');
  Test('swap("Hello, World!", -6, 1)', 'Wello, Horld!');
  Test('swap("Hello, World!", -13, 8)', 'Wello, Horld!');
  Test('swap("Hello, World!", -13, -6)', 'Wello, Horld!');
  Test('swap("Hello, World!", 8, -13)', 'Wello, Horld!');
  Test('swap("Hello, World!", -6, -13)', 'Wello, Horld!');

  Test('swap("Hello, World!", 6, -1)', 'Hello! World,');
  Test('swap("Hello, World!", -1, 6)', 'Hello! World,');
  Test('swap("Hello, World!", 6, 13)', 'Hello! World,');
  Test('swap("Hello, World!", 13, 6)', 'Hello! World,');

  Test('swap("Hello, World!", 0, -13)', failure, 'Index 0 out of bounds.');
  Test('swap("Hello, World!", 14, -13)', failure, 'Index 14 out of bounds.');
  Test('swap("Hello, World!", -14, -13)', failure, 'Index -14 out of bounds.');

  Test('swap("Hello, World!", 2, 0)', failure, 'Index 0 out of bounds.');
  Test('swap("Hello, World!", 3, 14)', failure, 'Index 14 out of bounds.');
  Test('swap("Hello, World!", 4, -14)', failure, 'Index -14 out of bounds.');

  Test('swap("Hello, World!", 1, 1)', 'Hello, World!');
  Test('swap("Hello, World!", 2, 2)', 'Hello, World!');
  Test('swap("Hello, World!", 13, 13)', 'Hello, World!');
  Test('swap("Hello, World!", -1, -1)', 'Hello, World!');
  Test('swap("Hello, World!", -1, 13)', 'Hello, World!');
  Test('swap("Hello, World!", 13, -1)', 'Hello, World!');
  Test('swap("Hello, World!", -13, 1)', 'Hello, World!');
  Test('swap("Hello, World!", 1, -13)', 'Hello, World!');

  Test('swap("Hello, World!", 2, 2, 5)', failure, 'Too many arguments');
  Test('swap("Hello, World!", 5)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('swap("Hello, World!")', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('swap("!", 1, 1)', '!');
  Test('swap("", 1, 1)', failure, 'Index 1 out of bounds.');
  Test('swap("", -1, 1)', failure, 'Index -1 out of bounds.');
  Test('swap("", -1, -1)', failure, 'Index -1 out of bounds.');

  Test('swap()', failure, 'Too few arguments. A required argument of type object is missing.');

  Test('swap(''(3, 1, 4, "one", "five", 9, 2), 2, 3)', asoarr([3, 4, 1, 'one', 'five', 9, 2]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), 2, 6)', asoarr([3, 9, 4, 'one', 'five', 1, 2]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), 2, 7)', asoarr([3, 2, 4, 'one', 'five', 9, 1]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), 7, 2)', asoarr([3, 2, 4, 'one', 'five', 9, 1]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), 2, -1)', asoarr([3, 2, 4, 'one', 'five', 9, 1]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), -1, 2)', asoarr([3, 2, 4, 'one', 'five', 9, 1]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), 1, 1)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), -1, -1)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), 1, 2)', asoarr([1, 3, 4, 'one', 'five', 9, 2]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), -1, -2)', asoarr([3, 1, 4, 'one', 'five', 2, 9]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), -2, -1)', asoarr([3, 1, 4, 'one', 'five', 2, 9]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), 3, -7)', asoarr([4, 1, 3, 'one', 'five', 9, 2]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), -7, 3)', asoarr([4, 1, 3, 'one', 'five', 9, 2]));
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), -7, -7)', asoarr([3, 1, 4, 'one', 'five', 9, 2]));

  Test('swap(''(3, 1, 4, "one", "five", 9, 2), -7, 8)', failure, 'Index 8 out of bounds.');
  Test('swap(''(3, 1, 4, "one", "five", 9, 2), -8, 5)', failure, 'Index -8 out of bounds.');

  Test('swap({1, 2, 3, 4}, 2, 3)', failure, 'Object type "set" is not a suitable container.');


  //
  // Shuffling and reversing
  //

  // Reversing

  Test('reverse(❨6, 0, 3, -4, 8, 2, -7, 1, 4, 10❩)', [10, 4, 1, -7, 2, 8, -4, 3, 0, 6]);
  Test('reverse(❨21, 22, 23❩)', [23, 22, 21]);
  Test('reverse(❨21, 22❩)', [22, 21]);
  Test('reverse(❨21❩)', [21]);
  Test('reverse(DebugObject("empty vector"))', TAlgosimRealVector.Create);

  Test('reverse(❨3, i, 5, i, 0, 1, 10, 11❩)', [11, 10, 1, 0, ImaginaryUnit, 5, ImaginaryUnit, 3]);
  Test('reverse(❨2, i, 3❩)', [3, ImaginaryUnit, 2]);
  Test('reverse(❨2, i❩)', [ImaginaryUnit, 2]);
  Test('reverse(❨i❩)', [ImaginaryUnit]);
  Test('reverse(ComplexVector(DebugObject("empty vector")))', TAlgosimComplexVector.Create);

  Test('reverse(❨❨5, 1, 2, 0❩, ❨7, 6, 5, 1❩❩)', 4, [1, 5, 6, 7, 0, 2, 1, 5]);
  Test('reverse(❨❨1, 2❩, ❨0, 5❩❩)', 2, [5, 0, 2, 1]);
  Test('reverse(❨❨1, 2, 3, 4, 5, 6, 7❩❩)', 7, [7, 6, 5, 4, 3, 2, 1]);
  Test('reverse(❨❨1, 2, 3, 4, 5, 6❩❩)', 6, [6, 5, 4, 3, 2, 1]);
  Test('reverse(❨❨1❩, ❨2❩, ❨3❩❩)', 1, [3, 2, 1]);
  Test('reverse(❨❨394❩❩)', 1, [394]);
  Test('reverse(DebugObject("empty matrix"))', TAlgosimRealMatrix.Create);

  Test('reverse(❨❨5, i, 2, 0❩, ❨7, 6, 5, 1❩❩)', 4, [1, 5, 6, 7, 0, 2, ImaginaryUnit, 5]);
  Test('reverse(❨❨i, 2❩, ❨0, 5❩❩)', 2, [5, 0, 2, ImaginaryUnit]);
  Test('reverse(❨❨i, 2, 3, 4, 5, 6, 7❩❩)', 7, [7, 6, 5, 4, 3, 2, ImaginaryUnit]);
  Test('reverse(❨❨i, 2, 3, 4, 5, 6❩❩)', 6, [6, 5, 4, 3, 2, ImaginaryUnit]);
  Test('reverse(❨❨i❩, ❨2❩, ❨3❩❩)', 1, [3, 2, ImaginaryUnit]);
  Test('reverse(❨❨394⋅i❩❩)', 1, [394*ImaginaryUnit]);
  Test('reverse(ComplexMatrix(DebugObject("empty matrix")))', TAlgosimComplexMatrix.Create);

  Test('reverse("Hello, World!")', '!dlroW ,olleH');
  Test('reverse("Hello, World")', 'dlroW ,olleH');
  Test('reverse("ABC")', 'CBA');
  Test('reverse("AB")', 'BA');
  Test('reverse("A")', 'A');
  Test('reverse("")', '');
  Test('reverse("∫⌬dx")', 'xd⌬∫');

  Test('reverse(''(1, 2, 3, 4, 5))', intarr([5, 4, 3, 2, 1]));
  Test('reverse(''(1, 2, 3, 4))', intarr([4, 3, 2, 1]));
  Test('reverse(''(1, 2, 3))', intarr([3, 2, 1]));
  Test('reverse(''(1, 2))', intarr([2, 1]));
  Test('reverse(''(1))', intarr([1]));
  Test('reverse(''())', intarr([]));

//  Test('reverse(123456)', 654321); // would be cool; optional second argument specifies the base
//  Test('reverse(-314159265)', -562951413);
//  Test('reverse(123456789123456)', 654321987654321);
//  Test('reverse(-123456789123456)', -654321987654321);

  Test('reverse({1, 2, 3})', failure, 'An object of type "set" cannot be reversed.');
  Test('reverse(3.1415)', failure, 'An object of type "real number" cannot be reversed.');

  // Shuffling

  Test('u ≔ ❨3, −5, 12, −7, 2, 0, 52, 9, 11, 36❩;', null);
  Test('v ≔ shuffle(u); sort(u) = sort(v)', True);

  Test('u ≔ ❨3, −5, 12, −7, 2, 0, 52, 9, 11, 36, 123❩;', null);
  Test('v ≔ shuffle(u); sort(u) = sort(v)', True);

  Test('u ≔ ❨3, −5, 12, −7, 2, 0, 52, 9, 11, 36, 123, 0❩;', null);
  Test('v ≔ shuffle(u); sort(u) = sort(v)', True);

  Test('u ≔ ❨3, −5, 12, −7, 2, 0, 52, 9, 11, 36, 123, 0, 0, 0, 0❩;', null);
  Test('v ≔ shuffle(u); sort(u) = sort(v)', True);

  Test('u ≔ ❨3, −5❩;', null);
  Test('v ≔ shuffle(u); sort(u) = sort(v)', True);

  Test('u ≔ ❨3❩;', null);
  Test('v ≔ shuffle(u); sort(u) = sort(v)', True);

  Test('u ≔ DebugObject("empty vector");', null);
  Test('v ≔ shuffle(u); sort(u) = sort(v)', True);

  Test('u ≔ ❨6, 1 + i, 2 − 5⋅i, 4 + 3⋅i, −i, −i, 123, 0, 55⋅i❩;', null);
  Test('v ≔ shuffle(u); sort(u, "real and imaginary parts") = sort(v, "real and imaginary parts")', True);

  Test('u ≔ ❨6, 1 + i, 2 − 5⋅i, 4 + 3⋅i, −i, −i, 123, 0, 55⋅i, 3❩;', null);
  Test('v ≔ shuffle(u); sort(u, "real and imaginary parts") = sort(v, "real and imaginary parts")', True);

  Test('u ≔ ❨6, 1 + i, 2 − 5⋅i, 4 + 3⋅i, −i, −i, 123, 0, 55⋅i, 3, 3, 3, 3, 3❩;', null);
  Test('v ≔ shuffle(u); sort(u, "real and imaginary parts") = sort(v, "real and imaginary parts")', True);

  Test('u ≔ ❨6, 1 + i❩;', null);
  Test('v ≔ shuffle(u); sort(u, "real and imaginary parts") = sort(v, "real and imaginary parts")', True);

  Test('u ≔ ❨1 + i❩;', null);
  Test('v ≔ shuffle(u); sort(u, "real and imaginary parts") = sort(v, "real and imaginary parts")', True);

  Test('u ≔ ComplexVector(DebugObject("empty vector"));', null);
  Test('v ≔ shuffle(u); sort(u, "real and imaginary parts") = sort(v, "real and imaginary parts")', True);

  Test('A ≔ ❨❨5, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩, ❨4, 9, 5, 1, 2❩, ❨5, 0, 3, 9, 6❩, ❨−4, 1, 5, 7, 2❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩, ❨4, 9, 5, 1, 2❩, ❨5, 0, 3, 9, 6❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩, ❨4, 9, 5, 1, 2❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5, 2, 3, 6, 7❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5, 2, 3, 6, 7, 1❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5, 2❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩, ❨4, 9, 5, 1, 2❩, ❨5, 0, 3, 9, 6❩, ❨−4, 1, 5, 7, 2❩, ❨1, 1, 10, -5, -7❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5❩, ❨4❩, ❨4❩, ❨5❩, ❨−4❩, ❨1❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨5❩, ❨4❩, ❨4❩, ❨5❩, ❨−4❩❩;', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ DebugObject("empty matrix");', null);
  Test('B ≔ shuffle(A); sort(A) = sort(B)', True);

  Test('A ≔ ❨❨i, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩, ❨4, 9, 5, 1, 2❩, ❨5, 0, 3, 9, 6❩, ❨−4, 1, 5, 7, 2❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩, ❨4, 9, 5, 1, 2❩, ❨5, 0, 3, 9, 6❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩, ❨4, 9, 5, 1, 2❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i, 2, 3, 6, 7❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i, 2, 3, 6, 7, 1❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i, 2❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i, 2, 3, 6, 7❩, ❨4, 0, −2, 4, 8❩, ❨4, 9, 5, 1, 2❩, ❨5, 0, 3, 9, 6❩, ❨−4, 1, 5, 7, 2❩, ❨1, 1, 10, -5, -7❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i❩, ❨4❩, ❨4❩, ❨5❩, ❨−4❩, ❨1❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ❨❨i❩, ❨4❩, ❨4❩, ❨5❩, ❨−4❩❩;', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('A ≔ ComplexMatrix(DebugObject("empty matrix"));', null);
  Test('B ≔ shuffle(A); sort(A, "real and imaginary parts") = sort(B, "real and imaginary parts")', True);

  Test('s ≔ "Hello, World!";', null);
  Test('t ≔ shuffle(s); sort(characters(s)) = sort(characters(t))', True);

  Test('s ≔ "Hello, Dogs!";', null);
  Test('t ≔ shuffle(s); sort(characters(s)) = sort(characters(t))', True);

  Test('s ≔ "It is the quality of one’s convictions that determines success, not the number of followers.";', null);
  Test('t ≔ shuffle(s); sort(characters(s)) = sort(characters(t))', True);

  Test('s ≔ "Hi!";', null);
  Test('t ≔ shuffle(s); sort(characters(s)) = sort(characters(t))', True);

  Test('s ≔ "Hi";', null);
  Test('t ≔ shuffle(s); sort(characters(s)) = sort(characters(t))', True);

  Test('s ≔ "H";', null);
  Test('t ≔ shuffle(s); sort(characters(s)) = sort(characters(t))', True);

  Test('s ≔ "";', null);
  Test('t ≔ shuffle(s); sort(characters(s)) = sort(characters(t))', True);

  Test('L ≔ ''(50, true, "cat", π, e^2, i, ❨1, 2, 3❩, IdentityMatrix(10), now(), 4, color("red"));', null);
  Test('M ≔ shuffle(L); sort(L) = sort(M)', True);

  Test('L ≔ ''(50, true, "cat", π, e^2, i, ❨1, 2, 3❩, IdentityMatrix(10), now(), 4, color("red"), 1);', null);
  Test('M ≔ shuffle(L); sort(L) = sort(M)', True);

  Test('L ≔ ''(50, true, color("red"));', null);
  Test('M ≔ shuffle(L); sort(L) = sort(M)', True);

  Test('L ≔ ''(50, color("red"));', null);
  Test('M ≔ shuffle(L); sort(L) = sort(M)', True);

  Test('L ≔ ''(color("red"));', null);
  Test('M ≔ shuffle(L); sort(L) = sort(M)', True);

  Test('L ≔ ''();', null);
  Test('M ≔ shuffle(L); sort(L) = sort(M)', True);

  Test('shuffle({3, 4, 1})', failure, 'An object of type "set" cannot be shuffled.');
  Test('shuffle(now())', failure, 'An object of type "typed structure" cannot be shuffled.');
  Test('shuffle(123)', failure, 'An object of type "integer" cannot be shuffled.'); // see remarks about reverse above
  Test('shuffle(123.456)', failure, 'An object of type "real number" cannot be shuffled.');
  Test('shuffle(123.456 + i)', failure, 'An object of type "complex number" cannot be shuffled.');
  Test('shuffle(true)', failure, 'An object of type "boolean" cannot be shuffled.');

  Test('shuffle()', failure, 'Too few arguments. A required argument of type object is missing.');
  Test('shuffle(L, 2)', failure, 'Too many arguments.');


  //
  // Duplicate removal
  //

  Test('unique(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, 1, 4, 1, 8, 9, 1, −1, 7, 123❩)',
    [1, 5, 0, 3, 6, 4, 2, 8, 9, -1, 7, 123]);

  Test('unique(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, 1, 4, 1, 8, 9, 1, −1, 7, 123, 1, 1, 0, 5, 6, 4, 1, 2, 0, 4, 8, 9, -1, 7, 123, 123, 1, 0, 123❩)',
    [1, 5, 0, 3, 6, 4, 2, 8, 9, -1, 7, 123]);

  Test('unique(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, 1, 4, 1, 8, 9, 1, −1, 7, 123, 1, 1, 0, 5, 6, 4, 1, 2, 0, 4, 8, 9, -1, 7, 123, 123, 1, 0, 123, 21❩)',
    [1, 5, 0, 3, 6, 4, 2, 8, 9, -1, 7, 123, 21]);

  Test('unique(❨1, 5, 0❩)',
    [1, 5, 0]);

  Test('unique(❨1, 5, 5❩)',
    [1, 5]);

  Test('unique(❨1, 5, 1❩)',
    [1, 5]);

  Test('unique(❨1, 5❩)',
    [1, 5]);

  Test('unique(❨1, 1❩)',
    [1]);

  Test('unique(❨1❩)',
    [1]);

  Test('unique(ZeroVector(10000))',
    [0]);

  Test('unique(BasisVector(10000, 6681))',
    [0, 1]);

  Test('unique(DebugObject("empty vector"))',
    TAlgosimRealVector.Create);

  Test('unique(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, i, 4, i, 8, 9, 1, −1, 7, 123❩)',
    [1, 5, 0, 3, 6, 4, 2, ImaginaryUnit, 8, 9, -1, 7, 123]);

  Test('unique(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, i, 4, i, 8, 9, 1, −1, 7, 123, 1, 1, 0, 5, 6, 4, 1, 2, 0, 4, 8, 9, -1, 7, 123, 123, 1, 0, 123❩)',
    [1, 5, 0, 3, 6, 4, 2, ImaginaryUnit, 8, 9, -1, 7, 123]);

  Test('unique(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, i, 4, i, 8, 9, 1, −1, 7, 123, 1, 1, 0, 5, 6, 4, 1, 2, 0, 4, 8, 9, -1, 7, 123, 123, 1, 0, 123, 21❩)',
    [1, 5, 0, 3, 6, 4, 2, ImaginaryUnit, 8, 9, -1, 7, 123, 21]);

  Test('unique(❨i, 5, 0❩)',
    [ImaginaryUnit, 5, 0]);

  Test('unique(❨i, 5, 5❩)',
    [ImaginaryUnit, 5]);

  Test('unique(❨i, 5, i❩)',
    [ImaginaryUnit, 5]);

  Test('unique(❨i, 5❩)',
    [ImaginaryUnit, 5]);

  Test('unique(❨i, i❩)',
    [ImaginaryUnit]);

  Test('unique(❨i❩)',
    [ImaginaryUnit]);

  Test('unique(ComplexZeroVector(10000))',
    [TASC(0)]);

  Test('unique(ComplexVector(BasisVector(10000, 6681)))',
    [TASC(0), 1]);

  Test('unique(ComplexVector(DebugObject("empty vector")))',
    TAlgosimComplexVector.Create);

  Test('unique(❨❨6, 3, 2, 5❩, ❨7, 0, 1, 3❩❩)', failure, 'Cannot remove duplicates from an object of type "real matrix".');
  Test('unique(❨❨6, 3, i, 5❩, ❨7, 0, 1, 3❩❩)', failure, 'Cannot remove duplicates from an object of type "complex matrix".');
  Test('unique({5, 2, 3, 6, 5, 2, 3, 7, 4, 0, 1})', failure, 'Cannot remove duplicates from an object of type "set".');

  Test('unique("Hello, World!")', 'Helo, Wrd!');
  Test('unique("This is nothing but a test.")', 'This notgbuae.');
  Test('unique("Aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...")', 'Aa.');
  Test('unique("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")', 'a');
  Test('unique("Skräptyg AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAB")', 'Skräptyg AB');
  Test('unique("")', '');

  Test('unique(''(21, 50, 32, 21, "cat", "dog", π, 50, 51, "rat", "dog", 21, true, color("red"), color("green"), color("blue"), 123, 100, e, π, π^2, color("red"), color("black"), 0))',
    asoarrex(
      [
        ASOInt(21),
        ASOInt(50),
        ASOInt(32),
        ASO('cat'),
        ASO('dog'),
        ASO(Pi),
        ASOInt(51),
        ASO('rat'),
        ASO(True),
        ASO(TRGB(clRed)),
        ASO(TRGB(clGreen)),
        ASO(TRGB(clBlue)),
        ASOInt(123),
        ASOInt(100),
        ASO(Exp(1)),
        ASO(Pi*Pi),
        ASO(TRGB(clBlack)),
        ASOInt(0)
      ]
    ));

  Test('unique(''(21, 50, 32, 21, "cat", "dog", π, 50, 51, "rat", "dog", 21, true, color("red"), color("green"), color("blue"), '+'123, 100, e, π, π^2, color("red"), color("black"), 0, 21, "dog", 100, e, color("red"), 50, 0, 99, color("black"), 32, 32, 32, color("red"), true, true))',
    asoarrex(
      [
        ASOInt(21),
        ASOInt(50),
        ASOInt(32),
        ASO('cat'),
        ASO('dog'),
        ASO(Pi),
        ASOInt(51),
        ASO('rat'),
        ASO(True),
        ASO(TRGB(clRed)),
        ASO(TRGB(clGreen)),
        ASO(TRGB(clBlue)),
        ASOInt(123),
        ASOInt(100),
        ASO(Exp(1)),
        ASO(Pi*Pi),
        ASO(TRGB(clBlack)),
        ASOInt(0),
        ASOInt(99)
      ]
    ));

  Test('unique(''(21, 50, 32, 21))',
    asoarrex(
      [
        ASOInt(21),
        ASOInt(50),
        ASOInt(32)
      ]
    ));

  Test('unique(''(21, 50, 21))',
    asoarrex(
      [
        ASOInt(21),
        ASOInt(50)
      ]
    ));

  Test('unique(''(21, 21))',
    asoarrex(
      [
        ASOInt(21)
      ]
    ));

  Test('unique(''(21))',
    asoarrex(
      [
        ASOInt(21)
      ]
    ));

  Test('unique(''())',
    asoarrex(
      [

      ]
    ));

  Test('RemoveDuplicates(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, 1, 4, 1, 8, 9, 1, −1, 7, 123❩)',
    [1, 5, 0, 3, 6, 4, 2, 8, 9, -1, 7, 123]);

  Test('RemoveDuplicates(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, 1, 4, 1, 8, 9, 1, −1, 7, 123, 1, 1, 0, 5, 6, 4, 1, 2, 0, 4, 8, 9, -1, 7, 123, 123, 1, 0, 123❩)',
    [1, 5, 0, 3, 6, 4, 2, 8, 9, -1, 7, 123]);

  Test('RemoveDuplicates(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, 1, 4, 1, 8, 9, 1, −1, 7, 123, 1, 1, 0, 5, 6, 4, 1, 2, 0, 4, 8, 9, -1, 7, 123, 123, 1, 0, 123, 21❩)',
    [1, 5, 0, 3, 6, 4, 2, 8, 9, -1, 7, 123, 21]);

  Test('RemoveDuplicates(❨1, 5, 0❩)',
    [1, 5, 0]);

  Test('RemoveDuplicates(❨1, 5, 5❩)',
    [1, 5]);

  Test('RemoveDuplicates(❨1, 5, 1❩)',
    [1, 5]);

  Test('RemoveDuplicates(❨1, 5❩)',
    [1, 5]);

  Test('RemoveDuplicates(❨1, 1❩)',
    [1]);

  Test('RemoveDuplicates(❨1❩)',
    [1]);

  Test('RemoveDuplicates(ZeroVector(10000))',
    [0]);

  Test('RemoveDuplicates(BasisVector(10000, 6681))',
    [0, 1]);

  Test('RemoveDuplicates(DebugObject("empty vector"))',
    TAlgosimRealVector.Create);

  Test('RemoveDuplicates(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, i, 4, i, 8, 9, 1, −1, 7, 123❩)',
    [1, 5, 0, 3, 6, 4, 2, ImaginaryUnit, 8, 9, -1, 7, 123]);

  Test('RemoveDuplicates(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, i, 4, i, 8, 9, 1, −1, 7, 123, 1, 1, 0, 5, 6, 4, 1, 2, 0, 4, 8, 9, -1, 7, 123, 123, 1, 0, 123❩)',
    [1, 5, 0, 3, 6, 4, 2, ImaginaryUnit, 8, 9, -1, 7, 123]);

  Test('RemoveDuplicates(❨1, 5, 0, 1, 3, 6, 4, 1, 2, 0, i, 4, i, 8, 9, 1, −1, 7, 123, 1, 1, 0, 5, 6, 4, 1, 2, 0, 4, 8, 9, -1, 7, 123, 123, 1, 0, 123, 21❩)',
    [1, 5, 0, 3, 6, 4, 2, ImaginaryUnit, 8, 9, -1, 7, 123, 21]);

  Test('RemoveDuplicates(❨i, 5, 0❩)',
    [ImaginaryUnit, 5, 0]);

  Test('RemoveDuplicates(❨i, 5, 5❩)',
    [ImaginaryUnit, 5]);

  Test('RemoveDuplicates(❨i, 5, i❩)',
    [ImaginaryUnit, 5]);

  Test('RemoveDuplicates(❨i, 5❩)',
    [ImaginaryUnit, 5]);

  Test('RemoveDuplicates(❨i, i❩)',
    [ImaginaryUnit]);

  Test('RemoveDuplicates(❨i❩)',
    [ImaginaryUnit]);

  Test('RemoveDuplicates(ComplexZeroVector(10000))',
    [TASC(0)]);

  Test('RemoveDuplicates(ComplexVector(BasisVector(10000, 6681)))',
    [TASC(0), 1]);

  Test('RemoveDuplicates(ComplexVector(DebugObject("empty vector")))',
    TAlgosimComplexVector.Create);

  Test('RemoveDuplicates(❨❨6, 3, 2, 5❩, ❨7, 0, 1, 3❩❩)', failure, 'Cannot remove duplicates from an object of type "real matrix".');
  Test('RemoveDuplicates(❨❨6, 3, i, 5❩, ❨7, 0, 1, 3❩❩)', failure, 'Cannot remove duplicates from an object of type "complex matrix".');
  Test('RemoveDuplicates({5, 2, 3, 6, 5, 2, 3, 7, 4, 0, 1})', failure, 'Cannot remove duplicates from an object of type "set".');

  Test('RemoveDuplicates("Hello, World!")', 'Helo, Wrd!');
  Test('RemoveDuplicates("This is nothing but a test.")', 'This notgbuae.');
  Test('RemoveDuplicates("Aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...")', 'Aa.');
  Test('RemoveDuplicates("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")', 'a');
  Test('RemoveDuplicates("Skräptyg AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAB")', 'Skräptyg AB');
  Test('RemoveDuplicates("")', '');

  Test('RemoveDuplicates(''(21, 50, 32, 21, "cat", "dog", π, 50, 51, "rat", "dog", 21, true, color("red"), color("green"), color("blue"), 123, 100, e, π, π^2, color("red"), color("black"), 0))',
    asoarrex(
      [
        ASOInt(21),
        ASOInt(50),
        ASOInt(32),
        ASO('cat'),
        ASO('dog'),
        ASO(Pi),
        ASOInt(51),
        ASO('rat'),
        ASO(True),
        ASO(TRGB(clRed)),
        ASO(TRGB(clGreen)),
        ASO(TRGB(clBlue)),
        ASOInt(123),
        ASOInt(100),
        ASO(Exp(1)),
        ASO(Pi*Pi),
        ASO(TRGB(clBlack)),
        ASOInt(0)
      ]
    ));

  Test('RemoveDuplicates(''(21, 50, 32, 21, "cat", "dog", π, 50, 51, "rat", "dog", 21, true, color("red"), color("green"), color("blue"), '+'123, 100, e, π, π^2, color("red"), color("black"), 0, 21, "dog", 100, e, color("red"), 50, 0, 99, color("black"), 32, 32, 32, color("red"), true, true))',
    asoarrex(
      [
        ASOInt(21),
        ASOInt(50),
        ASOInt(32),
        ASO('cat'),
        ASO('dog'),
        ASO(Pi),
        ASOInt(51),
        ASO('rat'),
        ASO(True),
        ASO(TRGB(clRed)),
        ASO(TRGB(clGreen)),
        ASO(TRGB(clBlue)),
        ASOInt(123),
        ASOInt(100),
        ASO(Exp(1)),
        ASO(Pi*Pi),
        ASO(TRGB(clBlack)),
        ASOInt(0),
        ASOInt(99)
      ]
    ));

  Test('RemoveDuplicates(''(21, 50, 32, 21))',
    asoarrex(
      [
        ASOInt(21),
        ASOInt(50),
        ASOInt(32)
      ]
    ));

  Test('RemoveDuplicates(''(21, 50, 21))',
    asoarrex(
      [
        ASOInt(21),
        ASOInt(50)
      ]
    ));

  Test('RemoveDuplicates(''(21, 21))',
    asoarrex(
      [
        ASOInt(21)
      ]
    ));

  Test('RemoveDuplicates(''(21))',
    asoarrex(
      [
        ASOInt(21)
      ]
    ));

  Test('RemoveDuplicates(''())',
    asoarrex(
      [

      ]
    ));

  // with epsilon

  Test('v ≔ ❨5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 5, 3, 2.99, −5, 1, 7, 0, 1, 5.00000001❩',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 5, 3, 2.99, -5, 1, 7, 0, 1, 5.00000001]);

  Test('unique(v)',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 2.99, -5, 1, 0, 5.00000001]);

  Test('unique(v, 2E-9)',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 2.99, -5, 1, 0, 5.00000001]);

  Test('unique(v, 2E-8)',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('unique(v, 2E-7)',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('unique(v, 2E-6)',
    [5, 3, 5.0001, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('unique(v, 2E-5)',
    [5, 3, 5.0001, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('unique(v, 2E-4)',
    [5, 3, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('unique(v, 2E-3)',
    [5, 3, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('unique(v, 2E-2)',
    [5, 3, 7, 21, 213, 8, -5, 1, 0]);

  Test('unique(v, 2E-1)',
    [5, 3, 7, 21, 213, 8, -5, 1, 0]);

  Test('unique(v, 2.5)',
    [5, 21, 213, 8, -5, 1]);

  Test('unique(v, 3.5)',
    [5, 21, 213, -5, 1]);

  Test('unique(v, 4.5)',
    [5, 21, 213, -5, 0]);

  Test('unique(v, 10.5)',
    [5, 21, 213]);

  Test('unique(v, 100)',
    [5, 213]);

  Test('unique(v, 300)',
    [5]);

  Test('unique(v, 300000000)',
    [5]);

  Test('unique(DebugObject("empty vector"), 0.001)', TAlgosimRealVector.Create);

  Test('v ≔ ❨3, i, 1 + i, 5, 2.99, 1.001 + i, 1 + 1.00001 ⋅ i, 7, 11, 5.001, 23, 123, 5 + 0.01⋅i, 50, 5, 0, 3❩',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 5, 0, 3]);

  Test('unique(v)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('unique(v, 2E-9)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('unique(v, 2E-8)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('unique(v, 2E-7)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('unique(v, 2E-6)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('unique(v, 2E-5)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('unique(v, 2E-4)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('unique(v, 2E-3)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 7, 11, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('unique(v, 2E-2)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 7, 11, 23, 123, 50, 0]);

  Test('unique(v, 2E-1)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 7, 11, 23, 123, 50, 0]);

  Test('unique(v, 2.5)',
    [3, ImaginaryUnit, 7, 11, 23, 123, 50]);

  Test('unique(v, 4.5)',
    [TASC(3), 11, 23, 123, 50]);

  Test('unique(v, 10)',
    [TASC(3), 23, 123, 50]);

  Test('unique(v, 25)',
    [TASC(3), 123, 50]);

  Test('unique(v, 50)',
    [TASC(3), 123]);

  Test('unique(v, 500)',
    [TASC(3)]);

  Test('unique(v, 500000000)',
    [TASC(3)]);

  Test('unique(ComplexVector(DebugObject("empty vector")), 0.001)', TAlgosimComplexVector.Create);

  Test('unique(❨❨6, 3, 2, 5❩, ❨7, 0, 1, 3❩❩, 0.0001)', failure, 'Cannot remove duplicates from an object of type "real matrix".');
  Test('unique(❨❨6, 3, i, 5❩, ❨7, 0, 1, 3❩❩, 0.0001)', failure, 'Cannot remove duplicates from an object of type "complex matrix".');

  Test('unique("test", 0.001)', failure, 'Cannot remove duplicates from an object of type "string".');

  Test('L ≔ ''(3, 2, 5, 6, 3.0001, 7, 12, 12.001, 5, 6.99, 5.00001, 123, 124, 124.5, 7.1, 2, 123 + 1E−10)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 5, 6.99, 5.00001, 123, 124, 124.5, 7.1, 2, 123 + 1E-10]));

  Test('unique(L)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1, 123 + 1E-10]));

  Test('unique(L, 2E-12)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1, 123 + 1E-10]));

  Test('unique(L, 2E-11)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1, 123 + 1E-10]));

  Test('unique(L, 2E-10)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('unique(L, 2E-9)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('unique(L, 2E-8)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('unique(L, 2E-7)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('unique(L, 2E-6)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('unique(L, 2E-5)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 123, 124, 124.5, 7.1]));

  Test('unique(L, 2E-4)',
    asoarr([3, 2, 5, 6, 7, 12, 12.001, 6.99, 123, 124, 124.5, 7.1]));

  Test('unique(L, 2E-3)',
    asoarr([3, 2, 5, 6, 7, 12, 6.99, 123, 124, 124.5, 7.1]));

  Test('unique(L, 2E-2)',
    asoarr([3, 2, 5, 6, 7, 12, 123, 124, 124.5, 7.1]));

  Test('unique(L, 2E-1)',
    asoarr([3, 2, 5, 6, 7, 12, 123, 124, 124.5]));

  Test('unique(L, 1.4)',
    asoarr([3, 5, 7, 12, 123, 124.5]));

  Test('unique(L, 2.4)',
    asoarr([3, 6, 12, 123]));

  Test('unique(L, 10)',
    asoarr([3, 123]));

  Test('unique(L, 100)',
    asoarr([3, 123]));

  Test('unique(L, 1000)',
    asoarr([3]));

  Test('unique(L, 1E50)',
    asoarr([3]));

  Test('L ≔ ''(3, 2, "cat", 5, 6, 3.0001, 7, 12, "cat", 12.001, 5, "dog", "rat", 6.99, 5.00001, 123, 124, 124.5, 7.1, 2, "rabbit", 123 + 1E−10, "thestral")',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 'cat', 12.001, 5, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 2, 'rabbit', 123 + 1E-10, 'thestral']));

  Test('unique(L)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 123 + 1E-10, 'thestral']));

  Test('unique(L, 2E-12)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 123 + 1E-10, 'thestral']));

  Test('unique(L, 2E-11)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 123 + 1E-10, 'thestral']));

  Test('unique(L, 2E-10)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('unique(L, 2E-9)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('unique(L, 2E-8)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('unique(L, 2E-7)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('unique(L, 2E-6)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('unique(L, 2E-5)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('unique(L, 2E-4)',
    asoarr([3, 2, 'cat', 5, 6, 7, 12, 12.001, 'dog', 'rat', 6.99, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('unique(L, 2E-3)',
    asoarr([3, 2, 'cat', 5, 6, 7, 12, 'dog', 'rat', 6.99, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('unique(L, 2E-2)',
    asoarr([3, 2, 'cat', 5, 6, 7, 12, 'dog', 'rat', 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('unique(L, 2E-1)',
    asoarr([3, 2, 'cat', 5, 6, 7, 12, 'dog', 'rat', 123, 124, 124.5, 'rabbit', 'thestral']));

  Test('unique(L, 1.4)',
    asoarr([3, 'cat', 5, 7, 12, 'dog', 'rat', 123, 124.5, 'rabbit', 'thestral']));

  Test('unique(L, 2.4)',
    asoarr([3, 'cat', 6, 12, 'dog', 'rat', 123, 'rabbit', 'thestral']));

  Test('unique(L, 5)',
    asoarr([3, 'cat', 12, 'dog', 'rat', 123, 'rabbit', 'thestral']));

  Test('unique(L, 10)',
    asoarr([3, 'cat', 'dog', 'rat', 123, 'rabbit', 'thestral']));

  Test('unique(L, 100)',
    asoarr([3, 'cat', 'dog', 'rat', 123, 'rabbit', 'thestral']));

  Test('unique(L, 1000)',
    asoarr([3, 'cat', 'dog', 'rat', 'rabbit', 'thestral']));

  Test('unique(L, 1E20)',
    asoarr([3, 'cat', 'dog', 'rat', 'rabbit', 'thestral']));

  Test('L ≔ ''(2 + 3⋅i, 5 − i, 2 + 3.001⋅i, 4.99 − i, 12, 2 + 3⋅i)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12), ASO(2 + 3*ImaginaryUnit)]));

  Test('unique(L)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 2E-6)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 2E-5)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 2E-4)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 2E-3)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 2E-2)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 2)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 3)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 4.2)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 9)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASOInt(12)]));

  Test('unique(L, 10.1)',
    asoarrex([ASO(2 + 3*ImaginaryUnit)]));

  Test('unique(L, 1E50)',
    asoarrex([ASO(2 + 3*ImaginaryUnit)]));

  Test('v ≔ ❨5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 5, 3, 2.99, −5, 1, 7, 0, 1, 5.00000001❩',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 5, 3, 2.99, -5, 1, 7, 0, 1, 5.00000001]);

  Test('RemoveDuplicates(v)',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 2.99, -5, 1, 0, 5.00000001]);

  Test('RemoveDuplicates(v, 2E-9)',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 2.99, -5, 1, 0, 5.00000001]);

  Test('RemoveDuplicates(v, 2E-8)',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('RemoveDuplicates(v, 2E-7)',
    [5, 3, 5.0001, 3.000001, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('RemoveDuplicates(v, 2E-6)',
    [5, 3, 5.0001, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('RemoveDuplicates(v, 2E-5)',
    [5, 3, 5.0001, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('RemoveDuplicates(v, 2E-4)',
    [5, 3, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('RemoveDuplicates(v, 2E-3)',
    [5, 3, 7, 21, 213, 8, 2.99, -5, 1, 0]);

  Test('RemoveDuplicates(v, 2E-2)',
    [5, 3, 7, 21, 213, 8, -5, 1, 0]);

  Test('RemoveDuplicates(v, 2E-1)',
    [5, 3, 7, 21, 213, 8, -5, 1, 0]);

  Test('RemoveDuplicates(v, 2.5)',
    [5, 21, 213, 8, -5, 1]);

  Test('RemoveDuplicates(v, 3.5)',
    [5, 21, 213, -5, 1]);

  Test('RemoveDuplicates(v, 4.5)',
    [5, 21, 213, -5, 0]);

  Test('RemoveDuplicates(v, 10.5)',
    [5, 21, 213]);

  Test('RemoveDuplicates(v, 100)',
    [5, 213]);

  Test('RemoveDuplicates(v, 300)',
    [5]);

  Test('RemoveDuplicates(v, 300000000)',
    [5]);

  Test('RemoveDuplicates(DebugObject("empty vector"), 0.001)', TAlgosimRealVector.Create);

  Test('v ≔ ❨3, i, 1 + i, 5, 2.99, 1.001 + i, 1 + 1.00001 ⋅ i, 7, 11, 5.001, 23, 123, 5 + 0.01⋅i, 50, 5, 0, 3❩',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 5, 0, 3]);

  Test('RemoveDuplicates(v)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('RemoveDuplicates(v, 2E-9)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('RemoveDuplicates(v, 2E-8)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('RemoveDuplicates(v, 2E-7)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('RemoveDuplicates(v, 2E-6)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 1 + 1.00001 * ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('RemoveDuplicates(v, 2E-5)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('RemoveDuplicates(v, 2E-4)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 1.001 + ImaginaryUnit, 7, 11, 5.001, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('RemoveDuplicates(v, 2E-3)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 2.99, 7, 11, 23, 123, 5 + 0.01*ImaginaryUnit, 50, 0]);

  Test('RemoveDuplicates(v, 2E-2)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 7, 11, 23, 123, 50, 0]);

  Test('RemoveDuplicates(v, 2E-1)',
    [3, ImaginaryUnit, 1 + ImaginaryUnit, 5, 7, 11, 23, 123, 50, 0]);

  Test('RemoveDuplicates(v, 2.5)',
    [3, ImaginaryUnit, 7, 11, 23, 123, 50]);

  Test('RemoveDuplicates(v, 4.5)',
    [TASC(3), 11, 23, 123, 50]);

  Test('RemoveDuplicates(v, 10)',
    [TASC(3), 23, 123, 50]);

  Test('RemoveDuplicates(v, 25)',
    [TASC(3), 123, 50]);

  Test('RemoveDuplicates(v, 50)',
    [TASC(3), 123]);

  Test('RemoveDuplicates(v, 500)',
    [TASC(3)]);

  Test('RemoveDuplicates(v, 500000000)',
    [TASC(3)]);

  Test('RemoveDuplicates(ComplexVector(DebugObject("empty vector")), 0.001)', TAlgosimComplexVector.Create);

  Test('RemoveDuplicates(❨❨6, 3, 2, 5❩, ❨7, 0, 1, 3❩❩, 0.0001)', failure, 'Cannot remove duplicates from an object of type "real matrix".');
  Test('RemoveDuplicates(❨❨6, 3, i, 5❩, ❨7, 0, 1, 3❩❩, 0.0001)', failure, 'Cannot remove duplicates from an object of type "complex matrix".');

  Test('RemoveDuplicates("test", 0.001)', failure, 'Cannot remove duplicates from an object of type "string".');

  Test('L ≔ ''(3, 2, 5, 6, 3.0001, 7, 12, 12.001, 5, 6.99, 5.00001, 123, 124, 124.5, 7.1, 2, 123 + 1E−10)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 5, 6.99, 5.00001, 123, 124, 124.5, 7.1, 2, 123 + 1E-10]));

  Test('RemoveDuplicates(L)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1, 123 + 1E-10]));

  Test('RemoveDuplicates(L, 2E-12)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1, 123 + 1E-10]));

  Test('RemoveDuplicates(L, 2E-11)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1, 123 + 1E-10]));

  Test('RemoveDuplicates(L, 2E-10)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('RemoveDuplicates(L, 2E-9)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('RemoveDuplicates(L, 2E-8)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('RemoveDuplicates(L, 2E-7)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('RemoveDuplicates(L, 2E-6)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 5.00001, 123, 124, 124.5, 7.1]));

  Test('RemoveDuplicates(L, 2E-5)',
    asoarr([3, 2, 5, 6, 3.0001, 7, 12, 12.001, 6.99, 123, 124, 124.5, 7.1]));

  Test('RemoveDuplicates(L, 2E-4)',
    asoarr([3, 2, 5, 6, 7, 12, 12.001, 6.99, 123, 124, 124.5, 7.1]));

  Test('RemoveDuplicates(L, 2E-3)',
    asoarr([3, 2, 5, 6, 7, 12, 6.99, 123, 124, 124.5, 7.1]));

  Test('RemoveDuplicates(L, 2E-2)',
    asoarr([3, 2, 5, 6, 7, 12, 123, 124, 124.5, 7.1]));

  Test('RemoveDuplicates(L, 2E-1)',
    asoarr([3, 2, 5, 6, 7, 12, 123, 124, 124.5]));

  Test('RemoveDuplicates(L, 1.4)',
    asoarr([3, 5, 7, 12, 123, 124.5]));

  Test('RemoveDuplicates(L, 2.4)',
    asoarr([3, 6, 12, 123]));

  Test('RemoveDuplicates(L, 10)',
    asoarr([3, 123]));

  Test('RemoveDuplicates(L, 100)',
    asoarr([3, 123]));

  Test('RemoveDuplicates(L, 1000)',
    asoarr([3]));

  Test('RemoveDuplicates(L, 1E50)',
    asoarr([3]));

  Test('L ≔ ''(3, 2, "cat", 5, 6, 3.0001, 7, 12, "cat", 12.001, 5, "dog", "rat", 6.99, 5.00001, 123, 124, 124.5, 7.1, 2, "rabbit", 123 + 1E−10, "thestral")',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 'cat', 12.001, 5, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 2, 'rabbit', 123 + 1E-10, 'thestral']));

  Test('RemoveDuplicates(L)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 123 + 1E-10, 'thestral']));

  Test('RemoveDuplicates(L, 2E-12)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 123 + 1E-10, 'thestral']));

  Test('RemoveDuplicates(L, 2E-11)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 123 + 1E-10, 'thestral']));

  Test('RemoveDuplicates(L, 2E-10)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2E-9)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2E-8)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2E-7)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2E-6)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 5.00001, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2E-5)',
    asoarr([3, 2, 'cat', 5, 6, 3.0001, 7, 12, 12.001, 'dog', 'rat', 6.99, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2E-4)',
    asoarr([3, 2, 'cat', 5, 6, 7, 12, 12.001, 'dog', 'rat', 6.99, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2E-3)',
    asoarr([3, 2, 'cat', 5, 6, 7, 12, 'dog', 'rat', 6.99, 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2E-2)',
    asoarr([3, 2, 'cat', 5, 6, 7, 12, 'dog', 'rat', 123, 124, 124.5, 7.1, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2E-1)',
    asoarr([3, 2, 'cat', 5, 6, 7, 12, 'dog', 'rat', 123, 124, 124.5, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 1.4)',
    asoarr([3, 'cat', 5, 7, 12, 'dog', 'rat', 123, 124.5, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 2.4)',
    asoarr([3, 'cat', 6, 12, 'dog', 'rat', 123, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 5)',
    asoarr([3, 'cat', 12, 'dog', 'rat', 123, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 10)',
    asoarr([3, 'cat', 'dog', 'rat', 123, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 100)',
    asoarr([3, 'cat', 'dog', 'rat', 123, 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 1000)',
    asoarr([3, 'cat', 'dog', 'rat', 'rabbit', 'thestral']));

  Test('RemoveDuplicates(L, 1E20)',
    asoarr([3, 'cat', 'dog', 'rat', 'rabbit', 'thestral']));

  Test('L ≔ ''(2 + 3⋅i, 5 − i, 2 + 3.001⋅i, 4.99 − i, 12, 2 + 3⋅i)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12), ASO(2 + 3*ImaginaryUnit)]));

  Test('RemoveDuplicates(L)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 2E-6)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 2E-5)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 2E-4)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(2 + 3.001*ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 2E-3)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASO(4.99 - ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 2E-2)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 2)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 3)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASO(5 - ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 4.2)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 9)',
    asoarrex([ASO(2 + 3*ImaginaryUnit), ASOInt(12)]));

  Test('RemoveDuplicates(L, 10.1)',
    asoarrex([ASO(2 + 3*ImaginaryUnit)]));

  Test('RemoveDuplicates(L, 1E50)',
    asoarrex([ASO(2 + 3*ImaginaryUnit)]));

  // adjacent duplicates

  Test('RemoveAdjacentDuplicates(❨4, 2, 3, 3, 4, 1, 5, 8, 0, 0, 0, 4, 7, 8, 9, 5, 5, 5, 3, 2, 4, 5, 1, 9, 5, 2❩)',
    [4, 2, 3, 4, 1, 5, 8, 0, 4, 7, 8, 9, 5, 3, 2, 4, 5, 1, 9, 5, 2]);

  Test('RemoveAdjacentDuplicates(❨4, 2, 3, 3, 4, 1, 5, 8, 0, 0, 0, 4, 7, 8, 9, 5, 5, 5, 3, 2, 4, 5, 1, 9, 5, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2❩)',
    [4, 2, 3, 4, 1, 5, 8, 0, 4, 7, 8, 9, 5, 3, 2, 4, 5, 1, 9, 5, 2]);

  Test('RemoveAdjacentDuplicates(❨1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 4, 2, 3, 3, 4, 1, 5, 8, 0, 0, 0, 4, 7, 8, 9, 5, 5, 5, 3, 2, 4, 5, 1, 9, 5, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2❩)',
    [1, 4, 2, 3, 4, 1, 5, 8, 0, 4, 7, 8, 9, 5, 3, 2, 4, 5, 1, 9, 5, 2]);

  Test('RemoveAdjacentDuplicates(❨3, 2, 0, 5, 8, 7, 4, 1, 0, 6, 3, 2, 5, 8, 6, 4, 7, 9, 1, 2, 3, 0, 5, 5, 8, 9❩)',
    [3, 2, 0, 5, 8, 7, 4, 1, 0, 6, 3, 2, 5, 8, 6, 4, 7, 9, 1, 2, 3, 0, 5, 8, 9]);

  Test('RemoveAdjacentDuplicates(❨1, 1, 1, 1, 1, 1, 1, 2, 1, 2, 2, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 1❩)',
    [1, 2, 1, 2, 3, 4, 1]);

  Test('RemoveAdjacentDuplicates(❨21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21❩)',
    [21]);

  Test('RemoveAdjacentDuplicates(ZeroVector(1000000))',
    [0]);

  Test('RemoveAdjacentDuplicates(BasisVector(1000000, 123456))',
    [0, 1, 0]);

  Test('RemoveAdjacentDuplicates(❨3, 6, 9, 2, 5, i, 4, 1, 5, 5, 2, 1, 4, 0, i, i, −i, 7, 0, 0, 4, 8, 2, 2, 3, i❩)',
    [3, 6, 9, 2, 5, ImaginaryUnit, 4, 1, 5, 2, 1, 4, 0, ImaginaryUnit, -ImaginaryUnit, 7, 0, 4, 8, 2, 3, ImaginaryUnit]);

  Test('RemoveAdjacentDuplicates(❨3, 6, 9, 2, 5, i, 4, 1, 5, 5, 2, 1, 4, 0, i, i, i, −i, 7, 0, 0, 4, 8, 2, 2, 3, i❩)',
    [3, 6, 9, 2, 5, ImaginaryUnit, 4, 1, 5, 2, 1, 4, 0, ImaginaryUnit, -ImaginaryUnit, 7, 0, 4, 8, 2, 3, ImaginaryUnit]);

  Test('RemoveAdjacentDuplicates(❨3, 6, 9, 2, 5, i, 4, 1, 5, 5, 5, 5, 5, 5, 5, 2, 1, 4, 0, i, i, i, −i, 7, 0, 0, 4, 8, 2, 2, 2, 2, 2, 2, 2, 2, 3, i❩)',
    [3, 6, 9, 2, 5, ImaginaryUnit, 4, 1, 5, 2, 1, 4, 0, ImaginaryUnit, -ImaginaryUnit, 7, 0, 4, 8, 2, 3, ImaginaryUnit]);

  Test('RemoveAdjacentDuplicates(❨3, 3, 3, 3, 3, 6, 9, 2, 5, i, 4, 1, 5, 5, 5, 5, 5, 5, 5, 2, 1, 4, 0, i, i, i, −i, 7, 0, 0, 4, 8, 2, 2, 2, 2, 2, 2, 2, 2, 3, i, i, i, i, i, i, i❩)',
    [3, 6, 9, 2, 5, ImaginaryUnit, 4, 1, 5, 2, 1, 4, 0, ImaginaryUnit, -ImaginaryUnit, 7, 0, 4, 8, 2, 3, ImaginaryUnit]);

  Test('RemoveAdjacentDuplicates(❨5, 5, 5, 5, 5, i, i, i, i, i, i, 5, 5, 5, i, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5❩)',
    [5, ImaginaryUnit, 5, ImaginaryUnit, 5]);

  Test('RemoveAdjacentDuplicates(❨5, 5, 5, 5, 5, i, i, i, i, i, i, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5❩)', // 16:55
    [5, ImaginaryUnit, 5]);

  Test('RemoveAdjacentDuplicates(❨5, 5, 5, 5, 5, i, i, i, i, i, i, i, i❩)',
    [5, ImaginaryUnit]);

  Test('RemoveAdjacentDuplicates(ComplexVector(❨5, 5, 5, 5, 5❩))',
    [TASC(5)]);

  Test('RemoveAdjacentDuplicates(ComplexVector(❨5, 5❩))',
    [TASC(5)]);

  Test('RemoveAdjacentDuplicates(ComplexVector(❨5❩))',
    [TASC(5)]);

  Test('RemoveAdjacentDuplicates(ComplexZeroVector(1000000))',
    [TASC(0)]);

  Test('RemoveAdjacentDuplicates(ComplexVector(BasisVector(1000000, 456123)))',
    [TASC(0), 1, 0]);

  Test('RemoveAdjacentDuplicates(ZeroMatrix(10))', failure,
    'Cannot remove adjacent duplicates from an object of type "real matrix".');

  Test('RemoveAdjacentDuplicates(ComplexZeroMatrix(10))', failure,
    'Cannot remove adjacent duplicates from an object of type "complex matrix".');

  Test('RemoveAdjacentDuplicates("Hello, World!")', 'Helo, World!');
  Test('RemoveAdjacentDuplicates("This is nothing but a test.")', 'This is nothing but a test.');
  Test('RemoveAdjacentDuplicates("He fell off his broomstick.")', 'He fel of his bromstick.');
  Test('RemoveAdjacentDuplicates("Aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...")', 'Aa.');
  Test('RemoveAdjacentDuplicates("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")', 'a');
  Test('RemoveAdjacentDuplicates("Skräptyg AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAB")', 'Skräptyg AB');
  Test('RemoveAdjacentDuplicates("")', '');

  Test('RemoveAdjacentDuplicates(''(4, 2, 1, 5, 5, 3, 3, 2, 6, 8, 9, 4, 1, 1, 1, 1, 0, 2, 5, 1, 10))',
    intarr([4, 2, 1, 5, 3, 2, 6, 8, 9, 4, 1, 0, 2, 5, 1, 10]));

  Test('RemoveAdjacentDuplicates(''(4, 2, 1, 5, 5, 3, 3, 2, 6, 8, 9, 4, 1, 1, 1, 1, 0, 2, 5, 1, 10, 10, 10, 10, 10, 10, 10, 10))',
    intarr([4, 2, 1, 5, 3, 2, 6, 8, 9, 4, 1, 0, 2, 5, 1, 10]));

  Test('RemoveAdjacentDuplicates(''(4, 4, 4, 4, 4, 4, 2, 1, 5, 5, 3, 3, 2, 6, 8, 9, 4, 1, 1, 1, 1, 0, 2, 5, 1, 10, 10, 10, 10, 10, 10, 10, 10))',
    intarr([4, 2, 1, 5, 3, 2, 6, 8, 9, 4, 1, 0, 2, 5, 1, 10]));

  Test('RemoveAdjacentDuplicates(''(4, 4, 4, 4, 4, 4, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 5, 5, 3, 3, 2, 6, 8, 9, 4, 1, 1, 1, 1, 0, 2, 5, 1, 10, 10, 10, 10, 10, 10, 10, 10))',
    intarr([4, 2, 1, 5, 3, 2, 6, 8, 9, 4, 1, 0, 2, 5, 1, 10]));

  Test('RemoveAdjacentDuplicates(''(4, 4, 4, 4, 4, 4, 2))',
    intarr([4, 2]));

  Test('RemoveAdjacentDuplicates(''(4, 4, 4, 4, 4, 4))',
    intarr([4]));

  Test('RemoveAdjacentDuplicates(''(4, 4, 4, 4, 4))',
    intarr([4]));

  Test('RemoveAdjacentDuplicates(''(4, 4, 4, 4))',
    intarr([4]));

  Test('RemoveAdjacentDuplicates(''(4, 4, 4))',
    intarr([4]));

  Test('RemoveAdjacentDuplicates(''(4, 4))',
    intarr([4]));

  Test('RemoveAdjacentDuplicates(''(4))',
    intarr([4]));

  Test('RemoveAdjacentDuplicates(''())',
    intarr([]));

  Test('RemoveAdjacentDuplicates(''(1, 2, 2, 2, 3, 4, 5, 5/2 + 5/2, 5.0, 5 + i - i, 6, 7, 8, 9, 10, 10))',
    intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]));

  Test('RemoveAdjacentDuplicates({5, 3, 0, 4, 1, 2, 2, 5, 7, 8, 9})', failure,
    'Cannot remove adjacent duplicates from an object of type "set".');

  Test('RemoveAdjacentDuplicates(3.14)', failure,
    'Cannot remove adjacent duplicates from an object of type "real number".');

  Test('RemoveAdjacentDuplicates(1033)', failure,
    'Cannot remove adjacent duplicates from an object of type "integer".');

  Test('RemoveAdjacentDuplicates(true)', failure,
    'Cannot remove adjacent duplicates from an object of type "boolean".');

  // adjacent duplicates with epsilon

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩)',
    [5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E-8, 4, 3.99, 3.999, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2E-10)',
    [5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E-8, 4, 3.99, 3.999, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2E-9)',
    [5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E-8, 4, 3.99, 3.999, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2E-8)',
    [5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E-8, 3.99, 3.999, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2E-7)',
    [5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E-8, 3.99, 3.999, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2E-6)',
    [5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E-8, 3.99, 3.999, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2E-5)',
    [5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E-8, 3.99, 3.999, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2E-4)',
    [5, 1, 2, 3, 2, 7, 4, 8, 7.999, 8.01, 4 + 1E-8, 3.99, 3.999, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2E-3)',
    [5, 1, 2, 3, 2, 7, 4, 8, 8.01, 4 + 1E-8, 3.99, 3.999, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2E-2)',
    [5, 1, 2, 3, 2, 7, 4, 8, 4 + 1E-8, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 0.5)',
    [5, 1, 2, 3, 2, 7, 4, 8, 4 + 1E-8, 5, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 1.5)',
    [5, 1, 3, 7, 4, 8, 4 + 1E-8, 0, 2]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 2.5)',
    [5, 1, 7, 4, 8, 4 + 1E-8, 0]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 5.5)',
    [5]);

  Test('RemoveAdjacentDuplicates(❨5, 1, 2, 3, 2, 2.0001, 7, 4, 8, 7.999, 8.01, 4 + 1E−8, 4, 3.99, 3.999, 5, 0, 0, 2❩, 10000000)',
    [5]);

  Test('RemoveAdjacentDuplicates(❨1, 0.99, 1.001❩)',
    [1, 0.99, 1.001]);

  Test('RemoveAdjacentDuplicates(❨1, 0.99, 1.001❩, 2E−4)',
    [1, 0.99, 1.001]);

  Test('RemoveAdjacentDuplicates(❨1, 0.99, 1.001❩, 2E−3)',
    [1, 0.99, 1.001]);

  Test('RemoveAdjacentDuplicates(❨1, 0.99, 1.001❩, 2E−2)',
    [1]);

  Test('RemoveAdjacentDuplicates(❨1, 0.99, 1.001❩, 200)',
    [1]);

  Test('RemoveAdjacentDuplicates(❨1❩, 200)',
    [1]);

  Test('RemoveAdjacentDuplicates(DebugObject("empty vector"), 0.001)',
    TAlgosimRealVector.Create);

  Test('RemoveAdjacentDuplicates(❨1, 2 + i, 1.99 + i, 1 + i/10000, 1, 5, 4.9, 4.9❩)',
    [1, 2 + ImaginaryUnit, 1.99 + ImaginaryUnit, 1 + 0.0001*ImaginaryUnit, 1, 5, 4.9]);

  Test('RemoveAdjacentDuplicates(❨1, 2 + i, 1.99 + i, 1 + i/10000, 1, 5, 4.9, 4.9❩, 2E-6)',
    [1, 2 + ImaginaryUnit, 1.99 + ImaginaryUnit, 1 + 0.0001*ImaginaryUnit, 1, 5, 4.9]);

  Test('RemoveAdjacentDuplicates(❨1, 2 + i, 1.99 + i, 1 + i/10000, 1, 5, 4.9, 4.9❩, 2E-5)',
    [1, 2 + ImaginaryUnit, 1.99 + ImaginaryUnit, 1 + 0.0001*ImaginaryUnit, 1, 5, 4.9]);

  Test('RemoveAdjacentDuplicates(❨1, 2 + i, 1.99 + i, 1 + i/10000, 1, 5, 4.9, 4.9❩, 2E-4)',
    [1, 2 + ImaginaryUnit, 1.99 + ImaginaryUnit, 1 + 0.0001*ImaginaryUnit, 5, 4.9]);

  Test('RemoveAdjacentDuplicates(❨1, 2 + i, 1.99 + i, 1 + i/10000, 1, 5, 4.9, 4.9❩, 2E-3)',
    [1, 2 + ImaginaryUnit, 1.99 + ImaginaryUnit, 1 + 0.0001*ImaginaryUnit, 5, 4.9]);

  Test('RemoveAdjacentDuplicates(❨1, 2 + i, 1.99 + i, 1 + i/10000, 1, 5, 4.9, 4.9❩, 2E-2)',
    [1, 2 + ImaginaryUnit, 1 + 0.0001*ImaginaryUnit, 5, 4.9]);

  Test('RemoveAdjacentDuplicates(❨1, 2 + i, 1.99 + i, 1 + i/10000, 1, 5, 4.9, 4.9❩, 2E-1)',
    [1, 2 + ImaginaryUnit, 1 + 0.0001*ImaginaryUnit, 5]);

  Test('RemoveAdjacentDuplicates(❨1, 2 + i, 1.99 + i, 1 + i/10000, 1, 5, 4.9, 4.9❩, 1.5)',
    [TASC(1), 5]);

  Test('RemoveAdjacentDuplicates(❨1, 2 + i, 1.99 + i, 1 + i/10000, 1, 5, 4.9, 4.9❩, 15)',
    [TASC(1)]);

  Test('RemoveAdjacentDuplicates(ComplexVector(DebugObject("empty vector")), 0.001)',
    TAlgosimComplexVector.Create);

  Test('RemoveAdjacentDuplicates(ZeroMatrix(10), 1E-6)', failure,
    'Cannot remove adjacent duplicates from an object of type "real matrix".');

  Test('RemoveAdjacentDuplicates(ComplexZeroMatrix(10), 1E-6)', failure,
    'Cannot remove adjacent duplicates from an object of type "complex matrix".');

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5))',
    asoarr([5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 1, 0, 5]));

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5), 2E-6)',
    asoarr([5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 1, 0, 5]));

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5), 2E-5)',
    asoarr([5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 1, 0, 5]));

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5), 2E-4)',
    asoarr([5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 1, 0, 5]));

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5), 2E-3)',
    asoarr([5, 3, 2, 1, 7, 6, 10, 9.99, 6, 1, 0, 5]));

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5), 2E-2)',
    asoarr([5, 3, 2, 1, 7, 6, 10, 6, 1, 0, 5]));

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5), 1.5)',
    asoarr([5, 3, 1, 7, 10, 6, 1, 5]));

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5), 2.5)',
    asoarr([5, 2, 7, 10, 6, 1, 5]));

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5), 5.5)',
    asoarr([5]));

  Test('RemoveAdjacentDuplicates(''(5, 3, 2, 1, 0.999, 0.9999, 7, 7.001, 6, 10, 9.99, 6, 6, 1, 0, 5, 5), 10!)',
    asoarr([5]));

  Test('RemoveAdjacentDuplicates(''(3, 7, "cat", 7, 6.999, 3, "dog", "dog", 4, 4.0001, "rat", 4, 4, 1))',
    asoarr([3, 7, 'cat', 7, 6.999, 3, 'dog', 4, 4.0001, 'rat', 4, 1]));

  Test('RemoveAdjacentDuplicates(''(3, 7, "cat", 7, 6.999, 3, "dog", "dog", 4, 4.0001, "rat", 4, 4, 1), 2E-6)',
    asoarr([3, 7, 'cat', 7, 6.999, 3, 'dog', 4, 4.0001, 'rat', 4, 1]));

  Test('RemoveAdjacentDuplicates(''(3, 7, "cat", 7, 6.999, 3, "dog", "dog", 4, 4.0001, "rat", 4, 4, 1), 2E-5)',
    asoarr([3, 7, 'cat', 7, 6.999, 3, 'dog', 4, 4.0001, 'rat', 4, 1]));

  Test('RemoveAdjacentDuplicates(''(3, 7, "cat", 7, 6.999, 3, "dog", "dog", 4, 4.0001, "rat", 4, 4, 1), 2E-4)',
    asoarr([3, 7, 'cat', 7, 6.999, 3, 'dog', 4, 'rat', 4, 1]));

  Test('RemoveAdjacentDuplicates(''(3, 7, "cat", 7, 6.999, 3, "dog", "dog", 4, 4.0001, "rat", 4, 4, 1), 2E-3)',
    asoarr([3, 7, 'cat', 7, 3, 'dog', 4, 'rat', 4, 1]));

  Test('RemoveAdjacentDuplicates(''(3, 7, "cat", 7, 6.999, 3, "dog", "dog", 4, 4.0001, "rat", 4, 4, 1), 100)',
    asoarr([3, 'cat', 7, 'dog', 4, 'rat', 4]));

  Test('RemoveAdjacentDuplicates(''(2, 2, 2), 1E-3)', intarr([2]));

  Test('RemoveAdjacentDuplicates(''(2, 2), 1E-3)', intarr([2]));

  Test('RemoveAdjacentDuplicates(''(2), 1E-3)', intarr([2]));

  Test('RemoveAdjacentDuplicates(''(), 1E-3)', intarr([]));

  Test('RemoveAdjacentDuplicates({1, 2, 3, 3, 4, 5, 6}, 1E-6)', failure,
    'Cannot remove adjacent duplicates from an object of type "set".');

  // trick cases

  Test('RemoveAdjacentDuplicates(❨8, 0, 5, 2, 6, −10❩, 3.5)', [8, 0, 5, -10]);
  Test('RemoveAdjacentDuplicates(❨8, i, 5, 2, 6, −10❩, 3.5)', [8, ImaginaryUnit, 5, -10]);
  Test('RemoveAdjacentDuplicates(''(8, 0, 5, 2, 6, −10), 3.5)', intarr([8, 0, 5, -10]));


  //
  // Frequencies
  //

  Test('frequencies(❨5, 3, 0, 4, 1, 2, 7, 8, 5, 2, 3, 6, 9, 2, 0, 1, 4, 5, 2, 0, 10, 8, 2, 6, 0, 4, 1, 2, 3, 6, 9, 5, 3, 0, 2, 1, 7, 4, 5, 2, 0, 3❩)',
    asoarrex(
      [
        freq(5, 5),
        freq(3, 5),
        freq(0, 6),
        freq(4, 4),
        freq(1, 4),
        freq(2, 8),
        freq(7, 2),
        freq(8, 2),
        freq(6, 3),
        freq(9, 2),
        freq(10, 1)
      ]
    )
  );

  Test('frequencies(❨1, 2, 3, 4, 5, 6, 7, 8, 9, 0❩)',
    asoarrex(
      [
        freq(1, 1),
        freq(2, 1),
        freq(3, 1),
        freq(4, 1),
        freq(5, 1),
        freq(6, 1),
        freq(7, 1),
        freq(8, 1),
        freq(9, 1),
        freq(0, 1)
      ]
    )
  );

  Test('frequencies(❨1, 1, 1, 1, 1, 1, 1, 1, 1, 1❩)',
    asoarrex(
      [
        freq(1, 10)
      ]
    )
  );

  Test('frequencies(❨1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1❩)',
    asoarrex(
      [
        freq(1, 10),
        freq(2, 1)
      ]
    )
  );

  Test('frequencies(ZeroVector(1000000))',
    asoarrex(
      [
        freq(0, 1000000)
      ]
    )
  );

  Test('frequencies(BasisVector(1000000, 654321))',
    asoarrex(
      [
        freq(0, 999999),
        freq(1, 1)
      ]
    )
  );

  Test('frequencies(DebugObject("empty vector"))',
    asoarrex(
      [
      ]
    )
  );

  Test('frequencies(❨5, 3, 0, i, 1, 2, 7, 8, 5, 2, 3, 6, 9, 2, 0, 1, 4, 5, 2, 0, 10, 8, 2, 6, 0, 4, 1, 2, 3, 6, 9, 5, 3, 0, 2, 1, 7, 4, 5, 2, 0, 3❩)',
    asoarrex(
      [
        freq(5, 5),
        freq(3, 5),
        freq(0, 6),
        freq(ImaginaryUnit, 1),
        freq(1, 4),
        freq(2, 8),
        freq(7, 2),
        freq(8, 2),
        freq(6, 3),
        freq(9, 2),
        freq(4, 3),
        freq(10, 1)
      ]
    )
  );

  Test('frequencies(❨1, 2, 3, 4, 5, 6, 7, 8, 9, i❩)',
    asoarrex(
      [
        freq(1, 1),
        freq(2, 1),
        freq(3, 1),
        freq(4, 1),
        freq(5, 1),
        freq(6, 1),
        freq(7, 1),
        freq(8, 1),
        freq(9, 1),
        freq(ImaginaryUnit, 1)
      ]
    )
  );

  Test('frequencies(❨i, i, i, i, i, i, i, i, i, i❩)',
    asoarrex(
      [
        freq(ImaginaryUnit, 10)
      ]
    )
  );

  Test('frequencies(❨i, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1❩)',
    asoarrex(
      [
        freq(ImaginaryUnit, 1),
        freq(1, 9),
        freq(2, 1)
      ]
    )
  );

  Test('frequencies(ComplexZeroVector(1000000))',
    asoarrex(
      [
        freq(TASC(0), 1000000)
      ]
    )
  );

  Test('frequencies(ComplexVector(BasisVector(1000000, 654321)))',
    asoarrex(
      [
        freq(TASC(0), 999999),
        freq(TASC(1), 1)
      ]
    )
  );

  Test('frequencies(ComplexVector(DebugObject("empty vector")))',
    asoarrex(
      [
      ]
    )
  );

  Test('frequencies(❨❨6, 2, 5, 4, 1, 2, 0, 1❩, ❨4, 5, 2, 0, 1, 4, 2, 5❩, '+'❨9, 5, 3, 2, 0, 1, 4, 7❩, ❨5, 5, 6, 0, 1, 4, 2, 1❩, ❨6, 2, 0, 5, 4, 8, 1, 0❩, ❨8, 2, 1, 4, 5, 4, 2, 1❩, ❨4, 1, 2, 1, 0, 2, 1, 1❩, ❨3, 2, 8, 2, 4, 5, 1, 0❩❩)',
    asoarrex(
      [
        freq(6, 3),
        freq(2, 13),
        freq(5, 9),
        freq(4, 10),
        freq(1, 14),
        freq(0, 8),
        freq(9, 1),
        freq(3, 2),
        freq(7, 1),
        freq(8, 3)
      ]
    )
  );

  Test('frequencies(❨❨6, 2, 5, 4, 1, 2, 0, 1❩❩)',
    asoarrex(
      [
        freq(6, 1),
        freq(2, 2),
        freq(5, 1),
        freq(4, 1),
        freq(1, 2),
        freq(0, 1)
      ]
    )
  );

  Test('frequencies(ZeroMatrix(1000))',
    asoarrex(
      [
        freq(0, 1000000)
      ]
    )
  );

  Test('frequencies(IdentityMatrix(1000))',
    asoarrex(
      [
        freq(1, 1000),
        freq(0, 1000000 - 1000)
      ]
    )
  );

  Test('frequencies(CirculantMatrix(❨1, 2, 3, 4, 5, 6, 7, 8, 9❩))',
    asoarrex(
      [
        freq(1, 9),
        freq(2, 9),
        freq(3, 9),
        freq(4, 9),
        freq(5, 9),
        freq(6, 9),
        freq(7, 9),
        freq(8, 9),
        freq(9, 9)
      ]
    )
  );

  Test('frequencies(ToeplitzMatrix(❨1, 2, 3, 4, 5, 6, 7, 8, 9❩, ❨1, 0, 0, 0, 0, 0, 0, 0, 0❩))',
    asoarrex(
      [
        freq(1, 9),
        freq(2, 8),
        freq(3, 7),
        freq(4, 6),
        freq(5, 5),
        freq(6, 4),
        freq(7, 3),
        freq(8, 2),
        freq(9, 1),
        freq(0, 8+7+6+5+4+3+2+1)
      ]
    )
  );

  Test('frequencies(❨❨21❩❩)',
    asoarrex(
      [
        freq(21, 1)
      ]
    )
  );

  Test('frequencies(DebugObject("empty matrix"))',
    asoarrex(
      [
      ]
    )
  );

  Test('frequencies(❨❨3, 6, 5, 2, 1, i❩, ❨4, 2, 0, i, 1, 4❩, ❨5, 3, 2, 5, 9, i❩, ❨4, 1, 2, 0, 1, −i❩❩)',
    asoarrex(
      [
        freq(3, 2),
        freq(6, 1),
        freq(5, 3),
        freq(2, 4),
        freq(1, 4),
        freq(ImaginaryUnit, 3),
        freq(4, 3),
        freq(0, 2),
        freq(9, 1),
        freq(-ImaginaryUnit, 1)
      ]
    )
  );

  Test('frequencies(❨❨i, i, i, i, i, i❩❩)',
    asoarrex(
      [
        freq(ImaginaryUnit, 6)
      ]
    )
  );

  Test('frequencies(ComplexZeroMatrix(1000))',
    asoarrex(
      [
        freq(TASC(0), 1000000)
      ]
    )
  );

  Test('frequencies(ComplexMatrix(IdentityMatrix(1000)))',
    asoarrex(
      [
        freq(TASC(1), 1000),
        freq(TASC(0), 1000000 - 1000)
      ]
    )
  );

  Test('frequencies(CirculantMatrix(❨i, 2, 3, 4, 5, 6, 7, 8, 9❩))',
    asoarrex(
      [
        freq(ImaginaryUnit, 9),
        freq(2, 9),
        freq(3, 9),
        freq(4, 9),
        freq(5, 9),
        freq(6, 9),
        freq(7, 9),
        freq(8, 9),
        freq(9, 9)
      ]
    )
  );

  Test('frequencies(ToeplitzMatrix(❨i, 2, 3, 4, 5, 6, 7, 8, 9❩, ❨i, 0, 0, 0, 0, 0, 0, 0, 0❩))',
    asoarrex(
      [
        freq(ImaginaryUnit, 9),
        freq(2, 8),
        freq(3, 7),
        freq(4, 6),
        freq(5, 5),
        freq(6, 4),
        freq(7, 3),
        freq(8, 2),
        freq(9, 1),
        freq(0, 8+7+6+5+4+3+2+1)
      ]
    )
  );

  Test('frequencies(ComplexMatrix(❨❨21❩❩))',
    asoarrex(
      [
        freq(TASC(21), 1)
      ]
    )
  );

  Test('frequencies(ComplexMatrix(DebugObject("empty matrix")))',
    asoarrex(
      [
      ]
    )
  );

  Test('frequencies(''(5, 3, 2, 0, 1, 4, 5, 7, 8, 9, 6, 5, 2, 3, 2, 1, 4, 5, 2, 0, 7, 2, 5, 8, 9, 6, 2, 1, 4, 5, 3, 2, 4, i − i, 5, 2, 0.0, 1, 4, 1, 0, 1/2 − 1/2, 2, 1, 0, 1, 2, 0, 1, 0, 1, 0, 1))',
    asoarrex(
      [
        freq(5, 7),
        freq(3, 3),
        freq(2, 10),
        freq(0, 10),
        freq(1, 10),
        freq(4, 5),
        freq(7, 2),
        freq(8, 2),
        freq(9, 2),
        freq(6, 2)
      ]
    )
  );

  Test('frequencies(''(1, 1, 1, 1, 1))',
    asoarrex(
      [
        freq(1, 5)
      ]
    )
  );

  Test('frequencies(''(1))',
    asoarrex(
      [
        freq(1, 1)
      ]
    )
  );

  Test('frequencies(''())',
    asoarrex(
      [
      ]
    )
  );

  Test('frequencies(''(1, "cat", 5, 1, 3, 2, "dog", "rat", 1, 2, 3, 1, 2, "dog", 1, 2, 3, "rabbit", "cat", "rat", 1, 2, 3, "guinea pig", "dog", 1, 2, 3, "dog"))',
    asoarrex(
      [
        freq(1, 7),
        freq('cat', 2),
        freq(5, 1),
        freq(3, 5),
        freq(2, 6),
        freq('dog', 4),
        freq('rat', 2),
        freq('rabbit', 1),
        freq('guinea pig', 1)
      ]
    )
  );

  Test('frequencies({1, "cat", 5, 1, 3, 2, "dog", "rat", 1, 2, 3, 1, 2, "dog", 1, 2, 3, "rabbit", "cat", "rat", 1, 2, 3, "guinea pig", "dog", 1, 2, 3, "dog"})',
    failure, 'Cannot compute member frequencies in an object of type "set".');

  Test('frequencies("Hello, World!")', failure, 'Cannot compute member frequencies in an object of type "string".');

  Test('frequencies(words("Hello, World!"))',
    asoarrex(
      [
        freq('Hello', 1),
        freq('World', 1)
      ]
    )
  );

  Test('frequencies(characters("Hello, World!"))',
    asoarrex(
      [
        freq('H', 1),
        freq('e', 1),
        freq('l', 3),
        freq('o', 2),
        freq(',', 1),
        freq(' ', 1),
        freq('W', 1),
        freq('r', 1),
        freq('d', 1),
        freq('!', 1)
      ]
    )
  );

  Test('frequencies(date())', failure, 'Cannot compute member frequencies in an object of type "typed structure".');
  Test('frequencies(true)', failure, 'Cannot compute member frequencies in an object of type "boolean".');
  Test('frequencies(123)', failure, 'Cannot compute member frequencies in an object of type "integer".');
  Test('frequencies(3.1415)', failure, 'Cannot compute member frequencies in an object of type "real number".');

  Test('frequencies()', failure, 'There is no version of this function that can take 0 argument(s).');

  // with epsilon

  Test('frequencies(❨1, 5, 3, 2, 5.0001, 1, 4, 8, 4.999, 1.01, 1, 10, 2.1❩)',
    asoarrex(
      [
        freq(1, 3),
        freq(5, 1),
        freq(3, 1),
        freq(2, 1),
        freq(5.0001, 1),
        freq(4, 1),
        freq(8, 1),
        freq(4.999, 1),
        freq(1.01, 1),
        freq(10, 1),
        freq(2.1, 1)
      ]
    )
  );

  Test('frequencies(❨1, 5, 3, 2, 5.0001, 1, 4, 8, 4.999, 1.01, 1, 10, 2.1❩, 2E-6)',
    asoarrex(
      [
        freq(1, 3),
        freq(5, 1),
        freq(3, 1),
        freq(2, 1),
        freq(5.0001, 1),
        freq(4, 1),
        freq(8, 1),
        freq(4.999, 1),
        freq(1.01, 1),
        freq(10, 1),
        freq(2.1, 1)
      ]
    )
  );

  Test('frequencies(❨1, 5, 3, 2, 5.0001, 1, 4, 8, 4.999, 1.01, 1, 10, 2.1❩, 2E-5)',
    asoarrex(
      [
        freq(1, 3),
        freq(5, 1),
        freq(3, 1),
        freq(2, 1),
        freq(5.0001, 1),
        freq(4, 1),
        freq(8, 1),
        freq(4.999, 1),
        freq(1.01, 1),
        freq(10, 1),
        freq(2.1, 1)
      ]
    )
  );

  Test('frequencies(❨1, 5, 3, 2, 5.0001, 1, 4, 8, 4.999, 1.01, 1, 10, 2.1❩, 2E-4)',
    asoarrex(
      [
        freq(1, 3),
        freq(5, 2),
        freq(3, 1),
        freq(2, 1),
        freq(4, 1),
        freq(8, 1),
        freq(4.999, 1),
        freq(1.01, 1),
        freq(10, 1),
        freq(2.1, 1)
      ]
    )
  );

  Test('frequencies(❨1, 5, 3, 2, 5.0001, 1, 4, 8, 4.999, 1.01, 1, 10, 2.1❩, 2E-3)',
    asoarrex(
      [
        freq(1, 3),
        freq(5, 3),
        freq(3, 1),
        freq(2, 1),
        freq(4, 1),
        freq(8, 1),
        freq(1.01, 1),
        freq(10, 1),
        freq(2.1, 1)
      ]
    )
  );

  Test('frequencies(❨1, 5, 3, 2, 5.0001, 1, 4, 8, 4.999, 1.01, 1, 10, 2.1❩, 2E-2)',
    asoarrex(
      [
        freq(1, 4),
        freq(5, 3),
        freq(3, 1),
        freq(2, 1),
        freq(4, 1),
        freq(8, 1),
        freq(10, 1),
        freq(2.1, 1)
      ]
    )
  );

  Test('frequencies(❨1, 5, 3, 2, 5.0001, 1, 4, 8, 4.999, 1.01, 1, 10, 2.1❩, 2E-1)',
    asoarrex(
      [
        freq(1, 4),
        freq(5, 3),
        freq(3, 1),
        freq(2, 2),
        freq(4, 1),
        freq(8, 1),
        freq(10, 1)
      ]
    )
  );

  Test('frequencies(❨1, 5, 3, 2, 5.0001, 1, 4, 8, 4.999, 1.01, 1, 10, 2.1❩, 1.5)',
    asoarrex(
      [
        freq(1, 6),
        freq(5, 4),
        freq(3, 1),
        freq(8, 1),
        freq(10, 1)
      ]
    )
  );

  Test('frequencies(❨1, 5, 3, 2, 5.0001, 1, 4, 8, 4.999, 1.01, 1, 10, 2.1❩, 10)',
    asoarrex(
      [
        freq(1, 13)
      ]
    )
  );

  Test('frequencies(❨1, 1, 1❩, 10)',
    asoarrex(
      [
        freq(1, 3)
      ]
    )
  );

  Test('frequencies(ZeroVector(1000), 0.001)',
    asoarrex(
      [
        freq(0, 1000)
      ]
    )
  );

  Test('frequencies(BasisVector(1000, 452), 0.001)',
    asoarrex(
      [
        freq(0, 1000 - 1),
        freq(1, 1)
      ]
    )
  );

  Test('frequencies(DebugObject("empty vector"), 0.001)',
    asoarrex(
      [
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 2),
        freq(1 + ImaginaryUnit, 1),
        freq(3, 1),
        freq(5.0001, 1),
        freq(0.999, 1),
        freq(1 + 0.99*ImaginaryUnit, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩, 2E-6)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 2),
        freq(1 + ImaginaryUnit, 1),
        freq(3, 1),
        freq(5.0001, 1),
        freq(0.999, 1),
        freq(1 + 0.99*ImaginaryUnit, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩, 2E-5)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 2),
        freq(1 + ImaginaryUnit, 1),
        freq(3, 1),
        freq(5.0001, 1),
        freq(0.999, 1),
        freq(1 + 0.99*ImaginaryUnit, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩, 2E-4)',
    asoarrex(
      [
        freq(5, 2),
        freq(1, 2),
        freq(1 + ImaginaryUnit, 1),
        freq(3, 1),
        freq(0.999, 1),
        freq(1 + 0.99*ImaginaryUnit, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩, 2E-3)',
    asoarrex(
      [
        freq(5, 2),
        freq(1, 3),
        freq(1 + ImaginaryUnit, 1),
        freq(3, 1),
        freq(1 + 0.99*ImaginaryUnit, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩, 2E-2)',
    asoarrex(
      [
        freq(5, 2),
        freq(1, 3),
        freq(1 + ImaginaryUnit, 2),
        freq(3, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩, 2E-1)',
    asoarrex(
      [
        freq(5, 3),
        freq(1, 3),
        freq(1 + ImaginaryUnit, 2),
        freq(3, 1)
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩, 1.5)',
    asoarrex(
      [
        freq(5, 3),
        freq(1, 5),
        freq(3, 1)
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩, 2.5)',
    asoarrex(
      [
        freq(5, 4),
        freq(1, 5)
      ]
    )
  );

  Test('frequencies(❨5, 1, 1 + i, 3, 5.0001, 0.999, 1 + 0.99⋅i, 1, 5.1❩, 5)',
    asoarrex(
      [
        freq(5, 9)
      ]
    )
  );

  Test('frequencies(❨i, i, i❩, 10)',
    asoarrex(
      [
        freq(ImaginaryUnit, 3)
      ]
    )
  );

  Test('frequencies(ComplexZeroVector(1000), 0.001)',
    asoarrex(
      [
        freq(TASC(0), 1000)
      ]
    )
  );

  Test('frequencies(ComplexVector(BasisVector(1000, 452)), 0.001)',
    asoarrex(
      [
        freq(TASC(0), 1000 - 1),
        freq(TASC(1), 1)
      ]
    )
  );

  Test('frequencies(ComplexVector(DebugObject("empty vector")), 0.001)',
    asoarrex(
      [
      ]
    )
  );

  Test('frequencies(❨❨2, 5, 2.001, 1.99❩, ❨5, 5.0001, 5.1, 2❩❩)',
    asoarrex(
      [
        freq(2, 2),
        freq(5, 2),
        freq(2.001, 1),
        freq(1.99, 1),
        freq(5.0001, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨❨2, 5, 2.001, 1.99❩, ❨5, 5.0001, 5.1, 2❩❩, 2E-6)',
    asoarrex(
      [
        freq(2, 2),
        freq(5, 2),
        freq(2.001, 1),
        freq(1.99, 1),
        freq(5.0001, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨❨2, 5, 2.001, 1.99❩, ❨5, 5.0001, 5.1, 2❩❩, 2E-5)',
    asoarrex(
      [
        freq(2, 2),
        freq(5, 2),
        freq(2.001, 1),
        freq(1.99, 1),
        freq(5.0001, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨❨2, 5, 2.001, 1.99❩, ❨5, 5.0001, 5.1, 2❩❩, 2E-4)',
    asoarrex(
      [
        freq(2, 2),
        freq(5, 3),
        freq(2.001, 1),
        freq(1.99, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨❨2, 5, 2.001, 1.99❩, ❨5, 5.0001, 5.1, 2❩❩, 2E-3)',
    asoarrex(
      [
        freq(2, 3),
        freq(5, 3),
        freq(1.99, 1),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨❨2, 5, 2.001, 1.99❩, ❨5, 5.0001, 5.1, 2❩❩, 2E-2)',
    asoarrex(
      [
        freq(2, 4),
        freq(5, 3),
        freq(5.1, 1)
      ]
    )
  );

  Test('frequencies(❨❨2, 5, 2.001, 1.99❩, ❨5, 5.0001, 5.1, 2❩❩, 2E-1)',
    asoarrex(
      [
        freq(2, 4),
        freq(5, 4)
      ]
    )
  );

  Test('frequencies(❨❨2, 5, 2.001, 1.99❩, ❨5, 5.0001, 5.1, 2❩❩, 3.5)',
    asoarrex(
      [
        freq(2, 8)
      ]
    )
  );

  Test('frequencies(ZeroMatrix(100), 0.001)',
    asoarrex(
      [
        freq(0, 10000)
      ]
    )
  );

  Test('frequencies(IdentityMatrix(100), 0.001)',
    asoarrex(
      [
        freq(1, 100),
        freq(0, 10000 - 100)
      ]
    )
  );

  Test('frequencies(CirculantMatrix(❨1, 2, 3, 4, 5, 6, 7, 8, 9❩), 0.5)',
    asoarrex(
      [
        freq(1, 9),
        freq(2, 9),
        freq(3, 9),
        freq(4, 9),
        freq(5, 9),
        freq(6, 9),
        freq(7, 9),
        freq(8, 9),
        freq(9, 9)
      ]
    )
  );

  Test('frequencies(CirculantMatrix(❨1, 2, 3, 4, 5, 6, 7, 8, 9❩), 1.2)',
    asoarrex(
      [
        freq(1, 18),
        freq(3, 18),
        freq(5, 18),
        freq(7, 18),
        freq(9, 9)
      ]
    )
  );

  Test('frequencies(ToeplitzMatrix(❨1, 2, 3, 4, 5, 6, 7, 8, 9❩, ❨1, 0, 0, 0, 0, 0, 0, 0, 0❩), 0.5)',
    asoarrex(
      [
        freq(1, 9),
        freq(2, 8),
        freq(3, 7),
        freq(4, 6),
        freq(5, 5),
        freq(6, 4),
        freq(7, 3),
        freq(8, 2),
        freq(9, 1),
        freq(0, 8+7+6+5+4+3+2+1)
      ]
    )
  );

  Test('frequencies(ToeplitzMatrix(❨1, 2, 3, 4, 5, 6, 7, 8, 9❩, ❨1, 0, 0, 0, 0, 0, 0, 0, 0❩), 1.5)',
    asoarrex(
      [
        freq(1, 9 + 8 + 8+7+6+5+4+3+2+1),
        freq(3, 7 + 6),
        freq(5, 5 + 4),
        freq(7, 3 + 2),
        freq(9, 1)
      ]
    )
  );

  Test('frequencies(DebugObject("empty matrix"), 0.001)',
    asoarrex(
      [
      ]
    )
  );

  Test('frequencies(❨❨4, i, 1 + i, 5❩, ❨3.999, 1 + 0.99⋅i, 5.0001, 4❩❩)',
    asoarrex(
      [
        freq(4, 2),
        freq(ImaginaryUnit, 1),
        freq(1 + ImaginaryUnit, 1),
        freq(5, 1),
        freq(3.999, 1),
        freq(1 + 0.99*ImaginaryUnit, 1),
        freq(5.0001, 1)
      ]
    )
  );

  Test('frequencies(❨❨4, i, 1 + i, 5❩, ❨3.999, 1 + 0.99⋅i, 5.0001, 4❩❩, 2E-6)',
    asoarrex(
      [
        freq(4, 2),
        freq(ImaginaryUnit, 1),
        freq(1 + ImaginaryUnit, 1),
        freq(5, 1),
        freq(3.999, 1),
        freq(1 + 0.99*ImaginaryUnit, 1),
        freq(5.0001, 1)
      ]
    )
  );

  Test('frequencies(❨❨4, i, 1 + i, 5❩, ❨3.999, 1 + 0.99⋅i, 5.0001, 4❩❩, 2E-5)',
    asoarrex(
      [
        freq(4, 2),
        freq(ImaginaryUnit, 1),
        freq(1 + ImaginaryUnit, 1),
        freq(5, 1),
        freq(3.999, 1),
        freq(1 + 0.99*ImaginaryUnit, 1),
        freq(5.0001, 1)
      ]
    )
  );

  Test('frequencies(❨❨4, i, 1 + i, 5❩, ❨3.999, 1 + 0.99⋅i, 5.0001, 4❩❩, 2E-4)',
    asoarrex(
      [
        freq(4, 2),
        freq(ImaginaryUnit, 1),
        freq(1 + ImaginaryUnit, 1),
        freq(5, 2),
        freq(3.999, 1),
        freq(1 + 0.99*ImaginaryUnit, 1)
      ]
    )
  );

  Test('frequencies(❨❨4, i, 1 + i, 5❩, ❨3.999, 1 + 0.99⋅i, 5.0001, 4❩❩, 2E-3)',
    asoarrex(
      [
        freq(4, 3),
        freq(ImaginaryUnit, 1),
        freq(1 + ImaginaryUnit, 1),
        freq(5, 2),
        freq(1 + 0.99*ImaginaryUnit, 1)
      ]
    )
  );

  Test('frequencies(❨❨4, i, 1 + i, 5❩, ❨3.999, 1 + 0.99⋅i, 5.0001, 4❩❩, 2E-2)',
    asoarrex(
      [
        freq(4, 3),
        freq(ImaginaryUnit, 1),
        freq(1 + ImaginaryUnit, 2),
        freq(5, 2)
      ]
    )
  );

  Test('frequencies(❨❨4, i, 1 + i, 5❩, ❨3.999, 1 + 0.99⋅i, 5.0001, 4❩❩, 2E-1)',
    asoarrex(
      [
        freq(4, 3),
        freq(ImaginaryUnit, 1),
        freq(1 + ImaginaryUnit, 2),
        freq(5, 2)
      ]
    )
  );

  Test('frequencies(❨❨4, i, 1 + i, 5❩, ❨3.999, 1 + 0.99⋅i, 5.0001, 4❩❩, 1.5)',
    asoarrex(
      [
        freq(4, 5),
        freq(ImaginaryUnit, 3)
      ]
    )
  );

  Test('frequencies(❨❨4, i, 1 + i, 5❩, ❨3.999, 1 + 0.99⋅i, 5.0001, 4❩❩, 5.5)',
    asoarrex(
      [
        freq(4, 8)
      ]
    )
  );

  Test('frequencies(ComplexZeroMatrix(100), 0.001)',
    asoarrex(
      [
        freq(0, 10000)
      ]
    )
  );

  Test('frequencies(ComplexMatrix(IdentityMatrix(100)), 0.001)',
    asoarrex(
      [
        freq(1, 100),
        freq(0, 10000 - 100)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7))',
    asoarrex(
      [
        freq(5, 1),
        freq(2, 2),
        freq('rat', 4),
        freq('dog', 2),
        freq(1.9999, 1),
        freq('cat', 1),
        freq(2.001, 1),
        freq('rabbit', 1),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(10, 1),
        freq(4.7, 1)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7), 2E-6)',
    asoarrex(
      [
        freq(5, 1),
        freq(2, 2),
        freq('rat', 4),
        freq('dog', 2),
        freq(1.9999, 1),
        freq('cat', 1),
        freq(2.001, 1),
        freq('rabbit', 1),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(10, 1),
        freq(4.7, 1)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7), 2E-5)',
    asoarrex(
      [
        freq(5, 1),
        freq(2, 2),
        freq('rat', 4),
        freq('dog', 2),
        freq(1.9999, 1),
        freq('cat', 1),
        freq(2.001, 1),
        freq('rabbit', 1),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(10, 1),
        freq(4.7, 1)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7), 2E-4)',
    asoarrex(
      [
        freq(5, 1),
        freq(2, 3),
        freq('rat', 4),
        freq('dog', 2),
        freq('cat', 1),
        freq(2.001, 1),
        freq('rabbit', 1),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(10, 1),
        freq(4.7, 1)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7), 2E-3)',
    asoarrex(
      [
        freq(5, 1),
        freq(2, 4),
        freq('rat', 4),
        freq('dog', 2),
        freq('cat', 1),
        freq('rabbit', 1),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(10, 1),
        freq(4.7, 1)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7), 2E-2)',
    asoarrex(
      [
        freq(5, 2),
        freq(2, 4),
        freq('rat', 4),
        freq('dog', 2),
        freq('cat', 1),
        freq('rabbit', 1),
        freq(5.1, 1),
        freq(10, 1),
        freq(4.7, 1)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7), 2E-1)',
    asoarrex(
      [
        freq(5, 3),
        freq(2, 4),
        freq('rat', 4),
        freq('dog', 2),
        freq('cat', 1),
        freq('rabbit', 1),
        freq(10, 1),
        freq(4.7, 1)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7), 4E-1)',
    asoarrex(
      [
        freq(5, 4),
        freq(2, 4),
        freq('rat', 4),
        freq('dog', 2),
        freq('cat', 1),
        freq('rabbit', 1),
        freq(10, 1)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7), 10)',
    asoarrex(
      [
        freq(5, 9),
        freq('rat', 4),
        freq('dog', 2),
        freq('cat', 1),
        freq('rabbit', 1)
      ]
    )
  );

  Test('frequencies(''(5, 2, "rat", 2, "dog", 1.9999, "cat", "dog", 2.001, "rat", "rabbit", 5.01, "rat", 5.1, 10, "rat", 4.7), 1E100)',
    asoarrex(
      [
        freq(5, 9),
        freq('rat', 4),
        freq('dog', 2),
        freq('cat', 1),
        freq('rabbit', 1)
      ]
    )
  );

  Test('frequencies(''(5, 5, 5), 0.1)',
    asoarrex(
      [
        freq(5, 3)
      ]
    )
  );

  Test('frequencies(''(), 0.1)',
    asoarrex(
      [
      ]
    )
  );

  Test('frequencies({5, 1, 2}, 1E-6)', failure,
    'Cannot compute member frequencies in an object of type "set".');

  Test('frequencies("nargle", 1E-6)', failure,
    'Cannot compute member frequencies in an object of type "string".');

  Test('frequencies(394, 1E-6)', failure,
    'Cannot compute member frequencies in an object of type "integer".');

  // adjacent (collapse sequencies)

  Test('collapse(❨5, 3, 2, 2, 1, 0, 2, 2, 2, 0, 4, 1, 2, 1, 1, 0, 3, 6❩)',
    asoarrex(
      [
        freq(5, 1),
        freq(3, 1),
        freq(2, 2),
        freq(1, 1),
        freq(0, 1),
        freq(2, 3),
        freq(0, 1),
        freq(4, 1),
        freq(1, 1),
        freq(2, 1),
        freq(1, 2),
        freq(0, 1),
        freq(3, 1),
        freq(6, 1)
      ]
    )
  );

  Test('collapse(❨5, 5, 5, 5, 5, 5, 2, 2, 2, 0, 4, 1, 2, 1, 1, 0, 3, 3❩)',
    asoarrex(
      [
        freq(5, 6),
        freq(2, 3),
        freq(0, 1),
        freq(4, 1),
        freq(1, 1),
        freq(2, 1),
        freq(1, 2),
        freq(0, 1),
        freq(3, 2)
      ]
    )
  );

  Test('collapse(❨5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 2, 2, 2, 2, 2, 2❩)',
    asoarrex(
      [
        freq(5, 12),
        freq(2, 6)
      ]
    )
  );

  Test('collapse(❨5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5❩)',
    asoarrex(
      [
        freq(5, 12)
      ]
    )
  );

  Test('collapse(❨5❩)',
    asoarrex(
      [
        freq(5, 1)
      ]
    )
  );

  Test('collapse(ZeroVector(1000000))',
    asoarrex(
      [
        freq(0, 1000000)
      ]
    )
  );

  Test('collapse(BasisVector(1000000, 654321))',
    asoarrex(
      [
        freq(0, 654320),
        freq(1, 1),
        freq(0, 1000000 - 654320 - 1)
      ]
    )
  );

  Test('collapse(DebugObject("empty vector"))',
    asoarrex(
      [
      ]
    )
  );

  Test('collapse(❨5, 3, i, i, 1, 0, 2, 2, 2, 0, 4, 1, 2, 1, 1, 0, 3, 6❩)',
    asoarrex(
      [
        freq(5, 1),
        freq(3, 1),
        freq(ImaginaryUnit, 2),
        freq(1, 1),
        freq(0, 1),
        freq(2, 3),
        freq(0, 1),
        freq(4, 1),
        freq(1, 1),
        freq(2, 1),
        freq(1, 2),
        freq(0, 1),
        freq(3, 1),
        freq(6, 1)
      ]
    )
  );

  Test('collapse(❨5, 5, 5, 5, 5, 5, i, i, i, 0, 4, 1, 2, 1, 1, 0, 3, 3❩)',
    asoarrex(
      [
        freq(5, 6),
        freq(ImaginaryUnit, 3),
        freq(0, 1),
        freq(4, 1),
        freq(1, 1),
        freq(2, 1),
        freq(1, 2),
        freq(0, 1),
        freq(3, 2)
      ]
    )
  );

  Test('collapse(❨5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, i, i, i, i, i, i❩)',
    asoarrex(
      [
        freq(5, 12),
        freq(ImaginaryUnit, 6)
      ]
    )
  );

  Test('collapse(❨i, i, i, i, i, i, i, i, i, i, i, i❩)',
    asoarrex(
      [
        freq(ImaginaryUnit, 12)
      ]
    )
  );

  Test('collapse(❨i❩)',
    asoarrex(
      [
        freq(ImaginaryUnit, 1)
      ]
    )
  );

  Test('collapse(ComplexZeroVector(1000000))',
    asoarrex(
      [
        freq(TASC(0), 1000000)
      ]
    )
  );

  Test('collapse(ComplexVector(BasisVector(1000000, 654321)))',
    asoarrex(
      [
        freq(0, 654320),
        freq(1, 1),
        freq(0, 1000000 - 654320 - 1)
      ]
    )
  );

  Test('collapse(ComplexVector(DebugObject("empty vector")))',
    asoarrex(
      [
      ]
    )
  );

  Test('collapse(❨❨3, 2, 2, 5❩, ❨5, 5, 5, 1❩❩)', failure,
    'Cannot compute adjacent member counts in an object of type "real matrix".');

  Test('collapse(❨❨3, 2, 2, 5❩, ❨5, 5, 5, i❩❩)', failure,
    'Cannot compute adjacent member counts in an object of type "complex matrix".');

  Test('collapse(''(3, 2, 2, 2, "rat", "cat", "dog", "dog", "cat", 6, 2, 1, true, true, false, true, true, true, true, 0, 1, 4, 4, 1, 1, 1))',
    asoarrex(
      [
        freq(3, 1),
        freq(2, 3),
        freq('rat', 1),
        freq('cat', 1),
        freq('dog', 2),
        freq('cat', 1),
        freq(6, 1),
        freq(2, 1),
        freq(1, 1),
        freq(True, 2),
        freq(False, 1),
        freq(True, 4),
        freq(0, 1),
        freq(1, 1),
        freq(4, 2),
        freq(1, 3)
      ]
    )
  );

  Test('collapse(''(3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, "rat", "cat", "dog", "dog", "cat", 6, 2, 1, true, true, false, true, true, true, true, 0, 1, 4, 4, 1, 1, 1))',
    asoarrex(
      [
        freq(3, 12),
        freq(2, 3),
        freq('rat', 1),
        freq('cat', 1),
        freq('dog', 2),
        freq('cat', 1),
        freq(6, 1),
        freq(2, 1),
        freq(1, 1),
        freq(True, 2),
        freq(False, 1),
        freq(True, 4),
        freq(0, 1),
        freq(1, 1),
        freq(4, 2),
        freq(1, 3)
      ]
    )
  );

  Test('collapse(''(3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2))',
    asoarrex(
      [
        freq(3, 12),
        freq(2, 3)
      ]
    )
  );

  Test('collapse(''(3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3))',
    asoarrex(
      [
        freq(3, 12)
      ]
    )
  );

  Test('collapse(''(3, 3))',
    asoarrex(
      [
        freq(3, 2)
      ]
    )
  );

  Test('collapse(''(3))',
    asoarrex(
      [
        freq(3, 1)
      ]
    )
  );

  Test('collapse(''())',
    asoarrex(
      [
      ]
    )
  );

  // adjacent (collapse sequencies) with epsilon

  Test('collapse(❨6, 5, 2, 4, 4, 4.001, 7, 6, 5, 4.9999, 5.01, 2❩)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(4, 2),
        freq(4.001, 1),
        freq(7, 1),
        freq(6, 1),
        freq(5, 1),
        freq(4.9999, 1),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, 4, 4, 4.001, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-6)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(4, 2),
        freq(4.001, 1),
        freq(7, 1),
        freq(6, 1),
        freq(5, 1),
        freq(4.9999, 1),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, 4, 4, 4.001, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-5)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(4, 2),
        freq(4.001, 1),
        freq(7, 1),
        freq(6, 1),
        freq(5, 1),
        freq(4.9999, 1),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, 4, 4, 4.001, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-4)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(4, 2),
        freq(4.001, 1),
        freq(7, 1),
        freq(6, 1),
        freq(5, 2),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, 4, 4, 4.001, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-3)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(4, 3),
        freq(7, 1),
        freq(6, 1),
        freq(5, 2),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, 4, 4, 4.001, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-2)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(4, 3),
        freq(7, 1),
        freq(6, 1),
        freq(5, 3),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, 4, 4, 4.001, 7, 6, 5, 4.9999, 5.01, 2❩, 1.5)',
    asoarrex(
      [
        freq(6, 2),
        freq(2, 1),
        freq(4, 3),
        freq(7, 2),
        freq(5, 3),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, 4, 4, 4.001, 7, 6, 5, 4.9999, 5.01, 2❩, 5.5)',
    asoarrex(
      [
        freq(6, 12)
      ]
    )
  );

  Test('collapse(ZeroVector(1000000), 0.01)',
    asoarrex(
      [
        freq(0, 1000000)
      ]
    )
  );

  Test('collapse(BasisVector(1000000, 654321), 0.01)',
    asoarrex(
      [
        freq(0, 654320),
        freq(1, 1),
        freq(0, 1000000 - 654320 - 1)
      ]
    )
  );

  Test('collapse(DebugObject("empty vector"), 0.01)',
    asoarrex(
      [
      ]
    )
  );

  Test('collapse(❨6, 5, 2, i, i, 1.001⋅i, 7, 6, 5, 4.9999, 5.01, 2❩)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(ImaginaryUnit, 2),
        freq(Imaginaryunit * 1.001, 1),
        freq(7, 1),
        freq(6, 1),
        freq(5, 1),
        freq(4.9999, 1),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, i, i, 1.001⋅i, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-6)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(ImaginaryUnit, 2),
        freq(Imaginaryunit * 1.001, 1),
        freq(7, 1),
        freq(6, 1),
        freq(5, 1),
        freq(4.9999, 1),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, i, i, 1.001⋅i, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-5)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(ImaginaryUnit, 2),
        freq(Imaginaryunit * 1.001, 1),
        freq(7, 1),
        freq(6, 1),
        freq(5, 1),
        freq(4.9999, 1),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, i, i, 1.001⋅i, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-4)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(ImaginaryUnit, 2),
        freq(Imaginaryunit * 1.001, 1),
        freq(7, 1),
        freq(6, 1),
        freq(5, 2),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, i, i, 1.001⋅i, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-3)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(ImaginaryUnit, 3),
        freq(7, 1),
        freq(6, 1),
        freq(5, 2),
        freq(5.01, 1),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, i, i, 1.001⋅i, 7, 6, 5, 4.9999, 5.01, 2❩, 2E-2)',
    asoarrex(
      [
        freq(6, 1),
        freq(5, 1),
        freq(2, 1),
        freq(ImaginaryUnit, 3),
        freq(7, 1),
        freq(6, 1),
        freq(5, 3),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, i, i, 1.001⋅i, 7, 6, 5, 4.9999, 5.01, 2❩, 1.5)',
    asoarrex(
      [
        freq(6, 2),
        freq(2, 1),
        freq(ImaginaryUnit, 3),
        freq(7, 2),
        freq(5, 3),
        freq(2, 1)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, i, i, 1.001⋅i, 7, 6, 5, 4.9999, 5.01, 2❩, 5.5)',
    asoarrex(
      [
        freq(6, 3),
        freq(ImaginaryUnit, 3),
        freq(7, 6)
      ]
    )
  );

  Test('collapse(❨6, 5, 2, i, i, 1.001⋅i, 7, 6, 5, 4.9999, 5.01, 2❩, 6.5)',
    asoarrex(
      [
        freq(6, 12)
      ]
    )
  );

  Test('collapse(ComplexZeroVector(1000000), 0.01)',
    asoarrex(
      [
        freq(0, 1000000)
      ]
    )
  );

  Test('collapse(ComplexVector(BasisVector(1000000, 654321)), 0.01)',
    asoarrex(
      [
        freq(0, 654320),
        freq(1, 1),
        freq(0, 1000000 - 654320 - 1)
      ]
    )
  );

  Test('collapse(ComplexVector(DebugObject("empty vector")), 0.01)',
    asoarrex(
      [
      ]
    )
  );

  Test('collapse(❨❨3, 2, 2, 5❩, ❨5, 5, 5, 1❩❩, 0.001)', failure,
    'Cannot compute adjacent member counts in an object of type "real matrix".');

  Test('collapse(❨❨3, 2, 2, 5❩, ❨5, 5, 5, i❩❩, 0.001)', failure,
    'Cannot compute adjacent member counts in an object of type "complex matrix".');

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9))',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 1),
        freq(2, 2),
        freq(2.0001, 1),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 1),
        freq(7, 1),
        freq(6.9, 1)
      ]
    )
  );

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9), 2E-6)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 1),
        freq(2, 2),
        freq(2.0001, 1),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 1),
        freq(7, 1),
        freq(6.9, 1)
      ]
    )
  );

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9), 2E-5)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 1),
        freq(2, 2),
        freq(2.0001, 1),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 1),
        freq(7, 1),
        freq(6.9, 1)
      ]
    )
  );

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9), 2E-4)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 1),
        freq(2, 3),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 1),
        freq(7, 1),
        freq(6.9, 1)
      ]
    )
  );

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9), 2E-3)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 1),
        freq(2, 3),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 1),
        freq(7, 1),
        freq(6.9, 1)
      ]
    )
  );

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9), 2E-2)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 1),
        freq(2, 3),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 1),
        freq(7, 1),
        freq(6.9, 1)
      ]
    )
  );

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9), 2E-1)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 1),
        freq(2, 3),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 1),
        freq(7, 2)
      ]
    )
  );

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9), 2.5)',
    asoarrex(
      [
        freq(5, 1),
        freq(1, 4),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 3)
      ]
    )
  );

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9), 5.5)',
    asoarrex(
      [
        freq(5, 5),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 3)
      ]
    )
  );

  Test('collapse(''(5, 1, 2, 2, 2.0001, "rat", "rat", "dog", 6, 7, 6.9), 1000000.001)',
    asoarrex(
      [
        freq(5, 5),
        freq('rat', 2),
        freq('dog', 1),
        freq(6, 3)
      ]
    )
  );

  Test('collapse(''(5, 5, 5, 5.0001, 5.001, 5.01, 5.1, 5))',
    asoarrex(
      [
        freq(5, 3),
        freq(5.0001, 1),
        freq(5.001, 1),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(5, 1)
      ]
    )
  );

  Test('collapse(''(5, 5, 5, 5.0001, 5.001, 5.01, 5.1, 5), 2E-6)',
    asoarrex(
      [
        freq(5, 3),
        freq(5.0001, 1),
        freq(5.001, 1),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(5, 1)
      ]
    )
  );

  Test('collapse(''(5, 5, 5, 5.0001, 5.001, 5.01, 5.1, 5), 2E-5)',
    asoarrex(
      [
        freq(5, 3),
        freq(5.0001, 1),
        freq(5.001, 1),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(5, 1)
      ]
    )
  );

  Test('collapse(''(5, 5, 5, 5.0001, 5.001, 5.01, 5.1, 5), 2E-4)',
    asoarrex(
      [
        freq(5, 4),
        freq(5.001, 1),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(5, 1)
      ]
    )
  );

  Test('collapse(''(5, 5, 5, 5.0001, 5.001, 5.01, 5.1, 5), 2E-3)',
    asoarrex(
      [
        freq(5, 5),
        freq(5.01, 1),
        freq(5.1, 1),
        freq(5, 1)
      ]
    )
  );

  Test('collapse(''(5, 5, 5, 5.0001, 5.001, 5.01, 5.1, 5), 2E-2)',
    asoarrex(
      [
        freq(5, 6),
        freq(5.1, 1),
        freq(5, 1)
      ]
    )
  );

  Test('collapse(''(5, 5, 5, 5.0001, 5.001, 5.01, 5.1, 5), 2E-1)',
    asoarrex(
      [
        freq(5, 8)
      ]
    )
  );

  Test('collapse(''(5, 5, 5, 5.0001, 5.001, 5.01, 5.1, 5), 123)',
    asoarrex(
      [
        freq(5, 8)
      ]
    )
  );

  Test('collapse(''(5, 5, 5), 0.1)',
    asoarrex(
      [
        freq(5, 3)
      ]
    )
  );

  Test('collapse(''(), 0.1)',
    asoarrex(
      [
      ]
    )
  );

  Test('collapse({1, 2, 2, 3}, 0.1)', failure, 'Cannot compute adjacent member counts in an object of type "set".');

  // trick cases

  Test('collapse(❨5, 8, 8.05, 7.92❩, 0.1)',
    asoarrex(
      [
        freq(5, 1),
        freq(8, 3)
      ]
    )
  );

  Test('collapse(ComplexVector(❨5, 8, 8.05, 7.92❩), 0.1)',
    asoarrex(
      [
        freq(5, 1),
        freq(8, 3)
      ]
    )
  );

  Test('collapse(''(5, 8, 8.05, 7.92), 0.1)',
    asoarrex(
      [
        freq(5, 1),
        freq(8, 3)
      ]
    )
  );


  //
  // SequenceList
  //

  Test('SequenceList(-100000000)',
    failure, 'A positive integer was expected as argument 1, but "-100000000" was given.');

  Test('SequenceList(-1)',
    failure, 'A positive integer was expected as argument 1, but "-1" was given.');

  Test('SequenceList(0)',
    failure, 'A positive integer was expected as argument 1, but "0" was given.');

  Test('SequenceList(1)',
    intarr([1]));

  Test('SequenceList(2)',
    intarr([1, 2]));

  Test('SequenceList(3)',
    intarr([1, 2, 3]));

  Test('SequenceList(4)',
    intarr([1, 2, 3, 4]));

  Test('SequenceList(5)',
    intarr([1, 2, 3, 4, 5]));

  Test('SequenceList(5, 8)',
    intarr([5, 6, 7, 8]));

  Test('SequenceList(10, 19)',
    intarr([10, 11, 12, 13, 14, 15, 16, 17, 18, 19]));

  Test('SequenceList(100, 110)',
    intarr([100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110]));

  Test('SequenceList(0, 10)',
    intarr([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]));

  Test('SequenceList(1, 10)',
    intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]));

  Test('SequenceList(-10, 10)',
    intarr([-10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]));

  Test('SequenceList(-99, -90)',
    intarr([-99, -98, -97, -96, -95, -94, -93, -92, -91, -90]));

  Test('SequenceList(7, 7)',
    intarr([7]));

  Test('SequenceList(0, 0)',
    intarr([0]));

  Test('SequenceList(-1, 1)',
    intarr([-1, 0, 1]));

  Test('SequenceList(1, 2)',
    intarr([1, 2]));

  Test('SequenceList(1, 3)',
    intarr([1, 2, 3]));

  Test('SequenceList(5, 2)',
    intarr([5, 4, 3, 2]));

  Test('SequenceList(50, 40)',
    intarr([50, 49, 48, 47, 46, 45, 44, 43, 42, 41, 40]));

  Test('SequenceList(10, 0)',
    intarr([10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0]));

  Test('SequenceList(10, 1)',
    intarr([10, 9, 8, 7, 6, 5, 4, 3, 2, 1]));

  Test('SequenceList(10, -10)',
    intarr([10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10]));

  Test('SequenceList(0, -10)',
    intarr([0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10]));

  Test('SequenceList(-1, -10)',
    intarr([-1, -2, -3, -4, -5, -6, -7, -8, -9, -10]));

  Test('SequenceList(-1, -2)',
    intarr([-1, -2]));

  Test('SequenceList(0, 10, 2)',
    intarr([0, 2, 4, 6, 8, 10]));

  Test('SequenceList(0, 10, 3)',
    intarr([0, 3, 6, 9]));

  Test('SequenceList(0, 10, 4)',
    intarr([0, 4, 8]));

  Test('SequenceList(0, 10, 5)',
    intarr([0, 5, 10]));

  Test('SequenceList(0, 10, 6)',
    intarr([0, 6]));

  Test('SequenceList(0, 10, 7)',
    intarr([0, 7]));

  Test('SequenceList(0, 10, 8)',
    intarr([0, 8]));

  Test('SequenceList(0, 10, 9)',
    intarr([0, 9]));

  Test('SequenceList(0, 10, 10)',
    intarr([0, 10]));

  Test('SequenceList(0, 10, 11)',
    intarr([0]));

  Test('SequenceList(0, 10, 12)',
    intarr([0]));

  Test('SequenceList(0, 100, 10)',
    intarr([0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]));

  Test('SequenceList(-20, 100, 10)',
    intarr([-20, -10, 0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]));

  Test('SequenceList(-30, 100, 10)',
    intarr([-30, -20, -10, 0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]));

  Test('SequenceList(-30, 100, 100000000)',
    intarr([-30]));

  Test('SequenceList(10, 0, 2)',
    intarr([10, 8, 6, 4, 2, 0]));

  Test('SequenceList(10, -10, 2)',
    intarr([10, 8, 6, 4, 2, 0, -2, -4, -6, -8, -10]));

  Test('SequenceList(10, 0, 3)',
    intarr([10, 7, 4, 1]));

  Test('SequenceList(10, 0, 5)',
    intarr([10, 5, 0]));

  Test('SequenceList(10, 0, 10)',
    intarr([10, 0]));

  Test('SequenceList(10, 0, 11)',
    intarr([10]));

  Test('SequenceList(123, 123, 123)',
    intarr([123]));

  Test('SequenceList(0, 10, 0)',
    failure, 'A positive integer was expected as argument 3, but "0" was given.');

  Test('SequenceList(0, 10, -1)',
    failure, 'A positive integer was expected as argument 3, but "-1" was given.');

  // SequenceVector should behave identically

  Test('SequenceVector(-100000000)',
    failure, 'A positive integer was expected as argument 1, but "-100000000" was given.');

  Test('SequenceVector(-1)',
    failure, 'A positive integer was expected as argument 1, but "-1" was given.');

  Test('SequenceVector(0)',
    failure, 'A positive integer was expected as argument 1, but "0" was given.');

  Test('SequenceVector(1)',
    [1]);

  Test('SequenceVector(2)',
    [1, 2]);

  Test('SequenceVector(3)',
    [1, 2, 3]);

  Test('SequenceVector(4)',
    [1, 2, 3, 4]);

  Test('SequenceVector(5)',
    [1, 2, 3, 4, 5]);

  Test('SequenceVector(5, 8)',
    [5, 6, 7, 8]);

  Test('SequenceVector(10, 19)',
    [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]);

  Test('SequenceVector(100, 110)',
    [100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110]);

  Test('SequenceVector(0, 10)',
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);

  Test('SequenceVector(1, 10)',
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);

  Test('SequenceVector(-10, 10)',
    [-10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);

  Test('SequenceVector(-99, -90)',
    [-99, -98, -97, -96, -95, -94, -93, -92, -91, -90]);

  Test('SequenceVector(7, 7)',
    [7]);

  Test('SequenceVector(0, 0)',
    [0]);

  Test('SequenceVector(-1, 1)',
    [-1, 0, 1]);

  Test('SequenceVector(1, 2)',
    [1, 2]);

  Test('SequenceVector(1, 3)',
    [1, 2, 3]);

  Test('SequenceVector(5, 2)',
    [5, 4, 3, 2]);

  Test('SequenceVector(50, 40)',
    [50, 49, 48, 47, 46, 45, 44, 43, 42, 41, 40]);

  Test('SequenceVector(10, 0)',
    [10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0]);

  Test('SequenceVector(10, 1)',
    [10, 9, 8, 7, 6, 5, 4, 3, 2, 1]);

  Test('SequenceVector(10, -10)',
    [10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10]);

  Test('SequenceVector(0, -10)',
    [0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10]);

  Test('SequenceVector(-1, -10)',
    [-1, -2, -3, -4, -5, -6, -7, -8, -9, -10]);

  Test('SequenceVector(-1, -2)',
    [-1, -2]);

  Test('SequenceVector(0, 10, 2)',
    [0, 2, 4, 6, 8, 10]);

  Test('SequenceVector(0, 10, 3)',
    [0, 3, 6, 9]);

  Test('SequenceVector(0, 10, 4)',
    [0, 4, 8]);

  Test('SequenceVector(0, 10, 5)',
    [0, 5, 10]);

  Test('SequenceVector(0, 10, 6)',
    [0, 6]);

  Test('SequenceVector(0, 10, 7)',
    [0, 7]);

  Test('SequenceVector(0, 10, 8)',
    [0, 8]);

  Test('SequenceVector(0, 10, 9)',
    [0, 9]);

  Test('SequenceVector(0, 10, 10)',
    [0, 10]);

  Test('SequenceVector(0, 10, 11)',
    [0]);

  Test('SequenceVector(0, 10, 12)',
    [0]);

  Test('SequenceVector(0, 100, 10)',
    [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]);

  Test('SequenceVector(-20, 100, 10)',
    [-20, -10, 0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]);

  Test('SequenceVector(-30, 100, 10)',
    [-30, -20, -10, 0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]);

  Test('SequenceVector(-30, 100, 100000000)',
    [-30]);

  Test('SequenceVector(10, 0, 2)',
    [10, 8, 6, 4, 2, 0]);

  Test('SequenceVector(10, -10, 2)',
    [10, 8, 6, 4, 2, 0, -2, -4, -6, -8, -10]);

  Test('SequenceVector(10, 0, 3)',
    [10, 7, 4, 1]);

  Test('SequenceVector(10, 0, 5)',
    [10, 5, 0]);

  Test('SequenceVector(10, 0, 10)',
    [10, 0]);

  Test('SequenceVector(10, 0, 11)',
    [10]);

  Test('SequenceVector(123, 123, 123)',
    [123]);

  Test('SequenceVector(0, 10, 0)',
    failure, 'A positive integer was expected as argument 3, but "0" was given.');

  Test('SequenceVector(0, 10, -1)',
    failure, 'A positive integer was expected as argument 3, but "-1" was given.');


  //
  // Random member
  //

  StartStopwatch;

  Test('v ≔ ❨0, 10, 20, 30, 40, 50❩; S ≔ ToSet(v); N ≔ 100000; x ≔ 0', 0);
  Test('for(i, 1, N, (r ≔ random(v); inc(x, r); assert(r ∈ S)))', null);
  Test('abs(x/N − 25) ≤ 0.5', True);

  Test('v ≔ i⋅❨0, 10, 20, 30, 40, 50❩; S ≔ ToSet(v); N ≔ 100000; x ≔ 0', 0);
  Test('for(i, 1, N, (r ≔ random(v); x ≔ x + r; assert(r ∈ S)))', null);
  Test('abs(x/(N⋅i) − 25) ≤ 0.5', True);

  Test('v ≔ ❨❨0, 10, 20❩, ❨30, 40, 50❩❩; S ≔ ToSet(v); N ≔ 100000; x ≔ 0', 0);
  Test('for(i, 1, N, (r ≔ random(v); inc(x, r); assert(r ∈ S)))', null);
  Test('abs(x/N − 25) ≤ 0.5', True);

  Test('v ≔ i⋅❨❨0, 10, 20❩, ❨30, 40, 50❩❩; S ≔ ToSet(v); N ≔ 100000; x ≔ 0', 0);
  Test('for(i, 1, N, (r ≔ random(v); x ≔ x + r; assert(r ∈ S)))', null);
  Test('abs(x/(N⋅i) − 25) ≤ 0.5', True);

  Test('L ≔ ''(0, 10, 20, 30, 40, 50); S ≔ ToSet(L); N ≔ 100000; x ≔ 0', 0);
  Test('for(i, 1, N, (r ≔ random(L); inc(x, r); assert(r ∈ S)))', null);
  Test('abs(x/N − 25) ≤ 0.5', True);

  Test('S ≔ {0, 10, 20, 30, 40, 50}; N ≔ 100000; x ≔ 0', 0);
  Test('for(i, 1, N, (r ≔ random(S); inc(x, r); assert(r ∈ S)))', null);
  Test('abs(x/N − 25) ≤ 0.5', True);

  StopStopwatch;
  TestDuration(12.5);

  Test('v ≔ ❨0, 10, 20, 30, 40, 50❩; S ≔ ToSet(v); N ≔ 10000; x ≔ 0', 0);
  Test('for(j, 1, N, (r ≔ random(❨0, 10, 20, 30, 40, 50❩); inc(x, r); assert(r ∈ S)))', null);
  Test('abs(x/N − 25) ≤ 5', True);

  Test('v ≔ i⋅❨0, 10, 20, 30, 40, 50❩; S ≔ ToSet(v); N ≔ 10000; x ≔ 0', 0);
  Test('for(j, 1, N, (r ≔ random(i⋅❨0, 10, 20, 30, 40, 50❩); x ≔ x + r; assert(r ∈ S)))', null);
  Test('abs(x/(N⋅i) − 25) ≤ 5', True);

  Test('v ≔ ❨❨0, 10, 20❩, ❨30, 40, 50❩❩; S ≔ ToSet(v); N ≔ 10000; x ≔ 0', 0);
  Test('for(j, 1, N, (r ≔ random(❨❨0, 10, 20❩, ❨30, 40, 50❩❩); inc(x, r); assert(r ∈ S)))', null);
  Test('abs(x/N − 25) ≤ 5', True);

  Test('v ≔ i⋅❨❨0, 10, 20❩, ❨30, 40, 50❩❩; S ≔ ToSet(v); N ≔ 10000; x ≔ 0', 0);
  Test('for(j, 1, N, (r ≔ random(i⋅❨❨0, 10, 20❩, ❨30, 40, 50❩❩); x ≔ x + r; assert(r ∈ S)))', null);
  Test('abs(x/(N⋅i) − 25) ≤ 5', True);

  Test('L ≔ ''(0, 10, 20, 30, 40, 50); S ≔ ToSet(L); N ≔ 10000; x ≔ 0', 0);
  Test('for(j, 1, N, (r ≔ random(''(0, 10, 20, 30, 40, 50)); inc(x, r); assert(r ∈ S)))', null);
  Test('abs(x/N − 25) ≤ 5', True);

  Test('S ≔ {0, 10, 20, 30, 40, 50}; N ≔ 10000; x ≔ 0', 0);
  Test('for(j, 1, N, (r ≔ random({0, 10, 20, 30, 40, 50}); inc(x, r); assert(r ∈ S)))', null);
  Test('abs(x/N − 25) ≤ 5', True);

  Test('for(j, 1, 1000, (v ≔ ❨1, 2, 3, 4, 5❩; random(v) ≔ 1000; assert(count(v, (x ↦ x = 1000)) = 1)))', null);
  Test('for(j, 1, 1000, (v ≔ ComplexVector(❨1, 2, 3, 4, 5❩); random(v) ≔ 1000; assert(count(v, (x ↦ x = 1000)) = 1)))', null);

  Test('for(j, 1, 1000, (v ≔ ❨❨1, 2, 3❩, ❨4, 5, 6❩❩; random(v) ≔ 1000; assert(count(v, (x ↦ x = 1000)) = 1)))', null);
  Test('for(j, 1, 1000, (v ≔ ComplexMatrix(❨❨1, 2, 3❩, ❨4, 5, 6❩❩); random(v) ≔ 1000; assert(count(v, (x ↦ x = 1000)) = 1)))', null);

  Test('for(j, 1, 1000, (v ≔ ''(1, 2, 3, 4, 5); random(v) ≔ 1000; assert(count(v, (x ↦ x = 1000)) = 1)))', null);
//  Test('for(j, 1, 1000, (v ≔ {1, 2, 3, 4, 5}; random(v) ≔ 1000; assert(count(v, (x ↦ x = 1000)) = 1)))', null); // not implemented

  Test('s ≔ "Hello, World!"', 'Hello, World!');
  Test('for(i, 1, 1000, assert(random(s) ∈ ToSet(characters(s))))', null); // you get only the chars of s
  Test('ForEach(characters(s), c, while(random(s) ≠ c, do()))', null); // you get all chars of s

  Test('v ≔ ❨21❩', [21]);
  Test('random(v)', 21.0);
  Test('random(v) ≔ 394', 394);
  Test('v', [394]);

  Test('v ≔ ❨21⋅i❩', [21*ImaginaryUnit]);
  Test('random(v)', 21*ImaginaryUnit);
  Test('random(v) ≔ 394⋅i', 394*ImaginaryUnit);
  Test('v', [394*ImaginaryUnit]);

  Test('A ≔ ❨❨21❩❩', 1, [21]);
  Test('random(A)', 21.0);
  Test('random(A) ≔ 394', 394);
  Test('A', 1, [394]);

  Test('A ≔ ❨❨21⋅i❩❩', 1, [21*ImaginaryUnit]);
  Test('random(A)', 21*ImaginaryUnit);
  Test('random(A) ≔ 394⋅i', 394*ImaginaryUnit);
  Test('A', 1, [394*ImaginaryUnit]);

  Test('L ≔ ''(21)', intarr([21]));
  Test('random(L)', 21);
  Test('random(L) ≔ 394', 394);
  Test('L', intarr([394]));

  Test('s ≔ "A"', 'A');
  Test('random(s)', 'A');
  Test('random(s) ≔ "B"', 'B');
  Test('s', 'B');

  Test('v ≔ DebugObject("empty vector");', null);
  Test('random(v)', failure, 'Index 1 out of bounds.');
  Test('random(v) ≔ 123', failure, 'Index 1 out of bounds.');

  Test('v ≔ ComplexVector(DebugObject("empty vector"));', null);
  Test('random(v)', failure, 'Index 1 out of bounds.');
  Test('random(v) ≔ 123', failure, 'Index 1 out of bounds.');

  Test('A ≔ DebugObject("empty matrix");', null);
  Test('random(A)', failure, 'Index 1 out of bounds.');
  Test('random(A) ≔ 123', failure, 'Index 1 out of bounds.');

  Test('A ≔ ComplexMatrix(DebugObject("empty matrix"));', null);
  Test('random(A)', failure, 'Index 1 out of bounds.');
  Test('random(A) ≔ 123', failure, 'Index 1 out of bounds.');

  Test('L ≔ ''();', null);
  Test('random(L)', failure, 'Index 1 out of bounds.');
  Test('random(L) ≔ 123', failure, 'Index 1 out of bounds.');

  Test('s ≔ "";', null);
  Test('random(s)', failure, 'Index 1 out of bounds.');
  Test('random(s) ≔ "A"', failure, 'Index 1 out of bounds.');


  //
  // Object- and predicate-based queries
  //

  // count

  Test('v ≔ ❨6, 0, −2, 7, 1, 4, 6, 5, 8, −7, 10, 123, −10, 3, 21, 7, 5, 2, 1024, 0❩',
    [6, 0, -2, 7, 1, 4, 6, 5, 8, -7, 10, 123, -10, 3, 21, 7, 5, 2, 1024, 0]);

  Test('count(v, even)', 11);
  Test('count(v, odd)', 9);
  Test('count(v, IsPositive)', 15);
  Test('count(v, IsNonNegative)', 17);
  Test('count(v, IsZero)', 2);
  Test('count(v, IsNonPositive)', 5);
  Test('count(v, IsNegative)', 3);
  Test('count(v, IsPrime)', 6);
  Test('count(v, IsNumber)', 20);
  Test('count(v, IsInteger)', 20);
  Test('count(v, IsBoolean)', 0);
  Test('count(v, x ↦ x ≥ 100)', 2);
  Test('count(v, x ↦ x ≥ 1000)', 1);
  Test('count(v, x ↦ x ≥ 10000)', 0);
  Test('count(v, x ↦ mod(x, 5) = 0)', 6);

  Test('count(❨6, 0, −2, 7, 1, 4, 6, 5, 8, −7, 10, 123, −10, 3, 21, 7, 5, 2, 1024, 0❩, odd)', 9);
  Test('count(DebugObject("empty vector"), odd)', 0);

  Test('v ≔ ❨2 + 3⋅i, 8, 5 − i, 4 + 2⋅i, −1 − i, 10 − 5⋅i, −4 + 6⋅i, 2⋅i, −4, 7 − 2⋅i, 0, −3⋅i, 1 + i, 21❩;', null);

  Test('count(v, IsReal)', 4);
  Test('count(v, IsComplex)', 10);
  Test('count(v, IsNumber)', 14);
  Test('count(v, IsSound)', 0);
  Test('count(v, IsZero)', 1);
  Test('count(v, z ↦ Re(z) = 0)', 3);
  Test('count(v, z ↦ arg(z) ≈ π/4)', 1);
  Test('count(v, z ↦ abs(z) > 10)', 2);
  Test('count(v, z ↦ abs(z) > 5)', 6);

  Test('count(❨2 + 3⋅i, 8, 5 − i, 4 + 2⋅i, −1 − i, 10 − 5⋅i, −4 + 6⋅i, 2⋅i, −4, 7 − 2⋅i, 0, −3⋅i, 1 + i, 21❩, z ↦ abs(z) > 5)', 6);
  Test('count(ComplexVector(DebugObject("empty vector")), z ↦ abs(z) > 5)', 0);

  Test('A ≔ ❨❨6, 0, −2, 7, 1, 4, 6, 5, 8, −7❩, ❨10, 123, −10, 3, 21, 7, 5, 2, 1024, 0❩❩', 10,
    [6, 0, -2, 7, 1, 4, 6, 5, 8, -7, 10, 123, -10, 3, 21, 7, 5, 2, 1024, 0]);

  Test('count(A, even)', 11);
  Test('count(A, odd)', 9);
  Test('count(A, IsPositive)', 15);
  Test('count(A, IsNonNegative)', 17);
  Test('count(A, IsZero)', 2);
  Test('count(A, IsNonPositive)', 5);
  Test('count(A, IsNegative)', 3);
  Test('count(A, IsPrime)', 6);
  Test('count(A, IsNumber)', 20);
  Test('count(A, IsInteger)', 20);
  Test('count(A, IsBoolean)', 0);
  Test('count(A, x ↦ x ≥ 100)', 2);
  Test('count(A, x ↦ x ≥ 1000)', 1);
  Test('count(A, x ↦ x ≥ 10000)', 0);
  Test('count(A, x ↦ mod(x, 5) = 0)', 6);

  Test('count(❨❨6, 0, −2, 7, 1, 4, 6, 5, 8, −7❩, ❨10, 123, −10, 3, 21, 7, 5, 2, 1024, 0❩❩, odd)', 9);
  Test('count(DebugObject("empty matrix"), odd)', 0);

  Test('A ≔ ❨❨2 + 3⋅i, 8, 5 − i, 4 + 2⋅i, −1 − i, 10 − 5⋅i, −4 + 6⋅i❩, ❨2⋅i, −4, 7 − 2⋅i, 0, −3⋅i, 1 + i, 21❩❩;', null);

  Test('count(A, IsReal)', 4);
  Test('count(A, IsComplex)', 10);
  Test('count(A, IsNumber)', 14);
  Test('count(A, IsSound)', 0);
  Test('count(A, IsZero)', 1);
  Test('count(A, z ↦ Re(z) = 0)', 3);
  Test('count(A, z ↦ arg(z) ≈ π/4)', 1);
  Test('count(A, z ↦ abs(z) > 10)', 2);
  Test('count(A, z ↦ abs(z) > 5)', 6);

  Test('count(❨❨2 + 3⋅i, 8, 5 − i, 4 + 2⋅i, −1 − i, 10 − 5⋅i, −4 + 6⋅i❩, ❨2⋅i, −4, 7 − 2⋅i, 0, −3⋅i, 1 + i, 21❩❩, z ↦ abs(z) > 5)', 6);
  Test('count(ComplexMatrix(DebugObject("empty matrix")), z ↦ abs(z) > 5)', 0);

  Test('s ≔ "It is produced from 7-dehydrocholesterol."', 'It is produced from 7-dehydrocholesterol.');
  Test('count(s, ChrIsLetter)', 34);
  Test('count(s, ChrIsWhitespace)', 4);
  Test('count(s, ChrIsPunctuation)', 2);
  Test('count(s, ChrIsDigit)', 1);
  Test('count(s, ChrIsControl)', 0);
  Test('count(s, IsCharacter)', 41);

  Test('count("It is produced from 7-dehydrocholesterol.", ChrIsWhitespace)', 4);
  Test('count("", ChrIsWhitespace)', 0);

  Test('L ≔ ''(21, "cat", "dog", π, i, 1+1=2, now(), 123, 1024, "rat");', null);
  Test('count(L, IsNumber)', 5);
  Test('count(L, IsText)', 3);
  Test('count(L, IsBoolean)', 1);
  Test('count(L, IsStructure)', 1);
  Test('count(L, IsPixmap)', 0);

  Test('count(''(21, "cat", "dog", π, i, 1+1=2, now(), 123, 1024, "rat"), IsNumber)', 5);
  Test('count(''(), IsNumber)', 0);

  Test('S ≔ {21, "cat", "dog", π, i, 1+1=2, now(), 123, 1024, "rat", "rat"};', null);
  Test('count(S, IsNumber)', 5);
  Test('count(S, IsText)', 3);
  Test('count(S, IsBoolean)', 1);
  Test('count(S, IsStructure)', 1);
  Test('count(S, IsPixmap)', 0);

  Test('count({21, "cat", "dog", π, i, 1+1=2, now(), 123, 1024, "rat", "rat"}, IsNumber)', 5);
  Test('count({}, IsNumber)', 0);

  Test('S ≔ struct("name": "Smith", "age": 58, "score": 100);', null);
  Test('count(S, IsText)', 1);
  Test('count(S, IsInteger)', 2);
  Test('count(S, IsPixmap)', 0);
  Test('count(now(), IsStructure)', 2);
  Test('count(struct(), IsStructure)', 0);

  Test('count(123, ChrIsDigit)', failure, 'Cannot count elements in an object of type "integer".');
  Test('count(123.456, ChrIsDigit)', failure, 'Cannot count elements in an object of type "real number".');
  Test('count(i, ChrIsDigit)', failure, 'Cannot count elements in an object of type "complex number".');
  Test('count(true, ChrIsDigit)', failure, 'Cannot count elements in an object of type "boolean".');

  Test('count(❨6, 0, −2, 7, 1, 4, 6, 5, 8, −7, 10.5, 123, −10, 3, 21, 7, 5, 2, 1024, 0❩, odd)',
    failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('v ≔ ❨2, 1, 0, 5, 3, 6, 2, 0, 4, 1, 2, 1, 5, 3, 0, 1, 2, 4, 1, 8❩;', null);

  Test('count(v, 0)', 3);
  Test('count(v, 1)', 5);
  Test('count(v, 2)', 4);
  Test('count(v, 3)', 2);
  Test('count(v, 5)', 2);
  Test('count(v, 6)', 1);
  Test('count(v, 7)', 0);
  Test('count(v, 8)', 1);
  Test('count(v, 9)', 0);

  Test('count(❨2, 1, 0, 5, 3, 6, 2, 0, 4, 1, 2, 1, 5, 3, 0, 1, 2, 4, 1, 8❩, 0)', 3);
  Test('count(DebugObject("empty vector"), 0)', 0);

  Test('v ≔ ❨2, i, 0, 5, 3, 6, 2, 0, 4, i, 2, 1, 5, 3, 0, 1, 2, 4, 1, 8❩;', null);

  Test('count(v, 0)', 3);
  Test('count(v, 1)', 3);
  Test('count(v, 2)', 4);
  Test('count(v, 3)', 2);
  Test('count(v, 5)', 2);
  Test('count(v, 6)', 1);
  Test('count(v, 7)', 0);
  Test('count(v, 8)', 1);
  Test('count(v, 9)', 0);
  Test('count(v, i)', 2);

  Test('count(❨2, i, 0, 5, 3, 6, 2, 0, 4, i, 2, 1, 5, 3, 0, 1, 2, 4, 1, 8❩, 0)', 3);
  Test('count(ComplexVector(DebugObject("empty vector")), 0)', 0);

  Test('v ≔ ❨❨2, 1, 0, 5, 3, 6, 2, 0, 4, 1❩, ❨2, 1, 5, 3, 0, 1, 2, 4, 1, 8❩❩;', null);

  Test('count(v, 0)', 3);
  Test('count(v, 1)', 5);
  Test('count(v, 2)', 4);
  Test('count(v, 3)', 2);
  Test('count(v, 5)', 2);
  Test('count(v, 6)', 1);
  Test('count(v, 7)', 0);
  Test('count(v, 8)', 1);
  Test('count(v, 9)', 0);

  Test('count(❨❨2, 1, 0, 5, 3, 6, 2, 0, 4, 1❩, ❨2, 1, 5, 3, 0, 1, 2, 4, 1, 8❩❩, 0)', 3);
  Test('count(DebugObject("empty matrix"), 0)', 0);

  Test('v ≔ ❨❨2, i, 0, 5, 3, 6, 2, 0, 4, i❩, ❨2, 1, 5, 3, 0, 1, 2, 4, 1, 8❩❩;', null);

  Test('count(v, 0)', 3);
  Test('count(v, 1)', 3);
  Test('count(v, 2)', 4);
  Test('count(v, 3)', 2);
  Test('count(v, 5)', 2);
  Test('count(v, 6)', 1);
  Test('count(v, 7)', 0);
  Test('count(v, 8)', 1);
  Test('count(v, 9)', 0);
  Test('count(v, i)', 2);

  Test('count(❨❨2, i, 0, 5, 3, 6, 2, 0, 4, i❩, ❨2, 1, 5, 3, 0, 1, 2, 4, 1, 8❩❩, 0)', 3);
  Test('count(ComplexMatrix(DebugObject("empty matrix")), 0)', 0);

  Test('s ≔ "the emotional range of a teaspoon";', null);
  Test('count(s, "a")', 4);
  Test('count(s, "e")', 4);
  Test('count(s, "o")', 5);
  Test('count(s, "t")', 3);
  Test('count(s, "h")', 1);
  Test('count(s, "i")', 1);
  Test('count(s, "m")', 1);
  Test('count(s, "n")', 3);
  Test('count(s, "f")', 1);
  Test('count(s, " ")', 5);
  Test('count(s, "⼔")', 0);
  Test('count(s, chr(0))', 0);

  Test('count("the emotional range of a teaspoon", "o")', 5);
  Test('count("", "o")', 0);

  Test('L ≔ ''(21, "cat", "dog", "dog", 123, π, i, 1+1=2, now(), 123, 1024, "dog", "rat", 123, "rat", "rabbit");', null);

  Test('count(L, 21)', 1);
  Test('count(L, "cat")', 1);
  Test('count(L, "dog")', 3);
  Test('count(L, 123)', 3);
  Test('count(L, π)', 1);
  Test('count(L, i)', 1);
  Test('count(L, true)', 1);
  Test('count(L, false)', 0);
  Test('count(L, 1024)', 1);
  Test('count(L, "rat")', 2);
  Test('count(L, "rabbit")', 1);
  Test('count(L, "muggle")', 0);

  Test('count(''(21, "cat", "dog", "dog", 123, π, i, 1+1=2, now(), 123, 1024, "dog", "rat", 123, "rat", "rabbit"), 123)', 3);
  Test('count(''(), 123)', 0);

  Test('S ≔ {21, "cat", "dog", "dog", 123, π, i, 1+1=2, now(), 123, 1024, "dog", "rat", 123, "rat", "rabbit"};', null);

  Test('count(S, 21)', 1);
  Test('count(S, "cat")', 1);
  Test('count(S, "dog")', 1);
  Test('count(S, 123)', 1);
  Test('count(S, π)', 1);
  Test('count(S, i)', 1);
  Test('count(S, true)', 1);
  Test('count(S, false)', 0);
  Test('count(S, 1024)', 1);
  Test('count(S, "rat")', 1);
  Test('count(S, "rabbit")', 1);
  Test('count(S, "muggle")', 0);

  Test('count({21, "cat", "dog", "dog", 123, π, i, 1+1=2, now(), 123, 1024, "dog", "rat", 123, "rat", "rabbit"}, 123)', 1);
  Test('count({}, 123)', 0);

  Test('count(123, 1)', failure, 'Cannot count elements in an object of type "integer".');
  Test('count(123.456, 1)', failure, 'Cannot count elements in an object of type "real number".');

  Test('v ≔ ❨2, 3, 1.999, 5, 8, 4, 5.001, 4, 1, 0, 0.00001, 3.01, 2.0001❩;', null);

  Test('count(v, 2)', 1);
  Test('count(v, 2, 2E-6)', 1);
  Test('count(v, 2, 2E-5)', 1);
  Test('count(v, 2, 2E-4)', 2);
  Test('count(v, 2, 2E-3)', 3);
  Test('count(v, 2, 2E-2)', 3);
  Test('count(v, 2, 2E-1)', 3);

  Test('count(v, 0)', 1);
  Test('count(v, 0, 2E-6)', 1);
  Test('count(v, 0, 2E-5)', 2);
  Test('count(v, 0, 2E-4)', 2);
  Test('count(v, 0, 2E-3)', 2);
  Test('count(v, 0, 2E-2)', 2);
  Test('count(v, 0, 2E-1)', 2);

  Test('count(v, 3)', 1);
  Test('count(v, 3, 2E-6)', 1);
  Test('count(v, 3, 2E-5)', 1);
  Test('count(v, 3, 2E-4)', 1);
  Test('count(v, 3, 2E-3)', 1);
  Test('count(v, 3, 2E-2)', 2);
  Test('count(v, 3, 2E-1)', 2);

  Test('count(v, 9)', 0);
  Test('count(v, 9, 2E-6)', 0);
  Test('count(v, 9, 2E-5)', 0);
  Test('count(v, 9, 2E-4)', 0);
  Test('count(v, 9, 2E-3)', 0);
  Test('count(v, 9, 2E-2)', 0);
  Test('count(v, 9, 2E-1)', 0);

  Test('count(❨2, 3, 1.999, 5, 8, 4, 5.001, 4, 1, 0, 0.00001, 3.01, 2.0001❩, 2, 2E-3)', 3);
  Test('count(DebugObject("empty vector"), 2, 2E-3)', 0);

  Test('v ≔ ❨5, 2, 1, 4.999, 5, 5.01, 5.1, 3 + i, 3.001 + i, 3.001 + 1.01⋅i, 0, 0.1❩;', null);

  Test('count(v, 5)', 2);
  Test('count(v, 5, 2E-6)', 2);
  Test('count(v, 5, 2E-5)', 2);
  Test('count(v, 5, 2E-4)', 2);
  Test('count(v, 5, 2E-3)', 3);
  Test('count(v, 5, 2E-2)', 4);
  Test('count(v, 5, 2E-1)', 5);
  Test('count(v, 5, 2.5)', 8);

  Test('count(v, 3 + i)', 1);
  Test('count(v, 3 + i, 2E-6)', 1);
  Test('count(v, 3 + i, 2E-5)', 1);
  Test('count(v, 3 + i, 2E-4)', 1);
  Test('count(v, 3 + i, 2E-3)', 2);
  Test('count(v, 3 + i, 2E-2)', 3);
  Test('count(v, 3 + i, 2E-1)', 3);
  Test('count(v, 3 + i, 2.5)', 10);

  Test('count(❨5, 2, 1, 4.999, 5, 5.01, 5.1, 3 + i, 3.001 + i, 3.001 + 1.01⋅i, 0, 0.1❩, 3 + i, 2E-3)', 2);
  Test('count(DebugObject("empty vector"), 3 + i, 2E-3)', 0);

  Test('A ≔ ❨❨2, 3, 1.999, 5, 8, 4, 5.001❩, ❨4, 1, 0, 0.00001, 3.01, 2.0001, 100❩❩;', null);

  Test('count(A, 2)', 1);
  Test('count(A, 2, 2E-6)', 1);
  Test('count(A, 2, 2E-5)', 1);
  Test('count(A, 2, 2E-4)', 2);
  Test('count(A, 2, 2E-3)', 3);
  Test('count(A, 2, 2E-2)', 3);
  Test('count(A, 2, 2E-1)', 3);

  Test('count(A, 0)', 1);
  Test('count(A, 0, 2E-6)', 1);
  Test('count(A, 0, 2E-5)', 2);
  Test('count(A, 0, 2E-4)', 2);
  Test('count(A, 0, 2E-3)', 2);
  Test('count(A, 0, 2E-2)', 2);
  Test('count(A, 0, 2E-1)', 2);

  Test('count(A, 3)', 1);
  Test('count(A, 3, 2E-6)', 1);
  Test('count(A, 3, 2E-5)', 1);
  Test('count(A, 3, 2E-4)', 1);
  Test('count(A, 3, 2E-3)', 1);
  Test('count(A, 3, 2E-2)', 2);
  Test('count(A, 3, 2E-1)', 2);

  Test('count(A, 9)', 0);
  Test('count(A, 9, 2E-6)', 0);
  Test('count(A, 9, 2E-5)', 0);
  Test('count(A, 9, 2E-4)', 0);
  Test('count(A, 9, 2E-3)', 0);
  Test('count(A, 9, 2E-2)', 0);
  Test('count(A, 9, 2E-1)', 0);

  Test('count(❨❨2, 3, 1.999, 5, 8, 4, 5.001❩, ❨4, 1, 0, 0.00001, 3.01, 2.0001, 100❩❩, 2, 2E-3)', 3);
  Test('count(DebugObject("empty matrix"), 2, 2E-3)', 0);

  Test('A ≔ ❨❨5, 2, 1, 4.999, 5, 5.01❩, ❨5.1, 3 + i, 3.001 + i, 3.001 + 1.01⋅i, 0, 0.1❩❩;', null);

  Test('count(A, 5)', 2);
  Test('count(A, 5, 2E-6)', 2);
  Test('count(A, 5, 2E-5)', 2);
  Test('count(A, 5, 2E-4)', 2);
  Test('count(A, 5, 2E-3)', 3);
  Test('count(A, 5, 2E-2)', 4);
  Test('count(A, 5, 2E-1)', 5);
  Test('count(A, 5, 2.5)', 8);

  Test('count(A, 3 + i)', 1);
  Test('count(A, 3 + i, 2E-6)', 1);
  Test('count(A, 3 + i, 2E-5)', 1);
  Test('count(A, 3 + i, 2E-4)', 1);
  Test('count(A, 3 + i, 2E-3)', 2);
  Test('count(A, 3 + i, 2E-2)', 3);
  Test('count(A, 3 + i, 2E-1)', 3);
  Test('count(A, 3 + i, 2.5)', 10);

  Test('count(❨❨5, 2, 1, 4.999, 5, 5.01❩, ❨5.1, 3 + i, 3.001 + i, 3.001 + 1.01⋅i, 0, 0.1❩❩, 3 + i, 2E-3)', 2);
  Test('count(ComplexMatrix(DebugObject("empty matrix")), 3 + i, 2E-3)', 0);

  Test('s ≔ "the emotional range of a teaspoon";', null);
  Test('count(s, "a", 1E-5)', 4);
  Test('count(s, "e", 1E-10)', 4);
  Test('count(s, "o", 5)', 5);
  Test('count(s, "t", 100)', 3);
  Test('count(s, "h", 500)', 1);
  Test('count(s, "i", 10E10)', 1);
  Test('count(s, "m", 10E100)', 1);
  Test('count(s, "n", 0)', 3);
  Test('count(s, "f", 2)', 1);
  Test('count(s, " ", 3)', 5);
  Test('count(s, "⼔", 4)', 0);
  Test('count(s, chr(0), 5)', 0);

  Test('count("the emotional range of a teaspoon", "o", 6)', 5);
  Test('count("", "o", 7)', 0);

  Test('L ≔ ''("cat", "dog", 1, 2, 1, 0.99, 5, 4, 1.001, 1 + i/10^6, 5 < 6);', null);

  Test('count(L, "cat")', 1);
  Test('count(L, "cat", 0.01)', 1);

  Test('count(L, 1)', 2);
  Test('count(L, 1, 2E-7)', 2);
  Test('count(L, 1, 2E-6)', 3);
  Test('count(L, 1, 2E-5)', 3);
  Test('count(L, 1, 2E-4)', 3);
  Test('count(L, 1, 2E-3)', 4);
  Test('count(L, 1, 2E-2)', 5);
  Test('count(L, 1, 2E-1)', 5);
  Test('count(L, 1, 1.5)', 6);

  Test('count(''("cat", "dog", 1, 2, 1, 0.99, 5, 4, 1.001, 1 + i/10^6, 5 < 6), 1, 2E-3)', 4);
  Test('count(''(), 1, 2E-3)', 0);

  Test('S ≔ {"cat", "dog", 1, 2, 1, 0.99, 5, 4, 1.001, 1 + i/10^6, 5 < 6};', null);

  Test('count(S, "cat")', 1);
  Test('count(S, "cat", 0.01)', 1);

  Test('count(S, 1)', 1);
  Test('count(S, 1, 2E-7)', 1);
  Test('count(S, 1, 2E-6)', 2);
  Test('count(S, 1, 2E-5)', 2);
  Test('count(S, 1, 2E-4)', 2);
  Test('count(S, 1, 2E-3)', 3);
  Test('count(S, 1, 2E-2)', 4);
  Test('count(S, 1, 2E-1)', 4);
  Test('count(S, 1, 1.5)', 5);

  Test('count({"cat", "dog", 1, 2, 1, 0.99, 5, 4, 1.001, 1 + i/10^6, 5 < 6}, 1, 2E-3)', 3);
  Test('count({}, 1, 2E-3)', 0);

  Test('count(123, 1, 0.001)', failure, 'Cannot count elements in an object of type "integer".');
  Test('count(123.456, 1, 0.001)', failure, 'Cannot count elements in an object of type "real number".');

  Test('count(v)', failure, 'There is no version of this function that can take 1 argument(s).');
  Test('count(A)', failure, 'There is no version of this function that can take 1 argument(s).');
  Test('count(L)', failure, 'There is no version of this function that can take 1 argument(s).');
  Test('count(S)', failure, 'There is no version of this function that can take 1 argument(s).');

  Test('count(S, 1, 2E-3, 3E-4)', failure, 'There is no version of this function that can take 4 argument(s).');
  Test('count(S, 1, -2E-4)', failure, 'A non-negative real number was expected as argument 3, but "-0.0002" was given.');

  Test('count(v, 1, -0.001)', failure, 'A non-negative real number was expected as argument 3, but "-0.001" was given.');

  // contains

  Test('v ≔ ❨6, 2, 0, −4, 1, 7, 3, 21, 123, 5, 6, 7, 6, 3, 0, 1024❩;', null);

  Test('contains(v, 1)', True);
  Test('contains(v, 2)', True);
  Test('contains(v, 3)', True);
  Test('contains(v, 4)', False);
  Test('contains(v, 5)', True);
  Test('contains(v, 6)', True);
  Test('contains(v, 7)', True);
  Test('contains(v, 8)', False);
  Test('contains(v, 9)', False);
  Test('contains(v, 10)', False);
  Test('contains(v, i)', False);
  Test('contains(v, -i)', False);
  Test('contains(v, 1 + i)', False);
  Test('contains(v, 2 + i)', False);
  Test('contains(v, 1024)', True);
  Test('contains(v, 123)', True);
  Test('contains(v, 123.456)', False);
  Test('contains(v, 0)', True);
  Test('contains(v, "cat")', False);
  Test('contains(v, true)', False);
  Test('contains(v, IdentityMatrix(6))', False);
  Test('contains(v, color("red"))', False);

  Test('contains(❨6, 2, 0, −4, 1, 7, 3, 21, 123, 5, 6, 7, 6, 3, 0, 1024❩, 123.456)', False);
  Test('contains(❨6, 2, 0, −4, 1, 7, 3, 21, 123, 5, 6, 7, 6, 3, 0, 1024❩, 0)', True);
  Test('contains(DebugObject("empty vector"), 0)', False);

  Test('v ≔ ❨3, i, 2, 5, 4, −i, 1 + i, 5, 10, 20, 123.456, 2048, 75, 10❩;', null);

  Test('contains(v, 1)', False);
  Test('contains(v, 2)', True);
  Test('contains(v, 3)', True);
  Test('contains(v, 4)', True);
  Test('contains(v, 5)', True);
  Test('contains(v, 6)', False);
  Test('contains(v, 7)', False);
  Test('contains(v, 8)', False);
  Test('contains(v, 9)', False);
  Test('contains(v, 10)', True);
  Test('contains(v, i)', True);
  Test('contains(v, -i)', True);
  Test('contains(v, 1 + i)', True);
  Test('contains(v, 2 + i)', False);
  Test('contains(v, 1024)', False);
  Test('contains(v, 123)', False);
  Test('contains(v, 123.456)', True);
  Test('contains(v, 0)', False);
  Test('contains(v, "cat")', False);
  Test('contains(v, true)', False);
  Test('contains(v, IdentityMatrix(6))', False);
  Test('contains(v, color("red"))', False);

  Test('contains(❨3, i, 2, 5, 4, −i, 1 + i, 5, 10, 20, 123.456, 2048, 75, 10❩, 123.456)', True);
  Test('contains(❨3, i, 2, 5, 4, −i, 1 + i, 5, 10, 20, 123.456, 2048, 75, 10❩, 0)', False);
  Test('contains(ComplexVector(DebugObject("empty vector")), 0)', False);

  Test('v ≔ ❨❨6, 2, 0, −4, 1, 7, 3, 21❩, ❨123, 5, 6, 7, 6, 3, 0, 1024❩❩;', null);

  Test('contains(v, 1)', True);
  Test('contains(v, 2)', True);
  Test('contains(v, 3)', True);
  Test('contains(v, 4)', False);
  Test('contains(v, 5)', True);
  Test('contains(v, 6)', True);
  Test('contains(v, 7)', True);
  Test('contains(v, 8)', False);
  Test('contains(v, 9)', False);
  Test('contains(v, 10)', False);
  Test('contains(v, i)', False);
  Test('contains(v, -i)', False);
  Test('contains(v, 1 + i)', False);
  Test('contains(v, 2 + i)', False);
  Test('contains(v, 1024)', True);
  Test('contains(v, 123)', True);
  Test('contains(v, 123.456)', False);
  Test('contains(v, 0)', True);
  Test('contains(v, "cat")', False);
  Test('contains(v, true)', False);
  Test('contains(v, IdentityMatrix(6))', False);
  Test('contains(v, color("red"))', False);

  Test('contains(❨❨6, 2, 0, −4, 1, 7, 3, 21❩, ❨123, 5, 6, 7, 6, 3, 0, 1024❩❩, 123.456)', False);
  Test('contains(❨❨6, 2, 0, −4, 1, 7, 3, 21❩, ❨123, 5, 6, 7, 6, 3, 0, 1024❩❩, 0)', True);
  Test('contains(DebugObject("empty matrix"), 0)', False);

  Test('v ≔ ❨❨3, i, 2, 5, 4, −i, 1 + i❩, ❨5, 10, 20, 123.456, 2048, 75, 10❩❩;', null);

  Test('contains(v, 1)', False);
  Test('contains(v, 2)', True);
  Test('contains(v, 3)', True);
  Test('contains(v, 4)', True);
  Test('contains(v, 5)', True);
  Test('contains(v, 6)', False);
  Test('contains(v, 7)', False);
  Test('contains(v, 8)', False);
  Test('contains(v, 9)', False);
  Test('contains(v, 10)', True);
  Test('contains(v, i)', True);
  Test('contains(v, -i)', True);
  Test('contains(v, 1 + i)', True);
  Test('contains(v, 2 + i)', False);
  Test('contains(v, 1024)', False);
  Test('contains(v, 123)', False);
  Test('contains(v, 123.456)', True);
  Test('contains(v, 0)', False);
  Test('contains(v, "cat")', False);
  Test('contains(v, true)', False);
  Test('contains(v, IdentityMatrix(6))', False);
  Test('contains(v, color("red"))', False);

  Test('contains(❨❨3, i, 2, 5, 4, −i, 1 + i❩, ❨5, 10, 20, 123.456, 2048, 75, 10❩❩, 123.456)', True);
  Test('contains(❨❨3, i, 2, 5, 4, −i, 1 + i❩, ❨5, 10, 20, 123.456, 2048, 75, 10❩❩, 0)', False);
  Test('contains(ComplexMatrix(DebugObject("empty matrix")), 0)', False);

  Test('s ≔ "the emotional range of a teaspoon";', null);

  Test('contains(s, "a")', True);
  Test('contains(s, "e")', True);
  Test('contains(s, "t")', True);
  Test('contains(s, " ")', True);
  Test('contains(s, "s")', True);
  Test('contains(s, "r")', True);
  Test('contains(s, "q")', False);
  Test('contains(s, "w")', False);
  Test('contains(s, "!")', False);
  Test('contains(s, "?")', False);
  Test('contains(s, "☃")', False);
  Test('contains(s, "the")', True);
  Test('contains(s, "emotional")', True);
  Test('contains(s, "emotion")', True);
  Test('contains(s, "intellectual")', False);
  Test('contains(s, "intellect")', False);
  Test('contains(s, "teaspoon")', True);
  Test('contains(s, "coffee spoon")', False);
  Test('contains(s, "unicorn")', False);
  Test('contains(s, true)', False);
  Test('contains(s, 123)', False);
  Test('contains(s, i)', False);
  Test('contains(s, color("silver"))', False);
  Test('contains(s, 12.5)', False);
  Test('contains(s, chr(0))', False);
  Test('contains(s, """")', False);

  Test('L ≔ ''(3, i, 2, 5, 4, −i, 1 + i, 5, 10, 20, 123.456, 2048, 75, 10, 2, 2);', null);

  Test('contains(L, 1)', False);
  Test('contains(L, 2)', True);
  Test('contains(L, 3)', True);
  Test('contains(L, 4)', True);
  Test('contains(L, 5)', True);
  Test('contains(L, 6)', False);
  Test('contains(L, 7)', False);
  Test('contains(L, 8)', False);
  Test('contains(L, 9)', False);
  Test('contains(L, 10)', True);
  Test('contains(L, i)', True);
  Test('contains(L, -i)', True);
  Test('contains(L, 1 + i)', True);
  Test('contains(L, 2 + i)', False);
  Test('contains(L, 1024)', False);
  Test('contains(L, 123)', False);
  Test('contains(L, 123.456)', True);
  Test('contains(L, 0)', False);
  Test('contains(L, "cat")', False);
  Test('contains(L, true)', False);
  Test('contains(L, IdentityMatrix(6))', False);
  Test('contains(L, color("red"))', False);

  Test('contains(''(3, i, 2, 5, 4, −i, 1 + i, 5, 10, 20, 123.456, 2048, 75, 10, 2, 2), 123.456)', True);
  Test('contains(''(3, i, 2, 5, 4, −i, 1 + i, 5, 10, 20, 123.456, 2048, 75, 10, 2, 2), 0)', False);
  Test('contains(''(), 0)', False);

  Test('S ≔ {3, i, 2, 5, 4, −i, 1 + i, 5, 10, 20, 123.456, 2048, 75, 10, 2, 2};', null);

  Test('contains(S, 1)', False);
  Test('contains(S, 2)', True);
  Test('contains(S, 3)', True);
  Test('contains(S, 4)', True);
  Test('contains(S, 5)', True);
  Test('contains(S, 6)', False);
  Test('contains(S, 7)', False);
  Test('contains(S, 8)', False);
  Test('contains(S, 9)', False);
  Test('contains(S, 10)', True);
  Test('contains(S, i)', True);
  Test('contains(S, -i)', True);
  Test('contains(S, 1 + i)', True);
  Test('contains(S, 2 + i)', False);
  Test('contains(S, 1024)', False);
  Test('contains(S, 123)', False);
  Test('contains(S, 123.456)', True);
  Test('contains(S, 0)', False);
  Test('contains(S, "cat")', False);
  Test('contains(S, true)', False);
  Test('contains(S, IdentityMatrix(6))', False);
  Test('contains(S, color("red"))', False);

  Test('contains({3, i, 2, 5, 4, −i, 1 + i, 5, 10, 20, 123.456, 2048, 75, 10, 2, 2}, 123.456)', True);
  Test('contains({3, i, 2, 5, 4, −i, 1 + i, 5, 10, 20, 123.456, 2048, 75, 10, 2, 2}, 0)', False);
  Test('contains({}, 0)', False);

  Test('S ≔ struct("name": "Sirius", "age": 43, "points": 152000);', null);

  Test('contains(S, "Remus")', False);
  Test('contains(S, "Arthur")', False);
  Test('contains(S, "Horace")', False);
  Test('contains(S, "Hagrid")', False);
  Test('contains(S, "Sirius")', True);
  Test('contains(S, "Minerva")', False);
  Test('contains(S, "Albus")', False);
  Test('contains(S, "Severus")', False);
  Test('contains(S, "Molly")', False);

  Test('contains(S, 100)', False);
  Test('contains(S, 25)', False);
  Test('contains(S, 50)', False);
  Test('contains(S, 57)', False);
  Test('contains(S, 53)', False);
  Test('contains(S, 43)', True);
  Test('contains(S, 80)', False);
  Test('contains(S, 84)', False);
  Test('contains(S, 121)', False);

  Test('contains(S, 150000)', False);
  Test('contains(S, 151000)', False);
  Test('contains(S, 152000)', True);
  Test('contains(S, 153000)', False);
  Test('contains(S, 154000)', False);

  Test('contains(struct("name": "Sirius", "age": 43, "points": 152000), "Hagrid")', False);
  Test('contains(struct("name": "Sirius", "age": 43, "points": 152000), "Sirius")', True);
  Test('contains(struct(), "Hagrid")', False);

  Test('contains(123, 1)', failure, 'Object type "integer" is not a suitable container.');
  Test('contains(123.456, 1)', failure, 'Object type "real number" is not a suitable container.');
  Test('contains(i, 1)', failure, 'Object type "complex number" is not a suitable container.');

  Test('v ≔ ❨5, 3, 2, 0.9999, 7.001, 10.01❩;', null);

  Test('contains(v, 5)', True);
  Test('contains(v, 5, 2E-8)', True);
  Test('contains(v, 5, 2E-7)', True);
  Test('contains(v, 5, 2E-6)', True);
  Test('contains(v, 5, 2E-5)', True);
  Test('contains(v, 5, 2E-4)', True);
  Test('contains(v, 5, 2E-3)', True);
  Test('contains(v, 5, 2E-2)', True);
  Test('contains(v, 5, 2E-1)', True);

  Test('contains(v, 1)', False);
  Test('contains(v, 1, 2E-8)', False);
  Test('contains(v, 1, 2E-7)', False);
  Test('contains(v, 1, 2E-6)', False);
  Test('contains(v, 1, 2E-5)', False);
  Test('contains(v, 1, 2E-4)', True);
  Test('contains(v, 1, 2E-3)', True);
  Test('contains(v, 1, 2E-2)', True);
  Test('contains(v, 1, 2E-1)', True);

  Test('contains(v, 7)', False);
  Test('contains(v, 7, 2E-8)', False);
  Test('contains(v, 7, 2E-7)', False);
  Test('contains(v, 7, 2E-6)', False);
  Test('contains(v, 7, 2E-5)', False);
  Test('contains(v, 7, 2E-4)', False);
  Test('contains(v, 7, 2E-3)', True);
  Test('contains(v, 7, 2E-2)', True);
  Test('contains(v, 7, 2E-1)', True);

  Test('contains(v, i, 2E-1)', False);
  Test('contains(v, "cat", 2E-1)', False);
  Test('contains(v, color("red"), 2E-1)', False);

  Test('contains(❨5, 3, 2, 0.9999, 7.001, 10.01❩, 7, 2E-4)', False);
  Test('contains(❨5, 3, 2, 0.9999, 7.001, 10.01❩, 7, 2E-3)', True);
  Test('contains(DebugObject("empty vector"), 7, 2E-5)', False);

  Test('v ≔ ❨5, 3, 2, 0.9999, 7.001⋅i, 10.01❩;', null);

  Test('contains(v, 5)', True);
  Test('contains(v, 5, 2E-8)', True);
  Test('contains(v, 5, 2E-7)', True);
  Test('contains(v, 5, 2E-6)', True);
  Test('contains(v, 5, 2E-5)', True);
  Test('contains(v, 5, 2E-4)', True);
  Test('contains(v, 5, 2E-3)', True);
  Test('contains(v, 5, 2E-2)', True);
  Test('contains(v, 5, 2E-1)', True);

  Test('contains(v, 1)', False);
  Test('contains(v, 1, 2E-8)', False);
  Test('contains(v, 1, 2E-7)', False);
  Test('contains(v, 1, 2E-6)', False);
  Test('contains(v, 1, 2E-5)', False);
  Test('contains(v, 1, 2E-4)', True);
  Test('contains(v, 1, 2E-3)', True);
  Test('contains(v, 1, 2E-2)', True);
  Test('contains(v, 1, 2E-1)', True);

  Test('contains(v, 7⋅i)', False);
  Test('contains(v, 7⋅i, 2E-8)', False);
  Test('contains(v, 7⋅i, 2E-7)', False);
  Test('contains(v, 7⋅i, 2E-6)', False);
  Test('contains(v, 7⋅i, 2E-5)', False);
  Test('contains(v, 7⋅i, 2E-4)', False);
  Test('contains(v, 7⋅i, 2E-3)', True);
  Test('contains(v, 7⋅i, 2E-2)', True);
  Test('contains(v, 7⋅i, 2E-1)', True);

  Test('contains(v, i, 2E-1)', False);
  Test('contains(v, "cat", 2E-1)', False);
  Test('contains(v, color("red"), 2E-1)', False);

  Test('contains(❨5, 3, 2, 0.9999, 7.001⋅i, 10.01❩, 7⋅i, 2E-4)', False);
  Test('contains(❨5, 3, 2, 0.9999, 7.001⋅i, 10.01❩, 7⋅i, 2E-3)', True);
  Test('contains(ComplexVector(DebugObject("empty vector")), 7⋅i, 2E-5)', False);

  Test('A ≔ ❨❨5, 3, 2❩, ❨0.9999, 7.001, 10.01❩❩;', null);

  Test('contains(A, 5)', True);
  Test('contains(A, 5, 2E-8)', True);
  Test('contains(A, 5, 2E-7)', True);
  Test('contains(A, 5, 2E-6)', True);
  Test('contains(A, 5, 2E-5)', True);
  Test('contains(A, 5, 2E-4)', True);
  Test('contains(A, 5, 2E-3)', True);
  Test('contains(A, 5, 2E-2)', True);
  Test('contains(A, 5, 2E-1)', True);

  Test('contains(A, 1)', False);
  Test('contains(A, 1, 2E-8)', False);
  Test('contains(A, 1, 2E-7)', False);
  Test('contains(A, 1, 2E-6)', False);
  Test('contains(A, 1, 2E-5)', False);
  Test('contains(A, 1, 2E-4)', True);
  Test('contains(A, 1, 2E-3)', True);
  Test('contains(A, 1, 2E-2)', True);
  Test('contains(A, 1, 2E-1)', True);

  Test('contains(A, 7)', False);
  Test('contains(A, 7, 2E-8)', False);
  Test('contains(A, 7, 2E-7)', False);
  Test('contains(A, 7, 2E-6)', False);
  Test('contains(A, 7, 2E-5)', False);
  Test('contains(A, 7, 2E-4)', False);
  Test('contains(A, 7, 2E-3)', True);
  Test('contains(A, 7, 2E-2)', True);
  Test('contains(A, 7, 2E-1)', True);

  Test('contains(A, i, 2E-1)', False);
  Test('contains(A, "cat", 2E-1)', False);
  Test('contains(A, color("red"), 2E-1)', False);

  Test('contains(❨❨5, 3, 2❩, ❨0.9999, 7.001, 10.01❩❩, 7, 2E-4)', False);
  Test('contains(❨❨5, 3, 2❩, ❨0.9999, 7.001, 10.01❩❩, 7, 2E-3)', True);
  Test('contains(DebugObject("empty matrix"), 7, 2E-5)', False);

  Test('A ≔ ❨❨5, 3, 2❩, ❨0.9999, 7.001⋅i, 10.01❩❩;', null);

  Test('contains(A, 5)', True);
  Test('contains(A, 5, 2E-8)', True);
  Test('contains(A, 5, 2E-7)', True);
  Test('contains(A, 5, 2E-6)', True);
  Test('contains(A, 5, 2E-5)', True);
  Test('contains(A, 5, 2E-4)', True);
  Test('contains(A, 5, 2E-3)', True);
  Test('contains(A, 5, 2E-2)', True);
  Test('contains(A, 5, 2E-1)', True);

  Test('contains(A, 1)', False);
  Test('contains(A, 1, 2E-8)', False);
  Test('contains(A, 1, 2E-7)', False);
  Test('contains(A, 1, 2E-6)', False);
  Test('contains(A, 1, 2E-5)', False);
  Test('contains(A, 1, 2E-4)', True);
  Test('contains(A, 1, 2E-3)', True);
  Test('contains(A, 1, 2E-2)', True);
  Test('contains(A, 1, 2E-1)', True);

  Test('contains(A, 7⋅i)', False);
  Test('contains(A, 7⋅i, 2E-8)', False);
  Test('contains(A, 7⋅i, 2E-7)', False);
  Test('contains(A, 7⋅i, 2E-6)', False);
  Test('contains(A, 7⋅i, 2E-5)', False);
  Test('contains(A, 7⋅i, 2E-4)', False);
  Test('contains(A, 7⋅i, 2E-3)', True);
  Test('contains(A, 7⋅i, 2E-2)', True);
  Test('contains(A, 7⋅i, 2E-1)', True);

  Test('contains(A, i, 2E-1)', False);
  Test('contains(A, "cat", 2E-1)', False);
  Test('contains(A, color("red"), 2E-1)', False);

  Test('contains(❨❨5, 3, 2❩, ❨0.9999, 7.001⋅i, 10.01❩❩, 7⋅i, 2E-4)', False);
  Test('contains(❨❨5, 3, 2❩, ❨0.9999, 7.001⋅i, 10.01❩❩, 7⋅i, 2E-3)', True);
  Test('contains(ComplexMatrix(DebugObject("empty matrix")), 7⋅i, 2E-5)', False);

  Test('s ≔ "the emotional range of a teaspoon";', null);

  Test('contains(s, "a", 0.001)', True);
  Test('contains(s, "e", 0.001)', True);
  Test('contains(s, "t", 0.001)', True);
  Test('contains(s, " ", 0.001)', True);
  Test('contains(s, "s", 0.001)', True);
  Test('contains(s, "r", 0.001)', True);
  Test('contains(s, "q", 0.001)', False);
  Test('contains(s, "w", 0.001)', False);
  Test('contains(s, "!", 0.001)', False);
  Test('contains(s, "?", 0.001)', False);
  Test('contains(s, "☃", 0.001)', False);
//  Test('contains(s, "the", 0.001)', True); // meaningless feature not implemented
//  Test('contains(s, "emotional", 0.001)', True);
//  Test('contains(s, "emotion", 0.001)', True);
  Test('contains(s, "intellectual", 0.001)', False);
  Test('contains(s, "intellect", 0.001)', False);
//  Test('contains(s, "teaspoon", 0.001)', True);
  Test('contains(s, "coffee spoon", 0.001)', False);
  Test('contains(s, "unicorn", 0.001)', False);
  Test('contains(s, true, 0.001)', False);
  Test('contains(s, 123, 0.001)', False);
  Test('contains(s, i, 0.001)', False);
  Test('contains(s, color("silver"), 0.001)', False);
  Test('contains(s, 12.5, 0.001)', False);
  Test('contains(s, chr(0), 0.001)', False);
  Test('contains(s, """", 0.001)', False);

  Test('L ≔ ''(5, 3, 2, 0.9999, 7.001, 10.01);', null);

  Test('contains(L, 5)', True);
  Test('contains(L, 5, 2E-8)', True);
  Test('contains(L, 5, 2E-7)', True);
  Test('contains(L, 5, 2E-6)', True);
  Test('contains(L, 5, 2E-5)', True);
  Test('contains(L, 5, 2E-4)', True);
  Test('contains(L, 5, 2E-3)', True);
  Test('contains(L, 5, 2E-2)', True);
  Test('contains(L, 5, 2E-1)', True);

  Test('contains(L, 1)', False);
  Test('contains(L, 1, 2E-8)', False);
  Test('contains(L, 1, 2E-7)', False);
  Test('contains(L, 1, 2E-6)', False);
  Test('contains(L, 1, 2E-5)', False);
  Test('contains(L, 1, 2E-4)', True);
  Test('contains(L, 1, 2E-3)', True);
  Test('contains(L, 1, 2E-2)', True);
  Test('contains(L, 1, 2E-1)', True);

  Test('contains(L, 7)', False);
  Test('contains(L, 7, 2E-8)', False);
  Test('contains(L, 7, 2E-7)', False);
  Test('contains(L, 7, 2E-6)', False);
  Test('contains(L, 7, 2E-5)', False);
  Test('contains(L, 7, 2E-4)', False);
  Test('contains(L, 7, 2E-3)', True);
  Test('contains(L, 7, 2E-2)', True);
  Test('contains(L, 7, 2E-1)', True);

  Test('contains(L, i, 2E-1)', False);
  Test('contains(L, "cat", 2E-1)', False);
  Test('contains(L, color("red"), 2E-1)', False);

  Test('contains(''(5, 3, 2, 0.9999, 7.001, 10.01), 7, 2E-4)', False);
  Test('contains(''(5, 3, 2, 0.9999, 7.001, 10.01), 7, 2E-3)', True);
  Test('contains(''(), 7, 2E-5)', False);

  Test('S ≔ {5, 3, 2, 0.9999, 7.001, 10.01};', null);

  Test('contains(S, 5)', True);
  Test('contains(S, 5, 2E-8)', True);
  Test('contains(S, 5, 2E-7)', True);
  Test('contains(S, 5, 2E-6)', True);
  Test('contains(S, 5, 2E-5)', True);
  Test('contains(S, 5, 2E-4)', True);
  Test('contains(S, 5, 2E-3)', True);
  Test('contains(S, 5, 2E-2)', True);
  Test('contains(S, 5, 2E-1)', True);

  Test('contains(S, 1)', False);
  Test('contains(S, 1, 2E-8)', False);
  Test('contains(S, 1, 2E-7)', False);
  Test('contains(S, 1, 2E-6)', False);
  Test('contains(S, 1, 2E-5)', False);
  Test('contains(S, 1, 2E-4)', True);
  Test('contains(S, 1, 2E-3)', True);
  Test('contains(S, 1, 2E-2)', True);
  Test('contains(S, 1, 2E-1)', True);

  Test('contains(S, 7)', False);
  Test('contains(S, 7, 2E-8)', False);
  Test('contains(S, 7, 2E-7)', False);
  Test('contains(S, 7, 2E-6)', False);
  Test('contains(S, 7, 2E-5)', False);
  Test('contains(S, 7, 2E-4)', False);
  Test('contains(S, 7, 2E-3)', True);
  Test('contains(S, 7, 2E-2)', True);
  Test('contains(S, 7, 2E-1)', True);

  Test('contains(S, i, 2E-1)', False);
  Test('contains(S, "cat", 2E-1)', False);
  Test('contains(S, color("red"), 2E-1)', False);

  Test('contains({5, 3, 2, 0.9999, 7.001, 10.01}, 7, 2E-4)', False);
  Test('contains({5, 3, 2, 0.9999, 7.001, 10.01}, 7, 2E-3)', True);
  Test('contains({}, 7, 2E-5)', False);

  Test('S ≔ struct("name": "Sirius", "age": 43, "points": 152000);', null);

  Test('contains(S, "Remus", 0.001)', False);
  Test('contains(S, "Arthur", 0.001)', False);
  Test('contains(S, "Horace", 0.001)', False);
  Test('contains(S, "Hagrid", 0.001)', False);
  Test('contains(S, "Sirius", 0.001)', True);
  Test('contains(S, "Minerva", 0.001)', False);
  Test('contains(S, "Albus", 0.001)', False);
  Test('contains(S, "Severus", 0.001)', False);
  Test('contains(S, "Molly", 0.001)', False);

  Test('contains(S, 100, 0.001)', False);
  Test('contains(S, 25, 0.001)', False);
  Test('contains(S, 50, 0.001)', False);
  Test('contains(S, 57, 0.001)', False);
  Test('contains(S, 53, 0.001)', False);
  Test('contains(S, 43, 0.001)', True);
  Test('contains(S, 80, 0.001)', False);
  Test('contains(S, 84, 0.001)', False);
  Test('contains(S, 121, 0.001)', False);

  Test('contains(S, 150000, 0.001)', False);
  Test('contains(S, 151000, 0.001)', False);
  Test('contains(S, 152000, 0.001)', True);
  Test('contains(S, 153000, 0.001)', False);
  Test('contains(S, 154000, 0.001)', False);

  Test('contains(S, 150000, 0.001)', False);
  Test('contains(S, 150000, 1)', False);
  Test('contains(S, 150000, 3)', False);
  Test('contains(S, 150000, 30)', False);
  Test('contains(S, 150000, 300)', False);
  Test('contains(S, 150000, 3000)', True);

  Test('contains(struct(), 150000, 0.001)', False);

  Test('contains(123, 1, 0.1)', failure, 'Object type "integer" is not a suitable container.');

  Test('contains(v, 1, -0.1)', failure, 'A non-negative real number was expected as argument 3, but "-0.1" was given.');

  Test('contains(v)', failure, 'There is no version of this function that can take 1 argument(s).');

  // exists

  Test('v ≔ ❨6, 3, 0, 1, 7, 5, 8, 9, 6, 10❩;', null);

  Test('∃(v, odd)', True);
  Test('∃(v, even)', True);
  Test('∃(v, IsPositive)', True);
  Test('∃(v, IsZero)', True);
  Test('∃(v, IsNegative)', False);
  Test('∃(v, IsPrime)', True);
  Test('∃(v, IsInteger)', True);
  Test('∃(v, x ↦ abs(x) ≥ 100)', False);

  Test('v ≔ ❨6, 2, 8, 10, 42, 68, 40, 100❩;', null);

  Test('∃(v, odd)', False);
  Test('∃(v, even)', True);
  Test('∃(v, IsPositive)', True);
  Test('∃(v, IsZero)', False);
  Test('∃(v, IsNegative)', False);
  Test('∃(v, IsPrime)', True);
  Test('∃(v, IsInteger)', True);
  Test('∃(v, x ↦ abs(x) ≥ 100)', True);

  Test('v ≔ ❨-7, -3, -51, 1, -91, -103❩;', null);

  Test('∃(v, odd)', True);
  Test('∃(v, even)', False);
  Test('∃(v, IsPositive)', True);
  Test('∃(v, IsZero)', False);
  Test('∃(v, IsNegative)', True);
  Test('∃(v, IsPrime)', False);
  Test('∃(v, IsInteger)', True);
  Test('∃(v, x ↦ abs(x) ≥ 100)', True);

  Test('v ≔ ❨-40, 2, 80, 102❩;', null);

  Test('∃(v, odd)', False);
  Test('∃(v, even)', True);
  Test('∃(v, IsPositive)', True);
  Test('∃(v, IsZero)', False);
  Test('∃(v, IsNegative)', True);
  Test('∃(v, IsPrime)', True);
  Test('∃(v, IsInteger)', True);
  Test('∃(v, x ↦ abs(x) ≥ 100)', True);

  Test('v ≔ DebugObject("empty vector");', null);

  Test('∃(v, odd)', False);
  Test('∃(v, even)', False);
  Test('∃(v, IsPositive)', False);
  Test('∃(v, IsZero)', False);
  Test('∃(v, IsNegative)', False);
  Test('∃(v, IsPrime)', False);
  Test('∃(v, IsInteger)', False);
  Test('∃(v, x ↦ abs(x) ≥ 100)', False);

  Test('∃(❨5.1, 2.3, 6.7, 1.2, 5.2, 4.3❩, x ↦ frac(x) > 0.5)', True);
  Test('∃(❨5.1, 2.3, 6.3, 1.2, 5.2, 4.3❩, x ↦ frac(x) > 0.5)', False);

  Test('v ≔ ❨6.6, 3.2, 0.4, 1.7, 7.1, 5.5, 8.1, 9.4, 6.1, 10.7❩;', null);

  Test('∃(v, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('A ≔ ❨❨6, 3, 0, 1, 7❩, ❨5, 8, 9, 6, 10❩❩;', null);

  Test('∃(A, odd)', True);
  Test('∃(A, even)', True);
  Test('∃(A, IsPositive)', True);
  Test('∃(A, IsZero)', True);
  Test('∃(A, IsNegative)', False);
  Test('∃(A, IsPrime)', True);
  Test('∃(A, IsInteger)', True);
  Test('∃(A, x ↦ abs(x) ≥ 100)', False);

  Test('A ≔ ❨❨6, 2, 8, 10❩, ❨42, 68, 40, 100❩❩;', null);

  Test('∃(A, odd)', False);
  Test('∃(A, even)', True);
  Test('∃(A, IsPositive)', True);
  Test('∃(A, IsZero)', False);
  Test('∃(A, IsNegative)', False);
  Test('∃(A, IsPrime)', True);
  Test('∃(A, IsInteger)', True);
  Test('∃(A, x ↦ abs(x) ≥ 100)', True);

  Test('A ≔ ❨❨-7, -3, -51❩, ❨1, -91, -103❩❩;', null);

  Test('∃(A, odd)', True);
  Test('∃(A, even)', False);
  Test('∃(A, IsPositive)', True);
  Test('∃(A, IsZero)', False);
  Test('∃(A, IsNegative)', True);
  Test('∃(A, IsPrime)', False);
  Test('∃(A, IsInteger)', True);
  Test('∃(A, x ↦ abs(x) ≥ 100)', True);

  Test('A ≔ ❨❨-40, 2❩, ❨80, 102❩❩;', null);

  Test('∃(A, odd)', False);
  Test('∃(A, even)', True);
  Test('∃(A, IsPositive)', True);
  Test('∃(A, IsZero)', False);
  Test('∃(A, IsNegative)', True);
  Test('∃(A, IsPrime)', True);
  Test('∃(A, IsInteger)', True);
  Test('∃(A, x ↦ abs(x) ≥ 100)', True);

  Test('A ≔ DebugObject("empty matrix");', null);

  Test('∃(A, odd)', False);
  Test('∃(A, even)', False);
  Test('∃(A, IsPositive)', False);
  Test('∃(A, IsZero)', False);
  Test('∃(A, IsNegative)', False);
  Test('∃(A, IsPrime)', False);
  Test('∃(A, IsInteger)', False);
  Test('∃(A, x ↦ abs(x) ≥ 100)', False);

  Test('∃(❨❨5.1, 2.3, 6.7❩, ❨1.2, 5.2, 4.3❩❩, x ↦ frac(x) > 0.5)', True);
  Test('∃(❨❨5.1, 2.3, 6.3❩, ❨1.2, 5.2, 4.3❩❩, x ↦ frac(x) > 0.5)', False);

  Test('v ≔ ❨❨6.6, 3.2, 0.4, 1.7, 7.1❩, ❨5.5, 8.1, 9.4, 6.1, 10.7❩❩;', null);

  Test('∃(v, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('v ≔ ❨3, 1 + i, 5 + 2⋅i, 2⋅i, −3 + 4⋅i, 5 + i❩;', null);

  Test('∃(v, IsReal)', True);
  Test('∃(v, IsInteger)', True);
  Test('∃(v, z ↦ Re(z) > 1)', True);
  Test('∃(v, z ↦ Re(z) > 5)', False);
  Test('∃(v, z ↦ Im(z) > 0)', True);
  Test('∃(v, z ↦ Im(z) < 0)', False);

  Test('v ≔ ❨3.5, 1 + i, 5 + 2⋅i, 2⋅i, −3 + 4⋅i, 8 + i❩;', null);

  Test('∃(v, IsReal)', True);
  Test('∃(v, IsInteger)', False);
  Test('∃(v, z ↦ Re(z) > 1)', True);
  Test('∃(v, z ↦ Re(z) > 5)', True);
  Test('∃(v, z ↦ Im(z) > 0)', True);
  Test('∃(v, z ↦ Im(z) < 0)', False);

  Test('v ≔ ❨3.5 − i, 1 + i, 5 + 2⋅i, 2⋅i, −3 + 4⋅i, 8 + i❩;', null);

  Test('∃(v, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');
  Test('∃(v, IsReal)', False);
  Test('∃(v, IsInteger)', False);
  Test('∃(v, z ↦ Re(z) > 1)', True);
  Test('∃(v, z ↦ Re(z) > 5)', True);
  Test('∃(v, z ↦ Im(z) > 0)', True);
  Test('∃(v, z ↦ Im(z) < 0)', True);

  Test('v ≔ ComplexVector(DebugObject("empty vector"));', null);

  Test('∃(v, IsReal)', False);
  Test('∃(v, IsInteger)', False);
  Test('∃(v, z ↦ Re(z) > 1)', False);
  Test('∃(v, z ↦ Re(z) > 5)', False);
  Test('∃(v, z ↦ Im(z) > 0)', False);
  Test('∃(v, z ↦ Im(z) < 0)', False);

  Test('∃(❨3.5, 1 + i, 5 + 2⋅i, 2⋅i, −3 + 4⋅i, 8 + i❩, z ↦ Im(z) > 0)', True);
  Test('∃(❨3.5, 1 + i, 5 + 2⋅i, 2⋅i, −3 + 4⋅i, 8 + i❩, z ↦ Im(z) < 0)', False);

  Test('A ≔ ❨❨3, 1 + i, 5 + 2⋅i❩, ❨2⋅i, −3 + 4⋅i, 5 + i❩❩;', null);

  Test('∃(A, IsReal)', True);
  Test('∃(A, IsInteger)', True);
  Test('∃(A, z ↦ Re(z) > 1)', True);
  Test('∃(A, z ↦ Re(z) > 5)', False);
  Test('∃(A, z ↦ Im(z) > 0)', True);
  Test('∃(A, z ↦ Im(z) < 0)', False);

  Test('A ≔ ❨❨3.5, 1 + i, 5 + 2⋅i❩, ❨2⋅i, −3 + 4⋅i, 8 + i❩❩;', null);

  Test('∃(A, IsReal)', True);
  Test('∃(A, IsInteger)', False);
  Test('∃(A, z ↦ Re(z) > 1)', True);
  Test('∃(A, z ↦ Re(z) > 5)', True);
  Test('∃(A, z ↦ Im(z) > 0)', True);
  Test('∃(A, z ↦ Im(z) < 0)', False);

  Test('A ≔ ❨❨3.5 − i, 1 + i, 5 + 2⋅i❩, ❨2⋅i, −3 + 4⋅i, 8 + i❩❩;', null);

  Test('∃(A, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');
  Test('∃(A, IsReal)', False);
  Test('∃(A, IsInteger)', False);
  Test('∃(A, z ↦ Re(z) > 1)', True);
  Test('∃(A, z ↦ Re(z) > 5)', True);
  Test('∃(A, z ↦ Im(z) > 0)', True);
  Test('∃(A, z ↦ Im(z) < 0)', True);

  Test('A ≔ ComplexMatrix(DebugObject("empty matrix"));', null);

  Test('∃(A, IsReal)', False);
  Test('∃(A, IsInteger)', False);
  Test('∃(A, z ↦ Re(z) > 1)', False);
  Test('∃(A, z ↦ Re(z) > 5)', False);
  Test('∃(A, z ↦ Im(z) > 0)', False);
  Test('∃(A, z ↦ Im(z) < 0)', False);

  Test('∃(❨❨3.5, 1 + i, 5 + 2⋅i❩, ❨2⋅i, −3 + 4⋅i, 8 + i❩❩, z ↦ Im(z) > 0)', True);
  Test('∃(❨❨3.5, 1 + i, 5 + 2⋅i❩, ❨2⋅i, −3 + 4⋅i, 8 + i❩❩, z ↦ Im(z) < 0)', False);

  Test('L ≔ ''(6, 3, 0, 1, 7, 5, 8, 9, 6, 10);', null);

  Test('∃(L, odd)', True);
  Test('∃(L, even)', True);
  Test('∃(L, IsPositive)', True);
  Test('∃(L, IsZero)', True);
  Test('∃(L, IsNegative)', False);
  Test('∃(L, IsPrime)', True);
  Test('∃(L, IsInteger)', True);
  Test('∃(L, x ↦ abs(x) ≥ 100)', False);

  Test('L ≔ ''(6, 2, 8, 10, 42, 68, 40, 100);', null);

  Test('∃(L, odd)', False);
  Test('∃(L, even)', True);
  Test('∃(L, IsPositive)', True);
  Test('∃(L, IsZero)', False);
  Test('∃(L, IsNegative)', False);
  Test('∃(L, IsPrime)', True);
  Test('∃(L, IsInteger)', True);
  Test('∃(L, x ↦ abs(x) ≥ 100)', True);

  Test('L ≔ ''(-7, -3, -51, 1, -91, -103);', null);

  Test('∃(L, odd)', True);
  Test('∃(L, even)', False);
  Test('∃(L, IsPositive)', True);
  Test('∃(L, IsZero)', False);
  Test('∃(L, IsNegative)', True);
  Test('∃(L, IsPrime)', False);
  Test('∃(L, IsInteger)', True);
  Test('∃(L, x ↦ abs(x) ≥ 100)', True);

  Test('L ≔ ''(-40, 2, 80, 102);', null);

  Test('∃(L, odd)', False);
  Test('∃(L, even)', True);
  Test('∃(L, IsPositive)', True);
  Test('∃(L, IsZero)', False);
  Test('∃(L, IsNegative)', True);
  Test('∃(L, IsPrime)', True);
  Test('∃(L, IsInteger)', True);
  Test('∃(L, x ↦ abs(x) ≥ 100)', True);

  Test('L ≔ ''();', null);

  Test('∃(L, odd)', False);
  Test('∃(L, even)', False);
  Test('∃(L, IsPositive)', False);
  Test('∃(L, IsZero)', False);
  Test('∃(L, IsNegative)', False);
  Test('∃(L, IsPrime)', False);
  Test('∃(L, IsInteger)', False);
  Test('∃(L, x ↦ abs(x) ≥ 100)', False);

  Test('∃(''(5.1, 2.3, 6.7, 1.2, 5.2, 4.3), x ↦ frac(x) > 0.5)', True);
  Test('∃(''(5.1, 2.3, 6.3, 1.2, 5.2, 4.3), x ↦ frac(x) > 0.5)', False);

  Test('L ≔ ''(6.5, 3.1, 0.7, 1.4, 7.6, 5.2, 8.1, 9.8, 6.4, 10.2);', null);

  Test('∃(L, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('S ≔ {6, 3, 0, 1, 7, 5, 8, 9, 6, 10};', null);

  Test('∃(S, odd)', True);
  Test('∃(S, even)', True);
  Test('∃(S, IsPositive)', True);
  Test('∃(S, IsZero)', True);
  Test('∃(S, IsNegative)', False);
  Test('∃(S, IsPrime)', True);
  Test('∃(S, IsInteger)', True);
  Test('∃(S, x ↦ abs(x) ≥ 100)', False);

  Test('S ≔ {6, 2, 8, 10, 42, 68, 40, 100};', null);

  Test('∃(S, odd)', False);
  Test('∃(S, even)', True);
  Test('∃(S, IsPositive)', True);
  Test('∃(S, IsZero)', False);
  Test('∃(S, IsNegative)', False);
  Test('∃(S, IsPrime)', True);
  Test('∃(S, IsInteger)', True);
  Test('∃(S, x ↦ abs(x) ≥ 100)', True);

  Test('S ≔ {-7, -3, -51, 1, -91, -103};', null);

  Test('∃(S, odd)', True);
  Test('∃(S, even)', False);
  Test('∃(S, IsPositive)', True);
  Test('∃(S, IsZero)', False);
  Test('∃(S, IsNegative)', True);
  Test('∃(S, IsPrime)', False);
  Test('∃(S, IsInteger)', True);
  Test('∃(S, x ↦ abs(x) ≥ 100)', True);

  Test('S ≔ {-40, 2, 80, 102};', null);

  Test('∃(S, odd)', False);
  Test('∃(S, even)', True);
  Test('∃(S, IsPositive)', True);
  Test('∃(S, IsZero)', False);
  Test('∃(S, IsNegative)', True);
  Test('∃(S, IsPrime)', True);
  Test('∃(S, IsInteger)', True);
  Test('∃(S, x ↦ abs(x) ≥ 100)', True);

  Test('S ≔ {};', null);

  Test('∃(S, odd)', False);
  Test('∃(S, even)', False);
  Test('∃(S, IsPositive)', False);
  Test('∃(S, IsZero)', False);
  Test('∃(S, IsNegative)', False);
  Test('∃(S, IsPrime)', False);
  Test('∃(S, IsInteger)', False);
  Test('∃(S, x ↦ abs(x) ≥ 100)', False);

  Test('∃({5.1, 2.3, 6.7, 1.2, 5.2, 4.3}, x ↦ frac(x) > 0.5)', True);
  Test('∃({5.1, 2.3, 6.3, 1.2, 5.2, 4.3}, x ↦ frac(x) > 0.5)', False);

  Test('S ≔ {6.5, 3.1, 0.7, 1.4, 7.6, 5.2, 8.1, 9.8, 6.4, 10.2};', null);

  Test('∃(S, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('s ≔ "the emotional range of a teaspoon";', null);

  Test('∃(s, ChrIsLetter)', True);
  Test('∃(s, ChrIsUpperCase)', False);
  Test('∃(s, ChrIsLowerCase)', True);
  Test('∃(s, ChrIsDigit)', False);
  Test('∃(s, ChrIsPunctuation)', False);
  Test('∃(s, ChrIsWhitespace)', True);

  Test('s ≔ "Hello, World!";', null);

  Test('∃(s, ChrIsLetter)', True);
  Test('∃(s, ChrIsUpperCase)', True);
  Test('∃(s, ChrIsLowerCase)', True);
  Test('∃(s, ChrIsDigit)', False);
  Test('∃(s, ChrIsPunctuation)', True);
  Test('∃(s, ChrIsWhitespace)', True);

  Test('∃("7-dehydrocholesterol", ChrIsLetter)', True);
  Test('∃("7-dehydrocholesterol", ChrIsUpperCase)', False);
  Test('∃("7-dehydrocholesterol", ChrIsLowerCase)', True);
  Test('∃("7-dehydrocholesterol", ChrIsDigit)', True);
  Test('∃("7-dehydrocholesterol", ChrIsPunctuation)', True);
  Test('∃("7-dehydrocholesterol", ChrIsWhitespace)', False);

  Test('∃("", ChrIsLetter)', False);
  Test('∃("", ChrIsUpperCase)', False);
  Test('∃("", ChrIsLowerCase)', False);
  Test('∃("", ChrIsDigit)', False);
  Test('∃("", ChrIsPunctuation)', False);
  Test('∃("", ChrIsWhitespace)', False);

  Test('∃(1234, ChrIsDigit)', failure, 'Cannot count elements in an object of type "integer".');
  Test('∃(1234.56789, ChrIsDigit)', failure, 'Cannot count elements in an object of type "real number".');

  // ForAll

  Test('v ≔ ❨5, 3, 0, 2, −1, 7, 5, 9, 10, 15, 85, −40❩;', null);

  Test('∀(v, IsNumber)', True);
  Test('∀(v, even)', False);
  Test('∀(v, odd)', False);
  Test('∀(v, IsPositive)', False);
  Test('∀(v, IsZero)', False);
  Test('∀(v, IsNegative)', False);
  Test('∀(v, IsPrime)', False);
  Test('∀(v, IsInteger)', True);
  Test('∀(v, x ↦ abs(x) ≤ 100)', True);

  Test('v ≔ ❨4, 8, 0, −22, 120, 16, −32, 68, 4, 6, 8, 1024❩;', null);

  Test('∀(v, IsNumber)', True);
  Test('∀(v, even)', True);
  Test('∀(v, odd)', False);
  Test('∀(v, IsPositive)', False);
  Test('∀(v, IsZero)', False);
  Test('∀(v, IsNegative)', False);
  Test('∀(v, IsPrime)', False);
  Test('∀(v, IsInteger)', True);
  Test('∀(v, x ↦ abs(x) ≤ 100)', False);

  Test('v ≔ ❨3, 9, 7, 13, 51, 53, 101, 55, 731, 1133, 191, 21❩;', null);

  Test('∀(v, IsNumber)', True);
  Test('∀(v, even)', False);
  Test('∀(v, odd)', True);
  Test('∀(v, IsPositive)', True);
  Test('∀(v, IsZero)', False);
  Test('∀(v, IsNegative)', False);
  Test('∀(v, IsPrime)', False);
  Test('∀(v, IsInteger)', True);
  Test('∀(v, x ↦ abs(x) ≤ 100)', False);

  Test('v ≔ ❨2, 5, 151, 53, 11, 7, 13, 101, 97, 71❩;', null);

  Test('∀(v, IsNumber)', True);
  Test('∀(v, even)', False);
  Test('∀(v, odd)', False);
  Test('∀(v, IsPositive)', True);
  Test('∀(v, IsZero)', False);
  Test('∀(v, IsNegative)', False);
  Test('∀(v, IsPrime)', True);
  Test('∀(v, IsInteger)', True);
  Test('∀(v, x ↦ abs(x) ≤ 100)', False);

  Test('∀(❨2, 5, 151, 53, 11, 7, 13, 101, 97, 71❩, IsPrime)', True);
  Test('∀(❨1, 5, 151, 53, 11, 7, 13, 101, 97, 71❩, IsPrime)', False);
  Test('∀(❨2, 5, 151, 53, 11, 7, 13, 101, 97, 72❩, IsPrime)', False);

  Test('∀(❨2.7, 5, 151, 53, 11, 7, 13, 101, 97, 71❩, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('∀(DebugObject("empty vector"), IsPrime)', True);

  Test('∀(DebugObject("empty vector"), odd)', True);
  Test('∀(DebugObject("empty vector"), even)', True);

  Test('∀(DebugObject("empty vector"), IsPositive)', True);
  Test('∀(DebugObject("empty vector"), IsNegative)', True);

  Test('A ≔ ❨❨5, 3, 0, 2, −1, 7❩, ❨5, 9, 10, 15, 85, −40❩❩;', null);

  Test('∀(A, IsNumber)', True);
  Test('∀(A, even)', False);
  Test('∀(A, odd)', False);
  Test('∀(A, IsPositive)', False);
  Test('∀(A, IsZero)', False);
  Test('∀(A, IsNegative)', False);
  Test('∀(A, IsPrime)', False);
  Test('∀(A, IsInteger)', True);
  Test('∀(A, x ↦ abs(x) ≤ 100)', True);

  Test('A ≔ ❨❨4, 8, 0, −22, 120, 16❩, ❨−32, 68, 4, 6, 8, 1024❩❩;', null);

  Test('∀(A, IsNumber)', True);
  Test('∀(A, even)', True);
  Test('∀(A, odd)', False);
  Test('∀(A, IsPositive)', False);
  Test('∀(A, IsZero)', False);
  Test('∀(A, IsNegative)', False);
  Test('∀(A, IsPrime)', False);
  Test('∀(A, IsInteger)', True);
  Test('∀(A, x ↦ abs(x) ≤ 100)', False);

  Test('A ≔ ❨❨3, 9, 7, 13, 51, 53❩, ❨101, 55, 731, 1133, 191, 21❩❩;', null);

  Test('∀(A, IsNumber)', True);
  Test('∀(A, even)', False);
  Test('∀(A, odd)', True);
  Test('∀(A, IsPositive)', True);
  Test('∀(A, IsZero)', False);
  Test('∀(A, IsNegative)', False);
  Test('∀(A, IsPrime)', False);
  Test('∀(A, IsInteger)', True);
  Test('∀(A, x ↦ abs(x) ≤ 100)', False);

  Test('A ≔ ❨❨2, 5, 151, 53, 11❩, ❨7, 13, 101, 97, 71❩❩;', null);

  Test('∀(A, IsNumber)', True);
  Test('∀(A, even)', False);
  Test('∀(A, odd)', False);
  Test('∀(A, IsPositive)', True);
  Test('∀(A, IsZero)', False);
  Test('∀(A, IsNegative)', False);
  Test('∀(A, IsPrime)', True);
  Test('∀(A, IsInteger)', True);
  Test('∀(A, x ↦ abs(x) ≤ 100)', False);

  Test('∀(❨❨2, 5, 151, 53, 11❩, ❨7, 13, 101, 97, 71❩❩, IsPrime)', True);
  Test('∀(❨❨1, 5, 151, 53, 11❩, ❨7, 13, 101, 97, 71❩❩, IsPrime)', False);
  Test('∀(❨❨2, 5, 151, 53, 11❩, ❨7, 13, 101, 97, 72❩❩, IsPrime)', False);

  Test('∀(❨❨2.5, 5.1, 151, 53, 11❩, ❨7, 13, 101, 97, 71❩❩, even)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('∀(DebugObject("empty matrix"), IsPrime)', True);

  Test('∀(DebugObject("empty matrix"), odd)', True);
  Test('∀(DebugObject("empty matrix"), even)', True);

  Test('∀(DebugObject("empty matrix"), IsPositive)', True);
  Test('∀(DebugObject("empty matrix"), IsNegative)', True);

  Test('v ≔ ❨10, i, 2, 5, 1 + i, 3 − i, 2 + i, 1 + 2⋅i❩;', null);

  Test('∀(v, IsNumber)', True);
  Test('∀(v, IsReal)', False);
  Test('∀(v, IsNonZero)', True);
  Test('∀(v, z ↦ Re(z) ≥ 0)', True);
  Test('∀(v, z ↦ Re(z) > 0)', False);
  Test('∀(v, z ↦ abs(z) ≤ 10)', True);

  Test('v ≔ ❨11, 5 + i, 2, 5, 1 + i, 3 − i, 2 + i, 1 + 2⋅i❩;', null);

  Test('∀(v, IsNumber)', True);
  Test('∀(v, IsReal)', False);
  Test('∀(v, IsNonZero)', True);
  Test('∀(v, z ↦ Re(z) ≥ 0)', True);
  Test('∀(v, z ↦ Re(z) > 0)', True);
  Test('∀(v, z ↦ abs(z) ≤ 10)', False);

  Test('v ≔ ❨11, 5 + i, 2, 5, 1 + i, 3 − i, 2 + i, 0❩;', null);

  Test('∀(v, IsNumber)', True);
  Test('∀(v, IsReal)', False);
  Test('∀(v, IsNonZero)', False);
  Test('∀(v, z ↦ Re(z) ≥ 0)', True);
  Test('∀(v, z ↦ Re(z) > 0)', False);
  Test('∀(v, z ↦ abs(z) ≤ 10)', False);

  Test('∀(❨10, i, 2, 5, 1 + i, 3 − i, 2 + i, 1 + 2⋅i❩, z ↦ abs(z) ≤ 10)', True);
  Test('∀(❨10, i, 2, 5, 1 + i, 3 − i, 2 + i, 1 + 20⋅i❩, z ↦ abs(z) ≤ 10)', False);

  Test('∀(❨10, i, 2, 5, 1 + i, 3 − i, 2 + i, 1 + 20⋅i❩, even)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('∀(ComplexVector(DebugObject("empty vector")), z ↦ Re(z) > 0)', True);
  Test('∀(ComplexVector(DebugObject("empty vector")), z ↦ Re(z) < 0)', True);

  Test('A ≔ ❨❨10, i, 2, 5❩, ❨1 + i, 3 − i, 2 + i, 1 + 2⋅i❩❩;', null);

  Test('∀(A, IsNumber)', True);
  Test('∀(A, IsReal)', False);
  Test('∀(A, IsNonZero)', True);
  Test('∀(A, z ↦ Re(z) ≥ 0)', True);
  Test('∀(A, z ↦ Re(z) > 0)', False);
  Test('∀(A, z ↦ abs(z) ≤ 10)', True);

  Test('A ≔ ❨❨11, 5 + i, 2, 5❩, ❨1 + i, 3 − i, 2 + i, 1 + 2⋅i❩❩;', null);

  Test('∀(A, IsNumber)', True);
  Test('∀(A, IsReal)', False);
  Test('∀(A, IsNonZero)', True);
  Test('∀(A, z ↦ Re(z) ≥ 0)', True);
  Test('∀(A, z ↦ Re(z) > 0)', True);
  Test('∀(A, z ↦ abs(z) ≤ 10)', False);

  Test('A ≔ ❨❨11, 5 + i, 2, 5❩, ❨1 + i, 3 − i, 2 + i, 0❩❩;', null);

  Test('∀(A, IsNumber)', True);
  Test('∀(A, IsReal)', False);
  Test('∀(A, IsNonZero)', False);
  Test('∀(A, z ↦ Re(z) ≥ 0)', True);
  Test('∀(A, z ↦ Re(z) > 0)', False);
  Test('∀(A, z ↦ abs(z) ≤ 10)', False);

  Test('∀(❨❨10, i, 2, 5❩, ❨1 + i, 3 − i, 2 + i, 1 + 2⋅i❩❩, z ↦ abs(z) ≤ 10)', True);
  Test('∀(❨❨10, i, 2, 5❩, ❨1 + i, 3 − i, 2 + i, 1 + 20⋅i❩❩, z ↦ abs(z) ≤ 10)', False);

  Test('∀(❨❨10, i, 2, 5❩, ❨1 + i, 3 − i, 2 + i, 1 + 20⋅i❩❩, even)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('∀(ComplexMatrix(DebugObject("empty matrix")), z ↦ Re(z) > 0)', True);
  Test('∀(ComplexMatrix(DebugObject("empty matrix")), z ↦ Re(z) < 0)', True);

  Test('s ≔ "the emotional range of a teaspoon";', null);

  Test('∀(s, IsCharacter)', True);
  Test('∀(s, ChrIsASCII)', True);
  Test('∀(s, ChrIsLetter)', False);
  Test('∀(s, ChrIsWhitespace)', False);
  Test('∀(s, c ↦ ChrIsLetter(c) ∨ ChrIsWhitespace(c))', True);

  Test('∀("7-dehydrocholesterol", c ↦ ChrIsLetterOrDigit(c))', False);
  Test('∀("7-dehydrocholesterol", c ↦ ChrIsLetterOrDigit(c) ∨ ChrIsPunctuation(c))', True);

  Test('∀("7-dehydrocholesterol", odd)', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');

  Test('∀("", ChrIsLetter)', True);
  Test('∀("", ChrIsWhitespace)', True);

  Test('L ≔ ''(❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 1❩❩);', null);

  Test('∀(L, IsMatrix)', True);
  Test('∀(L, IsBinary)', True);
  Test('∀(L, IsPermutation)', True);
  Test('∀(L, IsOrthogonal)', True);
  Test('∀(L, IsNormal)', True);
  Test('∀(L, IsIdentity)', False);
  Test('∀(L, IsPositive)', False);
  Test('∀(L, IsNonZero)', True);

  Test('L ≔ ''(❨❨1, 0, 1❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 1❩❩);', null);

  Test('∀(L, IsMatrix)', True);
  Test('∀(L, IsBinary)', True);
  Test('∀(L, IsPermutation)', False);
  Test('∀(L, IsOrthogonal)', False);
  Test('∀(L, IsNormal)', False);
  Test('∀(L, IsIdentity)', False);
  Test('∀(L, IsPositive)', False);
  Test('∀(L, IsNonZero)', True);

  Test('L ≔ ''(❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 2❩❩);', null);

  Test('∀(L, IsMatrix)', True);
  Test('∀(L, IsBinary)', False);
  Test('∀(L, IsPermutation)', False);
  Test('∀(L, IsOrthogonal)', False);
  Test('∀(L, IsNormal)', True);
  Test('∀(L, IsIdentity)', False);
  Test('∀(L, IsPositive)', False);
  Test('∀(L, IsNonZero)', True);

  Test('L ≔ ''(❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 2❩❩, "cat");', null);

  Test('∀(L, IsMatrix)', False);
  Test('∀(L, IsNormal)', failure, 'An object of type matrix was expected as argument 1, but an object of type string was given.');

  Test('∀(''(❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 1❩❩), IsNormal)', True);
  Test('∀(''(❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨2, 1, 3❩, ❨1, 0, 0❩, ❨0, 0, 1❩❩), IsNormal)', False);

  Test('∀(''(), IsSound)', True);
  Test('∀(''(), IsPixmap)', True);

  Test('S ≔ {❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 1❩❩};', null);

  Test('∀(S, IsMatrix)', True);
  Test('∀(S, IsBinary)', True);
  Test('∀(S, IsPermutation)', True);
  Test('∀(S, IsOrthogonal)', True);
  Test('∀(S, IsNormal)', True);
  Test('∀(S, IsIdentity)', False);
  Test('∀(S, IsPositive)', False);
  Test('∀(S, IsNonZero)', True);

  Test('S ≔ {❨❨1, 0, 1❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 1❩❩};', null);

  Test('∀(S, IsMatrix)', True);
  Test('∀(S, IsBinary)', True);
  Test('∀(S, IsPermutation)', False);
  Test('∀(S, IsOrthogonal)', False);
  Test('∀(S, IsNormal)', False);
  Test('∀(S, IsIdentity)', False);
  Test('∀(S, IsPositive)', False);
  Test('∀(S, IsNonZero)', True);

  Test('S ≔ {❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 2❩❩};', null);

  Test('∀(S, IsMatrix)', True);
  Test('∀(S, IsBinary)', False);
  Test('∀(S, IsPermutation)', False);
  Test('∀(S, IsOrthogonal)', False);
  Test('∀(S, IsNormal)', True);
  Test('∀(S, IsIdentity)', False);
  Test('∀(S, IsPositive)', False);
  Test('∀(S, IsNonZero)', True);

  Test('S ≔ {❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 2❩❩, "cat"};', null);

  Test('∀(S, IsMatrix)', False);
  Test('∀(S, IsNormal)', failure, 'An object of type matrix was expected as argument 1, but an object of type string was given.');

  Test('∀({❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨0, 1, 0❩, ❨1, 0, 0❩, ❨0, 0, 1❩❩}, IsNormal)', True);
  Test('∀({❨❨1, 0, 0❩, ❨0, 1, 0❩, ❨0, 0, 1❩❩, ❨❨0, 1❩, ❨1, 0❩❩, ❨❨2, 1, 3❩, ❨1, 0, 0❩, ❨0, 0, 1❩❩}, IsNormal)', False);

  Test('∀({}, IsSound)', True);
  Test('∀({}, IsPixmap)', True);

  // ExistsUnique

  Test('v ≔ ❨5, 3, 2, 0, 1, 4, 7, −5, 6, 9, 2, 3❩;', null);

  Test('ExistsUnique(v, IsPositive)', False);
  Test('ExistsUnique(v, IsZero)', True);
  Test('ExistsUnique(v, IsNegative)', True);
  Test('ExistsUnique(v, odd)', False);
  Test('ExistsUnique(v, even)', False);
  Test('ExistsUnique(v, IsPrime)', False);

  Test('v ≔ ❨0, 3, 2, 0, 1, 4, 7, −5, 6, 9, 2, −3❩;', null);

  Test('ExistsUnique(v, IsPositive)', False);
  Test('ExistsUnique(v, IsZero)', False);
  Test('ExistsUnique(v, IsNegative)', False);
  Test('ExistsUnique(v, odd)', False);
  Test('ExistsUnique(v, even)', False);
  Test('ExistsUnique(v, IsPrime)', False);

  Test('v ≔ ❨0, 3, 2, 2, 2, 4, 4, −8, 6, 6, 2, −6❩;', null);

  Test('ExistsUnique(v, IsPositive)', False);
  Test('ExistsUnique(v, IsZero)', True);
  Test('ExistsUnique(v, IsNegative)', False);
  Test('ExistsUnique(v, odd)', True);
  Test('ExistsUnique(v, even)', False);
  Test('ExistsUnique(v, IsPrime)', False);

  Test('v ≔ ❨0, −3, −2, −2, −2, −4, 4, −8, −6, −6, −2, −6❩;', null);

  Test('ExistsUnique(v, IsPositive)', True);
  Test('ExistsUnique(v, IsZero)', True);
  Test('ExistsUnique(v, IsNegative)', False);
  Test('ExistsUnique(v, odd)', True);
  Test('ExistsUnique(v, even)', False);
  Test('ExistsUnique(v, IsPrime)', False);

  Test('v ≔ ❨0, −3, −2, −2, −2, −4, 2, −8, −6, −6, −2, −6❩;', null);

  Test('ExistsUnique(v, IsPositive)', True);
  Test('ExistsUnique(v, IsZero)', True);
  Test('ExistsUnique(v, IsNegative)', False);
  Test('ExistsUnique(v, odd)', True);
  Test('ExistsUnique(v, even)', False);
  Test('ExistsUnique(v, IsPrime)', True);

  Test('ExistsUnique(❨5, 3, 2, 0, 1, 4, 7, −5, 6, 9, 2, 3❩, IsNegative)', True);
  Test('ExistsUnique(❨5, 3, 2, 0, 1, 4, 7, 5, 6, 9, 2, 3❩, IsNegative)', False);
  Test('ExistsUnique(❨5, 3, 2, 0, 1, −4, −7, 5, 6, 9, 2, 3❩, IsNegative)', False);

  Test('ExistsUnique(DebugObject("empty vector"), IsNegative)', False);
  Test('ExistsUnique(DebugObject("empty vector"), IsNumber)', False);
  Test('ExistsUnique(DebugObject("empty vector"), IsPositive)', False);

  Test('v ≔ ❨5.7, 3, 2, 0, 1, 4, 7, −5, 6, 9, 2, 3❩;', null);

  Test('ExistsUnique(v, IsPositive)', False);
  Test('ExistsUnique(v, IsZero)', True);
  Test('ExistsUnique(v, IsNegative)', True);
  Test('ExistsUnique(v, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('A ≔ ❨❨5, 3, 2, 0, 1, 4❩, ❨7, −5, 6, 9, 2, 3❩❩;', null);

  Test('ExistsUnique(A, IsPositive)', False);
  Test('ExistsUnique(A, IsZero)', True);
  Test('ExistsUnique(A, IsNegative)', True);
  Test('ExistsUnique(A, odd)', False);
  Test('ExistsUnique(A, even)', False);
  Test('ExistsUnique(A, IsPrime)', False);

  Test('A ≔ ❨❨0, 3, 2, 0, 1, 4❩, ❨7, −5, 6, 9, 2, −3❩❩;', null);

  Test('ExistsUnique(A, IsPositive)', False);
  Test('ExistsUnique(A, IsZero)', False);
  Test('ExistsUnique(A, IsNegative)', False);
  Test('ExistsUnique(A, odd)', False);
  Test('ExistsUnique(A, even)', False);
  Test('ExistsUnique(A, IsPrime)', False);

  Test('A ≔ ❨❨0, 3, 2, 2, 2, 4❩, ❨4, −8, 6, 6, 2, −6❩❩;', null);

  Test('ExistsUnique(A, IsPositive)', False);
  Test('ExistsUnique(A, IsZero)', True);
  Test('ExistsUnique(A, IsNegative)', False);
  Test('ExistsUnique(A, odd)', True);
  Test('ExistsUnique(A, even)', False);
  Test('ExistsUnique(A, IsPrime)', False);

  Test('A ≔ ❨❨0, −3, −2, −2, −2, −4❩, ❨4, −8, −6, −6, −2, −6❩❩;', null);

  Test('ExistsUnique(A, IsPositive)', True);
  Test('ExistsUnique(A, IsZero)', True);
  Test('ExistsUnique(A, IsNegative)', False);
  Test('ExistsUnique(A, odd)', True);
  Test('ExistsUnique(A, even)', False);
  Test('ExistsUnique(A, IsPrime)', False);

  Test('A ≔ ❨❨0, −3, −2, −2, −2, −4❩, ❨2, −8, −6, −6, −2, −6❩❩;', null);

  Test('ExistsUnique(A, IsPositive)', True);
  Test('ExistsUnique(A, IsZero)', True);
  Test('ExistsUnique(A, IsNegative)', False);
  Test('ExistsUnique(A, odd)', True);
  Test('ExistsUnique(A, even)', False);
  Test('ExistsUnique(A, IsPrime)', True);

  Test('ExistsUnique(❨❨5, 3, 2, 0, 1, 4❩, ❨7, −5, 6, 9, 2, 3❩❩, IsNegative)', True);
  Test('ExistsUnique(❨❨5, 3, 2, 0, 1, 4❩, ❨7, 5, 6, 9, 2, 3❩❩, IsNegative)', False);
  Test('ExistsUnique(❨❨5, 3, 2, 0, 1, −4❩, ❨−7, 5, 6, 9, 2, 3❩❩, IsNegative)', False);

  Test('ExistsUnique(DebugObject("empty matrix"), IsNegative)', False);
  Test('ExistsUnique(DebugObject("empty matrix"), IsNumber)', False);
  Test('ExistsUnique(DebugObject("empty matrix"), IsPositive)', False);

  Test('A ≔ ❨❨5.4, 3, 2, 0, 1, 4❩, ❨7, −5, 6, 9, 2, 3❩❩;', null);

  Test('ExistsUnique(A, IsPositive)', False);
  Test('ExistsUnique(A, IsZero)', True);
  Test('ExistsUnique(A, IsNegative)', True);
  Test('ExistsUnique(A, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('v ≔ ❨1 + i, 2 − i, 5, 3.5, −7.5, 2 + 3⋅i❩;', null);

  Test('ExistsUnique(v, IsNumber)', False);
  Test('ExistsUnique(v, IsReal)', False);
  Test('ExistsUnique(v, IsInteger)', True);
  Test('ExistsUnique(v, z ↦ Re(z) < 0)', True);
  Test('ExistsUnique(v, IsPrime)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('v ≔ ❨1 + i, 2 − i, 5 + i, 3.5 + i, −7.5, −2 + 3⋅i❩;', null);

  Test('ExistsUnique(v, IsNumber)', False);
  Test('ExistsUnique(v, IsReal)', True);
  Test('ExistsUnique(v, IsInteger)', False);
  Test('ExistsUnique(v, z ↦ Re(z) < 0)', False);
  Test('ExistsUnique(v, IsPrime)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('v ≔ ❨1, 2 − i, 5 + i, 3.5 + i, −7.5, −2 + 3⋅i❩;', null);

  Test('ExistsUnique(v, IsNumber)', False);
  Test('ExistsUnique(v, IsReal)', False);
  Test('ExistsUnique(v, IsInteger)', True);
  Test('ExistsUnique(v, z ↦ Re(z) < 0)', False);
  Test('ExistsUnique(v, IsPrime)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('v ≔ ❨1, 2 − i, 5 + i, 3.5 + i, 7.5, −2❩;', null);

  Test('ExistsUnique(v, IsNumber)', False);
  Test('ExistsUnique(v, IsReal)', False);
  Test('ExistsUnique(v, IsInteger)', False);
  Test('ExistsUnique(v, z ↦ Re(z) < 0)', True);
  Test('ExistsUnique(v, IsPrime)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('ExistsUnique(❨1, 2 − i, 5 + i, 3.5 + i, 7.5, −2❩, z ↦ Re(z) < 0)', True);
  Test('ExistsUnique(❨−1, 2 − i, 5 + i, 3.5 + i, 7.5, −2❩, z ↦ Re(z) < 0)', False);

  Test('ExistsUnique(ComplexVector(DebugObject("empty vector")), IsInteger)', False);
  Test('ExistsUnique(ComplexVector(DebugObject("empty vector")), IsNumber)', False);

  Test('A ≔ ❨❨1 + i, 2 − i, 5❩, ❨3.5, −7.5, 2 + 3⋅i❩❩;', null);

  Test('ExistsUnique(A, IsNumber)', False);
  Test('ExistsUnique(A, IsReal)', False);
  Test('ExistsUnique(A, IsInteger)', True);
  Test('ExistsUnique(A, z ↦ Re(z) < 0)', True);
  Test('ExistsUnique(A, IsPrime)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('A ≔ ❨❨1 + i, 2 − i, 5 + i❩, ❨3.5 + i, −7.5, −2 + 3⋅i❩❩;', null);

  Test('ExistsUnique(A, IsNumber)', False);
  Test('ExistsUnique(A, IsReal)', True);
  Test('ExistsUnique(A, IsInteger)', False);
  Test('ExistsUnique(A, z ↦ Re(z) < 0)', False);
  Test('ExistsUnique(A, IsPrime)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('A ≔ ❨❨1, 2 − i, 5 + i❩, ❨3.5 + i, −7.5, −2 + 3⋅i❩❩;', null);

  Test('ExistsUnique(A, IsNumber)', False);
  Test('ExistsUnique(A, IsReal)', False);
  Test('ExistsUnique(A, IsInteger)', True);
  Test('ExistsUnique(A, z ↦ Re(z) < 0)', False);
  Test('ExistsUnique(A, IsPrime)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('A ≔ ❨❨1, 2 − i, 5 + i❩, ❨3.5 + i, 7.5, −2❩❩;', null);

  Test('ExistsUnique(A, IsNumber)', False);
  Test('ExistsUnique(A, IsReal)', False);
  Test('ExistsUnique(A, IsInteger)', False);
  Test('ExistsUnique(A, z ↦ Re(z) < 0)', True);
  Test('ExistsUnique(A, IsPrime)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('ExistsUnique(❨❨1, 2 − i, 5 + i❩, ❨3.5 + i, 7.5, −2❩❩, z ↦ Re(z) < 0)', True);
  Test('ExistsUnique(❨❨−1, 2 − i, 5 + i❩, ❨3.5 + i, 7.5, −2❩❩, z ↦ Re(z) < 0)', False);

  Test('ExistsUnique(ComplexMatrix(DebugObject("empty matrix")), IsInteger)', False);
  Test('ExistsUnique(ComplexMatrix(DebugObject("empty matrix")), IsNumber)', False);

  Test('s ≔ "7-dehydrocholesterol";', null);

  Test('ExistsUnique(s, IsCharacter)', False);
  Test('ExistsUnique(s, IsPixmap)', False);
  Test('ExistsUnique(s, IsSound)', False);
  Test('ExistsUnique(s, ChrIsASCII)', False);
  Test('ExistsUnique(s, ChrIsLetter)', False);
  Test('ExistsUnique(s, ChrIsDigit)', True);
  Test('ExistsUnique(s, ChrIsPunctuation)', True);
  Test('ExistsUnique(s, ChrIsWhitespace)', False);

  Test('s ≔ "7-dehydrocholesterol!";', null);

  Test('ExistsUnique(s, IsCharacter)', False);
  Test('ExistsUnique(s, IsPixmap)', False);
  Test('ExistsUnique(s, IsSound)', False);
  Test('ExistsUnique(s, ChrIsASCII)', False);
  Test('ExistsUnique(s, ChrIsLetter)', False);
  Test('ExistsUnique(s, ChrIsDigit)', True);
  Test('ExistsUnique(s, ChrIsPunctuation)', False);
  Test('ExistsUnique(s, ChrIsWhitespace)', False);

  Test('ExistsUnique("7-dehydrocholesterol", IsCharacter)', False);
  Test('ExistsUnique("7-dehydrocholesterol", IsPixmap)', False);
  Test('ExistsUnique("7-dehydrocholesterol", IsSound)', False);
  Test('ExistsUnique("7-dehydrocholesterol", ChrIsASCII)', False);
  Test('ExistsUnique("7-dehydrocholesterol", ChrIsLetter)', False);
  Test('ExistsUnique("7-dehydrocholesterol", ChrIsDigit)', True);
  Test('ExistsUnique("7-dehydrocholesterol", ChrIsPunctuation)', True);
  Test('ExistsUnique("7-dehydrocholesterol", ChrIsWhitespace)', False);

  Test('ExistsUnique("", IsCharacter)', False);
  Test('ExistsUnique("", IsPixmap)', False);
  Test('ExistsUnique("", IsSound)', False);
  Test('ExistsUnique("", ChrIsASCII)', False);
  Test('ExistsUnique("", ChrIsLetter)', False);
  Test('ExistsUnique("", ChrIsDigit)', False);
  Test('ExistsUnique("", ChrIsPunctuation)', False);
  Test('ExistsUnique("", ChrIsWhitespace)', False);

  Test('ExistsUnique("7-dehydrocholesterol", odd)', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');

  Test('L ≔ ''(10, 2, −7, π, 2⋅π, 1024);', null);

  Test('ExistsUnique(L, x ↦ IsInteger(x) ∧ even(x))', False);
  Test('ExistsUnique(L, x ↦ IsInteger(x) ∧ odd(x))', True);
  Test('ExistsUnique(L, IsInteger)', False);
  Test('ExistsUnique(L, IsNegative)', True);
  Test('ExistsUnique(L, IsPositive)', False);
  Test('ExistsUnique(L, IsZero)', False);
  Test('ExistsUnique(L, x ↦ sin(x) ≈ 0)', False);
  Test('ExistsUnique(L, x ↦ sin(x/2) ≈ 0)', True);
  Test('ExistsUnique(L, x ↦ x ≥ 10000)', False);
  Test('ExistsUnique(L, x ↦ x ≥ 1000)', True);
  Test('ExistsUnique(L, x ↦ x ≥ 100)', True);
  Test('ExistsUnique(L, x ↦ x ≥ 10)', False);
  Test('ExistsUnique(L, x ↦ x ≥ 1)', False);
  Test('ExistsUnique(L, x ↦ x ≥ -100)', False);
  Test('ExistsUnique(L, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('ExistsUnique(''(10, 2, −7, π, 2⋅π, 1024), x ↦ sin(x) ≈ 0)', False);
  Test('ExistsUnique(''(10, 2, −7, π, 2⋅π, 1024), x ↦ sin(x/2) ≈ 0)', True);

  Test('ExistsUnique(''(), x ↦ sin(x) ≈ 0)', False);
  Test('ExistsUnique(''(), x ↦ true)', False);

  Test('S ≔ {10, 2, −7, −7, π, 2⋅π, 1024};', null);

  Test('ExistsUnique(S, x ↦ IsInteger(x) ∧ even(x))', False);
  Test('ExistsUnique(S, x ↦ IsInteger(x) ∧ odd(x))', True);
  Test('ExistsUnique(S, IsInteger)', False);
  Test('ExistsUnique(S, IsNegative)', True);
  Test('ExistsUnique(S, IsPositive)', False);
  Test('ExistsUnique(S, IsZero)', False);
  Test('ExistsUnique(S, x ↦ sin(x) ≈ 0)', False);
  Test('ExistsUnique(S, x ↦ sin(x/2) ≈ 0)', True);
  Test('ExistsUnique(S, x ↦ x ≥ 10000)', False);
  Test('ExistsUnique(S, x ↦ x ≥ 1000)', True);
  Test('ExistsUnique(S, x ↦ x ≥ 100)', True);
  Test('ExistsUnique(S, x ↦ x ≥ 10)', False);
  Test('ExistsUnique(S, x ↦ x ≥ 1)', False);
  Test('ExistsUnique(S, x ↦ x ≥ -100)', False);
  Test('ExistsUnique(S, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');

  Test('ExistsUnique({10, 2, −7, π, 2⋅π, 1024}, x ↦ sin(x) ≈ 0)', False);
  Test('ExistsUnique({10, 2, −7, π, 2⋅π, 1024}, x ↦ sin(x/2) ≈ 0)', True);

  Test('ExistsUnique({}, x ↦ sin(x) ≈ 0)', False);
  Test('ExistsUnique({}, x ↦ true)', False);

  // exists and ForAll performance

  // - best-case scenario

  Test('L ≔ SequenceList(10); for(i, 1, #L, (L[i] ≔ ZeroMatrix(1000); L[i][1, 1] ≔ π));', null);
  StartStopwatch;
  Test('for(i, 1, #L, assert(¬∀(L[i], IsZero)))', null);
  StopStopwatch;
  TestDuration(0.001);

  Test('L ≔ SequenceList(10); for(i, 1, #L, (L[i] ≔ ZeroMatrix(1000); L[i][1, 1] ≔ π));', null);
  StartStopwatch;
  Test('for(i, 1, #L, assert(∃(L[i], IsNonZero)))', null);
  StopStopwatch;
  TestDuration(0.001);

  Test('L ≔ SequenceList(10); for(i, 1, #L, (L[i] ≔ ZeroMatrix(1000); L[i][1, 1] ≔ π; L[i][1, 2] ≔ π));', null);
  StartStopwatch;
  Test('for(i, 1, #L, assert(¬ExistsUnique(L[i], IsNonZero)))', null);
  StopStopwatch;
  TestDuration(0.001);

  // - worst-case scenario

{$IFNDEF QuickTest}
  Test('L ≔ SequenceList(10); for(i, 1, #L, (L[i] ≔ ZeroMatrix(1000)));', null);
  StartStopwatch;
  Test('for(i, 1, #L, assert(∀(L[i], IsZero)))', null);
  Test('for(i, 1, #L, assert(¬∃(L[i], IsNonZero)))', null);
  Test('for(i, 1, #L, assert(¬ExistsUnique(L[i], IsNonZero)))', null);
  StopStopwatch;
  TestDuration(60.0);
{$ENDIF}

  // indices

  Test('v ≔ ❨6, 2, 1, 0, −4, 3, 7, 8, −9, 1, 0, 10❩;', null);

  Test('indices(v, IsNumber)', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]));
  Test('indices(v, IsInteger)', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]));
  Test('indices(v, odd)', intarr([3, 6, 7, 9, 10]));
  Test('indices(v, even)', intarr([1, 2, 4, 5, 8, 11, 12]));
  Test('indices(v, IsPositive)', intarr([1, 2, 3, 6, 7, 8, 10, 12]));
  Test('indices(v, IsZero)', intarr([4, 11]));
  Test('indices(v, IsNegative)', intarr([5, 9]));
  Test('indices(v, IsPrime)', intarr([2, 6, 7]));
  Test('indices(v, IsComplex)', intarr([]));
  Test('indices(v, IsPixmap)', intarr([]));
  Test('indices(v, n ↦ abs(n) > 5)', intarr([1, 7, 8, 9, 12]));
  Test('indices(v, n ↦ abs(n) > 10)', intarr([]));
  Test('indices(v, n ↦ n! > 5)', failure, 'A non-negative integer was expected as argument 1, but "-4" was given.');
  Test('indices(DebugObject("empty vector"), IsNumber)', intarr([]));
  Test('indices(❨6, 2, 1, 0, −4, 3, 7, 8, −9, 1, 0, 10❩, n ↦ abs(n) > 5)', intarr([1, 7, 8, 9, 12]));

  Test('A ≔ ❨❨6, 2, 1, 0, −4, 3❩, ❨7, 8, −9, 1, 0, 10❩❩;', null);

  Test('indices(A, IsNumber)', asoarrex([idx2(1, 1), idx2(1, 2), idx2(1, 3), idx2(1, 4), idx2(1, 5), idx2(1, 6), idx2(2, 1), idx2(2, 2), idx2(2, 3), idx2(2, 4), idx2(2, 5), idx2(2, 6)]));
  Test('indices(A, IsInteger)', asoarrex([idx2(1, 1), idx2(1, 2), idx2(1, 3), idx2(1, 4), idx2(1, 5), idx2(1, 6), idx2(2, 1), idx2(2, 2), idx2(2, 3), idx2(2, 4), idx2(2, 5), idx2(2, 6)]));
  Test('indices(A, odd)', asoarrex([idx2(1, 3), idx2(1, 6), idx2(2, 1), idx2(2, 3), idx2(2, 4)]));
  Test('indices(A, even)', asoarrex([idx2(1, 1), idx2(1, 2), idx2(1, 4), idx2(1, 5), idx2(2, 2), idx2(2, 5), idx2(2, 6)]));
  Test('indices(A, IsPositive)', asoarrex([idx2(1, 1), idx2(1, 2), idx2(1, 3), idx2(1, 6), idx2(2, 1), idx2(2, 2), idx2(2, 4), idx2(2, 6)]));
  Test('indices(A, IsZero)', asoarrex([idx2(1, 4), idx2(2, 5)]));
  Test('indices(A, IsNegative)', asoarrex([idx2(1, 5), idx2(2, 3)]));
  Test('indices(A, IsPrime)', asoarrex([idx2(1, 2), idx2(1, 6), idx2(2, 1)]));
  Test('indices(A, IsComplex)', asoarrex([]));
  Test('indices(A, IsPixmap)', asoarrex([]));
  Test('indices(A, n ↦ abs(n) > 5)', asoarrex([idx2(1, 1), idx2(2, 1), idx2(2, 2), idx2(2, 3), idx2(2, 6)]));
  Test('indices(A, n ↦ abs(n) > 10)', asoarrex([]));
  Test('indices(A, n ↦ n! > 5)', failure, 'A non-negative integer was expected as argument 1, but "-4" was given.');
  Test('indices(DebugObject("empty matrix"), IsNumber)', asoarrex([]));
  Test('indices(❨❨6, 2, 1, 0, −4, 3❩, ❨7, 8, −9, 1, 0, 10❩❩, IsPrime)', asoarrex([idx2(1, 2), idx2(1, 6), idx2(2, 1)]));

  Test('v ≔ ❨6, 2.5, 1, i, 1 + i, −i, 0, 5❩;', null);

  Test('indices(v, IsNumber)', intarr([1, 2, 3, 4, 5, 6, 7, 8]));
  Test('indices(v, IsReal)', intarr([1, 2, 3, 7, 8]));
  Test('indices(v, IsInteger)', intarr([1, 3, 7, 8]));
  Test('indices(v, IsZero)', intarr([7]));
  Test('indices(v, z ↦ abs(z) > 1)', intarr([1, 2, 5, 8]));
  Test('indices(v, z ↦ abs(z) > 10)', intarr([]));
  Test('indices(v, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');
  Test('indices(ComplexVector(DebugObject("empty vector")), IsNumber)', intarr([]));
  Test('indices(❨6, 2.5, 1, i, 1 + i, −i, 0, 5❩, z ↦ abs(z) > 1)', intarr([1, 2, 5, 8]));

  Test('A ≔ ❨❨6, 2.5, 1, i❩, ❨1 + i, −i, 0, 5❩❩;', null);

  Test('indices(A, IsNumber)', asoarrex([idx2(1, 1), idx2(1, 2), idx2(1, 3), idx2(1, 4), idx2(2, 1), idx2(2, 2), idx2(2, 3), idx2(2, 4)]));
  Test('indices(A, IsReal)', asoarrex([idx2(1, 1), idx2(1, 2), idx2(1, 3), idx2(2, 3), idx2(2, 4)]));
  Test('indices(A, IsInteger)', asoarrex([idx2(1, 1), idx2(1, 3), idx2(2, 3), idx2(2, 4)]));
  Test('indices(A, IsZero)', asoarrex([idx2(2, 3)]));
  Test('indices(A, z ↦ abs(z) > 1)', asoarrex([idx2(1, 1), idx2(1, 2), idx2(2, 1), idx2(2, 4)]));
  Test('indices(A, z ↦ abs(z) > 10)', asoarrex([]));
  Test('indices(A, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');
  Test('indices(ComplexMatrix(DebugObject("empty matrix")), IsNumber)', asoarrex([]));
  Test('indices(❨❨6, 2.5, 1, i❩, ❨1 + i, −i, 0, 5❩❩, z ↦ abs(z) > 1)', asoarrex([idx2(1, 1), idx2(1, 2), idx2(2, 1), idx2(2, 4)]));

  Test('s ≔ "Hello, World!";', null);

  Test('indices(s, IsCharacter)', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]));
  Test('indices(s, ChrIsASCII)', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]));
  Test('indices(s, ChrIsLetter)', intarr([1, 2, 3, 4, 5, 8, 9, 10, 11, 12]));
  Test('indices(s, ChrIsUpperCase)', intarr([1, 8]));
  Test('indices(s, ChrIsPunctuation)', intarr([6, 13]));
  Test('indices(s, ChrIsWhitespace)', intarr([7]));
  Test('indices(s, ChrIsControl)', intarr([]));

  Test('indices("Hello, World!", ChrIsPunctuation)', intarr([6, 13]));
  Test('indices("", IsCharacter)', intarr([]));
  Test('indices(s, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');

  Test('L ≔ ''(6, 2, 1, 0, −4, 3, 7, 8, −9, 1, 0, 10);', null);

  Test('indices(L, IsNumber)', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]));
  Test('indices(L, IsInteger)', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]));
  Test('indices(L, odd)', intarr([3, 6, 7, 9, 10]));
  Test('indices(L, even)', intarr([1, 2, 4, 5, 8, 11, 12]));
  Test('indices(L, IsPositive)', intarr([1, 2, 3, 6, 7, 8, 10, 12]));
  Test('indices(L, IsZero)', intarr([4, 11]));
  Test('indices(L, IsNegative)', intarr([5, 9]));
  Test('indices(L, IsPrime)', intarr([2, 6, 7]));
  Test('indices(L, IsComplex)', intarr([]));
  Test('indices(L, IsPixmap)', intarr([]));
  Test('indices(L, n ↦ abs(n) > 5)', intarr([1, 7, 8, 9, 12]));
  Test('indices(L, n ↦ abs(n) > 10)', intarr([]));
  Test('indices(L, n ↦ n! > 5)', failure, 'A non-negative integer was expected as argument 1, but "-4" was given.');
  Test('indices(''(), IsNumber)', intarr([]));
  Test('indices(''(6, 2, 1, 0, −4, 3, 7, 8, −9, 1, 0, 10), n ↦ abs(n) > 5)', intarr([1, 7, 8, 9, 12]));

  Test('S ≔ {6, 2, 1, 0, −4, 3, 7, 8, −9, 1, 0, 10};', null);

  Test('indices(S, IsNumber)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, IsInteger)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, odd)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, even)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, IsPositive)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, IsZero)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, IsNegative)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, IsPrime)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, IsComplex)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, IsPixmap)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, n ↦ abs(n) > 5)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, n ↦ abs(n) > 10)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, n ↦ n! > 5)', failure, 'Object type "set" is not a suitable container.');
  Test('indices({}, IsNumber)', failure, 'Object type "set" is not a suitable container.');
  Test('indices({6, 2, 1, 0, −4, 3, 7, 8, −9, 1, 0, 10}, n ↦ abs(n) > 5)', failure, 'Object type "set" is not a suitable container.');

  Test('S ≔ struct("name": "Albus Dumbledore", "age": 116, "role": "headmaster");', null);
  Test('indices(S, m ↦ type(m) = "string")', intarr([1, 3]));
  Test('indices(S, m ↦ type(m) = "integer")', intarr([2]));
  Test('indices(S, m ↦ type(m) = "pixmap")', intarr([]));

  Test('indices(struct("name": "Albus Dumbledore", "age": 116, "role": "headmaster"), m ↦ type(m) = "string")', intarr([1, 3]));

  Test('S ≔ struct();', null);
  Test('indices(S, m ↦ type(m) = "string")', intarr([]));
  Test('indices(S, m ↦ type(m) = "integer")', intarr([]));
  Test('indices(S, m ↦ type(m) = "pixmap")', intarr([]));

  Test('indices(123, IsNumber)', failure, 'Object type "integer" is not a suitable container.');

  Test('v ≔ ❨3, 0, 2, 5, −7, 4, 1, −9, 1, 0, 5, 2, 1, −3, 5, 4❩;', null);

  Test('indices(v, 0)', intarr([2, 10]));
  Test('indices(v, 1)', intarr([7, 9, 13]));
  Test('indices(v, 2)', intarr([3, 12]));
  Test('indices(v, 3)', intarr([1]));
  Test('indices(v, 4)', intarr([6, 16]));
  Test('indices(v, 5)', intarr([4, 11, 15]));
  Test('indices(v, 6)', intarr([]));
  Test('indices(v, 7)', intarr([]));
  Test('indices(v, 8)', intarr([]));
  Test('indices(v, 9)', intarr([]));
  Test('indices(v, "nargle")', intarr([]));

  Test('indices(❨3, 0, 2, 5, −7, 4, 1, −9, 1, 0, 5, 2, 1, −3, 5, 4❩, 5)', intarr([4, 11, 15]));
  Test('indices(DebugObject("empty vector"), 5)', intarr([]));

  Test('A ≔ ❨❨3, 0, 2, 5, −7, 4, 1, −9❩, ❨1, 0, 5, 2, 1, −3, 5, 4❩❩;', null);

  Test('indices(A, 0)', asoarrex([idx2(1, 2), idx2(2, 2)]));
  Test('indices(A, 1)', asoarrex([idx2(1, 7), idx2(2, 1), idx2(2, 5)]));
  Test('indices(A, 2)', asoarrex([idx2(1, 3), idx2(2, 4)]));
  Test('indices(A, 3)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 4)', asoarrex([idx2(1, 6), idx2(2, 8)]));
  Test('indices(A, 5)', asoarrex([idx2(1, 4), idx2(2, 3), idx2(2, 7)]));
  Test('indices(A, 6)', asoarrex([]));
  Test('indices(A, 7)', asoarrex([]));
  Test('indices(A, 8)', asoarrex([]));
  Test('indices(A, 9)', asoarrex([]));
  Test('indices(A, "nargle")', asoarrex([]));

  Test('indices(❨❨3, 0, 2, 5, −7, 4, 1, −9❩, ❨1, 0, 5, 2, 1, −3, 5, 4❩❩, 5)', asoarrex([idx2(1, 4), idx2(2, 3), idx2(2, 7)]));
  Test('indices(DebugObject("empty matrix"), 5)', asoarrex([]));

  Test('v ≔ ❨1, 5, i, 2, −i, 3, −1, 1 + i, 2, 3, 1, 0, 5, 2, 5, 5❩;', null);

  Test('indices(v, 0)', intarr([12]));
  Test('indices(v, 1)', intarr([1, 11]));
  Test('indices(v, 2)', intarr([4, 9, 14]));
  Test('indices(v, 3)', intarr([6, 10]));
  Test('indices(v, 4)', intarr([]));
  Test('indices(v, 5)', intarr([2, 13, 15, 16]));
  Test('indices(v, 6)', intarr([]));
  Test('indices(v, 7)', intarr([]));
  Test('indices(v, 8)', intarr([]));
  Test('indices(v, 9)', intarr([]));
  Test('indices(v, i)', intarr([3]));
  Test('indices(v, "nargle")', intarr([]));

  Test('indices(❨1, 5, i, 2, −i, 3, −1, 1 + i, 2, 3, 1, 0, 5, 2, 5, 5❩, 5)', intarr([2, 13, 15, 16]));
  Test('indices(ComplexVector(DebugObject("empty vector")), 8)', intarr([]));

  Test('v ≔ ❨❨1, 5, i, 2, −i, 3, −1, 1 + i❩, ❨2, 3, 1, 0, 5, 2, 5, 5❩❩;', null);

  Test('indices(v, 0)', asoarrex([idx2(2, 4)]));
  Test('indices(v, 1)', asoarrex([idx2(1, 1), idx2(2, 3)]));
  Test('indices(v, 2)', asoarrex([idx2(1, 4), idx2(2, 1), idx2(2, 6)]));
  Test('indices(v, 3)', asoarrex([idx2(1, 6), idx2(2, 2)]));
  Test('indices(v, 4)', asoarrex([]));
  Test('indices(v, 5)', asoarrex([idx2(1, 2), idx2(2, 5), idx2(2, 7), idx2(2, 8)]));
  Test('indices(v, 6)', asoarrex([]));
  Test('indices(v, 7)', asoarrex([]));
  Test('indices(v, 8)', asoarrex([]));
  Test('indices(v, 9)', asoarrex([]));
  Test('indices(v, i)', asoarrex([idx2(1, 3)]));
  Test('indices(v, "nargle")', asoarrex([]));

  Test('indices(❨❨1, 5, i, 2, −i, 3, −1, 1 + i❩, ❨2, 3, 1, 0, 5, 2, 5, 5❩❩, 5)', asoarrex([idx2(1, 2), idx2(2, 5), idx2(2, 7), idx2(2, 8)]));
  Test('indices(ComplexMatrix(DebugObject("empty matrix")), 8)', asoarrex([]));

  Test('s ≔ "the emotional range of a teaspoon";', null);

  Test('indices(s, "a")', intarr([12, 16, 24, 28]));
  Test('indices(s, "e")', intarr([3, 5, 19, 27]));
  Test('indices(s, "t")', intarr([1, 8, 26]));
  Test('indices(s, "y")', intarr([]));
  Test('indices(s, "m")', intarr([6]));
  Test('indices(s, "the")', intarr([]));
  Test('indices(s, 1024)', intarr([]));

  Test('indices("the emotional range of a teaspoon", "t")', intarr([1, 8, 26]));
  Test('indices("", "t")', intarr([]));

  Test('L ≔ ''(3, 0, 2, 5, −7, 4, 1, −9, 1, 0, 5, 2, 1, −3, 5, 4);', null);

  Test('indices(L, 0)', intarr([2, 10]));
  Test('indices(L, 1)', intarr([7, 9, 13]));
  Test('indices(L, 2)', intarr([3, 12]));
  Test('indices(L, 3)', intarr([1]));
  Test('indices(L, 4)', intarr([6, 16]));
  Test('indices(L, 5)', intarr([4, 11, 15]));
  Test('indices(L, 6)', intarr([]));
  Test('indices(L, 7)', intarr([]));
  Test('indices(L, 8)', intarr([]));
  Test('indices(L, 9)', intarr([]));
  Test('indices(L, "nargle")', intarr([]));

  Test('indices(''(3, 0, 2, 5, −7, 4, 1, −9, 1, 0, 5, 2, 1, −3, 5, 4), 5)', intarr([4, 11, 15]));
  Test('indices(''(), 5)', intarr([]));

  Test('S ≔ {3, 0, 2, 5, −7, 4, 1, −9, 1, 0, 5, 2, 1, −3, 5, 4};', null);

  Test('indices(S, 0)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 1)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 2)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 3)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 4)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 5)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 6)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 7)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 8)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 9)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, "nargle")', failure, 'Object type "set" is not a suitable container.');

  Test('indices({3, 0, 2, 5, −7, 4, 1, −9, 1, 0, 5, 2, 1, −3, 5, 4}, 5)', failure, 'Object type "set" is not a suitable container.');
  Test('indices({}, 5)', failure, 'Object type "set" is not a suitable container.');

  Test('S ≔ struct("name": "Albus Dumbledore", "age": 116, "role": "headmaster");', null);
  Test('indices(S, 116)', intarr([2]));

  Test('indices(123, 1)', failure, 'Object type "integer" is not a suitable container.');

  Test('v ≔ ❨3, 4, 5, 1, 0, 4.999, 1, 5.01, 1.1, 0.0001❩;', null);

  Test('indices(v, 0)', intarr([5]));
  Test('indices(v, 0, 2E-6)', intarr([5]));
  Test('indices(v, 0, 2E-5)', intarr([5]));
  Test('indices(v, 0, 2E-4)', intarr([5, 10]));
  Test('indices(v, 0, 2E-3)', intarr([5, 10]));
  Test('indices(v, 0, 2E-2)', intarr([5, 10]));
  Test('indices(v, 0, 2E-1)', intarr([5, 10]));

  Test('indices(v, 1)', intarr([4, 7]));
  Test('indices(v, 1, 2E-6)', intarr([4, 7]));
  Test('indices(v, 1, 2E-5)', intarr([4, 7]));
  Test('indices(v, 1, 2E-4)', intarr([4, 7]));
  Test('indices(v, 1, 2E-3)', intarr([4, 7]));
  Test('indices(v, 1, 2E-2)', intarr([4, 7]));
  Test('indices(v, 1, 2E-1)', intarr([4, 7, 9]));

  Test('indices(v, 5)', intarr([3]));
  Test('indices(v, 5, 2E-6)', intarr([3]));
  Test('indices(v, 5, 2E-5)', intarr([3]));
  Test('indices(v, 5, 2E-4)', intarr([3]));
  Test('indices(v, 5, 2E-3)', intarr([3, 6]));
  Test('indices(v, 5, 2E-2)', intarr([3, 6, 8]));
  Test('indices(v, 5, 2E-1)', intarr([3, 6, 8]));

  Test('indices(v, 3)', intarr([1]));
  Test('indices(v, 3, 2E-6)', intarr([1]));
  Test('indices(v, 3, 2E-5)', intarr([1]));
  Test('indices(v, 3, 2E-4)', intarr([1]));
  Test('indices(v, 3, 2E-3)', intarr([1]));
  Test('indices(v, 3, 2E-2)', intarr([1]));
  Test('indices(v, 3, 2E-1)', intarr([1]));

  Test('indices(v, 8)', intarr([]));
  Test('indices(v, 8, 2E-6)', intarr([]));
  Test('indices(v, 8, 2E-5)', intarr([]));
  Test('indices(v, 8, 2E-4)', intarr([]));
  Test('indices(v, 8, 2E-3)', intarr([]));
  Test('indices(v, 8, 2E-2)', intarr([]));
  Test('indices(v, 8, 2E-1)', intarr([]));

  Test('indices(❨3, 4, 5, 1, 0, 4.999, 1, 5.01, 1.1, 0.0001❩, 5, 2E-3)', intarr([3, 6]));
  Test('indices(DebugObject("empty vector"), 5, 2E-3)', intarr([]));
  Test('indices(v, "thestral", 2E-4)', intarr([]));
  Test('indices(v, 1, -2E-4)', failure, 'A non-negative real number was expected as argument 3, but "-0.0002" was given.');

  Test('A ≔ ❨❨3, 4, 5, 1, 0❩, ❨4.999, 1, 5.01, 1.1, 0.0001❩❩;', null);

  Test('indices(A, 0)', asoarrex([idx2(1, 5)]));
  Test('indices(A, 0, 2E-6)', asoarrex([idx2(1, 5)]));
  Test('indices(A, 0, 2E-5)', asoarrex([idx2(1, 5)]));
  Test('indices(A, 0, 2E-4)', asoarrex([idx2(1, 5), idx2(2, 5)]));
  Test('indices(A, 0, 2E-3)', asoarrex([idx2(1, 5), idx2(2, 5)]));
  Test('indices(A, 0, 2E-2)', asoarrex([idx2(1, 5), idx2(2, 5)]));
  Test('indices(A, 0, 2E-1)', asoarrex([idx2(1, 5), idx2(2, 5)]));

  Test('indices(A, 1)', asoarrex([idx2(1, 4), idx2(2, 2)]));
  Test('indices(A, 1, 2E-6)', asoarrex([idx2(1, 4), idx2(2, 2)]));
  Test('indices(A, 1, 2E-5)', asoarrex([idx2(1, 4), idx2(2, 2)]));
  Test('indices(A, 1, 2E-4)', asoarrex([idx2(1, 4), idx2(2, 2)]));
  Test('indices(A, 1, 2E-3)', asoarrex([idx2(1, 4), idx2(2, 2)]));
  Test('indices(A, 1, 2E-2)', asoarrex([idx2(1, 4), idx2(2, 2)]));
  Test('indices(A, 1, 2E-1)', asoarrex([idx2(1, 4), idx2(2, 2), idx2(2, 4)]));

  Test('indices(A, 5)', asoarrex([idx2(1, 3)]));
  Test('indices(A, 5, 2E-6)', asoarrex([idx2(1, 3)]));
  Test('indices(A, 5, 2E-5)', asoarrex([idx2(1, 3)]));
  Test('indices(A, 5, 2E-4)', asoarrex([idx2(1, 3)]));
  Test('indices(A, 5, 2E-3)', asoarrex([idx2(1, 3), idx2(2, 1)]));
  Test('indices(A, 5, 2E-2)', asoarrex([idx2(1, 3), idx2(2, 1), idx2(2, 3)]));
  Test('indices(A, 5, 2E-1)', asoarrex([idx2(1, 3), idx2(2, 1), idx2(2, 3)]));

  Test('indices(A, 3)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 3, 2E-6)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 3, 2E-5)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 3, 2E-4)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 3, 2E-3)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 3, 2E-2)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 3, 2E-1)', asoarrex([idx2(1, 1)]));

  Test('indices(A, 8)', asoarrex([]));
  Test('indices(A, 8, 2E-6)', asoarrex([]));
  Test('indices(A, 8, 2E-5)', asoarrex([]));
  Test('indices(A, 8, 2E-4)', asoarrex([]));
  Test('indices(A, 8, 2E-3)', asoarrex([]));
  Test('indices(A, 8, 2E-2)', asoarrex([]));
  Test('indices(A, 8, 2E-1)', asoarrex([]));

  Test('indices(❨❨3, 4, 5, 1, 0❩, ❨4.999, 1, 5.01, 1.1, 0.0001❩❩, 5, 2E-3)', asoarrex([idx2(1, 3), idx2(2, 1)]));
  Test('indices(DebugObject("empty matrix"), 5, 2E-3)', asoarrex([]));
  Test('indices(A, "thestral", 2E-4)', asoarrex([]));
  Test('indices(A, 1, -2E-4)', failure, 'A non-negative real number was expected as argument 3, but "-0.0002" was given.');

  Test('v ≔ ❨1, i, 0.99, 5, 1 + i, 1.0001 + i, 4.99, 5.001, 1 + 0.999⋅i, 5❩;', null);

  Test('indices(v, 1)', intarr([1]));
  Test('indices(v, 1, 2E-6)', intarr([1]));
  Test('indices(v, 1, 2E-5)', intarr([1]));
  Test('indices(v, 1, 2E-4)', intarr([1]));
  Test('indices(v, 1, 2E-3)', intarr([1]));
  Test('indices(v, 1, 2E-2)', intarr([1, 3]));
  Test('indices(v, 1, 2E-1)', intarr([1, 3]));

  Test('indices(v, 5)', intarr([4, 10]));
  Test('indices(v, 5, 2E-6)', intarr([4, 10]));
  Test('indices(v, 5, 2E-5)', intarr([4, 10]));
  Test('indices(v, 5, 2E-4)', intarr([4, 10]));
  Test('indices(v, 5, 2E-3)', intarr([4, 8, 10]));
  Test('indices(v, 5, 2E-2)', intarr([4, 7, 8, 10]));
  Test('indices(v, 5, 2E-1)', intarr([4, 7, 8, 10]));

  Test('indices(v, 1 + i)', intarr([5]));
  Test('indices(v, 1 + i, 2E-6)', intarr([5]));
  Test('indices(v, 1 + i, 2E-5)', intarr([5]));
  Test('indices(v, 1 + i, 2E-4)', intarr([5, 6]));
  Test('indices(v, 1 + i, 2E-3)', intarr([5, 6, 9]));
  Test('indices(v, 1 + i, 2E-2)', intarr([5, 6, 9]));
  Test('indices(v, 1 + i, 2E-1)', intarr([5, 6, 9]));

  Test('indices(v, -i, 2E-4)', intarr([]));
  Test('indices(v, "cat", 2E-4)', intarr([]));

  Test('indices(❨1, i, 0.99, 5, 1 + i, 1.0001 + i, 4.99, 5.001, 1 + 0.999⋅i, 5❩, 1 + i, 2E-3)', intarr([5, 6, 9]));
  Test('indices(ComplexVector(DebugObject("empty vector")), 1 + i, 2E-3)', intarr([]));

  Test('A ≔ ❨❨1, i, 0.99, 5, 1 + i❩, ❨1.0001 + i, 4.99, 5.001, 1 + 0.999⋅i, 5❩❩;', null);

  Test('indices(A, 1)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 1, 2E-6)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 1, 2E-5)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 1, 2E-4)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 1, 2E-3)', asoarrex([idx2(1, 1)]));
  Test('indices(A, 1, 2E-2)', asoarrex([idx2(1, 1), idx2(1, 3)]));
  Test('indices(A, 1, 2E-1)', asoarrex([idx2(1, 1), idx2(1, 3)]));

  Test('indices(A, 5)', asoarrex([idx2(1, 4), idx2(2, 5)]));
  Test('indices(A, 5, 2E-6)', asoarrex([idx2(1, 4), idx2(2, 5)]));
  Test('indices(A, 5, 2E-5)', asoarrex([idx2(1, 4), idx2(2, 5)]));
  Test('indices(A, 5, 2E-4)', asoarrex([idx2(1, 4), idx2(2, 5)]));
  Test('indices(A, 5, 2E-3)', asoarrex([idx2(1, 4), idx2(2, 3), idx2(2, 5)]));
  Test('indices(A, 5, 2E-2)', asoarrex([idx2(1, 4), idx2(2, 2), idx2(2, 3), idx2(2, 5)]));
  Test('indices(A, 5, 2E-1)', asoarrex([idx2(1, 4), idx2(2, 2), idx2(2, 3), idx2(2, 5)]));

  Test('indices(A, 1 + i)', asoarrex([idx2(1, 5)]));
  Test('indices(A, 1 + i, 2E-6)', asoarrex([idx2(1, 5)]));
  Test('indices(A, 1 + i, 2E-5)', asoarrex([idx2(1, 5)]));
  Test('indices(A, 1 + i, 2E-4)', asoarrex([idx2(1, 5), idx2(2, 1)]));
  Test('indices(A, 1 + i, 2E-3)', asoarrex([idx2(1, 5), idx2(2, 1), idx2(2, 4)]));
  Test('indices(A, 1 + i, 2E-2)', asoarrex([idx2(1, 5), idx2(2, 1), idx2(2, 4)]));
  Test('indices(A, 1 + i, 2E-1)', asoarrex([idx2(1, 5), idx2(2, 1), idx2(2, 4)]));

  Test('indices(A, -i, 2E-4)', asoarrex([]));
  Test('indices(A, "cat", 2E-4)', asoarrex([]));

  Test('indices(❨❨1, i, 0.99, 5, 1 + i❩, ❨1.0001 + i, 4.99, 5.001, 1 + 0.999⋅i, 5❩❩, 1 + i, 2E-3)', asoarrex([idx2(1, 5), idx2(2, 1), idx2(2, 4)]));
  Test('indices(ComplexMatrix(DebugObject("empty matrix")), 1 + i, 2E-3)', asoarrex([]));

  Test('s ≔ "the emotional range of a teaspoon";', null);

  Test('indices(s, "a", 0.001)', intarr([12, 16, 24, 28]));
  Test('indices(s, "e", 0.001)', intarr([3, 5, 19, 27]));
  Test('indices(s, "t", 0.001)', intarr([1, 8, 26]));
  Test('indices(s, "y", 0.001)', intarr([]));
  Test('indices(s, "m", 0.001)', intarr([6]));
  Test('indices(s, "the", 0.001)', intarr([]));
  Test('indices(s, 1024, 0.001)', intarr([]));

  Test('indices("the emotional range of a teaspoon", "t", 0.001)', intarr([1, 8, 26]));
  Test('indices("", "t", 0.001)', intarr([]));

  Test('L ≔ ''(3, 4, 5, 1, 0, 4.999, 1, 5.01, 1.1, 0.0001);', null);

  Test('indices(L, 0)', intarr([5]));
  Test('indices(L, 0, 2E-6)', intarr([5]));
  Test('indices(L, 0, 2E-5)', intarr([5]));
  Test('indices(L, 0, 2E-4)', intarr([5, 10]));
  Test('indices(L, 0, 2E-3)', intarr([5, 10]));
  Test('indices(L, 0, 2E-2)', intarr([5, 10]));
  Test('indices(L, 0, 2E-1)', intarr([5, 10]));

  Test('indices(L, 1)', intarr([4, 7]));
  Test('indices(L, 1, 2E-6)', intarr([4, 7]));
  Test('indices(L, 1, 2E-5)', intarr([4, 7]));
  Test('indices(L, 1, 2E-4)', intarr([4, 7]));
  Test('indices(L, 1, 2E-3)', intarr([4, 7]));
  Test('indices(L, 1, 2E-2)', intarr([4, 7]));
  Test('indices(L, 1, 2E-1)', intarr([4, 7, 9]));

  Test('indices(L, 5)', intarr([3]));
  Test('indices(L, 5, 2E-6)', intarr([3]));
  Test('indices(L, 5, 2E-5)', intarr([3]));
  Test('indices(L, 5, 2E-4)', intarr([3]));
  Test('indices(L, 5, 2E-3)', intarr([3, 6]));
  Test('indices(L, 5, 2E-2)', intarr([3, 6, 8]));
  Test('indices(L, 5, 2E-1)', intarr([3, 6, 8]));

  Test('indices(L, 3)', intarr([1]));
  Test('indices(L, 3, 2E-6)', intarr([1]));
  Test('indices(L, 3, 2E-5)', intarr([1]));
  Test('indices(L, 3, 2E-4)', intarr([1]));
  Test('indices(L, 3, 2E-3)', intarr([1]));
  Test('indices(L, 3, 2E-2)', intarr([1]));
  Test('indices(L, 3, 2E-1)', intarr([1]));

  Test('indices(L, 8)', intarr([]));
  Test('indices(L, 8, 2E-6)', intarr([]));
  Test('indices(L, 8, 2E-5)', intarr([]));
  Test('indices(L, 8, 2E-4)', intarr([]));
  Test('indices(L, 8, 2E-3)', intarr([]));
  Test('indices(L, 8, 2E-2)', intarr([]));
  Test('indices(L, 8, 2E-1)', intarr([]));

  Test('indices(''(3, 4, 5, 1, 0, 4.999, 1, 5.01, 1.1, 0.0001), 5, 2E-3)', intarr([3, 6]));
  Test('indices(''(), 5, 2E-3)', intarr([]));
  Test('indices(L, "thestral", 2E-4)', intarr([]));
  Test('indices(L, 1, -2E-4)', failure, 'A non-negative real number was expected as argument 3, but "-0.0002" was given.');

  Test('S ≔ {3, 4, 5, 1, 0, 4.999, 1, 5.01, 1.1, 0.0001};', null);

  Test('indices(S, 0)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 0, 2E-6)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 0, 2E-5)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 0, 2E-4)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 0, 2E-3)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 0, 2E-2)', failure, 'Object type "set" is not a suitable container.');
  Test('indices(S, 0, 2E-1)', failure, 'Object type "set" is not a suitable container.');

  Test('S ≔ struct("name": "Albus Dumbledore", "age": 116, "role": "headmaster", "points": 100);', null);
  Test('indices(S, 116, 0.001)', intarr([2]));
  Test('indices(S, 116, 15)', intarr([2]));
  Test('indices(S, 116, 20)', intarr([2, 4]));

  Test('indices(1024, 1, 0.01)', failure, 'Object type "integer" is not a suitable container.');

  Test('indices()', failure, 'There is no version of this function that can take 0 argument(s).');
  Test('indices(L)', failure, 'There is no version of this function that can take 1 argument(s).');

  Test('delete(v); delete(A); delete(s); delete(S); delete(L)', success);


  //
  // Filtering and picking
  //

  // Filtering

  Test('v ≔ ❨4, 2, −1, 3, 5, 8, 0, 4, −7, 9, 8, 12, 2, 11, −4, 7❩;', null);

  Test('filter(v, IsNumber)', [4, 2, -1, 3, 5, 8, 0, 4, -7, 9, 8, 12, 2, 11, -4, 7]);
  Test('filter(v, odd)', [-1, 3, 5, -7, 9, 11, 7]);
  Test('filter(v, even)', [4, 2, 8, 0, 4, 8, 12, 2, -4]);
  Test('filter(v, IsPositive)', [4, 2, 3, 5, 8, 4, 9, 8, 12, 2, 11, 7]);
  Test('filter(v, IsNegative)', [-1, -7, -4]);
  Test('filter(v, IsZero)', [0]);
  Test('filter(v, IsPrime)', [2, 3, 5, 2, 11, 7]);
  Test('filter(v, x ↦ x > 10)', [12, 11]);
  Test('filter(v, x ↦ x > 100)', TAlgosimRealVector.Create);

  Test('filter(❨4, 2, −1, 3, 5, 8, 0, 4, −7, 9, 8, 12, 2, 11, −4, 7❩, IsPrime)', [2, 3, 5, 2, 11, 7]);
  Test('filter(DebugObject("empty vector"), IsPrime)', TAlgosimRealVector.Create);
  Test('filter(v, x ↦ x! < 100)', failure, 'A non-negative integer was expected as argument 1, but "-1" was given.');

  Test('filter(❨❨5, 3, 2, 1❩, ❨5, 7, 8, 9❩❩, odd)', failure, 'Cannot filter elements in an object of type "real matrix".');

  Test('v ≔ ❨6, 5.2, i, 1 + i, 3 − i, 0, 1024❩;', null);

  Test('filter(v, IsNumber)', [6, 5.2, ImaginaryUnit, 1 + ImaginaryUnit, 3 - ImaginaryUnit, 0, 1024]);
  Test('filter(v, IsReal)', [TASC(6), 5.2, 0, 1024]);
  Test('filter(v, IsInteger)', [TASC(6), 0, 1024]);
  Test('filter(v, IsZero)', [TASC(0)]);
  Test('filter(v, z ↦ abs(z) > 1000)', [TASC(1024)]);
  Test('filter(v, z ↦ abs(z) > 2000)', TAlgosimComplexVector.Create);

  Test('filter(❨6, 5.2, i, 1 + i, 3 − i, 0, 1024❩, IsReal)', [TASC(6), 5.2, 0, 1024]);
  Test('filter(ComplexVector(DebugObject("empty vector")), IsReal)', TAlgosimComplexVector.Create);
  Test('filter(v, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');

  Test('filter(❨❨5, 3, 2, i❩, ❨5, 7, 8, 9❩❩, IsReal)', failure, 'Cannot filter elements in an object of type "complex matrix".');

  Test('s ≔ "the emotional range of a teaspoon";', null);

  Test('filter(s, IsCharacter)', 'the emotional range of a teaspoon');
  Test('filter(s, ChrIsASCII)', 'the emotional range of a teaspoon');
  Test('filter(s, ChrIsLetter)', 'theemotionalrangeofateaspoon');
  Test('filter(s, ChrIsWhitespace)', '     ');

  Test('filter("the emotional range of a teaspoon", ChrIsLetter)', 'theemotionalrangeofateaspoon');
  Test('filter("", ChrIsLetter)', '');

  Test('L ≔ ''(4, 2, −1, 3, 5, 8, 0, 4, −7, 9, 8, 12, 2, 11, −4, 7);', null);

  Test('filter(L, IsNumber)', intarr([4, 2, -1, 3, 5, 8, 0, 4, -7, 9, 8, 12, 2, 11, -4, 7]));
  Test('filter(L, odd)', intarr([-1, 3, 5, -7, 9, 11, 7]));
  Test('filter(L, even)', intarr([4, 2, 8, 0, 4, 8, 12, 2, -4]));
  Test('filter(L, IsPositive)', intarr([4, 2, 3, 5, 8, 4, 9, 8, 12, 2, 11, 7]));
  Test('filter(L, IsNegative)', intarr([-1, -7, -4]));
  Test('filter(L, IsZero)', intarr([0]));
  Test('filter(L, IsPrime)', intarr([2, 3, 5, 2, 11, 7]));
  Test('filter(L, x ↦ x > 10)', intarr([12, 11]));
  Test('filter(L, x ↦ x > 100)', intarr([]));

  Test('filter(''(4, 2, −1, 3, 5, 8, 0, 4, −7, 9, 8, 12, 2, 11, −4, 7), IsPrime)', intarr([2, 3, 5, 2, 11, 7]));
  Test('filter(''(), IsPrime)', intarr([]));
  Test('filter(L, x ↦ x! < 100)', failure, 'A non-negative integer was expected as argument 1, but "-1" was given.');

  Test('S ≔ {4, 2, −1, 3, 5, 8, 0, 4, −7, 9, 8, 12, 2, 11, −4, 7};', null);

  Test('filter(S, IsNumber)', intset([4, 2, -1, 3, 5, 8, 0, 4, -7, 9, 8, 12, 2, 11, -4, 7]));
  Test('filter(S, odd)', intset([-1, 3, 5, -7, 9, 11, 7]));
  Test('filter(S, even)', intset([4, 2, 8, 0, 4, 8, 12, 2, -4]));
  Test('filter(S, IsPositive)', intset([4, 2, 3, 5, 8, 4, 9, 8, 12, 2, 11, 7]));
  Test('filter(S, IsNegative)', intset([-1, -7, -4]));
  Test('filter(S, IsZero)', intset([0]));
  Test('filter(S, IsPrime)', intset([2, 3, 5, 2, 11, 7]));
  Test('filter(S, x ↦ x > 10)', intset([12, 11]));
  Test('filter(S, x ↦ x > 100)', intset([]));

  Test('filter({4, 2, −1, 3, 5, 8, 0, 4, −7, 9, 8, 12, 2, 11, −4, 7}, IsPrime)', intset([2, 3, 5, 2, 11, 7]));
  Test('filter({}, IsPrime)', intset([]));
  Test('filter(S, x ↦ x! < 100)', failure, 'A non-negative integer was expected as argument 1, but'); // you don't know in what order the elements are tried

  Test('S ≔ struct("name": "Albus Dumbledore", "age": 116, "role": "headmaster")', TAlgosimStructure.CreateWithValue(
    ['name', 'age', 'role'],
    [ASO('Albus Dumbledore'), ASOInt(116), ASO('headmaster')]
  ));

  Test('filter(S, IsText)', TAlgosimStructure.CreateWithValue(
    ['name', 'role'],
    [ASO('Albus Dumbledore'), ASO('headmaster')]
  ));

  Test('filter(S, IsNumber)', TAlgosimStructure.CreateWithValue(
    ['age'],
    [ASOInt(116)]
  ));

  Test('filter(S, IsPixmap)', TAlgosimStructure.CreateWithValue(
    [],
    []
  ));

  Test('filter(S, IsPositive)', failure, 'An object of type numeric entity was expected as argument 1, but an object of type string was given.');

  Test('filter(struct("name": "Albus Dumbledore", "age": 116, "role": "headmaster"), IsText)', TAlgosimStructure.CreateWithValue(
    ['name', 'role'],
    [ASO('Albus Dumbledore'), ASO('headmaster')]
  ));

  Test('filter(struct(), IsPixmap)', TAlgosimStructure.CreateWithValue(
    [],
    []
  ));

  Test('filter()', failure, '2 argument(s) expected, but 0 given.');
  Test('filter(S)', failure, '2 argument(s) expected, but 1 given.');

  Test('S', TAlgosimStructure.CreateWithValue(
    ['name', 'age', 'role'],
    [ASO('Albus Dumbledore'), ASOInt(116), ASO('headmaster')]
  ));

  // Picking

  Test('v ≔ ❨6, 3, 0, 2, 1, −7, 8, 10, 5, 4, −2, 5, 1, 9, 8, 11❩;', null);

  Test('pick(v, IsNumber)', intarr([6, 3, 0, 2, 1, -7, 8, 10, 5, 4, -2, 5, 1, 9, 8, 11]));
  Test('pick(v, IsPositive)', intarr([6, 3, 2, 1, 8, 10, 5, 4, 5, 1, 9, 8, 11]));
  Test('pick(v, IsZero)', intarr([0]));
  Test('pick(v, IsNegative)', intarr([-7, -2]));
  Test('pick(v, odd)', intarr([3, 1, -7, 5, 5, 1, 9, 11]));
  Test('pick(v, even)', intarr([6, 0, 2, 8, 10, 4, -2, 8]));
  Test('pick(v, IsPrime)', intarr([3, 2, 5, 5, 11]));
  Test('pick(v, IsSound)', intarr([]));
  Test('pick(v, x ↦ x ≥ 10)', intarr([10, 11]));

  Test('pick(❨6, 3, 0, 2, 1, −7, 8, 10, 5, 4, −2, 5, 1, 9, 8, 11❩, odd)', intarr([3, 1, -7, 5, 5, 1, 9, 11]));
  Test('pick(DebugObject("empty vector"), odd)', intarr([]));
  Test('pick(v, x ↦ x! ≥ 10)', failure, 'A non-negative integer was expected as argument 1, but "-7" was given.');
  Test('pick(v, sin)', failure, 'The predicate was expected to return a boolean, but an object of type real number was returned.');

  Test('A ≔ ❨❨6, 3, 0, 2, 1, −7, 8, 10❩, ❨5, 4, −2, 5, 1, 9, 8, 11❩❩;', null);

  Test('pick(A, IsNumber)', intarr([6, 3, 0, 2, 1, -7, 8, 10, 5, 4, -2, 5, 1, 9, 8, 11]));
  Test('pick(A, IsPositive)', intarr([6, 3, 2, 1, 8, 10, 5, 4, 5, 1, 9, 8, 11]));
  Test('pick(A, IsZero)', intarr([0]));
  Test('pick(A, IsNegative)', intarr([-7, -2]));
  Test('pick(A, odd)', intarr([3, 1, -7, 5, 5, 1, 9, 11]));
  Test('pick(A, even)', intarr([6, 0, 2, 8, 10, 4, -2, 8]));
  Test('pick(A, IsPrime)', intarr([3, 2, 5, 5, 11]));
  Test('pick(A, IsSound)', intarr([]));
  Test('pick(A, x ↦ x ≥ 10)', intarr([10, 11]));

  Test('pick(❨❨6, 3, 0, 2, 1, −7, 8, 10❩, ❨5, 4, −2, 5, 1, 9, 8, 11❩❩, odd)', intarr([3, 1, -7, 5, 5, 1, 9, 11]));
  Test('pick(DebugObject("empty matrix"), odd)', intarr([]));
  Test('pick(A, x ↦ x! ≥ 10)', failure, 'A non-negative integer was expected as argument 1, but "-7" was given.');
  Test('pick(A, sin)', failure, 'The predicate was expected to return a boolean, but an object of type real number was returned.');

  Test('v ≔ ❨6, −2, 5.2, i, −i, 1024, 100⋅i, 0❩;', null);

  Test('pick(v, IsNumber)', asoarrex([ASOInt(6), ASOInt(-2), ASO(5.2), ASO(ImaginaryUnit), ASO(-ImaginaryUnit), ASOInt(1024), ASO(100*ImaginaryUnit), ASOInt(0)]));
  Test('pick(v, IsReal)', asoarrex([ASOInt(6), ASOInt(-2), ASO(5.2), ASOInt(1024), ASOInt(0)]));
  Test('pick(v, IsInteger)', asoarrex([ASOInt(6), ASOInt(-2), ASOInt(1024), ASOInt(0)]));
  Test('pick(v, IsZero)', asoarrex([ASOInt(0)]));
  Test('pick(v, IsPixmap)', asoarrex([]));
  Test('pick(v, z ↦ abs(z) ≥ 100)', asoarrex([ASOInt(1024), ASO(100*ImaginaryUnit)]));
  Test('pick(v, z ↦ abs(z) ≥ 10000)', asoarrex([]));
  Test('pick(v, sin)', failure, 'The predicate was expected to return a boolean, but an object of type complex number was returned.');

  Test('pick(❨6, −2, 5.2, i, −i, 1024, 100⋅i, 0❩, IsInteger)', asoarrex([ASOInt(6), ASOInt(-2), ASOInt(1024), ASOInt(0)]));
  Test('pick(ComplexVector(DebugObject("empty vector")), IsInteger)', asoarrex([]));
  Test('pick(v, IsPositive)', failure, 'Cannot determine the sign of an object of type "complex number".');

  Test('A ≔ ❨❨6, −2, 5.2, i❩, ❨−i, 1024, 100⋅i, 0❩❩;', null);

  Test('pick(A, IsNumber)', asoarrex([ASOInt(6), ASOInt(-2), ASO(5.2), ASO(ImaginaryUnit), ASO(-ImaginaryUnit), ASOInt(1024), ASO(100*ImaginaryUnit), ASOInt(0)]));
  Test('pick(A, IsReal)', asoarrex([ASOInt(6), ASOInt(-2), ASO(5.2), ASOInt(1024), ASOInt(0)]));
  Test('pick(A, IsInteger)', asoarrex([ASOInt(6), ASOInt(-2), ASOInt(1024), ASOInt(0)]));
  Test('pick(A, IsZero)', asoarrex([ASOInt(0)]));
  Test('pick(A, IsPixmap)', asoarrex([]));
  Test('pick(A, z ↦ abs(z) ≥ 100)', asoarrex([ASOInt(1024), ASO(100*ImaginaryUnit)]));
  Test('pick(A, z ↦ abs(z) ≥ 10000)', asoarrex([]));

  Test('pick(❨❨6, −2, 5.2, i❩, ❨−i, 1024, 100⋅i, 0❩❩, IsInteger)', asoarrex([ASOInt(6), ASOInt(-2), ASOInt(1024), ASOInt(0)]));
  Test('pick(ComplexMatrix(DebugObject("empty matrix")), IsInteger)', asoarrex([]));
  Test('pick(A, IsPositive)', failure, 'Cannot determine the sign of an object of type "complex number".');
  Test('pick(A, sin)', failure, 'The predicate was expected to return a boolean, but an object of type complex number was returned.');

  Test('s ≔ "Turn to page 394.";', null);

  Test('pick(s, ChrIsASCII)', strarr(['T', 'u', 'r', 'n', ' ', 't', 'o', ' ', 'p', 'a', 'g', 'e', ' ', '3', '9', '4', '.']));
  Test('pick(s, ChrIsLetter)', strarr(['T', 'u', 'r', 'n', 't', 'o', 'p', 'a', 'g', 'e']));
  Test('pick(s, ChrIsDigit)', strarr(['3', '9', '4']));
  Test('pick(s, ChrIsPunctuation)', strarr(['.']));
  Test('pick(s, ChrIsControl)', strarr([]));

  Test('pick("Turn to page 394.", ChrIsDigit)', strarr(['3', '9', '4']));
  Test('pick("", ChrIsDigit)', strarr([]));
  Test('pick(s, odd)', failure, 'An object of type integer was expected as argument 1, but an object of type string was given.');
  Test('pick(s, length)', failure, 'The predicate was expected to return a boolean, but an object of type integer was returned.');

  Test('L ≔ ''(6, 3, 0, 2, 1, −7, 8, 10, 5, 4, −2, 5, 1, 9, 8, 11);', null);

  Test('pick(L, IsNumber)', intarr([6, 3, 0, 2, 1, -7, 8, 10, 5, 4, -2, 5, 1, 9, 8, 11]));
  Test('pick(L, IsPositive)', intarr([6, 3, 2, 1, 8, 10, 5, 4, 5, 1, 9, 8, 11]));
  Test('pick(L, IsZero)', intarr([0]));
  Test('pick(L, IsNegative)', intarr([-7, -2]));
  Test('pick(L, odd)', intarr([3, 1, -7, 5, 5, 1, 9, 11]));
  Test('pick(L, even)', intarr([6, 0, 2, 8, 10, 4, -2, 8]));
  Test('pick(L, IsPrime)', intarr([3, 2, 5, 5, 11]));
  Test('pick(L, IsSound)', intarr([]));
  Test('pick(L, x ↦ x ≥ 10)', intarr([10, 11]));

  Test('pick(''(6, 3, 0, 2, 1, −7, 8, 10, 5, 4, −2, 5, 1, 9, 8, 11), odd)', intarr([3, 1, -7, 5, 5, 1, 9, 11]));
  Test('pick(''(), odd)', intarr([]));
  Test('pick(L, x ↦ x! ≥ 10)', failure, 'A non-negative integer was expected as argument 1, but "-7" was given.');
  Test('pick(L, sin)', failure, 'The predicate was expected to return a boolean, but an object of type real number was returned.');

  Test('S ≔ {6, 3, 0, 2, 1, −7, 8, 10, 5, 4, −2, 5, 1, 9, 8, 11};', null);

  Test('ToSet(pick(S, IsNumber))', intset([6, 3, 0, 2, 1, -7, 8, 10, 5, 4, -2, 5, 1, 9, 8, 11]));
  Test('ToSet(pick(S, IsPositive))', intset([6, 3, 2, 1, 8, 10, 5, 4, 5, 1, 9, 8, 11]));
  Test('ToSet(pick(S, IsZero))', intset([0]));
  Test('ToSet(pick(S, IsNegative))', intset([-7, -2]));
  Test('ToSet(pick(S, odd))', intset([3, 1, -7, 5, 5, 1, 9, 11]));
  Test('ToSet(pick(S, even))', intset([6, 0, 2, 8, 10, 4, -2, 8]));
  Test('ToSet(pick(S, IsPrime))', intset([3, 2, 5, 5, 11]));
  Test('ToSet(pick(S, IsSound))', intset([]));
  Test('ToSet(pick(S, x ↦ x ≥ 10))', intset([10, 11]));

  Test('ToSet(pick({6, 3, 0, 2, 1, −7, 8, 10, 5, 4, −2, 5, 1, 9, 8, 11}, odd))', intset([3, 1, -7, 5, 5, 1, 9, 11]));
  Test('ToSet(pick({}, odd))', intset([]));
  Test('ToSet(pick(S, x ↦ x! ≥ 10))', failure, 'A non-negative integer was expected as argument 1, but "-7" was given.');

  Test('S ≔ struct("name": "Albus Dumbledore", "age": 116, "role": "headmaster")', TAlgosimStructure.CreateWithValue(
    ['name', 'age', 'role'],
    [ASO('Albus Dumbledore'), ASOInt(116), ASO('headmaster')]
  ));

  Test('pick(S, IsText)', strarr(['Albus Dumbledore', 'headmaster']));
  Test('pick(S, IsNumber)', intarr([116]));

  // - picking with level specification

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null);

  Test('pick(Mary, IsStructure)',
    asoarrex([Jane, Mike]));

  Test('pick(Mary, IsStructure, 1)',
    asoarrex([Jane, Mike]));

  Test('pick(Mary, IsStructure, 2)',
    asoarrex([Name('Jane', 'Smith'), Name('Mike', 'Doe')]));

  Test('pick(Mary, IsStructure, 2) @ member("first")',
    strarr(['Jane', 'Mike']));

  Test('pick(Mary, IsStructure, 3)',
    asoarrex([]));

  Test('pick(Mary, IsStructure, 0)',
    failure, 'A positive integer was expected as argument 3, but "0" was given.');

  Test('pick(Mary, sin, 1)',
    failure, 'An object of type number was expected as argument 1, but an object of type structure was given.');

  Test('pick(Mary, length, 1)',
    failure, 'The predicate was expected to return a boolean, but an object of type integer was returned.');

  Test('Mary ≔ ''(Jane, Mike);', null);

  Test('pick(Mary, IsStructure)',
    asoarrex([Jane, Mike]));

  Test('pick(Mary, IsStructure, 1)',
    asoarrex([Jane, Mike]));

  Test('pick(Mary, IsStructure, 2)',
    asoarrex([Name('Jane', 'Smith'), Name('Mike', 'Doe')]));

  Test('pick(Mary, IsStructure, 2) @ member("first")',
    strarr(['Jane', 'Mike']));

  Test('pick(Mary, IsStructure, 3)',
    asoarrex([]));

  Test('L ≔ ''(123, ''(10, ''(1024, 256)), ''(20, ''(512, 2048)));', null);

  Test('pick(L, IsInteger, 1)', intarr([123]));
  Test('pick(L, IsInteger, 2)', intarr([10, 20]));
  Test('pick(L, IsInteger, 3)', intarr([1024, 256, 512, 2048]));
  Test('pick(L, IsInteger, 4)', intarr([]));

  // - recursive picking

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null);

  Test('PickRecursive(Mary, IsInteger)', intarr([40, 45]));
  Test('PickRecursive(Mary, IsText)', strarr(['Jane', 'Smith', 'female', 'Mike', 'Doe', 'male']));

  Test('PickRecursive(Mary, sin)', failure, 'An object of type number was expected as argument 1, but an object of type structure was given.');
  Test('PickRecursive(Mary, type)', failure, 'The predicate was expected to return a boolean, but an object of type string was returned.');

  Test('L ≔ ''(123, ''(10, ''(1024, 256)), ''(20, ''(512, 2048)));', null);

  Test('PickRecursive(L, IsInteger)', intarr([123, 10, 1024, 256, 20, 512, 2048]));


  //
  // Application
  //

  Test('v ≔ SequenceVector(10)', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);

  Test('apply(v, sqr)', [1, 4, 9, 16, 25, 36, 49, 64, 81, 100]);
  Test('apply(v, inv)', [1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]);

  Test('apply(SequenceVector(10), sqr)', [1, 4, 9, 16, 25, 36, 49, 64, 81, 100]);
  Test('apply(DebugObject("empty vector"), sqr)', TAlgosimRealVector.Create);

  Test('v @ sqr', [1, 4, 9, 16, 25, 36, 49, 64, 81, 100]);
  Test('v @ inv', [1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]);

  Test('v @ type', failure, 'Real number application function must return a real number.');
  {But:} Test('ToList(v) @ type', strarr(['real number', 'real number', 'real number', 'real number', 'real number', 'real number', 'real number', 'real number', 'real number', 'real number']));

  Test('A ≔ ❨❨1, 2, 3, 4, 5❩, ❨6, 7, 8, 9, 10❩❩', 5, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);

  Test('apply(A, sqr)', 5, [1, 4, 9, 16, 25, 36, 49, 64, 81, 100]);
  Test('apply(A, inv)', 5, [1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]);

  Test('apply(❨❨1, 2, 3, 4, 5❩, ❨6, 7, 8, 9, 10❩❩, sqr)', 5, [1, 4, 9, 16, 25, 36, 49, 64, 81, 100]);
  Test('apply(DebugObject("empty matrix"), sqr)', TAlgosimRealMatrix.Create);

  Test('A @ sqr', 5, [1, 4, 9, 16, 25, 36, 49, 64, 81, 100]);
  Test('A @ inv', 5, [1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]);

  Test('A @ type', failure, 'Real number application function must return a real number.');

  Test('v ≔ ❨1, −1, i, 1, 2, 3❩', [1, -1, ImaginaryUnit, 1, 2, 3]);

  Test('apply(v, sqr)', [TASC(1), 1, -1, 1, 4, 9]);
  Test('apply(v, inv)', [1, -1, -ImaginaryUnit, 1, 1/2, 1/3]);

  Test('apply(❨1, −1, i, 1, 2, 3❩, sqr)', [TASC(1), 1, -1, 1, 4, 9]);
  Test('apply(ComplexVector(DebugObject("empty vector")), sqr)', TAlgosimComplexVector.Create);

  Test('v @ sqr', [TASC(1), 1, -1, 1, 4, 9]);
  Test('v @ inv', [1, -1, -ImaginaryUnit, 1, 1/2, 1/3]);

  Test('v @ type', failure, 'Complex number application function must return a complex number.');

  Test('A ≔ ❨❨1, −1, i❩, ❨1, 2, 3❩❩', 3, [1, -1, ImaginaryUnit, 1, 2, 3]);

  Test('apply(A, sqr)', 3, [TASC(1), 1, -1, 1, 4, 9]);
  Test('apply(A, inv)', 3, [1, -1, -ImaginaryUnit, 1, 1/2, 1/3]);

  Test('apply(❨❨1, −1, i❩, ❨1, 2, 3❩❩, sqr)', 3, [TASC(1), 1, -1, 1, 4, 9]);
  Test('apply(ComplexMatrix(DebugObject("empty matrix")), sqr)', TAlgosimComplexMatrix.Create);

  Test('A @ sqr', 3, [TASC(1), 1, -1, 1, 4, 9]);
  Test('A @ inv', 3, [1, -1, -ImaginaryUnit, 1, 1/2, 1/3]);

  Test('A @ type', failure, 'Complex number application function must return a complex number.');

  Test('s ≔ "Swimming across a voltage-gated ion channel"', 'Swimming across a voltage-gated ion channel');

  Test('apply(s, UpperCase)', 'SWIMMING ACROSS A VOLTAGE-GATED ION CHANNEL');
  Test('s @ UpperCase', 'SWIMMING ACROSS A VOLTAGE-GATED ION CHANNEL');

  Test('apply("Swimming across a voltage-gated ion channel", UpperCase)', 'SWIMMING ACROSS A VOLTAGE-GATED ION CHANNEL');
  Test('"Swimming across a voltage-gated ion channel" @ UpperCase', 'SWIMMING ACROSS A VOLTAGE-GATED ION CHANNEL');

  Test('apply("", UpperCase)', '');
  Test('"" @ UpperCase', '');

  Test('s @ length', failure, 'Character application function must return a single character.');
  Test('s @ (s ↦ 2⋅s)', failure, 'Character application function must return a single character.');

  Test('L ≔ SequenceList(10)', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]));

  Test('apply(L, sqr)', intarr([1, 4, 9, 16, 25, 36, 49, 64, 81, 100]));
  Test('apply(L, inv)', asoarr([1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]));

  Test('apply(SequenceList(10), sqr)', intarr([1, 4, 9, 16, 25, 36, 49, 64, 81, 100]));
  Test('apply(''(), sqr)', intarr([]));

  Test('L @ sqr', intarr([1, 4, 9, 16, 25, 36, 49, 64, 81, 100]));
  Test('L @ inv', asoarr([1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]));

  Test('S ≔ ToSet(SequenceList(10))', intset([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]));

  Test('apply(S, sqr)', intset([1, 4, 9, 16, 25, 36, 49, 64, 81, 100]));
  Test('apply(S, inv)', asoset([1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]));

  Test('apply(ToSet(SequenceList(10)), sqr)', intset([1, 4, 9, 16, 25, 36, 49, 64, 81, 100]));
  Test('apply({}, sqr)', intset([]));

  Test('S @ sqr', intset([1, 4, 9, 16, 25, 36, 49, 64, 81, 100]));
  Test('S @ inv', asoset([1, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9, 1/10]));

  // With level specification

  Test('L ≔ ''(''(1, 2, 3), ''(4, 5, 6));', null);

  Test('apply(L, sqr, 2)', asoarrex([intarr([1, 4, 9]), intarr([16, 25, 36])]));

  Test('apply(L, sqr, 1)', failure, 'An object of type numeric entity was expected as argument 1, but an object of type array was given.');
  Test('apply(L, sqr, 3)', failure, 'Cannot apply function to elements in an object of type "integer".');

  // Conditional application

  Test('v ≔ ❨3, 5, 4, −2, 1, 7, −2, 0, 5, 8, 1, 10, 11, 2❩;', null);

  Test('ApplyIf(v, IsNonNegative, factorial)', [6, 120, 24, -2, 1, 5040, -2, 1, 120, 40320, 1, 3628800, 39916800, 2]);
  Test('ApplyIf(v, IsZero, factorial)', [3, 5, 4, -2, 1, 7, -2, 1, 5, 8, 1, 10, 11, 2]);
  Test('ApplyIf(v, odd, x ↦ 2⋅x)', [6, 10, 4, -2, 2, 14, -2, 0, 10, 8, 2, 10, 22, 2]);
  Test('ApplyIf(v, IsPrime, NextPrime)', [5, 7, 4, -2, 1, 11, -2, 0, 7, 8, 1, 10, 13, 3]);

  Test('ApplyIf(v, IsPrime, type)', failure, 'Real number application function must return a real number.');
  Test('ApplyIf(v, IsNumber, factorial)', failure, 'A non-negative integer was expected as argument 1, but "-2" was given.');

  Test('ApplyIf(❨3, 5, 4, −2, 1, 7, −2, 0, 5, 8, 1, 10, 11, 2❩, IsPrime, NextPrime)', [5, 7, 4, -2, 1, 11, -2, 0, 7, 8, 1, 10, 13, 3]);
  Test('ApplyIf(DebugObject("empty vector"), IsPrime, NextPrime)', TAlgosimRealVector.Create);

  Test('A ≔ ❨❨3, 5, 4, −2, 1, 7, −2❩, ❨0, 5, 8, 1, 10, 11, 2❩❩;', null);

  Test('ApplyIf(A, IsNonNegative, factorial)', 7, [6, 120, 24, -2, 1, 5040, -2, 1, 120, 40320, 1, 3628800, 39916800, 2]);
  Test('ApplyIf(A, IsZero, factorial)', 7, [3, 5, 4, -2, 1, 7, -2, 1, 5, 8, 1, 10, 11, 2]);
  Test('ApplyIf(A, odd, x ↦ 2⋅x)', 7, [6, 10, 4, -2, 2, 14, -2, 0, 10, 8, 2, 10, 22, 2]);
  Test('ApplyIf(A, IsPrime, NextPrime)', 7, [5, 7, 4, -2, 1, 11, -2, 0, 7, 8, 1, 10, 13, 3]);

  Test('ApplyIf(A, IsPrime, type)', failure, 'Real number application function must return a real number.');
  Test('ApplyIf(A, IsNumber, factorial)', failure, 'A non-negative integer was expected as argument 1, but "-2" was given.');

  Test('ApplyIf(❨❨3, 5, 4, −2, 1, 7, −2❩, ❨0, 5, 8, 1, 10, 11, 2❩❩, IsPrime, NextPrime)', 7, [5, 7, 4, -2, 1, 11, -2, 0, 7, 8, 1, 10, 13, 3]);
  Test('ApplyIf(DebugObject("empty matrix"), IsPrime, NextPrime)', TAlgosimRealMatrix.Create);

  Test('v ≔ ❨5, 1.5, 0, −4, i, −i, 10, 1❩;', null);

  Test('ApplyIf(v, IsNumber, sqr)', [25, 2.25, 0, 16, -1, -1, 100, TASC(1)]);
  Test('ApplyIf(v, IsReal, sqr)', [25, 2.25, 0, 16, ImaginaryUnit, -ImaginaryUnit, 100, 1]);
  Test('ApplyIf(v, IsComplex, sqr)', [5, 1.5, 0, -4, -1, -1, 10, TASC(1)]);
  Test('ApplyIf(v, IsSound, sqr)', [5, 1.5, 0, -4, ImaginaryUnit, -ImaginaryUnit, 10, 1]);

  Test('ApplyIf(❨5, 1.5, 0, −4, i, −i, 10, 1❩, IsReal, sqr)', [25, 2.25, 0, 16, ImaginaryUnit, -ImaginaryUnit, 100, 1]);
  Test('ApplyIf(ComplexVector(DebugObject("empty vector")), IsReal, sqr)', TAlgosimComplexVector.Create);

  Test('ApplyIf(v, IsComplex, type)', failure, 'Complex number application function must return a complex number.');
  Test('ApplyIf(v, IsComplex, IsPositive)', failure, 'Cannot determine the sign of an object of type "complex number".');

  Test('A ≔ ❨❨5, 1.5, 0, −4❩, ❨i, −i, 10, 1❩❩;', null);

  Test('ApplyIf(A, IsNumber, sqr)', 4, [25, 2.25, 0, 16, -1, -1, 100, TASC(1)]);
  Test('ApplyIf(A, IsReal, sqr)', 4, [25, 2.25, 0, 16, ImaginaryUnit, -ImaginaryUnit, 100, 1]);
  Test('ApplyIf(A, IsComplex, sqr)', 4, [5, 1.5, 0, -4, -1, -1, 10, TASC(1)]);
  Test('ApplyIf(A, IsSound, sqr)', 4, [5, 1.5, 0, -4, ImaginaryUnit, -ImaginaryUnit, 10, 1]);

  Test('ApplyIf(❨❨5, 1.5, 0, −4❩, ❨i, −i, 10, 1❩❩, IsReal, sqr)', 4, [25, 2.25, 0, 16, ImaginaryUnit, -ImaginaryUnit, 100, 1]);
  Test('ApplyIf(ComplexMatrix(DebugObject("empty matrix")), IsReal, sqr)', TAlgosimComplexMatrix.Create);

  Test('ApplyIf(A, IsComplex, type)', failure, 'Complex number application function must return a complex number.');
  Test('ApplyIf(A, IsComplex, IsPositive)', failure, 'Cannot determine the sign of an object of type "complex number".');

  Test('ApplyIf("Hello, World!", ChrIsUpperCase, c ↦ chr(succ(ord(c))))', 'Iello, Xorld!');
  Test('ApplyIf("", ChrIsUpperCase, c ↦ chr(succ(ord(c))))', '');
  Test('ApplyIf("Hello, World!", ChrIsLetter, c ↦ chr(succ(ord(c))))', 'Ifmmp, Xpsme!');

  Test('L ≔ ''(3, 5, 4, −2, 1, 7, −2, 0, 5, 8, 1, 10, 11, 2);', null);

  Test('ApplyIf(L, IsNonNegative, factorial)', intarr([6, 120, 24, -2, 1, 5040, -2, 1, 120, 40320, 1, 3628800, 39916800, 2]));
  Test('ApplyIf(L, IsZero, factorial)', intarr([3, 5, 4, -2, 1, 7, -2, 1, 5, 8, 1, 10, 11, 2]));
  Test('ApplyIf(L, odd, x ↦ 2⋅x)', intarr([6, 10, 4, -2, 2, 14, -2, 0, 10, 8, 2, 10, 22, 2]));
  Test('ApplyIf(L, IsPrime, NextPrime)', intarr([5, 7, 4, -2, 1, 11, -2, 0, 7, 8, 1, 10, 13, 3]));

  Test('ApplyIf(L, IsPrime, type)', asoarr(['integer', 'integer', 4, -2, 1, 'integer', -2, 0, 'integer', 8, 1, 10, 'integer', 'integer']));
  Test('ApplyIf(L, IsNumber, factorial)', failure, 'A non-negative integer was expected as argument 1, but "-2" was given.');

  Test('ApplyIf(''(3, 5, 4, −2, 1, 7, −2, 0, 5, 8, 1, 10, 11, 2), IsPrime, NextPrime)', intarr([5, 7, 4, -2, 1, 11, -2, 0, 7, 8, 1, 10, 13, 3]));
  Test('ApplyIf(''(), IsPrime, NextPrime)', intarr([]));

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null);

  Test('ApplyIf(Jane, IsInteger, succ)',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Jane', 'Smith'), ASO('female'), ASOInt(41)]
    )
  );

  // Conditional application with level specification

  Test('L ≔ ''(''(1, 2, 3), ''(4, 5, 6));', null);

  Test('ApplyIf(L, odd, x ↦ 2⋅x, 2)',
    asoarrex(
      [
        intarr([2, 2, 6]),
        intarr([4, 10, 6])
      ]
    )
  );

  Test('ApplyIf(L, even, x ↦ x/2, 2)',
    asoarrex(
      [
        intarr([1, 1, 3]),
        intarr([2, 5, 3])
      ]
    )
  );


  //
  // Replacement
  //

  // Object-based replacement

  Test('v ≔ ❨4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6❩;', null);

  Test('ReplaceAll(v, 0, 10)', [4, 10, 2, 3, 5, 6, 8, 1, 2, 10, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(v, 2, 64)', [4, 0, 64, 3, 5, 6, 8, 1, 64, 0, 1, 64, 4, 3, 5, 6]);
  Test('ReplaceAll(v, 100, 64)', [4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(v, "cat", 64)', [4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(v, "", 64)', [4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6]);

  Test('ReplaceAll(v, 2, "cat")', failure, 'Object isn''t a real number.');

  Test('ReplaceAll(❨4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6❩, 2, 64)', [4, 0, 64, 3, 5, 6, 8, 1, 64, 0, 1, 64, 4, 3, 5, 6]);
  Test('ReplaceAll(DebugObject("empty vector"), 2, 64)', TAlgosimRealVector.Create);

  Test('A ≔ ❨❨4, 0, 2, 3, 5, 6, 8, 1❩, ❨2, 0, 1, 2, 4, 3, 5, 6❩❩;', null);

  Test('ReplaceAll(A, 0, 10)', 8, [4, 10, 2, 3, 5, 6, 8, 1, 2, 10, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(A, 2, 64)', 8, [4, 0, 64, 3, 5, 6, 8, 1, 64, 0, 1, 64, 4, 3, 5, 6]);
  Test('ReplaceAll(A, 100, 64)', 8, [4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(A, "cat", 64)', 8, [4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(A, "", 64)', 8, [4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6]);

  Test('ReplaceAll(A, 2, "cat")', failure, 'Object isn''t a real number.');

  Test('ReplaceAll(❨❨4, 0, 2, 3, 5, 6, 8, 1❩, ❨2, 0, 1, 2, 4, 3, 5, 6❩❩, 2, 64)', 8, [4, 0, 64, 3, 5, 6, 8, 1, 64, 0, 1, 64, 4, 3, 5, 6]);
  Test('ReplaceAll(DebugObject("empty matrix"), 2, 64)', TAlgosimRealMatrix.Create);

  Test('v ≔ ❨4, 0, 2, 3, 5, 6, i, 1, 2, 0, 1, 2, 4, 3, 5, 6❩;', null);

  Test('ReplaceAll(v, 0, 10)', [4, 10, 2, 3, 5, 6, ImaginaryUnit, 1, 2, 10, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(v, 2, 64)', [4, 0, 64, 3, 5, 6, ImaginaryUnit, 1, 64, 0, 1, 64, 4, 3, 5, 6]);
  Test('ReplaceAll(v, 2, i)', [4, 0, ImaginaryUnit, 3, 5, 6, ImaginaryUnit, 1, ImaginaryUnit, 0, 1, ImaginaryUnit, 4, 3, 5, 6]);
  Test('ReplaceAll(v, i, 5⋅i)', [4, 0, 2, 3, 5, 6, 5*ImaginaryUnit, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(v, 100, 64)', [4, 0, 2, 3, 5, 6, ImaginaryUnit, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(v, "cat", 64)', [4, 0, 2, 3, 5, 6, ImaginaryUnit, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(v, "", 64)', [4, 0, 2, 3, 5, 6, ImaginaryUnit, 1, 2, 0, 1, 2, 4, 3, 5, 6]);

  Test('ReplaceAll(v, 2, "cat")', failure, 'Object isn''t a complex number.');

  Test('ReplaceAll(❨4, 0, 2, 3, 5, 6, i, 1, 2, 0, 1, 2, 4, 3, 5, 6❩, 2, 64)', [4, 0, 64, 3, 5, 6, ImaginaryUnit, 1, 64, 0, 1, 64, 4, 3, 5, 6]);
  Test('ReplaceAll(ComplexVector(DebugObject("empty vector")), 2, 64)', TAlgosimComplexVector.Create);

  Test('A ≔ ❨❨4, 0, 2, 3, 5, 6, i, 1❩, ❨2, 0, 1, 2, 4, 3, 5, 6❩❩;', null);

  Test('ReplaceAll(A, 0, 10)', 8, [4, 10, 2, 3, 5, 6, ImaginaryUnit, 1, 2, 10, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(A, 2, 64)', 8, [4, 0, 64, 3, 5, 6, ImaginaryUnit, 1, 64, 0, 1, 64, 4, 3, 5, 6]);
  Test('ReplaceAll(A, 2, i)', 8, [4, 0, ImaginaryUnit, 3, 5, 6, ImaginaryUnit, 1, ImaginaryUnit, 0, 1, ImaginaryUnit, 4, 3, 5, 6]);
  Test('ReplaceAll(A, i, 5⋅i)', 8, [4, 0, 2, 3, 5, 6, 5*ImaginaryUnit, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(A, 100, 64)', 8, [4, 0, 2, 3, 5, 6, ImaginaryUnit, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(A, "cat", 64)', 8, [4, 0, 2, 3, 5, 6, ImaginaryUnit, 1, 2, 0, 1, 2, 4, 3, 5, 6]);
  Test('ReplaceAll(A, "", 64)', 8, [4, 0, 2, 3, 5, 6, ImaginaryUnit, 1, 2, 0, 1, 2, 4, 3, 5, 6]);

  Test('ReplaceAll(A, 2, "cat")', failure, 'Object isn''t a complex number.');

  Test('ReplaceAll(❨❨4, 0, 2, 3, 5, 6, i, 1❩, ❨2, 0, 1, 2, 4, 3, 5, 6❩❩, 2, 64)', 8, [4, 0, 64, 3, 5, 6, ImaginaryUnit, 1, 64, 0, 1, 64, 4, 3, 5, 6]);
  Test('ReplaceAll(ComplexMatrix(DebugObject("empty matrix")), 2, 64)', TAlgosimComplexMatrix.Create);

  Test('ReplaceAll("a, b, c, d, e", ",", ";")', 'a; b; c; d; e');
  Test('ReplaceAll("a, b, c, d, e", ";", "!")', 'a, b, c, d, e');
  Test('ReplaceAll("a, b, c, d, e", IdentityMatrix(100), "!")', 'a, b, c, d, e');
  Test('ReplaceAll("a, b, c, d, e", ",", IdentityMatrix(100))', failure, 'An object of type "real matrix" cannot be converted to a character.');
  Test('ReplaceAll("a, b, c, d, e", ",", "<>")', failure, 'A string of length 2 cannot be converted to a character.');
  Test('ReplaceAll("", ",", ";")', '');
  Test('ReplaceAll("a, b, c, d, e", "a, b", "@")', 'a, b, c, d, e'); {cf. StringReplace}

  Test('L ≔ ''(4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6);', null);

  Test('ReplaceAll(L, 0, 10)', intarr([4, 10, 2, 3, 5, 6, 8, 1, 2, 10, 1, 2, 4, 3, 5, 6]));
  Test('ReplaceAll(L, 2, 64)', intarr([4, 0, 64, 3, 5, 6, 8, 1, 64, 0, 1, 64, 4, 3, 5, 6]));
  Test('ReplaceAll(L, 100, 64)', intarr([4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6]));
  Test('ReplaceAll(L, "cat", 64)', intarr([4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6]));
  Test('ReplaceAll(L, "", 64)', intarr([4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6]));

  Test('ReplaceAll(L, 2, "cat")', asoarr([4, 0, 'cat', 3, 5, 6, 8, 1, 'cat', 0, 1, 'cat', 4, 3, 5, 6]));

  Test('ReplaceAll(''(4, 0, 2, 3, 5, 6, 8, 1, 2, 0, 1, 2, 4, 3, 5, 6), 2, 64)', intarr([4, 0, 64, 3, 5, 6, 8, 1, 64, 0, 1, 64, 4, 3, 5, 6]));
  Test('ReplaceAll(''(), 2, 64)', intarr([]));

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null);

  Test('ReplaceAll(Jane, 40, 5)',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Jane', 'Smith'), ASO('female'), ASOInt(5)]
    )
  );

  Test('ReplaceAll(ToSet(v), 1, 22)', failure, 'Object type "set" is not a suitable container.');

  // - with level specification

  Test('L ≔ ''(''(1, 2, 3), ''(4, 5, 6));', null);

  Test('ReplaceAll(L, 1, 10)',
    asoarrex(
      [
        intarr([1, 2, 3]),
        intarr([4, 5, 6])
      ]
    )
  );

  Test('ReplaceAll(L, 1, 10, 2)',
    asoarrex(
      [
        intarr([10, 2, 3]),
        intarr([4, 5, 6])
      ]
    )
  );

  // Predicate-based replacement

  Test('v ≔ ❨6, 0, 3, −2, 1, −4, 2, 0, 7, 10, −4, 5, 3, 2, 2, 11❩;', null);

  Test('ReplaceIf(v, IsNumber, 10)', [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]);
  Test('ReplaceIf(v, IsPositive, 64)', [64, 0, 64, -2, 64, -4, 64, 0, 64, 64, -4, 64, 64, 64, 64, 64]);
  Test('ReplaceIf(v, IsZero, 1.2)', [6, 1.2, 3, -2, 1, -4, 2, 1.2, 7, 10, -4, 5, 3, 2, 2, 11]);
  Test('ReplaceIf(v, IsNegative, 1024)', [6, 0, 3, 1024, 1, 1024, 2, 0, 7, 10, 1024, 5, 3, 2, 2, 11]);
  Test('ReplaceIf(v, odd, 3)', [6, 0, 3, -2, 3, -4, 2, 0, 3, 10, -4, 3, 3, 2, 2, 3]);
  Test('ReplaceIf(v, even, 1)', [1, 1, 3, 1, 1, 1, 1, 1, 7, 1, 1, 5, 3, 1, 1, 11]);
  Test('ReplaceIf(v, IsPrime, 1000)', [6, 0, 1000, -2, 1, -4, 1000, 0, 1000, 10, -4, 1000, 1000, 1000, 1000, 1000]);
  Test('ReplaceIf(v, IsSound, 1000)', [6, 0, 3, -2, 1, -4, 2, 0, 7, 10, -4, 5, 3, 2, 2, 11]);
  Test('ReplaceIf(v, x ↦ x ≥ 5, 5)', [5, 0, 3, -2, 1, -4, 2, 0, 5, 5, -4, 5, 3, 2, 2, 5]);
  Test('ReplaceIf(v, x ↦ x ≥ 5, "cat")', failure, 'Object isn''t a real number.');
  Test('ReplaceIf(v, x ↦ x! ≥ 5, 5)', failure, 'A non-negative integer was expected as argument 1, but "-2" was given.');

  Test('ReplaceIf(❨6, 0, 3, −2, 1, −4, 2, 0, 7, 10, −4, 5, 3, 2, 2, 11❩, x ↦ x ≥ 5, 5)', [5, 0, 3, -2, 1, -4, 2, 0, 5, 5, -4, 5, 3, 2, 2, 5]);
  Test('ReplaceIf(DebugObject("empty vector"), x ↦ x ≥ 5, 5)', TAlgosimRealVector.Create);

  Test('A ≔ ❨❨6, 0, 3, −2, 1, −4, 2, 0❩, ❨7, 10, −4, 5, 3, 2, 2, 11❩❩;', null);

  Test('ReplaceIf(A, IsNumber, 10)', 8, [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]);
  Test('ReplaceIf(A, IsPositive, 64)', 8, [64, 0, 64, -2, 64, -4, 64, 0, 64, 64, -4, 64, 64, 64, 64, 64]);
  Test('ReplaceIf(A, IsZero, 1.2)', 8, [6, 1.2, 3, -2, 1, -4, 2, 1.2, 7, 10, -4, 5, 3, 2, 2, 11]);
  Test('ReplaceIf(A, IsNegative, 1024)', 8, [6, 0, 3, 1024, 1, 1024, 2, 0, 7, 10, 1024, 5, 3, 2, 2, 11]);
  Test('ReplaceIf(A, odd, 3)', 8, [6, 0, 3, -2, 3, -4, 2, 0, 3, 10, -4, 3, 3, 2, 2, 3]);
  Test('ReplaceIf(A, even, 1)', 8, [1, 1, 3, 1, 1, 1, 1, 1, 7, 1, 1, 5, 3, 1, 1, 11]);
  Test('ReplaceIf(A, IsPrime, 1000)', 8, [6, 0, 1000, -2, 1, -4, 1000, 0, 1000, 10, -4, 1000, 1000, 1000, 1000, 1000]);
  Test('ReplaceIf(A, IsSound, 1000)', 8, [6, 0, 3, -2, 1, -4, 2, 0, 7, 10, -4, 5, 3, 2, 2, 11]);
  Test('ReplaceIf(A, x ↦ x ≥ 5, 5)', 8, [5, 0, 3, -2, 1, -4, 2, 0, 5, 5, -4, 5, 3, 2, 2, 5]);
  Test('ReplaceIf(A, x ↦ x ≥ 5, "cat")', failure, 'Object isn''t a real number.');
  Test('ReplaceIf(A, x ↦ x! ≥ 5, 5)', failure, 'A non-negative integer was expected as argument 1, but "-2" was given.');

  Test('ReplaceIf(❨❨6, 0, 3, −2, 1, −4, 2, 0❩, ❨7, 10, −4, 5, 3, 2, 2, 11❩❩, x ↦ x ≥ 5, 5)', 8, [5, 0, 3, -2, 1, -4, 2, 0, 5, 5, -4, 5, 3, 2, 2, 5]);
  Test('ReplaceIf(DebugObject("empty matrix"), x ↦ x ≥ 5, 5)', TAlgosimRealMatrix.Create);

  Test('v ≔ ❨1, 0, i, −i, −1, 2, 1, 0.5❩;', null);

  Test('ReplaceIf(v, IsNumber, 0)', [TASC(0), 0, 0, 0, 0, 0, 0, 0]);
  Test('ReplaceIf(v, IsReal, 1)', [1, 1, ImaginaryUnit, -ImaginaryUnit, 1, 1, 1, 1]);
  Test('ReplaceIf(v, IsInteger, 10)', [10, 10, ImaginaryUnit, -ImaginaryUnit, 10, 10, 10, 0.5]);
  Test('ReplaceIf(v, x ↦ IsInteger(x) ∧ IsPrime(x), 1024)', [1, 0, ImaginaryUnit, -ImaginaryUnit, -1, 1024, 1, 0.5]);
  Test('ReplaceIf(v, x ↦ IsInteger(x) ∧ odd(x), 12)', [12, 0, ImaginaryUnit, -ImaginaryUnit, 12, 2, 12, 0.5]);
  Test('ReplaceIf(v, x ↦ IsInteger(x) ∧ even(x), 99)', [1, 99, ImaginaryUnit, -ImaginaryUnit, -1, 99, 1, 0.5]);
  Test('ReplaceIf(v, IsSound, 12)', [1, 0, ImaginaryUnit, -ImaginaryUnit, -1, 2, 1, 0.5]);
  Test('ReplaceIf(v, even, 99)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');
  Test('ReplaceIf(v, x ↦ IsInteger(x) ∧ IsPrime(x), "nargle")', failure, 'Object isn''t a complex number.');

  Test('ReplaceIf(❨1, 0, i, −i, −1, 2, 1, 0.5❩, x ↦ IsInteger(x) ∧ IsPrime(x), 1024)', [1, 0, ImaginaryUnit, -ImaginaryUnit, -1, 1024, 1, 0.5]);
  Test('ReplaceIf(ComplexVector(DebugObject("empty vector")), x ↦ IsInteger(x) ∧ IsPrime(x), 1024)', TAlgosimComplexVector.Create);

  Test('A ≔ ❨❨1, 0, i, −i❩, ❨−1, 2, 1, 0.5❩❩;', null);

  Test('ReplaceIf(A, IsNumber, 0)', 4, [TASC(0), 0, 0, 0, 0, 0, 0, 0]);
  Test('ReplaceIf(A, IsReal, 1)', 4, [1, 1, ImaginaryUnit, -ImaginaryUnit, 1, 1, 1, 1]);
  Test('ReplaceIf(A, IsInteger, 10)', 4, [10, 10, ImaginaryUnit, -ImaginaryUnit, 10, 10, 10, 0.5]);
  Test('ReplaceIf(A, x ↦ IsInteger(x) ∧ IsPrime(x), 1024)', 4, [1, 0, ImaginaryUnit, -ImaginaryUnit, -1, 1024, 1, 0.5]);
  Test('ReplaceIf(A, x ↦ IsInteger(x) ∧ odd(x), 12)', 4, [12, 0, ImaginaryUnit, -ImaginaryUnit, 12, 2, 12, 0.5]);
  Test('ReplaceIf(A, x ↦ IsInteger(x) ∧ even(x), 99)', 4, [1, 99, ImaginaryUnit, -ImaginaryUnit, -1, 99, 1, 0.5]);
  Test('ReplaceIf(A, IsSound, 12)', 4, [1, 0, ImaginaryUnit, -ImaginaryUnit, -1, 2, 1, 0.5]);
  Test('ReplaceIf(A, even, 99)', failure, 'An object of type integer was expected as argument 1, but an object of type complex number was given.');
  Test('ReplaceIf(A, x ↦ IsInteger(x) ∧ IsPrime(x), "nargle")', failure, 'Object isn''t a complex number.');

  Test('ReplaceIf(❨❨1, 0, i, −i❩, ❨−1, 2, 1, 0.5❩❩, x ↦ IsInteger(x) ∧ IsPrime(x), 1024)', 4, [1, 0, ImaginaryUnit, -ImaginaryUnit, -1, 1024, 1, 0.5]);
  Test('ReplaceIf(ComplexMatrix(DebugObject("empty matrix")), x ↦ IsInteger(x) ∧ IsPrime(x), 1024)', TAlgosimComplexMatrix.Create);

  Test('ReplaceIf("Hello, World!", ChrIsUpperCase, "T")', 'Tello, Torld!');
  Test('ReplaceIf("Hello, World!", ChrIsPunctuation, ".")', 'Hello. World.');
  Test('ReplaceIf("Hello, World!", IsCharacter, "-")', '-------------');
  Test('ReplaceIf("Hello, World!", IsPixmap, ".")', 'Hello, World!');
  Test('ReplaceIf("", IsCharacter, "-")', '');
  Test('ReplaceIf("Hello, World!", ChrIsUpperCase, IdentityMatrix(10))', failure, 'An object of type "real matrix" cannot be converted to a character.');
  Test('ReplaceIf("Hello, World!", ChrIsUpperCase, "ABC")', failure, 'A string of length 3 cannot be converted to a character.');

  Test('L ≔ ''(6, 0, 3, −2, 1, −4, 2, 0, 7, 10, −4, 5, 3, 2, 2, 11);', null);

  Test('ReplaceIf(L, IsNumber, 10)', intarr([10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]));
  Test('ReplaceIf(L, IsPositive, 64)', intarr([64, 0, 64, -2, 64, -4, 64, 0, 64, 64, -4, 64, 64, 64, 64, 64]));
  Test('ReplaceIf(L, IsZero, 1.2)', asoarr([6, 1.2, 3, -2, 1, -4, 2, 1.2, 7, 10, -4, 5, 3, 2, 2, 11]));
  Test('ReplaceIf(L, IsNegative, 1024)', intarr([6, 0, 3, 1024, 1, 1024, 2, 0, 7, 10, 1024, 5, 3, 2, 2, 11]));
  Test('ReplaceIf(L, odd, 3)', intarr([6, 0, 3, -2, 3, -4, 2, 0, 3, 10, -4, 3, 3, 2, 2, 3]));
  Test('ReplaceIf(L, even, 1)', intarr([1, 1, 3, 1, 1, 1, 1, 1, 7, 1, 1, 5, 3, 1, 1, 11]));
  Test('ReplaceIf(L, IsPrime, 1000)', intarr([6, 0, 1000, -2, 1, -4, 1000, 0, 1000, 10, -4, 1000, 1000, 1000, 1000, 1000]));
  Test('ReplaceIf(L, IsSound, 1000)', intarr([6, 0, 3, -2, 1, -4, 2, 0, 7, 10, -4, 5, 3, 2, 2, 11]));
  Test('ReplaceIf(L, x ↦ x ≥ 5, 5)', intarr([5, 0, 3, -2, 1, -4, 2, 0, 5, 5, -4, 5, 3, 2, 2, 5]));
  Test('ReplaceIf(L, x ↦ x ≥ 5, "cat")', asoarr(['cat', 0, 3, -2, 1, -4, 2, 0, 'cat', 'cat', -4, 'cat', 3, 2, 2, 'cat']));
  Test('ReplaceIf(L, x ↦ x! ≥ 5, 5)', failure, 'A non-negative integer was expected as argument 1, but "-2" was given.');

  Test('ReplaceIf(''(6, 0, 3, −2, 1, −4, 2, 0, 7, 10, −4, 5, 3, 2, 2, 11), x ↦ x ≥ 5, 5)', intarr([5, 0, 3, -2, 1, -4, 2, 0, 5, 5, -4, 5, 3, 2, 2, 5]));
  Test('ReplaceIf(''(), x ↦ x ≥ 5, 5)', asoarr([]));

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null);

  Test('ReplaceIf(Jane, IsInteger, 50)',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [Name('Jane', 'Smith'), ASO('female'), ASOInt(50)]
    )
  );

  Test('ReplaceIf(ToSet(v), IsReal, 22)', failure, 'Object type "set" is not a suitable container.');

  // - with level specification

  Test('L ≔ ''(''(1, 2, 3), ''(4, 5, 6));', null);

  Test('ReplaceIf(L, x ↦ IsInteger(x) ∧ odd(x), 1024)',
    asoarrex(
      [
        intarr([1, 2, 3]),
        intarr([4, 5, 6])
      ]
    )
  );

  Test('ReplaceIf(L, x ↦ IsInteger(x) ∧ odd(x), 1024, 2)',
    asoarrex(
      [
        intarr([1024, 2, 1024]),
        intarr([4, 1024, 6])
      ]
    )
  );

  // Unconditional replacement

  Test('v ≔ ❨3, 0, 2, −4, 7, 5, −1, 2, 1, 4, 5, 2, 6, 3, 4, 0❩;', null);
  Test('ReplaceEvery(v, 5)', [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]);
  Test('ReplaceEvery(❨3, 0, 2, −4, 7, 5, −1, 2, 1, 4, 5, 2, 6, 3, 4, 0❩, 5)', [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]);
  Test('ReplaceEvery(DebugObject("empty vector"), 5)', TAlgosimRealVector.Create);
  Test('ReplaceEvery(v, "cat")', failure, 'Object isn''t a real number.');

  Test('A ≔ ❨❨3, 0, 2, −4, 7, 5, −1, 2❩, ❨1, 4, 5, 2, 6, 3, 4, 0❩❩;', null);
  Test('ReplaceEvery(A, 5)', 8, [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]);
  Test('ReplaceEvery(❨❨3, 0, 2, −4, 7, 5, −1, 2❩, ❨1, 4, 5, 2, 6, 3, 4, 0❩❩, 5)', 8, [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]);
  Test('ReplaceEvery(DebugObject("empty matrix"), 5)', TAlgosimRealMatrix.Create);
  Test('ReplaceEvery(A, "cat")', failure, 'Object isn''t a real number.');

  Test('v ≔ ❨i, 0, 2, −4, 7, 5, −1, 2, 1, 4, 5, 2, 6, 3, 4, 0❩;', null);
  Test('ReplaceEvery(v, 5)', [TASC(5), 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]);
  Test('ReplaceEvery(❨i, 0, 2, −4, 7, 5, −1, 2, 1, 4, 5, 2, 6, 3, 4, 0❩, 5)', [TASC(5), 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]);
  Test('ReplaceEvery(ComplexVector(DebugObject("empty vector")), 5)', TAlgosimComplexVector.Create);
  Test('ReplaceEvery(v, "cat")', failure, 'Object isn''t a complex number.');

  Test('A ≔ ❨❨i, 0, 2, −4, 7, 5, −1, 2❩, ❨1, 4, 5, 2, 6, 3, 4, 0❩❩;', null);
  Test('ReplaceEvery(A, 5)', 8, [TASC(5), 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]);
  Test('ReplaceEvery(❨❨i, 0, 2, −4, 7, 5, −1, 2❩, ❨1, 4, 5, 2, 6, 3, 4, 0❩❩, 5)', 8, [TASC(5), 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]);
  Test('ReplaceEvery(ComplexMatrix(DebugObject("empty matrix")), 5)', TAlgosimComplexMatrix.Create);
  Test('ReplaceEvery(A, "cat")', failure, 'Object isn''t a complex number.');

  Test('ReplaceEvery("Hello, World!", "-")', '-------------');
  Test('ReplaceEvery("", "-")', '');
  Test('ReplaceEvery("Hello, World!", "--")', failure, 'A string of length 2 cannot be converted to a character.');
  Test('ReplaceEvery("Hello, World!", 1024)', failure, 'An object of type "integer" cannot be converted to a character.');

  Test('L ≔ ''(3, 0, 2, −4, 7, 5, −1, 2, 1, 4, 5, 2, 6, 3, 4, 0);', null);
  Test('ReplaceEvery(L, 5)', intarr([5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]));
  Test('ReplaceEvery(''(3, 0, 2, −4, 7, 5, −1, 2, 1, 4, 5, 2, 6, 3, 4, 0), 5)', intarr([5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]));
  Test('ReplaceEvery(''(), 5)', intarr([]));
  Test('ReplaceEvery(L, "cat")', strarr(['cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'cat']));

  Test('ReplaceEvery(ToSet(L), "cat")', failure, 'Object type "set" is not a suitable container.');

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null);

  Test('ReplaceEvery(Jane, 123)',
    TAlgosimStructure.CreateWithValue(
      ['name', 'sex', 'age'],
      [ASOInt(123), ASOInt(123), ASOInt(123)]
    )
  );

  // - with level specification

  Test('L ≔ ''(''(1, 2, 3), ''(4, 5, 6));', null);

  Test('ReplaceEvery(L, 5)',
    asoarrex(
      [
        ASOInt(5),
        ASOInt(5)
      ]
    )
  );

  Test('ReplaceEvery(L, 5, 2)',
    asoarrex(
      [
        intarr([5, 5, 5]),
        intarr([5, 5, 5])
      ]
    )
  );


  //
  // Removal
  //

  // Object-based removal

  Test('v ≔ ❨6, 2, 0, 4, 1, −2, 3, 6, 2, 1, 0, 2, 1, 4, 1, 0❩;', null);

  Test('RemoveAll(v, 0)', [6, 2, 4, 1, -2, 3, 6, 2, 1, 2, 1, 4, 1]);
  Test('RemoveAll(v, 1)', [6, 2, 0, 4, -2, 3, 6, 2, 0, 2, 4, 0]);
  Test('RemoveAll(v, 2)', [6, 0, 4, 1, -2, 3, 6, 1, 0, 1, 4, 1, 0]);
  Test('RemoveAll(v, 3)', [6, 2, 0, 4, 1, -2, 6, 2, 1, 0, 2, 1, 4, 1, 0]);
  Test('RemoveAll(v, 4)', [6, 2, 0, 1, -2, 3, 6, 2, 1, 0, 2, 1, 1, 0]);
  Test('RemoveAll(v, 5)', [6, 2, 0, 4, 1, -2, 3, 6, 2, 1, 0, 2, 1, 4, 1, 0]);
  Test('RemoveAll(v, "cat")', [6, 2, 0, 4, 1, -2, 3, 6, 2, 1, 0, 2, 1, 4, 1, 0]);

  Test('RemoveAll(❨6, 2, 0, 4, 1, −2, 3, 6, 2, 1, 0, 2, 1, 4, 1, 0❩, 2)', [6, 0, 4, 1, -2, 3, 6, 1, 0, 1, 4, 1, 0]);
  Test('RemoveAll(DebugObject("empty vector"), 2)', TAlgosimRealVector.Create);

  Test('A ≔ ❨❨6, 2, 0, 4, 1, −2, 3, 6❩, ❨2, 1, 0, 2, 1, 4, 1, 0❩❩;', null);
  Test('RemoveAll(A, 0)', failure, 'Object type "real matrix" is not a suitable container.');

  Test('v ≔ ❨3, 2, i, 1, 2, i, 0, −1, 1, −i, i, 2, 1, 3, 0, 1❩;', null);

  Test('RemoveAll(v, 0)', [3, 2, ImaginaryUnit, 1, 2, ImaginaryUnit, -1, 1, -ImaginaryUnit, ImaginaryUnit, 2, 1, 3, 1]);
  Test('RemoveAll(v, 1)', [3, 2, ImaginaryUnit, 2, ImaginaryUnit, 0, -1, -ImaginaryUnit, ImaginaryUnit, 2, 3, 0]);
  Test('RemoveAll(v, 2)', [3, ImaginaryUnit, 1, ImaginaryUnit, 0, -1, 1, -ImaginaryUnit, ImaginaryUnit, 1, 3, 0, 1]);
  Test('RemoveAll(v, 3)', [2, ImaginaryUnit, 1, 2, ImaginaryUnit, 0, -1, 1, -ImaginaryUnit, ImaginaryUnit, 2, 1, 0, 1]);
  Test('RemoveAll(v, 4)', [3, 2, ImaginaryUnit, 1, 2, ImaginaryUnit, 0, -1, 1, -ImaginaryUnit, ImaginaryUnit, 2, 1, 3, 0, 1]);
  Test('RemoveAll(v, i)', [3, 2, 1, 2, 0, -1, 1, -ImaginaryUnit, 2, 1, 3, 0, 1]);
  Test('RemoveAll(v, "cat")', [3, 2, ImaginaryUnit, 1, 2, ImaginaryUnit, 0, -1, 1, -ImaginaryUnit, ImaginaryUnit, 2, 1, 3, 0, 1]);

  Test('RemoveAll(❨3, 2, i, 1, 2, i, 0, −1, 1, −i, i, 2, 1, 3, 0, 1❩, i)', [3, 2, 1, 2, 0, -1, 1, -ImaginaryUnit, 2, 1, 3, 0, 1]);
  Test('RemoveAll(ComplexVector(DebugObject("empty vector")), i)', TAlgosimComplexVector.Create);

  Test('A ≔ ❨❨3, 2, i, 1, 2, i, 0, −1❩, ❨1, −i, i, 2, 1, 3, 0, 1❩❩;', null);
  Test('RemoveAll(A, 0)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('RemoveAll("Hello, World!", "l")', 'Heo, Word!');
  Test('RemoveAll("Hello, World!", "Hello")', 'Hello, World!');
  Test('RemoveAll("Hello, World!", IdentityMatrix(100))', 'Hello, World!');
  Test('RemoveAll("", "l")', '');

  Test('L ≔ ''(6, 2, 0, 4, 1, −2, 3, 6, 2, 1, 0, 2, 1, 4, 1, 0);', null);

  Test('RemoveAll(L, 0)', intarr([6, 2, 4, 1, -2, 3, 6, 2, 1, 2, 1, 4, 1]));
  Test('RemoveAll(L, 1)', intarr([6, 2, 0, 4, -2, 3, 6, 2, 0, 2, 4, 0]));
  Test('RemoveAll(L, 2)', intarr([6, 0, 4, 1, -2, 3, 6, 1, 0, 1, 4, 1, 0]));
  Test('RemoveAll(L, 3)', intarr([6, 2, 0, 4, 1, -2, 6, 2, 1, 0, 2, 1, 4, 1, 0]));
  Test('RemoveAll(L, 4)', intarr([6, 2, 0, 1, -2, 3, 6, 2, 1, 0, 2, 1, 1, 0]));
  Test('RemoveAll(L, 5)', intarr([6, 2, 0, 4, 1, -2, 3, 6, 2, 1, 0, 2, 1, 4, 1, 0]));
  Test('RemoveAll(L, "cat")', intarr([6, 2, 0, 4, 1, -2, 3, 6, 2, 1, 0, 2, 1, 4, 1, 0]));

  Test('RemoveAll(''(6, 2, 0, 4, 1, −2, 3, 6, 2, 1, 0, 2, 1, 4, 1, 0), 2)', intarr([6, 0, 4, 1, -2, 3, 6, 1, 0, 1, 4, 1, 0]));
  Test('RemoveAll(''(), 2)', intarr([]));

  Test('RemoveAll(ToSet(L), 0)', failure, 'Object type "set" is not a suitable container.'); // use the set minus operator instead

//  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
//  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
//  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null);

//  Test('RemoveAll(Jane, 40)',
//    TAlgosimStructure.CreateWithValue(
//      ['name', 'sex'],
//      [Name('Jane', 'Smith'), ASO('female')]
//    )
//  ); // not implemented // TODO: Implement? If so, implement `remove(struct)` and this will come for free

  // - with level specification

  Test('L ≔ ''(''(1, 2, 3), ''(4, 5, 6));', null);

  Test('RemoveAll(L, 5)',
    asoarrex(
      [
        intarr([1, 2, 3]),
        intarr([4, 5, 6])
      ]
    )
  );

  Test('RemoveAll(L, 5, 2)',
    asoarrex(
      [
        intarr([1, 2, 3]),
        intarr([4, 6])
      ]
    )
  );

  // Predicate-based removal

  Test('v ≔ ❨2, −3, 1, 0, 4, 5, −6, 8, 9, 10, 11, −4, 7, 2, 1, 6❩;', null);

  Test('RemoveIf(v, IsNumber)', TAlgosimRealVector.Create);
  Test('RemoveIf(v, IsPositive)', [-3, 0, -6, -4]);
  Test('RemoveIf(v, IsZero)', [2, -3, 1, 4, 5, -6, 8, 9, 10, 11, -4, 7, 2, 1, 6]);
  Test('RemoveIf(v, IsNegative)', [2, 1, 0, 4, 5, 8, 9, 10, 11, 7, 2, 1, 6]);
  Test('RemoveIf(v, odd)', [2, 0, 4, -6, 8, 10, -4, 2, 6]);
  Test('RemoveIf(v, even)', [-3, 1, 5, 9, 11, 7, 1]);
  Test('RemoveIf(v, IsPrime)', [-3, 1, 0, 4, -6, 8, 9, 10, -4, 1, 6]);
  Test('RemoveIf(v, x ↦ x > 10)', [2, -3, 1, 0, 4, 5, -6, 8, 9, 10, -4, 7, 2, 1, 6]);
  Test('RemoveIf(v, x ↦ x > 100)', [2, -3, 1, 0, 4, 5, -6, 8, 9, 10, 11, -4, 7, 2, 1, 6]);
  Test('RemoveIf(v, x ↦ x! > 100)', failure, 'A non-negative integer was expected as argument 1, but "-3" was given.');
  Test('RemoveIf(v, sin)', failure, 'The predicate was expected to return a boolean, but an object of type real number was returned.');

  Test('RemoveIf(❨2, −3, 1, 0, 4, 5, −6, 8, 9, 10, 11, −4, 7, 2, 1, 6❩, IsNegative)', [2, 1, 0, 4, 5, 8, 9, 10, 11, 7, 2, 1, 6]);
  Test('RemoveIf(DebugObject("empty vector"), IsNegative)', TAlgosimRealVector.Create);

  Test('A ≔ ❨❨2, −3, 1, 0, 4, 5, −6, 8❩, ❨9, 10, 11, −4, 7, 2, 1, 6❩❩;', null);
  Test('RemoveIf(A, IsNegative)', failure, 'Object type "real matrix" is not a suitable container.');

  Test('v ≔ ❨5, 2.1, −4, 3, i, −i, 0, 3, 1, 5❩;', null);

  Test('RemoveIf(v, IsNumber)', TAlgosimComplexVector.Create);
  Test('RemoveIf(v, IsReal)', [ImaginaryUnit, -ImaginaryUnit]);
  Test('RemoveIf(v, IsInteger)', [2.1, ImaginaryUnit, -ImaginaryUnit]);
  Test('RemoveIf(v, IsSound)', [5, 2.1, -4, 3, ImaginaryUnit, -ImaginaryUnit, 0, 3, 1, 5]);
  Test('RemoveIf(v, sin)', failure, 'The predicate was expected to return a boolean, but an object of type complex number was returned.');

  Test('RemoveIf(❨5, 2.1, −4, 3, i, −i, 0, 3, 1, 5❩, IsInteger)', [2.1, ImaginaryUnit, -ImaginaryUnit]);
  Test('RemoveIf(ComplexVector(DebugObject("empty vector")), IsInteger)', TAlgosimComplexVector.Create);

  Test('A ≔ ❨❨5, 2.1, −4, 3, i❩, ❨−i, 0, 3, 1, 5❩❩;', null);
  Test('RemoveIf(A, IsInteger)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('RemoveIf("Hello, World!", ChrIsPunctuation)', 'Hello World');
  Test('RemoveIf("Hello, World!", ChrIsControl)', 'Hello, World!');
  Test('RemoveIf("Hello, World!", ChrIsASCII)', '');
  Test('RemoveIf("", IsCharacter)', '');
  Test('RemoveIf("Hello, World!", length)', failure, 'The predicate was expected to return a boolean, but an object of type integer was returned.');

  Test('L ≔ ''(2, −3, 1, 0, 4, 5, −6, 8, 9, 10, 11, −4, 7, 2, 1, 6);', null);

  Test('RemoveIf(L, IsNumber)', intarr([]));
  Test('RemoveIf(L, IsPositive)', intarr([-3, 0, -6, -4]));
  Test('RemoveIf(L, IsZero)', intarr([2, -3, 1, 4, 5, -6, 8, 9, 10, 11, -4, 7, 2, 1, 6]));
  Test('RemoveIf(L, IsNegative)', intarr([2, 1, 0, 4, 5, 8, 9, 10, 11, 7, 2, 1, 6]));
  Test('RemoveIf(L, odd)', intarr([2, 0, 4, -6, 8, 10, -4, 2, 6]));
  Test('RemoveIf(L, even)', intarr([-3, 1, 5, 9, 11, 7, 1]));
  Test('RemoveIf(L, IsPrime)', intarr([-3, 1, 0, 4, -6, 8, 9, 10, -4, 1, 6]));
  Test('RemoveIf(L, x ↦ x > 10)', intarr([2, -3, 1, 0, 4, 5, -6, 8, 9, 10, -4, 7, 2, 1, 6]));
  Test('RemoveIf(L, x ↦ x > 100)', intarr([2, -3, 1, 0, 4, 5, -6, 8, 9, 10, 11, -4, 7, 2, 1, 6]));
  Test('RemoveIf(L, x ↦ x! > 100)', failure, 'A non-negative integer was expected as argument 1, but "-3" was given.');
  Test('RemoveIf(L, sin)', failure, 'The predicate was expected to return a boolean, but an object of type real number was returned.');

  Test('RemoveIf(''(2, −3, 1, 0, 4, 5, −6, 8, 9, 10, 11, −4, 7, 2, 1, 6), IsNegative)', intarr([2, 1, 0, 4, 5, 8, 9, 10, 11, 7, 2, 1, 6]));
  Test('RemoveIf(''(), IsNegative)', intarr([]));

  Test('S ≔ {2, −3, 1, 0, 4, 5, −6, 8, 9, 10, 11, −4, 7, 2, 1, 6};', null);

  Test('RemoveIf(S, IsNumber)', intset([]));
  Test('RemoveIf(S, IsPositive)', intset([-3, 0, -6, -4]));
  Test('RemoveIf(S, IsZero)', intset([2, -3, 1, 4, 5, -6, 8, 9, 10, 11, -4, 7, 2, 1, 6]));
  Test('RemoveIf(S, IsNegative)', intset([2, 1, 0, 4, 5, 8, 9, 10, 11, 7, 2, 1, 6]));
  Test('RemoveIf(S, odd)', intset([2, 0, 4, -6, 8, 10, -4, 2, 6]));
  Test('RemoveIf(S, even)', intset([-3, 1, 5, 9, 11, 7, 1]));
  Test('RemoveIf(S, IsPrime)', intset([-3, 1, 0, 4, -6, 8, 9, 10, -4, 1, 6]));
  Test('RemoveIf(S, x ↦ x > 10)', intset([2, -3, 1, 0, 4, 5, -6, 8, 9, 10, -4, 7, 2, 1, 6]));
  Test('RemoveIf(S, x ↦ x > 100)', intset([2, -3, 1, 0, 4, 5, -6, 8, 9, 10, 11, -4, 7, 2, 1, 6]));
  Test('RemoveIf(S, x ↦ x! > 100)', failure, 'A non-negative integer was expected as argument 1, but '); // don't know the order
  Test('RemoveIf(S, sin)', failure, 'The predicate was expected to return a boolean, but an object of type real number was returned.');

  Test('RemoveIf({2, −3, 1, 0, 4, 5, −6, 8, 9, 10, 11, −4, 7, 2, 1, 6}, IsNegative)', intset([2, 1, 0, 4, 5, 8, 9, 10, 11, 7, 2, 1, 6]));
  Test('RemoveIf({}, IsNegative)', intset([]));

//  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
//  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
//  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null);

//  Test('RemoveIf(Jane, IsInteger)',
//    TAlgosimStructure.CreateWithValue(
//      ['name', 'sex'],
//      [Name('Jane', 'Smith'), ASO('female')]
//    )
//  ); // not implemented // TODO: Implement? If so, implement `remove(struct)` and this will come for free

  // - with level specification

  Test('L ≔ ''(''(1, 2, 3), ''(4, 5, 6));', null);

  Test('RemoveIf(L, x ↦ IsInteger(x) ∧ odd(x))',
    asoarrex(
      [
        intarr([1, 2, 3]),
        intarr([4, 5, 6])
      ]
    )
  );

  Test('RemoveIf(L, x ↦ IsInteger(x) ∧ odd(x), 2)',
    asoarrex(
      [
        intarr([2]),
        intarr([4, 6])
      ]
    )
  );


  //
  // Grouping and flattening
  //

  Test('L ≔ ''(1, 2, 3, 4, 5, 6, 7, 8, 9);', null);

  Test('group(L, 3)',
    asoarrex(
      [
        intarr([1, 2, 3]),
        intarr([4, 5, 6]),
        intarr([7, 8, 9])
      ]
    )
  );

  Test('L ≔ ''(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12);', null);

  Test('group(L, 1)',
    asoarrex(
      [
        intarr([1]),
        intarr([2]),
        intarr([3]),
        intarr([4]),
        intarr([5]),
        intarr([6]),
        intarr([7]),
        intarr([8]),
        intarr([9]),
        intarr([10]),
        intarr([11]),
        intarr([12])
      ]
    )
  );

  Test('group(L, 2)',
    asoarrex(
      [
        intarr([1, 2]),
        intarr([3, 4]),
        intarr([5, 6]),
        intarr([7, 8]),
        intarr([9, 10]),
        intarr([11, 12])
      ]
    )
  );

  Test('group(L, 3)',
    asoarrex(
      [
        intarr([1, 2, 3]),
        intarr([4, 5, 6]),
        intarr([7, 8, 9]),
        intarr([10, 11, 12])
      ]
    )
  );

  Test('group(L, 4)',
    asoarrex(
      [
        intarr([1, 2, 3, 4]),
        intarr([5, 6, 7, 8]),
        intarr([9, 10, 11, 12])
      ]
    )
  );

  Test('group(L, 6)',
    asoarrex(
      [
        intarr([1, 2, 3, 4, 5, 6]),
        intarr([7, 8, 9, 10, 11, 12])
      ]
    )
  );

  Test('group(L, 12)',
    asoarrex(
      [
        intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
      ]
    )
  );

  Test('group(L, 5)',
    asoarrex(
      [
        intarr([1, 2, 3, 4, 5]),
        intarr([6, 7, 8, 9, 10]),
        intarrN([11, 12, -1, -1, -1])
      ]
    )
  );

  Test('group(L, 7)',
    asoarrex(
      [
        intarr([1, 2, 3, 4, 5, 6, 7]),
        intarrN([8, 9, 10, 11, 12, -1, -1])
      ]
    )
  );

  Test('group(L, 11)',
    asoarrex(
      [
        intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]),
        intarrN([12, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])
      ]
    )
  );

  Test('group(L, 13)',
    asoarrex(
      [
        intarrN([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, -1])
      ]
    )
  );

  Test('group(L, 24)',
    asoarrex(
      [
        intarrN([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])
      ]
    )
  );

  Test('group(L, 0)', failure, 'A positive integer was expected as argument 2, but "0" was given.');
  Test('group(L, -1)', failure, 'A positive integer was expected as argument 2, but "-1" was given.');

  Test('L ≔ ''("alfa", 2, "beta", 3, "gamma", 4);', null);

  Test('group(L, 2)',
    asoarrex(
      [
        asoarr(['alfa', 2]),
        asoarr(['beta', 3]),
        asoarr(['gamma', 4])
      ]
    )
  );

  Test('flatten(''(''("alfa", 1), ''("beta", 2), ''("gamma", 3)))',
    asoarr(['alfa', 1, 'beta', 2, 'gamma', 3])
  );

  Test('flatten(''(''(1, 2, 3), ''(4, 5, 6), ''(7, 8, 9), ''(10, 11, 12)))',
    asoarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
  );

  Test('flatten(''(''("alfa", 1), ''("beta", 2), ''("gamma", 3, 4)))',
    asoarr(['alfa', 1, 'beta', 2, 'gamma', 3, 4])
  );

  Test('flatten(''(''(1, 2, 3, 3.5), ''(4, 5, 6), ''(7, 8), ''(10, 11, 12)))',
    asoarr([1, 2, 3, 3.5, 4, 5, 6, 7, 8, 10, 11, 12])
  );

  Test('flatten(''(1, ''(2, 3, ''(4, 5, ''(6, ''(7, ''(''(8))))))))',
    asoarr([1, 2, 3, 4, 5, 6, 7, 8])
  );

  Test('flatten(''(''(1, 2), 3, ''(''(4)), 5, ''(''(6), 7)))',
    asoarr([1, 2, 3, 4, 5, 6, 7])
  );


  //
  // Labelling
  //

  Test('label(''(1, 2, 3), ''("a", "b", "c"))',
    TAlgosimStructure.CreateWithValue(
      [
        sm('a', ASOInt(1)),
        sm('b', ASOInt(2)),
        sm('c', ASOInt(3))
      ]
    )
  );

  Test('label(''(1, 2, 3), ''("a", "b", true))',
    failure,
    'The array must contain only objects of type string, but it contains an object of type boolean.'
  );

  Test('label(''(1, 2, 3), ''("a", "b"))',
    failure,
    'The number of member names must equal the number of member values.'
  );

  Test('label(''(1, 2, 3), ''("a", "b", "c", "d"))',
    failure,
    'The number of member names must equal the number of member values.'
  );

  Test('label(''("Albus Dumbledore", 115, "headmaster"), ''("name", "age", "role"))',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(115)),
        sm('role', ASO('headmaster'))
      ]
    )
  );

  Test('label(''("Albus Dumbledore"), ''("name"))',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore'))
      ]
    )
  );

  Test('label(''(), ''())',
    TAlgosimStructure.CreateWithValue(
      [
      ]
    )
  );

  Test('label(''(1, 2, 3), ''("a", "b", ""))',
    failure,
    'Invalid member name: ""'
  );

  Test('label(''(1, 2, 3), ''("a", "b", "+"))',
    failure,
    'Invalid member name: "+"'
  );

  Test('label(''(1, 2, 3), ''("a", "b", "+c"))',
    failure,
    'Invalid member name: "+c"'
  );

  Test('label(''(1, 2, 3), ''("a", "b", " c"))',
    failure,
    'Invalid member name: " c"'
  );

  Test('label(''(1, 2, 3), ''("a", "b", "1"))',
    failure,
    'Invalid member name: "1"'
  );

  Test('label(''(1, 2, 3), ''("a", "b", "1c"))',
    failure,
    'Invalid member name: "1c"'
  );

  Test('label(''(1, 2, 3), ''("a", "b", "c1"))',
    TAlgosimStructure.CreateWithValue(
      [
        sm('a', ASOInt(1)),
        sm('b', ASOInt(2)),
        sm('c1', ASOInt(3))
      ]
    )
  );

  Test('delete(u); delete(v); delete(A); delete(B); delete(M); delete(s); delete(t); delete(L); delete(S); delete(name); delete(person); delete(Jane); delete(Mike); delete(Mary); delete(x); delete(N); delete(r)', success);


  //
  // Rotation
  //

  // Left

  Test('RotLeft(❨1, 2, 3, 4, 5❩)', [2, 3, 4, 5, 1]);
  Test('RotLeft(❨1, 2, 3❩)', [2, 3, 1]);
  Test('RotLeft(❨1, 2❩)', [2, 1]);
  Test('RotLeft(❨5❩)', [5]);
  Test('RotLeft(DebugObject("empty vector"))', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, 1)', [2, 3, 4, 5, 1]);
  Test('RotLeft(❨1, 2, 3❩, 1)', [2, 3, 1]);
  Test('RotLeft(❨1, 2❩, 1)', [2, 1]);
  Test('RotLeft(❨5❩, 1)', [5]);
  Test('RotLeft(DebugObject("empty vector"), 1)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, 2)', [3, 4, 5, 1, 2]);
  Test('RotLeft(❨1, 2, 3❩, 2)', [3, 1, 2]);
  Test('RotLeft(❨1, 2❩, 2)', [1, 2]);
  Test('RotLeft(❨5❩, 2)', [5]);
  Test('RotLeft(DebugObject("empty vector"), 2)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, 3)', [4, 5, 1, 2, 3]);
  Test('RotLeft(❨1, 2, 3❩, 3)', [1, 2, 3]);
  Test('RotLeft(❨1, 2❩, 3)', [2, 1]);
  Test('RotLeft(❨5❩, 3)', [5]);
  Test('RotLeft(DebugObject("empty vector"), 3)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, 4)', [5, 1, 2, 3, 4]);
  Test('RotLeft(❨1, 2, 3❩, 4)', [2, 3, 1]);
  Test('RotLeft(❨1, 2❩, 4)', [1, 2]);
  Test('RotLeft(❨5❩, 4)', [5]);
  Test('RotLeft(DebugObject("empty vector"), 4)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, 5)', [1, 2, 3, 4, 5]);
  Test('RotLeft(❨1, 2, 3❩, 5)', [3, 1, 2]);
  Test('RotLeft(❨1, 2❩, 5)', [2, 1]);
  Test('RotLeft(❨5❩, 5)', [5]);
  Test('RotLeft(DebugObject("empty vector"), 5)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, 10)', [1, 2, 3, 4, 5]);
  Test('RotLeft(❨1, 2, 3, 4, 5❩, 11)', [2, 3, 4, 5, 1]);
  Test('RotLeft(❨1, 2, 3, 4, 5❩, 12)', [3, 4, 5, 1, 2]);
  Test('RotLeft(❨1, 2, 3, 4, 5❩, 100)', [1, 2, 3, 4, 5]);
  Test('RotLeft(❨1, 2, 3, 4, 5❩, 101)', [2, 3, 4, 5, 1]);
  Test('RotLeft(❨1, 2, 3, 4, 5❩, 102)', [3, 4, 5, 1, 2]);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, 0)', [1, 2, 3, 4, 5]);
  Test('RotLeft(❨1, 2, 3❩, 0)', [1, 2, 3]);
  Test('RotLeft(❨1, 2❩, 0)', [1, 2]);
  Test('RotLeft(❨5❩, 0)', [5]);
  Test('RotLeft(DebugObject("empty vector"), 0)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, -1)', [5, 1, 2, 3, 4]);
  Test('RotLeft(❨1, 2, 3❩, -1)', [3, 1, 2]);
  Test('RotLeft(❨1, 2❩, -1)', [2, 1]);
  Test('RotLeft(❨5❩, -1)', [5]);
  Test('RotLeft(DebugObject("empty vector"), -1)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, -2)', [4, 5, 1, 2, 3]);
  Test('RotLeft(❨1, 2, 3❩, -2)', [2, 3, 1]);
  Test('RotLeft(❨1, 2❩, -2)', [1, 2]);
  Test('RotLeft(❨5❩, -2)', [5]);
  Test('RotLeft(DebugObject("empty vector"), -2)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, -3)', [3, 4, 5, 1, 2]);
  Test('RotLeft(❨1, 2, 3❩, -3)', [1, 2, 3]);
  Test('RotLeft(❨1, 2❩, -3)', [2, 1]);
  Test('RotLeft(❨5❩, -3)', [5]);
  Test('RotLeft(DebugObject("empty vector"), -3)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, -4)', [2, 3, 4, 5, 1]);
  Test('RotLeft(❨1, 2, 3❩, -4)', [3, 1, 2]);
  Test('RotLeft(❨1, 2❩, -4)', [1, 2]);
  Test('RotLeft(❨5❩, -4)', [5]);
  Test('RotLeft(DebugObject("empty vector"), -4)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, -5)', [1, 2, 3, 4, 5]);
  Test('RotLeft(❨1, 2, 3❩, -5)', [2, 3, 1]);
  Test('RotLeft(❨1, 2❩, -5)', [2, 1]);
  Test('RotLeft(❨5❩, -5)', [5]);
  Test('RotLeft(DebugObject("empty vector"), -5)', TAlgosimRealVector.Create);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, -10)', [1, 2, 3, 4, 5]);
  Test('RotLeft(❨1, 2, 3, 4, 5❩, -11)', [5, 1, 2, 3, 4]);
  Test('RotLeft(❨1, 2, 3, 4, 5❩, -12)', [4, 5, 1, 2, 3]);

  Test('RotLeft(❨1, 2, 3, 4, 5❩, -100)', [1, 2, 3, 4, 5]);
  Test('RotLeft(❨1, 2, 3, 4, 5❩, -101)', [5, 1, 2, 3, 4]);
  Test('RotLeft(❨1, 2, 3, 4, 5❩, -102)', [4, 5, 1, 2, 3]);

  Test('RotLeft(❨i, 2, 3, 4, 5❩)', [2, 3, 4, 5, ImaginaryUnit]);
  Test('RotLeft(❨i, 2, 3❩)', [2, 3, ImaginaryUnit]);
  Test('RotLeft(❨i, 2❩)', [2, ImaginaryUnit]);
  Test('RotLeft(❨i❩)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")))', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, 1)', [2, 3, 4, 5, ImaginaryUnit]);
  Test('RotLeft(❨i, 2, 3❩, 1)', [2, 3, ImaginaryUnit]);
  Test('RotLeft(❨i, 2❩, 1)', [2, ImaginaryUnit]);
  Test('RotLeft(❨i❩, 1)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), 1)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, 2)', [3, 4, 5, ImaginaryUnit, 2]);
  Test('RotLeft(❨i, 2, 3❩, 2)', [3, ImaginaryUnit, 2]);
  Test('RotLeft(❨i, 2❩, 2)', [ImaginaryUnit, 2]);
  Test('RotLeft(❨i❩, 2)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), 2)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, 3)', [4, 5, ImaginaryUnit, 2, 3]);
  Test('RotLeft(❨i, 2, 3❩, 3)', [ImaginaryUnit, 2, 3]);
  Test('RotLeft(❨i, 2❩, 3)', [2, ImaginaryUnit]);
  Test('RotLeft(❨i❩, 3)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), 3)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, 4)', [5, ImaginaryUnit, 2, 3, 4]);
  Test('RotLeft(❨i, 2, 3❩, 4)', [2, 3, ImaginaryUnit]);
  Test('RotLeft(❨i, 2❩, 4)', [ImaginaryUnit, 2]);
  Test('RotLeft(❨i❩, 4)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), 4)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, 5)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotLeft(❨i, 2, 3❩, 5)', [3, ImaginaryUnit, 2]);
  Test('RotLeft(❨i, 2❩, 5)', [2, ImaginaryUnit]);
  Test('RotLeft(❨i❩, 5)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), 5)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, 10)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotLeft(❨i, 2, 3, 4, 5❩, 11)', [2, 3, 4, 5, ImaginaryUnit]);
  Test('RotLeft(❨i, 2, 3, 4, 5❩, 12)', [3, 4, 5, ImaginaryUnit, 2]);
  Test('RotLeft(❨i, 2, 3, 4, 5❩, 100)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotLeft(❨i, 2, 3, 4, 5❩, 101)', [2, 3, 4, 5, ImaginaryUnit]);
  Test('RotLeft(❨i, 2, 3, 4, 5❩, 102)', [3, 4, 5, ImaginaryUnit, 2]);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, 0)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotLeft(❨i, 2, 3❩, 0)', [ImaginaryUnit, 2, 3]);
  Test('RotLeft(❨i, 2❩, 0)', [ImaginaryUnit, 2]);
  Test('RotLeft(❨i❩, 0)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), 0)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, -1)', [5, ImaginaryUnit, 2, 3, 4]);
  Test('RotLeft(❨i, 2, 3❩, -1)', [3, ImaginaryUnit, 2]);
  Test('RotLeft(❨i, 2❩, -1)', [2, ImaginaryUnit]);
  Test('RotLeft(❨i❩, -1)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), -1)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, -2)', [4, 5, ImaginaryUnit, 2, 3]);
  Test('RotLeft(❨i, 2, 3❩, -2)', [2, 3, ImaginaryUnit]);
  Test('RotLeft(❨i, 2❩, -2)', [ImaginaryUnit, 2]);
  Test('RotLeft(❨i❩, -2)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), -2)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, -3)', [3, 4, 5, ImaginaryUnit, 2]);
  Test('RotLeft(❨i, 2, 3❩, -3)', [ImaginaryUnit, 2, 3]);
  Test('RotLeft(❨i, 2❩, -3)', [2, ImaginaryUnit]);
  Test('RotLeft(❨i❩, -3)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), -3)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, -4)', [2, 3, 4, 5, ImaginaryUnit]);
  Test('RotLeft(❨i, 2, 3❩, -4)', [3, ImaginaryUnit, 2]);
  Test('RotLeft(❨i, 2❩, -4)', [ImaginaryUnit, 2]);
  Test('RotLeft(❨i❩, -4)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), -4)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, -5)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotLeft(❨i, 2, 3❩, -5)', [2, 3, ImaginaryUnit]);
  Test('RotLeft(❨i, 2❩, -5)', [2, ImaginaryUnit]);
  Test('RotLeft(❨i❩, -5)', [ImaginaryUnit]);
  Test('RotLeft(ComplexVector(DebugObject("empty vector")), -5)', TAlgosimComplexVector.Create);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, -10)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotLeft(❨i, 2, 3, 4, 5❩, -11)', [5, ImaginaryUnit, 2, 3, 4]);
  Test('RotLeft(❨i, 2, 3, 4, 5❩, -12)', [4, 5, ImaginaryUnit, 2, 3]);

  Test('RotLeft(❨i, 2, 3, 4, 5❩, -100)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotLeft(❨i, 2, 3, 4, 5❩, -101)', [5, ImaginaryUnit, 2, 3, 4]);
  Test('RotLeft(❨i, 2, 3, 4, 5❩, -102)', [4, 5, ImaginaryUnit, 2, 3]);

  Test('RotLeft("nargle")', 'arglen');
  Test('RotLeft("nargle", 1)', 'arglen');
  Test('RotLeft("nargle", 2)', 'rglena');
  Test('RotLeft("nargle", 3)', 'glenar');
  Test('RotLeft("nargle", 4)', 'lenarg');
  Test('RotLeft("nargle", 5)', 'enargl');
  Test('RotLeft("nargle", 6)', 'nargle');
  Test('RotLeft("nargle", 7)', 'arglen');
  Test('RotLeft("nargle", 8)', 'rglena');
  Test('RotLeft("nargle", 9)', 'glenar');
  Test('RotLeft("nargle", 10)', 'lenarg');
  Test('RotLeft("nargle", 11)', 'enargl');
  Test('RotLeft("nargle", 12)', 'nargle');

  Test('RotLeft("nargle", 60)', 'nargle');
  Test('RotLeft("nargle", 61)', 'arglen');
  Test('RotLeft("nargle", 62)', 'rglena');
  Test('RotLeft("nargle", 63)', 'glenar');
  Test('RotLeft("nargle", 64)', 'lenarg');
  Test('RotLeft("nargle", 65)', 'enargl');
  Test('RotLeft("nargle", 66)', 'nargle');

  Test('RotLeft("nargle", 0)', 'nargle');
  Test('RotLeft("nargle", -1)', 'enargl');
  Test('RotLeft("nargle", -2)', 'lenarg');
  Test('RotLeft("nargle", -3)', 'glenar');
  Test('RotLeft("nargle", -4)', 'rglena');
  Test('RotLeft("nargle", -5)', 'arglen');
  Test('RotLeft("nargle", -6)', 'nargle');
  Test('RotLeft("nargle", -7)', 'enargl');
  Test('RotLeft("nargle", -8)', 'lenarg');
  Test('RotLeft("nargle", -9)', 'glenar');
  Test('RotLeft("nargle", -10)', 'rglena');
  Test('RotLeft("nargle", -11)', 'arglen');
  Test('RotLeft("nargle", -12)', 'nargle');

  Test('RotLeft("nargle", -60)', 'nargle');
  Test('RotLeft("nargle", -61)', 'enargl');
  Test('RotLeft("nargle", -62)', 'lenarg');
  Test('RotLeft("nargle", -63)', 'glenar');
  Test('RotLeft("nargle", -64)', 'rglena');
  Test('RotLeft("nargle", -65)', 'arglen');
  Test('RotLeft("nargle", -66)', 'nargle');

  Test('RotLeft(''(1, 2, 3, 4, 5))', intarr([2, 3, 4, 5, 1]));
  Test('RotLeft(''(1, 2, 3))', intarr([2, 3, 1]));
  Test('RotLeft(''(1, 2))', intarr([2, 1]));
  Test('RotLeft(''(5))', intarr([5]));
  Test('RotLeft(''())', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), 1)', intarr([2, 3, 4, 5, 1]));
  Test('RotLeft(''(1, 2, 3), 1)', intarr([2, 3, 1]));
  Test('RotLeft(''(1, 2), 1)', intarr([2, 1]));
  Test('RotLeft(''(5), 1)', intarr([5]));
  Test('RotLeft(''(), 1)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), 2)', intarr([3, 4, 5, 1, 2]));
  Test('RotLeft(''(1, 2, 3), 2)', intarr([3, 1, 2]));
  Test('RotLeft(''(1, 2), 2)', intarr([1, 2]));
  Test('RotLeft(''(5), 2)', intarr([5]));
  Test('RotLeft(''(), 2)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), 3)', intarr([4, 5, 1, 2, 3]));
  Test('RotLeft(''(1, 2, 3), 3)', intarr([1, 2, 3]));
  Test('RotLeft(''(1, 2), 3)', intarr([2, 1]));
  Test('RotLeft(''(5), 3)', intarr([5]));
  Test('RotLeft(''(), 3)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), 4)', intarr([5, 1, 2, 3, 4]));
  Test('RotLeft(''(1, 2, 3), 4)', intarr([2, 3, 1]));
  Test('RotLeft(''(1, 2), 4)', intarr([1, 2]));
  Test('RotLeft(''(5), 4)', intarr([5]));
  Test('RotLeft(''(), 4)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), 5)', intarr([1, 2, 3, 4, 5]));
  Test('RotLeft(''(1, 2, 3), 5)', intarr([3, 1, 2]));
  Test('RotLeft(''(1, 2), 5)', intarr([2, 1]));
  Test('RotLeft(''(5), 5)', intarr([5]));
  Test('RotLeft(''(), 5)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), 10)', intarr([1, 2, 3, 4, 5]));
  Test('RotLeft(''(1, 2, 3, 4, 5), 11)', intarr([2, 3, 4, 5, 1]));
  Test('RotLeft(''(1, 2, 3, 4, 5), 12)', intarr([3, 4, 5, 1, 2]));
  Test('RotLeft(''(1, 2, 3, 4, 5), 100)', intarr([1, 2, 3, 4, 5]));
  Test('RotLeft(''(1, 2, 3, 4, 5), 101)', intarr([2, 3, 4, 5, 1]));
  Test('RotLeft(''(1, 2, 3, 4, 5), 102)', intarr([3, 4, 5, 1, 2]));

  Test('RotLeft(''(1, 2, 3, 4, 5), 0)', intarr([1, 2, 3, 4, 5]));
  Test('RotLeft(''(1, 2, 3), 0)', intarr([1, 2, 3]));
  Test('RotLeft(''(1, 2), 0)', intarr([1, 2]));
  Test('RotLeft(''(5), 0)', intarr([5]));
  Test('RotLeft(''(), 0)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), -1)', intarr([5, 1, 2, 3, 4]));
  Test('RotLeft(''(1, 2, 3), -1)', intarr([3, 1, 2]));
  Test('RotLeft(''(1, 2), -1)', intarr([2, 1]));
  Test('RotLeft(''(5), -1)', intarr([5]));
  Test('RotLeft(''(), -1)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), -2)', intarr([4, 5, 1, 2, 3]));
  Test('RotLeft(''(1, 2, 3), -2)', intarr([2, 3, 1]));
  Test('RotLeft(''(1, 2), -2)', intarr([1, 2]));
  Test('RotLeft(''(5), -2)', intarr([5]));
  Test('RotLeft(''(), -2)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), -3)', intarr([3, 4, 5, 1, 2]));
  Test('RotLeft(''(1, 2, 3), -3)', intarr([1, 2, 3]));
  Test('RotLeft(''(1, 2), -3)', intarr([2, 1]));
  Test('RotLeft(''(5), -3)', intarr([5]));
  Test('RotLeft(''(), -3)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), -4)', intarr([2, 3, 4, 5, 1]));
  Test('RotLeft(''(1, 2, 3), -4)', intarr([3, 1, 2]));
  Test('RotLeft(''(1, 2), -4)', intarr([1, 2]));
  Test('RotLeft(''(5), -4)', intarr([5]));
  Test('RotLeft(''(), -4)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), -5)', intarr([1, 2, 3, 4, 5]));
  Test('RotLeft(''(1, 2, 3), -5)', intarr([2, 3, 1]));
  Test('RotLeft(''(1, 2), -5)', intarr([2, 1]));
  Test('RotLeft(''(5), -5)', intarr([5]));
  Test('RotLeft(''(), -5)', intarr([]));

  Test('RotLeft(''(1, 2, 3, 4, 5), -10)', intarr([1, 2, 3, 4, 5]));
  Test('RotLeft(''(1, 2, 3, 4, 5), -11)', intarr([5, 1, 2, 3, 4]));
  Test('RotLeft(''(1, 2, 3, 4, 5), -12)', intarr([4, 5, 1, 2, 3]));

  Test('RotLeft(''(1, 2, 3, 4, 5), -100)', intarr([1, 2, 3, 4, 5]));
  Test('RotLeft(''(1, 2, 3, 4, 5), -101)', intarr([5, 1, 2, 3, 4]));
  Test('RotLeft(''(1, 2, 3, 4, 5), -102)', intarr([4, 5, 1, 2, 3]));

  // Right

  Test('RotRight(❨1, 2, 3, 4, 5❩)', [5, 1, 2, 3, 4]);
  Test('RotRight(❨1, 2, 3❩)', [3, 1, 2]);
  Test('RotRight(❨1, 2❩)', [2, 1]);
  Test('RotRight(❨5❩)', [5]);
  Test('RotRight(DebugObject("empty vector"))', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, -01)', [2, 3, 4, 5, 1]);
  Test('RotRight(❨1, 2, 3❩, -01)', [2, 3, 1]);
  Test('RotRight(❨1, 2❩, -01)', [2, 1]);
  Test('RotRight(❨5❩, -01)', [5]);
  Test('RotRight(DebugObject("empty vector"), -01)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, -02)', [3, 4, 5, 1, 2]);
  Test('RotRight(❨1, 2, 3❩, -02)', [3, 1, 2]);
  Test('RotRight(❨1, 2❩, -02)', [1, 2]);
  Test('RotRight(❨5❩, -02)', [5]);
  Test('RotRight(DebugObject("empty vector"), -02)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, -03)', [4, 5, 1, 2, 3]);
  Test('RotRight(❨1, 2, 3❩, -03)', [1, 2, 3]);
  Test('RotRight(❨1, 2❩, -03)', [2, 1]);
  Test('RotRight(❨5❩, -03)', [5]);
  Test('RotRight(DebugObject("empty vector"), -03)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, -04)', [5, 1, 2, 3, 4]);
  Test('RotRight(❨1, 2, 3❩, -04)', [2, 3, 1]);
  Test('RotRight(❨1, 2❩, -04)', [1, 2]);
  Test('RotRight(❨5❩, -04)', [5]);
  Test('RotRight(DebugObject("empty vector"), -04)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, -05)', [1, 2, 3, 4, 5]);
  Test('RotRight(❨1, 2, 3❩, -05)', [3, 1, 2]);
  Test('RotRight(❨1, 2❩, -05)', [2, 1]);
  Test('RotRight(❨5❩, -05)', [5]);
  Test('RotRight(DebugObject("empty vector"), -05)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, -010)', [1, 2, 3, 4, 5]);
  Test('RotRight(❨1, 2, 3, 4, 5❩, -011)', [2, 3, 4, 5, 1]);
  Test('RotRight(❨1, 2, 3, 4, 5❩, -012)', [3, 4, 5, 1, 2]);
  Test('RotRight(❨1, 2, 3, 4, 5❩, -0100)', [1, 2, 3, 4, 5]);
  Test('RotRight(❨1, 2, 3, 4, 5❩, -0101)', [2, 3, 4, 5, 1]);
  Test('RotRight(❨1, 2, 3, 4, 5❩, -0102)', [3, 4, 5, 1, 2]);

  Test('RotRight(❨1, 2, 3, 4, 5❩, 0)', [1, 2, 3, 4, 5]);
  Test('RotRight(❨1, 2, 3❩, 0)', [1, 2, 3]);
  Test('RotRight(❨1, 2❩, 0)', [1, 2]);
  Test('RotRight(❨5❩, 0)', [5]);
  Test('RotRight(DebugObject("empty vector"), 0)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, 01)', [5, 1, 2, 3, 4]);
  Test('RotRight(❨1, 2, 3❩, 01)', [3, 1, 2]);
  Test('RotRight(❨1, 2❩, 01)', [2, 1]);
  Test('RotRight(❨5❩, 01)', [5]);
  Test('RotRight(DebugObject("empty vector"), 01)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, 02)', [4, 5, 1, 2, 3]);
  Test('RotRight(❨1, 2, 3❩, 02)', [2, 3, 1]);
  Test('RotRight(❨1, 2❩, 02)', [1, 2]);
  Test('RotRight(❨5❩, 02)', [5]);
  Test('RotRight(DebugObject("empty vector"), 02)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, 03)', [3, 4, 5, 1, 2]);
  Test('RotRight(❨1, 2, 3❩, 03)', [1, 2, 3]);
  Test('RotRight(❨1, 2❩, 03)', [2, 1]);
  Test('RotRight(❨5❩, 03)', [5]);
  Test('RotRight(DebugObject("empty vector"), 03)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, 04)', [2, 3, 4, 5, 1]);
  Test('RotRight(❨1, 2, 3❩, 04)', [3, 1, 2]);
  Test('RotRight(❨1, 2❩, 04)', [1, 2]);
  Test('RotRight(❨5❩, 04)', [5]);
  Test('RotRight(DebugObject("empty vector"), 04)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, 05)', [1, 2, 3, 4, 5]);
  Test('RotRight(❨1, 2, 3❩, 05)', [2, 3, 1]);
  Test('RotRight(❨1, 2❩, 05)', [2, 1]);
  Test('RotRight(❨5❩, 05)', [5]);
  Test('RotRight(DebugObject("empty vector"), 05)', TAlgosimRealVector.Create);

  Test('RotRight(❨1, 2, 3, 4, 5❩, 010)', [1, 2, 3, 4, 5]);
  Test('RotRight(❨1, 2, 3, 4, 5❩, 011)', [5, 1, 2, 3, 4]);
  Test('RotRight(❨1, 2, 3, 4, 5❩, 012)', [4, 5, 1, 2, 3]);

  Test('RotRight(❨1, 2, 3, 4, 5❩, 0100)', [1, 2, 3, 4, 5]);
  Test('RotRight(❨1, 2, 3, 4, 5❩, 0101)', [5, 1, 2, 3, 4]);
  Test('RotRight(❨1, 2, 3, 4, 5❩, 0102)', [4, 5, 1, 2, 3]);

  Test('RotRight(❨i, 2, 3, 4, 5❩)', [5, ImaginaryUnit, 2, 3, 4]);
  Test('RotRight(❨i, 2, 3❩)', [3, ImaginaryUnit, 2]);
  Test('RotRight(❨i, 2❩)', [2, ImaginaryUnit]);
  Test('RotRight(❨i❩)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")))', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, -01)', [2, 3, 4, 5, ImaginaryUnit]);
  Test('RotRight(❨i, 2, 3❩, -01)', [2, 3, ImaginaryUnit]);
  Test('RotRight(❨i, 2❩, -01)', [2, ImaginaryUnit]);
  Test('RotRight(❨i❩, -01)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), 1)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, -02)', [3, 4, 5, ImaginaryUnit, 2]);
  Test('RotRight(❨i, 2, 3❩, -02)', [3, ImaginaryUnit, 2]);
  Test('RotRight(❨i, 2❩, -02)', [ImaginaryUnit, 2]);
  Test('RotRight(❨i❩, -02)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), 2)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, -03)', [4, 5, ImaginaryUnit, 2, 3]);
  Test('RotRight(❨i, 2, 3❩, -03)', [ImaginaryUnit, 2, 3]);
  Test('RotRight(❨i, 2❩, -03)', [2, ImaginaryUnit]);
  Test('RotRight(❨i❩, -03)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), 3)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, -04)', [5, ImaginaryUnit, 2, 3, 4]);
  Test('RotRight(❨i, 2, 3❩, -04)', [2, 3, ImaginaryUnit]);
  Test('RotRight(❨i, 2❩, -04)', [ImaginaryUnit, 2]);
  Test('RotRight(❨i❩, -04)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), 4)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, -05)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotRight(❨i, 2, 3❩, -05)', [3, ImaginaryUnit, 2]);
  Test('RotRight(❨i, 2❩, -05)', [2, ImaginaryUnit]);
  Test('RotRight(❨i❩, -05)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), 5)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, -010)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotRight(❨i, 2, 3, 4, 5❩, -011)', [2, 3, 4, 5, ImaginaryUnit]);
  Test('RotRight(❨i, 2, 3, 4, 5❩, -012)', [3, 4, 5, ImaginaryUnit, 2]);
  Test('RotRight(❨i, 2, 3, 4, 5❩, -0100)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotRight(❨i, 2, 3, 4, 5❩, -0101)', [2, 3, 4, 5, ImaginaryUnit]);
  Test('RotRight(❨i, 2, 3, 4, 5❩, -0102)', [3, 4, 5, ImaginaryUnit, 2]);

  Test('RotRight(❨i, 2, 3, 4, 5❩, 0)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotRight(❨i, 2, 3❩, 0)', [ImaginaryUnit, 2, 3]);
  Test('RotRight(❨i, 2❩, 0)', [ImaginaryUnit, 2]);
  Test('RotRight(❨i❩, 0)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), 0)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, 01)', [5, ImaginaryUnit, 2, 3, 4]);
  Test('RotRight(❨i, 2, 3❩, 01)', [3, ImaginaryUnit, 2]);
  Test('RotRight(❨i, 2❩, 01)', [2, ImaginaryUnit]);
  Test('RotRight(❨i❩, 01)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), -1)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, 02)', [4, 5, ImaginaryUnit, 2, 3]);
  Test('RotRight(❨i, 2, 3❩, 02)', [2, 3, ImaginaryUnit]);
  Test('RotRight(❨i, 2❩, 02)', [ImaginaryUnit, 2]);
  Test('RotRight(❨i❩, 02)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), -2)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, 03)', [3, 4, 5, ImaginaryUnit, 2]);
  Test('RotRight(❨i, 2, 3❩, 03)', [ImaginaryUnit, 2, 3]);
  Test('RotRight(❨i, 2❩, 03)', [2, ImaginaryUnit]);
  Test('RotRight(❨i❩, 03)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), -3)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, 04)', [2, 3, 4, 5, ImaginaryUnit]);
  Test('RotRight(❨i, 2, 3❩, 04)', [3, ImaginaryUnit, 2]);
  Test('RotRight(❨i, 2❩, 04)', [ImaginaryUnit, 2]);
  Test('RotRight(❨i❩, 04)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), -4)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, 05)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotRight(❨i, 2, 3❩, 05)', [2, 3, ImaginaryUnit]);
  Test('RotRight(❨i, 2❩, 05)', [2, ImaginaryUnit]);
  Test('RotRight(❨i❩, 05)', [ImaginaryUnit]);
  Test('RotRight(ComplexVector(DebugObject("empty vector")), -5)', TAlgosimComplexVector.Create);

  Test('RotRight(❨i, 2, 3, 4, 5❩, 010)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotRight(❨i, 2, 3, 4, 5❩, 011)', [5, ImaginaryUnit, 2, 3, 4]);
  Test('RotRight(❨i, 2, 3, 4, 5❩, 012)', [4, 5, ImaginaryUnit, 2, 3]);

  Test('RotRight(❨i, 2, 3, 4, 5❩, 0100)', [ImaginaryUnit, 2, 3, 4, 5]);
  Test('RotRight(❨i, 2, 3, 4, 5❩, 0101)', [5, ImaginaryUnit, 2, 3, 4]);
  Test('RotRight(❨i, 2, 3, 4, 5❩, 0102)', [4, 5, ImaginaryUnit, 2, 3]);

  Test('RotRight("nargle")', 'enargl');
  Test('RotRight("nargle", -1)', 'arglen');
  Test('RotRight("nargle", -2)', 'rglena');
  Test('RotRight("nargle", -3)', 'glenar');
  Test('RotRight("nargle", -4)', 'lenarg');
  Test('RotRight("nargle", -5)', 'enargl');
  Test('RotRight("nargle", -6)', 'nargle');
  Test('RotRight("nargle", -7)', 'arglen');
  Test('RotRight("nargle", -8)', 'rglena');
  Test('RotRight("nargle", -9)', 'glenar');
  Test('RotRight("nargle", -10)', 'lenarg');
  Test('RotRight("nargle", -11)', 'enargl');
  Test('RotRight("nargle", -12)', 'nargle');

  Test('RotRight("nargle", -60)', 'nargle');
  Test('RotRight("nargle", -61)', 'arglen');
  Test('RotRight("nargle", -62)', 'rglena');
  Test('RotRight("nargle", -63)', 'glenar');
  Test('RotRight("nargle", -64)', 'lenarg');
  Test('RotRight("nargle", -65)', 'enargl');
  Test('RotRight("nargle", -66)', 'nargle');

  Test('RotRight("nargle", 0)', 'nargle');
  Test('RotRight("nargle", 1)', 'enargl');
  Test('RotRight("nargle", 2)', 'lenarg');
  Test('RotRight("nargle", 3)', 'glenar');
  Test('RotRight("nargle", 4)', 'rglena');
  Test('RotRight("nargle", 5)', 'arglen');
  Test('RotRight("nargle", 6)', 'nargle');
  Test('RotRight("nargle", 7)', 'enargl');
  Test('RotRight("nargle", 8)', 'lenarg');
  Test('RotRight("nargle", 9)', 'glenar');
  Test('RotRight("nargle", 10)', 'rglena');
  Test('RotRight("nargle", 11)', 'arglen');
  Test('RotRight("nargle", 12)', 'nargle');

  Test('RotRight("nargle", 60)', 'nargle');
  Test('RotRight("nargle", 61)', 'enargl');
  Test('RotRight("nargle", 62)', 'lenarg');
  Test('RotRight("nargle", 63)', 'glenar');
  Test('RotRight("nargle", 64)', 'rglena');
  Test('RotRight("nargle", 65)', 'arglen');
  Test('RotRight("nargle", 66)', 'nargle');

  Test('RotRight(''(1, 2, 3, 4, 5))', intarr([5, 1, 2, 3, 4]));
  Test('RotRight(''(1, 2, 3))', intarr([3, 1, 2]));
  Test('RotRight(''(1, 2))', intarr([2, 1]));
  Test('RotRight(''(5))', intarr([5]));
  Test('RotRight(''())', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), -01)', intarr([2, 3, 4, 5, 1]));
  Test('RotRight(''(1, 2, 3), -01)', intarr([2, 3, 1]));
  Test('RotRight(''(1, 2), -01)', intarr([2, 1]));
  Test('RotRight(''(5), -01)', intarr([5]));
  Test('RotRight(''(), -01)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), -02)', intarr([3, 4, 5, 1, 2]));
  Test('RotRight(''(1, 2, 3), -02)', intarr([3, 1, 2]));
  Test('RotRight(''(1, 2), -02)', intarr([1, 2]));
  Test('RotRight(''(5), -02)', intarr([5]));
  Test('RotRight(''(), -02)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), -03)', intarr([4, 5, 1, 2, 3]));
  Test('RotRight(''(1, 2, 3), -03)', intarr([1, 2, 3]));
  Test('RotRight(''(1, 2), -03)', intarr([2, 1]));
  Test('RotRight(''(5), -03)', intarr([5]));
  Test('RotRight(''(), -03)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), -04)', intarr([5, 1, 2, 3, 4]));
  Test('RotRight(''(1, 2, 3), -04)', intarr([2, 3, 1]));
  Test('RotRight(''(1, 2), -04)', intarr([1, 2]));
  Test('RotRight(''(5), -04)', intarr([5]));
  Test('RotRight(''(), -04)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), -05)', intarr([1, 2, 3, 4, 5]));
  Test('RotRight(''(1, 2, 3), -05)', intarr([3, 1, 2]));
  Test('RotRight(''(1, 2), -05)', intarr([2, 1]));
  Test('RotRight(''(5), -05)', intarr([5]));
  Test('RotRight(''(), -05)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), -010)', intarr([1, 2, 3, 4, 5]));
  Test('RotRight(''(1, 2, 3, 4, 5), -011)', intarr([2, 3, 4, 5, 1]));
  Test('RotRight(''(1, 2, 3, 4, 5), -012)', intarr([3, 4, 5, 1, 2]));
  Test('RotRight(''(1, 2, 3, 4, 5), -0100)', intarr([1, 2, 3, 4, 5]));
  Test('RotRight(''(1, 2, 3, 4, 5), -0101)', intarr([2, 3, 4, 5, 1]));
  Test('RotRight(''(1, 2, 3, 4, 5), -0102)', intarr([3, 4, 5, 1, 2]));

  Test('RotRight(''(1, 2, 3, 4, 5), 0)', intarr([1, 2, 3, 4, 5]));
  Test('RotRight(''(1, 2, 3), 0)', intarr([1, 2, 3]));
  Test('RotRight(''(1, 2), 0)', intarr([1, 2]));
  Test('RotRight(''(5), 0)', intarr([5]));
  Test('RotRight(''(), 0)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), 01)', intarr([5, 1, 2, 3, 4]));
  Test('RotRight(''(1, 2, 3), 01)', intarr([3, 1, 2]));
  Test('RotRight(''(1, 2), 01)', intarr([2, 1]));
  Test('RotRight(''(5), 01)', intarr([5]));
  Test('RotRight(''(), 01)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), 02)', intarr([4, 5, 1, 2, 3]));
  Test('RotRight(''(1, 2, 3), 02)', intarr([2, 3, 1]));
  Test('RotRight(''(1, 2), 02)', intarr([1, 2]));
  Test('RotRight(''(5), 02)', intarr([5]));
  Test('RotRight(''(), 02)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), 03)', intarr([3, 4, 5, 1, 2]));
  Test('RotRight(''(1, 2, 3), 03)', intarr([1, 2, 3]));
  Test('RotRight(''(1, 2), 03)', intarr([2, 1]));
  Test('RotRight(''(5), 03)', intarr([5]));
  Test('RotRight(''(), 03)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), 04)', intarr([2, 3, 4, 5, 1]));
  Test('RotRight(''(1, 2, 3), 04)', intarr([3, 1, 2]));
  Test('RotRight(''(1, 2), 04)', intarr([1, 2]));
  Test('RotRight(''(5), 04)', intarr([5]));
  Test('RotRight(''(), 04)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), 05)', intarr([1, 2, 3, 4, 5]));
  Test('RotRight(''(1, 2, 3), 05)', intarr([2, 3, 1]));
  Test('RotRight(''(1, 2), 05)', intarr([2, 1]));
  Test('RotRight(''(5), 05)', intarr([5]));
  Test('RotRight(''(), 05)', intarr([]));

  Test('RotRight(''(1, 2, 3, 4, 5), 010)', intarr([1, 2, 3, 4, 5]));
  Test('RotRight(''(1, 2, 3, 4, 5), 011)', intarr([5, 1, 2, 3, 4]));
  Test('RotRight(''(1, 2, 3, 4, 5), 012)', intarr([4, 5, 1, 2, 3]));

  Test('RotRight(''(1, 2, 3, 4, 5), 0100)', intarr([1, 2, 3, 4, 5]));
  Test('RotRight(''(1, 2, 3, 4, 5), 0101)', intarr([5, 1, 2, 3, 4]));
  Test('RotRight(''(1, 2, 3, 4, 5), 0102)', intarr([4, 5, 1, 2, 3]));

  Test('A ≔ ❨❨6, 0, 2❩, ❨1, 4, 3❩, ❨8, 2, 3❩❩', 3, [6, 0, 2, 1, 4, 3, 8, 2, 3]);
  Test('MatFromCols(RotRight(cols(A)))', 3, [2, 6, 0, 3, 1, 4, 3, 8, 2]);





  //
  //
  //  CHAPTER 10
  //  Typecasts.
  //
  //

  Chapter('Typecasts');

  //
  // From integer
  //

  Test('number(0)', 0);
  Test('number(123)', 123);
  Test('number(-123)', -123);
  Test('number(123456789)', 123456789);
  Test('number(123456789123456)', 123456789123456);
  Test('number(-123456789)', -123456789);
  Test('number(-123456789123456)', -123456789123456);

  Test('integer(0)', 0);
  Test('integer(123)', 123);
  Test('integer(-123)', -123);
  Test('integer(123456789)', 123456789);
  Test('integer(123456789123456)', 123456789123456);
  Test('integer(-123456789)', -123456789);
  Test('integer(-123456789123456)', -123456789123456);

  Test('RealNumber(0)', 0.0);
  Test('RealNumber(123)', 123.0);
  Test('RealNumber(-123)', -123.0);
  Test('RealNumber(123456789)', 123456789.0);
  Test('RealNumber(123456789123456)', 123456789123456.0);
  Test('RealNumber(-123456789)', -123456789.0);
  Test('RealNumber(-123456789123456)', -123456789123456.0);

  Test('ComplexNumber(0)', TASC(0.0));
  Test('ComplexNumber(123)', TASC(123.0));
  Test('ComplexNumber(-123)', TASC(-123.0));
  Test('ComplexNumber(123456789)', TASC(123456789.0));
  Test('ComplexNumber(123456789123456)', TASC(123456789123456.0));
  Test('ComplexNumber(-123456789)', TASC(-123456789.0));
  Test('ComplexNumber(-123456789123456)', TASC(-123456789123456.0));

  Test('boolean(0)', False);
  Test('boolean(123)', True);
  Test('boolean(-123)', True);
  Test('boolean(123456789)', True);
  Test('boolean(123456789123456)', True);
  Test('boolean(-123456789)', True);
  Test('boolean(-123456789123456)', True);

  Test('string(0)', '0');
  Test('string(123)', '123');
  Test('string(-123)', '-123');
  Test('string(123456789)', '123456789');
  Test('string(123456789123456)', '123456789123456');
  Test('string(-123456789)', '-123456789');
  Test('string(-123456789123456)', '-123456789123456');

  Test('vector(0)', [0]);
  Test('vector(123)', [123]);
  Test('vector(-123)', [-123]);
  Test('vector(123456789)', [123456789]);
  Test('vector(123456789123456)', [123456789123456]);
  Test('vector(-123456789)', [-123456789]);
  Test('vector(-123456789123456)', [-123456789123456]);

  Test('RealVector(0)', [0]);
  Test('RealVector(123)', [123]);
  Test('RealVector(-123)', [-123]);
  Test('RealVector(123456789)', [123456789]);
  Test('RealVector(123456789123456)', [123456789123456]);
  Test('RealVector(-123456789)', [-123456789]);
  Test('RealVector(-123456789123456)', [-123456789123456]);

  Test('ComplexVector(0)', [TASC(0)]);
  Test('ComplexVector(123)', [TASC(123)]);
  Test('ComplexVector(-123)', [TASC(-123)]);
  Test('ComplexVector(123456789)', [TASC(123456789)]);
  Test('ComplexVector(123456789123456)', [TASC(123456789123456)]);
  Test('ComplexVector(-123456789)', [TASC(-123456789)]);
  Test('ComplexVector(-123456789123456)', [TASC(-123456789123456)]);

  Test('matrix(0)', 1, [0]);
  Test('matrix(123)', 1, [123]);
  Test('matrix(-123)', 1, [-123]);
  Test('matrix(123456789)', 1, [123456789]);
  Test('matrix(123456789123456)', 1, [123456789123456]);
  Test('matrix(-123456789)', 1, [-123456789]);
  Test('matrix(-123456789123456)', 1, [-123456789123456]);

  Test('RealMatrix(0)', 1, [0]);
  Test('RealMatrix(123)', 1, [123]);
  Test('RealMatrix(-123)', 1, [-123]);
  Test('RealMatrix(123456789)', 1, [123456789]);
  Test('RealMatrix(123456789123456)', 1, [123456789123456]);
  Test('RealMatrix(-123456789)', 1, [-123456789]);
  Test('RealMatrix(-123456789123456)', 1, [-123456789123456]);

  Test('ComplexMatrix(0)', 1, [TASC(0)]);
  Test('ComplexMatrix(123)', 1, [TASC(123)]);
  Test('ComplexMatrix(-123)', 1, [TASC(-123)]);
  Test('ComplexMatrix(123456789)', 1, [TASC(123456789)]);
  Test('ComplexMatrix(123456789123456)', 1, [TASC(123456789123456)]);
  Test('ComplexMatrix(-123456789)', 1, [TASC(-123456789)]);
  Test('ComplexMatrix(-123456789123456)', 1, [TASC(-123456789123456)]);

  Test('ToList(0)', intarr([0]));
  Test('ToList(123)', intarr([123]));
  Test('ToList(-123)', intarr([-123]));
  Test('ToList(123456789)', intarr([123456789]));
  Test('ToList(123456789123456)', intarr([123456789123456]));
  Test('ToList(-123456789)', intarr([-123456789]));
  Test('ToList(-123456789123456)', intarr([-123456789123456]));

  Test('ToSet(0)', intset([0]));
  Test('ToSet(123)', intset([123]));
  Test('ToSet(-123)', intset([-123]));
  Test('ToSet(123456789)', intset([123456789]));
  Test('ToSet(123456789123456)', intset([123456789123456]));
  Test('ToSet(-123456789)', intset([-123456789]));
  Test('ToSet(-123456789123456)', intset([-123456789123456]));

  Test('color(16#000000)', ASO(TRGB.Create(0, 0, 0)));
  Test('color(16#FF0000)', ASO(TRGB.Create(1, 0, 0)));
  Test('color(16#00FF00)', ASO(TRGB.Create(0, 1, 0)));
  Test('color(16#0000FF)', ASO(TRGB.Create(0, 0, 1)));
  Test('color(16#FFFF00)', ASO(TRGB.Create(1, 1, 0)));
  Test('color(16#FF00FF)', ASO(TRGB.Create(1, 0, 1)));
  Test('color(16#00FFFF)', ASO(TRGB.Create(0, 1, 1)));
  Test('color(16#FFFFFF)', ASO(TRGB.Create(1, 1, 1)));
  Test('color(16#808080)', ASO(TRGB.Create($80/$FF, $80/$FF, $80/$FF)));


  //
  // From rational number
  //

  Test('number(1/2 - 1/2)', rat(0, 1));
  Test('number(1/4)', rat(1, 4));
  Test('number(1/7)', rat(1, 7));
  Test('number(17/7 + 4/7)', rat(3, 1));
  Test('number(59/331)', rat(59, 331));
  Test('number(123456/123457)', rat(123456, 123457));
  Test('number(-1/4)', rat(-1, 4));
  Test('number(-1/7)', rat(-1, 7));
  Test('number(-17/7 - 4/7)', rat(-3, 1));
  Test('number(-59/331)', rat(-59, 331));
  Test('number(-123456/123457)', rat(-123456, 123457));

  Test('integer(1/2 - 1/2)', 0);
  Test('integer(1/4)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(1/7)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(17/7 + 4/7)', 3);
  Test('integer(59/331)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(123456/123457)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(-1/4)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(-1/7)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(-17/7 - 4/7)', -3);
  Test('integer(-59/331)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(-123456/123457)', failure, 'Object cannot be represented as a 64-bit integer.');

  Test('RealNumber(1/2 - 1/2)', 0.0);
  Test('RealNumber(1/4)', 0.25);
  Test('RealNumber(1/7)', 1/7);
  Test('RealNumber(17/7 + 4/7)', 3.0);
  Test('RealNumber(59/331)', 59/331);
  Test('RealNumber(123456/123457)', 123456/123457);
  Test('RealNumber(-1/4)', -0.25);
  Test('RealNumber(-1/7)', -1/7);
  Test('RealNumber(-17/7 - 4/7)', -3.0);
  Test('RealNumber(-59/331)', -59/331);
  Test('RealNumber(-123456/123457)', -123456/123457);

  Test('ComplexNumber(1/2 - 1/2)', TASC(0.0));
  Test('ComplexNumber(1/4)', TASC(0.25));
  Test('ComplexNumber(1/7)', TASC(1/7));
  Test('ComplexNumber(17/7 + 4/7)', TASC(3.0));
  Test('ComplexNumber(59/331)', TASC(59/331));
  Test('ComplexNumber(123456/123457)', TASC(123456/123457));
  Test('ComplexNumber(-1/4)', TASC(-0.25));
  Test('ComplexNumber(-1/7)', TASC(-1/7));
  Test('ComplexNumber(-17/7 - 4/7)', TASC(-3.0));
  Test('ComplexNumber(-59/331)', TASC(-59/331));
  Test('ComplexNumber(-123456/123457)', TASC(-123456/123457));

  Test('boolean(1/2 - 1/2)', False);
  Test('boolean(1/4)', True);
  Test('boolean(1/7)', True);
  Test('boolean(17/7 + 4/7)', True);
  Test('boolean(59/331)', True);
  Test('boolean(123456/123457)', True);
  Test('boolean(-1/4)', True);
  Test('boolean(-1/7)', True);
  Test('boolean(-17/7 - 4/7)', True);
  Test('boolean(-59/331)', True);
  Test('boolean(-123456/123457)', True);

  Test('string(1/2 - 1/2)', '0');
  Test('string(1/4)', '1/4');
  Test('string(1/7)', '1/7');
  Test('string(17/7 + 4/7)', '3');
  Test('string(59/331)', '59/331');
  Test('string(123456/123457)', '123456/123457');
  Test('string(-1/4)', '-1/4');
  Test('string(-1/7)', '-1/7');
  Test('string(-17/7 - 4/7)', '-3');
  Test('string(-59/331)', '-59/331');
  Test('string(-123456/123457)', '-123456/123457');

  Test('vector(1/2 - 1/2)', [0]);
  Test('vector(1/4)', [1/4]);
  Test('vector(1/7)', [1/7]);
  Test('vector(17/7 + 4/7)', [3]);
  Test('vector(59/331)', [59/331]);
  Test('vector(123456/123457)', [123456/123457]);
  Test('vector(-1/4)', [-1/4]);
  Test('vector(-1/7)', [-1/7]);
  Test('vector(-17/7 - 4/7)', [-3]);
  Test('vector(-59/331)', [-59/331]);
  Test('vector(-123456/123457)', [-123456/123457]);

  Test('RealVector(1/2 - 1/2)', [0]);
  Test('RealVector(1/4)', [1/4]);
  Test('RealVector(1/7)', [1/7]);
  Test('RealVector(17/7 + 4/7)', [3]);
  Test('RealVector(59/331)', [59/331]);
  Test('RealVector(123456/123457)', [123456/123457]);
  Test('RealVector(-1/4)', [-1/4]);
  Test('RealVector(-1/7)', [-1/7]);
  Test('RealVector(-17/7 - 4/7)', [-3]);
  Test('RealVector(-59/331)', [-59/331]);
  Test('RealVector(-123456/123457)', [-123456/123457]);

  Test('ComplexVector(1/2 - 1/2)', [TASC(0)]);
  Test('ComplexVector(1/4)', [TASC(1/4)]);
  Test('ComplexVector(1/7)', [TASC(1/7)]);
  Test('ComplexVector(17/7 + 4/7)', [TASC(3)]);
  Test('ComplexVector(59/331)', [TASC(59/331)]);
  Test('ComplexVector(123456/123457)', [TASC(123456/123457)]);
  Test('ComplexVector(-1/4)', [TASC(-1/4)]);
  Test('ComplexVector(-1/7)', [TASC(-1/7)]);
  Test('ComplexVector(-17/7 - 4/7)', [TASC(-3)]);
  Test('ComplexVector(-59/331)', [TASC(-59/331)]);
  Test('ComplexVector(-123456/123457)', [TASC(-123456/123457)]);

  Test('matrix(1/2 - 1/2)', 1, [0]);
  Test('matrix(1/4)', 1, [1/4]);
  Test('matrix(1/7)', 1, [1/7]);
  Test('matrix(17/7 + 4/7)', 1, [3]);
  Test('matrix(59/331)', 1, [59/331]);
  Test('matrix(123456/123457)', 1, [123456/123457]);
  Test('matrix(-1/4)', 1, [-1/4]);
  Test('matrix(-1/7)', 1, [-1/7]);
  Test('matrix(-17/7 - 4/7)', 1, [-3]);
  Test('matrix(-59/331)', 1, [-59/331]);
  Test('matrix(-123456/123457)', 1, [-123456/123457]);

  Test('RealMatrix(1/2 - 1/2)', 1, [0]);
  Test('RealMatrix(1/4)', 1, [1/4]);
  Test('RealMatrix(1/7)', 1, [1/7]);
  Test('RealMatrix(17/7 + 4/7)', 1, [3]);
  Test('RealMatrix(59/331)', 1, [59/331]);
  Test('RealMatrix(123456/123457)', 1, [123456/123457]);
  Test('RealMatrix(-1/4)', 1, [-1/4]);
  Test('RealMatrix(-1/7)', 1, [-1/7]);
  Test('RealMatrix(-17/7 - 4/7)', 1, [-3]);
  Test('RealMatrix(-59/331)', 1, [-59/331]);
  Test('RealMatrix(-123456/123457)', 1, [-123456/123457]);

  Test('ComplexMatrix(1/2 - 1/2)', 1, [TASC(0)]);
  Test('ComplexMatrix(1/4)', 1, [TASC(1/4)]);
  Test('ComplexMatrix(1/7)', 1, [TASC(1/7)]);
  Test('ComplexMatrix(17/7 + 4/7)', 1, [TASC(3)]);
  Test('ComplexMatrix(59/331)', 1, [TASC(59/331)]);
  Test('ComplexMatrix(123456/123457)', 1, [TASC(123456/123457)]);
  Test('ComplexMatrix(-1/4)', 1, [TASC(-1/4)]);
  Test('ComplexMatrix(-1/7)', 1, [TASC(-1/7)]);
  Test('ComplexMatrix(-17/7 - 4/7)', 1, [TASC(-3)]);
  Test('ComplexMatrix(-59/331)', 1, [TASC(-59/331)]);
  Test('ComplexMatrix(-123456/123457)', 1, [TASC(-123456/123457)]);

  Test('ToList(1/2 - 1/2)', asoarr([0]));
  Test('ToList(1/4)', asoarr([1/4]));
  Test('ToList(1/7)', asoarr([1/7]));
  Test('ToList(17/7 + 4/7)', asoarr([3]));
  Test('ToList(59/331)', asoarr([59/331]));
  Test('ToList(123456/123457)', asoarr([123456/123457]));
  Test('ToList(-1/4)', asoarr([-1/4]));
  Test('ToList(-1/7)', asoarr([-1/7]));
  Test('ToList(-17/7 - 4/7)', asoarr([-3]));
  Test('ToList(-59/331)', asoarr([-59/331]));
  Test('ToList(-123456/123457)', asoarr([-123456/123457]));

  Test('ToSet(1/2 - 1/2)', asoset([0]));
  Test('ToSet(1/4)', asoset([1/4]));
  Test('ToSet(1/7)', asoset([1/7]));
  Test('ToSet(17/7 + 4/7)', asoset([3]));
  Test('ToSet(59/331)', asoset([59/331]));
  Test('ToSet(123456/123457)', asoset([123456/123457]));
  Test('ToSet(-1/4)', asoset([-1/4]));
  Test('ToSet(-1/7)', asoset([-1/7]));
  Test('ToSet(-17/7 - 4/7)', asoset([-3]));
  Test('ToSet(-59/331)', asoset([-59/331]));
  Test('ToSet(-123456/123457)', asoset([-123456/123457]));

  Test('color(1/2 - 1/2)', ASO(TRGB.Create(0, 0, 0)));
  Test('color(1/2 - 1/2 + 255/1)', ASO(TRGB.Create(0, 0, 1)));
  Test('color(1/2)', failure, 'A number which isn''t a 32-bit integer cannot be converted to a colour.');


  //
  // From real number
  //

  Test('number(0.0)', 0.0);
  Test('number(21.0)', 21.0);
  Test('number(0.25)', 0.25);
  Test('number(123.456)', 123.456);
  Test('number(5.2E10)', 5.2E10);
  Test('number(5.2E100)', 5.2E100);
  Test('number(5.2E-10)', 5.2E-10);
  Test('number(5.2E-100)', 5.2E-100);
  Test('number(-21.0)', -21.0);
  Test('number(-0.25)', -0.25);
  Test('number(-123.456)', -123.456);
  Test('number(-5.2E10)', -5.2E10);
  Test('number(-5.2E100)', -5.2E100);
  Test('number(-5.2E-10)', -5.2E-10);
  Test('number(-5.2E-100)', -5.2E-100);

  Test('integer(0.0)', 0);
  Test('integer(21.0)', 21);
  Test('integer(0.25)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(123.456)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(5.2E10)', 52000000000);
  Test('integer(5.2E100)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(5.2E-10)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(5.2E-100)', 0);
  Test('integer(-21.0)', -21);
  Test('integer(-0.25)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(-123.456)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(-5.2E10)', -52000000000);
  Test('integer(-5.2E100)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(-5.2E-10)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(-5.2E-100)', 0);

  Test('RealNumber(0.0)', 0.0);
  Test('RealNumber(21.0)', 21.0);
  Test('RealNumber(0.25)', 0.25);
  Test('RealNumber(123.456)', 123.456);
  Test('RealNumber(5.2E10)', 5.2E10);
  Test('RealNumber(5.2E100)', 5.2E100);
  Test('RealNumber(5.2E-10)', 5.2E-10);
  Test('RealNumber(5.2E-100)', 5.2E-100);
  Test('RealNumber(-21.0)', -21.0);
  Test('RealNumber(-0.25)', -0.25);
  Test('RealNumber(-123.456)', -123.456);
  Test('RealNumber(-5.2E10)', -5.2E10);
  Test('RealNumber(-5.2E100)', -5.2E100);
  Test('RealNumber(-5.2E-10)', -5.2E-10);
  Test('RealNumber(-5.2E-100)', -5.2E-100);

  Test('ComplexNumber(0.0)', TASC(0.0));
  Test('ComplexNumber(21.0)', TASC(21.0));
  Test('ComplexNumber(0.25)', TASC(0.25));
  Test('ComplexNumber(123.456)', TASC(123.456));
  Test('ComplexNumber(5.2E10)', TASC(5.2E10));
  Test('ComplexNumber(5.2E100)', TASC(5.2E100));
  Test('ComplexNumber(5.2E-10)', TASC(5.2E-10));
  Test('ComplexNumber(5.2E-100)', TASC(5.2E-100));
  Test('ComplexNumber(-21.0)', TASC(-21.0));
  Test('ComplexNumber(-0.25)', TASC(-0.25));
  Test('ComplexNumber(-123.456)', TASC(-123.456));
  Test('ComplexNumber(-5.2E10)', TASC(-5.2E10));
  Test('ComplexNumber(-5.2E100)', TASC(-5.2E100));
  Test('ComplexNumber(-5.2E-10)', TASC(-5.2E-10));
  Test('ComplexNumber(-5.2E-100)', TASC(-5.2E-100));

  Test('boolean(0.0)', False);
  Test('boolean(21.0)', True);
  Test('boolean(0.25)', True);
  Test('boolean(123.456)', True);
  Test('boolean(5.2E10)', True);
  Test('boolean(5.2E100)', True);
  Test('boolean(5.2E-10)', True);
  Test('boolean(5.2E-100)', True);
  Test('boolean(-21.0)', True);
  Test('boolean(-0.25)', True);
  Test('boolean(-123.456)', True);
  Test('boolean(-5.2E10)', True);
  Test('boolean(-5.2E100)', True);
  Test('boolean(-5.2E-10)', True);
  Test('boolean(-5.2E-100)', True);

  Test('string(0.0)', '0');
  Test('string(21.0)', '21');
  Test('string(0.25)', '0.25');
  Test('string(123.456)', '123.456');
  Test('string(5.2E10)', '52000000000');
  Test('string(5.2E100)', '5.2E100');
  Test('string(5.2E-10)', '5.2E-10');
  Test('string(5.2E-100)', '5.2E-100');
  Test('string(-21.0)', '-21');
  Test('string(-0.25)', '-0.25');
  Test('string(-123.456)', '-123.456');
  Test('string(-5.2E10)', '-52000000000');
  Test('string(-5.2E100)', '-5.2E100');
  Test('string(-5.2E-10)', '-5.2E-10');
  Test('string(-5.2E-100)', '-5.2E-100');

  Test('vector(0.0)', [0]);
  Test('vector(21.0)', [21]);
  Test('vector(0.25)', [0.25]);
  Test('vector(123.456)', [123.456]);
  Test('vector(5.2E10)', [52000000000]);
  Test('vector(5.2E100)', [5.2E100]);
  Test('vector(5.2E-10)', [5.2E-10]);
  Test('vector(5.2E-100)', [5.2E-100]);
  Test('vector(-21.0)', [-21]);
  Test('vector(-0.25)', [-0.25]);
  Test('vector(-123.456)', [-123.456]);
  Test('vector(-5.2E10)', [-52000000000]);
  Test('vector(-5.2E100)', [-5.2E100]);
  Test('vector(-5.2E-10)', [-5.2E-10]);
  Test('vector(-5.2E-100)', [-5.2E-100]);

  Test('RealVector(0.0)', [0]);
  Test('RealVector(21.0)', [21]);
  Test('RealVector(0.25)', [0.25]);
  Test('RealVector(123.456)', [123.456]);
  Test('RealVector(5.2E10)', [52000000000]);
  Test('RealVector(5.2E100)', [5.2E100]);
  Test('RealVector(5.2E-10)', [5.2E-10]);
  Test('RealVector(5.2E-100)', [5.2E-100]);
  Test('RealVector(-21.0)', [-21]);
  Test('RealVector(-0.25)', [-0.25]);
  Test('RealVector(-123.456)', [-123.456]);
  Test('RealVector(-5.2E10)', [-52000000000]);
  Test('RealVector(-5.2E100)', [-5.2E100]);
  Test('RealVector(-5.2E-10)', [-5.2E-10]);
  Test('RealVector(-5.2E-100)', [-5.2E-100]);

  Test('ComplexVector(0.0)', [TASC(0)]);
  Test('ComplexVector(21.0)', [TASC(21)]);
  Test('ComplexVector(0.25)', [TASC(0.25)]);
  Test('ComplexVector(123.456)', [TASC(123.456)]);
  Test('ComplexVector(5.2E10)', [TASC(52000000000)]);
  Test('ComplexVector(5.2E100)', [TASC(5.2E100)]);
  Test('ComplexVector(5.2E-10)', [TASC(5.2E-10)]);
  Test('ComplexVector(5.2E-100)', [TASC(5.2E-100)]);
  Test('ComplexVector(-21.0)', [TASC(-21)]);
  Test('ComplexVector(-0.25)', [TASC(-0.25)]);
  Test('ComplexVector(-123.456)', [TASC(-123.456)]);
  Test('ComplexVector(-5.2E10)', [TASC(-52000000000)]);
  Test('ComplexVector(-5.2E100)', [TASC(-5.2E100)]);
  Test('ComplexVector(-5.2E-10)', [TASC(-5.2E-10)]);
  Test('ComplexVector(-5.2E-100)', [TASC(-5.2E-100)]);

  Test('matrix(0.0)', 1, [0]);
  Test('matrix(21.0)', 1, [21]);
  Test('matrix(0.25)', 1, [0.25]);
  Test('matrix(123.456)', 1, [123.456]);
  Test('matrix(5.2E10)', 1, [52000000000]);
  Test('matrix(5.2E100)', 1, [5.2E100]);
  Test('matrix(5.2E-10)', 1, [5.2E-10]);
  Test('matrix(5.2E-100)', 1, [5.2E-100]);
  Test('matrix(-21.0)', 1, [-21]);
  Test('matrix(-0.25)', 1, [-0.25]);
  Test('matrix(-123.456)', 1, [-123.456]);
  Test('matrix(-5.2E10)', 1, [-52000000000]);
  Test('matrix(-5.2E100)', 1, [-5.2E100]);
  Test('matrix(-5.2E-10)', 1, [-5.2E-10]);
  Test('matrix(-5.2E-100)', 1, [-5.2E-100]);

  Test('RealMatrix(0.0)', 1, [0]);
  Test('RealMatrix(21.0)', 1, [21]);
  Test('RealMatrix(0.25)', 1, [0.25]);
  Test('RealMatrix(123.456)', 1, [123.456]);
  Test('RealMatrix(5.2E10)', 1, [52000000000]);
  Test('RealMatrix(5.2E100)', 1, [5.2E100]);
  Test('RealMatrix(5.2E-10)', 1, [5.2E-10]);
  Test('RealMatrix(5.2E-100)', 1, [5.2E-100]);
  Test('RealMatrix(-21.0)', 1, [-21]);
  Test('RealMatrix(-0.25)', 1, [-0.25]);
  Test('RealMatrix(-123.456)', 1, [-123.456]);
  Test('RealMatrix(-5.2E10)', 1, [-52000000000]);
  Test('RealMatrix(-5.2E100)', 1, [-5.2E100]);
  Test('RealMatrix(-5.2E-10)', 1, [-5.2E-10]);
  Test('RealMatrix(-5.2E-100)', 1, [-5.2E-100]);

  Test('ComplexMatrix(0.0)', 1, [TASC(0)]);
  Test('ComplexMatrix(21.0)', 1, [TASC(21)]);
  Test('ComplexMatrix(0.25)', 1, [TASC(0.25)]);
  Test('ComplexMatrix(123.456)', 1, [TASC(123.456)]);
  Test('ComplexMatrix(5.2E10)', 1, [TASC(52000000000)]);
  Test('ComplexMatrix(5.2E100)', 1, [TASC(5.2E100)]);
  Test('ComplexMatrix(5.2E-10)', 1, [TASC(5.2E-10)]);
  Test('ComplexMatrix(5.2E-100)', 1, [TASC(5.2E-100)]);
  Test('ComplexMatrix(-21.0)', 1, [TASC(-21)]);
  Test('ComplexMatrix(-0.25)', 1, [TASC(-0.25)]);
  Test('ComplexMatrix(-123.456)', 1, [TASC(-123.456)]);
  Test('ComplexMatrix(-5.2E10)', 1, [TASC(-52000000000)]);
  Test('ComplexMatrix(-5.2E100)', 1, [TASC(-5.2E100)]);
  Test('ComplexMatrix(-5.2E-10)', 1, [TASC(-5.2E-10)]);
  Test('ComplexMatrix(-5.2E-100)', 1, [TASC(-5.2E-100)]);

  Test('ToList(0.0)', asoarr([0]));
  Test('ToList(21.0)', asoarr([21]));
  Test('ToList(0.25)', asoarr([0.25]));
  Test('ToList(123.456)', asoarr([123.456]));
  Test('ToList(5.2E10)', asoarr([52000000000]));
  Test('ToList(5.2E100)', asoarr([5.2E100]));
  Test('ToList(5.2E-10)', asoarr([5.2E-10]));
  Test('ToList(5.2E-100)', asoarr([5.2E-100]));
  Test('ToList(-21.0)', asoarr([-21]));
  Test('ToList(-0.25)', asoarr([-0.25]));
  Test('ToList(-123.456)', asoarr([-123.456]));
  Test('ToList(-5.2E10)', asoarr([-52000000000]));
  Test('ToList(-5.2E100)', asoarr([-5.2E100]));
  Test('ToList(-5.2E-10)', asoarr([-5.2E-10]));
  Test('ToList(-5.2E-100)', asoarr([-5.2E-100]));

  Test('ToSet(0.0)', asoset([0]));
  Test('ToSet(21.0)', asoset([21]));
  Test('ToSet(0.25)', asoset([0.25]));
  Test('ToSet(123.456)', asoset([123.456]));
  Test('ToSet(5.2E10)', asoset([52000000000]));
  Test('ToSet(5.2E100)', asoset([5.2E100]));
  Test('ToSet(5.2E-10)', asoset([5.2E-10]));
  Test('ToSet(5.2E-100)', asoset([5.2E-100]));
  Test('ToSet(-21.0)', asoset([-21]));
  Test('ToSet(-0.25)', asoset([-0.25]));
  Test('ToSet(-123.456)', asoset([-123.456]));
  Test('ToSet(-5.2E10)', asoset([-52000000000]));
  Test('ToSet(-5.2E100)', asoset([-5.2E100]));
  Test('ToSet(-5.2E-10)', asoset([-5.2E-10]));
  Test('ToSet(-5.2E-100)', asoset([-5.2E-100]));

  Test('color(0.0 + 16#000000)', ASO(TRGB.Create(0, 0, 0)));
  Test('color(0.0 + 16#FF0000)', ASO(TRGB.Create(1, 0, 0)));
  Test('color(0.0 + 16#00FF00)', ASO(TRGB.Create(0, 1, 0)));
  Test('color(0.0 + 16#0000FF)', ASO(TRGB.Create(0, 0, 1)));
  Test('color(0.0 + 16#FFFF00)', ASO(TRGB.Create(1, 1, 0)));
  Test('color(0.0 + 16#FF00FF)', ASO(TRGB.Create(1, 0, 1)));
  Test('color(0.0 + 16#00FFFF)', ASO(TRGB.Create(0, 1, 1)));
  Test('color(0.0 + 16#FFFFFF)', ASO(TRGB.Create(1, 1, 1)));
  Test('color(0.0 + 16#808080)', ASO(TRGB.Create($80/$FF, $80/$FF, $80/$FF)));
  Test('color(3.1415)', failure, 'A number which isn''t a 32-bit integer cannot be converted to a colour.');


  //
  // From complex number
  //

  Test('number(i - i)', TASC(0));
  Test('number(i - i + 5)', TASC(5));
  Test('number(i - i + 0.25)', TASC(0.25));
  Test('number(i - i + 123.456)', TASC(123.456));
  Test('number(i - i + 123.456E10)', TASC(123.456E10));
  Test('number(i)', ImaginaryUnit);
  Test('number(1 + i)', 1 + ImaginaryUnit);
  Test('number(5 - 2⋅i)', 5 - 2*ImaginaryUnit);

  Test('integer(i - i)', 0);
  Test('integer(i - i + 5)', 5);
  Test('integer(i - i + 0.25)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(i - i + 123.456)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(i - i + 123.456E10)', 1234560000000);
  Test('integer(i)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(1 + i)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(5 - 2⋅i)', failure, 'Object cannot be represented as a 64-bit integer.');

  Test('RealNumber(i - i)', 0.0);
  Test('RealNumber(i - i + 5)', 5.0);
  Test('RealNumber(i - i + 0.25)', 0.25);
  Test('RealNumber(i - i + 123.456)', 123.456);
  Test('RealNumber(i - i + 123.456E10)', 123.456E10);
  Test('RealNumber(i)', failure, 'Complex number i cannot be converted to a real number.');
  Test('RealNumber(1 + i)', failure, 'Complex number 1 + i cannot be converted to a real number.');
  Test('RealNumber(5 - 2⋅i)', failure, 'Complex number 5 − 2⋅i cannot be converted to a real number.');

  Test('ComplexNumber(i - i)', TASC(0));
  Test('ComplexNumber(i - i + 5)', TASC(5));
  Test('ComplexNumber(i - i + 0.25)', TASC(0.25));
  Test('ComplexNumber(i - i + 123.456)', TASC(123.456));
  Test('ComplexNumber(i - i + 123.456E10)', TASC(123.456E10));
  Test('ComplexNumber(i)', ImaginaryUnit);
  Test('ComplexNumber(1 + i)', 1 + ImaginaryUnit);
  Test('ComplexNumber(5 - 2⋅i)', 5 - 2*ImaginaryUnit);

  Test('boolean(i - i)', False);
  Test('boolean(i - i + 5)', True);
  Test('boolean(i - i + 0.25)', True);
  Test('boolean(i - i + 123.456)', True);
  Test('boolean(i - i + 123.456E10)', True);
  Test('boolean(i)', True);
  Test('boolean(1 + i)', True);
  Test('boolean(5 - 2⋅i)', True);
  Test('boolean(i/1E100)', True);
  Test('boolean(i - i + 1E-100)', True);

  Test('string(i - i)', '0');
  Test('string(i - i + 5)', '5');
  Test('string(i - i − 5)', '-5');
  Test('string(i - i + 0.25)', '0.25');
  Test('string(i - i + 123.456)', '123.456');
  Test('string(i - i + 123.456E10)', '1234560000000');
  Test('string(i)', 'i');
  Test('string(−i)', '-i');
  Test('string(1 + i)', '1 + i');
  Test('string(5 − 2⋅i)', '5 - 2⋅i');
  Test('string(−1 + i)', '-1 + i');
  Test('string(−5 − 2⋅i)', '-5 - 2⋅i');

  Test('vector(i - i)', [TASC(0)]);
  Test('vector(i - i + 5)', [TASC(5)]);
  Test('vector(i - i + 0.25)', [TASC(0.25)]);
  Test('vector(i - i + 123.456)', [TASC(123.456)]);
  Test('vector(i - i + 123.456E10)', [TASC(123.456E10)]);
  Test('vector(i)', [ImaginaryUnit]);
  Test('vector(1 + i)', [1 + ImaginaryUnit]);
  Test('vector(5 - 2⋅i)', [5 - 2*ImaginaryUnit]);

  Test('RealVector(i - i)', [0, 0]);
  Test('RealVector(i - i + 5)', [5, 0]);
  Test('RealVector(i - i + 0.25)', [0.25, 0]);
  Test('RealVector(i - i + 123.456)', [123.456, 0]);
  Test('RealVector(i - i + 123.456E10)', [123.456E10, 0]);
  Test('RealVector(i)', [0, 1]);
  Test('RealVector(1 + i)', [1, 1]);
  Test('RealVector(5 - 2⋅i)', [5, -2]);

  Test('RealVector(RealNumber(i - i))', [0]);
  Test('RealVector(RealNumber(i - i + 5))', [5]);
  Test('RealVector(RealNumber(i - i + 0.25))', [0.25]);
  Test('RealVector(RealNumber(i - i + 123.456))', [123.456]);
  Test('RealVector(RealNumber(i - i + 123.456E10))', [123.456E10]);
  Test('RealVector(RealNumber(i))', failure, 'Complex number i cannot be converted to a real number.');
  Test('RealVector(RealNumber(1 + i))', failure, 'Complex number 1 + i cannot be converted to a real number.');
  Test('RealVector(RealNumber(5 - 2⋅i))', failure, 'Complex number 5 − 2⋅i cannot be converted to a real number.');

  Test('ComplexVector(i - i)', [TASC(0)]);
  Test('ComplexVector(i - i + 5)', [TASC(5)]);
  Test('ComplexVector(i - i + 0.25)', [TASC(0.25)]);
  Test('ComplexVector(i - i + 123.456)', [TASC(123.456)]);
  Test('ComplexVector(i - i + 123.456E10)', [TASC(123.456E10)]);
  Test('ComplexVector(i)', [ImaginaryUnit]);
  Test('ComplexVector(1 + i)', [1 + ImaginaryUnit]);
  Test('ComplexVector(5 - 2⋅i)', [5 - 2*ImaginaryUnit]);

  Test('matrix(i - i)', 1, [TASC(0)]);
  Test('matrix(i - i + 5)', 1, [TASC(5)]);
  Test('matrix(i - i + 0.25)', 1, [TASC(0.25)]);
  Test('matrix(i - i + 123.456)', 1, [TASC(123.456)]);
  Test('matrix(i - i + 123.456E10)', 1, [TASC(123.456E10)]);
  Test('matrix(i)', 1, [ImaginaryUnit]);
  Test('matrix(1 + i)', 1, [1 + ImaginaryUnit]);
  Test('matrix(5 - 2⋅i)', 1, [5 - 2*ImaginaryUnit]);

  Test('RealMatrix(i - i)', 1, [0]);
  Test('RealMatrix(i - i + 5)', 1, [5]);
  Test('RealMatrix(i - i + 0.25)', 1, [0.25]);
  Test('RealMatrix(i - i + 123.456)', 1, [123.456]);
  Test('RealMatrix(i - i + 123.456E10)', 1, [123.456E10]);
  Test('RealMatrix(i)', failure, 'Complex number i cannot be converted to a real number.');
  Test('RealMatrix(1 + i)', failure, 'Complex number 1 + i cannot be converted to a real number.');
  Test('RealMatrix(5 - 2⋅i)', failure, 'Complex number 5 − 2⋅i cannot be converted to a real number.');

  Test('ComplexMatrix(i - i)', 1, [TASC(0)]);
  Test('ComplexMatrix(i - i + 5)', 1, [TASC(5)]);
  Test('ComplexMatrix(i - i + 0.25)', 1, [TASC(0.25)]);
  Test('ComplexMatrix(i - i + 123.456)', 1, [TASC(123.456)]);
  Test('ComplexMatrix(i - i + 123.456E10)', 1, [TASC(123.456E10)]);
  Test('ComplexMatrix(i)', 1, [ImaginaryUnit]);
  Test('ComplexMatrix(1 + i)', 1, [1 + ImaginaryUnit]);
  Test('ComplexMatrix(5 - 2⋅i)', 1, [5 - 2*ImaginaryUnit]);

  Test('ToList(i - i)', asoarrex([ASO(TASC(0))]));
  Test('ToList(i - i + 5)', asoarrex([ASO(TASC(5))]));
  Test('ToList(i - i + 0.25)', asoarrex([ASO(TASC(0.25))]));
  Test('ToList(i - i + 123.456)', asoarrex([ASO(TASC(123.456))]));
  Test('ToList(i - i + 123.456E10)', asoarrex([ASO(TASC(123.456E10))]));
  Test('ToList(i)', asoarrex([ASO(ImaginaryUnit)]));
  Test('ToList(1 + i)', asoarrex([ASO(1 + ImaginaryUnit)]));
  Test('ToList(5 - 2⋅i)', asoarrex([ASO(5 - 2*ImaginaryUnit)]));

  Test('ToSet(i - i)', asosetex([ASO(TASC(0))]));
  Test('ToSet(i - i + 5)', asosetex([ASO(TASC(5))]));
  Test('ToSet(i - i + 0.25)', asosetex([ASO(TASC(0.25))]));
  Test('ToSet(i - i + 123.456)', asosetex([ASO(TASC(123.456))]));
  Test('ToSet(i - i + 123.456E10)', asosetex([ASO(TASC(123.456E10))]));
  Test('ToSet(i)', asosetex([ASO(ImaginaryUnit)]));
  Test('ToSet(1 + i)', asosetex([ASO(1 + ImaginaryUnit)]));
  Test('ToSet(5 - 2⋅i)', asosetex([ASO(5 - 2*ImaginaryUnit)]));

  Test('color(i - i + 16#000000)', ASO(TRGB.Create(0, 0, 0)));
  Test('color(i - i + 16#FF0000)', ASO(TRGB.Create(1, 0, 0)));
  Test('color(i - i + 16#00FF00)', ASO(TRGB.Create(0, 1, 0)));
  Test('color(i - i + 16#0000FF)', ASO(TRGB.Create(0, 0, 1)));
  Test('color(i - i + 16#FFFF00)', ASO(TRGB.Create(1, 1, 0)));
  Test('color(i - i + 16#FF00FF)', ASO(TRGB.Create(1, 0, 1)));
  Test('color(i - i + 16#00FFFF)', ASO(TRGB.Create(0, 1, 1)));
  Test('color(i - i + 16#FFFFFF)', ASO(TRGB.Create(1, 1, 1)));
  Test('color(i - i + 16#808080)', ASO(TRGB.Create($80/$FF, $80/$FF, $80/$FF)));
  Test('color(i - i + 3.1415)', failure, 'A number which isn''t a 32-bit integer cannot be converted to a colour.');
  Test('color(i + 3.1415)', failure, 'A number which isn''t a 32-bit integer cannot be converted to a colour.');


  //
  // From boolean
  //

  Test('number(false)', 0);
  Test('number(true)', 1);

  Test('integer(false)', 0);
  Test('integer(true)', 1);

  Test('RealNumber(false)', 0.0);
  Test('RealNumber(true)', 1.0);

  Test('ComplexNumber(false)', TASC(0));
  Test('ComplexNumber(true)', TASC(1));

  Test('boolean(false)', False);
  Test('boolean(true)', True);

  Test('string(false)', 'false');
  Test('string(true)', 'true');

  Test('string(number(false))', '0');
  Test('string(number(true))', '1');

  Test('vector(false)', failure, 'An object of type "boolean" cannot be converted to a real vector.');
  Test('vector(true)', failure, 'An object of type "boolean" cannot be converted to a real vector.');
  Test('RealVector(false)', failure, 'An object of type "boolean" cannot be converted to a real vector.');
  Test('RealVector(true)', failure, 'An object of type "boolean" cannot be converted to a real vector.');
  Test('ComplexVector(false)', failure, 'An object of type "boolean" cannot be converted to a complex vector.');
  Test('ComplexVector(true)', failure, 'An object of type "boolean" cannot be converted to a complex vector.');

  Test('matrix(false)', failure, 'An object of type "boolean" cannot be converted to a real matrix.');
  Test('matrix(true)', failure, 'An object of type "boolean" cannot be converted to a real matrix.');
  Test('RealMatrix(false)', failure, 'An object of type "boolean" cannot be converted to a real matrix.');
  Test('RealMatrix(true)', failure, 'An object of type "boolean" cannot be converted to a real matrix.');
  Test('ComplexMatrix(false)', failure, 'An object of type "boolean" cannot be converted to a complex matrix.');
  Test('ComplexMatrix(true)', failure, 'An object of type "boolean" cannot be converted to a complex matrix.');

  Test('ToList(false)', asoarr([False]));
  Test('ToList(true)', asoarr([True]));

  Test('ToSet(false)', asoset([False]));
  Test('ToSet(true)', asoset([True]));

  Test('color(false)', failure, 'An object of type "boolean" cannot be converted to a colour.');
  Test('color(true)', failure, 'An object of type "boolean" cannot be converted to a colour.');


  //
  // From string
  //

  Test('number("0")', 0);
  Test('number("123")', 123);
  Test('number(" 123  ")', 123);  
  Test('number("123456789")', 123456789);
  Test('number("123456789123456")', 123456789123456);
  Test('number("-123")', -123);
  Test('number(" - 123  ")', -123);  
  Test('number("-123456789")', -123456789);
  Test('number("-123456789123456")', -123456789123456);

  Test('number("123 456 789")', 123456789);
  Test('number("123 456 789 123 456")', 123456789123456);
  Test('number("-123 456 789")', -123456789);
  Test('number("-123 456 789 123 456")', -123456789123456);

  Test('number("1/4")', Rat(1, 4));
  Test('number("123456/123457")', Rat(123456, 123457));
  Test('number("123 456/123 457")', Rat(123456, 123457));
  Test('number("-1/4")', Rat(-1, 4));
  Test('number("-123456/123457")', Rat(-123456, 123457));
  Test('number("-123 456/123 457")', Rat(-123456, 123457));

  Test('number("0.0")', 0.0);
  Test('number("0.000000")', 0.0);
  Test('number("0.123")', 0.123);
  Test('number("0.123456")', 0.123456);
  Test('number("0.123 456")', 0.123456);
  Test('number("58.223")', 58.223);
  Test('number("6.3E5")', 6.3E5);
  Test('number("6.3E10")', 6.3E10);
  Test('number("6.3E100")', 6.3E100);
  Test('number("6.3E+100")', 6.3E+100);
  Test('number("6.3E-5")', 6.3E-5);
  Test('number("6.3E-10")', 6.3E-10);
  Test('number("6.3E-100")', 6.3E-100);
  Test('number("-0.123")', -0.123);
  Test('number("-58.223")', -58.223);
  Test('number("-6.3E5")', -6.3E5);
  Test('number("-6.3E10")', -6.3E10);
  Test('number("-6.3E100")', -6.3E100);
  Test('number("-6.3E+100")', -6.3E+100);
  Test('number("-6.3E-5")', -6.3E-5);
  Test('number("-6.3E-10")', -6.3E-10);
  Test('number("-6.3E-100")', -6.3E-100);

  Test('number("5 + 2i")', ASC(5, 2));
  Test('number("5 + 2 i")', ASC(5, 2));
  Test('number("5 + 2*i")', ASC(5, 2));
  Test('number("5 + 2⋅i")', ASC(5, 2));
  Test('number("5 + 2j")', ASC(5, 2));
  Test('number("5 + 2 j")', ASC(5, 2));
  Test('number("5 + 2*j")', ASC(5, 2));
  Test('number("5 + 2⋅j")', ASC(5, 2));
  Test('number("5 + 2I")', ASC(5, 2));
  Test('number("5 + 2 I")', ASC(5, 2));
  Test('number("5 + 2*I")', ASC(5, 2));
  Test('number("5 + 2⋅I")', ASC(5, 2));
  Test('number("5 + 2J")', ASC(5, 2));
  Test('number("5 + 2 J")', ASC(5, 2));
  Test('number("5 + 2*J")', ASC(5, 2));
  Test('number("5 + 2⋅J")', ASC(5, 2));
  Test('number("5 - 2i")', ASC(5, -2));
  Test('number("5 - 2 i")', ASC(5, -2));
  Test('number("5 - 2*i")', ASC(5, -2));
  Test('number("5 - 2⋅i")', ASC(5, -2));
  Test('number("5 - 2j")', ASC(5, -2));
  Test('number("5 - 2 j")', ASC(5, -2));
  Test('number("5 - 2*j")', ASC(5, -2));
  Test('number("5 - 2⋅j")', ASC(5, -2));
  Test('number("5 - 2I")', ASC(5, -2));
  Test('number("5 - 2 I")', ASC(5, -2));
  Test('number("5 - 2*I")', ASC(5, -2));
  Test('number("5 - 2⋅I")', ASC(5, -2));
  Test('number("5 - 2J")', ASC(5, -2));
  Test('number("5 - 2 J")', ASC(5, -2));
  Test('number("5 - 2*J")', ASC(5, -2));
  Test('number("5 - 2⋅J")', ASC(5, -2));
  Test('number("2.41E-7 + 3.2 j")', ASC(2.41E-7, 3.2));

  Test('number("−123")', -123);
  Test('number(" − 123  ")', -123);  
  Test('number("−123456789")', -123456789);
  Test('number("−123 456 789")', -123456789);
  Test('number("−123456789123456")', -123456789123456);
  Test('number("−123 456 789 123 456")', -123456789123456);
  Test('number("−1/4")', Rat(-1, 4));
  Test('number("−123456/123457")', Rat(-123456, 123457));
  Test('number("−123 456/123 457")', Rat(-123456, 123457));
  Test('number("6.3E−5")', 6.3E-5);
  Test('number("6.3E−10")', 6.3E-10);
  Test('number("6.3E−100")', 6.3E-100);
  Test('number("−0.123")', -0.123);
  Test('number("−0.123456")', -0.123456);
  Test('number("−0.123 456")', -0.123456);
  Test('number("−58.223")', -58.223);
  Test('number("−6.3E5")', -6.3E5);
  Test('number("−6.3E10")', -6.3E10);
  Test('number("−6.3E100")', -6.3E100);
  Test('number("−6.3E+100")', -6.3E+100);
  Test('number("−6.3E−5")', -6.3E-5);
  Test('number("−6.3E−10")', -6.3E-10);
  Test('number("−6.3E−100")', -6.3E-100);
  Test('number("5 − 2i")', ASC(5, -2));
  Test('number("5 − 2 i")', ASC(5, -2));
  Test('number("5 − 2*i")', ASC(5, -2));
  Test('number("5 − 2⋅i")', ASC(5, -2));
  Test('number("5 − 2j")', ASC(5, -2));
  Test('number("5 − 2 j")', ASC(5, -2));
  Test('number("5 − 2*j")', ASC(5, -2));
  Test('number("5 − 2⋅j")', ASC(5, -2));
  Test('number("5 − 2I")', ASC(5, -2));
  Test('number("5 − 2 I")', ASC(5, -2));
  Test('number("5 − 2*I")', ASC(5, -2));
  Test('number("5 − 2⋅I")', ASC(5, -2));
  Test('number("5 − 2J")', ASC(5, -2));
  Test('number("5 − 2 J")', ASC(5, -2));
  Test('number("5 − 2*J")', ASC(5, -2));
  Test('number("5 − 2⋅J")', ASC(5, -2));
  Test('number("−2.41E−7 − 3.2 j")', ASC(-2.41E-7, -3.2));

  Test('number("123g")', failure, 'Cannot convert string "123g" to a complex number.');
  Test('number("123..")', failure, 'Cannot convert string "123.." to a complex number.');
  Test('number("3.14E5.1")', failure, 'Cannot convert string "3.14E5.1" to a complex number.');
  Test('number("6.5 + 3⋅i + 5⋅k")', failure, 'Cannot convert string "6.5 + 3⋅i + 5⋅k" to a complex number.');

  Test('integer("0")', 0);
  Test('integer("123")', 123);
  Test('integer(" 123  ")', 123);
  Test('integer("123456789")', 123456789);
  Test('integer("123456789123456")', 123456789123456);
  Test('integer("123 456 789")', 123456789);
  Test('integer("123 456 789 123 456")', 123456789123456);
  Test('integer("-123")', -123);
  Test('integer(" - 123  ")', -123);
  Test('integer("-123 456 789")', -123456789);
  Test('integer("-123 456 789 123 456")', -123456789123456);
  Test('integer("5.0")', 5);
  Test('integer("-5.0")', -5);
  Test('integer("123g")', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer("5.3")', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer("5.3E100")', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer("")', failure, 'Object cannot be represented as a 64-bit integer.');

  Test('integer("−123")', -123);
  Test('integer(" − 123  ")', -123);
  Test('integer("−123 456 789")', -123456789);
  Test('integer("−123 456 789 123 456")', -123456789123456);
  Test('integer("−5.0")', -5);  

  Test('RealNumber("0")', 0.0);
  Test('RealNumber("123")', 123.0);
  Test('RealNumber("123456789")', 123456789.0);
  Test('RealNumber("123456789123456")', 123456789123456.0);
  Test('RealNumber("-123")', -123.0);
  Test('RealNumber("-123456789")', -123456789.0);
  Test('RealNumber("-123456789123456")', -123456789123456.0);
  Test('RealNumber("0.123456789")', 0.123456789);
  Test('RealNumber("1.5")', 1.5);
  Test('RealNumber("123.456")', 123.456);
  Test('RealNumber("5.1E10")', 5.1E10);
  Test('RealNumber("5.1E100")', 5.1E100);
  Test('RealNumber("5.1E-10")', 5.1E-10);
  Test('RealNumber("5.1E-100")', 5.1E-100);
  Test('RealNumber("-0.123456789")', -0.123456789);
  Test('RealNumber("-1.5")', -1.5);
  Test('RealNumber("-123.456")', -123.456);
  Test('RealNumber("-5.1E10")', -5.1E10);
  Test('RealNumber("-5.1E100")', -5.1E100);
  Test('RealNumber("-5.1E-10")', -5.1E-10);
  Test('RealNumber("-5.1E-100")', -5.1E-100);
  Test('RealNumber("5.3g")', failure, 'Cannot convert string "5.3g" to a real number.');
  Test('RealNumber("5.3.2")', failure, 'Cannot convert string "5.3.2" to a real number.');
  Test('RealNumber("5.3E2.2")', failure, 'Cannot convert string "5.3E2.2" to a real number.');
  Test('RealNumber("")', failure, 'Cannot convert string "" to a real number.');

  Test('RealNumber("−123")', -123.0);
  Test('RealNumber("−123456789")', -123456789.0);
  Test('RealNumber("−123456789123456")', -123456789123456.0);
  Test('RealNumber("−0.123456789")', -0.123456789);
  Test('RealNumber("−1.5")', -1.5);
  Test('RealNumber("−123.456")', -123.456);
  Test('RealNumber("−5.1E10")', -5.1E10);
  Test('RealNumber("−5.1E100")', -5.1E100);
  Test('RealNumber("−5.1E−10")', -5.1E-10);
  Test('RealNumber("−5.1E−100")', -5.1E-100);

  Test('ComplexNumber("' +  '7 + 5 i'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5*i'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5⋅i'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5 j'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5*j'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5⋅j'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5 I'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5*I'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5⋅I'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5 J'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5*J'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5⋅J'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 - 5 i'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5*i'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5⋅i'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5 j'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5*j'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5⋅j'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5 I'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5*I'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5⋅I'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5 J'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5*J'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 - 5⋅J'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '-7 + 5 i'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5*i'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5⋅i'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5 j'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5*j'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5⋅j'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5 I'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5*I'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5⋅I'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5 J'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5*J'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 + 5⋅J'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '-7 - 5 i'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5*i'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5⋅i'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5 j'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5*j'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5⋅j'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5 I'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5*I'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5⋅I'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5 J'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5*J'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '-7 - 5⋅J'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '53.24 + 4.27 i'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27*i'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27⋅i'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 - 4.27 i'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 - 4.27*i'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 - 4.27⋅i'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27 i'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27*i'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27⋅i'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 - 4.27 i'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '-53.24 - 4.27*i'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '-53.24 - 4.27⋅i'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '53.24 + 4.27 j'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27*j'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27⋅j'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 - 4.27 j'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 - 4.27*j'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 - 4.27⋅j'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27 j'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27*j'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27⋅j'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 - 4.27 j'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '-53.24 - 4.27*j'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '-53.24 - 4.27⋅j'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '53.24 + 4.27 I'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27*I'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27⋅I'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 - 4.27 I'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 - 4.27*I'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 - 4.27⋅I'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27 I'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27*I'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27⋅I'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 - 4.27 I'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '-53.24 - 4.27*I'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '-53.24 - 4.27⋅I'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '53.24 + 4.27 J'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27*J'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27⋅J'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 - 4.27 J'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 - 4.27*J'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 - 4.27⋅J'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27 J'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27*J'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 + 4.27⋅J'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '-53.24 - 4.27 J'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '-53.24 - 4.27*J'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '-53.24 - 4.27⋅J'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '1 + 2 + 3 + 4 + 5 + 6 + i + j'                       + '")',  ASC(21, 2)                 );
  Test('ComplexNumber("' +  '1 + 2 + 3 + 4 + 5 + 6 + i + 10j'                     + '")',  ASC(21, 11)                 );
  Test('ComplexNumber("' +  '1.0 + 2. + 3e0 + 4.0e+0 + 5.e-0 + 6e00 + i + 10e0j'  + '")',  ASC(21, 11)                 );
  Eps; Test('ComplexNumber("' +  '1.001 + 5.321⋅i - 0.2 + i - 55*J'                    + '")',  ASC(0.801, -48.679)        );
  Test('ComplexNumber("' +  '0'                                                   + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0.0'                                                 + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-0'                                                  + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-0.0'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e0'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e-0'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e-10'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e+10'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e+10j'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-.0e0'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-.0e-0'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-.0e-10'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-.0e+10'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-.0e+10j'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '+.0e+10'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '+.0e+10j'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '1'                                                   + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0'                                                 + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0E0'                                               + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0e0'                                               + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0E+0'                                              + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0e+0'                                              + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0E-0'                                              + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0e-0'                                              + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '-1'                                                  + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '-1.0'                                                + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '-1.0E0'                                              + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '-1.0e0'                                              + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '-1.0E+0'                                             + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '-1.0e+0'                                             + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '-1.0E-0'                                             + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '-1.0e-0'                                             + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '26'                                                  + '")',  ASC(26, 0)                 );
  Test('ComplexNumber("' +  '-2'                                                  + '")',  ASC(-2, 0)                 );
  Test('ComplexNumber("' +  '3.9'                                                 + '")',  ASC(3.9, 0)                );
  Test('ComplexNumber("' +  '-5.10'                                               + '")',  ASC(-5.1, 0)               );
  Test('ComplexNumber("' +  '.8'                                                  + '")',  ASC(0.8, 0)                );
  Test('ComplexNumber("' +  '-.7'                                                 + '")',  ASC(-0.7, 0)               );
  Test('ComplexNumber("' +  '5.'                                                  + '")',  ASC(5, 0)                  );
  Test('ComplexNumber("' +  '-99.'                                                + '")',  ASC(-99, 0)                );
  Test('ComplexNumber("' +  '5.3E3'                                               + '")',  ASC(5300, 0)               );
  Test('ComplexNumber("' +  '-10e-1'                                              + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '1E6'                                                 + '")',  ASC(1000000, 0)            );
  Test('ComplexNumber("' +  '-1.2E-6'                                             + '")',  ASC(-0.0000012, 0)         );
  Test('ComplexNumber("' +  '0i'                                                  + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0i'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000 i '                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-0-0.0i'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0-.0i'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0.i'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  'i'                                                   + '")',  ASC(0, 1)                  );
  Test('ComplexNumber("' +  '-i'                                                  + '")',  ASC(0, -1)                 );
  Test('ComplexNumber("' +  '7i'                                                  + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '-55i'                                                + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3i'                                                + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '-99.2i'                                              + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5i'                                                + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '.5i'                                                 + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '-.9i'                                                + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '1+i'                                                 + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1-i'                                                 + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '-7-8i'                                               + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '1+2i'                                                + '")',  ASC(1, 2)                  );
  Test('ComplexNumber("' +  '2-9.2i'                                              + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92i'                                              + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-55.i'                                              + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2-55.2i'                                           + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800i'                                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4-0.2E2 i'                                         + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '1 + i'                                               + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1 - i'                                               + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '-7 - 8i'                                             + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 - 9.2 i'                                           + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92 i'                                           + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 - 55. i'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 - 55.2 i'                                        + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800i'                                          + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 - 0.2E2i'                                        + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '  1  +  i'                                           + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  ' 1 - i  '                                            + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '-7   - 8i '                                          + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  -  9.2 i'                                         + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92  i'                                      + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-  55.i'                                            + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  -  55.2i'                                      + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800   i   '                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  -    0.2E2 i'                                   + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '0*i'                                                 + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0*i'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000 *i '                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-0-0.0*i'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0-.0*i'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0.*i'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '7*i'                                                 + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '-55*i'                                               + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3*i'                                               + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '-99.2*i'                                             + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5*i'                                               + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '-.9*i'                                               + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '-7-8*i'                                              + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2-9.2*i'                                             + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92*i'                                             + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-55.*i'                                             + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2-55.2*i'                                          + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800*i'                                           + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4-0.2E2 *i'                                        + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '-7 - 8*i'                                            + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 - 9.2 *i'                                          + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92 *i'                                          + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 - 55. *i'                                          + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 - 55.2 *i'                                       + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800*i'                                         + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 - 0.2E2*i'                                       + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '-7   - 8*i '                                         + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  -  9.2 *i'                                        + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92  *i'                                     + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-  55.*i'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  -  55.2*i'                                     + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800   *i   '                           + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  -    0.2E2 *i'                                  + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '0I'                                                  + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0I'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000 I '                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-0-0.0I'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0-.0I'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0.I'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  'I'                                                   + '")',  ASC(0, 1)                  );
  Test('ComplexNumber("' +  '-I'                                                  + '")',  ASC(0, -1)                 );
  Test('ComplexNumber("' +  '7I'                                                  + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '-55I'                                                + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3I'                                                + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '-99.2I'                                              + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5I'                                                + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '-.9I'                                                + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '1+I'                                                 + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1-I'                                                 + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '-7-8I'                                               + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2-9.2I'                                              + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92I'                                              + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-55.I'                                              + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2-55.2I'                                           + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800I'                                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4-0.2E2 I'                                         + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '1 + I'                                               + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1 - I'                                               + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '-7 - 8I'                                             + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 - 9.2 I'                                           + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92 I'                                           + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 - 55. I'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 - 55.2 I'                                        + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800I'                                          + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 - 0.2E2I'                                        + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '  1  +  I'                                           + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  ' 1 - I  '                                            + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '-7   - 8I '                                          + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  -  9.2 I'                                         + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92  I'                                      + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-  55.I'                                            + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  -  55.2I'                                      + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800   I   '                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  -    0.2E2 I'                                   + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '0 ⋅ I'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0 ⋅ I'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000  ⋅ I '                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-0-0.0 ⋅ I'                                          + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0-.0 ⋅ I'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0. ⋅ I'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '7 ⋅ I'                                               + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '-55 ⋅ I'                                             + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3 ⋅ I'                                             + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '-99.2 ⋅ I'                                           + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5 ⋅ I'                                             + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '-.9 ⋅ I'                                             + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '-7-8 ⋅ I'                                            + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2-9.2 ⋅ I'                                           + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92 ⋅ I'                                           + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-55. ⋅ I'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2-55.2 ⋅ I'                                        + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800 ⋅ I'                                         + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4-0.2E2  ⋅ I'                                      + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '-7 - 8 ⋅ I'                                          + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 - 9.2  ⋅ I'                                        + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92  ⋅ I'                                        + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 - 55.  ⋅ I'                                        + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 - 55.2  ⋅ I'                                     + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800 ⋅ I'                                       + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 - 0.2E2 ⋅ I'                                     + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '-7   - 8 ⋅ I '                                       + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  -  9.2  ⋅ I'                                      + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92   ⋅ I'                                   + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-  55. ⋅ I'                                         + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  -  55.2 ⋅ I'                                   + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800    ⋅ I   '                         + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  -    0.2E2  ⋅ I'                                + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '0j'                                                  + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0j'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000 j '                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '-0-0.0j'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0-.0j'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0.j'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  'j'                                                   + '")',  ASC(0, 1)                  );
  Test('ComplexNumber("' +  '-j'                                                  + '")',  ASC(0, -1)                 );
  Test('ComplexNumber("' +  '7j'                                                  + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '-55j'                                                + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3j'                                                + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '-99.2j'                                              + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5j'                                                + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '-.9j'                                                + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '1+j'                                                 + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1-j'                                                 + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '-7-8j'                                               + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2-9.2j'                                              + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92j'                                              + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-55.j'                                              + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2-55.2j'                                           + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800j'                                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4-0.2E2 j'                                         + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '1 + j'                                               + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1 - j'                                               + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '-7 - 8j'                                             + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 - 9.2 j'                                           + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92 j'                                           + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 - 55. j'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 - 55.2 j'                                        + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800j'                                          + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 - 0.2E2j'                                        + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '  1  +  j'                                           + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  ' 1 - j  '                                            + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '-7   - 8j '                                          + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  -  9.2 j'                                         + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92  j'                                      + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0-  55.j'                                            + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  -  55.2j'                                      + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800   j   '                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  -    0.2E2 j'                                   + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '1+1'                                                 + '")',  ASC(2, 0)                  );
  Eps; Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10 - 2 - 2 - 1'       + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10i - 2 - 2 - 1'      + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10j - 2 - 2 - 1'      + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10I - 2 - 2 - 1'      + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10J - 2 - 2 - 1'      + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10⋅i - 2 - 2 - 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10⋅j - 2 - 2 - 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10⋅I - 2 - 2 - 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10⋅J - 2 - 2 - 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10*i - 2 - 2 - 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10*j - 2 - 2 - 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10*I - 2 - 2 - 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10*J - 2 - 2 - 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10 ⋅ i - 2 - 2 - 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10 ⋅ j - 2 - 2 - 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10 ⋅ I - 2 - 2 - 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10 ⋅ J - 2 - 2 - 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10 * i - 2 - 2 - 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10 * j - 2 - 2 - 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10 * I - 2 - 2 - 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 - 0.4 - 0.4 - 10 * J - 2 - 2 - 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1+i-2-2i'                                            + '")',  ASC(-1, -1)                );
  Test('ComplexNumber("' +  '4.2+i-0.2i'                                          + '")',  ASC(4.2, 0.8)              );
  Test('ComplexNumber("' +  '5 - 3*i'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '-3*i + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 - 3*i'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3*i+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3 * i   +    5 '                                   + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- .3e+1*i+5.0E-0'                                    + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '5 - 3⋅i'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '-3⋅i + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 - 3⋅i'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3⋅i+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3 ⋅ i   +    5 '                                   + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- .3e+1⋅i+5.0E-0'                                    + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '5 - 3⋅j'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '-3⋅j + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 - 3⋅j'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3⋅j+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3 ⋅ j   +    5 '                                   + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- .3e+1⋅j+5.0E-0'                                    + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '5 - 3*i'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '-3*i + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 - 3*i'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3*i+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3 * i   +    5 '                                   + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- .3e+1*i+5.0E-0'                                    + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '5 - 3*I'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '-3*J + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 - 3⋅i'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3⋅J+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- 3 J   +    5 '                                     + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '- .3e+1⋅I+5.E+0'                                     + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  'i - 1'                                               + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  'i - 1.0'                                             + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '+1i - 1e0'                                           + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 * i - 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 ⋅ i - 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 ⋅ j - 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 ⋅ I - 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 ⋅ J - 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 i - 1'                                           + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 I - 1'                                           + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 j - 1'                                           + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1. J - 1'                                            + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '7+3.22i'                                             + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22i'                                           + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22*i'                                          + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22 * i'                                        + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22i'                                           + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22⋅i'                                          + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22 ⋅ i'                                        + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7+3.22j'                                             + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22j'                                           + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22*j'                                          + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22 * j'                                        + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22j'                                           + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22⋅j'                                          + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22 ⋅ j'                                        + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7+3.22I'                                             + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '+7+3.22I'                                            + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7.e0 + 3.22E0j'                                      + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '1e1i'                                                + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '+1.0e+1*i'                                           + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '+1.0e+1*i-0.0e0'                                     + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '+1.0e+1⋅J-0.0E+0'                                    + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '+1.0000000000e+1*i-0.00000000000e00000'              + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '5.325632e22 + .25587i'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '5.325632e22 - .25587i'                               + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '.25587i + 5.325632e22'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '-.25587i + 5.325632e22'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '.25587i+5.325632e22'                                 + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '0.25587i+5.325632e22'                                + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '+.25587i+5.325632e22'                                + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '-.25587i+5.325632e22'                                + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '+0.25587i+5.325632e22'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '-0.25587i+5.325632e22'                               + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '5.325632e22 + .25587*I'                              + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '5.325632e22 - .25587*I'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '5.325632e22 - .25587*I'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '5.325632e22 - .25587*I'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '.25587*I + 5.325632e22'                              + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  #9'-.25587*I + 5.325632e22'                           + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '.25587*I+5.325632e22'                                + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '0.25587*I+5.325632e22'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '+.25587*I+5.325632e22'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '-.25587*I+5.325632e22'                               + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '+0.25587*I+5.325632e22'                              + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '-0.25587*I+5.325632e22'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '-0.25587*I+5.325632e22 - 2i'                         + '")',  ASC(5.325632e22, -2.25587) );
  Test('ComplexNumber("' +  '-0.25587*I+5.325632e22 - 2i + 0.1e-100'              + '")',  ASC(5.325632e22, -2.25587) );
  Test('ComplexNumber("' +  '-0.25587*I+5.325632e22 - 2i + 0.1e-100'              + '")',  ASC(5.325632e22, -2.25587) );
  Test('ComplexNumber("' +  #13#10#9'-0.25587*I+5.325632e22 -2i'                  + '")',  ASC(5.325632e22, -2.25587) );
  Test('ComplexNumber("' +  '10'                                                  + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  ' 10 '                                                + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  '  10  '                                              + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  '  10  '                                              + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  '10'                                                  + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  #9'10'                                                + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  #13#10'10'                                            + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  #13#10'i '                                            + '")',  ASC(0, 1)                  );
  Test('ComplexNumber("' +  #9'10'                                                + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  '10+i'                                                + '")',  ASC(10, 1)                 );
  Test('ComplexNumber("' +  '11+i'                                                + '")',  ASC(11, 1)                 );
  Test('ComplexNumber("' +  '11i'                                                 + '")',  ASC(0, 11)                 );
  Test('ComplexNumber("' +  '-5i + 3.1415'                                        + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '-5⋅i + 3.1415'                                       + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '-5⋅j + 3.1415'                                       + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '-5⋅I + 3.1415'                                       + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '-5⋅J + 3.1415'                                       + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '-5.0e0*I + 0.31415E1'                                + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '1e6 * i'                                             + '")',  ASC(0, 1E6)                );
  Test('ComplexNumber("' +  '-1e6 * i'                                            + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '1e-6 * i'                                            + '")',  ASC(0, 1E-6)               );
  Test('ComplexNumber("' +  '-1e-6 * i'                                           + '")',  ASC(0, -1E-6)              );
  Test('ComplexNumber("' +  '-1.0e6 * i'                                          + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '-1.0e6 * i + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '1e6 ⋅ i'                                             + '")',  ASC(0, 1E6)                );
  Test('ComplexNumber("' +  '-1e6 ⋅ i'                                            + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '1e-6 ⋅ i'                                            + '")',  ASC(0, 1E-6)               );
  Test('ComplexNumber("' +  '-1e-6 ⋅ i'                                           + '")',  ASC(0, -1E-6)              );
  Test('ComplexNumber("' +  '-1.0e6 ⋅ i'                                          + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '-1.0e6 ⋅ i + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '-1.0e6 ⋅ I + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '-1.0e6 ⋅ j + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '-1.0e6 ⋅ J + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '-1.0e6 ⋅ i'                                          + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '-1.0e6   i + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '-1.0e6   I + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '-1.0e6   j + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Eps; Test('ComplexNumber("' +  '-1.0e6   J + 78.90000000000000000000000000000000'    + '")',  ASC(78.9, -1E6)            );

  Test('ComplexNumber("' +  '7 + 5 i'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5*i'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5⋅i'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5 j'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5*j'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5⋅j'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5 I'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5*I'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5⋅I'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5 J'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5*J'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 + 5⋅J'                                             + '")',  ASC(7, 5)                  );
  Test('ComplexNumber("' +  '7 − 5 i'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5*i'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5⋅i'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5 j'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5*j'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5⋅j'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5 I'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5*I'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5⋅I'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5 J'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5*J'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '7 − 5⋅J'                                             + '")',  ASC(7, -5)                 );
  Test('ComplexNumber("' +  '−7 + 5 i'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5*i'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5⋅i'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5 j'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5*j'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5⋅j'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5 I'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5*I'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5⋅I'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5 J'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5*J'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 + 5⋅J'                                            + '")',  ASC(-7, 5)                 );
  Test('ComplexNumber("' +  '−7 − 5 i'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5*i'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5⋅i'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5 j'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5*j'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5⋅j'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5 I'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5*I'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5⋅I'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5 J'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5*J'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '−7 − 5⋅J'                                            + '")',  ASC(-7, -5)                );
  Test('ComplexNumber("' +  '53.24 + 4.27 i'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27*i'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27⋅i'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 − 4.27 i'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 − 4.27*i'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 − 4.27⋅i'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27 i'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27*i'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27⋅i'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 − 4.27 i'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '−53.24 − 4.27*i'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '−53.24 − 4.27⋅i'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '53.24 + 4.27 j'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27*j'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27⋅j'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 − 4.27 j'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 − 4.27*j'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 − 4.27⋅j'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27 j'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27*j'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27⋅j'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 − 4.27 j'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '−53.24 − 4.27*j'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '−53.24 − 4.27⋅j'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '53.24 + 4.27 I'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27*I'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27⋅I'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 − 4.27 I'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 − 4.27*I'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 − 4.27⋅I'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27 I'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27*I'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27⋅I'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 − 4.27 I'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '−53.24 − 4.27*I'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '−53.24 − 4.27⋅I'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '53.24 + 4.27 J'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27*J'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 + 4.27⋅J'                                      + '")',  ASC(53.24, 4.27)           );
  Test('ComplexNumber("' +  '53.24 − 4.27 J'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 − 4.27*J'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '53.24 − 4.27⋅J'                                      + '")',  ASC(53.24, -4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27 J'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27*J'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 + 4.27⋅J'                                     + '")',  ASC(-53.24, 4.27)          );
  Test('ComplexNumber("' +  '−53.24 − 4.27 J'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '−53.24 − 4.27*J'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '−53.24 − 4.27⋅J'                                     + '")',  ASC(-53.24, -4.27)         );
  Test('ComplexNumber("' +  '1 + 2 + 3 + 4 + 5 + 6 + i + j'                       + '")',  ASC(21, 2)                 );
  Test('ComplexNumber("' +  '1 + 2 + 3 + 4 + 5 + 6 + i + 10j'                     + '")',  ASC(21, 11)                 );
  Test('ComplexNumber("' +  '1.0 + 2. + 3e0 + 4.0e+0 + 5.e−0 + 6e00 + i + 10e0j'  + '")',  ASC(21, 11)                 );
  Eps; Test('ComplexNumber("' +  '1.001 + 5.321⋅i − 0.2 + i − 55*J'                    + '")',  ASC(0.801, -48.679)        );
  Test('ComplexNumber("' +  '0'                                                   + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0.0'                                                 + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−0'                                                  + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−0.0'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e0'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e−0'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e−10'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e+10'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '.0e+10j'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−.0e0'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−.0e−0'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−.0e−10'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−.0e+10'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−.0e+10j'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '+.0e+10'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '+.0e+10j'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '1'                                                   + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0'                                                 + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0E0'                                               + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0e0'                                               + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0E+0'                                              + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0e+0'                                              + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0E−0'                                              + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1.0e−0'                                              + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '−1'                                                  + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '−1.0'                                                + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '−1.0E0'                                              + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '−1.0e0'                                              + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '−1.0E+0'                                             + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '−1.0e+0'                                             + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '−1.0E−0'                                             + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '−1.0e−0'                                             + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '26'                                                  + '")',  ASC(26, 0)                 );
  Test('ComplexNumber("' +  '−2'                                                  + '")',  ASC(-2, 0)                 );
  Test('ComplexNumber("' +  '3.9'                                                 + '")',  ASC(3.9, 0)                );
  Test('ComplexNumber("' +  '−5.10'                                               + '")',  ASC(-5.1, 0)               );
  Test('ComplexNumber("' +  '.8'                                                  + '")',  ASC(0.8, 0)                );
  Test('ComplexNumber("' +  '−.7'                                                 + '")',  ASC(-0.7, 0)               );
  Test('ComplexNumber("' +  '5.'                                                  + '")',  ASC(5, 0)                  );
  Test('ComplexNumber("' +  '−99.'                                                + '")',  ASC(-99, 0)                );
  Test('ComplexNumber("' +  '5.3E3'                                               + '")',  ASC(5300, 0)               );
  Test('ComplexNumber("' +  '−10e−1'                                              + '")',  ASC(-1, 0)                 );
  Test('ComplexNumber("' +  '1E6'                                                 + '")',  ASC(1000000, 0)            );
  Test('ComplexNumber("' +  '−1.2E−6'                                             + '")',  ASC(-0.0000012, 0)         );
  Test('ComplexNumber("' +  '0i'                                                  + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0i'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000 i '                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−0−0.0i'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0−.0i'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0.i'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  'i'                                                   + '")',  ASC(0, 1)                  );
  Test('ComplexNumber("' +  '−i'                                                  + '")',  ASC(0, -1)                 );
  Test('ComplexNumber("' +  '7i'                                                  + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '−55i'                                                + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3i'                                                + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '−99.2i'                                              + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5i'                                                + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '.5i'                                                 + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '−.9i'                                                + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '1+i'                                                 + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1−i'                                                 + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '−7−8i'                                               + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '1+2i'                                                + '")',  ASC(1, 2)                  );
  Test('ComplexNumber("' +  '2−9.2i'                                              + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92i'                                              + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−55.i'                                              + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2−55.2i'                                           + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800i'                                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4−0.2E2 i'                                         + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '1 + i'                                               + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1 − i'                                               + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '−7 − 8i'                                             + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 − 9.2 i'                                           + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92 i'                                           + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 − 55. i'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 − 55.2 i'                                        + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800i'                                          + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 − 0.2E2i'                                        + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '  1  +  i'                                           + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  ' 1 − i  '                                            + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '−7   − 8i '                                          + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  −  9.2 i'                                         + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92  i'                                      + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−  55.i'                                            + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  −  55.2i'                                      + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800   i   '                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  −    0.2E2 i'                                   + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '0*i'                                                 + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0*i'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000 *i '                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−0−0.0*i'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0−.0*i'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0.*i'                                              + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '7*i'                                                 + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '−55*i'                                               + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3*i'                                               + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '−99.2*i'                                             + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5*i'                                               + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '−.9*i'                                               + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '−7−8*i'                                              + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2−9.2*i'                                             + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92*i'                                             + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−55.*i'                                             + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2−55.2*i'                                          + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800*i'                                           + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4−0.2E2 *i'                                        + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '−7 − 8*i'                                            + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 − 9.2 *i'                                          + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92 *i'                                          + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 − 55. *i'                                          + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 − 55.2 *i'                                       + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800*i'                                         + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 − 0.2E2*i'                                       + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '−7   − 8*i '                                         + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  −  9.2 *i'                                        + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92  *i'                                     + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−  55.*i'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  −  55.2*i'                                     + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800   *i   '                           + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  −    0.2E2 *i'                                  + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '0I'                                                  + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0I'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000 I '                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−0−0.0I'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0−.0I'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0.I'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  'I'                                                   + '")',  ASC(0, 1)                  );
  Test('ComplexNumber("' +  '−I'                                                  + '")',  ASC(0, -1)                 );
  Test('ComplexNumber("' +  '7I'                                                  + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '−55I'                                                + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3I'                                                + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '−99.2I'                                              + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5I'                                                + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '−.9I'                                                + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '1+I'                                                 + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1−I'                                                 + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '−7−8I'                                               + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2−9.2I'                                              + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92I'                                              + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−55.I'                                              + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2−55.2I'                                           + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800I'                                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4−0.2E2 I'                                         + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '1 + I'                                               + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1 − I'                                               + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '−7 − 8I'                                             + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 − 9.2 I'                                           + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92 I'                                           + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 − 55. I'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 − 55.2 I'                                        + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800I'                                          + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 − 0.2E2I'                                        + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '  1  +  I'                                           + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  ' 1 − I  '                                            + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '−7   − 8I '                                          + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  −  9.2 I'                                         + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92  I'                                      + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−  55.I'                                            + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  −  55.2I'                                      + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800   I   '                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  −    0.2E2 I'                                   + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '0 ⋅ I'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0 ⋅ I'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000  ⋅ I '                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−0−0.0 ⋅ I'                                          + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0−.0 ⋅ I'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0. ⋅ I'                                            + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '7 ⋅ I'                                               + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '−55 ⋅ I'                                             + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3 ⋅ I'                                             + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '−99.2 ⋅ I'                                           + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5 ⋅ I'                                             + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '−.9 ⋅ I'                                             + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '−7−8 ⋅ I'                                            + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2−9.2 ⋅ I'                                           + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92 ⋅ I'                                           + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−55. ⋅ I'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2−55.2 ⋅ I'                                        + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800 ⋅ I'                                         + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4−0.2E2  ⋅ I'                                      + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '−7 − 8 ⋅ I'                                          + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 − 9.2  ⋅ I'                                        + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92  ⋅ I'                                        + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 − 55.  ⋅ I'                                        + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 − 55.2  ⋅ I'                                     + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800 ⋅ I'                                       + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 − 0.2E2 ⋅ I'                                     + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '−7   − 8 ⋅ I '                                       + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  −  9.2  ⋅ I'                                      + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92   ⋅ I'                                   + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−  55. ⋅ I'                                         + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  −  55.2 ⋅ I'                                   + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800    ⋅ I   '                         + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  −    0.2E2  ⋅ I'                                + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '0j'                                                  + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0j'                                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  ' 0.0000E0 + 0.000 j '                                + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '−0−0.0j'                                             + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0−.0j'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  '0+0.j'                                               + '")',  ASC(0, 0)                  );
  Test('ComplexNumber("' +  'j'                                                   + '")',  ASC(0, 1)                  );
  Test('ComplexNumber("' +  '−j'                                                  + '")',  ASC(0, -1)                 );
  Test('ComplexNumber("' +  '7j'                                                  + '")',  ASC(0, 7)                  );
  Test('ComplexNumber("' +  '−55j'                                                + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '4.3j'                                                + '")',  ASC(0, 4.3)                );
  Test('ComplexNumber("' +  '−99.2j'                                              + '")',  ASC(0, -99.2)              );
  Test('ComplexNumber("' +  '0.5j'                                                + '")',  ASC(0, 0.5)                );
  Test('ComplexNumber("' +  '−.9j'                                                + '")',  ASC(0, -0.9)               );
  Test('ComplexNumber("' +  '1+j'                                                 + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1−j'                                                 + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '−7−8j'                                               + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2−9.2j'                                              + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0+.92j'                                              + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−55.j'                                              + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2−55.2j'                                           + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100+800j'                                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4−0.2E2 j'                                         + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '1 + j'                                               + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  '1 − j'                                               + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '−7 − 8j'                                             + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2 − 9.2 j'                                           + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '0 + .92 j'                                           + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0 − 55. j'                                           + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  '0.2 − 55.2 j'                                        + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '100 + 800j'                                          + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4 − 0.2E2j'                                        + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '  1  +  j'                                           + '")',  ASC(1, 1)                  );
  Test('ComplexNumber("' +  ' 1 − j  '                                            + '")',  ASC(1, -1)                 );
  Test('ComplexNumber("' +  '−7   − 8j '                                          + '")',  ASC(-7, -8)                );
  Test('ComplexNumber("' +  '2  −  9.2 j'                                         + '")',  ASC(2, -9.2)               );
  Test('ComplexNumber("' +  '   0   +.92  j'                                      + '")',  ASC(0, 0.92)               );
  Test('ComplexNumber("' +  '0−  55.j'                                            + '")',  ASC(0, -55)                );
  Test('ComplexNumber("' +  ' 0.2  −  55.2j'                                      + '")',  ASC(0.2, -55.2)            );
  Test('ComplexNumber("' +  '    100   +   800   j   '                            + '")',  ASC(100, 800)              );
  Test('ComplexNumber("' +  '2E4  −    0.2E2 j'                                   + '")',  ASC(20000, -20)            );
  Test('ComplexNumber("' +  '1+1'                                                 + '")',  ASC(2, 0)                  );
  Eps; Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10 − 2 − 2 − 1'       + '")',  ASC(1, 0)                  );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10i − 2 − 2 − 1'      + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10j − 2 − 2 − 1'      + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10I − 2 − 2 − 1'      + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10J − 2 − 2 − 1'      + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10⋅i − 2 − 2 − 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10⋅j − 2 − 2 − 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10⋅I − 2 − 2 − 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10⋅J − 2 − 2 − 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10*i − 2 − 2 − 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10*j − 2 − 2 − 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10*I − 2 − 2 − 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10*J − 2 − 2 − 1'     + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10 ⋅ i − 2 − 2 − 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10 ⋅ j − 2 − 2 − 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10 ⋅ I − 2 − 2 − 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10 ⋅ J − 2 − 2 − 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10 * i − 2 − 2 − 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10 * j − 2 − 2 − 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10 * I − 2 − 2 − 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1 + 5 + 0.8 + 10 − 0.4 − 0.4 − 10 * J − 2 − 2 − 1'   + '")',  ASC(11, -10)               );
  Test('ComplexNumber("' +  '1+i−2−2i'                                            + '")',  ASC(-1, -1)                );
  Test('ComplexNumber("' +  '4.2+i−0.2i'                                          + '")',  ASC(4.2, 0.8)              );
  Test('ComplexNumber("' +  '5 − 3*i'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '−3*i + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 − 3*i'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3*i+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3 * i   +    5 '                                   + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− .3e+1*i+5.0E−0'                                    + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '5 − 3⋅i'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '−3⋅i + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 − 3⋅i'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3⋅i+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3 ⋅ i   +    5 '                                   + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− .3e+1⋅i+5.0E−0'                                    + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '5 − 3⋅j'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '−3⋅j + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 − 3⋅j'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3⋅j+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3 ⋅ j   +    5 '                                   + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− .3e+1⋅j+5.0E−0'                                    + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '5 − 3*i'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '−3*i + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 − 3*i'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3*i+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3 * i   +    5 '                                   + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− .3e+1*i+5.0E−0'                                    + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '5 − 3*I'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '−3*J + 5'                                            + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '+ 5 − 3⋅i'                                           + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3⋅J+5'                                             + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− 3 J   +    5 '                                     + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  '− .3e+1⋅I+5.E+0'                                     + '")',  ASC(5, -3)                 );
  Test('ComplexNumber("' +  'i − 1'                                               + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  'i − 1.0'                                             + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '+1i − 1e0'                                           + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 * i − 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 ⋅ i − 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 ⋅ j − 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 ⋅ I − 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 ⋅ J − 1'                                         + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 i − 1'                                           + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 I − 1'                                           + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1.0 j − 1'                                           + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '1. J − 1'                                            + '")',  ASC(-1, 1)                 );
  Test('ComplexNumber("' +  '7+3.22i'                                             + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22i'                                           + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22*i'                                          + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22 * i'                                        + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22i'                                           + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22⋅i'                                          + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22 ⋅ i'                                        + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7+3.22j'                                             + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22j'                                           + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22*j'                                          + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22 * j'                                        + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22j'                                           + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22⋅j'                                          + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7 + 3.22 ⋅ j'                                        + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7+3.22I'                                             + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '+7+3.22I'                                            + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '7.e0 + 3.22E0j'                                      + '")',  ASC(7, 3.22)               );
  Test('ComplexNumber("' +  '1e1i'                                                + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '+1.0e+1*i'                                           + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '+1.0e+1*i−0.0e0'                                     + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '+1.0e+1⋅J−0.0E+0'                                    + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '+1.0000000000e+1*i−0.00000000000e00000'              + '")',  ASC(0, 10)                 );
  Test('ComplexNumber("' +  '5.325632e22 + .25587i'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '5.325632e22 − .25587i'                               + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '.25587i + 5.325632e22'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '−.25587i + 5.325632e22'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '.25587i+5.325632e22'                                 + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '0.25587i+5.325632e22'                                + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '+.25587i+5.325632e22'                                + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '−.25587i+5.325632e22'                                + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '+0.25587i+5.325632e22'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '−0.25587i+5.325632e22'                               + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '5.325632e22 + .25587*I'                              + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '5.325632e22 − .25587*I'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '5.325632e22 − .25587*I'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '5.325632e22 − .25587*I'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '.25587*I + 5.325632e22'                              + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  #9'−.25587*I + 5.325632e22'                           + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '.25587*I+5.325632e22'                                + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '0.25587*I+5.325632e22'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '+.25587*I+5.325632e22'                               + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '−.25587*I+5.325632e22'                               + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '+0.25587*I+5.325632e22'                              + '")',  ASC(5.325632e22, 0.25587)  );
  Test('ComplexNumber("' +  '−0.25587*I+5.325632e22'                              + '")',  ASC(5.325632e22, -0.25587) );
  Test('ComplexNumber("' +  '−0.25587*I+5.325632e22 − 2i'                         + '")',  ASC(5.325632e22, -2.25587) );
  Test('ComplexNumber("' +  '−0.25587*I+5.325632e22 − 2i + 0.1e−100'              + '")',  ASC(5.325632e22, -2.25587) );
  Test('ComplexNumber("' +  '−0.25587*I+5.325632e22 − 2i + 0.1e−100'              + '")',  ASC(5.325632e22, -2.25587) );
  Test('ComplexNumber("' +  #13#10#9'−0.25587*I+5.325632e22 −2i'                  + '")',  ASC(5.325632e22, -2.25587) );
  Test('ComplexNumber("' +  '10'                                                  + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  ' 10 '                                                + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  '  10  '                                              + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  '  10  '                                              + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  '10'                                                  + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  #9'10'                                                + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  #13#10'10'                                            + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  #13#10'i '                                            + '")',  ASC(0, 1)                  );
  Test('ComplexNumber("' +  #9'10'                                                + '")',  ASC(10, 0)                 );
  Test('ComplexNumber("' +  '10+i'                                                + '")',  ASC(10, 1)                 );
  Test('ComplexNumber("' +  '11+i'                                                + '")',  ASC(11, 1)                 );
  Test('ComplexNumber("' +  '11i'                                                 + '")',  ASC(0, 11)                 );
  Test('ComplexNumber("' +  '−5i + 3.1415'                                        + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '−5⋅i + 3.1415'                                       + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '−5⋅j + 3.1415'                                       + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '−5⋅I + 3.1415'                                       + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '−5⋅J + 3.1415'                                       + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '−5.0e0*I + 0.31415E1'                                + '")',  ASC(3.1415, -5)            );
  Test('ComplexNumber("' +  '1e6 * i'                                             + '")',  ASC(0, 1E6)                );
  Test('ComplexNumber("' +  '−1e6 * i'                                            + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '1e−6 * i'                                            + '")',  ASC(0, 1E-6)               );
  Test('ComplexNumber("' +  '−1e−6 * i'                                           + '")',  ASC(0, -1E-6)              );
  Test('ComplexNumber("' +  '−1.0e6 * i'                                          + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '−1.0e6 * i + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '1e6 ⋅ i'                                             + '")',  ASC(0, 1E6)                );
  Test('ComplexNumber("' +  '−1e6 ⋅ i'                                            + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '1e−6 ⋅ i'                                            + '")',  ASC(0, 1E-6)               );
  Test('ComplexNumber("' +  '−1e−6 ⋅ i'                                           + '")',  ASC(0, -1E-6)              );
  Test('ComplexNumber("' +  '−1.0e6 ⋅ i'                                          + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '−1.0e6 ⋅ i + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '−1.0e6 ⋅ I + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '−1.0e6 ⋅ j + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '−1.0e6 ⋅ J + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '−1.0e6 ⋅ i'                                          + '")',  ASC(0, -1E6)               );
  Test('ComplexNumber("' +  '−1.0e6   i + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '−1.0e6   I + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Test('ComplexNumber("' +  '−1.0e6   j + 78.90'                                  + '")',  ASC(78.9, -1E6)            );
  Eps; Test('ComplexNumber("' +  '−1.0e6   J + 78.90000000000000000000000000000000'    + '")',  ASC(78.9, -1E6)            );

  Test('ComplexNumber("' + #0                                                                + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + ''                                                                + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + ' '                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '  '                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '   '                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '     +     '                                                     + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + ' -- '                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5+'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5 +'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5+ '                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5 + '                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + ' 5 + '                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5+  '                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5 +  '                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5-'                                                              + '")', failure, 'Cannot convert string "5-" to a complex number.');
  Test('ComplexNumber("' + '5 -'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5- '                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5 - '                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + ' 5 - '                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5-  '                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5 -  '                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + ' 10 - '                                                          + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5+3*'                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5 + 3 * '                                                        + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5 + 3 *'                                                         + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5+3* '                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '5 - 3 * '                                                        + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-10-2⋅'                                                          + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-10 - 2⋅'                                                        + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-10 - 2 ⋅ '                                                      + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '*2'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '*'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '+'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '3+2i-'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-2-2i+'                                                          + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '2E'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-.9e'                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '1+e'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '1+6E'                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '2Ei'                                                             + '")', failure, 'Cannot convert string "2Ei" to a complex number.');
  Test('ComplexNumber("' + '-.3ei'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-2.4-.4E*i'                                                      + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '/'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '('                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '?'                                                               + '")', failure, 'Cannot convert string "?" to a complex number.');
  Test('ComplexNumber("' + ','                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '☃'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '⌬'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '∯'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '('                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + ')'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '()'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'E'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'e'                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '.E'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '.e'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '+E'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '+e'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '+.E'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '+.e'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-e'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-E'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-.e'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-.E'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'E+'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'e+'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'E-'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'e-'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'E+i'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'E+I'                                                             + '")', failure, 'Cannot convert string "E+I" to a complex number.');
  Test('ComplexNumber("' + 'e-i'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'e*i'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'I*e'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '+i*e'                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '+i*e+'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'e+i'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '.e+i'                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '.e.+i'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '+1.0000000000e+1.000000000*i-0.00000000000e00000'                + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-0.25587*I+5.325632e22 - 2i + 0.1f-100'                          + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'A'                                                               + '")', failure, 'Cannot convert string "A" to a complex number.');
  Test('ComplexNumber("' + 'A*I'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '1a'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '1+2*a'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '3..14'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '3..'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '..3'                                                             + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '3.1.4'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '+3.1.4'                                                          + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '-3.1.4'                                                          + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '3ee4'                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '3ee+4'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '3e--4'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '3e--4i'                                                          + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '4-3ei'                                                           + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'ie'                                                              + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + 'Microsoft Internet Explorer'                                     + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + #$222b                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + #8                                                                + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + #9                                                                + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + #27                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + #13#10                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + #13                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + #10                                                               + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '2+'#8                                                            + '")', failure, 'Cannot convert string "');
  Test('ComplexNumber("' + '10'#0                                                            + '")', failure, 'Cannot convert string "');

  Test('boolean("false")', False);
  Test('boolean("False")', False);
  Test('boolean(" False   ")', False);
  Test('boolean("FALSE")', False);
  Test('boolean(" FALSE ")', False);
  Test('boolean("true")', True);
  Test('boolean("True")', True);
  Test('boolean("TRUE")', True);
  Test('boolean("true ")', True);
  Test('boolean("  TRUE ")', True);
  Test('boolean("  TrUE")', True);
  Test('boolean("cat")', failure, 'String "cat" cannot be converted to a boolean.');
  Test('boolean("dog")', failure, 'String "dog" cannot be converted to a boolean.');
  Test('boolean("yes")', failure, 'String "yes" cannot be converted to a boolean.');
  Test('boolean("")', failure, 'String "" cannot be converted to a boolean.');
  Test('boolean("y")', failure, 'String "y" cannot be converted to a boolean.');
  Test('boolean("1")', failure, 'String "1" cannot be converted to a boolean.');
  Test('boolean("0")', failure, 'String "0" cannot be converted to a boolean.');

  Test('string("Sällan ses solen så storslaget skina, som stjärnan så småningom sken.")',
    'Sällan ses solen så storslaget skina, som stjärnan så småningom sken.');

  Test('string("")', '');

  Test('vector("1, 2, 3")', [1, 2, 3]);
  Test('vector("3.1415, 2.7182")', [3.1415, 2.7182]);
  Test('vector("  10,   20,   30, 40,50")', [10, 20, 30, 40, 50]);
  Test('vector("1; 2; 3")', [1, 2, 3]);
  Test('vector("3.1415; 2.7182")', [3.1415, 2.7182]);
  Test('vector("  10;   20;   30; 40;50")', [10, 20, 30, 40, 50]);
  Test('vector("1'#9'2'#9'3")', [1, 2, 3]);
  Test('vector("1'#13#10'2'#13#10'3")', [1, 2, 3]);
  Test('vector("1'#13#10'2'#13#10'3'#13#10'")', [1, 2, 3]);
  Test('vector("3.1415'#9'2.7182")', [3.1415, 2.7182]);
  Test('vector("  10'#9'   20'#9'   30'#9' 40'#9'50")', [10, 20, 30, 40, 50]);
  Test('vector("5, -3.8, 1.2E-10,0")', [5, -3.8, 1.2E-10, 0]);
  Test('vector("-1, -2, -3E-50, -4.0")', [-1, -2, -3E-50, -4]);
  Test('vector("5, -3.8, 1.2E−10,0")', [5, -3.8, 1.2E-10, 0]);
  Test('vector("−1, −2, −3E−50, −4.0")', [-1, -2, -3E-50, -4]);
  Test('vector("1, 2, 3, Vera Verto")', failure, 'Cannot convert string "Vera Verto" to a real number.');
  Test('vector("21")', [21]);
  Test('vector("")', TAlgosimRealVector.Create);
  Test('vector("1, 2, , 3")', failure, 'Cannot convert string "" to a real number.');

  Test('RealVector("1, 2, 3")', [1, 2, 3]);
  Test('RealVector("3.1415, 2.7182")', [3.1415, 2.7182]);
  Test('RealVector("  10,   20,   30, 40,50")', [10, 20, 30, 40, 50]);
  Test('RealVector("1; 2; 3")', [1, 2, 3]);
  Test('RealVector("3.1415; 2.7182")', [3.1415, 2.7182]);
  Test('RealVector("  10;   20;   30; 40;50")', [10, 20, 30, 40, 50]);
  Test('RealVector("1'#9'2'#9'3")', [1, 2, 3]);
  Test('RealVector("1'#13#10'2'#13#10'3")', [1, 2, 3]);
  Test('RealVector("1'#13#10'2'#13#10'3"'#13#10')', [1, 2, 3]);
  Test('RealVector("3.1415'#9'2.7182")', [3.1415, 2.7182]);
  Test('RealVector("  10'#9'   20'#9'   30'#9' 40'#9'50")', [10, 20, 30, 40, 50]);
  Test('RealVector("5, -3.8, 1.2E-10,0")', [5, -3.8, 1.2E-10, 0]);
  Test('RealVector("-1, -2, -3E-50, -4.0")', [-1, -2, -3E-50, -4]);
  Test('RealVector("5, -3.8, 1.2E−10,0")', [5, -3.8, 1.2E-10, 0]);
  Test('RealVector("−1, −2, −3E−50, −4.0")', [-1, -2, -3E-50, -4]);
  Test('RealVector("1, 2, 3, Vera Verto")', failure, 'Cannot convert string "Vera Verto" to a real number.');
  Test('RealVector("21")', [21]);
  Test('RealVector("")', TAlgosimRealVector.Create);
  Test('RealVector("1, 2, , 3")', failure, 'Cannot convert string "" to a real number.');

  Test('ComplexVector("1, 2, 3")', [TASC(1), 2, 3]);
  Test('ComplexVector("3.1415, 2.7182")', [TASC(3.1415), 2.7182]);
  Test('ComplexVector("  10,   20,   30, 40,50")', [TASC(10), 20, 30, 40, 50]);
  Test('ComplexVector("1; 2; 3")', [TASC(1), 2, 3]);
  Test('ComplexVector("3.1415; 2.7182")', [TASC(3.1415), 2.7182]);
  Test('ComplexVector("  10;   20;   30; 40;50")', [TASC(10), 20, 30, 40, 50]);
  Test('ComplexVector("1'#9'2'#9'3")', [TASC(1), 2, 3]);
  Test('ComplexVector("1'#13#10'2'#13#10'3")', [TASC(1), 2, 3]);
  Test('ComplexVector("1'#13#10'2'#13#10'3'#13#10'")', [TASC(1), 2, 3]);
  Test('ComplexVector("3.1415'#9'2.7182")', [TASC(3.1415), 2.7182]);
  Test('ComplexVector("  10'#9'   20'#9'   30'#9' 40'#9'50")', [TASC(10), 20, 30, 40, 50]);
  Test('ComplexVector("5, -3.8, 1.2E-10,0")', [TASC(5), -3.8, 1.2E-10, 0]);
  Test('ComplexVector("-1, -2, -3E-50, -4.0")', [TASC(-1), -2, -3E-50, -4]);
  Test('ComplexVector("5, -3.8, 1.2E−10,0")', [TASC(5), -3.8, 1.2E-10, 0]);
  Test('ComplexVector("−1, −2, −3E−50, −4.0")', [TASC(-1), -2, -3E-50, -4]);
  Test('ComplexVector("1, 2, 3, Vera Verto")', failure, 'Cannot convert string "Vera Verto" to a complex number.');
  Test('ComplexVector("21")', [TASC(21)]);
  Test('ComplexVector("")', TAlgosimComplexVector.Create);
  Test('ComplexVector("1, 2, , 3")', failure, 'Cannot convert string "" to a complex number.');

  Test('ComplexVector("1, 2, i, 3 + 4⋅i, -7, 5 − i, 2 + i + .5, .5 + j, 1.2E20 + 2.2E30 I")',
    [1, 2, ImaginaryUnit, 3 + 4*ImaginaryUnit, -7, 5 - ImaginaryUnit, 2.5 + ImaginaryUnit, 0.5 + ImaginaryUnit, 1.2E20 + 2.2E30*ImaginaryUnit]);

  _s :=
    '1, 2, -3, 4, 5'#13#10 +
    '3.1415, −10, 1.2E3, 5E−6, 6.'#13#10 +
    '.5, -.3, −.8E20, 3e3, .0'#13#10 +
    '7, 6; 5'#9'   4,    3.0e+0';

  Test('matrix("' + _s + '")', 5, [1, 2, -3, 4, 5, 3.1415, -10, 1200, 5E-6, 6, 0.5, -0.3, -0.8E20, 3000, 0, 7, 6, 5, 4, 3]);
  Test('RealMatrix("' + _s + '")', 5, [1, 2, -3, 4, 5, 3.1415, -10, 1200, 5E-6, 6, 0.5, -0.3, -0.8E20, 3000, 0, 7, 6, 5, 4, 3]);
  Test('ComplexMatrix("' + _s + '")', 5, [TASC(1), 2, -3, 4, 5, 3.1415, -10, 1200, 5E-6, 6, 0.5, -0.3, -0.8E20, 3000, 0, 7, 6, 5, 4, 3]);

  _s :=
    '1, 2, 3'#13#10 +
    '4, 5, 6';

  Test('matrix("' + _s + '")', 3, [1, 2, 3, 4, 5, 6]);
  Test('RealMatrix("' + _s + '")', 3, [1, 2, 3, 4, 5, 6]);
  Test('ComplexMatrix("' + _s + '")', 3, [TASC(1), 2, 3, 4, 5, 6]);

  _s :=
    '1, 2, 3'#13#10 +
    '4, 5';

  Test('matrix("' + _s + '")', failure, 'Attempt to create a non-rectangular matrix.');
  Test('RealMatrix("' + _s + '")', failure, 'Attempt to create a non-rectangular matrix.');
  Test('ComplexMatrix("' + _s + '")', failure, 'Attempt to create a non-rectangular matrix.');

  _s :=
    '1, 2, 3'#13#10 +
    '4, 5, 6'#13#10 +
    '7, 8, 9'#13#10 +
    '1, 0, 1';

  Test('matrix("' + _s + '")', 3, [1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 0, 1]);
  Test('RealMatrix("' + _s + '")', 3, [1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 0, 1]);
  Test('ComplexMatrix("' + _s + '")', 3, [TASC(1), 2, 3, 4, 5, 6, 7, 8, 9, 1, 0, 1]);

  _s :=
    '1, 2, 3';

  Test('matrix("' + _s + '")', 3, [1, 2, 3]);
  Test('RealMatrix("' + _s + '")', 3, [1, 2, 3]);
  Test('ComplexMatrix("' + _s + '")', 3, [TASC(1), 2, 3]);

  _s :=
    '1, 2, 3'#13#10 +
    '4, 5, 6'#13#10 +
    '7, nargle, 9'#13#10 +
    '1, 0, 1';

  Test('matrix("' + _s + '")', failure, 'Cannot convert string "nargle" to a real number.');
  Test('RealMatrix("' + _s + '")', failure, 'Cannot convert string "nargle" to a real number.');
  Test('ComplexMatrix("' + _s + '")', failure, 'Cannot convert string "nargle" to a complex number.');

  _s :=
    '1, 2, 3'#13#10 +
    '4, 5, 6'#13#10 +
    '7, , 9'#13#10 +
    '1, 0, 1';

  Test('matrix("' + _s + '")', failure, 'Cannot convert string "" to a real number.');
  Test('RealMatrix("' + _s + '")', failure, 'Cannot convert string "" to a real number.');
  Test('ComplexMatrix("' + _s + '")', failure, 'Cannot convert string "" to a complex number.');

  _s :=
    '1, 2, 3'#13#10 +
    '4, 5, 6'#13#10 +
    '7, 8,  '#13#10 +
    '1, 0, 1';

  Test('matrix("' + _s + '")', failure, 'Cannot convert string "" to a real number.');
  Test('RealMatrix("' + _s + '")', failure, 'Cannot convert string "" to a real number.');
  Test('ComplexMatrix("' + _s + '")', failure, 'Cannot convert string "" to a complex number.');

  _s :=
    'i, 2, 3'#13#10 +
    '4, 5, 6'#13#10 +
    '7, 8 + 3i, 9.5 - 2.1 i'#13#10 +
    '7−2⋅i, 5 - 3*i, 1 + 2 + i';

  Test('matrix("' + _s + '")', failure, 'Cannot convert string "i" to a real number.');
  Test('RealMatrix("' + _s + '")', failure, 'Cannot convert string "i" to a real number.');
  Test('ComplexMatrix("' + _s + '")', 3, [ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8 + 3*ImaginaryUnit, 9.5 - 2.1*ImaginaryUnit, 7 - 2*ImaginaryUnit, 5 - 3*ImaginaryUnit, 3 + ImaginaryUnit]);

  Test('ToList("the emotional range of a teaspoon")',
    TAlgosimArray.CreateWithValue('the emotional range of a teaspoon'.ToCharArray));

  Test('ToList("")', strarr([]));

  Test('ToSet("the emotional range of a teaspoon")', asoset(['t', 'h', 'e', ' ', 'e', 'm', 'o', 't', 'i', 'o', 'n', 'a', 'l', ' ', 'r', 'a', 'n', 'g', 'e', ' ', 'o', 'f', ' ', 'a', ' ', 't', 'e', 'a', 's', 'p', 'o', 'o', 'n']));

  Test('ToSet("")', asoset([]));

  Test('color("black")', ASO(TRGB.Create(0, 0, 0)));
  Test('color("red")', ASO(TRGB.Create(1, 0, 0)));
  Test('color("lime")', ASO(TRGB.Create(0, 1, 0)));
  Test('color("blue")', ASO(TRGB.Create(0, 0, 1)));
  Test('color("yellow")', ASO(TRGB.Create(1, 1, 0)));
  Test('color("white")', ASO(TRGB.Create(1, 1, 1)));
  Test('color("navy")', ASO(TRGB.Create(0, 0, $80/$FF)));
  Test('color("green")', ASO(TRGB.Create(0, $80/$FF, 0)));
  Test('color("gold")', ASO(TRGB.Create(1, $D7/$FF, 0)));

  Test('color("#000000")', ASO(TRGB.Create(0, 0, 0)));
  Test('color("#FF0000")', ASO(TRGB.Create(1, 0, 0)));
  Test('color("#00FF00")', ASO(TRGB.Create(0, 1, 0)));
  Test('color("#0000FF")', ASO(TRGB.Create(0, 0, 1)));
  Test('color("#FFFF00")', ASO(TRGB.Create(1, 1, 0)));
  Test('color("#FF00FF")', ASO(TRGB.Create(1, 0, 1)));
  Test('color("#00FFFF")', ASO(TRGB.Create(0, 1, 1)));
  Test('color("#FFFFFF")', ASO(TRGB.Create(1, 1, 1)));
  Test('color("#808080")', ASO(TRGB.Create($80/$FF, $80/$FF, $80/$FF)));


  //
  // From real vector
  //

  Test('number(DebugObject("empty vector"))', failure, 'An empty container cannot be converted to a number.');
  Test('number(❨1, 2, 3❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(❨21❩)', 21.0);
  Test('number(❨1/4❩)', 0.25);
  Test('number(❨123.456❩)', 123.456);
  Test('number(❨123.456E20❩)', 123.456E20);
  Test('number(❨123.456E-20❩)', 123.456E-20);

  Test('integer(DebugObject("empty vector"))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨1, 2, 3❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨21❩)', 21);
  Test('integer(❨1/4❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨123.456❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨123.456E20❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨123.456E-20❩)', 0);

  Test('RealNumber(DebugObject("empty vector"))', failure, 'An empty container cannot be converted to a number.');
  Test('RealNumber(❨1, 2, 3❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(❨21❩)', 21.0);
  Test('RealNumber(❨1/4❩)', 0.25);
  Test('RealNumber(❨123.456❩)', 123.456);
  Test('RealNumber(❨123.456E20❩)', 123.456E20);
  Test('RealNumber(❨123.456E-20❩)', 123.456E-20);

  Test('ComplexNumber(DebugObject("empty vector"))', failure, 'An empty container cannot be converted to a number.');
  Test('ComplexNumber(❨1, 2, 3❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(❨21❩)', TASC(21.0));
  Test('ComplexNumber(❨1/4❩)', TASC(0.25));
  Test('ComplexNumber(❨123.456❩)', TASC(123.456));
  Test('ComplexNumber(❨123.456E20❩)', TASC(123.456E20));
  Test('ComplexNumber(❨123.456E-20❩)', TASC(123.456E-20));

  Test('boolean(❨0❩)', True);
  Test('boolean(❨1❩)', True);
  Test('boolean(❨1, 2❩)', True);
  Test('boolean(❨1, 2, 3❩)', True);
  Test('boolean(❨1, 2, 3, 4❩)', True);
  Test('boolean(❨0, 0, 0, 0, 0, 0❩)', True);
  Test('boolean(DebugObject("empty vector"))', False);

  Test('string(❨1, 2, 3❩)', '1'#13#10'2'#13#10'3');
  Test('string(❨-1, 2, -3❩)', '-1'#13#10'2'#13#10'-3');
  Test('string(DebugObject("empty vector"))', '');

  Test('RealVector(❨1, 2, 3❩)', [1, 2, 3]);
  Test('RealVector(❨5, -2, 0, 9❩)', [5, -2, 0, 9]);
  Test('RealVector(❨21❩)', [21]);
  Test('RealVector(DebugObject("empty vector"))', TAlgosimRealVector.Create);

  Test('ComplexVector(❨1, 2, 3❩)', [TASC(1), 2, 3]);
  Test('ComplexVector(❨5, -2, 0, 9❩)', [TASC(5), -2, 0, 9]);
  Test('ComplexVector(❨21❩)', [TASC(21)]);
  Test('ComplexVector(DebugObject("empty vector"))', TAlgosimComplexVector.Create);

  Test('matrix(❨1, 2, 3❩)', 1, [1, 2, 3]);
  Test('matrix(❨5, -2, 0, 9❩)', 1, [5, -2, 0, 9]);
  Test('matrix(❨21❩)', 1, [21]);
  Test('matrix(DebugObject("empty vector"))', TAlgosimRealMatrix.Create);

  Test('RealMatrix(❨1, 2, 3❩)', 1, [1, 2, 3]);
  Test('RealMatrix(❨5, -2, 0, 9❩)', 1, [5, -2, 0, 9]);
  Test('RealMatrix(❨21❩)', 1, [21]);
  Test('RealMatrix(DebugObject("empty vector"))', TAlgosimRealMatrix.Create);

  Test('ComplexMatrix(❨1, 2, 3❩)', 1, [TASC(1), 2, 3]);
  Test('ComplexMatrix(❨5, -2, 0, 9❩)', 1, [TASC(5), -2, 0, 9]);
  Test('ComplexMatrix(❨21❩)', 1, [TASC(21)]);
  Test('ComplexMatrix(DebugObject("empty vector"))', TAlgosimComplexMatrix.Create);

  Test('ToList(❨1, 2, 3❩)', intarr([1, 2, 3]));
  Test('ToList(❨5, -2, 0, 9❩)', intarr([5, -2, 0, 9]));
  Test('ToList(❨5, -2, 0, 9, 0❩)', intarr([5, -2, 0, 9, 0]));
  Test('ToList(❨21❩)', intarr([21]));
  Test('ToList(DebugObject("empty vector"))', intarr([]));

  Test('ToSet(❨1, 2, 3❩)', intset([1, 2, 3]));
  Test('ToSet(❨5, -2, 0, 9❩)', intset([5, -2, 0, 9]));
  Test('ToSet(❨5, -2, 0, 9, 0❩)', intset([5, -2, 0, 9]));
  Test('ToSet(❨21❩)', intset([21]));
  Test('ToSet(DebugObject("empty vector"))', intset([]));

  Test('color(❨0.1, 0.2, 0.3❩)', failure, 'An object of type "real vector" cannot be converted to a colour.');


  //
  // From complex vector
  //

  Test('number(ComplexVector(DebugObject("empty vector")))', failure, 'An empty container cannot be converted to a number.');
  Test('number(❨5, i, 0, 1❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(❨5, i❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(❨5 + i - i❩)', TASC(5));
  Test('number(❨i❩)', ImaginaryUnit);
  Test('number(❨1 + i❩)', 1 + ImaginaryUnit);

  Test('integer(ComplexVector(DebugObject("empty vector")))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨5, i, 0, 1❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨5, i❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨5 + i - i❩)', 5);
  Test('integer(❨i❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨1 + i❩)', failure, 'Object cannot be represented as a 64-bit integer.');

  Test('RealNumber(ComplexVector(DebugObject("empty vector")))', failure, 'An empty container cannot be converted to a number.');
  Test('RealNumber(❨5, i, 0, 1❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(❨5, i❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(❨5 + i - i❩)', 5.0);
  Test('RealNumber(❨i❩)', failure, 'Complex number i cannot be converted to a real number.');
  Test('RealNumber(❨1 + i❩)', failure, 'Complex number 1 + i cannot be converted to a real number.');

  Test('ComplexNumber(ComplexVector(DebugObject("empty vector")))', failure, 'An empty container cannot be converted to a number.');
  Test('ComplexNumber(❨5, i, 0, 1❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(❨5, i❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(❨5 + i - i❩)', TASC(5));
  Test('ComplexNumber(❨i❩)', ImaginaryUnit);
  Test('ComplexNumber(❨1 + i❩)', 1 + ImaginaryUnit);

  Test('boolean(ComplexVector(DebugObject("empty vector")))', False);
  Test('boolean(❨5, i, 0, 1❩)', True);
  Test('boolean(❨5, i❩)', True);
  Test('boolean(❨5 + i - i❩)', True);
  Test('boolean(❨i❩)', True);
  Test('boolean(❨1 + i❩)', True);
  Test('boolean(❨0 + i - i❩)', True);

  Test('string(❨1, 2, i❩)', '1'#13#10'2'#13#10'i');
  Test('string(❨-1, 2, -i❩)', '-1'#13#10'2'#13#10'-i');
  Test('string(ComplexVector(DebugObject("empty vector")))', '');

  Test('RealVector(❨1, 2, i, 0❩)', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component i.');
  Test('RealVector(❨1, -2, -i, 0❩)', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component −i.');
  Test('RealVector(❨1 + i❩)', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component 1 + i.');
  Test('RealVector(❨1 + i - i, 2, 3, 4❩)', [1, 2, 3, 4]);
  Test('RealVector(❨21 + i - i❩)', [21]);

  Test('ComplexVector(❨1, 2, i, 0❩)', [1, 2, ImaginaryUnit, 0]);
  Test('ComplexVector(❨1, -2, -i, 0❩)', [1, -2, -ImaginaryUnit, 0]);
  Test('ComplexVector(❨1 + i❩)', [1 + ImaginaryUnit]);
  Test('ComplexVector(❨1 + i - i, 2, 3, 4❩)', [TASC(1), 2, 3, 4]);
  Test('ComplexVector(❨21 + i - i❩)', [TASC(21)]);

  Test('matrix(❨1, 2, i, 0❩)', 1, [1, 2, ImaginaryUnit, 0]);
  Test('matrix(❨1, -2, -i, 0❩)', 1, [1, -2, -ImaginaryUnit, 0]);
  Test('matrix(❨1 + i❩)', 1, [1 + ImaginaryUnit]);
  Test('matrix(❨1 + i - i, 2, 3, 4❩)', 1, [TASC(1), 2, 3, 4]);
  Test('matrix(❨21 + i - i❩)', 1, [TASC(21)]);

  Test('RealMatrix(❨1, 2, i, 0❩)', failure, 'Complex vector cannot be converted to real matrix because it contains a non-real component i.');
  Test('RealMatrix(❨1, -2, -i, 0❩)', failure, 'Complex vector cannot be converted to real matrix because it contains a non-real component −i.');
  Test('RealMatrix(❨1 + i❩)', failure, 'Complex vector cannot be converted to real matrix because it contains a non-real component 1 + i.');
  Test('RealMatrix(❨1 + i - i, 2, 3, 4❩)', 1, [1, 2, 3, 4]);
  Test('RealMatrix(❨21 + i - i❩)', 1, [21]);

  Test('ComplexMatrix(❨1, 2, i, 0❩)', 1, [1, 2, ImaginaryUnit, 0]);
  Test('ComplexMatrix(❨1, -2, -i, 0❩)', 1, [1, -2, -ImaginaryUnit, 0]);
  Test('ComplexMatrix(❨1 + i❩)', 1, [1 + ImaginaryUnit]);
  Test('ComplexMatrix(❨1 + i - i, 2, 3, 4❩)', 1, [TASC(1), 2, 3, 4]);
  Test('ComplexMatrix(❨21 + i - i❩)', 1, [TASC(21)]);

  Test('ToList(❨1, 2, i, 0, 0❩)', asoarrex([ASO(1), ASO(2), ASO(ImaginaryUnit), ASO(0), ASO(0)]));
  Test('ToList(❨1, -2, -i, 0❩)', asoarrex([ASO(1), ASO(-2), ASO(-ImaginaryUnit), ASO(0)]));
  Test('ToList(❨1 + i❩)', asoarrex([ASO(1 + ImaginaryUnit)]));
  Test('ToList(❨1 + i - i, 2, 3, 4❩)', asoarrex([ASO(1), ASO(2), ASO(3), ASO(4)]));
  Test('ToList(❨21 + i - i❩)', asoarrex([ASO(21)]));

  Test('ToSet(❨1, 2, i, 0, 0❩)', asosetex([ASO(1), ASO(2), ASO(ImaginaryUnit), ASO(0)]));
  Test('ToSet(❨1, -2, -i, 0❩)', asosetex([ASO(1), ASO(-2), ASO(-ImaginaryUnit), ASO(0)]));
  Test('ToSet(❨1 + i❩)', asosetex([ASO(1 + ImaginaryUnit)]));
  Test('ToSet(❨1 + i - i, 2, 3, 4❩)', asosetex([ASO(1), ASO(2), ASO(3), ASO(4)]));
  Test('ToSet(❨21 + i - i❩)', asosetex([ASO(21)]));

  Test('color(❨0.1 + i - i, 0.2, 0.3❩)', failure, 'An object of type "complex vector" cannot be converted to a colour.');


  //
  // From real matrix
  //

  Test('number(DebugObject("empty matrix"))', failure, 'An empty container cannot be converted to a number.');
  Test('number(❨❨5, 3❩, ❨1, 2❩❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(❨❨21❩❩)', 21.0);  
  Test('number(❨❨21.5❩❩)', 21.5);

  Test('integer(DebugObject("empty matrix"))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨❨5, 3❩, ❨1, 2❩❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨❨21❩❩)', 21);
  Test('integer(❨❨21.5❩❩)', failure, 'Object cannot be represented as a 64-bit integer.');

  Test('RealNumber(DebugObject("empty matrix"))', failure, 'An empty container cannot be converted to a number.');
  Test('RealNumber(❨❨5, 3❩, ❨1, 2❩❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(❨❨21❩❩)', 21.0);  
  Test('RealNumber(❨❨21.5❩❩)', 21.5);

  Test('ComplexNumber(DebugObject("empty matrix"))', failure, 'An empty container cannot be converted to a number.');
  Test('ComplexNumber(❨❨5, 3❩, ❨1, 2❩❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(❨❨21❩❩)', TASC(21.0));
  Test('ComplexNumber(❨❨21.5❩❩)', TASC(21.5));

  Test('boolean(DebugObject("empty matrix"))', False);
  Test('boolean(❨❨5, 3❩, ❨1, 2❩❩)', True);
  Test('boolean(❨❨21❩❩)', True);
  Test('boolean(❨❨21.5❩❩)', True);
  Test('boolean(❨❨0❩❩)', True);

  Test('string(DebugObject("empty matrix"))', '');
  Test('string(❨❨5, 3❩, ❨1, -2❩❩)', '5'#9'3'#13#10'1'#9'-2');
  Test('string(❨❨5, 3❩, ❨1, 2❩, ❨0, 4❩❩)', '5'#9'3'#13#10'1'#9'2'#13#10'0'#9'4');
  Test('string(❨❨21❩❩)', '21');
  Test('string(❨❨21.5❩❩)', '21.5');

  Test('vector(DebugObject("empty matrix"))', TAlgosimRealVector.Create);
  Test('vector(❨❨5, 3❩, ❨1, 2❩❩)', [5, 3, 1, 2]);
  Test('vector(❨❨5, 3❩, ❨1, 2❩, ❨0, 4❩❩)', [5, 3, 1, 2, 0, 4]);
  Test('vector(❨❨21❩❩)', [21]);
  Test('vector(❨❨21.5❩❩)', [21.5]);

  Test('RealVector(DebugObject("empty matrix"))', TAlgosimRealVector.Create);
  Test('RealVector(❨❨5, 3❩, ❨1, 2❩❩)', [5, 3, 1, 2]);
  Test('RealVector(❨❨5, 3❩, ❨1, 2❩, ❨0, 4❩❩)', [5, 3, 1, 2, 0, 4]);
  Test('RealVector(❨❨21❩❩)', [21]);
  Test('RealVector(❨❨21.5❩❩)', [21.5]);

  Test('ComplexVector(DebugObject("empty matrix"))', TAlgosimComplexVector.Create);
  Test('ComplexVector(❨❨5, 3❩, ❨1, 2❩❩)', [TASC(5), 3, 1, 2]);
  Test('ComplexVector(❨❨5, 3❩, ❨1, 2❩, ❨0, 4❩❩)', [TASC(5), 3, 1, 2, 0, 4]);
  Test('ComplexVector(❨❨21❩❩)', [TASC(21)]);
  Test('ComplexVector(❨❨21.5❩❩)', [TASC(21.5)]);

  Test('matrix(DebugObject("empty matrix"))', TAlgosimRealMatrix.Create);
  Test('matrix(❨❨5, 3❩, ❨1, 2❩❩)', 2, [5, 3, 1, 2]);
  Test('matrix(❨❨5, 3❩, ❨1, 2❩, ❨0, 4❩❩)', 2, [5, 3, 1, 2, 0, 4]);
  Test('matrix(❨❨21❩❩)', 1, [21]);
  Test('matrix(❨❨21.5❩❩)', 1, [21.5]);

  Test('RealMatrix(DebugObject("empty matrix"))', TAlgosimRealMatrix.Create);
  Test('RealMatrix(❨❨5, 3❩, ❨1, 2❩❩)', 2, [5, 3, 1, 2]);
  Test('RealMatrix(❨❨5, 3❩, ❨1, 2❩, ❨0, 4❩❩)', 2, [5, 3, 1, 2, 0, 4]);
  Test('RealMatrix(❨❨21❩❩)', 1, [21]);
  Test('RealMatrix(❨❨21.5❩❩)', 1, [21.5]);  

  Test('ComplexMatrix(DebugObject("empty matrix"))', TAlgosimComplexMatrix.Create);
  Test('ComplexMatrix(❨❨5, 3❩, ❨1, 2❩❩)', 2, [TASC(5), 3, 1, 2]);
  Test('ComplexMatrix(❨❨5, 3❩, ❨1, 2❩, ❨0, 4❩❩)', 2, [TASC(5), 3, 1, 2, 0, 4]);
  Test('ComplexMatrix(❨❨21❩❩)', 1, [TASC(21)]);
  Test('ComplexMatrix(❨❨21.5❩❩)', 1, [TASC(21.5)]);

  Test('ToList(DebugObject("empty matrix"))', asoarr([]));
  Test('ToList(❨❨5, 3❩, ❨1, 2❩❩)', asoarr([5, 3, 1, 2]));
  Test('ToList(❨❨5, 3❩, ❨1, 2❩, ❨0, 5❩❩)', asoarr([5, 3, 1, 2, 0, 5]));
  Test('ToList(❨❨21❩❩)', asoarr([21]));
  Test('ToList(❨❨21.5❩❩)', asoarr([21.5]));

  Test('ToSet(DebugObject("empty matrix"))', asoset([]));
  Test('ToSet(❨❨5, 3❩, ❨1, 2❩❩)', asoset([5, 3, 1, 2]));
  Test('ToSet(❨❨5, 3❩, ❨1, 2❩, ❨0, 5❩❩)', asoset([5, 3, 1, 2, 0]));
  Test('ToSet(❨❨21❩❩)', asoset([21]));
  Test('ToSet(❨❨21.5❩❩)', asoset([21.5]));

  Test('color(❨❨0.1, 0.2, 0.3❩❩)', failure, 'An object of type "real matrix" cannot be converted to a colour.');


  //
  // From complex matrix
  //

  Test('number(ComplexMatrix(DebugObject("empty matrix")))', failure, 'An empty container cannot be converted to a number.');
  Test('number(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(❨❨21 + i - i❩❩)', TASC(21));
  Test('number(❨❨21.5 + i - i❩❩)', TASC(21.5));
  Test('number(❨❨21 + i❩❩)', 21 + ImaginaryUnit);

  Test('integer(ComplexMatrix(DebugObject("empty matrix")))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨❨21 + i - i❩❩)', 21);
  Test('integer(❨❨21.5 + i - i❩❩)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(❨❨21 + i❩❩)', failure, 'Object cannot be represented as a 64-bit integer.');

  Test('RealNumber(ComplexMatrix(DebugObject("empty matrix")))', failure, 'An empty container cannot be converted to a number.');
  Test('RealNumber(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(❨❨21 + i - i❩❩)', 21.0);
  Test('RealNumber(❨❨21.5 + i - i❩❩)', 21.5);
  Test('RealNumber(❨❨21 + i❩❩)', failure, 'Complex number 21 + i cannot be converted to a real number.');

  Test('ComplexNumber(ComplexMatrix(DebugObject("empty matrix")))', failure, 'An empty container cannot be converted to a number.');
  Test('ComplexNumber(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(❨❨21 + i - i❩❩)', TASC(21));
  Test('ComplexNumber(❨❨21.5 + i - i❩❩)', TASC(21.5));
  Test('ComplexNumber(❨❨21 + i❩❩)', 21 + ImaginaryUnit);  

  Test('boolean(ComplexMatrix(DebugObject("empty matrix")))', False);
  Test('boolean(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', True);
  Test('boolean(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', True);
  Test('boolean(❨❨21 + i - i❩❩)', True);
  Test('boolean(❨❨21.5 + i - i❩❩)', True);
  Test('boolean(❨❨21 + i❩❩)', True);
  Test('boolean(❨❨0 + i - i❩❩)', True);

  Test('string(ComplexMatrix(DebugObject("empty matrix")))', '');
  Test('string(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', '5'#9'3'#9'i'#13#10'0'#9'1'#9'2');
  Test('string(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨-i, 7, -1❩❩)', '5'#9'3'#9'i'#13#10'0'#9'1'#9'2'#13#10'-i'#9'7'#9'-1');
  Test('string(❨❨21 + i - i❩❩)', '21');
  Test('string(❨❨-21 + i - i❩❩)', '-21');
  Test('string(❨❨21.5 + i - i❩❩)', '21.5');
  Test('string(❨❨21 + i❩❩)', '21 + i');
  Test('string(❨❨21 - i❩❩)', '21 - i');
  Test('string(❨❨21 + 5⋅i❩❩)', '21 + 5⋅i');
  Test('string(❨❨21 - 5⋅i❩❩)', '21 - 5⋅i');
  Test('string(❨❨-21 - 5⋅i❩❩)', '-21 - 5⋅i');
  Test('string(❨❨-5⋅i❩❩)', '-5⋅i');
  Test('string(❨❨0 + i - i❩❩)', '0');

  Test('vector(ComplexMatrix(DebugObject("empty matrix")))', TAlgosimComplexVector.Create);
  Test('vector(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', [5, 3, ImaginaryUnit, 0, 1, 2]);
  Test('vector(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', [5, 3, ImaginaryUnit, 0, 1, 2, ImaginaryUnit, 7, 1]);
  Test('vector(❨❨21 + i - i❩❩)', [TASC(21)]);
  Test('vector(❨❨21.5 + i - i❩❩)', [TASC(21.5)]);
  Test('vector(❨❨21 + i❩❩)', [21 + ImaginaryUnit]);
  Test('vector(❨❨0 + i - i❩❩)', [TASC(0)]);

  Test('RealVector(ComplexMatrix(DebugObject("empty matrix")))', TAlgosimRealVector.Create);
  Test('RealVector(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', failure, 'Complex matrix cannot be converted to real vector because it contains a non-real component i.');
  Test('RealVector(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', failure, 'Complex matrix cannot be converted to real vector because it contains a non-real component i.');
  Test('RealVector(❨❨21 + i - i❩❩)', [21]);
  Test('RealVector(❨❨21.5 + i - i❩❩)', [21.5]);
  Test('RealVector(❨❨21 + i❩❩)', failure, 'Complex matrix cannot be converted to real vector because it contains a non-real component 21 + i.');
  Test('RealVector(❨❨0 + i - i❩❩)', [0]);

  Test('ComplexVector(ComplexMatrix(DebugObject("empty matrix")))', TAlgosimComplexVector.Create);
  Test('ComplexVector(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', [5, 3, ImaginaryUnit, 0, 1, 2]);
  Test('ComplexVector(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', [5, 3, ImaginaryUnit, 0, 1, 2, ImaginaryUnit, 7, 1]);
  Test('ComplexVector(❨❨21 + i - i❩❩)', [TASC(21)]);
  Test('ComplexVector(❨❨21.5 + i - i❩❩)', [TASC(21.5)]);
  Test('ComplexVector(❨❨21 + i❩❩)', [21 + ImaginaryUnit]);
  Test('ComplexVector(❨❨0 + i - i❩❩)', [TASC(0)]);

  Test('matrix(ComplexMatrix(DebugObject("empty matrix")))', TAlgosimComplexMatrix.Create);
  Test('matrix(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', 3, [5, 3, ImaginaryUnit, 0, 1, 2]);
  Test('matrix(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', 3, [5, 3, ImaginaryUnit, 0, 1, 2, ImaginaryUnit, 7, 1]);
  Test('matrix(❨❨21 + i - i❩❩)', 1, [TASC(21)]);
  Test('matrix(❨❨21.5 + i - i❩❩)', 1, [TASC(21.5)]);
  Test('matrix(❨❨21 + i❩❩)', 1, [21 + ImaginaryUnit]);
  Test('matrix(❨❨0 + i - i❩❩)', 1, [TASC(0)]);

  Test('RealMatrix(ComplexMatrix(DebugObject("empty matrix")))', TAlgosimRealMatrix.Create);
  Test('RealMatrix(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', failure, 'Complex matrix cannot be converted to real matrix because it contains a non-real component i.');
  Test('RealMatrix(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', failure, 'Complex matrix cannot be converted to real matrix because it contains a non-real component i.');
  Test('RealMatrix(❨❨21 + i - i❩❩)', 1, [21]);
  Test('RealMatrix(❨❨21.5 + i - i❩❩)', 1, [21.5]);
  Test('RealMatrix(❨❨21 + i❩❩)', failure, 'Complex matrix cannot be converted to real matrix because it contains a non-real component 21 + i.');
  Test('RealMatrix(❨❨0 + i - i❩❩)', 1, [0]);
  
  Test('ComplexMatrix(ComplexMatrix(DebugObject("empty matrix")))', TAlgosimComplexMatrix.Create);
  Test('ComplexMatrix(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', 3, [5, 3, ImaginaryUnit, 0, 1, 2]);
  Test('ComplexMatrix(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', 3, [5, 3, ImaginaryUnit, 0, 1, 2, ImaginaryUnit, 7, 1]);
  Test('ComplexMatrix(❨❨21 + i - i❩❩)', 1, [TASC(21)]);
  Test('ComplexMatrix(❨❨21.5 + i - i❩❩)', 1, [TASC(21.5)]);
  Test('ComplexMatrix(❨❨21 + i❩❩)', 1, [21 + ImaginaryUnit]);
  Test('ComplexMatrix(❨❨0 + i - i❩❩)', 1, [TASC(0)]);

  Test('ToList(ComplexMatrix(DebugObject("empty matrix")))', asoarrex([]));
  Test('ToList(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', asoarrex([ASO(5), ASO(3), ASO(ImaginaryUnit), ASO(0), ASO(1), ASO(2)]));
  Test('ToList(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', asoarrex([ASO(5), ASO(3), ASO(ImaginaryUnit), ASO(0), ASO(1), ASO(2), ASO(ImaginaryUnit), ASO(7), ASO(1)]));
  Test('ToList(❨❨21 + i - i❩❩)', asoarrex([ASO(21)]));
  Test('ToList(❨❨21.5 + i - i❩❩)', asoarrex([ASO(21.5)]));
  Test('ToList(❨❨21 + i❩❩)', asoarrex([ASO(21 + ImaginaryUnit)]));
  Test('ToList(❨❨0 + i - i❩❩)', asoarrex([ASO(0)]));

  Test('ToSet(ComplexMatrix(DebugObject("empty matrix")))', asosetex([]));
  Test('ToSet(❨❨5, 3, i❩, ❨0, 1, 2❩❩)', asosetex([ASO(5), ASO(3), ASO(ImaginaryUnit), ASO(0), ASO(1), ASO(2)]));
  Test('ToSet(❨❨5, 3, i❩, ❨0, 1, 2❩, ❨i, 7, 1❩❩)', asosetex([ASO(5), ASO(3), ASO(ImaginaryUnit), ASO(0), ASO(1), ASO(2), ASO(7)]));
  Test('ToSet(❨❨21 + i - i❩❩)', asosetex([ASO(21)]));
  Test('ToSet(❨❨21.5 + i - i❩❩)', asosetex([ASO(21.5)]));
  Test('ToSet(❨❨21 + i❩❩)', asosetex([ASO(21 + ImaginaryUnit)]));
  Test('ToSet(❨❨0 + i - i❩❩)', asosetex([ASO(0)]));

  Test('color(❨❨0.1 + i - i, 0.2, 0.3❩❩)', failure, 'An object of type "complex matrix" cannot be converted to a colour.');


  //
  // From list
  //

  Test('number(''())', failure, 'An empty container cannot be converted to a number.');
  Test('number(''(3, 1, 4, 1, 5))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(''(3, 1, 4, 1, 5, 9, 2, 6, 5))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(''(3, i, 1, 0, -i, -1))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(''(3, i, 1, 0, -i, -1, "cat", "dog", "dog"))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(''(0))', 0);
  Test('number(''(123))', 123);
  Test('number(''(-123))', -123);
  Test('number(''(1/7))', rat(1, 7));
  Test('number(''(-1/7))', rat(-1, 7));
  Test('number(''(123.456))', 123.456);
  Test('number(''(-123.456))', -123.456);
  Test('number(''(2 + i))', 2 + ImaginaryUnit);
  Test('number(''(2 - i))', 2 - ImaginaryUnit);
  Test('number(''(2 + 3⋅i))', 2 + 3*ImaginaryUnit);
  Test('number(''(2 - 3⋅i))', 2 - 3*ImaginaryUnit);
  Test('number(''("cat"))', failure, 'Cannot convert string "cat" to a complex number.');
  Test('number(''(false))', 0);
  Test('number(''(true))', 1);
  Test('number(''(IdentityMatrix(4)))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(''(IdentityMatrix(1)))', 1.0);
  Test('number(''(ZeroVector(4)))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(''(ZeroVector(1)))', 0.0);
  Test('number(''(''(''(''(21)))))', 21);
  Test('number(''(''(''(''(21, 22)))))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(''(''(''(''()))))', failure, 'An empty container cannot be converted to a number.');
  Test('number(''(''(''(''(IdentityMatrix(1)/2)))))', 0.5);

  Test('integer(''())', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(3, 1, 4, 1, 5))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(3, 1, 4, 1, 5, 9, 2, 6, 5))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(3, i, 1, 0, -i, -1))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(3, i, 1, 0, -i, -1, "cat", "dog", "dog"))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(0))', 0);
  Test('integer(''(123))', 123);
  Test('integer(''(-123))', -123);
  Test('integer(''(1/7))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(-1/7))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(123.456))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(-123.456))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(2 + i))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(2 - i))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(2 + 3⋅i))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(2 - 3⋅i))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''("cat"))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(false))', 0);
  Test('integer(''(true))', 1);
  Test('integer(''(IdentityMatrix(4)))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(IdentityMatrix(1)))', 1);
  Test('integer(''(ZeroVector(4)))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(ZeroVector(1)))', 0);
  Test('integer(''(''(''(''(21)))))', 21);
  Test('integer(''(''(''(''(21, 22)))))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(''(''(''()))))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(''(''(''(''(IdentityMatrix(1)/2)))))', failure, 'Object cannot be represented as a 64-bit integer.');

  Test('RealNumber(''())', failure, 'An empty container cannot be converted to a number.');
  Test('RealNumber(''(3, 1, 4, 1, 5))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(''(3, 1, 4, 1, 5, 9, 2, 6, 5))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(''(3, i, 1, 0, -i, -1))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(''(3, i, 1, 0, -i, -1, "cat", "dog", "dog"))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(''(0))', 0.0);
  Test('RealNumber(''(123))', 123.0);
  Test('RealNumber(''(-123))', -123.0);
  Test('RealNumber(''(1/7))', 1/7);
  Test('RealNumber(''(-1/7))', -1/7);
  Test('RealNumber(''(123.456))', 123.456);
  Test('RealNumber(''(-123.456))', -123.456);
  Test('RealNumber(''(2 + i))', failure, 'Complex number 2 + i cannot be converted to a real number.');
  Test('RealNumber(''(2 - i))', failure, 'Complex number 2 − i cannot be converted to a real number.');
  Test('RealNumber(''(2 + 3⋅i))', failure, 'Complex number 2 + 3⋅i cannot be converted to a real number.');
  Test('RealNumber(''(2 - 3⋅i))', failure, 'Complex number 2 − 3⋅i cannot be converted to a real number.');
  Test('RealNumber(''("cat"))', failure, 'Cannot convert string "cat" to a real number.');
  Test('RealNumber(''(false))', 0.0);
  Test('RealNumber(''(true))', 1.0);
  Test('RealNumber(''(IdentityMatrix(4)))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(''(IdentityMatrix(1)))', 1.0);
  Test('RealNumber(''(ZeroVector(4)))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(''(ZeroVector(1)))', 0.0);
  Test('RealNumber(''(''(''(''(21)))))', 21.0);
  Test('RealNumber(''(''(''(''(21, 22)))))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(''(''(''(''()))))', failure, 'An empty container cannot be converted to a number.');
  Test('RealNumber(''(''(''(''(IdentityMatrix(1)/2)))))', 0.5);

  Test('ComplexNumber(''())', failure, 'An empty container cannot be converted to a number.');
  Test('ComplexNumber(''(3, 1, 4, 1, 5))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(''(3, 1, 4, 1, 5, 9, 2, 6, 5))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(''(3, i, 1, 0, -i, -1))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(''(3, i, 1, 0, -i, -1, "cat", "dog", "dog"))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(''(0))', TASC(0.0));
  Test('ComplexNumber(''(123))', TASC(123.0));
  Test('ComplexNumber(''(-123))', TASC(-123.0));
  Test('ComplexNumber(''(1/7))', TASC(1/7));
  Test('ComplexNumber(''(-1/7))', TASC(-1/7));
  Test('ComplexNumber(''(123.456))', TASC(123.456));
  Test('ComplexNumber(''(-123.456))', TASC(-123.456));
  Test('ComplexNumber(''(2 + i))', 2 + ImaginaryUnit);
  Test('ComplexNumber(''(2 - i))', 2 - ImaginaryUnit);
  Test('ComplexNumber(''(2 + 3⋅i))', 2 + 3*ImaginaryUnit);
  Test('ComplexNumber(''(2 - 3⋅i))', 2 - 3*ImaginaryUnit);
  Test('ComplexNumber(''("cat"))', failure, 'Cannot convert string "cat" to a complex number.');
  Test('ComplexNumber(''(false))', TASC(0.0));
  Test('ComplexNumber(''(true))', TASC(1.0));
  Test('ComplexNumber(''(IdentityMatrix(4)))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(''(IdentityMatrix(1)))', TASC(1.0));
  Test('ComplexNumber(''(ZeroVector(4)))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(''(ZeroVector(1)))', TASC(0.0));
  Test('ComplexNumber(''(''(''(''(21)))))', TASC(21.0));
  Test('ComplexNumber(''(''(''(''(21, 22)))))', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(''(''(''(''()))))', failure, 'An empty container cannot be converted to a number.');
  Test('ComplexNumber(''(''(''(''(IdentityMatrix(1)/2)))))', TASC(0.5));

  Test('boolean(''())', False);
  Test('boolean(''(3, 1, 4, 1, 5))', True);
  Test('boolean(''(3, 1, 4, 1, 5, 9, 2, 6, 5))', True);
  Test('boolean(''(3, i, 1, 0, -i, -1))', True);
  Test('boolean(''(3, i, 1, 0, -i, -1, "cat", "dog", "dog"))', True);
  Test('boolean(''(0))', True);
  Test('boolean(''(123))', True);
  Test('boolean(''(-123))', True);
  Test('boolean(''(1/7))', True);
  Test('boolean(''(-1/7))', True);
  Test('boolean(''(123.456))', True);
  Test('boolean(''(-123.456))', True);
  Test('boolean(''(2 + i))', True);
  Test('boolean(''(2 - i))', True);
  Test('boolean(''(2 + 3⋅i))', True);
  Test('boolean(''(2 - 3⋅i))', True);
  Test('boolean(''("cat"))', True);
  Test('boolean(''(false))', True);
  Test('boolean(''(true))', True);
  Test('boolean(''(IdentityMatrix(4)))', True);
  Test('boolean(''(IdentityMatrix(1)))', True);
  Test('boolean(''(ZeroVector(4)))', True);
  Test('boolean(''(ZeroVector(1)))', True);
  Test('boolean(''(''(''(''(21)))))', True);
  Test('boolean(''(''(''(''(21, 22)))))', True);
  Test('boolean(''(''(''(''()))))', True);
  Test('boolean(''(''(''(''(IdentityMatrix(1)/2)))))', True);

  Test('string(''())', '');
  Test('string(''(3, 1, 4, 1, 5))', '3'#13#10'1'#13#10'4'#13#10'1'#13#10'5');
  Test('string(''(3, 1, 4, 1, 5, 9, 2, 6, 5))', '3'#13#10'1'#13#10'4'#13#10'1'#13#10'5'#13#10'9'#13#10'2'#13#10'6'#13#10'5');
  Test('string(''(3, i, 1, 0, -i, -1))', '3'#13#10'i'#13#10'1'#13#10'0'#13#10'-i'#13#10'-1');
  Test('string(''(3, i, 1, 0, -i, -1, "cat", "dog", "dog"))', '3'#13#10'i'#13#10'1'#13#10'0'#13#10'-i'#13#10'-1'#13#10'"cat"'#13#10'"dog"'#13#10'"dog"');
  Test('string(''(0))', '0');
  Test('string(''(123))', '123');
  Test('string(''(-123))', '-123');
  Test('string(''(1/7))', '1/7');
  Test('string(''(-1/7))', '-1/7');
  Test('string(''(123.456))', '123.456');
  Test('string(''(-123.456))', '-123.456');
  Test('string(''(2 + i))', '2 + i');
  Test('string(''(2 - i))', '2 - i');
  Test('string(''(2 + 3⋅i))', '2 + 3⋅i');
  Test('string(''(2 - 3⋅i))', '2 - 3⋅i');
  Test('string(''("cat"))', '"cat"');
  Test('string(''(false))', 'false');
  Test('string(''(true))', 'true');
  Test('string(''(IdentityMatrix(4)))', '((1, 0, 0, 0), (0, 1, 0, 0), (0, 0, 1, 0), (0, 0, 0, 1))');
  Test('string(''(IdentityMatrix(1)))', '((1))');
  Test('string(''(ZeroVector(4)))', '(0, 0, 0, 0)');
  Test('string(''(ZeroVector(1)))', '(0)');
  Test('string(''(''(''(''(21)))))', '''(''(''(21)))');
  Test('string(''(''(''(''(21, 22)))))', '''(''(''(21, 22)))');
  Test('string(''(''(''(''()))))', '''(''(''()))');
  Test('string(''(''(''(''(IdentityMatrix(1)/2)))))', '''(''(''(((0.5)))))');

  Test('vector(''(1, 2, 3))', [1, 2, 3]);
  Test('vector(''(1, 2, 3, i))', [1, 2, 3, ImaginaryUnit]);
  Test('vector(''(21))', [21]);
  Test('vector(''(i))', [ImaginaryUnit]);
  Test('vector(''())', TAlgosimRealVector.Create);
  Test('vector(''(1, 2, 3, "dog", true, 4))', [1, 2, 3, 4]);
  Test('vector(''(1, 2, 3, "dog", true, 4, i))', [1, 2, 3, 4, ImaginaryUnit]);

  Test('RealVector(''(1, 2, 3))', [1, 2, 3]);
  Test('RealVector(''(1, 2, 3, i))', failure, 'Complex number i cannot be converted to a real number.');
  Test('RealVector(''(21))', [21]);
  Test('RealVector(''(i))', failure, 'Complex number i cannot be converted to a real number.');
  Test('RealVector(''())', TAlgosimRealVector.Create);
  Test('RealVector(''(1, 2, 3, "dog", true, 4))', [1, 2, 3, 4]);
  Test('RealVector(''(1, 2, 3, "dog", true, 4, i))', failure, 'Complex number i cannot be converted to a real number.');

  Test('ComplexVector(''(1, 2, 3))', [TASC(1), 2, 3]);
  Test('ComplexVector(''(1, 2, 3, i))', [1, 2, 3, ImaginaryUnit]);
  Test('ComplexVector(''(21))', [TASC(21)]);
  Test('ComplexVector(''(i))', [ImaginaryUnit]);
  Test('ComplexVector(''())', TAlgosimComplexVector.Create);
  Test('ComplexVector(''(1, 2, 3, "dog", true, 4))', [TASC(1), 2, 3, 4]);
  Test('ComplexVector(''(1, 2, 3, "dog", true, 4, i))', [1, 2, 3, 4, ImaginaryUnit]);

  Test('matrix(''(1, 2, 3))', failure, 'Cannot convert a list which is not 2D to a matrix.');
  Test('matrix(''(''(1, 2, 3), ''(4, 5, 6), ''(7, 8, 9)))', 3, [1, 2, 3, 4, 5, 6, 7, 8, 9]);
  Test('matrix(''(''(1, 2, 3), ''(4, 5), ''(7, 8, 9)))', failure, 'Cannot convert a list which is not 2D to a matrix.');
  Test('matrix(''(''(1, 2, 3), ''(4, 5, 6), ''(7, 8)))', failure, 'Cannot convert a list which is not 2D to a matrix.');
  Test('matrix(''(''(1, 2, 3), ''(4, 5, 6), ''(7, i, 9)))', 3, [1, 2, 3, 4, 5, 6, 7, ImaginaryUnit, 9]);
  Test('matrix(''(''(10), ''(20), ''(30)))', 1, [10, 20, 30]);
  Test('matrix(''(''(21)))', 1, [21]);
  Test('matrix(''(''()))', failure, 'A matrix must have size at least 1×1.');
  Test('matrix(''())', failure, 'A matrix must have size at least 1×1.');
  Test('matrix(''(''(1, 2, 3), ''(4, 5, 6), ''("cat", 8, 9)))', failure, 'Object isn''t a real number.');
  Test('matrix(''(''(1, 2, 3), ''(4, 5, 6), ''("cat", 8, i)))', failure, 'Object isn''t a complex number.');

  Test('RealMatrix(''(1, 2, 3))', failure, 'Cannot convert a list which is not 2D to a matrix.');
  Test('RealMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''(7, 8, 9)))', 3, [1, 2, 3, 4, 5, 6, 7, 8, 9]);
  Test('RealMatrix(''(''(1, 2, 3), ''(4, 5), ''(7, 8, 9)))', failure, 'Cannot convert a list which is not 2D to a matrix.');
  Test('RealMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''(7, 8)))', failure, 'Cannot convert a list which is not 2D to a matrix.');
  Test('RealMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''(7, i, 9)))', failure, 'Object isn''t a real number.');
  Test('RealMatrix(''(''(10), ''(20), ''(30)))', 1, [10, 20, 30]);
  Test('RealMatrix(''(''(21)))', 1, [21]);
  Test('RealMatrix(''(''()))', failure, 'A matrix must have size at least 1×1.');
  Test('RealMatrix(''())', failure, 'A matrix must have size at least 1×1.');
  Test('RealMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''("cat", 8, 9)))', failure, 'Object isn''t a real number.');
  Test('RealMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''("cat", 8, i)))', failure, 'Object isn''t a real number.');

  Test('ComplexMatrix(''(1, 2, 3))', failure, 'Cannot convert a list which is not 2D to a matrix.');
  Test('ComplexMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''(7, 8, 9)))', 3, [TASC(1), 2, 3, 4, 5, 6, 7, 8, 9]);
  Test('ComplexMatrix(''(''(1, 2, 3), ''(4, 5), ''(7, 8, 9)))', failure, 'Cannot convert a list which is not 2D to a matrix.');
  Test('ComplexMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''(7, 8)))', failure, 'Cannot convert a list which is not 2D to a matrix.');
  Test('ComplexMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''(7, i, 9)))', 3, [1, 2, 3, 4, 5, 6, 7, ImaginaryUnit, 9]);
  Test('ComplexMatrix(''(''(10), ''(20), ''(30)))', 1, [TASC(10), 20, 30]);
  Test('ComplexMatrix(''(''(21)))', 1, [TASC(21)]);
  Test('ComplexMatrix(''(''()))', failure, 'A matrix must have size at least 1×1.');
  Test('ComplexMatrix(''())', failure, 'A matrix must have size at least 1×1.');
  Test('ComplexMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''("cat", 8, 9)))', failure, 'Object isn''t a complex number.');
  Test('ComplexMatrix(''(''(1, 2, 3), ''(4, 5, 6), ''("cat", 8, i)))', failure, 'Object isn''t a complex number.');

  Test('ToList(''())', asoarr([]));
  Test('ToList(''(3, 1, 4, 1, 5))', asoarr([3, 1, 4, 1, 5]));
  Test('ToList(''(3, 1, 4, 1, 5, 9, 2, 6, 5))', asoarr([3, 1, 4, 1, 5, 9, 2, 6, 5]));
  Test('ToList(''(3, i, 1, 0, -i, -1))', asoarrex([ASOInt(3), ASO(ImaginaryUnit), ASOInt(1), ASOInt(0), ASO(-ImaginaryUnit), ASOInt(-1)]));
  Test('ToList(''(3, i, 1, 0, -i, -1, "cat", "dog", "dog"))', asoarrex([ASOInt(3), ASO(ImaginaryUnit), ASOInt(1), ASOInt(0), ASO(-ImaginaryUnit), ASOInt(-1), ASO('cat'), ASO('dog'), ASO('dog')]));
  Test('ToList(''(0))', asoarr([0]));
  Test('ToList(''(123))', asoarr([123]));
  Test('ToList(''(-123))', asoarr([-123]));
  Test('ToList(''(1/7))', asoarrex([ASORat(rat(1, 7))]));
  Test('ToList(''(-1/7))', asoarrex([ASORat(rat(-1, 7))]));
  Test('ToList(''(123.456))', asoarr([123.456]));
  Test('ToList(''(-123.456))', asoarr([-123.456]));
  Test('ToList(''(2 + i))', asoarrex([ASO(2 + ImaginaryUnit)]));
  Test('ToList(''(2 - i))', asoarrex([ASO(2 - ImaginaryUnit)]));
  Test('ToList(''(2 + 3⋅i))', asoarrex([ASO(2 + 3*ImaginaryUnit)]));
  Test('ToList(''(2 - 3⋅i))', asoarrex([ASO(2 - 3*ImaginaryUnit)]));
  Test('ToList(''("cat"))', asoarr(['cat']));
  Test('ToList(''(false))', asoarr([False]));
  Test('ToList(''(true))', asoarr([True]));
  Test('ToList(''(IdentityMatrix(4)))', asoarrex([ASO(IdentityMatrix(4))]));
  Test('ToList(''(IdentityMatrix(1)))', asoarrex([ASO(IdentityMatrix(1))]));
  Test('ToList(''(ZeroVector(4)))', asoarrex([ASO(ZeroVector(4))]));
  Test('ToList(''(ZeroVector(1)))', asoarrex([ASO(ZeroVector(1))]));
  Test('ToList(''(''(''(''(21)))))', asoarrex([asoarrex([asoarrex([asoarr([21])])])]));
  Test('ToList(''(''(''(''(21, 22)))))', asoarrex([asoarrex([asoarrex([asoarr([21, 22])])])]));
  Test('ToList(''(''(''(''()))))', asoarrex([asoarrex([asoarrex([asoarr([])])])]));
  Test('ToList(''(''(''(''(IdentityMatrix(1)/2)))))', asoarrex([asoarrex([asoarrex([asoarrex([ASO(IdentityMatrix(1)/2)])])])]));

  Test('ToSet(''())', asoset([]));
  Test('ToSet(''(3, 1, 4, 1, 5))', asoset([3, 1, 4, 5]));
  Test('ToSet(''(3, 1, 4, 1, 5, 9, 2, 6, 5))', asoset([3, 1, 4, 5, 9, 2, 6]));
  Test('ToSet(''(3, i, 1, 0, -i, -1))', asosetex([ASOInt(3), ASO(ImaginaryUnit), ASOInt(1), ASOInt(0), ASO(-ImaginaryUnit), ASOInt(-1)]));
  Test('ToSet(''(3, i, 1, 0, -i, -1, "cat", "dog", "dog"))', asosetex([ASOInt(3), ASO(ImaginaryUnit), ASOInt(1), ASOInt(0), ASO(-ImaginaryUnit), ASOInt(-1), ASO('cat'), ASO('dog')]));
  Test('ToSet(''(0))', asoset([0]));
  Test('ToSet(''(123))', asoset([123]));
  Test('ToSet(''(-123))', asoset([-123]));
  Test('ToSet(''(1/7))', asosetex([ASORat(rat(1, 7))]));
  Test('ToSet(''(-1/7))', asosetex([ASORat(rat(-1, 7))]));
  Test('ToSet(''(123.456))', asoset([123.456]));
  Test('ToSet(''(-123.456))', asoset([-123.456]));
  Test('ToSet(''(2 + i))', asosetex([ASO(2 + ImaginaryUnit)]));
  Test('ToSet(''(2 - i))', asosetex([ASO(2 - ImaginaryUnit)]));
  Test('ToSet(''(2 + 3⋅i))', asosetex([ASO(2 + 3*ImaginaryUnit)]));
  Test('ToSet(''(2 - 3⋅i))', asosetex([ASO(2 - 3*ImaginaryUnit)]));
  Test('ToSet(''("cat"))', asoset(['cat']));
  Test('ToSet(''(false))', asoset([False]));
  Test('ToSet(''(true))', asoset([True]));
  Test('ToSet(''(IdentityMatrix(4)))', asosetex([ASO(IdentityMatrix(4))]));
  Test('ToSet(''(IdentityMatrix(1)))', asosetex([ASO(IdentityMatrix(1))]));
  Test('ToSet(''(ZeroVector(4)))', asosetex([ASO(ZeroVector(4))]));
  Test('ToSet(''(ZeroVector(1)))', asosetex([ASO(ZeroVector(1))]));
  Test('ToSet(''(''(''(''(21)))))', asosetex([asoarrex([asoarrex([asoarr([21])])])]));
  Test('ToSet(''(''(''(''(21, 22)))))', asosetex([asoarrex([asoarrex([asoarr([21, 22])])])]));
  Test('ToSet(''(''(''(''()))))', asosetex([asoarrex([asoarrex([asoarr([])])])]));
  Test('ToSet(''(''(''(''(IdentityMatrix(1)/2)))))', asosetex([asoarrex([asoarrex([asoarrex([ASO(IdentityMatrix(1)/2)])])])]));

  Test('color(''(0.1, 0.2, 0.3))', failure, 'An object of type "array" cannot be converted to a colour.');


  //
  // From set
  //

  Test('number({})', failure, 'An empty container cannot be converted to a number.');
  Test('number({1, 2, 3})', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number({21})', 21);
  Test('number({1/7})', rat(1, 7));
  Test('number({21.5})', 21.5);
  Test('number({21 + i})', 21 + ImaginaryUnit);
  Test('number({false})', 0);
  Test('number({true})', 1);
  Test('number({"cat"})', failure, 'Cannot convert string "cat" to a complex number.');
  Test('number({''(21)})', 21);
  Test('number({❨1, 2, 3❩})', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number({❨394❩})', 394.0);

  Test('integer({})', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer({1, 2, 3})', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer({21})', 21);
  Test('integer({1/7})', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer({21.5})', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer({21 + i})', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer({false})', 0);
  Test('integer({true})', 1);
  Test('integer({"cat"})', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer({''(21)})', 21);
  Test('integer({❨1, 2, 3❩})', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer({❨394❩})', 394);

  Test('RealNumber({})', failure, 'An empty container cannot be converted to a number.');
  Test('RealNumber({1, 2, 3})', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber({21})', 21.0);
  Test('RealNumber({1/7})', 1/7);
  Test('RealNumber({21.5})', 21.5);
  Test('RealNumber({21 + i})', failure, 'Complex number 21 + i cannot be converted to a real number.');
  Test('RealNumber({false})', 0.0);
  Test('RealNumber({true})', 1.0);
  Test('RealNumber({"cat"})', failure, 'Cannot convert string "cat" to a real number.');
  Test('RealNumber({''(21)})', 21.0);
  Test('RealNumber({❨1, 2, 3❩})', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber({❨394❩})', 394.0);

  Test('ComplexNumber({})', failure, 'An empty container cannot be converted to a number.');
  Test('ComplexNumber({1, 2, 3})', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber({21})', TASC(21));
  Test('ComplexNumber({1/7})', TASC(1/7));
  Test('ComplexNumber({21.5})', TASC(21.5));
  Test('ComplexNumber({21 + i})', 21 + ImaginaryUnit);
  Test('ComplexNumber({false})', TASC(0));
  Test('ComplexNumber({true})', TASC(1));
  Test('ComplexNumber({"cat"})', failure, 'Cannot convert string "cat" to a complex number.');
  Test('ComplexNumber({''(21)})', TASC(21));
  Test('ComplexNumber({❨1, 2, 3❩})', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber({❨394❩})', TASC(394.0));

  Test('boolean({})', False);
  Test('boolean({1, 2, 3})', True);
  Test('boolean({21})', True);
  Test('boolean({1/7})', True);
  Test('boolean({21.5})', True);
  Test('boolean({21 + i})', True);
  Test('boolean({false})', True);
  Test('boolean({0})', True);
  Test('boolean({true})', True);
  Test('boolean({"cat"})', True);
  Test('boolean({''(21)})', True);
  Test('boolean({❨1, 2, 3❩})', True);
  Test('boolean({❨394❩})', True);

  Test('string({})', '∅');
  Test('string({1, 2, 3})', '{1, 2, 3}');
  Test('string({21})', '{21}');
  Test('string({1/7})', '{1/7}');
  Test('string({21.5})', '{21.5}');
  Test('string({21 + i})', '{21 + i}');
  Test('string({false})', '{false}');
  Test('string({0})', '{0}');
  Test('string({true})', '{true}');
  Test('string({"cat"})', '{"cat"}');
  Test('string({''(21)})', '{''(21)}');
  Test('string({❨1, 2, 3❩})', '{(1, 2, 3)}');
  Test('string({❨394❩})', '{(394)}');

  Test('s ≔ {1, 2, 3, 4, 5, 6}; type(vector(s)) = "real vector" ∧ ToSet(vector(s)) = s', True);
  Test('s ≔ {}; type(vector(s)) = "real vector" ∧ ToSet(vector(s)) = s', True);
  Test('s ≔ {1, 2, 3, 4, 5, 6, i}; type(vector(s)) = "complex vector" ∧ ToSet(vector(s)) = s', True);

  Test('s ≔ {1, 2, 3, 4, 5, 6}; type(RealVector(s)) = "real vector" ∧ ToSet(RealVector(s)) = s', True);
  Test('s ≔ {}; type(RealVector(s)) = "real vector" ∧ ToSet(RealVector(s)) = s', True);
  Test('s ≔ {1, 2, 3, 4, 5, 6, i}; RealVector(s)', failure, 'Complex number i cannot be converted to a real number.');

  Test('s ≔ {1, 2, 3, 4, 5, 6}; type(ComplexVector(s)) = "complex vector" ∧ ToSet(ComplexVector(s)) = s', True);
  Test('s ≔ {}; type(ComplexVector(s)) = "complex vector" ∧ ToSet(ComplexVector(s)) = s', True);
  Test('s ≔ {1, 2, 3, 4, 5, 6, i}; type(ComplexVector(s)) = "complex vector" ∧ ToSet(ComplexVector(s)) = s', True);

  Test('matrix({1, 2, 3})', failure, 'An object of type "set" cannot be converted to a real matrix.');
  Test('matrix({1, 2, i})', failure, 'An object of type "set" cannot be converted to a complex matrix.');

  Test('RealMatrix({1, 2, 3})', failure, 'An object of type "set" cannot be converted to a real matrix.');
  Test('ComplexMatrix({1, 2, 3})', failure, 'An object of type "set" cannot be converted to a complex matrix.');

  Test('s ≔ {1, 2, 3, 4, 5, 6}; type(ToList(s)) = "array" ∧ ToSet(ToList(s)) = s', True);
  Test('s ≔ {}; type(ToList(s)) = "array" ∧ ToSet(ToList(s)) = s', True);
  Test('s ≔ {1, 2, 3, 4, 5, 6, i}; type(ToList(s)) = "array" ∧ ToSet(ToList(s)) = s', True);
  Test('s ≔ {1, 2, 3, 4, 5, 6, i, "cat", "dog", 10, 1.23}; type(ToList(s)) = "array" ∧ ToSet(ToList(s)) = s', True);

  Test('s ≔ {1, 2, 3, 4, 5, 6}; ToSet(s) = s', True);
  Test('s ≔ {}; ToSet(s) = s', True);
  Test('s ≔ {1, 2, 3, 4, 5, 6, i}; ToSet(s) = s', True);
  Test('s ≔ {1, 2, 3, 4, 5, 6, i, "cat", "dog", 10, 1.23}; ToSet(s) = s', True);

  Test('delete(s)', success);

  Test('color({0.1, 0.2, 0.3})', failure, 'An object of type "set" cannot be converted to a colour.');


  //
  // From structure
  //

  Test('name ≔ (fn, ln) ↦ struct("first": fn, "last": ln); person ≔ (fn, ln, sex, age) ↦ struct("name": name(fn, ln), "sex": sex, "age": age);', null);
  Test('Jane ≔ person("Jane", "Smith", "female", 40); Mike ≔ person("Mike", "Doe", "male", 45);', null);
  Test('Mary ≔ struct("mother": Jane, "father": Mike);', null); 

  Test('number(struct())', failure, 'An empty container cannot be converted to a number.');
  Test('number(Jane)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('number(struct("value": 42))', 42); 
  Test('number(struct("value": 42/43))', rat(42, 43));
  Test('number(struct("value": 42.5))', 42.5);
  Test('number(struct("value": 42.5 + i))', 42.5 + ImaginaryUnit);
  Test('number(struct("value": "cat"))', failure, 'Cannot convert string "cat" to a complex number.');

  Test('integer(struct())', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(Jane)', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(struct("value": 42))', 42);
  Test('integer(struct("value": 42/43))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(struct("value": 42.5))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(struct("value": 42.5 + i))', failure, 'Object cannot be represented as a 64-bit integer.');
  Test('integer(struct("value": "cat"))', failure, 'Object cannot be represented as a 64-bit integer.');

  Test('RealNumber(struct())', failure, 'An empty container cannot be converted to a number.');
  Test('RealNumber(Jane)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('RealNumber(struct("value": 42))', 42.0);
  Test('RealNumber(struct("value": 42/43))', 42/43);
  Test('RealNumber(struct("value": 42.5))', 42.5);
  Test('RealNumber(struct("value": 42.5 + i))', failure, 'Complex number 42.5 + i cannot be converted to a real number.');
  Test('RealNumber(struct("value": "cat"))', failure, 'Cannot convert string "cat" to a real number.');

  Test('ComplexNumber(struct())', failure, 'An empty container cannot be converted to a number.');
  Test('ComplexNumber(Jane)', failure, 'A container with more than a single value cannot be converted to a number.');
  Test('ComplexNumber(struct("value": 42))', TASC(42.0));
  Test('ComplexNumber(struct("value": 42/43))', TASC(42/43));
  Test('ComplexNumber(struct("value": 42.5))', TASC(42.5));
  Test('ComplexNumber(struct("value": 42.5 + i))', 42.5 + ImaginaryUnit);
  Test('ComplexNumber(struct("value": "cat"))', failure, 'Cannot convert string "cat" to a complex number.');

  Test('boolean(struct())', False);
  Test('boolean(Jane)', True);
  Test('boolean(Mary)', True);
  Test('boolean(struct("value": 42))', True);

  Test('string(struct())', '');
  Test('string(Jane)', 'name.first: "Jane"'#13#10'name.last: "Smith"'#13#10'sex: "female"'#13#10'age: 40');

  Test('vector(struct())', TAlgosimRealVector.Create);
  Test('vector(Jane)', [40]);
  Test('vector(Mary)', TAlgosimRealVector.Create);
  Test('vector(struct("value": 42))', [42]);
  Test('vector(struct("alpha": 42, "beta": 10.3, "gamma": 1/2))', [42, 10.3, 0.5]);
  Test('vector(struct("alpha": 42, "beta": 10.3, "gamma": 1/2, "delta": 1 + i))', [42, 10.3, 0.5, 1 + ImaginaryUnit]);

  Test('RealVector(struct())', TAlgosimRealVector.Create);
  Test('RealVector(Jane)', [40]);
  Test('RealVector(Mary)', TAlgosimRealVector.Create);
  Test('RealVector(struct("value": 42))', [42]);
  Test('RealVector(struct("alpha": 42, "beta": 10.3, "gamma": 1/2))', [42, 10.3, 0.5]);
  Test('RealVector(struct("alpha": 42, "beta": 10.3, "gamma": 1/2; "delta": 1 + i))', failure, 'Complex number 1 + i cannot be converted to a real number.');

  Test('ComplexVector(struct())', TAlgosimComplexVector.Create);
  Test('ComplexVector(Jane)', [TASC(40)]);
  Test('ComplexVector(Mary)', TAlgosimComplexVector.Create);
  Test('ComplexVector(struct("value": 42))', [TASC(42)]);
  Test('ComplexVector(struct("alpha": 42, "beta": 10.3, "gamma": 1/2))', [TASC(42), 10.3, 0.5]);
  Test('ComplexVector(struct("alpha": 42, "beta": 10.3, "gamma": 1/2, "delta": 1 + i))', [42, 10.3, 0.5, 1 + ImaginaryUnit]);

  Test('matrix(Mary)', failure, 'An object of type "structure" cannot be converted to a real matrix.');
  Test('RealMatrix(Mary)', failure, 'An object of type "structure" cannot be converted to a real matrix.');
  Test('ComplexMatrix(Mary)', failure, 'An object of type "structure" cannot be converted to a complex matrix.');

  Test('ToList(struct("alpha": 42, "beta": 10.3, "gamma": "dog"))', asoarr([42, 10.3, 'dog']));
  Test('ToList(struct("alpha": 42))', asoarr([42]));
  Test('ToList(struct())', asoarr([]));
  Test('ToList(Jane)', asoarrex([Name('Jane', 'Smith'), ASO('female'), ASO(40)]));

  Test('ToSet(struct("alpha": 42, "beta": 10.3, "gamma": "dog"))', asoset([42, 10.3, 'dog']));
  Test('ToSet(struct("alpha": 42))', asoset([42]));
  Test('ToSet(struct())', asoset([]));
  Test('ToSet(Jane)', asosetex([Name('Jane', 'Smith'), ASO('female'), ASO(40)]));

  Test('delete(name); delete(person); delete(Jane); delete(Mike); delete(Mary)', success);

  Test('color(GetRGB(rgb(1, 0, 0)))', ASO(TRGB.Create(1, 0, 0)));
  Test('color(GetRGB(rgb(0, 1, 0)))', ASO(TRGB.Create(0, 1, 0)));
  Test('color(GetRGB(rgb(0, 0, 1)))', ASO(TRGB.Create(0, 0, 1)));
  Test('color(GetRGB(rgb(1, 1, 0)))', ASO(TRGB.Create(1, 1, 0)));
  Test('color(GetRGB(rgb(1, 1, 1)))', ASO(TRGB.Create(1, 1, 1)));

  Test('color(GetHSV(rgb(1, 0, 0)))', ASO(TRGB.Create(1, 0, 0)));
  Test('color(GetHSV(rgb(0, 1, 0)))', ASO(TRGB.Create(0, 1, 0)));
  Test('color(GetHSV(rgb(0, 0, 1)))', ASO(TRGB.Create(0, 0, 1)));
  Test('color(GetHSV(rgb(1, 1, 0)))', ASO(TRGB.Create(1, 1, 0)));
  Test('color(GetHSV(rgb(1, 1, 1)))', ASO(TRGB.Create(1, 1, 1)));

  Test('color(GetHSL(rgb(1, 0, 0)))', ASO(TRGB.Create(1, 0, 0)));
  Test('color(GetHSL(rgb(0, 1, 0)))', ASO(TRGB.Create(0, 1, 0)));
  Test('color(GetHSL(rgb(0, 0, 1)))', ASO(TRGB.Create(0, 0, 1)));
  Test('color(GetHSL(rgb(1, 1, 0)))', ASO(TRGB.Create(1, 1, 0)));
  Test('color(GetHSL(rgb(1, 1, 1)))', ASO(TRGB.Create(1, 1, 1)));

  Test('color(now())', failure, 'Invalid colour coordinates structure.');


  //
  // From colour
  //

  Test('c ≔ color("red")', ASOColor(clRed));
  Test('number(c)', failure, 'An object of type color cannot be converted to a real number.');
  Test('integer(c)', $FF0000);
  Test('RealNumber(c)', failure, 'An object of type color cannot be converted to a real number.');
  Test('ComplexNumber(c)', failure, 'An object of type color cannot be converted to a complex number.');
  Test('boolean(c)', True);
  Test('string(c)', 'rgba(1, 0, 0, 1)');
  Test('vector(c)', failure, 'An object of type "color" cannot be converted to a real vector.');
  Test('RealVector(c)', failure, 'An object of type "color" cannot be converted to a real vector.');
  Test('ComplexVector(c)', failure, 'An object of type "color" cannot be converted to a complex vector.');
  Test('matrix(c)', failure, 'An object of type "color" cannot be converted to a real matrix.');
  Test('RealMatrix(c)', failure, 'An object of type "color" cannot be converted to a real matrix.');
  Test('ComplexMatrix(c)', failure, 'An object of type "color" cannot be converted to a complex matrix.');
  Test('ToList(c)', asoarrex([ASOColor(clRed)]));
  Test('ToSet(c)', asosetex([ASOColor(clRed)]));
  Test('color(c)', ASOColor(clRed));





  //
  //
  //  CHAPTER 11
  //  System functions and miscellaneous functions.
  //
  //

  Chapter('System functions');

  //
  // Identity function
  //

  Test('identity(0)', 0);
  Test('identity(1/2-1/2)', rat(0, 1));
  Test('identity(0.0)', 0.0);
  Test('identity(i-i)', TASC(0));
  Test('identity(123)', 123);
  Test('identity(-123)', -123);
  Test('identity(123.456)', 123.456);
  Test('identity(1E50)', 1E50);
  Test('identity(true)', True);
  Test('identity(false)', False);
  Test('identity("cat")', 'cat');
  Test('identity("")', '');
  Test('identity(DebugObject("empty vector"))', TAlgosimRealVector.Create);
  Test('identity(ComplexVector(DebugObject("empty vector")))', TAlgosimComplexVector.Create);
  Test('identity(❨1, 2, 7, 5❩)', [1, 2, 7, 5]);
  Test('identity(❨21❩)', [21]);
  Test('identity(❨0❩)', [0]);
  Test('identity(❨0, 0, 0❩)', [0, 0, 0]);
  Test('identity(❨5, 0, i, 1❩)', [5, 0, ImaginaryUnit, 1]);
  Test('identity(❨❨4, 1❩, ❨2, 0❩❩)', 2, [4, 1, 2, 0]);
  Test('identity(❨❨4, 1, 7❩, ❨4, 2, 0❩❩)', 3, [4, 1, 7, 4, 2, 0]);
  Test('identity(❨❨4, 1, 7❩, ❨1 + i, 2, 0❩❩)', 3, [4, 1, 7, 1 + ImaginaryUnit, 2, 0]);
  Test('identity(❨❨394❩❩)', 1, [394]);
  Test('identity(DebugObject("empty matrix"))', TAlgosimRealMatrix.Create);
  Test('identity(ComplexMatrix(DebugObject("empty matrix")))', TAlgosimComplexMatrix.Create);
  Test('identity(''(5, "cat", true, 3.1415, 2, 5, 3))', asoarr([5, 'cat', true, 3.1415, 2, 5, 3]));
  Test('identity(''())', asoarr([]));
  Test('identity({5, "cat", true, 3.1415, 2, 5, 3})', asoset([5, 'cat', true, 3.1415, 2, 3]));
  Test('identity({})', asoset([]));
  Test('identity(0;)', null);
  Test('identity(''(''(''(''(''(1, 2), "cat"))), 4))',
    asoarrex([asoarrex([asoarrex([asoarrex([asoarrex([ASOInt(1), ASOInt(2)]), ASO('cat')])])]), ASOInt(4)]));


  //
  // The compute function
  //

  Test('compute(n, n, 1, 10)', intarr([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]));
  Test('compute(n^2, n, 1, 10)', intarr([1, 4, 9, 16, 25, 36, 49, 64, 81, 100]));
  Test('compute(prime(n), n, 1, 10)', intarr([2, 3, 5, 7, 11, 13, 17, 19, 23, 29]));
  Test('compute(2^n, n, 0, 10)', intarr([1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]));

  Test('matrix(compute(a⋅b, a, 1, 10, b, 1, 10))', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]
  );

  Test('compute(a⋅b⋅c, a, 1, 2, b, 1, 3, c, 1, 4)',
    asoarrex(
      [

        // a = 1
        asoarrex(
          [
            // b = 1
            intarr([1, 2, 3, 4]),
            // b = 2
            intarr([2, 4, 6, 8]),
            // b = 3
            intarr([3, 6, 9, 12])
          ]
        ),

        // a = 2
        asoarrex(
          [
            // b = 1
            intarr([2, 4, 6, 8]),
            // b = 2
            intarr([4, 8, 12, 16]),
            // b = 3
            intarr([6, 12, 18, 24])
          ]
        )

      ]
    )
  );

  Test('compute(n, n, 1, 1)', intarr([1]));
  Test('compute(n, n, 21, 21)', intarr([21]));
  Test('compute(n, n, 1, 0)', intarr([]));
  Test('compute(n, n, 21, 0)', intarr([]));
  Test('compute(n, n, 21, 10)', intarr([]));

  Test('compute(a + b, a, 5, 5, b, 9, 9)', asoarrex([intarr([14])]));
  Test('compute(a + b, a, 5, 5, b, 9, 10)', asoarrex([intarr([14, 15])]));
  Test('compute(a + b, a, 5, 5, b, 9, 8)', asoarrex([]));
  Test('compute(a + b, a, 5, 4, b, 9, 9)', asoarrex([]));
  Test('compute(a + b, a, 5, -4, b, 9, 1)', asoarrex([]));
  Test('compute(a + b, a, 5, -4, b, 9, 1, c, 10, 20, d, 40, 67, e, 11, 3)', asoarrex([]));

  Test('compute(1, a, 1, 1, b, 1, 1, c, 1, 1, d, 1, 1)', asoarrex([asoarrex([asoarrex([asoarr([1])])])]));
  Test('compute(1, a, 1, 1, b, 1, 1, c, 1, 1, d, 1, 3)', asoarrex([asoarrex([asoarrex([asoarr([1, 1, 1])])])]));
  Test('compute(1 + d, a, 1, 1, b, 1, 1, c, 1, 1, d, 1, 3)', asoarrex([asoarrex([asoarrex([asoarr([2, 3, 4])])])]));
  Test('compute(1 + d, a, 1, 1, b, 1, 1, c, 1, 2, d, 1, 3)', asoarrex([asoarrex([asoarrex([asoarr([2, 3, 4]), asoarr([2, 3, 4])])])]));
  Test('compute(1 + c + d, a, 1, 1, b, 1, 1, c, 1, 2, d, 1, 3)', asoarrex([asoarrex([asoarrex([asoarr([3, 4, 5]), asoarr([4, 5, 6])])])]));
  Test('compute(1 + c + d, a, 1, 0, b, 1, 1, c, 1, 2, d, 1, 3)', asoarrex([]));
  Test('compute(1 + c + d, a, 1, 1, b, 10, 1, c, 1, 2, d, 1, 3)', asoarrex([]));

  Test('compute(1 + c + d, a, 0.5, 1, b, 1, 1, c, 1, 2, d, 1, 3)', failure,
    'An object of type integer was expected as argument 3, but an object of type real number was given.');

  Test('compute(1 + c + d, a, 1, 1, b, 1, 1, c, 1, 2, d, 1, 3.1415)', failure,
    'An object of type integer was expected as argument 13, but an object of type real number was given.');

  Test('compute(c! + d!, a, 1, 1, b, 1, 1, c, 1, 2, d, 1, 3)', asoarrex([asoarrex([asoarrex([asoarr([2, 3, 7]), asoarr([3, 4, 8])])])]));
  Test('compute(c! + d!, a, 1, 1, b, 1, 1, c, -1, 0, d, 1, 3)', failure,
    'A non-negative integer was expected as argument 1, but "-1" was given.');
  Test('ErrorInfo(LastError()).stack', strarr(['factorial', 'add', 'compute', 'LastError']));


  //
  // Errors and error handling
  //

  Test('error("A nargle took your variable.")', failure, 'A nargle took your variable.');
  Test('error("")', failure);
  Test('error()', failure);
  Test('error(0)', failure, '0');

  Test('fail("A nargle took your variable.")', failure, 'A nargle took your variable.');
  Test('fail("")', failure);
  Test('fail()', failure);
  Test('fail(0)', failure, '0');

  Test('assert(1 + 1 = 2, "Addition issue detected.")', null);
  Test('assert(1 + 1 = 3, "Synergy deficiency detected.")', failure, 'Synergy deficiency detected.');

  Test('assert(123, "Integral thing.")', failure, 'An object of type boolean was expected as argument 1, but an object of type integer was given.');

  Test('N ≔ 2; assert(N ≥ 2, "N must be at least 2.")', null);
  Test('N ≔ 1; assert(N ≥ 2, "N must be at least 2.")', failure, 'N must be at least 2.');

  Test('p ≔ 3; assert(IsPrime(p), "p must be prime.")', null);
  Test('p ≔ 4; assert(IsPrime(p), "p must be prime.")', failure, 'p must be prime.');

  Test('L ≔ ''(1, 4, 2); assert(#L > 0, "L must be non-empty.")', null);
  Test('L ≔ ''(); assert(#L > 0, "L must be non-empty.")', failure, 'L must be non-empty.');

  Test('succeeded(5 / 3)', True);
  Test('succeeded(5 / 2)', True);
  Test('succeeded(5 / 1)', True);
  Test('succeeded(5 / 0)', False);

  Test('succeeded(fail("miserably"))', False);

  Test('try(5 / 3)', rat(5, 3));
  Test('try(5 / 2)', rat(5, 2));
  Test('try(5 / 1)', 5);
  Test('try(5 / 0)', null);

  Test('try(5 / 3, 21)', rat(5, 3));
  Test('try(5 / 2, 21)', rat(5, 2));
  Test('try(5 / 1, 21)', 5);
  Test('try(5 / 0, 21)', 21);

  Test('try(5 / 3, 21 / 0)', rat(5, 3));
  Test('try(5 / 2, 21 / 0)', rat(5, 2));
  Test('try(5 / 1, 21 / 0)', 5);
  Test('try(5 / 0, 21 / 0)', failure, 'Division by zero');

  Test('ErrorInfo(5 / 0)',
    ASOErrorInfo('Division by zero', [FCN_Divide]));

  Test('ErrorInfo(5 / 0)',
    stErrorInfo.New(
      [
        sm('reason', ASO('Division by zero')),
        sm('source', ASO('divide')),
        sm('stack', strarr(['divide']))
      ]
    )
  );

  Test('ErrorInfo(2 + 5 / 0)',
    ASOErrorInfo('Division by zero', [FCN_Divide, FCN_Add]));

  Test('ErrorInfo(2 + 5 / 0)',
    stErrorInfo.New(
      [
        sm('reason', ASO('Division by zero')),
        sm('source', ASO('divide')),
        sm('stack', strarr(['divide', 'add']))
      ]
    )
  );

  Test('ErrorInfo(3 + sin(7 + (3 + (-5)!)/2))',
    ASOErrorInfo('A non-negative integer was expected as argument 1, but "-5" was given.', [FCN_Factorial, FCN_Add, FCN_Divide, FCN_Add, FCN_Sin, FCN_Add]));

  Test('ErrorInfo(3 + sin(7 + (3 + (-5)!)/2))',
    stErrorInfo.New(
      [
        sm('reason', ASO('A non-negative integer was expected as argument 1, but "-5" was given.')),
        sm('source', ASO('factorial')),
        sm('stack', strarr(['factorial', 'add', 'divide', 'add', 'sin', 'add']))
      ]
    )
  );

  Test('s ≔ "the emotional range of a teaspoon"', 'the emotional range of a teaspoon');

  Test('ErrorInfo(succ(count(s, c ↦ ord(c))) / 2)',
    ASOErrorInfo('The predicate was expected to return a boolean, but an object of type integer was returned.', [FCN_Count, FCN_Succ, FCN_Divide]));

  Test('ErrorInfo(succ(count(s, c ↦ ord(c))) / 2)',
    stErrorInfo.New(
      [
        sm('reason', ASO('The predicate was expected to return a boolean, but an object of type integer was returned.')),
        sm('source', ASO('count')),
        sm('stack', strarr(['count', 'succ', 'divide']))
      ]
    )
  );

  Test('ErrorInfo(succ(count(s, c ↦ chr(ord(c + 5)))) / 2)',
    ASOErrorInfo('A character was expected as argument 1, but "t5" was given.', [FCN_ChrCode, FCN_Character, FCN_Count, FCN_Succ, FCN_Divide]));

  Test('ErrorInfo(succ(count(s, c ↦ chr(ord(c + 5)))) / 2)',
    stErrorInfo.New(
      [
        sm('reason', ASO('A character was expected as argument 1, but "t5" was given.')),
        sm('source', ASO('ord')),
        sm('stack', strarr(['ord', 'chr', 'count', 'succ', 'divide']))
      ]
    )
  );

  Test('ErrorInfo(π + (1/2)⋅∫(sin(x)^2/x + 1, x, 0, π))',
    ASOErrorInfo('Invalid floating point operation', [FCN_Divide, FCN_Add, FCN_Integrate, FCN_Multiply, FCN_Add]));

  Test('ErrorInfo(π + (1/2)⋅∫(sin(x)^2/x + 1, x, 0, π))',
    stErrorInfo.New(
      [
        sm('reason', ASO('Invalid floating point operation')),
        sm('source', ASO('divide')),
        sm('stack', strarr(['divide', 'add', 'integrate', 'multiply', 'add']))
      ]
    )
  );

  Test('5 / 0', failure);
  Test('ErrorInfo(LastError())',
    stErrorInfo.New(
      [
        sm('reason', ASO('Division by zero')),
        sm('source', ASO('divide')),
        sm('stack', strarr(['divide', 'LastError']))
      ]
    )
  );

  Test('2 + 5 / 0', failure);
  Test('ErrorInfo(LastError())',
    stErrorInfo.New(
      [
        sm('reason', ASO('Division by zero')),
        sm('source', ASO('divide')),
        sm('stack', strarr(['divide', 'add', 'LastError']))
      ]
    )
  );

  Test('3 + sin(7 + (3 + (-5)!)/2)', failure);
  Test('ErrorInfo(LastError())',
    stErrorInfo.New(
      [
        sm('reason', ASO('A non-negative integer was expected as argument 1, but "-5" was given.')),
        sm('source', ASO('factorial')),
        sm('stack', strarr(['factorial', 'add', 'divide', 'add', 'sin', 'add', 'LastError']))
      ]
    )
  );

  Test('succ(count(s, c ↦ ord(c))) / 2', failure);
  Test('ErrorInfo(LastError())',
    stErrorInfo.New(
      [
        sm('reason', ASO('The predicate was expected to return a boolean, but an object of type integer was returned.')),
        sm('source', ASO('count')),
        sm('stack', strarr(['count', 'succ', 'divide', 'LastError']))
      ]
    )
  );

  Test('succ(count(s, c ↦ chr(ord(c + 5)))) / 2', failure);
  Test('ErrorInfo(LastError())',
    stErrorInfo.New(
      [
        sm('reason', ASO('A character was expected as argument 1, but "t5" was given.')),
        sm('source', ASO('ord')),
        sm('stack', strarr(['ord', 'chr', 'count', 'succ', 'divide', 'LastError']))
      ]
    )
  );

  Test('π + (1/2)⋅∫(sin(x)^2/x + 1, x, 0, π)', failure);
  Test('ErrorInfo(LastError())',
    stErrorInfo.New(
      [
        sm('reason', ASO('Invalid floating point operation')),
        sm('source', ASO('divide')),
        sm('stack', strarr(['divide', 'add', 'integrate', 'multiply', 'add', 'LastError']))
      ]
    )
  );

  Test('delete(N); delete(p); delete(L); delete(s)', success);


  //
  // Object store and function manager access
  //

  // Object store

  Test('VarBase ≔ ToSet(variables()) ∪ { "VarBase" };', null);
  Test('{ "π", "true", "false", "i", "e", "ℎ", "ℏ", "∅", "ans" } ⊂ VarBase', True);
  Test('description(π)', 'The ratio of a circle''s circumference to its diameter.');
  Test('description(e)', 'The base of the natural logarithm.');
  Test('description(ans)', 'This is the result of the last computation.');
  Test('description(i)', 'The imaginary unit, a square root of negative one.');
  Test('description(∅)', 'The empty set.');
  Test('description(ℎ)', 'The Planck constant.');
  Test('description(ℏ)', 'The Planck constant divided by 2π.');

  Test('"InnerDiameter" ∈ VarBase', False);

  Test('InnerDiameter ≔ 1.7', 1.7);
  Test('VarBase ∪ { "InnerDiameter" } = ToSet(variables())', True);

  Test('var ≔ variable(InnerDiameter);', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('#var', 5);
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.7);
  Test('var.protected', False);

  Test('var ≔ variable("InnerDiameter");', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.7);
  Test('var.protected', False);

  Test('VarBase ∪ { "InnerDiameter", "var" } = ToSet(variables())', True);

  Test('InnerDiameter ≔ 1.8', 1.8);

  Test('var ≔ variable(InnerDiameter);', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.8);
  Test('var.protected', False);

  Test('var ≔ variable("InnerDiameter");', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.8);
  Test('var.protected', False);

  Test('object(InnerDiameter)', 1.8);
  Test('object("InnerDiameter")', 1.8);

  Test('protect(InnerDiameter)', success);

  Test('var ≔ variable(InnerDiameter);', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.8);
  Test('var.protected', True);

  Test('var ≔ variable("InnerDiameter");', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.8);
  Test('var.protected', True);

  Test('object(InnerDiameter)', 1.8);
  Test('object("InnerDiameter")', 1.8);

  Test('InnerDiameter ≔ 1.9', failure, 'Cannot modify or delete protected variable "InnerDiameter".');

  Test('var ≔ variable(InnerDiameter);', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.8);
  Test('var.protected', True);

  Test('var ≔ variable("InnerDiameter");', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.8);
  Test('var.protected', True);

  Test('object(InnerDiameter)', 1.8);
  Test('object("InnerDiameter")', 1.8);

  Test('InnerDiameter', 1.8);

  Test('unprotect(InnerDiameter)', success);

  Test('var ≔ variable(InnerDiameter);', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.8);
  Test('var.protected', False);

  Test('var ≔ variable("InnerDiameter");', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.8);
  Test('var.protected', False);

  Test('object(InnerDiameter)', 1.8);
  Test('object("InnerDiameter")', 1.8);

  Test('InnerDiameter ≔ 1.9', 1.9);

  Test('var ≔ variable(InnerDiameter);', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.9);
  Test('var.protected', False);

  Test('var ≔ variable("InnerDiameter");', null);
  Test('type(var)', 'structure of type "Variable"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', 1.9);
  Test('var.protected', False);

  Test('object(InnerDiameter)', 1.9);
  Test('object("InnerDiameter")', 1.9);

  Test('InnerDiameter ≔ 1.7', 1.7);

  Test('var ≔ metadata(InnerDiameter);', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('#var', 4);
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', failure, 'There is no member named "value".');
  Test('var.protected', False);

  Test('var ≔ metadata("InnerDiameter");', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.value', failure, 'There is no member named "value".');
  Test('var.protected', False);

  Test('object(InnerDiameter)', 1.7);
  Test('object("InnerDiameter")', 1.7);

  Test('InnerDiameter ≔ 1.8', 1.8);

  Test('var ≔ metadata(InnerDiameter);', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('#var', 4);
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', False);

  Test('var ≔ metadata("InnerDiameter");', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', False);

  Test('object(InnerDiameter)', 1.8);
  Test('object("InnerDiameter")', 1.8);

  Test('protect(InnerDiameter)', success);

  Test('var ≔ metadata(InnerDiameter);', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('#var', 4);
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', True);

  Test('var ≔ metadata("InnerDiameter");', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', True);

  Test('object(InnerDiameter)', 1.8);
  Test('object("InnerDiameter")', 1.8);

  Test('InnerDiameter ≔ 1.9', failure, 'Cannot modify or delete protected variable "InnerDiameter".');

  Test('var ≔ metadata(InnerDiameter);', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('#var', 4);
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', True);

  Test('var ≔ metadata("InnerDiameter");', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', True);

  Test('object(InnerDiameter)', 1.8);
  Test('object("InnerDiameter")', 1.8);
  Test('InnerDiameter', 1.8);

  Test('unprotect(InnerDiameter)', success);

  Test('var ≔ metadata(InnerDiameter);', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('#var', 4);
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', False);

  Test('var ≔ metadata("InnerDiameter");', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', False);

  Test('object(InnerDiameter)', 1.8);
  Test('object("InnerDiameter")', 1.8);

  Test('InnerDiameter ≔ 1.9', 1.9);

  Test('var ≔ metadata(InnerDiameter);', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('#var', 4);
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', False);

  Test('var ≔ metadata("InnerDiameter");', null);
  Test('type(var)', 'structure of type "VariableMetadata"');
  Test('var.name', 'InnerDiameter');
  Test('var.description', '');
  Test('var.protected', False);

  Test('object(InnerDiameter)', 1.9);
  Test('object("InnerDiameter")', 1.9);

  Test('variable(nargle)', failure, 'Unknown identifier "nargle".');
  Test('variable("nargle")', failure, 'Unknown identifier "nargle".');
  Test('metadata(nargle)', failure, 'Unknown identifier "nargle".');
  Test('metadata("nargle")', failure, 'Unknown identifier "nargle".');
  Test('object(nargle)', failure, 'Unknown identifier "nargle".');
  Test('object("nargle")', failure, 'Unknown identifier "nargle".');
  Test('description(nargle)', failure, 'Unknown identifier "nargle".');
  Test('description("nargle")', failure, 'Unknown identifier "nargle".');

  Test('variable()', failure, '1 argument(s) expected, but 0 given.');
  Test('variable("")', failure, 'Unknown identifier "".');
  Test('metadata()', failure, '1 argument(s) expected, but 0 given.');
  Test('metadata("")', failure, 'Unknown identifier "".');
  Test('object()', failure, '1 argument(s) expected, but 0 given.');
  Test('object("")', failure, 'Unknown identifier "".');
  Test('description()', failure, '1 argument(s) expected, but 0 given.');
  Test('description("")', failure, 'Unknown identifier "".');

  Test('ForEach(variables(), v, assert(object(+v) = variable(+v).value))', null);
  Test('ForEach(variables(), v, assert(description(+v) = variable(+v).description))', null);

  Test('VarBase ∪ { "InnerDiameter", "var" } = ToSet(variables())', True);
  Test('delete(var)', success);
  Test('VarBase ∪ { "InnerDiameter" } = ToSet(variables())', True);
  Test('delete(InnerDiameter)', success);
  Test('VarBase = ToSet(variables())', True);
  Test('delete(VarBase)', success);

  Test('ForEach(variables(), v, assert(object(+v) = variable(+v).value))', null);
  Test('ForEach(variables(), v, assert(description(+v) = variable(+v).description))', null);

  TestDefVars;
  Test('LoadDefVars()', success);
  Test('ForEach(variables(), v, (unprotect(+v); delete(+v)))', null);
  Test('e', failure, 'Unknown identifier "e".');
  Test('π', failure, 'Unknown identifier "π".');
  Test('i', failure, 'Unknown identifier "i".');
  Test('true', failure, 'Unknown identifier "true".');
  Test('LoadDefVars()', success);
  TestDefVars;
  Test('LoadDefVars()', success);
  TestDefVars;


  // Function manager

  Test('F ≔ functions(); S ≔ ToSet(F) @ function;', null);
  Test('type(F)', 'array');
  Test('∀(F, f ↦ type(f) = "string")', True);
  Test('F = sort(F)', True);
  Test('∀(S, f ↦ type(f) = "kernel function")', True);
  Test('#S ≤ #F', True);
  Test('{ sin, cos, tan, cot, sec, csc } ⊂ S', True);
  Test('{ sinh, cosh, tanh, coth, sech, csch } ⊂ S', True);
  Test('{ arcsin, arccos, arctan, arccot, arcsec, arccsc } ⊂ S', True);
  Test('{ arcsinh, arccosh, arctanh, arccoth, arcsech, arccsch } ⊂ S', True);
  Test('∀(F, f ↦ #categories(f) > 0)', True);
  Test('∀(S, f ↦ #categories(f) > 0)', True);
  Test('categories(sin)', strarr(['math', 'trigonometry']));
  Test('categories("sin")', strarr(['math', 'trigonometry']));
  Test('categories(sind)', failure, 'Unknown identifier "sind".');
  Test('categories("sind")', failure, 'Unknown identifier "sind".');
  Test('FcnName(sin)', 'sin');
  Test('FcnName("sin")', 'sin');
  Test('FcnNames(sin)', strarr(['sin']));
  Test('FcnNames("sin")', strarr(['sin']));
  Test('FcnName(length)', 'length');
  Test('FcnName("length")', 'length');
  Test('FcnName(cardinality)', 'length');
  Test('FcnName("cardinality")', 'length');
  Test('FcnName(card)', 'length');
  Test('FcnName("card")', 'length');
  Test('FcnNames(length)', strarr(['length', 'cardinality', 'card']));
  Test('FcnNames("length")', strarr(['length', 'cardinality', 'card']));
  Test('FcnNames(cardinality)', strarr(['length', 'cardinality', 'card']));
  Test('FcnNames("cardinality")', strarr(['length', 'cardinality', 'card']));
  Test('FcnNames(card)', strarr(['length', 'cardinality', 'card']));
  Test('FcnNames("card")', strarr(['length', 'cardinality', 'card']));
  Test('ForEach(F, f, assert(first(FcnNames(f)) = FcnName(f), FcnName(f) ~ " failed"))', null);
  Test('function(square)(5)', 25);
  Test('function("square")(5)', 25);
  Test('function(factorial)(5)', 120);
  Test('function("factorial")(5)', 120);
  Test('function(NargleShoeSize)("Betty")', failure, 'Unknown identifier "NargleShoeSize".');
  Test('function("NargleShoeSize")("Betty")', failure, 'Unknown identifier "NargleShoeSize".');
  Test('sin = sin', True);
  Test('sin = cos', False);
  Test('sin = tan', False);
  Test('cos = sin', False);
  Test('cos = cos', True);
  Test('cos = tan', False);
  Test('tan = sin', False);
  Test('tan = cos', False);
  Test('tan = tan', True);
  Test('length = length', True);
  Test('length = cardinality', True);
  Test('length = card', True);
  Test('cardinality = length', True);
  Test('cardinality = cardinality', True);
  Test('cardinality = card', True);
  Test('card = length', True);
  Test('card = cardinality', True);
  Test('card = card', True);
  Test('delete(F); delete(S)', success);

  Test('#operators()', 71);


  //
  // The beep function
  //

  Test('beep()', null);
  Test('beep("default")', null);
  Test('beep("information")', null);
  Test('beep("warning")', null);
  Test('beep("error")', null);
  Test('beep("question")', null);
  Test('beep("nargle noise")', failure, 'Unsupported beep sound type: "nargle noise"');

  StartStopwatch;
  Test('beep(200, 200)', null);
  StopStopwatch;
  TestDuration(0.19, 0.21);

  StartStopwatch;
  Test('beep(300, 200)', null);
  StopStopwatch;
  TestDuration(0.19, 0.21);

  StartStopwatch;
  Test('beep(400, 100)', null);
  StopStopwatch;
  TestDuration(0.09, 0.11);

  StartStopwatch;
  Test('beep(500, 100)', null);
  StopStopwatch;
  TestDuration(0.09, 0.11);

  StartStopwatch;
  Test('beep(600, 50)', null);
  StopStopwatch;
  TestDuration(0.04, 0.06);

  StartStopwatch;
  Test('beep(800, 50)', null);
  StopStopwatch;
  TestDuration(0.04, 0.06);

  StartStopwatch;
  Test('beep(1000, 50)', null);
  StopStopwatch;
  TestDuration(0.04, 0.06);

  StartStopwatch;
  Test('beep(2000, 50)', null);
  StopStopwatch;
  TestDuration(0.04, 0.06);

  StartStopwatch;
  Test('beep(4000, 50)', null);
  StopStopwatch;
  TestDuration(0.04, 0.06);

  StartStopwatch;
  Test('beep(100, 1000)', null);
  StopStopwatch;
  TestDuration(0.9, 1.1);

  StartStopwatch;
  Test('beep(4000, 0)', null);
  StopStopwatch;
  TestDuration(0.01);

  Test('beep(4000, -100)', failure, 'A non-negative integer was expected as argument 2, but "-100" was given.');

  Test('beep(10, 1000)', failure, 'Unsupported tone frequency: 10 Hz');
  Test('beep(50000, 1000)', failure, 'Unsupported tone frequency: 50000 Hz');


  //
  // The wait function
  //

  StartStopwatch;
  Test('wait(0.5)', null);
  StopStopwatch;
  TestDuration(0.4, 0.6);

  StartStopwatch;
  Test('wait(0.2)', null);
  StopStopwatch;
  TestDuration(0.1, 0.3);

  StartStopwatch;
  Test('wait(1)', null);
  StopStopwatch;
  TestDuration(0.9, 1.1);


  //
  // The TickCount function
  //

  Test('c ≔ TickCount();', null);
  Sleep(200);
  Test('abs(TickCount() - c - 200) < 100', True);

  Test('c ≔ TickCount();', null);
  Sleep(400);
  Test('abs(TickCount() - c - 400) < 100', True);

  Test('c ≔ TickCount();', null);
  Sleep(1000);
  Test('abs(TickCount() - c - 1000) < 100', True);

  Test('delete(c)', success);


  //
  // Colour functions
  //

  Test('type(rgb(1, 0, 0))', 'color');

  Test('rgb(1, 0, 0)', TRGB.Create(1, 0, 0));
  Test('rgb(0, 1, 0)', TRGB.Create(0, 1, 0));
  Test('rgb(0, 0, 1)', TRGB.Create(0, 0, 1));
  Test('rgb(0, 0, 0)', TRGB.Create(0, 0, 0));
  Test('rgb(1, 1, 1)', TRGB.Create(1, 1, 1));
  Test('rgb(0.5, 0.7, 0.1)', TRGB.Create(0.5, 0.7, 0.1));

  Test('type(hsv(240, 1, 1))', 'color');

  Test('hsv(240, 1, 1)', THSV.Create(240, 1, 1));
  Test('hsv(180, 1, 1)', THSV.Create(180, 1, 1));
  Test('hsv(10, 0.5, 1)', THSV.Create(10, 0.5, 1));
  Test('hsv(10, 1, 0.4)', THSV.Create(10, 1, 0.4));

  Test('type(hsl(240, 1, 0.5))', 'color');

  Test('hsl(240, 1, 0.5)', Thsl.Create(240, 1, 0.5));
  Test('hsl(180, 1, 0.5)', Thsl.Create(180, 1, 0.5));
  Test('hsl(10, 0.5, 0.6)', Thsl.Create(10, 0.5, 0.6));
  Test('hsl(10, 1, 0.4)', Thsl.Create(10, 1, 0.4));

  Test('black ≔ rgb(0, 0, 0); w ≔ rgb(1, 1, 1); r ≔ rgb(1, 0, 0); g ≔ rgb(0, 1, 0); b ≔ rgb(0, 0, 1); y ≔ rgb(1, 1, 0); x ≔ rgb(0.2, 0.4, 0.6);', null);

  Test('GetRGB(black)', ASORGB(TRGB.Create(0, 0, 0)));
  Test('GetRGB("black")', ASORGB(TRGB.Create(0, 0, 0)));
  Test('GetRGB(w)', ASORGB(TRGB.Create(1, 1, 1)));
  Test('GetRGB("white")', ASORGB(TRGB.Create(1, 1, 1)));
  Test('GetRGB("red")', ASORGB(TRGB.Create(1, 0, 0)));
  Test('GetRGB(r)', ASORGB(TRGB.Create(1, 0, 0)));
  Test('GetRGB(g)', ASORGB(TRGB.Create(0, 1, 0)));
  Test('GetRGB(b)', ASORGB(TRGB.Create(0, 0, 1)));
  Test('GetRGB(y)', ASORGB(TRGB.Create(1, 1, 0)));
  Test('GetRGB(x)', ASORGB(TRGB.Create(0.2, 0.4, 0.6)));

  Test('type(GetRGB(r))', 'structure of type "RGB"');

  Test('GetHSV(black)', ASOHSV(THSV.Create(0, 0, 0)));
  Test('GetHSV("black")', ASOHSV(THSV.Create(0, 0, 0)));
  Test('GetHSV(w)', ASOHSV(THSV.Create(0, 0, 1)));
  Test('GetHSV("white")', ASOHSV(THSV.Create(0, 0, 1)));
  Test('GetHSV(r)', ASOHSV(THSV.Create(0, 1, 1)));
  Test('GetHSV("red")', ASOHSV(THSV.Create(0, 1, 1)));
  Test('GetHSV(g)', ASOHSV(THSV.Create(120, 1, 1)));
  Test('GetHSV(b)', ASOHSV(THSV.Create(240, 1, 1)));
  Test('GetHSV(y)', ASOHSV(THSV.Create(60, 1, 1)));
  Test('GetHSV(x)', ASOHSV(THSV.Create(210, 2/3, 0.6)));

  Test('type(GetHSV(r))', 'structure of type "HSV"');

  Test('GetHSL(black)', ASOHSL(THSL.Create(0, 0, 0)));
  Test('GetHSL("black")', ASOHSL(THSL.Create(0, 0, 0)));
  Test('GetHSL(w)', ASOHSL(THSL.Create(0, 0, 1)));
  Test('GetHSL("white")', ASOHSL(THSL.Create(0, 0, 1)));
  Test('GetHSL(r)', ASOHSL(THSL.Create(0, 1, 0.5)));
  Test('GetHSL("red")', ASOHSL(THSL.Create(0, 1, 0.5)));
  Test('GetHSL(g)', ASOHSL(THSL.Create(120, 1, 0.5)));
  Test('GetHSL(b)', ASOHSL(THSL.Create(240, 1, 0.5)));
  Test('GetHSL(y)', ASOHSL(THSL.Create(60, 1, 0.5)));
  Eps; Test('GetHSL(x)', ASOHSL(THSL.Create(210, 0.50, 0.4)));

  Test('type(GetHSL(r))', 'structure of type "HSL"');

  Test('black ≔ hsv(0, 0, 0); w ≔ hsv(0, 0, 1); r ≔ hsv(0, 1, 1); g ≔ hsv(120, 1, 1); b ≔ hsv(240, 1, 1); y ≔ hsv(60, 1, 1); x ≔ hsv(210, 2/3, 0.6);', null);

  Test('GetRGB(black)', ASORGB(TRGB.Create(0, 0, 0)));
  Test('GetRGB(w)', ASORGB(TRGB.Create(1, 1, 1)));
  Test('GetRGB(r)', ASORGB(TRGB.Create(1, 0, 0)));
  Test('GetRGB(g)', ASORGB(TRGB.Create(0, 1, 0)));
  Test('GetRGB(b)', ASORGB(TRGB.Create(0, 0, 1)));
  Test('GetRGB(y)', ASORGB(TRGB.Create(1, 1, 0)));
  Test('GetRGB(x)', ASORGB(TRGB.Create(0.2, 0.4, 0.6)));

  Test('GetHSV(black)', ASOHSV(THSV.Create(0, 0, 0)));
  Test('GetHSV(w)', ASOHSV(THSV.Create(0, 0, 1)));
  Test('GetHSV(r)', ASOHSV(THSV.Create(0, 1, 1)));
  Test('GetHSV(g)', ASOHSV(THSV.Create(120, 1, 1)));
  Test('GetHSV(b)', ASOHSV(THSV.Create(240, 1, 1)));
  Test('GetHSV(y)', ASOHSV(THSV.Create(60, 1, 1)));
  Test('GetHSV(x)', ASOHSV(THSV.Create(210, 2/3, 0.6)));

  Test('GetHSL(black)', ASOHSL(THSL.Create(0, 0, 0)));
  Test('GetHSL(w)', ASOHSL(THSL.Create(0, 0, 1)));
  Test('GetHSL(r)', ASOHSL(THSL.Create(0, 1, 0.5)));
  Test('GetHSL(g)', ASOHSL(THSL.Create(120, 1, 0.5)));
  Test('GetHSL(b)', ASOHSL(THSL.Create(240, 1, 0.5)));
  Test('GetHSL(y)', ASOHSL(THSL.Create(60, 1, 0.5)));
  Eps; Test('GetHSL(x)', ASOHSL(THSL.Create(210, 0.50, 0.4)));

  Test('black ≔ hsl(0, 0, 0); w ≔ hsl(0, 0, 1); r ≔ hsl(0, 1, 0.5); g ≔ hsl(120, 1, 0.5); b ≔ hsl(240, 1, 0.5); y ≔ hsl(60, 1, 0.5); x ≔ hsl(210, 0.5, 0.4);', null);

  Eps(1E-13); Test('GetRGB(black)', ASORGB(TRGB.Create(0, 0, 0)));
  Eps(1E-13); Test('GetRGB(w)', ASORGB(TRGB.Create(1, 1, 1)));
  Eps(1E-13); Test('GetRGB(r)', ASORGB(TRGB.Create(1, 0, 0)));
  Eps(1E-13); Test('GetRGB(g)', ASORGB(TRGB.Create(0, 1, 0)));
  Eps(1E-13); Test('GetRGB(b)', ASORGB(TRGB.Create(0, 0, 1)));
  Eps(1E-13); Test('GetRGB(y)', ASORGB(TRGB.Create(1, 1, 0)));
  Eps(1E-13); Test('GetRGB(x)', ASORGB(TRGB.Create(0.2, 0.4, 0.6)));

  Eps(1E-13); Test('GetHSV(black)', ASOHSV(THSV.Create(0, 0, 0)));
  Eps(1E-13); Test('GetHSV(w)', ASOHSV(THSV.Create(0, 0, 1)));
  Eps(1E-13); Test('GetHSV(r)', ASOHSV(THSV.Create(0, 1, 1)));
  Eps(1E-13); Test('GetHSV(g)', ASOHSV(THSV.Create(120, 1, 1)));
  Eps(1E-13); Test('GetHSV(b)', ASOHSV(THSV.Create(240, 1, 1)));
  Eps(1E-13); Test('GetHSV(y)', ASOHSV(THSV.Create(60, 1, 1)));
  Eps(1E-13); Test('GetHSV(x)', ASOHSV(THSV.Create(210, 2/3, 0.6)));

  Eps(1E-13); Test('GetHSL(black)', ASOHSL(THSL.Create(0, 0, 0)));
  Eps(1E-13); Test('GetHSL(w)', ASOHSL(THSL.Create(0, 0, 1)));
  Eps(1E-13); Test('GetHSL(r)', ASOHSL(THSL.Create(0, 1, 0.5)));
  Eps(1E-13); Test('GetHSL(g)', ASOHSL(THSL.Create(120, 1, 0.5)));
  Eps(1E-13); Test('GetHSL(b)', ASOHSL(THSL.Create(240, 1, 0.5)));
  Eps(1E-13); Test('GetHSL(y)', ASOHSL(THSL.Create(60, 1, 0.5)));
  Eps(1E-13); Test('GetHSL(x)', ASOHSL(THSL.Create(210, 0.50, 0.4)));

  Test('color(GetRGB(rgb(0, 0, 0)))', TRGB.Create(0, 0, 0));
  Test('color(GetHSV(rgb(0, 0, 0)))', TRGB.Create(0, 0, 0));
  Test('color(GetHSL(rgb(0, 0, 0)))', TRGB.Create(0, 0, 0));
  Test('color(16#000000)', TRGB.Create(0, 0, 0));
  Test('color("black")', TRGB.Create(0, 0, 0));
  Test('color(color("black"))', TRGB.Create(0, 0, 0));
  Test('HexColorCode(color("black"))', '#000000');

  Test('color(GetRGB(rgb(1, 0, 0)))', TRGB.Create(1, 0, 0));
  Test('color(GetHSV(rgb(1, 0, 0)))', TRGB.Create(1, 0, 0));
  Test('color(GetHSL(rgb(1, 0, 0)))', TRGB.Create(1, 0, 0));
  Test('color(16#0000FF)', TRGB.Create(0, 0, 1));
  Test('color("red")', TRGB.Create(1, 0, 0));
  Test('color(color("red"))', TRGB.Create(1, 0, 0));
  Test('HexColorCode(color("red"))', '#FF0000');

  Test('color(GetRGB(rgb(1, 1, 0)))', TRGB.Create(1, 1, 0));
  Test('color(GetHSV(rgb(1, 1, 0)))', TRGB.Create(1, 1, 0));
  Test('color(GetHSL(rgb(1, 1, 0)))', TRGB.Create(1, 1, 0));
  Test('color(16#00FFFF)', TRGB.Create(0, 1, 1));
  Test('color("yellow")', TRGB.Create(1, 1, 0));
  Test('color(color("yellow"))', TRGB.Create(1, 1, 0));
  Test('HexColorCode(color("yellow"))', '#FFFF00');

  Test('color(GetRGB(rgb(1, 1, 1)))', TRGB.Create(1, 1, 1));
  Test('color(GetHSV(rgb(1, 1, 1)))', TRGB.Create(1, 1, 1));
  Test('color(GetHSL(rgb(1, 1, 1)))', TRGB.Create(1, 1, 1));
  Test('color(16#FFFFFF)', TRGB.Create(1, 1, 1));
  Test('color("white")', TRGB.Create(1, 1, 1));
  Test('color(color("white"))', TRGB.Create(1, 1, 1));
  Test('HexColorCode(color("white"))', '#FFFFFF');

  Test('color("white")', TRGB.Create(1, 1, 1));
  Test('color("#FFFFFF")', TRGB.Create(1, 1, 1));
  Test('color("black")', TRGB.Create(0, 0, 0));
  Test('color("#000000")', TRGB.Create(0, 0, 0));
  Test('color("red")', TRGB.Create(1, 0, 0));
  Test('color("#FF0000")', TRGB.Create(1, 0, 0));
  Test('color("lime")', TRGB.Create(0, 1, 0));
  Test('color("#00FF00")', TRGB.Create(0, 1, 0));
  Test('color("green")', TRGB.Create(0, 128/255, 0));
  Test('color("#008000")', TRGB.Create(0, 128/255, 0));
  Test('color("blue")', TRGB.Create(0, 0, 1));
  Test('color("#0000FF")', TRGB.Create(0, 0, 1));
  Test('color("navy")', TRGB.Create(0, 0, 128/255));
  Test('color("#000080")', TRGB.Create(0, 0, 128/255));
  Test('color("yellow")', TRGB.Create(1, 1, 0));
  Test('color("#FFFF00")', TRGB.Create(1, 1, 0));
  Test('color("teal")', TRGB.Create(0, 128/255, 128/255));
  Test('color("#008080")', TRGB.Create(0, 128/255, 128/255));
  Test('color("grey")', TRGB.Create(128/255, 128/255, 128/255));
  Test('color("#808080")', TRGB.Create(128/255, 128/255, 128/255));

  Test('InvertColor(rgb(0, 0, 0))', TRGB.Create(1, 1, 1));
  Test('InvertColor("black")', TRGB.Create(1, 1, 1));
  Test('InvertColor(rgb(1, 1, 1))', TRGB.Create(0, 0, 0));
  Test('InvertColor("white")', TRGB.Create(0, 0, 0));
  Test('InvertColor(rgb(1, 0, 0))', TRGB.Create(0, 1, 1));
  Test('InvertColor("red")', TRGB.Create(0, 1, 1));
  Test('InvertColor(rgb(0, 1, 0))', TRGB.Create(1, 0, 1));
  Test('InvertColor(rgb(0, 1, 1))', TRGB.Create(1, 0, 0));
  Test('InvertColor(rgb(0.5, 0.5, 0.5))', TRGB.Create(0.5, 0.5, 0.5));
  Test('InvertColor(rgb(0.4, 0.123, 0.72))', TRGB.Create(0.6, 0.877, 0.28));
  Test('InvertColor(rgb(0.3421, 0.7123, 0.1272))', TRGB.Create(0.6579, 0.2877, 0.8728));

  Test('InvertValue(hsv(0, 1, 1))', THSV.Create(0, 1, 0));
  Test('InvertValue("red")', THSV.Create(0, 1, 0));
  Test('InvertValue(hsv(0, 1, 0))', THSV.Create(0, 0, 1));
  Test('InvertValue(hsv(0, 1, 0.5))', THSV.Create(0, 1, 0.5));
  Test('InvertValue(hsv(0, 1, 0.8))', THSV.Create(0, 1, 0.2));
  Test('InvertValue(hsv(120, 0.7, 0.3))', THSV.Create(120, 0.7, 0.7));
  Test('InvertValue(hsv(120, 0.7, 0.4321))', THSV.Create(120, 0.7, 0.5679));
  Test('InvertValue(hsv(240, 0.5, 0.1))', THSV.Create(240, 0.5, 0.9));
  Test('InvertValue(hsv(60, 0.2, 0.99))', THSV.Create(60, 0.2, 0.01));
  Test('InvertValue(hsv(0, 0, 0.2))', THSV.Create(0, 0, 0.8));
  Test('InvertValue(hsv(0, 0, 0))', THSV.Create(0, 0, 1));
  Test('InvertValue(hsv(0, 0, 1))', THSV.Create(0, 0, 0));

  Test('InvertLightness(hsl(0, 0, 0))', THSL.Create(0, 0, 1));
  Test('InvertLightness("black")', THSL.Create(0, 0, 1));
  Test('InvertLightness(hsl(0, 0, 1))', THSL.Create(0, 0, 0));
  Test('InvertLightness("white")', THSL.Create(0, 0, 0));
  Test('InvertLightness(hsl(0, 1, 0.5))', THSL.Create(0, 1, 0.5));
  Test('InvertLightness("red")', THSL.Create(0, 1, 0.5));
  Test('InvertLightness(hsl(0, 1, 0.2))', THSL.Create(0, 1, 0.8));
  Test('InvertLightness(hsl(123, 1, 0.685))', THSL.Create(123, 1, 0.315));
  Test('InvertLightness(hsl(200, 0.7, 0.352))', THSL.Create(200, 0.7, 0.648));

  Test('darken(rgb(0.8, 0.4, 1))', TRGB.Create(0.4, 0.2, 0.5));
  Test('darken(rgb(0.5, 0.7, 0))', TRGB.Create(0.25, 0.35, 0.0));
  Test('darken(rgb(1, 1, 1))', TRGB.Create(0.5, 0.5, 0.5));
  Test('darken("white")', TRGB.Create(0.5, 0.5, 0.5));
  Test('darken(rgb(0, 0, 0))', TRGB.Create(0.0, 0.0, 0.0));
  Test('darken("black")', TRGB.Create(0.0, 0.0, 0.0));

  Test('whiten(rgb(0.5, 0.8, 0.2))', TRGB.Create(0.75, 0.9, 0.6));
  Test('whiten(rgb(0.1, 0.9, 0.4))', TRGB.Create(0.55, 0.95, 0.7));
  Test('whiten(rgb(0.0, 0.0, 1.0))', TRGB.Create(0.5, 0.5, 1.0));
  Test('whiten(rgb(1.0, 1.0, 1.0))', TRGB.Create(1.0, 1.0, 1.0));
  Test('whiten("white")', TRGB.Create(1.0, 1.0, 1.0));
  Test('whiten(rgb(0.0, 0.0, 0.0))', TRGB.Create(0.5, 0.5, 0.5));
  Test('whiten("black")', TRGB.Create(0.5, 0.5, 0.5));

  Test('FadeToColor(rgb(0.3, 0.5, 0.1), rgb(0.6, 0.2, 0.1), 0.0)', TRGB.Create(0.3, 0.5, 0.1));
  Test('FadeToColor(rgb(0.3, 0.5, 0.1), rgb(0.6, 0.2, 0.1), 0.5)', TRGB.Create(0.45, 0.35, 0.1));
  Test('FadeToColor(rgb(0.3, 0.5, 0.1), rgb(0.6, 0.2, 0.1), 1.0)', TRGB.Create(0.6, 0.2, 0.1));

  Test('FixHue(hsv(120, 0.3, 0.7), 200)', THSV.Create(200, 0.3, 0.7));
  Test('FixHue(hsv(240, 0.5, 0.1), 60)', THSV.Create(60, 0.5, 0.1));
  Test('FixHue(hsv(0, 0.72, 0.66), 300)', THSV.Create(300, 0.72, 0.66));

  Test('ToMonochromatic(hsv(120, 0.3, 0.7), 200)', THSV.Create(200, 1, 0.7));
  Test('ToMonochromatic(hsv(240, 0.5, 0.1), 60)', THSV.Create(60, 1, 0.1));
  Test('ToMonochromatic(hsv(0, 0.72, 0.66), 300)', THSV.Create(300, 1, 0.66));

  Test('ShiftHue(hsv(120, 0.3, 0.7), 200)', THSV.Create(320, 0.3, 0.7));
  Test('ShiftHue(hsv(240, 0.5, 0.1), 60)', THSV.Create(300, 0.5, 0.1));
  Test('ShiftHue(hsv(240, 0.5, 0.1), 120)', THSV.Create(0, 0.5, 0.1));
  Test('ShiftHue(hsv(240, 0.5, 0.1), 180)', THSV.Create(60, 0.5, 0.1));
  Test('ShiftHue(hsv(0, 0.72, 0.66), 300)', THSV.Create(300, 0.72, 0.66));

  Test('ToGreyscale(hsv(120, 0.3, 0.7))', THSV.Create(120, 0.0, 0.7));
  Test('ToGreyscale(hsv(240, 0.5, 0.1))', THSV.Create(240, 0.0, 0.1));
  Test('ToGreyscale(hsv(0, 0.72, 0.66))', THSV.Create(0, 0.0, 0.66));
  Test('ToGreyscale(hsv(0, 0.0, 0.66))', THSV.Create(0, 0.0, 0.66));
  Test('ToGreyscale(hsv(0, 0.0, 0.0))', THSV.Create(0, 0.0, 0.0));
  Test('ToGreyscale("black")', THSV.Create(0, 0.0, 0.0));
  Test('ToGreyscale(hsv(0, 0.0, 1.0))', THSV.Create(0, 0.0, 1.0));

  Test('IsDark(rgb(0, 0, 0))', True);
  Test('IsDark("black")', True);
  Test('IsDark(rgb(0.2, 0.1, 0.4))', True);
  Test('IsDark(rgb(0, 0.3, 0.5))', True);
  Test('IsDark(rgb(0.3, 0.4, 0.2))', True);

  Test('IsDark(rgb(1, 1, 1))', False);
  Test('IsDark("white")', False);
  Test('IsDark(rgb(0.8, 1, 1))', False);
  Test('IsDark(rgb(0.8, 0.7, 0.7))', False);
  Test('IsDark(rgb(0.6, 0.8, 0.7))', False);
  Test('IsDark(rgb(0, 1, 1))', False);
  Test('IsDark(rgb(1, 1, 0))', False);
  Test('IsDark(rgb(1, 1, 0.8))', False);
  Test('IsDark(rgb(0.7, 0.5, 1))', False);

  Test('∀(NamedColors(), c ↦ type(c) = "structure of type ""NamedColor""")', True);
  Test('∀(NamedColors(), c ↦ color(c.name) = c.color)', True);
  Test('#NamedColors() ≥ 140', True);

  Test('N, c ≔ 10000, ZeroVector(3); for(i, 1, N, (c ≔ c + vector(GetRGB(RandomColor())))); ∀(c/N, x ↦ 0.48 < x ∧ x < 0.52)', True);
  Test('delete(N); delete(c)', success);


  //
  // Structure functions
  //

  Test('S ≔ struct("name": "Albus Dumbledore", "age": 116)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(116))
      ]
    )
  );

  Test('AddMember(S, "role", "headmaster")',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(116)),
        sm('role', ASO('headmaster'))
      ]
    )
  );

  Test('AddMember(S, "FavouriteColour", color("darkgoldenrod"))',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(116)),
        sm('FavouriteColour', ASOColor(NamedColorsDict['darkgoldenrod']))
      ]
    )
  );

  Test('AddMember(S, "FavouriteColour2", color("goldenrod"))',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(116)),
        sm('FavouriteColour2', ASOColor(NamedColorsDict['goldenrod']))
      ]
    )
  );

  Test('AddMember(S, 1024, color("darkgoldenrod"))',
    failure,
    'An object of type string was expected as argument 2, but an object of type integer was given.'
  );

  Test('AddMember(S, "FavouriteColour", color("darkgoldenrod"), color("goldenrod"))',
    failure,
    'Too many arguments.'
  );

  Test('AddMember(S)',
    failure,
    'Too few arguments. A required argument of type string is missing.'
  );

  Test('AddMember(S, "favourite colour", color("darkgoldenrod"))',
    failure,
    'Invalid member name: "favourite colour".'
  );

  Test('AddMember(S, "", color("darkgoldenrod"))',
    failure,
    'Invalid member name: "".'
  );

  Test('AddMember(S, " ", color("darkgoldenrod"))',
    failure,
    'Invalid member name: " ".'
  );

  Test('AddMember(S, " FavouriteColour", color("darkgoldenrod"))',
    failure,
    'Invalid member name: " FavouriteColour".'
  );

  Test('AddMember(S, """FavouriteColour""", color("darkgoldenrod"))',
    failure,
    'Invalid member name: ""FavouriteColour"".'
  );

  Test('AddMember(S, "+", color("darkgoldenrod"))',
    failure,
    'Invalid member name: "+".'
  );

  Test('AddMember(S, "+1024", color("darkgoldenrod"))',
    failure,
    'Invalid member name: "+1024".'
  );

  Test('AddMember(S, "1024", color("darkgoldenrod"))',
    failure,
    'Invalid member name: "1024".'
  );

  Test('AddMember(S, "age", 117)',
    failure,
    'Structure already contains a member named "age".'
  );

  Test('AddMember("S", "age", 117)',
    failure,
    'An object of type structure was expected as argument 1, but an object of type string was given.'
  );

  Test('RemoveMember(S, "age")',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore'))
      ]
    )
  );

  Test('RemoveMember(S, "name")',
    TAlgosimStructure.CreateWithValue(
      [
        sm('age', ASOInt(116))
      ]
    )
  );

  Test('RemoveMember(S, 2)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore'))
      ]
    )
  );

  Test('RemoveMember(S, 1)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('age', ASOInt(116))
      ]
    )
  );

  Test('RemoveMember(S, -1)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore'))
      ]
    )
  );

  Test('RemoveMember(S, -2)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('age', ASOInt(116))
      ]
    )
  );

  Test('RemoveMember(RemoveMember(S, 1), 1)',
    TAlgosimStructure.CreateWithValue(
      [
      ]
    )
  );

  Test('RemoveMember(S, "FavouriteColour")',
    failure,
    'There is no member named "FavouriteColour".'
  );

  Test('RemoveMember(S, 3)',
    failure,
    'Index 3 out of bounds.'
  );

  Test('RemoveMember(S, -3)',
    failure,
    'Index -3 out of bounds.'
  );

  Test('RemoveMember(S, 0)',
    failure,
    'Index 0 out of bounds.'
  );

  Test('RemoveMember(S, true)',
    failure,
    'Invalid function arguments.'
  );

  Test('RemoveMember(S)',
    failure,
    'Too few arguments.'
  );

  Test('SetValue(S, "name", "Albus Percival Wulfric Brian Dumbledore")',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Percival Wulfric Brian Dumbledore')),
        sm('age', ASOInt(116))
      ]
    )
  );

  Test('SetValue(S, "age", 117)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(117))
      ]
    )
  );

  Test('SetValue(S, 1, "Albus Percival Wulfric Brian Dumbledore")',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Percival Wulfric Brian Dumbledore')),
        sm('age', ASOInt(116))
      ]
    )
  );

  Test('SetValue(S, 2, 117)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(117))
      ]
    )
  );

  Test('SetValue(S, -2, "Albus Percival Wulfric Brian Dumbledore")',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Percival Wulfric Brian Dumbledore')),
        sm('age', ASOInt(116))
      ]
    )
  );

  Test('SetValue(S, -1, 117)',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(117))
      ]
    )
  );

  Test('SetValue(S, "FavouriteColour", color("darkgoldenrod"))',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(116)),
        sm('FavouriteColour', ASOColor(NamedColorsDict['darkgoldenrod']))
      ]
    )
  );

  Test('SetValue(S, "", color("gold"))',
    failure,
    'Invalid member name: "".'
  );

  Test('SetValue(S, 3, color("gold"))',
    failure,
    'Index 3 out of bounds.'
  );

  Test('SetValue(S, -3, color("gold"))',
    failure,
    'Index -3 out of bounds.'
  );

  Test('SetValue(S, 0, color("gold"))',
    failure,
    'Index 0 out of bounds.'
  );

  Test('SetValue(S, true, color("gold"))',
    failure,
    'Invalid function arguments.'
  );

  Test('SetValue(S, "name")',
    failure,
    'Too few arguments.'
  );

  Test('SetValue(S)',
    failure,
    'Too few arguments.'
  );

  Test('SetValue()',
    failure,
    'Too few arguments. A required argument of type structure is missing.'
  );

  Test('RenameMember(S, "name", "FullName")',
    TAlgosimStructure.CreateWithValue(
      [
        sm('FullName', ASO('Albus Dumbledore')),
        sm('age', ASOInt(116))
      ]
    )
  );

  Test('RenameMember(S, "age", "Age")',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('Age', ASOInt(116))
      ]
    )
  );

  Test('RenameMember(S, "name", "name")',
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('Albus Dumbledore')),
        sm('age', ASOInt(116))
      ]
    )
  );

  Test('RenameMember(S, "name", "full name")',
    failure,
    'Invalid member name: "full name".'
  );

  Test('RenameMember(S, "name", "")',
    failure,
    'Invalid member name: "".'
  );

  Test('RenameMember(S, "name", "+")',
    failure,
    'Invalid member name: "+".'
  );

  Test('RenameMember(S, "name", " FullName")',
    failure,
    'Invalid member name: " FullName".'
  );

  Test('RenameMember(S, "name", "+name")',
    failure,
    'Invalid member name: "+name".'
  );

  Test('RenameMember(S, "name", """name""")',
    failure,
    'Invalid member name: ""name"".'
  );

  Test('RenameMember(S, "FirstName", "FullName")',
    failure,
    'There is no member named "FirstName".'
  );

  Test('RenameMember(S, "", "FullName")',
    failure,
    'There is no member named "".'
  );

  Test('RenameMember(S, "name", 123)',
    failure,
    'An object of type string was expected as argument 3, but an object of type integer was given.'
  );

  Test('RenameMember(S, "name")',
    failure,
    'Too few arguments. A required argument of type string is missing.'
  );

  Test('RenameMember(S)',
    failure,
    'Too few arguments. A required argument of type string is missing.'
  );

  Test('RenameMember()',
    failure,
    'Too few arguments. A required argument of type structure is missing.'
  );

  Test('RenameMember(123)',
    failure,
    'An object of type structure was expected as argument 1, but an object of type integer was given.'
  );

  Test('HasMember(S, "name")', True);
  Test('HasMember(S, "age")', True);
  Test('HasMember(S, "sex")', False);
  Test('HasMember(S, "role")', False);
  Test('HasMember(S, "")', False);
  Test('HasMember(S, " name")', False);
  Test('HasMember(S, "FullName")', False);

  Test('HasMember(struct("name": "Albus Dumbledore", "age": 116), "name")', True);
  Test('HasMember(struct("name": "Albus Dumbledore", "age": 116), "age")', True);
  Test('HasMember(struct("name": "Albus Dumbledore", "age": 116), "sex")', False);
  Test('HasMember(struct("name": "Albus Dumbledore", "age": 116), "role")', False);
  Test('HasMember(struct("name": "Albus Dumbledore", "age": 116), "")', False);
  Test('HasMember(struct("name": "Albus Dumbledore", "age": 116), " name")', False);
  Test('HasMember(struct("name": "Albus Dumbledore", "age": 116), "FullName")', False);

  Test('HasMember(S, 2)', failure, 'An object of type string was expected as argument 2, but an object of type integer was given.');
  Test('HasMember(true, "name")', failure, 'An object of type structure was expected as argument 1, but an object of type boolean was given.');
  Test('HasMember(S)', failure, '2 argument(s) expected, but 1 given.');
  Test('HasMember()', failure, '2 argument(s) expected, but 0 given.');

  Test('IndexOfName(S, "name")', 1);
  Test('IndexOfName(S, "age")', 2);
  Test('IndexOfName(S, "sex")', 0);
  Test('IndexOfName(S, "role")', 0);
  Test('IndexOfName(S, "")', 0);
  Test('IndexOfName(S, " name")', 0);
  Test('IndexOfName(S, "FullName")', 0);

  Test('IndexOfName(struct("name": "Albus Dumbledore", "age": 116), "name")', 1);
  Test('IndexOfName(struct("name": "Albus Dumbledore", "age": 116), "age")', 2);
  Test('IndexOfName(struct("name": "Albus Dumbledore", "age": 116), "sex")', 0);
  Test('IndexOfName(struct("name": "Albus Dumbledore", "age": 116), "role")', 0);
  Test('IndexOfName(struct("name": "Albus Dumbledore", "age": 116), "")', 0);
  Test('IndexOfName(struct("name": "Albus Dumbledore", "age": 116), " name")', 0);
  Test('IndexOfName(struct("name": "Albus Dumbledore", "age": 116), "FullName")', 0);

  Test('IndexOfName(S, 2)', failure, 'An object of type string was expected as argument 2, but an object of type integer was given.');
  Test('IndexOfName(true, "name")', failure, 'An object of type structure was expected as argument 1, but an object of type boolean was given.');
  Test('IndexOfName(S)', failure, '2 argument(s) expected, but 1 given.');
  Test('IndexOfName()', failure, '2 argument(s) expected, but 0 given.');

  for _st := Low(TStructType) to High(TStructType) do
  begin
    Test(Format('type(GetStructType("%s"))', [_st.Name]), 'structure type');
    Test(Format('GetStructType("%s")', [_st.Name]), _st.CreateTypeObject);
  end;

  Test('GetStructType("NargleData")', failure, 'Unknown structure type "NargleData".');

  with
    TAlgosimStructure.CreateWithValue(
        [
          sm('red', ASO(1)),
          sm('green', ASO(0)),
          sm('blue', ASO(0))
        ]
      )
  do
    try
      Test('GetStructType("RGB")',
        ToStructType('RGB')
      )
    finally
      Free;
    end;

  Test('rgbst ≔ GetStructType("RGB"); hsvst ≔ GetStructType("HSV");', null);
  Test('type(rgbst)', 'structure type');
  Test('type(hsvst)', 'structure type');
  Test('rgbst = rgbst', True);
  Test('hsvst = hsvst', True);
  Test('rgbst = hsvst', False);
  Test('rgbst = GetRGB(rgb(1, 0, 0))', True);
  Test('type(GetRGB(rgb(1, 0, 0)))', 'structure of type "RGB"');

  Test('person ≔ struct("name": "unnamed", "age": 0)',
    TAlgosimStructure.CreateWithValue(
        [
          sm('name', ASO('unnamed')),
          sm('age', ASOInt(0))
        ]
      )
    );

  Test('type(person)', 'structure');

  with
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('unnamed')),
        sm('age', ASOInt(0))
      ]
    )
  do
    try
      Test('pst ≔ ToStructType(person, "person")',
        ToStructType('person')
      );
    finally
      Free;
    end;

  Test('new(pst)',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('name', ASO('unnamed')),
        sm('age', ASOInt(0))
      ]
    )
  );

  Test('type(new(pst))', 'structure of type "person"');

  Test('new(pst, "name": "Severus Snape")',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('name', ASO('Severus Snape')),
        sm('age', ASOInt(0))
      ]
    )
  );

  Test('new(pst, "age": 50)',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('name', ASO('unnamed')),
        sm('age', ASOInt(50))
      ]
    )
  );

  Test('new(pst, "name": "Severus Snape", "age": 50)',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('name', ASO('Severus Snape')),
        sm('age', ASOInt(50))
      ]
    )
  );

  Test('new(pst, "name": "Severus Snape", "age": 50, "sex": "male")',
    failure,
    'Structure type "person" does not contain a member named "sex".'
  );

  Test('new(person, "name": "Severus Snape", "age": 50)',
    failure,
    'An object of type structure type was expected as argument 1, but an object of type structure was given.'
  );

  Test('new(rgbst, "red": 0.4, "green": 0.3, "blue": 0.7)',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('red', ASO(0.4)),
        sm('green', ASO(0.3)),
        sm('blue', ASO(0.7))
      ]
    )
  );

  Test('new(rgbst, "green": 0.3, "red": 0.4, "blue": 0.7)',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('red', ASO(0.4)),
        sm('green', ASO(0.3)),
        sm('blue', ASO(0.7))
      ]
    )
  );

  Test('type(new(rgbst, "red": 0.4, "green": 0.3, "blue": 0.7))',
    'structure of type "RGB"'
  );

  Test('new(rgbst, "red": 0.4, "green": 0.3, "blau": 0.7)',
    failure,
    'Structure type "RGB" does not contain a member named "blau".'
  );

  Test('pst2 ≔ GetStructType("person")',
    failure, 'Unknown structure type "person".');

  Test('RegisterStructType(pst)', success);

  with
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('unnamed')),
        sm('age', ASOInt(0))
      ]
    )
  do
    try
      Test('pst2 ≔ GetStructType("person")',
        ToStructType('person')
      );
    finally
      Free;
    end;

  Test('new(pst2)',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('name', ASO('unnamed')),
        sm('age', ASOInt(0))
      ]
    )
  );

  Test('type(new(pst2))', 'structure of type "person"');

  Test('new(pst2, "name": "Severus Snape")',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('name', ASO('Severus Snape')),
        sm('age', ASOInt(0))
      ]
    )
  );

  Test('new(pst2, "age": 50)',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('name', ASO('unnamed')),
        sm('age', ASOInt(50))
      ]
    )
  );

  Test('new(pst2, "name": "Severus Snape", "age": 50)',
    TAlgosimTypedStructure.CreateWithValue(
      [
        sm('name', ASO('Severus Snape')),
        sm('age', ASOInt(50))
      ]
    )
  );

  Test('new(pst2, "name": "Severus Snape", "age": 50, "sex": "male")',
    failure,
    'Structure type "person" does not contain a member named "sex".'
  );

  Test('UnregisterStructType("person")', success);

  Test('pst2 ≔ GetStructType("person")',
    failure, 'Unknown structure type "person".');

  Test('UnregisterStructType("RGB")', failure, 'Cannot unregister kernel structure type "RGB".');
  Test('UnregisterStructType("Date")', failure, 'Cannot unregister kernel structure type "Date".');
  Test('UnregisterStructType("Time")', failure, 'Cannot unregister kernel structure type "Time".');
  Test('UnregisterStructType("DateTime")', failure, 'Cannot unregister kernel structure type "DateTime".');
  Test('UnregisterStructType("Variable")', failure, 'Cannot unregister kernel structure type "Variable".');

  Test('UnregisterStructType("NargleData")', failure, 'Unknown structure type "NargleData"');


  //
  // Object types
  //

  Test('type(123)', 'integer');
  Test('type(1/2)', 'rational number');
  Test('type(5.2)', 'real number');
  Test('type(i)', 'complex number');
  Test('type(❨1, 0❩)', 'real vector');
  Test('type(❨i, 0❩)', 'complex vector');
  Test('type(❨❨1, 0❩, ❨0, 1❩❩)', 'real matrix');
  Test('type(❨❨i, 0❩, ❨0, i❩❩)', 'complex matrix');
  Test('type("niffler")', 'string');
  Test('type(true)', 'boolean');
  Test('type(''(1, 0))', 'array');
  Test('type({1, 0})', 'set');
  Test('type(struct("name": "Albus"))', 'structure');
  Test('type(date())', 'structure of type "Date"');
  Test('type(GetStructType("Date"))', 'structure type');
  Test('type(rgb(1, 0, 0))', 'color');
  Test('type(CreatePixmap(512, 512))', 'pixmap');
  Test('type(SineTone(400))', 'sound');
  Test('type(BinaryData(π))', 'binary data');
  Test('type(+sin)', 'kernel function');
  Test('type(n ↦ n! / n^n)', 'custom function');
  Test('type(0;)', 'null');
  Test('type(unprotect(ans))', 'success indication');

  // Integers

  Test('x ≔ 123;', null);
  Test('type(x)', 'integer');
  Test('ClassData(x).name', 'integer');
  Test('class(x)', 'integer');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Rational numbers

  Test('x ≔ 1/2;', null);
  Test('type(x)', 'rational number');
  Test('ClassData(x).name', 'rational number');
  Test('class(x)', 'rational number');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Real numbers

  Test('x ≔ 5.3;', null);
  Test('type(x)', 'real number');
  Test('ClassData(x).name', 'real number');
  Test('class(x)', 'real number');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Complex number

  Test('x ≔ 5 + i;', null);
  Test('type(x)', 'complex number');
  Test('ClassData(x).name', 'complex number');
  Test('class(x)', 'complex number');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', True);
  Test('IsComplexType(x)', True);
  Test('IsRealType(x)', False);
  Test('HasComplexType(x)', True);

  // Real vectors

  Test('x ≔ ❨1, 0, 0❩;', null);
  Test('type(x)', 'real vector');
  Test('ClassData(x).name', 'real vector');
  Test('class(x)', 'real vector');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', True);
  Test('IsValueContainer(x)', True);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Complex vectors

  Test('x ≔ ❨i, 0, 0❩;', null);
  Test('type(x)', 'complex vector');
  Test('ClassData(x).name', 'complex vector');
  Test('class(x)', 'complex vector');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', True);
  Test('IsValueContainer(x)', True);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', True);
  Test('IsComplexType(x)', True);
  Test('IsRealType(x)', False);
  Test('HasComplexType(x)', True);

  // Real matrices

  Test('x ≔ ❨❨3, 2❩, ❨1, 5❩❩;', null);
  Test('type(x)', 'real matrix');
  Test('ClassData(x).name', 'real matrix');
  Test('class(x)', 'real matrix');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', True);
  Test('IsValueContainer(x)', True);
  Test('ClassData(x).IsPlanarContainer', True);
  Test('IsPlanarContainer(x)', True);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Complex matrices

  Test('x ≔ ❨❨3, 2❩, ❨i, 5❩❩;', null);
  Test('type(x)', 'complex matrix');
  Test('ClassData(x).name', 'complex matrix');
  Test('class(x)', 'complex matrix');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', True);
  Test('IsValueContainer(x)', True);
  Test('ClassData(x).IsPlanarContainer', True);
  Test('IsPlanarContainer(x)', True);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', True);
  Test('IsComplexType(x)', True);
  Test('IsRealType(x)', False);
  Test('HasComplexType(x)', True);

  // Strings

  Test('x ≔ "Hello, World!";', null);
  Test('type(x)', 'string');
  Test('ClassData(x).name', 'string');
  Test('class(x)', 'string');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', True);
  Test('IsValueContainer(x)', True);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Booleans

  Test('x ≔ 1+1=2;', null);
  Test('type(x)', 'boolean');
  Test('ClassData(x).name', 'boolean');
  Test('class(x)', 'boolean');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Lists

  Test('x ≔ ''(1, 2, 3, "Vera Verto");', null);
  Test('type(x)', 'array');
  Test('ClassData(x).name', 'array');
  Test('class(x)', 'array');
  Test('ClassData(x).IsObjectContainer', True);
  Test('IsObjectContainer(x)', True);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  Test('x ≔ ''(i, "solemnly", "swear");', null);
  Test('type(x)', 'array');
  Test('ClassData(x).name', 'array');
  Test('class(x)', 'array');
  Test('ClassData(x).IsObjectContainer', True);
  Test('IsObjectContainer(x)', True);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', True);

  // Sets

  Test('x ≔ {1, 2, 3, "Vera Verto"};', null);
  Test('type(x)', 'set');
  Test('ClassData(x).name', 'set');
  Test('class(x)', 'set');
  Test('ClassData(x).IsObjectContainer', True);
  Test('IsObjectContainer(x)', True);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  Test('x ≔ {i, "solemnly", "swear"};', null);
  Test('type(x)', 'set');
  Test('ClassData(x).name', 'set');
  Test('class(x)', 'set');
  Test('ClassData(x).IsObjectContainer', True);
  Test('IsObjectContainer(x)', True);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', True);

  // Structures

  Test('x ≔ struct("name": "Albus Dumbledore", "age": 115);', null);
  Test('type(x)', 'structure');
  Test('ClassData(x).name', 'structure');
  Test('class(x)', 'structure');
  Test('ClassData(x).IsObjectContainer', True);
  Test('IsObjectContainer(x)', True);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  Test('x ≔ struct("name": "Albus Dumbledore", "age": i);', null);
  Test('type(x)', 'structure');
  Test('ClassData(x).name', 'structure');
  Test('class(x)', 'structure');
  Test('ClassData(x).IsObjectContainer', True);
  Test('IsObjectContainer(x)', True);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', True);

  with
    TAlgosimStructure.CreateWithValue(
      [
        sm('name', ASO('unnamed')),
        sm('age', ASOInt(0))
      ]
    )
  do
    try
      Test('pst ≔ ToStructType(person, "person")',
        ToStructType('person')
      );
    finally
      Free;
    end;

  Test('x ≔ new(pst, "name": "Albus Dumbledore", "age": 115);', null);
  Test('type(x)', 'structure of type "person"');
  Test('ClassData(x).name', 'typed structure');
  Test('class(x)', 'typed structure');
  Test('ClassData(x).IsObjectContainer', True);
  Test('IsObjectContainer(x)', True);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  Test('x ≔ new(pst, "name": "Albus Dumbledore", "age": i);', null);
  Test('type(x)', 'structure of type "person"');
  Test('ClassData(x).name', 'typed structure');
  Test('class(x)', 'typed structure');
  Test('ClassData(x).IsObjectContainer', True);
  Test('IsObjectContainer(x)', True);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', True);

  Test('x ≔ pst;', null);
  Test('type(x)', 'structure type');
  Test('ClassData(x).name', 'structure type');
  Test('class(x)', 'structure type');
  Test('ClassData(x).IsObjectContainer', True);
  Test('IsObjectContainer(x)', True);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Colours

  Test('x ≔ color("gold");', null);
  Test('type(x)', 'color');
  Test('ClassData(x).name', 'color');
  Test('class(x)', 'color');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('HasComplexType(x)', False);

  // Pixmaps

  Test('x ≔ CreatePixmap(512, 512);', null);
  Test('type(x)', 'pixmap');
  Test('ClassData(x).name', 'pixmap');
  Test('class(x)', 'pixmap');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', True);
  Test('IsValueContainer(x)', True);
  Test('ClassData(x).IsPlanarContainer', True);
  Test('IsPlanarContainer(x)', True);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Sounds

  Test('x ≔ SineTone(400);', null);
  Test('type(x)', 'sound');
  Test('ClassData(x).name', 'sound');
  Test('class(x)', 'sound');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Binary data

  Test('x ≔ BinaryData("nargle");', null);
  Test('type(x)', 'binary data');
  Test('ClassData(x).name', 'binary data');
  Test('class(x)', 'binary data');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', True);
  Test('IsValueContainer(x)', True);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', True);
  Test('IsOrderedContainer(x)', True);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Success indications

  Test('x ≔ unprotect("ans");', null);
  Test('type(x)', 'success indication');
  Test('ClassData(x).name', 'success indication');
  Test('class(x)', 'success indication');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Null objects

  Test('x ≔ (0;)', null);
  Test('type(x)', 'null');
  Test('ClassData(x).name', 'null');
  Test('class(x)', 'null');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Custom functions

  Test('x ≔ n ↦ n!/n^n;', null);
  Test('type(x)', 'custom function');
  Test('ClassData(x).name', 'custom function');
  Test('class(x)', 'custom function');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);

  // Kernel functions

  Test('x ≔ sin;', null);
  Test('type(x)', 'kernel function');
  Test('ClassData(x).name', 'kernel function');
  Test('class(x)', 'kernel function');
  Test('ClassData(x).IsObjectContainer', False);
  Test('IsObjectContainer(x)', False);
  Test('ClassData(x).IsValueContainer', False);
  Test('IsValueContainer(x)', False);
  Test('ClassData(x).IsPlanarContainer', False);
  Test('IsPlanarContainer(x)', False);
  Test('ClassData(x).IsOrderedContainer', False);
  Test('IsOrderedContainer(x)', False);
  Test('ClassData(x).IsComplexType', False);
  Test('IsComplexType(x)', False);
  Test('IsRealType(x)', True);
  Test('HasComplexType(x)', False);


  //
  // Format options
  //

  Test('NumDigits(π)', 0); // 0 = default
  Test('NumDigits(SetNumDigits(π, 3))', 3);
  Test('NumDigits(SetNumDigits(π, 5))', 5);
  Test('NumDigits(SetNumDigits(π, 10))', 10);
  Test('NumDigits(SetNumDigits(π, 12))', 12);
  Test('NumDigits(SetNumDigits(π, 0))', 0);

  Test('MinLength(123)', 0); // 0 = default
  Test('MinLength(SetMinLength(123, 2))', 2);
  Test('MinLength(SetMinLength(123, 5))', 5);
  Test('MinLength(SetMinLength(123, 7))', 7);
  Test('MinLength(SetMinLength(123, 10))', 10);

  Test('NumberFormat(123456.789)', 'inherited');
  Test('NumberFormat(SetNumberFormat(123456.789, "default"))', 'default');
  Test('NumberFormat(SetNumberFormat(123456.789, "fixed"))', 'fixed');
  Test('NumberFormat(SetNumberFormat(123456.789, "scientific"))', 'scientific');
  Test('NumberFormat(SetNumberFormat(123456.789, "inherited"))', 'inherited');

  Test('NumberBase(123)', 0); // 0 = default
  Test('NumberBase(SetNumberBase(123, 2))', 2);
  Test('NumberBase(SetNumberBase(123, 10))', 10);
  Test('NumberBase(SetNumberBase(123, 16))', 16);
  Test('NumberBase(SetNumberBase(123, 0))', 0);

  Test('FormatStyle(2)', 'default');
  Test('FormatStyle(SetFormatStyle(2, "day of week"))', 'day of week');
  Test('FormatStyle(SetFormatStyle(2, "MIDI instrument"))', 'MIDI instrument');
  Test('FormatStyle(SetFormatStyle(2, "default"))', 'default');

  Test('DigitGrouping(123456)', -1);
  Test('DigitGrouping(SetDigitGrouping(123456, 3))', 3);
  Test('DigitGrouping(SetDigitGrouping(123456, 6))', 6);
  Test('DigitGrouping(SetDigitGrouping(123456, 2))', 2);
  Test('DigitGrouping(SetDigitGrouping(123456, 1))', 1);
  Test('DigitGrouping(SetDigitGrouping(123456, 0))', 0);
  Test('DigitGrouping(SetDigitGrouping(123456, -1))', -1);

  Test('PrettyExp(1E20)', 'inherited');
  Test('PrettyExp(SetPrettyExp(1E20, true))', True);
  Test('PrettyExp(SetPrettyExp(1E20, false))', False);
  Test('PrettyExp(SetPrettyExp(1E20, "inherited"))', 'inherited');

  Test('MaxLen("Hello, World!")', 0);
  Test('MaxLen("Hello, World!" \ 5)', 5);
  Test('MaxLen("Hello, World!" \ 12)', 12);
  Test('MaxLen("Hello, World!" \ 0)', 0);


  //
  // Defuzzing
  //

  Test('defuzz(21)', 21);
  Test('defuzz(394)', 394);
  Test('defuzz(-19)', -19);
  Test('defuzz(0)', 0);

  Test('defuzz(21/22)', rat(21, 22));
  Test('defuzz(-1/4)', rat(-1, 4));
  Test('defuzz(1/4-1/4)', rat(0, 1));

  Test('defuzz(5.000000000785)', 5.0);
  Test('defuzz(5.00000785)', 5.00000785);
  Test('defuzz(5.00000785, 1E-8)', 5.00000785);
  Test('defuzz(5.00000785, 1E-7)', 5.00000785);
  Test('defuzz(5.00000785, 1E-6)', 5.00000785);
  Test('defuzz(5.00000785, 1E-5)', 5.0);
  Test('defuzz(5.00000785, 1E-4)', 5.0);
  Test('defuzz(5.00000785, 1E-3)', 5.0);
  Test('defuzz(5.00000785, 1E-2)', 5.0);
  Test('defuzz(5.00000785, 1E-1)', 5.0);
  Test('defuzz(5.00000785, 1)', 5.0);

  Test('defuzz(6.0000000036 − 5.99999999453 ⋅ i)', 6 - 6*ImaginaryUnit);
  Test('defuzz(5.0000256 − 1.999999985 ⋅ i)', 5.0000256 - 1.999999985 * ImaginaryUnit);
  Test('defuzz(5.0000256 − 1.999999985 ⋅ i, 1E-8)', 5.0000256 - 1.999999985 * ImaginaryUnit);
  Test('defuzz(5.0000256 − 1.999999985 ⋅ i, 1E-7)', 5.0000256 - 2 * ImaginaryUnit);
  Test('defuzz(5.0000256 − 1.999999985 ⋅ i, 1E-6)', 5.0000256 - 2 * ImaginaryUnit);
  Test('defuzz(5.0000256 − 1.999999985 ⋅ i, 1E-5)', 5.0000256 - 2 * ImaginaryUnit);
  Test('defuzz(5.0000256 − 1.999999985 ⋅ i, 1E-4)', 5 - 2 * ImaginaryUnit);
  Test('defuzz(5.0000256 − 1.999999985 ⋅ i, 1E-3)', 5 - 2 * ImaginaryUnit);
  Test('defuzz(5.0000256 − 1.999999985 ⋅ i, 1E-2)', 5 - 2 * ImaginaryUnit);
  Test('defuzz(5.0000256 − 1.999999985 ⋅ i, 1E-1)', 5 - 2 * ImaginaryUnit);

  Test('v ≔ ❨6.00000000031, 4.0000000071, −2.99999999563, 7.00000000952❩;', null);
  Test('defuzz(v)', [6, 4, -3, 7]);

  Test('v ≔ ❨6.0000145, 0.99999999214, 4.0000000317, 5.0000000012❩;', null);
  Test('defuzz(v)', [6.0000145, 1, 4.0000000317, 5]);
  Test('defuzz(v, 1E-9)', [6.0000145, 0.99999999214, 4.0000000317, 5.0000000012]);
  Test('defuzz(v, 1E-8)', [6.0000145, 1, 4.0000000317, 5]);
  Test('defuzz(v, 1E-7)', [6.0000145, 1, 4, 5]);
  Test('defuzz(v, 1E-6)', [6.0000145, 1, 4, 5]);
  Test('defuzz(v, 1E-5)', [6.0000145, 1, 4, 5]);
  Test('defuzz(v, 1E-4)', [6, 1, 4, 5]);
  Test('defuzz(v, 1E-3)', [6, 1, 4, 5]);
  Test('defuzz(v, 1E-2)', [6, 1, 4, 5]);
  Test('defuzz(v, 1E-1)', [6, 1, 4, 5]);

  Test('v ≔ ❨3.00000000152 + 0.99999999523 ⋅ i, 5.9999999912 − i, i❩;', null);
  Test('defuzz(v)', [3 + ImaginaryUnit, 6 - ImaginaryUnit, ImaginaryUnit]);

  Test('v ≔ ❨6.000035 − 0.000000007 ⋅ i, 2 + 4.99999921⋅i❩;', null);
  Test('defuzz(v)', [6.000035, 2 + 4.99999921*ImaginaryUnit]);
  Test('defuzz(v, 1E-8)', [6.000035, 2 + 4.99999921*ImaginaryUnit]);
  Test('defuzz(v, 1E-7)', [6.000035, 2 + 4.99999921*ImaginaryUnit]);
  Test('defuzz(v, 1E-6)', [6.000035, 2 + 5*ImaginaryUnit]);
  Test('defuzz(v, 1E-5)', [6.000035, 2 + 5*ImaginaryUnit]);
  Test('defuzz(v, 1E-4)', [6, 2 + 5*ImaginaryUnit]);
  Test('defuzz(v, 1E-3)', [6, 2 + 5*ImaginaryUnit]);
  Test('defuzz(v, 1E-2)', [6, 2 + 5*ImaginaryUnit]);
  Test('defuzz(v, 1E-1)', [6, 2 + 5*ImaginaryUnit]);

  Test('A ≔ ❨❨3.00000000012, 1.999999999997, 5❩, ❨−2.00000000053, 0, 0❩, ❨4.99999999987, 6.00000000021, 0.00000000000001❩❩;', null);
  Test('defuzz(A)', 3, [3, 2, 5, -2, 0, 0, 5, 6, 0]);

  Test('A ≔ ❨❨6.0000031, 4.000000017❩, ❨3.99999958, 0.00000025❩❩;', null);
  Test('defuzz(A)', 2, [6.0000031, 4.000000017, 3.99999958, 0.00000025]);
  Test('defuzz(A, 1E-8)', 2, [6.0000031, 4.000000017, 3.99999958, 0.00000025]);
  Test('defuzz(A, 1E-7)', 2, [6.0000031, 4, 3.99999958, 0.00000025]);
  Test('defuzz(A, 1E-6)', 2, [6.0000031, 4, 4, 0]);
  Test('defuzz(A, 1E-5)', 2, [6, 4, 4, 0]);
  Test('defuzz(A, 1E-4)', 2, [6, 4, 4, 0]);
  Test('defuzz(A, 1E-3)', 2, [6, 4, 4, 0]);
  Test('defuzz(A, 1E-2)', 2, [6, 4, 4, 0]);
  Test('defuzz(A, 1E-1)', 2, [6, 4, 4, 0]);

  Test('A ≔ ❨❨3.000000004 + 5.000000031⋅i, 0.99999924 − i❩, ❨5.0000000085⋅i, 3❩❩;', null);
  Test('defuzz(A)', 2, [3 + 5.000000031*ImaginaryUnit, 0.99999924 - ImaginaryUnit, 5*ImaginaryUnit, 3]);
  Test('defuzz(A, 1E-8)', 2, [3 + 5.000000031*ImaginaryUnit, 0.99999924 - ImaginaryUnit, 5*ImaginaryUnit, 3]);
  Test('defuzz(A, 1E-7)', 2, [3 + 5*ImaginaryUnit, 0.99999924 - ImaginaryUnit, 5*ImaginaryUnit, 3]);
  Test('defuzz(A, 1E-6)', 2, [3 + 5*ImaginaryUnit, 1 - ImaginaryUnit, 5*ImaginaryUnit, 3]);
  Test('defuzz(A, 1E-5)', 2, [3 + 5*ImaginaryUnit, 1 - ImaginaryUnit, 5*ImaginaryUnit, 3]);
  Test('defuzz(A, 1E-4)', 2, [3 + 5*ImaginaryUnit, 1 - ImaginaryUnit, 5*ImaginaryUnit, 3]);
  Test('defuzz(A, 1E-3)', 2, [3 + 5*ImaginaryUnit, 1 - ImaginaryUnit, 5*ImaginaryUnit, 3]);
  Test('defuzz(A, 1E-2)', 2, [3 + 5*ImaginaryUnit, 1 - ImaginaryUnit, 5*ImaginaryUnit, 3]);
  Test('defuzz(A, 1E-1)', 2, [3 + 5*ImaginaryUnit, 1 - ImaginaryUnit, 5*ImaginaryUnit, 3]);


  //
  // SameValue and CompareValue
  //

  Test('SameValue(5, 5)', True);
  Test('SameValue(127/128, 127/128)', True);
  Test('SameValue(3.1415, 3.1415)', True);
  Test('SameValue(-5.1, -5.1)', True);
  Test('SameValue(7E8, 7E8)', True);
  Test('SameValue(2.5E-20, 2.5E-20)', True);
  Test('SameValue(-5.1E10, -5.1E10)', True);
  Test('SameValue("cat", "cat")', True);
  Test('SameValue("hippogriff", "hippogriff")', True);
  Test('SameValue(true, true)', True);
  Test('SameValue(false, false)', True);
  Test('SameValue(❨1, 2, 3❩, ❨1, 2, 3❩)', True);
  Test('SameValue(❨❨5, 2❩, ❨0, 1❩❩, ❨❨5, 2❩, ❨0, 1❩❩)', True);
  Test('SameValue(❨i, 2, 3❩, ❨i, 2, 3❩)', True);
  Test('SameValue(❨❨i, 2❩, ❨0, 1❩❩, ❨❨i, 2❩, ❨0, 1❩❩)', True);
  Test('SameValue(''(1, 2, 3), ''(1, 2, 3))', True);
  Test('SameValue(''(1, 2, 3, "Vera Verto"), ''(1, 2, 3, "Vera Verto"))', True);
  Test('SameValue({1, 2, 3}, {1, 2, 3})', failure, 'Sets cannot be compared with epsilons.');
  Test('SameValue(color("red"), color("red"))', True);
  AvoidDateChange; Test('SameValue(date(), date())', True);
  Test('SameValue((0;), (0;))', True);
  Test('SameValue(ColorGradient(256, 256, color("red"), color("green")), ColorGradient(256, 256, color("red"), color("green")))', True);
  Test('SameValue(SineTone(400), SineTone(400))', True);
  Test('SameValue(unprotect(ans), unprotect(ans))', True);

  Test('SameValue(1/2 - 1/2, 0)', True);
  Test('SameValue(1.5 - 1.5, 0)', True);
  Test('SameValue(i - i, 0)', True);
  Test('SameValue("deathly" ~ " " ~ "hallows", "deathly hallows")', True);
  Test('SameValue(¬false, true)', True);
  Test('SameValue(❨1, 0, 0❩ + ❨0, 1, 0❩, ❨1, 1, 0❩)', True);
  Test('SameValue(color("red"), rgb(1, 0, 0))', True);
  Test('SameValue(color("red"), hsv(0, 1, 1))', True);
  Test('SameValue(color("red"), hsl(0, 1, 0.5))', True);

  Test('SameValue(7, 7.00000000000000027)', True);
  Test('SameValue(❨5, 1E−20❩, ❨5, 7E−21❩)', True);
  Test('SameValue(color("red"), rgb(0.999999999953, 0, 0))', True);

  Test('SameValue(❨5, 3, 6❩, ❨5, 3 + i − i, 6❩)', True);
  Test('SameValue(❨5, 3, 6 + i - i❩, ❨5, 3, 6❩)', True);
  Test('SameValue(❨❨5, 3, 6❩❩, ❨❨5, 3 + i − i, 6❩❩)', True);
  Test('SameValue(❨❨5, 3, 6 + i - i❩❩, ❨❨5, 3, 6❩❩)', True);

  Test('SameValue(❨5, 3, 6❩, ❨5, 3.0001 + i − i, 6❩)', False);
  Test('SameValue(❨5, 3, 6 + i - i❩, ❨5, 3.0001, 6❩)', False);
  Test('SameValue(❨❨5, 3, 6❩❩, ❨❨5, 3.0001 + i − i, 6❩❩)', False);
  Test('SameValue(❨❨5, 3, 6 + i - i❩❩, ❨❨5, 3.0001, 6❩❩)', False);

  Test('SameValue(❨5, 3, 6❩, ❨5, 3.0001 + i − i, 6❩, 2E-6)', False);
  Test('SameValue(❨5, 3, 6 + i - i❩, ❨5, 3.0001, 6❩, 2E-6)', False);
  Test('SameValue(❨❨5, 3, 6❩❩, ❨❨5, 3.0001 + i − i, 6❩❩, 2E-6)', False);
  Test('SameValue(❨❨5, 3, 6 + i - i❩❩, ❨❨5, 3.0001, 6❩❩, 2E-6)', False);

  Test('SameValue(❨5, 3, 6❩, ❨5, 3.0001 + i − i, 6❩, 2E-5)', False);
  Test('SameValue(❨5, 3, 6 + i - i❩, ❨5, 3.0001, 6❩, 2E-5)', False);
  Test('SameValue(❨❨5, 3, 6❩❩, ❨❨5, 3.0001 + i − i, 6❩❩, 2E-5)', False);
  Test('SameValue(❨❨5, 3, 6 + i - i❩❩, ❨❨5, 3.0001, 6❩❩, 2E-5)', False);

  Test('SameValue(❨5, 3, 6❩, ❨5, 3.0001 + i − i, 6❩, 2E-4)', True);
  Test('SameValue(❨5, 3, 6 + i - i❩, ❨5, 3.0001, 6❩, 2E-4)', True);
  Test('SameValue(❨❨5, 3, 6❩❩, ❨❨5, 3.0001 + i − i, 6❩❩, 2E-4)', True);
  Test('SameValue(❨❨5, 3, 6 + i - i❩❩, ❨❨5, 3.0001, 6❩❩, 2E-4)', True);

  Test('SameValue(5, 6)', False);
  Test('SameValue(127/128, 129/128)', False);
  Test('SameValue(127/128, -127/128)', False);
  Test('SameValue(3.1415, 3.141592653589793238)', False);
  Test('SameValue(-5.1, 5.1)', False);
  Test('SameValue(7E8, 7.00002E8)', False);
  Test('SameValue(2.5E20, 2.5E-20)', False);
  Test('SameValue(-5.1E10, -5.1000023E10)', False);
  Test('SameValue("cat", "dog")', False);
  Test('SameValue("hippogriff", "hippogriffs")', False);
  Test('SameValue(true, false)', False);
  Test('SameValue(false, true)', False);
  Test('SameValue(❨1, 2, 3❩, ❨-1, 2, 3❩)', False);
  Test('SameValue(❨❨5, 2❩, ❨1, 1❩❩, ❨❨5, 2❩, ❨0, 1❩❩)', False);
  Test('SameValue(❨i, 2, 3❩, ❨0, 2, 3❩)', False);
  Test('SameValue(❨❨i, 2❩, ❨i, 1❩❩, ❨❨i, 2❩, ❨0, 1❩❩)', False);
  Test('SameValue(''(1, 2.1, 3), ''(1, 2, 3))', False);
  Test('SameValue(''(1, 2, 3, "Vera Verto"), ''(i))', False);
  Test('SameValue({1, 2, 3}, {1, 2, 4})', failure, 'Sets cannot be compared with epsilons.');
  Test('SameValue(color("red"), color("green"))', False);
  Test('SameValue(date(), time())', False);
  Test('SameValue(0, (0;))', False);
  Test('SameValue(ColorGradient(256, 256, color("red"), color("green")), ColorGradient(256, 257, color("red"), color("green")))', False);
  Test('SameValue(ColorGradient(256, 256, color("red"), color("green")), ColorGradient(256, 256, color("red"), color("yellow")))', False);
  Test('SameValue(SineTone(400), SineTone(401))', False);
  Test('SameValue(unprotect(ans), 0)', False);

  Test('SameValue(1/2 - 1/3, 0)', False);
  Test('SameValue(1.5 - 1.6, 0)', False);
  Test('SameValue(i - 1, 0)', False);
  Test('SameValue("deathly" ~ " " ~ "hallows", "deadly hallows")', False);
  Test('SameValue(¬false, ¬¬¬true)', False);
  Test('SameValue(❨1, 0, 0❩ + ❨0, 0, 1❩, ❨1, 1, 0❩)', False);
  Test('SameValue(color("red"), rgb(0.8, 0, 0))', False);
  Test('SameValue(color("red"), hsv(0, 0.8, 1))', False);
  Test('SameValue(color("red"), hsl(0, 1, 0.6))', False);

  Test('SameValue(7, 7.0000000027)', False);
  Test('SameValue(❨5, 1E20❩, ❨5, 7E−21❩)', False);
  Test('SameValue(color("red"), rgb(0.99, 0, 0))', False);

  Test('SameValue(5, "rat")', False);
  Test('SameValue(127/128, "cat")', False);
  Test('SameValue(3.1415, "dog")', False);
  Test('SameValue(-5.1, "hippogriff")', False);
  Test('SameValue(7E8, 5)', False);
  Test('SameValue(2.5E-20, 1/2)', False);
  Test('SameValue(-5.1E10, true)', False);
  Test('SameValue("cat", ❨3, 5❩)', False);
  Test('SameValue("hippogriff", time())', False);
  Test('SameValue(true, "true")', False);
  Test('SameValue(false, 0)', False);
  Test('SameValue(❨1, 2, 3❩, ❨❨1, 2, 3❩, ❨1, 2, 3❩, ❨1, 2, 3❩❩)', False);
  Test('SameValue(❨❨5, 2❩, ❨0, 1❩❩, color("red"))', False);
  Test('SameValue(❨i, 2, 3❩, date())', False);
  Test('SameValue(❨❨i, 2❩, ❨0, 1❩❩, SineTone(400))', False);
  Test('SameValue(''(1, 2, 3), (0;))', False);
  Test('SameValue(''(1, 2, 3, "Vera Verto"), ''())', False);
  Test('SameValue(color("red"), 123)', False);
  Test('SameValue(date(), "date")', False);
  Test('SameValue((0;), false)', False);
  Test('SameValue(ColorGradient(256, 256, color("red"), color("green")), IdentityMatrix(100))', False);
  Test('SameValue(SineTone(400), exp(1))', False);
  Test('SameValue(unprotect(ans), 1)', False);

  Test('SameValue(7.6492341, 7.6495123)', False);
  Test('SameValue(7.6492341, 7.6495123, 1E-6)', False);
  Test('SameValue(7.6492341, 7.6495123, 1E-5)', False);
  Test('SameValue(7.6492341, 7.6495123, 1E-4)', False);
  Test('SameValue(7.6492341, 7.6495123, 1E-3)', True);
  Test('SameValue(7.6492341, 7.6495123, 1E-2)', True);
  Test('SameValue(7.6492341, 7.6495123, 1E-1)', True);

  Test('SameValue(3.1415, 3.1415926535897932384626)', False);
  Test('SameValue(3.1415, 3.1415926535897932384626, 1E-6)', False);
  Test('SameValue(3.1415, 3.1415926535897932384626, 1E-5)', False);
  Test('SameValue(3.1415, 3.1415926535897932384626, 1E-4)', True);
  Test('SameValue(3.1415, 3.1415926535897932384626, 1E-3)', True);
  Test('SameValue(3.1415, 3.1415926535897932384626, 1E-2)', True);
  Test('SameValue(3.1415, 3.1415926535897932384626, 1E-1)', True);
  Test('SameValue(3.1415, 3.1415926535897932384626, 1)', True);

  Test('SameValue(5 + i, 5.0000023 + i)', False);
  Test('SameValue(5 + i, 5.0000023 + i, 1E-6)', False);
  Test('SameValue(5 + i, 5.0000023 + i, 1E-5)', True);
  Test('SameValue(5 + i, 5.0000023 + i, 1E-4)', True);
  Test('SameValue(5 + i, 5.0000023 + i, 1E-3)', True);
  Test('SameValue(5 + i, 5.0000023 + i, 1E-2)', True);
  Test('SameValue(5 + i, 5.0000023 + i, 1E-1)', True);

  Test('SameValue(5 + i, 5 + 0.9999523⋅i)', False);
  Test('SameValue(5 + i, 5 + 0.9999523⋅i, 1E-6)', False);
  Test('SameValue(5 + i, 5 + 0.9999523⋅i, 1E-5)', False);
  Test('SameValue(5 + i, 5 + 0.9999523⋅i, 1E-4)', True);
  Test('SameValue(5 + i, 5 + 0.9999523⋅i, 1E-3)', True);
  Test('SameValue(5 + i, 5 + 0.9999523⋅i, 1E-2)', True);
  Test('SameValue(5 + i, 5 + 0.9999523⋅i, 1E-1)', True);

  Test('SameValue(❨0.515093362425, 0.691668689949, 0.0145804323256❩, ❨0.515057811471, 0.691663156165, 0.0145881118847❩)', False);
  Test('SameValue(❨0.515093362425, 0.691668689949, 0.0145804323256❩, ❨0.515057811471, 0.691663156165, 0.0145881118847❩, 1E-6)', False);
  Test('SameValue(❨0.515093362425, 0.691668689949, 0.0145804323256❩, ❨0.515057811471, 0.691663156165, 0.0145881118847❩, 1E-5)', False);
  Test('SameValue(❨0.515093362425, 0.691668689949, 0.0145804323256❩, ❨0.515057811471, 0.691663156165, 0.0145881118847❩, 1E-4)', True);
  Test('SameValue(❨0.515093362425, 0.691668689949, 0.0145804323256❩, ❨0.515057811471, 0.691663156165, 0.0145881118847❩, 1E-3)', True);
  Test('SameValue(❨0.515093362425, 0.691668689949, 0.0145804323256❩, ❨0.515057811471, 0.691663156165, 0.0145881118847❩, 1E-2)', True);
  Test('SameValue(❨0.515093362425, 0.691668689949, 0.0145804323256❩, ❨0.515057811471, 0.691663156165, 0.0145881118847❩, 1E-1)', True);

  Test('SameValue(❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩, ❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩)', False);
  Test('SameValue(❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩, ❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩, 1E-6)', False);
  Test('SameValue(❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩, ❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩, 1E-5)', False);
  Test('SameValue(❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩, ❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩, 1E-4)', True);
  Test('SameValue(❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩, ❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩, 1E-3)', True);
  Test('SameValue(❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩, ❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩, 1E-2)', True);
  Test('SameValue(❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩, ❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩, 1E-1)', True);

  Test('SameValue(❨❨0.515093362425, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471, 0.691663156165, 0.0145881118847❩❩)', False);
  Test('SameValue(❨❨0.515093362425, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471, 0.691663156165, 0.0145881118847❩❩, 1E-6)', False);
  Test('SameValue(❨❨0.515093362425, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471, 0.691663156165, 0.0145881118847❩❩, 1E-5)', False);
  Test('SameValue(❨❨0.515093362425, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471, 0.691663156165, 0.0145881118847❩❩, 1E-4)', True);
  Test('SameValue(❨❨0.515093362425, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471, 0.691663156165, 0.0145881118847❩❩, 1E-3)', True);
  Test('SameValue(❨❨0.515093362425, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471, 0.691663156165, 0.0145881118847❩❩, 1E-2)', True);
  Test('SameValue(❨❨0.515093362425, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471, 0.691663156165, 0.0145881118847❩❩, 1E-1)', True);

  Test('SameValue(❨❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩❩)', False);
  Test('SameValue(❨❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩❩, 1E-6)', False);
  Test('SameValue(❨❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩❩, 1E-5)', False);
  Test('SameValue(❨❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩❩, 1E-4)', True);
  Test('SameValue(❨❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩❩, 1E-3)', True);
  Test('SameValue(❨❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩❩, 1E-2)', True);
  Test('SameValue(❨❨0.515093362425⋅i, 0.691668689949, 0.0145804323256❩❩, ❨❨0.515057811471⋅i, 0.691663156165, 0.0145881118847❩❩, 1E-1)', True);

  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Technically, it’s a ferret."))', False);
  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Technically, it’s a ferret."), 1E-8)', False);
  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Technically, it’s a ferret."), 1E-7)', False);
  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Technically, it’s a ferret."), 1E-6)', True);
  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Technically, it’s a ferret."), 1E-5)', True);
  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Technically, it’s a ferret."), 1E-4)', True);
  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Technically, it’s a ferret."), 1E-3)', True);
  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Technically, it’s a ferret."), 1E-2)', True);
  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Technically, it’s a ferret."), 1E-1)', True);
  Test('SameValue(''(3.141592, true, "Technically, it’s a ferret."), ''(3.1415926535897932384626, true, "Is that a student?!"), 1E-1)', False);

  Test('CompareValue(5, 5)', 0);
  Test('CompareValue(127/128, 127/128)', 0);
  Test('CompareValue(3.1415, 3.1415)', 0);
  Test('CompareValue(-5.1, -5.1)', 0);
  Test('CompareValue(7E8, 7E8)', 0);
  Test('CompareValue(2.5E-20, 2.5E-20)', 0);
  Test('CompareValue(-5.1E10, -5.1E10)', 0);
  Test('CompareValue("cat", "cat")', 0);
  Test('CompareValue("hippogriff", "hippogriff")', 0);
  Test('CompareValue(true, true)', 0);
  Test('CompareValue(false, false)', 0);
  Test('CompareValue(❨1, 2, 3❩, ❨1, 2, 3❩)', 0);
  Test('CompareValue(❨❨5, 2❩, ❨0, 1❩❩, ❨❨5, 2❩, ❨0, 1❩❩)', 0);
  Test('CompareValue(❨i, 2, 3❩, ❨i, 2, 3❩)', 0);
  Test('CompareValue(❨❨i, 2❩, ❨0, 1❩❩, ❨❨i, 2❩, ❨0, 1❩❩)', 0);
  Test('CompareValue(''(1, 2, 3), ''(1, 2, 3))', 0);
  Test('CompareValue(''(1, 2, 3, "Vera Verto"), ''(1, 2, 3, "Vera Verto"))', 0);
  Test('CompareValue({1, 2, 3}, {1, 2, 3})', failure, 'Sets have no default order.');
  Test('CompareValue(color("red"), color("red"))', 0);
  AvoidDateChange; Test('CompareValue(date(), date())', 0);
  Test('CompareValue((0;), (0;))', 0);
  Test('CompareValue(ColorGradient(256, 256, color("red"), color("green")), ColorGradient(256, 256, color("red"), color("green")))', 0);
  Test('CompareValue(SineTone(400), SineTone(400))', 0);
  Test('CompareValue(unprotect(ans), unprotect(ans))', 0);

  Test('CompareValue(1/2 - 1/2, 0)', 0);
  Test('CompareValue(1.5 - 1.5, 0)', 0);
  Test('CompareValue(i - i, 0)', 0);
  Test('CompareValue("deathly" ~ " " ~ "hallows", "deathly hallows")', 0);
  Test('CompareValue(¬false, true)', 0);
  Test('CompareValue(❨1, 0, 0❩ + ❨0, 1, 0❩, ❨1, 1, 0❩)', 0);
  Test('CompareValue(color("red"), rgb(1, 0, 0))', 0);
  Test('CompareValue(color("red"), hsv(0, 1, 1))', 0);
  Test('CompareValue(color("red"), hsl(0, 1, 0.5))', 0);

  Test('CompareValue(7, 7.00000000000000027)', 0);
  Test('CompareValue(❨5, 1E−20❩, ❨5, 7E−21❩)', 0);

  Test('CompareValue(❨5, 3, 6❩, ❨5, 3 + i − i, 6❩)', 0);
  Test('CompareValue(❨5, 3, 6 + i - i❩, ❨5, 3, 6❩)', 0);
  Test('CompareValue(❨❨5, 3, 6❩❩, ❨❨5, 3 + i − i, 6❩❩)', 0);
  Test('CompareValue(❨❨5, 3, 6 + i - i❩❩, ❨❨5, 3, 6❩❩)', 0);

  Test('CompareValue(5, 6)', -1);
  Test('CompareValue(5, 5.0)', 0);
  Test('CompareValue(5, 4.0)', 1);
  Test('CompareValue(5.0, 6)', -1);
  Test('CompareValue(5.0, 5)', 0);
  Test('CompareValue(5.0, 4)', 1);

  Test('CompareValue(5, "cat")', -1);
  Test('CompareValue("dog", 7)', 1);
  Test('CompareValue("nargle", 0)', 1);

  Test('CompareValue("cat", "dog")', -1);
  Test('CompareValue("dog", "nargle")', -1);
  Test('CompareValue("nargle", "owl")', -1);
  Test('CompareValue("nargle", "dog")', 1);
  Test('CompareValue("nargle", "cat")', 1);
  Test('CompareValue("cat", "cat")', 0);

  Test('CompareValue(date(2020, 05, 09), date(2020, 05, 10))', -1);
  Test('CompareValue(date(2020, 05, 09), date(2020, 05, 09))', 0);
  Test('CompareValue(date(2020, 05, 10), date(2020, 05, 09))', 1);
  Test('CompareValue(date(2020, 04, 10), date(2020, 05, 09))', -1);
  Test('CompareValue(date(2019, 05, 10), date(2020, 05, 09))', -1);


  //
  // Additional L-value-modifying functions
  //

  // inc and dec

  Test('n ≔ 0', 0);
  Test('inc(n)', 1);
  Test('n', 1);
  Test('inc(n)', 2);
  Test('n', 2);
  Test('inc(n)', 3);
  Test('n', 3);
  Test('inc(n, 5)', 8);
  Test('n', 8);
  Test('inc(n, 2)', 10);
  Test('n', 10);
  Test('inc(n, -5)', 5);
  Test('n', 5);
  Test('inc(n)', 6);
  Test('n', 6);
  Test('inc(n, 10)', 16);
  Test('n', 16);
  Test('inc(n, -6)', 10);
  Test('n', 10);
  Test('inc(n, -10)', 0);
  Test('n', 0);
  Test('inc(n, -10)', -10);
  Test('n', -10);
  Test('inc(n, 3)', -7);
  Test('n', -7);
  Test('dec(n)', -8);
  Test('n', -8);
  Test('dec(n)', -9);
  Test('n', -9);
  Test('dec(n)', -10);
  Test('n', -10);
  Test('dec(n, 5)', -15);
  Test('n', -15);
  Test('dec(n, 5)', -20);
  Test('n', -20);
  Test('dec(n, -10)', -10);
  Test('n', -10);
  Test('dec(n, -5)', -5);
  Test('n', -5);
  Test('dec(n, -2)', -3);
  Test('n', -3);
  Test('dec(n, -3)', 0);
  Test('n', 0);
  Test('dec(n, -7)', 7);
  Test('n', 7);
  Test('dec(n)', 6);
  Test('n', 6);
  Test('dec(n)', 5);
  Test('n', 5);
  Test('dec(n, 3)', 2);
  Test('n', 2);
  Test('dec(n, 2)', 0);
  Test('n', 0);
  Test('dec(n, 1)', -1);
  Test('n', -1);
  Test('dec(n, -11)', 10);
  Test('n', 10);
  Test('dec(n)', 9);
  Test('n', 9);
  Test('inc(n, 1)', 10);
  Test('n', 10);
  Test('inc(n)', 11);
  Test('n', 11);
  Test('inc(n, 100)', 111);
  Test('n', 111);
  Test('dec(n, 11)', 100);
  Test('n', 100);

  Test('inc(n, 0.5)', failure, 'An object of type integer was expected as argument 2, but an object of type real number was given.');
  Test('n', 100);
  Test('inc(n, 1/0)', failure, 'Division by zero');
  Test('n', 100);

  Test('delete(n)', success);
  Test('inc(n)', failure, 'Unknown identifier "n".');
  Test('inc(n, 2)', failure, 'Unknown identifier "n".');
  Test('dec(n)', failure, 'Unknown identifier "n".');
  Test('dec(n, 2)', failure, 'Unknown identifier "n".');

  Test('n ≔ 5', 5);
  Test('inc(n)', 6);
  Test('dec(n, 4)', 2);
  Test('inc(n, 8)', 10);
  Test('dec(n)', 9);
  Test('n', 9);
  Test('protect(n)', success);
  Test('inc(n)', failure, 'Left side cannot be assigned to.'); // Would prefer "Cannot modify or delete protected variable "n"."
  Test('n', 9);
  Test('unprotect(n)', success);
  Test('inc(n)', 10);
  Test('n', 10);
  Test('protect(n)', success);
  Test('dec(n, 2)', failure, 'Left side cannot be assigned to.');
  Test('n', 10);
  Test('unprotect(n)', success);
  Test('dec(n, 2)', 8);
  Test('n', 8);

  Test('x ≔ 5.1', 5.1);
  Test('inc(x)', 6.1);
  Test('x', 6.1);
  Test('inc(x)', 7.1);
  Test('inc(x, 3)', 10.1);
  Test('dec(x)', 9.1);
  Test('dec(x)', 8.1);
  Eps; Test('dec(x, 6)', 2.1);
  Eps; Test('x', 2.1);

  Test('q ≔ 1/2', rat(1, 2));
  Test('inc(q)', rat(3, 2));
  Test('inc(q)', rat(5, 2));
  Test('inc(q)', rat(7, 2));
  Test('q', rat(7, 2));
  Test('inc(q, 2)', rat(11, 2));
  Test('inc(q, -1)', rat(9, 2));
  Test('dec(q)', rat(7, 2));
  Test('dec(q)', rat(5, 2));
  Test('q', rat(5, 2));
  Test('dec(q, 2)', rat(1, 2));
  Test('dec(q)', rat(-1, 2));
  Test('q', rat(-1, 2));

  Test('z ≔ 1 + i', 1 + ImaginaryUnit);
  Test('inc(z)', 2 + ImaginaryUnit);
  Test('z', 2 + ImaginaryUnit);
  Test('inc(z, 8)', 10 + ImaginaryUnit);
  Test('z', 10 + ImaginaryUnit);
  Test('dec(z)', 9 + ImaginaryUnit);
  Test('dec(z, 5)', 4 + ImaginaryUnit);
  Test('z', 4 + ImaginaryUnit);
  Test('inc(z, -4)', ImaginaryUnit);
  Test('z', ImaginaryUnit);

  Test('s ≔ "nargle"', 'nargle');
  Test('inc(s)', failure, 'An object of type number was expected as argument 1, but an object of type string was given.');

  Test('delete(n); delete(x); delete(q); delete(z); delete(s)', success);

  Test('inc(5)', failure, 'Left side cannot be assigned to.');

  // VarAppend

  Test('v ≔ ❨5, 3❩', [5, 3]);
  Test('VarAppend(v, 2)', success);
  Test('v', [5, 3, 2]);
  Test('VarAppend(v, -10)', success);
  Test('v', [5, 3, 2, -10]);
  Test('VarAppend(v, 1024)', success);
  Test('v', [5, 3, 2, -10, 1024]);
  Test('VarAppend(v, 5.2)', success);
  Test('v', [5, 3, 2, -10, 1024, 5.2]);
  Test('VarAppend(v, "cat")', failure, 'Object isn''t a real number.');
  Test('VarAppend(v, 1 + i)', failure, 'Object isn''t a real number.');
  Test('v', [5, 3, 2, -10, 1024, 5.2]);
  Test('protect(v)', success);
  Test('VarAppend(v, 3.14)', failure, 'Left side cannot be assigned to.');
  Test('v', [5, 3, 2, -10, 1024, 5.2]);
  Test('unprotect(v)', success);
  Test('VarAppend(v, 3.14)', success);
  Test('v', [5, 3, 2, -10, 1024, 5.2, 3.14]);
  Test('delete(v)', success);
  Test('VarAppend(v, 394)', failure, 'Unknown identifier "v".');

  Test('v ≔ ❨i, 3❩', [ImaginaryUnit, 3]);
  Test('VarAppend(v, 2)', success);
  Test('v', [ImaginaryUnit, 3, 2]);
  Test('VarAppend(v, -10)', success);
  Test('v', [ImaginaryUnit, 3, 2, -10]);
  Test('VarAppend(v, 1024)', success);
  Test('v', [ImaginaryUnit, 3, 2, -10, 1024]);
  Test('VarAppend(v, 5.2)', success);
  Test('v', [ImaginaryUnit, 3, 2, -10, 1024, 5.2]);
  Test('VarAppend(v, "cat")', failure, 'Object isn''t a complex number.');
  Test('VarAppend(v, 1 + i)', success);
  Test('v', [ImaginaryUnit, 3, 2, -10, 1024, 5.2, 1 + ImaginaryUnit]);
  Test('protect(v)', success);
  Test('VarAppend(v, 3.14)', failure, 'Left side cannot be assigned to.');
  Test('v', [ImaginaryUnit, 3, 2, -10, 1024, 5.2, 1 + ImaginaryUnit]);
  Test('unprotect(v)', success);
  Test('VarAppend(v, 3.14)', success);
  Test('v', [ImaginaryUnit, 3, 2, -10, 1024, 5.2, 1 + ImaginaryUnit, 3.14]);
  Test('delete(v)', success);
  Test('VarAppend(v, 394)', failure, 'Unknown identifier "v".');

  Test('A ≔ ❨❨5, 3❩, ❨4, 1❩❩', 2, [5, 3, 4, 1]);
  Test('VarAppend(A, 6)', failure, 'Object type "real matrix" is not a suitable container.');

  Test('A ≔ ❨❨5, 3❩, ❨4, i❩❩', 2, [5, 3, 4, ImaginaryUnit]);
  Test('VarAppend(A, 6)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('L ≔ ''(1, 2, 3)', asoarr([1, 2, 3]));
  Test('VarAppend(L, 5)', success);
  Test('L', asoarr([1, 2, 3, 5]));
  Test('VarAppend(L, "cat")', success);
  Test('L', asoarr([1, 2, 3, 5, 'cat']));
  Test('VarAppend(L, true)', success);
  Test('L', asoarr([1, 2, 3, 5, 'cat', True]));
  Test('protect(L)', success);
  Test('VarAppend(L, "dog")', failure, 'Left side cannot be assigned to.');
  Test('L', asoarr([1, 2, 3, 5, 'cat', True]));
  Test('unprotect(L)', success);
  Test('VarAppend(L, "dog")', success);
  Test('L', asoarr([1, 2, 3, 5, 'cat', True, 'dog']));
  Test('VarAppend(L, ''(3, 1, 4, 1, 5))', success);
  Test('L', asoarrex([ASOInt(1), ASOInt(2), ASOInt(3), ASOInt(5), ASO('cat'), ASO(True), ASO('dog'), intarr([3, 1, 4, 1, 5])]));

  Test('S ≔ {1, 2, 3}', asoset([1, 2, 3]));
  Test('VarAppend(S, 4)', failure, 'Object type "set" is not a suitable container.');

  Test('s ≔ "Hello, World"', 'Hello, World');
  Test('VarAppend(s, "!")', success);
  Test('s', 'Hello, World!');
  Test('VarAppend(s, "!!!")', failure, 'A string of length 3 cannot be converted to a character.');
  Test('VarAppend(s, true)', failure, 'An object of type "boolean" cannot be converted to a character.');

  Test('delete(A); delete(L); delete(S); delete(s)', success);

  // VarExtendWith

  Test('v ≔ ❨3, 1, 4❩', [3, 1, 4]);
  Test('VarExtendWith(v, ❨1, 5❩)', success);
  Test('v', [3, 1, 4, 1, 5]);
  Test('VarExtendWith(v, 9)', success);
  Test('v', [3, 1, 4, 1, 5, 9]);
  Test('VarExtendWith(v, 2)', success);
  Test('v', [3, 1, 4, 1, 5, 9, 2]);
  Test('VarExtendWith(v, ❨6, 5, 3, 5❩)', success);
  Test('v', [3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5]);
  Test('protect(v)', success);
  Test('VarExtendWith(v, ❨8, 9, 7, 9❩)', failure, 'Left side cannot be assigned to.');
  Test('v', [3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5]);
  Test('unprotect(v)', success);
  Test('VarExtendWith(v, ❨8, 9, 7, 9❩)', success);
  Test('v', [3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9]);
  Test('VarExtendWith(v, i)', failure, 'Object isn''t a real number.');
  Test('VarExtendWith(v, ❨8, 9, 7, i❩)', failure, 'Complex vector cannot be converted to real vector because it contains a non-real component i.');
  Test('VarExtendWith(v, "nargle")', failure, 'Cannot convert string "nargle" to a real number.');
  Test('VarExtendWith(v, true)', failure, 'An object of type "boolean" cannot be converted to a real vector.');
  Test('v', [3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9]);

  Test('v ≔ ❨i, 1, 4❩', [ImaginaryUnit, 1, 4]);
  Test('VarExtendWith(v, ❨1, 5❩)', success);
  Test('v', [ImaginaryUnit, 1, 4, 1, 5]);
  Test('VarExtendWith(v, 9)', success);
  Test('v', [ImaginaryUnit, 1, 4, 1, 5, 9]);
  Test('VarExtendWith(v, 2)', success);
  Test('v', [ImaginaryUnit, 1, 4, 1, 5, 9, 2]);
  Test('VarExtendWith(v, ❨6, 5, 3, 5❩)', success);
  Test('v', [ImaginaryUnit, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5]);
  Test('protect(v)', success);
  Test('VarExtendWith(v, ❨8, 9, 7, 9❩)', failure, 'Left side cannot be assigned to.');
  Test('v', [ImaginaryUnit, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5]);
  Test('unprotect(v)', success);
  Test('VarExtendWith(v, ❨8, 9, 7, 9❩)', success);
  Test('v', [ImaginaryUnit, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9]);
  Test('VarExtendWith(v, i)', success);
  Test('v', [ImaginaryUnit, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, ImaginaryUnit]);
  Test('VarExtendWith(v, ❨8, 9, 7, i❩)', success);
  Test('v', [ImaginaryUnit, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, ImaginaryUnit, 8, 9, 7, ImaginaryUnit]);
  Test('VarExtendWith(v, "nargle")', failure, 'Cannot convert string "nargle" to a complex number.');
  Test('VarExtendWith(v, true)', failure, 'An object of type "boolean" cannot be converted to a complex vector.');
  Test('v', [ImaginaryUnit, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, ImaginaryUnit, 8, 9, 7, ImaginaryUnit]);

  Test('A ≔ ❨❨3, 5❩, ❨1, 4❩❩;', null);
  Test('VarExtendWith(A, 6)', failure, 'Object type "real matrix" is not a suitable container.');
  Test('A ≔ ❨❨3, 5❩, ❨i, 4❩❩;', null);
  Test('VarExtendWith(A, 6)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('s ≔ "Hello, World"', 'Hello, World');
  Test('VarExtendWith(s, "!")', success);
  Test('s', 'Hello, World!');
  Test('VarExtendWith(s, " How are you?")', success);
  Test('s', 'Hello, World! How are you?');
  Test('VarExtendWith(s, " I''m fine!")', success);
  Test('s', 'Hello, World! How are you? I''m fine!');

  Test('L ≔ ''(1, 2)', intarr([1, 2]));
  Test('VarExtendWith(L, ''(3, "Vera Verto"))', success);
  Test('L', asoarr([1, 2, 3, 'Vera Verto']));
  Test('protect(L)', success);
  Test('VarExtendWith(L, ''("!"))', failure, 'Left side cannot be assigned to.');
  Test('L', asoarr([1, 2, 3, 'Vera Verto']));
  Test('unprotect(L)', success);
  Test('VarExtendWith(L, ''("!"))', success);
  Test('L', asoarr([1, 2, 3, 'Vera Verto', '!']));

  Test('S ≔ {1, 2, 3}', asoset([1, 2, 3]));
  Test('VarExtendWith(S, ''(4, 5, 6))', failure, 'Object type "set" is not a suitable container.');

  Test('VarExtendWith(❨1, 2, 3❩, ❨4, 5❩)', failure, 'Left side cannot be assigned to.');

  Test('delete(v); delete(A); delete(L); delete(S); delete(s)', success);

  // VarInsert

  Test('v ≔ ❨1, 2, 3, 4, 5, 6, 7, 8, 9, 10❩', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  Test('VarInsert(v, 1, 0)', success);
  Test('v', [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  Test('VarInsert(v, 12, 11)', success);
  Test('v', [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]);
  Test('VarInsert(v, 7, 5.5)', success);
  Test('v', [0, 1, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11]);
  Test('VarInsert(v, −1, 12)', success);
  Test('v', [0, 1, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11, 12]);
  Test('VarInsert(v, 16, 13)', failure, 'Index 16 out of bounds.');
  Test('VarInsert(v, -16, 13)', failure, 'Index -16 out of bounds.');
  Test('VarInsert(v, 0, 13)', failure, 'Index 0 out of bounds.');
  Test('VarInsert(v, 394, 13)', failure, 'Index 394 out of bounds.');
  Test('protect(v)', success);
  Test('VarInsert(v, 15, 13)', failure, 'Left side cannot be assigned to.');
  Test('v', [0, 1, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11, 12]);
  Test('unprotect(v)', success);
  Test('VarInsert(v, 15, 13)', success);
  Test('v', [0, 1, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11, 12, 13]);
  Test('VarInsert(v, −1, i)', failure, 'Object isn''t a real number.');
  Test('VarInsert(v, −1, "test")', failure, 'Object isn''t a real number.');
  Test('v', [0, 1, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11, 12, 13]);
  Test('VarInsert(❨1, 2, 3❩, 2, 10)', failure, 'Left side cannot be assigned to.');

  Test('v ≔ ❨i, 2, 3, 4, 5, 6, 7, 8, 9, 10❩', [ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  Test('VarInsert(v, 1, 0)', success);
  Test('v', [0, ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8, 9, 10]);
  Test('VarInsert(v, 12, 11)', success);
  Test('v', [0, ImaginaryUnit, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]);
  Test('VarInsert(v, 7, 5.5)', success);
  Test('v', [0, ImaginaryUnit, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11]);
  Test('VarInsert(v, −1, 12)', success);
  Test('v', [0, ImaginaryUnit, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11, 12]);
  Test('VarInsert(v, 16, 13)', failure, 'Index 16 out of bounds.');
  Test('VarInsert(v, -16, 13)', failure, 'Index -16 out of bounds.');
  Test('VarInsert(v, 0, 13)', failure, 'Index 0 out of bounds.');
  Test('VarInsert(v, 394, 13)', failure, 'Index 394 out of bounds.');
  Test('protect(v)', success);
  Test('VarInsert(v, 15, 13)', failure, 'Left side cannot be assigned to.');
  Test('v', [0, ImaginaryUnit, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11, 12]);
  Test('unprotect(v)', success);
  Test('VarInsert(v, 15, 13)', success);
  Test('v', [0, ImaginaryUnit, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11, 12, 13]);
  Test('VarInsert(v, −1, i)', success);
  Test('VarInsert(v, −1, "test")', failure, 'Object isn''t a complex number.');
  Test('v', [0, ImaginaryUnit, 2, 3, 4, 5, 5.5, 6, 7, 8, 9, 10, 11, 12, 13, ImaginaryUnit]);
  Test('VarInsert(❨1, 2, 3❩, 2, 10)', failure, 'Left side cannot be assigned to.');

  Test('A ≔ ❨❨3, 5❩, ❨1, 4❩❩', 2, [3, 5, 1, 4]);
  Test('VarInsert(A, 2, 10)', failure, 'Object type "real matrix" is not a suitable container.');
  Test('A ≔ ❨❨3, 5❩, ❨i, 4❩❩', 2, [3, 5, ImaginaryUnit, 4]);
  Test('VarInsert(A, 2, 10)', failure, 'Object type "complex matrix" is not a suitable container.');

  Test('s ≔ "Turn to 394"', 'Turn to 394');
  Test('VarInsert(s, 9, "page ")', success);
  Test('s', 'Turn to page 394');
  Test('VarInsert(s, −1, ".")', success);
  Test('s', 'Turn to page 394.');

  Test('s ≔ "Turn to page ."', 'Turn to page .');
  Test('VarInsert(s, −2, 394)', success);
  Test('s', 'Turn to page 394.');

  Test('L ≔ ''(10, 20, 30, 50, 60, 70)', intarr([10, 20, 30, 50, 60, 70]));
  Test('VarInsert(L, 4, "fourty")', success);
  Test('L', asoarr([10, 20, 30, 'fourty', 50, 60, 70]));
  Test('VarInsert(L, -1, 80)', success);
  Test('L', asoarr([10, 20, 30, 'fourty', 50, 60, 70, 80]));
  Test('VarInsert(L, -9, 0)', success);
  Test('L', asoarr([0, 10, 20, 30, 'fourty', 50, 60, 70, 80]));
  Test('VarInsert(L, -11, 0)', failure, 'Index -11 out of bounds.');
  Test('L', asoarr([0, 10, 20, 30, 'fourty', 50, 60, 70, 80]));

  Test('S ≔ {1, 2, 3}', asoset([1, 2, 3]));
  Test('VarInsert(S, 2, 4)', failure, 'Object type "set" is not a suitable container.');

  Test('VarInsert(❨1, 2, 3❩, 2, 1)', failure, 'Left side cannot be assigned to.');

  Test('delete(v); delete(A); delete(L); delete(S); delete(s)', success);

  // VarRemove

  Test('v ≔ ❨3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8❩',
    [3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8]);

  Test('VarRemove(v, 1)', success);
  Test('v', [1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8]);

  Test('VarRemove(v, -1)', success);
  Test('v', [1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3]);

  Test('VarRemove(v, range(5, 6))', success);
  Test('v', [1, 4, 1, 5, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3]);

  Test('VarRemove(v, ''(2, 4, 6))', success);
  Test('v', [1, 1, 6, 3, 5, 8, 9, 7, 9, 3, 2, 3]);

  Test('VarRemove(v, ''(1, 3, range(-1, -3)))', success);
  Test('v', [1, 3, 5, 8, 9, 7, 9]);

  Test('protect(v)', success);
  Test('VarRemove(v, ''(1, -1, range(3, 5)))', failure, 'Left side cannot be assigned to.');
  Test('v', [1, 3, 5, 8, 9, 7, 9]);

  Test('unprotect(v)', success);
  Test('VarRemove(v, ''(1, -1, range(3, 5)))', success);
  Test('v', [3, 7]);

  Test('VarRemove(v, ''(1, -1, range(3, 5)))', success);
  Test('v', TAlgosimRealVector.Create);

  Test('v ≔ ❨3, i, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8❩',
    [3, ImaginaryUnit, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8]);

  Test('VarRemove(v, 1)', success);
  Test('v', [ImaginaryUnit, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8]);

  Test('VarRemove(v, -1)', success);
  Test('v', [ImaginaryUnit, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3]);

  Test('VarRemove(v, range(5, 6))', success);
  Test('v', [ImaginaryUnit, 4, 1, 5, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3]);

  Test('VarRemove(v, ''(2, 4, 6))', success);
  Test('v', [ImaginaryUnit, 1, 6, 3, 5, 8, 9, 7, 9, 3, 2, 3]);

  Test('VarRemove(v, ''(1, 3, range(-1, -3)))', success);
  Test('v', [TASC(1), 3, 5, 8, 9, 7, 9]);

  Test('VarRemove(v, ''(1, -1, range(3, 5)))', success);
  Test('v', [TASC(3), 7]);

  Test('VarRemove(v, ''(1, -1, range(3, 5)))', success);
  Test('v', TAlgosimComplexVector.Create);

  Test('L ≔ ''(3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8)',
    intarr([3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8]));

  Test('VarRemove(L, 1)', success);
  Test('L', intarr([1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8]));

  Test('VarRemove(L, -1)', success);
  Test('L', intarr([1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3]));

  Test('VarRemove(L, range(5, 6))', success);
  Test('L', intarr([1, 4, 1, 5, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3]));

  Test('VarRemove(L, ''(2, 4, 6))', success);
  Test('L', intarr([1, 1, 6, 3, 5, 8, 9, 7, 9, 3, 2, 3]));

  Test('VarRemove(L, ''(1, 3, range(-1, -3)))', success);
  Test('L', intarr([1, 3, 5, 8, 9, 7, 9]));

  Test('VarRemove(L, ''(1, -1, range(3, 5)))', success);
  Test('L', intarr([3, 7]));

  Test('VarRemove(L, ''(1, -1, range(3, 5)))', success);
  Test('L', intarr([]));

  Test('s ≔ "Turn to pages 394–350, now."', 'Turn to pages 394–350, now.');
  Test('VarRemove(s, ''(13, range(18, 21), range(−2, −6)))', success);
  Test('s', 'Turn to page 394.');

  Test('S ≔ {1, 2, 3}', asoset([1, 2, 3]));
  Test('VarRemove(S, 2)', failure, 'Object type "set" is not a suitable container.');

  Test('VarRemove("Turn to pages 394–350, now.", ''(13, range(18, 21), range(−2, −6)))', failure, 'Left side cannot be assigned to.');

  Test('delete(v); delete(L); delete(S); delete(s)', success);

  // VarTruncate

  Test('v ≔ ❨3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8❩',
    [3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8]);

  Test('VarTruncate(v, 5)', success);
  Test('v', [3, 1, 4, 1, 5]);

  Test('VarTruncate(v, 5)', success);
  Test('v', [3, 1, 4, 1, 5]);

  Test('VarTruncate(v, -2)', failure, 'New length must be non-negative.');
  Test('v', [3, 1, 4, 1, 5]);

  Test('VarTruncate(v, 10)', success);
  Test('v', [3, 1, 4, 1, 5]);

  Test('protect(v)', success);
  Test('VarTruncate(v, 1)', failure, 'Left side cannot be assigned to.');
  Test('v', [3, 1, 4, 1, 5]);

  Test('unprotect(v)', success);
  Test('VarTruncate(v, 1)', success);
  Test('v', [3]);

  Test('VarTruncate(v, 0)', success);
  Test('v', TAlgosimRealVector.Create);

  Test('v ≔ ❨3, i, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8❩',
    [3, ImaginaryUnit, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8]);

  Test('VarTruncate(v, 5)', success);
  Test('v', [3, ImaginaryUnit, 4, 1, 5]);

  Test('VarTruncate(v, 5)', success);
  Test('v', [3, ImaginaryUnit, 4, 1, 5]);

  Test('VarTruncate(v, 10)', success);
  Test('v', [3, ImaginaryUnit, 4, 1, 5]);

  Test('protect(v)', success);
  Test('VarTruncate(v, 1)', failure, 'Left side cannot be assigned to.');
  Test('v', [3, ImaginaryUnit, 4, 1, 5]);
  Test('unprotect(v)', success);
  Test('VarTruncate(v, 1)', success);
  Test('v', [TASC(3)]);

  Test('VarTruncate(v, 0)', success);
  Test('v', TAlgosimComplexVector.Create);

  Test('s ≔ "Turn to page 394. Now."', 'Turn to page 394. Now.');
  Test('VarTruncate(s, 17)', success);
  Test('s', 'Turn to page 394.');

  Test('L ≔ ''(3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8)',
    intarr([3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9, 3, 2, 3, 8]));

  Test('VarTruncate(L, 5)', success);
  Test('L', intarr([3, 1, 4, 1, 5]));

  Test('VarTruncate(L, 5)', success);
  Test('L', intarr([3, 1, 4, 1, 5]));

  Test('VarTruncate(L, -2)', failure, 'New length must be non-negative.');
  Test('L', intarr([3, 1, 4, 1, 5]));

  Test('VarTruncate(L, 10)', success);
  Test('L', intarr([3, 1, 4, 1, 5]));

  Test('protect(L)', success);
  Test('VarTruncate(L, 1)', failure, 'Left side cannot be assigned to.');
  Test('L', intarr([3, 1, 4, 1, 5]));

  Test('unprotect(L)', success);
  Test('VarTruncate(L, 1)', success);
  Test('L', intarr([3]));

  Test('VarTruncate(L, 0)', success);
  Test('L', intarr([]));

  Test('S ≔ {1, 2, 3}', asoset([1, 2, 3]));
  Test('VarTruncate(S, 2)', failure, 'Object type "set" is not a suitable container.');

  Test('delete(v); delete(L); delete(S); delete(s)', success);

  // VarSwap

  Test('v ≔ ❨5, 3, 4, 2, 1, 8, 7, 5, 0, 4❩',
    [5, 3, 4, 2, 1, 8, 7, 5, 0, 4]);

  Test('VarSwap(v, 3, 6)', success);
  Test('v', [5, 3, 8, 2, 1, 4, 7, 5, 0, 4]);

  Test('VarSwap(v, 1, -1)', success);
  Test('v', [4, 3, 8, 2, 1, 4, 7, 5, 0, 5]);

  Test('VarSwap(v, 6, 2)', success);
  Test('v', [4, 4, 8, 2, 1, 3, 7, 5, 0, 5]);

  Test('VarSwap(v, 3, 8)', success);
  Test('v', [4, 4, 5, 2, 1, 3, 7, 8, 0, 5]);

  Test('VarSwap(v, -1, -2)', success);
  Test('v', [4, 4, 5, 2, 1, 3, 7, 8, 5, 0]);

  Test('VarSwap(v, 5, 11)', failure, 'Index 11 out of bounds.');
  Test('v', [4, 4, 5, 2, 1, 3, 7, 8, 5, 0]);

  Test('VarSwap(v, -12, 5)', failure, 'Index -12 out of bounds.');
  Test('v', [4, 4, 5, 2, 1, 3, 7, 8, 5, 0]);

  Test('VarSwap(v, 5, 0)', failure, 'Index 0 out of bounds.');
  Test('v', [4, 4, 5, 2, 1, 3, 7, 8, 5, 0]);

  Test('protect(v)', success);
  Test('VarSwap(v, 3, -1)', failure, 'Left side cannot be assigned to.');
  Test('v', [4, 4, 5, 2, 1, 3, 7, 8, 5, 0]);

  Test('unprotect(v)', success);
  Test('VarSwap(v, 3, -1)', success);
  Test('v', [4, 4, 0, 2, 1, 3, 7, 8, 5, 5]);

  Test('v ≔ ❨5, 3, 4, 2, 1, i, 7, 5, 0, 4❩',
    [5, 3, 4, 2, 1, ImaginaryUnit, 7, 5, 0, 4]);

  Test('VarSwap(v, 3, 6)', success);
  Test('v', [5, 3, ImaginaryUnit, 2, 1, 4, 7, 5, 0, 4]);

  Test('VarSwap(v, 1, -1)', success);
  Test('v', [4, 3, ImaginaryUnit, 2, 1, 4, 7, 5, 0, 5]);

  Test('VarSwap(v, 6, 2)', success);
  Test('v', [4, 4, ImaginaryUnit, 2, 1, 3, 7, 5, 0, 5]);

  Test('VarSwap(v, 3, 8)', success);
  Test('v', [4, 4, 5, 2, 1, 3, 7, ImaginaryUnit, 0, 5]);

  Test('VarSwap(v, -1, -2)', success);
  Test('v', [4, 4, 5, 2, 1, 3, 7, ImaginaryUnit, 5, 0]);

  Test('VarSwap(v, 5, 11)', failure, 'Index 11 out of bounds.');
  Test('v', [4, 4, 5, 2, 1, 3, 7, ImaginaryUnit, 5, 0]);

  Test('VarSwap(v, -12, 5)', failure, 'Index -12 out of bounds.');
  Test('v', [4, 4, 5, 2, 1, 3, 7, ImaginaryUnit, 5, 0]);

  Test('VarSwap(v, 5, 0)', failure, 'Index 0 out of bounds.');
  Test('v', [4, 4, 5, 2, 1, 3, 7, ImaginaryUnit, 5, 0]);

  Test('protect(v)', success);
  Test('VarSwap(v, 3, -1)', failure, 'Left side cannot be assigned to.');
  Test('v', [4, 4, 5, 2, 1, 3, 7, ImaginaryUnit, 5, 0]);

  Test('unprotect(v)', success);
  Test('VarSwap(v, 3, -1)', success);
  Test('v', [4, 4, 0, 2, 1, 3, 7, ImaginaryUnit, 5, 5]);

  Test('L ≔ ''(5, 3, 4, 2, 1, 8, 7, 5, 0, 4)',
    intarr([5, 3, 4, 2, 1, 8, 7, 5, 0, 4]));

  Test('VarSwap(L, 3, 6)', success);
  Test('L', intarr([5, 3, 8, 2, 1, 4, 7, 5, 0, 4]));

  Test('VarSwap(L, 1, -1)', success);
  Test('L', intarr([4, 3, 8, 2, 1, 4, 7, 5, 0, 5]));

  Test('VarSwap(L, 6, 2)', success);
  Test('L', intarr([4, 4, 8, 2, 1, 3, 7, 5, 0, 5]));

  Test('VarSwap(L, 3, 8)', success);
  Test('L', intarr([4, 4, 5, 2, 1, 3, 7, 8, 0, 5]));

  Test('VarSwap(L, -1, -2)', success);
  Test('L', intarr([4, 4, 5, 2, 1, 3, 7, 8, 5, 0]));

  Test('VarSwap(L, 5, 11)', failure, 'Index 11 out of bounds.');
  Test('L', intarr([4, 4, 5, 2, 1, 3, 7, 8, 5, 0]));

  Test('VarSwap(L, -12, 5)', failure, 'Index -12 out of bounds.');
  Test('L', intarr([4, 4, 5, 2, 1, 3, 7, 8, 5, 0]));

  Test('VarSwap(L, 5, 0)', failure, 'Index 0 out of bounds.');
  Test('L', intarr([4, 4, 5, 2, 1, 3, 7, 8, 5, 0]));

  Test('protect(L)', success);
  Test('VarSwap(L, 3, -1)', failure, 'Left side cannot be assigned to.');
  Test('L', intarr([4, 4, 5, 2, 1, 3, 7, 8, 5, 0]));

  Test('unprotect(L)', success);
  Test('VarSwap(L, 3, -1)', success);
  Test('L', intarr([4, 4, 0, 2, 1, 3, 7, 8, 5, 5]));

  Test('s ≔ "Turn to page 493."', 'Turn to page 493.');
  Test('VarSwap(s, -4, -2)', success);
  Test('s', 'Turn to page 394.');

  Test('S ≔ {1, 2, 3}', asoset([1, 2, 3]));
  Test('VarSwap(S, 2, 3)', failure, 'Object type "set" is not a suitable container.');

  Test('delete(v); delete(L); delete(S); delete(s)', success);


  //
  // Number extraction
  //

  Test('v ≔ ❨3, 1, 4, 1, 5, 9, 2, 6, 5❩;', null);
  Test('numbers(v)', asoarr([3, 1, 4, 1, 5, 9, 2, 6, 5]));
  Test('numbers(DebugObject("empty vector"))', asoarr([]));

  Test('v ≔ ❨i, 1, 4, 1, 5, 9, 2, 6, 5❩;', null);
  Test('numbers(v)', asoarrex([ASO(ImaginaryUnit), ASO(1), ASO(4), ASO(1), ASO(5), ASO(9), ASO(2), ASO(6), ASO(5)]));
  Test('numbers(ComplexVector(DebugObject("empty vector")))', asoarr([]));

  Test('A ≔ ❨❨5, 0, 1, 3❩, ❨−2, 1, 4, 8❩, ❨0, 1, 3, 9❩, ❨2, 1, 0, 7❩❩;', null);
  Test('numbers(A)', asoarr([5, 0, 1, 3, -2, 1, 4, 8, 0, 1, 3, 9, 2, 1, 0, 7]));
  Test('numbers(DebugObject("empty matrix"))', asoarr([]));

  Test('A ≔ ❨❨5, i, 1, 3❩, ❨−2, 1, 4, 8❩, ❨0, 1, 3, 9❩, ❨2, 1, 0, 7❩❩;', null);
  Test('numbers(A)', asoarrex([ASO(5), ASO(ImaginaryUnit), ASO(1), ASO(3), ASO(-2), ASO(1), ASO(4), ASO(8), ASO(0), ASO(1), ASO(3), ASO(9), ASO(2), ASO(1), ASO(0), ASO(7)]));
  Test('numbers(ComplexMatrix(DebugObject("empty matrix")))', asoarr([]));

  Test('L ≔ ''(3, 1, 4, 1, 5, 9, 2, 6, 5);', null);
  Test('numbers(L)', asoarr([3, 1, 4, 1, 5, 9, 2, 6, 5]));

  Test('L ≔ ''();', null);
  Test('numbers(L)', asoarr([]));

  Test('S ≔ {3, 1, 4, 1, 5, 9, 2, 6, 5};', null);
  Test('ToSet(numbers(S))', asoset([3, 1, 4, 5, 9, 2, 6]));

  Test('S ≔ {};', null);
  Test('numbers(S)', asoarr([]));

  Test('S ≔ struct("name": "Albus Dumbledore", "age": 116, "role": "headmaster");', null);
  Test('numbers(S)', intarr([116]));

  Test('S ≔ struct("name": "Albus Dumbledore", "role": "headmaster");', null);
  Test('numbers(S)', asoarr([]));

  Test('L ≔ ''(6, 2, 1, ❨❨5, 6, 3❩, ❨1, 4, 2❩❩, 6.5, ❨21, 5, 22❩, {1024, 2048}, ''(555, 666, ❨−10, −52, −40❩, ''(123456, 321654, {5040})), −999);', null);
  Test('numbers(L)', asoarr([6, 2, 1, 5, 6, 3, 1, 4, 2, 6.5, 21, 5, 22, 1024, 2048, 555, 666, -10, -52, -40, 123456, 321654, 5040, -999]));

  Test('numbers(''(3, 1, ❨7, 2, 1❩))', asoarr([3, 1, 7, 2, 1]));
  Test('numbers(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(394))))))))))))))))))))))))))', intarr([394]));
  Test('numbers(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''(''())))))))))))))))))))))))))', intarr([]));

  Test('delete(v); delete(A); delete(L); delete(S)', success);


  //
  // Images
  //

  Test('image(identity, 5)', 5);
  Test('image(factorial, 5)', 120);
  Test('image(square, 5)', 25);
  Test('image(inv, 5)', rat(1, 5));
  Test('image(gcd, 5265, 2550)', 15);
  Test('image(add, 100, 25)', 125);

  Test('image(divide, 5, 0)', failure, 'Division by zero');
  Test('image(divide, 5)', failure, 'Too few arguments. A required argument of type object is missing.');
  Test('image(list)', asoarr([]));
  Test('image(vector, 1, 2, 3, 4, 5)', [1, 2, 3, 4, 5]);

  Test('IteratedImage(n ↦ 2⋅n, 1, 0)', 1);
  Test('IteratedImage(n ↦ 2⋅n, 1, 1)', 2);
  Test('IteratedImage(n ↦ 2⋅n, 1, 2)', 4);
  Test('IteratedImage(n ↦ 2⋅n, 1, 3)', 8);
  Test('IteratedImage(n ↦ 2⋅n, 1, 4)', 16);
  Test('IteratedImage(n ↦ 2⋅n, 1, 5)', 32);
  Test('IteratedImage(n ↦ 2⋅n, 1, 6)', 64);
  Test('IteratedImage(n ↦ 2⋅n, 1, 7)', 128);
  Test('IteratedImage(n ↦ 2⋅n, 1, 8)', 256);
  Test('IteratedImage(n ↦ 2⋅n, 1, 9)', 512);
  Test('IteratedImage(n ↦ 2⋅n, 1, 10)', 1024);

  Test('IteratedImage(n ↦ 2⋅n, 1, -1)', failure, 'A non-negative integer was expected as argument 3, but "-1" was given.');

  Test('IteratedImage(sin, 1, 0)', 1);
  Eps; Test('IteratedImage(sin, 1, 1)', 0.841470984807896507);
  Eps; Test('IteratedImage(sin, 1, 2)', 0.745624141665557889);
  Eps; Test('IteratedImage(sin, 1, 3)', 0.678430477360740229);
  Eps; Test('IteratedImage(sin, 1, 4)', 0.627571832049159139);
  Eps; Test('IteratedImage(sin, 1, 5)', 0.587180996573430989);
  Eps; Test('IteratedImage(sin, 1, 6)', 0.55401639075562963);
  Eps; Test('IteratedImage(sin, 1, 7)', 0.526107075502841702);
  Eps; Test('IteratedImage(sin, 1, 8)', 0.502170676268555349);
  Eps; Test('IteratedImage(sin, 1, 9)', 0.481329355262346345);
  Eps; Test('IteratedImage(sin, 1, 10)', 0.462957898537811869);

  Test('IteratedImages(n ↦ 2⋅n, 1, 0)', intarr([1]));
  Test('IteratedImages(n ↦ 2⋅n, 1, 1)', intarr([1, 2]));
  Test('IteratedImages(n ↦ 2⋅n, 1, 10)', intarr([1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]));

  Test('IteratedImages(n ↦ 2⋅n, 1, -1)', failure, 'A non-negative integer was expected as argument 3, but "-1" was given.');

  Test('orbit(n ↦ 2⋅n, 1, 0)', intset([1]));
  Test('orbit(n ↦ 2⋅n, 1, 1)', intset([1, 2]));
  Test('orbit(n ↦ 2⋅n, 1, 10)', intset([1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]));

  Test('orbit(n ↦ 2⋅n, 1, -1)', failure, 'A non-negative integer was expected as argument 3, but "-1" was given.');

  Test('f ≔ n ↦ if(even(n), n/2, 3⋅n + 1);', null);

  Test('IteratedImages(f, 12, 9)', intarr([12, 6, 3, 10, 5, 16, 8, 4, 2, 1]));
  Test('IteratedImages(f, 19, 20)', intarr([19, 58, 29, 88, 44, 22, 11, 34, 17, 52, 26, 13, 40, 20, 10, 5, 16, 8, 4, 2, 1]));
  Test('IteratedImages(f, 27, 111)', intarr([27, 82, 41, 124, 62, 31, 94, 47, 142, 71, 214, 107, 322, 161, 484, 242, 121, 364, 182, 91, 274, 137, 412, 206, 103, 310, 155, 466, 233, 700, 350, 175, 526, 263, 790, 395, 1186, 593, 1780, 890, 445, 1336, 668, 334, 167, 502, 251, 754, 377, 1132, 566, 283, 850, 425, 1276, 638, 319, 958, 479, 1438, 719, 2158, 1079, 3238, 1619, 4858, 2429, 7288, 3644, 1822, 911, 2734, 1367, 4102, 2051, 6154, 3077, 9232, 4616, 2308, 1154, 577, 1732, 866, 433, 1300, 650, 325, 976, 488, 244, 122, 61, 184, 92, 46, 23, 70, 35, 106, 53, 160, 80, 40, 20, 10, 5, 16, 8, 4, 2, 1]));

  Test('orbit(f, 12, 9)', intset([12, 6, 3, 10, 5, 16, 8, 4, 2, 1]));
  Test('orbit(f, 19, 20)', intset([19, 58, 29, 88, 44, 22, 11, 34, 17, 52, 26, 13, 40, 20, 10, 5, 16, 8, 4, 2, 1]));
  Test('orbit(f, 27, 111)', intset([27, 82, 41, 124, 62, 31, 94, 47, 142, 71, 214, 107, 322, 161, 484, 242, 121, 364, 182, 91, 274, 137, 412, 206, 103, 310, 155, 466, 233, 700, 350, 175, 526, 263, 790, 395, 1186, 593, 1780, 890, 445, 1336, 668, 334, 167, 502, 251, 754, 377, 1132, 566, 283, 850, 425, 1276, 638, 319, 958, 479, 1438, 719, 2158, 1079, 3238, 1619, 4858, 2429, 7288, 3644, 1822, 911, 2734, 1367, 4102, 2051, 6154, 3077, 9232, 4616, 2308, 1154, 577, 1732, 866, 433, 1300, 650, 325, 976, 488, 244, 122, 61, 184, 92, 46, 23, 70, 35, 106, 53, 160, 80, 40, 20, 10, 5, 16, 8, 4, 2, 1]));

  Test('orbit(f, 12, 10000)', intset([12, 6, 3, 10, 5, 16, 8, 4, 2, 1]));
  Test('orbit(f, 19, 10000)', intset([19, 58, 29, 88, 44, 22, 11, 34, 17, 52, 26, 13, 40, 20, 10, 5, 16, 8, 4, 2, 1]));
  Test('orbit(f, 27, 10000)', intset([27, 82, 41, 124, 62, 31, 94, 47, 142, 71, 214, 107, 322, 161, 484, 242, 121, 364, 182, 91, 274, 137, 412, 206, 103, 310, 155, 466, 233, 700, 350, 175, 526, 263, 790, 395, 1186, 593, 1780, 890, 445, 1336, 668, 334, 167, 502, 251, 754, 377, 1132, 566, 283, 850, 425, 1276, 638, 319, 958, 479, 1438, 719, 2158, 1079, 3238, 1619, 4858, 2429, 7288, 3644, 1822, 911, 2734, 1367, 4102, 2051, 6154, 3077, 9232, 4616, 2308, 1154, 577, 1732, 866, 433, 1300, 650, 325, 976, 488, 244, 122, 61, 184, 92, 46, 23, 70, 35, 106, 53, 160, 80, 40, 20, 10, 5, 16, 8, 4, 2, 1]));

  Test('orbit(f, 12)', intset([12, 6, 3, 10, 5, 16, 8, 4, 2, 1]));
  Test('orbit(f, 19)', intset([19, 58, 29, 88, 44, 22, 11, 34, 17, 52, 26, 13, 40, 20, 10, 5, 16, 8, 4, 2, 1]));
  Test('orbit(f, 27)', intset([27, 82, 41, 124, 62, 31, 94, 47, 142, 71, 214, 107, 322, 161, 484, 242, 121, 364, 182, 91, 274, 137, 412, 206, 103, 310, 155, 466, 233, 700, 350, 175, 526, 263, 790, 395, 1186, 593, 1780, 890, 445, 1336, 668, 334, 167, 502, 251, 754, 377, 1132, 566, 283, 850, 425, 1276, 638, 319, 958, 479, 1438, 719, 2158, 1079, 3238, 1619, 4858, 2429, 7288, 3644, 1822, 911, 2734, 1367, 4102, 2051, 6154, 3077, 9232, 4616, 2308, 1154, 577, 1732, 866, 433, 1300, 650, 325, 976, 488, 244, 122, 61, 184, 92, 46, 23, 70, 35, 106, 53, 160, 80, 40, 20, 10, 5, 16, 8, 4, 2, 1]));


  //
  // IsX functions
  //

  Test('IsNumber(123)', True);
  Test('IsNumber(123.0)', True);
  Test('IsNumber(25.36)', True);
  Test('IsNumber(1/2)', True);
  Test('IsNumber(1 + i)', True);
  Test('IsNumber(❨1, 0❩)', False);
  Test('IsNumber(❨i, 0❩)', False);
  Test('IsNumber(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsNumber(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsNumber(''(1, 2, 3))', False);
  Test('IsNumber({1, 2, 3})', False);
  Test('IsNumber(true)', False);
  Test('IsNumber("nargle")', False);
  Test('IsNumber(CreatePixmap(512, 512))', False);
  Test('IsNumber(SineTone(400))', False);
  Test('IsNumber(color("red"))', False);
  Test('IsNumber("name": "Albus")', False);
  Test('IsNumber(BinaryData("nargle"))', False);

  Test('IsVector(123)', False);
  Test('IsVector(123.0)', False);
  Test('IsVector(25.36)', False);
  Test('IsVector(1/2)', False);
  Test('IsVector(1 + i)', False);
  Test('IsVector(❨1, 0❩)', True);
  Test('IsVector(❨i, 0❩)', True);
  Test('IsVector(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsVector(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsVector(''(1, 2, 3))', False);
  Test('IsVector({1, 2, 3})', False);
  Test('IsVector(true)', False);
  Test('IsVector("nargle")', False);
  Test('IsVector(CreatePixmap(512, 512))', False);
  Test('IsVector(SineTone(400))', False);
  Test('IsVector(color("red"))', False);
  Test('IsVector("name": "Albus")', False);
  Test('IsVector(BinaryData("nargle"))', False);

  Test('IsMatrix(123)', False);
  Test('IsMatrix(123.0)', False);
  Test('IsMatrix(25.36)', False);
  Test('IsMatrix(1/2)', False);
  Test('IsMatrix(1 + i)', False);
  Test('IsMatrix(❨1, 0❩)', False);
  Test('IsMatrix(❨i, 0❩)', False);
  Test('IsMatrix(❨❨3, 2❩, ❨0, 1❩❩)', True);
  Test('IsMatrix(❨❨3, i❩, ❨0, 1❩❩)', True);
  Test('IsMatrix(''(1, 2, 3))', False);
  Test('IsMatrix({1, 2, 3})', False);
  Test('IsMatrix(true)', False);
  Test('IsMatrix("nargle")', False);
  Test('IsMatrix(CreatePixmap(512, 512))', False);
  Test('IsMatrix(SineTone(400))', False);
  Test('IsMatrix(color("red"))', False);
  Test('IsMatrix("name": "Albus")', False);
  Test('IsMatrix(BinaryData("nargle"))', False);

  Test('IsText(123)', False);
  Test('IsText(123.0)', False);
  Test('IsText(25.36)', False);
  Test('IsText(1/2)', False);
  Test('IsText(1 + i)', False);
  Test('IsText(❨1, 0❩)', False);
  Test('IsText(❨i, 0❩)', False);
  Test('IsText(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsText(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsText(''(1, 2, 3))', False);
  Test('IsText({1, 2, 3})', False);
  Test('IsText(true)', False);
  Test('IsText("nargle")', True);
  Test('IsText(CreatePixmap(512, 512))', False);
  Test('IsText(SineTone(400))', False);
  Test('IsText(color("red"))', False);
  Test('IsText("name": "Albus")', False);
  Test('IsText(BinaryData("nargle"))', False);

  Test('IsCharacter(123)', False);
  Test('IsCharacter(123.0)', False);
  Test('IsCharacter(25.36)', False);
  Test('IsCharacter(1/2)', False);
  Test('IsCharacter(1 + i)', False);
  Test('IsCharacter(❨1, 0❩)', False);
  Test('IsCharacter(❨i, 0❩)', False);
  Test('IsCharacter(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsCharacter(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsCharacter(''(1, 2, 3))', False);
  Test('IsCharacter({1, 2, 3})', False);
  Test('IsCharacter(true)', False);
  Test('IsCharacter("nargle")', False);
  Test('IsCharacter("n")', True);
  Test('IsCharacter(" ")', True);
  Test('IsCharacter("!")', True);
  Test('IsCharacter("∮")', True);
  Test('IsCharacter("""")', True);
  Test('IsCharacter(chr(9))', True);
  Test('IsCharacter(CreatePixmap(512, 512))', False);
  Test('IsCharacter(SineTone(400))', False);
  Test('IsCharacter(color("red"))', False);
  Test('IsCharacter("name": "Albus")', False);
  Test('IsCharacter(BinaryData("nargle"))', False);

  Test('IsBoolean(123)', False);
  Test('IsBoolean(123.0)', False);
  Test('IsBoolean(25.36)', False);
  Test('IsBoolean(1/2)', False);
  Test('IsBoolean(1 + i)', False);
  Test('IsBoolean(❨1, 0❩)', False);
  Test('IsBoolean(❨i, 0❩)', False);
  Test('IsBoolean(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsBoolean(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsBoolean(''(1, 2, 3))', False);
  Test('IsBoolean({1, 2, 3})', False);
  Test('IsBoolean(true)', True);
  Test('IsBoolean("nargle")', False);
  Test('IsBoolean(CreatePixmap(512, 512))', False);
  Test('IsBoolean(SineTone(400))', False);
  Test('IsBoolean(color("red"))', False);
  Test('IsBoolean("name": "Albus")', False);
  Test('IsBoolean(BinaryData("nargle"))', False);

  Test('IsPixmap(123)', False);
  Test('IsPixmap(123.0)', False);
  Test('IsPixmap(25.36)', False);
  Test('IsPixmap(1/2)', False);
  Test('IsPixmap(1 + i)', False);
  Test('IsPixmap(❨1, 0❩)', False);
  Test('IsPixmap(❨i, 0❩)', False);
  Test('IsPixmap(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsPixmap(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsPixmap(''(1, 2, 3))', False);
  Test('IsPixmap({1, 2, 3})', False);
  Test('IsPixmap(true)', False);
  Test('IsPixmap("nargle")', False);
  Test('IsPixmap(CreatePixmap(512, 512))', True);
  Test('IsPixmap(SineTone(400))', False);
  Test('IsPixmap(color("red"))', False);
  Test('IsPixmap("name": "Albus")', False);
  Test('IsPixmap(BinaryData("nargle"))', False);

  Test('IsSound(123)', False);
  Test('IsSound(123.0)', False);
  Test('IsSound(25.36)', False);
  Test('IsSound(1/2)', False);
  Test('IsSound(1 + i)', False);
  Test('IsSound(❨1, 0❩)', False);
  Test('IsSound(❨i, 0❩)', False);
  Test('IsSound(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsSound(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsSound(''(1, 2, 3))', False);
  Test('IsSound({1, 2, 3})', False);
  Test('IsSound(true)', False);
  Test('IsSound("nargle")', False);
  Test('IsSound(CreatePixmap(512, 512))', False);
  Test('IsSound(SineTone(400))', True);
  Test('IsSound(color("red"))', False);
  Test('IsSound("name": "Albus")', False);
  Test('IsSound(BinaryData("nargle"))', False);

  Test('IsTable(123)', False);
  Test('IsTable(123.0)', False);
  Test('IsTable(25.36)', False);
  Test('IsTable(1/2)', False);
  Test('IsTable(1 + i)', False);
  Test('IsTable(❨1, 0❩)', False);
  Test('IsTable(❨i, 0❩)', False);
  Test('IsTable(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsTable(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsTable(''(1, 2, 3))', False);
  Test('IsTable({1, 2, 3})', False);
  Test('IsTable(true)', False);
  Test('IsTable("nargle")', False);
  Test('IsTable(CreatePixmap(512, 512))', False);
  Test('IsTable(SineTone(400))', False);
  Test('IsTable(color("red"))', False);
  Test('IsTable("name": "Albus")', False);
  Test('IsTable(BinaryData("nargle"))', False);

  Test('IsColor(123)', False);
  Test('IsColor(123.0)', False);
  Test('IsColor(25.36)', False);
  Test('IsColor(1/2)', False);
  Test('IsColor(1 + i)', False);
  Test('IsColor(❨1, 0❩)', False);
  Test('IsColor(❨i, 0❩)', False);
  Test('IsColor(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsColor(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsColor(''(1, 2, 3))', False);
  Test('IsColor({1, 2, 3})', False);
  Test('IsColor(true)', False);
  Test('IsColor("nargle")', False);
  Test('IsColor(CreatePixmap(512, 512))', False);
  Test('IsColor(SineTone(400))', False);
  Test('IsColor(color("red"))', True);
  Test('IsColor("name": "Albus")', False);
  Test('IsColor(BinaryData("nargle"))', False);

  Test('IsSet(123)', False);
  Test('IsSet(123.0)', False);
  Test('IsSet(25.36)', False);
  Test('IsSet(1/2)', False);
  Test('IsSet(1 + i)', False);
  Test('IsSet(❨1, 0❩)', False);
  Test('IsSet(❨i, 0❩)', False);
  Test('IsSet(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsSet(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsSet(''(1, 2, 3))', False);
  Test('IsSet({1, 2, 3})', True);
  Test('IsSet(true)', False);
  Test('IsSet("nargle")', False);
  Test('IsSet(CreatePixmap(512, 512))', False);
  Test('IsSet(SineTone(400))', False);
  Test('IsSet(color("red"))', False);
  Test('IsSet("name": "Albus")', False);
  Test('IsSet(BinaryData("nargle"))', False);

  Test('IsList(123)', False);
  Test('IsList(123.0)', False);
  Test('IsList(25.36)', False);
  Test('IsList(1/2)', False);
  Test('IsList(1 + i)', False);
  Test('IsList(❨1, 0❩)', False);
  Test('IsList(❨i, 0❩)', False);
  Test('IsList(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsList(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsList(''(1, 2, 3))', True);
  Test('IsList({1, 2, 3})', False);
  Test('IsList(true)', False);
  Test('IsList("nargle")', False);
  Test('IsList(CreatePixmap(512, 512))', False);
  Test('IsList(SineTone(400))', False);
  Test('IsList(color("red"))', False);
  Test('IsList("name": "Albus")', False);
  Test('IsList(BinaryData("nargle"))', False);

  Test('IsStructure(123)', False);
  Test('IsStructure(123.0)', False);
  Test('IsStructure(25.36)', False);
  Test('IsStructure(1/2)', False);
  Test('IsStructure(1 + i)', False);
  Test('IsStructure(❨1, 0❩)', False);
  Test('IsStructure(❨i, 0❩)', False);
  Test('IsStructure(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsStructure(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsStructure(''(1, 2, 3))', False);
  Test('IsStructure({1, 2, 3})', False);
  Test('IsStructure(true)', False);
  Test('IsStructure("nargle")', False);
  Test('IsStructure(CreatePixmap(512, 512))', False);
  Test('IsStructure(SineTone(400))', False);
  Test('IsStructure(color("red"))', False);
  Test('IsStructure("name": "Albus")', True);
  Test('IsStructure(BinaryData("nargle"))', False);

  Test('IsReal(123)', True);
  Test('IsReal(123.0)', True);
  Test('IsReal(123 + i - i)', True);
  Test('IsReal(25.36)', True);
  Test('IsReal(1/2)', True);
  Test('IsReal(1 + i)', False);
  Test('IsReal(❨1, 0❩)', True);
  Test('IsReal(❨1 + i - i, 0❩)', True);
  Test('IsReal(❨i, 0❩)', False);
  Test('IsReal(❨❨3, 2❩, ❨0, 1❩❩)', True);
  Test('IsReal(❨❨3, 2 + i - i❩, ❨0, 1❩❩)', True);
  Test('IsReal(❨❨3, i❩, ❨0, 1❩❩)', False);

  Test('IsComplex(123)', False);
  Test('IsComplex(123.0)', False);
  Test('IsComplex(123 + i - i)', False);
  Test('IsComplex(25.36)', False);
  Test('IsComplex(1/2)', False);
  Test('IsComplex(1 + i)', True);
  Test('IsComplex(❨1, 0❩)', False);
  Test('IsComplex(❨1 + i - i, 0❩)', False);
  Test('IsComplex(❨i, 0❩)', True);
  Test('IsComplex(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsComplex(❨❨3, 2 + i - i❩, ❨0, 1❩❩)', False);
  Test('IsComplex(❨❨3, i❩, ❨0, 1❩❩)', True);

  Test('IsInteger(123)', True);
  Test('IsInteger(123.0)', True);
  Test('IsInteger(25.36)', False);
  Test('IsInteger(1/2)', False);
  Test('IsInteger(1 + i)', False);
  Test('IsInteger(❨1, 0❩)', False);
  Test('IsInteger(❨i, 0❩)', False);
  Test('IsInteger(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsInteger(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsInteger(''(1, 2, 3))', False);
  Test('IsInteger({1, 2, 3})', False);
  Test('IsInteger(true)', False);
  Test('IsInteger("nargle")', False);
  Test('IsInteger(CreatePixmap(512, 512))', False);
  Test('IsInteger(SineTone(400))', False);
  Test('IsInteger(color("red"))', False);
  Test('IsInteger("name": "Albus")', False);
  Test('IsInteger(BinaryData("nargle"))', False);

  Test('IsNumericEntity(123)', True);
  Test('IsNumericEntity(123.0)', True);
  Test('IsNumericEntity(25.36)', True);
  Test('IsNumericEntity(1/2)', True);
  Test('IsNumericEntity(1 + i)', True);
  Test('IsNumericEntity(❨1, 0❩)', True);
  Test('IsNumericEntity(❨i, 0❩)', True);
  Test('IsNumericEntity(❨❨3, 2❩, ❨0, 1❩❩)', True);
  Test('IsNumericEntity(❨❨3, i❩, ❨0, 1❩❩)', True);
  Test('IsNumericEntity(''(1, 2, 3))', False);
  Test('IsNumericEntity({1, 2, 3})', False);
  Test('IsNumericEntity(true)', False);
  Test('IsNumericEntity("nargle")', False);
  Test('IsNumericEntity(CreatePixmap(512, 512))', False);
  Test('IsNumericEntity(SineTone(400))', False);
  Test('IsNumericEntity(color("red"))', False);
  Test('IsNumericEntity("name": "Albus")', False);
  Test('IsNumericEntity(BinaryData("nargle"))', False);

  Test('IsNumericArray(123)', False);
  Test('IsNumericArray(123.0)', False);
  Test('IsNumericArray(25.36)', False);
  Test('IsNumericArray(1/2)', False);
  Test('IsNumericArray(1 + i)', False);
  Test('IsNumericArray(❨1, 0❩)', True);
  Test('IsNumericArray(❨i, 0❩)', True);
  Test('IsNumericArray(❨❨3, 2❩, ❨0, 1❩❩)', True);
  Test('IsNumericArray(❨❨3, i❩, ❨0, 1❩❩)', True);
  Test('IsNumericArray(''(1, 2, 3))', False);
  Test('IsNumericArray({1, 2, 3})', False);
  Test('IsNumericArray(true)', False);
  Test('IsNumericArray("nargle")', False);
  Test('IsNumericArray(CreatePixmap(512, 512))', False);
  Test('IsNumericArray(SineTone(400))', False);
  Test('IsNumericArray(color("red"))', False);
  Test('IsNumericArray("name": "Albus")', False);
  Test('IsNumericArray(BinaryData("nargle"))', False);

  Test('IsBinaryData(123)', False);
  Test('IsBinaryData(123.0)', False);
  Test('IsBinaryData(25.36)', False);
  Test('IsBinaryData(1/2)', False);
  Test('IsBinaryData(1 + i)', False);
  Test('IsBinaryData(❨1, 0❩)', False);
  Test('IsBinaryData(❨i, 0❩)', False);
  Test('IsBinaryData(❨❨3, 2❩, ❨0, 1❩❩)', False);
  Test('IsBinaryData(❨❨3, i❩, ❨0, 1❩❩)', False);
  Test('IsBinaryData(''(1, 2, 3))', False);
  Test('IsBinaryData({1, 2, 3})', False);
  Test('IsBinaryData(true)', False);
  Test('IsBinaryData("nargle")', False);
  Test('IsBinaryData(CreatePixmap(512, 512))', False);
  Test('IsBinaryData(SineTone(400))', False);
  Test('IsBinaryData(color("red"))', False);
  Test('IsBinaryData("name": "Albus")', False);
  Test('IsBinaryData(BinaryData("nargle"))', True);


  //
  // Tokenizer and parser functions. Custom function expressions.
  //

  Test('tokenize("(2.1 + 3)⋅4! − sin(1 + 1/2^−n)")',
    strarr(
      [
        'Bracket ( – partner at idx 4',
        'Float literal 2.1',
        'Infix operator +',
        'Integer literal 3',
        'Bracket ) – partner at idx 0',
        'Infix operator ⋅',
        'Integer literal 4',
        'Postfix operator !',
        'Infix operator −',
        'Identifier "sin"',
        'Bracket ( applying on left value – partner at idx 19',
        'Integer literal 1',
        'Infix operator +',
        'Integer literal 1',
        'Infix operator /',
        'Integer literal 2',
        'Infix operator ^',
        'Prefix operator −',
        'Identifier "n"',
        'Bracket ) – partner at idx 10'
      ]
    )
  );

  Test('parse("(2.1 + 3)⋅4! − sin(1 + 1/2^−n)")',
    'FCN_Subtract'#13#10 +
    '    FCN_Multiply'#13#10 +
    '        FCN_Add'#13#10 +
    '            2.1'#13#10 +
    '            3'#13#10 +
    '        FCN_Factorial'#13#10 +
    '            4'#13#10 +
    '    FCN_sin'#13#10 +
    '        FCN_Add'#13#10 +
    '            1'#13#10 +
    '            FCN_Divide'#13#10 +
    '                1'#13#10 +
    '                FCN_Power'#13#10 +
    '                    2'#13#10 +
    '                    FCN_UnaryMinus'#13#10 +
    '                        n'
  );

  Test('f ≔ n ↦ n^2 + 1;', null);
  Test('FcnExpr(f)',
    'Arguments: n'#13#10 +
    'FCN_Add'#13#10 +
    '    FCN_Power'#13#10 +
    '        n'#13#10 +
    '        2'#13#10 +
    '    1'
  );
  Test('delete(f)', success);


  //
  // Buffer functions
  //

  Test('buffers()', strarr([]));

  Test('print("testbuf", "alpha")', null);

  Test('buffers()', strarr(['testbuf']));
  Test('BufferText("testbuf")', 'alpha'#13#10);

  Test('print("testbuf", "beta")', null);

  Test('buffers()', strarr(['testbuf']));
  Test('BufferText("testbuf")', 'alpha'#13#10'beta'#13#10);

  Test('print("testbuf", "gamma")', null);

  Test('buffers()', strarr(['testbuf']));
  Test('BufferText("testbuf")', 'alpha'#13#10'beta'#13#10'gamma'#13#10);

  Test('print("testbuf2", "delta")', null);

  Test('buffers()', strarr(['testbuf', 'testbuf2']));
  Test('BufferText("testbuf")', 'alpha'#13#10'beta'#13#10'gamma'#13#10);
  Test('BufferText("testbuf2")', 'delta'#13#10);

  Test('print("testbuf", "epsilon")', null);

  Test('buffers()', strarr(['testbuf', 'testbuf2']));
  Test('BufferText("testbuf")', 'alpha'#13#10'beta'#13#10'gamma'#13#10'epsilon'#13#10);
  Test('BufferText("testbuf2")', 'delta'#13#10);

  Test('RemoveBuffer("testbuf2")', null);

  Test('buffers()', strarr(['testbuf']));
  Test('BufferText("testbuf")', 'alpha'#13#10'beta'#13#10'gamma'#13#10'epsilon'#13#10);

  Test('RemoveBuffer("testbuf")', null);
  Test('buffers()', strarr([]));

  Test('RemoveBuffer("testbuf")', failure, 'There is no output buffer named "testbuf".');

  Test('print("!", "test")', failure, 'Invalid buffer name "!".');
  Test('print(" ", "test")', failure, 'Invalid buffer name " ".');
  Test('print("+buf", "test")', failure, 'Invalid buffer name "+buf".');
  Test('print(" buf", "test")', failure, 'Invalid buffer name " buf".');


  //
  // History functions
  //

  Test('SaveHistory(false)', null);
  Test('ClearHistory()', success);
  Test('1 + 1', 2);
  Test('5!', 120);
  Test('input(1)', 'ClearHistory()');
  Test('input(2)', '1 + 1');
  Test('input(3)', '5!');
  Test('output(1)', null);
  Test('output(2)', null);
  Test('output(3)', null);

  Test('HistoryLength()', 9);

  Test('SaveHistory(true)', null);
  Test('ClearHistory()', success);
  Test('1 + 1', 2);
  Test('5!', 120);
  Test('input(1)', 'ClearHistory()');
  Test('input(2)', '1 + 1');
  Test('input(3)', '5!');
  Test('output(1)', success);
  Test('output(2)', 2);
  Test('output(3)', 120);

  Test('SaveHistory(false)', null);
  Test('HistoryLength()', 10);


  //
  // Number base setting
  //

  Test('bases(10, 12)', null);
  Test('bases()', null);
  Test('bases(10)', null);


  //
  // Debug objects
  //

  Test('DebugObject("empty vector")', TAlgosimRealVector.Create);
  Test('DebugObject("empty matrix")', TAlgosimRealMatrix.Create);


  //
  // Property stores
  //

  Test('property("nargle")', failure, 'There is no property named "nargle" in store "global".');
  Test('property("")', failure, 'There is no property named "" in store "global".');
  Test('property("os")', failure, 'There is no property named "os" in store "global".');

  Test('type(property("os.UserName"))', 'string');
  Test('type(property("os.ComputerName"))', 'string');
  Test('type(property("os.WindowsVersion"))', 'structure');
  Test('type(property("os.MemoryStatus"))', 'structure');
  Test('type(property("os.CPUCount"))', 'integer');
  Test('type(property("os.DesktopWindow"))', 'integer');
  Test('type(property("os.LastError"))', 'integer');
  Test('type(property("os.focus"))', 'integer');
  Test('type(property("os.ForegroundWindow"))', 'integer');
  Test('type(property("os.ActiveWindow"))', 'integer');
  Test('type(property("os.CursorPos"))', 'structure');
  Test('type(property("os.CursorPos").x)', 'integer');
  Test('type(property("os.CursorPos").y)', 'integer');
  Test('type(property("os.MonitorCount"))', 'integer');
  Test('type(property("os.fonts"))', 'array');
  Test('∀(property("os.fonts"), x ↦ type(x) = "string")', True);
  Test('type(property("os.clipboard"))', 'string');
  Test('type(property("os.NargleCount"))', failure, 'There is no property named "NargleCount" in store "os".');

  Test('type(property("os.colors.WINDOW"))', 'color');
  Test('type(property("os.colors.WINDOWTEXT"))', 'color');
  Test('type(property("os.colors.BTNFACE"))', 'color');
  Test('type(property("os.colors.DESKTOP"))', 'color');
  Test('type(property("os.colors.HIGHLIGHT"))', 'color');
  Test('type(property("os.colors.HIGHLIGHTTEXT"))', 'color');
  Test('type(property("os.colors.NARGLEFACE"))', failure, 'There is no property named "NARGLEFACE" in store "colors".');

  Test('type(property("os.metrics.SM_CYICON"))', 'integer');
  Test('type(property("os.metrics.SM_MEDIACENTER"))', 'integer');
  Test('type(property("os.metrics.SM_REMOTESESSION"))', 'integer');
  Test('type(property("os.metrics.SM_MOUSEPRESENT"))', 'integer');
  Test('type(property("os.metrics.SM_SWAPBUTTON"))', 'integer');
  Test('type(property("os.metrics.SM_NARGLEWEIGHT"))', failure, 'There is no property named "SM_NARGLEWEIGHT" in store "metrics".');

  Test('type(property("os.keys.VK_F1"))', 'integer');
  Test('type(property("os.keys.VK_RETURN"))', 'integer');
  Test('type(property("os.keys.VK_BACK"))', 'integer');
  Test('type(property("os.keys.VK_HOME"))', 'integer');
  Test('type(property("os.keys.VK_DENARGLE"))', failure, 'There is no property named "VK_DENARGLE" in store "keys".');

  with TRegistry.Create do
    try
      RootKey := HKEY_CURRENT_USER;
      if OpenKey('\Software\Rejbrand\Algosim\SelfTest', True) then
      begin
        try
          _i := Random(10000000);
          WriteInteger('TestValue', _i);
          try
            Test('property("os.registry.HKEY_CURRENT_USER\Software\Rejbrand\Algosim\SelfTest\TestValue")', _i);
          finally
            DeleteValue('TestValue');
          end;
          WriteString('TestValue', 'nargles');
          try
            Test('property("os.registry.HKEY_CURRENT_USER\Software\Rejbrand\Algosim\SelfTest\TestValue")', 'nargles');
          finally
            DeleteValue('TestValue');
          end;
        finally
          DeleteKey('\Software\Rejbrand\Algosim\SelfTest');
        end;
      end;
    finally
      Free;
    end;

  if SetEnvironmentVariable('AlgosimSelfTest', 'nargle') then
    try
      Test('property("os.environment.AlgosimSelfTest")', 'nargle');
    finally
      SetEnvironmentVariable('AlgosimSelfTest', nil)
    end;





  //
  //
  //  CHAPTER 12
  //  Control flow.
  //
  //

  Chapter('Control flow');

  //
  // do
  //

  Test('do(123, true, 5.5, IdentityMatrix(10), color("red"), "page 394")', 'page 394');
  Test('do(123, "page 394")', 'page 394');
  Test('do("page 394")', 'page 394');
  Test('do()', null);


  //
  // if
  //

  Test('if(true, "cat")', 'cat');
  Test('if(false, "cat")', null);
  Test('if(true, "cat", "dog")', 'cat');
  Test('if(false, "cat", "dog")', 'dog');

  Test('if(true, "good", 1/0)', 'good');
  Test('if(false, 1/0, "good again")', 'good again');

  Test('if(true, 1/0, "dummy")', failure, 'Division by zero');
  Test('if(false, "dummy", 1/0)', failure, 'Division by zero');

  Test('if(1/0 > 5, "Interesting.", "You don''t say?")', failure, 'Division by zero');

  Test('f ≔ x ↦ if(x ≠ 0, 1/x, 0);', null);
  Test('f(5)', rat(1, 5));
  Test('f(-2)', rat(-1, 2));
  Test('f(0)', 0);

  Test('f ≔ z ↦ if(IsNumber(z), if(z ≈ 0, "zero", if(IsReal(z), "real", if(Re(z) ≈ 0, "purely imaginary", "mixed"))), "NaN");', null);
  Test('f(0)', 'zero');
  Test('f(4)', 'real');
  Test('f(−2)', 'real');
  Test('f(4⋅i)', 'purely imaginary');
  Test('f(−i)', 'purely imaginary');
  Test('f(i)', 'purely imaginary');
  Test('f(1 + i)', 'mixed');
  Test('f(5 − 3⋅i)', 'mixed');
  Test('f("rat")', 'NaN');

  Test('N ≔ 5; N ≔ N^2; if(N ≠ 0, (N ≔ 1/N)); N', rat(1, 25));
  Test('N ≔ 7; N ≔ N^2; if(N ≠ 0, (N ≔ 1/N)); N', rat(1, 49));
  Test('N ≔ 10; N ≔ N^2; if(N ≠ 0, (N ≔ 1/N)); N', rat(1, 100));
  Test('N ≔ 0; N ≔ N^2; if(N ≠ 0, (N ≔ 1/N)); N', 0);


  //
  // while
  //

  Test('n ≔ 0; while(n < 100, inc(n)); n', 100);
  Test('n ≔ 0; while(n! > 0, dec(n)); n', failure, 'A non-negative integer was expected as argument 1, but "-1" was given.');
  Test('ErrorInfo(LastError()).stack', strarr(['factorial', 'GreaterThan', 'while', 'do', 'LastError']));
  Test('while(5 > "cat", inc(n))', failure, 'Invalid function arguments.');
  Test('n ≔ 100; while(IsNumber(n), (n!; dec(n))); n', failure, 'A non-negative integer was expected as argument 1, but "-1" was given.');
  Test('ErrorInfo(LastError()).stack', strarr(['factorial', 'do', 'while', 'do', 'LastError']));

  Test('n ≔ 10; while(¬IsPrime(n), (n ≔ RandomInt(1000))); IsPrime(n)', True);

  Test('N ≔ 10; n, k ≔ 1; while(k ≤ N, (n ≔ n⋅k; inc(k))); n', 3628800);
  Test('N ≔ 5; n, k ≔ 1; while(k ≤ N, (n ≔ n⋅k; inc(k))); n', 120);
  Test('N ≔ 0; n, k ≔ 1; while(k ≤ N, (n ≔ n⋅k; inc(k))); n', 1);

  Test('N ≔ 50; k ≔ 2; f, fp ≔ 1; fpp ≔ 0; while(k ≤ N, (f ≔ fp + fpp; (fp, fpp) ≔ (f, fp); inc(k))); f', 12586269025);

  Test('fib ≔ N ↦ (k ≔ 2; f, fp ≔ 1; fpp ≔ 0; while(k ≤ N, (f ≔ fp + fpp; (fp, fpp) ≔ (f, fp); inc(k))); f);', null);
  Test('fib(10)', 55);
  Test('fib(20)', 6765);
  Test('fib(30)', 832040);
  Test('fib(40)', 102334155);
  Test('fib(50)', 12586269025);
  Test('fib(60)', 1548008755920);
  Test('fib(70)', 190392490709135);
  Test('fib(80)', 23416728348467685);
  Test('fib(90)', 2880067194370816120);

  Test('f ≔ n ↦ mod(71327⋅n + 32117, 95063);', null);
  Test('n ≔ 0; while(f(n) ≠ 10000, inc(n)); n', 12837);

  Test('s ≔ ""; y ≔ 0; while(y ≤ 9, (x ≔ 0; while(x ≤ 9, (VarExtendWith(s, string(y) ~ string(x)); inc(x))); VarExtendWith(s, ¶); inc(y))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '50515253545556575859'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('s ≔ 0; n ≔ 1; while(n ≤ 1000, (inc(s, n); inc(n))); s', 500500);
  Test('s ≔ 0; n ≔ 1; while(true, (inc(s, n); inc(n); if(n > 1000, break()))); s', 500500);
  Test('s ≔ 0; n ≔ 1; while(n ≤ 1000, (if(n = 500, (inc(n); continue())); inc(s, n); inc(n))); s', 500000);

  Test('s ≔ ""; y ≔ 0; while(y ≤ 9, (x ≔ 0; while(true, (VarExtendWith(s, string(y) ~ string(x)); inc(x); if(x = 10, break()))); VarExtendWith(s, ¶); inc(y))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '50515253545556575859'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('s ≔ ""; y ≔ 0; while(true, (x ≔ 0; while(x ≤ 9, (VarExtendWith(s, string(y) ~ string(x)); inc(x))); VarExtendWith(s, ¶); inc(y); if(y = 10, break()))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '50515253545556575859'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('s ≔ ""; y ≔ 0; while(y ≤ 9, (if(y = 5, (inc(y); continue())); x ≔ 0; while(x ≤ 9, (VarExtendWith(s, string(y) ~ string(x)); inc(x))); VarExtendWith(s, ¶); inc(y))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('s ≔ ""; y ≔ 0; while(y ≤ 9, (x ≔ 0; while(x ≤ 9, (if(x = 3, (inc(x); continue())); VarExtendWith(s, string(y) ~ string(x)); inc(x))); VarExtendWith(s, ¶); inc(y))); s',
    '000102040506070809'#13#10 +
    '101112141516171819'#13#10 +
    '202122242526272829'#13#10 +
    '303132343536373839'#13#10 +
    '404142444546474849'#13#10 +
    '505152545556575859'#13#10 +
    '606162646566676869'#13#10 +
    '707172747576777879'#13#10 +
    '808182848586878889'#13#10 +
    '909192949596979899'#13#10
  );

  Test('s ≔ ""; y ≔ 0; while(y ≤ 9, (if(y = 5, (inc(y); continue())); if(y = 8, break()); x ≔ 0; while(x ≤ 9, (VarExtendWith(s, string(y) ~ string(x)); inc(x))); VarExtendWith(s, ¶); inc(y))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10
  );

  Test('A ≔ VandermondeMatrix(SequenceVector(6));', null);
  Test('c ≔ 0; HasThreeDigit ≔ false; y ≔ 1; while(y ≤ height(A), (x ≔ 1; while(x ≤ width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true)); inc(x))); inc(y))); ''(HasThreeDigit, c)',
    asoarr([True, 36]));
  Test('c ≔ 0; HasThreeDigit ≔ false; y ≔ 1; while(y ≤ height(A), (x ≔ 1; while(x ≤ width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true; break(2))); inc(x))); inc(y))); ''(HasThreeDigit, c)',
    asoarr([True, 18]));

  Test('A ≔ VandermondeMatrix(❨2, 1, 0, 6, 2, 4❩);', null);
  Test('c ≔ 0; HasThreeDigit ≔ false; y ≔ 1; while(y ≤ height(A), (x ≔ 1; while(x ≤ width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true; break(2))); inc(x))); inc(y))); ''(HasThreeDigit, c)',
    asoarr([True, 22]));

  Test('A ≔ VandermondeMatrix(❨2, 1, 2, 2, 3❩);', null);
  Test('c ≔ 0; HasThreeDigit ≔ false; y ≔ 1; while(y ≤ height(A), (x ≔ 1; while(x ≤ width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true; break(2))); inc(x))); inc(y))); ''(HasThreeDigit, c)',
    asoarr([False, 25]));

  Test('A ≔ VandermondeMatrix(❨2, 1, 2, 2, 3, 7, 2❩);', null);
  Test('c ≔ 0; HasThreeDigit ≔ false; y ≔ 1; while(y ≤ height(A), (x ≔ 1; while(x ≤ width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true; break(2))); inc(x))); inc(y))); ''(HasThreeDigit, c)',
    asoarr([True, 34]));

  Test('n, s ≔ 0; while(n ≤ 10, (inc(s, n); inc(n); if(n = 11, continue()))); s', 55);


  //
  // until
  //

  Test('n ≔ 0; until(inc(n), n = 10); n', 10);
  Test('s, n ≔ 0; until((inc(n); inc(s, n)), n = 100); s', 5050);
  Test('until((n ≔ RandomInt(1000)), IsPrime(n)); IsPrime(n)', True);

  Test('until((n ≔ RandomInt(1000 / 0)), IsPrime(n)); IsPrime(n)', failure, 'Division by zero');
  Test('ErrorInfo(LastError()).stack', strarr(['divide', 'RandomInt', 'list node', 'assign', 'until', 'do', 'LastError']));

  Test('until((n ≔ RandomInt(1000)), IsPrime(n/π)); IsPrime(n)', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('ErrorInfo(LastError()).stack', strarr(['IsPrime', 'until', 'do', 'LastError']));

  Test('s, n ≔ 0; until((inc(n); inc(s, n); if(n = 100, break())), n = 1000); s', 5050);
  Test('s, n ≔ 0; until((inc(n); if(n = 50, continue()); inc(s, n)), n = 100); s', 5000);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; until((inc(y); x ≔ 0; until((inc(x); A[y, x] ≔ x⋅y), x = 10)), y = 10); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; until((inc(y); x ≔ 0; until((inc(x); if(x = 5, continue()); A[y, x] ≔ x⋅y), x = 10)), y = 10); A', 10,
    [
      1, 2, 3, 4, 0, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 0, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 0, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 0, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 0, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 0, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 0, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 0, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 0, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 0, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; until((inc(y); x ≔ 0; until((inc(x); if(x = 8, break()); A[y, x] ≔ x⋅y), x = 10)), y = 10); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 0, 0, 0,
      2, 4, 6, 8, 10, 12, 14, 0, 0, 0,
      3, 6, 9, 12, 15, 18, 21, 0, 0, 0,
      4, 8, 12, 16, 20, 24, 28, 0, 0, 0,
      5, 10, 15, 20, 25, 30, 35, 0, 0, 0,
      6, 12, 18, 24, 30, 36, 42, 0, 0, 0,
      7, 14, 21, 28, 35, 42, 49, 0, 0, 0,
      8, 16, 24, 32, 40, 48, 56, 0, 0, 0,
      9, 18, 27, 36, 45, 54, 63, 0, 0, 0,
      10, 20, 30, 40, 50, 60, 70, 0, 0, 0
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; until((inc(y); if(y = 3, continue()); x ≔ 0; until((inc(x); A[y, x] ≔ x⋅y), x = 10)), y = 10); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; until((inc(y); if(y = 9, break()); x ≔ 0; until((inc(x); A[y, x] ≔ x⋅y), x = 10)), y = 10); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; until((inc(y); x ≔ 0; until((inc(x); if(y = 7 ∧ x = 4, continue()); A[y, x] ≔ x⋅y), x = 10)), y = 10); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 0, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; until((inc(y); x ≔ 0; until((inc(x); if(y = 7 ∧ x = 4, break()); A[y, x] ≔ x⋅y), x = 10)), y = 10); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 0, 0, 0, 0, 0, 0, 0,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; until((inc(y); x ≔ 0; until((inc(x); if(y = 7 ∧ x = 4, break(2)); A[y, x] ≔ x⋅y), x = 10)), y = 10); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; until((inc(y); if(y = 3, continue()); if(y = 9, break()); x ≔ 0; until((inc(x); A[y, x] ≔ x⋅y), x = 10)), y = 10); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    ]);

  Test('n, s ≔ 0; until((inc(n); inc(s, n); if(n = 10, continue())), n = 10); s', 55);


  //
  // repeat
  //

  Test('n, s ≔ 0; repeat(inc(n); inc(s, n); if(n = 10, break())); s', 55);
  Test('n, s ≔ 0; repeat(inc(n); inc(s, n); if(n = 100, break())); s', 5050);
  Test('n, s ≔ 0; repeat(inc(n); inc(s, n); if(n = 1000, break())); s', 500500);

  Test('n, s ≔ 0; repeat(inc(n); if(n = 5, continue()); inc(s, n); if(n = 10, break())); s', 50);
  Test('n, s ≔ 0; repeat(inc(n); if(n = 50, continue()); inc(s, n); if(n = 100, break())); s', 5000);
  Test('n, s ≔ 0; repeat(inc(n); if(n = 500, continue()); inc(s, n); if(n = 1000, break())); s', 500000);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; repeat(inc(y); x ≔ 0; repeat(inc(x); A[y, x] ≔ y⋅x; if(x = 10, break())); if(y = 10, break())); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; repeat(inc(y); x ≔ 0; repeat(inc(x); if(x = 6, continue()); A[y, x] ≔ y⋅x; if(x = 10, break())); if(y = 10, break())); A', 10,
    [
      1, 2, 3, 4, 5, 0, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 0, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 0, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 0, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 0, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 0, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 0, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 0, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 0, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 0, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; repeat(inc(y); if(y = 3, continue()); x ≔ 0; repeat(inc(x); A[y, x] ≔ y⋅x; if(x = 10, break())); if(y = 10, break())); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; repeat(inc(y); x ≔ 0; repeat(inc(x); if(x = 7 ∧ y = 8, continue()); A[y, x] ≔ y⋅x; if(x = 10, break())); if(y = 10, break())); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 0, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; repeat(inc(y); x ≔ 0; repeat(inc(x); if(x = 7 ∧ y = 8, break()); A[y, x] ≔ y⋅x; if(x = 10, break())); if(y = 10, break())); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 0, 0, 0, 0,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; repeat(inc(y); x ≔ 0; repeat(inc(x); if(x = 7 ∧ y = 8, break(1)); A[y, x] ≔ y⋅x; if(x = 10, break())); if(y = 10, break())); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 0, 0, 0, 0,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); y ≔ 0; repeat(inc(y); x ≔ 0; repeat(inc(x); if(x = 7 ∧ y = 8, break(2)); A[y, x] ≔ y⋅x; if(x = 10, break())); if(y = 10, break())); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    ]);


  //
  // for
  //

  Test('s ≔ 0; for(i, 1, 1000, inc(s, i)); s', 500500);
  Test('s ≔ 0; for(i, 1, 1000000000000000, (inc(s, i); if(i = 1000, break()))); s', 500500);
  Test('s ≔ 0; for(i, 1, 1000, (if(i = 500, continue()); inc(s, i))); s', 500000);

  Test('N ≔ 10; n ≔ 1; for(k, 1, N, (n ≔ n ⋅ k)); n', 3628800);
  Test('N ≔ 5; n ≔ 1; for(k, 1, N, (n ≔ n ⋅ k)); n', 120);
  Test('N ≔ 0; n ≔ 1; for(k, 1, N, (n ≔ n ⋅ k)); n', 1);

  Test('for(i, 1, 10, 1 + i/0)', failure, 'Division by zero');
  Test('ErrorInfo(LastError()).stack', strarr(['divide', 'add', 'for', 'LastError']));

  Test('for(i, 1, 10/0, 1 + i)', failure, 'Division by zero');
  Test('ErrorInfo(LastError()).stack', strarr(['divide', 'for', 'LastError']));

  Test('for(i, 1/0, 10, 1 + i)', failure, 'Division by zero');
  Test('ErrorInfo(LastError()).stack', strarr(['divide', 'for', 'LastError']));

  Test('for(i/0, 1, 10, 1 + i)', failure, 'A symbol was expected as argument 1.');
  Test('for(i, "a", 10, i)', failure, 'An object of type integer was expected as argument 2, but an object of type string was given.');
  Test('for(i, 1, true, i)', failure, 'An object of type integer was expected as argument 3, but an object of type boolean was given.');

  Test('s ≔ ""; for(y, 0, 9, (for(x, 0, 9, VarExtendWith(s, string(y) ~ string(x))); VarExtendWith(s, ¶))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '50515253545556575859'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('s ≔ ""; for(y, 0, 9, (for(x, 0, 100000, (VarExtendWith(s, string(y) ~ string(x)); if(x = 9, break()))); VarExtendWith(s, ¶))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '50515253545556575859'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('s ≔ ""; for(y, 0, 100000, (for(x, 0, 9, VarExtendWith(s, string(y) ~ string(x))); VarExtendWith(s, ¶); if(y = 9, break()))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '50515253545556575859'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('s ≔ ""; for(y, 0, 9, (if(y = 5, continue()); for(x, 0, 9, VarExtendWith(s, string(y) ~ string(x))); VarExtendWith(s, ¶))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('s ≔ ""; for(y, 0, 9, (for(x, 0, 9, (if(x = 2, continue()); VarExtendWith(s, string(y) ~ string(x)))); VarExtendWith(s, ¶))); s',
    '000103040506070809'#13#10 +
    '101113141516171819'#13#10 +
    '202123242526272829'#13#10 +
    '303133343536373839'#13#10 +
    '404143444546474849'#13#10 +
    '505153545556575859'#13#10 +
    '606163646566676869'#13#10 +
    '707173747576777879'#13#10 +
    '808183848586878889'#13#10 +
    '909193949596979899'#13#10
  );

  Test('A ≔ VandermondeMatrix(SequenceVector(6));', null);
  Test('c ≔ 0; HasThreeDigit ≔ false; for(y, 1, height(A), (for(x, 1, width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true)))))); ''(HasThreeDigit, c)',
    asoarr([True, 36]));
  Test('c ≔ 0; HasThreeDigit ≔ false; for(y, 1, height(A), (for(x, 1, width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true; break(2))))))); ''(HasThreeDigit, c)',
    asoarr([True, 18]));

  Test('A ≔ VandermondeMatrix(❨2, 1, 0, 6, 2, 4❩);', null);
  Test('c ≔ 0; HasThreeDigit ≔ false; for(y, 1, height(A), (for(x, 1, width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true; break(2))))))); ''(HasThreeDigit, c)',
    asoarr([True, 22]));

  Test('A ≔ VandermondeMatrix(❨2, 1, 2, 2, 3❩);', null);
  Test('c ≔ 0; HasThreeDigit ≔ false; for(y, 1, height(A), (for(x, 1, width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true; break(2))))))); ''(HasThreeDigit, c)',
    asoarr([False, 25]));

  Test('A ≔ VandermondeMatrix(❨2, 1, 2, 2, 3, 7, 2❩);', null);
  Test('c ≔ 0; HasThreeDigit ≔ false; for(y, 1, height(A), (for(x, 1, width(A), (inc(c); if(A[y, x] ≥ 100, (HasThreeDigit ≔ true; break(2))))))); ''(HasThreeDigit, c)',
    asoarr([True, 34]));

  Test('A ≔ ZeroMatrix(10); for(i, 1, 10, for(j, 1, 10, (A[i, j] ≔ i⋅j))); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); for(i, 1, 10, for(j, 1, 10, (if(i = 7 ∧ j = 5, continue()); A[i, j] ≔ i⋅j))); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 0, 42, 49, 56, 63, 70,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); for(i, 1, 10, for(j, 1, 10, (if(i = 7 ∧ j = 5, break()); A[i, j] ≔ i⋅j))); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 0, 0, 0, 0, 0, 0,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); for(i, 1, 10, for(j, 1, 10, (if(i = 7 ∧ j = 5, break(1)); A[i, j] ≔ i⋅j))); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 0, 0, 0, 0, 0, 0,
      8, 16, 24, 32, 40, 48, 56, 64, 72, 80,
      9, 18, 27, 36, 45, 54, 63, 72, 81, 90,
      10, 20, 30, 40, 50, 60, 70, 80, 90, 100
    ]);

  Test('A ≔ ZeroMatrix(10); for(i, 1, 10, for(j, 1, 10, (if(i = 7 ∧ j = 5, break(2)); A[i, j] ≔ i⋅j))); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      2, 4, 6, 8, 10, 12, 14, 16, 18, 20,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    ]);

  Test('A ≔ ZeroMatrix(10); for(i, 1, 10, (if(i = 2, continue()); if(i = 8, break()); for(j, 1, 10, (A[i, j] ≔ i⋅j)))); A', 10,
    [
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      3, 6, 9, 12, 15, 18, 21, 24, 27, 30,
      4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
      5, 10, 15, 20, 25, 30, 35, 40, 45, 50,
      6, 12, 18, 24, 30, 36, 42, 48, 54, 60,
      7, 14, 21, 28, 35, 42, 49, 56, 63, 70,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    ]);

  Test('A ≔ ZeroMatrix(10); for(y, 1, 10, for(x, 1, 10, (x ≔ 4; A[y, x] ≔ y⋅x))); A', failure,
   'Left side cannot be assigned to.');

  Test('A ≔ ZeroMatrix(10); for(y, 1, 10, for(x, 1, 10, (dec(x); A[y, x] ≔ y⋅x))); A', failure,
   'Left side cannot be assigned to.');

  Test('A ≔ ZeroMatrix(10); for(y, 1, 10, for(x, 1, 10, (y ≔ 4; A[y, x] ≔ y⋅x))); A', failure,
   'Left side cannot be assigned to.');

  Test('A ≔ ZeroMatrix(10); for(y, 1, 10, for(x, 1, 10, (dec(y); A[y, x] ≔ y⋅x))); A', failure,
   'Left side cannot be assigned to.');


  //
  // ForEach
  //

  Test('L ≔ SequenceList(100); v ≔ SequenceVector(100); cv ≔ ComplexVector(v); S ≔ ToSet(L);', null);
  Test('A ≔ matrix(10, ToList(v)); CA ≔ ComplexMatrix(A);', null);

  Test('s ≔ 0; ForEach(L, x, inc(s, x)); s', 5050);
  Test('s ≔ 0; ForEach(v, x, inc(s, x)); s', 5050);
  Test('s ≔ 0; ForEach(cv, x, inc(s, x)); s', 5050);
  Test('s ≔ 0; ForEach(S, x, inc(s, x)); s', 5050);
  Test('s ≔ 0; ForEach(A, x, inc(s, x)); s', 5050);
  Test('s ≔ 0; ForEach(CA, x, inc(s, x)); s', 5050);

  Test('s ≔ 0; ForEach(L, x, (if(x = 50, continue()); inc(s, x))); s', 5000);
  Test('s ≔ 0; ForEach(v, x, (if(x = 50, continue()); inc(s, x))); s', 5000);
  Test('s ≔ 0; ForEach(cv, x, (if(x = 50, continue()); inc(s, x))); s', 5000);
  Test('s ≔ 0; ForEach(S, x, (if(x = 50, continue()); inc(s, x))); s', 5000);
  Test('s ≔ 0; ForEach(A, x, (if(x = 50, continue()); inc(s, x))); s', 5000);
  Test('s ≔ 0; ForEach(CA, x, (if(x = 50, continue()); inc(s, x))); s', 5000);

  Test('s ≔ 0; ForEach(L, x, (if(x = 50, break()); inc(s, x))); s', 1225);
  Test('s ≔ 0; ForEach(v, x, (if(x = 50, break()); inc(s, x))); s', 1225);
  Test('s ≔ 0; ForEach(cv, x, (if(x = 50, break()); inc(s, x))); s', 1225);
  Test('s ≔ 0; ForEach(A, x, (if(x = 50, break()); inc(s, x))); s', 1225);
  Test('s ≔ 0; ForEach(CA, x, (if(x = 50, break()); inc(s, x))); s', 1225);

  Test('S ≔ S ∪ {50.5};', null);
  Test('ForEach(S, x, odd(x))', failure, 'An object of type integer was expected as argument 1, but an object of type real number was given.');
  Test('ErrorInfo(LastError()).stack', strarr(['odd', 'ForEach', 'LastError']));

  Test('s ≔ 0; ForEach(L/0, x, inc(s, x)); s', failure, 'Invalid function arguments.');
  Test('s ≔ 0; ForEach(L, x/0, inc(s, x)); s', failure, 'A symbol was expected as argument 2.');

  Test('L ≔ SequenceList(0, 9); s ≔ ""; ForEach(L, y, (ForEach(L, x, VarExtendWith(s, string(y) ~ string(x))); VarExtendWith(s, ¶))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '50515253545556575859'#13#10 +
    '60616263646566676869'#13#10 +
    '70717273747576777879'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('L ≔ SequenceList(0, 9); s ≔ ""; ForEach(L, y, (if(y = 7, continue()); ForEach(L, x, VarExtendWith(s, string(y) ~ string(x))); VarExtendWith(s, ¶))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '50515253545556575859'#13#10 +
    '60616263646566676869'#13#10 +
    '80818283848586878889'#13#10 +
    '90919293949596979899'#13#10
  );

  Test('L ≔ SequenceList(0, 9); s ≔ ""; ForEach(L, y, (if(y = 7, break()); ForEach(L, x, VarExtendWith(s, string(y) ~ string(x))); VarExtendWith(s, ¶))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '40414243444546474849'#13#10 +
    '50515253545556575859'#13#10 +
    '60616263646566676869'#13#10
  );

  Test('L ≔ SequenceList(0, 9); s ≔ ""; ForEach(L, y, (ForEach(L, x, (if(x = 6, continue()); VarExtendWith(s, string(y) ~ string(x)))); VarExtendWith(s, ¶))); s',
    '000102030405070809'#13#10 +
    '101112131415171819'#13#10 +
    '202122232425272829'#13#10 +
    '303132333435373839'#13#10 +
    '404142434445474849'#13#10 +
    '505152535455575859'#13#10 +
    '606162636465676869'#13#10 +
    '707172737475777879'#13#10 +
    '808182838485878889'#13#10 +
    '909192939495979899'#13#10
  );

  Test('L ≔ SequenceList(0, 9); s ≔ ""; ForEach(L, y, (ForEach(L, x, (if(x = 6, break()); VarExtendWith(s, string(y) ~ string(x)))); VarExtendWith(s, ¶))); s',
    '000102030405'#13#10 +
    '101112131415'#13#10 +
    '202122232425'#13#10 +
    '303132333435'#13#10 +
    '404142434445'#13#10 +
    '505152535455'#13#10 +
    '606162636465'#13#10 +
    '707172737475'#13#10 +
    '808182838485'#13#10 +
    '909192939495'#13#10
  );

  Test('L ≔ SequenceList(0, 9); s ≔ ""; ForEach(L, y, (ForEach(L, x, (if(y = 4 ∧ x = 6, break(2)); VarExtendWith(s, string(y) ~ string(x)))); VarExtendWith(s, ¶))); s',
    '00010203040506070809'#13#10 +
    '10111213141516171819'#13#10 +
    '20212223242526272829'#13#10 +
    '30313233343536373839'#13#10 +
    '404142434445'
  );

  Test('s ≔ "It does not do to dwell on dreams and forget to live.";', null);
  Test('counts ≔ ''(0, 0, 0); ForEach(s, c, (if(ChrIsLetter(c), inc(counts[1])); if(ChrIsWhitespace(c), inc(counts[2])); if(ChrIsPunctuation(c), inc(counts[3])))); counts', intarr([41, 11, 1]));





  //
  //
  //  CHAPTER 13
  //  Sounds
  //
  //

  Chapter('Sounds');

  Test('a ≔ SineTone(400, 0.1, 5); b ≔ SineTone(800, 0.1, 5);', null);
  Test('type(a)', 'sound');
  Test('type(b)', 'sound');
  Test('AsSingleLine(a)', 'A 5.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(b)', 'A 5.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(a)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(5.0), ASOInt(48000), ASOInt(240000), ASOInt(960000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );
  Test('AudioMetadata(b)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(5.0), ASOInt(48000), ASOInt(240000), ASOInt(960000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );
  Test('a = b', False);

  Test('c ≔ a + b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 5.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(5.0), ASOInt(48000), ASOInt(240000), ASOInt(960000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a − b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 5.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(5.0), ASOInt(48000), ASOInt(240000), ASOInt(960000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a ⋅ b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 5.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(5.0), ASOInt(48000), ASOInt(240000), ASOInt(960000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('a ≔ SineTone(200, 0.1, 2); b ≔ SineTone(400, 0.1, 3);', null);
  Test('type(a)', 'sound');
  Test('type(b)', 'sound');
  Test('AsSingleLine(a)', 'A 2.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(b)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(a)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(2.0), ASOInt(48000), ASOInt(96000), ASOInt(384000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );
  Test('AudioMetadata(b)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a + b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a − b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a ⋅ b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ b + a;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ b − a;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ b ⋅ a;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('a ≔ SineTone(200, 0.1, 3); b ≔ DebugObject("empty sound");', null);

  Test('type(b)', 'sound');
  Test('AsSingleLine(b)', 'A 0.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(b)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(0.0), ASOInt(48000), ASOInt(0), ASOInt(0), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a + b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a − b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a ⋅ b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ b + a;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ b − a;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ b ⋅ a;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('a ≔ SineTone(100, 0.1, 2); b ≔ SineTone(400, 0.1, 1);', null);
  Test('c ≔ AppendSound(a, b);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('a ≔ SineTone(100, 0.1, 2); b ≔ DebugObject("empty sound");', null);
  Test('c ≔ AppendSound(a, b);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 2.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(2.0), ASOInt(48000), ASOInt(96000), ASOInt(384000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );
  Test('a = c', True);

  Test('c ≔ AppendSound(b, a);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 2.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(2.0), ASOInt(48000), ASOInt(96000), ASOInt(384000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );
  Test('a = c', True);

  Test('a ≔ SineTone(100, 0.1, 2); b ≔ 0.5⋅a;', null);
  Test('type(b)', 'sound');
  Test('AsSingleLine(b)', 'A 2.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(b)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(2.0), ASOInt(48000), ASOInt(96000), ASOInt(384000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );
  Test('a = b', False);

  Test('a ≔ SineTone(100, 0.1, 2); b ≔ 1⋅a;', null);
  Test('type(b)', 'sound');
  Test('AsSingleLine(b)', 'A 2.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(b)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(2.0), ASOInt(48000), ASOInt(96000), ASOInt(384000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );
  Test('a = b', True);

  Test('a ≔ SineTone(100, 0.1, 5); b ≔ SineTone(500, 0.1, 5);', null);
  Test('c ≔ a + b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 5.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(5.0), ASOInt(48000), ASOInt(240000), ASOInt(960000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a − b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 5.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(5.0), ASOInt(48000), ASOInt(240000), ASOInt(960000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ a ⋅ b;', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 5.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(5.0), ASOInt(48000), ASOInt(240000), ASOInt(960000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('a + b = superpose(a, b, 0, 1, 1)', True);
  Test('a − b = superpose(a, b, 0, 1, −1)', True);
  Test('a ⋅ b = superpose(a, b, 0, 1/2, 1/2)', True);

  Test('a + b = superpose(a, b, 0, 1, 0.8)', False);
  Test('a − b = superpose(a, b, 0, 1, 1)', False);
  Test('a ⋅ b = superpose(a, b, 0, 1/2, 1)', False);

  Test('c ≔ AppendSound(a, b);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 10.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(10.0), ASOInt(48000), ASOInt(480000), ASOInt(1920000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ superpose(a, b, 2.5, 1, 1);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 7.50-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(7.5), ASOInt(48000), ASOInt(360000), ASOInt(1440000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ superpose(a, b, 10, 1, 1);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 15.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(15), ASOInt(48000), ASOInt(720000), ASOInt(2880000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('AppendSound(a, b) = superpose(a, b, 5, 1, 1)', True);

  Test('c ≔ FadeSounds(a, b);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 9.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(9), ASOInt(48000), ASOInt(432000), ASOInt(1728000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('FadeSounds(a, b) = FadeSounds(a, b, 1)', True);

  Test('c ≔ FadeSounds(a, b, 2);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 8.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(8), ASOInt(48000), ASOInt(384000), ASOInt(1536000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ FadeSounds(a, b, 5);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 5.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(5), ASOInt(48000), ASOInt(240000), ASOInt(960000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ FadeSounds(a, b, 10);', failure, 'The duration of the overlap exceeds the duration of the first sound.');

  Test('a ≔ SineTone(100, 0.1, 5); b ≔ SineTone(500, 0.1, 2);', null);
  Test('FadeSounds(a, b);', null);
  Test('FadeSounds(a, b, 2);', null);
  Test('FadeSounds(a, b, 2.1);', failure, 'The duration of the overlap exceeds the duration of the second sound.');

  Test('FadeSounds(a, b, 0)', failure, 'A positive real number was expected as argument 3, but "0" was given.');

  Test('a ≔ SineTone(200, 0.1, 3); b ≔ DebugObject("empty sound");', null);

  Test('c ≔ reverse(a);', null);
  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 3.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.0), ASOInt(48000), ASOInt(144000), ASOInt(576000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ reverse(b);', null);
  Test('type(b)', 'sound');
  Test('AsSingleLine(b)', 'A 0.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(b)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(0.0), ASOInt(48000), ASOInt(0), ASOInt(0), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('c ≔ reverse(a, b);', failure, 'Too many arguments.');

  Test('a ≔ SineTone(100, 0.1, 5); b ≔ SineTone(100, 0.2, 5);', null);
  Eps(1E-8); Test('SoundMax(a)', 0.1);
  Eps(1E-8); Test('SoundMax(b)', 0.2);
  Eps(1E-8); Test('SoundMax(0.4 ⋅ a)', 0.04);
  Eps(1E-8); Test('SoundMax(a + b)', 0.3);
  Eps(1E-8); Test('SoundMax(a ⋅ b)', 0.15);
  Eps(1E-8); Test('SoundMax(a − b)', 0.1);
  Eps(1E-8); Test('SoundMax(a + a)', 0.2);
  Eps(1E-8); Test('SoundMax(a ⋅ a)', 0.1);
  Eps(1E-8); Test('SoundMax(a − a)', 0.0);

  Test('a ⋅ a = a', True);
  Test('0.5 ⋅ (a + a) = a', True);

  Test('echo(a);', null);
  Test('echo(a, 0.1);', null);
  Test('echo(a, 0.1, 0.4);', null);
  Test('echo(a, 0.5, 0.2);', null);
  Test('echo(a, 0.0, 0.0) = a', True);
  Test('echo(a, −0.1, 0.4);', failure, 'The delay cannot be negative.');
  Test('echo(a, 0.5, 1.2);', failure, 'The intensity of the echo must lie within [0, 1].');

  Test('reverse(a) = a', False);
  Test('reverse(reverse(a)) = a', True);

  Test('a ≔ SineTone(400, 0.1, 1);', null);
  Test('b ≔ ConvertAudio(a, 16);', null);
  Test('c ≔ ConvertAudio(a, 8);', null);
  Test('d ≔ ConvertAudio(a, 32);', null);

  Test('type(a)', 'sound');
  Test('AsSingleLine(a)', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(a)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(1.0), ASOInt(48000), ASOInt(48000), ASOInt(192000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('type(b)', 'sound');
  Test('AsSingleLine(b)', 'A 1.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(b)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(1.0), ASOInt(48000), ASOInt(48000), ASOInt(96000), ASOInt(16), ASOInt(1), ASOInt(4)]
    )
  );

  Test('type(c)', 'sound');
  Test('AsSingleLine(c)', 'A 1.00-second 8-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(c)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(1.0), ASOInt(48000), ASOInt(48000), ASOInt(48000), ASOInt(8), ASOInt(1), ASOInt(4)]
    )
  );

  Test('type(d)', 'sound');
  Test('AsSingleLine(d)', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(d)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(1.0), ASOInt(48000), ASOInt(48000), ASOInt(192000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('a = b', False);
  Test('a = c', False);
  Test('a = d', True);

  Test('a ≔ SineTone(100, 0.1, 2); b ≔ SineTone(500, 0.1, 2);', null);
  Test('C ≔ MultichannelAudio(a, b);', null);

  Test('type(C)', 'sound');
  Test('AsSingleLine(C)', 'A 2.00-second 32-bit 48000 Hz 2-channel sound.');
  Test('AudioMetadata(C)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(2.0), ASOInt(48000), ASOInt(96000), ASOInt(768000), ASOInt(32), ASOInt(2), ASOInt(3)]
    )
  );

  Test('D ≔ MultichannelAudio(b, a);', null);

  Test('type(D)', 'sound');
  Test('AsSingleLine(D)', 'A 2.00-second 32-bit 48000 Hz 2-channel sound.');
  Test('AudioMetadata(D)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(2.0), ASOInt(48000), ASOInt(96000), ASOInt(768000), ASOInt(32), ASOInt(2), ASOInt(3)]
    )
  );

  Test('E ≔ AppendSound(C, D);', null);

  Test('type(E)', 'sound');
  Test('AsSingleLine(E)', 'A 4.00-second 32-bit 48000 Hz 2-channel sound.');
  Test('AudioMetadata(E)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(4.0), ASOInt(48000), ASOInt(192000), ASOInt(1536000), ASOInt(32), ASOInt(2), ASOInt(3)]
    )
  );

  Test('a ≔ SineTone(100, 0.1, 2, 16, 22050);', null);

  Test('type(a)', 'sound');
  Test('AsSingleLine(a)', 'A 2.00-second 16-bit 22050 Hz 1-channel sound.');
  Test('AudioMetadata(a)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(2.0), ASOInt(22050), ASOInt(44100), ASOInt(88200), ASOInt(16), ASOInt(1), ASOInt(4)]
    )
  );

  Test('a ≔ WhiteNoise(0.5, 1);', null);

  Test('type(a)', 'sound');
  Test('AsSingleLine(a)', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AudioMetadata(a)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(1.0), ASOInt(48000), ASOInt(48000), ASOInt(192000), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('a ≔ WhiteNoise(0.7, 3.5, 800);', null);

  Test('type(a)', 'sound');
  Test('AsSingleLine(a)', 'A 3.50-second 32-bit 800 Hz 1-channel sound.');
  Test('AudioMetadata(a)',
    TAlgosimStructure.CreateWithValue(
      ['duration', 'SampleFreq', 'SampleLen', 'ByteLen', 'bitdepth', 'channels', 'ChannelMask'],
      [ASO(3.5), ASOInt(800), ASOInt(2800), ASOInt(11200), ASOInt(32), ASOInt(1), ASOInt(4)]
    )
  );

  Test('SineTone(100, 0.1, 1) = SineTone(100, 0.1, 1, 32)', True);
  Test('SineTone(100, 0.1, 1) = SineTone(100, 0.1, 1, 32, 48000)', True);

  Test('a ≔ SineTone(100, 0.1, 1); b ≔ SineTone(500, 0.1, 1);', null);
  Test('AsSingleLine(a + b)', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(a − b)', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(a ⋅ b)', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(AppendSound(a, b))', 'A 2.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(reverse(a))', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(echo(a))', 'A 1.40-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(MultichannelAudio(a, b))', 'A 1.00-second 32-bit 48000 Hz 2-channel sound.');
  Test('AsSingleLine(FadeSounds(a, b))', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 32))', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 16))', 'A 1.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 8))', 'A 1.00-second 8-bit 48000 Hz 1-channel sound.');
  Eps(1E-8); Test('SoundMax(a)', 0.1);

  Test('a ≔ SineTone(100, 0.1, 1, 16); b ≔ SineTone(500, 0.1, 1, 16);', null);
  Test('AsSingleLine(a + b)', 'A 1.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(a − b)', 'A 1.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(a ⋅ b)', 'A 1.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(AppendSound(a, b))', 'A 2.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(reverse(a))', 'A 1.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(echo(a))', 'A 1.40-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(MultichannelAudio(a, b))', 'A 1.00-second 16-bit 48000 Hz 2-channel sound.');
  Test('AsSingleLine(FadeSounds(a, b))', 'A 1.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 32))', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 16))', 'A 1.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 8))', 'A 1.00-second 8-bit 48000 Hz 1-channel sound.');
  Eps(1E-4); Test('SoundMax(a)', 0.1);

  Test('a ≔ SineTone(100, 0.1, 1, 8); b ≔ SineTone(500, 0.1, 1, 8);', null);
  Test('AsSingleLine(a + b)', 'A 1.00-second 8-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(a − b)', 'A 1.00-second 8-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(a ⋅ b)', 'A 1.00-second 8-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(AppendSound(a, b))', 'A 2.00-second 8-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(reverse(a))', 'A 1.00-second 8-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(MultichannelAudio(a, b))', 'A 1.00-second 8-bit 48000 Hz 2-channel sound.');
  Test('AsSingleLine(FadeSounds(a, b))', 'A 1.00-second 8-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 32))', 'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 16))', 'A 1.00-second 16-bit 48000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 8))', 'A 1.00-second 8-bit 48000 Hz 1-channel sound.');
  Eps(1E-2); Test('SoundMax(a)', 0.1);

  Test('a ≔ SineTone(100, 0.1, 1, 32, 24000); b ≔ SineTone(500, 0.1, 1, 32, 24000);', null);
  Test('AsSingleLine(a + b)', 'A 1.00-second 32-bit 24000 Hz 1-channel sound.');
  Test('AsSingleLine(a − b)', 'A 1.00-second 32-bit 24000 Hz 1-channel sound.');
  Test('AsSingleLine(a ⋅ b)', 'A 1.00-second 32-bit 24000 Hz 1-channel sound.');
  Test('AsSingleLine(AppendSound(a, b))', 'A 2.00-second 32-bit 24000 Hz 1-channel sound.');
  Test('AsSingleLine(reverse(a))', 'A 1.00-second 32-bit 24000 Hz 1-channel sound.');
  Test('AsSingleLine(echo(a))', 'A 1.40-second 32-bit 24000 Hz 1-channel sound.');
  Test('AsSingleLine(MultichannelAudio(a, b))', 'A 1.00-second 32-bit 24000 Hz 2-channel sound.');
  Test('AsSingleLine(FadeSounds(a, b))', 'A 1.00-second 32-bit 24000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 32))', 'A 1.00-second 32-bit 24000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 16))', 'A 1.00-second 16-bit 24000 Hz 1-channel sound.');
  Test('AsSingleLine(ConvertAudio(a, 8))', 'A 1.00-second 8-bit 24000 Hz 1-channel sound.');
  Eps(1E-8); Test('SoundMax(a)', 0.1);

  Test('a ≔ SineTone(100, 0.1, 1); b ≔ SineTone(500, 0.1, 1); C ≔ MultichannelAudio(a, b);', null);
  Test('a = ExtractChannel(C, 0) ∧ b = ExtractChannel(C, 1)', True);

  Test('ν ≔ 400; ω ≔ 2⋅π⋅ν; AsSingleLine(ComputeSound(sin(ω⋅t), t, 0, 1))',
    'A 1.00-second 32-bit 48000 Hz 1-channel sound.');
  Test('ν ≔ 400; ω ≔ 2⋅π⋅ν; ComputeSound(sin(ω⋅t, 4), t, 0, 1)', failure,
    'Too many arguments.');
  Test('ν ≔ 400; ω ≔ 2⋅π⋅ν; ComputeSound("piano", t, 0, 1)', failure,
    'Wave displacement must be a real number.');

  Test('delete(a); delete(b); delete(c); delete(d); delete(C); delete(D); delete(E)', success);





  //
  //
  //  CHAPTER 14
  //  Import and export
  //
  //

  Chapter('Import and export');

  //
  // Saving
  //

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try

    Test('SaveToFile(0, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.1', _s = '0');

    Test('SaveToFile(1024, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.2', _s = '1024');

    Test('SaveToFile(−1024, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.3', _s = '-1024');

    Test('SaveToFile(1/2-1/2, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.4', _s = '0');

    Test('SaveToFile(1/4, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.5', _s = '1/4');

    Test('SaveToFile(−1/4, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.6', _s = '-1/4');

    Test('SaveToFile(0.0, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.7', _s = '0');

    Test('SaveToFile(5.8, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.8', _s = '5.8');

    Test('SaveToFile(−5.8, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.9', _s = '-5.8');

    Test('SaveToFile(2 + 3⋅i, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.10', _s = '2 + 3⋅i');

    Test('SaveToFile(−2 − 3⋅i, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.11', _s = '-2 - 3⋅i');

    Test('SaveToFile(2 + 3⋅i − 3⋅i, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.12', _s = '2');

    Test('SaveToFile(−2 + 3⋅i − 3⋅i, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.13', _s = '-2');

    Test('SaveToFile(3⋅i, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.14', _s = '3⋅i');

    Test('SaveToFile(−3⋅i, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.15', _s = '-3⋅i');

    Test('SaveToFile(i, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.16', _s = 'i');

    Test('SaveToFile(−i, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.17', _s = '-i');

    Test('SaveToFile(3⋅i − 3⋅i, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.18', _s = '0');

    Test('SaveToFile(true, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.19', _s = 'true');

    Test('SaveToFile(false, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.20', _s = 'false');

    Test('SaveToFile("I don''t believe it was a particularly long conversation.", FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.21', _s = 'I don''t believe it was a particularly long conversation.');

    Test('SaveToFile("alpha"+¶+"beta"+¶+"gamma", FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.22', _s = 'alpha'#13#10'beta'#13#10'gamma');

    Test('SaveToFile("", FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.23', _s = '');

    Test('SaveToFile("What ""thing""?", FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.24', _s = 'What "thing"?');

    Test('SaveToFile(❨5, −2, 7.5, 0❩, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.25', _s = '5'#13#10'-2'#13#10'7.5'#13#10'0');

    Test('SaveToFile(❨5, −2, 7.5 + i, 0❩, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.26', _s = '5'#13#10'-2'#13#10'7.5 + i'#13#10'0');

    Test('SaveToFile(❨5, −2, 7.5 − i, 0❩, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.27', _s = '5'#13#10'-2'#13#10'7.5 - i'#13#10'0');

    Test('SaveToFile(❨❨2, 7, 5, 0❩, ❨6.3, −5, −1.2, 4❩❩, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.28', _s = '2'#9'7'#9'5'#9'0'#13#10'6.3'#9'-5'#9'-1.2'#9'4');

    Test('SaveToFile(❨❨2, 7, 5, −3⋅i❩, ❨6.3, −5 − 3⋅i, −1.2, i❩❩, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.29', _s = '2'#9'7'#9'5'#9'-3⋅i'#13#10'6.3'#9'-5 - 3⋅i'#9'-1.2'#9'i');

    Test('SaveToFile(''(123, −5, −6.7, 2 − 3⋅i, "teaspoon", true), FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.30', _s = '123'#13#10'-5'#13#10'-6.7'#13#10'2 - 3⋅i'#13#10'"teaspoon"'#13#10'true');

    Test('SaveToFile(''(123, −5, −6.7, 2 − 3⋅i, "what ""teaspoon""?", true), FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.31', _s = '123'#13#10'-5'#13#10'-6.7'#13#10'2 - 3⋅i'#13#10'"what ""teaspoon""?"'#13#10'true');

    Test('SaveToFile(''("nargle", "hippogriff", "boggart"), FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.32', _s = '"nargle"'#13#10'"hippogriff"'#13#10'"boggart"');

    Test('SaveToFile(''("alfa", "beta", ''(1, 2, 3, ''(1, 3, "gamma"), 5)), FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.33', _s = '"alfa"'#13#10'"beta"'#13#10'''(1, 2, 3, ''(1, 3, "gamma"), 5)');

    Test('SaveToFile(struct("name": "Albus Dumbledore", "role": "headmaster", "age": 115), FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.34', _s = 'name: "Albus Dumbledore"'#13#10'role: "headmaster"'#13#10'age: 115');

    Test('SaveToFile({-123}, FileName)', success);
    _s := TFile.ReadAllText(fn, TEncoding.UTF8);
    TestExternal('SaveToFile.35', _s = '{-123}');

  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.bmp';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('SaveToFile(ColorGradient(100, 100, "black", "yellow"), FileName)', success);
    _bin := TFile.ReadAllBytes(fn);
    TestExternal('saved bitmap: check size and magic word',
      (Length(_bin) > 40000) and (_bin[0] = Ord('B')) and (_bin[1] = Ord('M')));
    _bm := TBitmap.Create;
    try
      _bm.LoadFromFile(fn);
      TestExternal('saved bitmap: pixel-perfect match',
        TASPixmap.Create(_bm) = TASPixmap.CreateGradient(100, 100, clBlack, clYellow));
    finally
      _bm.Free;
    end;
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.png';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('SaveToFile(ColorGradient(100, 100, "black", "yellow"), FileName)', success);
    _bin := TFile.ReadAllBytes(fn);
    TestExternal('saved png: check size and magic word',
      (Length(_bin) > 500) and (Length(_bin) < 1000) and (PCardinal(_bin)^ = $474E5089));
    _png := TPngImage.Create;
    try
      _png.LoadFromFile(fn);
      _bm := TBitmap.Create;
      try
        _bm.Assign(_png);
        _bm.PixelFormat := pf32bit;
        TestExternal('saved png: pixel-perfect match',
          TASPixmap.Create(_bm) = TASPixmap.CreateGradient(100, 100, clBlack, clYellow));
      finally
        _bm.Free;
      end;
    finally
      _png.Free;
    end;
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.jpg';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('SaveToFile(ColorGradient(100, 100, "black", "yellow"), FileName)', success);
    _bin := TFile.ReadAllBytes(fn);
    TestExternal('saved jpg: check size and magic word',
      (Length(_bin) > 1000) and (Length(_bin) < 2000) and (PCardinal(PByte(_bin) + 6)^ = $4649464A));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.wav';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('SaveToFile(SineTone(400, 0.1, 5), FileName)', success);
    _bin := TFile.ReadAllBytes(fn);
    TestExternal('saved wav: check size and magic word',
      (Length(_bin) > 960000) and (Length(_bin) < 961000) and
      (PCardinal(_bin)^ = $46464952) and (PCardinal(PByte(_bin) + 8)^ = $45564157));
    TestExternal('saved wav: sample-perfect match',
      LoadSoundFromFile(fn) = SineTone(400, 0.1, 5));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.dat';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('SaveToFile(BinaryData("wand"), FileName)', success);
    _bin := TFile.ReadAllBytes(fn);
    TestExternal('saved binary data',
      (Length(_bin) = 8) and (PUInt64(_bin)^ = $0064006E00610077));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;


  //
  // Saving + Loading
  //

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 394; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "integer"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ −394; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "integer"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 0; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "integer"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 123456789123456; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "integer"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ −123456789123456; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "integer"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ −394; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "rational number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ −394; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ −394; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 1/2; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "rational number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ −1/2; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "rational number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 1/2 − 1/2; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "rational number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 12345678/12345679; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "rational number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ −12345678/12345679; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "rational number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5.3; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ −5.3; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5.3E85; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5.3E-85; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ -5.3E85; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ -5.3E-85; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 0.0; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5.3; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ −5.3; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5.3E85; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5.3E-85; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ -5.3E85; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ -5.3E-85; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 0.0; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5 + i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5 + 4⋅i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ -5 + 4⋅i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5 - 4⋅i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ -5 - 4⋅i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5.5 + 4.1⋅i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 5 + i - i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 4⋅i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ -4⋅i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ -i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ i - i; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex number"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨5, 3, 2, −1, 0, 2.5❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real vector"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨3, −2, 4, π❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real vector"); a ≈ b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨5, 3, 2, −1, 0, 2.5❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex vector"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨3, −2, 4, π❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex vector"); a ≈ b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨❨1, 5, 2❩, ❨1, 0, −2❩❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ IdentityMatrix(1000); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ IdentityMatrix(1); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨❨1, 5, 2❩, ❨1, 0, −2❩❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ IdentityMatrix(1000); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ IdentityMatrix(1); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨2 + i, 4 + 5⋅i, −2 + i, 4 − 3⋅i, 1.5⋅i, −3⋅i, −7 − 2.5⋅i, 4, −2, 7 − i, 4.009 + i❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex vector"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨2 + i❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex vector"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨❨2 + i, −i, 5 − 2⋅i❩, ❨−3, 0.5, −5 − 7⋅i❩, ❨−2 + i, −3 − i, −2⋅i❩❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ❨❨2.5 + i, −i, 5 − 2.991⋅i❩, ❨−3, 0.155, −5.153E2 − 7E-20⋅i❩, ❨−22.1E33 + i, −3 − i, −2⋅i❩❩; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ i⋅IdentityMatrix(1000); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ DebugObject("empty vector"); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real vector"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ComplexVector(DebugObject("empty vector")); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex vector"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ DebugObject("empty matrix"); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real matrix"); a = b', failure, 'A matrix must have size at least 1×1.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ComplexMatrix(DebugObject("empty matrix")); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', failure, 'A matrix must have size at least 1×1.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ "a teaspoon," ~ chr(13) ~ chr(10) ~ "    the emotional range of"; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "string"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ "No dawdling!"; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "string"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ""; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "string"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ''("alpha", 123, 1 + i); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "array"); ''(string(a)) = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.bmp';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ColorGradient(100, 100, "red", "black"); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "pixmap"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.png';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ColorGradient(100, 100, "red", "black"); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "pixmap"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.jpg';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ColorGradient(100, 100, "red", "black"); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "pixmap"); IsZero(RealVector(GetRGB(AverageColor(a))) − RealVector(GetRGB(AverageColor(b))), 1E-2)', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.wav';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ SineTone(400, 0.1, 5); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "sound"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;


  //
  // Loading
  //

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '0');
  try
    Test('LoadFromFile(FileName, "integer")', 0);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-0');
  try
    Test('LoadFromFile(FileName, "integer")', 0);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '0000000 0000000');
  try
    Test('LoadFromFile(FileName, "integer")', 0);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '123');
  try
    Test('LoadFromFile(FileName, "integer")', 123);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-123');
  try
    Test('LoadFromFile(FileName, "integer")', -123);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−123');
  try
    Test('LoadFromFile(FileName, "integer")', -123);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '123 456 789');
  try
    Test('LoadFromFile(FileName, "integer")', 123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '123 456 789'#13#10);
  try
    Test('LoadFromFile(FileName, "integer")', 123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-123 456 789');
  try
    Test('LoadFromFile(FileName, "integer")', -123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−123 456 789');
  try
    Test('LoadFromFile(FileName, "integer")', -123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '123 456 789');
  try
    Test('LoadFromFile(FileName, "integer")', 123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-123 456 789');
  try
    Test('LoadFromFile(FileName, "integer")', -123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−123 456 789');
  try
    Test('LoadFromFile(FileName, "integer")', -123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '123 456 789');
  try
    Test('LoadFromFile(FileName, "integer")', 123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-123 456 789');
  try
    Test('LoadFromFile(FileName, "integer")', -123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−123 456 789');
  try
    Test('LoadFromFile(FileName, "integer")', -123456789);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5.3');
  try
    Test('LoadFromFile(FileName, "integer")', failure, 'Object cannot be represented as a 64-bit integer.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '.3');
  try
    Test('LoadFromFile(FileName, "integer")', failure, 'Object cannot be represented as a 64-bit integer.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '');
  try
    Test('LoadFromFile(FileName, "integer")', failure, 'Object cannot be represented as a 64-bit integer.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, 'unicorn');
  try
    Test('LoadFromFile(FileName, "integer")', failure, 'Object cannot be represented as a 64-bit integer.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '536254175021596305541505125117456540');
  try
    Test('LoadFromFile(FileName, "integer")', failure, 'Object cannot be represented as a 64-bit integer.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '16#33CCFF');
  try
    Test('LoadFromFile(FileName, "integer")', failure, 'Object cannot be represented as a 64-bit integer.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, #0#0#0);
  try
    Test('LoadFromFile(FileName, "integer")', failure, 'Object cannot be represented as a 64-bit integer.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '394');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(394, 1));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-394');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-394, 1));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−394');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-394, 1));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '0');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(0, 1));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-0');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(0, 1));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−0');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-0, 1));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '1/2');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(1, 2));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-1/2');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-1, 2));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−1/2');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-1, 2));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3/4');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(3, 4));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '32/64');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(1, 2));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '53628178/53628179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-53628178/53628179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '53 628 178/53 628 179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '53 628 178/53 628 179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '53 628 178/53 628 179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-53 628 178/53 628 179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-53 628 178/53 628 179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '- 53 628 178/53 628 179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '− 53 628 178/53 628 179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−53 628 178/53 628 179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−53 628 178/53 628 179');
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−53 628 178/53 628 179'#13#10);
  try
    Test('LoadFromFile(FileName, "rational number")', rat(-53628178, 53628179));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '123');
  try
    Test('LoadFromFile(FileName, "real number")', 123.0);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '123.5');
  try
    Test('LoadFromFile(FileName, "real number")', 123.5);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-123');
  try
    Test('LoadFromFile(FileName, "real number")', -123.0);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-123.5');
  try
    Test('LoadFromFile(FileName, "real number")', -123.5);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−123');
  try
    Test('LoadFromFile(FileName, "real number")', -123.0);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−123.5');
  try
    Test('LoadFromFile(FileName, "real number")', -123.5);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−123.5'#13#10);
  try
    Test('LoadFromFile(FileName, "real number")', -123.5);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.14159265');
  try
    Test('LoadFromFile(FileName, "real number")', 3.14159265);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.1415926535897932384626');
  try
    Test('LoadFromFile(FileName, "real number")', 3.1415926535897932384626);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '1E6');
  try
    Test('LoadFromFile(FileName, "real number")', 1E6);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '7.532E25');
  try
    Test('LoadFromFile(FileName, "real number")', 7.532E25);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-7.532E25');
  try
    Test('LoadFromFile(FileName, "real number")', -7.532E25);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '7.532E-25');
  try
    Test('LoadFromFile(FileName, "real number")', 7.532E-25);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-7.532E-25');
  try
    Test('LoadFromFile(FileName, "real number")', -7.532E-25);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '+7.532E25');
  try
    Test('LoadFromFile(FileName, "real number")', 7.532E25);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '+7.532E+25');
  try
    Test('LoadFromFile(FileName, "real number")', 7.532E25);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '7.532E+25');
  try
    Test('LoadFromFile(FileName, "real number")', 7.532E25);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '56 352 471');
  try
    Test('LoadFromFile(FileName, "real number")', 56352471.0);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '56 352 471.5236');
  try
    Test('LoadFromFile(FileName, "real number")', 56352471.5236);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '56 352 471.526 412 475');
  try
    Test('LoadFromFile(FileName, "real number")', 56352471.526412475);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '- 56 352 471.526 412 475');
  try
    Test('LoadFromFile(FileName, "real number")', -56352471.526412475);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '- 56 352 471.526 412 475E-20');
  try
    Test('LoadFromFile(FileName, "real number")', -56352471.526412475E-20);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '.5');
  try
    Test('LoadFromFile(FileName, "real number")', 0.5);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-.5');
  try
    Test('LoadFromFile(FileName, "real number")', -0.5);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5269611515646514375253451651507449097105743634518465194104899911');
  try
    Test('LoadFromFile(FileName, "real number")', 5269611515646514375253451651507449097105743634518465194104899911.0);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '7.');
  try
    Test('LoadFromFile(FileName, "real number")', 7.0);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5.6.1');
  try
    Test('LoadFromFile(FileName, "real number")', failure, 'Cannot convert string "5.6.1" to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '');
  try
    Test('LoadFromFile(FileName, "real number")', failure, 'Cannot convert string "" to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5F20');
  try
    Test('LoadFromFile(FileName, "real number")', failure, 'Cannot convert string "5F20" to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '7.3+');
  try
    Test('LoadFromFile(FileName, "real number")', failure, 'Cannot convert string "7.3+" to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, 'I suspect nargles are behind it.');
  try
    Test('LoadFromFile(FileName, "real number")', failure, 'Cannot convert string "I suspect nargles are behind it." to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, 'abc');
  try
    Test('LoadFromFile(FileName, "real number")', failure, 'Cannot convert string "abc" to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '53E20E10');
  try
    Test('LoadFromFile(FileName, "real number")', failure, 'Cannot convert string "53E20E10" to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-');
  try
    Test('LoadFromFile(FileName, "real number")', failure, 'Cannot convert string "-" to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-5');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-5));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−5');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-5));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '0');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(0));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-0');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(0));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '123 456 789');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(123456789));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-52 000');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-52000));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.1415');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(3.1415));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-5.2');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-5.2));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.14 15 92 65 35 89 79 32 38');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(3.141592653589793238));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '7.2E20');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(7.2E20));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-7.2E20');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-7.2E20));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '7.2E-20');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(7.2E-20));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-7.2E-20');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-7.2E-20));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '.5');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(0.5));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '1 + i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(1 + ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 + 3⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5+3⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5+3i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, ' 5+3*i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, ' 5+3*i'#13#10);
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 - 3⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 - 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 − 3⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 - 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−5 − 3⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-5 - 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−5 + 3⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '−3⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, 'i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '-i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(-ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.22⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(3.22*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.22E20⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(3.22E20*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.22E-20⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(3.22E-20*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5E-10 + 3⋅i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5E-10 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 + 3*i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 + 3i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 + 3 i');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 + 3⋅j');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 + 3*j');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 + 3j');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5 + 3 j');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3i + 5');
  try
    Test('LoadFromFile(FileName, "complex number")', TASC(5 + 3*ImaginaryUnit));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '');
  try
    Test('LoadFromFile(FileName, "complex number")', failure, 'Cannot convert string "" to a complex number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, 'g');
  try
    Test('LoadFromFile(FileName, "complex number")', failure, 'Cannot convert string "g" to a complex number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5.36.2 + i');
  try
    Test('LoadFromFile(FileName, "complex number")', failure, 'Cannot convert string "5.36.2 + i" to a complex number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '1 + k');
  try
    Test('LoadFromFile(FileName, "complex number")', failure, 'Cannot convert string "1 + k" to a complex number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5/i');
  try
    Test('LoadFromFile(FileName, "complex number")', failure, 'Cannot convert string "5/i" to a complex number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5E2E2 + 3 i');
  try
    Test('LoadFromFile(FileName, "complex number")', failure, 'Cannot convert string "5E2E2 + 3 i" to a complex number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5'#13#10'6.1'#13#10'2.15E20'#13#10'-3.1415'#13#10'−7'#13#10'0');
  try
    Test('LoadFromFile(FileName, "real vector")', [5, 6.1, 2.15E20, -3.1415, -7, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5'#13#10'6.1'#13#10'2.15E20'#13#10'-3.1415'#13#10'−7'#13#10'0'#13#10);
  try
    Test('LoadFromFile(FileName, "real vector")', [5, 6.1, 2.15E20, -3.1415, -7, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5'#9'6.1'#9'2.15E20'#9'-3.1415'#9'−7'#9'0'#9);
  try
    Test('LoadFromFile(FileName, "real vector")', [5, 6.1, 2.15E20, -3.1415, -7, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5;6.1;2.15E20;-3.1415;−7;0');
  try
    Test('LoadFromFile(FileName, "real vector")', [5, 6.1, 2.15E20, -3.1415, -7, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5; 6.1; 2.15E20; -3.1415; −7; 0');
  try
    Test('LoadFromFile(FileName, "real vector")', [5, 6.1, 2.15E20, -3.1415, -7, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5,6.1,2.15E20,-3.1415,−7,0');
  try
    Test('LoadFromFile(FileName, "real vector")', [5, 6.1, 2.15E20, -3.1415, -7, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5'#13#10'6.1'#13#10'2.15E20'#13#10'-3.1415'#13#10'−7'#13#10'0');
  try
    Test('LoadFromFile(FileName, "complex vector")', [TASC(5), 6.1, 2.15E20, -3.1415, -7, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '5'#13#10'6.1'#13#10'2.15E20'#13#10'-3.1415'#13#10'−7'#13#10'0'#13#10#13#10);
  try
    Test('LoadFromFile(FileName, "complex vector")', [TASC(5), 6.1, 2.15E20, -3.1415, -7, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '10');
  try
    Test('LoadFromFile(FileName, "real vector")', [10]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '');
  try
    Test('LoadFromFile(FileName, "real vector")', TAlgosimRealVector.Create);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '10'#13#10);
  try
    Test('LoadFromFile(FileName, "real vector")', [10]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '10'#13#10#13#10);
  try
    Test('LoadFromFile(FileName, "real vector")', [10]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, #13#10);
  try
    Test('LoadFromFile(FileName, "real vector")', TAlgosimRealVector.Create);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.51'#13#10'3 + i'#13#10'7-4⋅i'#13#10'-4.2 + 5.12⋅i'#13#10'-i'#13#10'5+6i');
  try
    Test('LoadFromFile(FileName, "complex vector")', [3.51, 3 + ImaginaryUnit, 7 - 4*ImaginaryUnit, -4.2 + 5.12*ImaginaryUnit, -ImaginaryUnit, 5 + 6*ImaginaryUnit]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.51'#9'3 + i'#9'7-4⋅i'#9'-4.2 + 5.12⋅i'#9'-i'#9'5+6i');
  try
    Test('LoadFromFile(FileName, "complex vector")', [3.51, 3 + ImaginaryUnit, 7 - 4*ImaginaryUnit, -4.2 + 5.12*ImaginaryUnit, -ImaginaryUnit, 5 + 6*ImaginaryUnit]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.51; 3 + i; 7-4⋅i; -4.2 + 5.12⋅i; -i; 5+6i');
  try
    Test('LoadFromFile(FileName, "complex vector")', [3.51, 3 + ImaginaryUnit, 7 - 4*ImaginaryUnit, -4.2 + 5.12*ImaginaryUnit, -ImaginaryUnit, 5 + 6*ImaginaryUnit]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.51, 3 + i, 7-4⋅i, -4.2 + 5.12⋅i, -i, 5+6i');
  try
    Test('LoadFromFile(FileName, "complex vector")', [3.51, 3 + ImaginaryUnit, 7 - 4*ImaginaryUnit, -4.2 + 5.12*ImaginaryUnit, -ImaginaryUnit, 5 + 6*ImaginaryUnit]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.51,3 + i,7-4⋅i,-4.2 + 5.12⋅i,-i,5+6i');
  try
    Test('LoadFromFile(FileName, "complex vector")', [3.51, 3 + ImaginaryUnit, 7 - 4*ImaginaryUnit, -4.2 + 5.12*ImaginaryUnit, -ImaginaryUnit, 5 + 6*ImaginaryUnit]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.51, 3 + i, 7-4⋅i, -4.2 + 5.12⋅i, -i,5+6i'#13#10);
  try
    Test('LoadFromFile(FileName, "complex vector")', [3.51, 3 + ImaginaryUnit, 7 - 4*ImaginaryUnit, -4.2 + 5.12*ImaginaryUnit, -ImaginaryUnit, 5 + 6*ImaginaryUnit]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '3.51,3 + i, 7-4⋅i, -4.2 + 5.12⋅i;-i;5+6i'#13#10);
  try
    Test('LoadFromFile(FileName, "complex vector")', [3.51, 3 + ImaginaryUnit, 7 - 4*ImaginaryUnit, -4.2 + 5.12*ImaginaryUnit, -ImaginaryUnit, 5 + 6*ImaginaryUnit]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '0'#13#10'−5.2E10 − 3.1415E-10⋅i'#13#10'7i'#13#10'5-j'#13#10);
  try
    Test('LoadFromFile(FileName, "complex vector")', [0, -5.2E10 - 3.1415E-10 * ImaginaryUnit, 7 * ImaginaryUnit, 5 - ImaginaryUnit]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn, '0'#13#10'−5.2E10 − 3.1415E-10⋅i'#13#10'7i'#13#10'5-j');
  try
    Test('LoadFromFile(FileName, "real vector")', failure, 'Cannot convert string "−5.2E10 − 3.1415E-10⋅i" to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '0'
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 1, [0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '0'#9'123'#9'394'
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 3, [0, 123, 394]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '0, 123, 394'
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 3, [0, 123, 394]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '0,123,394'
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 3, [0, 123, 394]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '0; 123; 394'
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 3, [0, 123, 394]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '0;123;394'
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 3, [0, 123, 394]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '0, 123, 394'#13#10 +
    '5.2, 3.14159265, -5'
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 3, [0, 123, 394, 5.2, 3.14159265, -5]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '0, 123, 394'#13#10 +
    '5.2, 3.14159265, -5'#13#10 +
    '.58, -.59, −.60'#13#10 +
    '−9.99, 5., .5'
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 3, [0, 123, 394, 5.2, 3.14159265, -5, 0.58, -0.59, -0.60, -9.99, 5, 0.5]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '1;0;0;0;0'#13#10 +
    '0;1;0;0;0'#13#10 +
    '0;0;1;0;0'#13#10 +
    '0;0;0;1;0'#13#10 +
    '0;0;0;0;1'#13#10 +
    '0;0;0;1;0'#13#10 +
    '0;0;1;0;0'#13#10
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 5, [1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '1'#13#10 +
    '2'#13#10 +
    '3'#13#10 +
    '4'#13#10 +
    '5'#13#10 +
    '6'#13#10 +
    '7'#13#10
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', 1, [1, 2, 3, 4, 5, 6, 7]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '1;0;0;0;0'#13#10 +
    '0;1;0;0;0'#13#10 +
    '0;0;1;0;0'#13#10 +
    '0;0;0;1;0'#13#10 +
    '0;0;0;0;1'#13#10 +
    '0;0;0;1'#13#10 +
    '0;0;1;0;0'#13#10
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', failure, 'Attempt to create a non-rectangular matrix.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    '5 + i'#9'i'#9'3.1415'#13#10 +
    '-i'#9'7 − 6⋅i'#9'10i'
  );
  try
    Test('LoadFromFile(FileName, "complex matrix")', 3, [5 + ImaginaryUnit, ImaginaryUnit, 3.1415, -ImaginaryUnit, 7 - 6*ImaginaryUnit, 10*ImaginaryUnit]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'i;0;0;0;0'#13#10 +
    '0;i;0;0;0'#13#10 +
    '0;0;i;0;0'#13#10 +
    '0;0;0;i;0'#13#10 +
    '0;0;0;0;i'#13#10 +
    '0;0;0;i;0'#13#10 +
    '0;0;i;0;0'#13#10
  );
  try
    Test('LoadFromFile(FileName, "complex matrix")', 5, [ImaginaryUnit, 0, 0, 0, 0, 0, ImaginaryUnit, 0, 0, 0, 0, 0, ImaginaryUnit, 0, 0, 0, 0, 0, ImaginaryUnit, 0, 0, 0, 0, 0, ImaginaryUnit, 0, 0, 0, ImaginaryUnit, 0, 0, 0, ImaginaryUnit, 0, 0]);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'i;0;0;0;0'#13#10 +
    '0;i;0;0;0'#13#10 +
    '0;0;i;0;0'#13#10 +
    '0;0;0;i;0'#13#10 +
    '0;0;0;0;i'#13#10 +
    '0;0;0;i;0'#13#10 +
    '0;0;i;0'#13#10
  );
  try
    Test('LoadFromFile(FileName, "complex matrix")', failure, 'Attempt to create a non-rectangular matrix.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;


  //
  // Large files
  //

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ BasisVector(1000000, 1); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real vector"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ComplexVector(BasisVector(1000000, 1)); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex vector"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ IdentityMatrix(1000); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ComplexMatrix(IdentityMatrix(1000)); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ RealMatrix(BasisVector(1000000, 1)); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ComplexMatrix(BasisVector(1000000, 1)); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ RealMatrix(BasisVector(1000000, 1))*; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "real matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ComplexMatrix(BasisVector(1000000, 1))*; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "complex matrix"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.bmp';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ColorGradient(4000, 4000, "red", "gold"); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "pixmap"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.png';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ColorGradient(4000, 4000, "red", "gold"); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "pixmap"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.jpg';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ ColorGradient(4000, 4000, "red", "gold"); SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "pixmap"); norm(RealVector(GetRGB(AverageColor(a))) − RealVector(GetRGB(AverageColor(b)))) < 0.01', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  try
    Test('a ≔ 10000000⋅"Lorem ipsum. "; SaveToFile(a, FileName); b ≔ LoadFromFile(FileName, "string"); a = b', True);
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;


  //
  // Loading with type inference
  //

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName)', 'I suspect nargles are behind it.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.bmp';
  Test('FileName ≔ "' + fn + '"', fn);
  TASPixmap.CreateGradient(100, 100, clRed, clGreen).SaveToFile(fn, ifBitmap);
  try
    Test('LoadFromFile(FileName)', ASO(TASPixmap.CreateGradient(100, 100, clRed, clGreen)));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.png';
  Test('FileName ≔ "' + fn + '"', fn);
  TASPixmap.CreateGradient(100, 100, clRed, clGreen).SaveToFile(fn, ifPng);
  try
    Test('LoadFromFile(FileName)', ASO(TASPixmap.CreateGradient(100, 100, clRed, clGreen)));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.jpg';
  Test('FileName ≔ "' + fn + '"', fn);
  TASPixmap.CreateGradient(100, 100, clRed, clGreen).SaveToFile(fn, ifJpg);
  try
    Test('string(LoadFromFile(FileName))', 'A pixmap of size 100×100.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.wav';
  Test('FileName ≔ "' + fn + '"', fn);
  SaveSoundToFile(SineTone(400, 0.1, 5), fn);
  try
    Test('LoadFromFile(FileName)', ASO(SineTone(400, 0.1, 5)));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;


  //
  // Invalid arguments or file format
  //

  fn := TPath.GetTempPath + 'algosim_selftest.txt';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "nargle")', failure, 'Unknown class type: "nargle".');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName)', failure, 'Unknown file extension: "nargle".');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "integer")', failure, 'Object cannot be represented as a 64-bit integer.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "rational number")', failure, 'Object isn''t a rational number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "real number")', failure, 'Cannot convert string "I suspect nargles are behind it." to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "complex number")', failure, 'Cannot convert string "I suspect nargles are behind it." to a complex number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "real vector")', failure, 'Cannot convert string "I suspect nargles are behind it." to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "complex vector")', failure, 'Cannot convert string "I suspect nargles are behind it." to a complex number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "real matrix")', failure, 'Cannot convert string "I suspect nargles are behind it." to a real number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "complex matrix")', failure, 'Cannot convert string "I suspect nargles are behind it." to a complex number.');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "pixmap")', failure, 'Invalid image');
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;

  fn := TPath.GetTempPath + 'algosim_selftest.nargle';
  Test('FileName ≔ "' + fn + '"', fn);
  TFile.WriteAllText(fn,
    'I suspect nargles are behind it.'
  );
  try
    Test('LoadFromFile(FileName, "sound")', failure, Format('The specified file "%s" appears not to be a RIFF file.', [fn]));
  finally
    if TFile.Exists(fn) then
      TFile.Delete(fn);
  end;





  //
  //
  //  CHAPTER 15
  //  Pixmaps
  //
  //

  Chapter('Pixmaps');

  //
  // width, height, size, IsSquare, IsRow, IsCol
  //

  Test('width(CreatePixmap(50, 20, "red"))', 50);
  Test('height(CreatePixmap(50, 20, "red"))', 20);
  Test('size(CreatePixmap(50, 20, "red"))', ASOSize(20, 50));

  Test('width(CreatePixmap(150, 120, "red"))', 150);
  Test('height(CreatePixmap(150, 120, "red"))', 120);
  Test('size(CreatePixmap(150, 120, "red"))', ASOSize(120, 150));

  Test('IsSquare(CreatePixmap(50, 20, "red"))', False);
  Test('IsSquare(CreatePixmap(50, 25, "red"))', False);
  Test('IsSquare(CreatePixmap(50, 49, "red"))', False);
  Test('IsSquare(CreatePixmap(50, 50, "red"))', True);
  Test('IsSquare(CreatePixmap(50, 51, "red"))', False);
  Test('IsSquare(CreatePixmap(50, 55, "red"))', False);
  Test('IsSquare(CreatePixmap(50, 75, "red"))', False);
  Test('IsSquare(CreatePixmap(50, 100, "red"))', False);
  Test('IsSquare(CreatePixmap(50, 150, "red"))', False);

  Test('IsRow(CreatePixmap(50, 20, "red"))', False);
  Test('IsRow(CreatePixmap(50, 10, "red"))', False);
  Test('IsRow(CreatePixmap(50, 2, "red"))', False);
  Test('IsRow(CreatePixmap(50, 100, "red"))', False);
  Test('IsRow(CreatePixmap(50, 1, "red"))', True);
  Test('IsRow(CreatePixmap(50, 200, "red"))', False);

  Test('IsCol(CreatePixmap(500, 20, "red"))', False);
  Test('IsCol(CreatePixmap(10, 20, "red"))', False);
  Test('IsCol(CreatePixmap(5, 20, "red"))', False);
  Test('IsCol(CreatePixmap(2, 20, "red"))', False);
  Test('IsCol(CreatePixmap(1, 20, "red"))', True);
  Test('IsCol(CreatePixmap(100, 20, "red"))', False);
  Test('IsCol(CreatePixmap(20, 20, "red"))', False);


  //
  // AddBorder
  //

  Test('AddBorder(DebugObject("Sally"))', failure, 'Too few arguments. A required argument of type color is missing.');
  Test('AddBorder(DebugObject("Sally"), "black")', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('AddBorder(DebugObject("Sally"), "black", −4)', failure, 'A non-negative integer was expected as argument 3, but "-4" was given.');
  Test('AddBorder(DebugObject("Sally"), "black", 10)[0, 0]', ASOColor(clBlack));
  Test('AddBorder(DebugObject("Sally"), "red", 10)[0, 0]', ASOColor(clRed));

  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[0, 0]', ASOColor(clBlack));

  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, -1]', failure, 'Index (50, -1) out of bounds.');
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 0]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 1]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 2]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 3]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 4]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 5]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 6]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 7]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 8]', ASOColor(clWhite));

  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 110]', failure, 'Index (50, 110) out of bounds.');
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 109]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 108]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 107]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 106]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 105]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 104]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 103]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 102]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 101]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 100]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[50, 99]', ASOColor(clWhite));

  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[-1, 50]', failure, 'Index (-1, 50) out of bounds.');
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[0, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[1, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[2, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[3, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[4, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[5, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[6, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[7, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[8, 50]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[9, 50]', ASOColor(clWhite));

  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[116, 50]', failure, 'Index (116, 50) out of bounds.');
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[115, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[114, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[113, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[112, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[111, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[110, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[109, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[108, 50]', ASOColor(clBlack));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[107, 50]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[106, 50]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[105, 50]', ASOColor(clWhite));
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8)[104, 50]', ASOColor(clWhite));

  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8) = AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 7) ', False);
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8) = AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8) ', True);
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 8) = AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 9) ', False);
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5) = AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 4) ', False);
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5) = AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 5) ', True);
  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5) = AddBorder(CreatePixmap(100, 100, "white"), "black", 5, 6) ', False);

  Test('AddBorder(CreatePixmap(100, 100, "white"), "black", 5, -8)', failure, 'A non-negative integer was expected as argument 4, but "-8" was given.');

  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[0, 0]', ASOColor(clYellow));

  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, -1]', failure, 'Index (50, -1) out of bounds.');
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 0]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 1]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 2]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 3]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 4]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 5]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 6]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 7]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 8]', ASOColor(clGreen));

  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 110]', failure, 'Index (50, 110) out of bounds.');
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 109]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 108]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 107]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 106]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 105]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 104]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 103]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 102]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 101]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 100]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[50, 99]', ASOColor(clGreen));

  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[-1, 50]', failure, 'Index (-1, 50) out of bounds.');
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[0, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[1, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[2, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[3, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[4, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[5, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[6, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[7, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[8, 50]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[9, 50]', ASOColor(clGreen));

  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[116, 50]', failure, 'Index (116, 50) out of bounds.');
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[115, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[114, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[113, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[112, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[111, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[110, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[109, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[108, 50]', ASOColor(clYellow));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[107, 50]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[106, 50]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[105, 50]', ASOColor(clGreen));
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8)[104, 50]', ASOColor(clGreen));

  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8) = AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 7) ', False);
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8) = AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8) ', True);
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 8) = AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 9) ', False);
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5) = AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 4) ', False);
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5) = AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 5) ', True);
  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5) = AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, 6) ', False);

  Test('AddBorder(CreatePixmap(100, 100, "green"), "yellow", 5, -8)', failure, 'A non-negative integer was expected as argument 4, but "-8" was given.');

  Test('AddBorder(CreatePixmap(200, 100, "red"), "red", 10, 20) = CreatePixmap(220, 140, "red")', True);


  //
  // ExtendBorder
  //

  Test('ExtendBorder(ColorGradient(100, 100, "red", "blue"))', failure, 'A required argument of type integer is missing.');
  Test('ExtendBorder(ColorGradient(100, 100, "red", "blue"), −2)', failure, 'A non-negative integer was expected as argument 2, but "-2" was given.');
  Test('string(ExtendBorder(ColorGradient(100, 100, "red", "blue"), 10))', 'A pixmap of size 120×120.');
  Test('string(ExtendBorder(ColorGradient(100, 100, "red", "blue"), 10, 50))', 'A pixmap of size 120×200.');
  Test('string(ExtendBorder(ColorGradient(100, 100, "red", "blue"), 10, 50, 713))', failure, 'Too many arguments.');

  Test('ExtendBorder(CreatePixmap(200, 100, "red"), 10, 20) = CreatePixmap(220, 140, "red")', True);
  Test('ExtendBorder(ColorGradient(500, 100, "red", "blue"), 125, 0) = ColorGradient(750, 100, "red", "blue")', True);


  //
  // CreatePixmap, IsOnlyBackground, ExpandCanvas, auto cropping
  //

  Test('CreatePixmap(100)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('CreatePixmap(100, 100)', ASO(TASPixmap.Create(100, 100, clWhite)));
  Test('CreatePixmap(100, 100, "red")', ASO(TASPixmap.Create(100, 100, clRed)));
  Test('CreatePixmap(100, 100, "red", "blue")', ASO(TASPixmap.Create(100, 100, clRed)));

  Test('ExpandCanvas(CreatePixmap(100, 100, "red"), 10) = AddBorder(CreatePixmap(100, 100, "red"), "white", 10)', True);
  Test('ExpandCanvas(CreatePixmap(100, 100, "red"), 10, 2) = AddBorder(CreatePixmap(100, 100, "red"), "white", 10, 2)', True);

  Test('ExpandCanvas(CreatePixmap(100, 100, "red", "yellow"), 10) = AddBorder(CreatePixmap(100, 100, "red"), "yellow", 10)', True);
  Test('ExpandCanvas(CreatePixmap(100, 100, "red", "yellow"), 10, 2) = AddBorder(CreatePixmap(100, 100, "red"), "yellow", 10, 2)', True);

  Test('IsOnlyBackground(CreatePixmap(100, 100))', True);
  Test('IsOnlyBackground(CreatePixmap(100, 100, "red"))', False);
  Test('IsOnlyBackground(CreatePixmap(100, 100, "red", "red"))', True);

  Test('a ≔ FillRect(CreatePixmap(100, 100, "yellow"), 40, 40, 20, 20, "black"); AutoCrop(a) = a', True);
  Test('AutoCropRect(a)', ASOIntRect(0, 0, 100, 100));

  Test('a ≔ FillRect(CreatePixmap(100, 100, "yellow", "yellow"), 40, 40, 20, 20, "black");', null);
  Test('AutoCrop(a) = CreatePixmap(20, 20, "black")', True);
  Test('AutoCropRect(a)', ASOIntRect(40, 40, 20, 20));

  Test('a ≔ DrawLine(FillRect(CreatePixmap(100, 100), 10, 20, 5, 5, "black"), 60, 70, 70, 80, "black");', null);
  Test('AutoCropRect(a)', ASOIntRect(10, 20, 61, 61));


  //
  // FillRect
  //

  Test('FillRect(CreatePixmap(100, 100, "yellow"), 25, 25, 50, 50) = AddBorder(CreatePixmap(50, 50, "black"), "yellow", 25, 25)', True);
  Test('FillRect(CreatePixmap(100, 100, "yellow"), 25, 25, 50, 50, "blue") = AddBorder(CreatePixmap(50, 50, "blue"), "yellow", 25, 25)', True);

  Test('FillRect(CreatePixmap(200, 100, "blue"), 100, 0, 100, 100, "yellow") = FillRect(FillRect(CreatePixmap(200, 100), 0, 0, 100, 100, "blue"), 100, 0, 100, 100, "yellow")', True);
  Test('FillRect(CreatePixmap(200, 100, "blue"), 100, 0, 100, 100, "yellow") = FillRect(FillRect(FillRect(CreatePixmap(200, 100), 0, 0, 100, 100, "blue"), 100, 0, 100, 100, "yellow"), 200, 0, 100, 100, "red")', True);
  Test('FillRect(CreatePixmap(200, 100, "blue"), 100, 0, 100, 100, "yellow") = FillRect(FillRect(FillRect(CreatePixmap(200, 100), 0, 0, 100, 100, "blue"), 100, 0, 100, 100, "yellow"), 2000, 500000, 100, 100, "red")', True);
  Test('FillRect(CreatePixmap(200, 100, "blue"), 100, 0, 100, 100, "yellow") = FillRect(FillRect(FillRect(CreatePixmap(200, 100), 0, 0, 100, 100, "blue"), 100, 0, 100, 100, "yellow"), 2000, 500000, 100000, 100000, "red")', True);

  Test('FillRect(CreatePixmap(100, 200, "blue"), 50, 20, 200, 254, "blue") = CreatePixmap(100, 200, "blue")', True);
  Test('FillRect(CreatePixmap(100, 200, "blue"), 50, 20, 2000, 2540, "blue") = CreatePixmap(100, 200, "blue")', True);
  Test('FillRect(CreatePixmap(100, 200, "blue"), 50, 20, 1, 1, "blue") = CreatePixmap(100, 200, "blue")', True);
  Test('FillRect(CreatePixmap(100, 200, "blue"), 50, 20, 200, 0, "blue") = CreatePixmap(100, 200, "blue")', True);
  Test('FillRect(CreatePixmap(100, 200, "blue"), 50, 20, 200, 0, "red") = CreatePixmap(100, 200, "blue")', True);
  Test('FillRect(CreatePixmap(100, 200, "blue"), 50, 20, 0, 100, "red") = CreatePixmap(100, 200, "blue")', True);
  Test('FillRect(CreatePixmap(100, 200, "blue"), 50, 20, 0, 0, "red") = CreatePixmap(100, 200, "blue")', True);
  Test('FillRect(CreatePixmap(100, 200, "blue"), 50, 20, 1, 100, "red") = CreatePixmap(100, 200, "blue")', False);
  Test('FillRect(CreatePixmap(100, 200, "blue"), 20050, 10020, 15, 13, "blue") = CreatePixmap(100, 200, "blue")', True);


  //
  // DrawSquare, DrawSquares
  //

  Test('DrawSquare(CreatePixmap(10, 10, "black"))', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('DrawSquare(CreatePixmap(10, 10, "black"), 5)', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('AverageColor(DrawSquare(CreatePixmap(10, 10, "black"), 5, 5, "white", 1))', ASO(TRGB.Create(0.01, 0.01, 0.01)));
  Test('AverageColor(DrawSquare(CreatePixmap(10, 10, "black"), 5, 5, "white", 2))', ASO(TRGB.Create(0.04, 0.04, 0.04)));
  Test('AverageColor(DrawSquare(CreatePixmap(10, 10, "black"), 5, 5, "white", 3))', ASO(TRGB.Create(0.09, 0.09, 0.09)));
  Test('AverageColor(DrawSquare(CreatePixmap(10, 10, "black"), 5, 5, "white", 4))', ASO(TRGB.Create(0.16, 0.16, 0.16)));
  Test('AverageColor(DrawSquare(CreatePixmap(10, 10, "black"), 5, 5, "white", 5))', ASO(TRGB.Create(0.25, 0.25, 0.25)));
  Test('AverageColor(DrawSquare(CreatePixmap(10, 10, "black"), 5, 5, "black", 5))', ASOColor(clBlack));

  Test('AverageColor(DrawSquares(CreatePixmap(10, 10, "black"), 10, "red", 1))', failure, 'A list or set of integer points was expected as argument 2.');
  Test('AverageColor(DrawSquares(CreatePixmap(10, 10, "black"), ''(❨0, 0❩, ❨4, 4❩, ❨5, 5❩, ❨9, 9❩, ❨0, 9❩, ❨9, 0❩), "red", 1))', ASO(TRGB.Create(0.06, 0, 0)));
  Test('AverageColor(DrawSquares(CreatePixmap(10, 10, "black"), ''(❨0, 0❩, ❨4, 4❩, ❨5, 5❩, ❨9, 9❩, ❨0, 9❩, ❨9, 0❩, ❨10, 10❩, ❨10, 2❩, ❨5, 10❩, ❨156155612, 02156156❩), "red", 1))', ASO(TRGB.Create(0.06, 0, 0)));
  Test('AverageColor(DrawSquares(CreatePixmap(10, 10, "black"), ''(❨0, 0❩, ❨4, 4❩, ❨5, 5❩, ❨9, 9❩, ❨0, 9❩, ❨9, 0❩, ❨-10, 10❩, ❨-10, 2❩, ❨-5, 10❩, ❨-156155612, -02156156❩), "red", 1))', ASO(TRGB.Create(0.06, 0, 0)));


  //
  // DrawDisk, DrawDisks
  //

  Test('DrawDisk(CreatePixmap(10, 10, "black"))', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('DrawDisk(CreatePixmap(10, 10, "black"), 5)', failure, 'Too few arguments. A required argument of type integer is missing.');

  Test('max(abs(RealVector(GetRGB(AverageColor(DrawDisk(CreatePixmap(1000, 1000, "black"), 500, 500, "white", 250)))) − π⋅❨1, 1, 1❩/16)) < 0.0001', True);
  Test('max(abs(RealVector(GetRGB(AverageColor(DrawDisk(CreatePixmap(1000, 1000, "black"), 500, 500, "white", 500)))) − π⋅❨1, 1, 1❩/4)) < 0.0001', True);

  Test('max(abs(RealVector(GetRGB(AverageColor(DrawDisks(CreatePixmap(1000, 1000, "black"), ''(❨250, 250❩, ❨750, 250❩, ❨250, 750❩, ❨750, 750❩), "red", 250)))) − π⋅❨1, 0, 0❩/4)) < 0.0002', True);

  Test('max(abs(RealVector(GetRGB(AverageColor(DrawDisks(CreatePixmap(1000, 1000, "black"), ''(❨250, 250❩, ❨750, 250❩, ❨250, 750❩, ❨750, 750❩, ❨7500, 7500❩, ❨750, -750❩, ❨-750, 750❩), "red", 250)))) − π⋅❨1, 0, 0❩/4)) < 0.0002', True);


  //
  // DrawLine, DrawLines, DrawPolygon
  //

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 0, 0, 99, 99, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(x = y, color("white"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.01, 0.01)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 99, 99, 0, 0, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(x = y, color("white"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.01, 0.01)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 99, 0, 0, 99, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(x = 99 - y, color("white"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.01, 0.01)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 0, 99, 99, 0, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(x = 99 - y, color("white"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.01, 0.01)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 0, 50, 99, 50, "red");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(y = 50, color("red"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.00, 0.00)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 2, 0, 2, 99, "blue");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(x = 2, color("blue"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.00, 0.00, 0.01)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 5, 10, 10, 10, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(y = 10 ∧ x ≥ 5 ∧ x ≤ 10, color("white"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.0006, 0.0006, 0.0006)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 5, 10, 5, 10, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(y = 10 ∧ x = 5, color("white"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.0001, 0.0001, 0.0001)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 0, 0, 150, 150, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(x = y, color("white"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.01, 0.01)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), -200, -200, 99, 99, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(x = y, color("white"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.01, 0.01)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), -100, -100, 300, 300, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(x = y, color("white"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.01, 0.01)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), 0, 50, 500, 50, "red");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(y = 50, color("red"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.00, 0.00)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), -20, 50, 99, 50, "red");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(y = 50, color("red"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.00, 0.00)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), -500, 50, 200, 50, "red");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(y = 50, color("red"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.00, 0.00)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), -50000000, 50, 20000000, 50, "red");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(y = 50, color("red"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.00, 0.00)));

  Test('A ≔ DrawLine(CreatePixmap(100, 100, "black"), -50000000, 50, 1000000000, 50, "red");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(y = 50, color("red"), color("black")))));', null);
  Test('AverageColor(A)', ASO(TRGB.Create(0.01, 0.00, 0.00)));

  Test('DrawLines( CreatePixmap(100, 100, "red"), ''(❨0, 0❩, ❨99, 0❩, ❨99, 99❩, ❨0, 99❩, ❨0, 0❩), "white") = AddBorder(CreatePixmap(98, 98, "red"), "white", 1, 1)', True);
  Test('DrawPolygon( CreatePixmap(100, 100, "red"), ''(❨0, 0❩, ❨99, 0❩, ❨99, 99❩, ❨0, 99❩), "white") = AddBorder(CreatePixmap(98, 98, "red"), "white", 1, 1)', True);

  Test('DrawLines( CreatePixmap(100, 100, "red"), ''(❨50, 50❩, ❨60, 50❩, ❨60, 60❩, ❨50, 60❩, ❨50, 50❩), "white") = DrawPolygon( CreatePixmap(100, 100, "red"), ''(❨50, 50❩, ❨60, 50❩, ❨60, 60❩, ❨50, 60❩), "white")', True);


  //
  // Primitive rotations and reflections
  //

  Test('rot90p( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") ) = FillRect(CreatePixmap(100, 200, "black"), 0, 100, 100, 100, "red")', True);
  Test('rot90p(rot90p( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") )) = FillRect(CreatePixmap(200, 100, "black"), 100, 0, 100, 100, "red")', True);
  Test('rot90p(rot90p(rot90p( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") ))) = FillRect(CreatePixmap(100, 200, "black"), 0, 0, 100, 100, "red")', True);
  Test('rot90p(rot90p(rot90p(rot90p( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") )))) = FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red")', True);

  Test('rot90n( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") ) = FillRect(CreatePixmap(100, 200, "black"), 0, 0, 100, 100, "red")', True);
  Test('rot90n(rot90n( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") )) = FillRect(CreatePixmap(200, 100, "black"), 100, 0, 100, 100, "red")', True);
  Test('rot90n(rot90n(rot90n( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") ))) = FillRect(CreatePixmap(100, 200, "black"), 0, 100, 100, 100, "red")', True);
  Test('rot90n(rot90n(rot90n(rot90n( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") )))) = FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red")', True);

  Test('rot180( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") ) = FillRect(CreatePixmap(200, 100, "black"), 100, 0, 100, 100, "red")', True);
  Test('rot180(rot180( FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red") )) = FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red")', True);

  Test('flip(FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red")) = FillRect(CreatePixmap(200, 100, "black"), 100, 0, 100, 100, "red")', True);
  Test('flip(FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red"), "h") = FillRect(CreatePixmap(200, 100, "black"), 100, 0, 100, 100, "red")', True);
  Test('flip(FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red"), "v") = FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red")', True);

  Test('flip(FillRect(CreatePixmap(100, 200, "black"), 0, 0, 100, 100, "red")) = FillRect(CreatePixmap(100, 200, "black"), 0, 0, 100, 100, "red")', True);
  Test('flip(FillRect(CreatePixmap(100, 200, "black"), 0, 0, 100, 100, "red"), "h") = FillRect(CreatePixmap(100, 200, "black"), 0, 0, 100, 100, "red")', True);
  Test('flip(FillRect(CreatePixmap(100, 200, "black"), 0, 0, 100, 100, "red"), "v") = FillRect(CreatePixmap(100, 200, "black"), 0, 100, 100, 100, "red")', True);

  Test('flip(FillRect(CreatePixmap(100, 200, "black"), 0, 0, 100, 100, "red"), "s")', failure, 'Character "s" not allowed as argument. Expected one of "hv".');

  Test('a ≔ DebugObject("Sally"); b ≔ rot90p(a); a ≠ b ∧ AverageColor(a) = AverageColor(b)', True);
  Test('a ≔ DebugObject("Sally"); b ≔ rot90n(a); a ≠ b ∧ AverageColor(a) = AverageColor(b)', True);
  Test('a ≔ DebugObject("Sally"); b ≔ flip(a, "h"); a ≠ b ∧ AverageColor(a) = AverageColor(b)', True);
  Test('a ≔ DebugObject("Sally"); b ≔ flip(a, "v"); a ≠ b ∧ AverageColor(a) = AverageColor(b)', True);


  //
  // FloodFill and component fill
  //

  Test('A ≔ FloodFill(DrawLine(CreatePixmap(100, 100, "black"), 0, 0, 99, 99, "white"), 10, 50, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(y ≥ x, color("white"), color("black")))));', null);

  Test('A ≔ FloodFill(DrawLine(CreatePixmap(200, 100, "black"), 99, 0, 99, 99, "white"), 50, 50, "white");', null);
  Test('for(y, 0, 99, for(x, 0, 99, assert(A[y, x] = if(x ≤ 99, color("white"), color("black")))));', null);

  Test('FloodFill( DrawPolygon( CreatePixmap(100, 100, "black"), ''(❨10, 10❩, ❨10, 20❩, ❨20, 20❩, ❨20, 10❩), "red") , 15, 15, "red") = FillRect(CreatePixmap(100, 100, "black"), 10, 10, 11, 11, "red")', True);

  Test('a ≔ DrawLines(CreatePixmap(1000, 1000), ''(❨200, 0❩, ❨200, 700❩, ❨800, 700❩, ❨800, 200❩, ❨100, 200❩, ❨100, 500❩, ❨600, 500❩, ❨600, 1000❩), "black");', null);

  Test('b ≔ ComponentHighlight(a, n ↦ ''("#13031D", "#049CC8", "#89AFCE", "#8AC7A6", "#52E303")[n]);', null);
  Test('b = FloodFill(FloodFill(FloodFill(FloodFill(FloodFill(a, 10, 10, "#13031D"), 500, 10, "#049CC8"), 120, 300, "#89AFCE"), 220, 300, "#8AC7A6"), 220, 600, "#52E303")', True);

  Test('b ≔ ComponentHighlight(a, n ↦ ''("#9C7E76", "#F990A7", "#7417EB", "#EA8AF4", "#453AFF")[n]);', null);
  Test('b = FloodFill(FloodFill(FloodFill(FloodFill(FloodFill(a, 10, 10, "#9C7E76"), 500, 10, "#F990A7"), 120, 300, "#7417EB"), 220, 300, "#EA8AF4"), 220, 600, "#453AFF")', True);

  Test('b ≔ ComponentHighlight(a, n ↦ ''("#9C7E76", "#F990A7", "#7417EB", "#EA8AF4")[n]);', failure, 'Index 5 out of bounds.');
  Test('b ≔ ComponentHighlight(a, n ↦ ''("#9C7E76", "#F990A7", "#7417EB", "#EA8AF4", now())[n]);', failure, 'Invalid colour coordinates structure.');
  Test('b ≔ ComponentHighlight(a, n ↦ ''("#9C7E76", "#F990A7", "#7417EB", "#EA8AF4", true)[n]);', failure, 'An object of type "boolean" cannot be converted to a colour.');
  Test('b ≔ ComponentHighlight(a, n ↦ ''("#9C7E76", "#F990A7", "#7417EB", "#EA8AF4", 5.2)[n]);', failure, 'A number which isn''t a 32-bit integer cannot be converted to a colour.');

  Test('ColorCount(a)', 2);
  Test('ComponentHighlight(a) = ComponentHighlight(a)', False);
  Test('ComponentHighlight(a) = ComponentHighlight(a)', False);
  Test('ComponentHighlight(a) = ComponentHighlight(a)', False);
  Test('ColorCount(ComponentHighlight(a))', 6);
  Test('ColorCount(ComponentHighlight(a))', 6);
  Test('ColorCount(ComponentHighlight(a))', 6);


  //
  // shuffle and tiles
  //

  Test('a ≔ DebugObject("Sally"); b ≔ shuffle(a); a ≠ b ∧ AverageColor(a) = AverageColor(b)', True);

  Test('a ≔ ColorGradient(1000, 1000, "red", "black"); b ≔ tiles(a, 10, 10, 0); a ≠ b ∧ AverageColor(a) = AverageColor(b)', True);
  Test('a ≔ ColorGradient(1000, 1000, "red", "black"); b ≔ tiles(a, 10, 10, 0, false); a = b', True);
  Test('size(tiles(a, 10, 10, 20))', ASOSize(1220, 1220));


  //
  // IsGreyscale, ToGreyscale
  //

  Test('IsGreyscale(ColorGradient(1000, 1000, "white", "black"))', True);
  Test('IsGreyscale(ColorGradient(1000, 1000, "white", "silver"))', True);
  Test('IsGreyscale(ColorGradient(1000, 1000, "lightgrey", "darkgrey"))', True);
  Test('IsGreyscale(ColorGradient(1000, 1000, "gold", "silver"))', False);

  Test('a ≔ shuffle(ColorGradient(1000, 1000, "black", "white"));', null);
  Test('IsGreyscale(a)', True);
  Test('a[713, 394] ≔ rgb(0.22, 0.22, 0.23);', null);
  Test('IsGreyscale(a)', False);
  Test('a[713, 394] ≔ rgb(0.22, 0.22, 0.22);', null);
  Test('IsGreyscale(a)', True);

  Test('IsGreyscale(DebugObject("Sally"))', False);
  Test('IsGreyscale(DebugObject("harvestman"))', False);
  Test('IsGreyscale(DebugObject("rabbit"))', False);

  Test('a ≔ shuffle(ColorGradient(1000, 1000, "black", "white"));', null);
  Test('IsGreyscale(a)', True);
  Test('a[713, 394] ≔ rgb(0.22, 0.22, 0.23);', null);
  Test('IsGreyscale(a)', False);
  Test('a ≔ ToGreyscale(a);', null);
  Test('IsGreyscale(a)', True);

  Test('IsGreyscale(ToGreyscale(DebugObject("Sally")))', True);
  Test('IsGreyscale(ToGreyscale(DebugObject("harvestman")))', True);
  Test('IsGreyscale(ToGreyscale(DebugObject("rabbit")))', True);


  //
  // AverageColor, ColorCount, ColorFreqs
  //

  Test('a ≔ ColorGradient(1000, 1000, "black", "white");', null);
  Test('AverageColor(a)', ASO(TRGB.Create(0.5, 0.5, 0.5)));
  Test('ColorCount(a)', 256);
  Test('∀(ColorFreqs(a) @ (x ↦ x[2]), (n ↦ 1000 ∣ n))', True);

  Test('a ≔ ColorGradient(1000, 256, "black", "white");', null);
  Test('AverageColor(a)', ASO(TRGB.Create(0.5, 0.5, 0.5)));
  Test('ColorCount(a)', 256);
  Test('∀(ColorFreqs(a) @ (x ↦ x[2]), (n ↦ n = 1000))', True);

  Test('a ≔ FillRect(CreatePixmap(200, 100, "black"), 0, 0, 100, 100, "red");', null);
  Test('AverageColor(a)', ASO(TRGB.Create(0.5, 0.0, 0.0)));
  Test('ColorCount(a)', 2);
  Test('ColorFreqs(a)', asoarrex([
    asoarrex([ASO(TRGB.Create(0, 0, 0)), ASOInt(10000)]),
    asoarrex([ASO(TRGB.Create(1, 0, 0)), ASOInt(10000)])
  ]));

  Test('pm ≔ FillRect(FillRect(CreatePixmap(300, 100, "red"), 100, 0, 100, 100, "lime"), 200, 0, 100, 100, "blue");', null);
  Test('ColorCount(pm)', 3);
  Test('ColorFreqs(pm)', asoarrex([
    asoarrex([ASO(TRGB.Create(1, 0, 0)), ASOInt(10000)]),
    asoarrex([ASO(TRGB.Create(0, 1, 0)), ASOInt(10000)]),
    asoarrex([ASO(TRGB.Create(0, 0, 1)), ASOInt(10000)])
  ]));
  Test('AverageColor(pm)', ASO(TRGB.Create(1/3, 1/3, 1/3)));
  Test('AverageColor(pm, 0, 0, 100, 100)', ASO(TRGB.Create(1, 0, 0)));
  Test('AverageColor(pm, 100, 0, 100, 100)', ASO(TRGB.Create(0, 1, 0)));
  Test('AverageColor(pm, 200, 0, 100, 100)', ASO(TRGB.Create(0, 0, 1)));


  //
  // Image effects
  //

  Test('a ≔ DebugObject("Sally");', null);

  Test('b ≔ noise(a); a ≠ b', True);
  Test('b ≔ noise(a, 0.1); a ≠ b', True);
  Test('b ≔ noise(a, 0.9); a ≠ b', True);
  Test('b ≔ noise(a, 1); b = CreatePixmap(width(a), height(a), "white")', True);
  Test('b ≔ noise(a, 1, "red"); b = CreatePixmap(width(a), height(a), "red")', True);
  Test('b ≔ noise(a, 1, "red", "green")', failure, 'Too many arguments.');

  Test('b ≔ pixelate(a)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('b ≔ pixelate(a, 10); a ≠ b', True);
  Test('pixelate(a, 10) = pixelate(a, 10, 10)', True);
  Test('pixelate(a, 10) ≠ pixelate(a, 10, 20)', True);
  Test('b ≔ pixelate(a, ⌈width(a) / 5⌉, ⌈height(a) / 5⌉); ColorCount(b)', 25);
  Test('b ≔ pixelate(a, width(a), height(a)); ColorCount(b)', 1);
  Test('pixelate(CreatePixmap(1000, 1000, "red"), 10, 10) = CreatePixmap(1000, 1000, "red")', True);

  Test('b ≔ wind(a, 0, 10, 1); a = b', True);
  Test('b ≔ wind(a, 0.1, 0, 1); a = b', True);
  Test('b ≔ wind(a, 0.1, 10, 1); a = b', False);
  Test('b ≔ wind(a, -1, 10, 1); a = b', failure, 'Invalid probability.');
  Test('b ≔ wind(a, 2, 10, 1); a = b', failure, 'Invalid probability.');
  Test('b ≔ wind(a, 0.1, -5, 1); a = b', failure, 'Invalid distance.');
  Test('b ≔ wind(a, 0.1, 10, -1); a = b', failure, 'Invalid fading.');
  Test('b ≔ wind(a, 0.1, 10, 2); a = b', failure, 'Invalid fading.');

  Test('b ≔ ripple(a, 0, 200); a = b', True);
  Test('b ≔ ripple(a, 10, 200); a ≠ b', True);
  Test('b ≔ ripple(a, -10, 200); a ≠ b', True);

  Test('b ≔ ripple(a, 10, 0)', failure, 'The wavelength must be positive.');
  Test('b ≔ ripple(a, 10, -10)', failure, 'The wavelength must be positive.');

  Test('ripple(a, 10, 200) = ripple(a, 10, 200, "h")', True);
  Test('ripple(a, 10, 200, "v") = ripple(a, 10, 200, "h")', False);
  Test('ripple(a, 10, 200, "s")', failure, 'Character "s" not allowed as argument. Expected one of "hv".');

  Test('b ≔ emboss(a); a ≠ b ∧ IsGreyscale(b)', True);
  Test('emboss(a) = emboss(a, "hv")', True);
  Test('emboss(a) = emboss(a, "h")', False);
  Test('emboss(a) = emboss(a, "v")', False);

  Test('b ≔ EdgeDetect(a); a ≠ b ∧ IsGreyscale(b)', True);
  Test('EdgeDetect(a) = EdgeDetect(a, "hv")', True);
  Test('EdgeDetect(a) = EdgeDetect(a, "h")', False);
  Test('EdgeDetect(a) = EdgeDetect(a, "v")', False);

  Test('b ≔ GaussianBlur(a); a ≠ b', True);
  Test('GaussianBlur(a) = GaussianBlur(a, 2)', True);
  Test('GaussianBlur(a) = GaussianBlur(a, 2, 6)', True);
  Test('norm(RealVector(GetRGB(AverageColor(EdgeDetect(a))))) / norm(RealVector(GetRGB(AverageColor(EdgeDetect(GaussianBlur(a, 6)))))) > 10', True);

  Test('b ≔ BoxBlur(a); a ≠ b', True);
  Test('BoxBlur(a) = BoxBlur(a, 2)', True);
  Test('norm(RealVector(GetRGB(AverageColor(EdgeDetect(a))))) / norm(RealVector(GetRGB(AverageColor(EdgeDetect(BoxBlur(a, 6)))))) > 5', True);

  Test('b ≔ MotionBlur(a); a ≠ b', True);
  Test('MotionBlur(a, 0) = a', True);
  Test('MotionBlur(a, 0, "h") = a', True);
  Test('MotionBlur(a, 0, "v") = a', True);
  Test('MotionBlur(a) = MotionBlur(a, 32)', True);
  Test('MotionBlur(a) = MotionBlur(a, 32, "h")', True);
  Test('MotionBlur(a) = MotionBlur(a, 32, "v")', False);
  Test('MotionBlur(a, 32, "q")', failure, 'Character "q" not allowed as argument. Expected one of "hv".');
  Test('MotionBlur(a, -10)', failure, 'A non-negative integer was expected as argument 2, but "-10" was given.');

  Test('b ≔ DistortMetric(a)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('b ≔ DistortMetric(a, 10); a ≠ b', True);
  Test('b ≔ DistortMetric(a, 10, 20); a ≠ b', True);

  Test('b ≔ DistortColor(a)', failure, 'Too few arguments. A required argument of type integer is missing.');
  Test('b ≔ DistortColor(a, 10); a ≠ b', True);
  Test('b ≔ DistortColor(a, 10, 20); a ≠ b', True);
  Test('b ≔ DistortColor(a, 10, 20, 30); a ≠ b', True);
  Test('a = DistortColor(a, 0)', True);

  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", 1))', ASO(TRGB.Create(1/1, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", 2))', ASO(TRGB.Create(1/4, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", 4))', ASO(TRGB.Create(1/16, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", 5))', ASO(TRGB.Create(1/25, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", 10))', ASO(TRGB.Create(1/100, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", 100))', ASO(TRGB.Create(1/10000, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", 1000))', ASO(TRGB.Create(1/1000000, 0, 0)));

  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", -1))', ASO(TRGB.Create(0/1, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", -2))', ASO(TRGB.Create(3/4, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", -4))', ASO(TRGB.Create(15/16, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", -5))', ASO(TRGB.Create(24/25, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", -10))', ASO(TRGB.Create(99/100, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", -100))', ASO(TRGB.Create(9999/10000, 0, 0)));
  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", -1000))', ASO(TRGB.Create(999999/1000000, 0, 0)));

  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(1000, 1000, "black"), "red", 0))', failure, 'Denominator cannot be zero.');

  Test('AverageColor(DissolveToColorRegularly(CreatePixmap(100, 100, "black"), ((x, y) ↦ rgb(x/99, 0, 0)), 1))', ASO(TRGB.Create(0.5, 0, 0)));
  Test('norm(RealVector(GetRGB(AverageColor(DissolveToColorRegularly(CreatePixmap(100, 100, "black"), ((x, y) ↦ rgb(x/99, 0, 0)), 2)))) − ❨1/8, 0, 0❩) < 0.002', True);
  Test('norm(RealVector(GetRGB(AverageColor(DissolveToColorRegularly(CreatePixmap(100, 100, "black"), ((x, y) ↦ rgb(x/99, 0, 0)), 4)))) − ❨1/32, 0, 0❩) < 0.002', True);
  Test('norm(RealVector(GetRGB(AverageColor(DissolveToColorRegularly(CreatePixmap(100, 100, "black"), ((x, y) ↦ rgb(x/99, 0, 0)), −2)))) − ❨3/8, 0, 0❩) < 0.002', True);

  Test('p ≔ 0.0; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) = 0.000', True);
  Test('p ≔ 0.1; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.2; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.3; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.4; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.5; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.6; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.7; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.8; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.9; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) < 0.001', True);
  Test('p ≔ 1.0; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), "red", p)))) − ❨p, 0, 0❩) = 0.000', True);

  Test('p ≔ 0.0; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) = 0.000', True);
  Test('p ≔ 0.1; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.2; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.3; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.4; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.5; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.6; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.7; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.8; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) < 0.001', True);
  Test('p ≔ 0.9; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) < 0.001', True);
  Test('p ≔ 1.0; norm(RealVector(GetRGB(AverageColor(DissolveToColorStochastically(CreatePixmap(1000, 1000, "black"), ((x, y) ↦ rgb(x/999, 0, 0)), p)))) − ❨p/2, 0, 0❩) = 0.000', True);

  Test('ColorCount(binarize(DebugObject("Sally")))', 2);
  Test('ColorCount(binarize(DebugObject("rabbit")))', 2);
  Test('ColorCount(binarize(DebugObject("harvestman")))', 2);
  Test('a ≔ FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"); binarize(a) = a', True);

  Test('transformation(a, x, ❨x[1] + 4⋅sin(x[2]/2), x[2] + 4⋅sin(x[1]/2)❩);', null);
  Test('transformation(a, x, ❨x[1] + 4⋅sin(x[2]/2), x[2] + 4⋅sin(x[1]/0)❩);', failure, 'Division by zero');
  Test('transformation(a, x, true);', failure, 'Pixmap transformation function must return a two-dimensional real vector.');
  Test('transformation(a, x, ❨x[1] + 4⋅sin(x[2]/2), x[2] + 4⋅sin(x[1]/2), 5❩);', failure, 'Pixmap transformation function must return a two-dimensional real vector.');
  Test('transformation(a, x, ❨x[1] + 4⋅sin(x[2]/2)❩);', failure, 'Pixmap transformation function must return a two-dimensional real vector.');

  for _s in PredefinedKernelNames do
  begin
    Test('type(ConvolutionKernel("' + _s + '"))', 'real matrix');
    Test('convolve(DebugObject("Sally"), ConvolutionKernel("' + _s + '"));', null);
  end;

  Test('ConvolutionKernel("3×3 Nargle Naughtiness")', failure, 'Unknown convolution kernel "3×3 Nargle Naughtiness".');

  Test('a ≔ DebugObject("Sally"); convolve(a, ZeroMatrix(5)) = CreatePixmap(width(a), height(a), "black")', True);
  Test('a ≔ DebugObject("Sally"); convolve(a, ZeroMatrix(4))', failure, 'Convolution kernel must be square with an odd number of rows.');
  Test('a ≔ DebugObject("Sally"); convolve(a, ZeroMatrix(3, 5))', failure, 'Convolution kernel must be square with an odd number of rows.');


  //
  // Image generation
  //

  Test('a ≔ ColorGradient(1000, 1000, "white", "black");', null);
  Test('size(a)', ASOSize(1000, 1000));
  Test('a[0, 0]', ASOColor(clWhite));
  Test('a[999, 0]', ASOColor(clBlack));
  Test('AverageColor(a)', ASO(TRGB.Create(0.5, 0.5, 0.5)));

  Test('a ≔ voronoi(1000, 1001, ''(❨200, 300❩, ❨0, 400❩, ❨700, 600❩, ❨600, 300❩), ''("red", "green", "blue", "yellow"));', null);
  Test('size(a)', ASOSize(1001, 1000));
  Test('ColorCount(a)', 4);

  Test('a ≔ voronoi(1000, 1001, ''(❨200, 300❩, ❨0, 400❩, ❨700, 600❩), ''("red", "green", "blue"));', null);
  Test('size(a)', ASOSize(1001, 1000));
  Test('ColorCount(a)', 3);

  Test('a ≔ voronoi(1000, 1001, ''(❨200, 300❩, ❨0, 400❩), ''("red", "green"));', null);
  Test('size(a)', ASOSize(1001, 1000));
  Test('ColorCount(a)', 2);

  Test('a ≔ voronoi(1000, 1001, ''(❨200, 300❩), ''("red"));', null);
  Test('size(a)', ASOSize(1001, 1000));
  Test('ColorCount(a)', 1);

  Test('a ≔ voronoi(1000, 1001, ''(), ''());', null);
  Test('size(a)', ASOSize(1001, 1000));
  Test('a = CreatePixmap(1000, 1001)', True);

  Test('a ≔ voronoi(1, 1, ''(❨200, 300❩, ❨0, 400❩, ❨700, 600❩, ❨600, 300❩), ''("red", "green", "blue", "yellow"));', null);
  Test('size(a)', ASOSize(1, 1));
  Test('a = CreatePixmap(1, 1, "red")', True);


  //
  // Linear transformations
  //

  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/1000)', failure, 'Cannot create empty pixmap.');
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/100) = FillRect(CreatePixmap(2, 1, "white"), 0, 0, 1, 1, "black")', True);
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/10) = FillRect(CreatePixmap(20, 10, "white"), 0, 0, 10, 10, "black")', True);
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/4) = FillRect(CreatePixmap(50, 25, "white"), 0, 0, 25, 25, "black")', True);
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/2) = FillRect(CreatePixmap(100, 50, "white"), 0, 0, 50, 50, "black")', True);
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1) = FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black")', True);
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 2) = FillRect(CreatePixmap(400, 200, "white"), 0, 0, 200, 200, "black")', True);
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 3) = FillRect(CreatePixmap(600, 300, "white"), 0, 0, 300, 300, "black")', True);
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 4) = FillRect(CreatePixmap(800, 400, "white"), 0, 0, 400, 400, "black")', True);
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 10) = FillRect(CreatePixmap(2000, 1000, "white"), 0, 0, 1000, 1000, "black")', True);
  Test('scale(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 50) = FillRect(CreatePixmap(10000, 5000, "white"), 0, 0, 5000, 5000, "black")', True);

  Test('ColorCount(DebugObject("Sally")) = ColorCount(scale(DebugObject("Sally"), 2))', True);
  Test('ColorCount(DebugObject("Sally")) = ColorCount(scale(DebugObject("Sally"), 10))', True);

  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/1000)', failure, 'Cannot create empty pixmap.');
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/100) = FillRect(CreatePixmap(2, 1, "white"), 0, 0, 1, 1, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/10) = FillRect(CreatePixmap(20, 10, "white"), 0, 0, 10, 10, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/4) = FillRect(CreatePixmap(50, 25, "white"), 0, 0, 25, 25, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/2) = FillRect(CreatePixmap(100, 50, "white"), 0, 0, 50, 50, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1) = FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 2) = FillRect(CreatePixmap(400, 200, "white"), 0, 0, 200, 200, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 3) = FillRect(CreatePixmap(600, 300, "white"), 0, 0, 300, 300, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 4) = FillRect(CreatePixmap(800, 400, "white"), 0, 0, 400, 400, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 10) = FillRect(CreatePixmap(2000, 1000, "white"), 0, 0, 1000, 1000, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 50) = FillRect(CreatePixmap(10000, 5000, "white"), 0, 0, 5000, 5000, "black")', True);

  Test('ColorCount(DebugObject("Sally")) = ColorCount(stretch(DebugObject("Sally"), 2))', True);
  Test('ColorCount(DebugObject("Sally")) = ColorCount(stretch(DebugObject("Sally"), 10))', True);

  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1, 2) = FillRect(CreatePixmap(200, 200, "white"), 0, 0, 100, 200, "black")', True);
  Test('stretch(FillRect(CreatePixmap(200, 100, "white"), 0, 0, 100, 100, "black"), 1/2, 1) = FillRect(CreatePixmap(100, 100, "white"), 0, 0, 50, 100, "black")', True);

  Test('ColorCount(DebugObject("Sally")) = ColorCount(stretch(DebugObject("Sally"), 1, 2))', True);
  Test('ColorCount(DebugObject("Sally")) = ColorCount(stretch(DebugObject("Sally"), 5, 10))', True);

  Test('a ≔ DebugObject("Sally");', null);
  Test('rotate(a, 0) = a', True);
  Test('rotate(a, 2⋅π) = a', True);
  Test('rotate(a, −2⋅π) = a', True);
  Test('rotate(a, 4⋅π) = a', True);
  Test('rotate(a, −4⋅π) = a', True);
  Test('rotate(a, π/2) = rot90p(a)', True);
  Test('rotate(a, π) = rot180(a)', True);
  Test('rotate(a, −π/2) = rot90n(a)', True);

  Test('rotate(a, 0, 1)', failure, 'Too many arguments.');

  Test('sq ≔ CreatePixmap(1000, 1000, "black");', null);
  Test('size(rotate(sq, π/4))', ASOSize(1414, 1414));
  Test('norm(RealVector(GetRGB(AverageColor(rotate(sq, π/4)))) − ❨0.5, 0.5, 0.5❩) < 0.001', True);
  Test('size(AutoCrop(rotate(rotate(sq, π/4), π/4)))', ASOSize(1000, 1000));
  Test('norm(RealVector(GetRGB(AverageColor(AutoCrop(rotate(rotate(sq, π/4), π/4)))))) < 0.001', True);

  Test('norm(RealVector(GetRGB(AverageColor(rotate(CreatePixmap(1000, 1000, "black"), π/4)))) − ❨0.5, 0.5, 0.5❩) < 0.001', True);
  Test('norm(RealVector(GetRGB(AverageColor(rotate(CreatePixmap(1000, 1000, "black", "red"), π/4)))) − ❨0.5, 0, 0❩) < 0.001', True);

//  Nice-to-have properties:
//  Test('a ≔ rotate(CreatePixmap(100, 100, "black"), π/4); b ≔ FloodFill(DrawPolygon(CreatePixmap(141, 141, "white"), ''(❨140/2, 0❩, ❨0, 140/2❩, ❨140/2, 140❩, ❨140, 140/2❩), "black"), 70, 70, "black");', null);
//  Test('a = b', True);
//  Test('sq ≔ CreatePixmap(100, 100, "black"); AutoCrop(rotate(rotate(sq, π/4), π/4)) = sq', True);

  Test('shear(a, 0) = a', True);
  Test('shear(sq, 0) = sq', True);
  Test('b ≔ shear(sq, π/4);', null);
  Test('size(b)', ASOSize(1000, 2000));
  Test('AverageColor(b)', ASO(TRGB.Create(0.5, 0.5, 0.5)));
  Test('AutoCrop(shear(b, −π/4)) = sq', True);

  Test('norm(RealVector(GetRGB(AverageColor(shear(CreatePixmap(1000, 1000, "black"), π/4)))) − ❨0.5, 0.5, 0.5❩) < 0.001', True);
  Test('norm(RealVector(GetRGB(AverageColor(shear(CreatePixmap(1000, 1000, "black", "red"), π/4)))) − ❨0.5, 0, 0❩) < 0.001', True);

  Test('shear(sq, π/2)', failure, 'Skew angle must be between −π/2 and π/2.');

  Test('LinearTransformation(a, IdentityMatrix(2)) = a', True);
  Test('LinearTransformation(a, IdentityMatrix(3))', failure, 'Transformation matrix must be 2×2.');
  Test('LinearTransformation(a, ZeroMatrix(2))', failure, 'Transformation matrix must be invertible.');

  Test('norm(RealVector(GetRGB(AverageColor(LinearTransformation(a, ❨❨−1, 0❩, ❨0, 1❩❩)))) − RealVector(GetRGB(AverageColor(a)))) < 0.01', True);
  Test('norm(RealVector(GetRGB(AverageColor(LinearTransformation(a, ❨❨1, 0❩, ❨0, −1❩❩)))) − RealVector(GetRGB(AverageColor(a)))) < 0.01', True);
  Test('norm(RealVector(GetRGB(AverageColor(LinearTransformation(a, ❨❨−1, 0❩, ❨0, −1❩❩)))) − RealVector(GetRGB(AverageColor(a)))) < 0.01', True);

//  Nice-to-have properties:
//  Test('LinearTransformation(a, ❨❨−1, 0❩, ❨0, 1❩❩) = flip(a, "h")', True);
//  Test('LinearTransformation(a, ❨❨1, 0❩, ❨0, −1❩❩) = flip(a, "v")', True);
//  Test('LinearTransformation(a, ❨❨−1, 0❩, ❨0, −1❩❩) = rot180(a)', True);

  //
  // Scanline effects
  //

  Test('a ≔ DebugObject("Sally");', null);

  Test('ScanlineRotation(a, 0) = a', True);
  Test('ScanlineRotation(a, width(a)) = a', True);
  Test('ScanlineRotation(a, 2⋅width(a)) = a', True);
  Test('ScanlineRotation(a, 3⋅width(a)) = a', True);
  Test('ScanlineRotation(a, 1000⋅width(a)) = a', True);
  Test('ScanlineRotation(a, −width(a)) = a', True);
  Test('ScanlineRotation(a, −2⋅width(a)) = a', True);
  Test('ScanlineRotation(a, −3⋅width(a)) = a', True);
  Test('ScanlineRotation(a, −1000⋅width(a)) = a', True);

  Test('ScanlineRotation(a, 100) ≠ a', True);
  Test('AverageColor(ScanlineRotation(a, 100)) = AverageColor(a)', True);
  Test('ScanlineRotation(ScanlineRotation(a, 1), −1) = a', True);
  Test('ScanlineRotation(ScanlineRotation(a, 394), −394) = a', True);
  Test('ScanlineRotation(ScanlineRotation(a, 394000), −394000) = a', True);

  Test('ScanlineRotation(a, 100) = ScanlineRotation(a, 100, "h")', True);

  Test('ScanlineRotation(a, 0, "v") = a', True);
  Test('ScanlineRotation(a, height(a), "v") = a', True);
  Test('ScanlineRotation(a, 2⋅height(a), "v") = a', True);
  Test('ScanlineRotation(a, 3⋅height(a), "v") = a', True);
  Test('ScanlineRotation(a, 1000⋅height(a), "v") = a', True);
  Test('ScanlineRotation(a, −height(a), "v") = a', True);
  Test('ScanlineRotation(a, −2⋅height(a), "v") = a', True);
  Test('ScanlineRotation(a, −3⋅height(a), "v") = a', True);
  Test('ScanlineRotation(a, −1000⋅height(a), "v") = a', True);

  Test('ScanlineRotation(a, 100, "v") ≠ a', True);
  Test('AverageColor(ScanlineRotation(a, 100, "v")) = AverageColor(a)', True);
  Test('ScanlineRotation(ScanlineRotation(a, 1, "v"), −1, "v") = a', True);
  Test('ScanlineRotation(ScanlineRotation(a, 394, "v"), −394, "v") = a', True);
  Test('ScanlineRotation(ScanlineRotation(a, 394000, "v"), −394000, "v") = a', True);

  Test('ScanlineRotation(a, 100, "s")', failure, 'Character "s" not allowed as argument. Expected one of "hv".');
  Test('ScanlineRotation(a, 100, "v", 0)', failure, 'Too many arguments.');

  Test('RandomScanlineRotation(a, 0) = a', True);
  Test('RandomScanlineRotation(a, 0, "h") = a', True);
  Test('RandomScanlineRotation(a, 0, "v") = a', True);

  Test('AverageColor(RandomScanlineRotation(a)) = AverageColor(a)', True);
  Test('AverageColor(RandomScanlineRotation(a, 394)) = AverageColor(a)', True);
  Test('AverageColor(RandomScanlineRotation(a, 394, "h")) = AverageColor(a)', True);
  Test('AverageColor(RandomScanlineRotation(a, 394, "v")) = AverageColor(a)', True);
  Test('RandomScanlineRotation(a, -394)', failure, 'A non-negative integer was expected as argument 2, but "-394" was given.');

  Test('RandomScanlineRotation(a, 100, "s")', failure, 'Character "s" not allowed as argument. Expected one of "hv".');
  Test('RandomScanlineRotation(a, 100, "v", 0)', failure, 'Too many arguments.');

  Test('CustomScanlineRotation(a, y, 0) = a', True);
  Test('CustomScanlineRotation(a, y, 10) = ScanlineRotation(a, 10)', True);

  Test('AverageColor(CustomScanlineRotation(a, y, round(−√(height(a)^2 − y^2)))) = AverageColor(a)', True);
  Test('AverageColor(CustomScanlineRotation(a, y, round(−√(height(a)^2 − y^2/0)))) = AverageColor(a)', failure, 'Division by zero');
  Test('AverageColor(CustomScanlineRotation(a, y, −√(height(a)^2 − y^2))) = AverageColor(a)', failure, 'Expression must evaluate to an integer.');

  Test('SkewRotation(a, 0) = a', True);
  Test('SkewRotation(a, 100) = CustomScanlineRotation(a, y, y)', True);
  Test('SkewRotation(a, 200) = CustomScanlineRotation(a, y, 2⋅y)', True);
  Test('SkewRotation(a, 100) = SkewRotation(a, 100, 100)', True);
  Test('SkewRotation(a, 100, 100) = SkewRotation(a, 394, 394)', True);
  Test('SkewRotation(a, 100, 100) = SkewRotation(a, 713, 713, "h")', True);
  Test('SkewRotation(a, 22, 50) = CustomScanlineRotation(a, y, ⌊22⋅y/50⌋)', True);

  Test('SkewRotation(a, 0, 1, "v") = a', True);
  Test('AverageColor(SkewRotation(a, 1, 1, "v")) = AverageColor(a)', True);

  Test('SkewRotation(a, 0, 22) = a', True);
  Test('SkewRotation(a, 22, 0)', failure, 'Division by zero');


  //
  // Colour-related functions
  //

  Test('map ≔ (c1, c2, c3) ↦ FillRect(FillRect(CreatePixmap(100, 100, c1), 25, 0, 25, 100, c2), 50, 0, 50, 500, c3);', null);
  Test('a ≔ map("white", "black", "red");', null);

  Test('InvertColor(a) = map("black", "white", "aqua")', True);
  Test('InvertLightness(a) = map("black", "white", "red")', True);
  Test('InvertValue(a) = map("black", "white", "black")', True);

  Test('InvertColor(InvertColor(DebugObject("Sally"))) = DebugObject("Sally")', True);
  Test('InvertLightness(InvertLightness(DebugObject("Sally"))) = DebugObject("Sally")', True);
  Test('InvertValue(InvertValue(DebugObject("Sally"))) = DebugObject("Sally")', False); // because some parts have full value

  Test('FixHue(a)', failure, 'Too few arguments. A required argument of type real number is missing.');
  Test('FixHue(a, 0) = a', True);
  Test('FixHue(a, 60) = map("white", "black", "yellow")', True);
  Test('FixHue(a, 120) = map("white", "black", "lime")', True);
  Test('FixHue(a, 180) = map("white", "black", "aqua")', True);
  Test('FixHue(a, 240) = map("white", "black", "blue")', True);
  Test('FixHue(a, 300) = map("white", "black", "fuchsia")', True);
  Test('FixHue(a, 360) = a', True);

  Test('FixHue(a, 360 + 60) = map("white", "black", "yellow")', True);
  Test('FixHue(a, 360 + 120) = map("white", "black", "lime")', True);
  Test('FixHue(a, 360 + 180) = map("white", "black", "aqua")', True);
  Test('FixHue(a, 360 + 240) = map("white", "black", "blue")', True);
  Test('FixHue(a, 360 + 300) = map("white", "black", "fuchsia")', True);
  Test('FixHue(a, 360 + 360) = a', True);

  Test('∀(FixHue(DebugObject("Sally"), 120), c ↦ (hsv ≔ GetHSV(c); hsv.hue = 120 ∨ hsv.saturation = 0 ∨ hsv.value = 0))', True);

  Test('ToMonochromatic(a)', failure, 'Too few arguments. A required argument of type real number is missing.');
  Test('ToMonochromatic(a, 0) = map("red", "black", "red")', True);
  Test('ToMonochromatic(a, 60) = map("yellow", "black", "yellow")', True);
  Test('ToMonochromatic(a, 120) = map("lime", "black", "lime")', True);
  Test('ToMonochromatic(a, 180) = map("aqua", "black", "aqua")', True);
  Test('ToMonochromatic(a, 240) = map("blue", "black", "blue")', True);
  Test('ToMonochromatic(a, 300) = map("fuchsia", "black", "fuchsia")', True);
  Test('ToMonochromatic(a, 360) = map("red", "black", "red")', True);

  Test('ToMonochromatic(a, 360 + 60) = map("yellow", "black", "yellow")', True);
  Test('ToMonochromatic(a, 360 + 120) = map("lime", "black", "lime")', True);
  Test('ToMonochromatic(a, 360 + 180) = map("aqua", "black", "aqua")', True);
  Test('ToMonochromatic(a, 360 + 240) = map("blue", "black", "blue")', True);
  Test('ToMonochromatic(a, 360 + 300) = map("fuchsia", "black", "fuchsia")', True);
  Test('ToMonochromatic(a, 360 + 360) = map("red", "black", "red")', True);

  Test('∀(ToMonochromatic(DebugObject("Sally"), 120), c ↦ (hsv ≔ GetHSV(c); hsv.hue = 120 ∧ hsv.saturation = 1 ∨ hsv.value = 0))', True);

  Test('ShiftHue(a)', failure, 'Too few arguments. A required argument of type real number is missing.');
  Test('ShiftHue(a, 0) = a', True);
  Test('ShiftHue(a, 60) = map("white", "black", "yellow")', True);
  Test('ShiftHue(a, 120) = map("white", "black", "lime")', True);
  Test('ShiftHue(a, 180) = map("white", "black", "aqua")', True);
  Test('ShiftHue(a, 240) = map("white", "black", "blue")', True);
  Test('ShiftHue(a, 300) = map("white", "black", "fuchsia")', True);
  Test('ShiftHue(a, 360) = a', True);

  Test('ShiftHue(a, 360 + 60) = map("white", "black", "yellow")', True);
  Test('ShiftHue(a, 360 + 120) = map("white", "black", "lime")', True);
  Test('ShiftHue(a, 360 + 180) = map("white", "black", "aqua")', True);
  Test('ShiftHue(a, 360 + 240) = map("white", "black", "blue")', True);
  Test('ShiftHue(a, 360 + 300) = map("white", "black", "fuchsia")', True);
  Test('ShiftHue(a, 360 + 360) = a', True);

  Test('a ≔ map("white", "black", "yellow");', null);

  Test('ShiftHue(a, 0) = a', True);
  Test('ShiftHue(a, 60) = map("white", "black", "lime")', True);
  Test('ShiftHue(a, 120) = map("white", "black", "aqua")', True);
  Test('ShiftHue(a, 180) = map("white", "black", "blue")', True);
  Test('ShiftHue(a, 240) = map("white", "black", "fuchsia")', True);
  Test('ShiftHue(a, 300) = map("white", "black", "red")', True);
  Test('ShiftHue(a, 360) = a', True);

  Test('ShiftHue(a, 360 + 60) = map("white", "black", "lime")', True);
  Test('ShiftHue(a, 360 + 120) = map("white", "black", "aqua")', True);
  Test('ShiftHue(a, 360 + 180) = map("white", "black", "blue")', True);
  Test('ShiftHue(a, 360 + 240) = map("white", "black", "fuchsia")', True);
  Test('ShiftHue(a, 360 + 300) = map("white", "black", "red")', True);
  Test('ShiftHue(a, 360 + 360) = a', True);

  Test('a ≔ map("white", "black", "red");', null);

  Test('whiten(a) = map("white", "#7F7F7F", "#FF7F7F")', True);
  Test('darken(a) = map("#7F7F7F", "#000000", "#7F0000")', True);

  Test('FadeToColor(a)', failure, 'Too few arguments. A required argument of type color is missing.');

  Test('FadeToColor(a, "black") = map("grey", "black", "maroon")', True);
  Test('FadeToColor(a, "white") = map("white", "grey", "#FF8080")', True);
  Test('FadeToColor(a, "red") = map("#FF8080", "maroon", "red")', True);

  Test('AdjustRGB(a, ❨❨0, 0, 0, 0, 0❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 0, 0, 0❩❩) = map("black", "black", "black")', True);
  Test('AdjustRGB(a, ❨❨1, 0, 0, 0, 0❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 0, 0, 0❩❩) = map("red", "red", "red")', True);
  Test('AdjustRGB(a, ❨❨0, 0, 0, 0, 0❩, ❨1, 0, 0, 0, 0❩, ❨0, 0, 0, 0, 0❩❩) = map("lime", "lime", "lime")', True);
  Test('AdjustRGB(a, ❨❨0, 0, 0, 0, 0❩, ❨0, 0, 0, 0, 0❩, ❨1, 0, 0, 0, 0❩❩) = map("blue", "blue", "blue")', True);
  Test('AdjustRGB(a, ❨❨0, 1, 0, 0, 0❩, ❨0, 0, 1, 0, 0❩, ❨0, 0, 0, 1, 0❩❩) = a', True);
  Test('AdjustRGB(a, ❨❨0, 0, 0, 1, 0❩, ❨0, 0, 1, 0, 0❩, ❨0, 1, 0, 0, 0❩❩) = map("white", "black", "blue")', True);
  Test('AdjustRGB(a, ❨❨0, 1, 0, 0, 0❩, ❨0, 0, 1, 0, 0❩, ❨0, 1, 0, 1, 0❩❩) = map("white", "black", "fuchsia")', True);

  Test('AdjustHSV(DebugObject("Sally"), ❨❨0, 0, 0, 0, 0❩, ❨0, 0, 1, 0, 0❩, ❨0, 0, 0, 1, 0❩❩) = FixHue(DebugObject("Sally"), 0)', True);
  Test('AdjustHSV(DebugObject("Sally"), ❨❨60, 0, 0, 0, 0❩, ❨0, 0, 1, 0, 0❩, ❨0, 0, 0, 1, 0❩❩) = FixHue(DebugObject("Sally"), 60)', True);
  Test('AdjustHSV(DebugObject("Sally"), ❨❨0, 0, 0, 0, 0❩, ❨1, 0, 0, 0, 0❩, ❨0, 0, 0, 1, 0❩❩) = ToMonochromatic(DebugObject("Sally"), 0)', True);
  Test('AdjustHSV(DebugObject("Sally"), ❨❨120, 0, 0, 0, 0❩, ❨1, 0, 0, 0, 0❩, ❨0, 0, 0, 1, 0❩❩) = ToMonochromatic(DebugObject("Sally"), 120)', True);
  Test('AdjustHSV(DebugObject("Sally"), ❨❨0, 1, 0, 0, 0❩, ❨0, 0, 0, 0, 0❩, ❨0, 0, 0, 1, 0❩❩) = ToGreyscale(DebugObject("Sally"))', True);
  Test('AdjustHSV(DebugObject("Sally"), ❨❨120, 1, 0, 0, 1❩, ❨0, 0, 1, 0, 0❩, ❨0, 0, 0, 1, 0❩❩) = ShiftHue(DebugObject("Sally"), 120)', True);
  Test('AdjustHSV(DebugObject("Sally"), ❨❨360, 1, 0, 0, 1❩, ❨0, 0, 1, 0, 0❩, ❨0, 0, 0, 1, 0❩❩) = DebugObject("Sally")', True);
  Test('AdjustHSV(DebugObject("Sally"), ❨❨360, 1, 0, 0, 0❩, ❨0, 0, 1, 0, 0❩, ❨0, 0, 0, 1, 0❩❩) = FixHue(DebugObject("Sally"), 0)', True);

  Test('map ≔ (a, b, c, d, e, f) ↦ FillRect(FillRect(FillRect(FillRect(FillRect(CreatePixmap(600, 100, a), 100, 0, 100, 100, b), 200, 0, 100, 100, c), 300, 0, 100, 100, d), 400, 0, 100, 100, e), 500, 0, 100, 100, f);', null);
  Test('g ≔ v ↦ rgb(v, v, v);', null);
  Test('baseimg ≔ map("red", "green", "blue", "yellow", "white", "black");', null);

  Test('ExtractChannel(baseimg, "rgb", "r") = map("white", "black", "black", "white", "white", "black")', True);
  Test('ExtractChannel(baseimg, "rgb", "g") = map("black", "#808080", "black", "white", "white", "black")', True);
  Test('ExtractChannel(baseimg, "rgb", "b") = map("black", "black", "white", "black", "white", "black")', True);
  Test('ExtractChannel(baseimg, "rgb", "y")', failure, 'Character "y" not allowed as argument. Expected one of "rgb".');

  Test('ExtractChannel(baseimg, "hsv", "h") = map(g(0), g(120/360), g(240/360), g(60/360), g(0), g(0))', True);
  Test('ExtractChannel(baseimg, "hsv", "s") = map(g(1), g(1), g(1), g(1), g(0), g(0))', True);
  Test('ExtractChannel(baseimg, "hsv", "v") = map(g(1), g(0.5), g(1), g(1), g(1), g(0))', True);
  Test('ExtractChannel(baseimg, "hsv", "l")', failure, 'Character "l" not allowed as argument. Expected one of "hsv".');

  Test('ExtractChannel(baseimg, "hsl", "h") = map(g(0), g(120/360), g(240/360), g(60/360), g(0), g(0))', True);
  Test('ExtractChannel(baseimg, "hsl", "s") = map(g(1), g(1), g(1), g(1), g(0), g(0))', True);
  Test('ExtractChannel(baseimg, "hsl", "l") = map(g(0.5), g(0.25), g(0.5), g(0.5), g(1), g(0))', True);
  Test('ExtractChannel(baseimg, "hsl", "v")', failure, 'Character "v" not allowed as argument. Expected one of "hsl".');

  Test('ExtractChannel(baseimg, "hrg", "h")', failure, 'String "hrg" not allowed as argument. Expected one of "rgb", "hsv", "hsl".');

  Test('a ≔ DebugObject("Sally"); r ≔ ExtractChannel(a, "rgb", "r"); g ≔ ExtractChannel(a, "rgb", "g"); b ≔ ExtractChannel(a, "rgb", "b");', null);
  Test('CombineChannels("rgb", r, g, b) = a', True);

  Test('CombineChannels("rgb", r, g, DebugObject("harvestman"))', failure, 'Pixmaps are not of the same size.');
  Test('CombineChannels("rgb", DebugObject("harvestman"), g, DebugObject("harvestman"))', failure, 'Pixmaps are not of the same size.');

  Test('CombineChannels("rgb", b, g, r) = AdjustRGB(a, ❨❨0, 0, 0, 1, 0❩, ❨0, 0, 1, 0, 0❩, ❨0, 1, 0, 0, 0❩❩)', True);

  Test('a ≔ DebugObject("Sally"); h ≔ ExtractChannel(a, "hsv", "h"); s ≔ ExtractChannel(a, "hsv", "s"); v ≔ ExtractChannel(a, "hsv", "v");', null);
  Test('b ≔ CombineChannels("hsv", h, s, v);', null);
  Test('norm(RealVector(GetRGB(AverageColor(a))) − RealVector(GetRGB(AverageColor(b)))) < 0.00001', True);

  Test('CombineChannels("hsg", h, s, g)', failure, 'String "hsg" not allowed as argument. Expected one of "rgb", "hsv".');


  //
  // Blending
  //

  Test('a ≔ FillRect(FillRect(CreatePixmap(10, 10, "black"), 0, 0, 5, 5, "red"), 5, 5, 5, 5, "gold");', null);
  Test('GetRect(a, 0, 0, 5, 5) = CreatePixmap(5, 5, "red")', True);
  Test('GetRect(a, 5, 0, 5, 5) = CreatePixmap(5, 5, "black")', True);
  Test('GetRect(a, 0, 5, 5, 5) = CreatePixmap(5, 5, "black")', True);
  Test('GetRect(a, 5, 5, 5, 5) = CreatePixmap(5, 5, "gold")', True);
  Test('GetRect(a, 4, 4, 2, 2) = FillRect(FillRect(CreatePixmap(2, 2, "black"), 0, 0, 1, 1, "red"), 1, 1, 1, 1, "gold")', True);
  Test('scale(GetRect(a, 4, 4, 2, 2), 5) = a', True);
  Test('GetRect(a, 4, 4, 6, 6) = FillRect(FillRect(CreatePixmap(6, 6, "black"), 0, 0, 1, 1, "red"), 1, 1, 5, 5, "gold")', True);
  Test('GetRect(a, 4, 4, 60, 6) = FillRect(FillRect(CreatePixmap(6, 6, "black"), 0, 0, 1, 1, "red"), 1, 1, 5, 5, "gold")', True);
  Test('GetRect(a, 4, 4, 6, 60) = FillRect(FillRect(CreatePixmap(6, 6, "black"), 0, 0, 1, 1, "red"), 1, 1, 5, 5, "gold")', True);
  Test('GetRect(a, 4, 4, 60, 60) = FillRect(FillRect(CreatePixmap(6, 6, "black"), 0, 0, 1, 1, "red"), 1, 1, 5, 5, "gold")', True);
  Test('GetRect(a, 4, 4, 6000000, 600000000) = FillRect(FillRect(CreatePixmap(6, 6, "black"), 0, 0, 1, 1, "red"), 1, 1, 5, 5, "gold")', True);
  Test('GetRect(a, 10, 0, 3, 3)', failure, 'Rectangle is empty.');
  Test('GetRect(a, 0, 10, 3, 3)', failure, 'Rectangle is empty.');
  Test('GetRect(a, 10, 10, 3, 3)', failure, 'Rectangle is empty.');
  Test('GetRect(a, 10000, 10000, 3, 3)', failure, 'Rectangle is empty.');
  Test('GetRect(a, 9, 9, 3, 3) = CreatePixmap(1, 1, "gold")', True);
  Test('GetRect(a, 9, 9, 0, 3)', failure, 'Rectangle is empty.');
  Test('GetRect(a, 9, 9, 3, 0)', failure, 'Rectangle is empty.');
  Test('GetRect(a, 9, 9, 0, 0)', failure, 'Rectangle is empty.');

  Test('GetRect(a, -1, 0, 5, 5)', failure, 'A non-negative integer was expected as argument 2, but "-1" was given.');
  Test('GetRect(a, 0, -1, 5, 5)', failure, 'A non-negative integer was expected as argument 3, but "-1" was given.');
  Test('GetRect(a, 0, 0, -2, 5)', failure, 'A non-negative integer was expected as argument 4, but "-2" was given.');
  Test('GetRect(a, 0, 0, 5, -1000)', failure, 'A non-negative integer was expected as argument 5, but "-1000" was given.');

  Test('a ≔ CreatePixmap(100, 100, "gold"); b ≔ CreatePixmap(10, 15, "red");', null);
  Test('DrawRect(a, b, 0, 0) = FillRect(a, 0, 0, 10, 15, "red")', True);
  Test('DrawRect(a, b, 20, 40) = FillRect(a, 20, 40, 10, 15, "red")', True);
  Test('DrawRect(a, b, 90, 90) = FillRect(a, 90, 90, 123, 456, "red")', True);
  Test('DrawRect(a, b, 20, 40, 1) = FillRect(a, 20, 40, 10, 15, "red")', True);
  Test('DrawRect(a, b, 20, 40, 0) = a', True);
  Test('DrawRect(a, b, 20, 40, 0.5) = FillRect(a, 20, 40, 10, 15, "#FF6C00")', True);
  Test('(#BlendModes() > 40) ∧ ∀(BlendModes(), x ↦ type(x) = "string")', True);
  Test('ForEach(BlendModes(), bm, DrawRect(DebugObject("harvestman"), DebugObject("rabbit"), 0, 0, 1, bm))', null);
  Test('ForEach(BlendModes(), bm, assert(DrawRect(DebugObject("harvestman"), DebugObject("rabbit"), 10000, 100000, 1, bm) = DebugObject("harvestman")))', null);

  Test('a ≔ CreatePixmap(100, 100, "gold"); b ≔ CreatePixmap(10, 15, "red");', null);
  Test('AlphaDrawRect(a, b, 0, 0) = FillRect(a, 0, 0, 10, 15, "red")', True);
  Test('AlphaDrawRect(a, b, 20, 40) = FillRect(a, 20, 40, 10, 15, "red")', True);
  Test('AlphaDrawRect(a, b, 90, 90) = FillRect(a, 90, 90, 123, 456, "red")', True);
  Test('AlphaDrawRect(a, b, 20, 40, 1) = FillRect(a, 20, 40, 10, 15, "red")', True);
  Test('AlphaDrawRect(a, b, 20, 40, 0) = a', True);
  Test('AlphaDrawRect(a, b, 20, 40, 0.5) = FillRect(a, 20, 40, 10, 15, "#FF6C00")', True);
  Test('(#BlendModes() > 40) ∧ ∀(BlendModes(), x ↦ type(x) = "string")', True);
  Test('ForEach(BlendModes(), bm, AlphaDrawRect(DebugObject("harvestman"), DebugObject("rabbit"), 0, 0, 1, bm))', null);
  Test('ForEach(BlendModes(), bm, assert(AlphaDrawRect(DebugObject("harvestman"), DebugObject("rabbit"), 10000, 100000, 1, bm) = DebugObject("harvestman")))', null);

  Test('n ≔ 4; pm ≔ ComputePixmap(n, n, x, y, hsv(x⋅360/n, 1, succ(y)/n));', null);
  Test('size(pm)', ASOSize(4, 4));

  Test('pm[0, 0]', ASOColor(THSV.Create(360*0/4, 1, 1/4)));
  Test('pm[0, 1]', ASOColor(THSV.Create(360*1/4, 1, 1/4)));
  Test('pm[0, 2]', ASOColor(THSV.Create(360*2/4, 1, 1/4)));
  Test('pm[0, 3]', ASOColor(THSV.Create(360*3/4, 1, 1/4)));

  Test('pm[1, 0]', ASOColor(THSV.Create(360*0/4, 1, 2/4)));
  Test('pm[1, 1]', ASOColor(THSV.Create(360*1/4, 1, 2/4)));
  Test('pm[1, 2]', ASOColor(THSV.Create(360*2/4, 1, 2/4)));
  Test('pm[1, 3]', ASOColor(THSV.Create(360*3/4, 1, 2/4)));

  Test('pm[2, 0]', ASOColor(THSV.Create(360*0/4, 1, 3/4)));
  Test('pm[2, 1]', ASOColor(THSV.Create(360*1/4, 1, 3/4)));
  Test('pm[2, 2]', ASOColor(THSV.Create(360*2/4, 1, 3/4)));
  Test('pm[2, 3]', ASOColor(THSV.Create(360*3/4, 1, 3/4)));

  Test('pm[3, 0]', ASOColor(THSV.Create(360*0/4, 1, 4/4)));
  Test('pm[3, 1]', ASOColor(THSV.Create(360*1/4, 1, 4/4)));
  Test('pm[3, 2]', ASOColor(THSV.Create(360*2/4, 1, 4/4)));
  Test('pm[3, 3]', ASOColor(THSV.Create(360*3/4, 1, 4/4)));

  Test('l ≔ NamedColors(); pm ≔ ComputePixmap(#l, 1, x, y, l[succ(x)].color);', null);
  Test('size(pm)', ASOSize(1, Length(NamedColors)));
  Test('for(x, 0, pred(#l), assert(pm[0, x] = l[succ(x)].color))', null);

  StartStopwatch;
  Test('n ≔ 1000; ComputePixmap(n, n, x, y, hsv(x⋅360/pred(n), 1, y/pred(n)));', null);
  StopStopwatch;
  TestDuration(5.0);

  Test('n ≔ 1000; ComputePixmap(n, n, x, y, hsv(x⋅360/pred(n), y/pred(n)));', failure,
    'Too few arguments. A required argument of type real number is missing.');

  Test('delete(a); delete(b); delete(pm); delete(l); delete(n)', success);

  Test('MatrixPlot(❨❨0, 10, 5, 10❩, ❨1, 5, 10, 5❩❩) = FillRect(FillRect(FillRect(FillRect(FillRect(FillRect(FillRect('+'CreatePixmap(200, 100, "black"), 50, 0, 50, 50, "red"), 100, 0, 50, 50, rgb(0.5, 0, 0)), 150, 0, 50, 50, "red"), 0, 50, 50, 50, rgb(1/10, 0, 0)), 50, 50, 50, 50, rgb(0.5, 0, 0)), 100, 50, 50, 50, "red"), 150, 50, 50, 50, rgb(0.5, 0, 0))', True);
  Test('MatrixPlot(❨❨0, 10, 5, 10❩, ❨1, 5, 10, 5❩❩, "lime") = FillRect(FillRect(FillRect(FillRect(FillRect(FillRect('+'FillRect(CreatePixmap(200, 100, "black"), 50, 0, 50, 50, "lime"), 100, 0, 50, 50, rgb(0, 0.5, 0)), 150, 0, 50, 50, "lime"), 0, 50, 50, 50, rgb(0, 1/10, 0)), 50, 50, 50, 50, rgb(0, 0.5, 0)), 100, 50, 50, 50, "lime"), 150, 50, 50, 50, rgb(0, 0.5, 0))', True);
  Test('MatrixPlot(❨❨0, 10, 5, 10❩, ❨1, 5, 10, 5❩❩, "lime", false) = FillRect(FillRect(FillRect(FillRect(FillRect(FillRect('+'FillRect(CreatePixmap(200, 100, "black"), 50, 0, 50, 50, "lime"), 100, 0, 50, 50, rgb(0, 0.5, 0)), 150, 0, 50, 50, "lime"), 0, 50, 50, 50, rgb(0, 1/10, 0)), 50, 50, 50, 50, rgb(0, 0.5, 0)), 100, 50, 50, 50, "lime"), 150, 50, 50, 50, rgb(0, 0.5, 0))', True);
  Test('MatrixPlot(❨❨0, 10, 5, 10❩, ❨1, 5, 10, 5❩❩, "lime", false, 200) = FillRect(FillRect(FillRect(FillRect(FillRect(FillRect('+'FillRect(CreatePixmap(200, 100, "black"), 50, 0, 50, 50, "lime"), 100, 0, 50, 50, rgb(0, 0.5, 0)), 150, 0, 50, 50, "lime"), 0, 50, 50, 50, rgb(0, 1/10, 0)), 50, 50, 50, 50, rgb(0, 0.5, 0)), 100, 50, 50, 50, "lime"), 150, 50, 50, 50, rgb(0, 0.5, 0))', True);
  Test('MatrixPlot(❨❨0, 10, 5, 10❩, ❨1, 5, 10, 5❩❩, "lime", false, 200, 100) = FillRect(FillRect(FillRect(FillRect(FillRect(FillRect('+'FillRect(CreatePixmap(200, 100, "black"), 50, 0, 50, 50, "lime"), 100, 0, 50, 50, rgb(0, 0.5, 0)), 150, 0, 50, 50, "lime"), 0, 50, 50, 50, rgb(0, 1/10, 0)), 50, 50, 50, 50, rgb(0, 0.5, 0)), 100, 50, 50, 50, "lime"), 150, 50, 50, 50, rgb(0, 0.5, 0))', True);
  Test('MatrixPlot(❨❨0, 10, 5, 10❩, ❨1, 5, 10, 5❩❩, "lime", true, 200, 100) = FillRect(FillRect(FillRect(FillRect(FillRect('+'FillRect(FillRect('+'CreatePixmap(200, 100, rgb(0, 0.5, 0)), 50, 0, 50, 50, "lime"), 100, 0, 50, 50, rgb(0, 0.75, 0)), 150, 0, 50, 50, "lime"), 0, 50, 50, 50, rgb(0, 0.5 + 1/20, 0)), 50, 50, 50, 50, rgb(0, 0.75, 0)), 100, 50, 50, 50, "lime"), 150, 50, 50, 50, rgb(0, 0.75, 0))', True);
  Test('MatrixPlot(❨❨0, 10, 5, 10❩, ❨1, 5, 10, 5❩❩, "lime", true, 20) = FillRect(FillRect(FillRect(FillRect(FillRect('+'FillRect(FillRect(CreatePixmap(20, 10, rgb(0, 0.5, 0)), 5, 0, 5, 5, "lime"), 10, 0, 5, 5, rgb(0, 0.75, 0)), 15, 0, 5, 5, "lime"), 0, 5, 5, 5, rgb(0, 0.5 + 1/20, 0)), 5, 5, 5, 5, rgb(0, 0.75, 0)), 10, 5, 5, 5, "lime"), 15, 5, 5, 5, rgb(0, 0.75, 0))', True);
  Test('MatrixPlot(❨❨0, 10, 5, 10❩, ❨1, 5, 10, 5❩❩, "lime", true, 20, 20) = FillRect(FillRect(FillRect(FillRect(FillRect(FillRect(FillRect('+'CreatePixmap(20, 20, rgb(0, 0.5, 0)), 5, 0, 5, 10, "lime"), 10, 0, 5, 10, rgb(0, 0.75, 0)), 15, 0, 5, 10, "lime"), 0, 10, 5, 10, rgb(0, 0.5 + 1/20, 0)), 5, 10, 5, 10, rgb(0, 0.75, 0)), 10, 10, 5, 10, "lime"), 15, 10, 5, 10, rgb(0, 0.75, 0))', True);

  Test('block ≔ (a, b, c, d) ↦ (img ≔ CreatePixmap(2, 2); img[0, 0] ≔ color(a); img[0, 1] ≔ color(b); img[1, 0] ≔ color(c); img[1, 1] ≔ color(d); img);', null);
  Test('MatrixPlot(❨❨0, 1❩, ❨2, −1❩❩, "red", false, 2) = block(rgb(1/3, 0, 0), rgb(2/3, 0, 0), rgb(1, 0, 0), rgb(0, 0, 0))', True);
  Test('MatrixPlot(❨❨0, 1❩, ❨2, −1❩❩, "red", true, 2) = block(rgb(1/2, 0, 0), rgb(3/4, 0, 0), rgb(1, 0, 0), rgb(1/4, 0, 0))', True);
  Test('MatrixPlot(❨❨0, 1❩, ❨2, −1❩❩, ''("blue", "white", "red"), true, 2) = block("white", rgb(1, .5, .5), "red", rgb(.5, .5, 1))', True);
  Test('MatrixPlot(❨❨0, 1❩, ❨2, −2❩❩, ''("blue", "white", "red"), true, 2) = block("white", rgb(1, .5, .5), "red", "blue")', True);
  Test('MatrixPlot(❨❨0, 1❩, ❨2, −2❩❩, ''("blue", "white", "red"), false, 2) = block("white", rgb(1, .5, .5), "red", "blue")', True);
  Test('MatrixPlot(❨❨0, 1❩, ❨2, −1❩❩, ''("blue", "red"), false, 2) = block(rgb(1/3, 0, 2/3), rgb(2/3, 0, 1/3), "red", "blue")', True);
  Test('MatrixPlot(❨❨0, 1❩, ❨2, −1❩❩, ''("blue", "red"), true, 2) = block(rgb(.5, 0, .5), rgb(.75, 0, .25), "red", rgb(.25, 0, .75))', True);

  Test('MatrixPlot(❨❨0, 1❩, ❨2, −1❩❩, (t ↦ hsv(360⋅t, 1, 1)), true, 2) = block(hsv(180, 1, 1), hsv(270, 1, 1), hsv(0, 1, 1), hsv(90, 1, 1))', True);

  Test('MatrixPlot(❨❨0, 1❩, ❨2, −1❩❩, (t ↦ "cat"), true, 2)', failure, 'Unsupported colour descriptor "cat".');
  Test('MatrixPlot(❨❨0, 1❩, ❨2, −1❩❩, (t ↦ false), true, 2)', failure, 'An object of type "boolean" cannot be converted to a colour.');

  Test('delete(block)', success);





  //
  //
  //  CHAPTER 16
  //  MIDI functions.
  //
  //

  Chapter('MIDI functions');

  for _i := Ord(Low(TMidiInstrument)) to Ord(High(TMidiInstrument)) do
  begin
    Test('k ≔ ' + _i.ToString, _i);
    with GetMIDIInstrumentData(TMidiInstrument(_i)) do
    begin
      Test('InstrumentInfo(k)',
        TAlgosimStructure.CreateWithValue(
          [
            sm('InstrumentFamily', ASO(InstrumentFamily)),
            sm('InstrumentName', ASO(InstrumentName)),
            sm('IsPercussionChannel', ASO(IsPercussionChannel)),
            sm('ProgramNumber', ASOInt(ProgramNumber))
          ]
        )
      );
      Test('InstrumentInfo(InstrumentInfo(k).InstrumentName)',
        TAlgosimStructure.CreateWithValue(
          [
            sm('InstrumentFamily', ASO(InstrumentFamily)),
            sm('InstrumentName', ASO(InstrumentName)),
            sm('IsPercussionChannel', ASO(IsPercussionChannel)),
            sm('ProgramNumber', ASOInt(ProgramNumber))
          ]
        )
      );
      Test('InstrumentInfo(LowerCase(InstrumentInfo(k).InstrumentName))',
        TAlgosimStructure.CreateWithValue(
          [
            sm('InstrumentFamily', ASO(InstrumentFamily)),
            sm('InstrumentName', ASO(InstrumentName)),
            sm('IsPercussionChannel', ASO(False{!})),
            sm('ProgramNumber', ASOInt(_i{!}))
          ]
        )
      );
      if Odd(_i) then
        Test('SetInstrument(k)', success)
      else
        Test('SetInstrument(InstrumentInfo(k).InstrumentName)', success);
      Test('NoteOn(70, 100)', success);
      Sleep(100);
      Test('NoteOff(70, 100)', success);
      Test('GetInstrument()', _i);
      Test('AsSingleLine(GetInstrument())', InstrumentName);
    end;
  end;

  for _i := Ord(Low(TMidiInstrument)) to Ord(High(TMidiInstrument)) do
  begin
    Test('k ≔ ' + _i.ToString, _i);
    with GetMIDIInstrumentData(TMidiInstrument(_i)) do
    begin
      if not Odd(_i) then
        Test('SetInstrument(k)', success)
      else
        Test('SetInstrument(InstrumentInfo(k).InstrumentName)', success);
      Test('NoteOn(50, 100)', success);
      Sleep(100);
      Test('NoteOff(50, 100)', success);
    end;
  end;

  for _i := Ord(Low(TMidiInstrument)) to Ord(High(TMidiInstrument)) do
  begin
    Test('k ≔ ' + _i.ToString, _i);
    with GetMIDIInstrumentData(TMidiInstrument(_i)) do
    begin
      if not Odd(_i) then
        Test('SetInstrument(k)', success)
      else
        Test('SetInstrument(InstrumentInfo(k).InstrumentName)', success);
      Test('NoteOn(90, 100)', success);
      Sleep(100);
      Test('NoteOff(90, 100)', success);
    end;
  end;

  Test('InstrumentInfo("Nargle Chorus")', failure, 'Unknown instrument "Nargle Chorus".');
  Test('InstrumentInfo(' + Pred(Ord(Low(TMidiInstrument))).ToString + ')', failure, 'Invalid function arguments.');
  Test('InstrumentInfo(' + Succ(Ord(High(TMidiInstrument))).ToString + ')', failure, 'Invalid function arguments.');

  Test('SetInstrument("Nargle Chorus")', failure, 'Unknown instrument "Nargle Chorus".');
  Test('SetInstrument(' + Pred(Ord(Low(TMidiInstrument))).ToString + ')', failure, 'Invalid function arguments.');
  Test('SetInstrument(' + Succ(Ord(High(TMidiInstrument))).ToString + ')', failure, 'Invalid function arguments.');

  for _i := Ord(Low(TMidiPercussionKey)) to Ord(High(TMidiPercussionKey)) do
  begin
    Test('k ≔ ' + _i.ToString, _i);
    with GetMIDIPercussionInstrumentData(TMidiPercussionKey(_i)) do
    begin
      Test('PercussionInstrumentInfo(k)',
        TAlgosimStructure.CreateWithValue(
          [
            sm('InstrumentFamily', ASO(InstrumentFamily)),
            sm('InstrumentName', ASO(InstrumentName)),
            sm('IsPercussionChannel', ASO(IsPercussionChannel)),
            sm('NoteNumber', ASOInt(NoteNumber))
          ]
        )
      );
      Test('PercussionInstrumentInfo(PercussionInstrumentInfo(k).InstrumentName)',
        TAlgosimStructure.CreateWithValue(
          [
            sm('InstrumentFamily', ASO(InstrumentFamily)),
            sm('InstrumentName', ASO(InstrumentName)),
            sm('IsPercussionChannel', ASO(IsPercussionChannel)),
            sm('NoteNumber', ASOInt(NoteNumber))
          ]
        )
      );
      Test('PercussionInstrumentInfo(LowerCase(PercussionInstrumentInfo(k).InstrumentName))',
        TAlgosimStructure.CreateWithValue(
          [
            sm('InstrumentFamily', ASO('Percussion channel instrument')),
            sm('InstrumentName', ASO(InstrumentName)),
            sm('IsPercussionChannel', ASO(True{!})),
            sm('NoteNumber', ASOInt(_i{!}))
          ]
        )
      );
      if Odd(_i) then
        Test('PercussionNoteOn(k, 100)', success)
      else
        Test('PercussionNoteOn(PercussionInstrumentInfo(k).InstrumentName, 100)', success);
      Sleep(100);
    end;
  end;

  Test('PercussionInstrumentInfo("Niffler Drum")', failure, 'Unknown percussion instrument "Niffler Drum".');
  Test('PercussionInstrumentInfo(' + Pred(Ord(Low(TMidiPercussionKey))).ToString + ')', failure, 'Invalid function arguments.');
  Test('PercussionInstrumentInfo(' + Succ(Ord(High(TMidiPercussionKey))).ToString + ')', failure, 'Invalid function arguments.');

  Test('PercussionNoteOn("Niffler Drum", 100)', failure, 'Unknown percussion instrument "Niffler Drum".');
  Test('PercussionNoteOn(123456, 100)', failure, 'Invalid function arguments.');

  Test('SetInstrument(0)', success);
  for _i := 0 to 127 do
  begin
    Test('k ≔ ' + (512*_i).ToString, 512*_i);
    Test('SetVolume(k)', success);
    Test('NoteOn(40, 100)', success);
    Test('GetVolume()', 512*_i);
    Sleep(100);
  end;

  Test('SetInstrument(0)', success);
  for _i := 0 to 127 do
  begin
    Test('k ≔ ' + _i.ToString, _i);
    Test('NoteOn(60, k)', success);
    Sleep(100);
  end;





  //
  //
  //  CHAPTER 17
  //  Digits. Number formatting. Annotated output.
  //
  //

  Chapter('Digits and number formatting');

  //
  // Number formatting
  //

  TestNumFmt('0', nfDefault, ['0']);
  TestNumFmt('0', nfDefault, 10, ['0']);
  TestNumFmt('0', nfDefault, 20, ['0']);
  TestNumFmt('0', nfDefault, 100, ['0']);
  TestNumFmt('0', nfDefault, 1000, ['0']);

  TestNumFmt('0', nfFraction, ['0']);
  TestNumFmt('0', nfFraction, 10, ['0']);
  TestNumFmt('0', nfFraction, 20, ['0']);
  TestNumFmt('0', nfFraction, 100, ['0']);
  TestNumFmt('0', nfFraction, 1000, ['0']);

  TestNumFmt('0', nfFixed,     ['0', '0.000000000000']);
  TestNumFmt('0', nfFixed, 5, ['0', '0.00000']);
  TestNumFmt('0', nfFixed, 10, ['0', '0.0000000000']);
  TestNumFmt('0', nfFixed, 12, ['0', '0.000000000000']);
  TestNumFmt('0', nfFixed, 15, ['0', '0.000000000000000']);
  TestNumFmt('0', nfFixed, 50, ['0', '0.00000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0', nfFixed, 100, ['0', '0.0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0', nfExponent,     ['0', '0.00000000000E0']);
  TestNumFmt('0', nfExponent, 5, ['0', '0.0000E0']);
  TestNumFmt('0', nfExponent, 10, ['0', '0.000000000E0']);
  TestNumFmt('0', nfExponent, 12, ['0', '0.00000000000E0']);
  TestNumFmt('0', nfExponent, 15, ['0', '0.00000000000000E0']);
  TestNumFmt('0', nfExponent, 50, ['0', '0.0000000000000000000000000000000000000000000000000E0', '*']);
  TestNumFmt('0', nfExponent, 100, ['0', '0.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E0', '*']);

  // --

  TestNumFmt('5', nfDefault, ['5']);
  TestNumFmt('5', nfDefault, 1, ['5']);
  TestNumFmt('5', nfDefault, 2, ['5']);
  TestNumFmt('5', nfDefault, 10, ['5']);
  TestNumFmt('5', nfDefault, 20, ['5']);
  TestNumFmt('5', nfDefault, 100, ['5']);
  TestNumFmt('5', nfDefault, 1000, ['5']);

  TestNumFmt('5', nfFraction, ['5']);
  TestNumFmt('5', nfFraction, 1, ['5']);
  TestNumFmt('5', nfFraction, 2, ['5']);
  TestNumFmt('5', nfFraction, 10, ['5']);
  TestNumFmt('5', nfFraction, 20, ['5']);
  TestNumFmt('5', nfFraction, 100, ['5']);
  TestNumFmt('5', nfFraction, 1000, ['5']);

  TestNumFmt('5', nfFixed,     ['5', '5.000000000000']);
  TestNumFmt('5', nfFixed, 0, ['5', '5']);
  TestNumFmt('5', nfFixed, 1, ['5', '5.0']);
  TestNumFmt('5', nfFixed, 2, ['5', '5.00']);
  TestNumFmt('5', nfFixed, 5, ['5', '5.00000']);
  TestNumFmt('5', nfFixed, 10, ['5', '5.0000000000']);
  TestNumFmt('5', nfFixed, 12, ['5', '5.000000000000']);
  TestNumFmt('5', nfFixed, 15, ['5', '5.000000000000000']);
  TestNumFmt('5', nfFixed, 50, ['5', '5.00000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('5', nfFixed, 100, ['5', '5.0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('5', nfExponent,     ['5', '5.00000000000E0']);
  TestNumFmt('5', nfExponent, 2, ['5', '5.0E0']);
  TestNumFmt('5', nfExponent, 5, ['5', '5.0000E0']);
  TestNumFmt('5', nfExponent, 10, ['5', '5.000000000E0']);
  TestNumFmt('5', nfExponent, 12, ['5', '5.00000000000E0']);
  TestNumFmt('5', nfExponent, 15, ['5', '5.00000000000000E0']);
  TestNumFmt('5', nfExponent, 50, ['5', '5.0000000000000000000000000000000000000000000000000E0', '*']);
  TestNumFmt('5', nfExponent, 100, ['5', '5.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E0', '*']);

  // --

  TestNumFmt('9', nfDefault, ['9']);
  TestNumFmt('9', nfDefault, 1, ['9']);
  TestNumFmt('9', nfDefault, 2, ['9']);
  TestNumFmt('9', nfDefault, 10, ['9']);
  TestNumFmt('9', nfDefault, 20, ['9']);
  TestNumFmt('9', nfDefault, 100, ['9']);
  TestNumFmt('9', nfDefault, 1000, ['9']);

  TestNumFmt('9', nfFraction, ['9']);
  TestNumFmt('9', nfFraction, 1, ['9']);
  TestNumFmt('9', nfFraction, 2, ['9']);
  TestNumFmt('9', nfFraction, 10, ['9']);
  TestNumFmt('9', nfFraction, 20, ['9']);
  TestNumFmt('9', nfFraction, 100, ['9']);
  TestNumFmt('9', nfFraction, 1000, ['9']);

  TestNumFmt('9', nfFixed,     ['9', '9.000000000000']);
  TestNumFmt('9', nfFixed, 0, ['9', '9']);
  TestNumFmt('9', nfFixed, 1, ['9', '9.0']);
  TestNumFmt('9', nfFixed, 2, ['9', '9.00']);
  TestNumFmt('9', nfFixed, 5, ['9', '9.00000']);
  TestNumFmt('9', nfFixed, 10, ['9', '9.0000000000']);
  TestNumFmt('9', nfFixed, 12, ['9', '9.000000000000']);
  TestNumFmt('9', nfFixed, 15, ['9', '9.000000000000000']);
  TestNumFmt('9', nfFixed, 50, ['9', '9.00000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('9', nfFixed, 100, ['9', '9.0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('9', nfExponent,     ['9', '9.00000000000E0']);
  TestNumFmt('9', nfExponent, 2, ['9', '9.0E0']);
  TestNumFmt('9', nfExponent, 5, ['9', '9.0000E0']);
  TestNumFmt('9', nfExponent, 10, ['9', '9.000000000E0']);
  TestNumFmt('9', nfExponent, 12, ['9', '9.00000000000E0']);
  TestNumFmt('9', nfExponent, 15, ['9', '9.00000000000000E0']);
  TestNumFmt('9', nfExponent, 50, ['9', '9.0000000000000000000000000000000000000000000000000E0', '*']);
  TestNumFmt('9', nfExponent, 100, ['9', '9.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E0', '*']);

  // --

  TestNumFmt('10', nfDefault, ['10']);
  TestNumFmt('10', nfDefault, 2, ['10']);
  TestNumFmt('10', nfDefault, 10, ['10']);
  TestNumFmt('10', nfDefault, 20, ['10']);
  TestNumFmt('10', nfDefault, 100, ['10']);
  TestNumFmt('10', nfDefault, 1000, ['10']);

  TestNumFmt('10', nfFraction, ['10']);
  TestNumFmt('10', nfFraction, 2, ['10']);
  TestNumFmt('10', nfFraction, 10, ['10']);
  TestNumFmt('10', nfFraction, 20, ['10']);
  TestNumFmt('10', nfFraction, 100, ['10']);
  TestNumFmt('10', nfFraction, 1000, ['10']);

  TestNumFmt('10', nfFixed,     ['10', '10.000000000000']);
  TestNumFmt('10', nfFixed, 0, ['10', '10']);
  TestNumFmt('10', nfFixed, 1, ['10', '10.0']);
  TestNumFmt('10', nfFixed, 2, ['10', '10.00']);
  TestNumFmt('10', nfFixed, 5, ['10', '10.00000']);
  TestNumFmt('10', nfFixed, 10, ['10', '10.0000000000']);
  TestNumFmt('10', nfFixed, 12, ['10', '10.000000000000']);
  TestNumFmt('10', nfFixed, 15, ['10', '10.000000000000000']);
  TestNumFmt('10', nfFixed, 50, ['10', '10.00000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('10', nfFixed, 100, ['10', '10.0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('10', nfExponent,     ['10', '1.00000000000E1']);
  TestNumFmt('10', nfExponent, 1, ['10', '1E1', '1.0E1']);
  TestNumFmt('10', nfExponent, 2, ['10', '1.0E1']);
  TestNumFmt('10', nfExponent, 5, ['10', '1.0000E1']);
  TestNumFmt('10', nfExponent, 10, ['10', '1.000000000E1']);
  TestNumFmt('10', nfExponent, 12, ['10', '1.00000000000E1']);
  TestNumFmt('10', nfExponent, 15, ['10', '1.00000000000000E1']);
  TestNumFmt('10', nfExponent, 50, ['10', '1.0000000000000000000000000000000000000000000000000E1', '*']);
  TestNumFmt('10', nfExponent, 100, ['10', '1.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E1', '*']);

  // --

  TestNumFmt('12', nfDefault, ['12']);
  TestNumFmt('12', nfDefault, 2, ['12']);
  TestNumFmt('12', nfDefault, 10, ['12']);
  TestNumFmt('12', nfDefault, 20, ['12']);
  TestNumFmt('12', nfDefault, 100, ['12']);
  TestNumFmt('12', nfDefault, 1000, ['12']);

  TestNumFmt('12', nfFraction, ['12']);
  TestNumFmt('12', nfFraction, 2, ['12']);
  TestNumFmt('12', nfFraction, 10, ['12']);
  TestNumFmt('12', nfFraction, 20, ['12']);
  TestNumFmt('12', nfFraction, 100, ['12']);
  TestNumFmt('12', nfFraction, 1000, ['12']);

  TestNumFmt('12', nfFixed,     ['12', '12.000000000000']);
  TestNumFmt('12', nfFixed, 0, ['12', '12']);
  TestNumFmt('12', nfFixed, 1, ['12', '12.0']);
  TestNumFmt('12', nfFixed, 2, ['12', '12.00']);
  TestNumFmt('12', nfFixed, 5, ['12', '12.00000']);
  TestNumFmt('12', nfFixed, 10, ['12', '12.0000000000']);
  TestNumFmt('12', nfFixed, 12, ['12', '12.000000000000']);
  TestNumFmt('12', nfFixed, 15, ['12', '12.000000000000000']);
  TestNumFmt('12', nfFixed, 50, ['12', '12.00000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('12', nfFixed, 100, ['12', '12.0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('12', nfExponent,     ['12', '1.20000000000E1']);
  TestNumFmt('12', nfExponent, 1, ['12', '1E1', '1.2E1']);
  TestNumFmt('12', nfExponent, 2, ['12', '1.2E1']);
  TestNumFmt('12', nfExponent, 5, ['12', '1.2000E1']);
  TestNumFmt('12', nfExponent, 10, ['12', '1.200000000E1']);
  TestNumFmt('12', nfExponent, 12, ['12', '1.20000000000E1']);
  TestNumFmt('12', nfExponent, 15, ['12', '1.20000000000000E1']);
  TestNumFmt('12', nfExponent, 50, ['12', '1.2000000000000000000000000000000000000000000000000E1', '*']);
  TestNumFmt('12', nfExponent, 100, ['12', '1.200000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E1', '*']);

  // --

  TestNumFmt('18', nfDefault, ['18']);
  TestNumFmt('18', nfDefault, 2, ['18']);
  TestNumFmt('18', nfDefault, 10, ['18']);
  TestNumFmt('18', nfDefault, 20, ['18']);
  TestNumFmt('18', nfDefault, 100, ['18']);
  TestNumFmt('18', nfDefault, 1000, ['18']);

  TestNumFmt('18', nfFraction, ['18']);
  TestNumFmt('18', nfFraction, 2, ['18']);
  TestNumFmt('18', nfFraction, 10, ['18']);
  TestNumFmt('18', nfFraction, 20, ['18']);
  TestNumFmt('18', nfFraction, 100, ['18']);
  TestNumFmt('18', nfFraction, 1000, ['18']);

  TestNumFmt('18', nfFixed,     ['18', '18.000000000000']);
  TestNumFmt('18', nfFixed, 0, ['18', '18']);
  TestNumFmt('18', nfFixed, 1, ['18', '18.0']);
  TestNumFmt('18', nfFixed, 2, ['18', '18.00']);
  TestNumFmt('18', nfFixed, 5, ['18', '18.00000']);
  TestNumFmt('18', nfFixed, 10, ['18', '18.0000000000']);
  TestNumFmt('18', nfFixed, 12, ['18', '18.000000000000']);
  TestNumFmt('18', nfFixed, 15, ['18', '18.000000000000000']);
  TestNumFmt('18', nfFixed, 50, ['18', '18.00000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('18', nfFixed, 100, ['18', '18.0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('18', nfExponent,     ['18', '1.80000000000E1']);
  TestNumFmt('18', nfExponent, 1, ['18', '2E1', '1.8E1']);
  TestNumFmt('18', nfExponent, 2, ['18', '1.8E1']);
  TestNumFmt('18', nfExponent, 5, ['18', '1.8000E1']);
  TestNumFmt('18', nfExponent, 10, ['18', '1.800000000E1']);
  TestNumFmt('18', nfExponent, 12, ['18', '1.80000000000E1']);
  TestNumFmt('18', nfExponent, 15, ['18', '1.80000000000000E1']);
  TestNumFmt('18', nfExponent, 50, ['18', '1.8000000000000000000000000000000000000000000000000E1', '*']);
  TestNumFmt('18', nfExponent, 100, ['18', '1.800000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E1', '*']);

  // --

  TestNumFmt('18.7', nfDefault, ['!', '18.7']);
  TestNumFmt('18.7', nfDefault, 2, ['!', '19']);
  TestNumFmt('18.7', nfDefault, 10, ['!', '18.7']);
  TestNumFmt('18.7', nfDefault, 20, ['!', '18.7']);
  TestNumFmt('18.7', nfDefault, 100, ['!', '18.7']);
  TestNumFmt('18.7', nfDefault, 1000, ['!', '18.7']);

  TestNumFmt('18.7', nfFraction, ['!', '187/10', '18.7']);
  TestNumFmt('18.7', nfFraction, 2, ['!', '187/10', '19']);
  TestNumFmt('18.7', nfFraction, 10, ['!', '187/10', '18.7']);
  TestNumFmt('18.7', nfFraction, 20, ['!', '187/10', '18.7']);
  TestNumFmt('18.7', nfFraction, 100, ['!', '187/10', '18.7']);
  TestNumFmt('18.7', nfFraction, 1000, ['!', '187/10', '18.7']);

  TestNumFmt('18.7', nfFixed,     ['!', '18.700000000000']);
  TestNumFmt('18.7', nfFixed, 0, ['!', '19']);
  TestNumFmt('18.7', nfFixed, 1, ['!', '18.7']);
  TestNumFmt('18.7', nfFixed, 2, ['!', '18.70']);
  TestNumFmt('18.7', nfFixed, 5, ['!', '18.70000']);
  TestNumFmt('18.7', nfFixed, 10, ['!', '18.7000000000']);
  TestNumFmt('18.7', nfFixed, 12, ['!', '18.700000000000']);
  TestNumFmt('18.7', nfFixed, 15, ['!', '18.700000000000000']);
  TestNumFmt('18.7', nfFixed, 50, ['!', '18.70000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('18.7', nfFixed, 100, ['!', '18.7000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('18.7', nfExponent,     ['!', '1.87000000000E1']);
  TestNumFmt('18.7', nfExponent, 1, ['!', '2E1', '1.9E1']);
  TestNumFmt('18.7', nfExponent, 2, ['!', '1.9E1']);
  TestNumFmt('18.7', nfExponent, 3, ['!', '1.87E1']);
  TestNumFmt('18.7', nfExponent, 5, ['!', '1.8700E1']);
  TestNumFmt('18.7', nfExponent, 10, ['!', '1.870000000E1']);
  TestNumFmt('18.7', nfExponent, 12, ['!', '1.87000000000E1']);
  TestNumFmt('18.7', nfExponent, 15, ['!', '1.87000000000000E1']);
  TestNumFmt('18.7', nfExponent, 50, ['!', '1.8700000000000000000000000000000000000000000000000E1', '*']);
  TestNumFmt('18.7', nfExponent, 100, ['!', '1.870000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E1', '*']);

  // --

  TestNumFmt('18.4', nfDefault, ['!', '18.4']);
  TestNumFmt('18.4', nfDefault, 2, ['!', '18']);
  TestNumFmt('18.4', nfDefault, 10, ['!', '18.4']);
  TestNumFmt('18.4', nfDefault, 20, ['!', '18.4']);
  TestNumFmt('18.4', nfDefault, 100, ['!', '18.4']);
  TestNumFmt('18.4', nfDefault, 1000, ['!', '18.4']);

  TestNumFmt('18.4', nfFraction, ['!', '92/5', '18.4']);
  TestNumFmt('18.4', nfFraction, 2, ['!', '92/5', '18']);
  TestNumFmt('18.4', nfFraction, 10, ['!', '92/5', '18.4']);
  TestNumFmt('18.4', nfFraction, 20, ['!', '92/5', '18.4']);
  TestNumFmt('18.4', nfFraction, 100, ['!', '92/5', '18.4']);
  TestNumFmt('18.4', nfFraction, 1000, ['!', '92/5', '18.4']);

  TestNumFmt('18.4', nfFixed,     ['!', '18.400000000000']);
  TestNumFmt('18.4', nfFixed, 0, ['!', '18']);
  TestNumFmt('18.4', nfFixed, 1, ['!', '18.4']);
  TestNumFmt('18.4', nfFixed, 2, ['!', '18.40']);
  TestNumFmt('18.4', nfFixed, 5, ['!', '18.40000']);
  TestNumFmt('18.4', nfFixed, 10, ['!', '18.4000000000']);
  TestNumFmt('18.4', nfFixed, 12, ['!', '18.400000000000']);
  TestNumFmt('18.4', nfFixed, 15, ['!', '18.400000000000000']);
  TestNumFmt('18.4', nfFixed, 50, ['!', '18.40000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('18.4', nfFixed, 100, ['!', '18.4000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('18.4', nfExponent,     ['!', '1.84000000000E1']);
  TestNumFmt('18.4', nfExponent, 1, ['!', '2E1', '1.8E1']);
  TestNumFmt('18.4', nfExponent, 2, ['!', '1.8E1']);
  TestNumFmt('18.4', nfExponent, 3, ['!', '1.84E1']);
  TestNumFmt('18.4', nfExponent, 5, ['!', '1.8400E1']);
  TestNumFmt('18.4', nfExponent, 10, ['!', '1.840000000E1']);
  TestNumFmt('18.4', nfExponent, 12, ['!', '1.84000000000E1']);
  TestNumFmt('18.4', nfExponent, 15, ['!', '1.84000000000000E1']);
  TestNumFmt('18.4', nfExponent, 50, ['!', '1.8400000000000000000000000000000000000000000000000E1', '*']);
  TestNumFmt('18.4', nfExponent, 100, ['!', '1.840000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E1', '*']);

  // --

  TestNumFmt('0.5', nfDefault, ['!', '0.5']);
  TestNumFmt('0.5', nfDefault, 10, ['!', '0.5']);
  TestNumFmt('0.5', nfDefault, 20, ['!', '0.5']);
  TestNumFmt('0.5', nfDefault, 100, ['!', '0.5']);
  TestNumFmt('0.5', nfDefault, 1000, ['!', '0.5']);

  TestNumFmt('0.5', nfFraction, ['!', '1/2', '0.5']);
  TestNumFmt('0.5', nfFraction, 10, ['!', '1/2', '0.5']);
  TestNumFmt('0.5', nfFraction, 20, ['!', '1/2', '0.5']);
  TestNumFmt('0.5', nfFraction, 100, ['!', '1/2', '0.5']);
  TestNumFmt('0.5', nfFraction, 1000, ['!', '1/2', '0.5']);

  TestNumFmt('0.5', nfFixed,     ['!', '0.500000000000']);
  TestNumFmt('0.5', nfFixed, 5, ['!', '0.50000']);
  TestNumFmt('0.5', nfFixed, 10, ['!', '0.5000000000']);
  TestNumFmt('0.5', nfFixed, 12, ['!', '0.500000000000']);
  TestNumFmt('0.5', nfFixed, 15, ['!', '0.500000000000000']);
  TestNumFmt('0.5', nfFixed, 50, ['!', '0.50000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.5', nfFixed, 100, ['!', '0.5000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.5', nfExponent,     ['!', '5.00000000000E-1']);
  TestNumFmt('0.5', nfExponent, 5, ['!', '5.0000E-1']);
  TestNumFmt('0.5', nfExponent, 10, ['!', '5.000000000E-1']);
  TestNumFmt('0.5', nfExponent, 12, ['!', '5.00000000000E-1']);
  TestNumFmt('0.5', nfExponent, 15, ['!', '5.00000000000000E-1']);
  TestNumFmt('0.5', nfExponent, 50, ['!', '5.0000000000000000000000000000000000000000000000000E-1', '*']);
  TestNumFmt('0.5', nfExponent, 100, ['!', '5.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-1', '*']);

  // --

  TestNumFmt('0.1', nfDefault, ['!', '0.1']);
  TestNumFmt('0.1', nfDefault, 10, ['!', '0.1']);
  TestNumFmt('0.1', nfDefault, 20, ['!', '0.1']);
  TestNumFmt('0.1', nfDefault, 100, ['!', '0.1']);
  TestNumFmt('0.1', nfDefault, 1000, ['!', '0.1']);

  TestNumFmt('0.1', nfFraction, ['!', '1/10', '0.1']);
  TestNumFmt('0.1', nfFraction, 10, ['!', '1/10', '0.1']);
  TestNumFmt('0.1', nfFraction, 20, ['!', '1/10', '0.1']);
  TestNumFmt('0.1', nfFraction, 100, ['!', '1/10', '0.1']);
  TestNumFmt('0.1', nfFraction, 1000, ['!', '1/10', '0.1']);

  TestNumFmt('0.1', nfFixed,     ['!', '0.100000000000']);
  TestNumFmt('0.1', nfFixed, 5, ['!', '0.10000']);
  TestNumFmt('0.1', nfFixed, 10, ['!', '0.1000000000']);
  TestNumFmt('0.1', nfFixed, 12, ['!', '0.100000000000']);
  TestNumFmt('0.1', nfFixed, 15, ['!', '0.100000000000000']);
  TestNumFmt('0.1', nfFixed, 50, ['!', '0.10000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.1', nfFixed, 100, ['!', '0.1000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.1', nfExponent,     ['!', '1.00000000000E-1']);
  TestNumFmt('0.1', nfExponent, 5, ['!', '1.0000E-1']);
  TestNumFmt('0.1', nfExponent, 10, ['!', '1.000000000E-1']);
  TestNumFmt('0.1', nfExponent, 12, ['!', '1.00000000000E-1']);
  TestNumFmt('0.1', nfExponent, 15, ['!', '1.00000000000000E-1']);
  TestNumFmt('0.1', nfExponent, 50, ['!', '1.0000000000000000000000000000000000000000000000000E-1', '*']);
  TestNumFmt('0.1', nfExponent, 100, ['!', '1.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-1', '*']);

  // --

  TestNumFmt('0.2', nfDefault, ['!', '0.2']);
  TestNumFmt('0.2', nfDefault, 10, ['!', '0.2']);
  TestNumFmt('0.2', nfDefault, 20, ['!', '0.2']);
  TestNumFmt('0.2', nfDefault, 100, ['!', '0.2']);
  TestNumFmt('0.2', nfDefault, 1000, ['!', '0.2']);

  TestNumFmt('0.2', nfFraction, ['!', '1/5', '0.2']);
  TestNumFmt('0.2', nfFraction, 10, ['!', '1/5', '0.2']);
  TestNumFmt('0.2', nfFraction, 20, ['!', '1/5', '0.2']);
  TestNumFmt('0.2', nfFraction, 100, ['!', '1/5', '0.2']);
  TestNumFmt('0.2', nfFraction, 1000, ['!', '1/5', '0.2']);

  TestNumFmt('0.2', nfFixed,     ['!', '0.200000000000']);
  TestNumFmt('0.2', nfFixed, 5, ['!', '0.20000']);
  TestNumFmt('0.2', nfFixed, 10, ['!', '0.2000000000']);
  TestNumFmt('0.2', nfFixed, 12, ['!', '0.200000000000']);
  TestNumFmt('0.2', nfFixed, 15, ['!', '0.200000000000000']);
  TestNumFmt('0.2', nfFixed, 50, ['!', '0.20000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.2', nfFixed, 100, ['!', '0.2000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.2', nfExponent,     ['!', '2.00000000000E-1']);
  TestNumFmt('0.2', nfExponent, 5, ['!', '2.0000E-1']);
  TestNumFmt('0.2', nfExponent, 10, ['!', '2.000000000E-1']);
  TestNumFmt('0.2', nfExponent, 12, ['!', '2.00000000000E-1']);
  TestNumFmt('0.2', nfExponent, 15, ['!', '2.00000000000000E-1']);
  TestNumFmt('0.2', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-1', '*']);
  TestNumFmt('0.2', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-1', '*']);

  // --

  TestNumFmt('0.25', nfDefault, ['!', '0.25']);
  TestNumFmt('0.25', nfDefault, 10, ['!', '0.25']);
  TestNumFmt('0.25', nfDefault, 20, ['!', '0.25']);
  TestNumFmt('0.25', nfDefault, 100, ['!', '0.25']);
  TestNumFmt('0.25', nfDefault, 1000, ['!', '0.25']);

  TestNumFmt('0.25', nfFraction, ['!', '1/4', '0.25']);
  TestNumFmt('0.25', nfFraction, 10, ['!', '1/4', '0.25']);
  TestNumFmt('0.25', nfFraction, 20, ['!', '1/4', '0.25']);
  TestNumFmt('0.25', nfFraction, 100, ['!', '1/4', '0.25']);
  TestNumFmt('0.25', nfFraction, 1000, ['!', '1/4', '0.25']);

  TestNumFmt('0.25', nfFixed,     ['!', '0.250000000000']);
  TestNumFmt('0.25', nfFixed, 5, ['!', '0.25000']);
  TestNumFmt('0.25', nfFixed, 10, ['!', '0.2500000000']);
  TestNumFmt('0.25', nfFixed, 12, ['!', '0.250000000000']);
  TestNumFmt('0.25', nfFixed, 15, ['!', '0.250000000000000']);
  TestNumFmt('0.25', nfFixed, 50, ['!', '0.25000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.25', nfFixed, 100, ['!', '0.2500000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.25', nfExponent,     ['!', '2.50000000000E-1']);
  TestNumFmt('0.25', nfExponent, 5, ['!', '2.5000E-1']);
  TestNumFmt('0.25', nfExponent, 10, ['!', '2.500000000E-1']);
  TestNumFmt('0.25', nfExponent, 12, ['!', '2.50000000000E-1']);
  TestNumFmt('0.25', nfExponent, 15, ['!', '2.50000000000000E-1']);
  TestNumFmt('0.25', nfExponent, 50, ['!', '2.5000000000000000000000000000000000000000000000000E-1', '*']);
  TestNumFmt('0.25', nfExponent, 100, ['!', '2.500000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-1', '*']);

  // --

  TestNumFmt('0.02', nfDefault, ['!', '0.02']);
  TestNumFmt('0.02', nfDefault, 10, ['!', '0.02']);
  TestNumFmt('0.02', nfDefault, 20, ['!', '0.02']);
  TestNumFmt('0.02', nfDefault, 100, ['!', '0.02']);
  TestNumFmt('0.02', nfDefault, 1000, ['!', '0.02']);

  TestNumFmt('0.02', nfFraction, ['!', '1/50', '0.02']);
  TestNumFmt('0.02', nfFraction, 10, ['!', '1/50', '0.02']);
  TestNumFmt('0.02', nfFraction, 20, ['!', '1/50', '0.02']);
  TestNumFmt('0.02', nfFraction, 100, ['!', '1/50', '0.02']);
  TestNumFmt('0.02', nfFraction, 1000, ['!', '1/50', '0.02']);

  TestNumFmt('0.02', nfFixed,     ['!', '0.020000000000']);
  TestNumFmt('0.02', nfFixed, 5, ['!', '0.02000']);
  TestNumFmt('0.02', nfFixed, 10, ['!', '0.0200000000']);
  TestNumFmt('0.02', nfFixed, 12, ['!', '0.020000000000']);
  TestNumFmt('0.02', nfFixed, 15, ['!', '0.020000000000000']);
  TestNumFmt('0.02', nfFixed, 50, ['!', '0.02000000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.02', nfFixed, 100, ['!', '0.0200000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.02', nfExponent,     ['!', '2.00000000000E-2']);
  TestNumFmt('0.02', nfExponent, 5, ['!', '2.0000E-2']);
  TestNumFmt('0.02', nfExponent, 10, ['!', '2.000000000E-2']);
  TestNumFmt('0.02', nfExponent, 12, ['!', '2.00000000000E-2']);
  TestNumFmt('0.02', nfExponent, 15, ['!', '2.00000000000000E-2']);
  TestNumFmt('0.02', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-2', '*']);
  TestNumFmt('0.02', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-2', '*']);

  // --

  TestNumFmt('0.002', nfDefault, ['!', '0.002']);
  TestNumFmt('0.002', nfDefault, 10, ['!', '0.002']);
  TestNumFmt('0.002', nfDefault, 20, ['!', '0.002']);
  TestNumFmt('0.002', nfDefault, 100, ['!', '0.002']);
  TestNumFmt('0.002', nfDefault, 1000, ['!', '0.002']);

  TestNumFmt('0.002', nfFraction, ['!', '1/500', '0.002']);
  TestNumFmt('0.002', nfFraction, 10, ['!', '1/500', '0.002']);
  TestNumFmt('0.002', nfFraction, 20, ['!', '1/500', '0.002']);
  TestNumFmt('0.002', nfFraction, 100, ['!', '1/500', '0.002']);
  TestNumFmt('0.002', nfFraction, 1000, ['!', '1/500', '0.002']);

  TestNumFmt('0.002', nfFixed,     ['!', '0.002000000000']);
  TestNumFmt('0.002', nfFixed, 5, ['!', '0.00200']);
  TestNumFmt('0.002', nfFixed, 10, ['!', '0.0020000000']);
  TestNumFmt('0.002', nfFixed, 12, ['!', '0.002000000000']);
  TestNumFmt('0.002', nfFixed, 15, ['!', '0.002000000000000']);
  TestNumFmt('0.002', nfFixed, 50, ['!', '0.00200000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.002', nfFixed, 100, ['!', '0.0020000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.002', nfExponent,     ['!', '2.00000000000E-3']);
  TestNumFmt('0.002', nfExponent, 5, ['!', '2.0000E-3']);
  TestNumFmt('0.002', nfExponent, 10, ['!', '2.000000000E-3']);
  TestNumFmt('0.002', nfExponent, 12, ['!', '2.00000000000E-3']);
  TestNumFmt('0.002', nfExponent, 15, ['!', '2.00000000000000E-3']);
  TestNumFmt('0.002', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-3', '*']);
  TestNumFmt('0.002', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-3', '*']);

  // --

  TestNumFmt('0.0002', nfDefault, ['!', '0.0002']);
  TestNumFmt('0.0002', nfDefault, 10, ['!', '0.0002']);
  TestNumFmt('0.0002', nfDefault, 20, ['!', '0.0002']);
  TestNumFmt('0.0002', nfDefault, 100, ['!', '0.0002']);
  TestNumFmt('0.0002', nfDefault, 1000, ['!', '0.0002']);

  TestNumFmt('0.0002', nfFraction, ['!', '1/5000', '0.0002']);
  TestNumFmt('0.0002', nfFraction, 10, ['!', '1/5000', '0.0002']);
  TestNumFmt('0.0002', nfFraction, 20, ['!', '1/5000', '0.0002']);
  TestNumFmt('0.0002', nfFraction, 100, ['!', '1/5000', '0.0002']);
  TestNumFmt('0.0002', nfFraction, 1000, ['!', '1/5000', '0.0002']);

  TestNumFmt('0.0002', nfFixed,     ['!', '0.000200000000']);
  TestNumFmt('0.0002', nfFixed, 5, ['!', '0.00020']);
  TestNumFmt('0.0002', nfFixed, 10, ['!', '0.0002000000']);
  TestNumFmt('0.0002', nfFixed, 12, ['!', '0.000200000000']);
  TestNumFmt('0.0002', nfFixed, 15, ['!', '0.000200000000000']);
  TestNumFmt('0.0002', nfFixed, 50, ['!', '0.00020000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.0002', nfFixed, 100, ['!', '0.0002000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.0002', nfExponent,     ['!', '2.00000000000E-4']);
  TestNumFmt('0.0002', nfExponent, 5, ['!', '2.0000E-4']);
  TestNumFmt('0.0002', nfExponent, 10, ['!', '2.000000000E-4']);
  TestNumFmt('0.0002', nfExponent, 12, ['!', '2.00000000000E-4']);
  TestNumFmt('0.0002', nfExponent, 15, ['!', '2.00000000000000E-4']);
  TestNumFmt('0.0002', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-4', '*']);
  TestNumFmt('0.0002', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-4', '*']);

  // --

  TestNumFmt('0.00002', nfDefault, ['!', '2E-5']);
  TestNumFmt('0.00002', nfDefault, 10, ['!', '2E-5']);
  TestNumFmt('0.00002', nfDefault, 20, ['!', '2E-5']);
  TestNumFmt('0.00002', nfDefault, 100, ['!', '2E-5']);
  TestNumFmt('0.00002', nfDefault, 1000, ['!', '2E-5']);

  TestNumFmt('0.00002', nfFraction, ['!', '1/50000', '2E-5']);
  TestNumFmt('0.00002', nfFraction, 10, ['!', '1/50000', '2E-5']);
  TestNumFmt('0.00002', nfFraction, 20, ['!', '1/50000', '2E-5']);
  TestNumFmt('0.00002', nfFraction, 100, ['!', '1/50000', '2E-5']);
  TestNumFmt('0.00002', nfFraction, 1000, ['!', '1/50000', '2E-5']);

  TestNumFmt('0.00002', nfFixed,     ['!', '0.000020000000']);
  TestNumFmt('0.00002', nfFixed, 5, ['!', '0.00002']);
  TestNumFmt('0.00002', nfFixed, 10, ['!', '0.0000200000']);
  TestNumFmt('0.00002', nfFixed, 12, ['!', '0.000020000000']);
  TestNumFmt('0.00002', nfFixed, 15, ['!', '0.000020000000000']);
  TestNumFmt('0.00002', nfFixed, 50, ['!', '0.00002000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.00002', nfFixed, 100, ['!', '0.0000200000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.00002', nfExponent,     ['!', '2.00000000000E-5']);
  TestNumFmt('0.00002', nfExponent, 5, ['!', '2.0000E-5']);
  TestNumFmt('0.00002', nfExponent, 10, ['!', '2.000000000E-5']);
  TestNumFmt('0.00002', nfExponent, 12, ['!', '2.00000000000E-5']);
  TestNumFmt('0.00002', nfExponent, 15, ['!', '2.00000000000000E-5']);
  TestNumFmt('0.00002', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-5', '*']);
  TestNumFmt('0.00002', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-5', '*']);

  // --

  TestNumFmt('0.000002', nfDefault, ['!', '2E-6']);
  TestNumFmt('0.000002', nfDefault, 10, ['!', '2E-6']);
  TestNumFmt('0.000002', nfDefault, 20, ['!', '2E-6']);
  TestNumFmt('0.000002', nfDefault, 100, ['!', '2E-6']);
  TestNumFmt('0.000002', nfDefault, 1000, ['!', '2E-6']);

  TestNumFmt('0.000002', nfFraction, ['!', '1/500000', '2E-6']);
  TestNumFmt('0.000002', nfFraction, 10, ['!', '1/500000', '2E-6']);
  TestNumFmt('0.000002', nfFraction, 20, ['!', '1/500000', '2E-6']);
  TestNumFmt('0.000002', nfFraction, 100, ['!', '1/500000', '2E-6']);
  TestNumFmt('0.000002', nfFraction, 1000, ['!', '1/500000', '2E-6']);

  TestNumFmt('0.000002', nfFixed,     ['!', '0.000002000000']);
  TestNumFmt('0.000002', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('0.000002', nfFixed, 10, ['!', '0.0000020000']);
  TestNumFmt('0.000002', nfFixed, 12, ['!', '0.000002000000']);
  TestNumFmt('0.000002', nfFixed, 15, ['!', '0.000002000000000']);
  TestNumFmt('0.000002', nfFixed, 50, ['!', '0.00000200000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.000002', nfFixed, 100, ['!', '0.0000020000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.000002', nfExponent,     ['!', '2.00000000000E-6']);
  TestNumFmt('0.000002', nfExponent, 5, ['!', '2.0000E-6']);
  TestNumFmt('0.000002', nfExponent, 10, ['!', '2.000000000E-6']);
  TestNumFmt('0.000002', nfExponent, 12, ['!', '2.00000000000E-6']);
  TestNumFmt('0.000002', nfExponent, 15, ['!', '2.00000000000000E-6']);
  TestNumFmt('0.000002', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-6', '*']);
  TestNumFmt('0.000002', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-6', '*']);

  // --

  TestNumFmt('2/10000000', nfDefault, ['!', '2E-7']);
  TestNumFmt('2/10000000', nfDefault, 10, ['!', '2E-7']);
  TestNumFmt('2/10000000', nfDefault, 20, ['!', '2E-7']);
  TestNumFmt('2/10000000', nfDefault, 100, ['!', '2E-7']);
  TestNumFmt('2/10000000', nfDefault, 1000, ['!', '2E-7']);

  TestNumFmt('2/10000000', nfFraction, ['!', '1/5000000', '2E-7']);
  TestNumFmt('2/10000000', nfFraction, 10, ['!', '1/5000000', '2E-7']);
  TestNumFmt('2/10000000', nfFraction, 20, ['!', '1/5000000', '2E-7']);
  TestNumFmt('2/10000000', nfFraction, 100, ['!', '1/5000000', '2E-7']);
  TestNumFmt('2/10000000', nfFraction, 1000, ['!', '1/5000000', '2E-7']);

  TestNumFmt('2/10000000', nfFixed,     ['!', '0.000000200000']);
  TestNumFmt('2/10000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/10000000', nfFixed, 10, ['!', '0.0000002000']);
  TestNumFmt('2/10000000', nfFixed, 12, ['!', '0.000000200000']);
  TestNumFmt('2/10000000', nfFixed, 15, ['!', '0.000000200000000']);
  TestNumFmt('2/10000000', nfFixed, 50, ['!', '0.00000020000000000000000000000000000000000000000000', '*']);
  TestNumFmt('2/10000000', nfFixed, 100, ['!', '0.0000002000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/10000000', nfExponent,     ['!', '2.00000000000E-7']);
  TestNumFmt('2/10000000', nfExponent, 5, ['!', '2.0000E-7']);
  TestNumFmt('2/10000000', nfExponent, 10, ['!', '2.000000000E-7']);
  TestNumFmt('2/10000000', nfExponent, 12, ['!', '2.00000000000E-7']);
  TestNumFmt('2/10000000', nfExponent, 15, ['!', '2.00000000000000E-7']);
  TestNumFmt('2/10000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-7', '*']);
  TestNumFmt('2/10000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-7', '*']);

  // --

  TestNumFmt('2/100000000', nfDefault, ['!', '2E-8']);
  TestNumFmt('2/100000000', nfDefault, 10, ['!', '2E-8']);
  TestNumFmt('2/100000000', nfDefault, 20, ['!', '2E-8']);
  TestNumFmt('2/100000000', nfDefault, 100, ['!', '2E-8']);
  TestNumFmt('2/100000000', nfDefault, 1000, ['!', '2E-8']);

  TestNumFmt('2/100000000', nfFraction, ['!', '1/50000000', '2E-8']);
  TestNumFmt('2/100000000', nfFraction, 10, ['!', '1/50000000', '2E-8']);
  TestNumFmt('2/100000000', nfFraction, 20, ['!', '1/50000000', '2E-8']);
  TestNumFmt('2/100000000', nfFraction, 100, ['!', '1/50000000', '2E-8']);
  TestNumFmt('2/100000000', nfFraction, 1000, ['!', '1/50000000', '2E-8']);

  TestNumFmt('2/100000000', nfFixed,     ['!', '0.000000020000']);
  TestNumFmt('2/100000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/100000000', nfFixed, 10, ['!', '0.0000000200']);
  TestNumFmt('2/100000000', nfFixed, 12, ['!', '0.000000020000']);
  TestNumFmt('2/100000000', nfFixed, 15, ['!', '0.000000020000000']);
  TestNumFmt('2/100000000', nfFixed, 50, ['!', '0.00000002000000000000000000000000000000000000000000', '*']);
  TestNumFmt('2/100000000', nfFixed, 100, ['!', '0.0000000200000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/100000000', nfExponent,     ['!', '2.00000000000E-8']);
  TestNumFmt('2/100000000', nfExponent, 5, ['!', '2.0000E-8']);
  TestNumFmt('2/100000000', nfExponent, 10, ['!', '2.000000000E-8']);
  TestNumFmt('2/100000000', nfExponent, 12, ['!', '2.00000000000E-8']);
  TestNumFmt('2/100000000', nfExponent, 15, ['!', '2.00000000000000E-8']);
  TestNumFmt('2/100000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-8', '*']);
  TestNumFmt('2/100000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-8', '*']);

  // --

  TestNumFmt('2/1000000000', nfDefault, ['!', '2E-9']);
  TestNumFmt('2/1000000000', nfDefault, 10, ['!', '2E-9']);
  TestNumFmt('2/1000000000', nfDefault, 20, ['!', '2E-9']);
  TestNumFmt('2/1000000000', nfDefault, 100, ['!', '2E-9']);
  TestNumFmt('2/1000000000', nfDefault, 1000, ['!', '2E-9']);

  TestNumFmt('2/1000000000', nfFraction, ['!', '1/500000000', '2E-9']);
  TestNumFmt('2/1000000000', nfFraction, 10, ['!', '1/500000000', '2E-9']);
  TestNumFmt('2/1000000000', nfFraction, 20, ['!', '1/500000000', '2E-9']);
  TestNumFmt('2/1000000000', nfFraction, 100, ['!', '1/500000000', '2E-9']);
  TestNumFmt('2/1000000000', nfFraction, 1000, ['!', '1/500000000', '2E-9']);

  TestNumFmt('2/1000000000', nfFixed,     ['!', '0.000000002000']);
  TestNumFmt('2/1000000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/1000000000', nfFixed, 10, ['!', '0.0000000020']);
  TestNumFmt('2/1000000000', nfFixed, 12, ['!', '0.000000002000']);
  TestNumFmt('2/1000000000', nfFixed, 15, ['!', '0.000000002000000']);
  TestNumFmt('2/1000000000', nfFixed, 50, ['!', '0.00000000200000000000000000000000000000000000000000', '*']);
  TestNumFmt('2/1000000000', nfFixed, 100, ['!', '0.0000000020000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/1000000000', nfExponent,     ['!', '2.00000000000E-9']);
  TestNumFmt('2/1000000000', nfExponent, 5, ['!', '2.0000E-9']);
  TestNumFmt('2/1000000000', nfExponent, 10, ['!', '2.000000000E-9']);
  TestNumFmt('2/1000000000', nfExponent, 12, ['!', '2.00000000000E-9']);
  TestNumFmt('2/1000000000', nfExponent, 15, ['!', '2.00000000000000E-9']);
  TestNumFmt('2/1000000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-9', '*']);
  TestNumFmt('2/1000000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-9', '*']);

  // --

  TestNumFmt('2/10000000000', nfDefault, ['!', '2E-10']);
  TestNumFmt('2/10000000000', nfDefault, 10, ['!', '2E-10']);
  TestNumFmt('2/10000000000', nfDefault, 20, ['!', '2E-10']);
  TestNumFmt('2/10000000000', nfDefault, 100, ['!', '2E-10']);
  TestNumFmt('2/10000000000', nfDefault, 1000, ['!', '2E-10']);

  TestNumFmt('2/10000000000', nfFraction, ['!', '1/5000000000', '2E-10']);
  TestNumFmt('2/10000000000', nfFraction, 10, ['!', '1/5000000000', '2E-10']);
  TestNumFmt('2/10000000000', nfFraction, 20, ['!', '1/5000000000', '2E-10']);
  TestNumFmt('2/10000000000', nfFraction, 100, ['!', '1/5000000000', '2E-10']);
  TestNumFmt('2/10000000000', nfFraction, 1000, ['!', '1/5000000000', '2E-10']);

  TestNumFmt('2/10000000000', nfFixed,     ['!', '0.000000000200']);
  TestNumFmt('2/10000000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/10000000000', nfFixed, 10, ['!', '0.0000000002']);
  TestNumFmt('2/10000000000', nfFixed, 12, ['!', '0.000000000200']);
  TestNumFmt('2/10000000000', nfFixed, 15, ['!', '0.000000000200000']);
  TestNumFmt('2/10000000000', nfFixed, 50, ['!', '0.00000000020000000000000000000000000000000000000000', '*']);
  TestNumFmt('2/10000000000', nfFixed, 100, ['!', '0.0000000002000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/10000000000', nfExponent,     ['!', '2.00000000000E-10']);
  TestNumFmt('2/10000000000', nfExponent, 5, ['!', '2.0000E-10']);
  TestNumFmt('2/10000000000', nfExponent, 10, ['!', '2.000000000E-10']);
  TestNumFmt('2/10000000000', nfExponent, 12, ['!', '2.00000000000E-10']);
  TestNumFmt('2/10000000000', nfExponent, 15, ['!', '2.00000000000000E-10']);
  TestNumFmt('2/10000000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-10', '*']);
  TestNumFmt('2/10000000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-10', '*']);

  // --

  TestNumFmt('2/100000000000', nfDefault, ['!', '2E-11']);
  TestNumFmt('2/100000000000', nfDefault, 10, ['!', '2E-11']);
  TestNumFmt('2/100000000000', nfDefault, 20, ['!', '2E-11']);
  TestNumFmt('2/100000000000', nfDefault, 100, ['!', '2E-11']);
  TestNumFmt('2/100000000000', nfDefault, 1000, ['!', '2E-11']);

  TestNumFmt('2/100000000000', nfFraction, ['!', '1/50000000000', '2E-11']);
  TestNumFmt('2/100000000000', nfFraction, 10, ['!', '1/50000000000', '2E-11']);
  TestNumFmt('2/100000000000', nfFraction, 20, ['!', '1/50000000000', '2E-11']);
  TestNumFmt('2/100000000000', nfFraction, 100, ['!', '1/50000000000', '2E-11']);
  TestNumFmt('2/100000000000', nfFraction, 1000, ['!', '1/50000000000', '2E-11']);

  TestNumFmt('2/100000000000', nfFixed,     ['!', '0.000000000020']);
  TestNumFmt('2/100000000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/100000000000', nfFixed, 10, ['!', '0.0000000000']);
  TestNumFmt('2/100000000000', nfFixed, 12, ['!', '0.000000000020']);
  TestNumFmt('2/100000000000', nfFixed, 15, ['!', '0.000000000020000']);
  TestNumFmt('2/100000000000', nfFixed, 50, ['!', '0.00000000002000000000000000000000000000000000000000', '*']);
  TestNumFmt('2/100000000000', nfFixed, 100, ['!', '0.0000000000200000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/100000000000', nfExponent,     ['!', '2.00000000000E-11']);
  TestNumFmt('2/100000000000', nfExponent, 5, ['!', '2.0000E-11']);
  TestNumFmt('2/100000000000', nfExponent, 10, ['!', '2.000000000E-11']);
  TestNumFmt('2/100000000000', nfExponent, 12, ['!', '2.00000000000E-11']);
  TestNumFmt('2/100000000000', nfExponent, 15, ['!', '2.00000000000000E-11']);
  TestNumFmt('2/100000000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-11', '*']);
  TestNumFmt('2/100000000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-11', '*']);

  // --

  TestNumFmt('2/1000000000000', nfDefault, ['!', '2E-12']);
  TestNumFmt('2/1000000000000', nfDefault, 10, ['!', '2E-12']);
  TestNumFmt('2/1000000000000', nfDefault, 20, ['!', '2E-12']);
  TestNumFmt('2/1000000000000', nfDefault, 100, ['!', '2E-12']);
  TestNumFmt('2/1000000000000', nfDefault, 1000, ['!', '2E-12']);

  TestNumFmt('2/1000000000000', nfFraction, ['!', '1/500000000000', '2E-12']);
  TestNumFmt('2/1000000000000', nfFraction, 10, ['!', '1/500000000000', '2E-12']);
  TestNumFmt('2/1000000000000', nfFraction, 20, ['!', '1/500000000000', '2E-12']);
  TestNumFmt('2/1000000000000', nfFraction, 100, ['!', '1/500000000000', '2E-12']);
  TestNumFmt('2/1000000000000', nfFraction, 1000, ['!', '1/500000000000', '2E-12']);

  TestNumFmt('2/1000000000000', nfFixed,     ['!', '0.000000000002']);
  TestNumFmt('2/1000000000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/1000000000000', nfFixed, 10, ['!', '0.0000000000']);
  TestNumFmt('2/1000000000000', nfFixed, 12, ['!', '0.000000000002']);
  TestNumFmt('2/1000000000000', nfFixed, 15, ['!', '0.000000000002000']);
  TestNumFmt('2/1000000000000', nfFixed, 50, ['!', '0.00000000000200000000000000000000000000000000000000', '*']);
  TestNumFmt('2/1000000000000', nfFixed, 100, ['!', '0.0000000000020000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/1000000000000', nfExponent,     ['!', '2.00000000000E-12']);
  TestNumFmt('2/1000000000000', nfExponent, 5, ['!', '2.0000E-12']);
  TestNumFmt('2/1000000000000', nfExponent, 10, ['!', '2.000000000E-12']);
  TestNumFmt('2/1000000000000', nfExponent, 12, ['!', '2.00000000000E-12']);
  TestNumFmt('2/1000000000000', nfExponent, 15, ['!', '2.00000000000000E-12']);
  TestNumFmt('2/1000000000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-12', '*']);
  TestNumFmt('2/1000000000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-12', '*']);

  // --

  TestNumFmt('2/10000000000000', nfDefault, ['!', '2E-13']);
  TestNumFmt('2/10000000000000', nfDefault, 10, ['!', '2E-13']);
  TestNumFmt('2/10000000000000', nfDefault, 20, ['!', '2E-13']);
  TestNumFmt('2/10000000000000', nfDefault, 100, ['!', '2E-13']);
  TestNumFmt('2/10000000000000', nfDefault, 1000, ['!', '2E-13']);

  TestNumFmt('2/10000000000000', nfFraction, ['!', '1/5000000000000', '2E-13']);
  TestNumFmt('2/10000000000000', nfFraction, 10, ['!', '1/5000000000000', '2E-13']);
  TestNumFmt('2/10000000000000', nfFraction, 20, ['!', '1/5000000000000', '2E-13']);
  TestNumFmt('2/10000000000000', nfFraction, 100, ['!', '1/5000000000000', '2E-13']);
  TestNumFmt('2/10000000000000', nfFraction, 1000, ['!', '1/5000000000000', '2E-13']);

  TestNumFmt('2/10000000000000', nfFixed,     ['!', '0.000000000000']);
  TestNumFmt('2/10000000000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/10000000000000', nfFixed, 10, ['!', '0.0000000000']);
  TestNumFmt('2/10000000000000', nfFixed, 12, ['!', '0.000000000000']);
  TestNumFmt('2/10000000000000', nfFixed, 15, ['!', '0.000000000000200']);
  TestNumFmt('2/10000000000000', nfFixed, 50, ['!', '0.00000000000020000000000000000000000000000000000000', '*']);
  TestNumFmt('2/10000000000000', nfFixed, 100, ['!', '0.0000000000002000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/10000000000000', nfExponent,     ['!', '2.00000000000E-13']);
  TestNumFmt('2/10000000000000', nfExponent, 5, ['!', '2.0000E-13']);
  TestNumFmt('2/10000000000000', nfExponent, 10, ['!', '2.000000000E-13']);
  TestNumFmt('2/10000000000000', nfExponent, 12, ['!', '2.00000000000E-13']);
  TestNumFmt('2/10000000000000', nfExponent, 15, ['!', '2.00000000000000E-13']);
  TestNumFmt('2/10000000000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-13', '*']);
  TestNumFmt('2/10000000000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-13', '*']);

  // --

  TestNumFmt('2/100000000000000', nfDefault, ['!', '2E-14']);
  TestNumFmt('2/100000000000000', nfDefault, 10, ['!', '2E-14']);
  TestNumFmt('2/100000000000000', nfDefault, 20, ['!', '2E-14']);
  TestNumFmt('2/100000000000000', nfDefault, 100, ['!', '2E-14']);
  TestNumFmt('2/100000000000000', nfDefault, 1000, ['!', '2E-14']);

  TestNumFmt('2/100000000000000', nfFraction, ['!', '1/50000000000000', '2E-14']);
  TestNumFmt('2/100000000000000', nfFraction, 10, ['!', '1/50000000000000', '2E-14']);
  TestNumFmt('2/100000000000000', nfFraction, 20, ['!', '1/50000000000000', '2E-14']);
  TestNumFmt('2/100000000000000', nfFraction, 100, ['!', '1/50000000000000', '2E-14']);
  TestNumFmt('2/100000000000000', nfFraction, 1000, ['!', '1/50000000000000', '2E-14']);

  TestNumFmt('2/100000000000000', nfFixed,     ['!', '0.000000000000']);
  TestNumFmt('2/100000000000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/100000000000000', nfFixed, 10, ['!', '0.0000000000']);
  TestNumFmt('2/100000000000000', nfFixed, 12, ['!', '0.000000000000']);
  TestNumFmt('2/100000000000000', nfFixed, 15, ['!', '0.000000000000020']);
  TestNumFmt('2/100000000000000', nfFixed, 50, ['!', '0.00000000000002000000000000000000000000000000000000', '*']);
  TestNumFmt('2/100000000000000', nfFixed, 100, ['!', '0.0000000000000200000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/100000000000000', nfExponent,     ['!', '2.00000000000E-14']);
  TestNumFmt('2/100000000000000', nfExponent, 5, ['!', '2.0000E-14']);
  TestNumFmt('2/100000000000000', nfExponent, 10, ['!', '2.000000000E-14']);
  TestNumFmt('2/100000000000000', nfExponent, 12, ['!', '2.00000000000E-14']);
  TestNumFmt('2/100000000000000', nfExponent, 15, ['!', '2.00000000000000E-14']);
  TestNumFmt('2/100000000000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-14', '*']);
  TestNumFmt('2/100000000000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-14', '*']);

  // --

  TestNumFmt('2/1000000000000000', nfDefault, ['!', '2E-15']);
  TestNumFmt('2/1000000000000000', nfDefault, 10, ['!', '2E-15']);
  TestNumFmt('2/1000000000000000', nfDefault, 20, ['!', '2E-15']);
  TestNumFmt('2/1000000000000000', nfDefault, 100, ['!', '2E-15']);
  TestNumFmt('2/1000000000000000', nfDefault, 1000, ['!', '2E-15']);

  TestNumFmt('2/1000000000000000', nfFraction, ['!', '1/500000000000000', '2E-15']);
  TestNumFmt('2/1000000000000000', nfFraction, 10, ['!', '1/500000000000000', '2E-15']);
  TestNumFmt('2/1000000000000000', nfFraction, 20, ['!', '1/500000000000000', '2E-15']);
  TestNumFmt('2/1000000000000000', nfFraction, 100, ['!', '1/500000000000000', '2E-15']);
  TestNumFmt('2/1000000000000000', nfFraction, 1000, ['!', '1/500000000000000', '2E-15']);

  TestNumFmt('2/1000000000000000', nfFixed,     ['!', '0.000000000000']);
  TestNumFmt('2/1000000000000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/1000000000000000', nfFixed, 10, ['!', '0.0000000000']);
  TestNumFmt('2/1000000000000000', nfFixed, 12, ['!', '0.000000000000']);
  TestNumFmt('2/1000000000000000', nfFixed, 15, ['!', '0.000000000000002']);
  TestNumFmt('2/1000000000000000', nfFixed, 50, ['!', '0.00000000000000200000000000000000000000000000000000', '*']);
  TestNumFmt('2/1000000000000000', nfFixed, 100, ['!', '0.0000000000000020000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/1000000000000000', nfExponent,     ['!', '2.00000000000E-15']);
  TestNumFmt('2/1000000000000000', nfExponent, 5, ['!', '2.0000E-15']);
  TestNumFmt('2/1000000000000000', nfExponent, 10, ['!', '2.000000000E-15']);
  TestNumFmt('2/1000000000000000', nfExponent, 12, ['!', '2.00000000000E-15']);
  TestNumFmt('2/1000000000000000', nfExponent, 15, ['!', '2.00000000000000E-15']);
  TestNumFmt('2/1000000000000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-15', '*']);
  TestNumFmt('2/1000000000000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-15', '*']);

  // --

  TestNumFmt('2/10000000000000000', nfDefault, ['!', '2E-16']);
  TestNumFmt('2/10000000000000000', nfDefault, 10, ['!', '2E-16']);
  TestNumFmt('2/10000000000000000', nfDefault, 20, ['!', '2E-16']);
  TestNumFmt('2/10000000000000000', nfDefault, 100, ['!', '2E-16']);
  TestNumFmt('2/10000000000000000', nfDefault, 1000, ['!', '2E-16']);

  TestNumFmt('2/10000000000000000', nfFraction, ['!', '1/5000000000000000', '2E-16']);
  TestNumFmt('2/10000000000000000', nfFraction, 10, ['!', '1/5000000000000000', '2E-16']);
  TestNumFmt('2/10000000000000000', nfFraction, 20, ['!', '1/5000000000000000', '2E-16']);
  TestNumFmt('2/10000000000000000', nfFraction, 100, ['!', '1/5000000000000000', '2E-16']);
  TestNumFmt('2/10000000000000000', nfFraction, 1000, ['!', '1/5000000000000000', '2E-16']);

  TestNumFmt('2/10000000000000000', nfFixed,     ['!', '0.000000000000']);
  TestNumFmt('2/10000000000000000', nfFixed, 5, ['!', '0.00000']);
  TestNumFmt('2/10000000000000000', nfFixed, 10, ['!', '0.0000000000']);
  TestNumFmt('2/10000000000000000', nfFixed, 12, ['!', '0.000000000000']);
  TestNumFmt('2/10000000000000000', nfFixed, 15, ['!', '0.000000000000000']);
  TestNumFmt('2/10000000000000000', nfFixed, 50, ['!', '0.00000000000000020000000000000000000000000000000000', '*']);
  TestNumFmt('2/10000000000000000', nfFixed, 100, ['!', '0.0000000000000002000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('2/10000000000000000', nfExponent,     ['!', '2.00000000000E-16']);
  TestNumFmt('2/10000000000000000', nfExponent, 5, ['!', '2.0000E-16']);
  TestNumFmt('2/10000000000000000', nfExponent, 10, ['!', '2.000000000E-16']);
  TestNumFmt('2/10000000000000000', nfExponent, 12, ['!', '2.00000000000E-16']);
  TestNumFmt('2/10000000000000000', nfExponent, 15, ['!', '2.00000000000000E-16']);
  TestNumFmt('2/10000000000000000', nfExponent, 50, ['!', '2.0000000000000000000000000000000000000000000000000E-16', '*']);
  TestNumFmt('2/10000000000000000', nfExponent, 100, ['!', '2.000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-16', '*']);

  // --

  TestNumFmt('0.025', nfDefault, ['!', '0.025']);
  TestNumFmt('0.025', nfDefault, 10, ['!', '0.025']);
  TestNumFmt('0.025', nfDefault, 20, ['!', '0.025']);
  TestNumFmt('0.025', nfDefault, 100, ['!', '0.025']);
  TestNumFmt('0.025', nfDefault, 1000, ['!', '0.025']);

  TestNumFmt('0.025', nfFraction, ['!', '1/40', '0.025']);
  TestNumFmt('0.025', nfFraction, 10, ['!', '1/40', '0.025']);
  TestNumFmt('0.025', nfFraction, 20, ['!', '1/40', '0.025']);
  TestNumFmt('0.025', nfFraction, 100, ['!', '1/40', '0.025']);
  TestNumFmt('0.025', nfFraction, 1000, ['!', '1/40', '0.025']);

  TestNumFmt('0.025', nfFixed,     ['!', '0.025000000000']);
  TestNumFmt('0.025', nfFixed, 5, ['!', '0.02500']);
  TestNumFmt('0.025', nfFixed, 10, ['!', '0.0250000000']);
  TestNumFmt('0.025', nfFixed, 12, ['!', '0.025000000000']);
  TestNumFmt('0.025', nfFixed, 15, ['!', '0.025000000000000']);
  TestNumFmt('0.025', nfFixed, 50, ['!', '0.02500000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('0.025', nfFixed, 100, ['!', '0.0250000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', '*']);

  TestNumFmt('0.025', nfExponent,     ['!', '2.50000000000E-2']);
  TestNumFmt('0.025', nfExponent, 5, ['!', '2.5000E-2']);
  TestNumFmt('0.025', nfExponent, 10, ['!', '2.500000000E-2']);
  TestNumFmt('0.025', nfExponent, 12, ['!', '2.50000000000E-2']);
  TestNumFmt('0.025', nfExponent, 15, ['!', '2.50000000000000E-2']);
  TestNumFmt('0.025', nfExponent, 50, ['!', '2.5000000000000000000000000000000000000000000000000E-2', '*']);
  TestNumFmt('0.025', nfExponent, 100, ['!', '2.500000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000E-2', '*']);

  // --

  TestNumFmt('1/3', nfDefault, ['!', '0.333333333333']);
  TestNumFmt('1/3', nfDefault, 5, ['!', '0.33333']);
  TestNumFmt('1/3', nfDefault, 10, ['!', '0.3333333333']);
  TestNumFmt('1/3', nfDefault, 20, ['!', '0.33333333333333333333', '*']);
  TestNumFmt('1/3', nfDefault, 100, ['!', '0.3333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333', '*']);
  TestNumFmt('1/3', nfDefault, 1000, ['!', '0.' + StringOfChar('3', 1000), '*']);

  TestNumFmt('1/3', nfFraction, ['!', '1/3', '0.333333333333']);
  TestNumFmt('1/3', nfFraction, 5, ['!', '1/3', '0.33333']);
  TestNumFmt('1/3', nfFraction, 10, ['!', '1/3', '0.3333333333']);
  TestNumFmt('1/3', nfFraction, 20, ['!', '1/3', '*']);
  TestNumFmt('1/3', nfFraction, 100, ['!', '1/3', '*']);
  TestNumFmt('1/3', nfFraction, 1000, ['!', '1/3', '*']);

  TestNumFmt('1/3', nfFixed,     ['!', '0.333333333333']);
  TestNumFmt('1/3', nfFixed, 5, ['!', '0.33333']);
  TestNumFmt('1/3', nfFixed, 10, ['!', '0.3333333333']);
  TestNumFmt('1/3', nfFixed, 12, ['!', '0.333333333333']);
  TestNumFmt('1/3', nfFixed, 15, ['!', '0.333333333333333']);
  TestNumFmt('1/3', nfFixed, 50, ['!', '0.33333333333333333333333333333333333333333333333333', '*']);
  TestNumFmt('1/3', nfFixed, 100, ['!', '0.3333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333', '*']);

  TestNumFmt('1/3', nfExponent,     ['!', '3.33333333333E-1']);
  TestNumFmt('1/3', nfExponent, 5, ['!', '3.3333E-1']);
  TestNumFmt('1/3', nfExponent, 10, ['!', '3.333333333E-1']);
  TestNumFmt('1/3', nfExponent, 12, ['!', '3.33333333333E-1']);
  TestNumFmt('1/3', nfExponent, 15, ['!', '3.33333333333333E-1']);
  TestNumFmt('1/3', nfExponent, 50, ['!', '3.3333333333333333333333333333333333333333333333333E-1', '*']);
  TestNumFmt('1/3', nfExponent, 100, ['!', '3.333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333E-1', '*']);

  // --

  TestNumFmt('2/3', nfDefault, ['!', '0.666666666667']);
  TestNumFmt('2/3', nfDefault, 5, ['!', '0.66667']);
  TestNumFmt('2/3', nfDefault, 10, ['!', '0.6666666667']);
  TestNumFmt('2/3', nfDefault, 20, ['!', '0.66666666666666666667', '*']);
  TestNumFmt('2/3', nfDefault, 100, ['!', '0.6666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666667', '*']);
  TestNumFmt('2/3', nfDefault, 1000, ['!', '0.' + StringOfChar('6', 999) + '7', '*']);

  TestNumFmt('2/3', nfFraction, ['!', '2/3', '0.666666666667']);
  TestNumFmt('2/3', nfFraction, 5, ['!', '2/3', '0.66667']);
  TestNumFmt('2/3', nfFraction, 10, ['!', '2/3', '0.6666666667']);
  TestNumFmt('2/3', nfFraction, 20, ['!', '2/3', '*']);
  TestNumFmt('2/3', nfFraction, 100, ['!', '2/3', '*']);
  TestNumFmt('2/3', nfFraction, 1000, ['!', '2/3', '*']);

  TestNumFmt('2/3', nfFixed,     ['!', '0.666666666667']);
  TestNumFmt('2/3', nfFixed, 5, ['!', '0.66667']);
  TestNumFmt('2/3', nfFixed, 10, ['!', '0.6666666667']);
  TestNumFmt('2/3', nfFixed, 12, ['!', '0.666666666667']);
  TestNumFmt('2/3', nfFixed, 15, ['!', '0.666666666666667']);
  TestNumFmt('2/3', nfFixed, 50, ['!', '0.66666666666666666666666666666666666666666666666667', '*']);
  TestNumFmt('2/3', nfFixed, 100, ['!', '0.6666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666667', '*']);

  TestNumFmt('2/3', nfExponent,     ['!', '6.66666666667E-1']);
  TestNumFmt('2/3', nfExponent, 5, ['!', '6.6667E-1']);
  TestNumFmt('2/3', nfExponent, 10, ['!', '6.666666667E-1']);
  TestNumFmt('2/3', nfExponent, 12, ['!', '6.66666666667E-1']);
  TestNumFmt('2/3', nfExponent, 15, ['!', '6.66666666666667E-1']);
  TestNumFmt('2/3', nfExponent, 50, ['!', '6.6666666666666666666666666666666666666666666666667E-1', '*']);
  TestNumFmt('2/3', nfExponent, 100, ['!', '6.666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666667E-1', '*']);

  // --

  TestNumFmt('100/3', nfDefault, ['!', '33.3333333333']);
  TestNumFmt('100/3', nfDefault, 1, ['!', '3E1', '33']);
  TestNumFmt('100/3', nfDefault, 2, ['!', '33']);
  TestNumFmt('100/3', nfDefault, 3, ['!', '33.3']);
  TestNumFmt('100/3', nfDefault, 4, ['!', '33.33']);
  TestNumFmt('100/3', nfDefault, 5, ['!', '33.333']);
  TestNumFmt('100/3', nfDefault, 10, ['!', '33.33333333']);
  TestNumFmt('100/3', nfDefault, 20, ['!', '33.333333333333333333', '*']);
  TestNumFmt('100/3', nfDefault, 100, ['!', '33.33333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333', '*']);
  TestNumFmt('100/3', nfDefault, 1000, ['!', '33.' + StringOfChar('3', 1000 - 2), '*']);

  TestNumFmt('100/3', nfFraction, ['!', '100/3', '33.3333333333']);
  TestNumFmt('100/3', nfFraction, 1, ['!', '100/3', '33']);
  TestNumFmt('100/3', nfFraction, 2, ['!', '100/3', '33']);
  TestNumFmt('100/3', nfFraction, 3, ['!', '100/3', '33.3']);
  TestNumFmt('100/3', nfFraction, 4, ['!', '100/3', '33.33']);
  TestNumFmt('100/3', nfFraction, 5, ['!', '100/3', '33.333']);
  TestNumFmt('100/3', nfFraction, 10, ['!', '100/3', '33.33333333']);
  TestNumFmt('100/3', nfFraction, 20, ['!', '100/3', '*']);
  TestNumFmt('100/3', nfFraction, 100, ['!', '100/3', '*']);
  TestNumFmt('100/3', nfFraction, 1000, ['!', '100/3', '*']);

  TestNumFmt('100/3', nfFixed,     ['!', '33.333333333333']);
  TestNumFmt('100/3', nfFixed, 0, ['!', '33']);
  TestNumFmt('100/3', nfFixed, 1, ['!', '33.3']);
  TestNumFmt('100/3', nfFixed, 2, ['!', '33.33']);
  TestNumFmt('100/3', nfFixed, 3, ['!', '33.333']);
  TestNumFmt('100/3', nfFixed, 4, ['!', '33.3333']);
  TestNumFmt('100/3', nfFixed, 5, ['!', '33.33333']);
  TestNumFmt('100/3', nfFixed, 10, ['!', '33.3333333333']);
  TestNumFmt('100/3', nfFixed, 12, ['!', '33.333333333333']);
  TestNumFmt('100/3', nfFixed, 15, ['!', '33.333333333333333']);
  TestNumFmt('100/3', nfFixed, 50, ['!', '33.33333333333333333333333333333333333333333333333333', '*']);
  TestNumFmt('100/3', nfFixed, 100, ['!', '33.3333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333', '*']);

  TestNumFmt('100/3', nfExponent,     ['!', '3.33333333333E1']);
  TestNumFmt('100/3', nfExponent, 1, ['!', '3E1', '3.3E1']);
  TestNumFmt('100/3', nfExponent, 2, ['!', '3.3E1']);
  TestNumFmt('100/3', nfExponent, 3, ['!', '3.33E1']);
  TestNumFmt('100/3', nfExponent, 4, ['!', '3.333E1']);
  TestNumFmt('100/3', nfExponent, 5, ['!', '3.3333E1']);
  TestNumFmt('100/3', nfExponent, 10, ['!', '3.333333333E1']);
  TestNumFmt('100/3', nfExponent, 12, ['!', '3.33333333333E1']);
  TestNumFmt('100/3', nfExponent, 15, ['!', '3.33333333333333E1']);
  TestNumFmt('100/3', nfExponent, 50, ['!', '3.3333333333333333333333333333333333333333333333333E1', '*']);
  TestNumFmt('100/3', nfExponent, 100, ['!', '3.333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333E1', '*']);

  // --

  TestNumFmt('200/3', nfDefault, ['!', '66.6666666667']);
  TestNumFmt('200/3', nfDefault, 1, ['!', '7E1', '67']);
  TestNumFmt('200/3', nfDefault, 2, ['!', '67']);
  TestNumFmt('200/3', nfDefault, 3, ['!', '66.7']);
  TestNumFmt('200/3', nfDefault, 4, ['!', '66.67']);
  TestNumFmt('200/3', nfDefault, 5, ['!', '66.667']);
  TestNumFmt('200/3', nfDefault, 10, ['!', '66.66666667']);
  TestNumFmt('200/3', nfDefault, 20, ['!', '66.666666666666666667', '*']);
  TestNumFmt('200/3', nfDefault, 100, ['!', '66.66666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666667', '*']);
  TestNumFmt('200/3', nfDefault, 1000, ['!', '66.' + StringOfChar('6', 1000 - 2 - 1) + '7', '*']);

  TestNumFmt('200/3', nfFraction, ['!', '200/3', '66.6666666667']);
  TestNumFmt('200/3', nfFraction, 1, ['!', '200/3', '67']);
  TestNumFmt('200/3', nfFraction, 2, ['!', '200/3', '67']);
  TestNumFmt('200/3', nfFraction, 3, ['!', '200/3', '66.7']);
  TestNumFmt('200/3', nfFraction, 4, ['!', '200/3', '66.67']);
  TestNumFmt('200/3', nfFraction, 5, ['!', '200/3', '66.667']);
  TestNumFmt('200/3', nfFraction, 10, ['!', '200/3', '66.66666667']);
  TestNumFmt('200/3', nfFraction, 20, ['!', '200/3', '*']);
  TestNumFmt('200/3', nfFraction, 100, ['!', '200/3', '*']);
  TestNumFmt('200/3', nfFraction, 1000, ['!', '200/3', '*']);

  TestNumFmt('200/3', nfFixed,     ['!', '66.666666666667']);
  TestNumFmt('200/3', nfFixed, 0, ['!', '67']);
  TestNumFmt('200/3', nfFixed, 1, ['!', '66.7']);
  TestNumFmt('200/3', nfFixed, 2, ['!', '66.67']);
  TestNumFmt('200/3', nfFixed, 3, ['!', '66.667']);
  TestNumFmt('200/3', nfFixed, 4, ['!', '66.6667']);
  TestNumFmt('200/3', nfFixed, 5, ['!', '66.66667']);
  TestNumFmt('200/3', nfFixed, 10, ['!', '66.6666666667']);
  TestNumFmt('200/3', nfFixed, 12, ['!', '66.666666666667']);
  TestNumFmt('200/3', nfFixed, 15, ['!', '66.666666666666667']);
  TestNumFmt('200/3', nfFixed, 50, ['!', '66.66666666666666666666666666666666666666666666666667', '*']);
  TestNumFmt('200/3', nfFixed, 100, ['!', '66.6666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666667', '*']);

  TestNumFmt('200/3', nfExponent,     ['!', '6.66666666667E1']);
  TestNumFmt('200/3', nfExponent, 1, ['!', '7E1', '6.7E1']);
  TestNumFmt('200/3', nfExponent, 2, ['!', '6.7E1']);
  TestNumFmt('200/3', nfExponent, 3, ['!', '6.67E1']);
  TestNumFmt('200/3', nfExponent, 4, ['!', '6.667E1']);
  TestNumFmt('200/3', nfExponent, 5, ['!', '6.6667E1']);
  TestNumFmt('200/3', nfExponent, 10, ['!', '6.666666667E1']);
  TestNumFmt('200/3', nfExponent, 12, ['!', '6.66666666667E1']);
  TestNumFmt('200/3', nfExponent, 15, ['!', '6.66666666666667E1']);
  TestNumFmt('200/3', nfExponent, 50, ['!', '6.6666666666666666666666666666666666666666666666667E1', '*']);
  TestNumFmt('200/3', nfExponent, 100, ['!', '6.666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666666667E1', '*']);

  // --

  TestNumFmt('5.42832', nfDefault, ['!', '5.42832']);
  TestNumFmt('5.42832', nfDefault, 1, ['!', '5', '5.4']);
  TestNumFmt('5.42832', nfDefault, 2, ['!', '5.4']);
  TestNumFmt('5.42832', nfDefault, 3, ['!', '5.43']);
  TestNumFmt('5.42832', nfDefault, 4, ['!', '5.428']);
  TestNumFmt('5.42832', nfDefault, 5, ['!', '5.4283']);
  TestNumFmt('5.42832', nfDefault, 6, ['!', '5.42832']);
  TestNumFmt('5.42832', nfDefault, 10, ['!', '5.42832']);
  TestNumFmt('5.42832', nfDefault, 12, ['!', '5.42832']);
  TestNumFmt('5.42832', nfDefault, 50, ['!', '5.42832']);
  TestNumFmt('5.42832', nfDefault, 100, ['!', '5.42832']);
  TestNumFmt('5.42832', nfDefault, 1000, ['!', '5.42832']);

  TestNumFmt('5.42832', nfFraction, ['!', '33927/6250', '5.42832']);
  TestNumFmt('5.42832', nfFraction, 1, ['!', '33927/6250', '5.4']);
  TestNumFmt('5.42832', nfFraction, 2, ['!', '33927/6250', '5.4']);
  TestNumFmt('5.42832', nfFraction, 3, ['!', '33927/6250', '5.43']);
  TestNumFmt('5.42832', nfFraction, 4, ['!', '33927/6250', '5.428']);
  TestNumFmt('5.42832', nfFraction, 5, ['!', '33927/6250', '5.4283']);
  TestNumFmt('5.42832', nfFraction, 6, ['!', '33927/6250', '5.42832']);
  TestNumFmt('5.42832', nfFraction, 10, ['!', '33927/6250', '5.42832']);
  TestNumFmt('5.42832', nfFraction, 12, ['!', '33927/6250', '5.42832']);
  TestNumFmt('5.42832', nfFraction, 50, ['!', '33927/6250', '5.42832']);
  TestNumFmt('5.42832', nfFraction, 100, ['!', '33927/6250', '5.42832']);
  TestNumFmt('5.42832', nfFraction, 1000, ['!', '33927/6250', '5.42832']);

  TestNumFmt('5.42832', nfFixed, ['!', '5.428320000000']);
  TestNumFmt('5.42832', nfFixed, 0, ['!', '5']);
  TestNumFmt('5.42832', nfFixed, 1, ['!', '5.4']);
  TestNumFmt('5.42832', nfFixed, 2, ['!', '5.43']);
  TestNumFmt('5.42832', nfFixed, 3, ['!', '5.428']);
  TestNumFmt('5.42832', nfFixed, 4, ['!', '5.4283']);
  TestNumFmt('5.42832', nfFixed, 5, ['!', '5.42832']);
  TestNumFmt('5.42832', nfFixed, 6, ['!', '5.428320']);
  TestNumFmt('5.42832', nfFixed, 10, ['!', '5.4283200000']);
  TestNumFmt('5.42832', nfFixed, 12, ['!', '5.428320000000']);
  TestNumFmt('5.42832', nfFixed, 50, ['!', '5.42832000000000000000000000000000000000000000000000', '*']);
  TestNumFmt('5.42832', nfFixed, 100, ['!', '5.42832' + StringOfChar('0', 95), '*']);
  TestNumFmt('5.42832', nfFixed, 1000, ['!', '5.42832' + StringOfChar('0', 995), '*']);

  TestNumFmt('5.42832', nfExponent, ['!', '5.42832000000E0']);
  TestNumFmt('5.42832', nfExponent, 1, ['!', '5E0', '5.4E0']);
  TestNumFmt('5.42832', nfExponent, 2, ['!', '5.4E0']);
  TestNumFmt('5.42832', nfExponent, 3, ['!', '5.43E0']);
  TestNumFmt('5.42832', nfExponent, 4, ['!', '5.428E0']);
  TestNumFmt('5.42832', nfExponent, 5, ['!', '5.4283E0']);
  TestNumFmt('5.42832', nfExponent, 6, ['!', '5.42832E0']);
  TestNumFmt('5.42832', nfExponent, 10, ['!', '5.428320000E0']);
  TestNumFmt('5.42832', nfExponent, 12, ['!', '5.42832000000E0']);
  TestNumFmt('5.42832', nfExponent, 50, ['!', '5.4283200000000000000000000000000000000000000000000E0', '*']);
  TestNumFmt('5.42832', nfExponent, 100, ['!', '5.42832' + StringOfChar('0', 94) + 'E0', '*']);
  TestNumFmt('5.42832', nfExponent, 1000, ['!', '5.42832' + StringOfChar('0', 994) + 'E0', '*']);

  // --

  TestNumFmt('61.2187621', nfDefault, ['!', '61.2187621']);
  TestNumFmt('61.2187621', nfDefault, 1, ['!', '6E1', '61']);
  TestNumFmt('61.2187621', nfDefault, 2, ['!', '61']);
  TestNumFmt('61.2187621', nfDefault, 3, ['!', '61.2']);
  TestNumFmt('61.2187621', nfDefault, 4, ['!', '61.22']);
  TestNumFmt('61.2187621', nfDefault, 5, ['!', '61.219']);
  TestNumFmt('61.2187621', nfDefault, 6, ['!', '61.2188']);
  TestNumFmt('61.2187621', nfDefault, 7, ['!', '61.21876']);
  TestNumFmt('61.2187621', nfDefault, 8, ['!', '61.218762']);
  TestNumFmt('61.2187621', nfDefault, 9, ['!', '61.2187621']);
  TestNumFmt('61.2187621', nfDefault, 10, ['!', '61.2187621']);

  TestNumFmt('61.2187621', nfFraction, ['!', '612187621/10000000', '61.2187621']);
  TestNumFmt('61.2187621', nfFraction, 1, ['!', '612187621/10000000', '61']);
  TestNumFmt('61.2187621', nfFraction, 2, ['!', '612187621/10000000', '61']);
  TestNumFmt('61.2187621', nfFraction, 3, ['!', '612187621/10000000', '61.2']);
  TestNumFmt('61.2187621', nfFraction, 4, ['!', '612187621/10000000', '61.22']);
  TestNumFmt('61.2187621', nfFraction, 5, ['!', '612187621/10000000', '61.219']);
  TestNumFmt('61.2187621', nfFraction, 6, ['!', '612187621/10000000', '61.2188']);
  TestNumFmt('61.2187621', nfFraction, 7, ['!', '612187621/10000000', '61.21876']);
  TestNumFmt('61.2187621', nfFraction, 8, ['!', '612187621/10000000', '61.218762']);
  TestNumFmt('61.2187621', nfFraction, 9, ['!', '612187621/10000000', '61.2187621']);
  TestNumFmt('61.2187621', nfFraction, 10, ['!', '612187621/10000000', '61.2187621']);

  TestNumFmt('61.2187621', nfFixed, ['!', '61.218762100000']);
  TestNumFmt('61.2187621', nfFixed, 0, ['!', '61']);
  TestNumFmt('61.2187621', nfFixed, 1, ['!', '61.2']);
  TestNumFmt('61.2187621', nfFixed, 2, ['!', '61.22']);
  TestNumFmt('61.2187621', nfFixed, 3, ['!', '61.219']);
  TestNumFmt('61.2187621', nfFixed, 4, ['!', '61.2188']);
  TestNumFmt('61.2187621', nfFixed, 5, ['!', '61.21876']);
  TestNumFmt('61.2187621', nfFixed, 6, ['!', '61.218762']);
  TestNumFmt('61.2187621', nfFixed, 7, ['!', '61.2187621']);
  TestNumFmt('61.2187621', nfFixed, 8, ['!', '61.21876210']);
  TestNumFmt('61.2187621', nfFixed, 9, ['!', '61.218762100']);
  TestNumFmt('61.2187621', nfFixed, 10, ['!', '61.2187621000']);
  TestNumFmt('61.2187621', nfFixed, 12, ['!', '61.218762100000']);
  TestNumFmt('61.2187621', nfFixed, 20, ['!', '61.21876210000000000000', '*']);
  TestNumFmt('61.2187621', nfFixed, 30, ['!', '61.218762100000000000000000000000', '*']);

  TestNumFmt('61.2187621', nfExponent, ['!', '6.12187621000E1']);
  TestNumFmt('61.2187621', nfExponent, 1, ['!', '6E1', '6.1E1']);
  TestNumFmt('61.2187621', nfExponent, 2, ['!', '6.1E1']);
  TestNumFmt('61.2187621', nfExponent, 3, ['!', '6.12E1']);
  TestNumFmt('61.2187621', nfExponent, 4, ['!', '6.122E1']);
  TestNumFmt('61.2187621', nfExponent, 5, ['!', '6.1219E1']);
  TestNumFmt('61.2187621', nfExponent, 6, ['!', '6.12188E1']);
  TestNumFmt('61.2187621', nfExponent, 7, ['!', '6.121876E1']);
  TestNumFmt('61.2187621', nfExponent, 8, ['!', '6.1218762E1']);
  TestNumFmt('61.2187621', nfExponent, 9, ['!', '6.12187621E1']);
  TestNumFmt('61.2187621', nfExponent, 10, ['!', '6.121876210E1']);
  TestNumFmt('61.2187621', nfExponent, 12, ['!', '6.12187621000E1']);
  TestNumFmt('61.2187621', nfExponent, 20, ['!', '6.1218762100000000000E1', '*']);
  TestNumFmt('61.2187621', nfExponent, 30, ['!', '6.12187621000000000000000000000E1', '*']);

  // --

  TestNumFmt('512.69992', nfDefault, ['!', '512.69992']);
  TestNumFmt('512.69992', nfDefault, 1, ['!', '5E2', '5.1E2']);
  TestNumFmt('512.69992', nfDefault, 2, ['!', '5.1E2']);
  TestNumFmt('512.69992', nfDefault, 3, ['!', '513']);
  TestNumFmt('512.69992', nfDefault, 4, ['!', '512.7']);
  TestNumFmt('512.69992', nfDefault, 5, ['!', '512.7']);
  TestNumFmt('512.69992', nfDefault, 6, ['!', '512.7']);
  TestNumFmt('512.69992', nfDefault, 7, ['!', '512.6999']);
  TestNumFmt('512.69992', nfDefault, 8, ['!', '512.69992']);
  TestNumFmt('512.69992', nfDefault, 9, ['!', '512.69992']);
  TestNumFmt('512.69992', nfDefault, 10, ['!', '512.69992']);
  TestNumFmt('512.69992', nfDefault, 12, ['!', '512.69992']);
  TestNumFmt('512.69992', nfDefault, 20, ['!', '512.69992']);
  TestNumFmt('512.69992', nfDefault, 1000, ['!', '512.69992']);

  TestNumFmt('512.69992', nfFraction, ['!', '6408749/12500', '512.69992']);
  TestNumFmt('512.69992', nfFraction, 1, ['!', '6408749/12500', '5.1E2']);
  TestNumFmt('512.69992', nfFraction, 2, ['!', '6408749/12500', '5.1E2']);
  TestNumFmt('512.69992', nfFraction, 3, ['!', '6408749/12500', '513']);
  TestNumFmt('512.69992', nfFraction, 4, ['!', '6408749/12500', '512.7']);
  TestNumFmt('512.69992', nfFraction, 5, ['!', '6408749/12500', '512.7']);
  TestNumFmt('512.69992', nfFraction, 6, ['!', '6408749/12500', '512.7']);
  TestNumFmt('512.69992', nfFraction, 7, ['!', '6408749/12500', '512.6999']);
  TestNumFmt('512.69992', nfFraction, 8, ['!', '6408749/12500', '512.69992']);
  TestNumFmt('512.69992', nfFraction, 9, ['!', '6408749/12500', '512.69992']);
  TestNumFmt('512.69992', nfFraction, 10, ['!', '6408749/12500', '512.69992']);
  TestNumFmt('512.69992', nfFraction, 12, ['!', '6408749/12500', '512.69992']);
  TestNumFmt('512.69992', nfFraction, 20, ['!', '6408749/12500', '512.69992']);
  TestNumFmt('512.69992', nfFraction, 1000, ['!', '6408749/12500', '512.69992']);

  TestNumFmt('512.69992', nfFixed, ['!', '512.699920000000']);
  TestNumFmt('512.69992', nfFixed, 0, ['!', '513']);
  TestNumFmt('512.69992', nfFixed, 1, ['!', '512.7']);
  TestNumFmt('512.69992', nfFixed, 2, ['!', '512.70']);
  TestNumFmt('512.69992', nfFixed, 3, ['!', '512.700']);
  TestNumFmt('512.69992', nfFixed, 4, ['!', '512.6999']);
  TestNumFmt('512.69992', nfFixed, 5, ['!', '512.69992']);
  TestNumFmt('512.69992', nfFixed, 6, ['!', '512.699920']);
  TestNumFmt('512.69992', nfFixed, 7, ['!', '512.6999200']);
  TestNumFmt('512.69992', nfFixed, 8, ['!', '512.69992000']);
  TestNumFmt('512.69992', nfFixed, 9, ['!', '512.699920000']);
  TestNumFmt('512.69992', nfFixed, 10, ['!', '512.6999200000']);
  TestNumFmt('512.69992', nfFixed, 12, ['!', '512.699920000000']);
  TestNumFmt('512.69992', nfFixed, 20, ['!', '512.69992000000000000000', '*']);

  TestNumFmt('512.69992', nfExponent, ['!', '5.12699920000E2']);
  TestNumFmt('512.69992', nfExponent, 1, ['!', '5E2', '5.1E2']);
  TestNumFmt('512.69992', nfExponent, 2, ['!', '5.1E2']);
  TestNumFmt('512.69992', nfExponent, 3, ['!', '5.13E2']);
  TestNumFmt('512.69992', nfExponent, 4, ['!', '5.127E2']);
  TestNumFmt('512.69992', nfExponent, 5, ['!', '5.1270E2']);
  TestNumFmt('512.69992', nfExponent, 6, ['!', '5.12700E2']);
  TestNumFmt('512.69992', nfExponent, 7, ['!', '5.126999E2']);
  TestNumFmt('512.69992', nfExponent, 8, ['!', '5.1269992E2']);
  TestNumFmt('512.69992', nfExponent, 9, ['!', '5.12699920E2']);
  TestNumFmt('512.69992', nfExponent, 10, ['!', '5.126999200E2']);
  TestNumFmt('512.69992', nfExponent, 12, ['!', '5.12699920000E2']);
  TestNumFmt('512.69992', nfExponent, 20, ['!', '5.1269992000000000000E2', '*']);

  // --

  TestNumFmt('9999.9999', nfDefault, ['!', '9999.9999']);
  TestNumFmt('9999.9999', nfDefault, 1, ['!', '1E4']);
  TestNumFmt('9999.9999', nfDefault, 2, ['!', '1E4']);
  TestNumFmt('9999.9999', nfDefault, 3, ['!', '1E4']);
  TestNumFmt('9999.9999', nfDefault, 4, ['!', '1E4']);
  TestNumFmt('9999.9999', nfDefault, 5, ['!', '10000']);
  TestNumFmt('9999.9999', nfDefault, 6, ['!', '10000']);
  TestNumFmt('9999.9999', nfDefault, 7, ['!', '10000']);
  TestNumFmt('9999.9999', nfDefault, 8, ['!', '9999.9999']);
  TestNumFmt('9999.9999', nfDefault, 9, ['!', '9999.9999']);
  TestNumFmt('9999.9999', nfDefault, 10, ['!', '9999.9999']);
  TestNumFmt('9999.9999', nfDefault, 12, ['!', '9999.9999']);
  TestNumFmt('9999.9999', nfDefault, 20, ['!', '9999.9999']);
  TestNumFmt('9999.9999', nfDefault, 100, ['!', '9999.9999']);
  TestNumFmt('9999.9999', nfDefault, 1000, ['!', '9999.9999']);

  TestNumFmt('9999.9999', nfFraction, ['!', '99999999/10000', '9999.9999']);
  TestNumFmt('9999.9999', nfFraction, 1, ['!', '99999999/10000', '1E4']);
  TestNumFmt('9999.9999', nfFraction, 2, ['!', '99999999/10000', '1E4']);
  TestNumFmt('9999.9999', nfFraction, 3, ['!', '99999999/10000', '1E4']);
  TestNumFmt('9999.9999', nfFraction, 4, ['!', '99999999/10000', '1E4']);
  TestNumFmt('9999.9999', nfFraction, 5, ['!', '99999999/10000', '10000']);
  TestNumFmt('9999.9999', nfFraction, 6, ['!', '99999999/10000', '10000']);
  TestNumFmt('9999.9999', nfFraction, 7, ['!', '99999999/10000', '10000']);
  TestNumFmt('9999.9999', nfFraction, 8, ['!', '99999999/10000', '9999.9999']);
  TestNumFmt('9999.9999', nfFraction, 9, ['!', '99999999/10000', '9999.9999']);
  TestNumFmt('9999.9999', nfFraction, 10, ['!', '99999999/10000', '9999.9999']);
  TestNumFmt('9999.9999', nfFraction, 12, ['!', '99999999/10000', '9999.9999']);
  TestNumFmt('9999.9999', nfFraction, 20, ['!', '99999999/10000', '9999.9999']);
  TestNumFmt('9999.9999', nfFraction, 100, ['!', '99999999/10000', '9999.9999']);
  TestNumFmt('9999.9999', nfFraction, 1000, ['!', '99999999/10000', '9999.9999']);

  TestNumFmt('9999.9999', nfFixed, ['!', '9999.999900000000']);
  TestNumFmt('9999.9999', nfFixed, 0, ['!', '10000']);
  TestNumFmt('9999.9999', nfFixed, 1, ['!', '10000.0']);
  TestNumFmt('9999.9999', nfFixed, 2, ['!', '10000.00']);
  TestNumFmt('9999.9999', nfFixed, 3, ['!', '10000.000']);
  TestNumFmt('9999.9999', nfFixed, 4, ['!', '9999.9999']);
  TestNumFmt('9999.9999', nfFixed, 5, ['!', '9999.99990']);
  TestNumFmt('9999.9999', nfFixed, 6, ['!', '9999.999900']);
  TestNumFmt('9999.9999', nfFixed, 7, ['!', '9999.9999000']);
  TestNumFmt('9999.9999', nfFixed, 8, ['!', '9999.99990000']);
  TestNumFmt('9999.9999', nfFixed, 9, ['!', '9999.999900000']);
  TestNumFmt('9999.9999', nfFixed, 10, ['!', '9999.9999000000']);
  TestNumFmt('9999.9999', nfFixed, 12, ['!', '9999.999900000000']);
  TestNumFmt('9999.9999', nfFixed, 20, ['!', '9999.99990000000000000000', '*']);

  TestNumFmt('9999.9999', nfExponent, ['!', '9.99999990000E3']);
  TestNumFmt('9999.9999', nfExponent, 1, ['!', '1E4', '1.0E4']);
  TestNumFmt('9999.9999', nfExponent, 2, ['!', '1.0E4']);
  TestNumFmt('9999.9999', nfExponent, 3, ['!', '1.00E4']);
  TestNumFmt('9999.9999', nfExponent, 4, ['!', '1.000E4']);
  TestNumFmt('9999.9999', nfExponent, 5, ['!', '1.0000E4']);
  TestNumFmt('9999.9999', nfExponent, 6, ['!', '1.00000E4']);
  TestNumFmt('9999.9999', nfExponent, 7, ['!', '1.000000E4']);
  TestNumFmt('9999.9999', nfExponent, 8, ['!', '9.9999999E3']);
  TestNumFmt('9999.9999', nfExponent, 9, ['!', '9.99999990E3']);
  TestNumFmt('9999.9999', nfExponent, 10, ['!', '9.999999900E3']);
  TestNumFmt('9999.9999', nfExponent, 12, ['!', '9.99999990000E3']);
  TestNumFmt('9999.9999', nfExponent, 20, ['!', '9.9999999000000000000E3', '*']);

  // --

  TestNumFmt('9999', nfDefault, ['9999']);
  TestNumFmt('9999', nfDefault, 1, ['9999', '9999', '1E4']);
  TestNumFmt('9999', nfDefault, 2, ['9999', '9999', '1E4']);
  TestNumFmt('9999', nfDefault, 3, ['9999', '9999', '1E4']);
  TestNumFmt('9999', nfDefault, 4, ['9999']);
  TestNumFmt('9999', nfDefault, 6, ['9999']);
  TestNumFmt('9999', nfDefault, 10, ['9999']);
  TestNumFmt('9999', nfDefault, 12, ['9999']);
  TestNumFmt('9999', nfDefault, 20, ['9999']);

  TestNumFmt('9999', nfFraction, ['9999']);
  TestNumFmt('9999', nfFraction, 1, ['9999', '9999', '1E4']);
  TestNumFmt('9999', nfFraction, 2, ['9999', '9999', '1E4']);
  TestNumFmt('9999', nfFraction, 3, ['9999', '9999', '1E4']);
  TestNumFmt('9999', nfFraction, 4, ['9999']);
  TestNumFmt('9999', nfFraction, 6, ['9999']);
  TestNumFmt('9999', nfFraction, 10, ['9999']);
  TestNumFmt('9999', nfFraction, 12, ['9999']);
  TestNumFmt('9999', nfFraction, 20, ['9999']);

  TestNumFmt('9999', nfFixed, ['9999', '9999.000000000000']);
  TestNumFmt('9999', nfFixed, 0, ['9999', '9999']);
  TestNumFmt('9999', nfFixed, 1, ['9999', '9999.0']);
  TestNumFmt('9999', nfFixed, 2, ['9999', '9999.00']);
  TestNumFmt('9999', nfFixed, 3, ['9999', '9999.000']);
  TestNumFmt('9999', nfFixed, 4, ['9999', '9999.0000']);
  TestNumFmt('9999', nfFixed, 6, ['9999', '9999.000000']);
  TestNumFmt('9999', nfFixed, 10, ['9999', '9999.0000000000']);
  TestNumFmt('9999', nfFixed, 12, ['9999', '9999.000000000000']);
  TestNumFmt('9999', nfFixed, 20, ['9999', '9999.00000000000000000000', '*']);

  TestNumFmt('9999', nfExponent, ['9999', '9.99900000000E3']);
  TestNumFmt('9999', nfExponent, 1, ['9999', '1E4', '1.0E4']);
  TestNumFmt('9999', nfExponent, 2, ['9999', '1.0E4']);
  TestNumFmt('9999', nfExponent, 3, ['9999', '1.00E4']);
  TestNumFmt('9999', nfExponent, 4, ['9999', '9.999E3']);
  TestNumFmt('9999', nfExponent, 6, ['9999', '9.99900E3']);
  TestNumFmt('9999', nfExponent, 10, ['9999', '9.999000000E3']);
  TestNumFmt('9999', nfExponent, 12, ['9999', '9.99900000000E3']);
  TestNumFmt('9999', nfExponent, 20, ['9999', '9.9990000000000000000E3', '*']);

  // --

  TestNumFmt('999.9', nfDefault, ['!', '999.9']);
  TestNumFmt('999.9', nfDefault, 1, ['!', '1E3']);
  TestNumFmt('999.9', nfDefault, 2, ['!', '1E3']);
  TestNumFmt('999.9', nfDefault, 3, ['!', '1E3']);
  TestNumFmt('999.9', nfDefault, 4, ['!', '999.9']);
  TestNumFmt('999.9', nfDefault, 6, ['!', '999.9']);
  TestNumFmt('999.9', nfDefault, 10, ['!', '999.9']);
  TestNumFmt('999.9', nfDefault, 12, ['!', '999.9']);
  TestNumFmt('999.9', nfDefault, 20, ['!', '999.9']);

  TestNumFmt('999.9', nfFraction, ['!', '9999/10', '999.9']);
  TestNumFmt('999.9', nfFraction, 1, ['!', '9999/10', '1E3']);
  TestNumFmt('999.9', nfFraction, 2, ['!', '9999/10', '1E3']);
  TestNumFmt('999.9', nfFraction, 3, ['!', '9999/10', '1E3']);
  TestNumFmt('999.9', nfFraction, 4, ['!', '9999/10', '999.9']);
  TestNumFmt('999.9', nfFraction, 6, ['!', '9999/10', '999.9']);
  TestNumFmt('999.9', nfFraction, 10, ['!', '9999/10', '999.9']);
  TestNumFmt('999.9', nfFraction, 12, ['!', '9999/10', '999.9']);
  TestNumFmt('999.9', nfFraction, 20, ['!', '9999/10', '999.9']);

  TestNumFmt('999.9', nfFixed, ['!', '999.900000000000']);
  TestNumFmt('999.9', nfFixed, 0, ['!', '1000']);
  TestNumFmt('999.9', nfFixed, 1, ['!', '999.9']);
  TestNumFmt('999.9', nfFixed, 2, ['!', '999.90']);
  TestNumFmt('999.9', nfFixed, 3, ['!', '999.900']);
  TestNumFmt('999.9', nfFixed, 4, ['!', '999.9000']);
  TestNumFmt('999.9', nfFixed, 6, ['!', '999.900000']);
  TestNumFmt('999.9', nfFixed, 10, ['!', '999.9000000000']);
  TestNumFmt('999.9', nfFixed, 12, ['!', '999.900000000000']);
  TestNumFmt('999.9', nfFixed, 20, ['!', '999.90000000000000000000', '*']);

  TestNumFmt('999.9', nfExponent, ['!', '9.99900000000E2']);
  TestNumFmt('999.9', nfExponent, 1, ['!', '1E3', '1.0E3']);
  TestNumFmt('999.9', nfExponent, 2, ['!', '1.0E3']);
  TestNumFmt('999.9', nfExponent, 3, ['!', '1.00E3']);
  TestNumFmt('999.9', nfExponent, 4, ['!', '9.999E2']);
  TestNumFmt('999.9', nfExponent, 6, ['!', '9.99900E2']);
  TestNumFmt('999.9', nfExponent, 10, ['!', '9.999000000E2']);
  TestNumFmt('999.9', nfExponent, 12, ['!', '9.99900000000E2']);
  TestNumFmt('999.9', nfExponent, 20, ['!', '9.9990000000000000000E2', '*']);

  // --

  TestNumFmt('99.99', nfDefault, ['!', '99.99']);
  TestNumFmt('99.99', nfDefault, 1, ['!', '1E2']);
  TestNumFmt('99.99', nfDefault, 2, ['!', '1E2']);
  TestNumFmt('99.99', nfDefault, 3, ['!', '100']);
  TestNumFmt('99.99', nfDefault, 4, ['!', '99.99']);
  TestNumFmt('99.99', nfDefault, 6, ['!', '99.99']);
  TestNumFmt('99.99', nfDefault, 10, ['!', '99.99']);
  TestNumFmt('99.99', nfDefault, 12, ['!', '99.99']);
  TestNumFmt('99.99', nfDefault, 20, ['!', '99.99']);

  TestNumFmt('99.99', nfFraction, ['!', '9999/100', '99.99']);
  TestNumFmt('99.99', nfFraction, 1, ['!', '9999/100', '1E2']);
  TestNumFmt('99.99', nfFraction, 2, ['!', '9999/100', '1E2']);
  TestNumFmt('99.99', nfFraction, 3, ['!', '9999/100', '100']);
  TestNumFmt('99.99', nfFraction, 4, ['!', '9999/100', '99.99']);
  TestNumFmt('99.99', nfFraction, 6, ['!', '9999/100', '99.99']);
  TestNumFmt('99.99', nfFraction, 10, ['!', '9999/100', '99.99']);
  TestNumFmt('99.99', nfFraction, 12, ['!', '9999/100', '99.99']);
  TestNumFmt('99.99', nfFraction, 20, ['!', '9999/100', '99.99']);

  TestNumFmt('99.99', nfFixed, ['!', '99.990000000000']);
  TestNumFmt('99.99', nfFixed, 0, ['!', '100']);
  TestNumFmt('99.99', nfFixed, 1, ['!', '100.0']);
  TestNumFmt('99.99', nfFixed, 2, ['!', '99.99']);
  TestNumFmt('99.99', nfFixed, 3, ['!', '99.990']);
  TestNumFmt('99.99', nfFixed, 4, ['!', '99.9900']);
  TestNumFmt('99.99', nfFixed, 6, ['!', '99.990000']);
  TestNumFmt('99.99', nfFixed, 10, ['!', '99.9900000000']);
  TestNumFmt('99.99', nfFixed, 12, ['!', '99.990000000000']);
  TestNumFmt('99.99', nfFixed, 20, ['!', '99.99000000000000000000', '*']);

  TestNumFmt('99.99', nfExponent, ['!', '9.99900000000E1']);
  TestNumFmt('99.99', nfExponent, 1, ['!', '1E2', '1.0E2']);
  TestNumFmt('99.99', nfExponent, 2, ['!', '1.0E2']);
  TestNumFmt('99.99', nfExponent, 3, ['!', '1.00E2']);
  TestNumFmt('99.99', nfExponent, 4, ['!', '9.999E1']);
  TestNumFmt('99.99', nfExponent, 6, ['!', '9.99900E1']);
  TestNumFmt('99.99', nfExponent, 10, ['!', '9.999000000E1']);
  TestNumFmt('99.99', nfExponent, 12, ['!', '9.99900000000E1']);
  TestNumFmt('99.99', nfExponent, 20, ['!', '9.9990000000000000000E1', '*']);

  // --

  TestNumFmt('9.999', nfDefault, ['!', '9.999']);
  TestNumFmt('9.999', nfDefault, 1, ['!', '1E1', '10']);
  TestNumFmt('9.999', nfDefault, 2, ['!', '10']);
  TestNumFmt('9.999', nfDefault, 3, ['!', '10']);
  TestNumFmt('9.999', nfDefault, 4, ['!', '9.999']);
  TestNumFmt('9.999', nfDefault, 6, ['!', '9.999']);
  TestNumFmt('9.999', nfDefault, 10, ['!', '9.999']);
  TestNumFmt('9.999', nfDefault, 12, ['!', '9.999']);
  TestNumFmt('9.999', nfDefault, 20, ['!', '9.999']);

  TestNumFmt('9.999', nfFraction, ['!', '9999/1000', '9.999']);
  TestNumFmt('9.999', nfFraction, 1, ['!', '9999/1000', '10']);
  TestNumFmt('9.999', nfFraction, 2, ['!', '9999/1000', '10']);
  TestNumFmt('9.999', nfFraction, 3, ['!', '9999/1000', '10']);
  TestNumFmt('9.999', nfFraction, 4, ['!', '9999/1000', '9.999']);
  TestNumFmt('9.999', nfFraction, 6, ['!', '9999/1000', '9.999']);
  TestNumFmt('9.999', nfFraction, 10, ['!', '9999/1000', '9.999']);
  TestNumFmt('9.999', nfFraction, 12, ['!', '9999/1000', '9.999']);
  TestNumFmt('9.999', nfFraction, 20, ['!', '9999/1000', '9.999']);

  TestNumFmt('9.999', nfFixed, ['!', '9.999000000000']);
  TestNumFmt('9.999', nfFixed, 0, ['!', '10']);
  TestNumFmt('9.999', nfFixed, 1, ['!', '10.0']);
  TestNumFmt('9.999', nfFixed, 2, ['!', '10.00']);
  TestNumFmt('9.999', nfFixed, 3, ['!', '9.999']);
  TestNumFmt('9.999', nfFixed, 4, ['!', '9.9990']);
  TestNumFmt('9.999', nfFixed, 6, ['!', '9.999000']);
  TestNumFmt('9.999', nfFixed, 10, ['!', '9.9990000000']);
  TestNumFmt('9.999', nfFixed, 12, ['!', '9.999000000000']);
  TestNumFmt('9.999', nfFixed, 20, ['!', '9.99900000000000000000', '*']);

  TestNumFmt('9.999', nfExponent, ['!', '9.99900000000E0']);
  TestNumFmt('9.999', nfExponent, 1, ['!', '1E1', '1.0E1']);
  TestNumFmt('9.999', nfExponent, 2, ['!', '1.0E1']);
  TestNumFmt('9.999', nfExponent, 3, ['!', '1.00E1']);
  TestNumFmt('9.999', nfExponent, 4, ['!', '9.999E0']);
  TestNumFmt('9.999', nfExponent, 6, ['!', '9.99900E0']);
  TestNumFmt('9.999', nfExponent, 10, ['!', '9.999000000E0']);
  TestNumFmt('9.999', nfExponent, 12, ['!', '9.99900000000E0']);
  TestNumFmt('9.999', nfExponent, 20, ['!', '9.9990000000000000000E0', '*']);

  // --

  TestNumFmt('0.9999', nfDefault, ['!', '0.9999']);
  TestNumFmt('0.9999', nfDefault, 1, ['!', '1']);
  TestNumFmt('0.9999', nfDefault, 2, ['!', '1']);
  TestNumFmt('0.9999', nfDefault, 3, ['!', '1']);
  TestNumFmt('0.9999', nfDefault, 4, ['!', '0.9999']);
  TestNumFmt('0.9999', nfDefault, 6, ['!', '0.9999']);
  TestNumFmt('0.9999', nfDefault, 10, ['!', '0.9999']);
  TestNumFmt('0.9999', nfDefault, 12, ['!', '0.9999']);
  TestNumFmt('0.9999', nfDefault, 20, ['!', '0.9999']);

  TestNumFmt('0.9999', nfFraction, ['!', '9999/10000', '0.9999']);
  TestNumFmt('0.9999', nfFraction, 1, ['!', '9999/10000', '1']);
  TestNumFmt('0.9999', nfFraction, 2, ['!', '9999/10000', '1']);
  TestNumFmt('0.9999', nfFraction, 3, ['!', '9999/10000', '1']);
  TestNumFmt('0.9999', nfFraction, 4, ['!', '9999/10000', '0.9999']);
  TestNumFmt('0.9999', nfFraction, 6, ['!', '9999/10000', '0.9999']);
  TestNumFmt('0.9999', nfFraction, 10, ['!', '9999/10000', '0.9999']);
  TestNumFmt('0.9999', nfFraction, 12, ['!', '9999/10000', '0.9999']);
  TestNumFmt('0.9999', nfFraction, 20, ['!', '9999/10000', '0.9999']);

  TestNumFmt('0.9999', nfFixed, ['!', '0.999900000000']);
  TestNumFmt('0.9999', nfFixed, 0, ['!', '1']);
  TestNumFmt('0.9999', nfFixed, 1, ['!', '1.0']);
  TestNumFmt('0.9999', nfFixed, 2, ['!', '1.00']);
  TestNumFmt('0.9999', nfFixed, 3, ['!', '1.000']);
  TestNumFmt('0.9999', nfFixed, 4, ['!', '0.9999']);
  TestNumFmt('0.9999', nfFixed, 6, ['!', '0.999900']);
  TestNumFmt('0.9999', nfFixed, 10, ['!', '0.9999000000']);
  TestNumFmt('0.9999', nfFixed, 12, ['!', '0.999900000000']);
  TestNumFmt('0.9999', nfFixed, 20, ['!', '0.99990000000000000000', '*']);

  TestNumFmt('0.9999', nfExponent, ['!', '9.99900000000E-1']);
  TestNumFmt('0.9999', nfExponent, 1, ['!', '1E0', '1.0E0']);
  TestNumFmt('0.9999', nfExponent, 2, ['!', '1.0E0']);
  TestNumFmt('0.9999', nfExponent, 3, ['!', '1.00E0']);
  TestNumFmt('0.9999', nfExponent, 4, ['!', '9.999E-1']);
  TestNumFmt('0.9999', nfExponent, 6, ['!', '9.99900E-1']);
  TestNumFmt('0.9999', nfExponent, 10, ['!', '9.999000000E-1']);
  TestNumFmt('0.9999', nfExponent, 12, ['!', '9.99900000000E-1']);
  TestNumFmt('0.9999', nfExponent, 20, ['!', '9.9990000000000000000E-1', '*']);

  // --

  TestNumFmt('0.09999', nfDefault, ['!', '0.09999']);
  TestNumFmt('0.09999', nfDefault, 1, ['!', '0.1']);
  TestNumFmt('0.09999', nfDefault, 2, ['!', '0.1']);
  TestNumFmt('0.09999', nfDefault, 3, ['!', '0.1']);
  TestNumFmt('0.09999', nfDefault, 4, ['!', '0.09999']);
  TestNumFmt('0.09999', nfDefault, 6, ['!', '0.09999']);
  TestNumFmt('0.09999', nfDefault, 10, ['!', '0.09999']);
  TestNumFmt('0.09999', nfDefault, 12, ['!', '0.09999']);
  TestNumFmt('0.09999', nfDefault, 20, ['!', '0.09999']);

  TestNumFmt('0.09999', nfFraction, ['!', '9999/100000', '0.09999']);
  TestNumFmt('0.09999', nfFraction, 1, ['!', '9999/100000', '0.1']);
  TestNumFmt('0.09999', nfFraction, 2, ['!', '9999/100000', '0.1']);
  TestNumFmt('0.09999', nfFraction, 3, ['!', '9999/100000', '0.1']);
  TestNumFmt('0.09999', nfFraction, 4, ['!', '9999/100000', '0.09999']);
  TestNumFmt('0.09999', nfFraction, 6, ['!', '9999/100000', '0.09999']);
  TestNumFmt('0.09999', nfFraction, 10, ['!', '9999/100000', '0.09999']);
  TestNumFmt('0.09999', nfFraction, 12, ['!', '9999/100000', '0.09999']);
  TestNumFmt('0.09999', nfFraction, 20, ['!', '9999/100000', '0.09999']);

  TestNumFmt('0.09999', nfFixed, ['!', '0.099990000000']);
  TestNumFmt('0.09999', nfFixed, 0, ['!', '0']);
  TestNumFmt('0.09999', nfFixed, 1, ['!', '0.1']);
  TestNumFmt('0.09999', nfFixed, 2, ['!', '0.10']);
  TestNumFmt('0.09999', nfFixed, 3, ['!', '0.100']);
  TestNumFmt('0.09999', nfFixed, 4, ['!', '0.1000']);
  TestNumFmt('0.09999', nfFixed, 5, ['!', '0.09999']);
  TestNumFmt('0.09999', nfFixed, 6, ['!', '0.099990']);
  TestNumFmt('0.09999', nfFixed, 10, ['!', '0.0999900000']);
  TestNumFmt('0.09999', nfFixed, 12, ['!', '0.099990000000']);
  TestNumFmt('0.09999', nfFixed, 20, ['!', '0.09999000000000000000', '*']);

  TestNumFmt('0.09999', nfExponent, ['!', '9.99900000000E-2']);
  TestNumFmt('0.09999', nfExponent, 1, ['!', '1E-1', '1.0E-1']);
  TestNumFmt('0.09999', nfExponent, 2, ['!', '1.0E-1']);
  TestNumFmt('0.09999', nfExponent, 3, ['!', '1.00E-1']);
  TestNumFmt('0.09999', nfExponent, 4, ['!', '9.999E-2']);
  TestNumFmt('0.09999', nfExponent, 5, ['!', '9.9990E-2']);
  TestNumFmt('0.09999', nfExponent, 6, ['!', '9.99900E-2']);
  TestNumFmt('0.09999', nfExponent, 10, ['!', '9.999000000E-2']);
  TestNumFmt('0.09999', nfExponent, 12, ['!', '9.99900000000E-2']);
  TestNumFmt('0.09999', nfExponent, 20, ['!', '9.9990000000000000000E-2', '*']);

  // --

  TestNumFmt('0.009999', nfDefault, ['!', '0.009999']);
  TestNumFmt('0.009999', nfDefault, 1, ['!', '0.01']);
  TestNumFmt('0.009999', nfDefault, 2, ['!', '0.01']);
  TestNumFmt('0.009999', nfDefault, 3, ['!', '0.01']);
  TestNumFmt('0.009999', nfDefault, 4, ['!', '0.009999']);
  TestNumFmt('0.009999', nfDefault, 6, ['!', '0.009999']);
  TestNumFmt('0.009999', nfDefault, 10, ['!', '0.009999']);
  TestNumFmt('0.009999', nfDefault, 12, ['!', '0.009999']);
  TestNumFmt('0.009999', nfDefault, 20, ['!', '0.009999']);

  TestNumFmt('0.009999', nfFraction, ['!', '9999/1000000', '0.009999']);
  TestNumFmt('0.009999', nfFraction, 1, ['!', '9999/1000000', '0.01']);
  TestNumFmt('0.009999', nfFraction, 2, ['!', '9999/1000000', '0.01']);
  TestNumFmt('0.009999', nfFraction, 3, ['!', '9999/1000000', '0.01']);
  TestNumFmt('0.009999', nfFraction, 4, ['!', '9999/1000000', '0.009999']);
  TestNumFmt('0.009999', nfFraction, 6, ['!', '9999/1000000', '0.009999']);
  TestNumFmt('0.009999', nfFraction, 10, ['!', '9999/1000000', '0.009999']);
  TestNumFmt('0.009999', nfFraction, 12, ['!', '9999/1000000', '0.009999']);
  TestNumFmt('0.009999', nfFraction, 20, ['!', '9999/1000000', '0.009999']);

  TestNumFmt('0.009999', nfFixed, ['!', '0.009999000000']);
  TestNumFmt('0.009999', nfFixed, 0, ['!', '0']);
  TestNumFmt('0.009999', nfFixed, 1, ['!', '0.0']);
  TestNumFmt('0.009999', nfFixed, 2, ['!', '0.01']);
  TestNumFmt('0.009999', nfFixed, 3, ['!', '0.010']);
  TestNumFmt('0.009999', nfFixed, 4, ['!', '0.0100']);
  TestNumFmt('0.009999', nfFixed, 5, ['!', '0.01000']);
  TestNumFmt('0.009999', nfFixed, 6, ['!', '0.009999']);
  TestNumFmt('0.009999', nfFixed, 10, ['!', '0.0099990000']);
  TestNumFmt('0.009999', nfFixed, 12, ['!', '0.009999000000']);
  TestNumFmt('0.009999', nfFixed, 20, ['!', '0.00999900000000000000', '*']);

  TestNumFmt('0.009999', nfExponent, ['!', '9.99900000000E-3']);
  TestNumFmt('0.009999', nfExponent, 1, ['!', '1E-2', '1.0E-2']);
  TestNumFmt('0.009999', nfExponent, 2, ['!', '1.0E-2']);
  TestNumFmt('0.009999', nfExponent, 3, ['!', '1.00E-2']);
  TestNumFmt('0.009999', nfExponent, 4, ['!', '9.999E-3']);
  TestNumFmt('0.009999', nfExponent, 5, ['!', '9.9990E-3']);
  TestNumFmt('0.009999', nfExponent, 6, ['!', '9.99900E-3']);
  TestNumFmt('0.009999', nfExponent, 10, ['!', '9.999000000E-3']);
  TestNumFmt('0.009999', nfExponent, 12, ['!', '9.99900000000E-3']);
  TestNumFmt('0.009999', nfExponent, 20, ['!', '9.9990000000000000000E-3', '*']);

  // --

  TestNumFmt('5.9', nfDefault, ['!', '5.9']);
  TestNumFmt('5.9', nfDefault, 1, ['!', '6', '5.9']);
  TestNumFmt('5.9', nfDefault, 2, ['!', '5.9']);
  TestNumFmt('5.9', nfDefault, 4, ['!', '5.9']);
  TestNumFmt('5.9', nfDefault, 12, ['!', '5.9']);
  TestNumFmt('5.9', nfDefault, 20, ['!', '5.9']);

  TestNumFmt('5.9', nfFraction, ['!', '59/10', '5.9']);
  TestNumFmt('5.9', nfFraction, 1, ['!', '59/10', '5.9']);
  TestNumFmt('5.9', nfFraction, 2, ['!', '59/10', '5.9']);
  TestNumFmt('5.9', nfFraction, 4, ['!', '59/10', '5.9']);
  TestNumFmt('5.9', nfFraction, 12, ['!', '59/10', '5.9']);
  TestNumFmt('5.9', nfFraction, 20, ['!', '59/10', '5.9']);

  TestNumFmt('5.9', nfFixed, ['!', '5.900000000000']);
  TestNumFmt('5.9', nfFixed, 0, ['!', '6']);
  TestNumFmt('5.9', nfFixed, 1, ['!', '5.9']);
  TestNumFmt('5.9', nfFixed, 2, ['!', '5.90']);
  TestNumFmt('5.9', nfFixed, 3, ['!', '5.900']);
  TestNumFmt('5.9', nfFixed, 4, ['!', '5.9000']);
  TestNumFmt('5.9', nfFixed, 12, ['!', '5.900000000000']);
  TestNumFmt('5.9', nfFixed, 20, ['!', '5.90000000000000000000', '*']);

  TestNumFmt('5.9', nfExponent, ['!', '5.90000000000E0']);
  TestNumFmt('5.9', nfExponent, 1, ['!', '6E0', '5.9E0']);
  TestNumFmt('5.9', nfExponent, 2, ['!', '5.9E0']);
  TestNumFmt('5.9', nfExponent, 3, ['!', '5.90E0']);
  TestNumFmt('5.9', nfExponent, 4, ['!', '5.900E0']);
  TestNumFmt('5.9', nfExponent, 12, ['!', '5.90000000000E0']);
  TestNumFmt('5.9', nfExponent, 20, ['!', '5.9000000000000000000E0', '*']);

  // --

  TestNumFmt('5.5', nfDefault, ['!', '5.5']);
  TestNumFmt('5.5', nfDefault, 1, ['!', '6', '5.5']);
  TestNumFmt('5.5', nfDefault, 2, ['!', '5.5']);
  TestNumFmt('5.5', nfDefault, 4, ['!', '5.5']);
  TestNumFmt('5.5', nfDefault, 12, ['!', '5.5']);
  TestNumFmt('5.5', nfDefault, 20, ['!', '5.5']);

  TestNumFmt('5.5', nfFraction, ['!', '11/2', '5.5']);
  TestNumFmt('5.5', nfFraction, 1, ['!', '11/2', '5.5']);
  TestNumFmt('5.5', nfFraction, 2, ['!', '11/2', '5.5']);
  TestNumFmt('5.5', nfFraction, 4, ['!', '11/2', '5.5']);
  TestNumFmt('5.5', nfFraction, 12, ['!', '11/2', '5.5']);
  TestNumFmt('5.5', nfFraction, 20, ['!', '11/2', '5.5']);

  TestNumFmt('5.5', nfFixed, ['!', '5.500000000000']);
  TestNumFmt('5.5', nfFixed, 0, ['!', '6']);
  TestNumFmt('5.5', nfFixed, 1, ['!', '5.5']);
  TestNumFmt('5.5', nfFixed, 2, ['!', '5.50']);
  TestNumFmt('5.5', nfFixed, 3, ['!', '5.500']);
  TestNumFmt('5.5', nfFixed, 4, ['!', '5.5000']);
  TestNumFmt('5.5', nfFixed, 12, ['!', '5.500000000000']);
  TestNumFmt('5.5', nfFixed, 20, ['!', '5.50000000000000000000', '*']);

  TestNumFmt('5.5', nfExponent, ['!', '5.50000000000E0']);
  TestNumFmt('5.5', nfExponent, 1, ['!', '6E0', '5.5E0']);
  TestNumFmt('5.5', nfExponent, 2, ['!', '5.5E0']);
  TestNumFmt('5.5', nfExponent, 3, ['!', '5.50E0']);
  TestNumFmt('5.5', nfExponent, 4, ['!', '5.500E0']);
  TestNumFmt('5.5', nfExponent, 12, ['!', '5.50000000000E0']);
  TestNumFmt('5.5', nfExponent, 20, ['!', '5.5000000000000000000E0', '*']);

  // --

  TestNumFmt('5.1', nfDefault, ['!', '5.1']);
  TestNumFmt('5.1', nfDefault, 1, ['!', '5', '5.1']);
  TestNumFmt('5.1', nfDefault, 2, ['!', '5.1']);
  TestNumFmt('5.1', nfDefault, 4, ['!', '5.1']);
  TestNumFmt('5.1', nfDefault, 12, ['!', '5.1']);
  TestNumFmt('5.1', nfDefault, 20, ['!', '5.1']);

  TestNumFmt('5.1', nfFraction, ['!', '51/10', '5.1']);
  TestNumFmt('5.1', nfFraction, 1, ['!', '51/10', '5.1']);
  TestNumFmt('5.1', nfFraction, 2, ['!', '51/10', '5.1']);
  TestNumFmt('5.1', nfFraction, 4, ['!', '51/10', '5.1']);
  TestNumFmt('5.1', nfFraction, 12, ['!', '51/10', '5.1']);
  TestNumFmt('5.1', nfFraction, 20, ['!', '51/10', '5.1']);

  TestNumFmt('5.1', nfFixed, ['!', '5.100000000000']);
  TestNumFmt('5.1', nfFixed, 0, ['!', '5']);
  TestNumFmt('5.1', nfFixed, 1, ['!', '5.1']);
  TestNumFmt('5.1', nfFixed, 2, ['!', '5.10']);
  TestNumFmt('5.1', nfFixed, 3, ['!', '5.100']);
  TestNumFmt('5.1', nfFixed, 4, ['!', '5.1000']);
  TestNumFmt('5.1', nfFixed, 12, ['!', '5.100000000000']);
  TestNumFmt('5.1', nfFixed, 20, ['!', '5.10000000000000000000', '*']);

  TestNumFmt('5.1', nfExponent, ['!', '5.10000000000E0']);
  TestNumFmt('5.1', nfExponent, 1, ['!', '5E0', '5.1E0']);
  TestNumFmt('5.1', nfExponent, 2, ['!', '5.1E0']);
  TestNumFmt('5.1', nfExponent, 3, ['!', '5.10E0']);
  TestNumFmt('5.1', nfExponent, 4, ['!', '5.100E0']);
  TestNumFmt('5.1', nfExponent, 12, ['!', '5.10000000000E0']);
  TestNumFmt('5.1', nfExponent, 20, ['!', '5.1000000000000000000E0', '*']);

  // --

  TestNumFmt('0.0009', nfDefault, ['!', '0.0009']);
  TestNumFmt('0.0009', nfDefault, 1, ['!', '0.0009']);
  TestNumFmt('0.0009', nfDefault, 2, ['!', '0.0009']);
  TestNumFmt('0.0009', nfDefault, 3, ['!', '0.0009']);
  TestNumFmt('0.0009', nfDefault, 4, ['!', '0.0009']);
  TestNumFmt('0.0009', nfDefault, 6, ['!', '0.0009']);
  TestNumFmt('0.0009', nfDefault, 7, ['!', '0.0009']);
  TestNumFmt('0.0009', nfDefault, 8, ['!', '0.0009']);
  TestNumFmt('0.0009', nfDefault, 12, ['!', '0.0009']);
  TestNumFmt('0.0009', nfDefault, 32, ['!', '0.0009']);

  TestNumFmt('0.0009', nfFraction, ['!', '9/10000', '0.0009']);
  TestNumFmt('0.0009', nfFraction, 1, ['!', '9/10000', '0.0009']);
  TestNumFmt('0.0009', nfFraction, 2, ['!', '9/10000', '0.0009']);
  TestNumFmt('0.0009', nfFraction, 3, ['!', '9/10000', '0.0009']);
  TestNumFmt('0.0009', nfFraction, 4, ['!', '9/10000', '0.0009']);
  TestNumFmt('0.0009', nfFraction, 6, ['!', '9/10000', '0.0009']);
  TestNumFmt('0.0009', nfFraction, 7, ['!', '9/10000', '0.0009']);
  TestNumFmt('0.0009', nfFraction, 8, ['!', '9/10000', '0.0009']);
  TestNumFmt('0.0009', nfFraction, 12, ['!', '9/10000', '0.0009']);
  TestNumFmt('0.0009', nfFraction, 32, ['!', '9/10000', '0.0009']);

  TestNumFmt('0.0009', nfFixed, ['!', '0.000900000000']);
  TestNumFmt('0.0009', nfFixed, 0, ['!', '0']);
  TestNumFmt('0.0009', nfFixed, 1, ['!', '0.0']);
  TestNumFmt('0.0009', nfFixed, 2, ['!', '0.00']);
  TestNumFmt('0.0009', nfFixed, 3, ['!', '0.001']);
  TestNumFmt('0.0009', nfFixed, 4, ['!', '0.0009']);
  TestNumFmt('0.0009', nfFixed, 5, ['!', '0.00090']);
  TestNumFmt('0.0009', nfFixed, 6, ['!', '0.000900']);
  TestNumFmt('0.0009', nfFixed, 7, ['!', '0.0009000']);
  TestNumFmt('0.0009', nfFixed, 8, ['!', '0.00090000']);
  TestNumFmt('0.0009', nfFixed, 12, ['!', '0.000900000000']);
  TestNumFmt('0.0009', nfFixed, 32, ['!', '0.00090000000000000000000000000000', '*']);

  TestNumFmt('0.0009', nfExponent, ['!', '9.00000000000E-4']);
  TestNumFmt('0.0009', nfExponent, 1, ['!', '9E-4', '9.0E-4']);
  TestNumFmt('0.0009', nfExponent, 2, ['!', '9.0E-4']);
  TestNumFmt('0.0009', nfExponent, 3, ['!', '9.00E-4']);
  TestNumFmt('0.0009', nfExponent, 4, ['!', '9.000E-4']);
  TestNumFmt('0.0009', nfExponent, 5, ['!', '9.0000E-4']);
  TestNumFmt('0.0009', nfExponent, 6, ['!', '9.00000E-4']);
  TestNumFmt('0.0009', nfExponent, 7, ['!', '9.000000E-4']);
  TestNumFmt('0.0009', nfExponent, 8, ['!', '9.0000000E-4']);
  TestNumFmt('0.0009', nfExponent, 9, ['!', '9.00000000E-4']);
  TestNumFmt('0.0009', nfExponent, 10, ['!', '9.000000000E-4']);
  TestNumFmt('0.0009', nfExponent, 20, ['!', '9.0000000000000000000E-4', '*']);
  TestNumFmt('0.0009', nfExponent, 30, ['!', '9.00000000000000000000000000000E-4', '*']);

  // --

  TestNumFmt('0.00009', nfDefault, ['!', '9E-5']);
  TestNumFmt('0.00009', nfDefault, 1, ['!', '9E-5']);
  TestNumFmt('0.00009', nfDefault, 2, ['!', '9E-5']);
  TestNumFmt('0.00009', nfDefault, 3, ['!', '9E-5']);
  TestNumFmt('0.00009', nfDefault, 4, ['!', '9E-5']);
  TestNumFmt('0.00009', nfDefault, 6, ['!', '9E-5']);
  TestNumFmt('0.00009', nfDefault, 7, ['!', '9E-5']);
  TestNumFmt('0.00009', nfDefault, 8, ['!', '9E-5']);
  TestNumFmt('0.00009', nfDefault, 12, ['!', '9E-5']);
  TestNumFmt('0.00009', nfDefault, 32, ['!', '9E-5']);

  TestNumFmt('0.00009', nfFraction, ['!', '9/100000', '9E-5']);
  TestNumFmt('0.00009', nfFraction, 1, ['!', '9/100000', '9E-5']);
  TestNumFmt('0.00009', nfFraction, 2, ['!', '9/100000', '9E-5']);
  TestNumFmt('0.00009', nfFraction, 3, ['!', '9/100000', '9E-5']);
  TestNumFmt('0.00009', nfFraction, 4, ['!', '9/100000', '9E-5']);
  TestNumFmt('0.00009', nfFraction, 6, ['!', '9/100000', '9E-5']);
  TestNumFmt('0.00009', nfFraction, 7, ['!', '9/100000', '9E-5']);
  TestNumFmt('0.00009', nfFraction, 8, ['!', '9/100000', '9E-5']);
  TestNumFmt('0.00009', nfFraction, 12, ['!', '9/100000', '9E-5']);
  TestNumFmt('0.00009', nfFraction, 32, ['!', '9/100000', '9E-5']);

  TestNumFmt('0.00009', nfFixed, ['!', '0.000090000000']);
  TestNumFmt('0.00009', nfFixed, 0, ['!', '0']);
  TestNumFmt('0.00009', nfFixed, 1, ['!', '0.0']);
  TestNumFmt('0.00009', nfFixed, 2, ['!', '0.00']);
  TestNumFmt('0.00009', nfFixed, 3, ['!', '0.000']);
  TestNumFmt('0.00009', nfFixed, 4, ['!', '0.0001']);
  TestNumFmt('0.00009', nfFixed, 5, ['!', '0.00009']);
  TestNumFmt('0.00009', nfFixed, 6, ['!', '0.000090']);
  TestNumFmt('0.00009', nfFixed, 7, ['!', '0.0000900']);
  TestNumFmt('0.00009', nfFixed, 8, ['!', '0.00009000']);
  TestNumFmt('0.00009', nfFixed, 12, ['!', '0.000090000000']);
  TestNumFmt('0.00009', nfFixed, 32, ['!', '0.00009000000000000000000000000000', '*']);

  TestNumFmt('0.00009', nfExponent, ['!', '9.00000000000E-5']);
  TestNumFmt('0.00009', nfExponent, 1, ['!', '9E-5', '9.0E-5']);
  TestNumFmt('0.00009', nfExponent, 2, ['!', '9.0E-5']);
  TestNumFmt('0.00009', nfExponent, 3, ['!', '9.00E-5']);
  TestNumFmt('0.00009', nfExponent, 4, ['!', '9.000E-5']);
  TestNumFmt('0.00009', nfExponent, 5, ['!', '9.0000E-5']);
  TestNumFmt('0.00009', nfExponent, 6, ['!', '9.00000E-5']);
  TestNumFmt('0.00009', nfExponent, 7, ['!', '9.000000E-5']);
  TestNumFmt('0.00009', nfExponent, 8, ['!', '9.0000000E-5']);
  TestNumFmt('0.00009', nfExponent, 9, ['!', '9.00000000E-5']);
  TestNumFmt('0.00009', nfExponent, 10, ['!', '9.000000000E-5']);
  TestNumFmt('0.00009', nfExponent, 20, ['!', '9.0000000000000000000E-5', '*']);
  TestNumFmt('0.00009', nfExponent, 30, ['!', '9.00000000000000000000000000000E-5', '*']);

  // --

  TestNumFmt('0.9', nfDefault, ['!', '0.9']);
  TestNumFmt('0.9', nfDefault, 1, ['!', '0.9']);
  TestNumFmt('0.9', nfDefault, 2, ['!', '0.9']);
  TestNumFmt('0.9', nfDefault, 3, ['!', '0.9']);
  TestNumFmt('0.9', nfDefault, 4, ['!', '0.9']);
  TestNumFmt('0.9', nfDefault, 6, ['!', '0.9']);
  TestNumFmt('0.9', nfDefault, 10, ['!', '0.9']);
  TestNumFmt('0.9', nfDefault, 12, ['!', '0.9']);
  TestNumFmt('0.9', nfDefault, 20, ['!', '0.9']);

  TestNumFmt('0.9', nfFraction, ['!', '9/10', '0.9']);
  TestNumFmt('0.9', nfFraction, 1, ['!', '9/10', '0.9']);
  TestNumFmt('0.9', nfFraction, 2, ['!', '9/10', '0.9']);
  TestNumFmt('0.9', nfFraction, 3, ['!', '9/10', '0.9']);
  TestNumFmt('0.9', nfFraction, 4, ['!', '9/10', '0.9']);
  TestNumFmt('0.9', nfFraction, 6, ['!', '9/10', '0.9']);
  TestNumFmt('0.9', nfFraction, 10, ['!', '9/10', '0.9']);
  TestNumFmt('0.9', nfFraction, 12, ['!', '9/10', '0.9']);
  TestNumFmt('0.9', nfFraction, 20, ['!', '9/10', '0.9']);

  TestNumFmt('0.9', nfFixed, ['!', '0.900000000000']);
  TestNumFmt('0.9', nfFixed, 0, ['!', '1']);
  TestNumFmt('0.9', nfFixed, 1, ['!', '0.9']);
  TestNumFmt('0.9', nfFixed, 2, ['!', '0.90']);
  TestNumFmt('0.9', nfFixed, 3, ['!', '0.900']);
  TestNumFmt('0.9', nfFixed, 4, ['!', '0.9000']);
  TestNumFmt('0.9', nfFixed, 6, ['!', '0.900000']);
  TestNumFmt('0.9', nfFixed, 10, ['!', '0.9000000000']);
  TestNumFmt('0.9', nfFixed, 12, ['!', '0.900000000000']);
  TestNumFmt('0.9', nfFixed, 20, ['!', '0.90000000000000000000', '*']);

  TestNumFmt('0.9', nfExponent, ['!', '9.00000000000E-1']);
  TestNumFmt('0.9', nfExponent, 1, ['!', '9E-1', '9.0E-1']);
  TestNumFmt('0.9', nfExponent, 2, ['!', '9.0E-1']);
  TestNumFmt('0.9', nfExponent, 3, ['!', '9.00E-1']);
  TestNumFmt('0.9', nfExponent, 4, ['!', '9.000E-1']);
  TestNumFmt('0.9', nfExponent, 6, ['!', '9.00000E-1']);
  TestNumFmt('0.9', nfExponent, 10, ['!', '9.000000000E-1']);
  TestNumFmt('0.9', nfExponent, 12, ['!', '9.00000000000E-1']);
  TestNumFmt('0.9', nfExponent, 20, ['!', '9.0000000000000000000E-1', '*']);

  // --

  TestNumFmt('1/613', nfDefault, ['!', '0.00163132137031']);
  TestNumFmt('1/613', nfDefault, 1, ['!', '0.002', '0.0016']);
  TestNumFmt('1/613', nfDefault, 2, ['!', '0.0016']);
  TestNumFmt('1/613', nfDefault, 3, ['!', '0.00163']);
  TestNumFmt('1/613', nfDefault, 4, ['!', '0.001631']);
  TestNumFmt('1/613', nfDefault, 5, ['!', '0.0016313']);
  TestNumFmt('1/613', nfDefault, 6, ['!', '0.00163132']);
  TestNumFmt('1/613', nfDefault, 7, ['!', '0.001631321']);
  TestNumFmt('1/613', nfDefault, 8, ['!', '0.0016313214']);
  TestNumFmt('1/613', nfDefault, 9, ['!', '0.00163132137']);
  TestNumFmt('1/613', nfDefault, 10, ['!', '0.00163132137']);
  TestNumFmt('1/613', nfDefault, 11, ['!', '0.0016313213703']);
  TestNumFmt('1/613', nfDefault, 12, ['!', '0.00163132137031']);
  TestNumFmt('1/613', nfDefault, 20, ['!', '0.0016313213703099510604', '*']);
  TestNumFmt('1/613', nfDefault, 30, ['!', '0.00163132137030995106035889070147', '*']);
  TestNumFmt('1/613', nfDefault, 40, ['!', '0.001631321370309951060358890701468189233279', '*']);
  TestNumFmt('1/613', nfDefault, 50, ['!', '0.001631321370309951060358890701468189233278955954323', '*']);
  TestNumFmt('1/613', nfDefault, 60, ['!', '0.00163132137030995106035889070146818923327895595432300163132137', '*']);
  TestNumFmt('1/613', nfDefault, 70, ['!', '0.00163132137030995106035889070146818923327895595432300163132137030995106', '*']);
  TestNumFmt('1/613', nfDefault, 80, ['!', '0.0016313213703099510603588907014681892332789559543230016313213703099510603588907015', '*']);
  TestNumFmt('1/613', nfDefault, 90, ['!', '0.00163132137030995106035889070146818923327895595432300163132137030995106035889070146818923328', '*']);
  TestNumFmt('1/613', nfDefault, 100, ['!', '0.001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323', '*']);
  TestNumFmt('1/613', nfDefault, 150, ['!', '0.00163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432', '*']);
  TestNumFmt('1/613', nfDefault, 200, ['!', '0.0016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543', '*']);
  TestNumFmt('1/613', nfDefault, 250, ['!', '0.0016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313'+
  '21370309951060358890701468189233278955954', '*']);
  TestNumFmt('1/613', nfDefault, 1000, ['!', '0.001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631'+
  '32137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895'+'595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163'+
  '13213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313'+'21370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631'+
  '32137'+'0309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955'+'954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631'+
  '3213703099510'+'60358890701468', '*']);
  TestNumFmt('1/613', nfDefault, 2000, ['!', '0.0016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907'+'0146818923327895595432300163132137030995106035889070146818923327895595432300163'+
  '13213703099510603588907014681892'+'332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213'+'70309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001'+
  '6313'+'213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230'+'016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559'+'54323001631321370309951060358890701468189233278955954323'+
  '00163132137030995106035889070146818923327895595432300163132137030995106035889070146818'+'9233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309'+'95106035889070146818923327895595432300'+
  '1631321370309951060358890701468189'+
  '23327895595432300163132137030995106035889070'+'1468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358'+'89070146818923327895595432300163132137030995106035889070146818923'+
  '3278955954323001631321370309951060358890701468189233278955954323'+'00163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432'+'3001631321370309951060358890701468189233278'+
  '955954323001631321370309951060358890701468189233278955954323001631321370309951'+'06035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327'+'8955954323001631321370309951060358890701468189233278955954'+
  '323001631321370309951060358890701468189233278955954323001631321370309'+'951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954'+'3230016313213703', '*']);
  TestNumFmt('1/613', nfDefault, 3000, ['!', '0.0016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907'+'0146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995'+
  '10603588907014681892'+'332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213'+'703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313'+'21370309951060358'+
  '8907014681892332789559543230016313213703099510603588907014681892332789559543230'+'016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559'+'54323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189'+
  '23327895595432300163132137030995106035889070146818'+'9233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309'+'951060358890701468189233278955954323001631321370309951060358890701468189233'+
  '27895595432300163132137030995106035889070'+'1468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358'+'89070146818923327895595432300163132137030995106035889070146818923327895595432300'+
  '1631321370309951060358890701468189233278955954323'+'00163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432'+'3001631321370309951060358890701468189233278955954323001631321370309951'+
  '060358890701468189233278955954323001631321370309951'+'06035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327'+'8955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189'+
  '233278955954323001631321370309'+'951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954'+'3230016313213703'+
  '099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907'+'0146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300'+'1631321370309951060358890701468'+
  '189233278955954323001631321370309951060358890701468189233278955954323001631321370309'+'951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323'+'0016313213703099510603588907014681892332789559543230016'+
  '313213703099510603588907014681892332789559543230016313213703099510603588907014681892'+'3327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035'+'88907014681892332789559543230016313213703099510603588907'+
  '01468189233278955954323001631321370309951060358890701468189233278955'+'95432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327896', '*']);
  TestNumFmt('1/613', nfDefault, 4000, ['!', '0.0016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907'+'0146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818'+
  '92'+'332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213'+'703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313'+'21370309951060358890701468189233278955954323001631321'+
  '3703099510603588907014681892332789559543230'+'016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559'+'5432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818'+'9'+
  '233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309'+'9510603588907014681892332789'+
  '5595432300163132137030995106035889070146818923327895595432300163132137030995106035889070'+'1468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358'+'890701468189233278955954323001631321370309951060358'+
  '890701468189233278955954323001631321370309951060358890701468189233278955954323'+'00163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432'+'30016313213703099510603588907014681892332789559543230016313'+
  '21370309951060358890701468189233278955954323001631321370309951'+'06035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327'+'89559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789'+
  '55954323001631321370309'+'951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954'+'3230016313213703'+
  '099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907'+'0146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300'+'1631321370309951060358890701468189233278955954323'+
  '001631321370309951060358890701468189233278955954323001631321370309'+'951060358890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278955954323'+'0016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332'+
  '789559543230016313213703099510603588907014681892'+'3327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035'+'88907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014'+
  '68189233278955'+'95432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895'+
  '59543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213'+'703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543230016'+'313213703099510603588907014681892'+
  '33278955954323001631321370309951060358890701468189233278955954323001631321370309951060358'+'8907014681892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014'+'6818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146'+
  '81892332789559543230016'+'3132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030'+'995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327'+'895595432300163132137030995106035889'+
  '0701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951'+'060359', '*']);

  TestNumFmt('1/613', nfFraction, ['!', '1/613', '0.00163132137031']);
  TestNumFmt('1/613', nfFraction, 1, ['!', '1/613', '0.0016']);
  TestNumFmt('1/613', nfFraction, 2, ['!', '1/613', '0.0016']);
  TestNumFmt('1/613', nfFraction, 3, ['!', '1/613', '0.00163']);
  TestNumFmt('1/613', nfFraction, 4, ['!', '1/613', '0.001631']);
  TestNumFmt('1/613', nfFraction, 5, ['!', '1/613', '0.0016313']);
  TestNumFmt('1/613', nfFraction, 6, ['!', '1/613', '0.00163132']);
  TestNumFmt('1/613', nfFraction, 7, ['!', '1/613', '0.001631321']);
  TestNumFmt('1/613', nfFraction, 8, ['!', '1/613', '0.0016313214']);
  TestNumFmt('1/613', nfFraction, 9, ['!', '1/613', '0.00163132137']);
  TestNumFmt('1/613', nfFraction, 10, ['!', '1/613', '0.00163132137']);
  TestNumFmt('1/613', nfFraction, 11, ['!', '1/613', '0.0016313213703']);
  TestNumFmt('1/613', nfFraction, 12, ['!', '1/613', '0.00163132137031']);
  TestNumFmt('1/613', nfFraction, 20, ['!', '1/613', '*']);
  TestNumFmt('1/613', nfFraction, 30, ['!', '1/613', '*']);
  TestNumFmt('1/613', nfFraction, 40, ['!', '1/613', '*']);
  TestNumFmt('1/613', nfFraction, 50, ['!', '1/613', '*']);
  TestNumFmt('1/613', nfFraction, 4000, ['!', '1/613', '*']);

  TestNumFmt('1/613', nfFixed, ['!', '0.001631321370']);
  TestNumFmt('1/613', nfFixed,  0, ['!', '0']);
  TestNumFmt('1/613', nfFixed,  1, ['!', '0.0']);
  TestNumFmt('1/613', nfFixed,  2, ['!', '0.00']);
  TestNumFmt('1/613', nfFixed,  3, ['!', '0.002']);
  TestNumFmt('1/613', nfFixed,  4, ['!', '0.0016']);
  TestNumFmt('1/613', nfFixed,  5, ['!', '0.00163']);
  TestNumFmt('1/613', nfFixed,  6, ['!', '0.001631']);
  TestNumFmt('1/613', nfFixed,  7, ['!', '0.0016313']);
  TestNumFmt('1/613', nfFixed,  8, ['!', '0.00163132']);
  TestNumFmt('1/613', nfFixed,  9, ['!', '0.001631321']);
  TestNumFmt('1/613', nfFixed, 10, ['!', '0.0016313214']);
  TestNumFmt('1/613', nfFixed, 11, ['!', '0.00163132137']);
  TestNumFmt('1/613', nfFixed, 12, ['!', '0.001631321370']);
  TestNumFmt('1/613', nfFixed, 20, ['!', '0.00163132137030995106', '*']);
  TestNumFmt('1/613', nfFixed, 30, ['!', '0.001631321370309951060358890701', '*']);
  TestNumFmt('1/613', nfFixed, 40, ['!', '0.0016313213703099510603588907014681892333', '*']);
  TestNumFmt('1/613', nfFixed, 50, ['!', '0.00163132137030995106035889070146818923327895595432', '*']);
  TestNumFmt('1/613', nfFixed, 60, ['!', '0.001631321370309951060358890701468189233278955954323001631321', '*']);
  TestNumFmt('1/613', nfFixed, 70, ['!', '0.0016313213703099510603588907014681892332789559543230016313213703099511', '*']);
  TestNumFmt('1/613', nfFixed, 80, ['!', '0.00163132137030995106035889070146818923327895595432300163132137030995106035889070', '*']);
  TestNumFmt('1/613', nfFixed, 90, ['!', '0.001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233', '*']);
  TestNumFmt('1/613', nfFixed, 100, ['!', '0.0016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332789559543', '*']);
  TestNumFmt('1/613', nfFixed, 200, ['!', '0.00163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595', '*']);
  TestNumFmt('1/613', nfFixed, 300, ['!', '0.00163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035889070146818923327895595'+
  '4323001631321370309951060358890701468189233278955954323001631321370309951060358890701468189233278956', '*']);
  TestNumFmt('1/613', nfFixed, 400, ['!', '0.001631321370309951060358890701468189233278955954323001631321370309'+
  '95106035889070146818923327895595432300163132137030995106035889070146818923327895595432300163132137030995106035'+
  '8890701468189233278955954323001631321370309951060358890701468189233278955954323001631321370309951060358890701468'+
  '1892332789559543230016313213703099510603588907014681892332789559543230016313213703099510603588907014681892332790', '*']);

  TestNumFmt('1/613', nfExponent, ['!', '1.63132137031E-3']);
  TestNumFmt('1/613', nfExponent,  1, ['!', '2E-3', '1.6E-3']);
  TestNumFmt('1/613', nfExponent,  2, ['!', '1.6E-3']);
  TestNumFmt('1/613', nfExponent,  3, ['!', '1.63E-3']);
  TestNumFmt('1/613', nfExponent,  4, ['!', '1.631E-3']);
  TestNumFmt('1/613', nfExponent,  5, ['!', '1.6313E-3']);
  TestNumFmt('1/613', nfExponent,  6, ['!', '1.63132E-3']);
  TestNumFmt('1/613', nfExponent,  7, ['!', '1.631321E-3']);
  TestNumFmt('1/613', nfExponent,  8, ['!', '1.6313214E-3']);
  TestNumFmt('1/613', nfExponent,  9, ['!', '1.63132137E-3']);
  TestNumFmt('1/613', nfExponent, 10, ['!', '1.631321370E-3']);
  TestNumFmt('1/613', nfExponent, 11, ['!', '1.6313213703E-3']);
  TestNumFmt('1/613', nfExponent, 12, ['!', '1.63132137031E-3']);
  TestNumFmt('1/613', nfExponent, 20, ['!', '1.6313213703099510604E-3', '*']);
  TestNumFmt('1/613', nfExponent, 30, ['!', '1.63132137030995106035889070147E-3', '*']);
  TestNumFmt('1/613', nfExponent, 40, ['!', '1.631321370309951060358890701468189233279E-3', '*']);
  TestNumFmt('1/613', nfExponent, 50, ['!', '1.6313213703099510603588907014681892332789559543230E-3', '*']);
  TestNumFmt('1/613', nfExponent, 60, ['!', '1.63132137030995106035889070146818923327895595432300163132137E-3', '*']);
  TestNumFmt('1/613', nfExponent, 70, ['!', '1.631321370309951060358890701468189233278955954323001631321370309951060E-3', '*']);

  // --

  TestNumFmt('1/5137', nfDefault, ['!', '0.000194666147557']);
  TestNumFmt('1/5137', nfDefault,  1, ['!', '0.0002', '0.00019']);
  TestNumFmt('1/5137', nfDefault,  2, ['!', '0.00019']);
  TestNumFmt('1/5137', nfDefault,  3, ['!', '0.000195']);
  TestNumFmt('1/5137', nfDefault,  4, ['!', '0.0001947']);
  TestNumFmt('1/5137', nfDefault,  5, ['!', '0.00019467']);
  TestNumFmt('1/5137', nfDefault,  6, ['!', '0.000194666']);
  TestNumFmt('1/5137', nfDefault,  7, ['!', '0.0001946661']);
  TestNumFmt('1/5137', nfDefault,  8, ['!', '0.00019466615']);
  TestNumFmt('1/5137', nfDefault,  9, ['!', '0.000194666148']);
  TestNumFmt('1/5137', nfDefault, 10, ['!', '0.0001946661476']);
  TestNumFmt('1/5137', nfDefault, 11, ['!', '0.00019466614756']);
  TestNumFmt('1/5137', nfDefault, 12, ['!', '0.000194666147557']);
  TestNumFmt('1/5137', nfDefault, 20, ['!', '0.00019466614755693984816', '*']);
  TestNumFmt('1/5137', nfDefault, 30, ['!', '0.000194666147556939848160404905587', '*']);
  TestNumFmt('1/5137', nfDefault, 40, ['!', '0.0001946661475569398481604049055869184348842', '*']);
  TestNumFmt('1/5137', nfDefault, 50, ['!', '0.00019466614755693984816040490558691843488417364220362', '*']);
  TestNumFmt('1/5137', nfDefault, 60, ['!', '0.000194666147556939848160404905586918434884173642203620790344559', '*']);
  TestNumFmt('1/5137', nfDefault, 70, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835', '*']);
  TestNumFmt('1/5137', nfDefault, 80, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668', '*']);
  TestNumFmt('1/5137', nfDefault, 90, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884563', '*']);
  TestNumFmt('1/5137', nfDefault, 100, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467', '*']);
  TestNumFmt('1/5137', nfDefault, 110, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879891', '*']);
  TestNumFmt('1/5137', nfDefault, 120, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681', '*']);
  TestNumFmt('1/5137', nfDefault, 130, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017', '*']);
  TestNumFmt('1/5137', nfDefault, 140, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871326', '*']);
  TestNumFmt('1/5137', nfDefault, 150, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648628', '*']);
  TestNumFmt('1/5137', nfDefault, 160, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597236', '*']);
  TestNumFmt('1/5137', nfDefault, 170, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407046915', '*']);
  TestNumFmt('1/5137', nfDefault, 180, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612225', '*']);
  TestNumFmt('1/5137', nfDefault, 190, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340665758', '*']);
  TestNumFmt('1/5137', nfDefault, 200, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407046914541561222503406657582246447343', '*']);
  TestNumFmt('1/5137', nfDefault, 210, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612225034066575822464473428070858478', '*']);
  TestNumFmt('1/5137', nfDefault, 220, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340665758224644734280708584777107261047', '*']);
  TestNumFmt('1/5137', nfDefault, 230, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407046914541561222503406657582246447342807085847771072610473038738563', '*']);
  TestNumFmt('1/5137', nfDefault, 240, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612225034066575822464473428070858477710726104730387385633638310298', '*']);
  TestNumFmt('1/5137', nfDefault, 250, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340665758224644734280708584777107261047303873856336383102978392057621', '*']);

  TestNumFmt('1/5137', nfFraction, ['!', '1/5137', '0.000194666147557']);
  TestNumFmt('1/5137', nfFraction,  1, ['!', '1/5137', '0.00019']);
  TestNumFmt('1/5137', nfFraction,  2, ['!', '1/5137', '0.00019']);
  TestNumFmt('1/5137', nfFraction,  3, ['!', '1/5137', '0.000195']);
  TestNumFmt('1/5137', nfFraction,  4, ['!', '1/5137', '0.0001947']);
  TestNumFmt('1/5137', nfFraction,  5, ['!', '1/5137', '0.00019467']);
  TestNumFmt('1/5137', nfFraction,  6, ['!', '1/5137', '0.000194666']);
  TestNumFmt('1/5137', nfFraction,  7, ['!', '1/5137', '0.0001946661']);
  TestNumFmt('1/5137', nfFraction,  8, ['!', '1/5137', '0.00019466615']);
  TestNumFmt('1/5137', nfFraction,  9, ['!', '1/5137', '0.000194666148']);
  TestNumFmt('1/5137', nfFraction, 10, ['!', '1/5137', '0.0001946661476']);
  TestNumFmt('1/5137', nfFraction, 11, ['!', '1/5137', '0.00019466614756']);
  TestNumFmt('1/5137', nfFraction, 12, ['!', '1/5137', '0.000194666147557']);

  TestNumFmt('1/5137', nfFixed, ['!', '0.000194666148']);
  TestNumFmt('1/5137', nfFixed,  0, ['!', '0']);
  TestNumFmt('1/5137', nfFixed,  1, ['!', '0.0']);
  TestNumFmt('1/5137', nfFixed,  2, ['!', '0.00']);
  TestNumFmt('1/5137', nfFixed,  3, ['!', '0.000']);
  TestNumFmt('1/5137', nfFixed,  4, ['!', '0.0002']);
  TestNumFmt('1/5137', nfFixed,  5, ['!', '0.00019']);
  TestNumFmt('1/5137', nfFixed,  6, ['!', '0.000195']);
  TestNumFmt('1/5137', nfFixed,  7, ['!', '0.0001947']);
  TestNumFmt('1/5137', nfFixed,  8, ['!', '0.00019467']);
  TestNumFmt('1/5137', nfFixed,  9, ['!', '0.000194666']);
  TestNumFmt('1/5137', nfFixed, 10, ['!', '0.0001946661']);
  TestNumFmt('1/5137', nfFixed, 11, ['!', '0.00019466615']);
  TestNumFmt('1/5137', nfFixed, 12, ['!', '0.000194666148']);
  TestNumFmt('1/5137', nfFixed, 20, ['!', '0.00019466614755693985', '*']);
  TestNumFmt('1/5137', nfFixed, 30, ['!', '0.000194666147556939848160404906', '*']);
  TestNumFmt('1/5137', nfFixed, 40, ['!', '0.0001946661475569398481604049055869184349', '*']);
  TestNumFmt('1/5137', nfFixed, 50, ['!', '0.00019466614755693984816040490558691843488417364220', '*']);
  TestNumFmt('1/5137', nfFixed, 60, ['!', '0.000194666147556939848160404905586918434884173642203620790345', '*']);
  TestNumFmt('1/5137', nfFixed, 70, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811758', '*']);
  TestNumFmt('1/5137', nfFixed, 80, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124392', '*']);
  TestNumFmt('1/5137', nfFixed, 90, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888846', '*']);
  TestNumFmt('1/5137', nfFixed, 100, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873', '*']);
  TestNumFmt('1/5137', nfFixed, 110, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040880', '*']);
  TestNumFmt('1/5137', nfFixed, 120, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869574', '*']);
  TestNumFmt('1/5137', nfFixed, 130, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503', '*']);
  TestNumFmt('1/5137', nfFixed, 140, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871', '*']);
  TestNumFmt('1/5137', nfFixed, 150, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764649', '*']);
  TestNumFmt('1/5137', nfFixed, 160, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597', '*']);
  TestNumFmt('1/5137', nfFixed, 170, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407047', '*']);
  TestNumFmt('1/5137', nfFixed, 180, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612', '*']);
  TestNumFmt('1/5137', nfFixed, 190, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340666', '*']);
  TestNumFmt('1/5137', nfFixed, 200, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407046914541561222503406657582246447', '*']);
  TestNumFmt('1/5137', nfFixed, 210, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612225034066575822464473428070858', '*']);
  TestNumFmt('1/5137', nfFixed, 220, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340665758224644734280708584777107261', '*']);
  TestNumFmt('1/5137', nfFixed, 230, ['!', '0.00019466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407046914541561222503406657582246447342807085847771072610473038739', '*']);
  TestNumFmt('1/5137', nfFixed, 240, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612225034066575822464473428070858477710726104730387385633638310', '*']);
  TestNumFmt('1/5137', nfFixed, 250, ['!', '0.0001946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340665758224644734280708584777107261047303873856336383102978392058', '*']);
  TestNumFmt('1/5137', nfFixed, 1000, ['!', '0.000194666147556939848160404905586918434884173642203620790344559081175783531243'+'9166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340665758224644734280708584777107261047303873856336383102978392057621179676854195055479852053727856725715398092271753'+
  '941989488028031925248199338135098306404516254623321004477321393809616507689312828499124002335993770683278177924858867043021218610083706443449484134708974109402374927000194666147556939'+'848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612'+
  '2250340'+'66575822464473428070858477710726104730387385633638310297839205762117967685419505547985205372785672571539809227175394198948802803192524819933813509830640451625462332100447732'+'13938096165076893128284991240023359937706832781779248588670430212186100837064434494841347089741094023749270001946661475569398481604049055869184348841736422036207903'+
  '4455908118', '*']);

  TestNumFmt('1/5137', nfExponent, ['!', '1.94666147557E-4']);
  TestNumFmt('1/5137', nfExponent,  1, ['!', '2E-4', '1.9E-4']);
  TestNumFmt('1/5137', nfExponent,  2, ['!', '1.9E-4']);
  TestNumFmt('1/5137', nfExponent,  3, ['!', '1.95E-4']);
  TestNumFmt('1/5137', nfExponent,  4, ['!', '1.947E-4']);
  TestNumFmt('1/5137', nfExponent,  5, ['!', '1.9467E-4']);
  TestNumFmt('1/5137', nfExponent,  6, ['!', '1.94666E-4']);
  TestNumFmt('1/5137', nfExponent,  7, ['!', '1.946661E-4']);
  TestNumFmt('1/5137', nfExponent,  8, ['!', '1.9466615E-4']);
  TestNumFmt('1/5137', nfExponent,  9, ['!', '1.94666148E-4']);
  TestNumFmt('1/5137', nfExponent, 10, ['!', '1.946661476E-4']);
  TestNumFmt('1/5137', nfExponent, 11, ['!', '1.9466614756E-4']);
  TestNumFmt('1/5137', nfExponent, 12, ['!', '1.94666147557E-4']);
  TestNumFmt('1/5137', nfExponent, 13, ['!', '1.946661475569E-4']);
  TestNumFmt('1/5137', nfExponent, 14, ['!', '1.9466614755694E-4']);
  TestNumFmt('1/5137', nfExponent, 15, ['!', '1.94666147556940E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 16, ['!', '1.946661475569398E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 17, ['!', '1.9466614755693985E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 18, ['!', '1.94666147556939848E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 19, ['!', '1.946661475569398482E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 20, ['!', '1.9466614755693984816E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 30, ['!', '1.94666147556939848160404905587E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 40, ['!', '1.946661475569398481604049055869184348842E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 50, ['!', '1.9466614755693984816040490558691843488417364220362E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 60, ['!', '1.94666147556939848160404905586918434884173642203620790344559E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 70, ['!', '1.946661475569398481604049055869184348841736422036207903445590811757835E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 80, ['!', '1.9466614755693984816040490558691843488417364220362079034455908117578353124391668E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 90, ['!', '1.94666147556939848160404905586918434884173642203620790344559081175783531243916682888845630E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 100, ['!', '1.946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 110, ['!', '1.9466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879891E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 120, ['!', '1.94666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 130, ['!', '1.946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 140, ['!', '1.9466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871326E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 150, ['!', '1.94666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648628E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 160, ['!', '1.946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597236E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 170, ['!', '1.9466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407046915E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 180, ['!', '1.94666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612225E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 190, ['!', '1.946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340665758E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 200, ['!', '1.9466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407046914541561222503406657582246447343E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 210, ['!', '1.94666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612225034066575822464473428070858478E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 220, ['!', '1.946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340665758224644734280708584777107261047E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 230, ['!', '1.9466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407046914541561222503406657582246447342807085847771072610473038738563E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 240, ['!', '1.94666147556939848160404905586918434884173642203620790344559081175783531243916682888845629744987346700408798909869573681136850301732528713256764648627603659723574070469145415612225034066575822464473428070858477710726104730387385633638310298E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 250, ['!', '1.946661475569398481604049055869184348841736422036207903445590811757835312439166828888456297449873467004087989098695736811368503017325287132567646486276036597235740704691454156122250340665758224644734280708584777107261047303873856336383102978392057621E-4', '*']);
  TestNumFmt('1/5137', nfExponent, 1000, ['!', '1.9466614755693984816040490558691843488417364220362079034455908117578353124391668288884562974498734670040879890986957368113685030173252871325676464862760365972357407046914541561222503406657582246447342807085847771072610473038738563363831029783920576211'+
  '7967685419505547985205372785672571539'+'809227175394198948802803192524819933813509830640451625462332100447732139380961650768931282849912400233599377068327817792485886704302121861008370644344948413470897410940237492700019466614755693984816040490558691843488417364'+'22036207903445590811757835312439166828'+
  '8884562974498734670040879890986957368'+'113685030173252871325676464862760365972357407046914541561222503406657582246447342807085847771072610473038738563363831029783920576211796768541950554798520537278567257153980922717539419894880280319252481993381350983064045162'+'54623321004477321393809616507689312828'+
  '49912400233599377068327817792485886704302121861008370644344948413470897410940237492700019466614755693984816040490558691843488417364220362079034455908117578E-4', '*']);

  // --

  TestNumFmt('1/9', nfDefault, ['!', '0.111111111111']);
  TestNumFmt('1/9', nfDefault, 01, ['!', '0.1', '0.11']);
  TestNumFmt('1/9', nfDefault, 02, ['!', '0.11']);
  TestNumFmt('1/9', nfDefault, 03, ['!', '0.111']);
  TestNumFmt('1/9', nfDefault, 04, ['!', '0.1111']);
  TestNumFmt('1/9', nfDefault, 05, ['!', '0.11111']);
  TestNumFmt('1/9', nfDefault, 06, ['!', '0.111111']);
  TestNumFmt('1/9', nfDefault, 07, ['!', '0.1111111']);
  TestNumFmt('1/9', nfDefault, 08, ['!', '0.11111111']);
  TestNumFmt('1/9', nfDefault, 09, ['!', '0.111111111']);
  TestNumFmt('1/9', nfDefault, 10, ['!', '0.1111111111']);
  TestNumFmt('1/9', nfDefault, 11, ['!', '0.11111111111']);
  TestNumFmt('1/9', nfDefault, 12, ['!', '0.111111111111']);
  TestNumFmt('1/9', nfDefault, 100, ['!', '0.1111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111', '*']);
  TestNumFmt('1/9', nfDefault, 200, ['!', '0.11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111', '*']);

  TestNumFmt('1/9', nfFraction, ['!', '1/9', '0.111111111111']);
  TestNumFmt('1/9', nfFraction, 01, ['!', '1/9', '0.11']);
  TestNumFmt('1/9', nfFraction, 02, ['!', '1/9', '0.11']);
  TestNumFmt('1/9', nfFraction, 03, ['!', '1/9', '0.111']);
  TestNumFmt('1/9', nfFraction, 04, ['!', '1/9', '0.1111']);
  TestNumFmt('1/9', nfFraction, 05, ['!', '1/9', '0.11111']);
  TestNumFmt('1/9', nfFraction, 06, ['!', '1/9', '0.111111']);
  TestNumFmt('1/9', nfFraction, 07, ['!', '1/9', '0.1111111']);
  TestNumFmt('1/9', nfFraction, 08, ['!', '1/9', '0.11111111']);
  TestNumFmt('1/9', nfFraction, 09, ['!', '1/9', '0.111111111']);
  TestNumFmt('1/9', nfFraction, 10, ['!', '1/9', '0.1111111111']);
  TestNumFmt('1/9', nfFraction, 11, ['!', '1/9', '0.11111111111']);
  TestNumFmt('1/9', nfFraction, 12, ['!', '1/9', '0.111111111111']);
  TestNumFmt('1/9', nfFraction, 100, ['!', '1/9', '*']);
  TestNumFmt('1/9', nfFraction, 200, ['!', '1/9', '*']);

  TestNumFmt('1/9', nfFixed, ['!', '0.111111111111']);
  TestNumFmt('1/9', nfFixed, 00, ['!', '0']);
  TestNumFmt('1/9', nfFixed, 01, ['!', '0.1']);
  TestNumFmt('1/9', nfFixed, 02, ['!', '0.11']);
  TestNumFmt('1/9', nfFixed, 03, ['!', '0.111']);
  TestNumFmt('1/9', nfFixed, 04, ['!', '0.1111']);
  TestNumFmt('1/9', nfFixed, 05, ['!', '0.11111']);
  TestNumFmt('1/9', nfFixed, 06, ['!', '0.111111']);
  TestNumFmt('1/9', nfFixed, 07, ['!', '0.1111111']);
  TestNumFmt('1/9', nfFixed, 08, ['!', '0.11111111']);
  TestNumFmt('1/9', nfFixed, 09, ['!', '0.111111111']);
  TestNumFmt('1/9', nfFixed, 10, ['!', '0.1111111111']);
  TestNumFmt('1/9', nfFixed, 11, ['!', '0.11111111111']);
  TestNumFmt('1/9', nfFixed, 12, ['!', '0.111111111111']);
  TestNumFmt('1/9', nfFixed, 100, ['!', '0.1111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111', '*']);
  TestNumFmt('1/9', nfFixed, 200, ['!', '0.11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111', '*']);

  TestNumFmt('1/9', nfExponent, ['!', '1.11111111111E-1']);
  TestNumFmt('1/9', nfExponent, 01, ['!', '1E-1', '1.1E-1']);
  TestNumFmt('1/9', nfExponent, 02, ['!', '1.1E-1']);
  TestNumFmt('1/9', nfExponent, 03, ['!', '1.11E-1']);
  TestNumFmt('1/9', nfExponent, 04, ['!', '1.111E-1']);
  TestNumFmt('1/9', nfExponent, 05, ['!', '1.1111E-1']);
  TestNumFmt('1/9', nfExponent, 06, ['!', '1.11111E-1']);
  TestNumFmt('1/9', nfExponent, 07, ['!', '1.111111E-1']);
  TestNumFmt('1/9', nfExponent, 08, ['!', '1.1111111E-1']);
  TestNumFmt('1/9', nfExponent, 09, ['!', '1.11111111E-1']);
  TestNumFmt('1/9', nfExponent, 10, ['!', '1.111111111E-1']);
  TestNumFmt('1/9', nfExponent, 11, ['!', '1.1111111111E-1']);
  TestNumFmt('1/9', nfExponent, 12, ['!', '1.11111111111E-1']);
  TestNumFmt('1/9', nfExponent, 100, ['!', '1.111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111E-1', '*']);
  TestNumFmt('1/9', nfExponent, 200, ['!', '1.1111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111E-1', '*']);

  // --

  TestNumFmt('1/999', nfDefault, ['!', '0.001001001001']);
  TestNumFmt('1/999', nfDefault,  1, ['!', '0.001']);
  TestNumFmt('1/999', nfDefault,  2, ['!', '0.001']);
  TestNumFmt('1/999', nfDefault,  3, ['!', '0.001']);
  TestNumFmt('1/999', nfDefault,  4, ['!', '0.001001']);
  TestNumFmt('1/999', nfDefault,  5, ['!', '0.001001']);
  TestNumFmt('1/999', nfDefault,  6, ['!', '0.001001']);
  TestNumFmt('1/999', nfDefault,  7, ['!', '0.001001001']);
  TestNumFmt('1/999', nfDefault,  8, ['!', '0.001001001']);
  TestNumFmt('1/999', nfDefault,  9, ['!', '0.001001001']);
  TestNumFmt('1/999', nfDefault, 10, ['!', '0.001001001001']);
  TestNumFmt('1/999', nfDefault, 11, ['!', '0.001001001001']);
  TestNumFmt('1/999', nfDefault, 12, ['!', '0.001001001001']);

  TestNumFmt('1/999', nfFraction, ['!', '1/999', '0.001001001001']);
  TestNumFmt('1/999', nfFraction,  1, ['!', '1/999', '0.001']);
  TestNumFmt('1/999', nfFraction,  2, ['!', '1/999', '0.001']);
  TestNumFmt('1/999', nfFraction,  3, ['!', '1/999', '0.001']);
  TestNumFmt('1/999', nfFraction,  4, ['!', '1/999', '0.001001']);
  TestNumFmt('1/999', nfFraction,  5, ['!', '1/999', '0.001001']);
  TestNumFmt('1/999', nfFraction,  6, ['!', '1/999', '0.001001']);
  TestNumFmt('1/999', nfFraction,  7, ['!', '1/999', '0.001001001']);
  TestNumFmt('1/999', nfFraction,  8, ['!', '1/999', '0.001001001']);
  TestNumFmt('1/999', nfFraction,  9, ['!', '1/999', '0.001001001']);
  TestNumFmt('1/999', nfFraction, 10, ['!', '1/999', '0.001001001001']);
  TestNumFmt('1/999', nfFraction, 11, ['!', '1/999', '0.001001001001']);
  TestNumFmt('1/999', nfFraction, 12, ['!', '1/999', '0.001001001001']);

  TestNumFmt('1/999', nfFixed, ['!', '0.001001001001']);
  TestNumFmt('1/999', nfFixed,  0, ['!', '0']);
  TestNumFmt('1/999', nfFixed,  1, ['!', '0.0']);
  TestNumFmt('1/999', nfFixed,  2, ['!', '0.00']);
  TestNumFmt('1/999', nfFixed,  3, ['!', '0.001']);
  TestNumFmt('1/999', nfFixed,  4, ['!', '0.0010']);
  TestNumFmt('1/999', nfFixed,  5, ['!', '0.00100']);
  TestNumFmt('1/999', nfFixed,  6, ['!', '0.001001']);
  TestNumFmt('1/999', nfFixed,  7, ['!', '0.0010010']);
  TestNumFmt('1/999', nfFixed,  8, ['!', '0.00100100']);
  TestNumFmt('1/999', nfFixed,  9, ['!', '0.001001001']);
  TestNumFmt('1/999', nfFixed, 10, ['!', '0.0010010010']);
  TestNumFmt('1/999', nfFixed, 11, ['!', '0.00100100100']);
  TestNumFmt('1/999', nfFixed, 12, ['!', '0.001001001001']);

  TestNumFmt('1/999', nfExponent, ['!', '1.00100100100E-3']);
  TestNumFmt('1/999', nfExponent,  1, ['!', '1E-3', '1.0E-3']);
  TestNumFmt('1/999', nfExponent,  2, ['!', '1.0E-3']);
  TestNumFmt('1/999', nfExponent,  3, ['!', '1.00E-3']);
  TestNumFmt('1/999', nfExponent,  4, ['!', '1.001E-3']);
  TestNumFmt('1/999', nfExponent,  5, ['!', '1.0010E-3']);
  TestNumFmt('1/999', nfExponent,  6, ['!', '1.00100E-3']);
  TestNumFmt('1/999', nfExponent,  7, ['!', '1.001001E-3']);
  TestNumFmt('1/999', nfExponent,  8, ['!', '1.0010010E-3']);
  TestNumFmt('1/999', nfExponent,  9, ['!', '1.00100100E-3']);
  TestNumFmt('1/999', nfExponent, 10, ['!', '1.001001001E-3']);
  TestNumFmt('1/999', nfExponent, 11, ['!', '1.0010010010E-3']);
  TestNumFmt('1/999', nfExponent, 12, ['!', '1.00100100100E-3']);

  // --

  TestNumFmt('1/123456789', nfDefault, ['!', '8.10000007371E-9']);
  TestNumFmt('1/123456789', nfDefault,  1, ['!', '8E-9', '8.1E-9']);
  TestNumFmt('1/123456789', nfDefault,  2, ['!', '8.1E-9']);
  TestNumFmt('1/123456789', nfDefault,  3, ['!', '8.1E-9']);
  TestNumFmt('1/123456789', nfDefault,  4, ['!', '8.1E-9']);
  TestNumFmt('1/123456789', nfDefault,  5, ['!', '8.1E-9']);
  TestNumFmt('1/123456789', nfDefault,  6, ['!', '8.1E-9']);
  TestNumFmt('1/123456789', nfDefault,  7, ['!', '8.1E-9']);
  TestNumFmt('1/123456789', nfDefault,  8, ['!', '8.1000001E-9']);
  TestNumFmt('1/123456789', nfDefault,  9, ['!', '8.10000007E-9']);
  TestNumFmt('1/123456789', nfDefault, 10, ['!', '8.100000074E-9']);
  TestNumFmt('1/123456789', nfDefault, 11, ['!', '8.1000000737E-9']);
  TestNumFmt('1/123456789', nfDefault, 12, ['!', '8.10000007371E-9']);
  TestNumFmt('1/123456789', nfDefault, 20, ['!', '8.1000000737100006708E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 30, ['!', '8.10000007371000067076100610393E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 40, ['!', '8.100000073710000670761006103925155545719E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 50, ['!', '8.1000000737100006707610061039251555457189154660421E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 60, ['!', '8.10000007371000067076100610392515554571891546604213074098339E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 70, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948847E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 80, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 90, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 100, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327052E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 110, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 120, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315281E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 130, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690544E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 140, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 150, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315280701169054380638394863809393261E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 160, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690543806383948638093932606654786721E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 170, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486380939326066547867205585591571E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 180, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315280701169054380638394863809393260665478672055855915708288832945E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 190, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690543806383948638093932606654786720558559157082888329454283798034E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 200, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486380939326066547867205585591570828883294542837980339825621092E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 210, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486380939326066547867205585591570828883294542837980339825621092413151941E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 220, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690543806383948638093932606654786720558559157082888329454283798033982562109241315194095968266273E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 230, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486380939326066547867205585591570828883294542837980339825621092413151940959682662733112230871E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 240, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486380939326066547867205585591570828883294542837980339825621092413151940959682662733112230871321300929E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 250, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690543806383948638093932606654786720558559157082888329454283798033982562109241315194095968266273311223087132130092902383845412E-9', '*']);
  TestNumFmt('1/123456789', nfDefault, 500, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315280701169054380638394863809393260665478672055855915708288832945428379803398256210924131519409596826627331122308713213009290238384541169'+
  '29932464062385422967707349006136875955845571198194697903571750922502933394776693892467914421458021235267993240938738492542520282136934567446104563759551530211918924928462216849006173325956177266201213122431039414122458668514373883480802339675301291E-9', '*']);

  TestNumFmt('1/123456789', nfFraction, ['!', '1/123456789', '8.10000007371E-9']);
  TestNumFmt('1/123456789', nfFraction,  1, ['!', '1/123456789', '8.1E-9']);
  TestNumFmt('1/123456789', nfFraction,  2, ['!', '1/123456789', '8.1E-9']);
  TestNumFmt('1/123456789', nfFraction,  3, ['!', '1/123456789', '8.1E-9']);
  TestNumFmt('1/123456789', nfFraction,  4, ['!', '1/123456789', '8.1E-9']);
  TestNumFmt('1/123456789', nfFraction,  5, ['!', '1/123456789', '8.1E-9']);
  TestNumFmt('1/123456789', nfFraction,  6, ['!', '1/123456789', '8.1E-9']);
  TestNumFmt('1/123456789', nfFraction,  7, ['!', '1/123456789', '8.1E-9']);
  TestNumFmt('1/123456789', nfFraction,  8, ['!', '1/123456789', '8.1000001E-9']);
  TestNumFmt('1/123456789', nfFraction,  9, ['!', '1/123456789', '8.10000007E-9']);
  TestNumFmt('1/123456789', nfFraction, 10, ['!', '1/123456789', '8.100000074E-9']);
  TestNumFmt('1/123456789', nfFraction, 11, ['!', '1/123456789', '8.1000000737E-9']);
  TestNumFmt('1/123456789', nfFraction, 12, ['!', '1/123456789', '8.10000007371E-9']);

  TestNumFmt('1/123456789', nfFixed, ['!', '0.000000008100']);
  TestNumFmt('1/123456789', nfFixed,  0, ['!', '0']);
  TestNumFmt('1/123456789', nfFixed,  1, ['!', '0.0']);
  TestNumFmt('1/123456789', nfFixed,  2, ['!', '0.00']);
  TestNumFmt('1/123456789', nfFixed,  3, ['!', '0.000']);
  TestNumFmt('1/123456789', nfFixed,  4, ['!', '0.0000']);
  TestNumFmt('1/123456789', nfFixed,  5, ['!', '0.00000']);
  TestNumFmt('1/123456789', nfFixed,  6, ['!', '0.000000']);
  TestNumFmt('1/123456789', nfFixed,  7, ['!', '0.0000000']);
  TestNumFmt('1/123456789', nfFixed,  8, ['!', '0.00000001']);
  TestNumFmt('1/123456789', nfFixed,  9, ['!', '0.000000008']);
  TestNumFmt('1/123456789', nfFixed, 10, ['!', '0.0000000081']);
  TestNumFmt('1/123456789', nfFixed, 11, ['!', '0.00000000810']);
  TestNumFmt('1/123456789', nfFixed, 12, ['!', '0.000000008100']);
  TestNumFmt('1/123456789', nfFixed, 13, ['!', '0.0000000081000']);
  TestNumFmt('1/123456789', nfFixed, 14, ['!', '0.00000000810000']);
  TestNumFmt('1/123456789', nfFixed, 15, ['!', '0.000000008100000', '*']);
  TestNumFmt('1/123456789', nfFixed, 16, ['!', '0.0000000081000001', '*']);
  TestNumFmt('1/123456789', nfFixed, 17, ['!', '0.00000000810000007', '*']);
  TestNumFmt('1/123456789', nfFixed, 18, ['!', '0.000000008100000074', '*']);
  TestNumFmt('1/123456789', nfFixed, 19, ['!', '0.0000000081000000737', '*']);
  TestNumFmt('1/123456789', nfFixed, 20, ['!', '0.00000000810000007371', '*']);
  TestNumFmt('1/123456789', nfFixed, 30, ['!', '0.000000008100000073710000670761', '*']);
  TestNumFmt('1/123456789', nfFixed, 40, ['!', '0.0000000081000000737100006707610061039252', '*']);
  TestNumFmt('1/123456789', nfFixed, 50, ['!', '0.00000000810000007371000067076100610392515554571892', '*']);
  TestNumFmt('1/123456789', nfFixed, 100, ['!', '0.0000000081000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837', '*']);
  TestNumFmt('1/123456789', nfFixed, 150, ['!', '0.000000008100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690543806383948638', '*']);
  TestNumFmt('1/123456789', nfFixed, 200, ['!', '0.00000000810000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315280701169054380638394863809393260665478672055855915708288832945428379803398', '*']);
  TestNumFmt('1/123456789', nfFixed, 250, ['!', '0.0000000081000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486380939326066547867205585591570828883294542837980339825621092413151940959682662733112230871321300929024', '*']);

  TestNumFmt('1/123456789', nfExponent, ['!', '8.10000007371E-9']);
  TestNumFmt('1/123456789', nfExponent,  1, ['!', '8E-9', '8.1E-9']);
  TestNumFmt('1/123456789', nfExponent,  2, ['!', '8.1E-9']);
  TestNumFmt('1/123456789', nfExponent,  3, ['!', '8.10E-9']);
  TestNumFmt('1/123456789', nfExponent,  4, ['!', '8.100E-9']);
  TestNumFmt('1/123456789', nfExponent,  5, ['!', '8.1000E-9']);
  TestNumFmt('1/123456789', nfExponent,  6, ['!', '8.10000E-9']);
  TestNumFmt('1/123456789', nfExponent,  7, ['!', '8.100000E-9']);
  TestNumFmt('1/123456789', nfExponent,  8, ['!', '8.1000001E-9']);
  TestNumFmt('1/123456789', nfExponent,  9, ['!', '8.10000007E-9']);
  TestNumFmt('1/123456789', nfExponent, 10, ['!', '8.100000074E-9']);
  TestNumFmt('1/123456789', nfExponent, 11, ['!', '8.1000000737E-9']);
  TestNumFmt('1/123456789', nfExponent, 12, ['!', '8.10000007371E-9']);
  TestNumFmt('1/123456789', nfExponent, 20, ['!', '8.1000000737100006708E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 30, ['!', '8.10000007371000067076100610393E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 40, ['!', '8.100000073710000670761006103925155545719E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 50, ['!', '8.1000000737100006707610061039251555457189154660421E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 60, ['!', '8.10000007371000067076100610392515554571891546604213074098339E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 70, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948847E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 80, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 90, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 100, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327052E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 110, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 120, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315281E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 130, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690544E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 140, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 150, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315280701169054380638394863809393261E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 160, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690543806383948638093932606654786721E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 170, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486380939326066547867205585591571E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 180, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315280701169054380638394863809393260665478672055855915708288832945E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 190, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690543806383948638093932606654786720558559157082888329454283798034E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 200, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486380939326066547867205585591570828883294542837980339825621092E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 210, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315280701169054380638394863809393260665478672055855915708288832945428379803398256210924131519410E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 220, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690543806383948638093932606654786720558559157082888329454283798033982562109241315194095968266273E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 230, ['!', '8.1000000737100006707610061039251555457189154660421307409833897429488466608345046135939919837053270517184761706381331528070116905438063839486380939326066547867205585591570828883294542837980339825621092413151940959682662733112230871E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 240, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315280701169054380638394863809393260665478672055855915708288832945428379803398256210924131519409596826627331122308713213009290E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 250, ['!', '8.100000073710000670761006103925155545718915466042130740983389742948846660834504613593991983705327051718476170638133152807011690543806383948638093932606654786720558559157082888329454283798033982562109241315194095968266273311223087132130092902383845412E-9', '*']);
  TestNumFmt('1/123456789', nfExponent, 500, ['!', '8.10000007371000067076100610392515554571891546604213074098338974294884666083450461359399198370532705171847617063813315280701169054380638394863809393260665478672055855915708288832945428379803398256210924131519409596826627331122308713213009290238384541169'+
  '29932464062385422967707349006136875955845571198194697903571750922502933394776693892467914421458021235267993240938738492542520282136934567446104563759551530211918924928462216849006173325956177266201213122431039414122458668514373883480802339675301291E-9', '*']);

  // --

  TestNumFmt('1234567890123/2', nfDefault, ['!', '617283945062']);
  TestNumFmt('1234567890123/2', nfDefault,  1, ['!', '6E11', '6.2E11']);
  TestNumFmt('1234567890123/2', nfDefault,  2, ['!', '6.2E11']);
  TestNumFmt('1234567890123/2', nfDefault,  3, ['!', '6.17E11']);
  TestNumFmt('1234567890123/2', nfDefault,  4, ['!', '6.173E11']);
  TestNumFmt('1234567890123/2', nfDefault,  5, ['!', '6.1728E11']);
  TestNumFmt('1234567890123/2', nfDefault,  6, ['!', '6.17284E11']);
  TestNumFmt('1234567890123/2', nfDefault,  7, ['!', '6.172839E11']);
  TestNumFmt('1234567890123/2', nfDefault,  8, ['!', '6.1728395E11']);
  TestNumFmt('1234567890123/2', nfDefault,  9, ['!', '6.17283945E11']);
  TestNumFmt('1234567890123/2', nfDefault, 10, ['!', '6.172839451E11']);
  TestNumFmt('1234567890123/2', nfDefault, 11, ['!', '6.1728394506E11']);
  TestNumFmt('1234567890123/2', nfDefault, 12, ['!', '617283945062']);
  TestNumFmt('1234567890123/2', nfDefault, 13, ['!', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfDefault, 14, ['!', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfDefault, 15, ['!', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfDefault, 20, ['!', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfDefault, 100, ['!', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfDefault, 1000, ['!', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfDefault, 5000, ['!', '617283945061.5']);

  TestNumFmt('1234567890123/2', nfFraction, ['!', '1234567890123/2', '617283945062']);
  TestNumFmt('1234567890123/2', nfFraction,  1, ['!', '1234567890123/2', '6.2E11']);
  TestNumFmt('1234567890123/2', nfFraction,  2, ['!', '1234567890123/2', '6.2E11']);
  TestNumFmt('1234567890123/2', nfFraction,  3, ['!', '1234567890123/2', '6.17E11']);
  TestNumFmt('1234567890123/2', nfFraction,  4, ['!', '1234567890123/2', '6.173E11']);
  TestNumFmt('1234567890123/2', nfFraction,  5, ['!', '1234567890123/2', '6.1728E11']);
  TestNumFmt('1234567890123/2', nfFraction,  6, ['!', '1234567890123/2', '6.17284E11']);
  TestNumFmt('1234567890123/2', nfFraction,  7, ['!', '1234567890123/2', '6.172839E11']);
  TestNumFmt('1234567890123/2', nfFraction,  8, ['!', '1234567890123/2', '6.1728395E11']);
  TestNumFmt('1234567890123/2', nfFraction,  9, ['!', '1234567890123/2', '6.17283945E11']);
  TestNumFmt('1234567890123/2', nfFraction, 10, ['!', '1234567890123/2', '6.172839451E11']);
  TestNumFmt('1234567890123/2', nfFraction, 11, ['!', '1234567890123/2', '6.1728394506E11']);
  TestNumFmt('1234567890123/2', nfFraction, 12, ['!', '1234567890123/2', '617283945062']);
  TestNumFmt('1234567890123/2', nfFraction, 13, ['!', '1234567890123/2', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfFraction, 14, ['!', '1234567890123/2', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfFraction, 15, ['!', '1234567890123/2', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfFraction, 20, ['!', '1234567890123/2', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfFraction, 100, ['!', '1234567890123/2', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfFraction, 1000, ['!', '1234567890123/2', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfFraction, 5000, ['!', '1234567890123/2', '617283945061.5']);

  TestNumFmt('1234567890123/2', nfFixed, ['!', '617283945061.500000000000']);
  TestNumFmt('1234567890123/2', nfFixed,  0, ['!', '617283945062']);
  TestNumFmt('1234567890123/2', nfFixed,  1, ['!', '617283945061.5']);
  TestNumFmt('1234567890123/2', nfFixed,  2, ['!', '617283945061.50']);
  TestNumFmt('1234567890123/2', nfFixed,  3, ['!', '617283945061.500']);
  TestNumFmt('1234567890123/2', nfFixed,  4, ['!', '617283945061.5000']);
  TestNumFmt('1234567890123/2', nfFixed,  5, ['!', '617283945061.50000']);
  TestNumFmt('1234567890123/2', nfFixed,  6, ['!', '617283945061.500000']);
  TestNumFmt('1234567890123/2', nfFixed,  7, ['!', '617283945061.5000000']);
  TestNumFmt('1234567890123/2', nfFixed,  8, ['!', '617283945061.50000000']);
  TestNumFmt('1234567890123/2', nfFixed,  9, ['!', '617283945061.500000000']);
  TestNumFmt('1234567890123/2', nfFixed, 10, ['!', '617283945061.5000000000']);
  TestNumFmt('1234567890123/2', nfFixed, 11, ['!', '617283945061.50000000000']);
  TestNumFmt('1234567890123/2', nfFixed, 12, ['!', '617283945061.500000000000']);
  TestNumFmt('1234567890123/2', nfFixed, 20, ['!', '617283945061.50000000000000000000', '*']);

  TestNumFmt('1234567890123/2', nfExponent, ['!', '6.17283945062E11']);
  TestNumFmt('1234567890123/2', nfExponent,  1, ['!', '6E11', '6.2E11']);
  TestNumFmt('1234567890123/2', nfExponent,  2, ['!', '6.2E11']);
  TestNumFmt('1234567890123/2', nfExponent,  3, ['!', '6.17E11']);
  TestNumFmt('1234567890123/2', nfExponent,  4, ['!', '6.173E11']);
  TestNumFmt('1234567890123/2', nfExponent,  5, ['!', '6.1728E11']);
  TestNumFmt('1234567890123/2', nfExponent,  6, ['!', '6.17284E11']);
  TestNumFmt('1234567890123/2', nfExponent,  7, ['!', '6.172839E11']);
  TestNumFmt('1234567890123/2', nfExponent,  8, ['!', '6.1728395E11']);
  TestNumFmt('1234567890123/2', nfExponent,  9, ['!', '6.17283945E11']);
  TestNumFmt('1234567890123/2', nfExponent, 10, ['!', '6.172839451E11']);
  TestNumFmt('1234567890123/2', nfExponent, 11, ['!', '6.1728394506E11']);
  TestNumFmt('1234567890123/2', nfExponent, 12, ['!', '6.17283945062E11']);
  TestNumFmt('1234567890123/2', nfExponent, 13, ['!', '6.172839450615E11']);
  TestNumFmt('1234567890123/2', nfExponent, 14, ['!', '6.1728394506150E11']);
  TestNumFmt('1234567890123/2', nfExponent, 15, ['!', '6.17283945061500E11']);
  TestNumFmt('1234567890123/2', nfExponent, 20, ['!', '6.1728394506150000000E11', '*']);

  // --

  TestNumFmt('123456789012345', nfDefault, ['123456789012345', '123456789012345', '1.23456789012E14']);
  TestNumFmt('123456789012345', nfDefault,  1, ['123456789012345', '123456789012345', '1.2E14']);
  TestNumFmt('123456789012345', nfDefault,  2, ['123456789012345', '123456789012345', '1.2E14']);
  TestNumFmt('123456789012345', nfDefault,  3, ['123456789012345', '123456789012345', '1.23E14']);
  TestNumFmt('123456789012345', nfDefault,  4, ['123456789012345', '123456789012345', '1.235E14']);
  TestNumFmt('123456789012345', nfDefault,  5, ['123456789012345', '123456789012345', '1.2346E14']);
  TestNumFmt('123456789012345', nfDefault,  6, ['123456789012345', '123456789012345', '1.23457E14']);
  TestNumFmt('123456789012345', nfDefault,  7, ['123456789012345', '123456789012345', '1.234568E14']);
  TestNumFmt('123456789012345', nfDefault,  8, ['123456789012345', '123456789012345', '1.2345679E14']);
  TestNumFmt('123456789012345', nfDefault,  9, ['123456789012345', '123456789012345', '1.23456789E14']);
  TestNumFmt('123456789012345', nfDefault, 10, ['123456789012345', '123456789012345', '1.23456789E14']);
  TestNumFmt('123456789012345', nfDefault, 11, ['123456789012345', '123456789012345', '1.2345678901E14']);
  TestNumFmt('123456789012345', nfDefault, 12, ['123456789012345', '123456789012345', '1.23456789012E14']);
  TestNumFmt('123456789012345', nfDefault, 13, ['123456789012345', '123456789012345', '1.234567890123E14']);
  TestNumFmt('123456789012345', nfDefault, 14, ['123456789012345', '123456789012345', '1.2345678901235E14']);
  TestNumFmt('123456789012345', nfDefault, 15, ['123456789012345', '123456789012345', '123456789012345']);

  TestNumFmt('123456789012345', nfFraction, ['123456789012345', '123456789012345', '1.23456789012E14']);
  TestNumFmt('123456789012345', nfFraction,  1, ['123456789012345', '123456789012345', '1.2E14']);
  TestNumFmt('123456789012345', nfFraction,  2, ['123456789012345', '123456789012345', '1.2E14']);
  TestNumFmt('123456789012345', nfFraction,  3, ['123456789012345', '123456789012345', '1.23E14']);
  TestNumFmt('123456789012345', nfFraction,  4, ['123456789012345', '123456789012345', '1.235E14']);
  TestNumFmt('123456789012345', nfFraction,  5, ['123456789012345', '123456789012345', '1.2346E14']);
  TestNumFmt('123456789012345', nfFraction,  6, ['123456789012345', '123456789012345', '1.23457E14']);
  TestNumFmt('123456789012345', nfFraction,  7, ['123456789012345', '123456789012345', '1.234568E14']);
  TestNumFmt('123456789012345', nfFraction,  8, ['123456789012345', '123456789012345', '1.2345679E14']);
  TestNumFmt('123456789012345', nfFraction,  9, ['123456789012345', '123456789012345', '1.23456789E14']);
  TestNumFmt('123456789012345', nfFraction, 10, ['123456789012345', '123456789012345', '1.23456789E14']);
  TestNumFmt('123456789012345', nfFraction, 11, ['123456789012345', '123456789012345', '1.2345678901E14']);
  TestNumFmt('123456789012345', nfFraction, 12, ['123456789012345', '123456789012345', '1.23456789012E14']);
  TestNumFmt('123456789012345', nfFraction, 13, ['123456789012345', '123456789012345', '1.234567890123E14']);
  TestNumFmt('123456789012345', nfFraction, 14, ['123456789012345', '123456789012345', '1.2345678901235E14']);
  TestNumFmt('123456789012345', nfFraction, 15, ['123456789012345', '123456789012345', '123456789012345']);

  TestNumFmt('123456789012345', nfFixed, ['123456789012345', '123456789012345.000000000000', '123456789012345.000000000000']);
  TestNumFmt('123456789012345', nfFixed,  0, ['123456789012345', '123456789012345', '123456789012345']);
  TestNumFmt('123456789012345', nfFixed,  1, ['123456789012345', '123456789012345.0', '123456789012345.0']);
  TestNumFmt('123456789012345', nfFixed,  2, ['123456789012345', '123456789012345.00', '123456789012345.00']);
  TestNumFmt('123456789012345', nfFixed,  3, ['123456789012345', '123456789012345.000', '123456789012345.000']);
  TestNumFmt('123456789012345', nfFixed,  4, ['123456789012345', '123456789012345.0000', '123456789012345.0000']);
  TestNumFmt('123456789012345', nfFixed,  5, ['123456789012345', '123456789012345.00000', '123456789012345.00000']);
  TestNumFmt('123456789012345', nfFixed,  6, ['123456789012345', '123456789012345.000000', '123456789012345.000000']);
  TestNumFmt('123456789012345', nfFixed,  7, ['123456789012345', '123456789012345.0000000', '123456789012345.0000000']);
  TestNumFmt('123456789012345', nfFixed,  8, ['123456789012345', '123456789012345.00000000', '123456789012345.00000000']);
  TestNumFmt('123456789012345', nfFixed,  9, ['123456789012345', '123456789012345.000000000', '123456789012345.000000000']);
  TestNumFmt('123456789012345', nfFixed, 10, ['123456789012345', '123456789012345.0000000000', '123456789012345.0000000000']);
  TestNumFmt('123456789012345', nfFixed, 11, ['123456789012345', '123456789012345.00000000000', '123456789012345.00000000000']);
  TestNumFmt('123456789012345', nfFixed, 12, ['123456789012345', '123456789012345.000000000000', '123456789012345.000000000000']);
  TestNumFmt('123456789012345', nfFixed, 13, ['123456789012345', '123456789012345.0000000000000', '123456789012345.0000000000000']);
  TestNumFmt('123456789012345', nfFixed, 14, ['123456789012345', '123456789012345.00000000000000', '123456789012345.00000000000000']);
  TestNumFmt('123456789012345', nfFixed, 15, ['123456789012345', '123456789012345.000000000000000', '123456789012345.000000000000000']);

  TestNumFmt('123456789012345', nfExponent, ['123456789012345', '1.23456789012E14']);
  TestNumFmt('123456789012345', nfExponent,  1, ['123456789012345', '1E14', '1.2E14']);
  TestNumFmt('123456789012345', nfExponent,  2, ['123456789012345', '1.2E14']);
  TestNumFmt('123456789012345', nfExponent,  3, ['123456789012345', '1.23E14']);
  TestNumFmt('123456789012345', nfExponent,  4, ['123456789012345', '1.235E14']);
  TestNumFmt('123456789012345', nfExponent,  5, ['123456789012345', '1.2346E14']);
  TestNumFmt('123456789012345', nfExponent,  6, ['123456789012345', '1.23457E14']);
  TestNumFmt('123456789012345', nfExponent,  7, ['123456789012345', '1.234568E14']);
  TestNumFmt('123456789012345', nfExponent,  8, ['123456789012345', '1.2345679E14']);
  TestNumFmt('123456789012345', nfExponent,  9, ['123456789012345', '1.23456789E14']);
  TestNumFmt('123456789012345', nfExponent, 10, ['123456789012345', '1.234567890E14']);
  TestNumFmt('123456789012345', nfExponent, 11, ['123456789012345', '1.2345678901E14']);
  TestNumFmt('123456789012345', nfExponent, 12, ['123456789012345', '1.23456789012E14']);
  TestNumFmt('123456789012345', nfExponent, 13, ['123456789012345', '1.234567890123E14']);
  TestNumFmt('123456789012345', nfExponent, 14, ['123456789012345', '1.2345678901235E14']);
  TestNumFmt('123456789012345', nfExponent, 15, ['123456789012345', '1.23456789012345E14']);
  TestNumFmt('123456789012345', nfExponent, 16, ['123456789012345', '1.234567890123450E14', '*']);
  TestNumFmt('123456789012345', nfExponent, 17, ['123456789012345', '1.2345678901234500E14', '*']);
  TestNumFmt('123456789012345', nfExponent, 18, ['123456789012345', '1.23456789012345000E14', '*']);
  TestNumFmt('123456789012345', nfExponent, 19, ['123456789012345', '1.234567890123450000E14', '*']);
  TestNumFmt('123456789012345', nfExponent, 20, ['123456789012345', '1.2345678901234500000E14', '*']);

  // --

  TestNumFmt('123456789012345/2', nfDefault, ['!', '6.17283945062E13']);
  TestNumFmt('123456789012345/2', nfDefault,  1, ['!', '6E13', '6.2E13']);
  TestNumFmt('123456789012345/2', nfDefault,  2, ['!', '6.2E13']);
  TestNumFmt('123456789012345/2', nfDefault,  3, ['!', '6.17E13']);
  TestNumFmt('123456789012345/2', nfDefault,  4, ['!', '6.173E13']);
  TestNumFmt('123456789012345/2', nfDefault,  5, ['!', '6.1728E13']);
  TestNumFmt('123456789012345/2', nfDefault,  6, ['!', '6.17284E13']);
  TestNumFmt('123456789012345/2', nfDefault,  7, ['!', '6.172839E13']);
  TestNumFmt('123456789012345/2', nfDefault,  8, ['!', '6.1728395E13']);
  TestNumFmt('123456789012345/2', nfDefault,  9, ['!', '6.17283945E13']);
  TestNumFmt('123456789012345/2', nfDefault, 10, ['!', '6.172839451E13']);
  TestNumFmt('123456789012345/2', nfDefault, 11, ['!', '6.1728394506E13']);
  TestNumFmt('123456789012345/2', nfDefault, 12, ['!', '6.17283945062E13']);
  TestNumFmt('123456789012345/2', nfDefault, 13, ['!', '6.172839450617E13']);
  TestNumFmt('123456789012345/2', nfDefault, 14, ['!', '61728394506173']);
  TestNumFmt('123456789012345/2', nfDefault, 15, ['!', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfDefault, 16, ['!', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfDefault, 17, ['!', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfDefault, 18, ['!', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfDefault, 19, ['!', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfDefault, 20, ['!', '61728394506172.5']);

  TestNumFmt('123456789012345/2', nfFraction, ['!', '123456789012345/2', '6.17283945062E13']);
  TestNumFmt('123456789012345/2', nfFraction,  1, ['!', '123456789012345/2', '6.2E13']);
  TestNumFmt('123456789012345/2', nfFraction,  2, ['!', '123456789012345/2', '6.2E13']);
  TestNumFmt('123456789012345/2', nfFraction,  3, ['!', '123456789012345/2', '6.17E13']);
  TestNumFmt('123456789012345/2', nfFraction,  4, ['!', '123456789012345/2', '6.173E13']);
  TestNumFmt('123456789012345/2', nfFraction,  5, ['!', '123456789012345/2', '6.1728E13']);
  TestNumFmt('123456789012345/2', nfFraction,  6, ['!', '123456789012345/2', '6.17284E13']);
  TestNumFmt('123456789012345/2', nfFraction,  7, ['!', '123456789012345/2', '6.172839E13']);
  TestNumFmt('123456789012345/2', nfFraction,  8, ['!', '123456789012345/2', '6.1728395E13']);
  TestNumFmt('123456789012345/2', nfFraction,  9, ['!', '123456789012345/2', '6.17283945E13']);
  TestNumFmt('123456789012345/2', nfFraction, 10, ['!', '123456789012345/2', '6.172839451E13']);
  TestNumFmt('123456789012345/2', nfFraction, 11, ['!', '123456789012345/2', '6.1728394506E13']);
  TestNumFmt('123456789012345/2', nfFraction, 12, ['!', '123456789012345/2', '6.17283945062E13']);
  TestNumFmt('123456789012345/2', nfFraction, 13, ['!', '123456789012345/2', '6.172839450617E13']);
  TestNumFmt('123456789012345/2', nfFraction, 14, ['!', '123456789012345/2', '61728394506173']);
  TestNumFmt('123456789012345/2', nfFraction, 15, ['!', '123456789012345/2', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfFraction, 16, ['!', '123456789012345/2', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfFraction, 17, ['!', '123456789012345/2', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfFraction, 18, ['!', '123456789012345/2', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfFraction, 19, ['!', '123456789012345/2', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfFraction, 20, ['!', '123456789012345/2', '61728394506172.5']);

  TestNumFmt('123456789012345/2', nfFixed, ['!', '61728394506172.500000000000']);
  TestNumFmt('123456789012345/2', nfFixed,  0, ['!', '61728394506173']);
  TestNumFmt('123456789012345/2', nfFixed,  1, ['!', '61728394506172.5']);
  TestNumFmt('123456789012345/2', nfFixed,  2, ['!', '61728394506172.50']);
  TestNumFmt('123456789012345/2', nfFixed,  3, ['!', '61728394506172.500']);
  TestNumFmt('123456789012345/2', nfFixed,  4, ['!', '61728394506172.5000']);
  TestNumFmt('123456789012345/2', nfFixed,  5, ['!', '61728394506172.50000']);
  TestNumFmt('123456789012345/2', nfFixed,  6, ['!', '61728394506172.500000']);
  TestNumFmt('123456789012345/2', nfFixed,  7, ['!', '61728394506172.5000000']);
  TestNumFmt('123456789012345/2', nfFixed,  8, ['!', '61728394506172.50000000']);
  TestNumFmt('123456789012345/2', nfFixed,  9, ['!', '61728394506172.500000000']);
  TestNumFmt('123456789012345/2', nfFixed, 10, ['!', '61728394506172.5000000000']);
  TestNumFmt('123456789012345/2', nfFixed, 11, ['!', '61728394506172.50000000000']);
  TestNumFmt('123456789012345/2', nfFixed, 12, ['!', '61728394506172.500000000000']);
  TestNumFmt('123456789012345/2', nfFixed, 13, ['!', '61728394506172.5000000000000']);
  TestNumFmt('123456789012345/2', nfFixed, 14, ['!', '61728394506172.50000000000000']);
  TestNumFmt('123456789012345/2', nfFixed, 15, ['!', '61728394506172.500000000000000']);
  TestNumFmt('123456789012345/2', nfFixed, 16, ['!', '61728394506172.5000000000000000', '*']);
  TestNumFmt('123456789012345/2', nfFixed, 17, ['!', '61728394506172.50000000000000000', '*']);
  TestNumFmt('123456789012345/2', nfFixed, 18, ['!', '61728394506172.500000000000000000', '*']);
  TestNumFmt('123456789012345/2', nfFixed, 19, ['!', '61728394506172.5000000000000000000', '*']);
  TestNumFmt('123456789012345/2', nfFixed, 20, ['!', '61728394506172.50000000000000000000', '*']);

  TestNumFmt('123456789012345/2', nfExponent, ['!', '6.17283945062E13']);
  TestNumFmt('123456789012345/2', nfExponent, 1, ['!', '6E13', '6.2E13']);
  TestNumFmt('123456789012345/2', nfExponent, 2, ['!', '6.2E13']);
  TestNumFmt('123456789012345/2', nfExponent, 3, ['!', '6.17E13']);
  TestNumFmt('123456789012345/2', nfExponent, 4, ['!', '6.173E13']);
  TestNumFmt('123456789012345/2', nfExponent, 5, ['!', '6.1728E13']);
  TestNumFmt('123456789012345/2', nfExponent, 6, ['!', '6.17284E13']);
  TestNumFmt('123456789012345/2', nfExponent, 7, ['!', '6.172839E13']);
  TestNumFmt('123456789012345/2', nfExponent, 8, ['!', '6.1728395E13']);
  TestNumFmt('123456789012345/2', nfExponent, 9, ['!', '6.17283945E13']);
  TestNumFmt('123456789012345/2', nfExponent, 10, ['!', '6.172839451E13']);
  TestNumFmt('123456789012345/2', nfExponent, 11, ['!', '6.1728394506E13']);
  TestNumFmt('123456789012345/2', nfExponent, 12, ['!', '6.17283945062E13']);
  TestNumFmt('123456789012345/2', nfExponent, 13, ['!', '6.172839450617E13']);
  TestNumFmt('123456789012345/2', nfExponent, 14, ['!', '6.1728394506173E13']);
  TestNumFmt('123456789012345/2', nfExponent, 15, ['!', '6.17283945061725E13']);
  TestNumFmt('123456789012345/2', nfExponent, 16, ['!', '6.172839450617250E13', '*']);
  TestNumFmt('123456789012345/2', nfExponent, 17, ['!', '6.1728394506172500E13', '*']);
  TestNumFmt('123456789012345/2', nfExponent, 18, ['!', '6.17283945061725000E13', '*']);
  TestNumFmt('123456789012345/2', nfExponent, 19, ['!', '6.172839450617250000E13', '*']);
  TestNumFmt('123456789012345/2', nfExponent, 20, ['!', '6.1728394506172500000E13', '*']);

  _i := RandSeed;
  try
    RandSeed := 1482311;
    for _j := 1 to 20000 do
      Test(
        'x ≔ RandomInt(-100000000, 100000000)/RandomInt(1, 100000000); ' +
        'for(d, 0, 4, for(k, 2, 10, if(' +
        '  AsSingleLine(x\k) = AsSingleLine(RealNumber(x)\k) ' +
        '∧ AsSingleLine(SetDigitGrouping(SetNumberFormat(x\k, "default"), d)) = AsSingleLine(SetDigitGrouping(SetNumberFormat(RealNumber(x), "default")\k, d)) ' +
        '∧ AsSingleLine(SetDigitGrouping(SetNumberFormat(x\k, "fixed"), d)) = AsSingleLine(SetDigitGrouping(SetNumberFormat(RealNumber(x), "fixed")\k, d)) ' +
        '∧ AsSingleLine(SetDigitGrouping(SetNumberFormat(x\k, "scientific"), d)) = AsSingleLine(SetDigitGrouping(SetNumberFormat(RealNumber(x), "scientific")\k, d)), 0, fail(string(x)))))',
      null);
      Test(
        '  if(RealNumber(AsSingleLine(x\100)) ≈ x' +
        '∧ RealNumber(AsSingleLine(SetNumberFormat(x, "default")\100)) ≈ x' +
        '∧ RealNumber(AsSingleLine(SetNumberFormat(x, "fixed")\100)) ≈ x' +
        '∧ RealNumber(AsSingleLine(SetNumberFormat(SetPrettyExp(x, false), "scientific")\100)) ≈ x, true, x)',
      True);
  finally
    RandSeed := _i;
  end;


  //
  // Digits
  //

  Test('digits(5321407)', intarr([5, 3, 2, 1, 4, 0, 7]));
  Test('digits(−28441105319810423)', intarr([2, 8, 4, 4, 1, 1, 0,5 , 3, 1, 9, 8, 1, 0, 4, 2, 3]));
  Test('digits(0)', intarr([0]));

  Test('digits(1/4)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr([2, 5]))
      ]
    ));

  Test('digits(1/2)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr([5]))
      ]
    ));

  Test('digits(1/10)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr([1]))
      ]
    ));

  Test('digits(1/100)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr([0, 1]))
      ]
    ));

  Test('digits(15327/10)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([1, 5, 3, 2])),
        sm('frac', intarr([7]))
      ]
    ));

  Test('digits(15327632/10000)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([1, 5, 3, 2])),
        sm('frac', intarr([7, 6, 3, 2]))
      ]
    ));

  Test('digits(1/3)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr([3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3]))
      ]
    ));

  Test('digits(100/3)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([3, 3])),
        sm('frac', intarr([3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3]))
      ]
    ));

  Test('digits(1/123456789, 4160)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr(
          [
            0, 0, 0, 0, 0, 0, 0, 0, 8, 1, 0, 0, 0, 0, 0, 0, 7, 3, 7, 1, 0, 0, 0, 0, 6, 7, 0,
            7, 6, 1, 0, 0, 6, 1, 0, 3, 9, 2, 5, 1, 5, 5, 5, 4, 5, 7, 1, 8, 9, 1, 5, 4, 6, 6,
            0, 4, 2, 1, 3, 0, 7, 4, 0, 9, 8, 3, 3, 8, 9, 7, 4, 2, 9, 4, 8, 8, 4, 6, 6, 6, 0,
            8, 3, 4, 5, 0, 4, 6, 1, 3, 5, 9, 3, 9, 9, 1, 9, 8, 3, 7, 0, 5, 3, 2, 7, 0, 5, 1,
            7, 1, 8, 4, 7, 6, 1, 7, 0, 6, 3, 8, 1, 3, 3, 1, 5, 2, 8, 0, 7, 0, 1, 1, 6, 9, 0,
            5, 4, 3, 8, 0, 6, 3, 8, 3, 9, 4, 8, 6, 3, 8, 0, 9, 3, 9, 3, 2, 6, 0, 6, 6, 5, 4,
            7, 8, 6, 7, 2, 0, 5, 5, 8, 5, 5, 9, 1, 5, 7, 0, 8, 2, 8, 8, 8, 3, 2, 9, 4, 5, 4,
            2, 8, 3, 7, 9, 8, 0, 3, 3, 9, 8, 2, 5, 6, 2, 1, 0, 9, 2, 4, 1, 3, 1, 5, 1, 9, 4,
            0, 9, 5, 9, 6, 8, 2, 6, 6, 2, 7, 3, 3, 1, 1, 2, 2, 3, 0, 8, 7, 1, 3, 2, 1, 3, 0,
            0, 9, 2, 9, 0, 2, 3, 8, 3, 8, 4, 5, 4, 1, 1, 6, 9, 2, 9, 9, 3, 2, 4, 6, 4, 0, 6,
            2, 3, 8, 5, 4, 2, 2, 9, 6, 7, 7, 0, 7, 3, 4, 9, 0, 0, 6, 1, 3, 6, 8, 7, 5, 9, 5,
            5, 8, 4, 5, 5, 7, 1, 1, 9, 8, 1, 9, 4, 6, 9, 7, 9, 0, 3, 5, 7, 1, 7, 5, 0, 9, 2,
            2, 5, 0, 2, 9, 3, 3, 3, 9, 4, 7, 7, 6, 6, 9, 3, 8, 9, 2, 4, 6, 7, 9, 1, 4, 4, 2,
            1, 4, 5, 8, 0, 2, 1, 2, 3, 5, 2, 6, 7, 9, 9, 3, 2, 4, 0, 9, 3, 8, 7, 3, 8, 4, 9,
            2, 5, 4, 2, 5, 2, 0, 2, 8, 2, 1, 3, 6, 9, 3, 4, 5, 6, 7, 4, 4, 6, 1, 0, 4, 5, 6,
            3, 7, 5, 9, 5, 5, 1, 5, 3, 0, 2, 1, 1, 9, 1, 8, 9, 2, 4, 9, 2, 8, 4, 6, 2, 2, 1,
            6, 8, 4, 9, 0, 0, 6, 1, 7, 3, 3, 2, 5, 9, 5, 6, 1, 7, 7, 2, 6, 6, 2, 0, 1, 2, 1,
            3, 1, 2, 2, 4, 3, 1, 0, 3, 9, 4, 1, 4, 1, 2, 2, 4, 5, 8, 6, 6, 8, 5, 1, 4, 3, 7,
            3, 8, 8, 3, 4, 8, 0, 8, 0, 2, 3, 3, 9, 6, 7, 5, 3, 0, 1, 2, 9, 1, 0, 4, 5, 2, 4,
            1, 7, 4, 8, 5, 1, 1, 6, 9, 9, 9, 1, 1, 4, 5, 6, 4, 6, 9, 1, 9, 4, 2, 5, 3, 8, 6,
            9, 6, 6, 7, 7, 1, 0, 2, 1, 3, 9, 7, 6, 1, 6, 2, 9, 4, 7, 1, 8, 3, 0, 8, 2, 8, 1,
            9, 3, 6, 6, 0, 5, 3, 6, 5, 6, 2, 3, 1, 0, 8, 8, 2, 7, 1, 7, 0, 2, 9, 0, 3, 2, 7,
            2, 4, 9, 6, 4, 1, 9, 7, 7, 9, 7, 1, 7, 4, 1, 9, 9, 9, 5, 4, 2, 8, 5, 2, 1, 9, 5,
            8, 3, 9, 9, 5, 4, 9, 8, 2, 1, 4, 3, 5, 9, 0, 3, 3, 7, 5, 0, 6, 6, 7, 2, 0, 7, 1,
            3, 1, 0, 7, 1, 5, 8, 4, 8, 9, 2, 7, 5, 1, 4, 2, 2, 5, 2, 4, 0, 3, 7, 9, 4, 4, 9,
            6, 8, 7, 4, 5, 2, 9, 9, 2, 1, 5, 5, 8, 2, 2, 2, 2, 8, 6, 1, 7, 9, 8, 2, 2, 8, 0,
            4, 2, 3, 6, 3, 8, 7, 5, 1, 8, 5, 5, 1, 1, 2, 6, 4, 1, 8, 8, 1, 5, 2, 5, 0, 4, 1,
            1, 2, 1, 8, 7, 7, 8, 7, 4, 2, 0, 9, 0, 8, 8, 6, 5, 5, 3, 0, 2, 7, 0, 6, 7, 6, 3,
            2, 5, 4, 6, 3, 1, 5, 4, 5, 6, 1, 7, 1, 4, 7, 0, 6, 5, 1, 1, 6, 0, 3, 8, 2, 9, 2,
            5, 5, 5, 9, 4, 8, 4, 6, 2, 2, 5, 9, 1, 3, 1, 0, 0, 6, 5, 5, 8, 0, 9, 2, 1, 5, 9,
            6, 7, 8, 6, 3, 8, 6, 5, 3, 0, 7, 5, 6, 1, 1, 7, 4, 2, 9, 8, 8, 0, 6, 6, 8, 6, 1,
            1, 9, 1, 4, 0, 8, 4, 3, 6, 8, 4, 1, 8, 1, 6, 7, 7, 5, 2, 6, 0, 5, 3, 2, 6, 5, 4,
            8, 7, 0, 8, 4, 7, 1, 5, 9, 3, 2, 4, 7, 0, 9, 1, 4, 9, 8, 5, 4, 8, 5, 3, 2, 6, 3,
            6, 7, 9, 1, 6, 4, 6, 9, 9, 4, 8, 0, 3, 9, 8, 7, 6, 5, 2, 7, 1, 6, 2, 8, 7, 6, 3,
            9, 7, 1, 8, 2, 1, 7, 5, 2, 1, 4, 3, 5, 7, 7, 9, 4, 4, 5, 0, 6, 5, 5, 9, 2, 9, 5,
            0, 0, 9, 6, 8, 9, 5, 8, 4, 5, 8, 8, 1, 7, 5, 2, 1, 9, 7, 5, 2, 3, 9, 4, 4, 9, 9,
            7, 4, 6, 7, 8, 9, 9, 4, 7, 6, 9, 5, 7, 8, 8, 5, 2, 4, 0, 3, 1, 6, 7, 5, 5, 6, 8,
            6, 8, 8, 2, 4, 7, 6, 7, 5, 0, 6, 3, 0, 5, 3, 8, 4, 3, 0, 7, 3, 7, 8, 9, 9, 7, 1,
            9, 7, 1, 4, 8, 8, 7, 4, 4, 9, 4, 0, 5, 4, 7, 5, 7, 8, 9, 5, 8, 9, 8, 2, 9, 6, 8,
            5, 2, 6, 7, 4, 5, 0, 1, 3, 5, 9, 3, 3, 7, 9, 6, 2, 3, 6, 9, 9, 7, 5, 4, 5, 7, 5,
            6, 6, 7, 7, 6, 6, 6, 3, 8, 5, 7, 6, 6, 7, 6, 4, 1, 1, 0, 4, 7, 7, 5, 5, 3, 4, 0,
            5, 3, 4, 5, 7, 3, 5, 9, 8, 8, 6, 4, 6, 1, 9, 7, 4, 9, 6, 6, 8, 0, 3, 9, 7, 2, 1,
            9, 7, 9, 1, 6, 1, 4, 7, 0, 0, 1, 0, 3, 6, 9, 3, 7, 7, 0, 9, 4, 3, 6, 1, 3, 3, 1,
            5, 5, 8, 6, 8, 8, 1, 1, 7, 1, 8, 4, 0, 6, 1, 8, 6, 6, 3, 7, 4, 9, 6, 2, 9, 8, 4,
            0, 1, 2, 1, 6, 3, 1, 5, 4, 5, 1, 0, 6, 8, 4, 7, 0, 6, 0, 4, 7, 2, 3, 0, 8, 2, 5,
            0, 2, 9, 8, 0, 0, 5, 0, 7, 7, 7, 1, 1, 8, 4, 6, 2, 0, 7, 1, 7, 7, 8, 0, 0, 4, 8,
            5, 3, 1, 7, 9, 8, 4, 4, 1, 6, 3, 9, 3, 6, 5, 8, 1, 8, 9, 1, 8, 2, 2, 8, 9, 5, 2,
            1, 5, 5, 8, 8, 3, 4, 6, 4, 6, 1, 8, 5, 3, 9, 5, 2, 8, 0, 2, 8, 7, 0, 9, 7, 0, 5,
            0, 6, 1, 2, 5, 8, 3, 1, 6, 0, 5, 7, 4, 5, 0, 6, 7, 6, 1, 2, 2, 8, 0, 1, 1, 5, 2,
            7, 1, 7, 4, 9, 0, 4, 8, 9, 7, 2, 9, 1, 6, 3, 4, 5, 6, 5, 3, 5, 3, 8, 7, 4, 5, 4,
            4, 7, 2, 0, 2, 5, 8, 3, 5, 6, 9, 5, 4, 3, 5, 1, 0, 4, 8, 2, 8, 4, 5, 9, 4, 5, 3,
            9, 3, 8, 9, 8, 1, 0, 3, 0, 8, 4, 4, 7, 2, 7, 3, 8, 0, 6, 8, 7, 0, 1, 9, 1, 6, 4,
            2, 5, 1, 8, 7, 4, 3, 9, 4, 6, 9, 2, 0, 5, 6, 9, 9, 1, 6, 9, 7, 7, 1, 8, 6, 2, 4,
            4, 4, 9, 2, 3, 9, 4, 8, 2, 4, 8, 8, 0, 7, 9, 2, 9, 0, 6, 4, 1, 5, 2, 1, 5, 4, 4,
            8, 3, 7, 8, 4, 6, 0, 5, 8, 0, 2, 4, 3, 9, 9, 1, 2, 8, 0, 2, 2, 0, 3, 2, 0, 6, 5,
            0, 0, 0, 4, 9, 1, 7, 9, 1, 5, 0, 4, 4, 7, 5, 3, 0, 2, 6, 9, 0, 7, 2, 5, 2, 5, 4,
            4, 8, 5, 5, 9, 9, 8, 1, 5, 8, 1, 8, 9, 5, 8, 3, 2, 3, 9, 5, 2, 5, 2, 0, 7, 4, 7,
            9, 6, 7, 9, 3, 8, 8, 0, 6, 5, 0, 8, 2, 4, 3, 1, 3, 9, 2, 2, 5, 0, 1, 2, 5, 6, 6,
            9, 4, 7, 6, 1, 4, 3, 5, 9, 2, 2, 3, 2, 9, 0, 6, 6, 8, 9, 3, 1, 9, 4, 5, 0, 8, 7,
            2, 8, 0, 7, 0, 0, 2, 9, 4, 2, 5, 4, 3, 7, 2, 6, 7, 7, 7, 1, 4, 7, 9, 1, 3, 6, 7,
            2, 0, 4, 6, 0, 1, 4, 4, 1, 5, 6, 1, 8, 7, 3, 1, 1, 8, 2, 1, 3, 0, 4, 5, 3, 7, 5,
            7, 3, 8, 7, 1, 2, 9, 1, 9, 2, 2, 2, 2, 8, 7, 5, 6, 4, 9, 2, 2, 8, 1, 6, 8, 4, 0,
            7, 9, 7, 6, 3, 3, 2, 5, 1, 2, 5, 8, 4, 6, 2, 5, 8, 6, 4, 5, 2, 0, 0, 9, 5, 3, 6,
            7, 1, 3, 2, 8, 6, 7, 8, 4, 0, 9, 0, 9, 0, 9, 7, 3, 5, 2, 2, 7, 2, 7, 8, 5, 9, 0,
            5, 6, 8, 2, 3, 5, 1, 7, 4, 1, 7, 0, 9, 4, 0, 0, 8, 4, 9, 5, 5, 5, 5, 4, 7, 7, 3,
            0, 9, 5, 5, 4, 8, 4, 3, 5, 1, 6, 9, 4, 9, 0, 7, 6, 0, 0, 4, 2, 3, 6, 5, 9, 1, 6,
            3, 8, 5, 5, 2, 9, 8, 3, 9, 1, 0, 8, 3, 2, 1, 5, 3, 5, 8, 8, 5, 7, 2, 5, 9, 7, 6,
            5, 6, 0, 1, 0, 6, 3, 8, 6, 6, 9, 6, 9, 6, 8, 1, 1, 8, 9, 4, 2, 4, 0, 9, 8, 8, 2,
            3, 7, 5, 9, 2, 9, 9, 2, 9, 6, 2, 0, 9, 6, 2, 3, 5, 9, 5, 5, 0, 7, 5, 7, 4, 7, 1,
            9, 1, 1, 8, 9, 2, 9, 9, 4, 3, 9, 8, 2, 2, 6, 2, 4, 9, 0, 2, 3, 8, 5, 8, 8, 6, 6,
            1, 1, 7, 1, 1, 5, 6, 8, 1, 6, 6, 5, 7, 5, 2, 7, 0, 3, 1, 5, 8, 3, 4, 9, 5, 9, 8,
            7, 4, 0, 9, 8, 1, 3, 4, 8, 5, 4, 2, 9, 3, 0, 2, 7, 1, 7, 4, 0, 6, 6, 5, 4, 7, 2,
            8, 4, 0, 0, 5, 5, 8, 0, 2, 8, 4, 4, 5, 0, 7, 8, 0, 5, 8, 8, 5, 0, 2, 1, 0, 3, 3,
            5, 5, 3, 6, 9, 1, 4, 0, 5, 3, 3, 8, 5, 9, 1, 7, 8, 8, 5, 8, 1, 1, 8, 5, 2, 7, 6,
            0, 8, 8, 7, 8, 6, 0, 1, 2, 4, 0, 7, 9, 5, 2, 7, 1, 2, 9, 1, 2, 3, 6, 9, 6, 8, 7,
            5, 0, 2, 5, 6, 4, 1, 5, 6, 2, 7, 3, 3, 3, 3, 8, 2, 2, 0, 8, 7, 3, 3, 7, 7, 8, 0,
            9, 9, 4, 7, 7, 3, 8, 0, 7, 0, 5, 2, 4, 4, 1, 6, 4, 4, 1, 7, 7, 2, 1, 8, 9, 6, 2,
            0, 1, 2, 6, 9, 2, 5, 5, 4, 3, 1, 5, 5, 0, 2, 2, 4, 4, 2, 7, 1, 0, 7, 0, 4, 2, 2,
            8, 6, 6, 7, 4, 0, 8, 4, 8, 0, 8, 7, 3, 4, 1, 7, 1, 7, 5, 9, 4, 8, 0, 9, 6, 3, 0,
            1, 1, 2, 7, 6, 7, 6, 3, 4, 0, 2, 6, 1, 8, 5, 4, 6, 9, 6, 3, 8, 2, 8, 7, 7, 7, 3,
            7, 0, 8, 4, 1, 8, 7, 4, 0, 7, 4, 6, 6, 1, 0, 5, 4, 0, 7, 9, 4, 1, 5, 5, 9, 2, 1,
            2, 2, 6, 8, 1, 8, 8, 8, 3, 1, 6, 4, 0, 5, 1, 8, 3, 6, 7, 9, 2, 8, 7, 1, 7, 1, 4,
            8, 1, 5, 1, 3, 2, 6, 0, 4, 8, 1, 7, 7, 0, 6, 7, 0, 3, 8, 4, 1, 1, 3, 1, 0, 0, 4,
            9, 5, 4, 2, 9, 2, 1, 4, 5, 0, 8, 4, 0, 5, 8, 5, 2, 0, 2, 6, 4, 9, 3, 2, 5, 3, 4,
            4, 1, 0, 8, 8, 6, 0, 6, 3, 1, 3, 9, 0, 6, 3, 1, 7, 4, 5, 6, 5, 4, 7, 4, 8, 8, 8,
            5, 4, 5, 8, 2, 1, 4, 8, 5, 7, 6, 6, 9, 7, 5, 5, 2, 0, 4, 7, 9, 4, 7, 7, 2, 3, 6,
            3, 6, 3, 2, 4, 2, 8, 5, 0, 9, 0, 5, 5, 0, 9, 9, 4, 3, 2, 4, 0, 1, 4, 0, 4, 8, 3,
            4, 8, 5, 2, 7, 8, 3, 9, 9, 7, 1, 6, 0, 3, 3, 4, 3, 7, 4, 1, 5, 9, 0, 4, 2, 8, 0,
            4, 8, 4, 7, 2, 8, 9, 5, 2, 4, 1, 1, 0, 3, 3, 4, 6, 6, 9, 4, 0, 4, 0, 4, 5, 4, 9,
            1, 5, 7, 6, 8, 1, 3, 9, 7, 3, 3, 4, 9, 0, 0, 7, 1, 5, 7, 4, 7, 5, 9, 6, 5, 1, 3,
            3, 0, 3, 1, 2, 8, 2, 7, 1, 0, 5, 8, 4, 6, 7, 2, 6, 6, 6, 3, 2, 0, 5, 2, 1, 2, 6,
            3, 5, 1, 6, 7, 4, 3, 4, 9, 8, 0, 0, 2, 3, 6, 5, 8, 3, 1, 8, 2, 1, 5, 2, 9, 0, 6,
            9, 5, 7, 5, 9, 1, 4, 5, 3, 3, 1, 4, 0, 8, 2, 2, 2, 5, 1, 5, 8, 1, 4, 8, 2, 4, 8,
            9, 3, 9, 1, 4, 9, 0, 6, 5, 3, 4, 6, 2, 5, 6, 4, 9, 4, 6, 5, 0, 9, 3, 4, 1, 0, 1,
            3, 2, 3, 5, 0, 0, 3, 2, 2, 0, 4, 3, 8, 5, 2, 9, 3, 0, 5, 9, 9, 0, 6, 1, 6, 6, 8,
            4, 5, 1, 4, 6, 1, 1, 8, 2, 9, 0, 8, 2, 9, 6, 7, 6, 4, 4, 6, 5, 5, 0, 0, 5, 5, 6,
            6, 3, 6, 0, 5, 5, 0, 6, 5, 3, 8, 8, 1, 0, 1, 0, 9, 5, 0, 3, 1, 7, 1, 9, 9, 6, 4,
            7, 8, 8, 6, 5, 1, 6, 7, 9, 5, 7, 6, 7, 3, 0, 2, 8, 4, 1, 4, 8, 2, 4, 5, 5, 8, 5,
            7, 4, 9, 0, 3, 4, 8, 3, 0, 3, 1, 6, 2, 1, 6, 9, 5, 5, 8, 7, 7, 5, 7, 4, 2, 9, 8,
            4, 8, 5, 9, 2, 6, 1, 1, 6, 2, 2, 1, 9, 2, 7, 6, 5, 7, 6, 1, 9, 5, 4, 1, 6, 8, 4,
            3, 3, 7, 8, 2, 9, 3, 2, 7, 4, 7, 4, 2, 4, 6, 8, 8, 0, 0, 1, 5, 6, 4, 6, 6, 0, 8,
            1, 4, 2, 3, 8, 4, 1, 3, 4, 0, 9, 5, 6, 9, 5, 6, 2, 0, 2, 7, 0, 8, 3, 0, 1, 4, 4,
            4, 6, 4, 5, 5, 4, 3, 1, 4, 6, 2, 7, 4, 4, 4, 2, 6, 3, 1, 0, 9, 7, 4, 2, 7, 9, 4,
            2, 9, 8, 6, 5, 9, 4, 2, 8, 1, 1, 7, 8, 0, 0, 7, 9, 5, 8, 7, 1, 9, 8, 7, 2, 4, 2,
            4, 3, 5, 0, 8, 3, 9, 0, 6, 1, 5, 9, 2, 6, 3, 5, 4, 6, 0, 4, 9, 2, 9, 8, 2, 6, 9,
            0, 4, 8, 6, 1, 4, 2, 4, 8, 3, 4, 2, 3, 8, 9, 6, 5, 9, 9, 1, 5, 7, 4, 5, 9, 0, 5,
            2, 3, 3, 2, 8, 7, 7, 3, 7, 6, 2, 2, 9, 1, 8, 4, 1, 2, 3, 6, 8, 5, 5, 7, 5, 5, 2,
            5, 5, 3, 8, 7, 3, 7, 2, 8, 2, 4, 0, 2, 5, 0, 9, 2, 6, 9, 8, 6, 2, 8, 3, 4, 3, 5,
            5, 7, 5, 1, 7, 9, 2, 6, 3, 7, 3, 4, 1, 3, 1, 2, 9, 9, 9, 8, 0, 5, 9, 4, 8, 2, 9,
            8, 2, 3, 4, 1, 2, 9, 5, 1, 3, 9, 3, 0, 5, 7, 8, 5, 7, 6, 7, 6, 8, 2, 6, 5, 0, 4,
            8, 5, 9, 1, 2, 1, 1, 9, 4, 2, 1, 8, 0, 0, 2, 8, 6, 7, 3, 8, 3, 8, 2, 6, 0, 9, 3,
            1, 9, 2, 8, 1, 7, 4, 4, 8, 0, 5, 4, 6, 3, 8, 7, 7, 7, 2, 9, 7, 2, 1, 2, 8, 7, 3,
            4, 0, 4, 6, 3, 7, 1, 4, 7, 9, 8, 2, 1, 9, 8, 0, 4, 6, 6, 3, 8, 0, 0, 2, 2, 2, 4,
            4, 0, 5, 8, 2, 0, 2, 4, 2, 0, 9, 2, 9, 6, 4, 2, 0, 3, 0, 4, 5, 9, 7, 4, 2, 4, 7,
            7, 1, 8, 3, 6, 5, 6, 5, 4, 2, 3, 7, 1, 2, 7, 4, 5, 3, 5, 5, 7, 8, 5, 9, 8, 2, 7,
            3, 7, 6, 5, 2, 4, 4, 2, 9, 1, 2, 6, 3, 7, 2, 3, 0, 5, 0, 4, 9, 9, 8, 7, 9, 7, 5,
            9, 5, 4, 8, 9, 0, 5, 8, 1, 1, 8, 9, 5, 0, 4, 2, 8, 8, 8, 2, 4, 4, 8, 9, 0, 2, 8,
            3, 0, 2, 8, 5, 0, 1, 5, 7, 5, 5, 5, 9, 3, 6, 4, 3, 3, 7, 5, 9, 0, 2, 1, 5, 4, 7,
            2, 0, 7, 0, 9, 6, 0, 7, 9, 5, 8, 4, 5, 7, 4, 3, 2, 4, 2, 1, 9, 6, 2, 6, 3, 5, 0,
            3, 9, 8, 5, 9, 9, 7, 8, 8, 6, 2, 7, 2, 5, 8, 0, 7, 6, 5, 0, 8, 0, 4, 8, 4, 9, 6,
            2, 2, 3, 2, 4, 1, 3, 1, 5, 6, 3, 1, 4, 9, 5, 9, 7, 2, 2, 4, 6, 6, 1, 3, 3, 4, 7,
            4, 4, 4, 1, 8, 1, 4, 6, 1, 7, 4, 2, 0, 5, 1, 3, 0, 1, 8, 5, 2, 6, 6, 6, 8, 4, 6,
            8, 5, 9, 2, 6, 8, 3, 0, 6, 4, 1, 9, 3, 4, 1, 5, 8, 8, 4, 1, 6, 0, 0, 8, 4, 5, 4,
            5, 8, 5, 6, 7, 6, 9, 3, 6, 7, 2, 9, 6, 6, 0, 1, 2, 4, 2, 3, 9, 9, 0, 7, 1, 3, 0,
            5, 8, 3, 1, 5, 4, 8, 8, 8, 3, 0, 6, 7, 0, 9, 4, 8, 3, 5, 9, 1, 0, 5, 6, 3, 0, 0,
            6, 7, 8, 6, 1, 2, 3, 3, 6, 1, 7, 5, 3, 7, 2, 2, 5, 9, 1, 9, 5, 8, 8, 7, 5, 5, 8,
            6, 8, 2, 5, 7, 6, 7, 8, 4, 0, 1, 1, 4, 4, 8, 7, 3, 4, 5, 0, 4, 1, 8, 3, 4, 8, 3,
            9, 8, 8, 0, 6, 9, 7, 0, 4, 2, 9, 1, 4, 3, 4, 3, 0, 9, 0, 5, 2, 0, 5, 2, 2, 1, 2,
            3, 7, 3, 6, 7, 5, 1, 3, 2, 6, 0, 0, 4, 4, 3, 7, 0, 6, 6, 6, 4, 0, 3, 7, 7, 3, 0,
            6, 4, 2, 7, 4, 3, 3, 4, 8, 8, 4, 8, 9, 6, 4, 4, 7, 4, 5, 2, 5, 5, 7, 6, 7, 1, 8,
            1, 8, 2, 7, 4, 8, 1, 3, 5, 4, 6, 3, 0, 0, 8, 0, 3, 2, 7, 1, 3, 3, 7, 3, 0, 9, 7,
            6, 9, 1, 6, 9, 5, 1, 8, 8, 9, 9, 4, 4, 2, 6, 2, 1, 9, 8, 4, 9, 2, 7, 8, 6, 0, 0,
            6, 2, 8, 4, 3, 5, 2, 6, 5, 7, 1, 8, 7, 6, 0, 9, 1, 8, 0, 4, 0, 7, 2, 4, 3, 5, 4,
            1, 7, 0, 5, 9, 1, 6, 2, 2, 9, 5, 2, 3, 8, 3, 7, 6, 8, 8, 6, 6, 6, 9, 2, 2, 9, 6,
            6, 8, 6, 8, 9, 9, 8, 9, 9, 8, 5, 0, 7, 8, 9, 0, 8, 8, 6, 4, 2, 1, 8, 0, 7, 0, 6,
            6, 4, 3, 8, 4, 4, 4, 3, 0, 4, 5, 8, 9, 8, 4, 3, 1, 7, 1, 7, 6, 7, 5, 7, 2, 8, 6,
            3, 0, 8, 4, 9, 1, 3, 0, 5, 4, 0, 7, 2, 7, 0, 8, 7, 9, 2, 0, 6, 1, 6, 5, 0, 0, 0,
            7, 7, 6, 1, 0, 1, 5, 0, 7, 0, 6, 2, 5, 2, 3, 7, 1, 4, 2, 6, 8, 9, 6, 5, 7, 9, 9,
            8, 4, 7, 5, 8, 8, 7, 7, 8, 6, 1, 3, 0, 5, 7, 8, 8, 5, 3, 7, 8, 8, 2, 6, 7, 5, 6,
            9, 4, 7, 3, 2, 3, 4, 8, 8, 2, 2, 0, 6, 4, 3, 7, 4, 2, 8, 0, 7, 8, 5, 8, 0, 5, 9,
            5, 5, 1, 5, 0, 8, 3, 4, 1, 9, 1, 8, 7, 2, 5, 9, 1, 1, 4, 6, 0, 4, 0, 5, 7, 9, 4,
            2, 8, 9, 6, 9, 2, 7, 2, 8, 0, 3, 6, 2, 0, 3, 8, 2, 5, 1, 2, 9, 4, 5, 4, 8, 0, 8,
            6, 7, 8, 0, 3, 8, 7, 5, 8, 9, 7, 0, 1, 5, 2, 7, 0, 6, 6, 2, 8, 3, 8, 9, 6, 3, 0,
            3, 1, 8, 3, 4, 5, 6, 3, 5, 8, 9, 6, 9, 4, 5, 2, 8, 6, 6, 6, 2, 2, 0, 2, 1, 0, 8,
            6, 2, 6, 0, 3, 9, 1, 8, 8, 4, 9, 6, 9, 5, 6, 6, 1, 5, 3, 2, 2, 3, 0, 5, 1, 9, 9,
            4, 3, 2, 9, 7, 7, 3, 1, 4, 8, 4, 0, 0, 9, 3, 5, 6, 5, 0, 4, 4, 8, 5, 1, 4, 4, 1,
            9, 0, 8, 1, 4, 8, 1, 2, 1, 3, 6, 4, 1, 4, 7, 9, 0, 4, 4, 1, 3, 7, 4, 5, 9, 3, 0,
            1, 6, 5, 0, 8, 7, 9, 6, 4, 5, 0, 2, 3, 0, 0, 4, 7, 6, 9, 7, 0, 9, 3, 4, 3, 4, 0,
            4, 3, 5, 5, 0, 2, 4, 9, 7, 9, 6, 3, 0, 7, 2, 7, 3, 1, 4, 6, 3, 9, 6, 1, 8, 5, 6,
            3, 2, 2, 0, 5, 2, 8, 9, 2, 5, 3, 0, 6, 8, 1, 3, 2, 2, 0, 2, 9, 2, 0, 0, 0, 3, 0,
            4, 6, 5, 7, 2, 0, 2, 7, 7, 2, 3, 8, 0, 5, 4, 5, 2, 2, 8, 6, 6, 2, 9, 6, 1, 5, 8,
            0, 8, 3, 2, 9, 5, 0, 3, 8, 5, 5, 7, 9, 8, 4, 8, 5, 0, 8, 7, 7, 6, 6, 2, 1, 4, 2,
            9, 8, 6, 7, 2, 5, 5, 0, 1, 1, 7, 9, 2, 0, 2, 0, 6, 0, 7, 3, 0, 7, 3, 8, 7, 5, 2,
            6, 4, 9, 7, 2, 2, 6, 4, 9, 1, 1, 2, 4, 7, 6, 1, 0, 6, 9, 2, 3, 5, 3, 2, 5, 7, 3{,
            0, 0}
          ]
        ))
      ]
    ));

  Test('digits(1/123456789, 20)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr(
          [
            0, 0, 0, 0, 0, 0, 0, 0, 8, 1, 0, 0, 0, 0, 0, 0, 7, 3, 7, 1
          ]
        ))
      ]
    ));

  Test('digits(1/123456789, 20, "fractional")',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr(
          [
            0, 0, 0, 0, 0, 0, 0, 0, 8, 1, 0, 0, 0, 0, 0, 0, 7, 3, 7, 1
          ]
        ))
      ]
    ));

  Test('digits(1/123456789, 20, "significant")',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr(
          [
            0, 0, 0, 0, 0, 0, 0, 0, 8, 1, 0, 0, 0, 0, 0, 0, 7, 3, 7, 1, 0, 0, 0, 0, 6, 7, 0,
            7
          ]
        ))
      ]
    ));

  Test('digits(1/613, 100)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr(
          [
            0, 0, 1, 6, 3, 1, 3, 2, 1, 3, 7, 0, 3, 0, 9, 9, 5, 1, 0, 6, 0, 3, 5, 8, 8, 9, 0,
            7, 0, 1, 4, 6, 8, 1, 8, 9, 2, 3, 3, 2, 7, 8, 9, 5, 5, 9, 5, 4, 3, 2, 3, 0, 0, 1,
            6, 3, 1, 3, 2, 1, 3, 7, 0, 3, 0, 9, 9, 5, 1, 0, 6, 0, 3, 5, 8, 8, 9, 0, 7, 0, 1,
            4, 6, 8, 1, 8, 9, 2, 3, 3, 2, 7, 8, 9, 5, 5, 9, 5, 4, 3
          ]
        ))
      ]
    ));

  Test('digits(1/613, 100, "fractional")',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr(
          [
            0, 0, 1, 6, 3, 1, 3, 2, 1, 3, 7, 0, 3, 0, 9, 9, 5, 1, 0, 6, 0, 3, 5, 8, 8, 9, 0,
            7, 0, 1, 4, 6, 8, 1, 8, 9, 2, 3, 3, 2, 7, 8, 9, 5, 5, 9, 5, 4, 3, 2, 3, 0, 0, 1,
            6, 3, 1, 3, 2, 1, 3, 7, 0, 3, 0, 9, 9, 5, 1, 0, 6, 0, 3, 5, 8, 8, 9, 0, 7, 0, 1,
            4, 6, 8, 1, 8, 9, 2, 3, 3, 2, 7, 8, 9, 5, 5, 9, 5, 4, 3
          ]
        ))
      ]
    ));

  Test('digits(1/613, 100, "significant")',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([0])),
        sm('frac', intarr(
          [
            0, 0, 1, 6, 3, 1, 3, 2, 1, 3, 7, 0, 3, 0, 9, 9, 5, 1, 0, 6, 0, 3, 5, 8, 8, 9, 0,
            7, 0, 1, 4, 6, 8, 1, 8, 9, 2, 3, 3, 2, 7, 8, 9, 5, 5, 9, 5, 4, 3, 2, 3, 0, 0, 1,
            6, 3, 1, 3, 2, 1, 3, 7, 0, 3, 0, 9, 9, 5, 1, 0, 6, 0, 3, 5, 8, 8, 9, 0, 7, 0, 1,
            4, 6, 8, 1, 8, 9, 2, 3, 3, 2, 7, 8, 9, 5, 5, 9, 5, 4, 3, 2, 3
          ]
        ))
      ]
    ));

  Test('digits(-37153/7, 50, "fractional")',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([5, 3, 0, 7])),
        sm('frac', intarr(
          [
            5, 7, 1, 4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1,
            4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7
          ]
        ))
      ]
    ));

  Test('digits(37153/7, 50, "significant")',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([5, 3, 0, 7])),
        sm('frac', intarr(
          [
            5, 7, 1, 4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1,
            4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1, 4, 2, 8, 5, 7, 1, 4
          ]
        ))
      ]
    ));

  Test('digits(-91523745821/9731, 100)',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([9, 4, 0, 5, 3, 7, 9])),
        sm('frac', intarr(
          [
            2, 8, 4, 8, 6, 2, 8, 0, 9, 5, 7, 7, 6, 3, 8, 4, 7, 4, 9, 7, 6, 8, 7, 8, 0, 1, 8,
            7, 0, 3, 1, 1, 3, 7, 6, 0, 1, 4, 7, 9, 8, 0, 6, 8, 0, 3, 0, 0, 0, 7, 1, 9, 3, 5,
            0, 5, 2, 9, 2, 3, 6, 4, 6, 0, 7, 9, 5, 3, 9, 6, 1, 5, 6, 6, 1, 2, 8, 8, 6, 6, 5,
            0, 9, 0, 9, 4, 6, 4, 5, 9, 7, 6, 7, 7, 5, 2, 5, 4, 3, 4
          ]
        ))
      ]
    ));

  Test('digits(91523745821/9731, 100, "fractional")',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([9, 4, 0, 5, 3, 7, 9])),
        sm('frac', intarr(
          [
            2, 8, 4, 8, 6, 2, 8, 0, 9, 5, 7, 7, 6, 3, 8, 4, 7, 4, 9, 7, 6, 8, 7, 8, 0, 1, 8,
            7, 0, 3, 1, 1, 3, 7, 6, 0, 1, 4, 7, 9, 8, 0, 6, 8, 0, 3, 0, 0, 0, 7, 1, 9, 3, 5,
            0, 5, 2, 9, 2, 3, 6, 4, 6, 0, 7, 9, 5, 3, 9, 6, 1, 5, 6, 6, 1, 2, 8, 8, 6, 6, 5,
            0, 9, 0, 9, 4, 6, 4, 5, 9, 7, 6, 7, 7, 5, 2, 5, 4, 3, 4
          ]
        ))
      ]
    ));

  Test('digits(-91523745821/9731, 100, "significant")',
    TAlgosimStructure.CreateWithValue
    (
      [
        sm('int', intarr([9, 4, 0, 5, 3, 7, 9])),
        sm('frac', intarr(
          [
            2, 8, 4, 8, 6, 2, 8, 0, 9, 5, 7, 7, 6, 3, 8, 4, 7, 4, 9, 7, 6, 8, 7, 8, 0, 1, 8,
            7, 0, 3, 1, 1, 3, 7, 6, 0, 1, 4, 7, 9, 8, 0, 6, 8, 0, 3, 0, 0, 0, 7, 1, 9, 3, 5,
            0, 5, 2, 9, 2, 3, 6, 4, 6, 0, 7, 9, 5, 3, 9, 6, 1, 5, 6, 6, 1, 2, 8, 8, 6, 6, 5,
            0, 9, 0, 9, 4, 6, 4, 5, 9, 7, 6, 7
          ]
        ))
      ]
    ));


  //
  // Fraction parts
  //

  Test('FractionParts(1/2 - 1/2)', ASORationalNumber(rat(0, 1)));
  Test('FractionParts(1/2)', ASORationalNumber(rat(1, 2)));
  Test('FractionParts(1/3)', ASORationalNumber(rat(1, 3)));
  Test('FractionParts(1/4)', ASORationalNumber(rat(1, 4)));
  Test('FractionParts(1/5)', ASORationalNumber(rat(1, 5)));
  Test('FractionParts(1/6)', ASORationalNumber(rat(1, 6)));
  Test('FractionParts(1/7)', ASORationalNumber(rat(1, 7)));
  Test('FractionParts(1/8)', ASORationalNumber(rat(1, 8)));
  Test('FractionParts(1/9)', ASORationalNumber(rat(1, 9)));
  Test('FractionParts(1/10)', ASORationalNumber(rat(1, 10)));
  Test('FractionParts(1/100)', ASORationalNumber(rat(1, 100)));
  Test('FractionParts(1/10000000)', ASORationalNumber(rat(1, 10000000)));
  Test('FractionParts(2/10000000)', ASORationalNumber(rat(1, 5000000)));
  Test('FractionParts(1/123456789123)', ASORationalNumber(rat(1, 123456789123)));
  Test('FractionParts(2/12345678912345)', ASORationalNumber(rat(2, 12345678912345)));
  Test('FractionParts(1/613)', ASORationalNumber(rat(1, 613)));

  Test('FractionParts(2/3)', ASORationalNumber(rat(2, 3)));
  Test('FractionParts(3/4)', ASORationalNumber(rat(3, 4)));
  Test('FractionParts(57/59)', ASORationalNumber(rat(57, 59)));
  Test('FractionParts(15243978/15243979)', ASORationalNumber(rat(15243978, 15243979)));
  Test('FractionParts(824517365895213/82451757789123)', ASORationalNumber(rat(274839121965071, 27483919263041)));
  Test('FractionParts(824517365895213/2)', ASORationalNumber(rat(824517365895213, 2)));

  Test('FractionParts(-1/2)', ASORationalNumber(rat(-1, 2)));
  Test('FractionParts(-1/3)', ASORationalNumber(rat(-1, 3)));
  Test('FractionParts(-1/4)', ASORationalNumber(rat(-1, 4)));
  Test('FractionParts(-1/5)', ASORationalNumber(rat(-1, 5)));
  Test('FractionParts(-1/6)', ASORationalNumber(rat(-1, 6)));
  Test('FractionParts(-1/7)', ASORationalNumber(rat(-1, 7)));
  Test('FractionParts(-1/8)', ASORationalNumber(rat(-1, 8)));
  Test('FractionParts(-1/9)', ASORationalNumber(rat(-1, 9)));
  Test('FractionParts(-1/10)', ASORationalNumber(rat(-1, 10)));
  Test('FractionParts(-1/100)', ASORationalNumber(rat(-1, 100)));
  Test('FractionParts(-1/10000000)', ASORationalNumber(rat(-1, 10000000)));
  Test('FractionParts(-2/10000000)', ASORationalNumber(rat(-1, 5000000)));
  Test('FractionParts(-1/123456789123)', ASORationalNumber(rat(-1, 123456789123)));
  Test('FractionParts(-2/12345678912345)', ASORationalNumber(rat(-2, 12345678912345)));
  Test('FractionParts(-1/613)', ASORationalNumber(rat(-1, 613)));

  Test('FractionParts(-2/3)', ASORationalNumber(rat(-2, 3)));
  Test('FractionParts(-3/4)', ASORationalNumber(rat(-3, 4)));
  Test('FractionParts(-57/59)', ASORationalNumber(rat(-57, 59)));
  Test('FractionParts(-15243978/15243979)', ASORationalNumber(rat(-15243978, 15243979)));
  Test('FractionParts(-824517365895213/82451757789123)', ASORationalNumber(rat(-274839121965071, 27483919263041)));
  Test('FractionParts(-824517365895213/2)', ASORationalNumber(rat(-824517365895213, 2)));





  //
  //
  //  CHAPTER 18
  //  File system functions
  //
  //

  Chapter('File system functions');

  fn := TPath.Combine(TPath.GetTempPath, 'asfs');
  if TDirectory.Exists(fn) then
    TDirectory.Delete(fn, True);
  TDirectory.CreateDirectory(fn);
  try
    TDirectory.CreateDirectory(TPath.Combine(fn, 'cats'));
    TDirectory.CreateDirectory(TPath.Combine(fn, 'dogs'));
    TDirectory.CreateDirectory(TPath.Combine(fn, 'unicorns'));
    TFile.WriteAllText(TPath.Combine(fn, 'data.txt'), 'test data', TEncoding.UTF8);
    TFile.WriteAllText(TPath.Combine(fn, 'more data.txt'), 'more test data', TEncoding.UTF8);
    TFile.WriteAllText(TPath.Combine(fn, 'largefile.txt'), StringOfChar('A', 1024*1024), TEncoding.UTF8);
    Test('fn ≔ "' + fn + '"', fn);
    Test('DirectoryExists(fn)', True);
    Test('DirectoryExists(fn + "\cats")', True);
    Test('DirectoryExists(fn + "\dogs")', True);
    Test('DirectoryExists(fn + "\unicorns")', True);
    Test('DirectoryExists(fn + "\horses")', False);
    Test('DirectoryExists(fn + "\rats")', False);
    Test('CreateDirectory(fn + "\rats")', success);
    Test('DirectoryExists(fn + "\rats")', True);
    Test('DirectoryList(fn)', strarr([TPath.Combine(fn, 'cats'), TPath.Combine(fn, 'dogs'), TPath.Combine(fn, 'rats'), TPath.Combine(fn, 'unicorns')]));
    Test('DeleteDirectory(fn + "\unicorns")', success);
    Test('DirectoryList(fn)', strarr([TPath.Combine(fn, 'cats'), TPath.Combine(fn, 'dogs'), TPath.Combine(fn, 'rats')]));
    Test('FileExists(fn + "\data.txt")', True);
    Test('FileExists(fn + "\more data.txt")', True);
    Test('FileExists(fn + "\largefile.txt")', True);
    Test('FileExists(fn + "\nargles.txt")', False);
    Test('FileSize(fn + "\data.txt")', 9 + 3);
    Test('FileSize(fn + "\more data.txt")', 14 + 3);
    Test('FileSize(fn + "\largefile.txt")', 1024*1024 + 3);
    Test('FileSize(fn + "\nargles.txt")', failure, 'File not found.');
    Test('FileList(fn)', strarr([TPath.Combine(fn, 'data.txt'), TPath.Combine(fn, 'largefile.txt'), TPath.Combine(fn, 'more data.txt')]));
    Test('DeleteFile(fn + "\data.txt")', success);
    Test('DeleteFile(fn + "\nargles.txt")', failure);
    Test('FileList(fn)', strarr([TPath.Combine(fn, 'largefile.txt'), TPath.Combine(fn, 'more data.txt')]));
    Test('FileExists(fn + "\data.txt")', False);
    Test('DeleteDirectory(fn)', failure, 'The specified directory is not empty');
    Test('fn ≔ "C:\Users\Luna Lovegood\Documents\The Quibbler\June 2015\Articles\More nargles found in southern England.html";', null);
    Test('FileName(fn)', 'More nargles found in southern England.html');
    Test('PrettyFileName(fn)', 'More nargles found in southern England');
    Test('FileExt(fn)', '.html');
    Test('FilePath(fn)', 'C:\Users\Luna Lovegood\Documents\The Quibbler\June 2015\Articles\');
    Test('delete(fn)', success);
    Test('DeleteFile( property("FrontEnd.ExeName") )', failure);
  finally
    if TDirectory.Exists(fn) then
      TDirectory.Delete(fn, True);
  end;





  //
  //
  //  CHAPTER 19
  //  Pathological expressions
  //
  //

  Bookmark: Chapter('Pathological expressions');

  Test('CustomSort(''(0, 1), (left, right) ↦ left − right)', failure,
    'The comparer was expected to return a real number, but an object of type custom function was returned.');

  Test('F ≔ ξ ↦ (n ↦ ξ⋅n); F(5)(7)', 35);
  Test('F ≔ δ ↦ (ξ ↦ (n ↦ ξ⋅n + δ)); F(3)(5)(7)', 38);
  Test('F ≔ ξ ↦ (ξ ↦ ξ⋅ξ); F(5)(7)', 49);

  Test('∑((n ↦ n^2; 7), n, 1, 10)', 70);
  Test('∑((n ↦ n^2)(3), n, 1, 10)', 90);
  Test('∑((n ↦ n^2)(n), n, 1, 10)', 385);
  Test('∑(∑((n ↦ n^2)(n), n, 1, 10), n, 1, 10)', 3850);
  Test('∑(∑(∑((n ↦ n^2)(n), n, 1, 10), n, 1, 10), n, 1, 10)', 38500);
  Test('(n ↦ ∑(∑(∑((n ↦ n^2)(n), n, 1, 10), n, 1, 10), n, 1, 10))(true)', 38500);
  Test('f ≔ n ↦ ∑(∑(∑((n ↦ n^2)(n), n, 1, 10), n, 1, 10), n, 1, 10); f(2)', 38500);

  Test('f ≔ n ↦ f(n); f(1)', failure, 'Too deep recursion.');
  Test('a ≔ n ↦ b(n); b ≔ n ↦ a(n); a(1)', failure, 'Too deep recursion.');
  Test('a ≔ n ↦ b(n); b ≔ n ↦ c(n); c ≔ n ↦ a(n); a(1)', failure, 'Too deep recursion.');
  Test('f ≔ x ↦ ∫(g(t), t, 0, x); g ≔ x ↦ ∫(f(t), t, 0, x); f(1)', failure, 'Too deep recursion.');
  Test('f ≔ x ↦ ∫(g(x), x, 0, x); g ≔ x ↦ ∫(f(x), x, 0, x); f(1)', failure, 'Too deep recursion.');

  Test(StringOfChar('+', 1000) + '1', EParseException, 'Expression is too deep.');
  Test(StringOfChar('+', 10000) + '1', EParseException, 'Expression is too deep.');

  Test(StringOfChar('(', 1000) + '1' + StringOfChar(')', 1000), 1);

  Test('+/⋅−+/*−+−+/−+/−⋅/−+//+⋅−/−/+−+⋅+⋅−+/−', ESyntaxException);
  Test('−−+−+−⋅/−+//+⋅−/−/+−+⋅+⋅−+/−', ESyntaxException);
  Test('/−+/−+/−⋅/−+//+⋅−/−/+−+⋅+⋅−+/−', ESyntaxException);
  Test('*−+−+/−+/−⋅/−+//+⋅−/−/+−+⋅+⋅−+/−', ESyntaxException);
  Test('⋅−+*−−/−/−−+//+⋅−/−/+−+⋅+⋅−+/−', ESyntaxException);
  Test('+/⋅−+/*−+−+/−+/−⋅/−+/−/+−+⋅+⋅−+/−', ESyntaxException);
  Test('+', ESyntaxException, 'Operand missing after operator + at column 1.');

  Test('x, x, x ≔ (x, x ≔ 5)', 5);
  Test('x', 5);

  Test('AtZero ≔ f ↦ f(0); AtZero(sin) ≈ 0 ∧ AtZero(cos) ≈ 1', True);
  Test('f ≔ x ↦ x(x); f(type)', 'kernel function');
  Test('f(f)', failure, 'Too deep recursion.');

  Test('f ≔ x ↦ (y ↦ y(y) + x); f(type)(type)', 'kernel functiontype');
  Test('(x ↦ (y ↦ y(y) + x))(type)(type)', 'kernel functiontype');
  Test('(x ↦ (y ↦ y(y)))(0)(type)', 'kernel function');
  Test('string((x ↦ (y ↦ y(y)))(0)(x ↦ x))', 'custom function');

  Test('(x ↦ (y ↦ y(y)))(0)(x ↦ x)(x ↦ x)(394)', 394);
  Test('(x ↦ (x ↦ x(x)))(0)(x ↦ x)(x ↦ x)(394)', 394);

  Test('(x ↦ (x ↦ x(x)))(x ↦ x)(x ↦ x)(x ↦ x)(394)', 394);

  Test('repeat(repeat(repeat(break(3)))); 2', 2);

  Test('∫(break(), x, 0, 1)', failure, 'Integrand must be a real-valued function.');
  Test('∑(break(), n, 0, 10)', ASO(_break));

  Eps; Test('[[−1, ''(−1, ''(−1, ''(1, −1)[1])[−[−[−1, 1, 1][0! + [[[0, 1][1]]!]]]!])[−1]][[−1, [1]][−[−1, 1][−1]]], 10^⌈π⌉][−[0!]] − '+'(−[1, 10^⌈e⌉][−1]⋅⌊e⌋ − [[10⋅[⌊π + e⌋ + [1, [digits(e)[−1][11], [1, 10, 1][−digits(π)[1][1]]][1]][−1]]]⋅⌊π⌋ − (−[1, ''(sqr, sqrt)[digits(1)[1]](10), 1][40] − '+'[1, [1, 10, 1][[1, [1, ''(sqr)[−1](10), 1][−1], 1][[1, digits(e)[−1][11]][−[1, digits(e)[1][1]][1]]]], 1][−[[0]!]!])])',
    12345.0);





  //
  //
  //  CHAPTER 20
  //  Benchmarks
  //
  //

  Chapter('Benchmarks');
  begin

    ASNum.ClearCaches;

    var c20bmtc1 := GetTickCount64;

    Test('for(n, 1, 40, eigenvalues(RandomMatrix(n)))', null);
    Test('eigenvectors(RandomMatrix(40));', null);
    Test('LU(RandomMatrix(150));', null);
    Test('det(IdentityMatrix(1000))', 1.0);
    Test('QR(IdentityMatrix(200));', null);
    Test('for(n, 1, 10, assert(A ≔ RandomMatrix(250); IsIdentity(A ⋅ inv(A), 1E−12)))', null);

    Test('n ≔ 1000; ComputePixmap(n, n, x, y, hsv(x⋅360/pred(n), 1, y/pred(n)));', null);
    Test('ComputePixmap(1000, 1000, x, y, rgb(x/999, y/999, 0.5));', null);
    Test('ComputePixmap(500, 500, x, y, hsv(x^2/100 + y/10 − sin((x−20⋅cos(y/25))/200)⋅x⋅y/50, abs(cos(x/500)), abs(cos(x/500))));', null);
    Test('CustomScanlineRotation(ExampleData("Sally"), y, round(100⋅cos(sin(y/20)⋅y^2/300^2)));', null);
    Test('ν ≔ 400; ω ≔ 2⋅π⋅ν; ComputeSound(sin(ω⋅sin(t)), t, 0, 10, 1);', null);

    Test('∫(FresnelC(t), t, 0, 1);', null);
    Test('∫(FresnelS(t), t, 0, 1);', null);
    Test('∫(erf(x), x, −2, 2);', null);
    Test('∑(sin(n)/cos(n), n, 1, 500000);', null);

    Test('sort(frequencies(compute(RandomInt(1000000)^2, n, 1, 1000000) @ digits @ first));', null);

    var c20bmtc2 := GetTickCount64;

    TSelfTestCUI.OutputText(FormatFloat('#.0', 100 * (c20bmtc2 - c20bmtc1) / 60000) + '%');

  end;




end;

end.
